// Code generated by gocc; DO NOT EDIT.

package parser

type (
	actionTable [numStates]actionRow
	actionRow   struct {
		canRecover bool
		actions    [numSymbols]action
	}
)

var actionTab = actionTable{
	actionRow{ // S0
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(2),  // $, reduce: StatementBlock
			nil,        // empty
			shift(6),   // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			shift(30),  // unset
			shift(33),  // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			shift(34),  // print
			nil,        // ,
			shift(35),  // printn
			shift(36),  // eprint
			shift(37),  // eprintn
			shift(38),  // dump
			shift(39),  // edump
			shift(40),  // tee
			shift(42),  // emitf
			shift(43),  // emit1
			shift(44),  // emit
			shift(45),  // (
			nil,        // )
			shift(46),  // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			shift(71),  // arr
			shift(72),  // bool
			shift(73),  // float
			shift(74),  // int
			shift(75),  // map
			shift(76),  // num
			shift(77),  // str
			shift(78),  // var
			shift(79),  // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			shift(141), // call
			shift(150), // begin
			shift(151), // end
			shift(154), // if
			nil,        // elif
			nil,        // else
			shift(155), // while
			shift(156), // do
			shift(161), // for
			nil,        // in
			shift(162), // break
			shift(163), // continue
			shift(164), // func
			shift(165), // subr
			shift(166), // return
		},
	},
	actionRow{ // S1
		canRecover: false,
		actions: [numSymbols]action{
			nil,          // INVALID
			accept(true), // $
			nil,          // empty
			nil,          // ;
			nil,          // {
			nil,          // }
			nil,          // =
			nil,          // unset
			nil,          // filter
			nil,          // >
			nil,          // >>
			nil,          // |
			nil,          // stdout
			nil,          // stderr
			nil,          // print
			nil,          // ,
			nil,          // printn
			nil,          // eprint
			nil,          // eprintn
			nil,          // dump
			nil,          // edump
			nil,          // tee
			nil,          // emitf
			nil,          // emit1
			nil,          // emit
			nil,          // (
			nil,          // )
			nil,          // emitp
			nil,          // field_name
			nil,          // $[
			nil,          // ]
			nil,          // braced_field_name
			nil,          // $[[
			nil,          // $[[[
			nil,          // full_srec
			nil,          // oosvar_name
			nil,          // @[
			nil,          // braced_oosvar_name
			nil,          // full_oosvar
			nil,          // all
			nil,          // non_sigil_name
			nil,          // arr
			nil,          // bool
			nil,          // float
			nil,          // int
			nil,          // map
			nil,          // num
			nil,          // str
			nil,          // var
			nil,          // funct
			nil,          // ||=
			nil,          // ^^=
			nil,          // &&=
			nil,          // ??=
			nil,          // ???=
			nil,          // |=
			nil,          // &=
			nil,          // ^=
			nil,          // <<=
			nil,          // >>=
			nil,          // >>>=
			nil,          // +=
			nil,          // .=
			nil,          // -=
			nil,          // *=
			nil,          // /=
			nil,          // //=
			nil,          // %=
			nil,          // **=
			nil,          // ?
			nil,          // :
			nil,          // ||
			nil,          // ^^
			nil,          // &&
			nil,          // ??
			nil,          // ???
			nil,          // =~
			nil,          // !=~
			nil,          // ==
			nil,          // !=
			nil,          // <=>
			nil,          // >=
			nil,          // <
			nil,          // <=
			nil,          // ^
			nil,          // &
			nil,          // <<
			nil,          // >>>
			nil,          // +
			nil,          // -
			nil,          // .+
			nil,          // .-
			nil,          // *
			nil,          // /
			nil,          // //
			nil,          // %
			nil,          // .*
			nil,          // ./
			nil,          // .//
			nil,          // .
			nil,          // !
			nil,          // ~
			nil,          // **
			nil,          // string_literal
			nil,          // regex_case_insensitive
			nil,          // int_literal
			nil,          // float_literal
			nil,          // boolean_literal
			nil,          // null_literal
			nil,          // inf_literal
			nil,          // nan_literal
			nil,          // const_M_PI
			nil,          // const_M_E
			nil,          // panic
			nil,          // [
			nil,          // ctx_IPS
			nil,          // ctx_IFS
			nil,          // ctx_IRS
			nil,          // ctx_OPS
			nil,          // ctx_OFS
			nil,          // ctx_ORS
			nil,          // ctx_FLATSEP
			nil,          // ctx_NF
			nil,          // ctx_NR
			nil,          // ctx_FNR
			nil,          // ctx_FILENAME
			nil,          // ctx_FILENUM
			nil,          // env
			nil,          // [[
			nil,          // [[[
			nil,          // call
			nil,          // begin
			nil,          // end
			nil,          // if
			nil,          // elif
			nil,          // else
			nil,          // while
			nil,          // do
			nil,          // for
			nil,          // in
			nil,          // break
			nil,          // continue
			nil,          // func
			nil,          // subr
			nil,          // return
		},
	},
	actionRow{ // S2
		canRecover: false,
		actions: [numSymbols]action{
			nil,       // INVALID
			reduce(1), // $, reduce: Root
			nil,       // empty
			nil,       // ;
			nil,       // {
			nil,       // }
			nil,       // =
			nil,       // unset
			nil,       // filter
			nil,       // >
			nil,       // >>
			nil,       // |
			nil,       // stdout
			nil,       // stderr
			nil,       // print
			nil,       // ,
			nil,       // printn
			nil,       // eprint
			nil,       // eprintn
			nil,       // dump
			nil,       // edump
			nil,       // tee
			nil,       // emitf
			nil,       // emit1
			nil,       // emit
			nil,       // (
			nil,       // )
			nil,       // emitp
			nil,       // field_name
			nil,       // $[
			nil,       // ]
			nil,       // braced_field_name
			nil,       // $[[
			nil,       // $[[[
			nil,       // full_srec
			nil,       // oosvar_name
			nil,       // @[
			nil,       // braced_oosvar_name
			nil,       // full_oosvar
			nil,       // all
			nil,       // non_sigil_name
			nil,       // arr
			nil,       // bool
			nil,       // float
			nil,       // int
			nil,       // map
			nil,       // num
			nil,       // str
			nil,       // var
			nil,       // funct
			nil,       // ||=
			nil,       // ^^=
			nil,       // &&=
			nil,       // ??=
			nil,       // ???=
			nil,       // |=
			nil,       // &=
			nil,       // ^=
			nil,       // <<=
			nil,       // >>=
			nil,       // >>>=
			nil,       // +=
			nil,       // .=
			nil,       // -=
			nil,       // *=
			nil,       // /=
			nil,       // //=
			nil,       // %=
			nil,       // **=
			nil,       // ?
			nil,       // :
			nil,       // ||
			nil,       // ^^
			nil,       // &&
			nil,       // ??
			nil,       // ???
			nil,       // =~
			nil,       // !=~
			nil,       // ==
			nil,       // !=
			nil,       // <=>
			nil,       // >=
			nil,       // <
			nil,       // <=
			nil,       // ^
			nil,       // &
			nil,       // <<
			nil,       // >>>
			nil,       // +
			nil,       // -
			nil,       // .+
			nil,       // .-
			nil,       // *
			nil,       // /
			nil,       // //
			nil,       // %
			nil,       // .*
			nil,       // ./
			nil,       // .//
			nil,       // .
			nil,       // !
			nil,       // ~
			nil,       // **
			nil,       // string_literal
			nil,       // regex_case_insensitive
			nil,       // int_literal
			nil,       // float_literal
			nil,       // boolean_literal
			nil,       // null_literal
			nil,       // inf_literal
			nil,       // nan_literal
			nil,       // const_M_PI
			nil,       // const_M_E
			nil,       // panic
			nil,       // [
			nil,       // ctx_IPS
			nil,       // ctx_IFS
			nil,       // ctx_IRS
			nil,       // ctx_OPS
			nil,       // ctx_OFS
			nil,       // ctx_ORS
			nil,       // ctx_FLATSEP
			nil,       // ctx_NF
			nil,       // ctx_NR
			nil,       // ctx_FNR
			nil,       // ctx_FILENAME
			nil,       // ctx_FILENUM
			nil,       // env
			nil,       // [[
			nil,       // [[[
			nil,       // call
			nil,       // begin
			nil,       // end
			nil,       // if
			nil,       // elif
			nil,       // else
			nil,       // while
			nil,       // do
			nil,       // for
			nil,       // in
			nil,       // break
			nil,       // continue
			nil,       // func
			nil,       // subr
			nil,       // return
		},
	},
	actionRow{ // S3
		canRecover: false,
		actions: [numSymbols]action{
			nil,       // INVALID
			reduce(3), // $, reduce: StatementBlock
			nil,       // empty
			nil,       // ;
			nil,       // {
			nil,       // }
			nil,       // =
			nil,       // unset
			nil,       // filter
			nil,       // >
			nil,       // >>
			nil,       // |
			nil,       // stdout
			nil,       // stderr
			nil,       // print
			nil,       // ,
			nil,       // printn
			nil,       // eprint
			nil,       // eprintn
			nil,       // dump
			nil,       // edump
			nil,       // tee
			nil,       // emitf
			nil,       // emit1
			nil,       // emit
			nil,       // (
			nil,       // )
			nil,       // emitp
			nil,       // field_name
			nil,       // $[
			nil,       // ]
			nil,       // braced_field_name
			nil,       // $[[
			nil,       // $[[[
			nil,       // full_srec
			nil,       // oosvar_name
			nil,       // @[
			nil,       // braced_oosvar_name
			nil,       // full_oosvar
			nil,       // all
			nil,       // non_sigil_name
			nil,       // arr
			nil,       // bool
			nil,       // float
			nil,       // int
			nil,       // map
			nil,       // num
			nil,       // str
			nil,       // var
			nil,       // funct
			nil,       // ||=
			nil,       // ^^=
			nil,       // &&=
			nil,       // ??=
			nil,       // ???=
			nil,       // |=
			nil,       // &=
			nil,       // ^=
			nil,       // <<=
			nil,       // >>=
			nil,       // >>>=
			nil,       // +=
			nil,       // .=
			nil,       // -=
			nil,       // *=
			nil,       // /=
			nil,       // //=
			nil,       // %=
			nil,       // **=
			nil,       // ?
			nil,       // :
			nil,       // ||
			nil,       // ^^
			nil,       // &&
			nil,       // ??
			nil,       // ???
			nil,       // =~
			nil,       // !=~
			nil,       // ==
			nil,       // !=
			nil,       // <=>
			nil,       // >=
			nil,       // <
			nil,       // <=
			nil,       // ^
			nil,       // &
			nil,       // <<
			nil,       // >>>
			nil,       // +
			nil,       // -
			nil,       // .+
			nil,       // .-
			nil,       // *
			nil,       // /
			nil,       // //
			nil,       // %
			nil,       // .*
			nil,       // ./
			nil,       // .//
			nil,       // .
			nil,       // !
			nil,       // ~
			nil,       // **
			nil,       // string_literal
			nil,       // regex_case_insensitive
			nil,       // int_literal
			nil,       // float_literal
			nil,       // boolean_literal
			nil,       // null_literal
			nil,       // inf_literal
			nil,       // nan_literal
			nil,       // const_M_PI
			nil,       // const_M_E
			nil,       // panic
			nil,       // [
			nil,       // ctx_IPS
			nil,       // ctx_IFS
			nil,       // ctx_IRS
			nil,       // ctx_OPS
			nil,       // ctx_OFS
			nil,       // ctx_ORS
			nil,       // ctx_FLATSEP
			nil,       // ctx_NF
			nil,       // ctx_NR
			nil,       // ctx_FNR
			nil,       // ctx_FILENAME
			nil,       // ctx_FILENUM
			nil,       // env
			nil,       // [[
			nil,       // [[[
			nil,       // call
			nil,       // begin
			nil,       // end
			nil,       // if
			nil,       // elif
			nil,       // else
			nil,       // while
			nil,       // do
			nil,       // for
			nil,       // in
			nil,       // break
			nil,       // continue
			nil,       // func
			nil,       // subr
			nil,       // return
		},
	},
	actionRow{ // S4
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(4),  // $, reduce: NonEmptyStatementBlock
			nil,        // empty
			shift(167), // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S5
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(5),  // $, reduce: NonEmptyStatementBlock
			nil,        // empty
			shift(170), // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			shift(30),  // unset
			shift(33),  // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			shift(34),  // print
			nil,        // ,
			shift(35),  // printn
			shift(36),  // eprint
			shift(37),  // eprintn
			shift(38),  // dump
			shift(39),  // edump
			shift(40),  // tee
			shift(42),  // emitf
			shift(43),  // emit1
			shift(44),  // emit
			shift(45),  // (
			nil,        // )
			shift(46),  // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			shift(71),  // arr
			shift(72),  // bool
			shift(73),  // float
			shift(74),  // int
			shift(75),  // map
			shift(76),  // num
			shift(77),  // str
			shift(78),  // var
			shift(79),  // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			shift(141), // call
			shift(150), // begin
			shift(151), // end
			shift(154), // if
			nil,        // elif
			nil,        // else
			shift(155), // while
			shift(156), // do
			shift(161), // for
			nil,        // in
			shift(162), // break
			shift(163), // continue
			shift(164), // func
			shift(165), // subr
			shift(166), // return
		},
	},
	actionRow{ // S6
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(2),  // $, reduce: StatementBlock
			nil,        // empty
			shift(6),   // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			shift(30),  // unset
			shift(33),  // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			shift(34),  // print
			nil,        // ,
			shift(35),  // printn
			shift(36),  // eprint
			shift(37),  // eprintn
			shift(38),  // dump
			shift(39),  // edump
			shift(40),  // tee
			shift(42),  // emitf
			shift(43),  // emit1
			shift(44),  // emit
			shift(45),  // (
			nil,        // )
			shift(46),  // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			shift(71),  // arr
			shift(72),  // bool
			shift(73),  // float
			shift(74),  // int
			shift(75),  // map
			shift(76),  // num
			shift(77),  // str
			shift(78),  // var
			shift(79),  // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			shift(141), // call
			shift(150), // begin
			shift(151), // end
			shift(154), // if
			nil,        // elif
			nil,        // else
			shift(155), // while
			shift(156), // do
			shift(161), // for
			nil,        // in
			shift(162), // break
			shift(163), // continue
			shift(164), // func
			shift(165), // subr
			shift(166), // return
		},
	},
	actionRow{ // S7
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			shift(173), // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S8
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(14), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(14), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S9
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(15), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(15), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S10
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(16), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(16), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S11
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(17), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(17), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S12
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(18), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(18), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S13
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(19), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(19), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S14
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(20), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(20), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S15
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(21), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(21), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S16
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(22), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(22), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S17
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(23), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(23), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S18
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(24), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(24), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S19
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(25), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(25), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S20
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(26), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(26), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S21
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(27), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(27), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S22
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(28), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(28), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S23
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(29), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(29), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S24
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(30), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(30), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S25
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(31), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(31), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S26
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(32), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(32), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S27
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(33), // $, reduce: BracelessStatement
			nil,        // empty
			reduce(33), // ;, reduce: BracelessStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S28
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			shift(268), // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			shift(269), // ||=
			shift(270), // ^^=
			shift(271), // &&=
			shift(272), // ??=
			shift(273), // ???=
			shift(274), // |=
			shift(275), // &=
			shift(276), // ^=
			shift(277), // <<=
			shift(278), // >>=
			shift(279), // >>>=
			shift(280), // +=
			shift(281), // .=
			shift(282), // -=
			shift(283), // *=
			shift(284), // /=
			shift(285), // //=
			shift(286), // %=
			shift(287), // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S29
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(38), // $, reduce: BareBoolean
			nil,        // empty
			reduce(38), // ;, reduce: BareBoolean
			shift(289), // {
			nil,        // }
			reduce(36), // =, reduce: Lvalue
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			reduce(36), // ||=, reduce: Lvalue
			reduce(36), // ^^=, reduce: Lvalue
			reduce(36), // &&=, reduce: Lvalue
			reduce(36), // ??=, reduce: Lvalue
			reduce(36), // ???=, reduce: Lvalue
			reduce(36), // |=, reduce: Lvalue
			reduce(36), // &=, reduce: Lvalue
			reduce(36), // ^=, reduce: Lvalue
			reduce(36), // <<=, reduce: Lvalue
			reduce(36), // >>=, reduce: Lvalue
			reduce(36), // >>>=, reduce: Lvalue
			reduce(36), // +=, reduce: Lvalue
			reduce(36), // .=, reduce: Lvalue
			reduce(36), // -=, reduce: Lvalue
			reduce(36), // *=, reduce: Lvalue
			reduce(36), // /=, reduce: Lvalue
			reduce(36), // //=, reduce: Lvalue
			reduce(36), // %=, reduce: Lvalue
			reduce(36), // **=, reduce: Lvalue
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S30
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S31
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			shift(383), // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S32
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(214), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(214), // ;, reduce: MlrvalOrFunction
			reduce(214), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(214), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(214), // ||=, reduce: MlrvalOrFunction
			reduce(214), // ^^=, reduce: MlrvalOrFunction
			reduce(214), // &&=, reduce: MlrvalOrFunction
			reduce(214), // ??=, reduce: MlrvalOrFunction
			reduce(214), // ???=, reduce: MlrvalOrFunction
			reduce(214), // |=, reduce: MlrvalOrFunction
			reduce(214), // &=, reduce: MlrvalOrFunction
			reduce(214), // ^=, reduce: MlrvalOrFunction
			reduce(214), // <<=, reduce: MlrvalOrFunction
			reduce(214), // >>=, reduce: MlrvalOrFunction
			reduce(214), // >>>=, reduce: MlrvalOrFunction
			reduce(214), // +=, reduce: MlrvalOrFunction
			reduce(214), // .=, reduce: MlrvalOrFunction
			reduce(214), // -=, reduce: MlrvalOrFunction
			reduce(214), // *=, reduce: MlrvalOrFunction
			reduce(214), // /=, reduce: MlrvalOrFunction
			reduce(214), // //=, reduce: MlrvalOrFunction
			reduce(214), // %=, reduce: MlrvalOrFunction
			reduce(214), // **=, reduce: MlrvalOrFunction
			reduce(214), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S33
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S34
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(46), // $, reduce: PrintStatement
			nil,        // empty
			reduce(46), // ;, reduce: PrintStatement
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			shift(477), // >
			shift(478), // >>
			shift(479), // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S35
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(50), // $, reduce: PrintnStatement
			nil,        // empty
			reduce(50), // ;, reduce: PrintnStatement
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			shift(477), // >
			shift(478), // >>
			shift(479), // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S36
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(54), // $, reduce: EprintStatement
			nil,        // empty
			reduce(54), // ;, reduce: EprintStatement
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S37
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(56), // $, reduce: EprintnStatement
			nil,        // empty
			reduce(56), // ;, reduce: EprintnStatement
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S38
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(58), // $, reduce: DumpStatement
			nil,        // empty
			reduce(58), // ;, reduce: DumpStatement
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			shift(477), // >
			shift(478), // >>
			shift(479), // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S39
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(62), // $, reduce: EdumpStatement
			nil,        // empty
			reduce(62), // ;, reduce: EdumpStatement
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S40
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			shift(488), // >
			shift(489), // >>
			shift(490), // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S41
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(211), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(211), // ;, reduce: MlrvalOrFunction
			reduce(211), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(211), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(211), // >, reduce: MlrvalOrFunction
			reduce(211), // >>, reduce: MlrvalOrFunction
			reduce(211), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(211), // ||=, reduce: MlrvalOrFunction
			reduce(211), // ^^=, reduce: MlrvalOrFunction
			reduce(211), // &&=, reduce: MlrvalOrFunction
			reduce(211), // ??=, reduce: MlrvalOrFunction
			reduce(211), // ???=, reduce: MlrvalOrFunction
			reduce(211), // |=, reduce: MlrvalOrFunction
			reduce(211), // &=, reduce: MlrvalOrFunction
			reduce(211), // ^=, reduce: MlrvalOrFunction
			reduce(211), // <<=, reduce: MlrvalOrFunction
			reduce(211), // >>=, reduce: MlrvalOrFunction
			reduce(211), // >>>=, reduce: MlrvalOrFunction
			reduce(211), // +=, reduce: MlrvalOrFunction
			reduce(211), // .=, reduce: MlrvalOrFunction
			reduce(211), // -=, reduce: MlrvalOrFunction
			reduce(211), // *=, reduce: MlrvalOrFunction
			reduce(211), // /=, reduce: MlrvalOrFunction
			reduce(211), // //=, reduce: MlrvalOrFunction
			reduce(211), // %=, reduce: MlrvalOrFunction
			reduce(211), // **=, reduce: MlrvalOrFunction
			reduce(211), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(211), // ||, reduce: MlrvalOrFunction
			reduce(211), // ^^, reduce: MlrvalOrFunction
			reduce(211), // &&, reduce: MlrvalOrFunction
			reduce(211), // ??, reduce: MlrvalOrFunction
			reduce(211), // ???, reduce: MlrvalOrFunction
			reduce(211), // =~, reduce: MlrvalOrFunction
			reduce(211), // !=~, reduce: MlrvalOrFunction
			reduce(211), // ==, reduce: MlrvalOrFunction
			reduce(211), // !=, reduce: MlrvalOrFunction
			reduce(211), // <=>, reduce: MlrvalOrFunction
			reduce(211), // >=, reduce: MlrvalOrFunction
			reduce(211), // <, reduce: MlrvalOrFunction
			reduce(211), // <=, reduce: MlrvalOrFunction
			reduce(211), // ^, reduce: MlrvalOrFunction
			reduce(211), // &, reduce: MlrvalOrFunction
			reduce(211), // <<, reduce: MlrvalOrFunction
			reduce(211), // >>>, reduce: MlrvalOrFunction
			reduce(211), // +, reduce: MlrvalOrFunction
			reduce(211), // -, reduce: MlrvalOrFunction
			reduce(211), // .+, reduce: MlrvalOrFunction
			reduce(211), // .-, reduce: MlrvalOrFunction
			reduce(211), // *, reduce: MlrvalOrFunction
			reduce(211), // /, reduce: MlrvalOrFunction
			reduce(211), // //, reduce: MlrvalOrFunction
			reduce(211), // %, reduce: MlrvalOrFunction
			reduce(211), // .*, reduce: MlrvalOrFunction
			reduce(211), // ./, reduce: MlrvalOrFunction
			reduce(211), // .//, reduce: MlrvalOrFunction
			reduce(211), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(211), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(211), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(211), // [[, reduce: MlrvalOrFunction
			reduce(211), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S42
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(491), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			shift(488), // >
			shift(489), // >>
			shift(490), // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			shift(505), // field_name
			shift(506), // $[
			nil,        // ]
			shift(507), // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			shift(508), // full_srec
			shift(509), // oosvar_name
			shift(510), // @[
			shift(511), // braced_oosvar_name
			shift(512), // full_oosvar
			shift(513), // all
			shift(514), // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S43
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S44
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(491), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			shift(488), // >
			shift(489), // >>
			shift(490), // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(518), // (
			nil,        // )
			nil,        // emitp
			shift(505), // field_name
			shift(506), // $[
			nil,        // ]
			shift(507), // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			shift(508), // full_srec
			shift(509), // oosvar_name
			shift(510), // @[
			shift(511), // braced_oosvar_name
			shift(512), // full_oosvar
			shift(513), // all
			shift(514), // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S45
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S46
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(491), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			shift(488), // >
			shift(489), // >>
			shift(490), // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(613), // (
			nil,        // )
			nil,        // emitp
			shift(505), // field_name
			shift(506), // $[
			nil,        // ]
			shift(507), // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			shift(508), // full_srec
			shift(509), // oosvar_name
			shift(510), // @[
			shift(511), // braced_oosvar_name
			shift(512), // full_oosvar
			shift(513), // all
			shift(514), // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S47
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(110), // $, reduce: OosvarValue
			nil,         // empty
			reduce(110), // ;, reduce: OosvarValue
			reduce(110), // {, reduce: OosvarValue
			nil,         // }
			reduce(110), // =, reduce: OosvarValue
			nil,         // unset
			nil,         // filter
			reduce(110), // >, reduce: OosvarValue
			reduce(110), // >>, reduce: OosvarValue
			reduce(110), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(110), // ||=, reduce: OosvarValue
			reduce(110), // ^^=, reduce: OosvarValue
			reduce(110), // &&=, reduce: OosvarValue
			reduce(110), // ??=, reduce: OosvarValue
			reduce(110), // ???=, reduce: OosvarValue
			reduce(110), // |=, reduce: OosvarValue
			reduce(110), // &=, reduce: OosvarValue
			reduce(110), // ^=, reduce: OosvarValue
			reduce(110), // <<=, reduce: OosvarValue
			reduce(110), // >>=, reduce: OosvarValue
			reduce(110), // >>>=, reduce: OosvarValue
			reduce(110), // +=, reduce: OosvarValue
			reduce(110), // .=, reduce: OosvarValue
			reduce(110), // -=, reduce: OosvarValue
			reduce(110), // *=, reduce: OosvarValue
			reduce(110), // /=, reduce: OosvarValue
			reduce(110), // //=, reduce: OosvarValue
			reduce(110), // %=, reduce: OosvarValue
			reduce(110), // **=, reduce: OosvarValue
			reduce(110), // ?, reduce: OosvarValue
			nil,         // :
			reduce(110), // ||, reduce: OosvarValue
			reduce(110), // ^^, reduce: OosvarValue
			reduce(110), // &&, reduce: OosvarValue
			reduce(110), // ??, reduce: OosvarValue
			reduce(110), // ???, reduce: OosvarValue
			reduce(110), // =~, reduce: OosvarValue
			reduce(110), // !=~, reduce: OosvarValue
			reduce(110), // ==, reduce: OosvarValue
			reduce(110), // !=, reduce: OosvarValue
			reduce(110), // <=>, reduce: OosvarValue
			reduce(110), // >=, reduce: OosvarValue
			reduce(110), // <, reduce: OosvarValue
			reduce(110), // <=, reduce: OosvarValue
			reduce(110), // ^, reduce: OosvarValue
			reduce(110), // &, reduce: OosvarValue
			reduce(110), // <<, reduce: OosvarValue
			reduce(110), // >>>, reduce: OosvarValue
			reduce(110), // +, reduce: OosvarValue
			reduce(110), // -, reduce: OosvarValue
			reduce(110), // .+, reduce: OosvarValue
			reduce(110), // .-, reduce: OosvarValue
			reduce(110), // *, reduce: OosvarValue
			reduce(110), // /, reduce: OosvarValue
			reduce(110), // //, reduce: OosvarValue
			reduce(110), // %, reduce: OosvarValue
			reduce(110), // .*, reduce: OosvarValue
			reduce(110), // ./, reduce: OosvarValue
			reduce(110), // .//, reduce: OosvarValue
			reduce(110), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(110), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(110), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(110), // [[, reduce: OosvarValue
			reduce(110), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S48
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(112), // $, reduce: OosvarValue
			nil,         // empty
			reduce(112), // ;, reduce: OosvarValue
			reduce(112), // {, reduce: OosvarValue
			nil,         // }
			reduce(112), // =, reduce: OosvarValue
			nil,         // unset
			nil,         // filter
			reduce(112), // >, reduce: OosvarValue
			reduce(112), // >>, reduce: OosvarValue
			reduce(112), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(112), // ||=, reduce: OosvarValue
			reduce(112), // ^^=, reduce: OosvarValue
			reduce(112), // &&=, reduce: OosvarValue
			reduce(112), // ??=, reduce: OosvarValue
			reduce(112), // ???=, reduce: OosvarValue
			reduce(112), // |=, reduce: OosvarValue
			reduce(112), // &=, reduce: OosvarValue
			reduce(112), // ^=, reduce: OosvarValue
			reduce(112), // <<=, reduce: OosvarValue
			reduce(112), // >>=, reduce: OosvarValue
			reduce(112), // >>>=, reduce: OosvarValue
			reduce(112), // +=, reduce: OosvarValue
			reduce(112), // .=, reduce: OosvarValue
			reduce(112), // -=, reduce: OosvarValue
			reduce(112), // *=, reduce: OosvarValue
			reduce(112), // /=, reduce: OosvarValue
			reduce(112), // //=, reduce: OosvarValue
			reduce(112), // %=, reduce: OosvarValue
			reduce(112), // **=, reduce: OosvarValue
			reduce(112), // ?, reduce: OosvarValue
			nil,         // :
			reduce(112), // ||, reduce: OosvarValue
			reduce(112), // ^^, reduce: OosvarValue
			reduce(112), // &&, reduce: OosvarValue
			reduce(112), // ??, reduce: OosvarValue
			reduce(112), // ???, reduce: OosvarValue
			reduce(112), // =~, reduce: OosvarValue
			reduce(112), // !=~, reduce: OosvarValue
			reduce(112), // ==, reduce: OosvarValue
			reduce(112), // !=, reduce: OosvarValue
			reduce(112), // <=>, reduce: OosvarValue
			reduce(112), // >=, reduce: OosvarValue
			reduce(112), // <, reduce: OosvarValue
			reduce(112), // <=, reduce: OosvarValue
			reduce(112), // ^, reduce: OosvarValue
			reduce(112), // &, reduce: OosvarValue
			reduce(112), // <<, reduce: OosvarValue
			reduce(112), // >>>, reduce: OosvarValue
			reduce(112), // +, reduce: OosvarValue
			reduce(112), // -, reduce: OosvarValue
			reduce(112), // .+, reduce: OosvarValue
			reduce(112), // .-, reduce: OosvarValue
			reduce(112), // *, reduce: OosvarValue
			reduce(112), // /, reduce: OosvarValue
			reduce(112), // //, reduce: OosvarValue
			reduce(112), // %, reduce: OosvarValue
			reduce(112), // .*, reduce: OosvarValue
			reduce(112), // ./, reduce: OosvarValue
			reduce(112), // .//, reduce: OosvarValue
			reduce(112), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(112), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(112), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(112), // [[, reduce: OosvarValue
			reduce(112), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S49
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(111), // $, reduce: OosvarValue
			nil,         // empty
			reduce(111), // ;, reduce: OosvarValue
			reduce(111), // {, reduce: OosvarValue
			nil,         // }
			reduce(111), // =, reduce: OosvarValue
			nil,         // unset
			nil,         // filter
			reduce(111), // >, reduce: OosvarValue
			reduce(111), // >>, reduce: OosvarValue
			reduce(111), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(111), // ||=, reduce: OosvarValue
			reduce(111), // ^^=, reduce: OosvarValue
			reduce(111), // &&=, reduce: OosvarValue
			reduce(111), // ??=, reduce: OosvarValue
			reduce(111), // ???=, reduce: OosvarValue
			reduce(111), // |=, reduce: OosvarValue
			reduce(111), // &=, reduce: OosvarValue
			reduce(111), // ^=, reduce: OosvarValue
			reduce(111), // <<=, reduce: OosvarValue
			reduce(111), // >>=, reduce: OosvarValue
			reduce(111), // >>>=, reduce: OosvarValue
			reduce(111), // +=, reduce: OosvarValue
			reduce(111), // .=, reduce: OosvarValue
			reduce(111), // -=, reduce: OosvarValue
			reduce(111), // *=, reduce: OosvarValue
			reduce(111), // /=, reduce: OosvarValue
			reduce(111), // //=, reduce: OosvarValue
			reduce(111), // %=, reduce: OosvarValue
			reduce(111), // **=, reduce: OosvarValue
			reduce(111), // ?, reduce: OosvarValue
			nil,         // :
			reduce(111), // ||, reduce: OosvarValue
			reduce(111), // ^^, reduce: OosvarValue
			reduce(111), // &&, reduce: OosvarValue
			reduce(111), // ??, reduce: OosvarValue
			reduce(111), // ???, reduce: OosvarValue
			reduce(111), // =~, reduce: OosvarValue
			reduce(111), // !=~, reduce: OosvarValue
			reduce(111), // ==, reduce: OosvarValue
			reduce(111), // !=, reduce: OosvarValue
			reduce(111), // <=>, reduce: OosvarValue
			reduce(111), // >=, reduce: OosvarValue
			reduce(111), // <, reduce: OosvarValue
			reduce(111), // <=, reduce: OosvarValue
			reduce(111), // ^, reduce: OosvarValue
			reduce(111), // &, reduce: OosvarValue
			reduce(111), // <<, reduce: OosvarValue
			reduce(111), // >>>, reduce: OosvarValue
			reduce(111), // +, reduce: OosvarValue
			reduce(111), // -, reduce: OosvarValue
			reduce(111), // .+, reduce: OosvarValue
			reduce(111), // .-, reduce: OosvarValue
			reduce(111), // *, reduce: OosvarValue
			reduce(111), // /, reduce: OosvarValue
			reduce(111), // //, reduce: OosvarValue
			reduce(111), // %, reduce: OosvarValue
			reduce(111), // .*, reduce: OosvarValue
			reduce(111), // ./, reduce: OosvarValue
			reduce(111), // .//, reduce: OosvarValue
			reduce(111), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(111), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(111), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(111), // [[, reduce: OosvarValue
			reduce(111), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S50
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(99), // $, reduce: FieldValue
			nil,        // empty
			reduce(99), // ;, reduce: FieldValue
			reduce(99), // {, reduce: FieldValue
			nil,        // }
			reduce(99), // =, reduce: FieldValue
			nil,        // unset
			nil,        // filter
			reduce(99), // >, reduce: FieldValue
			reduce(99), // >>, reduce: FieldValue
			reduce(99), // |, reduce: FieldValue
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			reduce(99), // ||=, reduce: FieldValue
			reduce(99), // ^^=, reduce: FieldValue
			reduce(99), // &&=, reduce: FieldValue
			reduce(99), // ??=, reduce: FieldValue
			reduce(99), // ???=, reduce: FieldValue
			reduce(99), // |=, reduce: FieldValue
			reduce(99), // &=, reduce: FieldValue
			reduce(99), // ^=, reduce: FieldValue
			reduce(99), // <<=, reduce: FieldValue
			reduce(99), // >>=, reduce: FieldValue
			reduce(99), // >>>=, reduce: FieldValue
			reduce(99), // +=, reduce: FieldValue
			reduce(99), // .=, reduce: FieldValue
			reduce(99), // -=, reduce: FieldValue
			reduce(99), // *=, reduce: FieldValue
			reduce(99), // /=, reduce: FieldValue
			reduce(99), // //=, reduce: FieldValue
			reduce(99), // %=, reduce: FieldValue
			reduce(99), // **=, reduce: FieldValue
			reduce(99), // ?, reduce: FieldValue
			nil,        // :
			reduce(99), // ||, reduce: FieldValue
			reduce(99), // ^^, reduce: FieldValue
			reduce(99), // &&, reduce: FieldValue
			reduce(99), // ??, reduce: FieldValue
			reduce(99), // ???, reduce: FieldValue
			reduce(99), // =~, reduce: FieldValue
			reduce(99), // !=~, reduce: FieldValue
			reduce(99), // ==, reduce: FieldValue
			reduce(99), // !=, reduce: FieldValue
			reduce(99), // <=>, reduce: FieldValue
			reduce(99), // >=, reduce: FieldValue
			reduce(99), // <, reduce: FieldValue
			reduce(99), // <=, reduce: FieldValue
			reduce(99), // ^, reduce: FieldValue
			reduce(99), // &, reduce: FieldValue
			reduce(99), // <<, reduce: FieldValue
			reduce(99), // >>>, reduce: FieldValue
			reduce(99), // +, reduce: FieldValue
			reduce(99), // -, reduce: FieldValue
			reduce(99), // .+, reduce: FieldValue
			reduce(99), // .-, reduce: FieldValue
			reduce(99), // *, reduce: FieldValue
			reduce(99), // /, reduce: FieldValue
			reduce(99), // //, reduce: FieldValue
			reduce(99), // %, reduce: FieldValue
			reduce(99), // .*, reduce: FieldValue
			reduce(99), // ./, reduce: FieldValue
			reduce(99), // .//, reduce: FieldValue
			reduce(99), // ., reduce: FieldValue
			nil,        // !
			nil,        // ~
			reduce(99), // **, reduce: FieldValue
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(99), // [, reduce: FieldValue
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(99), // [[, reduce: FieldValue
			reduce(99), // [[[, reduce: FieldValue
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S51
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(101), // $, reduce: FieldValue
			nil,         // empty
			reduce(101), // ;, reduce: FieldValue
			reduce(101), // {, reduce: FieldValue
			nil,         // }
			reduce(101), // =, reduce: FieldValue
			nil,         // unset
			nil,         // filter
			reduce(101), // >, reduce: FieldValue
			reduce(101), // >>, reduce: FieldValue
			reduce(101), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(101), // ||=, reduce: FieldValue
			reduce(101), // ^^=, reduce: FieldValue
			reduce(101), // &&=, reduce: FieldValue
			reduce(101), // ??=, reduce: FieldValue
			reduce(101), // ???=, reduce: FieldValue
			reduce(101), // |=, reduce: FieldValue
			reduce(101), // &=, reduce: FieldValue
			reduce(101), // ^=, reduce: FieldValue
			reduce(101), // <<=, reduce: FieldValue
			reduce(101), // >>=, reduce: FieldValue
			reduce(101), // >>>=, reduce: FieldValue
			reduce(101), // +=, reduce: FieldValue
			reduce(101), // .=, reduce: FieldValue
			reduce(101), // -=, reduce: FieldValue
			reduce(101), // *=, reduce: FieldValue
			reduce(101), // /=, reduce: FieldValue
			reduce(101), // //=, reduce: FieldValue
			reduce(101), // %=, reduce: FieldValue
			reduce(101), // **=, reduce: FieldValue
			reduce(101), // ?, reduce: FieldValue
			nil,         // :
			reduce(101), // ||, reduce: FieldValue
			reduce(101), // ^^, reduce: FieldValue
			reduce(101), // &&, reduce: FieldValue
			reduce(101), // ??, reduce: FieldValue
			reduce(101), // ???, reduce: FieldValue
			reduce(101), // =~, reduce: FieldValue
			reduce(101), // !=~, reduce: FieldValue
			reduce(101), // ==, reduce: FieldValue
			reduce(101), // !=, reduce: FieldValue
			reduce(101), // <=>, reduce: FieldValue
			reduce(101), // >=, reduce: FieldValue
			reduce(101), // <, reduce: FieldValue
			reduce(101), // <=, reduce: FieldValue
			reduce(101), // ^, reduce: FieldValue
			reduce(101), // &, reduce: FieldValue
			reduce(101), // <<, reduce: FieldValue
			reduce(101), // >>>, reduce: FieldValue
			reduce(101), // +, reduce: FieldValue
			reduce(101), // -, reduce: FieldValue
			reduce(101), // .+, reduce: FieldValue
			reduce(101), // .-, reduce: FieldValue
			reduce(101), // *, reduce: FieldValue
			reduce(101), // /, reduce: FieldValue
			reduce(101), // //, reduce: FieldValue
			reduce(101), // %, reduce: FieldValue
			reduce(101), // .*, reduce: FieldValue
			reduce(101), // ./, reduce: FieldValue
			reduce(101), // .//, reduce: FieldValue
			reduce(101), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(101), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(101), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(101), // [[, reduce: FieldValue
			reduce(101), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S52
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(100), // $, reduce: FieldValue
			nil,         // empty
			reduce(100), // ;, reduce: FieldValue
			reduce(100), // {, reduce: FieldValue
			nil,         // }
			reduce(100), // =, reduce: FieldValue
			nil,         // unset
			nil,         // filter
			reduce(100), // >, reduce: FieldValue
			reduce(100), // >>, reduce: FieldValue
			reduce(100), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(100), // ||=, reduce: FieldValue
			reduce(100), // ^^=, reduce: FieldValue
			reduce(100), // &&=, reduce: FieldValue
			reduce(100), // ??=, reduce: FieldValue
			reduce(100), // ???=, reduce: FieldValue
			reduce(100), // |=, reduce: FieldValue
			reduce(100), // &=, reduce: FieldValue
			reduce(100), // ^=, reduce: FieldValue
			reduce(100), // <<=, reduce: FieldValue
			reduce(100), // >>=, reduce: FieldValue
			reduce(100), // >>>=, reduce: FieldValue
			reduce(100), // +=, reduce: FieldValue
			reduce(100), // .=, reduce: FieldValue
			reduce(100), // -=, reduce: FieldValue
			reduce(100), // *=, reduce: FieldValue
			reduce(100), // /=, reduce: FieldValue
			reduce(100), // //=, reduce: FieldValue
			reduce(100), // %=, reduce: FieldValue
			reduce(100), // **=, reduce: FieldValue
			reduce(100), // ?, reduce: FieldValue
			nil,         // :
			reduce(100), // ||, reduce: FieldValue
			reduce(100), // ^^, reduce: FieldValue
			reduce(100), // &&, reduce: FieldValue
			reduce(100), // ??, reduce: FieldValue
			reduce(100), // ???, reduce: FieldValue
			reduce(100), // =~, reduce: FieldValue
			reduce(100), // !=~, reduce: FieldValue
			reduce(100), // ==, reduce: FieldValue
			reduce(100), // !=, reduce: FieldValue
			reduce(100), // <=>, reduce: FieldValue
			reduce(100), // >=, reduce: FieldValue
			reduce(100), // <, reduce: FieldValue
			reduce(100), // <=, reduce: FieldValue
			reduce(100), // ^, reduce: FieldValue
			reduce(100), // &, reduce: FieldValue
			reduce(100), // <<, reduce: FieldValue
			reduce(100), // >>>, reduce: FieldValue
			reduce(100), // +, reduce: FieldValue
			reduce(100), // -, reduce: FieldValue
			reduce(100), // .+, reduce: FieldValue
			reduce(100), // .-, reduce: FieldValue
			reduce(100), // *, reduce: FieldValue
			reduce(100), // /, reduce: FieldValue
			reduce(100), // //, reduce: FieldValue
			reduce(100), // %, reduce: FieldValue
			reduce(100), // .*, reduce: FieldValue
			reduce(100), // ./, reduce: FieldValue
			reduce(100), // .//, reduce: FieldValue
			reduce(100), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(100), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(100), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(100), // [[, reduce: FieldValue
			reduce(100), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S53
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(213), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(213), // ;, reduce: MlrvalOrFunction
			reduce(213), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(213), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(213), // >, reduce: MlrvalOrFunction
			reduce(213), // >>, reduce: MlrvalOrFunction
			reduce(213), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(213), // ||=, reduce: MlrvalOrFunction
			reduce(213), // ^^=, reduce: MlrvalOrFunction
			reduce(213), // &&=, reduce: MlrvalOrFunction
			reduce(213), // ??=, reduce: MlrvalOrFunction
			reduce(213), // ???=, reduce: MlrvalOrFunction
			reduce(213), // |=, reduce: MlrvalOrFunction
			reduce(213), // &=, reduce: MlrvalOrFunction
			reduce(213), // ^=, reduce: MlrvalOrFunction
			reduce(213), // <<=, reduce: MlrvalOrFunction
			reduce(213), // >>=, reduce: MlrvalOrFunction
			reduce(213), // >>>=, reduce: MlrvalOrFunction
			reduce(213), // +=, reduce: MlrvalOrFunction
			reduce(213), // .=, reduce: MlrvalOrFunction
			reduce(213), // -=, reduce: MlrvalOrFunction
			reduce(213), // *=, reduce: MlrvalOrFunction
			reduce(213), // /=, reduce: MlrvalOrFunction
			reduce(213), // //=, reduce: MlrvalOrFunction
			reduce(213), // %=, reduce: MlrvalOrFunction
			reduce(213), // **=, reduce: MlrvalOrFunction
			reduce(213), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(213), // ||, reduce: MlrvalOrFunction
			reduce(213), // ^^, reduce: MlrvalOrFunction
			reduce(213), // &&, reduce: MlrvalOrFunction
			reduce(213), // ??, reduce: MlrvalOrFunction
			reduce(213), // ???, reduce: MlrvalOrFunction
			reduce(213), // =~, reduce: MlrvalOrFunction
			reduce(213), // !=~, reduce: MlrvalOrFunction
			reduce(213), // ==, reduce: MlrvalOrFunction
			reduce(213), // !=, reduce: MlrvalOrFunction
			reduce(213), // <=>, reduce: MlrvalOrFunction
			reduce(213), // >=, reduce: MlrvalOrFunction
			reduce(213), // <, reduce: MlrvalOrFunction
			reduce(213), // <=, reduce: MlrvalOrFunction
			reduce(213), // ^, reduce: MlrvalOrFunction
			reduce(213), // &, reduce: MlrvalOrFunction
			reduce(213), // <<, reduce: MlrvalOrFunction
			reduce(213), // >>>, reduce: MlrvalOrFunction
			reduce(213), // +, reduce: MlrvalOrFunction
			reduce(213), // -, reduce: MlrvalOrFunction
			reduce(213), // .+, reduce: MlrvalOrFunction
			reduce(213), // .-, reduce: MlrvalOrFunction
			reduce(213), // *, reduce: MlrvalOrFunction
			reduce(213), // /, reduce: MlrvalOrFunction
			reduce(213), // //, reduce: MlrvalOrFunction
			reduce(213), // %, reduce: MlrvalOrFunction
			reduce(213), // .*, reduce: MlrvalOrFunction
			reduce(213), // ./, reduce: MlrvalOrFunction
			reduce(213), // .//, reduce: MlrvalOrFunction
			reduce(213), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(213), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(213), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(213), // [[, reduce: MlrvalOrFunction
			reduce(213), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S54
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(233), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(233), // ;, reduce: MlrvalOrFunction
			reduce(233), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(233), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(233), // >, reduce: MlrvalOrFunction
			reduce(233), // >>, reduce: MlrvalOrFunction
			reduce(233), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(233), // ||=, reduce: MlrvalOrFunction
			reduce(233), // ^^=, reduce: MlrvalOrFunction
			reduce(233), // &&=, reduce: MlrvalOrFunction
			reduce(233), // ??=, reduce: MlrvalOrFunction
			reduce(233), // ???=, reduce: MlrvalOrFunction
			reduce(233), // |=, reduce: MlrvalOrFunction
			reduce(233), // &=, reduce: MlrvalOrFunction
			reduce(233), // ^=, reduce: MlrvalOrFunction
			reduce(233), // <<=, reduce: MlrvalOrFunction
			reduce(233), // >>=, reduce: MlrvalOrFunction
			reduce(233), // >>>=, reduce: MlrvalOrFunction
			reduce(233), // +=, reduce: MlrvalOrFunction
			reduce(233), // .=, reduce: MlrvalOrFunction
			reduce(233), // -=, reduce: MlrvalOrFunction
			reduce(233), // *=, reduce: MlrvalOrFunction
			reduce(233), // /=, reduce: MlrvalOrFunction
			reduce(233), // //=, reduce: MlrvalOrFunction
			reduce(233), // %=, reduce: MlrvalOrFunction
			reduce(233), // **=, reduce: MlrvalOrFunction
			reduce(233), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(233), // ||, reduce: MlrvalOrFunction
			reduce(233), // ^^, reduce: MlrvalOrFunction
			reduce(233), // &&, reduce: MlrvalOrFunction
			reduce(233), // ??, reduce: MlrvalOrFunction
			reduce(233), // ???, reduce: MlrvalOrFunction
			reduce(233), // =~, reduce: MlrvalOrFunction
			reduce(233), // !=~, reduce: MlrvalOrFunction
			reduce(233), // ==, reduce: MlrvalOrFunction
			reduce(233), // !=, reduce: MlrvalOrFunction
			reduce(233), // <=>, reduce: MlrvalOrFunction
			reduce(233), // >=, reduce: MlrvalOrFunction
			reduce(233), // <, reduce: MlrvalOrFunction
			reduce(233), // <=, reduce: MlrvalOrFunction
			reduce(233), // ^, reduce: MlrvalOrFunction
			reduce(233), // &, reduce: MlrvalOrFunction
			reduce(233), // <<, reduce: MlrvalOrFunction
			reduce(233), // >>>, reduce: MlrvalOrFunction
			reduce(233), // +, reduce: MlrvalOrFunction
			reduce(233), // -, reduce: MlrvalOrFunction
			reduce(233), // .+, reduce: MlrvalOrFunction
			reduce(233), // .-, reduce: MlrvalOrFunction
			reduce(233), // *, reduce: MlrvalOrFunction
			reduce(233), // /, reduce: MlrvalOrFunction
			reduce(233), // //, reduce: MlrvalOrFunction
			reduce(233), // %, reduce: MlrvalOrFunction
			reduce(233), // .*, reduce: MlrvalOrFunction
			reduce(233), // ./, reduce: MlrvalOrFunction
			reduce(233), // .//, reduce: MlrvalOrFunction
			reduce(233), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(233), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(233), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(233), // [[, reduce: MlrvalOrFunction
			reduce(233), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S55
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(210), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(210), // ;, reduce: MlrvalOrFunction
			reduce(210), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(210), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(210), // >, reduce: MlrvalOrFunction
			reduce(210), // >>, reduce: MlrvalOrFunction
			reduce(210), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(210), // ||=, reduce: MlrvalOrFunction
			reduce(210), // ^^=, reduce: MlrvalOrFunction
			reduce(210), // &&=, reduce: MlrvalOrFunction
			reduce(210), // ??=, reduce: MlrvalOrFunction
			reduce(210), // ???=, reduce: MlrvalOrFunction
			reduce(210), // |=, reduce: MlrvalOrFunction
			reduce(210), // &=, reduce: MlrvalOrFunction
			reduce(210), // ^=, reduce: MlrvalOrFunction
			reduce(210), // <<=, reduce: MlrvalOrFunction
			reduce(210), // >>=, reduce: MlrvalOrFunction
			reduce(210), // >>>=, reduce: MlrvalOrFunction
			reduce(210), // +=, reduce: MlrvalOrFunction
			reduce(210), // .=, reduce: MlrvalOrFunction
			reduce(210), // -=, reduce: MlrvalOrFunction
			reduce(210), // *=, reduce: MlrvalOrFunction
			reduce(210), // /=, reduce: MlrvalOrFunction
			reduce(210), // //=, reduce: MlrvalOrFunction
			reduce(210), // %=, reduce: MlrvalOrFunction
			reduce(210), // **=, reduce: MlrvalOrFunction
			reduce(210), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(210), // ||, reduce: MlrvalOrFunction
			reduce(210), // ^^, reduce: MlrvalOrFunction
			reduce(210), // &&, reduce: MlrvalOrFunction
			reduce(210), // ??, reduce: MlrvalOrFunction
			reduce(210), // ???, reduce: MlrvalOrFunction
			reduce(210), // =~, reduce: MlrvalOrFunction
			reduce(210), // !=~, reduce: MlrvalOrFunction
			reduce(210), // ==, reduce: MlrvalOrFunction
			reduce(210), // !=, reduce: MlrvalOrFunction
			reduce(210), // <=>, reduce: MlrvalOrFunction
			reduce(210), // >=, reduce: MlrvalOrFunction
			reduce(210), // <, reduce: MlrvalOrFunction
			reduce(210), // <=, reduce: MlrvalOrFunction
			reduce(210), // ^, reduce: MlrvalOrFunction
			reduce(210), // &, reduce: MlrvalOrFunction
			reduce(210), // <<, reduce: MlrvalOrFunction
			reduce(210), // >>>, reduce: MlrvalOrFunction
			reduce(210), // +, reduce: MlrvalOrFunction
			reduce(210), // -, reduce: MlrvalOrFunction
			reduce(210), // .+, reduce: MlrvalOrFunction
			reduce(210), // .-, reduce: MlrvalOrFunction
			reduce(210), // *, reduce: MlrvalOrFunction
			reduce(210), // /, reduce: MlrvalOrFunction
			reduce(210), // //, reduce: MlrvalOrFunction
			reduce(210), // %, reduce: MlrvalOrFunction
			reduce(210), // .*, reduce: MlrvalOrFunction
			reduce(210), // ./, reduce: MlrvalOrFunction
			reduce(210), // .//, reduce: MlrvalOrFunction
			reduce(210), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(210), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(210), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(210), // [[, reduce: MlrvalOrFunction
			reduce(210), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S56
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(102), // $, reduce: FieldValue
			nil,         // empty
			reduce(102), // ;, reduce: FieldValue
			reduce(102), // {, reduce: FieldValue
			nil,         // }
			reduce(102), // =, reduce: FieldValue
			nil,         // unset
			nil,         // filter
			reduce(102), // >, reduce: FieldValue
			reduce(102), // >>, reduce: FieldValue
			reduce(102), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(102), // ||=, reduce: FieldValue
			reduce(102), // ^^=, reduce: FieldValue
			reduce(102), // &&=, reduce: FieldValue
			reduce(102), // ??=, reduce: FieldValue
			reduce(102), // ???=, reduce: FieldValue
			reduce(102), // |=, reduce: FieldValue
			reduce(102), // &=, reduce: FieldValue
			reduce(102), // ^=, reduce: FieldValue
			reduce(102), // <<=, reduce: FieldValue
			reduce(102), // >>=, reduce: FieldValue
			reduce(102), // >>>=, reduce: FieldValue
			reduce(102), // +=, reduce: FieldValue
			reduce(102), // .=, reduce: FieldValue
			reduce(102), // -=, reduce: FieldValue
			reduce(102), // *=, reduce: FieldValue
			reduce(102), // /=, reduce: FieldValue
			reduce(102), // //=, reduce: FieldValue
			reduce(102), // %=, reduce: FieldValue
			reduce(102), // **=, reduce: FieldValue
			reduce(102), // ?, reduce: FieldValue
			nil,         // :
			reduce(102), // ||, reduce: FieldValue
			reduce(102), // ^^, reduce: FieldValue
			reduce(102), // &&, reduce: FieldValue
			reduce(102), // ??, reduce: FieldValue
			reduce(102), // ???, reduce: FieldValue
			reduce(102), // =~, reduce: FieldValue
			reduce(102), // !=~, reduce: FieldValue
			reduce(102), // ==, reduce: FieldValue
			reduce(102), // !=, reduce: FieldValue
			reduce(102), // <=>, reduce: FieldValue
			reduce(102), // >=, reduce: FieldValue
			reduce(102), // <, reduce: FieldValue
			reduce(102), // <=, reduce: FieldValue
			reduce(102), // ^, reduce: FieldValue
			reduce(102), // &, reduce: FieldValue
			reduce(102), // <<, reduce: FieldValue
			reduce(102), // >>>, reduce: FieldValue
			reduce(102), // +, reduce: FieldValue
			reduce(102), // -, reduce: FieldValue
			reduce(102), // .+, reduce: FieldValue
			reduce(102), // .-, reduce: FieldValue
			reduce(102), // *, reduce: FieldValue
			reduce(102), // /, reduce: FieldValue
			reduce(102), // //, reduce: FieldValue
			reduce(102), // %, reduce: FieldValue
			reduce(102), // .*, reduce: FieldValue
			reduce(102), // ./, reduce: FieldValue
			reduce(102), // .//, reduce: FieldValue
			reduce(102), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(102), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(102), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(102), // [[, reduce: FieldValue
			reduce(102), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S57
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(103), // $, reduce: FieldValue
			nil,         // empty
			reduce(103), // ;, reduce: FieldValue
			reduce(103), // {, reduce: FieldValue
			nil,         // }
			reduce(103), // =, reduce: FieldValue
			nil,         // unset
			nil,         // filter
			reduce(103), // >, reduce: FieldValue
			reduce(103), // >>, reduce: FieldValue
			reduce(103), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(103), // ||=, reduce: FieldValue
			reduce(103), // ^^=, reduce: FieldValue
			reduce(103), // &&=, reduce: FieldValue
			reduce(103), // ??=, reduce: FieldValue
			reduce(103), // ???=, reduce: FieldValue
			reduce(103), // |=, reduce: FieldValue
			reduce(103), // &=, reduce: FieldValue
			reduce(103), // ^=, reduce: FieldValue
			reduce(103), // <<=, reduce: FieldValue
			reduce(103), // >>=, reduce: FieldValue
			reduce(103), // >>>=, reduce: FieldValue
			reduce(103), // +=, reduce: FieldValue
			reduce(103), // .=, reduce: FieldValue
			reduce(103), // -=, reduce: FieldValue
			reduce(103), // *=, reduce: FieldValue
			reduce(103), // /=, reduce: FieldValue
			reduce(103), // //=, reduce: FieldValue
			reduce(103), // %=, reduce: FieldValue
			reduce(103), // **=, reduce: FieldValue
			reduce(103), // ?, reduce: FieldValue
			nil,         // :
			reduce(103), // ||, reduce: FieldValue
			reduce(103), // ^^, reduce: FieldValue
			reduce(103), // &&, reduce: FieldValue
			reduce(103), // ??, reduce: FieldValue
			reduce(103), // ???, reduce: FieldValue
			reduce(103), // =~, reduce: FieldValue
			reduce(103), // !=~, reduce: FieldValue
			reduce(103), // ==, reduce: FieldValue
			reduce(103), // !=, reduce: FieldValue
			reduce(103), // <=>, reduce: FieldValue
			reduce(103), // >=, reduce: FieldValue
			reduce(103), // <, reduce: FieldValue
			reduce(103), // <=, reduce: FieldValue
			reduce(103), // ^, reduce: FieldValue
			reduce(103), // &, reduce: FieldValue
			reduce(103), // <<, reduce: FieldValue
			reduce(103), // >>>, reduce: FieldValue
			reduce(103), // +, reduce: FieldValue
			reduce(103), // -, reduce: FieldValue
			reduce(103), // .+, reduce: FieldValue
			reduce(103), // .-, reduce: FieldValue
			reduce(103), // *, reduce: FieldValue
			reduce(103), // /, reduce: FieldValue
			reduce(103), // //, reduce: FieldValue
			reduce(103), // %, reduce: FieldValue
			reduce(103), // .*, reduce: FieldValue
			reduce(103), // ./, reduce: FieldValue
			reduce(103), // .//, reduce: FieldValue
			reduce(103), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(103), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(103), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(103), // [[, reduce: FieldValue
			reduce(103), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S58
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(104), // $, reduce: DirectFieldValue
			nil,         // empty
			reduce(104), // ;, reduce: DirectFieldValue
			reduce(104), // {, reduce: DirectFieldValue
			nil,         // }
			reduce(104), // =, reduce: DirectFieldValue
			nil,         // unset
			nil,         // filter
			reduce(104), // >, reduce: DirectFieldValue
			reduce(104), // >>, reduce: DirectFieldValue
			reduce(104), // |, reduce: DirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(104), // ||=, reduce: DirectFieldValue
			reduce(104), // ^^=, reduce: DirectFieldValue
			reduce(104), // &&=, reduce: DirectFieldValue
			reduce(104), // ??=, reduce: DirectFieldValue
			reduce(104), // ???=, reduce: DirectFieldValue
			reduce(104), // |=, reduce: DirectFieldValue
			reduce(104), // &=, reduce: DirectFieldValue
			reduce(104), // ^=, reduce: DirectFieldValue
			reduce(104), // <<=, reduce: DirectFieldValue
			reduce(104), // >>=, reduce: DirectFieldValue
			reduce(104), // >>>=, reduce: DirectFieldValue
			reduce(104), // +=, reduce: DirectFieldValue
			reduce(104), // .=, reduce: DirectFieldValue
			reduce(104), // -=, reduce: DirectFieldValue
			reduce(104), // *=, reduce: DirectFieldValue
			reduce(104), // /=, reduce: DirectFieldValue
			reduce(104), // //=, reduce: DirectFieldValue
			reduce(104), // %=, reduce: DirectFieldValue
			reduce(104), // **=, reduce: DirectFieldValue
			reduce(104), // ?, reduce: DirectFieldValue
			nil,         // :
			reduce(104), // ||, reduce: DirectFieldValue
			reduce(104), // ^^, reduce: DirectFieldValue
			reduce(104), // &&, reduce: DirectFieldValue
			reduce(104), // ??, reduce: DirectFieldValue
			reduce(104), // ???, reduce: DirectFieldValue
			reduce(104), // =~, reduce: DirectFieldValue
			reduce(104), // !=~, reduce: DirectFieldValue
			reduce(104), // ==, reduce: DirectFieldValue
			reduce(104), // !=, reduce: DirectFieldValue
			reduce(104), // <=>, reduce: DirectFieldValue
			reduce(104), // >=, reduce: DirectFieldValue
			reduce(104), // <, reduce: DirectFieldValue
			reduce(104), // <=, reduce: DirectFieldValue
			reduce(104), // ^, reduce: DirectFieldValue
			reduce(104), // &, reduce: DirectFieldValue
			reduce(104), // <<, reduce: DirectFieldValue
			reduce(104), // >>>, reduce: DirectFieldValue
			reduce(104), // +, reduce: DirectFieldValue
			reduce(104), // -, reduce: DirectFieldValue
			reduce(104), // .+, reduce: DirectFieldValue
			reduce(104), // .-, reduce: DirectFieldValue
			reduce(104), // *, reduce: DirectFieldValue
			reduce(104), // /, reduce: DirectFieldValue
			reduce(104), // //, reduce: DirectFieldValue
			reduce(104), // %, reduce: DirectFieldValue
			reduce(104), // .*, reduce: DirectFieldValue
			reduce(104), // ./, reduce: DirectFieldValue
			reduce(104), // .//, reduce: DirectFieldValue
			reduce(104), // ., reduce: DirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(104), // **, reduce: DirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(104), // [, reduce: DirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(104), // [[, reduce: DirectFieldValue
			reduce(104), // [[[, reduce: DirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S59
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S60
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(106), // $, reduce: BracedFieldValue
			nil,         // empty
			reduce(106), // ;, reduce: BracedFieldValue
			reduce(106), // {, reduce: BracedFieldValue
			nil,         // }
			reduce(106), // =, reduce: BracedFieldValue
			nil,         // unset
			nil,         // filter
			reduce(106), // >, reduce: BracedFieldValue
			reduce(106), // >>, reduce: BracedFieldValue
			reduce(106), // |, reduce: BracedFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(106), // ||=, reduce: BracedFieldValue
			reduce(106), // ^^=, reduce: BracedFieldValue
			reduce(106), // &&=, reduce: BracedFieldValue
			reduce(106), // ??=, reduce: BracedFieldValue
			reduce(106), // ???=, reduce: BracedFieldValue
			reduce(106), // |=, reduce: BracedFieldValue
			reduce(106), // &=, reduce: BracedFieldValue
			reduce(106), // ^=, reduce: BracedFieldValue
			reduce(106), // <<=, reduce: BracedFieldValue
			reduce(106), // >>=, reduce: BracedFieldValue
			reduce(106), // >>>=, reduce: BracedFieldValue
			reduce(106), // +=, reduce: BracedFieldValue
			reduce(106), // .=, reduce: BracedFieldValue
			reduce(106), // -=, reduce: BracedFieldValue
			reduce(106), // *=, reduce: BracedFieldValue
			reduce(106), // /=, reduce: BracedFieldValue
			reduce(106), // //=, reduce: BracedFieldValue
			reduce(106), // %=, reduce: BracedFieldValue
			reduce(106), // **=, reduce: BracedFieldValue
			reduce(106), // ?, reduce: BracedFieldValue
			nil,         // :
			reduce(106), // ||, reduce: BracedFieldValue
			reduce(106), // ^^, reduce: BracedFieldValue
			reduce(106), // &&, reduce: BracedFieldValue
			reduce(106), // ??, reduce: BracedFieldValue
			reduce(106), // ???, reduce: BracedFieldValue
			reduce(106), // =~, reduce: BracedFieldValue
			reduce(106), // !=~, reduce: BracedFieldValue
			reduce(106), // ==, reduce: BracedFieldValue
			reduce(106), // !=, reduce: BracedFieldValue
			reduce(106), // <=>, reduce: BracedFieldValue
			reduce(106), // >=, reduce: BracedFieldValue
			reduce(106), // <, reduce: BracedFieldValue
			reduce(106), // <=, reduce: BracedFieldValue
			reduce(106), // ^, reduce: BracedFieldValue
			reduce(106), // &, reduce: BracedFieldValue
			reduce(106), // <<, reduce: BracedFieldValue
			reduce(106), // >>>, reduce: BracedFieldValue
			reduce(106), // +, reduce: BracedFieldValue
			reduce(106), // -, reduce: BracedFieldValue
			reduce(106), // .+, reduce: BracedFieldValue
			reduce(106), // .-, reduce: BracedFieldValue
			reduce(106), // *, reduce: BracedFieldValue
			reduce(106), // /, reduce: BracedFieldValue
			reduce(106), // //, reduce: BracedFieldValue
			reduce(106), // %, reduce: BracedFieldValue
			reduce(106), // .*, reduce: BracedFieldValue
			reduce(106), // ./, reduce: BracedFieldValue
			reduce(106), // .//, reduce: BracedFieldValue
			reduce(106), // ., reduce: BracedFieldValue
			nil,         // !
			nil,         // ~
			reduce(106), // **, reduce: BracedFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(106), // [, reduce: BracedFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(106), // [[, reduce: BracedFieldValue
			reduce(106), // [[[, reduce: BracedFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S61
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S62
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S63
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(109), // $, reduce: FullSrec
			nil,         // empty
			reduce(109), // ;, reduce: FullSrec
			reduce(109), // {, reduce: FullSrec
			nil,         // }
			reduce(109), // =, reduce: FullSrec
			nil,         // unset
			nil,         // filter
			reduce(109), // >, reduce: FullSrec
			reduce(109), // >>, reduce: FullSrec
			reduce(109), // |, reduce: FullSrec
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(109), // ||=, reduce: FullSrec
			reduce(109), // ^^=, reduce: FullSrec
			reduce(109), // &&=, reduce: FullSrec
			reduce(109), // ??=, reduce: FullSrec
			reduce(109), // ???=, reduce: FullSrec
			reduce(109), // |=, reduce: FullSrec
			reduce(109), // &=, reduce: FullSrec
			reduce(109), // ^=, reduce: FullSrec
			reduce(109), // <<=, reduce: FullSrec
			reduce(109), // >>=, reduce: FullSrec
			reduce(109), // >>>=, reduce: FullSrec
			reduce(109), // +=, reduce: FullSrec
			reduce(109), // .=, reduce: FullSrec
			reduce(109), // -=, reduce: FullSrec
			reduce(109), // *=, reduce: FullSrec
			reduce(109), // /=, reduce: FullSrec
			reduce(109), // //=, reduce: FullSrec
			reduce(109), // %=, reduce: FullSrec
			reduce(109), // **=, reduce: FullSrec
			reduce(109), // ?, reduce: FullSrec
			nil,         // :
			reduce(109), // ||, reduce: FullSrec
			reduce(109), // ^^, reduce: FullSrec
			reduce(109), // &&, reduce: FullSrec
			reduce(109), // ??, reduce: FullSrec
			reduce(109), // ???, reduce: FullSrec
			reduce(109), // =~, reduce: FullSrec
			reduce(109), // !=~, reduce: FullSrec
			reduce(109), // ==, reduce: FullSrec
			reduce(109), // !=, reduce: FullSrec
			reduce(109), // <=>, reduce: FullSrec
			reduce(109), // >=, reduce: FullSrec
			reduce(109), // <, reduce: FullSrec
			reduce(109), // <=, reduce: FullSrec
			reduce(109), // ^, reduce: FullSrec
			reduce(109), // &, reduce: FullSrec
			reduce(109), // <<, reduce: FullSrec
			reduce(109), // >>>, reduce: FullSrec
			reduce(109), // +, reduce: FullSrec
			reduce(109), // -, reduce: FullSrec
			reduce(109), // .+, reduce: FullSrec
			reduce(109), // .-, reduce: FullSrec
			reduce(109), // *, reduce: FullSrec
			reduce(109), // /, reduce: FullSrec
			reduce(109), // //, reduce: FullSrec
			reduce(109), // %, reduce: FullSrec
			reduce(109), // .*, reduce: FullSrec
			reduce(109), // ./, reduce: FullSrec
			reduce(109), // .//, reduce: FullSrec
			reduce(109), // ., reduce: FullSrec
			nil,         // !
			nil,         // ~
			reduce(109), // **, reduce: FullSrec
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(109), // [, reduce: FullSrec
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(109), // [[, reduce: FullSrec
			reduce(109), // [[[, reduce: FullSrec
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S64
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(212), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(212), // ;, reduce: MlrvalOrFunction
			reduce(212), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(212), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(212), // >, reduce: MlrvalOrFunction
			reduce(212), // >>, reduce: MlrvalOrFunction
			reduce(212), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(212), // ||=, reduce: MlrvalOrFunction
			reduce(212), // ^^=, reduce: MlrvalOrFunction
			reduce(212), // &&=, reduce: MlrvalOrFunction
			reduce(212), // ??=, reduce: MlrvalOrFunction
			reduce(212), // ???=, reduce: MlrvalOrFunction
			reduce(212), // |=, reduce: MlrvalOrFunction
			reduce(212), // &=, reduce: MlrvalOrFunction
			reduce(212), // ^=, reduce: MlrvalOrFunction
			reduce(212), // <<=, reduce: MlrvalOrFunction
			reduce(212), // >>=, reduce: MlrvalOrFunction
			reduce(212), // >>>=, reduce: MlrvalOrFunction
			reduce(212), // +=, reduce: MlrvalOrFunction
			reduce(212), // .=, reduce: MlrvalOrFunction
			reduce(212), // -=, reduce: MlrvalOrFunction
			reduce(212), // *=, reduce: MlrvalOrFunction
			reduce(212), // /=, reduce: MlrvalOrFunction
			reduce(212), // //=, reduce: MlrvalOrFunction
			reduce(212), // %=, reduce: MlrvalOrFunction
			reduce(212), // **=, reduce: MlrvalOrFunction
			reduce(212), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(212), // ||, reduce: MlrvalOrFunction
			reduce(212), // ^^, reduce: MlrvalOrFunction
			reduce(212), // &&, reduce: MlrvalOrFunction
			reduce(212), // ??, reduce: MlrvalOrFunction
			reduce(212), // ???, reduce: MlrvalOrFunction
			reduce(212), // =~, reduce: MlrvalOrFunction
			reduce(212), // !=~, reduce: MlrvalOrFunction
			reduce(212), // ==, reduce: MlrvalOrFunction
			reduce(212), // !=, reduce: MlrvalOrFunction
			reduce(212), // <=>, reduce: MlrvalOrFunction
			reduce(212), // >=, reduce: MlrvalOrFunction
			reduce(212), // <, reduce: MlrvalOrFunction
			reduce(212), // <=, reduce: MlrvalOrFunction
			reduce(212), // ^, reduce: MlrvalOrFunction
			reduce(212), // &, reduce: MlrvalOrFunction
			reduce(212), // <<, reduce: MlrvalOrFunction
			reduce(212), // >>>, reduce: MlrvalOrFunction
			reduce(212), // +, reduce: MlrvalOrFunction
			reduce(212), // -, reduce: MlrvalOrFunction
			reduce(212), // .+, reduce: MlrvalOrFunction
			reduce(212), // .-, reduce: MlrvalOrFunction
			reduce(212), // *, reduce: MlrvalOrFunction
			reduce(212), // /, reduce: MlrvalOrFunction
			reduce(212), // //, reduce: MlrvalOrFunction
			reduce(212), // %, reduce: MlrvalOrFunction
			reduce(212), // .*, reduce: MlrvalOrFunction
			reduce(212), // ./, reduce: MlrvalOrFunction
			reduce(212), // .//, reduce: MlrvalOrFunction
			reduce(212), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(212), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(212), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(212), // [[, reduce: MlrvalOrFunction
			reduce(212), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S65
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(113), // $, reduce: DirectOosvarValue
			nil,         // empty
			reduce(113), // ;, reduce: DirectOosvarValue
			reduce(113), // {, reduce: DirectOosvarValue
			nil,         // }
			reduce(113), // =, reduce: DirectOosvarValue
			nil,         // unset
			nil,         // filter
			reduce(113), // >, reduce: DirectOosvarValue
			reduce(113), // >>, reduce: DirectOosvarValue
			reduce(113), // |, reduce: DirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(113), // ||=, reduce: DirectOosvarValue
			reduce(113), // ^^=, reduce: DirectOosvarValue
			reduce(113), // &&=, reduce: DirectOosvarValue
			reduce(113), // ??=, reduce: DirectOosvarValue
			reduce(113), // ???=, reduce: DirectOosvarValue
			reduce(113), // |=, reduce: DirectOosvarValue
			reduce(113), // &=, reduce: DirectOosvarValue
			reduce(113), // ^=, reduce: DirectOosvarValue
			reduce(113), // <<=, reduce: DirectOosvarValue
			reduce(113), // >>=, reduce: DirectOosvarValue
			reduce(113), // >>>=, reduce: DirectOosvarValue
			reduce(113), // +=, reduce: DirectOosvarValue
			reduce(113), // .=, reduce: DirectOosvarValue
			reduce(113), // -=, reduce: DirectOosvarValue
			reduce(113), // *=, reduce: DirectOosvarValue
			reduce(113), // /=, reduce: DirectOosvarValue
			reduce(113), // //=, reduce: DirectOosvarValue
			reduce(113), // %=, reduce: DirectOosvarValue
			reduce(113), // **=, reduce: DirectOosvarValue
			reduce(113), // ?, reduce: DirectOosvarValue
			nil,         // :
			reduce(113), // ||, reduce: DirectOosvarValue
			reduce(113), // ^^, reduce: DirectOosvarValue
			reduce(113), // &&, reduce: DirectOosvarValue
			reduce(113), // ??, reduce: DirectOosvarValue
			reduce(113), // ???, reduce: DirectOosvarValue
			reduce(113), // =~, reduce: DirectOosvarValue
			reduce(113), // !=~, reduce: DirectOosvarValue
			reduce(113), // ==, reduce: DirectOosvarValue
			reduce(113), // !=, reduce: DirectOosvarValue
			reduce(113), // <=>, reduce: DirectOosvarValue
			reduce(113), // >=, reduce: DirectOosvarValue
			reduce(113), // <, reduce: DirectOosvarValue
			reduce(113), // <=, reduce: DirectOosvarValue
			reduce(113), // ^, reduce: DirectOosvarValue
			reduce(113), // &, reduce: DirectOosvarValue
			reduce(113), // <<, reduce: DirectOosvarValue
			reduce(113), // >>>, reduce: DirectOosvarValue
			reduce(113), // +, reduce: DirectOosvarValue
			reduce(113), // -, reduce: DirectOosvarValue
			reduce(113), // .+, reduce: DirectOosvarValue
			reduce(113), // .-, reduce: DirectOosvarValue
			reduce(113), // *, reduce: DirectOosvarValue
			reduce(113), // /, reduce: DirectOosvarValue
			reduce(113), // //, reduce: DirectOosvarValue
			reduce(113), // %, reduce: DirectOosvarValue
			reduce(113), // .*, reduce: DirectOosvarValue
			reduce(113), // ./, reduce: DirectOosvarValue
			reduce(113), // .//, reduce: DirectOosvarValue
			reduce(113), // ., reduce: DirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(113), // **, reduce: DirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(113), // [, reduce: DirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(113), // [[, reduce: DirectOosvarValue
			reduce(113), // [[[, reduce: DirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S66
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S67
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(115), // $, reduce: BracedOosvarValue
			nil,         // empty
			reduce(115), // ;, reduce: BracedOosvarValue
			reduce(115), // {, reduce: BracedOosvarValue
			nil,         // }
			reduce(115), // =, reduce: BracedOosvarValue
			nil,         // unset
			nil,         // filter
			reduce(115), // >, reduce: BracedOosvarValue
			reduce(115), // >>, reduce: BracedOosvarValue
			reduce(115), // |, reduce: BracedOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(115), // ||=, reduce: BracedOosvarValue
			reduce(115), // ^^=, reduce: BracedOosvarValue
			reduce(115), // &&=, reduce: BracedOosvarValue
			reduce(115), // ??=, reduce: BracedOosvarValue
			reduce(115), // ???=, reduce: BracedOosvarValue
			reduce(115), // |=, reduce: BracedOosvarValue
			reduce(115), // &=, reduce: BracedOosvarValue
			reduce(115), // ^=, reduce: BracedOosvarValue
			reduce(115), // <<=, reduce: BracedOosvarValue
			reduce(115), // >>=, reduce: BracedOosvarValue
			reduce(115), // >>>=, reduce: BracedOosvarValue
			reduce(115), // +=, reduce: BracedOosvarValue
			reduce(115), // .=, reduce: BracedOosvarValue
			reduce(115), // -=, reduce: BracedOosvarValue
			reduce(115), // *=, reduce: BracedOosvarValue
			reduce(115), // /=, reduce: BracedOosvarValue
			reduce(115), // //=, reduce: BracedOosvarValue
			reduce(115), // %=, reduce: BracedOosvarValue
			reduce(115), // **=, reduce: BracedOosvarValue
			reduce(115), // ?, reduce: BracedOosvarValue
			nil,         // :
			reduce(115), // ||, reduce: BracedOosvarValue
			reduce(115), // ^^, reduce: BracedOosvarValue
			reduce(115), // &&, reduce: BracedOosvarValue
			reduce(115), // ??, reduce: BracedOosvarValue
			reduce(115), // ???, reduce: BracedOosvarValue
			reduce(115), // =~, reduce: BracedOosvarValue
			reduce(115), // !=~, reduce: BracedOosvarValue
			reduce(115), // ==, reduce: BracedOosvarValue
			reduce(115), // !=, reduce: BracedOosvarValue
			reduce(115), // <=>, reduce: BracedOosvarValue
			reduce(115), // >=, reduce: BracedOosvarValue
			reduce(115), // <, reduce: BracedOosvarValue
			reduce(115), // <=, reduce: BracedOosvarValue
			reduce(115), // ^, reduce: BracedOosvarValue
			reduce(115), // &, reduce: BracedOosvarValue
			reduce(115), // <<, reduce: BracedOosvarValue
			reduce(115), // >>>, reduce: BracedOosvarValue
			reduce(115), // +, reduce: BracedOosvarValue
			reduce(115), // -, reduce: BracedOosvarValue
			reduce(115), // .+, reduce: BracedOosvarValue
			reduce(115), // .-, reduce: BracedOosvarValue
			reduce(115), // *, reduce: BracedOosvarValue
			reduce(115), // /, reduce: BracedOosvarValue
			reduce(115), // //, reduce: BracedOosvarValue
			reduce(115), // %, reduce: BracedOosvarValue
			reduce(115), // .*, reduce: BracedOosvarValue
			reduce(115), // ./, reduce: BracedOosvarValue
			reduce(115), // .//, reduce: BracedOosvarValue
			reduce(115), // ., reduce: BracedOosvarValue
			nil,         // !
			nil,         // ~
			reduce(115), // **, reduce: BracedOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(115), // [, reduce: BracedOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(115), // [[, reduce: BracedOosvarValue
			reduce(115), // [[[, reduce: BracedOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S68
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(116), // $, reduce: FullOosvar
			nil,         // empty
			reduce(116), // ;, reduce: FullOosvar
			reduce(116), // {, reduce: FullOosvar
			nil,         // }
			reduce(116), // =, reduce: FullOosvar
			nil,         // unset
			nil,         // filter
			reduce(116), // >, reduce: FullOosvar
			reduce(116), // >>, reduce: FullOosvar
			reduce(116), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(116), // ||=, reduce: FullOosvar
			reduce(116), // ^^=, reduce: FullOosvar
			reduce(116), // &&=, reduce: FullOosvar
			reduce(116), // ??=, reduce: FullOosvar
			reduce(116), // ???=, reduce: FullOosvar
			reduce(116), // |=, reduce: FullOosvar
			reduce(116), // &=, reduce: FullOosvar
			reduce(116), // ^=, reduce: FullOosvar
			reduce(116), // <<=, reduce: FullOosvar
			reduce(116), // >>=, reduce: FullOosvar
			reduce(116), // >>>=, reduce: FullOosvar
			reduce(116), // +=, reduce: FullOosvar
			reduce(116), // .=, reduce: FullOosvar
			reduce(116), // -=, reduce: FullOosvar
			reduce(116), // *=, reduce: FullOosvar
			reduce(116), // /=, reduce: FullOosvar
			reduce(116), // //=, reduce: FullOosvar
			reduce(116), // %=, reduce: FullOosvar
			reduce(116), // **=, reduce: FullOosvar
			reduce(116), // ?, reduce: FullOosvar
			nil,         // :
			reduce(116), // ||, reduce: FullOosvar
			reduce(116), // ^^, reduce: FullOosvar
			reduce(116), // &&, reduce: FullOosvar
			reduce(116), // ??, reduce: FullOosvar
			reduce(116), // ???, reduce: FullOosvar
			reduce(116), // =~, reduce: FullOosvar
			reduce(116), // !=~, reduce: FullOosvar
			reduce(116), // ==, reduce: FullOosvar
			reduce(116), // !=, reduce: FullOosvar
			reduce(116), // <=>, reduce: FullOosvar
			reduce(116), // >=, reduce: FullOosvar
			reduce(116), // <, reduce: FullOosvar
			reduce(116), // <=, reduce: FullOosvar
			reduce(116), // ^, reduce: FullOosvar
			reduce(116), // &, reduce: FullOosvar
			reduce(116), // <<, reduce: FullOosvar
			reduce(116), // >>>, reduce: FullOosvar
			reduce(116), // +, reduce: FullOosvar
			reduce(116), // -, reduce: FullOosvar
			reduce(116), // .+, reduce: FullOosvar
			reduce(116), // .-, reduce: FullOosvar
			reduce(116), // *, reduce: FullOosvar
			reduce(116), // /, reduce: FullOosvar
			reduce(116), // //, reduce: FullOosvar
			reduce(116), // %, reduce: FullOosvar
			reduce(116), // .*, reduce: FullOosvar
			reduce(116), // ./, reduce: FullOosvar
			reduce(116), // .//, reduce: FullOosvar
			reduce(116), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(116), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(116), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(116), // [[, reduce: FullOosvar
			reduce(116), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S69
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(117), // $, reduce: FullOosvar
			nil,         // empty
			reduce(117), // ;, reduce: FullOosvar
			reduce(117), // {, reduce: FullOosvar
			nil,         // }
			reduce(117), // =, reduce: FullOosvar
			nil,         // unset
			nil,         // filter
			reduce(117), // >, reduce: FullOosvar
			reduce(117), // >>, reduce: FullOosvar
			reduce(117), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(117), // ||=, reduce: FullOosvar
			reduce(117), // ^^=, reduce: FullOosvar
			reduce(117), // &&=, reduce: FullOosvar
			reduce(117), // ??=, reduce: FullOosvar
			reduce(117), // ???=, reduce: FullOosvar
			reduce(117), // |=, reduce: FullOosvar
			reduce(117), // &=, reduce: FullOosvar
			reduce(117), // ^=, reduce: FullOosvar
			reduce(117), // <<=, reduce: FullOosvar
			reduce(117), // >>=, reduce: FullOosvar
			reduce(117), // >>>=, reduce: FullOosvar
			reduce(117), // +=, reduce: FullOosvar
			reduce(117), // .=, reduce: FullOosvar
			reduce(117), // -=, reduce: FullOosvar
			reduce(117), // *=, reduce: FullOosvar
			reduce(117), // /=, reduce: FullOosvar
			reduce(117), // //=, reduce: FullOosvar
			reduce(117), // %=, reduce: FullOosvar
			reduce(117), // **=, reduce: FullOosvar
			reduce(117), // ?, reduce: FullOosvar
			nil,         // :
			reduce(117), // ||, reduce: FullOosvar
			reduce(117), // ^^, reduce: FullOosvar
			reduce(117), // &&, reduce: FullOosvar
			reduce(117), // ??, reduce: FullOosvar
			reduce(117), // ???, reduce: FullOosvar
			reduce(117), // =~, reduce: FullOosvar
			reduce(117), // !=~, reduce: FullOosvar
			reduce(117), // ==, reduce: FullOosvar
			reduce(117), // !=, reduce: FullOosvar
			reduce(117), // <=>, reduce: FullOosvar
			reduce(117), // >=, reduce: FullOosvar
			reduce(117), // <, reduce: FullOosvar
			reduce(117), // <=, reduce: FullOosvar
			reduce(117), // ^, reduce: FullOosvar
			reduce(117), // &, reduce: FullOosvar
			reduce(117), // <<, reduce: FullOosvar
			reduce(117), // >>>, reduce: FullOosvar
			reduce(117), // +, reduce: FullOosvar
			reduce(117), // -, reduce: FullOosvar
			reduce(117), // .+, reduce: FullOosvar
			reduce(117), // .-, reduce: FullOosvar
			reduce(117), // *, reduce: FullOosvar
			reduce(117), // /, reduce: FullOosvar
			reduce(117), // //, reduce: FullOosvar
			reduce(117), // %, reduce: FullOosvar
			reduce(117), // .*, reduce: FullOosvar
			reduce(117), // ./, reduce: FullOosvar
			reduce(117), // .//, reduce: FullOosvar
			reduce(117), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(117), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(117), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(117), // [[, reduce: FullOosvar
			reduce(117), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S70
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(118), // $, reduce: LocalVariable
			nil,         // empty
			reduce(118), // ;, reduce: LocalVariable
			reduce(118), // {, reduce: LocalVariable
			nil,         // }
			reduce(118), // =, reduce: LocalVariable
			nil,         // unset
			nil,         // filter
			reduce(118), // >, reduce: LocalVariable
			reduce(118), // >>, reduce: LocalVariable
			reduce(118), // |, reduce: LocalVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(270), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(118), // ||=, reduce: LocalVariable
			reduce(118), // ^^=, reduce: LocalVariable
			reduce(118), // &&=, reduce: LocalVariable
			reduce(118), // ??=, reduce: LocalVariable
			reduce(118), // ???=, reduce: LocalVariable
			reduce(118), // |=, reduce: LocalVariable
			reduce(118), // &=, reduce: LocalVariable
			reduce(118), // ^=, reduce: LocalVariable
			reduce(118), // <<=, reduce: LocalVariable
			reduce(118), // >>=, reduce: LocalVariable
			reduce(118), // >>>=, reduce: LocalVariable
			reduce(118), // +=, reduce: LocalVariable
			reduce(118), // .=, reduce: LocalVariable
			reduce(118), // -=, reduce: LocalVariable
			reduce(118), // *=, reduce: LocalVariable
			reduce(118), // /=, reduce: LocalVariable
			reduce(118), // //=, reduce: LocalVariable
			reduce(118), // %=, reduce: LocalVariable
			reduce(118), // **=, reduce: LocalVariable
			reduce(118), // ?, reduce: LocalVariable
			nil,         // :
			reduce(118), // ||, reduce: LocalVariable
			reduce(118), // ^^, reduce: LocalVariable
			reduce(118), // &&, reduce: LocalVariable
			reduce(118), // ??, reduce: LocalVariable
			reduce(118), // ???, reduce: LocalVariable
			reduce(118), // =~, reduce: LocalVariable
			reduce(118), // !=~, reduce: LocalVariable
			reduce(118), // ==, reduce: LocalVariable
			reduce(118), // !=, reduce: LocalVariable
			reduce(118), // <=>, reduce: LocalVariable
			reduce(118), // >=, reduce: LocalVariable
			reduce(118), // <, reduce: LocalVariable
			reduce(118), // <=, reduce: LocalVariable
			reduce(118), // ^, reduce: LocalVariable
			reduce(118), // &, reduce: LocalVariable
			reduce(118), // <<, reduce: LocalVariable
			reduce(118), // >>>, reduce: LocalVariable
			reduce(118), // +, reduce: LocalVariable
			reduce(118), // -, reduce: LocalVariable
			reduce(118), // .+, reduce: LocalVariable
			reduce(118), // .-, reduce: LocalVariable
			reduce(118), // *, reduce: LocalVariable
			reduce(118), // /, reduce: LocalVariable
			reduce(118), // //, reduce: LocalVariable
			reduce(118), // %, reduce: LocalVariable
			reduce(118), // .*, reduce: LocalVariable
			reduce(118), // ./, reduce: LocalVariable
			reduce(118), // .//, reduce: LocalVariable
			reduce(118), // ., reduce: LocalVariable
			nil,         // !
			nil,         // ~
			reduce(118), // **, reduce: LocalVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(118), // [, reduce: LocalVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(118), // [[, reduce: LocalVariable
			reduce(118), // [[[, reduce: LocalVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S71
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			reduce(119), // non_sigil_name, reduce: Typedecl
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S72
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			reduce(120), // non_sigil_name, reduce: Typedecl
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S73
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(272), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			reduce(121), // non_sigil_name, reduce: Typedecl
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S74
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(271), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			reduce(122), // non_sigil_name, reduce: Typedecl
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S75
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			reduce(123), // non_sigil_name, reduce: Typedecl
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S76
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			reduce(124), // non_sigil_name, reduce: Typedecl
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S77
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			reduce(125), // non_sigil_name, reduce: Typedecl
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S78
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			reduce(126), // non_sigil_name, reduce: Typedecl
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S79
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			reduce(127), // non_sigil_name, reduce: Typedecl
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S80
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(147), // $, reduce: Rvalue
			nil,         // empty
			reduce(147), // ;, reduce: Rvalue
			reduce(147), // {, reduce: Rvalue
			nil,         // }
			reduce(147), // =, reduce: Rvalue
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(147), // ||=, reduce: Rvalue
			reduce(147), // ^^=, reduce: Rvalue
			reduce(147), // &&=, reduce: Rvalue
			reduce(147), // ??=, reduce: Rvalue
			reduce(147), // ???=, reduce: Rvalue
			reduce(147), // |=, reduce: Rvalue
			reduce(147), // &=, reduce: Rvalue
			reduce(147), // ^=, reduce: Rvalue
			reduce(147), // <<=, reduce: Rvalue
			reduce(147), // >>=, reduce: Rvalue
			reduce(147), // >>>=, reduce: Rvalue
			reduce(147), // +=, reduce: Rvalue
			reduce(147), // .=, reduce: Rvalue
			reduce(147), // -=, reduce: Rvalue
			reduce(147), // *=, reduce: Rvalue
			reduce(147), // /=, reduce: Rvalue
			reduce(147), // //=, reduce: Rvalue
			reduce(147), // %=, reduce: Rvalue
			reduce(147), // **=, reduce: Rvalue
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S81
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(148), // $, reduce: PrecedenceChainStart
			nil,         // empty
			reduce(148), // ;, reduce: PrecedenceChainStart
			reduce(148), // {, reduce: PrecedenceChainStart
			nil,         // }
			reduce(148), // =, reduce: PrecedenceChainStart
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(148), // ||=, reduce: PrecedenceChainStart
			reduce(148), // ^^=, reduce: PrecedenceChainStart
			reduce(148), // &&=, reduce: PrecedenceChainStart
			reduce(148), // ??=, reduce: PrecedenceChainStart
			reduce(148), // ???=, reduce: PrecedenceChainStart
			reduce(148), // |=, reduce: PrecedenceChainStart
			reduce(148), // &=, reduce: PrecedenceChainStart
			reduce(148), // ^=, reduce: PrecedenceChainStart
			reduce(148), // <<=, reduce: PrecedenceChainStart
			reduce(148), // >>=, reduce: PrecedenceChainStart
			reduce(148), // >>>=, reduce: PrecedenceChainStart
			reduce(148), // +=, reduce: PrecedenceChainStart
			reduce(148), // .=, reduce: PrecedenceChainStart
			reduce(148), // -=, reduce: PrecedenceChainStart
			reduce(148), // *=, reduce: PrecedenceChainStart
			reduce(148), // /=, reduce: PrecedenceChainStart
			reduce(148), // //=, reduce: PrecedenceChainStart
			reduce(148), // %=, reduce: PrecedenceChainStart
			reduce(148), // **=, reduce: PrecedenceChainStart
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S82
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(150), // $, reduce: TernaryTerm
			nil,         // empty
			reduce(150), // ;, reduce: TernaryTerm
			reduce(150), // {, reduce: TernaryTerm
			nil,         // }
			reduce(150), // =, reduce: TernaryTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(150), // ||=, reduce: TernaryTerm
			reduce(150), // ^^=, reduce: TernaryTerm
			reduce(150), // &&=, reduce: TernaryTerm
			reduce(150), // ??=, reduce: TernaryTerm
			reduce(150), // ???=, reduce: TernaryTerm
			reduce(150), // |=, reduce: TernaryTerm
			reduce(150), // &=, reduce: TernaryTerm
			reduce(150), // ^=, reduce: TernaryTerm
			reduce(150), // <<=, reduce: TernaryTerm
			reduce(150), // >>=, reduce: TernaryTerm
			reduce(150), // >>>=, reduce: TernaryTerm
			reduce(150), // +=, reduce: TernaryTerm
			reduce(150), // .=, reduce: TernaryTerm
			reduce(150), // -=, reduce: TernaryTerm
			reduce(150), // *=, reduce: TernaryTerm
			reduce(150), // /=, reduce: TernaryTerm
			reduce(150), // //=, reduce: TernaryTerm
			reduce(150), // %=, reduce: TernaryTerm
			reduce(150), // **=, reduce: TernaryTerm
			shift(708),  // ?
			nil,         // :
			shift(709),  // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S83
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(152), // $, reduce: LogicalOrTerm
			nil,         // empty
			reduce(152), // ;, reduce: LogicalOrTerm
			reduce(152), // {, reduce: LogicalOrTerm
			nil,         // }
			reduce(152), // =, reduce: LogicalOrTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(152), // ||=, reduce: LogicalOrTerm
			reduce(152), // ^^=, reduce: LogicalOrTerm
			reduce(152), // &&=, reduce: LogicalOrTerm
			reduce(152), // ??=, reduce: LogicalOrTerm
			reduce(152), // ???=, reduce: LogicalOrTerm
			reduce(152), // |=, reduce: LogicalOrTerm
			reduce(152), // &=, reduce: LogicalOrTerm
			reduce(152), // ^=, reduce: LogicalOrTerm
			reduce(152), // <<=, reduce: LogicalOrTerm
			reduce(152), // >>=, reduce: LogicalOrTerm
			reduce(152), // >>>=, reduce: LogicalOrTerm
			reduce(152), // +=, reduce: LogicalOrTerm
			reduce(152), // .=, reduce: LogicalOrTerm
			reduce(152), // -=, reduce: LogicalOrTerm
			reduce(152), // *=, reduce: LogicalOrTerm
			reduce(152), // /=, reduce: LogicalOrTerm
			reduce(152), // //=, reduce: LogicalOrTerm
			reduce(152), // %=, reduce: LogicalOrTerm
			reduce(152), // **=, reduce: LogicalOrTerm
			reduce(152), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(152), // ||, reduce: LogicalOrTerm
			shift(710),  // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S84
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(154), // $, reduce: LogicalXORTerm
			nil,         // empty
			reduce(154), // ;, reduce: LogicalXORTerm
			reduce(154), // {, reduce: LogicalXORTerm
			nil,         // }
			reduce(154), // =, reduce: LogicalXORTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(154), // ||=, reduce: LogicalXORTerm
			reduce(154), // ^^=, reduce: LogicalXORTerm
			reduce(154), // &&=, reduce: LogicalXORTerm
			reduce(154), // ??=, reduce: LogicalXORTerm
			reduce(154), // ???=, reduce: LogicalXORTerm
			reduce(154), // |=, reduce: LogicalXORTerm
			reduce(154), // &=, reduce: LogicalXORTerm
			reduce(154), // ^=, reduce: LogicalXORTerm
			reduce(154), // <<=, reduce: LogicalXORTerm
			reduce(154), // >>=, reduce: LogicalXORTerm
			reduce(154), // >>>=, reduce: LogicalXORTerm
			reduce(154), // +=, reduce: LogicalXORTerm
			reduce(154), // .=, reduce: LogicalXORTerm
			reduce(154), // -=, reduce: LogicalXORTerm
			reduce(154), // *=, reduce: LogicalXORTerm
			reduce(154), // /=, reduce: LogicalXORTerm
			reduce(154), // //=, reduce: LogicalXORTerm
			reduce(154), // %=, reduce: LogicalXORTerm
			reduce(154), // **=, reduce: LogicalXORTerm
			reduce(154), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(154), // ||, reduce: LogicalXORTerm
			reduce(154), // ^^, reduce: LogicalXORTerm
			shift(711),  // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S85
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(156), // $, reduce: LogicalAndTerm
			nil,         // empty
			reduce(156), // ;, reduce: LogicalAndTerm
			reduce(156), // {, reduce: LogicalAndTerm
			nil,         // }
			reduce(156), // =, reduce: LogicalAndTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(156), // ||=, reduce: LogicalAndTerm
			reduce(156), // ^^=, reduce: LogicalAndTerm
			reduce(156), // &&=, reduce: LogicalAndTerm
			reduce(156), // ??=, reduce: LogicalAndTerm
			reduce(156), // ???=, reduce: LogicalAndTerm
			reduce(156), // |=, reduce: LogicalAndTerm
			reduce(156), // &=, reduce: LogicalAndTerm
			reduce(156), // ^=, reduce: LogicalAndTerm
			reduce(156), // <<=, reduce: LogicalAndTerm
			reduce(156), // >>=, reduce: LogicalAndTerm
			reduce(156), // >>>=, reduce: LogicalAndTerm
			reduce(156), // +=, reduce: LogicalAndTerm
			reduce(156), // .=, reduce: LogicalAndTerm
			reduce(156), // -=, reduce: LogicalAndTerm
			reduce(156), // *=, reduce: LogicalAndTerm
			reduce(156), // /=, reduce: LogicalAndTerm
			reduce(156), // //=, reduce: LogicalAndTerm
			reduce(156), // %=, reduce: LogicalAndTerm
			reduce(156), // **=, reduce: LogicalAndTerm
			reduce(156), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(156), // ||, reduce: LogicalAndTerm
			reduce(156), // ^^, reduce: LogicalAndTerm
			reduce(156), // &&, reduce: LogicalAndTerm
			shift(712),  // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S86
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(158), // $, reduce: AbsentCoalesceTerm
			nil,         // empty
			reduce(158), // ;, reduce: AbsentCoalesceTerm
			reduce(158), // {, reduce: AbsentCoalesceTerm
			nil,         // }
			reduce(158), // =, reduce: AbsentCoalesceTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(158), // ||=, reduce: AbsentCoalesceTerm
			reduce(158), // ^^=, reduce: AbsentCoalesceTerm
			reduce(158), // &&=, reduce: AbsentCoalesceTerm
			reduce(158), // ??=, reduce: AbsentCoalesceTerm
			reduce(158), // ???=, reduce: AbsentCoalesceTerm
			reduce(158), // |=, reduce: AbsentCoalesceTerm
			reduce(158), // &=, reduce: AbsentCoalesceTerm
			reduce(158), // ^=, reduce: AbsentCoalesceTerm
			reduce(158), // <<=, reduce: AbsentCoalesceTerm
			reduce(158), // >>=, reduce: AbsentCoalesceTerm
			reduce(158), // >>>=, reduce: AbsentCoalesceTerm
			reduce(158), // +=, reduce: AbsentCoalesceTerm
			reduce(158), // .=, reduce: AbsentCoalesceTerm
			reduce(158), // -=, reduce: AbsentCoalesceTerm
			reduce(158), // *=, reduce: AbsentCoalesceTerm
			reduce(158), // /=, reduce: AbsentCoalesceTerm
			reduce(158), // //=, reduce: AbsentCoalesceTerm
			reduce(158), // %=, reduce: AbsentCoalesceTerm
			reduce(158), // **=, reduce: AbsentCoalesceTerm
			reduce(158), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(158), // ||, reduce: AbsentCoalesceTerm
			reduce(158), // ^^, reduce: AbsentCoalesceTerm
			reduce(158), // &&, reduce: AbsentCoalesceTerm
			reduce(158), // ??, reduce: AbsentCoalesceTerm
			shift(713),  // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S87
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(160), // $, reduce: EmptyCoalesceTerm
			nil,         // empty
			reduce(160), // ;, reduce: EmptyCoalesceTerm
			reduce(160), // {, reduce: EmptyCoalesceTerm
			nil,         // }
			reduce(160), // =, reduce: EmptyCoalesceTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(160), // ||=, reduce: EmptyCoalesceTerm
			reduce(160), // ^^=, reduce: EmptyCoalesceTerm
			reduce(160), // &&=, reduce: EmptyCoalesceTerm
			reduce(160), // ??=, reduce: EmptyCoalesceTerm
			reduce(160), // ???=, reduce: EmptyCoalesceTerm
			reduce(160), // |=, reduce: EmptyCoalesceTerm
			reduce(160), // &=, reduce: EmptyCoalesceTerm
			reduce(160), // ^=, reduce: EmptyCoalesceTerm
			reduce(160), // <<=, reduce: EmptyCoalesceTerm
			reduce(160), // >>=, reduce: EmptyCoalesceTerm
			reduce(160), // >>>=, reduce: EmptyCoalesceTerm
			reduce(160), // +=, reduce: EmptyCoalesceTerm
			reduce(160), // .=, reduce: EmptyCoalesceTerm
			reduce(160), // -=, reduce: EmptyCoalesceTerm
			reduce(160), // *=, reduce: EmptyCoalesceTerm
			reduce(160), // /=, reduce: EmptyCoalesceTerm
			reduce(160), // //=, reduce: EmptyCoalesceTerm
			reduce(160), // %=, reduce: EmptyCoalesceTerm
			reduce(160), // **=, reduce: EmptyCoalesceTerm
			reduce(160), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(160), // ||, reduce: EmptyCoalesceTerm
			reduce(160), // ^^, reduce: EmptyCoalesceTerm
			reduce(160), // &&, reduce: EmptyCoalesceTerm
			reduce(160), // ??, reduce: EmptyCoalesceTerm
			reduce(160), // ???, reduce: EmptyCoalesceTerm
			shift(714),  // =~
			shift(715),  // !=~
			shift(716),  // ==
			shift(717),  // !=
			shift(718),  // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S88
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(166), // $, reduce: EqneTerm
			nil,         // empty
			reduce(166), // ;, reduce: EqneTerm
			reduce(166), // {, reduce: EqneTerm
			nil,         // }
			reduce(166), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(719),  // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(166), // ||=, reduce: EqneTerm
			reduce(166), // ^^=, reduce: EqneTerm
			reduce(166), // &&=, reduce: EqneTerm
			reduce(166), // ??=, reduce: EqneTerm
			reduce(166), // ???=, reduce: EqneTerm
			reduce(166), // |=, reduce: EqneTerm
			reduce(166), // &=, reduce: EqneTerm
			reduce(166), // ^=, reduce: EqneTerm
			reduce(166), // <<=, reduce: EqneTerm
			reduce(166), // >>=, reduce: EqneTerm
			reduce(166), // >>>=, reduce: EqneTerm
			reduce(166), // +=, reduce: EqneTerm
			reduce(166), // .=, reduce: EqneTerm
			reduce(166), // -=, reduce: EqneTerm
			reduce(166), // *=, reduce: EqneTerm
			reduce(166), // /=, reduce: EqneTerm
			reduce(166), // //=, reduce: EqneTerm
			reduce(166), // %=, reduce: EqneTerm
			reduce(166), // **=, reduce: EqneTerm
			reduce(166), // ?, reduce: EqneTerm
			nil,         // :
			reduce(166), // ||, reduce: EqneTerm
			reduce(166), // ^^, reduce: EqneTerm
			reduce(166), // &&, reduce: EqneTerm
			reduce(166), // ??, reduce: EqneTerm
			reduce(166), // ???, reduce: EqneTerm
			reduce(166), // =~, reduce: EqneTerm
			reduce(166), // !=~, reduce: EqneTerm
			reduce(166), // ==, reduce: EqneTerm
			reduce(166), // !=, reduce: EqneTerm
			reduce(166), // <=>, reduce: EqneTerm
			shift(720),  // >=
			shift(721),  // <
			shift(722),  // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S89
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(171), // $, reduce: CmpTerm
			nil,         // empty
			reduce(171), // ;, reduce: CmpTerm
			reduce(171), // {, reduce: CmpTerm
			nil,         // }
			reduce(171), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(171), // >, reduce: CmpTerm
			nil,         // >>
			shift(723),  // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(171), // ||=, reduce: CmpTerm
			reduce(171), // ^^=, reduce: CmpTerm
			reduce(171), // &&=, reduce: CmpTerm
			reduce(171), // ??=, reduce: CmpTerm
			reduce(171), // ???=, reduce: CmpTerm
			reduce(171), // |=, reduce: CmpTerm
			reduce(171), // &=, reduce: CmpTerm
			reduce(171), // ^=, reduce: CmpTerm
			reduce(171), // <<=, reduce: CmpTerm
			reduce(171), // >>=, reduce: CmpTerm
			reduce(171), // >>>=, reduce: CmpTerm
			reduce(171), // +=, reduce: CmpTerm
			reduce(171), // .=, reduce: CmpTerm
			reduce(171), // -=, reduce: CmpTerm
			reduce(171), // *=, reduce: CmpTerm
			reduce(171), // /=, reduce: CmpTerm
			reduce(171), // //=, reduce: CmpTerm
			reduce(171), // %=, reduce: CmpTerm
			reduce(171), // **=, reduce: CmpTerm
			reduce(171), // ?, reduce: CmpTerm
			nil,         // :
			reduce(171), // ||, reduce: CmpTerm
			reduce(171), // ^^, reduce: CmpTerm
			reduce(171), // &&, reduce: CmpTerm
			reduce(171), // ??, reduce: CmpTerm
			reduce(171), // ???, reduce: CmpTerm
			reduce(171), // =~, reduce: CmpTerm
			reduce(171), // !=~, reduce: CmpTerm
			reduce(171), // ==, reduce: CmpTerm
			reduce(171), // !=, reduce: CmpTerm
			reduce(171), // <=>, reduce: CmpTerm
			reduce(171), // >=, reduce: CmpTerm
			reduce(171), // <, reduce: CmpTerm
			reduce(171), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S90
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(173), // $, reduce: BitwiseORTerm
			nil,         // empty
			reduce(173), // ;, reduce: BitwiseORTerm
			reduce(173), // {, reduce: BitwiseORTerm
			nil,         // }
			reduce(173), // =, reduce: BitwiseORTerm
			nil,         // unset
			nil,         // filter
			reduce(173), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(173), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(173), // ||=, reduce: BitwiseORTerm
			reduce(173), // ^^=, reduce: BitwiseORTerm
			reduce(173), // &&=, reduce: BitwiseORTerm
			reduce(173), // ??=, reduce: BitwiseORTerm
			reduce(173), // ???=, reduce: BitwiseORTerm
			reduce(173), // |=, reduce: BitwiseORTerm
			reduce(173), // &=, reduce: BitwiseORTerm
			reduce(173), // ^=, reduce: BitwiseORTerm
			reduce(173), // <<=, reduce: BitwiseORTerm
			reduce(173), // >>=, reduce: BitwiseORTerm
			reduce(173), // >>>=, reduce: BitwiseORTerm
			reduce(173), // +=, reduce: BitwiseORTerm
			reduce(173), // .=, reduce: BitwiseORTerm
			reduce(173), // -=, reduce: BitwiseORTerm
			reduce(173), // *=, reduce: BitwiseORTerm
			reduce(173), // /=, reduce: BitwiseORTerm
			reduce(173), // //=, reduce: BitwiseORTerm
			reduce(173), // %=, reduce: BitwiseORTerm
			reduce(173), // **=, reduce: BitwiseORTerm
			reduce(173), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(173), // ||, reduce: BitwiseORTerm
			reduce(173), // ^^, reduce: BitwiseORTerm
			reduce(173), // &&, reduce: BitwiseORTerm
			reduce(173), // ??, reduce: BitwiseORTerm
			reduce(173), // ???, reduce: BitwiseORTerm
			reduce(173), // =~, reduce: BitwiseORTerm
			reduce(173), // !=~, reduce: BitwiseORTerm
			reduce(173), // ==, reduce: BitwiseORTerm
			reduce(173), // !=, reduce: BitwiseORTerm
			reduce(173), // <=>, reduce: BitwiseORTerm
			reduce(173), // >=, reduce: BitwiseORTerm
			reduce(173), // <, reduce: BitwiseORTerm
			reduce(173), // <=, reduce: BitwiseORTerm
			shift(724),  // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S91
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(175), // $, reduce: BitwiseXORTerm
			nil,         // empty
			reduce(175), // ;, reduce: BitwiseXORTerm
			reduce(175), // {, reduce: BitwiseXORTerm
			nil,         // }
			reduce(175), // =, reduce: BitwiseXORTerm
			nil,         // unset
			nil,         // filter
			reduce(175), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(175), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(175), // ||=, reduce: BitwiseXORTerm
			reduce(175), // ^^=, reduce: BitwiseXORTerm
			reduce(175), // &&=, reduce: BitwiseXORTerm
			reduce(175), // ??=, reduce: BitwiseXORTerm
			reduce(175), // ???=, reduce: BitwiseXORTerm
			reduce(175), // |=, reduce: BitwiseXORTerm
			reduce(175), // &=, reduce: BitwiseXORTerm
			reduce(175), // ^=, reduce: BitwiseXORTerm
			reduce(175), // <<=, reduce: BitwiseXORTerm
			reduce(175), // >>=, reduce: BitwiseXORTerm
			reduce(175), // >>>=, reduce: BitwiseXORTerm
			reduce(175), // +=, reduce: BitwiseXORTerm
			reduce(175), // .=, reduce: BitwiseXORTerm
			reduce(175), // -=, reduce: BitwiseXORTerm
			reduce(175), // *=, reduce: BitwiseXORTerm
			reduce(175), // /=, reduce: BitwiseXORTerm
			reduce(175), // //=, reduce: BitwiseXORTerm
			reduce(175), // %=, reduce: BitwiseXORTerm
			reduce(175), // **=, reduce: BitwiseXORTerm
			reduce(175), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(175), // ||, reduce: BitwiseXORTerm
			reduce(175), // ^^, reduce: BitwiseXORTerm
			reduce(175), // &&, reduce: BitwiseXORTerm
			reduce(175), // ??, reduce: BitwiseXORTerm
			reduce(175), // ???, reduce: BitwiseXORTerm
			reduce(175), // =~, reduce: BitwiseXORTerm
			reduce(175), // !=~, reduce: BitwiseXORTerm
			reduce(175), // ==, reduce: BitwiseXORTerm
			reduce(175), // !=, reduce: BitwiseXORTerm
			reduce(175), // <=>, reduce: BitwiseXORTerm
			reduce(175), // >=, reduce: BitwiseXORTerm
			reduce(175), // <, reduce: BitwiseXORTerm
			reduce(175), // <=, reduce: BitwiseXORTerm
			reduce(175), // ^, reduce: BitwiseXORTerm
			shift(725),  // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S92
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(177), // $, reduce: BitwiseANDTerm
			nil,         // empty
			reduce(177), // ;, reduce: BitwiseANDTerm
			reduce(177), // {, reduce: BitwiseANDTerm
			nil,         // }
			reduce(177), // =, reduce: BitwiseANDTerm
			nil,         // unset
			nil,         // filter
			reduce(177), // >, reduce: BitwiseANDTerm
			shift(726),  // >>
			reduce(177), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(177), // ||=, reduce: BitwiseANDTerm
			reduce(177), // ^^=, reduce: BitwiseANDTerm
			reduce(177), // &&=, reduce: BitwiseANDTerm
			reduce(177), // ??=, reduce: BitwiseANDTerm
			reduce(177), // ???=, reduce: BitwiseANDTerm
			reduce(177), // |=, reduce: BitwiseANDTerm
			reduce(177), // &=, reduce: BitwiseANDTerm
			reduce(177), // ^=, reduce: BitwiseANDTerm
			reduce(177), // <<=, reduce: BitwiseANDTerm
			reduce(177), // >>=, reduce: BitwiseANDTerm
			reduce(177), // >>>=, reduce: BitwiseANDTerm
			reduce(177), // +=, reduce: BitwiseANDTerm
			reduce(177), // .=, reduce: BitwiseANDTerm
			reduce(177), // -=, reduce: BitwiseANDTerm
			reduce(177), // *=, reduce: BitwiseANDTerm
			reduce(177), // /=, reduce: BitwiseANDTerm
			reduce(177), // //=, reduce: BitwiseANDTerm
			reduce(177), // %=, reduce: BitwiseANDTerm
			reduce(177), // **=, reduce: BitwiseANDTerm
			reduce(177), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(177), // ||, reduce: BitwiseANDTerm
			reduce(177), // ^^, reduce: BitwiseANDTerm
			reduce(177), // &&, reduce: BitwiseANDTerm
			reduce(177), // ??, reduce: BitwiseANDTerm
			reduce(177), // ???, reduce: BitwiseANDTerm
			reduce(177), // =~, reduce: BitwiseANDTerm
			reduce(177), // !=~, reduce: BitwiseANDTerm
			reduce(177), // ==, reduce: BitwiseANDTerm
			reduce(177), // !=, reduce: BitwiseANDTerm
			reduce(177), // <=>, reduce: BitwiseANDTerm
			reduce(177), // >=, reduce: BitwiseANDTerm
			reduce(177), // <, reduce: BitwiseANDTerm
			reduce(177), // <=, reduce: BitwiseANDTerm
			reduce(177), // ^, reduce: BitwiseANDTerm
			reduce(177), // &, reduce: BitwiseANDTerm
			shift(727),  // <<
			shift(728),  // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S93
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(181), // $, reduce: BitwiseShiftTerm
			nil,         // empty
			reduce(181), // ;, reduce: BitwiseShiftTerm
			reduce(181), // {, reduce: BitwiseShiftTerm
			nil,         // }
			reduce(181), // =, reduce: BitwiseShiftTerm
			nil,         // unset
			nil,         // filter
			reduce(181), // >, reduce: BitwiseShiftTerm
			reduce(181), // >>, reduce: BitwiseShiftTerm
			reduce(181), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(181), // ||=, reduce: BitwiseShiftTerm
			reduce(181), // ^^=, reduce: BitwiseShiftTerm
			reduce(181), // &&=, reduce: BitwiseShiftTerm
			reduce(181), // ??=, reduce: BitwiseShiftTerm
			reduce(181), // ???=, reduce: BitwiseShiftTerm
			reduce(181), // |=, reduce: BitwiseShiftTerm
			reduce(181), // &=, reduce: BitwiseShiftTerm
			reduce(181), // ^=, reduce: BitwiseShiftTerm
			reduce(181), // <<=, reduce: BitwiseShiftTerm
			reduce(181), // >>=, reduce: BitwiseShiftTerm
			reduce(181), // >>>=, reduce: BitwiseShiftTerm
			reduce(181), // +=, reduce: BitwiseShiftTerm
			reduce(181), // .=, reduce: BitwiseShiftTerm
			reduce(181), // -=, reduce: BitwiseShiftTerm
			reduce(181), // *=, reduce: BitwiseShiftTerm
			reduce(181), // /=, reduce: BitwiseShiftTerm
			reduce(181), // //=, reduce: BitwiseShiftTerm
			reduce(181), // %=, reduce: BitwiseShiftTerm
			reduce(181), // **=, reduce: BitwiseShiftTerm
			reduce(181), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(181), // ||, reduce: BitwiseShiftTerm
			reduce(181), // ^^, reduce: BitwiseShiftTerm
			reduce(181), // &&, reduce: BitwiseShiftTerm
			reduce(181), // ??, reduce: BitwiseShiftTerm
			reduce(181), // ???, reduce: BitwiseShiftTerm
			reduce(181), // =~, reduce: BitwiseShiftTerm
			reduce(181), // !=~, reduce: BitwiseShiftTerm
			reduce(181), // ==, reduce: BitwiseShiftTerm
			reduce(181), // !=, reduce: BitwiseShiftTerm
			reduce(181), // <=>, reduce: BitwiseShiftTerm
			reduce(181), // >=, reduce: BitwiseShiftTerm
			reduce(181), // <, reduce: BitwiseShiftTerm
			reduce(181), // <=, reduce: BitwiseShiftTerm
			reduce(181), // ^, reduce: BitwiseShiftTerm
			reduce(181), // &, reduce: BitwiseShiftTerm
			reduce(181), // <<, reduce: BitwiseShiftTerm
			reduce(181), // >>>, reduce: BitwiseShiftTerm
			shift(729),  // +
			shift(730),  // -
			shift(731),  // .+
			shift(732),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S94
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S95
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(186), // $, reduce: AddsubdotTerm
			nil,         // empty
			reduce(186), // ;, reduce: AddsubdotTerm
			reduce(186), // {, reduce: AddsubdotTerm
			nil,         // }
			reduce(186), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(186), // >, reduce: AddsubdotTerm
			reduce(186), // >>, reduce: AddsubdotTerm
			reduce(186), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(186), // ||=, reduce: AddsubdotTerm
			reduce(186), // ^^=, reduce: AddsubdotTerm
			reduce(186), // &&=, reduce: AddsubdotTerm
			reduce(186), // ??=, reduce: AddsubdotTerm
			reduce(186), // ???=, reduce: AddsubdotTerm
			reduce(186), // |=, reduce: AddsubdotTerm
			reduce(186), // &=, reduce: AddsubdotTerm
			reduce(186), // ^=, reduce: AddsubdotTerm
			reduce(186), // <<=, reduce: AddsubdotTerm
			reduce(186), // >>=, reduce: AddsubdotTerm
			reduce(186), // >>>=, reduce: AddsubdotTerm
			reduce(186), // +=, reduce: AddsubdotTerm
			reduce(186), // .=, reduce: AddsubdotTerm
			reduce(186), // -=, reduce: AddsubdotTerm
			reduce(186), // *=, reduce: AddsubdotTerm
			reduce(186), // /=, reduce: AddsubdotTerm
			reduce(186), // //=, reduce: AddsubdotTerm
			reduce(186), // %=, reduce: AddsubdotTerm
			reduce(186), // **=, reduce: AddsubdotTerm
			reduce(186), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(186), // ||, reduce: AddsubdotTerm
			reduce(186), // ^^, reduce: AddsubdotTerm
			reduce(186), // &&, reduce: AddsubdotTerm
			reduce(186), // ??, reduce: AddsubdotTerm
			reduce(186), // ???, reduce: AddsubdotTerm
			reduce(186), // =~, reduce: AddsubdotTerm
			reduce(186), // !=~, reduce: AddsubdotTerm
			reduce(186), // ==, reduce: AddsubdotTerm
			reduce(186), // !=, reduce: AddsubdotTerm
			reduce(186), // <=>, reduce: AddsubdotTerm
			reduce(186), // >=, reduce: AddsubdotTerm
			reduce(186), // <, reduce: AddsubdotTerm
			reduce(186), // <=, reduce: AddsubdotTerm
			reduce(186), // ^, reduce: AddsubdotTerm
			reduce(186), // &, reduce: AddsubdotTerm
			reduce(186), // <<, reduce: AddsubdotTerm
			reduce(186), // >>>, reduce: AddsubdotTerm
			reduce(186), // +, reduce: AddsubdotTerm
			reduce(186), // -, reduce: AddsubdotTerm
			reduce(186), // .+, reduce: AddsubdotTerm
			reduce(186), // .-, reduce: AddsubdotTerm
			shift(735),  // *
			shift(736),  // /
			shift(737),  // //
			shift(738),  // %
			shift(739),  // .*
			shift(740),  // ./
			shift(741),  // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S96
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S97
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S98
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S99
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(194), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(194), // ;, reduce: MuldivTerm
			reduce(194), // {, reduce: MuldivTerm
			nil,         // }
			reduce(194), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(194), // >, reduce: MuldivTerm
			reduce(194), // >>, reduce: MuldivTerm
			reduce(194), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(194), // ||=, reduce: MuldivTerm
			reduce(194), // ^^=, reduce: MuldivTerm
			reduce(194), // &&=, reduce: MuldivTerm
			reduce(194), // ??=, reduce: MuldivTerm
			reduce(194), // ???=, reduce: MuldivTerm
			reduce(194), // |=, reduce: MuldivTerm
			reduce(194), // &=, reduce: MuldivTerm
			reduce(194), // ^=, reduce: MuldivTerm
			reduce(194), // <<=, reduce: MuldivTerm
			reduce(194), // >>=, reduce: MuldivTerm
			reduce(194), // >>>=, reduce: MuldivTerm
			reduce(194), // +=, reduce: MuldivTerm
			reduce(194), // .=, reduce: MuldivTerm
			reduce(194), // -=, reduce: MuldivTerm
			reduce(194), // *=, reduce: MuldivTerm
			reduce(194), // /=, reduce: MuldivTerm
			reduce(194), // //=, reduce: MuldivTerm
			reduce(194), // %=, reduce: MuldivTerm
			reduce(194), // **=, reduce: MuldivTerm
			reduce(194), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(194), // ||, reduce: MuldivTerm
			reduce(194), // ^^, reduce: MuldivTerm
			reduce(194), // &&, reduce: MuldivTerm
			reduce(194), // ??, reduce: MuldivTerm
			reduce(194), // ???, reduce: MuldivTerm
			reduce(194), // =~, reduce: MuldivTerm
			reduce(194), // !=~, reduce: MuldivTerm
			reduce(194), // ==, reduce: MuldivTerm
			reduce(194), // !=, reduce: MuldivTerm
			reduce(194), // <=>, reduce: MuldivTerm
			reduce(194), // >=, reduce: MuldivTerm
			reduce(194), // <, reduce: MuldivTerm
			reduce(194), // <=, reduce: MuldivTerm
			reduce(194), // ^, reduce: MuldivTerm
			reduce(194), // &, reduce: MuldivTerm
			reduce(194), // <<, reduce: MuldivTerm
			reduce(194), // >>>, reduce: MuldivTerm
			reduce(194), // +, reduce: MuldivTerm
			reduce(194), // -, reduce: MuldivTerm
			reduce(194), // .+, reduce: MuldivTerm
			reduce(194), // .-, reduce: MuldivTerm
			reduce(194), // *, reduce: MuldivTerm
			reduce(194), // /, reduce: MuldivTerm
			reduce(194), // //, reduce: MuldivTerm
			reduce(194), // %, reduce: MuldivTerm
			reduce(194), // .*, reduce: MuldivTerm
			reduce(194), // ./, reduce: MuldivTerm
			reduce(194), // .//, reduce: MuldivTerm
			shift(745),  // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S100
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(196), // $, reduce: DotTerm
			nil,         // empty
			reduce(196), // ;, reduce: DotTerm
			reduce(196), // {, reduce: DotTerm
			nil,         // }
			reduce(196), // =, reduce: DotTerm
			nil,         // unset
			nil,         // filter
			reduce(196), // >, reduce: DotTerm
			reduce(196), // >>, reduce: DotTerm
			reduce(196), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(196), // ||=, reduce: DotTerm
			reduce(196), // ^^=, reduce: DotTerm
			reduce(196), // &&=, reduce: DotTerm
			reduce(196), // ??=, reduce: DotTerm
			reduce(196), // ???=, reduce: DotTerm
			reduce(196), // |=, reduce: DotTerm
			reduce(196), // &=, reduce: DotTerm
			reduce(196), // ^=, reduce: DotTerm
			reduce(196), // <<=, reduce: DotTerm
			reduce(196), // >>=, reduce: DotTerm
			reduce(196), // >>>=, reduce: DotTerm
			reduce(196), // +=, reduce: DotTerm
			reduce(196), // .=, reduce: DotTerm
			reduce(196), // -=, reduce: DotTerm
			reduce(196), // *=, reduce: DotTerm
			reduce(196), // /=, reduce: DotTerm
			reduce(196), // //=, reduce: DotTerm
			reduce(196), // %=, reduce: DotTerm
			reduce(196), // **=, reduce: DotTerm
			reduce(196), // ?, reduce: DotTerm
			nil,         // :
			reduce(196), // ||, reduce: DotTerm
			reduce(196), // ^^, reduce: DotTerm
			reduce(196), // &&, reduce: DotTerm
			reduce(196), // ??, reduce: DotTerm
			reduce(196), // ???, reduce: DotTerm
			reduce(196), // =~, reduce: DotTerm
			reduce(196), // !=~, reduce: DotTerm
			reduce(196), // ==, reduce: DotTerm
			reduce(196), // !=, reduce: DotTerm
			reduce(196), // <=>, reduce: DotTerm
			reduce(196), // >=, reduce: DotTerm
			reduce(196), // <, reduce: DotTerm
			reduce(196), // <=, reduce: DotTerm
			reduce(196), // ^, reduce: DotTerm
			reduce(196), // &, reduce: DotTerm
			reduce(196), // <<, reduce: DotTerm
			reduce(196), // >>>, reduce: DotTerm
			reduce(196), // +, reduce: DotTerm
			reduce(196), // -, reduce: DotTerm
			reduce(196), // .+, reduce: DotTerm
			reduce(196), // .-, reduce: DotTerm
			reduce(196), // *, reduce: DotTerm
			reduce(196), // /, reduce: DotTerm
			reduce(196), // //, reduce: DotTerm
			reduce(196), // %, reduce: DotTerm
			reduce(196), // .*, reduce: DotTerm
			reduce(196), // ./, reduce: DotTerm
			reduce(196), // .//, reduce: DotTerm
			reduce(196), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S101
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S102
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S103
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(203), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(203), // ;, reduce: UnaryOpTerm
			reduce(203), // {, reduce: UnaryOpTerm
			nil,         // }
			reduce(203), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(203), // >, reduce: UnaryOpTerm
			reduce(203), // >>, reduce: UnaryOpTerm
			reduce(203), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(203), // ||=, reduce: UnaryOpTerm
			reduce(203), // ^^=, reduce: UnaryOpTerm
			reduce(203), // &&=, reduce: UnaryOpTerm
			reduce(203), // ??=, reduce: UnaryOpTerm
			reduce(203), // ???=, reduce: UnaryOpTerm
			reduce(203), // |=, reduce: UnaryOpTerm
			reduce(203), // &=, reduce: UnaryOpTerm
			reduce(203), // ^=, reduce: UnaryOpTerm
			reduce(203), // <<=, reduce: UnaryOpTerm
			reduce(203), // >>=, reduce: UnaryOpTerm
			reduce(203), // >>>=, reduce: UnaryOpTerm
			reduce(203), // +=, reduce: UnaryOpTerm
			reduce(203), // .=, reduce: UnaryOpTerm
			reduce(203), // -=, reduce: UnaryOpTerm
			reduce(203), // *=, reduce: UnaryOpTerm
			reduce(203), // /=, reduce: UnaryOpTerm
			reduce(203), // //=, reduce: UnaryOpTerm
			reduce(203), // %=, reduce: UnaryOpTerm
			reduce(203), // **=, reduce: UnaryOpTerm
			reduce(203), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(203), // ||, reduce: UnaryOpTerm
			reduce(203), // ^^, reduce: UnaryOpTerm
			reduce(203), // &&, reduce: UnaryOpTerm
			reduce(203), // ??, reduce: UnaryOpTerm
			reduce(203), // ???, reduce: UnaryOpTerm
			reduce(203), // =~, reduce: UnaryOpTerm
			reduce(203), // !=~, reduce: UnaryOpTerm
			reduce(203), // ==, reduce: UnaryOpTerm
			reduce(203), // !=, reduce: UnaryOpTerm
			reduce(203), // <=>, reduce: UnaryOpTerm
			reduce(203), // >=, reduce: UnaryOpTerm
			reduce(203), // <, reduce: UnaryOpTerm
			reduce(203), // <=, reduce: UnaryOpTerm
			reduce(203), // ^, reduce: UnaryOpTerm
			reduce(203), // &, reduce: UnaryOpTerm
			reduce(203), // <<, reduce: UnaryOpTerm
			reduce(203), // >>>, reduce: UnaryOpTerm
			reduce(203), // +, reduce: UnaryOpTerm
			reduce(203), // -, reduce: UnaryOpTerm
			reduce(203), // .+, reduce: UnaryOpTerm
			reduce(203), // .-, reduce: UnaryOpTerm
			reduce(203), // *, reduce: UnaryOpTerm
			reduce(203), // /, reduce: UnaryOpTerm
			reduce(203), // //, reduce: UnaryOpTerm
			reduce(203), // %, reduce: UnaryOpTerm
			reduce(203), // .*, reduce: UnaryOpTerm
			reduce(203), // ./, reduce: UnaryOpTerm
			reduce(203), // .//, reduce: UnaryOpTerm
			reduce(203), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S104
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(207), // $, reduce: PowTerm
			nil,         // empty
			reduce(207), // ;, reduce: PowTerm
			reduce(207), // {, reduce: PowTerm
			nil,         // }
			reduce(207), // =, reduce: PowTerm
			nil,         // unset
			nil,         // filter
			reduce(207), // >, reduce: PowTerm
			reduce(207), // >>, reduce: PowTerm
			reduce(207), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(207), // ||=, reduce: PowTerm
			reduce(207), // ^^=, reduce: PowTerm
			reduce(207), // &&=, reduce: PowTerm
			reduce(207), // ??=, reduce: PowTerm
			reduce(207), // ???=, reduce: PowTerm
			reduce(207), // |=, reduce: PowTerm
			reduce(207), // &=, reduce: PowTerm
			reduce(207), // ^=, reduce: PowTerm
			reduce(207), // <<=, reduce: PowTerm
			reduce(207), // >>=, reduce: PowTerm
			reduce(207), // >>>=, reduce: PowTerm
			reduce(207), // +=, reduce: PowTerm
			reduce(207), // .=, reduce: PowTerm
			reduce(207), // -=, reduce: PowTerm
			reduce(207), // *=, reduce: PowTerm
			reduce(207), // /=, reduce: PowTerm
			reduce(207), // //=, reduce: PowTerm
			reduce(207), // %=, reduce: PowTerm
			reduce(207), // **=, reduce: PowTerm
			reduce(207), // ?, reduce: PowTerm
			nil,         // :
			reduce(207), // ||, reduce: PowTerm
			reduce(207), // ^^, reduce: PowTerm
			reduce(207), // &&, reduce: PowTerm
			reduce(207), // ??, reduce: PowTerm
			reduce(207), // ???, reduce: PowTerm
			reduce(207), // =~, reduce: PowTerm
			reduce(207), // !=~, reduce: PowTerm
			reduce(207), // ==, reduce: PowTerm
			reduce(207), // !=, reduce: PowTerm
			reduce(207), // <=>, reduce: PowTerm
			reduce(207), // >=, reduce: PowTerm
			reduce(207), // <, reduce: PowTerm
			reduce(207), // <=, reduce: PowTerm
			reduce(207), // ^, reduce: PowTerm
			reduce(207), // &, reduce: PowTerm
			reduce(207), // <<, reduce: PowTerm
			reduce(207), // >>>, reduce: PowTerm
			reduce(207), // +, reduce: PowTerm
			reduce(207), // -, reduce: PowTerm
			reduce(207), // .+, reduce: PowTerm
			reduce(207), // .-, reduce: PowTerm
			reduce(207), // *, reduce: PowTerm
			reduce(207), // /, reduce: PowTerm
			reduce(207), // //, reduce: PowTerm
			reduce(207), // %, reduce: PowTerm
			reduce(207), // .*, reduce: PowTerm
			reduce(207), // ./, reduce: PowTerm
			reduce(207), // .//, reduce: PowTerm
			reduce(207), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			shift(748),  // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S105
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(209), // $, reduce: PrecedenceChainEnd
			nil,         // empty
			reduce(209), // ;, reduce: PrecedenceChainEnd
			reduce(209), // {, reduce: PrecedenceChainEnd
			nil,         // }
			reduce(209), // =, reduce: PrecedenceChainEnd
			nil,         // unset
			nil,         // filter
			reduce(209), // >, reduce: PrecedenceChainEnd
			reduce(209), // >>, reduce: PrecedenceChainEnd
			reduce(209), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(209), // ||=, reduce: PrecedenceChainEnd
			reduce(209), // ^^=, reduce: PrecedenceChainEnd
			reduce(209), // &&=, reduce: PrecedenceChainEnd
			reduce(209), // ??=, reduce: PrecedenceChainEnd
			reduce(209), // ???=, reduce: PrecedenceChainEnd
			reduce(209), // |=, reduce: PrecedenceChainEnd
			reduce(209), // &=, reduce: PrecedenceChainEnd
			reduce(209), // ^=, reduce: PrecedenceChainEnd
			reduce(209), // <<=, reduce: PrecedenceChainEnd
			reduce(209), // >>=, reduce: PrecedenceChainEnd
			reduce(209), // >>>=, reduce: PrecedenceChainEnd
			reduce(209), // +=, reduce: PrecedenceChainEnd
			reduce(209), // .=, reduce: PrecedenceChainEnd
			reduce(209), // -=, reduce: PrecedenceChainEnd
			reduce(209), // *=, reduce: PrecedenceChainEnd
			reduce(209), // /=, reduce: PrecedenceChainEnd
			reduce(209), // //=, reduce: PrecedenceChainEnd
			reduce(209), // %=, reduce: PrecedenceChainEnd
			reduce(209), // **=, reduce: PrecedenceChainEnd
			reduce(209), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(209), // ||, reduce: PrecedenceChainEnd
			reduce(209), // ^^, reduce: PrecedenceChainEnd
			reduce(209), // &&, reduce: PrecedenceChainEnd
			reduce(209), // ??, reduce: PrecedenceChainEnd
			reduce(209), // ???, reduce: PrecedenceChainEnd
			reduce(209), // =~, reduce: PrecedenceChainEnd
			reduce(209), // !=~, reduce: PrecedenceChainEnd
			reduce(209), // ==, reduce: PrecedenceChainEnd
			reduce(209), // !=, reduce: PrecedenceChainEnd
			reduce(209), // <=>, reduce: PrecedenceChainEnd
			reduce(209), // >=, reduce: PrecedenceChainEnd
			reduce(209), // <, reduce: PrecedenceChainEnd
			reduce(209), // <=, reduce: PrecedenceChainEnd
			reduce(209), // ^, reduce: PrecedenceChainEnd
			reduce(209), // &, reduce: PrecedenceChainEnd
			reduce(209), // <<, reduce: PrecedenceChainEnd
			reduce(209), // >>>, reduce: PrecedenceChainEnd
			reduce(209), // +, reduce: PrecedenceChainEnd
			reduce(209), // -, reduce: PrecedenceChainEnd
			reduce(209), // .+, reduce: PrecedenceChainEnd
			reduce(209), // .-, reduce: PrecedenceChainEnd
			reduce(209), // *, reduce: PrecedenceChainEnd
			reduce(209), // /, reduce: PrecedenceChainEnd
			reduce(209), // //, reduce: PrecedenceChainEnd
			reduce(209), // %, reduce: PrecedenceChainEnd
			reduce(209), // .*, reduce: PrecedenceChainEnd
			reduce(209), // ./, reduce: PrecedenceChainEnd
			reduce(209), // .//, reduce: PrecedenceChainEnd
			reduce(209), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(209), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(749),  // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			shift(750),  // [[
			shift(751),  // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S106
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(215), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(215), // ;, reduce: MlrvalOrFunction
			reduce(215), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(215), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(215), // >, reduce: MlrvalOrFunction
			reduce(215), // >>, reduce: MlrvalOrFunction
			reduce(215), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(215), // ||=, reduce: MlrvalOrFunction
			reduce(215), // ^^=, reduce: MlrvalOrFunction
			reduce(215), // &&=, reduce: MlrvalOrFunction
			reduce(215), // ??=, reduce: MlrvalOrFunction
			reduce(215), // ???=, reduce: MlrvalOrFunction
			reduce(215), // |=, reduce: MlrvalOrFunction
			reduce(215), // &=, reduce: MlrvalOrFunction
			reduce(215), // ^=, reduce: MlrvalOrFunction
			reduce(215), // <<=, reduce: MlrvalOrFunction
			reduce(215), // >>=, reduce: MlrvalOrFunction
			reduce(215), // >>>=, reduce: MlrvalOrFunction
			reduce(215), // +=, reduce: MlrvalOrFunction
			reduce(215), // .=, reduce: MlrvalOrFunction
			reduce(215), // -=, reduce: MlrvalOrFunction
			reduce(215), // *=, reduce: MlrvalOrFunction
			reduce(215), // /=, reduce: MlrvalOrFunction
			reduce(215), // //=, reduce: MlrvalOrFunction
			reduce(215), // %=, reduce: MlrvalOrFunction
			reduce(215), // **=, reduce: MlrvalOrFunction
			reduce(215), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(215), // ||, reduce: MlrvalOrFunction
			reduce(215), // ^^, reduce: MlrvalOrFunction
			reduce(215), // &&, reduce: MlrvalOrFunction
			reduce(215), // ??, reduce: MlrvalOrFunction
			reduce(215), // ???, reduce: MlrvalOrFunction
			reduce(215), // =~, reduce: MlrvalOrFunction
			reduce(215), // !=~, reduce: MlrvalOrFunction
			reduce(215), // ==, reduce: MlrvalOrFunction
			reduce(215), // !=, reduce: MlrvalOrFunction
			reduce(215), // <=>, reduce: MlrvalOrFunction
			reduce(215), // >=, reduce: MlrvalOrFunction
			reduce(215), // <, reduce: MlrvalOrFunction
			reduce(215), // <=, reduce: MlrvalOrFunction
			reduce(215), // ^, reduce: MlrvalOrFunction
			reduce(215), // &, reduce: MlrvalOrFunction
			reduce(215), // <<, reduce: MlrvalOrFunction
			reduce(215), // >>>, reduce: MlrvalOrFunction
			reduce(215), // +, reduce: MlrvalOrFunction
			reduce(215), // -, reduce: MlrvalOrFunction
			reduce(215), // .+, reduce: MlrvalOrFunction
			reduce(215), // .-, reduce: MlrvalOrFunction
			reduce(215), // *, reduce: MlrvalOrFunction
			reduce(215), // /, reduce: MlrvalOrFunction
			reduce(215), // //, reduce: MlrvalOrFunction
			reduce(215), // %, reduce: MlrvalOrFunction
			reduce(215), // .*, reduce: MlrvalOrFunction
			reduce(215), // ./, reduce: MlrvalOrFunction
			reduce(215), // .//, reduce: MlrvalOrFunction
			reduce(215), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(215), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(215), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(215), // [[, reduce: MlrvalOrFunction
			reduce(215), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S107
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(216), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(216), // ;, reduce: MlrvalOrFunction
			reduce(216), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(216), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(216), // >, reduce: MlrvalOrFunction
			reduce(216), // >>, reduce: MlrvalOrFunction
			reduce(216), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(216), // ||=, reduce: MlrvalOrFunction
			reduce(216), // ^^=, reduce: MlrvalOrFunction
			reduce(216), // &&=, reduce: MlrvalOrFunction
			reduce(216), // ??=, reduce: MlrvalOrFunction
			reduce(216), // ???=, reduce: MlrvalOrFunction
			reduce(216), // |=, reduce: MlrvalOrFunction
			reduce(216), // &=, reduce: MlrvalOrFunction
			reduce(216), // ^=, reduce: MlrvalOrFunction
			reduce(216), // <<=, reduce: MlrvalOrFunction
			reduce(216), // >>=, reduce: MlrvalOrFunction
			reduce(216), // >>>=, reduce: MlrvalOrFunction
			reduce(216), // +=, reduce: MlrvalOrFunction
			reduce(216), // .=, reduce: MlrvalOrFunction
			reduce(216), // -=, reduce: MlrvalOrFunction
			reduce(216), // *=, reduce: MlrvalOrFunction
			reduce(216), // /=, reduce: MlrvalOrFunction
			reduce(216), // //=, reduce: MlrvalOrFunction
			reduce(216), // %=, reduce: MlrvalOrFunction
			reduce(216), // **=, reduce: MlrvalOrFunction
			reduce(216), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(216), // ||, reduce: MlrvalOrFunction
			reduce(216), // ^^, reduce: MlrvalOrFunction
			reduce(216), // &&, reduce: MlrvalOrFunction
			reduce(216), // ??, reduce: MlrvalOrFunction
			reduce(216), // ???, reduce: MlrvalOrFunction
			reduce(216), // =~, reduce: MlrvalOrFunction
			reduce(216), // !=~, reduce: MlrvalOrFunction
			reduce(216), // ==, reduce: MlrvalOrFunction
			reduce(216), // !=, reduce: MlrvalOrFunction
			reduce(216), // <=>, reduce: MlrvalOrFunction
			reduce(216), // >=, reduce: MlrvalOrFunction
			reduce(216), // <, reduce: MlrvalOrFunction
			reduce(216), // <=, reduce: MlrvalOrFunction
			reduce(216), // ^, reduce: MlrvalOrFunction
			reduce(216), // &, reduce: MlrvalOrFunction
			reduce(216), // <<, reduce: MlrvalOrFunction
			reduce(216), // >>>, reduce: MlrvalOrFunction
			reduce(216), // +, reduce: MlrvalOrFunction
			reduce(216), // -, reduce: MlrvalOrFunction
			reduce(216), // .+, reduce: MlrvalOrFunction
			reduce(216), // .-, reduce: MlrvalOrFunction
			reduce(216), // *, reduce: MlrvalOrFunction
			reduce(216), // /, reduce: MlrvalOrFunction
			reduce(216), // //, reduce: MlrvalOrFunction
			reduce(216), // %, reduce: MlrvalOrFunction
			reduce(216), // .*, reduce: MlrvalOrFunction
			reduce(216), // ./, reduce: MlrvalOrFunction
			reduce(216), // .//, reduce: MlrvalOrFunction
			reduce(216), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(216), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(216), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(216), // [[, reduce: MlrvalOrFunction
			reduce(216), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S108
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(217), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(217), // ;, reduce: MlrvalOrFunction
			reduce(217), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(217), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(217), // >, reduce: MlrvalOrFunction
			reduce(217), // >>, reduce: MlrvalOrFunction
			reduce(217), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(217), // ||=, reduce: MlrvalOrFunction
			reduce(217), // ^^=, reduce: MlrvalOrFunction
			reduce(217), // &&=, reduce: MlrvalOrFunction
			reduce(217), // ??=, reduce: MlrvalOrFunction
			reduce(217), // ???=, reduce: MlrvalOrFunction
			reduce(217), // |=, reduce: MlrvalOrFunction
			reduce(217), // &=, reduce: MlrvalOrFunction
			reduce(217), // ^=, reduce: MlrvalOrFunction
			reduce(217), // <<=, reduce: MlrvalOrFunction
			reduce(217), // >>=, reduce: MlrvalOrFunction
			reduce(217), // >>>=, reduce: MlrvalOrFunction
			reduce(217), // +=, reduce: MlrvalOrFunction
			reduce(217), // .=, reduce: MlrvalOrFunction
			reduce(217), // -=, reduce: MlrvalOrFunction
			reduce(217), // *=, reduce: MlrvalOrFunction
			reduce(217), // /=, reduce: MlrvalOrFunction
			reduce(217), // //=, reduce: MlrvalOrFunction
			reduce(217), // %=, reduce: MlrvalOrFunction
			reduce(217), // **=, reduce: MlrvalOrFunction
			reduce(217), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(217), // ||, reduce: MlrvalOrFunction
			reduce(217), // ^^, reduce: MlrvalOrFunction
			reduce(217), // &&, reduce: MlrvalOrFunction
			reduce(217), // ??, reduce: MlrvalOrFunction
			reduce(217), // ???, reduce: MlrvalOrFunction
			reduce(217), // =~, reduce: MlrvalOrFunction
			reduce(217), // !=~, reduce: MlrvalOrFunction
			reduce(217), // ==, reduce: MlrvalOrFunction
			reduce(217), // !=, reduce: MlrvalOrFunction
			reduce(217), // <=>, reduce: MlrvalOrFunction
			reduce(217), // >=, reduce: MlrvalOrFunction
			reduce(217), // <, reduce: MlrvalOrFunction
			reduce(217), // <=, reduce: MlrvalOrFunction
			reduce(217), // ^, reduce: MlrvalOrFunction
			reduce(217), // &, reduce: MlrvalOrFunction
			reduce(217), // <<, reduce: MlrvalOrFunction
			reduce(217), // >>>, reduce: MlrvalOrFunction
			reduce(217), // +, reduce: MlrvalOrFunction
			reduce(217), // -, reduce: MlrvalOrFunction
			reduce(217), // .+, reduce: MlrvalOrFunction
			reduce(217), // .-, reduce: MlrvalOrFunction
			reduce(217), // *, reduce: MlrvalOrFunction
			reduce(217), // /, reduce: MlrvalOrFunction
			reduce(217), // //, reduce: MlrvalOrFunction
			reduce(217), // %, reduce: MlrvalOrFunction
			reduce(217), // .*, reduce: MlrvalOrFunction
			reduce(217), // ./, reduce: MlrvalOrFunction
			reduce(217), // .//, reduce: MlrvalOrFunction
			reduce(217), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(217), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(217), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(217), // [[, reduce: MlrvalOrFunction
			reduce(217), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S109
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(218), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(218), // ;, reduce: MlrvalOrFunction
			reduce(218), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(218), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(218), // >, reduce: MlrvalOrFunction
			reduce(218), // >>, reduce: MlrvalOrFunction
			reduce(218), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(218), // ||=, reduce: MlrvalOrFunction
			reduce(218), // ^^=, reduce: MlrvalOrFunction
			reduce(218), // &&=, reduce: MlrvalOrFunction
			reduce(218), // ??=, reduce: MlrvalOrFunction
			reduce(218), // ???=, reduce: MlrvalOrFunction
			reduce(218), // |=, reduce: MlrvalOrFunction
			reduce(218), // &=, reduce: MlrvalOrFunction
			reduce(218), // ^=, reduce: MlrvalOrFunction
			reduce(218), // <<=, reduce: MlrvalOrFunction
			reduce(218), // >>=, reduce: MlrvalOrFunction
			reduce(218), // >>>=, reduce: MlrvalOrFunction
			reduce(218), // +=, reduce: MlrvalOrFunction
			reduce(218), // .=, reduce: MlrvalOrFunction
			reduce(218), // -=, reduce: MlrvalOrFunction
			reduce(218), // *=, reduce: MlrvalOrFunction
			reduce(218), // /=, reduce: MlrvalOrFunction
			reduce(218), // //=, reduce: MlrvalOrFunction
			reduce(218), // %=, reduce: MlrvalOrFunction
			reduce(218), // **=, reduce: MlrvalOrFunction
			reduce(218), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(218), // ||, reduce: MlrvalOrFunction
			reduce(218), // ^^, reduce: MlrvalOrFunction
			reduce(218), // &&, reduce: MlrvalOrFunction
			reduce(218), // ??, reduce: MlrvalOrFunction
			reduce(218), // ???, reduce: MlrvalOrFunction
			reduce(218), // =~, reduce: MlrvalOrFunction
			reduce(218), // !=~, reduce: MlrvalOrFunction
			reduce(218), // ==, reduce: MlrvalOrFunction
			reduce(218), // !=, reduce: MlrvalOrFunction
			reduce(218), // <=>, reduce: MlrvalOrFunction
			reduce(218), // >=, reduce: MlrvalOrFunction
			reduce(218), // <, reduce: MlrvalOrFunction
			reduce(218), // <=, reduce: MlrvalOrFunction
			reduce(218), // ^, reduce: MlrvalOrFunction
			reduce(218), // &, reduce: MlrvalOrFunction
			reduce(218), // <<, reduce: MlrvalOrFunction
			reduce(218), // >>>, reduce: MlrvalOrFunction
			reduce(218), // +, reduce: MlrvalOrFunction
			reduce(218), // -, reduce: MlrvalOrFunction
			reduce(218), // .+, reduce: MlrvalOrFunction
			reduce(218), // .-, reduce: MlrvalOrFunction
			reduce(218), // *, reduce: MlrvalOrFunction
			reduce(218), // /, reduce: MlrvalOrFunction
			reduce(218), // //, reduce: MlrvalOrFunction
			reduce(218), // %, reduce: MlrvalOrFunction
			reduce(218), // .*, reduce: MlrvalOrFunction
			reduce(218), // ./, reduce: MlrvalOrFunction
			reduce(218), // .//, reduce: MlrvalOrFunction
			reduce(218), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(218), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(218), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(218), // [[, reduce: MlrvalOrFunction
			reduce(218), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S110
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(219), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(219), // ;, reduce: MlrvalOrFunction
			reduce(219), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(219), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(219), // >, reduce: MlrvalOrFunction
			reduce(219), // >>, reduce: MlrvalOrFunction
			reduce(219), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(219), // ||=, reduce: MlrvalOrFunction
			reduce(219), // ^^=, reduce: MlrvalOrFunction
			reduce(219), // &&=, reduce: MlrvalOrFunction
			reduce(219), // ??=, reduce: MlrvalOrFunction
			reduce(219), // ???=, reduce: MlrvalOrFunction
			reduce(219), // |=, reduce: MlrvalOrFunction
			reduce(219), // &=, reduce: MlrvalOrFunction
			reduce(219), // ^=, reduce: MlrvalOrFunction
			reduce(219), // <<=, reduce: MlrvalOrFunction
			reduce(219), // >>=, reduce: MlrvalOrFunction
			reduce(219), // >>>=, reduce: MlrvalOrFunction
			reduce(219), // +=, reduce: MlrvalOrFunction
			reduce(219), // .=, reduce: MlrvalOrFunction
			reduce(219), // -=, reduce: MlrvalOrFunction
			reduce(219), // *=, reduce: MlrvalOrFunction
			reduce(219), // /=, reduce: MlrvalOrFunction
			reduce(219), // //=, reduce: MlrvalOrFunction
			reduce(219), // %=, reduce: MlrvalOrFunction
			reduce(219), // **=, reduce: MlrvalOrFunction
			reduce(219), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(219), // ||, reduce: MlrvalOrFunction
			reduce(219), // ^^, reduce: MlrvalOrFunction
			reduce(219), // &&, reduce: MlrvalOrFunction
			reduce(219), // ??, reduce: MlrvalOrFunction
			reduce(219), // ???, reduce: MlrvalOrFunction
			reduce(219), // =~, reduce: MlrvalOrFunction
			reduce(219), // !=~, reduce: MlrvalOrFunction
			reduce(219), // ==, reduce: MlrvalOrFunction
			reduce(219), // !=, reduce: MlrvalOrFunction
			reduce(219), // <=>, reduce: MlrvalOrFunction
			reduce(219), // >=, reduce: MlrvalOrFunction
			reduce(219), // <, reduce: MlrvalOrFunction
			reduce(219), // <=, reduce: MlrvalOrFunction
			reduce(219), // ^, reduce: MlrvalOrFunction
			reduce(219), // &, reduce: MlrvalOrFunction
			reduce(219), // <<, reduce: MlrvalOrFunction
			reduce(219), // >>>, reduce: MlrvalOrFunction
			reduce(219), // +, reduce: MlrvalOrFunction
			reduce(219), // -, reduce: MlrvalOrFunction
			reduce(219), // .+, reduce: MlrvalOrFunction
			reduce(219), // .-, reduce: MlrvalOrFunction
			reduce(219), // *, reduce: MlrvalOrFunction
			reduce(219), // /, reduce: MlrvalOrFunction
			reduce(219), // //, reduce: MlrvalOrFunction
			reduce(219), // %, reduce: MlrvalOrFunction
			reduce(219), // .*, reduce: MlrvalOrFunction
			reduce(219), // ./, reduce: MlrvalOrFunction
			reduce(219), // .//, reduce: MlrvalOrFunction
			reduce(219), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(219), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(219), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(219), // [[, reduce: MlrvalOrFunction
			reduce(219), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S111
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(220), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(220), // ;, reduce: MlrvalOrFunction
			reduce(220), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(220), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(220), // >, reduce: MlrvalOrFunction
			reduce(220), // >>, reduce: MlrvalOrFunction
			reduce(220), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(220), // ||=, reduce: MlrvalOrFunction
			reduce(220), // ^^=, reduce: MlrvalOrFunction
			reduce(220), // &&=, reduce: MlrvalOrFunction
			reduce(220), // ??=, reduce: MlrvalOrFunction
			reduce(220), // ???=, reduce: MlrvalOrFunction
			reduce(220), // |=, reduce: MlrvalOrFunction
			reduce(220), // &=, reduce: MlrvalOrFunction
			reduce(220), // ^=, reduce: MlrvalOrFunction
			reduce(220), // <<=, reduce: MlrvalOrFunction
			reduce(220), // >>=, reduce: MlrvalOrFunction
			reduce(220), // >>>=, reduce: MlrvalOrFunction
			reduce(220), // +=, reduce: MlrvalOrFunction
			reduce(220), // .=, reduce: MlrvalOrFunction
			reduce(220), // -=, reduce: MlrvalOrFunction
			reduce(220), // *=, reduce: MlrvalOrFunction
			reduce(220), // /=, reduce: MlrvalOrFunction
			reduce(220), // //=, reduce: MlrvalOrFunction
			reduce(220), // %=, reduce: MlrvalOrFunction
			reduce(220), // **=, reduce: MlrvalOrFunction
			reduce(220), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(220), // ||, reduce: MlrvalOrFunction
			reduce(220), // ^^, reduce: MlrvalOrFunction
			reduce(220), // &&, reduce: MlrvalOrFunction
			reduce(220), // ??, reduce: MlrvalOrFunction
			reduce(220), // ???, reduce: MlrvalOrFunction
			reduce(220), // =~, reduce: MlrvalOrFunction
			reduce(220), // !=~, reduce: MlrvalOrFunction
			reduce(220), // ==, reduce: MlrvalOrFunction
			reduce(220), // !=, reduce: MlrvalOrFunction
			reduce(220), // <=>, reduce: MlrvalOrFunction
			reduce(220), // >=, reduce: MlrvalOrFunction
			reduce(220), // <, reduce: MlrvalOrFunction
			reduce(220), // <=, reduce: MlrvalOrFunction
			reduce(220), // ^, reduce: MlrvalOrFunction
			reduce(220), // &, reduce: MlrvalOrFunction
			reduce(220), // <<, reduce: MlrvalOrFunction
			reduce(220), // >>>, reduce: MlrvalOrFunction
			reduce(220), // +, reduce: MlrvalOrFunction
			reduce(220), // -, reduce: MlrvalOrFunction
			reduce(220), // .+, reduce: MlrvalOrFunction
			reduce(220), // .-, reduce: MlrvalOrFunction
			reduce(220), // *, reduce: MlrvalOrFunction
			reduce(220), // /, reduce: MlrvalOrFunction
			reduce(220), // //, reduce: MlrvalOrFunction
			reduce(220), // %, reduce: MlrvalOrFunction
			reduce(220), // .*, reduce: MlrvalOrFunction
			reduce(220), // ./, reduce: MlrvalOrFunction
			reduce(220), // .//, reduce: MlrvalOrFunction
			reduce(220), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(220), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(220), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(220), // [[, reduce: MlrvalOrFunction
			reduce(220), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S112
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(221), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(221), // ;, reduce: MlrvalOrFunction
			reduce(221), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(221), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(221), // >, reduce: MlrvalOrFunction
			reduce(221), // >>, reduce: MlrvalOrFunction
			reduce(221), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(221), // ||=, reduce: MlrvalOrFunction
			reduce(221), // ^^=, reduce: MlrvalOrFunction
			reduce(221), // &&=, reduce: MlrvalOrFunction
			reduce(221), // ??=, reduce: MlrvalOrFunction
			reduce(221), // ???=, reduce: MlrvalOrFunction
			reduce(221), // |=, reduce: MlrvalOrFunction
			reduce(221), // &=, reduce: MlrvalOrFunction
			reduce(221), // ^=, reduce: MlrvalOrFunction
			reduce(221), // <<=, reduce: MlrvalOrFunction
			reduce(221), // >>=, reduce: MlrvalOrFunction
			reduce(221), // >>>=, reduce: MlrvalOrFunction
			reduce(221), // +=, reduce: MlrvalOrFunction
			reduce(221), // .=, reduce: MlrvalOrFunction
			reduce(221), // -=, reduce: MlrvalOrFunction
			reduce(221), // *=, reduce: MlrvalOrFunction
			reduce(221), // /=, reduce: MlrvalOrFunction
			reduce(221), // //=, reduce: MlrvalOrFunction
			reduce(221), // %=, reduce: MlrvalOrFunction
			reduce(221), // **=, reduce: MlrvalOrFunction
			reduce(221), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(221), // ||, reduce: MlrvalOrFunction
			reduce(221), // ^^, reduce: MlrvalOrFunction
			reduce(221), // &&, reduce: MlrvalOrFunction
			reduce(221), // ??, reduce: MlrvalOrFunction
			reduce(221), // ???, reduce: MlrvalOrFunction
			reduce(221), // =~, reduce: MlrvalOrFunction
			reduce(221), // !=~, reduce: MlrvalOrFunction
			reduce(221), // ==, reduce: MlrvalOrFunction
			reduce(221), // !=, reduce: MlrvalOrFunction
			reduce(221), // <=>, reduce: MlrvalOrFunction
			reduce(221), // >=, reduce: MlrvalOrFunction
			reduce(221), // <, reduce: MlrvalOrFunction
			reduce(221), // <=, reduce: MlrvalOrFunction
			reduce(221), // ^, reduce: MlrvalOrFunction
			reduce(221), // &, reduce: MlrvalOrFunction
			reduce(221), // <<, reduce: MlrvalOrFunction
			reduce(221), // >>>, reduce: MlrvalOrFunction
			reduce(221), // +, reduce: MlrvalOrFunction
			reduce(221), // -, reduce: MlrvalOrFunction
			reduce(221), // .+, reduce: MlrvalOrFunction
			reduce(221), // .-, reduce: MlrvalOrFunction
			reduce(221), // *, reduce: MlrvalOrFunction
			reduce(221), // /, reduce: MlrvalOrFunction
			reduce(221), // //, reduce: MlrvalOrFunction
			reduce(221), // %, reduce: MlrvalOrFunction
			reduce(221), // .*, reduce: MlrvalOrFunction
			reduce(221), // ./, reduce: MlrvalOrFunction
			reduce(221), // .//, reduce: MlrvalOrFunction
			reduce(221), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(221), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(221), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(221), // [[, reduce: MlrvalOrFunction
			reduce(221), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S113
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(222), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(222), // ;, reduce: MlrvalOrFunction
			reduce(222), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(222), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(222), // >, reduce: MlrvalOrFunction
			reduce(222), // >>, reduce: MlrvalOrFunction
			reduce(222), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(222), // ||=, reduce: MlrvalOrFunction
			reduce(222), // ^^=, reduce: MlrvalOrFunction
			reduce(222), // &&=, reduce: MlrvalOrFunction
			reduce(222), // ??=, reduce: MlrvalOrFunction
			reduce(222), // ???=, reduce: MlrvalOrFunction
			reduce(222), // |=, reduce: MlrvalOrFunction
			reduce(222), // &=, reduce: MlrvalOrFunction
			reduce(222), // ^=, reduce: MlrvalOrFunction
			reduce(222), // <<=, reduce: MlrvalOrFunction
			reduce(222), // >>=, reduce: MlrvalOrFunction
			reduce(222), // >>>=, reduce: MlrvalOrFunction
			reduce(222), // +=, reduce: MlrvalOrFunction
			reduce(222), // .=, reduce: MlrvalOrFunction
			reduce(222), // -=, reduce: MlrvalOrFunction
			reduce(222), // *=, reduce: MlrvalOrFunction
			reduce(222), // /=, reduce: MlrvalOrFunction
			reduce(222), // //=, reduce: MlrvalOrFunction
			reduce(222), // %=, reduce: MlrvalOrFunction
			reduce(222), // **=, reduce: MlrvalOrFunction
			reduce(222), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(222), // ||, reduce: MlrvalOrFunction
			reduce(222), // ^^, reduce: MlrvalOrFunction
			reduce(222), // &&, reduce: MlrvalOrFunction
			reduce(222), // ??, reduce: MlrvalOrFunction
			reduce(222), // ???, reduce: MlrvalOrFunction
			reduce(222), // =~, reduce: MlrvalOrFunction
			reduce(222), // !=~, reduce: MlrvalOrFunction
			reduce(222), // ==, reduce: MlrvalOrFunction
			reduce(222), // !=, reduce: MlrvalOrFunction
			reduce(222), // <=>, reduce: MlrvalOrFunction
			reduce(222), // >=, reduce: MlrvalOrFunction
			reduce(222), // <, reduce: MlrvalOrFunction
			reduce(222), // <=, reduce: MlrvalOrFunction
			reduce(222), // ^, reduce: MlrvalOrFunction
			reduce(222), // &, reduce: MlrvalOrFunction
			reduce(222), // <<, reduce: MlrvalOrFunction
			reduce(222), // >>>, reduce: MlrvalOrFunction
			reduce(222), // +, reduce: MlrvalOrFunction
			reduce(222), // -, reduce: MlrvalOrFunction
			reduce(222), // .+, reduce: MlrvalOrFunction
			reduce(222), // .-, reduce: MlrvalOrFunction
			reduce(222), // *, reduce: MlrvalOrFunction
			reduce(222), // /, reduce: MlrvalOrFunction
			reduce(222), // //, reduce: MlrvalOrFunction
			reduce(222), // %, reduce: MlrvalOrFunction
			reduce(222), // .*, reduce: MlrvalOrFunction
			reduce(222), // ./, reduce: MlrvalOrFunction
			reduce(222), // .//, reduce: MlrvalOrFunction
			reduce(222), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(222), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(222), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(222), // [[, reduce: MlrvalOrFunction
			reduce(222), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S114
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(223), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(223), // ;, reduce: MlrvalOrFunction
			reduce(223), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(223), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(223), // >, reduce: MlrvalOrFunction
			reduce(223), // >>, reduce: MlrvalOrFunction
			reduce(223), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(223), // ||=, reduce: MlrvalOrFunction
			reduce(223), // ^^=, reduce: MlrvalOrFunction
			reduce(223), // &&=, reduce: MlrvalOrFunction
			reduce(223), // ??=, reduce: MlrvalOrFunction
			reduce(223), // ???=, reduce: MlrvalOrFunction
			reduce(223), // |=, reduce: MlrvalOrFunction
			reduce(223), // &=, reduce: MlrvalOrFunction
			reduce(223), // ^=, reduce: MlrvalOrFunction
			reduce(223), // <<=, reduce: MlrvalOrFunction
			reduce(223), // >>=, reduce: MlrvalOrFunction
			reduce(223), // >>>=, reduce: MlrvalOrFunction
			reduce(223), // +=, reduce: MlrvalOrFunction
			reduce(223), // .=, reduce: MlrvalOrFunction
			reduce(223), // -=, reduce: MlrvalOrFunction
			reduce(223), // *=, reduce: MlrvalOrFunction
			reduce(223), // /=, reduce: MlrvalOrFunction
			reduce(223), // //=, reduce: MlrvalOrFunction
			reduce(223), // %=, reduce: MlrvalOrFunction
			reduce(223), // **=, reduce: MlrvalOrFunction
			reduce(223), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(223), // ||, reduce: MlrvalOrFunction
			reduce(223), // ^^, reduce: MlrvalOrFunction
			reduce(223), // &&, reduce: MlrvalOrFunction
			reduce(223), // ??, reduce: MlrvalOrFunction
			reduce(223), // ???, reduce: MlrvalOrFunction
			reduce(223), // =~, reduce: MlrvalOrFunction
			reduce(223), // !=~, reduce: MlrvalOrFunction
			reduce(223), // ==, reduce: MlrvalOrFunction
			reduce(223), // !=, reduce: MlrvalOrFunction
			reduce(223), // <=>, reduce: MlrvalOrFunction
			reduce(223), // >=, reduce: MlrvalOrFunction
			reduce(223), // <, reduce: MlrvalOrFunction
			reduce(223), // <=, reduce: MlrvalOrFunction
			reduce(223), // ^, reduce: MlrvalOrFunction
			reduce(223), // &, reduce: MlrvalOrFunction
			reduce(223), // <<, reduce: MlrvalOrFunction
			reduce(223), // >>>, reduce: MlrvalOrFunction
			reduce(223), // +, reduce: MlrvalOrFunction
			reduce(223), // -, reduce: MlrvalOrFunction
			reduce(223), // .+, reduce: MlrvalOrFunction
			reduce(223), // .-, reduce: MlrvalOrFunction
			reduce(223), // *, reduce: MlrvalOrFunction
			reduce(223), // /, reduce: MlrvalOrFunction
			reduce(223), // //, reduce: MlrvalOrFunction
			reduce(223), // %, reduce: MlrvalOrFunction
			reduce(223), // .*, reduce: MlrvalOrFunction
			reduce(223), // ./, reduce: MlrvalOrFunction
			reduce(223), // .//, reduce: MlrvalOrFunction
			reduce(223), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(223), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(223), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(223), // [[, reduce: MlrvalOrFunction
			reduce(223), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S115
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(224), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(224), // ;, reduce: MlrvalOrFunction
			reduce(224), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(224), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(224), // >, reduce: MlrvalOrFunction
			reduce(224), // >>, reduce: MlrvalOrFunction
			reduce(224), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(224), // ||=, reduce: MlrvalOrFunction
			reduce(224), // ^^=, reduce: MlrvalOrFunction
			reduce(224), // &&=, reduce: MlrvalOrFunction
			reduce(224), // ??=, reduce: MlrvalOrFunction
			reduce(224), // ???=, reduce: MlrvalOrFunction
			reduce(224), // |=, reduce: MlrvalOrFunction
			reduce(224), // &=, reduce: MlrvalOrFunction
			reduce(224), // ^=, reduce: MlrvalOrFunction
			reduce(224), // <<=, reduce: MlrvalOrFunction
			reduce(224), // >>=, reduce: MlrvalOrFunction
			reduce(224), // >>>=, reduce: MlrvalOrFunction
			reduce(224), // +=, reduce: MlrvalOrFunction
			reduce(224), // .=, reduce: MlrvalOrFunction
			reduce(224), // -=, reduce: MlrvalOrFunction
			reduce(224), // *=, reduce: MlrvalOrFunction
			reduce(224), // /=, reduce: MlrvalOrFunction
			reduce(224), // //=, reduce: MlrvalOrFunction
			reduce(224), // %=, reduce: MlrvalOrFunction
			reduce(224), // **=, reduce: MlrvalOrFunction
			reduce(224), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(224), // ||, reduce: MlrvalOrFunction
			reduce(224), // ^^, reduce: MlrvalOrFunction
			reduce(224), // &&, reduce: MlrvalOrFunction
			reduce(224), // ??, reduce: MlrvalOrFunction
			reduce(224), // ???, reduce: MlrvalOrFunction
			reduce(224), // =~, reduce: MlrvalOrFunction
			reduce(224), // !=~, reduce: MlrvalOrFunction
			reduce(224), // ==, reduce: MlrvalOrFunction
			reduce(224), // !=, reduce: MlrvalOrFunction
			reduce(224), // <=>, reduce: MlrvalOrFunction
			reduce(224), // >=, reduce: MlrvalOrFunction
			reduce(224), // <, reduce: MlrvalOrFunction
			reduce(224), // <=, reduce: MlrvalOrFunction
			reduce(224), // ^, reduce: MlrvalOrFunction
			reduce(224), // &, reduce: MlrvalOrFunction
			reduce(224), // <<, reduce: MlrvalOrFunction
			reduce(224), // >>>, reduce: MlrvalOrFunction
			reduce(224), // +, reduce: MlrvalOrFunction
			reduce(224), // -, reduce: MlrvalOrFunction
			reduce(224), // .+, reduce: MlrvalOrFunction
			reduce(224), // .-, reduce: MlrvalOrFunction
			reduce(224), // *, reduce: MlrvalOrFunction
			reduce(224), // /, reduce: MlrvalOrFunction
			reduce(224), // //, reduce: MlrvalOrFunction
			reduce(224), // %, reduce: MlrvalOrFunction
			reduce(224), // .*, reduce: MlrvalOrFunction
			reduce(224), // ./, reduce: MlrvalOrFunction
			reduce(224), // .//, reduce: MlrvalOrFunction
			reduce(224), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(224), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(224), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(224), // [[, reduce: MlrvalOrFunction
			reduce(224), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S116
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(225), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(225), // ;, reduce: MlrvalOrFunction
			reduce(225), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(225), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(225), // >, reduce: MlrvalOrFunction
			reduce(225), // >>, reduce: MlrvalOrFunction
			reduce(225), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(225), // ||=, reduce: MlrvalOrFunction
			reduce(225), // ^^=, reduce: MlrvalOrFunction
			reduce(225), // &&=, reduce: MlrvalOrFunction
			reduce(225), // ??=, reduce: MlrvalOrFunction
			reduce(225), // ???=, reduce: MlrvalOrFunction
			reduce(225), // |=, reduce: MlrvalOrFunction
			reduce(225), // &=, reduce: MlrvalOrFunction
			reduce(225), // ^=, reduce: MlrvalOrFunction
			reduce(225), // <<=, reduce: MlrvalOrFunction
			reduce(225), // >>=, reduce: MlrvalOrFunction
			reduce(225), // >>>=, reduce: MlrvalOrFunction
			reduce(225), // +=, reduce: MlrvalOrFunction
			reduce(225), // .=, reduce: MlrvalOrFunction
			reduce(225), // -=, reduce: MlrvalOrFunction
			reduce(225), // *=, reduce: MlrvalOrFunction
			reduce(225), // /=, reduce: MlrvalOrFunction
			reduce(225), // //=, reduce: MlrvalOrFunction
			reduce(225), // %=, reduce: MlrvalOrFunction
			reduce(225), // **=, reduce: MlrvalOrFunction
			reduce(225), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(225), // ||, reduce: MlrvalOrFunction
			reduce(225), // ^^, reduce: MlrvalOrFunction
			reduce(225), // &&, reduce: MlrvalOrFunction
			reduce(225), // ??, reduce: MlrvalOrFunction
			reduce(225), // ???, reduce: MlrvalOrFunction
			reduce(225), // =~, reduce: MlrvalOrFunction
			reduce(225), // !=~, reduce: MlrvalOrFunction
			reduce(225), // ==, reduce: MlrvalOrFunction
			reduce(225), // !=, reduce: MlrvalOrFunction
			reduce(225), // <=>, reduce: MlrvalOrFunction
			reduce(225), // >=, reduce: MlrvalOrFunction
			reduce(225), // <, reduce: MlrvalOrFunction
			reduce(225), // <=, reduce: MlrvalOrFunction
			reduce(225), // ^, reduce: MlrvalOrFunction
			reduce(225), // &, reduce: MlrvalOrFunction
			reduce(225), // <<, reduce: MlrvalOrFunction
			reduce(225), // >>>, reduce: MlrvalOrFunction
			reduce(225), // +, reduce: MlrvalOrFunction
			reduce(225), // -, reduce: MlrvalOrFunction
			reduce(225), // .+, reduce: MlrvalOrFunction
			reduce(225), // .-, reduce: MlrvalOrFunction
			reduce(225), // *, reduce: MlrvalOrFunction
			reduce(225), // /, reduce: MlrvalOrFunction
			reduce(225), // //, reduce: MlrvalOrFunction
			reduce(225), // %, reduce: MlrvalOrFunction
			reduce(225), // .*, reduce: MlrvalOrFunction
			reduce(225), // ./, reduce: MlrvalOrFunction
			reduce(225), // .//, reduce: MlrvalOrFunction
			reduce(225), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(225), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(225), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(225), // [[, reduce: MlrvalOrFunction
			reduce(225), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S117
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(226), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(226), // ;, reduce: MlrvalOrFunction
			reduce(226), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(226), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(226), // >, reduce: MlrvalOrFunction
			reduce(226), // >>, reduce: MlrvalOrFunction
			reduce(226), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(226), // ||=, reduce: MlrvalOrFunction
			reduce(226), // ^^=, reduce: MlrvalOrFunction
			reduce(226), // &&=, reduce: MlrvalOrFunction
			reduce(226), // ??=, reduce: MlrvalOrFunction
			reduce(226), // ???=, reduce: MlrvalOrFunction
			reduce(226), // |=, reduce: MlrvalOrFunction
			reduce(226), // &=, reduce: MlrvalOrFunction
			reduce(226), // ^=, reduce: MlrvalOrFunction
			reduce(226), // <<=, reduce: MlrvalOrFunction
			reduce(226), // >>=, reduce: MlrvalOrFunction
			reduce(226), // >>>=, reduce: MlrvalOrFunction
			reduce(226), // +=, reduce: MlrvalOrFunction
			reduce(226), // .=, reduce: MlrvalOrFunction
			reduce(226), // -=, reduce: MlrvalOrFunction
			reduce(226), // *=, reduce: MlrvalOrFunction
			reduce(226), // /=, reduce: MlrvalOrFunction
			reduce(226), // //=, reduce: MlrvalOrFunction
			reduce(226), // %=, reduce: MlrvalOrFunction
			reduce(226), // **=, reduce: MlrvalOrFunction
			reduce(226), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(226), // ||, reduce: MlrvalOrFunction
			reduce(226), // ^^, reduce: MlrvalOrFunction
			reduce(226), // &&, reduce: MlrvalOrFunction
			reduce(226), // ??, reduce: MlrvalOrFunction
			reduce(226), // ???, reduce: MlrvalOrFunction
			reduce(226), // =~, reduce: MlrvalOrFunction
			reduce(226), // !=~, reduce: MlrvalOrFunction
			reduce(226), // ==, reduce: MlrvalOrFunction
			reduce(226), // !=, reduce: MlrvalOrFunction
			reduce(226), // <=>, reduce: MlrvalOrFunction
			reduce(226), // >=, reduce: MlrvalOrFunction
			reduce(226), // <, reduce: MlrvalOrFunction
			reduce(226), // <=, reduce: MlrvalOrFunction
			reduce(226), // ^, reduce: MlrvalOrFunction
			reduce(226), // &, reduce: MlrvalOrFunction
			reduce(226), // <<, reduce: MlrvalOrFunction
			reduce(226), // >>>, reduce: MlrvalOrFunction
			reduce(226), // +, reduce: MlrvalOrFunction
			reduce(226), // -, reduce: MlrvalOrFunction
			reduce(226), // .+, reduce: MlrvalOrFunction
			reduce(226), // .-, reduce: MlrvalOrFunction
			reduce(226), // *, reduce: MlrvalOrFunction
			reduce(226), // /, reduce: MlrvalOrFunction
			reduce(226), // //, reduce: MlrvalOrFunction
			reduce(226), // %, reduce: MlrvalOrFunction
			reduce(226), // .*, reduce: MlrvalOrFunction
			reduce(226), // ./, reduce: MlrvalOrFunction
			reduce(226), // .//, reduce: MlrvalOrFunction
			reduce(226), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(226), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(226), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(226), // [[, reduce: MlrvalOrFunction
			reduce(226), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S118
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(227), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(227), // ;, reduce: MlrvalOrFunction
			reduce(227), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(227), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(227), // >, reduce: MlrvalOrFunction
			reduce(227), // >>, reduce: MlrvalOrFunction
			reduce(227), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(227), // ||=, reduce: MlrvalOrFunction
			reduce(227), // ^^=, reduce: MlrvalOrFunction
			reduce(227), // &&=, reduce: MlrvalOrFunction
			reduce(227), // ??=, reduce: MlrvalOrFunction
			reduce(227), // ???=, reduce: MlrvalOrFunction
			reduce(227), // |=, reduce: MlrvalOrFunction
			reduce(227), // &=, reduce: MlrvalOrFunction
			reduce(227), // ^=, reduce: MlrvalOrFunction
			reduce(227), // <<=, reduce: MlrvalOrFunction
			reduce(227), // >>=, reduce: MlrvalOrFunction
			reduce(227), // >>>=, reduce: MlrvalOrFunction
			reduce(227), // +=, reduce: MlrvalOrFunction
			reduce(227), // .=, reduce: MlrvalOrFunction
			reduce(227), // -=, reduce: MlrvalOrFunction
			reduce(227), // *=, reduce: MlrvalOrFunction
			reduce(227), // /=, reduce: MlrvalOrFunction
			reduce(227), // //=, reduce: MlrvalOrFunction
			reduce(227), // %=, reduce: MlrvalOrFunction
			reduce(227), // **=, reduce: MlrvalOrFunction
			reduce(227), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(227), // ||, reduce: MlrvalOrFunction
			reduce(227), // ^^, reduce: MlrvalOrFunction
			reduce(227), // &&, reduce: MlrvalOrFunction
			reduce(227), // ??, reduce: MlrvalOrFunction
			reduce(227), // ???, reduce: MlrvalOrFunction
			reduce(227), // =~, reduce: MlrvalOrFunction
			reduce(227), // !=~, reduce: MlrvalOrFunction
			reduce(227), // ==, reduce: MlrvalOrFunction
			reduce(227), // !=, reduce: MlrvalOrFunction
			reduce(227), // <=>, reduce: MlrvalOrFunction
			reduce(227), // >=, reduce: MlrvalOrFunction
			reduce(227), // <, reduce: MlrvalOrFunction
			reduce(227), // <=, reduce: MlrvalOrFunction
			reduce(227), // ^, reduce: MlrvalOrFunction
			reduce(227), // &, reduce: MlrvalOrFunction
			reduce(227), // <<, reduce: MlrvalOrFunction
			reduce(227), // >>>, reduce: MlrvalOrFunction
			reduce(227), // +, reduce: MlrvalOrFunction
			reduce(227), // -, reduce: MlrvalOrFunction
			reduce(227), // .+, reduce: MlrvalOrFunction
			reduce(227), // .-, reduce: MlrvalOrFunction
			reduce(227), // *, reduce: MlrvalOrFunction
			reduce(227), // /, reduce: MlrvalOrFunction
			reduce(227), // //, reduce: MlrvalOrFunction
			reduce(227), // %, reduce: MlrvalOrFunction
			reduce(227), // .*, reduce: MlrvalOrFunction
			reduce(227), // ./, reduce: MlrvalOrFunction
			reduce(227), // .//, reduce: MlrvalOrFunction
			reduce(227), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(227), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(227), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(227), // [[, reduce: MlrvalOrFunction
			reduce(227), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S119
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			shift(770), // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S120
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(240), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(240), // ;, reduce: MlrvalOrFunction
			reduce(240), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(240), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(240), // >, reduce: MlrvalOrFunction
			reduce(240), // >>, reduce: MlrvalOrFunction
			reduce(240), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(240), // ||=, reduce: MlrvalOrFunction
			reduce(240), // ^^=, reduce: MlrvalOrFunction
			reduce(240), // &&=, reduce: MlrvalOrFunction
			reduce(240), // ??=, reduce: MlrvalOrFunction
			reduce(240), // ???=, reduce: MlrvalOrFunction
			reduce(240), // |=, reduce: MlrvalOrFunction
			reduce(240), // &=, reduce: MlrvalOrFunction
			reduce(240), // ^=, reduce: MlrvalOrFunction
			reduce(240), // <<=, reduce: MlrvalOrFunction
			reduce(240), // >>=, reduce: MlrvalOrFunction
			reduce(240), // >>>=, reduce: MlrvalOrFunction
			reduce(240), // +=, reduce: MlrvalOrFunction
			reduce(240), // .=, reduce: MlrvalOrFunction
			reduce(240), // -=, reduce: MlrvalOrFunction
			reduce(240), // *=, reduce: MlrvalOrFunction
			reduce(240), // /=, reduce: MlrvalOrFunction
			reduce(240), // //=, reduce: MlrvalOrFunction
			reduce(240), // %=, reduce: MlrvalOrFunction
			reduce(240), // **=, reduce: MlrvalOrFunction
			reduce(240), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(240), // ||, reduce: MlrvalOrFunction
			reduce(240), // ^^, reduce: MlrvalOrFunction
			reduce(240), // &&, reduce: MlrvalOrFunction
			reduce(240), // ??, reduce: MlrvalOrFunction
			reduce(240), // ???, reduce: MlrvalOrFunction
			reduce(240), // =~, reduce: MlrvalOrFunction
			reduce(240), // !=~, reduce: MlrvalOrFunction
			reduce(240), // ==, reduce: MlrvalOrFunction
			reduce(240), // !=, reduce: MlrvalOrFunction
			reduce(240), // <=>, reduce: MlrvalOrFunction
			reduce(240), // >=, reduce: MlrvalOrFunction
			reduce(240), // <, reduce: MlrvalOrFunction
			reduce(240), // <=, reduce: MlrvalOrFunction
			reduce(240), // ^, reduce: MlrvalOrFunction
			reduce(240), // &, reduce: MlrvalOrFunction
			reduce(240), // <<, reduce: MlrvalOrFunction
			reduce(240), // >>>, reduce: MlrvalOrFunction
			reduce(240), // +, reduce: MlrvalOrFunction
			reduce(240), // -, reduce: MlrvalOrFunction
			reduce(240), // .+, reduce: MlrvalOrFunction
			reduce(240), // .-, reduce: MlrvalOrFunction
			reduce(240), // *, reduce: MlrvalOrFunction
			reduce(240), // /, reduce: MlrvalOrFunction
			reduce(240), // //, reduce: MlrvalOrFunction
			reduce(240), // %, reduce: MlrvalOrFunction
			reduce(240), // .*, reduce: MlrvalOrFunction
			reduce(240), // ./, reduce: MlrvalOrFunction
			reduce(240), // .//, reduce: MlrvalOrFunction
			reduce(240), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(240), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(240), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(240), // [[, reduce: MlrvalOrFunction
			reduce(240), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S121
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(241), // $, reduce: ContextVariable
			nil,         // empty
			reduce(241), // ;, reduce: ContextVariable
			reduce(241), // {, reduce: ContextVariable
			nil,         // }
			reduce(241), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(241), // >, reduce: ContextVariable
			reduce(241), // >>, reduce: ContextVariable
			reduce(241), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(241), // ||=, reduce: ContextVariable
			reduce(241), // ^^=, reduce: ContextVariable
			reduce(241), // &&=, reduce: ContextVariable
			reduce(241), // ??=, reduce: ContextVariable
			reduce(241), // ???=, reduce: ContextVariable
			reduce(241), // |=, reduce: ContextVariable
			reduce(241), // &=, reduce: ContextVariable
			reduce(241), // ^=, reduce: ContextVariable
			reduce(241), // <<=, reduce: ContextVariable
			reduce(241), // >>=, reduce: ContextVariable
			reduce(241), // >>>=, reduce: ContextVariable
			reduce(241), // +=, reduce: ContextVariable
			reduce(241), // .=, reduce: ContextVariable
			reduce(241), // -=, reduce: ContextVariable
			reduce(241), // *=, reduce: ContextVariable
			reduce(241), // /=, reduce: ContextVariable
			reduce(241), // //=, reduce: ContextVariable
			reduce(241), // %=, reduce: ContextVariable
			reduce(241), // **=, reduce: ContextVariable
			reduce(241), // ?, reduce: ContextVariable
			nil,         // :
			reduce(241), // ||, reduce: ContextVariable
			reduce(241), // ^^, reduce: ContextVariable
			reduce(241), // &&, reduce: ContextVariable
			reduce(241), // ??, reduce: ContextVariable
			reduce(241), // ???, reduce: ContextVariable
			reduce(241), // =~, reduce: ContextVariable
			reduce(241), // !=~, reduce: ContextVariable
			reduce(241), // ==, reduce: ContextVariable
			reduce(241), // !=, reduce: ContextVariable
			reduce(241), // <=>, reduce: ContextVariable
			reduce(241), // >=, reduce: ContextVariable
			reduce(241), // <, reduce: ContextVariable
			reduce(241), // <=, reduce: ContextVariable
			reduce(241), // ^, reduce: ContextVariable
			reduce(241), // &, reduce: ContextVariable
			reduce(241), // <<, reduce: ContextVariable
			reduce(241), // >>>, reduce: ContextVariable
			reduce(241), // +, reduce: ContextVariable
			reduce(241), // -, reduce: ContextVariable
			reduce(241), // .+, reduce: ContextVariable
			reduce(241), // .-, reduce: ContextVariable
			reduce(241), // *, reduce: ContextVariable
			reduce(241), // /, reduce: ContextVariable
			reduce(241), // //, reduce: ContextVariable
			reduce(241), // %, reduce: ContextVariable
			reduce(241), // .*, reduce: ContextVariable
			reduce(241), // ./, reduce: ContextVariable
			reduce(241), // .//, reduce: ContextVariable
			reduce(241), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(241), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(241), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(241), // [[, reduce: ContextVariable
			reduce(241), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S122
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(242), // $, reduce: ContextVariable
			nil,         // empty
			reduce(242), // ;, reduce: ContextVariable
			reduce(242), // {, reduce: ContextVariable
			nil,         // }
			reduce(242), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(242), // >, reduce: ContextVariable
			reduce(242), // >>, reduce: ContextVariable
			reduce(242), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(242), // ||=, reduce: ContextVariable
			reduce(242), // ^^=, reduce: ContextVariable
			reduce(242), // &&=, reduce: ContextVariable
			reduce(242), // ??=, reduce: ContextVariable
			reduce(242), // ???=, reduce: ContextVariable
			reduce(242), // |=, reduce: ContextVariable
			reduce(242), // &=, reduce: ContextVariable
			reduce(242), // ^=, reduce: ContextVariable
			reduce(242), // <<=, reduce: ContextVariable
			reduce(242), // >>=, reduce: ContextVariable
			reduce(242), // >>>=, reduce: ContextVariable
			reduce(242), // +=, reduce: ContextVariable
			reduce(242), // .=, reduce: ContextVariable
			reduce(242), // -=, reduce: ContextVariable
			reduce(242), // *=, reduce: ContextVariable
			reduce(242), // /=, reduce: ContextVariable
			reduce(242), // //=, reduce: ContextVariable
			reduce(242), // %=, reduce: ContextVariable
			reduce(242), // **=, reduce: ContextVariable
			reduce(242), // ?, reduce: ContextVariable
			nil,         // :
			reduce(242), // ||, reduce: ContextVariable
			reduce(242), // ^^, reduce: ContextVariable
			reduce(242), // &&, reduce: ContextVariable
			reduce(242), // ??, reduce: ContextVariable
			reduce(242), // ???, reduce: ContextVariable
			reduce(242), // =~, reduce: ContextVariable
			reduce(242), // !=~, reduce: ContextVariable
			reduce(242), // ==, reduce: ContextVariable
			reduce(242), // !=, reduce: ContextVariable
			reduce(242), // <=>, reduce: ContextVariable
			reduce(242), // >=, reduce: ContextVariable
			reduce(242), // <, reduce: ContextVariable
			reduce(242), // <=, reduce: ContextVariable
			reduce(242), // ^, reduce: ContextVariable
			reduce(242), // &, reduce: ContextVariable
			reduce(242), // <<, reduce: ContextVariable
			reduce(242), // >>>, reduce: ContextVariable
			reduce(242), // +, reduce: ContextVariable
			reduce(242), // -, reduce: ContextVariable
			reduce(242), // .+, reduce: ContextVariable
			reduce(242), // .-, reduce: ContextVariable
			reduce(242), // *, reduce: ContextVariable
			reduce(242), // /, reduce: ContextVariable
			reduce(242), // //, reduce: ContextVariable
			reduce(242), // %, reduce: ContextVariable
			reduce(242), // .*, reduce: ContextVariable
			reduce(242), // ./, reduce: ContextVariable
			reduce(242), // .//, reduce: ContextVariable
			reduce(242), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(242), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(242), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(242), // [[, reduce: ContextVariable
			reduce(242), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S123
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(243), // $, reduce: ContextVariable
			nil,         // empty
			reduce(243), // ;, reduce: ContextVariable
			reduce(243), // {, reduce: ContextVariable
			nil,         // }
			reduce(243), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(243), // >, reduce: ContextVariable
			reduce(243), // >>, reduce: ContextVariable
			reduce(243), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(243), // ||=, reduce: ContextVariable
			reduce(243), // ^^=, reduce: ContextVariable
			reduce(243), // &&=, reduce: ContextVariable
			reduce(243), // ??=, reduce: ContextVariable
			reduce(243), // ???=, reduce: ContextVariable
			reduce(243), // |=, reduce: ContextVariable
			reduce(243), // &=, reduce: ContextVariable
			reduce(243), // ^=, reduce: ContextVariable
			reduce(243), // <<=, reduce: ContextVariable
			reduce(243), // >>=, reduce: ContextVariable
			reduce(243), // >>>=, reduce: ContextVariable
			reduce(243), // +=, reduce: ContextVariable
			reduce(243), // .=, reduce: ContextVariable
			reduce(243), // -=, reduce: ContextVariable
			reduce(243), // *=, reduce: ContextVariable
			reduce(243), // /=, reduce: ContextVariable
			reduce(243), // //=, reduce: ContextVariable
			reduce(243), // %=, reduce: ContextVariable
			reduce(243), // **=, reduce: ContextVariable
			reduce(243), // ?, reduce: ContextVariable
			nil,         // :
			reduce(243), // ||, reduce: ContextVariable
			reduce(243), // ^^, reduce: ContextVariable
			reduce(243), // &&, reduce: ContextVariable
			reduce(243), // ??, reduce: ContextVariable
			reduce(243), // ???, reduce: ContextVariable
			reduce(243), // =~, reduce: ContextVariable
			reduce(243), // !=~, reduce: ContextVariable
			reduce(243), // ==, reduce: ContextVariable
			reduce(243), // !=, reduce: ContextVariable
			reduce(243), // <=>, reduce: ContextVariable
			reduce(243), // >=, reduce: ContextVariable
			reduce(243), // <, reduce: ContextVariable
			reduce(243), // <=, reduce: ContextVariable
			reduce(243), // ^, reduce: ContextVariable
			reduce(243), // &, reduce: ContextVariable
			reduce(243), // <<, reduce: ContextVariable
			reduce(243), // >>>, reduce: ContextVariable
			reduce(243), // +, reduce: ContextVariable
			reduce(243), // -, reduce: ContextVariable
			reduce(243), // .+, reduce: ContextVariable
			reduce(243), // .-, reduce: ContextVariable
			reduce(243), // *, reduce: ContextVariable
			reduce(243), // /, reduce: ContextVariable
			reduce(243), // //, reduce: ContextVariable
			reduce(243), // %, reduce: ContextVariable
			reduce(243), // .*, reduce: ContextVariable
			reduce(243), // ./, reduce: ContextVariable
			reduce(243), // .//, reduce: ContextVariable
			reduce(243), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(243), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(243), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(243), // [[, reduce: ContextVariable
			reduce(243), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S124
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(244), // $, reduce: ContextVariable
			nil,         // empty
			reduce(244), // ;, reduce: ContextVariable
			reduce(244), // {, reduce: ContextVariable
			nil,         // }
			reduce(244), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(244), // >, reduce: ContextVariable
			reduce(244), // >>, reduce: ContextVariable
			reduce(244), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(244), // ||=, reduce: ContextVariable
			reduce(244), // ^^=, reduce: ContextVariable
			reduce(244), // &&=, reduce: ContextVariable
			reduce(244), // ??=, reduce: ContextVariable
			reduce(244), // ???=, reduce: ContextVariable
			reduce(244), // |=, reduce: ContextVariable
			reduce(244), // &=, reduce: ContextVariable
			reduce(244), // ^=, reduce: ContextVariable
			reduce(244), // <<=, reduce: ContextVariable
			reduce(244), // >>=, reduce: ContextVariable
			reduce(244), // >>>=, reduce: ContextVariable
			reduce(244), // +=, reduce: ContextVariable
			reduce(244), // .=, reduce: ContextVariable
			reduce(244), // -=, reduce: ContextVariable
			reduce(244), // *=, reduce: ContextVariable
			reduce(244), // /=, reduce: ContextVariable
			reduce(244), // //=, reduce: ContextVariable
			reduce(244), // %=, reduce: ContextVariable
			reduce(244), // **=, reduce: ContextVariable
			reduce(244), // ?, reduce: ContextVariable
			nil,         // :
			reduce(244), // ||, reduce: ContextVariable
			reduce(244), // ^^, reduce: ContextVariable
			reduce(244), // &&, reduce: ContextVariable
			reduce(244), // ??, reduce: ContextVariable
			reduce(244), // ???, reduce: ContextVariable
			reduce(244), // =~, reduce: ContextVariable
			reduce(244), // !=~, reduce: ContextVariable
			reduce(244), // ==, reduce: ContextVariable
			reduce(244), // !=, reduce: ContextVariable
			reduce(244), // <=>, reduce: ContextVariable
			reduce(244), // >=, reduce: ContextVariable
			reduce(244), // <, reduce: ContextVariable
			reduce(244), // <=, reduce: ContextVariable
			reduce(244), // ^, reduce: ContextVariable
			reduce(244), // &, reduce: ContextVariable
			reduce(244), // <<, reduce: ContextVariable
			reduce(244), // >>>, reduce: ContextVariable
			reduce(244), // +, reduce: ContextVariable
			reduce(244), // -, reduce: ContextVariable
			reduce(244), // .+, reduce: ContextVariable
			reduce(244), // .-, reduce: ContextVariable
			reduce(244), // *, reduce: ContextVariable
			reduce(244), // /, reduce: ContextVariable
			reduce(244), // //, reduce: ContextVariable
			reduce(244), // %, reduce: ContextVariable
			reduce(244), // .*, reduce: ContextVariable
			reduce(244), // ./, reduce: ContextVariable
			reduce(244), // .//, reduce: ContextVariable
			reduce(244), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(244), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(244), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(244), // [[, reduce: ContextVariable
			reduce(244), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S125
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(245), // $, reduce: ContextVariable
			nil,         // empty
			reduce(245), // ;, reduce: ContextVariable
			reduce(245), // {, reduce: ContextVariable
			nil,         // }
			reduce(245), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(245), // >, reduce: ContextVariable
			reduce(245), // >>, reduce: ContextVariable
			reduce(245), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(245), // ||=, reduce: ContextVariable
			reduce(245), // ^^=, reduce: ContextVariable
			reduce(245), // &&=, reduce: ContextVariable
			reduce(245), // ??=, reduce: ContextVariable
			reduce(245), // ???=, reduce: ContextVariable
			reduce(245), // |=, reduce: ContextVariable
			reduce(245), // &=, reduce: ContextVariable
			reduce(245), // ^=, reduce: ContextVariable
			reduce(245), // <<=, reduce: ContextVariable
			reduce(245), // >>=, reduce: ContextVariable
			reduce(245), // >>>=, reduce: ContextVariable
			reduce(245), // +=, reduce: ContextVariable
			reduce(245), // .=, reduce: ContextVariable
			reduce(245), // -=, reduce: ContextVariable
			reduce(245), // *=, reduce: ContextVariable
			reduce(245), // /=, reduce: ContextVariable
			reduce(245), // //=, reduce: ContextVariable
			reduce(245), // %=, reduce: ContextVariable
			reduce(245), // **=, reduce: ContextVariable
			reduce(245), // ?, reduce: ContextVariable
			nil,         // :
			reduce(245), // ||, reduce: ContextVariable
			reduce(245), // ^^, reduce: ContextVariable
			reduce(245), // &&, reduce: ContextVariable
			reduce(245), // ??, reduce: ContextVariable
			reduce(245), // ???, reduce: ContextVariable
			reduce(245), // =~, reduce: ContextVariable
			reduce(245), // !=~, reduce: ContextVariable
			reduce(245), // ==, reduce: ContextVariable
			reduce(245), // !=, reduce: ContextVariable
			reduce(245), // <=>, reduce: ContextVariable
			reduce(245), // >=, reduce: ContextVariable
			reduce(245), // <, reduce: ContextVariable
			reduce(245), // <=, reduce: ContextVariable
			reduce(245), // ^, reduce: ContextVariable
			reduce(245), // &, reduce: ContextVariable
			reduce(245), // <<, reduce: ContextVariable
			reduce(245), // >>>, reduce: ContextVariable
			reduce(245), // +, reduce: ContextVariable
			reduce(245), // -, reduce: ContextVariable
			reduce(245), // .+, reduce: ContextVariable
			reduce(245), // .-, reduce: ContextVariable
			reduce(245), // *, reduce: ContextVariable
			reduce(245), // /, reduce: ContextVariable
			reduce(245), // //, reduce: ContextVariable
			reduce(245), // %, reduce: ContextVariable
			reduce(245), // .*, reduce: ContextVariable
			reduce(245), // ./, reduce: ContextVariable
			reduce(245), // .//, reduce: ContextVariable
			reduce(245), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(245), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(245), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(245), // [[, reduce: ContextVariable
			reduce(245), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S126
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(246), // $, reduce: ContextVariable
			nil,         // empty
			reduce(246), // ;, reduce: ContextVariable
			reduce(246), // {, reduce: ContextVariable
			nil,         // }
			reduce(246), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(246), // >, reduce: ContextVariable
			reduce(246), // >>, reduce: ContextVariable
			reduce(246), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(246), // ||=, reduce: ContextVariable
			reduce(246), // ^^=, reduce: ContextVariable
			reduce(246), // &&=, reduce: ContextVariable
			reduce(246), // ??=, reduce: ContextVariable
			reduce(246), // ???=, reduce: ContextVariable
			reduce(246), // |=, reduce: ContextVariable
			reduce(246), // &=, reduce: ContextVariable
			reduce(246), // ^=, reduce: ContextVariable
			reduce(246), // <<=, reduce: ContextVariable
			reduce(246), // >>=, reduce: ContextVariable
			reduce(246), // >>>=, reduce: ContextVariable
			reduce(246), // +=, reduce: ContextVariable
			reduce(246), // .=, reduce: ContextVariable
			reduce(246), // -=, reduce: ContextVariable
			reduce(246), // *=, reduce: ContextVariable
			reduce(246), // /=, reduce: ContextVariable
			reduce(246), // //=, reduce: ContextVariable
			reduce(246), // %=, reduce: ContextVariable
			reduce(246), // **=, reduce: ContextVariable
			reduce(246), // ?, reduce: ContextVariable
			nil,         // :
			reduce(246), // ||, reduce: ContextVariable
			reduce(246), // ^^, reduce: ContextVariable
			reduce(246), // &&, reduce: ContextVariable
			reduce(246), // ??, reduce: ContextVariable
			reduce(246), // ???, reduce: ContextVariable
			reduce(246), // =~, reduce: ContextVariable
			reduce(246), // !=~, reduce: ContextVariable
			reduce(246), // ==, reduce: ContextVariable
			reduce(246), // !=, reduce: ContextVariable
			reduce(246), // <=>, reduce: ContextVariable
			reduce(246), // >=, reduce: ContextVariable
			reduce(246), // <, reduce: ContextVariable
			reduce(246), // <=, reduce: ContextVariable
			reduce(246), // ^, reduce: ContextVariable
			reduce(246), // &, reduce: ContextVariable
			reduce(246), // <<, reduce: ContextVariable
			reduce(246), // >>>, reduce: ContextVariable
			reduce(246), // +, reduce: ContextVariable
			reduce(246), // -, reduce: ContextVariable
			reduce(246), // .+, reduce: ContextVariable
			reduce(246), // .-, reduce: ContextVariable
			reduce(246), // *, reduce: ContextVariable
			reduce(246), // /, reduce: ContextVariable
			reduce(246), // //, reduce: ContextVariable
			reduce(246), // %, reduce: ContextVariable
			reduce(246), // .*, reduce: ContextVariable
			reduce(246), // ./, reduce: ContextVariable
			reduce(246), // .//, reduce: ContextVariable
			reduce(246), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(246), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(246), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(246), // [[, reduce: ContextVariable
			reduce(246), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S127
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(247), // $, reduce: ContextVariable
			nil,         // empty
			reduce(247), // ;, reduce: ContextVariable
			reduce(247), // {, reduce: ContextVariable
			nil,         // }
			reduce(247), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(247), // >, reduce: ContextVariable
			reduce(247), // >>, reduce: ContextVariable
			reduce(247), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(247), // ||=, reduce: ContextVariable
			reduce(247), // ^^=, reduce: ContextVariable
			reduce(247), // &&=, reduce: ContextVariable
			reduce(247), // ??=, reduce: ContextVariable
			reduce(247), // ???=, reduce: ContextVariable
			reduce(247), // |=, reduce: ContextVariable
			reduce(247), // &=, reduce: ContextVariable
			reduce(247), // ^=, reduce: ContextVariable
			reduce(247), // <<=, reduce: ContextVariable
			reduce(247), // >>=, reduce: ContextVariable
			reduce(247), // >>>=, reduce: ContextVariable
			reduce(247), // +=, reduce: ContextVariable
			reduce(247), // .=, reduce: ContextVariable
			reduce(247), // -=, reduce: ContextVariable
			reduce(247), // *=, reduce: ContextVariable
			reduce(247), // /=, reduce: ContextVariable
			reduce(247), // //=, reduce: ContextVariable
			reduce(247), // %=, reduce: ContextVariable
			reduce(247), // **=, reduce: ContextVariable
			reduce(247), // ?, reduce: ContextVariable
			nil,         // :
			reduce(247), // ||, reduce: ContextVariable
			reduce(247), // ^^, reduce: ContextVariable
			reduce(247), // &&, reduce: ContextVariable
			reduce(247), // ??, reduce: ContextVariable
			reduce(247), // ???, reduce: ContextVariable
			reduce(247), // =~, reduce: ContextVariable
			reduce(247), // !=~, reduce: ContextVariable
			reduce(247), // ==, reduce: ContextVariable
			reduce(247), // !=, reduce: ContextVariable
			reduce(247), // <=>, reduce: ContextVariable
			reduce(247), // >=, reduce: ContextVariable
			reduce(247), // <, reduce: ContextVariable
			reduce(247), // <=, reduce: ContextVariable
			reduce(247), // ^, reduce: ContextVariable
			reduce(247), // &, reduce: ContextVariable
			reduce(247), // <<, reduce: ContextVariable
			reduce(247), // >>>, reduce: ContextVariable
			reduce(247), // +, reduce: ContextVariable
			reduce(247), // -, reduce: ContextVariable
			reduce(247), // .+, reduce: ContextVariable
			reduce(247), // .-, reduce: ContextVariable
			reduce(247), // *, reduce: ContextVariable
			reduce(247), // /, reduce: ContextVariable
			reduce(247), // //, reduce: ContextVariable
			reduce(247), // %, reduce: ContextVariable
			reduce(247), // .*, reduce: ContextVariable
			reduce(247), // ./, reduce: ContextVariable
			reduce(247), // .//, reduce: ContextVariable
			reduce(247), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(247), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(247), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(247), // [[, reduce: ContextVariable
			reduce(247), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S128
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(248), // $, reduce: ContextVariable
			nil,         // empty
			reduce(248), // ;, reduce: ContextVariable
			reduce(248), // {, reduce: ContextVariable
			nil,         // }
			reduce(248), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(248), // >, reduce: ContextVariable
			reduce(248), // >>, reduce: ContextVariable
			reduce(248), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(248), // ||=, reduce: ContextVariable
			reduce(248), // ^^=, reduce: ContextVariable
			reduce(248), // &&=, reduce: ContextVariable
			reduce(248), // ??=, reduce: ContextVariable
			reduce(248), // ???=, reduce: ContextVariable
			reduce(248), // |=, reduce: ContextVariable
			reduce(248), // &=, reduce: ContextVariable
			reduce(248), // ^=, reduce: ContextVariable
			reduce(248), // <<=, reduce: ContextVariable
			reduce(248), // >>=, reduce: ContextVariable
			reduce(248), // >>>=, reduce: ContextVariable
			reduce(248), // +=, reduce: ContextVariable
			reduce(248), // .=, reduce: ContextVariable
			reduce(248), // -=, reduce: ContextVariable
			reduce(248), // *=, reduce: ContextVariable
			reduce(248), // /=, reduce: ContextVariable
			reduce(248), // //=, reduce: ContextVariable
			reduce(248), // %=, reduce: ContextVariable
			reduce(248), // **=, reduce: ContextVariable
			reduce(248), // ?, reduce: ContextVariable
			nil,         // :
			reduce(248), // ||, reduce: ContextVariable
			reduce(248), // ^^, reduce: ContextVariable
			reduce(248), // &&, reduce: ContextVariable
			reduce(248), // ??, reduce: ContextVariable
			reduce(248), // ???, reduce: ContextVariable
			reduce(248), // =~, reduce: ContextVariable
			reduce(248), // !=~, reduce: ContextVariable
			reduce(248), // ==, reduce: ContextVariable
			reduce(248), // !=, reduce: ContextVariable
			reduce(248), // <=>, reduce: ContextVariable
			reduce(248), // >=, reduce: ContextVariable
			reduce(248), // <, reduce: ContextVariable
			reduce(248), // <=, reduce: ContextVariable
			reduce(248), // ^, reduce: ContextVariable
			reduce(248), // &, reduce: ContextVariable
			reduce(248), // <<, reduce: ContextVariable
			reduce(248), // >>>, reduce: ContextVariable
			reduce(248), // +, reduce: ContextVariable
			reduce(248), // -, reduce: ContextVariable
			reduce(248), // .+, reduce: ContextVariable
			reduce(248), // .-, reduce: ContextVariable
			reduce(248), // *, reduce: ContextVariable
			reduce(248), // /, reduce: ContextVariable
			reduce(248), // //, reduce: ContextVariable
			reduce(248), // %, reduce: ContextVariable
			reduce(248), // .*, reduce: ContextVariable
			reduce(248), // ./, reduce: ContextVariable
			reduce(248), // .//, reduce: ContextVariable
			reduce(248), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(248), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(248), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(248), // [[, reduce: ContextVariable
			reduce(248), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S129
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(249), // $, reduce: ContextVariable
			nil,         // empty
			reduce(249), // ;, reduce: ContextVariable
			reduce(249), // {, reduce: ContextVariable
			nil,         // }
			reduce(249), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(249), // >, reduce: ContextVariable
			reduce(249), // >>, reduce: ContextVariable
			reduce(249), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(249), // ||=, reduce: ContextVariable
			reduce(249), // ^^=, reduce: ContextVariable
			reduce(249), // &&=, reduce: ContextVariable
			reduce(249), // ??=, reduce: ContextVariable
			reduce(249), // ???=, reduce: ContextVariable
			reduce(249), // |=, reduce: ContextVariable
			reduce(249), // &=, reduce: ContextVariable
			reduce(249), // ^=, reduce: ContextVariable
			reduce(249), // <<=, reduce: ContextVariable
			reduce(249), // >>=, reduce: ContextVariable
			reduce(249), // >>>=, reduce: ContextVariable
			reduce(249), // +=, reduce: ContextVariable
			reduce(249), // .=, reduce: ContextVariable
			reduce(249), // -=, reduce: ContextVariable
			reduce(249), // *=, reduce: ContextVariable
			reduce(249), // /=, reduce: ContextVariable
			reduce(249), // //=, reduce: ContextVariable
			reduce(249), // %=, reduce: ContextVariable
			reduce(249), // **=, reduce: ContextVariable
			reduce(249), // ?, reduce: ContextVariable
			nil,         // :
			reduce(249), // ||, reduce: ContextVariable
			reduce(249), // ^^, reduce: ContextVariable
			reduce(249), // &&, reduce: ContextVariable
			reduce(249), // ??, reduce: ContextVariable
			reduce(249), // ???, reduce: ContextVariable
			reduce(249), // =~, reduce: ContextVariable
			reduce(249), // !=~, reduce: ContextVariable
			reduce(249), // ==, reduce: ContextVariable
			reduce(249), // !=, reduce: ContextVariable
			reduce(249), // <=>, reduce: ContextVariable
			reduce(249), // >=, reduce: ContextVariable
			reduce(249), // <, reduce: ContextVariable
			reduce(249), // <=, reduce: ContextVariable
			reduce(249), // ^, reduce: ContextVariable
			reduce(249), // &, reduce: ContextVariable
			reduce(249), // <<, reduce: ContextVariable
			reduce(249), // >>>, reduce: ContextVariable
			reduce(249), // +, reduce: ContextVariable
			reduce(249), // -, reduce: ContextVariable
			reduce(249), // .+, reduce: ContextVariable
			reduce(249), // .-, reduce: ContextVariable
			reduce(249), // *, reduce: ContextVariable
			reduce(249), // /, reduce: ContextVariable
			reduce(249), // //, reduce: ContextVariable
			reduce(249), // %, reduce: ContextVariable
			reduce(249), // .*, reduce: ContextVariable
			reduce(249), // ./, reduce: ContextVariable
			reduce(249), // .//, reduce: ContextVariable
			reduce(249), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(249), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(249), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(249), // [[, reduce: ContextVariable
			reduce(249), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S130
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(250), // $, reduce: ContextVariable
			nil,         // empty
			reduce(250), // ;, reduce: ContextVariable
			reduce(250), // {, reduce: ContextVariable
			nil,         // }
			reduce(250), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(250), // >, reduce: ContextVariable
			reduce(250), // >>, reduce: ContextVariable
			reduce(250), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(250), // ||=, reduce: ContextVariable
			reduce(250), // ^^=, reduce: ContextVariable
			reduce(250), // &&=, reduce: ContextVariable
			reduce(250), // ??=, reduce: ContextVariable
			reduce(250), // ???=, reduce: ContextVariable
			reduce(250), // |=, reduce: ContextVariable
			reduce(250), // &=, reduce: ContextVariable
			reduce(250), // ^=, reduce: ContextVariable
			reduce(250), // <<=, reduce: ContextVariable
			reduce(250), // >>=, reduce: ContextVariable
			reduce(250), // >>>=, reduce: ContextVariable
			reduce(250), // +=, reduce: ContextVariable
			reduce(250), // .=, reduce: ContextVariable
			reduce(250), // -=, reduce: ContextVariable
			reduce(250), // *=, reduce: ContextVariable
			reduce(250), // /=, reduce: ContextVariable
			reduce(250), // //=, reduce: ContextVariable
			reduce(250), // %=, reduce: ContextVariable
			reduce(250), // **=, reduce: ContextVariable
			reduce(250), // ?, reduce: ContextVariable
			nil,         // :
			reduce(250), // ||, reduce: ContextVariable
			reduce(250), // ^^, reduce: ContextVariable
			reduce(250), // &&, reduce: ContextVariable
			reduce(250), // ??, reduce: ContextVariable
			reduce(250), // ???, reduce: ContextVariable
			reduce(250), // =~, reduce: ContextVariable
			reduce(250), // !=~, reduce: ContextVariable
			reduce(250), // ==, reduce: ContextVariable
			reduce(250), // !=, reduce: ContextVariable
			reduce(250), // <=>, reduce: ContextVariable
			reduce(250), // >=, reduce: ContextVariable
			reduce(250), // <, reduce: ContextVariable
			reduce(250), // <=, reduce: ContextVariable
			reduce(250), // ^, reduce: ContextVariable
			reduce(250), // &, reduce: ContextVariable
			reduce(250), // <<, reduce: ContextVariable
			reduce(250), // >>>, reduce: ContextVariable
			reduce(250), // +, reduce: ContextVariable
			reduce(250), // -, reduce: ContextVariable
			reduce(250), // .+, reduce: ContextVariable
			reduce(250), // .-, reduce: ContextVariable
			reduce(250), // *, reduce: ContextVariable
			reduce(250), // /, reduce: ContextVariable
			reduce(250), // //, reduce: ContextVariable
			reduce(250), // %, reduce: ContextVariable
			reduce(250), // .*, reduce: ContextVariable
			reduce(250), // ./, reduce: ContextVariable
			reduce(250), // .//, reduce: ContextVariable
			reduce(250), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(250), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(250), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(250), // [[, reduce: ContextVariable
			reduce(250), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S131
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(251), // $, reduce: ContextVariable
			nil,         // empty
			reduce(251), // ;, reduce: ContextVariable
			reduce(251), // {, reduce: ContextVariable
			nil,         // }
			reduce(251), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(251), // >, reduce: ContextVariable
			reduce(251), // >>, reduce: ContextVariable
			reduce(251), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(251), // ||=, reduce: ContextVariable
			reduce(251), // ^^=, reduce: ContextVariable
			reduce(251), // &&=, reduce: ContextVariable
			reduce(251), // ??=, reduce: ContextVariable
			reduce(251), // ???=, reduce: ContextVariable
			reduce(251), // |=, reduce: ContextVariable
			reduce(251), // &=, reduce: ContextVariable
			reduce(251), // ^=, reduce: ContextVariable
			reduce(251), // <<=, reduce: ContextVariable
			reduce(251), // >>=, reduce: ContextVariable
			reduce(251), // >>>=, reduce: ContextVariable
			reduce(251), // +=, reduce: ContextVariable
			reduce(251), // .=, reduce: ContextVariable
			reduce(251), // -=, reduce: ContextVariable
			reduce(251), // *=, reduce: ContextVariable
			reduce(251), // /=, reduce: ContextVariable
			reduce(251), // //=, reduce: ContextVariable
			reduce(251), // %=, reduce: ContextVariable
			reduce(251), // **=, reduce: ContextVariable
			reduce(251), // ?, reduce: ContextVariable
			nil,         // :
			reduce(251), // ||, reduce: ContextVariable
			reduce(251), // ^^, reduce: ContextVariable
			reduce(251), // &&, reduce: ContextVariable
			reduce(251), // ??, reduce: ContextVariable
			reduce(251), // ???, reduce: ContextVariable
			reduce(251), // =~, reduce: ContextVariable
			reduce(251), // !=~, reduce: ContextVariable
			reduce(251), // ==, reduce: ContextVariable
			reduce(251), // !=, reduce: ContextVariable
			reduce(251), // <=>, reduce: ContextVariable
			reduce(251), // >=, reduce: ContextVariable
			reduce(251), // <, reduce: ContextVariable
			reduce(251), // <=, reduce: ContextVariable
			reduce(251), // ^, reduce: ContextVariable
			reduce(251), // &, reduce: ContextVariable
			reduce(251), // <<, reduce: ContextVariable
			reduce(251), // >>>, reduce: ContextVariable
			reduce(251), // +, reduce: ContextVariable
			reduce(251), // -, reduce: ContextVariable
			reduce(251), // .+, reduce: ContextVariable
			reduce(251), // .-, reduce: ContextVariable
			reduce(251), // *, reduce: ContextVariable
			reduce(251), // /, reduce: ContextVariable
			reduce(251), // //, reduce: ContextVariable
			reduce(251), // %, reduce: ContextVariable
			reduce(251), // .*, reduce: ContextVariable
			reduce(251), // ./, reduce: ContextVariable
			reduce(251), // .//, reduce: ContextVariable
			reduce(251), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(251), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(251), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(251), // [[, reduce: ContextVariable
			reduce(251), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S132
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(252), // $, reduce: ContextVariable
			nil,         // empty
			reduce(252), // ;, reduce: ContextVariable
			reduce(252), // {, reduce: ContextVariable
			nil,         // }
			reduce(252), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(252), // >, reduce: ContextVariable
			reduce(252), // >>, reduce: ContextVariable
			reduce(252), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(252), // ||=, reduce: ContextVariable
			reduce(252), // ^^=, reduce: ContextVariable
			reduce(252), // &&=, reduce: ContextVariable
			reduce(252), // ??=, reduce: ContextVariable
			reduce(252), // ???=, reduce: ContextVariable
			reduce(252), // |=, reduce: ContextVariable
			reduce(252), // &=, reduce: ContextVariable
			reduce(252), // ^=, reduce: ContextVariable
			reduce(252), // <<=, reduce: ContextVariable
			reduce(252), // >>=, reduce: ContextVariable
			reduce(252), // >>>=, reduce: ContextVariable
			reduce(252), // +=, reduce: ContextVariable
			reduce(252), // .=, reduce: ContextVariable
			reduce(252), // -=, reduce: ContextVariable
			reduce(252), // *=, reduce: ContextVariable
			reduce(252), // /=, reduce: ContextVariable
			reduce(252), // //=, reduce: ContextVariable
			reduce(252), // %=, reduce: ContextVariable
			reduce(252), // **=, reduce: ContextVariable
			reduce(252), // ?, reduce: ContextVariable
			nil,         // :
			reduce(252), // ||, reduce: ContextVariable
			reduce(252), // ^^, reduce: ContextVariable
			reduce(252), // &&, reduce: ContextVariable
			reduce(252), // ??, reduce: ContextVariable
			reduce(252), // ???, reduce: ContextVariable
			reduce(252), // =~, reduce: ContextVariable
			reduce(252), // !=~, reduce: ContextVariable
			reduce(252), // ==, reduce: ContextVariable
			reduce(252), // !=, reduce: ContextVariable
			reduce(252), // <=>, reduce: ContextVariable
			reduce(252), // >=, reduce: ContextVariable
			reduce(252), // <, reduce: ContextVariable
			reduce(252), // <=, reduce: ContextVariable
			reduce(252), // ^, reduce: ContextVariable
			reduce(252), // &, reduce: ContextVariable
			reduce(252), // <<, reduce: ContextVariable
			reduce(252), // >>>, reduce: ContextVariable
			reduce(252), // +, reduce: ContextVariable
			reduce(252), // -, reduce: ContextVariable
			reduce(252), // .+, reduce: ContextVariable
			reduce(252), // .-, reduce: ContextVariable
			reduce(252), // *, reduce: ContextVariable
			reduce(252), // /, reduce: ContextVariable
			reduce(252), // //, reduce: ContextVariable
			reduce(252), // %, reduce: ContextVariable
			reduce(252), // .*, reduce: ContextVariable
			reduce(252), // ./, reduce: ContextVariable
			reduce(252), // .//, reduce: ContextVariable
			reduce(252), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(252), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(252), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(252), // [[, reduce: ContextVariable
			reduce(252), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S133
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(253), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(253), // ;, reduce: MlrvalOrFunction
			reduce(253), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(253), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(253), // >, reduce: MlrvalOrFunction
			reduce(253), // >>, reduce: MlrvalOrFunction
			reduce(253), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(253), // ||=, reduce: MlrvalOrFunction
			reduce(253), // ^^=, reduce: MlrvalOrFunction
			reduce(253), // &&=, reduce: MlrvalOrFunction
			reduce(253), // ??=, reduce: MlrvalOrFunction
			reduce(253), // ???=, reduce: MlrvalOrFunction
			reduce(253), // |=, reduce: MlrvalOrFunction
			reduce(253), // &=, reduce: MlrvalOrFunction
			reduce(253), // ^=, reduce: MlrvalOrFunction
			reduce(253), // <<=, reduce: MlrvalOrFunction
			reduce(253), // >>=, reduce: MlrvalOrFunction
			reduce(253), // >>>=, reduce: MlrvalOrFunction
			reduce(253), // +=, reduce: MlrvalOrFunction
			reduce(253), // .=, reduce: MlrvalOrFunction
			reduce(253), // -=, reduce: MlrvalOrFunction
			reduce(253), // *=, reduce: MlrvalOrFunction
			reduce(253), // /=, reduce: MlrvalOrFunction
			reduce(253), // //=, reduce: MlrvalOrFunction
			reduce(253), // %=, reduce: MlrvalOrFunction
			reduce(253), // **=, reduce: MlrvalOrFunction
			reduce(253), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(253), // ||, reduce: MlrvalOrFunction
			reduce(253), // ^^, reduce: MlrvalOrFunction
			reduce(253), // &&, reduce: MlrvalOrFunction
			reduce(253), // ??, reduce: MlrvalOrFunction
			reduce(253), // ???, reduce: MlrvalOrFunction
			reduce(253), // =~, reduce: MlrvalOrFunction
			reduce(253), // !=~, reduce: MlrvalOrFunction
			reduce(253), // ==, reduce: MlrvalOrFunction
			reduce(253), // !=, reduce: MlrvalOrFunction
			reduce(253), // <=>, reduce: MlrvalOrFunction
			reduce(253), // >=, reduce: MlrvalOrFunction
			reduce(253), // <, reduce: MlrvalOrFunction
			reduce(253), // <=, reduce: MlrvalOrFunction
			reduce(253), // ^, reduce: MlrvalOrFunction
			reduce(253), // &, reduce: MlrvalOrFunction
			reduce(253), // <<, reduce: MlrvalOrFunction
			reduce(253), // >>>, reduce: MlrvalOrFunction
			reduce(253), // +, reduce: MlrvalOrFunction
			reduce(253), // -, reduce: MlrvalOrFunction
			reduce(253), // .+, reduce: MlrvalOrFunction
			reduce(253), // .-, reduce: MlrvalOrFunction
			reduce(253), // *, reduce: MlrvalOrFunction
			reduce(253), // /, reduce: MlrvalOrFunction
			reduce(253), // //, reduce: MlrvalOrFunction
			reduce(253), // %, reduce: MlrvalOrFunction
			reduce(253), // .*, reduce: MlrvalOrFunction
			reduce(253), // ./, reduce: MlrvalOrFunction
			reduce(253), // .//, reduce: MlrvalOrFunction
			reduce(253), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(253), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(253), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(253), // [[, reduce: MlrvalOrFunction
			reduce(253), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S134
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			shift(845), // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			shift(846), // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S135
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(256), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(256), // ;, reduce: MlrvalOrFunction
			reduce(256), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(256), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(256), // >, reduce: MlrvalOrFunction
			reduce(256), // >>, reduce: MlrvalOrFunction
			reduce(256), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(256), // ||=, reduce: MlrvalOrFunction
			reduce(256), // ^^=, reduce: MlrvalOrFunction
			reduce(256), // &&=, reduce: MlrvalOrFunction
			reduce(256), // ??=, reduce: MlrvalOrFunction
			reduce(256), // ???=, reduce: MlrvalOrFunction
			reduce(256), // |=, reduce: MlrvalOrFunction
			reduce(256), // &=, reduce: MlrvalOrFunction
			reduce(256), // ^=, reduce: MlrvalOrFunction
			reduce(256), // <<=, reduce: MlrvalOrFunction
			reduce(256), // >>=, reduce: MlrvalOrFunction
			reduce(256), // >>>=, reduce: MlrvalOrFunction
			reduce(256), // +=, reduce: MlrvalOrFunction
			reduce(256), // .=, reduce: MlrvalOrFunction
			reduce(256), // -=, reduce: MlrvalOrFunction
			reduce(256), // *=, reduce: MlrvalOrFunction
			reduce(256), // /=, reduce: MlrvalOrFunction
			reduce(256), // //=, reduce: MlrvalOrFunction
			reduce(256), // %=, reduce: MlrvalOrFunction
			reduce(256), // **=, reduce: MlrvalOrFunction
			reduce(256), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(256), // ||, reduce: MlrvalOrFunction
			reduce(256), // ^^, reduce: MlrvalOrFunction
			reduce(256), // &&, reduce: MlrvalOrFunction
			reduce(256), // ??, reduce: MlrvalOrFunction
			reduce(256), // ???, reduce: MlrvalOrFunction
			reduce(256), // =~, reduce: MlrvalOrFunction
			reduce(256), // !=~, reduce: MlrvalOrFunction
			reduce(256), // ==, reduce: MlrvalOrFunction
			reduce(256), // !=, reduce: MlrvalOrFunction
			reduce(256), // <=>, reduce: MlrvalOrFunction
			reduce(256), // >=, reduce: MlrvalOrFunction
			reduce(256), // <, reduce: MlrvalOrFunction
			reduce(256), // <=, reduce: MlrvalOrFunction
			reduce(256), // ^, reduce: MlrvalOrFunction
			reduce(256), // &, reduce: MlrvalOrFunction
			reduce(256), // <<, reduce: MlrvalOrFunction
			reduce(256), // >>>, reduce: MlrvalOrFunction
			reduce(256), // +, reduce: MlrvalOrFunction
			reduce(256), // -, reduce: MlrvalOrFunction
			reduce(256), // .+, reduce: MlrvalOrFunction
			reduce(256), // .-, reduce: MlrvalOrFunction
			reduce(256), // *, reduce: MlrvalOrFunction
			reduce(256), // /, reduce: MlrvalOrFunction
			reduce(256), // //, reduce: MlrvalOrFunction
			reduce(256), // %, reduce: MlrvalOrFunction
			reduce(256), // .*, reduce: MlrvalOrFunction
			reduce(256), // ./, reduce: MlrvalOrFunction
			reduce(256), // .//, reduce: MlrvalOrFunction
			reduce(256), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(256), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(256), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(256), // [[, reduce: MlrvalOrFunction
			reduce(256), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S136
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(257), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(257), // ;, reduce: MlrvalOrFunction
			reduce(257), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(257), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(257), // >, reduce: MlrvalOrFunction
			reduce(257), // >>, reduce: MlrvalOrFunction
			reduce(257), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(257), // ||=, reduce: MlrvalOrFunction
			reduce(257), // ^^=, reduce: MlrvalOrFunction
			reduce(257), // &&=, reduce: MlrvalOrFunction
			reduce(257), // ??=, reduce: MlrvalOrFunction
			reduce(257), // ???=, reduce: MlrvalOrFunction
			reduce(257), // |=, reduce: MlrvalOrFunction
			reduce(257), // &=, reduce: MlrvalOrFunction
			reduce(257), // ^=, reduce: MlrvalOrFunction
			reduce(257), // <<=, reduce: MlrvalOrFunction
			reduce(257), // >>=, reduce: MlrvalOrFunction
			reduce(257), // >>>=, reduce: MlrvalOrFunction
			reduce(257), // +=, reduce: MlrvalOrFunction
			reduce(257), // .=, reduce: MlrvalOrFunction
			reduce(257), // -=, reduce: MlrvalOrFunction
			reduce(257), // *=, reduce: MlrvalOrFunction
			reduce(257), // /=, reduce: MlrvalOrFunction
			reduce(257), // //=, reduce: MlrvalOrFunction
			reduce(257), // %=, reduce: MlrvalOrFunction
			reduce(257), // **=, reduce: MlrvalOrFunction
			reduce(257), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(257), // ||, reduce: MlrvalOrFunction
			reduce(257), // ^^, reduce: MlrvalOrFunction
			reduce(257), // &&, reduce: MlrvalOrFunction
			reduce(257), // ??, reduce: MlrvalOrFunction
			reduce(257), // ???, reduce: MlrvalOrFunction
			reduce(257), // =~, reduce: MlrvalOrFunction
			reduce(257), // !=~, reduce: MlrvalOrFunction
			reduce(257), // ==, reduce: MlrvalOrFunction
			reduce(257), // !=, reduce: MlrvalOrFunction
			reduce(257), // <=>, reduce: MlrvalOrFunction
			reduce(257), // >=, reduce: MlrvalOrFunction
			reduce(257), // <, reduce: MlrvalOrFunction
			reduce(257), // <=, reduce: MlrvalOrFunction
			reduce(257), // ^, reduce: MlrvalOrFunction
			reduce(257), // &, reduce: MlrvalOrFunction
			reduce(257), // <<, reduce: MlrvalOrFunction
			reduce(257), // >>>, reduce: MlrvalOrFunction
			reduce(257), // +, reduce: MlrvalOrFunction
			reduce(257), // -, reduce: MlrvalOrFunction
			reduce(257), // .+, reduce: MlrvalOrFunction
			reduce(257), // .-, reduce: MlrvalOrFunction
			reduce(257), // *, reduce: MlrvalOrFunction
			reduce(257), // /, reduce: MlrvalOrFunction
			reduce(257), // //, reduce: MlrvalOrFunction
			reduce(257), // %, reduce: MlrvalOrFunction
			reduce(257), // .*, reduce: MlrvalOrFunction
			reduce(257), // ./, reduce: MlrvalOrFunction
			reduce(257), // .//, reduce: MlrvalOrFunction
			reduce(257), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(257), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(257), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(257), // [[, reduce: MlrvalOrFunction
			reduce(257), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S137
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(258), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(258), // ;, reduce: MlrvalOrFunction
			reduce(258), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(258), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(258), // >, reduce: MlrvalOrFunction
			reduce(258), // >>, reduce: MlrvalOrFunction
			reduce(258), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(258), // ||=, reduce: MlrvalOrFunction
			reduce(258), // ^^=, reduce: MlrvalOrFunction
			reduce(258), // &&=, reduce: MlrvalOrFunction
			reduce(258), // ??=, reduce: MlrvalOrFunction
			reduce(258), // ???=, reduce: MlrvalOrFunction
			reduce(258), // |=, reduce: MlrvalOrFunction
			reduce(258), // &=, reduce: MlrvalOrFunction
			reduce(258), // ^=, reduce: MlrvalOrFunction
			reduce(258), // <<=, reduce: MlrvalOrFunction
			reduce(258), // >>=, reduce: MlrvalOrFunction
			reduce(258), // >>>=, reduce: MlrvalOrFunction
			reduce(258), // +=, reduce: MlrvalOrFunction
			reduce(258), // .=, reduce: MlrvalOrFunction
			reduce(258), // -=, reduce: MlrvalOrFunction
			reduce(258), // *=, reduce: MlrvalOrFunction
			reduce(258), // /=, reduce: MlrvalOrFunction
			reduce(258), // //=, reduce: MlrvalOrFunction
			reduce(258), // %=, reduce: MlrvalOrFunction
			reduce(258), // **=, reduce: MlrvalOrFunction
			reduce(258), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(258), // ||, reduce: MlrvalOrFunction
			reduce(258), // ^^, reduce: MlrvalOrFunction
			reduce(258), // &&, reduce: MlrvalOrFunction
			reduce(258), // ??, reduce: MlrvalOrFunction
			reduce(258), // ???, reduce: MlrvalOrFunction
			reduce(258), // =~, reduce: MlrvalOrFunction
			reduce(258), // !=~, reduce: MlrvalOrFunction
			reduce(258), // ==, reduce: MlrvalOrFunction
			reduce(258), // !=, reduce: MlrvalOrFunction
			reduce(258), // <=>, reduce: MlrvalOrFunction
			reduce(258), // >=, reduce: MlrvalOrFunction
			reduce(258), // <, reduce: MlrvalOrFunction
			reduce(258), // <=, reduce: MlrvalOrFunction
			reduce(258), // ^, reduce: MlrvalOrFunction
			reduce(258), // &, reduce: MlrvalOrFunction
			reduce(258), // <<, reduce: MlrvalOrFunction
			reduce(258), // >>>, reduce: MlrvalOrFunction
			reduce(258), // +, reduce: MlrvalOrFunction
			reduce(258), // -, reduce: MlrvalOrFunction
			reduce(258), // .+, reduce: MlrvalOrFunction
			reduce(258), // .-, reduce: MlrvalOrFunction
			reduce(258), // *, reduce: MlrvalOrFunction
			reduce(258), // /, reduce: MlrvalOrFunction
			reduce(258), // //, reduce: MlrvalOrFunction
			reduce(258), // %, reduce: MlrvalOrFunction
			reduce(258), // .*, reduce: MlrvalOrFunction
			reduce(258), // ./, reduce: MlrvalOrFunction
			reduce(258), // .//, reduce: MlrvalOrFunction
			reduce(258), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(258), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(258), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(258), // [[, reduce: MlrvalOrFunction
			reduce(258), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S138
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(259), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(259), // ;, reduce: MlrvalOrFunction
			reduce(259), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(259), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(259), // >, reduce: MlrvalOrFunction
			reduce(259), // >>, reduce: MlrvalOrFunction
			reduce(259), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(259), // ||=, reduce: MlrvalOrFunction
			reduce(259), // ^^=, reduce: MlrvalOrFunction
			reduce(259), // &&=, reduce: MlrvalOrFunction
			reduce(259), // ??=, reduce: MlrvalOrFunction
			reduce(259), // ???=, reduce: MlrvalOrFunction
			reduce(259), // |=, reduce: MlrvalOrFunction
			reduce(259), // &=, reduce: MlrvalOrFunction
			reduce(259), // ^=, reduce: MlrvalOrFunction
			reduce(259), // <<=, reduce: MlrvalOrFunction
			reduce(259), // >>=, reduce: MlrvalOrFunction
			reduce(259), // >>>=, reduce: MlrvalOrFunction
			reduce(259), // +=, reduce: MlrvalOrFunction
			reduce(259), // .=, reduce: MlrvalOrFunction
			reduce(259), // -=, reduce: MlrvalOrFunction
			reduce(259), // *=, reduce: MlrvalOrFunction
			reduce(259), // /=, reduce: MlrvalOrFunction
			reduce(259), // //=, reduce: MlrvalOrFunction
			reduce(259), // %=, reduce: MlrvalOrFunction
			reduce(259), // **=, reduce: MlrvalOrFunction
			reduce(259), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(259), // ||, reduce: MlrvalOrFunction
			reduce(259), // ^^, reduce: MlrvalOrFunction
			reduce(259), // &&, reduce: MlrvalOrFunction
			reduce(259), // ??, reduce: MlrvalOrFunction
			reduce(259), // ???, reduce: MlrvalOrFunction
			reduce(259), // =~, reduce: MlrvalOrFunction
			reduce(259), // !=~, reduce: MlrvalOrFunction
			reduce(259), // ==, reduce: MlrvalOrFunction
			reduce(259), // !=, reduce: MlrvalOrFunction
			reduce(259), // <=>, reduce: MlrvalOrFunction
			reduce(259), // >=, reduce: MlrvalOrFunction
			reduce(259), // <, reduce: MlrvalOrFunction
			reduce(259), // <=, reduce: MlrvalOrFunction
			reduce(259), // ^, reduce: MlrvalOrFunction
			reduce(259), // &, reduce: MlrvalOrFunction
			reduce(259), // <<, reduce: MlrvalOrFunction
			reduce(259), // >>>, reduce: MlrvalOrFunction
			reduce(259), // +, reduce: MlrvalOrFunction
			reduce(259), // -, reduce: MlrvalOrFunction
			reduce(259), // .+, reduce: MlrvalOrFunction
			reduce(259), // .-, reduce: MlrvalOrFunction
			reduce(259), // *, reduce: MlrvalOrFunction
			reduce(259), // /, reduce: MlrvalOrFunction
			reduce(259), // //, reduce: MlrvalOrFunction
			reduce(259), // %, reduce: MlrvalOrFunction
			reduce(259), // .*, reduce: MlrvalOrFunction
			reduce(259), // ./, reduce: MlrvalOrFunction
			reduce(259), // .//, reduce: MlrvalOrFunction
			reduce(259), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(259), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(259), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(259), // [[, reduce: MlrvalOrFunction
			reduce(259), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S139
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(267), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(267), // ;, reduce: MlrvalOrFunction
			reduce(267), // {, reduce: MlrvalOrFunction
			nil,         // }
			reduce(267), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(267), // >, reduce: MlrvalOrFunction
			reduce(267), // >>, reduce: MlrvalOrFunction
			reduce(267), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(267), // ||=, reduce: MlrvalOrFunction
			reduce(267), // ^^=, reduce: MlrvalOrFunction
			reduce(267), // &&=, reduce: MlrvalOrFunction
			reduce(267), // ??=, reduce: MlrvalOrFunction
			reduce(267), // ???=, reduce: MlrvalOrFunction
			reduce(267), // |=, reduce: MlrvalOrFunction
			reduce(267), // &=, reduce: MlrvalOrFunction
			reduce(267), // ^=, reduce: MlrvalOrFunction
			reduce(267), // <<=, reduce: MlrvalOrFunction
			reduce(267), // >>=, reduce: MlrvalOrFunction
			reduce(267), // >>>=, reduce: MlrvalOrFunction
			reduce(267), // +=, reduce: MlrvalOrFunction
			reduce(267), // .=, reduce: MlrvalOrFunction
			reduce(267), // -=, reduce: MlrvalOrFunction
			reduce(267), // *=, reduce: MlrvalOrFunction
			reduce(267), // /=, reduce: MlrvalOrFunction
			reduce(267), // //=, reduce: MlrvalOrFunction
			reduce(267), // %=, reduce: MlrvalOrFunction
			reduce(267), // **=, reduce: MlrvalOrFunction
			reduce(267), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(267), // ||, reduce: MlrvalOrFunction
			reduce(267), // ^^, reduce: MlrvalOrFunction
			reduce(267), // &&, reduce: MlrvalOrFunction
			reduce(267), // ??, reduce: MlrvalOrFunction
			reduce(267), // ???, reduce: MlrvalOrFunction
			reduce(267), // =~, reduce: MlrvalOrFunction
			reduce(267), // !=~, reduce: MlrvalOrFunction
			reduce(267), // ==, reduce: MlrvalOrFunction
			reduce(267), // !=, reduce: MlrvalOrFunction
			reduce(267), // <=>, reduce: MlrvalOrFunction
			reduce(267), // >=, reduce: MlrvalOrFunction
			reduce(267), // <, reduce: MlrvalOrFunction
			reduce(267), // <=, reduce: MlrvalOrFunction
			reduce(267), // ^, reduce: MlrvalOrFunction
			reduce(267), // &, reduce: MlrvalOrFunction
			reduce(267), // <<, reduce: MlrvalOrFunction
			reduce(267), // >>>, reduce: MlrvalOrFunction
			reduce(267), // +, reduce: MlrvalOrFunction
			reduce(267), // -, reduce: MlrvalOrFunction
			reduce(267), // .+, reduce: MlrvalOrFunction
			reduce(267), // .-, reduce: MlrvalOrFunction
			reduce(267), // *, reduce: MlrvalOrFunction
			reduce(267), // /, reduce: MlrvalOrFunction
			reduce(267), // //, reduce: MlrvalOrFunction
			reduce(267), // %, reduce: MlrvalOrFunction
			reduce(267), // .*, reduce: MlrvalOrFunction
			reduce(267), // ./, reduce: MlrvalOrFunction
			reduce(267), // .//, reduce: MlrvalOrFunction
			reduce(267), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(267), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(267), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(267), // [[, reduce: MlrvalOrFunction
			reduce(267), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S140
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(847), // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S141
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			shift(848), // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S142
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(279), // $, reduce: BracefulStatement
			nil,         // empty
			reduce(279), // ;, reduce: BracefulStatement
			reduce(279), // {, reduce: BracefulStatement
			nil,         // }
			nil,         // =
			reduce(279), // unset, reduce: BracefulStatement
			reduce(279), // filter, reduce: BracefulStatement
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(279), // print, reduce: BracefulStatement
			nil,         // ,
			reduce(279), // printn, reduce: BracefulStatement
			reduce(279), // eprint, reduce: BracefulStatement
			reduce(279), // eprintn, reduce: BracefulStatement
			reduce(279), // dump, reduce: BracefulStatement
			reduce(279), // edump, reduce: BracefulStatement
			reduce(279), // tee, reduce: BracefulStatement
			reduce(279), // emitf, reduce: BracefulStatement
			reduce(279), // emit1, reduce: BracefulStatement
			reduce(279), // emit, reduce: BracefulStatement
			reduce(279), // (, reduce: BracefulStatement
			nil,         // )
			reduce(279), // emitp, reduce: BracefulStatement
			reduce(279), // field_name, reduce: BracefulStatement
			reduce(279), // $[, reduce: BracefulStatement
			nil,         // ]
			reduce(279), // braced_field_name, reduce: BracefulStatement
			reduce(279), // $[[, reduce: BracefulStatement
			reduce(279), // $[[[, reduce: BracefulStatement
			reduce(279), // full_srec, reduce: BracefulStatement
			reduce(279), // oosvar_name, reduce: BracefulStatement
			reduce(279), // @[, reduce: BracefulStatement
			reduce(279), // braced_oosvar_name, reduce: BracefulStatement
			reduce(279), // full_oosvar, reduce: BracefulStatement
			reduce(279), // all, reduce: BracefulStatement
			reduce(279), // non_sigil_name, reduce: BracefulStatement
			reduce(279), // arr, reduce: BracefulStatement
			reduce(279), // bool, reduce: BracefulStatement
			reduce(279), // float, reduce: BracefulStatement
			reduce(279), // int, reduce: BracefulStatement
			reduce(279), // map, reduce: BracefulStatement
			reduce(279), // num, reduce: BracefulStatement
			reduce(279), // str, reduce: BracefulStatement
			reduce(279), // var, reduce: BracefulStatement
			reduce(279), // funct, reduce: BracefulStatement
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(279), // +, reduce: BracefulStatement
			reduce(279), // -, reduce: BracefulStatement
			reduce(279), // .+, reduce: BracefulStatement
			reduce(279), // .-, reduce: BracefulStatement
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(279), // !, reduce: BracefulStatement
			reduce(279), // ~, reduce: BracefulStatement
			nil,         // **
			reduce(279), // string_literal, reduce: BracefulStatement
			reduce(279), // regex_case_insensitive, reduce: BracefulStatement
			reduce(279), // int_literal, reduce: BracefulStatement
			reduce(279), // float_literal, reduce: BracefulStatement
			reduce(279), // boolean_literal, reduce: BracefulStatement
			reduce(279), // null_literal, reduce: BracefulStatement
			reduce(279), // inf_literal, reduce: BracefulStatement
			reduce(279), // nan_literal, reduce: BracefulStatement
			reduce(279), // const_M_PI, reduce: BracefulStatement
			reduce(279), // const_M_E, reduce: BracefulStatement
			reduce(279), // panic, reduce: BracefulStatement
			reduce(279), // [, reduce: BracefulStatement
			reduce(279), // ctx_IPS, reduce: BracefulStatement
			reduce(279), // ctx_IFS, reduce: BracefulStatement
			reduce(279), // ctx_IRS, reduce: BracefulStatement
			reduce(279), // ctx_OPS, reduce: BracefulStatement
			reduce(279), // ctx_OFS, reduce: BracefulStatement
			reduce(279), // ctx_ORS, reduce: BracefulStatement
			reduce(279), // ctx_FLATSEP, reduce: BracefulStatement
			reduce(279), // ctx_NF, reduce: BracefulStatement
			reduce(279), // ctx_NR, reduce: BracefulStatement
			reduce(279), // ctx_FNR, reduce: BracefulStatement
			reduce(279), // ctx_FILENAME, reduce: BracefulStatement
			reduce(279), // ctx_FILENUM, reduce: BracefulStatement
			reduce(279), // env, reduce: BracefulStatement
			nil,         // [[
			nil,         // [[[
			reduce(279), // call, reduce: BracefulStatement
			reduce(279), // begin, reduce: BracefulStatement
			reduce(279), // end, reduce: BracefulStatement
			reduce(279), // if, reduce: BracefulStatement
			nil,         // elif
			nil,         // else
			reduce(279), // while, reduce: BracefulStatement
			reduce(279), // do, reduce: BracefulStatement
			reduce(279), // for, reduce: BracefulStatement
			nil,         // in
			reduce(279), // break, reduce: BracefulStatement
			reduce(279), // continue, reduce: BracefulStatement
			reduce(279), // func, reduce: BracefulStatement
			reduce(279), // subr, reduce: BracefulStatement
			reduce(279), // return, reduce: BracefulStatement
		},
	},
	actionRow{ // S143
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(280), // $, reduce: BracefulStatement
			nil,         // empty
			reduce(280), // ;, reduce: BracefulStatement
			reduce(280), // {, reduce: BracefulStatement
			nil,         // }
			nil,         // =
			reduce(280), // unset, reduce: BracefulStatement
			reduce(280), // filter, reduce: BracefulStatement
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(280), // print, reduce: BracefulStatement
			nil,         // ,
			reduce(280), // printn, reduce: BracefulStatement
			reduce(280), // eprint, reduce: BracefulStatement
			reduce(280), // eprintn, reduce: BracefulStatement
			reduce(280), // dump, reduce: BracefulStatement
			reduce(280), // edump, reduce: BracefulStatement
			reduce(280), // tee, reduce: BracefulStatement
			reduce(280), // emitf, reduce: BracefulStatement
			reduce(280), // emit1, reduce: BracefulStatement
			reduce(280), // emit, reduce: BracefulStatement
			reduce(280), // (, reduce: BracefulStatement
			nil,         // )
			reduce(280), // emitp, reduce: BracefulStatement
			reduce(280), // field_name, reduce: BracefulStatement
			reduce(280), // $[, reduce: BracefulStatement
			nil,         // ]
			reduce(280), // braced_field_name, reduce: BracefulStatement
			reduce(280), // $[[, reduce: BracefulStatement
			reduce(280), // $[[[, reduce: BracefulStatement
			reduce(280), // full_srec, reduce: BracefulStatement
			reduce(280), // oosvar_name, reduce: BracefulStatement
			reduce(280), // @[, reduce: BracefulStatement
			reduce(280), // braced_oosvar_name, reduce: BracefulStatement
			reduce(280), // full_oosvar, reduce: BracefulStatement
			reduce(280), // all, reduce: BracefulStatement
			reduce(280), // non_sigil_name, reduce: BracefulStatement
			reduce(280), // arr, reduce: BracefulStatement
			reduce(280), // bool, reduce: BracefulStatement
			reduce(280), // float, reduce: BracefulStatement
			reduce(280), // int, reduce: BracefulStatement
			reduce(280), // map, reduce: BracefulStatement
			reduce(280), // num, reduce: BracefulStatement
			reduce(280), // str, reduce: BracefulStatement
			reduce(280), // var, reduce: BracefulStatement
			reduce(280), // funct, reduce: BracefulStatement
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(280), // +, reduce: BracefulStatement
			reduce(280), // -, reduce: BracefulStatement
			reduce(280), // .+, reduce: BracefulStatement
			reduce(280), // .-, reduce: BracefulStatement
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(280), // !, reduce: BracefulStatement
			reduce(280), // ~, reduce: BracefulStatement
			nil,         // **
			reduce(280), // string_literal, reduce: BracefulStatement
			reduce(280), // regex_case_insensitive, reduce: BracefulStatement
			reduce(280), // int_literal, reduce: BracefulStatement
			reduce(280), // float_literal, reduce: BracefulStatement
			reduce(280), // boolean_literal, reduce: BracefulStatement
			reduce(280), // null_literal, reduce: BracefulStatement
			reduce(280), // inf_literal, reduce: BracefulStatement
			reduce(280), // nan_literal, reduce: BracefulStatement
			reduce(280), // const_M_PI, reduce: BracefulStatement
			reduce(280), // const_M_E, reduce: BracefulStatement
			reduce(280), // panic, reduce: BracefulStatement
			reduce(280), // [, reduce: BracefulStatement
			reduce(280), // ctx_IPS, reduce: BracefulStatement
			reduce(280), // ctx_IFS, reduce: BracefulStatement
			reduce(280), // ctx_IRS, reduce: BracefulStatement
			reduce(280), // ctx_OPS, reduce: BracefulStatement
			reduce(280), // ctx_OFS, reduce: BracefulStatement
			reduce(280), // ctx_ORS, reduce: BracefulStatement
			reduce(280), // ctx_FLATSEP, reduce: BracefulStatement
			reduce(280), // ctx_NF, reduce: BracefulStatement
			reduce(280), // ctx_NR, reduce: BracefulStatement
			reduce(280), // ctx_FNR, reduce: BracefulStatement
			reduce(280), // ctx_FILENAME, reduce: BracefulStatement
			reduce(280), // ctx_FILENUM, reduce: BracefulStatement
			reduce(280), // env, reduce: BracefulStatement
			nil,         // [[
			nil,         // [[[
			reduce(280), // call, reduce: BracefulStatement
			reduce(280), // begin, reduce: BracefulStatement
			reduce(280), // end, reduce: BracefulStatement
			reduce(280), // if, reduce: BracefulStatement
			nil,         // elif
			nil,         // else
			reduce(280), // while, reduce: BracefulStatement
			reduce(280), // do, reduce: BracefulStatement
			reduce(280), // for, reduce: BracefulStatement
			nil,         // in
			reduce(280), // break, reduce: BracefulStatement
			reduce(280), // continue, reduce: BracefulStatement
			reduce(280), // func, reduce: BracefulStatement
			reduce(280), // subr, reduce: BracefulStatement
			reduce(280), // return, reduce: BracefulStatement
		},
	},
	actionRow{ // S144
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(281), // $, reduce: BracefulStatement
			nil,         // empty
			reduce(281), // ;, reduce: BracefulStatement
			reduce(281), // {, reduce: BracefulStatement
			nil,         // }
			nil,         // =
			reduce(281), // unset, reduce: BracefulStatement
			reduce(281), // filter, reduce: BracefulStatement
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(281), // print, reduce: BracefulStatement
			nil,         // ,
			reduce(281), // printn, reduce: BracefulStatement
			reduce(281), // eprint, reduce: BracefulStatement
			reduce(281), // eprintn, reduce: BracefulStatement
			reduce(281), // dump, reduce: BracefulStatement
			reduce(281), // edump, reduce: BracefulStatement
			reduce(281), // tee, reduce: BracefulStatement
			reduce(281), // emitf, reduce: BracefulStatement
			reduce(281), // emit1, reduce: BracefulStatement
			reduce(281), // emit, reduce: BracefulStatement
			reduce(281), // (, reduce: BracefulStatement
			nil,         // )
			reduce(281), // emitp, reduce: BracefulStatement
			reduce(281), // field_name, reduce: BracefulStatement
			reduce(281), // $[, reduce: BracefulStatement
			nil,         // ]
			reduce(281), // braced_field_name, reduce: BracefulStatement
			reduce(281), // $[[, reduce: BracefulStatement
			reduce(281), // $[[[, reduce: BracefulStatement
			reduce(281), // full_srec, reduce: BracefulStatement
			reduce(281), // oosvar_name, reduce: BracefulStatement
			reduce(281), // @[, reduce: BracefulStatement
			reduce(281), // braced_oosvar_name, reduce: BracefulStatement
			reduce(281), // full_oosvar, reduce: BracefulStatement
			reduce(281), // all, reduce: BracefulStatement
			reduce(281), // non_sigil_name, reduce: BracefulStatement
			reduce(281), // arr, reduce: BracefulStatement
			reduce(281), // bool, reduce: BracefulStatement
			reduce(281), // float, reduce: BracefulStatement
			reduce(281), // int, reduce: BracefulStatement
			reduce(281), // map, reduce: BracefulStatement
			reduce(281), // num, reduce: BracefulStatement
			reduce(281), // str, reduce: BracefulStatement
			reduce(281), // var, reduce: BracefulStatement
			reduce(281), // funct, reduce: BracefulStatement
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(281), // +, reduce: BracefulStatement
			reduce(281), // -, reduce: BracefulStatement
			reduce(281), // .+, reduce: BracefulStatement
			reduce(281), // .-, reduce: BracefulStatement
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(281), // !, reduce: BracefulStatement
			reduce(281), // ~, reduce: BracefulStatement
			nil,         // **
			reduce(281), // string_literal, reduce: BracefulStatement
			reduce(281), // regex_case_insensitive, reduce: BracefulStatement
			reduce(281), // int_literal, reduce: BracefulStatement
			reduce(281), // float_literal, reduce: BracefulStatement
			reduce(281), // boolean_literal, reduce: BracefulStatement
			reduce(281), // null_literal, reduce: BracefulStatement
			reduce(281), // inf_literal, reduce: BracefulStatement
			reduce(281), // nan_literal, reduce: BracefulStatement
			reduce(281), // const_M_PI, reduce: BracefulStatement
			reduce(281), // const_M_E, reduce: BracefulStatement
			reduce(281), // panic, reduce: BracefulStatement
			reduce(281), // [, reduce: BracefulStatement
			reduce(281), // ctx_IPS, reduce: BracefulStatement
			reduce(281), // ctx_IFS, reduce: BracefulStatement
			reduce(281), // ctx_IRS, reduce: BracefulStatement
			reduce(281), // ctx_OPS, reduce: BracefulStatement
			reduce(281), // ctx_OFS, reduce: BracefulStatement
			reduce(281), // ctx_ORS, reduce: BracefulStatement
			reduce(281), // ctx_FLATSEP, reduce: BracefulStatement
			reduce(281), // ctx_NF, reduce: BracefulStatement
			reduce(281), // ctx_NR, reduce: BracefulStatement
			reduce(281), // ctx_FNR, reduce: BracefulStatement
			reduce(281), // ctx_FILENAME, reduce: BracefulStatement
			reduce(281), // ctx_FILENUM, reduce: BracefulStatement
			reduce(281), // env, reduce: BracefulStatement
			nil,         // [[
			nil,         // [[[
			reduce(281), // call, reduce: BracefulStatement
			reduce(281), // begin, reduce: BracefulStatement
			reduce(281), // end, reduce: BracefulStatement
			reduce(281), // if, reduce: BracefulStatement
			nil,         // elif
			nil,         // else
			reduce(281), // while, reduce: BracefulStatement
			reduce(281), // do, reduce: BracefulStatement
			reduce(281), // for, reduce: BracefulStatement
			nil,         // in
			reduce(281), // break, reduce: BracefulStatement
			reduce(281), // continue, reduce: BracefulStatement
			reduce(281), // func, reduce: BracefulStatement
			reduce(281), // subr, reduce: BracefulStatement
			reduce(281), // return, reduce: BracefulStatement
		},
	},
	actionRow{ // S145
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(282), // $, reduce: BracefulStatement
			nil,         // empty
			reduce(282), // ;, reduce: BracefulStatement
			reduce(282), // {, reduce: BracefulStatement
			nil,         // }
			nil,         // =
			reduce(282), // unset, reduce: BracefulStatement
			reduce(282), // filter, reduce: BracefulStatement
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(282), // print, reduce: BracefulStatement
			nil,         // ,
			reduce(282), // printn, reduce: BracefulStatement
			reduce(282), // eprint, reduce: BracefulStatement
			reduce(282), // eprintn, reduce: BracefulStatement
			reduce(282), // dump, reduce: BracefulStatement
			reduce(282), // edump, reduce: BracefulStatement
			reduce(282), // tee, reduce: BracefulStatement
			reduce(282), // emitf, reduce: BracefulStatement
			reduce(282), // emit1, reduce: BracefulStatement
			reduce(282), // emit, reduce: BracefulStatement
			reduce(282), // (, reduce: BracefulStatement
			nil,         // )
			reduce(282), // emitp, reduce: BracefulStatement
			reduce(282), // field_name, reduce: BracefulStatement
			reduce(282), // $[, reduce: BracefulStatement
			nil,         // ]
			reduce(282), // braced_field_name, reduce: BracefulStatement
			reduce(282), // $[[, reduce: BracefulStatement
			reduce(282), // $[[[, reduce: BracefulStatement
			reduce(282), // full_srec, reduce: BracefulStatement
			reduce(282), // oosvar_name, reduce: BracefulStatement
			reduce(282), // @[, reduce: BracefulStatement
			reduce(282), // braced_oosvar_name, reduce: BracefulStatement
			reduce(282), // full_oosvar, reduce: BracefulStatement
			reduce(282), // all, reduce: BracefulStatement
			reduce(282), // non_sigil_name, reduce: BracefulStatement
			reduce(282), // arr, reduce: BracefulStatement
			reduce(282), // bool, reduce: BracefulStatement
			reduce(282), // float, reduce: BracefulStatement
			reduce(282), // int, reduce: BracefulStatement
			reduce(282), // map, reduce: BracefulStatement
			reduce(282), // num, reduce: BracefulStatement
			reduce(282), // str, reduce: BracefulStatement
			reduce(282), // var, reduce: BracefulStatement
			reduce(282), // funct, reduce: BracefulStatement
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(282), // +, reduce: BracefulStatement
			reduce(282), // -, reduce: BracefulStatement
			reduce(282), // .+, reduce: BracefulStatement
			reduce(282), // .-, reduce: BracefulStatement
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(282), // !, reduce: BracefulStatement
			reduce(282), // ~, reduce: BracefulStatement
			nil,         // **
			reduce(282), // string_literal, reduce: BracefulStatement
			reduce(282), // regex_case_insensitive, reduce: BracefulStatement
			reduce(282), // int_literal, reduce: BracefulStatement
			reduce(282), // float_literal, reduce: BracefulStatement
			reduce(282), // boolean_literal, reduce: BracefulStatement
			reduce(282), // null_literal, reduce: BracefulStatement
			reduce(282), // inf_literal, reduce: BracefulStatement
			reduce(282), // nan_literal, reduce: BracefulStatement
			reduce(282), // const_M_PI, reduce: BracefulStatement
			reduce(282), // const_M_E, reduce: BracefulStatement
			reduce(282), // panic, reduce: BracefulStatement
			reduce(282), // [, reduce: BracefulStatement
			reduce(282), // ctx_IPS, reduce: BracefulStatement
			reduce(282), // ctx_IFS, reduce: BracefulStatement
			reduce(282), // ctx_IRS, reduce: BracefulStatement
			reduce(282), // ctx_OPS, reduce: BracefulStatement
			reduce(282), // ctx_OFS, reduce: BracefulStatement
			reduce(282), // ctx_ORS, reduce: BracefulStatement
			reduce(282), // ctx_FLATSEP, reduce: BracefulStatement
			reduce(282), // ctx_NF, reduce: BracefulStatement
			reduce(282), // ctx_NR, reduce: BracefulStatement
			reduce(282), // ctx_FNR, reduce: BracefulStatement
			reduce(282), // ctx_FILENAME, reduce: BracefulStatement
			reduce(282), // ctx_FILENUM, reduce: BracefulStatement
			reduce(282), // env, reduce: BracefulStatement
			nil,         // [[
			nil,         // [[[
			reduce(282), // call, reduce: BracefulStatement
			reduce(282), // begin, reduce: BracefulStatement
			reduce(282), // end, reduce: BracefulStatement
			reduce(282), // if, reduce: BracefulStatement
			nil,         // elif
			nil,         // else
			reduce(282), // while, reduce: BracefulStatement
			reduce(282), // do, reduce: BracefulStatement
			reduce(282), // for, reduce: BracefulStatement
			nil,         // in
			reduce(282), // break, reduce: BracefulStatement
			reduce(282), // continue, reduce: BracefulStatement
			reduce(282), // func, reduce: BracefulStatement
			reduce(282), // subr, reduce: BracefulStatement
			reduce(282), // return, reduce: BracefulStatement
		},
	},
	actionRow{ // S146
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(283), // $, reduce: BracefulStatement
			nil,         // empty
			reduce(283), // ;, reduce: BracefulStatement
			reduce(283), // {, reduce: BracefulStatement
			nil,         // }
			nil,         // =
			reduce(283), // unset, reduce: BracefulStatement
			reduce(283), // filter, reduce: BracefulStatement
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(283), // print, reduce: BracefulStatement
			nil,         // ,
			reduce(283), // printn, reduce: BracefulStatement
			reduce(283), // eprint, reduce: BracefulStatement
			reduce(283), // eprintn, reduce: BracefulStatement
			reduce(283), // dump, reduce: BracefulStatement
			reduce(283), // edump, reduce: BracefulStatement
			reduce(283), // tee, reduce: BracefulStatement
			reduce(283), // emitf, reduce: BracefulStatement
			reduce(283), // emit1, reduce: BracefulStatement
			reduce(283), // emit, reduce: BracefulStatement
			reduce(283), // (, reduce: BracefulStatement
			nil,         // )
			reduce(283), // emitp, reduce: BracefulStatement
			reduce(283), // field_name, reduce: BracefulStatement
			reduce(283), // $[, reduce: BracefulStatement
			nil,         // ]
			reduce(283), // braced_field_name, reduce: BracefulStatement
			reduce(283), // $[[, reduce: BracefulStatement
			reduce(283), // $[[[, reduce: BracefulStatement
			reduce(283), // full_srec, reduce: BracefulStatement
			reduce(283), // oosvar_name, reduce: BracefulStatement
			reduce(283), // @[, reduce: BracefulStatement
			reduce(283), // braced_oosvar_name, reduce: BracefulStatement
			reduce(283), // full_oosvar, reduce: BracefulStatement
			reduce(283), // all, reduce: BracefulStatement
			reduce(283), // non_sigil_name, reduce: BracefulStatement
			reduce(283), // arr, reduce: BracefulStatement
			reduce(283), // bool, reduce: BracefulStatement
			reduce(283), // float, reduce: BracefulStatement
			reduce(283), // int, reduce: BracefulStatement
			reduce(283), // map, reduce: BracefulStatement
			reduce(283), // num, reduce: BracefulStatement
			reduce(283), // str, reduce: BracefulStatement
			reduce(283), // var, reduce: BracefulStatement
			reduce(283), // funct, reduce: BracefulStatement
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(283), // +, reduce: BracefulStatement
			reduce(283), // -, reduce: BracefulStatement
			reduce(283), // .+, reduce: BracefulStatement
			reduce(283), // .-, reduce: BracefulStatement
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(283), // !, reduce: BracefulStatement
			reduce(283), // ~, reduce: BracefulStatement
			nil,         // **
			reduce(283), // string_literal, reduce: BracefulStatement
			reduce(283), // regex_case_insensitive, reduce: BracefulStatement
			reduce(283), // int_literal, reduce: BracefulStatement
			reduce(283), // float_literal, reduce: BracefulStatement
			reduce(283), // boolean_literal, reduce: BracefulStatement
			reduce(283), // null_literal, reduce: BracefulStatement
			reduce(283), // inf_literal, reduce: BracefulStatement
			reduce(283), // nan_literal, reduce: BracefulStatement
			reduce(283), // const_M_PI, reduce: BracefulStatement
			reduce(283), // const_M_E, reduce: BracefulStatement
			reduce(283), // panic, reduce: BracefulStatement
			reduce(283), // [, reduce: BracefulStatement
			reduce(283), // ctx_IPS, reduce: BracefulStatement
			reduce(283), // ctx_IFS, reduce: BracefulStatement
			reduce(283), // ctx_IRS, reduce: BracefulStatement
			reduce(283), // ctx_OPS, reduce: BracefulStatement
			reduce(283), // ctx_OFS, reduce: BracefulStatement
			reduce(283), // ctx_ORS, reduce: BracefulStatement
			reduce(283), // ctx_FLATSEP, reduce: BracefulStatement
			reduce(283), // ctx_NF, reduce: BracefulStatement
			reduce(283), // ctx_NR, reduce: BracefulStatement
			reduce(283), // ctx_FNR, reduce: BracefulStatement
			reduce(283), // ctx_FILENAME, reduce: BracefulStatement
			reduce(283), // ctx_FILENUM, reduce: BracefulStatement
			reduce(283), // env, reduce: BracefulStatement
			nil,         // [[
			nil,         // [[[
			reduce(283), // call, reduce: BracefulStatement
			reduce(283), // begin, reduce: BracefulStatement
			reduce(283), // end, reduce: BracefulStatement
			reduce(283), // if, reduce: BracefulStatement
			nil,         // elif
			nil,         // else
			reduce(283), // while, reduce: BracefulStatement
			reduce(283), // do, reduce: BracefulStatement
			reduce(283), // for, reduce: BracefulStatement
			nil,         // in
			reduce(283), // break, reduce: BracefulStatement
			reduce(283), // continue, reduce: BracefulStatement
			reduce(283), // func, reduce: BracefulStatement
			reduce(283), // subr, reduce: BracefulStatement
			reduce(283), // return, reduce: BracefulStatement
		},
	},
	actionRow{ // S147
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(284), // $, reduce: BracefulStatement
			nil,         // empty
			reduce(284), // ;, reduce: BracefulStatement
			reduce(284), // {, reduce: BracefulStatement
			nil,         // }
			nil,         // =
			reduce(284), // unset, reduce: BracefulStatement
			reduce(284), // filter, reduce: BracefulStatement
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(284), // print, reduce: BracefulStatement
			nil,         // ,
			reduce(284), // printn, reduce: BracefulStatement
			reduce(284), // eprint, reduce: BracefulStatement
			reduce(284), // eprintn, reduce: BracefulStatement
			reduce(284), // dump, reduce: BracefulStatement
			reduce(284), // edump, reduce: BracefulStatement
			reduce(284), // tee, reduce: BracefulStatement
			reduce(284), // emitf, reduce: BracefulStatement
			reduce(284), // emit1, reduce: BracefulStatement
			reduce(284), // emit, reduce: BracefulStatement
			reduce(284), // (, reduce: BracefulStatement
			nil,         // )
			reduce(284), // emitp, reduce: BracefulStatement
			reduce(284), // field_name, reduce: BracefulStatement
			reduce(284), // $[, reduce: BracefulStatement
			nil,         // ]
			reduce(284), // braced_field_name, reduce: BracefulStatement
			reduce(284), // $[[, reduce: BracefulStatement
			reduce(284), // $[[[, reduce: BracefulStatement
			reduce(284), // full_srec, reduce: BracefulStatement
			reduce(284), // oosvar_name, reduce: BracefulStatement
			reduce(284), // @[, reduce: BracefulStatement
			reduce(284), // braced_oosvar_name, reduce: BracefulStatement
			reduce(284), // full_oosvar, reduce: BracefulStatement
			reduce(284), // all, reduce: BracefulStatement
			reduce(284), // non_sigil_name, reduce: BracefulStatement
			reduce(284), // arr, reduce: BracefulStatement
			reduce(284), // bool, reduce: BracefulStatement
			reduce(284), // float, reduce: BracefulStatement
			reduce(284), // int, reduce: BracefulStatement
			reduce(284), // map, reduce: BracefulStatement
			reduce(284), // num, reduce: BracefulStatement
			reduce(284), // str, reduce: BracefulStatement
			reduce(284), // var, reduce: BracefulStatement
			reduce(284), // funct, reduce: BracefulStatement
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(284), // +, reduce: BracefulStatement
			reduce(284), // -, reduce: BracefulStatement
			reduce(284), // .+, reduce: BracefulStatement
			reduce(284), // .-, reduce: BracefulStatement
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(284), // !, reduce: BracefulStatement
			reduce(284), // ~, reduce: BracefulStatement
			nil,         // **
			reduce(284), // string_literal, reduce: BracefulStatement
			reduce(284), // regex_case_insensitive, reduce: BracefulStatement
			reduce(284), // int_literal, reduce: BracefulStatement
			reduce(284), // float_literal, reduce: BracefulStatement
			reduce(284), // boolean_literal, reduce: BracefulStatement
			reduce(284), // null_literal, reduce: BracefulStatement
			reduce(284), // inf_literal, reduce: BracefulStatement
			reduce(284), // nan_literal, reduce: BracefulStatement
			reduce(284), // const_M_PI, reduce: BracefulStatement
			reduce(284), // const_M_E, reduce: BracefulStatement
			reduce(284), // panic, reduce: BracefulStatement
			reduce(284), // [, reduce: BracefulStatement
			reduce(284), // ctx_IPS, reduce: BracefulStatement
			reduce(284), // ctx_IFS, reduce: BracefulStatement
			reduce(284), // ctx_IRS, reduce: BracefulStatement
			reduce(284), // ctx_OPS, reduce: BracefulStatement
			reduce(284), // ctx_OFS, reduce: BracefulStatement
			reduce(284), // ctx_ORS, reduce: BracefulStatement
			reduce(284), // ctx_FLATSEP, reduce: BracefulStatement
			reduce(284), // ctx_NF, reduce: BracefulStatement
			reduce(284), // ctx_NR, reduce: BracefulStatement
			reduce(284), // ctx_FNR, reduce: BracefulStatement
			reduce(284), // ctx_FILENAME, reduce: BracefulStatement
			reduce(284), // ctx_FILENUM, reduce: BracefulStatement
			reduce(284), // env, reduce: BracefulStatement
			nil,         // [[
			nil,         // [[[
			reduce(284), // call, reduce: BracefulStatement
			reduce(284), // begin, reduce: BracefulStatement
			reduce(284), // end, reduce: BracefulStatement
			reduce(284), // if, reduce: BracefulStatement
			nil,         // elif
			nil,         // else
			reduce(284), // while, reduce: BracefulStatement
			reduce(284), // do, reduce: BracefulStatement
			reduce(284), // for, reduce: BracefulStatement
			nil,         // in
			reduce(284), // break, reduce: BracefulStatement
			reduce(284), // continue, reduce: BracefulStatement
			reduce(284), // func, reduce: BracefulStatement
			reduce(284), // subr, reduce: BracefulStatement
			reduce(284), // return, reduce: BracefulStatement
		},
	},
	actionRow{ // S148
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(285), // $, reduce: BracefulStatement
			nil,         // empty
			reduce(285), // ;, reduce: BracefulStatement
			reduce(285), // {, reduce: BracefulStatement
			nil,         // }
			nil,         // =
			reduce(285), // unset, reduce: BracefulStatement
			reduce(285), // filter, reduce: BracefulStatement
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(285), // print, reduce: BracefulStatement
			nil,         // ,
			reduce(285), // printn, reduce: BracefulStatement
			reduce(285), // eprint, reduce: BracefulStatement
			reduce(285), // eprintn, reduce: BracefulStatement
			reduce(285), // dump, reduce: BracefulStatement
			reduce(285), // edump, reduce: BracefulStatement
			reduce(285), // tee, reduce: BracefulStatement
			reduce(285), // emitf, reduce: BracefulStatement
			reduce(285), // emit1, reduce: BracefulStatement
			reduce(285), // emit, reduce: BracefulStatement
			reduce(285), // (, reduce: BracefulStatement
			nil,         // )
			reduce(285), // emitp, reduce: BracefulStatement
			reduce(285), // field_name, reduce: BracefulStatement
			reduce(285), // $[, reduce: BracefulStatement
			nil,         // ]
			reduce(285), // braced_field_name, reduce: BracefulStatement
			reduce(285), // $[[, reduce: BracefulStatement
			reduce(285), // $[[[, reduce: BracefulStatement
			reduce(285), // full_srec, reduce: BracefulStatement
			reduce(285), // oosvar_name, reduce: BracefulStatement
			reduce(285), // @[, reduce: BracefulStatement
			reduce(285), // braced_oosvar_name, reduce: BracefulStatement
			reduce(285), // full_oosvar, reduce: BracefulStatement
			reduce(285), // all, reduce: BracefulStatement
			reduce(285), // non_sigil_name, reduce: BracefulStatement
			reduce(285), // arr, reduce: BracefulStatement
			reduce(285), // bool, reduce: BracefulStatement
			reduce(285), // float, reduce: BracefulStatement
			reduce(285), // int, reduce: BracefulStatement
			reduce(285), // map, reduce: BracefulStatement
			reduce(285), // num, reduce: BracefulStatement
			reduce(285), // str, reduce: BracefulStatement
			reduce(285), // var, reduce: BracefulStatement
			reduce(285), // funct, reduce: BracefulStatement
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(285), // +, reduce: BracefulStatement
			reduce(285), // -, reduce: BracefulStatement
			reduce(285), // .+, reduce: BracefulStatement
			reduce(285), // .-, reduce: BracefulStatement
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(285), // !, reduce: BracefulStatement
			reduce(285), // ~, reduce: BracefulStatement
			nil,         // **
			reduce(285), // string_literal, reduce: BracefulStatement
			reduce(285), // regex_case_insensitive, reduce: BracefulStatement
			reduce(285), // int_literal, reduce: BracefulStatement
			reduce(285), // float_literal, reduce: BracefulStatement
			reduce(285), // boolean_literal, reduce: BracefulStatement
			reduce(285), // null_literal, reduce: BracefulStatement
			reduce(285), // inf_literal, reduce: BracefulStatement
			reduce(285), // nan_literal, reduce: BracefulStatement
			reduce(285), // const_M_PI, reduce: BracefulStatement
			reduce(285), // const_M_E, reduce: BracefulStatement
			reduce(285), // panic, reduce: BracefulStatement
			reduce(285), // [, reduce: BracefulStatement
			reduce(285), // ctx_IPS, reduce: BracefulStatement
			reduce(285), // ctx_IFS, reduce: BracefulStatement
			reduce(285), // ctx_IRS, reduce: BracefulStatement
			reduce(285), // ctx_OPS, reduce: BracefulStatement
			reduce(285), // ctx_OFS, reduce: BracefulStatement
			reduce(285), // ctx_ORS, reduce: BracefulStatement
			reduce(285), // ctx_FLATSEP, reduce: BracefulStatement
			reduce(285), // ctx_NF, reduce: BracefulStatement
			reduce(285), // ctx_NR, reduce: BracefulStatement
			reduce(285), // ctx_FNR, reduce: BracefulStatement
			reduce(285), // ctx_FILENAME, reduce: BracefulStatement
			reduce(285), // ctx_FILENUM, reduce: BracefulStatement
			reduce(285), // env, reduce: BracefulStatement
			nil,         // [[
			nil,         // [[[
			reduce(285), // call, reduce: BracefulStatement
			reduce(285), // begin, reduce: BracefulStatement
			reduce(285), // end, reduce: BracefulStatement
			reduce(285), // if, reduce: BracefulStatement
			nil,         // elif
			nil,         // else
			reduce(285), // while, reduce: BracefulStatement
			reduce(285), // do, reduce: BracefulStatement
			reduce(285), // for, reduce: BracefulStatement
			nil,         // in
			reduce(285), // break, reduce: BracefulStatement
			reduce(285), // continue, reduce: BracefulStatement
			reduce(285), // func, reduce: BracefulStatement
			reduce(285), // subr, reduce: BracefulStatement
			reduce(285), // return, reduce: BracefulStatement
		},
	},
	actionRow{ // S149
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(286), // $, reduce: BracefulStatement
			nil,         // empty
			reduce(286), // ;, reduce: BracefulStatement
			reduce(286), // {, reduce: BracefulStatement
			nil,         // }
			nil,         // =
			reduce(286), // unset, reduce: BracefulStatement
			reduce(286), // filter, reduce: BracefulStatement
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(286), // print, reduce: BracefulStatement
			nil,         // ,
			reduce(286), // printn, reduce: BracefulStatement
			reduce(286), // eprint, reduce: BracefulStatement
			reduce(286), // eprintn, reduce: BracefulStatement
			reduce(286), // dump, reduce: BracefulStatement
			reduce(286), // edump, reduce: BracefulStatement
			reduce(286), // tee, reduce: BracefulStatement
			reduce(286), // emitf, reduce: BracefulStatement
			reduce(286), // emit1, reduce: BracefulStatement
			reduce(286), // emit, reduce: BracefulStatement
			reduce(286), // (, reduce: BracefulStatement
			nil,         // )
			reduce(286), // emitp, reduce: BracefulStatement
			reduce(286), // field_name, reduce: BracefulStatement
			reduce(286), // $[, reduce: BracefulStatement
			nil,         // ]
			reduce(286), // braced_field_name, reduce: BracefulStatement
			reduce(286), // $[[, reduce: BracefulStatement
			reduce(286), // $[[[, reduce: BracefulStatement
			reduce(286), // full_srec, reduce: BracefulStatement
			reduce(286), // oosvar_name, reduce: BracefulStatement
			reduce(286), // @[, reduce: BracefulStatement
			reduce(286), // braced_oosvar_name, reduce: BracefulStatement
			reduce(286), // full_oosvar, reduce: BracefulStatement
			reduce(286), // all, reduce: BracefulStatement
			reduce(286), // non_sigil_name, reduce: BracefulStatement
			reduce(286), // arr, reduce: BracefulStatement
			reduce(286), // bool, reduce: BracefulStatement
			reduce(286), // float, reduce: BracefulStatement
			reduce(286), // int, reduce: BracefulStatement
			reduce(286), // map, reduce: BracefulStatement
			reduce(286), // num, reduce: BracefulStatement
			reduce(286), // str, reduce: BracefulStatement
			reduce(286), // var, reduce: BracefulStatement
			reduce(286), // funct, reduce: BracefulStatement
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(286), // +, reduce: BracefulStatement
			reduce(286), // -, reduce: BracefulStatement
			reduce(286), // .+, reduce: BracefulStatement
			reduce(286), // .-, reduce: BracefulStatement
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(286), // !, reduce: BracefulStatement
			reduce(286), // ~, reduce: BracefulStatement
			nil,         // **
			reduce(286), // string_literal, reduce: BracefulStatement
			reduce(286), // regex_case_insensitive, reduce: BracefulStatement
			reduce(286), // int_literal, reduce: BracefulStatement
			reduce(286), // float_literal, reduce: BracefulStatement
			reduce(286), // boolean_literal, reduce: BracefulStatement
			reduce(286), // null_literal, reduce: BracefulStatement
			reduce(286), // inf_literal, reduce: BracefulStatement
			reduce(286), // nan_literal, reduce: BracefulStatement
			reduce(286), // const_M_PI, reduce: BracefulStatement
			reduce(286), // const_M_E, reduce: BracefulStatement
			reduce(286), // panic, reduce: BracefulStatement
			reduce(286), // [, reduce: BracefulStatement
			reduce(286), // ctx_IPS, reduce: BracefulStatement
			reduce(286), // ctx_IFS, reduce: BracefulStatement
			reduce(286), // ctx_IRS, reduce: BracefulStatement
			reduce(286), // ctx_OPS, reduce: BracefulStatement
			reduce(286), // ctx_OFS, reduce: BracefulStatement
			reduce(286), // ctx_ORS, reduce: BracefulStatement
			reduce(286), // ctx_FLATSEP, reduce: BracefulStatement
			reduce(286), // ctx_NF, reduce: BracefulStatement
			reduce(286), // ctx_NR, reduce: BracefulStatement
			reduce(286), // ctx_FNR, reduce: BracefulStatement
			reduce(286), // ctx_FILENAME, reduce: BracefulStatement
			reduce(286), // ctx_FILENUM, reduce: BracefulStatement
			reduce(286), // env, reduce: BracefulStatement
			nil,         // [[
			nil,         // [[[
			reduce(286), // call, reduce: BracefulStatement
			reduce(286), // begin, reduce: BracefulStatement
			reduce(286), // end, reduce: BracefulStatement
			reduce(286), // if, reduce: BracefulStatement
			nil,         // elif
			nil,         // else
			reduce(286), // while, reduce: BracefulStatement
			reduce(286), // do, reduce: BracefulStatement
			reduce(286), // for, reduce: BracefulStatement
			nil,         // in
			reduce(286), // break, reduce: BracefulStatement
			reduce(286), // continue, reduce: BracefulStatement
			reduce(286), // func, reduce: BracefulStatement
			reduce(286), // subr, reduce: BracefulStatement
			reduce(286), // return, reduce: BracefulStatement
		},
	},
	actionRow{ // S150
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(289), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S151
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(289), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S152
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(290), // $, reduce: IfChain
			nil,         // empty
			reduce(290), // ;, reduce: IfChain
			reduce(290), // {, reduce: IfChain
			nil,         // }
			nil,         // =
			reduce(290), // unset, reduce: IfChain
			reduce(290), // filter, reduce: IfChain
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(290), // print, reduce: IfChain
			nil,         // ,
			reduce(290), // printn, reduce: IfChain
			reduce(290), // eprint, reduce: IfChain
			reduce(290), // eprintn, reduce: IfChain
			reduce(290), // dump, reduce: IfChain
			reduce(290), // edump, reduce: IfChain
			reduce(290), // tee, reduce: IfChain
			reduce(290), // emitf, reduce: IfChain
			reduce(290), // emit1, reduce: IfChain
			reduce(290), // emit, reduce: IfChain
			reduce(290), // (, reduce: IfChain
			nil,         // )
			reduce(290), // emitp, reduce: IfChain
			reduce(290), // field_name, reduce: IfChain
			reduce(290), // $[, reduce: IfChain
			nil,         // ]
			reduce(290), // braced_field_name, reduce: IfChain
			reduce(290), // $[[, reduce: IfChain
			reduce(290), // $[[[, reduce: IfChain
			reduce(290), // full_srec, reduce: IfChain
			reduce(290), // oosvar_name, reduce: IfChain
			reduce(290), // @[, reduce: IfChain
			reduce(290), // braced_oosvar_name, reduce: IfChain
			reduce(290), // full_oosvar, reduce: IfChain
			reduce(290), // all, reduce: IfChain
			reduce(290), // non_sigil_name, reduce: IfChain
			reduce(290), // arr, reduce: IfChain
			reduce(290), // bool, reduce: IfChain
			reduce(290), // float, reduce: IfChain
			reduce(290), // int, reduce: IfChain
			reduce(290), // map, reduce: IfChain
			reduce(290), // num, reduce: IfChain
			reduce(290), // str, reduce: IfChain
			reduce(290), // var, reduce: IfChain
			reduce(290), // funct, reduce: IfChain
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(290), // +, reduce: IfChain
			reduce(290), // -, reduce: IfChain
			reduce(290), // .+, reduce: IfChain
			reduce(290), // .-, reduce: IfChain
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(290), // !, reduce: IfChain
			reduce(290), // ~, reduce: IfChain
			nil,         // **
			reduce(290), // string_literal, reduce: IfChain
			reduce(290), // regex_case_insensitive, reduce: IfChain
			reduce(290), // int_literal, reduce: IfChain
			reduce(290), // float_literal, reduce: IfChain
			reduce(290), // boolean_literal, reduce: IfChain
			reduce(290), // null_literal, reduce: IfChain
			reduce(290), // inf_literal, reduce: IfChain
			reduce(290), // nan_literal, reduce: IfChain
			reduce(290), // const_M_PI, reduce: IfChain
			reduce(290), // const_M_E, reduce: IfChain
			reduce(290), // panic, reduce: IfChain
			reduce(290), // [, reduce: IfChain
			reduce(290), // ctx_IPS, reduce: IfChain
			reduce(290), // ctx_IFS, reduce: IfChain
			reduce(290), // ctx_IRS, reduce: IfChain
			reduce(290), // ctx_OPS, reduce: IfChain
			reduce(290), // ctx_OFS, reduce: IfChain
			reduce(290), // ctx_ORS, reduce: IfChain
			reduce(290), // ctx_FLATSEP, reduce: IfChain
			reduce(290), // ctx_NF, reduce: IfChain
			reduce(290), // ctx_NR, reduce: IfChain
			reduce(290), // ctx_FNR, reduce: IfChain
			reduce(290), // ctx_FILENAME, reduce: IfChain
			reduce(290), // ctx_FILENUM, reduce: IfChain
			reduce(290), // env, reduce: IfChain
			nil,         // [[
			nil,         // [[[
			reduce(290), // call, reduce: IfChain
			reduce(290), // begin, reduce: IfChain
			reduce(290), // end, reduce: IfChain
			reduce(290), // if, reduce: IfChain
			shift(854),  // elif
			shift(855),  // else
			reduce(290), // while, reduce: IfChain
			reduce(290), // do, reduce: IfChain
			reduce(290), // for, reduce: IfChain
			nil,         // in
			reduce(290), // break, reduce: IfChain
			reduce(290), // continue, reduce: IfChain
			reduce(290), // func, reduce: IfChain
			reduce(290), // subr, reduce: IfChain
			reduce(290), // return, reduce: IfChain
		},
	},
	actionRow{ // S153
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(292), // $, reduce: IfElifStar
			nil,         // empty
			reduce(292), // ;, reduce: IfElifStar
			reduce(292), // {, reduce: IfElifStar
			nil,         // }
			nil,         // =
			reduce(292), // unset, reduce: IfElifStar
			reduce(292), // filter, reduce: IfElifStar
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(292), // print, reduce: IfElifStar
			nil,         // ,
			reduce(292), // printn, reduce: IfElifStar
			reduce(292), // eprint, reduce: IfElifStar
			reduce(292), // eprintn, reduce: IfElifStar
			reduce(292), // dump, reduce: IfElifStar
			reduce(292), // edump, reduce: IfElifStar
			reduce(292), // tee, reduce: IfElifStar
			reduce(292), // emitf, reduce: IfElifStar
			reduce(292), // emit1, reduce: IfElifStar
			reduce(292), // emit, reduce: IfElifStar
			reduce(292), // (, reduce: IfElifStar
			nil,         // )
			reduce(292), // emitp, reduce: IfElifStar
			reduce(292), // field_name, reduce: IfElifStar
			reduce(292), // $[, reduce: IfElifStar
			nil,         // ]
			reduce(292), // braced_field_name, reduce: IfElifStar
			reduce(292), // $[[, reduce: IfElifStar
			reduce(292), // $[[[, reduce: IfElifStar
			reduce(292), // full_srec, reduce: IfElifStar
			reduce(292), // oosvar_name, reduce: IfElifStar
			reduce(292), // @[, reduce: IfElifStar
			reduce(292), // braced_oosvar_name, reduce: IfElifStar
			reduce(292), // full_oosvar, reduce: IfElifStar
			reduce(292), // all, reduce: IfElifStar
			reduce(292), // non_sigil_name, reduce: IfElifStar
			reduce(292), // arr, reduce: IfElifStar
			reduce(292), // bool, reduce: IfElifStar
			reduce(292), // float, reduce: IfElifStar
			reduce(292), // int, reduce: IfElifStar
			reduce(292), // map, reduce: IfElifStar
			reduce(292), // num, reduce: IfElifStar
			reduce(292), // str, reduce: IfElifStar
			reduce(292), // var, reduce: IfElifStar
			reduce(292), // funct, reduce: IfElifStar
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(292), // +, reduce: IfElifStar
			reduce(292), // -, reduce: IfElifStar
			reduce(292), // .+, reduce: IfElifStar
			reduce(292), // .-, reduce: IfElifStar
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(292), // !, reduce: IfElifStar
			reduce(292), // ~, reduce: IfElifStar
			nil,         // **
			reduce(292), // string_literal, reduce: IfElifStar
			reduce(292), // regex_case_insensitive, reduce: IfElifStar
			reduce(292), // int_literal, reduce: IfElifStar
			reduce(292), // float_literal, reduce: IfElifStar
			reduce(292), // boolean_literal, reduce: IfElifStar
			reduce(292), // null_literal, reduce: IfElifStar
			reduce(292), // inf_literal, reduce: IfElifStar
			reduce(292), // nan_literal, reduce: IfElifStar
			reduce(292), // const_M_PI, reduce: IfElifStar
			reduce(292), // const_M_E, reduce: IfElifStar
			reduce(292), // panic, reduce: IfElifStar
			reduce(292), // [, reduce: IfElifStar
			reduce(292), // ctx_IPS, reduce: IfElifStar
			reduce(292), // ctx_IFS, reduce: IfElifStar
			reduce(292), // ctx_IRS, reduce: IfElifStar
			reduce(292), // ctx_OPS, reduce: IfElifStar
			reduce(292), // ctx_OFS, reduce: IfElifStar
			reduce(292), // ctx_ORS, reduce: IfElifStar
			reduce(292), // ctx_FLATSEP, reduce: IfElifStar
			reduce(292), // ctx_NF, reduce: IfElifStar
			reduce(292), // ctx_NR, reduce: IfElifStar
			reduce(292), // ctx_FNR, reduce: IfElifStar
			reduce(292), // ctx_FILENAME, reduce: IfElifStar
			reduce(292), // ctx_FILENUM, reduce: IfElifStar
			reduce(292), // env, reduce: IfElifStar
			nil,         // [[
			nil,         // [[[
			reduce(292), // call, reduce: IfElifStar
			reduce(292), // begin, reduce: IfElifStar
			reduce(292), // end, reduce: IfElifStar
			reduce(292), // if, reduce: IfElifStar
			reduce(292), // elif, reduce: IfElifStar
			reduce(292), // else, reduce: IfElifStar
			reduce(292), // while, reduce: IfElifStar
			reduce(292), // do, reduce: IfElifStar
			reduce(292), // for, reduce: IfElifStar
			nil,         // in
			reduce(292), // break, reduce: IfElifStar
			reduce(292), // continue, reduce: IfElifStar
			reduce(292), // func, reduce: IfElifStar
			reduce(292), // subr, reduce: IfElifStar
			reduce(292), // return, reduce: IfElifStar
		},
	},
	actionRow{ // S154
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(856), // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S155
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(857), // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S156
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(859), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S157
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(299), // $, reduce: ForLoop
			nil,         // empty
			reduce(299), // ;, reduce: ForLoop
			reduce(299), // {, reduce: ForLoop
			nil,         // }
			nil,         // =
			reduce(299), // unset, reduce: ForLoop
			reduce(299), // filter, reduce: ForLoop
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(299), // print, reduce: ForLoop
			nil,         // ,
			reduce(299), // printn, reduce: ForLoop
			reduce(299), // eprint, reduce: ForLoop
			reduce(299), // eprintn, reduce: ForLoop
			reduce(299), // dump, reduce: ForLoop
			reduce(299), // edump, reduce: ForLoop
			reduce(299), // tee, reduce: ForLoop
			reduce(299), // emitf, reduce: ForLoop
			reduce(299), // emit1, reduce: ForLoop
			reduce(299), // emit, reduce: ForLoop
			reduce(299), // (, reduce: ForLoop
			nil,         // )
			reduce(299), // emitp, reduce: ForLoop
			reduce(299), // field_name, reduce: ForLoop
			reduce(299), // $[, reduce: ForLoop
			nil,         // ]
			reduce(299), // braced_field_name, reduce: ForLoop
			reduce(299), // $[[, reduce: ForLoop
			reduce(299), // $[[[, reduce: ForLoop
			reduce(299), // full_srec, reduce: ForLoop
			reduce(299), // oosvar_name, reduce: ForLoop
			reduce(299), // @[, reduce: ForLoop
			reduce(299), // braced_oosvar_name, reduce: ForLoop
			reduce(299), // full_oosvar, reduce: ForLoop
			reduce(299), // all, reduce: ForLoop
			reduce(299), // non_sigil_name, reduce: ForLoop
			reduce(299), // arr, reduce: ForLoop
			reduce(299), // bool, reduce: ForLoop
			reduce(299), // float, reduce: ForLoop
			reduce(299), // int, reduce: ForLoop
			reduce(299), // map, reduce: ForLoop
			reduce(299), // num, reduce: ForLoop
			reduce(299), // str, reduce: ForLoop
			reduce(299), // var, reduce: ForLoop
			reduce(299), // funct, reduce: ForLoop
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(299), // +, reduce: ForLoop
			reduce(299), // -, reduce: ForLoop
			reduce(299), // .+, reduce: ForLoop
			reduce(299), // .-, reduce: ForLoop
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(299), // !, reduce: ForLoop
			reduce(299), // ~, reduce: ForLoop
			nil,         // **
			reduce(299), // string_literal, reduce: ForLoop
			reduce(299), // regex_case_insensitive, reduce: ForLoop
			reduce(299), // int_literal, reduce: ForLoop
			reduce(299), // float_literal, reduce: ForLoop
			reduce(299), // boolean_literal, reduce: ForLoop
			reduce(299), // null_literal, reduce: ForLoop
			reduce(299), // inf_literal, reduce: ForLoop
			reduce(299), // nan_literal, reduce: ForLoop
			reduce(299), // const_M_PI, reduce: ForLoop
			reduce(299), // const_M_E, reduce: ForLoop
			reduce(299), // panic, reduce: ForLoop
			reduce(299), // [, reduce: ForLoop
			reduce(299), // ctx_IPS, reduce: ForLoop
			reduce(299), // ctx_IFS, reduce: ForLoop
			reduce(299), // ctx_IRS, reduce: ForLoop
			reduce(299), // ctx_OPS, reduce: ForLoop
			reduce(299), // ctx_OFS, reduce: ForLoop
			reduce(299), // ctx_ORS, reduce: ForLoop
			reduce(299), // ctx_FLATSEP, reduce: ForLoop
			reduce(299), // ctx_NF, reduce: ForLoop
			reduce(299), // ctx_NR, reduce: ForLoop
			reduce(299), // ctx_FNR, reduce: ForLoop
			reduce(299), // ctx_FILENAME, reduce: ForLoop
			reduce(299), // ctx_FILENUM, reduce: ForLoop
			reduce(299), // env, reduce: ForLoop
			nil,         // [[
			nil,         // [[[
			reduce(299), // call, reduce: ForLoop
			reduce(299), // begin, reduce: ForLoop
			reduce(299), // end, reduce: ForLoop
			reduce(299), // if, reduce: ForLoop
			nil,         // elif
			nil,         // else
			reduce(299), // while, reduce: ForLoop
			reduce(299), // do, reduce: ForLoop
			reduce(299), // for, reduce: ForLoop
			nil,         // in
			reduce(299), // break, reduce: ForLoop
			reduce(299), // continue, reduce: ForLoop
			reduce(299), // func, reduce: ForLoop
			reduce(299), // subr, reduce: ForLoop
			reduce(299), // return, reduce: ForLoop
		},
	},
	actionRow{ // S158
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(300), // $, reduce: ForLoop
			nil,         // empty
			reduce(300), // ;, reduce: ForLoop
			reduce(300), // {, reduce: ForLoop
			nil,         // }
			nil,         // =
			reduce(300), // unset, reduce: ForLoop
			reduce(300), // filter, reduce: ForLoop
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(300), // print, reduce: ForLoop
			nil,         // ,
			reduce(300), // printn, reduce: ForLoop
			reduce(300), // eprint, reduce: ForLoop
			reduce(300), // eprintn, reduce: ForLoop
			reduce(300), // dump, reduce: ForLoop
			reduce(300), // edump, reduce: ForLoop
			reduce(300), // tee, reduce: ForLoop
			reduce(300), // emitf, reduce: ForLoop
			reduce(300), // emit1, reduce: ForLoop
			reduce(300), // emit, reduce: ForLoop
			reduce(300), // (, reduce: ForLoop
			nil,         // )
			reduce(300), // emitp, reduce: ForLoop
			reduce(300), // field_name, reduce: ForLoop
			reduce(300), // $[, reduce: ForLoop
			nil,         // ]
			reduce(300), // braced_field_name, reduce: ForLoop
			reduce(300), // $[[, reduce: ForLoop
			reduce(300), // $[[[, reduce: ForLoop
			reduce(300), // full_srec, reduce: ForLoop
			reduce(300), // oosvar_name, reduce: ForLoop
			reduce(300), // @[, reduce: ForLoop
			reduce(300), // braced_oosvar_name, reduce: ForLoop
			reduce(300), // full_oosvar, reduce: ForLoop
			reduce(300), // all, reduce: ForLoop
			reduce(300), // non_sigil_name, reduce: ForLoop
			reduce(300), // arr, reduce: ForLoop
			reduce(300), // bool, reduce: ForLoop
			reduce(300), // float, reduce: ForLoop
			reduce(300), // int, reduce: ForLoop
			reduce(300), // map, reduce: ForLoop
			reduce(300), // num, reduce: ForLoop
			reduce(300), // str, reduce: ForLoop
			reduce(300), // var, reduce: ForLoop
			reduce(300), // funct, reduce: ForLoop
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(300), // +, reduce: ForLoop
			reduce(300), // -, reduce: ForLoop
			reduce(300), // .+, reduce: ForLoop
			reduce(300), // .-, reduce: ForLoop
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(300), // !, reduce: ForLoop
			reduce(300), // ~, reduce: ForLoop
			nil,         // **
			reduce(300), // string_literal, reduce: ForLoop
			reduce(300), // regex_case_insensitive, reduce: ForLoop
			reduce(300), // int_literal, reduce: ForLoop
			reduce(300), // float_literal, reduce: ForLoop
			reduce(300), // boolean_literal, reduce: ForLoop
			reduce(300), // null_literal, reduce: ForLoop
			reduce(300), // inf_literal, reduce: ForLoop
			reduce(300), // nan_literal, reduce: ForLoop
			reduce(300), // const_M_PI, reduce: ForLoop
			reduce(300), // const_M_E, reduce: ForLoop
			reduce(300), // panic, reduce: ForLoop
			reduce(300), // [, reduce: ForLoop
			reduce(300), // ctx_IPS, reduce: ForLoop
			reduce(300), // ctx_IFS, reduce: ForLoop
			reduce(300), // ctx_IRS, reduce: ForLoop
			reduce(300), // ctx_OPS, reduce: ForLoop
			reduce(300), // ctx_OFS, reduce: ForLoop
			reduce(300), // ctx_ORS, reduce: ForLoop
			reduce(300), // ctx_FLATSEP, reduce: ForLoop
			reduce(300), // ctx_NF, reduce: ForLoop
			reduce(300), // ctx_NR, reduce: ForLoop
			reduce(300), // ctx_FNR, reduce: ForLoop
			reduce(300), // ctx_FILENAME, reduce: ForLoop
			reduce(300), // ctx_FILENUM, reduce: ForLoop
			reduce(300), // env, reduce: ForLoop
			nil,         // [[
			nil,         // [[[
			reduce(300), // call, reduce: ForLoop
			reduce(300), // begin, reduce: ForLoop
			reduce(300), // end, reduce: ForLoop
			reduce(300), // if, reduce: ForLoop
			nil,         // elif
			nil,         // else
			reduce(300), // while, reduce: ForLoop
			reduce(300), // do, reduce: ForLoop
			reduce(300), // for, reduce: ForLoop
			nil,         // in
			reduce(300), // break, reduce: ForLoop
			reduce(300), // continue, reduce: ForLoop
			reduce(300), // func, reduce: ForLoop
			reduce(300), // subr, reduce: ForLoop
			reduce(300), // return, reduce: ForLoop
		},
	},
	actionRow{ // S159
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(301), // $, reduce: ForLoop
			nil,         // empty
			reduce(301), // ;, reduce: ForLoop
			reduce(301), // {, reduce: ForLoop
			nil,         // }
			nil,         // =
			reduce(301), // unset, reduce: ForLoop
			reduce(301), // filter, reduce: ForLoop
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(301), // print, reduce: ForLoop
			nil,         // ,
			reduce(301), // printn, reduce: ForLoop
			reduce(301), // eprint, reduce: ForLoop
			reduce(301), // eprintn, reduce: ForLoop
			reduce(301), // dump, reduce: ForLoop
			reduce(301), // edump, reduce: ForLoop
			reduce(301), // tee, reduce: ForLoop
			reduce(301), // emitf, reduce: ForLoop
			reduce(301), // emit1, reduce: ForLoop
			reduce(301), // emit, reduce: ForLoop
			reduce(301), // (, reduce: ForLoop
			nil,         // )
			reduce(301), // emitp, reduce: ForLoop
			reduce(301), // field_name, reduce: ForLoop
			reduce(301), // $[, reduce: ForLoop
			nil,         // ]
			reduce(301), // braced_field_name, reduce: ForLoop
			reduce(301), // $[[, reduce: ForLoop
			reduce(301), // $[[[, reduce: ForLoop
			reduce(301), // full_srec, reduce: ForLoop
			reduce(301), // oosvar_name, reduce: ForLoop
			reduce(301), // @[, reduce: ForLoop
			reduce(301), // braced_oosvar_name, reduce: ForLoop
			reduce(301), // full_oosvar, reduce: ForLoop
			reduce(301), // all, reduce: ForLoop
			reduce(301), // non_sigil_name, reduce: ForLoop
			reduce(301), // arr, reduce: ForLoop
			reduce(301), // bool, reduce: ForLoop
			reduce(301), // float, reduce: ForLoop
			reduce(301), // int, reduce: ForLoop
			reduce(301), // map, reduce: ForLoop
			reduce(301), // num, reduce: ForLoop
			reduce(301), // str, reduce: ForLoop
			reduce(301), // var, reduce: ForLoop
			reduce(301), // funct, reduce: ForLoop
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(301), // +, reduce: ForLoop
			reduce(301), // -, reduce: ForLoop
			reduce(301), // .+, reduce: ForLoop
			reduce(301), // .-, reduce: ForLoop
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(301), // !, reduce: ForLoop
			reduce(301), // ~, reduce: ForLoop
			nil,         // **
			reduce(301), // string_literal, reduce: ForLoop
			reduce(301), // regex_case_insensitive, reduce: ForLoop
			reduce(301), // int_literal, reduce: ForLoop
			reduce(301), // float_literal, reduce: ForLoop
			reduce(301), // boolean_literal, reduce: ForLoop
			reduce(301), // null_literal, reduce: ForLoop
			reduce(301), // inf_literal, reduce: ForLoop
			reduce(301), // nan_literal, reduce: ForLoop
			reduce(301), // const_M_PI, reduce: ForLoop
			reduce(301), // const_M_E, reduce: ForLoop
			reduce(301), // panic, reduce: ForLoop
			reduce(301), // [, reduce: ForLoop
			reduce(301), // ctx_IPS, reduce: ForLoop
			reduce(301), // ctx_IFS, reduce: ForLoop
			reduce(301), // ctx_IRS, reduce: ForLoop
			reduce(301), // ctx_OPS, reduce: ForLoop
			reduce(301), // ctx_OFS, reduce: ForLoop
			reduce(301), // ctx_ORS, reduce: ForLoop
			reduce(301), // ctx_FLATSEP, reduce: ForLoop
			reduce(301), // ctx_NF, reduce: ForLoop
			reduce(301), // ctx_NR, reduce: ForLoop
			reduce(301), // ctx_FNR, reduce: ForLoop
			reduce(301), // ctx_FILENAME, reduce: ForLoop
			reduce(301), // ctx_FILENUM, reduce: ForLoop
			reduce(301), // env, reduce: ForLoop
			nil,         // [[
			nil,         // [[[
			reduce(301), // call, reduce: ForLoop
			reduce(301), // begin, reduce: ForLoop
			reduce(301), // end, reduce: ForLoop
			reduce(301), // if, reduce: ForLoop
			nil,         // elif
			nil,         // else
			reduce(301), // while, reduce: ForLoop
			reduce(301), // do, reduce: ForLoop
			reduce(301), // for, reduce: ForLoop
			nil,         // in
			reduce(301), // break, reduce: ForLoop
			reduce(301), // continue, reduce: ForLoop
			reduce(301), // func, reduce: ForLoop
			reduce(301), // subr, reduce: ForLoop
			reduce(301), // return, reduce: ForLoop
		},
	},
	actionRow{ // S160
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(302), // $, reduce: ForLoop
			nil,         // empty
			reduce(302), // ;, reduce: ForLoop
			reduce(302), // {, reduce: ForLoop
			nil,         // }
			nil,         // =
			reduce(302), // unset, reduce: ForLoop
			reduce(302), // filter, reduce: ForLoop
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(302), // print, reduce: ForLoop
			nil,         // ,
			reduce(302), // printn, reduce: ForLoop
			reduce(302), // eprint, reduce: ForLoop
			reduce(302), // eprintn, reduce: ForLoop
			reduce(302), // dump, reduce: ForLoop
			reduce(302), // edump, reduce: ForLoop
			reduce(302), // tee, reduce: ForLoop
			reduce(302), // emitf, reduce: ForLoop
			reduce(302), // emit1, reduce: ForLoop
			reduce(302), // emit, reduce: ForLoop
			reduce(302), // (, reduce: ForLoop
			nil,         // )
			reduce(302), // emitp, reduce: ForLoop
			reduce(302), // field_name, reduce: ForLoop
			reduce(302), // $[, reduce: ForLoop
			nil,         // ]
			reduce(302), // braced_field_name, reduce: ForLoop
			reduce(302), // $[[, reduce: ForLoop
			reduce(302), // $[[[, reduce: ForLoop
			reduce(302), // full_srec, reduce: ForLoop
			reduce(302), // oosvar_name, reduce: ForLoop
			reduce(302), // @[, reduce: ForLoop
			reduce(302), // braced_oosvar_name, reduce: ForLoop
			reduce(302), // full_oosvar, reduce: ForLoop
			reduce(302), // all, reduce: ForLoop
			reduce(302), // non_sigil_name, reduce: ForLoop
			reduce(302), // arr, reduce: ForLoop
			reduce(302), // bool, reduce: ForLoop
			reduce(302), // float, reduce: ForLoop
			reduce(302), // int, reduce: ForLoop
			reduce(302), // map, reduce: ForLoop
			reduce(302), // num, reduce: ForLoop
			reduce(302), // str, reduce: ForLoop
			reduce(302), // var, reduce: ForLoop
			reduce(302), // funct, reduce: ForLoop
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(302), // +, reduce: ForLoop
			reduce(302), // -, reduce: ForLoop
			reduce(302), // .+, reduce: ForLoop
			reduce(302), // .-, reduce: ForLoop
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(302), // !, reduce: ForLoop
			reduce(302), // ~, reduce: ForLoop
			nil,         // **
			reduce(302), // string_literal, reduce: ForLoop
			reduce(302), // regex_case_insensitive, reduce: ForLoop
			reduce(302), // int_literal, reduce: ForLoop
			reduce(302), // float_literal, reduce: ForLoop
			reduce(302), // boolean_literal, reduce: ForLoop
			reduce(302), // null_literal, reduce: ForLoop
			reduce(302), // inf_literal, reduce: ForLoop
			reduce(302), // nan_literal, reduce: ForLoop
			reduce(302), // const_M_PI, reduce: ForLoop
			reduce(302), // const_M_E, reduce: ForLoop
			reduce(302), // panic, reduce: ForLoop
			reduce(302), // [, reduce: ForLoop
			reduce(302), // ctx_IPS, reduce: ForLoop
			reduce(302), // ctx_IFS, reduce: ForLoop
			reduce(302), // ctx_IRS, reduce: ForLoop
			reduce(302), // ctx_OPS, reduce: ForLoop
			reduce(302), // ctx_OFS, reduce: ForLoop
			reduce(302), // ctx_ORS, reduce: ForLoop
			reduce(302), // ctx_FLATSEP, reduce: ForLoop
			reduce(302), // ctx_NF, reduce: ForLoop
			reduce(302), // ctx_NR, reduce: ForLoop
			reduce(302), // ctx_FNR, reduce: ForLoop
			reduce(302), // ctx_FILENAME, reduce: ForLoop
			reduce(302), // ctx_FILENUM, reduce: ForLoop
			reduce(302), // env, reduce: ForLoop
			nil,         // [[
			nil,         // [[[
			reduce(302), // call, reduce: ForLoop
			reduce(302), // begin, reduce: ForLoop
			reduce(302), // end, reduce: ForLoop
			reduce(302), // if, reduce: ForLoop
			nil,         // elif
			nil,         // else
			reduce(302), // while, reduce: ForLoop
			reduce(302), // do, reduce: ForLoop
			reduce(302), // for, reduce: ForLoop
			nil,         // in
			reduce(302), // break, reduce: ForLoop
			reduce(302), // continue, reduce: ForLoop
			reduce(302), // func, reduce: ForLoop
			reduce(302), // subr, reduce: ForLoop
			reduce(302), // return, reduce: ForLoop
		},
	},
	actionRow{ // S161
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(860), // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S162
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(320), // $, reduce: BreakStatement
			nil,         // empty
			reduce(320), // ;, reduce: BreakStatement
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S163
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(321), // $, reduce: ContinueStatement
			nil,         // empty
			reduce(321), // ;, reduce: ContinueStatement
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S164
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(861), // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			shift(862), // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S165
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			shift(863), // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S166
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(337), // $, reduce: ReturnStatement
			nil,         // empty
			reduce(337), // ;, reduce: ReturnStatement
			shift(384),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(388),  // (
			nil,         // )
			nil,         // emitp
			shift(400),  // field_name
			shift(401),  // $[
			nil,         // ]
			shift(402),  // braced_field_name
			shift(403),  // $[[
			shift(404),  // $[[[
			shift(405),  // full_srec
			shift(407),  // oosvar_name
			shift(408),  // @[
			shift(409),  // braced_oosvar_name
			shift(410),  // full_oosvar
			shift(411),  // all
			shift(412),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(427),  // +
			shift(429),  // -
			shift(430),  // .+
			shift(431),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(434),  // !
			shift(435),  // ~
			nil,         // **
			shift(440),  // string_literal
			shift(441),  // regex_case_insensitive
			shift(442),  // int_literal
			shift(443),  // float_literal
			shift(444),  // boolean_literal
			shift(445),  // null_literal
			shift(446),  // inf_literal
			shift(447),  // nan_literal
			shift(448),  // const_M_PI
			shift(449),  // const_M_E
			shift(450),  // panic
			shift(452),  // [
			shift(454),  // ctx_IPS
			shift(455),  // ctx_IFS
			shift(456),  // ctx_IRS
			shift(457),  // ctx_OPS
			shift(458),  // ctx_OFS
			shift(459),  // ctx_ORS
			shift(460),  // ctx_FLATSEP
			shift(461),  // ctx_NF
			shift(462),  // ctx_NR
			shift(463),  // ctx_FNR
			shift(464),  // ctx_FILENAME
			shift(465),  // ctx_FILENUM
			shift(467),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(474),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S167
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(2),  // $, reduce: StatementBlock
			nil,        // empty
			shift(6),   // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			shift(30),  // unset
			shift(33),  // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			shift(34),  // print
			nil,        // ,
			shift(35),  // printn
			shift(36),  // eprint
			shift(37),  // eprintn
			shift(38),  // dump
			shift(39),  // edump
			shift(40),  // tee
			shift(42),  // emitf
			shift(43),  // emit1
			shift(44),  // emit
			shift(45),  // (
			nil,        // )
			shift(46),  // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			shift(71),  // arr
			shift(72),  // bool
			shift(73),  // float
			shift(74),  // int
			shift(75),  // map
			shift(76),  // num
			shift(77),  // str
			shift(78),  // var
			shift(79),  // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			shift(141), // call
			shift(150), // begin
			shift(151), // end
			shift(154), // if
			nil,        // elif
			nil,        // else
			shift(155), // while
			shift(156), // do
			shift(161), // for
			nil,        // in
			shift(162), // break
			shift(163), // continue
			shift(164), // func
			shift(165), // subr
			shift(166), // return
		},
	},
	actionRow{ // S168
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(10), // $, reduce: NonEmptyStatementBlock
			nil,        // empty
			shift(866), // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S169
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(2),  // $, reduce: StatementBlock
			nil,        // empty
			shift(6),   // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			shift(30),  // unset
			shift(33),  // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			shift(34),  // print
			nil,        // ,
			shift(35),  // printn
			shift(36),  // eprint
			shift(37),  // eprintn
			shift(38),  // dump
			shift(39),  // edump
			shift(40),  // tee
			shift(42),  // emitf
			shift(43),  // emit1
			shift(44),  // emit
			shift(45),  // (
			nil,        // )
			shift(46),  // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			shift(71),  // arr
			shift(72),  // bool
			shift(73),  // float
			shift(74),  // int
			shift(75),  // map
			shift(76),  // num
			shift(77),  // str
			shift(78),  // var
			shift(79),  // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			shift(141), // call
			shift(150), // begin
			shift(151), // end
			shift(154), // if
			nil,        // elif
			nil,        // else
			shift(155), // while
			shift(156), // do
			shift(161), // for
			nil,        // in
			shift(162), // break
			shift(163), // continue
			shift(164), // func
			shift(165), // subr
			shift(166), // return
		},
	},
	actionRow{ // S170
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(2),  // $, reduce: StatementBlock
			nil,        // empty
			shift(6),   // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			shift(30),  // unset
			shift(33),  // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			shift(34),  // print
			nil,        // ,
			shift(35),  // printn
			shift(36),  // eprint
			shift(37),  // eprintn
			shift(38),  // dump
			shift(39),  // edump
			shift(40),  // tee
			shift(42),  // emitf
			shift(43),  // emit1
			shift(44),  // emit
			shift(45),  // (
			nil,        // )
			shift(46),  // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			shift(71),  // arr
			shift(72),  // bool
			shift(73),  // float
			shift(74),  // int
			shift(75),  // map
			shift(76),  // num
			shift(77),  // str
			shift(78),  // var
			shift(79),  // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			shift(141), // call
			shift(150), // begin
			shift(151), // end
			shift(154), // if
			nil,        // elif
			nil,        // else
			shift(155), // while
			shift(156), // do
			shift(161), // for
			nil,        // in
			shift(162), // break
			shift(163), // continue
			shift(164), // func
			shift(165), // subr
			shift(166), // return
		},
	},
	actionRow{ // S171
		canRecover: false,
		actions: [numSymbols]action{
			nil,       // INVALID
			reduce(6), // $, reduce: NonEmptyStatementBlock
			nil,       // empty
			nil,       // ;
			nil,       // {
			nil,       // }
			nil,       // =
			nil,       // unset
			nil,       // filter
			nil,       // >
			nil,       // >>
			nil,       // |
			nil,       // stdout
			nil,       // stderr
			nil,       // print
			nil,       // ,
			nil,       // printn
			nil,       // eprint
			nil,       // eprintn
			nil,       // dump
			nil,       // edump
			nil,       // tee
			nil,       // emitf
			nil,       // emit1
			nil,       // emit
			nil,       // (
			nil,       // )
			nil,       // emitp
			nil,       // field_name
			nil,       // $[
			nil,       // ]
			nil,       // braced_field_name
			nil,       // $[[
			nil,       // $[[[
			nil,       // full_srec
			nil,       // oosvar_name
			nil,       // @[
			nil,       // braced_oosvar_name
			nil,       // full_oosvar
			nil,       // all
			nil,       // non_sigil_name
			nil,       // arr
			nil,       // bool
			nil,       // float
			nil,       // int
			nil,       // map
			nil,       // num
			nil,       // str
			nil,       // var
			nil,       // funct
			nil,       // ||=
			nil,       // ^^=
			nil,       // &&=
			nil,       // ??=
			nil,       // ???=
			nil,       // |=
			nil,       // &=
			nil,       // ^=
			nil,       // <<=
			nil,       // >>=
			nil,       // >>>=
			nil,       // +=
			nil,       // .=
			nil,       // -=
			nil,       // *=
			nil,       // /=
			nil,       // //=
			nil,       // %=
			nil,       // **=
			nil,       // ?
			nil,       // :
			nil,       // ||
			nil,       // ^^
			nil,       // &&
			nil,       // ??
			nil,       // ???
			nil,       // =~
			nil,       // !=~
			nil,       // ==
			nil,       // !=
			nil,       // <=>
			nil,       // >=
			nil,       // <
			nil,       // <=
			nil,       // ^
			nil,       // &
			nil,       // <<
			nil,       // >>>
			nil,       // +
			nil,       // -
			nil,       // .+
			nil,       // .-
			nil,       // *
			nil,       // /
			nil,       // //
			nil,       // %
			nil,       // .*
			nil,       // ./
			nil,       // .//
			nil,       // .
			nil,       // !
			nil,       // ~
			nil,       // **
			nil,       // string_literal
			nil,       // regex_case_insensitive
			nil,       // int_literal
			nil,       // float_literal
			nil,       // boolean_literal
			nil,       // null_literal
			nil,       // inf_literal
			nil,       // nan_literal
			nil,       // const_M_PI
			nil,       // const_M_E
			nil,       // panic
			nil,       // [
			nil,       // ctx_IPS
			nil,       // ctx_IFS
			nil,       // ctx_IRS
			nil,       // ctx_OPS
			nil,       // ctx_OFS
			nil,       // ctx_ORS
			nil,       // ctx_FLATSEP
			nil,       // ctx_NF
			nil,       // ctx_NR
			nil,       // ctx_FNR
			nil,       // ctx_FILENAME
			nil,       // ctx_FILENUM
			nil,       // env
			nil,       // [[
			nil,       // [[[
			nil,       // call
			nil,       // begin
			nil,       // end
			nil,       // if
			nil,       // elif
			nil,       // else
			nil,       // while
			nil,       // do
			nil,       // for
			nil,       // in
			nil,       // break
			nil,       // continue
			nil,       // func
			nil,       // subr
			nil,       // return
		},
	},
	actionRow{ // S172
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			shift(869), // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S173
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(234), // $, reduce: MapLiteral
			nil,         // empty
			reduce(234), // ;, reduce: MapLiteral
			reduce(234), // {, reduce: MapLiteral
			nil,         // }
			reduce(234), // =, reduce: MapLiteral
			nil,         // unset
			nil,         // filter
			reduce(234), // >, reduce: MapLiteral
			reduce(234), // >>, reduce: MapLiteral
			reduce(234), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(234), // ||=, reduce: MapLiteral
			reduce(234), // ^^=, reduce: MapLiteral
			reduce(234), // &&=, reduce: MapLiteral
			reduce(234), // ??=, reduce: MapLiteral
			reduce(234), // ???=, reduce: MapLiteral
			reduce(234), // |=, reduce: MapLiteral
			reduce(234), // &=, reduce: MapLiteral
			reduce(234), // ^=, reduce: MapLiteral
			reduce(234), // <<=, reduce: MapLiteral
			reduce(234), // >>=, reduce: MapLiteral
			reduce(234), // >>>=, reduce: MapLiteral
			reduce(234), // +=, reduce: MapLiteral
			reduce(234), // .=, reduce: MapLiteral
			reduce(234), // -=, reduce: MapLiteral
			reduce(234), // *=, reduce: MapLiteral
			reduce(234), // /=, reduce: MapLiteral
			reduce(234), // //=, reduce: MapLiteral
			reduce(234), // %=, reduce: MapLiteral
			reduce(234), // **=, reduce: MapLiteral
			reduce(234), // ?, reduce: MapLiteral
			nil,         // :
			reduce(234), // ||, reduce: MapLiteral
			reduce(234), // ^^, reduce: MapLiteral
			reduce(234), // &&, reduce: MapLiteral
			reduce(234), // ??, reduce: MapLiteral
			reduce(234), // ???, reduce: MapLiteral
			reduce(234), // =~, reduce: MapLiteral
			reduce(234), // !=~, reduce: MapLiteral
			reduce(234), // ==, reduce: MapLiteral
			reduce(234), // !=, reduce: MapLiteral
			reduce(234), // <=>, reduce: MapLiteral
			reduce(234), // >=, reduce: MapLiteral
			reduce(234), // <, reduce: MapLiteral
			reduce(234), // <=, reduce: MapLiteral
			reduce(234), // ^, reduce: MapLiteral
			reduce(234), // &, reduce: MapLiteral
			reduce(234), // <<, reduce: MapLiteral
			reduce(234), // >>>, reduce: MapLiteral
			reduce(234), // +, reduce: MapLiteral
			reduce(234), // -, reduce: MapLiteral
			reduce(234), // .+, reduce: MapLiteral
			reduce(234), // .-, reduce: MapLiteral
			reduce(234), // *, reduce: MapLiteral
			reduce(234), // /, reduce: MapLiteral
			reduce(234), // //, reduce: MapLiteral
			reduce(234), // %, reduce: MapLiteral
			reduce(234), // .*, reduce: MapLiteral
			reduce(234), // ./, reduce: MapLiteral
			reduce(234), // .//, reduce: MapLiteral
			reduce(234), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(234), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(234), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(234), // [[, reduce: MapLiteral
			reduce(234), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S174
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			shift(871), // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S175
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(214), // ?, reduce: MlrvalOrFunction
			reduce(214), // :, reduce: MlrvalOrFunction
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S176
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(211), // >, reduce: MlrvalOrFunction
			reduce(211), // >>, reduce: MlrvalOrFunction
			reduce(211), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(211), // ?, reduce: MlrvalOrFunction
			reduce(211), // :, reduce: MlrvalOrFunction
			reduce(211), // ||, reduce: MlrvalOrFunction
			reduce(211), // ^^, reduce: MlrvalOrFunction
			reduce(211), // &&, reduce: MlrvalOrFunction
			reduce(211), // ??, reduce: MlrvalOrFunction
			reduce(211), // ???, reduce: MlrvalOrFunction
			reduce(211), // =~, reduce: MlrvalOrFunction
			reduce(211), // !=~, reduce: MlrvalOrFunction
			reduce(211), // ==, reduce: MlrvalOrFunction
			reduce(211), // !=, reduce: MlrvalOrFunction
			reduce(211), // <=>, reduce: MlrvalOrFunction
			reduce(211), // >=, reduce: MlrvalOrFunction
			reduce(211), // <, reduce: MlrvalOrFunction
			reduce(211), // <=, reduce: MlrvalOrFunction
			reduce(211), // ^, reduce: MlrvalOrFunction
			reduce(211), // &, reduce: MlrvalOrFunction
			reduce(211), // <<, reduce: MlrvalOrFunction
			reduce(211), // >>>, reduce: MlrvalOrFunction
			reduce(211), // +, reduce: MlrvalOrFunction
			reduce(211), // -, reduce: MlrvalOrFunction
			reduce(211), // .+, reduce: MlrvalOrFunction
			reduce(211), // .-, reduce: MlrvalOrFunction
			reduce(211), // *, reduce: MlrvalOrFunction
			reduce(211), // /, reduce: MlrvalOrFunction
			reduce(211), // //, reduce: MlrvalOrFunction
			reduce(211), // %, reduce: MlrvalOrFunction
			reduce(211), // .*, reduce: MlrvalOrFunction
			reduce(211), // ./, reduce: MlrvalOrFunction
			reduce(211), // .//, reduce: MlrvalOrFunction
			reduce(211), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(211), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(211), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(211), // [[, reduce: MlrvalOrFunction
			reduce(211), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S177
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S178
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(110), // >, reduce: OosvarValue
			reduce(110), // >>, reduce: OosvarValue
			reduce(110), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(110), // ?, reduce: OosvarValue
			reduce(110), // :, reduce: OosvarValue
			reduce(110), // ||, reduce: OosvarValue
			reduce(110), // ^^, reduce: OosvarValue
			reduce(110), // &&, reduce: OosvarValue
			reduce(110), // ??, reduce: OosvarValue
			reduce(110), // ???, reduce: OosvarValue
			reduce(110), // =~, reduce: OosvarValue
			reduce(110), // !=~, reduce: OosvarValue
			reduce(110), // ==, reduce: OosvarValue
			reduce(110), // !=, reduce: OosvarValue
			reduce(110), // <=>, reduce: OosvarValue
			reduce(110), // >=, reduce: OosvarValue
			reduce(110), // <, reduce: OosvarValue
			reduce(110), // <=, reduce: OosvarValue
			reduce(110), // ^, reduce: OosvarValue
			reduce(110), // &, reduce: OosvarValue
			reduce(110), // <<, reduce: OosvarValue
			reduce(110), // >>>, reduce: OosvarValue
			reduce(110), // +, reduce: OosvarValue
			reduce(110), // -, reduce: OosvarValue
			reduce(110), // .+, reduce: OosvarValue
			reduce(110), // .-, reduce: OosvarValue
			reduce(110), // *, reduce: OosvarValue
			reduce(110), // /, reduce: OosvarValue
			reduce(110), // //, reduce: OosvarValue
			reduce(110), // %, reduce: OosvarValue
			reduce(110), // .*, reduce: OosvarValue
			reduce(110), // ./, reduce: OosvarValue
			reduce(110), // .//, reduce: OosvarValue
			reduce(110), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(110), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(110), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(110), // [[, reduce: OosvarValue
			reduce(110), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S179
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(112), // >, reduce: OosvarValue
			reduce(112), // >>, reduce: OosvarValue
			reduce(112), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(112), // ?, reduce: OosvarValue
			reduce(112), // :, reduce: OosvarValue
			reduce(112), // ||, reduce: OosvarValue
			reduce(112), // ^^, reduce: OosvarValue
			reduce(112), // &&, reduce: OosvarValue
			reduce(112), // ??, reduce: OosvarValue
			reduce(112), // ???, reduce: OosvarValue
			reduce(112), // =~, reduce: OosvarValue
			reduce(112), // !=~, reduce: OosvarValue
			reduce(112), // ==, reduce: OosvarValue
			reduce(112), // !=, reduce: OosvarValue
			reduce(112), // <=>, reduce: OosvarValue
			reduce(112), // >=, reduce: OosvarValue
			reduce(112), // <, reduce: OosvarValue
			reduce(112), // <=, reduce: OosvarValue
			reduce(112), // ^, reduce: OosvarValue
			reduce(112), // &, reduce: OosvarValue
			reduce(112), // <<, reduce: OosvarValue
			reduce(112), // >>>, reduce: OosvarValue
			reduce(112), // +, reduce: OosvarValue
			reduce(112), // -, reduce: OosvarValue
			reduce(112), // .+, reduce: OosvarValue
			reduce(112), // .-, reduce: OosvarValue
			reduce(112), // *, reduce: OosvarValue
			reduce(112), // /, reduce: OosvarValue
			reduce(112), // //, reduce: OosvarValue
			reduce(112), // %, reduce: OosvarValue
			reduce(112), // .*, reduce: OosvarValue
			reduce(112), // ./, reduce: OosvarValue
			reduce(112), // .//, reduce: OosvarValue
			reduce(112), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(112), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(112), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(112), // [[, reduce: OosvarValue
			reduce(112), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S180
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(111), // >, reduce: OosvarValue
			reduce(111), // >>, reduce: OosvarValue
			reduce(111), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(111), // ?, reduce: OosvarValue
			reduce(111), // :, reduce: OosvarValue
			reduce(111), // ||, reduce: OosvarValue
			reduce(111), // ^^, reduce: OosvarValue
			reduce(111), // &&, reduce: OosvarValue
			reduce(111), // ??, reduce: OosvarValue
			reduce(111), // ???, reduce: OosvarValue
			reduce(111), // =~, reduce: OosvarValue
			reduce(111), // !=~, reduce: OosvarValue
			reduce(111), // ==, reduce: OosvarValue
			reduce(111), // !=, reduce: OosvarValue
			reduce(111), // <=>, reduce: OosvarValue
			reduce(111), // >=, reduce: OosvarValue
			reduce(111), // <, reduce: OosvarValue
			reduce(111), // <=, reduce: OosvarValue
			reduce(111), // ^, reduce: OosvarValue
			reduce(111), // &, reduce: OosvarValue
			reduce(111), // <<, reduce: OosvarValue
			reduce(111), // >>>, reduce: OosvarValue
			reduce(111), // +, reduce: OosvarValue
			reduce(111), // -, reduce: OosvarValue
			reduce(111), // .+, reduce: OosvarValue
			reduce(111), // .-, reduce: OosvarValue
			reduce(111), // *, reduce: OosvarValue
			reduce(111), // /, reduce: OosvarValue
			reduce(111), // //, reduce: OosvarValue
			reduce(111), // %, reduce: OosvarValue
			reduce(111), // .*, reduce: OosvarValue
			reduce(111), // ./, reduce: OosvarValue
			reduce(111), // .//, reduce: OosvarValue
			reduce(111), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(111), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(111), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(111), // [[, reduce: OosvarValue
			reduce(111), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S181
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(99), // >, reduce: FieldValue
			reduce(99), // >>, reduce: FieldValue
			reduce(99), // |, reduce: FieldValue
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(99), // ?, reduce: FieldValue
			reduce(99), // :, reduce: FieldValue
			reduce(99), // ||, reduce: FieldValue
			reduce(99), // ^^, reduce: FieldValue
			reduce(99), // &&, reduce: FieldValue
			reduce(99), // ??, reduce: FieldValue
			reduce(99), // ???, reduce: FieldValue
			reduce(99), // =~, reduce: FieldValue
			reduce(99), // !=~, reduce: FieldValue
			reduce(99), // ==, reduce: FieldValue
			reduce(99), // !=, reduce: FieldValue
			reduce(99), // <=>, reduce: FieldValue
			reduce(99), // >=, reduce: FieldValue
			reduce(99), // <, reduce: FieldValue
			reduce(99), // <=, reduce: FieldValue
			reduce(99), // ^, reduce: FieldValue
			reduce(99), // &, reduce: FieldValue
			reduce(99), // <<, reduce: FieldValue
			reduce(99), // >>>, reduce: FieldValue
			reduce(99), // +, reduce: FieldValue
			reduce(99), // -, reduce: FieldValue
			reduce(99), // .+, reduce: FieldValue
			reduce(99), // .-, reduce: FieldValue
			reduce(99), // *, reduce: FieldValue
			reduce(99), // /, reduce: FieldValue
			reduce(99), // //, reduce: FieldValue
			reduce(99), // %, reduce: FieldValue
			reduce(99), // .*, reduce: FieldValue
			reduce(99), // ./, reduce: FieldValue
			reduce(99), // .//, reduce: FieldValue
			reduce(99), // ., reduce: FieldValue
			nil,        // !
			nil,        // ~
			reduce(99), // **, reduce: FieldValue
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(99), // [, reduce: FieldValue
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(99), // [[, reduce: FieldValue
			reduce(99), // [[[, reduce: FieldValue
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S182
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(101), // >, reduce: FieldValue
			reduce(101), // >>, reduce: FieldValue
			reduce(101), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(101), // ?, reduce: FieldValue
			reduce(101), // :, reduce: FieldValue
			reduce(101), // ||, reduce: FieldValue
			reduce(101), // ^^, reduce: FieldValue
			reduce(101), // &&, reduce: FieldValue
			reduce(101), // ??, reduce: FieldValue
			reduce(101), // ???, reduce: FieldValue
			reduce(101), // =~, reduce: FieldValue
			reduce(101), // !=~, reduce: FieldValue
			reduce(101), // ==, reduce: FieldValue
			reduce(101), // !=, reduce: FieldValue
			reduce(101), // <=>, reduce: FieldValue
			reduce(101), // >=, reduce: FieldValue
			reduce(101), // <, reduce: FieldValue
			reduce(101), // <=, reduce: FieldValue
			reduce(101), // ^, reduce: FieldValue
			reduce(101), // &, reduce: FieldValue
			reduce(101), // <<, reduce: FieldValue
			reduce(101), // >>>, reduce: FieldValue
			reduce(101), // +, reduce: FieldValue
			reduce(101), // -, reduce: FieldValue
			reduce(101), // .+, reduce: FieldValue
			reduce(101), // .-, reduce: FieldValue
			reduce(101), // *, reduce: FieldValue
			reduce(101), // /, reduce: FieldValue
			reduce(101), // //, reduce: FieldValue
			reduce(101), // %, reduce: FieldValue
			reduce(101), // .*, reduce: FieldValue
			reduce(101), // ./, reduce: FieldValue
			reduce(101), // .//, reduce: FieldValue
			reduce(101), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(101), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(101), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(101), // [[, reduce: FieldValue
			reduce(101), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S183
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(100), // >, reduce: FieldValue
			reduce(100), // >>, reduce: FieldValue
			reduce(100), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(100), // ?, reduce: FieldValue
			reduce(100), // :, reduce: FieldValue
			reduce(100), // ||, reduce: FieldValue
			reduce(100), // ^^, reduce: FieldValue
			reduce(100), // &&, reduce: FieldValue
			reduce(100), // ??, reduce: FieldValue
			reduce(100), // ???, reduce: FieldValue
			reduce(100), // =~, reduce: FieldValue
			reduce(100), // !=~, reduce: FieldValue
			reduce(100), // ==, reduce: FieldValue
			reduce(100), // !=, reduce: FieldValue
			reduce(100), // <=>, reduce: FieldValue
			reduce(100), // >=, reduce: FieldValue
			reduce(100), // <, reduce: FieldValue
			reduce(100), // <=, reduce: FieldValue
			reduce(100), // ^, reduce: FieldValue
			reduce(100), // &, reduce: FieldValue
			reduce(100), // <<, reduce: FieldValue
			reduce(100), // >>>, reduce: FieldValue
			reduce(100), // +, reduce: FieldValue
			reduce(100), // -, reduce: FieldValue
			reduce(100), // .+, reduce: FieldValue
			reduce(100), // .-, reduce: FieldValue
			reduce(100), // *, reduce: FieldValue
			reduce(100), // /, reduce: FieldValue
			reduce(100), // //, reduce: FieldValue
			reduce(100), // %, reduce: FieldValue
			reduce(100), // .*, reduce: FieldValue
			reduce(100), // ./, reduce: FieldValue
			reduce(100), // .//, reduce: FieldValue
			reduce(100), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(100), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(100), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(100), // [[, reduce: FieldValue
			reduce(100), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S184
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(213), // >, reduce: MlrvalOrFunction
			reduce(213), // >>, reduce: MlrvalOrFunction
			reduce(213), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(213), // ?, reduce: MlrvalOrFunction
			reduce(213), // :, reduce: MlrvalOrFunction
			reduce(213), // ||, reduce: MlrvalOrFunction
			reduce(213), // ^^, reduce: MlrvalOrFunction
			reduce(213), // &&, reduce: MlrvalOrFunction
			reduce(213), // ??, reduce: MlrvalOrFunction
			reduce(213), // ???, reduce: MlrvalOrFunction
			reduce(213), // =~, reduce: MlrvalOrFunction
			reduce(213), // !=~, reduce: MlrvalOrFunction
			reduce(213), // ==, reduce: MlrvalOrFunction
			reduce(213), // !=, reduce: MlrvalOrFunction
			reduce(213), // <=>, reduce: MlrvalOrFunction
			reduce(213), // >=, reduce: MlrvalOrFunction
			reduce(213), // <, reduce: MlrvalOrFunction
			reduce(213), // <=, reduce: MlrvalOrFunction
			reduce(213), // ^, reduce: MlrvalOrFunction
			reduce(213), // &, reduce: MlrvalOrFunction
			reduce(213), // <<, reduce: MlrvalOrFunction
			reduce(213), // >>>, reduce: MlrvalOrFunction
			reduce(213), // +, reduce: MlrvalOrFunction
			reduce(213), // -, reduce: MlrvalOrFunction
			reduce(213), // .+, reduce: MlrvalOrFunction
			reduce(213), // .-, reduce: MlrvalOrFunction
			reduce(213), // *, reduce: MlrvalOrFunction
			reduce(213), // /, reduce: MlrvalOrFunction
			reduce(213), // //, reduce: MlrvalOrFunction
			reduce(213), // %, reduce: MlrvalOrFunction
			reduce(213), // .*, reduce: MlrvalOrFunction
			reduce(213), // ./, reduce: MlrvalOrFunction
			reduce(213), // .//, reduce: MlrvalOrFunction
			reduce(213), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(213), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(213), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(213), // [[, reduce: MlrvalOrFunction
			reduce(213), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S185
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(233), // >, reduce: MlrvalOrFunction
			reduce(233), // >>, reduce: MlrvalOrFunction
			reduce(233), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(233), // ?, reduce: MlrvalOrFunction
			reduce(233), // :, reduce: MlrvalOrFunction
			reduce(233), // ||, reduce: MlrvalOrFunction
			reduce(233), // ^^, reduce: MlrvalOrFunction
			reduce(233), // &&, reduce: MlrvalOrFunction
			reduce(233), // ??, reduce: MlrvalOrFunction
			reduce(233), // ???, reduce: MlrvalOrFunction
			reduce(233), // =~, reduce: MlrvalOrFunction
			reduce(233), // !=~, reduce: MlrvalOrFunction
			reduce(233), // ==, reduce: MlrvalOrFunction
			reduce(233), // !=, reduce: MlrvalOrFunction
			reduce(233), // <=>, reduce: MlrvalOrFunction
			reduce(233), // >=, reduce: MlrvalOrFunction
			reduce(233), // <, reduce: MlrvalOrFunction
			reduce(233), // <=, reduce: MlrvalOrFunction
			reduce(233), // ^, reduce: MlrvalOrFunction
			reduce(233), // &, reduce: MlrvalOrFunction
			reduce(233), // <<, reduce: MlrvalOrFunction
			reduce(233), // >>>, reduce: MlrvalOrFunction
			reduce(233), // +, reduce: MlrvalOrFunction
			reduce(233), // -, reduce: MlrvalOrFunction
			reduce(233), // .+, reduce: MlrvalOrFunction
			reduce(233), // .-, reduce: MlrvalOrFunction
			reduce(233), // *, reduce: MlrvalOrFunction
			reduce(233), // /, reduce: MlrvalOrFunction
			reduce(233), // //, reduce: MlrvalOrFunction
			reduce(233), // %, reduce: MlrvalOrFunction
			reduce(233), // .*, reduce: MlrvalOrFunction
			reduce(233), // ./, reduce: MlrvalOrFunction
			reduce(233), // .//, reduce: MlrvalOrFunction
			reduce(233), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(233), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(233), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(233), // [[, reduce: MlrvalOrFunction
			reduce(233), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S186
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(210), // >, reduce: MlrvalOrFunction
			reduce(210), // >>, reduce: MlrvalOrFunction
			reduce(210), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(210), // ?, reduce: MlrvalOrFunction
			reduce(210), // :, reduce: MlrvalOrFunction
			reduce(210), // ||, reduce: MlrvalOrFunction
			reduce(210), // ^^, reduce: MlrvalOrFunction
			reduce(210), // &&, reduce: MlrvalOrFunction
			reduce(210), // ??, reduce: MlrvalOrFunction
			reduce(210), // ???, reduce: MlrvalOrFunction
			reduce(210), // =~, reduce: MlrvalOrFunction
			reduce(210), // !=~, reduce: MlrvalOrFunction
			reduce(210), // ==, reduce: MlrvalOrFunction
			reduce(210), // !=, reduce: MlrvalOrFunction
			reduce(210), // <=>, reduce: MlrvalOrFunction
			reduce(210), // >=, reduce: MlrvalOrFunction
			reduce(210), // <, reduce: MlrvalOrFunction
			reduce(210), // <=, reduce: MlrvalOrFunction
			reduce(210), // ^, reduce: MlrvalOrFunction
			reduce(210), // &, reduce: MlrvalOrFunction
			reduce(210), // <<, reduce: MlrvalOrFunction
			reduce(210), // >>>, reduce: MlrvalOrFunction
			reduce(210), // +, reduce: MlrvalOrFunction
			reduce(210), // -, reduce: MlrvalOrFunction
			reduce(210), // .+, reduce: MlrvalOrFunction
			reduce(210), // .-, reduce: MlrvalOrFunction
			reduce(210), // *, reduce: MlrvalOrFunction
			reduce(210), // /, reduce: MlrvalOrFunction
			reduce(210), // //, reduce: MlrvalOrFunction
			reduce(210), // %, reduce: MlrvalOrFunction
			reduce(210), // .*, reduce: MlrvalOrFunction
			reduce(210), // ./, reduce: MlrvalOrFunction
			reduce(210), // .//, reduce: MlrvalOrFunction
			reduce(210), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(210), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(210), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(210), // [[, reduce: MlrvalOrFunction
			reduce(210), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S187
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(102), // >, reduce: FieldValue
			reduce(102), // >>, reduce: FieldValue
			reduce(102), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(102), // ?, reduce: FieldValue
			reduce(102), // :, reduce: FieldValue
			reduce(102), // ||, reduce: FieldValue
			reduce(102), // ^^, reduce: FieldValue
			reduce(102), // &&, reduce: FieldValue
			reduce(102), // ??, reduce: FieldValue
			reduce(102), // ???, reduce: FieldValue
			reduce(102), // =~, reduce: FieldValue
			reduce(102), // !=~, reduce: FieldValue
			reduce(102), // ==, reduce: FieldValue
			reduce(102), // !=, reduce: FieldValue
			reduce(102), // <=>, reduce: FieldValue
			reduce(102), // >=, reduce: FieldValue
			reduce(102), // <, reduce: FieldValue
			reduce(102), // <=, reduce: FieldValue
			reduce(102), // ^, reduce: FieldValue
			reduce(102), // &, reduce: FieldValue
			reduce(102), // <<, reduce: FieldValue
			reduce(102), // >>>, reduce: FieldValue
			reduce(102), // +, reduce: FieldValue
			reduce(102), // -, reduce: FieldValue
			reduce(102), // .+, reduce: FieldValue
			reduce(102), // .-, reduce: FieldValue
			reduce(102), // *, reduce: FieldValue
			reduce(102), // /, reduce: FieldValue
			reduce(102), // //, reduce: FieldValue
			reduce(102), // %, reduce: FieldValue
			reduce(102), // .*, reduce: FieldValue
			reduce(102), // ./, reduce: FieldValue
			reduce(102), // .//, reduce: FieldValue
			reduce(102), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(102), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(102), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(102), // [[, reduce: FieldValue
			reduce(102), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S188
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(103), // >, reduce: FieldValue
			reduce(103), // >>, reduce: FieldValue
			reduce(103), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(103), // ?, reduce: FieldValue
			reduce(103), // :, reduce: FieldValue
			reduce(103), // ||, reduce: FieldValue
			reduce(103), // ^^, reduce: FieldValue
			reduce(103), // &&, reduce: FieldValue
			reduce(103), // ??, reduce: FieldValue
			reduce(103), // ???, reduce: FieldValue
			reduce(103), // =~, reduce: FieldValue
			reduce(103), // !=~, reduce: FieldValue
			reduce(103), // ==, reduce: FieldValue
			reduce(103), // !=, reduce: FieldValue
			reduce(103), // <=>, reduce: FieldValue
			reduce(103), // >=, reduce: FieldValue
			reduce(103), // <, reduce: FieldValue
			reduce(103), // <=, reduce: FieldValue
			reduce(103), // ^, reduce: FieldValue
			reduce(103), // &, reduce: FieldValue
			reduce(103), // <<, reduce: FieldValue
			reduce(103), // >>>, reduce: FieldValue
			reduce(103), // +, reduce: FieldValue
			reduce(103), // -, reduce: FieldValue
			reduce(103), // .+, reduce: FieldValue
			reduce(103), // .-, reduce: FieldValue
			reduce(103), // *, reduce: FieldValue
			reduce(103), // /, reduce: FieldValue
			reduce(103), // //, reduce: FieldValue
			reduce(103), // %, reduce: FieldValue
			reduce(103), // .*, reduce: FieldValue
			reduce(103), // ./, reduce: FieldValue
			reduce(103), // .//, reduce: FieldValue
			reduce(103), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(103), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(103), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(103), // [[, reduce: FieldValue
			reduce(103), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S189
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(104), // >, reduce: DirectFieldValue
			reduce(104), // >>, reduce: DirectFieldValue
			reduce(104), // |, reduce: DirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(104), // ?, reduce: DirectFieldValue
			reduce(104), // :, reduce: DirectFieldValue
			reduce(104), // ||, reduce: DirectFieldValue
			reduce(104), // ^^, reduce: DirectFieldValue
			reduce(104), // &&, reduce: DirectFieldValue
			reduce(104), // ??, reduce: DirectFieldValue
			reduce(104), // ???, reduce: DirectFieldValue
			reduce(104), // =~, reduce: DirectFieldValue
			reduce(104), // !=~, reduce: DirectFieldValue
			reduce(104), // ==, reduce: DirectFieldValue
			reduce(104), // !=, reduce: DirectFieldValue
			reduce(104), // <=>, reduce: DirectFieldValue
			reduce(104), // >=, reduce: DirectFieldValue
			reduce(104), // <, reduce: DirectFieldValue
			reduce(104), // <=, reduce: DirectFieldValue
			reduce(104), // ^, reduce: DirectFieldValue
			reduce(104), // &, reduce: DirectFieldValue
			reduce(104), // <<, reduce: DirectFieldValue
			reduce(104), // >>>, reduce: DirectFieldValue
			reduce(104), // +, reduce: DirectFieldValue
			reduce(104), // -, reduce: DirectFieldValue
			reduce(104), // .+, reduce: DirectFieldValue
			reduce(104), // .-, reduce: DirectFieldValue
			reduce(104), // *, reduce: DirectFieldValue
			reduce(104), // /, reduce: DirectFieldValue
			reduce(104), // //, reduce: DirectFieldValue
			reduce(104), // %, reduce: DirectFieldValue
			reduce(104), // .*, reduce: DirectFieldValue
			reduce(104), // ./, reduce: DirectFieldValue
			reduce(104), // .//, reduce: DirectFieldValue
			reduce(104), // ., reduce: DirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(104), // **, reduce: DirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(104), // [, reduce: DirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(104), // [[, reduce: DirectFieldValue
			reduce(104), // [[[, reduce: DirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S190
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S191
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(106), // >, reduce: BracedFieldValue
			reduce(106), // >>, reduce: BracedFieldValue
			reduce(106), // |, reduce: BracedFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(106), // ?, reduce: BracedFieldValue
			reduce(106), // :, reduce: BracedFieldValue
			reduce(106), // ||, reduce: BracedFieldValue
			reduce(106), // ^^, reduce: BracedFieldValue
			reduce(106), // &&, reduce: BracedFieldValue
			reduce(106), // ??, reduce: BracedFieldValue
			reduce(106), // ???, reduce: BracedFieldValue
			reduce(106), // =~, reduce: BracedFieldValue
			reduce(106), // !=~, reduce: BracedFieldValue
			reduce(106), // ==, reduce: BracedFieldValue
			reduce(106), // !=, reduce: BracedFieldValue
			reduce(106), // <=>, reduce: BracedFieldValue
			reduce(106), // >=, reduce: BracedFieldValue
			reduce(106), // <, reduce: BracedFieldValue
			reduce(106), // <=, reduce: BracedFieldValue
			reduce(106), // ^, reduce: BracedFieldValue
			reduce(106), // &, reduce: BracedFieldValue
			reduce(106), // <<, reduce: BracedFieldValue
			reduce(106), // >>>, reduce: BracedFieldValue
			reduce(106), // +, reduce: BracedFieldValue
			reduce(106), // -, reduce: BracedFieldValue
			reduce(106), // .+, reduce: BracedFieldValue
			reduce(106), // .-, reduce: BracedFieldValue
			reduce(106), // *, reduce: BracedFieldValue
			reduce(106), // /, reduce: BracedFieldValue
			reduce(106), // //, reduce: BracedFieldValue
			reduce(106), // %, reduce: BracedFieldValue
			reduce(106), // .*, reduce: BracedFieldValue
			reduce(106), // ./, reduce: BracedFieldValue
			reduce(106), // .//, reduce: BracedFieldValue
			reduce(106), // ., reduce: BracedFieldValue
			nil,         // !
			nil,         // ~
			reduce(106), // **, reduce: BracedFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(106), // [, reduce: BracedFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(106), // [[, reduce: BracedFieldValue
			reduce(106), // [[[, reduce: BracedFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S192
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S193
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S194
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(109), // >, reduce: FullSrec
			reduce(109), // >>, reduce: FullSrec
			reduce(109), // |, reduce: FullSrec
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(109), // ?, reduce: FullSrec
			reduce(109), // :, reduce: FullSrec
			reduce(109), // ||, reduce: FullSrec
			reduce(109), // ^^, reduce: FullSrec
			reduce(109), // &&, reduce: FullSrec
			reduce(109), // ??, reduce: FullSrec
			reduce(109), // ???, reduce: FullSrec
			reduce(109), // =~, reduce: FullSrec
			reduce(109), // !=~, reduce: FullSrec
			reduce(109), // ==, reduce: FullSrec
			reduce(109), // !=, reduce: FullSrec
			reduce(109), // <=>, reduce: FullSrec
			reduce(109), // >=, reduce: FullSrec
			reduce(109), // <, reduce: FullSrec
			reduce(109), // <=, reduce: FullSrec
			reduce(109), // ^, reduce: FullSrec
			reduce(109), // &, reduce: FullSrec
			reduce(109), // <<, reduce: FullSrec
			reduce(109), // >>>, reduce: FullSrec
			reduce(109), // +, reduce: FullSrec
			reduce(109), // -, reduce: FullSrec
			reduce(109), // .+, reduce: FullSrec
			reduce(109), // .-, reduce: FullSrec
			reduce(109), // *, reduce: FullSrec
			reduce(109), // /, reduce: FullSrec
			reduce(109), // //, reduce: FullSrec
			reduce(109), // %, reduce: FullSrec
			reduce(109), // .*, reduce: FullSrec
			reduce(109), // ./, reduce: FullSrec
			reduce(109), // .//, reduce: FullSrec
			reduce(109), // ., reduce: FullSrec
			nil,         // !
			nil,         // ~
			reduce(109), // **, reduce: FullSrec
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(109), // [, reduce: FullSrec
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(109), // [[, reduce: FullSrec
			reduce(109), // [[[, reduce: FullSrec
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S195
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(212), // >, reduce: MlrvalOrFunction
			reduce(212), // >>, reduce: MlrvalOrFunction
			reduce(212), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(212), // ?, reduce: MlrvalOrFunction
			reduce(212), // :, reduce: MlrvalOrFunction
			reduce(212), // ||, reduce: MlrvalOrFunction
			reduce(212), // ^^, reduce: MlrvalOrFunction
			reduce(212), // &&, reduce: MlrvalOrFunction
			reduce(212), // ??, reduce: MlrvalOrFunction
			reduce(212), // ???, reduce: MlrvalOrFunction
			reduce(212), // =~, reduce: MlrvalOrFunction
			reduce(212), // !=~, reduce: MlrvalOrFunction
			reduce(212), // ==, reduce: MlrvalOrFunction
			reduce(212), // !=, reduce: MlrvalOrFunction
			reduce(212), // <=>, reduce: MlrvalOrFunction
			reduce(212), // >=, reduce: MlrvalOrFunction
			reduce(212), // <, reduce: MlrvalOrFunction
			reduce(212), // <=, reduce: MlrvalOrFunction
			reduce(212), // ^, reduce: MlrvalOrFunction
			reduce(212), // &, reduce: MlrvalOrFunction
			reduce(212), // <<, reduce: MlrvalOrFunction
			reduce(212), // >>>, reduce: MlrvalOrFunction
			reduce(212), // +, reduce: MlrvalOrFunction
			reduce(212), // -, reduce: MlrvalOrFunction
			reduce(212), // .+, reduce: MlrvalOrFunction
			reduce(212), // .-, reduce: MlrvalOrFunction
			reduce(212), // *, reduce: MlrvalOrFunction
			reduce(212), // /, reduce: MlrvalOrFunction
			reduce(212), // //, reduce: MlrvalOrFunction
			reduce(212), // %, reduce: MlrvalOrFunction
			reduce(212), // .*, reduce: MlrvalOrFunction
			reduce(212), // ./, reduce: MlrvalOrFunction
			reduce(212), // .//, reduce: MlrvalOrFunction
			reduce(212), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(212), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(212), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(212), // [[, reduce: MlrvalOrFunction
			reduce(212), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S196
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(113), // >, reduce: DirectOosvarValue
			reduce(113), // >>, reduce: DirectOosvarValue
			reduce(113), // |, reduce: DirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(113), // ?, reduce: DirectOosvarValue
			reduce(113), // :, reduce: DirectOosvarValue
			reduce(113), // ||, reduce: DirectOosvarValue
			reduce(113), // ^^, reduce: DirectOosvarValue
			reduce(113), // &&, reduce: DirectOosvarValue
			reduce(113), // ??, reduce: DirectOosvarValue
			reduce(113), // ???, reduce: DirectOosvarValue
			reduce(113), // =~, reduce: DirectOosvarValue
			reduce(113), // !=~, reduce: DirectOosvarValue
			reduce(113), // ==, reduce: DirectOosvarValue
			reduce(113), // !=, reduce: DirectOosvarValue
			reduce(113), // <=>, reduce: DirectOosvarValue
			reduce(113), // >=, reduce: DirectOosvarValue
			reduce(113), // <, reduce: DirectOosvarValue
			reduce(113), // <=, reduce: DirectOosvarValue
			reduce(113), // ^, reduce: DirectOosvarValue
			reduce(113), // &, reduce: DirectOosvarValue
			reduce(113), // <<, reduce: DirectOosvarValue
			reduce(113), // >>>, reduce: DirectOosvarValue
			reduce(113), // +, reduce: DirectOosvarValue
			reduce(113), // -, reduce: DirectOosvarValue
			reduce(113), // .+, reduce: DirectOosvarValue
			reduce(113), // .-, reduce: DirectOosvarValue
			reduce(113), // *, reduce: DirectOosvarValue
			reduce(113), // /, reduce: DirectOosvarValue
			reduce(113), // //, reduce: DirectOosvarValue
			reduce(113), // %, reduce: DirectOosvarValue
			reduce(113), // .*, reduce: DirectOosvarValue
			reduce(113), // ./, reduce: DirectOosvarValue
			reduce(113), // .//, reduce: DirectOosvarValue
			reduce(113), // ., reduce: DirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(113), // **, reduce: DirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(113), // [, reduce: DirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(113), // [[, reduce: DirectOosvarValue
			reduce(113), // [[[, reduce: DirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S197
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S198
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(115), // >, reduce: BracedOosvarValue
			reduce(115), // >>, reduce: BracedOosvarValue
			reduce(115), // |, reduce: BracedOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(115), // ?, reduce: BracedOosvarValue
			reduce(115), // :, reduce: BracedOosvarValue
			reduce(115), // ||, reduce: BracedOosvarValue
			reduce(115), // ^^, reduce: BracedOosvarValue
			reduce(115), // &&, reduce: BracedOosvarValue
			reduce(115), // ??, reduce: BracedOosvarValue
			reduce(115), // ???, reduce: BracedOosvarValue
			reduce(115), // =~, reduce: BracedOosvarValue
			reduce(115), // !=~, reduce: BracedOosvarValue
			reduce(115), // ==, reduce: BracedOosvarValue
			reduce(115), // !=, reduce: BracedOosvarValue
			reduce(115), // <=>, reduce: BracedOosvarValue
			reduce(115), // >=, reduce: BracedOosvarValue
			reduce(115), // <, reduce: BracedOosvarValue
			reduce(115), // <=, reduce: BracedOosvarValue
			reduce(115), // ^, reduce: BracedOosvarValue
			reduce(115), // &, reduce: BracedOosvarValue
			reduce(115), // <<, reduce: BracedOosvarValue
			reduce(115), // >>>, reduce: BracedOosvarValue
			reduce(115), // +, reduce: BracedOosvarValue
			reduce(115), // -, reduce: BracedOosvarValue
			reduce(115), // .+, reduce: BracedOosvarValue
			reduce(115), // .-, reduce: BracedOosvarValue
			reduce(115), // *, reduce: BracedOosvarValue
			reduce(115), // /, reduce: BracedOosvarValue
			reduce(115), // //, reduce: BracedOosvarValue
			reduce(115), // %, reduce: BracedOosvarValue
			reduce(115), // .*, reduce: BracedOosvarValue
			reduce(115), // ./, reduce: BracedOosvarValue
			reduce(115), // .//, reduce: BracedOosvarValue
			reduce(115), // ., reduce: BracedOosvarValue
			nil,         // !
			nil,         // ~
			reduce(115), // **, reduce: BracedOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(115), // [, reduce: BracedOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(115), // [[, reduce: BracedOosvarValue
			reduce(115), // [[[, reduce: BracedOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S199
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(116), // >, reduce: FullOosvar
			reduce(116), // >>, reduce: FullOosvar
			reduce(116), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(116), // ?, reduce: FullOosvar
			reduce(116), // :, reduce: FullOosvar
			reduce(116), // ||, reduce: FullOosvar
			reduce(116), // ^^, reduce: FullOosvar
			reduce(116), // &&, reduce: FullOosvar
			reduce(116), // ??, reduce: FullOosvar
			reduce(116), // ???, reduce: FullOosvar
			reduce(116), // =~, reduce: FullOosvar
			reduce(116), // !=~, reduce: FullOosvar
			reduce(116), // ==, reduce: FullOosvar
			reduce(116), // !=, reduce: FullOosvar
			reduce(116), // <=>, reduce: FullOosvar
			reduce(116), // >=, reduce: FullOosvar
			reduce(116), // <, reduce: FullOosvar
			reduce(116), // <=, reduce: FullOosvar
			reduce(116), // ^, reduce: FullOosvar
			reduce(116), // &, reduce: FullOosvar
			reduce(116), // <<, reduce: FullOosvar
			reduce(116), // >>>, reduce: FullOosvar
			reduce(116), // +, reduce: FullOosvar
			reduce(116), // -, reduce: FullOosvar
			reduce(116), // .+, reduce: FullOosvar
			reduce(116), // .-, reduce: FullOosvar
			reduce(116), // *, reduce: FullOosvar
			reduce(116), // /, reduce: FullOosvar
			reduce(116), // //, reduce: FullOosvar
			reduce(116), // %, reduce: FullOosvar
			reduce(116), // .*, reduce: FullOosvar
			reduce(116), // ./, reduce: FullOosvar
			reduce(116), // .//, reduce: FullOosvar
			reduce(116), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(116), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(116), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(116), // [[, reduce: FullOosvar
			reduce(116), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S200
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(117), // >, reduce: FullOosvar
			reduce(117), // >>, reduce: FullOosvar
			reduce(117), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(117), // ?, reduce: FullOosvar
			reduce(117), // :, reduce: FullOosvar
			reduce(117), // ||, reduce: FullOosvar
			reduce(117), // ^^, reduce: FullOosvar
			reduce(117), // &&, reduce: FullOosvar
			reduce(117), // ??, reduce: FullOosvar
			reduce(117), // ???, reduce: FullOosvar
			reduce(117), // =~, reduce: FullOosvar
			reduce(117), // !=~, reduce: FullOosvar
			reduce(117), // ==, reduce: FullOosvar
			reduce(117), // !=, reduce: FullOosvar
			reduce(117), // <=>, reduce: FullOosvar
			reduce(117), // >=, reduce: FullOosvar
			reduce(117), // <, reduce: FullOosvar
			reduce(117), // <=, reduce: FullOosvar
			reduce(117), // ^, reduce: FullOosvar
			reduce(117), // &, reduce: FullOosvar
			reduce(117), // <<, reduce: FullOosvar
			reduce(117), // >>>, reduce: FullOosvar
			reduce(117), // +, reduce: FullOosvar
			reduce(117), // -, reduce: FullOosvar
			reduce(117), // .+, reduce: FullOosvar
			reduce(117), // .-, reduce: FullOosvar
			reduce(117), // *, reduce: FullOosvar
			reduce(117), // /, reduce: FullOosvar
			reduce(117), // //, reduce: FullOosvar
			reduce(117), // %, reduce: FullOosvar
			reduce(117), // .*, reduce: FullOosvar
			reduce(117), // ./, reduce: FullOosvar
			reduce(117), // .//, reduce: FullOosvar
			reduce(117), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(117), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(117), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(117), // [[, reduce: FullOosvar
			reduce(117), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S201
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(118), // >, reduce: LocalVariable
			reduce(118), // >>, reduce: LocalVariable
			reduce(118), // |, reduce: LocalVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(270), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(118), // ?, reduce: LocalVariable
			reduce(118), // :, reduce: LocalVariable
			reduce(118), // ||, reduce: LocalVariable
			reduce(118), // ^^, reduce: LocalVariable
			reduce(118), // &&, reduce: LocalVariable
			reduce(118), // ??, reduce: LocalVariable
			reduce(118), // ???, reduce: LocalVariable
			reduce(118), // =~, reduce: LocalVariable
			reduce(118), // !=~, reduce: LocalVariable
			reduce(118), // ==, reduce: LocalVariable
			reduce(118), // !=, reduce: LocalVariable
			reduce(118), // <=>, reduce: LocalVariable
			reduce(118), // >=, reduce: LocalVariable
			reduce(118), // <, reduce: LocalVariable
			reduce(118), // <=, reduce: LocalVariable
			reduce(118), // ^, reduce: LocalVariable
			reduce(118), // &, reduce: LocalVariable
			reduce(118), // <<, reduce: LocalVariable
			reduce(118), // >>>, reduce: LocalVariable
			reduce(118), // +, reduce: LocalVariable
			reduce(118), // -, reduce: LocalVariable
			reduce(118), // .+, reduce: LocalVariable
			reduce(118), // .-, reduce: LocalVariable
			reduce(118), // *, reduce: LocalVariable
			reduce(118), // /, reduce: LocalVariable
			reduce(118), // //, reduce: LocalVariable
			reduce(118), // %, reduce: LocalVariable
			reduce(118), // .*, reduce: LocalVariable
			reduce(118), // ./, reduce: LocalVariable
			reduce(118), // .//, reduce: LocalVariable
			reduce(118), // ., reduce: LocalVariable
			nil,         // !
			nil,         // ~
			reduce(118), // **, reduce: LocalVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(118), // [, reduce: LocalVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(118), // [[, reduce: LocalVariable
			reduce(118), // [[[, reduce: LocalVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S202
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(272), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S203
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(271), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S204
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			reduce(147), // :, reduce: Rvalue
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S205
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			reduce(148), // :, reduce: PrecedenceChainStart
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S206
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			shift(877),  // ?
			reduce(150), // :, reduce: TernaryTerm
			shift(878),  // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S207
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(152), // ?, reduce: LogicalOrTerm
			reduce(152), // :, reduce: LogicalOrTerm
			reduce(152), // ||, reduce: LogicalOrTerm
			shift(879),  // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S208
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(154), // ?, reduce: LogicalXORTerm
			reduce(154), // :, reduce: LogicalXORTerm
			reduce(154), // ||, reduce: LogicalXORTerm
			reduce(154), // ^^, reduce: LogicalXORTerm
			shift(880),  // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S209
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(156), // ?, reduce: LogicalAndTerm
			reduce(156), // :, reduce: LogicalAndTerm
			reduce(156), // ||, reduce: LogicalAndTerm
			reduce(156), // ^^, reduce: LogicalAndTerm
			reduce(156), // &&, reduce: LogicalAndTerm
			shift(881),  // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S210
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(158), // ?, reduce: AbsentCoalesceTerm
			reduce(158), // :, reduce: AbsentCoalesceTerm
			reduce(158), // ||, reduce: AbsentCoalesceTerm
			reduce(158), // ^^, reduce: AbsentCoalesceTerm
			reduce(158), // &&, reduce: AbsentCoalesceTerm
			reduce(158), // ??, reduce: AbsentCoalesceTerm
			shift(882),  // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S211
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(160), // ?, reduce: EmptyCoalesceTerm
			reduce(160), // :, reduce: EmptyCoalesceTerm
			reduce(160), // ||, reduce: EmptyCoalesceTerm
			reduce(160), // ^^, reduce: EmptyCoalesceTerm
			reduce(160), // &&, reduce: EmptyCoalesceTerm
			reduce(160), // ??, reduce: EmptyCoalesceTerm
			reduce(160), // ???, reduce: EmptyCoalesceTerm
			shift(883),  // =~
			shift(884),  // !=~
			shift(885),  // ==
			shift(886),  // !=
			shift(887),  // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S212
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(888),  // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(166), // ?, reduce: EqneTerm
			reduce(166), // :, reduce: EqneTerm
			reduce(166), // ||, reduce: EqneTerm
			reduce(166), // ^^, reduce: EqneTerm
			reduce(166), // &&, reduce: EqneTerm
			reduce(166), // ??, reduce: EqneTerm
			reduce(166), // ???, reduce: EqneTerm
			reduce(166), // =~, reduce: EqneTerm
			reduce(166), // !=~, reduce: EqneTerm
			reduce(166), // ==, reduce: EqneTerm
			reduce(166), // !=, reduce: EqneTerm
			reduce(166), // <=>, reduce: EqneTerm
			shift(889),  // >=
			shift(890),  // <
			shift(891),  // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S213
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(171), // >, reduce: CmpTerm
			nil,         // >>
			shift(892),  // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(171), // ?, reduce: CmpTerm
			reduce(171), // :, reduce: CmpTerm
			reduce(171), // ||, reduce: CmpTerm
			reduce(171), // ^^, reduce: CmpTerm
			reduce(171), // &&, reduce: CmpTerm
			reduce(171), // ??, reduce: CmpTerm
			reduce(171), // ???, reduce: CmpTerm
			reduce(171), // =~, reduce: CmpTerm
			reduce(171), // !=~, reduce: CmpTerm
			reduce(171), // ==, reduce: CmpTerm
			reduce(171), // !=, reduce: CmpTerm
			reduce(171), // <=>, reduce: CmpTerm
			reduce(171), // >=, reduce: CmpTerm
			reduce(171), // <, reduce: CmpTerm
			reduce(171), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S214
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(173), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(173), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(173), // ?, reduce: BitwiseORTerm
			reduce(173), // :, reduce: BitwiseORTerm
			reduce(173), // ||, reduce: BitwiseORTerm
			reduce(173), // ^^, reduce: BitwiseORTerm
			reduce(173), // &&, reduce: BitwiseORTerm
			reduce(173), // ??, reduce: BitwiseORTerm
			reduce(173), // ???, reduce: BitwiseORTerm
			reduce(173), // =~, reduce: BitwiseORTerm
			reduce(173), // !=~, reduce: BitwiseORTerm
			reduce(173), // ==, reduce: BitwiseORTerm
			reduce(173), // !=, reduce: BitwiseORTerm
			reduce(173), // <=>, reduce: BitwiseORTerm
			reduce(173), // >=, reduce: BitwiseORTerm
			reduce(173), // <, reduce: BitwiseORTerm
			reduce(173), // <=, reduce: BitwiseORTerm
			shift(893),  // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S215
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(175), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(175), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(175), // ?, reduce: BitwiseXORTerm
			reduce(175), // :, reduce: BitwiseXORTerm
			reduce(175), // ||, reduce: BitwiseXORTerm
			reduce(175), // ^^, reduce: BitwiseXORTerm
			reduce(175), // &&, reduce: BitwiseXORTerm
			reduce(175), // ??, reduce: BitwiseXORTerm
			reduce(175), // ???, reduce: BitwiseXORTerm
			reduce(175), // =~, reduce: BitwiseXORTerm
			reduce(175), // !=~, reduce: BitwiseXORTerm
			reduce(175), // ==, reduce: BitwiseXORTerm
			reduce(175), // !=, reduce: BitwiseXORTerm
			reduce(175), // <=>, reduce: BitwiseXORTerm
			reduce(175), // >=, reduce: BitwiseXORTerm
			reduce(175), // <, reduce: BitwiseXORTerm
			reduce(175), // <=, reduce: BitwiseXORTerm
			reduce(175), // ^, reduce: BitwiseXORTerm
			shift(894),  // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S216
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(177), // >, reduce: BitwiseANDTerm
			shift(895),  // >>
			reduce(177), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(177), // ?, reduce: BitwiseANDTerm
			reduce(177), // :, reduce: BitwiseANDTerm
			reduce(177), // ||, reduce: BitwiseANDTerm
			reduce(177), // ^^, reduce: BitwiseANDTerm
			reduce(177), // &&, reduce: BitwiseANDTerm
			reduce(177), // ??, reduce: BitwiseANDTerm
			reduce(177), // ???, reduce: BitwiseANDTerm
			reduce(177), // =~, reduce: BitwiseANDTerm
			reduce(177), // !=~, reduce: BitwiseANDTerm
			reduce(177), // ==, reduce: BitwiseANDTerm
			reduce(177), // !=, reduce: BitwiseANDTerm
			reduce(177), // <=>, reduce: BitwiseANDTerm
			reduce(177), // >=, reduce: BitwiseANDTerm
			reduce(177), // <, reduce: BitwiseANDTerm
			reduce(177), // <=, reduce: BitwiseANDTerm
			reduce(177), // ^, reduce: BitwiseANDTerm
			reduce(177), // &, reduce: BitwiseANDTerm
			shift(896),  // <<
			shift(897),  // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S217
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(181), // >, reduce: BitwiseShiftTerm
			reduce(181), // >>, reduce: BitwiseShiftTerm
			reduce(181), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(181), // ?, reduce: BitwiseShiftTerm
			reduce(181), // :, reduce: BitwiseShiftTerm
			reduce(181), // ||, reduce: BitwiseShiftTerm
			reduce(181), // ^^, reduce: BitwiseShiftTerm
			reduce(181), // &&, reduce: BitwiseShiftTerm
			reduce(181), // ??, reduce: BitwiseShiftTerm
			reduce(181), // ???, reduce: BitwiseShiftTerm
			reduce(181), // =~, reduce: BitwiseShiftTerm
			reduce(181), // !=~, reduce: BitwiseShiftTerm
			reduce(181), // ==, reduce: BitwiseShiftTerm
			reduce(181), // !=, reduce: BitwiseShiftTerm
			reduce(181), // <=>, reduce: BitwiseShiftTerm
			reduce(181), // >=, reduce: BitwiseShiftTerm
			reduce(181), // <, reduce: BitwiseShiftTerm
			reduce(181), // <=, reduce: BitwiseShiftTerm
			reduce(181), // ^, reduce: BitwiseShiftTerm
			reduce(181), // &, reduce: BitwiseShiftTerm
			reduce(181), // <<, reduce: BitwiseShiftTerm
			reduce(181), // >>>, reduce: BitwiseShiftTerm
			shift(898),  // +
			shift(899),  // -
			shift(900),  // .+
			shift(901),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S218
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S219
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(186), // >, reduce: AddsubdotTerm
			reduce(186), // >>, reduce: AddsubdotTerm
			reduce(186), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(186), // ?, reduce: AddsubdotTerm
			reduce(186), // :, reduce: AddsubdotTerm
			reduce(186), // ||, reduce: AddsubdotTerm
			reduce(186), // ^^, reduce: AddsubdotTerm
			reduce(186), // &&, reduce: AddsubdotTerm
			reduce(186), // ??, reduce: AddsubdotTerm
			reduce(186), // ???, reduce: AddsubdotTerm
			reduce(186), // =~, reduce: AddsubdotTerm
			reduce(186), // !=~, reduce: AddsubdotTerm
			reduce(186), // ==, reduce: AddsubdotTerm
			reduce(186), // !=, reduce: AddsubdotTerm
			reduce(186), // <=>, reduce: AddsubdotTerm
			reduce(186), // >=, reduce: AddsubdotTerm
			reduce(186), // <, reduce: AddsubdotTerm
			reduce(186), // <=, reduce: AddsubdotTerm
			reduce(186), // ^, reduce: AddsubdotTerm
			reduce(186), // &, reduce: AddsubdotTerm
			reduce(186), // <<, reduce: AddsubdotTerm
			reduce(186), // >>>, reduce: AddsubdotTerm
			reduce(186), // +, reduce: AddsubdotTerm
			reduce(186), // -, reduce: AddsubdotTerm
			reduce(186), // .+, reduce: AddsubdotTerm
			reduce(186), // .-, reduce: AddsubdotTerm
			shift(903),  // *
			shift(904),  // /
			shift(905),  // //
			shift(906),  // %
			shift(907),  // .*
			shift(908),  // ./
			shift(909),  // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S220
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S221
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S222
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S223
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(194), // >, reduce: MuldivTerm
			reduce(194), // >>, reduce: MuldivTerm
			reduce(194), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(194), // ?, reduce: MuldivTerm
			reduce(194), // :, reduce: MuldivTerm
			reduce(194), // ||, reduce: MuldivTerm
			reduce(194), // ^^, reduce: MuldivTerm
			reduce(194), // &&, reduce: MuldivTerm
			reduce(194), // ??, reduce: MuldivTerm
			reduce(194), // ???, reduce: MuldivTerm
			reduce(194), // =~, reduce: MuldivTerm
			reduce(194), // !=~, reduce: MuldivTerm
			reduce(194), // ==, reduce: MuldivTerm
			reduce(194), // !=, reduce: MuldivTerm
			reduce(194), // <=>, reduce: MuldivTerm
			reduce(194), // >=, reduce: MuldivTerm
			reduce(194), // <, reduce: MuldivTerm
			reduce(194), // <=, reduce: MuldivTerm
			reduce(194), // ^, reduce: MuldivTerm
			reduce(194), // &, reduce: MuldivTerm
			reduce(194), // <<, reduce: MuldivTerm
			reduce(194), // >>>, reduce: MuldivTerm
			reduce(194), // +, reduce: MuldivTerm
			reduce(194), // -, reduce: MuldivTerm
			reduce(194), // .+, reduce: MuldivTerm
			reduce(194), // .-, reduce: MuldivTerm
			reduce(194), // *, reduce: MuldivTerm
			reduce(194), // /, reduce: MuldivTerm
			reduce(194), // //, reduce: MuldivTerm
			reduce(194), // %, reduce: MuldivTerm
			reduce(194), // .*, reduce: MuldivTerm
			reduce(194), // ./, reduce: MuldivTerm
			reduce(194), // .//, reduce: MuldivTerm
			shift(913),  // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S224
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(196), // >, reduce: DotTerm
			reduce(196), // >>, reduce: DotTerm
			reduce(196), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(196), // ?, reduce: DotTerm
			reduce(196), // :, reduce: DotTerm
			reduce(196), // ||, reduce: DotTerm
			reduce(196), // ^^, reduce: DotTerm
			reduce(196), // &&, reduce: DotTerm
			reduce(196), // ??, reduce: DotTerm
			reduce(196), // ???, reduce: DotTerm
			reduce(196), // =~, reduce: DotTerm
			reduce(196), // !=~, reduce: DotTerm
			reduce(196), // ==, reduce: DotTerm
			reduce(196), // !=, reduce: DotTerm
			reduce(196), // <=>, reduce: DotTerm
			reduce(196), // >=, reduce: DotTerm
			reduce(196), // <, reduce: DotTerm
			reduce(196), // <=, reduce: DotTerm
			reduce(196), // ^, reduce: DotTerm
			reduce(196), // &, reduce: DotTerm
			reduce(196), // <<, reduce: DotTerm
			reduce(196), // >>>, reduce: DotTerm
			reduce(196), // +, reduce: DotTerm
			reduce(196), // -, reduce: DotTerm
			reduce(196), // .+, reduce: DotTerm
			reduce(196), // .-, reduce: DotTerm
			reduce(196), // *, reduce: DotTerm
			reduce(196), // /, reduce: DotTerm
			reduce(196), // //, reduce: DotTerm
			reduce(196), // %, reduce: DotTerm
			reduce(196), // .*, reduce: DotTerm
			reduce(196), // ./, reduce: DotTerm
			reduce(196), // .//, reduce: DotTerm
			reduce(196), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S225
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S226
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S227
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(203), // >, reduce: UnaryOpTerm
			reduce(203), // >>, reduce: UnaryOpTerm
			reduce(203), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(203), // ?, reduce: UnaryOpTerm
			reduce(203), // :, reduce: UnaryOpTerm
			reduce(203), // ||, reduce: UnaryOpTerm
			reduce(203), // ^^, reduce: UnaryOpTerm
			reduce(203), // &&, reduce: UnaryOpTerm
			reduce(203), // ??, reduce: UnaryOpTerm
			reduce(203), // ???, reduce: UnaryOpTerm
			reduce(203), // =~, reduce: UnaryOpTerm
			reduce(203), // !=~, reduce: UnaryOpTerm
			reduce(203), // ==, reduce: UnaryOpTerm
			reduce(203), // !=, reduce: UnaryOpTerm
			reduce(203), // <=>, reduce: UnaryOpTerm
			reduce(203), // >=, reduce: UnaryOpTerm
			reduce(203), // <, reduce: UnaryOpTerm
			reduce(203), // <=, reduce: UnaryOpTerm
			reduce(203), // ^, reduce: UnaryOpTerm
			reduce(203), // &, reduce: UnaryOpTerm
			reduce(203), // <<, reduce: UnaryOpTerm
			reduce(203), // >>>, reduce: UnaryOpTerm
			reduce(203), // +, reduce: UnaryOpTerm
			reduce(203), // -, reduce: UnaryOpTerm
			reduce(203), // .+, reduce: UnaryOpTerm
			reduce(203), // .-, reduce: UnaryOpTerm
			reduce(203), // *, reduce: UnaryOpTerm
			reduce(203), // /, reduce: UnaryOpTerm
			reduce(203), // //, reduce: UnaryOpTerm
			reduce(203), // %, reduce: UnaryOpTerm
			reduce(203), // .*, reduce: UnaryOpTerm
			reduce(203), // ./, reduce: UnaryOpTerm
			reduce(203), // .//, reduce: UnaryOpTerm
			reduce(203), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S228
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(207), // >, reduce: PowTerm
			reduce(207), // >>, reduce: PowTerm
			reduce(207), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(207), // ?, reduce: PowTerm
			reduce(207), // :, reduce: PowTerm
			reduce(207), // ||, reduce: PowTerm
			reduce(207), // ^^, reduce: PowTerm
			reduce(207), // &&, reduce: PowTerm
			reduce(207), // ??, reduce: PowTerm
			reduce(207), // ???, reduce: PowTerm
			reduce(207), // =~, reduce: PowTerm
			reduce(207), // !=~, reduce: PowTerm
			reduce(207), // ==, reduce: PowTerm
			reduce(207), // !=, reduce: PowTerm
			reduce(207), // <=>, reduce: PowTerm
			reduce(207), // >=, reduce: PowTerm
			reduce(207), // <, reduce: PowTerm
			reduce(207), // <=, reduce: PowTerm
			reduce(207), // ^, reduce: PowTerm
			reduce(207), // &, reduce: PowTerm
			reduce(207), // <<, reduce: PowTerm
			reduce(207), // >>>, reduce: PowTerm
			reduce(207), // +, reduce: PowTerm
			reduce(207), // -, reduce: PowTerm
			reduce(207), // .+, reduce: PowTerm
			reduce(207), // .-, reduce: PowTerm
			reduce(207), // *, reduce: PowTerm
			reduce(207), // /, reduce: PowTerm
			reduce(207), // //, reduce: PowTerm
			reduce(207), // %, reduce: PowTerm
			reduce(207), // .*, reduce: PowTerm
			reduce(207), // ./, reduce: PowTerm
			reduce(207), // .//, reduce: PowTerm
			reduce(207), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			shift(916),  // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S229
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(209), // >, reduce: PrecedenceChainEnd
			reduce(209), // >>, reduce: PrecedenceChainEnd
			reduce(209), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(209), // ?, reduce: PrecedenceChainEnd
			reduce(209), // :, reduce: PrecedenceChainEnd
			reduce(209), // ||, reduce: PrecedenceChainEnd
			reduce(209), // ^^, reduce: PrecedenceChainEnd
			reduce(209), // &&, reduce: PrecedenceChainEnd
			reduce(209), // ??, reduce: PrecedenceChainEnd
			reduce(209), // ???, reduce: PrecedenceChainEnd
			reduce(209), // =~, reduce: PrecedenceChainEnd
			reduce(209), // !=~, reduce: PrecedenceChainEnd
			reduce(209), // ==, reduce: PrecedenceChainEnd
			reduce(209), // !=, reduce: PrecedenceChainEnd
			reduce(209), // <=>, reduce: PrecedenceChainEnd
			reduce(209), // >=, reduce: PrecedenceChainEnd
			reduce(209), // <, reduce: PrecedenceChainEnd
			reduce(209), // <=, reduce: PrecedenceChainEnd
			reduce(209), // ^, reduce: PrecedenceChainEnd
			reduce(209), // &, reduce: PrecedenceChainEnd
			reduce(209), // <<, reduce: PrecedenceChainEnd
			reduce(209), // >>>, reduce: PrecedenceChainEnd
			reduce(209), // +, reduce: PrecedenceChainEnd
			reduce(209), // -, reduce: PrecedenceChainEnd
			reduce(209), // .+, reduce: PrecedenceChainEnd
			reduce(209), // .-, reduce: PrecedenceChainEnd
			reduce(209), // *, reduce: PrecedenceChainEnd
			reduce(209), // /, reduce: PrecedenceChainEnd
			reduce(209), // //, reduce: PrecedenceChainEnd
			reduce(209), // %, reduce: PrecedenceChainEnd
			reduce(209), // .*, reduce: PrecedenceChainEnd
			reduce(209), // ./, reduce: PrecedenceChainEnd
			reduce(209), // .//, reduce: PrecedenceChainEnd
			reduce(209), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(209), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(917),  // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			shift(918),  // [[
			shift(919),  // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S230
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(215), // >, reduce: MlrvalOrFunction
			reduce(215), // >>, reduce: MlrvalOrFunction
			reduce(215), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(215), // ?, reduce: MlrvalOrFunction
			reduce(215), // :, reduce: MlrvalOrFunction
			reduce(215), // ||, reduce: MlrvalOrFunction
			reduce(215), // ^^, reduce: MlrvalOrFunction
			reduce(215), // &&, reduce: MlrvalOrFunction
			reduce(215), // ??, reduce: MlrvalOrFunction
			reduce(215), // ???, reduce: MlrvalOrFunction
			reduce(215), // =~, reduce: MlrvalOrFunction
			reduce(215), // !=~, reduce: MlrvalOrFunction
			reduce(215), // ==, reduce: MlrvalOrFunction
			reduce(215), // !=, reduce: MlrvalOrFunction
			reduce(215), // <=>, reduce: MlrvalOrFunction
			reduce(215), // >=, reduce: MlrvalOrFunction
			reduce(215), // <, reduce: MlrvalOrFunction
			reduce(215), // <=, reduce: MlrvalOrFunction
			reduce(215), // ^, reduce: MlrvalOrFunction
			reduce(215), // &, reduce: MlrvalOrFunction
			reduce(215), // <<, reduce: MlrvalOrFunction
			reduce(215), // >>>, reduce: MlrvalOrFunction
			reduce(215), // +, reduce: MlrvalOrFunction
			reduce(215), // -, reduce: MlrvalOrFunction
			reduce(215), // .+, reduce: MlrvalOrFunction
			reduce(215), // .-, reduce: MlrvalOrFunction
			reduce(215), // *, reduce: MlrvalOrFunction
			reduce(215), // /, reduce: MlrvalOrFunction
			reduce(215), // //, reduce: MlrvalOrFunction
			reduce(215), // %, reduce: MlrvalOrFunction
			reduce(215), // .*, reduce: MlrvalOrFunction
			reduce(215), // ./, reduce: MlrvalOrFunction
			reduce(215), // .//, reduce: MlrvalOrFunction
			reduce(215), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(215), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(215), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(215), // [[, reduce: MlrvalOrFunction
			reduce(215), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S231
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(216), // >, reduce: MlrvalOrFunction
			reduce(216), // >>, reduce: MlrvalOrFunction
			reduce(216), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(216), // ?, reduce: MlrvalOrFunction
			reduce(216), // :, reduce: MlrvalOrFunction
			reduce(216), // ||, reduce: MlrvalOrFunction
			reduce(216), // ^^, reduce: MlrvalOrFunction
			reduce(216), // &&, reduce: MlrvalOrFunction
			reduce(216), // ??, reduce: MlrvalOrFunction
			reduce(216), // ???, reduce: MlrvalOrFunction
			reduce(216), // =~, reduce: MlrvalOrFunction
			reduce(216), // !=~, reduce: MlrvalOrFunction
			reduce(216), // ==, reduce: MlrvalOrFunction
			reduce(216), // !=, reduce: MlrvalOrFunction
			reduce(216), // <=>, reduce: MlrvalOrFunction
			reduce(216), // >=, reduce: MlrvalOrFunction
			reduce(216), // <, reduce: MlrvalOrFunction
			reduce(216), // <=, reduce: MlrvalOrFunction
			reduce(216), // ^, reduce: MlrvalOrFunction
			reduce(216), // &, reduce: MlrvalOrFunction
			reduce(216), // <<, reduce: MlrvalOrFunction
			reduce(216), // >>>, reduce: MlrvalOrFunction
			reduce(216), // +, reduce: MlrvalOrFunction
			reduce(216), // -, reduce: MlrvalOrFunction
			reduce(216), // .+, reduce: MlrvalOrFunction
			reduce(216), // .-, reduce: MlrvalOrFunction
			reduce(216), // *, reduce: MlrvalOrFunction
			reduce(216), // /, reduce: MlrvalOrFunction
			reduce(216), // //, reduce: MlrvalOrFunction
			reduce(216), // %, reduce: MlrvalOrFunction
			reduce(216), // .*, reduce: MlrvalOrFunction
			reduce(216), // ./, reduce: MlrvalOrFunction
			reduce(216), // .//, reduce: MlrvalOrFunction
			reduce(216), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(216), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(216), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(216), // [[, reduce: MlrvalOrFunction
			reduce(216), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S232
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(217), // >, reduce: MlrvalOrFunction
			reduce(217), // >>, reduce: MlrvalOrFunction
			reduce(217), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(217), // ?, reduce: MlrvalOrFunction
			reduce(217), // :, reduce: MlrvalOrFunction
			reduce(217), // ||, reduce: MlrvalOrFunction
			reduce(217), // ^^, reduce: MlrvalOrFunction
			reduce(217), // &&, reduce: MlrvalOrFunction
			reduce(217), // ??, reduce: MlrvalOrFunction
			reduce(217), // ???, reduce: MlrvalOrFunction
			reduce(217), // =~, reduce: MlrvalOrFunction
			reduce(217), // !=~, reduce: MlrvalOrFunction
			reduce(217), // ==, reduce: MlrvalOrFunction
			reduce(217), // !=, reduce: MlrvalOrFunction
			reduce(217), // <=>, reduce: MlrvalOrFunction
			reduce(217), // >=, reduce: MlrvalOrFunction
			reduce(217), // <, reduce: MlrvalOrFunction
			reduce(217), // <=, reduce: MlrvalOrFunction
			reduce(217), // ^, reduce: MlrvalOrFunction
			reduce(217), // &, reduce: MlrvalOrFunction
			reduce(217), // <<, reduce: MlrvalOrFunction
			reduce(217), // >>>, reduce: MlrvalOrFunction
			reduce(217), // +, reduce: MlrvalOrFunction
			reduce(217), // -, reduce: MlrvalOrFunction
			reduce(217), // .+, reduce: MlrvalOrFunction
			reduce(217), // .-, reduce: MlrvalOrFunction
			reduce(217), // *, reduce: MlrvalOrFunction
			reduce(217), // /, reduce: MlrvalOrFunction
			reduce(217), // //, reduce: MlrvalOrFunction
			reduce(217), // %, reduce: MlrvalOrFunction
			reduce(217), // .*, reduce: MlrvalOrFunction
			reduce(217), // ./, reduce: MlrvalOrFunction
			reduce(217), // .//, reduce: MlrvalOrFunction
			reduce(217), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(217), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(217), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(217), // [[, reduce: MlrvalOrFunction
			reduce(217), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S233
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(218), // >, reduce: MlrvalOrFunction
			reduce(218), // >>, reduce: MlrvalOrFunction
			reduce(218), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(218), // ?, reduce: MlrvalOrFunction
			reduce(218), // :, reduce: MlrvalOrFunction
			reduce(218), // ||, reduce: MlrvalOrFunction
			reduce(218), // ^^, reduce: MlrvalOrFunction
			reduce(218), // &&, reduce: MlrvalOrFunction
			reduce(218), // ??, reduce: MlrvalOrFunction
			reduce(218), // ???, reduce: MlrvalOrFunction
			reduce(218), // =~, reduce: MlrvalOrFunction
			reduce(218), // !=~, reduce: MlrvalOrFunction
			reduce(218), // ==, reduce: MlrvalOrFunction
			reduce(218), // !=, reduce: MlrvalOrFunction
			reduce(218), // <=>, reduce: MlrvalOrFunction
			reduce(218), // >=, reduce: MlrvalOrFunction
			reduce(218), // <, reduce: MlrvalOrFunction
			reduce(218), // <=, reduce: MlrvalOrFunction
			reduce(218), // ^, reduce: MlrvalOrFunction
			reduce(218), // &, reduce: MlrvalOrFunction
			reduce(218), // <<, reduce: MlrvalOrFunction
			reduce(218), // >>>, reduce: MlrvalOrFunction
			reduce(218), // +, reduce: MlrvalOrFunction
			reduce(218), // -, reduce: MlrvalOrFunction
			reduce(218), // .+, reduce: MlrvalOrFunction
			reduce(218), // .-, reduce: MlrvalOrFunction
			reduce(218), // *, reduce: MlrvalOrFunction
			reduce(218), // /, reduce: MlrvalOrFunction
			reduce(218), // //, reduce: MlrvalOrFunction
			reduce(218), // %, reduce: MlrvalOrFunction
			reduce(218), // .*, reduce: MlrvalOrFunction
			reduce(218), // ./, reduce: MlrvalOrFunction
			reduce(218), // .//, reduce: MlrvalOrFunction
			reduce(218), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(218), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(218), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(218), // [[, reduce: MlrvalOrFunction
			reduce(218), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S234
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(219), // >, reduce: MlrvalOrFunction
			reduce(219), // >>, reduce: MlrvalOrFunction
			reduce(219), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(219), // ?, reduce: MlrvalOrFunction
			reduce(219), // :, reduce: MlrvalOrFunction
			reduce(219), // ||, reduce: MlrvalOrFunction
			reduce(219), // ^^, reduce: MlrvalOrFunction
			reduce(219), // &&, reduce: MlrvalOrFunction
			reduce(219), // ??, reduce: MlrvalOrFunction
			reduce(219), // ???, reduce: MlrvalOrFunction
			reduce(219), // =~, reduce: MlrvalOrFunction
			reduce(219), // !=~, reduce: MlrvalOrFunction
			reduce(219), // ==, reduce: MlrvalOrFunction
			reduce(219), // !=, reduce: MlrvalOrFunction
			reduce(219), // <=>, reduce: MlrvalOrFunction
			reduce(219), // >=, reduce: MlrvalOrFunction
			reduce(219), // <, reduce: MlrvalOrFunction
			reduce(219), // <=, reduce: MlrvalOrFunction
			reduce(219), // ^, reduce: MlrvalOrFunction
			reduce(219), // &, reduce: MlrvalOrFunction
			reduce(219), // <<, reduce: MlrvalOrFunction
			reduce(219), // >>>, reduce: MlrvalOrFunction
			reduce(219), // +, reduce: MlrvalOrFunction
			reduce(219), // -, reduce: MlrvalOrFunction
			reduce(219), // .+, reduce: MlrvalOrFunction
			reduce(219), // .-, reduce: MlrvalOrFunction
			reduce(219), // *, reduce: MlrvalOrFunction
			reduce(219), // /, reduce: MlrvalOrFunction
			reduce(219), // //, reduce: MlrvalOrFunction
			reduce(219), // %, reduce: MlrvalOrFunction
			reduce(219), // .*, reduce: MlrvalOrFunction
			reduce(219), // ./, reduce: MlrvalOrFunction
			reduce(219), // .//, reduce: MlrvalOrFunction
			reduce(219), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(219), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(219), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(219), // [[, reduce: MlrvalOrFunction
			reduce(219), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S235
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(220), // >, reduce: MlrvalOrFunction
			reduce(220), // >>, reduce: MlrvalOrFunction
			reduce(220), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(220), // ?, reduce: MlrvalOrFunction
			reduce(220), // :, reduce: MlrvalOrFunction
			reduce(220), // ||, reduce: MlrvalOrFunction
			reduce(220), // ^^, reduce: MlrvalOrFunction
			reduce(220), // &&, reduce: MlrvalOrFunction
			reduce(220), // ??, reduce: MlrvalOrFunction
			reduce(220), // ???, reduce: MlrvalOrFunction
			reduce(220), // =~, reduce: MlrvalOrFunction
			reduce(220), // !=~, reduce: MlrvalOrFunction
			reduce(220), // ==, reduce: MlrvalOrFunction
			reduce(220), // !=, reduce: MlrvalOrFunction
			reduce(220), // <=>, reduce: MlrvalOrFunction
			reduce(220), // >=, reduce: MlrvalOrFunction
			reduce(220), // <, reduce: MlrvalOrFunction
			reduce(220), // <=, reduce: MlrvalOrFunction
			reduce(220), // ^, reduce: MlrvalOrFunction
			reduce(220), // &, reduce: MlrvalOrFunction
			reduce(220), // <<, reduce: MlrvalOrFunction
			reduce(220), // >>>, reduce: MlrvalOrFunction
			reduce(220), // +, reduce: MlrvalOrFunction
			reduce(220), // -, reduce: MlrvalOrFunction
			reduce(220), // .+, reduce: MlrvalOrFunction
			reduce(220), // .-, reduce: MlrvalOrFunction
			reduce(220), // *, reduce: MlrvalOrFunction
			reduce(220), // /, reduce: MlrvalOrFunction
			reduce(220), // //, reduce: MlrvalOrFunction
			reduce(220), // %, reduce: MlrvalOrFunction
			reduce(220), // .*, reduce: MlrvalOrFunction
			reduce(220), // ./, reduce: MlrvalOrFunction
			reduce(220), // .//, reduce: MlrvalOrFunction
			reduce(220), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(220), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(220), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(220), // [[, reduce: MlrvalOrFunction
			reduce(220), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S236
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(221), // >, reduce: MlrvalOrFunction
			reduce(221), // >>, reduce: MlrvalOrFunction
			reduce(221), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(221), // ?, reduce: MlrvalOrFunction
			reduce(221), // :, reduce: MlrvalOrFunction
			reduce(221), // ||, reduce: MlrvalOrFunction
			reduce(221), // ^^, reduce: MlrvalOrFunction
			reduce(221), // &&, reduce: MlrvalOrFunction
			reduce(221), // ??, reduce: MlrvalOrFunction
			reduce(221), // ???, reduce: MlrvalOrFunction
			reduce(221), // =~, reduce: MlrvalOrFunction
			reduce(221), // !=~, reduce: MlrvalOrFunction
			reduce(221), // ==, reduce: MlrvalOrFunction
			reduce(221), // !=, reduce: MlrvalOrFunction
			reduce(221), // <=>, reduce: MlrvalOrFunction
			reduce(221), // >=, reduce: MlrvalOrFunction
			reduce(221), // <, reduce: MlrvalOrFunction
			reduce(221), // <=, reduce: MlrvalOrFunction
			reduce(221), // ^, reduce: MlrvalOrFunction
			reduce(221), // &, reduce: MlrvalOrFunction
			reduce(221), // <<, reduce: MlrvalOrFunction
			reduce(221), // >>>, reduce: MlrvalOrFunction
			reduce(221), // +, reduce: MlrvalOrFunction
			reduce(221), // -, reduce: MlrvalOrFunction
			reduce(221), // .+, reduce: MlrvalOrFunction
			reduce(221), // .-, reduce: MlrvalOrFunction
			reduce(221), // *, reduce: MlrvalOrFunction
			reduce(221), // /, reduce: MlrvalOrFunction
			reduce(221), // //, reduce: MlrvalOrFunction
			reduce(221), // %, reduce: MlrvalOrFunction
			reduce(221), // .*, reduce: MlrvalOrFunction
			reduce(221), // ./, reduce: MlrvalOrFunction
			reduce(221), // .//, reduce: MlrvalOrFunction
			reduce(221), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(221), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(221), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(221), // [[, reduce: MlrvalOrFunction
			reduce(221), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S237
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(222), // >, reduce: MlrvalOrFunction
			reduce(222), // >>, reduce: MlrvalOrFunction
			reduce(222), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(222), // ?, reduce: MlrvalOrFunction
			reduce(222), // :, reduce: MlrvalOrFunction
			reduce(222), // ||, reduce: MlrvalOrFunction
			reduce(222), // ^^, reduce: MlrvalOrFunction
			reduce(222), // &&, reduce: MlrvalOrFunction
			reduce(222), // ??, reduce: MlrvalOrFunction
			reduce(222), // ???, reduce: MlrvalOrFunction
			reduce(222), // =~, reduce: MlrvalOrFunction
			reduce(222), // !=~, reduce: MlrvalOrFunction
			reduce(222), // ==, reduce: MlrvalOrFunction
			reduce(222), // !=, reduce: MlrvalOrFunction
			reduce(222), // <=>, reduce: MlrvalOrFunction
			reduce(222), // >=, reduce: MlrvalOrFunction
			reduce(222), // <, reduce: MlrvalOrFunction
			reduce(222), // <=, reduce: MlrvalOrFunction
			reduce(222), // ^, reduce: MlrvalOrFunction
			reduce(222), // &, reduce: MlrvalOrFunction
			reduce(222), // <<, reduce: MlrvalOrFunction
			reduce(222), // >>>, reduce: MlrvalOrFunction
			reduce(222), // +, reduce: MlrvalOrFunction
			reduce(222), // -, reduce: MlrvalOrFunction
			reduce(222), // .+, reduce: MlrvalOrFunction
			reduce(222), // .-, reduce: MlrvalOrFunction
			reduce(222), // *, reduce: MlrvalOrFunction
			reduce(222), // /, reduce: MlrvalOrFunction
			reduce(222), // //, reduce: MlrvalOrFunction
			reduce(222), // %, reduce: MlrvalOrFunction
			reduce(222), // .*, reduce: MlrvalOrFunction
			reduce(222), // ./, reduce: MlrvalOrFunction
			reduce(222), // .//, reduce: MlrvalOrFunction
			reduce(222), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(222), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(222), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(222), // [[, reduce: MlrvalOrFunction
			reduce(222), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S238
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(223), // >, reduce: MlrvalOrFunction
			reduce(223), // >>, reduce: MlrvalOrFunction
			reduce(223), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(223), // ?, reduce: MlrvalOrFunction
			reduce(223), // :, reduce: MlrvalOrFunction
			reduce(223), // ||, reduce: MlrvalOrFunction
			reduce(223), // ^^, reduce: MlrvalOrFunction
			reduce(223), // &&, reduce: MlrvalOrFunction
			reduce(223), // ??, reduce: MlrvalOrFunction
			reduce(223), // ???, reduce: MlrvalOrFunction
			reduce(223), // =~, reduce: MlrvalOrFunction
			reduce(223), // !=~, reduce: MlrvalOrFunction
			reduce(223), // ==, reduce: MlrvalOrFunction
			reduce(223), // !=, reduce: MlrvalOrFunction
			reduce(223), // <=>, reduce: MlrvalOrFunction
			reduce(223), // >=, reduce: MlrvalOrFunction
			reduce(223), // <, reduce: MlrvalOrFunction
			reduce(223), // <=, reduce: MlrvalOrFunction
			reduce(223), // ^, reduce: MlrvalOrFunction
			reduce(223), // &, reduce: MlrvalOrFunction
			reduce(223), // <<, reduce: MlrvalOrFunction
			reduce(223), // >>>, reduce: MlrvalOrFunction
			reduce(223), // +, reduce: MlrvalOrFunction
			reduce(223), // -, reduce: MlrvalOrFunction
			reduce(223), // .+, reduce: MlrvalOrFunction
			reduce(223), // .-, reduce: MlrvalOrFunction
			reduce(223), // *, reduce: MlrvalOrFunction
			reduce(223), // /, reduce: MlrvalOrFunction
			reduce(223), // //, reduce: MlrvalOrFunction
			reduce(223), // %, reduce: MlrvalOrFunction
			reduce(223), // .*, reduce: MlrvalOrFunction
			reduce(223), // ./, reduce: MlrvalOrFunction
			reduce(223), // .//, reduce: MlrvalOrFunction
			reduce(223), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(223), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(223), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(223), // [[, reduce: MlrvalOrFunction
			reduce(223), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S239
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(224), // >, reduce: MlrvalOrFunction
			reduce(224), // >>, reduce: MlrvalOrFunction
			reduce(224), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(224), // ?, reduce: MlrvalOrFunction
			reduce(224), // :, reduce: MlrvalOrFunction
			reduce(224), // ||, reduce: MlrvalOrFunction
			reduce(224), // ^^, reduce: MlrvalOrFunction
			reduce(224), // &&, reduce: MlrvalOrFunction
			reduce(224), // ??, reduce: MlrvalOrFunction
			reduce(224), // ???, reduce: MlrvalOrFunction
			reduce(224), // =~, reduce: MlrvalOrFunction
			reduce(224), // !=~, reduce: MlrvalOrFunction
			reduce(224), // ==, reduce: MlrvalOrFunction
			reduce(224), // !=, reduce: MlrvalOrFunction
			reduce(224), // <=>, reduce: MlrvalOrFunction
			reduce(224), // >=, reduce: MlrvalOrFunction
			reduce(224), // <, reduce: MlrvalOrFunction
			reduce(224), // <=, reduce: MlrvalOrFunction
			reduce(224), // ^, reduce: MlrvalOrFunction
			reduce(224), // &, reduce: MlrvalOrFunction
			reduce(224), // <<, reduce: MlrvalOrFunction
			reduce(224), // >>>, reduce: MlrvalOrFunction
			reduce(224), // +, reduce: MlrvalOrFunction
			reduce(224), // -, reduce: MlrvalOrFunction
			reduce(224), // .+, reduce: MlrvalOrFunction
			reduce(224), // .-, reduce: MlrvalOrFunction
			reduce(224), // *, reduce: MlrvalOrFunction
			reduce(224), // /, reduce: MlrvalOrFunction
			reduce(224), // //, reduce: MlrvalOrFunction
			reduce(224), // %, reduce: MlrvalOrFunction
			reduce(224), // .*, reduce: MlrvalOrFunction
			reduce(224), // ./, reduce: MlrvalOrFunction
			reduce(224), // .//, reduce: MlrvalOrFunction
			reduce(224), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(224), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(224), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(224), // [[, reduce: MlrvalOrFunction
			reduce(224), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S240
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(225), // >, reduce: MlrvalOrFunction
			reduce(225), // >>, reduce: MlrvalOrFunction
			reduce(225), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(225), // ?, reduce: MlrvalOrFunction
			reduce(225), // :, reduce: MlrvalOrFunction
			reduce(225), // ||, reduce: MlrvalOrFunction
			reduce(225), // ^^, reduce: MlrvalOrFunction
			reduce(225), // &&, reduce: MlrvalOrFunction
			reduce(225), // ??, reduce: MlrvalOrFunction
			reduce(225), // ???, reduce: MlrvalOrFunction
			reduce(225), // =~, reduce: MlrvalOrFunction
			reduce(225), // !=~, reduce: MlrvalOrFunction
			reduce(225), // ==, reduce: MlrvalOrFunction
			reduce(225), // !=, reduce: MlrvalOrFunction
			reduce(225), // <=>, reduce: MlrvalOrFunction
			reduce(225), // >=, reduce: MlrvalOrFunction
			reduce(225), // <, reduce: MlrvalOrFunction
			reduce(225), // <=, reduce: MlrvalOrFunction
			reduce(225), // ^, reduce: MlrvalOrFunction
			reduce(225), // &, reduce: MlrvalOrFunction
			reduce(225), // <<, reduce: MlrvalOrFunction
			reduce(225), // >>>, reduce: MlrvalOrFunction
			reduce(225), // +, reduce: MlrvalOrFunction
			reduce(225), // -, reduce: MlrvalOrFunction
			reduce(225), // .+, reduce: MlrvalOrFunction
			reduce(225), // .-, reduce: MlrvalOrFunction
			reduce(225), // *, reduce: MlrvalOrFunction
			reduce(225), // /, reduce: MlrvalOrFunction
			reduce(225), // //, reduce: MlrvalOrFunction
			reduce(225), // %, reduce: MlrvalOrFunction
			reduce(225), // .*, reduce: MlrvalOrFunction
			reduce(225), // ./, reduce: MlrvalOrFunction
			reduce(225), // .//, reduce: MlrvalOrFunction
			reduce(225), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(225), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(225), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(225), // [[, reduce: MlrvalOrFunction
			reduce(225), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S241
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(226), // >, reduce: MlrvalOrFunction
			reduce(226), // >>, reduce: MlrvalOrFunction
			reduce(226), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(226), // ?, reduce: MlrvalOrFunction
			reduce(226), // :, reduce: MlrvalOrFunction
			reduce(226), // ||, reduce: MlrvalOrFunction
			reduce(226), // ^^, reduce: MlrvalOrFunction
			reduce(226), // &&, reduce: MlrvalOrFunction
			reduce(226), // ??, reduce: MlrvalOrFunction
			reduce(226), // ???, reduce: MlrvalOrFunction
			reduce(226), // =~, reduce: MlrvalOrFunction
			reduce(226), // !=~, reduce: MlrvalOrFunction
			reduce(226), // ==, reduce: MlrvalOrFunction
			reduce(226), // !=, reduce: MlrvalOrFunction
			reduce(226), // <=>, reduce: MlrvalOrFunction
			reduce(226), // >=, reduce: MlrvalOrFunction
			reduce(226), // <, reduce: MlrvalOrFunction
			reduce(226), // <=, reduce: MlrvalOrFunction
			reduce(226), // ^, reduce: MlrvalOrFunction
			reduce(226), // &, reduce: MlrvalOrFunction
			reduce(226), // <<, reduce: MlrvalOrFunction
			reduce(226), // >>>, reduce: MlrvalOrFunction
			reduce(226), // +, reduce: MlrvalOrFunction
			reduce(226), // -, reduce: MlrvalOrFunction
			reduce(226), // .+, reduce: MlrvalOrFunction
			reduce(226), // .-, reduce: MlrvalOrFunction
			reduce(226), // *, reduce: MlrvalOrFunction
			reduce(226), // /, reduce: MlrvalOrFunction
			reduce(226), // //, reduce: MlrvalOrFunction
			reduce(226), // %, reduce: MlrvalOrFunction
			reduce(226), // .*, reduce: MlrvalOrFunction
			reduce(226), // ./, reduce: MlrvalOrFunction
			reduce(226), // .//, reduce: MlrvalOrFunction
			reduce(226), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(226), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(226), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(226), // [[, reduce: MlrvalOrFunction
			reduce(226), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S242
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(227), // >, reduce: MlrvalOrFunction
			reduce(227), // >>, reduce: MlrvalOrFunction
			reduce(227), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(227), // ?, reduce: MlrvalOrFunction
			reduce(227), // :, reduce: MlrvalOrFunction
			reduce(227), // ||, reduce: MlrvalOrFunction
			reduce(227), // ^^, reduce: MlrvalOrFunction
			reduce(227), // &&, reduce: MlrvalOrFunction
			reduce(227), // ??, reduce: MlrvalOrFunction
			reduce(227), // ???, reduce: MlrvalOrFunction
			reduce(227), // =~, reduce: MlrvalOrFunction
			reduce(227), // !=~, reduce: MlrvalOrFunction
			reduce(227), // ==, reduce: MlrvalOrFunction
			reduce(227), // !=, reduce: MlrvalOrFunction
			reduce(227), // <=>, reduce: MlrvalOrFunction
			reduce(227), // >=, reduce: MlrvalOrFunction
			reduce(227), // <, reduce: MlrvalOrFunction
			reduce(227), // <=, reduce: MlrvalOrFunction
			reduce(227), // ^, reduce: MlrvalOrFunction
			reduce(227), // &, reduce: MlrvalOrFunction
			reduce(227), // <<, reduce: MlrvalOrFunction
			reduce(227), // >>>, reduce: MlrvalOrFunction
			reduce(227), // +, reduce: MlrvalOrFunction
			reduce(227), // -, reduce: MlrvalOrFunction
			reduce(227), // .+, reduce: MlrvalOrFunction
			reduce(227), // .-, reduce: MlrvalOrFunction
			reduce(227), // *, reduce: MlrvalOrFunction
			reduce(227), // /, reduce: MlrvalOrFunction
			reduce(227), // //, reduce: MlrvalOrFunction
			reduce(227), // %, reduce: MlrvalOrFunction
			reduce(227), // .*, reduce: MlrvalOrFunction
			reduce(227), // ./, reduce: MlrvalOrFunction
			reduce(227), // .//, reduce: MlrvalOrFunction
			reduce(227), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(227), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(227), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(227), // [[, reduce: MlrvalOrFunction
			reduce(227), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S243
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			shift(920), // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S244
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			shift(922), // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S245
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(236), // }, reduce: MapLiteralKeyValuePairs
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(923),  // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S246
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(240), // >, reduce: MlrvalOrFunction
			reduce(240), // >>, reduce: MlrvalOrFunction
			reduce(240), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(240), // ?, reduce: MlrvalOrFunction
			reduce(240), // :, reduce: MlrvalOrFunction
			reduce(240), // ||, reduce: MlrvalOrFunction
			reduce(240), // ^^, reduce: MlrvalOrFunction
			reduce(240), // &&, reduce: MlrvalOrFunction
			reduce(240), // ??, reduce: MlrvalOrFunction
			reduce(240), // ???, reduce: MlrvalOrFunction
			reduce(240), // =~, reduce: MlrvalOrFunction
			reduce(240), // !=~, reduce: MlrvalOrFunction
			reduce(240), // ==, reduce: MlrvalOrFunction
			reduce(240), // !=, reduce: MlrvalOrFunction
			reduce(240), // <=>, reduce: MlrvalOrFunction
			reduce(240), // >=, reduce: MlrvalOrFunction
			reduce(240), // <, reduce: MlrvalOrFunction
			reduce(240), // <=, reduce: MlrvalOrFunction
			reduce(240), // ^, reduce: MlrvalOrFunction
			reduce(240), // &, reduce: MlrvalOrFunction
			reduce(240), // <<, reduce: MlrvalOrFunction
			reduce(240), // >>>, reduce: MlrvalOrFunction
			reduce(240), // +, reduce: MlrvalOrFunction
			reduce(240), // -, reduce: MlrvalOrFunction
			reduce(240), // .+, reduce: MlrvalOrFunction
			reduce(240), // .-, reduce: MlrvalOrFunction
			reduce(240), // *, reduce: MlrvalOrFunction
			reduce(240), // /, reduce: MlrvalOrFunction
			reduce(240), // //, reduce: MlrvalOrFunction
			reduce(240), // %, reduce: MlrvalOrFunction
			reduce(240), // .*, reduce: MlrvalOrFunction
			reduce(240), // ./, reduce: MlrvalOrFunction
			reduce(240), // .//, reduce: MlrvalOrFunction
			reduce(240), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(240), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(240), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(240), // [[, reduce: MlrvalOrFunction
			reduce(240), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S247
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(241), // >, reduce: ContextVariable
			reduce(241), // >>, reduce: ContextVariable
			reduce(241), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(241), // ?, reduce: ContextVariable
			reduce(241), // :, reduce: ContextVariable
			reduce(241), // ||, reduce: ContextVariable
			reduce(241), // ^^, reduce: ContextVariable
			reduce(241), // &&, reduce: ContextVariable
			reduce(241), // ??, reduce: ContextVariable
			reduce(241), // ???, reduce: ContextVariable
			reduce(241), // =~, reduce: ContextVariable
			reduce(241), // !=~, reduce: ContextVariable
			reduce(241), // ==, reduce: ContextVariable
			reduce(241), // !=, reduce: ContextVariable
			reduce(241), // <=>, reduce: ContextVariable
			reduce(241), // >=, reduce: ContextVariable
			reduce(241), // <, reduce: ContextVariable
			reduce(241), // <=, reduce: ContextVariable
			reduce(241), // ^, reduce: ContextVariable
			reduce(241), // &, reduce: ContextVariable
			reduce(241), // <<, reduce: ContextVariable
			reduce(241), // >>>, reduce: ContextVariable
			reduce(241), // +, reduce: ContextVariable
			reduce(241), // -, reduce: ContextVariable
			reduce(241), // .+, reduce: ContextVariable
			reduce(241), // .-, reduce: ContextVariable
			reduce(241), // *, reduce: ContextVariable
			reduce(241), // /, reduce: ContextVariable
			reduce(241), // //, reduce: ContextVariable
			reduce(241), // %, reduce: ContextVariable
			reduce(241), // .*, reduce: ContextVariable
			reduce(241), // ./, reduce: ContextVariable
			reduce(241), // .//, reduce: ContextVariable
			reduce(241), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(241), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(241), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(241), // [[, reduce: ContextVariable
			reduce(241), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S248
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(242), // >, reduce: ContextVariable
			reduce(242), // >>, reduce: ContextVariable
			reduce(242), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(242), // ?, reduce: ContextVariable
			reduce(242), // :, reduce: ContextVariable
			reduce(242), // ||, reduce: ContextVariable
			reduce(242), // ^^, reduce: ContextVariable
			reduce(242), // &&, reduce: ContextVariable
			reduce(242), // ??, reduce: ContextVariable
			reduce(242), // ???, reduce: ContextVariable
			reduce(242), // =~, reduce: ContextVariable
			reduce(242), // !=~, reduce: ContextVariable
			reduce(242), // ==, reduce: ContextVariable
			reduce(242), // !=, reduce: ContextVariable
			reduce(242), // <=>, reduce: ContextVariable
			reduce(242), // >=, reduce: ContextVariable
			reduce(242), // <, reduce: ContextVariable
			reduce(242), // <=, reduce: ContextVariable
			reduce(242), // ^, reduce: ContextVariable
			reduce(242), // &, reduce: ContextVariable
			reduce(242), // <<, reduce: ContextVariable
			reduce(242), // >>>, reduce: ContextVariable
			reduce(242), // +, reduce: ContextVariable
			reduce(242), // -, reduce: ContextVariable
			reduce(242), // .+, reduce: ContextVariable
			reduce(242), // .-, reduce: ContextVariable
			reduce(242), // *, reduce: ContextVariable
			reduce(242), // /, reduce: ContextVariable
			reduce(242), // //, reduce: ContextVariable
			reduce(242), // %, reduce: ContextVariable
			reduce(242), // .*, reduce: ContextVariable
			reduce(242), // ./, reduce: ContextVariable
			reduce(242), // .//, reduce: ContextVariable
			reduce(242), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(242), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(242), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(242), // [[, reduce: ContextVariable
			reduce(242), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S249
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(243), // >, reduce: ContextVariable
			reduce(243), // >>, reduce: ContextVariable
			reduce(243), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(243), // ?, reduce: ContextVariable
			reduce(243), // :, reduce: ContextVariable
			reduce(243), // ||, reduce: ContextVariable
			reduce(243), // ^^, reduce: ContextVariable
			reduce(243), // &&, reduce: ContextVariable
			reduce(243), // ??, reduce: ContextVariable
			reduce(243), // ???, reduce: ContextVariable
			reduce(243), // =~, reduce: ContextVariable
			reduce(243), // !=~, reduce: ContextVariable
			reduce(243), // ==, reduce: ContextVariable
			reduce(243), // !=, reduce: ContextVariable
			reduce(243), // <=>, reduce: ContextVariable
			reduce(243), // >=, reduce: ContextVariable
			reduce(243), // <, reduce: ContextVariable
			reduce(243), // <=, reduce: ContextVariable
			reduce(243), // ^, reduce: ContextVariable
			reduce(243), // &, reduce: ContextVariable
			reduce(243), // <<, reduce: ContextVariable
			reduce(243), // >>>, reduce: ContextVariable
			reduce(243), // +, reduce: ContextVariable
			reduce(243), // -, reduce: ContextVariable
			reduce(243), // .+, reduce: ContextVariable
			reduce(243), // .-, reduce: ContextVariable
			reduce(243), // *, reduce: ContextVariable
			reduce(243), // /, reduce: ContextVariable
			reduce(243), // //, reduce: ContextVariable
			reduce(243), // %, reduce: ContextVariable
			reduce(243), // .*, reduce: ContextVariable
			reduce(243), // ./, reduce: ContextVariable
			reduce(243), // .//, reduce: ContextVariable
			reduce(243), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(243), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(243), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(243), // [[, reduce: ContextVariable
			reduce(243), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S250
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(244), // >, reduce: ContextVariable
			reduce(244), // >>, reduce: ContextVariable
			reduce(244), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(244), // ?, reduce: ContextVariable
			reduce(244), // :, reduce: ContextVariable
			reduce(244), // ||, reduce: ContextVariable
			reduce(244), // ^^, reduce: ContextVariable
			reduce(244), // &&, reduce: ContextVariable
			reduce(244), // ??, reduce: ContextVariable
			reduce(244), // ???, reduce: ContextVariable
			reduce(244), // =~, reduce: ContextVariable
			reduce(244), // !=~, reduce: ContextVariable
			reduce(244), // ==, reduce: ContextVariable
			reduce(244), // !=, reduce: ContextVariable
			reduce(244), // <=>, reduce: ContextVariable
			reduce(244), // >=, reduce: ContextVariable
			reduce(244), // <, reduce: ContextVariable
			reduce(244), // <=, reduce: ContextVariable
			reduce(244), // ^, reduce: ContextVariable
			reduce(244), // &, reduce: ContextVariable
			reduce(244), // <<, reduce: ContextVariable
			reduce(244), // >>>, reduce: ContextVariable
			reduce(244), // +, reduce: ContextVariable
			reduce(244), // -, reduce: ContextVariable
			reduce(244), // .+, reduce: ContextVariable
			reduce(244), // .-, reduce: ContextVariable
			reduce(244), // *, reduce: ContextVariable
			reduce(244), // /, reduce: ContextVariable
			reduce(244), // //, reduce: ContextVariable
			reduce(244), // %, reduce: ContextVariable
			reduce(244), // .*, reduce: ContextVariable
			reduce(244), // ./, reduce: ContextVariable
			reduce(244), // .//, reduce: ContextVariable
			reduce(244), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(244), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(244), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(244), // [[, reduce: ContextVariable
			reduce(244), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S251
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(245), // >, reduce: ContextVariable
			reduce(245), // >>, reduce: ContextVariable
			reduce(245), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(245), // ?, reduce: ContextVariable
			reduce(245), // :, reduce: ContextVariable
			reduce(245), // ||, reduce: ContextVariable
			reduce(245), // ^^, reduce: ContextVariable
			reduce(245), // &&, reduce: ContextVariable
			reduce(245), // ??, reduce: ContextVariable
			reduce(245), // ???, reduce: ContextVariable
			reduce(245), // =~, reduce: ContextVariable
			reduce(245), // !=~, reduce: ContextVariable
			reduce(245), // ==, reduce: ContextVariable
			reduce(245), // !=, reduce: ContextVariable
			reduce(245), // <=>, reduce: ContextVariable
			reduce(245), // >=, reduce: ContextVariable
			reduce(245), // <, reduce: ContextVariable
			reduce(245), // <=, reduce: ContextVariable
			reduce(245), // ^, reduce: ContextVariable
			reduce(245), // &, reduce: ContextVariable
			reduce(245), // <<, reduce: ContextVariable
			reduce(245), // >>>, reduce: ContextVariable
			reduce(245), // +, reduce: ContextVariable
			reduce(245), // -, reduce: ContextVariable
			reduce(245), // .+, reduce: ContextVariable
			reduce(245), // .-, reduce: ContextVariable
			reduce(245), // *, reduce: ContextVariable
			reduce(245), // /, reduce: ContextVariable
			reduce(245), // //, reduce: ContextVariable
			reduce(245), // %, reduce: ContextVariable
			reduce(245), // .*, reduce: ContextVariable
			reduce(245), // ./, reduce: ContextVariable
			reduce(245), // .//, reduce: ContextVariable
			reduce(245), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(245), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(245), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(245), // [[, reduce: ContextVariable
			reduce(245), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S252
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(246), // >, reduce: ContextVariable
			reduce(246), // >>, reduce: ContextVariable
			reduce(246), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(246), // ?, reduce: ContextVariable
			reduce(246), // :, reduce: ContextVariable
			reduce(246), // ||, reduce: ContextVariable
			reduce(246), // ^^, reduce: ContextVariable
			reduce(246), // &&, reduce: ContextVariable
			reduce(246), // ??, reduce: ContextVariable
			reduce(246), // ???, reduce: ContextVariable
			reduce(246), // =~, reduce: ContextVariable
			reduce(246), // !=~, reduce: ContextVariable
			reduce(246), // ==, reduce: ContextVariable
			reduce(246), // !=, reduce: ContextVariable
			reduce(246), // <=>, reduce: ContextVariable
			reduce(246), // >=, reduce: ContextVariable
			reduce(246), // <, reduce: ContextVariable
			reduce(246), // <=, reduce: ContextVariable
			reduce(246), // ^, reduce: ContextVariable
			reduce(246), // &, reduce: ContextVariable
			reduce(246), // <<, reduce: ContextVariable
			reduce(246), // >>>, reduce: ContextVariable
			reduce(246), // +, reduce: ContextVariable
			reduce(246), // -, reduce: ContextVariable
			reduce(246), // .+, reduce: ContextVariable
			reduce(246), // .-, reduce: ContextVariable
			reduce(246), // *, reduce: ContextVariable
			reduce(246), // /, reduce: ContextVariable
			reduce(246), // //, reduce: ContextVariable
			reduce(246), // %, reduce: ContextVariable
			reduce(246), // .*, reduce: ContextVariable
			reduce(246), // ./, reduce: ContextVariable
			reduce(246), // .//, reduce: ContextVariable
			reduce(246), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(246), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(246), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(246), // [[, reduce: ContextVariable
			reduce(246), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S253
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(247), // >, reduce: ContextVariable
			reduce(247), // >>, reduce: ContextVariable
			reduce(247), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(247), // ?, reduce: ContextVariable
			reduce(247), // :, reduce: ContextVariable
			reduce(247), // ||, reduce: ContextVariable
			reduce(247), // ^^, reduce: ContextVariable
			reduce(247), // &&, reduce: ContextVariable
			reduce(247), // ??, reduce: ContextVariable
			reduce(247), // ???, reduce: ContextVariable
			reduce(247), // =~, reduce: ContextVariable
			reduce(247), // !=~, reduce: ContextVariable
			reduce(247), // ==, reduce: ContextVariable
			reduce(247), // !=, reduce: ContextVariable
			reduce(247), // <=>, reduce: ContextVariable
			reduce(247), // >=, reduce: ContextVariable
			reduce(247), // <, reduce: ContextVariable
			reduce(247), // <=, reduce: ContextVariable
			reduce(247), // ^, reduce: ContextVariable
			reduce(247), // &, reduce: ContextVariable
			reduce(247), // <<, reduce: ContextVariable
			reduce(247), // >>>, reduce: ContextVariable
			reduce(247), // +, reduce: ContextVariable
			reduce(247), // -, reduce: ContextVariable
			reduce(247), // .+, reduce: ContextVariable
			reduce(247), // .-, reduce: ContextVariable
			reduce(247), // *, reduce: ContextVariable
			reduce(247), // /, reduce: ContextVariable
			reduce(247), // //, reduce: ContextVariable
			reduce(247), // %, reduce: ContextVariable
			reduce(247), // .*, reduce: ContextVariable
			reduce(247), // ./, reduce: ContextVariable
			reduce(247), // .//, reduce: ContextVariable
			reduce(247), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(247), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(247), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(247), // [[, reduce: ContextVariable
			reduce(247), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S254
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(248), // >, reduce: ContextVariable
			reduce(248), // >>, reduce: ContextVariable
			reduce(248), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(248), // ?, reduce: ContextVariable
			reduce(248), // :, reduce: ContextVariable
			reduce(248), // ||, reduce: ContextVariable
			reduce(248), // ^^, reduce: ContextVariable
			reduce(248), // &&, reduce: ContextVariable
			reduce(248), // ??, reduce: ContextVariable
			reduce(248), // ???, reduce: ContextVariable
			reduce(248), // =~, reduce: ContextVariable
			reduce(248), // !=~, reduce: ContextVariable
			reduce(248), // ==, reduce: ContextVariable
			reduce(248), // !=, reduce: ContextVariable
			reduce(248), // <=>, reduce: ContextVariable
			reduce(248), // >=, reduce: ContextVariable
			reduce(248), // <, reduce: ContextVariable
			reduce(248), // <=, reduce: ContextVariable
			reduce(248), // ^, reduce: ContextVariable
			reduce(248), // &, reduce: ContextVariable
			reduce(248), // <<, reduce: ContextVariable
			reduce(248), // >>>, reduce: ContextVariable
			reduce(248), // +, reduce: ContextVariable
			reduce(248), // -, reduce: ContextVariable
			reduce(248), // .+, reduce: ContextVariable
			reduce(248), // .-, reduce: ContextVariable
			reduce(248), // *, reduce: ContextVariable
			reduce(248), // /, reduce: ContextVariable
			reduce(248), // //, reduce: ContextVariable
			reduce(248), // %, reduce: ContextVariable
			reduce(248), // .*, reduce: ContextVariable
			reduce(248), // ./, reduce: ContextVariable
			reduce(248), // .//, reduce: ContextVariable
			reduce(248), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(248), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(248), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(248), // [[, reduce: ContextVariable
			reduce(248), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S255
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(249), // >, reduce: ContextVariable
			reduce(249), // >>, reduce: ContextVariable
			reduce(249), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(249), // ?, reduce: ContextVariable
			reduce(249), // :, reduce: ContextVariable
			reduce(249), // ||, reduce: ContextVariable
			reduce(249), // ^^, reduce: ContextVariable
			reduce(249), // &&, reduce: ContextVariable
			reduce(249), // ??, reduce: ContextVariable
			reduce(249), // ???, reduce: ContextVariable
			reduce(249), // =~, reduce: ContextVariable
			reduce(249), // !=~, reduce: ContextVariable
			reduce(249), // ==, reduce: ContextVariable
			reduce(249), // !=, reduce: ContextVariable
			reduce(249), // <=>, reduce: ContextVariable
			reduce(249), // >=, reduce: ContextVariable
			reduce(249), // <, reduce: ContextVariable
			reduce(249), // <=, reduce: ContextVariable
			reduce(249), // ^, reduce: ContextVariable
			reduce(249), // &, reduce: ContextVariable
			reduce(249), // <<, reduce: ContextVariable
			reduce(249), // >>>, reduce: ContextVariable
			reduce(249), // +, reduce: ContextVariable
			reduce(249), // -, reduce: ContextVariable
			reduce(249), // .+, reduce: ContextVariable
			reduce(249), // .-, reduce: ContextVariable
			reduce(249), // *, reduce: ContextVariable
			reduce(249), // /, reduce: ContextVariable
			reduce(249), // //, reduce: ContextVariable
			reduce(249), // %, reduce: ContextVariable
			reduce(249), // .*, reduce: ContextVariable
			reduce(249), // ./, reduce: ContextVariable
			reduce(249), // .//, reduce: ContextVariable
			reduce(249), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(249), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(249), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(249), // [[, reduce: ContextVariable
			reduce(249), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S256
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(250), // >, reduce: ContextVariable
			reduce(250), // >>, reduce: ContextVariable
			reduce(250), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(250), // ?, reduce: ContextVariable
			reduce(250), // :, reduce: ContextVariable
			reduce(250), // ||, reduce: ContextVariable
			reduce(250), // ^^, reduce: ContextVariable
			reduce(250), // &&, reduce: ContextVariable
			reduce(250), // ??, reduce: ContextVariable
			reduce(250), // ???, reduce: ContextVariable
			reduce(250), // =~, reduce: ContextVariable
			reduce(250), // !=~, reduce: ContextVariable
			reduce(250), // ==, reduce: ContextVariable
			reduce(250), // !=, reduce: ContextVariable
			reduce(250), // <=>, reduce: ContextVariable
			reduce(250), // >=, reduce: ContextVariable
			reduce(250), // <, reduce: ContextVariable
			reduce(250), // <=, reduce: ContextVariable
			reduce(250), // ^, reduce: ContextVariable
			reduce(250), // &, reduce: ContextVariable
			reduce(250), // <<, reduce: ContextVariable
			reduce(250), // >>>, reduce: ContextVariable
			reduce(250), // +, reduce: ContextVariable
			reduce(250), // -, reduce: ContextVariable
			reduce(250), // .+, reduce: ContextVariable
			reduce(250), // .-, reduce: ContextVariable
			reduce(250), // *, reduce: ContextVariable
			reduce(250), // /, reduce: ContextVariable
			reduce(250), // //, reduce: ContextVariable
			reduce(250), // %, reduce: ContextVariable
			reduce(250), // .*, reduce: ContextVariable
			reduce(250), // ./, reduce: ContextVariable
			reduce(250), // .//, reduce: ContextVariable
			reduce(250), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(250), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(250), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(250), // [[, reduce: ContextVariable
			reduce(250), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S257
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(251), // >, reduce: ContextVariable
			reduce(251), // >>, reduce: ContextVariable
			reduce(251), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(251), // ?, reduce: ContextVariable
			reduce(251), // :, reduce: ContextVariable
			reduce(251), // ||, reduce: ContextVariable
			reduce(251), // ^^, reduce: ContextVariable
			reduce(251), // &&, reduce: ContextVariable
			reduce(251), // ??, reduce: ContextVariable
			reduce(251), // ???, reduce: ContextVariable
			reduce(251), // =~, reduce: ContextVariable
			reduce(251), // !=~, reduce: ContextVariable
			reduce(251), // ==, reduce: ContextVariable
			reduce(251), // !=, reduce: ContextVariable
			reduce(251), // <=>, reduce: ContextVariable
			reduce(251), // >=, reduce: ContextVariable
			reduce(251), // <, reduce: ContextVariable
			reduce(251), // <=, reduce: ContextVariable
			reduce(251), // ^, reduce: ContextVariable
			reduce(251), // &, reduce: ContextVariable
			reduce(251), // <<, reduce: ContextVariable
			reduce(251), // >>>, reduce: ContextVariable
			reduce(251), // +, reduce: ContextVariable
			reduce(251), // -, reduce: ContextVariable
			reduce(251), // .+, reduce: ContextVariable
			reduce(251), // .-, reduce: ContextVariable
			reduce(251), // *, reduce: ContextVariable
			reduce(251), // /, reduce: ContextVariable
			reduce(251), // //, reduce: ContextVariable
			reduce(251), // %, reduce: ContextVariable
			reduce(251), // .*, reduce: ContextVariable
			reduce(251), // ./, reduce: ContextVariable
			reduce(251), // .//, reduce: ContextVariable
			reduce(251), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(251), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(251), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(251), // [[, reduce: ContextVariable
			reduce(251), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S258
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(252), // >, reduce: ContextVariable
			reduce(252), // >>, reduce: ContextVariable
			reduce(252), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(252), // ?, reduce: ContextVariable
			reduce(252), // :, reduce: ContextVariable
			reduce(252), // ||, reduce: ContextVariable
			reduce(252), // ^^, reduce: ContextVariable
			reduce(252), // &&, reduce: ContextVariable
			reduce(252), // ??, reduce: ContextVariable
			reduce(252), // ???, reduce: ContextVariable
			reduce(252), // =~, reduce: ContextVariable
			reduce(252), // !=~, reduce: ContextVariable
			reduce(252), // ==, reduce: ContextVariable
			reduce(252), // !=, reduce: ContextVariable
			reduce(252), // <=>, reduce: ContextVariable
			reduce(252), // >=, reduce: ContextVariable
			reduce(252), // <, reduce: ContextVariable
			reduce(252), // <=, reduce: ContextVariable
			reduce(252), // ^, reduce: ContextVariable
			reduce(252), // &, reduce: ContextVariable
			reduce(252), // <<, reduce: ContextVariable
			reduce(252), // >>>, reduce: ContextVariable
			reduce(252), // +, reduce: ContextVariable
			reduce(252), // -, reduce: ContextVariable
			reduce(252), // .+, reduce: ContextVariable
			reduce(252), // .-, reduce: ContextVariable
			reduce(252), // *, reduce: ContextVariable
			reduce(252), // /, reduce: ContextVariable
			reduce(252), // //, reduce: ContextVariable
			reduce(252), // %, reduce: ContextVariable
			reduce(252), // .*, reduce: ContextVariable
			reduce(252), // ./, reduce: ContextVariable
			reduce(252), // .//, reduce: ContextVariable
			reduce(252), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(252), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(252), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(252), // [[, reduce: ContextVariable
			reduce(252), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S259
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(253), // >, reduce: MlrvalOrFunction
			reduce(253), // >>, reduce: MlrvalOrFunction
			reduce(253), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(253), // ?, reduce: MlrvalOrFunction
			reduce(253), // :, reduce: MlrvalOrFunction
			reduce(253), // ||, reduce: MlrvalOrFunction
			reduce(253), // ^^, reduce: MlrvalOrFunction
			reduce(253), // &&, reduce: MlrvalOrFunction
			reduce(253), // ??, reduce: MlrvalOrFunction
			reduce(253), // ???, reduce: MlrvalOrFunction
			reduce(253), // =~, reduce: MlrvalOrFunction
			reduce(253), // !=~, reduce: MlrvalOrFunction
			reduce(253), // ==, reduce: MlrvalOrFunction
			reduce(253), // !=, reduce: MlrvalOrFunction
			reduce(253), // <=>, reduce: MlrvalOrFunction
			reduce(253), // >=, reduce: MlrvalOrFunction
			reduce(253), // <, reduce: MlrvalOrFunction
			reduce(253), // <=, reduce: MlrvalOrFunction
			reduce(253), // ^, reduce: MlrvalOrFunction
			reduce(253), // &, reduce: MlrvalOrFunction
			reduce(253), // <<, reduce: MlrvalOrFunction
			reduce(253), // >>>, reduce: MlrvalOrFunction
			reduce(253), // +, reduce: MlrvalOrFunction
			reduce(253), // -, reduce: MlrvalOrFunction
			reduce(253), // .+, reduce: MlrvalOrFunction
			reduce(253), // .-, reduce: MlrvalOrFunction
			reduce(253), // *, reduce: MlrvalOrFunction
			reduce(253), // /, reduce: MlrvalOrFunction
			reduce(253), // //, reduce: MlrvalOrFunction
			reduce(253), // %, reduce: MlrvalOrFunction
			reduce(253), // .*, reduce: MlrvalOrFunction
			reduce(253), // ./, reduce: MlrvalOrFunction
			reduce(253), // .//, reduce: MlrvalOrFunction
			reduce(253), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(253), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(253), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(253), // [[, reduce: MlrvalOrFunction
			reduce(253), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S260
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			shift(924), // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			shift(925), // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S261
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(256), // >, reduce: MlrvalOrFunction
			reduce(256), // >>, reduce: MlrvalOrFunction
			reduce(256), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(256), // ?, reduce: MlrvalOrFunction
			reduce(256), // :, reduce: MlrvalOrFunction
			reduce(256), // ||, reduce: MlrvalOrFunction
			reduce(256), // ^^, reduce: MlrvalOrFunction
			reduce(256), // &&, reduce: MlrvalOrFunction
			reduce(256), // ??, reduce: MlrvalOrFunction
			reduce(256), // ???, reduce: MlrvalOrFunction
			reduce(256), // =~, reduce: MlrvalOrFunction
			reduce(256), // !=~, reduce: MlrvalOrFunction
			reduce(256), // ==, reduce: MlrvalOrFunction
			reduce(256), // !=, reduce: MlrvalOrFunction
			reduce(256), // <=>, reduce: MlrvalOrFunction
			reduce(256), // >=, reduce: MlrvalOrFunction
			reduce(256), // <, reduce: MlrvalOrFunction
			reduce(256), // <=, reduce: MlrvalOrFunction
			reduce(256), // ^, reduce: MlrvalOrFunction
			reduce(256), // &, reduce: MlrvalOrFunction
			reduce(256), // <<, reduce: MlrvalOrFunction
			reduce(256), // >>>, reduce: MlrvalOrFunction
			reduce(256), // +, reduce: MlrvalOrFunction
			reduce(256), // -, reduce: MlrvalOrFunction
			reduce(256), // .+, reduce: MlrvalOrFunction
			reduce(256), // .-, reduce: MlrvalOrFunction
			reduce(256), // *, reduce: MlrvalOrFunction
			reduce(256), // /, reduce: MlrvalOrFunction
			reduce(256), // //, reduce: MlrvalOrFunction
			reduce(256), // %, reduce: MlrvalOrFunction
			reduce(256), // .*, reduce: MlrvalOrFunction
			reduce(256), // ./, reduce: MlrvalOrFunction
			reduce(256), // .//, reduce: MlrvalOrFunction
			reduce(256), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(256), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(256), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(256), // [[, reduce: MlrvalOrFunction
			reduce(256), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S262
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(257), // >, reduce: MlrvalOrFunction
			reduce(257), // >>, reduce: MlrvalOrFunction
			reduce(257), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(257), // ?, reduce: MlrvalOrFunction
			reduce(257), // :, reduce: MlrvalOrFunction
			reduce(257), // ||, reduce: MlrvalOrFunction
			reduce(257), // ^^, reduce: MlrvalOrFunction
			reduce(257), // &&, reduce: MlrvalOrFunction
			reduce(257), // ??, reduce: MlrvalOrFunction
			reduce(257), // ???, reduce: MlrvalOrFunction
			reduce(257), // =~, reduce: MlrvalOrFunction
			reduce(257), // !=~, reduce: MlrvalOrFunction
			reduce(257), // ==, reduce: MlrvalOrFunction
			reduce(257), // !=, reduce: MlrvalOrFunction
			reduce(257), // <=>, reduce: MlrvalOrFunction
			reduce(257), // >=, reduce: MlrvalOrFunction
			reduce(257), // <, reduce: MlrvalOrFunction
			reduce(257), // <=, reduce: MlrvalOrFunction
			reduce(257), // ^, reduce: MlrvalOrFunction
			reduce(257), // &, reduce: MlrvalOrFunction
			reduce(257), // <<, reduce: MlrvalOrFunction
			reduce(257), // >>>, reduce: MlrvalOrFunction
			reduce(257), // +, reduce: MlrvalOrFunction
			reduce(257), // -, reduce: MlrvalOrFunction
			reduce(257), // .+, reduce: MlrvalOrFunction
			reduce(257), // .-, reduce: MlrvalOrFunction
			reduce(257), // *, reduce: MlrvalOrFunction
			reduce(257), // /, reduce: MlrvalOrFunction
			reduce(257), // //, reduce: MlrvalOrFunction
			reduce(257), // %, reduce: MlrvalOrFunction
			reduce(257), // .*, reduce: MlrvalOrFunction
			reduce(257), // ./, reduce: MlrvalOrFunction
			reduce(257), // .//, reduce: MlrvalOrFunction
			reduce(257), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(257), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(257), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(257), // [[, reduce: MlrvalOrFunction
			reduce(257), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S263
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(258), // >, reduce: MlrvalOrFunction
			reduce(258), // >>, reduce: MlrvalOrFunction
			reduce(258), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(258), // ?, reduce: MlrvalOrFunction
			reduce(258), // :, reduce: MlrvalOrFunction
			reduce(258), // ||, reduce: MlrvalOrFunction
			reduce(258), // ^^, reduce: MlrvalOrFunction
			reduce(258), // &&, reduce: MlrvalOrFunction
			reduce(258), // ??, reduce: MlrvalOrFunction
			reduce(258), // ???, reduce: MlrvalOrFunction
			reduce(258), // =~, reduce: MlrvalOrFunction
			reduce(258), // !=~, reduce: MlrvalOrFunction
			reduce(258), // ==, reduce: MlrvalOrFunction
			reduce(258), // !=, reduce: MlrvalOrFunction
			reduce(258), // <=>, reduce: MlrvalOrFunction
			reduce(258), // >=, reduce: MlrvalOrFunction
			reduce(258), // <, reduce: MlrvalOrFunction
			reduce(258), // <=, reduce: MlrvalOrFunction
			reduce(258), // ^, reduce: MlrvalOrFunction
			reduce(258), // &, reduce: MlrvalOrFunction
			reduce(258), // <<, reduce: MlrvalOrFunction
			reduce(258), // >>>, reduce: MlrvalOrFunction
			reduce(258), // +, reduce: MlrvalOrFunction
			reduce(258), // -, reduce: MlrvalOrFunction
			reduce(258), // .+, reduce: MlrvalOrFunction
			reduce(258), // .-, reduce: MlrvalOrFunction
			reduce(258), // *, reduce: MlrvalOrFunction
			reduce(258), // /, reduce: MlrvalOrFunction
			reduce(258), // //, reduce: MlrvalOrFunction
			reduce(258), // %, reduce: MlrvalOrFunction
			reduce(258), // .*, reduce: MlrvalOrFunction
			reduce(258), // ./, reduce: MlrvalOrFunction
			reduce(258), // .//, reduce: MlrvalOrFunction
			reduce(258), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(258), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(258), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(258), // [[, reduce: MlrvalOrFunction
			reduce(258), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S264
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(259), // >, reduce: MlrvalOrFunction
			reduce(259), // >>, reduce: MlrvalOrFunction
			reduce(259), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(259), // ?, reduce: MlrvalOrFunction
			reduce(259), // :, reduce: MlrvalOrFunction
			reduce(259), // ||, reduce: MlrvalOrFunction
			reduce(259), // ^^, reduce: MlrvalOrFunction
			reduce(259), // &&, reduce: MlrvalOrFunction
			reduce(259), // ??, reduce: MlrvalOrFunction
			reduce(259), // ???, reduce: MlrvalOrFunction
			reduce(259), // =~, reduce: MlrvalOrFunction
			reduce(259), // !=~, reduce: MlrvalOrFunction
			reduce(259), // ==, reduce: MlrvalOrFunction
			reduce(259), // !=, reduce: MlrvalOrFunction
			reduce(259), // <=>, reduce: MlrvalOrFunction
			reduce(259), // >=, reduce: MlrvalOrFunction
			reduce(259), // <, reduce: MlrvalOrFunction
			reduce(259), // <=, reduce: MlrvalOrFunction
			reduce(259), // ^, reduce: MlrvalOrFunction
			reduce(259), // &, reduce: MlrvalOrFunction
			reduce(259), // <<, reduce: MlrvalOrFunction
			reduce(259), // >>>, reduce: MlrvalOrFunction
			reduce(259), // +, reduce: MlrvalOrFunction
			reduce(259), // -, reduce: MlrvalOrFunction
			reduce(259), // .+, reduce: MlrvalOrFunction
			reduce(259), // .-, reduce: MlrvalOrFunction
			reduce(259), // *, reduce: MlrvalOrFunction
			reduce(259), // /, reduce: MlrvalOrFunction
			reduce(259), // //, reduce: MlrvalOrFunction
			reduce(259), // %, reduce: MlrvalOrFunction
			reduce(259), // .*, reduce: MlrvalOrFunction
			reduce(259), // ./, reduce: MlrvalOrFunction
			reduce(259), // .//, reduce: MlrvalOrFunction
			reduce(259), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(259), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(259), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(259), // [[, reduce: MlrvalOrFunction
			reduce(259), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S265
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(267), // >, reduce: MlrvalOrFunction
			reduce(267), // >>, reduce: MlrvalOrFunction
			reduce(267), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(267), // ?, reduce: MlrvalOrFunction
			reduce(267), // :, reduce: MlrvalOrFunction
			reduce(267), // ||, reduce: MlrvalOrFunction
			reduce(267), // ^^, reduce: MlrvalOrFunction
			reduce(267), // &&, reduce: MlrvalOrFunction
			reduce(267), // ??, reduce: MlrvalOrFunction
			reduce(267), // ???, reduce: MlrvalOrFunction
			reduce(267), // =~, reduce: MlrvalOrFunction
			reduce(267), // !=~, reduce: MlrvalOrFunction
			reduce(267), // ==, reduce: MlrvalOrFunction
			reduce(267), // !=, reduce: MlrvalOrFunction
			reduce(267), // <=>, reduce: MlrvalOrFunction
			reduce(267), // >=, reduce: MlrvalOrFunction
			reduce(267), // <, reduce: MlrvalOrFunction
			reduce(267), // <=, reduce: MlrvalOrFunction
			reduce(267), // ^, reduce: MlrvalOrFunction
			reduce(267), // &, reduce: MlrvalOrFunction
			reduce(267), // <<, reduce: MlrvalOrFunction
			reduce(267), // >>>, reduce: MlrvalOrFunction
			reduce(267), // +, reduce: MlrvalOrFunction
			reduce(267), // -, reduce: MlrvalOrFunction
			reduce(267), // .+, reduce: MlrvalOrFunction
			reduce(267), // .-, reduce: MlrvalOrFunction
			reduce(267), // *, reduce: MlrvalOrFunction
			reduce(267), // /, reduce: MlrvalOrFunction
			reduce(267), // //, reduce: MlrvalOrFunction
			reduce(267), // %, reduce: MlrvalOrFunction
			reduce(267), // .*, reduce: MlrvalOrFunction
			reduce(267), // ./, reduce: MlrvalOrFunction
			reduce(267), // .//, reduce: MlrvalOrFunction
			reduce(267), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(267), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(267), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(267), // [[, reduce: MlrvalOrFunction
			reduce(267), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S266
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(926), // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S267
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(927), // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S268
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S269
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S270
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S271
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S272
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S273
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S274
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S275
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S276
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S277
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S278
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S279
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S280
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S281
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S282
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S283
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S284
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S285
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S286
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S287
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S288
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(289), // $, reduce: CondBlock
			nil,         // empty
			reduce(289), // ;, reduce: CondBlock
			reduce(289), // {, reduce: CondBlock
			nil,         // }
			nil,         // =
			reduce(289), // unset, reduce: CondBlock
			reduce(289), // filter, reduce: CondBlock
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(289), // print, reduce: CondBlock
			nil,         // ,
			reduce(289), // printn, reduce: CondBlock
			reduce(289), // eprint, reduce: CondBlock
			reduce(289), // eprintn, reduce: CondBlock
			reduce(289), // dump, reduce: CondBlock
			reduce(289), // edump, reduce: CondBlock
			reduce(289), // tee, reduce: CondBlock
			reduce(289), // emitf, reduce: CondBlock
			reduce(289), // emit1, reduce: CondBlock
			reduce(289), // emit, reduce: CondBlock
			reduce(289), // (, reduce: CondBlock
			nil,         // )
			reduce(289), // emitp, reduce: CondBlock
			reduce(289), // field_name, reduce: CondBlock
			reduce(289), // $[, reduce: CondBlock
			nil,         // ]
			reduce(289), // braced_field_name, reduce: CondBlock
			reduce(289), // $[[, reduce: CondBlock
			reduce(289), // $[[[, reduce: CondBlock
			reduce(289), // full_srec, reduce: CondBlock
			reduce(289), // oosvar_name, reduce: CondBlock
			reduce(289), // @[, reduce: CondBlock
			reduce(289), // braced_oosvar_name, reduce: CondBlock
			reduce(289), // full_oosvar, reduce: CondBlock
			reduce(289), // all, reduce: CondBlock
			reduce(289), // non_sigil_name, reduce: CondBlock
			reduce(289), // arr, reduce: CondBlock
			reduce(289), // bool, reduce: CondBlock
			reduce(289), // float, reduce: CondBlock
			reduce(289), // int, reduce: CondBlock
			reduce(289), // map, reduce: CondBlock
			reduce(289), // num, reduce: CondBlock
			reduce(289), // str, reduce: CondBlock
			reduce(289), // var, reduce: CondBlock
			reduce(289), // funct, reduce: CondBlock
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(289), // +, reduce: CondBlock
			reduce(289), // -, reduce: CondBlock
			reduce(289), // .+, reduce: CondBlock
			reduce(289), // .-, reduce: CondBlock
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(289), // !, reduce: CondBlock
			reduce(289), // ~, reduce: CondBlock
			nil,         // **
			reduce(289), // string_literal, reduce: CondBlock
			reduce(289), // regex_case_insensitive, reduce: CondBlock
			reduce(289), // int_literal, reduce: CondBlock
			reduce(289), // float_literal, reduce: CondBlock
			reduce(289), // boolean_literal, reduce: CondBlock
			reduce(289), // null_literal, reduce: CondBlock
			reduce(289), // inf_literal, reduce: CondBlock
			reduce(289), // nan_literal, reduce: CondBlock
			reduce(289), // const_M_PI, reduce: CondBlock
			reduce(289), // const_M_E, reduce: CondBlock
			reduce(289), // panic, reduce: CondBlock
			reduce(289), // [, reduce: CondBlock
			reduce(289), // ctx_IPS, reduce: CondBlock
			reduce(289), // ctx_IFS, reduce: CondBlock
			reduce(289), // ctx_IRS, reduce: CondBlock
			reduce(289), // ctx_OPS, reduce: CondBlock
			reduce(289), // ctx_OFS, reduce: CondBlock
			reduce(289), // ctx_ORS, reduce: CondBlock
			reduce(289), // ctx_FLATSEP, reduce: CondBlock
			reduce(289), // ctx_NF, reduce: CondBlock
			reduce(289), // ctx_NR, reduce: CondBlock
			reduce(289), // ctx_FNR, reduce: CondBlock
			reduce(289), // ctx_FILENAME, reduce: CondBlock
			reduce(289), // ctx_FILENUM, reduce: CondBlock
			reduce(289), // env, reduce: CondBlock
			nil,         // [[
			nil,         // [[[
			reduce(289), // call, reduce: CondBlock
			reduce(289), // begin, reduce: CondBlock
			reduce(289), // end, reduce: CondBlock
			reduce(289), // if, reduce: CondBlock
			nil,         // elif
			nil,         // else
			reduce(289), // while, reduce: CondBlock
			reduce(289), // do, reduce: CondBlock
			reduce(289), // for, reduce: CondBlock
			nil,         // in
			reduce(289), // break, reduce: CondBlock
			reduce(289), // continue, reduce: CondBlock
			reduce(289), // func, reduce: CondBlock
			reduce(289), // subr, reduce: CondBlock
			reduce(289), // return, reduce: CondBlock
		},
	},
	actionRow{ // S289
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S290
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			shift(1103), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S291
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(273), // $, reduce: FcnArgs
			nil,         // empty
			reduce(273), // ;, reduce: FcnArgs
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(1105), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S292
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(35), // $, reduce: Unset
			nil,        // empty
			reduce(35), // ;, reduce: Unset
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S293
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(214), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(214), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(214), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(214), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S294
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(211), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(211), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(211), // >, reduce: MlrvalOrFunction
			reduce(211), // >>, reduce: MlrvalOrFunction
			reduce(211), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(211), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(211), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(211), // ||, reduce: MlrvalOrFunction
			reduce(211), // ^^, reduce: MlrvalOrFunction
			reduce(211), // &&, reduce: MlrvalOrFunction
			reduce(211), // ??, reduce: MlrvalOrFunction
			reduce(211), // ???, reduce: MlrvalOrFunction
			reduce(211), // =~, reduce: MlrvalOrFunction
			reduce(211), // !=~, reduce: MlrvalOrFunction
			reduce(211), // ==, reduce: MlrvalOrFunction
			reduce(211), // !=, reduce: MlrvalOrFunction
			reduce(211), // <=>, reduce: MlrvalOrFunction
			reduce(211), // >=, reduce: MlrvalOrFunction
			reduce(211), // <, reduce: MlrvalOrFunction
			reduce(211), // <=, reduce: MlrvalOrFunction
			reduce(211), // ^, reduce: MlrvalOrFunction
			reduce(211), // &, reduce: MlrvalOrFunction
			reduce(211), // <<, reduce: MlrvalOrFunction
			reduce(211), // >>>, reduce: MlrvalOrFunction
			reduce(211), // +, reduce: MlrvalOrFunction
			reduce(211), // -, reduce: MlrvalOrFunction
			reduce(211), // .+, reduce: MlrvalOrFunction
			reduce(211), // .-, reduce: MlrvalOrFunction
			reduce(211), // *, reduce: MlrvalOrFunction
			reduce(211), // /, reduce: MlrvalOrFunction
			reduce(211), // //, reduce: MlrvalOrFunction
			reduce(211), // %, reduce: MlrvalOrFunction
			reduce(211), // .*, reduce: MlrvalOrFunction
			reduce(211), // ./, reduce: MlrvalOrFunction
			reduce(211), // .//, reduce: MlrvalOrFunction
			reduce(211), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(211), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(211), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(211), // [[, reduce: MlrvalOrFunction
			reduce(211), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S295
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S296
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(110), // $, reduce: OosvarValue
			nil,         // empty
			reduce(110), // ;, reduce: OosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(110), // >, reduce: OosvarValue
			reduce(110), // >>, reduce: OosvarValue
			reduce(110), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(110), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(110), // ?, reduce: OosvarValue
			nil,         // :
			reduce(110), // ||, reduce: OosvarValue
			reduce(110), // ^^, reduce: OosvarValue
			reduce(110), // &&, reduce: OosvarValue
			reduce(110), // ??, reduce: OosvarValue
			reduce(110), // ???, reduce: OosvarValue
			reduce(110), // =~, reduce: OosvarValue
			reduce(110), // !=~, reduce: OosvarValue
			reduce(110), // ==, reduce: OosvarValue
			reduce(110), // !=, reduce: OosvarValue
			reduce(110), // <=>, reduce: OosvarValue
			reduce(110), // >=, reduce: OosvarValue
			reduce(110), // <, reduce: OosvarValue
			reduce(110), // <=, reduce: OosvarValue
			reduce(110), // ^, reduce: OosvarValue
			reduce(110), // &, reduce: OosvarValue
			reduce(110), // <<, reduce: OosvarValue
			reduce(110), // >>>, reduce: OosvarValue
			reduce(110), // +, reduce: OosvarValue
			reduce(110), // -, reduce: OosvarValue
			reduce(110), // .+, reduce: OosvarValue
			reduce(110), // .-, reduce: OosvarValue
			reduce(110), // *, reduce: OosvarValue
			reduce(110), // /, reduce: OosvarValue
			reduce(110), // //, reduce: OosvarValue
			reduce(110), // %, reduce: OosvarValue
			reduce(110), // .*, reduce: OosvarValue
			reduce(110), // ./, reduce: OosvarValue
			reduce(110), // .//, reduce: OosvarValue
			reduce(110), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(110), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(110), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(110), // [[, reduce: OosvarValue
			reduce(110), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S297
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(112), // $, reduce: OosvarValue
			nil,         // empty
			reduce(112), // ;, reduce: OosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(112), // >, reduce: OosvarValue
			reduce(112), // >>, reduce: OosvarValue
			reduce(112), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(112), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(112), // ?, reduce: OosvarValue
			nil,         // :
			reduce(112), // ||, reduce: OosvarValue
			reduce(112), // ^^, reduce: OosvarValue
			reduce(112), // &&, reduce: OosvarValue
			reduce(112), // ??, reduce: OosvarValue
			reduce(112), // ???, reduce: OosvarValue
			reduce(112), // =~, reduce: OosvarValue
			reduce(112), // !=~, reduce: OosvarValue
			reduce(112), // ==, reduce: OosvarValue
			reduce(112), // !=, reduce: OosvarValue
			reduce(112), // <=>, reduce: OosvarValue
			reduce(112), // >=, reduce: OosvarValue
			reduce(112), // <, reduce: OosvarValue
			reduce(112), // <=, reduce: OosvarValue
			reduce(112), // ^, reduce: OosvarValue
			reduce(112), // &, reduce: OosvarValue
			reduce(112), // <<, reduce: OosvarValue
			reduce(112), // >>>, reduce: OosvarValue
			reduce(112), // +, reduce: OosvarValue
			reduce(112), // -, reduce: OosvarValue
			reduce(112), // .+, reduce: OosvarValue
			reduce(112), // .-, reduce: OosvarValue
			reduce(112), // *, reduce: OosvarValue
			reduce(112), // /, reduce: OosvarValue
			reduce(112), // //, reduce: OosvarValue
			reduce(112), // %, reduce: OosvarValue
			reduce(112), // .*, reduce: OosvarValue
			reduce(112), // ./, reduce: OosvarValue
			reduce(112), // .//, reduce: OosvarValue
			reduce(112), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(112), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(112), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(112), // [[, reduce: OosvarValue
			reduce(112), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S298
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(111), // $, reduce: OosvarValue
			nil,         // empty
			reduce(111), // ;, reduce: OosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(111), // >, reduce: OosvarValue
			reduce(111), // >>, reduce: OosvarValue
			reduce(111), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(111), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(111), // ?, reduce: OosvarValue
			nil,         // :
			reduce(111), // ||, reduce: OosvarValue
			reduce(111), // ^^, reduce: OosvarValue
			reduce(111), // &&, reduce: OosvarValue
			reduce(111), // ??, reduce: OosvarValue
			reduce(111), // ???, reduce: OosvarValue
			reduce(111), // =~, reduce: OosvarValue
			reduce(111), // !=~, reduce: OosvarValue
			reduce(111), // ==, reduce: OosvarValue
			reduce(111), // !=, reduce: OosvarValue
			reduce(111), // <=>, reduce: OosvarValue
			reduce(111), // >=, reduce: OosvarValue
			reduce(111), // <, reduce: OosvarValue
			reduce(111), // <=, reduce: OosvarValue
			reduce(111), // ^, reduce: OosvarValue
			reduce(111), // &, reduce: OosvarValue
			reduce(111), // <<, reduce: OosvarValue
			reduce(111), // >>>, reduce: OosvarValue
			reduce(111), // +, reduce: OosvarValue
			reduce(111), // -, reduce: OosvarValue
			reduce(111), // .+, reduce: OosvarValue
			reduce(111), // .-, reduce: OosvarValue
			reduce(111), // *, reduce: OosvarValue
			reduce(111), // /, reduce: OosvarValue
			reduce(111), // //, reduce: OosvarValue
			reduce(111), // %, reduce: OosvarValue
			reduce(111), // .*, reduce: OosvarValue
			reduce(111), // ./, reduce: OosvarValue
			reduce(111), // .//, reduce: OosvarValue
			reduce(111), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(111), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(111), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(111), // [[, reduce: OosvarValue
			reduce(111), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S299
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(99), // $, reduce: FieldValue
			nil,        // empty
			reduce(99), // ;, reduce: FieldValue
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(99), // >, reduce: FieldValue
			reduce(99), // >>, reduce: FieldValue
			reduce(99), // |, reduce: FieldValue
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(99), // ,, reduce: FieldValue
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(99), // ?, reduce: FieldValue
			nil,        // :
			reduce(99), // ||, reduce: FieldValue
			reduce(99), // ^^, reduce: FieldValue
			reduce(99), // &&, reduce: FieldValue
			reduce(99), // ??, reduce: FieldValue
			reduce(99), // ???, reduce: FieldValue
			reduce(99), // =~, reduce: FieldValue
			reduce(99), // !=~, reduce: FieldValue
			reduce(99), // ==, reduce: FieldValue
			reduce(99), // !=, reduce: FieldValue
			reduce(99), // <=>, reduce: FieldValue
			reduce(99), // >=, reduce: FieldValue
			reduce(99), // <, reduce: FieldValue
			reduce(99), // <=, reduce: FieldValue
			reduce(99), // ^, reduce: FieldValue
			reduce(99), // &, reduce: FieldValue
			reduce(99), // <<, reduce: FieldValue
			reduce(99), // >>>, reduce: FieldValue
			reduce(99), // +, reduce: FieldValue
			reduce(99), // -, reduce: FieldValue
			reduce(99), // .+, reduce: FieldValue
			reduce(99), // .-, reduce: FieldValue
			reduce(99), // *, reduce: FieldValue
			reduce(99), // /, reduce: FieldValue
			reduce(99), // //, reduce: FieldValue
			reduce(99), // %, reduce: FieldValue
			reduce(99), // .*, reduce: FieldValue
			reduce(99), // ./, reduce: FieldValue
			reduce(99), // .//, reduce: FieldValue
			reduce(99), // ., reduce: FieldValue
			nil,        // !
			nil,        // ~
			reduce(99), // **, reduce: FieldValue
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(99), // [, reduce: FieldValue
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(99), // [[, reduce: FieldValue
			reduce(99), // [[[, reduce: FieldValue
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S300
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(101), // $, reduce: FieldValue
			nil,         // empty
			reduce(101), // ;, reduce: FieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(101), // >, reduce: FieldValue
			reduce(101), // >>, reduce: FieldValue
			reduce(101), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(101), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(101), // ?, reduce: FieldValue
			nil,         // :
			reduce(101), // ||, reduce: FieldValue
			reduce(101), // ^^, reduce: FieldValue
			reduce(101), // &&, reduce: FieldValue
			reduce(101), // ??, reduce: FieldValue
			reduce(101), // ???, reduce: FieldValue
			reduce(101), // =~, reduce: FieldValue
			reduce(101), // !=~, reduce: FieldValue
			reduce(101), // ==, reduce: FieldValue
			reduce(101), // !=, reduce: FieldValue
			reduce(101), // <=>, reduce: FieldValue
			reduce(101), // >=, reduce: FieldValue
			reduce(101), // <, reduce: FieldValue
			reduce(101), // <=, reduce: FieldValue
			reduce(101), // ^, reduce: FieldValue
			reduce(101), // &, reduce: FieldValue
			reduce(101), // <<, reduce: FieldValue
			reduce(101), // >>>, reduce: FieldValue
			reduce(101), // +, reduce: FieldValue
			reduce(101), // -, reduce: FieldValue
			reduce(101), // .+, reduce: FieldValue
			reduce(101), // .-, reduce: FieldValue
			reduce(101), // *, reduce: FieldValue
			reduce(101), // /, reduce: FieldValue
			reduce(101), // //, reduce: FieldValue
			reduce(101), // %, reduce: FieldValue
			reduce(101), // .*, reduce: FieldValue
			reduce(101), // ./, reduce: FieldValue
			reduce(101), // .//, reduce: FieldValue
			reduce(101), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(101), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(101), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(101), // [[, reduce: FieldValue
			reduce(101), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S301
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(100), // $, reduce: FieldValue
			nil,         // empty
			reduce(100), // ;, reduce: FieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(100), // >, reduce: FieldValue
			reduce(100), // >>, reduce: FieldValue
			reduce(100), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(100), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(100), // ?, reduce: FieldValue
			nil,         // :
			reduce(100), // ||, reduce: FieldValue
			reduce(100), // ^^, reduce: FieldValue
			reduce(100), // &&, reduce: FieldValue
			reduce(100), // ??, reduce: FieldValue
			reduce(100), // ???, reduce: FieldValue
			reduce(100), // =~, reduce: FieldValue
			reduce(100), // !=~, reduce: FieldValue
			reduce(100), // ==, reduce: FieldValue
			reduce(100), // !=, reduce: FieldValue
			reduce(100), // <=>, reduce: FieldValue
			reduce(100), // >=, reduce: FieldValue
			reduce(100), // <, reduce: FieldValue
			reduce(100), // <=, reduce: FieldValue
			reduce(100), // ^, reduce: FieldValue
			reduce(100), // &, reduce: FieldValue
			reduce(100), // <<, reduce: FieldValue
			reduce(100), // >>>, reduce: FieldValue
			reduce(100), // +, reduce: FieldValue
			reduce(100), // -, reduce: FieldValue
			reduce(100), // .+, reduce: FieldValue
			reduce(100), // .-, reduce: FieldValue
			reduce(100), // *, reduce: FieldValue
			reduce(100), // /, reduce: FieldValue
			reduce(100), // //, reduce: FieldValue
			reduce(100), // %, reduce: FieldValue
			reduce(100), // .*, reduce: FieldValue
			reduce(100), // ./, reduce: FieldValue
			reduce(100), // .//, reduce: FieldValue
			reduce(100), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(100), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(100), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(100), // [[, reduce: FieldValue
			reduce(100), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S302
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(213), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(213), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(213), // >, reduce: MlrvalOrFunction
			reduce(213), // >>, reduce: MlrvalOrFunction
			reduce(213), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(213), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(213), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(213), // ||, reduce: MlrvalOrFunction
			reduce(213), // ^^, reduce: MlrvalOrFunction
			reduce(213), // &&, reduce: MlrvalOrFunction
			reduce(213), // ??, reduce: MlrvalOrFunction
			reduce(213), // ???, reduce: MlrvalOrFunction
			reduce(213), // =~, reduce: MlrvalOrFunction
			reduce(213), // !=~, reduce: MlrvalOrFunction
			reduce(213), // ==, reduce: MlrvalOrFunction
			reduce(213), // !=, reduce: MlrvalOrFunction
			reduce(213), // <=>, reduce: MlrvalOrFunction
			reduce(213), // >=, reduce: MlrvalOrFunction
			reduce(213), // <, reduce: MlrvalOrFunction
			reduce(213), // <=, reduce: MlrvalOrFunction
			reduce(213), // ^, reduce: MlrvalOrFunction
			reduce(213), // &, reduce: MlrvalOrFunction
			reduce(213), // <<, reduce: MlrvalOrFunction
			reduce(213), // >>>, reduce: MlrvalOrFunction
			reduce(213), // +, reduce: MlrvalOrFunction
			reduce(213), // -, reduce: MlrvalOrFunction
			reduce(213), // .+, reduce: MlrvalOrFunction
			reduce(213), // .-, reduce: MlrvalOrFunction
			reduce(213), // *, reduce: MlrvalOrFunction
			reduce(213), // /, reduce: MlrvalOrFunction
			reduce(213), // //, reduce: MlrvalOrFunction
			reduce(213), // %, reduce: MlrvalOrFunction
			reduce(213), // .*, reduce: MlrvalOrFunction
			reduce(213), // ./, reduce: MlrvalOrFunction
			reduce(213), // .//, reduce: MlrvalOrFunction
			reduce(213), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(213), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(213), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(213), // [[, reduce: MlrvalOrFunction
			reduce(213), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S303
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(233), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(233), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(233), // >, reduce: MlrvalOrFunction
			reduce(233), // >>, reduce: MlrvalOrFunction
			reduce(233), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(233), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(233), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(233), // ||, reduce: MlrvalOrFunction
			reduce(233), // ^^, reduce: MlrvalOrFunction
			reduce(233), // &&, reduce: MlrvalOrFunction
			reduce(233), // ??, reduce: MlrvalOrFunction
			reduce(233), // ???, reduce: MlrvalOrFunction
			reduce(233), // =~, reduce: MlrvalOrFunction
			reduce(233), // !=~, reduce: MlrvalOrFunction
			reduce(233), // ==, reduce: MlrvalOrFunction
			reduce(233), // !=, reduce: MlrvalOrFunction
			reduce(233), // <=>, reduce: MlrvalOrFunction
			reduce(233), // >=, reduce: MlrvalOrFunction
			reduce(233), // <, reduce: MlrvalOrFunction
			reduce(233), // <=, reduce: MlrvalOrFunction
			reduce(233), // ^, reduce: MlrvalOrFunction
			reduce(233), // &, reduce: MlrvalOrFunction
			reduce(233), // <<, reduce: MlrvalOrFunction
			reduce(233), // >>>, reduce: MlrvalOrFunction
			reduce(233), // +, reduce: MlrvalOrFunction
			reduce(233), // -, reduce: MlrvalOrFunction
			reduce(233), // .+, reduce: MlrvalOrFunction
			reduce(233), // .-, reduce: MlrvalOrFunction
			reduce(233), // *, reduce: MlrvalOrFunction
			reduce(233), // /, reduce: MlrvalOrFunction
			reduce(233), // //, reduce: MlrvalOrFunction
			reduce(233), // %, reduce: MlrvalOrFunction
			reduce(233), // .*, reduce: MlrvalOrFunction
			reduce(233), // ./, reduce: MlrvalOrFunction
			reduce(233), // .//, reduce: MlrvalOrFunction
			reduce(233), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(233), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(233), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(233), // [[, reduce: MlrvalOrFunction
			reduce(233), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S304
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(210), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(210), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(210), // >, reduce: MlrvalOrFunction
			reduce(210), // >>, reduce: MlrvalOrFunction
			reduce(210), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(210), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(210), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(210), // ||, reduce: MlrvalOrFunction
			reduce(210), // ^^, reduce: MlrvalOrFunction
			reduce(210), // &&, reduce: MlrvalOrFunction
			reduce(210), // ??, reduce: MlrvalOrFunction
			reduce(210), // ???, reduce: MlrvalOrFunction
			reduce(210), // =~, reduce: MlrvalOrFunction
			reduce(210), // !=~, reduce: MlrvalOrFunction
			reduce(210), // ==, reduce: MlrvalOrFunction
			reduce(210), // !=, reduce: MlrvalOrFunction
			reduce(210), // <=>, reduce: MlrvalOrFunction
			reduce(210), // >=, reduce: MlrvalOrFunction
			reduce(210), // <, reduce: MlrvalOrFunction
			reduce(210), // <=, reduce: MlrvalOrFunction
			reduce(210), // ^, reduce: MlrvalOrFunction
			reduce(210), // &, reduce: MlrvalOrFunction
			reduce(210), // <<, reduce: MlrvalOrFunction
			reduce(210), // >>>, reduce: MlrvalOrFunction
			reduce(210), // +, reduce: MlrvalOrFunction
			reduce(210), // -, reduce: MlrvalOrFunction
			reduce(210), // .+, reduce: MlrvalOrFunction
			reduce(210), // .-, reduce: MlrvalOrFunction
			reduce(210), // *, reduce: MlrvalOrFunction
			reduce(210), // /, reduce: MlrvalOrFunction
			reduce(210), // //, reduce: MlrvalOrFunction
			reduce(210), // %, reduce: MlrvalOrFunction
			reduce(210), // .*, reduce: MlrvalOrFunction
			reduce(210), // ./, reduce: MlrvalOrFunction
			reduce(210), // .//, reduce: MlrvalOrFunction
			reduce(210), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(210), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(210), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(210), // [[, reduce: MlrvalOrFunction
			reduce(210), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S305
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(102), // $, reduce: FieldValue
			nil,         // empty
			reduce(102), // ;, reduce: FieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(102), // >, reduce: FieldValue
			reduce(102), // >>, reduce: FieldValue
			reduce(102), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(102), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(102), // ?, reduce: FieldValue
			nil,         // :
			reduce(102), // ||, reduce: FieldValue
			reduce(102), // ^^, reduce: FieldValue
			reduce(102), // &&, reduce: FieldValue
			reduce(102), // ??, reduce: FieldValue
			reduce(102), // ???, reduce: FieldValue
			reduce(102), // =~, reduce: FieldValue
			reduce(102), // !=~, reduce: FieldValue
			reduce(102), // ==, reduce: FieldValue
			reduce(102), // !=, reduce: FieldValue
			reduce(102), // <=>, reduce: FieldValue
			reduce(102), // >=, reduce: FieldValue
			reduce(102), // <, reduce: FieldValue
			reduce(102), // <=, reduce: FieldValue
			reduce(102), // ^, reduce: FieldValue
			reduce(102), // &, reduce: FieldValue
			reduce(102), // <<, reduce: FieldValue
			reduce(102), // >>>, reduce: FieldValue
			reduce(102), // +, reduce: FieldValue
			reduce(102), // -, reduce: FieldValue
			reduce(102), // .+, reduce: FieldValue
			reduce(102), // .-, reduce: FieldValue
			reduce(102), // *, reduce: FieldValue
			reduce(102), // /, reduce: FieldValue
			reduce(102), // //, reduce: FieldValue
			reduce(102), // %, reduce: FieldValue
			reduce(102), // .*, reduce: FieldValue
			reduce(102), // ./, reduce: FieldValue
			reduce(102), // .//, reduce: FieldValue
			reduce(102), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(102), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(102), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(102), // [[, reduce: FieldValue
			reduce(102), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S306
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(103), // $, reduce: FieldValue
			nil,         // empty
			reduce(103), // ;, reduce: FieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(103), // >, reduce: FieldValue
			reduce(103), // >>, reduce: FieldValue
			reduce(103), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(103), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(103), // ?, reduce: FieldValue
			nil,         // :
			reduce(103), // ||, reduce: FieldValue
			reduce(103), // ^^, reduce: FieldValue
			reduce(103), // &&, reduce: FieldValue
			reduce(103), // ??, reduce: FieldValue
			reduce(103), // ???, reduce: FieldValue
			reduce(103), // =~, reduce: FieldValue
			reduce(103), // !=~, reduce: FieldValue
			reduce(103), // ==, reduce: FieldValue
			reduce(103), // !=, reduce: FieldValue
			reduce(103), // <=>, reduce: FieldValue
			reduce(103), // >=, reduce: FieldValue
			reduce(103), // <, reduce: FieldValue
			reduce(103), // <=, reduce: FieldValue
			reduce(103), // ^, reduce: FieldValue
			reduce(103), // &, reduce: FieldValue
			reduce(103), // <<, reduce: FieldValue
			reduce(103), // >>>, reduce: FieldValue
			reduce(103), // +, reduce: FieldValue
			reduce(103), // -, reduce: FieldValue
			reduce(103), // .+, reduce: FieldValue
			reduce(103), // .-, reduce: FieldValue
			reduce(103), // *, reduce: FieldValue
			reduce(103), // /, reduce: FieldValue
			reduce(103), // //, reduce: FieldValue
			reduce(103), // %, reduce: FieldValue
			reduce(103), // .*, reduce: FieldValue
			reduce(103), // ./, reduce: FieldValue
			reduce(103), // .//, reduce: FieldValue
			reduce(103), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(103), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(103), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(103), // [[, reduce: FieldValue
			reduce(103), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S307
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(104), // $, reduce: DirectFieldValue
			nil,         // empty
			reduce(104), // ;, reduce: DirectFieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(104), // >, reduce: DirectFieldValue
			reduce(104), // >>, reduce: DirectFieldValue
			reduce(104), // |, reduce: DirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(104), // ,, reduce: DirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(104), // ?, reduce: DirectFieldValue
			nil,         // :
			reduce(104), // ||, reduce: DirectFieldValue
			reduce(104), // ^^, reduce: DirectFieldValue
			reduce(104), // &&, reduce: DirectFieldValue
			reduce(104), // ??, reduce: DirectFieldValue
			reduce(104), // ???, reduce: DirectFieldValue
			reduce(104), // =~, reduce: DirectFieldValue
			reduce(104), // !=~, reduce: DirectFieldValue
			reduce(104), // ==, reduce: DirectFieldValue
			reduce(104), // !=, reduce: DirectFieldValue
			reduce(104), // <=>, reduce: DirectFieldValue
			reduce(104), // >=, reduce: DirectFieldValue
			reduce(104), // <, reduce: DirectFieldValue
			reduce(104), // <=, reduce: DirectFieldValue
			reduce(104), // ^, reduce: DirectFieldValue
			reduce(104), // &, reduce: DirectFieldValue
			reduce(104), // <<, reduce: DirectFieldValue
			reduce(104), // >>>, reduce: DirectFieldValue
			reduce(104), // +, reduce: DirectFieldValue
			reduce(104), // -, reduce: DirectFieldValue
			reduce(104), // .+, reduce: DirectFieldValue
			reduce(104), // .-, reduce: DirectFieldValue
			reduce(104), // *, reduce: DirectFieldValue
			reduce(104), // /, reduce: DirectFieldValue
			reduce(104), // //, reduce: DirectFieldValue
			reduce(104), // %, reduce: DirectFieldValue
			reduce(104), // .*, reduce: DirectFieldValue
			reduce(104), // ./, reduce: DirectFieldValue
			reduce(104), // .//, reduce: DirectFieldValue
			reduce(104), // ., reduce: DirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(104), // **, reduce: DirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(104), // [, reduce: DirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(104), // [[, reduce: DirectFieldValue
			reduce(104), // [[[, reduce: DirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S308
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S309
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(106), // $, reduce: BracedFieldValue
			nil,         // empty
			reduce(106), // ;, reduce: BracedFieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(106), // >, reduce: BracedFieldValue
			reduce(106), // >>, reduce: BracedFieldValue
			reduce(106), // |, reduce: BracedFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(106), // ,, reduce: BracedFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(106), // ?, reduce: BracedFieldValue
			nil,         // :
			reduce(106), // ||, reduce: BracedFieldValue
			reduce(106), // ^^, reduce: BracedFieldValue
			reduce(106), // &&, reduce: BracedFieldValue
			reduce(106), // ??, reduce: BracedFieldValue
			reduce(106), // ???, reduce: BracedFieldValue
			reduce(106), // =~, reduce: BracedFieldValue
			reduce(106), // !=~, reduce: BracedFieldValue
			reduce(106), // ==, reduce: BracedFieldValue
			reduce(106), // !=, reduce: BracedFieldValue
			reduce(106), // <=>, reduce: BracedFieldValue
			reduce(106), // >=, reduce: BracedFieldValue
			reduce(106), // <, reduce: BracedFieldValue
			reduce(106), // <=, reduce: BracedFieldValue
			reduce(106), // ^, reduce: BracedFieldValue
			reduce(106), // &, reduce: BracedFieldValue
			reduce(106), // <<, reduce: BracedFieldValue
			reduce(106), // >>>, reduce: BracedFieldValue
			reduce(106), // +, reduce: BracedFieldValue
			reduce(106), // -, reduce: BracedFieldValue
			reduce(106), // .+, reduce: BracedFieldValue
			reduce(106), // .-, reduce: BracedFieldValue
			reduce(106), // *, reduce: BracedFieldValue
			reduce(106), // /, reduce: BracedFieldValue
			reduce(106), // //, reduce: BracedFieldValue
			reduce(106), // %, reduce: BracedFieldValue
			reduce(106), // .*, reduce: BracedFieldValue
			reduce(106), // ./, reduce: BracedFieldValue
			reduce(106), // .//, reduce: BracedFieldValue
			reduce(106), // ., reduce: BracedFieldValue
			nil,         // !
			nil,         // ~
			reduce(106), // **, reduce: BracedFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(106), // [, reduce: BracedFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(106), // [[, reduce: BracedFieldValue
			reduce(106), // [[[, reduce: BracedFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S310
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S311
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S312
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(109), // $, reduce: FullSrec
			nil,         // empty
			reduce(109), // ;, reduce: FullSrec
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(109), // >, reduce: FullSrec
			reduce(109), // >>, reduce: FullSrec
			reduce(109), // |, reduce: FullSrec
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(109), // ,, reduce: FullSrec
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(109), // ?, reduce: FullSrec
			nil,         // :
			reduce(109), // ||, reduce: FullSrec
			reduce(109), // ^^, reduce: FullSrec
			reduce(109), // &&, reduce: FullSrec
			reduce(109), // ??, reduce: FullSrec
			reduce(109), // ???, reduce: FullSrec
			reduce(109), // =~, reduce: FullSrec
			reduce(109), // !=~, reduce: FullSrec
			reduce(109), // ==, reduce: FullSrec
			reduce(109), // !=, reduce: FullSrec
			reduce(109), // <=>, reduce: FullSrec
			reduce(109), // >=, reduce: FullSrec
			reduce(109), // <, reduce: FullSrec
			reduce(109), // <=, reduce: FullSrec
			reduce(109), // ^, reduce: FullSrec
			reduce(109), // &, reduce: FullSrec
			reduce(109), // <<, reduce: FullSrec
			reduce(109), // >>>, reduce: FullSrec
			reduce(109), // +, reduce: FullSrec
			reduce(109), // -, reduce: FullSrec
			reduce(109), // .+, reduce: FullSrec
			reduce(109), // .-, reduce: FullSrec
			reduce(109), // *, reduce: FullSrec
			reduce(109), // /, reduce: FullSrec
			reduce(109), // //, reduce: FullSrec
			reduce(109), // %, reduce: FullSrec
			reduce(109), // .*, reduce: FullSrec
			reduce(109), // ./, reduce: FullSrec
			reduce(109), // .//, reduce: FullSrec
			reduce(109), // ., reduce: FullSrec
			nil,         // !
			nil,         // ~
			reduce(109), // **, reduce: FullSrec
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(109), // [, reduce: FullSrec
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(109), // [[, reduce: FullSrec
			reduce(109), // [[[, reduce: FullSrec
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S313
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(212), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(212), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(212), // >, reduce: MlrvalOrFunction
			reduce(212), // >>, reduce: MlrvalOrFunction
			reduce(212), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(212), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(212), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(212), // ||, reduce: MlrvalOrFunction
			reduce(212), // ^^, reduce: MlrvalOrFunction
			reduce(212), // &&, reduce: MlrvalOrFunction
			reduce(212), // ??, reduce: MlrvalOrFunction
			reduce(212), // ???, reduce: MlrvalOrFunction
			reduce(212), // =~, reduce: MlrvalOrFunction
			reduce(212), // !=~, reduce: MlrvalOrFunction
			reduce(212), // ==, reduce: MlrvalOrFunction
			reduce(212), // !=, reduce: MlrvalOrFunction
			reduce(212), // <=>, reduce: MlrvalOrFunction
			reduce(212), // >=, reduce: MlrvalOrFunction
			reduce(212), // <, reduce: MlrvalOrFunction
			reduce(212), // <=, reduce: MlrvalOrFunction
			reduce(212), // ^, reduce: MlrvalOrFunction
			reduce(212), // &, reduce: MlrvalOrFunction
			reduce(212), // <<, reduce: MlrvalOrFunction
			reduce(212), // >>>, reduce: MlrvalOrFunction
			reduce(212), // +, reduce: MlrvalOrFunction
			reduce(212), // -, reduce: MlrvalOrFunction
			reduce(212), // .+, reduce: MlrvalOrFunction
			reduce(212), // .-, reduce: MlrvalOrFunction
			reduce(212), // *, reduce: MlrvalOrFunction
			reduce(212), // /, reduce: MlrvalOrFunction
			reduce(212), // //, reduce: MlrvalOrFunction
			reduce(212), // %, reduce: MlrvalOrFunction
			reduce(212), // .*, reduce: MlrvalOrFunction
			reduce(212), // ./, reduce: MlrvalOrFunction
			reduce(212), // .//, reduce: MlrvalOrFunction
			reduce(212), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(212), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(212), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(212), // [[, reduce: MlrvalOrFunction
			reduce(212), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S314
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(113), // $, reduce: DirectOosvarValue
			nil,         // empty
			reduce(113), // ;, reduce: DirectOosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(113), // >, reduce: DirectOosvarValue
			reduce(113), // >>, reduce: DirectOosvarValue
			reduce(113), // |, reduce: DirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(113), // ,, reduce: DirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(113), // ?, reduce: DirectOosvarValue
			nil,         // :
			reduce(113), // ||, reduce: DirectOosvarValue
			reduce(113), // ^^, reduce: DirectOosvarValue
			reduce(113), // &&, reduce: DirectOosvarValue
			reduce(113), // ??, reduce: DirectOosvarValue
			reduce(113), // ???, reduce: DirectOosvarValue
			reduce(113), // =~, reduce: DirectOosvarValue
			reduce(113), // !=~, reduce: DirectOosvarValue
			reduce(113), // ==, reduce: DirectOosvarValue
			reduce(113), // !=, reduce: DirectOosvarValue
			reduce(113), // <=>, reduce: DirectOosvarValue
			reduce(113), // >=, reduce: DirectOosvarValue
			reduce(113), // <, reduce: DirectOosvarValue
			reduce(113), // <=, reduce: DirectOosvarValue
			reduce(113), // ^, reduce: DirectOosvarValue
			reduce(113), // &, reduce: DirectOosvarValue
			reduce(113), // <<, reduce: DirectOosvarValue
			reduce(113), // >>>, reduce: DirectOosvarValue
			reduce(113), // +, reduce: DirectOosvarValue
			reduce(113), // -, reduce: DirectOosvarValue
			reduce(113), // .+, reduce: DirectOosvarValue
			reduce(113), // .-, reduce: DirectOosvarValue
			reduce(113), // *, reduce: DirectOosvarValue
			reduce(113), // /, reduce: DirectOosvarValue
			reduce(113), // //, reduce: DirectOosvarValue
			reduce(113), // %, reduce: DirectOosvarValue
			reduce(113), // .*, reduce: DirectOosvarValue
			reduce(113), // ./, reduce: DirectOosvarValue
			reduce(113), // .//, reduce: DirectOosvarValue
			reduce(113), // ., reduce: DirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(113), // **, reduce: DirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(113), // [, reduce: DirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(113), // [[, reduce: DirectOosvarValue
			reduce(113), // [[[, reduce: DirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S315
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S316
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(115), // $, reduce: BracedOosvarValue
			nil,         // empty
			reduce(115), // ;, reduce: BracedOosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(115), // >, reduce: BracedOosvarValue
			reduce(115), // >>, reduce: BracedOosvarValue
			reduce(115), // |, reduce: BracedOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(115), // ,, reduce: BracedOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(115), // ?, reduce: BracedOosvarValue
			nil,         // :
			reduce(115), // ||, reduce: BracedOosvarValue
			reduce(115), // ^^, reduce: BracedOosvarValue
			reduce(115), // &&, reduce: BracedOosvarValue
			reduce(115), // ??, reduce: BracedOosvarValue
			reduce(115), // ???, reduce: BracedOosvarValue
			reduce(115), // =~, reduce: BracedOosvarValue
			reduce(115), // !=~, reduce: BracedOosvarValue
			reduce(115), // ==, reduce: BracedOosvarValue
			reduce(115), // !=, reduce: BracedOosvarValue
			reduce(115), // <=>, reduce: BracedOosvarValue
			reduce(115), // >=, reduce: BracedOosvarValue
			reduce(115), // <, reduce: BracedOosvarValue
			reduce(115), // <=, reduce: BracedOosvarValue
			reduce(115), // ^, reduce: BracedOosvarValue
			reduce(115), // &, reduce: BracedOosvarValue
			reduce(115), // <<, reduce: BracedOosvarValue
			reduce(115), // >>>, reduce: BracedOosvarValue
			reduce(115), // +, reduce: BracedOosvarValue
			reduce(115), // -, reduce: BracedOosvarValue
			reduce(115), // .+, reduce: BracedOosvarValue
			reduce(115), // .-, reduce: BracedOosvarValue
			reduce(115), // *, reduce: BracedOosvarValue
			reduce(115), // /, reduce: BracedOosvarValue
			reduce(115), // //, reduce: BracedOosvarValue
			reduce(115), // %, reduce: BracedOosvarValue
			reduce(115), // .*, reduce: BracedOosvarValue
			reduce(115), // ./, reduce: BracedOosvarValue
			reduce(115), // .//, reduce: BracedOosvarValue
			reduce(115), // ., reduce: BracedOosvarValue
			nil,         // !
			nil,         // ~
			reduce(115), // **, reduce: BracedOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(115), // [, reduce: BracedOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(115), // [[, reduce: BracedOosvarValue
			reduce(115), // [[[, reduce: BracedOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S317
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(116), // $, reduce: FullOosvar
			nil,         // empty
			reduce(116), // ;, reduce: FullOosvar
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(116), // >, reduce: FullOosvar
			reduce(116), // >>, reduce: FullOosvar
			reduce(116), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(116), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(116), // ?, reduce: FullOosvar
			nil,         // :
			reduce(116), // ||, reduce: FullOosvar
			reduce(116), // ^^, reduce: FullOosvar
			reduce(116), // &&, reduce: FullOosvar
			reduce(116), // ??, reduce: FullOosvar
			reduce(116), // ???, reduce: FullOosvar
			reduce(116), // =~, reduce: FullOosvar
			reduce(116), // !=~, reduce: FullOosvar
			reduce(116), // ==, reduce: FullOosvar
			reduce(116), // !=, reduce: FullOosvar
			reduce(116), // <=>, reduce: FullOosvar
			reduce(116), // >=, reduce: FullOosvar
			reduce(116), // <, reduce: FullOosvar
			reduce(116), // <=, reduce: FullOosvar
			reduce(116), // ^, reduce: FullOosvar
			reduce(116), // &, reduce: FullOosvar
			reduce(116), // <<, reduce: FullOosvar
			reduce(116), // >>>, reduce: FullOosvar
			reduce(116), // +, reduce: FullOosvar
			reduce(116), // -, reduce: FullOosvar
			reduce(116), // .+, reduce: FullOosvar
			reduce(116), // .-, reduce: FullOosvar
			reduce(116), // *, reduce: FullOosvar
			reduce(116), // /, reduce: FullOosvar
			reduce(116), // //, reduce: FullOosvar
			reduce(116), // %, reduce: FullOosvar
			reduce(116), // .*, reduce: FullOosvar
			reduce(116), // ./, reduce: FullOosvar
			reduce(116), // .//, reduce: FullOosvar
			reduce(116), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(116), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(116), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(116), // [[, reduce: FullOosvar
			reduce(116), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S318
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(117), // $, reduce: FullOosvar
			nil,         // empty
			reduce(117), // ;, reduce: FullOosvar
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(117), // >, reduce: FullOosvar
			reduce(117), // >>, reduce: FullOosvar
			reduce(117), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(117), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(117), // ?, reduce: FullOosvar
			nil,         // :
			reduce(117), // ||, reduce: FullOosvar
			reduce(117), // ^^, reduce: FullOosvar
			reduce(117), // &&, reduce: FullOosvar
			reduce(117), // ??, reduce: FullOosvar
			reduce(117), // ???, reduce: FullOosvar
			reduce(117), // =~, reduce: FullOosvar
			reduce(117), // !=~, reduce: FullOosvar
			reduce(117), // ==, reduce: FullOosvar
			reduce(117), // !=, reduce: FullOosvar
			reduce(117), // <=>, reduce: FullOosvar
			reduce(117), // >=, reduce: FullOosvar
			reduce(117), // <, reduce: FullOosvar
			reduce(117), // <=, reduce: FullOosvar
			reduce(117), // ^, reduce: FullOosvar
			reduce(117), // &, reduce: FullOosvar
			reduce(117), // <<, reduce: FullOosvar
			reduce(117), // >>>, reduce: FullOosvar
			reduce(117), // +, reduce: FullOosvar
			reduce(117), // -, reduce: FullOosvar
			reduce(117), // .+, reduce: FullOosvar
			reduce(117), // .-, reduce: FullOosvar
			reduce(117), // *, reduce: FullOosvar
			reduce(117), // /, reduce: FullOosvar
			reduce(117), // //, reduce: FullOosvar
			reduce(117), // %, reduce: FullOosvar
			reduce(117), // .*, reduce: FullOosvar
			reduce(117), // ./, reduce: FullOosvar
			reduce(117), // .//, reduce: FullOosvar
			reduce(117), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(117), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(117), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(117), // [[, reduce: FullOosvar
			reduce(117), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S319
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(118), // $, reduce: LocalVariable
			nil,         // empty
			reduce(118), // ;, reduce: LocalVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(118), // >, reduce: LocalVariable
			reduce(118), // >>, reduce: LocalVariable
			reduce(118), // |, reduce: LocalVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(118), // ,, reduce: LocalVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(270), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(118), // ?, reduce: LocalVariable
			nil,         // :
			reduce(118), // ||, reduce: LocalVariable
			reduce(118), // ^^, reduce: LocalVariable
			reduce(118), // &&, reduce: LocalVariable
			reduce(118), // ??, reduce: LocalVariable
			reduce(118), // ???, reduce: LocalVariable
			reduce(118), // =~, reduce: LocalVariable
			reduce(118), // !=~, reduce: LocalVariable
			reduce(118), // ==, reduce: LocalVariable
			reduce(118), // !=, reduce: LocalVariable
			reduce(118), // <=>, reduce: LocalVariable
			reduce(118), // >=, reduce: LocalVariable
			reduce(118), // <, reduce: LocalVariable
			reduce(118), // <=, reduce: LocalVariable
			reduce(118), // ^, reduce: LocalVariable
			reduce(118), // &, reduce: LocalVariable
			reduce(118), // <<, reduce: LocalVariable
			reduce(118), // >>>, reduce: LocalVariable
			reduce(118), // +, reduce: LocalVariable
			reduce(118), // -, reduce: LocalVariable
			reduce(118), // .+, reduce: LocalVariable
			reduce(118), // .-, reduce: LocalVariable
			reduce(118), // *, reduce: LocalVariable
			reduce(118), // /, reduce: LocalVariable
			reduce(118), // //, reduce: LocalVariable
			reduce(118), // %, reduce: LocalVariable
			reduce(118), // .*, reduce: LocalVariable
			reduce(118), // ./, reduce: LocalVariable
			reduce(118), // .//, reduce: LocalVariable
			reduce(118), // ., reduce: LocalVariable
			nil,         // !
			nil,         // ~
			reduce(118), // **, reduce: LocalVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(118), // [, reduce: LocalVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(118), // [[, reduce: LocalVariable
			reduce(118), // [[[, reduce: LocalVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S320
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(147), // $, reduce: Rvalue
			nil,         // empty
			reduce(147), // ;, reduce: Rvalue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(147), // ,, reduce: Rvalue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S321
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(148), // $, reduce: PrecedenceChainStart
			nil,         // empty
			reduce(148), // ;, reduce: PrecedenceChainStart
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(148), // ,, reduce: PrecedenceChainStart
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S322
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(150), // $, reduce: TernaryTerm
			nil,         // empty
			reduce(150), // ;, reduce: TernaryTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(150), // ,, reduce: TernaryTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			shift(1111), // ?
			nil,         // :
			shift(1112), // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S323
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(152), // $, reduce: LogicalOrTerm
			nil,         // empty
			reduce(152), // ;, reduce: LogicalOrTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(152), // ,, reduce: LogicalOrTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(152), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(152), // ||, reduce: LogicalOrTerm
			shift(1113), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S324
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(154), // $, reduce: LogicalXORTerm
			nil,         // empty
			reduce(154), // ;, reduce: LogicalXORTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(154), // ,, reduce: LogicalXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(154), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(154), // ||, reduce: LogicalXORTerm
			reduce(154), // ^^, reduce: LogicalXORTerm
			shift(1114), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S325
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(156), // $, reduce: LogicalAndTerm
			nil,         // empty
			reduce(156), // ;, reduce: LogicalAndTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(156), // ,, reduce: LogicalAndTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(156), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(156), // ||, reduce: LogicalAndTerm
			reduce(156), // ^^, reduce: LogicalAndTerm
			reduce(156), // &&, reduce: LogicalAndTerm
			shift(1115), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S326
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(158), // $, reduce: AbsentCoalesceTerm
			nil,         // empty
			reduce(158), // ;, reduce: AbsentCoalesceTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(158), // ,, reduce: AbsentCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(158), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(158), // ||, reduce: AbsentCoalesceTerm
			reduce(158), // ^^, reduce: AbsentCoalesceTerm
			reduce(158), // &&, reduce: AbsentCoalesceTerm
			reduce(158), // ??, reduce: AbsentCoalesceTerm
			shift(1116), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S327
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(160), // $, reduce: EmptyCoalesceTerm
			nil,         // empty
			reduce(160), // ;, reduce: EmptyCoalesceTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(160), // ,, reduce: EmptyCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(160), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(160), // ||, reduce: EmptyCoalesceTerm
			reduce(160), // ^^, reduce: EmptyCoalesceTerm
			reduce(160), // &&, reduce: EmptyCoalesceTerm
			reduce(160), // ??, reduce: EmptyCoalesceTerm
			reduce(160), // ???, reduce: EmptyCoalesceTerm
			shift(1117), // =~
			shift(1118), // !=~
			shift(1119), // ==
			shift(1120), // !=
			shift(1121), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S328
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(166), // $, reduce: EqneTerm
			nil,         // empty
			reduce(166), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1122), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(166), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(166), // ?, reduce: EqneTerm
			nil,         // :
			reduce(166), // ||, reduce: EqneTerm
			reduce(166), // ^^, reduce: EqneTerm
			reduce(166), // &&, reduce: EqneTerm
			reduce(166), // ??, reduce: EqneTerm
			reduce(166), // ???, reduce: EqneTerm
			reduce(166), // =~, reduce: EqneTerm
			reduce(166), // !=~, reduce: EqneTerm
			reduce(166), // ==, reduce: EqneTerm
			reduce(166), // !=, reduce: EqneTerm
			reduce(166), // <=>, reduce: EqneTerm
			shift(1123), // >=
			shift(1124), // <
			shift(1125), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S329
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(171), // $, reduce: CmpTerm
			nil,         // empty
			reduce(171), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(171), // >, reduce: CmpTerm
			nil,         // >>
			shift(1126), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(171), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(171), // ?, reduce: CmpTerm
			nil,         // :
			reduce(171), // ||, reduce: CmpTerm
			reduce(171), // ^^, reduce: CmpTerm
			reduce(171), // &&, reduce: CmpTerm
			reduce(171), // ??, reduce: CmpTerm
			reduce(171), // ???, reduce: CmpTerm
			reduce(171), // =~, reduce: CmpTerm
			reduce(171), // !=~, reduce: CmpTerm
			reduce(171), // ==, reduce: CmpTerm
			reduce(171), // !=, reduce: CmpTerm
			reduce(171), // <=>, reduce: CmpTerm
			reduce(171), // >=, reduce: CmpTerm
			reduce(171), // <, reduce: CmpTerm
			reduce(171), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S330
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(173), // $, reduce: BitwiseORTerm
			nil,         // empty
			reduce(173), // ;, reduce: BitwiseORTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(173), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(173), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(173), // ,, reduce: BitwiseORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(173), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(173), // ||, reduce: BitwiseORTerm
			reduce(173), // ^^, reduce: BitwiseORTerm
			reduce(173), // &&, reduce: BitwiseORTerm
			reduce(173), // ??, reduce: BitwiseORTerm
			reduce(173), // ???, reduce: BitwiseORTerm
			reduce(173), // =~, reduce: BitwiseORTerm
			reduce(173), // !=~, reduce: BitwiseORTerm
			reduce(173), // ==, reduce: BitwiseORTerm
			reduce(173), // !=, reduce: BitwiseORTerm
			reduce(173), // <=>, reduce: BitwiseORTerm
			reduce(173), // >=, reduce: BitwiseORTerm
			reduce(173), // <, reduce: BitwiseORTerm
			reduce(173), // <=, reduce: BitwiseORTerm
			shift(1127), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S331
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(175), // $, reduce: BitwiseXORTerm
			nil,         // empty
			reduce(175), // ;, reduce: BitwiseXORTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(175), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(175), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(175), // ,, reduce: BitwiseXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(175), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(175), // ||, reduce: BitwiseXORTerm
			reduce(175), // ^^, reduce: BitwiseXORTerm
			reduce(175), // &&, reduce: BitwiseXORTerm
			reduce(175), // ??, reduce: BitwiseXORTerm
			reduce(175), // ???, reduce: BitwiseXORTerm
			reduce(175), // =~, reduce: BitwiseXORTerm
			reduce(175), // !=~, reduce: BitwiseXORTerm
			reduce(175), // ==, reduce: BitwiseXORTerm
			reduce(175), // !=, reduce: BitwiseXORTerm
			reduce(175), // <=>, reduce: BitwiseXORTerm
			reduce(175), // >=, reduce: BitwiseXORTerm
			reduce(175), // <, reduce: BitwiseXORTerm
			reduce(175), // <=, reduce: BitwiseXORTerm
			reduce(175), // ^, reduce: BitwiseXORTerm
			shift(1128), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S332
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(177), // $, reduce: BitwiseANDTerm
			nil,         // empty
			reduce(177), // ;, reduce: BitwiseANDTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(177), // >, reduce: BitwiseANDTerm
			shift(1129), // >>
			reduce(177), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(177), // ,, reduce: BitwiseANDTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(177), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(177), // ||, reduce: BitwiseANDTerm
			reduce(177), // ^^, reduce: BitwiseANDTerm
			reduce(177), // &&, reduce: BitwiseANDTerm
			reduce(177), // ??, reduce: BitwiseANDTerm
			reduce(177), // ???, reduce: BitwiseANDTerm
			reduce(177), // =~, reduce: BitwiseANDTerm
			reduce(177), // !=~, reduce: BitwiseANDTerm
			reduce(177), // ==, reduce: BitwiseANDTerm
			reduce(177), // !=, reduce: BitwiseANDTerm
			reduce(177), // <=>, reduce: BitwiseANDTerm
			reduce(177), // >=, reduce: BitwiseANDTerm
			reduce(177), // <, reduce: BitwiseANDTerm
			reduce(177), // <=, reduce: BitwiseANDTerm
			reduce(177), // ^, reduce: BitwiseANDTerm
			reduce(177), // &, reduce: BitwiseANDTerm
			shift(1130), // <<
			shift(1131), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S333
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(181), // $, reduce: BitwiseShiftTerm
			nil,         // empty
			reduce(181), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(181), // >, reduce: BitwiseShiftTerm
			reduce(181), // >>, reduce: BitwiseShiftTerm
			reduce(181), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(181), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(181), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(181), // ||, reduce: BitwiseShiftTerm
			reduce(181), // ^^, reduce: BitwiseShiftTerm
			reduce(181), // &&, reduce: BitwiseShiftTerm
			reduce(181), // ??, reduce: BitwiseShiftTerm
			reduce(181), // ???, reduce: BitwiseShiftTerm
			reduce(181), // =~, reduce: BitwiseShiftTerm
			reduce(181), // !=~, reduce: BitwiseShiftTerm
			reduce(181), // ==, reduce: BitwiseShiftTerm
			reduce(181), // !=, reduce: BitwiseShiftTerm
			reduce(181), // <=>, reduce: BitwiseShiftTerm
			reduce(181), // >=, reduce: BitwiseShiftTerm
			reduce(181), // <, reduce: BitwiseShiftTerm
			reduce(181), // <=, reduce: BitwiseShiftTerm
			reduce(181), // ^, reduce: BitwiseShiftTerm
			reduce(181), // &, reduce: BitwiseShiftTerm
			reduce(181), // <<, reduce: BitwiseShiftTerm
			reduce(181), // >>>, reduce: BitwiseShiftTerm
			shift(1132), // +
			shift(1133), // -
			shift(1134), // .+
			shift(1135), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S334
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S335
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(186), // $, reduce: AddsubdotTerm
			nil,         // empty
			reduce(186), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(186), // >, reduce: AddsubdotTerm
			reduce(186), // >>, reduce: AddsubdotTerm
			reduce(186), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(186), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(186), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(186), // ||, reduce: AddsubdotTerm
			reduce(186), // ^^, reduce: AddsubdotTerm
			reduce(186), // &&, reduce: AddsubdotTerm
			reduce(186), // ??, reduce: AddsubdotTerm
			reduce(186), // ???, reduce: AddsubdotTerm
			reduce(186), // =~, reduce: AddsubdotTerm
			reduce(186), // !=~, reduce: AddsubdotTerm
			reduce(186), // ==, reduce: AddsubdotTerm
			reduce(186), // !=, reduce: AddsubdotTerm
			reduce(186), // <=>, reduce: AddsubdotTerm
			reduce(186), // >=, reduce: AddsubdotTerm
			reduce(186), // <, reduce: AddsubdotTerm
			reduce(186), // <=, reduce: AddsubdotTerm
			reduce(186), // ^, reduce: AddsubdotTerm
			reduce(186), // &, reduce: AddsubdotTerm
			reduce(186), // <<, reduce: AddsubdotTerm
			reduce(186), // >>>, reduce: AddsubdotTerm
			reduce(186), // +, reduce: AddsubdotTerm
			reduce(186), // -, reduce: AddsubdotTerm
			reduce(186), // .+, reduce: AddsubdotTerm
			reduce(186), // .-, reduce: AddsubdotTerm
			shift(1137), // *
			shift(1138), // /
			shift(1139), // //
			shift(1140), // %
			shift(1141), // .*
			shift(1142), // ./
			shift(1143), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S336
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S337
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S338
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S339
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(194), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(194), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(194), // >, reduce: MuldivTerm
			reduce(194), // >>, reduce: MuldivTerm
			reduce(194), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(194), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(194), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(194), // ||, reduce: MuldivTerm
			reduce(194), // ^^, reduce: MuldivTerm
			reduce(194), // &&, reduce: MuldivTerm
			reduce(194), // ??, reduce: MuldivTerm
			reduce(194), // ???, reduce: MuldivTerm
			reduce(194), // =~, reduce: MuldivTerm
			reduce(194), // !=~, reduce: MuldivTerm
			reduce(194), // ==, reduce: MuldivTerm
			reduce(194), // !=, reduce: MuldivTerm
			reduce(194), // <=>, reduce: MuldivTerm
			reduce(194), // >=, reduce: MuldivTerm
			reduce(194), // <, reduce: MuldivTerm
			reduce(194), // <=, reduce: MuldivTerm
			reduce(194), // ^, reduce: MuldivTerm
			reduce(194), // &, reduce: MuldivTerm
			reduce(194), // <<, reduce: MuldivTerm
			reduce(194), // >>>, reduce: MuldivTerm
			reduce(194), // +, reduce: MuldivTerm
			reduce(194), // -, reduce: MuldivTerm
			reduce(194), // .+, reduce: MuldivTerm
			reduce(194), // .-, reduce: MuldivTerm
			reduce(194), // *, reduce: MuldivTerm
			reduce(194), // /, reduce: MuldivTerm
			reduce(194), // //, reduce: MuldivTerm
			reduce(194), // %, reduce: MuldivTerm
			reduce(194), // .*, reduce: MuldivTerm
			reduce(194), // ./, reduce: MuldivTerm
			reduce(194), // .//, reduce: MuldivTerm
			shift(1147), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S340
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(196), // $, reduce: DotTerm
			nil,         // empty
			reduce(196), // ;, reduce: DotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(196), // >, reduce: DotTerm
			reduce(196), // >>, reduce: DotTerm
			reduce(196), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(196), // ,, reduce: DotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(196), // ?, reduce: DotTerm
			nil,         // :
			reduce(196), // ||, reduce: DotTerm
			reduce(196), // ^^, reduce: DotTerm
			reduce(196), // &&, reduce: DotTerm
			reduce(196), // ??, reduce: DotTerm
			reduce(196), // ???, reduce: DotTerm
			reduce(196), // =~, reduce: DotTerm
			reduce(196), // !=~, reduce: DotTerm
			reduce(196), // ==, reduce: DotTerm
			reduce(196), // !=, reduce: DotTerm
			reduce(196), // <=>, reduce: DotTerm
			reduce(196), // >=, reduce: DotTerm
			reduce(196), // <, reduce: DotTerm
			reduce(196), // <=, reduce: DotTerm
			reduce(196), // ^, reduce: DotTerm
			reduce(196), // &, reduce: DotTerm
			reduce(196), // <<, reduce: DotTerm
			reduce(196), // >>>, reduce: DotTerm
			reduce(196), // +, reduce: DotTerm
			reduce(196), // -, reduce: DotTerm
			reduce(196), // .+, reduce: DotTerm
			reduce(196), // .-, reduce: DotTerm
			reduce(196), // *, reduce: DotTerm
			reduce(196), // /, reduce: DotTerm
			reduce(196), // //, reduce: DotTerm
			reduce(196), // %, reduce: DotTerm
			reduce(196), // .*, reduce: DotTerm
			reduce(196), // ./, reduce: DotTerm
			reduce(196), // .//, reduce: DotTerm
			reduce(196), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S341
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S342
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S343
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(203), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(203), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(203), // >, reduce: UnaryOpTerm
			reduce(203), // >>, reduce: UnaryOpTerm
			reduce(203), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(203), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(203), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(203), // ||, reduce: UnaryOpTerm
			reduce(203), // ^^, reduce: UnaryOpTerm
			reduce(203), // &&, reduce: UnaryOpTerm
			reduce(203), // ??, reduce: UnaryOpTerm
			reduce(203), // ???, reduce: UnaryOpTerm
			reduce(203), // =~, reduce: UnaryOpTerm
			reduce(203), // !=~, reduce: UnaryOpTerm
			reduce(203), // ==, reduce: UnaryOpTerm
			reduce(203), // !=, reduce: UnaryOpTerm
			reduce(203), // <=>, reduce: UnaryOpTerm
			reduce(203), // >=, reduce: UnaryOpTerm
			reduce(203), // <, reduce: UnaryOpTerm
			reduce(203), // <=, reduce: UnaryOpTerm
			reduce(203), // ^, reduce: UnaryOpTerm
			reduce(203), // &, reduce: UnaryOpTerm
			reduce(203), // <<, reduce: UnaryOpTerm
			reduce(203), // >>>, reduce: UnaryOpTerm
			reduce(203), // +, reduce: UnaryOpTerm
			reduce(203), // -, reduce: UnaryOpTerm
			reduce(203), // .+, reduce: UnaryOpTerm
			reduce(203), // .-, reduce: UnaryOpTerm
			reduce(203), // *, reduce: UnaryOpTerm
			reduce(203), // /, reduce: UnaryOpTerm
			reduce(203), // //, reduce: UnaryOpTerm
			reduce(203), // %, reduce: UnaryOpTerm
			reduce(203), // .*, reduce: UnaryOpTerm
			reduce(203), // ./, reduce: UnaryOpTerm
			reduce(203), // .//, reduce: UnaryOpTerm
			reduce(203), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S344
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(207), // $, reduce: PowTerm
			nil,         // empty
			reduce(207), // ;, reduce: PowTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(207), // >, reduce: PowTerm
			reduce(207), // >>, reduce: PowTerm
			reduce(207), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(207), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(207), // ?, reduce: PowTerm
			nil,         // :
			reduce(207), // ||, reduce: PowTerm
			reduce(207), // ^^, reduce: PowTerm
			reduce(207), // &&, reduce: PowTerm
			reduce(207), // ??, reduce: PowTerm
			reduce(207), // ???, reduce: PowTerm
			reduce(207), // =~, reduce: PowTerm
			reduce(207), // !=~, reduce: PowTerm
			reduce(207), // ==, reduce: PowTerm
			reduce(207), // !=, reduce: PowTerm
			reduce(207), // <=>, reduce: PowTerm
			reduce(207), // >=, reduce: PowTerm
			reduce(207), // <, reduce: PowTerm
			reduce(207), // <=, reduce: PowTerm
			reduce(207), // ^, reduce: PowTerm
			reduce(207), // &, reduce: PowTerm
			reduce(207), // <<, reduce: PowTerm
			reduce(207), // >>>, reduce: PowTerm
			reduce(207), // +, reduce: PowTerm
			reduce(207), // -, reduce: PowTerm
			reduce(207), // .+, reduce: PowTerm
			reduce(207), // .-, reduce: PowTerm
			reduce(207), // *, reduce: PowTerm
			reduce(207), // /, reduce: PowTerm
			reduce(207), // //, reduce: PowTerm
			reduce(207), // %, reduce: PowTerm
			reduce(207), // .*, reduce: PowTerm
			reduce(207), // ./, reduce: PowTerm
			reduce(207), // .//, reduce: PowTerm
			reduce(207), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			shift(1150), // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S345
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(209), // $, reduce: PrecedenceChainEnd
			nil,         // empty
			reduce(209), // ;, reduce: PrecedenceChainEnd
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(209), // >, reduce: PrecedenceChainEnd
			reduce(209), // >>, reduce: PrecedenceChainEnd
			reduce(209), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(209), // ,, reduce: PrecedenceChainEnd
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(209), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(209), // ||, reduce: PrecedenceChainEnd
			reduce(209), // ^^, reduce: PrecedenceChainEnd
			reduce(209), // &&, reduce: PrecedenceChainEnd
			reduce(209), // ??, reduce: PrecedenceChainEnd
			reduce(209), // ???, reduce: PrecedenceChainEnd
			reduce(209), // =~, reduce: PrecedenceChainEnd
			reduce(209), // !=~, reduce: PrecedenceChainEnd
			reduce(209), // ==, reduce: PrecedenceChainEnd
			reduce(209), // !=, reduce: PrecedenceChainEnd
			reduce(209), // <=>, reduce: PrecedenceChainEnd
			reduce(209), // >=, reduce: PrecedenceChainEnd
			reduce(209), // <, reduce: PrecedenceChainEnd
			reduce(209), // <=, reduce: PrecedenceChainEnd
			reduce(209), // ^, reduce: PrecedenceChainEnd
			reduce(209), // &, reduce: PrecedenceChainEnd
			reduce(209), // <<, reduce: PrecedenceChainEnd
			reduce(209), // >>>, reduce: PrecedenceChainEnd
			reduce(209), // +, reduce: PrecedenceChainEnd
			reduce(209), // -, reduce: PrecedenceChainEnd
			reduce(209), // .+, reduce: PrecedenceChainEnd
			reduce(209), // .-, reduce: PrecedenceChainEnd
			reduce(209), // *, reduce: PrecedenceChainEnd
			reduce(209), // /, reduce: PrecedenceChainEnd
			reduce(209), // //, reduce: PrecedenceChainEnd
			reduce(209), // %, reduce: PrecedenceChainEnd
			reduce(209), // .*, reduce: PrecedenceChainEnd
			reduce(209), // ./, reduce: PrecedenceChainEnd
			reduce(209), // .//, reduce: PrecedenceChainEnd
			reduce(209), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(209), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(1151), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			shift(1152), // [[
			shift(1153), // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S346
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(215), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(215), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(215), // >, reduce: MlrvalOrFunction
			reduce(215), // >>, reduce: MlrvalOrFunction
			reduce(215), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(215), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(215), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(215), // ||, reduce: MlrvalOrFunction
			reduce(215), // ^^, reduce: MlrvalOrFunction
			reduce(215), // &&, reduce: MlrvalOrFunction
			reduce(215), // ??, reduce: MlrvalOrFunction
			reduce(215), // ???, reduce: MlrvalOrFunction
			reduce(215), // =~, reduce: MlrvalOrFunction
			reduce(215), // !=~, reduce: MlrvalOrFunction
			reduce(215), // ==, reduce: MlrvalOrFunction
			reduce(215), // !=, reduce: MlrvalOrFunction
			reduce(215), // <=>, reduce: MlrvalOrFunction
			reduce(215), // >=, reduce: MlrvalOrFunction
			reduce(215), // <, reduce: MlrvalOrFunction
			reduce(215), // <=, reduce: MlrvalOrFunction
			reduce(215), // ^, reduce: MlrvalOrFunction
			reduce(215), // &, reduce: MlrvalOrFunction
			reduce(215), // <<, reduce: MlrvalOrFunction
			reduce(215), // >>>, reduce: MlrvalOrFunction
			reduce(215), // +, reduce: MlrvalOrFunction
			reduce(215), // -, reduce: MlrvalOrFunction
			reduce(215), // .+, reduce: MlrvalOrFunction
			reduce(215), // .-, reduce: MlrvalOrFunction
			reduce(215), // *, reduce: MlrvalOrFunction
			reduce(215), // /, reduce: MlrvalOrFunction
			reduce(215), // //, reduce: MlrvalOrFunction
			reduce(215), // %, reduce: MlrvalOrFunction
			reduce(215), // .*, reduce: MlrvalOrFunction
			reduce(215), // ./, reduce: MlrvalOrFunction
			reduce(215), // .//, reduce: MlrvalOrFunction
			reduce(215), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(215), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(215), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(215), // [[, reduce: MlrvalOrFunction
			reduce(215), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S347
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(216), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(216), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(216), // >, reduce: MlrvalOrFunction
			reduce(216), // >>, reduce: MlrvalOrFunction
			reduce(216), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(216), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(216), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(216), // ||, reduce: MlrvalOrFunction
			reduce(216), // ^^, reduce: MlrvalOrFunction
			reduce(216), // &&, reduce: MlrvalOrFunction
			reduce(216), // ??, reduce: MlrvalOrFunction
			reduce(216), // ???, reduce: MlrvalOrFunction
			reduce(216), // =~, reduce: MlrvalOrFunction
			reduce(216), // !=~, reduce: MlrvalOrFunction
			reduce(216), // ==, reduce: MlrvalOrFunction
			reduce(216), // !=, reduce: MlrvalOrFunction
			reduce(216), // <=>, reduce: MlrvalOrFunction
			reduce(216), // >=, reduce: MlrvalOrFunction
			reduce(216), // <, reduce: MlrvalOrFunction
			reduce(216), // <=, reduce: MlrvalOrFunction
			reduce(216), // ^, reduce: MlrvalOrFunction
			reduce(216), // &, reduce: MlrvalOrFunction
			reduce(216), // <<, reduce: MlrvalOrFunction
			reduce(216), // >>>, reduce: MlrvalOrFunction
			reduce(216), // +, reduce: MlrvalOrFunction
			reduce(216), // -, reduce: MlrvalOrFunction
			reduce(216), // .+, reduce: MlrvalOrFunction
			reduce(216), // .-, reduce: MlrvalOrFunction
			reduce(216), // *, reduce: MlrvalOrFunction
			reduce(216), // /, reduce: MlrvalOrFunction
			reduce(216), // //, reduce: MlrvalOrFunction
			reduce(216), // %, reduce: MlrvalOrFunction
			reduce(216), // .*, reduce: MlrvalOrFunction
			reduce(216), // ./, reduce: MlrvalOrFunction
			reduce(216), // .//, reduce: MlrvalOrFunction
			reduce(216), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(216), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(216), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(216), // [[, reduce: MlrvalOrFunction
			reduce(216), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S348
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(217), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(217), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(217), // >, reduce: MlrvalOrFunction
			reduce(217), // >>, reduce: MlrvalOrFunction
			reduce(217), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(217), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(217), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(217), // ||, reduce: MlrvalOrFunction
			reduce(217), // ^^, reduce: MlrvalOrFunction
			reduce(217), // &&, reduce: MlrvalOrFunction
			reduce(217), // ??, reduce: MlrvalOrFunction
			reduce(217), // ???, reduce: MlrvalOrFunction
			reduce(217), // =~, reduce: MlrvalOrFunction
			reduce(217), // !=~, reduce: MlrvalOrFunction
			reduce(217), // ==, reduce: MlrvalOrFunction
			reduce(217), // !=, reduce: MlrvalOrFunction
			reduce(217), // <=>, reduce: MlrvalOrFunction
			reduce(217), // >=, reduce: MlrvalOrFunction
			reduce(217), // <, reduce: MlrvalOrFunction
			reduce(217), // <=, reduce: MlrvalOrFunction
			reduce(217), // ^, reduce: MlrvalOrFunction
			reduce(217), // &, reduce: MlrvalOrFunction
			reduce(217), // <<, reduce: MlrvalOrFunction
			reduce(217), // >>>, reduce: MlrvalOrFunction
			reduce(217), // +, reduce: MlrvalOrFunction
			reduce(217), // -, reduce: MlrvalOrFunction
			reduce(217), // .+, reduce: MlrvalOrFunction
			reduce(217), // .-, reduce: MlrvalOrFunction
			reduce(217), // *, reduce: MlrvalOrFunction
			reduce(217), // /, reduce: MlrvalOrFunction
			reduce(217), // //, reduce: MlrvalOrFunction
			reduce(217), // %, reduce: MlrvalOrFunction
			reduce(217), // .*, reduce: MlrvalOrFunction
			reduce(217), // ./, reduce: MlrvalOrFunction
			reduce(217), // .//, reduce: MlrvalOrFunction
			reduce(217), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(217), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(217), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(217), // [[, reduce: MlrvalOrFunction
			reduce(217), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S349
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(218), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(218), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(218), // >, reduce: MlrvalOrFunction
			reduce(218), // >>, reduce: MlrvalOrFunction
			reduce(218), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(218), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(218), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(218), // ||, reduce: MlrvalOrFunction
			reduce(218), // ^^, reduce: MlrvalOrFunction
			reduce(218), // &&, reduce: MlrvalOrFunction
			reduce(218), // ??, reduce: MlrvalOrFunction
			reduce(218), // ???, reduce: MlrvalOrFunction
			reduce(218), // =~, reduce: MlrvalOrFunction
			reduce(218), // !=~, reduce: MlrvalOrFunction
			reduce(218), // ==, reduce: MlrvalOrFunction
			reduce(218), // !=, reduce: MlrvalOrFunction
			reduce(218), // <=>, reduce: MlrvalOrFunction
			reduce(218), // >=, reduce: MlrvalOrFunction
			reduce(218), // <, reduce: MlrvalOrFunction
			reduce(218), // <=, reduce: MlrvalOrFunction
			reduce(218), // ^, reduce: MlrvalOrFunction
			reduce(218), // &, reduce: MlrvalOrFunction
			reduce(218), // <<, reduce: MlrvalOrFunction
			reduce(218), // >>>, reduce: MlrvalOrFunction
			reduce(218), // +, reduce: MlrvalOrFunction
			reduce(218), // -, reduce: MlrvalOrFunction
			reduce(218), // .+, reduce: MlrvalOrFunction
			reduce(218), // .-, reduce: MlrvalOrFunction
			reduce(218), // *, reduce: MlrvalOrFunction
			reduce(218), // /, reduce: MlrvalOrFunction
			reduce(218), // //, reduce: MlrvalOrFunction
			reduce(218), // %, reduce: MlrvalOrFunction
			reduce(218), // .*, reduce: MlrvalOrFunction
			reduce(218), // ./, reduce: MlrvalOrFunction
			reduce(218), // .//, reduce: MlrvalOrFunction
			reduce(218), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(218), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(218), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(218), // [[, reduce: MlrvalOrFunction
			reduce(218), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S350
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(219), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(219), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(219), // >, reduce: MlrvalOrFunction
			reduce(219), // >>, reduce: MlrvalOrFunction
			reduce(219), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(219), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(219), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(219), // ||, reduce: MlrvalOrFunction
			reduce(219), // ^^, reduce: MlrvalOrFunction
			reduce(219), // &&, reduce: MlrvalOrFunction
			reduce(219), // ??, reduce: MlrvalOrFunction
			reduce(219), // ???, reduce: MlrvalOrFunction
			reduce(219), // =~, reduce: MlrvalOrFunction
			reduce(219), // !=~, reduce: MlrvalOrFunction
			reduce(219), // ==, reduce: MlrvalOrFunction
			reduce(219), // !=, reduce: MlrvalOrFunction
			reduce(219), // <=>, reduce: MlrvalOrFunction
			reduce(219), // >=, reduce: MlrvalOrFunction
			reduce(219), // <, reduce: MlrvalOrFunction
			reduce(219), // <=, reduce: MlrvalOrFunction
			reduce(219), // ^, reduce: MlrvalOrFunction
			reduce(219), // &, reduce: MlrvalOrFunction
			reduce(219), // <<, reduce: MlrvalOrFunction
			reduce(219), // >>>, reduce: MlrvalOrFunction
			reduce(219), // +, reduce: MlrvalOrFunction
			reduce(219), // -, reduce: MlrvalOrFunction
			reduce(219), // .+, reduce: MlrvalOrFunction
			reduce(219), // .-, reduce: MlrvalOrFunction
			reduce(219), // *, reduce: MlrvalOrFunction
			reduce(219), // /, reduce: MlrvalOrFunction
			reduce(219), // //, reduce: MlrvalOrFunction
			reduce(219), // %, reduce: MlrvalOrFunction
			reduce(219), // .*, reduce: MlrvalOrFunction
			reduce(219), // ./, reduce: MlrvalOrFunction
			reduce(219), // .//, reduce: MlrvalOrFunction
			reduce(219), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(219), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(219), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(219), // [[, reduce: MlrvalOrFunction
			reduce(219), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S351
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(220), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(220), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(220), // >, reduce: MlrvalOrFunction
			reduce(220), // >>, reduce: MlrvalOrFunction
			reduce(220), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(220), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(220), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(220), // ||, reduce: MlrvalOrFunction
			reduce(220), // ^^, reduce: MlrvalOrFunction
			reduce(220), // &&, reduce: MlrvalOrFunction
			reduce(220), // ??, reduce: MlrvalOrFunction
			reduce(220), // ???, reduce: MlrvalOrFunction
			reduce(220), // =~, reduce: MlrvalOrFunction
			reduce(220), // !=~, reduce: MlrvalOrFunction
			reduce(220), // ==, reduce: MlrvalOrFunction
			reduce(220), // !=, reduce: MlrvalOrFunction
			reduce(220), // <=>, reduce: MlrvalOrFunction
			reduce(220), // >=, reduce: MlrvalOrFunction
			reduce(220), // <, reduce: MlrvalOrFunction
			reduce(220), // <=, reduce: MlrvalOrFunction
			reduce(220), // ^, reduce: MlrvalOrFunction
			reduce(220), // &, reduce: MlrvalOrFunction
			reduce(220), // <<, reduce: MlrvalOrFunction
			reduce(220), // >>>, reduce: MlrvalOrFunction
			reduce(220), // +, reduce: MlrvalOrFunction
			reduce(220), // -, reduce: MlrvalOrFunction
			reduce(220), // .+, reduce: MlrvalOrFunction
			reduce(220), // .-, reduce: MlrvalOrFunction
			reduce(220), // *, reduce: MlrvalOrFunction
			reduce(220), // /, reduce: MlrvalOrFunction
			reduce(220), // //, reduce: MlrvalOrFunction
			reduce(220), // %, reduce: MlrvalOrFunction
			reduce(220), // .*, reduce: MlrvalOrFunction
			reduce(220), // ./, reduce: MlrvalOrFunction
			reduce(220), // .//, reduce: MlrvalOrFunction
			reduce(220), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(220), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(220), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(220), // [[, reduce: MlrvalOrFunction
			reduce(220), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S352
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(221), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(221), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(221), // >, reduce: MlrvalOrFunction
			reduce(221), // >>, reduce: MlrvalOrFunction
			reduce(221), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(221), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(221), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(221), // ||, reduce: MlrvalOrFunction
			reduce(221), // ^^, reduce: MlrvalOrFunction
			reduce(221), // &&, reduce: MlrvalOrFunction
			reduce(221), // ??, reduce: MlrvalOrFunction
			reduce(221), // ???, reduce: MlrvalOrFunction
			reduce(221), // =~, reduce: MlrvalOrFunction
			reduce(221), // !=~, reduce: MlrvalOrFunction
			reduce(221), // ==, reduce: MlrvalOrFunction
			reduce(221), // !=, reduce: MlrvalOrFunction
			reduce(221), // <=>, reduce: MlrvalOrFunction
			reduce(221), // >=, reduce: MlrvalOrFunction
			reduce(221), // <, reduce: MlrvalOrFunction
			reduce(221), // <=, reduce: MlrvalOrFunction
			reduce(221), // ^, reduce: MlrvalOrFunction
			reduce(221), // &, reduce: MlrvalOrFunction
			reduce(221), // <<, reduce: MlrvalOrFunction
			reduce(221), // >>>, reduce: MlrvalOrFunction
			reduce(221), // +, reduce: MlrvalOrFunction
			reduce(221), // -, reduce: MlrvalOrFunction
			reduce(221), // .+, reduce: MlrvalOrFunction
			reduce(221), // .-, reduce: MlrvalOrFunction
			reduce(221), // *, reduce: MlrvalOrFunction
			reduce(221), // /, reduce: MlrvalOrFunction
			reduce(221), // //, reduce: MlrvalOrFunction
			reduce(221), // %, reduce: MlrvalOrFunction
			reduce(221), // .*, reduce: MlrvalOrFunction
			reduce(221), // ./, reduce: MlrvalOrFunction
			reduce(221), // .//, reduce: MlrvalOrFunction
			reduce(221), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(221), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(221), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(221), // [[, reduce: MlrvalOrFunction
			reduce(221), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S353
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(222), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(222), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(222), // >, reduce: MlrvalOrFunction
			reduce(222), // >>, reduce: MlrvalOrFunction
			reduce(222), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(222), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(222), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(222), // ||, reduce: MlrvalOrFunction
			reduce(222), // ^^, reduce: MlrvalOrFunction
			reduce(222), // &&, reduce: MlrvalOrFunction
			reduce(222), // ??, reduce: MlrvalOrFunction
			reduce(222), // ???, reduce: MlrvalOrFunction
			reduce(222), // =~, reduce: MlrvalOrFunction
			reduce(222), // !=~, reduce: MlrvalOrFunction
			reduce(222), // ==, reduce: MlrvalOrFunction
			reduce(222), // !=, reduce: MlrvalOrFunction
			reduce(222), // <=>, reduce: MlrvalOrFunction
			reduce(222), // >=, reduce: MlrvalOrFunction
			reduce(222), // <, reduce: MlrvalOrFunction
			reduce(222), // <=, reduce: MlrvalOrFunction
			reduce(222), // ^, reduce: MlrvalOrFunction
			reduce(222), // &, reduce: MlrvalOrFunction
			reduce(222), // <<, reduce: MlrvalOrFunction
			reduce(222), // >>>, reduce: MlrvalOrFunction
			reduce(222), // +, reduce: MlrvalOrFunction
			reduce(222), // -, reduce: MlrvalOrFunction
			reduce(222), // .+, reduce: MlrvalOrFunction
			reduce(222), // .-, reduce: MlrvalOrFunction
			reduce(222), // *, reduce: MlrvalOrFunction
			reduce(222), // /, reduce: MlrvalOrFunction
			reduce(222), // //, reduce: MlrvalOrFunction
			reduce(222), // %, reduce: MlrvalOrFunction
			reduce(222), // .*, reduce: MlrvalOrFunction
			reduce(222), // ./, reduce: MlrvalOrFunction
			reduce(222), // .//, reduce: MlrvalOrFunction
			reduce(222), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(222), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(222), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(222), // [[, reduce: MlrvalOrFunction
			reduce(222), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S354
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(223), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(223), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(223), // >, reduce: MlrvalOrFunction
			reduce(223), // >>, reduce: MlrvalOrFunction
			reduce(223), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(223), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(223), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(223), // ||, reduce: MlrvalOrFunction
			reduce(223), // ^^, reduce: MlrvalOrFunction
			reduce(223), // &&, reduce: MlrvalOrFunction
			reduce(223), // ??, reduce: MlrvalOrFunction
			reduce(223), // ???, reduce: MlrvalOrFunction
			reduce(223), // =~, reduce: MlrvalOrFunction
			reduce(223), // !=~, reduce: MlrvalOrFunction
			reduce(223), // ==, reduce: MlrvalOrFunction
			reduce(223), // !=, reduce: MlrvalOrFunction
			reduce(223), // <=>, reduce: MlrvalOrFunction
			reduce(223), // >=, reduce: MlrvalOrFunction
			reduce(223), // <, reduce: MlrvalOrFunction
			reduce(223), // <=, reduce: MlrvalOrFunction
			reduce(223), // ^, reduce: MlrvalOrFunction
			reduce(223), // &, reduce: MlrvalOrFunction
			reduce(223), // <<, reduce: MlrvalOrFunction
			reduce(223), // >>>, reduce: MlrvalOrFunction
			reduce(223), // +, reduce: MlrvalOrFunction
			reduce(223), // -, reduce: MlrvalOrFunction
			reduce(223), // .+, reduce: MlrvalOrFunction
			reduce(223), // .-, reduce: MlrvalOrFunction
			reduce(223), // *, reduce: MlrvalOrFunction
			reduce(223), // /, reduce: MlrvalOrFunction
			reduce(223), // //, reduce: MlrvalOrFunction
			reduce(223), // %, reduce: MlrvalOrFunction
			reduce(223), // .*, reduce: MlrvalOrFunction
			reduce(223), // ./, reduce: MlrvalOrFunction
			reduce(223), // .//, reduce: MlrvalOrFunction
			reduce(223), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(223), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(223), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(223), // [[, reduce: MlrvalOrFunction
			reduce(223), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S355
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(224), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(224), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(224), // >, reduce: MlrvalOrFunction
			reduce(224), // >>, reduce: MlrvalOrFunction
			reduce(224), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(224), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(224), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(224), // ||, reduce: MlrvalOrFunction
			reduce(224), // ^^, reduce: MlrvalOrFunction
			reduce(224), // &&, reduce: MlrvalOrFunction
			reduce(224), // ??, reduce: MlrvalOrFunction
			reduce(224), // ???, reduce: MlrvalOrFunction
			reduce(224), // =~, reduce: MlrvalOrFunction
			reduce(224), // !=~, reduce: MlrvalOrFunction
			reduce(224), // ==, reduce: MlrvalOrFunction
			reduce(224), // !=, reduce: MlrvalOrFunction
			reduce(224), // <=>, reduce: MlrvalOrFunction
			reduce(224), // >=, reduce: MlrvalOrFunction
			reduce(224), // <, reduce: MlrvalOrFunction
			reduce(224), // <=, reduce: MlrvalOrFunction
			reduce(224), // ^, reduce: MlrvalOrFunction
			reduce(224), // &, reduce: MlrvalOrFunction
			reduce(224), // <<, reduce: MlrvalOrFunction
			reduce(224), // >>>, reduce: MlrvalOrFunction
			reduce(224), // +, reduce: MlrvalOrFunction
			reduce(224), // -, reduce: MlrvalOrFunction
			reduce(224), // .+, reduce: MlrvalOrFunction
			reduce(224), // .-, reduce: MlrvalOrFunction
			reduce(224), // *, reduce: MlrvalOrFunction
			reduce(224), // /, reduce: MlrvalOrFunction
			reduce(224), // //, reduce: MlrvalOrFunction
			reduce(224), // %, reduce: MlrvalOrFunction
			reduce(224), // .*, reduce: MlrvalOrFunction
			reduce(224), // ./, reduce: MlrvalOrFunction
			reduce(224), // .//, reduce: MlrvalOrFunction
			reduce(224), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(224), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(224), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(224), // [[, reduce: MlrvalOrFunction
			reduce(224), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S356
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(225), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(225), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(225), // >, reduce: MlrvalOrFunction
			reduce(225), // >>, reduce: MlrvalOrFunction
			reduce(225), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(225), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(225), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(225), // ||, reduce: MlrvalOrFunction
			reduce(225), // ^^, reduce: MlrvalOrFunction
			reduce(225), // &&, reduce: MlrvalOrFunction
			reduce(225), // ??, reduce: MlrvalOrFunction
			reduce(225), // ???, reduce: MlrvalOrFunction
			reduce(225), // =~, reduce: MlrvalOrFunction
			reduce(225), // !=~, reduce: MlrvalOrFunction
			reduce(225), // ==, reduce: MlrvalOrFunction
			reduce(225), // !=, reduce: MlrvalOrFunction
			reduce(225), // <=>, reduce: MlrvalOrFunction
			reduce(225), // >=, reduce: MlrvalOrFunction
			reduce(225), // <, reduce: MlrvalOrFunction
			reduce(225), // <=, reduce: MlrvalOrFunction
			reduce(225), // ^, reduce: MlrvalOrFunction
			reduce(225), // &, reduce: MlrvalOrFunction
			reduce(225), // <<, reduce: MlrvalOrFunction
			reduce(225), // >>>, reduce: MlrvalOrFunction
			reduce(225), // +, reduce: MlrvalOrFunction
			reduce(225), // -, reduce: MlrvalOrFunction
			reduce(225), // .+, reduce: MlrvalOrFunction
			reduce(225), // .-, reduce: MlrvalOrFunction
			reduce(225), // *, reduce: MlrvalOrFunction
			reduce(225), // /, reduce: MlrvalOrFunction
			reduce(225), // //, reduce: MlrvalOrFunction
			reduce(225), // %, reduce: MlrvalOrFunction
			reduce(225), // .*, reduce: MlrvalOrFunction
			reduce(225), // ./, reduce: MlrvalOrFunction
			reduce(225), // .//, reduce: MlrvalOrFunction
			reduce(225), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(225), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(225), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(225), // [[, reduce: MlrvalOrFunction
			reduce(225), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S357
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(226), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(226), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(226), // >, reduce: MlrvalOrFunction
			reduce(226), // >>, reduce: MlrvalOrFunction
			reduce(226), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(226), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(226), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(226), // ||, reduce: MlrvalOrFunction
			reduce(226), // ^^, reduce: MlrvalOrFunction
			reduce(226), // &&, reduce: MlrvalOrFunction
			reduce(226), // ??, reduce: MlrvalOrFunction
			reduce(226), // ???, reduce: MlrvalOrFunction
			reduce(226), // =~, reduce: MlrvalOrFunction
			reduce(226), // !=~, reduce: MlrvalOrFunction
			reduce(226), // ==, reduce: MlrvalOrFunction
			reduce(226), // !=, reduce: MlrvalOrFunction
			reduce(226), // <=>, reduce: MlrvalOrFunction
			reduce(226), // >=, reduce: MlrvalOrFunction
			reduce(226), // <, reduce: MlrvalOrFunction
			reduce(226), // <=, reduce: MlrvalOrFunction
			reduce(226), // ^, reduce: MlrvalOrFunction
			reduce(226), // &, reduce: MlrvalOrFunction
			reduce(226), // <<, reduce: MlrvalOrFunction
			reduce(226), // >>>, reduce: MlrvalOrFunction
			reduce(226), // +, reduce: MlrvalOrFunction
			reduce(226), // -, reduce: MlrvalOrFunction
			reduce(226), // .+, reduce: MlrvalOrFunction
			reduce(226), // .-, reduce: MlrvalOrFunction
			reduce(226), // *, reduce: MlrvalOrFunction
			reduce(226), // /, reduce: MlrvalOrFunction
			reduce(226), // //, reduce: MlrvalOrFunction
			reduce(226), // %, reduce: MlrvalOrFunction
			reduce(226), // .*, reduce: MlrvalOrFunction
			reduce(226), // ./, reduce: MlrvalOrFunction
			reduce(226), // .//, reduce: MlrvalOrFunction
			reduce(226), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(226), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(226), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(226), // [[, reduce: MlrvalOrFunction
			reduce(226), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S358
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(227), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(227), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(227), // >, reduce: MlrvalOrFunction
			reduce(227), // >>, reduce: MlrvalOrFunction
			reduce(227), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(227), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(227), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(227), // ||, reduce: MlrvalOrFunction
			reduce(227), // ^^, reduce: MlrvalOrFunction
			reduce(227), // &&, reduce: MlrvalOrFunction
			reduce(227), // ??, reduce: MlrvalOrFunction
			reduce(227), // ???, reduce: MlrvalOrFunction
			reduce(227), // =~, reduce: MlrvalOrFunction
			reduce(227), // !=~, reduce: MlrvalOrFunction
			reduce(227), // ==, reduce: MlrvalOrFunction
			reduce(227), // !=, reduce: MlrvalOrFunction
			reduce(227), // <=>, reduce: MlrvalOrFunction
			reduce(227), // >=, reduce: MlrvalOrFunction
			reduce(227), // <, reduce: MlrvalOrFunction
			reduce(227), // <=, reduce: MlrvalOrFunction
			reduce(227), // ^, reduce: MlrvalOrFunction
			reduce(227), // &, reduce: MlrvalOrFunction
			reduce(227), // <<, reduce: MlrvalOrFunction
			reduce(227), // >>>, reduce: MlrvalOrFunction
			reduce(227), // +, reduce: MlrvalOrFunction
			reduce(227), // -, reduce: MlrvalOrFunction
			reduce(227), // .+, reduce: MlrvalOrFunction
			reduce(227), // .-, reduce: MlrvalOrFunction
			reduce(227), // *, reduce: MlrvalOrFunction
			reduce(227), // /, reduce: MlrvalOrFunction
			reduce(227), // //, reduce: MlrvalOrFunction
			reduce(227), // %, reduce: MlrvalOrFunction
			reduce(227), // .*, reduce: MlrvalOrFunction
			reduce(227), // ./, reduce: MlrvalOrFunction
			reduce(227), // .//, reduce: MlrvalOrFunction
			reduce(227), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(227), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(227), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(227), // [[, reduce: MlrvalOrFunction
			reduce(227), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S359
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(752),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(756),  // (
			nil,         // )
			nil,         // emitp
			shift(768),  // field_name
			shift(769),  // $[
			shift(1154), // ]
			shift(771),  // braced_field_name
			shift(772),  // $[[
			shift(773),  // $[[[
			shift(774),  // full_srec
			shift(776),  // oosvar_name
			shift(777),  // @[
			shift(778),  // braced_oosvar_name
			shift(779),  // full_oosvar
			shift(780),  // all
			shift(781),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(796),  // +
			shift(798),  // -
			shift(799),  // .+
			shift(800),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(803),  // !
			shift(804),  // ~
			nil,         // **
			shift(809),  // string_literal
			shift(810),  // regex_case_insensitive
			shift(811),  // int_literal
			shift(812),  // float_literal
			shift(813),  // boolean_literal
			shift(814),  // null_literal
			shift(815),  // inf_literal
			shift(816),  // nan_literal
			shift(817),  // const_M_PI
			shift(818),  // const_M_E
			shift(819),  // panic
			shift(821),  // [
			shift(824),  // ctx_IPS
			shift(825),  // ctx_IFS
			shift(826),  // ctx_IRS
			shift(827),  // ctx_OPS
			shift(828),  // ctx_OFS
			shift(829),  // ctx_ORS
			shift(830),  // ctx_FLATSEP
			shift(831),  // ctx_NF
			shift(832),  // ctx_NR
			shift(833),  // ctx_FNR
			shift(834),  // ctx_FILENAME
			shift(835),  // ctx_FILENUM
			shift(837),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(844),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S360
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(240), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(240), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(240), // >, reduce: MlrvalOrFunction
			reduce(240), // >>, reduce: MlrvalOrFunction
			reduce(240), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(240), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(240), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(240), // ||, reduce: MlrvalOrFunction
			reduce(240), // ^^, reduce: MlrvalOrFunction
			reduce(240), // &&, reduce: MlrvalOrFunction
			reduce(240), // ??, reduce: MlrvalOrFunction
			reduce(240), // ???, reduce: MlrvalOrFunction
			reduce(240), // =~, reduce: MlrvalOrFunction
			reduce(240), // !=~, reduce: MlrvalOrFunction
			reduce(240), // ==, reduce: MlrvalOrFunction
			reduce(240), // !=, reduce: MlrvalOrFunction
			reduce(240), // <=>, reduce: MlrvalOrFunction
			reduce(240), // >=, reduce: MlrvalOrFunction
			reduce(240), // <, reduce: MlrvalOrFunction
			reduce(240), // <=, reduce: MlrvalOrFunction
			reduce(240), // ^, reduce: MlrvalOrFunction
			reduce(240), // &, reduce: MlrvalOrFunction
			reduce(240), // <<, reduce: MlrvalOrFunction
			reduce(240), // >>>, reduce: MlrvalOrFunction
			reduce(240), // +, reduce: MlrvalOrFunction
			reduce(240), // -, reduce: MlrvalOrFunction
			reduce(240), // .+, reduce: MlrvalOrFunction
			reduce(240), // .-, reduce: MlrvalOrFunction
			reduce(240), // *, reduce: MlrvalOrFunction
			reduce(240), // /, reduce: MlrvalOrFunction
			reduce(240), // //, reduce: MlrvalOrFunction
			reduce(240), // %, reduce: MlrvalOrFunction
			reduce(240), // .*, reduce: MlrvalOrFunction
			reduce(240), // ./, reduce: MlrvalOrFunction
			reduce(240), // .//, reduce: MlrvalOrFunction
			reduce(240), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(240), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(240), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(240), // [[, reduce: MlrvalOrFunction
			reduce(240), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S361
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(241), // $, reduce: ContextVariable
			nil,         // empty
			reduce(241), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(241), // >, reduce: ContextVariable
			reduce(241), // >>, reduce: ContextVariable
			reduce(241), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(241), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(241), // ?, reduce: ContextVariable
			nil,         // :
			reduce(241), // ||, reduce: ContextVariable
			reduce(241), // ^^, reduce: ContextVariable
			reduce(241), // &&, reduce: ContextVariable
			reduce(241), // ??, reduce: ContextVariable
			reduce(241), // ???, reduce: ContextVariable
			reduce(241), // =~, reduce: ContextVariable
			reduce(241), // !=~, reduce: ContextVariable
			reduce(241), // ==, reduce: ContextVariable
			reduce(241), // !=, reduce: ContextVariable
			reduce(241), // <=>, reduce: ContextVariable
			reduce(241), // >=, reduce: ContextVariable
			reduce(241), // <, reduce: ContextVariable
			reduce(241), // <=, reduce: ContextVariable
			reduce(241), // ^, reduce: ContextVariable
			reduce(241), // &, reduce: ContextVariable
			reduce(241), // <<, reduce: ContextVariable
			reduce(241), // >>>, reduce: ContextVariable
			reduce(241), // +, reduce: ContextVariable
			reduce(241), // -, reduce: ContextVariable
			reduce(241), // .+, reduce: ContextVariable
			reduce(241), // .-, reduce: ContextVariable
			reduce(241), // *, reduce: ContextVariable
			reduce(241), // /, reduce: ContextVariable
			reduce(241), // //, reduce: ContextVariable
			reduce(241), // %, reduce: ContextVariable
			reduce(241), // .*, reduce: ContextVariable
			reduce(241), // ./, reduce: ContextVariable
			reduce(241), // .//, reduce: ContextVariable
			reduce(241), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(241), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(241), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(241), // [[, reduce: ContextVariable
			reduce(241), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S362
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(242), // $, reduce: ContextVariable
			nil,         // empty
			reduce(242), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(242), // >, reduce: ContextVariable
			reduce(242), // >>, reduce: ContextVariable
			reduce(242), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(242), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(242), // ?, reduce: ContextVariable
			nil,         // :
			reduce(242), // ||, reduce: ContextVariable
			reduce(242), // ^^, reduce: ContextVariable
			reduce(242), // &&, reduce: ContextVariable
			reduce(242), // ??, reduce: ContextVariable
			reduce(242), // ???, reduce: ContextVariable
			reduce(242), // =~, reduce: ContextVariable
			reduce(242), // !=~, reduce: ContextVariable
			reduce(242), // ==, reduce: ContextVariable
			reduce(242), // !=, reduce: ContextVariable
			reduce(242), // <=>, reduce: ContextVariable
			reduce(242), // >=, reduce: ContextVariable
			reduce(242), // <, reduce: ContextVariable
			reduce(242), // <=, reduce: ContextVariable
			reduce(242), // ^, reduce: ContextVariable
			reduce(242), // &, reduce: ContextVariable
			reduce(242), // <<, reduce: ContextVariable
			reduce(242), // >>>, reduce: ContextVariable
			reduce(242), // +, reduce: ContextVariable
			reduce(242), // -, reduce: ContextVariable
			reduce(242), // .+, reduce: ContextVariable
			reduce(242), // .-, reduce: ContextVariable
			reduce(242), // *, reduce: ContextVariable
			reduce(242), // /, reduce: ContextVariable
			reduce(242), // //, reduce: ContextVariable
			reduce(242), // %, reduce: ContextVariable
			reduce(242), // .*, reduce: ContextVariable
			reduce(242), // ./, reduce: ContextVariable
			reduce(242), // .//, reduce: ContextVariable
			reduce(242), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(242), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(242), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(242), // [[, reduce: ContextVariable
			reduce(242), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S363
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(243), // $, reduce: ContextVariable
			nil,         // empty
			reduce(243), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(243), // >, reduce: ContextVariable
			reduce(243), // >>, reduce: ContextVariable
			reduce(243), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(243), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(243), // ?, reduce: ContextVariable
			nil,         // :
			reduce(243), // ||, reduce: ContextVariable
			reduce(243), // ^^, reduce: ContextVariable
			reduce(243), // &&, reduce: ContextVariable
			reduce(243), // ??, reduce: ContextVariable
			reduce(243), // ???, reduce: ContextVariable
			reduce(243), // =~, reduce: ContextVariable
			reduce(243), // !=~, reduce: ContextVariable
			reduce(243), // ==, reduce: ContextVariable
			reduce(243), // !=, reduce: ContextVariable
			reduce(243), // <=>, reduce: ContextVariable
			reduce(243), // >=, reduce: ContextVariable
			reduce(243), // <, reduce: ContextVariable
			reduce(243), // <=, reduce: ContextVariable
			reduce(243), // ^, reduce: ContextVariable
			reduce(243), // &, reduce: ContextVariable
			reduce(243), // <<, reduce: ContextVariable
			reduce(243), // >>>, reduce: ContextVariable
			reduce(243), // +, reduce: ContextVariable
			reduce(243), // -, reduce: ContextVariable
			reduce(243), // .+, reduce: ContextVariable
			reduce(243), // .-, reduce: ContextVariable
			reduce(243), // *, reduce: ContextVariable
			reduce(243), // /, reduce: ContextVariable
			reduce(243), // //, reduce: ContextVariable
			reduce(243), // %, reduce: ContextVariable
			reduce(243), // .*, reduce: ContextVariable
			reduce(243), // ./, reduce: ContextVariable
			reduce(243), // .//, reduce: ContextVariable
			reduce(243), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(243), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(243), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(243), // [[, reduce: ContextVariable
			reduce(243), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S364
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(244), // $, reduce: ContextVariable
			nil,         // empty
			reduce(244), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(244), // >, reduce: ContextVariable
			reduce(244), // >>, reduce: ContextVariable
			reduce(244), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(244), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(244), // ?, reduce: ContextVariable
			nil,         // :
			reduce(244), // ||, reduce: ContextVariable
			reduce(244), // ^^, reduce: ContextVariable
			reduce(244), // &&, reduce: ContextVariable
			reduce(244), // ??, reduce: ContextVariable
			reduce(244), // ???, reduce: ContextVariable
			reduce(244), // =~, reduce: ContextVariable
			reduce(244), // !=~, reduce: ContextVariable
			reduce(244), // ==, reduce: ContextVariable
			reduce(244), // !=, reduce: ContextVariable
			reduce(244), // <=>, reduce: ContextVariable
			reduce(244), // >=, reduce: ContextVariable
			reduce(244), // <, reduce: ContextVariable
			reduce(244), // <=, reduce: ContextVariable
			reduce(244), // ^, reduce: ContextVariable
			reduce(244), // &, reduce: ContextVariable
			reduce(244), // <<, reduce: ContextVariable
			reduce(244), // >>>, reduce: ContextVariable
			reduce(244), // +, reduce: ContextVariable
			reduce(244), // -, reduce: ContextVariable
			reduce(244), // .+, reduce: ContextVariable
			reduce(244), // .-, reduce: ContextVariable
			reduce(244), // *, reduce: ContextVariable
			reduce(244), // /, reduce: ContextVariable
			reduce(244), // //, reduce: ContextVariable
			reduce(244), // %, reduce: ContextVariable
			reduce(244), // .*, reduce: ContextVariable
			reduce(244), // ./, reduce: ContextVariable
			reduce(244), // .//, reduce: ContextVariable
			reduce(244), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(244), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(244), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(244), // [[, reduce: ContextVariable
			reduce(244), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S365
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(245), // $, reduce: ContextVariable
			nil,         // empty
			reduce(245), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(245), // >, reduce: ContextVariable
			reduce(245), // >>, reduce: ContextVariable
			reduce(245), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(245), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(245), // ?, reduce: ContextVariable
			nil,         // :
			reduce(245), // ||, reduce: ContextVariable
			reduce(245), // ^^, reduce: ContextVariable
			reduce(245), // &&, reduce: ContextVariable
			reduce(245), // ??, reduce: ContextVariable
			reduce(245), // ???, reduce: ContextVariable
			reduce(245), // =~, reduce: ContextVariable
			reduce(245), // !=~, reduce: ContextVariable
			reduce(245), // ==, reduce: ContextVariable
			reduce(245), // !=, reduce: ContextVariable
			reduce(245), // <=>, reduce: ContextVariable
			reduce(245), // >=, reduce: ContextVariable
			reduce(245), // <, reduce: ContextVariable
			reduce(245), // <=, reduce: ContextVariable
			reduce(245), // ^, reduce: ContextVariable
			reduce(245), // &, reduce: ContextVariable
			reduce(245), // <<, reduce: ContextVariable
			reduce(245), // >>>, reduce: ContextVariable
			reduce(245), // +, reduce: ContextVariable
			reduce(245), // -, reduce: ContextVariable
			reduce(245), // .+, reduce: ContextVariable
			reduce(245), // .-, reduce: ContextVariable
			reduce(245), // *, reduce: ContextVariable
			reduce(245), // /, reduce: ContextVariable
			reduce(245), // //, reduce: ContextVariable
			reduce(245), // %, reduce: ContextVariable
			reduce(245), // .*, reduce: ContextVariable
			reduce(245), // ./, reduce: ContextVariable
			reduce(245), // .//, reduce: ContextVariable
			reduce(245), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(245), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(245), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(245), // [[, reduce: ContextVariable
			reduce(245), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S366
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(246), // $, reduce: ContextVariable
			nil,         // empty
			reduce(246), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(246), // >, reduce: ContextVariable
			reduce(246), // >>, reduce: ContextVariable
			reduce(246), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(246), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(246), // ?, reduce: ContextVariable
			nil,         // :
			reduce(246), // ||, reduce: ContextVariable
			reduce(246), // ^^, reduce: ContextVariable
			reduce(246), // &&, reduce: ContextVariable
			reduce(246), // ??, reduce: ContextVariable
			reduce(246), // ???, reduce: ContextVariable
			reduce(246), // =~, reduce: ContextVariable
			reduce(246), // !=~, reduce: ContextVariable
			reduce(246), // ==, reduce: ContextVariable
			reduce(246), // !=, reduce: ContextVariable
			reduce(246), // <=>, reduce: ContextVariable
			reduce(246), // >=, reduce: ContextVariable
			reduce(246), // <, reduce: ContextVariable
			reduce(246), // <=, reduce: ContextVariable
			reduce(246), // ^, reduce: ContextVariable
			reduce(246), // &, reduce: ContextVariable
			reduce(246), // <<, reduce: ContextVariable
			reduce(246), // >>>, reduce: ContextVariable
			reduce(246), // +, reduce: ContextVariable
			reduce(246), // -, reduce: ContextVariable
			reduce(246), // .+, reduce: ContextVariable
			reduce(246), // .-, reduce: ContextVariable
			reduce(246), // *, reduce: ContextVariable
			reduce(246), // /, reduce: ContextVariable
			reduce(246), // //, reduce: ContextVariable
			reduce(246), // %, reduce: ContextVariable
			reduce(246), // .*, reduce: ContextVariable
			reduce(246), // ./, reduce: ContextVariable
			reduce(246), // .//, reduce: ContextVariable
			reduce(246), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(246), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(246), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(246), // [[, reduce: ContextVariable
			reduce(246), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S367
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(247), // $, reduce: ContextVariable
			nil,         // empty
			reduce(247), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(247), // >, reduce: ContextVariable
			reduce(247), // >>, reduce: ContextVariable
			reduce(247), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(247), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(247), // ?, reduce: ContextVariable
			nil,         // :
			reduce(247), // ||, reduce: ContextVariable
			reduce(247), // ^^, reduce: ContextVariable
			reduce(247), // &&, reduce: ContextVariable
			reduce(247), // ??, reduce: ContextVariable
			reduce(247), // ???, reduce: ContextVariable
			reduce(247), // =~, reduce: ContextVariable
			reduce(247), // !=~, reduce: ContextVariable
			reduce(247), // ==, reduce: ContextVariable
			reduce(247), // !=, reduce: ContextVariable
			reduce(247), // <=>, reduce: ContextVariable
			reduce(247), // >=, reduce: ContextVariable
			reduce(247), // <, reduce: ContextVariable
			reduce(247), // <=, reduce: ContextVariable
			reduce(247), // ^, reduce: ContextVariable
			reduce(247), // &, reduce: ContextVariable
			reduce(247), // <<, reduce: ContextVariable
			reduce(247), // >>>, reduce: ContextVariable
			reduce(247), // +, reduce: ContextVariable
			reduce(247), // -, reduce: ContextVariable
			reduce(247), // .+, reduce: ContextVariable
			reduce(247), // .-, reduce: ContextVariable
			reduce(247), // *, reduce: ContextVariable
			reduce(247), // /, reduce: ContextVariable
			reduce(247), // //, reduce: ContextVariable
			reduce(247), // %, reduce: ContextVariable
			reduce(247), // .*, reduce: ContextVariable
			reduce(247), // ./, reduce: ContextVariable
			reduce(247), // .//, reduce: ContextVariable
			reduce(247), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(247), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(247), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(247), // [[, reduce: ContextVariable
			reduce(247), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S368
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(248), // $, reduce: ContextVariable
			nil,         // empty
			reduce(248), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(248), // >, reduce: ContextVariable
			reduce(248), // >>, reduce: ContextVariable
			reduce(248), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(248), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(248), // ?, reduce: ContextVariable
			nil,         // :
			reduce(248), // ||, reduce: ContextVariable
			reduce(248), // ^^, reduce: ContextVariable
			reduce(248), // &&, reduce: ContextVariable
			reduce(248), // ??, reduce: ContextVariable
			reduce(248), // ???, reduce: ContextVariable
			reduce(248), // =~, reduce: ContextVariable
			reduce(248), // !=~, reduce: ContextVariable
			reduce(248), // ==, reduce: ContextVariable
			reduce(248), // !=, reduce: ContextVariable
			reduce(248), // <=>, reduce: ContextVariable
			reduce(248), // >=, reduce: ContextVariable
			reduce(248), // <, reduce: ContextVariable
			reduce(248), // <=, reduce: ContextVariable
			reduce(248), // ^, reduce: ContextVariable
			reduce(248), // &, reduce: ContextVariable
			reduce(248), // <<, reduce: ContextVariable
			reduce(248), // >>>, reduce: ContextVariable
			reduce(248), // +, reduce: ContextVariable
			reduce(248), // -, reduce: ContextVariable
			reduce(248), // .+, reduce: ContextVariable
			reduce(248), // .-, reduce: ContextVariable
			reduce(248), // *, reduce: ContextVariable
			reduce(248), // /, reduce: ContextVariable
			reduce(248), // //, reduce: ContextVariable
			reduce(248), // %, reduce: ContextVariable
			reduce(248), // .*, reduce: ContextVariable
			reduce(248), // ./, reduce: ContextVariable
			reduce(248), // .//, reduce: ContextVariable
			reduce(248), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(248), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(248), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(248), // [[, reduce: ContextVariable
			reduce(248), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S369
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(249), // $, reduce: ContextVariable
			nil,         // empty
			reduce(249), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(249), // >, reduce: ContextVariable
			reduce(249), // >>, reduce: ContextVariable
			reduce(249), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(249), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(249), // ?, reduce: ContextVariable
			nil,         // :
			reduce(249), // ||, reduce: ContextVariable
			reduce(249), // ^^, reduce: ContextVariable
			reduce(249), // &&, reduce: ContextVariable
			reduce(249), // ??, reduce: ContextVariable
			reduce(249), // ???, reduce: ContextVariable
			reduce(249), // =~, reduce: ContextVariable
			reduce(249), // !=~, reduce: ContextVariable
			reduce(249), // ==, reduce: ContextVariable
			reduce(249), // !=, reduce: ContextVariable
			reduce(249), // <=>, reduce: ContextVariable
			reduce(249), // >=, reduce: ContextVariable
			reduce(249), // <, reduce: ContextVariable
			reduce(249), // <=, reduce: ContextVariable
			reduce(249), // ^, reduce: ContextVariable
			reduce(249), // &, reduce: ContextVariable
			reduce(249), // <<, reduce: ContextVariable
			reduce(249), // >>>, reduce: ContextVariable
			reduce(249), // +, reduce: ContextVariable
			reduce(249), // -, reduce: ContextVariable
			reduce(249), // .+, reduce: ContextVariable
			reduce(249), // .-, reduce: ContextVariable
			reduce(249), // *, reduce: ContextVariable
			reduce(249), // /, reduce: ContextVariable
			reduce(249), // //, reduce: ContextVariable
			reduce(249), // %, reduce: ContextVariable
			reduce(249), // .*, reduce: ContextVariable
			reduce(249), // ./, reduce: ContextVariable
			reduce(249), // .//, reduce: ContextVariable
			reduce(249), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(249), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(249), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(249), // [[, reduce: ContextVariable
			reduce(249), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S370
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(250), // $, reduce: ContextVariable
			nil,         // empty
			reduce(250), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(250), // >, reduce: ContextVariable
			reduce(250), // >>, reduce: ContextVariable
			reduce(250), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(250), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(250), // ?, reduce: ContextVariable
			nil,         // :
			reduce(250), // ||, reduce: ContextVariable
			reduce(250), // ^^, reduce: ContextVariable
			reduce(250), // &&, reduce: ContextVariable
			reduce(250), // ??, reduce: ContextVariable
			reduce(250), // ???, reduce: ContextVariable
			reduce(250), // =~, reduce: ContextVariable
			reduce(250), // !=~, reduce: ContextVariable
			reduce(250), // ==, reduce: ContextVariable
			reduce(250), // !=, reduce: ContextVariable
			reduce(250), // <=>, reduce: ContextVariable
			reduce(250), // >=, reduce: ContextVariable
			reduce(250), // <, reduce: ContextVariable
			reduce(250), // <=, reduce: ContextVariable
			reduce(250), // ^, reduce: ContextVariable
			reduce(250), // &, reduce: ContextVariable
			reduce(250), // <<, reduce: ContextVariable
			reduce(250), // >>>, reduce: ContextVariable
			reduce(250), // +, reduce: ContextVariable
			reduce(250), // -, reduce: ContextVariable
			reduce(250), // .+, reduce: ContextVariable
			reduce(250), // .-, reduce: ContextVariable
			reduce(250), // *, reduce: ContextVariable
			reduce(250), // /, reduce: ContextVariable
			reduce(250), // //, reduce: ContextVariable
			reduce(250), // %, reduce: ContextVariable
			reduce(250), // .*, reduce: ContextVariable
			reduce(250), // ./, reduce: ContextVariable
			reduce(250), // .//, reduce: ContextVariable
			reduce(250), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(250), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(250), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(250), // [[, reduce: ContextVariable
			reduce(250), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S371
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(251), // $, reduce: ContextVariable
			nil,         // empty
			reduce(251), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(251), // >, reduce: ContextVariable
			reduce(251), // >>, reduce: ContextVariable
			reduce(251), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(251), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(251), // ?, reduce: ContextVariable
			nil,         // :
			reduce(251), // ||, reduce: ContextVariable
			reduce(251), // ^^, reduce: ContextVariable
			reduce(251), // &&, reduce: ContextVariable
			reduce(251), // ??, reduce: ContextVariable
			reduce(251), // ???, reduce: ContextVariable
			reduce(251), // =~, reduce: ContextVariable
			reduce(251), // !=~, reduce: ContextVariable
			reduce(251), // ==, reduce: ContextVariable
			reduce(251), // !=, reduce: ContextVariable
			reduce(251), // <=>, reduce: ContextVariable
			reduce(251), // >=, reduce: ContextVariable
			reduce(251), // <, reduce: ContextVariable
			reduce(251), // <=, reduce: ContextVariable
			reduce(251), // ^, reduce: ContextVariable
			reduce(251), // &, reduce: ContextVariable
			reduce(251), // <<, reduce: ContextVariable
			reduce(251), // >>>, reduce: ContextVariable
			reduce(251), // +, reduce: ContextVariable
			reduce(251), // -, reduce: ContextVariable
			reduce(251), // .+, reduce: ContextVariable
			reduce(251), // .-, reduce: ContextVariable
			reduce(251), // *, reduce: ContextVariable
			reduce(251), // /, reduce: ContextVariable
			reduce(251), // //, reduce: ContextVariable
			reduce(251), // %, reduce: ContextVariable
			reduce(251), // .*, reduce: ContextVariable
			reduce(251), // ./, reduce: ContextVariable
			reduce(251), // .//, reduce: ContextVariable
			reduce(251), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(251), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(251), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(251), // [[, reduce: ContextVariable
			reduce(251), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S372
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(252), // $, reduce: ContextVariable
			nil,         // empty
			reduce(252), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(252), // >, reduce: ContextVariable
			reduce(252), // >>, reduce: ContextVariable
			reduce(252), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(252), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(252), // ?, reduce: ContextVariable
			nil,         // :
			reduce(252), // ||, reduce: ContextVariable
			reduce(252), // ^^, reduce: ContextVariable
			reduce(252), // &&, reduce: ContextVariable
			reduce(252), // ??, reduce: ContextVariable
			reduce(252), // ???, reduce: ContextVariable
			reduce(252), // =~, reduce: ContextVariable
			reduce(252), // !=~, reduce: ContextVariable
			reduce(252), // ==, reduce: ContextVariable
			reduce(252), // !=, reduce: ContextVariable
			reduce(252), // <=>, reduce: ContextVariable
			reduce(252), // >=, reduce: ContextVariable
			reduce(252), // <, reduce: ContextVariable
			reduce(252), // <=, reduce: ContextVariable
			reduce(252), // ^, reduce: ContextVariable
			reduce(252), // &, reduce: ContextVariable
			reduce(252), // <<, reduce: ContextVariable
			reduce(252), // >>>, reduce: ContextVariable
			reduce(252), // +, reduce: ContextVariable
			reduce(252), // -, reduce: ContextVariable
			reduce(252), // .+, reduce: ContextVariable
			reduce(252), // .-, reduce: ContextVariable
			reduce(252), // *, reduce: ContextVariable
			reduce(252), // /, reduce: ContextVariable
			reduce(252), // //, reduce: ContextVariable
			reduce(252), // %, reduce: ContextVariable
			reduce(252), // .*, reduce: ContextVariable
			reduce(252), // ./, reduce: ContextVariable
			reduce(252), // .//, reduce: ContextVariable
			reduce(252), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(252), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(252), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(252), // [[, reduce: ContextVariable
			reduce(252), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S373
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(253), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(253), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(253), // >, reduce: MlrvalOrFunction
			reduce(253), // >>, reduce: MlrvalOrFunction
			reduce(253), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(253), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(253), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(253), // ||, reduce: MlrvalOrFunction
			reduce(253), // ^^, reduce: MlrvalOrFunction
			reduce(253), // &&, reduce: MlrvalOrFunction
			reduce(253), // ??, reduce: MlrvalOrFunction
			reduce(253), // ???, reduce: MlrvalOrFunction
			reduce(253), // =~, reduce: MlrvalOrFunction
			reduce(253), // !=~, reduce: MlrvalOrFunction
			reduce(253), // ==, reduce: MlrvalOrFunction
			reduce(253), // !=, reduce: MlrvalOrFunction
			reduce(253), // <=>, reduce: MlrvalOrFunction
			reduce(253), // >=, reduce: MlrvalOrFunction
			reduce(253), // <, reduce: MlrvalOrFunction
			reduce(253), // <=, reduce: MlrvalOrFunction
			reduce(253), // ^, reduce: MlrvalOrFunction
			reduce(253), // &, reduce: MlrvalOrFunction
			reduce(253), // <<, reduce: MlrvalOrFunction
			reduce(253), // >>>, reduce: MlrvalOrFunction
			reduce(253), // +, reduce: MlrvalOrFunction
			reduce(253), // -, reduce: MlrvalOrFunction
			reduce(253), // .+, reduce: MlrvalOrFunction
			reduce(253), // .-, reduce: MlrvalOrFunction
			reduce(253), // *, reduce: MlrvalOrFunction
			reduce(253), // /, reduce: MlrvalOrFunction
			reduce(253), // //, reduce: MlrvalOrFunction
			reduce(253), // %, reduce: MlrvalOrFunction
			reduce(253), // .*, reduce: MlrvalOrFunction
			reduce(253), // ./, reduce: MlrvalOrFunction
			reduce(253), // .//, reduce: MlrvalOrFunction
			reduce(253), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(253), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(253), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(253), // [[, reduce: MlrvalOrFunction
			reduce(253), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S374
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			shift(1156), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(1157), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S375
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(256), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(256), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(256), // >, reduce: MlrvalOrFunction
			reduce(256), // >>, reduce: MlrvalOrFunction
			reduce(256), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(256), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(256), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(256), // ||, reduce: MlrvalOrFunction
			reduce(256), // ^^, reduce: MlrvalOrFunction
			reduce(256), // &&, reduce: MlrvalOrFunction
			reduce(256), // ??, reduce: MlrvalOrFunction
			reduce(256), // ???, reduce: MlrvalOrFunction
			reduce(256), // =~, reduce: MlrvalOrFunction
			reduce(256), // !=~, reduce: MlrvalOrFunction
			reduce(256), // ==, reduce: MlrvalOrFunction
			reduce(256), // !=, reduce: MlrvalOrFunction
			reduce(256), // <=>, reduce: MlrvalOrFunction
			reduce(256), // >=, reduce: MlrvalOrFunction
			reduce(256), // <, reduce: MlrvalOrFunction
			reduce(256), // <=, reduce: MlrvalOrFunction
			reduce(256), // ^, reduce: MlrvalOrFunction
			reduce(256), // &, reduce: MlrvalOrFunction
			reduce(256), // <<, reduce: MlrvalOrFunction
			reduce(256), // >>>, reduce: MlrvalOrFunction
			reduce(256), // +, reduce: MlrvalOrFunction
			reduce(256), // -, reduce: MlrvalOrFunction
			reduce(256), // .+, reduce: MlrvalOrFunction
			reduce(256), // .-, reduce: MlrvalOrFunction
			reduce(256), // *, reduce: MlrvalOrFunction
			reduce(256), // /, reduce: MlrvalOrFunction
			reduce(256), // //, reduce: MlrvalOrFunction
			reduce(256), // %, reduce: MlrvalOrFunction
			reduce(256), // .*, reduce: MlrvalOrFunction
			reduce(256), // ./, reduce: MlrvalOrFunction
			reduce(256), // .//, reduce: MlrvalOrFunction
			reduce(256), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(256), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(256), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(256), // [[, reduce: MlrvalOrFunction
			reduce(256), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S376
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(257), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(257), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(257), // >, reduce: MlrvalOrFunction
			reduce(257), // >>, reduce: MlrvalOrFunction
			reduce(257), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(257), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(257), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(257), // ||, reduce: MlrvalOrFunction
			reduce(257), // ^^, reduce: MlrvalOrFunction
			reduce(257), // &&, reduce: MlrvalOrFunction
			reduce(257), // ??, reduce: MlrvalOrFunction
			reduce(257), // ???, reduce: MlrvalOrFunction
			reduce(257), // =~, reduce: MlrvalOrFunction
			reduce(257), // !=~, reduce: MlrvalOrFunction
			reduce(257), // ==, reduce: MlrvalOrFunction
			reduce(257), // !=, reduce: MlrvalOrFunction
			reduce(257), // <=>, reduce: MlrvalOrFunction
			reduce(257), // >=, reduce: MlrvalOrFunction
			reduce(257), // <, reduce: MlrvalOrFunction
			reduce(257), // <=, reduce: MlrvalOrFunction
			reduce(257), // ^, reduce: MlrvalOrFunction
			reduce(257), // &, reduce: MlrvalOrFunction
			reduce(257), // <<, reduce: MlrvalOrFunction
			reduce(257), // >>>, reduce: MlrvalOrFunction
			reduce(257), // +, reduce: MlrvalOrFunction
			reduce(257), // -, reduce: MlrvalOrFunction
			reduce(257), // .+, reduce: MlrvalOrFunction
			reduce(257), // .-, reduce: MlrvalOrFunction
			reduce(257), // *, reduce: MlrvalOrFunction
			reduce(257), // /, reduce: MlrvalOrFunction
			reduce(257), // //, reduce: MlrvalOrFunction
			reduce(257), // %, reduce: MlrvalOrFunction
			reduce(257), // .*, reduce: MlrvalOrFunction
			reduce(257), // ./, reduce: MlrvalOrFunction
			reduce(257), // .//, reduce: MlrvalOrFunction
			reduce(257), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(257), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(257), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(257), // [[, reduce: MlrvalOrFunction
			reduce(257), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S377
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(258), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(258), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(258), // >, reduce: MlrvalOrFunction
			reduce(258), // >>, reduce: MlrvalOrFunction
			reduce(258), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(258), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(258), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(258), // ||, reduce: MlrvalOrFunction
			reduce(258), // ^^, reduce: MlrvalOrFunction
			reduce(258), // &&, reduce: MlrvalOrFunction
			reduce(258), // ??, reduce: MlrvalOrFunction
			reduce(258), // ???, reduce: MlrvalOrFunction
			reduce(258), // =~, reduce: MlrvalOrFunction
			reduce(258), // !=~, reduce: MlrvalOrFunction
			reduce(258), // ==, reduce: MlrvalOrFunction
			reduce(258), // !=, reduce: MlrvalOrFunction
			reduce(258), // <=>, reduce: MlrvalOrFunction
			reduce(258), // >=, reduce: MlrvalOrFunction
			reduce(258), // <, reduce: MlrvalOrFunction
			reduce(258), // <=, reduce: MlrvalOrFunction
			reduce(258), // ^, reduce: MlrvalOrFunction
			reduce(258), // &, reduce: MlrvalOrFunction
			reduce(258), // <<, reduce: MlrvalOrFunction
			reduce(258), // >>>, reduce: MlrvalOrFunction
			reduce(258), // +, reduce: MlrvalOrFunction
			reduce(258), // -, reduce: MlrvalOrFunction
			reduce(258), // .+, reduce: MlrvalOrFunction
			reduce(258), // .-, reduce: MlrvalOrFunction
			reduce(258), // *, reduce: MlrvalOrFunction
			reduce(258), // /, reduce: MlrvalOrFunction
			reduce(258), // //, reduce: MlrvalOrFunction
			reduce(258), // %, reduce: MlrvalOrFunction
			reduce(258), // .*, reduce: MlrvalOrFunction
			reduce(258), // ./, reduce: MlrvalOrFunction
			reduce(258), // .//, reduce: MlrvalOrFunction
			reduce(258), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(258), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(258), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(258), // [[, reduce: MlrvalOrFunction
			reduce(258), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S378
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(259), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(259), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(259), // >, reduce: MlrvalOrFunction
			reduce(259), // >>, reduce: MlrvalOrFunction
			reduce(259), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(259), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(259), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(259), // ||, reduce: MlrvalOrFunction
			reduce(259), // ^^, reduce: MlrvalOrFunction
			reduce(259), // &&, reduce: MlrvalOrFunction
			reduce(259), // ??, reduce: MlrvalOrFunction
			reduce(259), // ???, reduce: MlrvalOrFunction
			reduce(259), // =~, reduce: MlrvalOrFunction
			reduce(259), // !=~, reduce: MlrvalOrFunction
			reduce(259), // ==, reduce: MlrvalOrFunction
			reduce(259), // !=, reduce: MlrvalOrFunction
			reduce(259), // <=>, reduce: MlrvalOrFunction
			reduce(259), // >=, reduce: MlrvalOrFunction
			reduce(259), // <, reduce: MlrvalOrFunction
			reduce(259), // <=, reduce: MlrvalOrFunction
			reduce(259), // ^, reduce: MlrvalOrFunction
			reduce(259), // &, reduce: MlrvalOrFunction
			reduce(259), // <<, reduce: MlrvalOrFunction
			reduce(259), // >>>, reduce: MlrvalOrFunction
			reduce(259), // +, reduce: MlrvalOrFunction
			reduce(259), // -, reduce: MlrvalOrFunction
			reduce(259), // .+, reduce: MlrvalOrFunction
			reduce(259), // .-, reduce: MlrvalOrFunction
			reduce(259), // *, reduce: MlrvalOrFunction
			reduce(259), // /, reduce: MlrvalOrFunction
			reduce(259), // //, reduce: MlrvalOrFunction
			reduce(259), // %, reduce: MlrvalOrFunction
			reduce(259), // .*, reduce: MlrvalOrFunction
			reduce(259), // ./, reduce: MlrvalOrFunction
			reduce(259), // .//, reduce: MlrvalOrFunction
			reduce(259), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(259), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(259), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(259), // [[, reduce: MlrvalOrFunction
			reduce(259), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S379
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(267), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(267), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(267), // >, reduce: MlrvalOrFunction
			reduce(267), // >>, reduce: MlrvalOrFunction
			reduce(267), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(267), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(267), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(267), // ||, reduce: MlrvalOrFunction
			reduce(267), // ^^, reduce: MlrvalOrFunction
			reduce(267), // &&, reduce: MlrvalOrFunction
			reduce(267), // ??, reduce: MlrvalOrFunction
			reduce(267), // ???, reduce: MlrvalOrFunction
			reduce(267), // =~, reduce: MlrvalOrFunction
			reduce(267), // !=~, reduce: MlrvalOrFunction
			reduce(267), // ==, reduce: MlrvalOrFunction
			reduce(267), // !=, reduce: MlrvalOrFunction
			reduce(267), // <=>, reduce: MlrvalOrFunction
			reduce(267), // >=, reduce: MlrvalOrFunction
			reduce(267), // <, reduce: MlrvalOrFunction
			reduce(267), // <=, reduce: MlrvalOrFunction
			reduce(267), // ^, reduce: MlrvalOrFunction
			reduce(267), // &, reduce: MlrvalOrFunction
			reduce(267), // <<, reduce: MlrvalOrFunction
			reduce(267), // >>>, reduce: MlrvalOrFunction
			reduce(267), // +, reduce: MlrvalOrFunction
			reduce(267), // -, reduce: MlrvalOrFunction
			reduce(267), // .+, reduce: MlrvalOrFunction
			reduce(267), // .-, reduce: MlrvalOrFunction
			reduce(267), // *, reduce: MlrvalOrFunction
			reduce(267), // /, reduce: MlrvalOrFunction
			reduce(267), // //, reduce: MlrvalOrFunction
			reduce(267), // %, reduce: MlrvalOrFunction
			reduce(267), // .*, reduce: MlrvalOrFunction
			reduce(267), // ./, reduce: MlrvalOrFunction
			reduce(267), // .//, reduce: MlrvalOrFunction
			reduce(267), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(267), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(267), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(267), // [[, reduce: MlrvalOrFunction
			reduce(267), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S380
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1158), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S381
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1159), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S382
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			reduce(37), // =, reduce: Lvalue
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			reduce(37), // ||=, reduce: Lvalue
			reduce(37), // ^^=, reduce: Lvalue
			reduce(37), // &&=, reduce: Lvalue
			reduce(37), // ??=, reduce: Lvalue
			reduce(37), // ???=, reduce: Lvalue
			reduce(37), // |=, reduce: Lvalue
			reduce(37), // &=, reduce: Lvalue
			reduce(37), // ^=, reduce: Lvalue
			reduce(37), // <<=, reduce: Lvalue
			reduce(37), // >>=, reduce: Lvalue
			reduce(37), // >>>=, reduce: Lvalue
			reduce(37), // +=, reduce: Lvalue
			reduce(37), // .=, reduce: Lvalue
			reduce(37), // -=, reduce: Lvalue
			reduce(37), // *=, reduce: Lvalue
			reduce(37), // /=, reduce: Lvalue
			reduce(37), // //=, reduce: Lvalue
			reduce(37), // %=, reduce: Lvalue
			reduce(37), // **=, reduce: Lvalue
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S383
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(118), // =, reduce: LocalVariable
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(118), // ||=, reduce: LocalVariable
			reduce(118), // ^^=, reduce: LocalVariable
			reduce(118), // &&=, reduce: LocalVariable
			reduce(118), // ??=, reduce: LocalVariable
			reduce(118), // ???=, reduce: LocalVariable
			reduce(118), // |=, reduce: LocalVariable
			reduce(118), // &=, reduce: LocalVariable
			reduce(118), // ^=, reduce: LocalVariable
			reduce(118), // <<=, reduce: LocalVariable
			reduce(118), // >>=, reduce: LocalVariable
			reduce(118), // >>>=, reduce: LocalVariable
			reduce(118), // +=, reduce: LocalVariable
			reduce(118), // .=, reduce: LocalVariable
			reduce(118), // -=, reduce: LocalVariable
			reduce(118), // *=, reduce: LocalVariable
			reduce(118), // /=, reduce: LocalVariable
			reduce(118), // //=, reduce: LocalVariable
			reduce(118), // %=, reduce: LocalVariable
			reduce(118), // **=, reduce: LocalVariable
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S384
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			shift(1160), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S385
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(39), // $, reduce: FilterStatement
			nil,        // empty
			reduce(39), // ;, reduce: FilterStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S386
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(214), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(214), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(214), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S387
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(211), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(211), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(211), // >, reduce: MlrvalOrFunction
			reduce(211), // >>, reduce: MlrvalOrFunction
			reduce(211), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(211), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(211), // ||, reduce: MlrvalOrFunction
			reduce(211), // ^^, reduce: MlrvalOrFunction
			reduce(211), // &&, reduce: MlrvalOrFunction
			reduce(211), // ??, reduce: MlrvalOrFunction
			reduce(211), // ???, reduce: MlrvalOrFunction
			reduce(211), // =~, reduce: MlrvalOrFunction
			reduce(211), // !=~, reduce: MlrvalOrFunction
			reduce(211), // ==, reduce: MlrvalOrFunction
			reduce(211), // !=, reduce: MlrvalOrFunction
			reduce(211), // <=>, reduce: MlrvalOrFunction
			reduce(211), // >=, reduce: MlrvalOrFunction
			reduce(211), // <, reduce: MlrvalOrFunction
			reduce(211), // <=, reduce: MlrvalOrFunction
			reduce(211), // ^, reduce: MlrvalOrFunction
			reduce(211), // &, reduce: MlrvalOrFunction
			reduce(211), // <<, reduce: MlrvalOrFunction
			reduce(211), // >>>, reduce: MlrvalOrFunction
			reduce(211), // +, reduce: MlrvalOrFunction
			reduce(211), // -, reduce: MlrvalOrFunction
			reduce(211), // .+, reduce: MlrvalOrFunction
			reduce(211), // .-, reduce: MlrvalOrFunction
			reduce(211), // *, reduce: MlrvalOrFunction
			reduce(211), // /, reduce: MlrvalOrFunction
			reduce(211), // //, reduce: MlrvalOrFunction
			reduce(211), // %, reduce: MlrvalOrFunction
			reduce(211), // .*, reduce: MlrvalOrFunction
			reduce(211), // ./, reduce: MlrvalOrFunction
			reduce(211), // .//, reduce: MlrvalOrFunction
			reduce(211), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(211), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(211), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(211), // [[, reduce: MlrvalOrFunction
			reduce(211), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S388
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S389
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(110), // $, reduce: OosvarValue
			nil,         // empty
			reduce(110), // ;, reduce: OosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(110), // >, reduce: OosvarValue
			reduce(110), // >>, reduce: OosvarValue
			reduce(110), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(110), // ?, reduce: OosvarValue
			nil,         // :
			reduce(110), // ||, reduce: OosvarValue
			reduce(110), // ^^, reduce: OosvarValue
			reduce(110), // &&, reduce: OosvarValue
			reduce(110), // ??, reduce: OosvarValue
			reduce(110), // ???, reduce: OosvarValue
			reduce(110), // =~, reduce: OosvarValue
			reduce(110), // !=~, reduce: OosvarValue
			reduce(110), // ==, reduce: OosvarValue
			reduce(110), // !=, reduce: OosvarValue
			reduce(110), // <=>, reduce: OosvarValue
			reduce(110), // >=, reduce: OosvarValue
			reduce(110), // <, reduce: OosvarValue
			reduce(110), // <=, reduce: OosvarValue
			reduce(110), // ^, reduce: OosvarValue
			reduce(110), // &, reduce: OosvarValue
			reduce(110), // <<, reduce: OosvarValue
			reduce(110), // >>>, reduce: OosvarValue
			reduce(110), // +, reduce: OosvarValue
			reduce(110), // -, reduce: OosvarValue
			reduce(110), // .+, reduce: OosvarValue
			reduce(110), // .-, reduce: OosvarValue
			reduce(110), // *, reduce: OosvarValue
			reduce(110), // /, reduce: OosvarValue
			reduce(110), // //, reduce: OosvarValue
			reduce(110), // %, reduce: OosvarValue
			reduce(110), // .*, reduce: OosvarValue
			reduce(110), // ./, reduce: OosvarValue
			reduce(110), // .//, reduce: OosvarValue
			reduce(110), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(110), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(110), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(110), // [[, reduce: OosvarValue
			reduce(110), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S390
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(112), // $, reduce: OosvarValue
			nil,         // empty
			reduce(112), // ;, reduce: OosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(112), // >, reduce: OosvarValue
			reduce(112), // >>, reduce: OosvarValue
			reduce(112), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(112), // ?, reduce: OosvarValue
			nil,         // :
			reduce(112), // ||, reduce: OosvarValue
			reduce(112), // ^^, reduce: OosvarValue
			reduce(112), // &&, reduce: OosvarValue
			reduce(112), // ??, reduce: OosvarValue
			reduce(112), // ???, reduce: OosvarValue
			reduce(112), // =~, reduce: OosvarValue
			reduce(112), // !=~, reduce: OosvarValue
			reduce(112), // ==, reduce: OosvarValue
			reduce(112), // !=, reduce: OosvarValue
			reduce(112), // <=>, reduce: OosvarValue
			reduce(112), // >=, reduce: OosvarValue
			reduce(112), // <, reduce: OosvarValue
			reduce(112), // <=, reduce: OosvarValue
			reduce(112), // ^, reduce: OosvarValue
			reduce(112), // &, reduce: OosvarValue
			reduce(112), // <<, reduce: OosvarValue
			reduce(112), // >>>, reduce: OosvarValue
			reduce(112), // +, reduce: OosvarValue
			reduce(112), // -, reduce: OosvarValue
			reduce(112), // .+, reduce: OosvarValue
			reduce(112), // .-, reduce: OosvarValue
			reduce(112), // *, reduce: OosvarValue
			reduce(112), // /, reduce: OosvarValue
			reduce(112), // //, reduce: OosvarValue
			reduce(112), // %, reduce: OosvarValue
			reduce(112), // .*, reduce: OosvarValue
			reduce(112), // ./, reduce: OosvarValue
			reduce(112), // .//, reduce: OosvarValue
			reduce(112), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(112), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(112), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(112), // [[, reduce: OosvarValue
			reduce(112), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S391
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(111), // $, reduce: OosvarValue
			nil,         // empty
			reduce(111), // ;, reduce: OosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(111), // >, reduce: OosvarValue
			reduce(111), // >>, reduce: OosvarValue
			reduce(111), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(111), // ?, reduce: OosvarValue
			nil,         // :
			reduce(111), // ||, reduce: OosvarValue
			reduce(111), // ^^, reduce: OosvarValue
			reduce(111), // &&, reduce: OosvarValue
			reduce(111), // ??, reduce: OosvarValue
			reduce(111), // ???, reduce: OosvarValue
			reduce(111), // =~, reduce: OosvarValue
			reduce(111), // !=~, reduce: OosvarValue
			reduce(111), // ==, reduce: OosvarValue
			reduce(111), // !=, reduce: OosvarValue
			reduce(111), // <=>, reduce: OosvarValue
			reduce(111), // >=, reduce: OosvarValue
			reduce(111), // <, reduce: OosvarValue
			reduce(111), // <=, reduce: OosvarValue
			reduce(111), // ^, reduce: OosvarValue
			reduce(111), // &, reduce: OosvarValue
			reduce(111), // <<, reduce: OosvarValue
			reduce(111), // >>>, reduce: OosvarValue
			reduce(111), // +, reduce: OosvarValue
			reduce(111), // -, reduce: OosvarValue
			reduce(111), // .+, reduce: OosvarValue
			reduce(111), // .-, reduce: OosvarValue
			reduce(111), // *, reduce: OosvarValue
			reduce(111), // /, reduce: OosvarValue
			reduce(111), // //, reduce: OosvarValue
			reduce(111), // %, reduce: OosvarValue
			reduce(111), // .*, reduce: OosvarValue
			reduce(111), // ./, reduce: OosvarValue
			reduce(111), // .//, reduce: OosvarValue
			reduce(111), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(111), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(111), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(111), // [[, reduce: OosvarValue
			reduce(111), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S392
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(99), // $, reduce: FieldValue
			nil,        // empty
			reduce(99), // ;, reduce: FieldValue
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(99), // >, reduce: FieldValue
			reduce(99), // >>, reduce: FieldValue
			reduce(99), // |, reduce: FieldValue
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(99), // ?, reduce: FieldValue
			nil,        // :
			reduce(99), // ||, reduce: FieldValue
			reduce(99), // ^^, reduce: FieldValue
			reduce(99), // &&, reduce: FieldValue
			reduce(99), // ??, reduce: FieldValue
			reduce(99), // ???, reduce: FieldValue
			reduce(99), // =~, reduce: FieldValue
			reduce(99), // !=~, reduce: FieldValue
			reduce(99), // ==, reduce: FieldValue
			reduce(99), // !=, reduce: FieldValue
			reduce(99), // <=>, reduce: FieldValue
			reduce(99), // >=, reduce: FieldValue
			reduce(99), // <, reduce: FieldValue
			reduce(99), // <=, reduce: FieldValue
			reduce(99), // ^, reduce: FieldValue
			reduce(99), // &, reduce: FieldValue
			reduce(99), // <<, reduce: FieldValue
			reduce(99), // >>>, reduce: FieldValue
			reduce(99), // +, reduce: FieldValue
			reduce(99), // -, reduce: FieldValue
			reduce(99), // .+, reduce: FieldValue
			reduce(99), // .-, reduce: FieldValue
			reduce(99), // *, reduce: FieldValue
			reduce(99), // /, reduce: FieldValue
			reduce(99), // //, reduce: FieldValue
			reduce(99), // %, reduce: FieldValue
			reduce(99), // .*, reduce: FieldValue
			reduce(99), // ./, reduce: FieldValue
			reduce(99), // .//, reduce: FieldValue
			reduce(99), // ., reduce: FieldValue
			nil,        // !
			nil,        // ~
			reduce(99), // **, reduce: FieldValue
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(99), // [, reduce: FieldValue
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(99), // [[, reduce: FieldValue
			reduce(99), // [[[, reduce: FieldValue
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S393
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(101), // $, reduce: FieldValue
			nil,         // empty
			reduce(101), // ;, reduce: FieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(101), // >, reduce: FieldValue
			reduce(101), // >>, reduce: FieldValue
			reduce(101), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(101), // ?, reduce: FieldValue
			nil,         // :
			reduce(101), // ||, reduce: FieldValue
			reduce(101), // ^^, reduce: FieldValue
			reduce(101), // &&, reduce: FieldValue
			reduce(101), // ??, reduce: FieldValue
			reduce(101), // ???, reduce: FieldValue
			reduce(101), // =~, reduce: FieldValue
			reduce(101), // !=~, reduce: FieldValue
			reduce(101), // ==, reduce: FieldValue
			reduce(101), // !=, reduce: FieldValue
			reduce(101), // <=>, reduce: FieldValue
			reduce(101), // >=, reduce: FieldValue
			reduce(101), // <, reduce: FieldValue
			reduce(101), // <=, reduce: FieldValue
			reduce(101), // ^, reduce: FieldValue
			reduce(101), // &, reduce: FieldValue
			reduce(101), // <<, reduce: FieldValue
			reduce(101), // >>>, reduce: FieldValue
			reduce(101), // +, reduce: FieldValue
			reduce(101), // -, reduce: FieldValue
			reduce(101), // .+, reduce: FieldValue
			reduce(101), // .-, reduce: FieldValue
			reduce(101), // *, reduce: FieldValue
			reduce(101), // /, reduce: FieldValue
			reduce(101), // //, reduce: FieldValue
			reduce(101), // %, reduce: FieldValue
			reduce(101), // .*, reduce: FieldValue
			reduce(101), // ./, reduce: FieldValue
			reduce(101), // .//, reduce: FieldValue
			reduce(101), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(101), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(101), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(101), // [[, reduce: FieldValue
			reduce(101), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S394
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(100), // $, reduce: FieldValue
			nil,         // empty
			reduce(100), // ;, reduce: FieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(100), // >, reduce: FieldValue
			reduce(100), // >>, reduce: FieldValue
			reduce(100), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(100), // ?, reduce: FieldValue
			nil,         // :
			reduce(100), // ||, reduce: FieldValue
			reduce(100), // ^^, reduce: FieldValue
			reduce(100), // &&, reduce: FieldValue
			reduce(100), // ??, reduce: FieldValue
			reduce(100), // ???, reduce: FieldValue
			reduce(100), // =~, reduce: FieldValue
			reduce(100), // !=~, reduce: FieldValue
			reduce(100), // ==, reduce: FieldValue
			reduce(100), // !=, reduce: FieldValue
			reduce(100), // <=>, reduce: FieldValue
			reduce(100), // >=, reduce: FieldValue
			reduce(100), // <, reduce: FieldValue
			reduce(100), // <=, reduce: FieldValue
			reduce(100), // ^, reduce: FieldValue
			reduce(100), // &, reduce: FieldValue
			reduce(100), // <<, reduce: FieldValue
			reduce(100), // >>>, reduce: FieldValue
			reduce(100), // +, reduce: FieldValue
			reduce(100), // -, reduce: FieldValue
			reduce(100), // .+, reduce: FieldValue
			reduce(100), // .-, reduce: FieldValue
			reduce(100), // *, reduce: FieldValue
			reduce(100), // /, reduce: FieldValue
			reduce(100), // //, reduce: FieldValue
			reduce(100), // %, reduce: FieldValue
			reduce(100), // .*, reduce: FieldValue
			reduce(100), // ./, reduce: FieldValue
			reduce(100), // .//, reduce: FieldValue
			reduce(100), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(100), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(100), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(100), // [[, reduce: FieldValue
			reduce(100), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S395
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(213), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(213), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(213), // >, reduce: MlrvalOrFunction
			reduce(213), // >>, reduce: MlrvalOrFunction
			reduce(213), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(213), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(213), // ||, reduce: MlrvalOrFunction
			reduce(213), // ^^, reduce: MlrvalOrFunction
			reduce(213), // &&, reduce: MlrvalOrFunction
			reduce(213), // ??, reduce: MlrvalOrFunction
			reduce(213), // ???, reduce: MlrvalOrFunction
			reduce(213), // =~, reduce: MlrvalOrFunction
			reduce(213), // !=~, reduce: MlrvalOrFunction
			reduce(213), // ==, reduce: MlrvalOrFunction
			reduce(213), // !=, reduce: MlrvalOrFunction
			reduce(213), // <=>, reduce: MlrvalOrFunction
			reduce(213), // >=, reduce: MlrvalOrFunction
			reduce(213), // <, reduce: MlrvalOrFunction
			reduce(213), // <=, reduce: MlrvalOrFunction
			reduce(213), // ^, reduce: MlrvalOrFunction
			reduce(213), // &, reduce: MlrvalOrFunction
			reduce(213), // <<, reduce: MlrvalOrFunction
			reduce(213), // >>>, reduce: MlrvalOrFunction
			reduce(213), // +, reduce: MlrvalOrFunction
			reduce(213), // -, reduce: MlrvalOrFunction
			reduce(213), // .+, reduce: MlrvalOrFunction
			reduce(213), // .-, reduce: MlrvalOrFunction
			reduce(213), // *, reduce: MlrvalOrFunction
			reduce(213), // /, reduce: MlrvalOrFunction
			reduce(213), // //, reduce: MlrvalOrFunction
			reduce(213), // %, reduce: MlrvalOrFunction
			reduce(213), // .*, reduce: MlrvalOrFunction
			reduce(213), // ./, reduce: MlrvalOrFunction
			reduce(213), // .//, reduce: MlrvalOrFunction
			reduce(213), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(213), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(213), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(213), // [[, reduce: MlrvalOrFunction
			reduce(213), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S396
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(233), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(233), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(233), // >, reduce: MlrvalOrFunction
			reduce(233), // >>, reduce: MlrvalOrFunction
			reduce(233), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(233), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(233), // ||, reduce: MlrvalOrFunction
			reduce(233), // ^^, reduce: MlrvalOrFunction
			reduce(233), // &&, reduce: MlrvalOrFunction
			reduce(233), // ??, reduce: MlrvalOrFunction
			reduce(233), // ???, reduce: MlrvalOrFunction
			reduce(233), // =~, reduce: MlrvalOrFunction
			reduce(233), // !=~, reduce: MlrvalOrFunction
			reduce(233), // ==, reduce: MlrvalOrFunction
			reduce(233), // !=, reduce: MlrvalOrFunction
			reduce(233), // <=>, reduce: MlrvalOrFunction
			reduce(233), // >=, reduce: MlrvalOrFunction
			reduce(233), // <, reduce: MlrvalOrFunction
			reduce(233), // <=, reduce: MlrvalOrFunction
			reduce(233), // ^, reduce: MlrvalOrFunction
			reduce(233), // &, reduce: MlrvalOrFunction
			reduce(233), // <<, reduce: MlrvalOrFunction
			reduce(233), // >>>, reduce: MlrvalOrFunction
			reduce(233), // +, reduce: MlrvalOrFunction
			reduce(233), // -, reduce: MlrvalOrFunction
			reduce(233), // .+, reduce: MlrvalOrFunction
			reduce(233), // .-, reduce: MlrvalOrFunction
			reduce(233), // *, reduce: MlrvalOrFunction
			reduce(233), // /, reduce: MlrvalOrFunction
			reduce(233), // //, reduce: MlrvalOrFunction
			reduce(233), // %, reduce: MlrvalOrFunction
			reduce(233), // .*, reduce: MlrvalOrFunction
			reduce(233), // ./, reduce: MlrvalOrFunction
			reduce(233), // .//, reduce: MlrvalOrFunction
			reduce(233), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(233), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(233), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(233), // [[, reduce: MlrvalOrFunction
			reduce(233), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S397
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(210), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(210), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(210), // >, reduce: MlrvalOrFunction
			reduce(210), // >>, reduce: MlrvalOrFunction
			reduce(210), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(210), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(210), // ||, reduce: MlrvalOrFunction
			reduce(210), // ^^, reduce: MlrvalOrFunction
			reduce(210), // &&, reduce: MlrvalOrFunction
			reduce(210), // ??, reduce: MlrvalOrFunction
			reduce(210), // ???, reduce: MlrvalOrFunction
			reduce(210), // =~, reduce: MlrvalOrFunction
			reduce(210), // !=~, reduce: MlrvalOrFunction
			reduce(210), // ==, reduce: MlrvalOrFunction
			reduce(210), // !=, reduce: MlrvalOrFunction
			reduce(210), // <=>, reduce: MlrvalOrFunction
			reduce(210), // >=, reduce: MlrvalOrFunction
			reduce(210), // <, reduce: MlrvalOrFunction
			reduce(210), // <=, reduce: MlrvalOrFunction
			reduce(210), // ^, reduce: MlrvalOrFunction
			reduce(210), // &, reduce: MlrvalOrFunction
			reduce(210), // <<, reduce: MlrvalOrFunction
			reduce(210), // >>>, reduce: MlrvalOrFunction
			reduce(210), // +, reduce: MlrvalOrFunction
			reduce(210), // -, reduce: MlrvalOrFunction
			reduce(210), // .+, reduce: MlrvalOrFunction
			reduce(210), // .-, reduce: MlrvalOrFunction
			reduce(210), // *, reduce: MlrvalOrFunction
			reduce(210), // /, reduce: MlrvalOrFunction
			reduce(210), // //, reduce: MlrvalOrFunction
			reduce(210), // %, reduce: MlrvalOrFunction
			reduce(210), // .*, reduce: MlrvalOrFunction
			reduce(210), // ./, reduce: MlrvalOrFunction
			reduce(210), // .//, reduce: MlrvalOrFunction
			reduce(210), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(210), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(210), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(210), // [[, reduce: MlrvalOrFunction
			reduce(210), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S398
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(102), // $, reduce: FieldValue
			nil,         // empty
			reduce(102), // ;, reduce: FieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(102), // >, reduce: FieldValue
			reduce(102), // >>, reduce: FieldValue
			reduce(102), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(102), // ?, reduce: FieldValue
			nil,         // :
			reduce(102), // ||, reduce: FieldValue
			reduce(102), // ^^, reduce: FieldValue
			reduce(102), // &&, reduce: FieldValue
			reduce(102), // ??, reduce: FieldValue
			reduce(102), // ???, reduce: FieldValue
			reduce(102), // =~, reduce: FieldValue
			reduce(102), // !=~, reduce: FieldValue
			reduce(102), // ==, reduce: FieldValue
			reduce(102), // !=, reduce: FieldValue
			reduce(102), // <=>, reduce: FieldValue
			reduce(102), // >=, reduce: FieldValue
			reduce(102), // <, reduce: FieldValue
			reduce(102), // <=, reduce: FieldValue
			reduce(102), // ^, reduce: FieldValue
			reduce(102), // &, reduce: FieldValue
			reduce(102), // <<, reduce: FieldValue
			reduce(102), // >>>, reduce: FieldValue
			reduce(102), // +, reduce: FieldValue
			reduce(102), // -, reduce: FieldValue
			reduce(102), // .+, reduce: FieldValue
			reduce(102), // .-, reduce: FieldValue
			reduce(102), // *, reduce: FieldValue
			reduce(102), // /, reduce: FieldValue
			reduce(102), // //, reduce: FieldValue
			reduce(102), // %, reduce: FieldValue
			reduce(102), // .*, reduce: FieldValue
			reduce(102), // ./, reduce: FieldValue
			reduce(102), // .//, reduce: FieldValue
			reduce(102), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(102), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(102), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(102), // [[, reduce: FieldValue
			reduce(102), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S399
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(103), // $, reduce: FieldValue
			nil,         // empty
			reduce(103), // ;, reduce: FieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(103), // >, reduce: FieldValue
			reduce(103), // >>, reduce: FieldValue
			reduce(103), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(103), // ?, reduce: FieldValue
			nil,         // :
			reduce(103), // ||, reduce: FieldValue
			reduce(103), // ^^, reduce: FieldValue
			reduce(103), // &&, reduce: FieldValue
			reduce(103), // ??, reduce: FieldValue
			reduce(103), // ???, reduce: FieldValue
			reduce(103), // =~, reduce: FieldValue
			reduce(103), // !=~, reduce: FieldValue
			reduce(103), // ==, reduce: FieldValue
			reduce(103), // !=, reduce: FieldValue
			reduce(103), // <=>, reduce: FieldValue
			reduce(103), // >=, reduce: FieldValue
			reduce(103), // <, reduce: FieldValue
			reduce(103), // <=, reduce: FieldValue
			reduce(103), // ^, reduce: FieldValue
			reduce(103), // &, reduce: FieldValue
			reduce(103), // <<, reduce: FieldValue
			reduce(103), // >>>, reduce: FieldValue
			reduce(103), // +, reduce: FieldValue
			reduce(103), // -, reduce: FieldValue
			reduce(103), // .+, reduce: FieldValue
			reduce(103), // .-, reduce: FieldValue
			reduce(103), // *, reduce: FieldValue
			reduce(103), // /, reduce: FieldValue
			reduce(103), // //, reduce: FieldValue
			reduce(103), // %, reduce: FieldValue
			reduce(103), // .*, reduce: FieldValue
			reduce(103), // ./, reduce: FieldValue
			reduce(103), // .//, reduce: FieldValue
			reduce(103), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(103), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(103), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(103), // [[, reduce: FieldValue
			reduce(103), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S400
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(104), // $, reduce: DirectFieldValue
			nil,         // empty
			reduce(104), // ;, reduce: DirectFieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(104), // >, reduce: DirectFieldValue
			reduce(104), // >>, reduce: DirectFieldValue
			reduce(104), // |, reduce: DirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(104), // ?, reduce: DirectFieldValue
			nil,         // :
			reduce(104), // ||, reduce: DirectFieldValue
			reduce(104), // ^^, reduce: DirectFieldValue
			reduce(104), // &&, reduce: DirectFieldValue
			reduce(104), // ??, reduce: DirectFieldValue
			reduce(104), // ???, reduce: DirectFieldValue
			reduce(104), // =~, reduce: DirectFieldValue
			reduce(104), // !=~, reduce: DirectFieldValue
			reduce(104), // ==, reduce: DirectFieldValue
			reduce(104), // !=, reduce: DirectFieldValue
			reduce(104), // <=>, reduce: DirectFieldValue
			reduce(104), // >=, reduce: DirectFieldValue
			reduce(104), // <, reduce: DirectFieldValue
			reduce(104), // <=, reduce: DirectFieldValue
			reduce(104), // ^, reduce: DirectFieldValue
			reduce(104), // &, reduce: DirectFieldValue
			reduce(104), // <<, reduce: DirectFieldValue
			reduce(104), // >>>, reduce: DirectFieldValue
			reduce(104), // +, reduce: DirectFieldValue
			reduce(104), // -, reduce: DirectFieldValue
			reduce(104), // .+, reduce: DirectFieldValue
			reduce(104), // .-, reduce: DirectFieldValue
			reduce(104), // *, reduce: DirectFieldValue
			reduce(104), // /, reduce: DirectFieldValue
			reduce(104), // //, reduce: DirectFieldValue
			reduce(104), // %, reduce: DirectFieldValue
			reduce(104), // .*, reduce: DirectFieldValue
			reduce(104), // ./, reduce: DirectFieldValue
			reduce(104), // .//, reduce: DirectFieldValue
			reduce(104), // ., reduce: DirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(104), // **, reduce: DirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(104), // [, reduce: DirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(104), // [[, reduce: DirectFieldValue
			reduce(104), // [[[, reduce: DirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S401
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S402
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(106), // $, reduce: BracedFieldValue
			nil,         // empty
			reduce(106), // ;, reduce: BracedFieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(106), // >, reduce: BracedFieldValue
			reduce(106), // >>, reduce: BracedFieldValue
			reduce(106), // |, reduce: BracedFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(106), // ?, reduce: BracedFieldValue
			nil,         // :
			reduce(106), // ||, reduce: BracedFieldValue
			reduce(106), // ^^, reduce: BracedFieldValue
			reduce(106), // &&, reduce: BracedFieldValue
			reduce(106), // ??, reduce: BracedFieldValue
			reduce(106), // ???, reduce: BracedFieldValue
			reduce(106), // =~, reduce: BracedFieldValue
			reduce(106), // !=~, reduce: BracedFieldValue
			reduce(106), // ==, reduce: BracedFieldValue
			reduce(106), // !=, reduce: BracedFieldValue
			reduce(106), // <=>, reduce: BracedFieldValue
			reduce(106), // >=, reduce: BracedFieldValue
			reduce(106), // <, reduce: BracedFieldValue
			reduce(106), // <=, reduce: BracedFieldValue
			reduce(106), // ^, reduce: BracedFieldValue
			reduce(106), // &, reduce: BracedFieldValue
			reduce(106), // <<, reduce: BracedFieldValue
			reduce(106), // >>>, reduce: BracedFieldValue
			reduce(106), // +, reduce: BracedFieldValue
			reduce(106), // -, reduce: BracedFieldValue
			reduce(106), // .+, reduce: BracedFieldValue
			reduce(106), // .-, reduce: BracedFieldValue
			reduce(106), // *, reduce: BracedFieldValue
			reduce(106), // /, reduce: BracedFieldValue
			reduce(106), // //, reduce: BracedFieldValue
			reduce(106), // %, reduce: BracedFieldValue
			reduce(106), // .*, reduce: BracedFieldValue
			reduce(106), // ./, reduce: BracedFieldValue
			reduce(106), // .//, reduce: BracedFieldValue
			reduce(106), // ., reduce: BracedFieldValue
			nil,         // !
			nil,         // ~
			reduce(106), // **, reduce: BracedFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(106), // [, reduce: BracedFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(106), // [[, reduce: BracedFieldValue
			reduce(106), // [[[, reduce: BracedFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S403
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S404
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S405
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(109), // $, reduce: FullSrec
			nil,         // empty
			reduce(109), // ;, reduce: FullSrec
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(109), // >, reduce: FullSrec
			reduce(109), // >>, reduce: FullSrec
			reduce(109), // |, reduce: FullSrec
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(109), // ?, reduce: FullSrec
			nil,         // :
			reduce(109), // ||, reduce: FullSrec
			reduce(109), // ^^, reduce: FullSrec
			reduce(109), // &&, reduce: FullSrec
			reduce(109), // ??, reduce: FullSrec
			reduce(109), // ???, reduce: FullSrec
			reduce(109), // =~, reduce: FullSrec
			reduce(109), // !=~, reduce: FullSrec
			reduce(109), // ==, reduce: FullSrec
			reduce(109), // !=, reduce: FullSrec
			reduce(109), // <=>, reduce: FullSrec
			reduce(109), // >=, reduce: FullSrec
			reduce(109), // <, reduce: FullSrec
			reduce(109), // <=, reduce: FullSrec
			reduce(109), // ^, reduce: FullSrec
			reduce(109), // &, reduce: FullSrec
			reduce(109), // <<, reduce: FullSrec
			reduce(109), // >>>, reduce: FullSrec
			reduce(109), // +, reduce: FullSrec
			reduce(109), // -, reduce: FullSrec
			reduce(109), // .+, reduce: FullSrec
			reduce(109), // .-, reduce: FullSrec
			reduce(109), // *, reduce: FullSrec
			reduce(109), // /, reduce: FullSrec
			reduce(109), // //, reduce: FullSrec
			reduce(109), // %, reduce: FullSrec
			reduce(109), // .*, reduce: FullSrec
			reduce(109), // ./, reduce: FullSrec
			reduce(109), // .//, reduce: FullSrec
			reduce(109), // ., reduce: FullSrec
			nil,         // !
			nil,         // ~
			reduce(109), // **, reduce: FullSrec
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(109), // [, reduce: FullSrec
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(109), // [[, reduce: FullSrec
			reduce(109), // [[[, reduce: FullSrec
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S406
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(212), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(212), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(212), // >, reduce: MlrvalOrFunction
			reduce(212), // >>, reduce: MlrvalOrFunction
			reduce(212), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(212), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(212), // ||, reduce: MlrvalOrFunction
			reduce(212), // ^^, reduce: MlrvalOrFunction
			reduce(212), // &&, reduce: MlrvalOrFunction
			reduce(212), // ??, reduce: MlrvalOrFunction
			reduce(212), // ???, reduce: MlrvalOrFunction
			reduce(212), // =~, reduce: MlrvalOrFunction
			reduce(212), // !=~, reduce: MlrvalOrFunction
			reduce(212), // ==, reduce: MlrvalOrFunction
			reduce(212), // !=, reduce: MlrvalOrFunction
			reduce(212), // <=>, reduce: MlrvalOrFunction
			reduce(212), // >=, reduce: MlrvalOrFunction
			reduce(212), // <, reduce: MlrvalOrFunction
			reduce(212), // <=, reduce: MlrvalOrFunction
			reduce(212), // ^, reduce: MlrvalOrFunction
			reduce(212), // &, reduce: MlrvalOrFunction
			reduce(212), // <<, reduce: MlrvalOrFunction
			reduce(212), // >>>, reduce: MlrvalOrFunction
			reduce(212), // +, reduce: MlrvalOrFunction
			reduce(212), // -, reduce: MlrvalOrFunction
			reduce(212), // .+, reduce: MlrvalOrFunction
			reduce(212), // .-, reduce: MlrvalOrFunction
			reduce(212), // *, reduce: MlrvalOrFunction
			reduce(212), // /, reduce: MlrvalOrFunction
			reduce(212), // //, reduce: MlrvalOrFunction
			reduce(212), // %, reduce: MlrvalOrFunction
			reduce(212), // .*, reduce: MlrvalOrFunction
			reduce(212), // ./, reduce: MlrvalOrFunction
			reduce(212), // .//, reduce: MlrvalOrFunction
			reduce(212), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(212), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(212), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(212), // [[, reduce: MlrvalOrFunction
			reduce(212), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S407
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(113), // $, reduce: DirectOosvarValue
			nil,         // empty
			reduce(113), // ;, reduce: DirectOosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(113), // >, reduce: DirectOosvarValue
			reduce(113), // >>, reduce: DirectOosvarValue
			reduce(113), // |, reduce: DirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(113), // ?, reduce: DirectOosvarValue
			nil,         // :
			reduce(113), // ||, reduce: DirectOosvarValue
			reduce(113), // ^^, reduce: DirectOosvarValue
			reduce(113), // &&, reduce: DirectOosvarValue
			reduce(113), // ??, reduce: DirectOosvarValue
			reduce(113), // ???, reduce: DirectOosvarValue
			reduce(113), // =~, reduce: DirectOosvarValue
			reduce(113), // !=~, reduce: DirectOosvarValue
			reduce(113), // ==, reduce: DirectOosvarValue
			reduce(113), // !=, reduce: DirectOosvarValue
			reduce(113), // <=>, reduce: DirectOosvarValue
			reduce(113), // >=, reduce: DirectOosvarValue
			reduce(113), // <, reduce: DirectOosvarValue
			reduce(113), // <=, reduce: DirectOosvarValue
			reduce(113), // ^, reduce: DirectOosvarValue
			reduce(113), // &, reduce: DirectOosvarValue
			reduce(113), // <<, reduce: DirectOosvarValue
			reduce(113), // >>>, reduce: DirectOosvarValue
			reduce(113), // +, reduce: DirectOosvarValue
			reduce(113), // -, reduce: DirectOosvarValue
			reduce(113), // .+, reduce: DirectOosvarValue
			reduce(113), // .-, reduce: DirectOosvarValue
			reduce(113), // *, reduce: DirectOosvarValue
			reduce(113), // /, reduce: DirectOosvarValue
			reduce(113), // //, reduce: DirectOosvarValue
			reduce(113), // %, reduce: DirectOosvarValue
			reduce(113), // .*, reduce: DirectOosvarValue
			reduce(113), // ./, reduce: DirectOosvarValue
			reduce(113), // .//, reduce: DirectOosvarValue
			reduce(113), // ., reduce: DirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(113), // **, reduce: DirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(113), // [, reduce: DirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(113), // [[, reduce: DirectOosvarValue
			reduce(113), // [[[, reduce: DirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S408
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S409
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(115), // $, reduce: BracedOosvarValue
			nil,         // empty
			reduce(115), // ;, reduce: BracedOosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(115), // >, reduce: BracedOosvarValue
			reduce(115), // >>, reduce: BracedOosvarValue
			reduce(115), // |, reduce: BracedOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(115), // ?, reduce: BracedOosvarValue
			nil,         // :
			reduce(115), // ||, reduce: BracedOosvarValue
			reduce(115), // ^^, reduce: BracedOosvarValue
			reduce(115), // &&, reduce: BracedOosvarValue
			reduce(115), // ??, reduce: BracedOosvarValue
			reduce(115), // ???, reduce: BracedOosvarValue
			reduce(115), // =~, reduce: BracedOosvarValue
			reduce(115), // !=~, reduce: BracedOosvarValue
			reduce(115), // ==, reduce: BracedOosvarValue
			reduce(115), // !=, reduce: BracedOosvarValue
			reduce(115), // <=>, reduce: BracedOosvarValue
			reduce(115), // >=, reduce: BracedOosvarValue
			reduce(115), // <, reduce: BracedOosvarValue
			reduce(115), // <=, reduce: BracedOosvarValue
			reduce(115), // ^, reduce: BracedOosvarValue
			reduce(115), // &, reduce: BracedOosvarValue
			reduce(115), // <<, reduce: BracedOosvarValue
			reduce(115), // >>>, reduce: BracedOosvarValue
			reduce(115), // +, reduce: BracedOosvarValue
			reduce(115), // -, reduce: BracedOosvarValue
			reduce(115), // .+, reduce: BracedOosvarValue
			reduce(115), // .-, reduce: BracedOosvarValue
			reduce(115), // *, reduce: BracedOosvarValue
			reduce(115), // /, reduce: BracedOosvarValue
			reduce(115), // //, reduce: BracedOosvarValue
			reduce(115), // %, reduce: BracedOosvarValue
			reduce(115), // .*, reduce: BracedOosvarValue
			reduce(115), // ./, reduce: BracedOosvarValue
			reduce(115), // .//, reduce: BracedOosvarValue
			reduce(115), // ., reduce: BracedOosvarValue
			nil,         // !
			nil,         // ~
			reduce(115), // **, reduce: BracedOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(115), // [, reduce: BracedOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(115), // [[, reduce: BracedOosvarValue
			reduce(115), // [[[, reduce: BracedOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S410
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(116), // $, reduce: FullOosvar
			nil,         // empty
			reduce(116), // ;, reduce: FullOosvar
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(116), // >, reduce: FullOosvar
			reduce(116), // >>, reduce: FullOosvar
			reduce(116), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(116), // ?, reduce: FullOosvar
			nil,         // :
			reduce(116), // ||, reduce: FullOosvar
			reduce(116), // ^^, reduce: FullOosvar
			reduce(116), // &&, reduce: FullOosvar
			reduce(116), // ??, reduce: FullOosvar
			reduce(116), // ???, reduce: FullOosvar
			reduce(116), // =~, reduce: FullOosvar
			reduce(116), // !=~, reduce: FullOosvar
			reduce(116), // ==, reduce: FullOosvar
			reduce(116), // !=, reduce: FullOosvar
			reduce(116), // <=>, reduce: FullOosvar
			reduce(116), // >=, reduce: FullOosvar
			reduce(116), // <, reduce: FullOosvar
			reduce(116), // <=, reduce: FullOosvar
			reduce(116), // ^, reduce: FullOosvar
			reduce(116), // &, reduce: FullOosvar
			reduce(116), // <<, reduce: FullOosvar
			reduce(116), // >>>, reduce: FullOosvar
			reduce(116), // +, reduce: FullOosvar
			reduce(116), // -, reduce: FullOosvar
			reduce(116), // .+, reduce: FullOosvar
			reduce(116), // .-, reduce: FullOosvar
			reduce(116), // *, reduce: FullOosvar
			reduce(116), // /, reduce: FullOosvar
			reduce(116), // //, reduce: FullOosvar
			reduce(116), // %, reduce: FullOosvar
			reduce(116), // .*, reduce: FullOosvar
			reduce(116), // ./, reduce: FullOosvar
			reduce(116), // .//, reduce: FullOosvar
			reduce(116), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(116), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(116), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(116), // [[, reduce: FullOosvar
			reduce(116), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S411
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(117), // $, reduce: FullOosvar
			nil,         // empty
			reduce(117), // ;, reduce: FullOosvar
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(117), // >, reduce: FullOosvar
			reduce(117), // >>, reduce: FullOosvar
			reduce(117), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(117), // ?, reduce: FullOosvar
			nil,         // :
			reduce(117), // ||, reduce: FullOosvar
			reduce(117), // ^^, reduce: FullOosvar
			reduce(117), // &&, reduce: FullOosvar
			reduce(117), // ??, reduce: FullOosvar
			reduce(117), // ???, reduce: FullOosvar
			reduce(117), // =~, reduce: FullOosvar
			reduce(117), // !=~, reduce: FullOosvar
			reduce(117), // ==, reduce: FullOosvar
			reduce(117), // !=, reduce: FullOosvar
			reduce(117), // <=>, reduce: FullOosvar
			reduce(117), // >=, reduce: FullOosvar
			reduce(117), // <, reduce: FullOosvar
			reduce(117), // <=, reduce: FullOosvar
			reduce(117), // ^, reduce: FullOosvar
			reduce(117), // &, reduce: FullOosvar
			reduce(117), // <<, reduce: FullOosvar
			reduce(117), // >>>, reduce: FullOosvar
			reduce(117), // +, reduce: FullOosvar
			reduce(117), // -, reduce: FullOosvar
			reduce(117), // .+, reduce: FullOosvar
			reduce(117), // .-, reduce: FullOosvar
			reduce(117), // *, reduce: FullOosvar
			reduce(117), // /, reduce: FullOosvar
			reduce(117), // //, reduce: FullOosvar
			reduce(117), // %, reduce: FullOosvar
			reduce(117), // .*, reduce: FullOosvar
			reduce(117), // ./, reduce: FullOosvar
			reduce(117), // .//, reduce: FullOosvar
			reduce(117), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(117), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(117), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(117), // [[, reduce: FullOosvar
			reduce(117), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S412
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(118), // $, reduce: LocalVariable
			nil,         // empty
			reduce(118), // ;, reduce: LocalVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(118), // >, reduce: LocalVariable
			reduce(118), // >>, reduce: LocalVariable
			reduce(118), // |, reduce: LocalVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(270), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(118), // ?, reduce: LocalVariable
			nil,         // :
			reduce(118), // ||, reduce: LocalVariable
			reduce(118), // ^^, reduce: LocalVariable
			reduce(118), // &&, reduce: LocalVariable
			reduce(118), // ??, reduce: LocalVariable
			reduce(118), // ???, reduce: LocalVariable
			reduce(118), // =~, reduce: LocalVariable
			reduce(118), // !=~, reduce: LocalVariable
			reduce(118), // ==, reduce: LocalVariable
			reduce(118), // !=, reduce: LocalVariable
			reduce(118), // <=>, reduce: LocalVariable
			reduce(118), // >=, reduce: LocalVariable
			reduce(118), // <, reduce: LocalVariable
			reduce(118), // <=, reduce: LocalVariable
			reduce(118), // ^, reduce: LocalVariable
			reduce(118), // &, reduce: LocalVariable
			reduce(118), // <<, reduce: LocalVariable
			reduce(118), // >>>, reduce: LocalVariable
			reduce(118), // +, reduce: LocalVariable
			reduce(118), // -, reduce: LocalVariable
			reduce(118), // .+, reduce: LocalVariable
			reduce(118), // .-, reduce: LocalVariable
			reduce(118), // *, reduce: LocalVariable
			reduce(118), // /, reduce: LocalVariable
			reduce(118), // //, reduce: LocalVariable
			reduce(118), // %, reduce: LocalVariable
			reduce(118), // .*, reduce: LocalVariable
			reduce(118), // ./, reduce: LocalVariable
			reduce(118), // .//, reduce: LocalVariable
			reduce(118), // ., reduce: LocalVariable
			nil,         // !
			nil,         // ~
			reduce(118), // **, reduce: LocalVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(118), // [, reduce: LocalVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(118), // [[, reduce: LocalVariable
			reduce(118), // [[[, reduce: LocalVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S413
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(147), // $, reduce: Rvalue
			nil,         // empty
			reduce(147), // ;, reduce: Rvalue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S414
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(148), // $, reduce: PrecedenceChainStart
			nil,         // empty
			reduce(148), // ;, reduce: PrecedenceChainStart
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S415
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(150), // $, reduce: TernaryTerm
			nil,         // empty
			reduce(150), // ;, reduce: TernaryTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			shift(1167), // ?
			nil,         // :
			shift(1168), // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S416
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(152), // $, reduce: LogicalOrTerm
			nil,         // empty
			reduce(152), // ;, reduce: LogicalOrTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(152), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(152), // ||, reduce: LogicalOrTerm
			shift(1169), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S417
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(154), // $, reduce: LogicalXORTerm
			nil,         // empty
			reduce(154), // ;, reduce: LogicalXORTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(154), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(154), // ||, reduce: LogicalXORTerm
			reduce(154), // ^^, reduce: LogicalXORTerm
			shift(1170), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S418
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(156), // $, reduce: LogicalAndTerm
			nil,         // empty
			reduce(156), // ;, reduce: LogicalAndTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(156), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(156), // ||, reduce: LogicalAndTerm
			reduce(156), // ^^, reduce: LogicalAndTerm
			reduce(156), // &&, reduce: LogicalAndTerm
			shift(1171), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S419
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(158), // $, reduce: AbsentCoalesceTerm
			nil,         // empty
			reduce(158), // ;, reduce: AbsentCoalesceTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(158), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(158), // ||, reduce: AbsentCoalesceTerm
			reduce(158), // ^^, reduce: AbsentCoalesceTerm
			reduce(158), // &&, reduce: AbsentCoalesceTerm
			reduce(158), // ??, reduce: AbsentCoalesceTerm
			shift(1172), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S420
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(160), // $, reduce: EmptyCoalesceTerm
			nil,         // empty
			reduce(160), // ;, reduce: EmptyCoalesceTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(160), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(160), // ||, reduce: EmptyCoalesceTerm
			reduce(160), // ^^, reduce: EmptyCoalesceTerm
			reduce(160), // &&, reduce: EmptyCoalesceTerm
			reduce(160), // ??, reduce: EmptyCoalesceTerm
			reduce(160), // ???, reduce: EmptyCoalesceTerm
			shift(1173), // =~
			shift(1174), // !=~
			shift(1175), // ==
			shift(1176), // !=
			shift(1177), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S421
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(166), // $, reduce: EqneTerm
			nil,         // empty
			reduce(166), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1178), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(166), // ?, reduce: EqneTerm
			nil,         // :
			reduce(166), // ||, reduce: EqneTerm
			reduce(166), // ^^, reduce: EqneTerm
			reduce(166), // &&, reduce: EqneTerm
			reduce(166), // ??, reduce: EqneTerm
			reduce(166), // ???, reduce: EqneTerm
			reduce(166), // =~, reduce: EqneTerm
			reduce(166), // !=~, reduce: EqneTerm
			reduce(166), // ==, reduce: EqneTerm
			reduce(166), // !=, reduce: EqneTerm
			reduce(166), // <=>, reduce: EqneTerm
			shift(1179), // >=
			shift(1180), // <
			shift(1181), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S422
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(171), // $, reduce: CmpTerm
			nil,         // empty
			reduce(171), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(171), // >, reduce: CmpTerm
			nil,         // >>
			shift(1182), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(171), // ?, reduce: CmpTerm
			nil,         // :
			reduce(171), // ||, reduce: CmpTerm
			reduce(171), // ^^, reduce: CmpTerm
			reduce(171), // &&, reduce: CmpTerm
			reduce(171), // ??, reduce: CmpTerm
			reduce(171), // ???, reduce: CmpTerm
			reduce(171), // =~, reduce: CmpTerm
			reduce(171), // !=~, reduce: CmpTerm
			reduce(171), // ==, reduce: CmpTerm
			reduce(171), // !=, reduce: CmpTerm
			reduce(171), // <=>, reduce: CmpTerm
			reduce(171), // >=, reduce: CmpTerm
			reduce(171), // <, reduce: CmpTerm
			reduce(171), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S423
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(173), // $, reduce: BitwiseORTerm
			nil,         // empty
			reduce(173), // ;, reduce: BitwiseORTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(173), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(173), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(173), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(173), // ||, reduce: BitwiseORTerm
			reduce(173), // ^^, reduce: BitwiseORTerm
			reduce(173), // &&, reduce: BitwiseORTerm
			reduce(173), // ??, reduce: BitwiseORTerm
			reduce(173), // ???, reduce: BitwiseORTerm
			reduce(173), // =~, reduce: BitwiseORTerm
			reduce(173), // !=~, reduce: BitwiseORTerm
			reduce(173), // ==, reduce: BitwiseORTerm
			reduce(173), // !=, reduce: BitwiseORTerm
			reduce(173), // <=>, reduce: BitwiseORTerm
			reduce(173), // >=, reduce: BitwiseORTerm
			reduce(173), // <, reduce: BitwiseORTerm
			reduce(173), // <=, reduce: BitwiseORTerm
			shift(1183), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S424
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(175), // $, reduce: BitwiseXORTerm
			nil,         // empty
			reduce(175), // ;, reduce: BitwiseXORTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(175), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(175), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(175), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(175), // ||, reduce: BitwiseXORTerm
			reduce(175), // ^^, reduce: BitwiseXORTerm
			reduce(175), // &&, reduce: BitwiseXORTerm
			reduce(175), // ??, reduce: BitwiseXORTerm
			reduce(175), // ???, reduce: BitwiseXORTerm
			reduce(175), // =~, reduce: BitwiseXORTerm
			reduce(175), // !=~, reduce: BitwiseXORTerm
			reduce(175), // ==, reduce: BitwiseXORTerm
			reduce(175), // !=, reduce: BitwiseXORTerm
			reduce(175), // <=>, reduce: BitwiseXORTerm
			reduce(175), // >=, reduce: BitwiseXORTerm
			reduce(175), // <, reduce: BitwiseXORTerm
			reduce(175), // <=, reduce: BitwiseXORTerm
			reduce(175), // ^, reduce: BitwiseXORTerm
			shift(1184), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S425
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(177), // $, reduce: BitwiseANDTerm
			nil,         // empty
			reduce(177), // ;, reduce: BitwiseANDTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(177), // >, reduce: BitwiseANDTerm
			shift(1185), // >>
			reduce(177), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(177), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(177), // ||, reduce: BitwiseANDTerm
			reduce(177), // ^^, reduce: BitwiseANDTerm
			reduce(177), // &&, reduce: BitwiseANDTerm
			reduce(177), // ??, reduce: BitwiseANDTerm
			reduce(177), // ???, reduce: BitwiseANDTerm
			reduce(177), // =~, reduce: BitwiseANDTerm
			reduce(177), // !=~, reduce: BitwiseANDTerm
			reduce(177), // ==, reduce: BitwiseANDTerm
			reduce(177), // !=, reduce: BitwiseANDTerm
			reduce(177), // <=>, reduce: BitwiseANDTerm
			reduce(177), // >=, reduce: BitwiseANDTerm
			reduce(177), // <, reduce: BitwiseANDTerm
			reduce(177), // <=, reduce: BitwiseANDTerm
			reduce(177), // ^, reduce: BitwiseANDTerm
			reduce(177), // &, reduce: BitwiseANDTerm
			shift(1186), // <<
			shift(1187), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S426
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(181), // $, reduce: BitwiseShiftTerm
			nil,         // empty
			reduce(181), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(181), // >, reduce: BitwiseShiftTerm
			reduce(181), // >>, reduce: BitwiseShiftTerm
			reduce(181), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(181), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(181), // ||, reduce: BitwiseShiftTerm
			reduce(181), // ^^, reduce: BitwiseShiftTerm
			reduce(181), // &&, reduce: BitwiseShiftTerm
			reduce(181), // ??, reduce: BitwiseShiftTerm
			reduce(181), // ???, reduce: BitwiseShiftTerm
			reduce(181), // =~, reduce: BitwiseShiftTerm
			reduce(181), // !=~, reduce: BitwiseShiftTerm
			reduce(181), // ==, reduce: BitwiseShiftTerm
			reduce(181), // !=, reduce: BitwiseShiftTerm
			reduce(181), // <=>, reduce: BitwiseShiftTerm
			reduce(181), // >=, reduce: BitwiseShiftTerm
			reduce(181), // <, reduce: BitwiseShiftTerm
			reduce(181), // <=, reduce: BitwiseShiftTerm
			reduce(181), // ^, reduce: BitwiseShiftTerm
			reduce(181), // &, reduce: BitwiseShiftTerm
			reduce(181), // <<, reduce: BitwiseShiftTerm
			reduce(181), // >>>, reduce: BitwiseShiftTerm
			shift(1188), // +
			shift(1189), // -
			shift(1190), // .+
			shift(1191), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S427
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S428
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(186), // $, reduce: AddsubdotTerm
			nil,         // empty
			reduce(186), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(186), // >, reduce: AddsubdotTerm
			reduce(186), // >>, reduce: AddsubdotTerm
			reduce(186), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(186), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(186), // ||, reduce: AddsubdotTerm
			reduce(186), // ^^, reduce: AddsubdotTerm
			reduce(186), // &&, reduce: AddsubdotTerm
			reduce(186), // ??, reduce: AddsubdotTerm
			reduce(186), // ???, reduce: AddsubdotTerm
			reduce(186), // =~, reduce: AddsubdotTerm
			reduce(186), // !=~, reduce: AddsubdotTerm
			reduce(186), // ==, reduce: AddsubdotTerm
			reduce(186), // !=, reduce: AddsubdotTerm
			reduce(186), // <=>, reduce: AddsubdotTerm
			reduce(186), // >=, reduce: AddsubdotTerm
			reduce(186), // <, reduce: AddsubdotTerm
			reduce(186), // <=, reduce: AddsubdotTerm
			reduce(186), // ^, reduce: AddsubdotTerm
			reduce(186), // &, reduce: AddsubdotTerm
			reduce(186), // <<, reduce: AddsubdotTerm
			reduce(186), // >>>, reduce: AddsubdotTerm
			reduce(186), // +, reduce: AddsubdotTerm
			reduce(186), // -, reduce: AddsubdotTerm
			reduce(186), // .+, reduce: AddsubdotTerm
			reduce(186), // .-, reduce: AddsubdotTerm
			shift(1193), // *
			shift(1194), // /
			shift(1195), // //
			shift(1196), // %
			shift(1197), // .*
			shift(1198), // ./
			shift(1199), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S429
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S430
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S431
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S432
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(194), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(194), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(194), // >, reduce: MuldivTerm
			reduce(194), // >>, reduce: MuldivTerm
			reduce(194), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(194), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(194), // ||, reduce: MuldivTerm
			reduce(194), // ^^, reduce: MuldivTerm
			reduce(194), // &&, reduce: MuldivTerm
			reduce(194), // ??, reduce: MuldivTerm
			reduce(194), // ???, reduce: MuldivTerm
			reduce(194), // =~, reduce: MuldivTerm
			reduce(194), // !=~, reduce: MuldivTerm
			reduce(194), // ==, reduce: MuldivTerm
			reduce(194), // !=, reduce: MuldivTerm
			reduce(194), // <=>, reduce: MuldivTerm
			reduce(194), // >=, reduce: MuldivTerm
			reduce(194), // <, reduce: MuldivTerm
			reduce(194), // <=, reduce: MuldivTerm
			reduce(194), // ^, reduce: MuldivTerm
			reduce(194), // &, reduce: MuldivTerm
			reduce(194), // <<, reduce: MuldivTerm
			reduce(194), // >>>, reduce: MuldivTerm
			reduce(194), // +, reduce: MuldivTerm
			reduce(194), // -, reduce: MuldivTerm
			reduce(194), // .+, reduce: MuldivTerm
			reduce(194), // .-, reduce: MuldivTerm
			reduce(194), // *, reduce: MuldivTerm
			reduce(194), // /, reduce: MuldivTerm
			reduce(194), // //, reduce: MuldivTerm
			reduce(194), // %, reduce: MuldivTerm
			reduce(194), // .*, reduce: MuldivTerm
			reduce(194), // ./, reduce: MuldivTerm
			reduce(194), // .//, reduce: MuldivTerm
			shift(1203), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S433
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(196), // $, reduce: DotTerm
			nil,         // empty
			reduce(196), // ;, reduce: DotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(196), // >, reduce: DotTerm
			reduce(196), // >>, reduce: DotTerm
			reduce(196), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(196), // ?, reduce: DotTerm
			nil,         // :
			reduce(196), // ||, reduce: DotTerm
			reduce(196), // ^^, reduce: DotTerm
			reduce(196), // &&, reduce: DotTerm
			reduce(196), // ??, reduce: DotTerm
			reduce(196), // ???, reduce: DotTerm
			reduce(196), // =~, reduce: DotTerm
			reduce(196), // !=~, reduce: DotTerm
			reduce(196), // ==, reduce: DotTerm
			reduce(196), // !=, reduce: DotTerm
			reduce(196), // <=>, reduce: DotTerm
			reduce(196), // >=, reduce: DotTerm
			reduce(196), // <, reduce: DotTerm
			reduce(196), // <=, reduce: DotTerm
			reduce(196), // ^, reduce: DotTerm
			reduce(196), // &, reduce: DotTerm
			reduce(196), // <<, reduce: DotTerm
			reduce(196), // >>>, reduce: DotTerm
			reduce(196), // +, reduce: DotTerm
			reduce(196), // -, reduce: DotTerm
			reduce(196), // .+, reduce: DotTerm
			reduce(196), // .-, reduce: DotTerm
			reduce(196), // *, reduce: DotTerm
			reduce(196), // /, reduce: DotTerm
			reduce(196), // //, reduce: DotTerm
			reduce(196), // %, reduce: DotTerm
			reduce(196), // .*, reduce: DotTerm
			reduce(196), // ./, reduce: DotTerm
			reduce(196), // .//, reduce: DotTerm
			reduce(196), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S434
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S435
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S436
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(203), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(203), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(203), // >, reduce: UnaryOpTerm
			reduce(203), // >>, reduce: UnaryOpTerm
			reduce(203), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(203), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(203), // ||, reduce: UnaryOpTerm
			reduce(203), // ^^, reduce: UnaryOpTerm
			reduce(203), // &&, reduce: UnaryOpTerm
			reduce(203), // ??, reduce: UnaryOpTerm
			reduce(203), // ???, reduce: UnaryOpTerm
			reduce(203), // =~, reduce: UnaryOpTerm
			reduce(203), // !=~, reduce: UnaryOpTerm
			reduce(203), // ==, reduce: UnaryOpTerm
			reduce(203), // !=, reduce: UnaryOpTerm
			reduce(203), // <=>, reduce: UnaryOpTerm
			reduce(203), // >=, reduce: UnaryOpTerm
			reduce(203), // <, reduce: UnaryOpTerm
			reduce(203), // <=, reduce: UnaryOpTerm
			reduce(203), // ^, reduce: UnaryOpTerm
			reduce(203), // &, reduce: UnaryOpTerm
			reduce(203), // <<, reduce: UnaryOpTerm
			reduce(203), // >>>, reduce: UnaryOpTerm
			reduce(203), // +, reduce: UnaryOpTerm
			reduce(203), // -, reduce: UnaryOpTerm
			reduce(203), // .+, reduce: UnaryOpTerm
			reduce(203), // .-, reduce: UnaryOpTerm
			reduce(203), // *, reduce: UnaryOpTerm
			reduce(203), // /, reduce: UnaryOpTerm
			reduce(203), // //, reduce: UnaryOpTerm
			reduce(203), // %, reduce: UnaryOpTerm
			reduce(203), // .*, reduce: UnaryOpTerm
			reduce(203), // ./, reduce: UnaryOpTerm
			reduce(203), // .//, reduce: UnaryOpTerm
			reduce(203), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S437
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(207), // $, reduce: PowTerm
			nil,         // empty
			reduce(207), // ;, reduce: PowTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(207), // >, reduce: PowTerm
			reduce(207), // >>, reduce: PowTerm
			reduce(207), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(207), // ?, reduce: PowTerm
			nil,         // :
			reduce(207), // ||, reduce: PowTerm
			reduce(207), // ^^, reduce: PowTerm
			reduce(207), // &&, reduce: PowTerm
			reduce(207), // ??, reduce: PowTerm
			reduce(207), // ???, reduce: PowTerm
			reduce(207), // =~, reduce: PowTerm
			reduce(207), // !=~, reduce: PowTerm
			reduce(207), // ==, reduce: PowTerm
			reduce(207), // !=, reduce: PowTerm
			reduce(207), // <=>, reduce: PowTerm
			reduce(207), // >=, reduce: PowTerm
			reduce(207), // <, reduce: PowTerm
			reduce(207), // <=, reduce: PowTerm
			reduce(207), // ^, reduce: PowTerm
			reduce(207), // &, reduce: PowTerm
			reduce(207), // <<, reduce: PowTerm
			reduce(207), // >>>, reduce: PowTerm
			reduce(207), // +, reduce: PowTerm
			reduce(207), // -, reduce: PowTerm
			reduce(207), // .+, reduce: PowTerm
			reduce(207), // .-, reduce: PowTerm
			reduce(207), // *, reduce: PowTerm
			reduce(207), // /, reduce: PowTerm
			reduce(207), // //, reduce: PowTerm
			reduce(207), // %, reduce: PowTerm
			reduce(207), // .*, reduce: PowTerm
			reduce(207), // ./, reduce: PowTerm
			reduce(207), // .//, reduce: PowTerm
			reduce(207), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			shift(1206), // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S438
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(209), // $, reduce: PrecedenceChainEnd
			nil,         // empty
			reduce(209), // ;, reduce: PrecedenceChainEnd
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(209), // >, reduce: PrecedenceChainEnd
			reduce(209), // >>, reduce: PrecedenceChainEnd
			reduce(209), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(209), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(209), // ||, reduce: PrecedenceChainEnd
			reduce(209), // ^^, reduce: PrecedenceChainEnd
			reduce(209), // &&, reduce: PrecedenceChainEnd
			reduce(209), // ??, reduce: PrecedenceChainEnd
			reduce(209), // ???, reduce: PrecedenceChainEnd
			reduce(209), // =~, reduce: PrecedenceChainEnd
			reduce(209), // !=~, reduce: PrecedenceChainEnd
			reduce(209), // ==, reduce: PrecedenceChainEnd
			reduce(209), // !=, reduce: PrecedenceChainEnd
			reduce(209), // <=>, reduce: PrecedenceChainEnd
			reduce(209), // >=, reduce: PrecedenceChainEnd
			reduce(209), // <, reduce: PrecedenceChainEnd
			reduce(209), // <=, reduce: PrecedenceChainEnd
			reduce(209), // ^, reduce: PrecedenceChainEnd
			reduce(209), // &, reduce: PrecedenceChainEnd
			reduce(209), // <<, reduce: PrecedenceChainEnd
			reduce(209), // >>>, reduce: PrecedenceChainEnd
			reduce(209), // +, reduce: PrecedenceChainEnd
			reduce(209), // -, reduce: PrecedenceChainEnd
			reduce(209), // .+, reduce: PrecedenceChainEnd
			reduce(209), // .-, reduce: PrecedenceChainEnd
			reduce(209), // *, reduce: PrecedenceChainEnd
			reduce(209), // /, reduce: PrecedenceChainEnd
			reduce(209), // //, reduce: PrecedenceChainEnd
			reduce(209), // %, reduce: PrecedenceChainEnd
			reduce(209), // .*, reduce: PrecedenceChainEnd
			reduce(209), // ./, reduce: PrecedenceChainEnd
			reduce(209), // .//, reduce: PrecedenceChainEnd
			reduce(209), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(209), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(1207), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			shift(1208), // [[
			shift(1209), // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S439
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(215), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(215), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(215), // >, reduce: MlrvalOrFunction
			reduce(215), // >>, reduce: MlrvalOrFunction
			reduce(215), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(215), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(215), // ||, reduce: MlrvalOrFunction
			reduce(215), // ^^, reduce: MlrvalOrFunction
			reduce(215), // &&, reduce: MlrvalOrFunction
			reduce(215), // ??, reduce: MlrvalOrFunction
			reduce(215), // ???, reduce: MlrvalOrFunction
			reduce(215), // =~, reduce: MlrvalOrFunction
			reduce(215), // !=~, reduce: MlrvalOrFunction
			reduce(215), // ==, reduce: MlrvalOrFunction
			reduce(215), // !=, reduce: MlrvalOrFunction
			reduce(215), // <=>, reduce: MlrvalOrFunction
			reduce(215), // >=, reduce: MlrvalOrFunction
			reduce(215), // <, reduce: MlrvalOrFunction
			reduce(215), // <=, reduce: MlrvalOrFunction
			reduce(215), // ^, reduce: MlrvalOrFunction
			reduce(215), // &, reduce: MlrvalOrFunction
			reduce(215), // <<, reduce: MlrvalOrFunction
			reduce(215), // >>>, reduce: MlrvalOrFunction
			reduce(215), // +, reduce: MlrvalOrFunction
			reduce(215), // -, reduce: MlrvalOrFunction
			reduce(215), // .+, reduce: MlrvalOrFunction
			reduce(215), // .-, reduce: MlrvalOrFunction
			reduce(215), // *, reduce: MlrvalOrFunction
			reduce(215), // /, reduce: MlrvalOrFunction
			reduce(215), // //, reduce: MlrvalOrFunction
			reduce(215), // %, reduce: MlrvalOrFunction
			reduce(215), // .*, reduce: MlrvalOrFunction
			reduce(215), // ./, reduce: MlrvalOrFunction
			reduce(215), // .//, reduce: MlrvalOrFunction
			reduce(215), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(215), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(215), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(215), // [[, reduce: MlrvalOrFunction
			reduce(215), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S440
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(216), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(216), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(216), // >, reduce: MlrvalOrFunction
			reduce(216), // >>, reduce: MlrvalOrFunction
			reduce(216), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(216), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(216), // ||, reduce: MlrvalOrFunction
			reduce(216), // ^^, reduce: MlrvalOrFunction
			reduce(216), // &&, reduce: MlrvalOrFunction
			reduce(216), // ??, reduce: MlrvalOrFunction
			reduce(216), // ???, reduce: MlrvalOrFunction
			reduce(216), // =~, reduce: MlrvalOrFunction
			reduce(216), // !=~, reduce: MlrvalOrFunction
			reduce(216), // ==, reduce: MlrvalOrFunction
			reduce(216), // !=, reduce: MlrvalOrFunction
			reduce(216), // <=>, reduce: MlrvalOrFunction
			reduce(216), // >=, reduce: MlrvalOrFunction
			reduce(216), // <, reduce: MlrvalOrFunction
			reduce(216), // <=, reduce: MlrvalOrFunction
			reduce(216), // ^, reduce: MlrvalOrFunction
			reduce(216), // &, reduce: MlrvalOrFunction
			reduce(216), // <<, reduce: MlrvalOrFunction
			reduce(216), // >>>, reduce: MlrvalOrFunction
			reduce(216), // +, reduce: MlrvalOrFunction
			reduce(216), // -, reduce: MlrvalOrFunction
			reduce(216), // .+, reduce: MlrvalOrFunction
			reduce(216), // .-, reduce: MlrvalOrFunction
			reduce(216), // *, reduce: MlrvalOrFunction
			reduce(216), // /, reduce: MlrvalOrFunction
			reduce(216), // //, reduce: MlrvalOrFunction
			reduce(216), // %, reduce: MlrvalOrFunction
			reduce(216), // .*, reduce: MlrvalOrFunction
			reduce(216), // ./, reduce: MlrvalOrFunction
			reduce(216), // .//, reduce: MlrvalOrFunction
			reduce(216), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(216), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(216), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(216), // [[, reduce: MlrvalOrFunction
			reduce(216), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S441
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(217), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(217), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(217), // >, reduce: MlrvalOrFunction
			reduce(217), // >>, reduce: MlrvalOrFunction
			reduce(217), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(217), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(217), // ||, reduce: MlrvalOrFunction
			reduce(217), // ^^, reduce: MlrvalOrFunction
			reduce(217), // &&, reduce: MlrvalOrFunction
			reduce(217), // ??, reduce: MlrvalOrFunction
			reduce(217), // ???, reduce: MlrvalOrFunction
			reduce(217), // =~, reduce: MlrvalOrFunction
			reduce(217), // !=~, reduce: MlrvalOrFunction
			reduce(217), // ==, reduce: MlrvalOrFunction
			reduce(217), // !=, reduce: MlrvalOrFunction
			reduce(217), // <=>, reduce: MlrvalOrFunction
			reduce(217), // >=, reduce: MlrvalOrFunction
			reduce(217), // <, reduce: MlrvalOrFunction
			reduce(217), // <=, reduce: MlrvalOrFunction
			reduce(217), // ^, reduce: MlrvalOrFunction
			reduce(217), // &, reduce: MlrvalOrFunction
			reduce(217), // <<, reduce: MlrvalOrFunction
			reduce(217), // >>>, reduce: MlrvalOrFunction
			reduce(217), // +, reduce: MlrvalOrFunction
			reduce(217), // -, reduce: MlrvalOrFunction
			reduce(217), // .+, reduce: MlrvalOrFunction
			reduce(217), // .-, reduce: MlrvalOrFunction
			reduce(217), // *, reduce: MlrvalOrFunction
			reduce(217), // /, reduce: MlrvalOrFunction
			reduce(217), // //, reduce: MlrvalOrFunction
			reduce(217), // %, reduce: MlrvalOrFunction
			reduce(217), // .*, reduce: MlrvalOrFunction
			reduce(217), // ./, reduce: MlrvalOrFunction
			reduce(217), // .//, reduce: MlrvalOrFunction
			reduce(217), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(217), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(217), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(217), // [[, reduce: MlrvalOrFunction
			reduce(217), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S442
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(218), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(218), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(218), // >, reduce: MlrvalOrFunction
			reduce(218), // >>, reduce: MlrvalOrFunction
			reduce(218), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(218), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(218), // ||, reduce: MlrvalOrFunction
			reduce(218), // ^^, reduce: MlrvalOrFunction
			reduce(218), // &&, reduce: MlrvalOrFunction
			reduce(218), // ??, reduce: MlrvalOrFunction
			reduce(218), // ???, reduce: MlrvalOrFunction
			reduce(218), // =~, reduce: MlrvalOrFunction
			reduce(218), // !=~, reduce: MlrvalOrFunction
			reduce(218), // ==, reduce: MlrvalOrFunction
			reduce(218), // !=, reduce: MlrvalOrFunction
			reduce(218), // <=>, reduce: MlrvalOrFunction
			reduce(218), // >=, reduce: MlrvalOrFunction
			reduce(218), // <, reduce: MlrvalOrFunction
			reduce(218), // <=, reduce: MlrvalOrFunction
			reduce(218), // ^, reduce: MlrvalOrFunction
			reduce(218), // &, reduce: MlrvalOrFunction
			reduce(218), // <<, reduce: MlrvalOrFunction
			reduce(218), // >>>, reduce: MlrvalOrFunction
			reduce(218), // +, reduce: MlrvalOrFunction
			reduce(218), // -, reduce: MlrvalOrFunction
			reduce(218), // .+, reduce: MlrvalOrFunction
			reduce(218), // .-, reduce: MlrvalOrFunction
			reduce(218), // *, reduce: MlrvalOrFunction
			reduce(218), // /, reduce: MlrvalOrFunction
			reduce(218), // //, reduce: MlrvalOrFunction
			reduce(218), // %, reduce: MlrvalOrFunction
			reduce(218), // .*, reduce: MlrvalOrFunction
			reduce(218), // ./, reduce: MlrvalOrFunction
			reduce(218), // .//, reduce: MlrvalOrFunction
			reduce(218), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(218), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(218), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(218), // [[, reduce: MlrvalOrFunction
			reduce(218), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S443
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(219), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(219), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(219), // >, reduce: MlrvalOrFunction
			reduce(219), // >>, reduce: MlrvalOrFunction
			reduce(219), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(219), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(219), // ||, reduce: MlrvalOrFunction
			reduce(219), // ^^, reduce: MlrvalOrFunction
			reduce(219), // &&, reduce: MlrvalOrFunction
			reduce(219), // ??, reduce: MlrvalOrFunction
			reduce(219), // ???, reduce: MlrvalOrFunction
			reduce(219), // =~, reduce: MlrvalOrFunction
			reduce(219), // !=~, reduce: MlrvalOrFunction
			reduce(219), // ==, reduce: MlrvalOrFunction
			reduce(219), // !=, reduce: MlrvalOrFunction
			reduce(219), // <=>, reduce: MlrvalOrFunction
			reduce(219), // >=, reduce: MlrvalOrFunction
			reduce(219), // <, reduce: MlrvalOrFunction
			reduce(219), // <=, reduce: MlrvalOrFunction
			reduce(219), // ^, reduce: MlrvalOrFunction
			reduce(219), // &, reduce: MlrvalOrFunction
			reduce(219), // <<, reduce: MlrvalOrFunction
			reduce(219), // >>>, reduce: MlrvalOrFunction
			reduce(219), // +, reduce: MlrvalOrFunction
			reduce(219), // -, reduce: MlrvalOrFunction
			reduce(219), // .+, reduce: MlrvalOrFunction
			reduce(219), // .-, reduce: MlrvalOrFunction
			reduce(219), // *, reduce: MlrvalOrFunction
			reduce(219), // /, reduce: MlrvalOrFunction
			reduce(219), // //, reduce: MlrvalOrFunction
			reduce(219), // %, reduce: MlrvalOrFunction
			reduce(219), // .*, reduce: MlrvalOrFunction
			reduce(219), // ./, reduce: MlrvalOrFunction
			reduce(219), // .//, reduce: MlrvalOrFunction
			reduce(219), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(219), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(219), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(219), // [[, reduce: MlrvalOrFunction
			reduce(219), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S444
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(220), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(220), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(220), // >, reduce: MlrvalOrFunction
			reduce(220), // >>, reduce: MlrvalOrFunction
			reduce(220), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(220), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(220), // ||, reduce: MlrvalOrFunction
			reduce(220), // ^^, reduce: MlrvalOrFunction
			reduce(220), // &&, reduce: MlrvalOrFunction
			reduce(220), // ??, reduce: MlrvalOrFunction
			reduce(220), // ???, reduce: MlrvalOrFunction
			reduce(220), // =~, reduce: MlrvalOrFunction
			reduce(220), // !=~, reduce: MlrvalOrFunction
			reduce(220), // ==, reduce: MlrvalOrFunction
			reduce(220), // !=, reduce: MlrvalOrFunction
			reduce(220), // <=>, reduce: MlrvalOrFunction
			reduce(220), // >=, reduce: MlrvalOrFunction
			reduce(220), // <, reduce: MlrvalOrFunction
			reduce(220), // <=, reduce: MlrvalOrFunction
			reduce(220), // ^, reduce: MlrvalOrFunction
			reduce(220), // &, reduce: MlrvalOrFunction
			reduce(220), // <<, reduce: MlrvalOrFunction
			reduce(220), // >>>, reduce: MlrvalOrFunction
			reduce(220), // +, reduce: MlrvalOrFunction
			reduce(220), // -, reduce: MlrvalOrFunction
			reduce(220), // .+, reduce: MlrvalOrFunction
			reduce(220), // .-, reduce: MlrvalOrFunction
			reduce(220), // *, reduce: MlrvalOrFunction
			reduce(220), // /, reduce: MlrvalOrFunction
			reduce(220), // //, reduce: MlrvalOrFunction
			reduce(220), // %, reduce: MlrvalOrFunction
			reduce(220), // .*, reduce: MlrvalOrFunction
			reduce(220), // ./, reduce: MlrvalOrFunction
			reduce(220), // .//, reduce: MlrvalOrFunction
			reduce(220), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(220), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(220), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(220), // [[, reduce: MlrvalOrFunction
			reduce(220), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S445
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(221), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(221), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(221), // >, reduce: MlrvalOrFunction
			reduce(221), // >>, reduce: MlrvalOrFunction
			reduce(221), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(221), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(221), // ||, reduce: MlrvalOrFunction
			reduce(221), // ^^, reduce: MlrvalOrFunction
			reduce(221), // &&, reduce: MlrvalOrFunction
			reduce(221), // ??, reduce: MlrvalOrFunction
			reduce(221), // ???, reduce: MlrvalOrFunction
			reduce(221), // =~, reduce: MlrvalOrFunction
			reduce(221), // !=~, reduce: MlrvalOrFunction
			reduce(221), // ==, reduce: MlrvalOrFunction
			reduce(221), // !=, reduce: MlrvalOrFunction
			reduce(221), // <=>, reduce: MlrvalOrFunction
			reduce(221), // >=, reduce: MlrvalOrFunction
			reduce(221), // <, reduce: MlrvalOrFunction
			reduce(221), // <=, reduce: MlrvalOrFunction
			reduce(221), // ^, reduce: MlrvalOrFunction
			reduce(221), // &, reduce: MlrvalOrFunction
			reduce(221), // <<, reduce: MlrvalOrFunction
			reduce(221), // >>>, reduce: MlrvalOrFunction
			reduce(221), // +, reduce: MlrvalOrFunction
			reduce(221), // -, reduce: MlrvalOrFunction
			reduce(221), // .+, reduce: MlrvalOrFunction
			reduce(221), // .-, reduce: MlrvalOrFunction
			reduce(221), // *, reduce: MlrvalOrFunction
			reduce(221), // /, reduce: MlrvalOrFunction
			reduce(221), // //, reduce: MlrvalOrFunction
			reduce(221), // %, reduce: MlrvalOrFunction
			reduce(221), // .*, reduce: MlrvalOrFunction
			reduce(221), // ./, reduce: MlrvalOrFunction
			reduce(221), // .//, reduce: MlrvalOrFunction
			reduce(221), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(221), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(221), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(221), // [[, reduce: MlrvalOrFunction
			reduce(221), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S446
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(222), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(222), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(222), // >, reduce: MlrvalOrFunction
			reduce(222), // >>, reduce: MlrvalOrFunction
			reduce(222), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(222), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(222), // ||, reduce: MlrvalOrFunction
			reduce(222), // ^^, reduce: MlrvalOrFunction
			reduce(222), // &&, reduce: MlrvalOrFunction
			reduce(222), // ??, reduce: MlrvalOrFunction
			reduce(222), // ???, reduce: MlrvalOrFunction
			reduce(222), // =~, reduce: MlrvalOrFunction
			reduce(222), // !=~, reduce: MlrvalOrFunction
			reduce(222), // ==, reduce: MlrvalOrFunction
			reduce(222), // !=, reduce: MlrvalOrFunction
			reduce(222), // <=>, reduce: MlrvalOrFunction
			reduce(222), // >=, reduce: MlrvalOrFunction
			reduce(222), // <, reduce: MlrvalOrFunction
			reduce(222), // <=, reduce: MlrvalOrFunction
			reduce(222), // ^, reduce: MlrvalOrFunction
			reduce(222), // &, reduce: MlrvalOrFunction
			reduce(222), // <<, reduce: MlrvalOrFunction
			reduce(222), // >>>, reduce: MlrvalOrFunction
			reduce(222), // +, reduce: MlrvalOrFunction
			reduce(222), // -, reduce: MlrvalOrFunction
			reduce(222), // .+, reduce: MlrvalOrFunction
			reduce(222), // .-, reduce: MlrvalOrFunction
			reduce(222), // *, reduce: MlrvalOrFunction
			reduce(222), // /, reduce: MlrvalOrFunction
			reduce(222), // //, reduce: MlrvalOrFunction
			reduce(222), // %, reduce: MlrvalOrFunction
			reduce(222), // .*, reduce: MlrvalOrFunction
			reduce(222), // ./, reduce: MlrvalOrFunction
			reduce(222), // .//, reduce: MlrvalOrFunction
			reduce(222), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(222), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(222), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(222), // [[, reduce: MlrvalOrFunction
			reduce(222), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S447
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(223), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(223), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(223), // >, reduce: MlrvalOrFunction
			reduce(223), // >>, reduce: MlrvalOrFunction
			reduce(223), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(223), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(223), // ||, reduce: MlrvalOrFunction
			reduce(223), // ^^, reduce: MlrvalOrFunction
			reduce(223), // &&, reduce: MlrvalOrFunction
			reduce(223), // ??, reduce: MlrvalOrFunction
			reduce(223), // ???, reduce: MlrvalOrFunction
			reduce(223), // =~, reduce: MlrvalOrFunction
			reduce(223), // !=~, reduce: MlrvalOrFunction
			reduce(223), // ==, reduce: MlrvalOrFunction
			reduce(223), // !=, reduce: MlrvalOrFunction
			reduce(223), // <=>, reduce: MlrvalOrFunction
			reduce(223), // >=, reduce: MlrvalOrFunction
			reduce(223), // <, reduce: MlrvalOrFunction
			reduce(223), // <=, reduce: MlrvalOrFunction
			reduce(223), // ^, reduce: MlrvalOrFunction
			reduce(223), // &, reduce: MlrvalOrFunction
			reduce(223), // <<, reduce: MlrvalOrFunction
			reduce(223), // >>>, reduce: MlrvalOrFunction
			reduce(223), // +, reduce: MlrvalOrFunction
			reduce(223), // -, reduce: MlrvalOrFunction
			reduce(223), // .+, reduce: MlrvalOrFunction
			reduce(223), // .-, reduce: MlrvalOrFunction
			reduce(223), // *, reduce: MlrvalOrFunction
			reduce(223), // /, reduce: MlrvalOrFunction
			reduce(223), // //, reduce: MlrvalOrFunction
			reduce(223), // %, reduce: MlrvalOrFunction
			reduce(223), // .*, reduce: MlrvalOrFunction
			reduce(223), // ./, reduce: MlrvalOrFunction
			reduce(223), // .//, reduce: MlrvalOrFunction
			reduce(223), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(223), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(223), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(223), // [[, reduce: MlrvalOrFunction
			reduce(223), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S448
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(224), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(224), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(224), // >, reduce: MlrvalOrFunction
			reduce(224), // >>, reduce: MlrvalOrFunction
			reduce(224), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(224), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(224), // ||, reduce: MlrvalOrFunction
			reduce(224), // ^^, reduce: MlrvalOrFunction
			reduce(224), // &&, reduce: MlrvalOrFunction
			reduce(224), // ??, reduce: MlrvalOrFunction
			reduce(224), // ???, reduce: MlrvalOrFunction
			reduce(224), // =~, reduce: MlrvalOrFunction
			reduce(224), // !=~, reduce: MlrvalOrFunction
			reduce(224), // ==, reduce: MlrvalOrFunction
			reduce(224), // !=, reduce: MlrvalOrFunction
			reduce(224), // <=>, reduce: MlrvalOrFunction
			reduce(224), // >=, reduce: MlrvalOrFunction
			reduce(224), // <, reduce: MlrvalOrFunction
			reduce(224), // <=, reduce: MlrvalOrFunction
			reduce(224), // ^, reduce: MlrvalOrFunction
			reduce(224), // &, reduce: MlrvalOrFunction
			reduce(224), // <<, reduce: MlrvalOrFunction
			reduce(224), // >>>, reduce: MlrvalOrFunction
			reduce(224), // +, reduce: MlrvalOrFunction
			reduce(224), // -, reduce: MlrvalOrFunction
			reduce(224), // .+, reduce: MlrvalOrFunction
			reduce(224), // .-, reduce: MlrvalOrFunction
			reduce(224), // *, reduce: MlrvalOrFunction
			reduce(224), // /, reduce: MlrvalOrFunction
			reduce(224), // //, reduce: MlrvalOrFunction
			reduce(224), // %, reduce: MlrvalOrFunction
			reduce(224), // .*, reduce: MlrvalOrFunction
			reduce(224), // ./, reduce: MlrvalOrFunction
			reduce(224), // .//, reduce: MlrvalOrFunction
			reduce(224), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(224), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(224), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(224), // [[, reduce: MlrvalOrFunction
			reduce(224), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S449
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(225), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(225), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(225), // >, reduce: MlrvalOrFunction
			reduce(225), // >>, reduce: MlrvalOrFunction
			reduce(225), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(225), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(225), // ||, reduce: MlrvalOrFunction
			reduce(225), // ^^, reduce: MlrvalOrFunction
			reduce(225), // &&, reduce: MlrvalOrFunction
			reduce(225), // ??, reduce: MlrvalOrFunction
			reduce(225), // ???, reduce: MlrvalOrFunction
			reduce(225), // =~, reduce: MlrvalOrFunction
			reduce(225), // !=~, reduce: MlrvalOrFunction
			reduce(225), // ==, reduce: MlrvalOrFunction
			reduce(225), // !=, reduce: MlrvalOrFunction
			reduce(225), // <=>, reduce: MlrvalOrFunction
			reduce(225), // >=, reduce: MlrvalOrFunction
			reduce(225), // <, reduce: MlrvalOrFunction
			reduce(225), // <=, reduce: MlrvalOrFunction
			reduce(225), // ^, reduce: MlrvalOrFunction
			reduce(225), // &, reduce: MlrvalOrFunction
			reduce(225), // <<, reduce: MlrvalOrFunction
			reduce(225), // >>>, reduce: MlrvalOrFunction
			reduce(225), // +, reduce: MlrvalOrFunction
			reduce(225), // -, reduce: MlrvalOrFunction
			reduce(225), // .+, reduce: MlrvalOrFunction
			reduce(225), // .-, reduce: MlrvalOrFunction
			reduce(225), // *, reduce: MlrvalOrFunction
			reduce(225), // /, reduce: MlrvalOrFunction
			reduce(225), // //, reduce: MlrvalOrFunction
			reduce(225), // %, reduce: MlrvalOrFunction
			reduce(225), // .*, reduce: MlrvalOrFunction
			reduce(225), // ./, reduce: MlrvalOrFunction
			reduce(225), // .//, reduce: MlrvalOrFunction
			reduce(225), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(225), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(225), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(225), // [[, reduce: MlrvalOrFunction
			reduce(225), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S450
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(226), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(226), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(226), // >, reduce: MlrvalOrFunction
			reduce(226), // >>, reduce: MlrvalOrFunction
			reduce(226), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(226), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(226), // ||, reduce: MlrvalOrFunction
			reduce(226), // ^^, reduce: MlrvalOrFunction
			reduce(226), // &&, reduce: MlrvalOrFunction
			reduce(226), // ??, reduce: MlrvalOrFunction
			reduce(226), // ???, reduce: MlrvalOrFunction
			reduce(226), // =~, reduce: MlrvalOrFunction
			reduce(226), // !=~, reduce: MlrvalOrFunction
			reduce(226), // ==, reduce: MlrvalOrFunction
			reduce(226), // !=, reduce: MlrvalOrFunction
			reduce(226), // <=>, reduce: MlrvalOrFunction
			reduce(226), // >=, reduce: MlrvalOrFunction
			reduce(226), // <, reduce: MlrvalOrFunction
			reduce(226), // <=, reduce: MlrvalOrFunction
			reduce(226), // ^, reduce: MlrvalOrFunction
			reduce(226), // &, reduce: MlrvalOrFunction
			reduce(226), // <<, reduce: MlrvalOrFunction
			reduce(226), // >>>, reduce: MlrvalOrFunction
			reduce(226), // +, reduce: MlrvalOrFunction
			reduce(226), // -, reduce: MlrvalOrFunction
			reduce(226), // .+, reduce: MlrvalOrFunction
			reduce(226), // .-, reduce: MlrvalOrFunction
			reduce(226), // *, reduce: MlrvalOrFunction
			reduce(226), // /, reduce: MlrvalOrFunction
			reduce(226), // //, reduce: MlrvalOrFunction
			reduce(226), // %, reduce: MlrvalOrFunction
			reduce(226), // .*, reduce: MlrvalOrFunction
			reduce(226), // ./, reduce: MlrvalOrFunction
			reduce(226), // .//, reduce: MlrvalOrFunction
			reduce(226), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(226), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(226), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(226), // [[, reduce: MlrvalOrFunction
			reduce(226), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S451
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(227), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(227), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(227), // >, reduce: MlrvalOrFunction
			reduce(227), // >>, reduce: MlrvalOrFunction
			reduce(227), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(227), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(227), // ||, reduce: MlrvalOrFunction
			reduce(227), // ^^, reduce: MlrvalOrFunction
			reduce(227), // &&, reduce: MlrvalOrFunction
			reduce(227), // ??, reduce: MlrvalOrFunction
			reduce(227), // ???, reduce: MlrvalOrFunction
			reduce(227), // =~, reduce: MlrvalOrFunction
			reduce(227), // !=~, reduce: MlrvalOrFunction
			reduce(227), // ==, reduce: MlrvalOrFunction
			reduce(227), // !=, reduce: MlrvalOrFunction
			reduce(227), // <=>, reduce: MlrvalOrFunction
			reduce(227), // >=, reduce: MlrvalOrFunction
			reduce(227), // <, reduce: MlrvalOrFunction
			reduce(227), // <=, reduce: MlrvalOrFunction
			reduce(227), // ^, reduce: MlrvalOrFunction
			reduce(227), // &, reduce: MlrvalOrFunction
			reduce(227), // <<, reduce: MlrvalOrFunction
			reduce(227), // >>>, reduce: MlrvalOrFunction
			reduce(227), // +, reduce: MlrvalOrFunction
			reduce(227), // -, reduce: MlrvalOrFunction
			reduce(227), // .+, reduce: MlrvalOrFunction
			reduce(227), // .-, reduce: MlrvalOrFunction
			reduce(227), // *, reduce: MlrvalOrFunction
			reduce(227), // /, reduce: MlrvalOrFunction
			reduce(227), // //, reduce: MlrvalOrFunction
			reduce(227), // %, reduce: MlrvalOrFunction
			reduce(227), // .*, reduce: MlrvalOrFunction
			reduce(227), // ./, reduce: MlrvalOrFunction
			reduce(227), // .//, reduce: MlrvalOrFunction
			reduce(227), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(227), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(227), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(227), // [[, reduce: MlrvalOrFunction
			reduce(227), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S452
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(752),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(756),  // (
			nil,         // )
			nil,         // emitp
			shift(768),  // field_name
			shift(769),  // $[
			shift(1210), // ]
			shift(771),  // braced_field_name
			shift(772),  // $[[
			shift(773),  // $[[[
			shift(774),  // full_srec
			shift(776),  // oosvar_name
			shift(777),  // @[
			shift(778),  // braced_oosvar_name
			shift(779),  // full_oosvar
			shift(780),  // all
			shift(781),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(796),  // +
			shift(798),  // -
			shift(799),  // .+
			shift(800),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(803),  // !
			shift(804),  // ~
			nil,         // **
			shift(809),  // string_literal
			shift(810),  // regex_case_insensitive
			shift(811),  // int_literal
			shift(812),  // float_literal
			shift(813),  // boolean_literal
			shift(814),  // null_literal
			shift(815),  // inf_literal
			shift(816),  // nan_literal
			shift(817),  // const_M_PI
			shift(818),  // const_M_E
			shift(819),  // panic
			shift(821),  // [
			shift(824),  // ctx_IPS
			shift(825),  // ctx_IFS
			shift(826),  // ctx_IRS
			shift(827),  // ctx_OPS
			shift(828),  // ctx_OFS
			shift(829),  // ctx_ORS
			shift(830),  // ctx_FLATSEP
			shift(831),  // ctx_NF
			shift(832),  // ctx_NR
			shift(833),  // ctx_FNR
			shift(834),  // ctx_FILENAME
			shift(835),  // ctx_FILENUM
			shift(837),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(844),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S453
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(240), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(240), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(240), // >, reduce: MlrvalOrFunction
			reduce(240), // >>, reduce: MlrvalOrFunction
			reduce(240), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(240), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(240), // ||, reduce: MlrvalOrFunction
			reduce(240), // ^^, reduce: MlrvalOrFunction
			reduce(240), // &&, reduce: MlrvalOrFunction
			reduce(240), // ??, reduce: MlrvalOrFunction
			reduce(240), // ???, reduce: MlrvalOrFunction
			reduce(240), // =~, reduce: MlrvalOrFunction
			reduce(240), // !=~, reduce: MlrvalOrFunction
			reduce(240), // ==, reduce: MlrvalOrFunction
			reduce(240), // !=, reduce: MlrvalOrFunction
			reduce(240), // <=>, reduce: MlrvalOrFunction
			reduce(240), // >=, reduce: MlrvalOrFunction
			reduce(240), // <, reduce: MlrvalOrFunction
			reduce(240), // <=, reduce: MlrvalOrFunction
			reduce(240), // ^, reduce: MlrvalOrFunction
			reduce(240), // &, reduce: MlrvalOrFunction
			reduce(240), // <<, reduce: MlrvalOrFunction
			reduce(240), // >>>, reduce: MlrvalOrFunction
			reduce(240), // +, reduce: MlrvalOrFunction
			reduce(240), // -, reduce: MlrvalOrFunction
			reduce(240), // .+, reduce: MlrvalOrFunction
			reduce(240), // .-, reduce: MlrvalOrFunction
			reduce(240), // *, reduce: MlrvalOrFunction
			reduce(240), // /, reduce: MlrvalOrFunction
			reduce(240), // //, reduce: MlrvalOrFunction
			reduce(240), // %, reduce: MlrvalOrFunction
			reduce(240), // .*, reduce: MlrvalOrFunction
			reduce(240), // ./, reduce: MlrvalOrFunction
			reduce(240), // .//, reduce: MlrvalOrFunction
			reduce(240), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(240), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(240), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(240), // [[, reduce: MlrvalOrFunction
			reduce(240), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S454
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(241), // $, reduce: ContextVariable
			nil,         // empty
			reduce(241), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(241), // >, reduce: ContextVariable
			reduce(241), // >>, reduce: ContextVariable
			reduce(241), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(241), // ?, reduce: ContextVariable
			nil,         // :
			reduce(241), // ||, reduce: ContextVariable
			reduce(241), // ^^, reduce: ContextVariable
			reduce(241), // &&, reduce: ContextVariable
			reduce(241), // ??, reduce: ContextVariable
			reduce(241), // ???, reduce: ContextVariable
			reduce(241), // =~, reduce: ContextVariable
			reduce(241), // !=~, reduce: ContextVariable
			reduce(241), // ==, reduce: ContextVariable
			reduce(241), // !=, reduce: ContextVariable
			reduce(241), // <=>, reduce: ContextVariable
			reduce(241), // >=, reduce: ContextVariable
			reduce(241), // <, reduce: ContextVariable
			reduce(241), // <=, reduce: ContextVariable
			reduce(241), // ^, reduce: ContextVariable
			reduce(241), // &, reduce: ContextVariable
			reduce(241), // <<, reduce: ContextVariable
			reduce(241), // >>>, reduce: ContextVariable
			reduce(241), // +, reduce: ContextVariable
			reduce(241), // -, reduce: ContextVariable
			reduce(241), // .+, reduce: ContextVariable
			reduce(241), // .-, reduce: ContextVariable
			reduce(241), // *, reduce: ContextVariable
			reduce(241), // /, reduce: ContextVariable
			reduce(241), // //, reduce: ContextVariable
			reduce(241), // %, reduce: ContextVariable
			reduce(241), // .*, reduce: ContextVariable
			reduce(241), // ./, reduce: ContextVariable
			reduce(241), // .//, reduce: ContextVariable
			reduce(241), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(241), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(241), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(241), // [[, reduce: ContextVariable
			reduce(241), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S455
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(242), // $, reduce: ContextVariable
			nil,         // empty
			reduce(242), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(242), // >, reduce: ContextVariable
			reduce(242), // >>, reduce: ContextVariable
			reduce(242), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(242), // ?, reduce: ContextVariable
			nil,         // :
			reduce(242), // ||, reduce: ContextVariable
			reduce(242), // ^^, reduce: ContextVariable
			reduce(242), // &&, reduce: ContextVariable
			reduce(242), // ??, reduce: ContextVariable
			reduce(242), // ???, reduce: ContextVariable
			reduce(242), // =~, reduce: ContextVariable
			reduce(242), // !=~, reduce: ContextVariable
			reduce(242), // ==, reduce: ContextVariable
			reduce(242), // !=, reduce: ContextVariable
			reduce(242), // <=>, reduce: ContextVariable
			reduce(242), // >=, reduce: ContextVariable
			reduce(242), // <, reduce: ContextVariable
			reduce(242), // <=, reduce: ContextVariable
			reduce(242), // ^, reduce: ContextVariable
			reduce(242), // &, reduce: ContextVariable
			reduce(242), // <<, reduce: ContextVariable
			reduce(242), // >>>, reduce: ContextVariable
			reduce(242), // +, reduce: ContextVariable
			reduce(242), // -, reduce: ContextVariable
			reduce(242), // .+, reduce: ContextVariable
			reduce(242), // .-, reduce: ContextVariable
			reduce(242), // *, reduce: ContextVariable
			reduce(242), // /, reduce: ContextVariable
			reduce(242), // //, reduce: ContextVariable
			reduce(242), // %, reduce: ContextVariable
			reduce(242), // .*, reduce: ContextVariable
			reduce(242), // ./, reduce: ContextVariable
			reduce(242), // .//, reduce: ContextVariable
			reduce(242), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(242), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(242), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(242), // [[, reduce: ContextVariable
			reduce(242), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S456
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(243), // $, reduce: ContextVariable
			nil,         // empty
			reduce(243), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(243), // >, reduce: ContextVariable
			reduce(243), // >>, reduce: ContextVariable
			reduce(243), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(243), // ?, reduce: ContextVariable
			nil,         // :
			reduce(243), // ||, reduce: ContextVariable
			reduce(243), // ^^, reduce: ContextVariable
			reduce(243), // &&, reduce: ContextVariable
			reduce(243), // ??, reduce: ContextVariable
			reduce(243), // ???, reduce: ContextVariable
			reduce(243), // =~, reduce: ContextVariable
			reduce(243), // !=~, reduce: ContextVariable
			reduce(243), // ==, reduce: ContextVariable
			reduce(243), // !=, reduce: ContextVariable
			reduce(243), // <=>, reduce: ContextVariable
			reduce(243), // >=, reduce: ContextVariable
			reduce(243), // <, reduce: ContextVariable
			reduce(243), // <=, reduce: ContextVariable
			reduce(243), // ^, reduce: ContextVariable
			reduce(243), // &, reduce: ContextVariable
			reduce(243), // <<, reduce: ContextVariable
			reduce(243), // >>>, reduce: ContextVariable
			reduce(243), // +, reduce: ContextVariable
			reduce(243), // -, reduce: ContextVariable
			reduce(243), // .+, reduce: ContextVariable
			reduce(243), // .-, reduce: ContextVariable
			reduce(243), // *, reduce: ContextVariable
			reduce(243), // /, reduce: ContextVariable
			reduce(243), // //, reduce: ContextVariable
			reduce(243), // %, reduce: ContextVariable
			reduce(243), // .*, reduce: ContextVariable
			reduce(243), // ./, reduce: ContextVariable
			reduce(243), // .//, reduce: ContextVariable
			reduce(243), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(243), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(243), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(243), // [[, reduce: ContextVariable
			reduce(243), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S457
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(244), // $, reduce: ContextVariable
			nil,         // empty
			reduce(244), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(244), // >, reduce: ContextVariable
			reduce(244), // >>, reduce: ContextVariable
			reduce(244), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(244), // ?, reduce: ContextVariable
			nil,         // :
			reduce(244), // ||, reduce: ContextVariable
			reduce(244), // ^^, reduce: ContextVariable
			reduce(244), // &&, reduce: ContextVariable
			reduce(244), // ??, reduce: ContextVariable
			reduce(244), // ???, reduce: ContextVariable
			reduce(244), // =~, reduce: ContextVariable
			reduce(244), // !=~, reduce: ContextVariable
			reduce(244), // ==, reduce: ContextVariable
			reduce(244), // !=, reduce: ContextVariable
			reduce(244), // <=>, reduce: ContextVariable
			reduce(244), // >=, reduce: ContextVariable
			reduce(244), // <, reduce: ContextVariable
			reduce(244), // <=, reduce: ContextVariable
			reduce(244), // ^, reduce: ContextVariable
			reduce(244), // &, reduce: ContextVariable
			reduce(244), // <<, reduce: ContextVariable
			reduce(244), // >>>, reduce: ContextVariable
			reduce(244), // +, reduce: ContextVariable
			reduce(244), // -, reduce: ContextVariable
			reduce(244), // .+, reduce: ContextVariable
			reduce(244), // .-, reduce: ContextVariable
			reduce(244), // *, reduce: ContextVariable
			reduce(244), // /, reduce: ContextVariable
			reduce(244), // //, reduce: ContextVariable
			reduce(244), // %, reduce: ContextVariable
			reduce(244), // .*, reduce: ContextVariable
			reduce(244), // ./, reduce: ContextVariable
			reduce(244), // .//, reduce: ContextVariable
			reduce(244), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(244), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(244), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(244), // [[, reduce: ContextVariable
			reduce(244), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S458
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(245), // $, reduce: ContextVariable
			nil,         // empty
			reduce(245), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(245), // >, reduce: ContextVariable
			reduce(245), // >>, reduce: ContextVariable
			reduce(245), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(245), // ?, reduce: ContextVariable
			nil,         // :
			reduce(245), // ||, reduce: ContextVariable
			reduce(245), // ^^, reduce: ContextVariable
			reduce(245), // &&, reduce: ContextVariable
			reduce(245), // ??, reduce: ContextVariable
			reduce(245), // ???, reduce: ContextVariable
			reduce(245), // =~, reduce: ContextVariable
			reduce(245), // !=~, reduce: ContextVariable
			reduce(245), // ==, reduce: ContextVariable
			reduce(245), // !=, reduce: ContextVariable
			reduce(245), // <=>, reduce: ContextVariable
			reduce(245), // >=, reduce: ContextVariable
			reduce(245), // <, reduce: ContextVariable
			reduce(245), // <=, reduce: ContextVariable
			reduce(245), // ^, reduce: ContextVariable
			reduce(245), // &, reduce: ContextVariable
			reduce(245), // <<, reduce: ContextVariable
			reduce(245), // >>>, reduce: ContextVariable
			reduce(245), // +, reduce: ContextVariable
			reduce(245), // -, reduce: ContextVariable
			reduce(245), // .+, reduce: ContextVariable
			reduce(245), // .-, reduce: ContextVariable
			reduce(245), // *, reduce: ContextVariable
			reduce(245), // /, reduce: ContextVariable
			reduce(245), // //, reduce: ContextVariable
			reduce(245), // %, reduce: ContextVariable
			reduce(245), // .*, reduce: ContextVariable
			reduce(245), // ./, reduce: ContextVariable
			reduce(245), // .//, reduce: ContextVariable
			reduce(245), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(245), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(245), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(245), // [[, reduce: ContextVariable
			reduce(245), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S459
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(246), // $, reduce: ContextVariable
			nil,         // empty
			reduce(246), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(246), // >, reduce: ContextVariable
			reduce(246), // >>, reduce: ContextVariable
			reduce(246), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(246), // ?, reduce: ContextVariable
			nil,         // :
			reduce(246), // ||, reduce: ContextVariable
			reduce(246), // ^^, reduce: ContextVariable
			reduce(246), // &&, reduce: ContextVariable
			reduce(246), // ??, reduce: ContextVariable
			reduce(246), // ???, reduce: ContextVariable
			reduce(246), // =~, reduce: ContextVariable
			reduce(246), // !=~, reduce: ContextVariable
			reduce(246), // ==, reduce: ContextVariable
			reduce(246), // !=, reduce: ContextVariable
			reduce(246), // <=>, reduce: ContextVariable
			reduce(246), // >=, reduce: ContextVariable
			reduce(246), // <, reduce: ContextVariable
			reduce(246), // <=, reduce: ContextVariable
			reduce(246), // ^, reduce: ContextVariable
			reduce(246), // &, reduce: ContextVariable
			reduce(246), // <<, reduce: ContextVariable
			reduce(246), // >>>, reduce: ContextVariable
			reduce(246), // +, reduce: ContextVariable
			reduce(246), // -, reduce: ContextVariable
			reduce(246), // .+, reduce: ContextVariable
			reduce(246), // .-, reduce: ContextVariable
			reduce(246), // *, reduce: ContextVariable
			reduce(246), // /, reduce: ContextVariable
			reduce(246), // //, reduce: ContextVariable
			reduce(246), // %, reduce: ContextVariable
			reduce(246), // .*, reduce: ContextVariable
			reduce(246), // ./, reduce: ContextVariable
			reduce(246), // .//, reduce: ContextVariable
			reduce(246), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(246), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(246), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(246), // [[, reduce: ContextVariable
			reduce(246), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S460
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(247), // $, reduce: ContextVariable
			nil,         // empty
			reduce(247), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(247), // >, reduce: ContextVariable
			reduce(247), // >>, reduce: ContextVariable
			reduce(247), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(247), // ?, reduce: ContextVariable
			nil,         // :
			reduce(247), // ||, reduce: ContextVariable
			reduce(247), // ^^, reduce: ContextVariable
			reduce(247), // &&, reduce: ContextVariable
			reduce(247), // ??, reduce: ContextVariable
			reduce(247), // ???, reduce: ContextVariable
			reduce(247), // =~, reduce: ContextVariable
			reduce(247), // !=~, reduce: ContextVariable
			reduce(247), // ==, reduce: ContextVariable
			reduce(247), // !=, reduce: ContextVariable
			reduce(247), // <=>, reduce: ContextVariable
			reduce(247), // >=, reduce: ContextVariable
			reduce(247), // <, reduce: ContextVariable
			reduce(247), // <=, reduce: ContextVariable
			reduce(247), // ^, reduce: ContextVariable
			reduce(247), // &, reduce: ContextVariable
			reduce(247), // <<, reduce: ContextVariable
			reduce(247), // >>>, reduce: ContextVariable
			reduce(247), // +, reduce: ContextVariable
			reduce(247), // -, reduce: ContextVariable
			reduce(247), // .+, reduce: ContextVariable
			reduce(247), // .-, reduce: ContextVariable
			reduce(247), // *, reduce: ContextVariable
			reduce(247), // /, reduce: ContextVariable
			reduce(247), // //, reduce: ContextVariable
			reduce(247), // %, reduce: ContextVariable
			reduce(247), // .*, reduce: ContextVariable
			reduce(247), // ./, reduce: ContextVariable
			reduce(247), // .//, reduce: ContextVariable
			reduce(247), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(247), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(247), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(247), // [[, reduce: ContextVariable
			reduce(247), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S461
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(248), // $, reduce: ContextVariable
			nil,         // empty
			reduce(248), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(248), // >, reduce: ContextVariable
			reduce(248), // >>, reduce: ContextVariable
			reduce(248), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(248), // ?, reduce: ContextVariable
			nil,         // :
			reduce(248), // ||, reduce: ContextVariable
			reduce(248), // ^^, reduce: ContextVariable
			reduce(248), // &&, reduce: ContextVariable
			reduce(248), // ??, reduce: ContextVariable
			reduce(248), // ???, reduce: ContextVariable
			reduce(248), // =~, reduce: ContextVariable
			reduce(248), // !=~, reduce: ContextVariable
			reduce(248), // ==, reduce: ContextVariable
			reduce(248), // !=, reduce: ContextVariable
			reduce(248), // <=>, reduce: ContextVariable
			reduce(248), // >=, reduce: ContextVariable
			reduce(248), // <, reduce: ContextVariable
			reduce(248), // <=, reduce: ContextVariable
			reduce(248), // ^, reduce: ContextVariable
			reduce(248), // &, reduce: ContextVariable
			reduce(248), // <<, reduce: ContextVariable
			reduce(248), // >>>, reduce: ContextVariable
			reduce(248), // +, reduce: ContextVariable
			reduce(248), // -, reduce: ContextVariable
			reduce(248), // .+, reduce: ContextVariable
			reduce(248), // .-, reduce: ContextVariable
			reduce(248), // *, reduce: ContextVariable
			reduce(248), // /, reduce: ContextVariable
			reduce(248), // //, reduce: ContextVariable
			reduce(248), // %, reduce: ContextVariable
			reduce(248), // .*, reduce: ContextVariable
			reduce(248), // ./, reduce: ContextVariable
			reduce(248), // .//, reduce: ContextVariable
			reduce(248), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(248), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(248), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(248), // [[, reduce: ContextVariable
			reduce(248), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S462
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(249), // $, reduce: ContextVariable
			nil,         // empty
			reduce(249), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(249), // >, reduce: ContextVariable
			reduce(249), // >>, reduce: ContextVariable
			reduce(249), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(249), // ?, reduce: ContextVariable
			nil,         // :
			reduce(249), // ||, reduce: ContextVariable
			reduce(249), // ^^, reduce: ContextVariable
			reduce(249), // &&, reduce: ContextVariable
			reduce(249), // ??, reduce: ContextVariable
			reduce(249), // ???, reduce: ContextVariable
			reduce(249), // =~, reduce: ContextVariable
			reduce(249), // !=~, reduce: ContextVariable
			reduce(249), // ==, reduce: ContextVariable
			reduce(249), // !=, reduce: ContextVariable
			reduce(249), // <=>, reduce: ContextVariable
			reduce(249), // >=, reduce: ContextVariable
			reduce(249), // <, reduce: ContextVariable
			reduce(249), // <=, reduce: ContextVariable
			reduce(249), // ^, reduce: ContextVariable
			reduce(249), // &, reduce: ContextVariable
			reduce(249), // <<, reduce: ContextVariable
			reduce(249), // >>>, reduce: ContextVariable
			reduce(249), // +, reduce: ContextVariable
			reduce(249), // -, reduce: ContextVariable
			reduce(249), // .+, reduce: ContextVariable
			reduce(249), // .-, reduce: ContextVariable
			reduce(249), // *, reduce: ContextVariable
			reduce(249), // /, reduce: ContextVariable
			reduce(249), // //, reduce: ContextVariable
			reduce(249), // %, reduce: ContextVariable
			reduce(249), // .*, reduce: ContextVariable
			reduce(249), // ./, reduce: ContextVariable
			reduce(249), // .//, reduce: ContextVariable
			reduce(249), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(249), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(249), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(249), // [[, reduce: ContextVariable
			reduce(249), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S463
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(250), // $, reduce: ContextVariable
			nil,         // empty
			reduce(250), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(250), // >, reduce: ContextVariable
			reduce(250), // >>, reduce: ContextVariable
			reduce(250), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(250), // ?, reduce: ContextVariable
			nil,         // :
			reduce(250), // ||, reduce: ContextVariable
			reduce(250), // ^^, reduce: ContextVariable
			reduce(250), // &&, reduce: ContextVariable
			reduce(250), // ??, reduce: ContextVariable
			reduce(250), // ???, reduce: ContextVariable
			reduce(250), // =~, reduce: ContextVariable
			reduce(250), // !=~, reduce: ContextVariable
			reduce(250), // ==, reduce: ContextVariable
			reduce(250), // !=, reduce: ContextVariable
			reduce(250), // <=>, reduce: ContextVariable
			reduce(250), // >=, reduce: ContextVariable
			reduce(250), // <, reduce: ContextVariable
			reduce(250), // <=, reduce: ContextVariable
			reduce(250), // ^, reduce: ContextVariable
			reduce(250), // &, reduce: ContextVariable
			reduce(250), // <<, reduce: ContextVariable
			reduce(250), // >>>, reduce: ContextVariable
			reduce(250), // +, reduce: ContextVariable
			reduce(250), // -, reduce: ContextVariable
			reduce(250), // .+, reduce: ContextVariable
			reduce(250), // .-, reduce: ContextVariable
			reduce(250), // *, reduce: ContextVariable
			reduce(250), // /, reduce: ContextVariable
			reduce(250), // //, reduce: ContextVariable
			reduce(250), // %, reduce: ContextVariable
			reduce(250), // .*, reduce: ContextVariable
			reduce(250), // ./, reduce: ContextVariable
			reduce(250), // .//, reduce: ContextVariable
			reduce(250), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(250), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(250), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(250), // [[, reduce: ContextVariable
			reduce(250), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S464
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(251), // $, reduce: ContextVariable
			nil,         // empty
			reduce(251), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(251), // >, reduce: ContextVariable
			reduce(251), // >>, reduce: ContextVariable
			reduce(251), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(251), // ?, reduce: ContextVariable
			nil,         // :
			reduce(251), // ||, reduce: ContextVariable
			reduce(251), // ^^, reduce: ContextVariable
			reduce(251), // &&, reduce: ContextVariable
			reduce(251), // ??, reduce: ContextVariable
			reduce(251), // ???, reduce: ContextVariable
			reduce(251), // =~, reduce: ContextVariable
			reduce(251), // !=~, reduce: ContextVariable
			reduce(251), // ==, reduce: ContextVariable
			reduce(251), // !=, reduce: ContextVariable
			reduce(251), // <=>, reduce: ContextVariable
			reduce(251), // >=, reduce: ContextVariable
			reduce(251), // <, reduce: ContextVariable
			reduce(251), // <=, reduce: ContextVariable
			reduce(251), // ^, reduce: ContextVariable
			reduce(251), // &, reduce: ContextVariable
			reduce(251), // <<, reduce: ContextVariable
			reduce(251), // >>>, reduce: ContextVariable
			reduce(251), // +, reduce: ContextVariable
			reduce(251), // -, reduce: ContextVariable
			reduce(251), // .+, reduce: ContextVariable
			reduce(251), // .-, reduce: ContextVariable
			reduce(251), // *, reduce: ContextVariable
			reduce(251), // /, reduce: ContextVariable
			reduce(251), // //, reduce: ContextVariable
			reduce(251), // %, reduce: ContextVariable
			reduce(251), // .*, reduce: ContextVariable
			reduce(251), // ./, reduce: ContextVariable
			reduce(251), // .//, reduce: ContextVariable
			reduce(251), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(251), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(251), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(251), // [[, reduce: ContextVariable
			reduce(251), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S465
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(252), // $, reduce: ContextVariable
			nil,         // empty
			reduce(252), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(252), // >, reduce: ContextVariable
			reduce(252), // >>, reduce: ContextVariable
			reduce(252), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(252), // ?, reduce: ContextVariable
			nil,         // :
			reduce(252), // ||, reduce: ContextVariable
			reduce(252), // ^^, reduce: ContextVariable
			reduce(252), // &&, reduce: ContextVariable
			reduce(252), // ??, reduce: ContextVariable
			reduce(252), // ???, reduce: ContextVariable
			reduce(252), // =~, reduce: ContextVariable
			reduce(252), // !=~, reduce: ContextVariable
			reduce(252), // ==, reduce: ContextVariable
			reduce(252), // !=, reduce: ContextVariable
			reduce(252), // <=>, reduce: ContextVariable
			reduce(252), // >=, reduce: ContextVariable
			reduce(252), // <, reduce: ContextVariable
			reduce(252), // <=, reduce: ContextVariable
			reduce(252), // ^, reduce: ContextVariable
			reduce(252), // &, reduce: ContextVariable
			reduce(252), // <<, reduce: ContextVariable
			reduce(252), // >>>, reduce: ContextVariable
			reduce(252), // +, reduce: ContextVariable
			reduce(252), // -, reduce: ContextVariable
			reduce(252), // .+, reduce: ContextVariable
			reduce(252), // .-, reduce: ContextVariable
			reduce(252), // *, reduce: ContextVariable
			reduce(252), // /, reduce: ContextVariable
			reduce(252), // //, reduce: ContextVariable
			reduce(252), // %, reduce: ContextVariable
			reduce(252), // .*, reduce: ContextVariable
			reduce(252), // ./, reduce: ContextVariable
			reduce(252), // .//, reduce: ContextVariable
			reduce(252), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(252), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(252), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(252), // [[, reduce: ContextVariable
			reduce(252), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S466
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(253), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(253), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(253), // >, reduce: MlrvalOrFunction
			reduce(253), // >>, reduce: MlrvalOrFunction
			reduce(253), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(253), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(253), // ||, reduce: MlrvalOrFunction
			reduce(253), // ^^, reduce: MlrvalOrFunction
			reduce(253), // &&, reduce: MlrvalOrFunction
			reduce(253), // ??, reduce: MlrvalOrFunction
			reduce(253), // ???, reduce: MlrvalOrFunction
			reduce(253), // =~, reduce: MlrvalOrFunction
			reduce(253), // !=~, reduce: MlrvalOrFunction
			reduce(253), // ==, reduce: MlrvalOrFunction
			reduce(253), // !=, reduce: MlrvalOrFunction
			reduce(253), // <=>, reduce: MlrvalOrFunction
			reduce(253), // >=, reduce: MlrvalOrFunction
			reduce(253), // <, reduce: MlrvalOrFunction
			reduce(253), // <=, reduce: MlrvalOrFunction
			reduce(253), // ^, reduce: MlrvalOrFunction
			reduce(253), // &, reduce: MlrvalOrFunction
			reduce(253), // <<, reduce: MlrvalOrFunction
			reduce(253), // >>>, reduce: MlrvalOrFunction
			reduce(253), // +, reduce: MlrvalOrFunction
			reduce(253), // -, reduce: MlrvalOrFunction
			reduce(253), // .+, reduce: MlrvalOrFunction
			reduce(253), // .-, reduce: MlrvalOrFunction
			reduce(253), // *, reduce: MlrvalOrFunction
			reduce(253), // /, reduce: MlrvalOrFunction
			reduce(253), // //, reduce: MlrvalOrFunction
			reduce(253), // %, reduce: MlrvalOrFunction
			reduce(253), // .*, reduce: MlrvalOrFunction
			reduce(253), // ./, reduce: MlrvalOrFunction
			reduce(253), // .//, reduce: MlrvalOrFunction
			reduce(253), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(253), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(253), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(253), // [[, reduce: MlrvalOrFunction
			reduce(253), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S467
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			shift(1212), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(1213), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S468
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(256), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(256), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(256), // >, reduce: MlrvalOrFunction
			reduce(256), // >>, reduce: MlrvalOrFunction
			reduce(256), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(256), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(256), // ||, reduce: MlrvalOrFunction
			reduce(256), // ^^, reduce: MlrvalOrFunction
			reduce(256), // &&, reduce: MlrvalOrFunction
			reduce(256), // ??, reduce: MlrvalOrFunction
			reduce(256), // ???, reduce: MlrvalOrFunction
			reduce(256), // =~, reduce: MlrvalOrFunction
			reduce(256), // !=~, reduce: MlrvalOrFunction
			reduce(256), // ==, reduce: MlrvalOrFunction
			reduce(256), // !=, reduce: MlrvalOrFunction
			reduce(256), // <=>, reduce: MlrvalOrFunction
			reduce(256), // >=, reduce: MlrvalOrFunction
			reduce(256), // <, reduce: MlrvalOrFunction
			reduce(256), // <=, reduce: MlrvalOrFunction
			reduce(256), // ^, reduce: MlrvalOrFunction
			reduce(256), // &, reduce: MlrvalOrFunction
			reduce(256), // <<, reduce: MlrvalOrFunction
			reduce(256), // >>>, reduce: MlrvalOrFunction
			reduce(256), // +, reduce: MlrvalOrFunction
			reduce(256), // -, reduce: MlrvalOrFunction
			reduce(256), // .+, reduce: MlrvalOrFunction
			reduce(256), // .-, reduce: MlrvalOrFunction
			reduce(256), // *, reduce: MlrvalOrFunction
			reduce(256), // /, reduce: MlrvalOrFunction
			reduce(256), // //, reduce: MlrvalOrFunction
			reduce(256), // %, reduce: MlrvalOrFunction
			reduce(256), // .*, reduce: MlrvalOrFunction
			reduce(256), // ./, reduce: MlrvalOrFunction
			reduce(256), // .//, reduce: MlrvalOrFunction
			reduce(256), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(256), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(256), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(256), // [[, reduce: MlrvalOrFunction
			reduce(256), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S469
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(257), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(257), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(257), // >, reduce: MlrvalOrFunction
			reduce(257), // >>, reduce: MlrvalOrFunction
			reduce(257), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(257), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(257), // ||, reduce: MlrvalOrFunction
			reduce(257), // ^^, reduce: MlrvalOrFunction
			reduce(257), // &&, reduce: MlrvalOrFunction
			reduce(257), // ??, reduce: MlrvalOrFunction
			reduce(257), // ???, reduce: MlrvalOrFunction
			reduce(257), // =~, reduce: MlrvalOrFunction
			reduce(257), // !=~, reduce: MlrvalOrFunction
			reduce(257), // ==, reduce: MlrvalOrFunction
			reduce(257), // !=, reduce: MlrvalOrFunction
			reduce(257), // <=>, reduce: MlrvalOrFunction
			reduce(257), // >=, reduce: MlrvalOrFunction
			reduce(257), // <, reduce: MlrvalOrFunction
			reduce(257), // <=, reduce: MlrvalOrFunction
			reduce(257), // ^, reduce: MlrvalOrFunction
			reduce(257), // &, reduce: MlrvalOrFunction
			reduce(257), // <<, reduce: MlrvalOrFunction
			reduce(257), // >>>, reduce: MlrvalOrFunction
			reduce(257), // +, reduce: MlrvalOrFunction
			reduce(257), // -, reduce: MlrvalOrFunction
			reduce(257), // .+, reduce: MlrvalOrFunction
			reduce(257), // .-, reduce: MlrvalOrFunction
			reduce(257), // *, reduce: MlrvalOrFunction
			reduce(257), // /, reduce: MlrvalOrFunction
			reduce(257), // //, reduce: MlrvalOrFunction
			reduce(257), // %, reduce: MlrvalOrFunction
			reduce(257), // .*, reduce: MlrvalOrFunction
			reduce(257), // ./, reduce: MlrvalOrFunction
			reduce(257), // .//, reduce: MlrvalOrFunction
			reduce(257), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(257), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(257), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(257), // [[, reduce: MlrvalOrFunction
			reduce(257), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S470
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(258), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(258), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(258), // >, reduce: MlrvalOrFunction
			reduce(258), // >>, reduce: MlrvalOrFunction
			reduce(258), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(258), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(258), // ||, reduce: MlrvalOrFunction
			reduce(258), // ^^, reduce: MlrvalOrFunction
			reduce(258), // &&, reduce: MlrvalOrFunction
			reduce(258), // ??, reduce: MlrvalOrFunction
			reduce(258), // ???, reduce: MlrvalOrFunction
			reduce(258), // =~, reduce: MlrvalOrFunction
			reduce(258), // !=~, reduce: MlrvalOrFunction
			reduce(258), // ==, reduce: MlrvalOrFunction
			reduce(258), // !=, reduce: MlrvalOrFunction
			reduce(258), // <=>, reduce: MlrvalOrFunction
			reduce(258), // >=, reduce: MlrvalOrFunction
			reduce(258), // <, reduce: MlrvalOrFunction
			reduce(258), // <=, reduce: MlrvalOrFunction
			reduce(258), // ^, reduce: MlrvalOrFunction
			reduce(258), // &, reduce: MlrvalOrFunction
			reduce(258), // <<, reduce: MlrvalOrFunction
			reduce(258), // >>>, reduce: MlrvalOrFunction
			reduce(258), // +, reduce: MlrvalOrFunction
			reduce(258), // -, reduce: MlrvalOrFunction
			reduce(258), // .+, reduce: MlrvalOrFunction
			reduce(258), // .-, reduce: MlrvalOrFunction
			reduce(258), // *, reduce: MlrvalOrFunction
			reduce(258), // /, reduce: MlrvalOrFunction
			reduce(258), // //, reduce: MlrvalOrFunction
			reduce(258), // %, reduce: MlrvalOrFunction
			reduce(258), // .*, reduce: MlrvalOrFunction
			reduce(258), // ./, reduce: MlrvalOrFunction
			reduce(258), // .//, reduce: MlrvalOrFunction
			reduce(258), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(258), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(258), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(258), // [[, reduce: MlrvalOrFunction
			reduce(258), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S471
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(259), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(259), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(259), // >, reduce: MlrvalOrFunction
			reduce(259), // >>, reduce: MlrvalOrFunction
			reduce(259), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(259), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(259), // ||, reduce: MlrvalOrFunction
			reduce(259), // ^^, reduce: MlrvalOrFunction
			reduce(259), // &&, reduce: MlrvalOrFunction
			reduce(259), // ??, reduce: MlrvalOrFunction
			reduce(259), // ???, reduce: MlrvalOrFunction
			reduce(259), // =~, reduce: MlrvalOrFunction
			reduce(259), // !=~, reduce: MlrvalOrFunction
			reduce(259), // ==, reduce: MlrvalOrFunction
			reduce(259), // !=, reduce: MlrvalOrFunction
			reduce(259), // <=>, reduce: MlrvalOrFunction
			reduce(259), // >=, reduce: MlrvalOrFunction
			reduce(259), // <, reduce: MlrvalOrFunction
			reduce(259), // <=, reduce: MlrvalOrFunction
			reduce(259), // ^, reduce: MlrvalOrFunction
			reduce(259), // &, reduce: MlrvalOrFunction
			reduce(259), // <<, reduce: MlrvalOrFunction
			reduce(259), // >>>, reduce: MlrvalOrFunction
			reduce(259), // +, reduce: MlrvalOrFunction
			reduce(259), // -, reduce: MlrvalOrFunction
			reduce(259), // .+, reduce: MlrvalOrFunction
			reduce(259), // .-, reduce: MlrvalOrFunction
			reduce(259), // *, reduce: MlrvalOrFunction
			reduce(259), // /, reduce: MlrvalOrFunction
			reduce(259), // //, reduce: MlrvalOrFunction
			reduce(259), // %, reduce: MlrvalOrFunction
			reduce(259), // .*, reduce: MlrvalOrFunction
			reduce(259), // ./, reduce: MlrvalOrFunction
			reduce(259), // .//, reduce: MlrvalOrFunction
			reduce(259), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(259), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(259), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(259), // [[, reduce: MlrvalOrFunction
			reduce(259), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S472
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(267), // $, reduce: MlrvalOrFunction
			nil,         // empty
			reduce(267), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(267), // >, reduce: MlrvalOrFunction
			reduce(267), // >>, reduce: MlrvalOrFunction
			reduce(267), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(267), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(267), // ||, reduce: MlrvalOrFunction
			reduce(267), // ^^, reduce: MlrvalOrFunction
			reduce(267), // &&, reduce: MlrvalOrFunction
			reduce(267), // ??, reduce: MlrvalOrFunction
			reduce(267), // ???, reduce: MlrvalOrFunction
			reduce(267), // =~, reduce: MlrvalOrFunction
			reduce(267), // !=~, reduce: MlrvalOrFunction
			reduce(267), // ==, reduce: MlrvalOrFunction
			reduce(267), // !=, reduce: MlrvalOrFunction
			reduce(267), // <=>, reduce: MlrvalOrFunction
			reduce(267), // >=, reduce: MlrvalOrFunction
			reduce(267), // <, reduce: MlrvalOrFunction
			reduce(267), // <=, reduce: MlrvalOrFunction
			reduce(267), // ^, reduce: MlrvalOrFunction
			reduce(267), // &, reduce: MlrvalOrFunction
			reduce(267), // <<, reduce: MlrvalOrFunction
			reduce(267), // >>>, reduce: MlrvalOrFunction
			reduce(267), // +, reduce: MlrvalOrFunction
			reduce(267), // -, reduce: MlrvalOrFunction
			reduce(267), // .+, reduce: MlrvalOrFunction
			reduce(267), // .-, reduce: MlrvalOrFunction
			reduce(267), // *, reduce: MlrvalOrFunction
			reduce(267), // /, reduce: MlrvalOrFunction
			reduce(267), // //, reduce: MlrvalOrFunction
			reduce(267), // %, reduce: MlrvalOrFunction
			reduce(267), // .*, reduce: MlrvalOrFunction
			reduce(267), // ./, reduce: MlrvalOrFunction
			reduce(267), // .//, reduce: MlrvalOrFunction
			reduce(267), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(267), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(267), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(267), // [[, reduce: MlrvalOrFunction
			reduce(267), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S473
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1214), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S474
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1215), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S475
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(48), // $, reduce: PrintStatement
			nil,        // empty
			reduce(48), // ;, reduce: PrintStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S476
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(47),  // $, reduce: PrintStatement
			nil,         // empty
			reduce(47),  // ;, reduce: PrintStatement
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(1216), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S477
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(290),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			shift(1219), // stdout
			shift(1220), // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(295),  // (
			nil,         // )
			nil,         // emitp
			shift(307),  // field_name
			shift(308),  // $[
			nil,         // ]
			shift(309),  // braced_field_name
			shift(310),  // $[[
			shift(311),  // $[[[
			shift(312),  // full_srec
			shift(314),  // oosvar_name
			shift(315),  // @[
			shift(316),  // braced_oosvar_name
			shift(317),  // full_oosvar
			shift(318),  // all
			shift(319),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(334),  // +
			shift(336),  // -
			shift(337),  // .+
			shift(338),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(341),  // !
			shift(342),  // ~
			nil,         // **
			shift(347),  // string_literal
			shift(348),  // regex_case_insensitive
			shift(349),  // int_literal
			shift(350),  // float_literal
			shift(351),  // boolean_literal
			shift(352),  // null_literal
			shift(353),  // inf_literal
			shift(354),  // nan_literal
			shift(355),  // const_M_PI
			shift(356),  // const_M_E
			shift(357),  // panic
			shift(359),  // [
			shift(361),  // ctx_IPS
			shift(362),  // ctx_IFS
			shift(363),  // ctx_IRS
			shift(364),  // ctx_OPS
			shift(365),  // ctx_OFS
			shift(366),  // ctx_ORS
			shift(367),  // ctx_FLATSEP
			shift(368),  // ctx_NF
			shift(369),  // ctx_NR
			shift(370),  // ctx_FNR
			shift(371),  // ctx_FILENAME
			shift(372),  // ctx_FILENUM
			shift(374),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(381),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S478
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(290),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			shift(1219), // stdout
			shift(1220), // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(295),  // (
			nil,         // )
			nil,         // emitp
			shift(307),  // field_name
			shift(308),  // $[
			nil,         // ]
			shift(309),  // braced_field_name
			shift(310),  // $[[
			shift(311),  // $[[[
			shift(312),  // full_srec
			shift(314),  // oosvar_name
			shift(315),  // @[
			shift(316),  // braced_oosvar_name
			shift(317),  // full_oosvar
			shift(318),  // all
			shift(319),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(334),  // +
			shift(336),  // -
			shift(337),  // .+
			shift(338),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(341),  // !
			shift(342),  // ~
			nil,         // **
			shift(347),  // string_literal
			shift(348),  // regex_case_insensitive
			shift(349),  // int_literal
			shift(350),  // float_literal
			shift(351),  // boolean_literal
			shift(352),  // null_literal
			shift(353),  // inf_literal
			shift(354),  // nan_literal
			shift(355),  // const_M_PI
			shift(356),  // const_M_E
			shift(357),  // panic
			shift(359),  // [
			shift(361),  // ctx_IPS
			shift(362),  // ctx_IFS
			shift(363),  // ctx_IRS
			shift(364),  // ctx_OPS
			shift(365),  // ctx_OFS
			shift(366),  // ctx_ORS
			shift(367),  // ctx_FLATSEP
			shift(368),  // ctx_NF
			shift(369),  // ctx_NR
			shift(370),  // ctx_FNR
			shift(371),  // ctx_FILENAME
			shift(372),  // ctx_FILENUM
			shift(374),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(381),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S479
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(290),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			shift(1219), // stdout
			shift(1220), // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(295),  // (
			nil,         // )
			nil,         // emitp
			shift(307),  // field_name
			shift(308),  // $[
			nil,         // ]
			shift(309),  // braced_field_name
			shift(310),  // $[[
			shift(311),  // $[[[
			shift(312),  // full_srec
			shift(314),  // oosvar_name
			shift(315),  // @[
			shift(316),  // braced_oosvar_name
			shift(317),  // full_oosvar
			shift(318),  // all
			shift(319),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(334),  // +
			shift(336),  // -
			shift(337),  // .+
			shift(338),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(341),  // !
			shift(342),  // ~
			nil,         // **
			shift(347),  // string_literal
			shift(348),  // regex_case_insensitive
			shift(349),  // int_literal
			shift(350),  // float_literal
			shift(351),  // boolean_literal
			shift(352),  // null_literal
			shift(353),  // inf_literal
			shift(354),  // nan_literal
			shift(355),  // const_M_PI
			shift(356),  // const_M_E
			shift(357),  // panic
			shift(359),  // [
			shift(361),  // ctx_IPS
			shift(362),  // ctx_IFS
			shift(363),  // ctx_IRS
			shift(364),  // ctx_OPS
			shift(365),  // ctx_OFS
			shift(366),  // ctx_ORS
			shift(367),  // ctx_FLATSEP
			shift(368),  // ctx_NF
			shift(369),  // ctx_NR
			shift(370),  // ctx_FNR
			shift(371),  // ctx_FILENAME
			shift(372),  // ctx_FILENUM
			shift(374),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(381),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S480
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(52), // $, reduce: PrintnStatement
			nil,        // empty
			reduce(52), // ;, reduce: PrintnStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S481
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(51),  // $, reduce: PrintnStatement
			nil,         // empty
			reduce(51),  // ;, reduce: PrintnStatement
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(1223), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S482
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(55), // $, reduce: EprintStatement
			nil,        // empty
			reduce(55), // ;, reduce: EprintStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S483
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(57), // $, reduce: EprintnStatement
			nil,        // empty
			reduce(57), // ;, reduce: EprintnStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S484
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(60), // $, reduce: DumpStatement
			nil,        // empty
			reduce(60), // ;, reduce: DumpStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S485
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(59),  // $, reduce: DumpStatement
			nil,         // empty
			reduce(59),  // ;, reduce: DumpStatement
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(1224), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S486
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(63), // $, reduce: EdumpStatement
			nil,        // empty
			reduce(63), // ;, reduce: EdumpStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S487
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(1225), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S488
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			shift(1230), // stdout
			shift(1231), // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S489
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			shift(1230), // stdout
			shift(1231), // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S490
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			shift(1230), // stdout
			shift(1231), // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S491
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			shift(1322), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S492
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(87), // $, reduce: Emittable
			nil,        // empty
			reduce(87), // ;, reduce: Emittable
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(87), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S493
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(1324), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S494
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(94), // $, reduce: Emittable
			nil,        // empty
			reduce(94), // ;, reduce: Emittable
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(94), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S495
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(65), // $, reduce: EmitFStatement
			nil,        // empty
			reduce(65), // ;, reduce: EmitFStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S496
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(84),  // $, reduce: EmittableList
			nil,         // empty
			reduce(84),  // ;, reduce: EmittableList
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(1325), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S497
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(88), // $, reduce: Emittable
			nil,        // empty
			reduce(88), // ;, reduce: Emittable
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(88), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S498
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(89), // $, reduce: Emittable
			nil,        // empty
			reduce(89), // ;, reduce: Emittable
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(89), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S499
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(90), // $, reduce: Emittable
			nil,        // empty
			reduce(90), // ;, reduce: Emittable
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(90), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S500
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(91), // $, reduce: Emittable
			nil,        // empty
			reduce(91), // ;, reduce: Emittable
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(91), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S501
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(92), // $, reduce: Emittable
			nil,        // empty
			reduce(92), // ;, reduce: Emittable
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(92), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S502
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(93), // $, reduce: Emittable
			nil,        // empty
			reduce(93), // ;, reduce: Emittable
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(93), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S503
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(95), // $, reduce: Emittable
			nil,        // empty
			reduce(95), // ;, reduce: Emittable
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(95), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S504
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(96), // $, reduce: Emittable
			nil,        // empty
			reduce(96), // ;, reduce: Emittable
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(96), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S505
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(104), // $, reduce: DirectFieldValue
			nil,         // empty
			reduce(104), // ;, reduce: DirectFieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(104), // ,, reduce: DirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S506
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S507
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(106), // $, reduce: BracedFieldValue
			nil,         // empty
			reduce(106), // ;, reduce: BracedFieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(106), // ,, reduce: BracedFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S508
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(109), // $, reduce: FullSrec
			nil,         // empty
			reduce(109), // ;, reduce: FullSrec
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(109), // ,, reduce: FullSrec
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S509
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(113), // $, reduce: DirectOosvarValue
			nil,         // empty
			reduce(113), // ;, reduce: DirectOosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(113), // ,, reduce: DirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S510
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S511
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(115), // $, reduce: BracedOosvarValue
			nil,         // empty
			reduce(115), // ;, reduce: BracedOosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(115), // ,, reduce: BracedOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S512
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(116), // $, reduce: FullOosvar
			nil,         // empty
			reduce(116), // ;, reduce: FullOosvar
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(116), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S513
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(117), // $, reduce: FullOosvar
			nil,         // empty
			reduce(117), // ;, reduce: FullOosvar
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(117), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S514
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(118), // $, reduce: LocalVariable
			nil,         // empty
			reduce(118), // ;, reduce: LocalVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(118), // ,, reduce: LocalVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S515
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(67), // $, reduce: Emit1Statement
			nil,        // empty
			reduce(67), // ;, reduce: Emit1Statement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S516
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(1328), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S517
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(68),  // $, reduce: EmitStatement
			nil,         // empty
			reduce(68),  // ;, reduce: EmitStatement
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(1329), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S518
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1330), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			shift(1343), // field_name
			shift(1344), // $[
			nil,         // ]
			shift(1345), // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(1346), // full_srec
			shift(1347), // oosvar_name
			shift(1348), // @[
			shift(1349), // braced_oosvar_name
			shift(1350), // full_oosvar
			shift(1351), // all
			shift(1352), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S519
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(86), // $, reduce: EmittableAsList
			nil,        // empty
			reduce(86), // ;, reduce: EmittableAsList
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(86), // ,, reduce: EmittableAsList
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S520
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			shift(1353), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S521
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(1355), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S522
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(214), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(214), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S523
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(211), // >, reduce: MlrvalOrFunction
			reduce(211), // >>, reduce: MlrvalOrFunction
			reduce(211), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(211), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(211), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(211), // ||, reduce: MlrvalOrFunction
			reduce(211), // ^^, reduce: MlrvalOrFunction
			reduce(211), // &&, reduce: MlrvalOrFunction
			reduce(211), // ??, reduce: MlrvalOrFunction
			reduce(211), // ???, reduce: MlrvalOrFunction
			reduce(211), // =~, reduce: MlrvalOrFunction
			reduce(211), // !=~, reduce: MlrvalOrFunction
			reduce(211), // ==, reduce: MlrvalOrFunction
			reduce(211), // !=, reduce: MlrvalOrFunction
			reduce(211), // <=>, reduce: MlrvalOrFunction
			reduce(211), // >=, reduce: MlrvalOrFunction
			reduce(211), // <, reduce: MlrvalOrFunction
			reduce(211), // <=, reduce: MlrvalOrFunction
			reduce(211), // ^, reduce: MlrvalOrFunction
			reduce(211), // &, reduce: MlrvalOrFunction
			reduce(211), // <<, reduce: MlrvalOrFunction
			reduce(211), // >>>, reduce: MlrvalOrFunction
			reduce(211), // +, reduce: MlrvalOrFunction
			reduce(211), // -, reduce: MlrvalOrFunction
			reduce(211), // .+, reduce: MlrvalOrFunction
			reduce(211), // .-, reduce: MlrvalOrFunction
			reduce(211), // *, reduce: MlrvalOrFunction
			reduce(211), // /, reduce: MlrvalOrFunction
			reduce(211), // //, reduce: MlrvalOrFunction
			reduce(211), // %, reduce: MlrvalOrFunction
			reduce(211), // .*, reduce: MlrvalOrFunction
			reduce(211), // ./, reduce: MlrvalOrFunction
			reduce(211), // .//, reduce: MlrvalOrFunction
			reduce(211), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(211), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(211), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(211), // [[, reduce: MlrvalOrFunction
			reduce(211), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S524
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S525
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(110), // >, reduce: OosvarValue
			reduce(110), // >>, reduce: OosvarValue
			reduce(110), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(110), // ), reduce: OosvarValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(110), // ?, reduce: OosvarValue
			nil,         // :
			reduce(110), // ||, reduce: OosvarValue
			reduce(110), // ^^, reduce: OosvarValue
			reduce(110), // &&, reduce: OosvarValue
			reduce(110), // ??, reduce: OosvarValue
			reduce(110), // ???, reduce: OosvarValue
			reduce(110), // =~, reduce: OosvarValue
			reduce(110), // !=~, reduce: OosvarValue
			reduce(110), // ==, reduce: OosvarValue
			reduce(110), // !=, reduce: OosvarValue
			reduce(110), // <=>, reduce: OosvarValue
			reduce(110), // >=, reduce: OosvarValue
			reduce(110), // <, reduce: OosvarValue
			reduce(110), // <=, reduce: OosvarValue
			reduce(110), // ^, reduce: OosvarValue
			reduce(110), // &, reduce: OosvarValue
			reduce(110), // <<, reduce: OosvarValue
			reduce(110), // >>>, reduce: OosvarValue
			reduce(110), // +, reduce: OosvarValue
			reduce(110), // -, reduce: OosvarValue
			reduce(110), // .+, reduce: OosvarValue
			reduce(110), // .-, reduce: OosvarValue
			reduce(110), // *, reduce: OosvarValue
			reduce(110), // /, reduce: OosvarValue
			reduce(110), // //, reduce: OosvarValue
			reduce(110), // %, reduce: OosvarValue
			reduce(110), // .*, reduce: OosvarValue
			reduce(110), // ./, reduce: OosvarValue
			reduce(110), // .//, reduce: OosvarValue
			reduce(110), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(110), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(110), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(110), // [[, reduce: OosvarValue
			reduce(110), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S526
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(112), // >, reduce: OosvarValue
			reduce(112), // >>, reduce: OosvarValue
			reduce(112), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(112), // ), reduce: OosvarValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(112), // ?, reduce: OosvarValue
			nil,         // :
			reduce(112), // ||, reduce: OosvarValue
			reduce(112), // ^^, reduce: OosvarValue
			reduce(112), // &&, reduce: OosvarValue
			reduce(112), // ??, reduce: OosvarValue
			reduce(112), // ???, reduce: OosvarValue
			reduce(112), // =~, reduce: OosvarValue
			reduce(112), // !=~, reduce: OosvarValue
			reduce(112), // ==, reduce: OosvarValue
			reduce(112), // !=, reduce: OosvarValue
			reduce(112), // <=>, reduce: OosvarValue
			reduce(112), // >=, reduce: OosvarValue
			reduce(112), // <, reduce: OosvarValue
			reduce(112), // <=, reduce: OosvarValue
			reduce(112), // ^, reduce: OosvarValue
			reduce(112), // &, reduce: OosvarValue
			reduce(112), // <<, reduce: OosvarValue
			reduce(112), // >>>, reduce: OosvarValue
			reduce(112), // +, reduce: OosvarValue
			reduce(112), // -, reduce: OosvarValue
			reduce(112), // .+, reduce: OosvarValue
			reduce(112), // .-, reduce: OosvarValue
			reduce(112), // *, reduce: OosvarValue
			reduce(112), // /, reduce: OosvarValue
			reduce(112), // //, reduce: OosvarValue
			reduce(112), // %, reduce: OosvarValue
			reduce(112), // .*, reduce: OosvarValue
			reduce(112), // ./, reduce: OosvarValue
			reduce(112), // .//, reduce: OosvarValue
			reduce(112), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(112), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(112), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(112), // [[, reduce: OosvarValue
			reduce(112), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S527
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(111), // >, reduce: OosvarValue
			reduce(111), // >>, reduce: OosvarValue
			reduce(111), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(111), // ), reduce: OosvarValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(111), // ?, reduce: OosvarValue
			nil,         // :
			reduce(111), // ||, reduce: OosvarValue
			reduce(111), // ^^, reduce: OosvarValue
			reduce(111), // &&, reduce: OosvarValue
			reduce(111), // ??, reduce: OosvarValue
			reduce(111), // ???, reduce: OosvarValue
			reduce(111), // =~, reduce: OosvarValue
			reduce(111), // !=~, reduce: OosvarValue
			reduce(111), // ==, reduce: OosvarValue
			reduce(111), // !=, reduce: OosvarValue
			reduce(111), // <=>, reduce: OosvarValue
			reduce(111), // >=, reduce: OosvarValue
			reduce(111), // <, reduce: OosvarValue
			reduce(111), // <=, reduce: OosvarValue
			reduce(111), // ^, reduce: OosvarValue
			reduce(111), // &, reduce: OosvarValue
			reduce(111), // <<, reduce: OosvarValue
			reduce(111), // >>>, reduce: OosvarValue
			reduce(111), // +, reduce: OosvarValue
			reduce(111), // -, reduce: OosvarValue
			reduce(111), // .+, reduce: OosvarValue
			reduce(111), // .-, reduce: OosvarValue
			reduce(111), // *, reduce: OosvarValue
			reduce(111), // /, reduce: OosvarValue
			reduce(111), // //, reduce: OosvarValue
			reduce(111), // %, reduce: OosvarValue
			reduce(111), // .*, reduce: OosvarValue
			reduce(111), // ./, reduce: OosvarValue
			reduce(111), // .//, reduce: OosvarValue
			reduce(111), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(111), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(111), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(111), // [[, reduce: OosvarValue
			reduce(111), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S528
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(99), // >, reduce: FieldValue
			reduce(99), // >>, reduce: FieldValue
			reduce(99), // |, reduce: FieldValue
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			reduce(99), // ), reduce: FieldValue
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(99), // ?, reduce: FieldValue
			nil,        // :
			reduce(99), // ||, reduce: FieldValue
			reduce(99), // ^^, reduce: FieldValue
			reduce(99), // &&, reduce: FieldValue
			reduce(99), // ??, reduce: FieldValue
			reduce(99), // ???, reduce: FieldValue
			reduce(99), // =~, reduce: FieldValue
			reduce(99), // !=~, reduce: FieldValue
			reduce(99), // ==, reduce: FieldValue
			reduce(99), // !=, reduce: FieldValue
			reduce(99), // <=>, reduce: FieldValue
			reduce(99), // >=, reduce: FieldValue
			reduce(99), // <, reduce: FieldValue
			reduce(99), // <=, reduce: FieldValue
			reduce(99), // ^, reduce: FieldValue
			reduce(99), // &, reduce: FieldValue
			reduce(99), // <<, reduce: FieldValue
			reduce(99), // >>>, reduce: FieldValue
			reduce(99), // +, reduce: FieldValue
			reduce(99), // -, reduce: FieldValue
			reduce(99), // .+, reduce: FieldValue
			reduce(99), // .-, reduce: FieldValue
			reduce(99), // *, reduce: FieldValue
			reduce(99), // /, reduce: FieldValue
			reduce(99), // //, reduce: FieldValue
			reduce(99), // %, reduce: FieldValue
			reduce(99), // .*, reduce: FieldValue
			reduce(99), // ./, reduce: FieldValue
			reduce(99), // .//, reduce: FieldValue
			reduce(99), // ., reduce: FieldValue
			nil,        // !
			nil,        // ~
			reduce(99), // **, reduce: FieldValue
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(99), // [, reduce: FieldValue
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(99), // [[, reduce: FieldValue
			reduce(99), // [[[, reduce: FieldValue
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S529
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(101), // >, reduce: FieldValue
			reduce(101), // >>, reduce: FieldValue
			reduce(101), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(101), // ), reduce: FieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(101), // ?, reduce: FieldValue
			nil,         // :
			reduce(101), // ||, reduce: FieldValue
			reduce(101), // ^^, reduce: FieldValue
			reduce(101), // &&, reduce: FieldValue
			reduce(101), // ??, reduce: FieldValue
			reduce(101), // ???, reduce: FieldValue
			reduce(101), // =~, reduce: FieldValue
			reduce(101), // !=~, reduce: FieldValue
			reduce(101), // ==, reduce: FieldValue
			reduce(101), // !=, reduce: FieldValue
			reduce(101), // <=>, reduce: FieldValue
			reduce(101), // >=, reduce: FieldValue
			reduce(101), // <, reduce: FieldValue
			reduce(101), // <=, reduce: FieldValue
			reduce(101), // ^, reduce: FieldValue
			reduce(101), // &, reduce: FieldValue
			reduce(101), // <<, reduce: FieldValue
			reduce(101), // >>>, reduce: FieldValue
			reduce(101), // +, reduce: FieldValue
			reduce(101), // -, reduce: FieldValue
			reduce(101), // .+, reduce: FieldValue
			reduce(101), // .-, reduce: FieldValue
			reduce(101), // *, reduce: FieldValue
			reduce(101), // /, reduce: FieldValue
			reduce(101), // //, reduce: FieldValue
			reduce(101), // %, reduce: FieldValue
			reduce(101), // .*, reduce: FieldValue
			reduce(101), // ./, reduce: FieldValue
			reduce(101), // .//, reduce: FieldValue
			reduce(101), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(101), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(101), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(101), // [[, reduce: FieldValue
			reduce(101), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S530
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(100), // >, reduce: FieldValue
			reduce(100), // >>, reduce: FieldValue
			reduce(100), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(100), // ), reduce: FieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(100), // ?, reduce: FieldValue
			nil,         // :
			reduce(100), // ||, reduce: FieldValue
			reduce(100), // ^^, reduce: FieldValue
			reduce(100), // &&, reduce: FieldValue
			reduce(100), // ??, reduce: FieldValue
			reduce(100), // ???, reduce: FieldValue
			reduce(100), // =~, reduce: FieldValue
			reduce(100), // !=~, reduce: FieldValue
			reduce(100), // ==, reduce: FieldValue
			reduce(100), // !=, reduce: FieldValue
			reduce(100), // <=>, reduce: FieldValue
			reduce(100), // >=, reduce: FieldValue
			reduce(100), // <, reduce: FieldValue
			reduce(100), // <=, reduce: FieldValue
			reduce(100), // ^, reduce: FieldValue
			reduce(100), // &, reduce: FieldValue
			reduce(100), // <<, reduce: FieldValue
			reduce(100), // >>>, reduce: FieldValue
			reduce(100), // +, reduce: FieldValue
			reduce(100), // -, reduce: FieldValue
			reduce(100), // .+, reduce: FieldValue
			reduce(100), // .-, reduce: FieldValue
			reduce(100), // *, reduce: FieldValue
			reduce(100), // /, reduce: FieldValue
			reduce(100), // //, reduce: FieldValue
			reduce(100), // %, reduce: FieldValue
			reduce(100), // .*, reduce: FieldValue
			reduce(100), // ./, reduce: FieldValue
			reduce(100), // .//, reduce: FieldValue
			reduce(100), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(100), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(100), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(100), // [[, reduce: FieldValue
			reduce(100), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S531
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(213), // >, reduce: MlrvalOrFunction
			reduce(213), // >>, reduce: MlrvalOrFunction
			reduce(213), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(213), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(213), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(213), // ||, reduce: MlrvalOrFunction
			reduce(213), // ^^, reduce: MlrvalOrFunction
			reduce(213), // &&, reduce: MlrvalOrFunction
			reduce(213), // ??, reduce: MlrvalOrFunction
			reduce(213), // ???, reduce: MlrvalOrFunction
			reduce(213), // =~, reduce: MlrvalOrFunction
			reduce(213), // !=~, reduce: MlrvalOrFunction
			reduce(213), // ==, reduce: MlrvalOrFunction
			reduce(213), // !=, reduce: MlrvalOrFunction
			reduce(213), // <=>, reduce: MlrvalOrFunction
			reduce(213), // >=, reduce: MlrvalOrFunction
			reduce(213), // <, reduce: MlrvalOrFunction
			reduce(213), // <=, reduce: MlrvalOrFunction
			reduce(213), // ^, reduce: MlrvalOrFunction
			reduce(213), // &, reduce: MlrvalOrFunction
			reduce(213), // <<, reduce: MlrvalOrFunction
			reduce(213), // >>>, reduce: MlrvalOrFunction
			reduce(213), // +, reduce: MlrvalOrFunction
			reduce(213), // -, reduce: MlrvalOrFunction
			reduce(213), // .+, reduce: MlrvalOrFunction
			reduce(213), // .-, reduce: MlrvalOrFunction
			reduce(213), // *, reduce: MlrvalOrFunction
			reduce(213), // /, reduce: MlrvalOrFunction
			reduce(213), // //, reduce: MlrvalOrFunction
			reduce(213), // %, reduce: MlrvalOrFunction
			reduce(213), // .*, reduce: MlrvalOrFunction
			reduce(213), // ./, reduce: MlrvalOrFunction
			reduce(213), // .//, reduce: MlrvalOrFunction
			reduce(213), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(213), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(213), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(213), // [[, reduce: MlrvalOrFunction
			reduce(213), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S532
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(233), // >, reduce: MlrvalOrFunction
			reduce(233), // >>, reduce: MlrvalOrFunction
			reduce(233), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(233), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(233), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(233), // ||, reduce: MlrvalOrFunction
			reduce(233), // ^^, reduce: MlrvalOrFunction
			reduce(233), // &&, reduce: MlrvalOrFunction
			reduce(233), // ??, reduce: MlrvalOrFunction
			reduce(233), // ???, reduce: MlrvalOrFunction
			reduce(233), // =~, reduce: MlrvalOrFunction
			reduce(233), // !=~, reduce: MlrvalOrFunction
			reduce(233), // ==, reduce: MlrvalOrFunction
			reduce(233), // !=, reduce: MlrvalOrFunction
			reduce(233), // <=>, reduce: MlrvalOrFunction
			reduce(233), // >=, reduce: MlrvalOrFunction
			reduce(233), // <, reduce: MlrvalOrFunction
			reduce(233), // <=, reduce: MlrvalOrFunction
			reduce(233), // ^, reduce: MlrvalOrFunction
			reduce(233), // &, reduce: MlrvalOrFunction
			reduce(233), // <<, reduce: MlrvalOrFunction
			reduce(233), // >>>, reduce: MlrvalOrFunction
			reduce(233), // +, reduce: MlrvalOrFunction
			reduce(233), // -, reduce: MlrvalOrFunction
			reduce(233), // .+, reduce: MlrvalOrFunction
			reduce(233), // .-, reduce: MlrvalOrFunction
			reduce(233), // *, reduce: MlrvalOrFunction
			reduce(233), // /, reduce: MlrvalOrFunction
			reduce(233), // //, reduce: MlrvalOrFunction
			reduce(233), // %, reduce: MlrvalOrFunction
			reduce(233), // .*, reduce: MlrvalOrFunction
			reduce(233), // ./, reduce: MlrvalOrFunction
			reduce(233), // .//, reduce: MlrvalOrFunction
			reduce(233), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(233), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(233), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(233), // [[, reduce: MlrvalOrFunction
			reduce(233), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S533
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(210), // >, reduce: MlrvalOrFunction
			reduce(210), // >>, reduce: MlrvalOrFunction
			reduce(210), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(210), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(210), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(210), // ||, reduce: MlrvalOrFunction
			reduce(210), // ^^, reduce: MlrvalOrFunction
			reduce(210), // &&, reduce: MlrvalOrFunction
			reduce(210), // ??, reduce: MlrvalOrFunction
			reduce(210), // ???, reduce: MlrvalOrFunction
			reduce(210), // =~, reduce: MlrvalOrFunction
			reduce(210), // !=~, reduce: MlrvalOrFunction
			reduce(210), // ==, reduce: MlrvalOrFunction
			reduce(210), // !=, reduce: MlrvalOrFunction
			reduce(210), // <=>, reduce: MlrvalOrFunction
			reduce(210), // >=, reduce: MlrvalOrFunction
			reduce(210), // <, reduce: MlrvalOrFunction
			reduce(210), // <=, reduce: MlrvalOrFunction
			reduce(210), // ^, reduce: MlrvalOrFunction
			reduce(210), // &, reduce: MlrvalOrFunction
			reduce(210), // <<, reduce: MlrvalOrFunction
			reduce(210), // >>>, reduce: MlrvalOrFunction
			reduce(210), // +, reduce: MlrvalOrFunction
			reduce(210), // -, reduce: MlrvalOrFunction
			reduce(210), // .+, reduce: MlrvalOrFunction
			reduce(210), // .-, reduce: MlrvalOrFunction
			reduce(210), // *, reduce: MlrvalOrFunction
			reduce(210), // /, reduce: MlrvalOrFunction
			reduce(210), // //, reduce: MlrvalOrFunction
			reduce(210), // %, reduce: MlrvalOrFunction
			reduce(210), // .*, reduce: MlrvalOrFunction
			reduce(210), // ./, reduce: MlrvalOrFunction
			reduce(210), // .//, reduce: MlrvalOrFunction
			reduce(210), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(210), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(210), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(210), // [[, reduce: MlrvalOrFunction
			reduce(210), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S534
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(102), // >, reduce: FieldValue
			reduce(102), // >>, reduce: FieldValue
			reduce(102), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(102), // ), reduce: FieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(102), // ?, reduce: FieldValue
			nil,         // :
			reduce(102), // ||, reduce: FieldValue
			reduce(102), // ^^, reduce: FieldValue
			reduce(102), // &&, reduce: FieldValue
			reduce(102), // ??, reduce: FieldValue
			reduce(102), // ???, reduce: FieldValue
			reduce(102), // =~, reduce: FieldValue
			reduce(102), // !=~, reduce: FieldValue
			reduce(102), // ==, reduce: FieldValue
			reduce(102), // !=, reduce: FieldValue
			reduce(102), // <=>, reduce: FieldValue
			reduce(102), // >=, reduce: FieldValue
			reduce(102), // <, reduce: FieldValue
			reduce(102), // <=, reduce: FieldValue
			reduce(102), // ^, reduce: FieldValue
			reduce(102), // &, reduce: FieldValue
			reduce(102), // <<, reduce: FieldValue
			reduce(102), // >>>, reduce: FieldValue
			reduce(102), // +, reduce: FieldValue
			reduce(102), // -, reduce: FieldValue
			reduce(102), // .+, reduce: FieldValue
			reduce(102), // .-, reduce: FieldValue
			reduce(102), // *, reduce: FieldValue
			reduce(102), // /, reduce: FieldValue
			reduce(102), // //, reduce: FieldValue
			reduce(102), // %, reduce: FieldValue
			reduce(102), // .*, reduce: FieldValue
			reduce(102), // ./, reduce: FieldValue
			reduce(102), // .//, reduce: FieldValue
			reduce(102), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(102), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(102), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(102), // [[, reduce: FieldValue
			reduce(102), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S535
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(103), // >, reduce: FieldValue
			reduce(103), // >>, reduce: FieldValue
			reduce(103), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(103), // ), reduce: FieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(103), // ?, reduce: FieldValue
			nil,         // :
			reduce(103), // ||, reduce: FieldValue
			reduce(103), // ^^, reduce: FieldValue
			reduce(103), // &&, reduce: FieldValue
			reduce(103), // ??, reduce: FieldValue
			reduce(103), // ???, reduce: FieldValue
			reduce(103), // =~, reduce: FieldValue
			reduce(103), // !=~, reduce: FieldValue
			reduce(103), // ==, reduce: FieldValue
			reduce(103), // !=, reduce: FieldValue
			reduce(103), // <=>, reduce: FieldValue
			reduce(103), // >=, reduce: FieldValue
			reduce(103), // <, reduce: FieldValue
			reduce(103), // <=, reduce: FieldValue
			reduce(103), // ^, reduce: FieldValue
			reduce(103), // &, reduce: FieldValue
			reduce(103), // <<, reduce: FieldValue
			reduce(103), // >>>, reduce: FieldValue
			reduce(103), // +, reduce: FieldValue
			reduce(103), // -, reduce: FieldValue
			reduce(103), // .+, reduce: FieldValue
			reduce(103), // .-, reduce: FieldValue
			reduce(103), // *, reduce: FieldValue
			reduce(103), // /, reduce: FieldValue
			reduce(103), // //, reduce: FieldValue
			reduce(103), // %, reduce: FieldValue
			reduce(103), // .*, reduce: FieldValue
			reduce(103), // ./, reduce: FieldValue
			reduce(103), // .//, reduce: FieldValue
			reduce(103), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(103), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(103), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(103), // [[, reduce: FieldValue
			reduce(103), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S536
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(104), // >, reduce: DirectFieldValue
			reduce(104), // >>, reduce: DirectFieldValue
			reduce(104), // |, reduce: DirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(104), // ), reduce: DirectFieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(104), // ?, reduce: DirectFieldValue
			nil,         // :
			reduce(104), // ||, reduce: DirectFieldValue
			reduce(104), // ^^, reduce: DirectFieldValue
			reduce(104), // &&, reduce: DirectFieldValue
			reduce(104), // ??, reduce: DirectFieldValue
			reduce(104), // ???, reduce: DirectFieldValue
			reduce(104), // =~, reduce: DirectFieldValue
			reduce(104), // !=~, reduce: DirectFieldValue
			reduce(104), // ==, reduce: DirectFieldValue
			reduce(104), // !=, reduce: DirectFieldValue
			reduce(104), // <=>, reduce: DirectFieldValue
			reduce(104), // >=, reduce: DirectFieldValue
			reduce(104), // <, reduce: DirectFieldValue
			reduce(104), // <=, reduce: DirectFieldValue
			reduce(104), // ^, reduce: DirectFieldValue
			reduce(104), // &, reduce: DirectFieldValue
			reduce(104), // <<, reduce: DirectFieldValue
			reduce(104), // >>>, reduce: DirectFieldValue
			reduce(104), // +, reduce: DirectFieldValue
			reduce(104), // -, reduce: DirectFieldValue
			reduce(104), // .+, reduce: DirectFieldValue
			reduce(104), // .-, reduce: DirectFieldValue
			reduce(104), // *, reduce: DirectFieldValue
			reduce(104), // /, reduce: DirectFieldValue
			reduce(104), // //, reduce: DirectFieldValue
			reduce(104), // %, reduce: DirectFieldValue
			reduce(104), // .*, reduce: DirectFieldValue
			reduce(104), // ./, reduce: DirectFieldValue
			reduce(104), // .//, reduce: DirectFieldValue
			reduce(104), // ., reduce: DirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(104), // **, reduce: DirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(104), // [, reduce: DirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(104), // [[, reduce: DirectFieldValue
			reduce(104), // [[[, reduce: DirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S537
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S538
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(106), // >, reduce: BracedFieldValue
			reduce(106), // >>, reduce: BracedFieldValue
			reduce(106), // |, reduce: BracedFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(106), // ), reduce: BracedFieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(106), // ?, reduce: BracedFieldValue
			nil,         // :
			reduce(106), // ||, reduce: BracedFieldValue
			reduce(106), // ^^, reduce: BracedFieldValue
			reduce(106), // &&, reduce: BracedFieldValue
			reduce(106), // ??, reduce: BracedFieldValue
			reduce(106), // ???, reduce: BracedFieldValue
			reduce(106), // =~, reduce: BracedFieldValue
			reduce(106), // !=~, reduce: BracedFieldValue
			reduce(106), // ==, reduce: BracedFieldValue
			reduce(106), // !=, reduce: BracedFieldValue
			reduce(106), // <=>, reduce: BracedFieldValue
			reduce(106), // >=, reduce: BracedFieldValue
			reduce(106), // <, reduce: BracedFieldValue
			reduce(106), // <=, reduce: BracedFieldValue
			reduce(106), // ^, reduce: BracedFieldValue
			reduce(106), // &, reduce: BracedFieldValue
			reduce(106), // <<, reduce: BracedFieldValue
			reduce(106), // >>>, reduce: BracedFieldValue
			reduce(106), // +, reduce: BracedFieldValue
			reduce(106), // -, reduce: BracedFieldValue
			reduce(106), // .+, reduce: BracedFieldValue
			reduce(106), // .-, reduce: BracedFieldValue
			reduce(106), // *, reduce: BracedFieldValue
			reduce(106), // /, reduce: BracedFieldValue
			reduce(106), // //, reduce: BracedFieldValue
			reduce(106), // %, reduce: BracedFieldValue
			reduce(106), // .*, reduce: BracedFieldValue
			reduce(106), // ./, reduce: BracedFieldValue
			reduce(106), // .//, reduce: BracedFieldValue
			reduce(106), // ., reduce: BracedFieldValue
			nil,         // !
			nil,         // ~
			reduce(106), // **, reduce: BracedFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(106), // [, reduce: BracedFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(106), // [[, reduce: BracedFieldValue
			reduce(106), // [[[, reduce: BracedFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S539
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S540
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S541
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(109), // >, reduce: FullSrec
			reduce(109), // >>, reduce: FullSrec
			reduce(109), // |, reduce: FullSrec
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(109), // ), reduce: FullSrec
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(109), // ?, reduce: FullSrec
			nil,         // :
			reduce(109), // ||, reduce: FullSrec
			reduce(109), // ^^, reduce: FullSrec
			reduce(109), // &&, reduce: FullSrec
			reduce(109), // ??, reduce: FullSrec
			reduce(109), // ???, reduce: FullSrec
			reduce(109), // =~, reduce: FullSrec
			reduce(109), // !=~, reduce: FullSrec
			reduce(109), // ==, reduce: FullSrec
			reduce(109), // !=, reduce: FullSrec
			reduce(109), // <=>, reduce: FullSrec
			reduce(109), // >=, reduce: FullSrec
			reduce(109), // <, reduce: FullSrec
			reduce(109), // <=, reduce: FullSrec
			reduce(109), // ^, reduce: FullSrec
			reduce(109), // &, reduce: FullSrec
			reduce(109), // <<, reduce: FullSrec
			reduce(109), // >>>, reduce: FullSrec
			reduce(109), // +, reduce: FullSrec
			reduce(109), // -, reduce: FullSrec
			reduce(109), // .+, reduce: FullSrec
			reduce(109), // .-, reduce: FullSrec
			reduce(109), // *, reduce: FullSrec
			reduce(109), // /, reduce: FullSrec
			reduce(109), // //, reduce: FullSrec
			reduce(109), // %, reduce: FullSrec
			reduce(109), // .*, reduce: FullSrec
			reduce(109), // ./, reduce: FullSrec
			reduce(109), // .//, reduce: FullSrec
			reduce(109), // ., reduce: FullSrec
			nil,         // !
			nil,         // ~
			reduce(109), // **, reduce: FullSrec
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(109), // [, reduce: FullSrec
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(109), // [[, reduce: FullSrec
			reduce(109), // [[[, reduce: FullSrec
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S542
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(212), // >, reduce: MlrvalOrFunction
			reduce(212), // >>, reduce: MlrvalOrFunction
			reduce(212), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(212), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(212), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(212), // ||, reduce: MlrvalOrFunction
			reduce(212), // ^^, reduce: MlrvalOrFunction
			reduce(212), // &&, reduce: MlrvalOrFunction
			reduce(212), // ??, reduce: MlrvalOrFunction
			reduce(212), // ???, reduce: MlrvalOrFunction
			reduce(212), // =~, reduce: MlrvalOrFunction
			reduce(212), // !=~, reduce: MlrvalOrFunction
			reduce(212), // ==, reduce: MlrvalOrFunction
			reduce(212), // !=, reduce: MlrvalOrFunction
			reduce(212), // <=>, reduce: MlrvalOrFunction
			reduce(212), // >=, reduce: MlrvalOrFunction
			reduce(212), // <, reduce: MlrvalOrFunction
			reduce(212), // <=, reduce: MlrvalOrFunction
			reduce(212), // ^, reduce: MlrvalOrFunction
			reduce(212), // &, reduce: MlrvalOrFunction
			reduce(212), // <<, reduce: MlrvalOrFunction
			reduce(212), // >>>, reduce: MlrvalOrFunction
			reduce(212), // +, reduce: MlrvalOrFunction
			reduce(212), // -, reduce: MlrvalOrFunction
			reduce(212), // .+, reduce: MlrvalOrFunction
			reduce(212), // .-, reduce: MlrvalOrFunction
			reduce(212), // *, reduce: MlrvalOrFunction
			reduce(212), // /, reduce: MlrvalOrFunction
			reduce(212), // //, reduce: MlrvalOrFunction
			reduce(212), // %, reduce: MlrvalOrFunction
			reduce(212), // .*, reduce: MlrvalOrFunction
			reduce(212), // ./, reduce: MlrvalOrFunction
			reduce(212), // .//, reduce: MlrvalOrFunction
			reduce(212), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(212), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(212), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(212), // [[, reduce: MlrvalOrFunction
			reduce(212), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S543
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(113), // >, reduce: DirectOosvarValue
			reduce(113), // >>, reduce: DirectOosvarValue
			reduce(113), // |, reduce: DirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(113), // ), reduce: DirectOosvarValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(113), // ?, reduce: DirectOosvarValue
			nil,         // :
			reduce(113), // ||, reduce: DirectOosvarValue
			reduce(113), // ^^, reduce: DirectOosvarValue
			reduce(113), // &&, reduce: DirectOosvarValue
			reduce(113), // ??, reduce: DirectOosvarValue
			reduce(113), // ???, reduce: DirectOosvarValue
			reduce(113), // =~, reduce: DirectOosvarValue
			reduce(113), // !=~, reduce: DirectOosvarValue
			reduce(113), // ==, reduce: DirectOosvarValue
			reduce(113), // !=, reduce: DirectOosvarValue
			reduce(113), // <=>, reduce: DirectOosvarValue
			reduce(113), // >=, reduce: DirectOosvarValue
			reduce(113), // <, reduce: DirectOosvarValue
			reduce(113), // <=, reduce: DirectOosvarValue
			reduce(113), // ^, reduce: DirectOosvarValue
			reduce(113), // &, reduce: DirectOosvarValue
			reduce(113), // <<, reduce: DirectOosvarValue
			reduce(113), // >>>, reduce: DirectOosvarValue
			reduce(113), // +, reduce: DirectOosvarValue
			reduce(113), // -, reduce: DirectOosvarValue
			reduce(113), // .+, reduce: DirectOosvarValue
			reduce(113), // .-, reduce: DirectOosvarValue
			reduce(113), // *, reduce: DirectOosvarValue
			reduce(113), // /, reduce: DirectOosvarValue
			reduce(113), // //, reduce: DirectOosvarValue
			reduce(113), // %, reduce: DirectOosvarValue
			reduce(113), // .*, reduce: DirectOosvarValue
			reduce(113), // ./, reduce: DirectOosvarValue
			reduce(113), // .//, reduce: DirectOosvarValue
			reduce(113), // ., reduce: DirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(113), // **, reduce: DirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(113), // [, reduce: DirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(113), // [[, reduce: DirectOosvarValue
			reduce(113), // [[[, reduce: DirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S544
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S545
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(115), // >, reduce: BracedOosvarValue
			reduce(115), // >>, reduce: BracedOosvarValue
			reduce(115), // |, reduce: BracedOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(115), // ), reduce: BracedOosvarValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(115), // ?, reduce: BracedOosvarValue
			nil,         // :
			reduce(115), // ||, reduce: BracedOosvarValue
			reduce(115), // ^^, reduce: BracedOosvarValue
			reduce(115), // &&, reduce: BracedOosvarValue
			reduce(115), // ??, reduce: BracedOosvarValue
			reduce(115), // ???, reduce: BracedOosvarValue
			reduce(115), // =~, reduce: BracedOosvarValue
			reduce(115), // !=~, reduce: BracedOosvarValue
			reduce(115), // ==, reduce: BracedOosvarValue
			reduce(115), // !=, reduce: BracedOosvarValue
			reduce(115), // <=>, reduce: BracedOosvarValue
			reduce(115), // >=, reduce: BracedOosvarValue
			reduce(115), // <, reduce: BracedOosvarValue
			reduce(115), // <=, reduce: BracedOosvarValue
			reduce(115), // ^, reduce: BracedOosvarValue
			reduce(115), // &, reduce: BracedOosvarValue
			reduce(115), // <<, reduce: BracedOosvarValue
			reduce(115), // >>>, reduce: BracedOosvarValue
			reduce(115), // +, reduce: BracedOosvarValue
			reduce(115), // -, reduce: BracedOosvarValue
			reduce(115), // .+, reduce: BracedOosvarValue
			reduce(115), // .-, reduce: BracedOosvarValue
			reduce(115), // *, reduce: BracedOosvarValue
			reduce(115), // /, reduce: BracedOosvarValue
			reduce(115), // //, reduce: BracedOosvarValue
			reduce(115), // %, reduce: BracedOosvarValue
			reduce(115), // .*, reduce: BracedOosvarValue
			reduce(115), // ./, reduce: BracedOosvarValue
			reduce(115), // .//, reduce: BracedOosvarValue
			reduce(115), // ., reduce: BracedOosvarValue
			nil,         // !
			nil,         // ~
			reduce(115), // **, reduce: BracedOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(115), // [, reduce: BracedOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(115), // [[, reduce: BracedOosvarValue
			reduce(115), // [[[, reduce: BracedOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S546
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(116), // >, reduce: FullOosvar
			reduce(116), // >>, reduce: FullOosvar
			reduce(116), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(116), // ), reduce: FullOosvar
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(116), // ?, reduce: FullOosvar
			nil,         // :
			reduce(116), // ||, reduce: FullOosvar
			reduce(116), // ^^, reduce: FullOosvar
			reduce(116), // &&, reduce: FullOosvar
			reduce(116), // ??, reduce: FullOosvar
			reduce(116), // ???, reduce: FullOosvar
			reduce(116), // =~, reduce: FullOosvar
			reduce(116), // !=~, reduce: FullOosvar
			reduce(116), // ==, reduce: FullOosvar
			reduce(116), // !=, reduce: FullOosvar
			reduce(116), // <=>, reduce: FullOosvar
			reduce(116), // >=, reduce: FullOosvar
			reduce(116), // <, reduce: FullOosvar
			reduce(116), // <=, reduce: FullOosvar
			reduce(116), // ^, reduce: FullOosvar
			reduce(116), // &, reduce: FullOosvar
			reduce(116), // <<, reduce: FullOosvar
			reduce(116), // >>>, reduce: FullOosvar
			reduce(116), // +, reduce: FullOosvar
			reduce(116), // -, reduce: FullOosvar
			reduce(116), // .+, reduce: FullOosvar
			reduce(116), // .-, reduce: FullOosvar
			reduce(116), // *, reduce: FullOosvar
			reduce(116), // /, reduce: FullOosvar
			reduce(116), // //, reduce: FullOosvar
			reduce(116), // %, reduce: FullOosvar
			reduce(116), // .*, reduce: FullOosvar
			reduce(116), // ./, reduce: FullOosvar
			reduce(116), // .//, reduce: FullOosvar
			reduce(116), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(116), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(116), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(116), // [[, reduce: FullOosvar
			reduce(116), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S547
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(117), // >, reduce: FullOosvar
			reduce(117), // >>, reduce: FullOosvar
			reduce(117), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(117), // ), reduce: FullOosvar
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(117), // ?, reduce: FullOosvar
			nil,         // :
			reduce(117), // ||, reduce: FullOosvar
			reduce(117), // ^^, reduce: FullOosvar
			reduce(117), // &&, reduce: FullOosvar
			reduce(117), // ??, reduce: FullOosvar
			reduce(117), // ???, reduce: FullOosvar
			reduce(117), // =~, reduce: FullOosvar
			reduce(117), // !=~, reduce: FullOosvar
			reduce(117), // ==, reduce: FullOosvar
			reduce(117), // !=, reduce: FullOosvar
			reduce(117), // <=>, reduce: FullOosvar
			reduce(117), // >=, reduce: FullOosvar
			reduce(117), // <, reduce: FullOosvar
			reduce(117), // <=, reduce: FullOosvar
			reduce(117), // ^, reduce: FullOosvar
			reduce(117), // &, reduce: FullOosvar
			reduce(117), // <<, reduce: FullOosvar
			reduce(117), // >>>, reduce: FullOosvar
			reduce(117), // +, reduce: FullOosvar
			reduce(117), // -, reduce: FullOosvar
			reduce(117), // .+, reduce: FullOosvar
			reduce(117), // .-, reduce: FullOosvar
			reduce(117), // *, reduce: FullOosvar
			reduce(117), // /, reduce: FullOosvar
			reduce(117), // //, reduce: FullOosvar
			reduce(117), // %, reduce: FullOosvar
			reduce(117), // .*, reduce: FullOosvar
			reduce(117), // ./, reduce: FullOosvar
			reduce(117), // .//, reduce: FullOosvar
			reduce(117), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(117), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(117), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(117), // [[, reduce: FullOosvar
			reduce(117), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S548
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(118), // >, reduce: LocalVariable
			reduce(118), // >>, reduce: LocalVariable
			reduce(118), // |, reduce: LocalVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(270), // (, reduce: FunctionName
			reduce(118), // ), reduce: LocalVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(118), // ?, reduce: LocalVariable
			nil,         // :
			reduce(118), // ||, reduce: LocalVariable
			reduce(118), // ^^, reduce: LocalVariable
			reduce(118), // &&, reduce: LocalVariable
			reduce(118), // ??, reduce: LocalVariable
			reduce(118), // ???, reduce: LocalVariable
			reduce(118), // =~, reduce: LocalVariable
			reduce(118), // !=~, reduce: LocalVariable
			reduce(118), // ==, reduce: LocalVariable
			reduce(118), // !=, reduce: LocalVariable
			reduce(118), // <=>, reduce: LocalVariable
			reduce(118), // >=, reduce: LocalVariable
			reduce(118), // <, reduce: LocalVariable
			reduce(118), // <=, reduce: LocalVariable
			reduce(118), // ^, reduce: LocalVariable
			reduce(118), // &, reduce: LocalVariable
			reduce(118), // <<, reduce: LocalVariable
			reduce(118), // >>>, reduce: LocalVariable
			reduce(118), // +, reduce: LocalVariable
			reduce(118), // -, reduce: LocalVariable
			reduce(118), // .+, reduce: LocalVariable
			reduce(118), // .-, reduce: LocalVariable
			reduce(118), // *, reduce: LocalVariable
			reduce(118), // /, reduce: LocalVariable
			reduce(118), // //, reduce: LocalVariable
			reduce(118), // %, reduce: LocalVariable
			reduce(118), // .*, reduce: LocalVariable
			reduce(118), // ./, reduce: LocalVariable
			reduce(118), // .//, reduce: LocalVariable
			reduce(118), // ., reduce: LocalVariable
			nil,         // !
			nil,         // ~
			reduce(118), // **, reduce: LocalVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(118), // [, reduce: LocalVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(118), // [[, reduce: LocalVariable
			reduce(118), // [[[, reduce: LocalVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S549
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(147), // ), reduce: Rvalue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S550
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(148), // ), reduce: PrecedenceChainStart
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S551
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(150), // ), reduce: TernaryTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			shift(1361), // ?
			nil,         // :
			shift(1362), // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S552
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(152), // ), reduce: LogicalOrTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(152), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(152), // ||, reduce: LogicalOrTerm
			shift(1363), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S553
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(154), // ), reduce: LogicalXORTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(154), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(154), // ||, reduce: LogicalXORTerm
			reduce(154), // ^^, reduce: LogicalXORTerm
			shift(1364), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S554
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(156), // ), reduce: LogicalAndTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(156), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(156), // ||, reduce: LogicalAndTerm
			reduce(156), // ^^, reduce: LogicalAndTerm
			reduce(156), // &&, reduce: LogicalAndTerm
			shift(1365), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S555
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(158), // ), reduce: AbsentCoalesceTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(158), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(158), // ||, reduce: AbsentCoalesceTerm
			reduce(158), // ^^, reduce: AbsentCoalesceTerm
			reduce(158), // &&, reduce: AbsentCoalesceTerm
			reduce(158), // ??, reduce: AbsentCoalesceTerm
			shift(1366), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S556
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(160), // ), reduce: EmptyCoalesceTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(160), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(160), // ||, reduce: EmptyCoalesceTerm
			reduce(160), // ^^, reduce: EmptyCoalesceTerm
			reduce(160), // &&, reduce: EmptyCoalesceTerm
			reduce(160), // ??, reduce: EmptyCoalesceTerm
			reduce(160), // ???, reduce: EmptyCoalesceTerm
			shift(1367), // =~
			shift(1368), // !=~
			shift(1369), // ==
			shift(1370), // !=
			shift(1371), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S557
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1372), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(166), // ), reduce: EqneTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(166), // ?, reduce: EqneTerm
			nil,         // :
			reduce(166), // ||, reduce: EqneTerm
			reduce(166), // ^^, reduce: EqneTerm
			reduce(166), // &&, reduce: EqneTerm
			reduce(166), // ??, reduce: EqneTerm
			reduce(166), // ???, reduce: EqneTerm
			reduce(166), // =~, reduce: EqneTerm
			reduce(166), // !=~, reduce: EqneTerm
			reduce(166), // ==, reduce: EqneTerm
			reduce(166), // !=, reduce: EqneTerm
			reduce(166), // <=>, reduce: EqneTerm
			shift(1373), // >=
			shift(1374), // <
			shift(1375), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S558
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(171), // >, reduce: CmpTerm
			nil,         // >>
			shift(1376), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(171), // ), reduce: CmpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(171), // ?, reduce: CmpTerm
			nil,         // :
			reduce(171), // ||, reduce: CmpTerm
			reduce(171), // ^^, reduce: CmpTerm
			reduce(171), // &&, reduce: CmpTerm
			reduce(171), // ??, reduce: CmpTerm
			reduce(171), // ???, reduce: CmpTerm
			reduce(171), // =~, reduce: CmpTerm
			reduce(171), // !=~, reduce: CmpTerm
			reduce(171), // ==, reduce: CmpTerm
			reduce(171), // !=, reduce: CmpTerm
			reduce(171), // <=>, reduce: CmpTerm
			reduce(171), // >=, reduce: CmpTerm
			reduce(171), // <, reduce: CmpTerm
			reduce(171), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S559
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(173), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(173), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(173), // ), reduce: BitwiseORTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(173), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(173), // ||, reduce: BitwiseORTerm
			reduce(173), // ^^, reduce: BitwiseORTerm
			reduce(173), // &&, reduce: BitwiseORTerm
			reduce(173), // ??, reduce: BitwiseORTerm
			reduce(173), // ???, reduce: BitwiseORTerm
			reduce(173), // =~, reduce: BitwiseORTerm
			reduce(173), // !=~, reduce: BitwiseORTerm
			reduce(173), // ==, reduce: BitwiseORTerm
			reduce(173), // !=, reduce: BitwiseORTerm
			reduce(173), // <=>, reduce: BitwiseORTerm
			reduce(173), // >=, reduce: BitwiseORTerm
			reduce(173), // <, reduce: BitwiseORTerm
			reduce(173), // <=, reduce: BitwiseORTerm
			shift(1377), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S560
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(175), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(175), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(175), // ), reduce: BitwiseXORTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(175), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(175), // ||, reduce: BitwiseXORTerm
			reduce(175), // ^^, reduce: BitwiseXORTerm
			reduce(175), // &&, reduce: BitwiseXORTerm
			reduce(175), // ??, reduce: BitwiseXORTerm
			reduce(175), // ???, reduce: BitwiseXORTerm
			reduce(175), // =~, reduce: BitwiseXORTerm
			reduce(175), // !=~, reduce: BitwiseXORTerm
			reduce(175), // ==, reduce: BitwiseXORTerm
			reduce(175), // !=, reduce: BitwiseXORTerm
			reduce(175), // <=>, reduce: BitwiseXORTerm
			reduce(175), // >=, reduce: BitwiseXORTerm
			reduce(175), // <, reduce: BitwiseXORTerm
			reduce(175), // <=, reduce: BitwiseXORTerm
			reduce(175), // ^, reduce: BitwiseXORTerm
			shift(1378), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S561
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(177), // >, reduce: BitwiseANDTerm
			shift(1379), // >>
			reduce(177), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(177), // ), reduce: BitwiseANDTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(177), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(177), // ||, reduce: BitwiseANDTerm
			reduce(177), // ^^, reduce: BitwiseANDTerm
			reduce(177), // &&, reduce: BitwiseANDTerm
			reduce(177), // ??, reduce: BitwiseANDTerm
			reduce(177), // ???, reduce: BitwiseANDTerm
			reduce(177), // =~, reduce: BitwiseANDTerm
			reduce(177), // !=~, reduce: BitwiseANDTerm
			reduce(177), // ==, reduce: BitwiseANDTerm
			reduce(177), // !=, reduce: BitwiseANDTerm
			reduce(177), // <=>, reduce: BitwiseANDTerm
			reduce(177), // >=, reduce: BitwiseANDTerm
			reduce(177), // <, reduce: BitwiseANDTerm
			reduce(177), // <=, reduce: BitwiseANDTerm
			reduce(177), // ^, reduce: BitwiseANDTerm
			reduce(177), // &, reduce: BitwiseANDTerm
			shift(1380), // <<
			shift(1381), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S562
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(181), // >, reduce: BitwiseShiftTerm
			reduce(181), // >>, reduce: BitwiseShiftTerm
			reduce(181), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(181), // ), reduce: BitwiseShiftTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(181), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(181), // ||, reduce: BitwiseShiftTerm
			reduce(181), // ^^, reduce: BitwiseShiftTerm
			reduce(181), // &&, reduce: BitwiseShiftTerm
			reduce(181), // ??, reduce: BitwiseShiftTerm
			reduce(181), // ???, reduce: BitwiseShiftTerm
			reduce(181), // =~, reduce: BitwiseShiftTerm
			reduce(181), // !=~, reduce: BitwiseShiftTerm
			reduce(181), // ==, reduce: BitwiseShiftTerm
			reduce(181), // !=, reduce: BitwiseShiftTerm
			reduce(181), // <=>, reduce: BitwiseShiftTerm
			reduce(181), // >=, reduce: BitwiseShiftTerm
			reduce(181), // <, reduce: BitwiseShiftTerm
			reduce(181), // <=, reduce: BitwiseShiftTerm
			reduce(181), // ^, reduce: BitwiseShiftTerm
			reduce(181), // &, reduce: BitwiseShiftTerm
			reduce(181), // <<, reduce: BitwiseShiftTerm
			reduce(181), // >>>, reduce: BitwiseShiftTerm
			shift(1382), // +
			shift(1383), // -
			shift(1384), // .+
			shift(1385), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S563
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S564
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(186), // >, reduce: AddsubdotTerm
			reduce(186), // >>, reduce: AddsubdotTerm
			reduce(186), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(186), // ), reduce: AddsubdotTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(186), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(186), // ||, reduce: AddsubdotTerm
			reduce(186), // ^^, reduce: AddsubdotTerm
			reduce(186), // &&, reduce: AddsubdotTerm
			reduce(186), // ??, reduce: AddsubdotTerm
			reduce(186), // ???, reduce: AddsubdotTerm
			reduce(186), // =~, reduce: AddsubdotTerm
			reduce(186), // !=~, reduce: AddsubdotTerm
			reduce(186), // ==, reduce: AddsubdotTerm
			reduce(186), // !=, reduce: AddsubdotTerm
			reduce(186), // <=>, reduce: AddsubdotTerm
			reduce(186), // >=, reduce: AddsubdotTerm
			reduce(186), // <, reduce: AddsubdotTerm
			reduce(186), // <=, reduce: AddsubdotTerm
			reduce(186), // ^, reduce: AddsubdotTerm
			reduce(186), // &, reduce: AddsubdotTerm
			reduce(186), // <<, reduce: AddsubdotTerm
			reduce(186), // >>>, reduce: AddsubdotTerm
			reduce(186), // +, reduce: AddsubdotTerm
			reduce(186), // -, reduce: AddsubdotTerm
			reduce(186), // .+, reduce: AddsubdotTerm
			reduce(186), // .-, reduce: AddsubdotTerm
			shift(1387), // *
			shift(1388), // /
			shift(1389), // //
			shift(1390), // %
			shift(1391), // .*
			shift(1392), // ./
			shift(1393), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S565
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S566
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S567
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S568
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(194), // >, reduce: MuldivTerm
			reduce(194), // >>, reduce: MuldivTerm
			reduce(194), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(194), // ), reduce: MuldivTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(194), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(194), // ||, reduce: MuldivTerm
			reduce(194), // ^^, reduce: MuldivTerm
			reduce(194), // &&, reduce: MuldivTerm
			reduce(194), // ??, reduce: MuldivTerm
			reduce(194), // ???, reduce: MuldivTerm
			reduce(194), // =~, reduce: MuldivTerm
			reduce(194), // !=~, reduce: MuldivTerm
			reduce(194), // ==, reduce: MuldivTerm
			reduce(194), // !=, reduce: MuldivTerm
			reduce(194), // <=>, reduce: MuldivTerm
			reduce(194), // >=, reduce: MuldivTerm
			reduce(194), // <, reduce: MuldivTerm
			reduce(194), // <=, reduce: MuldivTerm
			reduce(194), // ^, reduce: MuldivTerm
			reduce(194), // &, reduce: MuldivTerm
			reduce(194), // <<, reduce: MuldivTerm
			reduce(194), // >>>, reduce: MuldivTerm
			reduce(194), // +, reduce: MuldivTerm
			reduce(194), // -, reduce: MuldivTerm
			reduce(194), // .+, reduce: MuldivTerm
			reduce(194), // .-, reduce: MuldivTerm
			reduce(194), // *, reduce: MuldivTerm
			reduce(194), // /, reduce: MuldivTerm
			reduce(194), // //, reduce: MuldivTerm
			reduce(194), // %, reduce: MuldivTerm
			reduce(194), // .*, reduce: MuldivTerm
			reduce(194), // ./, reduce: MuldivTerm
			reduce(194), // .//, reduce: MuldivTerm
			shift(1397), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S569
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(196), // >, reduce: DotTerm
			reduce(196), // >>, reduce: DotTerm
			reduce(196), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(196), // ), reduce: DotTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(196), // ?, reduce: DotTerm
			nil,         // :
			reduce(196), // ||, reduce: DotTerm
			reduce(196), // ^^, reduce: DotTerm
			reduce(196), // &&, reduce: DotTerm
			reduce(196), // ??, reduce: DotTerm
			reduce(196), // ???, reduce: DotTerm
			reduce(196), // =~, reduce: DotTerm
			reduce(196), // !=~, reduce: DotTerm
			reduce(196), // ==, reduce: DotTerm
			reduce(196), // !=, reduce: DotTerm
			reduce(196), // <=>, reduce: DotTerm
			reduce(196), // >=, reduce: DotTerm
			reduce(196), // <, reduce: DotTerm
			reduce(196), // <=, reduce: DotTerm
			reduce(196), // ^, reduce: DotTerm
			reduce(196), // &, reduce: DotTerm
			reduce(196), // <<, reduce: DotTerm
			reduce(196), // >>>, reduce: DotTerm
			reduce(196), // +, reduce: DotTerm
			reduce(196), // -, reduce: DotTerm
			reduce(196), // .+, reduce: DotTerm
			reduce(196), // .-, reduce: DotTerm
			reduce(196), // *, reduce: DotTerm
			reduce(196), // /, reduce: DotTerm
			reduce(196), // //, reduce: DotTerm
			reduce(196), // %, reduce: DotTerm
			reduce(196), // .*, reduce: DotTerm
			reduce(196), // ./, reduce: DotTerm
			reduce(196), // .//, reduce: DotTerm
			reduce(196), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S570
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S571
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S572
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(203), // >, reduce: UnaryOpTerm
			reduce(203), // >>, reduce: UnaryOpTerm
			reduce(203), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(203), // ), reduce: UnaryOpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(203), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(203), // ||, reduce: UnaryOpTerm
			reduce(203), // ^^, reduce: UnaryOpTerm
			reduce(203), // &&, reduce: UnaryOpTerm
			reduce(203), // ??, reduce: UnaryOpTerm
			reduce(203), // ???, reduce: UnaryOpTerm
			reduce(203), // =~, reduce: UnaryOpTerm
			reduce(203), // !=~, reduce: UnaryOpTerm
			reduce(203), // ==, reduce: UnaryOpTerm
			reduce(203), // !=, reduce: UnaryOpTerm
			reduce(203), // <=>, reduce: UnaryOpTerm
			reduce(203), // >=, reduce: UnaryOpTerm
			reduce(203), // <, reduce: UnaryOpTerm
			reduce(203), // <=, reduce: UnaryOpTerm
			reduce(203), // ^, reduce: UnaryOpTerm
			reduce(203), // &, reduce: UnaryOpTerm
			reduce(203), // <<, reduce: UnaryOpTerm
			reduce(203), // >>>, reduce: UnaryOpTerm
			reduce(203), // +, reduce: UnaryOpTerm
			reduce(203), // -, reduce: UnaryOpTerm
			reduce(203), // .+, reduce: UnaryOpTerm
			reduce(203), // .-, reduce: UnaryOpTerm
			reduce(203), // *, reduce: UnaryOpTerm
			reduce(203), // /, reduce: UnaryOpTerm
			reduce(203), // //, reduce: UnaryOpTerm
			reduce(203), // %, reduce: UnaryOpTerm
			reduce(203), // .*, reduce: UnaryOpTerm
			reduce(203), // ./, reduce: UnaryOpTerm
			reduce(203), // .//, reduce: UnaryOpTerm
			reduce(203), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S573
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(207), // >, reduce: PowTerm
			reduce(207), // >>, reduce: PowTerm
			reduce(207), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(207), // ), reduce: PowTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(207), // ?, reduce: PowTerm
			nil,         // :
			reduce(207), // ||, reduce: PowTerm
			reduce(207), // ^^, reduce: PowTerm
			reduce(207), // &&, reduce: PowTerm
			reduce(207), // ??, reduce: PowTerm
			reduce(207), // ???, reduce: PowTerm
			reduce(207), // =~, reduce: PowTerm
			reduce(207), // !=~, reduce: PowTerm
			reduce(207), // ==, reduce: PowTerm
			reduce(207), // !=, reduce: PowTerm
			reduce(207), // <=>, reduce: PowTerm
			reduce(207), // >=, reduce: PowTerm
			reduce(207), // <, reduce: PowTerm
			reduce(207), // <=, reduce: PowTerm
			reduce(207), // ^, reduce: PowTerm
			reduce(207), // &, reduce: PowTerm
			reduce(207), // <<, reduce: PowTerm
			reduce(207), // >>>, reduce: PowTerm
			reduce(207), // +, reduce: PowTerm
			reduce(207), // -, reduce: PowTerm
			reduce(207), // .+, reduce: PowTerm
			reduce(207), // .-, reduce: PowTerm
			reduce(207), // *, reduce: PowTerm
			reduce(207), // /, reduce: PowTerm
			reduce(207), // //, reduce: PowTerm
			reduce(207), // %, reduce: PowTerm
			reduce(207), // .*, reduce: PowTerm
			reduce(207), // ./, reduce: PowTerm
			reduce(207), // .//, reduce: PowTerm
			reduce(207), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			shift(1400), // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S574
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(209), // >, reduce: PrecedenceChainEnd
			reduce(209), // >>, reduce: PrecedenceChainEnd
			reduce(209), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(209), // ), reduce: PrecedenceChainEnd
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(209), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(209), // ||, reduce: PrecedenceChainEnd
			reduce(209), // ^^, reduce: PrecedenceChainEnd
			reduce(209), // &&, reduce: PrecedenceChainEnd
			reduce(209), // ??, reduce: PrecedenceChainEnd
			reduce(209), // ???, reduce: PrecedenceChainEnd
			reduce(209), // =~, reduce: PrecedenceChainEnd
			reduce(209), // !=~, reduce: PrecedenceChainEnd
			reduce(209), // ==, reduce: PrecedenceChainEnd
			reduce(209), // !=, reduce: PrecedenceChainEnd
			reduce(209), // <=>, reduce: PrecedenceChainEnd
			reduce(209), // >=, reduce: PrecedenceChainEnd
			reduce(209), // <, reduce: PrecedenceChainEnd
			reduce(209), // <=, reduce: PrecedenceChainEnd
			reduce(209), // ^, reduce: PrecedenceChainEnd
			reduce(209), // &, reduce: PrecedenceChainEnd
			reduce(209), // <<, reduce: PrecedenceChainEnd
			reduce(209), // >>>, reduce: PrecedenceChainEnd
			reduce(209), // +, reduce: PrecedenceChainEnd
			reduce(209), // -, reduce: PrecedenceChainEnd
			reduce(209), // .+, reduce: PrecedenceChainEnd
			reduce(209), // .-, reduce: PrecedenceChainEnd
			reduce(209), // *, reduce: PrecedenceChainEnd
			reduce(209), // /, reduce: PrecedenceChainEnd
			reduce(209), // //, reduce: PrecedenceChainEnd
			reduce(209), // %, reduce: PrecedenceChainEnd
			reduce(209), // .*, reduce: PrecedenceChainEnd
			reduce(209), // ./, reduce: PrecedenceChainEnd
			reduce(209), // .//, reduce: PrecedenceChainEnd
			reduce(209), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(209), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(1401), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			shift(1402), // [[
			shift(1403), // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S575
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(215), // >, reduce: MlrvalOrFunction
			reduce(215), // >>, reduce: MlrvalOrFunction
			reduce(215), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(215), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(215), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(215), // ||, reduce: MlrvalOrFunction
			reduce(215), // ^^, reduce: MlrvalOrFunction
			reduce(215), // &&, reduce: MlrvalOrFunction
			reduce(215), // ??, reduce: MlrvalOrFunction
			reduce(215), // ???, reduce: MlrvalOrFunction
			reduce(215), // =~, reduce: MlrvalOrFunction
			reduce(215), // !=~, reduce: MlrvalOrFunction
			reduce(215), // ==, reduce: MlrvalOrFunction
			reduce(215), // !=, reduce: MlrvalOrFunction
			reduce(215), // <=>, reduce: MlrvalOrFunction
			reduce(215), // >=, reduce: MlrvalOrFunction
			reduce(215), // <, reduce: MlrvalOrFunction
			reduce(215), // <=, reduce: MlrvalOrFunction
			reduce(215), // ^, reduce: MlrvalOrFunction
			reduce(215), // &, reduce: MlrvalOrFunction
			reduce(215), // <<, reduce: MlrvalOrFunction
			reduce(215), // >>>, reduce: MlrvalOrFunction
			reduce(215), // +, reduce: MlrvalOrFunction
			reduce(215), // -, reduce: MlrvalOrFunction
			reduce(215), // .+, reduce: MlrvalOrFunction
			reduce(215), // .-, reduce: MlrvalOrFunction
			reduce(215), // *, reduce: MlrvalOrFunction
			reduce(215), // /, reduce: MlrvalOrFunction
			reduce(215), // //, reduce: MlrvalOrFunction
			reduce(215), // %, reduce: MlrvalOrFunction
			reduce(215), // .*, reduce: MlrvalOrFunction
			reduce(215), // ./, reduce: MlrvalOrFunction
			reduce(215), // .//, reduce: MlrvalOrFunction
			reduce(215), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(215), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(215), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(215), // [[, reduce: MlrvalOrFunction
			reduce(215), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S576
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(216), // >, reduce: MlrvalOrFunction
			reduce(216), // >>, reduce: MlrvalOrFunction
			reduce(216), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(216), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(216), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(216), // ||, reduce: MlrvalOrFunction
			reduce(216), // ^^, reduce: MlrvalOrFunction
			reduce(216), // &&, reduce: MlrvalOrFunction
			reduce(216), // ??, reduce: MlrvalOrFunction
			reduce(216), // ???, reduce: MlrvalOrFunction
			reduce(216), // =~, reduce: MlrvalOrFunction
			reduce(216), // !=~, reduce: MlrvalOrFunction
			reduce(216), // ==, reduce: MlrvalOrFunction
			reduce(216), // !=, reduce: MlrvalOrFunction
			reduce(216), // <=>, reduce: MlrvalOrFunction
			reduce(216), // >=, reduce: MlrvalOrFunction
			reduce(216), // <, reduce: MlrvalOrFunction
			reduce(216), // <=, reduce: MlrvalOrFunction
			reduce(216), // ^, reduce: MlrvalOrFunction
			reduce(216), // &, reduce: MlrvalOrFunction
			reduce(216), // <<, reduce: MlrvalOrFunction
			reduce(216), // >>>, reduce: MlrvalOrFunction
			reduce(216), // +, reduce: MlrvalOrFunction
			reduce(216), // -, reduce: MlrvalOrFunction
			reduce(216), // .+, reduce: MlrvalOrFunction
			reduce(216), // .-, reduce: MlrvalOrFunction
			reduce(216), // *, reduce: MlrvalOrFunction
			reduce(216), // /, reduce: MlrvalOrFunction
			reduce(216), // //, reduce: MlrvalOrFunction
			reduce(216), // %, reduce: MlrvalOrFunction
			reduce(216), // .*, reduce: MlrvalOrFunction
			reduce(216), // ./, reduce: MlrvalOrFunction
			reduce(216), // .//, reduce: MlrvalOrFunction
			reduce(216), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(216), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(216), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(216), // [[, reduce: MlrvalOrFunction
			reduce(216), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S577
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(217), // >, reduce: MlrvalOrFunction
			reduce(217), // >>, reduce: MlrvalOrFunction
			reduce(217), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(217), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(217), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(217), // ||, reduce: MlrvalOrFunction
			reduce(217), // ^^, reduce: MlrvalOrFunction
			reduce(217), // &&, reduce: MlrvalOrFunction
			reduce(217), // ??, reduce: MlrvalOrFunction
			reduce(217), // ???, reduce: MlrvalOrFunction
			reduce(217), // =~, reduce: MlrvalOrFunction
			reduce(217), // !=~, reduce: MlrvalOrFunction
			reduce(217), // ==, reduce: MlrvalOrFunction
			reduce(217), // !=, reduce: MlrvalOrFunction
			reduce(217), // <=>, reduce: MlrvalOrFunction
			reduce(217), // >=, reduce: MlrvalOrFunction
			reduce(217), // <, reduce: MlrvalOrFunction
			reduce(217), // <=, reduce: MlrvalOrFunction
			reduce(217), // ^, reduce: MlrvalOrFunction
			reduce(217), // &, reduce: MlrvalOrFunction
			reduce(217), // <<, reduce: MlrvalOrFunction
			reduce(217), // >>>, reduce: MlrvalOrFunction
			reduce(217), // +, reduce: MlrvalOrFunction
			reduce(217), // -, reduce: MlrvalOrFunction
			reduce(217), // .+, reduce: MlrvalOrFunction
			reduce(217), // .-, reduce: MlrvalOrFunction
			reduce(217), // *, reduce: MlrvalOrFunction
			reduce(217), // /, reduce: MlrvalOrFunction
			reduce(217), // //, reduce: MlrvalOrFunction
			reduce(217), // %, reduce: MlrvalOrFunction
			reduce(217), // .*, reduce: MlrvalOrFunction
			reduce(217), // ./, reduce: MlrvalOrFunction
			reduce(217), // .//, reduce: MlrvalOrFunction
			reduce(217), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(217), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(217), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(217), // [[, reduce: MlrvalOrFunction
			reduce(217), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S578
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(218), // >, reduce: MlrvalOrFunction
			reduce(218), // >>, reduce: MlrvalOrFunction
			reduce(218), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(218), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(218), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(218), // ||, reduce: MlrvalOrFunction
			reduce(218), // ^^, reduce: MlrvalOrFunction
			reduce(218), // &&, reduce: MlrvalOrFunction
			reduce(218), // ??, reduce: MlrvalOrFunction
			reduce(218), // ???, reduce: MlrvalOrFunction
			reduce(218), // =~, reduce: MlrvalOrFunction
			reduce(218), // !=~, reduce: MlrvalOrFunction
			reduce(218), // ==, reduce: MlrvalOrFunction
			reduce(218), // !=, reduce: MlrvalOrFunction
			reduce(218), // <=>, reduce: MlrvalOrFunction
			reduce(218), // >=, reduce: MlrvalOrFunction
			reduce(218), // <, reduce: MlrvalOrFunction
			reduce(218), // <=, reduce: MlrvalOrFunction
			reduce(218), // ^, reduce: MlrvalOrFunction
			reduce(218), // &, reduce: MlrvalOrFunction
			reduce(218), // <<, reduce: MlrvalOrFunction
			reduce(218), // >>>, reduce: MlrvalOrFunction
			reduce(218), // +, reduce: MlrvalOrFunction
			reduce(218), // -, reduce: MlrvalOrFunction
			reduce(218), // .+, reduce: MlrvalOrFunction
			reduce(218), // .-, reduce: MlrvalOrFunction
			reduce(218), // *, reduce: MlrvalOrFunction
			reduce(218), // /, reduce: MlrvalOrFunction
			reduce(218), // //, reduce: MlrvalOrFunction
			reduce(218), // %, reduce: MlrvalOrFunction
			reduce(218), // .*, reduce: MlrvalOrFunction
			reduce(218), // ./, reduce: MlrvalOrFunction
			reduce(218), // .//, reduce: MlrvalOrFunction
			reduce(218), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(218), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(218), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(218), // [[, reduce: MlrvalOrFunction
			reduce(218), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S579
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(219), // >, reduce: MlrvalOrFunction
			reduce(219), // >>, reduce: MlrvalOrFunction
			reduce(219), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(219), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(219), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(219), // ||, reduce: MlrvalOrFunction
			reduce(219), // ^^, reduce: MlrvalOrFunction
			reduce(219), // &&, reduce: MlrvalOrFunction
			reduce(219), // ??, reduce: MlrvalOrFunction
			reduce(219), // ???, reduce: MlrvalOrFunction
			reduce(219), // =~, reduce: MlrvalOrFunction
			reduce(219), // !=~, reduce: MlrvalOrFunction
			reduce(219), // ==, reduce: MlrvalOrFunction
			reduce(219), // !=, reduce: MlrvalOrFunction
			reduce(219), // <=>, reduce: MlrvalOrFunction
			reduce(219), // >=, reduce: MlrvalOrFunction
			reduce(219), // <, reduce: MlrvalOrFunction
			reduce(219), // <=, reduce: MlrvalOrFunction
			reduce(219), // ^, reduce: MlrvalOrFunction
			reduce(219), // &, reduce: MlrvalOrFunction
			reduce(219), // <<, reduce: MlrvalOrFunction
			reduce(219), // >>>, reduce: MlrvalOrFunction
			reduce(219), // +, reduce: MlrvalOrFunction
			reduce(219), // -, reduce: MlrvalOrFunction
			reduce(219), // .+, reduce: MlrvalOrFunction
			reduce(219), // .-, reduce: MlrvalOrFunction
			reduce(219), // *, reduce: MlrvalOrFunction
			reduce(219), // /, reduce: MlrvalOrFunction
			reduce(219), // //, reduce: MlrvalOrFunction
			reduce(219), // %, reduce: MlrvalOrFunction
			reduce(219), // .*, reduce: MlrvalOrFunction
			reduce(219), // ./, reduce: MlrvalOrFunction
			reduce(219), // .//, reduce: MlrvalOrFunction
			reduce(219), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(219), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(219), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(219), // [[, reduce: MlrvalOrFunction
			reduce(219), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S580
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(220), // >, reduce: MlrvalOrFunction
			reduce(220), // >>, reduce: MlrvalOrFunction
			reduce(220), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(220), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(220), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(220), // ||, reduce: MlrvalOrFunction
			reduce(220), // ^^, reduce: MlrvalOrFunction
			reduce(220), // &&, reduce: MlrvalOrFunction
			reduce(220), // ??, reduce: MlrvalOrFunction
			reduce(220), // ???, reduce: MlrvalOrFunction
			reduce(220), // =~, reduce: MlrvalOrFunction
			reduce(220), // !=~, reduce: MlrvalOrFunction
			reduce(220), // ==, reduce: MlrvalOrFunction
			reduce(220), // !=, reduce: MlrvalOrFunction
			reduce(220), // <=>, reduce: MlrvalOrFunction
			reduce(220), // >=, reduce: MlrvalOrFunction
			reduce(220), // <, reduce: MlrvalOrFunction
			reduce(220), // <=, reduce: MlrvalOrFunction
			reduce(220), // ^, reduce: MlrvalOrFunction
			reduce(220), // &, reduce: MlrvalOrFunction
			reduce(220), // <<, reduce: MlrvalOrFunction
			reduce(220), // >>>, reduce: MlrvalOrFunction
			reduce(220), // +, reduce: MlrvalOrFunction
			reduce(220), // -, reduce: MlrvalOrFunction
			reduce(220), // .+, reduce: MlrvalOrFunction
			reduce(220), // .-, reduce: MlrvalOrFunction
			reduce(220), // *, reduce: MlrvalOrFunction
			reduce(220), // /, reduce: MlrvalOrFunction
			reduce(220), // //, reduce: MlrvalOrFunction
			reduce(220), // %, reduce: MlrvalOrFunction
			reduce(220), // .*, reduce: MlrvalOrFunction
			reduce(220), // ./, reduce: MlrvalOrFunction
			reduce(220), // .//, reduce: MlrvalOrFunction
			reduce(220), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(220), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(220), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(220), // [[, reduce: MlrvalOrFunction
			reduce(220), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S581
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(221), // >, reduce: MlrvalOrFunction
			reduce(221), // >>, reduce: MlrvalOrFunction
			reduce(221), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(221), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(221), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(221), // ||, reduce: MlrvalOrFunction
			reduce(221), // ^^, reduce: MlrvalOrFunction
			reduce(221), // &&, reduce: MlrvalOrFunction
			reduce(221), // ??, reduce: MlrvalOrFunction
			reduce(221), // ???, reduce: MlrvalOrFunction
			reduce(221), // =~, reduce: MlrvalOrFunction
			reduce(221), // !=~, reduce: MlrvalOrFunction
			reduce(221), // ==, reduce: MlrvalOrFunction
			reduce(221), // !=, reduce: MlrvalOrFunction
			reduce(221), // <=>, reduce: MlrvalOrFunction
			reduce(221), // >=, reduce: MlrvalOrFunction
			reduce(221), // <, reduce: MlrvalOrFunction
			reduce(221), // <=, reduce: MlrvalOrFunction
			reduce(221), // ^, reduce: MlrvalOrFunction
			reduce(221), // &, reduce: MlrvalOrFunction
			reduce(221), // <<, reduce: MlrvalOrFunction
			reduce(221), // >>>, reduce: MlrvalOrFunction
			reduce(221), // +, reduce: MlrvalOrFunction
			reduce(221), // -, reduce: MlrvalOrFunction
			reduce(221), // .+, reduce: MlrvalOrFunction
			reduce(221), // .-, reduce: MlrvalOrFunction
			reduce(221), // *, reduce: MlrvalOrFunction
			reduce(221), // /, reduce: MlrvalOrFunction
			reduce(221), // //, reduce: MlrvalOrFunction
			reduce(221), // %, reduce: MlrvalOrFunction
			reduce(221), // .*, reduce: MlrvalOrFunction
			reduce(221), // ./, reduce: MlrvalOrFunction
			reduce(221), // .//, reduce: MlrvalOrFunction
			reduce(221), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(221), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(221), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(221), // [[, reduce: MlrvalOrFunction
			reduce(221), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S582
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(222), // >, reduce: MlrvalOrFunction
			reduce(222), // >>, reduce: MlrvalOrFunction
			reduce(222), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(222), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(222), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(222), // ||, reduce: MlrvalOrFunction
			reduce(222), // ^^, reduce: MlrvalOrFunction
			reduce(222), // &&, reduce: MlrvalOrFunction
			reduce(222), // ??, reduce: MlrvalOrFunction
			reduce(222), // ???, reduce: MlrvalOrFunction
			reduce(222), // =~, reduce: MlrvalOrFunction
			reduce(222), // !=~, reduce: MlrvalOrFunction
			reduce(222), // ==, reduce: MlrvalOrFunction
			reduce(222), // !=, reduce: MlrvalOrFunction
			reduce(222), // <=>, reduce: MlrvalOrFunction
			reduce(222), // >=, reduce: MlrvalOrFunction
			reduce(222), // <, reduce: MlrvalOrFunction
			reduce(222), // <=, reduce: MlrvalOrFunction
			reduce(222), // ^, reduce: MlrvalOrFunction
			reduce(222), // &, reduce: MlrvalOrFunction
			reduce(222), // <<, reduce: MlrvalOrFunction
			reduce(222), // >>>, reduce: MlrvalOrFunction
			reduce(222), // +, reduce: MlrvalOrFunction
			reduce(222), // -, reduce: MlrvalOrFunction
			reduce(222), // .+, reduce: MlrvalOrFunction
			reduce(222), // .-, reduce: MlrvalOrFunction
			reduce(222), // *, reduce: MlrvalOrFunction
			reduce(222), // /, reduce: MlrvalOrFunction
			reduce(222), // //, reduce: MlrvalOrFunction
			reduce(222), // %, reduce: MlrvalOrFunction
			reduce(222), // .*, reduce: MlrvalOrFunction
			reduce(222), // ./, reduce: MlrvalOrFunction
			reduce(222), // .//, reduce: MlrvalOrFunction
			reduce(222), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(222), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(222), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(222), // [[, reduce: MlrvalOrFunction
			reduce(222), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S583
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(223), // >, reduce: MlrvalOrFunction
			reduce(223), // >>, reduce: MlrvalOrFunction
			reduce(223), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(223), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(223), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(223), // ||, reduce: MlrvalOrFunction
			reduce(223), // ^^, reduce: MlrvalOrFunction
			reduce(223), // &&, reduce: MlrvalOrFunction
			reduce(223), // ??, reduce: MlrvalOrFunction
			reduce(223), // ???, reduce: MlrvalOrFunction
			reduce(223), // =~, reduce: MlrvalOrFunction
			reduce(223), // !=~, reduce: MlrvalOrFunction
			reduce(223), // ==, reduce: MlrvalOrFunction
			reduce(223), // !=, reduce: MlrvalOrFunction
			reduce(223), // <=>, reduce: MlrvalOrFunction
			reduce(223), // >=, reduce: MlrvalOrFunction
			reduce(223), // <, reduce: MlrvalOrFunction
			reduce(223), // <=, reduce: MlrvalOrFunction
			reduce(223), // ^, reduce: MlrvalOrFunction
			reduce(223), // &, reduce: MlrvalOrFunction
			reduce(223), // <<, reduce: MlrvalOrFunction
			reduce(223), // >>>, reduce: MlrvalOrFunction
			reduce(223), // +, reduce: MlrvalOrFunction
			reduce(223), // -, reduce: MlrvalOrFunction
			reduce(223), // .+, reduce: MlrvalOrFunction
			reduce(223), // .-, reduce: MlrvalOrFunction
			reduce(223), // *, reduce: MlrvalOrFunction
			reduce(223), // /, reduce: MlrvalOrFunction
			reduce(223), // //, reduce: MlrvalOrFunction
			reduce(223), // %, reduce: MlrvalOrFunction
			reduce(223), // .*, reduce: MlrvalOrFunction
			reduce(223), // ./, reduce: MlrvalOrFunction
			reduce(223), // .//, reduce: MlrvalOrFunction
			reduce(223), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(223), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(223), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(223), // [[, reduce: MlrvalOrFunction
			reduce(223), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S584
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(224), // >, reduce: MlrvalOrFunction
			reduce(224), // >>, reduce: MlrvalOrFunction
			reduce(224), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(224), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(224), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(224), // ||, reduce: MlrvalOrFunction
			reduce(224), // ^^, reduce: MlrvalOrFunction
			reduce(224), // &&, reduce: MlrvalOrFunction
			reduce(224), // ??, reduce: MlrvalOrFunction
			reduce(224), // ???, reduce: MlrvalOrFunction
			reduce(224), // =~, reduce: MlrvalOrFunction
			reduce(224), // !=~, reduce: MlrvalOrFunction
			reduce(224), // ==, reduce: MlrvalOrFunction
			reduce(224), // !=, reduce: MlrvalOrFunction
			reduce(224), // <=>, reduce: MlrvalOrFunction
			reduce(224), // >=, reduce: MlrvalOrFunction
			reduce(224), // <, reduce: MlrvalOrFunction
			reduce(224), // <=, reduce: MlrvalOrFunction
			reduce(224), // ^, reduce: MlrvalOrFunction
			reduce(224), // &, reduce: MlrvalOrFunction
			reduce(224), // <<, reduce: MlrvalOrFunction
			reduce(224), // >>>, reduce: MlrvalOrFunction
			reduce(224), // +, reduce: MlrvalOrFunction
			reduce(224), // -, reduce: MlrvalOrFunction
			reduce(224), // .+, reduce: MlrvalOrFunction
			reduce(224), // .-, reduce: MlrvalOrFunction
			reduce(224), // *, reduce: MlrvalOrFunction
			reduce(224), // /, reduce: MlrvalOrFunction
			reduce(224), // //, reduce: MlrvalOrFunction
			reduce(224), // %, reduce: MlrvalOrFunction
			reduce(224), // .*, reduce: MlrvalOrFunction
			reduce(224), // ./, reduce: MlrvalOrFunction
			reduce(224), // .//, reduce: MlrvalOrFunction
			reduce(224), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(224), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(224), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(224), // [[, reduce: MlrvalOrFunction
			reduce(224), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S585
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(225), // >, reduce: MlrvalOrFunction
			reduce(225), // >>, reduce: MlrvalOrFunction
			reduce(225), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(225), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(225), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(225), // ||, reduce: MlrvalOrFunction
			reduce(225), // ^^, reduce: MlrvalOrFunction
			reduce(225), // &&, reduce: MlrvalOrFunction
			reduce(225), // ??, reduce: MlrvalOrFunction
			reduce(225), // ???, reduce: MlrvalOrFunction
			reduce(225), // =~, reduce: MlrvalOrFunction
			reduce(225), // !=~, reduce: MlrvalOrFunction
			reduce(225), // ==, reduce: MlrvalOrFunction
			reduce(225), // !=, reduce: MlrvalOrFunction
			reduce(225), // <=>, reduce: MlrvalOrFunction
			reduce(225), // >=, reduce: MlrvalOrFunction
			reduce(225), // <, reduce: MlrvalOrFunction
			reduce(225), // <=, reduce: MlrvalOrFunction
			reduce(225), // ^, reduce: MlrvalOrFunction
			reduce(225), // &, reduce: MlrvalOrFunction
			reduce(225), // <<, reduce: MlrvalOrFunction
			reduce(225), // >>>, reduce: MlrvalOrFunction
			reduce(225), // +, reduce: MlrvalOrFunction
			reduce(225), // -, reduce: MlrvalOrFunction
			reduce(225), // .+, reduce: MlrvalOrFunction
			reduce(225), // .-, reduce: MlrvalOrFunction
			reduce(225), // *, reduce: MlrvalOrFunction
			reduce(225), // /, reduce: MlrvalOrFunction
			reduce(225), // //, reduce: MlrvalOrFunction
			reduce(225), // %, reduce: MlrvalOrFunction
			reduce(225), // .*, reduce: MlrvalOrFunction
			reduce(225), // ./, reduce: MlrvalOrFunction
			reduce(225), // .//, reduce: MlrvalOrFunction
			reduce(225), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(225), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(225), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(225), // [[, reduce: MlrvalOrFunction
			reduce(225), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S586
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(226), // >, reduce: MlrvalOrFunction
			reduce(226), // >>, reduce: MlrvalOrFunction
			reduce(226), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(226), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(226), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(226), // ||, reduce: MlrvalOrFunction
			reduce(226), // ^^, reduce: MlrvalOrFunction
			reduce(226), // &&, reduce: MlrvalOrFunction
			reduce(226), // ??, reduce: MlrvalOrFunction
			reduce(226), // ???, reduce: MlrvalOrFunction
			reduce(226), // =~, reduce: MlrvalOrFunction
			reduce(226), // !=~, reduce: MlrvalOrFunction
			reduce(226), // ==, reduce: MlrvalOrFunction
			reduce(226), // !=, reduce: MlrvalOrFunction
			reduce(226), // <=>, reduce: MlrvalOrFunction
			reduce(226), // >=, reduce: MlrvalOrFunction
			reduce(226), // <, reduce: MlrvalOrFunction
			reduce(226), // <=, reduce: MlrvalOrFunction
			reduce(226), // ^, reduce: MlrvalOrFunction
			reduce(226), // &, reduce: MlrvalOrFunction
			reduce(226), // <<, reduce: MlrvalOrFunction
			reduce(226), // >>>, reduce: MlrvalOrFunction
			reduce(226), // +, reduce: MlrvalOrFunction
			reduce(226), // -, reduce: MlrvalOrFunction
			reduce(226), // .+, reduce: MlrvalOrFunction
			reduce(226), // .-, reduce: MlrvalOrFunction
			reduce(226), // *, reduce: MlrvalOrFunction
			reduce(226), // /, reduce: MlrvalOrFunction
			reduce(226), // //, reduce: MlrvalOrFunction
			reduce(226), // %, reduce: MlrvalOrFunction
			reduce(226), // .*, reduce: MlrvalOrFunction
			reduce(226), // ./, reduce: MlrvalOrFunction
			reduce(226), // .//, reduce: MlrvalOrFunction
			reduce(226), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(226), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(226), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(226), // [[, reduce: MlrvalOrFunction
			reduce(226), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S587
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(227), // >, reduce: MlrvalOrFunction
			reduce(227), // >>, reduce: MlrvalOrFunction
			reduce(227), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(227), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(227), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(227), // ||, reduce: MlrvalOrFunction
			reduce(227), // ^^, reduce: MlrvalOrFunction
			reduce(227), // &&, reduce: MlrvalOrFunction
			reduce(227), // ??, reduce: MlrvalOrFunction
			reduce(227), // ???, reduce: MlrvalOrFunction
			reduce(227), // =~, reduce: MlrvalOrFunction
			reduce(227), // !=~, reduce: MlrvalOrFunction
			reduce(227), // ==, reduce: MlrvalOrFunction
			reduce(227), // !=, reduce: MlrvalOrFunction
			reduce(227), // <=>, reduce: MlrvalOrFunction
			reduce(227), // >=, reduce: MlrvalOrFunction
			reduce(227), // <, reduce: MlrvalOrFunction
			reduce(227), // <=, reduce: MlrvalOrFunction
			reduce(227), // ^, reduce: MlrvalOrFunction
			reduce(227), // &, reduce: MlrvalOrFunction
			reduce(227), // <<, reduce: MlrvalOrFunction
			reduce(227), // >>>, reduce: MlrvalOrFunction
			reduce(227), // +, reduce: MlrvalOrFunction
			reduce(227), // -, reduce: MlrvalOrFunction
			reduce(227), // .+, reduce: MlrvalOrFunction
			reduce(227), // .-, reduce: MlrvalOrFunction
			reduce(227), // *, reduce: MlrvalOrFunction
			reduce(227), // /, reduce: MlrvalOrFunction
			reduce(227), // //, reduce: MlrvalOrFunction
			reduce(227), // %, reduce: MlrvalOrFunction
			reduce(227), // .*, reduce: MlrvalOrFunction
			reduce(227), // ./, reduce: MlrvalOrFunction
			reduce(227), // .//, reduce: MlrvalOrFunction
			reduce(227), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(227), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(227), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(227), // [[, reduce: MlrvalOrFunction
			reduce(227), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S588
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(752),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(756),  // (
			nil,         // )
			nil,         // emitp
			shift(768),  // field_name
			shift(769),  // $[
			shift(1404), // ]
			shift(771),  // braced_field_name
			shift(772),  // $[[
			shift(773),  // $[[[
			shift(774),  // full_srec
			shift(776),  // oosvar_name
			shift(777),  // @[
			shift(778),  // braced_oosvar_name
			shift(779),  // full_oosvar
			shift(780),  // all
			shift(781),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(796),  // +
			shift(798),  // -
			shift(799),  // .+
			shift(800),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(803),  // !
			shift(804),  // ~
			nil,         // **
			shift(809),  // string_literal
			shift(810),  // regex_case_insensitive
			shift(811),  // int_literal
			shift(812),  // float_literal
			shift(813),  // boolean_literal
			shift(814),  // null_literal
			shift(815),  // inf_literal
			shift(816),  // nan_literal
			shift(817),  // const_M_PI
			shift(818),  // const_M_E
			shift(819),  // panic
			shift(821),  // [
			shift(824),  // ctx_IPS
			shift(825),  // ctx_IFS
			shift(826),  // ctx_IRS
			shift(827),  // ctx_OPS
			shift(828),  // ctx_OFS
			shift(829),  // ctx_ORS
			shift(830),  // ctx_FLATSEP
			shift(831),  // ctx_NF
			shift(832),  // ctx_NR
			shift(833),  // ctx_FNR
			shift(834),  // ctx_FILENAME
			shift(835),  // ctx_FILENUM
			shift(837),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(844),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S589
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(240), // >, reduce: MlrvalOrFunction
			reduce(240), // >>, reduce: MlrvalOrFunction
			reduce(240), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(240), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(240), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(240), // ||, reduce: MlrvalOrFunction
			reduce(240), // ^^, reduce: MlrvalOrFunction
			reduce(240), // &&, reduce: MlrvalOrFunction
			reduce(240), // ??, reduce: MlrvalOrFunction
			reduce(240), // ???, reduce: MlrvalOrFunction
			reduce(240), // =~, reduce: MlrvalOrFunction
			reduce(240), // !=~, reduce: MlrvalOrFunction
			reduce(240), // ==, reduce: MlrvalOrFunction
			reduce(240), // !=, reduce: MlrvalOrFunction
			reduce(240), // <=>, reduce: MlrvalOrFunction
			reduce(240), // >=, reduce: MlrvalOrFunction
			reduce(240), // <, reduce: MlrvalOrFunction
			reduce(240), // <=, reduce: MlrvalOrFunction
			reduce(240), // ^, reduce: MlrvalOrFunction
			reduce(240), // &, reduce: MlrvalOrFunction
			reduce(240), // <<, reduce: MlrvalOrFunction
			reduce(240), // >>>, reduce: MlrvalOrFunction
			reduce(240), // +, reduce: MlrvalOrFunction
			reduce(240), // -, reduce: MlrvalOrFunction
			reduce(240), // .+, reduce: MlrvalOrFunction
			reduce(240), // .-, reduce: MlrvalOrFunction
			reduce(240), // *, reduce: MlrvalOrFunction
			reduce(240), // /, reduce: MlrvalOrFunction
			reduce(240), // //, reduce: MlrvalOrFunction
			reduce(240), // %, reduce: MlrvalOrFunction
			reduce(240), // .*, reduce: MlrvalOrFunction
			reduce(240), // ./, reduce: MlrvalOrFunction
			reduce(240), // .//, reduce: MlrvalOrFunction
			reduce(240), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(240), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(240), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(240), // [[, reduce: MlrvalOrFunction
			reduce(240), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S590
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(241), // >, reduce: ContextVariable
			reduce(241), // >>, reduce: ContextVariable
			reduce(241), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(241), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(241), // ?, reduce: ContextVariable
			nil,         // :
			reduce(241), // ||, reduce: ContextVariable
			reduce(241), // ^^, reduce: ContextVariable
			reduce(241), // &&, reduce: ContextVariable
			reduce(241), // ??, reduce: ContextVariable
			reduce(241), // ???, reduce: ContextVariable
			reduce(241), // =~, reduce: ContextVariable
			reduce(241), // !=~, reduce: ContextVariable
			reduce(241), // ==, reduce: ContextVariable
			reduce(241), // !=, reduce: ContextVariable
			reduce(241), // <=>, reduce: ContextVariable
			reduce(241), // >=, reduce: ContextVariable
			reduce(241), // <, reduce: ContextVariable
			reduce(241), // <=, reduce: ContextVariable
			reduce(241), // ^, reduce: ContextVariable
			reduce(241), // &, reduce: ContextVariable
			reduce(241), // <<, reduce: ContextVariable
			reduce(241), // >>>, reduce: ContextVariable
			reduce(241), // +, reduce: ContextVariable
			reduce(241), // -, reduce: ContextVariable
			reduce(241), // .+, reduce: ContextVariable
			reduce(241), // .-, reduce: ContextVariable
			reduce(241), // *, reduce: ContextVariable
			reduce(241), // /, reduce: ContextVariable
			reduce(241), // //, reduce: ContextVariable
			reduce(241), // %, reduce: ContextVariable
			reduce(241), // .*, reduce: ContextVariable
			reduce(241), // ./, reduce: ContextVariable
			reduce(241), // .//, reduce: ContextVariable
			reduce(241), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(241), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(241), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(241), // [[, reduce: ContextVariable
			reduce(241), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S591
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(242), // >, reduce: ContextVariable
			reduce(242), // >>, reduce: ContextVariable
			reduce(242), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(242), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(242), // ?, reduce: ContextVariable
			nil,         // :
			reduce(242), // ||, reduce: ContextVariable
			reduce(242), // ^^, reduce: ContextVariable
			reduce(242), // &&, reduce: ContextVariable
			reduce(242), // ??, reduce: ContextVariable
			reduce(242), // ???, reduce: ContextVariable
			reduce(242), // =~, reduce: ContextVariable
			reduce(242), // !=~, reduce: ContextVariable
			reduce(242), // ==, reduce: ContextVariable
			reduce(242), // !=, reduce: ContextVariable
			reduce(242), // <=>, reduce: ContextVariable
			reduce(242), // >=, reduce: ContextVariable
			reduce(242), // <, reduce: ContextVariable
			reduce(242), // <=, reduce: ContextVariable
			reduce(242), // ^, reduce: ContextVariable
			reduce(242), // &, reduce: ContextVariable
			reduce(242), // <<, reduce: ContextVariable
			reduce(242), // >>>, reduce: ContextVariable
			reduce(242), // +, reduce: ContextVariable
			reduce(242), // -, reduce: ContextVariable
			reduce(242), // .+, reduce: ContextVariable
			reduce(242), // .-, reduce: ContextVariable
			reduce(242), // *, reduce: ContextVariable
			reduce(242), // /, reduce: ContextVariable
			reduce(242), // //, reduce: ContextVariable
			reduce(242), // %, reduce: ContextVariable
			reduce(242), // .*, reduce: ContextVariable
			reduce(242), // ./, reduce: ContextVariable
			reduce(242), // .//, reduce: ContextVariable
			reduce(242), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(242), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(242), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(242), // [[, reduce: ContextVariable
			reduce(242), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S592
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(243), // >, reduce: ContextVariable
			reduce(243), // >>, reduce: ContextVariable
			reduce(243), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(243), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(243), // ?, reduce: ContextVariable
			nil,         // :
			reduce(243), // ||, reduce: ContextVariable
			reduce(243), // ^^, reduce: ContextVariable
			reduce(243), // &&, reduce: ContextVariable
			reduce(243), // ??, reduce: ContextVariable
			reduce(243), // ???, reduce: ContextVariable
			reduce(243), // =~, reduce: ContextVariable
			reduce(243), // !=~, reduce: ContextVariable
			reduce(243), // ==, reduce: ContextVariable
			reduce(243), // !=, reduce: ContextVariable
			reduce(243), // <=>, reduce: ContextVariable
			reduce(243), // >=, reduce: ContextVariable
			reduce(243), // <, reduce: ContextVariable
			reduce(243), // <=, reduce: ContextVariable
			reduce(243), // ^, reduce: ContextVariable
			reduce(243), // &, reduce: ContextVariable
			reduce(243), // <<, reduce: ContextVariable
			reduce(243), // >>>, reduce: ContextVariable
			reduce(243), // +, reduce: ContextVariable
			reduce(243), // -, reduce: ContextVariable
			reduce(243), // .+, reduce: ContextVariable
			reduce(243), // .-, reduce: ContextVariable
			reduce(243), // *, reduce: ContextVariable
			reduce(243), // /, reduce: ContextVariable
			reduce(243), // //, reduce: ContextVariable
			reduce(243), // %, reduce: ContextVariable
			reduce(243), // .*, reduce: ContextVariable
			reduce(243), // ./, reduce: ContextVariable
			reduce(243), // .//, reduce: ContextVariable
			reduce(243), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(243), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(243), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(243), // [[, reduce: ContextVariable
			reduce(243), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S593
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(244), // >, reduce: ContextVariable
			reduce(244), // >>, reduce: ContextVariable
			reduce(244), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(244), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(244), // ?, reduce: ContextVariable
			nil,         // :
			reduce(244), // ||, reduce: ContextVariable
			reduce(244), // ^^, reduce: ContextVariable
			reduce(244), // &&, reduce: ContextVariable
			reduce(244), // ??, reduce: ContextVariable
			reduce(244), // ???, reduce: ContextVariable
			reduce(244), // =~, reduce: ContextVariable
			reduce(244), // !=~, reduce: ContextVariable
			reduce(244), // ==, reduce: ContextVariable
			reduce(244), // !=, reduce: ContextVariable
			reduce(244), // <=>, reduce: ContextVariable
			reduce(244), // >=, reduce: ContextVariable
			reduce(244), // <, reduce: ContextVariable
			reduce(244), // <=, reduce: ContextVariable
			reduce(244), // ^, reduce: ContextVariable
			reduce(244), // &, reduce: ContextVariable
			reduce(244), // <<, reduce: ContextVariable
			reduce(244), // >>>, reduce: ContextVariable
			reduce(244), // +, reduce: ContextVariable
			reduce(244), // -, reduce: ContextVariable
			reduce(244), // .+, reduce: ContextVariable
			reduce(244), // .-, reduce: ContextVariable
			reduce(244), // *, reduce: ContextVariable
			reduce(244), // /, reduce: ContextVariable
			reduce(244), // //, reduce: ContextVariable
			reduce(244), // %, reduce: ContextVariable
			reduce(244), // .*, reduce: ContextVariable
			reduce(244), // ./, reduce: ContextVariable
			reduce(244), // .//, reduce: ContextVariable
			reduce(244), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(244), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(244), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(244), // [[, reduce: ContextVariable
			reduce(244), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S594
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(245), // >, reduce: ContextVariable
			reduce(245), // >>, reduce: ContextVariable
			reduce(245), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(245), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(245), // ?, reduce: ContextVariable
			nil,         // :
			reduce(245), // ||, reduce: ContextVariable
			reduce(245), // ^^, reduce: ContextVariable
			reduce(245), // &&, reduce: ContextVariable
			reduce(245), // ??, reduce: ContextVariable
			reduce(245), // ???, reduce: ContextVariable
			reduce(245), // =~, reduce: ContextVariable
			reduce(245), // !=~, reduce: ContextVariable
			reduce(245), // ==, reduce: ContextVariable
			reduce(245), // !=, reduce: ContextVariable
			reduce(245), // <=>, reduce: ContextVariable
			reduce(245), // >=, reduce: ContextVariable
			reduce(245), // <, reduce: ContextVariable
			reduce(245), // <=, reduce: ContextVariable
			reduce(245), // ^, reduce: ContextVariable
			reduce(245), // &, reduce: ContextVariable
			reduce(245), // <<, reduce: ContextVariable
			reduce(245), // >>>, reduce: ContextVariable
			reduce(245), // +, reduce: ContextVariable
			reduce(245), // -, reduce: ContextVariable
			reduce(245), // .+, reduce: ContextVariable
			reduce(245), // .-, reduce: ContextVariable
			reduce(245), // *, reduce: ContextVariable
			reduce(245), // /, reduce: ContextVariable
			reduce(245), // //, reduce: ContextVariable
			reduce(245), // %, reduce: ContextVariable
			reduce(245), // .*, reduce: ContextVariable
			reduce(245), // ./, reduce: ContextVariable
			reduce(245), // .//, reduce: ContextVariable
			reduce(245), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(245), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(245), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(245), // [[, reduce: ContextVariable
			reduce(245), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S595
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(246), // >, reduce: ContextVariable
			reduce(246), // >>, reduce: ContextVariable
			reduce(246), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(246), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(246), // ?, reduce: ContextVariable
			nil,         // :
			reduce(246), // ||, reduce: ContextVariable
			reduce(246), // ^^, reduce: ContextVariable
			reduce(246), // &&, reduce: ContextVariable
			reduce(246), // ??, reduce: ContextVariable
			reduce(246), // ???, reduce: ContextVariable
			reduce(246), // =~, reduce: ContextVariable
			reduce(246), // !=~, reduce: ContextVariable
			reduce(246), // ==, reduce: ContextVariable
			reduce(246), // !=, reduce: ContextVariable
			reduce(246), // <=>, reduce: ContextVariable
			reduce(246), // >=, reduce: ContextVariable
			reduce(246), // <, reduce: ContextVariable
			reduce(246), // <=, reduce: ContextVariable
			reduce(246), // ^, reduce: ContextVariable
			reduce(246), // &, reduce: ContextVariable
			reduce(246), // <<, reduce: ContextVariable
			reduce(246), // >>>, reduce: ContextVariable
			reduce(246), // +, reduce: ContextVariable
			reduce(246), // -, reduce: ContextVariable
			reduce(246), // .+, reduce: ContextVariable
			reduce(246), // .-, reduce: ContextVariable
			reduce(246), // *, reduce: ContextVariable
			reduce(246), // /, reduce: ContextVariable
			reduce(246), // //, reduce: ContextVariable
			reduce(246), // %, reduce: ContextVariable
			reduce(246), // .*, reduce: ContextVariable
			reduce(246), // ./, reduce: ContextVariable
			reduce(246), // .//, reduce: ContextVariable
			reduce(246), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(246), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(246), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(246), // [[, reduce: ContextVariable
			reduce(246), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S596
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(247), // >, reduce: ContextVariable
			reduce(247), // >>, reduce: ContextVariable
			reduce(247), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(247), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(247), // ?, reduce: ContextVariable
			nil,         // :
			reduce(247), // ||, reduce: ContextVariable
			reduce(247), // ^^, reduce: ContextVariable
			reduce(247), // &&, reduce: ContextVariable
			reduce(247), // ??, reduce: ContextVariable
			reduce(247), // ???, reduce: ContextVariable
			reduce(247), // =~, reduce: ContextVariable
			reduce(247), // !=~, reduce: ContextVariable
			reduce(247), // ==, reduce: ContextVariable
			reduce(247), // !=, reduce: ContextVariable
			reduce(247), // <=>, reduce: ContextVariable
			reduce(247), // >=, reduce: ContextVariable
			reduce(247), // <, reduce: ContextVariable
			reduce(247), // <=, reduce: ContextVariable
			reduce(247), // ^, reduce: ContextVariable
			reduce(247), // &, reduce: ContextVariable
			reduce(247), // <<, reduce: ContextVariable
			reduce(247), // >>>, reduce: ContextVariable
			reduce(247), // +, reduce: ContextVariable
			reduce(247), // -, reduce: ContextVariable
			reduce(247), // .+, reduce: ContextVariable
			reduce(247), // .-, reduce: ContextVariable
			reduce(247), // *, reduce: ContextVariable
			reduce(247), // /, reduce: ContextVariable
			reduce(247), // //, reduce: ContextVariable
			reduce(247), // %, reduce: ContextVariable
			reduce(247), // .*, reduce: ContextVariable
			reduce(247), // ./, reduce: ContextVariable
			reduce(247), // .//, reduce: ContextVariable
			reduce(247), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(247), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(247), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(247), // [[, reduce: ContextVariable
			reduce(247), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S597
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(248), // >, reduce: ContextVariable
			reduce(248), // >>, reduce: ContextVariable
			reduce(248), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(248), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(248), // ?, reduce: ContextVariable
			nil,         // :
			reduce(248), // ||, reduce: ContextVariable
			reduce(248), // ^^, reduce: ContextVariable
			reduce(248), // &&, reduce: ContextVariable
			reduce(248), // ??, reduce: ContextVariable
			reduce(248), // ???, reduce: ContextVariable
			reduce(248), // =~, reduce: ContextVariable
			reduce(248), // !=~, reduce: ContextVariable
			reduce(248), // ==, reduce: ContextVariable
			reduce(248), // !=, reduce: ContextVariable
			reduce(248), // <=>, reduce: ContextVariable
			reduce(248), // >=, reduce: ContextVariable
			reduce(248), // <, reduce: ContextVariable
			reduce(248), // <=, reduce: ContextVariable
			reduce(248), // ^, reduce: ContextVariable
			reduce(248), // &, reduce: ContextVariable
			reduce(248), // <<, reduce: ContextVariable
			reduce(248), // >>>, reduce: ContextVariable
			reduce(248), // +, reduce: ContextVariable
			reduce(248), // -, reduce: ContextVariable
			reduce(248), // .+, reduce: ContextVariable
			reduce(248), // .-, reduce: ContextVariable
			reduce(248), // *, reduce: ContextVariable
			reduce(248), // /, reduce: ContextVariable
			reduce(248), // //, reduce: ContextVariable
			reduce(248), // %, reduce: ContextVariable
			reduce(248), // .*, reduce: ContextVariable
			reduce(248), // ./, reduce: ContextVariable
			reduce(248), // .//, reduce: ContextVariable
			reduce(248), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(248), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(248), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(248), // [[, reduce: ContextVariable
			reduce(248), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S598
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(249), // >, reduce: ContextVariable
			reduce(249), // >>, reduce: ContextVariable
			reduce(249), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(249), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(249), // ?, reduce: ContextVariable
			nil,         // :
			reduce(249), // ||, reduce: ContextVariable
			reduce(249), // ^^, reduce: ContextVariable
			reduce(249), // &&, reduce: ContextVariable
			reduce(249), // ??, reduce: ContextVariable
			reduce(249), // ???, reduce: ContextVariable
			reduce(249), // =~, reduce: ContextVariable
			reduce(249), // !=~, reduce: ContextVariable
			reduce(249), // ==, reduce: ContextVariable
			reduce(249), // !=, reduce: ContextVariable
			reduce(249), // <=>, reduce: ContextVariable
			reduce(249), // >=, reduce: ContextVariable
			reduce(249), // <, reduce: ContextVariable
			reduce(249), // <=, reduce: ContextVariable
			reduce(249), // ^, reduce: ContextVariable
			reduce(249), // &, reduce: ContextVariable
			reduce(249), // <<, reduce: ContextVariable
			reduce(249), // >>>, reduce: ContextVariable
			reduce(249), // +, reduce: ContextVariable
			reduce(249), // -, reduce: ContextVariable
			reduce(249), // .+, reduce: ContextVariable
			reduce(249), // .-, reduce: ContextVariable
			reduce(249), // *, reduce: ContextVariable
			reduce(249), // /, reduce: ContextVariable
			reduce(249), // //, reduce: ContextVariable
			reduce(249), // %, reduce: ContextVariable
			reduce(249), // .*, reduce: ContextVariable
			reduce(249), // ./, reduce: ContextVariable
			reduce(249), // .//, reduce: ContextVariable
			reduce(249), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(249), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(249), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(249), // [[, reduce: ContextVariable
			reduce(249), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S599
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(250), // >, reduce: ContextVariable
			reduce(250), // >>, reduce: ContextVariable
			reduce(250), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(250), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(250), // ?, reduce: ContextVariable
			nil,         // :
			reduce(250), // ||, reduce: ContextVariable
			reduce(250), // ^^, reduce: ContextVariable
			reduce(250), // &&, reduce: ContextVariable
			reduce(250), // ??, reduce: ContextVariable
			reduce(250), // ???, reduce: ContextVariable
			reduce(250), // =~, reduce: ContextVariable
			reduce(250), // !=~, reduce: ContextVariable
			reduce(250), // ==, reduce: ContextVariable
			reduce(250), // !=, reduce: ContextVariable
			reduce(250), // <=>, reduce: ContextVariable
			reduce(250), // >=, reduce: ContextVariable
			reduce(250), // <, reduce: ContextVariable
			reduce(250), // <=, reduce: ContextVariable
			reduce(250), // ^, reduce: ContextVariable
			reduce(250), // &, reduce: ContextVariable
			reduce(250), // <<, reduce: ContextVariable
			reduce(250), // >>>, reduce: ContextVariable
			reduce(250), // +, reduce: ContextVariable
			reduce(250), // -, reduce: ContextVariable
			reduce(250), // .+, reduce: ContextVariable
			reduce(250), // .-, reduce: ContextVariable
			reduce(250), // *, reduce: ContextVariable
			reduce(250), // /, reduce: ContextVariable
			reduce(250), // //, reduce: ContextVariable
			reduce(250), // %, reduce: ContextVariable
			reduce(250), // .*, reduce: ContextVariable
			reduce(250), // ./, reduce: ContextVariable
			reduce(250), // .//, reduce: ContextVariable
			reduce(250), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(250), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(250), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(250), // [[, reduce: ContextVariable
			reduce(250), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S600
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(251), // >, reduce: ContextVariable
			reduce(251), // >>, reduce: ContextVariable
			reduce(251), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(251), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(251), // ?, reduce: ContextVariable
			nil,         // :
			reduce(251), // ||, reduce: ContextVariable
			reduce(251), // ^^, reduce: ContextVariable
			reduce(251), // &&, reduce: ContextVariable
			reduce(251), // ??, reduce: ContextVariable
			reduce(251), // ???, reduce: ContextVariable
			reduce(251), // =~, reduce: ContextVariable
			reduce(251), // !=~, reduce: ContextVariable
			reduce(251), // ==, reduce: ContextVariable
			reduce(251), // !=, reduce: ContextVariable
			reduce(251), // <=>, reduce: ContextVariable
			reduce(251), // >=, reduce: ContextVariable
			reduce(251), // <, reduce: ContextVariable
			reduce(251), // <=, reduce: ContextVariable
			reduce(251), // ^, reduce: ContextVariable
			reduce(251), // &, reduce: ContextVariable
			reduce(251), // <<, reduce: ContextVariable
			reduce(251), // >>>, reduce: ContextVariable
			reduce(251), // +, reduce: ContextVariable
			reduce(251), // -, reduce: ContextVariable
			reduce(251), // .+, reduce: ContextVariable
			reduce(251), // .-, reduce: ContextVariable
			reduce(251), // *, reduce: ContextVariable
			reduce(251), // /, reduce: ContextVariable
			reduce(251), // //, reduce: ContextVariable
			reduce(251), // %, reduce: ContextVariable
			reduce(251), // .*, reduce: ContextVariable
			reduce(251), // ./, reduce: ContextVariable
			reduce(251), // .//, reduce: ContextVariable
			reduce(251), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(251), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(251), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(251), // [[, reduce: ContextVariable
			reduce(251), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S601
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(252), // >, reduce: ContextVariable
			reduce(252), // >>, reduce: ContextVariable
			reduce(252), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(252), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(252), // ?, reduce: ContextVariable
			nil,         // :
			reduce(252), // ||, reduce: ContextVariable
			reduce(252), // ^^, reduce: ContextVariable
			reduce(252), // &&, reduce: ContextVariable
			reduce(252), // ??, reduce: ContextVariable
			reduce(252), // ???, reduce: ContextVariable
			reduce(252), // =~, reduce: ContextVariable
			reduce(252), // !=~, reduce: ContextVariable
			reduce(252), // ==, reduce: ContextVariable
			reduce(252), // !=, reduce: ContextVariable
			reduce(252), // <=>, reduce: ContextVariable
			reduce(252), // >=, reduce: ContextVariable
			reduce(252), // <, reduce: ContextVariable
			reduce(252), // <=, reduce: ContextVariable
			reduce(252), // ^, reduce: ContextVariable
			reduce(252), // &, reduce: ContextVariable
			reduce(252), // <<, reduce: ContextVariable
			reduce(252), // >>>, reduce: ContextVariable
			reduce(252), // +, reduce: ContextVariable
			reduce(252), // -, reduce: ContextVariable
			reduce(252), // .+, reduce: ContextVariable
			reduce(252), // .-, reduce: ContextVariable
			reduce(252), // *, reduce: ContextVariable
			reduce(252), // /, reduce: ContextVariable
			reduce(252), // //, reduce: ContextVariable
			reduce(252), // %, reduce: ContextVariable
			reduce(252), // .*, reduce: ContextVariable
			reduce(252), // ./, reduce: ContextVariable
			reduce(252), // .//, reduce: ContextVariable
			reduce(252), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(252), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(252), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(252), // [[, reduce: ContextVariable
			reduce(252), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S602
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(253), // >, reduce: MlrvalOrFunction
			reduce(253), // >>, reduce: MlrvalOrFunction
			reduce(253), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(253), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(253), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(253), // ||, reduce: MlrvalOrFunction
			reduce(253), // ^^, reduce: MlrvalOrFunction
			reduce(253), // &&, reduce: MlrvalOrFunction
			reduce(253), // ??, reduce: MlrvalOrFunction
			reduce(253), // ???, reduce: MlrvalOrFunction
			reduce(253), // =~, reduce: MlrvalOrFunction
			reduce(253), // !=~, reduce: MlrvalOrFunction
			reduce(253), // ==, reduce: MlrvalOrFunction
			reduce(253), // !=, reduce: MlrvalOrFunction
			reduce(253), // <=>, reduce: MlrvalOrFunction
			reduce(253), // >=, reduce: MlrvalOrFunction
			reduce(253), // <, reduce: MlrvalOrFunction
			reduce(253), // <=, reduce: MlrvalOrFunction
			reduce(253), // ^, reduce: MlrvalOrFunction
			reduce(253), // &, reduce: MlrvalOrFunction
			reduce(253), // <<, reduce: MlrvalOrFunction
			reduce(253), // >>>, reduce: MlrvalOrFunction
			reduce(253), // +, reduce: MlrvalOrFunction
			reduce(253), // -, reduce: MlrvalOrFunction
			reduce(253), // .+, reduce: MlrvalOrFunction
			reduce(253), // .-, reduce: MlrvalOrFunction
			reduce(253), // *, reduce: MlrvalOrFunction
			reduce(253), // /, reduce: MlrvalOrFunction
			reduce(253), // //, reduce: MlrvalOrFunction
			reduce(253), // %, reduce: MlrvalOrFunction
			reduce(253), // .*, reduce: MlrvalOrFunction
			reduce(253), // ./, reduce: MlrvalOrFunction
			reduce(253), // .//, reduce: MlrvalOrFunction
			reduce(253), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(253), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(253), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(253), // [[, reduce: MlrvalOrFunction
			reduce(253), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S603
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			shift(1406), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(1407), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S604
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(256), // >, reduce: MlrvalOrFunction
			reduce(256), // >>, reduce: MlrvalOrFunction
			reduce(256), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(256), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(256), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(256), // ||, reduce: MlrvalOrFunction
			reduce(256), // ^^, reduce: MlrvalOrFunction
			reduce(256), // &&, reduce: MlrvalOrFunction
			reduce(256), // ??, reduce: MlrvalOrFunction
			reduce(256), // ???, reduce: MlrvalOrFunction
			reduce(256), // =~, reduce: MlrvalOrFunction
			reduce(256), // !=~, reduce: MlrvalOrFunction
			reduce(256), // ==, reduce: MlrvalOrFunction
			reduce(256), // !=, reduce: MlrvalOrFunction
			reduce(256), // <=>, reduce: MlrvalOrFunction
			reduce(256), // >=, reduce: MlrvalOrFunction
			reduce(256), // <, reduce: MlrvalOrFunction
			reduce(256), // <=, reduce: MlrvalOrFunction
			reduce(256), // ^, reduce: MlrvalOrFunction
			reduce(256), // &, reduce: MlrvalOrFunction
			reduce(256), // <<, reduce: MlrvalOrFunction
			reduce(256), // >>>, reduce: MlrvalOrFunction
			reduce(256), // +, reduce: MlrvalOrFunction
			reduce(256), // -, reduce: MlrvalOrFunction
			reduce(256), // .+, reduce: MlrvalOrFunction
			reduce(256), // .-, reduce: MlrvalOrFunction
			reduce(256), // *, reduce: MlrvalOrFunction
			reduce(256), // /, reduce: MlrvalOrFunction
			reduce(256), // //, reduce: MlrvalOrFunction
			reduce(256), // %, reduce: MlrvalOrFunction
			reduce(256), // .*, reduce: MlrvalOrFunction
			reduce(256), // ./, reduce: MlrvalOrFunction
			reduce(256), // .//, reduce: MlrvalOrFunction
			reduce(256), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(256), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(256), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(256), // [[, reduce: MlrvalOrFunction
			reduce(256), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S605
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(257), // >, reduce: MlrvalOrFunction
			reduce(257), // >>, reduce: MlrvalOrFunction
			reduce(257), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(257), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(257), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(257), // ||, reduce: MlrvalOrFunction
			reduce(257), // ^^, reduce: MlrvalOrFunction
			reduce(257), // &&, reduce: MlrvalOrFunction
			reduce(257), // ??, reduce: MlrvalOrFunction
			reduce(257), // ???, reduce: MlrvalOrFunction
			reduce(257), // =~, reduce: MlrvalOrFunction
			reduce(257), // !=~, reduce: MlrvalOrFunction
			reduce(257), // ==, reduce: MlrvalOrFunction
			reduce(257), // !=, reduce: MlrvalOrFunction
			reduce(257), // <=>, reduce: MlrvalOrFunction
			reduce(257), // >=, reduce: MlrvalOrFunction
			reduce(257), // <, reduce: MlrvalOrFunction
			reduce(257), // <=, reduce: MlrvalOrFunction
			reduce(257), // ^, reduce: MlrvalOrFunction
			reduce(257), // &, reduce: MlrvalOrFunction
			reduce(257), // <<, reduce: MlrvalOrFunction
			reduce(257), // >>>, reduce: MlrvalOrFunction
			reduce(257), // +, reduce: MlrvalOrFunction
			reduce(257), // -, reduce: MlrvalOrFunction
			reduce(257), // .+, reduce: MlrvalOrFunction
			reduce(257), // .-, reduce: MlrvalOrFunction
			reduce(257), // *, reduce: MlrvalOrFunction
			reduce(257), // /, reduce: MlrvalOrFunction
			reduce(257), // //, reduce: MlrvalOrFunction
			reduce(257), // %, reduce: MlrvalOrFunction
			reduce(257), // .*, reduce: MlrvalOrFunction
			reduce(257), // ./, reduce: MlrvalOrFunction
			reduce(257), // .//, reduce: MlrvalOrFunction
			reduce(257), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(257), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(257), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(257), // [[, reduce: MlrvalOrFunction
			reduce(257), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S606
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(258), // >, reduce: MlrvalOrFunction
			reduce(258), // >>, reduce: MlrvalOrFunction
			reduce(258), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(258), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(258), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(258), // ||, reduce: MlrvalOrFunction
			reduce(258), // ^^, reduce: MlrvalOrFunction
			reduce(258), // &&, reduce: MlrvalOrFunction
			reduce(258), // ??, reduce: MlrvalOrFunction
			reduce(258), // ???, reduce: MlrvalOrFunction
			reduce(258), // =~, reduce: MlrvalOrFunction
			reduce(258), // !=~, reduce: MlrvalOrFunction
			reduce(258), // ==, reduce: MlrvalOrFunction
			reduce(258), // !=, reduce: MlrvalOrFunction
			reduce(258), // <=>, reduce: MlrvalOrFunction
			reduce(258), // >=, reduce: MlrvalOrFunction
			reduce(258), // <, reduce: MlrvalOrFunction
			reduce(258), // <=, reduce: MlrvalOrFunction
			reduce(258), // ^, reduce: MlrvalOrFunction
			reduce(258), // &, reduce: MlrvalOrFunction
			reduce(258), // <<, reduce: MlrvalOrFunction
			reduce(258), // >>>, reduce: MlrvalOrFunction
			reduce(258), // +, reduce: MlrvalOrFunction
			reduce(258), // -, reduce: MlrvalOrFunction
			reduce(258), // .+, reduce: MlrvalOrFunction
			reduce(258), // .-, reduce: MlrvalOrFunction
			reduce(258), // *, reduce: MlrvalOrFunction
			reduce(258), // /, reduce: MlrvalOrFunction
			reduce(258), // //, reduce: MlrvalOrFunction
			reduce(258), // %, reduce: MlrvalOrFunction
			reduce(258), // .*, reduce: MlrvalOrFunction
			reduce(258), // ./, reduce: MlrvalOrFunction
			reduce(258), // .//, reduce: MlrvalOrFunction
			reduce(258), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(258), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(258), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(258), // [[, reduce: MlrvalOrFunction
			reduce(258), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S607
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(259), // >, reduce: MlrvalOrFunction
			reduce(259), // >>, reduce: MlrvalOrFunction
			reduce(259), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(259), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(259), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(259), // ||, reduce: MlrvalOrFunction
			reduce(259), // ^^, reduce: MlrvalOrFunction
			reduce(259), // &&, reduce: MlrvalOrFunction
			reduce(259), // ??, reduce: MlrvalOrFunction
			reduce(259), // ???, reduce: MlrvalOrFunction
			reduce(259), // =~, reduce: MlrvalOrFunction
			reduce(259), // !=~, reduce: MlrvalOrFunction
			reduce(259), // ==, reduce: MlrvalOrFunction
			reduce(259), // !=, reduce: MlrvalOrFunction
			reduce(259), // <=>, reduce: MlrvalOrFunction
			reduce(259), // >=, reduce: MlrvalOrFunction
			reduce(259), // <, reduce: MlrvalOrFunction
			reduce(259), // <=, reduce: MlrvalOrFunction
			reduce(259), // ^, reduce: MlrvalOrFunction
			reduce(259), // &, reduce: MlrvalOrFunction
			reduce(259), // <<, reduce: MlrvalOrFunction
			reduce(259), // >>>, reduce: MlrvalOrFunction
			reduce(259), // +, reduce: MlrvalOrFunction
			reduce(259), // -, reduce: MlrvalOrFunction
			reduce(259), // .+, reduce: MlrvalOrFunction
			reduce(259), // .-, reduce: MlrvalOrFunction
			reduce(259), // *, reduce: MlrvalOrFunction
			reduce(259), // /, reduce: MlrvalOrFunction
			reduce(259), // //, reduce: MlrvalOrFunction
			reduce(259), // %, reduce: MlrvalOrFunction
			reduce(259), // .*, reduce: MlrvalOrFunction
			reduce(259), // ./, reduce: MlrvalOrFunction
			reduce(259), // .//, reduce: MlrvalOrFunction
			reduce(259), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(259), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(259), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(259), // [[, reduce: MlrvalOrFunction
			reduce(259), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S608
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(267), // >, reduce: MlrvalOrFunction
			reduce(267), // >>, reduce: MlrvalOrFunction
			reduce(267), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(267), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(267), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(267), // ||, reduce: MlrvalOrFunction
			reduce(267), // ^^, reduce: MlrvalOrFunction
			reduce(267), // &&, reduce: MlrvalOrFunction
			reduce(267), // ??, reduce: MlrvalOrFunction
			reduce(267), // ???, reduce: MlrvalOrFunction
			reduce(267), // =~, reduce: MlrvalOrFunction
			reduce(267), // !=~, reduce: MlrvalOrFunction
			reduce(267), // ==, reduce: MlrvalOrFunction
			reduce(267), // !=, reduce: MlrvalOrFunction
			reduce(267), // <=>, reduce: MlrvalOrFunction
			reduce(267), // >=, reduce: MlrvalOrFunction
			reduce(267), // <, reduce: MlrvalOrFunction
			reduce(267), // <=, reduce: MlrvalOrFunction
			reduce(267), // ^, reduce: MlrvalOrFunction
			reduce(267), // &, reduce: MlrvalOrFunction
			reduce(267), // <<, reduce: MlrvalOrFunction
			reduce(267), // >>>, reduce: MlrvalOrFunction
			reduce(267), // +, reduce: MlrvalOrFunction
			reduce(267), // -, reduce: MlrvalOrFunction
			reduce(267), // .+, reduce: MlrvalOrFunction
			reduce(267), // .-, reduce: MlrvalOrFunction
			reduce(267), // *, reduce: MlrvalOrFunction
			reduce(267), // /, reduce: MlrvalOrFunction
			reduce(267), // //, reduce: MlrvalOrFunction
			reduce(267), // %, reduce: MlrvalOrFunction
			reduce(267), // .*, reduce: MlrvalOrFunction
			reduce(267), // ./, reduce: MlrvalOrFunction
			reduce(267), // .//, reduce: MlrvalOrFunction
			reduce(267), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(267), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(267), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(267), // [[, reduce: MlrvalOrFunction
			reduce(267), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S609
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1408), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S610
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1409), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S611
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(1410), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S612
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(76),  // $, reduce: EmitPStatement
			nil,         // empty
			reduce(76),  // ;, reduce: EmitPStatement
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(1411), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S613
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1330), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			shift(1343), // field_name
			shift(1344), // $[
			nil,         // ]
			shift(1345), // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(1346), // full_srec
			shift(1347), // oosvar_name
			shift(1348), // @[
			shift(1349), // braced_oosvar_name
			shift(1350), // full_oosvar
			shift(1351), // all
			shift(1352), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S614
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			shift(1413), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S615
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(1415), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S616
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(214), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(214), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S617
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(211), // >, reduce: MlrvalOrFunction
			reduce(211), // >>, reduce: MlrvalOrFunction
			reduce(211), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(211), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(211), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(211), // ||, reduce: MlrvalOrFunction
			reduce(211), // ^^, reduce: MlrvalOrFunction
			reduce(211), // &&, reduce: MlrvalOrFunction
			reduce(211), // ??, reduce: MlrvalOrFunction
			reduce(211), // ???, reduce: MlrvalOrFunction
			reduce(211), // =~, reduce: MlrvalOrFunction
			reduce(211), // !=~, reduce: MlrvalOrFunction
			reduce(211), // ==, reduce: MlrvalOrFunction
			reduce(211), // !=, reduce: MlrvalOrFunction
			reduce(211), // <=>, reduce: MlrvalOrFunction
			reduce(211), // >=, reduce: MlrvalOrFunction
			reduce(211), // <, reduce: MlrvalOrFunction
			reduce(211), // <=, reduce: MlrvalOrFunction
			reduce(211), // ^, reduce: MlrvalOrFunction
			reduce(211), // &, reduce: MlrvalOrFunction
			reduce(211), // <<, reduce: MlrvalOrFunction
			reduce(211), // >>>, reduce: MlrvalOrFunction
			reduce(211), // +, reduce: MlrvalOrFunction
			reduce(211), // -, reduce: MlrvalOrFunction
			reduce(211), // .+, reduce: MlrvalOrFunction
			reduce(211), // .-, reduce: MlrvalOrFunction
			reduce(211), // *, reduce: MlrvalOrFunction
			reduce(211), // /, reduce: MlrvalOrFunction
			reduce(211), // //, reduce: MlrvalOrFunction
			reduce(211), // %, reduce: MlrvalOrFunction
			reduce(211), // .*, reduce: MlrvalOrFunction
			reduce(211), // ./, reduce: MlrvalOrFunction
			reduce(211), // .//, reduce: MlrvalOrFunction
			reduce(211), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(211), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(211), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(211), // [[, reduce: MlrvalOrFunction
			reduce(211), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S618
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S619
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(110), // >, reduce: OosvarValue
			reduce(110), // >>, reduce: OosvarValue
			reduce(110), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(110), // ], reduce: OosvarValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(110), // ?, reduce: OosvarValue
			nil,         // :
			reduce(110), // ||, reduce: OosvarValue
			reduce(110), // ^^, reduce: OosvarValue
			reduce(110), // &&, reduce: OosvarValue
			reduce(110), // ??, reduce: OosvarValue
			reduce(110), // ???, reduce: OosvarValue
			reduce(110), // =~, reduce: OosvarValue
			reduce(110), // !=~, reduce: OosvarValue
			reduce(110), // ==, reduce: OosvarValue
			reduce(110), // !=, reduce: OosvarValue
			reduce(110), // <=>, reduce: OosvarValue
			reduce(110), // >=, reduce: OosvarValue
			reduce(110), // <, reduce: OosvarValue
			reduce(110), // <=, reduce: OosvarValue
			reduce(110), // ^, reduce: OosvarValue
			reduce(110), // &, reduce: OosvarValue
			reduce(110), // <<, reduce: OosvarValue
			reduce(110), // >>>, reduce: OosvarValue
			reduce(110), // +, reduce: OosvarValue
			reduce(110), // -, reduce: OosvarValue
			reduce(110), // .+, reduce: OosvarValue
			reduce(110), // .-, reduce: OosvarValue
			reduce(110), // *, reduce: OosvarValue
			reduce(110), // /, reduce: OosvarValue
			reduce(110), // //, reduce: OosvarValue
			reduce(110), // %, reduce: OosvarValue
			reduce(110), // .*, reduce: OosvarValue
			reduce(110), // ./, reduce: OosvarValue
			reduce(110), // .//, reduce: OosvarValue
			reduce(110), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(110), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(110), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(110), // [[, reduce: OosvarValue
			reduce(110), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S620
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(112), // >, reduce: OosvarValue
			reduce(112), // >>, reduce: OosvarValue
			reduce(112), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(112), // ], reduce: OosvarValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(112), // ?, reduce: OosvarValue
			nil,         // :
			reduce(112), // ||, reduce: OosvarValue
			reduce(112), // ^^, reduce: OosvarValue
			reduce(112), // &&, reduce: OosvarValue
			reduce(112), // ??, reduce: OosvarValue
			reduce(112), // ???, reduce: OosvarValue
			reduce(112), // =~, reduce: OosvarValue
			reduce(112), // !=~, reduce: OosvarValue
			reduce(112), // ==, reduce: OosvarValue
			reduce(112), // !=, reduce: OosvarValue
			reduce(112), // <=>, reduce: OosvarValue
			reduce(112), // >=, reduce: OosvarValue
			reduce(112), // <, reduce: OosvarValue
			reduce(112), // <=, reduce: OosvarValue
			reduce(112), // ^, reduce: OosvarValue
			reduce(112), // &, reduce: OosvarValue
			reduce(112), // <<, reduce: OosvarValue
			reduce(112), // >>>, reduce: OosvarValue
			reduce(112), // +, reduce: OosvarValue
			reduce(112), // -, reduce: OosvarValue
			reduce(112), // .+, reduce: OosvarValue
			reduce(112), // .-, reduce: OosvarValue
			reduce(112), // *, reduce: OosvarValue
			reduce(112), // /, reduce: OosvarValue
			reduce(112), // //, reduce: OosvarValue
			reduce(112), // %, reduce: OosvarValue
			reduce(112), // .*, reduce: OosvarValue
			reduce(112), // ./, reduce: OosvarValue
			reduce(112), // .//, reduce: OosvarValue
			reduce(112), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(112), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(112), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(112), // [[, reduce: OosvarValue
			reduce(112), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S621
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(111), // >, reduce: OosvarValue
			reduce(111), // >>, reduce: OosvarValue
			reduce(111), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(111), // ], reduce: OosvarValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(111), // ?, reduce: OosvarValue
			nil,         // :
			reduce(111), // ||, reduce: OosvarValue
			reduce(111), // ^^, reduce: OosvarValue
			reduce(111), // &&, reduce: OosvarValue
			reduce(111), // ??, reduce: OosvarValue
			reduce(111), // ???, reduce: OosvarValue
			reduce(111), // =~, reduce: OosvarValue
			reduce(111), // !=~, reduce: OosvarValue
			reduce(111), // ==, reduce: OosvarValue
			reduce(111), // !=, reduce: OosvarValue
			reduce(111), // <=>, reduce: OosvarValue
			reduce(111), // >=, reduce: OosvarValue
			reduce(111), // <, reduce: OosvarValue
			reduce(111), // <=, reduce: OosvarValue
			reduce(111), // ^, reduce: OosvarValue
			reduce(111), // &, reduce: OosvarValue
			reduce(111), // <<, reduce: OosvarValue
			reduce(111), // >>>, reduce: OosvarValue
			reduce(111), // +, reduce: OosvarValue
			reduce(111), // -, reduce: OosvarValue
			reduce(111), // .+, reduce: OosvarValue
			reduce(111), // .-, reduce: OosvarValue
			reduce(111), // *, reduce: OosvarValue
			reduce(111), // /, reduce: OosvarValue
			reduce(111), // //, reduce: OosvarValue
			reduce(111), // %, reduce: OosvarValue
			reduce(111), // .*, reduce: OosvarValue
			reduce(111), // ./, reduce: OosvarValue
			reduce(111), // .//, reduce: OosvarValue
			reduce(111), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(111), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(111), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(111), // [[, reduce: OosvarValue
			reduce(111), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S622
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(99), // >, reduce: FieldValue
			reduce(99), // >>, reduce: FieldValue
			reduce(99), // |, reduce: FieldValue
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			reduce(99), // ], reduce: FieldValue
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(99), // ?, reduce: FieldValue
			nil,        // :
			reduce(99), // ||, reduce: FieldValue
			reduce(99), // ^^, reduce: FieldValue
			reduce(99), // &&, reduce: FieldValue
			reduce(99), // ??, reduce: FieldValue
			reduce(99), // ???, reduce: FieldValue
			reduce(99), // =~, reduce: FieldValue
			reduce(99), // !=~, reduce: FieldValue
			reduce(99), // ==, reduce: FieldValue
			reduce(99), // !=, reduce: FieldValue
			reduce(99), // <=>, reduce: FieldValue
			reduce(99), // >=, reduce: FieldValue
			reduce(99), // <, reduce: FieldValue
			reduce(99), // <=, reduce: FieldValue
			reduce(99), // ^, reduce: FieldValue
			reduce(99), // &, reduce: FieldValue
			reduce(99), // <<, reduce: FieldValue
			reduce(99), // >>>, reduce: FieldValue
			reduce(99), // +, reduce: FieldValue
			reduce(99), // -, reduce: FieldValue
			reduce(99), // .+, reduce: FieldValue
			reduce(99), // .-, reduce: FieldValue
			reduce(99), // *, reduce: FieldValue
			reduce(99), // /, reduce: FieldValue
			reduce(99), // //, reduce: FieldValue
			reduce(99), // %, reduce: FieldValue
			reduce(99), // .*, reduce: FieldValue
			reduce(99), // ./, reduce: FieldValue
			reduce(99), // .//, reduce: FieldValue
			reduce(99), // ., reduce: FieldValue
			nil,        // !
			nil,        // ~
			reduce(99), // **, reduce: FieldValue
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(99), // [, reduce: FieldValue
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(99), // [[, reduce: FieldValue
			reduce(99), // [[[, reduce: FieldValue
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S623
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(101), // >, reduce: FieldValue
			reduce(101), // >>, reduce: FieldValue
			reduce(101), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(101), // ], reduce: FieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(101), // ?, reduce: FieldValue
			nil,         // :
			reduce(101), // ||, reduce: FieldValue
			reduce(101), // ^^, reduce: FieldValue
			reduce(101), // &&, reduce: FieldValue
			reduce(101), // ??, reduce: FieldValue
			reduce(101), // ???, reduce: FieldValue
			reduce(101), // =~, reduce: FieldValue
			reduce(101), // !=~, reduce: FieldValue
			reduce(101), // ==, reduce: FieldValue
			reduce(101), // !=, reduce: FieldValue
			reduce(101), // <=>, reduce: FieldValue
			reduce(101), // >=, reduce: FieldValue
			reduce(101), // <, reduce: FieldValue
			reduce(101), // <=, reduce: FieldValue
			reduce(101), // ^, reduce: FieldValue
			reduce(101), // &, reduce: FieldValue
			reduce(101), // <<, reduce: FieldValue
			reduce(101), // >>>, reduce: FieldValue
			reduce(101), // +, reduce: FieldValue
			reduce(101), // -, reduce: FieldValue
			reduce(101), // .+, reduce: FieldValue
			reduce(101), // .-, reduce: FieldValue
			reduce(101), // *, reduce: FieldValue
			reduce(101), // /, reduce: FieldValue
			reduce(101), // //, reduce: FieldValue
			reduce(101), // %, reduce: FieldValue
			reduce(101), // .*, reduce: FieldValue
			reduce(101), // ./, reduce: FieldValue
			reduce(101), // .//, reduce: FieldValue
			reduce(101), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(101), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(101), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(101), // [[, reduce: FieldValue
			reduce(101), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S624
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(100), // >, reduce: FieldValue
			reduce(100), // >>, reduce: FieldValue
			reduce(100), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(100), // ], reduce: FieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(100), // ?, reduce: FieldValue
			nil,         // :
			reduce(100), // ||, reduce: FieldValue
			reduce(100), // ^^, reduce: FieldValue
			reduce(100), // &&, reduce: FieldValue
			reduce(100), // ??, reduce: FieldValue
			reduce(100), // ???, reduce: FieldValue
			reduce(100), // =~, reduce: FieldValue
			reduce(100), // !=~, reduce: FieldValue
			reduce(100), // ==, reduce: FieldValue
			reduce(100), // !=, reduce: FieldValue
			reduce(100), // <=>, reduce: FieldValue
			reduce(100), // >=, reduce: FieldValue
			reduce(100), // <, reduce: FieldValue
			reduce(100), // <=, reduce: FieldValue
			reduce(100), // ^, reduce: FieldValue
			reduce(100), // &, reduce: FieldValue
			reduce(100), // <<, reduce: FieldValue
			reduce(100), // >>>, reduce: FieldValue
			reduce(100), // +, reduce: FieldValue
			reduce(100), // -, reduce: FieldValue
			reduce(100), // .+, reduce: FieldValue
			reduce(100), // .-, reduce: FieldValue
			reduce(100), // *, reduce: FieldValue
			reduce(100), // /, reduce: FieldValue
			reduce(100), // //, reduce: FieldValue
			reduce(100), // %, reduce: FieldValue
			reduce(100), // .*, reduce: FieldValue
			reduce(100), // ./, reduce: FieldValue
			reduce(100), // .//, reduce: FieldValue
			reduce(100), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(100), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(100), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(100), // [[, reduce: FieldValue
			reduce(100), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S625
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(213), // >, reduce: MlrvalOrFunction
			reduce(213), // >>, reduce: MlrvalOrFunction
			reduce(213), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(213), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(213), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(213), // ||, reduce: MlrvalOrFunction
			reduce(213), // ^^, reduce: MlrvalOrFunction
			reduce(213), // &&, reduce: MlrvalOrFunction
			reduce(213), // ??, reduce: MlrvalOrFunction
			reduce(213), // ???, reduce: MlrvalOrFunction
			reduce(213), // =~, reduce: MlrvalOrFunction
			reduce(213), // !=~, reduce: MlrvalOrFunction
			reduce(213), // ==, reduce: MlrvalOrFunction
			reduce(213), // !=, reduce: MlrvalOrFunction
			reduce(213), // <=>, reduce: MlrvalOrFunction
			reduce(213), // >=, reduce: MlrvalOrFunction
			reduce(213), // <, reduce: MlrvalOrFunction
			reduce(213), // <=, reduce: MlrvalOrFunction
			reduce(213), // ^, reduce: MlrvalOrFunction
			reduce(213), // &, reduce: MlrvalOrFunction
			reduce(213), // <<, reduce: MlrvalOrFunction
			reduce(213), // >>>, reduce: MlrvalOrFunction
			reduce(213), // +, reduce: MlrvalOrFunction
			reduce(213), // -, reduce: MlrvalOrFunction
			reduce(213), // .+, reduce: MlrvalOrFunction
			reduce(213), // .-, reduce: MlrvalOrFunction
			reduce(213), // *, reduce: MlrvalOrFunction
			reduce(213), // /, reduce: MlrvalOrFunction
			reduce(213), // //, reduce: MlrvalOrFunction
			reduce(213), // %, reduce: MlrvalOrFunction
			reduce(213), // .*, reduce: MlrvalOrFunction
			reduce(213), // ./, reduce: MlrvalOrFunction
			reduce(213), // .//, reduce: MlrvalOrFunction
			reduce(213), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(213), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(213), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(213), // [[, reduce: MlrvalOrFunction
			reduce(213), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S626
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(233), // >, reduce: MlrvalOrFunction
			reduce(233), // >>, reduce: MlrvalOrFunction
			reduce(233), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(233), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(233), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(233), // ||, reduce: MlrvalOrFunction
			reduce(233), // ^^, reduce: MlrvalOrFunction
			reduce(233), // &&, reduce: MlrvalOrFunction
			reduce(233), // ??, reduce: MlrvalOrFunction
			reduce(233), // ???, reduce: MlrvalOrFunction
			reduce(233), // =~, reduce: MlrvalOrFunction
			reduce(233), // !=~, reduce: MlrvalOrFunction
			reduce(233), // ==, reduce: MlrvalOrFunction
			reduce(233), // !=, reduce: MlrvalOrFunction
			reduce(233), // <=>, reduce: MlrvalOrFunction
			reduce(233), // >=, reduce: MlrvalOrFunction
			reduce(233), // <, reduce: MlrvalOrFunction
			reduce(233), // <=, reduce: MlrvalOrFunction
			reduce(233), // ^, reduce: MlrvalOrFunction
			reduce(233), // &, reduce: MlrvalOrFunction
			reduce(233), // <<, reduce: MlrvalOrFunction
			reduce(233), // >>>, reduce: MlrvalOrFunction
			reduce(233), // +, reduce: MlrvalOrFunction
			reduce(233), // -, reduce: MlrvalOrFunction
			reduce(233), // .+, reduce: MlrvalOrFunction
			reduce(233), // .-, reduce: MlrvalOrFunction
			reduce(233), // *, reduce: MlrvalOrFunction
			reduce(233), // /, reduce: MlrvalOrFunction
			reduce(233), // //, reduce: MlrvalOrFunction
			reduce(233), // %, reduce: MlrvalOrFunction
			reduce(233), // .*, reduce: MlrvalOrFunction
			reduce(233), // ./, reduce: MlrvalOrFunction
			reduce(233), // .//, reduce: MlrvalOrFunction
			reduce(233), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(233), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(233), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(233), // [[, reduce: MlrvalOrFunction
			reduce(233), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S627
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(210), // >, reduce: MlrvalOrFunction
			reduce(210), // >>, reduce: MlrvalOrFunction
			reduce(210), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(210), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(210), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(210), // ||, reduce: MlrvalOrFunction
			reduce(210), // ^^, reduce: MlrvalOrFunction
			reduce(210), // &&, reduce: MlrvalOrFunction
			reduce(210), // ??, reduce: MlrvalOrFunction
			reduce(210), // ???, reduce: MlrvalOrFunction
			reduce(210), // =~, reduce: MlrvalOrFunction
			reduce(210), // !=~, reduce: MlrvalOrFunction
			reduce(210), // ==, reduce: MlrvalOrFunction
			reduce(210), // !=, reduce: MlrvalOrFunction
			reduce(210), // <=>, reduce: MlrvalOrFunction
			reduce(210), // >=, reduce: MlrvalOrFunction
			reduce(210), // <, reduce: MlrvalOrFunction
			reduce(210), // <=, reduce: MlrvalOrFunction
			reduce(210), // ^, reduce: MlrvalOrFunction
			reduce(210), // &, reduce: MlrvalOrFunction
			reduce(210), // <<, reduce: MlrvalOrFunction
			reduce(210), // >>>, reduce: MlrvalOrFunction
			reduce(210), // +, reduce: MlrvalOrFunction
			reduce(210), // -, reduce: MlrvalOrFunction
			reduce(210), // .+, reduce: MlrvalOrFunction
			reduce(210), // .-, reduce: MlrvalOrFunction
			reduce(210), // *, reduce: MlrvalOrFunction
			reduce(210), // /, reduce: MlrvalOrFunction
			reduce(210), // //, reduce: MlrvalOrFunction
			reduce(210), // %, reduce: MlrvalOrFunction
			reduce(210), // .*, reduce: MlrvalOrFunction
			reduce(210), // ./, reduce: MlrvalOrFunction
			reduce(210), // .//, reduce: MlrvalOrFunction
			reduce(210), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(210), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(210), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(210), // [[, reduce: MlrvalOrFunction
			reduce(210), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S628
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(102), // >, reduce: FieldValue
			reduce(102), // >>, reduce: FieldValue
			reduce(102), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(102), // ], reduce: FieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(102), // ?, reduce: FieldValue
			nil,         // :
			reduce(102), // ||, reduce: FieldValue
			reduce(102), // ^^, reduce: FieldValue
			reduce(102), // &&, reduce: FieldValue
			reduce(102), // ??, reduce: FieldValue
			reduce(102), // ???, reduce: FieldValue
			reduce(102), // =~, reduce: FieldValue
			reduce(102), // !=~, reduce: FieldValue
			reduce(102), // ==, reduce: FieldValue
			reduce(102), // !=, reduce: FieldValue
			reduce(102), // <=>, reduce: FieldValue
			reduce(102), // >=, reduce: FieldValue
			reduce(102), // <, reduce: FieldValue
			reduce(102), // <=, reduce: FieldValue
			reduce(102), // ^, reduce: FieldValue
			reduce(102), // &, reduce: FieldValue
			reduce(102), // <<, reduce: FieldValue
			reduce(102), // >>>, reduce: FieldValue
			reduce(102), // +, reduce: FieldValue
			reduce(102), // -, reduce: FieldValue
			reduce(102), // .+, reduce: FieldValue
			reduce(102), // .-, reduce: FieldValue
			reduce(102), // *, reduce: FieldValue
			reduce(102), // /, reduce: FieldValue
			reduce(102), // //, reduce: FieldValue
			reduce(102), // %, reduce: FieldValue
			reduce(102), // .*, reduce: FieldValue
			reduce(102), // ./, reduce: FieldValue
			reduce(102), // .//, reduce: FieldValue
			reduce(102), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(102), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(102), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(102), // [[, reduce: FieldValue
			reduce(102), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S629
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(103), // >, reduce: FieldValue
			reduce(103), // >>, reduce: FieldValue
			reduce(103), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(103), // ], reduce: FieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(103), // ?, reduce: FieldValue
			nil,         // :
			reduce(103), // ||, reduce: FieldValue
			reduce(103), // ^^, reduce: FieldValue
			reduce(103), // &&, reduce: FieldValue
			reduce(103), // ??, reduce: FieldValue
			reduce(103), // ???, reduce: FieldValue
			reduce(103), // =~, reduce: FieldValue
			reduce(103), // !=~, reduce: FieldValue
			reduce(103), // ==, reduce: FieldValue
			reduce(103), // !=, reduce: FieldValue
			reduce(103), // <=>, reduce: FieldValue
			reduce(103), // >=, reduce: FieldValue
			reduce(103), // <, reduce: FieldValue
			reduce(103), // <=, reduce: FieldValue
			reduce(103), // ^, reduce: FieldValue
			reduce(103), // &, reduce: FieldValue
			reduce(103), // <<, reduce: FieldValue
			reduce(103), // >>>, reduce: FieldValue
			reduce(103), // +, reduce: FieldValue
			reduce(103), // -, reduce: FieldValue
			reduce(103), // .+, reduce: FieldValue
			reduce(103), // .-, reduce: FieldValue
			reduce(103), // *, reduce: FieldValue
			reduce(103), // /, reduce: FieldValue
			reduce(103), // //, reduce: FieldValue
			reduce(103), // %, reduce: FieldValue
			reduce(103), // .*, reduce: FieldValue
			reduce(103), // ./, reduce: FieldValue
			reduce(103), // .//, reduce: FieldValue
			reduce(103), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(103), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(103), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(103), // [[, reduce: FieldValue
			reduce(103), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S630
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(104), // >, reduce: DirectFieldValue
			reduce(104), // >>, reduce: DirectFieldValue
			reduce(104), // |, reduce: DirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(104), // ], reduce: DirectFieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(104), // ?, reduce: DirectFieldValue
			nil,         // :
			reduce(104), // ||, reduce: DirectFieldValue
			reduce(104), // ^^, reduce: DirectFieldValue
			reduce(104), // &&, reduce: DirectFieldValue
			reduce(104), // ??, reduce: DirectFieldValue
			reduce(104), // ???, reduce: DirectFieldValue
			reduce(104), // =~, reduce: DirectFieldValue
			reduce(104), // !=~, reduce: DirectFieldValue
			reduce(104), // ==, reduce: DirectFieldValue
			reduce(104), // !=, reduce: DirectFieldValue
			reduce(104), // <=>, reduce: DirectFieldValue
			reduce(104), // >=, reduce: DirectFieldValue
			reduce(104), // <, reduce: DirectFieldValue
			reduce(104), // <=, reduce: DirectFieldValue
			reduce(104), // ^, reduce: DirectFieldValue
			reduce(104), // &, reduce: DirectFieldValue
			reduce(104), // <<, reduce: DirectFieldValue
			reduce(104), // >>>, reduce: DirectFieldValue
			reduce(104), // +, reduce: DirectFieldValue
			reduce(104), // -, reduce: DirectFieldValue
			reduce(104), // .+, reduce: DirectFieldValue
			reduce(104), // .-, reduce: DirectFieldValue
			reduce(104), // *, reduce: DirectFieldValue
			reduce(104), // /, reduce: DirectFieldValue
			reduce(104), // //, reduce: DirectFieldValue
			reduce(104), // %, reduce: DirectFieldValue
			reduce(104), // .*, reduce: DirectFieldValue
			reduce(104), // ./, reduce: DirectFieldValue
			reduce(104), // .//, reduce: DirectFieldValue
			reduce(104), // ., reduce: DirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(104), // **, reduce: DirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(104), // [, reduce: DirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(104), // [[, reduce: DirectFieldValue
			reduce(104), // [[[, reduce: DirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S631
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S632
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(106), // >, reduce: BracedFieldValue
			reduce(106), // >>, reduce: BracedFieldValue
			reduce(106), // |, reduce: BracedFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(106), // ], reduce: BracedFieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(106), // ?, reduce: BracedFieldValue
			nil,         // :
			reduce(106), // ||, reduce: BracedFieldValue
			reduce(106), // ^^, reduce: BracedFieldValue
			reduce(106), // &&, reduce: BracedFieldValue
			reduce(106), // ??, reduce: BracedFieldValue
			reduce(106), // ???, reduce: BracedFieldValue
			reduce(106), // =~, reduce: BracedFieldValue
			reduce(106), // !=~, reduce: BracedFieldValue
			reduce(106), // ==, reduce: BracedFieldValue
			reduce(106), // !=, reduce: BracedFieldValue
			reduce(106), // <=>, reduce: BracedFieldValue
			reduce(106), // >=, reduce: BracedFieldValue
			reduce(106), // <, reduce: BracedFieldValue
			reduce(106), // <=, reduce: BracedFieldValue
			reduce(106), // ^, reduce: BracedFieldValue
			reduce(106), // &, reduce: BracedFieldValue
			reduce(106), // <<, reduce: BracedFieldValue
			reduce(106), // >>>, reduce: BracedFieldValue
			reduce(106), // +, reduce: BracedFieldValue
			reduce(106), // -, reduce: BracedFieldValue
			reduce(106), // .+, reduce: BracedFieldValue
			reduce(106), // .-, reduce: BracedFieldValue
			reduce(106), // *, reduce: BracedFieldValue
			reduce(106), // /, reduce: BracedFieldValue
			reduce(106), // //, reduce: BracedFieldValue
			reduce(106), // %, reduce: BracedFieldValue
			reduce(106), // .*, reduce: BracedFieldValue
			reduce(106), // ./, reduce: BracedFieldValue
			reduce(106), // .//, reduce: BracedFieldValue
			reduce(106), // ., reduce: BracedFieldValue
			nil,         // !
			nil,         // ~
			reduce(106), // **, reduce: BracedFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(106), // [, reduce: BracedFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(106), // [[, reduce: BracedFieldValue
			reduce(106), // [[[, reduce: BracedFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S633
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S634
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S635
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(109), // >, reduce: FullSrec
			reduce(109), // >>, reduce: FullSrec
			reduce(109), // |, reduce: FullSrec
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(109), // ], reduce: FullSrec
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(109), // ?, reduce: FullSrec
			nil,         // :
			reduce(109), // ||, reduce: FullSrec
			reduce(109), // ^^, reduce: FullSrec
			reduce(109), // &&, reduce: FullSrec
			reduce(109), // ??, reduce: FullSrec
			reduce(109), // ???, reduce: FullSrec
			reduce(109), // =~, reduce: FullSrec
			reduce(109), // !=~, reduce: FullSrec
			reduce(109), // ==, reduce: FullSrec
			reduce(109), // !=, reduce: FullSrec
			reduce(109), // <=>, reduce: FullSrec
			reduce(109), // >=, reduce: FullSrec
			reduce(109), // <, reduce: FullSrec
			reduce(109), // <=, reduce: FullSrec
			reduce(109), // ^, reduce: FullSrec
			reduce(109), // &, reduce: FullSrec
			reduce(109), // <<, reduce: FullSrec
			reduce(109), // >>>, reduce: FullSrec
			reduce(109), // +, reduce: FullSrec
			reduce(109), // -, reduce: FullSrec
			reduce(109), // .+, reduce: FullSrec
			reduce(109), // .-, reduce: FullSrec
			reduce(109), // *, reduce: FullSrec
			reduce(109), // /, reduce: FullSrec
			reduce(109), // //, reduce: FullSrec
			reduce(109), // %, reduce: FullSrec
			reduce(109), // .*, reduce: FullSrec
			reduce(109), // ./, reduce: FullSrec
			reduce(109), // .//, reduce: FullSrec
			reduce(109), // ., reduce: FullSrec
			nil,         // !
			nil,         // ~
			reduce(109), // **, reduce: FullSrec
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(109), // [, reduce: FullSrec
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(109), // [[, reduce: FullSrec
			reduce(109), // [[[, reduce: FullSrec
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S636
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(212), // >, reduce: MlrvalOrFunction
			reduce(212), // >>, reduce: MlrvalOrFunction
			reduce(212), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(212), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(212), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(212), // ||, reduce: MlrvalOrFunction
			reduce(212), // ^^, reduce: MlrvalOrFunction
			reduce(212), // &&, reduce: MlrvalOrFunction
			reduce(212), // ??, reduce: MlrvalOrFunction
			reduce(212), // ???, reduce: MlrvalOrFunction
			reduce(212), // =~, reduce: MlrvalOrFunction
			reduce(212), // !=~, reduce: MlrvalOrFunction
			reduce(212), // ==, reduce: MlrvalOrFunction
			reduce(212), // !=, reduce: MlrvalOrFunction
			reduce(212), // <=>, reduce: MlrvalOrFunction
			reduce(212), // >=, reduce: MlrvalOrFunction
			reduce(212), // <, reduce: MlrvalOrFunction
			reduce(212), // <=, reduce: MlrvalOrFunction
			reduce(212), // ^, reduce: MlrvalOrFunction
			reduce(212), // &, reduce: MlrvalOrFunction
			reduce(212), // <<, reduce: MlrvalOrFunction
			reduce(212), // >>>, reduce: MlrvalOrFunction
			reduce(212), // +, reduce: MlrvalOrFunction
			reduce(212), // -, reduce: MlrvalOrFunction
			reduce(212), // .+, reduce: MlrvalOrFunction
			reduce(212), // .-, reduce: MlrvalOrFunction
			reduce(212), // *, reduce: MlrvalOrFunction
			reduce(212), // /, reduce: MlrvalOrFunction
			reduce(212), // //, reduce: MlrvalOrFunction
			reduce(212), // %, reduce: MlrvalOrFunction
			reduce(212), // .*, reduce: MlrvalOrFunction
			reduce(212), // ./, reduce: MlrvalOrFunction
			reduce(212), // .//, reduce: MlrvalOrFunction
			reduce(212), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(212), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(212), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(212), // [[, reduce: MlrvalOrFunction
			reduce(212), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S637
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(113), // >, reduce: DirectOosvarValue
			reduce(113), // >>, reduce: DirectOosvarValue
			reduce(113), // |, reduce: DirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(113), // ], reduce: DirectOosvarValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(113), // ?, reduce: DirectOosvarValue
			nil,         // :
			reduce(113), // ||, reduce: DirectOosvarValue
			reduce(113), // ^^, reduce: DirectOosvarValue
			reduce(113), // &&, reduce: DirectOosvarValue
			reduce(113), // ??, reduce: DirectOosvarValue
			reduce(113), // ???, reduce: DirectOosvarValue
			reduce(113), // =~, reduce: DirectOosvarValue
			reduce(113), // !=~, reduce: DirectOosvarValue
			reduce(113), // ==, reduce: DirectOosvarValue
			reduce(113), // !=, reduce: DirectOosvarValue
			reduce(113), // <=>, reduce: DirectOosvarValue
			reduce(113), // >=, reduce: DirectOosvarValue
			reduce(113), // <, reduce: DirectOosvarValue
			reduce(113), // <=, reduce: DirectOosvarValue
			reduce(113), // ^, reduce: DirectOosvarValue
			reduce(113), // &, reduce: DirectOosvarValue
			reduce(113), // <<, reduce: DirectOosvarValue
			reduce(113), // >>>, reduce: DirectOosvarValue
			reduce(113), // +, reduce: DirectOosvarValue
			reduce(113), // -, reduce: DirectOosvarValue
			reduce(113), // .+, reduce: DirectOosvarValue
			reduce(113), // .-, reduce: DirectOosvarValue
			reduce(113), // *, reduce: DirectOosvarValue
			reduce(113), // /, reduce: DirectOosvarValue
			reduce(113), // //, reduce: DirectOosvarValue
			reduce(113), // %, reduce: DirectOosvarValue
			reduce(113), // .*, reduce: DirectOosvarValue
			reduce(113), // ./, reduce: DirectOosvarValue
			reduce(113), // .//, reduce: DirectOosvarValue
			reduce(113), // ., reduce: DirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(113), // **, reduce: DirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(113), // [, reduce: DirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(113), // [[, reduce: DirectOosvarValue
			reduce(113), // [[[, reduce: DirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S638
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S639
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(115), // >, reduce: BracedOosvarValue
			reduce(115), // >>, reduce: BracedOosvarValue
			reduce(115), // |, reduce: BracedOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(115), // ], reduce: BracedOosvarValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(115), // ?, reduce: BracedOosvarValue
			nil,         // :
			reduce(115), // ||, reduce: BracedOosvarValue
			reduce(115), // ^^, reduce: BracedOosvarValue
			reduce(115), // &&, reduce: BracedOosvarValue
			reduce(115), // ??, reduce: BracedOosvarValue
			reduce(115), // ???, reduce: BracedOosvarValue
			reduce(115), // =~, reduce: BracedOosvarValue
			reduce(115), // !=~, reduce: BracedOosvarValue
			reduce(115), // ==, reduce: BracedOosvarValue
			reduce(115), // !=, reduce: BracedOosvarValue
			reduce(115), // <=>, reduce: BracedOosvarValue
			reduce(115), // >=, reduce: BracedOosvarValue
			reduce(115), // <, reduce: BracedOosvarValue
			reduce(115), // <=, reduce: BracedOosvarValue
			reduce(115), // ^, reduce: BracedOosvarValue
			reduce(115), // &, reduce: BracedOosvarValue
			reduce(115), // <<, reduce: BracedOosvarValue
			reduce(115), // >>>, reduce: BracedOosvarValue
			reduce(115), // +, reduce: BracedOosvarValue
			reduce(115), // -, reduce: BracedOosvarValue
			reduce(115), // .+, reduce: BracedOosvarValue
			reduce(115), // .-, reduce: BracedOosvarValue
			reduce(115), // *, reduce: BracedOosvarValue
			reduce(115), // /, reduce: BracedOosvarValue
			reduce(115), // //, reduce: BracedOosvarValue
			reduce(115), // %, reduce: BracedOosvarValue
			reduce(115), // .*, reduce: BracedOosvarValue
			reduce(115), // ./, reduce: BracedOosvarValue
			reduce(115), // .//, reduce: BracedOosvarValue
			reduce(115), // ., reduce: BracedOosvarValue
			nil,         // !
			nil,         // ~
			reduce(115), // **, reduce: BracedOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(115), // [, reduce: BracedOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(115), // [[, reduce: BracedOosvarValue
			reduce(115), // [[[, reduce: BracedOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S640
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(116), // >, reduce: FullOosvar
			reduce(116), // >>, reduce: FullOosvar
			reduce(116), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(116), // ], reduce: FullOosvar
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(116), // ?, reduce: FullOosvar
			nil,         // :
			reduce(116), // ||, reduce: FullOosvar
			reduce(116), // ^^, reduce: FullOosvar
			reduce(116), // &&, reduce: FullOosvar
			reduce(116), // ??, reduce: FullOosvar
			reduce(116), // ???, reduce: FullOosvar
			reduce(116), // =~, reduce: FullOosvar
			reduce(116), // !=~, reduce: FullOosvar
			reduce(116), // ==, reduce: FullOosvar
			reduce(116), // !=, reduce: FullOosvar
			reduce(116), // <=>, reduce: FullOosvar
			reduce(116), // >=, reduce: FullOosvar
			reduce(116), // <, reduce: FullOosvar
			reduce(116), // <=, reduce: FullOosvar
			reduce(116), // ^, reduce: FullOosvar
			reduce(116), // &, reduce: FullOosvar
			reduce(116), // <<, reduce: FullOosvar
			reduce(116), // >>>, reduce: FullOosvar
			reduce(116), // +, reduce: FullOosvar
			reduce(116), // -, reduce: FullOosvar
			reduce(116), // .+, reduce: FullOosvar
			reduce(116), // .-, reduce: FullOosvar
			reduce(116), // *, reduce: FullOosvar
			reduce(116), // /, reduce: FullOosvar
			reduce(116), // //, reduce: FullOosvar
			reduce(116), // %, reduce: FullOosvar
			reduce(116), // .*, reduce: FullOosvar
			reduce(116), // ./, reduce: FullOosvar
			reduce(116), // .//, reduce: FullOosvar
			reduce(116), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(116), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(116), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(116), // [[, reduce: FullOosvar
			reduce(116), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S641
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(117), // >, reduce: FullOosvar
			reduce(117), // >>, reduce: FullOosvar
			reduce(117), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(117), // ], reduce: FullOosvar
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(117), // ?, reduce: FullOosvar
			nil,         // :
			reduce(117), // ||, reduce: FullOosvar
			reduce(117), // ^^, reduce: FullOosvar
			reduce(117), // &&, reduce: FullOosvar
			reduce(117), // ??, reduce: FullOosvar
			reduce(117), // ???, reduce: FullOosvar
			reduce(117), // =~, reduce: FullOosvar
			reduce(117), // !=~, reduce: FullOosvar
			reduce(117), // ==, reduce: FullOosvar
			reduce(117), // !=, reduce: FullOosvar
			reduce(117), // <=>, reduce: FullOosvar
			reduce(117), // >=, reduce: FullOosvar
			reduce(117), // <, reduce: FullOosvar
			reduce(117), // <=, reduce: FullOosvar
			reduce(117), // ^, reduce: FullOosvar
			reduce(117), // &, reduce: FullOosvar
			reduce(117), // <<, reduce: FullOosvar
			reduce(117), // >>>, reduce: FullOosvar
			reduce(117), // +, reduce: FullOosvar
			reduce(117), // -, reduce: FullOosvar
			reduce(117), // .+, reduce: FullOosvar
			reduce(117), // .-, reduce: FullOosvar
			reduce(117), // *, reduce: FullOosvar
			reduce(117), // /, reduce: FullOosvar
			reduce(117), // //, reduce: FullOosvar
			reduce(117), // %, reduce: FullOosvar
			reduce(117), // .*, reduce: FullOosvar
			reduce(117), // ./, reduce: FullOosvar
			reduce(117), // .//, reduce: FullOosvar
			reduce(117), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(117), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(117), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(117), // [[, reduce: FullOosvar
			reduce(117), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S642
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(118), // >, reduce: LocalVariable
			reduce(118), // >>, reduce: LocalVariable
			reduce(118), // |, reduce: LocalVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(270), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(118), // ], reduce: LocalVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(118), // ?, reduce: LocalVariable
			nil,         // :
			reduce(118), // ||, reduce: LocalVariable
			reduce(118), // ^^, reduce: LocalVariable
			reduce(118), // &&, reduce: LocalVariable
			reduce(118), // ??, reduce: LocalVariable
			reduce(118), // ???, reduce: LocalVariable
			reduce(118), // =~, reduce: LocalVariable
			reduce(118), // !=~, reduce: LocalVariable
			reduce(118), // ==, reduce: LocalVariable
			reduce(118), // !=, reduce: LocalVariable
			reduce(118), // <=>, reduce: LocalVariable
			reduce(118), // >=, reduce: LocalVariable
			reduce(118), // <, reduce: LocalVariable
			reduce(118), // <=, reduce: LocalVariable
			reduce(118), // ^, reduce: LocalVariable
			reduce(118), // &, reduce: LocalVariable
			reduce(118), // <<, reduce: LocalVariable
			reduce(118), // >>>, reduce: LocalVariable
			reduce(118), // +, reduce: LocalVariable
			reduce(118), // -, reduce: LocalVariable
			reduce(118), // .+, reduce: LocalVariable
			reduce(118), // .-, reduce: LocalVariable
			reduce(118), // *, reduce: LocalVariable
			reduce(118), // /, reduce: LocalVariable
			reduce(118), // //, reduce: LocalVariable
			reduce(118), // %, reduce: LocalVariable
			reduce(118), // .*, reduce: LocalVariable
			reduce(118), // ./, reduce: LocalVariable
			reduce(118), // .//, reduce: LocalVariable
			reduce(118), // ., reduce: LocalVariable
			nil,         // !
			nil,         // ~
			reduce(118), // **, reduce: LocalVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(118), // [, reduce: LocalVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(118), // [[, reduce: LocalVariable
			reduce(118), // [[[, reduce: LocalVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S643
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(147), // ], reduce: Rvalue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S644
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(148), // ], reduce: PrecedenceChainStart
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S645
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(150), // ], reduce: TernaryTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			shift(1421), // ?
			nil,         // :
			shift(1422), // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S646
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(152), // ], reduce: LogicalOrTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(152), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(152), // ||, reduce: LogicalOrTerm
			shift(1423), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S647
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(154), // ], reduce: LogicalXORTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(154), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(154), // ||, reduce: LogicalXORTerm
			reduce(154), // ^^, reduce: LogicalXORTerm
			shift(1424), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S648
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(156), // ], reduce: LogicalAndTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(156), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(156), // ||, reduce: LogicalAndTerm
			reduce(156), // ^^, reduce: LogicalAndTerm
			reduce(156), // &&, reduce: LogicalAndTerm
			shift(1425), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S649
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(158), // ], reduce: AbsentCoalesceTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(158), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(158), // ||, reduce: AbsentCoalesceTerm
			reduce(158), // ^^, reduce: AbsentCoalesceTerm
			reduce(158), // &&, reduce: AbsentCoalesceTerm
			reduce(158), // ??, reduce: AbsentCoalesceTerm
			shift(1426), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S650
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(160), // ], reduce: EmptyCoalesceTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(160), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(160), // ||, reduce: EmptyCoalesceTerm
			reduce(160), // ^^, reduce: EmptyCoalesceTerm
			reduce(160), // &&, reduce: EmptyCoalesceTerm
			reduce(160), // ??, reduce: EmptyCoalesceTerm
			reduce(160), // ???, reduce: EmptyCoalesceTerm
			shift(1427), // =~
			shift(1428), // !=~
			shift(1429), // ==
			shift(1430), // !=
			shift(1431), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S651
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1432), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(166), // ], reduce: EqneTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(166), // ?, reduce: EqneTerm
			nil,         // :
			reduce(166), // ||, reduce: EqneTerm
			reduce(166), // ^^, reduce: EqneTerm
			reduce(166), // &&, reduce: EqneTerm
			reduce(166), // ??, reduce: EqneTerm
			reduce(166), // ???, reduce: EqneTerm
			reduce(166), // =~, reduce: EqneTerm
			reduce(166), // !=~, reduce: EqneTerm
			reduce(166), // ==, reduce: EqneTerm
			reduce(166), // !=, reduce: EqneTerm
			reduce(166), // <=>, reduce: EqneTerm
			shift(1433), // >=
			shift(1434), // <
			shift(1435), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S652
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(171), // >, reduce: CmpTerm
			nil,         // >>
			shift(1436), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(171), // ], reduce: CmpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(171), // ?, reduce: CmpTerm
			nil,         // :
			reduce(171), // ||, reduce: CmpTerm
			reduce(171), // ^^, reduce: CmpTerm
			reduce(171), // &&, reduce: CmpTerm
			reduce(171), // ??, reduce: CmpTerm
			reduce(171), // ???, reduce: CmpTerm
			reduce(171), // =~, reduce: CmpTerm
			reduce(171), // !=~, reduce: CmpTerm
			reduce(171), // ==, reduce: CmpTerm
			reduce(171), // !=, reduce: CmpTerm
			reduce(171), // <=>, reduce: CmpTerm
			reduce(171), // >=, reduce: CmpTerm
			reduce(171), // <, reduce: CmpTerm
			reduce(171), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S653
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(173), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(173), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(173), // ], reduce: BitwiseORTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(173), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(173), // ||, reduce: BitwiseORTerm
			reduce(173), // ^^, reduce: BitwiseORTerm
			reduce(173), // &&, reduce: BitwiseORTerm
			reduce(173), // ??, reduce: BitwiseORTerm
			reduce(173), // ???, reduce: BitwiseORTerm
			reduce(173), // =~, reduce: BitwiseORTerm
			reduce(173), // !=~, reduce: BitwiseORTerm
			reduce(173), // ==, reduce: BitwiseORTerm
			reduce(173), // !=, reduce: BitwiseORTerm
			reduce(173), // <=>, reduce: BitwiseORTerm
			reduce(173), // >=, reduce: BitwiseORTerm
			reduce(173), // <, reduce: BitwiseORTerm
			reduce(173), // <=, reduce: BitwiseORTerm
			shift(1437), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S654
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(175), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(175), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(175), // ], reduce: BitwiseXORTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(175), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(175), // ||, reduce: BitwiseXORTerm
			reduce(175), // ^^, reduce: BitwiseXORTerm
			reduce(175), // &&, reduce: BitwiseXORTerm
			reduce(175), // ??, reduce: BitwiseXORTerm
			reduce(175), // ???, reduce: BitwiseXORTerm
			reduce(175), // =~, reduce: BitwiseXORTerm
			reduce(175), // !=~, reduce: BitwiseXORTerm
			reduce(175), // ==, reduce: BitwiseXORTerm
			reduce(175), // !=, reduce: BitwiseXORTerm
			reduce(175), // <=>, reduce: BitwiseXORTerm
			reduce(175), // >=, reduce: BitwiseXORTerm
			reduce(175), // <, reduce: BitwiseXORTerm
			reduce(175), // <=, reduce: BitwiseXORTerm
			reduce(175), // ^, reduce: BitwiseXORTerm
			shift(1438), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S655
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(177), // >, reduce: BitwiseANDTerm
			shift(1439), // >>
			reduce(177), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(177), // ], reduce: BitwiseANDTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(177), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(177), // ||, reduce: BitwiseANDTerm
			reduce(177), // ^^, reduce: BitwiseANDTerm
			reduce(177), // &&, reduce: BitwiseANDTerm
			reduce(177), // ??, reduce: BitwiseANDTerm
			reduce(177), // ???, reduce: BitwiseANDTerm
			reduce(177), // =~, reduce: BitwiseANDTerm
			reduce(177), // !=~, reduce: BitwiseANDTerm
			reduce(177), // ==, reduce: BitwiseANDTerm
			reduce(177), // !=, reduce: BitwiseANDTerm
			reduce(177), // <=>, reduce: BitwiseANDTerm
			reduce(177), // >=, reduce: BitwiseANDTerm
			reduce(177), // <, reduce: BitwiseANDTerm
			reduce(177), // <=, reduce: BitwiseANDTerm
			reduce(177), // ^, reduce: BitwiseANDTerm
			reduce(177), // &, reduce: BitwiseANDTerm
			shift(1440), // <<
			shift(1441), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S656
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(181), // >, reduce: BitwiseShiftTerm
			reduce(181), // >>, reduce: BitwiseShiftTerm
			reduce(181), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(181), // ], reduce: BitwiseShiftTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(181), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(181), // ||, reduce: BitwiseShiftTerm
			reduce(181), // ^^, reduce: BitwiseShiftTerm
			reduce(181), // &&, reduce: BitwiseShiftTerm
			reduce(181), // ??, reduce: BitwiseShiftTerm
			reduce(181), // ???, reduce: BitwiseShiftTerm
			reduce(181), // =~, reduce: BitwiseShiftTerm
			reduce(181), // !=~, reduce: BitwiseShiftTerm
			reduce(181), // ==, reduce: BitwiseShiftTerm
			reduce(181), // !=, reduce: BitwiseShiftTerm
			reduce(181), // <=>, reduce: BitwiseShiftTerm
			reduce(181), // >=, reduce: BitwiseShiftTerm
			reduce(181), // <, reduce: BitwiseShiftTerm
			reduce(181), // <=, reduce: BitwiseShiftTerm
			reduce(181), // ^, reduce: BitwiseShiftTerm
			reduce(181), // &, reduce: BitwiseShiftTerm
			reduce(181), // <<, reduce: BitwiseShiftTerm
			reduce(181), // >>>, reduce: BitwiseShiftTerm
			shift(1442), // +
			shift(1443), // -
			shift(1444), // .+
			shift(1445), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S657
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S658
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(186), // >, reduce: AddsubdotTerm
			reduce(186), // >>, reduce: AddsubdotTerm
			reduce(186), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(186), // ], reduce: AddsubdotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(186), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(186), // ||, reduce: AddsubdotTerm
			reduce(186), // ^^, reduce: AddsubdotTerm
			reduce(186), // &&, reduce: AddsubdotTerm
			reduce(186), // ??, reduce: AddsubdotTerm
			reduce(186), // ???, reduce: AddsubdotTerm
			reduce(186), // =~, reduce: AddsubdotTerm
			reduce(186), // !=~, reduce: AddsubdotTerm
			reduce(186), // ==, reduce: AddsubdotTerm
			reduce(186), // !=, reduce: AddsubdotTerm
			reduce(186), // <=>, reduce: AddsubdotTerm
			reduce(186), // >=, reduce: AddsubdotTerm
			reduce(186), // <, reduce: AddsubdotTerm
			reduce(186), // <=, reduce: AddsubdotTerm
			reduce(186), // ^, reduce: AddsubdotTerm
			reduce(186), // &, reduce: AddsubdotTerm
			reduce(186), // <<, reduce: AddsubdotTerm
			reduce(186), // >>>, reduce: AddsubdotTerm
			reduce(186), // +, reduce: AddsubdotTerm
			reduce(186), // -, reduce: AddsubdotTerm
			reduce(186), // .+, reduce: AddsubdotTerm
			reduce(186), // .-, reduce: AddsubdotTerm
			shift(1447), // *
			shift(1448), // /
			shift(1449), // //
			shift(1450), // %
			shift(1451), // .*
			shift(1452), // ./
			shift(1453), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S659
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S660
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S661
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S662
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(194), // >, reduce: MuldivTerm
			reduce(194), // >>, reduce: MuldivTerm
			reduce(194), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(194), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(194), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(194), // ||, reduce: MuldivTerm
			reduce(194), // ^^, reduce: MuldivTerm
			reduce(194), // &&, reduce: MuldivTerm
			reduce(194), // ??, reduce: MuldivTerm
			reduce(194), // ???, reduce: MuldivTerm
			reduce(194), // =~, reduce: MuldivTerm
			reduce(194), // !=~, reduce: MuldivTerm
			reduce(194), // ==, reduce: MuldivTerm
			reduce(194), // !=, reduce: MuldivTerm
			reduce(194), // <=>, reduce: MuldivTerm
			reduce(194), // >=, reduce: MuldivTerm
			reduce(194), // <, reduce: MuldivTerm
			reduce(194), // <=, reduce: MuldivTerm
			reduce(194), // ^, reduce: MuldivTerm
			reduce(194), // &, reduce: MuldivTerm
			reduce(194), // <<, reduce: MuldivTerm
			reduce(194), // >>>, reduce: MuldivTerm
			reduce(194), // +, reduce: MuldivTerm
			reduce(194), // -, reduce: MuldivTerm
			reduce(194), // .+, reduce: MuldivTerm
			reduce(194), // .-, reduce: MuldivTerm
			reduce(194), // *, reduce: MuldivTerm
			reduce(194), // /, reduce: MuldivTerm
			reduce(194), // //, reduce: MuldivTerm
			reduce(194), // %, reduce: MuldivTerm
			reduce(194), // .*, reduce: MuldivTerm
			reduce(194), // ./, reduce: MuldivTerm
			reduce(194), // .//, reduce: MuldivTerm
			shift(1457), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S663
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(196), // >, reduce: DotTerm
			reduce(196), // >>, reduce: DotTerm
			reduce(196), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(196), // ], reduce: DotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(196), // ?, reduce: DotTerm
			nil,         // :
			reduce(196), // ||, reduce: DotTerm
			reduce(196), // ^^, reduce: DotTerm
			reduce(196), // &&, reduce: DotTerm
			reduce(196), // ??, reduce: DotTerm
			reduce(196), // ???, reduce: DotTerm
			reduce(196), // =~, reduce: DotTerm
			reduce(196), // !=~, reduce: DotTerm
			reduce(196), // ==, reduce: DotTerm
			reduce(196), // !=, reduce: DotTerm
			reduce(196), // <=>, reduce: DotTerm
			reduce(196), // >=, reduce: DotTerm
			reduce(196), // <, reduce: DotTerm
			reduce(196), // <=, reduce: DotTerm
			reduce(196), // ^, reduce: DotTerm
			reduce(196), // &, reduce: DotTerm
			reduce(196), // <<, reduce: DotTerm
			reduce(196), // >>>, reduce: DotTerm
			reduce(196), // +, reduce: DotTerm
			reduce(196), // -, reduce: DotTerm
			reduce(196), // .+, reduce: DotTerm
			reduce(196), // .-, reduce: DotTerm
			reduce(196), // *, reduce: DotTerm
			reduce(196), // /, reduce: DotTerm
			reduce(196), // //, reduce: DotTerm
			reduce(196), // %, reduce: DotTerm
			reduce(196), // .*, reduce: DotTerm
			reduce(196), // ./, reduce: DotTerm
			reduce(196), // .//, reduce: DotTerm
			reduce(196), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S664
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S665
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S666
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(203), // >, reduce: UnaryOpTerm
			reduce(203), // >>, reduce: UnaryOpTerm
			reduce(203), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(203), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(203), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(203), // ||, reduce: UnaryOpTerm
			reduce(203), // ^^, reduce: UnaryOpTerm
			reduce(203), // &&, reduce: UnaryOpTerm
			reduce(203), // ??, reduce: UnaryOpTerm
			reduce(203), // ???, reduce: UnaryOpTerm
			reduce(203), // =~, reduce: UnaryOpTerm
			reduce(203), // !=~, reduce: UnaryOpTerm
			reduce(203), // ==, reduce: UnaryOpTerm
			reduce(203), // !=, reduce: UnaryOpTerm
			reduce(203), // <=>, reduce: UnaryOpTerm
			reduce(203), // >=, reduce: UnaryOpTerm
			reduce(203), // <, reduce: UnaryOpTerm
			reduce(203), // <=, reduce: UnaryOpTerm
			reduce(203), // ^, reduce: UnaryOpTerm
			reduce(203), // &, reduce: UnaryOpTerm
			reduce(203), // <<, reduce: UnaryOpTerm
			reduce(203), // >>>, reduce: UnaryOpTerm
			reduce(203), // +, reduce: UnaryOpTerm
			reduce(203), // -, reduce: UnaryOpTerm
			reduce(203), // .+, reduce: UnaryOpTerm
			reduce(203), // .-, reduce: UnaryOpTerm
			reduce(203), // *, reduce: UnaryOpTerm
			reduce(203), // /, reduce: UnaryOpTerm
			reduce(203), // //, reduce: UnaryOpTerm
			reduce(203), // %, reduce: UnaryOpTerm
			reduce(203), // .*, reduce: UnaryOpTerm
			reduce(203), // ./, reduce: UnaryOpTerm
			reduce(203), // .//, reduce: UnaryOpTerm
			reduce(203), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S667
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(207), // >, reduce: PowTerm
			reduce(207), // >>, reduce: PowTerm
			reduce(207), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(207), // ], reduce: PowTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(207), // ?, reduce: PowTerm
			nil,         // :
			reduce(207), // ||, reduce: PowTerm
			reduce(207), // ^^, reduce: PowTerm
			reduce(207), // &&, reduce: PowTerm
			reduce(207), // ??, reduce: PowTerm
			reduce(207), // ???, reduce: PowTerm
			reduce(207), // =~, reduce: PowTerm
			reduce(207), // !=~, reduce: PowTerm
			reduce(207), // ==, reduce: PowTerm
			reduce(207), // !=, reduce: PowTerm
			reduce(207), // <=>, reduce: PowTerm
			reduce(207), // >=, reduce: PowTerm
			reduce(207), // <, reduce: PowTerm
			reduce(207), // <=, reduce: PowTerm
			reduce(207), // ^, reduce: PowTerm
			reduce(207), // &, reduce: PowTerm
			reduce(207), // <<, reduce: PowTerm
			reduce(207), // >>>, reduce: PowTerm
			reduce(207), // +, reduce: PowTerm
			reduce(207), // -, reduce: PowTerm
			reduce(207), // .+, reduce: PowTerm
			reduce(207), // .-, reduce: PowTerm
			reduce(207), // *, reduce: PowTerm
			reduce(207), // /, reduce: PowTerm
			reduce(207), // //, reduce: PowTerm
			reduce(207), // %, reduce: PowTerm
			reduce(207), // .*, reduce: PowTerm
			reduce(207), // ./, reduce: PowTerm
			reduce(207), // .//, reduce: PowTerm
			reduce(207), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			shift(1460), // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S668
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(209), // >, reduce: PrecedenceChainEnd
			reduce(209), // >>, reduce: PrecedenceChainEnd
			reduce(209), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(209), // ], reduce: PrecedenceChainEnd
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(209), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(209), // ||, reduce: PrecedenceChainEnd
			reduce(209), // ^^, reduce: PrecedenceChainEnd
			reduce(209), // &&, reduce: PrecedenceChainEnd
			reduce(209), // ??, reduce: PrecedenceChainEnd
			reduce(209), // ???, reduce: PrecedenceChainEnd
			reduce(209), // =~, reduce: PrecedenceChainEnd
			reduce(209), // !=~, reduce: PrecedenceChainEnd
			reduce(209), // ==, reduce: PrecedenceChainEnd
			reduce(209), // !=, reduce: PrecedenceChainEnd
			reduce(209), // <=>, reduce: PrecedenceChainEnd
			reduce(209), // >=, reduce: PrecedenceChainEnd
			reduce(209), // <, reduce: PrecedenceChainEnd
			reduce(209), // <=, reduce: PrecedenceChainEnd
			reduce(209), // ^, reduce: PrecedenceChainEnd
			reduce(209), // &, reduce: PrecedenceChainEnd
			reduce(209), // <<, reduce: PrecedenceChainEnd
			reduce(209), // >>>, reduce: PrecedenceChainEnd
			reduce(209), // +, reduce: PrecedenceChainEnd
			reduce(209), // -, reduce: PrecedenceChainEnd
			reduce(209), // .+, reduce: PrecedenceChainEnd
			reduce(209), // .-, reduce: PrecedenceChainEnd
			reduce(209), // *, reduce: PrecedenceChainEnd
			reduce(209), // /, reduce: PrecedenceChainEnd
			reduce(209), // //, reduce: PrecedenceChainEnd
			reduce(209), // %, reduce: PrecedenceChainEnd
			reduce(209), // .*, reduce: PrecedenceChainEnd
			reduce(209), // ./, reduce: PrecedenceChainEnd
			reduce(209), // .//, reduce: PrecedenceChainEnd
			reduce(209), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(209), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(1461), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			shift(1462), // [[
			shift(1463), // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S669
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(215), // >, reduce: MlrvalOrFunction
			reduce(215), // >>, reduce: MlrvalOrFunction
			reduce(215), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(215), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(215), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(215), // ||, reduce: MlrvalOrFunction
			reduce(215), // ^^, reduce: MlrvalOrFunction
			reduce(215), // &&, reduce: MlrvalOrFunction
			reduce(215), // ??, reduce: MlrvalOrFunction
			reduce(215), // ???, reduce: MlrvalOrFunction
			reduce(215), // =~, reduce: MlrvalOrFunction
			reduce(215), // !=~, reduce: MlrvalOrFunction
			reduce(215), // ==, reduce: MlrvalOrFunction
			reduce(215), // !=, reduce: MlrvalOrFunction
			reduce(215), // <=>, reduce: MlrvalOrFunction
			reduce(215), // >=, reduce: MlrvalOrFunction
			reduce(215), // <, reduce: MlrvalOrFunction
			reduce(215), // <=, reduce: MlrvalOrFunction
			reduce(215), // ^, reduce: MlrvalOrFunction
			reduce(215), // &, reduce: MlrvalOrFunction
			reduce(215), // <<, reduce: MlrvalOrFunction
			reduce(215), // >>>, reduce: MlrvalOrFunction
			reduce(215), // +, reduce: MlrvalOrFunction
			reduce(215), // -, reduce: MlrvalOrFunction
			reduce(215), // .+, reduce: MlrvalOrFunction
			reduce(215), // .-, reduce: MlrvalOrFunction
			reduce(215), // *, reduce: MlrvalOrFunction
			reduce(215), // /, reduce: MlrvalOrFunction
			reduce(215), // //, reduce: MlrvalOrFunction
			reduce(215), // %, reduce: MlrvalOrFunction
			reduce(215), // .*, reduce: MlrvalOrFunction
			reduce(215), // ./, reduce: MlrvalOrFunction
			reduce(215), // .//, reduce: MlrvalOrFunction
			reduce(215), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(215), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(215), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(215), // [[, reduce: MlrvalOrFunction
			reduce(215), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S670
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(216), // >, reduce: MlrvalOrFunction
			reduce(216), // >>, reduce: MlrvalOrFunction
			reduce(216), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(216), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(216), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(216), // ||, reduce: MlrvalOrFunction
			reduce(216), // ^^, reduce: MlrvalOrFunction
			reduce(216), // &&, reduce: MlrvalOrFunction
			reduce(216), // ??, reduce: MlrvalOrFunction
			reduce(216), // ???, reduce: MlrvalOrFunction
			reduce(216), // =~, reduce: MlrvalOrFunction
			reduce(216), // !=~, reduce: MlrvalOrFunction
			reduce(216), // ==, reduce: MlrvalOrFunction
			reduce(216), // !=, reduce: MlrvalOrFunction
			reduce(216), // <=>, reduce: MlrvalOrFunction
			reduce(216), // >=, reduce: MlrvalOrFunction
			reduce(216), // <, reduce: MlrvalOrFunction
			reduce(216), // <=, reduce: MlrvalOrFunction
			reduce(216), // ^, reduce: MlrvalOrFunction
			reduce(216), // &, reduce: MlrvalOrFunction
			reduce(216), // <<, reduce: MlrvalOrFunction
			reduce(216), // >>>, reduce: MlrvalOrFunction
			reduce(216), // +, reduce: MlrvalOrFunction
			reduce(216), // -, reduce: MlrvalOrFunction
			reduce(216), // .+, reduce: MlrvalOrFunction
			reduce(216), // .-, reduce: MlrvalOrFunction
			reduce(216), // *, reduce: MlrvalOrFunction
			reduce(216), // /, reduce: MlrvalOrFunction
			reduce(216), // //, reduce: MlrvalOrFunction
			reduce(216), // %, reduce: MlrvalOrFunction
			reduce(216), // .*, reduce: MlrvalOrFunction
			reduce(216), // ./, reduce: MlrvalOrFunction
			reduce(216), // .//, reduce: MlrvalOrFunction
			reduce(216), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(216), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(216), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(216), // [[, reduce: MlrvalOrFunction
			reduce(216), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S671
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(217), // >, reduce: MlrvalOrFunction
			reduce(217), // >>, reduce: MlrvalOrFunction
			reduce(217), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(217), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(217), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(217), // ||, reduce: MlrvalOrFunction
			reduce(217), // ^^, reduce: MlrvalOrFunction
			reduce(217), // &&, reduce: MlrvalOrFunction
			reduce(217), // ??, reduce: MlrvalOrFunction
			reduce(217), // ???, reduce: MlrvalOrFunction
			reduce(217), // =~, reduce: MlrvalOrFunction
			reduce(217), // !=~, reduce: MlrvalOrFunction
			reduce(217), // ==, reduce: MlrvalOrFunction
			reduce(217), // !=, reduce: MlrvalOrFunction
			reduce(217), // <=>, reduce: MlrvalOrFunction
			reduce(217), // >=, reduce: MlrvalOrFunction
			reduce(217), // <, reduce: MlrvalOrFunction
			reduce(217), // <=, reduce: MlrvalOrFunction
			reduce(217), // ^, reduce: MlrvalOrFunction
			reduce(217), // &, reduce: MlrvalOrFunction
			reduce(217), // <<, reduce: MlrvalOrFunction
			reduce(217), // >>>, reduce: MlrvalOrFunction
			reduce(217), // +, reduce: MlrvalOrFunction
			reduce(217), // -, reduce: MlrvalOrFunction
			reduce(217), // .+, reduce: MlrvalOrFunction
			reduce(217), // .-, reduce: MlrvalOrFunction
			reduce(217), // *, reduce: MlrvalOrFunction
			reduce(217), // /, reduce: MlrvalOrFunction
			reduce(217), // //, reduce: MlrvalOrFunction
			reduce(217), // %, reduce: MlrvalOrFunction
			reduce(217), // .*, reduce: MlrvalOrFunction
			reduce(217), // ./, reduce: MlrvalOrFunction
			reduce(217), // .//, reduce: MlrvalOrFunction
			reduce(217), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(217), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(217), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(217), // [[, reduce: MlrvalOrFunction
			reduce(217), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S672
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(218), // >, reduce: MlrvalOrFunction
			reduce(218), // >>, reduce: MlrvalOrFunction
			reduce(218), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(218), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(218), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(218), // ||, reduce: MlrvalOrFunction
			reduce(218), // ^^, reduce: MlrvalOrFunction
			reduce(218), // &&, reduce: MlrvalOrFunction
			reduce(218), // ??, reduce: MlrvalOrFunction
			reduce(218), // ???, reduce: MlrvalOrFunction
			reduce(218), // =~, reduce: MlrvalOrFunction
			reduce(218), // !=~, reduce: MlrvalOrFunction
			reduce(218), // ==, reduce: MlrvalOrFunction
			reduce(218), // !=, reduce: MlrvalOrFunction
			reduce(218), // <=>, reduce: MlrvalOrFunction
			reduce(218), // >=, reduce: MlrvalOrFunction
			reduce(218), // <, reduce: MlrvalOrFunction
			reduce(218), // <=, reduce: MlrvalOrFunction
			reduce(218), // ^, reduce: MlrvalOrFunction
			reduce(218), // &, reduce: MlrvalOrFunction
			reduce(218), // <<, reduce: MlrvalOrFunction
			reduce(218), // >>>, reduce: MlrvalOrFunction
			reduce(218), // +, reduce: MlrvalOrFunction
			reduce(218), // -, reduce: MlrvalOrFunction
			reduce(218), // .+, reduce: MlrvalOrFunction
			reduce(218), // .-, reduce: MlrvalOrFunction
			reduce(218), // *, reduce: MlrvalOrFunction
			reduce(218), // /, reduce: MlrvalOrFunction
			reduce(218), // //, reduce: MlrvalOrFunction
			reduce(218), // %, reduce: MlrvalOrFunction
			reduce(218), // .*, reduce: MlrvalOrFunction
			reduce(218), // ./, reduce: MlrvalOrFunction
			reduce(218), // .//, reduce: MlrvalOrFunction
			reduce(218), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(218), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(218), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(218), // [[, reduce: MlrvalOrFunction
			reduce(218), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S673
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(219), // >, reduce: MlrvalOrFunction
			reduce(219), // >>, reduce: MlrvalOrFunction
			reduce(219), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(219), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(219), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(219), // ||, reduce: MlrvalOrFunction
			reduce(219), // ^^, reduce: MlrvalOrFunction
			reduce(219), // &&, reduce: MlrvalOrFunction
			reduce(219), // ??, reduce: MlrvalOrFunction
			reduce(219), // ???, reduce: MlrvalOrFunction
			reduce(219), // =~, reduce: MlrvalOrFunction
			reduce(219), // !=~, reduce: MlrvalOrFunction
			reduce(219), // ==, reduce: MlrvalOrFunction
			reduce(219), // !=, reduce: MlrvalOrFunction
			reduce(219), // <=>, reduce: MlrvalOrFunction
			reduce(219), // >=, reduce: MlrvalOrFunction
			reduce(219), // <, reduce: MlrvalOrFunction
			reduce(219), // <=, reduce: MlrvalOrFunction
			reduce(219), // ^, reduce: MlrvalOrFunction
			reduce(219), // &, reduce: MlrvalOrFunction
			reduce(219), // <<, reduce: MlrvalOrFunction
			reduce(219), // >>>, reduce: MlrvalOrFunction
			reduce(219), // +, reduce: MlrvalOrFunction
			reduce(219), // -, reduce: MlrvalOrFunction
			reduce(219), // .+, reduce: MlrvalOrFunction
			reduce(219), // .-, reduce: MlrvalOrFunction
			reduce(219), // *, reduce: MlrvalOrFunction
			reduce(219), // /, reduce: MlrvalOrFunction
			reduce(219), // //, reduce: MlrvalOrFunction
			reduce(219), // %, reduce: MlrvalOrFunction
			reduce(219), // .*, reduce: MlrvalOrFunction
			reduce(219), // ./, reduce: MlrvalOrFunction
			reduce(219), // .//, reduce: MlrvalOrFunction
			reduce(219), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(219), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(219), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(219), // [[, reduce: MlrvalOrFunction
			reduce(219), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S674
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(220), // >, reduce: MlrvalOrFunction
			reduce(220), // >>, reduce: MlrvalOrFunction
			reduce(220), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(220), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(220), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(220), // ||, reduce: MlrvalOrFunction
			reduce(220), // ^^, reduce: MlrvalOrFunction
			reduce(220), // &&, reduce: MlrvalOrFunction
			reduce(220), // ??, reduce: MlrvalOrFunction
			reduce(220), // ???, reduce: MlrvalOrFunction
			reduce(220), // =~, reduce: MlrvalOrFunction
			reduce(220), // !=~, reduce: MlrvalOrFunction
			reduce(220), // ==, reduce: MlrvalOrFunction
			reduce(220), // !=, reduce: MlrvalOrFunction
			reduce(220), // <=>, reduce: MlrvalOrFunction
			reduce(220), // >=, reduce: MlrvalOrFunction
			reduce(220), // <, reduce: MlrvalOrFunction
			reduce(220), // <=, reduce: MlrvalOrFunction
			reduce(220), // ^, reduce: MlrvalOrFunction
			reduce(220), // &, reduce: MlrvalOrFunction
			reduce(220), // <<, reduce: MlrvalOrFunction
			reduce(220), // >>>, reduce: MlrvalOrFunction
			reduce(220), // +, reduce: MlrvalOrFunction
			reduce(220), // -, reduce: MlrvalOrFunction
			reduce(220), // .+, reduce: MlrvalOrFunction
			reduce(220), // .-, reduce: MlrvalOrFunction
			reduce(220), // *, reduce: MlrvalOrFunction
			reduce(220), // /, reduce: MlrvalOrFunction
			reduce(220), // //, reduce: MlrvalOrFunction
			reduce(220), // %, reduce: MlrvalOrFunction
			reduce(220), // .*, reduce: MlrvalOrFunction
			reduce(220), // ./, reduce: MlrvalOrFunction
			reduce(220), // .//, reduce: MlrvalOrFunction
			reduce(220), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(220), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(220), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(220), // [[, reduce: MlrvalOrFunction
			reduce(220), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S675
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(221), // >, reduce: MlrvalOrFunction
			reduce(221), // >>, reduce: MlrvalOrFunction
			reduce(221), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(221), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(221), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(221), // ||, reduce: MlrvalOrFunction
			reduce(221), // ^^, reduce: MlrvalOrFunction
			reduce(221), // &&, reduce: MlrvalOrFunction
			reduce(221), // ??, reduce: MlrvalOrFunction
			reduce(221), // ???, reduce: MlrvalOrFunction
			reduce(221), // =~, reduce: MlrvalOrFunction
			reduce(221), // !=~, reduce: MlrvalOrFunction
			reduce(221), // ==, reduce: MlrvalOrFunction
			reduce(221), // !=, reduce: MlrvalOrFunction
			reduce(221), // <=>, reduce: MlrvalOrFunction
			reduce(221), // >=, reduce: MlrvalOrFunction
			reduce(221), // <, reduce: MlrvalOrFunction
			reduce(221), // <=, reduce: MlrvalOrFunction
			reduce(221), // ^, reduce: MlrvalOrFunction
			reduce(221), // &, reduce: MlrvalOrFunction
			reduce(221), // <<, reduce: MlrvalOrFunction
			reduce(221), // >>>, reduce: MlrvalOrFunction
			reduce(221), // +, reduce: MlrvalOrFunction
			reduce(221), // -, reduce: MlrvalOrFunction
			reduce(221), // .+, reduce: MlrvalOrFunction
			reduce(221), // .-, reduce: MlrvalOrFunction
			reduce(221), // *, reduce: MlrvalOrFunction
			reduce(221), // /, reduce: MlrvalOrFunction
			reduce(221), // //, reduce: MlrvalOrFunction
			reduce(221), // %, reduce: MlrvalOrFunction
			reduce(221), // .*, reduce: MlrvalOrFunction
			reduce(221), // ./, reduce: MlrvalOrFunction
			reduce(221), // .//, reduce: MlrvalOrFunction
			reduce(221), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(221), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(221), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(221), // [[, reduce: MlrvalOrFunction
			reduce(221), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S676
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(222), // >, reduce: MlrvalOrFunction
			reduce(222), // >>, reduce: MlrvalOrFunction
			reduce(222), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(222), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(222), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(222), // ||, reduce: MlrvalOrFunction
			reduce(222), // ^^, reduce: MlrvalOrFunction
			reduce(222), // &&, reduce: MlrvalOrFunction
			reduce(222), // ??, reduce: MlrvalOrFunction
			reduce(222), // ???, reduce: MlrvalOrFunction
			reduce(222), // =~, reduce: MlrvalOrFunction
			reduce(222), // !=~, reduce: MlrvalOrFunction
			reduce(222), // ==, reduce: MlrvalOrFunction
			reduce(222), // !=, reduce: MlrvalOrFunction
			reduce(222), // <=>, reduce: MlrvalOrFunction
			reduce(222), // >=, reduce: MlrvalOrFunction
			reduce(222), // <, reduce: MlrvalOrFunction
			reduce(222), // <=, reduce: MlrvalOrFunction
			reduce(222), // ^, reduce: MlrvalOrFunction
			reduce(222), // &, reduce: MlrvalOrFunction
			reduce(222), // <<, reduce: MlrvalOrFunction
			reduce(222), // >>>, reduce: MlrvalOrFunction
			reduce(222), // +, reduce: MlrvalOrFunction
			reduce(222), // -, reduce: MlrvalOrFunction
			reduce(222), // .+, reduce: MlrvalOrFunction
			reduce(222), // .-, reduce: MlrvalOrFunction
			reduce(222), // *, reduce: MlrvalOrFunction
			reduce(222), // /, reduce: MlrvalOrFunction
			reduce(222), // //, reduce: MlrvalOrFunction
			reduce(222), // %, reduce: MlrvalOrFunction
			reduce(222), // .*, reduce: MlrvalOrFunction
			reduce(222), // ./, reduce: MlrvalOrFunction
			reduce(222), // .//, reduce: MlrvalOrFunction
			reduce(222), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(222), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(222), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(222), // [[, reduce: MlrvalOrFunction
			reduce(222), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S677
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(223), // >, reduce: MlrvalOrFunction
			reduce(223), // >>, reduce: MlrvalOrFunction
			reduce(223), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(223), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(223), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(223), // ||, reduce: MlrvalOrFunction
			reduce(223), // ^^, reduce: MlrvalOrFunction
			reduce(223), // &&, reduce: MlrvalOrFunction
			reduce(223), // ??, reduce: MlrvalOrFunction
			reduce(223), // ???, reduce: MlrvalOrFunction
			reduce(223), // =~, reduce: MlrvalOrFunction
			reduce(223), // !=~, reduce: MlrvalOrFunction
			reduce(223), // ==, reduce: MlrvalOrFunction
			reduce(223), // !=, reduce: MlrvalOrFunction
			reduce(223), // <=>, reduce: MlrvalOrFunction
			reduce(223), // >=, reduce: MlrvalOrFunction
			reduce(223), // <, reduce: MlrvalOrFunction
			reduce(223), // <=, reduce: MlrvalOrFunction
			reduce(223), // ^, reduce: MlrvalOrFunction
			reduce(223), // &, reduce: MlrvalOrFunction
			reduce(223), // <<, reduce: MlrvalOrFunction
			reduce(223), // >>>, reduce: MlrvalOrFunction
			reduce(223), // +, reduce: MlrvalOrFunction
			reduce(223), // -, reduce: MlrvalOrFunction
			reduce(223), // .+, reduce: MlrvalOrFunction
			reduce(223), // .-, reduce: MlrvalOrFunction
			reduce(223), // *, reduce: MlrvalOrFunction
			reduce(223), // /, reduce: MlrvalOrFunction
			reduce(223), // //, reduce: MlrvalOrFunction
			reduce(223), // %, reduce: MlrvalOrFunction
			reduce(223), // .*, reduce: MlrvalOrFunction
			reduce(223), // ./, reduce: MlrvalOrFunction
			reduce(223), // .//, reduce: MlrvalOrFunction
			reduce(223), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(223), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(223), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(223), // [[, reduce: MlrvalOrFunction
			reduce(223), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S678
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(224), // >, reduce: MlrvalOrFunction
			reduce(224), // >>, reduce: MlrvalOrFunction
			reduce(224), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(224), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(224), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(224), // ||, reduce: MlrvalOrFunction
			reduce(224), // ^^, reduce: MlrvalOrFunction
			reduce(224), // &&, reduce: MlrvalOrFunction
			reduce(224), // ??, reduce: MlrvalOrFunction
			reduce(224), // ???, reduce: MlrvalOrFunction
			reduce(224), // =~, reduce: MlrvalOrFunction
			reduce(224), // !=~, reduce: MlrvalOrFunction
			reduce(224), // ==, reduce: MlrvalOrFunction
			reduce(224), // !=, reduce: MlrvalOrFunction
			reduce(224), // <=>, reduce: MlrvalOrFunction
			reduce(224), // >=, reduce: MlrvalOrFunction
			reduce(224), // <, reduce: MlrvalOrFunction
			reduce(224), // <=, reduce: MlrvalOrFunction
			reduce(224), // ^, reduce: MlrvalOrFunction
			reduce(224), // &, reduce: MlrvalOrFunction
			reduce(224), // <<, reduce: MlrvalOrFunction
			reduce(224), // >>>, reduce: MlrvalOrFunction
			reduce(224), // +, reduce: MlrvalOrFunction
			reduce(224), // -, reduce: MlrvalOrFunction
			reduce(224), // .+, reduce: MlrvalOrFunction
			reduce(224), // .-, reduce: MlrvalOrFunction
			reduce(224), // *, reduce: MlrvalOrFunction
			reduce(224), // /, reduce: MlrvalOrFunction
			reduce(224), // //, reduce: MlrvalOrFunction
			reduce(224), // %, reduce: MlrvalOrFunction
			reduce(224), // .*, reduce: MlrvalOrFunction
			reduce(224), // ./, reduce: MlrvalOrFunction
			reduce(224), // .//, reduce: MlrvalOrFunction
			reduce(224), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(224), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(224), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(224), // [[, reduce: MlrvalOrFunction
			reduce(224), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S679
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(225), // >, reduce: MlrvalOrFunction
			reduce(225), // >>, reduce: MlrvalOrFunction
			reduce(225), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(225), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(225), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(225), // ||, reduce: MlrvalOrFunction
			reduce(225), // ^^, reduce: MlrvalOrFunction
			reduce(225), // &&, reduce: MlrvalOrFunction
			reduce(225), // ??, reduce: MlrvalOrFunction
			reduce(225), // ???, reduce: MlrvalOrFunction
			reduce(225), // =~, reduce: MlrvalOrFunction
			reduce(225), // !=~, reduce: MlrvalOrFunction
			reduce(225), // ==, reduce: MlrvalOrFunction
			reduce(225), // !=, reduce: MlrvalOrFunction
			reduce(225), // <=>, reduce: MlrvalOrFunction
			reduce(225), // >=, reduce: MlrvalOrFunction
			reduce(225), // <, reduce: MlrvalOrFunction
			reduce(225), // <=, reduce: MlrvalOrFunction
			reduce(225), // ^, reduce: MlrvalOrFunction
			reduce(225), // &, reduce: MlrvalOrFunction
			reduce(225), // <<, reduce: MlrvalOrFunction
			reduce(225), // >>>, reduce: MlrvalOrFunction
			reduce(225), // +, reduce: MlrvalOrFunction
			reduce(225), // -, reduce: MlrvalOrFunction
			reduce(225), // .+, reduce: MlrvalOrFunction
			reduce(225), // .-, reduce: MlrvalOrFunction
			reduce(225), // *, reduce: MlrvalOrFunction
			reduce(225), // /, reduce: MlrvalOrFunction
			reduce(225), // //, reduce: MlrvalOrFunction
			reduce(225), // %, reduce: MlrvalOrFunction
			reduce(225), // .*, reduce: MlrvalOrFunction
			reduce(225), // ./, reduce: MlrvalOrFunction
			reduce(225), // .//, reduce: MlrvalOrFunction
			reduce(225), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(225), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(225), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(225), // [[, reduce: MlrvalOrFunction
			reduce(225), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S680
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(226), // >, reduce: MlrvalOrFunction
			reduce(226), // >>, reduce: MlrvalOrFunction
			reduce(226), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(226), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(226), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(226), // ||, reduce: MlrvalOrFunction
			reduce(226), // ^^, reduce: MlrvalOrFunction
			reduce(226), // &&, reduce: MlrvalOrFunction
			reduce(226), // ??, reduce: MlrvalOrFunction
			reduce(226), // ???, reduce: MlrvalOrFunction
			reduce(226), // =~, reduce: MlrvalOrFunction
			reduce(226), // !=~, reduce: MlrvalOrFunction
			reduce(226), // ==, reduce: MlrvalOrFunction
			reduce(226), // !=, reduce: MlrvalOrFunction
			reduce(226), // <=>, reduce: MlrvalOrFunction
			reduce(226), // >=, reduce: MlrvalOrFunction
			reduce(226), // <, reduce: MlrvalOrFunction
			reduce(226), // <=, reduce: MlrvalOrFunction
			reduce(226), // ^, reduce: MlrvalOrFunction
			reduce(226), // &, reduce: MlrvalOrFunction
			reduce(226), // <<, reduce: MlrvalOrFunction
			reduce(226), // >>>, reduce: MlrvalOrFunction
			reduce(226), // +, reduce: MlrvalOrFunction
			reduce(226), // -, reduce: MlrvalOrFunction
			reduce(226), // .+, reduce: MlrvalOrFunction
			reduce(226), // .-, reduce: MlrvalOrFunction
			reduce(226), // *, reduce: MlrvalOrFunction
			reduce(226), // /, reduce: MlrvalOrFunction
			reduce(226), // //, reduce: MlrvalOrFunction
			reduce(226), // %, reduce: MlrvalOrFunction
			reduce(226), // .*, reduce: MlrvalOrFunction
			reduce(226), // ./, reduce: MlrvalOrFunction
			reduce(226), // .//, reduce: MlrvalOrFunction
			reduce(226), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(226), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(226), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(226), // [[, reduce: MlrvalOrFunction
			reduce(226), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S681
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(227), // >, reduce: MlrvalOrFunction
			reduce(227), // >>, reduce: MlrvalOrFunction
			reduce(227), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(227), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(227), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(227), // ||, reduce: MlrvalOrFunction
			reduce(227), // ^^, reduce: MlrvalOrFunction
			reduce(227), // &&, reduce: MlrvalOrFunction
			reduce(227), // ??, reduce: MlrvalOrFunction
			reduce(227), // ???, reduce: MlrvalOrFunction
			reduce(227), // =~, reduce: MlrvalOrFunction
			reduce(227), // !=~, reduce: MlrvalOrFunction
			reduce(227), // ==, reduce: MlrvalOrFunction
			reduce(227), // !=, reduce: MlrvalOrFunction
			reduce(227), // <=>, reduce: MlrvalOrFunction
			reduce(227), // >=, reduce: MlrvalOrFunction
			reduce(227), // <, reduce: MlrvalOrFunction
			reduce(227), // <=, reduce: MlrvalOrFunction
			reduce(227), // ^, reduce: MlrvalOrFunction
			reduce(227), // &, reduce: MlrvalOrFunction
			reduce(227), // <<, reduce: MlrvalOrFunction
			reduce(227), // >>>, reduce: MlrvalOrFunction
			reduce(227), // +, reduce: MlrvalOrFunction
			reduce(227), // -, reduce: MlrvalOrFunction
			reduce(227), // .+, reduce: MlrvalOrFunction
			reduce(227), // .-, reduce: MlrvalOrFunction
			reduce(227), // *, reduce: MlrvalOrFunction
			reduce(227), // /, reduce: MlrvalOrFunction
			reduce(227), // //, reduce: MlrvalOrFunction
			reduce(227), // %, reduce: MlrvalOrFunction
			reduce(227), // .*, reduce: MlrvalOrFunction
			reduce(227), // ./, reduce: MlrvalOrFunction
			reduce(227), // .//, reduce: MlrvalOrFunction
			reduce(227), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(227), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(227), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(227), // [[, reduce: MlrvalOrFunction
			reduce(227), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S682
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(752),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(756),  // (
			nil,         // )
			nil,         // emitp
			shift(768),  // field_name
			shift(769),  // $[
			shift(1464), // ]
			shift(771),  // braced_field_name
			shift(772),  // $[[
			shift(773),  // $[[[
			shift(774),  // full_srec
			shift(776),  // oosvar_name
			shift(777),  // @[
			shift(778),  // braced_oosvar_name
			shift(779),  // full_oosvar
			shift(780),  // all
			shift(781),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(796),  // +
			shift(798),  // -
			shift(799),  // .+
			shift(800),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(803),  // !
			shift(804),  // ~
			nil,         // **
			shift(809),  // string_literal
			shift(810),  // regex_case_insensitive
			shift(811),  // int_literal
			shift(812),  // float_literal
			shift(813),  // boolean_literal
			shift(814),  // null_literal
			shift(815),  // inf_literal
			shift(816),  // nan_literal
			shift(817),  // const_M_PI
			shift(818),  // const_M_E
			shift(819),  // panic
			shift(821),  // [
			shift(824),  // ctx_IPS
			shift(825),  // ctx_IFS
			shift(826),  // ctx_IRS
			shift(827),  // ctx_OPS
			shift(828),  // ctx_OFS
			shift(829),  // ctx_ORS
			shift(830),  // ctx_FLATSEP
			shift(831),  // ctx_NF
			shift(832),  // ctx_NR
			shift(833),  // ctx_FNR
			shift(834),  // ctx_FILENAME
			shift(835),  // ctx_FILENUM
			shift(837),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(844),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S683
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(240), // >, reduce: MlrvalOrFunction
			reduce(240), // >>, reduce: MlrvalOrFunction
			reduce(240), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(240), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(240), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(240), // ||, reduce: MlrvalOrFunction
			reduce(240), // ^^, reduce: MlrvalOrFunction
			reduce(240), // &&, reduce: MlrvalOrFunction
			reduce(240), // ??, reduce: MlrvalOrFunction
			reduce(240), // ???, reduce: MlrvalOrFunction
			reduce(240), // =~, reduce: MlrvalOrFunction
			reduce(240), // !=~, reduce: MlrvalOrFunction
			reduce(240), // ==, reduce: MlrvalOrFunction
			reduce(240), // !=, reduce: MlrvalOrFunction
			reduce(240), // <=>, reduce: MlrvalOrFunction
			reduce(240), // >=, reduce: MlrvalOrFunction
			reduce(240), // <, reduce: MlrvalOrFunction
			reduce(240), // <=, reduce: MlrvalOrFunction
			reduce(240), // ^, reduce: MlrvalOrFunction
			reduce(240), // &, reduce: MlrvalOrFunction
			reduce(240), // <<, reduce: MlrvalOrFunction
			reduce(240), // >>>, reduce: MlrvalOrFunction
			reduce(240), // +, reduce: MlrvalOrFunction
			reduce(240), // -, reduce: MlrvalOrFunction
			reduce(240), // .+, reduce: MlrvalOrFunction
			reduce(240), // .-, reduce: MlrvalOrFunction
			reduce(240), // *, reduce: MlrvalOrFunction
			reduce(240), // /, reduce: MlrvalOrFunction
			reduce(240), // //, reduce: MlrvalOrFunction
			reduce(240), // %, reduce: MlrvalOrFunction
			reduce(240), // .*, reduce: MlrvalOrFunction
			reduce(240), // ./, reduce: MlrvalOrFunction
			reduce(240), // .//, reduce: MlrvalOrFunction
			reduce(240), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(240), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(240), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(240), // [[, reduce: MlrvalOrFunction
			reduce(240), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S684
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(241), // >, reduce: ContextVariable
			reduce(241), // >>, reduce: ContextVariable
			reduce(241), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(241), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(241), // ?, reduce: ContextVariable
			nil,         // :
			reduce(241), // ||, reduce: ContextVariable
			reduce(241), // ^^, reduce: ContextVariable
			reduce(241), // &&, reduce: ContextVariable
			reduce(241), // ??, reduce: ContextVariable
			reduce(241), // ???, reduce: ContextVariable
			reduce(241), // =~, reduce: ContextVariable
			reduce(241), // !=~, reduce: ContextVariable
			reduce(241), // ==, reduce: ContextVariable
			reduce(241), // !=, reduce: ContextVariable
			reduce(241), // <=>, reduce: ContextVariable
			reduce(241), // >=, reduce: ContextVariable
			reduce(241), // <, reduce: ContextVariable
			reduce(241), // <=, reduce: ContextVariable
			reduce(241), // ^, reduce: ContextVariable
			reduce(241), // &, reduce: ContextVariable
			reduce(241), // <<, reduce: ContextVariable
			reduce(241), // >>>, reduce: ContextVariable
			reduce(241), // +, reduce: ContextVariable
			reduce(241), // -, reduce: ContextVariable
			reduce(241), // .+, reduce: ContextVariable
			reduce(241), // .-, reduce: ContextVariable
			reduce(241), // *, reduce: ContextVariable
			reduce(241), // /, reduce: ContextVariable
			reduce(241), // //, reduce: ContextVariable
			reduce(241), // %, reduce: ContextVariable
			reduce(241), // .*, reduce: ContextVariable
			reduce(241), // ./, reduce: ContextVariable
			reduce(241), // .//, reduce: ContextVariable
			reduce(241), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(241), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(241), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(241), // [[, reduce: ContextVariable
			reduce(241), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S685
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(242), // >, reduce: ContextVariable
			reduce(242), // >>, reduce: ContextVariable
			reduce(242), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(242), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(242), // ?, reduce: ContextVariable
			nil,         // :
			reduce(242), // ||, reduce: ContextVariable
			reduce(242), // ^^, reduce: ContextVariable
			reduce(242), // &&, reduce: ContextVariable
			reduce(242), // ??, reduce: ContextVariable
			reduce(242), // ???, reduce: ContextVariable
			reduce(242), // =~, reduce: ContextVariable
			reduce(242), // !=~, reduce: ContextVariable
			reduce(242), // ==, reduce: ContextVariable
			reduce(242), // !=, reduce: ContextVariable
			reduce(242), // <=>, reduce: ContextVariable
			reduce(242), // >=, reduce: ContextVariable
			reduce(242), // <, reduce: ContextVariable
			reduce(242), // <=, reduce: ContextVariable
			reduce(242), // ^, reduce: ContextVariable
			reduce(242), // &, reduce: ContextVariable
			reduce(242), // <<, reduce: ContextVariable
			reduce(242), // >>>, reduce: ContextVariable
			reduce(242), // +, reduce: ContextVariable
			reduce(242), // -, reduce: ContextVariable
			reduce(242), // .+, reduce: ContextVariable
			reduce(242), // .-, reduce: ContextVariable
			reduce(242), // *, reduce: ContextVariable
			reduce(242), // /, reduce: ContextVariable
			reduce(242), // //, reduce: ContextVariable
			reduce(242), // %, reduce: ContextVariable
			reduce(242), // .*, reduce: ContextVariable
			reduce(242), // ./, reduce: ContextVariable
			reduce(242), // .//, reduce: ContextVariable
			reduce(242), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(242), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(242), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(242), // [[, reduce: ContextVariable
			reduce(242), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S686
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(243), // >, reduce: ContextVariable
			reduce(243), // >>, reduce: ContextVariable
			reduce(243), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(243), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(243), // ?, reduce: ContextVariable
			nil,         // :
			reduce(243), // ||, reduce: ContextVariable
			reduce(243), // ^^, reduce: ContextVariable
			reduce(243), // &&, reduce: ContextVariable
			reduce(243), // ??, reduce: ContextVariable
			reduce(243), // ???, reduce: ContextVariable
			reduce(243), // =~, reduce: ContextVariable
			reduce(243), // !=~, reduce: ContextVariable
			reduce(243), // ==, reduce: ContextVariable
			reduce(243), // !=, reduce: ContextVariable
			reduce(243), // <=>, reduce: ContextVariable
			reduce(243), // >=, reduce: ContextVariable
			reduce(243), // <, reduce: ContextVariable
			reduce(243), // <=, reduce: ContextVariable
			reduce(243), // ^, reduce: ContextVariable
			reduce(243), // &, reduce: ContextVariable
			reduce(243), // <<, reduce: ContextVariable
			reduce(243), // >>>, reduce: ContextVariable
			reduce(243), // +, reduce: ContextVariable
			reduce(243), // -, reduce: ContextVariable
			reduce(243), // .+, reduce: ContextVariable
			reduce(243), // .-, reduce: ContextVariable
			reduce(243), // *, reduce: ContextVariable
			reduce(243), // /, reduce: ContextVariable
			reduce(243), // //, reduce: ContextVariable
			reduce(243), // %, reduce: ContextVariable
			reduce(243), // .*, reduce: ContextVariable
			reduce(243), // ./, reduce: ContextVariable
			reduce(243), // .//, reduce: ContextVariable
			reduce(243), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(243), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(243), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(243), // [[, reduce: ContextVariable
			reduce(243), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S687
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(244), // >, reduce: ContextVariable
			reduce(244), // >>, reduce: ContextVariable
			reduce(244), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(244), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(244), // ?, reduce: ContextVariable
			nil,         // :
			reduce(244), // ||, reduce: ContextVariable
			reduce(244), // ^^, reduce: ContextVariable
			reduce(244), // &&, reduce: ContextVariable
			reduce(244), // ??, reduce: ContextVariable
			reduce(244), // ???, reduce: ContextVariable
			reduce(244), // =~, reduce: ContextVariable
			reduce(244), // !=~, reduce: ContextVariable
			reduce(244), // ==, reduce: ContextVariable
			reduce(244), // !=, reduce: ContextVariable
			reduce(244), // <=>, reduce: ContextVariable
			reduce(244), // >=, reduce: ContextVariable
			reduce(244), // <, reduce: ContextVariable
			reduce(244), // <=, reduce: ContextVariable
			reduce(244), // ^, reduce: ContextVariable
			reduce(244), // &, reduce: ContextVariable
			reduce(244), // <<, reduce: ContextVariable
			reduce(244), // >>>, reduce: ContextVariable
			reduce(244), // +, reduce: ContextVariable
			reduce(244), // -, reduce: ContextVariable
			reduce(244), // .+, reduce: ContextVariable
			reduce(244), // .-, reduce: ContextVariable
			reduce(244), // *, reduce: ContextVariable
			reduce(244), // /, reduce: ContextVariable
			reduce(244), // //, reduce: ContextVariable
			reduce(244), // %, reduce: ContextVariable
			reduce(244), // .*, reduce: ContextVariable
			reduce(244), // ./, reduce: ContextVariable
			reduce(244), // .//, reduce: ContextVariable
			reduce(244), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(244), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(244), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(244), // [[, reduce: ContextVariable
			reduce(244), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S688
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(245), // >, reduce: ContextVariable
			reduce(245), // >>, reduce: ContextVariable
			reduce(245), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(245), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(245), // ?, reduce: ContextVariable
			nil,         // :
			reduce(245), // ||, reduce: ContextVariable
			reduce(245), // ^^, reduce: ContextVariable
			reduce(245), // &&, reduce: ContextVariable
			reduce(245), // ??, reduce: ContextVariable
			reduce(245), // ???, reduce: ContextVariable
			reduce(245), // =~, reduce: ContextVariable
			reduce(245), // !=~, reduce: ContextVariable
			reduce(245), // ==, reduce: ContextVariable
			reduce(245), // !=, reduce: ContextVariable
			reduce(245), // <=>, reduce: ContextVariable
			reduce(245), // >=, reduce: ContextVariable
			reduce(245), // <, reduce: ContextVariable
			reduce(245), // <=, reduce: ContextVariable
			reduce(245), // ^, reduce: ContextVariable
			reduce(245), // &, reduce: ContextVariable
			reduce(245), // <<, reduce: ContextVariable
			reduce(245), // >>>, reduce: ContextVariable
			reduce(245), // +, reduce: ContextVariable
			reduce(245), // -, reduce: ContextVariable
			reduce(245), // .+, reduce: ContextVariable
			reduce(245), // .-, reduce: ContextVariable
			reduce(245), // *, reduce: ContextVariable
			reduce(245), // /, reduce: ContextVariable
			reduce(245), // //, reduce: ContextVariable
			reduce(245), // %, reduce: ContextVariable
			reduce(245), // .*, reduce: ContextVariable
			reduce(245), // ./, reduce: ContextVariable
			reduce(245), // .//, reduce: ContextVariable
			reduce(245), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(245), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(245), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(245), // [[, reduce: ContextVariable
			reduce(245), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S689
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(246), // >, reduce: ContextVariable
			reduce(246), // >>, reduce: ContextVariable
			reduce(246), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(246), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(246), // ?, reduce: ContextVariable
			nil,         // :
			reduce(246), // ||, reduce: ContextVariable
			reduce(246), // ^^, reduce: ContextVariable
			reduce(246), // &&, reduce: ContextVariable
			reduce(246), // ??, reduce: ContextVariable
			reduce(246), // ???, reduce: ContextVariable
			reduce(246), // =~, reduce: ContextVariable
			reduce(246), // !=~, reduce: ContextVariable
			reduce(246), // ==, reduce: ContextVariable
			reduce(246), // !=, reduce: ContextVariable
			reduce(246), // <=>, reduce: ContextVariable
			reduce(246), // >=, reduce: ContextVariable
			reduce(246), // <, reduce: ContextVariable
			reduce(246), // <=, reduce: ContextVariable
			reduce(246), // ^, reduce: ContextVariable
			reduce(246), // &, reduce: ContextVariable
			reduce(246), // <<, reduce: ContextVariable
			reduce(246), // >>>, reduce: ContextVariable
			reduce(246), // +, reduce: ContextVariable
			reduce(246), // -, reduce: ContextVariable
			reduce(246), // .+, reduce: ContextVariable
			reduce(246), // .-, reduce: ContextVariable
			reduce(246), // *, reduce: ContextVariable
			reduce(246), // /, reduce: ContextVariable
			reduce(246), // //, reduce: ContextVariable
			reduce(246), // %, reduce: ContextVariable
			reduce(246), // .*, reduce: ContextVariable
			reduce(246), // ./, reduce: ContextVariable
			reduce(246), // .//, reduce: ContextVariable
			reduce(246), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(246), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(246), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(246), // [[, reduce: ContextVariable
			reduce(246), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S690
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(247), // >, reduce: ContextVariable
			reduce(247), // >>, reduce: ContextVariable
			reduce(247), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(247), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(247), // ?, reduce: ContextVariable
			nil,         // :
			reduce(247), // ||, reduce: ContextVariable
			reduce(247), // ^^, reduce: ContextVariable
			reduce(247), // &&, reduce: ContextVariable
			reduce(247), // ??, reduce: ContextVariable
			reduce(247), // ???, reduce: ContextVariable
			reduce(247), // =~, reduce: ContextVariable
			reduce(247), // !=~, reduce: ContextVariable
			reduce(247), // ==, reduce: ContextVariable
			reduce(247), // !=, reduce: ContextVariable
			reduce(247), // <=>, reduce: ContextVariable
			reduce(247), // >=, reduce: ContextVariable
			reduce(247), // <, reduce: ContextVariable
			reduce(247), // <=, reduce: ContextVariable
			reduce(247), // ^, reduce: ContextVariable
			reduce(247), // &, reduce: ContextVariable
			reduce(247), // <<, reduce: ContextVariable
			reduce(247), // >>>, reduce: ContextVariable
			reduce(247), // +, reduce: ContextVariable
			reduce(247), // -, reduce: ContextVariable
			reduce(247), // .+, reduce: ContextVariable
			reduce(247), // .-, reduce: ContextVariable
			reduce(247), // *, reduce: ContextVariable
			reduce(247), // /, reduce: ContextVariable
			reduce(247), // //, reduce: ContextVariable
			reduce(247), // %, reduce: ContextVariable
			reduce(247), // .*, reduce: ContextVariable
			reduce(247), // ./, reduce: ContextVariable
			reduce(247), // .//, reduce: ContextVariable
			reduce(247), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(247), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(247), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(247), // [[, reduce: ContextVariable
			reduce(247), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S691
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(248), // >, reduce: ContextVariable
			reduce(248), // >>, reduce: ContextVariable
			reduce(248), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(248), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(248), // ?, reduce: ContextVariable
			nil,         // :
			reduce(248), // ||, reduce: ContextVariable
			reduce(248), // ^^, reduce: ContextVariable
			reduce(248), // &&, reduce: ContextVariable
			reduce(248), // ??, reduce: ContextVariable
			reduce(248), // ???, reduce: ContextVariable
			reduce(248), // =~, reduce: ContextVariable
			reduce(248), // !=~, reduce: ContextVariable
			reduce(248), // ==, reduce: ContextVariable
			reduce(248), // !=, reduce: ContextVariable
			reduce(248), // <=>, reduce: ContextVariable
			reduce(248), // >=, reduce: ContextVariable
			reduce(248), // <, reduce: ContextVariable
			reduce(248), // <=, reduce: ContextVariable
			reduce(248), // ^, reduce: ContextVariable
			reduce(248), // &, reduce: ContextVariable
			reduce(248), // <<, reduce: ContextVariable
			reduce(248), // >>>, reduce: ContextVariable
			reduce(248), // +, reduce: ContextVariable
			reduce(248), // -, reduce: ContextVariable
			reduce(248), // .+, reduce: ContextVariable
			reduce(248), // .-, reduce: ContextVariable
			reduce(248), // *, reduce: ContextVariable
			reduce(248), // /, reduce: ContextVariable
			reduce(248), // //, reduce: ContextVariable
			reduce(248), // %, reduce: ContextVariable
			reduce(248), // .*, reduce: ContextVariable
			reduce(248), // ./, reduce: ContextVariable
			reduce(248), // .//, reduce: ContextVariable
			reduce(248), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(248), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(248), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(248), // [[, reduce: ContextVariable
			reduce(248), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S692
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(249), // >, reduce: ContextVariable
			reduce(249), // >>, reduce: ContextVariable
			reduce(249), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(249), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(249), // ?, reduce: ContextVariable
			nil,         // :
			reduce(249), // ||, reduce: ContextVariable
			reduce(249), // ^^, reduce: ContextVariable
			reduce(249), // &&, reduce: ContextVariable
			reduce(249), // ??, reduce: ContextVariable
			reduce(249), // ???, reduce: ContextVariable
			reduce(249), // =~, reduce: ContextVariable
			reduce(249), // !=~, reduce: ContextVariable
			reduce(249), // ==, reduce: ContextVariable
			reduce(249), // !=, reduce: ContextVariable
			reduce(249), // <=>, reduce: ContextVariable
			reduce(249), // >=, reduce: ContextVariable
			reduce(249), // <, reduce: ContextVariable
			reduce(249), // <=, reduce: ContextVariable
			reduce(249), // ^, reduce: ContextVariable
			reduce(249), // &, reduce: ContextVariable
			reduce(249), // <<, reduce: ContextVariable
			reduce(249), // >>>, reduce: ContextVariable
			reduce(249), // +, reduce: ContextVariable
			reduce(249), // -, reduce: ContextVariable
			reduce(249), // .+, reduce: ContextVariable
			reduce(249), // .-, reduce: ContextVariable
			reduce(249), // *, reduce: ContextVariable
			reduce(249), // /, reduce: ContextVariable
			reduce(249), // //, reduce: ContextVariable
			reduce(249), // %, reduce: ContextVariable
			reduce(249), // .*, reduce: ContextVariable
			reduce(249), // ./, reduce: ContextVariable
			reduce(249), // .//, reduce: ContextVariable
			reduce(249), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(249), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(249), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(249), // [[, reduce: ContextVariable
			reduce(249), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S693
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(250), // >, reduce: ContextVariable
			reduce(250), // >>, reduce: ContextVariable
			reduce(250), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(250), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(250), // ?, reduce: ContextVariable
			nil,         // :
			reduce(250), // ||, reduce: ContextVariable
			reduce(250), // ^^, reduce: ContextVariable
			reduce(250), // &&, reduce: ContextVariable
			reduce(250), // ??, reduce: ContextVariable
			reduce(250), // ???, reduce: ContextVariable
			reduce(250), // =~, reduce: ContextVariable
			reduce(250), // !=~, reduce: ContextVariable
			reduce(250), // ==, reduce: ContextVariable
			reduce(250), // !=, reduce: ContextVariable
			reduce(250), // <=>, reduce: ContextVariable
			reduce(250), // >=, reduce: ContextVariable
			reduce(250), // <, reduce: ContextVariable
			reduce(250), // <=, reduce: ContextVariable
			reduce(250), // ^, reduce: ContextVariable
			reduce(250), // &, reduce: ContextVariable
			reduce(250), // <<, reduce: ContextVariable
			reduce(250), // >>>, reduce: ContextVariable
			reduce(250), // +, reduce: ContextVariable
			reduce(250), // -, reduce: ContextVariable
			reduce(250), // .+, reduce: ContextVariable
			reduce(250), // .-, reduce: ContextVariable
			reduce(250), // *, reduce: ContextVariable
			reduce(250), // /, reduce: ContextVariable
			reduce(250), // //, reduce: ContextVariable
			reduce(250), // %, reduce: ContextVariable
			reduce(250), // .*, reduce: ContextVariable
			reduce(250), // ./, reduce: ContextVariable
			reduce(250), // .//, reduce: ContextVariable
			reduce(250), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(250), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(250), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(250), // [[, reduce: ContextVariable
			reduce(250), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S694
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(251), // >, reduce: ContextVariable
			reduce(251), // >>, reduce: ContextVariable
			reduce(251), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(251), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(251), // ?, reduce: ContextVariable
			nil,         // :
			reduce(251), // ||, reduce: ContextVariable
			reduce(251), // ^^, reduce: ContextVariable
			reduce(251), // &&, reduce: ContextVariable
			reduce(251), // ??, reduce: ContextVariable
			reduce(251), // ???, reduce: ContextVariable
			reduce(251), // =~, reduce: ContextVariable
			reduce(251), // !=~, reduce: ContextVariable
			reduce(251), // ==, reduce: ContextVariable
			reduce(251), // !=, reduce: ContextVariable
			reduce(251), // <=>, reduce: ContextVariable
			reduce(251), // >=, reduce: ContextVariable
			reduce(251), // <, reduce: ContextVariable
			reduce(251), // <=, reduce: ContextVariable
			reduce(251), // ^, reduce: ContextVariable
			reduce(251), // &, reduce: ContextVariable
			reduce(251), // <<, reduce: ContextVariable
			reduce(251), // >>>, reduce: ContextVariable
			reduce(251), // +, reduce: ContextVariable
			reduce(251), // -, reduce: ContextVariable
			reduce(251), // .+, reduce: ContextVariable
			reduce(251), // .-, reduce: ContextVariable
			reduce(251), // *, reduce: ContextVariable
			reduce(251), // /, reduce: ContextVariable
			reduce(251), // //, reduce: ContextVariable
			reduce(251), // %, reduce: ContextVariable
			reduce(251), // .*, reduce: ContextVariable
			reduce(251), // ./, reduce: ContextVariable
			reduce(251), // .//, reduce: ContextVariable
			reduce(251), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(251), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(251), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(251), // [[, reduce: ContextVariable
			reduce(251), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S695
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(252), // >, reduce: ContextVariable
			reduce(252), // >>, reduce: ContextVariable
			reduce(252), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(252), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(252), // ?, reduce: ContextVariable
			nil,         // :
			reduce(252), // ||, reduce: ContextVariable
			reduce(252), // ^^, reduce: ContextVariable
			reduce(252), // &&, reduce: ContextVariable
			reduce(252), // ??, reduce: ContextVariable
			reduce(252), // ???, reduce: ContextVariable
			reduce(252), // =~, reduce: ContextVariable
			reduce(252), // !=~, reduce: ContextVariable
			reduce(252), // ==, reduce: ContextVariable
			reduce(252), // !=, reduce: ContextVariable
			reduce(252), // <=>, reduce: ContextVariable
			reduce(252), // >=, reduce: ContextVariable
			reduce(252), // <, reduce: ContextVariable
			reduce(252), // <=, reduce: ContextVariable
			reduce(252), // ^, reduce: ContextVariable
			reduce(252), // &, reduce: ContextVariable
			reduce(252), // <<, reduce: ContextVariable
			reduce(252), // >>>, reduce: ContextVariable
			reduce(252), // +, reduce: ContextVariable
			reduce(252), // -, reduce: ContextVariable
			reduce(252), // .+, reduce: ContextVariable
			reduce(252), // .-, reduce: ContextVariable
			reduce(252), // *, reduce: ContextVariable
			reduce(252), // /, reduce: ContextVariable
			reduce(252), // //, reduce: ContextVariable
			reduce(252), // %, reduce: ContextVariable
			reduce(252), // .*, reduce: ContextVariable
			reduce(252), // ./, reduce: ContextVariable
			reduce(252), // .//, reduce: ContextVariable
			reduce(252), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(252), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(252), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(252), // [[, reduce: ContextVariable
			reduce(252), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S696
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(253), // >, reduce: MlrvalOrFunction
			reduce(253), // >>, reduce: MlrvalOrFunction
			reduce(253), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(253), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(253), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(253), // ||, reduce: MlrvalOrFunction
			reduce(253), // ^^, reduce: MlrvalOrFunction
			reduce(253), // &&, reduce: MlrvalOrFunction
			reduce(253), // ??, reduce: MlrvalOrFunction
			reduce(253), // ???, reduce: MlrvalOrFunction
			reduce(253), // =~, reduce: MlrvalOrFunction
			reduce(253), // !=~, reduce: MlrvalOrFunction
			reduce(253), // ==, reduce: MlrvalOrFunction
			reduce(253), // !=, reduce: MlrvalOrFunction
			reduce(253), // <=>, reduce: MlrvalOrFunction
			reduce(253), // >=, reduce: MlrvalOrFunction
			reduce(253), // <, reduce: MlrvalOrFunction
			reduce(253), // <=, reduce: MlrvalOrFunction
			reduce(253), // ^, reduce: MlrvalOrFunction
			reduce(253), // &, reduce: MlrvalOrFunction
			reduce(253), // <<, reduce: MlrvalOrFunction
			reduce(253), // >>>, reduce: MlrvalOrFunction
			reduce(253), // +, reduce: MlrvalOrFunction
			reduce(253), // -, reduce: MlrvalOrFunction
			reduce(253), // .+, reduce: MlrvalOrFunction
			reduce(253), // .-, reduce: MlrvalOrFunction
			reduce(253), // *, reduce: MlrvalOrFunction
			reduce(253), // /, reduce: MlrvalOrFunction
			reduce(253), // //, reduce: MlrvalOrFunction
			reduce(253), // %, reduce: MlrvalOrFunction
			reduce(253), // .*, reduce: MlrvalOrFunction
			reduce(253), // ./, reduce: MlrvalOrFunction
			reduce(253), // .//, reduce: MlrvalOrFunction
			reduce(253), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(253), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(253), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(253), // [[, reduce: MlrvalOrFunction
			reduce(253), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S697
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			shift(1466), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(1467), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S698
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(256), // >, reduce: MlrvalOrFunction
			reduce(256), // >>, reduce: MlrvalOrFunction
			reduce(256), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(256), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(256), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(256), // ||, reduce: MlrvalOrFunction
			reduce(256), // ^^, reduce: MlrvalOrFunction
			reduce(256), // &&, reduce: MlrvalOrFunction
			reduce(256), // ??, reduce: MlrvalOrFunction
			reduce(256), // ???, reduce: MlrvalOrFunction
			reduce(256), // =~, reduce: MlrvalOrFunction
			reduce(256), // !=~, reduce: MlrvalOrFunction
			reduce(256), // ==, reduce: MlrvalOrFunction
			reduce(256), // !=, reduce: MlrvalOrFunction
			reduce(256), // <=>, reduce: MlrvalOrFunction
			reduce(256), // >=, reduce: MlrvalOrFunction
			reduce(256), // <, reduce: MlrvalOrFunction
			reduce(256), // <=, reduce: MlrvalOrFunction
			reduce(256), // ^, reduce: MlrvalOrFunction
			reduce(256), // &, reduce: MlrvalOrFunction
			reduce(256), // <<, reduce: MlrvalOrFunction
			reduce(256), // >>>, reduce: MlrvalOrFunction
			reduce(256), // +, reduce: MlrvalOrFunction
			reduce(256), // -, reduce: MlrvalOrFunction
			reduce(256), // .+, reduce: MlrvalOrFunction
			reduce(256), // .-, reduce: MlrvalOrFunction
			reduce(256), // *, reduce: MlrvalOrFunction
			reduce(256), // /, reduce: MlrvalOrFunction
			reduce(256), // //, reduce: MlrvalOrFunction
			reduce(256), // %, reduce: MlrvalOrFunction
			reduce(256), // .*, reduce: MlrvalOrFunction
			reduce(256), // ./, reduce: MlrvalOrFunction
			reduce(256), // .//, reduce: MlrvalOrFunction
			reduce(256), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(256), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(256), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(256), // [[, reduce: MlrvalOrFunction
			reduce(256), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S699
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(257), // >, reduce: MlrvalOrFunction
			reduce(257), // >>, reduce: MlrvalOrFunction
			reduce(257), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(257), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(257), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(257), // ||, reduce: MlrvalOrFunction
			reduce(257), // ^^, reduce: MlrvalOrFunction
			reduce(257), // &&, reduce: MlrvalOrFunction
			reduce(257), // ??, reduce: MlrvalOrFunction
			reduce(257), // ???, reduce: MlrvalOrFunction
			reduce(257), // =~, reduce: MlrvalOrFunction
			reduce(257), // !=~, reduce: MlrvalOrFunction
			reduce(257), // ==, reduce: MlrvalOrFunction
			reduce(257), // !=, reduce: MlrvalOrFunction
			reduce(257), // <=>, reduce: MlrvalOrFunction
			reduce(257), // >=, reduce: MlrvalOrFunction
			reduce(257), // <, reduce: MlrvalOrFunction
			reduce(257), // <=, reduce: MlrvalOrFunction
			reduce(257), // ^, reduce: MlrvalOrFunction
			reduce(257), // &, reduce: MlrvalOrFunction
			reduce(257), // <<, reduce: MlrvalOrFunction
			reduce(257), // >>>, reduce: MlrvalOrFunction
			reduce(257), // +, reduce: MlrvalOrFunction
			reduce(257), // -, reduce: MlrvalOrFunction
			reduce(257), // .+, reduce: MlrvalOrFunction
			reduce(257), // .-, reduce: MlrvalOrFunction
			reduce(257), // *, reduce: MlrvalOrFunction
			reduce(257), // /, reduce: MlrvalOrFunction
			reduce(257), // //, reduce: MlrvalOrFunction
			reduce(257), // %, reduce: MlrvalOrFunction
			reduce(257), // .*, reduce: MlrvalOrFunction
			reduce(257), // ./, reduce: MlrvalOrFunction
			reduce(257), // .//, reduce: MlrvalOrFunction
			reduce(257), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(257), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(257), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(257), // [[, reduce: MlrvalOrFunction
			reduce(257), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S700
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(258), // >, reduce: MlrvalOrFunction
			reduce(258), // >>, reduce: MlrvalOrFunction
			reduce(258), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(258), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(258), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(258), // ||, reduce: MlrvalOrFunction
			reduce(258), // ^^, reduce: MlrvalOrFunction
			reduce(258), // &&, reduce: MlrvalOrFunction
			reduce(258), // ??, reduce: MlrvalOrFunction
			reduce(258), // ???, reduce: MlrvalOrFunction
			reduce(258), // =~, reduce: MlrvalOrFunction
			reduce(258), // !=~, reduce: MlrvalOrFunction
			reduce(258), // ==, reduce: MlrvalOrFunction
			reduce(258), // !=, reduce: MlrvalOrFunction
			reduce(258), // <=>, reduce: MlrvalOrFunction
			reduce(258), // >=, reduce: MlrvalOrFunction
			reduce(258), // <, reduce: MlrvalOrFunction
			reduce(258), // <=, reduce: MlrvalOrFunction
			reduce(258), // ^, reduce: MlrvalOrFunction
			reduce(258), // &, reduce: MlrvalOrFunction
			reduce(258), // <<, reduce: MlrvalOrFunction
			reduce(258), // >>>, reduce: MlrvalOrFunction
			reduce(258), // +, reduce: MlrvalOrFunction
			reduce(258), // -, reduce: MlrvalOrFunction
			reduce(258), // .+, reduce: MlrvalOrFunction
			reduce(258), // .-, reduce: MlrvalOrFunction
			reduce(258), // *, reduce: MlrvalOrFunction
			reduce(258), // /, reduce: MlrvalOrFunction
			reduce(258), // //, reduce: MlrvalOrFunction
			reduce(258), // %, reduce: MlrvalOrFunction
			reduce(258), // .*, reduce: MlrvalOrFunction
			reduce(258), // ./, reduce: MlrvalOrFunction
			reduce(258), // .//, reduce: MlrvalOrFunction
			reduce(258), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(258), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(258), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(258), // [[, reduce: MlrvalOrFunction
			reduce(258), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S701
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(259), // >, reduce: MlrvalOrFunction
			reduce(259), // >>, reduce: MlrvalOrFunction
			reduce(259), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(259), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(259), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(259), // ||, reduce: MlrvalOrFunction
			reduce(259), // ^^, reduce: MlrvalOrFunction
			reduce(259), // &&, reduce: MlrvalOrFunction
			reduce(259), // ??, reduce: MlrvalOrFunction
			reduce(259), // ???, reduce: MlrvalOrFunction
			reduce(259), // =~, reduce: MlrvalOrFunction
			reduce(259), // !=~, reduce: MlrvalOrFunction
			reduce(259), // ==, reduce: MlrvalOrFunction
			reduce(259), // !=, reduce: MlrvalOrFunction
			reduce(259), // <=>, reduce: MlrvalOrFunction
			reduce(259), // >=, reduce: MlrvalOrFunction
			reduce(259), // <, reduce: MlrvalOrFunction
			reduce(259), // <=, reduce: MlrvalOrFunction
			reduce(259), // ^, reduce: MlrvalOrFunction
			reduce(259), // &, reduce: MlrvalOrFunction
			reduce(259), // <<, reduce: MlrvalOrFunction
			reduce(259), // >>>, reduce: MlrvalOrFunction
			reduce(259), // +, reduce: MlrvalOrFunction
			reduce(259), // -, reduce: MlrvalOrFunction
			reduce(259), // .+, reduce: MlrvalOrFunction
			reduce(259), // .-, reduce: MlrvalOrFunction
			reduce(259), // *, reduce: MlrvalOrFunction
			reduce(259), // /, reduce: MlrvalOrFunction
			reduce(259), // //, reduce: MlrvalOrFunction
			reduce(259), // %, reduce: MlrvalOrFunction
			reduce(259), // .*, reduce: MlrvalOrFunction
			reduce(259), // ./, reduce: MlrvalOrFunction
			reduce(259), // .//, reduce: MlrvalOrFunction
			reduce(259), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(259), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(259), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(259), // [[, reduce: MlrvalOrFunction
			reduce(259), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S702
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(267), // >, reduce: MlrvalOrFunction
			reduce(267), // >>, reduce: MlrvalOrFunction
			reduce(267), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(267), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(267), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(267), // ||, reduce: MlrvalOrFunction
			reduce(267), // ^^, reduce: MlrvalOrFunction
			reduce(267), // &&, reduce: MlrvalOrFunction
			reduce(267), // ??, reduce: MlrvalOrFunction
			reduce(267), // ???, reduce: MlrvalOrFunction
			reduce(267), // =~, reduce: MlrvalOrFunction
			reduce(267), // !=~, reduce: MlrvalOrFunction
			reduce(267), // ==, reduce: MlrvalOrFunction
			reduce(267), // !=, reduce: MlrvalOrFunction
			reduce(267), // <=>, reduce: MlrvalOrFunction
			reduce(267), // >=, reduce: MlrvalOrFunction
			reduce(267), // <, reduce: MlrvalOrFunction
			reduce(267), // <=, reduce: MlrvalOrFunction
			reduce(267), // ^, reduce: MlrvalOrFunction
			reduce(267), // &, reduce: MlrvalOrFunction
			reduce(267), // <<, reduce: MlrvalOrFunction
			reduce(267), // >>>, reduce: MlrvalOrFunction
			reduce(267), // +, reduce: MlrvalOrFunction
			reduce(267), // -, reduce: MlrvalOrFunction
			reduce(267), // .+, reduce: MlrvalOrFunction
			reduce(267), // .-, reduce: MlrvalOrFunction
			reduce(267), // *, reduce: MlrvalOrFunction
			reduce(267), // /, reduce: MlrvalOrFunction
			reduce(267), // //, reduce: MlrvalOrFunction
			reduce(267), // %, reduce: MlrvalOrFunction
			reduce(267), // .*, reduce: MlrvalOrFunction
			reduce(267), // ./, reduce: MlrvalOrFunction
			reduce(267), // .//, reduce: MlrvalOrFunction
			reduce(267), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(267), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(267), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(267), // [[, reduce: MlrvalOrFunction
			reduce(267), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S703
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1468), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S704
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1469), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S705
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(1470), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S706
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(1471), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S707
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(1472), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S708
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S709
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S710
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S711
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S712
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S713
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S714
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S715
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S716
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S717
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S718
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S719
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S720
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S721
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S722
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S723
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S724
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S725
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S726
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S727
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S728
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S729
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S730
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S731
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S732
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S733
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(197), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(197), // ;, reduce: UnaryOpTerm
			reduce(197), // {, reduce: UnaryOpTerm
			nil,         // }
			reduce(197), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(197), // >, reduce: UnaryOpTerm
			reduce(197), // >>, reduce: UnaryOpTerm
			reduce(197), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(197), // ||=, reduce: UnaryOpTerm
			reduce(197), // ^^=, reduce: UnaryOpTerm
			reduce(197), // &&=, reduce: UnaryOpTerm
			reduce(197), // ??=, reduce: UnaryOpTerm
			reduce(197), // ???=, reduce: UnaryOpTerm
			reduce(197), // |=, reduce: UnaryOpTerm
			reduce(197), // &=, reduce: UnaryOpTerm
			reduce(197), // ^=, reduce: UnaryOpTerm
			reduce(197), // <<=, reduce: UnaryOpTerm
			reduce(197), // >>=, reduce: UnaryOpTerm
			reduce(197), // >>>=, reduce: UnaryOpTerm
			reduce(197), // +=, reduce: UnaryOpTerm
			reduce(197), // .=, reduce: UnaryOpTerm
			reduce(197), // -=, reduce: UnaryOpTerm
			reduce(197), // *=, reduce: UnaryOpTerm
			reduce(197), // /=, reduce: UnaryOpTerm
			reduce(197), // //=, reduce: UnaryOpTerm
			reduce(197), // %=, reduce: UnaryOpTerm
			reduce(197), // **=, reduce: UnaryOpTerm
			reduce(197), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(197), // ||, reduce: UnaryOpTerm
			reduce(197), // ^^, reduce: UnaryOpTerm
			reduce(197), // &&, reduce: UnaryOpTerm
			reduce(197), // ??, reduce: UnaryOpTerm
			reduce(197), // ???, reduce: UnaryOpTerm
			reduce(197), // =~, reduce: UnaryOpTerm
			reduce(197), // !=~, reduce: UnaryOpTerm
			reduce(197), // ==, reduce: UnaryOpTerm
			reduce(197), // !=, reduce: UnaryOpTerm
			reduce(197), // <=>, reduce: UnaryOpTerm
			reduce(197), // >=, reduce: UnaryOpTerm
			reduce(197), // <, reduce: UnaryOpTerm
			reduce(197), // <=, reduce: UnaryOpTerm
			reduce(197), // ^, reduce: UnaryOpTerm
			reduce(197), // &, reduce: UnaryOpTerm
			reduce(197), // <<, reduce: UnaryOpTerm
			reduce(197), // >>>, reduce: UnaryOpTerm
			reduce(197), // +, reduce: UnaryOpTerm
			reduce(197), // -, reduce: UnaryOpTerm
			reduce(197), // .+, reduce: UnaryOpTerm
			reduce(197), // .-, reduce: UnaryOpTerm
			reduce(197), // *, reduce: UnaryOpTerm
			reduce(197), // /, reduce: UnaryOpTerm
			reduce(197), // //, reduce: UnaryOpTerm
			reduce(197), // %, reduce: UnaryOpTerm
			reduce(197), // .*, reduce: UnaryOpTerm
			reduce(197), // ./, reduce: UnaryOpTerm
			reduce(197), // .//, reduce: UnaryOpTerm
			reduce(197), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S734
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(861), // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S735
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S736
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S737
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S738
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S739
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S740
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S741
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S742
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(198), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(198), // ;, reduce: UnaryOpTerm
			reduce(198), // {, reduce: UnaryOpTerm
			nil,         // }
			reduce(198), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(198), // >, reduce: UnaryOpTerm
			reduce(198), // >>, reduce: UnaryOpTerm
			reduce(198), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(198), // ||=, reduce: UnaryOpTerm
			reduce(198), // ^^=, reduce: UnaryOpTerm
			reduce(198), // &&=, reduce: UnaryOpTerm
			reduce(198), // ??=, reduce: UnaryOpTerm
			reduce(198), // ???=, reduce: UnaryOpTerm
			reduce(198), // |=, reduce: UnaryOpTerm
			reduce(198), // &=, reduce: UnaryOpTerm
			reduce(198), // ^=, reduce: UnaryOpTerm
			reduce(198), // <<=, reduce: UnaryOpTerm
			reduce(198), // >>=, reduce: UnaryOpTerm
			reduce(198), // >>>=, reduce: UnaryOpTerm
			reduce(198), // +=, reduce: UnaryOpTerm
			reduce(198), // .=, reduce: UnaryOpTerm
			reduce(198), // -=, reduce: UnaryOpTerm
			reduce(198), // *=, reduce: UnaryOpTerm
			reduce(198), // /=, reduce: UnaryOpTerm
			reduce(198), // //=, reduce: UnaryOpTerm
			reduce(198), // %=, reduce: UnaryOpTerm
			reduce(198), // **=, reduce: UnaryOpTerm
			reduce(198), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(198), // ||, reduce: UnaryOpTerm
			reduce(198), // ^^, reduce: UnaryOpTerm
			reduce(198), // &&, reduce: UnaryOpTerm
			reduce(198), // ??, reduce: UnaryOpTerm
			reduce(198), // ???, reduce: UnaryOpTerm
			reduce(198), // =~, reduce: UnaryOpTerm
			reduce(198), // !=~, reduce: UnaryOpTerm
			reduce(198), // ==, reduce: UnaryOpTerm
			reduce(198), // !=, reduce: UnaryOpTerm
			reduce(198), // <=>, reduce: UnaryOpTerm
			reduce(198), // >=, reduce: UnaryOpTerm
			reduce(198), // <, reduce: UnaryOpTerm
			reduce(198), // <=, reduce: UnaryOpTerm
			reduce(198), // ^, reduce: UnaryOpTerm
			reduce(198), // &, reduce: UnaryOpTerm
			reduce(198), // <<, reduce: UnaryOpTerm
			reduce(198), // >>>, reduce: UnaryOpTerm
			reduce(198), // +, reduce: UnaryOpTerm
			reduce(198), // -, reduce: UnaryOpTerm
			reduce(198), // .+, reduce: UnaryOpTerm
			reduce(198), // .-, reduce: UnaryOpTerm
			reduce(198), // *, reduce: UnaryOpTerm
			reduce(198), // /, reduce: UnaryOpTerm
			reduce(198), // //, reduce: UnaryOpTerm
			reduce(198), // %, reduce: UnaryOpTerm
			reduce(198), // .*, reduce: UnaryOpTerm
			reduce(198), // ./, reduce: UnaryOpTerm
			reduce(198), // .//, reduce: UnaryOpTerm
			reduce(198), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S743
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(199), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(199), // ;, reduce: UnaryOpTerm
			reduce(199), // {, reduce: UnaryOpTerm
			nil,         // }
			reduce(199), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(199), // >, reduce: UnaryOpTerm
			reduce(199), // >>, reduce: UnaryOpTerm
			reduce(199), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(199), // ||=, reduce: UnaryOpTerm
			reduce(199), // ^^=, reduce: UnaryOpTerm
			reduce(199), // &&=, reduce: UnaryOpTerm
			reduce(199), // ??=, reduce: UnaryOpTerm
			reduce(199), // ???=, reduce: UnaryOpTerm
			reduce(199), // |=, reduce: UnaryOpTerm
			reduce(199), // &=, reduce: UnaryOpTerm
			reduce(199), // ^=, reduce: UnaryOpTerm
			reduce(199), // <<=, reduce: UnaryOpTerm
			reduce(199), // >>=, reduce: UnaryOpTerm
			reduce(199), // >>>=, reduce: UnaryOpTerm
			reduce(199), // +=, reduce: UnaryOpTerm
			reduce(199), // .=, reduce: UnaryOpTerm
			reduce(199), // -=, reduce: UnaryOpTerm
			reduce(199), // *=, reduce: UnaryOpTerm
			reduce(199), // /=, reduce: UnaryOpTerm
			reduce(199), // //=, reduce: UnaryOpTerm
			reduce(199), // %=, reduce: UnaryOpTerm
			reduce(199), // **=, reduce: UnaryOpTerm
			reduce(199), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(199), // ||, reduce: UnaryOpTerm
			reduce(199), // ^^, reduce: UnaryOpTerm
			reduce(199), // &&, reduce: UnaryOpTerm
			reduce(199), // ??, reduce: UnaryOpTerm
			reduce(199), // ???, reduce: UnaryOpTerm
			reduce(199), // =~, reduce: UnaryOpTerm
			reduce(199), // !=~, reduce: UnaryOpTerm
			reduce(199), // ==, reduce: UnaryOpTerm
			reduce(199), // !=, reduce: UnaryOpTerm
			reduce(199), // <=>, reduce: UnaryOpTerm
			reduce(199), // >=, reduce: UnaryOpTerm
			reduce(199), // <, reduce: UnaryOpTerm
			reduce(199), // <=, reduce: UnaryOpTerm
			reduce(199), // ^, reduce: UnaryOpTerm
			reduce(199), // &, reduce: UnaryOpTerm
			reduce(199), // <<, reduce: UnaryOpTerm
			reduce(199), // >>>, reduce: UnaryOpTerm
			reduce(199), // +, reduce: UnaryOpTerm
			reduce(199), // -, reduce: UnaryOpTerm
			reduce(199), // .+, reduce: UnaryOpTerm
			reduce(199), // .-, reduce: UnaryOpTerm
			reduce(199), // *, reduce: UnaryOpTerm
			reduce(199), // /, reduce: UnaryOpTerm
			reduce(199), // //, reduce: UnaryOpTerm
			reduce(199), // %, reduce: UnaryOpTerm
			reduce(199), // .*, reduce: UnaryOpTerm
			reduce(199), // ./, reduce: UnaryOpTerm
			reduce(199), // .//, reduce: UnaryOpTerm
			reduce(199), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S744
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(200), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(200), // ;, reduce: UnaryOpTerm
			reduce(200), // {, reduce: UnaryOpTerm
			nil,         // }
			reduce(200), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(200), // >, reduce: UnaryOpTerm
			reduce(200), // >>, reduce: UnaryOpTerm
			reduce(200), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(200), // ||=, reduce: UnaryOpTerm
			reduce(200), // ^^=, reduce: UnaryOpTerm
			reduce(200), // &&=, reduce: UnaryOpTerm
			reduce(200), // ??=, reduce: UnaryOpTerm
			reduce(200), // ???=, reduce: UnaryOpTerm
			reduce(200), // |=, reduce: UnaryOpTerm
			reduce(200), // &=, reduce: UnaryOpTerm
			reduce(200), // ^=, reduce: UnaryOpTerm
			reduce(200), // <<=, reduce: UnaryOpTerm
			reduce(200), // >>=, reduce: UnaryOpTerm
			reduce(200), // >>>=, reduce: UnaryOpTerm
			reduce(200), // +=, reduce: UnaryOpTerm
			reduce(200), // .=, reduce: UnaryOpTerm
			reduce(200), // -=, reduce: UnaryOpTerm
			reduce(200), // *=, reduce: UnaryOpTerm
			reduce(200), // /=, reduce: UnaryOpTerm
			reduce(200), // //=, reduce: UnaryOpTerm
			reduce(200), // %=, reduce: UnaryOpTerm
			reduce(200), // **=, reduce: UnaryOpTerm
			reduce(200), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(200), // ||, reduce: UnaryOpTerm
			reduce(200), // ^^, reduce: UnaryOpTerm
			reduce(200), // &&, reduce: UnaryOpTerm
			reduce(200), // ??, reduce: UnaryOpTerm
			reduce(200), // ???, reduce: UnaryOpTerm
			reduce(200), // =~, reduce: UnaryOpTerm
			reduce(200), // !=~, reduce: UnaryOpTerm
			reduce(200), // ==, reduce: UnaryOpTerm
			reduce(200), // !=, reduce: UnaryOpTerm
			reduce(200), // <=>, reduce: UnaryOpTerm
			reduce(200), // >=, reduce: UnaryOpTerm
			reduce(200), // <, reduce: UnaryOpTerm
			reduce(200), // <=, reduce: UnaryOpTerm
			reduce(200), // ^, reduce: UnaryOpTerm
			reduce(200), // &, reduce: UnaryOpTerm
			reduce(200), // <<, reduce: UnaryOpTerm
			reduce(200), // >>>, reduce: UnaryOpTerm
			reduce(200), // +, reduce: UnaryOpTerm
			reduce(200), // -, reduce: UnaryOpTerm
			reduce(200), // .+, reduce: UnaryOpTerm
			reduce(200), // .-, reduce: UnaryOpTerm
			reduce(200), // *, reduce: UnaryOpTerm
			reduce(200), // /, reduce: UnaryOpTerm
			reduce(200), // //, reduce: UnaryOpTerm
			reduce(200), // %, reduce: UnaryOpTerm
			reduce(200), // .*, reduce: UnaryOpTerm
			reduce(200), // ./, reduce: UnaryOpTerm
			reduce(200), // .//, reduce: UnaryOpTerm
			reduce(200), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S745
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S746
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(201), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(201), // ;, reduce: UnaryOpTerm
			reduce(201), // {, reduce: UnaryOpTerm
			nil,         // }
			reduce(201), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(201), // >, reduce: UnaryOpTerm
			reduce(201), // >>, reduce: UnaryOpTerm
			reduce(201), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(201), // ||=, reduce: UnaryOpTerm
			reduce(201), // ^^=, reduce: UnaryOpTerm
			reduce(201), // &&=, reduce: UnaryOpTerm
			reduce(201), // ??=, reduce: UnaryOpTerm
			reduce(201), // ???=, reduce: UnaryOpTerm
			reduce(201), // |=, reduce: UnaryOpTerm
			reduce(201), // &=, reduce: UnaryOpTerm
			reduce(201), // ^=, reduce: UnaryOpTerm
			reduce(201), // <<=, reduce: UnaryOpTerm
			reduce(201), // >>=, reduce: UnaryOpTerm
			reduce(201), // >>>=, reduce: UnaryOpTerm
			reduce(201), // +=, reduce: UnaryOpTerm
			reduce(201), // .=, reduce: UnaryOpTerm
			reduce(201), // -=, reduce: UnaryOpTerm
			reduce(201), // *=, reduce: UnaryOpTerm
			reduce(201), // /=, reduce: UnaryOpTerm
			reduce(201), // //=, reduce: UnaryOpTerm
			reduce(201), // %=, reduce: UnaryOpTerm
			reduce(201), // **=, reduce: UnaryOpTerm
			reduce(201), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(201), // ||, reduce: UnaryOpTerm
			reduce(201), // ^^, reduce: UnaryOpTerm
			reduce(201), // &&, reduce: UnaryOpTerm
			reduce(201), // ??, reduce: UnaryOpTerm
			reduce(201), // ???, reduce: UnaryOpTerm
			reduce(201), // =~, reduce: UnaryOpTerm
			reduce(201), // !=~, reduce: UnaryOpTerm
			reduce(201), // ==, reduce: UnaryOpTerm
			reduce(201), // !=, reduce: UnaryOpTerm
			reduce(201), // <=>, reduce: UnaryOpTerm
			reduce(201), // >=, reduce: UnaryOpTerm
			reduce(201), // <, reduce: UnaryOpTerm
			reduce(201), // <=, reduce: UnaryOpTerm
			reduce(201), // ^, reduce: UnaryOpTerm
			reduce(201), // &, reduce: UnaryOpTerm
			reduce(201), // <<, reduce: UnaryOpTerm
			reduce(201), // >>>, reduce: UnaryOpTerm
			reduce(201), // +, reduce: UnaryOpTerm
			reduce(201), // -, reduce: UnaryOpTerm
			reduce(201), // .+, reduce: UnaryOpTerm
			reduce(201), // .-, reduce: UnaryOpTerm
			reduce(201), // *, reduce: UnaryOpTerm
			reduce(201), // /, reduce: UnaryOpTerm
			reduce(201), // //, reduce: UnaryOpTerm
			reduce(201), // %, reduce: UnaryOpTerm
			reduce(201), // .*, reduce: UnaryOpTerm
			reduce(201), // ./, reduce: UnaryOpTerm
			reduce(201), // .//, reduce: UnaryOpTerm
			reduce(201), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S747
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(202), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(202), // ;, reduce: UnaryOpTerm
			reduce(202), // {, reduce: UnaryOpTerm
			nil,         // }
			reduce(202), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(202), // >, reduce: UnaryOpTerm
			reduce(202), // >>, reduce: UnaryOpTerm
			reduce(202), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(202), // ||=, reduce: UnaryOpTerm
			reduce(202), // ^^=, reduce: UnaryOpTerm
			reduce(202), // &&=, reduce: UnaryOpTerm
			reduce(202), // ??=, reduce: UnaryOpTerm
			reduce(202), // ???=, reduce: UnaryOpTerm
			reduce(202), // |=, reduce: UnaryOpTerm
			reduce(202), // &=, reduce: UnaryOpTerm
			reduce(202), // ^=, reduce: UnaryOpTerm
			reduce(202), // <<=, reduce: UnaryOpTerm
			reduce(202), // >>=, reduce: UnaryOpTerm
			reduce(202), // >>>=, reduce: UnaryOpTerm
			reduce(202), // +=, reduce: UnaryOpTerm
			reduce(202), // .=, reduce: UnaryOpTerm
			reduce(202), // -=, reduce: UnaryOpTerm
			reduce(202), // *=, reduce: UnaryOpTerm
			reduce(202), // /=, reduce: UnaryOpTerm
			reduce(202), // //=, reduce: UnaryOpTerm
			reduce(202), // %=, reduce: UnaryOpTerm
			reduce(202), // **=, reduce: UnaryOpTerm
			reduce(202), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(202), // ||, reduce: UnaryOpTerm
			reduce(202), // ^^, reduce: UnaryOpTerm
			reduce(202), // &&, reduce: UnaryOpTerm
			reduce(202), // ??, reduce: UnaryOpTerm
			reduce(202), // ???, reduce: UnaryOpTerm
			reduce(202), // =~, reduce: UnaryOpTerm
			reduce(202), // !=~, reduce: UnaryOpTerm
			reduce(202), // ==, reduce: UnaryOpTerm
			reduce(202), // !=, reduce: UnaryOpTerm
			reduce(202), // <=>, reduce: UnaryOpTerm
			reduce(202), // >=, reduce: UnaryOpTerm
			reduce(202), // <, reduce: UnaryOpTerm
			reduce(202), // <=, reduce: UnaryOpTerm
			reduce(202), // ^, reduce: UnaryOpTerm
			reduce(202), // &, reduce: UnaryOpTerm
			reduce(202), // <<, reduce: UnaryOpTerm
			reduce(202), // >>>, reduce: UnaryOpTerm
			reduce(202), // +, reduce: UnaryOpTerm
			reduce(202), // -, reduce: UnaryOpTerm
			reduce(202), // .+, reduce: UnaryOpTerm
			reduce(202), // .-, reduce: UnaryOpTerm
			reduce(202), // *, reduce: UnaryOpTerm
			reduce(202), // /, reduce: UnaryOpTerm
			reduce(202), // //, reduce: UnaryOpTerm
			reduce(202), // %, reduce: UnaryOpTerm
			reduce(202), // .*, reduce: UnaryOpTerm
			reduce(202), // ./, reduce: UnaryOpTerm
			reduce(202), // .//, reduce: UnaryOpTerm
			reduce(202), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S748
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(7),    // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(45),   // (
			nil,         // )
			nil,         // emitp
			shift(58),   // field_name
			shift(59),   // $[
			nil,         // ]
			shift(60),   // braced_field_name
			shift(61),   // $[[
			shift(62),   // $[[[
			shift(63),   // full_srec
			shift(65),   // oosvar_name
			shift(66),   // @[
			shift(67),   // braced_oosvar_name
			shift(68),   // full_oosvar
			shift(69),   // all
			shift(70),   // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1506), // +
			shift(1507), // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(107),  // string_literal
			shift(108),  // regex_case_insensitive
			shift(109),  // int_literal
			shift(110),  // float_literal
			shift(111),  // boolean_literal
			shift(112),  // null_literal
			shift(113),  // inf_literal
			shift(114),  // nan_literal
			shift(115),  // const_M_PI
			shift(116),  // const_M_E
			shift(117),  // panic
			shift(119),  // [
			shift(121),  // ctx_IPS
			shift(122),  // ctx_IFS
			shift(123),  // ctx_IRS
			shift(124),  // ctx_OPS
			shift(125),  // ctx_OFS
			shift(126),  // ctx_ORS
			shift(127),  // ctx_FLATSEP
			shift(128),  // ctx_NF
			shift(129),  // ctx_NR
			shift(130),  // ctx_FNR
			shift(131),  // ctx_FILENAME
			shift(132),  // ctx_FILENUM
			shift(134),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(734),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S749
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(1541), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S750
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S751
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S752
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			shift(1603), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S753
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(1605), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(230), // ], reduce: ArrayLiteralElements
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S754
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(214), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(214), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(214), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S755
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(211), // >, reduce: MlrvalOrFunction
			reduce(211), // >>, reduce: MlrvalOrFunction
			reduce(211), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(211), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(211), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(211), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(211), // ||, reduce: MlrvalOrFunction
			reduce(211), // ^^, reduce: MlrvalOrFunction
			reduce(211), // &&, reduce: MlrvalOrFunction
			reduce(211), // ??, reduce: MlrvalOrFunction
			reduce(211), // ???, reduce: MlrvalOrFunction
			reduce(211), // =~, reduce: MlrvalOrFunction
			reduce(211), // !=~, reduce: MlrvalOrFunction
			reduce(211), // ==, reduce: MlrvalOrFunction
			reduce(211), // !=, reduce: MlrvalOrFunction
			reduce(211), // <=>, reduce: MlrvalOrFunction
			reduce(211), // >=, reduce: MlrvalOrFunction
			reduce(211), // <, reduce: MlrvalOrFunction
			reduce(211), // <=, reduce: MlrvalOrFunction
			reduce(211), // ^, reduce: MlrvalOrFunction
			reduce(211), // &, reduce: MlrvalOrFunction
			reduce(211), // <<, reduce: MlrvalOrFunction
			reduce(211), // >>>, reduce: MlrvalOrFunction
			reduce(211), // +, reduce: MlrvalOrFunction
			reduce(211), // -, reduce: MlrvalOrFunction
			reduce(211), // .+, reduce: MlrvalOrFunction
			reduce(211), // .-, reduce: MlrvalOrFunction
			reduce(211), // *, reduce: MlrvalOrFunction
			reduce(211), // /, reduce: MlrvalOrFunction
			reduce(211), // //, reduce: MlrvalOrFunction
			reduce(211), // %, reduce: MlrvalOrFunction
			reduce(211), // .*, reduce: MlrvalOrFunction
			reduce(211), // ./, reduce: MlrvalOrFunction
			reduce(211), // .//, reduce: MlrvalOrFunction
			reduce(211), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(211), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(211), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(211), // [[, reduce: MlrvalOrFunction
			reduce(211), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S756
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S757
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(110), // >, reduce: OosvarValue
			reduce(110), // >>, reduce: OosvarValue
			reduce(110), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(110), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(110), // ], reduce: OosvarValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(110), // ?, reduce: OosvarValue
			nil,         // :
			reduce(110), // ||, reduce: OosvarValue
			reduce(110), // ^^, reduce: OosvarValue
			reduce(110), // &&, reduce: OosvarValue
			reduce(110), // ??, reduce: OosvarValue
			reduce(110), // ???, reduce: OosvarValue
			reduce(110), // =~, reduce: OosvarValue
			reduce(110), // !=~, reduce: OosvarValue
			reduce(110), // ==, reduce: OosvarValue
			reduce(110), // !=, reduce: OosvarValue
			reduce(110), // <=>, reduce: OosvarValue
			reduce(110), // >=, reduce: OosvarValue
			reduce(110), // <, reduce: OosvarValue
			reduce(110), // <=, reduce: OosvarValue
			reduce(110), // ^, reduce: OosvarValue
			reduce(110), // &, reduce: OosvarValue
			reduce(110), // <<, reduce: OosvarValue
			reduce(110), // >>>, reduce: OosvarValue
			reduce(110), // +, reduce: OosvarValue
			reduce(110), // -, reduce: OosvarValue
			reduce(110), // .+, reduce: OosvarValue
			reduce(110), // .-, reduce: OosvarValue
			reduce(110), // *, reduce: OosvarValue
			reduce(110), // /, reduce: OosvarValue
			reduce(110), // //, reduce: OosvarValue
			reduce(110), // %, reduce: OosvarValue
			reduce(110), // .*, reduce: OosvarValue
			reduce(110), // ./, reduce: OosvarValue
			reduce(110), // .//, reduce: OosvarValue
			reduce(110), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(110), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(110), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(110), // [[, reduce: OosvarValue
			reduce(110), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S758
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(112), // >, reduce: OosvarValue
			reduce(112), // >>, reduce: OosvarValue
			reduce(112), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(112), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(112), // ], reduce: OosvarValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(112), // ?, reduce: OosvarValue
			nil,         // :
			reduce(112), // ||, reduce: OosvarValue
			reduce(112), // ^^, reduce: OosvarValue
			reduce(112), // &&, reduce: OosvarValue
			reduce(112), // ??, reduce: OosvarValue
			reduce(112), // ???, reduce: OosvarValue
			reduce(112), // =~, reduce: OosvarValue
			reduce(112), // !=~, reduce: OosvarValue
			reduce(112), // ==, reduce: OosvarValue
			reduce(112), // !=, reduce: OosvarValue
			reduce(112), // <=>, reduce: OosvarValue
			reduce(112), // >=, reduce: OosvarValue
			reduce(112), // <, reduce: OosvarValue
			reduce(112), // <=, reduce: OosvarValue
			reduce(112), // ^, reduce: OosvarValue
			reduce(112), // &, reduce: OosvarValue
			reduce(112), // <<, reduce: OosvarValue
			reduce(112), // >>>, reduce: OosvarValue
			reduce(112), // +, reduce: OosvarValue
			reduce(112), // -, reduce: OosvarValue
			reduce(112), // .+, reduce: OosvarValue
			reduce(112), // .-, reduce: OosvarValue
			reduce(112), // *, reduce: OosvarValue
			reduce(112), // /, reduce: OosvarValue
			reduce(112), // //, reduce: OosvarValue
			reduce(112), // %, reduce: OosvarValue
			reduce(112), // .*, reduce: OosvarValue
			reduce(112), // ./, reduce: OosvarValue
			reduce(112), // .//, reduce: OosvarValue
			reduce(112), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(112), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(112), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(112), // [[, reduce: OosvarValue
			reduce(112), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S759
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(111), // >, reduce: OosvarValue
			reduce(111), // >>, reduce: OosvarValue
			reduce(111), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(111), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(111), // ], reduce: OosvarValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(111), // ?, reduce: OosvarValue
			nil,         // :
			reduce(111), // ||, reduce: OosvarValue
			reduce(111), // ^^, reduce: OosvarValue
			reduce(111), // &&, reduce: OosvarValue
			reduce(111), // ??, reduce: OosvarValue
			reduce(111), // ???, reduce: OosvarValue
			reduce(111), // =~, reduce: OosvarValue
			reduce(111), // !=~, reduce: OosvarValue
			reduce(111), // ==, reduce: OosvarValue
			reduce(111), // !=, reduce: OosvarValue
			reduce(111), // <=>, reduce: OosvarValue
			reduce(111), // >=, reduce: OosvarValue
			reduce(111), // <, reduce: OosvarValue
			reduce(111), // <=, reduce: OosvarValue
			reduce(111), // ^, reduce: OosvarValue
			reduce(111), // &, reduce: OosvarValue
			reduce(111), // <<, reduce: OosvarValue
			reduce(111), // >>>, reduce: OosvarValue
			reduce(111), // +, reduce: OosvarValue
			reduce(111), // -, reduce: OosvarValue
			reduce(111), // .+, reduce: OosvarValue
			reduce(111), // .-, reduce: OosvarValue
			reduce(111), // *, reduce: OosvarValue
			reduce(111), // /, reduce: OosvarValue
			reduce(111), // //, reduce: OosvarValue
			reduce(111), // %, reduce: OosvarValue
			reduce(111), // .*, reduce: OosvarValue
			reduce(111), // ./, reduce: OosvarValue
			reduce(111), // .//, reduce: OosvarValue
			reduce(111), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(111), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(111), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(111), // [[, reduce: OosvarValue
			reduce(111), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S760
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(99), // >, reduce: FieldValue
			reduce(99), // >>, reduce: FieldValue
			reduce(99), // |, reduce: FieldValue
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(99), // ,, reduce: FieldValue
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			reduce(99), // ], reduce: FieldValue
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(99), // ?, reduce: FieldValue
			nil,        // :
			reduce(99), // ||, reduce: FieldValue
			reduce(99), // ^^, reduce: FieldValue
			reduce(99), // &&, reduce: FieldValue
			reduce(99), // ??, reduce: FieldValue
			reduce(99), // ???, reduce: FieldValue
			reduce(99), // =~, reduce: FieldValue
			reduce(99), // !=~, reduce: FieldValue
			reduce(99), // ==, reduce: FieldValue
			reduce(99), // !=, reduce: FieldValue
			reduce(99), // <=>, reduce: FieldValue
			reduce(99), // >=, reduce: FieldValue
			reduce(99), // <, reduce: FieldValue
			reduce(99), // <=, reduce: FieldValue
			reduce(99), // ^, reduce: FieldValue
			reduce(99), // &, reduce: FieldValue
			reduce(99), // <<, reduce: FieldValue
			reduce(99), // >>>, reduce: FieldValue
			reduce(99), // +, reduce: FieldValue
			reduce(99), // -, reduce: FieldValue
			reduce(99), // .+, reduce: FieldValue
			reduce(99), // .-, reduce: FieldValue
			reduce(99), // *, reduce: FieldValue
			reduce(99), // /, reduce: FieldValue
			reduce(99), // //, reduce: FieldValue
			reduce(99), // %, reduce: FieldValue
			reduce(99), // .*, reduce: FieldValue
			reduce(99), // ./, reduce: FieldValue
			reduce(99), // .//, reduce: FieldValue
			reduce(99), // ., reduce: FieldValue
			nil,        // !
			nil,        // ~
			reduce(99), // **, reduce: FieldValue
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(99), // [, reduce: FieldValue
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(99), // [[, reduce: FieldValue
			reduce(99), // [[[, reduce: FieldValue
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S761
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(101), // >, reduce: FieldValue
			reduce(101), // >>, reduce: FieldValue
			reduce(101), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(101), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(101), // ], reduce: FieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(101), // ?, reduce: FieldValue
			nil,         // :
			reduce(101), // ||, reduce: FieldValue
			reduce(101), // ^^, reduce: FieldValue
			reduce(101), // &&, reduce: FieldValue
			reduce(101), // ??, reduce: FieldValue
			reduce(101), // ???, reduce: FieldValue
			reduce(101), // =~, reduce: FieldValue
			reduce(101), // !=~, reduce: FieldValue
			reduce(101), // ==, reduce: FieldValue
			reduce(101), // !=, reduce: FieldValue
			reduce(101), // <=>, reduce: FieldValue
			reduce(101), // >=, reduce: FieldValue
			reduce(101), // <, reduce: FieldValue
			reduce(101), // <=, reduce: FieldValue
			reduce(101), // ^, reduce: FieldValue
			reduce(101), // &, reduce: FieldValue
			reduce(101), // <<, reduce: FieldValue
			reduce(101), // >>>, reduce: FieldValue
			reduce(101), // +, reduce: FieldValue
			reduce(101), // -, reduce: FieldValue
			reduce(101), // .+, reduce: FieldValue
			reduce(101), // .-, reduce: FieldValue
			reduce(101), // *, reduce: FieldValue
			reduce(101), // /, reduce: FieldValue
			reduce(101), // //, reduce: FieldValue
			reduce(101), // %, reduce: FieldValue
			reduce(101), // .*, reduce: FieldValue
			reduce(101), // ./, reduce: FieldValue
			reduce(101), // .//, reduce: FieldValue
			reduce(101), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(101), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(101), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(101), // [[, reduce: FieldValue
			reduce(101), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S762
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(100), // >, reduce: FieldValue
			reduce(100), // >>, reduce: FieldValue
			reduce(100), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(100), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(100), // ], reduce: FieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(100), // ?, reduce: FieldValue
			nil,         // :
			reduce(100), // ||, reduce: FieldValue
			reduce(100), // ^^, reduce: FieldValue
			reduce(100), // &&, reduce: FieldValue
			reduce(100), // ??, reduce: FieldValue
			reduce(100), // ???, reduce: FieldValue
			reduce(100), // =~, reduce: FieldValue
			reduce(100), // !=~, reduce: FieldValue
			reduce(100), // ==, reduce: FieldValue
			reduce(100), // !=, reduce: FieldValue
			reduce(100), // <=>, reduce: FieldValue
			reduce(100), // >=, reduce: FieldValue
			reduce(100), // <, reduce: FieldValue
			reduce(100), // <=, reduce: FieldValue
			reduce(100), // ^, reduce: FieldValue
			reduce(100), // &, reduce: FieldValue
			reduce(100), // <<, reduce: FieldValue
			reduce(100), // >>>, reduce: FieldValue
			reduce(100), // +, reduce: FieldValue
			reduce(100), // -, reduce: FieldValue
			reduce(100), // .+, reduce: FieldValue
			reduce(100), // .-, reduce: FieldValue
			reduce(100), // *, reduce: FieldValue
			reduce(100), // /, reduce: FieldValue
			reduce(100), // //, reduce: FieldValue
			reduce(100), // %, reduce: FieldValue
			reduce(100), // .*, reduce: FieldValue
			reduce(100), // ./, reduce: FieldValue
			reduce(100), // .//, reduce: FieldValue
			reduce(100), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(100), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(100), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(100), // [[, reduce: FieldValue
			reduce(100), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S763
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(213), // >, reduce: MlrvalOrFunction
			reduce(213), // >>, reduce: MlrvalOrFunction
			reduce(213), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(213), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(213), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(213), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(213), // ||, reduce: MlrvalOrFunction
			reduce(213), // ^^, reduce: MlrvalOrFunction
			reduce(213), // &&, reduce: MlrvalOrFunction
			reduce(213), // ??, reduce: MlrvalOrFunction
			reduce(213), // ???, reduce: MlrvalOrFunction
			reduce(213), // =~, reduce: MlrvalOrFunction
			reduce(213), // !=~, reduce: MlrvalOrFunction
			reduce(213), // ==, reduce: MlrvalOrFunction
			reduce(213), // !=, reduce: MlrvalOrFunction
			reduce(213), // <=>, reduce: MlrvalOrFunction
			reduce(213), // >=, reduce: MlrvalOrFunction
			reduce(213), // <, reduce: MlrvalOrFunction
			reduce(213), // <=, reduce: MlrvalOrFunction
			reduce(213), // ^, reduce: MlrvalOrFunction
			reduce(213), // &, reduce: MlrvalOrFunction
			reduce(213), // <<, reduce: MlrvalOrFunction
			reduce(213), // >>>, reduce: MlrvalOrFunction
			reduce(213), // +, reduce: MlrvalOrFunction
			reduce(213), // -, reduce: MlrvalOrFunction
			reduce(213), // .+, reduce: MlrvalOrFunction
			reduce(213), // .-, reduce: MlrvalOrFunction
			reduce(213), // *, reduce: MlrvalOrFunction
			reduce(213), // /, reduce: MlrvalOrFunction
			reduce(213), // //, reduce: MlrvalOrFunction
			reduce(213), // %, reduce: MlrvalOrFunction
			reduce(213), // .*, reduce: MlrvalOrFunction
			reduce(213), // ./, reduce: MlrvalOrFunction
			reduce(213), // .//, reduce: MlrvalOrFunction
			reduce(213), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(213), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(213), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(213), // [[, reduce: MlrvalOrFunction
			reduce(213), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S764
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(233), // >, reduce: MlrvalOrFunction
			reduce(233), // >>, reduce: MlrvalOrFunction
			reduce(233), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(233), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(233), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(233), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(233), // ||, reduce: MlrvalOrFunction
			reduce(233), // ^^, reduce: MlrvalOrFunction
			reduce(233), // &&, reduce: MlrvalOrFunction
			reduce(233), // ??, reduce: MlrvalOrFunction
			reduce(233), // ???, reduce: MlrvalOrFunction
			reduce(233), // =~, reduce: MlrvalOrFunction
			reduce(233), // !=~, reduce: MlrvalOrFunction
			reduce(233), // ==, reduce: MlrvalOrFunction
			reduce(233), // !=, reduce: MlrvalOrFunction
			reduce(233), // <=>, reduce: MlrvalOrFunction
			reduce(233), // >=, reduce: MlrvalOrFunction
			reduce(233), // <, reduce: MlrvalOrFunction
			reduce(233), // <=, reduce: MlrvalOrFunction
			reduce(233), // ^, reduce: MlrvalOrFunction
			reduce(233), // &, reduce: MlrvalOrFunction
			reduce(233), // <<, reduce: MlrvalOrFunction
			reduce(233), // >>>, reduce: MlrvalOrFunction
			reduce(233), // +, reduce: MlrvalOrFunction
			reduce(233), // -, reduce: MlrvalOrFunction
			reduce(233), // .+, reduce: MlrvalOrFunction
			reduce(233), // .-, reduce: MlrvalOrFunction
			reduce(233), // *, reduce: MlrvalOrFunction
			reduce(233), // /, reduce: MlrvalOrFunction
			reduce(233), // //, reduce: MlrvalOrFunction
			reduce(233), // %, reduce: MlrvalOrFunction
			reduce(233), // .*, reduce: MlrvalOrFunction
			reduce(233), // ./, reduce: MlrvalOrFunction
			reduce(233), // .//, reduce: MlrvalOrFunction
			reduce(233), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(233), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(233), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(233), // [[, reduce: MlrvalOrFunction
			reduce(233), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S765
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(210), // >, reduce: MlrvalOrFunction
			reduce(210), // >>, reduce: MlrvalOrFunction
			reduce(210), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(210), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(210), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(210), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(210), // ||, reduce: MlrvalOrFunction
			reduce(210), // ^^, reduce: MlrvalOrFunction
			reduce(210), // &&, reduce: MlrvalOrFunction
			reduce(210), // ??, reduce: MlrvalOrFunction
			reduce(210), // ???, reduce: MlrvalOrFunction
			reduce(210), // =~, reduce: MlrvalOrFunction
			reduce(210), // !=~, reduce: MlrvalOrFunction
			reduce(210), // ==, reduce: MlrvalOrFunction
			reduce(210), // !=, reduce: MlrvalOrFunction
			reduce(210), // <=>, reduce: MlrvalOrFunction
			reduce(210), // >=, reduce: MlrvalOrFunction
			reduce(210), // <, reduce: MlrvalOrFunction
			reduce(210), // <=, reduce: MlrvalOrFunction
			reduce(210), // ^, reduce: MlrvalOrFunction
			reduce(210), // &, reduce: MlrvalOrFunction
			reduce(210), // <<, reduce: MlrvalOrFunction
			reduce(210), // >>>, reduce: MlrvalOrFunction
			reduce(210), // +, reduce: MlrvalOrFunction
			reduce(210), // -, reduce: MlrvalOrFunction
			reduce(210), // .+, reduce: MlrvalOrFunction
			reduce(210), // .-, reduce: MlrvalOrFunction
			reduce(210), // *, reduce: MlrvalOrFunction
			reduce(210), // /, reduce: MlrvalOrFunction
			reduce(210), // //, reduce: MlrvalOrFunction
			reduce(210), // %, reduce: MlrvalOrFunction
			reduce(210), // .*, reduce: MlrvalOrFunction
			reduce(210), // ./, reduce: MlrvalOrFunction
			reduce(210), // .//, reduce: MlrvalOrFunction
			reduce(210), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(210), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(210), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(210), // [[, reduce: MlrvalOrFunction
			reduce(210), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S766
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(102), // >, reduce: FieldValue
			reduce(102), // >>, reduce: FieldValue
			reduce(102), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(102), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(102), // ], reduce: FieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(102), // ?, reduce: FieldValue
			nil,         // :
			reduce(102), // ||, reduce: FieldValue
			reduce(102), // ^^, reduce: FieldValue
			reduce(102), // &&, reduce: FieldValue
			reduce(102), // ??, reduce: FieldValue
			reduce(102), // ???, reduce: FieldValue
			reduce(102), // =~, reduce: FieldValue
			reduce(102), // !=~, reduce: FieldValue
			reduce(102), // ==, reduce: FieldValue
			reduce(102), // !=, reduce: FieldValue
			reduce(102), // <=>, reduce: FieldValue
			reduce(102), // >=, reduce: FieldValue
			reduce(102), // <, reduce: FieldValue
			reduce(102), // <=, reduce: FieldValue
			reduce(102), // ^, reduce: FieldValue
			reduce(102), // &, reduce: FieldValue
			reduce(102), // <<, reduce: FieldValue
			reduce(102), // >>>, reduce: FieldValue
			reduce(102), // +, reduce: FieldValue
			reduce(102), // -, reduce: FieldValue
			reduce(102), // .+, reduce: FieldValue
			reduce(102), // .-, reduce: FieldValue
			reduce(102), // *, reduce: FieldValue
			reduce(102), // /, reduce: FieldValue
			reduce(102), // //, reduce: FieldValue
			reduce(102), // %, reduce: FieldValue
			reduce(102), // .*, reduce: FieldValue
			reduce(102), // ./, reduce: FieldValue
			reduce(102), // .//, reduce: FieldValue
			reduce(102), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(102), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(102), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(102), // [[, reduce: FieldValue
			reduce(102), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S767
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(103), // >, reduce: FieldValue
			reduce(103), // >>, reduce: FieldValue
			reduce(103), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(103), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(103), // ], reduce: FieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(103), // ?, reduce: FieldValue
			nil,         // :
			reduce(103), // ||, reduce: FieldValue
			reduce(103), // ^^, reduce: FieldValue
			reduce(103), // &&, reduce: FieldValue
			reduce(103), // ??, reduce: FieldValue
			reduce(103), // ???, reduce: FieldValue
			reduce(103), // =~, reduce: FieldValue
			reduce(103), // !=~, reduce: FieldValue
			reduce(103), // ==, reduce: FieldValue
			reduce(103), // !=, reduce: FieldValue
			reduce(103), // <=>, reduce: FieldValue
			reduce(103), // >=, reduce: FieldValue
			reduce(103), // <, reduce: FieldValue
			reduce(103), // <=, reduce: FieldValue
			reduce(103), // ^, reduce: FieldValue
			reduce(103), // &, reduce: FieldValue
			reduce(103), // <<, reduce: FieldValue
			reduce(103), // >>>, reduce: FieldValue
			reduce(103), // +, reduce: FieldValue
			reduce(103), // -, reduce: FieldValue
			reduce(103), // .+, reduce: FieldValue
			reduce(103), // .-, reduce: FieldValue
			reduce(103), // *, reduce: FieldValue
			reduce(103), // /, reduce: FieldValue
			reduce(103), // //, reduce: FieldValue
			reduce(103), // %, reduce: FieldValue
			reduce(103), // .*, reduce: FieldValue
			reduce(103), // ./, reduce: FieldValue
			reduce(103), // .//, reduce: FieldValue
			reduce(103), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(103), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(103), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(103), // [[, reduce: FieldValue
			reduce(103), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S768
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(104), // >, reduce: DirectFieldValue
			reduce(104), // >>, reduce: DirectFieldValue
			reduce(104), // |, reduce: DirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(104), // ,, reduce: DirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(104), // ], reduce: DirectFieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(104), // ?, reduce: DirectFieldValue
			nil,         // :
			reduce(104), // ||, reduce: DirectFieldValue
			reduce(104), // ^^, reduce: DirectFieldValue
			reduce(104), // &&, reduce: DirectFieldValue
			reduce(104), // ??, reduce: DirectFieldValue
			reduce(104), // ???, reduce: DirectFieldValue
			reduce(104), // =~, reduce: DirectFieldValue
			reduce(104), // !=~, reduce: DirectFieldValue
			reduce(104), // ==, reduce: DirectFieldValue
			reduce(104), // !=, reduce: DirectFieldValue
			reduce(104), // <=>, reduce: DirectFieldValue
			reduce(104), // >=, reduce: DirectFieldValue
			reduce(104), // <, reduce: DirectFieldValue
			reduce(104), // <=, reduce: DirectFieldValue
			reduce(104), // ^, reduce: DirectFieldValue
			reduce(104), // &, reduce: DirectFieldValue
			reduce(104), // <<, reduce: DirectFieldValue
			reduce(104), // >>>, reduce: DirectFieldValue
			reduce(104), // +, reduce: DirectFieldValue
			reduce(104), // -, reduce: DirectFieldValue
			reduce(104), // .+, reduce: DirectFieldValue
			reduce(104), // .-, reduce: DirectFieldValue
			reduce(104), // *, reduce: DirectFieldValue
			reduce(104), // /, reduce: DirectFieldValue
			reduce(104), // //, reduce: DirectFieldValue
			reduce(104), // %, reduce: DirectFieldValue
			reduce(104), // .*, reduce: DirectFieldValue
			reduce(104), // ./, reduce: DirectFieldValue
			reduce(104), // .//, reduce: DirectFieldValue
			reduce(104), // ., reduce: DirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(104), // **, reduce: DirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(104), // [, reduce: DirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(104), // [[, reduce: DirectFieldValue
			reduce(104), // [[[, reduce: DirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S769
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S770
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(228), // $, reduce: ArrayLiteral
			nil,         // empty
			reduce(228), // ;, reduce: ArrayLiteral
			reduce(228), // {, reduce: ArrayLiteral
			nil,         // }
			reduce(228), // =, reduce: ArrayLiteral
			nil,         // unset
			nil,         // filter
			reduce(228), // >, reduce: ArrayLiteral
			reduce(228), // >>, reduce: ArrayLiteral
			reduce(228), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(228), // ||=, reduce: ArrayLiteral
			reduce(228), // ^^=, reduce: ArrayLiteral
			reduce(228), // &&=, reduce: ArrayLiteral
			reduce(228), // ??=, reduce: ArrayLiteral
			reduce(228), // ???=, reduce: ArrayLiteral
			reduce(228), // |=, reduce: ArrayLiteral
			reduce(228), // &=, reduce: ArrayLiteral
			reduce(228), // ^=, reduce: ArrayLiteral
			reduce(228), // <<=, reduce: ArrayLiteral
			reduce(228), // >>=, reduce: ArrayLiteral
			reduce(228), // >>>=, reduce: ArrayLiteral
			reduce(228), // +=, reduce: ArrayLiteral
			reduce(228), // .=, reduce: ArrayLiteral
			reduce(228), // -=, reduce: ArrayLiteral
			reduce(228), // *=, reduce: ArrayLiteral
			reduce(228), // /=, reduce: ArrayLiteral
			reduce(228), // //=, reduce: ArrayLiteral
			reduce(228), // %=, reduce: ArrayLiteral
			reduce(228), // **=, reduce: ArrayLiteral
			reduce(228), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(228), // ||, reduce: ArrayLiteral
			reduce(228), // ^^, reduce: ArrayLiteral
			reduce(228), // &&, reduce: ArrayLiteral
			reduce(228), // ??, reduce: ArrayLiteral
			reduce(228), // ???, reduce: ArrayLiteral
			reduce(228), // =~, reduce: ArrayLiteral
			reduce(228), // !=~, reduce: ArrayLiteral
			reduce(228), // ==, reduce: ArrayLiteral
			reduce(228), // !=, reduce: ArrayLiteral
			reduce(228), // <=>, reduce: ArrayLiteral
			reduce(228), // >=, reduce: ArrayLiteral
			reduce(228), // <, reduce: ArrayLiteral
			reduce(228), // <=, reduce: ArrayLiteral
			reduce(228), // ^, reduce: ArrayLiteral
			reduce(228), // &, reduce: ArrayLiteral
			reduce(228), // <<, reduce: ArrayLiteral
			reduce(228), // >>>, reduce: ArrayLiteral
			reduce(228), // +, reduce: ArrayLiteral
			reduce(228), // -, reduce: ArrayLiteral
			reduce(228), // .+, reduce: ArrayLiteral
			reduce(228), // .-, reduce: ArrayLiteral
			reduce(228), // *, reduce: ArrayLiteral
			reduce(228), // /, reduce: ArrayLiteral
			reduce(228), // //, reduce: ArrayLiteral
			reduce(228), // %, reduce: ArrayLiteral
			reduce(228), // .*, reduce: ArrayLiteral
			reduce(228), // ./, reduce: ArrayLiteral
			reduce(228), // .//, reduce: ArrayLiteral
			reduce(228), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(228), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(228), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(228), // [[, reduce: ArrayLiteral
			reduce(228), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S771
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(106), // >, reduce: BracedFieldValue
			reduce(106), // >>, reduce: BracedFieldValue
			reduce(106), // |, reduce: BracedFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(106), // ,, reduce: BracedFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(106), // ], reduce: BracedFieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(106), // ?, reduce: BracedFieldValue
			nil,         // :
			reduce(106), // ||, reduce: BracedFieldValue
			reduce(106), // ^^, reduce: BracedFieldValue
			reduce(106), // &&, reduce: BracedFieldValue
			reduce(106), // ??, reduce: BracedFieldValue
			reduce(106), // ???, reduce: BracedFieldValue
			reduce(106), // =~, reduce: BracedFieldValue
			reduce(106), // !=~, reduce: BracedFieldValue
			reduce(106), // ==, reduce: BracedFieldValue
			reduce(106), // !=, reduce: BracedFieldValue
			reduce(106), // <=>, reduce: BracedFieldValue
			reduce(106), // >=, reduce: BracedFieldValue
			reduce(106), // <, reduce: BracedFieldValue
			reduce(106), // <=, reduce: BracedFieldValue
			reduce(106), // ^, reduce: BracedFieldValue
			reduce(106), // &, reduce: BracedFieldValue
			reduce(106), // <<, reduce: BracedFieldValue
			reduce(106), // >>>, reduce: BracedFieldValue
			reduce(106), // +, reduce: BracedFieldValue
			reduce(106), // -, reduce: BracedFieldValue
			reduce(106), // .+, reduce: BracedFieldValue
			reduce(106), // .-, reduce: BracedFieldValue
			reduce(106), // *, reduce: BracedFieldValue
			reduce(106), // /, reduce: BracedFieldValue
			reduce(106), // //, reduce: BracedFieldValue
			reduce(106), // %, reduce: BracedFieldValue
			reduce(106), // .*, reduce: BracedFieldValue
			reduce(106), // ./, reduce: BracedFieldValue
			reduce(106), // .//, reduce: BracedFieldValue
			reduce(106), // ., reduce: BracedFieldValue
			nil,         // !
			nil,         // ~
			reduce(106), // **, reduce: BracedFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(106), // [, reduce: BracedFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(106), // [[, reduce: BracedFieldValue
			reduce(106), // [[[, reduce: BracedFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S772
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S773
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S774
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(109), // >, reduce: FullSrec
			reduce(109), // >>, reduce: FullSrec
			reduce(109), // |, reduce: FullSrec
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(109), // ,, reduce: FullSrec
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(109), // ], reduce: FullSrec
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(109), // ?, reduce: FullSrec
			nil,         // :
			reduce(109), // ||, reduce: FullSrec
			reduce(109), // ^^, reduce: FullSrec
			reduce(109), // &&, reduce: FullSrec
			reduce(109), // ??, reduce: FullSrec
			reduce(109), // ???, reduce: FullSrec
			reduce(109), // =~, reduce: FullSrec
			reduce(109), // !=~, reduce: FullSrec
			reduce(109), // ==, reduce: FullSrec
			reduce(109), // !=, reduce: FullSrec
			reduce(109), // <=>, reduce: FullSrec
			reduce(109), // >=, reduce: FullSrec
			reduce(109), // <, reduce: FullSrec
			reduce(109), // <=, reduce: FullSrec
			reduce(109), // ^, reduce: FullSrec
			reduce(109), // &, reduce: FullSrec
			reduce(109), // <<, reduce: FullSrec
			reduce(109), // >>>, reduce: FullSrec
			reduce(109), // +, reduce: FullSrec
			reduce(109), // -, reduce: FullSrec
			reduce(109), // .+, reduce: FullSrec
			reduce(109), // .-, reduce: FullSrec
			reduce(109), // *, reduce: FullSrec
			reduce(109), // /, reduce: FullSrec
			reduce(109), // //, reduce: FullSrec
			reduce(109), // %, reduce: FullSrec
			reduce(109), // .*, reduce: FullSrec
			reduce(109), // ./, reduce: FullSrec
			reduce(109), // .//, reduce: FullSrec
			reduce(109), // ., reduce: FullSrec
			nil,         // !
			nil,         // ~
			reduce(109), // **, reduce: FullSrec
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(109), // [, reduce: FullSrec
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(109), // [[, reduce: FullSrec
			reduce(109), // [[[, reduce: FullSrec
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S775
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(212), // >, reduce: MlrvalOrFunction
			reduce(212), // >>, reduce: MlrvalOrFunction
			reduce(212), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(212), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(212), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(212), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(212), // ||, reduce: MlrvalOrFunction
			reduce(212), // ^^, reduce: MlrvalOrFunction
			reduce(212), // &&, reduce: MlrvalOrFunction
			reduce(212), // ??, reduce: MlrvalOrFunction
			reduce(212), // ???, reduce: MlrvalOrFunction
			reduce(212), // =~, reduce: MlrvalOrFunction
			reduce(212), // !=~, reduce: MlrvalOrFunction
			reduce(212), // ==, reduce: MlrvalOrFunction
			reduce(212), // !=, reduce: MlrvalOrFunction
			reduce(212), // <=>, reduce: MlrvalOrFunction
			reduce(212), // >=, reduce: MlrvalOrFunction
			reduce(212), // <, reduce: MlrvalOrFunction
			reduce(212), // <=, reduce: MlrvalOrFunction
			reduce(212), // ^, reduce: MlrvalOrFunction
			reduce(212), // &, reduce: MlrvalOrFunction
			reduce(212), // <<, reduce: MlrvalOrFunction
			reduce(212), // >>>, reduce: MlrvalOrFunction
			reduce(212), // +, reduce: MlrvalOrFunction
			reduce(212), // -, reduce: MlrvalOrFunction
			reduce(212), // .+, reduce: MlrvalOrFunction
			reduce(212), // .-, reduce: MlrvalOrFunction
			reduce(212), // *, reduce: MlrvalOrFunction
			reduce(212), // /, reduce: MlrvalOrFunction
			reduce(212), // //, reduce: MlrvalOrFunction
			reduce(212), // %, reduce: MlrvalOrFunction
			reduce(212), // .*, reduce: MlrvalOrFunction
			reduce(212), // ./, reduce: MlrvalOrFunction
			reduce(212), // .//, reduce: MlrvalOrFunction
			reduce(212), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(212), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(212), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(212), // [[, reduce: MlrvalOrFunction
			reduce(212), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S776
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(113), // >, reduce: DirectOosvarValue
			reduce(113), // >>, reduce: DirectOosvarValue
			reduce(113), // |, reduce: DirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(113), // ,, reduce: DirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(113), // ], reduce: DirectOosvarValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(113), // ?, reduce: DirectOosvarValue
			nil,         // :
			reduce(113), // ||, reduce: DirectOosvarValue
			reduce(113), // ^^, reduce: DirectOosvarValue
			reduce(113), // &&, reduce: DirectOosvarValue
			reduce(113), // ??, reduce: DirectOosvarValue
			reduce(113), // ???, reduce: DirectOosvarValue
			reduce(113), // =~, reduce: DirectOosvarValue
			reduce(113), // !=~, reduce: DirectOosvarValue
			reduce(113), // ==, reduce: DirectOosvarValue
			reduce(113), // !=, reduce: DirectOosvarValue
			reduce(113), // <=>, reduce: DirectOosvarValue
			reduce(113), // >=, reduce: DirectOosvarValue
			reduce(113), // <, reduce: DirectOosvarValue
			reduce(113), // <=, reduce: DirectOosvarValue
			reduce(113), // ^, reduce: DirectOosvarValue
			reduce(113), // &, reduce: DirectOosvarValue
			reduce(113), // <<, reduce: DirectOosvarValue
			reduce(113), // >>>, reduce: DirectOosvarValue
			reduce(113), // +, reduce: DirectOosvarValue
			reduce(113), // -, reduce: DirectOosvarValue
			reduce(113), // .+, reduce: DirectOosvarValue
			reduce(113), // .-, reduce: DirectOosvarValue
			reduce(113), // *, reduce: DirectOosvarValue
			reduce(113), // /, reduce: DirectOosvarValue
			reduce(113), // //, reduce: DirectOosvarValue
			reduce(113), // %, reduce: DirectOosvarValue
			reduce(113), // .*, reduce: DirectOosvarValue
			reduce(113), // ./, reduce: DirectOosvarValue
			reduce(113), // .//, reduce: DirectOosvarValue
			reduce(113), // ., reduce: DirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(113), // **, reduce: DirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(113), // [, reduce: DirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(113), // [[, reduce: DirectOosvarValue
			reduce(113), // [[[, reduce: DirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S777
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S778
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(115), // >, reduce: BracedOosvarValue
			reduce(115), // >>, reduce: BracedOosvarValue
			reduce(115), // |, reduce: BracedOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(115), // ,, reduce: BracedOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(115), // ], reduce: BracedOosvarValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(115), // ?, reduce: BracedOosvarValue
			nil,         // :
			reduce(115), // ||, reduce: BracedOosvarValue
			reduce(115), // ^^, reduce: BracedOosvarValue
			reduce(115), // &&, reduce: BracedOosvarValue
			reduce(115), // ??, reduce: BracedOosvarValue
			reduce(115), // ???, reduce: BracedOosvarValue
			reduce(115), // =~, reduce: BracedOosvarValue
			reduce(115), // !=~, reduce: BracedOosvarValue
			reduce(115), // ==, reduce: BracedOosvarValue
			reduce(115), // !=, reduce: BracedOosvarValue
			reduce(115), // <=>, reduce: BracedOosvarValue
			reduce(115), // >=, reduce: BracedOosvarValue
			reduce(115), // <, reduce: BracedOosvarValue
			reduce(115), // <=, reduce: BracedOosvarValue
			reduce(115), // ^, reduce: BracedOosvarValue
			reduce(115), // &, reduce: BracedOosvarValue
			reduce(115), // <<, reduce: BracedOosvarValue
			reduce(115), // >>>, reduce: BracedOosvarValue
			reduce(115), // +, reduce: BracedOosvarValue
			reduce(115), // -, reduce: BracedOosvarValue
			reduce(115), // .+, reduce: BracedOosvarValue
			reduce(115), // .-, reduce: BracedOosvarValue
			reduce(115), // *, reduce: BracedOosvarValue
			reduce(115), // /, reduce: BracedOosvarValue
			reduce(115), // //, reduce: BracedOosvarValue
			reduce(115), // %, reduce: BracedOosvarValue
			reduce(115), // .*, reduce: BracedOosvarValue
			reduce(115), // ./, reduce: BracedOosvarValue
			reduce(115), // .//, reduce: BracedOosvarValue
			reduce(115), // ., reduce: BracedOosvarValue
			nil,         // !
			nil,         // ~
			reduce(115), // **, reduce: BracedOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(115), // [, reduce: BracedOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(115), // [[, reduce: BracedOosvarValue
			reduce(115), // [[[, reduce: BracedOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S779
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(116), // >, reduce: FullOosvar
			reduce(116), // >>, reduce: FullOosvar
			reduce(116), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(116), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(116), // ], reduce: FullOosvar
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(116), // ?, reduce: FullOosvar
			nil,         // :
			reduce(116), // ||, reduce: FullOosvar
			reduce(116), // ^^, reduce: FullOosvar
			reduce(116), // &&, reduce: FullOosvar
			reduce(116), // ??, reduce: FullOosvar
			reduce(116), // ???, reduce: FullOosvar
			reduce(116), // =~, reduce: FullOosvar
			reduce(116), // !=~, reduce: FullOosvar
			reduce(116), // ==, reduce: FullOosvar
			reduce(116), // !=, reduce: FullOosvar
			reduce(116), // <=>, reduce: FullOosvar
			reduce(116), // >=, reduce: FullOosvar
			reduce(116), // <, reduce: FullOosvar
			reduce(116), // <=, reduce: FullOosvar
			reduce(116), // ^, reduce: FullOosvar
			reduce(116), // &, reduce: FullOosvar
			reduce(116), // <<, reduce: FullOosvar
			reduce(116), // >>>, reduce: FullOosvar
			reduce(116), // +, reduce: FullOosvar
			reduce(116), // -, reduce: FullOosvar
			reduce(116), // .+, reduce: FullOosvar
			reduce(116), // .-, reduce: FullOosvar
			reduce(116), // *, reduce: FullOosvar
			reduce(116), // /, reduce: FullOosvar
			reduce(116), // //, reduce: FullOosvar
			reduce(116), // %, reduce: FullOosvar
			reduce(116), // .*, reduce: FullOosvar
			reduce(116), // ./, reduce: FullOosvar
			reduce(116), // .//, reduce: FullOosvar
			reduce(116), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(116), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(116), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(116), // [[, reduce: FullOosvar
			reduce(116), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S780
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(117), // >, reduce: FullOosvar
			reduce(117), // >>, reduce: FullOosvar
			reduce(117), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(117), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(117), // ], reduce: FullOosvar
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(117), // ?, reduce: FullOosvar
			nil,         // :
			reduce(117), // ||, reduce: FullOosvar
			reduce(117), // ^^, reduce: FullOosvar
			reduce(117), // &&, reduce: FullOosvar
			reduce(117), // ??, reduce: FullOosvar
			reduce(117), // ???, reduce: FullOosvar
			reduce(117), // =~, reduce: FullOosvar
			reduce(117), // !=~, reduce: FullOosvar
			reduce(117), // ==, reduce: FullOosvar
			reduce(117), // !=, reduce: FullOosvar
			reduce(117), // <=>, reduce: FullOosvar
			reduce(117), // >=, reduce: FullOosvar
			reduce(117), // <, reduce: FullOosvar
			reduce(117), // <=, reduce: FullOosvar
			reduce(117), // ^, reduce: FullOosvar
			reduce(117), // &, reduce: FullOosvar
			reduce(117), // <<, reduce: FullOosvar
			reduce(117), // >>>, reduce: FullOosvar
			reduce(117), // +, reduce: FullOosvar
			reduce(117), // -, reduce: FullOosvar
			reduce(117), // .+, reduce: FullOosvar
			reduce(117), // .-, reduce: FullOosvar
			reduce(117), // *, reduce: FullOosvar
			reduce(117), // /, reduce: FullOosvar
			reduce(117), // //, reduce: FullOosvar
			reduce(117), // %, reduce: FullOosvar
			reduce(117), // .*, reduce: FullOosvar
			reduce(117), // ./, reduce: FullOosvar
			reduce(117), // .//, reduce: FullOosvar
			reduce(117), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(117), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(117), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(117), // [[, reduce: FullOosvar
			reduce(117), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S781
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(118), // >, reduce: LocalVariable
			reduce(118), // >>, reduce: LocalVariable
			reduce(118), // |, reduce: LocalVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(118), // ,, reduce: LocalVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(270), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(118), // ], reduce: LocalVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(118), // ?, reduce: LocalVariable
			nil,         // :
			reduce(118), // ||, reduce: LocalVariable
			reduce(118), // ^^, reduce: LocalVariable
			reduce(118), // &&, reduce: LocalVariable
			reduce(118), // ??, reduce: LocalVariable
			reduce(118), // ???, reduce: LocalVariable
			reduce(118), // =~, reduce: LocalVariable
			reduce(118), // !=~, reduce: LocalVariable
			reduce(118), // ==, reduce: LocalVariable
			reduce(118), // !=, reduce: LocalVariable
			reduce(118), // <=>, reduce: LocalVariable
			reduce(118), // >=, reduce: LocalVariable
			reduce(118), // <, reduce: LocalVariable
			reduce(118), // <=, reduce: LocalVariable
			reduce(118), // ^, reduce: LocalVariable
			reduce(118), // &, reduce: LocalVariable
			reduce(118), // <<, reduce: LocalVariable
			reduce(118), // >>>, reduce: LocalVariable
			reduce(118), // +, reduce: LocalVariable
			reduce(118), // -, reduce: LocalVariable
			reduce(118), // .+, reduce: LocalVariable
			reduce(118), // .-, reduce: LocalVariable
			reduce(118), // *, reduce: LocalVariable
			reduce(118), // /, reduce: LocalVariable
			reduce(118), // //, reduce: LocalVariable
			reduce(118), // %, reduce: LocalVariable
			reduce(118), // .*, reduce: LocalVariable
			reduce(118), // ./, reduce: LocalVariable
			reduce(118), // .//, reduce: LocalVariable
			reduce(118), // ., reduce: LocalVariable
			nil,         // !
			nil,         // ~
			reduce(118), // **, reduce: LocalVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(118), // [, reduce: LocalVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(118), // [[, reduce: LocalVariable
			reduce(118), // [[[, reduce: LocalVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S782
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(147), // ,, reduce: Rvalue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(147), // ], reduce: Rvalue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S783
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(148), // ,, reduce: PrecedenceChainStart
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(148), // ], reduce: PrecedenceChainStart
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S784
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(150), // ,, reduce: TernaryTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(150), // ], reduce: TernaryTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			shift(1611), // ?
			nil,         // :
			shift(1612), // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S785
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(152), // ,, reduce: LogicalOrTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(152), // ], reduce: LogicalOrTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(152), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(152), // ||, reduce: LogicalOrTerm
			shift(1613), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S786
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(154), // ,, reduce: LogicalXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(154), // ], reduce: LogicalXORTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(154), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(154), // ||, reduce: LogicalXORTerm
			reduce(154), // ^^, reduce: LogicalXORTerm
			shift(1614), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S787
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(156), // ,, reduce: LogicalAndTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(156), // ], reduce: LogicalAndTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(156), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(156), // ||, reduce: LogicalAndTerm
			reduce(156), // ^^, reduce: LogicalAndTerm
			reduce(156), // &&, reduce: LogicalAndTerm
			shift(1615), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S788
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(158), // ,, reduce: AbsentCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(158), // ], reduce: AbsentCoalesceTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(158), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(158), // ||, reduce: AbsentCoalesceTerm
			reduce(158), // ^^, reduce: AbsentCoalesceTerm
			reduce(158), // &&, reduce: AbsentCoalesceTerm
			reduce(158), // ??, reduce: AbsentCoalesceTerm
			shift(1616), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S789
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(160), // ,, reduce: EmptyCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(160), // ], reduce: EmptyCoalesceTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(160), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(160), // ||, reduce: EmptyCoalesceTerm
			reduce(160), // ^^, reduce: EmptyCoalesceTerm
			reduce(160), // &&, reduce: EmptyCoalesceTerm
			reduce(160), // ??, reduce: EmptyCoalesceTerm
			reduce(160), // ???, reduce: EmptyCoalesceTerm
			shift(1617), // =~
			shift(1618), // !=~
			shift(1619), // ==
			shift(1620), // !=
			shift(1621), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S790
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1622), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(166), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(166), // ], reduce: EqneTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(166), // ?, reduce: EqneTerm
			nil,         // :
			reduce(166), // ||, reduce: EqneTerm
			reduce(166), // ^^, reduce: EqneTerm
			reduce(166), // &&, reduce: EqneTerm
			reduce(166), // ??, reduce: EqneTerm
			reduce(166), // ???, reduce: EqneTerm
			reduce(166), // =~, reduce: EqneTerm
			reduce(166), // !=~, reduce: EqneTerm
			reduce(166), // ==, reduce: EqneTerm
			reduce(166), // !=, reduce: EqneTerm
			reduce(166), // <=>, reduce: EqneTerm
			shift(1623), // >=
			shift(1624), // <
			shift(1625), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S791
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(171), // >, reduce: CmpTerm
			nil,         // >>
			shift(1626), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(171), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(171), // ], reduce: CmpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(171), // ?, reduce: CmpTerm
			nil,         // :
			reduce(171), // ||, reduce: CmpTerm
			reduce(171), // ^^, reduce: CmpTerm
			reduce(171), // &&, reduce: CmpTerm
			reduce(171), // ??, reduce: CmpTerm
			reduce(171), // ???, reduce: CmpTerm
			reduce(171), // =~, reduce: CmpTerm
			reduce(171), // !=~, reduce: CmpTerm
			reduce(171), // ==, reduce: CmpTerm
			reduce(171), // !=, reduce: CmpTerm
			reduce(171), // <=>, reduce: CmpTerm
			reduce(171), // >=, reduce: CmpTerm
			reduce(171), // <, reduce: CmpTerm
			reduce(171), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S792
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(173), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(173), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(173), // ,, reduce: BitwiseORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(173), // ], reduce: BitwiseORTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(173), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(173), // ||, reduce: BitwiseORTerm
			reduce(173), // ^^, reduce: BitwiseORTerm
			reduce(173), // &&, reduce: BitwiseORTerm
			reduce(173), // ??, reduce: BitwiseORTerm
			reduce(173), // ???, reduce: BitwiseORTerm
			reduce(173), // =~, reduce: BitwiseORTerm
			reduce(173), // !=~, reduce: BitwiseORTerm
			reduce(173), // ==, reduce: BitwiseORTerm
			reduce(173), // !=, reduce: BitwiseORTerm
			reduce(173), // <=>, reduce: BitwiseORTerm
			reduce(173), // >=, reduce: BitwiseORTerm
			reduce(173), // <, reduce: BitwiseORTerm
			reduce(173), // <=, reduce: BitwiseORTerm
			shift(1627), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S793
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(175), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(175), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(175), // ,, reduce: BitwiseXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(175), // ], reduce: BitwiseXORTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(175), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(175), // ||, reduce: BitwiseXORTerm
			reduce(175), // ^^, reduce: BitwiseXORTerm
			reduce(175), // &&, reduce: BitwiseXORTerm
			reduce(175), // ??, reduce: BitwiseXORTerm
			reduce(175), // ???, reduce: BitwiseXORTerm
			reduce(175), // =~, reduce: BitwiseXORTerm
			reduce(175), // !=~, reduce: BitwiseXORTerm
			reduce(175), // ==, reduce: BitwiseXORTerm
			reduce(175), // !=, reduce: BitwiseXORTerm
			reduce(175), // <=>, reduce: BitwiseXORTerm
			reduce(175), // >=, reduce: BitwiseXORTerm
			reduce(175), // <, reduce: BitwiseXORTerm
			reduce(175), // <=, reduce: BitwiseXORTerm
			reduce(175), // ^, reduce: BitwiseXORTerm
			shift(1628), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S794
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(177), // >, reduce: BitwiseANDTerm
			shift(1629), // >>
			reduce(177), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(177), // ,, reduce: BitwiseANDTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(177), // ], reduce: BitwiseANDTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(177), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(177), // ||, reduce: BitwiseANDTerm
			reduce(177), // ^^, reduce: BitwiseANDTerm
			reduce(177), // &&, reduce: BitwiseANDTerm
			reduce(177), // ??, reduce: BitwiseANDTerm
			reduce(177), // ???, reduce: BitwiseANDTerm
			reduce(177), // =~, reduce: BitwiseANDTerm
			reduce(177), // !=~, reduce: BitwiseANDTerm
			reduce(177), // ==, reduce: BitwiseANDTerm
			reduce(177), // !=, reduce: BitwiseANDTerm
			reduce(177), // <=>, reduce: BitwiseANDTerm
			reduce(177), // >=, reduce: BitwiseANDTerm
			reduce(177), // <, reduce: BitwiseANDTerm
			reduce(177), // <=, reduce: BitwiseANDTerm
			reduce(177), // ^, reduce: BitwiseANDTerm
			reduce(177), // &, reduce: BitwiseANDTerm
			shift(1630), // <<
			shift(1631), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S795
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(181), // >, reduce: BitwiseShiftTerm
			reduce(181), // >>, reduce: BitwiseShiftTerm
			reduce(181), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(181), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(181), // ], reduce: BitwiseShiftTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(181), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(181), // ||, reduce: BitwiseShiftTerm
			reduce(181), // ^^, reduce: BitwiseShiftTerm
			reduce(181), // &&, reduce: BitwiseShiftTerm
			reduce(181), // ??, reduce: BitwiseShiftTerm
			reduce(181), // ???, reduce: BitwiseShiftTerm
			reduce(181), // =~, reduce: BitwiseShiftTerm
			reduce(181), // !=~, reduce: BitwiseShiftTerm
			reduce(181), // ==, reduce: BitwiseShiftTerm
			reduce(181), // !=, reduce: BitwiseShiftTerm
			reduce(181), // <=>, reduce: BitwiseShiftTerm
			reduce(181), // >=, reduce: BitwiseShiftTerm
			reduce(181), // <, reduce: BitwiseShiftTerm
			reduce(181), // <=, reduce: BitwiseShiftTerm
			reduce(181), // ^, reduce: BitwiseShiftTerm
			reduce(181), // &, reduce: BitwiseShiftTerm
			reduce(181), // <<, reduce: BitwiseShiftTerm
			reduce(181), // >>>, reduce: BitwiseShiftTerm
			shift(1632), // +
			shift(1633), // -
			shift(1634), // .+
			shift(1635), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S796
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S797
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(186), // >, reduce: AddsubdotTerm
			reduce(186), // >>, reduce: AddsubdotTerm
			reduce(186), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(186), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(186), // ], reduce: AddsubdotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(186), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(186), // ||, reduce: AddsubdotTerm
			reduce(186), // ^^, reduce: AddsubdotTerm
			reduce(186), // &&, reduce: AddsubdotTerm
			reduce(186), // ??, reduce: AddsubdotTerm
			reduce(186), // ???, reduce: AddsubdotTerm
			reduce(186), // =~, reduce: AddsubdotTerm
			reduce(186), // !=~, reduce: AddsubdotTerm
			reduce(186), // ==, reduce: AddsubdotTerm
			reduce(186), // !=, reduce: AddsubdotTerm
			reduce(186), // <=>, reduce: AddsubdotTerm
			reduce(186), // >=, reduce: AddsubdotTerm
			reduce(186), // <, reduce: AddsubdotTerm
			reduce(186), // <=, reduce: AddsubdotTerm
			reduce(186), // ^, reduce: AddsubdotTerm
			reduce(186), // &, reduce: AddsubdotTerm
			reduce(186), // <<, reduce: AddsubdotTerm
			reduce(186), // >>>, reduce: AddsubdotTerm
			reduce(186), // +, reduce: AddsubdotTerm
			reduce(186), // -, reduce: AddsubdotTerm
			reduce(186), // .+, reduce: AddsubdotTerm
			reduce(186), // .-, reduce: AddsubdotTerm
			shift(1637), // *
			shift(1638), // /
			shift(1639), // //
			shift(1640), // %
			shift(1641), // .*
			shift(1642), // ./
			shift(1643), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S798
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S799
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S800
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S801
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(194), // >, reduce: MuldivTerm
			reduce(194), // >>, reduce: MuldivTerm
			reduce(194), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(194), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(194), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(194), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(194), // ||, reduce: MuldivTerm
			reduce(194), // ^^, reduce: MuldivTerm
			reduce(194), // &&, reduce: MuldivTerm
			reduce(194), // ??, reduce: MuldivTerm
			reduce(194), // ???, reduce: MuldivTerm
			reduce(194), // =~, reduce: MuldivTerm
			reduce(194), // !=~, reduce: MuldivTerm
			reduce(194), // ==, reduce: MuldivTerm
			reduce(194), // !=, reduce: MuldivTerm
			reduce(194), // <=>, reduce: MuldivTerm
			reduce(194), // >=, reduce: MuldivTerm
			reduce(194), // <, reduce: MuldivTerm
			reduce(194), // <=, reduce: MuldivTerm
			reduce(194), // ^, reduce: MuldivTerm
			reduce(194), // &, reduce: MuldivTerm
			reduce(194), // <<, reduce: MuldivTerm
			reduce(194), // >>>, reduce: MuldivTerm
			reduce(194), // +, reduce: MuldivTerm
			reduce(194), // -, reduce: MuldivTerm
			reduce(194), // .+, reduce: MuldivTerm
			reduce(194), // .-, reduce: MuldivTerm
			reduce(194), // *, reduce: MuldivTerm
			reduce(194), // /, reduce: MuldivTerm
			reduce(194), // //, reduce: MuldivTerm
			reduce(194), // %, reduce: MuldivTerm
			reduce(194), // .*, reduce: MuldivTerm
			reduce(194), // ./, reduce: MuldivTerm
			reduce(194), // .//, reduce: MuldivTerm
			shift(1647), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S802
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(196), // >, reduce: DotTerm
			reduce(196), // >>, reduce: DotTerm
			reduce(196), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(196), // ,, reduce: DotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(196), // ], reduce: DotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(196), // ?, reduce: DotTerm
			nil,         // :
			reduce(196), // ||, reduce: DotTerm
			reduce(196), // ^^, reduce: DotTerm
			reduce(196), // &&, reduce: DotTerm
			reduce(196), // ??, reduce: DotTerm
			reduce(196), // ???, reduce: DotTerm
			reduce(196), // =~, reduce: DotTerm
			reduce(196), // !=~, reduce: DotTerm
			reduce(196), // ==, reduce: DotTerm
			reduce(196), // !=, reduce: DotTerm
			reduce(196), // <=>, reduce: DotTerm
			reduce(196), // >=, reduce: DotTerm
			reduce(196), // <, reduce: DotTerm
			reduce(196), // <=, reduce: DotTerm
			reduce(196), // ^, reduce: DotTerm
			reduce(196), // &, reduce: DotTerm
			reduce(196), // <<, reduce: DotTerm
			reduce(196), // >>>, reduce: DotTerm
			reduce(196), // +, reduce: DotTerm
			reduce(196), // -, reduce: DotTerm
			reduce(196), // .+, reduce: DotTerm
			reduce(196), // .-, reduce: DotTerm
			reduce(196), // *, reduce: DotTerm
			reduce(196), // /, reduce: DotTerm
			reduce(196), // //, reduce: DotTerm
			reduce(196), // %, reduce: DotTerm
			reduce(196), // .*, reduce: DotTerm
			reduce(196), // ./, reduce: DotTerm
			reduce(196), // .//, reduce: DotTerm
			reduce(196), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S803
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S804
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S805
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(203), // >, reduce: UnaryOpTerm
			reduce(203), // >>, reduce: UnaryOpTerm
			reduce(203), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(203), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(203), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(203), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(203), // ||, reduce: UnaryOpTerm
			reduce(203), // ^^, reduce: UnaryOpTerm
			reduce(203), // &&, reduce: UnaryOpTerm
			reduce(203), // ??, reduce: UnaryOpTerm
			reduce(203), // ???, reduce: UnaryOpTerm
			reduce(203), // =~, reduce: UnaryOpTerm
			reduce(203), // !=~, reduce: UnaryOpTerm
			reduce(203), // ==, reduce: UnaryOpTerm
			reduce(203), // !=, reduce: UnaryOpTerm
			reduce(203), // <=>, reduce: UnaryOpTerm
			reduce(203), // >=, reduce: UnaryOpTerm
			reduce(203), // <, reduce: UnaryOpTerm
			reduce(203), // <=, reduce: UnaryOpTerm
			reduce(203), // ^, reduce: UnaryOpTerm
			reduce(203), // &, reduce: UnaryOpTerm
			reduce(203), // <<, reduce: UnaryOpTerm
			reduce(203), // >>>, reduce: UnaryOpTerm
			reduce(203), // +, reduce: UnaryOpTerm
			reduce(203), // -, reduce: UnaryOpTerm
			reduce(203), // .+, reduce: UnaryOpTerm
			reduce(203), // .-, reduce: UnaryOpTerm
			reduce(203), // *, reduce: UnaryOpTerm
			reduce(203), // /, reduce: UnaryOpTerm
			reduce(203), // //, reduce: UnaryOpTerm
			reduce(203), // %, reduce: UnaryOpTerm
			reduce(203), // .*, reduce: UnaryOpTerm
			reduce(203), // ./, reduce: UnaryOpTerm
			reduce(203), // .//, reduce: UnaryOpTerm
			reduce(203), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S806
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(207), // >, reduce: PowTerm
			reduce(207), // >>, reduce: PowTerm
			reduce(207), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(207), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(207), // ], reduce: PowTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(207), // ?, reduce: PowTerm
			nil,         // :
			reduce(207), // ||, reduce: PowTerm
			reduce(207), // ^^, reduce: PowTerm
			reduce(207), // &&, reduce: PowTerm
			reduce(207), // ??, reduce: PowTerm
			reduce(207), // ???, reduce: PowTerm
			reduce(207), // =~, reduce: PowTerm
			reduce(207), // !=~, reduce: PowTerm
			reduce(207), // ==, reduce: PowTerm
			reduce(207), // !=, reduce: PowTerm
			reduce(207), // <=>, reduce: PowTerm
			reduce(207), // >=, reduce: PowTerm
			reduce(207), // <, reduce: PowTerm
			reduce(207), // <=, reduce: PowTerm
			reduce(207), // ^, reduce: PowTerm
			reduce(207), // &, reduce: PowTerm
			reduce(207), // <<, reduce: PowTerm
			reduce(207), // >>>, reduce: PowTerm
			reduce(207), // +, reduce: PowTerm
			reduce(207), // -, reduce: PowTerm
			reduce(207), // .+, reduce: PowTerm
			reduce(207), // .-, reduce: PowTerm
			reduce(207), // *, reduce: PowTerm
			reduce(207), // /, reduce: PowTerm
			reduce(207), // //, reduce: PowTerm
			reduce(207), // %, reduce: PowTerm
			reduce(207), // .*, reduce: PowTerm
			reduce(207), // ./, reduce: PowTerm
			reduce(207), // .//, reduce: PowTerm
			reduce(207), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			shift(1650), // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S807
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(209), // >, reduce: PrecedenceChainEnd
			reduce(209), // >>, reduce: PrecedenceChainEnd
			reduce(209), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(209), // ,, reduce: PrecedenceChainEnd
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(209), // ], reduce: PrecedenceChainEnd
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(209), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(209), // ||, reduce: PrecedenceChainEnd
			reduce(209), // ^^, reduce: PrecedenceChainEnd
			reduce(209), // &&, reduce: PrecedenceChainEnd
			reduce(209), // ??, reduce: PrecedenceChainEnd
			reduce(209), // ???, reduce: PrecedenceChainEnd
			reduce(209), // =~, reduce: PrecedenceChainEnd
			reduce(209), // !=~, reduce: PrecedenceChainEnd
			reduce(209), // ==, reduce: PrecedenceChainEnd
			reduce(209), // !=, reduce: PrecedenceChainEnd
			reduce(209), // <=>, reduce: PrecedenceChainEnd
			reduce(209), // >=, reduce: PrecedenceChainEnd
			reduce(209), // <, reduce: PrecedenceChainEnd
			reduce(209), // <=, reduce: PrecedenceChainEnd
			reduce(209), // ^, reduce: PrecedenceChainEnd
			reduce(209), // &, reduce: PrecedenceChainEnd
			reduce(209), // <<, reduce: PrecedenceChainEnd
			reduce(209), // >>>, reduce: PrecedenceChainEnd
			reduce(209), // +, reduce: PrecedenceChainEnd
			reduce(209), // -, reduce: PrecedenceChainEnd
			reduce(209), // .+, reduce: PrecedenceChainEnd
			reduce(209), // .-, reduce: PrecedenceChainEnd
			reduce(209), // *, reduce: PrecedenceChainEnd
			reduce(209), // /, reduce: PrecedenceChainEnd
			reduce(209), // //, reduce: PrecedenceChainEnd
			reduce(209), // %, reduce: PrecedenceChainEnd
			reduce(209), // .*, reduce: PrecedenceChainEnd
			reduce(209), // ./, reduce: PrecedenceChainEnd
			reduce(209), // .//, reduce: PrecedenceChainEnd
			reduce(209), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(209), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(1651), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			shift(1652), // [[
			shift(1653), // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S808
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(215), // >, reduce: MlrvalOrFunction
			reduce(215), // >>, reduce: MlrvalOrFunction
			reduce(215), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(215), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(215), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(215), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(215), // ||, reduce: MlrvalOrFunction
			reduce(215), // ^^, reduce: MlrvalOrFunction
			reduce(215), // &&, reduce: MlrvalOrFunction
			reduce(215), // ??, reduce: MlrvalOrFunction
			reduce(215), // ???, reduce: MlrvalOrFunction
			reduce(215), // =~, reduce: MlrvalOrFunction
			reduce(215), // !=~, reduce: MlrvalOrFunction
			reduce(215), // ==, reduce: MlrvalOrFunction
			reduce(215), // !=, reduce: MlrvalOrFunction
			reduce(215), // <=>, reduce: MlrvalOrFunction
			reduce(215), // >=, reduce: MlrvalOrFunction
			reduce(215), // <, reduce: MlrvalOrFunction
			reduce(215), // <=, reduce: MlrvalOrFunction
			reduce(215), // ^, reduce: MlrvalOrFunction
			reduce(215), // &, reduce: MlrvalOrFunction
			reduce(215), // <<, reduce: MlrvalOrFunction
			reduce(215), // >>>, reduce: MlrvalOrFunction
			reduce(215), // +, reduce: MlrvalOrFunction
			reduce(215), // -, reduce: MlrvalOrFunction
			reduce(215), // .+, reduce: MlrvalOrFunction
			reduce(215), // .-, reduce: MlrvalOrFunction
			reduce(215), // *, reduce: MlrvalOrFunction
			reduce(215), // /, reduce: MlrvalOrFunction
			reduce(215), // //, reduce: MlrvalOrFunction
			reduce(215), // %, reduce: MlrvalOrFunction
			reduce(215), // .*, reduce: MlrvalOrFunction
			reduce(215), // ./, reduce: MlrvalOrFunction
			reduce(215), // .//, reduce: MlrvalOrFunction
			reduce(215), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(215), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(215), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(215), // [[, reduce: MlrvalOrFunction
			reduce(215), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S809
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(216), // >, reduce: MlrvalOrFunction
			reduce(216), // >>, reduce: MlrvalOrFunction
			reduce(216), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(216), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(216), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(216), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(216), // ||, reduce: MlrvalOrFunction
			reduce(216), // ^^, reduce: MlrvalOrFunction
			reduce(216), // &&, reduce: MlrvalOrFunction
			reduce(216), // ??, reduce: MlrvalOrFunction
			reduce(216), // ???, reduce: MlrvalOrFunction
			reduce(216), // =~, reduce: MlrvalOrFunction
			reduce(216), // !=~, reduce: MlrvalOrFunction
			reduce(216), // ==, reduce: MlrvalOrFunction
			reduce(216), // !=, reduce: MlrvalOrFunction
			reduce(216), // <=>, reduce: MlrvalOrFunction
			reduce(216), // >=, reduce: MlrvalOrFunction
			reduce(216), // <, reduce: MlrvalOrFunction
			reduce(216), // <=, reduce: MlrvalOrFunction
			reduce(216), // ^, reduce: MlrvalOrFunction
			reduce(216), // &, reduce: MlrvalOrFunction
			reduce(216), // <<, reduce: MlrvalOrFunction
			reduce(216), // >>>, reduce: MlrvalOrFunction
			reduce(216), // +, reduce: MlrvalOrFunction
			reduce(216), // -, reduce: MlrvalOrFunction
			reduce(216), // .+, reduce: MlrvalOrFunction
			reduce(216), // .-, reduce: MlrvalOrFunction
			reduce(216), // *, reduce: MlrvalOrFunction
			reduce(216), // /, reduce: MlrvalOrFunction
			reduce(216), // //, reduce: MlrvalOrFunction
			reduce(216), // %, reduce: MlrvalOrFunction
			reduce(216), // .*, reduce: MlrvalOrFunction
			reduce(216), // ./, reduce: MlrvalOrFunction
			reduce(216), // .//, reduce: MlrvalOrFunction
			reduce(216), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(216), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(216), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(216), // [[, reduce: MlrvalOrFunction
			reduce(216), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S810
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(217), // >, reduce: MlrvalOrFunction
			reduce(217), // >>, reduce: MlrvalOrFunction
			reduce(217), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(217), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(217), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(217), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(217), // ||, reduce: MlrvalOrFunction
			reduce(217), // ^^, reduce: MlrvalOrFunction
			reduce(217), // &&, reduce: MlrvalOrFunction
			reduce(217), // ??, reduce: MlrvalOrFunction
			reduce(217), // ???, reduce: MlrvalOrFunction
			reduce(217), // =~, reduce: MlrvalOrFunction
			reduce(217), // !=~, reduce: MlrvalOrFunction
			reduce(217), // ==, reduce: MlrvalOrFunction
			reduce(217), // !=, reduce: MlrvalOrFunction
			reduce(217), // <=>, reduce: MlrvalOrFunction
			reduce(217), // >=, reduce: MlrvalOrFunction
			reduce(217), // <, reduce: MlrvalOrFunction
			reduce(217), // <=, reduce: MlrvalOrFunction
			reduce(217), // ^, reduce: MlrvalOrFunction
			reduce(217), // &, reduce: MlrvalOrFunction
			reduce(217), // <<, reduce: MlrvalOrFunction
			reduce(217), // >>>, reduce: MlrvalOrFunction
			reduce(217), // +, reduce: MlrvalOrFunction
			reduce(217), // -, reduce: MlrvalOrFunction
			reduce(217), // .+, reduce: MlrvalOrFunction
			reduce(217), // .-, reduce: MlrvalOrFunction
			reduce(217), // *, reduce: MlrvalOrFunction
			reduce(217), // /, reduce: MlrvalOrFunction
			reduce(217), // //, reduce: MlrvalOrFunction
			reduce(217), // %, reduce: MlrvalOrFunction
			reduce(217), // .*, reduce: MlrvalOrFunction
			reduce(217), // ./, reduce: MlrvalOrFunction
			reduce(217), // .//, reduce: MlrvalOrFunction
			reduce(217), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(217), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(217), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(217), // [[, reduce: MlrvalOrFunction
			reduce(217), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S811
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(218), // >, reduce: MlrvalOrFunction
			reduce(218), // >>, reduce: MlrvalOrFunction
			reduce(218), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(218), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(218), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(218), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(218), // ||, reduce: MlrvalOrFunction
			reduce(218), // ^^, reduce: MlrvalOrFunction
			reduce(218), // &&, reduce: MlrvalOrFunction
			reduce(218), // ??, reduce: MlrvalOrFunction
			reduce(218), // ???, reduce: MlrvalOrFunction
			reduce(218), // =~, reduce: MlrvalOrFunction
			reduce(218), // !=~, reduce: MlrvalOrFunction
			reduce(218), // ==, reduce: MlrvalOrFunction
			reduce(218), // !=, reduce: MlrvalOrFunction
			reduce(218), // <=>, reduce: MlrvalOrFunction
			reduce(218), // >=, reduce: MlrvalOrFunction
			reduce(218), // <, reduce: MlrvalOrFunction
			reduce(218), // <=, reduce: MlrvalOrFunction
			reduce(218), // ^, reduce: MlrvalOrFunction
			reduce(218), // &, reduce: MlrvalOrFunction
			reduce(218), // <<, reduce: MlrvalOrFunction
			reduce(218), // >>>, reduce: MlrvalOrFunction
			reduce(218), // +, reduce: MlrvalOrFunction
			reduce(218), // -, reduce: MlrvalOrFunction
			reduce(218), // .+, reduce: MlrvalOrFunction
			reduce(218), // .-, reduce: MlrvalOrFunction
			reduce(218), // *, reduce: MlrvalOrFunction
			reduce(218), // /, reduce: MlrvalOrFunction
			reduce(218), // //, reduce: MlrvalOrFunction
			reduce(218), // %, reduce: MlrvalOrFunction
			reduce(218), // .*, reduce: MlrvalOrFunction
			reduce(218), // ./, reduce: MlrvalOrFunction
			reduce(218), // .//, reduce: MlrvalOrFunction
			reduce(218), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(218), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(218), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(218), // [[, reduce: MlrvalOrFunction
			reduce(218), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S812
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(219), // >, reduce: MlrvalOrFunction
			reduce(219), // >>, reduce: MlrvalOrFunction
			reduce(219), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(219), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(219), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(219), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(219), // ||, reduce: MlrvalOrFunction
			reduce(219), // ^^, reduce: MlrvalOrFunction
			reduce(219), // &&, reduce: MlrvalOrFunction
			reduce(219), // ??, reduce: MlrvalOrFunction
			reduce(219), // ???, reduce: MlrvalOrFunction
			reduce(219), // =~, reduce: MlrvalOrFunction
			reduce(219), // !=~, reduce: MlrvalOrFunction
			reduce(219), // ==, reduce: MlrvalOrFunction
			reduce(219), // !=, reduce: MlrvalOrFunction
			reduce(219), // <=>, reduce: MlrvalOrFunction
			reduce(219), // >=, reduce: MlrvalOrFunction
			reduce(219), // <, reduce: MlrvalOrFunction
			reduce(219), // <=, reduce: MlrvalOrFunction
			reduce(219), // ^, reduce: MlrvalOrFunction
			reduce(219), // &, reduce: MlrvalOrFunction
			reduce(219), // <<, reduce: MlrvalOrFunction
			reduce(219), // >>>, reduce: MlrvalOrFunction
			reduce(219), // +, reduce: MlrvalOrFunction
			reduce(219), // -, reduce: MlrvalOrFunction
			reduce(219), // .+, reduce: MlrvalOrFunction
			reduce(219), // .-, reduce: MlrvalOrFunction
			reduce(219), // *, reduce: MlrvalOrFunction
			reduce(219), // /, reduce: MlrvalOrFunction
			reduce(219), // //, reduce: MlrvalOrFunction
			reduce(219), // %, reduce: MlrvalOrFunction
			reduce(219), // .*, reduce: MlrvalOrFunction
			reduce(219), // ./, reduce: MlrvalOrFunction
			reduce(219), // .//, reduce: MlrvalOrFunction
			reduce(219), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(219), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(219), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(219), // [[, reduce: MlrvalOrFunction
			reduce(219), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S813
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(220), // >, reduce: MlrvalOrFunction
			reduce(220), // >>, reduce: MlrvalOrFunction
			reduce(220), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(220), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(220), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(220), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(220), // ||, reduce: MlrvalOrFunction
			reduce(220), // ^^, reduce: MlrvalOrFunction
			reduce(220), // &&, reduce: MlrvalOrFunction
			reduce(220), // ??, reduce: MlrvalOrFunction
			reduce(220), // ???, reduce: MlrvalOrFunction
			reduce(220), // =~, reduce: MlrvalOrFunction
			reduce(220), // !=~, reduce: MlrvalOrFunction
			reduce(220), // ==, reduce: MlrvalOrFunction
			reduce(220), // !=, reduce: MlrvalOrFunction
			reduce(220), // <=>, reduce: MlrvalOrFunction
			reduce(220), // >=, reduce: MlrvalOrFunction
			reduce(220), // <, reduce: MlrvalOrFunction
			reduce(220), // <=, reduce: MlrvalOrFunction
			reduce(220), // ^, reduce: MlrvalOrFunction
			reduce(220), // &, reduce: MlrvalOrFunction
			reduce(220), // <<, reduce: MlrvalOrFunction
			reduce(220), // >>>, reduce: MlrvalOrFunction
			reduce(220), // +, reduce: MlrvalOrFunction
			reduce(220), // -, reduce: MlrvalOrFunction
			reduce(220), // .+, reduce: MlrvalOrFunction
			reduce(220), // .-, reduce: MlrvalOrFunction
			reduce(220), // *, reduce: MlrvalOrFunction
			reduce(220), // /, reduce: MlrvalOrFunction
			reduce(220), // //, reduce: MlrvalOrFunction
			reduce(220), // %, reduce: MlrvalOrFunction
			reduce(220), // .*, reduce: MlrvalOrFunction
			reduce(220), // ./, reduce: MlrvalOrFunction
			reduce(220), // .//, reduce: MlrvalOrFunction
			reduce(220), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(220), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(220), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(220), // [[, reduce: MlrvalOrFunction
			reduce(220), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S814
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(221), // >, reduce: MlrvalOrFunction
			reduce(221), // >>, reduce: MlrvalOrFunction
			reduce(221), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(221), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(221), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(221), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(221), // ||, reduce: MlrvalOrFunction
			reduce(221), // ^^, reduce: MlrvalOrFunction
			reduce(221), // &&, reduce: MlrvalOrFunction
			reduce(221), // ??, reduce: MlrvalOrFunction
			reduce(221), // ???, reduce: MlrvalOrFunction
			reduce(221), // =~, reduce: MlrvalOrFunction
			reduce(221), // !=~, reduce: MlrvalOrFunction
			reduce(221), // ==, reduce: MlrvalOrFunction
			reduce(221), // !=, reduce: MlrvalOrFunction
			reduce(221), // <=>, reduce: MlrvalOrFunction
			reduce(221), // >=, reduce: MlrvalOrFunction
			reduce(221), // <, reduce: MlrvalOrFunction
			reduce(221), // <=, reduce: MlrvalOrFunction
			reduce(221), // ^, reduce: MlrvalOrFunction
			reduce(221), // &, reduce: MlrvalOrFunction
			reduce(221), // <<, reduce: MlrvalOrFunction
			reduce(221), // >>>, reduce: MlrvalOrFunction
			reduce(221), // +, reduce: MlrvalOrFunction
			reduce(221), // -, reduce: MlrvalOrFunction
			reduce(221), // .+, reduce: MlrvalOrFunction
			reduce(221), // .-, reduce: MlrvalOrFunction
			reduce(221), // *, reduce: MlrvalOrFunction
			reduce(221), // /, reduce: MlrvalOrFunction
			reduce(221), // //, reduce: MlrvalOrFunction
			reduce(221), // %, reduce: MlrvalOrFunction
			reduce(221), // .*, reduce: MlrvalOrFunction
			reduce(221), // ./, reduce: MlrvalOrFunction
			reduce(221), // .//, reduce: MlrvalOrFunction
			reduce(221), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(221), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(221), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(221), // [[, reduce: MlrvalOrFunction
			reduce(221), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S815
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(222), // >, reduce: MlrvalOrFunction
			reduce(222), // >>, reduce: MlrvalOrFunction
			reduce(222), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(222), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(222), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(222), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(222), // ||, reduce: MlrvalOrFunction
			reduce(222), // ^^, reduce: MlrvalOrFunction
			reduce(222), // &&, reduce: MlrvalOrFunction
			reduce(222), // ??, reduce: MlrvalOrFunction
			reduce(222), // ???, reduce: MlrvalOrFunction
			reduce(222), // =~, reduce: MlrvalOrFunction
			reduce(222), // !=~, reduce: MlrvalOrFunction
			reduce(222), // ==, reduce: MlrvalOrFunction
			reduce(222), // !=, reduce: MlrvalOrFunction
			reduce(222), // <=>, reduce: MlrvalOrFunction
			reduce(222), // >=, reduce: MlrvalOrFunction
			reduce(222), // <, reduce: MlrvalOrFunction
			reduce(222), // <=, reduce: MlrvalOrFunction
			reduce(222), // ^, reduce: MlrvalOrFunction
			reduce(222), // &, reduce: MlrvalOrFunction
			reduce(222), // <<, reduce: MlrvalOrFunction
			reduce(222), // >>>, reduce: MlrvalOrFunction
			reduce(222), // +, reduce: MlrvalOrFunction
			reduce(222), // -, reduce: MlrvalOrFunction
			reduce(222), // .+, reduce: MlrvalOrFunction
			reduce(222), // .-, reduce: MlrvalOrFunction
			reduce(222), // *, reduce: MlrvalOrFunction
			reduce(222), // /, reduce: MlrvalOrFunction
			reduce(222), // //, reduce: MlrvalOrFunction
			reduce(222), // %, reduce: MlrvalOrFunction
			reduce(222), // .*, reduce: MlrvalOrFunction
			reduce(222), // ./, reduce: MlrvalOrFunction
			reduce(222), // .//, reduce: MlrvalOrFunction
			reduce(222), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(222), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(222), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(222), // [[, reduce: MlrvalOrFunction
			reduce(222), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S816
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(223), // >, reduce: MlrvalOrFunction
			reduce(223), // >>, reduce: MlrvalOrFunction
			reduce(223), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(223), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(223), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(223), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(223), // ||, reduce: MlrvalOrFunction
			reduce(223), // ^^, reduce: MlrvalOrFunction
			reduce(223), // &&, reduce: MlrvalOrFunction
			reduce(223), // ??, reduce: MlrvalOrFunction
			reduce(223), // ???, reduce: MlrvalOrFunction
			reduce(223), // =~, reduce: MlrvalOrFunction
			reduce(223), // !=~, reduce: MlrvalOrFunction
			reduce(223), // ==, reduce: MlrvalOrFunction
			reduce(223), // !=, reduce: MlrvalOrFunction
			reduce(223), // <=>, reduce: MlrvalOrFunction
			reduce(223), // >=, reduce: MlrvalOrFunction
			reduce(223), // <, reduce: MlrvalOrFunction
			reduce(223), // <=, reduce: MlrvalOrFunction
			reduce(223), // ^, reduce: MlrvalOrFunction
			reduce(223), // &, reduce: MlrvalOrFunction
			reduce(223), // <<, reduce: MlrvalOrFunction
			reduce(223), // >>>, reduce: MlrvalOrFunction
			reduce(223), // +, reduce: MlrvalOrFunction
			reduce(223), // -, reduce: MlrvalOrFunction
			reduce(223), // .+, reduce: MlrvalOrFunction
			reduce(223), // .-, reduce: MlrvalOrFunction
			reduce(223), // *, reduce: MlrvalOrFunction
			reduce(223), // /, reduce: MlrvalOrFunction
			reduce(223), // //, reduce: MlrvalOrFunction
			reduce(223), // %, reduce: MlrvalOrFunction
			reduce(223), // .*, reduce: MlrvalOrFunction
			reduce(223), // ./, reduce: MlrvalOrFunction
			reduce(223), // .//, reduce: MlrvalOrFunction
			reduce(223), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(223), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(223), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(223), // [[, reduce: MlrvalOrFunction
			reduce(223), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S817
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(224), // >, reduce: MlrvalOrFunction
			reduce(224), // >>, reduce: MlrvalOrFunction
			reduce(224), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(224), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(224), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(224), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(224), // ||, reduce: MlrvalOrFunction
			reduce(224), // ^^, reduce: MlrvalOrFunction
			reduce(224), // &&, reduce: MlrvalOrFunction
			reduce(224), // ??, reduce: MlrvalOrFunction
			reduce(224), // ???, reduce: MlrvalOrFunction
			reduce(224), // =~, reduce: MlrvalOrFunction
			reduce(224), // !=~, reduce: MlrvalOrFunction
			reduce(224), // ==, reduce: MlrvalOrFunction
			reduce(224), // !=, reduce: MlrvalOrFunction
			reduce(224), // <=>, reduce: MlrvalOrFunction
			reduce(224), // >=, reduce: MlrvalOrFunction
			reduce(224), // <, reduce: MlrvalOrFunction
			reduce(224), // <=, reduce: MlrvalOrFunction
			reduce(224), // ^, reduce: MlrvalOrFunction
			reduce(224), // &, reduce: MlrvalOrFunction
			reduce(224), // <<, reduce: MlrvalOrFunction
			reduce(224), // >>>, reduce: MlrvalOrFunction
			reduce(224), // +, reduce: MlrvalOrFunction
			reduce(224), // -, reduce: MlrvalOrFunction
			reduce(224), // .+, reduce: MlrvalOrFunction
			reduce(224), // .-, reduce: MlrvalOrFunction
			reduce(224), // *, reduce: MlrvalOrFunction
			reduce(224), // /, reduce: MlrvalOrFunction
			reduce(224), // //, reduce: MlrvalOrFunction
			reduce(224), // %, reduce: MlrvalOrFunction
			reduce(224), // .*, reduce: MlrvalOrFunction
			reduce(224), // ./, reduce: MlrvalOrFunction
			reduce(224), // .//, reduce: MlrvalOrFunction
			reduce(224), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(224), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(224), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(224), // [[, reduce: MlrvalOrFunction
			reduce(224), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S818
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(225), // >, reduce: MlrvalOrFunction
			reduce(225), // >>, reduce: MlrvalOrFunction
			reduce(225), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(225), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(225), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(225), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(225), // ||, reduce: MlrvalOrFunction
			reduce(225), // ^^, reduce: MlrvalOrFunction
			reduce(225), // &&, reduce: MlrvalOrFunction
			reduce(225), // ??, reduce: MlrvalOrFunction
			reduce(225), // ???, reduce: MlrvalOrFunction
			reduce(225), // =~, reduce: MlrvalOrFunction
			reduce(225), // !=~, reduce: MlrvalOrFunction
			reduce(225), // ==, reduce: MlrvalOrFunction
			reduce(225), // !=, reduce: MlrvalOrFunction
			reduce(225), // <=>, reduce: MlrvalOrFunction
			reduce(225), // >=, reduce: MlrvalOrFunction
			reduce(225), // <, reduce: MlrvalOrFunction
			reduce(225), // <=, reduce: MlrvalOrFunction
			reduce(225), // ^, reduce: MlrvalOrFunction
			reduce(225), // &, reduce: MlrvalOrFunction
			reduce(225), // <<, reduce: MlrvalOrFunction
			reduce(225), // >>>, reduce: MlrvalOrFunction
			reduce(225), // +, reduce: MlrvalOrFunction
			reduce(225), // -, reduce: MlrvalOrFunction
			reduce(225), // .+, reduce: MlrvalOrFunction
			reduce(225), // .-, reduce: MlrvalOrFunction
			reduce(225), // *, reduce: MlrvalOrFunction
			reduce(225), // /, reduce: MlrvalOrFunction
			reduce(225), // //, reduce: MlrvalOrFunction
			reduce(225), // %, reduce: MlrvalOrFunction
			reduce(225), // .*, reduce: MlrvalOrFunction
			reduce(225), // ./, reduce: MlrvalOrFunction
			reduce(225), // .//, reduce: MlrvalOrFunction
			reduce(225), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(225), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(225), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(225), // [[, reduce: MlrvalOrFunction
			reduce(225), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S819
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(226), // >, reduce: MlrvalOrFunction
			reduce(226), // >>, reduce: MlrvalOrFunction
			reduce(226), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(226), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(226), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(226), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(226), // ||, reduce: MlrvalOrFunction
			reduce(226), // ^^, reduce: MlrvalOrFunction
			reduce(226), // &&, reduce: MlrvalOrFunction
			reduce(226), // ??, reduce: MlrvalOrFunction
			reduce(226), // ???, reduce: MlrvalOrFunction
			reduce(226), // =~, reduce: MlrvalOrFunction
			reduce(226), // !=~, reduce: MlrvalOrFunction
			reduce(226), // ==, reduce: MlrvalOrFunction
			reduce(226), // !=, reduce: MlrvalOrFunction
			reduce(226), // <=>, reduce: MlrvalOrFunction
			reduce(226), // >=, reduce: MlrvalOrFunction
			reduce(226), // <, reduce: MlrvalOrFunction
			reduce(226), // <=, reduce: MlrvalOrFunction
			reduce(226), // ^, reduce: MlrvalOrFunction
			reduce(226), // &, reduce: MlrvalOrFunction
			reduce(226), // <<, reduce: MlrvalOrFunction
			reduce(226), // >>>, reduce: MlrvalOrFunction
			reduce(226), // +, reduce: MlrvalOrFunction
			reduce(226), // -, reduce: MlrvalOrFunction
			reduce(226), // .+, reduce: MlrvalOrFunction
			reduce(226), // .-, reduce: MlrvalOrFunction
			reduce(226), // *, reduce: MlrvalOrFunction
			reduce(226), // /, reduce: MlrvalOrFunction
			reduce(226), // //, reduce: MlrvalOrFunction
			reduce(226), // %, reduce: MlrvalOrFunction
			reduce(226), // .*, reduce: MlrvalOrFunction
			reduce(226), // ./, reduce: MlrvalOrFunction
			reduce(226), // .//, reduce: MlrvalOrFunction
			reduce(226), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(226), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(226), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(226), // [[, reduce: MlrvalOrFunction
			reduce(226), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S820
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(227), // >, reduce: MlrvalOrFunction
			reduce(227), // >>, reduce: MlrvalOrFunction
			reduce(227), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(227), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(227), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(227), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(227), // ||, reduce: MlrvalOrFunction
			reduce(227), // ^^, reduce: MlrvalOrFunction
			reduce(227), // &&, reduce: MlrvalOrFunction
			reduce(227), // ??, reduce: MlrvalOrFunction
			reduce(227), // ???, reduce: MlrvalOrFunction
			reduce(227), // =~, reduce: MlrvalOrFunction
			reduce(227), // !=~, reduce: MlrvalOrFunction
			reduce(227), // ==, reduce: MlrvalOrFunction
			reduce(227), // !=, reduce: MlrvalOrFunction
			reduce(227), // <=>, reduce: MlrvalOrFunction
			reduce(227), // >=, reduce: MlrvalOrFunction
			reduce(227), // <, reduce: MlrvalOrFunction
			reduce(227), // <=, reduce: MlrvalOrFunction
			reduce(227), // ^, reduce: MlrvalOrFunction
			reduce(227), // &, reduce: MlrvalOrFunction
			reduce(227), // <<, reduce: MlrvalOrFunction
			reduce(227), // >>>, reduce: MlrvalOrFunction
			reduce(227), // +, reduce: MlrvalOrFunction
			reduce(227), // -, reduce: MlrvalOrFunction
			reduce(227), // .+, reduce: MlrvalOrFunction
			reduce(227), // .-, reduce: MlrvalOrFunction
			reduce(227), // *, reduce: MlrvalOrFunction
			reduce(227), // /, reduce: MlrvalOrFunction
			reduce(227), // //, reduce: MlrvalOrFunction
			reduce(227), // %, reduce: MlrvalOrFunction
			reduce(227), // .*, reduce: MlrvalOrFunction
			reduce(227), // ./, reduce: MlrvalOrFunction
			reduce(227), // .//, reduce: MlrvalOrFunction
			reduce(227), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(227), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(227), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(227), // [[, reduce: MlrvalOrFunction
			reduce(227), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S821
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(752),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(756),  // (
			nil,         // )
			nil,         // emitp
			shift(768),  // field_name
			shift(769),  // $[
			shift(1654), // ]
			shift(771),  // braced_field_name
			shift(772),  // $[[
			shift(773),  // $[[[
			shift(774),  // full_srec
			shift(776),  // oosvar_name
			shift(777),  // @[
			shift(778),  // braced_oosvar_name
			shift(779),  // full_oosvar
			shift(780),  // all
			shift(781),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(796),  // +
			shift(798),  // -
			shift(799),  // .+
			shift(800),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(803),  // !
			shift(804),  // ~
			nil,         // **
			shift(809),  // string_literal
			shift(810),  // regex_case_insensitive
			shift(811),  // int_literal
			shift(812),  // float_literal
			shift(813),  // boolean_literal
			shift(814),  // null_literal
			shift(815),  // inf_literal
			shift(816),  // nan_literal
			shift(817),  // const_M_PI
			shift(818),  // const_M_E
			shift(819),  // panic
			shift(821),  // [
			shift(824),  // ctx_IPS
			shift(825),  // ctx_IFS
			shift(826),  // ctx_IRS
			shift(827),  // ctx_OPS
			shift(828),  // ctx_OFS
			shift(829),  // ctx_ORS
			shift(830),  // ctx_FLATSEP
			shift(831),  // ctx_NF
			shift(832),  // ctx_NR
			shift(833),  // ctx_FNR
			shift(834),  // ctx_FILENAME
			shift(835),  // ctx_FILENUM
			shift(837),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(844),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S822
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(1656), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S823
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(240), // >, reduce: MlrvalOrFunction
			reduce(240), // >>, reduce: MlrvalOrFunction
			reduce(240), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(240), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(240), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(240), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(240), // ||, reduce: MlrvalOrFunction
			reduce(240), // ^^, reduce: MlrvalOrFunction
			reduce(240), // &&, reduce: MlrvalOrFunction
			reduce(240), // ??, reduce: MlrvalOrFunction
			reduce(240), // ???, reduce: MlrvalOrFunction
			reduce(240), // =~, reduce: MlrvalOrFunction
			reduce(240), // !=~, reduce: MlrvalOrFunction
			reduce(240), // ==, reduce: MlrvalOrFunction
			reduce(240), // !=, reduce: MlrvalOrFunction
			reduce(240), // <=>, reduce: MlrvalOrFunction
			reduce(240), // >=, reduce: MlrvalOrFunction
			reduce(240), // <, reduce: MlrvalOrFunction
			reduce(240), // <=, reduce: MlrvalOrFunction
			reduce(240), // ^, reduce: MlrvalOrFunction
			reduce(240), // &, reduce: MlrvalOrFunction
			reduce(240), // <<, reduce: MlrvalOrFunction
			reduce(240), // >>>, reduce: MlrvalOrFunction
			reduce(240), // +, reduce: MlrvalOrFunction
			reduce(240), // -, reduce: MlrvalOrFunction
			reduce(240), // .+, reduce: MlrvalOrFunction
			reduce(240), // .-, reduce: MlrvalOrFunction
			reduce(240), // *, reduce: MlrvalOrFunction
			reduce(240), // /, reduce: MlrvalOrFunction
			reduce(240), // //, reduce: MlrvalOrFunction
			reduce(240), // %, reduce: MlrvalOrFunction
			reduce(240), // .*, reduce: MlrvalOrFunction
			reduce(240), // ./, reduce: MlrvalOrFunction
			reduce(240), // .//, reduce: MlrvalOrFunction
			reduce(240), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(240), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(240), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(240), // [[, reduce: MlrvalOrFunction
			reduce(240), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S824
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(241), // >, reduce: ContextVariable
			reduce(241), // >>, reduce: ContextVariable
			reduce(241), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(241), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(241), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(241), // ?, reduce: ContextVariable
			nil,         // :
			reduce(241), // ||, reduce: ContextVariable
			reduce(241), // ^^, reduce: ContextVariable
			reduce(241), // &&, reduce: ContextVariable
			reduce(241), // ??, reduce: ContextVariable
			reduce(241), // ???, reduce: ContextVariable
			reduce(241), // =~, reduce: ContextVariable
			reduce(241), // !=~, reduce: ContextVariable
			reduce(241), // ==, reduce: ContextVariable
			reduce(241), // !=, reduce: ContextVariable
			reduce(241), // <=>, reduce: ContextVariable
			reduce(241), // >=, reduce: ContextVariable
			reduce(241), // <, reduce: ContextVariable
			reduce(241), // <=, reduce: ContextVariable
			reduce(241), // ^, reduce: ContextVariable
			reduce(241), // &, reduce: ContextVariable
			reduce(241), // <<, reduce: ContextVariable
			reduce(241), // >>>, reduce: ContextVariable
			reduce(241), // +, reduce: ContextVariable
			reduce(241), // -, reduce: ContextVariable
			reduce(241), // .+, reduce: ContextVariable
			reduce(241), // .-, reduce: ContextVariable
			reduce(241), // *, reduce: ContextVariable
			reduce(241), // /, reduce: ContextVariable
			reduce(241), // //, reduce: ContextVariable
			reduce(241), // %, reduce: ContextVariable
			reduce(241), // .*, reduce: ContextVariable
			reduce(241), // ./, reduce: ContextVariable
			reduce(241), // .//, reduce: ContextVariable
			reduce(241), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(241), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(241), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(241), // [[, reduce: ContextVariable
			reduce(241), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S825
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(242), // >, reduce: ContextVariable
			reduce(242), // >>, reduce: ContextVariable
			reduce(242), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(242), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(242), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(242), // ?, reduce: ContextVariable
			nil,         // :
			reduce(242), // ||, reduce: ContextVariable
			reduce(242), // ^^, reduce: ContextVariable
			reduce(242), // &&, reduce: ContextVariable
			reduce(242), // ??, reduce: ContextVariable
			reduce(242), // ???, reduce: ContextVariable
			reduce(242), // =~, reduce: ContextVariable
			reduce(242), // !=~, reduce: ContextVariable
			reduce(242), // ==, reduce: ContextVariable
			reduce(242), // !=, reduce: ContextVariable
			reduce(242), // <=>, reduce: ContextVariable
			reduce(242), // >=, reduce: ContextVariable
			reduce(242), // <, reduce: ContextVariable
			reduce(242), // <=, reduce: ContextVariable
			reduce(242), // ^, reduce: ContextVariable
			reduce(242), // &, reduce: ContextVariable
			reduce(242), // <<, reduce: ContextVariable
			reduce(242), // >>>, reduce: ContextVariable
			reduce(242), // +, reduce: ContextVariable
			reduce(242), // -, reduce: ContextVariable
			reduce(242), // .+, reduce: ContextVariable
			reduce(242), // .-, reduce: ContextVariable
			reduce(242), // *, reduce: ContextVariable
			reduce(242), // /, reduce: ContextVariable
			reduce(242), // //, reduce: ContextVariable
			reduce(242), // %, reduce: ContextVariable
			reduce(242), // .*, reduce: ContextVariable
			reduce(242), // ./, reduce: ContextVariable
			reduce(242), // .//, reduce: ContextVariable
			reduce(242), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(242), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(242), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(242), // [[, reduce: ContextVariable
			reduce(242), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S826
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(243), // >, reduce: ContextVariable
			reduce(243), // >>, reduce: ContextVariable
			reduce(243), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(243), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(243), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(243), // ?, reduce: ContextVariable
			nil,         // :
			reduce(243), // ||, reduce: ContextVariable
			reduce(243), // ^^, reduce: ContextVariable
			reduce(243), // &&, reduce: ContextVariable
			reduce(243), // ??, reduce: ContextVariable
			reduce(243), // ???, reduce: ContextVariable
			reduce(243), // =~, reduce: ContextVariable
			reduce(243), // !=~, reduce: ContextVariable
			reduce(243), // ==, reduce: ContextVariable
			reduce(243), // !=, reduce: ContextVariable
			reduce(243), // <=>, reduce: ContextVariable
			reduce(243), // >=, reduce: ContextVariable
			reduce(243), // <, reduce: ContextVariable
			reduce(243), // <=, reduce: ContextVariable
			reduce(243), // ^, reduce: ContextVariable
			reduce(243), // &, reduce: ContextVariable
			reduce(243), // <<, reduce: ContextVariable
			reduce(243), // >>>, reduce: ContextVariable
			reduce(243), // +, reduce: ContextVariable
			reduce(243), // -, reduce: ContextVariable
			reduce(243), // .+, reduce: ContextVariable
			reduce(243), // .-, reduce: ContextVariable
			reduce(243), // *, reduce: ContextVariable
			reduce(243), // /, reduce: ContextVariable
			reduce(243), // //, reduce: ContextVariable
			reduce(243), // %, reduce: ContextVariable
			reduce(243), // .*, reduce: ContextVariable
			reduce(243), // ./, reduce: ContextVariable
			reduce(243), // .//, reduce: ContextVariable
			reduce(243), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(243), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(243), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(243), // [[, reduce: ContextVariable
			reduce(243), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S827
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(244), // >, reduce: ContextVariable
			reduce(244), // >>, reduce: ContextVariable
			reduce(244), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(244), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(244), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(244), // ?, reduce: ContextVariable
			nil,         // :
			reduce(244), // ||, reduce: ContextVariable
			reduce(244), // ^^, reduce: ContextVariable
			reduce(244), // &&, reduce: ContextVariable
			reduce(244), // ??, reduce: ContextVariable
			reduce(244), // ???, reduce: ContextVariable
			reduce(244), // =~, reduce: ContextVariable
			reduce(244), // !=~, reduce: ContextVariable
			reduce(244), // ==, reduce: ContextVariable
			reduce(244), // !=, reduce: ContextVariable
			reduce(244), // <=>, reduce: ContextVariable
			reduce(244), // >=, reduce: ContextVariable
			reduce(244), // <, reduce: ContextVariable
			reduce(244), // <=, reduce: ContextVariable
			reduce(244), // ^, reduce: ContextVariable
			reduce(244), // &, reduce: ContextVariable
			reduce(244), // <<, reduce: ContextVariable
			reduce(244), // >>>, reduce: ContextVariable
			reduce(244), // +, reduce: ContextVariable
			reduce(244), // -, reduce: ContextVariable
			reduce(244), // .+, reduce: ContextVariable
			reduce(244), // .-, reduce: ContextVariable
			reduce(244), // *, reduce: ContextVariable
			reduce(244), // /, reduce: ContextVariable
			reduce(244), // //, reduce: ContextVariable
			reduce(244), // %, reduce: ContextVariable
			reduce(244), // .*, reduce: ContextVariable
			reduce(244), // ./, reduce: ContextVariable
			reduce(244), // .//, reduce: ContextVariable
			reduce(244), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(244), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(244), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(244), // [[, reduce: ContextVariable
			reduce(244), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S828
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(245), // >, reduce: ContextVariable
			reduce(245), // >>, reduce: ContextVariable
			reduce(245), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(245), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(245), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(245), // ?, reduce: ContextVariable
			nil,         // :
			reduce(245), // ||, reduce: ContextVariable
			reduce(245), // ^^, reduce: ContextVariable
			reduce(245), // &&, reduce: ContextVariable
			reduce(245), // ??, reduce: ContextVariable
			reduce(245), // ???, reduce: ContextVariable
			reduce(245), // =~, reduce: ContextVariable
			reduce(245), // !=~, reduce: ContextVariable
			reduce(245), // ==, reduce: ContextVariable
			reduce(245), // !=, reduce: ContextVariable
			reduce(245), // <=>, reduce: ContextVariable
			reduce(245), // >=, reduce: ContextVariable
			reduce(245), // <, reduce: ContextVariable
			reduce(245), // <=, reduce: ContextVariable
			reduce(245), // ^, reduce: ContextVariable
			reduce(245), // &, reduce: ContextVariable
			reduce(245), // <<, reduce: ContextVariable
			reduce(245), // >>>, reduce: ContextVariable
			reduce(245), // +, reduce: ContextVariable
			reduce(245), // -, reduce: ContextVariable
			reduce(245), // .+, reduce: ContextVariable
			reduce(245), // .-, reduce: ContextVariable
			reduce(245), // *, reduce: ContextVariable
			reduce(245), // /, reduce: ContextVariable
			reduce(245), // //, reduce: ContextVariable
			reduce(245), // %, reduce: ContextVariable
			reduce(245), // .*, reduce: ContextVariable
			reduce(245), // ./, reduce: ContextVariable
			reduce(245), // .//, reduce: ContextVariable
			reduce(245), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(245), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(245), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(245), // [[, reduce: ContextVariable
			reduce(245), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S829
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(246), // >, reduce: ContextVariable
			reduce(246), // >>, reduce: ContextVariable
			reduce(246), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(246), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(246), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(246), // ?, reduce: ContextVariable
			nil,         // :
			reduce(246), // ||, reduce: ContextVariable
			reduce(246), // ^^, reduce: ContextVariable
			reduce(246), // &&, reduce: ContextVariable
			reduce(246), // ??, reduce: ContextVariable
			reduce(246), // ???, reduce: ContextVariable
			reduce(246), // =~, reduce: ContextVariable
			reduce(246), // !=~, reduce: ContextVariable
			reduce(246), // ==, reduce: ContextVariable
			reduce(246), // !=, reduce: ContextVariable
			reduce(246), // <=>, reduce: ContextVariable
			reduce(246), // >=, reduce: ContextVariable
			reduce(246), // <, reduce: ContextVariable
			reduce(246), // <=, reduce: ContextVariable
			reduce(246), // ^, reduce: ContextVariable
			reduce(246), // &, reduce: ContextVariable
			reduce(246), // <<, reduce: ContextVariable
			reduce(246), // >>>, reduce: ContextVariable
			reduce(246), // +, reduce: ContextVariable
			reduce(246), // -, reduce: ContextVariable
			reduce(246), // .+, reduce: ContextVariable
			reduce(246), // .-, reduce: ContextVariable
			reduce(246), // *, reduce: ContextVariable
			reduce(246), // /, reduce: ContextVariable
			reduce(246), // //, reduce: ContextVariable
			reduce(246), // %, reduce: ContextVariable
			reduce(246), // .*, reduce: ContextVariable
			reduce(246), // ./, reduce: ContextVariable
			reduce(246), // .//, reduce: ContextVariable
			reduce(246), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(246), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(246), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(246), // [[, reduce: ContextVariable
			reduce(246), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S830
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(247), // >, reduce: ContextVariable
			reduce(247), // >>, reduce: ContextVariable
			reduce(247), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(247), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(247), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(247), // ?, reduce: ContextVariable
			nil,         // :
			reduce(247), // ||, reduce: ContextVariable
			reduce(247), // ^^, reduce: ContextVariable
			reduce(247), // &&, reduce: ContextVariable
			reduce(247), // ??, reduce: ContextVariable
			reduce(247), // ???, reduce: ContextVariable
			reduce(247), // =~, reduce: ContextVariable
			reduce(247), // !=~, reduce: ContextVariable
			reduce(247), // ==, reduce: ContextVariable
			reduce(247), // !=, reduce: ContextVariable
			reduce(247), // <=>, reduce: ContextVariable
			reduce(247), // >=, reduce: ContextVariable
			reduce(247), // <, reduce: ContextVariable
			reduce(247), // <=, reduce: ContextVariable
			reduce(247), // ^, reduce: ContextVariable
			reduce(247), // &, reduce: ContextVariable
			reduce(247), // <<, reduce: ContextVariable
			reduce(247), // >>>, reduce: ContextVariable
			reduce(247), // +, reduce: ContextVariable
			reduce(247), // -, reduce: ContextVariable
			reduce(247), // .+, reduce: ContextVariable
			reduce(247), // .-, reduce: ContextVariable
			reduce(247), // *, reduce: ContextVariable
			reduce(247), // /, reduce: ContextVariable
			reduce(247), // //, reduce: ContextVariable
			reduce(247), // %, reduce: ContextVariable
			reduce(247), // .*, reduce: ContextVariable
			reduce(247), // ./, reduce: ContextVariable
			reduce(247), // .//, reduce: ContextVariable
			reduce(247), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(247), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(247), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(247), // [[, reduce: ContextVariable
			reduce(247), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S831
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(248), // >, reduce: ContextVariable
			reduce(248), // >>, reduce: ContextVariable
			reduce(248), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(248), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(248), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(248), // ?, reduce: ContextVariable
			nil,         // :
			reduce(248), // ||, reduce: ContextVariable
			reduce(248), // ^^, reduce: ContextVariable
			reduce(248), // &&, reduce: ContextVariable
			reduce(248), // ??, reduce: ContextVariable
			reduce(248), // ???, reduce: ContextVariable
			reduce(248), // =~, reduce: ContextVariable
			reduce(248), // !=~, reduce: ContextVariable
			reduce(248), // ==, reduce: ContextVariable
			reduce(248), // !=, reduce: ContextVariable
			reduce(248), // <=>, reduce: ContextVariable
			reduce(248), // >=, reduce: ContextVariable
			reduce(248), // <, reduce: ContextVariable
			reduce(248), // <=, reduce: ContextVariable
			reduce(248), // ^, reduce: ContextVariable
			reduce(248), // &, reduce: ContextVariable
			reduce(248), // <<, reduce: ContextVariable
			reduce(248), // >>>, reduce: ContextVariable
			reduce(248), // +, reduce: ContextVariable
			reduce(248), // -, reduce: ContextVariable
			reduce(248), // .+, reduce: ContextVariable
			reduce(248), // .-, reduce: ContextVariable
			reduce(248), // *, reduce: ContextVariable
			reduce(248), // /, reduce: ContextVariable
			reduce(248), // //, reduce: ContextVariable
			reduce(248), // %, reduce: ContextVariable
			reduce(248), // .*, reduce: ContextVariable
			reduce(248), // ./, reduce: ContextVariable
			reduce(248), // .//, reduce: ContextVariable
			reduce(248), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(248), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(248), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(248), // [[, reduce: ContextVariable
			reduce(248), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S832
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(249), // >, reduce: ContextVariable
			reduce(249), // >>, reduce: ContextVariable
			reduce(249), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(249), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(249), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(249), // ?, reduce: ContextVariable
			nil,         // :
			reduce(249), // ||, reduce: ContextVariable
			reduce(249), // ^^, reduce: ContextVariable
			reduce(249), // &&, reduce: ContextVariable
			reduce(249), // ??, reduce: ContextVariable
			reduce(249), // ???, reduce: ContextVariable
			reduce(249), // =~, reduce: ContextVariable
			reduce(249), // !=~, reduce: ContextVariable
			reduce(249), // ==, reduce: ContextVariable
			reduce(249), // !=, reduce: ContextVariable
			reduce(249), // <=>, reduce: ContextVariable
			reduce(249), // >=, reduce: ContextVariable
			reduce(249), // <, reduce: ContextVariable
			reduce(249), // <=, reduce: ContextVariable
			reduce(249), // ^, reduce: ContextVariable
			reduce(249), // &, reduce: ContextVariable
			reduce(249), // <<, reduce: ContextVariable
			reduce(249), // >>>, reduce: ContextVariable
			reduce(249), // +, reduce: ContextVariable
			reduce(249), // -, reduce: ContextVariable
			reduce(249), // .+, reduce: ContextVariable
			reduce(249), // .-, reduce: ContextVariable
			reduce(249), // *, reduce: ContextVariable
			reduce(249), // /, reduce: ContextVariable
			reduce(249), // //, reduce: ContextVariable
			reduce(249), // %, reduce: ContextVariable
			reduce(249), // .*, reduce: ContextVariable
			reduce(249), // ./, reduce: ContextVariable
			reduce(249), // .//, reduce: ContextVariable
			reduce(249), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(249), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(249), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(249), // [[, reduce: ContextVariable
			reduce(249), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S833
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(250), // >, reduce: ContextVariable
			reduce(250), // >>, reduce: ContextVariable
			reduce(250), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(250), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(250), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(250), // ?, reduce: ContextVariable
			nil,         // :
			reduce(250), // ||, reduce: ContextVariable
			reduce(250), // ^^, reduce: ContextVariable
			reduce(250), // &&, reduce: ContextVariable
			reduce(250), // ??, reduce: ContextVariable
			reduce(250), // ???, reduce: ContextVariable
			reduce(250), // =~, reduce: ContextVariable
			reduce(250), // !=~, reduce: ContextVariable
			reduce(250), // ==, reduce: ContextVariable
			reduce(250), // !=, reduce: ContextVariable
			reduce(250), // <=>, reduce: ContextVariable
			reduce(250), // >=, reduce: ContextVariable
			reduce(250), // <, reduce: ContextVariable
			reduce(250), // <=, reduce: ContextVariable
			reduce(250), // ^, reduce: ContextVariable
			reduce(250), // &, reduce: ContextVariable
			reduce(250), // <<, reduce: ContextVariable
			reduce(250), // >>>, reduce: ContextVariable
			reduce(250), // +, reduce: ContextVariable
			reduce(250), // -, reduce: ContextVariable
			reduce(250), // .+, reduce: ContextVariable
			reduce(250), // .-, reduce: ContextVariable
			reduce(250), // *, reduce: ContextVariable
			reduce(250), // /, reduce: ContextVariable
			reduce(250), // //, reduce: ContextVariable
			reduce(250), // %, reduce: ContextVariable
			reduce(250), // .*, reduce: ContextVariable
			reduce(250), // ./, reduce: ContextVariable
			reduce(250), // .//, reduce: ContextVariable
			reduce(250), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(250), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(250), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(250), // [[, reduce: ContextVariable
			reduce(250), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S834
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(251), // >, reduce: ContextVariable
			reduce(251), // >>, reduce: ContextVariable
			reduce(251), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(251), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(251), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(251), // ?, reduce: ContextVariable
			nil,         // :
			reduce(251), // ||, reduce: ContextVariable
			reduce(251), // ^^, reduce: ContextVariable
			reduce(251), // &&, reduce: ContextVariable
			reduce(251), // ??, reduce: ContextVariable
			reduce(251), // ???, reduce: ContextVariable
			reduce(251), // =~, reduce: ContextVariable
			reduce(251), // !=~, reduce: ContextVariable
			reduce(251), // ==, reduce: ContextVariable
			reduce(251), // !=, reduce: ContextVariable
			reduce(251), // <=>, reduce: ContextVariable
			reduce(251), // >=, reduce: ContextVariable
			reduce(251), // <, reduce: ContextVariable
			reduce(251), // <=, reduce: ContextVariable
			reduce(251), // ^, reduce: ContextVariable
			reduce(251), // &, reduce: ContextVariable
			reduce(251), // <<, reduce: ContextVariable
			reduce(251), // >>>, reduce: ContextVariable
			reduce(251), // +, reduce: ContextVariable
			reduce(251), // -, reduce: ContextVariable
			reduce(251), // .+, reduce: ContextVariable
			reduce(251), // .-, reduce: ContextVariable
			reduce(251), // *, reduce: ContextVariable
			reduce(251), // /, reduce: ContextVariable
			reduce(251), // //, reduce: ContextVariable
			reduce(251), // %, reduce: ContextVariable
			reduce(251), // .*, reduce: ContextVariable
			reduce(251), // ./, reduce: ContextVariable
			reduce(251), // .//, reduce: ContextVariable
			reduce(251), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(251), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(251), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(251), // [[, reduce: ContextVariable
			reduce(251), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S835
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(252), // >, reduce: ContextVariable
			reduce(252), // >>, reduce: ContextVariable
			reduce(252), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(252), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(252), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(252), // ?, reduce: ContextVariable
			nil,         // :
			reduce(252), // ||, reduce: ContextVariable
			reduce(252), // ^^, reduce: ContextVariable
			reduce(252), // &&, reduce: ContextVariable
			reduce(252), // ??, reduce: ContextVariable
			reduce(252), // ???, reduce: ContextVariable
			reduce(252), // =~, reduce: ContextVariable
			reduce(252), // !=~, reduce: ContextVariable
			reduce(252), // ==, reduce: ContextVariable
			reduce(252), // !=, reduce: ContextVariable
			reduce(252), // <=>, reduce: ContextVariable
			reduce(252), // >=, reduce: ContextVariable
			reduce(252), // <, reduce: ContextVariable
			reduce(252), // <=, reduce: ContextVariable
			reduce(252), // ^, reduce: ContextVariable
			reduce(252), // &, reduce: ContextVariable
			reduce(252), // <<, reduce: ContextVariable
			reduce(252), // >>>, reduce: ContextVariable
			reduce(252), // +, reduce: ContextVariable
			reduce(252), // -, reduce: ContextVariable
			reduce(252), // .+, reduce: ContextVariable
			reduce(252), // .-, reduce: ContextVariable
			reduce(252), // *, reduce: ContextVariable
			reduce(252), // /, reduce: ContextVariable
			reduce(252), // //, reduce: ContextVariable
			reduce(252), // %, reduce: ContextVariable
			reduce(252), // .*, reduce: ContextVariable
			reduce(252), // ./, reduce: ContextVariable
			reduce(252), // .//, reduce: ContextVariable
			reduce(252), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(252), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(252), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(252), // [[, reduce: ContextVariable
			reduce(252), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S836
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(253), // >, reduce: MlrvalOrFunction
			reduce(253), // >>, reduce: MlrvalOrFunction
			reduce(253), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(253), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(253), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(253), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(253), // ||, reduce: MlrvalOrFunction
			reduce(253), // ^^, reduce: MlrvalOrFunction
			reduce(253), // &&, reduce: MlrvalOrFunction
			reduce(253), // ??, reduce: MlrvalOrFunction
			reduce(253), // ???, reduce: MlrvalOrFunction
			reduce(253), // =~, reduce: MlrvalOrFunction
			reduce(253), // !=~, reduce: MlrvalOrFunction
			reduce(253), // ==, reduce: MlrvalOrFunction
			reduce(253), // !=, reduce: MlrvalOrFunction
			reduce(253), // <=>, reduce: MlrvalOrFunction
			reduce(253), // >=, reduce: MlrvalOrFunction
			reduce(253), // <, reduce: MlrvalOrFunction
			reduce(253), // <=, reduce: MlrvalOrFunction
			reduce(253), // ^, reduce: MlrvalOrFunction
			reduce(253), // &, reduce: MlrvalOrFunction
			reduce(253), // <<, reduce: MlrvalOrFunction
			reduce(253), // >>>, reduce: MlrvalOrFunction
			reduce(253), // +, reduce: MlrvalOrFunction
			reduce(253), // -, reduce: MlrvalOrFunction
			reduce(253), // .+, reduce: MlrvalOrFunction
			reduce(253), // .-, reduce: MlrvalOrFunction
			reduce(253), // *, reduce: MlrvalOrFunction
			reduce(253), // /, reduce: MlrvalOrFunction
			reduce(253), // //, reduce: MlrvalOrFunction
			reduce(253), // %, reduce: MlrvalOrFunction
			reduce(253), // .*, reduce: MlrvalOrFunction
			reduce(253), // ./, reduce: MlrvalOrFunction
			reduce(253), // .//, reduce: MlrvalOrFunction
			reduce(253), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(253), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(253), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(253), // [[, reduce: MlrvalOrFunction
			reduce(253), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S837
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			shift(1657), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(1658), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S838
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(256), // >, reduce: MlrvalOrFunction
			reduce(256), // >>, reduce: MlrvalOrFunction
			reduce(256), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(256), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(256), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(256), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(256), // ||, reduce: MlrvalOrFunction
			reduce(256), // ^^, reduce: MlrvalOrFunction
			reduce(256), // &&, reduce: MlrvalOrFunction
			reduce(256), // ??, reduce: MlrvalOrFunction
			reduce(256), // ???, reduce: MlrvalOrFunction
			reduce(256), // =~, reduce: MlrvalOrFunction
			reduce(256), // !=~, reduce: MlrvalOrFunction
			reduce(256), // ==, reduce: MlrvalOrFunction
			reduce(256), // !=, reduce: MlrvalOrFunction
			reduce(256), // <=>, reduce: MlrvalOrFunction
			reduce(256), // >=, reduce: MlrvalOrFunction
			reduce(256), // <, reduce: MlrvalOrFunction
			reduce(256), // <=, reduce: MlrvalOrFunction
			reduce(256), // ^, reduce: MlrvalOrFunction
			reduce(256), // &, reduce: MlrvalOrFunction
			reduce(256), // <<, reduce: MlrvalOrFunction
			reduce(256), // >>>, reduce: MlrvalOrFunction
			reduce(256), // +, reduce: MlrvalOrFunction
			reduce(256), // -, reduce: MlrvalOrFunction
			reduce(256), // .+, reduce: MlrvalOrFunction
			reduce(256), // .-, reduce: MlrvalOrFunction
			reduce(256), // *, reduce: MlrvalOrFunction
			reduce(256), // /, reduce: MlrvalOrFunction
			reduce(256), // //, reduce: MlrvalOrFunction
			reduce(256), // %, reduce: MlrvalOrFunction
			reduce(256), // .*, reduce: MlrvalOrFunction
			reduce(256), // ./, reduce: MlrvalOrFunction
			reduce(256), // .//, reduce: MlrvalOrFunction
			reduce(256), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(256), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(256), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(256), // [[, reduce: MlrvalOrFunction
			reduce(256), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S839
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(257), // >, reduce: MlrvalOrFunction
			reduce(257), // >>, reduce: MlrvalOrFunction
			reduce(257), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(257), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(257), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(257), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(257), // ||, reduce: MlrvalOrFunction
			reduce(257), // ^^, reduce: MlrvalOrFunction
			reduce(257), // &&, reduce: MlrvalOrFunction
			reduce(257), // ??, reduce: MlrvalOrFunction
			reduce(257), // ???, reduce: MlrvalOrFunction
			reduce(257), // =~, reduce: MlrvalOrFunction
			reduce(257), // !=~, reduce: MlrvalOrFunction
			reduce(257), // ==, reduce: MlrvalOrFunction
			reduce(257), // !=, reduce: MlrvalOrFunction
			reduce(257), // <=>, reduce: MlrvalOrFunction
			reduce(257), // >=, reduce: MlrvalOrFunction
			reduce(257), // <, reduce: MlrvalOrFunction
			reduce(257), // <=, reduce: MlrvalOrFunction
			reduce(257), // ^, reduce: MlrvalOrFunction
			reduce(257), // &, reduce: MlrvalOrFunction
			reduce(257), // <<, reduce: MlrvalOrFunction
			reduce(257), // >>>, reduce: MlrvalOrFunction
			reduce(257), // +, reduce: MlrvalOrFunction
			reduce(257), // -, reduce: MlrvalOrFunction
			reduce(257), // .+, reduce: MlrvalOrFunction
			reduce(257), // .-, reduce: MlrvalOrFunction
			reduce(257), // *, reduce: MlrvalOrFunction
			reduce(257), // /, reduce: MlrvalOrFunction
			reduce(257), // //, reduce: MlrvalOrFunction
			reduce(257), // %, reduce: MlrvalOrFunction
			reduce(257), // .*, reduce: MlrvalOrFunction
			reduce(257), // ./, reduce: MlrvalOrFunction
			reduce(257), // .//, reduce: MlrvalOrFunction
			reduce(257), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(257), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(257), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(257), // [[, reduce: MlrvalOrFunction
			reduce(257), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S840
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(258), // >, reduce: MlrvalOrFunction
			reduce(258), // >>, reduce: MlrvalOrFunction
			reduce(258), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(258), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(258), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(258), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(258), // ||, reduce: MlrvalOrFunction
			reduce(258), // ^^, reduce: MlrvalOrFunction
			reduce(258), // &&, reduce: MlrvalOrFunction
			reduce(258), // ??, reduce: MlrvalOrFunction
			reduce(258), // ???, reduce: MlrvalOrFunction
			reduce(258), // =~, reduce: MlrvalOrFunction
			reduce(258), // !=~, reduce: MlrvalOrFunction
			reduce(258), // ==, reduce: MlrvalOrFunction
			reduce(258), // !=, reduce: MlrvalOrFunction
			reduce(258), // <=>, reduce: MlrvalOrFunction
			reduce(258), // >=, reduce: MlrvalOrFunction
			reduce(258), // <, reduce: MlrvalOrFunction
			reduce(258), // <=, reduce: MlrvalOrFunction
			reduce(258), // ^, reduce: MlrvalOrFunction
			reduce(258), // &, reduce: MlrvalOrFunction
			reduce(258), // <<, reduce: MlrvalOrFunction
			reduce(258), // >>>, reduce: MlrvalOrFunction
			reduce(258), // +, reduce: MlrvalOrFunction
			reduce(258), // -, reduce: MlrvalOrFunction
			reduce(258), // .+, reduce: MlrvalOrFunction
			reduce(258), // .-, reduce: MlrvalOrFunction
			reduce(258), // *, reduce: MlrvalOrFunction
			reduce(258), // /, reduce: MlrvalOrFunction
			reduce(258), // //, reduce: MlrvalOrFunction
			reduce(258), // %, reduce: MlrvalOrFunction
			reduce(258), // .*, reduce: MlrvalOrFunction
			reduce(258), // ./, reduce: MlrvalOrFunction
			reduce(258), // .//, reduce: MlrvalOrFunction
			reduce(258), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(258), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(258), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(258), // [[, reduce: MlrvalOrFunction
			reduce(258), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S841
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(259), // >, reduce: MlrvalOrFunction
			reduce(259), // >>, reduce: MlrvalOrFunction
			reduce(259), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(259), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(259), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(259), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(259), // ||, reduce: MlrvalOrFunction
			reduce(259), // ^^, reduce: MlrvalOrFunction
			reduce(259), // &&, reduce: MlrvalOrFunction
			reduce(259), // ??, reduce: MlrvalOrFunction
			reduce(259), // ???, reduce: MlrvalOrFunction
			reduce(259), // =~, reduce: MlrvalOrFunction
			reduce(259), // !=~, reduce: MlrvalOrFunction
			reduce(259), // ==, reduce: MlrvalOrFunction
			reduce(259), // !=, reduce: MlrvalOrFunction
			reduce(259), // <=>, reduce: MlrvalOrFunction
			reduce(259), // >=, reduce: MlrvalOrFunction
			reduce(259), // <, reduce: MlrvalOrFunction
			reduce(259), // <=, reduce: MlrvalOrFunction
			reduce(259), // ^, reduce: MlrvalOrFunction
			reduce(259), // &, reduce: MlrvalOrFunction
			reduce(259), // <<, reduce: MlrvalOrFunction
			reduce(259), // >>>, reduce: MlrvalOrFunction
			reduce(259), // +, reduce: MlrvalOrFunction
			reduce(259), // -, reduce: MlrvalOrFunction
			reduce(259), // .+, reduce: MlrvalOrFunction
			reduce(259), // .-, reduce: MlrvalOrFunction
			reduce(259), // *, reduce: MlrvalOrFunction
			reduce(259), // /, reduce: MlrvalOrFunction
			reduce(259), // //, reduce: MlrvalOrFunction
			reduce(259), // %, reduce: MlrvalOrFunction
			reduce(259), // .*, reduce: MlrvalOrFunction
			reduce(259), // ./, reduce: MlrvalOrFunction
			reduce(259), // .//, reduce: MlrvalOrFunction
			reduce(259), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(259), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(259), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(259), // [[, reduce: MlrvalOrFunction
			reduce(259), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S842
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(267), // >, reduce: MlrvalOrFunction
			reduce(267), // >>, reduce: MlrvalOrFunction
			reduce(267), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(267), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(267), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(267), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(267), // ||, reduce: MlrvalOrFunction
			reduce(267), // ^^, reduce: MlrvalOrFunction
			reduce(267), // &&, reduce: MlrvalOrFunction
			reduce(267), // ??, reduce: MlrvalOrFunction
			reduce(267), // ???, reduce: MlrvalOrFunction
			reduce(267), // =~, reduce: MlrvalOrFunction
			reduce(267), // !=~, reduce: MlrvalOrFunction
			reduce(267), // ==, reduce: MlrvalOrFunction
			reduce(267), // !=, reduce: MlrvalOrFunction
			reduce(267), // <=>, reduce: MlrvalOrFunction
			reduce(267), // >=, reduce: MlrvalOrFunction
			reduce(267), // <, reduce: MlrvalOrFunction
			reduce(267), // <=, reduce: MlrvalOrFunction
			reduce(267), // ^, reduce: MlrvalOrFunction
			reduce(267), // &, reduce: MlrvalOrFunction
			reduce(267), // <<, reduce: MlrvalOrFunction
			reduce(267), // >>>, reduce: MlrvalOrFunction
			reduce(267), // +, reduce: MlrvalOrFunction
			reduce(267), // -, reduce: MlrvalOrFunction
			reduce(267), // .+, reduce: MlrvalOrFunction
			reduce(267), // .-, reduce: MlrvalOrFunction
			reduce(267), // *, reduce: MlrvalOrFunction
			reduce(267), // /, reduce: MlrvalOrFunction
			reduce(267), // //, reduce: MlrvalOrFunction
			reduce(267), // %, reduce: MlrvalOrFunction
			reduce(267), // .*, reduce: MlrvalOrFunction
			reduce(267), // ./, reduce: MlrvalOrFunction
			reduce(267), // .//, reduce: MlrvalOrFunction
			reduce(267), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(267), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(267), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(267), // [[, reduce: MlrvalOrFunction
			reduce(267), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S843
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1659), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S844
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1660), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S845
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1661), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S846
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S847
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(1669), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S848
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(278), // (, reduce: SubroutineName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S849
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1756), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S850
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(287), // $, reduce: BeginBlock
			nil,         // empty
			reduce(287), // ;, reduce: BeginBlock
			reduce(287), // {, reduce: BeginBlock
			nil,         // }
			nil,         // =
			reduce(287), // unset, reduce: BeginBlock
			reduce(287), // filter, reduce: BeginBlock
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(287), // print, reduce: BeginBlock
			nil,         // ,
			reduce(287), // printn, reduce: BeginBlock
			reduce(287), // eprint, reduce: BeginBlock
			reduce(287), // eprintn, reduce: BeginBlock
			reduce(287), // dump, reduce: BeginBlock
			reduce(287), // edump, reduce: BeginBlock
			reduce(287), // tee, reduce: BeginBlock
			reduce(287), // emitf, reduce: BeginBlock
			reduce(287), // emit1, reduce: BeginBlock
			reduce(287), // emit, reduce: BeginBlock
			reduce(287), // (, reduce: BeginBlock
			nil,         // )
			reduce(287), // emitp, reduce: BeginBlock
			reduce(287), // field_name, reduce: BeginBlock
			reduce(287), // $[, reduce: BeginBlock
			nil,         // ]
			reduce(287), // braced_field_name, reduce: BeginBlock
			reduce(287), // $[[, reduce: BeginBlock
			reduce(287), // $[[[, reduce: BeginBlock
			reduce(287), // full_srec, reduce: BeginBlock
			reduce(287), // oosvar_name, reduce: BeginBlock
			reduce(287), // @[, reduce: BeginBlock
			reduce(287), // braced_oosvar_name, reduce: BeginBlock
			reduce(287), // full_oosvar, reduce: BeginBlock
			reduce(287), // all, reduce: BeginBlock
			reduce(287), // non_sigil_name, reduce: BeginBlock
			reduce(287), // arr, reduce: BeginBlock
			reduce(287), // bool, reduce: BeginBlock
			reduce(287), // float, reduce: BeginBlock
			reduce(287), // int, reduce: BeginBlock
			reduce(287), // map, reduce: BeginBlock
			reduce(287), // num, reduce: BeginBlock
			reduce(287), // str, reduce: BeginBlock
			reduce(287), // var, reduce: BeginBlock
			reduce(287), // funct, reduce: BeginBlock
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(287), // +, reduce: BeginBlock
			reduce(287), // -, reduce: BeginBlock
			reduce(287), // .+, reduce: BeginBlock
			reduce(287), // .-, reduce: BeginBlock
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(287), // !, reduce: BeginBlock
			reduce(287), // ~, reduce: BeginBlock
			nil,         // **
			reduce(287), // string_literal, reduce: BeginBlock
			reduce(287), // regex_case_insensitive, reduce: BeginBlock
			reduce(287), // int_literal, reduce: BeginBlock
			reduce(287), // float_literal, reduce: BeginBlock
			reduce(287), // boolean_literal, reduce: BeginBlock
			reduce(287), // null_literal, reduce: BeginBlock
			reduce(287), // inf_literal, reduce: BeginBlock
			reduce(287), // nan_literal, reduce: BeginBlock
			reduce(287), // const_M_PI, reduce: BeginBlock
			reduce(287), // const_M_E, reduce: BeginBlock
			reduce(287), // panic, reduce: BeginBlock
			reduce(287), // [, reduce: BeginBlock
			reduce(287), // ctx_IPS, reduce: BeginBlock
			reduce(287), // ctx_IFS, reduce: BeginBlock
			reduce(287), // ctx_IRS, reduce: BeginBlock
			reduce(287), // ctx_OPS, reduce: BeginBlock
			reduce(287), // ctx_OFS, reduce: BeginBlock
			reduce(287), // ctx_ORS, reduce: BeginBlock
			reduce(287), // ctx_FLATSEP, reduce: BeginBlock
			reduce(287), // ctx_NF, reduce: BeginBlock
			reduce(287), // ctx_NR, reduce: BeginBlock
			reduce(287), // ctx_FNR, reduce: BeginBlock
			reduce(287), // ctx_FILENAME, reduce: BeginBlock
			reduce(287), // ctx_FILENUM, reduce: BeginBlock
			reduce(287), // env, reduce: BeginBlock
			nil,         // [[
			nil,         // [[[
			reduce(287), // call, reduce: BeginBlock
			reduce(287), // begin, reduce: BeginBlock
			reduce(287), // end, reduce: BeginBlock
			reduce(287), // if, reduce: BeginBlock
			nil,         // elif
			nil,         // else
			reduce(287), // while, reduce: BeginBlock
			reduce(287), // do, reduce: BeginBlock
			reduce(287), // for, reduce: BeginBlock
			nil,         // in
			reduce(287), // break, reduce: BeginBlock
			reduce(287), // continue, reduce: BeginBlock
			reduce(287), // func, reduce: BeginBlock
			reduce(287), // subr, reduce: BeginBlock
			reduce(287), // return, reduce: BeginBlock
		},
	},
	actionRow{ // S851
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(288), // $, reduce: EndBlock
			nil,         // empty
			reduce(288), // ;, reduce: EndBlock
			reduce(288), // {, reduce: EndBlock
			nil,         // }
			nil,         // =
			reduce(288), // unset, reduce: EndBlock
			reduce(288), // filter, reduce: EndBlock
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(288), // print, reduce: EndBlock
			nil,         // ,
			reduce(288), // printn, reduce: EndBlock
			reduce(288), // eprint, reduce: EndBlock
			reduce(288), // eprintn, reduce: EndBlock
			reduce(288), // dump, reduce: EndBlock
			reduce(288), // edump, reduce: EndBlock
			reduce(288), // tee, reduce: EndBlock
			reduce(288), // emitf, reduce: EndBlock
			reduce(288), // emit1, reduce: EndBlock
			reduce(288), // emit, reduce: EndBlock
			reduce(288), // (, reduce: EndBlock
			nil,         // )
			reduce(288), // emitp, reduce: EndBlock
			reduce(288), // field_name, reduce: EndBlock
			reduce(288), // $[, reduce: EndBlock
			nil,         // ]
			reduce(288), // braced_field_name, reduce: EndBlock
			reduce(288), // $[[, reduce: EndBlock
			reduce(288), // $[[[, reduce: EndBlock
			reduce(288), // full_srec, reduce: EndBlock
			reduce(288), // oosvar_name, reduce: EndBlock
			reduce(288), // @[, reduce: EndBlock
			reduce(288), // braced_oosvar_name, reduce: EndBlock
			reduce(288), // full_oosvar, reduce: EndBlock
			reduce(288), // all, reduce: EndBlock
			reduce(288), // non_sigil_name, reduce: EndBlock
			reduce(288), // arr, reduce: EndBlock
			reduce(288), // bool, reduce: EndBlock
			reduce(288), // float, reduce: EndBlock
			reduce(288), // int, reduce: EndBlock
			reduce(288), // map, reduce: EndBlock
			reduce(288), // num, reduce: EndBlock
			reduce(288), // str, reduce: EndBlock
			reduce(288), // var, reduce: EndBlock
			reduce(288), // funct, reduce: EndBlock
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(288), // +, reduce: EndBlock
			reduce(288), // -, reduce: EndBlock
			reduce(288), // .+, reduce: EndBlock
			reduce(288), // .-, reduce: EndBlock
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(288), // !, reduce: EndBlock
			reduce(288), // ~, reduce: EndBlock
			nil,         // **
			reduce(288), // string_literal, reduce: EndBlock
			reduce(288), // regex_case_insensitive, reduce: EndBlock
			reduce(288), // int_literal, reduce: EndBlock
			reduce(288), // float_literal, reduce: EndBlock
			reduce(288), // boolean_literal, reduce: EndBlock
			reduce(288), // null_literal, reduce: EndBlock
			reduce(288), // inf_literal, reduce: EndBlock
			reduce(288), // nan_literal, reduce: EndBlock
			reduce(288), // const_M_PI, reduce: EndBlock
			reduce(288), // const_M_E, reduce: EndBlock
			reduce(288), // panic, reduce: EndBlock
			reduce(288), // [, reduce: EndBlock
			reduce(288), // ctx_IPS, reduce: EndBlock
			reduce(288), // ctx_IFS, reduce: EndBlock
			reduce(288), // ctx_IRS, reduce: EndBlock
			reduce(288), // ctx_OPS, reduce: EndBlock
			reduce(288), // ctx_OFS, reduce: EndBlock
			reduce(288), // ctx_ORS, reduce: EndBlock
			reduce(288), // ctx_FLATSEP, reduce: EndBlock
			reduce(288), // ctx_NF, reduce: EndBlock
			reduce(288), // ctx_NR, reduce: EndBlock
			reduce(288), // ctx_FNR, reduce: EndBlock
			reduce(288), // ctx_FILENAME, reduce: EndBlock
			reduce(288), // ctx_FILENUM, reduce: EndBlock
			reduce(288), // env, reduce: EndBlock
			nil,         // [[
			nil,         // [[[
			reduce(288), // call, reduce: EndBlock
			reduce(288), // begin, reduce: EndBlock
			reduce(288), // end, reduce: EndBlock
			reduce(288), // if, reduce: EndBlock
			nil,         // elif
			nil,         // else
			reduce(288), // while, reduce: EndBlock
			reduce(288), // do, reduce: EndBlock
			reduce(288), // for, reduce: EndBlock
			nil,         // in
			reduce(288), // break, reduce: EndBlock
			reduce(288), // continue, reduce: EndBlock
			reduce(288), // func, reduce: EndBlock
			reduce(288), // subr, reduce: EndBlock
			reduce(288), // return, reduce: EndBlock
		},
	},
	actionRow{ // S852
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(291), // $, reduce: IfChain
			nil,         // empty
			reduce(291), // ;, reduce: IfChain
			reduce(291), // {, reduce: IfChain
			nil,         // }
			nil,         // =
			reduce(291), // unset, reduce: IfChain
			reduce(291), // filter, reduce: IfChain
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(291), // print, reduce: IfChain
			nil,         // ,
			reduce(291), // printn, reduce: IfChain
			reduce(291), // eprint, reduce: IfChain
			reduce(291), // eprintn, reduce: IfChain
			reduce(291), // dump, reduce: IfChain
			reduce(291), // edump, reduce: IfChain
			reduce(291), // tee, reduce: IfChain
			reduce(291), // emitf, reduce: IfChain
			reduce(291), // emit1, reduce: IfChain
			reduce(291), // emit, reduce: IfChain
			reduce(291), // (, reduce: IfChain
			nil,         // )
			reduce(291), // emitp, reduce: IfChain
			reduce(291), // field_name, reduce: IfChain
			reduce(291), // $[, reduce: IfChain
			nil,         // ]
			reduce(291), // braced_field_name, reduce: IfChain
			reduce(291), // $[[, reduce: IfChain
			reduce(291), // $[[[, reduce: IfChain
			reduce(291), // full_srec, reduce: IfChain
			reduce(291), // oosvar_name, reduce: IfChain
			reduce(291), // @[, reduce: IfChain
			reduce(291), // braced_oosvar_name, reduce: IfChain
			reduce(291), // full_oosvar, reduce: IfChain
			reduce(291), // all, reduce: IfChain
			reduce(291), // non_sigil_name, reduce: IfChain
			reduce(291), // arr, reduce: IfChain
			reduce(291), // bool, reduce: IfChain
			reduce(291), // float, reduce: IfChain
			reduce(291), // int, reduce: IfChain
			reduce(291), // map, reduce: IfChain
			reduce(291), // num, reduce: IfChain
			reduce(291), // str, reduce: IfChain
			reduce(291), // var, reduce: IfChain
			reduce(291), // funct, reduce: IfChain
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(291), // +, reduce: IfChain
			reduce(291), // -, reduce: IfChain
			reduce(291), // .+, reduce: IfChain
			reduce(291), // .-, reduce: IfChain
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(291), // !, reduce: IfChain
			reduce(291), // ~, reduce: IfChain
			nil,         // **
			reduce(291), // string_literal, reduce: IfChain
			reduce(291), // regex_case_insensitive, reduce: IfChain
			reduce(291), // int_literal, reduce: IfChain
			reduce(291), // float_literal, reduce: IfChain
			reduce(291), // boolean_literal, reduce: IfChain
			reduce(291), // null_literal, reduce: IfChain
			reduce(291), // inf_literal, reduce: IfChain
			reduce(291), // nan_literal, reduce: IfChain
			reduce(291), // const_M_PI, reduce: IfChain
			reduce(291), // const_M_E, reduce: IfChain
			reduce(291), // panic, reduce: IfChain
			reduce(291), // [, reduce: IfChain
			reduce(291), // ctx_IPS, reduce: IfChain
			reduce(291), // ctx_IFS, reduce: IfChain
			reduce(291), // ctx_IRS, reduce: IfChain
			reduce(291), // ctx_OPS, reduce: IfChain
			reduce(291), // ctx_OFS, reduce: IfChain
			reduce(291), // ctx_ORS, reduce: IfChain
			reduce(291), // ctx_FLATSEP, reduce: IfChain
			reduce(291), // ctx_NF, reduce: IfChain
			reduce(291), // ctx_NR, reduce: IfChain
			reduce(291), // ctx_FNR, reduce: IfChain
			reduce(291), // ctx_FILENAME, reduce: IfChain
			reduce(291), // ctx_FILENUM, reduce: IfChain
			reduce(291), // env, reduce: IfChain
			nil,         // [[
			nil,         // [[[
			reduce(291), // call, reduce: IfChain
			reduce(291), // begin, reduce: IfChain
			reduce(291), // end, reduce: IfChain
			reduce(291), // if, reduce: IfChain
			nil,         // elif
			nil,         // else
			reduce(291), // while, reduce: IfChain
			reduce(291), // do, reduce: IfChain
			reduce(291), // for, reduce: IfChain
			nil,         // in
			reduce(291), // break, reduce: IfChain
			reduce(291), // continue, reduce: IfChain
			reduce(291), // func, reduce: IfChain
			reduce(291), // subr, reduce: IfChain
			reduce(291), // return, reduce: IfChain
		},
	},
	actionRow{ // S853
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(293), // $, reduce: IfElifStar
			nil,         // empty
			reduce(293), // ;, reduce: IfElifStar
			reduce(293), // {, reduce: IfElifStar
			nil,         // }
			nil,         // =
			reduce(293), // unset, reduce: IfElifStar
			reduce(293), // filter, reduce: IfElifStar
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(293), // print, reduce: IfElifStar
			nil,         // ,
			reduce(293), // printn, reduce: IfElifStar
			reduce(293), // eprint, reduce: IfElifStar
			reduce(293), // eprintn, reduce: IfElifStar
			reduce(293), // dump, reduce: IfElifStar
			reduce(293), // edump, reduce: IfElifStar
			reduce(293), // tee, reduce: IfElifStar
			reduce(293), // emitf, reduce: IfElifStar
			reduce(293), // emit1, reduce: IfElifStar
			reduce(293), // emit, reduce: IfElifStar
			reduce(293), // (, reduce: IfElifStar
			nil,         // )
			reduce(293), // emitp, reduce: IfElifStar
			reduce(293), // field_name, reduce: IfElifStar
			reduce(293), // $[, reduce: IfElifStar
			nil,         // ]
			reduce(293), // braced_field_name, reduce: IfElifStar
			reduce(293), // $[[, reduce: IfElifStar
			reduce(293), // $[[[, reduce: IfElifStar
			reduce(293), // full_srec, reduce: IfElifStar
			reduce(293), // oosvar_name, reduce: IfElifStar
			reduce(293), // @[, reduce: IfElifStar
			reduce(293), // braced_oosvar_name, reduce: IfElifStar
			reduce(293), // full_oosvar, reduce: IfElifStar
			reduce(293), // all, reduce: IfElifStar
			reduce(293), // non_sigil_name, reduce: IfElifStar
			reduce(293), // arr, reduce: IfElifStar
			reduce(293), // bool, reduce: IfElifStar
			reduce(293), // float, reduce: IfElifStar
			reduce(293), // int, reduce: IfElifStar
			reduce(293), // map, reduce: IfElifStar
			reduce(293), // num, reduce: IfElifStar
			reduce(293), // str, reduce: IfElifStar
			reduce(293), // var, reduce: IfElifStar
			reduce(293), // funct, reduce: IfElifStar
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(293), // +, reduce: IfElifStar
			reduce(293), // -, reduce: IfElifStar
			reduce(293), // .+, reduce: IfElifStar
			reduce(293), // .-, reduce: IfElifStar
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(293), // !, reduce: IfElifStar
			reduce(293), // ~, reduce: IfElifStar
			nil,         // **
			reduce(293), // string_literal, reduce: IfElifStar
			reduce(293), // regex_case_insensitive, reduce: IfElifStar
			reduce(293), // int_literal, reduce: IfElifStar
			reduce(293), // float_literal, reduce: IfElifStar
			reduce(293), // boolean_literal, reduce: IfElifStar
			reduce(293), // null_literal, reduce: IfElifStar
			reduce(293), // inf_literal, reduce: IfElifStar
			reduce(293), // nan_literal, reduce: IfElifStar
			reduce(293), // const_M_PI, reduce: IfElifStar
			reduce(293), // const_M_E, reduce: IfElifStar
			reduce(293), // panic, reduce: IfElifStar
			reduce(293), // [, reduce: IfElifStar
			reduce(293), // ctx_IPS, reduce: IfElifStar
			reduce(293), // ctx_IFS, reduce: IfElifStar
			reduce(293), // ctx_IRS, reduce: IfElifStar
			reduce(293), // ctx_OPS, reduce: IfElifStar
			reduce(293), // ctx_OFS, reduce: IfElifStar
			reduce(293), // ctx_ORS, reduce: IfElifStar
			reduce(293), // ctx_FLATSEP, reduce: IfElifStar
			reduce(293), // ctx_NF, reduce: IfElifStar
			reduce(293), // ctx_NR, reduce: IfElifStar
			reduce(293), // ctx_FNR, reduce: IfElifStar
			reduce(293), // ctx_FILENAME, reduce: IfElifStar
			reduce(293), // ctx_FILENUM, reduce: IfElifStar
			reduce(293), // env, reduce: IfElifStar
			nil,         // [[
			nil,         // [[[
			reduce(293), // call, reduce: IfElifStar
			reduce(293), // begin, reduce: IfElifStar
			reduce(293), // end, reduce: IfElifStar
			reduce(293), // if, reduce: IfElifStar
			reduce(293), // elif, reduce: IfElifStar
			reduce(293), // else, reduce: IfElifStar
			reduce(293), // while, reduce: IfElifStar
			reduce(293), // do, reduce: IfElifStar
			reduce(293), // for, reduce: IfElifStar
			nil,         // in
			reduce(293), // break, reduce: IfElifStar
			reduce(293), // continue, reduce: IfElifStar
			reduce(293), // func, reduce: IfElifStar
			reduce(293), // subr, reduce: IfElifStar
			reduce(293), // return, reduce: IfElifStar
		},
	},
	actionRow{ // S854
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1757), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S855
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(289), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S856
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S857
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S858
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			shift(1761), // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S859
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S860
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(309), // ;, reduce: TripleForStart
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(309), // ,, reduce: TripleForStart
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1769), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(1793), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S861
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S862
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1866), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S863
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1867), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S864
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(336), // $, reduce: ReturnStatement
			nil,         // empty
			reduce(336), // ;, reduce: ReturnStatement
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S865
		canRecover: false,
		actions: [numSymbols]action{
			nil,       // INVALID
			reduce(7), // $, reduce: NonEmptyStatementBlock
			nil,       // empty
			nil,       // ;
			nil,       // {
			nil,       // }
			nil,       // =
			nil,       // unset
			nil,       // filter
			nil,       // >
			nil,       // >>
			nil,       // |
			nil,       // stdout
			nil,       // stderr
			nil,       // print
			nil,       // ,
			nil,       // printn
			nil,       // eprint
			nil,       // eprintn
			nil,       // dump
			nil,       // edump
			nil,       // tee
			nil,       // emitf
			nil,       // emit1
			nil,       // emit
			nil,       // (
			nil,       // )
			nil,       // emitp
			nil,       // field_name
			nil,       // $[
			nil,       // ]
			nil,       // braced_field_name
			nil,       // $[[
			nil,       // $[[[
			nil,       // full_srec
			nil,       // oosvar_name
			nil,       // @[
			nil,       // braced_oosvar_name
			nil,       // full_oosvar
			nil,       // all
			nil,       // non_sigil_name
			nil,       // arr
			nil,       // bool
			nil,       // float
			nil,       // int
			nil,       // map
			nil,       // num
			nil,       // str
			nil,       // var
			nil,       // funct
			nil,       // ||=
			nil,       // ^^=
			nil,       // &&=
			nil,       // ??=
			nil,       // ???=
			nil,       // |=
			nil,       // &=
			nil,       // ^=
			nil,       // <<=
			nil,       // >>=
			nil,       // >>>=
			nil,       // +=
			nil,       // .=
			nil,       // -=
			nil,       // *=
			nil,       // /=
			nil,       // //=
			nil,       // %=
			nil,       // **=
			nil,       // ?
			nil,       // :
			nil,       // ||
			nil,       // ^^
			nil,       // &&
			nil,       // ??
			nil,       // ???
			nil,       // =~
			nil,       // !=~
			nil,       // ==
			nil,       // !=
			nil,       // <=>
			nil,       // >=
			nil,       // <
			nil,       // <=
			nil,       // ^
			nil,       // &
			nil,       // <<
			nil,       // >>>
			nil,       // +
			nil,       // -
			nil,       // .+
			nil,       // .-
			nil,       // *
			nil,       // /
			nil,       // //
			nil,       // %
			nil,       // .*
			nil,       // ./
			nil,       // .//
			nil,       // .
			nil,       // !
			nil,       // ~
			nil,       // **
			nil,       // string_literal
			nil,       // regex_case_insensitive
			nil,       // int_literal
			nil,       // float_literal
			nil,       // boolean_literal
			nil,       // null_literal
			nil,       // inf_literal
			nil,       // nan_literal
			nil,       // const_M_PI
			nil,       // const_M_E
			nil,       // panic
			nil,       // [
			nil,       // ctx_IPS
			nil,       // ctx_IFS
			nil,       // ctx_IRS
			nil,       // ctx_OPS
			nil,       // ctx_OFS
			nil,       // ctx_ORS
			nil,       // ctx_FLATSEP
			nil,       // ctx_NF
			nil,       // ctx_NR
			nil,       // ctx_FNR
			nil,       // ctx_FILENAME
			nil,       // ctx_FILENUM
			nil,       // env
			nil,       // [[
			nil,       // [[[
			nil,       // call
			nil,       // begin
			nil,       // end
			nil,       // if
			nil,       // elif
			nil,       // else
			nil,       // while
			nil,       // do
			nil,       // for
			nil,       // in
			nil,       // break
			nil,       // continue
			nil,       // func
			nil,       // subr
			nil,       // return
		},
	},
	actionRow{ // S866
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(11), // $, reduce: NonEmptyStatementBlock
			nil,        // empty
			shift(6),   // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			shift(30),  // unset
			shift(33),  // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			shift(34),  // print
			nil,        // ,
			shift(35),  // printn
			shift(36),  // eprint
			shift(37),  // eprintn
			shift(38),  // dump
			shift(39),  // edump
			shift(40),  // tee
			shift(42),  // emitf
			shift(43),  // emit1
			shift(44),  // emit
			shift(45),  // (
			nil,        // )
			shift(46),  // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			shift(71),  // arr
			shift(72),  // bool
			shift(73),  // float
			shift(74),  // int
			shift(75),  // map
			shift(76),  // num
			shift(77),  // str
			shift(78),  // var
			shift(79),  // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			shift(141), // call
			shift(150), // begin
			shift(151), // end
			shift(154), // if
			nil,        // elif
			nil,        // else
			shift(155), // while
			shift(156), // do
			shift(161), // for
			nil,        // in
			shift(162), // break
			shift(163), // continue
			shift(164), // func
			shift(165), // subr
			shift(166), // return
		},
	},
	actionRow{ // S867
		canRecover: false,
		actions: [numSymbols]action{
			nil,       // INVALID
			reduce(9), // $, reduce: NonEmptyStatementBlock
			nil,       // empty
			nil,       // ;
			nil,       // {
			nil,       // }
			nil,       // =
			nil,       // unset
			nil,       // filter
			nil,       // >
			nil,       // >>
			nil,       // |
			nil,       // stdout
			nil,       // stderr
			nil,       // print
			nil,       // ,
			nil,       // printn
			nil,       // eprint
			nil,       // eprintn
			nil,       // dump
			nil,       // edump
			nil,       // tee
			nil,       // emitf
			nil,       // emit1
			nil,       // emit
			nil,       // (
			nil,       // )
			nil,       // emitp
			nil,       // field_name
			nil,       // $[
			nil,       // ]
			nil,       // braced_field_name
			nil,       // $[[
			nil,       // $[[[
			nil,       // full_srec
			nil,       // oosvar_name
			nil,       // @[
			nil,       // braced_oosvar_name
			nil,       // full_oosvar
			nil,       // all
			nil,       // non_sigil_name
			nil,       // arr
			nil,       // bool
			nil,       // float
			nil,       // int
			nil,       // map
			nil,       // num
			nil,       // str
			nil,       // var
			nil,       // funct
			nil,       // ||=
			nil,       // ^^=
			nil,       // &&=
			nil,       // ??=
			nil,       // ???=
			nil,       // |=
			nil,       // &=
			nil,       // ^=
			nil,       // <<=
			nil,       // >>=
			nil,       // >>>=
			nil,       // +=
			nil,       // .=
			nil,       // -=
			nil,       // *=
			nil,       // /=
			nil,       // //=
			nil,       // %=
			nil,       // **=
			nil,       // ?
			nil,       // :
			nil,       // ||
			nil,       // ^^
			nil,       // &&
			nil,       // ??
			nil,       // ???
			nil,       // =~
			nil,       // !=~
			nil,       // ==
			nil,       // !=
			nil,       // <=>
			nil,       // >=
			nil,       // <
			nil,       // <=
			nil,       // ^
			nil,       // &
			nil,       // <<
			nil,       // >>>
			nil,       // +
			nil,       // -
			nil,       // .+
			nil,       // .-
			nil,       // *
			nil,       // /
			nil,       // //
			nil,       // %
			nil,       // .*
			nil,       // ./
			nil,       // .//
			nil,       // .
			nil,       // !
			nil,       // ~
			nil,       // **
			nil,       // string_literal
			nil,       // regex_case_insensitive
			nil,       // int_literal
			nil,       // float_literal
			nil,       // boolean_literal
			nil,       // null_literal
			nil,       // inf_literal
			nil,       // nan_literal
			nil,       // const_M_PI
			nil,       // const_M_E
			nil,       // panic
			nil,       // [
			nil,       // ctx_IPS
			nil,       // ctx_IFS
			nil,       // ctx_IRS
			nil,       // ctx_OPS
			nil,       // ctx_OFS
			nil,       // ctx_ORS
			nil,       // ctx_FLATSEP
			nil,       // ctx_NF
			nil,       // ctx_NR
			nil,       // ctx_FNR
			nil,       // ctx_FILENAME
			nil,       // ctx_FILENUM
			nil,       // env
			nil,       // [[
			nil,       // [[[
			nil,       // call
			nil,       // begin
			nil,       // end
			nil,       // if
			nil,       // elif
			nil,       // else
			nil,       // while
			nil,       // do
			nil,       // for
			nil,       // in
			nil,       // break
			nil,       // continue
			nil,       // func
			nil,       // subr
			nil,       // return
		},
	},
	actionRow{ // S868
		canRecover: false,
		actions: [numSymbols]action{
			nil,       // INVALID
			reduce(8), // $, reduce: NonEmptyStatementBlock
			nil,       // empty
			nil,       // ;
			nil,       // {
			nil,       // }
			nil,       // =
			nil,       // unset
			nil,       // filter
			nil,       // >
			nil,       // >>
			nil,       // |
			nil,       // stdout
			nil,       // stderr
			nil,       // print
			nil,       // ,
			nil,       // printn
			nil,       // eprint
			nil,       // eprintn
			nil,       // dump
			nil,       // edump
			nil,       // tee
			nil,       // emitf
			nil,       // emit1
			nil,       // emit
			nil,       // (
			nil,       // )
			nil,       // emitp
			nil,       // field_name
			nil,       // $[
			nil,       // ]
			nil,       // braced_field_name
			nil,       // $[[
			nil,       // $[[[
			nil,       // full_srec
			nil,       // oosvar_name
			nil,       // @[
			nil,       // braced_oosvar_name
			nil,       // full_oosvar
			nil,       // all
			nil,       // non_sigil_name
			nil,       // arr
			nil,       // bool
			nil,       // float
			nil,       // int
			nil,       // map
			nil,       // num
			nil,       // str
			nil,       // var
			nil,       // funct
			nil,       // ||=
			nil,       // ^^=
			nil,       // &&=
			nil,       // ??=
			nil,       // ???=
			nil,       // |=
			nil,       // &=
			nil,       // ^=
			nil,       // <<=
			nil,       // >>=
			nil,       // >>>=
			nil,       // +=
			nil,       // .=
			nil,       // -=
			nil,       // *=
			nil,       // /=
			nil,       // //=
			nil,       // %=
			nil,       // **=
			nil,       // ?
			nil,       // :
			nil,       // ||
			nil,       // ^^
			nil,       // &&
			nil,       // ??
			nil,       // ???
			nil,       // =~
			nil,       // !=~
			nil,       // ==
			nil,       // !=
			nil,       // <=>
			nil,       // >=
			nil,       // <
			nil,       // <=
			nil,       // ^
			nil,       // &
			nil,       // <<
			nil,       // >>>
			nil,       // +
			nil,       // -
			nil,       // .+
			nil,       // .-
			nil,       // *
			nil,       // /
			nil,       // //
			nil,       // %
			nil,       // .*
			nil,       // ./
			nil,       // .//
			nil,       // .
			nil,       // !
			nil,       // ~
			nil,       // **
			nil,       // string_literal
			nil,       // regex_case_insensitive
			nil,       // int_literal
			nil,       // float_literal
			nil,       // boolean_literal
			nil,       // null_literal
			nil,       // inf_literal
			nil,       // nan_literal
			nil,       // const_M_PI
			nil,       // const_M_E
			nil,       // panic
			nil,       // [
			nil,       // ctx_IPS
			nil,       // ctx_IFS
			nil,       // ctx_IRS
			nil,       // ctx_OPS
			nil,       // ctx_OFS
			nil,       // ctx_ORS
			nil,       // ctx_FLATSEP
			nil,       // ctx_NF
			nil,       // ctx_NR
			nil,       // ctx_FNR
			nil,       // ctx_FILENAME
			nil,       // ctx_FILENUM
			nil,       // env
			nil,       // [[
			nil,       // [[[
			nil,       // call
			nil,       // begin
			nil,       // end
			nil,       // if
			nil,       // elif
			nil,       // else
			nil,       // while
			nil,       // do
			nil,       // for
			nil,       // in
			nil,       // break
			nil,       // continue
			nil,       // func
			nil,       // subr
			nil,       // return
		},
	},
	actionRow{ // S869
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(234), // >, reduce: MapLiteral
			reduce(234), // >>, reduce: MapLiteral
			reduce(234), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(234), // ?, reduce: MapLiteral
			reduce(234), // :, reduce: MapLiteral
			reduce(234), // ||, reduce: MapLiteral
			reduce(234), // ^^, reduce: MapLiteral
			reduce(234), // &&, reduce: MapLiteral
			reduce(234), // ??, reduce: MapLiteral
			reduce(234), // ???, reduce: MapLiteral
			reduce(234), // =~, reduce: MapLiteral
			reduce(234), // !=~, reduce: MapLiteral
			reduce(234), // ==, reduce: MapLiteral
			reduce(234), // !=, reduce: MapLiteral
			reduce(234), // <=>, reduce: MapLiteral
			reduce(234), // >=, reduce: MapLiteral
			reduce(234), // <, reduce: MapLiteral
			reduce(234), // <=, reduce: MapLiteral
			reduce(234), // ^, reduce: MapLiteral
			reduce(234), // &, reduce: MapLiteral
			reduce(234), // <<, reduce: MapLiteral
			reduce(234), // >>>, reduce: MapLiteral
			reduce(234), // +, reduce: MapLiteral
			reduce(234), // -, reduce: MapLiteral
			reduce(234), // .+, reduce: MapLiteral
			reduce(234), // .-, reduce: MapLiteral
			reduce(234), // *, reduce: MapLiteral
			reduce(234), // /, reduce: MapLiteral
			reduce(234), // //, reduce: MapLiteral
			reduce(234), // %, reduce: MapLiteral
			reduce(234), // .*, reduce: MapLiteral
			reduce(234), // ./, reduce: MapLiteral
			reduce(234), // .//, reduce: MapLiteral
			reduce(234), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(234), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(234), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(234), // [[, reduce: MapLiteral
			reduce(234), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S870
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(1869), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S871
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S872
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(1961), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S873
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(1962), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S874
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(1963), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S875
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(1964), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S876
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(1965), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S877
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S878
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S879
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S880
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S881
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S882
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S883
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S884
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S885
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S886
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S887
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S888
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S889
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S890
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S891
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S892
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S893
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S894
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S895
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S896
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S897
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S898
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S899
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S900
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S901
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S902
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(197), // >, reduce: UnaryOpTerm
			reduce(197), // >>, reduce: UnaryOpTerm
			reduce(197), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(197), // ?, reduce: UnaryOpTerm
			reduce(197), // :, reduce: UnaryOpTerm
			reduce(197), // ||, reduce: UnaryOpTerm
			reduce(197), // ^^, reduce: UnaryOpTerm
			reduce(197), // &&, reduce: UnaryOpTerm
			reduce(197), // ??, reduce: UnaryOpTerm
			reduce(197), // ???, reduce: UnaryOpTerm
			reduce(197), // =~, reduce: UnaryOpTerm
			reduce(197), // !=~, reduce: UnaryOpTerm
			reduce(197), // ==, reduce: UnaryOpTerm
			reduce(197), // !=, reduce: UnaryOpTerm
			reduce(197), // <=>, reduce: UnaryOpTerm
			reduce(197), // >=, reduce: UnaryOpTerm
			reduce(197), // <, reduce: UnaryOpTerm
			reduce(197), // <=, reduce: UnaryOpTerm
			reduce(197), // ^, reduce: UnaryOpTerm
			reduce(197), // &, reduce: UnaryOpTerm
			reduce(197), // <<, reduce: UnaryOpTerm
			reduce(197), // >>>, reduce: UnaryOpTerm
			reduce(197), // +, reduce: UnaryOpTerm
			reduce(197), // -, reduce: UnaryOpTerm
			reduce(197), // .+, reduce: UnaryOpTerm
			reduce(197), // .-, reduce: UnaryOpTerm
			reduce(197), // *, reduce: UnaryOpTerm
			reduce(197), // /, reduce: UnaryOpTerm
			reduce(197), // //, reduce: UnaryOpTerm
			reduce(197), // %, reduce: UnaryOpTerm
			reduce(197), // .*, reduce: UnaryOpTerm
			reduce(197), // ./, reduce: UnaryOpTerm
			reduce(197), // .//, reduce: UnaryOpTerm
			reduce(197), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S903
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S904
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S905
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S906
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S907
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S908
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S909
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S910
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(198), // >, reduce: UnaryOpTerm
			reduce(198), // >>, reduce: UnaryOpTerm
			reduce(198), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(198), // ?, reduce: UnaryOpTerm
			reduce(198), // :, reduce: UnaryOpTerm
			reduce(198), // ||, reduce: UnaryOpTerm
			reduce(198), // ^^, reduce: UnaryOpTerm
			reduce(198), // &&, reduce: UnaryOpTerm
			reduce(198), // ??, reduce: UnaryOpTerm
			reduce(198), // ???, reduce: UnaryOpTerm
			reduce(198), // =~, reduce: UnaryOpTerm
			reduce(198), // !=~, reduce: UnaryOpTerm
			reduce(198), // ==, reduce: UnaryOpTerm
			reduce(198), // !=, reduce: UnaryOpTerm
			reduce(198), // <=>, reduce: UnaryOpTerm
			reduce(198), // >=, reduce: UnaryOpTerm
			reduce(198), // <, reduce: UnaryOpTerm
			reduce(198), // <=, reduce: UnaryOpTerm
			reduce(198), // ^, reduce: UnaryOpTerm
			reduce(198), // &, reduce: UnaryOpTerm
			reduce(198), // <<, reduce: UnaryOpTerm
			reduce(198), // >>>, reduce: UnaryOpTerm
			reduce(198), // +, reduce: UnaryOpTerm
			reduce(198), // -, reduce: UnaryOpTerm
			reduce(198), // .+, reduce: UnaryOpTerm
			reduce(198), // .-, reduce: UnaryOpTerm
			reduce(198), // *, reduce: UnaryOpTerm
			reduce(198), // /, reduce: UnaryOpTerm
			reduce(198), // //, reduce: UnaryOpTerm
			reduce(198), // %, reduce: UnaryOpTerm
			reduce(198), // .*, reduce: UnaryOpTerm
			reduce(198), // ./, reduce: UnaryOpTerm
			reduce(198), // .//, reduce: UnaryOpTerm
			reduce(198), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S911
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(199), // >, reduce: UnaryOpTerm
			reduce(199), // >>, reduce: UnaryOpTerm
			reduce(199), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(199), // ?, reduce: UnaryOpTerm
			reduce(199), // :, reduce: UnaryOpTerm
			reduce(199), // ||, reduce: UnaryOpTerm
			reduce(199), // ^^, reduce: UnaryOpTerm
			reduce(199), // &&, reduce: UnaryOpTerm
			reduce(199), // ??, reduce: UnaryOpTerm
			reduce(199), // ???, reduce: UnaryOpTerm
			reduce(199), // =~, reduce: UnaryOpTerm
			reduce(199), // !=~, reduce: UnaryOpTerm
			reduce(199), // ==, reduce: UnaryOpTerm
			reduce(199), // !=, reduce: UnaryOpTerm
			reduce(199), // <=>, reduce: UnaryOpTerm
			reduce(199), // >=, reduce: UnaryOpTerm
			reduce(199), // <, reduce: UnaryOpTerm
			reduce(199), // <=, reduce: UnaryOpTerm
			reduce(199), // ^, reduce: UnaryOpTerm
			reduce(199), // &, reduce: UnaryOpTerm
			reduce(199), // <<, reduce: UnaryOpTerm
			reduce(199), // >>>, reduce: UnaryOpTerm
			reduce(199), // +, reduce: UnaryOpTerm
			reduce(199), // -, reduce: UnaryOpTerm
			reduce(199), // .+, reduce: UnaryOpTerm
			reduce(199), // .-, reduce: UnaryOpTerm
			reduce(199), // *, reduce: UnaryOpTerm
			reduce(199), // /, reduce: UnaryOpTerm
			reduce(199), // //, reduce: UnaryOpTerm
			reduce(199), // %, reduce: UnaryOpTerm
			reduce(199), // .*, reduce: UnaryOpTerm
			reduce(199), // ./, reduce: UnaryOpTerm
			reduce(199), // .//, reduce: UnaryOpTerm
			reduce(199), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S912
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(200), // >, reduce: UnaryOpTerm
			reduce(200), // >>, reduce: UnaryOpTerm
			reduce(200), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(200), // ?, reduce: UnaryOpTerm
			reduce(200), // :, reduce: UnaryOpTerm
			reduce(200), // ||, reduce: UnaryOpTerm
			reduce(200), // ^^, reduce: UnaryOpTerm
			reduce(200), // &&, reduce: UnaryOpTerm
			reduce(200), // ??, reduce: UnaryOpTerm
			reduce(200), // ???, reduce: UnaryOpTerm
			reduce(200), // =~, reduce: UnaryOpTerm
			reduce(200), // !=~, reduce: UnaryOpTerm
			reduce(200), // ==, reduce: UnaryOpTerm
			reduce(200), // !=, reduce: UnaryOpTerm
			reduce(200), // <=>, reduce: UnaryOpTerm
			reduce(200), // >=, reduce: UnaryOpTerm
			reduce(200), // <, reduce: UnaryOpTerm
			reduce(200), // <=, reduce: UnaryOpTerm
			reduce(200), // ^, reduce: UnaryOpTerm
			reduce(200), // &, reduce: UnaryOpTerm
			reduce(200), // <<, reduce: UnaryOpTerm
			reduce(200), // >>>, reduce: UnaryOpTerm
			reduce(200), // +, reduce: UnaryOpTerm
			reduce(200), // -, reduce: UnaryOpTerm
			reduce(200), // .+, reduce: UnaryOpTerm
			reduce(200), // .-, reduce: UnaryOpTerm
			reduce(200), // *, reduce: UnaryOpTerm
			reduce(200), // /, reduce: UnaryOpTerm
			reduce(200), // //, reduce: UnaryOpTerm
			reduce(200), // %, reduce: UnaryOpTerm
			reduce(200), // .*, reduce: UnaryOpTerm
			reduce(200), // ./, reduce: UnaryOpTerm
			reduce(200), // .//, reduce: UnaryOpTerm
			reduce(200), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S913
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S914
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(201), // >, reduce: UnaryOpTerm
			reduce(201), // >>, reduce: UnaryOpTerm
			reduce(201), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(201), // ?, reduce: UnaryOpTerm
			reduce(201), // :, reduce: UnaryOpTerm
			reduce(201), // ||, reduce: UnaryOpTerm
			reduce(201), // ^^, reduce: UnaryOpTerm
			reduce(201), // &&, reduce: UnaryOpTerm
			reduce(201), // ??, reduce: UnaryOpTerm
			reduce(201), // ???, reduce: UnaryOpTerm
			reduce(201), // =~, reduce: UnaryOpTerm
			reduce(201), // !=~, reduce: UnaryOpTerm
			reduce(201), // ==, reduce: UnaryOpTerm
			reduce(201), // !=, reduce: UnaryOpTerm
			reduce(201), // <=>, reduce: UnaryOpTerm
			reduce(201), // >=, reduce: UnaryOpTerm
			reduce(201), // <, reduce: UnaryOpTerm
			reduce(201), // <=, reduce: UnaryOpTerm
			reduce(201), // ^, reduce: UnaryOpTerm
			reduce(201), // &, reduce: UnaryOpTerm
			reduce(201), // <<, reduce: UnaryOpTerm
			reduce(201), // >>>, reduce: UnaryOpTerm
			reduce(201), // +, reduce: UnaryOpTerm
			reduce(201), // -, reduce: UnaryOpTerm
			reduce(201), // .+, reduce: UnaryOpTerm
			reduce(201), // .-, reduce: UnaryOpTerm
			reduce(201), // *, reduce: UnaryOpTerm
			reduce(201), // /, reduce: UnaryOpTerm
			reduce(201), // //, reduce: UnaryOpTerm
			reduce(201), // %, reduce: UnaryOpTerm
			reduce(201), // .*, reduce: UnaryOpTerm
			reduce(201), // ./, reduce: UnaryOpTerm
			reduce(201), // .//, reduce: UnaryOpTerm
			reduce(201), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S915
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(202), // >, reduce: UnaryOpTerm
			reduce(202), // >>, reduce: UnaryOpTerm
			reduce(202), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(202), // ?, reduce: UnaryOpTerm
			reduce(202), // :, reduce: UnaryOpTerm
			reduce(202), // ||, reduce: UnaryOpTerm
			reduce(202), // ^^, reduce: UnaryOpTerm
			reduce(202), // &&, reduce: UnaryOpTerm
			reduce(202), // ??, reduce: UnaryOpTerm
			reduce(202), // ???, reduce: UnaryOpTerm
			reduce(202), // =~, reduce: UnaryOpTerm
			reduce(202), // !=~, reduce: UnaryOpTerm
			reduce(202), // ==, reduce: UnaryOpTerm
			reduce(202), // !=, reduce: UnaryOpTerm
			reduce(202), // <=>, reduce: UnaryOpTerm
			reduce(202), // >=, reduce: UnaryOpTerm
			reduce(202), // <, reduce: UnaryOpTerm
			reduce(202), // <=, reduce: UnaryOpTerm
			reduce(202), // ^, reduce: UnaryOpTerm
			reduce(202), // &, reduce: UnaryOpTerm
			reduce(202), // <<, reduce: UnaryOpTerm
			reduce(202), // >>>, reduce: UnaryOpTerm
			reduce(202), // +, reduce: UnaryOpTerm
			reduce(202), // -, reduce: UnaryOpTerm
			reduce(202), // .+, reduce: UnaryOpTerm
			reduce(202), // .-, reduce: UnaryOpTerm
			reduce(202), // *, reduce: UnaryOpTerm
			reduce(202), // /, reduce: UnaryOpTerm
			reduce(202), // //, reduce: UnaryOpTerm
			reduce(202), // %, reduce: UnaryOpTerm
			reduce(202), // .*, reduce: UnaryOpTerm
			reduce(202), // ./, reduce: UnaryOpTerm
			reduce(202), // .//, reduce: UnaryOpTerm
			reduce(202), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S916
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1999), // +
			shift(2000), // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S917
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(2003), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S918
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S919
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S920
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(228), // >, reduce: ArrayLiteral
			reduce(228), // >>, reduce: ArrayLiteral
			reduce(228), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(228), // ?, reduce: ArrayLiteral
			reduce(228), // :, reduce: ArrayLiteral
			reduce(228), // ||, reduce: ArrayLiteral
			reduce(228), // ^^, reduce: ArrayLiteral
			reduce(228), // &&, reduce: ArrayLiteral
			reduce(228), // ??, reduce: ArrayLiteral
			reduce(228), // ???, reduce: ArrayLiteral
			reduce(228), // =~, reduce: ArrayLiteral
			reduce(228), // !=~, reduce: ArrayLiteral
			reduce(228), // ==, reduce: ArrayLiteral
			reduce(228), // !=, reduce: ArrayLiteral
			reduce(228), // <=>, reduce: ArrayLiteral
			reduce(228), // >=, reduce: ArrayLiteral
			reduce(228), // <, reduce: ArrayLiteral
			reduce(228), // <=, reduce: ArrayLiteral
			reduce(228), // ^, reduce: ArrayLiteral
			reduce(228), // &, reduce: ArrayLiteral
			reduce(228), // <<, reduce: ArrayLiteral
			reduce(228), // >>>, reduce: ArrayLiteral
			reduce(228), // +, reduce: ArrayLiteral
			reduce(228), // -, reduce: ArrayLiteral
			reduce(228), // .+, reduce: ArrayLiteral
			reduce(228), // .-, reduce: ArrayLiteral
			reduce(228), // *, reduce: ArrayLiteral
			reduce(228), // /, reduce: ArrayLiteral
			reduce(228), // //, reduce: ArrayLiteral
			reduce(228), // %, reduce: ArrayLiteral
			reduce(228), // .*, reduce: ArrayLiteral
			reduce(228), // ./, reduce: ArrayLiteral
			reduce(228), // .//, reduce: ArrayLiteral
			reduce(228), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(228), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(228), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(228), // [[, reduce: ArrayLiteral
			reduce(228), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S921
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2006), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S922
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(235), // $, reduce: MapLiteral
			nil,         // empty
			reduce(235), // ;, reduce: MapLiteral
			reduce(235), // {, reduce: MapLiteral
			nil,         // }
			reduce(235), // =, reduce: MapLiteral
			nil,         // unset
			nil,         // filter
			reduce(235), // >, reduce: MapLiteral
			reduce(235), // >>, reduce: MapLiteral
			reduce(235), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(235), // ||=, reduce: MapLiteral
			reduce(235), // ^^=, reduce: MapLiteral
			reduce(235), // &&=, reduce: MapLiteral
			reduce(235), // ??=, reduce: MapLiteral
			reduce(235), // ???=, reduce: MapLiteral
			reduce(235), // |=, reduce: MapLiteral
			reduce(235), // &=, reduce: MapLiteral
			reduce(235), // ^=, reduce: MapLiteral
			reduce(235), // <<=, reduce: MapLiteral
			reduce(235), // >>=, reduce: MapLiteral
			reduce(235), // >>>=, reduce: MapLiteral
			reduce(235), // +=, reduce: MapLiteral
			reduce(235), // .=, reduce: MapLiteral
			reduce(235), // -=, reduce: MapLiteral
			reduce(235), // *=, reduce: MapLiteral
			reduce(235), // /=, reduce: MapLiteral
			reduce(235), // //=, reduce: MapLiteral
			reduce(235), // %=, reduce: MapLiteral
			reduce(235), // **=, reduce: MapLiteral
			reduce(235), // ?, reduce: MapLiteral
			nil,         // :
			reduce(235), // ||, reduce: MapLiteral
			reduce(235), // ^^, reduce: MapLiteral
			reduce(235), // &&, reduce: MapLiteral
			reduce(235), // ??, reduce: MapLiteral
			reduce(235), // ???, reduce: MapLiteral
			reduce(235), // =~, reduce: MapLiteral
			reduce(235), // !=~, reduce: MapLiteral
			reduce(235), // ==, reduce: MapLiteral
			reduce(235), // !=, reduce: MapLiteral
			reduce(235), // <=>, reduce: MapLiteral
			reduce(235), // >=, reduce: MapLiteral
			reduce(235), // <, reduce: MapLiteral
			reduce(235), // <=, reduce: MapLiteral
			reduce(235), // ^, reduce: MapLiteral
			reduce(235), // &, reduce: MapLiteral
			reduce(235), // <<, reduce: MapLiteral
			reduce(235), // >>>, reduce: MapLiteral
			reduce(235), // +, reduce: MapLiteral
			reduce(235), // -, reduce: MapLiteral
			reduce(235), // .+, reduce: MapLiteral
			reduce(235), // .-, reduce: MapLiteral
			reduce(235), // *, reduce: MapLiteral
			reduce(235), // /, reduce: MapLiteral
			reduce(235), // //, reduce: MapLiteral
			reduce(235), // %, reduce: MapLiteral
			reduce(235), // .*, reduce: MapLiteral
			reduce(235), // ./, reduce: MapLiteral
			reduce(235), // .//, reduce: MapLiteral
			reduce(235), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(235), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(235), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(235), // [[, reduce: MapLiteral
			reduce(235), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S923
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			reduce(237), // }, reduce: MapLiteralKeyValuePairs
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S924
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(2008), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S925
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S926
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(2011), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S927
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S928
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(34), // $, reduce: Assignment
			nil,        // empty
			reduce(34), // ;, reduce: Assignment
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S929
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(128), // $, reduce: Assignment
			nil,         // empty
			reduce(128), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S930
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(129), // $, reduce: Assignment
			nil,         // empty
			reduce(129), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S931
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(130), // $, reduce: Assignment
			nil,         // empty
			reduce(130), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S932
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(131), // $, reduce: Assignment
			nil,         // empty
			reduce(131), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S933
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(132), // $, reduce: Assignment
			nil,         // empty
			reduce(132), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S934
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(133), // $, reduce: Assignment
			nil,         // empty
			reduce(133), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S935
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(134), // $, reduce: Assignment
			nil,         // empty
			reduce(134), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S936
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(135), // $, reduce: Assignment
			nil,         // empty
			reduce(135), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S937
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(136), // $, reduce: Assignment
			nil,         // empty
			reduce(136), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S938
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(137), // $, reduce: Assignment
			nil,         // empty
			reduce(137), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S939
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(138), // $, reduce: Assignment
			nil,         // empty
			reduce(138), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S940
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(139), // $, reduce: Assignment
			nil,         // empty
			reduce(139), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S941
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(140), // $, reduce: Assignment
			nil,         // empty
			reduce(140), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S942
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(141), // $, reduce: Assignment
			nil,         // empty
			reduce(141), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S943
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(142), // $, reduce: Assignment
			nil,         // empty
			reduce(142), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S944
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(143), // $, reduce: Assignment
			nil,         // empty
			reduce(143), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S945
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(144), // $, reduce: Assignment
			nil,         // empty
			reduce(144), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S946
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(145), // $, reduce: Assignment
			nil,         // empty
			reduce(145), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S947
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(146), // $, reduce: Assignment
			nil,         // empty
			reduce(146), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S948
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(2013), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S949
		canRecover: false,
		actions: [numSymbols]action{
			nil,       // INVALID
			nil,       // $
			nil,       // empty
			nil,       // ;
			nil,       // {
			reduce(3), // }, reduce: StatementBlock
			nil,       // =
			nil,       // unset
			nil,       // filter
			nil,       // >
			nil,       // >>
			nil,       // |
			nil,       // stdout
			nil,       // stderr
			nil,       // print
			nil,       // ,
			nil,       // printn
			nil,       // eprint
			nil,       // eprintn
			nil,       // dump
			nil,       // edump
			nil,       // tee
			nil,       // emitf
			nil,       // emit1
			nil,       // emit
			nil,       // (
			nil,       // )
			nil,       // emitp
			nil,       // field_name
			nil,       // $[
			nil,       // ]
			nil,       // braced_field_name
			nil,       // $[[
			nil,       // $[[[
			nil,       // full_srec
			nil,       // oosvar_name
			nil,       // @[
			nil,       // braced_oosvar_name
			nil,       // full_oosvar
			nil,       // all
			nil,       // non_sigil_name
			nil,       // arr
			nil,       // bool
			nil,       // float
			nil,       // int
			nil,       // map
			nil,       // num
			nil,       // str
			nil,       // var
			nil,       // funct
			nil,       // ||=
			nil,       // ^^=
			nil,       // &&=
			nil,       // ??=
			nil,       // ???=
			nil,       // |=
			nil,       // &=
			nil,       // ^=
			nil,       // <<=
			nil,       // >>=
			nil,       // >>>=
			nil,       // +=
			nil,       // .=
			nil,       // -=
			nil,       // *=
			nil,       // /=
			nil,       // //=
			nil,       // %=
			nil,       // **=
			nil,       // ?
			nil,       // :
			nil,       // ||
			nil,       // ^^
			nil,       // &&
			nil,       // ??
			nil,       // ???
			nil,       // =~
			nil,       // !=~
			nil,       // ==
			nil,       // !=
			nil,       // <=>
			nil,       // >=
			nil,       // <
			nil,       // <=
			nil,       // ^
			nil,       // &
			nil,       // <<
			nil,       // >>>
			nil,       // +
			nil,       // -
			nil,       // .+
			nil,       // .-
			nil,       // *
			nil,       // /
			nil,       // //
			nil,       // %
			nil,       // .*
			nil,       // ./
			nil,       // .//
			nil,       // .
			nil,       // !
			nil,       // ~
			nil,       // **
			nil,       // string_literal
			nil,       // regex_case_insensitive
			nil,       // int_literal
			nil,       // float_literal
			nil,       // boolean_literal
			nil,       // null_literal
			nil,       // inf_literal
			nil,       // nan_literal
			nil,       // const_M_PI
			nil,       // const_M_E
			nil,       // panic
			nil,       // [
			nil,       // ctx_IPS
			nil,       // ctx_IFS
			nil,       // ctx_IRS
			nil,       // ctx_OPS
			nil,       // ctx_OFS
			nil,       // ctx_ORS
			nil,       // ctx_FLATSEP
			nil,       // ctx_NF
			nil,       // ctx_NR
			nil,       // ctx_FNR
			nil,       // ctx_FILENAME
			nil,       // ctx_FILENUM
			nil,       // env
			nil,       // [[
			nil,       // [[[
			nil,       // call
			nil,       // begin
			nil,       // end
			nil,       // if
			nil,       // elif
			nil,       // else
			nil,       // while
			nil,       // do
			nil,       // for
			nil,       // in
			nil,       // break
			nil,       // continue
			nil,       // func
			nil,       // subr
			nil,       // return
		},
	},
	actionRow{ // S950
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(2014), // ;
			nil,         // {
			reduce(4),   // }, reduce: NonEmptyStatementBlock
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S951
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(2017), // ;
			shift(953),  // {
			reduce(5),   // }, reduce: NonEmptyStatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S952
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S953
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			shift(2019), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S954
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(14), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(14), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S955
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(15), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(15), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S956
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(16), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(16), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S957
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(17), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(17), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S958
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(18), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(18), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S959
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(19), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(19), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S960
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(20), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(20), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S961
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(21), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(21), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S962
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(22), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(22), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S963
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(23), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(23), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S964
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(24), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(24), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S965
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(25), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(25), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S966
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(26), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(26), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S967
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(27), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(27), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S968
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(28), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(28), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S969
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(29), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(29), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S970
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(30), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(30), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S971
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(31), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(31), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S972
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(32), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(32), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S973
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(33), // ;, reduce: BracelessStatement
			nil,        // {
			reduce(33), // }, reduce: BracelessStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S974
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			shift(2021), // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			shift(2022), // ||=
			shift(2023), // ^^=
			shift(2024), // &&=
			shift(2025), // ??=
			shift(2026), // ???=
			shift(2027), // |=
			shift(2028), // &=
			shift(2029), // ^=
			shift(2030), // <<=
			shift(2031), // >>=
			shift(2032), // >>>=
			shift(2033), // +=
			shift(2034), // .=
			shift(2035), // -=
			shift(2036), // *=
			shift(2037), // /=
			shift(2038), // //=
			shift(2039), // %=
			shift(2040), // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S975
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(38),  // ;, reduce: BareBoolean
			shift(2042), // {
			reduce(38),  // }, reduce: BareBoolean
			reduce(36),  // =, reduce: Lvalue
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(36),  // ||=, reduce: Lvalue
			reduce(36),  // ^^=, reduce: Lvalue
			reduce(36),  // &&=, reduce: Lvalue
			reduce(36),  // ??=, reduce: Lvalue
			reduce(36),  // ???=, reduce: Lvalue
			reduce(36),  // |=, reduce: Lvalue
			reduce(36),  // &=, reduce: Lvalue
			reduce(36),  // ^=, reduce: Lvalue
			reduce(36),  // <<=, reduce: Lvalue
			reduce(36),  // >>=, reduce: Lvalue
			reduce(36),  // >>>=, reduce: Lvalue
			reduce(36),  // +=, reduce: Lvalue
			reduce(36),  // .=, reduce: Lvalue
			reduce(36),  // -=, reduce: Lvalue
			reduce(36),  // *=, reduce: Lvalue
			reduce(36),  // /=, reduce: Lvalue
			reduce(36),  // //=, reduce: Lvalue
			reduce(36),  // %=, reduce: Lvalue
			reduce(36),  // **=, reduce: Lvalue
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S976
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S977
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(214), // ;, reduce: MlrvalOrFunction
			reduce(214), // {, reduce: MlrvalOrFunction
			reduce(214), // }, reduce: MlrvalOrFunction
			reduce(214), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(214), // ||=, reduce: MlrvalOrFunction
			reduce(214), // ^^=, reduce: MlrvalOrFunction
			reduce(214), // &&=, reduce: MlrvalOrFunction
			reduce(214), // ??=, reduce: MlrvalOrFunction
			reduce(214), // ???=, reduce: MlrvalOrFunction
			reduce(214), // |=, reduce: MlrvalOrFunction
			reduce(214), // &=, reduce: MlrvalOrFunction
			reduce(214), // ^=, reduce: MlrvalOrFunction
			reduce(214), // <<=, reduce: MlrvalOrFunction
			reduce(214), // >>=, reduce: MlrvalOrFunction
			reduce(214), // >>>=, reduce: MlrvalOrFunction
			reduce(214), // +=, reduce: MlrvalOrFunction
			reduce(214), // .=, reduce: MlrvalOrFunction
			reduce(214), // -=, reduce: MlrvalOrFunction
			reduce(214), // *=, reduce: MlrvalOrFunction
			reduce(214), // /=, reduce: MlrvalOrFunction
			reduce(214), // //=, reduce: MlrvalOrFunction
			reduce(214), // %=, reduce: MlrvalOrFunction
			reduce(214), // **=, reduce: MlrvalOrFunction
			reduce(214), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S978
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S979
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(46),  // ;, reduce: PrintStatement
			shift(2043), // {
			reduce(46),  // }, reduce: PrintStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2228), // >
			shift(2229), // >>
			shift(2230), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S980
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(50),  // ;, reduce: PrintnStatement
			shift(2043), // {
			reduce(50),  // }, reduce: PrintnStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2228), // >
			shift(2229), // >>
			shift(2230), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S981
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(54),  // ;, reduce: EprintStatement
			shift(2043), // {
			reduce(54),  // }, reduce: EprintStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S982
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(56),  // ;, reduce: EprintnStatement
			shift(2043), // {
			reduce(56),  // }, reduce: EprintnStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S983
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(58),  // ;, reduce: DumpStatement
			shift(2043), // {
			reduce(58),  // }, reduce: DumpStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2228), // >
			shift(2229), // >>
			shift(2230), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S984
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(62),  // ;, reduce: EdumpStatement
			shift(2043), // {
			reduce(62),  // }, reduce: EdumpStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S985
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			shift(488), // >
			shift(489), // >>
			shift(490), // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S986
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(211), // ;, reduce: MlrvalOrFunction
			reduce(211), // {, reduce: MlrvalOrFunction
			reduce(211), // }, reduce: MlrvalOrFunction
			reduce(211), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(211), // >, reduce: MlrvalOrFunction
			reduce(211), // >>, reduce: MlrvalOrFunction
			reduce(211), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(211), // ||=, reduce: MlrvalOrFunction
			reduce(211), // ^^=, reduce: MlrvalOrFunction
			reduce(211), // &&=, reduce: MlrvalOrFunction
			reduce(211), // ??=, reduce: MlrvalOrFunction
			reduce(211), // ???=, reduce: MlrvalOrFunction
			reduce(211), // |=, reduce: MlrvalOrFunction
			reduce(211), // &=, reduce: MlrvalOrFunction
			reduce(211), // ^=, reduce: MlrvalOrFunction
			reduce(211), // <<=, reduce: MlrvalOrFunction
			reduce(211), // >>=, reduce: MlrvalOrFunction
			reduce(211), // >>>=, reduce: MlrvalOrFunction
			reduce(211), // +=, reduce: MlrvalOrFunction
			reduce(211), // .=, reduce: MlrvalOrFunction
			reduce(211), // -=, reduce: MlrvalOrFunction
			reduce(211), // *=, reduce: MlrvalOrFunction
			reduce(211), // /=, reduce: MlrvalOrFunction
			reduce(211), // //=, reduce: MlrvalOrFunction
			reduce(211), // %=, reduce: MlrvalOrFunction
			reduce(211), // **=, reduce: MlrvalOrFunction
			reduce(211), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(211), // ||, reduce: MlrvalOrFunction
			reduce(211), // ^^, reduce: MlrvalOrFunction
			reduce(211), // &&, reduce: MlrvalOrFunction
			reduce(211), // ??, reduce: MlrvalOrFunction
			reduce(211), // ???, reduce: MlrvalOrFunction
			reduce(211), // =~, reduce: MlrvalOrFunction
			reduce(211), // !=~, reduce: MlrvalOrFunction
			reduce(211), // ==, reduce: MlrvalOrFunction
			reduce(211), // !=, reduce: MlrvalOrFunction
			reduce(211), // <=>, reduce: MlrvalOrFunction
			reduce(211), // >=, reduce: MlrvalOrFunction
			reduce(211), // <, reduce: MlrvalOrFunction
			reduce(211), // <=, reduce: MlrvalOrFunction
			reduce(211), // ^, reduce: MlrvalOrFunction
			reduce(211), // &, reduce: MlrvalOrFunction
			reduce(211), // <<, reduce: MlrvalOrFunction
			reduce(211), // >>>, reduce: MlrvalOrFunction
			reduce(211), // +, reduce: MlrvalOrFunction
			reduce(211), // -, reduce: MlrvalOrFunction
			reduce(211), // .+, reduce: MlrvalOrFunction
			reduce(211), // .-, reduce: MlrvalOrFunction
			reduce(211), // *, reduce: MlrvalOrFunction
			reduce(211), // /, reduce: MlrvalOrFunction
			reduce(211), // //, reduce: MlrvalOrFunction
			reduce(211), // %, reduce: MlrvalOrFunction
			reduce(211), // .*, reduce: MlrvalOrFunction
			reduce(211), // ./, reduce: MlrvalOrFunction
			reduce(211), // .//, reduce: MlrvalOrFunction
			reduce(211), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(211), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(211), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(211), // [[, reduce: MlrvalOrFunction
			reduce(211), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S987
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2239), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(488),  // >
			shift(489),  // >>
			shift(490),  // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			shift(2253), // field_name
			shift(2254), // $[
			nil,         // ]
			shift(2255), // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(2256), // full_srec
			shift(2257), // oosvar_name
			shift(2258), // @[
			shift(2259), // braced_oosvar_name
			shift(2260), // full_oosvar
			shift(2261), // all
			shift(2262), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S988
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S989
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2239), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(488),  // >
			shift(489),  // >>
			shift(490),  // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2266), // (
			nil,         // )
			nil,         // emitp
			shift(2253), // field_name
			shift(2254), // $[
			nil,         // ]
			shift(2255), // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(2256), // full_srec
			shift(2257), // oosvar_name
			shift(2258), // @[
			shift(2259), // braced_oosvar_name
			shift(2260), // full_oosvar
			shift(2261), // all
			shift(2262), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S990
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S991
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2239), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(488),  // >
			shift(489),  // >>
			shift(490),  // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2271), // (
			nil,         // )
			nil,         // emitp
			shift(2253), // field_name
			shift(2254), // $[
			nil,         // ]
			shift(2255), // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(2256), // full_srec
			shift(2257), // oosvar_name
			shift(2258), // @[
			shift(2259), // braced_oosvar_name
			shift(2260), // full_oosvar
			shift(2261), // all
			shift(2262), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S992
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(110), // ;, reduce: OosvarValue
			reduce(110), // {, reduce: OosvarValue
			reduce(110), // }, reduce: OosvarValue
			reduce(110), // =, reduce: OosvarValue
			nil,         // unset
			nil,         // filter
			reduce(110), // >, reduce: OosvarValue
			reduce(110), // >>, reduce: OosvarValue
			reduce(110), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(110), // ||=, reduce: OosvarValue
			reduce(110), // ^^=, reduce: OosvarValue
			reduce(110), // &&=, reduce: OosvarValue
			reduce(110), // ??=, reduce: OosvarValue
			reduce(110), // ???=, reduce: OosvarValue
			reduce(110), // |=, reduce: OosvarValue
			reduce(110), // &=, reduce: OosvarValue
			reduce(110), // ^=, reduce: OosvarValue
			reduce(110), // <<=, reduce: OosvarValue
			reduce(110), // >>=, reduce: OosvarValue
			reduce(110), // >>>=, reduce: OosvarValue
			reduce(110), // +=, reduce: OosvarValue
			reduce(110), // .=, reduce: OosvarValue
			reduce(110), // -=, reduce: OosvarValue
			reduce(110), // *=, reduce: OosvarValue
			reduce(110), // /=, reduce: OosvarValue
			reduce(110), // //=, reduce: OosvarValue
			reduce(110), // %=, reduce: OosvarValue
			reduce(110), // **=, reduce: OosvarValue
			reduce(110), // ?, reduce: OosvarValue
			nil,         // :
			reduce(110), // ||, reduce: OosvarValue
			reduce(110), // ^^, reduce: OosvarValue
			reduce(110), // &&, reduce: OosvarValue
			reduce(110), // ??, reduce: OosvarValue
			reduce(110), // ???, reduce: OosvarValue
			reduce(110), // =~, reduce: OosvarValue
			reduce(110), // !=~, reduce: OosvarValue
			reduce(110), // ==, reduce: OosvarValue
			reduce(110), // !=, reduce: OosvarValue
			reduce(110), // <=>, reduce: OosvarValue
			reduce(110), // >=, reduce: OosvarValue
			reduce(110), // <, reduce: OosvarValue
			reduce(110), // <=, reduce: OosvarValue
			reduce(110), // ^, reduce: OosvarValue
			reduce(110), // &, reduce: OosvarValue
			reduce(110), // <<, reduce: OosvarValue
			reduce(110), // >>>, reduce: OosvarValue
			reduce(110), // +, reduce: OosvarValue
			reduce(110), // -, reduce: OosvarValue
			reduce(110), // .+, reduce: OosvarValue
			reduce(110), // .-, reduce: OosvarValue
			reduce(110), // *, reduce: OosvarValue
			reduce(110), // /, reduce: OosvarValue
			reduce(110), // //, reduce: OosvarValue
			reduce(110), // %, reduce: OosvarValue
			reduce(110), // .*, reduce: OosvarValue
			reduce(110), // ./, reduce: OosvarValue
			reduce(110), // .//, reduce: OosvarValue
			reduce(110), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(110), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(110), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(110), // [[, reduce: OosvarValue
			reduce(110), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S993
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(112), // ;, reduce: OosvarValue
			reduce(112), // {, reduce: OosvarValue
			reduce(112), // }, reduce: OosvarValue
			reduce(112), // =, reduce: OosvarValue
			nil,         // unset
			nil,         // filter
			reduce(112), // >, reduce: OosvarValue
			reduce(112), // >>, reduce: OosvarValue
			reduce(112), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(112), // ||=, reduce: OosvarValue
			reduce(112), // ^^=, reduce: OosvarValue
			reduce(112), // &&=, reduce: OosvarValue
			reduce(112), // ??=, reduce: OosvarValue
			reduce(112), // ???=, reduce: OosvarValue
			reduce(112), // |=, reduce: OosvarValue
			reduce(112), // &=, reduce: OosvarValue
			reduce(112), // ^=, reduce: OosvarValue
			reduce(112), // <<=, reduce: OosvarValue
			reduce(112), // >>=, reduce: OosvarValue
			reduce(112), // >>>=, reduce: OosvarValue
			reduce(112), // +=, reduce: OosvarValue
			reduce(112), // .=, reduce: OosvarValue
			reduce(112), // -=, reduce: OosvarValue
			reduce(112), // *=, reduce: OosvarValue
			reduce(112), // /=, reduce: OosvarValue
			reduce(112), // //=, reduce: OosvarValue
			reduce(112), // %=, reduce: OosvarValue
			reduce(112), // **=, reduce: OosvarValue
			reduce(112), // ?, reduce: OosvarValue
			nil,         // :
			reduce(112), // ||, reduce: OosvarValue
			reduce(112), // ^^, reduce: OosvarValue
			reduce(112), // &&, reduce: OosvarValue
			reduce(112), // ??, reduce: OosvarValue
			reduce(112), // ???, reduce: OosvarValue
			reduce(112), // =~, reduce: OosvarValue
			reduce(112), // !=~, reduce: OosvarValue
			reduce(112), // ==, reduce: OosvarValue
			reduce(112), // !=, reduce: OosvarValue
			reduce(112), // <=>, reduce: OosvarValue
			reduce(112), // >=, reduce: OosvarValue
			reduce(112), // <, reduce: OosvarValue
			reduce(112), // <=, reduce: OosvarValue
			reduce(112), // ^, reduce: OosvarValue
			reduce(112), // &, reduce: OosvarValue
			reduce(112), // <<, reduce: OosvarValue
			reduce(112), // >>>, reduce: OosvarValue
			reduce(112), // +, reduce: OosvarValue
			reduce(112), // -, reduce: OosvarValue
			reduce(112), // .+, reduce: OosvarValue
			reduce(112), // .-, reduce: OosvarValue
			reduce(112), // *, reduce: OosvarValue
			reduce(112), // /, reduce: OosvarValue
			reduce(112), // //, reduce: OosvarValue
			reduce(112), // %, reduce: OosvarValue
			reduce(112), // .*, reduce: OosvarValue
			reduce(112), // ./, reduce: OosvarValue
			reduce(112), // .//, reduce: OosvarValue
			reduce(112), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(112), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(112), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(112), // [[, reduce: OosvarValue
			reduce(112), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S994
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(111), // ;, reduce: OosvarValue
			reduce(111), // {, reduce: OosvarValue
			reduce(111), // }, reduce: OosvarValue
			reduce(111), // =, reduce: OosvarValue
			nil,         // unset
			nil,         // filter
			reduce(111), // >, reduce: OosvarValue
			reduce(111), // >>, reduce: OosvarValue
			reduce(111), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(111), // ||=, reduce: OosvarValue
			reduce(111), // ^^=, reduce: OosvarValue
			reduce(111), // &&=, reduce: OosvarValue
			reduce(111), // ??=, reduce: OosvarValue
			reduce(111), // ???=, reduce: OosvarValue
			reduce(111), // |=, reduce: OosvarValue
			reduce(111), // &=, reduce: OosvarValue
			reduce(111), // ^=, reduce: OosvarValue
			reduce(111), // <<=, reduce: OosvarValue
			reduce(111), // >>=, reduce: OosvarValue
			reduce(111), // >>>=, reduce: OosvarValue
			reduce(111), // +=, reduce: OosvarValue
			reduce(111), // .=, reduce: OosvarValue
			reduce(111), // -=, reduce: OosvarValue
			reduce(111), // *=, reduce: OosvarValue
			reduce(111), // /=, reduce: OosvarValue
			reduce(111), // //=, reduce: OosvarValue
			reduce(111), // %=, reduce: OosvarValue
			reduce(111), // **=, reduce: OosvarValue
			reduce(111), // ?, reduce: OosvarValue
			nil,         // :
			reduce(111), // ||, reduce: OosvarValue
			reduce(111), // ^^, reduce: OosvarValue
			reduce(111), // &&, reduce: OosvarValue
			reduce(111), // ??, reduce: OosvarValue
			reduce(111), // ???, reduce: OosvarValue
			reduce(111), // =~, reduce: OosvarValue
			reduce(111), // !=~, reduce: OosvarValue
			reduce(111), // ==, reduce: OosvarValue
			reduce(111), // !=, reduce: OosvarValue
			reduce(111), // <=>, reduce: OosvarValue
			reduce(111), // >=, reduce: OosvarValue
			reduce(111), // <, reduce: OosvarValue
			reduce(111), // <=, reduce: OosvarValue
			reduce(111), // ^, reduce: OosvarValue
			reduce(111), // &, reduce: OosvarValue
			reduce(111), // <<, reduce: OosvarValue
			reduce(111), // >>>, reduce: OosvarValue
			reduce(111), // +, reduce: OosvarValue
			reduce(111), // -, reduce: OosvarValue
			reduce(111), // .+, reduce: OosvarValue
			reduce(111), // .-, reduce: OosvarValue
			reduce(111), // *, reduce: OosvarValue
			reduce(111), // /, reduce: OosvarValue
			reduce(111), // //, reduce: OosvarValue
			reduce(111), // %, reduce: OosvarValue
			reduce(111), // .*, reduce: OosvarValue
			reduce(111), // ./, reduce: OosvarValue
			reduce(111), // .//, reduce: OosvarValue
			reduce(111), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(111), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(111), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(111), // [[, reduce: OosvarValue
			reduce(111), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S995
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(99), // ;, reduce: FieldValue
			reduce(99), // {, reduce: FieldValue
			reduce(99), // }, reduce: FieldValue
			reduce(99), // =, reduce: FieldValue
			nil,        // unset
			nil,        // filter
			reduce(99), // >, reduce: FieldValue
			reduce(99), // >>, reduce: FieldValue
			reduce(99), // |, reduce: FieldValue
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			reduce(99), // ||=, reduce: FieldValue
			reduce(99), // ^^=, reduce: FieldValue
			reduce(99), // &&=, reduce: FieldValue
			reduce(99), // ??=, reduce: FieldValue
			reduce(99), // ???=, reduce: FieldValue
			reduce(99), // |=, reduce: FieldValue
			reduce(99), // &=, reduce: FieldValue
			reduce(99), // ^=, reduce: FieldValue
			reduce(99), // <<=, reduce: FieldValue
			reduce(99), // >>=, reduce: FieldValue
			reduce(99), // >>>=, reduce: FieldValue
			reduce(99), // +=, reduce: FieldValue
			reduce(99), // .=, reduce: FieldValue
			reduce(99), // -=, reduce: FieldValue
			reduce(99), // *=, reduce: FieldValue
			reduce(99), // /=, reduce: FieldValue
			reduce(99), // //=, reduce: FieldValue
			reduce(99), // %=, reduce: FieldValue
			reduce(99), // **=, reduce: FieldValue
			reduce(99), // ?, reduce: FieldValue
			nil,        // :
			reduce(99), // ||, reduce: FieldValue
			reduce(99), // ^^, reduce: FieldValue
			reduce(99), // &&, reduce: FieldValue
			reduce(99), // ??, reduce: FieldValue
			reduce(99), // ???, reduce: FieldValue
			reduce(99), // =~, reduce: FieldValue
			reduce(99), // !=~, reduce: FieldValue
			reduce(99), // ==, reduce: FieldValue
			reduce(99), // !=, reduce: FieldValue
			reduce(99), // <=>, reduce: FieldValue
			reduce(99), // >=, reduce: FieldValue
			reduce(99), // <, reduce: FieldValue
			reduce(99), // <=, reduce: FieldValue
			reduce(99), // ^, reduce: FieldValue
			reduce(99), // &, reduce: FieldValue
			reduce(99), // <<, reduce: FieldValue
			reduce(99), // >>>, reduce: FieldValue
			reduce(99), // +, reduce: FieldValue
			reduce(99), // -, reduce: FieldValue
			reduce(99), // .+, reduce: FieldValue
			reduce(99), // .-, reduce: FieldValue
			reduce(99), // *, reduce: FieldValue
			reduce(99), // /, reduce: FieldValue
			reduce(99), // //, reduce: FieldValue
			reduce(99), // %, reduce: FieldValue
			reduce(99), // .*, reduce: FieldValue
			reduce(99), // ./, reduce: FieldValue
			reduce(99), // .//, reduce: FieldValue
			reduce(99), // ., reduce: FieldValue
			nil,        // !
			nil,        // ~
			reduce(99), // **, reduce: FieldValue
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(99), // [, reduce: FieldValue
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(99), // [[, reduce: FieldValue
			reduce(99), // [[[, reduce: FieldValue
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S996
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(101), // ;, reduce: FieldValue
			reduce(101), // {, reduce: FieldValue
			reduce(101), // }, reduce: FieldValue
			reduce(101), // =, reduce: FieldValue
			nil,         // unset
			nil,         // filter
			reduce(101), // >, reduce: FieldValue
			reduce(101), // >>, reduce: FieldValue
			reduce(101), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(101), // ||=, reduce: FieldValue
			reduce(101), // ^^=, reduce: FieldValue
			reduce(101), // &&=, reduce: FieldValue
			reduce(101), // ??=, reduce: FieldValue
			reduce(101), // ???=, reduce: FieldValue
			reduce(101), // |=, reduce: FieldValue
			reduce(101), // &=, reduce: FieldValue
			reduce(101), // ^=, reduce: FieldValue
			reduce(101), // <<=, reduce: FieldValue
			reduce(101), // >>=, reduce: FieldValue
			reduce(101), // >>>=, reduce: FieldValue
			reduce(101), // +=, reduce: FieldValue
			reduce(101), // .=, reduce: FieldValue
			reduce(101), // -=, reduce: FieldValue
			reduce(101), // *=, reduce: FieldValue
			reduce(101), // /=, reduce: FieldValue
			reduce(101), // //=, reduce: FieldValue
			reduce(101), // %=, reduce: FieldValue
			reduce(101), // **=, reduce: FieldValue
			reduce(101), // ?, reduce: FieldValue
			nil,         // :
			reduce(101), // ||, reduce: FieldValue
			reduce(101), // ^^, reduce: FieldValue
			reduce(101), // &&, reduce: FieldValue
			reduce(101), // ??, reduce: FieldValue
			reduce(101), // ???, reduce: FieldValue
			reduce(101), // =~, reduce: FieldValue
			reduce(101), // !=~, reduce: FieldValue
			reduce(101), // ==, reduce: FieldValue
			reduce(101), // !=, reduce: FieldValue
			reduce(101), // <=>, reduce: FieldValue
			reduce(101), // >=, reduce: FieldValue
			reduce(101), // <, reduce: FieldValue
			reduce(101), // <=, reduce: FieldValue
			reduce(101), // ^, reduce: FieldValue
			reduce(101), // &, reduce: FieldValue
			reduce(101), // <<, reduce: FieldValue
			reduce(101), // >>>, reduce: FieldValue
			reduce(101), // +, reduce: FieldValue
			reduce(101), // -, reduce: FieldValue
			reduce(101), // .+, reduce: FieldValue
			reduce(101), // .-, reduce: FieldValue
			reduce(101), // *, reduce: FieldValue
			reduce(101), // /, reduce: FieldValue
			reduce(101), // //, reduce: FieldValue
			reduce(101), // %, reduce: FieldValue
			reduce(101), // .*, reduce: FieldValue
			reduce(101), // ./, reduce: FieldValue
			reduce(101), // .//, reduce: FieldValue
			reduce(101), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(101), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(101), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(101), // [[, reduce: FieldValue
			reduce(101), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S997
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(100), // ;, reduce: FieldValue
			reduce(100), // {, reduce: FieldValue
			reduce(100), // }, reduce: FieldValue
			reduce(100), // =, reduce: FieldValue
			nil,         // unset
			nil,         // filter
			reduce(100), // >, reduce: FieldValue
			reduce(100), // >>, reduce: FieldValue
			reduce(100), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(100), // ||=, reduce: FieldValue
			reduce(100), // ^^=, reduce: FieldValue
			reduce(100), // &&=, reduce: FieldValue
			reduce(100), // ??=, reduce: FieldValue
			reduce(100), // ???=, reduce: FieldValue
			reduce(100), // |=, reduce: FieldValue
			reduce(100), // &=, reduce: FieldValue
			reduce(100), // ^=, reduce: FieldValue
			reduce(100), // <<=, reduce: FieldValue
			reduce(100), // >>=, reduce: FieldValue
			reduce(100), // >>>=, reduce: FieldValue
			reduce(100), // +=, reduce: FieldValue
			reduce(100), // .=, reduce: FieldValue
			reduce(100), // -=, reduce: FieldValue
			reduce(100), // *=, reduce: FieldValue
			reduce(100), // /=, reduce: FieldValue
			reduce(100), // //=, reduce: FieldValue
			reduce(100), // %=, reduce: FieldValue
			reduce(100), // **=, reduce: FieldValue
			reduce(100), // ?, reduce: FieldValue
			nil,         // :
			reduce(100), // ||, reduce: FieldValue
			reduce(100), // ^^, reduce: FieldValue
			reduce(100), // &&, reduce: FieldValue
			reduce(100), // ??, reduce: FieldValue
			reduce(100), // ???, reduce: FieldValue
			reduce(100), // =~, reduce: FieldValue
			reduce(100), // !=~, reduce: FieldValue
			reduce(100), // ==, reduce: FieldValue
			reduce(100), // !=, reduce: FieldValue
			reduce(100), // <=>, reduce: FieldValue
			reduce(100), // >=, reduce: FieldValue
			reduce(100), // <, reduce: FieldValue
			reduce(100), // <=, reduce: FieldValue
			reduce(100), // ^, reduce: FieldValue
			reduce(100), // &, reduce: FieldValue
			reduce(100), // <<, reduce: FieldValue
			reduce(100), // >>>, reduce: FieldValue
			reduce(100), // +, reduce: FieldValue
			reduce(100), // -, reduce: FieldValue
			reduce(100), // .+, reduce: FieldValue
			reduce(100), // .-, reduce: FieldValue
			reduce(100), // *, reduce: FieldValue
			reduce(100), // /, reduce: FieldValue
			reduce(100), // //, reduce: FieldValue
			reduce(100), // %, reduce: FieldValue
			reduce(100), // .*, reduce: FieldValue
			reduce(100), // ./, reduce: FieldValue
			reduce(100), // .//, reduce: FieldValue
			reduce(100), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(100), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(100), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(100), // [[, reduce: FieldValue
			reduce(100), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S998
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(213), // ;, reduce: MlrvalOrFunction
			reduce(213), // {, reduce: MlrvalOrFunction
			reduce(213), // }, reduce: MlrvalOrFunction
			reduce(213), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(213), // >, reduce: MlrvalOrFunction
			reduce(213), // >>, reduce: MlrvalOrFunction
			reduce(213), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(213), // ||=, reduce: MlrvalOrFunction
			reduce(213), // ^^=, reduce: MlrvalOrFunction
			reduce(213), // &&=, reduce: MlrvalOrFunction
			reduce(213), // ??=, reduce: MlrvalOrFunction
			reduce(213), // ???=, reduce: MlrvalOrFunction
			reduce(213), // |=, reduce: MlrvalOrFunction
			reduce(213), // &=, reduce: MlrvalOrFunction
			reduce(213), // ^=, reduce: MlrvalOrFunction
			reduce(213), // <<=, reduce: MlrvalOrFunction
			reduce(213), // >>=, reduce: MlrvalOrFunction
			reduce(213), // >>>=, reduce: MlrvalOrFunction
			reduce(213), // +=, reduce: MlrvalOrFunction
			reduce(213), // .=, reduce: MlrvalOrFunction
			reduce(213), // -=, reduce: MlrvalOrFunction
			reduce(213), // *=, reduce: MlrvalOrFunction
			reduce(213), // /=, reduce: MlrvalOrFunction
			reduce(213), // //=, reduce: MlrvalOrFunction
			reduce(213), // %=, reduce: MlrvalOrFunction
			reduce(213), // **=, reduce: MlrvalOrFunction
			reduce(213), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(213), // ||, reduce: MlrvalOrFunction
			reduce(213), // ^^, reduce: MlrvalOrFunction
			reduce(213), // &&, reduce: MlrvalOrFunction
			reduce(213), // ??, reduce: MlrvalOrFunction
			reduce(213), // ???, reduce: MlrvalOrFunction
			reduce(213), // =~, reduce: MlrvalOrFunction
			reduce(213), // !=~, reduce: MlrvalOrFunction
			reduce(213), // ==, reduce: MlrvalOrFunction
			reduce(213), // !=, reduce: MlrvalOrFunction
			reduce(213), // <=>, reduce: MlrvalOrFunction
			reduce(213), // >=, reduce: MlrvalOrFunction
			reduce(213), // <, reduce: MlrvalOrFunction
			reduce(213), // <=, reduce: MlrvalOrFunction
			reduce(213), // ^, reduce: MlrvalOrFunction
			reduce(213), // &, reduce: MlrvalOrFunction
			reduce(213), // <<, reduce: MlrvalOrFunction
			reduce(213), // >>>, reduce: MlrvalOrFunction
			reduce(213), // +, reduce: MlrvalOrFunction
			reduce(213), // -, reduce: MlrvalOrFunction
			reduce(213), // .+, reduce: MlrvalOrFunction
			reduce(213), // .-, reduce: MlrvalOrFunction
			reduce(213), // *, reduce: MlrvalOrFunction
			reduce(213), // /, reduce: MlrvalOrFunction
			reduce(213), // //, reduce: MlrvalOrFunction
			reduce(213), // %, reduce: MlrvalOrFunction
			reduce(213), // .*, reduce: MlrvalOrFunction
			reduce(213), // ./, reduce: MlrvalOrFunction
			reduce(213), // .//, reduce: MlrvalOrFunction
			reduce(213), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(213), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(213), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(213), // [[, reduce: MlrvalOrFunction
			reduce(213), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S999
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(233), // ;, reduce: MlrvalOrFunction
			reduce(233), // {, reduce: MlrvalOrFunction
			reduce(233), // }, reduce: MlrvalOrFunction
			reduce(233), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(233), // >, reduce: MlrvalOrFunction
			reduce(233), // >>, reduce: MlrvalOrFunction
			reduce(233), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(233), // ||=, reduce: MlrvalOrFunction
			reduce(233), // ^^=, reduce: MlrvalOrFunction
			reduce(233), // &&=, reduce: MlrvalOrFunction
			reduce(233), // ??=, reduce: MlrvalOrFunction
			reduce(233), // ???=, reduce: MlrvalOrFunction
			reduce(233), // |=, reduce: MlrvalOrFunction
			reduce(233), // &=, reduce: MlrvalOrFunction
			reduce(233), // ^=, reduce: MlrvalOrFunction
			reduce(233), // <<=, reduce: MlrvalOrFunction
			reduce(233), // >>=, reduce: MlrvalOrFunction
			reduce(233), // >>>=, reduce: MlrvalOrFunction
			reduce(233), // +=, reduce: MlrvalOrFunction
			reduce(233), // .=, reduce: MlrvalOrFunction
			reduce(233), // -=, reduce: MlrvalOrFunction
			reduce(233), // *=, reduce: MlrvalOrFunction
			reduce(233), // /=, reduce: MlrvalOrFunction
			reduce(233), // //=, reduce: MlrvalOrFunction
			reduce(233), // %=, reduce: MlrvalOrFunction
			reduce(233), // **=, reduce: MlrvalOrFunction
			reduce(233), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(233), // ||, reduce: MlrvalOrFunction
			reduce(233), // ^^, reduce: MlrvalOrFunction
			reduce(233), // &&, reduce: MlrvalOrFunction
			reduce(233), // ??, reduce: MlrvalOrFunction
			reduce(233), // ???, reduce: MlrvalOrFunction
			reduce(233), // =~, reduce: MlrvalOrFunction
			reduce(233), // !=~, reduce: MlrvalOrFunction
			reduce(233), // ==, reduce: MlrvalOrFunction
			reduce(233), // !=, reduce: MlrvalOrFunction
			reduce(233), // <=>, reduce: MlrvalOrFunction
			reduce(233), // >=, reduce: MlrvalOrFunction
			reduce(233), // <, reduce: MlrvalOrFunction
			reduce(233), // <=, reduce: MlrvalOrFunction
			reduce(233), // ^, reduce: MlrvalOrFunction
			reduce(233), // &, reduce: MlrvalOrFunction
			reduce(233), // <<, reduce: MlrvalOrFunction
			reduce(233), // >>>, reduce: MlrvalOrFunction
			reduce(233), // +, reduce: MlrvalOrFunction
			reduce(233), // -, reduce: MlrvalOrFunction
			reduce(233), // .+, reduce: MlrvalOrFunction
			reduce(233), // .-, reduce: MlrvalOrFunction
			reduce(233), // *, reduce: MlrvalOrFunction
			reduce(233), // /, reduce: MlrvalOrFunction
			reduce(233), // //, reduce: MlrvalOrFunction
			reduce(233), // %, reduce: MlrvalOrFunction
			reduce(233), // .*, reduce: MlrvalOrFunction
			reduce(233), // ./, reduce: MlrvalOrFunction
			reduce(233), // .//, reduce: MlrvalOrFunction
			reduce(233), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(233), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(233), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(233), // [[, reduce: MlrvalOrFunction
			reduce(233), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1000
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(210), // ;, reduce: MlrvalOrFunction
			reduce(210), // {, reduce: MlrvalOrFunction
			reduce(210), // }, reduce: MlrvalOrFunction
			reduce(210), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(210), // >, reduce: MlrvalOrFunction
			reduce(210), // >>, reduce: MlrvalOrFunction
			reduce(210), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(210), // ||=, reduce: MlrvalOrFunction
			reduce(210), // ^^=, reduce: MlrvalOrFunction
			reduce(210), // &&=, reduce: MlrvalOrFunction
			reduce(210), // ??=, reduce: MlrvalOrFunction
			reduce(210), // ???=, reduce: MlrvalOrFunction
			reduce(210), // |=, reduce: MlrvalOrFunction
			reduce(210), // &=, reduce: MlrvalOrFunction
			reduce(210), // ^=, reduce: MlrvalOrFunction
			reduce(210), // <<=, reduce: MlrvalOrFunction
			reduce(210), // >>=, reduce: MlrvalOrFunction
			reduce(210), // >>>=, reduce: MlrvalOrFunction
			reduce(210), // +=, reduce: MlrvalOrFunction
			reduce(210), // .=, reduce: MlrvalOrFunction
			reduce(210), // -=, reduce: MlrvalOrFunction
			reduce(210), // *=, reduce: MlrvalOrFunction
			reduce(210), // /=, reduce: MlrvalOrFunction
			reduce(210), // //=, reduce: MlrvalOrFunction
			reduce(210), // %=, reduce: MlrvalOrFunction
			reduce(210), // **=, reduce: MlrvalOrFunction
			reduce(210), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(210), // ||, reduce: MlrvalOrFunction
			reduce(210), // ^^, reduce: MlrvalOrFunction
			reduce(210), // &&, reduce: MlrvalOrFunction
			reduce(210), // ??, reduce: MlrvalOrFunction
			reduce(210), // ???, reduce: MlrvalOrFunction
			reduce(210), // =~, reduce: MlrvalOrFunction
			reduce(210), // !=~, reduce: MlrvalOrFunction
			reduce(210), // ==, reduce: MlrvalOrFunction
			reduce(210), // !=, reduce: MlrvalOrFunction
			reduce(210), // <=>, reduce: MlrvalOrFunction
			reduce(210), // >=, reduce: MlrvalOrFunction
			reduce(210), // <, reduce: MlrvalOrFunction
			reduce(210), // <=, reduce: MlrvalOrFunction
			reduce(210), // ^, reduce: MlrvalOrFunction
			reduce(210), // &, reduce: MlrvalOrFunction
			reduce(210), // <<, reduce: MlrvalOrFunction
			reduce(210), // >>>, reduce: MlrvalOrFunction
			reduce(210), // +, reduce: MlrvalOrFunction
			reduce(210), // -, reduce: MlrvalOrFunction
			reduce(210), // .+, reduce: MlrvalOrFunction
			reduce(210), // .-, reduce: MlrvalOrFunction
			reduce(210), // *, reduce: MlrvalOrFunction
			reduce(210), // /, reduce: MlrvalOrFunction
			reduce(210), // //, reduce: MlrvalOrFunction
			reduce(210), // %, reduce: MlrvalOrFunction
			reduce(210), // .*, reduce: MlrvalOrFunction
			reduce(210), // ./, reduce: MlrvalOrFunction
			reduce(210), // .//, reduce: MlrvalOrFunction
			reduce(210), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(210), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(210), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(210), // [[, reduce: MlrvalOrFunction
			reduce(210), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1001
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(102), // ;, reduce: FieldValue
			reduce(102), // {, reduce: FieldValue
			reduce(102), // }, reduce: FieldValue
			reduce(102), // =, reduce: FieldValue
			nil,         // unset
			nil,         // filter
			reduce(102), // >, reduce: FieldValue
			reduce(102), // >>, reduce: FieldValue
			reduce(102), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(102), // ||=, reduce: FieldValue
			reduce(102), // ^^=, reduce: FieldValue
			reduce(102), // &&=, reduce: FieldValue
			reduce(102), // ??=, reduce: FieldValue
			reduce(102), // ???=, reduce: FieldValue
			reduce(102), // |=, reduce: FieldValue
			reduce(102), // &=, reduce: FieldValue
			reduce(102), // ^=, reduce: FieldValue
			reduce(102), // <<=, reduce: FieldValue
			reduce(102), // >>=, reduce: FieldValue
			reduce(102), // >>>=, reduce: FieldValue
			reduce(102), // +=, reduce: FieldValue
			reduce(102), // .=, reduce: FieldValue
			reduce(102), // -=, reduce: FieldValue
			reduce(102), // *=, reduce: FieldValue
			reduce(102), // /=, reduce: FieldValue
			reduce(102), // //=, reduce: FieldValue
			reduce(102), // %=, reduce: FieldValue
			reduce(102), // **=, reduce: FieldValue
			reduce(102), // ?, reduce: FieldValue
			nil,         // :
			reduce(102), // ||, reduce: FieldValue
			reduce(102), // ^^, reduce: FieldValue
			reduce(102), // &&, reduce: FieldValue
			reduce(102), // ??, reduce: FieldValue
			reduce(102), // ???, reduce: FieldValue
			reduce(102), // =~, reduce: FieldValue
			reduce(102), // !=~, reduce: FieldValue
			reduce(102), // ==, reduce: FieldValue
			reduce(102), // !=, reduce: FieldValue
			reduce(102), // <=>, reduce: FieldValue
			reduce(102), // >=, reduce: FieldValue
			reduce(102), // <, reduce: FieldValue
			reduce(102), // <=, reduce: FieldValue
			reduce(102), // ^, reduce: FieldValue
			reduce(102), // &, reduce: FieldValue
			reduce(102), // <<, reduce: FieldValue
			reduce(102), // >>>, reduce: FieldValue
			reduce(102), // +, reduce: FieldValue
			reduce(102), // -, reduce: FieldValue
			reduce(102), // .+, reduce: FieldValue
			reduce(102), // .-, reduce: FieldValue
			reduce(102), // *, reduce: FieldValue
			reduce(102), // /, reduce: FieldValue
			reduce(102), // //, reduce: FieldValue
			reduce(102), // %, reduce: FieldValue
			reduce(102), // .*, reduce: FieldValue
			reduce(102), // ./, reduce: FieldValue
			reduce(102), // .//, reduce: FieldValue
			reduce(102), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(102), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(102), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(102), // [[, reduce: FieldValue
			reduce(102), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1002
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(103), // ;, reduce: FieldValue
			reduce(103), // {, reduce: FieldValue
			reduce(103), // }, reduce: FieldValue
			reduce(103), // =, reduce: FieldValue
			nil,         // unset
			nil,         // filter
			reduce(103), // >, reduce: FieldValue
			reduce(103), // >>, reduce: FieldValue
			reduce(103), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(103), // ||=, reduce: FieldValue
			reduce(103), // ^^=, reduce: FieldValue
			reduce(103), // &&=, reduce: FieldValue
			reduce(103), // ??=, reduce: FieldValue
			reduce(103), // ???=, reduce: FieldValue
			reduce(103), // |=, reduce: FieldValue
			reduce(103), // &=, reduce: FieldValue
			reduce(103), // ^=, reduce: FieldValue
			reduce(103), // <<=, reduce: FieldValue
			reduce(103), // >>=, reduce: FieldValue
			reduce(103), // >>>=, reduce: FieldValue
			reduce(103), // +=, reduce: FieldValue
			reduce(103), // .=, reduce: FieldValue
			reduce(103), // -=, reduce: FieldValue
			reduce(103), // *=, reduce: FieldValue
			reduce(103), // /=, reduce: FieldValue
			reduce(103), // //=, reduce: FieldValue
			reduce(103), // %=, reduce: FieldValue
			reduce(103), // **=, reduce: FieldValue
			reduce(103), // ?, reduce: FieldValue
			nil,         // :
			reduce(103), // ||, reduce: FieldValue
			reduce(103), // ^^, reduce: FieldValue
			reduce(103), // &&, reduce: FieldValue
			reduce(103), // ??, reduce: FieldValue
			reduce(103), // ???, reduce: FieldValue
			reduce(103), // =~, reduce: FieldValue
			reduce(103), // !=~, reduce: FieldValue
			reduce(103), // ==, reduce: FieldValue
			reduce(103), // !=, reduce: FieldValue
			reduce(103), // <=>, reduce: FieldValue
			reduce(103), // >=, reduce: FieldValue
			reduce(103), // <, reduce: FieldValue
			reduce(103), // <=, reduce: FieldValue
			reduce(103), // ^, reduce: FieldValue
			reduce(103), // &, reduce: FieldValue
			reduce(103), // <<, reduce: FieldValue
			reduce(103), // >>>, reduce: FieldValue
			reduce(103), // +, reduce: FieldValue
			reduce(103), // -, reduce: FieldValue
			reduce(103), // .+, reduce: FieldValue
			reduce(103), // .-, reduce: FieldValue
			reduce(103), // *, reduce: FieldValue
			reduce(103), // /, reduce: FieldValue
			reduce(103), // //, reduce: FieldValue
			reduce(103), // %, reduce: FieldValue
			reduce(103), // .*, reduce: FieldValue
			reduce(103), // ./, reduce: FieldValue
			reduce(103), // .//, reduce: FieldValue
			reduce(103), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(103), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(103), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(103), // [[, reduce: FieldValue
			reduce(103), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1003
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(104), // ;, reduce: DirectFieldValue
			reduce(104), // {, reduce: DirectFieldValue
			reduce(104), // }, reduce: DirectFieldValue
			reduce(104), // =, reduce: DirectFieldValue
			nil,         // unset
			nil,         // filter
			reduce(104), // >, reduce: DirectFieldValue
			reduce(104), // >>, reduce: DirectFieldValue
			reduce(104), // |, reduce: DirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(104), // ||=, reduce: DirectFieldValue
			reduce(104), // ^^=, reduce: DirectFieldValue
			reduce(104), // &&=, reduce: DirectFieldValue
			reduce(104), // ??=, reduce: DirectFieldValue
			reduce(104), // ???=, reduce: DirectFieldValue
			reduce(104), // |=, reduce: DirectFieldValue
			reduce(104), // &=, reduce: DirectFieldValue
			reduce(104), // ^=, reduce: DirectFieldValue
			reduce(104), // <<=, reduce: DirectFieldValue
			reduce(104), // >>=, reduce: DirectFieldValue
			reduce(104), // >>>=, reduce: DirectFieldValue
			reduce(104), // +=, reduce: DirectFieldValue
			reduce(104), // .=, reduce: DirectFieldValue
			reduce(104), // -=, reduce: DirectFieldValue
			reduce(104), // *=, reduce: DirectFieldValue
			reduce(104), // /=, reduce: DirectFieldValue
			reduce(104), // //=, reduce: DirectFieldValue
			reduce(104), // %=, reduce: DirectFieldValue
			reduce(104), // **=, reduce: DirectFieldValue
			reduce(104), // ?, reduce: DirectFieldValue
			nil,         // :
			reduce(104), // ||, reduce: DirectFieldValue
			reduce(104), // ^^, reduce: DirectFieldValue
			reduce(104), // &&, reduce: DirectFieldValue
			reduce(104), // ??, reduce: DirectFieldValue
			reduce(104), // ???, reduce: DirectFieldValue
			reduce(104), // =~, reduce: DirectFieldValue
			reduce(104), // !=~, reduce: DirectFieldValue
			reduce(104), // ==, reduce: DirectFieldValue
			reduce(104), // !=, reduce: DirectFieldValue
			reduce(104), // <=>, reduce: DirectFieldValue
			reduce(104), // >=, reduce: DirectFieldValue
			reduce(104), // <, reduce: DirectFieldValue
			reduce(104), // <=, reduce: DirectFieldValue
			reduce(104), // ^, reduce: DirectFieldValue
			reduce(104), // &, reduce: DirectFieldValue
			reduce(104), // <<, reduce: DirectFieldValue
			reduce(104), // >>>, reduce: DirectFieldValue
			reduce(104), // +, reduce: DirectFieldValue
			reduce(104), // -, reduce: DirectFieldValue
			reduce(104), // .+, reduce: DirectFieldValue
			reduce(104), // .-, reduce: DirectFieldValue
			reduce(104), // *, reduce: DirectFieldValue
			reduce(104), // /, reduce: DirectFieldValue
			reduce(104), // //, reduce: DirectFieldValue
			reduce(104), // %, reduce: DirectFieldValue
			reduce(104), // .*, reduce: DirectFieldValue
			reduce(104), // ./, reduce: DirectFieldValue
			reduce(104), // .//, reduce: DirectFieldValue
			reduce(104), // ., reduce: DirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(104), // **, reduce: DirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(104), // [, reduce: DirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(104), // [[, reduce: DirectFieldValue
			reduce(104), // [[[, reduce: DirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1004
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1005
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(106), // ;, reduce: BracedFieldValue
			reduce(106), // {, reduce: BracedFieldValue
			reduce(106), // }, reduce: BracedFieldValue
			reduce(106), // =, reduce: BracedFieldValue
			nil,         // unset
			nil,         // filter
			reduce(106), // >, reduce: BracedFieldValue
			reduce(106), // >>, reduce: BracedFieldValue
			reduce(106), // |, reduce: BracedFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(106), // ||=, reduce: BracedFieldValue
			reduce(106), // ^^=, reduce: BracedFieldValue
			reduce(106), // &&=, reduce: BracedFieldValue
			reduce(106), // ??=, reduce: BracedFieldValue
			reduce(106), // ???=, reduce: BracedFieldValue
			reduce(106), // |=, reduce: BracedFieldValue
			reduce(106), // &=, reduce: BracedFieldValue
			reduce(106), // ^=, reduce: BracedFieldValue
			reduce(106), // <<=, reduce: BracedFieldValue
			reduce(106), // >>=, reduce: BracedFieldValue
			reduce(106), // >>>=, reduce: BracedFieldValue
			reduce(106), // +=, reduce: BracedFieldValue
			reduce(106), // .=, reduce: BracedFieldValue
			reduce(106), // -=, reduce: BracedFieldValue
			reduce(106), // *=, reduce: BracedFieldValue
			reduce(106), // /=, reduce: BracedFieldValue
			reduce(106), // //=, reduce: BracedFieldValue
			reduce(106), // %=, reduce: BracedFieldValue
			reduce(106), // **=, reduce: BracedFieldValue
			reduce(106), // ?, reduce: BracedFieldValue
			nil,         // :
			reduce(106), // ||, reduce: BracedFieldValue
			reduce(106), // ^^, reduce: BracedFieldValue
			reduce(106), // &&, reduce: BracedFieldValue
			reduce(106), // ??, reduce: BracedFieldValue
			reduce(106), // ???, reduce: BracedFieldValue
			reduce(106), // =~, reduce: BracedFieldValue
			reduce(106), // !=~, reduce: BracedFieldValue
			reduce(106), // ==, reduce: BracedFieldValue
			reduce(106), // !=, reduce: BracedFieldValue
			reduce(106), // <=>, reduce: BracedFieldValue
			reduce(106), // >=, reduce: BracedFieldValue
			reduce(106), // <, reduce: BracedFieldValue
			reduce(106), // <=, reduce: BracedFieldValue
			reduce(106), // ^, reduce: BracedFieldValue
			reduce(106), // &, reduce: BracedFieldValue
			reduce(106), // <<, reduce: BracedFieldValue
			reduce(106), // >>>, reduce: BracedFieldValue
			reduce(106), // +, reduce: BracedFieldValue
			reduce(106), // -, reduce: BracedFieldValue
			reduce(106), // .+, reduce: BracedFieldValue
			reduce(106), // .-, reduce: BracedFieldValue
			reduce(106), // *, reduce: BracedFieldValue
			reduce(106), // /, reduce: BracedFieldValue
			reduce(106), // //, reduce: BracedFieldValue
			reduce(106), // %, reduce: BracedFieldValue
			reduce(106), // .*, reduce: BracedFieldValue
			reduce(106), // ./, reduce: BracedFieldValue
			reduce(106), // .//, reduce: BracedFieldValue
			reduce(106), // ., reduce: BracedFieldValue
			nil,         // !
			nil,         // ~
			reduce(106), // **, reduce: BracedFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(106), // [, reduce: BracedFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(106), // [[, reduce: BracedFieldValue
			reduce(106), // [[[, reduce: BracedFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1006
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1007
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1008
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(109), // ;, reduce: FullSrec
			reduce(109), // {, reduce: FullSrec
			reduce(109), // }, reduce: FullSrec
			reduce(109), // =, reduce: FullSrec
			nil,         // unset
			nil,         // filter
			reduce(109), // >, reduce: FullSrec
			reduce(109), // >>, reduce: FullSrec
			reduce(109), // |, reduce: FullSrec
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(109), // ||=, reduce: FullSrec
			reduce(109), // ^^=, reduce: FullSrec
			reduce(109), // &&=, reduce: FullSrec
			reduce(109), // ??=, reduce: FullSrec
			reduce(109), // ???=, reduce: FullSrec
			reduce(109), // |=, reduce: FullSrec
			reduce(109), // &=, reduce: FullSrec
			reduce(109), // ^=, reduce: FullSrec
			reduce(109), // <<=, reduce: FullSrec
			reduce(109), // >>=, reduce: FullSrec
			reduce(109), // >>>=, reduce: FullSrec
			reduce(109), // +=, reduce: FullSrec
			reduce(109), // .=, reduce: FullSrec
			reduce(109), // -=, reduce: FullSrec
			reduce(109), // *=, reduce: FullSrec
			reduce(109), // /=, reduce: FullSrec
			reduce(109), // //=, reduce: FullSrec
			reduce(109), // %=, reduce: FullSrec
			reduce(109), // **=, reduce: FullSrec
			reduce(109), // ?, reduce: FullSrec
			nil,         // :
			reduce(109), // ||, reduce: FullSrec
			reduce(109), // ^^, reduce: FullSrec
			reduce(109), // &&, reduce: FullSrec
			reduce(109), // ??, reduce: FullSrec
			reduce(109), // ???, reduce: FullSrec
			reduce(109), // =~, reduce: FullSrec
			reduce(109), // !=~, reduce: FullSrec
			reduce(109), // ==, reduce: FullSrec
			reduce(109), // !=, reduce: FullSrec
			reduce(109), // <=>, reduce: FullSrec
			reduce(109), // >=, reduce: FullSrec
			reduce(109), // <, reduce: FullSrec
			reduce(109), // <=, reduce: FullSrec
			reduce(109), // ^, reduce: FullSrec
			reduce(109), // &, reduce: FullSrec
			reduce(109), // <<, reduce: FullSrec
			reduce(109), // >>>, reduce: FullSrec
			reduce(109), // +, reduce: FullSrec
			reduce(109), // -, reduce: FullSrec
			reduce(109), // .+, reduce: FullSrec
			reduce(109), // .-, reduce: FullSrec
			reduce(109), // *, reduce: FullSrec
			reduce(109), // /, reduce: FullSrec
			reduce(109), // //, reduce: FullSrec
			reduce(109), // %, reduce: FullSrec
			reduce(109), // .*, reduce: FullSrec
			reduce(109), // ./, reduce: FullSrec
			reduce(109), // .//, reduce: FullSrec
			reduce(109), // ., reduce: FullSrec
			nil,         // !
			nil,         // ~
			reduce(109), // **, reduce: FullSrec
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(109), // [, reduce: FullSrec
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(109), // [[, reduce: FullSrec
			reduce(109), // [[[, reduce: FullSrec
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1009
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(212), // ;, reduce: MlrvalOrFunction
			reduce(212), // {, reduce: MlrvalOrFunction
			reduce(212), // }, reduce: MlrvalOrFunction
			reduce(212), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(212), // >, reduce: MlrvalOrFunction
			reduce(212), // >>, reduce: MlrvalOrFunction
			reduce(212), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(212), // ||=, reduce: MlrvalOrFunction
			reduce(212), // ^^=, reduce: MlrvalOrFunction
			reduce(212), // &&=, reduce: MlrvalOrFunction
			reduce(212), // ??=, reduce: MlrvalOrFunction
			reduce(212), // ???=, reduce: MlrvalOrFunction
			reduce(212), // |=, reduce: MlrvalOrFunction
			reduce(212), // &=, reduce: MlrvalOrFunction
			reduce(212), // ^=, reduce: MlrvalOrFunction
			reduce(212), // <<=, reduce: MlrvalOrFunction
			reduce(212), // >>=, reduce: MlrvalOrFunction
			reduce(212), // >>>=, reduce: MlrvalOrFunction
			reduce(212), // +=, reduce: MlrvalOrFunction
			reduce(212), // .=, reduce: MlrvalOrFunction
			reduce(212), // -=, reduce: MlrvalOrFunction
			reduce(212), // *=, reduce: MlrvalOrFunction
			reduce(212), // /=, reduce: MlrvalOrFunction
			reduce(212), // //=, reduce: MlrvalOrFunction
			reduce(212), // %=, reduce: MlrvalOrFunction
			reduce(212), // **=, reduce: MlrvalOrFunction
			reduce(212), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(212), // ||, reduce: MlrvalOrFunction
			reduce(212), // ^^, reduce: MlrvalOrFunction
			reduce(212), // &&, reduce: MlrvalOrFunction
			reduce(212), // ??, reduce: MlrvalOrFunction
			reduce(212), // ???, reduce: MlrvalOrFunction
			reduce(212), // =~, reduce: MlrvalOrFunction
			reduce(212), // !=~, reduce: MlrvalOrFunction
			reduce(212), // ==, reduce: MlrvalOrFunction
			reduce(212), // !=, reduce: MlrvalOrFunction
			reduce(212), // <=>, reduce: MlrvalOrFunction
			reduce(212), // >=, reduce: MlrvalOrFunction
			reduce(212), // <, reduce: MlrvalOrFunction
			reduce(212), // <=, reduce: MlrvalOrFunction
			reduce(212), // ^, reduce: MlrvalOrFunction
			reduce(212), // &, reduce: MlrvalOrFunction
			reduce(212), // <<, reduce: MlrvalOrFunction
			reduce(212), // >>>, reduce: MlrvalOrFunction
			reduce(212), // +, reduce: MlrvalOrFunction
			reduce(212), // -, reduce: MlrvalOrFunction
			reduce(212), // .+, reduce: MlrvalOrFunction
			reduce(212), // .-, reduce: MlrvalOrFunction
			reduce(212), // *, reduce: MlrvalOrFunction
			reduce(212), // /, reduce: MlrvalOrFunction
			reduce(212), // //, reduce: MlrvalOrFunction
			reduce(212), // %, reduce: MlrvalOrFunction
			reduce(212), // .*, reduce: MlrvalOrFunction
			reduce(212), // ./, reduce: MlrvalOrFunction
			reduce(212), // .//, reduce: MlrvalOrFunction
			reduce(212), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(212), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(212), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(212), // [[, reduce: MlrvalOrFunction
			reduce(212), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1010
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(113), // ;, reduce: DirectOosvarValue
			reduce(113), // {, reduce: DirectOosvarValue
			reduce(113), // }, reduce: DirectOosvarValue
			reduce(113), // =, reduce: DirectOosvarValue
			nil,         // unset
			nil,         // filter
			reduce(113), // >, reduce: DirectOosvarValue
			reduce(113), // >>, reduce: DirectOosvarValue
			reduce(113), // |, reduce: DirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(113), // ||=, reduce: DirectOosvarValue
			reduce(113), // ^^=, reduce: DirectOosvarValue
			reduce(113), // &&=, reduce: DirectOosvarValue
			reduce(113), // ??=, reduce: DirectOosvarValue
			reduce(113), // ???=, reduce: DirectOosvarValue
			reduce(113), // |=, reduce: DirectOosvarValue
			reduce(113), // &=, reduce: DirectOosvarValue
			reduce(113), // ^=, reduce: DirectOosvarValue
			reduce(113), // <<=, reduce: DirectOosvarValue
			reduce(113), // >>=, reduce: DirectOosvarValue
			reduce(113), // >>>=, reduce: DirectOosvarValue
			reduce(113), // +=, reduce: DirectOosvarValue
			reduce(113), // .=, reduce: DirectOosvarValue
			reduce(113), // -=, reduce: DirectOosvarValue
			reduce(113), // *=, reduce: DirectOosvarValue
			reduce(113), // /=, reduce: DirectOosvarValue
			reduce(113), // //=, reduce: DirectOosvarValue
			reduce(113), // %=, reduce: DirectOosvarValue
			reduce(113), // **=, reduce: DirectOosvarValue
			reduce(113), // ?, reduce: DirectOosvarValue
			nil,         // :
			reduce(113), // ||, reduce: DirectOosvarValue
			reduce(113), // ^^, reduce: DirectOosvarValue
			reduce(113), // &&, reduce: DirectOosvarValue
			reduce(113), // ??, reduce: DirectOosvarValue
			reduce(113), // ???, reduce: DirectOosvarValue
			reduce(113), // =~, reduce: DirectOosvarValue
			reduce(113), // !=~, reduce: DirectOosvarValue
			reduce(113), // ==, reduce: DirectOosvarValue
			reduce(113), // !=, reduce: DirectOosvarValue
			reduce(113), // <=>, reduce: DirectOosvarValue
			reduce(113), // >=, reduce: DirectOosvarValue
			reduce(113), // <, reduce: DirectOosvarValue
			reduce(113), // <=, reduce: DirectOosvarValue
			reduce(113), // ^, reduce: DirectOosvarValue
			reduce(113), // &, reduce: DirectOosvarValue
			reduce(113), // <<, reduce: DirectOosvarValue
			reduce(113), // >>>, reduce: DirectOosvarValue
			reduce(113), // +, reduce: DirectOosvarValue
			reduce(113), // -, reduce: DirectOosvarValue
			reduce(113), // .+, reduce: DirectOosvarValue
			reduce(113), // .-, reduce: DirectOosvarValue
			reduce(113), // *, reduce: DirectOosvarValue
			reduce(113), // /, reduce: DirectOosvarValue
			reduce(113), // //, reduce: DirectOosvarValue
			reduce(113), // %, reduce: DirectOosvarValue
			reduce(113), // .*, reduce: DirectOosvarValue
			reduce(113), // ./, reduce: DirectOosvarValue
			reduce(113), // .//, reduce: DirectOosvarValue
			reduce(113), // ., reduce: DirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(113), // **, reduce: DirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(113), // [, reduce: DirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(113), // [[, reduce: DirectOosvarValue
			reduce(113), // [[[, reduce: DirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1011
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1012
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(115), // ;, reduce: BracedOosvarValue
			reduce(115), // {, reduce: BracedOosvarValue
			reduce(115), // }, reduce: BracedOosvarValue
			reduce(115), // =, reduce: BracedOosvarValue
			nil,         // unset
			nil,         // filter
			reduce(115), // >, reduce: BracedOosvarValue
			reduce(115), // >>, reduce: BracedOosvarValue
			reduce(115), // |, reduce: BracedOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(115), // ||=, reduce: BracedOosvarValue
			reduce(115), // ^^=, reduce: BracedOosvarValue
			reduce(115), // &&=, reduce: BracedOosvarValue
			reduce(115), // ??=, reduce: BracedOosvarValue
			reduce(115), // ???=, reduce: BracedOosvarValue
			reduce(115), // |=, reduce: BracedOosvarValue
			reduce(115), // &=, reduce: BracedOosvarValue
			reduce(115), // ^=, reduce: BracedOosvarValue
			reduce(115), // <<=, reduce: BracedOosvarValue
			reduce(115), // >>=, reduce: BracedOosvarValue
			reduce(115), // >>>=, reduce: BracedOosvarValue
			reduce(115), // +=, reduce: BracedOosvarValue
			reduce(115), // .=, reduce: BracedOosvarValue
			reduce(115), // -=, reduce: BracedOosvarValue
			reduce(115), // *=, reduce: BracedOosvarValue
			reduce(115), // /=, reduce: BracedOosvarValue
			reduce(115), // //=, reduce: BracedOosvarValue
			reduce(115), // %=, reduce: BracedOosvarValue
			reduce(115), // **=, reduce: BracedOosvarValue
			reduce(115), // ?, reduce: BracedOosvarValue
			nil,         // :
			reduce(115), // ||, reduce: BracedOosvarValue
			reduce(115), // ^^, reduce: BracedOosvarValue
			reduce(115), // &&, reduce: BracedOosvarValue
			reduce(115), // ??, reduce: BracedOosvarValue
			reduce(115), // ???, reduce: BracedOosvarValue
			reduce(115), // =~, reduce: BracedOosvarValue
			reduce(115), // !=~, reduce: BracedOosvarValue
			reduce(115), // ==, reduce: BracedOosvarValue
			reduce(115), // !=, reduce: BracedOosvarValue
			reduce(115), // <=>, reduce: BracedOosvarValue
			reduce(115), // >=, reduce: BracedOosvarValue
			reduce(115), // <, reduce: BracedOosvarValue
			reduce(115), // <=, reduce: BracedOosvarValue
			reduce(115), // ^, reduce: BracedOosvarValue
			reduce(115), // &, reduce: BracedOosvarValue
			reduce(115), // <<, reduce: BracedOosvarValue
			reduce(115), // >>>, reduce: BracedOosvarValue
			reduce(115), // +, reduce: BracedOosvarValue
			reduce(115), // -, reduce: BracedOosvarValue
			reduce(115), // .+, reduce: BracedOosvarValue
			reduce(115), // .-, reduce: BracedOosvarValue
			reduce(115), // *, reduce: BracedOosvarValue
			reduce(115), // /, reduce: BracedOosvarValue
			reduce(115), // //, reduce: BracedOosvarValue
			reduce(115), // %, reduce: BracedOosvarValue
			reduce(115), // .*, reduce: BracedOosvarValue
			reduce(115), // ./, reduce: BracedOosvarValue
			reduce(115), // .//, reduce: BracedOosvarValue
			reduce(115), // ., reduce: BracedOosvarValue
			nil,         // !
			nil,         // ~
			reduce(115), // **, reduce: BracedOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(115), // [, reduce: BracedOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(115), // [[, reduce: BracedOosvarValue
			reduce(115), // [[[, reduce: BracedOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1013
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(116), // ;, reduce: FullOosvar
			reduce(116), // {, reduce: FullOosvar
			reduce(116), // }, reduce: FullOosvar
			reduce(116), // =, reduce: FullOosvar
			nil,         // unset
			nil,         // filter
			reduce(116), // >, reduce: FullOosvar
			reduce(116), // >>, reduce: FullOosvar
			reduce(116), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(116), // ||=, reduce: FullOosvar
			reduce(116), // ^^=, reduce: FullOosvar
			reduce(116), // &&=, reduce: FullOosvar
			reduce(116), // ??=, reduce: FullOosvar
			reduce(116), // ???=, reduce: FullOosvar
			reduce(116), // |=, reduce: FullOosvar
			reduce(116), // &=, reduce: FullOosvar
			reduce(116), // ^=, reduce: FullOosvar
			reduce(116), // <<=, reduce: FullOosvar
			reduce(116), // >>=, reduce: FullOosvar
			reduce(116), // >>>=, reduce: FullOosvar
			reduce(116), // +=, reduce: FullOosvar
			reduce(116), // .=, reduce: FullOosvar
			reduce(116), // -=, reduce: FullOosvar
			reduce(116), // *=, reduce: FullOosvar
			reduce(116), // /=, reduce: FullOosvar
			reduce(116), // //=, reduce: FullOosvar
			reduce(116), // %=, reduce: FullOosvar
			reduce(116), // **=, reduce: FullOosvar
			reduce(116), // ?, reduce: FullOosvar
			nil,         // :
			reduce(116), // ||, reduce: FullOosvar
			reduce(116), // ^^, reduce: FullOosvar
			reduce(116), // &&, reduce: FullOosvar
			reduce(116), // ??, reduce: FullOosvar
			reduce(116), // ???, reduce: FullOosvar
			reduce(116), // =~, reduce: FullOosvar
			reduce(116), // !=~, reduce: FullOosvar
			reduce(116), // ==, reduce: FullOosvar
			reduce(116), // !=, reduce: FullOosvar
			reduce(116), // <=>, reduce: FullOosvar
			reduce(116), // >=, reduce: FullOosvar
			reduce(116), // <, reduce: FullOosvar
			reduce(116), // <=, reduce: FullOosvar
			reduce(116), // ^, reduce: FullOosvar
			reduce(116), // &, reduce: FullOosvar
			reduce(116), // <<, reduce: FullOosvar
			reduce(116), // >>>, reduce: FullOosvar
			reduce(116), // +, reduce: FullOosvar
			reduce(116), // -, reduce: FullOosvar
			reduce(116), // .+, reduce: FullOosvar
			reduce(116), // .-, reduce: FullOosvar
			reduce(116), // *, reduce: FullOosvar
			reduce(116), // /, reduce: FullOosvar
			reduce(116), // //, reduce: FullOosvar
			reduce(116), // %, reduce: FullOosvar
			reduce(116), // .*, reduce: FullOosvar
			reduce(116), // ./, reduce: FullOosvar
			reduce(116), // .//, reduce: FullOosvar
			reduce(116), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(116), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(116), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(116), // [[, reduce: FullOosvar
			reduce(116), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1014
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(117), // ;, reduce: FullOosvar
			reduce(117), // {, reduce: FullOosvar
			reduce(117), // }, reduce: FullOosvar
			reduce(117), // =, reduce: FullOosvar
			nil,         // unset
			nil,         // filter
			reduce(117), // >, reduce: FullOosvar
			reduce(117), // >>, reduce: FullOosvar
			reduce(117), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(117), // ||=, reduce: FullOosvar
			reduce(117), // ^^=, reduce: FullOosvar
			reduce(117), // &&=, reduce: FullOosvar
			reduce(117), // ??=, reduce: FullOosvar
			reduce(117), // ???=, reduce: FullOosvar
			reduce(117), // |=, reduce: FullOosvar
			reduce(117), // &=, reduce: FullOosvar
			reduce(117), // ^=, reduce: FullOosvar
			reduce(117), // <<=, reduce: FullOosvar
			reduce(117), // >>=, reduce: FullOosvar
			reduce(117), // >>>=, reduce: FullOosvar
			reduce(117), // +=, reduce: FullOosvar
			reduce(117), // .=, reduce: FullOosvar
			reduce(117), // -=, reduce: FullOosvar
			reduce(117), // *=, reduce: FullOosvar
			reduce(117), // /=, reduce: FullOosvar
			reduce(117), // //=, reduce: FullOosvar
			reduce(117), // %=, reduce: FullOosvar
			reduce(117), // **=, reduce: FullOosvar
			reduce(117), // ?, reduce: FullOosvar
			nil,         // :
			reduce(117), // ||, reduce: FullOosvar
			reduce(117), // ^^, reduce: FullOosvar
			reduce(117), // &&, reduce: FullOosvar
			reduce(117), // ??, reduce: FullOosvar
			reduce(117), // ???, reduce: FullOosvar
			reduce(117), // =~, reduce: FullOosvar
			reduce(117), // !=~, reduce: FullOosvar
			reduce(117), // ==, reduce: FullOosvar
			reduce(117), // !=, reduce: FullOosvar
			reduce(117), // <=>, reduce: FullOosvar
			reduce(117), // >=, reduce: FullOosvar
			reduce(117), // <, reduce: FullOosvar
			reduce(117), // <=, reduce: FullOosvar
			reduce(117), // ^, reduce: FullOosvar
			reduce(117), // &, reduce: FullOosvar
			reduce(117), // <<, reduce: FullOosvar
			reduce(117), // >>>, reduce: FullOosvar
			reduce(117), // +, reduce: FullOosvar
			reduce(117), // -, reduce: FullOosvar
			reduce(117), // .+, reduce: FullOosvar
			reduce(117), // .-, reduce: FullOosvar
			reduce(117), // *, reduce: FullOosvar
			reduce(117), // /, reduce: FullOosvar
			reduce(117), // //, reduce: FullOosvar
			reduce(117), // %, reduce: FullOosvar
			reduce(117), // .*, reduce: FullOosvar
			reduce(117), // ./, reduce: FullOosvar
			reduce(117), // .//, reduce: FullOosvar
			reduce(117), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(117), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(117), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(117), // [[, reduce: FullOosvar
			reduce(117), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1015
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(118), // ;, reduce: LocalVariable
			reduce(118), // {, reduce: LocalVariable
			reduce(118), // }, reduce: LocalVariable
			reduce(118), // =, reduce: LocalVariable
			nil,         // unset
			nil,         // filter
			reduce(118), // >, reduce: LocalVariable
			reduce(118), // >>, reduce: LocalVariable
			reduce(118), // |, reduce: LocalVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(270), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(118), // ||=, reduce: LocalVariable
			reduce(118), // ^^=, reduce: LocalVariable
			reduce(118), // &&=, reduce: LocalVariable
			reduce(118), // ??=, reduce: LocalVariable
			reduce(118), // ???=, reduce: LocalVariable
			reduce(118), // |=, reduce: LocalVariable
			reduce(118), // &=, reduce: LocalVariable
			reduce(118), // ^=, reduce: LocalVariable
			reduce(118), // <<=, reduce: LocalVariable
			reduce(118), // >>=, reduce: LocalVariable
			reduce(118), // >>>=, reduce: LocalVariable
			reduce(118), // +=, reduce: LocalVariable
			reduce(118), // .=, reduce: LocalVariable
			reduce(118), // -=, reduce: LocalVariable
			reduce(118), // *=, reduce: LocalVariable
			reduce(118), // /=, reduce: LocalVariable
			reduce(118), // //=, reduce: LocalVariable
			reduce(118), // %=, reduce: LocalVariable
			reduce(118), // **=, reduce: LocalVariable
			reduce(118), // ?, reduce: LocalVariable
			nil,         // :
			reduce(118), // ||, reduce: LocalVariable
			reduce(118), // ^^, reduce: LocalVariable
			reduce(118), // &&, reduce: LocalVariable
			reduce(118), // ??, reduce: LocalVariable
			reduce(118), // ???, reduce: LocalVariable
			reduce(118), // =~, reduce: LocalVariable
			reduce(118), // !=~, reduce: LocalVariable
			reduce(118), // ==, reduce: LocalVariable
			reduce(118), // !=, reduce: LocalVariable
			reduce(118), // <=>, reduce: LocalVariable
			reduce(118), // >=, reduce: LocalVariable
			reduce(118), // <, reduce: LocalVariable
			reduce(118), // <=, reduce: LocalVariable
			reduce(118), // ^, reduce: LocalVariable
			reduce(118), // &, reduce: LocalVariable
			reduce(118), // <<, reduce: LocalVariable
			reduce(118), // >>>, reduce: LocalVariable
			reduce(118), // +, reduce: LocalVariable
			reduce(118), // -, reduce: LocalVariable
			reduce(118), // .+, reduce: LocalVariable
			reduce(118), // .-, reduce: LocalVariable
			reduce(118), // *, reduce: LocalVariable
			reduce(118), // /, reduce: LocalVariable
			reduce(118), // //, reduce: LocalVariable
			reduce(118), // %, reduce: LocalVariable
			reduce(118), // .*, reduce: LocalVariable
			reduce(118), // ./, reduce: LocalVariable
			reduce(118), // .//, reduce: LocalVariable
			reduce(118), // ., reduce: LocalVariable
			nil,         // !
			nil,         // ~
			reduce(118), // **, reduce: LocalVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(118), // [, reduce: LocalVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(118), // [[, reduce: LocalVariable
			reduce(118), // [[[, reduce: LocalVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1016
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(147), // ;, reduce: Rvalue
			reduce(147), // {, reduce: Rvalue
			reduce(147), // }, reduce: Rvalue
			reduce(147), // =, reduce: Rvalue
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(147), // ||=, reduce: Rvalue
			reduce(147), // ^^=, reduce: Rvalue
			reduce(147), // &&=, reduce: Rvalue
			reduce(147), // ??=, reduce: Rvalue
			reduce(147), // ???=, reduce: Rvalue
			reduce(147), // |=, reduce: Rvalue
			reduce(147), // &=, reduce: Rvalue
			reduce(147), // ^=, reduce: Rvalue
			reduce(147), // <<=, reduce: Rvalue
			reduce(147), // >>=, reduce: Rvalue
			reduce(147), // >>>=, reduce: Rvalue
			reduce(147), // +=, reduce: Rvalue
			reduce(147), // .=, reduce: Rvalue
			reduce(147), // -=, reduce: Rvalue
			reduce(147), // *=, reduce: Rvalue
			reduce(147), // /=, reduce: Rvalue
			reduce(147), // //=, reduce: Rvalue
			reduce(147), // %=, reduce: Rvalue
			reduce(147), // **=, reduce: Rvalue
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1017
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(148), // ;, reduce: PrecedenceChainStart
			reduce(148), // {, reduce: PrecedenceChainStart
			reduce(148), // }, reduce: PrecedenceChainStart
			reduce(148), // =, reduce: PrecedenceChainStart
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(148), // ||=, reduce: PrecedenceChainStart
			reduce(148), // ^^=, reduce: PrecedenceChainStart
			reduce(148), // &&=, reduce: PrecedenceChainStart
			reduce(148), // ??=, reduce: PrecedenceChainStart
			reduce(148), // ???=, reduce: PrecedenceChainStart
			reduce(148), // |=, reduce: PrecedenceChainStart
			reduce(148), // &=, reduce: PrecedenceChainStart
			reduce(148), // ^=, reduce: PrecedenceChainStart
			reduce(148), // <<=, reduce: PrecedenceChainStart
			reduce(148), // >>=, reduce: PrecedenceChainStart
			reduce(148), // >>>=, reduce: PrecedenceChainStart
			reduce(148), // +=, reduce: PrecedenceChainStart
			reduce(148), // .=, reduce: PrecedenceChainStart
			reduce(148), // -=, reduce: PrecedenceChainStart
			reduce(148), // *=, reduce: PrecedenceChainStart
			reduce(148), // /=, reduce: PrecedenceChainStart
			reduce(148), // //=, reduce: PrecedenceChainStart
			reduce(148), // %=, reduce: PrecedenceChainStart
			reduce(148), // **=, reduce: PrecedenceChainStart
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1018
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(150), // ;, reduce: TernaryTerm
			reduce(150), // {, reduce: TernaryTerm
			reduce(150), // }, reduce: TernaryTerm
			reduce(150), // =, reduce: TernaryTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(150), // ||=, reduce: TernaryTerm
			reduce(150), // ^^=, reduce: TernaryTerm
			reduce(150), // &&=, reduce: TernaryTerm
			reduce(150), // ??=, reduce: TernaryTerm
			reduce(150), // ???=, reduce: TernaryTerm
			reduce(150), // |=, reduce: TernaryTerm
			reduce(150), // &=, reduce: TernaryTerm
			reduce(150), // ^=, reduce: TernaryTerm
			reduce(150), // <<=, reduce: TernaryTerm
			reduce(150), // >>=, reduce: TernaryTerm
			reduce(150), // >>>=, reduce: TernaryTerm
			reduce(150), // +=, reduce: TernaryTerm
			reduce(150), // .=, reduce: TernaryTerm
			reduce(150), // -=, reduce: TernaryTerm
			reduce(150), // *=, reduce: TernaryTerm
			reduce(150), // /=, reduce: TernaryTerm
			reduce(150), // //=, reduce: TernaryTerm
			reduce(150), // %=, reduce: TernaryTerm
			reduce(150), // **=, reduce: TernaryTerm
			shift(2276), // ?
			nil,         // :
			shift(2277), // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1019
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(152), // ;, reduce: LogicalOrTerm
			reduce(152), // {, reduce: LogicalOrTerm
			reduce(152), // }, reduce: LogicalOrTerm
			reduce(152), // =, reduce: LogicalOrTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(152), // ||=, reduce: LogicalOrTerm
			reduce(152), // ^^=, reduce: LogicalOrTerm
			reduce(152), // &&=, reduce: LogicalOrTerm
			reduce(152), // ??=, reduce: LogicalOrTerm
			reduce(152), // ???=, reduce: LogicalOrTerm
			reduce(152), // |=, reduce: LogicalOrTerm
			reduce(152), // &=, reduce: LogicalOrTerm
			reduce(152), // ^=, reduce: LogicalOrTerm
			reduce(152), // <<=, reduce: LogicalOrTerm
			reduce(152), // >>=, reduce: LogicalOrTerm
			reduce(152), // >>>=, reduce: LogicalOrTerm
			reduce(152), // +=, reduce: LogicalOrTerm
			reduce(152), // .=, reduce: LogicalOrTerm
			reduce(152), // -=, reduce: LogicalOrTerm
			reduce(152), // *=, reduce: LogicalOrTerm
			reduce(152), // /=, reduce: LogicalOrTerm
			reduce(152), // //=, reduce: LogicalOrTerm
			reduce(152), // %=, reduce: LogicalOrTerm
			reduce(152), // **=, reduce: LogicalOrTerm
			reduce(152), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(152), // ||, reduce: LogicalOrTerm
			shift(2278), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1020
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(154), // ;, reduce: LogicalXORTerm
			reduce(154), // {, reduce: LogicalXORTerm
			reduce(154), // }, reduce: LogicalXORTerm
			reduce(154), // =, reduce: LogicalXORTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(154), // ||=, reduce: LogicalXORTerm
			reduce(154), // ^^=, reduce: LogicalXORTerm
			reduce(154), // &&=, reduce: LogicalXORTerm
			reduce(154), // ??=, reduce: LogicalXORTerm
			reduce(154), // ???=, reduce: LogicalXORTerm
			reduce(154), // |=, reduce: LogicalXORTerm
			reduce(154), // &=, reduce: LogicalXORTerm
			reduce(154), // ^=, reduce: LogicalXORTerm
			reduce(154), // <<=, reduce: LogicalXORTerm
			reduce(154), // >>=, reduce: LogicalXORTerm
			reduce(154), // >>>=, reduce: LogicalXORTerm
			reduce(154), // +=, reduce: LogicalXORTerm
			reduce(154), // .=, reduce: LogicalXORTerm
			reduce(154), // -=, reduce: LogicalXORTerm
			reduce(154), // *=, reduce: LogicalXORTerm
			reduce(154), // /=, reduce: LogicalXORTerm
			reduce(154), // //=, reduce: LogicalXORTerm
			reduce(154), // %=, reduce: LogicalXORTerm
			reduce(154), // **=, reduce: LogicalXORTerm
			reduce(154), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(154), // ||, reduce: LogicalXORTerm
			reduce(154), // ^^, reduce: LogicalXORTerm
			shift(2279), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1021
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(156), // ;, reduce: LogicalAndTerm
			reduce(156), // {, reduce: LogicalAndTerm
			reduce(156), // }, reduce: LogicalAndTerm
			reduce(156), // =, reduce: LogicalAndTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(156), // ||=, reduce: LogicalAndTerm
			reduce(156), // ^^=, reduce: LogicalAndTerm
			reduce(156), // &&=, reduce: LogicalAndTerm
			reduce(156), // ??=, reduce: LogicalAndTerm
			reduce(156), // ???=, reduce: LogicalAndTerm
			reduce(156), // |=, reduce: LogicalAndTerm
			reduce(156), // &=, reduce: LogicalAndTerm
			reduce(156), // ^=, reduce: LogicalAndTerm
			reduce(156), // <<=, reduce: LogicalAndTerm
			reduce(156), // >>=, reduce: LogicalAndTerm
			reduce(156), // >>>=, reduce: LogicalAndTerm
			reduce(156), // +=, reduce: LogicalAndTerm
			reduce(156), // .=, reduce: LogicalAndTerm
			reduce(156), // -=, reduce: LogicalAndTerm
			reduce(156), // *=, reduce: LogicalAndTerm
			reduce(156), // /=, reduce: LogicalAndTerm
			reduce(156), // //=, reduce: LogicalAndTerm
			reduce(156), // %=, reduce: LogicalAndTerm
			reduce(156), // **=, reduce: LogicalAndTerm
			reduce(156), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(156), // ||, reduce: LogicalAndTerm
			reduce(156), // ^^, reduce: LogicalAndTerm
			reduce(156), // &&, reduce: LogicalAndTerm
			shift(2280), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1022
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(158), // ;, reduce: AbsentCoalesceTerm
			reduce(158), // {, reduce: AbsentCoalesceTerm
			reduce(158), // }, reduce: AbsentCoalesceTerm
			reduce(158), // =, reduce: AbsentCoalesceTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(158), // ||=, reduce: AbsentCoalesceTerm
			reduce(158), // ^^=, reduce: AbsentCoalesceTerm
			reduce(158), // &&=, reduce: AbsentCoalesceTerm
			reduce(158), // ??=, reduce: AbsentCoalesceTerm
			reduce(158), // ???=, reduce: AbsentCoalesceTerm
			reduce(158), // |=, reduce: AbsentCoalesceTerm
			reduce(158), // &=, reduce: AbsentCoalesceTerm
			reduce(158), // ^=, reduce: AbsentCoalesceTerm
			reduce(158), // <<=, reduce: AbsentCoalesceTerm
			reduce(158), // >>=, reduce: AbsentCoalesceTerm
			reduce(158), // >>>=, reduce: AbsentCoalesceTerm
			reduce(158), // +=, reduce: AbsentCoalesceTerm
			reduce(158), // .=, reduce: AbsentCoalesceTerm
			reduce(158), // -=, reduce: AbsentCoalesceTerm
			reduce(158), // *=, reduce: AbsentCoalesceTerm
			reduce(158), // /=, reduce: AbsentCoalesceTerm
			reduce(158), // //=, reduce: AbsentCoalesceTerm
			reduce(158), // %=, reduce: AbsentCoalesceTerm
			reduce(158), // **=, reduce: AbsentCoalesceTerm
			reduce(158), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(158), // ||, reduce: AbsentCoalesceTerm
			reduce(158), // ^^, reduce: AbsentCoalesceTerm
			reduce(158), // &&, reduce: AbsentCoalesceTerm
			reduce(158), // ??, reduce: AbsentCoalesceTerm
			shift(2281), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1023
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(160), // ;, reduce: EmptyCoalesceTerm
			reduce(160), // {, reduce: EmptyCoalesceTerm
			reduce(160), // }, reduce: EmptyCoalesceTerm
			reduce(160), // =, reduce: EmptyCoalesceTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(160), // ||=, reduce: EmptyCoalesceTerm
			reduce(160), // ^^=, reduce: EmptyCoalesceTerm
			reduce(160), // &&=, reduce: EmptyCoalesceTerm
			reduce(160), // ??=, reduce: EmptyCoalesceTerm
			reduce(160), // ???=, reduce: EmptyCoalesceTerm
			reduce(160), // |=, reduce: EmptyCoalesceTerm
			reduce(160), // &=, reduce: EmptyCoalesceTerm
			reduce(160), // ^=, reduce: EmptyCoalesceTerm
			reduce(160), // <<=, reduce: EmptyCoalesceTerm
			reduce(160), // >>=, reduce: EmptyCoalesceTerm
			reduce(160), // >>>=, reduce: EmptyCoalesceTerm
			reduce(160), // +=, reduce: EmptyCoalesceTerm
			reduce(160), // .=, reduce: EmptyCoalesceTerm
			reduce(160), // -=, reduce: EmptyCoalesceTerm
			reduce(160), // *=, reduce: EmptyCoalesceTerm
			reduce(160), // /=, reduce: EmptyCoalesceTerm
			reduce(160), // //=, reduce: EmptyCoalesceTerm
			reduce(160), // %=, reduce: EmptyCoalesceTerm
			reduce(160), // **=, reduce: EmptyCoalesceTerm
			reduce(160), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(160), // ||, reduce: EmptyCoalesceTerm
			reduce(160), // ^^, reduce: EmptyCoalesceTerm
			reduce(160), // &&, reduce: EmptyCoalesceTerm
			reduce(160), // ??, reduce: EmptyCoalesceTerm
			reduce(160), // ???, reduce: EmptyCoalesceTerm
			shift(2282), // =~
			shift(2283), // !=~
			shift(2284), // ==
			shift(2285), // !=
			shift(2286), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1024
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(166), // ;, reduce: EqneTerm
			reduce(166), // {, reduce: EqneTerm
			reduce(166), // }, reduce: EqneTerm
			reduce(166), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(2287), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(166), // ||=, reduce: EqneTerm
			reduce(166), // ^^=, reduce: EqneTerm
			reduce(166), // &&=, reduce: EqneTerm
			reduce(166), // ??=, reduce: EqneTerm
			reduce(166), // ???=, reduce: EqneTerm
			reduce(166), // |=, reduce: EqneTerm
			reduce(166), // &=, reduce: EqneTerm
			reduce(166), // ^=, reduce: EqneTerm
			reduce(166), // <<=, reduce: EqneTerm
			reduce(166), // >>=, reduce: EqneTerm
			reduce(166), // >>>=, reduce: EqneTerm
			reduce(166), // +=, reduce: EqneTerm
			reduce(166), // .=, reduce: EqneTerm
			reduce(166), // -=, reduce: EqneTerm
			reduce(166), // *=, reduce: EqneTerm
			reduce(166), // /=, reduce: EqneTerm
			reduce(166), // //=, reduce: EqneTerm
			reduce(166), // %=, reduce: EqneTerm
			reduce(166), // **=, reduce: EqneTerm
			reduce(166), // ?, reduce: EqneTerm
			nil,         // :
			reduce(166), // ||, reduce: EqneTerm
			reduce(166), // ^^, reduce: EqneTerm
			reduce(166), // &&, reduce: EqneTerm
			reduce(166), // ??, reduce: EqneTerm
			reduce(166), // ???, reduce: EqneTerm
			reduce(166), // =~, reduce: EqneTerm
			reduce(166), // !=~, reduce: EqneTerm
			reduce(166), // ==, reduce: EqneTerm
			reduce(166), // !=, reduce: EqneTerm
			reduce(166), // <=>, reduce: EqneTerm
			shift(2288), // >=
			shift(2289), // <
			shift(2290), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1025
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(171), // ;, reduce: CmpTerm
			reduce(171), // {, reduce: CmpTerm
			reduce(171), // }, reduce: CmpTerm
			reduce(171), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(171), // >, reduce: CmpTerm
			nil,         // >>
			shift(2291), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(171), // ||=, reduce: CmpTerm
			reduce(171), // ^^=, reduce: CmpTerm
			reduce(171), // &&=, reduce: CmpTerm
			reduce(171), // ??=, reduce: CmpTerm
			reduce(171), // ???=, reduce: CmpTerm
			reduce(171), // |=, reduce: CmpTerm
			reduce(171), // &=, reduce: CmpTerm
			reduce(171), // ^=, reduce: CmpTerm
			reduce(171), // <<=, reduce: CmpTerm
			reduce(171), // >>=, reduce: CmpTerm
			reduce(171), // >>>=, reduce: CmpTerm
			reduce(171), // +=, reduce: CmpTerm
			reduce(171), // .=, reduce: CmpTerm
			reduce(171), // -=, reduce: CmpTerm
			reduce(171), // *=, reduce: CmpTerm
			reduce(171), // /=, reduce: CmpTerm
			reduce(171), // //=, reduce: CmpTerm
			reduce(171), // %=, reduce: CmpTerm
			reduce(171), // **=, reduce: CmpTerm
			reduce(171), // ?, reduce: CmpTerm
			nil,         // :
			reduce(171), // ||, reduce: CmpTerm
			reduce(171), // ^^, reduce: CmpTerm
			reduce(171), // &&, reduce: CmpTerm
			reduce(171), // ??, reduce: CmpTerm
			reduce(171), // ???, reduce: CmpTerm
			reduce(171), // =~, reduce: CmpTerm
			reduce(171), // !=~, reduce: CmpTerm
			reduce(171), // ==, reduce: CmpTerm
			reduce(171), // !=, reduce: CmpTerm
			reduce(171), // <=>, reduce: CmpTerm
			reduce(171), // >=, reduce: CmpTerm
			reduce(171), // <, reduce: CmpTerm
			reduce(171), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1026
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(173), // ;, reduce: BitwiseORTerm
			reduce(173), // {, reduce: BitwiseORTerm
			reduce(173), // }, reduce: BitwiseORTerm
			reduce(173), // =, reduce: BitwiseORTerm
			nil,         // unset
			nil,         // filter
			reduce(173), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(173), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(173), // ||=, reduce: BitwiseORTerm
			reduce(173), // ^^=, reduce: BitwiseORTerm
			reduce(173), // &&=, reduce: BitwiseORTerm
			reduce(173), // ??=, reduce: BitwiseORTerm
			reduce(173), // ???=, reduce: BitwiseORTerm
			reduce(173), // |=, reduce: BitwiseORTerm
			reduce(173), // &=, reduce: BitwiseORTerm
			reduce(173), // ^=, reduce: BitwiseORTerm
			reduce(173), // <<=, reduce: BitwiseORTerm
			reduce(173), // >>=, reduce: BitwiseORTerm
			reduce(173), // >>>=, reduce: BitwiseORTerm
			reduce(173), // +=, reduce: BitwiseORTerm
			reduce(173), // .=, reduce: BitwiseORTerm
			reduce(173), // -=, reduce: BitwiseORTerm
			reduce(173), // *=, reduce: BitwiseORTerm
			reduce(173), // /=, reduce: BitwiseORTerm
			reduce(173), // //=, reduce: BitwiseORTerm
			reduce(173), // %=, reduce: BitwiseORTerm
			reduce(173), // **=, reduce: BitwiseORTerm
			reduce(173), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(173), // ||, reduce: BitwiseORTerm
			reduce(173), // ^^, reduce: BitwiseORTerm
			reduce(173), // &&, reduce: BitwiseORTerm
			reduce(173), // ??, reduce: BitwiseORTerm
			reduce(173), // ???, reduce: BitwiseORTerm
			reduce(173), // =~, reduce: BitwiseORTerm
			reduce(173), // !=~, reduce: BitwiseORTerm
			reduce(173), // ==, reduce: BitwiseORTerm
			reduce(173), // !=, reduce: BitwiseORTerm
			reduce(173), // <=>, reduce: BitwiseORTerm
			reduce(173), // >=, reduce: BitwiseORTerm
			reduce(173), // <, reduce: BitwiseORTerm
			reduce(173), // <=, reduce: BitwiseORTerm
			shift(2292), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1027
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(175), // ;, reduce: BitwiseXORTerm
			reduce(175), // {, reduce: BitwiseXORTerm
			reduce(175), // }, reduce: BitwiseXORTerm
			reduce(175), // =, reduce: BitwiseXORTerm
			nil,         // unset
			nil,         // filter
			reduce(175), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(175), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(175), // ||=, reduce: BitwiseXORTerm
			reduce(175), // ^^=, reduce: BitwiseXORTerm
			reduce(175), // &&=, reduce: BitwiseXORTerm
			reduce(175), // ??=, reduce: BitwiseXORTerm
			reduce(175), // ???=, reduce: BitwiseXORTerm
			reduce(175), // |=, reduce: BitwiseXORTerm
			reduce(175), // &=, reduce: BitwiseXORTerm
			reduce(175), // ^=, reduce: BitwiseXORTerm
			reduce(175), // <<=, reduce: BitwiseXORTerm
			reduce(175), // >>=, reduce: BitwiseXORTerm
			reduce(175), // >>>=, reduce: BitwiseXORTerm
			reduce(175), // +=, reduce: BitwiseXORTerm
			reduce(175), // .=, reduce: BitwiseXORTerm
			reduce(175), // -=, reduce: BitwiseXORTerm
			reduce(175), // *=, reduce: BitwiseXORTerm
			reduce(175), // /=, reduce: BitwiseXORTerm
			reduce(175), // //=, reduce: BitwiseXORTerm
			reduce(175), // %=, reduce: BitwiseXORTerm
			reduce(175), // **=, reduce: BitwiseXORTerm
			reduce(175), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(175), // ||, reduce: BitwiseXORTerm
			reduce(175), // ^^, reduce: BitwiseXORTerm
			reduce(175), // &&, reduce: BitwiseXORTerm
			reduce(175), // ??, reduce: BitwiseXORTerm
			reduce(175), // ???, reduce: BitwiseXORTerm
			reduce(175), // =~, reduce: BitwiseXORTerm
			reduce(175), // !=~, reduce: BitwiseXORTerm
			reduce(175), // ==, reduce: BitwiseXORTerm
			reduce(175), // !=, reduce: BitwiseXORTerm
			reduce(175), // <=>, reduce: BitwiseXORTerm
			reduce(175), // >=, reduce: BitwiseXORTerm
			reduce(175), // <, reduce: BitwiseXORTerm
			reduce(175), // <=, reduce: BitwiseXORTerm
			reduce(175), // ^, reduce: BitwiseXORTerm
			shift(2293), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1028
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(177), // ;, reduce: BitwiseANDTerm
			reduce(177), // {, reduce: BitwiseANDTerm
			reduce(177), // }, reduce: BitwiseANDTerm
			reduce(177), // =, reduce: BitwiseANDTerm
			nil,         // unset
			nil,         // filter
			reduce(177), // >, reduce: BitwiseANDTerm
			shift(2294), // >>
			reduce(177), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(177), // ||=, reduce: BitwiseANDTerm
			reduce(177), // ^^=, reduce: BitwiseANDTerm
			reduce(177), // &&=, reduce: BitwiseANDTerm
			reduce(177), // ??=, reduce: BitwiseANDTerm
			reduce(177), // ???=, reduce: BitwiseANDTerm
			reduce(177), // |=, reduce: BitwiseANDTerm
			reduce(177), // &=, reduce: BitwiseANDTerm
			reduce(177), // ^=, reduce: BitwiseANDTerm
			reduce(177), // <<=, reduce: BitwiseANDTerm
			reduce(177), // >>=, reduce: BitwiseANDTerm
			reduce(177), // >>>=, reduce: BitwiseANDTerm
			reduce(177), // +=, reduce: BitwiseANDTerm
			reduce(177), // .=, reduce: BitwiseANDTerm
			reduce(177), // -=, reduce: BitwiseANDTerm
			reduce(177), // *=, reduce: BitwiseANDTerm
			reduce(177), // /=, reduce: BitwiseANDTerm
			reduce(177), // //=, reduce: BitwiseANDTerm
			reduce(177), // %=, reduce: BitwiseANDTerm
			reduce(177), // **=, reduce: BitwiseANDTerm
			reduce(177), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(177), // ||, reduce: BitwiseANDTerm
			reduce(177), // ^^, reduce: BitwiseANDTerm
			reduce(177), // &&, reduce: BitwiseANDTerm
			reduce(177), // ??, reduce: BitwiseANDTerm
			reduce(177), // ???, reduce: BitwiseANDTerm
			reduce(177), // =~, reduce: BitwiseANDTerm
			reduce(177), // !=~, reduce: BitwiseANDTerm
			reduce(177), // ==, reduce: BitwiseANDTerm
			reduce(177), // !=, reduce: BitwiseANDTerm
			reduce(177), // <=>, reduce: BitwiseANDTerm
			reduce(177), // >=, reduce: BitwiseANDTerm
			reduce(177), // <, reduce: BitwiseANDTerm
			reduce(177), // <=, reduce: BitwiseANDTerm
			reduce(177), // ^, reduce: BitwiseANDTerm
			reduce(177), // &, reduce: BitwiseANDTerm
			shift(2295), // <<
			shift(2296), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1029
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(181), // ;, reduce: BitwiseShiftTerm
			reduce(181), // {, reduce: BitwiseShiftTerm
			reduce(181), // }, reduce: BitwiseShiftTerm
			reduce(181), // =, reduce: BitwiseShiftTerm
			nil,         // unset
			nil,         // filter
			reduce(181), // >, reduce: BitwiseShiftTerm
			reduce(181), // >>, reduce: BitwiseShiftTerm
			reduce(181), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(181), // ||=, reduce: BitwiseShiftTerm
			reduce(181), // ^^=, reduce: BitwiseShiftTerm
			reduce(181), // &&=, reduce: BitwiseShiftTerm
			reduce(181), // ??=, reduce: BitwiseShiftTerm
			reduce(181), // ???=, reduce: BitwiseShiftTerm
			reduce(181), // |=, reduce: BitwiseShiftTerm
			reduce(181), // &=, reduce: BitwiseShiftTerm
			reduce(181), // ^=, reduce: BitwiseShiftTerm
			reduce(181), // <<=, reduce: BitwiseShiftTerm
			reduce(181), // >>=, reduce: BitwiseShiftTerm
			reduce(181), // >>>=, reduce: BitwiseShiftTerm
			reduce(181), // +=, reduce: BitwiseShiftTerm
			reduce(181), // .=, reduce: BitwiseShiftTerm
			reduce(181), // -=, reduce: BitwiseShiftTerm
			reduce(181), // *=, reduce: BitwiseShiftTerm
			reduce(181), // /=, reduce: BitwiseShiftTerm
			reduce(181), // //=, reduce: BitwiseShiftTerm
			reduce(181), // %=, reduce: BitwiseShiftTerm
			reduce(181), // **=, reduce: BitwiseShiftTerm
			reduce(181), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(181), // ||, reduce: BitwiseShiftTerm
			reduce(181), // ^^, reduce: BitwiseShiftTerm
			reduce(181), // &&, reduce: BitwiseShiftTerm
			reduce(181), // ??, reduce: BitwiseShiftTerm
			reduce(181), // ???, reduce: BitwiseShiftTerm
			reduce(181), // =~, reduce: BitwiseShiftTerm
			reduce(181), // !=~, reduce: BitwiseShiftTerm
			reduce(181), // ==, reduce: BitwiseShiftTerm
			reduce(181), // !=, reduce: BitwiseShiftTerm
			reduce(181), // <=>, reduce: BitwiseShiftTerm
			reduce(181), // >=, reduce: BitwiseShiftTerm
			reduce(181), // <, reduce: BitwiseShiftTerm
			reduce(181), // <=, reduce: BitwiseShiftTerm
			reduce(181), // ^, reduce: BitwiseShiftTerm
			reduce(181), // &, reduce: BitwiseShiftTerm
			reduce(181), // <<, reduce: BitwiseShiftTerm
			reduce(181), // >>>, reduce: BitwiseShiftTerm
			shift(2297), // +
			shift(2298), // -
			shift(2299), // .+
			shift(2300), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1030
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1031
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(186), // ;, reduce: AddsubdotTerm
			reduce(186), // {, reduce: AddsubdotTerm
			reduce(186), // }, reduce: AddsubdotTerm
			reduce(186), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(186), // >, reduce: AddsubdotTerm
			reduce(186), // >>, reduce: AddsubdotTerm
			reduce(186), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(186), // ||=, reduce: AddsubdotTerm
			reduce(186), // ^^=, reduce: AddsubdotTerm
			reduce(186), // &&=, reduce: AddsubdotTerm
			reduce(186), // ??=, reduce: AddsubdotTerm
			reduce(186), // ???=, reduce: AddsubdotTerm
			reduce(186), // |=, reduce: AddsubdotTerm
			reduce(186), // &=, reduce: AddsubdotTerm
			reduce(186), // ^=, reduce: AddsubdotTerm
			reduce(186), // <<=, reduce: AddsubdotTerm
			reduce(186), // >>=, reduce: AddsubdotTerm
			reduce(186), // >>>=, reduce: AddsubdotTerm
			reduce(186), // +=, reduce: AddsubdotTerm
			reduce(186), // .=, reduce: AddsubdotTerm
			reduce(186), // -=, reduce: AddsubdotTerm
			reduce(186), // *=, reduce: AddsubdotTerm
			reduce(186), // /=, reduce: AddsubdotTerm
			reduce(186), // //=, reduce: AddsubdotTerm
			reduce(186), // %=, reduce: AddsubdotTerm
			reduce(186), // **=, reduce: AddsubdotTerm
			reduce(186), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(186), // ||, reduce: AddsubdotTerm
			reduce(186), // ^^, reduce: AddsubdotTerm
			reduce(186), // &&, reduce: AddsubdotTerm
			reduce(186), // ??, reduce: AddsubdotTerm
			reduce(186), // ???, reduce: AddsubdotTerm
			reduce(186), // =~, reduce: AddsubdotTerm
			reduce(186), // !=~, reduce: AddsubdotTerm
			reduce(186), // ==, reduce: AddsubdotTerm
			reduce(186), // !=, reduce: AddsubdotTerm
			reduce(186), // <=>, reduce: AddsubdotTerm
			reduce(186), // >=, reduce: AddsubdotTerm
			reduce(186), // <, reduce: AddsubdotTerm
			reduce(186), // <=, reduce: AddsubdotTerm
			reduce(186), // ^, reduce: AddsubdotTerm
			reduce(186), // &, reduce: AddsubdotTerm
			reduce(186), // <<, reduce: AddsubdotTerm
			reduce(186), // >>>, reduce: AddsubdotTerm
			reduce(186), // +, reduce: AddsubdotTerm
			reduce(186), // -, reduce: AddsubdotTerm
			reduce(186), // .+, reduce: AddsubdotTerm
			reduce(186), // .-, reduce: AddsubdotTerm
			shift(2303), // *
			shift(2304), // /
			shift(2305), // //
			shift(2306), // %
			shift(2307), // .*
			shift(2308), // ./
			shift(2309), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1032
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1033
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1034
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1035
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(194), // ;, reduce: MuldivTerm
			reduce(194), // {, reduce: MuldivTerm
			reduce(194), // }, reduce: MuldivTerm
			reduce(194), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(194), // >, reduce: MuldivTerm
			reduce(194), // >>, reduce: MuldivTerm
			reduce(194), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(194), // ||=, reduce: MuldivTerm
			reduce(194), // ^^=, reduce: MuldivTerm
			reduce(194), // &&=, reduce: MuldivTerm
			reduce(194), // ??=, reduce: MuldivTerm
			reduce(194), // ???=, reduce: MuldivTerm
			reduce(194), // |=, reduce: MuldivTerm
			reduce(194), // &=, reduce: MuldivTerm
			reduce(194), // ^=, reduce: MuldivTerm
			reduce(194), // <<=, reduce: MuldivTerm
			reduce(194), // >>=, reduce: MuldivTerm
			reduce(194), // >>>=, reduce: MuldivTerm
			reduce(194), // +=, reduce: MuldivTerm
			reduce(194), // .=, reduce: MuldivTerm
			reduce(194), // -=, reduce: MuldivTerm
			reduce(194), // *=, reduce: MuldivTerm
			reduce(194), // /=, reduce: MuldivTerm
			reduce(194), // //=, reduce: MuldivTerm
			reduce(194), // %=, reduce: MuldivTerm
			reduce(194), // **=, reduce: MuldivTerm
			reduce(194), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(194), // ||, reduce: MuldivTerm
			reduce(194), // ^^, reduce: MuldivTerm
			reduce(194), // &&, reduce: MuldivTerm
			reduce(194), // ??, reduce: MuldivTerm
			reduce(194), // ???, reduce: MuldivTerm
			reduce(194), // =~, reduce: MuldivTerm
			reduce(194), // !=~, reduce: MuldivTerm
			reduce(194), // ==, reduce: MuldivTerm
			reduce(194), // !=, reduce: MuldivTerm
			reduce(194), // <=>, reduce: MuldivTerm
			reduce(194), // >=, reduce: MuldivTerm
			reduce(194), // <, reduce: MuldivTerm
			reduce(194), // <=, reduce: MuldivTerm
			reduce(194), // ^, reduce: MuldivTerm
			reduce(194), // &, reduce: MuldivTerm
			reduce(194), // <<, reduce: MuldivTerm
			reduce(194), // >>>, reduce: MuldivTerm
			reduce(194), // +, reduce: MuldivTerm
			reduce(194), // -, reduce: MuldivTerm
			reduce(194), // .+, reduce: MuldivTerm
			reduce(194), // .-, reduce: MuldivTerm
			reduce(194), // *, reduce: MuldivTerm
			reduce(194), // /, reduce: MuldivTerm
			reduce(194), // //, reduce: MuldivTerm
			reduce(194), // %, reduce: MuldivTerm
			reduce(194), // .*, reduce: MuldivTerm
			reduce(194), // ./, reduce: MuldivTerm
			reduce(194), // .//, reduce: MuldivTerm
			shift(2313), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1036
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(196), // ;, reduce: DotTerm
			reduce(196), // {, reduce: DotTerm
			reduce(196), // }, reduce: DotTerm
			reduce(196), // =, reduce: DotTerm
			nil,         // unset
			nil,         // filter
			reduce(196), // >, reduce: DotTerm
			reduce(196), // >>, reduce: DotTerm
			reduce(196), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(196), // ||=, reduce: DotTerm
			reduce(196), // ^^=, reduce: DotTerm
			reduce(196), // &&=, reduce: DotTerm
			reduce(196), // ??=, reduce: DotTerm
			reduce(196), // ???=, reduce: DotTerm
			reduce(196), // |=, reduce: DotTerm
			reduce(196), // &=, reduce: DotTerm
			reduce(196), // ^=, reduce: DotTerm
			reduce(196), // <<=, reduce: DotTerm
			reduce(196), // >>=, reduce: DotTerm
			reduce(196), // >>>=, reduce: DotTerm
			reduce(196), // +=, reduce: DotTerm
			reduce(196), // .=, reduce: DotTerm
			reduce(196), // -=, reduce: DotTerm
			reduce(196), // *=, reduce: DotTerm
			reduce(196), // /=, reduce: DotTerm
			reduce(196), // //=, reduce: DotTerm
			reduce(196), // %=, reduce: DotTerm
			reduce(196), // **=, reduce: DotTerm
			reduce(196), // ?, reduce: DotTerm
			nil,         // :
			reduce(196), // ||, reduce: DotTerm
			reduce(196), // ^^, reduce: DotTerm
			reduce(196), // &&, reduce: DotTerm
			reduce(196), // ??, reduce: DotTerm
			reduce(196), // ???, reduce: DotTerm
			reduce(196), // =~, reduce: DotTerm
			reduce(196), // !=~, reduce: DotTerm
			reduce(196), // ==, reduce: DotTerm
			reduce(196), // !=, reduce: DotTerm
			reduce(196), // <=>, reduce: DotTerm
			reduce(196), // >=, reduce: DotTerm
			reduce(196), // <, reduce: DotTerm
			reduce(196), // <=, reduce: DotTerm
			reduce(196), // ^, reduce: DotTerm
			reduce(196), // &, reduce: DotTerm
			reduce(196), // <<, reduce: DotTerm
			reduce(196), // >>>, reduce: DotTerm
			reduce(196), // +, reduce: DotTerm
			reduce(196), // -, reduce: DotTerm
			reduce(196), // .+, reduce: DotTerm
			reduce(196), // .-, reduce: DotTerm
			reduce(196), // *, reduce: DotTerm
			reduce(196), // /, reduce: DotTerm
			reduce(196), // //, reduce: DotTerm
			reduce(196), // %, reduce: DotTerm
			reduce(196), // .*, reduce: DotTerm
			reduce(196), // ./, reduce: DotTerm
			reduce(196), // .//, reduce: DotTerm
			reduce(196), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1037
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1038
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1039
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(203), // ;, reduce: UnaryOpTerm
			reduce(203), // {, reduce: UnaryOpTerm
			reduce(203), // }, reduce: UnaryOpTerm
			reduce(203), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(203), // >, reduce: UnaryOpTerm
			reduce(203), // >>, reduce: UnaryOpTerm
			reduce(203), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(203), // ||=, reduce: UnaryOpTerm
			reduce(203), // ^^=, reduce: UnaryOpTerm
			reduce(203), // &&=, reduce: UnaryOpTerm
			reduce(203), // ??=, reduce: UnaryOpTerm
			reduce(203), // ???=, reduce: UnaryOpTerm
			reduce(203), // |=, reduce: UnaryOpTerm
			reduce(203), // &=, reduce: UnaryOpTerm
			reduce(203), // ^=, reduce: UnaryOpTerm
			reduce(203), // <<=, reduce: UnaryOpTerm
			reduce(203), // >>=, reduce: UnaryOpTerm
			reduce(203), // >>>=, reduce: UnaryOpTerm
			reduce(203), // +=, reduce: UnaryOpTerm
			reduce(203), // .=, reduce: UnaryOpTerm
			reduce(203), // -=, reduce: UnaryOpTerm
			reduce(203), // *=, reduce: UnaryOpTerm
			reduce(203), // /=, reduce: UnaryOpTerm
			reduce(203), // //=, reduce: UnaryOpTerm
			reduce(203), // %=, reduce: UnaryOpTerm
			reduce(203), // **=, reduce: UnaryOpTerm
			reduce(203), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(203), // ||, reduce: UnaryOpTerm
			reduce(203), // ^^, reduce: UnaryOpTerm
			reduce(203), // &&, reduce: UnaryOpTerm
			reduce(203), // ??, reduce: UnaryOpTerm
			reduce(203), // ???, reduce: UnaryOpTerm
			reduce(203), // =~, reduce: UnaryOpTerm
			reduce(203), // !=~, reduce: UnaryOpTerm
			reduce(203), // ==, reduce: UnaryOpTerm
			reduce(203), // !=, reduce: UnaryOpTerm
			reduce(203), // <=>, reduce: UnaryOpTerm
			reduce(203), // >=, reduce: UnaryOpTerm
			reduce(203), // <, reduce: UnaryOpTerm
			reduce(203), // <=, reduce: UnaryOpTerm
			reduce(203), // ^, reduce: UnaryOpTerm
			reduce(203), // &, reduce: UnaryOpTerm
			reduce(203), // <<, reduce: UnaryOpTerm
			reduce(203), // >>>, reduce: UnaryOpTerm
			reduce(203), // +, reduce: UnaryOpTerm
			reduce(203), // -, reduce: UnaryOpTerm
			reduce(203), // .+, reduce: UnaryOpTerm
			reduce(203), // .-, reduce: UnaryOpTerm
			reduce(203), // *, reduce: UnaryOpTerm
			reduce(203), // /, reduce: UnaryOpTerm
			reduce(203), // //, reduce: UnaryOpTerm
			reduce(203), // %, reduce: UnaryOpTerm
			reduce(203), // .*, reduce: UnaryOpTerm
			reduce(203), // ./, reduce: UnaryOpTerm
			reduce(203), // .//, reduce: UnaryOpTerm
			reduce(203), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1040
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(207), // ;, reduce: PowTerm
			reduce(207), // {, reduce: PowTerm
			reduce(207), // }, reduce: PowTerm
			reduce(207), // =, reduce: PowTerm
			nil,         // unset
			nil,         // filter
			reduce(207), // >, reduce: PowTerm
			reduce(207), // >>, reduce: PowTerm
			reduce(207), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(207), // ||=, reduce: PowTerm
			reduce(207), // ^^=, reduce: PowTerm
			reduce(207), // &&=, reduce: PowTerm
			reduce(207), // ??=, reduce: PowTerm
			reduce(207), // ???=, reduce: PowTerm
			reduce(207), // |=, reduce: PowTerm
			reduce(207), // &=, reduce: PowTerm
			reduce(207), // ^=, reduce: PowTerm
			reduce(207), // <<=, reduce: PowTerm
			reduce(207), // >>=, reduce: PowTerm
			reduce(207), // >>>=, reduce: PowTerm
			reduce(207), // +=, reduce: PowTerm
			reduce(207), // .=, reduce: PowTerm
			reduce(207), // -=, reduce: PowTerm
			reduce(207), // *=, reduce: PowTerm
			reduce(207), // /=, reduce: PowTerm
			reduce(207), // //=, reduce: PowTerm
			reduce(207), // %=, reduce: PowTerm
			reduce(207), // **=, reduce: PowTerm
			reduce(207), // ?, reduce: PowTerm
			nil,         // :
			reduce(207), // ||, reduce: PowTerm
			reduce(207), // ^^, reduce: PowTerm
			reduce(207), // &&, reduce: PowTerm
			reduce(207), // ??, reduce: PowTerm
			reduce(207), // ???, reduce: PowTerm
			reduce(207), // =~, reduce: PowTerm
			reduce(207), // !=~, reduce: PowTerm
			reduce(207), // ==, reduce: PowTerm
			reduce(207), // !=, reduce: PowTerm
			reduce(207), // <=>, reduce: PowTerm
			reduce(207), // >=, reduce: PowTerm
			reduce(207), // <, reduce: PowTerm
			reduce(207), // <=, reduce: PowTerm
			reduce(207), // ^, reduce: PowTerm
			reduce(207), // &, reduce: PowTerm
			reduce(207), // <<, reduce: PowTerm
			reduce(207), // >>>, reduce: PowTerm
			reduce(207), // +, reduce: PowTerm
			reduce(207), // -, reduce: PowTerm
			reduce(207), // .+, reduce: PowTerm
			reduce(207), // .-, reduce: PowTerm
			reduce(207), // *, reduce: PowTerm
			reduce(207), // /, reduce: PowTerm
			reduce(207), // //, reduce: PowTerm
			reduce(207), // %, reduce: PowTerm
			reduce(207), // .*, reduce: PowTerm
			reduce(207), // ./, reduce: PowTerm
			reduce(207), // .//, reduce: PowTerm
			reduce(207), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			shift(2316), // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1041
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(209), // ;, reduce: PrecedenceChainEnd
			reduce(209), // {, reduce: PrecedenceChainEnd
			reduce(209), // }, reduce: PrecedenceChainEnd
			reduce(209), // =, reduce: PrecedenceChainEnd
			nil,         // unset
			nil,         // filter
			reduce(209), // >, reduce: PrecedenceChainEnd
			reduce(209), // >>, reduce: PrecedenceChainEnd
			reduce(209), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(209), // ||=, reduce: PrecedenceChainEnd
			reduce(209), // ^^=, reduce: PrecedenceChainEnd
			reduce(209), // &&=, reduce: PrecedenceChainEnd
			reduce(209), // ??=, reduce: PrecedenceChainEnd
			reduce(209), // ???=, reduce: PrecedenceChainEnd
			reduce(209), // |=, reduce: PrecedenceChainEnd
			reduce(209), // &=, reduce: PrecedenceChainEnd
			reduce(209), // ^=, reduce: PrecedenceChainEnd
			reduce(209), // <<=, reduce: PrecedenceChainEnd
			reduce(209), // >>=, reduce: PrecedenceChainEnd
			reduce(209), // >>>=, reduce: PrecedenceChainEnd
			reduce(209), // +=, reduce: PrecedenceChainEnd
			reduce(209), // .=, reduce: PrecedenceChainEnd
			reduce(209), // -=, reduce: PrecedenceChainEnd
			reduce(209), // *=, reduce: PrecedenceChainEnd
			reduce(209), // /=, reduce: PrecedenceChainEnd
			reduce(209), // //=, reduce: PrecedenceChainEnd
			reduce(209), // %=, reduce: PrecedenceChainEnd
			reduce(209), // **=, reduce: PrecedenceChainEnd
			reduce(209), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(209), // ||, reduce: PrecedenceChainEnd
			reduce(209), // ^^, reduce: PrecedenceChainEnd
			reduce(209), // &&, reduce: PrecedenceChainEnd
			reduce(209), // ??, reduce: PrecedenceChainEnd
			reduce(209), // ???, reduce: PrecedenceChainEnd
			reduce(209), // =~, reduce: PrecedenceChainEnd
			reduce(209), // !=~, reduce: PrecedenceChainEnd
			reduce(209), // ==, reduce: PrecedenceChainEnd
			reduce(209), // !=, reduce: PrecedenceChainEnd
			reduce(209), // <=>, reduce: PrecedenceChainEnd
			reduce(209), // >=, reduce: PrecedenceChainEnd
			reduce(209), // <, reduce: PrecedenceChainEnd
			reduce(209), // <=, reduce: PrecedenceChainEnd
			reduce(209), // ^, reduce: PrecedenceChainEnd
			reduce(209), // &, reduce: PrecedenceChainEnd
			reduce(209), // <<, reduce: PrecedenceChainEnd
			reduce(209), // >>>, reduce: PrecedenceChainEnd
			reduce(209), // +, reduce: PrecedenceChainEnd
			reduce(209), // -, reduce: PrecedenceChainEnd
			reduce(209), // .+, reduce: PrecedenceChainEnd
			reduce(209), // .-, reduce: PrecedenceChainEnd
			reduce(209), // *, reduce: PrecedenceChainEnd
			reduce(209), // /, reduce: PrecedenceChainEnd
			reduce(209), // //, reduce: PrecedenceChainEnd
			reduce(209), // %, reduce: PrecedenceChainEnd
			reduce(209), // .*, reduce: PrecedenceChainEnd
			reduce(209), // ./, reduce: PrecedenceChainEnd
			reduce(209), // .//, reduce: PrecedenceChainEnd
			reduce(209), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(209), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(2317), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			shift(2318), // [[
			shift(2319), // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1042
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(215), // ;, reduce: MlrvalOrFunction
			reduce(215), // {, reduce: MlrvalOrFunction
			reduce(215), // }, reduce: MlrvalOrFunction
			reduce(215), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(215), // >, reduce: MlrvalOrFunction
			reduce(215), // >>, reduce: MlrvalOrFunction
			reduce(215), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(215), // ||=, reduce: MlrvalOrFunction
			reduce(215), // ^^=, reduce: MlrvalOrFunction
			reduce(215), // &&=, reduce: MlrvalOrFunction
			reduce(215), // ??=, reduce: MlrvalOrFunction
			reduce(215), // ???=, reduce: MlrvalOrFunction
			reduce(215), // |=, reduce: MlrvalOrFunction
			reduce(215), // &=, reduce: MlrvalOrFunction
			reduce(215), // ^=, reduce: MlrvalOrFunction
			reduce(215), // <<=, reduce: MlrvalOrFunction
			reduce(215), // >>=, reduce: MlrvalOrFunction
			reduce(215), // >>>=, reduce: MlrvalOrFunction
			reduce(215), // +=, reduce: MlrvalOrFunction
			reduce(215), // .=, reduce: MlrvalOrFunction
			reduce(215), // -=, reduce: MlrvalOrFunction
			reduce(215), // *=, reduce: MlrvalOrFunction
			reduce(215), // /=, reduce: MlrvalOrFunction
			reduce(215), // //=, reduce: MlrvalOrFunction
			reduce(215), // %=, reduce: MlrvalOrFunction
			reduce(215), // **=, reduce: MlrvalOrFunction
			reduce(215), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(215), // ||, reduce: MlrvalOrFunction
			reduce(215), // ^^, reduce: MlrvalOrFunction
			reduce(215), // &&, reduce: MlrvalOrFunction
			reduce(215), // ??, reduce: MlrvalOrFunction
			reduce(215), // ???, reduce: MlrvalOrFunction
			reduce(215), // =~, reduce: MlrvalOrFunction
			reduce(215), // !=~, reduce: MlrvalOrFunction
			reduce(215), // ==, reduce: MlrvalOrFunction
			reduce(215), // !=, reduce: MlrvalOrFunction
			reduce(215), // <=>, reduce: MlrvalOrFunction
			reduce(215), // >=, reduce: MlrvalOrFunction
			reduce(215), // <, reduce: MlrvalOrFunction
			reduce(215), // <=, reduce: MlrvalOrFunction
			reduce(215), // ^, reduce: MlrvalOrFunction
			reduce(215), // &, reduce: MlrvalOrFunction
			reduce(215), // <<, reduce: MlrvalOrFunction
			reduce(215), // >>>, reduce: MlrvalOrFunction
			reduce(215), // +, reduce: MlrvalOrFunction
			reduce(215), // -, reduce: MlrvalOrFunction
			reduce(215), // .+, reduce: MlrvalOrFunction
			reduce(215), // .-, reduce: MlrvalOrFunction
			reduce(215), // *, reduce: MlrvalOrFunction
			reduce(215), // /, reduce: MlrvalOrFunction
			reduce(215), // //, reduce: MlrvalOrFunction
			reduce(215), // %, reduce: MlrvalOrFunction
			reduce(215), // .*, reduce: MlrvalOrFunction
			reduce(215), // ./, reduce: MlrvalOrFunction
			reduce(215), // .//, reduce: MlrvalOrFunction
			reduce(215), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(215), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(215), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(215), // [[, reduce: MlrvalOrFunction
			reduce(215), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1043
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(216), // ;, reduce: MlrvalOrFunction
			reduce(216), // {, reduce: MlrvalOrFunction
			reduce(216), // }, reduce: MlrvalOrFunction
			reduce(216), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(216), // >, reduce: MlrvalOrFunction
			reduce(216), // >>, reduce: MlrvalOrFunction
			reduce(216), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(216), // ||=, reduce: MlrvalOrFunction
			reduce(216), // ^^=, reduce: MlrvalOrFunction
			reduce(216), // &&=, reduce: MlrvalOrFunction
			reduce(216), // ??=, reduce: MlrvalOrFunction
			reduce(216), // ???=, reduce: MlrvalOrFunction
			reduce(216), // |=, reduce: MlrvalOrFunction
			reduce(216), // &=, reduce: MlrvalOrFunction
			reduce(216), // ^=, reduce: MlrvalOrFunction
			reduce(216), // <<=, reduce: MlrvalOrFunction
			reduce(216), // >>=, reduce: MlrvalOrFunction
			reduce(216), // >>>=, reduce: MlrvalOrFunction
			reduce(216), // +=, reduce: MlrvalOrFunction
			reduce(216), // .=, reduce: MlrvalOrFunction
			reduce(216), // -=, reduce: MlrvalOrFunction
			reduce(216), // *=, reduce: MlrvalOrFunction
			reduce(216), // /=, reduce: MlrvalOrFunction
			reduce(216), // //=, reduce: MlrvalOrFunction
			reduce(216), // %=, reduce: MlrvalOrFunction
			reduce(216), // **=, reduce: MlrvalOrFunction
			reduce(216), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(216), // ||, reduce: MlrvalOrFunction
			reduce(216), // ^^, reduce: MlrvalOrFunction
			reduce(216), // &&, reduce: MlrvalOrFunction
			reduce(216), // ??, reduce: MlrvalOrFunction
			reduce(216), // ???, reduce: MlrvalOrFunction
			reduce(216), // =~, reduce: MlrvalOrFunction
			reduce(216), // !=~, reduce: MlrvalOrFunction
			reduce(216), // ==, reduce: MlrvalOrFunction
			reduce(216), // !=, reduce: MlrvalOrFunction
			reduce(216), // <=>, reduce: MlrvalOrFunction
			reduce(216), // >=, reduce: MlrvalOrFunction
			reduce(216), // <, reduce: MlrvalOrFunction
			reduce(216), // <=, reduce: MlrvalOrFunction
			reduce(216), // ^, reduce: MlrvalOrFunction
			reduce(216), // &, reduce: MlrvalOrFunction
			reduce(216), // <<, reduce: MlrvalOrFunction
			reduce(216), // >>>, reduce: MlrvalOrFunction
			reduce(216), // +, reduce: MlrvalOrFunction
			reduce(216), // -, reduce: MlrvalOrFunction
			reduce(216), // .+, reduce: MlrvalOrFunction
			reduce(216), // .-, reduce: MlrvalOrFunction
			reduce(216), // *, reduce: MlrvalOrFunction
			reduce(216), // /, reduce: MlrvalOrFunction
			reduce(216), // //, reduce: MlrvalOrFunction
			reduce(216), // %, reduce: MlrvalOrFunction
			reduce(216), // .*, reduce: MlrvalOrFunction
			reduce(216), // ./, reduce: MlrvalOrFunction
			reduce(216), // .//, reduce: MlrvalOrFunction
			reduce(216), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(216), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(216), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(216), // [[, reduce: MlrvalOrFunction
			reduce(216), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1044
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(217), // ;, reduce: MlrvalOrFunction
			reduce(217), // {, reduce: MlrvalOrFunction
			reduce(217), // }, reduce: MlrvalOrFunction
			reduce(217), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(217), // >, reduce: MlrvalOrFunction
			reduce(217), // >>, reduce: MlrvalOrFunction
			reduce(217), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(217), // ||=, reduce: MlrvalOrFunction
			reduce(217), // ^^=, reduce: MlrvalOrFunction
			reduce(217), // &&=, reduce: MlrvalOrFunction
			reduce(217), // ??=, reduce: MlrvalOrFunction
			reduce(217), // ???=, reduce: MlrvalOrFunction
			reduce(217), // |=, reduce: MlrvalOrFunction
			reduce(217), // &=, reduce: MlrvalOrFunction
			reduce(217), // ^=, reduce: MlrvalOrFunction
			reduce(217), // <<=, reduce: MlrvalOrFunction
			reduce(217), // >>=, reduce: MlrvalOrFunction
			reduce(217), // >>>=, reduce: MlrvalOrFunction
			reduce(217), // +=, reduce: MlrvalOrFunction
			reduce(217), // .=, reduce: MlrvalOrFunction
			reduce(217), // -=, reduce: MlrvalOrFunction
			reduce(217), // *=, reduce: MlrvalOrFunction
			reduce(217), // /=, reduce: MlrvalOrFunction
			reduce(217), // //=, reduce: MlrvalOrFunction
			reduce(217), // %=, reduce: MlrvalOrFunction
			reduce(217), // **=, reduce: MlrvalOrFunction
			reduce(217), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(217), // ||, reduce: MlrvalOrFunction
			reduce(217), // ^^, reduce: MlrvalOrFunction
			reduce(217), // &&, reduce: MlrvalOrFunction
			reduce(217), // ??, reduce: MlrvalOrFunction
			reduce(217), // ???, reduce: MlrvalOrFunction
			reduce(217), // =~, reduce: MlrvalOrFunction
			reduce(217), // !=~, reduce: MlrvalOrFunction
			reduce(217), // ==, reduce: MlrvalOrFunction
			reduce(217), // !=, reduce: MlrvalOrFunction
			reduce(217), // <=>, reduce: MlrvalOrFunction
			reduce(217), // >=, reduce: MlrvalOrFunction
			reduce(217), // <, reduce: MlrvalOrFunction
			reduce(217), // <=, reduce: MlrvalOrFunction
			reduce(217), // ^, reduce: MlrvalOrFunction
			reduce(217), // &, reduce: MlrvalOrFunction
			reduce(217), // <<, reduce: MlrvalOrFunction
			reduce(217), // >>>, reduce: MlrvalOrFunction
			reduce(217), // +, reduce: MlrvalOrFunction
			reduce(217), // -, reduce: MlrvalOrFunction
			reduce(217), // .+, reduce: MlrvalOrFunction
			reduce(217), // .-, reduce: MlrvalOrFunction
			reduce(217), // *, reduce: MlrvalOrFunction
			reduce(217), // /, reduce: MlrvalOrFunction
			reduce(217), // //, reduce: MlrvalOrFunction
			reduce(217), // %, reduce: MlrvalOrFunction
			reduce(217), // .*, reduce: MlrvalOrFunction
			reduce(217), // ./, reduce: MlrvalOrFunction
			reduce(217), // .//, reduce: MlrvalOrFunction
			reduce(217), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(217), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(217), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(217), // [[, reduce: MlrvalOrFunction
			reduce(217), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1045
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(218), // ;, reduce: MlrvalOrFunction
			reduce(218), // {, reduce: MlrvalOrFunction
			reduce(218), // }, reduce: MlrvalOrFunction
			reduce(218), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(218), // >, reduce: MlrvalOrFunction
			reduce(218), // >>, reduce: MlrvalOrFunction
			reduce(218), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(218), // ||=, reduce: MlrvalOrFunction
			reduce(218), // ^^=, reduce: MlrvalOrFunction
			reduce(218), // &&=, reduce: MlrvalOrFunction
			reduce(218), // ??=, reduce: MlrvalOrFunction
			reduce(218), // ???=, reduce: MlrvalOrFunction
			reduce(218), // |=, reduce: MlrvalOrFunction
			reduce(218), // &=, reduce: MlrvalOrFunction
			reduce(218), // ^=, reduce: MlrvalOrFunction
			reduce(218), // <<=, reduce: MlrvalOrFunction
			reduce(218), // >>=, reduce: MlrvalOrFunction
			reduce(218), // >>>=, reduce: MlrvalOrFunction
			reduce(218), // +=, reduce: MlrvalOrFunction
			reduce(218), // .=, reduce: MlrvalOrFunction
			reduce(218), // -=, reduce: MlrvalOrFunction
			reduce(218), // *=, reduce: MlrvalOrFunction
			reduce(218), // /=, reduce: MlrvalOrFunction
			reduce(218), // //=, reduce: MlrvalOrFunction
			reduce(218), // %=, reduce: MlrvalOrFunction
			reduce(218), // **=, reduce: MlrvalOrFunction
			reduce(218), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(218), // ||, reduce: MlrvalOrFunction
			reduce(218), // ^^, reduce: MlrvalOrFunction
			reduce(218), // &&, reduce: MlrvalOrFunction
			reduce(218), // ??, reduce: MlrvalOrFunction
			reduce(218), // ???, reduce: MlrvalOrFunction
			reduce(218), // =~, reduce: MlrvalOrFunction
			reduce(218), // !=~, reduce: MlrvalOrFunction
			reduce(218), // ==, reduce: MlrvalOrFunction
			reduce(218), // !=, reduce: MlrvalOrFunction
			reduce(218), // <=>, reduce: MlrvalOrFunction
			reduce(218), // >=, reduce: MlrvalOrFunction
			reduce(218), // <, reduce: MlrvalOrFunction
			reduce(218), // <=, reduce: MlrvalOrFunction
			reduce(218), // ^, reduce: MlrvalOrFunction
			reduce(218), // &, reduce: MlrvalOrFunction
			reduce(218), // <<, reduce: MlrvalOrFunction
			reduce(218), // >>>, reduce: MlrvalOrFunction
			reduce(218), // +, reduce: MlrvalOrFunction
			reduce(218), // -, reduce: MlrvalOrFunction
			reduce(218), // .+, reduce: MlrvalOrFunction
			reduce(218), // .-, reduce: MlrvalOrFunction
			reduce(218), // *, reduce: MlrvalOrFunction
			reduce(218), // /, reduce: MlrvalOrFunction
			reduce(218), // //, reduce: MlrvalOrFunction
			reduce(218), // %, reduce: MlrvalOrFunction
			reduce(218), // .*, reduce: MlrvalOrFunction
			reduce(218), // ./, reduce: MlrvalOrFunction
			reduce(218), // .//, reduce: MlrvalOrFunction
			reduce(218), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(218), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(218), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(218), // [[, reduce: MlrvalOrFunction
			reduce(218), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1046
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(219), // ;, reduce: MlrvalOrFunction
			reduce(219), // {, reduce: MlrvalOrFunction
			reduce(219), // }, reduce: MlrvalOrFunction
			reduce(219), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(219), // >, reduce: MlrvalOrFunction
			reduce(219), // >>, reduce: MlrvalOrFunction
			reduce(219), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(219), // ||=, reduce: MlrvalOrFunction
			reduce(219), // ^^=, reduce: MlrvalOrFunction
			reduce(219), // &&=, reduce: MlrvalOrFunction
			reduce(219), // ??=, reduce: MlrvalOrFunction
			reduce(219), // ???=, reduce: MlrvalOrFunction
			reduce(219), // |=, reduce: MlrvalOrFunction
			reduce(219), // &=, reduce: MlrvalOrFunction
			reduce(219), // ^=, reduce: MlrvalOrFunction
			reduce(219), // <<=, reduce: MlrvalOrFunction
			reduce(219), // >>=, reduce: MlrvalOrFunction
			reduce(219), // >>>=, reduce: MlrvalOrFunction
			reduce(219), // +=, reduce: MlrvalOrFunction
			reduce(219), // .=, reduce: MlrvalOrFunction
			reduce(219), // -=, reduce: MlrvalOrFunction
			reduce(219), // *=, reduce: MlrvalOrFunction
			reduce(219), // /=, reduce: MlrvalOrFunction
			reduce(219), // //=, reduce: MlrvalOrFunction
			reduce(219), // %=, reduce: MlrvalOrFunction
			reduce(219), // **=, reduce: MlrvalOrFunction
			reduce(219), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(219), // ||, reduce: MlrvalOrFunction
			reduce(219), // ^^, reduce: MlrvalOrFunction
			reduce(219), // &&, reduce: MlrvalOrFunction
			reduce(219), // ??, reduce: MlrvalOrFunction
			reduce(219), // ???, reduce: MlrvalOrFunction
			reduce(219), // =~, reduce: MlrvalOrFunction
			reduce(219), // !=~, reduce: MlrvalOrFunction
			reduce(219), // ==, reduce: MlrvalOrFunction
			reduce(219), // !=, reduce: MlrvalOrFunction
			reduce(219), // <=>, reduce: MlrvalOrFunction
			reduce(219), // >=, reduce: MlrvalOrFunction
			reduce(219), // <, reduce: MlrvalOrFunction
			reduce(219), // <=, reduce: MlrvalOrFunction
			reduce(219), // ^, reduce: MlrvalOrFunction
			reduce(219), // &, reduce: MlrvalOrFunction
			reduce(219), // <<, reduce: MlrvalOrFunction
			reduce(219), // >>>, reduce: MlrvalOrFunction
			reduce(219), // +, reduce: MlrvalOrFunction
			reduce(219), // -, reduce: MlrvalOrFunction
			reduce(219), // .+, reduce: MlrvalOrFunction
			reduce(219), // .-, reduce: MlrvalOrFunction
			reduce(219), // *, reduce: MlrvalOrFunction
			reduce(219), // /, reduce: MlrvalOrFunction
			reduce(219), // //, reduce: MlrvalOrFunction
			reduce(219), // %, reduce: MlrvalOrFunction
			reduce(219), // .*, reduce: MlrvalOrFunction
			reduce(219), // ./, reduce: MlrvalOrFunction
			reduce(219), // .//, reduce: MlrvalOrFunction
			reduce(219), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(219), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(219), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(219), // [[, reduce: MlrvalOrFunction
			reduce(219), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1047
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(220), // ;, reduce: MlrvalOrFunction
			reduce(220), // {, reduce: MlrvalOrFunction
			reduce(220), // }, reduce: MlrvalOrFunction
			reduce(220), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(220), // >, reduce: MlrvalOrFunction
			reduce(220), // >>, reduce: MlrvalOrFunction
			reduce(220), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(220), // ||=, reduce: MlrvalOrFunction
			reduce(220), // ^^=, reduce: MlrvalOrFunction
			reduce(220), // &&=, reduce: MlrvalOrFunction
			reduce(220), // ??=, reduce: MlrvalOrFunction
			reduce(220), // ???=, reduce: MlrvalOrFunction
			reduce(220), // |=, reduce: MlrvalOrFunction
			reduce(220), // &=, reduce: MlrvalOrFunction
			reduce(220), // ^=, reduce: MlrvalOrFunction
			reduce(220), // <<=, reduce: MlrvalOrFunction
			reduce(220), // >>=, reduce: MlrvalOrFunction
			reduce(220), // >>>=, reduce: MlrvalOrFunction
			reduce(220), // +=, reduce: MlrvalOrFunction
			reduce(220), // .=, reduce: MlrvalOrFunction
			reduce(220), // -=, reduce: MlrvalOrFunction
			reduce(220), // *=, reduce: MlrvalOrFunction
			reduce(220), // /=, reduce: MlrvalOrFunction
			reduce(220), // //=, reduce: MlrvalOrFunction
			reduce(220), // %=, reduce: MlrvalOrFunction
			reduce(220), // **=, reduce: MlrvalOrFunction
			reduce(220), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(220), // ||, reduce: MlrvalOrFunction
			reduce(220), // ^^, reduce: MlrvalOrFunction
			reduce(220), // &&, reduce: MlrvalOrFunction
			reduce(220), // ??, reduce: MlrvalOrFunction
			reduce(220), // ???, reduce: MlrvalOrFunction
			reduce(220), // =~, reduce: MlrvalOrFunction
			reduce(220), // !=~, reduce: MlrvalOrFunction
			reduce(220), // ==, reduce: MlrvalOrFunction
			reduce(220), // !=, reduce: MlrvalOrFunction
			reduce(220), // <=>, reduce: MlrvalOrFunction
			reduce(220), // >=, reduce: MlrvalOrFunction
			reduce(220), // <, reduce: MlrvalOrFunction
			reduce(220), // <=, reduce: MlrvalOrFunction
			reduce(220), // ^, reduce: MlrvalOrFunction
			reduce(220), // &, reduce: MlrvalOrFunction
			reduce(220), // <<, reduce: MlrvalOrFunction
			reduce(220), // >>>, reduce: MlrvalOrFunction
			reduce(220), // +, reduce: MlrvalOrFunction
			reduce(220), // -, reduce: MlrvalOrFunction
			reduce(220), // .+, reduce: MlrvalOrFunction
			reduce(220), // .-, reduce: MlrvalOrFunction
			reduce(220), // *, reduce: MlrvalOrFunction
			reduce(220), // /, reduce: MlrvalOrFunction
			reduce(220), // //, reduce: MlrvalOrFunction
			reduce(220), // %, reduce: MlrvalOrFunction
			reduce(220), // .*, reduce: MlrvalOrFunction
			reduce(220), // ./, reduce: MlrvalOrFunction
			reduce(220), // .//, reduce: MlrvalOrFunction
			reduce(220), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(220), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(220), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(220), // [[, reduce: MlrvalOrFunction
			reduce(220), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1048
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(221), // ;, reduce: MlrvalOrFunction
			reduce(221), // {, reduce: MlrvalOrFunction
			reduce(221), // }, reduce: MlrvalOrFunction
			reduce(221), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(221), // >, reduce: MlrvalOrFunction
			reduce(221), // >>, reduce: MlrvalOrFunction
			reduce(221), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(221), // ||=, reduce: MlrvalOrFunction
			reduce(221), // ^^=, reduce: MlrvalOrFunction
			reduce(221), // &&=, reduce: MlrvalOrFunction
			reduce(221), // ??=, reduce: MlrvalOrFunction
			reduce(221), // ???=, reduce: MlrvalOrFunction
			reduce(221), // |=, reduce: MlrvalOrFunction
			reduce(221), // &=, reduce: MlrvalOrFunction
			reduce(221), // ^=, reduce: MlrvalOrFunction
			reduce(221), // <<=, reduce: MlrvalOrFunction
			reduce(221), // >>=, reduce: MlrvalOrFunction
			reduce(221), // >>>=, reduce: MlrvalOrFunction
			reduce(221), // +=, reduce: MlrvalOrFunction
			reduce(221), // .=, reduce: MlrvalOrFunction
			reduce(221), // -=, reduce: MlrvalOrFunction
			reduce(221), // *=, reduce: MlrvalOrFunction
			reduce(221), // /=, reduce: MlrvalOrFunction
			reduce(221), // //=, reduce: MlrvalOrFunction
			reduce(221), // %=, reduce: MlrvalOrFunction
			reduce(221), // **=, reduce: MlrvalOrFunction
			reduce(221), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(221), // ||, reduce: MlrvalOrFunction
			reduce(221), // ^^, reduce: MlrvalOrFunction
			reduce(221), // &&, reduce: MlrvalOrFunction
			reduce(221), // ??, reduce: MlrvalOrFunction
			reduce(221), // ???, reduce: MlrvalOrFunction
			reduce(221), // =~, reduce: MlrvalOrFunction
			reduce(221), // !=~, reduce: MlrvalOrFunction
			reduce(221), // ==, reduce: MlrvalOrFunction
			reduce(221), // !=, reduce: MlrvalOrFunction
			reduce(221), // <=>, reduce: MlrvalOrFunction
			reduce(221), // >=, reduce: MlrvalOrFunction
			reduce(221), // <, reduce: MlrvalOrFunction
			reduce(221), // <=, reduce: MlrvalOrFunction
			reduce(221), // ^, reduce: MlrvalOrFunction
			reduce(221), // &, reduce: MlrvalOrFunction
			reduce(221), // <<, reduce: MlrvalOrFunction
			reduce(221), // >>>, reduce: MlrvalOrFunction
			reduce(221), // +, reduce: MlrvalOrFunction
			reduce(221), // -, reduce: MlrvalOrFunction
			reduce(221), // .+, reduce: MlrvalOrFunction
			reduce(221), // .-, reduce: MlrvalOrFunction
			reduce(221), // *, reduce: MlrvalOrFunction
			reduce(221), // /, reduce: MlrvalOrFunction
			reduce(221), // //, reduce: MlrvalOrFunction
			reduce(221), // %, reduce: MlrvalOrFunction
			reduce(221), // .*, reduce: MlrvalOrFunction
			reduce(221), // ./, reduce: MlrvalOrFunction
			reduce(221), // .//, reduce: MlrvalOrFunction
			reduce(221), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(221), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(221), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(221), // [[, reduce: MlrvalOrFunction
			reduce(221), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1049
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(222), // ;, reduce: MlrvalOrFunction
			reduce(222), // {, reduce: MlrvalOrFunction
			reduce(222), // }, reduce: MlrvalOrFunction
			reduce(222), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(222), // >, reduce: MlrvalOrFunction
			reduce(222), // >>, reduce: MlrvalOrFunction
			reduce(222), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(222), // ||=, reduce: MlrvalOrFunction
			reduce(222), // ^^=, reduce: MlrvalOrFunction
			reduce(222), // &&=, reduce: MlrvalOrFunction
			reduce(222), // ??=, reduce: MlrvalOrFunction
			reduce(222), // ???=, reduce: MlrvalOrFunction
			reduce(222), // |=, reduce: MlrvalOrFunction
			reduce(222), // &=, reduce: MlrvalOrFunction
			reduce(222), // ^=, reduce: MlrvalOrFunction
			reduce(222), // <<=, reduce: MlrvalOrFunction
			reduce(222), // >>=, reduce: MlrvalOrFunction
			reduce(222), // >>>=, reduce: MlrvalOrFunction
			reduce(222), // +=, reduce: MlrvalOrFunction
			reduce(222), // .=, reduce: MlrvalOrFunction
			reduce(222), // -=, reduce: MlrvalOrFunction
			reduce(222), // *=, reduce: MlrvalOrFunction
			reduce(222), // /=, reduce: MlrvalOrFunction
			reduce(222), // //=, reduce: MlrvalOrFunction
			reduce(222), // %=, reduce: MlrvalOrFunction
			reduce(222), // **=, reduce: MlrvalOrFunction
			reduce(222), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(222), // ||, reduce: MlrvalOrFunction
			reduce(222), // ^^, reduce: MlrvalOrFunction
			reduce(222), // &&, reduce: MlrvalOrFunction
			reduce(222), // ??, reduce: MlrvalOrFunction
			reduce(222), // ???, reduce: MlrvalOrFunction
			reduce(222), // =~, reduce: MlrvalOrFunction
			reduce(222), // !=~, reduce: MlrvalOrFunction
			reduce(222), // ==, reduce: MlrvalOrFunction
			reduce(222), // !=, reduce: MlrvalOrFunction
			reduce(222), // <=>, reduce: MlrvalOrFunction
			reduce(222), // >=, reduce: MlrvalOrFunction
			reduce(222), // <, reduce: MlrvalOrFunction
			reduce(222), // <=, reduce: MlrvalOrFunction
			reduce(222), // ^, reduce: MlrvalOrFunction
			reduce(222), // &, reduce: MlrvalOrFunction
			reduce(222), // <<, reduce: MlrvalOrFunction
			reduce(222), // >>>, reduce: MlrvalOrFunction
			reduce(222), // +, reduce: MlrvalOrFunction
			reduce(222), // -, reduce: MlrvalOrFunction
			reduce(222), // .+, reduce: MlrvalOrFunction
			reduce(222), // .-, reduce: MlrvalOrFunction
			reduce(222), // *, reduce: MlrvalOrFunction
			reduce(222), // /, reduce: MlrvalOrFunction
			reduce(222), // //, reduce: MlrvalOrFunction
			reduce(222), // %, reduce: MlrvalOrFunction
			reduce(222), // .*, reduce: MlrvalOrFunction
			reduce(222), // ./, reduce: MlrvalOrFunction
			reduce(222), // .//, reduce: MlrvalOrFunction
			reduce(222), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(222), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(222), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(222), // [[, reduce: MlrvalOrFunction
			reduce(222), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1050
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(223), // ;, reduce: MlrvalOrFunction
			reduce(223), // {, reduce: MlrvalOrFunction
			reduce(223), // }, reduce: MlrvalOrFunction
			reduce(223), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(223), // >, reduce: MlrvalOrFunction
			reduce(223), // >>, reduce: MlrvalOrFunction
			reduce(223), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(223), // ||=, reduce: MlrvalOrFunction
			reduce(223), // ^^=, reduce: MlrvalOrFunction
			reduce(223), // &&=, reduce: MlrvalOrFunction
			reduce(223), // ??=, reduce: MlrvalOrFunction
			reduce(223), // ???=, reduce: MlrvalOrFunction
			reduce(223), // |=, reduce: MlrvalOrFunction
			reduce(223), // &=, reduce: MlrvalOrFunction
			reduce(223), // ^=, reduce: MlrvalOrFunction
			reduce(223), // <<=, reduce: MlrvalOrFunction
			reduce(223), // >>=, reduce: MlrvalOrFunction
			reduce(223), // >>>=, reduce: MlrvalOrFunction
			reduce(223), // +=, reduce: MlrvalOrFunction
			reduce(223), // .=, reduce: MlrvalOrFunction
			reduce(223), // -=, reduce: MlrvalOrFunction
			reduce(223), // *=, reduce: MlrvalOrFunction
			reduce(223), // /=, reduce: MlrvalOrFunction
			reduce(223), // //=, reduce: MlrvalOrFunction
			reduce(223), // %=, reduce: MlrvalOrFunction
			reduce(223), // **=, reduce: MlrvalOrFunction
			reduce(223), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(223), // ||, reduce: MlrvalOrFunction
			reduce(223), // ^^, reduce: MlrvalOrFunction
			reduce(223), // &&, reduce: MlrvalOrFunction
			reduce(223), // ??, reduce: MlrvalOrFunction
			reduce(223), // ???, reduce: MlrvalOrFunction
			reduce(223), // =~, reduce: MlrvalOrFunction
			reduce(223), // !=~, reduce: MlrvalOrFunction
			reduce(223), // ==, reduce: MlrvalOrFunction
			reduce(223), // !=, reduce: MlrvalOrFunction
			reduce(223), // <=>, reduce: MlrvalOrFunction
			reduce(223), // >=, reduce: MlrvalOrFunction
			reduce(223), // <, reduce: MlrvalOrFunction
			reduce(223), // <=, reduce: MlrvalOrFunction
			reduce(223), // ^, reduce: MlrvalOrFunction
			reduce(223), // &, reduce: MlrvalOrFunction
			reduce(223), // <<, reduce: MlrvalOrFunction
			reduce(223), // >>>, reduce: MlrvalOrFunction
			reduce(223), // +, reduce: MlrvalOrFunction
			reduce(223), // -, reduce: MlrvalOrFunction
			reduce(223), // .+, reduce: MlrvalOrFunction
			reduce(223), // .-, reduce: MlrvalOrFunction
			reduce(223), // *, reduce: MlrvalOrFunction
			reduce(223), // /, reduce: MlrvalOrFunction
			reduce(223), // //, reduce: MlrvalOrFunction
			reduce(223), // %, reduce: MlrvalOrFunction
			reduce(223), // .*, reduce: MlrvalOrFunction
			reduce(223), // ./, reduce: MlrvalOrFunction
			reduce(223), // .//, reduce: MlrvalOrFunction
			reduce(223), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(223), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(223), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(223), // [[, reduce: MlrvalOrFunction
			reduce(223), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1051
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(224), // ;, reduce: MlrvalOrFunction
			reduce(224), // {, reduce: MlrvalOrFunction
			reduce(224), // }, reduce: MlrvalOrFunction
			reduce(224), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(224), // >, reduce: MlrvalOrFunction
			reduce(224), // >>, reduce: MlrvalOrFunction
			reduce(224), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(224), // ||=, reduce: MlrvalOrFunction
			reduce(224), // ^^=, reduce: MlrvalOrFunction
			reduce(224), // &&=, reduce: MlrvalOrFunction
			reduce(224), // ??=, reduce: MlrvalOrFunction
			reduce(224), // ???=, reduce: MlrvalOrFunction
			reduce(224), // |=, reduce: MlrvalOrFunction
			reduce(224), // &=, reduce: MlrvalOrFunction
			reduce(224), // ^=, reduce: MlrvalOrFunction
			reduce(224), // <<=, reduce: MlrvalOrFunction
			reduce(224), // >>=, reduce: MlrvalOrFunction
			reduce(224), // >>>=, reduce: MlrvalOrFunction
			reduce(224), // +=, reduce: MlrvalOrFunction
			reduce(224), // .=, reduce: MlrvalOrFunction
			reduce(224), // -=, reduce: MlrvalOrFunction
			reduce(224), // *=, reduce: MlrvalOrFunction
			reduce(224), // /=, reduce: MlrvalOrFunction
			reduce(224), // //=, reduce: MlrvalOrFunction
			reduce(224), // %=, reduce: MlrvalOrFunction
			reduce(224), // **=, reduce: MlrvalOrFunction
			reduce(224), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(224), // ||, reduce: MlrvalOrFunction
			reduce(224), // ^^, reduce: MlrvalOrFunction
			reduce(224), // &&, reduce: MlrvalOrFunction
			reduce(224), // ??, reduce: MlrvalOrFunction
			reduce(224), // ???, reduce: MlrvalOrFunction
			reduce(224), // =~, reduce: MlrvalOrFunction
			reduce(224), // !=~, reduce: MlrvalOrFunction
			reduce(224), // ==, reduce: MlrvalOrFunction
			reduce(224), // !=, reduce: MlrvalOrFunction
			reduce(224), // <=>, reduce: MlrvalOrFunction
			reduce(224), // >=, reduce: MlrvalOrFunction
			reduce(224), // <, reduce: MlrvalOrFunction
			reduce(224), // <=, reduce: MlrvalOrFunction
			reduce(224), // ^, reduce: MlrvalOrFunction
			reduce(224), // &, reduce: MlrvalOrFunction
			reduce(224), // <<, reduce: MlrvalOrFunction
			reduce(224), // >>>, reduce: MlrvalOrFunction
			reduce(224), // +, reduce: MlrvalOrFunction
			reduce(224), // -, reduce: MlrvalOrFunction
			reduce(224), // .+, reduce: MlrvalOrFunction
			reduce(224), // .-, reduce: MlrvalOrFunction
			reduce(224), // *, reduce: MlrvalOrFunction
			reduce(224), // /, reduce: MlrvalOrFunction
			reduce(224), // //, reduce: MlrvalOrFunction
			reduce(224), // %, reduce: MlrvalOrFunction
			reduce(224), // .*, reduce: MlrvalOrFunction
			reduce(224), // ./, reduce: MlrvalOrFunction
			reduce(224), // .//, reduce: MlrvalOrFunction
			reduce(224), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(224), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(224), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(224), // [[, reduce: MlrvalOrFunction
			reduce(224), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1052
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(225), // ;, reduce: MlrvalOrFunction
			reduce(225), // {, reduce: MlrvalOrFunction
			reduce(225), // }, reduce: MlrvalOrFunction
			reduce(225), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(225), // >, reduce: MlrvalOrFunction
			reduce(225), // >>, reduce: MlrvalOrFunction
			reduce(225), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(225), // ||=, reduce: MlrvalOrFunction
			reduce(225), // ^^=, reduce: MlrvalOrFunction
			reduce(225), // &&=, reduce: MlrvalOrFunction
			reduce(225), // ??=, reduce: MlrvalOrFunction
			reduce(225), // ???=, reduce: MlrvalOrFunction
			reduce(225), // |=, reduce: MlrvalOrFunction
			reduce(225), // &=, reduce: MlrvalOrFunction
			reduce(225), // ^=, reduce: MlrvalOrFunction
			reduce(225), // <<=, reduce: MlrvalOrFunction
			reduce(225), // >>=, reduce: MlrvalOrFunction
			reduce(225), // >>>=, reduce: MlrvalOrFunction
			reduce(225), // +=, reduce: MlrvalOrFunction
			reduce(225), // .=, reduce: MlrvalOrFunction
			reduce(225), // -=, reduce: MlrvalOrFunction
			reduce(225), // *=, reduce: MlrvalOrFunction
			reduce(225), // /=, reduce: MlrvalOrFunction
			reduce(225), // //=, reduce: MlrvalOrFunction
			reduce(225), // %=, reduce: MlrvalOrFunction
			reduce(225), // **=, reduce: MlrvalOrFunction
			reduce(225), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(225), // ||, reduce: MlrvalOrFunction
			reduce(225), // ^^, reduce: MlrvalOrFunction
			reduce(225), // &&, reduce: MlrvalOrFunction
			reduce(225), // ??, reduce: MlrvalOrFunction
			reduce(225), // ???, reduce: MlrvalOrFunction
			reduce(225), // =~, reduce: MlrvalOrFunction
			reduce(225), // !=~, reduce: MlrvalOrFunction
			reduce(225), // ==, reduce: MlrvalOrFunction
			reduce(225), // !=, reduce: MlrvalOrFunction
			reduce(225), // <=>, reduce: MlrvalOrFunction
			reduce(225), // >=, reduce: MlrvalOrFunction
			reduce(225), // <, reduce: MlrvalOrFunction
			reduce(225), // <=, reduce: MlrvalOrFunction
			reduce(225), // ^, reduce: MlrvalOrFunction
			reduce(225), // &, reduce: MlrvalOrFunction
			reduce(225), // <<, reduce: MlrvalOrFunction
			reduce(225), // >>>, reduce: MlrvalOrFunction
			reduce(225), // +, reduce: MlrvalOrFunction
			reduce(225), // -, reduce: MlrvalOrFunction
			reduce(225), // .+, reduce: MlrvalOrFunction
			reduce(225), // .-, reduce: MlrvalOrFunction
			reduce(225), // *, reduce: MlrvalOrFunction
			reduce(225), // /, reduce: MlrvalOrFunction
			reduce(225), // //, reduce: MlrvalOrFunction
			reduce(225), // %, reduce: MlrvalOrFunction
			reduce(225), // .*, reduce: MlrvalOrFunction
			reduce(225), // ./, reduce: MlrvalOrFunction
			reduce(225), // .//, reduce: MlrvalOrFunction
			reduce(225), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(225), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(225), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(225), // [[, reduce: MlrvalOrFunction
			reduce(225), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1053
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(226), // ;, reduce: MlrvalOrFunction
			reduce(226), // {, reduce: MlrvalOrFunction
			reduce(226), // }, reduce: MlrvalOrFunction
			reduce(226), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(226), // >, reduce: MlrvalOrFunction
			reduce(226), // >>, reduce: MlrvalOrFunction
			reduce(226), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(226), // ||=, reduce: MlrvalOrFunction
			reduce(226), // ^^=, reduce: MlrvalOrFunction
			reduce(226), // &&=, reduce: MlrvalOrFunction
			reduce(226), // ??=, reduce: MlrvalOrFunction
			reduce(226), // ???=, reduce: MlrvalOrFunction
			reduce(226), // |=, reduce: MlrvalOrFunction
			reduce(226), // &=, reduce: MlrvalOrFunction
			reduce(226), // ^=, reduce: MlrvalOrFunction
			reduce(226), // <<=, reduce: MlrvalOrFunction
			reduce(226), // >>=, reduce: MlrvalOrFunction
			reduce(226), // >>>=, reduce: MlrvalOrFunction
			reduce(226), // +=, reduce: MlrvalOrFunction
			reduce(226), // .=, reduce: MlrvalOrFunction
			reduce(226), // -=, reduce: MlrvalOrFunction
			reduce(226), // *=, reduce: MlrvalOrFunction
			reduce(226), // /=, reduce: MlrvalOrFunction
			reduce(226), // //=, reduce: MlrvalOrFunction
			reduce(226), // %=, reduce: MlrvalOrFunction
			reduce(226), // **=, reduce: MlrvalOrFunction
			reduce(226), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(226), // ||, reduce: MlrvalOrFunction
			reduce(226), // ^^, reduce: MlrvalOrFunction
			reduce(226), // &&, reduce: MlrvalOrFunction
			reduce(226), // ??, reduce: MlrvalOrFunction
			reduce(226), // ???, reduce: MlrvalOrFunction
			reduce(226), // =~, reduce: MlrvalOrFunction
			reduce(226), // !=~, reduce: MlrvalOrFunction
			reduce(226), // ==, reduce: MlrvalOrFunction
			reduce(226), // !=, reduce: MlrvalOrFunction
			reduce(226), // <=>, reduce: MlrvalOrFunction
			reduce(226), // >=, reduce: MlrvalOrFunction
			reduce(226), // <, reduce: MlrvalOrFunction
			reduce(226), // <=, reduce: MlrvalOrFunction
			reduce(226), // ^, reduce: MlrvalOrFunction
			reduce(226), // &, reduce: MlrvalOrFunction
			reduce(226), // <<, reduce: MlrvalOrFunction
			reduce(226), // >>>, reduce: MlrvalOrFunction
			reduce(226), // +, reduce: MlrvalOrFunction
			reduce(226), // -, reduce: MlrvalOrFunction
			reduce(226), // .+, reduce: MlrvalOrFunction
			reduce(226), // .-, reduce: MlrvalOrFunction
			reduce(226), // *, reduce: MlrvalOrFunction
			reduce(226), // /, reduce: MlrvalOrFunction
			reduce(226), // //, reduce: MlrvalOrFunction
			reduce(226), // %, reduce: MlrvalOrFunction
			reduce(226), // .*, reduce: MlrvalOrFunction
			reduce(226), // ./, reduce: MlrvalOrFunction
			reduce(226), // .//, reduce: MlrvalOrFunction
			reduce(226), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(226), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(226), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(226), // [[, reduce: MlrvalOrFunction
			reduce(226), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1054
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(227), // ;, reduce: MlrvalOrFunction
			reduce(227), // {, reduce: MlrvalOrFunction
			reduce(227), // }, reduce: MlrvalOrFunction
			reduce(227), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(227), // >, reduce: MlrvalOrFunction
			reduce(227), // >>, reduce: MlrvalOrFunction
			reduce(227), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(227), // ||=, reduce: MlrvalOrFunction
			reduce(227), // ^^=, reduce: MlrvalOrFunction
			reduce(227), // &&=, reduce: MlrvalOrFunction
			reduce(227), // ??=, reduce: MlrvalOrFunction
			reduce(227), // ???=, reduce: MlrvalOrFunction
			reduce(227), // |=, reduce: MlrvalOrFunction
			reduce(227), // &=, reduce: MlrvalOrFunction
			reduce(227), // ^=, reduce: MlrvalOrFunction
			reduce(227), // <<=, reduce: MlrvalOrFunction
			reduce(227), // >>=, reduce: MlrvalOrFunction
			reduce(227), // >>>=, reduce: MlrvalOrFunction
			reduce(227), // +=, reduce: MlrvalOrFunction
			reduce(227), // .=, reduce: MlrvalOrFunction
			reduce(227), // -=, reduce: MlrvalOrFunction
			reduce(227), // *=, reduce: MlrvalOrFunction
			reduce(227), // /=, reduce: MlrvalOrFunction
			reduce(227), // //=, reduce: MlrvalOrFunction
			reduce(227), // %=, reduce: MlrvalOrFunction
			reduce(227), // **=, reduce: MlrvalOrFunction
			reduce(227), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(227), // ||, reduce: MlrvalOrFunction
			reduce(227), // ^^, reduce: MlrvalOrFunction
			reduce(227), // &&, reduce: MlrvalOrFunction
			reduce(227), // ??, reduce: MlrvalOrFunction
			reduce(227), // ???, reduce: MlrvalOrFunction
			reduce(227), // =~, reduce: MlrvalOrFunction
			reduce(227), // !=~, reduce: MlrvalOrFunction
			reduce(227), // ==, reduce: MlrvalOrFunction
			reduce(227), // !=, reduce: MlrvalOrFunction
			reduce(227), // <=>, reduce: MlrvalOrFunction
			reduce(227), // >=, reduce: MlrvalOrFunction
			reduce(227), // <, reduce: MlrvalOrFunction
			reduce(227), // <=, reduce: MlrvalOrFunction
			reduce(227), // ^, reduce: MlrvalOrFunction
			reduce(227), // &, reduce: MlrvalOrFunction
			reduce(227), // <<, reduce: MlrvalOrFunction
			reduce(227), // >>>, reduce: MlrvalOrFunction
			reduce(227), // +, reduce: MlrvalOrFunction
			reduce(227), // -, reduce: MlrvalOrFunction
			reduce(227), // .+, reduce: MlrvalOrFunction
			reduce(227), // .-, reduce: MlrvalOrFunction
			reduce(227), // *, reduce: MlrvalOrFunction
			reduce(227), // /, reduce: MlrvalOrFunction
			reduce(227), // //, reduce: MlrvalOrFunction
			reduce(227), // %, reduce: MlrvalOrFunction
			reduce(227), // .*, reduce: MlrvalOrFunction
			reduce(227), // ./, reduce: MlrvalOrFunction
			reduce(227), // .//, reduce: MlrvalOrFunction
			reduce(227), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(227), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(227), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(227), // [[, reduce: MlrvalOrFunction
			reduce(227), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1055
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(752),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(756),  // (
			nil,         // )
			nil,         // emitp
			shift(768),  // field_name
			shift(769),  // $[
			shift(2320), // ]
			shift(771),  // braced_field_name
			shift(772),  // $[[
			shift(773),  // $[[[
			shift(774),  // full_srec
			shift(776),  // oosvar_name
			shift(777),  // @[
			shift(778),  // braced_oosvar_name
			shift(779),  // full_oosvar
			shift(780),  // all
			shift(781),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(796),  // +
			shift(798),  // -
			shift(799),  // .+
			shift(800),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(803),  // !
			shift(804),  // ~
			nil,         // **
			shift(809),  // string_literal
			shift(810),  // regex_case_insensitive
			shift(811),  // int_literal
			shift(812),  // float_literal
			shift(813),  // boolean_literal
			shift(814),  // null_literal
			shift(815),  // inf_literal
			shift(816),  // nan_literal
			shift(817),  // const_M_PI
			shift(818),  // const_M_E
			shift(819),  // panic
			shift(821),  // [
			shift(824),  // ctx_IPS
			shift(825),  // ctx_IFS
			shift(826),  // ctx_IRS
			shift(827),  // ctx_OPS
			shift(828),  // ctx_OFS
			shift(829),  // ctx_ORS
			shift(830),  // ctx_FLATSEP
			shift(831),  // ctx_NF
			shift(832),  // ctx_NR
			shift(833),  // ctx_FNR
			shift(834),  // ctx_FILENAME
			shift(835),  // ctx_FILENUM
			shift(837),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(844),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1056
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(240), // ;, reduce: MlrvalOrFunction
			reduce(240), // {, reduce: MlrvalOrFunction
			reduce(240), // }, reduce: MlrvalOrFunction
			reduce(240), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(240), // >, reduce: MlrvalOrFunction
			reduce(240), // >>, reduce: MlrvalOrFunction
			reduce(240), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(240), // ||=, reduce: MlrvalOrFunction
			reduce(240), // ^^=, reduce: MlrvalOrFunction
			reduce(240), // &&=, reduce: MlrvalOrFunction
			reduce(240), // ??=, reduce: MlrvalOrFunction
			reduce(240), // ???=, reduce: MlrvalOrFunction
			reduce(240), // |=, reduce: MlrvalOrFunction
			reduce(240), // &=, reduce: MlrvalOrFunction
			reduce(240), // ^=, reduce: MlrvalOrFunction
			reduce(240), // <<=, reduce: MlrvalOrFunction
			reduce(240), // >>=, reduce: MlrvalOrFunction
			reduce(240), // >>>=, reduce: MlrvalOrFunction
			reduce(240), // +=, reduce: MlrvalOrFunction
			reduce(240), // .=, reduce: MlrvalOrFunction
			reduce(240), // -=, reduce: MlrvalOrFunction
			reduce(240), // *=, reduce: MlrvalOrFunction
			reduce(240), // /=, reduce: MlrvalOrFunction
			reduce(240), // //=, reduce: MlrvalOrFunction
			reduce(240), // %=, reduce: MlrvalOrFunction
			reduce(240), // **=, reduce: MlrvalOrFunction
			reduce(240), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(240), // ||, reduce: MlrvalOrFunction
			reduce(240), // ^^, reduce: MlrvalOrFunction
			reduce(240), // &&, reduce: MlrvalOrFunction
			reduce(240), // ??, reduce: MlrvalOrFunction
			reduce(240), // ???, reduce: MlrvalOrFunction
			reduce(240), // =~, reduce: MlrvalOrFunction
			reduce(240), // !=~, reduce: MlrvalOrFunction
			reduce(240), // ==, reduce: MlrvalOrFunction
			reduce(240), // !=, reduce: MlrvalOrFunction
			reduce(240), // <=>, reduce: MlrvalOrFunction
			reduce(240), // >=, reduce: MlrvalOrFunction
			reduce(240), // <, reduce: MlrvalOrFunction
			reduce(240), // <=, reduce: MlrvalOrFunction
			reduce(240), // ^, reduce: MlrvalOrFunction
			reduce(240), // &, reduce: MlrvalOrFunction
			reduce(240), // <<, reduce: MlrvalOrFunction
			reduce(240), // >>>, reduce: MlrvalOrFunction
			reduce(240), // +, reduce: MlrvalOrFunction
			reduce(240), // -, reduce: MlrvalOrFunction
			reduce(240), // .+, reduce: MlrvalOrFunction
			reduce(240), // .-, reduce: MlrvalOrFunction
			reduce(240), // *, reduce: MlrvalOrFunction
			reduce(240), // /, reduce: MlrvalOrFunction
			reduce(240), // //, reduce: MlrvalOrFunction
			reduce(240), // %, reduce: MlrvalOrFunction
			reduce(240), // .*, reduce: MlrvalOrFunction
			reduce(240), // ./, reduce: MlrvalOrFunction
			reduce(240), // .//, reduce: MlrvalOrFunction
			reduce(240), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(240), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(240), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(240), // [[, reduce: MlrvalOrFunction
			reduce(240), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1057
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(241), // ;, reduce: ContextVariable
			reduce(241), // {, reduce: ContextVariable
			reduce(241), // }, reduce: ContextVariable
			reduce(241), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(241), // >, reduce: ContextVariable
			reduce(241), // >>, reduce: ContextVariable
			reduce(241), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(241), // ||=, reduce: ContextVariable
			reduce(241), // ^^=, reduce: ContextVariable
			reduce(241), // &&=, reduce: ContextVariable
			reduce(241), // ??=, reduce: ContextVariable
			reduce(241), // ???=, reduce: ContextVariable
			reduce(241), // |=, reduce: ContextVariable
			reduce(241), // &=, reduce: ContextVariable
			reduce(241), // ^=, reduce: ContextVariable
			reduce(241), // <<=, reduce: ContextVariable
			reduce(241), // >>=, reduce: ContextVariable
			reduce(241), // >>>=, reduce: ContextVariable
			reduce(241), // +=, reduce: ContextVariable
			reduce(241), // .=, reduce: ContextVariable
			reduce(241), // -=, reduce: ContextVariable
			reduce(241), // *=, reduce: ContextVariable
			reduce(241), // /=, reduce: ContextVariable
			reduce(241), // //=, reduce: ContextVariable
			reduce(241), // %=, reduce: ContextVariable
			reduce(241), // **=, reduce: ContextVariable
			reduce(241), // ?, reduce: ContextVariable
			nil,         // :
			reduce(241), // ||, reduce: ContextVariable
			reduce(241), // ^^, reduce: ContextVariable
			reduce(241), // &&, reduce: ContextVariable
			reduce(241), // ??, reduce: ContextVariable
			reduce(241), // ???, reduce: ContextVariable
			reduce(241), // =~, reduce: ContextVariable
			reduce(241), // !=~, reduce: ContextVariable
			reduce(241), // ==, reduce: ContextVariable
			reduce(241), // !=, reduce: ContextVariable
			reduce(241), // <=>, reduce: ContextVariable
			reduce(241), // >=, reduce: ContextVariable
			reduce(241), // <, reduce: ContextVariable
			reduce(241), // <=, reduce: ContextVariable
			reduce(241), // ^, reduce: ContextVariable
			reduce(241), // &, reduce: ContextVariable
			reduce(241), // <<, reduce: ContextVariable
			reduce(241), // >>>, reduce: ContextVariable
			reduce(241), // +, reduce: ContextVariable
			reduce(241), // -, reduce: ContextVariable
			reduce(241), // .+, reduce: ContextVariable
			reduce(241), // .-, reduce: ContextVariable
			reduce(241), // *, reduce: ContextVariable
			reduce(241), // /, reduce: ContextVariable
			reduce(241), // //, reduce: ContextVariable
			reduce(241), // %, reduce: ContextVariable
			reduce(241), // .*, reduce: ContextVariable
			reduce(241), // ./, reduce: ContextVariable
			reduce(241), // .//, reduce: ContextVariable
			reduce(241), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(241), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(241), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(241), // [[, reduce: ContextVariable
			reduce(241), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1058
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(242), // ;, reduce: ContextVariable
			reduce(242), // {, reduce: ContextVariable
			reduce(242), // }, reduce: ContextVariable
			reduce(242), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(242), // >, reduce: ContextVariable
			reduce(242), // >>, reduce: ContextVariable
			reduce(242), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(242), // ||=, reduce: ContextVariable
			reduce(242), // ^^=, reduce: ContextVariable
			reduce(242), // &&=, reduce: ContextVariable
			reduce(242), // ??=, reduce: ContextVariable
			reduce(242), // ???=, reduce: ContextVariable
			reduce(242), // |=, reduce: ContextVariable
			reduce(242), // &=, reduce: ContextVariable
			reduce(242), // ^=, reduce: ContextVariable
			reduce(242), // <<=, reduce: ContextVariable
			reduce(242), // >>=, reduce: ContextVariable
			reduce(242), // >>>=, reduce: ContextVariable
			reduce(242), // +=, reduce: ContextVariable
			reduce(242), // .=, reduce: ContextVariable
			reduce(242), // -=, reduce: ContextVariable
			reduce(242), // *=, reduce: ContextVariable
			reduce(242), // /=, reduce: ContextVariable
			reduce(242), // //=, reduce: ContextVariable
			reduce(242), // %=, reduce: ContextVariable
			reduce(242), // **=, reduce: ContextVariable
			reduce(242), // ?, reduce: ContextVariable
			nil,         // :
			reduce(242), // ||, reduce: ContextVariable
			reduce(242), // ^^, reduce: ContextVariable
			reduce(242), // &&, reduce: ContextVariable
			reduce(242), // ??, reduce: ContextVariable
			reduce(242), // ???, reduce: ContextVariable
			reduce(242), // =~, reduce: ContextVariable
			reduce(242), // !=~, reduce: ContextVariable
			reduce(242), // ==, reduce: ContextVariable
			reduce(242), // !=, reduce: ContextVariable
			reduce(242), // <=>, reduce: ContextVariable
			reduce(242), // >=, reduce: ContextVariable
			reduce(242), // <, reduce: ContextVariable
			reduce(242), // <=, reduce: ContextVariable
			reduce(242), // ^, reduce: ContextVariable
			reduce(242), // &, reduce: ContextVariable
			reduce(242), // <<, reduce: ContextVariable
			reduce(242), // >>>, reduce: ContextVariable
			reduce(242), // +, reduce: ContextVariable
			reduce(242), // -, reduce: ContextVariable
			reduce(242), // .+, reduce: ContextVariable
			reduce(242), // .-, reduce: ContextVariable
			reduce(242), // *, reduce: ContextVariable
			reduce(242), // /, reduce: ContextVariable
			reduce(242), // //, reduce: ContextVariable
			reduce(242), // %, reduce: ContextVariable
			reduce(242), // .*, reduce: ContextVariable
			reduce(242), // ./, reduce: ContextVariable
			reduce(242), // .//, reduce: ContextVariable
			reduce(242), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(242), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(242), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(242), // [[, reduce: ContextVariable
			reduce(242), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1059
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(243), // ;, reduce: ContextVariable
			reduce(243), // {, reduce: ContextVariable
			reduce(243), // }, reduce: ContextVariable
			reduce(243), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(243), // >, reduce: ContextVariable
			reduce(243), // >>, reduce: ContextVariable
			reduce(243), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(243), // ||=, reduce: ContextVariable
			reduce(243), // ^^=, reduce: ContextVariable
			reduce(243), // &&=, reduce: ContextVariable
			reduce(243), // ??=, reduce: ContextVariable
			reduce(243), // ???=, reduce: ContextVariable
			reduce(243), // |=, reduce: ContextVariable
			reduce(243), // &=, reduce: ContextVariable
			reduce(243), // ^=, reduce: ContextVariable
			reduce(243), // <<=, reduce: ContextVariable
			reduce(243), // >>=, reduce: ContextVariable
			reduce(243), // >>>=, reduce: ContextVariable
			reduce(243), // +=, reduce: ContextVariable
			reduce(243), // .=, reduce: ContextVariable
			reduce(243), // -=, reduce: ContextVariable
			reduce(243), // *=, reduce: ContextVariable
			reduce(243), // /=, reduce: ContextVariable
			reduce(243), // //=, reduce: ContextVariable
			reduce(243), // %=, reduce: ContextVariable
			reduce(243), // **=, reduce: ContextVariable
			reduce(243), // ?, reduce: ContextVariable
			nil,         // :
			reduce(243), // ||, reduce: ContextVariable
			reduce(243), // ^^, reduce: ContextVariable
			reduce(243), // &&, reduce: ContextVariable
			reduce(243), // ??, reduce: ContextVariable
			reduce(243), // ???, reduce: ContextVariable
			reduce(243), // =~, reduce: ContextVariable
			reduce(243), // !=~, reduce: ContextVariable
			reduce(243), // ==, reduce: ContextVariable
			reduce(243), // !=, reduce: ContextVariable
			reduce(243), // <=>, reduce: ContextVariable
			reduce(243), // >=, reduce: ContextVariable
			reduce(243), // <, reduce: ContextVariable
			reduce(243), // <=, reduce: ContextVariable
			reduce(243), // ^, reduce: ContextVariable
			reduce(243), // &, reduce: ContextVariable
			reduce(243), // <<, reduce: ContextVariable
			reduce(243), // >>>, reduce: ContextVariable
			reduce(243), // +, reduce: ContextVariable
			reduce(243), // -, reduce: ContextVariable
			reduce(243), // .+, reduce: ContextVariable
			reduce(243), // .-, reduce: ContextVariable
			reduce(243), // *, reduce: ContextVariable
			reduce(243), // /, reduce: ContextVariable
			reduce(243), // //, reduce: ContextVariable
			reduce(243), // %, reduce: ContextVariable
			reduce(243), // .*, reduce: ContextVariable
			reduce(243), // ./, reduce: ContextVariable
			reduce(243), // .//, reduce: ContextVariable
			reduce(243), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(243), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(243), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(243), // [[, reduce: ContextVariable
			reduce(243), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1060
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(244), // ;, reduce: ContextVariable
			reduce(244), // {, reduce: ContextVariable
			reduce(244), // }, reduce: ContextVariable
			reduce(244), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(244), // >, reduce: ContextVariable
			reduce(244), // >>, reduce: ContextVariable
			reduce(244), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(244), // ||=, reduce: ContextVariable
			reduce(244), // ^^=, reduce: ContextVariable
			reduce(244), // &&=, reduce: ContextVariable
			reduce(244), // ??=, reduce: ContextVariable
			reduce(244), // ???=, reduce: ContextVariable
			reduce(244), // |=, reduce: ContextVariable
			reduce(244), // &=, reduce: ContextVariable
			reduce(244), // ^=, reduce: ContextVariable
			reduce(244), // <<=, reduce: ContextVariable
			reduce(244), // >>=, reduce: ContextVariable
			reduce(244), // >>>=, reduce: ContextVariable
			reduce(244), // +=, reduce: ContextVariable
			reduce(244), // .=, reduce: ContextVariable
			reduce(244), // -=, reduce: ContextVariable
			reduce(244), // *=, reduce: ContextVariable
			reduce(244), // /=, reduce: ContextVariable
			reduce(244), // //=, reduce: ContextVariable
			reduce(244), // %=, reduce: ContextVariable
			reduce(244), // **=, reduce: ContextVariable
			reduce(244), // ?, reduce: ContextVariable
			nil,         // :
			reduce(244), // ||, reduce: ContextVariable
			reduce(244), // ^^, reduce: ContextVariable
			reduce(244), // &&, reduce: ContextVariable
			reduce(244), // ??, reduce: ContextVariable
			reduce(244), // ???, reduce: ContextVariable
			reduce(244), // =~, reduce: ContextVariable
			reduce(244), // !=~, reduce: ContextVariable
			reduce(244), // ==, reduce: ContextVariable
			reduce(244), // !=, reduce: ContextVariable
			reduce(244), // <=>, reduce: ContextVariable
			reduce(244), // >=, reduce: ContextVariable
			reduce(244), // <, reduce: ContextVariable
			reduce(244), // <=, reduce: ContextVariable
			reduce(244), // ^, reduce: ContextVariable
			reduce(244), // &, reduce: ContextVariable
			reduce(244), // <<, reduce: ContextVariable
			reduce(244), // >>>, reduce: ContextVariable
			reduce(244), // +, reduce: ContextVariable
			reduce(244), // -, reduce: ContextVariable
			reduce(244), // .+, reduce: ContextVariable
			reduce(244), // .-, reduce: ContextVariable
			reduce(244), // *, reduce: ContextVariable
			reduce(244), // /, reduce: ContextVariable
			reduce(244), // //, reduce: ContextVariable
			reduce(244), // %, reduce: ContextVariable
			reduce(244), // .*, reduce: ContextVariable
			reduce(244), // ./, reduce: ContextVariable
			reduce(244), // .//, reduce: ContextVariable
			reduce(244), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(244), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(244), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(244), // [[, reduce: ContextVariable
			reduce(244), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1061
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(245), // ;, reduce: ContextVariable
			reduce(245), // {, reduce: ContextVariable
			reduce(245), // }, reduce: ContextVariable
			reduce(245), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(245), // >, reduce: ContextVariable
			reduce(245), // >>, reduce: ContextVariable
			reduce(245), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(245), // ||=, reduce: ContextVariable
			reduce(245), // ^^=, reduce: ContextVariable
			reduce(245), // &&=, reduce: ContextVariable
			reduce(245), // ??=, reduce: ContextVariable
			reduce(245), // ???=, reduce: ContextVariable
			reduce(245), // |=, reduce: ContextVariable
			reduce(245), // &=, reduce: ContextVariable
			reduce(245), // ^=, reduce: ContextVariable
			reduce(245), // <<=, reduce: ContextVariable
			reduce(245), // >>=, reduce: ContextVariable
			reduce(245), // >>>=, reduce: ContextVariable
			reduce(245), // +=, reduce: ContextVariable
			reduce(245), // .=, reduce: ContextVariable
			reduce(245), // -=, reduce: ContextVariable
			reduce(245), // *=, reduce: ContextVariable
			reduce(245), // /=, reduce: ContextVariable
			reduce(245), // //=, reduce: ContextVariable
			reduce(245), // %=, reduce: ContextVariable
			reduce(245), // **=, reduce: ContextVariable
			reduce(245), // ?, reduce: ContextVariable
			nil,         // :
			reduce(245), // ||, reduce: ContextVariable
			reduce(245), // ^^, reduce: ContextVariable
			reduce(245), // &&, reduce: ContextVariable
			reduce(245), // ??, reduce: ContextVariable
			reduce(245), // ???, reduce: ContextVariable
			reduce(245), // =~, reduce: ContextVariable
			reduce(245), // !=~, reduce: ContextVariable
			reduce(245), // ==, reduce: ContextVariable
			reduce(245), // !=, reduce: ContextVariable
			reduce(245), // <=>, reduce: ContextVariable
			reduce(245), // >=, reduce: ContextVariable
			reduce(245), // <, reduce: ContextVariable
			reduce(245), // <=, reduce: ContextVariable
			reduce(245), // ^, reduce: ContextVariable
			reduce(245), // &, reduce: ContextVariable
			reduce(245), // <<, reduce: ContextVariable
			reduce(245), // >>>, reduce: ContextVariable
			reduce(245), // +, reduce: ContextVariable
			reduce(245), // -, reduce: ContextVariable
			reduce(245), // .+, reduce: ContextVariable
			reduce(245), // .-, reduce: ContextVariable
			reduce(245), // *, reduce: ContextVariable
			reduce(245), // /, reduce: ContextVariable
			reduce(245), // //, reduce: ContextVariable
			reduce(245), // %, reduce: ContextVariable
			reduce(245), // .*, reduce: ContextVariable
			reduce(245), // ./, reduce: ContextVariable
			reduce(245), // .//, reduce: ContextVariable
			reduce(245), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(245), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(245), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(245), // [[, reduce: ContextVariable
			reduce(245), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1062
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(246), // ;, reduce: ContextVariable
			reduce(246), // {, reduce: ContextVariable
			reduce(246), // }, reduce: ContextVariable
			reduce(246), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(246), // >, reduce: ContextVariable
			reduce(246), // >>, reduce: ContextVariable
			reduce(246), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(246), // ||=, reduce: ContextVariable
			reduce(246), // ^^=, reduce: ContextVariable
			reduce(246), // &&=, reduce: ContextVariable
			reduce(246), // ??=, reduce: ContextVariable
			reduce(246), // ???=, reduce: ContextVariable
			reduce(246), // |=, reduce: ContextVariable
			reduce(246), // &=, reduce: ContextVariable
			reduce(246), // ^=, reduce: ContextVariable
			reduce(246), // <<=, reduce: ContextVariable
			reduce(246), // >>=, reduce: ContextVariable
			reduce(246), // >>>=, reduce: ContextVariable
			reduce(246), // +=, reduce: ContextVariable
			reduce(246), // .=, reduce: ContextVariable
			reduce(246), // -=, reduce: ContextVariable
			reduce(246), // *=, reduce: ContextVariable
			reduce(246), // /=, reduce: ContextVariable
			reduce(246), // //=, reduce: ContextVariable
			reduce(246), // %=, reduce: ContextVariable
			reduce(246), // **=, reduce: ContextVariable
			reduce(246), // ?, reduce: ContextVariable
			nil,         // :
			reduce(246), // ||, reduce: ContextVariable
			reduce(246), // ^^, reduce: ContextVariable
			reduce(246), // &&, reduce: ContextVariable
			reduce(246), // ??, reduce: ContextVariable
			reduce(246), // ???, reduce: ContextVariable
			reduce(246), // =~, reduce: ContextVariable
			reduce(246), // !=~, reduce: ContextVariable
			reduce(246), // ==, reduce: ContextVariable
			reduce(246), // !=, reduce: ContextVariable
			reduce(246), // <=>, reduce: ContextVariable
			reduce(246), // >=, reduce: ContextVariable
			reduce(246), // <, reduce: ContextVariable
			reduce(246), // <=, reduce: ContextVariable
			reduce(246), // ^, reduce: ContextVariable
			reduce(246), // &, reduce: ContextVariable
			reduce(246), // <<, reduce: ContextVariable
			reduce(246), // >>>, reduce: ContextVariable
			reduce(246), // +, reduce: ContextVariable
			reduce(246), // -, reduce: ContextVariable
			reduce(246), // .+, reduce: ContextVariable
			reduce(246), // .-, reduce: ContextVariable
			reduce(246), // *, reduce: ContextVariable
			reduce(246), // /, reduce: ContextVariable
			reduce(246), // //, reduce: ContextVariable
			reduce(246), // %, reduce: ContextVariable
			reduce(246), // .*, reduce: ContextVariable
			reduce(246), // ./, reduce: ContextVariable
			reduce(246), // .//, reduce: ContextVariable
			reduce(246), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(246), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(246), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(246), // [[, reduce: ContextVariable
			reduce(246), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1063
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(247), // ;, reduce: ContextVariable
			reduce(247), // {, reduce: ContextVariable
			reduce(247), // }, reduce: ContextVariable
			reduce(247), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(247), // >, reduce: ContextVariable
			reduce(247), // >>, reduce: ContextVariable
			reduce(247), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(247), // ||=, reduce: ContextVariable
			reduce(247), // ^^=, reduce: ContextVariable
			reduce(247), // &&=, reduce: ContextVariable
			reduce(247), // ??=, reduce: ContextVariable
			reduce(247), // ???=, reduce: ContextVariable
			reduce(247), // |=, reduce: ContextVariable
			reduce(247), // &=, reduce: ContextVariable
			reduce(247), // ^=, reduce: ContextVariable
			reduce(247), // <<=, reduce: ContextVariable
			reduce(247), // >>=, reduce: ContextVariable
			reduce(247), // >>>=, reduce: ContextVariable
			reduce(247), // +=, reduce: ContextVariable
			reduce(247), // .=, reduce: ContextVariable
			reduce(247), // -=, reduce: ContextVariable
			reduce(247), // *=, reduce: ContextVariable
			reduce(247), // /=, reduce: ContextVariable
			reduce(247), // //=, reduce: ContextVariable
			reduce(247), // %=, reduce: ContextVariable
			reduce(247), // **=, reduce: ContextVariable
			reduce(247), // ?, reduce: ContextVariable
			nil,         // :
			reduce(247), // ||, reduce: ContextVariable
			reduce(247), // ^^, reduce: ContextVariable
			reduce(247), // &&, reduce: ContextVariable
			reduce(247), // ??, reduce: ContextVariable
			reduce(247), // ???, reduce: ContextVariable
			reduce(247), // =~, reduce: ContextVariable
			reduce(247), // !=~, reduce: ContextVariable
			reduce(247), // ==, reduce: ContextVariable
			reduce(247), // !=, reduce: ContextVariable
			reduce(247), // <=>, reduce: ContextVariable
			reduce(247), // >=, reduce: ContextVariable
			reduce(247), // <, reduce: ContextVariable
			reduce(247), // <=, reduce: ContextVariable
			reduce(247), // ^, reduce: ContextVariable
			reduce(247), // &, reduce: ContextVariable
			reduce(247), // <<, reduce: ContextVariable
			reduce(247), // >>>, reduce: ContextVariable
			reduce(247), // +, reduce: ContextVariable
			reduce(247), // -, reduce: ContextVariable
			reduce(247), // .+, reduce: ContextVariable
			reduce(247), // .-, reduce: ContextVariable
			reduce(247), // *, reduce: ContextVariable
			reduce(247), // /, reduce: ContextVariable
			reduce(247), // //, reduce: ContextVariable
			reduce(247), // %, reduce: ContextVariable
			reduce(247), // .*, reduce: ContextVariable
			reduce(247), // ./, reduce: ContextVariable
			reduce(247), // .//, reduce: ContextVariable
			reduce(247), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(247), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(247), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(247), // [[, reduce: ContextVariable
			reduce(247), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1064
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(248), // ;, reduce: ContextVariable
			reduce(248), // {, reduce: ContextVariable
			reduce(248), // }, reduce: ContextVariable
			reduce(248), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(248), // >, reduce: ContextVariable
			reduce(248), // >>, reduce: ContextVariable
			reduce(248), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(248), // ||=, reduce: ContextVariable
			reduce(248), // ^^=, reduce: ContextVariable
			reduce(248), // &&=, reduce: ContextVariable
			reduce(248), // ??=, reduce: ContextVariable
			reduce(248), // ???=, reduce: ContextVariable
			reduce(248), // |=, reduce: ContextVariable
			reduce(248), // &=, reduce: ContextVariable
			reduce(248), // ^=, reduce: ContextVariable
			reduce(248), // <<=, reduce: ContextVariable
			reduce(248), // >>=, reduce: ContextVariable
			reduce(248), // >>>=, reduce: ContextVariable
			reduce(248), // +=, reduce: ContextVariable
			reduce(248), // .=, reduce: ContextVariable
			reduce(248), // -=, reduce: ContextVariable
			reduce(248), // *=, reduce: ContextVariable
			reduce(248), // /=, reduce: ContextVariable
			reduce(248), // //=, reduce: ContextVariable
			reduce(248), // %=, reduce: ContextVariable
			reduce(248), // **=, reduce: ContextVariable
			reduce(248), // ?, reduce: ContextVariable
			nil,         // :
			reduce(248), // ||, reduce: ContextVariable
			reduce(248), // ^^, reduce: ContextVariable
			reduce(248), // &&, reduce: ContextVariable
			reduce(248), // ??, reduce: ContextVariable
			reduce(248), // ???, reduce: ContextVariable
			reduce(248), // =~, reduce: ContextVariable
			reduce(248), // !=~, reduce: ContextVariable
			reduce(248), // ==, reduce: ContextVariable
			reduce(248), // !=, reduce: ContextVariable
			reduce(248), // <=>, reduce: ContextVariable
			reduce(248), // >=, reduce: ContextVariable
			reduce(248), // <, reduce: ContextVariable
			reduce(248), // <=, reduce: ContextVariable
			reduce(248), // ^, reduce: ContextVariable
			reduce(248), // &, reduce: ContextVariable
			reduce(248), // <<, reduce: ContextVariable
			reduce(248), // >>>, reduce: ContextVariable
			reduce(248), // +, reduce: ContextVariable
			reduce(248), // -, reduce: ContextVariable
			reduce(248), // .+, reduce: ContextVariable
			reduce(248), // .-, reduce: ContextVariable
			reduce(248), // *, reduce: ContextVariable
			reduce(248), // /, reduce: ContextVariable
			reduce(248), // //, reduce: ContextVariable
			reduce(248), // %, reduce: ContextVariable
			reduce(248), // .*, reduce: ContextVariable
			reduce(248), // ./, reduce: ContextVariable
			reduce(248), // .//, reduce: ContextVariable
			reduce(248), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(248), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(248), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(248), // [[, reduce: ContextVariable
			reduce(248), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1065
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(249), // ;, reduce: ContextVariable
			reduce(249), // {, reduce: ContextVariable
			reduce(249), // }, reduce: ContextVariable
			reduce(249), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(249), // >, reduce: ContextVariable
			reduce(249), // >>, reduce: ContextVariable
			reduce(249), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(249), // ||=, reduce: ContextVariable
			reduce(249), // ^^=, reduce: ContextVariable
			reduce(249), // &&=, reduce: ContextVariable
			reduce(249), // ??=, reduce: ContextVariable
			reduce(249), // ???=, reduce: ContextVariable
			reduce(249), // |=, reduce: ContextVariable
			reduce(249), // &=, reduce: ContextVariable
			reduce(249), // ^=, reduce: ContextVariable
			reduce(249), // <<=, reduce: ContextVariable
			reduce(249), // >>=, reduce: ContextVariable
			reduce(249), // >>>=, reduce: ContextVariable
			reduce(249), // +=, reduce: ContextVariable
			reduce(249), // .=, reduce: ContextVariable
			reduce(249), // -=, reduce: ContextVariable
			reduce(249), // *=, reduce: ContextVariable
			reduce(249), // /=, reduce: ContextVariable
			reduce(249), // //=, reduce: ContextVariable
			reduce(249), // %=, reduce: ContextVariable
			reduce(249), // **=, reduce: ContextVariable
			reduce(249), // ?, reduce: ContextVariable
			nil,         // :
			reduce(249), // ||, reduce: ContextVariable
			reduce(249), // ^^, reduce: ContextVariable
			reduce(249), // &&, reduce: ContextVariable
			reduce(249), // ??, reduce: ContextVariable
			reduce(249), // ???, reduce: ContextVariable
			reduce(249), // =~, reduce: ContextVariable
			reduce(249), // !=~, reduce: ContextVariable
			reduce(249), // ==, reduce: ContextVariable
			reduce(249), // !=, reduce: ContextVariable
			reduce(249), // <=>, reduce: ContextVariable
			reduce(249), // >=, reduce: ContextVariable
			reduce(249), // <, reduce: ContextVariable
			reduce(249), // <=, reduce: ContextVariable
			reduce(249), // ^, reduce: ContextVariable
			reduce(249), // &, reduce: ContextVariable
			reduce(249), // <<, reduce: ContextVariable
			reduce(249), // >>>, reduce: ContextVariable
			reduce(249), // +, reduce: ContextVariable
			reduce(249), // -, reduce: ContextVariable
			reduce(249), // .+, reduce: ContextVariable
			reduce(249), // .-, reduce: ContextVariable
			reduce(249), // *, reduce: ContextVariable
			reduce(249), // /, reduce: ContextVariable
			reduce(249), // //, reduce: ContextVariable
			reduce(249), // %, reduce: ContextVariable
			reduce(249), // .*, reduce: ContextVariable
			reduce(249), // ./, reduce: ContextVariable
			reduce(249), // .//, reduce: ContextVariable
			reduce(249), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(249), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(249), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(249), // [[, reduce: ContextVariable
			reduce(249), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1066
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(250), // ;, reduce: ContextVariable
			reduce(250), // {, reduce: ContextVariable
			reduce(250), // }, reduce: ContextVariable
			reduce(250), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(250), // >, reduce: ContextVariable
			reduce(250), // >>, reduce: ContextVariable
			reduce(250), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(250), // ||=, reduce: ContextVariable
			reduce(250), // ^^=, reduce: ContextVariable
			reduce(250), // &&=, reduce: ContextVariable
			reduce(250), // ??=, reduce: ContextVariable
			reduce(250), // ???=, reduce: ContextVariable
			reduce(250), // |=, reduce: ContextVariable
			reduce(250), // &=, reduce: ContextVariable
			reduce(250), // ^=, reduce: ContextVariable
			reduce(250), // <<=, reduce: ContextVariable
			reduce(250), // >>=, reduce: ContextVariable
			reduce(250), // >>>=, reduce: ContextVariable
			reduce(250), // +=, reduce: ContextVariable
			reduce(250), // .=, reduce: ContextVariable
			reduce(250), // -=, reduce: ContextVariable
			reduce(250), // *=, reduce: ContextVariable
			reduce(250), // /=, reduce: ContextVariable
			reduce(250), // //=, reduce: ContextVariable
			reduce(250), // %=, reduce: ContextVariable
			reduce(250), // **=, reduce: ContextVariable
			reduce(250), // ?, reduce: ContextVariable
			nil,         // :
			reduce(250), // ||, reduce: ContextVariable
			reduce(250), // ^^, reduce: ContextVariable
			reduce(250), // &&, reduce: ContextVariable
			reduce(250), // ??, reduce: ContextVariable
			reduce(250), // ???, reduce: ContextVariable
			reduce(250), // =~, reduce: ContextVariable
			reduce(250), // !=~, reduce: ContextVariable
			reduce(250), // ==, reduce: ContextVariable
			reduce(250), // !=, reduce: ContextVariable
			reduce(250), // <=>, reduce: ContextVariable
			reduce(250), // >=, reduce: ContextVariable
			reduce(250), // <, reduce: ContextVariable
			reduce(250), // <=, reduce: ContextVariable
			reduce(250), // ^, reduce: ContextVariable
			reduce(250), // &, reduce: ContextVariable
			reduce(250), // <<, reduce: ContextVariable
			reduce(250), // >>>, reduce: ContextVariable
			reduce(250), // +, reduce: ContextVariable
			reduce(250), // -, reduce: ContextVariable
			reduce(250), // .+, reduce: ContextVariable
			reduce(250), // .-, reduce: ContextVariable
			reduce(250), // *, reduce: ContextVariable
			reduce(250), // /, reduce: ContextVariable
			reduce(250), // //, reduce: ContextVariable
			reduce(250), // %, reduce: ContextVariable
			reduce(250), // .*, reduce: ContextVariable
			reduce(250), // ./, reduce: ContextVariable
			reduce(250), // .//, reduce: ContextVariable
			reduce(250), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(250), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(250), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(250), // [[, reduce: ContextVariable
			reduce(250), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1067
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(251), // ;, reduce: ContextVariable
			reduce(251), // {, reduce: ContextVariable
			reduce(251), // }, reduce: ContextVariable
			reduce(251), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(251), // >, reduce: ContextVariable
			reduce(251), // >>, reduce: ContextVariable
			reduce(251), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(251), // ||=, reduce: ContextVariable
			reduce(251), // ^^=, reduce: ContextVariable
			reduce(251), // &&=, reduce: ContextVariable
			reduce(251), // ??=, reduce: ContextVariable
			reduce(251), // ???=, reduce: ContextVariable
			reduce(251), // |=, reduce: ContextVariable
			reduce(251), // &=, reduce: ContextVariable
			reduce(251), // ^=, reduce: ContextVariable
			reduce(251), // <<=, reduce: ContextVariable
			reduce(251), // >>=, reduce: ContextVariable
			reduce(251), // >>>=, reduce: ContextVariable
			reduce(251), // +=, reduce: ContextVariable
			reduce(251), // .=, reduce: ContextVariable
			reduce(251), // -=, reduce: ContextVariable
			reduce(251), // *=, reduce: ContextVariable
			reduce(251), // /=, reduce: ContextVariable
			reduce(251), // //=, reduce: ContextVariable
			reduce(251), // %=, reduce: ContextVariable
			reduce(251), // **=, reduce: ContextVariable
			reduce(251), // ?, reduce: ContextVariable
			nil,         // :
			reduce(251), // ||, reduce: ContextVariable
			reduce(251), // ^^, reduce: ContextVariable
			reduce(251), // &&, reduce: ContextVariable
			reduce(251), // ??, reduce: ContextVariable
			reduce(251), // ???, reduce: ContextVariable
			reduce(251), // =~, reduce: ContextVariable
			reduce(251), // !=~, reduce: ContextVariable
			reduce(251), // ==, reduce: ContextVariable
			reduce(251), // !=, reduce: ContextVariable
			reduce(251), // <=>, reduce: ContextVariable
			reduce(251), // >=, reduce: ContextVariable
			reduce(251), // <, reduce: ContextVariable
			reduce(251), // <=, reduce: ContextVariable
			reduce(251), // ^, reduce: ContextVariable
			reduce(251), // &, reduce: ContextVariable
			reduce(251), // <<, reduce: ContextVariable
			reduce(251), // >>>, reduce: ContextVariable
			reduce(251), // +, reduce: ContextVariable
			reduce(251), // -, reduce: ContextVariable
			reduce(251), // .+, reduce: ContextVariable
			reduce(251), // .-, reduce: ContextVariable
			reduce(251), // *, reduce: ContextVariable
			reduce(251), // /, reduce: ContextVariable
			reduce(251), // //, reduce: ContextVariable
			reduce(251), // %, reduce: ContextVariable
			reduce(251), // .*, reduce: ContextVariable
			reduce(251), // ./, reduce: ContextVariable
			reduce(251), // .//, reduce: ContextVariable
			reduce(251), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(251), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(251), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(251), // [[, reduce: ContextVariable
			reduce(251), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1068
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(252), // ;, reduce: ContextVariable
			reduce(252), // {, reduce: ContextVariable
			reduce(252), // }, reduce: ContextVariable
			reduce(252), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(252), // >, reduce: ContextVariable
			reduce(252), // >>, reduce: ContextVariable
			reduce(252), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(252), // ||=, reduce: ContextVariable
			reduce(252), // ^^=, reduce: ContextVariable
			reduce(252), // &&=, reduce: ContextVariable
			reduce(252), // ??=, reduce: ContextVariable
			reduce(252), // ???=, reduce: ContextVariable
			reduce(252), // |=, reduce: ContextVariable
			reduce(252), // &=, reduce: ContextVariable
			reduce(252), // ^=, reduce: ContextVariable
			reduce(252), // <<=, reduce: ContextVariable
			reduce(252), // >>=, reduce: ContextVariable
			reduce(252), // >>>=, reduce: ContextVariable
			reduce(252), // +=, reduce: ContextVariable
			reduce(252), // .=, reduce: ContextVariable
			reduce(252), // -=, reduce: ContextVariable
			reduce(252), // *=, reduce: ContextVariable
			reduce(252), // /=, reduce: ContextVariable
			reduce(252), // //=, reduce: ContextVariable
			reduce(252), // %=, reduce: ContextVariable
			reduce(252), // **=, reduce: ContextVariable
			reduce(252), // ?, reduce: ContextVariable
			nil,         // :
			reduce(252), // ||, reduce: ContextVariable
			reduce(252), // ^^, reduce: ContextVariable
			reduce(252), // &&, reduce: ContextVariable
			reduce(252), // ??, reduce: ContextVariable
			reduce(252), // ???, reduce: ContextVariable
			reduce(252), // =~, reduce: ContextVariable
			reduce(252), // !=~, reduce: ContextVariable
			reduce(252), // ==, reduce: ContextVariable
			reduce(252), // !=, reduce: ContextVariable
			reduce(252), // <=>, reduce: ContextVariable
			reduce(252), // >=, reduce: ContextVariable
			reduce(252), // <, reduce: ContextVariable
			reduce(252), // <=, reduce: ContextVariable
			reduce(252), // ^, reduce: ContextVariable
			reduce(252), // &, reduce: ContextVariable
			reduce(252), // <<, reduce: ContextVariable
			reduce(252), // >>>, reduce: ContextVariable
			reduce(252), // +, reduce: ContextVariable
			reduce(252), // -, reduce: ContextVariable
			reduce(252), // .+, reduce: ContextVariable
			reduce(252), // .-, reduce: ContextVariable
			reduce(252), // *, reduce: ContextVariable
			reduce(252), // /, reduce: ContextVariable
			reduce(252), // //, reduce: ContextVariable
			reduce(252), // %, reduce: ContextVariable
			reduce(252), // .*, reduce: ContextVariable
			reduce(252), // ./, reduce: ContextVariable
			reduce(252), // .//, reduce: ContextVariable
			reduce(252), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(252), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(252), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(252), // [[, reduce: ContextVariable
			reduce(252), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1069
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(253), // ;, reduce: MlrvalOrFunction
			reduce(253), // {, reduce: MlrvalOrFunction
			reduce(253), // }, reduce: MlrvalOrFunction
			reduce(253), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(253), // >, reduce: MlrvalOrFunction
			reduce(253), // >>, reduce: MlrvalOrFunction
			reduce(253), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(253), // ||=, reduce: MlrvalOrFunction
			reduce(253), // ^^=, reduce: MlrvalOrFunction
			reduce(253), // &&=, reduce: MlrvalOrFunction
			reduce(253), // ??=, reduce: MlrvalOrFunction
			reduce(253), // ???=, reduce: MlrvalOrFunction
			reduce(253), // |=, reduce: MlrvalOrFunction
			reduce(253), // &=, reduce: MlrvalOrFunction
			reduce(253), // ^=, reduce: MlrvalOrFunction
			reduce(253), // <<=, reduce: MlrvalOrFunction
			reduce(253), // >>=, reduce: MlrvalOrFunction
			reduce(253), // >>>=, reduce: MlrvalOrFunction
			reduce(253), // +=, reduce: MlrvalOrFunction
			reduce(253), // .=, reduce: MlrvalOrFunction
			reduce(253), // -=, reduce: MlrvalOrFunction
			reduce(253), // *=, reduce: MlrvalOrFunction
			reduce(253), // /=, reduce: MlrvalOrFunction
			reduce(253), // //=, reduce: MlrvalOrFunction
			reduce(253), // %=, reduce: MlrvalOrFunction
			reduce(253), // **=, reduce: MlrvalOrFunction
			reduce(253), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(253), // ||, reduce: MlrvalOrFunction
			reduce(253), // ^^, reduce: MlrvalOrFunction
			reduce(253), // &&, reduce: MlrvalOrFunction
			reduce(253), // ??, reduce: MlrvalOrFunction
			reduce(253), // ???, reduce: MlrvalOrFunction
			reduce(253), // =~, reduce: MlrvalOrFunction
			reduce(253), // !=~, reduce: MlrvalOrFunction
			reduce(253), // ==, reduce: MlrvalOrFunction
			reduce(253), // !=, reduce: MlrvalOrFunction
			reduce(253), // <=>, reduce: MlrvalOrFunction
			reduce(253), // >=, reduce: MlrvalOrFunction
			reduce(253), // <, reduce: MlrvalOrFunction
			reduce(253), // <=, reduce: MlrvalOrFunction
			reduce(253), // ^, reduce: MlrvalOrFunction
			reduce(253), // &, reduce: MlrvalOrFunction
			reduce(253), // <<, reduce: MlrvalOrFunction
			reduce(253), // >>>, reduce: MlrvalOrFunction
			reduce(253), // +, reduce: MlrvalOrFunction
			reduce(253), // -, reduce: MlrvalOrFunction
			reduce(253), // .+, reduce: MlrvalOrFunction
			reduce(253), // .-, reduce: MlrvalOrFunction
			reduce(253), // *, reduce: MlrvalOrFunction
			reduce(253), // /, reduce: MlrvalOrFunction
			reduce(253), // //, reduce: MlrvalOrFunction
			reduce(253), // %, reduce: MlrvalOrFunction
			reduce(253), // .*, reduce: MlrvalOrFunction
			reduce(253), // ./, reduce: MlrvalOrFunction
			reduce(253), // .//, reduce: MlrvalOrFunction
			reduce(253), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(253), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(253), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(253), // [[, reduce: MlrvalOrFunction
			reduce(253), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1070
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			shift(2322), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(2323), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1071
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(256), // ;, reduce: MlrvalOrFunction
			reduce(256), // {, reduce: MlrvalOrFunction
			reduce(256), // }, reduce: MlrvalOrFunction
			reduce(256), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(256), // >, reduce: MlrvalOrFunction
			reduce(256), // >>, reduce: MlrvalOrFunction
			reduce(256), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(256), // ||=, reduce: MlrvalOrFunction
			reduce(256), // ^^=, reduce: MlrvalOrFunction
			reduce(256), // &&=, reduce: MlrvalOrFunction
			reduce(256), // ??=, reduce: MlrvalOrFunction
			reduce(256), // ???=, reduce: MlrvalOrFunction
			reduce(256), // |=, reduce: MlrvalOrFunction
			reduce(256), // &=, reduce: MlrvalOrFunction
			reduce(256), // ^=, reduce: MlrvalOrFunction
			reduce(256), // <<=, reduce: MlrvalOrFunction
			reduce(256), // >>=, reduce: MlrvalOrFunction
			reduce(256), // >>>=, reduce: MlrvalOrFunction
			reduce(256), // +=, reduce: MlrvalOrFunction
			reduce(256), // .=, reduce: MlrvalOrFunction
			reduce(256), // -=, reduce: MlrvalOrFunction
			reduce(256), // *=, reduce: MlrvalOrFunction
			reduce(256), // /=, reduce: MlrvalOrFunction
			reduce(256), // //=, reduce: MlrvalOrFunction
			reduce(256), // %=, reduce: MlrvalOrFunction
			reduce(256), // **=, reduce: MlrvalOrFunction
			reduce(256), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(256), // ||, reduce: MlrvalOrFunction
			reduce(256), // ^^, reduce: MlrvalOrFunction
			reduce(256), // &&, reduce: MlrvalOrFunction
			reduce(256), // ??, reduce: MlrvalOrFunction
			reduce(256), // ???, reduce: MlrvalOrFunction
			reduce(256), // =~, reduce: MlrvalOrFunction
			reduce(256), // !=~, reduce: MlrvalOrFunction
			reduce(256), // ==, reduce: MlrvalOrFunction
			reduce(256), // !=, reduce: MlrvalOrFunction
			reduce(256), // <=>, reduce: MlrvalOrFunction
			reduce(256), // >=, reduce: MlrvalOrFunction
			reduce(256), // <, reduce: MlrvalOrFunction
			reduce(256), // <=, reduce: MlrvalOrFunction
			reduce(256), // ^, reduce: MlrvalOrFunction
			reduce(256), // &, reduce: MlrvalOrFunction
			reduce(256), // <<, reduce: MlrvalOrFunction
			reduce(256), // >>>, reduce: MlrvalOrFunction
			reduce(256), // +, reduce: MlrvalOrFunction
			reduce(256), // -, reduce: MlrvalOrFunction
			reduce(256), // .+, reduce: MlrvalOrFunction
			reduce(256), // .-, reduce: MlrvalOrFunction
			reduce(256), // *, reduce: MlrvalOrFunction
			reduce(256), // /, reduce: MlrvalOrFunction
			reduce(256), // //, reduce: MlrvalOrFunction
			reduce(256), // %, reduce: MlrvalOrFunction
			reduce(256), // .*, reduce: MlrvalOrFunction
			reduce(256), // ./, reduce: MlrvalOrFunction
			reduce(256), // .//, reduce: MlrvalOrFunction
			reduce(256), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(256), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(256), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(256), // [[, reduce: MlrvalOrFunction
			reduce(256), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1072
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(257), // ;, reduce: MlrvalOrFunction
			reduce(257), // {, reduce: MlrvalOrFunction
			reduce(257), // }, reduce: MlrvalOrFunction
			reduce(257), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(257), // >, reduce: MlrvalOrFunction
			reduce(257), // >>, reduce: MlrvalOrFunction
			reduce(257), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(257), // ||=, reduce: MlrvalOrFunction
			reduce(257), // ^^=, reduce: MlrvalOrFunction
			reduce(257), // &&=, reduce: MlrvalOrFunction
			reduce(257), // ??=, reduce: MlrvalOrFunction
			reduce(257), // ???=, reduce: MlrvalOrFunction
			reduce(257), // |=, reduce: MlrvalOrFunction
			reduce(257), // &=, reduce: MlrvalOrFunction
			reduce(257), // ^=, reduce: MlrvalOrFunction
			reduce(257), // <<=, reduce: MlrvalOrFunction
			reduce(257), // >>=, reduce: MlrvalOrFunction
			reduce(257), // >>>=, reduce: MlrvalOrFunction
			reduce(257), // +=, reduce: MlrvalOrFunction
			reduce(257), // .=, reduce: MlrvalOrFunction
			reduce(257), // -=, reduce: MlrvalOrFunction
			reduce(257), // *=, reduce: MlrvalOrFunction
			reduce(257), // /=, reduce: MlrvalOrFunction
			reduce(257), // //=, reduce: MlrvalOrFunction
			reduce(257), // %=, reduce: MlrvalOrFunction
			reduce(257), // **=, reduce: MlrvalOrFunction
			reduce(257), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(257), // ||, reduce: MlrvalOrFunction
			reduce(257), // ^^, reduce: MlrvalOrFunction
			reduce(257), // &&, reduce: MlrvalOrFunction
			reduce(257), // ??, reduce: MlrvalOrFunction
			reduce(257), // ???, reduce: MlrvalOrFunction
			reduce(257), // =~, reduce: MlrvalOrFunction
			reduce(257), // !=~, reduce: MlrvalOrFunction
			reduce(257), // ==, reduce: MlrvalOrFunction
			reduce(257), // !=, reduce: MlrvalOrFunction
			reduce(257), // <=>, reduce: MlrvalOrFunction
			reduce(257), // >=, reduce: MlrvalOrFunction
			reduce(257), // <, reduce: MlrvalOrFunction
			reduce(257), // <=, reduce: MlrvalOrFunction
			reduce(257), // ^, reduce: MlrvalOrFunction
			reduce(257), // &, reduce: MlrvalOrFunction
			reduce(257), // <<, reduce: MlrvalOrFunction
			reduce(257), // >>>, reduce: MlrvalOrFunction
			reduce(257), // +, reduce: MlrvalOrFunction
			reduce(257), // -, reduce: MlrvalOrFunction
			reduce(257), // .+, reduce: MlrvalOrFunction
			reduce(257), // .-, reduce: MlrvalOrFunction
			reduce(257), // *, reduce: MlrvalOrFunction
			reduce(257), // /, reduce: MlrvalOrFunction
			reduce(257), // //, reduce: MlrvalOrFunction
			reduce(257), // %, reduce: MlrvalOrFunction
			reduce(257), // .*, reduce: MlrvalOrFunction
			reduce(257), // ./, reduce: MlrvalOrFunction
			reduce(257), // .//, reduce: MlrvalOrFunction
			reduce(257), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(257), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(257), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(257), // [[, reduce: MlrvalOrFunction
			reduce(257), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1073
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(258), // ;, reduce: MlrvalOrFunction
			reduce(258), // {, reduce: MlrvalOrFunction
			reduce(258), // }, reduce: MlrvalOrFunction
			reduce(258), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(258), // >, reduce: MlrvalOrFunction
			reduce(258), // >>, reduce: MlrvalOrFunction
			reduce(258), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(258), // ||=, reduce: MlrvalOrFunction
			reduce(258), // ^^=, reduce: MlrvalOrFunction
			reduce(258), // &&=, reduce: MlrvalOrFunction
			reduce(258), // ??=, reduce: MlrvalOrFunction
			reduce(258), // ???=, reduce: MlrvalOrFunction
			reduce(258), // |=, reduce: MlrvalOrFunction
			reduce(258), // &=, reduce: MlrvalOrFunction
			reduce(258), // ^=, reduce: MlrvalOrFunction
			reduce(258), // <<=, reduce: MlrvalOrFunction
			reduce(258), // >>=, reduce: MlrvalOrFunction
			reduce(258), // >>>=, reduce: MlrvalOrFunction
			reduce(258), // +=, reduce: MlrvalOrFunction
			reduce(258), // .=, reduce: MlrvalOrFunction
			reduce(258), // -=, reduce: MlrvalOrFunction
			reduce(258), // *=, reduce: MlrvalOrFunction
			reduce(258), // /=, reduce: MlrvalOrFunction
			reduce(258), // //=, reduce: MlrvalOrFunction
			reduce(258), // %=, reduce: MlrvalOrFunction
			reduce(258), // **=, reduce: MlrvalOrFunction
			reduce(258), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(258), // ||, reduce: MlrvalOrFunction
			reduce(258), // ^^, reduce: MlrvalOrFunction
			reduce(258), // &&, reduce: MlrvalOrFunction
			reduce(258), // ??, reduce: MlrvalOrFunction
			reduce(258), // ???, reduce: MlrvalOrFunction
			reduce(258), // =~, reduce: MlrvalOrFunction
			reduce(258), // !=~, reduce: MlrvalOrFunction
			reduce(258), // ==, reduce: MlrvalOrFunction
			reduce(258), // !=, reduce: MlrvalOrFunction
			reduce(258), // <=>, reduce: MlrvalOrFunction
			reduce(258), // >=, reduce: MlrvalOrFunction
			reduce(258), // <, reduce: MlrvalOrFunction
			reduce(258), // <=, reduce: MlrvalOrFunction
			reduce(258), // ^, reduce: MlrvalOrFunction
			reduce(258), // &, reduce: MlrvalOrFunction
			reduce(258), // <<, reduce: MlrvalOrFunction
			reduce(258), // >>>, reduce: MlrvalOrFunction
			reduce(258), // +, reduce: MlrvalOrFunction
			reduce(258), // -, reduce: MlrvalOrFunction
			reduce(258), // .+, reduce: MlrvalOrFunction
			reduce(258), // .-, reduce: MlrvalOrFunction
			reduce(258), // *, reduce: MlrvalOrFunction
			reduce(258), // /, reduce: MlrvalOrFunction
			reduce(258), // //, reduce: MlrvalOrFunction
			reduce(258), // %, reduce: MlrvalOrFunction
			reduce(258), // .*, reduce: MlrvalOrFunction
			reduce(258), // ./, reduce: MlrvalOrFunction
			reduce(258), // .//, reduce: MlrvalOrFunction
			reduce(258), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(258), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(258), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(258), // [[, reduce: MlrvalOrFunction
			reduce(258), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1074
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(259), // ;, reduce: MlrvalOrFunction
			reduce(259), // {, reduce: MlrvalOrFunction
			reduce(259), // }, reduce: MlrvalOrFunction
			reduce(259), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(259), // >, reduce: MlrvalOrFunction
			reduce(259), // >>, reduce: MlrvalOrFunction
			reduce(259), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(259), // ||=, reduce: MlrvalOrFunction
			reduce(259), // ^^=, reduce: MlrvalOrFunction
			reduce(259), // &&=, reduce: MlrvalOrFunction
			reduce(259), // ??=, reduce: MlrvalOrFunction
			reduce(259), // ???=, reduce: MlrvalOrFunction
			reduce(259), // |=, reduce: MlrvalOrFunction
			reduce(259), // &=, reduce: MlrvalOrFunction
			reduce(259), // ^=, reduce: MlrvalOrFunction
			reduce(259), // <<=, reduce: MlrvalOrFunction
			reduce(259), // >>=, reduce: MlrvalOrFunction
			reduce(259), // >>>=, reduce: MlrvalOrFunction
			reduce(259), // +=, reduce: MlrvalOrFunction
			reduce(259), // .=, reduce: MlrvalOrFunction
			reduce(259), // -=, reduce: MlrvalOrFunction
			reduce(259), // *=, reduce: MlrvalOrFunction
			reduce(259), // /=, reduce: MlrvalOrFunction
			reduce(259), // //=, reduce: MlrvalOrFunction
			reduce(259), // %=, reduce: MlrvalOrFunction
			reduce(259), // **=, reduce: MlrvalOrFunction
			reduce(259), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(259), // ||, reduce: MlrvalOrFunction
			reduce(259), // ^^, reduce: MlrvalOrFunction
			reduce(259), // &&, reduce: MlrvalOrFunction
			reduce(259), // ??, reduce: MlrvalOrFunction
			reduce(259), // ???, reduce: MlrvalOrFunction
			reduce(259), // =~, reduce: MlrvalOrFunction
			reduce(259), // !=~, reduce: MlrvalOrFunction
			reduce(259), // ==, reduce: MlrvalOrFunction
			reduce(259), // !=, reduce: MlrvalOrFunction
			reduce(259), // <=>, reduce: MlrvalOrFunction
			reduce(259), // >=, reduce: MlrvalOrFunction
			reduce(259), // <, reduce: MlrvalOrFunction
			reduce(259), // <=, reduce: MlrvalOrFunction
			reduce(259), // ^, reduce: MlrvalOrFunction
			reduce(259), // &, reduce: MlrvalOrFunction
			reduce(259), // <<, reduce: MlrvalOrFunction
			reduce(259), // >>>, reduce: MlrvalOrFunction
			reduce(259), // +, reduce: MlrvalOrFunction
			reduce(259), // -, reduce: MlrvalOrFunction
			reduce(259), // .+, reduce: MlrvalOrFunction
			reduce(259), // .-, reduce: MlrvalOrFunction
			reduce(259), // *, reduce: MlrvalOrFunction
			reduce(259), // /, reduce: MlrvalOrFunction
			reduce(259), // //, reduce: MlrvalOrFunction
			reduce(259), // %, reduce: MlrvalOrFunction
			reduce(259), // .*, reduce: MlrvalOrFunction
			reduce(259), // ./, reduce: MlrvalOrFunction
			reduce(259), // .//, reduce: MlrvalOrFunction
			reduce(259), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(259), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(259), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(259), // [[, reduce: MlrvalOrFunction
			reduce(259), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1075
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(267), // ;, reduce: MlrvalOrFunction
			reduce(267), // {, reduce: MlrvalOrFunction
			reduce(267), // }, reduce: MlrvalOrFunction
			reduce(267), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(267), // >, reduce: MlrvalOrFunction
			reduce(267), // >>, reduce: MlrvalOrFunction
			reduce(267), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(267), // ||=, reduce: MlrvalOrFunction
			reduce(267), // ^^=, reduce: MlrvalOrFunction
			reduce(267), // &&=, reduce: MlrvalOrFunction
			reduce(267), // ??=, reduce: MlrvalOrFunction
			reduce(267), // ???=, reduce: MlrvalOrFunction
			reduce(267), // |=, reduce: MlrvalOrFunction
			reduce(267), // &=, reduce: MlrvalOrFunction
			reduce(267), // ^=, reduce: MlrvalOrFunction
			reduce(267), // <<=, reduce: MlrvalOrFunction
			reduce(267), // >>=, reduce: MlrvalOrFunction
			reduce(267), // >>>=, reduce: MlrvalOrFunction
			reduce(267), // +=, reduce: MlrvalOrFunction
			reduce(267), // .=, reduce: MlrvalOrFunction
			reduce(267), // -=, reduce: MlrvalOrFunction
			reduce(267), // *=, reduce: MlrvalOrFunction
			reduce(267), // /=, reduce: MlrvalOrFunction
			reduce(267), // //=, reduce: MlrvalOrFunction
			reduce(267), // %=, reduce: MlrvalOrFunction
			reduce(267), // **=, reduce: MlrvalOrFunction
			reduce(267), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(267), // ||, reduce: MlrvalOrFunction
			reduce(267), // ^^, reduce: MlrvalOrFunction
			reduce(267), // &&, reduce: MlrvalOrFunction
			reduce(267), // ??, reduce: MlrvalOrFunction
			reduce(267), // ???, reduce: MlrvalOrFunction
			reduce(267), // =~, reduce: MlrvalOrFunction
			reduce(267), // !=~, reduce: MlrvalOrFunction
			reduce(267), // ==, reduce: MlrvalOrFunction
			reduce(267), // !=, reduce: MlrvalOrFunction
			reduce(267), // <=>, reduce: MlrvalOrFunction
			reduce(267), // >=, reduce: MlrvalOrFunction
			reduce(267), // <, reduce: MlrvalOrFunction
			reduce(267), // <=, reduce: MlrvalOrFunction
			reduce(267), // ^, reduce: MlrvalOrFunction
			reduce(267), // &, reduce: MlrvalOrFunction
			reduce(267), // <<, reduce: MlrvalOrFunction
			reduce(267), // >>>, reduce: MlrvalOrFunction
			reduce(267), // +, reduce: MlrvalOrFunction
			reduce(267), // -, reduce: MlrvalOrFunction
			reduce(267), // .+, reduce: MlrvalOrFunction
			reduce(267), // .-, reduce: MlrvalOrFunction
			reduce(267), // *, reduce: MlrvalOrFunction
			reduce(267), // /, reduce: MlrvalOrFunction
			reduce(267), // //, reduce: MlrvalOrFunction
			reduce(267), // %, reduce: MlrvalOrFunction
			reduce(267), // .*, reduce: MlrvalOrFunction
			reduce(267), // ./, reduce: MlrvalOrFunction
			reduce(267), // .//, reduce: MlrvalOrFunction
			reduce(267), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(267), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(267), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(267), // [[, reduce: MlrvalOrFunction
			reduce(267), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1076
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2324), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1077
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			shift(848), // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1078
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(279), // ;, reduce: BracefulStatement
			reduce(279), // {, reduce: BracefulStatement
			reduce(279), // }, reduce: BracefulStatement
			nil,         // =
			reduce(279), // unset, reduce: BracefulStatement
			reduce(279), // filter, reduce: BracefulStatement
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(279), // print, reduce: BracefulStatement
			nil,         // ,
			reduce(279), // printn, reduce: BracefulStatement
			reduce(279), // eprint, reduce: BracefulStatement
			reduce(279), // eprintn, reduce: BracefulStatement
			reduce(279), // dump, reduce: BracefulStatement
			reduce(279), // edump, reduce: BracefulStatement
			reduce(279), // tee, reduce: BracefulStatement
			reduce(279), // emitf, reduce: BracefulStatement
			reduce(279), // emit1, reduce: BracefulStatement
			reduce(279), // emit, reduce: BracefulStatement
			reduce(279), // (, reduce: BracefulStatement
			nil,         // )
			reduce(279), // emitp, reduce: BracefulStatement
			reduce(279), // field_name, reduce: BracefulStatement
			reduce(279), // $[, reduce: BracefulStatement
			nil,         // ]
			reduce(279), // braced_field_name, reduce: BracefulStatement
			reduce(279), // $[[, reduce: BracefulStatement
			reduce(279), // $[[[, reduce: BracefulStatement
			reduce(279), // full_srec, reduce: BracefulStatement
			reduce(279), // oosvar_name, reduce: BracefulStatement
			reduce(279), // @[, reduce: BracefulStatement
			reduce(279), // braced_oosvar_name, reduce: BracefulStatement
			reduce(279), // full_oosvar, reduce: BracefulStatement
			reduce(279), // all, reduce: BracefulStatement
			reduce(279), // non_sigil_name, reduce: BracefulStatement
			reduce(279), // arr, reduce: BracefulStatement
			reduce(279), // bool, reduce: BracefulStatement
			reduce(279), // float, reduce: BracefulStatement
			reduce(279), // int, reduce: BracefulStatement
			reduce(279), // map, reduce: BracefulStatement
			reduce(279), // num, reduce: BracefulStatement
			reduce(279), // str, reduce: BracefulStatement
			reduce(279), // var, reduce: BracefulStatement
			reduce(279), // funct, reduce: BracefulStatement
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(279), // +, reduce: BracefulStatement
			reduce(279), // -, reduce: BracefulStatement
			reduce(279), // .+, reduce: BracefulStatement
			reduce(279), // .-, reduce: BracefulStatement
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(279), // !, reduce: BracefulStatement
			reduce(279), // ~, reduce: BracefulStatement
			nil,         // **
			reduce(279), // string_literal, reduce: BracefulStatement
			reduce(279), // regex_case_insensitive, reduce: BracefulStatement
			reduce(279), // int_literal, reduce: BracefulStatement
			reduce(279), // float_literal, reduce: BracefulStatement
			reduce(279), // boolean_literal, reduce: BracefulStatement
			reduce(279), // null_literal, reduce: BracefulStatement
			reduce(279), // inf_literal, reduce: BracefulStatement
			reduce(279), // nan_literal, reduce: BracefulStatement
			reduce(279), // const_M_PI, reduce: BracefulStatement
			reduce(279), // const_M_E, reduce: BracefulStatement
			reduce(279), // panic, reduce: BracefulStatement
			reduce(279), // [, reduce: BracefulStatement
			reduce(279), // ctx_IPS, reduce: BracefulStatement
			reduce(279), // ctx_IFS, reduce: BracefulStatement
			reduce(279), // ctx_IRS, reduce: BracefulStatement
			reduce(279), // ctx_OPS, reduce: BracefulStatement
			reduce(279), // ctx_OFS, reduce: BracefulStatement
			reduce(279), // ctx_ORS, reduce: BracefulStatement
			reduce(279), // ctx_FLATSEP, reduce: BracefulStatement
			reduce(279), // ctx_NF, reduce: BracefulStatement
			reduce(279), // ctx_NR, reduce: BracefulStatement
			reduce(279), // ctx_FNR, reduce: BracefulStatement
			reduce(279), // ctx_FILENAME, reduce: BracefulStatement
			reduce(279), // ctx_FILENUM, reduce: BracefulStatement
			reduce(279), // env, reduce: BracefulStatement
			nil,         // [[
			nil,         // [[[
			reduce(279), // call, reduce: BracefulStatement
			reduce(279), // begin, reduce: BracefulStatement
			reduce(279), // end, reduce: BracefulStatement
			reduce(279), // if, reduce: BracefulStatement
			nil,         // elif
			nil,         // else
			reduce(279), // while, reduce: BracefulStatement
			reduce(279), // do, reduce: BracefulStatement
			reduce(279), // for, reduce: BracefulStatement
			nil,         // in
			reduce(279), // break, reduce: BracefulStatement
			reduce(279), // continue, reduce: BracefulStatement
			reduce(279), // func, reduce: BracefulStatement
			reduce(279), // subr, reduce: BracefulStatement
			reduce(279), // return, reduce: BracefulStatement
		},
	},
	actionRow{ // S1079
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(280), // ;, reduce: BracefulStatement
			reduce(280), // {, reduce: BracefulStatement
			reduce(280), // }, reduce: BracefulStatement
			nil,         // =
			reduce(280), // unset, reduce: BracefulStatement
			reduce(280), // filter, reduce: BracefulStatement
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(280), // print, reduce: BracefulStatement
			nil,         // ,
			reduce(280), // printn, reduce: BracefulStatement
			reduce(280), // eprint, reduce: BracefulStatement
			reduce(280), // eprintn, reduce: BracefulStatement
			reduce(280), // dump, reduce: BracefulStatement
			reduce(280), // edump, reduce: BracefulStatement
			reduce(280), // tee, reduce: BracefulStatement
			reduce(280), // emitf, reduce: BracefulStatement
			reduce(280), // emit1, reduce: BracefulStatement
			reduce(280), // emit, reduce: BracefulStatement
			reduce(280), // (, reduce: BracefulStatement
			nil,         // )
			reduce(280), // emitp, reduce: BracefulStatement
			reduce(280), // field_name, reduce: BracefulStatement
			reduce(280), // $[, reduce: BracefulStatement
			nil,         // ]
			reduce(280), // braced_field_name, reduce: BracefulStatement
			reduce(280), // $[[, reduce: BracefulStatement
			reduce(280), // $[[[, reduce: BracefulStatement
			reduce(280), // full_srec, reduce: BracefulStatement
			reduce(280), // oosvar_name, reduce: BracefulStatement
			reduce(280), // @[, reduce: BracefulStatement
			reduce(280), // braced_oosvar_name, reduce: BracefulStatement
			reduce(280), // full_oosvar, reduce: BracefulStatement
			reduce(280), // all, reduce: BracefulStatement
			reduce(280), // non_sigil_name, reduce: BracefulStatement
			reduce(280), // arr, reduce: BracefulStatement
			reduce(280), // bool, reduce: BracefulStatement
			reduce(280), // float, reduce: BracefulStatement
			reduce(280), // int, reduce: BracefulStatement
			reduce(280), // map, reduce: BracefulStatement
			reduce(280), // num, reduce: BracefulStatement
			reduce(280), // str, reduce: BracefulStatement
			reduce(280), // var, reduce: BracefulStatement
			reduce(280), // funct, reduce: BracefulStatement
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(280), // +, reduce: BracefulStatement
			reduce(280), // -, reduce: BracefulStatement
			reduce(280), // .+, reduce: BracefulStatement
			reduce(280), // .-, reduce: BracefulStatement
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(280), // !, reduce: BracefulStatement
			reduce(280), // ~, reduce: BracefulStatement
			nil,         // **
			reduce(280), // string_literal, reduce: BracefulStatement
			reduce(280), // regex_case_insensitive, reduce: BracefulStatement
			reduce(280), // int_literal, reduce: BracefulStatement
			reduce(280), // float_literal, reduce: BracefulStatement
			reduce(280), // boolean_literal, reduce: BracefulStatement
			reduce(280), // null_literal, reduce: BracefulStatement
			reduce(280), // inf_literal, reduce: BracefulStatement
			reduce(280), // nan_literal, reduce: BracefulStatement
			reduce(280), // const_M_PI, reduce: BracefulStatement
			reduce(280), // const_M_E, reduce: BracefulStatement
			reduce(280), // panic, reduce: BracefulStatement
			reduce(280), // [, reduce: BracefulStatement
			reduce(280), // ctx_IPS, reduce: BracefulStatement
			reduce(280), // ctx_IFS, reduce: BracefulStatement
			reduce(280), // ctx_IRS, reduce: BracefulStatement
			reduce(280), // ctx_OPS, reduce: BracefulStatement
			reduce(280), // ctx_OFS, reduce: BracefulStatement
			reduce(280), // ctx_ORS, reduce: BracefulStatement
			reduce(280), // ctx_FLATSEP, reduce: BracefulStatement
			reduce(280), // ctx_NF, reduce: BracefulStatement
			reduce(280), // ctx_NR, reduce: BracefulStatement
			reduce(280), // ctx_FNR, reduce: BracefulStatement
			reduce(280), // ctx_FILENAME, reduce: BracefulStatement
			reduce(280), // ctx_FILENUM, reduce: BracefulStatement
			reduce(280), // env, reduce: BracefulStatement
			nil,         // [[
			nil,         // [[[
			reduce(280), // call, reduce: BracefulStatement
			reduce(280), // begin, reduce: BracefulStatement
			reduce(280), // end, reduce: BracefulStatement
			reduce(280), // if, reduce: BracefulStatement
			nil,         // elif
			nil,         // else
			reduce(280), // while, reduce: BracefulStatement
			reduce(280), // do, reduce: BracefulStatement
			reduce(280), // for, reduce: BracefulStatement
			nil,         // in
			reduce(280), // break, reduce: BracefulStatement
			reduce(280), // continue, reduce: BracefulStatement
			reduce(280), // func, reduce: BracefulStatement
			reduce(280), // subr, reduce: BracefulStatement
			reduce(280), // return, reduce: BracefulStatement
		},
	},
	actionRow{ // S1080
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(281), // ;, reduce: BracefulStatement
			reduce(281), // {, reduce: BracefulStatement
			reduce(281), // }, reduce: BracefulStatement
			nil,         // =
			reduce(281), // unset, reduce: BracefulStatement
			reduce(281), // filter, reduce: BracefulStatement
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(281), // print, reduce: BracefulStatement
			nil,         // ,
			reduce(281), // printn, reduce: BracefulStatement
			reduce(281), // eprint, reduce: BracefulStatement
			reduce(281), // eprintn, reduce: BracefulStatement
			reduce(281), // dump, reduce: BracefulStatement
			reduce(281), // edump, reduce: BracefulStatement
			reduce(281), // tee, reduce: BracefulStatement
			reduce(281), // emitf, reduce: BracefulStatement
			reduce(281), // emit1, reduce: BracefulStatement
			reduce(281), // emit, reduce: BracefulStatement
			reduce(281), // (, reduce: BracefulStatement
			nil,         // )
			reduce(281), // emitp, reduce: BracefulStatement
			reduce(281), // field_name, reduce: BracefulStatement
			reduce(281), // $[, reduce: BracefulStatement
			nil,         // ]
			reduce(281), // braced_field_name, reduce: BracefulStatement
			reduce(281), // $[[, reduce: BracefulStatement
			reduce(281), // $[[[, reduce: BracefulStatement
			reduce(281), // full_srec, reduce: BracefulStatement
			reduce(281), // oosvar_name, reduce: BracefulStatement
			reduce(281), // @[, reduce: BracefulStatement
			reduce(281), // braced_oosvar_name, reduce: BracefulStatement
			reduce(281), // full_oosvar, reduce: BracefulStatement
			reduce(281), // all, reduce: BracefulStatement
			reduce(281), // non_sigil_name, reduce: BracefulStatement
			reduce(281), // arr, reduce: BracefulStatement
			reduce(281), // bool, reduce: BracefulStatement
			reduce(281), // float, reduce: BracefulStatement
			reduce(281), // int, reduce: BracefulStatement
			reduce(281), // map, reduce: BracefulStatement
			reduce(281), // num, reduce: BracefulStatement
			reduce(281), // str, reduce: BracefulStatement
			reduce(281), // var, reduce: BracefulStatement
			reduce(281), // funct, reduce: BracefulStatement
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(281), // +, reduce: BracefulStatement
			reduce(281), // -, reduce: BracefulStatement
			reduce(281), // .+, reduce: BracefulStatement
			reduce(281), // .-, reduce: BracefulStatement
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(281), // !, reduce: BracefulStatement
			reduce(281), // ~, reduce: BracefulStatement
			nil,         // **
			reduce(281), // string_literal, reduce: BracefulStatement
			reduce(281), // regex_case_insensitive, reduce: BracefulStatement
			reduce(281), // int_literal, reduce: BracefulStatement
			reduce(281), // float_literal, reduce: BracefulStatement
			reduce(281), // boolean_literal, reduce: BracefulStatement
			reduce(281), // null_literal, reduce: BracefulStatement
			reduce(281), // inf_literal, reduce: BracefulStatement
			reduce(281), // nan_literal, reduce: BracefulStatement
			reduce(281), // const_M_PI, reduce: BracefulStatement
			reduce(281), // const_M_E, reduce: BracefulStatement
			reduce(281), // panic, reduce: BracefulStatement
			reduce(281), // [, reduce: BracefulStatement
			reduce(281), // ctx_IPS, reduce: BracefulStatement
			reduce(281), // ctx_IFS, reduce: BracefulStatement
			reduce(281), // ctx_IRS, reduce: BracefulStatement
			reduce(281), // ctx_OPS, reduce: BracefulStatement
			reduce(281), // ctx_OFS, reduce: BracefulStatement
			reduce(281), // ctx_ORS, reduce: BracefulStatement
			reduce(281), // ctx_FLATSEP, reduce: BracefulStatement
			reduce(281), // ctx_NF, reduce: BracefulStatement
			reduce(281), // ctx_NR, reduce: BracefulStatement
			reduce(281), // ctx_FNR, reduce: BracefulStatement
			reduce(281), // ctx_FILENAME, reduce: BracefulStatement
			reduce(281), // ctx_FILENUM, reduce: BracefulStatement
			reduce(281), // env, reduce: BracefulStatement
			nil,         // [[
			nil,         // [[[
			reduce(281), // call, reduce: BracefulStatement
			reduce(281), // begin, reduce: BracefulStatement
			reduce(281), // end, reduce: BracefulStatement
			reduce(281), // if, reduce: BracefulStatement
			nil,         // elif
			nil,         // else
			reduce(281), // while, reduce: BracefulStatement
			reduce(281), // do, reduce: BracefulStatement
			reduce(281), // for, reduce: BracefulStatement
			nil,         // in
			reduce(281), // break, reduce: BracefulStatement
			reduce(281), // continue, reduce: BracefulStatement
			reduce(281), // func, reduce: BracefulStatement
			reduce(281), // subr, reduce: BracefulStatement
			reduce(281), // return, reduce: BracefulStatement
		},
	},
	actionRow{ // S1081
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(282), // ;, reduce: BracefulStatement
			reduce(282), // {, reduce: BracefulStatement
			reduce(282), // }, reduce: BracefulStatement
			nil,         // =
			reduce(282), // unset, reduce: BracefulStatement
			reduce(282), // filter, reduce: BracefulStatement
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(282), // print, reduce: BracefulStatement
			nil,         // ,
			reduce(282), // printn, reduce: BracefulStatement
			reduce(282), // eprint, reduce: BracefulStatement
			reduce(282), // eprintn, reduce: BracefulStatement
			reduce(282), // dump, reduce: BracefulStatement
			reduce(282), // edump, reduce: BracefulStatement
			reduce(282), // tee, reduce: BracefulStatement
			reduce(282), // emitf, reduce: BracefulStatement
			reduce(282), // emit1, reduce: BracefulStatement
			reduce(282), // emit, reduce: BracefulStatement
			reduce(282), // (, reduce: BracefulStatement
			nil,         // )
			reduce(282), // emitp, reduce: BracefulStatement
			reduce(282), // field_name, reduce: BracefulStatement
			reduce(282), // $[, reduce: BracefulStatement
			nil,         // ]
			reduce(282), // braced_field_name, reduce: BracefulStatement
			reduce(282), // $[[, reduce: BracefulStatement
			reduce(282), // $[[[, reduce: BracefulStatement
			reduce(282), // full_srec, reduce: BracefulStatement
			reduce(282), // oosvar_name, reduce: BracefulStatement
			reduce(282), // @[, reduce: BracefulStatement
			reduce(282), // braced_oosvar_name, reduce: BracefulStatement
			reduce(282), // full_oosvar, reduce: BracefulStatement
			reduce(282), // all, reduce: BracefulStatement
			reduce(282), // non_sigil_name, reduce: BracefulStatement
			reduce(282), // arr, reduce: BracefulStatement
			reduce(282), // bool, reduce: BracefulStatement
			reduce(282), // float, reduce: BracefulStatement
			reduce(282), // int, reduce: BracefulStatement
			reduce(282), // map, reduce: BracefulStatement
			reduce(282), // num, reduce: BracefulStatement
			reduce(282), // str, reduce: BracefulStatement
			reduce(282), // var, reduce: BracefulStatement
			reduce(282), // funct, reduce: BracefulStatement
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(282), // +, reduce: BracefulStatement
			reduce(282), // -, reduce: BracefulStatement
			reduce(282), // .+, reduce: BracefulStatement
			reduce(282), // .-, reduce: BracefulStatement
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(282), // !, reduce: BracefulStatement
			reduce(282), // ~, reduce: BracefulStatement
			nil,         // **
			reduce(282), // string_literal, reduce: BracefulStatement
			reduce(282), // regex_case_insensitive, reduce: BracefulStatement
			reduce(282), // int_literal, reduce: BracefulStatement
			reduce(282), // float_literal, reduce: BracefulStatement
			reduce(282), // boolean_literal, reduce: BracefulStatement
			reduce(282), // null_literal, reduce: BracefulStatement
			reduce(282), // inf_literal, reduce: BracefulStatement
			reduce(282), // nan_literal, reduce: BracefulStatement
			reduce(282), // const_M_PI, reduce: BracefulStatement
			reduce(282), // const_M_E, reduce: BracefulStatement
			reduce(282), // panic, reduce: BracefulStatement
			reduce(282), // [, reduce: BracefulStatement
			reduce(282), // ctx_IPS, reduce: BracefulStatement
			reduce(282), // ctx_IFS, reduce: BracefulStatement
			reduce(282), // ctx_IRS, reduce: BracefulStatement
			reduce(282), // ctx_OPS, reduce: BracefulStatement
			reduce(282), // ctx_OFS, reduce: BracefulStatement
			reduce(282), // ctx_ORS, reduce: BracefulStatement
			reduce(282), // ctx_FLATSEP, reduce: BracefulStatement
			reduce(282), // ctx_NF, reduce: BracefulStatement
			reduce(282), // ctx_NR, reduce: BracefulStatement
			reduce(282), // ctx_FNR, reduce: BracefulStatement
			reduce(282), // ctx_FILENAME, reduce: BracefulStatement
			reduce(282), // ctx_FILENUM, reduce: BracefulStatement
			reduce(282), // env, reduce: BracefulStatement
			nil,         // [[
			nil,         // [[[
			reduce(282), // call, reduce: BracefulStatement
			reduce(282), // begin, reduce: BracefulStatement
			reduce(282), // end, reduce: BracefulStatement
			reduce(282), // if, reduce: BracefulStatement
			nil,         // elif
			nil,         // else
			reduce(282), // while, reduce: BracefulStatement
			reduce(282), // do, reduce: BracefulStatement
			reduce(282), // for, reduce: BracefulStatement
			nil,         // in
			reduce(282), // break, reduce: BracefulStatement
			reduce(282), // continue, reduce: BracefulStatement
			reduce(282), // func, reduce: BracefulStatement
			reduce(282), // subr, reduce: BracefulStatement
			reduce(282), // return, reduce: BracefulStatement
		},
	},
	actionRow{ // S1082
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(283), // ;, reduce: BracefulStatement
			reduce(283), // {, reduce: BracefulStatement
			reduce(283), // }, reduce: BracefulStatement
			nil,         // =
			reduce(283), // unset, reduce: BracefulStatement
			reduce(283), // filter, reduce: BracefulStatement
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(283), // print, reduce: BracefulStatement
			nil,         // ,
			reduce(283), // printn, reduce: BracefulStatement
			reduce(283), // eprint, reduce: BracefulStatement
			reduce(283), // eprintn, reduce: BracefulStatement
			reduce(283), // dump, reduce: BracefulStatement
			reduce(283), // edump, reduce: BracefulStatement
			reduce(283), // tee, reduce: BracefulStatement
			reduce(283), // emitf, reduce: BracefulStatement
			reduce(283), // emit1, reduce: BracefulStatement
			reduce(283), // emit, reduce: BracefulStatement
			reduce(283), // (, reduce: BracefulStatement
			nil,         // )
			reduce(283), // emitp, reduce: BracefulStatement
			reduce(283), // field_name, reduce: BracefulStatement
			reduce(283), // $[, reduce: BracefulStatement
			nil,         // ]
			reduce(283), // braced_field_name, reduce: BracefulStatement
			reduce(283), // $[[, reduce: BracefulStatement
			reduce(283), // $[[[, reduce: BracefulStatement
			reduce(283), // full_srec, reduce: BracefulStatement
			reduce(283), // oosvar_name, reduce: BracefulStatement
			reduce(283), // @[, reduce: BracefulStatement
			reduce(283), // braced_oosvar_name, reduce: BracefulStatement
			reduce(283), // full_oosvar, reduce: BracefulStatement
			reduce(283), // all, reduce: BracefulStatement
			reduce(283), // non_sigil_name, reduce: BracefulStatement
			reduce(283), // arr, reduce: BracefulStatement
			reduce(283), // bool, reduce: BracefulStatement
			reduce(283), // float, reduce: BracefulStatement
			reduce(283), // int, reduce: BracefulStatement
			reduce(283), // map, reduce: BracefulStatement
			reduce(283), // num, reduce: BracefulStatement
			reduce(283), // str, reduce: BracefulStatement
			reduce(283), // var, reduce: BracefulStatement
			reduce(283), // funct, reduce: BracefulStatement
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(283), // +, reduce: BracefulStatement
			reduce(283), // -, reduce: BracefulStatement
			reduce(283), // .+, reduce: BracefulStatement
			reduce(283), // .-, reduce: BracefulStatement
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(283), // !, reduce: BracefulStatement
			reduce(283), // ~, reduce: BracefulStatement
			nil,         // **
			reduce(283), // string_literal, reduce: BracefulStatement
			reduce(283), // regex_case_insensitive, reduce: BracefulStatement
			reduce(283), // int_literal, reduce: BracefulStatement
			reduce(283), // float_literal, reduce: BracefulStatement
			reduce(283), // boolean_literal, reduce: BracefulStatement
			reduce(283), // null_literal, reduce: BracefulStatement
			reduce(283), // inf_literal, reduce: BracefulStatement
			reduce(283), // nan_literal, reduce: BracefulStatement
			reduce(283), // const_M_PI, reduce: BracefulStatement
			reduce(283), // const_M_E, reduce: BracefulStatement
			reduce(283), // panic, reduce: BracefulStatement
			reduce(283), // [, reduce: BracefulStatement
			reduce(283), // ctx_IPS, reduce: BracefulStatement
			reduce(283), // ctx_IFS, reduce: BracefulStatement
			reduce(283), // ctx_IRS, reduce: BracefulStatement
			reduce(283), // ctx_OPS, reduce: BracefulStatement
			reduce(283), // ctx_OFS, reduce: BracefulStatement
			reduce(283), // ctx_ORS, reduce: BracefulStatement
			reduce(283), // ctx_FLATSEP, reduce: BracefulStatement
			reduce(283), // ctx_NF, reduce: BracefulStatement
			reduce(283), // ctx_NR, reduce: BracefulStatement
			reduce(283), // ctx_FNR, reduce: BracefulStatement
			reduce(283), // ctx_FILENAME, reduce: BracefulStatement
			reduce(283), // ctx_FILENUM, reduce: BracefulStatement
			reduce(283), // env, reduce: BracefulStatement
			nil,         // [[
			nil,         // [[[
			reduce(283), // call, reduce: BracefulStatement
			reduce(283), // begin, reduce: BracefulStatement
			reduce(283), // end, reduce: BracefulStatement
			reduce(283), // if, reduce: BracefulStatement
			nil,         // elif
			nil,         // else
			reduce(283), // while, reduce: BracefulStatement
			reduce(283), // do, reduce: BracefulStatement
			reduce(283), // for, reduce: BracefulStatement
			nil,         // in
			reduce(283), // break, reduce: BracefulStatement
			reduce(283), // continue, reduce: BracefulStatement
			reduce(283), // func, reduce: BracefulStatement
			reduce(283), // subr, reduce: BracefulStatement
			reduce(283), // return, reduce: BracefulStatement
		},
	},
	actionRow{ // S1083
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(284), // ;, reduce: BracefulStatement
			reduce(284), // {, reduce: BracefulStatement
			reduce(284), // }, reduce: BracefulStatement
			nil,         // =
			reduce(284), // unset, reduce: BracefulStatement
			reduce(284), // filter, reduce: BracefulStatement
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(284), // print, reduce: BracefulStatement
			nil,         // ,
			reduce(284), // printn, reduce: BracefulStatement
			reduce(284), // eprint, reduce: BracefulStatement
			reduce(284), // eprintn, reduce: BracefulStatement
			reduce(284), // dump, reduce: BracefulStatement
			reduce(284), // edump, reduce: BracefulStatement
			reduce(284), // tee, reduce: BracefulStatement
			reduce(284), // emitf, reduce: BracefulStatement
			reduce(284), // emit1, reduce: BracefulStatement
			reduce(284), // emit, reduce: BracefulStatement
			reduce(284), // (, reduce: BracefulStatement
			nil,         // )
			reduce(284), // emitp, reduce: BracefulStatement
			reduce(284), // field_name, reduce: BracefulStatement
			reduce(284), // $[, reduce: BracefulStatement
			nil,         // ]
			reduce(284), // braced_field_name, reduce: BracefulStatement
			reduce(284), // $[[, reduce: BracefulStatement
			reduce(284), // $[[[, reduce: BracefulStatement
			reduce(284), // full_srec, reduce: BracefulStatement
			reduce(284), // oosvar_name, reduce: BracefulStatement
			reduce(284), // @[, reduce: BracefulStatement
			reduce(284), // braced_oosvar_name, reduce: BracefulStatement
			reduce(284), // full_oosvar, reduce: BracefulStatement
			reduce(284), // all, reduce: BracefulStatement
			reduce(284), // non_sigil_name, reduce: BracefulStatement
			reduce(284), // arr, reduce: BracefulStatement
			reduce(284), // bool, reduce: BracefulStatement
			reduce(284), // float, reduce: BracefulStatement
			reduce(284), // int, reduce: BracefulStatement
			reduce(284), // map, reduce: BracefulStatement
			reduce(284), // num, reduce: BracefulStatement
			reduce(284), // str, reduce: BracefulStatement
			reduce(284), // var, reduce: BracefulStatement
			reduce(284), // funct, reduce: BracefulStatement
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(284), // +, reduce: BracefulStatement
			reduce(284), // -, reduce: BracefulStatement
			reduce(284), // .+, reduce: BracefulStatement
			reduce(284), // .-, reduce: BracefulStatement
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(284), // !, reduce: BracefulStatement
			reduce(284), // ~, reduce: BracefulStatement
			nil,         // **
			reduce(284), // string_literal, reduce: BracefulStatement
			reduce(284), // regex_case_insensitive, reduce: BracefulStatement
			reduce(284), // int_literal, reduce: BracefulStatement
			reduce(284), // float_literal, reduce: BracefulStatement
			reduce(284), // boolean_literal, reduce: BracefulStatement
			reduce(284), // null_literal, reduce: BracefulStatement
			reduce(284), // inf_literal, reduce: BracefulStatement
			reduce(284), // nan_literal, reduce: BracefulStatement
			reduce(284), // const_M_PI, reduce: BracefulStatement
			reduce(284), // const_M_E, reduce: BracefulStatement
			reduce(284), // panic, reduce: BracefulStatement
			reduce(284), // [, reduce: BracefulStatement
			reduce(284), // ctx_IPS, reduce: BracefulStatement
			reduce(284), // ctx_IFS, reduce: BracefulStatement
			reduce(284), // ctx_IRS, reduce: BracefulStatement
			reduce(284), // ctx_OPS, reduce: BracefulStatement
			reduce(284), // ctx_OFS, reduce: BracefulStatement
			reduce(284), // ctx_ORS, reduce: BracefulStatement
			reduce(284), // ctx_FLATSEP, reduce: BracefulStatement
			reduce(284), // ctx_NF, reduce: BracefulStatement
			reduce(284), // ctx_NR, reduce: BracefulStatement
			reduce(284), // ctx_FNR, reduce: BracefulStatement
			reduce(284), // ctx_FILENAME, reduce: BracefulStatement
			reduce(284), // ctx_FILENUM, reduce: BracefulStatement
			reduce(284), // env, reduce: BracefulStatement
			nil,         // [[
			nil,         // [[[
			reduce(284), // call, reduce: BracefulStatement
			reduce(284), // begin, reduce: BracefulStatement
			reduce(284), // end, reduce: BracefulStatement
			reduce(284), // if, reduce: BracefulStatement
			nil,         // elif
			nil,         // else
			reduce(284), // while, reduce: BracefulStatement
			reduce(284), // do, reduce: BracefulStatement
			reduce(284), // for, reduce: BracefulStatement
			nil,         // in
			reduce(284), // break, reduce: BracefulStatement
			reduce(284), // continue, reduce: BracefulStatement
			reduce(284), // func, reduce: BracefulStatement
			reduce(284), // subr, reduce: BracefulStatement
			reduce(284), // return, reduce: BracefulStatement
		},
	},
	actionRow{ // S1084
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(285), // ;, reduce: BracefulStatement
			reduce(285), // {, reduce: BracefulStatement
			reduce(285), // }, reduce: BracefulStatement
			nil,         // =
			reduce(285), // unset, reduce: BracefulStatement
			reduce(285), // filter, reduce: BracefulStatement
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(285), // print, reduce: BracefulStatement
			nil,         // ,
			reduce(285), // printn, reduce: BracefulStatement
			reduce(285), // eprint, reduce: BracefulStatement
			reduce(285), // eprintn, reduce: BracefulStatement
			reduce(285), // dump, reduce: BracefulStatement
			reduce(285), // edump, reduce: BracefulStatement
			reduce(285), // tee, reduce: BracefulStatement
			reduce(285), // emitf, reduce: BracefulStatement
			reduce(285), // emit1, reduce: BracefulStatement
			reduce(285), // emit, reduce: BracefulStatement
			reduce(285), // (, reduce: BracefulStatement
			nil,         // )
			reduce(285), // emitp, reduce: BracefulStatement
			reduce(285), // field_name, reduce: BracefulStatement
			reduce(285), // $[, reduce: BracefulStatement
			nil,         // ]
			reduce(285), // braced_field_name, reduce: BracefulStatement
			reduce(285), // $[[, reduce: BracefulStatement
			reduce(285), // $[[[, reduce: BracefulStatement
			reduce(285), // full_srec, reduce: BracefulStatement
			reduce(285), // oosvar_name, reduce: BracefulStatement
			reduce(285), // @[, reduce: BracefulStatement
			reduce(285), // braced_oosvar_name, reduce: BracefulStatement
			reduce(285), // full_oosvar, reduce: BracefulStatement
			reduce(285), // all, reduce: BracefulStatement
			reduce(285), // non_sigil_name, reduce: BracefulStatement
			reduce(285), // arr, reduce: BracefulStatement
			reduce(285), // bool, reduce: BracefulStatement
			reduce(285), // float, reduce: BracefulStatement
			reduce(285), // int, reduce: BracefulStatement
			reduce(285), // map, reduce: BracefulStatement
			reduce(285), // num, reduce: BracefulStatement
			reduce(285), // str, reduce: BracefulStatement
			reduce(285), // var, reduce: BracefulStatement
			reduce(285), // funct, reduce: BracefulStatement
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(285), // +, reduce: BracefulStatement
			reduce(285), // -, reduce: BracefulStatement
			reduce(285), // .+, reduce: BracefulStatement
			reduce(285), // .-, reduce: BracefulStatement
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(285), // !, reduce: BracefulStatement
			reduce(285), // ~, reduce: BracefulStatement
			nil,         // **
			reduce(285), // string_literal, reduce: BracefulStatement
			reduce(285), // regex_case_insensitive, reduce: BracefulStatement
			reduce(285), // int_literal, reduce: BracefulStatement
			reduce(285), // float_literal, reduce: BracefulStatement
			reduce(285), // boolean_literal, reduce: BracefulStatement
			reduce(285), // null_literal, reduce: BracefulStatement
			reduce(285), // inf_literal, reduce: BracefulStatement
			reduce(285), // nan_literal, reduce: BracefulStatement
			reduce(285), // const_M_PI, reduce: BracefulStatement
			reduce(285), // const_M_E, reduce: BracefulStatement
			reduce(285), // panic, reduce: BracefulStatement
			reduce(285), // [, reduce: BracefulStatement
			reduce(285), // ctx_IPS, reduce: BracefulStatement
			reduce(285), // ctx_IFS, reduce: BracefulStatement
			reduce(285), // ctx_IRS, reduce: BracefulStatement
			reduce(285), // ctx_OPS, reduce: BracefulStatement
			reduce(285), // ctx_OFS, reduce: BracefulStatement
			reduce(285), // ctx_ORS, reduce: BracefulStatement
			reduce(285), // ctx_FLATSEP, reduce: BracefulStatement
			reduce(285), // ctx_NF, reduce: BracefulStatement
			reduce(285), // ctx_NR, reduce: BracefulStatement
			reduce(285), // ctx_FNR, reduce: BracefulStatement
			reduce(285), // ctx_FILENAME, reduce: BracefulStatement
			reduce(285), // ctx_FILENUM, reduce: BracefulStatement
			reduce(285), // env, reduce: BracefulStatement
			nil,         // [[
			nil,         // [[[
			reduce(285), // call, reduce: BracefulStatement
			reduce(285), // begin, reduce: BracefulStatement
			reduce(285), // end, reduce: BracefulStatement
			reduce(285), // if, reduce: BracefulStatement
			nil,         // elif
			nil,         // else
			reduce(285), // while, reduce: BracefulStatement
			reduce(285), // do, reduce: BracefulStatement
			reduce(285), // for, reduce: BracefulStatement
			nil,         // in
			reduce(285), // break, reduce: BracefulStatement
			reduce(285), // continue, reduce: BracefulStatement
			reduce(285), // func, reduce: BracefulStatement
			reduce(285), // subr, reduce: BracefulStatement
			reduce(285), // return, reduce: BracefulStatement
		},
	},
	actionRow{ // S1085
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(286), // ;, reduce: BracefulStatement
			reduce(286), // {, reduce: BracefulStatement
			reduce(286), // }, reduce: BracefulStatement
			nil,         // =
			reduce(286), // unset, reduce: BracefulStatement
			reduce(286), // filter, reduce: BracefulStatement
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(286), // print, reduce: BracefulStatement
			nil,         // ,
			reduce(286), // printn, reduce: BracefulStatement
			reduce(286), // eprint, reduce: BracefulStatement
			reduce(286), // eprintn, reduce: BracefulStatement
			reduce(286), // dump, reduce: BracefulStatement
			reduce(286), // edump, reduce: BracefulStatement
			reduce(286), // tee, reduce: BracefulStatement
			reduce(286), // emitf, reduce: BracefulStatement
			reduce(286), // emit1, reduce: BracefulStatement
			reduce(286), // emit, reduce: BracefulStatement
			reduce(286), // (, reduce: BracefulStatement
			nil,         // )
			reduce(286), // emitp, reduce: BracefulStatement
			reduce(286), // field_name, reduce: BracefulStatement
			reduce(286), // $[, reduce: BracefulStatement
			nil,         // ]
			reduce(286), // braced_field_name, reduce: BracefulStatement
			reduce(286), // $[[, reduce: BracefulStatement
			reduce(286), // $[[[, reduce: BracefulStatement
			reduce(286), // full_srec, reduce: BracefulStatement
			reduce(286), // oosvar_name, reduce: BracefulStatement
			reduce(286), // @[, reduce: BracefulStatement
			reduce(286), // braced_oosvar_name, reduce: BracefulStatement
			reduce(286), // full_oosvar, reduce: BracefulStatement
			reduce(286), // all, reduce: BracefulStatement
			reduce(286), // non_sigil_name, reduce: BracefulStatement
			reduce(286), // arr, reduce: BracefulStatement
			reduce(286), // bool, reduce: BracefulStatement
			reduce(286), // float, reduce: BracefulStatement
			reduce(286), // int, reduce: BracefulStatement
			reduce(286), // map, reduce: BracefulStatement
			reduce(286), // num, reduce: BracefulStatement
			reduce(286), // str, reduce: BracefulStatement
			reduce(286), // var, reduce: BracefulStatement
			reduce(286), // funct, reduce: BracefulStatement
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(286), // +, reduce: BracefulStatement
			reduce(286), // -, reduce: BracefulStatement
			reduce(286), // .+, reduce: BracefulStatement
			reduce(286), // .-, reduce: BracefulStatement
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(286), // !, reduce: BracefulStatement
			reduce(286), // ~, reduce: BracefulStatement
			nil,         // **
			reduce(286), // string_literal, reduce: BracefulStatement
			reduce(286), // regex_case_insensitive, reduce: BracefulStatement
			reduce(286), // int_literal, reduce: BracefulStatement
			reduce(286), // float_literal, reduce: BracefulStatement
			reduce(286), // boolean_literal, reduce: BracefulStatement
			reduce(286), // null_literal, reduce: BracefulStatement
			reduce(286), // inf_literal, reduce: BracefulStatement
			reduce(286), // nan_literal, reduce: BracefulStatement
			reduce(286), // const_M_PI, reduce: BracefulStatement
			reduce(286), // const_M_E, reduce: BracefulStatement
			reduce(286), // panic, reduce: BracefulStatement
			reduce(286), // [, reduce: BracefulStatement
			reduce(286), // ctx_IPS, reduce: BracefulStatement
			reduce(286), // ctx_IFS, reduce: BracefulStatement
			reduce(286), // ctx_IRS, reduce: BracefulStatement
			reduce(286), // ctx_OPS, reduce: BracefulStatement
			reduce(286), // ctx_OFS, reduce: BracefulStatement
			reduce(286), // ctx_ORS, reduce: BracefulStatement
			reduce(286), // ctx_FLATSEP, reduce: BracefulStatement
			reduce(286), // ctx_NF, reduce: BracefulStatement
			reduce(286), // ctx_NR, reduce: BracefulStatement
			reduce(286), // ctx_FNR, reduce: BracefulStatement
			reduce(286), // ctx_FILENAME, reduce: BracefulStatement
			reduce(286), // ctx_FILENUM, reduce: BracefulStatement
			reduce(286), // env, reduce: BracefulStatement
			nil,         // [[
			nil,         // [[[
			reduce(286), // call, reduce: BracefulStatement
			reduce(286), // begin, reduce: BracefulStatement
			reduce(286), // end, reduce: BracefulStatement
			reduce(286), // if, reduce: BracefulStatement
			nil,         // elif
			nil,         // else
			reduce(286), // while, reduce: BracefulStatement
			reduce(286), // do, reduce: BracefulStatement
			reduce(286), // for, reduce: BracefulStatement
			nil,         // in
			reduce(286), // break, reduce: BracefulStatement
			reduce(286), // continue, reduce: BracefulStatement
			reduce(286), // func, reduce: BracefulStatement
			reduce(286), // subr, reduce: BracefulStatement
			reduce(286), // return, reduce: BracefulStatement
		},
	},
	actionRow{ // S1086
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2042), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1087
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2042), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1088
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(290), // ;, reduce: IfChain
			reduce(290), // {, reduce: IfChain
			reduce(290), // }, reduce: IfChain
			nil,         // =
			reduce(290), // unset, reduce: IfChain
			reduce(290), // filter, reduce: IfChain
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(290), // print, reduce: IfChain
			nil,         // ,
			reduce(290), // printn, reduce: IfChain
			reduce(290), // eprint, reduce: IfChain
			reduce(290), // eprintn, reduce: IfChain
			reduce(290), // dump, reduce: IfChain
			reduce(290), // edump, reduce: IfChain
			reduce(290), // tee, reduce: IfChain
			reduce(290), // emitf, reduce: IfChain
			reduce(290), // emit1, reduce: IfChain
			reduce(290), // emit, reduce: IfChain
			reduce(290), // (, reduce: IfChain
			nil,         // )
			reduce(290), // emitp, reduce: IfChain
			reduce(290), // field_name, reduce: IfChain
			reduce(290), // $[, reduce: IfChain
			nil,         // ]
			reduce(290), // braced_field_name, reduce: IfChain
			reduce(290), // $[[, reduce: IfChain
			reduce(290), // $[[[, reduce: IfChain
			reduce(290), // full_srec, reduce: IfChain
			reduce(290), // oosvar_name, reduce: IfChain
			reduce(290), // @[, reduce: IfChain
			reduce(290), // braced_oosvar_name, reduce: IfChain
			reduce(290), // full_oosvar, reduce: IfChain
			reduce(290), // all, reduce: IfChain
			reduce(290), // non_sigil_name, reduce: IfChain
			reduce(290), // arr, reduce: IfChain
			reduce(290), // bool, reduce: IfChain
			reduce(290), // float, reduce: IfChain
			reduce(290), // int, reduce: IfChain
			reduce(290), // map, reduce: IfChain
			reduce(290), // num, reduce: IfChain
			reduce(290), // str, reduce: IfChain
			reduce(290), // var, reduce: IfChain
			reduce(290), // funct, reduce: IfChain
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(290), // +, reduce: IfChain
			reduce(290), // -, reduce: IfChain
			reduce(290), // .+, reduce: IfChain
			reduce(290), // .-, reduce: IfChain
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(290), // !, reduce: IfChain
			reduce(290), // ~, reduce: IfChain
			nil,         // **
			reduce(290), // string_literal, reduce: IfChain
			reduce(290), // regex_case_insensitive, reduce: IfChain
			reduce(290), // int_literal, reduce: IfChain
			reduce(290), // float_literal, reduce: IfChain
			reduce(290), // boolean_literal, reduce: IfChain
			reduce(290), // null_literal, reduce: IfChain
			reduce(290), // inf_literal, reduce: IfChain
			reduce(290), // nan_literal, reduce: IfChain
			reduce(290), // const_M_PI, reduce: IfChain
			reduce(290), // const_M_E, reduce: IfChain
			reduce(290), // panic, reduce: IfChain
			reduce(290), // [, reduce: IfChain
			reduce(290), // ctx_IPS, reduce: IfChain
			reduce(290), // ctx_IFS, reduce: IfChain
			reduce(290), // ctx_IRS, reduce: IfChain
			reduce(290), // ctx_OPS, reduce: IfChain
			reduce(290), // ctx_OFS, reduce: IfChain
			reduce(290), // ctx_ORS, reduce: IfChain
			reduce(290), // ctx_FLATSEP, reduce: IfChain
			reduce(290), // ctx_NF, reduce: IfChain
			reduce(290), // ctx_NR, reduce: IfChain
			reduce(290), // ctx_FNR, reduce: IfChain
			reduce(290), // ctx_FILENAME, reduce: IfChain
			reduce(290), // ctx_FILENUM, reduce: IfChain
			reduce(290), // env, reduce: IfChain
			nil,         // [[
			nil,         // [[[
			reduce(290), // call, reduce: IfChain
			reduce(290), // begin, reduce: IfChain
			reduce(290), // end, reduce: IfChain
			reduce(290), // if, reduce: IfChain
			shift(2330), // elif
			shift(2331), // else
			reduce(290), // while, reduce: IfChain
			reduce(290), // do, reduce: IfChain
			reduce(290), // for, reduce: IfChain
			nil,         // in
			reduce(290), // break, reduce: IfChain
			reduce(290), // continue, reduce: IfChain
			reduce(290), // func, reduce: IfChain
			reduce(290), // subr, reduce: IfChain
			reduce(290), // return, reduce: IfChain
		},
	},
	actionRow{ // S1089
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(292), // ;, reduce: IfElifStar
			reduce(292), // {, reduce: IfElifStar
			reduce(292), // }, reduce: IfElifStar
			nil,         // =
			reduce(292), // unset, reduce: IfElifStar
			reduce(292), // filter, reduce: IfElifStar
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(292), // print, reduce: IfElifStar
			nil,         // ,
			reduce(292), // printn, reduce: IfElifStar
			reduce(292), // eprint, reduce: IfElifStar
			reduce(292), // eprintn, reduce: IfElifStar
			reduce(292), // dump, reduce: IfElifStar
			reduce(292), // edump, reduce: IfElifStar
			reduce(292), // tee, reduce: IfElifStar
			reduce(292), // emitf, reduce: IfElifStar
			reduce(292), // emit1, reduce: IfElifStar
			reduce(292), // emit, reduce: IfElifStar
			reduce(292), // (, reduce: IfElifStar
			nil,         // )
			reduce(292), // emitp, reduce: IfElifStar
			reduce(292), // field_name, reduce: IfElifStar
			reduce(292), // $[, reduce: IfElifStar
			nil,         // ]
			reduce(292), // braced_field_name, reduce: IfElifStar
			reduce(292), // $[[, reduce: IfElifStar
			reduce(292), // $[[[, reduce: IfElifStar
			reduce(292), // full_srec, reduce: IfElifStar
			reduce(292), // oosvar_name, reduce: IfElifStar
			reduce(292), // @[, reduce: IfElifStar
			reduce(292), // braced_oosvar_name, reduce: IfElifStar
			reduce(292), // full_oosvar, reduce: IfElifStar
			reduce(292), // all, reduce: IfElifStar
			reduce(292), // non_sigil_name, reduce: IfElifStar
			reduce(292), // arr, reduce: IfElifStar
			reduce(292), // bool, reduce: IfElifStar
			reduce(292), // float, reduce: IfElifStar
			reduce(292), // int, reduce: IfElifStar
			reduce(292), // map, reduce: IfElifStar
			reduce(292), // num, reduce: IfElifStar
			reduce(292), // str, reduce: IfElifStar
			reduce(292), // var, reduce: IfElifStar
			reduce(292), // funct, reduce: IfElifStar
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(292), // +, reduce: IfElifStar
			reduce(292), // -, reduce: IfElifStar
			reduce(292), // .+, reduce: IfElifStar
			reduce(292), // .-, reduce: IfElifStar
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(292), // !, reduce: IfElifStar
			reduce(292), // ~, reduce: IfElifStar
			nil,         // **
			reduce(292), // string_literal, reduce: IfElifStar
			reduce(292), // regex_case_insensitive, reduce: IfElifStar
			reduce(292), // int_literal, reduce: IfElifStar
			reduce(292), // float_literal, reduce: IfElifStar
			reduce(292), // boolean_literal, reduce: IfElifStar
			reduce(292), // null_literal, reduce: IfElifStar
			reduce(292), // inf_literal, reduce: IfElifStar
			reduce(292), // nan_literal, reduce: IfElifStar
			reduce(292), // const_M_PI, reduce: IfElifStar
			reduce(292), // const_M_E, reduce: IfElifStar
			reduce(292), // panic, reduce: IfElifStar
			reduce(292), // [, reduce: IfElifStar
			reduce(292), // ctx_IPS, reduce: IfElifStar
			reduce(292), // ctx_IFS, reduce: IfElifStar
			reduce(292), // ctx_IRS, reduce: IfElifStar
			reduce(292), // ctx_OPS, reduce: IfElifStar
			reduce(292), // ctx_OFS, reduce: IfElifStar
			reduce(292), // ctx_ORS, reduce: IfElifStar
			reduce(292), // ctx_FLATSEP, reduce: IfElifStar
			reduce(292), // ctx_NF, reduce: IfElifStar
			reduce(292), // ctx_NR, reduce: IfElifStar
			reduce(292), // ctx_FNR, reduce: IfElifStar
			reduce(292), // ctx_FILENAME, reduce: IfElifStar
			reduce(292), // ctx_FILENUM, reduce: IfElifStar
			reduce(292), // env, reduce: IfElifStar
			nil,         // [[
			nil,         // [[[
			reduce(292), // call, reduce: IfElifStar
			reduce(292), // begin, reduce: IfElifStar
			reduce(292), // end, reduce: IfElifStar
			reduce(292), // if, reduce: IfElifStar
			reduce(292), // elif, reduce: IfElifStar
			reduce(292), // else, reduce: IfElifStar
			reduce(292), // while, reduce: IfElifStar
			reduce(292), // do, reduce: IfElifStar
			reduce(292), // for, reduce: IfElifStar
			nil,         // in
			reduce(292), // break, reduce: IfElifStar
			reduce(292), // continue, reduce: IfElifStar
			reduce(292), // func, reduce: IfElifStar
			reduce(292), // subr, reduce: IfElifStar
			reduce(292), // return, reduce: IfElifStar
		},
	},
	actionRow{ // S1090
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2332), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1091
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2333), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1092
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(859), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1093
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(299), // ;, reduce: ForLoop
			reduce(299), // {, reduce: ForLoop
			reduce(299), // }, reduce: ForLoop
			nil,         // =
			reduce(299), // unset, reduce: ForLoop
			reduce(299), // filter, reduce: ForLoop
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(299), // print, reduce: ForLoop
			nil,         // ,
			reduce(299), // printn, reduce: ForLoop
			reduce(299), // eprint, reduce: ForLoop
			reduce(299), // eprintn, reduce: ForLoop
			reduce(299), // dump, reduce: ForLoop
			reduce(299), // edump, reduce: ForLoop
			reduce(299), // tee, reduce: ForLoop
			reduce(299), // emitf, reduce: ForLoop
			reduce(299), // emit1, reduce: ForLoop
			reduce(299), // emit, reduce: ForLoop
			reduce(299), // (, reduce: ForLoop
			nil,         // )
			reduce(299), // emitp, reduce: ForLoop
			reduce(299), // field_name, reduce: ForLoop
			reduce(299), // $[, reduce: ForLoop
			nil,         // ]
			reduce(299), // braced_field_name, reduce: ForLoop
			reduce(299), // $[[, reduce: ForLoop
			reduce(299), // $[[[, reduce: ForLoop
			reduce(299), // full_srec, reduce: ForLoop
			reduce(299), // oosvar_name, reduce: ForLoop
			reduce(299), // @[, reduce: ForLoop
			reduce(299), // braced_oosvar_name, reduce: ForLoop
			reduce(299), // full_oosvar, reduce: ForLoop
			reduce(299), // all, reduce: ForLoop
			reduce(299), // non_sigil_name, reduce: ForLoop
			reduce(299), // arr, reduce: ForLoop
			reduce(299), // bool, reduce: ForLoop
			reduce(299), // float, reduce: ForLoop
			reduce(299), // int, reduce: ForLoop
			reduce(299), // map, reduce: ForLoop
			reduce(299), // num, reduce: ForLoop
			reduce(299), // str, reduce: ForLoop
			reduce(299), // var, reduce: ForLoop
			reduce(299), // funct, reduce: ForLoop
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(299), // +, reduce: ForLoop
			reduce(299), // -, reduce: ForLoop
			reduce(299), // .+, reduce: ForLoop
			reduce(299), // .-, reduce: ForLoop
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(299), // !, reduce: ForLoop
			reduce(299), // ~, reduce: ForLoop
			nil,         // **
			reduce(299), // string_literal, reduce: ForLoop
			reduce(299), // regex_case_insensitive, reduce: ForLoop
			reduce(299), // int_literal, reduce: ForLoop
			reduce(299), // float_literal, reduce: ForLoop
			reduce(299), // boolean_literal, reduce: ForLoop
			reduce(299), // null_literal, reduce: ForLoop
			reduce(299), // inf_literal, reduce: ForLoop
			reduce(299), // nan_literal, reduce: ForLoop
			reduce(299), // const_M_PI, reduce: ForLoop
			reduce(299), // const_M_E, reduce: ForLoop
			reduce(299), // panic, reduce: ForLoop
			reduce(299), // [, reduce: ForLoop
			reduce(299), // ctx_IPS, reduce: ForLoop
			reduce(299), // ctx_IFS, reduce: ForLoop
			reduce(299), // ctx_IRS, reduce: ForLoop
			reduce(299), // ctx_OPS, reduce: ForLoop
			reduce(299), // ctx_OFS, reduce: ForLoop
			reduce(299), // ctx_ORS, reduce: ForLoop
			reduce(299), // ctx_FLATSEP, reduce: ForLoop
			reduce(299), // ctx_NF, reduce: ForLoop
			reduce(299), // ctx_NR, reduce: ForLoop
			reduce(299), // ctx_FNR, reduce: ForLoop
			reduce(299), // ctx_FILENAME, reduce: ForLoop
			reduce(299), // ctx_FILENUM, reduce: ForLoop
			reduce(299), // env, reduce: ForLoop
			nil,         // [[
			nil,         // [[[
			reduce(299), // call, reduce: ForLoop
			reduce(299), // begin, reduce: ForLoop
			reduce(299), // end, reduce: ForLoop
			reduce(299), // if, reduce: ForLoop
			nil,         // elif
			nil,         // else
			reduce(299), // while, reduce: ForLoop
			reduce(299), // do, reduce: ForLoop
			reduce(299), // for, reduce: ForLoop
			nil,         // in
			reduce(299), // break, reduce: ForLoop
			reduce(299), // continue, reduce: ForLoop
			reduce(299), // func, reduce: ForLoop
			reduce(299), // subr, reduce: ForLoop
			reduce(299), // return, reduce: ForLoop
		},
	},
	actionRow{ // S1094
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(300), // ;, reduce: ForLoop
			reduce(300), // {, reduce: ForLoop
			reduce(300), // }, reduce: ForLoop
			nil,         // =
			reduce(300), // unset, reduce: ForLoop
			reduce(300), // filter, reduce: ForLoop
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(300), // print, reduce: ForLoop
			nil,         // ,
			reduce(300), // printn, reduce: ForLoop
			reduce(300), // eprint, reduce: ForLoop
			reduce(300), // eprintn, reduce: ForLoop
			reduce(300), // dump, reduce: ForLoop
			reduce(300), // edump, reduce: ForLoop
			reduce(300), // tee, reduce: ForLoop
			reduce(300), // emitf, reduce: ForLoop
			reduce(300), // emit1, reduce: ForLoop
			reduce(300), // emit, reduce: ForLoop
			reduce(300), // (, reduce: ForLoop
			nil,         // )
			reduce(300), // emitp, reduce: ForLoop
			reduce(300), // field_name, reduce: ForLoop
			reduce(300), // $[, reduce: ForLoop
			nil,         // ]
			reduce(300), // braced_field_name, reduce: ForLoop
			reduce(300), // $[[, reduce: ForLoop
			reduce(300), // $[[[, reduce: ForLoop
			reduce(300), // full_srec, reduce: ForLoop
			reduce(300), // oosvar_name, reduce: ForLoop
			reduce(300), // @[, reduce: ForLoop
			reduce(300), // braced_oosvar_name, reduce: ForLoop
			reduce(300), // full_oosvar, reduce: ForLoop
			reduce(300), // all, reduce: ForLoop
			reduce(300), // non_sigil_name, reduce: ForLoop
			reduce(300), // arr, reduce: ForLoop
			reduce(300), // bool, reduce: ForLoop
			reduce(300), // float, reduce: ForLoop
			reduce(300), // int, reduce: ForLoop
			reduce(300), // map, reduce: ForLoop
			reduce(300), // num, reduce: ForLoop
			reduce(300), // str, reduce: ForLoop
			reduce(300), // var, reduce: ForLoop
			reduce(300), // funct, reduce: ForLoop
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(300), // +, reduce: ForLoop
			reduce(300), // -, reduce: ForLoop
			reduce(300), // .+, reduce: ForLoop
			reduce(300), // .-, reduce: ForLoop
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(300), // !, reduce: ForLoop
			reduce(300), // ~, reduce: ForLoop
			nil,         // **
			reduce(300), // string_literal, reduce: ForLoop
			reduce(300), // regex_case_insensitive, reduce: ForLoop
			reduce(300), // int_literal, reduce: ForLoop
			reduce(300), // float_literal, reduce: ForLoop
			reduce(300), // boolean_literal, reduce: ForLoop
			reduce(300), // null_literal, reduce: ForLoop
			reduce(300), // inf_literal, reduce: ForLoop
			reduce(300), // nan_literal, reduce: ForLoop
			reduce(300), // const_M_PI, reduce: ForLoop
			reduce(300), // const_M_E, reduce: ForLoop
			reduce(300), // panic, reduce: ForLoop
			reduce(300), // [, reduce: ForLoop
			reduce(300), // ctx_IPS, reduce: ForLoop
			reduce(300), // ctx_IFS, reduce: ForLoop
			reduce(300), // ctx_IRS, reduce: ForLoop
			reduce(300), // ctx_OPS, reduce: ForLoop
			reduce(300), // ctx_OFS, reduce: ForLoop
			reduce(300), // ctx_ORS, reduce: ForLoop
			reduce(300), // ctx_FLATSEP, reduce: ForLoop
			reduce(300), // ctx_NF, reduce: ForLoop
			reduce(300), // ctx_NR, reduce: ForLoop
			reduce(300), // ctx_FNR, reduce: ForLoop
			reduce(300), // ctx_FILENAME, reduce: ForLoop
			reduce(300), // ctx_FILENUM, reduce: ForLoop
			reduce(300), // env, reduce: ForLoop
			nil,         // [[
			nil,         // [[[
			reduce(300), // call, reduce: ForLoop
			reduce(300), // begin, reduce: ForLoop
			reduce(300), // end, reduce: ForLoop
			reduce(300), // if, reduce: ForLoop
			nil,         // elif
			nil,         // else
			reduce(300), // while, reduce: ForLoop
			reduce(300), // do, reduce: ForLoop
			reduce(300), // for, reduce: ForLoop
			nil,         // in
			reduce(300), // break, reduce: ForLoop
			reduce(300), // continue, reduce: ForLoop
			reduce(300), // func, reduce: ForLoop
			reduce(300), // subr, reduce: ForLoop
			reduce(300), // return, reduce: ForLoop
		},
	},
	actionRow{ // S1095
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(301), // ;, reduce: ForLoop
			reduce(301), // {, reduce: ForLoop
			reduce(301), // }, reduce: ForLoop
			nil,         // =
			reduce(301), // unset, reduce: ForLoop
			reduce(301), // filter, reduce: ForLoop
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(301), // print, reduce: ForLoop
			nil,         // ,
			reduce(301), // printn, reduce: ForLoop
			reduce(301), // eprint, reduce: ForLoop
			reduce(301), // eprintn, reduce: ForLoop
			reduce(301), // dump, reduce: ForLoop
			reduce(301), // edump, reduce: ForLoop
			reduce(301), // tee, reduce: ForLoop
			reduce(301), // emitf, reduce: ForLoop
			reduce(301), // emit1, reduce: ForLoop
			reduce(301), // emit, reduce: ForLoop
			reduce(301), // (, reduce: ForLoop
			nil,         // )
			reduce(301), // emitp, reduce: ForLoop
			reduce(301), // field_name, reduce: ForLoop
			reduce(301), // $[, reduce: ForLoop
			nil,         // ]
			reduce(301), // braced_field_name, reduce: ForLoop
			reduce(301), // $[[, reduce: ForLoop
			reduce(301), // $[[[, reduce: ForLoop
			reduce(301), // full_srec, reduce: ForLoop
			reduce(301), // oosvar_name, reduce: ForLoop
			reduce(301), // @[, reduce: ForLoop
			reduce(301), // braced_oosvar_name, reduce: ForLoop
			reduce(301), // full_oosvar, reduce: ForLoop
			reduce(301), // all, reduce: ForLoop
			reduce(301), // non_sigil_name, reduce: ForLoop
			reduce(301), // arr, reduce: ForLoop
			reduce(301), // bool, reduce: ForLoop
			reduce(301), // float, reduce: ForLoop
			reduce(301), // int, reduce: ForLoop
			reduce(301), // map, reduce: ForLoop
			reduce(301), // num, reduce: ForLoop
			reduce(301), // str, reduce: ForLoop
			reduce(301), // var, reduce: ForLoop
			reduce(301), // funct, reduce: ForLoop
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(301), // +, reduce: ForLoop
			reduce(301), // -, reduce: ForLoop
			reduce(301), // .+, reduce: ForLoop
			reduce(301), // .-, reduce: ForLoop
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(301), // !, reduce: ForLoop
			reduce(301), // ~, reduce: ForLoop
			nil,         // **
			reduce(301), // string_literal, reduce: ForLoop
			reduce(301), // regex_case_insensitive, reduce: ForLoop
			reduce(301), // int_literal, reduce: ForLoop
			reduce(301), // float_literal, reduce: ForLoop
			reduce(301), // boolean_literal, reduce: ForLoop
			reduce(301), // null_literal, reduce: ForLoop
			reduce(301), // inf_literal, reduce: ForLoop
			reduce(301), // nan_literal, reduce: ForLoop
			reduce(301), // const_M_PI, reduce: ForLoop
			reduce(301), // const_M_E, reduce: ForLoop
			reduce(301), // panic, reduce: ForLoop
			reduce(301), // [, reduce: ForLoop
			reduce(301), // ctx_IPS, reduce: ForLoop
			reduce(301), // ctx_IFS, reduce: ForLoop
			reduce(301), // ctx_IRS, reduce: ForLoop
			reduce(301), // ctx_OPS, reduce: ForLoop
			reduce(301), // ctx_OFS, reduce: ForLoop
			reduce(301), // ctx_ORS, reduce: ForLoop
			reduce(301), // ctx_FLATSEP, reduce: ForLoop
			reduce(301), // ctx_NF, reduce: ForLoop
			reduce(301), // ctx_NR, reduce: ForLoop
			reduce(301), // ctx_FNR, reduce: ForLoop
			reduce(301), // ctx_FILENAME, reduce: ForLoop
			reduce(301), // ctx_FILENUM, reduce: ForLoop
			reduce(301), // env, reduce: ForLoop
			nil,         // [[
			nil,         // [[[
			reduce(301), // call, reduce: ForLoop
			reduce(301), // begin, reduce: ForLoop
			reduce(301), // end, reduce: ForLoop
			reduce(301), // if, reduce: ForLoop
			nil,         // elif
			nil,         // else
			reduce(301), // while, reduce: ForLoop
			reduce(301), // do, reduce: ForLoop
			reduce(301), // for, reduce: ForLoop
			nil,         // in
			reduce(301), // break, reduce: ForLoop
			reduce(301), // continue, reduce: ForLoop
			reduce(301), // func, reduce: ForLoop
			reduce(301), // subr, reduce: ForLoop
			reduce(301), // return, reduce: ForLoop
		},
	},
	actionRow{ // S1096
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(302), // ;, reduce: ForLoop
			reduce(302), // {, reduce: ForLoop
			reduce(302), // }, reduce: ForLoop
			nil,         // =
			reduce(302), // unset, reduce: ForLoop
			reduce(302), // filter, reduce: ForLoop
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(302), // print, reduce: ForLoop
			nil,         // ,
			reduce(302), // printn, reduce: ForLoop
			reduce(302), // eprint, reduce: ForLoop
			reduce(302), // eprintn, reduce: ForLoop
			reduce(302), // dump, reduce: ForLoop
			reduce(302), // edump, reduce: ForLoop
			reduce(302), // tee, reduce: ForLoop
			reduce(302), // emitf, reduce: ForLoop
			reduce(302), // emit1, reduce: ForLoop
			reduce(302), // emit, reduce: ForLoop
			reduce(302), // (, reduce: ForLoop
			nil,         // )
			reduce(302), // emitp, reduce: ForLoop
			reduce(302), // field_name, reduce: ForLoop
			reduce(302), // $[, reduce: ForLoop
			nil,         // ]
			reduce(302), // braced_field_name, reduce: ForLoop
			reduce(302), // $[[, reduce: ForLoop
			reduce(302), // $[[[, reduce: ForLoop
			reduce(302), // full_srec, reduce: ForLoop
			reduce(302), // oosvar_name, reduce: ForLoop
			reduce(302), // @[, reduce: ForLoop
			reduce(302), // braced_oosvar_name, reduce: ForLoop
			reduce(302), // full_oosvar, reduce: ForLoop
			reduce(302), // all, reduce: ForLoop
			reduce(302), // non_sigil_name, reduce: ForLoop
			reduce(302), // arr, reduce: ForLoop
			reduce(302), // bool, reduce: ForLoop
			reduce(302), // float, reduce: ForLoop
			reduce(302), // int, reduce: ForLoop
			reduce(302), // map, reduce: ForLoop
			reduce(302), // num, reduce: ForLoop
			reduce(302), // str, reduce: ForLoop
			reduce(302), // var, reduce: ForLoop
			reduce(302), // funct, reduce: ForLoop
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(302), // +, reduce: ForLoop
			reduce(302), // -, reduce: ForLoop
			reduce(302), // .+, reduce: ForLoop
			reduce(302), // .-, reduce: ForLoop
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(302), // !, reduce: ForLoop
			reduce(302), // ~, reduce: ForLoop
			nil,         // **
			reduce(302), // string_literal, reduce: ForLoop
			reduce(302), // regex_case_insensitive, reduce: ForLoop
			reduce(302), // int_literal, reduce: ForLoop
			reduce(302), // float_literal, reduce: ForLoop
			reduce(302), // boolean_literal, reduce: ForLoop
			reduce(302), // null_literal, reduce: ForLoop
			reduce(302), // inf_literal, reduce: ForLoop
			reduce(302), // nan_literal, reduce: ForLoop
			reduce(302), // const_M_PI, reduce: ForLoop
			reduce(302), // const_M_E, reduce: ForLoop
			reduce(302), // panic, reduce: ForLoop
			reduce(302), // [, reduce: ForLoop
			reduce(302), // ctx_IPS, reduce: ForLoop
			reduce(302), // ctx_IFS, reduce: ForLoop
			reduce(302), // ctx_IRS, reduce: ForLoop
			reduce(302), // ctx_OPS, reduce: ForLoop
			reduce(302), // ctx_OFS, reduce: ForLoop
			reduce(302), // ctx_ORS, reduce: ForLoop
			reduce(302), // ctx_FLATSEP, reduce: ForLoop
			reduce(302), // ctx_NF, reduce: ForLoop
			reduce(302), // ctx_NR, reduce: ForLoop
			reduce(302), // ctx_FNR, reduce: ForLoop
			reduce(302), // ctx_FILENAME, reduce: ForLoop
			reduce(302), // ctx_FILENUM, reduce: ForLoop
			reduce(302), // env, reduce: ForLoop
			nil,         // [[
			nil,         // [[[
			reduce(302), // call, reduce: ForLoop
			reduce(302), // begin, reduce: ForLoop
			reduce(302), // end, reduce: ForLoop
			reduce(302), // if, reduce: ForLoop
			nil,         // elif
			nil,         // else
			reduce(302), // while, reduce: ForLoop
			reduce(302), // do, reduce: ForLoop
			reduce(302), // for, reduce: ForLoop
			nil,         // in
			reduce(302), // break, reduce: ForLoop
			reduce(302), // continue, reduce: ForLoop
			reduce(302), // func, reduce: ForLoop
			reduce(302), // subr, reduce: ForLoop
			reduce(302), // return, reduce: ForLoop
		},
	},
	actionRow{ // S1097
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2335), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1098
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(320), // ;, reduce: BreakStatement
			nil,         // {
			reduce(320), // }, reduce: BreakStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1099
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(321), // ;, reduce: ContinueStatement
			nil,         // {
			reduce(321), // }, reduce: ContinueStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1100
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2336), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(2337), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1101
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(2338), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1102
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(337), // ;, reduce: ReturnStatement
			shift(2135), // {
			reduce(337), // }, reduce: ReturnStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1103
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(234), // $, reduce: MapLiteral
			nil,         // empty
			reduce(234), // ;, reduce: MapLiteral
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(234), // >, reduce: MapLiteral
			reduce(234), // >>, reduce: MapLiteral
			reduce(234), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(234), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(234), // ?, reduce: MapLiteral
			nil,         // :
			reduce(234), // ||, reduce: MapLiteral
			reduce(234), // ^^, reduce: MapLiteral
			reduce(234), // &&, reduce: MapLiteral
			reduce(234), // ??, reduce: MapLiteral
			reduce(234), // ???, reduce: MapLiteral
			reduce(234), // =~, reduce: MapLiteral
			reduce(234), // !=~, reduce: MapLiteral
			reduce(234), // ==, reduce: MapLiteral
			reduce(234), // !=, reduce: MapLiteral
			reduce(234), // <=>, reduce: MapLiteral
			reduce(234), // >=, reduce: MapLiteral
			reduce(234), // <, reduce: MapLiteral
			reduce(234), // <=, reduce: MapLiteral
			reduce(234), // ^, reduce: MapLiteral
			reduce(234), // &, reduce: MapLiteral
			reduce(234), // <<, reduce: MapLiteral
			reduce(234), // >>>, reduce: MapLiteral
			reduce(234), // +, reduce: MapLiteral
			reduce(234), // -, reduce: MapLiteral
			reduce(234), // .+, reduce: MapLiteral
			reduce(234), // .-, reduce: MapLiteral
			reduce(234), // *, reduce: MapLiteral
			reduce(234), // /, reduce: MapLiteral
			reduce(234), // //, reduce: MapLiteral
			reduce(234), // %, reduce: MapLiteral
			reduce(234), // .*, reduce: MapLiteral
			reduce(234), // ./, reduce: MapLiteral
			reduce(234), // .//, reduce: MapLiteral
			reduce(234), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(234), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(234), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(234), // [[, reduce: MapLiteral
			reduce(234), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1104
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(2340), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1105
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(274), // $, reduce: FcnArgs
			nil,         // empty
			reduce(274), // ;, reduce: FcnArgs
			shift(290),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(295),  // (
			nil,         // )
			nil,         // emitp
			shift(307),  // field_name
			shift(308),  // $[
			nil,         // ]
			shift(309),  // braced_field_name
			shift(310),  // $[[
			shift(311),  // $[[[
			shift(312),  // full_srec
			shift(314),  // oosvar_name
			shift(315),  // @[
			shift(316),  // braced_oosvar_name
			shift(317),  // full_oosvar
			shift(318),  // all
			shift(319),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(334),  // +
			shift(336),  // -
			shift(337),  // .+
			shift(338),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(341),  // !
			shift(342),  // ~
			nil,         // **
			shift(347),  // string_literal
			shift(348),  // regex_case_insensitive
			shift(349),  // int_literal
			shift(350),  // float_literal
			shift(351),  // boolean_literal
			shift(352),  // null_literal
			shift(353),  // inf_literal
			shift(354),  // nan_literal
			shift(355),  // const_M_PI
			shift(356),  // const_M_E
			shift(357),  // panic
			shift(359),  // [
			shift(361),  // ctx_IPS
			shift(362),  // ctx_IFS
			shift(363),  // ctx_IRS
			shift(364),  // ctx_OPS
			shift(365),  // ctx_OFS
			shift(366),  // ctx_ORS
			shift(367),  // ctx_FLATSEP
			shift(368),  // ctx_NF
			shift(369),  // ctx_NR
			shift(370),  // ctx_FNR
			shift(371),  // ctx_FILENAME
			shift(372),  // ctx_FILENUM
			shift(374),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(381),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1106
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(2342), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1107
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2343), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1108
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2344), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1109
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2345), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1110
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2346), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1111
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1112
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1113
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1114
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1115
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1116
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1117
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1118
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1119
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1120
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1121
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1122
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1123
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1124
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1125
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1126
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1127
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1128
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1129
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1130
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1131
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1132
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1133
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1134
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1135
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1136
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(197), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(197), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(197), // >, reduce: UnaryOpTerm
			reduce(197), // >>, reduce: UnaryOpTerm
			reduce(197), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(197), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(197), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(197), // ||, reduce: UnaryOpTerm
			reduce(197), // ^^, reduce: UnaryOpTerm
			reduce(197), // &&, reduce: UnaryOpTerm
			reduce(197), // ??, reduce: UnaryOpTerm
			reduce(197), // ???, reduce: UnaryOpTerm
			reduce(197), // =~, reduce: UnaryOpTerm
			reduce(197), // !=~, reduce: UnaryOpTerm
			reduce(197), // ==, reduce: UnaryOpTerm
			reduce(197), // !=, reduce: UnaryOpTerm
			reduce(197), // <=>, reduce: UnaryOpTerm
			reduce(197), // >=, reduce: UnaryOpTerm
			reduce(197), // <, reduce: UnaryOpTerm
			reduce(197), // <=, reduce: UnaryOpTerm
			reduce(197), // ^, reduce: UnaryOpTerm
			reduce(197), // &, reduce: UnaryOpTerm
			reduce(197), // <<, reduce: UnaryOpTerm
			reduce(197), // >>>, reduce: UnaryOpTerm
			reduce(197), // +, reduce: UnaryOpTerm
			reduce(197), // -, reduce: UnaryOpTerm
			reduce(197), // .+, reduce: UnaryOpTerm
			reduce(197), // .-, reduce: UnaryOpTerm
			reduce(197), // *, reduce: UnaryOpTerm
			reduce(197), // /, reduce: UnaryOpTerm
			reduce(197), // //, reduce: UnaryOpTerm
			reduce(197), // %, reduce: UnaryOpTerm
			reduce(197), // .*, reduce: UnaryOpTerm
			reduce(197), // ./, reduce: UnaryOpTerm
			reduce(197), // .//, reduce: UnaryOpTerm
			reduce(197), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1137
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1138
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1139
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1140
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1141
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1142
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1143
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1144
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(198), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(198), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(198), // >, reduce: UnaryOpTerm
			reduce(198), // >>, reduce: UnaryOpTerm
			reduce(198), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(198), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(198), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(198), // ||, reduce: UnaryOpTerm
			reduce(198), // ^^, reduce: UnaryOpTerm
			reduce(198), // &&, reduce: UnaryOpTerm
			reduce(198), // ??, reduce: UnaryOpTerm
			reduce(198), // ???, reduce: UnaryOpTerm
			reduce(198), // =~, reduce: UnaryOpTerm
			reduce(198), // !=~, reduce: UnaryOpTerm
			reduce(198), // ==, reduce: UnaryOpTerm
			reduce(198), // !=, reduce: UnaryOpTerm
			reduce(198), // <=>, reduce: UnaryOpTerm
			reduce(198), // >=, reduce: UnaryOpTerm
			reduce(198), // <, reduce: UnaryOpTerm
			reduce(198), // <=, reduce: UnaryOpTerm
			reduce(198), // ^, reduce: UnaryOpTerm
			reduce(198), // &, reduce: UnaryOpTerm
			reduce(198), // <<, reduce: UnaryOpTerm
			reduce(198), // >>>, reduce: UnaryOpTerm
			reduce(198), // +, reduce: UnaryOpTerm
			reduce(198), // -, reduce: UnaryOpTerm
			reduce(198), // .+, reduce: UnaryOpTerm
			reduce(198), // .-, reduce: UnaryOpTerm
			reduce(198), // *, reduce: UnaryOpTerm
			reduce(198), // /, reduce: UnaryOpTerm
			reduce(198), // //, reduce: UnaryOpTerm
			reduce(198), // %, reduce: UnaryOpTerm
			reduce(198), // .*, reduce: UnaryOpTerm
			reduce(198), // ./, reduce: UnaryOpTerm
			reduce(198), // .//, reduce: UnaryOpTerm
			reduce(198), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1145
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(199), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(199), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(199), // >, reduce: UnaryOpTerm
			reduce(199), // >>, reduce: UnaryOpTerm
			reduce(199), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(199), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(199), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(199), // ||, reduce: UnaryOpTerm
			reduce(199), // ^^, reduce: UnaryOpTerm
			reduce(199), // &&, reduce: UnaryOpTerm
			reduce(199), // ??, reduce: UnaryOpTerm
			reduce(199), // ???, reduce: UnaryOpTerm
			reduce(199), // =~, reduce: UnaryOpTerm
			reduce(199), // !=~, reduce: UnaryOpTerm
			reduce(199), // ==, reduce: UnaryOpTerm
			reduce(199), // !=, reduce: UnaryOpTerm
			reduce(199), // <=>, reduce: UnaryOpTerm
			reduce(199), // >=, reduce: UnaryOpTerm
			reduce(199), // <, reduce: UnaryOpTerm
			reduce(199), // <=, reduce: UnaryOpTerm
			reduce(199), // ^, reduce: UnaryOpTerm
			reduce(199), // &, reduce: UnaryOpTerm
			reduce(199), // <<, reduce: UnaryOpTerm
			reduce(199), // >>>, reduce: UnaryOpTerm
			reduce(199), // +, reduce: UnaryOpTerm
			reduce(199), // -, reduce: UnaryOpTerm
			reduce(199), // .+, reduce: UnaryOpTerm
			reduce(199), // .-, reduce: UnaryOpTerm
			reduce(199), // *, reduce: UnaryOpTerm
			reduce(199), // /, reduce: UnaryOpTerm
			reduce(199), // //, reduce: UnaryOpTerm
			reduce(199), // %, reduce: UnaryOpTerm
			reduce(199), // .*, reduce: UnaryOpTerm
			reduce(199), // ./, reduce: UnaryOpTerm
			reduce(199), // .//, reduce: UnaryOpTerm
			reduce(199), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1146
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(200), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(200), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(200), // >, reduce: UnaryOpTerm
			reduce(200), // >>, reduce: UnaryOpTerm
			reduce(200), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(200), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(200), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(200), // ||, reduce: UnaryOpTerm
			reduce(200), // ^^, reduce: UnaryOpTerm
			reduce(200), // &&, reduce: UnaryOpTerm
			reduce(200), // ??, reduce: UnaryOpTerm
			reduce(200), // ???, reduce: UnaryOpTerm
			reduce(200), // =~, reduce: UnaryOpTerm
			reduce(200), // !=~, reduce: UnaryOpTerm
			reduce(200), // ==, reduce: UnaryOpTerm
			reduce(200), // !=, reduce: UnaryOpTerm
			reduce(200), // <=>, reduce: UnaryOpTerm
			reduce(200), // >=, reduce: UnaryOpTerm
			reduce(200), // <, reduce: UnaryOpTerm
			reduce(200), // <=, reduce: UnaryOpTerm
			reduce(200), // ^, reduce: UnaryOpTerm
			reduce(200), // &, reduce: UnaryOpTerm
			reduce(200), // <<, reduce: UnaryOpTerm
			reduce(200), // >>>, reduce: UnaryOpTerm
			reduce(200), // +, reduce: UnaryOpTerm
			reduce(200), // -, reduce: UnaryOpTerm
			reduce(200), // .+, reduce: UnaryOpTerm
			reduce(200), // .-, reduce: UnaryOpTerm
			reduce(200), // *, reduce: UnaryOpTerm
			reduce(200), // /, reduce: UnaryOpTerm
			reduce(200), // //, reduce: UnaryOpTerm
			reduce(200), // %, reduce: UnaryOpTerm
			reduce(200), // .*, reduce: UnaryOpTerm
			reduce(200), // ./, reduce: UnaryOpTerm
			reduce(200), // .//, reduce: UnaryOpTerm
			reduce(200), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1147
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1148
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(201), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(201), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(201), // >, reduce: UnaryOpTerm
			reduce(201), // >>, reduce: UnaryOpTerm
			reduce(201), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(201), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(201), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(201), // ||, reduce: UnaryOpTerm
			reduce(201), // ^^, reduce: UnaryOpTerm
			reduce(201), // &&, reduce: UnaryOpTerm
			reduce(201), // ??, reduce: UnaryOpTerm
			reduce(201), // ???, reduce: UnaryOpTerm
			reduce(201), // =~, reduce: UnaryOpTerm
			reduce(201), // !=~, reduce: UnaryOpTerm
			reduce(201), // ==, reduce: UnaryOpTerm
			reduce(201), // !=, reduce: UnaryOpTerm
			reduce(201), // <=>, reduce: UnaryOpTerm
			reduce(201), // >=, reduce: UnaryOpTerm
			reduce(201), // <, reduce: UnaryOpTerm
			reduce(201), // <=, reduce: UnaryOpTerm
			reduce(201), // ^, reduce: UnaryOpTerm
			reduce(201), // &, reduce: UnaryOpTerm
			reduce(201), // <<, reduce: UnaryOpTerm
			reduce(201), // >>>, reduce: UnaryOpTerm
			reduce(201), // +, reduce: UnaryOpTerm
			reduce(201), // -, reduce: UnaryOpTerm
			reduce(201), // .+, reduce: UnaryOpTerm
			reduce(201), // .-, reduce: UnaryOpTerm
			reduce(201), // *, reduce: UnaryOpTerm
			reduce(201), // /, reduce: UnaryOpTerm
			reduce(201), // //, reduce: UnaryOpTerm
			reduce(201), // %, reduce: UnaryOpTerm
			reduce(201), // .*, reduce: UnaryOpTerm
			reduce(201), // ./, reduce: UnaryOpTerm
			reduce(201), // .//, reduce: UnaryOpTerm
			reduce(201), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1149
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(202), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(202), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(202), // >, reduce: UnaryOpTerm
			reduce(202), // >>, reduce: UnaryOpTerm
			reduce(202), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(202), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(202), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(202), // ||, reduce: UnaryOpTerm
			reduce(202), // ^^, reduce: UnaryOpTerm
			reduce(202), // &&, reduce: UnaryOpTerm
			reduce(202), // ??, reduce: UnaryOpTerm
			reduce(202), // ???, reduce: UnaryOpTerm
			reduce(202), // =~, reduce: UnaryOpTerm
			reduce(202), // !=~, reduce: UnaryOpTerm
			reduce(202), // ==, reduce: UnaryOpTerm
			reduce(202), // !=, reduce: UnaryOpTerm
			reduce(202), // <=>, reduce: UnaryOpTerm
			reduce(202), // >=, reduce: UnaryOpTerm
			reduce(202), // <, reduce: UnaryOpTerm
			reduce(202), // <=, reduce: UnaryOpTerm
			reduce(202), // ^, reduce: UnaryOpTerm
			reduce(202), // &, reduce: UnaryOpTerm
			reduce(202), // <<, reduce: UnaryOpTerm
			reduce(202), // >>>, reduce: UnaryOpTerm
			reduce(202), // +, reduce: UnaryOpTerm
			reduce(202), // -, reduce: UnaryOpTerm
			reduce(202), // .+, reduce: UnaryOpTerm
			reduce(202), // .-, reduce: UnaryOpTerm
			reduce(202), // *, reduce: UnaryOpTerm
			reduce(202), // /, reduce: UnaryOpTerm
			reduce(202), // //, reduce: UnaryOpTerm
			reduce(202), // %, reduce: UnaryOpTerm
			reduce(202), // .*, reduce: UnaryOpTerm
			reduce(202), // ./, reduce: UnaryOpTerm
			reduce(202), // .//, reduce: UnaryOpTerm
			reduce(202), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1150
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(290),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(295),  // (
			nil,         // )
			nil,         // emitp
			shift(307),  // field_name
			shift(308),  // $[
			nil,         // ]
			shift(309),  // braced_field_name
			shift(310),  // $[[
			shift(311),  // $[[[
			shift(312),  // full_srec
			shift(314),  // oosvar_name
			shift(315),  // @[
			shift(316),  // braced_oosvar_name
			shift(317),  // full_oosvar
			shift(318),  // all
			shift(319),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2380), // +
			shift(2381), // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(347),  // string_literal
			shift(348),  // regex_case_insensitive
			shift(349),  // int_literal
			shift(350),  // float_literal
			shift(351),  // boolean_literal
			shift(352),  // null_literal
			shift(353),  // inf_literal
			shift(354),  // nan_literal
			shift(355),  // const_M_PI
			shift(356),  // const_M_E
			shift(357),  // panic
			shift(359),  // [
			shift(361),  // ctx_IPS
			shift(362),  // ctx_IFS
			shift(363),  // ctx_IRS
			shift(364),  // ctx_OPS
			shift(365),  // ctx_OFS
			shift(366),  // ctx_ORS
			shift(367),  // ctx_FLATSEP
			shift(368),  // ctx_NF
			shift(369),  // ctx_NR
			shift(370),  // ctx_FNR
			shift(371),  // ctx_FILENAME
			shift(372),  // ctx_FILENUM
			shift(374),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(381),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1151
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(2384), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1152
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1153
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1154
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(228), // $, reduce: ArrayLiteral
			nil,         // empty
			reduce(228), // ;, reduce: ArrayLiteral
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(228), // >, reduce: ArrayLiteral
			reduce(228), // >>, reduce: ArrayLiteral
			reduce(228), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(228), // ,, reduce: ArrayLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(228), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(228), // ||, reduce: ArrayLiteral
			reduce(228), // ^^, reduce: ArrayLiteral
			reduce(228), // &&, reduce: ArrayLiteral
			reduce(228), // ??, reduce: ArrayLiteral
			reduce(228), // ???, reduce: ArrayLiteral
			reduce(228), // =~, reduce: ArrayLiteral
			reduce(228), // !=~, reduce: ArrayLiteral
			reduce(228), // ==, reduce: ArrayLiteral
			reduce(228), // !=, reduce: ArrayLiteral
			reduce(228), // <=>, reduce: ArrayLiteral
			reduce(228), // >=, reduce: ArrayLiteral
			reduce(228), // <, reduce: ArrayLiteral
			reduce(228), // <=, reduce: ArrayLiteral
			reduce(228), // ^, reduce: ArrayLiteral
			reduce(228), // &, reduce: ArrayLiteral
			reduce(228), // <<, reduce: ArrayLiteral
			reduce(228), // >>>, reduce: ArrayLiteral
			reduce(228), // +, reduce: ArrayLiteral
			reduce(228), // -, reduce: ArrayLiteral
			reduce(228), // .+, reduce: ArrayLiteral
			reduce(228), // .-, reduce: ArrayLiteral
			reduce(228), // *, reduce: ArrayLiteral
			reduce(228), // /, reduce: ArrayLiteral
			reduce(228), // //, reduce: ArrayLiteral
			reduce(228), // %, reduce: ArrayLiteral
			reduce(228), // .*, reduce: ArrayLiteral
			reduce(228), // ./, reduce: ArrayLiteral
			reduce(228), // .//, reduce: ArrayLiteral
			reduce(228), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(228), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(228), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(228), // [[, reduce: ArrayLiteral
			reduce(228), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1155
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2387), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1156
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(2388), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1157
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1158
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(2391), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1159
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1160
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(234), // $, reduce: MapLiteral
			nil,         // empty
			reduce(234), // ;, reduce: MapLiteral
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(234), // >, reduce: MapLiteral
			reduce(234), // >>, reduce: MapLiteral
			reduce(234), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(234), // ?, reduce: MapLiteral
			nil,         // :
			reduce(234), // ||, reduce: MapLiteral
			reduce(234), // ^^, reduce: MapLiteral
			reduce(234), // &&, reduce: MapLiteral
			reduce(234), // ??, reduce: MapLiteral
			reduce(234), // ???, reduce: MapLiteral
			reduce(234), // =~, reduce: MapLiteral
			reduce(234), // !=~, reduce: MapLiteral
			reduce(234), // ==, reduce: MapLiteral
			reduce(234), // !=, reduce: MapLiteral
			reduce(234), // <=>, reduce: MapLiteral
			reduce(234), // >=, reduce: MapLiteral
			reduce(234), // <, reduce: MapLiteral
			reduce(234), // <=, reduce: MapLiteral
			reduce(234), // ^, reduce: MapLiteral
			reduce(234), // &, reduce: MapLiteral
			reduce(234), // <<, reduce: MapLiteral
			reduce(234), // >>>, reduce: MapLiteral
			reduce(234), // +, reduce: MapLiteral
			reduce(234), // -, reduce: MapLiteral
			reduce(234), // .+, reduce: MapLiteral
			reduce(234), // .-, reduce: MapLiteral
			reduce(234), // *, reduce: MapLiteral
			reduce(234), // /, reduce: MapLiteral
			reduce(234), // //, reduce: MapLiteral
			reduce(234), // %, reduce: MapLiteral
			reduce(234), // .*, reduce: MapLiteral
			reduce(234), // ./, reduce: MapLiteral
			reduce(234), // .//, reduce: MapLiteral
			reduce(234), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(234), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(234), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(234), // [[, reduce: MapLiteral
			reduce(234), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1161
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(2393), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1162
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(2394), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1163
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2395), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1164
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2396), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1165
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2397), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1166
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2398), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1167
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1168
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1169
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1170
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1171
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1172
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1173
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1174
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1175
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1176
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1177
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1178
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1179
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1180
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1181
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1182
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1183
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1184
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1185
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1186
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1187
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1188
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1189
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1190
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1191
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1192
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(197), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(197), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(197), // >, reduce: UnaryOpTerm
			reduce(197), // >>, reduce: UnaryOpTerm
			reduce(197), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(197), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(197), // ||, reduce: UnaryOpTerm
			reduce(197), // ^^, reduce: UnaryOpTerm
			reduce(197), // &&, reduce: UnaryOpTerm
			reduce(197), // ??, reduce: UnaryOpTerm
			reduce(197), // ???, reduce: UnaryOpTerm
			reduce(197), // =~, reduce: UnaryOpTerm
			reduce(197), // !=~, reduce: UnaryOpTerm
			reduce(197), // ==, reduce: UnaryOpTerm
			reduce(197), // !=, reduce: UnaryOpTerm
			reduce(197), // <=>, reduce: UnaryOpTerm
			reduce(197), // >=, reduce: UnaryOpTerm
			reduce(197), // <, reduce: UnaryOpTerm
			reduce(197), // <=, reduce: UnaryOpTerm
			reduce(197), // ^, reduce: UnaryOpTerm
			reduce(197), // &, reduce: UnaryOpTerm
			reduce(197), // <<, reduce: UnaryOpTerm
			reduce(197), // >>>, reduce: UnaryOpTerm
			reduce(197), // +, reduce: UnaryOpTerm
			reduce(197), // -, reduce: UnaryOpTerm
			reduce(197), // .+, reduce: UnaryOpTerm
			reduce(197), // .-, reduce: UnaryOpTerm
			reduce(197), // *, reduce: UnaryOpTerm
			reduce(197), // /, reduce: UnaryOpTerm
			reduce(197), // //, reduce: UnaryOpTerm
			reduce(197), // %, reduce: UnaryOpTerm
			reduce(197), // .*, reduce: UnaryOpTerm
			reduce(197), // ./, reduce: UnaryOpTerm
			reduce(197), // .//, reduce: UnaryOpTerm
			reduce(197), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1193
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1194
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1195
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1196
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1197
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1198
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1199
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1200
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(198), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(198), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(198), // >, reduce: UnaryOpTerm
			reduce(198), // >>, reduce: UnaryOpTerm
			reduce(198), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(198), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(198), // ||, reduce: UnaryOpTerm
			reduce(198), // ^^, reduce: UnaryOpTerm
			reduce(198), // &&, reduce: UnaryOpTerm
			reduce(198), // ??, reduce: UnaryOpTerm
			reduce(198), // ???, reduce: UnaryOpTerm
			reduce(198), // =~, reduce: UnaryOpTerm
			reduce(198), // !=~, reduce: UnaryOpTerm
			reduce(198), // ==, reduce: UnaryOpTerm
			reduce(198), // !=, reduce: UnaryOpTerm
			reduce(198), // <=>, reduce: UnaryOpTerm
			reduce(198), // >=, reduce: UnaryOpTerm
			reduce(198), // <, reduce: UnaryOpTerm
			reduce(198), // <=, reduce: UnaryOpTerm
			reduce(198), // ^, reduce: UnaryOpTerm
			reduce(198), // &, reduce: UnaryOpTerm
			reduce(198), // <<, reduce: UnaryOpTerm
			reduce(198), // >>>, reduce: UnaryOpTerm
			reduce(198), // +, reduce: UnaryOpTerm
			reduce(198), // -, reduce: UnaryOpTerm
			reduce(198), // .+, reduce: UnaryOpTerm
			reduce(198), // .-, reduce: UnaryOpTerm
			reduce(198), // *, reduce: UnaryOpTerm
			reduce(198), // /, reduce: UnaryOpTerm
			reduce(198), // //, reduce: UnaryOpTerm
			reduce(198), // %, reduce: UnaryOpTerm
			reduce(198), // .*, reduce: UnaryOpTerm
			reduce(198), // ./, reduce: UnaryOpTerm
			reduce(198), // .//, reduce: UnaryOpTerm
			reduce(198), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1201
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(199), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(199), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(199), // >, reduce: UnaryOpTerm
			reduce(199), // >>, reduce: UnaryOpTerm
			reduce(199), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(199), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(199), // ||, reduce: UnaryOpTerm
			reduce(199), // ^^, reduce: UnaryOpTerm
			reduce(199), // &&, reduce: UnaryOpTerm
			reduce(199), // ??, reduce: UnaryOpTerm
			reduce(199), // ???, reduce: UnaryOpTerm
			reduce(199), // =~, reduce: UnaryOpTerm
			reduce(199), // !=~, reduce: UnaryOpTerm
			reduce(199), // ==, reduce: UnaryOpTerm
			reduce(199), // !=, reduce: UnaryOpTerm
			reduce(199), // <=>, reduce: UnaryOpTerm
			reduce(199), // >=, reduce: UnaryOpTerm
			reduce(199), // <, reduce: UnaryOpTerm
			reduce(199), // <=, reduce: UnaryOpTerm
			reduce(199), // ^, reduce: UnaryOpTerm
			reduce(199), // &, reduce: UnaryOpTerm
			reduce(199), // <<, reduce: UnaryOpTerm
			reduce(199), // >>>, reduce: UnaryOpTerm
			reduce(199), // +, reduce: UnaryOpTerm
			reduce(199), // -, reduce: UnaryOpTerm
			reduce(199), // .+, reduce: UnaryOpTerm
			reduce(199), // .-, reduce: UnaryOpTerm
			reduce(199), // *, reduce: UnaryOpTerm
			reduce(199), // /, reduce: UnaryOpTerm
			reduce(199), // //, reduce: UnaryOpTerm
			reduce(199), // %, reduce: UnaryOpTerm
			reduce(199), // .*, reduce: UnaryOpTerm
			reduce(199), // ./, reduce: UnaryOpTerm
			reduce(199), // .//, reduce: UnaryOpTerm
			reduce(199), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1202
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(200), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(200), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(200), // >, reduce: UnaryOpTerm
			reduce(200), // >>, reduce: UnaryOpTerm
			reduce(200), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(200), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(200), // ||, reduce: UnaryOpTerm
			reduce(200), // ^^, reduce: UnaryOpTerm
			reduce(200), // &&, reduce: UnaryOpTerm
			reduce(200), // ??, reduce: UnaryOpTerm
			reduce(200), // ???, reduce: UnaryOpTerm
			reduce(200), // =~, reduce: UnaryOpTerm
			reduce(200), // !=~, reduce: UnaryOpTerm
			reduce(200), // ==, reduce: UnaryOpTerm
			reduce(200), // !=, reduce: UnaryOpTerm
			reduce(200), // <=>, reduce: UnaryOpTerm
			reduce(200), // >=, reduce: UnaryOpTerm
			reduce(200), // <, reduce: UnaryOpTerm
			reduce(200), // <=, reduce: UnaryOpTerm
			reduce(200), // ^, reduce: UnaryOpTerm
			reduce(200), // &, reduce: UnaryOpTerm
			reduce(200), // <<, reduce: UnaryOpTerm
			reduce(200), // >>>, reduce: UnaryOpTerm
			reduce(200), // +, reduce: UnaryOpTerm
			reduce(200), // -, reduce: UnaryOpTerm
			reduce(200), // .+, reduce: UnaryOpTerm
			reduce(200), // .-, reduce: UnaryOpTerm
			reduce(200), // *, reduce: UnaryOpTerm
			reduce(200), // /, reduce: UnaryOpTerm
			reduce(200), // //, reduce: UnaryOpTerm
			reduce(200), // %, reduce: UnaryOpTerm
			reduce(200), // .*, reduce: UnaryOpTerm
			reduce(200), // ./, reduce: UnaryOpTerm
			reduce(200), // .//, reduce: UnaryOpTerm
			reduce(200), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1203
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1204
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(201), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(201), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(201), // >, reduce: UnaryOpTerm
			reduce(201), // >>, reduce: UnaryOpTerm
			reduce(201), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(201), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(201), // ||, reduce: UnaryOpTerm
			reduce(201), // ^^, reduce: UnaryOpTerm
			reduce(201), // &&, reduce: UnaryOpTerm
			reduce(201), // ??, reduce: UnaryOpTerm
			reduce(201), // ???, reduce: UnaryOpTerm
			reduce(201), // =~, reduce: UnaryOpTerm
			reduce(201), // !=~, reduce: UnaryOpTerm
			reduce(201), // ==, reduce: UnaryOpTerm
			reduce(201), // !=, reduce: UnaryOpTerm
			reduce(201), // <=>, reduce: UnaryOpTerm
			reduce(201), // >=, reduce: UnaryOpTerm
			reduce(201), // <, reduce: UnaryOpTerm
			reduce(201), // <=, reduce: UnaryOpTerm
			reduce(201), // ^, reduce: UnaryOpTerm
			reduce(201), // &, reduce: UnaryOpTerm
			reduce(201), // <<, reduce: UnaryOpTerm
			reduce(201), // >>>, reduce: UnaryOpTerm
			reduce(201), // +, reduce: UnaryOpTerm
			reduce(201), // -, reduce: UnaryOpTerm
			reduce(201), // .+, reduce: UnaryOpTerm
			reduce(201), // .-, reduce: UnaryOpTerm
			reduce(201), // *, reduce: UnaryOpTerm
			reduce(201), // /, reduce: UnaryOpTerm
			reduce(201), // //, reduce: UnaryOpTerm
			reduce(201), // %, reduce: UnaryOpTerm
			reduce(201), // .*, reduce: UnaryOpTerm
			reduce(201), // ./, reduce: UnaryOpTerm
			reduce(201), // .//, reduce: UnaryOpTerm
			reduce(201), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1205
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(202), // $, reduce: UnaryOpTerm
			nil,         // empty
			reduce(202), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(202), // >, reduce: UnaryOpTerm
			reduce(202), // >>, reduce: UnaryOpTerm
			reduce(202), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(202), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(202), // ||, reduce: UnaryOpTerm
			reduce(202), // ^^, reduce: UnaryOpTerm
			reduce(202), // &&, reduce: UnaryOpTerm
			reduce(202), // ??, reduce: UnaryOpTerm
			reduce(202), // ???, reduce: UnaryOpTerm
			reduce(202), // =~, reduce: UnaryOpTerm
			reduce(202), // !=~, reduce: UnaryOpTerm
			reduce(202), // ==, reduce: UnaryOpTerm
			reduce(202), // !=, reduce: UnaryOpTerm
			reduce(202), // <=>, reduce: UnaryOpTerm
			reduce(202), // >=, reduce: UnaryOpTerm
			reduce(202), // <, reduce: UnaryOpTerm
			reduce(202), // <=, reduce: UnaryOpTerm
			reduce(202), // ^, reduce: UnaryOpTerm
			reduce(202), // &, reduce: UnaryOpTerm
			reduce(202), // <<, reduce: UnaryOpTerm
			reduce(202), // >>>, reduce: UnaryOpTerm
			reduce(202), // +, reduce: UnaryOpTerm
			reduce(202), // -, reduce: UnaryOpTerm
			reduce(202), // .+, reduce: UnaryOpTerm
			reduce(202), // .-, reduce: UnaryOpTerm
			reduce(202), // *, reduce: UnaryOpTerm
			reduce(202), // /, reduce: UnaryOpTerm
			reduce(202), // //, reduce: UnaryOpTerm
			reduce(202), // %, reduce: UnaryOpTerm
			reduce(202), // .*, reduce: UnaryOpTerm
			reduce(202), // ./, reduce: UnaryOpTerm
			reduce(202), // .//, reduce: UnaryOpTerm
			reduce(202), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1206
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(384),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(388),  // (
			nil,         // )
			nil,         // emitp
			shift(400),  // field_name
			shift(401),  // $[
			nil,         // ]
			shift(402),  // braced_field_name
			shift(403),  // $[[
			shift(404),  // $[[[
			shift(405),  // full_srec
			shift(407),  // oosvar_name
			shift(408),  // @[
			shift(409),  // braced_oosvar_name
			shift(410),  // full_oosvar
			shift(411),  // all
			shift(412),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2432), // +
			shift(2433), // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(440),  // string_literal
			shift(441),  // regex_case_insensitive
			shift(442),  // int_literal
			shift(443),  // float_literal
			shift(444),  // boolean_literal
			shift(445),  // null_literal
			shift(446),  // inf_literal
			shift(447),  // nan_literal
			shift(448),  // const_M_PI
			shift(449),  // const_M_E
			shift(450),  // panic
			shift(452),  // [
			shift(454),  // ctx_IPS
			shift(455),  // ctx_IFS
			shift(456),  // ctx_IRS
			shift(457),  // ctx_OPS
			shift(458),  // ctx_OFS
			shift(459),  // ctx_ORS
			shift(460),  // ctx_FLATSEP
			shift(461),  // ctx_NF
			shift(462),  // ctx_NR
			shift(463),  // ctx_FNR
			shift(464),  // ctx_FILENAME
			shift(465),  // ctx_FILENUM
			shift(467),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(474),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1207
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(2436), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1208
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1209
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1210
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(228), // $, reduce: ArrayLiteral
			nil,         // empty
			reduce(228), // ;, reduce: ArrayLiteral
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(228), // >, reduce: ArrayLiteral
			reduce(228), // >>, reduce: ArrayLiteral
			reduce(228), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(228), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(228), // ||, reduce: ArrayLiteral
			reduce(228), // ^^, reduce: ArrayLiteral
			reduce(228), // &&, reduce: ArrayLiteral
			reduce(228), // ??, reduce: ArrayLiteral
			reduce(228), // ???, reduce: ArrayLiteral
			reduce(228), // =~, reduce: ArrayLiteral
			reduce(228), // !=~, reduce: ArrayLiteral
			reduce(228), // ==, reduce: ArrayLiteral
			reduce(228), // !=, reduce: ArrayLiteral
			reduce(228), // <=>, reduce: ArrayLiteral
			reduce(228), // >=, reduce: ArrayLiteral
			reduce(228), // <, reduce: ArrayLiteral
			reduce(228), // <=, reduce: ArrayLiteral
			reduce(228), // ^, reduce: ArrayLiteral
			reduce(228), // &, reduce: ArrayLiteral
			reduce(228), // <<, reduce: ArrayLiteral
			reduce(228), // >>>, reduce: ArrayLiteral
			reduce(228), // +, reduce: ArrayLiteral
			reduce(228), // -, reduce: ArrayLiteral
			reduce(228), // .+, reduce: ArrayLiteral
			reduce(228), // .-, reduce: ArrayLiteral
			reduce(228), // *, reduce: ArrayLiteral
			reduce(228), // /, reduce: ArrayLiteral
			reduce(228), // //, reduce: ArrayLiteral
			reduce(228), // %, reduce: ArrayLiteral
			reduce(228), // .*, reduce: ArrayLiteral
			reduce(228), // ./, reduce: ArrayLiteral
			reduce(228), // .//, reduce: ArrayLiteral
			reduce(228), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(228), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(228), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(228), // [[, reduce: ArrayLiteral
			reduce(228), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1211
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2439), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1212
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(2440), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1213
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1214
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(2443), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1215
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1216
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1217
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(45), // $, reduce: RedirectTarget
			nil,        // empty
			reduce(45), // ;, reduce: RedirectTarget
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(45), // ,, reduce: RedirectTarget
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1218
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(40), // $, reduce: Redirector
			nil,        // empty
			reduce(40), // ;, reduce: Redirector
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(40), // ,, reduce: Redirector
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1219
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(43), // $, reduce: RedirectTarget
			nil,        // empty
			reduce(43), // ;, reduce: RedirectTarget
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(43), // ,, reduce: RedirectTarget
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1220
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(44), // $, reduce: RedirectTarget
			nil,        // empty
			reduce(44), // ;, reduce: RedirectTarget
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(44), // ,, reduce: RedirectTarget
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1221
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(41), // $, reduce: Redirector
			nil,        // empty
			reduce(41), // ;, reduce: Redirector
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(41), // ,, reduce: Redirector
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1222
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(42), // $, reduce: Redirector
			nil,        // empty
			reduce(42), // ;, reduce: Redirector
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(42), // ,, reduce: Redirector
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1223
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1224
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1225
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(2449), // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1226
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			shift(2450), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1227
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(45), // ,, reduce: RedirectTarget
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1228
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(214), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(214), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1229
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(40), // ,, reduce: Redirector
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1230
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(43), // ,, reduce: RedirectTarget
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1231
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(44), // ,, reduce: RedirectTarget
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1232
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(211), // >, reduce: MlrvalOrFunction
			reduce(211), // >>, reduce: MlrvalOrFunction
			reduce(211), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(211), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(211), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(211), // ||, reduce: MlrvalOrFunction
			reduce(211), // ^^, reduce: MlrvalOrFunction
			reduce(211), // &&, reduce: MlrvalOrFunction
			reduce(211), // ??, reduce: MlrvalOrFunction
			reduce(211), // ???, reduce: MlrvalOrFunction
			reduce(211), // =~, reduce: MlrvalOrFunction
			reduce(211), // !=~, reduce: MlrvalOrFunction
			reduce(211), // ==, reduce: MlrvalOrFunction
			reduce(211), // !=, reduce: MlrvalOrFunction
			reduce(211), // <=>, reduce: MlrvalOrFunction
			reduce(211), // >=, reduce: MlrvalOrFunction
			reduce(211), // <, reduce: MlrvalOrFunction
			reduce(211), // <=, reduce: MlrvalOrFunction
			reduce(211), // ^, reduce: MlrvalOrFunction
			reduce(211), // &, reduce: MlrvalOrFunction
			reduce(211), // <<, reduce: MlrvalOrFunction
			reduce(211), // >>>, reduce: MlrvalOrFunction
			reduce(211), // +, reduce: MlrvalOrFunction
			reduce(211), // -, reduce: MlrvalOrFunction
			reduce(211), // .+, reduce: MlrvalOrFunction
			reduce(211), // .-, reduce: MlrvalOrFunction
			reduce(211), // *, reduce: MlrvalOrFunction
			reduce(211), // /, reduce: MlrvalOrFunction
			reduce(211), // //, reduce: MlrvalOrFunction
			reduce(211), // %, reduce: MlrvalOrFunction
			reduce(211), // .*, reduce: MlrvalOrFunction
			reduce(211), // ./, reduce: MlrvalOrFunction
			reduce(211), // .//, reduce: MlrvalOrFunction
			reduce(211), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(211), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(211), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(211), // [[, reduce: MlrvalOrFunction
			reduce(211), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1233
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1234
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(110), // >, reduce: OosvarValue
			reduce(110), // >>, reduce: OosvarValue
			reduce(110), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(110), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(110), // ?, reduce: OosvarValue
			nil,         // :
			reduce(110), // ||, reduce: OosvarValue
			reduce(110), // ^^, reduce: OosvarValue
			reduce(110), // &&, reduce: OosvarValue
			reduce(110), // ??, reduce: OosvarValue
			reduce(110), // ???, reduce: OosvarValue
			reduce(110), // =~, reduce: OosvarValue
			reduce(110), // !=~, reduce: OosvarValue
			reduce(110), // ==, reduce: OosvarValue
			reduce(110), // !=, reduce: OosvarValue
			reduce(110), // <=>, reduce: OosvarValue
			reduce(110), // >=, reduce: OosvarValue
			reduce(110), // <, reduce: OosvarValue
			reduce(110), // <=, reduce: OosvarValue
			reduce(110), // ^, reduce: OosvarValue
			reduce(110), // &, reduce: OosvarValue
			reduce(110), // <<, reduce: OosvarValue
			reduce(110), // >>>, reduce: OosvarValue
			reduce(110), // +, reduce: OosvarValue
			reduce(110), // -, reduce: OosvarValue
			reduce(110), // .+, reduce: OosvarValue
			reduce(110), // .-, reduce: OosvarValue
			reduce(110), // *, reduce: OosvarValue
			reduce(110), // /, reduce: OosvarValue
			reduce(110), // //, reduce: OosvarValue
			reduce(110), // %, reduce: OosvarValue
			reduce(110), // .*, reduce: OosvarValue
			reduce(110), // ./, reduce: OosvarValue
			reduce(110), // .//, reduce: OosvarValue
			reduce(110), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(110), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(110), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(110), // [[, reduce: OosvarValue
			reduce(110), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1235
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(112), // >, reduce: OosvarValue
			reduce(112), // >>, reduce: OosvarValue
			reduce(112), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(112), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(112), // ?, reduce: OosvarValue
			nil,         // :
			reduce(112), // ||, reduce: OosvarValue
			reduce(112), // ^^, reduce: OosvarValue
			reduce(112), // &&, reduce: OosvarValue
			reduce(112), // ??, reduce: OosvarValue
			reduce(112), // ???, reduce: OosvarValue
			reduce(112), // =~, reduce: OosvarValue
			reduce(112), // !=~, reduce: OosvarValue
			reduce(112), // ==, reduce: OosvarValue
			reduce(112), // !=, reduce: OosvarValue
			reduce(112), // <=>, reduce: OosvarValue
			reduce(112), // >=, reduce: OosvarValue
			reduce(112), // <, reduce: OosvarValue
			reduce(112), // <=, reduce: OosvarValue
			reduce(112), // ^, reduce: OosvarValue
			reduce(112), // &, reduce: OosvarValue
			reduce(112), // <<, reduce: OosvarValue
			reduce(112), // >>>, reduce: OosvarValue
			reduce(112), // +, reduce: OosvarValue
			reduce(112), // -, reduce: OosvarValue
			reduce(112), // .+, reduce: OosvarValue
			reduce(112), // .-, reduce: OosvarValue
			reduce(112), // *, reduce: OosvarValue
			reduce(112), // /, reduce: OosvarValue
			reduce(112), // //, reduce: OosvarValue
			reduce(112), // %, reduce: OosvarValue
			reduce(112), // .*, reduce: OosvarValue
			reduce(112), // ./, reduce: OosvarValue
			reduce(112), // .//, reduce: OosvarValue
			reduce(112), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(112), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(112), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(112), // [[, reduce: OosvarValue
			reduce(112), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1236
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(111), // >, reduce: OosvarValue
			reduce(111), // >>, reduce: OosvarValue
			reduce(111), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(111), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(111), // ?, reduce: OosvarValue
			nil,         // :
			reduce(111), // ||, reduce: OosvarValue
			reduce(111), // ^^, reduce: OosvarValue
			reduce(111), // &&, reduce: OosvarValue
			reduce(111), // ??, reduce: OosvarValue
			reduce(111), // ???, reduce: OosvarValue
			reduce(111), // =~, reduce: OosvarValue
			reduce(111), // !=~, reduce: OosvarValue
			reduce(111), // ==, reduce: OosvarValue
			reduce(111), // !=, reduce: OosvarValue
			reduce(111), // <=>, reduce: OosvarValue
			reduce(111), // >=, reduce: OosvarValue
			reduce(111), // <, reduce: OosvarValue
			reduce(111), // <=, reduce: OosvarValue
			reduce(111), // ^, reduce: OosvarValue
			reduce(111), // &, reduce: OosvarValue
			reduce(111), // <<, reduce: OosvarValue
			reduce(111), // >>>, reduce: OosvarValue
			reduce(111), // +, reduce: OosvarValue
			reduce(111), // -, reduce: OosvarValue
			reduce(111), // .+, reduce: OosvarValue
			reduce(111), // .-, reduce: OosvarValue
			reduce(111), // *, reduce: OosvarValue
			reduce(111), // /, reduce: OosvarValue
			reduce(111), // //, reduce: OosvarValue
			reduce(111), // %, reduce: OosvarValue
			reduce(111), // .*, reduce: OosvarValue
			reduce(111), // ./, reduce: OosvarValue
			reduce(111), // .//, reduce: OosvarValue
			reduce(111), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(111), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(111), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(111), // [[, reduce: OosvarValue
			reduce(111), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1237
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(99), // >, reduce: FieldValue
			reduce(99), // >>, reduce: FieldValue
			reduce(99), // |, reduce: FieldValue
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(99), // ,, reduce: FieldValue
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(99), // ?, reduce: FieldValue
			nil,        // :
			reduce(99), // ||, reduce: FieldValue
			reduce(99), // ^^, reduce: FieldValue
			reduce(99), // &&, reduce: FieldValue
			reduce(99), // ??, reduce: FieldValue
			reduce(99), // ???, reduce: FieldValue
			reduce(99), // =~, reduce: FieldValue
			reduce(99), // !=~, reduce: FieldValue
			reduce(99), // ==, reduce: FieldValue
			reduce(99), // !=, reduce: FieldValue
			reduce(99), // <=>, reduce: FieldValue
			reduce(99), // >=, reduce: FieldValue
			reduce(99), // <, reduce: FieldValue
			reduce(99), // <=, reduce: FieldValue
			reduce(99), // ^, reduce: FieldValue
			reduce(99), // &, reduce: FieldValue
			reduce(99), // <<, reduce: FieldValue
			reduce(99), // >>>, reduce: FieldValue
			reduce(99), // +, reduce: FieldValue
			reduce(99), // -, reduce: FieldValue
			reduce(99), // .+, reduce: FieldValue
			reduce(99), // .-, reduce: FieldValue
			reduce(99), // *, reduce: FieldValue
			reduce(99), // /, reduce: FieldValue
			reduce(99), // //, reduce: FieldValue
			reduce(99), // %, reduce: FieldValue
			reduce(99), // .*, reduce: FieldValue
			reduce(99), // ./, reduce: FieldValue
			reduce(99), // .//, reduce: FieldValue
			reduce(99), // ., reduce: FieldValue
			nil,        // !
			nil,        // ~
			reduce(99), // **, reduce: FieldValue
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(99), // [, reduce: FieldValue
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(99), // [[, reduce: FieldValue
			reduce(99), // [[[, reduce: FieldValue
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1238
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(101), // >, reduce: FieldValue
			reduce(101), // >>, reduce: FieldValue
			reduce(101), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(101), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(101), // ?, reduce: FieldValue
			nil,         // :
			reduce(101), // ||, reduce: FieldValue
			reduce(101), // ^^, reduce: FieldValue
			reduce(101), // &&, reduce: FieldValue
			reduce(101), // ??, reduce: FieldValue
			reduce(101), // ???, reduce: FieldValue
			reduce(101), // =~, reduce: FieldValue
			reduce(101), // !=~, reduce: FieldValue
			reduce(101), // ==, reduce: FieldValue
			reduce(101), // !=, reduce: FieldValue
			reduce(101), // <=>, reduce: FieldValue
			reduce(101), // >=, reduce: FieldValue
			reduce(101), // <, reduce: FieldValue
			reduce(101), // <=, reduce: FieldValue
			reduce(101), // ^, reduce: FieldValue
			reduce(101), // &, reduce: FieldValue
			reduce(101), // <<, reduce: FieldValue
			reduce(101), // >>>, reduce: FieldValue
			reduce(101), // +, reduce: FieldValue
			reduce(101), // -, reduce: FieldValue
			reduce(101), // .+, reduce: FieldValue
			reduce(101), // .-, reduce: FieldValue
			reduce(101), // *, reduce: FieldValue
			reduce(101), // /, reduce: FieldValue
			reduce(101), // //, reduce: FieldValue
			reduce(101), // %, reduce: FieldValue
			reduce(101), // .*, reduce: FieldValue
			reduce(101), // ./, reduce: FieldValue
			reduce(101), // .//, reduce: FieldValue
			reduce(101), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(101), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(101), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(101), // [[, reduce: FieldValue
			reduce(101), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1239
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(100), // >, reduce: FieldValue
			reduce(100), // >>, reduce: FieldValue
			reduce(100), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(100), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(100), // ?, reduce: FieldValue
			nil,         // :
			reduce(100), // ||, reduce: FieldValue
			reduce(100), // ^^, reduce: FieldValue
			reduce(100), // &&, reduce: FieldValue
			reduce(100), // ??, reduce: FieldValue
			reduce(100), // ???, reduce: FieldValue
			reduce(100), // =~, reduce: FieldValue
			reduce(100), // !=~, reduce: FieldValue
			reduce(100), // ==, reduce: FieldValue
			reduce(100), // !=, reduce: FieldValue
			reduce(100), // <=>, reduce: FieldValue
			reduce(100), // >=, reduce: FieldValue
			reduce(100), // <, reduce: FieldValue
			reduce(100), // <=, reduce: FieldValue
			reduce(100), // ^, reduce: FieldValue
			reduce(100), // &, reduce: FieldValue
			reduce(100), // <<, reduce: FieldValue
			reduce(100), // >>>, reduce: FieldValue
			reduce(100), // +, reduce: FieldValue
			reduce(100), // -, reduce: FieldValue
			reduce(100), // .+, reduce: FieldValue
			reduce(100), // .-, reduce: FieldValue
			reduce(100), // *, reduce: FieldValue
			reduce(100), // /, reduce: FieldValue
			reduce(100), // //, reduce: FieldValue
			reduce(100), // %, reduce: FieldValue
			reduce(100), // .*, reduce: FieldValue
			reduce(100), // ./, reduce: FieldValue
			reduce(100), // .//, reduce: FieldValue
			reduce(100), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(100), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(100), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(100), // [[, reduce: FieldValue
			reduce(100), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1240
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(213), // >, reduce: MlrvalOrFunction
			reduce(213), // >>, reduce: MlrvalOrFunction
			reduce(213), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(213), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(213), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(213), // ||, reduce: MlrvalOrFunction
			reduce(213), // ^^, reduce: MlrvalOrFunction
			reduce(213), // &&, reduce: MlrvalOrFunction
			reduce(213), // ??, reduce: MlrvalOrFunction
			reduce(213), // ???, reduce: MlrvalOrFunction
			reduce(213), // =~, reduce: MlrvalOrFunction
			reduce(213), // !=~, reduce: MlrvalOrFunction
			reduce(213), // ==, reduce: MlrvalOrFunction
			reduce(213), // !=, reduce: MlrvalOrFunction
			reduce(213), // <=>, reduce: MlrvalOrFunction
			reduce(213), // >=, reduce: MlrvalOrFunction
			reduce(213), // <, reduce: MlrvalOrFunction
			reduce(213), // <=, reduce: MlrvalOrFunction
			reduce(213), // ^, reduce: MlrvalOrFunction
			reduce(213), // &, reduce: MlrvalOrFunction
			reduce(213), // <<, reduce: MlrvalOrFunction
			reduce(213), // >>>, reduce: MlrvalOrFunction
			reduce(213), // +, reduce: MlrvalOrFunction
			reduce(213), // -, reduce: MlrvalOrFunction
			reduce(213), // .+, reduce: MlrvalOrFunction
			reduce(213), // .-, reduce: MlrvalOrFunction
			reduce(213), // *, reduce: MlrvalOrFunction
			reduce(213), // /, reduce: MlrvalOrFunction
			reduce(213), // //, reduce: MlrvalOrFunction
			reduce(213), // %, reduce: MlrvalOrFunction
			reduce(213), // .*, reduce: MlrvalOrFunction
			reduce(213), // ./, reduce: MlrvalOrFunction
			reduce(213), // .//, reduce: MlrvalOrFunction
			reduce(213), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(213), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(213), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(213), // [[, reduce: MlrvalOrFunction
			reduce(213), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1241
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(233), // >, reduce: MlrvalOrFunction
			reduce(233), // >>, reduce: MlrvalOrFunction
			reduce(233), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(233), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(233), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(233), // ||, reduce: MlrvalOrFunction
			reduce(233), // ^^, reduce: MlrvalOrFunction
			reduce(233), // &&, reduce: MlrvalOrFunction
			reduce(233), // ??, reduce: MlrvalOrFunction
			reduce(233), // ???, reduce: MlrvalOrFunction
			reduce(233), // =~, reduce: MlrvalOrFunction
			reduce(233), // !=~, reduce: MlrvalOrFunction
			reduce(233), // ==, reduce: MlrvalOrFunction
			reduce(233), // !=, reduce: MlrvalOrFunction
			reduce(233), // <=>, reduce: MlrvalOrFunction
			reduce(233), // >=, reduce: MlrvalOrFunction
			reduce(233), // <, reduce: MlrvalOrFunction
			reduce(233), // <=, reduce: MlrvalOrFunction
			reduce(233), // ^, reduce: MlrvalOrFunction
			reduce(233), // &, reduce: MlrvalOrFunction
			reduce(233), // <<, reduce: MlrvalOrFunction
			reduce(233), // >>>, reduce: MlrvalOrFunction
			reduce(233), // +, reduce: MlrvalOrFunction
			reduce(233), // -, reduce: MlrvalOrFunction
			reduce(233), // .+, reduce: MlrvalOrFunction
			reduce(233), // .-, reduce: MlrvalOrFunction
			reduce(233), // *, reduce: MlrvalOrFunction
			reduce(233), // /, reduce: MlrvalOrFunction
			reduce(233), // //, reduce: MlrvalOrFunction
			reduce(233), // %, reduce: MlrvalOrFunction
			reduce(233), // .*, reduce: MlrvalOrFunction
			reduce(233), // ./, reduce: MlrvalOrFunction
			reduce(233), // .//, reduce: MlrvalOrFunction
			reduce(233), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(233), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(233), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(233), // [[, reduce: MlrvalOrFunction
			reduce(233), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1242
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(210), // >, reduce: MlrvalOrFunction
			reduce(210), // >>, reduce: MlrvalOrFunction
			reduce(210), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(210), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(210), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(210), // ||, reduce: MlrvalOrFunction
			reduce(210), // ^^, reduce: MlrvalOrFunction
			reduce(210), // &&, reduce: MlrvalOrFunction
			reduce(210), // ??, reduce: MlrvalOrFunction
			reduce(210), // ???, reduce: MlrvalOrFunction
			reduce(210), // =~, reduce: MlrvalOrFunction
			reduce(210), // !=~, reduce: MlrvalOrFunction
			reduce(210), // ==, reduce: MlrvalOrFunction
			reduce(210), // !=, reduce: MlrvalOrFunction
			reduce(210), // <=>, reduce: MlrvalOrFunction
			reduce(210), // >=, reduce: MlrvalOrFunction
			reduce(210), // <, reduce: MlrvalOrFunction
			reduce(210), // <=, reduce: MlrvalOrFunction
			reduce(210), // ^, reduce: MlrvalOrFunction
			reduce(210), // &, reduce: MlrvalOrFunction
			reduce(210), // <<, reduce: MlrvalOrFunction
			reduce(210), // >>>, reduce: MlrvalOrFunction
			reduce(210), // +, reduce: MlrvalOrFunction
			reduce(210), // -, reduce: MlrvalOrFunction
			reduce(210), // .+, reduce: MlrvalOrFunction
			reduce(210), // .-, reduce: MlrvalOrFunction
			reduce(210), // *, reduce: MlrvalOrFunction
			reduce(210), // /, reduce: MlrvalOrFunction
			reduce(210), // //, reduce: MlrvalOrFunction
			reduce(210), // %, reduce: MlrvalOrFunction
			reduce(210), // .*, reduce: MlrvalOrFunction
			reduce(210), // ./, reduce: MlrvalOrFunction
			reduce(210), // .//, reduce: MlrvalOrFunction
			reduce(210), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(210), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(210), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(210), // [[, reduce: MlrvalOrFunction
			reduce(210), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1243
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(102), // >, reduce: FieldValue
			reduce(102), // >>, reduce: FieldValue
			reduce(102), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(102), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(102), // ?, reduce: FieldValue
			nil,         // :
			reduce(102), // ||, reduce: FieldValue
			reduce(102), // ^^, reduce: FieldValue
			reduce(102), // &&, reduce: FieldValue
			reduce(102), // ??, reduce: FieldValue
			reduce(102), // ???, reduce: FieldValue
			reduce(102), // =~, reduce: FieldValue
			reduce(102), // !=~, reduce: FieldValue
			reduce(102), // ==, reduce: FieldValue
			reduce(102), // !=, reduce: FieldValue
			reduce(102), // <=>, reduce: FieldValue
			reduce(102), // >=, reduce: FieldValue
			reduce(102), // <, reduce: FieldValue
			reduce(102), // <=, reduce: FieldValue
			reduce(102), // ^, reduce: FieldValue
			reduce(102), // &, reduce: FieldValue
			reduce(102), // <<, reduce: FieldValue
			reduce(102), // >>>, reduce: FieldValue
			reduce(102), // +, reduce: FieldValue
			reduce(102), // -, reduce: FieldValue
			reduce(102), // .+, reduce: FieldValue
			reduce(102), // .-, reduce: FieldValue
			reduce(102), // *, reduce: FieldValue
			reduce(102), // /, reduce: FieldValue
			reduce(102), // //, reduce: FieldValue
			reduce(102), // %, reduce: FieldValue
			reduce(102), // .*, reduce: FieldValue
			reduce(102), // ./, reduce: FieldValue
			reduce(102), // .//, reduce: FieldValue
			reduce(102), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(102), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(102), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(102), // [[, reduce: FieldValue
			reduce(102), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1244
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(103), // >, reduce: FieldValue
			reduce(103), // >>, reduce: FieldValue
			reduce(103), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(103), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(103), // ?, reduce: FieldValue
			nil,         // :
			reduce(103), // ||, reduce: FieldValue
			reduce(103), // ^^, reduce: FieldValue
			reduce(103), // &&, reduce: FieldValue
			reduce(103), // ??, reduce: FieldValue
			reduce(103), // ???, reduce: FieldValue
			reduce(103), // =~, reduce: FieldValue
			reduce(103), // !=~, reduce: FieldValue
			reduce(103), // ==, reduce: FieldValue
			reduce(103), // !=, reduce: FieldValue
			reduce(103), // <=>, reduce: FieldValue
			reduce(103), // >=, reduce: FieldValue
			reduce(103), // <, reduce: FieldValue
			reduce(103), // <=, reduce: FieldValue
			reduce(103), // ^, reduce: FieldValue
			reduce(103), // &, reduce: FieldValue
			reduce(103), // <<, reduce: FieldValue
			reduce(103), // >>>, reduce: FieldValue
			reduce(103), // +, reduce: FieldValue
			reduce(103), // -, reduce: FieldValue
			reduce(103), // .+, reduce: FieldValue
			reduce(103), // .-, reduce: FieldValue
			reduce(103), // *, reduce: FieldValue
			reduce(103), // /, reduce: FieldValue
			reduce(103), // //, reduce: FieldValue
			reduce(103), // %, reduce: FieldValue
			reduce(103), // .*, reduce: FieldValue
			reduce(103), // ./, reduce: FieldValue
			reduce(103), // .//, reduce: FieldValue
			reduce(103), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(103), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(103), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(103), // [[, reduce: FieldValue
			reduce(103), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1245
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(104), // >, reduce: DirectFieldValue
			reduce(104), // >>, reduce: DirectFieldValue
			reduce(104), // |, reduce: DirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(104), // ,, reduce: DirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(104), // ?, reduce: DirectFieldValue
			nil,         // :
			reduce(104), // ||, reduce: DirectFieldValue
			reduce(104), // ^^, reduce: DirectFieldValue
			reduce(104), // &&, reduce: DirectFieldValue
			reduce(104), // ??, reduce: DirectFieldValue
			reduce(104), // ???, reduce: DirectFieldValue
			reduce(104), // =~, reduce: DirectFieldValue
			reduce(104), // !=~, reduce: DirectFieldValue
			reduce(104), // ==, reduce: DirectFieldValue
			reduce(104), // !=, reduce: DirectFieldValue
			reduce(104), // <=>, reduce: DirectFieldValue
			reduce(104), // >=, reduce: DirectFieldValue
			reduce(104), // <, reduce: DirectFieldValue
			reduce(104), // <=, reduce: DirectFieldValue
			reduce(104), // ^, reduce: DirectFieldValue
			reduce(104), // &, reduce: DirectFieldValue
			reduce(104), // <<, reduce: DirectFieldValue
			reduce(104), // >>>, reduce: DirectFieldValue
			reduce(104), // +, reduce: DirectFieldValue
			reduce(104), // -, reduce: DirectFieldValue
			reduce(104), // .+, reduce: DirectFieldValue
			reduce(104), // .-, reduce: DirectFieldValue
			reduce(104), // *, reduce: DirectFieldValue
			reduce(104), // /, reduce: DirectFieldValue
			reduce(104), // //, reduce: DirectFieldValue
			reduce(104), // %, reduce: DirectFieldValue
			reduce(104), // .*, reduce: DirectFieldValue
			reduce(104), // ./, reduce: DirectFieldValue
			reduce(104), // .//, reduce: DirectFieldValue
			reduce(104), // ., reduce: DirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(104), // **, reduce: DirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(104), // [, reduce: DirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(104), // [[, reduce: DirectFieldValue
			reduce(104), // [[[, reduce: DirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1246
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1247
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(106), // >, reduce: BracedFieldValue
			reduce(106), // >>, reduce: BracedFieldValue
			reduce(106), // |, reduce: BracedFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(106), // ,, reduce: BracedFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(106), // ?, reduce: BracedFieldValue
			nil,         // :
			reduce(106), // ||, reduce: BracedFieldValue
			reduce(106), // ^^, reduce: BracedFieldValue
			reduce(106), // &&, reduce: BracedFieldValue
			reduce(106), // ??, reduce: BracedFieldValue
			reduce(106), // ???, reduce: BracedFieldValue
			reduce(106), // =~, reduce: BracedFieldValue
			reduce(106), // !=~, reduce: BracedFieldValue
			reduce(106), // ==, reduce: BracedFieldValue
			reduce(106), // !=, reduce: BracedFieldValue
			reduce(106), // <=>, reduce: BracedFieldValue
			reduce(106), // >=, reduce: BracedFieldValue
			reduce(106), // <, reduce: BracedFieldValue
			reduce(106), // <=, reduce: BracedFieldValue
			reduce(106), // ^, reduce: BracedFieldValue
			reduce(106), // &, reduce: BracedFieldValue
			reduce(106), // <<, reduce: BracedFieldValue
			reduce(106), // >>>, reduce: BracedFieldValue
			reduce(106), // +, reduce: BracedFieldValue
			reduce(106), // -, reduce: BracedFieldValue
			reduce(106), // .+, reduce: BracedFieldValue
			reduce(106), // .-, reduce: BracedFieldValue
			reduce(106), // *, reduce: BracedFieldValue
			reduce(106), // /, reduce: BracedFieldValue
			reduce(106), // //, reduce: BracedFieldValue
			reduce(106), // %, reduce: BracedFieldValue
			reduce(106), // .*, reduce: BracedFieldValue
			reduce(106), // ./, reduce: BracedFieldValue
			reduce(106), // .//, reduce: BracedFieldValue
			reduce(106), // ., reduce: BracedFieldValue
			nil,         // !
			nil,         // ~
			reduce(106), // **, reduce: BracedFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(106), // [, reduce: BracedFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(106), // [[, reduce: BracedFieldValue
			reduce(106), // [[[, reduce: BracedFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1248
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1249
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1250
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(109), // >, reduce: FullSrec
			reduce(109), // >>, reduce: FullSrec
			reduce(109), // |, reduce: FullSrec
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(109), // ,, reduce: FullSrec
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(109), // ?, reduce: FullSrec
			nil,         // :
			reduce(109), // ||, reduce: FullSrec
			reduce(109), // ^^, reduce: FullSrec
			reduce(109), // &&, reduce: FullSrec
			reduce(109), // ??, reduce: FullSrec
			reduce(109), // ???, reduce: FullSrec
			reduce(109), // =~, reduce: FullSrec
			reduce(109), // !=~, reduce: FullSrec
			reduce(109), // ==, reduce: FullSrec
			reduce(109), // !=, reduce: FullSrec
			reduce(109), // <=>, reduce: FullSrec
			reduce(109), // >=, reduce: FullSrec
			reduce(109), // <, reduce: FullSrec
			reduce(109), // <=, reduce: FullSrec
			reduce(109), // ^, reduce: FullSrec
			reduce(109), // &, reduce: FullSrec
			reduce(109), // <<, reduce: FullSrec
			reduce(109), // >>>, reduce: FullSrec
			reduce(109), // +, reduce: FullSrec
			reduce(109), // -, reduce: FullSrec
			reduce(109), // .+, reduce: FullSrec
			reduce(109), // .-, reduce: FullSrec
			reduce(109), // *, reduce: FullSrec
			reduce(109), // /, reduce: FullSrec
			reduce(109), // //, reduce: FullSrec
			reduce(109), // %, reduce: FullSrec
			reduce(109), // .*, reduce: FullSrec
			reduce(109), // ./, reduce: FullSrec
			reduce(109), // .//, reduce: FullSrec
			reduce(109), // ., reduce: FullSrec
			nil,         // !
			nil,         // ~
			reduce(109), // **, reduce: FullSrec
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(109), // [, reduce: FullSrec
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(109), // [[, reduce: FullSrec
			reduce(109), // [[[, reduce: FullSrec
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1251
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(212), // >, reduce: MlrvalOrFunction
			reduce(212), // >>, reduce: MlrvalOrFunction
			reduce(212), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(212), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(212), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(212), // ||, reduce: MlrvalOrFunction
			reduce(212), // ^^, reduce: MlrvalOrFunction
			reduce(212), // &&, reduce: MlrvalOrFunction
			reduce(212), // ??, reduce: MlrvalOrFunction
			reduce(212), // ???, reduce: MlrvalOrFunction
			reduce(212), // =~, reduce: MlrvalOrFunction
			reduce(212), // !=~, reduce: MlrvalOrFunction
			reduce(212), // ==, reduce: MlrvalOrFunction
			reduce(212), // !=, reduce: MlrvalOrFunction
			reduce(212), // <=>, reduce: MlrvalOrFunction
			reduce(212), // >=, reduce: MlrvalOrFunction
			reduce(212), // <, reduce: MlrvalOrFunction
			reduce(212), // <=, reduce: MlrvalOrFunction
			reduce(212), // ^, reduce: MlrvalOrFunction
			reduce(212), // &, reduce: MlrvalOrFunction
			reduce(212), // <<, reduce: MlrvalOrFunction
			reduce(212), // >>>, reduce: MlrvalOrFunction
			reduce(212), // +, reduce: MlrvalOrFunction
			reduce(212), // -, reduce: MlrvalOrFunction
			reduce(212), // .+, reduce: MlrvalOrFunction
			reduce(212), // .-, reduce: MlrvalOrFunction
			reduce(212), // *, reduce: MlrvalOrFunction
			reduce(212), // /, reduce: MlrvalOrFunction
			reduce(212), // //, reduce: MlrvalOrFunction
			reduce(212), // %, reduce: MlrvalOrFunction
			reduce(212), // .*, reduce: MlrvalOrFunction
			reduce(212), // ./, reduce: MlrvalOrFunction
			reduce(212), // .//, reduce: MlrvalOrFunction
			reduce(212), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(212), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(212), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(212), // [[, reduce: MlrvalOrFunction
			reduce(212), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1252
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(113), // >, reduce: DirectOosvarValue
			reduce(113), // >>, reduce: DirectOosvarValue
			reduce(113), // |, reduce: DirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(113), // ,, reduce: DirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(113), // ?, reduce: DirectOosvarValue
			nil,         // :
			reduce(113), // ||, reduce: DirectOosvarValue
			reduce(113), // ^^, reduce: DirectOosvarValue
			reduce(113), // &&, reduce: DirectOosvarValue
			reduce(113), // ??, reduce: DirectOosvarValue
			reduce(113), // ???, reduce: DirectOosvarValue
			reduce(113), // =~, reduce: DirectOosvarValue
			reduce(113), // !=~, reduce: DirectOosvarValue
			reduce(113), // ==, reduce: DirectOosvarValue
			reduce(113), // !=, reduce: DirectOosvarValue
			reduce(113), // <=>, reduce: DirectOosvarValue
			reduce(113), // >=, reduce: DirectOosvarValue
			reduce(113), // <, reduce: DirectOosvarValue
			reduce(113), // <=, reduce: DirectOosvarValue
			reduce(113), // ^, reduce: DirectOosvarValue
			reduce(113), // &, reduce: DirectOosvarValue
			reduce(113), // <<, reduce: DirectOosvarValue
			reduce(113), // >>>, reduce: DirectOosvarValue
			reduce(113), // +, reduce: DirectOosvarValue
			reduce(113), // -, reduce: DirectOosvarValue
			reduce(113), // .+, reduce: DirectOosvarValue
			reduce(113), // .-, reduce: DirectOosvarValue
			reduce(113), // *, reduce: DirectOosvarValue
			reduce(113), // /, reduce: DirectOosvarValue
			reduce(113), // //, reduce: DirectOosvarValue
			reduce(113), // %, reduce: DirectOosvarValue
			reduce(113), // .*, reduce: DirectOosvarValue
			reduce(113), // ./, reduce: DirectOosvarValue
			reduce(113), // .//, reduce: DirectOosvarValue
			reduce(113), // ., reduce: DirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(113), // **, reduce: DirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(113), // [, reduce: DirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(113), // [[, reduce: DirectOosvarValue
			reduce(113), // [[[, reduce: DirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1253
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1254
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(115), // >, reduce: BracedOosvarValue
			reduce(115), // >>, reduce: BracedOosvarValue
			reduce(115), // |, reduce: BracedOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(115), // ,, reduce: BracedOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(115), // ?, reduce: BracedOosvarValue
			nil,         // :
			reduce(115), // ||, reduce: BracedOosvarValue
			reduce(115), // ^^, reduce: BracedOosvarValue
			reduce(115), // &&, reduce: BracedOosvarValue
			reduce(115), // ??, reduce: BracedOosvarValue
			reduce(115), // ???, reduce: BracedOosvarValue
			reduce(115), // =~, reduce: BracedOosvarValue
			reduce(115), // !=~, reduce: BracedOosvarValue
			reduce(115), // ==, reduce: BracedOosvarValue
			reduce(115), // !=, reduce: BracedOosvarValue
			reduce(115), // <=>, reduce: BracedOosvarValue
			reduce(115), // >=, reduce: BracedOosvarValue
			reduce(115), // <, reduce: BracedOosvarValue
			reduce(115), // <=, reduce: BracedOosvarValue
			reduce(115), // ^, reduce: BracedOosvarValue
			reduce(115), // &, reduce: BracedOosvarValue
			reduce(115), // <<, reduce: BracedOosvarValue
			reduce(115), // >>>, reduce: BracedOosvarValue
			reduce(115), // +, reduce: BracedOosvarValue
			reduce(115), // -, reduce: BracedOosvarValue
			reduce(115), // .+, reduce: BracedOosvarValue
			reduce(115), // .-, reduce: BracedOosvarValue
			reduce(115), // *, reduce: BracedOosvarValue
			reduce(115), // /, reduce: BracedOosvarValue
			reduce(115), // //, reduce: BracedOosvarValue
			reduce(115), // %, reduce: BracedOosvarValue
			reduce(115), // .*, reduce: BracedOosvarValue
			reduce(115), // ./, reduce: BracedOosvarValue
			reduce(115), // .//, reduce: BracedOosvarValue
			reduce(115), // ., reduce: BracedOosvarValue
			nil,         // !
			nil,         // ~
			reduce(115), // **, reduce: BracedOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(115), // [, reduce: BracedOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(115), // [[, reduce: BracedOosvarValue
			reduce(115), // [[[, reduce: BracedOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1255
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(116), // >, reduce: FullOosvar
			reduce(116), // >>, reduce: FullOosvar
			reduce(116), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(116), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(116), // ?, reduce: FullOosvar
			nil,         // :
			reduce(116), // ||, reduce: FullOosvar
			reduce(116), // ^^, reduce: FullOosvar
			reduce(116), // &&, reduce: FullOosvar
			reduce(116), // ??, reduce: FullOosvar
			reduce(116), // ???, reduce: FullOosvar
			reduce(116), // =~, reduce: FullOosvar
			reduce(116), // !=~, reduce: FullOosvar
			reduce(116), // ==, reduce: FullOosvar
			reduce(116), // !=, reduce: FullOosvar
			reduce(116), // <=>, reduce: FullOosvar
			reduce(116), // >=, reduce: FullOosvar
			reduce(116), // <, reduce: FullOosvar
			reduce(116), // <=, reduce: FullOosvar
			reduce(116), // ^, reduce: FullOosvar
			reduce(116), // &, reduce: FullOosvar
			reduce(116), // <<, reduce: FullOosvar
			reduce(116), // >>>, reduce: FullOosvar
			reduce(116), // +, reduce: FullOosvar
			reduce(116), // -, reduce: FullOosvar
			reduce(116), // .+, reduce: FullOosvar
			reduce(116), // .-, reduce: FullOosvar
			reduce(116), // *, reduce: FullOosvar
			reduce(116), // /, reduce: FullOosvar
			reduce(116), // //, reduce: FullOosvar
			reduce(116), // %, reduce: FullOosvar
			reduce(116), // .*, reduce: FullOosvar
			reduce(116), // ./, reduce: FullOosvar
			reduce(116), // .//, reduce: FullOosvar
			reduce(116), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(116), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(116), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(116), // [[, reduce: FullOosvar
			reduce(116), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1256
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(117), // >, reduce: FullOosvar
			reduce(117), // >>, reduce: FullOosvar
			reduce(117), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(117), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(117), // ?, reduce: FullOosvar
			nil,         // :
			reduce(117), // ||, reduce: FullOosvar
			reduce(117), // ^^, reduce: FullOosvar
			reduce(117), // &&, reduce: FullOosvar
			reduce(117), // ??, reduce: FullOosvar
			reduce(117), // ???, reduce: FullOosvar
			reduce(117), // =~, reduce: FullOosvar
			reduce(117), // !=~, reduce: FullOosvar
			reduce(117), // ==, reduce: FullOosvar
			reduce(117), // !=, reduce: FullOosvar
			reduce(117), // <=>, reduce: FullOosvar
			reduce(117), // >=, reduce: FullOosvar
			reduce(117), // <, reduce: FullOosvar
			reduce(117), // <=, reduce: FullOosvar
			reduce(117), // ^, reduce: FullOosvar
			reduce(117), // &, reduce: FullOosvar
			reduce(117), // <<, reduce: FullOosvar
			reduce(117), // >>>, reduce: FullOosvar
			reduce(117), // +, reduce: FullOosvar
			reduce(117), // -, reduce: FullOosvar
			reduce(117), // .+, reduce: FullOosvar
			reduce(117), // .-, reduce: FullOosvar
			reduce(117), // *, reduce: FullOosvar
			reduce(117), // /, reduce: FullOosvar
			reduce(117), // //, reduce: FullOosvar
			reduce(117), // %, reduce: FullOosvar
			reduce(117), // .*, reduce: FullOosvar
			reduce(117), // ./, reduce: FullOosvar
			reduce(117), // .//, reduce: FullOosvar
			reduce(117), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(117), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(117), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(117), // [[, reduce: FullOosvar
			reduce(117), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1257
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(118), // >, reduce: LocalVariable
			reduce(118), // >>, reduce: LocalVariable
			reduce(118), // |, reduce: LocalVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(118), // ,, reduce: LocalVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(270), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(118), // ?, reduce: LocalVariable
			nil,         // :
			reduce(118), // ||, reduce: LocalVariable
			reduce(118), // ^^, reduce: LocalVariable
			reduce(118), // &&, reduce: LocalVariable
			reduce(118), // ??, reduce: LocalVariable
			reduce(118), // ???, reduce: LocalVariable
			reduce(118), // =~, reduce: LocalVariable
			reduce(118), // !=~, reduce: LocalVariable
			reduce(118), // ==, reduce: LocalVariable
			reduce(118), // !=, reduce: LocalVariable
			reduce(118), // <=>, reduce: LocalVariable
			reduce(118), // >=, reduce: LocalVariable
			reduce(118), // <, reduce: LocalVariable
			reduce(118), // <=, reduce: LocalVariable
			reduce(118), // ^, reduce: LocalVariable
			reduce(118), // &, reduce: LocalVariable
			reduce(118), // <<, reduce: LocalVariable
			reduce(118), // >>>, reduce: LocalVariable
			reduce(118), // +, reduce: LocalVariable
			reduce(118), // -, reduce: LocalVariable
			reduce(118), // .+, reduce: LocalVariable
			reduce(118), // .-, reduce: LocalVariable
			reduce(118), // *, reduce: LocalVariable
			reduce(118), // /, reduce: LocalVariable
			reduce(118), // //, reduce: LocalVariable
			reduce(118), // %, reduce: LocalVariable
			reduce(118), // .*, reduce: LocalVariable
			reduce(118), // ./, reduce: LocalVariable
			reduce(118), // .//, reduce: LocalVariable
			reduce(118), // ., reduce: LocalVariable
			nil,         // !
			nil,         // ~
			reduce(118), // **, reduce: LocalVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(118), // [, reduce: LocalVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(118), // [[, reduce: LocalVariable
			reduce(118), // [[[, reduce: LocalVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1258
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(147), // ,, reduce: Rvalue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1259
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(148), // ,, reduce: PrecedenceChainStart
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1260
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(150), // ,, reduce: TernaryTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			shift(2457), // ?
			nil,         // :
			shift(2458), // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1261
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(152), // ,, reduce: LogicalOrTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(152), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(152), // ||, reduce: LogicalOrTerm
			shift(2459), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1262
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(154), // ,, reduce: LogicalXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(154), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(154), // ||, reduce: LogicalXORTerm
			reduce(154), // ^^, reduce: LogicalXORTerm
			shift(2460), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1263
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(156), // ,, reduce: LogicalAndTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(156), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(156), // ||, reduce: LogicalAndTerm
			reduce(156), // ^^, reduce: LogicalAndTerm
			reduce(156), // &&, reduce: LogicalAndTerm
			shift(2461), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1264
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(158), // ,, reduce: AbsentCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(158), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(158), // ||, reduce: AbsentCoalesceTerm
			reduce(158), // ^^, reduce: AbsentCoalesceTerm
			reduce(158), // &&, reduce: AbsentCoalesceTerm
			reduce(158), // ??, reduce: AbsentCoalesceTerm
			shift(2462), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1265
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(160), // ,, reduce: EmptyCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(160), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(160), // ||, reduce: EmptyCoalesceTerm
			reduce(160), // ^^, reduce: EmptyCoalesceTerm
			reduce(160), // &&, reduce: EmptyCoalesceTerm
			reduce(160), // ??, reduce: EmptyCoalesceTerm
			reduce(160), // ???, reduce: EmptyCoalesceTerm
			shift(2463), // =~
			shift(2464), // !=~
			shift(2465), // ==
			shift(2466), // !=
			shift(2467), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1266
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2468), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(166), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(166), // ?, reduce: EqneTerm
			nil,         // :
			reduce(166), // ||, reduce: EqneTerm
			reduce(166), // ^^, reduce: EqneTerm
			reduce(166), // &&, reduce: EqneTerm
			reduce(166), // ??, reduce: EqneTerm
			reduce(166), // ???, reduce: EqneTerm
			reduce(166), // =~, reduce: EqneTerm
			reduce(166), // !=~, reduce: EqneTerm
			reduce(166), // ==, reduce: EqneTerm
			reduce(166), // !=, reduce: EqneTerm
			reduce(166), // <=>, reduce: EqneTerm
			shift(2469), // >=
			shift(2470), // <
			shift(2471), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1267
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(171), // >, reduce: CmpTerm
			nil,         // >>
			shift(2472), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(171), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(171), // ?, reduce: CmpTerm
			nil,         // :
			reduce(171), // ||, reduce: CmpTerm
			reduce(171), // ^^, reduce: CmpTerm
			reduce(171), // &&, reduce: CmpTerm
			reduce(171), // ??, reduce: CmpTerm
			reduce(171), // ???, reduce: CmpTerm
			reduce(171), // =~, reduce: CmpTerm
			reduce(171), // !=~, reduce: CmpTerm
			reduce(171), // ==, reduce: CmpTerm
			reduce(171), // !=, reduce: CmpTerm
			reduce(171), // <=>, reduce: CmpTerm
			reduce(171), // >=, reduce: CmpTerm
			reduce(171), // <, reduce: CmpTerm
			reduce(171), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1268
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(173), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(173), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(173), // ,, reduce: BitwiseORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(173), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(173), // ||, reduce: BitwiseORTerm
			reduce(173), // ^^, reduce: BitwiseORTerm
			reduce(173), // &&, reduce: BitwiseORTerm
			reduce(173), // ??, reduce: BitwiseORTerm
			reduce(173), // ???, reduce: BitwiseORTerm
			reduce(173), // =~, reduce: BitwiseORTerm
			reduce(173), // !=~, reduce: BitwiseORTerm
			reduce(173), // ==, reduce: BitwiseORTerm
			reduce(173), // !=, reduce: BitwiseORTerm
			reduce(173), // <=>, reduce: BitwiseORTerm
			reduce(173), // >=, reduce: BitwiseORTerm
			reduce(173), // <, reduce: BitwiseORTerm
			reduce(173), // <=, reduce: BitwiseORTerm
			shift(2473), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1269
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(175), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(175), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(175), // ,, reduce: BitwiseXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(175), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(175), // ||, reduce: BitwiseXORTerm
			reduce(175), // ^^, reduce: BitwiseXORTerm
			reduce(175), // &&, reduce: BitwiseXORTerm
			reduce(175), // ??, reduce: BitwiseXORTerm
			reduce(175), // ???, reduce: BitwiseXORTerm
			reduce(175), // =~, reduce: BitwiseXORTerm
			reduce(175), // !=~, reduce: BitwiseXORTerm
			reduce(175), // ==, reduce: BitwiseXORTerm
			reduce(175), // !=, reduce: BitwiseXORTerm
			reduce(175), // <=>, reduce: BitwiseXORTerm
			reduce(175), // >=, reduce: BitwiseXORTerm
			reduce(175), // <, reduce: BitwiseXORTerm
			reduce(175), // <=, reduce: BitwiseXORTerm
			reduce(175), // ^, reduce: BitwiseXORTerm
			shift(2474), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1270
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(177), // >, reduce: BitwiseANDTerm
			shift(2475), // >>
			reduce(177), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(177), // ,, reduce: BitwiseANDTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(177), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(177), // ||, reduce: BitwiseANDTerm
			reduce(177), // ^^, reduce: BitwiseANDTerm
			reduce(177), // &&, reduce: BitwiseANDTerm
			reduce(177), // ??, reduce: BitwiseANDTerm
			reduce(177), // ???, reduce: BitwiseANDTerm
			reduce(177), // =~, reduce: BitwiseANDTerm
			reduce(177), // !=~, reduce: BitwiseANDTerm
			reduce(177), // ==, reduce: BitwiseANDTerm
			reduce(177), // !=, reduce: BitwiseANDTerm
			reduce(177), // <=>, reduce: BitwiseANDTerm
			reduce(177), // >=, reduce: BitwiseANDTerm
			reduce(177), // <, reduce: BitwiseANDTerm
			reduce(177), // <=, reduce: BitwiseANDTerm
			reduce(177), // ^, reduce: BitwiseANDTerm
			reduce(177), // &, reduce: BitwiseANDTerm
			shift(2476), // <<
			shift(2477), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1271
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(181), // >, reduce: BitwiseShiftTerm
			reduce(181), // >>, reduce: BitwiseShiftTerm
			reduce(181), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(181), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(181), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(181), // ||, reduce: BitwiseShiftTerm
			reduce(181), // ^^, reduce: BitwiseShiftTerm
			reduce(181), // &&, reduce: BitwiseShiftTerm
			reduce(181), // ??, reduce: BitwiseShiftTerm
			reduce(181), // ???, reduce: BitwiseShiftTerm
			reduce(181), // =~, reduce: BitwiseShiftTerm
			reduce(181), // !=~, reduce: BitwiseShiftTerm
			reduce(181), // ==, reduce: BitwiseShiftTerm
			reduce(181), // !=, reduce: BitwiseShiftTerm
			reduce(181), // <=>, reduce: BitwiseShiftTerm
			reduce(181), // >=, reduce: BitwiseShiftTerm
			reduce(181), // <, reduce: BitwiseShiftTerm
			reduce(181), // <=, reduce: BitwiseShiftTerm
			reduce(181), // ^, reduce: BitwiseShiftTerm
			reduce(181), // &, reduce: BitwiseShiftTerm
			reduce(181), // <<, reduce: BitwiseShiftTerm
			reduce(181), // >>>, reduce: BitwiseShiftTerm
			shift(2478), // +
			shift(2479), // -
			shift(2480), // .+
			shift(2481), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1272
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1273
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(186), // >, reduce: AddsubdotTerm
			reduce(186), // >>, reduce: AddsubdotTerm
			reduce(186), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(186), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(186), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(186), // ||, reduce: AddsubdotTerm
			reduce(186), // ^^, reduce: AddsubdotTerm
			reduce(186), // &&, reduce: AddsubdotTerm
			reduce(186), // ??, reduce: AddsubdotTerm
			reduce(186), // ???, reduce: AddsubdotTerm
			reduce(186), // =~, reduce: AddsubdotTerm
			reduce(186), // !=~, reduce: AddsubdotTerm
			reduce(186), // ==, reduce: AddsubdotTerm
			reduce(186), // !=, reduce: AddsubdotTerm
			reduce(186), // <=>, reduce: AddsubdotTerm
			reduce(186), // >=, reduce: AddsubdotTerm
			reduce(186), // <, reduce: AddsubdotTerm
			reduce(186), // <=, reduce: AddsubdotTerm
			reduce(186), // ^, reduce: AddsubdotTerm
			reduce(186), // &, reduce: AddsubdotTerm
			reduce(186), // <<, reduce: AddsubdotTerm
			reduce(186), // >>>, reduce: AddsubdotTerm
			reduce(186), // +, reduce: AddsubdotTerm
			reduce(186), // -, reduce: AddsubdotTerm
			reduce(186), // .+, reduce: AddsubdotTerm
			reduce(186), // .-, reduce: AddsubdotTerm
			shift(2483), // *
			shift(2484), // /
			shift(2485), // //
			shift(2486), // %
			shift(2487), // .*
			shift(2488), // ./
			shift(2489), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1274
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1275
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1276
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1277
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(194), // >, reduce: MuldivTerm
			reduce(194), // >>, reduce: MuldivTerm
			reduce(194), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(194), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(194), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(194), // ||, reduce: MuldivTerm
			reduce(194), // ^^, reduce: MuldivTerm
			reduce(194), // &&, reduce: MuldivTerm
			reduce(194), // ??, reduce: MuldivTerm
			reduce(194), // ???, reduce: MuldivTerm
			reduce(194), // =~, reduce: MuldivTerm
			reduce(194), // !=~, reduce: MuldivTerm
			reduce(194), // ==, reduce: MuldivTerm
			reduce(194), // !=, reduce: MuldivTerm
			reduce(194), // <=>, reduce: MuldivTerm
			reduce(194), // >=, reduce: MuldivTerm
			reduce(194), // <, reduce: MuldivTerm
			reduce(194), // <=, reduce: MuldivTerm
			reduce(194), // ^, reduce: MuldivTerm
			reduce(194), // &, reduce: MuldivTerm
			reduce(194), // <<, reduce: MuldivTerm
			reduce(194), // >>>, reduce: MuldivTerm
			reduce(194), // +, reduce: MuldivTerm
			reduce(194), // -, reduce: MuldivTerm
			reduce(194), // .+, reduce: MuldivTerm
			reduce(194), // .-, reduce: MuldivTerm
			reduce(194), // *, reduce: MuldivTerm
			reduce(194), // /, reduce: MuldivTerm
			reduce(194), // //, reduce: MuldivTerm
			reduce(194), // %, reduce: MuldivTerm
			reduce(194), // .*, reduce: MuldivTerm
			reduce(194), // ./, reduce: MuldivTerm
			reduce(194), // .//, reduce: MuldivTerm
			shift(2493), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1278
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(196), // >, reduce: DotTerm
			reduce(196), // >>, reduce: DotTerm
			reduce(196), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(196), // ,, reduce: DotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(196), // ?, reduce: DotTerm
			nil,         // :
			reduce(196), // ||, reduce: DotTerm
			reduce(196), // ^^, reduce: DotTerm
			reduce(196), // &&, reduce: DotTerm
			reduce(196), // ??, reduce: DotTerm
			reduce(196), // ???, reduce: DotTerm
			reduce(196), // =~, reduce: DotTerm
			reduce(196), // !=~, reduce: DotTerm
			reduce(196), // ==, reduce: DotTerm
			reduce(196), // !=, reduce: DotTerm
			reduce(196), // <=>, reduce: DotTerm
			reduce(196), // >=, reduce: DotTerm
			reduce(196), // <, reduce: DotTerm
			reduce(196), // <=, reduce: DotTerm
			reduce(196), // ^, reduce: DotTerm
			reduce(196), // &, reduce: DotTerm
			reduce(196), // <<, reduce: DotTerm
			reduce(196), // >>>, reduce: DotTerm
			reduce(196), // +, reduce: DotTerm
			reduce(196), // -, reduce: DotTerm
			reduce(196), // .+, reduce: DotTerm
			reduce(196), // .-, reduce: DotTerm
			reduce(196), // *, reduce: DotTerm
			reduce(196), // /, reduce: DotTerm
			reduce(196), // //, reduce: DotTerm
			reduce(196), // %, reduce: DotTerm
			reduce(196), // .*, reduce: DotTerm
			reduce(196), // ./, reduce: DotTerm
			reduce(196), // .//, reduce: DotTerm
			reduce(196), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1279
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1280
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1281
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(203), // >, reduce: UnaryOpTerm
			reduce(203), // >>, reduce: UnaryOpTerm
			reduce(203), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(203), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(203), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(203), // ||, reduce: UnaryOpTerm
			reduce(203), // ^^, reduce: UnaryOpTerm
			reduce(203), // &&, reduce: UnaryOpTerm
			reduce(203), // ??, reduce: UnaryOpTerm
			reduce(203), // ???, reduce: UnaryOpTerm
			reduce(203), // =~, reduce: UnaryOpTerm
			reduce(203), // !=~, reduce: UnaryOpTerm
			reduce(203), // ==, reduce: UnaryOpTerm
			reduce(203), // !=, reduce: UnaryOpTerm
			reduce(203), // <=>, reduce: UnaryOpTerm
			reduce(203), // >=, reduce: UnaryOpTerm
			reduce(203), // <, reduce: UnaryOpTerm
			reduce(203), // <=, reduce: UnaryOpTerm
			reduce(203), // ^, reduce: UnaryOpTerm
			reduce(203), // &, reduce: UnaryOpTerm
			reduce(203), // <<, reduce: UnaryOpTerm
			reduce(203), // >>>, reduce: UnaryOpTerm
			reduce(203), // +, reduce: UnaryOpTerm
			reduce(203), // -, reduce: UnaryOpTerm
			reduce(203), // .+, reduce: UnaryOpTerm
			reduce(203), // .-, reduce: UnaryOpTerm
			reduce(203), // *, reduce: UnaryOpTerm
			reduce(203), // /, reduce: UnaryOpTerm
			reduce(203), // //, reduce: UnaryOpTerm
			reduce(203), // %, reduce: UnaryOpTerm
			reduce(203), // .*, reduce: UnaryOpTerm
			reduce(203), // ./, reduce: UnaryOpTerm
			reduce(203), // .//, reduce: UnaryOpTerm
			reduce(203), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1282
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(207), // >, reduce: PowTerm
			reduce(207), // >>, reduce: PowTerm
			reduce(207), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(207), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(207), // ?, reduce: PowTerm
			nil,         // :
			reduce(207), // ||, reduce: PowTerm
			reduce(207), // ^^, reduce: PowTerm
			reduce(207), // &&, reduce: PowTerm
			reduce(207), // ??, reduce: PowTerm
			reduce(207), // ???, reduce: PowTerm
			reduce(207), // =~, reduce: PowTerm
			reduce(207), // !=~, reduce: PowTerm
			reduce(207), // ==, reduce: PowTerm
			reduce(207), // !=, reduce: PowTerm
			reduce(207), // <=>, reduce: PowTerm
			reduce(207), // >=, reduce: PowTerm
			reduce(207), // <, reduce: PowTerm
			reduce(207), // <=, reduce: PowTerm
			reduce(207), // ^, reduce: PowTerm
			reduce(207), // &, reduce: PowTerm
			reduce(207), // <<, reduce: PowTerm
			reduce(207), // >>>, reduce: PowTerm
			reduce(207), // +, reduce: PowTerm
			reduce(207), // -, reduce: PowTerm
			reduce(207), // .+, reduce: PowTerm
			reduce(207), // .-, reduce: PowTerm
			reduce(207), // *, reduce: PowTerm
			reduce(207), // /, reduce: PowTerm
			reduce(207), // //, reduce: PowTerm
			reduce(207), // %, reduce: PowTerm
			reduce(207), // .*, reduce: PowTerm
			reduce(207), // ./, reduce: PowTerm
			reduce(207), // .//, reduce: PowTerm
			reduce(207), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			shift(2496), // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1283
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(209), // >, reduce: PrecedenceChainEnd
			reduce(209), // >>, reduce: PrecedenceChainEnd
			reduce(209), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(209), // ,, reduce: PrecedenceChainEnd
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(209), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(209), // ||, reduce: PrecedenceChainEnd
			reduce(209), // ^^, reduce: PrecedenceChainEnd
			reduce(209), // &&, reduce: PrecedenceChainEnd
			reduce(209), // ??, reduce: PrecedenceChainEnd
			reduce(209), // ???, reduce: PrecedenceChainEnd
			reduce(209), // =~, reduce: PrecedenceChainEnd
			reduce(209), // !=~, reduce: PrecedenceChainEnd
			reduce(209), // ==, reduce: PrecedenceChainEnd
			reduce(209), // !=, reduce: PrecedenceChainEnd
			reduce(209), // <=>, reduce: PrecedenceChainEnd
			reduce(209), // >=, reduce: PrecedenceChainEnd
			reduce(209), // <, reduce: PrecedenceChainEnd
			reduce(209), // <=, reduce: PrecedenceChainEnd
			reduce(209), // ^, reduce: PrecedenceChainEnd
			reduce(209), // &, reduce: PrecedenceChainEnd
			reduce(209), // <<, reduce: PrecedenceChainEnd
			reduce(209), // >>>, reduce: PrecedenceChainEnd
			reduce(209), // +, reduce: PrecedenceChainEnd
			reduce(209), // -, reduce: PrecedenceChainEnd
			reduce(209), // .+, reduce: PrecedenceChainEnd
			reduce(209), // .-, reduce: PrecedenceChainEnd
			reduce(209), // *, reduce: PrecedenceChainEnd
			reduce(209), // /, reduce: PrecedenceChainEnd
			reduce(209), // //, reduce: PrecedenceChainEnd
			reduce(209), // %, reduce: PrecedenceChainEnd
			reduce(209), // .*, reduce: PrecedenceChainEnd
			reduce(209), // ./, reduce: PrecedenceChainEnd
			reduce(209), // .//, reduce: PrecedenceChainEnd
			reduce(209), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(209), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(2497), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			shift(2498), // [[
			shift(2499), // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1284
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(215), // >, reduce: MlrvalOrFunction
			reduce(215), // >>, reduce: MlrvalOrFunction
			reduce(215), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(215), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(215), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(215), // ||, reduce: MlrvalOrFunction
			reduce(215), // ^^, reduce: MlrvalOrFunction
			reduce(215), // &&, reduce: MlrvalOrFunction
			reduce(215), // ??, reduce: MlrvalOrFunction
			reduce(215), // ???, reduce: MlrvalOrFunction
			reduce(215), // =~, reduce: MlrvalOrFunction
			reduce(215), // !=~, reduce: MlrvalOrFunction
			reduce(215), // ==, reduce: MlrvalOrFunction
			reduce(215), // !=, reduce: MlrvalOrFunction
			reduce(215), // <=>, reduce: MlrvalOrFunction
			reduce(215), // >=, reduce: MlrvalOrFunction
			reduce(215), // <, reduce: MlrvalOrFunction
			reduce(215), // <=, reduce: MlrvalOrFunction
			reduce(215), // ^, reduce: MlrvalOrFunction
			reduce(215), // &, reduce: MlrvalOrFunction
			reduce(215), // <<, reduce: MlrvalOrFunction
			reduce(215), // >>>, reduce: MlrvalOrFunction
			reduce(215), // +, reduce: MlrvalOrFunction
			reduce(215), // -, reduce: MlrvalOrFunction
			reduce(215), // .+, reduce: MlrvalOrFunction
			reduce(215), // .-, reduce: MlrvalOrFunction
			reduce(215), // *, reduce: MlrvalOrFunction
			reduce(215), // /, reduce: MlrvalOrFunction
			reduce(215), // //, reduce: MlrvalOrFunction
			reduce(215), // %, reduce: MlrvalOrFunction
			reduce(215), // .*, reduce: MlrvalOrFunction
			reduce(215), // ./, reduce: MlrvalOrFunction
			reduce(215), // .//, reduce: MlrvalOrFunction
			reduce(215), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(215), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(215), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(215), // [[, reduce: MlrvalOrFunction
			reduce(215), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1285
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(216), // >, reduce: MlrvalOrFunction
			reduce(216), // >>, reduce: MlrvalOrFunction
			reduce(216), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(216), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(216), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(216), // ||, reduce: MlrvalOrFunction
			reduce(216), // ^^, reduce: MlrvalOrFunction
			reduce(216), // &&, reduce: MlrvalOrFunction
			reduce(216), // ??, reduce: MlrvalOrFunction
			reduce(216), // ???, reduce: MlrvalOrFunction
			reduce(216), // =~, reduce: MlrvalOrFunction
			reduce(216), // !=~, reduce: MlrvalOrFunction
			reduce(216), // ==, reduce: MlrvalOrFunction
			reduce(216), // !=, reduce: MlrvalOrFunction
			reduce(216), // <=>, reduce: MlrvalOrFunction
			reduce(216), // >=, reduce: MlrvalOrFunction
			reduce(216), // <, reduce: MlrvalOrFunction
			reduce(216), // <=, reduce: MlrvalOrFunction
			reduce(216), // ^, reduce: MlrvalOrFunction
			reduce(216), // &, reduce: MlrvalOrFunction
			reduce(216), // <<, reduce: MlrvalOrFunction
			reduce(216), // >>>, reduce: MlrvalOrFunction
			reduce(216), // +, reduce: MlrvalOrFunction
			reduce(216), // -, reduce: MlrvalOrFunction
			reduce(216), // .+, reduce: MlrvalOrFunction
			reduce(216), // .-, reduce: MlrvalOrFunction
			reduce(216), // *, reduce: MlrvalOrFunction
			reduce(216), // /, reduce: MlrvalOrFunction
			reduce(216), // //, reduce: MlrvalOrFunction
			reduce(216), // %, reduce: MlrvalOrFunction
			reduce(216), // .*, reduce: MlrvalOrFunction
			reduce(216), // ./, reduce: MlrvalOrFunction
			reduce(216), // .//, reduce: MlrvalOrFunction
			reduce(216), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(216), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(216), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(216), // [[, reduce: MlrvalOrFunction
			reduce(216), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1286
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(217), // >, reduce: MlrvalOrFunction
			reduce(217), // >>, reduce: MlrvalOrFunction
			reduce(217), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(217), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(217), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(217), // ||, reduce: MlrvalOrFunction
			reduce(217), // ^^, reduce: MlrvalOrFunction
			reduce(217), // &&, reduce: MlrvalOrFunction
			reduce(217), // ??, reduce: MlrvalOrFunction
			reduce(217), // ???, reduce: MlrvalOrFunction
			reduce(217), // =~, reduce: MlrvalOrFunction
			reduce(217), // !=~, reduce: MlrvalOrFunction
			reduce(217), // ==, reduce: MlrvalOrFunction
			reduce(217), // !=, reduce: MlrvalOrFunction
			reduce(217), // <=>, reduce: MlrvalOrFunction
			reduce(217), // >=, reduce: MlrvalOrFunction
			reduce(217), // <, reduce: MlrvalOrFunction
			reduce(217), // <=, reduce: MlrvalOrFunction
			reduce(217), // ^, reduce: MlrvalOrFunction
			reduce(217), // &, reduce: MlrvalOrFunction
			reduce(217), // <<, reduce: MlrvalOrFunction
			reduce(217), // >>>, reduce: MlrvalOrFunction
			reduce(217), // +, reduce: MlrvalOrFunction
			reduce(217), // -, reduce: MlrvalOrFunction
			reduce(217), // .+, reduce: MlrvalOrFunction
			reduce(217), // .-, reduce: MlrvalOrFunction
			reduce(217), // *, reduce: MlrvalOrFunction
			reduce(217), // /, reduce: MlrvalOrFunction
			reduce(217), // //, reduce: MlrvalOrFunction
			reduce(217), // %, reduce: MlrvalOrFunction
			reduce(217), // .*, reduce: MlrvalOrFunction
			reduce(217), // ./, reduce: MlrvalOrFunction
			reduce(217), // .//, reduce: MlrvalOrFunction
			reduce(217), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(217), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(217), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(217), // [[, reduce: MlrvalOrFunction
			reduce(217), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1287
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(218), // >, reduce: MlrvalOrFunction
			reduce(218), // >>, reduce: MlrvalOrFunction
			reduce(218), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(218), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(218), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(218), // ||, reduce: MlrvalOrFunction
			reduce(218), // ^^, reduce: MlrvalOrFunction
			reduce(218), // &&, reduce: MlrvalOrFunction
			reduce(218), // ??, reduce: MlrvalOrFunction
			reduce(218), // ???, reduce: MlrvalOrFunction
			reduce(218), // =~, reduce: MlrvalOrFunction
			reduce(218), // !=~, reduce: MlrvalOrFunction
			reduce(218), // ==, reduce: MlrvalOrFunction
			reduce(218), // !=, reduce: MlrvalOrFunction
			reduce(218), // <=>, reduce: MlrvalOrFunction
			reduce(218), // >=, reduce: MlrvalOrFunction
			reduce(218), // <, reduce: MlrvalOrFunction
			reduce(218), // <=, reduce: MlrvalOrFunction
			reduce(218), // ^, reduce: MlrvalOrFunction
			reduce(218), // &, reduce: MlrvalOrFunction
			reduce(218), // <<, reduce: MlrvalOrFunction
			reduce(218), // >>>, reduce: MlrvalOrFunction
			reduce(218), // +, reduce: MlrvalOrFunction
			reduce(218), // -, reduce: MlrvalOrFunction
			reduce(218), // .+, reduce: MlrvalOrFunction
			reduce(218), // .-, reduce: MlrvalOrFunction
			reduce(218), // *, reduce: MlrvalOrFunction
			reduce(218), // /, reduce: MlrvalOrFunction
			reduce(218), // //, reduce: MlrvalOrFunction
			reduce(218), // %, reduce: MlrvalOrFunction
			reduce(218), // .*, reduce: MlrvalOrFunction
			reduce(218), // ./, reduce: MlrvalOrFunction
			reduce(218), // .//, reduce: MlrvalOrFunction
			reduce(218), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(218), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(218), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(218), // [[, reduce: MlrvalOrFunction
			reduce(218), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1288
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(219), // >, reduce: MlrvalOrFunction
			reduce(219), // >>, reduce: MlrvalOrFunction
			reduce(219), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(219), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(219), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(219), // ||, reduce: MlrvalOrFunction
			reduce(219), // ^^, reduce: MlrvalOrFunction
			reduce(219), // &&, reduce: MlrvalOrFunction
			reduce(219), // ??, reduce: MlrvalOrFunction
			reduce(219), // ???, reduce: MlrvalOrFunction
			reduce(219), // =~, reduce: MlrvalOrFunction
			reduce(219), // !=~, reduce: MlrvalOrFunction
			reduce(219), // ==, reduce: MlrvalOrFunction
			reduce(219), // !=, reduce: MlrvalOrFunction
			reduce(219), // <=>, reduce: MlrvalOrFunction
			reduce(219), // >=, reduce: MlrvalOrFunction
			reduce(219), // <, reduce: MlrvalOrFunction
			reduce(219), // <=, reduce: MlrvalOrFunction
			reduce(219), // ^, reduce: MlrvalOrFunction
			reduce(219), // &, reduce: MlrvalOrFunction
			reduce(219), // <<, reduce: MlrvalOrFunction
			reduce(219), // >>>, reduce: MlrvalOrFunction
			reduce(219), // +, reduce: MlrvalOrFunction
			reduce(219), // -, reduce: MlrvalOrFunction
			reduce(219), // .+, reduce: MlrvalOrFunction
			reduce(219), // .-, reduce: MlrvalOrFunction
			reduce(219), // *, reduce: MlrvalOrFunction
			reduce(219), // /, reduce: MlrvalOrFunction
			reduce(219), // //, reduce: MlrvalOrFunction
			reduce(219), // %, reduce: MlrvalOrFunction
			reduce(219), // .*, reduce: MlrvalOrFunction
			reduce(219), // ./, reduce: MlrvalOrFunction
			reduce(219), // .//, reduce: MlrvalOrFunction
			reduce(219), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(219), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(219), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(219), // [[, reduce: MlrvalOrFunction
			reduce(219), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1289
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(220), // >, reduce: MlrvalOrFunction
			reduce(220), // >>, reduce: MlrvalOrFunction
			reduce(220), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(220), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(220), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(220), // ||, reduce: MlrvalOrFunction
			reduce(220), // ^^, reduce: MlrvalOrFunction
			reduce(220), // &&, reduce: MlrvalOrFunction
			reduce(220), // ??, reduce: MlrvalOrFunction
			reduce(220), // ???, reduce: MlrvalOrFunction
			reduce(220), // =~, reduce: MlrvalOrFunction
			reduce(220), // !=~, reduce: MlrvalOrFunction
			reduce(220), // ==, reduce: MlrvalOrFunction
			reduce(220), // !=, reduce: MlrvalOrFunction
			reduce(220), // <=>, reduce: MlrvalOrFunction
			reduce(220), // >=, reduce: MlrvalOrFunction
			reduce(220), // <, reduce: MlrvalOrFunction
			reduce(220), // <=, reduce: MlrvalOrFunction
			reduce(220), // ^, reduce: MlrvalOrFunction
			reduce(220), // &, reduce: MlrvalOrFunction
			reduce(220), // <<, reduce: MlrvalOrFunction
			reduce(220), // >>>, reduce: MlrvalOrFunction
			reduce(220), // +, reduce: MlrvalOrFunction
			reduce(220), // -, reduce: MlrvalOrFunction
			reduce(220), // .+, reduce: MlrvalOrFunction
			reduce(220), // .-, reduce: MlrvalOrFunction
			reduce(220), // *, reduce: MlrvalOrFunction
			reduce(220), // /, reduce: MlrvalOrFunction
			reduce(220), // //, reduce: MlrvalOrFunction
			reduce(220), // %, reduce: MlrvalOrFunction
			reduce(220), // .*, reduce: MlrvalOrFunction
			reduce(220), // ./, reduce: MlrvalOrFunction
			reduce(220), // .//, reduce: MlrvalOrFunction
			reduce(220), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(220), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(220), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(220), // [[, reduce: MlrvalOrFunction
			reduce(220), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1290
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(221), // >, reduce: MlrvalOrFunction
			reduce(221), // >>, reduce: MlrvalOrFunction
			reduce(221), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(221), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(221), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(221), // ||, reduce: MlrvalOrFunction
			reduce(221), // ^^, reduce: MlrvalOrFunction
			reduce(221), // &&, reduce: MlrvalOrFunction
			reduce(221), // ??, reduce: MlrvalOrFunction
			reduce(221), // ???, reduce: MlrvalOrFunction
			reduce(221), // =~, reduce: MlrvalOrFunction
			reduce(221), // !=~, reduce: MlrvalOrFunction
			reduce(221), // ==, reduce: MlrvalOrFunction
			reduce(221), // !=, reduce: MlrvalOrFunction
			reduce(221), // <=>, reduce: MlrvalOrFunction
			reduce(221), // >=, reduce: MlrvalOrFunction
			reduce(221), // <, reduce: MlrvalOrFunction
			reduce(221), // <=, reduce: MlrvalOrFunction
			reduce(221), // ^, reduce: MlrvalOrFunction
			reduce(221), // &, reduce: MlrvalOrFunction
			reduce(221), // <<, reduce: MlrvalOrFunction
			reduce(221), // >>>, reduce: MlrvalOrFunction
			reduce(221), // +, reduce: MlrvalOrFunction
			reduce(221), // -, reduce: MlrvalOrFunction
			reduce(221), // .+, reduce: MlrvalOrFunction
			reduce(221), // .-, reduce: MlrvalOrFunction
			reduce(221), // *, reduce: MlrvalOrFunction
			reduce(221), // /, reduce: MlrvalOrFunction
			reduce(221), // //, reduce: MlrvalOrFunction
			reduce(221), // %, reduce: MlrvalOrFunction
			reduce(221), // .*, reduce: MlrvalOrFunction
			reduce(221), // ./, reduce: MlrvalOrFunction
			reduce(221), // .//, reduce: MlrvalOrFunction
			reduce(221), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(221), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(221), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(221), // [[, reduce: MlrvalOrFunction
			reduce(221), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1291
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(222), // >, reduce: MlrvalOrFunction
			reduce(222), // >>, reduce: MlrvalOrFunction
			reduce(222), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(222), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(222), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(222), // ||, reduce: MlrvalOrFunction
			reduce(222), // ^^, reduce: MlrvalOrFunction
			reduce(222), // &&, reduce: MlrvalOrFunction
			reduce(222), // ??, reduce: MlrvalOrFunction
			reduce(222), // ???, reduce: MlrvalOrFunction
			reduce(222), // =~, reduce: MlrvalOrFunction
			reduce(222), // !=~, reduce: MlrvalOrFunction
			reduce(222), // ==, reduce: MlrvalOrFunction
			reduce(222), // !=, reduce: MlrvalOrFunction
			reduce(222), // <=>, reduce: MlrvalOrFunction
			reduce(222), // >=, reduce: MlrvalOrFunction
			reduce(222), // <, reduce: MlrvalOrFunction
			reduce(222), // <=, reduce: MlrvalOrFunction
			reduce(222), // ^, reduce: MlrvalOrFunction
			reduce(222), // &, reduce: MlrvalOrFunction
			reduce(222), // <<, reduce: MlrvalOrFunction
			reduce(222), // >>>, reduce: MlrvalOrFunction
			reduce(222), // +, reduce: MlrvalOrFunction
			reduce(222), // -, reduce: MlrvalOrFunction
			reduce(222), // .+, reduce: MlrvalOrFunction
			reduce(222), // .-, reduce: MlrvalOrFunction
			reduce(222), // *, reduce: MlrvalOrFunction
			reduce(222), // /, reduce: MlrvalOrFunction
			reduce(222), // //, reduce: MlrvalOrFunction
			reduce(222), // %, reduce: MlrvalOrFunction
			reduce(222), // .*, reduce: MlrvalOrFunction
			reduce(222), // ./, reduce: MlrvalOrFunction
			reduce(222), // .//, reduce: MlrvalOrFunction
			reduce(222), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(222), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(222), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(222), // [[, reduce: MlrvalOrFunction
			reduce(222), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1292
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(223), // >, reduce: MlrvalOrFunction
			reduce(223), // >>, reduce: MlrvalOrFunction
			reduce(223), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(223), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(223), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(223), // ||, reduce: MlrvalOrFunction
			reduce(223), // ^^, reduce: MlrvalOrFunction
			reduce(223), // &&, reduce: MlrvalOrFunction
			reduce(223), // ??, reduce: MlrvalOrFunction
			reduce(223), // ???, reduce: MlrvalOrFunction
			reduce(223), // =~, reduce: MlrvalOrFunction
			reduce(223), // !=~, reduce: MlrvalOrFunction
			reduce(223), // ==, reduce: MlrvalOrFunction
			reduce(223), // !=, reduce: MlrvalOrFunction
			reduce(223), // <=>, reduce: MlrvalOrFunction
			reduce(223), // >=, reduce: MlrvalOrFunction
			reduce(223), // <, reduce: MlrvalOrFunction
			reduce(223), // <=, reduce: MlrvalOrFunction
			reduce(223), // ^, reduce: MlrvalOrFunction
			reduce(223), // &, reduce: MlrvalOrFunction
			reduce(223), // <<, reduce: MlrvalOrFunction
			reduce(223), // >>>, reduce: MlrvalOrFunction
			reduce(223), // +, reduce: MlrvalOrFunction
			reduce(223), // -, reduce: MlrvalOrFunction
			reduce(223), // .+, reduce: MlrvalOrFunction
			reduce(223), // .-, reduce: MlrvalOrFunction
			reduce(223), // *, reduce: MlrvalOrFunction
			reduce(223), // /, reduce: MlrvalOrFunction
			reduce(223), // //, reduce: MlrvalOrFunction
			reduce(223), // %, reduce: MlrvalOrFunction
			reduce(223), // .*, reduce: MlrvalOrFunction
			reduce(223), // ./, reduce: MlrvalOrFunction
			reduce(223), // .//, reduce: MlrvalOrFunction
			reduce(223), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(223), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(223), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(223), // [[, reduce: MlrvalOrFunction
			reduce(223), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1293
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(224), // >, reduce: MlrvalOrFunction
			reduce(224), // >>, reduce: MlrvalOrFunction
			reduce(224), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(224), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(224), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(224), // ||, reduce: MlrvalOrFunction
			reduce(224), // ^^, reduce: MlrvalOrFunction
			reduce(224), // &&, reduce: MlrvalOrFunction
			reduce(224), // ??, reduce: MlrvalOrFunction
			reduce(224), // ???, reduce: MlrvalOrFunction
			reduce(224), // =~, reduce: MlrvalOrFunction
			reduce(224), // !=~, reduce: MlrvalOrFunction
			reduce(224), // ==, reduce: MlrvalOrFunction
			reduce(224), // !=, reduce: MlrvalOrFunction
			reduce(224), // <=>, reduce: MlrvalOrFunction
			reduce(224), // >=, reduce: MlrvalOrFunction
			reduce(224), // <, reduce: MlrvalOrFunction
			reduce(224), // <=, reduce: MlrvalOrFunction
			reduce(224), // ^, reduce: MlrvalOrFunction
			reduce(224), // &, reduce: MlrvalOrFunction
			reduce(224), // <<, reduce: MlrvalOrFunction
			reduce(224), // >>>, reduce: MlrvalOrFunction
			reduce(224), // +, reduce: MlrvalOrFunction
			reduce(224), // -, reduce: MlrvalOrFunction
			reduce(224), // .+, reduce: MlrvalOrFunction
			reduce(224), // .-, reduce: MlrvalOrFunction
			reduce(224), // *, reduce: MlrvalOrFunction
			reduce(224), // /, reduce: MlrvalOrFunction
			reduce(224), // //, reduce: MlrvalOrFunction
			reduce(224), // %, reduce: MlrvalOrFunction
			reduce(224), // .*, reduce: MlrvalOrFunction
			reduce(224), // ./, reduce: MlrvalOrFunction
			reduce(224), // .//, reduce: MlrvalOrFunction
			reduce(224), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(224), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(224), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(224), // [[, reduce: MlrvalOrFunction
			reduce(224), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1294
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(225), // >, reduce: MlrvalOrFunction
			reduce(225), // >>, reduce: MlrvalOrFunction
			reduce(225), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(225), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(225), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(225), // ||, reduce: MlrvalOrFunction
			reduce(225), // ^^, reduce: MlrvalOrFunction
			reduce(225), // &&, reduce: MlrvalOrFunction
			reduce(225), // ??, reduce: MlrvalOrFunction
			reduce(225), // ???, reduce: MlrvalOrFunction
			reduce(225), // =~, reduce: MlrvalOrFunction
			reduce(225), // !=~, reduce: MlrvalOrFunction
			reduce(225), // ==, reduce: MlrvalOrFunction
			reduce(225), // !=, reduce: MlrvalOrFunction
			reduce(225), // <=>, reduce: MlrvalOrFunction
			reduce(225), // >=, reduce: MlrvalOrFunction
			reduce(225), // <, reduce: MlrvalOrFunction
			reduce(225), // <=, reduce: MlrvalOrFunction
			reduce(225), // ^, reduce: MlrvalOrFunction
			reduce(225), // &, reduce: MlrvalOrFunction
			reduce(225), // <<, reduce: MlrvalOrFunction
			reduce(225), // >>>, reduce: MlrvalOrFunction
			reduce(225), // +, reduce: MlrvalOrFunction
			reduce(225), // -, reduce: MlrvalOrFunction
			reduce(225), // .+, reduce: MlrvalOrFunction
			reduce(225), // .-, reduce: MlrvalOrFunction
			reduce(225), // *, reduce: MlrvalOrFunction
			reduce(225), // /, reduce: MlrvalOrFunction
			reduce(225), // //, reduce: MlrvalOrFunction
			reduce(225), // %, reduce: MlrvalOrFunction
			reduce(225), // .*, reduce: MlrvalOrFunction
			reduce(225), // ./, reduce: MlrvalOrFunction
			reduce(225), // .//, reduce: MlrvalOrFunction
			reduce(225), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(225), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(225), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(225), // [[, reduce: MlrvalOrFunction
			reduce(225), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1295
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(226), // >, reduce: MlrvalOrFunction
			reduce(226), // >>, reduce: MlrvalOrFunction
			reduce(226), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(226), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(226), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(226), // ||, reduce: MlrvalOrFunction
			reduce(226), // ^^, reduce: MlrvalOrFunction
			reduce(226), // &&, reduce: MlrvalOrFunction
			reduce(226), // ??, reduce: MlrvalOrFunction
			reduce(226), // ???, reduce: MlrvalOrFunction
			reduce(226), // =~, reduce: MlrvalOrFunction
			reduce(226), // !=~, reduce: MlrvalOrFunction
			reduce(226), // ==, reduce: MlrvalOrFunction
			reduce(226), // !=, reduce: MlrvalOrFunction
			reduce(226), // <=>, reduce: MlrvalOrFunction
			reduce(226), // >=, reduce: MlrvalOrFunction
			reduce(226), // <, reduce: MlrvalOrFunction
			reduce(226), // <=, reduce: MlrvalOrFunction
			reduce(226), // ^, reduce: MlrvalOrFunction
			reduce(226), // &, reduce: MlrvalOrFunction
			reduce(226), // <<, reduce: MlrvalOrFunction
			reduce(226), // >>>, reduce: MlrvalOrFunction
			reduce(226), // +, reduce: MlrvalOrFunction
			reduce(226), // -, reduce: MlrvalOrFunction
			reduce(226), // .+, reduce: MlrvalOrFunction
			reduce(226), // .-, reduce: MlrvalOrFunction
			reduce(226), // *, reduce: MlrvalOrFunction
			reduce(226), // /, reduce: MlrvalOrFunction
			reduce(226), // //, reduce: MlrvalOrFunction
			reduce(226), // %, reduce: MlrvalOrFunction
			reduce(226), // .*, reduce: MlrvalOrFunction
			reduce(226), // ./, reduce: MlrvalOrFunction
			reduce(226), // .//, reduce: MlrvalOrFunction
			reduce(226), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(226), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(226), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(226), // [[, reduce: MlrvalOrFunction
			reduce(226), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1296
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(227), // >, reduce: MlrvalOrFunction
			reduce(227), // >>, reduce: MlrvalOrFunction
			reduce(227), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(227), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(227), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(227), // ||, reduce: MlrvalOrFunction
			reduce(227), // ^^, reduce: MlrvalOrFunction
			reduce(227), // &&, reduce: MlrvalOrFunction
			reduce(227), // ??, reduce: MlrvalOrFunction
			reduce(227), // ???, reduce: MlrvalOrFunction
			reduce(227), // =~, reduce: MlrvalOrFunction
			reduce(227), // !=~, reduce: MlrvalOrFunction
			reduce(227), // ==, reduce: MlrvalOrFunction
			reduce(227), // !=, reduce: MlrvalOrFunction
			reduce(227), // <=>, reduce: MlrvalOrFunction
			reduce(227), // >=, reduce: MlrvalOrFunction
			reduce(227), // <, reduce: MlrvalOrFunction
			reduce(227), // <=, reduce: MlrvalOrFunction
			reduce(227), // ^, reduce: MlrvalOrFunction
			reduce(227), // &, reduce: MlrvalOrFunction
			reduce(227), // <<, reduce: MlrvalOrFunction
			reduce(227), // >>>, reduce: MlrvalOrFunction
			reduce(227), // +, reduce: MlrvalOrFunction
			reduce(227), // -, reduce: MlrvalOrFunction
			reduce(227), // .+, reduce: MlrvalOrFunction
			reduce(227), // .-, reduce: MlrvalOrFunction
			reduce(227), // *, reduce: MlrvalOrFunction
			reduce(227), // /, reduce: MlrvalOrFunction
			reduce(227), // //, reduce: MlrvalOrFunction
			reduce(227), // %, reduce: MlrvalOrFunction
			reduce(227), // .*, reduce: MlrvalOrFunction
			reduce(227), // ./, reduce: MlrvalOrFunction
			reduce(227), // .//, reduce: MlrvalOrFunction
			reduce(227), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(227), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(227), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(227), // [[, reduce: MlrvalOrFunction
			reduce(227), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1297
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(752),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(756),  // (
			nil,         // )
			nil,         // emitp
			shift(768),  // field_name
			shift(769),  // $[
			shift(2500), // ]
			shift(771),  // braced_field_name
			shift(772),  // $[[
			shift(773),  // $[[[
			shift(774),  // full_srec
			shift(776),  // oosvar_name
			shift(777),  // @[
			shift(778),  // braced_oosvar_name
			shift(779),  // full_oosvar
			shift(780),  // all
			shift(781),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(796),  // +
			shift(798),  // -
			shift(799),  // .+
			shift(800),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(803),  // !
			shift(804),  // ~
			nil,         // **
			shift(809),  // string_literal
			shift(810),  // regex_case_insensitive
			shift(811),  // int_literal
			shift(812),  // float_literal
			shift(813),  // boolean_literal
			shift(814),  // null_literal
			shift(815),  // inf_literal
			shift(816),  // nan_literal
			shift(817),  // const_M_PI
			shift(818),  // const_M_E
			shift(819),  // panic
			shift(821),  // [
			shift(824),  // ctx_IPS
			shift(825),  // ctx_IFS
			shift(826),  // ctx_IRS
			shift(827),  // ctx_OPS
			shift(828),  // ctx_OFS
			shift(829),  // ctx_ORS
			shift(830),  // ctx_FLATSEP
			shift(831),  // ctx_NF
			shift(832),  // ctx_NR
			shift(833),  // ctx_FNR
			shift(834),  // ctx_FILENAME
			shift(835),  // ctx_FILENUM
			shift(837),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(844),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1298
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(240), // >, reduce: MlrvalOrFunction
			reduce(240), // >>, reduce: MlrvalOrFunction
			reduce(240), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(240), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(240), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(240), // ||, reduce: MlrvalOrFunction
			reduce(240), // ^^, reduce: MlrvalOrFunction
			reduce(240), // &&, reduce: MlrvalOrFunction
			reduce(240), // ??, reduce: MlrvalOrFunction
			reduce(240), // ???, reduce: MlrvalOrFunction
			reduce(240), // =~, reduce: MlrvalOrFunction
			reduce(240), // !=~, reduce: MlrvalOrFunction
			reduce(240), // ==, reduce: MlrvalOrFunction
			reduce(240), // !=, reduce: MlrvalOrFunction
			reduce(240), // <=>, reduce: MlrvalOrFunction
			reduce(240), // >=, reduce: MlrvalOrFunction
			reduce(240), // <, reduce: MlrvalOrFunction
			reduce(240), // <=, reduce: MlrvalOrFunction
			reduce(240), // ^, reduce: MlrvalOrFunction
			reduce(240), // &, reduce: MlrvalOrFunction
			reduce(240), // <<, reduce: MlrvalOrFunction
			reduce(240), // >>>, reduce: MlrvalOrFunction
			reduce(240), // +, reduce: MlrvalOrFunction
			reduce(240), // -, reduce: MlrvalOrFunction
			reduce(240), // .+, reduce: MlrvalOrFunction
			reduce(240), // .-, reduce: MlrvalOrFunction
			reduce(240), // *, reduce: MlrvalOrFunction
			reduce(240), // /, reduce: MlrvalOrFunction
			reduce(240), // //, reduce: MlrvalOrFunction
			reduce(240), // %, reduce: MlrvalOrFunction
			reduce(240), // .*, reduce: MlrvalOrFunction
			reduce(240), // ./, reduce: MlrvalOrFunction
			reduce(240), // .//, reduce: MlrvalOrFunction
			reduce(240), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(240), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(240), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(240), // [[, reduce: MlrvalOrFunction
			reduce(240), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1299
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(241), // >, reduce: ContextVariable
			reduce(241), // >>, reduce: ContextVariable
			reduce(241), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(241), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(241), // ?, reduce: ContextVariable
			nil,         // :
			reduce(241), // ||, reduce: ContextVariable
			reduce(241), // ^^, reduce: ContextVariable
			reduce(241), // &&, reduce: ContextVariable
			reduce(241), // ??, reduce: ContextVariable
			reduce(241), // ???, reduce: ContextVariable
			reduce(241), // =~, reduce: ContextVariable
			reduce(241), // !=~, reduce: ContextVariable
			reduce(241), // ==, reduce: ContextVariable
			reduce(241), // !=, reduce: ContextVariable
			reduce(241), // <=>, reduce: ContextVariable
			reduce(241), // >=, reduce: ContextVariable
			reduce(241), // <, reduce: ContextVariable
			reduce(241), // <=, reduce: ContextVariable
			reduce(241), // ^, reduce: ContextVariable
			reduce(241), // &, reduce: ContextVariable
			reduce(241), // <<, reduce: ContextVariable
			reduce(241), // >>>, reduce: ContextVariable
			reduce(241), // +, reduce: ContextVariable
			reduce(241), // -, reduce: ContextVariable
			reduce(241), // .+, reduce: ContextVariable
			reduce(241), // .-, reduce: ContextVariable
			reduce(241), // *, reduce: ContextVariable
			reduce(241), // /, reduce: ContextVariable
			reduce(241), // //, reduce: ContextVariable
			reduce(241), // %, reduce: ContextVariable
			reduce(241), // .*, reduce: ContextVariable
			reduce(241), // ./, reduce: ContextVariable
			reduce(241), // .//, reduce: ContextVariable
			reduce(241), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(241), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(241), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(241), // [[, reduce: ContextVariable
			reduce(241), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1300
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(242), // >, reduce: ContextVariable
			reduce(242), // >>, reduce: ContextVariable
			reduce(242), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(242), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(242), // ?, reduce: ContextVariable
			nil,         // :
			reduce(242), // ||, reduce: ContextVariable
			reduce(242), // ^^, reduce: ContextVariable
			reduce(242), // &&, reduce: ContextVariable
			reduce(242), // ??, reduce: ContextVariable
			reduce(242), // ???, reduce: ContextVariable
			reduce(242), // =~, reduce: ContextVariable
			reduce(242), // !=~, reduce: ContextVariable
			reduce(242), // ==, reduce: ContextVariable
			reduce(242), // !=, reduce: ContextVariable
			reduce(242), // <=>, reduce: ContextVariable
			reduce(242), // >=, reduce: ContextVariable
			reduce(242), // <, reduce: ContextVariable
			reduce(242), // <=, reduce: ContextVariable
			reduce(242), // ^, reduce: ContextVariable
			reduce(242), // &, reduce: ContextVariable
			reduce(242), // <<, reduce: ContextVariable
			reduce(242), // >>>, reduce: ContextVariable
			reduce(242), // +, reduce: ContextVariable
			reduce(242), // -, reduce: ContextVariable
			reduce(242), // .+, reduce: ContextVariable
			reduce(242), // .-, reduce: ContextVariable
			reduce(242), // *, reduce: ContextVariable
			reduce(242), // /, reduce: ContextVariable
			reduce(242), // //, reduce: ContextVariable
			reduce(242), // %, reduce: ContextVariable
			reduce(242), // .*, reduce: ContextVariable
			reduce(242), // ./, reduce: ContextVariable
			reduce(242), // .//, reduce: ContextVariable
			reduce(242), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(242), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(242), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(242), // [[, reduce: ContextVariable
			reduce(242), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1301
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(243), // >, reduce: ContextVariable
			reduce(243), // >>, reduce: ContextVariable
			reduce(243), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(243), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(243), // ?, reduce: ContextVariable
			nil,         // :
			reduce(243), // ||, reduce: ContextVariable
			reduce(243), // ^^, reduce: ContextVariable
			reduce(243), // &&, reduce: ContextVariable
			reduce(243), // ??, reduce: ContextVariable
			reduce(243), // ???, reduce: ContextVariable
			reduce(243), // =~, reduce: ContextVariable
			reduce(243), // !=~, reduce: ContextVariable
			reduce(243), // ==, reduce: ContextVariable
			reduce(243), // !=, reduce: ContextVariable
			reduce(243), // <=>, reduce: ContextVariable
			reduce(243), // >=, reduce: ContextVariable
			reduce(243), // <, reduce: ContextVariable
			reduce(243), // <=, reduce: ContextVariable
			reduce(243), // ^, reduce: ContextVariable
			reduce(243), // &, reduce: ContextVariable
			reduce(243), // <<, reduce: ContextVariable
			reduce(243), // >>>, reduce: ContextVariable
			reduce(243), // +, reduce: ContextVariable
			reduce(243), // -, reduce: ContextVariable
			reduce(243), // .+, reduce: ContextVariable
			reduce(243), // .-, reduce: ContextVariable
			reduce(243), // *, reduce: ContextVariable
			reduce(243), // /, reduce: ContextVariable
			reduce(243), // //, reduce: ContextVariable
			reduce(243), // %, reduce: ContextVariable
			reduce(243), // .*, reduce: ContextVariable
			reduce(243), // ./, reduce: ContextVariable
			reduce(243), // .//, reduce: ContextVariable
			reduce(243), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(243), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(243), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(243), // [[, reduce: ContextVariable
			reduce(243), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1302
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(244), // >, reduce: ContextVariable
			reduce(244), // >>, reduce: ContextVariable
			reduce(244), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(244), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(244), // ?, reduce: ContextVariable
			nil,         // :
			reduce(244), // ||, reduce: ContextVariable
			reduce(244), // ^^, reduce: ContextVariable
			reduce(244), // &&, reduce: ContextVariable
			reduce(244), // ??, reduce: ContextVariable
			reduce(244), // ???, reduce: ContextVariable
			reduce(244), // =~, reduce: ContextVariable
			reduce(244), // !=~, reduce: ContextVariable
			reduce(244), // ==, reduce: ContextVariable
			reduce(244), // !=, reduce: ContextVariable
			reduce(244), // <=>, reduce: ContextVariable
			reduce(244), // >=, reduce: ContextVariable
			reduce(244), // <, reduce: ContextVariable
			reduce(244), // <=, reduce: ContextVariable
			reduce(244), // ^, reduce: ContextVariable
			reduce(244), // &, reduce: ContextVariable
			reduce(244), // <<, reduce: ContextVariable
			reduce(244), // >>>, reduce: ContextVariable
			reduce(244), // +, reduce: ContextVariable
			reduce(244), // -, reduce: ContextVariable
			reduce(244), // .+, reduce: ContextVariable
			reduce(244), // .-, reduce: ContextVariable
			reduce(244), // *, reduce: ContextVariable
			reduce(244), // /, reduce: ContextVariable
			reduce(244), // //, reduce: ContextVariable
			reduce(244), // %, reduce: ContextVariable
			reduce(244), // .*, reduce: ContextVariable
			reduce(244), // ./, reduce: ContextVariable
			reduce(244), // .//, reduce: ContextVariable
			reduce(244), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(244), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(244), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(244), // [[, reduce: ContextVariable
			reduce(244), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1303
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(245), // >, reduce: ContextVariable
			reduce(245), // >>, reduce: ContextVariable
			reduce(245), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(245), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(245), // ?, reduce: ContextVariable
			nil,         // :
			reduce(245), // ||, reduce: ContextVariable
			reduce(245), // ^^, reduce: ContextVariable
			reduce(245), // &&, reduce: ContextVariable
			reduce(245), // ??, reduce: ContextVariable
			reduce(245), // ???, reduce: ContextVariable
			reduce(245), // =~, reduce: ContextVariable
			reduce(245), // !=~, reduce: ContextVariable
			reduce(245), // ==, reduce: ContextVariable
			reduce(245), // !=, reduce: ContextVariable
			reduce(245), // <=>, reduce: ContextVariable
			reduce(245), // >=, reduce: ContextVariable
			reduce(245), // <, reduce: ContextVariable
			reduce(245), // <=, reduce: ContextVariable
			reduce(245), // ^, reduce: ContextVariable
			reduce(245), // &, reduce: ContextVariable
			reduce(245), // <<, reduce: ContextVariable
			reduce(245), // >>>, reduce: ContextVariable
			reduce(245), // +, reduce: ContextVariable
			reduce(245), // -, reduce: ContextVariable
			reduce(245), // .+, reduce: ContextVariable
			reduce(245), // .-, reduce: ContextVariable
			reduce(245), // *, reduce: ContextVariable
			reduce(245), // /, reduce: ContextVariable
			reduce(245), // //, reduce: ContextVariable
			reduce(245), // %, reduce: ContextVariable
			reduce(245), // .*, reduce: ContextVariable
			reduce(245), // ./, reduce: ContextVariable
			reduce(245), // .//, reduce: ContextVariable
			reduce(245), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(245), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(245), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(245), // [[, reduce: ContextVariable
			reduce(245), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1304
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(246), // >, reduce: ContextVariable
			reduce(246), // >>, reduce: ContextVariable
			reduce(246), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(246), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(246), // ?, reduce: ContextVariable
			nil,         // :
			reduce(246), // ||, reduce: ContextVariable
			reduce(246), // ^^, reduce: ContextVariable
			reduce(246), // &&, reduce: ContextVariable
			reduce(246), // ??, reduce: ContextVariable
			reduce(246), // ???, reduce: ContextVariable
			reduce(246), // =~, reduce: ContextVariable
			reduce(246), // !=~, reduce: ContextVariable
			reduce(246), // ==, reduce: ContextVariable
			reduce(246), // !=, reduce: ContextVariable
			reduce(246), // <=>, reduce: ContextVariable
			reduce(246), // >=, reduce: ContextVariable
			reduce(246), // <, reduce: ContextVariable
			reduce(246), // <=, reduce: ContextVariable
			reduce(246), // ^, reduce: ContextVariable
			reduce(246), // &, reduce: ContextVariable
			reduce(246), // <<, reduce: ContextVariable
			reduce(246), // >>>, reduce: ContextVariable
			reduce(246), // +, reduce: ContextVariable
			reduce(246), // -, reduce: ContextVariable
			reduce(246), // .+, reduce: ContextVariable
			reduce(246), // .-, reduce: ContextVariable
			reduce(246), // *, reduce: ContextVariable
			reduce(246), // /, reduce: ContextVariable
			reduce(246), // //, reduce: ContextVariable
			reduce(246), // %, reduce: ContextVariable
			reduce(246), // .*, reduce: ContextVariable
			reduce(246), // ./, reduce: ContextVariable
			reduce(246), // .//, reduce: ContextVariable
			reduce(246), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(246), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(246), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(246), // [[, reduce: ContextVariable
			reduce(246), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1305
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(247), // >, reduce: ContextVariable
			reduce(247), // >>, reduce: ContextVariable
			reduce(247), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(247), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(247), // ?, reduce: ContextVariable
			nil,         // :
			reduce(247), // ||, reduce: ContextVariable
			reduce(247), // ^^, reduce: ContextVariable
			reduce(247), // &&, reduce: ContextVariable
			reduce(247), // ??, reduce: ContextVariable
			reduce(247), // ???, reduce: ContextVariable
			reduce(247), // =~, reduce: ContextVariable
			reduce(247), // !=~, reduce: ContextVariable
			reduce(247), // ==, reduce: ContextVariable
			reduce(247), // !=, reduce: ContextVariable
			reduce(247), // <=>, reduce: ContextVariable
			reduce(247), // >=, reduce: ContextVariable
			reduce(247), // <, reduce: ContextVariable
			reduce(247), // <=, reduce: ContextVariable
			reduce(247), // ^, reduce: ContextVariable
			reduce(247), // &, reduce: ContextVariable
			reduce(247), // <<, reduce: ContextVariable
			reduce(247), // >>>, reduce: ContextVariable
			reduce(247), // +, reduce: ContextVariable
			reduce(247), // -, reduce: ContextVariable
			reduce(247), // .+, reduce: ContextVariable
			reduce(247), // .-, reduce: ContextVariable
			reduce(247), // *, reduce: ContextVariable
			reduce(247), // /, reduce: ContextVariable
			reduce(247), // //, reduce: ContextVariable
			reduce(247), // %, reduce: ContextVariable
			reduce(247), // .*, reduce: ContextVariable
			reduce(247), // ./, reduce: ContextVariable
			reduce(247), // .//, reduce: ContextVariable
			reduce(247), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(247), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(247), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(247), // [[, reduce: ContextVariable
			reduce(247), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1306
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(248), // >, reduce: ContextVariable
			reduce(248), // >>, reduce: ContextVariable
			reduce(248), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(248), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(248), // ?, reduce: ContextVariable
			nil,         // :
			reduce(248), // ||, reduce: ContextVariable
			reduce(248), // ^^, reduce: ContextVariable
			reduce(248), // &&, reduce: ContextVariable
			reduce(248), // ??, reduce: ContextVariable
			reduce(248), // ???, reduce: ContextVariable
			reduce(248), // =~, reduce: ContextVariable
			reduce(248), // !=~, reduce: ContextVariable
			reduce(248), // ==, reduce: ContextVariable
			reduce(248), // !=, reduce: ContextVariable
			reduce(248), // <=>, reduce: ContextVariable
			reduce(248), // >=, reduce: ContextVariable
			reduce(248), // <, reduce: ContextVariable
			reduce(248), // <=, reduce: ContextVariable
			reduce(248), // ^, reduce: ContextVariable
			reduce(248), // &, reduce: ContextVariable
			reduce(248), // <<, reduce: ContextVariable
			reduce(248), // >>>, reduce: ContextVariable
			reduce(248), // +, reduce: ContextVariable
			reduce(248), // -, reduce: ContextVariable
			reduce(248), // .+, reduce: ContextVariable
			reduce(248), // .-, reduce: ContextVariable
			reduce(248), // *, reduce: ContextVariable
			reduce(248), // /, reduce: ContextVariable
			reduce(248), // //, reduce: ContextVariable
			reduce(248), // %, reduce: ContextVariable
			reduce(248), // .*, reduce: ContextVariable
			reduce(248), // ./, reduce: ContextVariable
			reduce(248), // .//, reduce: ContextVariable
			reduce(248), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(248), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(248), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(248), // [[, reduce: ContextVariable
			reduce(248), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1307
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(249), // >, reduce: ContextVariable
			reduce(249), // >>, reduce: ContextVariable
			reduce(249), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(249), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(249), // ?, reduce: ContextVariable
			nil,         // :
			reduce(249), // ||, reduce: ContextVariable
			reduce(249), // ^^, reduce: ContextVariable
			reduce(249), // &&, reduce: ContextVariable
			reduce(249), // ??, reduce: ContextVariable
			reduce(249), // ???, reduce: ContextVariable
			reduce(249), // =~, reduce: ContextVariable
			reduce(249), // !=~, reduce: ContextVariable
			reduce(249), // ==, reduce: ContextVariable
			reduce(249), // !=, reduce: ContextVariable
			reduce(249), // <=>, reduce: ContextVariable
			reduce(249), // >=, reduce: ContextVariable
			reduce(249), // <, reduce: ContextVariable
			reduce(249), // <=, reduce: ContextVariable
			reduce(249), // ^, reduce: ContextVariable
			reduce(249), // &, reduce: ContextVariable
			reduce(249), // <<, reduce: ContextVariable
			reduce(249), // >>>, reduce: ContextVariable
			reduce(249), // +, reduce: ContextVariable
			reduce(249), // -, reduce: ContextVariable
			reduce(249), // .+, reduce: ContextVariable
			reduce(249), // .-, reduce: ContextVariable
			reduce(249), // *, reduce: ContextVariable
			reduce(249), // /, reduce: ContextVariable
			reduce(249), // //, reduce: ContextVariable
			reduce(249), // %, reduce: ContextVariable
			reduce(249), // .*, reduce: ContextVariable
			reduce(249), // ./, reduce: ContextVariable
			reduce(249), // .//, reduce: ContextVariable
			reduce(249), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(249), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(249), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(249), // [[, reduce: ContextVariable
			reduce(249), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1308
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(250), // >, reduce: ContextVariable
			reduce(250), // >>, reduce: ContextVariable
			reduce(250), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(250), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(250), // ?, reduce: ContextVariable
			nil,         // :
			reduce(250), // ||, reduce: ContextVariable
			reduce(250), // ^^, reduce: ContextVariable
			reduce(250), // &&, reduce: ContextVariable
			reduce(250), // ??, reduce: ContextVariable
			reduce(250), // ???, reduce: ContextVariable
			reduce(250), // =~, reduce: ContextVariable
			reduce(250), // !=~, reduce: ContextVariable
			reduce(250), // ==, reduce: ContextVariable
			reduce(250), // !=, reduce: ContextVariable
			reduce(250), // <=>, reduce: ContextVariable
			reduce(250), // >=, reduce: ContextVariable
			reduce(250), // <, reduce: ContextVariable
			reduce(250), // <=, reduce: ContextVariable
			reduce(250), // ^, reduce: ContextVariable
			reduce(250), // &, reduce: ContextVariable
			reduce(250), // <<, reduce: ContextVariable
			reduce(250), // >>>, reduce: ContextVariable
			reduce(250), // +, reduce: ContextVariable
			reduce(250), // -, reduce: ContextVariable
			reduce(250), // .+, reduce: ContextVariable
			reduce(250), // .-, reduce: ContextVariable
			reduce(250), // *, reduce: ContextVariable
			reduce(250), // /, reduce: ContextVariable
			reduce(250), // //, reduce: ContextVariable
			reduce(250), // %, reduce: ContextVariable
			reduce(250), // .*, reduce: ContextVariable
			reduce(250), // ./, reduce: ContextVariable
			reduce(250), // .//, reduce: ContextVariable
			reduce(250), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(250), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(250), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(250), // [[, reduce: ContextVariable
			reduce(250), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1309
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(251), // >, reduce: ContextVariable
			reduce(251), // >>, reduce: ContextVariable
			reduce(251), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(251), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(251), // ?, reduce: ContextVariable
			nil,         // :
			reduce(251), // ||, reduce: ContextVariable
			reduce(251), // ^^, reduce: ContextVariable
			reduce(251), // &&, reduce: ContextVariable
			reduce(251), // ??, reduce: ContextVariable
			reduce(251), // ???, reduce: ContextVariable
			reduce(251), // =~, reduce: ContextVariable
			reduce(251), // !=~, reduce: ContextVariable
			reduce(251), // ==, reduce: ContextVariable
			reduce(251), // !=, reduce: ContextVariable
			reduce(251), // <=>, reduce: ContextVariable
			reduce(251), // >=, reduce: ContextVariable
			reduce(251), // <, reduce: ContextVariable
			reduce(251), // <=, reduce: ContextVariable
			reduce(251), // ^, reduce: ContextVariable
			reduce(251), // &, reduce: ContextVariable
			reduce(251), // <<, reduce: ContextVariable
			reduce(251), // >>>, reduce: ContextVariable
			reduce(251), // +, reduce: ContextVariable
			reduce(251), // -, reduce: ContextVariable
			reduce(251), // .+, reduce: ContextVariable
			reduce(251), // .-, reduce: ContextVariable
			reduce(251), // *, reduce: ContextVariable
			reduce(251), // /, reduce: ContextVariable
			reduce(251), // //, reduce: ContextVariable
			reduce(251), // %, reduce: ContextVariable
			reduce(251), // .*, reduce: ContextVariable
			reduce(251), // ./, reduce: ContextVariable
			reduce(251), // .//, reduce: ContextVariable
			reduce(251), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(251), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(251), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(251), // [[, reduce: ContextVariable
			reduce(251), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1310
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(252), // >, reduce: ContextVariable
			reduce(252), // >>, reduce: ContextVariable
			reduce(252), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(252), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(252), // ?, reduce: ContextVariable
			nil,         // :
			reduce(252), // ||, reduce: ContextVariable
			reduce(252), // ^^, reduce: ContextVariable
			reduce(252), // &&, reduce: ContextVariable
			reduce(252), // ??, reduce: ContextVariable
			reduce(252), // ???, reduce: ContextVariable
			reduce(252), // =~, reduce: ContextVariable
			reduce(252), // !=~, reduce: ContextVariable
			reduce(252), // ==, reduce: ContextVariable
			reduce(252), // !=, reduce: ContextVariable
			reduce(252), // <=>, reduce: ContextVariable
			reduce(252), // >=, reduce: ContextVariable
			reduce(252), // <, reduce: ContextVariable
			reduce(252), // <=, reduce: ContextVariable
			reduce(252), // ^, reduce: ContextVariable
			reduce(252), // &, reduce: ContextVariable
			reduce(252), // <<, reduce: ContextVariable
			reduce(252), // >>>, reduce: ContextVariable
			reduce(252), // +, reduce: ContextVariable
			reduce(252), // -, reduce: ContextVariable
			reduce(252), // .+, reduce: ContextVariable
			reduce(252), // .-, reduce: ContextVariable
			reduce(252), // *, reduce: ContextVariable
			reduce(252), // /, reduce: ContextVariable
			reduce(252), // //, reduce: ContextVariable
			reduce(252), // %, reduce: ContextVariable
			reduce(252), // .*, reduce: ContextVariable
			reduce(252), // ./, reduce: ContextVariable
			reduce(252), // .//, reduce: ContextVariable
			reduce(252), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(252), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(252), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(252), // [[, reduce: ContextVariable
			reduce(252), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1311
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(253), // >, reduce: MlrvalOrFunction
			reduce(253), // >>, reduce: MlrvalOrFunction
			reduce(253), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(253), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(253), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(253), // ||, reduce: MlrvalOrFunction
			reduce(253), // ^^, reduce: MlrvalOrFunction
			reduce(253), // &&, reduce: MlrvalOrFunction
			reduce(253), // ??, reduce: MlrvalOrFunction
			reduce(253), // ???, reduce: MlrvalOrFunction
			reduce(253), // =~, reduce: MlrvalOrFunction
			reduce(253), // !=~, reduce: MlrvalOrFunction
			reduce(253), // ==, reduce: MlrvalOrFunction
			reduce(253), // !=, reduce: MlrvalOrFunction
			reduce(253), // <=>, reduce: MlrvalOrFunction
			reduce(253), // >=, reduce: MlrvalOrFunction
			reduce(253), // <, reduce: MlrvalOrFunction
			reduce(253), // <=, reduce: MlrvalOrFunction
			reduce(253), // ^, reduce: MlrvalOrFunction
			reduce(253), // &, reduce: MlrvalOrFunction
			reduce(253), // <<, reduce: MlrvalOrFunction
			reduce(253), // >>>, reduce: MlrvalOrFunction
			reduce(253), // +, reduce: MlrvalOrFunction
			reduce(253), // -, reduce: MlrvalOrFunction
			reduce(253), // .+, reduce: MlrvalOrFunction
			reduce(253), // .-, reduce: MlrvalOrFunction
			reduce(253), // *, reduce: MlrvalOrFunction
			reduce(253), // /, reduce: MlrvalOrFunction
			reduce(253), // //, reduce: MlrvalOrFunction
			reduce(253), // %, reduce: MlrvalOrFunction
			reduce(253), // .*, reduce: MlrvalOrFunction
			reduce(253), // ./, reduce: MlrvalOrFunction
			reduce(253), // .//, reduce: MlrvalOrFunction
			reduce(253), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(253), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(253), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(253), // [[, reduce: MlrvalOrFunction
			reduce(253), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1312
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			shift(2502), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(2503), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1313
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(256), // >, reduce: MlrvalOrFunction
			reduce(256), // >>, reduce: MlrvalOrFunction
			reduce(256), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(256), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(256), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(256), // ||, reduce: MlrvalOrFunction
			reduce(256), // ^^, reduce: MlrvalOrFunction
			reduce(256), // &&, reduce: MlrvalOrFunction
			reduce(256), // ??, reduce: MlrvalOrFunction
			reduce(256), // ???, reduce: MlrvalOrFunction
			reduce(256), // =~, reduce: MlrvalOrFunction
			reduce(256), // !=~, reduce: MlrvalOrFunction
			reduce(256), // ==, reduce: MlrvalOrFunction
			reduce(256), // !=, reduce: MlrvalOrFunction
			reduce(256), // <=>, reduce: MlrvalOrFunction
			reduce(256), // >=, reduce: MlrvalOrFunction
			reduce(256), // <, reduce: MlrvalOrFunction
			reduce(256), // <=, reduce: MlrvalOrFunction
			reduce(256), // ^, reduce: MlrvalOrFunction
			reduce(256), // &, reduce: MlrvalOrFunction
			reduce(256), // <<, reduce: MlrvalOrFunction
			reduce(256), // >>>, reduce: MlrvalOrFunction
			reduce(256), // +, reduce: MlrvalOrFunction
			reduce(256), // -, reduce: MlrvalOrFunction
			reduce(256), // .+, reduce: MlrvalOrFunction
			reduce(256), // .-, reduce: MlrvalOrFunction
			reduce(256), // *, reduce: MlrvalOrFunction
			reduce(256), // /, reduce: MlrvalOrFunction
			reduce(256), // //, reduce: MlrvalOrFunction
			reduce(256), // %, reduce: MlrvalOrFunction
			reduce(256), // .*, reduce: MlrvalOrFunction
			reduce(256), // ./, reduce: MlrvalOrFunction
			reduce(256), // .//, reduce: MlrvalOrFunction
			reduce(256), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(256), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(256), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(256), // [[, reduce: MlrvalOrFunction
			reduce(256), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1314
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(257), // >, reduce: MlrvalOrFunction
			reduce(257), // >>, reduce: MlrvalOrFunction
			reduce(257), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(257), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(257), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(257), // ||, reduce: MlrvalOrFunction
			reduce(257), // ^^, reduce: MlrvalOrFunction
			reduce(257), // &&, reduce: MlrvalOrFunction
			reduce(257), // ??, reduce: MlrvalOrFunction
			reduce(257), // ???, reduce: MlrvalOrFunction
			reduce(257), // =~, reduce: MlrvalOrFunction
			reduce(257), // !=~, reduce: MlrvalOrFunction
			reduce(257), // ==, reduce: MlrvalOrFunction
			reduce(257), // !=, reduce: MlrvalOrFunction
			reduce(257), // <=>, reduce: MlrvalOrFunction
			reduce(257), // >=, reduce: MlrvalOrFunction
			reduce(257), // <, reduce: MlrvalOrFunction
			reduce(257), // <=, reduce: MlrvalOrFunction
			reduce(257), // ^, reduce: MlrvalOrFunction
			reduce(257), // &, reduce: MlrvalOrFunction
			reduce(257), // <<, reduce: MlrvalOrFunction
			reduce(257), // >>>, reduce: MlrvalOrFunction
			reduce(257), // +, reduce: MlrvalOrFunction
			reduce(257), // -, reduce: MlrvalOrFunction
			reduce(257), // .+, reduce: MlrvalOrFunction
			reduce(257), // .-, reduce: MlrvalOrFunction
			reduce(257), // *, reduce: MlrvalOrFunction
			reduce(257), // /, reduce: MlrvalOrFunction
			reduce(257), // //, reduce: MlrvalOrFunction
			reduce(257), // %, reduce: MlrvalOrFunction
			reduce(257), // .*, reduce: MlrvalOrFunction
			reduce(257), // ./, reduce: MlrvalOrFunction
			reduce(257), // .//, reduce: MlrvalOrFunction
			reduce(257), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(257), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(257), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(257), // [[, reduce: MlrvalOrFunction
			reduce(257), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1315
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(258), // >, reduce: MlrvalOrFunction
			reduce(258), // >>, reduce: MlrvalOrFunction
			reduce(258), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(258), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(258), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(258), // ||, reduce: MlrvalOrFunction
			reduce(258), // ^^, reduce: MlrvalOrFunction
			reduce(258), // &&, reduce: MlrvalOrFunction
			reduce(258), // ??, reduce: MlrvalOrFunction
			reduce(258), // ???, reduce: MlrvalOrFunction
			reduce(258), // =~, reduce: MlrvalOrFunction
			reduce(258), // !=~, reduce: MlrvalOrFunction
			reduce(258), // ==, reduce: MlrvalOrFunction
			reduce(258), // !=, reduce: MlrvalOrFunction
			reduce(258), // <=>, reduce: MlrvalOrFunction
			reduce(258), // >=, reduce: MlrvalOrFunction
			reduce(258), // <, reduce: MlrvalOrFunction
			reduce(258), // <=, reduce: MlrvalOrFunction
			reduce(258), // ^, reduce: MlrvalOrFunction
			reduce(258), // &, reduce: MlrvalOrFunction
			reduce(258), // <<, reduce: MlrvalOrFunction
			reduce(258), // >>>, reduce: MlrvalOrFunction
			reduce(258), // +, reduce: MlrvalOrFunction
			reduce(258), // -, reduce: MlrvalOrFunction
			reduce(258), // .+, reduce: MlrvalOrFunction
			reduce(258), // .-, reduce: MlrvalOrFunction
			reduce(258), // *, reduce: MlrvalOrFunction
			reduce(258), // /, reduce: MlrvalOrFunction
			reduce(258), // //, reduce: MlrvalOrFunction
			reduce(258), // %, reduce: MlrvalOrFunction
			reduce(258), // .*, reduce: MlrvalOrFunction
			reduce(258), // ./, reduce: MlrvalOrFunction
			reduce(258), // .//, reduce: MlrvalOrFunction
			reduce(258), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(258), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(258), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(258), // [[, reduce: MlrvalOrFunction
			reduce(258), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1316
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(259), // >, reduce: MlrvalOrFunction
			reduce(259), // >>, reduce: MlrvalOrFunction
			reduce(259), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(259), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(259), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(259), // ||, reduce: MlrvalOrFunction
			reduce(259), // ^^, reduce: MlrvalOrFunction
			reduce(259), // &&, reduce: MlrvalOrFunction
			reduce(259), // ??, reduce: MlrvalOrFunction
			reduce(259), // ???, reduce: MlrvalOrFunction
			reduce(259), // =~, reduce: MlrvalOrFunction
			reduce(259), // !=~, reduce: MlrvalOrFunction
			reduce(259), // ==, reduce: MlrvalOrFunction
			reduce(259), // !=, reduce: MlrvalOrFunction
			reduce(259), // <=>, reduce: MlrvalOrFunction
			reduce(259), // >=, reduce: MlrvalOrFunction
			reduce(259), // <, reduce: MlrvalOrFunction
			reduce(259), // <=, reduce: MlrvalOrFunction
			reduce(259), // ^, reduce: MlrvalOrFunction
			reduce(259), // &, reduce: MlrvalOrFunction
			reduce(259), // <<, reduce: MlrvalOrFunction
			reduce(259), // >>>, reduce: MlrvalOrFunction
			reduce(259), // +, reduce: MlrvalOrFunction
			reduce(259), // -, reduce: MlrvalOrFunction
			reduce(259), // .+, reduce: MlrvalOrFunction
			reduce(259), // .-, reduce: MlrvalOrFunction
			reduce(259), // *, reduce: MlrvalOrFunction
			reduce(259), // /, reduce: MlrvalOrFunction
			reduce(259), // //, reduce: MlrvalOrFunction
			reduce(259), // %, reduce: MlrvalOrFunction
			reduce(259), // .*, reduce: MlrvalOrFunction
			reduce(259), // ./, reduce: MlrvalOrFunction
			reduce(259), // .//, reduce: MlrvalOrFunction
			reduce(259), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(259), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(259), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(259), // [[, reduce: MlrvalOrFunction
			reduce(259), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1317
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(267), // >, reduce: MlrvalOrFunction
			reduce(267), // >>, reduce: MlrvalOrFunction
			reduce(267), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(267), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(267), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(267), // ||, reduce: MlrvalOrFunction
			reduce(267), // ^^, reduce: MlrvalOrFunction
			reduce(267), // &&, reduce: MlrvalOrFunction
			reduce(267), // ??, reduce: MlrvalOrFunction
			reduce(267), // ???, reduce: MlrvalOrFunction
			reduce(267), // =~, reduce: MlrvalOrFunction
			reduce(267), // !=~, reduce: MlrvalOrFunction
			reduce(267), // ==, reduce: MlrvalOrFunction
			reduce(267), // !=, reduce: MlrvalOrFunction
			reduce(267), // <=>, reduce: MlrvalOrFunction
			reduce(267), // >=, reduce: MlrvalOrFunction
			reduce(267), // <, reduce: MlrvalOrFunction
			reduce(267), // <=, reduce: MlrvalOrFunction
			reduce(267), // ^, reduce: MlrvalOrFunction
			reduce(267), // &, reduce: MlrvalOrFunction
			reduce(267), // <<, reduce: MlrvalOrFunction
			reduce(267), // >>>, reduce: MlrvalOrFunction
			reduce(267), // +, reduce: MlrvalOrFunction
			reduce(267), // -, reduce: MlrvalOrFunction
			reduce(267), // .+, reduce: MlrvalOrFunction
			reduce(267), // .-, reduce: MlrvalOrFunction
			reduce(267), // *, reduce: MlrvalOrFunction
			reduce(267), // /, reduce: MlrvalOrFunction
			reduce(267), // //, reduce: MlrvalOrFunction
			reduce(267), // %, reduce: MlrvalOrFunction
			reduce(267), // .*, reduce: MlrvalOrFunction
			reduce(267), // ./, reduce: MlrvalOrFunction
			reduce(267), // .//, reduce: MlrvalOrFunction
			reduce(267), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(267), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(267), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(267), // [[, reduce: MlrvalOrFunction
			reduce(267), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1318
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2504), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1319
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2505), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1320
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(41), // ,, reduce: Redirector
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1321
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(42), // ,, reduce: Redirector
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1322
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(234), // $, reduce: MapLiteral
			nil,         // empty
			reduce(234), // ;, reduce: MapLiteral
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(234), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1323
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(2506), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1324
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(491), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			shift(505), // field_name
			shift(506), // $[
			nil,        // ]
			shift(507), // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			shift(508), // full_srec
			shift(509), // oosvar_name
			shift(510), // @[
			shift(511), // braced_oosvar_name
			shift(512), // full_oosvar
			shift(513), // all
			shift(514), // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1325
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(491), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			shift(505), // field_name
			shift(506), // $[
			nil,        // ]
			shift(507), // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			shift(508), // full_srec
			shift(509), // oosvar_name
			shift(510), // @[
			shift(511), // braced_oosvar_name
			shift(512), // full_oosvar
			shift(513), // all
			shift(514), // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1326
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2509), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1327
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2510), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1328
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(491),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2512), // (
			nil,         // )
			nil,         // emitp
			shift(505),  // field_name
			shift(506),  // $[
			nil,         // ]
			shift(507),  // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(508),  // full_srec
			shift(509),  // oosvar_name
			shift(510),  // @[
			shift(511),  // braced_oosvar_name
			shift(512),  // full_oosvar
			shift(513),  // all
			shift(514),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1329
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1330
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			shift(2515), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1331
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(87), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			reduce(87), // ), reduce: Emittable
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1332
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(94), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			reduce(94), // ), reduce: Emittable
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1333
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(2517), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1334
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(2518), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(84),  // ), reduce: EmittableList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1335
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(88), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			reduce(88), // ), reduce: Emittable
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1336
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(89), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			reduce(89), // ), reduce: Emittable
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1337
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(90), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			reduce(90), // ), reduce: Emittable
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1338
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(91), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			reduce(91), // ), reduce: Emittable
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1339
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(92), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			reduce(92), // ), reduce: Emittable
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1340
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(93), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			reduce(93), // ), reduce: Emittable
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1341
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(95), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			reduce(95), // ), reduce: Emittable
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1342
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(96), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			reduce(96), // ), reduce: Emittable
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1343
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(104), // ,, reduce: DirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(104), // ), reduce: DirectFieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1344
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1345
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(106), // ,, reduce: BracedFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(106), // ), reduce: BracedFieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1346
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(109), // ,, reduce: FullSrec
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(109), // ), reduce: FullSrec
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1347
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(113), // ,, reduce: DirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(113), // ), reduce: DirectOosvarValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1348
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1349
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(115), // ,, reduce: BracedOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(115), // ), reduce: BracedOosvarValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1350
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(116), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(116), // ), reduce: FullOosvar
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1351
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(117), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(117), // ), reduce: FullOosvar
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1352
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(118), // ,, reduce: LocalVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(118), // ), reduce: LocalVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1353
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(234), // >, reduce: MapLiteral
			reduce(234), // >>, reduce: MapLiteral
			reduce(234), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(234), // ), reduce: MapLiteral
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(234), // ?, reduce: MapLiteral
			nil,         // :
			reduce(234), // ||, reduce: MapLiteral
			reduce(234), // ^^, reduce: MapLiteral
			reduce(234), // &&, reduce: MapLiteral
			reduce(234), // ??, reduce: MapLiteral
			reduce(234), // ???, reduce: MapLiteral
			reduce(234), // =~, reduce: MapLiteral
			reduce(234), // !=~, reduce: MapLiteral
			reduce(234), // ==, reduce: MapLiteral
			reduce(234), // !=, reduce: MapLiteral
			reduce(234), // <=>, reduce: MapLiteral
			reduce(234), // >=, reduce: MapLiteral
			reduce(234), // <, reduce: MapLiteral
			reduce(234), // <=, reduce: MapLiteral
			reduce(234), // ^, reduce: MapLiteral
			reduce(234), // &, reduce: MapLiteral
			reduce(234), // <<, reduce: MapLiteral
			reduce(234), // >>>, reduce: MapLiteral
			reduce(234), // +, reduce: MapLiteral
			reduce(234), // -, reduce: MapLiteral
			reduce(234), // .+, reduce: MapLiteral
			reduce(234), // .-, reduce: MapLiteral
			reduce(234), // *, reduce: MapLiteral
			reduce(234), // /, reduce: MapLiteral
			reduce(234), // //, reduce: MapLiteral
			reduce(234), // %, reduce: MapLiteral
			reduce(234), // .*, reduce: MapLiteral
			reduce(234), // ./, reduce: MapLiteral
			reduce(234), // .//, reduce: MapLiteral
			reduce(234), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(234), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(234), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(234), // [[, reduce: MapLiteral
			reduce(234), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1354
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(2521), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1355
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(208), // $, reduce: PrecedenceChainEnd
			nil,         // empty
			reduce(208), // ;, reduce: PrecedenceChainEnd
			reduce(208), // {, reduce: PrecedenceChainEnd
			nil,         // }
			reduce(208), // =, reduce: PrecedenceChainEnd
			nil,         // unset
			nil,         // filter
			reduce(208), // >, reduce: PrecedenceChainEnd
			reduce(208), // >>, reduce: PrecedenceChainEnd
			reduce(208), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(208), // ||=, reduce: PrecedenceChainEnd
			reduce(208), // ^^=, reduce: PrecedenceChainEnd
			reduce(208), // &&=, reduce: PrecedenceChainEnd
			reduce(208), // ??=, reduce: PrecedenceChainEnd
			reduce(208), // ???=, reduce: PrecedenceChainEnd
			reduce(208), // |=, reduce: PrecedenceChainEnd
			reduce(208), // &=, reduce: PrecedenceChainEnd
			reduce(208), // ^=, reduce: PrecedenceChainEnd
			reduce(208), // <<=, reduce: PrecedenceChainEnd
			reduce(208), // >>=, reduce: PrecedenceChainEnd
			reduce(208), // >>>=, reduce: PrecedenceChainEnd
			reduce(208), // +=, reduce: PrecedenceChainEnd
			reduce(208), // .=, reduce: PrecedenceChainEnd
			reduce(208), // -=, reduce: PrecedenceChainEnd
			reduce(208), // *=, reduce: PrecedenceChainEnd
			reduce(208), // /=, reduce: PrecedenceChainEnd
			reduce(208), // //=, reduce: PrecedenceChainEnd
			reduce(208), // %=, reduce: PrecedenceChainEnd
			reduce(208), // **=, reduce: PrecedenceChainEnd
			reduce(208), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(208), // ||, reduce: PrecedenceChainEnd
			reduce(208), // ^^, reduce: PrecedenceChainEnd
			reduce(208), // &&, reduce: PrecedenceChainEnd
			reduce(208), // ??, reduce: PrecedenceChainEnd
			reduce(208), // ???, reduce: PrecedenceChainEnd
			reduce(208), // =~, reduce: PrecedenceChainEnd
			reduce(208), // !=~, reduce: PrecedenceChainEnd
			reduce(208), // ==, reduce: PrecedenceChainEnd
			reduce(208), // !=, reduce: PrecedenceChainEnd
			reduce(208), // <=>, reduce: PrecedenceChainEnd
			reduce(208), // >=, reduce: PrecedenceChainEnd
			reduce(208), // <, reduce: PrecedenceChainEnd
			reduce(208), // <=, reduce: PrecedenceChainEnd
			reduce(208), // ^, reduce: PrecedenceChainEnd
			reduce(208), // &, reduce: PrecedenceChainEnd
			reduce(208), // <<, reduce: PrecedenceChainEnd
			reduce(208), // >>>, reduce: PrecedenceChainEnd
			reduce(208), // +, reduce: PrecedenceChainEnd
			reduce(208), // -, reduce: PrecedenceChainEnd
			reduce(208), // .+, reduce: PrecedenceChainEnd
			reduce(208), // .-, reduce: PrecedenceChainEnd
			reduce(208), // *, reduce: PrecedenceChainEnd
			reduce(208), // /, reduce: PrecedenceChainEnd
			reduce(208), // //, reduce: PrecedenceChainEnd
			reduce(208), // %, reduce: PrecedenceChainEnd
			reduce(208), // .*, reduce: PrecedenceChainEnd
			reduce(208), // ./, reduce: PrecedenceChainEnd
			reduce(208), // .//, reduce: PrecedenceChainEnd
			reduce(208), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(208), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1356
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(2522), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1357
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2523), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1358
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2524), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1359
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2525), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1360
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2526), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1361
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1362
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1363
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1364
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1365
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1366
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1367
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1368
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1369
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1370
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1371
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1372
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1373
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1374
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1375
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1376
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1377
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1378
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1379
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1380
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1381
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1382
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1383
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1384
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1385
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1386
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(197), // >, reduce: UnaryOpTerm
			reduce(197), // >>, reduce: UnaryOpTerm
			reduce(197), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(197), // ), reduce: UnaryOpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(197), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(197), // ||, reduce: UnaryOpTerm
			reduce(197), // ^^, reduce: UnaryOpTerm
			reduce(197), // &&, reduce: UnaryOpTerm
			reduce(197), // ??, reduce: UnaryOpTerm
			reduce(197), // ???, reduce: UnaryOpTerm
			reduce(197), // =~, reduce: UnaryOpTerm
			reduce(197), // !=~, reduce: UnaryOpTerm
			reduce(197), // ==, reduce: UnaryOpTerm
			reduce(197), // !=, reduce: UnaryOpTerm
			reduce(197), // <=>, reduce: UnaryOpTerm
			reduce(197), // >=, reduce: UnaryOpTerm
			reduce(197), // <, reduce: UnaryOpTerm
			reduce(197), // <=, reduce: UnaryOpTerm
			reduce(197), // ^, reduce: UnaryOpTerm
			reduce(197), // &, reduce: UnaryOpTerm
			reduce(197), // <<, reduce: UnaryOpTerm
			reduce(197), // >>>, reduce: UnaryOpTerm
			reduce(197), // +, reduce: UnaryOpTerm
			reduce(197), // -, reduce: UnaryOpTerm
			reduce(197), // .+, reduce: UnaryOpTerm
			reduce(197), // .-, reduce: UnaryOpTerm
			reduce(197), // *, reduce: UnaryOpTerm
			reduce(197), // /, reduce: UnaryOpTerm
			reduce(197), // //, reduce: UnaryOpTerm
			reduce(197), // %, reduce: UnaryOpTerm
			reduce(197), // .*, reduce: UnaryOpTerm
			reduce(197), // ./, reduce: UnaryOpTerm
			reduce(197), // .//, reduce: UnaryOpTerm
			reduce(197), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1387
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1388
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1389
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1390
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1391
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1392
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1393
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1394
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(198), // >, reduce: UnaryOpTerm
			reduce(198), // >>, reduce: UnaryOpTerm
			reduce(198), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(198), // ), reduce: UnaryOpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(198), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(198), // ||, reduce: UnaryOpTerm
			reduce(198), // ^^, reduce: UnaryOpTerm
			reduce(198), // &&, reduce: UnaryOpTerm
			reduce(198), // ??, reduce: UnaryOpTerm
			reduce(198), // ???, reduce: UnaryOpTerm
			reduce(198), // =~, reduce: UnaryOpTerm
			reduce(198), // !=~, reduce: UnaryOpTerm
			reduce(198), // ==, reduce: UnaryOpTerm
			reduce(198), // !=, reduce: UnaryOpTerm
			reduce(198), // <=>, reduce: UnaryOpTerm
			reduce(198), // >=, reduce: UnaryOpTerm
			reduce(198), // <, reduce: UnaryOpTerm
			reduce(198), // <=, reduce: UnaryOpTerm
			reduce(198), // ^, reduce: UnaryOpTerm
			reduce(198), // &, reduce: UnaryOpTerm
			reduce(198), // <<, reduce: UnaryOpTerm
			reduce(198), // >>>, reduce: UnaryOpTerm
			reduce(198), // +, reduce: UnaryOpTerm
			reduce(198), // -, reduce: UnaryOpTerm
			reduce(198), // .+, reduce: UnaryOpTerm
			reduce(198), // .-, reduce: UnaryOpTerm
			reduce(198), // *, reduce: UnaryOpTerm
			reduce(198), // /, reduce: UnaryOpTerm
			reduce(198), // //, reduce: UnaryOpTerm
			reduce(198), // %, reduce: UnaryOpTerm
			reduce(198), // .*, reduce: UnaryOpTerm
			reduce(198), // ./, reduce: UnaryOpTerm
			reduce(198), // .//, reduce: UnaryOpTerm
			reduce(198), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1395
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(199), // >, reduce: UnaryOpTerm
			reduce(199), // >>, reduce: UnaryOpTerm
			reduce(199), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(199), // ), reduce: UnaryOpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(199), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(199), // ||, reduce: UnaryOpTerm
			reduce(199), // ^^, reduce: UnaryOpTerm
			reduce(199), // &&, reduce: UnaryOpTerm
			reduce(199), // ??, reduce: UnaryOpTerm
			reduce(199), // ???, reduce: UnaryOpTerm
			reduce(199), // =~, reduce: UnaryOpTerm
			reduce(199), // !=~, reduce: UnaryOpTerm
			reduce(199), // ==, reduce: UnaryOpTerm
			reduce(199), // !=, reduce: UnaryOpTerm
			reduce(199), // <=>, reduce: UnaryOpTerm
			reduce(199), // >=, reduce: UnaryOpTerm
			reduce(199), // <, reduce: UnaryOpTerm
			reduce(199), // <=, reduce: UnaryOpTerm
			reduce(199), // ^, reduce: UnaryOpTerm
			reduce(199), // &, reduce: UnaryOpTerm
			reduce(199), // <<, reduce: UnaryOpTerm
			reduce(199), // >>>, reduce: UnaryOpTerm
			reduce(199), // +, reduce: UnaryOpTerm
			reduce(199), // -, reduce: UnaryOpTerm
			reduce(199), // .+, reduce: UnaryOpTerm
			reduce(199), // .-, reduce: UnaryOpTerm
			reduce(199), // *, reduce: UnaryOpTerm
			reduce(199), // /, reduce: UnaryOpTerm
			reduce(199), // //, reduce: UnaryOpTerm
			reduce(199), // %, reduce: UnaryOpTerm
			reduce(199), // .*, reduce: UnaryOpTerm
			reduce(199), // ./, reduce: UnaryOpTerm
			reduce(199), // .//, reduce: UnaryOpTerm
			reduce(199), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1396
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(200), // >, reduce: UnaryOpTerm
			reduce(200), // >>, reduce: UnaryOpTerm
			reduce(200), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(200), // ), reduce: UnaryOpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(200), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(200), // ||, reduce: UnaryOpTerm
			reduce(200), // ^^, reduce: UnaryOpTerm
			reduce(200), // &&, reduce: UnaryOpTerm
			reduce(200), // ??, reduce: UnaryOpTerm
			reduce(200), // ???, reduce: UnaryOpTerm
			reduce(200), // =~, reduce: UnaryOpTerm
			reduce(200), // !=~, reduce: UnaryOpTerm
			reduce(200), // ==, reduce: UnaryOpTerm
			reduce(200), // !=, reduce: UnaryOpTerm
			reduce(200), // <=>, reduce: UnaryOpTerm
			reduce(200), // >=, reduce: UnaryOpTerm
			reduce(200), // <, reduce: UnaryOpTerm
			reduce(200), // <=, reduce: UnaryOpTerm
			reduce(200), // ^, reduce: UnaryOpTerm
			reduce(200), // &, reduce: UnaryOpTerm
			reduce(200), // <<, reduce: UnaryOpTerm
			reduce(200), // >>>, reduce: UnaryOpTerm
			reduce(200), // +, reduce: UnaryOpTerm
			reduce(200), // -, reduce: UnaryOpTerm
			reduce(200), // .+, reduce: UnaryOpTerm
			reduce(200), // .-, reduce: UnaryOpTerm
			reduce(200), // *, reduce: UnaryOpTerm
			reduce(200), // /, reduce: UnaryOpTerm
			reduce(200), // //, reduce: UnaryOpTerm
			reduce(200), // %, reduce: UnaryOpTerm
			reduce(200), // .*, reduce: UnaryOpTerm
			reduce(200), // ./, reduce: UnaryOpTerm
			reduce(200), // .//, reduce: UnaryOpTerm
			reduce(200), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1397
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1398
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(201), // >, reduce: UnaryOpTerm
			reduce(201), // >>, reduce: UnaryOpTerm
			reduce(201), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(201), // ), reduce: UnaryOpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(201), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(201), // ||, reduce: UnaryOpTerm
			reduce(201), // ^^, reduce: UnaryOpTerm
			reduce(201), // &&, reduce: UnaryOpTerm
			reduce(201), // ??, reduce: UnaryOpTerm
			reduce(201), // ???, reduce: UnaryOpTerm
			reduce(201), // =~, reduce: UnaryOpTerm
			reduce(201), // !=~, reduce: UnaryOpTerm
			reduce(201), // ==, reduce: UnaryOpTerm
			reduce(201), // !=, reduce: UnaryOpTerm
			reduce(201), // <=>, reduce: UnaryOpTerm
			reduce(201), // >=, reduce: UnaryOpTerm
			reduce(201), // <, reduce: UnaryOpTerm
			reduce(201), // <=, reduce: UnaryOpTerm
			reduce(201), // ^, reduce: UnaryOpTerm
			reduce(201), // &, reduce: UnaryOpTerm
			reduce(201), // <<, reduce: UnaryOpTerm
			reduce(201), // >>>, reduce: UnaryOpTerm
			reduce(201), // +, reduce: UnaryOpTerm
			reduce(201), // -, reduce: UnaryOpTerm
			reduce(201), // .+, reduce: UnaryOpTerm
			reduce(201), // .-, reduce: UnaryOpTerm
			reduce(201), // *, reduce: UnaryOpTerm
			reduce(201), // /, reduce: UnaryOpTerm
			reduce(201), // //, reduce: UnaryOpTerm
			reduce(201), // %, reduce: UnaryOpTerm
			reduce(201), // .*, reduce: UnaryOpTerm
			reduce(201), // ./, reduce: UnaryOpTerm
			reduce(201), // .//, reduce: UnaryOpTerm
			reduce(201), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1399
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(202), // >, reduce: UnaryOpTerm
			reduce(202), // >>, reduce: UnaryOpTerm
			reduce(202), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(202), // ), reduce: UnaryOpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(202), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(202), // ||, reduce: UnaryOpTerm
			reduce(202), // ^^, reduce: UnaryOpTerm
			reduce(202), // &&, reduce: UnaryOpTerm
			reduce(202), // ??, reduce: UnaryOpTerm
			reduce(202), // ???, reduce: UnaryOpTerm
			reduce(202), // =~, reduce: UnaryOpTerm
			reduce(202), // !=~, reduce: UnaryOpTerm
			reduce(202), // ==, reduce: UnaryOpTerm
			reduce(202), // !=, reduce: UnaryOpTerm
			reduce(202), // <=>, reduce: UnaryOpTerm
			reduce(202), // >=, reduce: UnaryOpTerm
			reduce(202), // <, reduce: UnaryOpTerm
			reduce(202), // <=, reduce: UnaryOpTerm
			reduce(202), // ^, reduce: UnaryOpTerm
			reduce(202), // &, reduce: UnaryOpTerm
			reduce(202), // <<, reduce: UnaryOpTerm
			reduce(202), // >>>, reduce: UnaryOpTerm
			reduce(202), // +, reduce: UnaryOpTerm
			reduce(202), // -, reduce: UnaryOpTerm
			reduce(202), // .+, reduce: UnaryOpTerm
			reduce(202), // .-, reduce: UnaryOpTerm
			reduce(202), // *, reduce: UnaryOpTerm
			reduce(202), // /, reduce: UnaryOpTerm
			reduce(202), // //, reduce: UnaryOpTerm
			reduce(202), // %, reduce: UnaryOpTerm
			reduce(202), // .*, reduce: UnaryOpTerm
			reduce(202), // ./, reduce: UnaryOpTerm
			reduce(202), // .//, reduce: UnaryOpTerm
			reduce(202), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1400
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(520),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(524),  // (
			nil,         // )
			nil,         // emitp
			shift(536),  // field_name
			shift(537),  // $[
			nil,         // ]
			shift(538),  // braced_field_name
			shift(539),  // $[[
			shift(540),  // $[[[
			shift(541),  // full_srec
			shift(543),  // oosvar_name
			shift(544),  // @[
			shift(545),  // braced_oosvar_name
			shift(546),  // full_oosvar
			shift(547),  // all
			shift(548),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2560), // +
			shift(2561), // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(576),  // string_literal
			shift(577),  // regex_case_insensitive
			shift(578),  // int_literal
			shift(579),  // float_literal
			shift(580),  // boolean_literal
			shift(581),  // null_literal
			shift(582),  // inf_literal
			shift(583),  // nan_literal
			shift(584),  // const_M_PI
			shift(585),  // const_M_E
			shift(586),  // panic
			shift(588),  // [
			shift(590),  // ctx_IPS
			shift(591),  // ctx_IFS
			shift(592),  // ctx_IRS
			shift(593),  // ctx_OPS
			shift(594),  // ctx_OFS
			shift(595),  // ctx_ORS
			shift(596),  // ctx_FLATSEP
			shift(597),  // ctx_NF
			shift(598),  // ctx_NR
			shift(599),  // ctx_FNR
			shift(600),  // ctx_FILENAME
			shift(601),  // ctx_FILENUM
			shift(603),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(610),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1401
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(2564), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1402
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1403
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1404
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(228), // >, reduce: ArrayLiteral
			reduce(228), // >>, reduce: ArrayLiteral
			reduce(228), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(228), // ), reduce: ArrayLiteral
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(228), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(228), // ||, reduce: ArrayLiteral
			reduce(228), // ^^, reduce: ArrayLiteral
			reduce(228), // &&, reduce: ArrayLiteral
			reduce(228), // ??, reduce: ArrayLiteral
			reduce(228), // ???, reduce: ArrayLiteral
			reduce(228), // =~, reduce: ArrayLiteral
			reduce(228), // !=~, reduce: ArrayLiteral
			reduce(228), // ==, reduce: ArrayLiteral
			reduce(228), // !=, reduce: ArrayLiteral
			reduce(228), // <=>, reduce: ArrayLiteral
			reduce(228), // >=, reduce: ArrayLiteral
			reduce(228), // <, reduce: ArrayLiteral
			reduce(228), // <=, reduce: ArrayLiteral
			reduce(228), // ^, reduce: ArrayLiteral
			reduce(228), // &, reduce: ArrayLiteral
			reduce(228), // <<, reduce: ArrayLiteral
			reduce(228), // >>>, reduce: ArrayLiteral
			reduce(228), // +, reduce: ArrayLiteral
			reduce(228), // -, reduce: ArrayLiteral
			reduce(228), // .+, reduce: ArrayLiteral
			reduce(228), // .-, reduce: ArrayLiteral
			reduce(228), // *, reduce: ArrayLiteral
			reduce(228), // /, reduce: ArrayLiteral
			reduce(228), // //, reduce: ArrayLiteral
			reduce(228), // %, reduce: ArrayLiteral
			reduce(228), // .*, reduce: ArrayLiteral
			reduce(228), // ./, reduce: ArrayLiteral
			reduce(228), // .//, reduce: ArrayLiteral
			reduce(228), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(228), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(228), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(228), // [[, reduce: ArrayLiteral
			reduce(228), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1405
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2567), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1406
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(2568), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1407
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1408
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(2571), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1409
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1410
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(491),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2574), // (
			nil,         // )
			nil,         // emitp
			shift(505),  // field_name
			shift(506),  // $[
			nil,         // ]
			shift(507),  // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(508),  // full_srec
			shift(509),  // oosvar_name
			shift(510),  // @[
			shift(511),  // braced_oosvar_name
			shift(512),  // full_oosvar
			shift(513),  // all
			shift(514),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1411
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1412
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(2576), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1413
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(234), // >, reduce: MapLiteral
			reduce(234), // >>, reduce: MapLiteral
			reduce(234), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(234), // ], reduce: MapLiteral
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(234), // ?, reduce: MapLiteral
			nil,         // :
			reduce(234), // ||, reduce: MapLiteral
			reduce(234), // ^^, reduce: MapLiteral
			reduce(234), // &&, reduce: MapLiteral
			reduce(234), // ??, reduce: MapLiteral
			reduce(234), // ???, reduce: MapLiteral
			reduce(234), // =~, reduce: MapLiteral
			reduce(234), // !=~, reduce: MapLiteral
			reduce(234), // ==, reduce: MapLiteral
			reduce(234), // !=, reduce: MapLiteral
			reduce(234), // <=>, reduce: MapLiteral
			reduce(234), // >=, reduce: MapLiteral
			reduce(234), // <, reduce: MapLiteral
			reduce(234), // <=, reduce: MapLiteral
			reduce(234), // ^, reduce: MapLiteral
			reduce(234), // &, reduce: MapLiteral
			reduce(234), // <<, reduce: MapLiteral
			reduce(234), // >>>, reduce: MapLiteral
			reduce(234), // +, reduce: MapLiteral
			reduce(234), // -, reduce: MapLiteral
			reduce(234), // .+, reduce: MapLiteral
			reduce(234), // .-, reduce: MapLiteral
			reduce(234), // *, reduce: MapLiteral
			reduce(234), // /, reduce: MapLiteral
			reduce(234), // //, reduce: MapLiteral
			reduce(234), // %, reduce: MapLiteral
			reduce(234), // .*, reduce: MapLiteral
			reduce(234), // ./, reduce: MapLiteral
			reduce(234), // .//, reduce: MapLiteral
			reduce(234), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(234), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(234), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(234), // [[, reduce: MapLiteral
			reduce(234), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1414
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(2577), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1415
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(105), // $, reduce: IndirectFieldValue
			nil,         // empty
			reduce(105), // ;, reduce: IndirectFieldValue
			reduce(105), // {, reduce: IndirectFieldValue
			nil,         // }
			reduce(105), // =, reduce: IndirectFieldValue
			nil,         // unset
			nil,         // filter
			reduce(105), // >, reduce: IndirectFieldValue
			reduce(105), // >>, reduce: IndirectFieldValue
			reduce(105), // |, reduce: IndirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(105), // ||=, reduce: IndirectFieldValue
			reduce(105), // ^^=, reduce: IndirectFieldValue
			reduce(105), // &&=, reduce: IndirectFieldValue
			reduce(105), // ??=, reduce: IndirectFieldValue
			reduce(105), // ???=, reduce: IndirectFieldValue
			reduce(105), // |=, reduce: IndirectFieldValue
			reduce(105), // &=, reduce: IndirectFieldValue
			reduce(105), // ^=, reduce: IndirectFieldValue
			reduce(105), // <<=, reduce: IndirectFieldValue
			reduce(105), // >>=, reduce: IndirectFieldValue
			reduce(105), // >>>=, reduce: IndirectFieldValue
			reduce(105), // +=, reduce: IndirectFieldValue
			reduce(105), // .=, reduce: IndirectFieldValue
			reduce(105), // -=, reduce: IndirectFieldValue
			reduce(105), // *=, reduce: IndirectFieldValue
			reduce(105), // /=, reduce: IndirectFieldValue
			reduce(105), // //=, reduce: IndirectFieldValue
			reduce(105), // %=, reduce: IndirectFieldValue
			reduce(105), // **=, reduce: IndirectFieldValue
			reduce(105), // ?, reduce: IndirectFieldValue
			nil,         // :
			reduce(105), // ||, reduce: IndirectFieldValue
			reduce(105), // ^^, reduce: IndirectFieldValue
			reduce(105), // &&, reduce: IndirectFieldValue
			reduce(105), // ??, reduce: IndirectFieldValue
			reduce(105), // ???, reduce: IndirectFieldValue
			reduce(105), // =~, reduce: IndirectFieldValue
			reduce(105), // !=~, reduce: IndirectFieldValue
			reduce(105), // ==, reduce: IndirectFieldValue
			reduce(105), // !=, reduce: IndirectFieldValue
			reduce(105), // <=>, reduce: IndirectFieldValue
			reduce(105), // >=, reduce: IndirectFieldValue
			reduce(105), // <, reduce: IndirectFieldValue
			reduce(105), // <=, reduce: IndirectFieldValue
			reduce(105), // ^, reduce: IndirectFieldValue
			reduce(105), // &, reduce: IndirectFieldValue
			reduce(105), // <<, reduce: IndirectFieldValue
			reduce(105), // >>>, reduce: IndirectFieldValue
			reduce(105), // +, reduce: IndirectFieldValue
			reduce(105), // -, reduce: IndirectFieldValue
			reduce(105), // .+, reduce: IndirectFieldValue
			reduce(105), // .-, reduce: IndirectFieldValue
			reduce(105), // *, reduce: IndirectFieldValue
			reduce(105), // /, reduce: IndirectFieldValue
			reduce(105), // //, reduce: IndirectFieldValue
			reduce(105), // %, reduce: IndirectFieldValue
			reduce(105), // .*, reduce: IndirectFieldValue
			reduce(105), // ./, reduce: IndirectFieldValue
			reduce(105), // .//, reduce: IndirectFieldValue
			reduce(105), // ., reduce: IndirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(105), // **, reduce: IndirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(105), // [, reduce: IndirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(105), // [[, reduce: IndirectFieldValue
			reduce(105), // [[[, reduce: IndirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1416
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(2578), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1417
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2579), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1418
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2580), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1419
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2581), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1420
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2582), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1421
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1422
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1423
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1424
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1425
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1426
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1427
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1428
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1429
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1430
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1431
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1432
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1433
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1434
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1435
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1436
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1437
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1438
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1439
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1440
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1441
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1442
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1443
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1444
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1445
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1446
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(197), // >, reduce: UnaryOpTerm
			reduce(197), // >>, reduce: UnaryOpTerm
			reduce(197), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(197), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(197), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(197), // ||, reduce: UnaryOpTerm
			reduce(197), // ^^, reduce: UnaryOpTerm
			reduce(197), // &&, reduce: UnaryOpTerm
			reduce(197), // ??, reduce: UnaryOpTerm
			reduce(197), // ???, reduce: UnaryOpTerm
			reduce(197), // =~, reduce: UnaryOpTerm
			reduce(197), // !=~, reduce: UnaryOpTerm
			reduce(197), // ==, reduce: UnaryOpTerm
			reduce(197), // !=, reduce: UnaryOpTerm
			reduce(197), // <=>, reduce: UnaryOpTerm
			reduce(197), // >=, reduce: UnaryOpTerm
			reduce(197), // <, reduce: UnaryOpTerm
			reduce(197), // <=, reduce: UnaryOpTerm
			reduce(197), // ^, reduce: UnaryOpTerm
			reduce(197), // &, reduce: UnaryOpTerm
			reduce(197), // <<, reduce: UnaryOpTerm
			reduce(197), // >>>, reduce: UnaryOpTerm
			reduce(197), // +, reduce: UnaryOpTerm
			reduce(197), // -, reduce: UnaryOpTerm
			reduce(197), // .+, reduce: UnaryOpTerm
			reduce(197), // .-, reduce: UnaryOpTerm
			reduce(197), // *, reduce: UnaryOpTerm
			reduce(197), // /, reduce: UnaryOpTerm
			reduce(197), // //, reduce: UnaryOpTerm
			reduce(197), // %, reduce: UnaryOpTerm
			reduce(197), // .*, reduce: UnaryOpTerm
			reduce(197), // ./, reduce: UnaryOpTerm
			reduce(197), // .//, reduce: UnaryOpTerm
			reduce(197), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1447
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1448
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1449
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1450
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1451
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1452
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1453
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1454
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(198), // >, reduce: UnaryOpTerm
			reduce(198), // >>, reduce: UnaryOpTerm
			reduce(198), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(198), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(198), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(198), // ||, reduce: UnaryOpTerm
			reduce(198), // ^^, reduce: UnaryOpTerm
			reduce(198), // &&, reduce: UnaryOpTerm
			reduce(198), // ??, reduce: UnaryOpTerm
			reduce(198), // ???, reduce: UnaryOpTerm
			reduce(198), // =~, reduce: UnaryOpTerm
			reduce(198), // !=~, reduce: UnaryOpTerm
			reduce(198), // ==, reduce: UnaryOpTerm
			reduce(198), // !=, reduce: UnaryOpTerm
			reduce(198), // <=>, reduce: UnaryOpTerm
			reduce(198), // >=, reduce: UnaryOpTerm
			reduce(198), // <, reduce: UnaryOpTerm
			reduce(198), // <=, reduce: UnaryOpTerm
			reduce(198), // ^, reduce: UnaryOpTerm
			reduce(198), // &, reduce: UnaryOpTerm
			reduce(198), // <<, reduce: UnaryOpTerm
			reduce(198), // >>>, reduce: UnaryOpTerm
			reduce(198), // +, reduce: UnaryOpTerm
			reduce(198), // -, reduce: UnaryOpTerm
			reduce(198), // .+, reduce: UnaryOpTerm
			reduce(198), // .-, reduce: UnaryOpTerm
			reduce(198), // *, reduce: UnaryOpTerm
			reduce(198), // /, reduce: UnaryOpTerm
			reduce(198), // //, reduce: UnaryOpTerm
			reduce(198), // %, reduce: UnaryOpTerm
			reduce(198), // .*, reduce: UnaryOpTerm
			reduce(198), // ./, reduce: UnaryOpTerm
			reduce(198), // .//, reduce: UnaryOpTerm
			reduce(198), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1455
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(199), // >, reduce: UnaryOpTerm
			reduce(199), // >>, reduce: UnaryOpTerm
			reduce(199), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(199), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(199), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(199), // ||, reduce: UnaryOpTerm
			reduce(199), // ^^, reduce: UnaryOpTerm
			reduce(199), // &&, reduce: UnaryOpTerm
			reduce(199), // ??, reduce: UnaryOpTerm
			reduce(199), // ???, reduce: UnaryOpTerm
			reduce(199), // =~, reduce: UnaryOpTerm
			reduce(199), // !=~, reduce: UnaryOpTerm
			reduce(199), // ==, reduce: UnaryOpTerm
			reduce(199), // !=, reduce: UnaryOpTerm
			reduce(199), // <=>, reduce: UnaryOpTerm
			reduce(199), // >=, reduce: UnaryOpTerm
			reduce(199), // <, reduce: UnaryOpTerm
			reduce(199), // <=, reduce: UnaryOpTerm
			reduce(199), // ^, reduce: UnaryOpTerm
			reduce(199), // &, reduce: UnaryOpTerm
			reduce(199), // <<, reduce: UnaryOpTerm
			reduce(199), // >>>, reduce: UnaryOpTerm
			reduce(199), // +, reduce: UnaryOpTerm
			reduce(199), // -, reduce: UnaryOpTerm
			reduce(199), // .+, reduce: UnaryOpTerm
			reduce(199), // .-, reduce: UnaryOpTerm
			reduce(199), // *, reduce: UnaryOpTerm
			reduce(199), // /, reduce: UnaryOpTerm
			reduce(199), // //, reduce: UnaryOpTerm
			reduce(199), // %, reduce: UnaryOpTerm
			reduce(199), // .*, reduce: UnaryOpTerm
			reduce(199), // ./, reduce: UnaryOpTerm
			reduce(199), // .//, reduce: UnaryOpTerm
			reduce(199), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1456
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(200), // >, reduce: UnaryOpTerm
			reduce(200), // >>, reduce: UnaryOpTerm
			reduce(200), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(200), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(200), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(200), // ||, reduce: UnaryOpTerm
			reduce(200), // ^^, reduce: UnaryOpTerm
			reduce(200), // &&, reduce: UnaryOpTerm
			reduce(200), // ??, reduce: UnaryOpTerm
			reduce(200), // ???, reduce: UnaryOpTerm
			reduce(200), // =~, reduce: UnaryOpTerm
			reduce(200), // !=~, reduce: UnaryOpTerm
			reduce(200), // ==, reduce: UnaryOpTerm
			reduce(200), // !=, reduce: UnaryOpTerm
			reduce(200), // <=>, reduce: UnaryOpTerm
			reduce(200), // >=, reduce: UnaryOpTerm
			reduce(200), // <, reduce: UnaryOpTerm
			reduce(200), // <=, reduce: UnaryOpTerm
			reduce(200), // ^, reduce: UnaryOpTerm
			reduce(200), // &, reduce: UnaryOpTerm
			reduce(200), // <<, reduce: UnaryOpTerm
			reduce(200), // >>>, reduce: UnaryOpTerm
			reduce(200), // +, reduce: UnaryOpTerm
			reduce(200), // -, reduce: UnaryOpTerm
			reduce(200), // .+, reduce: UnaryOpTerm
			reduce(200), // .-, reduce: UnaryOpTerm
			reduce(200), // *, reduce: UnaryOpTerm
			reduce(200), // /, reduce: UnaryOpTerm
			reduce(200), // //, reduce: UnaryOpTerm
			reduce(200), // %, reduce: UnaryOpTerm
			reduce(200), // .*, reduce: UnaryOpTerm
			reduce(200), // ./, reduce: UnaryOpTerm
			reduce(200), // .//, reduce: UnaryOpTerm
			reduce(200), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1457
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1458
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(201), // >, reduce: UnaryOpTerm
			reduce(201), // >>, reduce: UnaryOpTerm
			reduce(201), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(201), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(201), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(201), // ||, reduce: UnaryOpTerm
			reduce(201), // ^^, reduce: UnaryOpTerm
			reduce(201), // &&, reduce: UnaryOpTerm
			reduce(201), // ??, reduce: UnaryOpTerm
			reduce(201), // ???, reduce: UnaryOpTerm
			reduce(201), // =~, reduce: UnaryOpTerm
			reduce(201), // !=~, reduce: UnaryOpTerm
			reduce(201), // ==, reduce: UnaryOpTerm
			reduce(201), // !=, reduce: UnaryOpTerm
			reduce(201), // <=>, reduce: UnaryOpTerm
			reduce(201), // >=, reduce: UnaryOpTerm
			reduce(201), // <, reduce: UnaryOpTerm
			reduce(201), // <=, reduce: UnaryOpTerm
			reduce(201), // ^, reduce: UnaryOpTerm
			reduce(201), // &, reduce: UnaryOpTerm
			reduce(201), // <<, reduce: UnaryOpTerm
			reduce(201), // >>>, reduce: UnaryOpTerm
			reduce(201), // +, reduce: UnaryOpTerm
			reduce(201), // -, reduce: UnaryOpTerm
			reduce(201), // .+, reduce: UnaryOpTerm
			reduce(201), // .-, reduce: UnaryOpTerm
			reduce(201), // *, reduce: UnaryOpTerm
			reduce(201), // /, reduce: UnaryOpTerm
			reduce(201), // //, reduce: UnaryOpTerm
			reduce(201), // %, reduce: UnaryOpTerm
			reduce(201), // .*, reduce: UnaryOpTerm
			reduce(201), // ./, reduce: UnaryOpTerm
			reduce(201), // .//, reduce: UnaryOpTerm
			reduce(201), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1459
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(202), // >, reduce: UnaryOpTerm
			reduce(202), // >>, reduce: UnaryOpTerm
			reduce(202), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(202), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(202), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(202), // ||, reduce: UnaryOpTerm
			reduce(202), // ^^, reduce: UnaryOpTerm
			reduce(202), // &&, reduce: UnaryOpTerm
			reduce(202), // ??, reduce: UnaryOpTerm
			reduce(202), // ???, reduce: UnaryOpTerm
			reduce(202), // =~, reduce: UnaryOpTerm
			reduce(202), // !=~, reduce: UnaryOpTerm
			reduce(202), // ==, reduce: UnaryOpTerm
			reduce(202), // !=, reduce: UnaryOpTerm
			reduce(202), // <=>, reduce: UnaryOpTerm
			reduce(202), // >=, reduce: UnaryOpTerm
			reduce(202), // <, reduce: UnaryOpTerm
			reduce(202), // <=, reduce: UnaryOpTerm
			reduce(202), // ^, reduce: UnaryOpTerm
			reduce(202), // &, reduce: UnaryOpTerm
			reduce(202), // <<, reduce: UnaryOpTerm
			reduce(202), // >>>, reduce: UnaryOpTerm
			reduce(202), // +, reduce: UnaryOpTerm
			reduce(202), // -, reduce: UnaryOpTerm
			reduce(202), // .+, reduce: UnaryOpTerm
			reduce(202), // .-, reduce: UnaryOpTerm
			reduce(202), // *, reduce: UnaryOpTerm
			reduce(202), // /, reduce: UnaryOpTerm
			reduce(202), // //, reduce: UnaryOpTerm
			reduce(202), // %, reduce: UnaryOpTerm
			reduce(202), // .*, reduce: UnaryOpTerm
			reduce(202), // ./, reduce: UnaryOpTerm
			reduce(202), // .//, reduce: UnaryOpTerm
			reduce(202), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1460
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			nil,         // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2616), // +
			shift(2617), // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1461
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(2620), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1462
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1463
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1464
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(228), // >, reduce: ArrayLiteral
			reduce(228), // >>, reduce: ArrayLiteral
			reduce(228), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(228), // ], reduce: ArrayLiteral
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(228), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(228), // ||, reduce: ArrayLiteral
			reduce(228), // ^^, reduce: ArrayLiteral
			reduce(228), // &&, reduce: ArrayLiteral
			reduce(228), // ??, reduce: ArrayLiteral
			reduce(228), // ???, reduce: ArrayLiteral
			reduce(228), // =~, reduce: ArrayLiteral
			reduce(228), // !=~, reduce: ArrayLiteral
			reduce(228), // ==, reduce: ArrayLiteral
			reduce(228), // !=, reduce: ArrayLiteral
			reduce(228), // <=>, reduce: ArrayLiteral
			reduce(228), // >=, reduce: ArrayLiteral
			reduce(228), // <, reduce: ArrayLiteral
			reduce(228), // <=, reduce: ArrayLiteral
			reduce(228), // ^, reduce: ArrayLiteral
			reduce(228), // &, reduce: ArrayLiteral
			reduce(228), // <<, reduce: ArrayLiteral
			reduce(228), // >>>, reduce: ArrayLiteral
			reduce(228), // +, reduce: ArrayLiteral
			reduce(228), // -, reduce: ArrayLiteral
			reduce(228), // .+, reduce: ArrayLiteral
			reduce(228), // .-, reduce: ArrayLiteral
			reduce(228), // *, reduce: ArrayLiteral
			reduce(228), // /, reduce: ArrayLiteral
			reduce(228), // //, reduce: ArrayLiteral
			reduce(228), // %, reduce: ArrayLiteral
			reduce(228), // .*, reduce: ArrayLiteral
			reduce(228), // ./, reduce: ArrayLiteral
			reduce(228), // .//, reduce: ArrayLiteral
			reduce(228), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(228), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(228), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(228), // [[, reduce: ArrayLiteral
			reduce(228), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1465
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2623), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1466
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(2624), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1467
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1468
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(2627), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1469
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1470
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2629), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1471
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2630), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1472
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(114), // $, reduce: IndirectOosvarValue
			nil,         // empty
			reduce(114), // ;, reduce: IndirectOosvarValue
			reduce(114), // {, reduce: IndirectOosvarValue
			nil,         // }
			reduce(114), // =, reduce: IndirectOosvarValue
			nil,         // unset
			nil,         // filter
			reduce(114), // >, reduce: IndirectOosvarValue
			reduce(114), // >>, reduce: IndirectOosvarValue
			reduce(114), // |, reduce: IndirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(114), // ||=, reduce: IndirectOosvarValue
			reduce(114), // ^^=, reduce: IndirectOosvarValue
			reduce(114), // &&=, reduce: IndirectOosvarValue
			reduce(114), // ??=, reduce: IndirectOosvarValue
			reduce(114), // ???=, reduce: IndirectOosvarValue
			reduce(114), // |=, reduce: IndirectOosvarValue
			reduce(114), // &=, reduce: IndirectOosvarValue
			reduce(114), // ^=, reduce: IndirectOosvarValue
			reduce(114), // <<=, reduce: IndirectOosvarValue
			reduce(114), // >>=, reduce: IndirectOosvarValue
			reduce(114), // >>>=, reduce: IndirectOosvarValue
			reduce(114), // +=, reduce: IndirectOosvarValue
			reduce(114), // .=, reduce: IndirectOosvarValue
			reduce(114), // -=, reduce: IndirectOosvarValue
			reduce(114), // *=, reduce: IndirectOosvarValue
			reduce(114), // /=, reduce: IndirectOosvarValue
			reduce(114), // //=, reduce: IndirectOosvarValue
			reduce(114), // %=, reduce: IndirectOosvarValue
			reduce(114), // **=, reduce: IndirectOosvarValue
			reduce(114), // ?, reduce: IndirectOosvarValue
			nil,         // :
			reduce(114), // ||, reduce: IndirectOosvarValue
			reduce(114), // ^^, reduce: IndirectOosvarValue
			reduce(114), // &&, reduce: IndirectOosvarValue
			reduce(114), // ??, reduce: IndirectOosvarValue
			reduce(114), // ???, reduce: IndirectOosvarValue
			reduce(114), // =~, reduce: IndirectOosvarValue
			reduce(114), // !=~, reduce: IndirectOosvarValue
			reduce(114), // ==, reduce: IndirectOosvarValue
			reduce(114), // !=, reduce: IndirectOosvarValue
			reduce(114), // <=>, reduce: IndirectOosvarValue
			reduce(114), // >=, reduce: IndirectOosvarValue
			reduce(114), // <, reduce: IndirectOosvarValue
			reduce(114), // <=, reduce: IndirectOosvarValue
			reduce(114), // ^, reduce: IndirectOosvarValue
			reduce(114), // &, reduce: IndirectOosvarValue
			reduce(114), // <<, reduce: IndirectOosvarValue
			reduce(114), // >>>, reduce: IndirectOosvarValue
			reduce(114), // +, reduce: IndirectOosvarValue
			reduce(114), // -, reduce: IndirectOosvarValue
			reduce(114), // .+, reduce: IndirectOosvarValue
			reduce(114), // .-, reduce: IndirectOosvarValue
			reduce(114), // *, reduce: IndirectOosvarValue
			reduce(114), // /, reduce: IndirectOosvarValue
			reduce(114), // //, reduce: IndirectOosvarValue
			reduce(114), // %, reduce: IndirectOosvarValue
			reduce(114), // .*, reduce: IndirectOosvarValue
			reduce(114), // ./, reduce: IndirectOosvarValue
			reduce(114), // .//, reduce: IndirectOosvarValue
			reduce(114), // ., reduce: IndirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(114), // **, reduce: IndirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(114), // [, reduce: IndirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(114), // [[, reduce: IndirectOosvarValue
			reduce(114), // [[[, reduce: IndirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1473
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(2631), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1474
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(151), // $, reduce: LogicalOrTerm
			nil,         // empty
			reduce(151), // ;, reduce: LogicalOrTerm
			reduce(151), // {, reduce: LogicalOrTerm
			nil,         // }
			reduce(151), // =, reduce: LogicalOrTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(151), // ||=, reduce: LogicalOrTerm
			reduce(151), // ^^=, reduce: LogicalOrTerm
			reduce(151), // &&=, reduce: LogicalOrTerm
			reduce(151), // ??=, reduce: LogicalOrTerm
			reduce(151), // ???=, reduce: LogicalOrTerm
			reduce(151), // |=, reduce: LogicalOrTerm
			reduce(151), // &=, reduce: LogicalOrTerm
			reduce(151), // ^=, reduce: LogicalOrTerm
			reduce(151), // <<=, reduce: LogicalOrTerm
			reduce(151), // >>=, reduce: LogicalOrTerm
			reduce(151), // >>>=, reduce: LogicalOrTerm
			reduce(151), // +=, reduce: LogicalOrTerm
			reduce(151), // .=, reduce: LogicalOrTerm
			reduce(151), // -=, reduce: LogicalOrTerm
			reduce(151), // *=, reduce: LogicalOrTerm
			reduce(151), // /=, reduce: LogicalOrTerm
			reduce(151), // //=, reduce: LogicalOrTerm
			reduce(151), // %=, reduce: LogicalOrTerm
			reduce(151), // **=, reduce: LogicalOrTerm
			reduce(151), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(151), // ||, reduce: LogicalOrTerm
			shift(710),  // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1475
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(153), // $, reduce: LogicalXORTerm
			nil,         // empty
			reduce(153), // ;, reduce: LogicalXORTerm
			reduce(153), // {, reduce: LogicalXORTerm
			nil,         // }
			reduce(153), // =, reduce: LogicalXORTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(153), // ||=, reduce: LogicalXORTerm
			reduce(153), // ^^=, reduce: LogicalXORTerm
			reduce(153), // &&=, reduce: LogicalXORTerm
			reduce(153), // ??=, reduce: LogicalXORTerm
			reduce(153), // ???=, reduce: LogicalXORTerm
			reduce(153), // |=, reduce: LogicalXORTerm
			reduce(153), // &=, reduce: LogicalXORTerm
			reduce(153), // ^=, reduce: LogicalXORTerm
			reduce(153), // <<=, reduce: LogicalXORTerm
			reduce(153), // >>=, reduce: LogicalXORTerm
			reduce(153), // >>>=, reduce: LogicalXORTerm
			reduce(153), // +=, reduce: LogicalXORTerm
			reduce(153), // .=, reduce: LogicalXORTerm
			reduce(153), // -=, reduce: LogicalXORTerm
			reduce(153), // *=, reduce: LogicalXORTerm
			reduce(153), // /=, reduce: LogicalXORTerm
			reduce(153), // //=, reduce: LogicalXORTerm
			reduce(153), // %=, reduce: LogicalXORTerm
			reduce(153), // **=, reduce: LogicalXORTerm
			reduce(153), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(153), // ||, reduce: LogicalXORTerm
			reduce(153), // ^^, reduce: LogicalXORTerm
			shift(711),  // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1476
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(155), // $, reduce: LogicalAndTerm
			nil,         // empty
			reduce(155), // ;, reduce: LogicalAndTerm
			reduce(155), // {, reduce: LogicalAndTerm
			nil,         // }
			reduce(155), // =, reduce: LogicalAndTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(155), // ||=, reduce: LogicalAndTerm
			reduce(155), // ^^=, reduce: LogicalAndTerm
			reduce(155), // &&=, reduce: LogicalAndTerm
			reduce(155), // ??=, reduce: LogicalAndTerm
			reduce(155), // ???=, reduce: LogicalAndTerm
			reduce(155), // |=, reduce: LogicalAndTerm
			reduce(155), // &=, reduce: LogicalAndTerm
			reduce(155), // ^=, reduce: LogicalAndTerm
			reduce(155), // <<=, reduce: LogicalAndTerm
			reduce(155), // >>=, reduce: LogicalAndTerm
			reduce(155), // >>>=, reduce: LogicalAndTerm
			reduce(155), // +=, reduce: LogicalAndTerm
			reduce(155), // .=, reduce: LogicalAndTerm
			reduce(155), // -=, reduce: LogicalAndTerm
			reduce(155), // *=, reduce: LogicalAndTerm
			reduce(155), // /=, reduce: LogicalAndTerm
			reduce(155), // //=, reduce: LogicalAndTerm
			reduce(155), // %=, reduce: LogicalAndTerm
			reduce(155), // **=, reduce: LogicalAndTerm
			reduce(155), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(155), // ||, reduce: LogicalAndTerm
			reduce(155), // ^^, reduce: LogicalAndTerm
			reduce(155), // &&, reduce: LogicalAndTerm
			shift(712),  // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1477
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(157), // $, reduce: AbsentCoalesceTerm
			nil,         // empty
			reduce(157), // ;, reduce: AbsentCoalesceTerm
			reduce(157), // {, reduce: AbsentCoalesceTerm
			nil,         // }
			reduce(157), // =, reduce: AbsentCoalesceTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(157), // ||=, reduce: AbsentCoalesceTerm
			reduce(157), // ^^=, reduce: AbsentCoalesceTerm
			reduce(157), // &&=, reduce: AbsentCoalesceTerm
			reduce(157), // ??=, reduce: AbsentCoalesceTerm
			reduce(157), // ???=, reduce: AbsentCoalesceTerm
			reduce(157), // |=, reduce: AbsentCoalesceTerm
			reduce(157), // &=, reduce: AbsentCoalesceTerm
			reduce(157), // ^=, reduce: AbsentCoalesceTerm
			reduce(157), // <<=, reduce: AbsentCoalesceTerm
			reduce(157), // >>=, reduce: AbsentCoalesceTerm
			reduce(157), // >>>=, reduce: AbsentCoalesceTerm
			reduce(157), // +=, reduce: AbsentCoalesceTerm
			reduce(157), // .=, reduce: AbsentCoalesceTerm
			reduce(157), // -=, reduce: AbsentCoalesceTerm
			reduce(157), // *=, reduce: AbsentCoalesceTerm
			reduce(157), // /=, reduce: AbsentCoalesceTerm
			reduce(157), // //=, reduce: AbsentCoalesceTerm
			reduce(157), // %=, reduce: AbsentCoalesceTerm
			reduce(157), // **=, reduce: AbsentCoalesceTerm
			reduce(157), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(157), // ||, reduce: AbsentCoalesceTerm
			reduce(157), // ^^, reduce: AbsentCoalesceTerm
			reduce(157), // &&, reduce: AbsentCoalesceTerm
			reduce(157), // ??, reduce: AbsentCoalesceTerm
			shift(713),  // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1478
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(159), // $, reduce: EmptyCoalesceTerm
			nil,         // empty
			reduce(159), // ;, reduce: EmptyCoalesceTerm
			reduce(159), // {, reduce: EmptyCoalesceTerm
			nil,         // }
			reduce(159), // =, reduce: EmptyCoalesceTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(159), // ||=, reduce: EmptyCoalesceTerm
			reduce(159), // ^^=, reduce: EmptyCoalesceTerm
			reduce(159), // &&=, reduce: EmptyCoalesceTerm
			reduce(159), // ??=, reduce: EmptyCoalesceTerm
			reduce(159), // ???=, reduce: EmptyCoalesceTerm
			reduce(159), // |=, reduce: EmptyCoalesceTerm
			reduce(159), // &=, reduce: EmptyCoalesceTerm
			reduce(159), // ^=, reduce: EmptyCoalesceTerm
			reduce(159), // <<=, reduce: EmptyCoalesceTerm
			reduce(159), // >>=, reduce: EmptyCoalesceTerm
			reduce(159), // >>>=, reduce: EmptyCoalesceTerm
			reduce(159), // +=, reduce: EmptyCoalesceTerm
			reduce(159), // .=, reduce: EmptyCoalesceTerm
			reduce(159), // -=, reduce: EmptyCoalesceTerm
			reduce(159), // *=, reduce: EmptyCoalesceTerm
			reduce(159), // /=, reduce: EmptyCoalesceTerm
			reduce(159), // //=, reduce: EmptyCoalesceTerm
			reduce(159), // %=, reduce: EmptyCoalesceTerm
			reduce(159), // **=, reduce: EmptyCoalesceTerm
			reduce(159), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(159), // ||, reduce: EmptyCoalesceTerm
			reduce(159), // ^^, reduce: EmptyCoalesceTerm
			reduce(159), // &&, reduce: EmptyCoalesceTerm
			reduce(159), // ??, reduce: EmptyCoalesceTerm
			reduce(159), // ???, reduce: EmptyCoalesceTerm
			shift(714),  // =~
			shift(715),  // !=~
			shift(716),  // ==
			shift(717),  // !=
			shift(718),  // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1479
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(161), // $, reduce: EqneTerm
			nil,         // empty
			reduce(161), // ;, reduce: EqneTerm
			reduce(161), // {, reduce: EqneTerm
			nil,         // }
			reduce(161), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(719),  // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(161), // ||=, reduce: EqneTerm
			reduce(161), // ^^=, reduce: EqneTerm
			reduce(161), // &&=, reduce: EqneTerm
			reduce(161), // ??=, reduce: EqneTerm
			reduce(161), // ???=, reduce: EqneTerm
			reduce(161), // |=, reduce: EqneTerm
			reduce(161), // &=, reduce: EqneTerm
			reduce(161), // ^=, reduce: EqneTerm
			reduce(161), // <<=, reduce: EqneTerm
			reduce(161), // >>=, reduce: EqneTerm
			reduce(161), // >>>=, reduce: EqneTerm
			reduce(161), // +=, reduce: EqneTerm
			reduce(161), // .=, reduce: EqneTerm
			reduce(161), // -=, reduce: EqneTerm
			reduce(161), // *=, reduce: EqneTerm
			reduce(161), // /=, reduce: EqneTerm
			reduce(161), // //=, reduce: EqneTerm
			reduce(161), // %=, reduce: EqneTerm
			reduce(161), // **=, reduce: EqneTerm
			reduce(161), // ?, reduce: EqneTerm
			nil,         // :
			reduce(161), // ||, reduce: EqneTerm
			reduce(161), // ^^, reduce: EqneTerm
			reduce(161), // &&, reduce: EqneTerm
			reduce(161), // ??, reduce: EqneTerm
			reduce(161), // ???, reduce: EqneTerm
			reduce(161), // =~, reduce: EqneTerm
			reduce(161), // !=~, reduce: EqneTerm
			reduce(161), // ==, reduce: EqneTerm
			reduce(161), // !=, reduce: EqneTerm
			reduce(161), // <=>, reduce: EqneTerm
			shift(720),  // >=
			shift(721),  // <
			shift(722),  // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1480
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(162), // $, reduce: EqneTerm
			nil,         // empty
			reduce(162), // ;, reduce: EqneTerm
			reduce(162), // {, reduce: EqneTerm
			nil,         // }
			reduce(162), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(719),  // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(162), // ||=, reduce: EqneTerm
			reduce(162), // ^^=, reduce: EqneTerm
			reduce(162), // &&=, reduce: EqneTerm
			reduce(162), // ??=, reduce: EqneTerm
			reduce(162), // ???=, reduce: EqneTerm
			reduce(162), // |=, reduce: EqneTerm
			reduce(162), // &=, reduce: EqneTerm
			reduce(162), // ^=, reduce: EqneTerm
			reduce(162), // <<=, reduce: EqneTerm
			reduce(162), // >>=, reduce: EqneTerm
			reduce(162), // >>>=, reduce: EqneTerm
			reduce(162), // +=, reduce: EqneTerm
			reduce(162), // .=, reduce: EqneTerm
			reduce(162), // -=, reduce: EqneTerm
			reduce(162), // *=, reduce: EqneTerm
			reduce(162), // /=, reduce: EqneTerm
			reduce(162), // //=, reduce: EqneTerm
			reduce(162), // %=, reduce: EqneTerm
			reduce(162), // **=, reduce: EqneTerm
			reduce(162), // ?, reduce: EqneTerm
			nil,         // :
			reduce(162), // ||, reduce: EqneTerm
			reduce(162), // ^^, reduce: EqneTerm
			reduce(162), // &&, reduce: EqneTerm
			reduce(162), // ??, reduce: EqneTerm
			reduce(162), // ???, reduce: EqneTerm
			reduce(162), // =~, reduce: EqneTerm
			reduce(162), // !=~, reduce: EqneTerm
			reduce(162), // ==, reduce: EqneTerm
			reduce(162), // !=, reduce: EqneTerm
			reduce(162), // <=>, reduce: EqneTerm
			shift(720),  // >=
			shift(721),  // <
			shift(722),  // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1481
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(163), // $, reduce: EqneTerm
			nil,         // empty
			reduce(163), // ;, reduce: EqneTerm
			reduce(163), // {, reduce: EqneTerm
			nil,         // }
			reduce(163), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(719),  // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(163), // ||=, reduce: EqneTerm
			reduce(163), // ^^=, reduce: EqneTerm
			reduce(163), // &&=, reduce: EqneTerm
			reduce(163), // ??=, reduce: EqneTerm
			reduce(163), // ???=, reduce: EqneTerm
			reduce(163), // |=, reduce: EqneTerm
			reduce(163), // &=, reduce: EqneTerm
			reduce(163), // ^=, reduce: EqneTerm
			reduce(163), // <<=, reduce: EqneTerm
			reduce(163), // >>=, reduce: EqneTerm
			reduce(163), // >>>=, reduce: EqneTerm
			reduce(163), // +=, reduce: EqneTerm
			reduce(163), // .=, reduce: EqneTerm
			reduce(163), // -=, reduce: EqneTerm
			reduce(163), // *=, reduce: EqneTerm
			reduce(163), // /=, reduce: EqneTerm
			reduce(163), // //=, reduce: EqneTerm
			reduce(163), // %=, reduce: EqneTerm
			reduce(163), // **=, reduce: EqneTerm
			reduce(163), // ?, reduce: EqneTerm
			nil,         // :
			reduce(163), // ||, reduce: EqneTerm
			reduce(163), // ^^, reduce: EqneTerm
			reduce(163), // &&, reduce: EqneTerm
			reduce(163), // ??, reduce: EqneTerm
			reduce(163), // ???, reduce: EqneTerm
			reduce(163), // =~, reduce: EqneTerm
			reduce(163), // !=~, reduce: EqneTerm
			reduce(163), // ==, reduce: EqneTerm
			reduce(163), // !=, reduce: EqneTerm
			reduce(163), // <=>, reduce: EqneTerm
			shift(720),  // >=
			shift(721),  // <
			shift(722),  // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1482
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(164), // $, reduce: EqneTerm
			nil,         // empty
			reduce(164), // ;, reduce: EqneTerm
			reduce(164), // {, reduce: EqneTerm
			nil,         // }
			reduce(164), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(719),  // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(164), // ||=, reduce: EqneTerm
			reduce(164), // ^^=, reduce: EqneTerm
			reduce(164), // &&=, reduce: EqneTerm
			reduce(164), // ??=, reduce: EqneTerm
			reduce(164), // ???=, reduce: EqneTerm
			reduce(164), // |=, reduce: EqneTerm
			reduce(164), // &=, reduce: EqneTerm
			reduce(164), // ^=, reduce: EqneTerm
			reduce(164), // <<=, reduce: EqneTerm
			reduce(164), // >>=, reduce: EqneTerm
			reduce(164), // >>>=, reduce: EqneTerm
			reduce(164), // +=, reduce: EqneTerm
			reduce(164), // .=, reduce: EqneTerm
			reduce(164), // -=, reduce: EqneTerm
			reduce(164), // *=, reduce: EqneTerm
			reduce(164), // /=, reduce: EqneTerm
			reduce(164), // //=, reduce: EqneTerm
			reduce(164), // %=, reduce: EqneTerm
			reduce(164), // **=, reduce: EqneTerm
			reduce(164), // ?, reduce: EqneTerm
			nil,         // :
			reduce(164), // ||, reduce: EqneTerm
			reduce(164), // ^^, reduce: EqneTerm
			reduce(164), // &&, reduce: EqneTerm
			reduce(164), // ??, reduce: EqneTerm
			reduce(164), // ???, reduce: EqneTerm
			reduce(164), // =~, reduce: EqneTerm
			reduce(164), // !=~, reduce: EqneTerm
			reduce(164), // ==, reduce: EqneTerm
			reduce(164), // !=, reduce: EqneTerm
			reduce(164), // <=>, reduce: EqneTerm
			shift(720),  // >=
			shift(721),  // <
			shift(722),  // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1483
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(165), // $, reduce: EqneTerm
			nil,         // empty
			reduce(165), // ;, reduce: EqneTerm
			reduce(165), // {, reduce: EqneTerm
			nil,         // }
			reduce(165), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(719),  // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(165), // ||=, reduce: EqneTerm
			reduce(165), // ^^=, reduce: EqneTerm
			reduce(165), // &&=, reduce: EqneTerm
			reduce(165), // ??=, reduce: EqneTerm
			reduce(165), // ???=, reduce: EqneTerm
			reduce(165), // |=, reduce: EqneTerm
			reduce(165), // &=, reduce: EqneTerm
			reduce(165), // ^=, reduce: EqneTerm
			reduce(165), // <<=, reduce: EqneTerm
			reduce(165), // >>=, reduce: EqneTerm
			reduce(165), // >>>=, reduce: EqneTerm
			reduce(165), // +=, reduce: EqneTerm
			reduce(165), // .=, reduce: EqneTerm
			reduce(165), // -=, reduce: EqneTerm
			reduce(165), // *=, reduce: EqneTerm
			reduce(165), // /=, reduce: EqneTerm
			reduce(165), // //=, reduce: EqneTerm
			reduce(165), // %=, reduce: EqneTerm
			reduce(165), // **=, reduce: EqneTerm
			reduce(165), // ?, reduce: EqneTerm
			nil,         // :
			reduce(165), // ||, reduce: EqneTerm
			reduce(165), // ^^, reduce: EqneTerm
			reduce(165), // &&, reduce: EqneTerm
			reduce(165), // ??, reduce: EqneTerm
			reduce(165), // ???, reduce: EqneTerm
			reduce(165), // =~, reduce: EqneTerm
			reduce(165), // !=~, reduce: EqneTerm
			reduce(165), // ==, reduce: EqneTerm
			reduce(165), // !=, reduce: EqneTerm
			reduce(165), // <=>, reduce: EqneTerm
			shift(720),  // >=
			shift(721),  // <
			shift(722),  // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1484
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(167), // $, reduce: CmpTerm
			nil,         // empty
			reduce(167), // ;, reduce: CmpTerm
			reduce(167), // {, reduce: CmpTerm
			nil,         // }
			reduce(167), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(167), // >, reduce: CmpTerm
			nil,         // >>
			shift(723),  // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(167), // ||=, reduce: CmpTerm
			reduce(167), // ^^=, reduce: CmpTerm
			reduce(167), // &&=, reduce: CmpTerm
			reduce(167), // ??=, reduce: CmpTerm
			reduce(167), // ???=, reduce: CmpTerm
			reduce(167), // |=, reduce: CmpTerm
			reduce(167), // &=, reduce: CmpTerm
			reduce(167), // ^=, reduce: CmpTerm
			reduce(167), // <<=, reduce: CmpTerm
			reduce(167), // >>=, reduce: CmpTerm
			reduce(167), // >>>=, reduce: CmpTerm
			reduce(167), // +=, reduce: CmpTerm
			reduce(167), // .=, reduce: CmpTerm
			reduce(167), // -=, reduce: CmpTerm
			reduce(167), // *=, reduce: CmpTerm
			reduce(167), // /=, reduce: CmpTerm
			reduce(167), // //=, reduce: CmpTerm
			reduce(167), // %=, reduce: CmpTerm
			reduce(167), // **=, reduce: CmpTerm
			reduce(167), // ?, reduce: CmpTerm
			nil,         // :
			reduce(167), // ||, reduce: CmpTerm
			reduce(167), // ^^, reduce: CmpTerm
			reduce(167), // &&, reduce: CmpTerm
			reduce(167), // ??, reduce: CmpTerm
			reduce(167), // ???, reduce: CmpTerm
			reduce(167), // =~, reduce: CmpTerm
			reduce(167), // !=~, reduce: CmpTerm
			reduce(167), // ==, reduce: CmpTerm
			reduce(167), // !=, reduce: CmpTerm
			reduce(167), // <=>, reduce: CmpTerm
			reduce(167), // >=, reduce: CmpTerm
			reduce(167), // <, reduce: CmpTerm
			reduce(167), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1485
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(168), // $, reduce: CmpTerm
			nil,         // empty
			reduce(168), // ;, reduce: CmpTerm
			reduce(168), // {, reduce: CmpTerm
			nil,         // }
			reduce(168), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(168), // >, reduce: CmpTerm
			nil,         // >>
			shift(723),  // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(168), // ||=, reduce: CmpTerm
			reduce(168), // ^^=, reduce: CmpTerm
			reduce(168), // &&=, reduce: CmpTerm
			reduce(168), // ??=, reduce: CmpTerm
			reduce(168), // ???=, reduce: CmpTerm
			reduce(168), // |=, reduce: CmpTerm
			reduce(168), // &=, reduce: CmpTerm
			reduce(168), // ^=, reduce: CmpTerm
			reduce(168), // <<=, reduce: CmpTerm
			reduce(168), // >>=, reduce: CmpTerm
			reduce(168), // >>>=, reduce: CmpTerm
			reduce(168), // +=, reduce: CmpTerm
			reduce(168), // .=, reduce: CmpTerm
			reduce(168), // -=, reduce: CmpTerm
			reduce(168), // *=, reduce: CmpTerm
			reduce(168), // /=, reduce: CmpTerm
			reduce(168), // //=, reduce: CmpTerm
			reduce(168), // %=, reduce: CmpTerm
			reduce(168), // **=, reduce: CmpTerm
			reduce(168), // ?, reduce: CmpTerm
			nil,         // :
			reduce(168), // ||, reduce: CmpTerm
			reduce(168), // ^^, reduce: CmpTerm
			reduce(168), // &&, reduce: CmpTerm
			reduce(168), // ??, reduce: CmpTerm
			reduce(168), // ???, reduce: CmpTerm
			reduce(168), // =~, reduce: CmpTerm
			reduce(168), // !=~, reduce: CmpTerm
			reduce(168), // ==, reduce: CmpTerm
			reduce(168), // !=, reduce: CmpTerm
			reduce(168), // <=>, reduce: CmpTerm
			reduce(168), // >=, reduce: CmpTerm
			reduce(168), // <, reduce: CmpTerm
			reduce(168), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1486
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(169), // $, reduce: CmpTerm
			nil,         // empty
			reduce(169), // ;, reduce: CmpTerm
			reduce(169), // {, reduce: CmpTerm
			nil,         // }
			reduce(169), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(169), // >, reduce: CmpTerm
			nil,         // >>
			shift(723),  // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(169), // ||=, reduce: CmpTerm
			reduce(169), // ^^=, reduce: CmpTerm
			reduce(169), // &&=, reduce: CmpTerm
			reduce(169), // ??=, reduce: CmpTerm
			reduce(169), // ???=, reduce: CmpTerm
			reduce(169), // |=, reduce: CmpTerm
			reduce(169), // &=, reduce: CmpTerm
			reduce(169), // ^=, reduce: CmpTerm
			reduce(169), // <<=, reduce: CmpTerm
			reduce(169), // >>=, reduce: CmpTerm
			reduce(169), // >>>=, reduce: CmpTerm
			reduce(169), // +=, reduce: CmpTerm
			reduce(169), // .=, reduce: CmpTerm
			reduce(169), // -=, reduce: CmpTerm
			reduce(169), // *=, reduce: CmpTerm
			reduce(169), // /=, reduce: CmpTerm
			reduce(169), // //=, reduce: CmpTerm
			reduce(169), // %=, reduce: CmpTerm
			reduce(169), // **=, reduce: CmpTerm
			reduce(169), // ?, reduce: CmpTerm
			nil,         // :
			reduce(169), // ||, reduce: CmpTerm
			reduce(169), // ^^, reduce: CmpTerm
			reduce(169), // &&, reduce: CmpTerm
			reduce(169), // ??, reduce: CmpTerm
			reduce(169), // ???, reduce: CmpTerm
			reduce(169), // =~, reduce: CmpTerm
			reduce(169), // !=~, reduce: CmpTerm
			reduce(169), // ==, reduce: CmpTerm
			reduce(169), // !=, reduce: CmpTerm
			reduce(169), // <=>, reduce: CmpTerm
			reduce(169), // >=, reduce: CmpTerm
			reduce(169), // <, reduce: CmpTerm
			reduce(169), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1487
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(170), // $, reduce: CmpTerm
			nil,         // empty
			reduce(170), // ;, reduce: CmpTerm
			reduce(170), // {, reduce: CmpTerm
			nil,         // }
			reduce(170), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(170), // >, reduce: CmpTerm
			nil,         // >>
			shift(723),  // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(170), // ||=, reduce: CmpTerm
			reduce(170), // ^^=, reduce: CmpTerm
			reduce(170), // &&=, reduce: CmpTerm
			reduce(170), // ??=, reduce: CmpTerm
			reduce(170), // ???=, reduce: CmpTerm
			reduce(170), // |=, reduce: CmpTerm
			reduce(170), // &=, reduce: CmpTerm
			reduce(170), // ^=, reduce: CmpTerm
			reduce(170), // <<=, reduce: CmpTerm
			reduce(170), // >>=, reduce: CmpTerm
			reduce(170), // >>>=, reduce: CmpTerm
			reduce(170), // +=, reduce: CmpTerm
			reduce(170), // .=, reduce: CmpTerm
			reduce(170), // -=, reduce: CmpTerm
			reduce(170), // *=, reduce: CmpTerm
			reduce(170), // /=, reduce: CmpTerm
			reduce(170), // //=, reduce: CmpTerm
			reduce(170), // %=, reduce: CmpTerm
			reduce(170), // **=, reduce: CmpTerm
			reduce(170), // ?, reduce: CmpTerm
			nil,         // :
			reduce(170), // ||, reduce: CmpTerm
			reduce(170), // ^^, reduce: CmpTerm
			reduce(170), // &&, reduce: CmpTerm
			reduce(170), // ??, reduce: CmpTerm
			reduce(170), // ???, reduce: CmpTerm
			reduce(170), // =~, reduce: CmpTerm
			reduce(170), // !=~, reduce: CmpTerm
			reduce(170), // ==, reduce: CmpTerm
			reduce(170), // !=, reduce: CmpTerm
			reduce(170), // <=>, reduce: CmpTerm
			reduce(170), // >=, reduce: CmpTerm
			reduce(170), // <, reduce: CmpTerm
			reduce(170), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1488
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(172), // $, reduce: BitwiseORTerm
			nil,         // empty
			reduce(172), // ;, reduce: BitwiseORTerm
			reduce(172), // {, reduce: BitwiseORTerm
			nil,         // }
			reduce(172), // =, reduce: BitwiseORTerm
			nil,         // unset
			nil,         // filter
			reduce(172), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(172), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(172), // ||=, reduce: BitwiseORTerm
			reduce(172), // ^^=, reduce: BitwiseORTerm
			reduce(172), // &&=, reduce: BitwiseORTerm
			reduce(172), // ??=, reduce: BitwiseORTerm
			reduce(172), // ???=, reduce: BitwiseORTerm
			reduce(172), // |=, reduce: BitwiseORTerm
			reduce(172), // &=, reduce: BitwiseORTerm
			reduce(172), // ^=, reduce: BitwiseORTerm
			reduce(172), // <<=, reduce: BitwiseORTerm
			reduce(172), // >>=, reduce: BitwiseORTerm
			reduce(172), // >>>=, reduce: BitwiseORTerm
			reduce(172), // +=, reduce: BitwiseORTerm
			reduce(172), // .=, reduce: BitwiseORTerm
			reduce(172), // -=, reduce: BitwiseORTerm
			reduce(172), // *=, reduce: BitwiseORTerm
			reduce(172), // /=, reduce: BitwiseORTerm
			reduce(172), // //=, reduce: BitwiseORTerm
			reduce(172), // %=, reduce: BitwiseORTerm
			reduce(172), // **=, reduce: BitwiseORTerm
			reduce(172), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(172), // ||, reduce: BitwiseORTerm
			reduce(172), // ^^, reduce: BitwiseORTerm
			reduce(172), // &&, reduce: BitwiseORTerm
			reduce(172), // ??, reduce: BitwiseORTerm
			reduce(172), // ???, reduce: BitwiseORTerm
			reduce(172), // =~, reduce: BitwiseORTerm
			reduce(172), // !=~, reduce: BitwiseORTerm
			reduce(172), // ==, reduce: BitwiseORTerm
			reduce(172), // !=, reduce: BitwiseORTerm
			reduce(172), // <=>, reduce: BitwiseORTerm
			reduce(172), // >=, reduce: BitwiseORTerm
			reduce(172), // <, reduce: BitwiseORTerm
			reduce(172), // <=, reduce: BitwiseORTerm
			shift(724),  // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1489
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(174), // $, reduce: BitwiseXORTerm
			nil,         // empty
			reduce(174), // ;, reduce: BitwiseXORTerm
			reduce(174), // {, reduce: BitwiseXORTerm
			nil,         // }
			reduce(174), // =, reduce: BitwiseXORTerm
			nil,         // unset
			nil,         // filter
			reduce(174), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(174), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(174), // ||=, reduce: BitwiseXORTerm
			reduce(174), // ^^=, reduce: BitwiseXORTerm
			reduce(174), // &&=, reduce: BitwiseXORTerm
			reduce(174), // ??=, reduce: BitwiseXORTerm
			reduce(174), // ???=, reduce: BitwiseXORTerm
			reduce(174), // |=, reduce: BitwiseXORTerm
			reduce(174), // &=, reduce: BitwiseXORTerm
			reduce(174), // ^=, reduce: BitwiseXORTerm
			reduce(174), // <<=, reduce: BitwiseXORTerm
			reduce(174), // >>=, reduce: BitwiseXORTerm
			reduce(174), // >>>=, reduce: BitwiseXORTerm
			reduce(174), // +=, reduce: BitwiseXORTerm
			reduce(174), // .=, reduce: BitwiseXORTerm
			reduce(174), // -=, reduce: BitwiseXORTerm
			reduce(174), // *=, reduce: BitwiseXORTerm
			reduce(174), // /=, reduce: BitwiseXORTerm
			reduce(174), // //=, reduce: BitwiseXORTerm
			reduce(174), // %=, reduce: BitwiseXORTerm
			reduce(174), // **=, reduce: BitwiseXORTerm
			reduce(174), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(174), // ||, reduce: BitwiseXORTerm
			reduce(174), // ^^, reduce: BitwiseXORTerm
			reduce(174), // &&, reduce: BitwiseXORTerm
			reduce(174), // ??, reduce: BitwiseXORTerm
			reduce(174), // ???, reduce: BitwiseXORTerm
			reduce(174), // =~, reduce: BitwiseXORTerm
			reduce(174), // !=~, reduce: BitwiseXORTerm
			reduce(174), // ==, reduce: BitwiseXORTerm
			reduce(174), // !=, reduce: BitwiseXORTerm
			reduce(174), // <=>, reduce: BitwiseXORTerm
			reduce(174), // >=, reduce: BitwiseXORTerm
			reduce(174), // <, reduce: BitwiseXORTerm
			reduce(174), // <=, reduce: BitwiseXORTerm
			reduce(174), // ^, reduce: BitwiseXORTerm
			shift(725),  // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1490
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(176), // $, reduce: BitwiseANDTerm
			nil,         // empty
			reduce(176), // ;, reduce: BitwiseANDTerm
			reduce(176), // {, reduce: BitwiseANDTerm
			nil,         // }
			reduce(176), // =, reduce: BitwiseANDTerm
			nil,         // unset
			nil,         // filter
			reduce(176), // >, reduce: BitwiseANDTerm
			shift(726),  // >>
			reduce(176), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(176), // ||=, reduce: BitwiseANDTerm
			reduce(176), // ^^=, reduce: BitwiseANDTerm
			reduce(176), // &&=, reduce: BitwiseANDTerm
			reduce(176), // ??=, reduce: BitwiseANDTerm
			reduce(176), // ???=, reduce: BitwiseANDTerm
			reduce(176), // |=, reduce: BitwiseANDTerm
			reduce(176), // &=, reduce: BitwiseANDTerm
			reduce(176), // ^=, reduce: BitwiseANDTerm
			reduce(176), // <<=, reduce: BitwiseANDTerm
			reduce(176), // >>=, reduce: BitwiseANDTerm
			reduce(176), // >>>=, reduce: BitwiseANDTerm
			reduce(176), // +=, reduce: BitwiseANDTerm
			reduce(176), // .=, reduce: BitwiseANDTerm
			reduce(176), // -=, reduce: BitwiseANDTerm
			reduce(176), // *=, reduce: BitwiseANDTerm
			reduce(176), // /=, reduce: BitwiseANDTerm
			reduce(176), // //=, reduce: BitwiseANDTerm
			reduce(176), // %=, reduce: BitwiseANDTerm
			reduce(176), // **=, reduce: BitwiseANDTerm
			reduce(176), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(176), // ||, reduce: BitwiseANDTerm
			reduce(176), // ^^, reduce: BitwiseANDTerm
			reduce(176), // &&, reduce: BitwiseANDTerm
			reduce(176), // ??, reduce: BitwiseANDTerm
			reduce(176), // ???, reduce: BitwiseANDTerm
			reduce(176), // =~, reduce: BitwiseANDTerm
			reduce(176), // !=~, reduce: BitwiseANDTerm
			reduce(176), // ==, reduce: BitwiseANDTerm
			reduce(176), // !=, reduce: BitwiseANDTerm
			reduce(176), // <=>, reduce: BitwiseANDTerm
			reduce(176), // >=, reduce: BitwiseANDTerm
			reduce(176), // <, reduce: BitwiseANDTerm
			reduce(176), // <=, reduce: BitwiseANDTerm
			reduce(176), // ^, reduce: BitwiseANDTerm
			reduce(176), // &, reduce: BitwiseANDTerm
			shift(727),  // <<
			shift(728),  // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1491
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(179), // $, reduce: BitwiseShiftTerm
			nil,         // empty
			reduce(179), // ;, reduce: BitwiseShiftTerm
			reduce(179), // {, reduce: BitwiseShiftTerm
			nil,         // }
			reduce(179), // =, reduce: BitwiseShiftTerm
			nil,         // unset
			nil,         // filter
			reduce(179), // >, reduce: BitwiseShiftTerm
			reduce(179), // >>, reduce: BitwiseShiftTerm
			reduce(179), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(179), // ||=, reduce: BitwiseShiftTerm
			reduce(179), // ^^=, reduce: BitwiseShiftTerm
			reduce(179), // &&=, reduce: BitwiseShiftTerm
			reduce(179), // ??=, reduce: BitwiseShiftTerm
			reduce(179), // ???=, reduce: BitwiseShiftTerm
			reduce(179), // |=, reduce: BitwiseShiftTerm
			reduce(179), // &=, reduce: BitwiseShiftTerm
			reduce(179), // ^=, reduce: BitwiseShiftTerm
			reduce(179), // <<=, reduce: BitwiseShiftTerm
			reduce(179), // >>=, reduce: BitwiseShiftTerm
			reduce(179), // >>>=, reduce: BitwiseShiftTerm
			reduce(179), // +=, reduce: BitwiseShiftTerm
			reduce(179), // .=, reduce: BitwiseShiftTerm
			reduce(179), // -=, reduce: BitwiseShiftTerm
			reduce(179), // *=, reduce: BitwiseShiftTerm
			reduce(179), // /=, reduce: BitwiseShiftTerm
			reduce(179), // //=, reduce: BitwiseShiftTerm
			reduce(179), // %=, reduce: BitwiseShiftTerm
			reduce(179), // **=, reduce: BitwiseShiftTerm
			reduce(179), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(179), // ||, reduce: BitwiseShiftTerm
			reduce(179), // ^^, reduce: BitwiseShiftTerm
			reduce(179), // &&, reduce: BitwiseShiftTerm
			reduce(179), // ??, reduce: BitwiseShiftTerm
			reduce(179), // ???, reduce: BitwiseShiftTerm
			reduce(179), // =~, reduce: BitwiseShiftTerm
			reduce(179), // !=~, reduce: BitwiseShiftTerm
			reduce(179), // ==, reduce: BitwiseShiftTerm
			reduce(179), // !=, reduce: BitwiseShiftTerm
			reduce(179), // <=>, reduce: BitwiseShiftTerm
			reduce(179), // >=, reduce: BitwiseShiftTerm
			reduce(179), // <, reduce: BitwiseShiftTerm
			reduce(179), // <=, reduce: BitwiseShiftTerm
			reduce(179), // ^, reduce: BitwiseShiftTerm
			reduce(179), // &, reduce: BitwiseShiftTerm
			reduce(179), // <<, reduce: BitwiseShiftTerm
			reduce(179), // >>>, reduce: BitwiseShiftTerm
			shift(729),  // +
			shift(730),  // -
			shift(731),  // .+
			shift(732),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1492
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(178), // $, reduce: BitwiseShiftTerm
			nil,         // empty
			reduce(178), // ;, reduce: BitwiseShiftTerm
			reduce(178), // {, reduce: BitwiseShiftTerm
			nil,         // }
			reduce(178), // =, reduce: BitwiseShiftTerm
			nil,         // unset
			nil,         // filter
			reduce(178), // >, reduce: BitwiseShiftTerm
			reduce(178), // >>, reduce: BitwiseShiftTerm
			reduce(178), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(178), // ||=, reduce: BitwiseShiftTerm
			reduce(178), // ^^=, reduce: BitwiseShiftTerm
			reduce(178), // &&=, reduce: BitwiseShiftTerm
			reduce(178), // ??=, reduce: BitwiseShiftTerm
			reduce(178), // ???=, reduce: BitwiseShiftTerm
			reduce(178), // |=, reduce: BitwiseShiftTerm
			reduce(178), // &=, reduce: BitwiseShiftTerm
			reduce(178), // ^=, reduce: BitwiseShiftTerm
			reduce(178), // <<=, reduce: BitwiseShiftTerm
			reduce(178), // >>=, reduce: BitwiseShiftTerm
			reduce(178), // >>>=, reduce: BitwiseShiftTerm
			reduce(178), // +=, reduce: BitwiseShiftTerm
			reduce(178), // .=, reduce: BitwiseShiftTerm
			reduce(178), // -=, reduce: BitwiseShiftTerm
			reduce(178), // *=, reduce: BitwiseShiftTerm
			reduce(178), // /=, reduce: BitwiseShiftTerm
			reduce(178), // //=, reduce: BitwiseShiftTerm
			reduce(178), // %=, reduce: BitwiseShiftTerm
			reduce(178), // **=, reduce: BitwiseShiftTerm
			reduce(178), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(178), // ||, reduce: BitwiseShiftTerm
			reduce(178), // ^^, reduce: BitwiseShiftTerm
			reduce(178), // &&, reduce: BitwiseShiftTerm
			reduce(178), // ??, reduce: BitwiseShiftTerm
			reduce(178), // ???, reduce: BitwiseShiftTerm
			reduce(178), // =~, reduce: BitwiseShiftTerm
			reduce(178), // !=~, reduce: BitwiseShiftTerm
			reduce(178), // ==, reduce: BitwiseShiftTerm
			reduce(178), // !=, reduce: BitwiseShiftTerm
			reduce(178), // <=>, reduce: BitwiseShiftTerm
			reduce(178), // >=, reduce: BitwiseShiftTerm
			reduce(178), // <, reduce: BitwiseShiftTerm
			reduce(178), // <=, reduce: BitwiseShiftTerm
			reduce(178), // ^, reduce: BitwiseShiftTerm
			reduce(178), // &, reduce: BitwiseShiftTerm
			reduce(178), // <<, reduce: BitwiseShiftTerm
			reduce(178), // >>>, reduce: BitwiseShiftTerm
			shift(729),  // +
			shift(730),  // -
			shift(731),  // .+
			shift(732),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1493
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(180), // $, reduce: BitwiseShiftTerm
			nil,         // empty
			reduce(180), // ;, reduce: BitwiseShiftTerm
			reduce(180), // {, reduce: BitwiseShiftTerm
			nil,         // }
			reduce(180), // =, reduce: BitwiseShiftTerm
			nil,         // unset
			nil,         // filter
			reduce(180), // >, reduce: BitwiseShiftTerm
			reduce(180), // >>, reduce: BitwiseShiftTerm
			reduce(180), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(180), // ||=, reduce: BitwiseShiftTerm
			reduce(180), // ^^=, reduce: BitwiseShiftTerm
			reduce(180), // &&=, reduce: BitwiseShiftTerm
			reduce(180), // ??=, reduce: BitwiseShiftTerm
			reduce(180), // ???=, reduce: BitwiseShiftTerm
			reduce(180), // |=, reduce: BitwiseShiftTerm
			reduce(180), // &=, reduce: BitwiseShiftTerm
			reduce(180), // ^=, reduce: BitwiseShiftTerm
			reduce(180), // <<=, reduce: BitwiseShiftTerm
			reduce(180), // >>=, reduce: BitwiseShiftTerm
			reduce(180), // >>>=, reduce: BitwiseShiftTerm
			reduce(180), // +=, reduce: BitwiseShiftTerm
			reduce(180), // .=, reduce: BitwiseShiftTerm
			reduce(180), // -=, reduce: BitwiseShiftTerm
			reduce(180), // *=, reduce: BitwiseShiftTerm
			reduce(180), // /=, reduce: BitwiseShiftTerm
			reduce(180), // //=, reduce: BitwiseShiftTerm
			reduce(180), // %=, reduce: BitwiseShiftTerm
			reduce(180), // **=, reduce: BitwiseShiftTerm
			reduce(180), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(180), // ||, reduce: BitwiseShiftTerm
			reduce(180), // ^^, reduce: BitwiseShiftTerm
			reduce(180), // &&, reduce: BitwiseShiftTerm
			reduce(180), // ??, reduce: BitwiseShiftTerm
			reduce(180), // ???, reduce: BitwiseShiftTerm
			reduce(180), // =~, reduce: BitwiseShiftTerm
			reduce(180), // !=~, reduce: BitwiseShiftTerm
			reduce(180), // ==, reduce: BitwiseShiftTerm
			reduce(180), // !=, reduce: BitwiseShiftTerm
			reduce(180), // <=>, reduce: BitwiseShiftTerm
			reduce(180), // >=, reduce: BitwiseShiftTerm
			reduce(180), // <, reduce: BitwiseShiftTerm
			reduce(180), // <=, reduce: BitwiseShiftTerm
			reduce(180), // ^, reduce: BitwiseShiftTerm
			reduce(180), // &, reduce: BitwiseShiftTerm
			reduce(180), // <<, reduce: BitwiseShiftTerm
			reduce(180), // >>>, reduce: BitwiseShiftTerm
			shift(729),  // +
			shift(730),  // -
			shift(731),  // .+
			shift(732),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1494
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(182), // $, reduce: AddsubdotTerm
			nil,         // empty
			reduce(182), // ;, reduce: AddsubdotTerm
			reduce(182), // {, reduce: AddsubdotTerm
			nil,         // }
			reduce(182), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(182), // >, reduce: AddsubdotTerm
			reduce(182), // >>, reduce: AddsubdotTerm
			reduce(182), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(182), // ||=, reduce: AddsubdotTerm
			reduce(182), // ^^=, reduce: AddsubdotTerm
			reduce(182), // &&=, reduce: AddsubdotTerm
			reduce(182), // ??=, reduce: AddsubdotTerm
			reduce(182), // ???=, reduce: AddsubdotTerm
			reduce(182), // |=, reduce: AddsubdotTerm
			reduce(182), // &=, reduce: AddsubdotTerm
			reduce(182), // ^=, reduce: AddsubdotTerm
			reduce(182), // <<=, reduce: AddsubdotTerm
			reduce(182), // >>=, reduce: AddsubdotTerm
			reduce(182), // >>>=, reduce: AddsubdotTerm
			reduce(182), // +=, reduce: AddsubdotTerm
			reduce(182), // .=, reduce: AddsubdotTerm
			reduce(182), // -=, reduce: AddsubdotTerm
			reduce(182), // *=, reduce: AddsubdotTerm
			reduce(182), // /=, reduce: AddsubdotTerm
			reduce(182), // //=, reduce: AddsubdotTerm
			reduce(182), // %=, reduce: AddsubdotTerm
			reduce(182), // **=, reduce: AddsubdotTerm
			reduce(182), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(182), // ||, reduce: AddsubdotTerm
			reduce(182), // ^^, reduce: AddsubdotTerm
			reduce(182), // &&, reduce: AddsubdotTerm
			reduce(182), // ??, reduce: AddsubdotTerm
			reduce(182), // ???, reduce: AddsubdotTerm
			reduce(182), // =~, reduce: AddsubdotTerm
			reduce(182), // !=~, reduce: AddsubdotTerm
			reduce(182), // ==, reduce: AddsubdotTerm
			reduce(182), // !=, reduce: AddsubdotTerm
			reduce(182), // <=>, reduce: AddsubdotTerm
			reduce(182), // >=, reduce: AddsubdotTerm
			reduce(182), // <, reduce: AddsubdotTerm
			reduce(182), // <=, reduce: AddsubdotTerm
			reduce(182), // ^, reduce: AddsubdotTerm
			reduce(182), // &, reduce: AddsubdotTerm
			reduce(182), // <<, reduce: AddsubdotTerm
			reduce(182), // >>>, reduce: AddsubdotTerm
			reduce(182), // +, reduce: AddsubdotTerm
			reduce(182), // -, reduce: AddsubdotTerm
			reduce(182), // .+, reduce: AddsubdotTerm
			reduce(182), // .-, reduce: AddsubdotTerm
			shift(735),  // *
			shift(736),  // /
			shift(737),  // //
			shift(738),  // %
			shift(739),  // .*
			shift(740),  // ./
			shift(741),  // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1495
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(183), // $, reduce: AddsubdotTerm
			nil,         // empty
			reduce(183), // ;, reduce: AddsubdotTerm
			reduce(183), // {, reduce: AddsubdotTerm
			nil,         // }
			reduce(183), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(183), // >, reduce: AddsubdotTerm
			reduce(183), // >>, reduce: AddsubdotTerm
			reduce(183), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(183), // ||=, reduce: AddsubdotTerm
			reduce(183), // ^^=, reduce: AddsubdotTerm
			reduce(183), // &&=, reduce: AddsubdotTerm
			reduce(183), // ??=, reduce: AddsubdotTerm
			reduce(183), // ???=, reduce: AddsubdotTerm
			reduce(183), // |=, reduce: AddsubdotTerm
			reduce(183), // &=, reduce: AddsubdotTerm
			reduce(183), // ^=, reduce: AddsubdotTerm
			reduce(183), // <<=, reduce: AddsubdotTerm
			reduce(183), // >>=, reduce: AddsubdotTerm
			reduce(183), // >>>=, reduce: AddsubdotTerm
			reduce(183), // +=, reduce: AddsubdotTerm
			reduce(183), // .=, reduce: AddsubdotTerm
			reduce(183), // -=, reduce: AddsubdotTerm
			reduce(183), // *=, reduce: AddsubdotTerm
			reduce(183), // /=, reduce: AddsubdotTerm
			reduce(183), // //=, reduce: AddsubdotTerm
			reduce(183), // %=, reduce: AddsubdotTerm
			reduce(183), // **=, reduce: AddsubdotTerm
			reduce(183), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(183), // ||, reduce: AddsubdotTerm
			reduce(183), // ^^, reduce: AddsubdotTerm
			reduce(183), // &&, reduce: AddsubdotTerm
			reduce(183), // ??, reduce: AddsubdotTerm
			reduce(183), // ???, reduce: AddsubdotTerm
			reduce(183), // =~, reduce: AddsubdotTerm
			reduce(183), // !=~, reduce: AddsubdotTerm
			reduce(183), // ==, reduce: AddsubdotTerm
			reduce(183), // !=, reduce: AddsubdotTerm
			reduce(183), // <=>, reduce: AddsubdotTerm
			reduce(183), // >=, reduce: AddsubdotTerm
			reduce(183), // <, reduce: AddsubdotTerm
			reduce(183), // <=, reduce: AddsubdotTerm
			reduce(183), // ^, reduce: AddsubdotTerm
			reduce(183), // &, reduce: AddsubdotTerm
			reduce(183), // <<, reduce: AddsubdotTerm
			reduce(183), // >>>, reduce: AddsubdotTerm
			reduce(183), // +, reduce: AddsubdotTerm
			reduce(183), // -, reduce: AddsubdotTerm
			reduce(183), // .+, reduce: AddsubdotTerm
			reduce(183), // .-, reduce: AddsubdotTerm
			shift(735),  // *
			shift(736),  // /
			shift(737),  // //
			shift(738),  // %
			shift(739),  // .*
			shift(740),  // ./
			shift(741),  // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1496
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(184), // $, reduce: AddsubdotTerm
			nil,         // empty
			reduce(184), // ;, reduce: AddsubdotTerm
			reduce(184), // {, reduce: AddsubdotTerm
			nil,         // }
			reduce(184), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(184), // >, reduce: AddsubdotTerm
			reduce(184), // >>, reduce: AddsubdotTerm
			reduce(184), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(184), // ||=, reduce: AddsubdotTerm
			reduce(184), // ^^=, reduce: AddsubdotTerm
			reduce(184), // &&=, reduce: AddsubdotTerm
			reduce(184), // ??=, reduce: AddsubdotTerm
			reduce(184), // ???=, reduce: AddsubdotTerm
			reduce(184), // |=, reduce: AddsubdotTerm
			reduce(184), // &=, reduce: AddsubdotTerm
			reduce(184), // ^=, reduce: AddsubdotTerm
			reduce(184), // <<=, reduce: AddsubdotTerm
			reduce(184), // >>=, reduce: AddsubdotTerm
			reduce(184), // >>>=, reduce: AddsubdotTerm
			reduce(184), // +=, reduce: AddsubdotTerm
			reduce(184), // .=, reduce: AddsubdotTerm
			reduce(184), // -=, reduce: AddsubdotTerm
			reduce(184), // *=, reduce: AddsubdotTerm
			reduce(184), // /=, reduce: AddsubdotTerm
			reduce(184), // //=, reduce: AddsubdotTerm
			reduce(184), // %=, reduce: AddsubdotTerm
			reduce(184), // **=, reduce: AddsubdotTerm
			reduce(184), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(184), // ||, reduce: AddsubdotTerm
			reduce(184), // ^^, reduce: AddsubdotTerm
			reduce(184), // &&, reduce: AddsubdotTerm
			reduce(184), // ??, reduce: AddsubdotTerm
			reduce(184), // ???, reduce: AddsubdotTerm
			reduce(184), // =~, reduce: AddsubdotTerm
			reduce(184), // !=~, reduce: AddsubdotTerm
			reduce(184), // ==, reduce: AddsubdotTerm
			reduce(184), // !=, reduce: AddsubdotTerm
			reduce(184), // <=>, reduce: AddsubdotTerm
			reduce(184), // >=, reduce: AddsubdotTerm
			reduce(184), // <, reduce: AddsubdotTerm
			reduce(184), // <=, reduce: AddsubdotTerm
			reduce(184), // ^, reduce: AddsubdotTerm
			reduce(184), // &, reduce: AddsubdotTerm
			reduce(184), // <<, reduce: AddsubdotTerm
			reduce(184), // >>>, reduce: AddsubdotTerm
			reduce(184), // +, reduce: AddsubdotTerm
			reduce(184), // -, reduce: AddsubdotTerm
			reduce(184), // .+, reduce: AddsubdotTerm
			reduce(184), // .-, reduce: AddsubdotTerm
			shift(735),  // *
			shift(736),  // /
			shift(737),  // //
			shift(738),  // %
			shift(739),  // .*
			shift(740),  // ./
			shift(741),  // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1497
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(185), // $, reduce: AddsubdotTerm
			nil,         // empty
			reduce(185), // ;, reduce: AddsubdotTerm
			reduce(185), // {, reduce: AddsubdotTerm
			nil,         // }
			reduce(185), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(185), // >, reduce: AddsubdotTerm
			reduce(185), // >>, reduce: AddsubdotTerm
			reduce(185), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(185), // ||=, reduce: AddsubdotTerm
			reduce(185), // ^^=, reduce: AddsubdotTerm
			reduce(185), // &&=, reduce: AddsubdotTerm
			reduce(185), // ??=, reduce: AddsubdotTerm
			reduce(185), // ???=, reduce: AddsubdotTerm
			reduce(185), // |=, reduce: AddsubdotTerm
			reduce(185), // &=, reduce: AddsubdotTerm
			reduce(185), // ^=, reduce: AddsubdotTerm
			reduce(185), // <<=, reduce: AddsubdotTerm
			reduce(185), // >>=, reduce: AddsubdotTerm
			reduce(185), // >>>=, reduce: AddsubdotTerm
			reduce(185), // +=, reduce: AddsubdotTerm
			reduce(185), // .=, reduce: AddsubdotTerm
			reduce(185), // -=, reduce: AddsubdotTerm
			reduce(185), // *=, reduce: AddsubdotTerm
			reduce(185), // /=, reduce: AddsubdotTerm
			reduce(185), // //=, reduce: AddsubdotTerm
			reduce(185), // %=, reduce: AddsubdotTerm
			reduce(185), // **=, reduce: AddsubdotTerm
			reduce(185), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(185), // ||, reduce: AddsubdotTerm
			reduce(185), // ^^, reduce: AddsubdotTerm
			reduce(185), // &&, reduce: AddsubdotTerm
			reduce(185), // ??, reduce: AddsubdotTerm
			reduce(185), // ???, reduce: AddsubdotTerm
			reduce(185), // =~, reduce: AddsubdotTerm
			reduce(185), // !=~, reduce: AddsubdotTerm
			reduce(185), // ==, reduce: AddsubdotTerm
			reduce(185), // !=, reduce: AddsubdotTerm
			reduce(185), // <=>, reduce: AddsubdotTerm
			reduce(185), // >=, reduce: AddsubdotTerm
			reduce(185), // <, reduce: AddsubdotTerm
			reduce(185), // <=, reduce: AddsubdotTerm
			reduce(185), // ^, reduce: AddsubdotTerm
			reduce(185), // &, reduce: AddsubdotTerm
			reduce(185), // <<, reduce: AddsubdotTerm
			reduce(185), // >>>, reduce: AddsubdotTerm
			reduce(185), // +, reduce: AddsubdotTerm
			reduce(185), // -, reduce: AddsubdotTerm
			reduce(185), // .+, reduce: AddsubdotTerm
			reduce(185), // .-, reduce: AddsubdotTerm
			shift(735),  // *
			shift(736),  // /
			shift(737),  // //
			shift(738),  // %
			shift(739),  // .*
			shift(740),  // ./
			shift(741),  // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1498
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(187), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(187), // ;, reduce: MuldivTerm
			reduce(187), // {, reduce: MuldivTerm
			nil,         // }
			reduce(187), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(187), // >, reduce: MuldivTerm
			reduce(187), // >>, reduce: MuldivTerm
			reduce(187), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(187), // ||=, reduce: MuldivTerm
			reduce(187), // ^^=, reduce: MuldivTerm
			reduce(187), // &&=, reduce: MuldivTerm
			reduce(187), // ??=, reduce: MuldivTerm
			reduce(187), // ???=, reduce: MuldivTerm
			reduce(187), // |=, reduce: MuldivTerm
			reduce(187), // &=, reduce: MuldivTerm
			reduce(187), // ^=, reduce: MuldivTerm
			reduce(187), // <<=, reduce: MuldivTerm
			reduce(187), // >>=, reduce: MuldivTerm
			reduce(187), // >>>=, reduce: MuldivTerm
			reduce(187), // +=, reduce: MuldivTerm
			reduce(187), // .=, reduce: MuldivTerm
			reduce(187), // -=, reduce: MuldivTerm
			reduce(187), // *=, reduce: MuldivTerm
			reduce(187), // /=, reduce: MuldivTerm
			reduce(187), // //=, reduce: MuldivTerm
			reduce(187), // %=, reduce: MuldivTerm
			reduce(187), // **=, reduce: MuldivTerm
			reduce(187), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(187), // ||, reduce: MuldivTerm
			reduce(187), // ^^, reduce: MuldivTerm
			reduce(187), // &&, reduce: MuldivTerm
			reduce(187), // ??, reduce: MuldivTerm
			reduce(187), // ???, reduce: MuldivTerm
			reduce(187), // =~, reduce: MuldivTerm
			reduce(187), // !=~, reduce: MuldivTerm
			reduce(187), // ==, reduce: MuldivTerm
			reduce(187), // !=, reduce: MuldivTerm
			reduce(187), // <=>, reduce: MuldivTerm
			reduce(187), // >=, reduce: MuldivTerm
			reduce(187), // <, reduce: MuldivTerm
			reduce(187), // <=, reduce: MuldivTerm
			reduce(187), // ^, reduce: MuldivTerm
			reduce(187), // &, reduce: MuldivTerm
			reduce(187), // <<, reduce: MuldivTerm
			reduce(187), // >>>, reduce: MuldivTerm
			reduce(187), // +, reduce: MuldivTerm
			reduce(187), // -, reduce: MuldivTerm
			reduce(187), // .+, reduce: MuldivTerm
			reduce(187), // .-, reduce: MuldivTerm
			reduce(187), // *, reduce: MuldivTerm
			reduce(187), // /, reduce: MuldivTerm
			reduce(187), // //, reduce: MuldivTerm
			reduce(187), // %, reduce: MuldivTerm
			reduce(187), // .*, reduce: MuldivTerm
			reduce(187), // ./, reduce: MuldivTerm
			reduce(187), // .//, reduce: MuldivTerm
			shift(745),  // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1499
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(188), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(188), // ;, reduce: MuldivTerm
			reduce(188), // {, reduce: MuldivTerm
			nil,         // }
			reduce(188), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(188), // >, reduce: MuldivTerm
			reduce(188), // >>, reduce: MuldivTerm
			reduce(188), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(188), // ||=, reduce: MuldivTerm
			reduce(188), // ^^=, reduce: MuldivTerm
			reduce(188), // &&=, reduce: MuldivTerm
			reduce(188), // ??=, reduce: MuldivTerm
			reduce(188), // ???=, reduce: MuldivTerm
			reduce(188), // |=, reduce: MuldivTerm
			reduce(188), // &=, reduce: MuldivTerm
			reduce(188), // ^=, reduce: MuldivTerm
			reduce(188), // <<=, reduce: MuldivTerm
			reduce(188), // >>=, reduce: MuldivTerm
			reduce(188), // >>>=, reduce: MuldivTerm
			reduce(188), // +=, reduce: MuldivTerm
			reduce(188), // .=, reduce: MuldivTerm
			reduce(188), // -=, reduce: MuldivTerm
			reduce(188), // *=, reduce: MuldivTerm
			reduce(188), // /=, reduce: MuldivTerm
			reduce(188), // //=, reduce: MuldivTerm
			reduce(188), // %=, reduce: MuldivTerm
			reduce(188), // **=, reduce: MuldivTerm
			reduce(188), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(188), // ||, reduce: MuldivTerm
			reduce(188), // ^^, reduce: MuldivTerm
			reduce(188), // &&, reduce: MuldivTerm
			reduce(188), // ??, reduce: MuldivTerm
			reduce(188), // ???, reduce: MuldivTerm
			reduce(188), // =~, reduce: MuldivTerm
			reduce(188), // !=~, reduce: MuldivTerm
			reduce(188), // ==, reduce: MuldivTerm
			reduce(188), // !=, reduce: MuldivTerm
			reduce(188), // <=>, reduce: MuldivTerm
			reduce(188), // >=, reduce: MuldivTerm
			reduce(188), // <, reduce: MuldivTerm
			reduce(188), // <=, reduce: MuldivTerm
			reduce(188), // ^, reduce: MuldivTerm
			reduce(188), // &, reduce: MuldivTerm
			reduce(188), // <<, reduce: MuldivTerm
			reduce(188), // >>>, reduce: MuldivTerm
			reduce(188), // +, reduce: MuldivTerm
			reduce(188), // -, reduce: MuldivTerm
			reduce(188), // .+, reduce: MuldivTerm
			reduce(188), // .-, reduce: MuldivTerm
			reduce(188), // *, reduce: MuldivTerm
			reduce(188), // /, reduce: MuldivTerm
			reduce(188), // //, reduce: MuldivTerm
			reduce(188), // %, reduce: MuldivTerm
			reduce(188), // .*, reduce: MuldivTerm
			reduce(188), // ./, reduce: MuldivTerm
			reduce(188), // .//, reduce: MuldivTerm
			shift(745),  // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1500
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(189), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(189), // ;, reduce: MuldivTerm
			reduce(189), // {, reduce: MuldivTerm
			nil,         // }
			reduce(189), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(189), // >, reduce: MuldivTerm
			reduce(189), // >>, reduce: MuldivTerm
			reduce(189), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(189), // ||=, reduce: MuldivTerm
			reduce(189), // ^^=, reduce: MuldivTerm
			reduce(189), // &&=, reduce: MuldivTerm
			reduce(189), // ??=, reduce: MuldivTerm
			reduce(189), // ???=, reduce: MuldivTerm
			reduce(189), // |=, reduce: MuldivTerm
			reduce(189), // &=, reduce: MuldivTerm
			reduce(189), // ^=, reduce: MuldivTerm
			reduce(189), // <<=, reduce: MuldivTerm
			reduce(189), // >>=, reduce: MuldivTerm
			reduce(189), // >>>=, reduce: MuldivTerm
			reduce(189), // +=, reduce: MuldivTerm
			reduce(189), // .=, reduce: MuldivTerm
			reduce(189), // -=, reduce: MuldivTerm
			reduce(189), // *=, reduce: MuldivTerm
			reduce(189), // /=, reduce: MuldivTerm
			reduce(189), // //=, reduce: MuldivTerm
			reduce(189), // %=, reduce: MuldivTerm
			reduce(189), // **=, reduce: MuldivTerm
			reduce(189), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(189), // ||, reduce: MuldivTerm
			reduce(189), // ^^, reduce: MuldivTerm
			reduce(189), // &&, reduce: MuldivTerm
			reduce(189), // ??, reduce: MuldivTerm
			reduce(189), // ???, reduce: MuldivTerm
			reduce(189), // =~, reduce: MuldivTerm
			reduce(189), // !=~, reduce: MuldivTerm
			reduce(189), // ==, reduce: MuldivTerm
			reduce(189), // !=, reduce: MuldivTerm
			reduce(189), // <=>, reduce: MuldivTerm
			reduce(189), // >=, reduce: MuldivTerm
			reduce(189), // <, reduce: MuldivTerm
			reduce(189), // <=, reduce: MuldivTerm
			reduce(189), // ^, reduce: MuldivTerm
			reduce(189), // &, reduce: MuldivTerm
			reduce(189), // <<, reduce: MuldivTerm
			reduce(189), // >>>, reduce: MuldivTerm
			reduce(189), // +, reduce: MuldivTerm
			reduce(189), // -, reduce: MuldivTerm
			reduce(189), // .+, reduce: MuldivTerm
			reduce(189), // .-, reduce: MuldivTerm
			reduce(189), // *, reduce: MuldivTerm
			reduce(189), // /, reduce: MuldivTerm
			reduce(189), // //, reduce: MuldivTerm
			reduce(189), // %, reduce: MuldivTerm
			reduce(189), // .*, reduce: MuldivTerm
			reduce(189), // ./, reduce: MuldivTerm
			reduce(189), // .//, reduce: MuldivTerm
			shift(745),  // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1501
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(190), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(190), // ;, reduce: MuldivTerm
			reduce(190), // {, reduce: MuldivTerm
			nil,         // }
			reduce(190), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(190), // >, reduce: MuldivTerm
			reduce(190), // >>, reduce: MuldivTerm
			reduce(190), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(190), // ||=, reduce: MuldivTerm
			reduce(190), // ^^=, reduce: MuldivTerm
			reduce(190), // &&=, reduce: MuldivTerm
			reduce(190), // ??=, reduce: MuldivTerm
			reduce(190), // ???=, reduce: MuldivTerm
			reduce(190), // |=, reduce: MuldivTerm
			reduce(190), // &=, reduce: MuldivTerm
			reduce(190), // ^=, reduce: MuldivTerm
			reduce(190), // <<=, reduce: MuldivTerm
			reduce(190), // >>=, reduce: MuldivTerm
			reduce(190), // >>>=, reduce: MuldivTerm
			reduce(190), // +=, reduce: MuldivTerm
			reduce(190), // .=, reduce: MuldivTerm
			reduce(190), // -=, reduce: MuldivTerm
			reduce(190), // *=, reduce: MuldivTerm
			reduce(190), // /=, reduce: MuldivTerm
			reduce(190), // //=, reduce: MuldivTerm
			reduce(190), // %=, reduce: MuldivTerm
			reduce(190), // **=, reduce: MuldivTerm
			reduce(190), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(190), // ||, reduce: MuldivTerm
			reduce(190), // ^^, reduce: MuldivTerm
			reduce(190), // &&, reduce: MuldivTerm
			reduce(190), // ??, reduce: MuldivTerm
			reduce(190), // ???, reduce: MuldivTerm
			reduce(190), // =~, reduce: MuldivTerm
			reduce(190), // !=~, reduce: MuldivTerm
			reduce(190), // ==, reduce: MuldivTerm
			reduce(190), // !=, reduce: MuldivTerm
			reduce(190), // <=>, reduce: MuldivTerm
			reduce(190), // >=, reduce: MuldivTerm
			reduce(190), // <, reduce: MuldivTerm
			reduce(190), // <=, reduce: MuldivTerm
			reduce(190), // ^, reduce: MuldivTerm
			reduce(190), // &, reduce: MuldivTerm
			reduce(190), // <<, reduce: MuldivTerm
			reduce(190), // >>>, reduce: MuldivTerm
			reduce(190), // +, reduce: MuldivTerm
			reduce(190), // -, reduce: MuldivTerm
			reduce(190), // .+, reduce: MuldivTerm
			reduce(190), // .-, reduce: MuldivTerm
			reduce(190), // *, reduce: MuldivTerm
			reduce(190), // /, reduce: MuldivTerm
			reduce(190), // //, reduce: MuldivTerm
			reduce(190), // %, reduce: MuldivTerm
			reduce(190), // .*, reduce: MuldivTerm
			reduce(190), // ./, reduce: MuldivTerm
			reduce(190), // .//, reduce: MuldivTerm
			shift(745),  // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1502
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(191), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(191), // ;, reduce: MuldivTerm
			reduce(191), // {, reduce: MuldivTerm
			nil,         // }
			reduce(191), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(191), // >, reduce: MuldivTerm
			reduce(191), // >>, reduce: MuldivTerm
			reduce(191), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(191), // ||=, reduce: MuldivTerm
			reduce(191), // ^^=, reduce: MuldivTerm
			reduce(191), // &&=, reduce: MuldivTerm
			reduce(191), // ??=, reduce: MuldivTerm
			reduce(191), // ???=, reduce: MuldivTerm
			reduce(191), // |=, reduce: MuldivTerm
			reduce(191), // &=, reduce: MuldivTerm
			reduce(191), // ^=, reduce: MuldivTerm
			reduce(191), // <<=, reduce: MuldivTerm
			reduce(191), // >>=, reduce: MuldivTerm
			reduce(191), // >>>=, reduce: MuldivTerm
			reduce(191), // +=, reduce: MuldivTerm
			reduce(191), // .=, reduce: MuldivTerm
			reduce(191), // -=, reduce: MuldivTerm
			reduce(191), // *=, reduce: MuldivTerm
			reduce(191), // /=, reduce: MuldivTerm
			reduce(191), // //=, reduce: MuldivTerm
			reduce(191), // %=, reduce: MuldivTerm
			reduce(191), // **=, reduce: MuldivTerm
			reduce(191), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(191), // ||, reduce: MuldivTerm
			reduce(191), // ^^, reduce: MuldivTerm
			reduce(191), // &&, reduce: MuldivTerm
			reduce(191), // ??, reduce: MuldivTerm
			reduce(191), // ???, reduce: MuldivTerm
			reduce(191), // =~, reduce: MuldivTerm
			reduce(191), // !=~, reduce: MuldivTerm
			reduce(191), // ==, reduce: MuldivTerm
			reduce(191), // !=, reduce: MuldivTerm
			reduce(191), // <=>, reduce: MuldivTerm
			reduce(191), // >=, reduce: MuldivTerm
			reduce(191), // <, reduce: MuldivTerm
			reduce(191), // <=, reduce: MuldivTerm
			reduce(191), // ^, reduce: MuldivTerm
			reduce(191), // &, reduce: MuldivTerm
			reduce(191), // <<, reduce: MuldivTerm
			reduce(191), // >>>, reduce: MuldivTerm
			reduce(191), // +, reduce: MuldivTerm
			reduce(191), // -, reduce: MuldivTerm
			reduce(191), // .+, reduce: MuldivTerm
			reduce(191), // .-, reduce: MuldivTerm
			reduce(191), // *, reduce: MuldivTerm
			reduce(191), // /, reduce: MuldivTerm
			reduce(191), // //, reduce: MuldivTerm
			reduce(191), // %, reduce: MuldivTerm
			reduce(191), // .*, reduce: MuldivTerm
			reduce(191), // ./, reduce: MuldivTerm
			reduce(191), // .//, reduce: MuldivTerm
			shift(745),  // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1503
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(192), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(192), // ;, reduce: MuldivTerm
			reduce(192), // {, reduce: MuldivTerm
			nil,         // }
			reduce(192), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(192), // >, reduce: MuldivTerm
			reduce(192), // >>, reduce: MuldivTerm
			reduce(192), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(192), // ||=, reduce: MuldivTerm
			reduce(192), // ^^=, reduce: MuldivTerm
			reduce(192), // &&=, reduce: MuldivTerm
			reduce(192), // ??=, reduce: MuldivTerm
			reduce(192), // ???=, reduce: MuldivTerm
			reduce(192), // |=, reduce: MuldivTerm
			reduce(192), // &=, reduce: MuldivTerm
			reduce(192), // ^=, reduce: MuldivTerm
			reduce(192), // <<=, reduce: MuldivTerm
			reduce(192), // >>=, reduce: MuldivTerm
			reduce(192), // >>>=, reduce: MuldivTerm
			reduce(192), // +=, reduce: MuldivTerm
			reduce(192), // .=, reduce: MuldivTerm
			reduce(192), // -=, reduce: MuldivTerm
			reduce(192), // *=, reduce: MuldivTerm
			reduce(192), // /=, reduce: MuldivTerm
			reduce(192), // //=, reduce: MuldivTerm
			reduce(192), // %=, reduce: MuldivTerm
			reduce(192), // **=, reduce: MuldivTerm
			reduce(192), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(192), // ||, reduce: MuldivTerm
			reduce(192), // ^^, reduce: MuldivTerm
			reduce(192), // &&, reduce: MuldivTerm
			reduce(192), // ??, reduce: MuldivTerm
			reduce(192), // ???, reduce: MuldivTerm
			reduce(192), // =~, reduce: MuldivTerm
			reduce(192), // !=~, reduce: MuldivTerm
			reduce(192), // ==, reduce: MuldivTerm
			reduce(192), // !=, reduce: MuldivTerm
			reduce(192), // <=>, reduce: MuldivTerm
			reduce(192), // >=, reduce: MuldivTerm
			reduce(192), // <, reduce: MuldivTerm
			reduce(192), // <=, reduce: MuldivTerm
			reduce(192), // ^, reduce: MuldivTerm
			reduce(192), // &, reduce: MuldivTerm
			reduce(192), // <<, reduce: MuldivTerm
			reduce(192), // >>>, reduce: MuldivTerm
			reduce(192), // +, reduce: MuldivTerm
			reduce(192), // -, reduce: MuldivTerm
			reduce(192), // .+, reduce: MuldivTerm
			reduce(192), // .-, reduce: MuldivTerm
			reduce(192), // *, reduce: MuldivTerm
			reduce(192), // /, reduce: MuldivTerm
			reduce(192), // //, reduce: MuldivTerm
			reduce(192), // %, reduce: MuldivTerm
			reduce(192), // .*, reduce: MuldivTerm
			reduce(192), // ./, reduce: MuldivTerm
			reduce(192), // .//, reduce: MuldivTerm
			shift(745),  // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1504
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(193), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(193), // ;, reduce: MuldivTerm
			reduce(193), // {, reduce: MuldivTerm
			nil,         // }
			reduce(193), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(193), // >, reduce: MuldivTerm
			reduce(193), // >>, reduce: MuldivTerm
			reduce(193), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(193), // ||=, reduce: MuldivTerm
			reduce(193), // ^^=, reduce: MuldivTerm
			reduce(193), // &&=, reduce: MuldivTerm
			reduce(193), // ??=, reduce: MuldivTerm
			reduce(193), // ???=, reduce: MuldivTerm
			reduce(193), // |=, reduce: MuldivTerm
			reduce(193), // &=, reduce: MuldivTerm
			reduce(193), // ^=, reduce: MuldivTerm
			reduce(193), // <<=, reduce: MuldivTerm
			reduce(193), // >>=, reduce: MuldivTerm
			reduce(193), // >>>=, reduce: MuldivTerm
			reduce(193), // +=, reduce: MuldivTerm
			reduce(193), // .=, reduce: MuldivTerm
			reduce(193), // -=, reduce: MuldivTerm
			reduce(193), // *=, reduce: MuldivTerm
			reduce(193), // /=, reduce: MuldivTerm
			reduce(193), // //=, reduce: MuldivTerm
			reduce(193), // %=, reduce: MuldivTerm
			reduce(193), // **=, reduce: MuldivTerm
			reduce(193), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(193), // ||, reduce: MuldivTerm
			reduce(193), // ^^, reduce: MuldivTerm
			reduce(193), // &&, reduce: MuldivTerm
			reduce(193), // ??, reduce: MuldivTerm
			reduce(193), // ???, reduce: MuldivTerm
			reduce(193), // =~, reduce: MuldivTerm
			reduce(193), // !=~, reduce: MuldivTerm
			reduce(193), // ==, reduce: MuldivTerm
			reduce(193), // !=, reduce: MuldivTerm
			reduce(193), // <=>, reduce: MuldivTerm
			reduce(193), // >=, reduce: MuldivTerm
			reduce(193), // <, reduce: MuldivTerm
			reduce(193), // <=, reduce: MuldivTerm
			reduce(193), // ^, reduce: MuldivTerm
			reduce(193), // &, reduce: MuldivTerm
			reduce(193), // <<, reduce: MuldivTerm
			reduce(193), // >>>, reduce: MuldivTerm
			reduce(193), // +, reduce: MuldivTerm
			reduce(193), // -, reduce: MuldivTerm
			reduce(193), // .+, reduce: MuldivTerm
			reduce(193), // .-, reduce: MuldivTerm
			reduce(193), // *, reduce: MuldivTerm
			reduce(193), // /, reduce: MuldivTerm
			reduce(193), // //, reduce: MuldivTerm
			reduce(193), // %, reduce: MuldivTerm
			reduce(193), // .*, reduce: MuldivTerm
			reduce(193), // ./, reduce: MuldivTerm
			reduce(193), // .//, reduce: MuldivTerm
			shift(745),  // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1505
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(195), // $, reduce: DotTerm
			nil,         // empty
			reduce(195), // ;, reduce: DotTerm
			reduce(195), // {, reduce: DotTerm
			nil,         // }
			reduce(195), // =, reduce: DotTerm
			nil,         // unset
			nil,         // filter
			reduce(195), // >, reduce: DotTerm
			reduce(195), // >>, reduce: DotTerm
			reduce(195), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(195), // ||=, reduce: DotTerm
			reduce(195), // ^^=, reduce: DotTerm
			reduce(195), // &&=, reduce: DotTerm
			reduce(195), // ??=, reduce: DotTerm
			reduce(195), // ???=, reduce: DotTerm
			reduce(195), // |=, reduce: DotTerm
			reduce(195), // &=, reduce: DotTerm
			reduce(195), // ^=, reduce: DotTerm
			reduce(195), // <<=, reduce: DotTerm
			reduce(195), // >>=, reduce: DotTerm
			reduce(195), // >>>=, reduce: DotTerm
			reduce(195), // +=, reduce: DotTerm
			reduce(195), // .=, reduce: DotTerm
			reduce(195), // -=, reduce: DotTerm
			reduce(195), // *=, reduce: DotTerm
			reduce(195), // /=, reduce: DotTerm
			reduce(195), // //=, reduce: DotTerm
			reduce(195), // %=, reduce: DotTerm
			reduce(195), // **=, reduce: DotTerm
			reduce(195), // ?, reduce: DotTerm
			nil,         // :
			reduce(195), // ||, reduce: DotTerm
			reduce(195), // ^^, reduce: DotTerm
			reduce(195), // &&, reduce: DotTerm
			reduce(195), // ??, reduce: DotTerm
			reduce(195), // ???, reduce: DotTerm
			reduce(195), // =~, reduce: DotTerm
			reduce(195), // !=~, reduce: DotTerm
			reduce(195), // ==, reduce: DotTerm
			reduce(195), // !=, reduce: DotTerm
			reduce(195), // <=>, reduce: DotTerm
			reduce(195), // >=, reduce: DotTerm
			reduce(195), // <, reduce: DotTerm
			reduce(195), // <=, reduce: DotTerm
			reduce(195), // ^, reduce: DotTerm
			reduce(195), // &, reduce: DotTerm
			reduce(195), // <<, reduce: DotTerm
			reduce(195), // >>>, reduce: DotTerm
			reduce(195), // +, reduce: DotTerm
			reduce(195), // -, reduce: DotTerm
			reduce(195), // .+, reduce: DotTerm
			reduce(195), // .-, reduce: DotTerm
			reduce(195), // *, reduce: DotTerm
			reduce(195), // /, reduce: DotTerm
			reduce(195), // //, reduce: DotTerm
			reduce(195), // %, reduce: DotTerm
			reduce(195), // .*, reduce: DotTerm
			reduce(195), // ./, reduce: DotTerm
			reduce(195), // .//, reduce: DotTerm
			reduce(195), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1506
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1507
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1508
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(204), // $, reduce: PowTerm
			nil,         // empty
			reduce(204), // ;, reduce: PowTerm
			reduce(204), // {, reduce: PowTerm
			nil,         // }
			reduce(204), // =, reduce: PowTerm
			nil,         // unset
			nil,         // filter
			reduce(204), // >, reduce: PowTerm
			reduce(204), // >>, reduce: PowTerm
			reduce(204), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(204), // ||=, reduce: PowTerm
			reduce(204), // ^^=, reduce: PowTerm
			reduce(204), // &&=, reduce: PowTerm
			reduce(204), // ??=, reduce: PowTerm
			reduce(204), // ???=, reduce: PowTerm
			reduce(204), // |=, reduce: PowTerm
			reduce(204), // &=, reduce: PowTerm
			reduce(204), // ^=, reduce: PowTerm
			reduce(204), // <<=, reduce: PowTerm
			reduce(204), // >>=, reduce: PowTerm
			reduce(204), // >>>=, reduce: PowTerm
			reduce(204), // +=, reduce: PowTerm
			reduce(204), // .=, reduce: PowTerm
			reduce(204), // -=, reduce: PowTerm
			reduce(204), // *=, reduce: PowTerm
			reduce(204), // /=, reduce: PowTerm
			reduce(204), // //=, reduce: PowTerm
			reduce(204), // %=, reduce: PowTerm
			reduce(204), // **=, reduce: PowTerm
			reduce(204), // ?, reduce: PowTerm
			nil,         // :
			reduce(204), // ||, reduce: PowTerm
			reduce(204), // ^^, reduce: PowTerm
			reduce(204), // &&, reduce: PowTerm
			reduce(204), // ??, reduce: PowTerm
			reduce(204), // ???, reduce: PowTerm
			reduce(204), // =~, reduce: PowTerm
			reduce(204), // !=~, reduce: PowTerm
			reduce(204), // ==, reduce: PowTerm
			reduce(204), // !=, reduce: PowTerm
			reduce(204), // <=>, reduce: PowTerm
			reduce(204), // >=, reduce: PowTerm
			reduce(204), // <, reduce: PowTerm
			reduce(204), // <=, reduce: PowTerm
			reduce(204), // ^, reduce: PowTerm
			reduce(204), // &, reduce: PowTerm
			reduce(204), // <<, reduce: PowTerm
			reduce(204), // >>>, reduce: PowTerm
			reduce(204), // +, reduce: PowTerm
			reduce(204), // -, reduce: PowTerm
			reduce(204), // .+, reduce: PowTerm
			reduce(204), // .-, reduce: PowTerm
			reduce(204), // *, reduce: PowTerm
			reduce(204), // /, reduce: PowTerm
			reduce(204), // //, reduce: PowTerm
			reduce(204), // %, reduce: PowTerm
			reduce(204), // .*, reduce: PowTerm
			reduce(204), // ./, reduce: PowTerm
			reduce(204), // .//, reduce: PowTerm
			reduce(204), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1509
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			shift(2634), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1510
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2636), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(2637), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1511
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(214), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(214), // ?, reduce: MlrvalOrFunction
			reduce(214), // :, reduce: MlrvalOrFunction
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1512
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(211), // >, reduce: MlrvalOrFunction
			reduce(211), // >>, reduce: MlrvalOrFunction
			reduce(211), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(211), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(211), // ?, reduce: MlrvalOrFunction
			reduce(211), // :, reduce: MlrvalOrFunction
			reduce(211), // ||, reduce: MlrvalOrFunction
			reduce(211), // ^^, reduce: MlrvalOrFunction
			reduce(211), // &&, reduce: MlrvalOrFunction
			reduce(211), // ??, reduce: MlrvalOrFunction
			reduce(211), // ???, reduce: MlrvalOrFunction
			reduce(211), // =~, reduce: MlrvalOrFunction
			reduce(211), // !=~, reduce: MlrvalOrFunction
			reduce(211), // ==, reduce: MlrvalOrFunction
			reduce(211), // !=, reduce: MlrvalOrFunction
			reduce(211), // <=>, reduce: MlrvalOrFunction
			reduce(211), // >=, reduce: MlrvalOrFunction
			reduce(211), // <, reduce: MlrvalOrFunction
			reduce(211), // <=, reduce: MlrvalOrFunction
			reduce(211), // ^, reduce: MlrvalOrFunction
			reduce(211), // &, reduce: MlrvalOrFunction
			reduce(211), // <<, reduce: MlrvalOrFunction
			reduce(211), // >>>, reduce: MlrvalOrFunction
			reduce(211), // +, reduce: MlrvalOrFunction
			reduce(211), // -, reduce: MlrvalOrFunction
			reduce(211), // .+, reduce: MlrvalOrFunction
			reduce(211), // .-, reduce: MlrvalOrFunction
			reduce(211), // *, reduce: MlrvalOrFunction
			reduce(211), // /, reduce: MlrvalOrFunction
			reduce(211), // //, reduce: MlrvalOrFunction
			reduce(211), // %, reduce: MlrvalOrFunction
			reduce(211), // .*, reduce: MlrvalOrFunction
			reduce(211), // ./, reduce: MlrvalOrFunction
			reduce(211), // .//, reduce: MlrvalOrFunction
			reduce(211), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(211), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(211), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(211), // [[, reduce: MlrvalOrFunction
			reduce(211), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1513
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1514
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(110), // >, reduce: OosvarValue
			reduce(110), // >>, reduce: OosvarValue
			reduce(110), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(110), // ], reduce: OosvarValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(110), // ?, reduce: OosvarValue
			reduce(110), // :, reduce: OosvarValue
			reduce(110), // ||, reduce: OosvarValue
			reduce(110), // ^^, reduce: OosvarValue
			reduce(110), // &&, reduce: OosvarValue
			reduce(110), // ??, reduce: OosvarValue
			reduce(110), // ???, reduce: OosvarValue
			reduce(110), // =~, reduce: OosvarValue
			reduce(110), // !=~, reduce: OosvarValue
			reduce(110), // ==, reduce: OosvarValue
			reduce(110), // !=, reduce: OosvarValue
			reduce(110), // <=>, reduce: OosvarValue
			reduce(110), // >=, reduce: OosvarValue
			reduce(110), // <, reduce: OosvarValue
			reduce(110), // <=, reduce: OosvarValue
			reduce(110), // ^, reduce: OosvarValue
			reduce(110), // &, reduce: OosvarValue
			reduce(110), // <<, reduce: OosvarValue
			reduce(110), // >>>, reduce: OosvarValue
			reduce(110), // +, reduce: OosvarValue
			reduce(110), // -, reduce: OosvarValue
			reduce(110), // .+, reduce: OosvarValue
			reduce(110), // .-, reduce: OosvarValue
			reduce(110), // *, reduce: OosvarValue
			reduce(110), // /, reduce: OosvarValue
			reduce(110), // //, reduce: OosvarValue
			reduce(110), // %, reduce: OosvarValue
			reduce(110), // .*, reduce: OosvarValue
			reduce(110), // ./, reduce: OosvarValue
			reduce(110), // .//, reduce: OosvarValue
			reduce(110), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(110), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(110), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(110), // [[, reduce: OosvarValue
			reduce(110), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1515
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(112), // >, reduce: OosvarValue
			reduce(112), // >>, reduce: OosvarValue
			reduce(112), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(112), // ], reduce: OosvarValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(112), // ?, reduce: OosvarValue
			reduce(112), // :, reduce: OosvarValue
			reduce(112), // ||, reduce: OosvarValue
			reduce(112), // ^^, reduce: OosvarValue
			reduce(112), // &&, reduce: OosvarValue
			reduce(112), // ??, reduce: OosvarValue
			reduce(112), // ???, reduce: OosvarValue
			reduce(112), // =~, reduce: OosvarValue
			reduce(112), // !=~, reduce: OosvarValue
			reduce(112), // ==, reduce: OosvarValue
			reduce(112), // !=, reduce: OosvarValue
			reduce(112), // <=>, reduce: OosvarValue
			reduce(112), // >=, reduce: OosvarValue
			reduce(112), // <, reduce: OosvarValue
			reduce(112), // <=, reduce: OosvarValue
			reduce(112), // ^, reduce: OosvarValue
			reduce(112), // &, reduce: OosvarValue
			reduce(112), // <<, reduce: OosvarValue
			reduce(112), // >>>, reduce: OosvarValue
			reduce(112), // +, reduce: OosvarValue
			reduce(112), // -, reduce: OosvarValue
			reduce(112), // .+, reduce: OosvarValue
			reduce(112), // .-, reduce: OosvarValue
			reduce(112), // *, reduce: OosvarValue
			reduce(112), // /, reduce: OosvarValue
			reduce(112), // //, reduce: OosvarValue
			reduce(112), // %, reduce: OosvarValue
			reduce(112), // .*, reduce: OosvarValue
			reduce(112), // ./, reduce: OosvarValue
			reduce(112), // .//, reduce: OosvarValue
			reduce(112), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(112), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(112), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(112), // [[, reduce: OosvarValue
			reduce(112), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1516
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(111), // >, reduce: OosvarValue
			reduce(111), // >>, reduce: OosvarValue
			reduce(111), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(111), // ], reduce: OosvarValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(111), // ?, reduce: OosvarValue
			reduce(111), // :, reduce: OosvarValue
			reduce(111), // ||, reduce: OosvarValue
			reduce(111), // ^^, reduce: OosvarValue
			reduce(111), // &&, reduce: OosvarValue
			reduce(111), // ??, reduce: OosvarValue
			reduce(111), // ???, reduce: OosvarValue
			reduce(111), // =~, reduce: OosvarValue
			reduce(111), // !=~, reduce: OosvarValue
			reduce(111), // ==, reduce: OosvarValue
			reduce(111), // !=, reduce: OosvarValue
			reduce(111), // <=>, reduce: OosvarValue
			reduce(111), // >=, reduce: OosvarValue
			reduce(111), // <, reduce: OosvarValue
			reduce(111), // <=, reduce: OosvarValue
			reduce(111), // ^, reduce: OosvarValue
			reduce(111), // &, reduce: OosvarValue
			reduce(111), // <<, reduce: OosvarValue
			reduce(111), // >>>, reduce: OosvarValue
			reduce(111), // +, reduce: OosvarValue
			reduce(111), // -, reduce: OosvarValue
			reduce(111), // .+, reduce: OosvarValue
			reduce(111), // .-, reduce: OosvarValue
			reduce(111), // *, reduce: OosvarValue
			reduce(111), // /, reduce: OosvarValue
			reduce(111), // //, reduce: OosvarValue
			reduce(111), // %, reduce: OosvarValue
			reduce(111), // .*, reduce: OosvarValue
			reduce(111), // ./, reduce: OosvarValue
			reduce(111), // .//, reduce: OosvarValue
			reduce(111), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(111), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(111), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(111), // [[, reduce: OosvarValue
			reduce(111), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1517
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(99), // >, reduce: FieldValue
			reduce(99), // >>, reduce: FieldValue
			reduce(99), // |, reduce: FieldValue
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			reduce(99), // ], reduce: FieldValue
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(99), // ?, reduce: FieldValue
			reduce(99), // :, reduce: FieldValue
			reduce(99), // ||, reduce: FieldValue
			reduce(99), // ^^, reduce: FieldValue
			reduce(99), // &&, reduce: FieldValue
			reduce(99), // ??, reduce: FieldValue
			reduce(99), // ???, reduce: FieldValue
			reduce(99), // =~, reduce: FieldValue
			reduce(99), // !=~, reduce: FieldValue
			reduce(99), // ==, reduce: FieldValue
			reduce(99), // !=, reduce: FieldValue
			reduce(99), // <=>, reduce: FieldValue
			reduce(99), // >=, reduce: FieldValue
			reduce(99), // <, reduce: FieldValue
			reduce(99), // <=, reduce: FieldValue
			reduce(99), // ^, reduce: FieldValue
			reduce(99), // &, reduce: FieldValue
			reduce(99), // <<, reduce: FieldValue
			reduce(99), // >>>, reduce: FieldValue
			reduce(99), // +, reduce: FieldValue
			reduce(99), // -, reduce: FieldValue
			reduce(99), // .+, reduce: FieldValue
			reduce(99), // .-, reduce: FieldValue
			reduce(99), // *, reduce: FieldValue
			reduce(99), // /, reduce: FieldValue
			reduce(99), // //, reduce: FieldValue
			reduce(99), // %, reduce: FieldValue
			reduce(99), // .*, reduce: FieldValue
			reduce(99), // ./, reduce: FieldValue
			reduce(99), // .//, reduce: FieldValue
			reduce(99), // ., reduce: FieldValue
			nil,        // !
			nil,        // ~
			reduce(99), // **, reduce: FieldValue
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(99), // [, reduce: FieldValue
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(99), // [[, reduce: FieldValue
			reduce(99), // [[[, reduce: FieldValue
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1518
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(101), // >, reduce: FieldValue
			reduce(101), // >>, reduce: FieldValue
			reduce(101), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(101), // ], reduce: FieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(101), // ?, reduce: FieldValue
			reduce(101), // :, reduce: FieldValue
			reduce(101), // ||, reduce: FieldValue
			reduce(101), // ^^, reduce: FieldValue
			reduce(101), // &&, reduce: FieldValue
			reduce(101), // ??, reduce: FieldValue
			reduce(101), // ???, reduce: FieldValue
			reduce(101), // =~, reduce: FieldValue
			reduce(101), // !=~, reduce: FieldValue
			reduce(101), // ==, reduce: FieldValue
			reduce(101), // !=, reduce: FieldValue
			reduce(101), // <=>, reduce: FieldValue
			reduce(101), // >=, reduce: FieldValue
			reduce(101), // <, reduce: FieldValue
			reduce(101), // <=, reduce: FieldValue
			reduce(101), // ^, reduce: FieldValue
			reduce(101), // &, reduce: FieldValue
			reduce(101), // <<, reduce: FieldValue
			reduce(101), // >>>, reduce: FieldValue
			reduce(101), // +, reduce: FieldValue
			reduce(101), // -, reduce: FieldValue
			reduce(101), // .+, reduce: FieldValue
			reduce(101), // .-, reduce: FieldValue
			reduce(101), // *, reduce: FieldValue
			reduce(101), // /, reduce: FieldValue
			reduce(101), // //, reduce: FieldValue
			reduce(101), // %, reduce: FieldValue
			reduce(101), // .*, reduce: FieldValue
			reduce(101), // ./, reduce: FieldValue
			reduce(101), // .//, reduce: FieldValue
			reduce(101), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(101), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(101), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(101), // [[, reduce: FieldValue
			reduce(101), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1519
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(100), // >, reduce: FieldValue
			reduce(100), // >>, reduce: FieldValue
			reduce(100), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(100), // ], reduce: FieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(100), // ?, reduce: FieldValue
			reduce(100), // :, reduce: FieldValue
			reduce(100), // ||, reduce: FieldValue
			reduce(100), // ^^, reduce: FieldValue
			reduce(100), // &&, reduce: FieldValue
			reduce(100), // ??, reduce: FieldValue
			reduce(100), // ???, reduce: FieldValue
			reduce(100), // =~, reduce: FieldValue
			reduce(100), // !=~, reduce: FieldValue
			reduce(100), // ==, reduce: FieldValue
			reduce(100), // !=, reduce: FieldValue
			reduce(100), // <=>, reduce: FieldValue
			reduce(100), // >=, reduce: FieldValue
			reduce(100), // <, reduce: FieldValue
			reduce(100), // <=, reduce: FieldValue
			reduce(100), // ^, reduce: FieldValue
			reduce(100), // &, reduce: FieldValue
			reduce(100), // <<, reduce: FieldValue
			reduce(100), // >>>, reduce: FieldValue
			reduce(100), // +, reduce: FieldValue
			reduce(100), // -, reduce: FieldValue
			reduce(100), // .+, reduce: FieldValue
			reduce(100), // .-, reduce: FieldValue
			reduce(100), // *, reduce: FieldValue
			reduce(100), // /, reduce: FieldValue
			reduce(100), // //, reduce: FieldValue
			reduce(100), // %, reduce: FieldValue
			reduce(100), // .*, reduce: FieldValue
			reduce(100), // ./, reduce: FieldValue
			reduce(100), // .//, reduce: FieldValue
			reduce(100), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(100), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(100), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(100), // [[, reduce: FieldValue
			reduce(100), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1520
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(213), // >, reduce: MlrvalOrFunction
			reduce(213), // >>, reduce: MlrvalOrFunction
			reduce(213), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(213), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(213), // ?, reduce: MlrvalOrFunction
			reduce(213), // :, reduce: MlrvalOrFunction
			reduce(213), // ||, reduce: MlrvalOrFunction
			reduce(213), // ^^, reduce: MlrvalOrFunction
			reduce(213), // &&, reduce: MlrvalOrFunction
			reduce(213), // ??, reduce: MlrvalOrFunction
			reduce(213), // ???, reduce: MlrvalOrFunction
			reduce(213), // =~, reduce: MlrvalOrFunction
			reduce(213), // !=~, reduce: MlrvalOrFunction
			reduce(213), // ==, reduce: MlrvalOrFunction
			reduce(213), // !=, reduce: MlrvalOrFunction
			reduce(213), // <=>, reduce: MlrvalOrFunction
			reduce(213), // >=, reduce: MlrvalOrFunction
			reduce(213), // <, reduce: MlrvalOrFunction
			reduce(213), // <=, reduce: MlrvalOrFunction
			reduce(213), // ^, reduce: MlrvalOrFunction
			reduce(213), // &, reduce: MlrvalOrFunction
			reduce(213), // <<, reduce: MlrvalOrFunction
			reduce(213), // >>>, reduce: MlrvalOrFunction
			reduce(213), // +, reduce: MlrvalOrFunction
			reduce(213), // -, reduce: MlrvalOrFunction
			reduce(213), // .+, reduce: MlrvalOrFunction
			reduce(213), // .-, reduce: MlrvalOrFunction
			reduce(213), // *, reduce: MlrvalOrFunction
			reduce(213), // /, reduce: MlrvalOrFunction
			reduce(213), // //, reduce: MlrvalOrFunction
			reduce(213), // %, reduce: MlrvalOrFunction
			reduce(213), // .*, reduce: MlrvalOrFunction
			reduce(213), // ./, reduce: MlrvalOrFunction
			reduce(213), // .//, reduce: MlrvalOrFunction
			reduce(213), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(213), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(213), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(213), // [[, reduce: MlrvalOrFunction
			reduce(213), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1521
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(233), // >, reduce: MlrvalOrFunction
			reduce(233), // >>, reduce: MlrvalOrFunction
			reduce(233), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(233), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(233), // ?, reduce: MlrvalOrFunction
			reduce(233), // :, reduce: MlrvalOrFunction
			reduce(233), // ||, reduce: MlrvalOrFunction
			reduce(233), // ^^, reduce: MlrvalOrFunction
			reduce(233), // &&, reduce: MlrvalOrFunction
			reduce(233), // ??, reduce: MlrvalOrFunction
			reduce(233), // ???, reduce: MlrvalOrFunction
			reduce(233), // =~, reduce: MlrvalOrFunction
			reduce(233), // !=~, reduce: MlrvalOrFunction
			reduce(233), // ==, reduce: MlrvalOrFunction
			reduce(233), // !=, reduce: MlrvalOrFunction
			reduce(233), // <=>, reduce: MlrvalOrFunction
			reduce(233), // >=, reduce: MlrvalOrFunction
			reduce(233), // <, reduce: MlrvalOrFunction
			reduce(233), // <=, reduce: MlrvalOrFunction
			reduce(233), // ^, reduce: MlrvalOrFunction
			reduce(233), // &, reduce: MlrvalOrFunction
			reduce(233), // <<, reduce: MlrvalOrFunction
			reduce(233), // >>>, reduce: MlrvalOrFunction
			reduce(233), // +, reduce: MlrvalOrFunction
			reduce(233), // -, reduce: MlrvalOrFunction
			reduce(233), // .+, reduce: MlrvalOrFunction
			reduce(233), // .-, reduce: MlrvalOrFunction
			reduce(233), // *, reduce: MlrvalOrFunction
			reduce(233), // /, reduce: MlrvalOrFunction
			reduce(233), // //, reduce: MlrvalOrFunction
			reduce(233), // %, reduce: MlrvalOrFunction
			reduce(233), // .*, reduce: MlrvalOrFunction
			reduce(233), // ./, reduce: MlrvalOrFunction
			reduce(233), // .//, reduce: MlrvalOrFunction
			reduce(233), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(233), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(233), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(233), // [[, reduce: MlrvalOrFunction
			reduce(233), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1522
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(210), // >, reduce: MlrvalOrFunction
			reduce(210), // >>, reduce: MlrvalOrFunction
			reduce(210), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(210), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(210), // ?, reduce: MlrvalOrFunction
			reduce(210), // :, reduce: MlrvalOrFunction
			reduce(210), // ||, reduce: MlrvalOrFunction
			reduce(210), // ^^, reduce: MlrvalOrFunction
			reduce(210), // &&, reduce: MlrvalOrFunction
			reduce(210), // ??, reduce: MlrvalOrFunction
			reduce(210), // ???, reduce: MlrvalOrFunction
			reduce(210), // =~, reduce: MlrvalOrFunction
			reduce(210), // !=~, reduce: MlrvalOrFunction
			reduce(210), // ==, reduce: MlrvalOrFunction
			reduce(210), // !=, reduce: MlrvalOrFunction
			reduce(210), // <=>, reduce: MlrvalOrFunction
			reduce(210), // >=, reduce: MlrvalOrFunction
			reduce(210), // <, reduce: MlrvalOrFunction
			reduce(210), // <=, reduce: MlrvalOrFunction
			reduce(210), // ^, reduce: MlrvalOrFunction
			reduce(210), // &, reduce: MlrvalOrFunction
			reduce(210), // <<, reduce: MlrvalOrFunction
			reduce(210), // >>>, reduce: MlrvalOrFunction
			reduce(210), // +, reduce: MlrvalOrFunction
			reduce(210), // -, reduce: MlrvalOrFunction
			reduce(210), // .+, reduce: MlrvalOrFunction
			reduce(210), // .-, reduce: MlrvalOrFunction
			reduce(210), // *, reduce: MlrvalOrFunction
			reduce(210), // /, reduce: MlrvalOrFunction
			reduce(210), // //, reduce: MlrvalOrFunction
			reduce(210), // %, reduce: MlrvalOrFunction
			reduce(210), // .*, reduce: MlrvalOrFunction
			reduce(210), // ./, reduce: MlrvalOrFunction
			reduce(210), // .//, reduce: MlrvalOrFunction
			reduce(210), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(210), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(210), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(210), // [[, reduce: MlrvalOrFunction
			reduce(210), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1523
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(102), // >, reduce: FieldValue
			reduce(102), // >>, reduce: FieldValue
			reduce(102), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(102), // ], reduce: FieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(102), // ?, reduce: FieldValue
			reduce(102), // :, reduce: FieldValue
			reduce(102), // ||, reduce: FieldValue
			reduce(102), // ^^, reduce: FieldValue
			reduce(102), // &&, reduce: FieldValue
			reduce(102), // ??, reduce: FieldValue
			reduce(102), // ???, reduce: FieldValue
			reduce(102), // =~, reduce: FieldValue
			reduce(102), // !=~, reduce: FieldValue
			reduce(102), // ==, reduce: FieldValue
			reduce(102), // !=, reduce: FieldValue
			reduce(102), // <=>, reduce: FieldValue
			reduce(102), // >=, reduce: FieldValue
			reduce(102), // <, reduce: FieldValue
			reduce(102), // <=, reduce: FieldValue
			reduce(102), // ^, reduce: FieldValue
			reduce(102), // &, reduce: FieldValue
			reduce(102), // <<, reduce: FieldValue
			reduce(102), // >>>, reduce: FieldValue
			reduce(102), // +, reduce: FieldValue
			reduce(102), // -, reduce: FieldValue
			reduce(102), // .+, reduce: FieldValue
			reduce(102), // .-, reduce: FieldValue
			reduce(102), // *, reduce: FieldValue
			reduce(102), // /, reduce: FieldValue
			reduce(102), // //, reduce: FieldValue
			reduce(102), // %, reduce: FieldValue
			reduce(102), // .*, reduce: FieldValue
			reduce(102), // ./, reduce: FieldValue
			reduce(102), // .//, reduce: FieldValue
			reduce(102), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(102), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(102), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(102), // [[, reduce: FieldValue
			reduce(102), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1524
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(103), // >, reduce: FieldValue
			reduce(103), // >>, reduce: FieldValue
			reduce(103), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(103), // ], reduce: FieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(103), // ?, reduce: FieldValue
			reduce(103), // :, reduce: FieldValue
			reduce(103), // ||, reduce: FieldValue
			reduce(103), // ^^, reduce: FieldValue
			reduce(103), // &&, reduce: FieldValue
			reduce(103), // ??, reduce: FieldValue
			reduce(103), // ???, reduce: FieldValue
			reduce(103), // =~, reduce: FieldValue
			reduce(103), // !=~, reduce: FieldValue
			reduce(103), // ==, reduce: FieldValue
			reduce(103), // !=, reduce: FieldValue
			reduce(103), // <=>, reduce: FieldValue
			reduce(103), // >=, reduce: FieldValue
			reduce(103), // <, reduce: FieldValue
			reduce(103), // <=, reduce: FieldValue
			reduce(103), // ^, reduce: FieldValue
			reduce(103), // &, reduce: FieldValue
			reduce(103), // <<, reduce: FieldValue
			reduce(103), // >>>, reduce: FieldValue
			reduce(103), // +, reduce: FieldValue
			reduce(103), // -, reduce: FieldValue
			reduce(103), // .+, reduce: FieldValue
			reduce(103), // .-, reduce: FieldValue
			reduce(103), // *, reduce: FieldValue
			reduce(103), // /, reduce: FieldValue
			reduce(103), // //, reduce: FieldValue
			reduce(103), // %, reduce: FieldValue
			reduce(103), // .*, reduce: FieldValue
			reduce(103), // ./, reduce: FieldValue
			reduce(103), // .//, reduce: FieldValue
			reduce(103), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(103), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(103), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(103), // [[, reduce: FieldValue
			reduce(103), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1525
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(104), // >, reduce: DirectFieldValue
			reduce(104), // >>, reduce: DirectFieldValue
			reduce(104), // |, reduce: DirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(104), // ], reduce: DirectFieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(104), // ?, reduce: DirectFieldValue
			reduce(104), // :, reduce: DirectFieldValue
			reduce(104), // ||, reduce: DirectFieldValue
			reduce(104), // ^^, reduce: DirectFieldValue
			reduce(104), // &&, reduce: DirectFieldValue
			reduce(104), // ??, reduce: DirectFieldValue
			reduce(104), // ???, reduce: DirectFieldValue
			reduce(104), // =~, reduce: DirectFieldValue
			reduce(104), // !=~, reduce: DirectFieldValue
			reduce(104), // ==, reduce: DirectFieldValue
			reduce(104), // !=, reduce: DirectFieldValue
			reduce(104), // <=>, reduce: DirectFieldValue
			reduce(104), // >=, reduce: DirectFieldValue
			reduce(104), // <, reduce: DirectFieldValue
			reduce(104), // <=, reduce: DirectFieldValue
			reduce(104), // ^, reduce: DirectFieldValue
			reduce(104), // &, reduce: DirectFieldValue
			reduce(104), // <<, reduce: DirectFieldValue
			reduce(104), // >>>, reduce: DirectFieldValue
			reduce(104), // +, reduce: DirectFieldValue
			reduce(104), // -, reduce: DirectFieldValue
			reduce(104), // .+, reduce: DirectFieldValue
			reduce(104), // .-, reduce: DirectFieldValue
			reduce(104), // *, reduce: DirectFieldValue
			reduce(104), // /, reduce: DirectFieldValue
			reduce(104), // //, reduce: DirectFieldValue
			reduce(104), // %, reduce: DirectFieldValue
			reduce(104), // .*, reduce: DirectFieldValue
			reduce(104), // ./, reduce: DirectFieldValue
			reduce(104), // .//, reduce: DirectFieldValue
			reduce(104), // ., reduce: DirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(104), // **, reduce: DirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(104), // [, reduce: DirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(104), // [[, reduce: DirectFieldValue
			reduce(104), // [[[, reduce: DirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1526
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1527
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(106), // >, reduce: BracedFieldValue
			reduce(106), // >>, reduce: BracedFieldValue
			reduce(106), // |, reduce: BracedFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(106), // ], reduce: BracedFieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(106), // ?, reduce: BracedFieldValue
			reduce(106), // :, reduce: BracedFieldValue
			reduce(106), // ||, reduce: BracedFieldValue
			reduce(106), // ^^, reduce: BracedFieldValue
			reduce(106), // &&, reduce: BracedFieldValue
			reduce(106), // ??, reduce: BracedFieldValue
			reduce(106), // ???, reduce: BracedFieldValue
			reduce(106), // =~, reduce: BracedFieldValue
			reduce(106), // !=~, reduce: BracedFieldValue
			reduce(106), // ==, reduce: BracedFieldValue
			reduce(106), // !=, reduce: BracedFieldValue
			reduce(106), // <=>, reduce: BracedFieldValue
			reduce(106), // >=, reduce: BracedFieldValue
			reduce(106), // <, reduce: BracedFieldValue
			reduce(106), // <=, reduce: BracedFieldValue
			reduce(106), // ^, reduce: BracedFieldValue
			reduce(106), // &, reduce: BracedFieldValue
			reduce(106), // <<, reduce: BracedFieldValue
			reduce(106), // >>>, reduce: BracedFieldValue
			reduce(106), // +, reduce: BracedFieldValue
			reduce(106), // -, reduce: BracedFieldValue
			reduce(106), // .+, reduce: BracedFieldValue
			reduce(106), // .-, reduce: BracedFieldValue
			reduce(106), // *, reduce: BracedFieldValue
			reduce(106), // /, reduce: BracedFieldValue
			reduce(106), // //, reduce: BracedFieldValue
			reduce(106), // %, reduce: BracedFieldValue
			reduce(106), // .*, reduce: BracedFieldValue
			reduce(106), // ./, reduce: BracedFieldValue
			reduce(106), // .//, reduce: BracedFieldValue
			reduce(106), // ., reduce: BracedFieldValue
			nil,         // !
			nil,         // ~
			reduce(106), // **, reduce: BracedFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(106), // [, reduce: BracedFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(106), // [[, reduce: BracedFieldValue
			reduce(106), // [[[, reduce: BracedFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1528
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1529
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1530
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(109), // >, reduce: FullSrec
			reduce(109), // >>, reduce: FullSrec
			reduce(109), // |, reduce: FullSrec
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(109), // ], reduce: FullSrec
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(109), // ?, reduce: FullSrec
			reduce(109), // :, reduce: FullSrec
			reduce(109), // ||, reduce: FullSrec
			reduce(109), // ^^, reduce: FullSrec
			reduce(109), // &&, reduce: FullSrec
			reduce(109), // ??, reduce: FullSrec
			reduce(109), // ???, reduce: FullSrec
			reduce(109), // =~, reduce: FullSrec
			reduce(109), // !=~, reduce: FullSrec
			reduce(109), // ==, reduce: FullSrec
			reduce(109), // !=, reduce: FullSrec
			reduce(109), // <=>, reduce: FullSrec
			reduce(109), // >=, reduce: FullSrec
			reduce(109), // <, reduce: FullSrec
			reduce(109), // <=, reduce: FullSrec
			reduce(109), // ^, reduce: FullSrec
			reduce(109), // &, reduce: FullSrec
			reduce(109), // <<, reduce: FullSrec
			reduce(109), // >>>, reduce: FullSrec
			reduce(109), // +, reduce: FullSrec
			reduce(109), // -, reduce: FullSrec
			reduce(109), // .+, reduce: FullSrec
			reduce(109), // .-, reduce: FullSrec
			reduce(109), // *, reduce: FullSrec
			reduce(109), // /, reduce: FullSrec
			reduce(109), // //, reduce: FullSrec
			reduce(109), // %, reduce: FullSrec
			reduce(109), // .*, reduce: FullSrec
			reduce(109), // ./, reduce: FullSrec
			reduce(109), // .//, reduce: FullSrec
			reduce(109), // ., reduce: FullSrec
			nil,         // !
			nil,         // ~
			reduce(109), // **, reduce: FullSrec
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(109), // [, reduce: FullSrec
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(109), // [[, reduce: FullSrec
			reduce(109), // [[[, reduce: FullSrec
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1531
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(212), // >, reduce: MlrvalOrFunction
			reduce(212), // >>, reduce: MlrvalOrFunction
			reduce(212), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(212), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(212), // ?, reduce: MlrvalOrFunction
			reduce(212), // :, reduce: MlrvalOrFunction
			reduce(212), // ||, reduce: MlrvalOrFunction
			reduce(212), // ^^, reduce: MlrvalOrFunction
			reduce(212), // &&, reduce: MlrvalOrFunction
			reduce(212), // ??, reduce: MlrvalOrFunction
			reduce(212), // ???, reduce: MlrvalOrFunction
			reduce(212), // =~, reduce: MlrvalOrFunction
			reduce(212), // !=~, reduce: MlrvalOrFunction
			reduce(212), // ==, reduce: MlrvalOrFunction
			reduce(212), // !=, reduce: MlrvalOrFunction
			reduce(212), // <=>, reduce: MlrvalOrFunction
			reduce(212), // >=, reduce: MlrvalOrFunction
			reduce(212), // <, reduce: MlrvalOrFunction
			reduce(212), // <=, reduce: MlrvalOrFunction
			reduce(212), // ^, reduce: MlrvalOrFunction
			reduce(212), // &, reduce: MlrvalOrFunction
			reduce(212), // <<, reduce: MlrvalOrFunction
			reduce(212), // >>>, reduce: MlrvalOrFunction
			reduce(212), // +, reduce: MlrvalOrFunction
			reduce(212), // -, reduce: MlrvalOrFunction
			reduce(212), // .+, reduce: MlrvalOrFunction
			reduce(212), // .-, reduce: MlrvalOrFunction
			reduce(212), // *, reduce: MlrvalOrFunction
			reduce(212), // /, reduce: MlrvalOrFunction
			reduce(212), // //, reduce: MlrvalOrFunction
			reduce(212), // %, reduce: MlrvalOrFunction
			reduce(212), // .*, reduce: MlrvalOrFunction
			reduce(212), // ./, reduce: MlrvalOrFunction
			reduce(212), // .//, reduce: MlrvalOrFunction
			reduce(212), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(212), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(212), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(212), // [[, reduce: MlrvalOrFunction
			reduce(212), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1532
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(113), // >, reduce: DirectOosvarValue
			reduce(113), // >>, reduce: DirectOosvarValue
			reduce(113), // |, reduce: DirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(113), // ], reduce: DirectOosvarValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(113), // ?, reduce: DirectOosvarValue
			reduce(113), // :, reduce: DirectOosvarValue
			reduce(113), // ||, reduce: DirectOosvarValue
			reduce(113), // ^^, reduce: DirectOosvarValue
			reduce(113), // &&, reduce: DirectOosvarValue
			reduce(113), // ??, reduce: DirectOosvarValue
			reduce(113), // ???, reduce: DirectOosvarValue
			reduce(113), // =~, reduce: DirectOosvarValue
			reduce(113), // !=~, reduce: DirectOosvarValue
			reduce(113), // ==, reduce: DirectOosvarValue
			reduce(113), // !=, reduce: DirectOosvarValue
			reduce(113), // <=>, reduce: DirectOosvarValue
			reduce(113), // >=, reduce: DirectOosvarValue
			reduce(113), // <, reduce: DirectOosvarValue
			reduce(113), // <=, reduce: DirectOosvarValue
			reduce(113), // ^, reduce: DirectOosvarValue
			reduce(113), // &, reduce: DirectOosvarValue
			reduce(113), // <<, reduce: DirectOosvarValue
			reduce(113), // >>>, reduce: DirectOosvarValue
			reduce(113), // +, reduce: DirectOosvarValue
			reduce(113), // -, reduce: DirectOosvarValue
			reduce(113), // .+, reduce: DirectOosvarValue
			reduce(113), // .-, reduce: DirectOosvarValue
			reduce(113), // *, reduce: DirectOosvarValue
			reduce(113), // /, reduce: DirectOosvarValue
			reduce(113), // //, reduce: DirectOosvarValue
			reduce(113), // %, reduce: DirectOosvarValue
			reduce(113), // .*, reduce: DirectOosvarValue
			reduce(113), // ./, reduce: DirectOosvarValue
			reduce(113), // .//, reduce: DirectOosvarValue
			reduce(113), // ., reduce: DirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(113), // **, reduce: DirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(113), // [, reduce: DirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(113), // [[, reduce: DirectOosvarValue
			reduce(113), // [[[, reduce: DirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1533
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1534
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(115), // >, reduce: BracedOosvarValue
			reduce(115), // >>, reduce: BracedOosvarValue
			reduce(115), // |, reduce: BracedOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(115), // ], reduce: BracedOosvarValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(115), // ?, reduce: BracedOosvarValue
			reduce(115), // :, reduce: BracedOosvarValue
			reduce(115), // ||, reduce: BracedOosvarValue
			reduce(115), // ^^, reduce: BracedOosvarValue
			reduce(115), // &&, reduce: BracedOosvarValue
			reduce(115), // ??, reduce: BracedOosvarValue
			reduce(115), // ???, reduce: BracedOosvarValue
			reduce(115), // =~, reduce: BracedOosvarValue
			reduce(115), // !=~, reduce: BracedOosvarValue
			reduce(115), // ==, reduce: BracedOosvarValue
			reduce(115), // !=, reduce: BracedOosvarValue
			reduce(115), // <=>, reduce: BracedOosvarValue
			reduce(115), // >=, reduce: BracedOosvarValue
			reduce(115), // <, reduce: BracedOosvarValue
			reduce(115), // <=, reduce: BracedOosvarValue
			reduce(115), // ^, reduce: BracedOosvarValue
			reduce(115), // &, reduce: BracedOosvarValue
			reduce(115), // <<, reduce: BracedOosvarValue
			reduce(115), // >>>, reduce: BracedOosvarValue
			reduce(115), // +, reduce: BracedOosvarValue
			reduce(115), // -, reduce: BracedOosvarValue
			reduce(115), // .+, reduce: BracedOosvarValue
			reduce(115), // .-, reduce: BracedOosvarValue
			reduce(115), // *, reduce: BracedOosvarValue
			reduce(115), // /, reduce: BracedOosvarValue
			reduce(115), // //, reduce: BracedOosvarValue
			reduce(115), // %, reduce: BracedOosvarValue
			reduce(115), // .*, reduce: BracedOosvarValue
			reduce(115), // ./, reduce: BracedOosvarValue
			reduce(115), // .//, reduce: BracedOosvarValue
			reduce(115), // ., reduce: BracedOosvarValue
			nil,         // !
			nil,         // ~
			reduce(115), // **, reduce: BracedOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(115), // [, reduce: BracedOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(115), // [[, reduce: BracedOosvarValue
			reduce(115), // [[[, reduce: BracedOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1535
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(116), // >, reduce: FullOosvar
			reduce(116), // >>, reduce: FullOosvar
			reduce(116), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(116), // ], reduce: FullOosvar
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(116), // ?, reduce: FullOosvar
			reduce(116), // :, reduce: FullOosvar
			reduce(116), // ||, reduce: FullOosvar
			reduce(116), // ^^, reduce: FullOosvar
			reduce(116), // &&, reduce: FullOosvar
			reduce(116), // ??, reduce: FullOosvar
			reduce(116), // ???, reduce: FullOosvar
			reduce(116), // =~, reduce: FullOosvar
			reduce(116), // !=~, reduce: FullOosvar
			reduce(116), // ==, reduce: FullOosvar
			reduce(116), // !=, reduce: FullOosvar
			reduce(116), // <=>, reduce: FullOosvar
			reduce(116), // >=, reduce: FullOosvar
			reduce(116), // <, reduce: FullOosvar
			reduce(116), // <=, reduce: FullOosvar
			reduce(116), // ^, reduce: FullOosvar
			reduce(116), // &, reduce: FullOosvar
			reduce(116), // <<, reduce: FullOosvar
			reduce(116), // >>>, reduce: FullOosvar
			reduce(116), // +, reduce: FullOosvar
			reduce(116), // -, reduce: FullOosvar
			reduce(116), // .+, reduce: FullOosvar
			reduce(116), // .-, reduce: FullOosvar
			reduce(116), // *, reduce: FullOosvar
			reduce(116), // /, reduce: FullOosvar
			reduce(116), // //, reduce: FullOosvar
			reduce(116), // %, reduce: FullOosvar
			reduce(116), // .*, reduce: FullOosvar
			reduce(116), // ./, reduce: FullOosvar
			reduce(116), // .//, reduce: FullOosvar
			reduce(116), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(116), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(116), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(116), // [[, reduce: FullOosvar
			reduce(116), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1536
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(117), // >, reduce: FullOosvar
			reduce(117), // >>, reduce: FullOosvar
			reduce(117), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(117), // ], reduce: FullOosvar
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(117), // ?, reduce: FullOosvar
			reduce(117), // :, reduce: FullOosvar
			reduce(117), // ||, reduce: FullOosvar
			reduce(117), // ^^, reduce: FullOosvar
			reduce(117), // &&, reduce: FullOosvar
			reduce(117), // ??, reduce: FullOosvar
			reduce(117), // ???, reduce: FullOosvar
			reduce(117), // =~, reduce: FullOosvar
			reduce(117), // !=~, reduce: FullOosvar
			reduce(117), // ==, reduce: FullOosvar
			reduce(117), // !=, reduce: FullOosvar
			reduce(117), // <=>, reduce: FullOosvar
			reduce(117), // >=, reduce: FullOosvar
			reduce(117), // <, reduce: FullOosvar
			reduce(117), // <=, reduce: FullOosvar
			reduce(117), // ^, reduce: FullOosvar
			reduce(117), // &, reduce: FullOosvar
			reduce(117), // <<, reduce: FullOosvar
			reduce(117), // >>>, reduce: FullOosvar
			reduce(117), // +, reduce: FullOosvar
			reduce(117), // -, reduce: FullOosvar
			reduce(117), // .+, reduce: FullOosvar
			reduce(117), // .-, reduce: FullOosvar
			reduce(117), // *, reduce: FullOosvar
			reduce(117), // /, reduce: FullOosvar
			reduce(117), // //, reduce: FullOosvar
			reduce(117), // %, reduce: FullOosvar
			reduce(117), // .*, reduce: FullOosvar
			reduce(117), // ./, reduce: FullOosvar
			reduce(117), // .//, reduce: FullOosvar
			reduce(117), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(117), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(117), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(117), // [[, reduce: FullOosvar
			reduce(117), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1537
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(118), // >, reduce: LocalVariable
			reduce(118), // >>, reduce: LocalVariable
			reduce(118), // |, reduce: LocalVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(270), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(118), // ], reduce: LocalVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(118), // ?, reduce: LocalVariable
			reduce(118), // :, reduce: LocalVariable
			reduce(118), // ||, reduce: LocalVariable
			reduce(118), // ^^, reduce: LocalVariable
			reduce(118), // &&, reduce: LocalVariable
			reduce(118), // ??, reduce: LocalVariable
			reduce(118), // ???, reduce: LocalVariable
			reduce(118), // =~, reduce: LocalVariable
			reduce(118), // !=~, reduce: LocalVariable
			reduce(118), // ==, reduce: LocalVariable
			reduce(118), // !=, reduce: LocalVariable
			reduce(118), // <=>, reduce: LocalVariable
			reduce(118), // >=, reduce: LocalVariable
			reduce(118), // <, reduce: LocalVariable
			reduce(118), // <=, reduce: LocalVariable
			reduce(118), // ^, reduce: LocalVariable
			reduce(118), // &, reduce: LocalVariable
			reduce(118), // <<, reduce: LocalVariable
			reduce(118), // >>>, reduce: LocalVariable
			reduce(118), // +, reduce: LocalVariable
			reduce(118), // -, reduce: LocalVariable
			reduce(118), // .+, reduce: LocalVariable
			reduce(118), // .-, reduce: LocalVariable
			reduce(118), // *, reduce: LocalVariable
			reduce(118), // /, reduce: LocalVariable
			reduce(118), // //, reduce: LocalVariable
			reduce(118), // %, reduce: LocalVariable
			reduce(118), // .*, reduce: LocalVariable
			reduce(118), // ./, reduce: LocalVariable
			reduce(118), // .//, reduce: LocalVariable
			reduce(118), // ., reduce: LocalVariable
			nil,         // !
			nil,         // ~
			reduce(118), // **, reduce: LocalVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(118), // [, reduce: LocalVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(118), // [[, reduce: LocalVariable
			reduce(118), // [[[, reduce: LocalVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1538
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(147), // ], reduce: Rvalue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			reduce(147), // :, reduce: Rvalue
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1539
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(148), // ], reduce: PrecedenceChainStart
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			reduce(148), // :, reduce: PrecedenceChainStart
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1540
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(150), // ], reduce: TernaryTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			shift(2643), // ?
			reduce(150), // :, reduce: TernaryTerm
			shift(2644), // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1541
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(2646), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1542
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(152), // ], reduce: LogicalOrTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(152), // ?, reduce: LogicalOrTerm
			reduce(152), // :, reduce: LogicalOrTerm
			reduce(152), // ||, reduce: LogicalOrTerm
			shift(2647), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1543
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(154), // ], reduce: LogicalXORTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(154), // ?, reduce: LogicalXORTerm
			reduce(154), // :, reduce: LogicalXORTerm
			reduce(154), // ||, reduce: LogicalXORTerm
			reduce(154), // ^^, reduce: LogicalXORTerm
			shift(2648), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1544
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(156), // ], reduce: LogicalAndTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(156), // ?, reduce: LogicalAndTerm
			reduce(156), // :, reduce: LogicalAndTerm
			reduce(156), // ||, reduce: LogicalAndTerm
			reduce(156), // ^^, reduce: LogicalAndTerm
			reduce(156), // &&, reduce: LogicalAndTerm
			shift(2649), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1545
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(158), // ], reduce: AbsentCoalesceTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(158), // ?, reduce: AbsentCoalesceTerm
			reduce(158), // :, reduce: AbsentCoalesceTerm
			reduce(158), // ||, reduce: AbsentCoalesceTerm
			reduce(158), // ^^, reduce: AbsentCoalesceTerm
			reduce(158), // &&, reduce: AbsentCoalesceTerm
			reduce(158), // ??, reduce: AbsentCoalesceTerm
			shift(2650), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1546
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(160), // ], reduce: EmptyCoalesceTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(160), // ?, reduce: EmptyCoalesceTerm
			reduce(160), // :, reduce: EmptyCoalesceTerm
			reduce(160), // ||, reduce: EmptyCoalesceTerm
			reduce(160), // ^^, reduce: EmptyCoalesceTerm
			reduce(160), // &&, reduce: EmptyCoalesceTerm
			reduce(160), // ??, reduce: EmptyCoalesceTerm
			reduce(160), // ???, reduce: EmptyCoalesceTerm
			shift(2651), // =~
			shift(2652), // !=~
			shift(2653), // ==
			shift(2654), // !=
			shift(2655), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1547
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2656), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(166), // ], reduce: EqneTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(166), // ?, reduce: EqneTerm
			reduce(166), // :, reduce: EqneTerm
			reduce(166), // ||, reduce: EqneTerm
			reduce(166), // ^^, reduce: EqneTerm
			reduce(166), // &&, reduce: EqneTerm
			reduce(166), // ??, reduce: EqneTerm
			reduce(166), // ???, reduce: EqneTerm
			reduce(166), // =~, reduce: EqneTerm
			reduce(166), // !=~, reduce: EqneTerm
			reduce(166), // ==, reduce: EqneTerm
			reduce(166), // !=, reduce: EqneTerm
			reduce(166), // <=>, reduce: EqneTerm
			shift(2657), // >=
			shift(2658), // <
			shift(2659), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1548
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(171), // >, reduce: CmpTerm
			nil,         // >>
			shift(2660), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(171), // ], reduce: CmpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(171), // ?, reduce: CmpTerm
			reduce(171), // :, reduce: CmpTerm
			reduce(171), // ||, reduce: CmpTerm
			reduce(171), // ^^, reduce: CmpTerm
			reduce(171), // &&, reduce: CmpTerm
			reduce(171), // ??, reduce: CmpTerm
			reduce(171), // ???, reduce: CmpTerm
			reduce(171), // =~, reduce: CmpTerm
			reduce(171), // !=~, reduce: CmpTerm
			reduce(171), // ==, reduce: CmpTerm
			reduce(171), // !=, reduce: CmpTerm
			reduce(171), // <=>, reduce: CmpTerm
			reduce(171), // >=, reduce: CmpTerm
			reduce(171), // <, reduce: CmpTerm
			reduce(171), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1549
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(173), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(173), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(173), // ], reduce: BitwiseORTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(173), // ?, reduce: BitwiseORTerm
			reduce(173), // :, reduce: BitwiseORTerm
			reduce(173), // ||, reduce: BitwiseORTerm
			reduce(173), // ^^, reduce: BitwiseORTerm
			reduce(173), // &&, reduce: BitwiseORTerm
			reduce(173), // ??, reduce: BitwiseORTerm
			reduce(173), // ???, reduce: BitwiseORTerm
			reduce(173), // =~, reduce: BitwiseORTerm
			reduce(173), // !=~, reduce: BitwiseORTerm
			reduce(173), // ==, reduce: BitwiseORTerm
			reduce(173), // !=, reduce: BitwiseORTerm
			reduce(173), // <=>, reduce: BitwiseORTerm
			reduce(173), // >=, reduce: BitwiseORTerm
			reduce(173), // <, reduce: BitwiseORTerm
			reduce(173), // <=, reduce: BitwiseORTerm
			shift(2661), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1550
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(175), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(175), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(175), // ], reduce: BitwiseXORTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(175), // ?, reduce: BitwiseXORTerm
			reduce(175), // :, reduce: BitwiseXORTerm
			reduce(175), // ||, reduce: BitwiseXORTerm
			reduce(175), // ^^, reduce: BitwiseXORTerm
			reduce(175), // &&, reduce: BitwiseXORTerm
			reduce(175), // ??, reduce: BitwiseXORTerm
			reduce(175), // ???, reduce: BitwiseXORTerm
			reduce(175), // =~, reduce: BitwiseXORTerm
			reduce(175), // !=~, reduce: BitwiseXORTerm
			reduce(175), // ==, reduce: BitwiseXORTerm
			reduce(175), // !=, reduce: BitwiseXORTerm
			reduce(175), // <=>, reduce: BitwiseXORTerm
			reduce(175), // >=, reduce: BitwiseXORTerm
			reduce(175), // <, reduce: BitwiseXORTerm
			reduce(175), // <=, reduce: BitwiseXORTerm
			reduce(175), // ^, reduce: BitwiseXORTerm
			shift(2662), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1551
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(177), // >, reduce: BitwiseANDTerm
			shift(2663), // >>
			reduce(177), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(177), // ], reduce: BitwiseANDTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(177), // ?, reduce: BitwiseANDTerm
			reduce(177), // :, reduce: BitwiseANDTerm
			reduce(177), // ||, reduce: BitwiseANDTerm
			reduce(177), // ^^, reduce: BitwiseANDTerm
			reduce(177), // &&, reduce: BitwiseANDTerm
			reduce(177), // ??, reduce: BitwiseANDTerm
			reduce(177), // ???, reduce: BitwiseANDTerm
			reduce(177), // =~, reduce: BitwiseANDTerm
			reduce(177), // !=~, reduce: BitwiseANDTerm
			reduce(177), // ==, reduce: BitwiseANDTerm
			reduce(177), // !=, reduce: BitwiseANDTerm
			reduce(177), // <=>, reduce: BitwiseANDTerm
			reduce(177), // >=, reduce: BitwiseANDTerm
			reduce(177), // <, reduce: BitwiseANDTerm
			reduce(177), // <=, reduce: BitwiseANDTerm
			reduce(177), // ^, reduce: BitwiseANDTerm
			reduce(177), // &, reduce: BitwiseANDTerm
			shift(2664), // <<
			shift(2665), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1552
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(181), // >, reduce: BitwiseShiftTerm
			reduce(181), // >>, reduce: BitwiseShiftTerm
			reduce(181), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(181), // ], reduce: BitwiseShiftTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(181), // ?, reduce: BitwiseShiftTerm
			reduce(181), // :, reduce: BitwiseShiftTerm
			reduce(181), // ||, reduce: BitwiseShiftTerm
			reduce(181), // ^^, reduce: BitwiseShiftTerm
			reduce(181), // &&, reduce: BitwiseShiftTerm
			reduce(181), // ??, reduce: BitwiseShiftTerm
			reduce(181), // ???, reduce: BitwiseShiftTerm
			reduce(181), // =~, reduce: BitwiseShiftTerm
			reduce(181), // !=~, reduce: BitwiseShiftTerm
			reduce(181), // ==, reduce: BitwiseShiftTerm
			reduce(181), // !=, reduce: BitwiseShiftTerm
			reduce(181), // <=>, reduce: BitwiseShiftTerm
			reduce(181), // >=, reduce: BitwiseShiftTerm
			reduce(181), // <, reduce: BitwiseShiftTerm
			reduce(181), // <=, reduce: BitwiseShiftTerm
			reduce(181), // ^, reduce: BitwiseShiftTerm
			reduce(181), // &, reduce: BitwiseShiftTerm
			reduce(181), // <<, reduce: BitwiseShiftTerm
			reduce(181), // >>>, reduce: BitwiseShiftTerm
			shift(2666), // +
			shift(2667), // -
			shift(2668), // .+
			shift(2669), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1553
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1554
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(186), // >, reduce: AddsubdotTerm
			reduce(186), // >>, reduce: AddsubdotTerm
			reduce(186), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(186), // ], reduce: AddsubdotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(186), // ?, reduce: AddsubdotTerm
			reduce(186), // :, reduce: AddsubdotTerm
			reduce(186), // ||, reduce: AddsubdotTerm
			reduce(186), // ^^, reduce: AddsubdotTerm
			reduce(186), // &&, reduce: AddsubdotTerm
			reduce(186), // ??, reduce: AddsubdotTerm
			reduce(186), // ???, reduce: AddsubdotTerm
			reduce(186), // =~, reduce: AddsubdotTerm
			reduce(186), // !=~, reduce: AddsubdotTerm
			reduce(186), // ==, reduce: AddsubdotTerm
			reduce(186), // !=, reduce: AddsubdotTerm
			reduce(186), // <=>, reduce: AddsubdotTerm
			reduce(186), // >=, reduce: AddsubdotTerm
			reduce(186), // <, reduce: AddsubdotTerm
			reduce(186), // <=, reduce: AddsubdotTerm
			reduce(186), // ^, reduce: AddsubdotTerm
			reduce(186), // &, reduce: AddsubdotTerm
			reduce(186), // <<, reduce: AddsubdotTerm
			reduce(186), // >>>, reduce: AddsubdotTerm
			reduce(186), // +, reduce: AddsubdotTerm
			reduce(186), // -, reduce: AddsubdotTerm
			reduce(186), // .+, reduce: AddsubdotTerm
			reduce(186), // .-, reduce: AddsubdotTerm
			shift(2671), // *
			shift(2672), // /
			shift(2673), // //
			shift(2674), // %
			shift(2675), // .*
			shift(2676), // ./
			shift(2677), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1555
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1556
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1557
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1558
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(194), // >, reduce: MuldivTerm
			reduce(194), // >>, reduce: MuldivTerm
			reduce(194), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(194), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(194), // ?, reduce: MuldivTerm
			reduce(194), // :, reduce: MuldivTerm
			reduce(194), // ||, reduce: MuldivTerm
			reduce(194), // ^^, reduce: MuldivTerm
			reduce(194), // &&, reduce: MuldivTerm
			reduce(194), // ??, reduce: MuldivTerm
			reduce(194), // ???, reduce: MuldivTerm
			reduce(194), // =~, reduce: MuldivTerm
			reduce(194), // !=~, reduce: MuldivTerm
			reduce(194), // ==, reduce: MuldivTerm
			reduce(194), // !=, reduce: MuldivTerm
			reduce(194), // <=>, reduce: MuldivTerm
			reduce(194), // >=, reduce: MuldivTerm
			reduce(194), // <, reduce: MuldivTerm
			reduce(194), // <=, reduce: MuldivTerm
			reduce(194), // ^, reduce: MuldivTerm
			reduce(194), // &, reduce: MuldivTerm
			reduce(194), // <<, reduce: MuldivTerm
			reduce(194), // >>>, reduce: MuldivTerm
			reduce(194), // +, reduce: MuldivTerm
			reduce(194), // -, reduce: MuldivTerm
			reduce(194), // .+, reduce: MuldivTerm
			reduce(194), // .-, reduce: MuldivTerm
			reduce(194), // *, reduce: MuldivTerm
			reduce(194), // /, reduce: MuldivTerm
			reduce(194), // //, reduce: MuldivTerm
			reduce(194), // %, reduce: MuldivTerm
			reduce(194), // .*, reduce: MuldivTerm
			reduce(194), // ./, reduce: MuldivTerm
			reduce(194), // .//, reduce: MuldivTerm
			shift(2681), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1559
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(196), // >, reduce: DotTerm
			reduce(196), // >>, reduce: DotTerm
			reduce(196), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(196), // ], reduce: DotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(196), // ?, reduce: DotTerm
			reduce(196), // :, reduce: DotTerm
			reduce(196), // ||, reduce: DotTerm
			reduce(196), // ^^, reduce: DotTerm
			reduce(196), // &&, reduce: DotTerm
			reduce(196), // ??, reduce: DotTerm
			reduce(196), // ???, reduce: DotTerm
			reduce(196), // =~, reduce: DotTerm
			reduce(196), // !=~, reduce: DotTerm
			reduce(196), // ==, reduce: DotTerm
			reduce(196), // !=, reduce: DotTerm
			reduce(196), // <=>, reduce: DotTerm
			reduce(196), // >=, reduce: DotTerm
			reduce(196), // <, reduce: DotTerm
			reduce(196), // <=, reduce: DotTerm
			reduce(196), // ^, reduce: DotTerm
			reduce(196), // &, reduce: DotTerm
			reduce(196), // <<, reduce: DotTerm
			reduce(196), // >>>, reduce: DotTerm
			reduce(196), // +, reduce: DotTerm
			reduce(196), // -, reduce: DotTerm
			reduce(196), // .+, reduce: DotTerm
			reduce(196), // .-, reduce: DotTerm
			reduce(196), // *, reduce: DotTerm
			reduce(196), // /, reduce: DotTerm
			reduce(196), // //, reduce: DotTerm
			reduce(196), // %, reduce: DotTerm
			reduce(196), // .*, reduce: DotTerm
			reduce(196), // ./, reduce: DotTerm
			reduce(196), // .//, reduce: DotTerm
			reduce(196), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1560
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1561
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1562
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(203), // >, reduce: UnaryOpTerm
			reduce(203), // >>, reduce: UnaryOpTerm
			reduce(203), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(203), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(203), // ?, reduce: UnaryOpTerm
			reduce(203), // :, reduce: UnaryOpTerm
			reduce(203), // ||, reduce: UnaryOpTerm
			reduce(203), // ^^, reduce: UnaryOpTerm
			reduce(203), // &&, reduce: UnaryOpTerm
			reduce(203), // ??, reduce: UnaryOpTerm
			reduce(203), // ???, reduce: UnaryOpTerm
			reduce(203), // =~, reduce: UnaryOpTerm
			reduce(203), // !=~, reduce: UnaryOpTerm
			reduce(203), // ==, reduce: UnaryOpTerm
			reduce(203), // !=, reduce: UnaryOpTerm
			reduce(203), // <=>, reduce: UnaryOpTerm
			reduce(203), // >=, reduce: UnaryOpTerm
			reduce(203), // <, reduce: UnaryOpTerm
			reduce(203), // <=, reduce: UnaryOpTerm
			reduce(203), // ^, reduce: UnaryOpTerm
			reduce(203), // &, reduce: UnaryOpTerm
			reduce(203), // <<, reduce: UnaryOpTerm
			reduce(203), // >>>, reduce: UnaryOpTerm
			reduce(203), // +, reduce: UnaryOpTerm
			reduce(203), // -, reduce: UnaryOpTerm
			reduce(203), // .+, reduce: UnaryOpTerm
			reduce(203), // .-, reduce: UnaryOpTerm
			reduce(203), // *, reduce: UnaryOpTerm
			reduce(203), // /, reduce: UnaryOpTerm
			reduce(203), // //, reduce: UnaryOpTerm
			reduce(203), // %, reduce: UnaryOpTerm
			reduce(203), // .*, reduce: UnaryOpTerm
			reduce(203), // ./, reduce: UnaryOpTerm
			reduce(203), // .//, reduce: UnaryOpTerm
			reduce(203), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1563
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(207), // >, reduce: PowTerm
			reduce(207), // >>, reduce: PowTerm
			reduce(207), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(207), // ], reduce: PowTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(207), // ?, reduce: PowTerm
			reduce(207), // :, reduce: PowTerm
			reduce(207), // ||, reduce: PowTerm
			reduce(207), // ^^, reduce: PowTerm
			reduce(207), // &&, reduce: PowTerm
			reduce(207), // ??, reduce: PowTerm
			reduce(207), // ???, reduce: PowTerm
			reduce(207), // =~, reduce: PowTerm
			reduce(207), // !=~, reduce: PowTerm
			reduce(207), // ==, reduce: PowTerm
			reduce(207), // !=, reduce: PowTerm
			reduce(207), // <=>, reduce: PowTerm
			reduce(207), // >=, reduce: PowTerm
			reduce(207), // <, reduce: PowTerm
			reduce(207), // <=, reduce: PowTerm
			reduce(207), // ^, reduce: PowTerm
			reduce(207), // &, reduce: PowTerm
			reduce(207), // <<, reduce: PowTerm
			reduce(207), // >>>, reduce: PowTerm
			reduce(207), // +, reduce: PowTerm
			reduce(207), // -, reduce: PowTerm
			reduce(207), // .+, reduce: PowTerm
			reduce(207), // .-, reduce: PowTerm
			reduce(207), // *, reduce: PowTerm
			reduce(207), // /, reduce: PowTerm
			reduce(207), // //, reduce: PowTerm
			reduce(207), // %, reduce: PowTerm
			reduce(207), // .*, reduce: PowTerm
			reduce(207), // ./, reduce: PowTerm
			reduce(207), // .//, reduce: PowTerm
			reduce(207), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			shift(2684), // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1564
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(209), // >, reduce: PrecedenceChainEnd
			reduce(209), // >>, reduce: PrecedenceChainEnd
			reduce(209), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(209), // ], reduce: PrecedenceChainEnd
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(209), // ?, reduce: PrecedenceChainEnd
			reduce(209), // :, reduce: PrecedenceChainEnd
			reduce(209), // ||, reduce: PrecedenceChainEnd
			reduce(209), // ^^, reduce: PrecedenceChainEnd
			reduce(209), // &&, reduce: PrecedenceChainEnd
			reduce(209), // ??, reduce: PrecedenceChainEnd
			reduce(209), // ???, reduce: PrecedenceChainEnd
			reduce(209), // =~, reduce: PrecedenceChainEnd
			reduce(209), // !=~, reduce: PrecedenceChainEnd
			reduce(209), // ==, reduce: PrecedenceChainEnd
			reduce(209), // !=, reduce: PrecedenceChainEnd
			reduce(209), // <=>, reduce: PrecedenceChainEnd
			reduce(209), // >=, reduce: PrecedenceChainEnd
			reduce(209), // <, reduce: PrecedenceChainEnd
			reduce(209), // <=, reduce: PrecedenceChainEnd
			reduce(209), // ^, reduce: PrecedenceChainEnd
			reduce(209), // &, reduce: PrecedenceChainEnd
			reduce(209), // <<, reduce: PrecedenceChainEnd
			reduce(209), // >>>, reduce: PrecedenceChainEnd
			reduce(209), // +, reduce: PrecedenceChainEnd
			reduce(209), // -, reduce: PrecedenceChainEnd
			reduce(209), // .+, reduce: PrecedenceChainEnd
			reduce(209), // .-, reduce: PrecedenceChainEnd
			reduce(209), // *, reduce: PrecedenceChainEnd
			reduce(209), // /, reduce: PrecedenceChainEnd
			reduce(209), // //, reduce: PrecedenceChainEnd
			reduce(209), // %, reduce: PrecedenceChainEnd
			reduce(209), // .*, reduce: PrecedenceChainEnd
			reduce(209), // ./, reduce: PrecedenceChainEnd
			reduce(209), // .//, reduce: PrecedenceChainEnd
			reduce(209), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(209), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(2685), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			shift(2686), // [[
			shift(2687), // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1565
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(215), // >, reduce: MlrvalOrFunction
			reduce(215), // >>, reduce: MlrvalOrFunction
			reduce(215), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(215), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(215), // ?, reduce: MlrvalOrFunction
			reduce(215), // :, reduce: MlrvalOrFunction
			reduce(215), // ||, reduce: MlrvalOrFunction
			reduce(215), // ^^, reduce: MlrvalOrFunction
			reduce(215), // &&, reduce: MlrvalOrFunction
			reduce(215), // ??, reduce: MlrvalOrFunction
			reduce(215), // ???, reduce: MlrvalOrFunction
			reduce(215), // =~, reduce: MlrvalOrFunction
			reduce(215), // !=~, reduce: MlrvalOrFunction
			reduce(215), // ==, reduce: MlrvalOrFunction
			reduce(215), // !=, reduce: MlrvalOrFunction
			reduce(215), // <=>, reduce: MlrvalOrFunction
			reduce(215), // >=, reduce: MlrvalOrFunction
			reduce(215), // <, reduce: MlrvalOrFunction
			reduce(215), // <=, reduce: MlrvalOrFunction
			reduce(215), // ^, reduce: MlrvalOrFunction
			reduce(215), // &, reduce: MlrvalOrFunction
			reduce(215), // <<, reduce: MlrvalOrFunction
			reduce(215), // >>>, reduce: MlrvalOrFunction
			reduce(215), // +, reduce: MlrvalOrFunction
			reduce(215), // -, reduce: MlrvalOrFunction
			reduce(215), // .+, reduce: MlrvalOrFunction
			reduce(215), // .-, reduce: MlrvalOrFunction
			reduce(215), // *, reduce: MlrvalOrFunction
			reduce(215), // /, reduce: MlrvalOrFunction
			reduce(215), // //, reduce: MlrvalOrFunction
			reduce(215), // %, reduce: MlrvalOrFunction
			reduce(215), // .*, reduce: MlrvalOrFunction
			reduce(215), // ./, reduce: MlrvalOrFunction
			reduce(215), // .//, reduce: MlrvalOrFunction
			reduce(215), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(215), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(215), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(215), // [[, reduce: MlrvalOrFunction
			reduce(215), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1566
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(216), // >, reduce: MlrvalOrFunction
			reduce(216), // >>, reduce: MlrvalOrFunction
			reduce(216), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(216), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(216), // ?, reduce: MlrvalOrFunction
			reduce(216), // :, reduce: MlrvalOrFunction
			reduce(216), // ||, reduce: MlrvalOrFunction
			reduce(216), // ^^, reduce: MlrvalOrFunction
			reduce(216), // &&, reduce: MlrvalOrFunction
			reduce(216), // ??, reduce: MlrvalOrFunction
			reduce(216), // ???, reduce: MlrvalOrFunction
			reduce(216), // =~, reduce: MlrvalOrFunction
			reduce(216), // !=~, reduce: MlrvalOrFunction
			reduce(216), // ==, reduce: MlrvalOrFunction
			reduce(216), // !=, reduce: MlrvalOrFunction
			reduce(216), // <=>, reduce: MlrvalOrFunction
			reduce(216), // >=, reduce: MlrvalOrFunction
			reduce(216), // <, reduce: MlrvalOrFunction
			reduce(216), // <=, reduce: MlrvalOrFunction
			reduce(216), // ^, reduce: MlrvalOrFunction
			reduce(216), // &, reduce: MlrvalOrFunction
			reduce(216), // <<, reduce: MlrvalOrFunction
			reduce(216), // >>>, reduce: MlrvalOrFunction
			reduce(216), // +, reduce: MlrvalOrFunction
			reduce(216), // -, reduce: MlrvalOrFunction
			reduce(216), // .+, reduce: MlrvalOrFunction
			reduce(216), // .-, reduce: MlrvalOrFunction
			reduce(216), // *, reduce: MlrvalOrFunction
			reduce(216), // /, reduce: MlrvalOrFunction
			reduce(216), // //, reduce: MlrvalOrFunction
			reduce(216), // %, reduce: MlrvalOrFunction
			reduce(216), // .*, reduce: MlrvalOrFunction
			reduce(216), // ./, reduce: MlrvalOrFunction
			reduce(216), // .//, reduce: MlrvalOrFunction
			reduce(216), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(216), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(216), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(216), // [[, reduce: MlrvalOrFunction
			reduce(216), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1567
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(217), // >, reduce: MlrvalOrFunction
			reduce(217), // >>, reduce: MlrvalOrFunction
			reduce(217), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(217), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(217), // ?, reduce: MlrvalOrFunction
			reduce(217), // :, reduce: MlrvalOrFunction
			reduce(217), // ||, reduce: MlrvalOrFunction
			reduce(217), // ^^, reduce: MlrvalOrFunction
			reduce(217), // &&, reduce: MlrvalOrFunction
			reduce(217), // ??, reduce: MlrvalOrFunction
			reduce(217), // ???, reduce: MlrvalOrFunction
			reduce(217), // =~, reduce: MlrvalOrFunction
			reduce(217), // !=~, reduce: MlrvalOrFunction
			reduce(217), // ==, reduce: MlrvalOrFunction
			reduce(217), // !=, reduce: MlrvalOrFunction
			reduce(217), // <=>, reduce: MlrvalOrFunction
			reduce(217), // >=, reduce: MlrvalOrFunction
			reduce(217), // <, reduce: MlrvalOrFunction
			reduce(217), // <=, reduce: MlrvalOrFunction
			reduce(217), // ^, reduce: MlrvalOrFunction
			reduce(217), // &, reduce: MlrvalOrFunction
			reduce(217), // <<, reduce: MlrvalOrFunction
			reduce(217), // >>>, reduce: MlrvalOrFunction
			reduce(217), // +, reduce: MlrvalOrFunction
			reduce(217), // -, reduce: MlrvalOrFunction
			reduce(217), // .+, reduce: MlrvalOrFunction
			reduce(217), // .-, reduce: MlrvalOrFunction
			reduce(217), // *, reduce: MlrvalOrFunction
			reduce(217), // /, reduce: MlrvalOrFunction
			reduce(217), // //, reduce: MlrvalOrFunction
			reduce(217), // %, reduce: MlrvalOrFunction
			reduce(217), // .*, reduce: MlrvalOrFunction
			reduce(217), // ./, reduce: MlrvalOrFunction
			reduce(217), // .//, reduce: MlrvalOrFunction
			reduce(217), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(217), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(217), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(217), // [[, reduce: MlrvalOrFunction
			reduce(217), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1568
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(218), // >, reduce: MlrvalOrFunction
			reduce(218), // >>, reduce: MlrvalOrFunction
			reduce(218), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(218), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(218), // ?, reduce: MlrvalOrFunction
			reduce(218), // :, reduce: MlrvalOrFunction
			reduce(218), // ||, reduce: MlrvalOrFunction
			reduce(218), // ^^, reduce: MlrvalOrFunction
			reduce(218), // &&, reduce: MlrvalOrFunction
			reduce(218), // ??, reduce: MlrvalOrFunction
			reduce(218), // ???, reduce: MlrvalOrFunction
			reduce(218), // =~, reduce: MlrvalOrFunction
			reduce(218), // !=~, reduce: MlrvalOrFunction
			reduce(218), // ==, reduce: MlrvalOrFunction
			reduce(218), // !=, reduce: MlrvalOrFunction
			reduce(218), // <=>, reduce: MlrvalOrFunction
			reduce(218), // >=, reduce: MlrvalOrFunction
			reduce(218), // <, reduce: MlrvalOrFunction
			reduce(218), // <=, reduce: MlrvalOrFunction
			reduce(218), // ^, reduce: MlrvalOrFunction
			reduce(218), // &, reduce: MlrvalOrFunction
			reduce(218), // <<, reduce: MlrvalOrFunction
			reduce(218), // >>>, reduce: MlrvalOrFunction
			reduce(218), // +, reduce: MlrvalOrFunction
			reduce(218), // -, reduce: MlrvalOrFunction
			reduce(218), // .+, reduce: MlrvalOrFunction
			reduce(218), // .-, reduce: MlrvalOrFunction
			reduce(218), // *, reduce: MlrvalOrFunction
			reduce(218), // /, reduce: MlrvalOrFunction
			reduce(218), // //, reduce: MlrvalOrFunction
			reduce(218), // %, reduce: MlrvalOrFunction
			reduce(218), // .*, reduce: MlrvalOrFunction
			reduce(218), // ./, reduce: MlrvalOrFunction
			reduce(218), // .//, reduce: MlrvalOrFunction
			reduce(218), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(218), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(218), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(218), // [[, reduce: MlrvalOrFunction
			reduce(218), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1569
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(219), // >, reduce: MlrvalOrFunction
			reduce(219), // >>, reduce: MlrvalOrFunction
			reduce(219), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(219), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(219), // ?, reduce: MlrvalOrFunction
			reduce(219), // :, reduce: MlrvalOrFunction
			reduce(219), // ||, reduce: MlrvalOrFunction
			reduce(219), // ^^, reduce: MlrvalOrFunction
			reduce(219), // &&, reduce: MlrvalOrFunction
			reduce(219), // ??, reduce: MlrvalOrFunction
			reduce(219), // ???, reduce: MlrvalOrFunction
			reduce(219), // =~, reduce: MlrvalOrFunction
			reduce(219), // !=~, reduce: MlrvalOrFunction
			reduce(219), // ==, reduce: MlrvalOrFunction
			reduce(219), // !=, reduce: MlrvalOrFunction
			reduce(219), // <=>, reduce: MlrvalOrFunction
			reduce(219), // >=, reduce: MlrvalOrFunction
			reduce(219), // <, reduce: MlrvalOrFunction
			reduce(219), // <=, reduce: MlrvalOrFunction
			reduce(219), // ^, reduce: MlrvalOrFunction
			reduce(219), // &, reduce: MlrvalOrFunction
			reduce(219), // <<, reduce: MlrvalOrFunction
			reduce(219), // >>>, reduce: MlrvalOrFunction
			reduce(219), // +, reduce: MlrvalOrFunction
			reduce(219), // -, reduce: MlrvalOrFunction
			reduce(219), // .+, reduce: MlrvalOrFunction
			reduce(219), // .-, reduce: MlrvalOrFunction
			reduce(219), // *, reduce: MlrvalOrFunction
			reduce(219), // /, reduce: MlrvalOrFunction
			reduce(219), // //, reduce: MlrvalOrFunction
			reduce(219), // %, reduce: MlrvalOrFunction
			reduce(219), // .*, reduce: MlrvalOrFunction
			reduce(219), // ./, reduce: MlrvalOrFunction
			reduce(219), // .//, reduce: MlrvalOrFunction
			reduce(219), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(219), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(219), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(219), // [[, reduce: MlrvalOrFunction
			reduce(219), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1570
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(220), // >, reduce: MlrvalOrFunction
			reduce(220), // >>, reduce: MlrvalOrFunction
			reduce(220), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(220), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(220), // ?, reduce: MlrvalOrFunction
			reduce(220), // :, reduce: MlrvalOrFunction
			reduce(220), // ||, reduce: MlrvalOrFunction
			reduce(220), // ^^, reduce: MlrvalOrFunction
			reduce(220), // &&, reduce: MlrvalOrFunction
			reduce(220), // ??, reduce: MlrvalOrFunction
			reduce(220), // ???, reduce: MlrvalOrFunction
			reduce(220), // =~, reduce: MlrvalOrFunction
			reduce(220), // !=~, reduce: MlrvalOrFunction
			reduce(220), // ==, reduce: MlrvalOrFunction
			reduce(220), // !=, reduce: MlrvalOrFunction
			reduce(220), // <=>, reduce: MlrvalOrFunction
			reduce(220), // >=, reduce: MlrvalOrFunction
			reduce(220), // <, reduce: MlrvalOrFunction
			reduce(220), // <=, reduce: MlrvalOrFunction
			reduce(220), // ^, reduce: MlrvalOrFunction
			reduce(220), // &, reduce: MlrvalOrFunction
			reduce(220), // <<, reduce: MlrvalOrFunction
			reduce(220), // >>>, reduce: MlrvalOrFunction
			reduce(220), // +, reduce: MlrvalOrFunction
			reduce(220), // -, reduce: MlrvalOrFunction
			reduce(220), // .+, reduce: MlrvalOrFunction
			reduce(220), // .-, reduce: MlrvalOrFunction
			reduce(220), // *, reduce: MlrvalOrFunction
			reduce(220), // /, reduce: MlrvalOrFunction
			reduce(220), // //, reduce: MlrvalOrFunction
			reduce(220), // %, reduce: MlrvalOrFunction
			reduce(220), // .*, reduce: MlrvalOrFunction
			reduce(220), // ./, reduce: MlrvalOrFunction
			reduce(220), // .//, reduce: MlrvalOrFunction
			reduce(220), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(220), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(220), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(220), // [[, reduce: MlrvalOrFunction
			reduce(220), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1571
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(221), // >, reduce: MlrvalOrFunction
			reduce(221), // >>, reduce: MlrvalOrFunction
			reduce(221), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(221), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(221), // ?, reduce: MlrvalOrFunction
			reduce(221), // :, reduce: MlrvalOrFunction
			reduce(221), // ||, reduce: MlrvalOrFunction
			reduce(221), // ^^, reduce: MlrvalOrFunction
			reduce(221), // &&, reduce: MlrvalOrFunction
			reduce(221), // ??, reduce: MlrvalOrFunction
			reduce(221), // ???, reduce: MlrvalOrFunction
			reduce(221), // =~, reduce: MlrvalOrFunction
			reduce(221), // !=~, reduce: MlrvalOrFunction
			reduce(221), // ==, reduce: MlrvalOrFunction
			reduce(221), // !=, reduce: MlrvalOrFunction
			reduce(221), // <=>, reduce: MlrvalOrFunction
			reduce(221), // >=, reduce: MlrvalOrFunction
			reduce(221), // <, reduce: MlrvalOrFunction
			reduce(221), // <=, reduce: MlrvalOrFunction
			reduce(221), // ^, reduce: MlrvalOrFunction
			reduce(221), // &, reduce: MlrvalOrFunction
			reduce(221), // <<, reduce: MlrvalOrFunction
			reduce(221), // >>>, reduce: MlrvalOrFunction
			reduce(221), // +, reduce: MlrvalOrFunction
			reduce(221), // -, reduce: MlrvalOrFunction
			reduce(221), // .+, reduce: MlrvalOrFunction
			reduce(221), // .-, reduce: MlrvalOrFunction
			reduce(221), // *, reduce: MlrvalOrFunction
			reduce(221), // /, reduce: MlrvalOrFunction
			reduce(221), // //, reduce: MlrvalOrFunction
			reduce(221), // %, reduce: MlrvalOrFunction
			reduce(221), // .*, reduce: MlrvalOrFunction
			reduce(221), // ./, reduce: MlrvalOrFunction
			reduce(221), // .//, reduce: MlrvalOrFunction
			reduce(221), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(221), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(221), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(221), // [[, reduce: MlrvalOrFunction
			reduce(221), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1572
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(222), // >, reduce: MlrvalOrFunction
			reduce(222), // >>, reduce: MlrvalOrFunction
			reduce(222), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(222), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(222), // ?, reduce: MlrvalOrFunction
			reduce(222), // :, reduce: MlrvalOrFunction
			reduce(222), // ||, reduce: MlrvalOrFunction
			reduce(222), // ^^, reduce: MlrvalOrFunction
			reduce(222), // &&, reduce: MlrvalOrFunction
			reduce(222), // ??, reduce: MlrvalOrFunction
			reduce(222), // ???, reduce: MlrvalOrFunction
			reduce(222), // =~, reduce: MlrvalOrFunction
			reduce(222), // !=~, reduce: MlrvalOrFunction
			reduce(222), // ==, reduce: MlrvalOrFunction
			reduce(222), // !=, reduce: MlrvalOrFunction
			reduce(222), // <=>, reduce: MlrvalOrFunction
			reduce(222), // >=, reduce: MlrvalOrFunction
			reduce(222), // <, reduce: MlrvalOrFunction
			reduce(222), // <=, reduce: MlrvalOrFunction
			reduce(222), // ^, reduce: MlrvalOrFunction
			reduce(222), // &, reduce: MlrvalOrFunction
			reduce(222), // <<, reduce: MlrvalOrFunction
			reduce(222), // >>>, reduce: MlrvalOrFunction
			reduce(222), // +, reduce: MlrvalOrFunction
			reduce(222), // -, reduce: MlrvalOrFunction
			reduce(222), // .+, reduce: MlrvalOrFunction
			reduce(222), // .-, reduce: MlrvalOrFunction
			reduce(222), // *, reduce: MlrvalOrFunction
			reduce(222), // /, reduce: MlrvalOrFunction
			reduce(222), // //, reduce: MlrvalOrFunction
			reduce(222), // %, reduce: MlrvalOrFunction
			reduce(222), // .*, reduce: MlrvalOrFunction
			reduce(222), // ./, reduce: MlrvalOrFunction
			reduce(222), // .//, reduce: MlrvalOrFunction
			reduce(222), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(222), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(222), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(222), // [[, reduce: MlrvalOrFunction
			reduce(222), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1573
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(223), // >, reduce: MlrvalOrFunction
			reduce(223), // >>, reduce: MlrvalOrFunction
			reduce(223), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(223), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(223), // ?, reduce: MlrvalOrFunction
			reduce(223), // :, reduce: MlrvalOrFunction
			reduce(223), // ||, reduce: MlrvalOrFunction
			reduce(223), // ^^, reduce: MlrvalOrFunction
			reduce(223), // &&, reduce: MlrvalOrFunction
			reduce(223), // ??, reduce: MlrvalOrFunction
			reduce(223), // ???, reduce: MlrvalOrFunction
			reduce(223), // =~, reduce: MlrvalOrFunction
			reduce(223), // !=~, reduce: MlrvalOrFunction
			reduce(223), // ==, reduce: MlrvalOrFunction
			reduce(223), // !=, reduce: MlrvalOrFunction
			reduce(223), // <=>, reduce: MlrvalOrFunction
			reduce(223), // >=, reduce: MlrvalOrFunction
			reduce(223), // <, reduce: MlrvalOrFunction
			reduce(223), // <=, reduce: MlrvalOrFunction
			reduce(223), // ^, reduce: MlrvalOrFunction
			reduce(223), // &, reduce: MlrvalOrFunction
			reduce(223), // <<, reduce: MlrvalOrFunction
			reduce(223), // >>>, reduce: MlrvalOrFunction
			reduce(223), // +, reduce: MlrvalOrFunction
			reduce(223), // -, reduce: MlrvalOrFunction
			reduce(223), // .+, reduce: MlrvalOrFunction
			reduce(223), // .-, reduce: MlrvalOrFunction
			reduce(223), // *, reduce: MlrvalOrFunction
			reduce(223), // /, reduce: MlrvalOrFunction
			reduce(223), // //, reduce: MlrvalOrFunction
			reduce(223), // %, reduce: MlrvalOrFunction
			reduce(223), // .*, reduce: MlrvalOrFunction
			reduce(223), // ./, reduce: MlrvalOrFunction
			reduce(223), // .//, reduce: MlrvalOrFunction
			reduce(223), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(223), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(223), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(223), // [[, reduce: MlrvalOrFunction
			reduce(223), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1574
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(224), // >, reduce: MlrvalOrFunction
			reduce(224), // >>, reduce: MlrvalOrFunction
			reduce(224), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(224), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(224), // ?, reduce: MlrvalOrFunction
			reduce(224), // :, reduce: MlrvalOrFunction
			reduce(224), // ||, reduce: MlrvalOrFunction
			reduce(224), // ^^, reduce: MlrvalOrFunction
			reduce(224), // &&, reduce: MlrvalOrFunction
			reduce(224), // ??, reduce: MlrvalOrFunction
			reduce(224), // ???, reduce: MlrvalOrFunction
			reduce(224), // =~, reduce: MlrvalOrFunction
			reduce(224), // !=~, reduce: MlrvalOrFunction
			reduce(224), // ==, reduce: MlrvalOrFunction
			reduce(224), // !=, reduce: MlrvalOrFunction
			reduce(224), // <=>, reduce: MlrvalOrFunction
			reduce(224), // >=, reduce: MlrvalOrFunction
			reduce(224), // <, reduce: MlrvalOrFunction
			reduce(224), // <=, reduce: MlrvalOrFunction
			reduce(224), // ^, reduce: MlrvalOrFunction
			reduce(224), // &, reduce: MlrvalOrFunction
			reduce(224), // <<, reduce: MlrvalOrFunction
			reduce(224), // >>>, reduce: MlrvalOrFunction
			reduce(224), // +, reduce: MlrvalOrFunction
			reduce(224), // -, reduce: MlrvalOrFunction
			reduce(224), // .+, reduce: MlrvalOrFunction
			reduce(224), // .-, reduce: MlrvalOrFunction
			reduce(224), // *, reduce: MlrvalOrFunction
			reduce(224), // /, reduce: MlrvalOrFunction
			reduce(224), // //, reduce: MlrvalOrFunction
			reduce(224), // %, reduce: MlrvalOrFunction
			reduce(224), // .*, reduce: MlrvalOrFunction
			reduce(224), // ./, reduce: MlrvalOrFunction
			reduce(224), // .//, reduce: MlrvalOrFunction
			reduce(224), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(224), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(224), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(224), // [[, reduce: MlrvalOrFunction
			reduce(224), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1575
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(225), // >, reduce: MlrvalOrFunction
			reduce(225), // >>, reduce: MlrvalOrFunction
			reduce(225), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(225), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(225), // ?, reduce: MlrvalOrFunction
			reduce(225), // :, reduce: MlrvalOrFunction
			reduce(225), // ||, reduce: MlrvalOrFunction
			reduce(225), // ^^, reduce: MlrvalOrFunction
			reduce(225), // &&, reduce: MlrvalOrFunction
			reduce(225), // ??, reduce: MlrvalOrFunction
			reduce(225), // ???, reduce: MlrvalOrFunction
			reduce(225), // =~, reduce: MlrvalOrFunction
			reduce(225), // !=~, reduce: MlrvalOrFunction
			reduce(225), // ==, reduce: MlrvalOrFunction
			reduce(225), // !=, reduce: MlrvalOrFunction
			reduce(225), // <=>, reduce: MlrvalOrFunction
			reduce(225), // >=, reduce: MlrvalOrFunction
			reduce(225), // <, reduce: MlrvalOrFunction
			reduce(225), // <=, reduce: MlrvalOrFunction
			reduce(225), // ^, reduce: MlrvalOrFunction
			reduce(225), // &, reduce: MlrvalOrFunction
			reduce(225), // <<, reduce: MlrvalOrFunction
			reduce(225), // >>>, reduce: MlrvalOrFunction
			reduce(225), // +, reduce: MlrvalOrFunction
			reduce(225), // -, reduce: MlrvalOrFunction
			reduce(225), // .+, reduce: MlrvalOrFunction
			reduce(225), // .-, reduce: MlrvalOrFunction
			reduce(225), // *, reduce: MlrvalOrFunction
			reduce(225), // /, reduce: MlrvalOrFunction
			reduce(225), // //, reduce: MlrvalOrFunction
			reduce(225), // %, reduce: MlrvalOrFunction
			reduce(225), // .*, reduce: MlrvalOrFunction
			reduce(225), // ./, reduce: MlrvalOrFunction
			reduce(225), // .//, reduce: MlrvalOrFunction
			reduce(225), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(225), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(225), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(225), // [[, reduce: MlrvalOrFunction
			reduce(225), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1576
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(226), // >, reduce: MlrvalOrFunction
			reduce(226), // >>, reduce: MlrvalOrFunction
			reduce(226), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(226), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(226), // ?, reduce: MlrvalOrFunction
			reduce(226), // :, reduce: MlrvalOrFunction
			reduce(226), // ||, reduce: MlrvalOrFunction
			reduce(226), // ^^, reduce: MlrvalOrFunction
			reduce(226), // &&, reduce: MlrvalOrFunction
			reduce(226), // ??, reduce: MlrvalOrFunction
			reduce(226), // ???, reduce: MlrvalOrFunction
			reduce(226), // =~, reduce: MlrvalOrFunction
			reduce(226), // !=~, reduce: MlrvalOrFunction
			reduce(226), // ==, reduce: MlrvalOrFunction
			reduce(226), // !=, reduce: MlrvalOrFunction
			reduce(226), // <=>, reduce: MlrvalOrFunction
			reduce(226), // >=, reduce: MlrvalOrFunction
			reduce(226), // <, reduce: MlrvalOrFunction
			reduce(226), // <=, reduce: MlrvalOrFunction
			reduce(226), // ^, reduce: MlrvalOrFunction
			reduce(226), // &, reduce: MlrvalOrFunction
			reduce(226), // <<, reduce: MlrvalOrFunction
			reduce(226), // >>>, reduce: MlrvalOrFunction
			reduce(226), // +, reduce: MlrvalOrFunction
			reduce(226), // -, reduce: MlrvalOrFunction
			reduce(226), // .+, reduce: MlrvalOrFunction
			reduce(226), // .-, reduce: MlrvalOrFunction
			reduce(226), // *, reduce: MlrvalOrFunction
			reduce(226), // /, reduce: MlrvalOrFunction
			reduce(226), // //, reduce: MlrvalOrFunction
			reduce(226), // %, reduce: MlrvalOrFunction
			reduce(226), // .*, reduce: MlrvalOrFunction
			reduce(226), // ./, reduce: MlrvalOrFunction
			reduce(226), // .//, reduce: MlrvalOrFunction
			reduce(226), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(226), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(226), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(226), // [[, reduce: MlrvalOrFunction
			reduce(226), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1577
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(227), // >, reduce: MlrvalOrFunction
			reduce(227), // >>, reduce: MlrvalOrFunction
			reduce(227), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(227), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(227), // ?, reduce: MlrvalOrFunction
			reduce(227), // :, reduce: MlrvalOrFunction
			reduce(227), // ||, reduce: MlrvalOrFunction
			reduce(227), // ^^, reduce: MlrvalOrFunction
			reduce(227), // &&, reduce: MlrvalOrFunction
			reduce(227), // ??, reduce: MlrvalOrFunction
			reduce(227), // ???, reduce: MlrvalOrFunction
			reduce(227), // =~, reduce: MlrvalOrFunction
			reduce(227), // !=~, reduce: MlrvalOrFunction
			reduce(227), // ==, reduce: MlrvalOrFunction
			reduce(227), // !=, reduce: MlrvalOrFunction
			reduce(227), // <=>, reduce: MlrvalOrFunction
			reduce(227), // >=, reduce: MlrvalOrFunction
			reduce(227), // <, reduce: MlrvalOrFunction
			reduce(227), // <=, reduce: MlrvalOrFunction
			reduce(227), // ^, reduce: MlrvalOrFunction
			reduce(227), // &, reduce: MlrvalOrFunction
			reduce(227), // <<, reduce: MlrvalOrFunction
			reduce(227), // >>>, reduce: MlrvalOrFunction
			reduce(227), // +, reduce: MlrvalOrFunction
			reduce(227), // -, reduce: MlrvalOrFunction
			reduce(227), // .+, reduce: MlrvalOrFunction
			reduce(227), // .-, reduce: MlrvalOrFunction
			reduce(227), // *, reduce: MlrvalOrFunction
			reduce(227), // /, reduce: MlrvalOrFunction
			reduce(227), // //, reduce: MlrvalOrFunction
			reduce(227), // %, reduce: MlrvalOrFunction
			reduce(227), // .*, reduce: MlrvalOrFunction
			reduce(227), // ./, reduce: MlrvalOrFunction
			reduce(227), // .//, reduce: MlrvalOrFunction
			reduce(227), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(227), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(227), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(227), // [[, reduce: MlrvalOrFunction
			reduce(227), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1578
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(752),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(756),  // (
			nil,         // )
			nil,         // emitp
			shift(768),  // field_name
			shift(769),  // $[
			shift(2688), // ]
			shift(771),  // braced_field_name
			shift(772),  // $[[
			shift(773),  // $[[[
			shift(774),  // full_srec
			shift(776),  // oosvar_name
			shift(777),  // @[
			shift(778),  // braced_oosvar_name
			shift(779),  // full_oosvar
			shift(780),  // all
			shift(781),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(796),  // +
			shift(798),  // -
			shift(799),  // .+
			shift(800),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(803),  // !
			shift(804),  // ~
			nil,         // **
			shift(809),  // string_literal
			shift(810),  // regex_case_insensitive
			shift(811),  // int_literal
			shift(812),  // float_literal
			shift(813),  // boolean_literal
			shift(814),  // null_literal
			shift(815),  // inf_literal
			shift(816),  // nan_literal
			shift(817),  // const_M_PI
			shift(818),  // const_M_E
			shift(819),  // panic
			shift(821),  // [
			shift(824),  // ctx_IPS
			shift(825),  // ctx_IFS
			shift(826),  // ctx_IRS
			shift(827),  // ctx_OPS
			shift(828),  // ctx_OFS
			shift(829),  // ctx_ORS
			shift(830),  // ctx_FLATSEP
			shift(831),  // ctx_NF
			shift(832),  // ctx_NR
			shift(833),  // ctx_FNR
			shift(834),  // ctx_FILENAME
			shift(835),  // ctx_FILENUM
			shift(837),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(844),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1579
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(240), // >, reduce: MlrvalOrFunction
			reduce(240), // >>, reduce: MlrvalOrFunction
			reduce(240), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(240), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(240), // ?, reduce: MlrvalOrFunction
			reduce(240), // :, reduce: MlrvalOrFunction
			reduce(240), // ||, reduce: MlrvalOrFunction
			reduce(240), // ^^, reduce: MlrvalOrFunction
			reduce(240), // &&, reduce: MlrvalOrFunction
			reduce(240), // ??, reduce: MlrvalOrFunction
			reduce(240), // ???, reduce: MlrvalOrFunction
			reduce(240), // =~, reduce: MlrvalOrFunction
			reduce(240), // !=~, reduce: MlrvalOrFunction
			reduce(240), // ==, reduce: MlrvalOrFunction
			reduce(240), // !=, reduce: MlrvalOrFunction
			reduce(240), // <=>, reduce: MlrvalOrFunction
			reduce(240), // >=, reduce: MlrvalOrFunction
			reduce(240), // <, reduce: MlrvalOrFunction
			reduce(240), // <=, reduce: MlrvalOrFunction
			reduce(240), // ^, reduce: MlrvalOrFunction
			reduce(240), // &, reduce: MlrvalOrFunction
			reduce(240), // <<, reduce: MlrvalOrFunction
			reduce(240), // >>>, reduce: MlrvalOrFunction
			reduce(240), // +, reduce: MlrvalOrFunction
			reduce(240), // -, reduce: MlrvalOrFunction
			reduce(240), // .+, reduce: MlrvalOrFunction
			reduce(240), // .-, reduce: MlrvalOrFunction
			reduce(240), // *, reduce: MlrvalOrFunction
			reduce(240), // /, reduce: MlrvalOrFunction
			reduce(240), // //, reduce: MlrvalOrFunction
			reduce(240), // %, reduce: MlrvalOrFunction
			reduce(240), // .*, reduce: MlrvalOrFunction
			reduce(240), // ./, reduce: MlrvalOrFunction
			reduce(240), // .//, reduce: MlrvalOrFunction
			reduce(240), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(240), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(240), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(240), // [[, reduce: MlrvalOrFunction
			reduce(240), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1580
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(241), // >, reduce: ContextVariable
			reduce(241), // >>, reduce: ContextVariable
			reduce(241), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(241), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(241), // ?, reduce: ContextVariable
			reduce(241), // :, reduce: ContextVariable
			reduce(241), // ||, reduce: ContextVariable
			reduce(241), // ^^, reduce: ContextVariable
			reduce(241), // &&, reduce: ContextVariable
			reduce(241), // ??, reduce: ContextVariable
			reduce(241), // ???, reduce: ContextVariable
			reduce(241), // =~, reduce: ContextVariable
			reduce(241), // !=~, reduce: ContextVariable
			reduce(241), // ==, reduce: ContextVariable
			reduce(241), // !=, reduce: ContextVariable
			reduce(241), // <=>, reduce: ContextVariable
			reduce(241), // >=, reduce: ContextVariable
			reduce(241), // <, reduce: ContextVariable
			reduce(241), // <=, reduce: ContextVariable
			reduce(241), // ^, reduce: ContextVariable
			reduce(241), // &, reduce: ContextVariable
			reduce(241), // <<, reduce: ContextVariable
			reduce(241), // >>>, reduce: ContextVariable
			reduce(241), // +, reduce: ContextVariable
			reduce(241), // -, reduce: ContextVariable
			reduce(241), // .+, reduce: ContextVariable
			reduce(241), // .-, reduce: ContextVariable
			reduce(241), // *, reduce: ContextVariable
			reduce(241), // /, reduce: ContextVariable
			reduce(241), // //, reduce: ContextVariable
			reduce(241), // %, reduce: ContextVariable
			reduce(241), // .*, reduce: ContextVariable
			reduce(241), // ./, reduce: ContextVariable
			reduce(241), // .//, reduce: ContextVariable
			reduce(241), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(241), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(241), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(241), // [[, reduce: ContextVariable
			reduce(241), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1581
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(242), // >, reduce: ContextVariable
			reduce(242), // >>, reduce: ContextVariable
			reduce(242), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(242), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(242), // ?, reduce: ContextVariable
			reduce(242), // :, reduce: ContextVariable
			reduce(242), // ||, reduce: ContextVariable
			reduce(242), // ^^, reduce: ContextVariable
			reduce(242), // &&, reduce: ContextVariable
			reduce(242), // ??, reduce: ContextVariable
			reduce(242), // ???, reduce: ContextVariable
			reduce(242), // =~, reduce: ContextVariable
			reduce(242), // !=~, reduce: ContextVariable
			reduce(242), // ==, reduce: ContextVariable
			reduce(242), // !=, reduce: ContextVariable
			reduce(242), // <=>, reduce: ContextVariable
			reduce(242), // >=, reduce: ContextVariable
			reduce(242), // <, reduce: ContextVariable
			reduce(242), // <=, reduce: ContextVariable
			reduce(242), // ^, reduce: ContextVariable
			reduce(242), // &, reduce: ContextVariable
			reduce(242), // <<, reduce: ContextVariable
			reduce(242), // >>>, reduce: ContextVariable
			reduce(242), // +, reduce: ContextVariable
			reduce(242), // -, reduce: ContextVariable
			reduce(242), // .+, reduce: ContextVariable
			reduce(242), // .-, reduce: ContextVariable
			reduce(242), // *, reduce: ContextVariable
			reduce(242), // /, reduce: ContextVariable
			reduce(242), // //, reduce: ContextVariable
			reduce(242), // %, reduce: ContextVariable
			reduce(242), // .*, reduce: ContextVariable
			reduce(242), // ./, reduce: ContextVariable
			reduce(242), // .//, reduce: ContextVariable
			reduce(242), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(242), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(242), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(242), // [[, reduce: ContextVariable
			reduce(242), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1582
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(243), // >, reduce: ContextVariable
			reduce(243), // >>, reduce: ContextVariable
			reduce(243), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(243), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(243), // ?, reduce: ContextVariable
			reduce(243), // :, reduce: ContextVariable
			reduce(243), // ||, reduce: ContextVariable
			reduce(243), // ^^, reduce: ContextVariable
			reduce(243), // &&, reduce: ContextVariable
			reduce(243), // ??, reduce: ContextVariable
			reduce(243), // ???, reduce: ContextVariable
			reduce(243), // =~, reduce: ContextVariable
			reduce(243), // !=~, reduce: ContextVariable
			reduce(243), // ==, reduce: ContextVariable
			reduce(243), // !=, reduce: ContextVariable
			reduce(243), // <=>, reduce: ContextVariable
			reduce(243), // >=, reduce: ContextVariable
			reduce(243), // <, reduce: ContextVariable
			reduce(243), // <=, reduce: ContextVariable
			reduce(243), // ^, reduce: ContextVariable
			reduce(243), // &, reduce: ContextVariable
			reduce(243), // <<, reduce: ContextVariable
			reduce(243), // >>>, reduce: ContextVariable
			reduce(243), // +, reduce: ContextVariable
			reduce(243), // -, reduce: ContextVariable
			reduce(243), // .+, reduce: ContextVariable
			reduce(243), // .-, reduce: ContextVariable
			reduce(243), // *, reduce: ContextVariable
			reduce(243), // /, reduce: ContextVariable
			reduce(243), // //, reduce: ContextVariable
			reduce(243), // %, reduce: ContextVariable
			reduce(243), // .*, reduce: ContextVariable
			reduce(243), // ./, reduce: ContextVariable
			reduce(243), // .//, reduce: ContextVariable
			reduce(243), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(243), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(243), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(243), // [[, reduce: ContextVariable
			reduce(243), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1583
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(244), // >, reduce: ContextVariable
			reduce(244), // >>, reduce: ContextVariable
			reduce(244), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(244), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(244), // ?, reduce: ContextVariable
			reduce(244), // :, reduce: ContextVariable
			reduce(244), // ||, reduce: ContextVariable
			reduce(244), // ^^, reduce: ContextVariable
			reduce(244), // &&, reduce: ContextVariable
			reduce(244), // ??, reduce: ContextVariable
			reduce(244), // ???, reduce: ContextVariable
			reduce(244), // =~, reduce: ContextVariable
			reduce(244), // !=~, reduce: ContextVariable
			reduce(244), // ==, reduce: ContextVariable
			reduce(244), // !=, reduce: ContextVariable
			reduce(244), // <=>, reduce: ContextVariable
			reduce(244), // >=, reduce: ContextVariable
			reduce(244), // <, reduce: ContextVariable
			reduce(244), // <=, reduce: ContextVariable
			reduce(244), // ^, reduce: ContextVariable
			reduce(244), // &, reduce: ContextVariable
			reduce(244), // <<, reduce: ContextVariable
			reduce(244), // >>>, reduce: ContextVariable
			reduce(244), // +, reduce: ContextVariable
			reduce(244), // -, reduce: ContextVariable
			reduce(244), // .+, reduce: ContextVariable
			reduce(244), // .-, reduce: ContextVariable
			reduce(244), // *, reduce: ContextVariable
			reduce(244), // /, reduce: ContextVariable
			reduce(244), // //, reduce: ContextVariable
			reduce(244), // %, reduce: ContextVariable
			reduce(244), // .*, reduce: ContextVariable
			reduce(244), // ./, reduce: ContextVariable
			reduce(244), // .//, reduce: ContextVariable
			reduce(244), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(244), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(244), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(244), // [[, reduce: ContextVariable
			reduce(244), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1584
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(245), // >, reduce: ContextVariable
			reduce(245), // >>, reduce: ContextVariable
			reduce(245), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(245), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(245), // ?, reduce: ContextVariable
			reduce(245), // :, reduce: ContextVariable
			reduce(245), // ||, reduce: ContextVariable
			reduce(245), // ^^, reduce: ContextVariable
			reduce(245), // &&, reduce: ContextVariable
			reduce(245), // ??, reduce: ContextVariable
			reduce(245), // ???, reduce: ContextVariable
			reduce(245), // =~, reduce: ContextVariable
			reduce(245), // !=~, reduce: ContextVariable
			reduce(245), // ==, reduce: ContextVariable
			reduce(245), // !=, reduce: ContextVariable
			reduce(245), // <=>, reduce: ContextVariable
			reduce(245), // >=, reduce: ContextVariable
			reduce(245), // <, reduce: ContextVariable
			reduce(245), // <=, reduce: ContextVariable
			reduce(245), // ^, reduce: ContextVariable
			reduce(245), // &, reduce: ContextVariable
			reduce(245), // <<, reduce: ContextVariable
			reduce(245), // >>>, reduce: ContextVariable
			reduce(245), // +, reduce: ContextVariable
			reduce(245), // -, reduce: ContextVariable
			reduce(245), // .+, reduce: ContextVariable
			reduce(245), // .-, reduce: ContextVariable
			reduce(245), // *, reduce: ContextVariable
			reduce(245), // /, reduce: ContextVariable
			reduce(245), // //, reduce: ContextVariable
			reduce(245), // %, reduce: ContextVariable
			reduce(245), // .*, reduce: ContextVariable
			reduce(245), // ./, reduce: ContextVariable
			reduce(245), // .//, reduce: ContextVariable
			reduce(245), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(245), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(245), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(245), // [[, reduce: ContextVariable
			reduce(245), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1585
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(246), // >, reduce: ContextVariable
			reduce(246), // >>, reduce: ContextVariable
			reduce(246), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(246), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(246), // ?, reduce: ContextVariable
			reduce(246), // :, reduce: ContextVariable
			reduce(246), // ||, reduce: ContextVariable
			reduce(246), // ^^, reduce: ContextVariable
			reduce(246), // &&, reduce: ContextVariable
			reduce(246), // ??, reduce: ContextVariable
			reduce(246), // ???, reduce: ContextVariable
			reduce(246), // =~, reduce: ContextVariable
			reduce(246), // !=~, reduce: ContextVariable
			reduce(246), // ==, reduce: ContextVariable
			reduce(246), // !=, reduce: ContextVariable
			reduce(246), // <=>, reduce: ContextVariable
			reduce(246), // >=, reduce: ContextVariable
			reduce(246), // <, reduce: ContextVariable
			reduce(246), // <=, reduce: ContextVariable
			reduce(246), // ^, reduce: ContextVariable
			reduce(246), // &, reduce: ContextVariable
			reduce(246), // <<, reduce: ContextVariable
			reduce(246), // >>>, reduce: ContextVariable
			reduce(246), // +, reduce: ContextVariable
			reduce(246), // -, reduce: ContextVariable
			reduce(246), // .+, reduce: ContextVariable
			reduce(246), // .-, reduce: ContextVariable
			reduce(246), // *, reduce: ContextVariable
			reduce(246), // /, reduce: ContextVariable
			reduce(246), // //, reduce: ContextVariable
			reduce(246), // %, reduce: ContextVariable
			reduce(246), // .*, reduce: ContextVariable
			reduce(246), // ./, reduce: ContextVariable
			reduce(246), // .//, reduce: ContextVariable
			reduce(246), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(246), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(246), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(246), // [[, reduce: ContextVariable
			reduce(246), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1586
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(247), // >, reduce: ContextVariable
			reduce(247), // >>, reduce: ContextVariable
			reduce(247), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(247), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(247), // ?, reduce: ContextVariable
			reduce(247), // :, reduce: ContextVariable
			reduce(247), // ||, reduce: ContextVariable
			reduce(247), // ^^, reduce: ContextVariable
			reduce(247), // &&, reduce: ContextVariable
			reduce(247), // ??, reduce: ContextVariable
			reduce(247), // ???, reduce: ContextVariable
			reduce(247), // =~, reduce: ContextVariable
			reduce(247), // !=~, reduce: ContextVariable
			reduce(247), // ==, reduce: ContextVariable
			reduce(247), // !=, reduce: ContextVariable
			reduce(247), // <=>, reduce: ContextVariable
			reduce(247), // >=, reduce: ContextVariable
			reduce(247), // <, reduce: ContextVariable
			reduce(247), // <=, reduce: ContextVariable
			reduce(247), // ^, reduce: ContextVariable
			reduce(247), // &, reduce: ContextVariable
			reduce(247), // <<, reduce: ContextVariable
			reduce(247), // >>>, reduce: ContextVariable
			reduce(247), // +, reduce: ContextVariable
			reduce(247), // -, reduce: ContextVariable
			reduce(247), // .+, reduce: ContextVariable
			reduce(247), // .-, reduce: ContextVariable
			reduce(247), // *, reduce: ContextVariable
			reduce(247), // /, reduce: ContextVariable
			reduce(247), // //, reduce: ContextVariable
			reduce(247), // %, reduce: ContextVariable
			reduce(247), // .*, reduce: ContextVariable
			reduce(247), // ./, reduce: ContextVariable
			reduce(247), // .//, reduce: ContextVariable
			reduce(247), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(247), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(247), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(247), // [[, reduce: ContextVariable
			reduce(247), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1587
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(248), // >, reduce: ContextVariable
			reduce(248), // >>, reduce: ContextVariable
			reduce(248), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(248), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(248), // ?, reduce: ContextVariable
			reduce(248), // :, reduce: ContextVariable
			reduce(248), // ||, reduce: ContextVariable
			reduce(248), // ^^, reduce: ContextVariable
			reduce(248), // &&, reduce: ContextVariable
			reduce(248), // ??, reduce: ContextVariable
			reduce(248), // ???, reduce: ContextVariable
			reduce(248), // =~, reduce: ContextVariable
			reduce(248), // !=~, reduce: ContextVariable
			reduce(248), // ==, reduce: ContextVariable
			reduce(248), // !=, reduce: ContextVariable
			reduce(248), // <=>, reduce: ContextVariable
			reduce(248), // >=, reduce: ContextVariable
			reduce(248), // <, reduce: ContextVariable
			reduce(248), // <=, reduce: ContextVariable
			reduce(248), // ^, reduce: ContextVariable
			reduce(248), // &, reduce: ContextVariable
			reduce(248), // <<, reduce: ContextVariable
			reduce(248), // >>>, reduce: ContextVariable
			reduce(248), // +, reduce: ContextVariable
			reduce(248), // -, reduce: ContextVariable
			reduce(248), // .+, reduce: ContextVariable
			reduce(248), // .-, reduce: ContextVariable
			reduce(248), // *, reduce: ContextVariable
			reduce(248), // /, reduce: ContextVariable
			reduce(248), // //, reduce: ContextVariable
			reduce(248), // %, reduce: ContextVariable
			reduce(248), // .*, reduce: ContextVariable
			reduce(248), // ./, reduce: ContextVariable
			reduce(248), // .//, reduce: ContextVariable
			reduce(248), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(248), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(248), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(248), // [[, reduce: ContextVariable
			reduce(248), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1588
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(249), // >, reduce: ContextVariable
			reduce(249), // >>, reduce: ContextVariable
			reduce(249), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(249), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(249), // ?, reduce: ContextVariable
			reduce(249), // :, reduce: ContextVariable
			reduce(249), // ||, reduce: ContextVariable
			reduce(249), // ^^, reduce: ContextVariable
			reduce(249), // &&, reduce: ContextVariable
			reduce(249), // ??, reduce: ContextVariable
			reduce(249), // ???, reduce: ContextVariable
			reduce(249), // =~, reduce: ContextVariable
			reduce(249), // !=~, reduce: ContextVariable
			reduce(249), // ==, reduce: ContextVariable
			reduce(249), // !=, reduce: ContextVariable
			reduce(249), // <=>, reduce: ContextVariable
			reduce(249), // >=, reduce: ContextVariable
			reduce(249), // <, reduce: ContextVariable
			reduce(249), // <=, reduce: ContextVariable
			reduce(249), // ^, reduce: ContextVariable
			reduce(249), // &, reduce: ContextVariable
			reduce(249), // <<, reduce: ContextVariable
			reduce(249), // >>>, reduce: ContextVariable
			reduce(249), // +, reduce: ContextVariable
			reduce(249), // -, reduce: ContextVariable
			reduce(249), // .+, reduce: ContextVariable
			reduce(249), // .-, reduce: ContextVariable
			reduce(249), // *, reduce: ContextVariable
			reduce(249), // /, reduce: ContextVariable
			reduce(249), // //, reduce: ContextVariable
			reduce(249), // %, reduce: ContextVariable
			reduce(249), // .*, reduce: ContextVariable
			reduce(249), // ./, reduce: ContextVariable
			reduce(249), // .//, reduce: ContextVariable
			reduce(249), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(249), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(249), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(249), // [[, reduce: ContextVariable
			reduce(249), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1589
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(250), // >, reduce: ContextVariable
			reduce(250), // >>, reduce: ContextVariable
			reduce(250), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(250), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(250), // ?, reduce: ContextVariable
			reduce(250), // :, reduce: ContextVariable
			reduce(250), // ||, reduce: ContextVariable
			reduce(250), // ^^, reduce: ContextVariable
			reduce(250), // &&, reduce: ContextVariable
			reduce(250), // ??, reduce: ContextVariable
			reduce(250), // ???, reduce: ContextVariable
			reduce(250), // =~, reduce: ContextVariable
			reduce(250), // !=~, reduce: ContextVariable
			reduce(250), // ==, reduce: ContextVariable
			reduce(250), // !=, reduce: ContextVariable
			reduce(250), // <=>, reduce: ContextVariable
			reduce(250), // >=, reduce: ContextVariable
			reduce(250), // <, reduce: ContextVariable
			reduce(250), // <=, reduce: ContextVariable
			reduce(250), // ^, reduce: ContextVariable
			reduce(250), // &, reduce: ContextVariable
			reduce(250), // <<, reduce: ContextVariable
			reduce(250), // >>>, reduce: ContextVariable
			reduce(250), // +, reduce: ContextVariable
			reduce(250), // -, reduce: ContextVariable
			reduce(250), // .+, reduce: ContextVariable
			reduce(250), // .-, reduce: ContextVariable
			reduce(250), // *, reduce: ContextVariable
			reduce(250), // /, reduce: ContextVariable
			reduce(250), // //, reduce: ContextVariable
			reduce(250), // %, reduce: ContextVariable
			reduce(250), // .*, reduce: ContextVariable
			reduce(250), // ./, reduce: ContextVariable
			reduce(250), // .//, reduce: ContextVariable
			reduce(250), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(250), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(250), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(250), // [[, reduce: ContextVariable
			reduce(250), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1590
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(251), // >, reduce: ContextVariable
			reduce(251), // >>, reduce: ContextVariable
			reduce(251), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(251), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(251), // ?, reduce: ContextVariable
			reduce(251), // :, reduce: ContextVariable
			reduce(251), // ||, reduce: ContextVariable
			reduce(251), // ^^, reduce: ContextVariable
			reduce(251), // &&, reduce: ContextVariable
			reduce(251), // ??, reduce: ContextVariable
			reduce(251), // ???, reduce: ContextVariable
			reduce(251), // =~, reduce: ContextVariable
			reduce(251), // !=~, reduce: ContextVariable
			reduce(251), // ==, reduce: ContextVariable
			reduce(251), // !=, reduce: ContextVariable
			reduce(251), // <=>, reduce: ContextVariable
			reduce(251), // >=, reduce: ContextVariable
			reduce(251), // <, reduce: ContextVariable
			reduce(251), // <=, reduce: ContextVariable
			reduce(251), // ^, reduce: ContextVariable
			reduce(251), // &, reduce: ContextVariable
			reduce(251), // <<, reduce: ContextVariable
			reduce(251), // >>>, reduce: ContextVariable
			reduce(251), // +, reduce: ContextVariable
			reduce(251), // -, reduce: ContextVariable
			reduce(251), // .+, reduce: ContextVariable
			reduce(251), // .-, reduce: ContextVariable
			reduce(251), // *, reduce: ContextVariable
			reduce(251), // /, reduce: ContextVariable
			reduce(251), // //, reduce: ContextVariable
			reduce(251), // %, reduce: ContextVariable
			reduce(251), // .*, reduce: ContextVariable
			reduce(251), // ./, reduce: ContextVariable
			reduce(251), // .//, reduce: ContextVariable
			reduce(251), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(251), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(251), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(251), // [[, reduce: ContextVariable
			reduce(251), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1591
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(252), // >, reduce: ContextVariable
			reduce(252), // >>, reduce: ContextVariable
			reduce(252), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(252), // ], reduce: ContextVariable
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(252), // ?, reduce: ContextVariable
			reduce(252), // :, reduce: ContextVariable
			reduce(252), // ||, reduce: ContextVariable
			reduce(252), // ^^, reduce: ContextVariable
			reduce(252), // &&, reduce: ContextVariable
			reduce(252), // ??, reduce: ContextVariable
			reduce(252), // ???, reduce: ContextVariable
			reduce(252), // =~, reduce: ContextVariable
			reduce(252), // !=~, reduce: ContextVariable
			reduce(252), // ==, reduce: ContextVariable
			reduce(252), // !=, reduce: ContextVariable
			reduce(252), // <=>, reduce: ContextVariable
			reduce(252), // >=, reduce: ContextVariable
			reduce(252), // <, reduce: ContextVariable
			reduce(252), // <=, reduce: ContextVariable
			reduce(252), // ^, reduce: ContextVariable
			reduce(252), // &, reduce: ContextVariable
			reduce(252), // <<, reduce: ContextVariable
			reduce(252), // >>>, reduce: ContextVariable
			reduce(252), // +, reduce: ContextVariable
			reduce(252), // -, reduce: ContextVariable
			reduce(252), // .+, reduce: ContextVariable
			reduce(252), // .-, reduce: ContextVariable
			reduce(252), // *, reduce: ContextVariable
			reduce(252), // /, reduce: ContextVariable
			reduce(252), // //, reduce: ContextVariable
			reduce(252), // %, reduce: ContextVariable
			reduce(252), // .*, reduce: ContextVariable
			reduce(252), // ./, reduce: ContextVariable
			reduce(252), // .//, reduce: ContextVariable
			reduce(252), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(252), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(252), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(252), // [[, reduce: ContextVariable
			reduce(252), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1592
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(253), // >, reduce: MlrvalOrFunction
			reduce(253), // >>, reduce: MlrvalOrFunction
			reduce(253), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(253), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(253), // ?, reduce: MlrvalOrFunction
			reduce(253), // :, reduce: MlrvalOrFunction
			reduce(253), // ||, reduce: MlrvalOrFunction
			reduce(253), // ^^, reduce: MlrvalOrFunction
			reduce(253), // &&, reduce: MlrvalOrFunction
			reduce(253), // ??, reduce: MlrvalOrFunction
			reduce(253), // ???, reduce: MlrvalOrFunction
			reduce(253), // =~, reduce: MlrvalOrFunction
			reduce(253), // !=~, reduce: MlrvalOrFunction
			reduce(253), // ==, reduce: MlrvalOrFunction
			reduce(253), // !=, reduce: MlrvalOrFunction
			reduce(253), // <=>, reduce: MlrvalOrFunction
			reduce(253), // >=, reduce: MlrvalOrFunction
			reduce(253), // <, reduce: MlrvalOrFunction
			reduce(253), // <=, reduce: MlrvalOrFunction
			reduce(253), // ^, reduce: MlrvalOrFunction
			reduce(253), // &, reduce: MlrvalOrFunction
			reduce(253), // <<, reduce: MlrvalOrFunction
			reduce(253), // >>>, reduce: MlrvalOrFunction
			reduce(253), // +, reduce: MlrvalOrFunction
			reduce(253), // -, reduce: MlrvalOrFunction
			reduce(253), // .+, reduce: MlrvalOrFunction
			reduce(253), // .-, reduce: MlrvalOrFunction
			reduce(253), // *, reduce: MlrvalOrFunction
			reduce(253), // /, reduce: MlrvalOrFunction
			reduce(253), // //, reduce: MlrvalOrFunction
			reduce(253), // %, reduce: MlrvalOrFunction
			reduce(253), // .*, reduce: MlrvalOrFunction
			reduce(253), // ./, reduce: MlrvalOrFunction
			reduce(253), // .//, reduce: MlrvalOrFunction
			reduce(253), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(253), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(253), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(253), // [[, reduce: MlrvalOrFunction
			reduce(253), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1593
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			shift(2690), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(2691), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1594
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(256), // >, reduce: MlrvalOrFunction
			reduce(256), // >>, reduce: MlrvalOrFunction
			reduce(256), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(256), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(256), // ?, reduce: MlrvalOrFunction
			reduce(256), // :, reduce: MlrvalOrFunction
			reduce(256), // ||, reduce: MlrvalOrFunction
			reduce(256), // ^^, reduce: MlrvalOrFunction
			reduce(256), // &&, reduce: MlrvalOrFunction
			reduce(256), // ??, reduce: MlrvalOrFunction
			reduce(256), // ???, reduce: MlrvalOrFunction
			reduce(256), // =~, reduce: MlrvalOrFunction
			reduce(256), // !=~, reduce: MlrvalOrFunction
			reduce(256), // ==, reduce: MlrvalOrFunction
			reduce(256), // !=, reduce: MlrvalOrFunction
			reduce(256), // <=>, reduce: MlrvalOrFunction
			reduce(256), // >=, reduce: MlrvalOrFunction
			reduce(256), // <, reduce: MlrvalOrFunction
			reduce(256), // <=, reduce: MlrvalOrFunction
			reduce(256), // ^, reduce: MlrvalOrFunction
			reduce(256), // &, reduce: MlrvalOrFunction
			reduce(256), // <<, reduce: MlrvalOrFunction
			reduce(256), // >>>, reduce: MlrvalOrFunction
			reduce(256), // +, reduce: MlrvalOrFunction
			reduce(256), // -, reduce: MlrvalOrFunction
			reduce(256), // .+, reduce: MlrvalOrFunction
			reduce(256), // .-, reduce: MlrvalOrFunction
			reduce(256), // *, reduce: MlrvalOrFunction
			reduce(256), // /, reduce: MlrvalOrFunction
			reduce(256), // //, reduce: MlrvalOrFunction
			reduce(256), // %, reduce: MlrvalOrFunction
			reduce(256), // .*, reduce: MlrvalOrFunction
			reduce(256), // ./, reduce: MlrvalOrFunction
			reduce(256), // .//, reduce: MlrvalOrFunction
			reduce(256), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(256), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(256), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(256), // [[, reduce: MlrvalOrFunction
			reduce(256), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1595
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(257), // >, reduce: MlrvalOrFunction
			reduce(257), // >>, reduce: MlrvalOrFunction
			reduce(257), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(257), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(257), // ?, reduce: MlrvalOrFunction
			reduce(257), // :, reduce: MlrvalOrFunction
			reduce(257), // ||, reduce: MlrvalOrFunction
			reduce(257), // ^^, reduce: MlrvalOrFunction
			reduce(257), // &&, reduce: MlrvalOrFunction
			reduce(257), // ??, reduce: MlrvalOrFunction
			reduce(257), // ???, reduce: MlrvalOrFunction
			reduce(257), // =~, reduce: MlrvalOrFunction
			reduce(257), // !=~, reduce: MlrvalOrFunction
			reduce(257), // ==, reduce: MlrvalOrFunction
			reduce(257), // !=, reduce: MlrvalOrFunction
			reduce(257), // <=>, reduce: MlrvalOrFunction
			reduce(257), // >=, reduce: MlrvalOrFunction
			reduce(257), // <, reduce: MlrvalOrFunction
			reduce(257), // <=, reduce: MlrvalOrFunction
			reduce(257), // ^, reduce: MlrvalOrFunction
			reduce(257), // &, reduce: MlrvalOrFunction
			reduce(257), // <<, reduce: MlrvalOrFunction
			reduce(257), // >>>, reduce: MlrvalOrFunction
			reduce(257), // +, reduce: MlrvalOrFunction
			reduce(257), // -, reduce: MlrvalOrFunction
			reduce(257), // .+, reduce: MlrvalOrFunction
			reduce(257), // .-, reduce: MlrvalOrFunction
			reduce(257), // *, reduce: MlrvalOrFunction
			reduce(257), // /, reduce: MlrvalOrFunction
			reduce(257), // //, reduce: MlrvalOrFunction
			reduce(257), // %, reduce: MlrvalOrFunction
			reduce(257), // .*, reduce: MlrvalOrFunction
			reduce(257), // ./, reduce: MlrvalOrFunction
			reduce(257), // .//, reduce: MlrvalOrFunction
			reduce(257), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(257), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(257), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(257), // [[, reduce: MlrvalOrFunction
			reduce(257), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1596
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(258), // >, reduce: MlrvalOrFunction
			reduce(258), // >>, reduce: MlrvalOrFunction
			reduce(258), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(258), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(258), // ?, reduce: MlrvalOrFunction
			reduce(258), // :, reduce: MlrvalOrFunction
			reduce(258), // ||, reduce: MlrvalOrFunction
			reduce(258), // ^^, reduce: MlrvalOrFunction
			reduce(258), // &&, reduce: MlrvalOrFunction
			reduce(258), // ??, reduce: MlrvalOrFunction
			reduce(258), // ???, reduce: MlrvalOrFunction
			reduce(258), // =~, reduce: MlrvalOrFunction
			reduce(258), // !=~, reduce: MlrvalOrFunction
			reduce(258), // ==, reduce: MlrvalOrFunction
			reduce(258), // !=, reduce: MlrvalOrFunction
			reduce(258), // <=>, reduce: MlrvalOrFunction
			reduce(258), // >=, reduce: MlrvalOrFunction
			reduce(258), // <, reduce: MlrvalOrFunction
			reduce(258), // <=, reduce: MlrvalOrFunction
			reduce(258), // ^, reduce: MlrvalOrFunction
			reduce(258), // &, reduce: MlrvalOrFunction
			reduce(258), // <<, reduce: MlrvalOrFunction
			reduce(258), // >>>, reduce: MlrvalOrFunction
			reduce(258), // +, reduce: MlrvalOrFunction
			reduce(258), // -, reduce: MlrvalOrFunction
			reduce(258), // .+, reduce: MlrvalOrFunction
			reduce(258), // .-, reduce: MlrvalOrFunction
			reduce(258), // *, reduce: MlrvalOrFunction
			reduce(258), // /, reduce: MlrvalOrFunction
			reduce(258), // //, reduce: MlrvalOrFunction
			reduce(258), // %, reduce: MlrvalOrFunction
			reduce(258), // .*, reduce: MlrvalOrFunction
			reduce(258), // ./, reduce: MlrvalOrFunction
			reduce(258), // .//, reduce: MlrvalOrFunction
			reduce(258), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(258), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(258), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(258), // [[, reduce: MlrvalOrFunction
			reduce(258), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1597
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(259), // >, reduce: MlrvalOrFunction
			reduce(259), // >>, reduce: MlrvalOrFunction
			reduce(259), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(259), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(259), // ?, reduce: MlrvalOrFunction
			reduce(259), // :, reduce: MlrvalOrFunction
			reduce(259), // ||, reduce: MlrvalOrFunction
			reduce(259), // ^^, reduce: MlrvalOrFunction
			reduce(259), // &&, reduce: MlrvalOrFunction
			reduce(259), // ??, reduce: MlrvalOrFunction
			reduce(259), // ???, reduce: MlrvalOrFunction
			reduce(259), // =~, reduce: MlrvalOrFunction
			reduce(259), // !=~, reduce: MlrvalOrFunction
			reduce(259), // ==, reduce: MlrvalOrFunction
			reduce(259), // !=, reduce: MlrvalOrFunction
			reduce(259), // <=>, reduce: MlrvalOrFunction
			reduce(259), // >=, reduce: MlrvalOrFunction
			reduce(259), // <, reduce: MlrvalOrFunction
			reduce(259), // <=, reduce: MlrvalOrFunction
			reduce(259), // ^, reduce: MlrvalOrFunction
			reduce(259), // &, reduce: MlrvalOrFunction
			reduce(259), // <<, reduce: MlrvalOrFunction
			reduce(259), // >>>, reduce: MlrvalOrFunction
			reduce(259), // +, reduce: MlrvalOrFunction
			reduce(259), // -, reduce: MlrvalOrFunction
			reduce(259), // .+, reduce: MlrvalOrFunction
			reduce(259), // .-, reduce: MlrvalOrFunction
			reduce(259), // *, reduce: MlrvalOrFunction
			reduce(259), // /, reduce: MlrvalOrFunction
			reduce(259), // //, reduce: MlrvalOrFunction
			reduce(259), // %, reduce: MlrvalOrFunction
			reduce(259), // .*, reduce: MlrvalOrFunction
			reduce(259), // ./, reduce: MlrvalOrFunction
			reduce(259), // .//, reduce: MlrvalOrFunction
			reduce(259), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(259), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(259), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(259), // [[, reduce: MlrvalOrFunction
			reduce(259), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1598
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(267), // >, reduce: MlrvalOrFunction
			reduce(267), // >>, reduce: MlrvalOrFunction
			reduce(267), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(267), // ], reduce: MlrvalOrFunction
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(267), // ?, reduce: MlrvalOrFunction
			reduce(267), // :, reduce: MlrvalOrFunction
			reduce(267), // ||, reduce: MlrvalOrFunction
			reduce(267), // ^^, reduce: MlrvalOrFunction
			reduce(267), // &&, reduce: MlrvalOrFunction
			reduce(267), // ??, reduce: MlrvalOrFunction
			reduce(267), // ???, reduce: MlrvalOrFunction
			reduce(267), // =~, reduce: MlrvalOrFunction
			reduce(267), // !=~, reduce: MlrvalOrFunction
			reduce(267), // ==, reduce: MlrvalOrFunction
			reduce(267), // !=, reduce: MlrvalOrFunction
			reduce(267), // <=>, reduce: MlrvalOrFunction
			reduce(267), // >=, reduce: MlrvalOrFunction
			reduce(267), // <, reduce: MlrvalOrFunction
			reduce(267), // <=, reduce: MlrvalOrFunction
			reduce(267), // ^, reduce: MlrvalOrFunction
			reduce(267), // &, reduce: MlrvalOrFunction
			reduce(267), // <<, reduce: MlrvalOrFunction
			reduce(267), // >>>, reduce: MlrvalOrFunction
			reduce(267), // +, reduce: MlrvalOrFunction
			reduce(267), // -, reduce: MlrvalOrFunction
			reduce(267), // .+, reduce: MlrvalOrFunction
			reduce(267), // .-, reduce: MlrvalOrFunction
			reduce(267), // *, reduce: MlrvalOrFunction
			reduce(267), // /, reduce: MlrvalOrFunction
			reduce(267), // //, reduce: MlrvalOrFunction
			reduce(267), // %, reduce: MlrvalOrFunction
			reduce(267), // .*, reduce: MlrvalOrFunction
			reduce(267), // ./, reduce: MlrvalOrFunction
			reduce(267), // .//, reduce: MlrvalOrFunction
			reduce(267), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(267), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(267), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(267), // [[, reduce: MlrvalOrFunction
			reduce(267), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1599
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2692), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1600
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2693), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1601
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2694), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1602
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2695), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1603
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(234), // >, reduce: MapLiteral
			reduce(234), // >>, reduce: MapLiteral
			reduce(234), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(234), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(234), // ], reduce: MapLiteral
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(234), // ?, reduce: MapLiteral
			nil,         // :
			reduce(234), // ||, reduce: MapLiteral
			reduce(234), // ^^, reduce: MapLiteral
			reduce(234), // &&, reduce: MapLiteral
			reduce(234), // ??, reduce: MapLiteral
			reduce(234), // ???, reduce: MapLiteral
			reduce(234), // =~, reduce: MapLiteral
			reduce(234), // !=~, reduce: MapLiteral
			reduce(234), // ==, reduce: MapLiteral
			reduce(234), // !=, reduce: MapLiteral
			reduce(234), // <=>, reduce: MapLiteral
			reduce(234), // >=, reduce: MapLiteral
			reduce(234), // <, reduce: MapLiteral
			reduce(234), // <=, reduce: MapLiteral
			reduce(234), // ^, reduce: MapLiteral
			reduce(234), // &, reduce: MapLiteral
			reduce(234), // <<, reduce: MapLiteral
			reduce(234), // >>>, reduce: MapLiteral
			reduce(234), // +, reduce: MapLiteral
			reduce(234), // -, reduce: MapLiteral
			reduce(234), // .+, reduce: MapLiteral
			reduce(234), // .-, reduce: MapLiteral
			reduce(234), // *, reduce: MapLiteral
			reduce(234), // /, reduce: MapLiteral
			reduce(234), // //, reduce: MapLiteral
			reduce(234), // %, reduce: MapLiteral
			reduce(234), // .*, reduce: MapLiteral
			reduce(234), // ./, reduce: MapLiteral
			reduce(234), // .//, reduce: MapLiteral
			reduce(234), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(234), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(234), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(234), // [[, reduce: MapLiteral
			reduce(234), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1604
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(2696), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1605
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(752),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(756),  // (
			nil,         // )
			nil,         // emitp
			shift(768),  // field_name
			shift(769),  // $[
			reduce(231), // ], reduce: ArrayLiteralElements
			shift(771),  // braced_field_name
			shift(772),  // $[[
			shift(773),  // $[[[
			shift(774),  // full_srec
			shift(776),  // oosvar_name
			shift(777),  // @[
			shift(778),  // braced_oosvar_name
			shift(779),  // full_oosvar
			shift(780),  // all
			shift(781),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(796),  // +
			shift(798),  // -
			shift(799),  // .+
			shift(800),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(803),  // !
			shift(804),  // ~
			nil,         // **
			shift(809),  // string_literal
			shift(810),  // regex_case_insensitive
			shift(811),  // int_literal
			shift(812),  // float_literal
			shift(813),  // boolean_literal
			shift(814),  // null_literal
			shift(815),  // inf_literal
			shift(816),  // nan_literal
			shift(817),  // const_M_PI
			shift(818),  // const_M_E
			shift(819),  // panic
			shift(821),  // [
			shift(824),  // ctx_IPS
			shift(825),  // ctx_IFS
			shift(826),  // ctx_IRS
			shift(827),  // ctx_OPS
			shift(828),  // ctx_OFS
			shift(829),  // ctx_ORS
			shift(830),  // ctx_FLATSEP
			shift(831),  // ctx_NF
			shift(832),  // ctx_NR
			shift(833),  // ctx_FNR
			shift(834),  // ctx_FILENAME
			shift(835),  // ctx_FILENUM
			shift(837),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(844),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1606
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(2698), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1607
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2699), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1608
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2700), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1609
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2701), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1610
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2702), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1611
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1612
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1613
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1614
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1615
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1616
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1617
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1618
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1619
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1620
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1621
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1622
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1623
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1624
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1625
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1626
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1627
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1628
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1629
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1630
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1631
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1632
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1633
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1634
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1635
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1636
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(197), // >, reduce: UnaryOpTerm
			reduce(197), // >>, reduce: UnaryOpTerm
			reduce(197), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(197), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(197), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(197), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(197), // ||, reduce: UnaryOpTerm
			reduce(197), // ^^, reduce: UnaryOpTerm
			reduce(197), // &&, reduce: UnaryOpTerm
			reduce(197), // ??, reduce: UnaryOpTerm
			reduce(197), // ???, reduce: UnaryOpTerm
			reduce(197), // =~, reduce: UnaryOpTerm
			reduce(197), // !=~, reduce: UnaryOpTerm
			reduce(197), // ==, reduce: UnaryOpTerm
			reduce(197), // !=, reduce: UnaryOpTerm
			reduce(197), // <=>, reduce: UnaryOpTerm
			reduce(197), // >=, reduce: UnaryOpTerm
			reduce(197), // <, reduce: UnaryOpTerm
			reduce(197), // <=, reduce: UnaryOpTerm
			reduce(197), // ^, reduce: UnaryOpTerm
			reduce(197), // &, reduce: UnaryOpTerm
			reduce(197), // <<, reduce: UnaryOpTerm
			reduce(197), // >>>, reduce: UnaryOpTerm
			reduce(197), // +, reduce: UnaryOpTerm
			reduce(197), // -, reduce: UnaryOpTerm
			reduce(197), // .+, reduce: UnaryOpTerm
			reduce(197), // .-, reduce: UnaryOpTerm
			reduce(197), // *, reduce: UnaryOpTerm
			reduce(197), // /, reduce: UnaryOpTerm
			reduce(197), // //, reduce: UnaryOpTerm
			reduce(197), // %, reduce: UnaryOpTerm
			reduce(197), // .*, reduce: UnaryOpTerm
			reduce(197), // ./, reduce: UnaryOpTerm
			reduce(197), // .//, reduce: UnaryOpTerm
			reduce(197), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1637
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1638
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1639
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1640
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1641
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1642
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1643
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1644
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(198), // >, reduce: UnaryOpTerm
			reduce(198), // >>, reduce: UnaryOpTerm
			reduce(198), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(198), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(198), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(198), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(198), // ||, reduce: UnaryOpTerm
			reduce(198), // ^^, reduce: UnaryOpTerm
			reduce(198), // &&, reduce: UnaryOpTerm
			reduce(198), // ??, reduce: UnaryOpTerm
			reduce(198), // ???, reduce: UnaryOpTerm
			reduce(198), // =~, reduce: UnaryOpTerm
			reduce(198), // !=~, reduce: UnaryOpTerm
			reduce(198), // ==, reduce: UnaryOpTerm
			reduce(198), // !=, reduce: UnaryOpTerm
			reduce(198), // <=>, reduce: UnaryOpTerm
			reduce(198), // >=, reduce: UnaryOpTerm
			reduce(198), // <, reduce: UnaryOpTerm
			reduce(198), // <=, reduce: UnaryOpTerm
			reduce(198), // ^, reduce: UnaryOpTerm
			reduce(198), // &, reduce: UnaryOpTerm
			reduce(198), // <<, reduce: UnaryOpTerm
			reduce(198), // >>>, reduce: UnaryOpTerm
			reduce(198), // +, reduce: UnaryOpTerm
			reduce(198), // -, reduce: UnaryOpTerm
			reduce(198), // .+, reduce: UnaryOpTerm
			reduce(198), // .-, reduce: UnaryOpTerm
			reduce(198), // *, reduce: UnaryOpTerm
			reduce(198), // /, reduce: UnaryOpTerm
			reduce(198), // //, reduce: UnaryOpTerm
			reduce(198), // %, reduce: UnaryOpTerm
			reduce(198), // .*, reduce: UnaryOpTerm
			reduce(198), // ./, reduce: UnaryOpTerm
			reduce(198), // .//, reduce: UnaryOpTerm
			reduce(198), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1645
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(199), // >, reduce: UnaryOpTerm
			reduce(199), // >>, reduce: UnaryOpTerm
			reduce(199), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(199), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(199), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(199), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(199), // ||, reduce: UnaryOpTerm
			reduce(199), // ^^, reduce: UnaryOpTerm
			reduce(199), // &&, reduce: UnaryOpTerm
			reduce(199), // ??, reduce: UnaryOpTerm
			reduce(199), // ???, reduce: UnaryOpTerm
			reduce(199), // =~, reduce: UnaryOpTerm
			reduce(199), // !=~, reduce: UnaryOpTerm
			reduce(199), // ==, reduce: UnaryOpTerm
			reduce(199), // !=, reduce: UnaryOpTerm
			reduce(199), // <=>, reduce: UnaryOpTerm
			reduce(199), // >=, reduce: UnaryOpTerm
			reduce(199), // <, reduce: UnaryOpTerm
			reduce(199), // <=, reduce: UnaryOpTerm
			reduce(199), // ^, reduce: UnaryOpTerm
			reduce(199), // &, reduce: UnaryOpTerm
			reduce(199), // <<, reduce: UnaryOpTerm
			reduce(199), // >>>, reduce: UnaryOpTerm
			reduce(199), // +, reduce: UnaryOpTerm
			reduce(199), // -, reduce: UnaryOpTerm
			reduce(199), // .+, reduce: UnaryOpTerm
			reduce(199), // .-, reduce: UnaryOpTerm
			reduce(199), // *, reduce: UnaryOpTerm
			reduce(199), // /, reduce: UnaryOpTerm
			reduce(199), // //, reduce: UnaryOpTerm
			reduce(199), // %, reduce: UnaryOpTerm
			reduce(199), // .*, reduce: UnaryOpTerm
			reduce(199), // ./, reduce: UnaryOpTerm
			reduce(199), // .//, reduce: UnaryOpTerm
			reduce(199), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1646
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(200), // >, reduce: UnaryOpTerm
			reduce(200), // >>, reduce: UnaryOpTerm
			reduce(200), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(200), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(200), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(200), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(200), // ||, reduce: UnaryOpTerm
			reduce(200), // ^^, reduce: UnaryOpTerm
			reduce(200), // &&, reduce: UnaryOpTerm
			reduce(200), // ??, reduce: UnaryOpTerm
			reduce(200), // ???, reduce: UnaryOpTerm
			reduce(200), // =~, reduce: UnaryOpTerm
			reduce(200), // !=~, reduce: UnaryOpTerm
			reduce(200), // ==, reduce: UnaryOpTerm
			reduce(200), // !=, reduce: UnaryOpTerm
			reduce(200), // <=>, reduce: UnaryOpTerm
			reduce(200), // >=, reduce: UnaryOpTerm
			reduce(200), // <, reduce: UnaryOpTerm
			reduce(200), // <=, reduce: UnaryOpTerm
			reduce(200), // ^, reduce: UnaryOpTerm
			reduce(200), // &, reduce: UnaryOpTerm
			reduce(200), // <<, reduce: UnaryOpTerm
			reduce(200), // >>>, reduce: UnaryOpTerm
			reduce(200), // +, reduce: UnaryOpTerm
			reduce(200), // -, reduce: UnaryOpTerm
			reduce(200), // .+, reduce: UnaryOpTerm
			reduce(200), // .-, reduce: UnaryOpTerm
			reduce(200), // *, reduce: UnaryOpTerm
			reduce(200), // /, reduce: UnaryOpTerm
			reduce(200), // //, reduce: UnaryOpTerm
			reduce(200), // %, reduce: UnaryOpTerm
			reduce(200), // .*, reduce: UnaryOpTerm
			reduce(200), // ./, reduce: UnaryOpTerm
			reduce(200), // .//, reduce: UnaryOpTerm
			reduce(200), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1647
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1648
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(201), // >, reduce: UnaryOpTerm
			reduce(201), // >>, reduce: UnaryOpTerm
			reduce(201), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(201), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(201), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(201), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(201), // ||, reduce: UnaryOpTerm
			reduce(201), // ^^, reduce: UnaryOpTerm
			reduce(201), // &&, reduce: UnaryOpTerm
			reduce(201), // ??, reduce: UnaryOpTerm
			reduce(201), // ???, reduce: UnaryOpTerm
			reduce(201), // =~, reduce: UnaryOpTerm
			reduce(201), // !=~, reduce: UnaryOpTerm
			reduce(201), // ==, reduce: UnaryOpTerm
			reduce(201), // !=, reduce: UnaryOpTerm
			reduce(201), // <=>, reduce: UnaryOpTerm
			reduce(201), // >=, reduce: UnaryOpTerm
			reduce(201), // <, reduce: UnaryOpTerm
			reduce(201), // <=, reduce: UnaryOpTerm
			reduce(201), // ^, reduce: UnaryOpTerm
			reduce(201), // &, reduce: UnaryOpTerm
			reduce(201), // <<, reduce: UnaryOpTerm
			reduce(201), // >>>, reduce: UnaryOpTerm
			reduce(201), // +, reduce: UnaryOpTerm
			reduce(201), // -, reduce: UnaryOpTerm
			reduce(201), // .+, reduce: UnaryOpTerm
			reduce(201), // .-, reduce: UnaryOpTerm
			reduce(201), // *, reduce: UnaryOpTerm
			reduce(201), // /, reduce: UnaryOpTerm
			reduce(201), // //, reduce: UnaryOpTerm
			reduce(201), // %, reduce: UnaryOpTerm
			reduce(201), // .*, reduce: UnaryOpTerm
			reduce(201), // ./, reduce: UnaryOpTerm
			reduce(201), // .//, reduce: UnaryOpTerm
			reduce(201), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1649
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(202), // >, reduce: UnaryOpTerm
			reduce(202), // >>, reduce: UnaryOpTerm
			reduce(202), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(202), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(202), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(202), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(202), // ||, reduce: UnaryOpTerm
			reduce(202), // ^^, reduce: UnaryOpTerm
			reduce(202), // &&, reduce: UnaryOpTerm
			reduce(202), // ??, reduce: UnaryOpTerm
			reduce(202), // ???, reduce: UnaryOpTerm
			reduce(202), // =~, reduce: UnaryOpTerm
			reduce(202), // !=~, reduce: UnaryOpTerm
			reduce(202), // ==, reduce: UnaryOpTerm
			reduce(202), // !=, reduce: UnaryOpTerm
			reduce(202), // <=>, reduce: UnaryOpTerm
			reduce(202), // >=, reduce: UnaryOpTerm
			reduce(202), // <, reduce: UnaryOpTerm
			reduce(202), // <=, reduce: UnaryOpTerm
			reduce(202), // ^, reduce: UnaryOpTerm
			reduce(202), // &, reduce: UnaryOpTerm
			reduce(202), // <<, reduce: UnaryOpTerm
			reduce(202), // >>>, reduce: UnaryOpTerm
			reduce(202), // +, reduce: UnaryOpTerm
			reduce(202), // -, reduce: UnaryOpTerm
			reduce(202), // .+, reduce: UnaryOpTerm
			reduce(202), // .-, reduce: UnaryOpTerm
			reduce(202), // *, reduce: UnaryOpTerm
			reduce(202), // /, reduce: UnaryOpTerm
			reduce(202), // //, reduce: UnaryOpTerm
			reduce(202), // %, reduce: UnaryOpTerm
			reduce(202), // .*, reduce: UnaryOpTerm
			reduce(202), // ./, reduce: UnaryOpTerm
			reduce(202), // .//, reduce: UnaryOpTerm
			reduce(202), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1650
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(752),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(756),  // (
			nil,         // )
			nil,         // emitp
			shift(768),  // field_name
			shift(769),  // $[
			nil,         // ]
			shift(771),  // braced_field_name
			shift(772),  // $[[
			shift(773),  // $[[[
			shift(774),  // full_srec
			shift(776),  // oosvar_name
			shift(777),  // @[
			shift(778),  // braced_oosvar_name
			shift(779),  // full_oosvar
			shift(780),  // all
			shift(781),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2736), // +
			shift(2737), // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(809),  // string_literal
			shift(810),  // regex_case_insensitive
			shift(811),  // int_literal
			shift(812),  // float_literal
			shift(813),  // boolean_literal
			shift(814),  // null_literal
			shift(815),  // inf_literal
			shift(816),  // nan_literal
			shift(817),  // const_M_PI
			shift(818),  // const_M_E
			shift(819),  // panic
			shift(821),  // [
			shift(824),  // ctx_IPS
			shift(825),  // ctx_IFS
			shift(826),  // ctx_IRS
			shift(827),  // ctx_OPS
			shift(828),  // ctx_OFS
			shift(829),  // ctx_ORS
			shift(830),  // ctx_FLATSEP
			shift(831),  // ctx_NF
			shift(832),  // ctx_NR
			shift(833),  // ctx_FNR
			shift(834),  // ctx_FILENAME
			shift(835),  // ctx_FILENUM
			shift(837),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(844),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1651
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(2740), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1652
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1653
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1654
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(228), // >, reduce: ArrayLiteral
			reduce(228), // >>, reduce: ArrayLiteral
			reduce(228), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(228), // ,, reduce: ArrayLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(228), // ], reduce: ArrayLiteral
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(228), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(228), // ||, reduce: ArrayLiteral
			reduce(228), // ^^, reduce: ArrayLiteral
			reduce(228), // &&, reduce: ArrayLiteral
			reduce(228), // ??, reduce: ArrayLiteral
			reduce(228), // ???, reduce: ArrayLiteral
			reduce(228), // =~, reduce: ArrayLiteral
			reduce(228), // !=~, reduce: ArrayLiteral
			reduce(228), // ==, reduce: ArrayLiteral
			reduce(228), // !=, reduce: ArrayLiteral
			reduce(228), // <=>, reduce: ArrayLiteral
			reduce(228), // >=, reduce: ArrayLiteral
			reduce(228), // <, reduce: ArrayLiteral
			reduce(228), // <=, reduce: ArrayLiteral
			reduce(228), // ^, reduce: ArrayLiteral
			reduce(228), // &, reduce: ArrayLiteral
			reduce(228), // <<, reduce: ArrayLiteral
			reduce(228), // >>>, reduce: ArrayLiteral
			reduce(228), // +, reduce: ArrayLiteral
			reduce(228), // -, reduce: ArrayLiteral
			reduce(228), // .+, reduce: ArrayLiteral
			reduce(228), // .-, reduce: ArrayLiteral
			reduce(228), // *, reduce: ArrayLiteral
			reduce(228), // /, reduce: ArrayLiteral
			reduce(228), // //, reduce: ArrayLiteral
			reduce(228), // %, reduce: ArrayLiteral
			reduce(228), // .*, reduce: ArrayLiteral
			reduce(228), // ./, reduce: ArrayLiteral
			reduce(228), // .//, reduce: ArrayLiteral
			reduce(228), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(228), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(228), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(228), // [[, reduce: ArrayLiteral
			reduce(228), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1655
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2743), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1656
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(229), // $, reduce: ArrayLiteral
			nil,         // empty
			reduce(229), // ;, reduce: ArrayLiteral
			reduce(229), // {, reduce: ArrayLiteral
			nil,         // }
			reduce(229), // =, reduce: ArrayLiteral
			nil,         // unset
			nil,         // filter
			reduce(229), // >, reduce: ArrayLiteral
			reduce(229), // >>, reduce: ArrayLiteral
			reduce(229), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(229), // ||=, reduce: ArrayLiteral
			reduce(229), // ^^=, reduce: ArrayLiteral
			reduce(229), // &&=, reduce: ArrayLiteral
			reduce(229), // ??=, reduce: ArrayLiteral
			reduce(229), // ???=, reduce: ArrayLiteral
			reduce(229), // |=, reduce: ArrayLiteral
			reduce(229), // &=, reduce: ArrayLiteral
			reduce(229), // ^=, reduce: ArrayLiteral
			reduce(229), // <<=, reduce: ArrayLiteral
			reduce(229), // >>=, reduce: ArrayLiteral
			reduce(229), // >>>=, reduce: ArrayLiteral
			reduce(229), // +=, reduce: ArrayLiteral
			reduce(229), // .=, reduce: ArrayLiteral
			reduce(229), // -=, reduce: ArrayLiteral
			reduce(229), // *=, reduce: ArrayLiteral
			reduce(229), // /=, reduce: ArrayLiteral
			reduce(229), // //=, reduce: ArrayLiteral
			reduce(229), // %=, reduce: ArrayLiteral
			reduce(229), // **=, reduce: ArrayLiteral
			reduce(229), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(229), // ||, reduce: ArrayLiteral
			reduce(229), // ^^, reduce: ArrayLiteral
			reduce(229), // &&, reduce: ArrayLiteral
			reduce(229), // ??, reduce: ArrayLiteral
			reduce(229), // ???, reduce: ArrayLiteral
			reduce(229), // =~, reduce: ArrayLiteral
			reduce(229), // !=~, reduce: ArrayLiteral
			reduce(229), // ==, reduce: ArrayLiteral
			reduce(229), // !=, reduce: ArrayLiteral
			reduce(229), // <=>, reduce: ArrayLiteral
			reduce(229), // >=, reduce: ArrayLiteral
			reduce(229), // <, reduce: ArrayLiteral
			reduce(229), // <=, reduce: ArrayLiteral
			reduce(229), // ^, reduce: ArrayLiteral
			reduce(229), // &, reduce: ArrayLiteral
			reduce(229), // <<, reduce: ArrayLiteral
			reduce(229), // >>>, reduce: ArrayLiteral
			reduce(229), // +, reduce: ArrayLiteral
			reduce(229), // -, reduce: ArrayLiteral
			reduce(229), // .+, reduce: ArrayLiteral
			reduce(229), // .-, reduce: ArrayLiteral
			reduce(229), // *, reduce: ArrayLiteral
			reduce(229), // /, reduce: ArrayLiteral
			reduce(229), // //, reduce: ArrayLiteral
			reduce(229), // %, reduce: ArrayLiteral
			reduce(229), // .*, reduce: ArrayLiteral
			reduce(229), // ./, reduce: ArrayLiteral
			reduce(229), // .//, reduce: ArrayLiteral
			reduce(229), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(229), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(229), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(229), // [[, reduce: ArrayLiteral
			reduce(229), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1657
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(2744), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1658
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1659
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(2747), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1660
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1661
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(255), // $, reduce: ENV
			nil,         // empty
			reduce(255), // ;, reduce: ENV
			reduce(255), // {, reduce: ENV
			nil,         // }
			reduce(255), // =, reduce: ENV
			nil,         // unset
			nil,         // filter
			reduce(255), // >, reduce: ENV
			reduce(255), // >>, reduce: ENV
			reduce(255), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(255), // ||=, reduce: ENV
			reduce(255), // ^^=, reduce: ENV
			reduce(255), // &&=, reduce: ENV
			reduce(255), // ??=, reduce: ENV
			reduce(255), // ???=, reduce: ENV
			reduce(255), // |=, reduce: ENV
			reduce(255), // &=, reduce: ENV
			reduce(255), // ^=, reduce: ENV
			reduce(255), // <<=, reduce: ENV
			reduce(255), // >>=, reduce: ENV
			reduce(255), // >>>=, reduce: ENV
			reduce(255), // +=, reduce: ENV
			reduce(255), // .=, reduce: ENV
			reduce(255), // -=, reduce: ENV
			reduce(255), // *=, reduce: ENV
			reduce(255), // /=, reduce: ENV
			reduce(255), // //=, reduce: ENV
			reduce(255), // %=, reduce: ENV
			reduce(255), // **=, reduce: ENV
			reduce(255), // ?, reduce: ENV
			nil,         // :
			reduce(255), // ||, reduce: ENV
			reduce(255), // ^^, reduce: ENV
			reduce(255), // &&, reduce: ENV
			reduce(255), // ??, reduce: ENV
			reduce(255), // ???, reduce: ENV
			reduce(255), // =~, reduce: ENV
			reduce(255), // !=~, reduce: ENV
			reduce(255), // ==, reduce: ENV
			reduce(255), // !=, reduce: ENV
			reduce(255), // <=>, reduce: ENV
			reduce(255), // >=, reduce: ENV
			reduce(255), // <, reduce: ENV
			reduce(255), // <=, reduce: ENV
			reduce(255), // ^, reduce: ENV
			reduce(255), // &, reduce: ENV
			reduce(255), // <<, reduce: ENV
			reduce(255), // >>>, reduce: ENV
			reduce(255), // +, reduce: ENV
			reduce(255), // -, reduce: ENV
			reduce(255), // .+, reduce: ENV
			reduce(255), // .-, reduce: ENV
			reduce(255), // *, reduce: ENV
			reduce(255), // /, reduce: ENV
			reduce(255), // //, reduce: ENV
			reduce(255), // %, reduce: ENV
			reduce(255), // .*, reduce: ENV
			reduce(255), // ./, reduce: ENV
			reduce(255), // .//, reduce: ENV
			reduce(255), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(255), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(255), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(255), // [[, reduce: ENV
			reduce(255), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1662
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2749), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1663
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			shift(2750), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1664
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(2752), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(273), // ), reduce: FcnArgs
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1665
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(2753), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1666
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(214), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(214), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(214), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1667
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(211), // >, reduce: MlrvalOrFunction
			reduce(211), // >>, reduce: MlrvalOrFunction
			reduce(211), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(211), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(211), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(211), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(211), // ||, reduce: MlrvalOrFunction
			reduce(211), // ^^, reduce: MlrvalOrFunction
			reduce(211), // &&, reduce: MlrvalOrFunction
			reduce(211), // ??, reduce: MlrvalOrFunction
			reduce(211), // ???, reduce: MlrvalOrFunction
			reduce(211), // =~, reduce: MlrvalOrFunction
			reduce(211), // !=~, reduce: MlrvalOrFunction
			reduce(211), // ==, reduce: MlrvalOrFunction
			reduce(211), // !=, reduce: MlrvalOrFunction
			reduce(211), // <=>, reduce: MlrvalOrFunction
			reduce(211), // >=, reduce: MlrvalOrFunction
			reduce(211), // <, reduce: MlrvalOrFunction
			reduce(211), // <=, reduce: MlrvalOrFunction
			reduce(211), // ^, reduce: MlrvalOrFunction
			reduce(211), // &, reduce: MlrvalOrFunction
			reduce(211), // <<, reduce: MlrvalOrFunction
			reduce(211), // >>>, reduce: MlrvalOrFunction
			reduce(211), // +, reduce: MlrvalOrFunction
			reduce(211), // -, reduce: MlrvalOrFunction
			reduce(211), // .+, reduce: MlrvalOrFunction
			reduce(211), // .-, reduce: MlrvalOrFunction
			reduce(211), // *, reduce: MlrvalOrFunction
			reduce(211), // /, reduce: MlrvalOrFunction
			reduce(211), // //, reduce: MlrvalOrFunction
			reduce(211), // %, reduce: MlrvalOrFunction
			reduce(211), // .*, reduce: MlrvalOrFunction
			reduce(211), // ./, reduce: MlrvalOrFunction
			reduce(211), // .//, reduce: MlrvalOrFunction
			reduce(211), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(211), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(211), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(211), // [[, reduce: MlrvalOrFunction
			reduce(211), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1668
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1669
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(268), // $, reduce: FunctionCallsite
			nil,         // empty
			reduce(268), // ;, reduce: FunctionCallsite
			reduce(268), // {, reduce: FunctionCallsite
			nil,         // }
			reduce(268), // =, reduce: FunctionCallsite
			nil,         // unset
			nil,         // filter
			reduce(268), // >, reduce: FunctionCallsite
			reduce(268), // >>, reduce: FunctionCallsite
			reduce(268), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(268), // ||=, reduce: FunctionCallsite
			reduce(268), // ^^=, reduce: FunctionCallsite
			reduce(268), // &&=, reduce: FunctionCallsite
			reduce(268), // ??=, reduce: FunctionCallsite
			reduce(268), // ???=, reduce: FunctionCallsite
			reduce(268), // |=, reduce: FunctionCallsite
			reduce(268), // &=, reduce: FunctionCallsite
			reduce(268), // ^=, reduce: FunctionCallsite
			reduce(268), // <<=, reduce: FunctionCallsite
			reduce(268), // >>=, reduce: FunctionCallsite
			reduce(268), // >>>=, reduce: FunctionCallsite
			reduce(268), // +=, reduce: FunctionCallsite
			reduce(268), // .=, reduce: FunctionCallsite
			reduce(268), // -=, reduce: FunctionCallsite
			reduce(268), // *=, reduce: FunctionCallsite
			reduce(268), // /=, reduce: FunctionCallsite
			reduce(268), // //=, reduce: FunctionCallsite
			reduce(268), // %=, reduce: FunctionCallsite
			reduce(268), // **=, reduce: FunctionCallsite
			reduce(268), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(268), // ||, reduce: FunctionCallsite
			reduce(268), // ^^, reduce: FunctionCallsite
			reduce(268), // &&, reduce: FunctionCallsite
			reduce(268), // ??, reduce: FunctionCallsite
			reduce(268), // ???, reduce: FunctionCallsite
			reduce(268), // =~, reduce: FunctionCallsite
			reduce(268), // !=~, reduce: FunctionCallsite
			reduce(268), // ==, reduce: FunctionCallsite
			reduce(268), // !=, reduce: FunctionCallsite
			reduce(268), // <=>, reduce: FunctionCallsite
			reduce(268), // >=, reduce: FunctionCallsite
			reduce(268), // <, reduce: FunctionCallsite
			reduce(268), // <=, reduce: FunctionCallsite
			reduce(268), // ^, reduce: FunctionCallsite
			reduce(268), // &, reduce: FunctionCallsite
			reduce(268), // <<, reduce: FunctionCallsite
			reduce(268), // >>>, reduce: FunctionCallsite
			reduce(268), // +, reduce: FunctionCallsite
			reduce(268), // -, reduce: FunctionCallsite
			reduce(268), // .+, reduce: FunctionCallsite
			reduce(268), // .-, reduce: FunctionCallsite
			reduce(268), // *, reduce: FunctionCallsite
			reduce(268), // /, reduce: FunctionCallsite
			reduce(268), // //, reduce: FunctionCallsite
			reduce(268), // %, reduce: FunctionCallsite
			reduce(268), // .*, reduce: FunctionCallsite
			reduce(268), // ./, reduce: FunctionCallsite
			reduce(268), // .//, reduce: FunctionCallsite
			reduce(268), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(268), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(268), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(268), // [[, reduce: FunctionCallsite
			reduce(268), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1670
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(110), // >, reduce: OosvarValue
			reduce(110), // >>, reduce: OosvarValue
			reduce(110), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(110), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(110), // ), reduce: OosvarValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(110), // ?, reduce: OosvarValue
			nil,         // :
			reduce(110), // ||, reduce: OosvarValue
			reduce(110), // ^^, reduce: OosvarValue
			reduce(110), // &&, reduce: OosvarValue
			reduce(110), // ??, reduce: OosvarValue
			reduce(110), // ???, reduce: OosvarValue
			reduce(110), // =~, reduce: OosvarValue
			reduce(110), // !=~, reduce: OosvarValue
			reduce(110), // ==, reduce: OosvarValue
			reduce(110), // !=, reduce: OosvarValue
			reduce(110), // <=>, reduce: OosvarValue
			reduce(110), // >=, reduce: OosvarValue
			reduce(110), // <, reduce: OosvarValue
			reduce(110), // <=, reduce: OosvarValue
			reduce(110), // ^, reduce: OosvarValue
			reduce(110), // &, reduce: OosvarValue
			reduce(110), // <<, reduce: OosvarValue
			reduce(110), // >>>, reduce: OosvarValue
			reduce(110), // +, reduce: OosvarValue
			reduce(110), // -, reduce: OosvarValue
			reduce(110), // .+, reduce: OosvarValue
			reduce(110), // .-, reduce: OosvarValue
			reduce(110), // *, reduce: OosvarValue
			reduce(110), // /, reduce: OosvarValue
			reduce(110), // //, reduce: OosvarValue
			reduce(110), // %, reduce: OosvarValue
			reduce(110), // .*, reduce: OosvarValue
			reduce(110), // ./, reduce: OosvarValue
			reduce(110), // .//, reduce: OosvarValue
			reduce(110), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(110), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(110), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(110), // [[, reduce: OosvarValue
			reduce(110), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1671
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(112), // >, reduce: OosvarValue
			reduce(112), // >>, reduce: OosvarValue
			reduce(112), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(112), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(112), // ), reduce: OosvarValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(112), // ?, reduce: OosvarValue
			nil,         // :
			reduce(112), // ||, reduce: OosvarValue
			reduce(112), // ^^, reduce: OosvarValue
			reduce(112), // &&, reduce: OosvarValue
			reduce(112), // ??, reduce: OosvarValue
			reduce(112), // ???, reduce: OosvarValue
			reduce(112), // =~, reduce: OosvarValue
			reduce(112), // !=~, reduce: OosvarValue
			reduce(112), // ==, reduce: OosvarValue
			reduce(112), // !=, reduce: OosvarValue
			reduce(112), // <=>, reduce: OosvarValue
			reduce(112), // >=, reduce: OosvarValue
			reduce(112), // <, reduce: OosvarValue
			reduce(112), // <=, reduce: OosvarValue
			reduce(112), // ^, reduce: OosvarValue
			reduce(112), // &, reduce: OosvarValue
			reduce(112), // <<, reduce: OosvarValue
			reduce(112), // >>>, reduce: OosvarValue
			reduce(112), // +, reduce: OosvarValue
			reduce(112), // -, reduce: OosvarValue
			reduce(112), // .+, reduce: OosvarValue
			reduce(112), // .-, reduce: OosvarValue
			reduce(112), // *, reduce: OosvarValue
			reduce(112), // /, reduce: OosvarValue
			reduce(112), // //, reduce: OosvarValue
			reduce(112), // %, reduce: OosvarValue
			reduce(112), // .*, reduce: OosvarValue
			reduce(112), // ./, reduce: OosvarValue
			reduce(112), // .//, reduce: OosvarValue
			reduce(112), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(112), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(112), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(112), // [[, reduce: OosvarValue
			reduce(112), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1672
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(111), // >, reduce: OosvarValue
			reduce(111), // >>, reduce: OosvarValue
			reduce(111), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(111), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(111), // ), reduce: OosvarValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(111), // ?, reduce: OosvarValue
			nil,         // :
			reduce(111), // ||, reduce: OosvarValue
			reduce(111), // ^^, reduce: OosvarValue
			reduce(111), // &&, reduce: OosvarValue
			reduce(111), // ??, reduce: OosvarValue
			reduce(111), // ???, reduce: OosvarValue
			reduce(111), // =~, reduce: OosvarValue
			reduce(111), // !=~, reduce: OosvarValue
			reduce(111), // ==, reduce: OosvarValue
			reduce(111), // !=, reduce: OosvarValue
			reduce(111), // <=>, reduce: OosvarValue
			reduce(111), // >=, reduce: OosvarValue
			reduce(111), // <, reduce: OosvarValue
			reduce(111), // <=, reduce: OosvarValue
			reduce(111), // ^, reduce: OosvarValue
			reduce(111), // &, reduce: OosvarValue
			reduce(111), // <<, reduce: OosvarValue
			reduce(111), // >>>, reduce: OosvarValue
			reduce(111), // +, reduce: OosvarValue
			reduce(111), // -, reduce: OosvarValue
			reduce(111), // .+, reduce: OosvarValue
			reduce(111), // .-, reduce: OosvarValue
			reduce(111), // *, reduce: OosvarValue
			reduce(111), // /, reduce: OosvarValue
			reduce(111), // //, reduce: OosvarValue
			reduce(111), // %, reduce: OosvarValue
			reduce(111), // .*, reduce: OosvarValue
			reduce(111), // ./, reduce: OosvarValue
			reduce(111), // .//, reduce: OosvarValue
			reduce(111), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(111), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(111), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(111), // [[, reduce: OosvarValue
			reduce(111), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1673
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(99), // >, reduce: FieldValue
			reduce(99), // >>, reduce: FieldValue
			reduce(99), // |, reduce: FieldValue
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(99), // ,, reduce: FieldValue
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			reduce(99), // ), reduce: FieldValue
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(99), // ?, reduce: FieldValue
			nil,        // :
			reduce(99), // ||, reduce: FieldValue
			reduce(99), // ^^, reduce: FieldValue
			reduce(99), // &&, reduce: FieldValue
			reduce(99), // ??, reduce: FieldValue
			reduce(99), // ???, reduce: FieldValue
			reduce(99), // =~, reduce: FieldValue
			reduce(99), // !=~, reduce: FieldValue
			reduce(99), // ==, reduce: FieldValue
			reduce(99), // !=, reduce: FieldValue
			reduce(99), // <=>, reduce: FieldValue
			reduce(99), // >=, reduce: FieldValue
			reduce(99), // <, reduce: FieldValue
			reduce(99), // <=, reduce: FieldValue
			reduce(99), // ^, reduce: FieldValue
			reduce(99), // &, reduce: FieldValue
			reduce(99), // <<, reduce: FieldValue
			reduce(99), // >>>, reduce: FieldValue
			reduce(99), // +, reduce: FieldValue
			reduce(99), // -, reduce: FieldValue
			reduce(99), // .+, reduce: FieldValue
			reduce(99), // .-, reduce: FieldValue
			reduce(99), // *, reduce: FieldValue
			reduce(99), // /, reduce: FieldValue
			reduce(99), // //, reduce: FieldValue
			reduce(99), // %, reduce: FieldValue
			reduce(99), // .*, reduce: FieldValue
			reduce(99), // ./, reduce: FieldValue
			reduce(99), // .//, reduce: FieldValue
			reduce(99), // ., reduce: FieldValue
			nil,        // !
			nil,        // ~
			reduce(99), // **, reduce: FieldValue
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(99), // [, reduce: FieldValue
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(99), // [[, reduce: FieldValue
			reduce(99), // [[[, reduce: FieldValue
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1674
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(101), // >, reduce: FieldValue
			reduce(101), // >>, reduce: FieldValue
			reduce(101), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(101), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(101), // ), reduce: FieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(101), // ?, reduce: FieldValue
			nil,         // :
			reduce(101), // ||, reduce: FieldValue
			reduce(101), // ^^, reduce: FieldValue
			reduce(101), // &&, reduce: FieldValue
			reduce(101), // ??, reduce: FieldValue
			reduce(101), // ???, reduce: FieldValue
			reduce(101), // =~, reduce: FieldValue
			reduce(101), // !=~, reduce: FieldValue
			reduce(101), // ==, reduce: FieldValue
			reduce(101), // !=, reduce: FieldValue
			reduce(101), // <=>, reduce: FieldValue
			reduce(101), // >=, reduce: FieldValue
			reduce(101), // <, reduce: FieldValue
			reduce(101), // <=, reduce: FieldValue
			reduce(101), // ^, reduce: FieldValue
			reduce(101), // &, reduce: FieldValue
			reduce(101), // <<, reduce: FieldValue
			reduce(101), // >>>, reduce: FieldValue
			reduce(101), // +, reduce: FieldValue
			reduce(101), // -, reduce: FieldValue
			reduce(101), // .+, reduce: FieldValue
			reduce(101), // .-, reduce: FieldValue
			reduce(101), // *, reduce: FieldValue
			reduce(101), // /, reduce: FieldValue
			reduce(101), // //, reduce: FieldValue
			reduce(101), // %, reduce: FieldValue
			reduce(101), // .*, reduce: FieldValue
			reduce(101), // ./, reduce: FieldValue
			reduce(101), // .//, reduce: FieldValue
			reduce(101), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(101), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(101), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(101), // [[, reduce: FieldValue
			reduce(101), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1675
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(100), // >, reduce: FieldValue
			reduce(100), // >>, reduce: FieldValue
			reduce(100), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(100), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(100), // ), reduce: FieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(100), // ?, reduce: FieldValue
			nil,         // :
			reduce(100), // ||, reduce: FieldValue
			reduce(100), // ^^, reduce: FieldValue
			reduce(100), // &&, reduce: FieldValue
			reduce(100), // ??, reduce: FieldValue
			reduce(100), // ???, reduce: FieldValue
			reduce(100), // =~, reduce: FieldValue
			reduce(100), // !=~, reduce: FieldValue
			reduce(100), // ==, reduce: FieldValue
			reduce(100), // !=, reduce: FieldValue
			reduce(100), // <=>, reduce: FieldValue
			reduce(100), // >=, reduce: FieldValue
			reduce(100), // <, reduce: FieldValue
			reduce(100), // <=, reduce: FieldValue
			reduce(100), // ^, reduce: FieldValue
			reduce(100), // &, reduce: FieldValue
			reduce(100), // <<, reduce: FieldValue
			reduce(100), // >>>, reduce: FieldValue
			reduce(100), // +, reduce: FieldValue
			reduce(100), // -, reduce: FieldValue
			reduce(100), // .+, reduce: FieldValue
			reduce(100), // .-, reduce: FieldValue
			reduce(100), // *, reduce: FieldValue
			reduce(100), // /, reduce: FieldValue
			reduce(100), // //, reduce: FieldValue
			reduce(100), // %, reduce: FieldValue
			reduce(100), // .*, reduce: FieldValue
			reduce(100), // ./, reduce: FieldValue
			reduce(100), // .//, reduce: FieldValue
			reduce(100), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(100), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(100), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(100), // [[, reduce: FieldValue
			reduce(100), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1676
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(213), // >, reduce: MlrvalOrFunction
			reduce(213), // >>, reduce: MlrvalOrFunction
			reduce(213), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(213), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(213), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(213), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(213), // ||, reduce: MlrvalOrFunction
			reduce(213), // ^^, reduce: MlrvalOrFunction
			reduce(213), // &&, reduce: MlrvalOrFunction
			reduce(213), // ??, reduce: MlrvalOrFunction
			reduce(213), // ???, reduce: MlrvalOrFunction
			reduce(213), // =~, reduce: MlrvalOrFunction
			reduce(213), // !=~, reduce: MlrvalOrFunction
			reduce(213), // ==, reduce: MlrvalOrFunction
			reduce(213), // !=, reduce: MlrvalOrFunction
			reduce(213), // <=>, reduce: MlrvalOrFunction
			reduce(213), // >=, reduce: MlrvalOrFunction
			reduce(213), // <, reduce: MlrvalOrFunction
			reduce(213), // <=, reduce: MlrvalOrFunction
			reduce(213), // ^, reduce: MlrvalOrFunction
			reduce(213), // &, reduce: MlrvalOrFunction
			reduce(213), // <<, reduce: MlrvalOrFunction
			reduce(213), // >>>, reduce: MlrvalOrFunction
			reduce(213), // +, reduce: MlrvalOrFunction
			reduce(213), // -, reduce: MlrvalOrFunction
			reduce(213), // .+, reduce: MlrvalOrFunction
			reduce(213), // .-, reduce: MlrvalOrFunction
			reduce(213), // *, reduce: MlrvalOrFunction
			reduce(213), // /, reduce: MlrvalOrFunction
			reduce(213), // //, reduce: MlrvalOrFunction
			reduce(213), // %, reduce: MlrvalOrFunction
			reduce(213), // .*, reduce: MlrvalOrFunction
			reduce(213), // ./, reduce: MlrvalOrFunction
			reduce(213), // .//, reduce: MlrvalOrFunction
			reduce(213), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(213), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(213), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(213), // [[, reduce: MlrvalOrFunction
			reduce(213), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1677
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(233), // >, reduce: MlrvalOrFunction
			reduce(233), // >>, reduce: MlrvalOrFunction
			reduce(233), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(233), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(233), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(233), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(233), // ||, reduce: MlrvalOrFunction
			reduce(233), // ^^, reduce: MlrvalOrFunction
			reduce(233), // &&, reduce: MlrvalOrFunction
			reduce(233), // ??, reduce: MlrvalOrFunction
			reduce(233), // ???, reduce: MlrvalOrFunction
			reduce(233), // =~, reduce: MlrvalOrFunction
			reduce(233), // !=~, reduce: MlrvalOrFunction
			reduce(233), // ==, reduce: MlrvalOrFunction
			reduce(233), // !=, reduce: MlrvalOrFunction
			reduce(233), // <=>, reduce: MlrvalOrFunction
			reduce(233), // >=, reduce: MlrvalOrFunction
			reduce(233), // <, reduce: MlrvalOrFunction
			reduce(233), // <=, reduce: MlrvalOrFunction
			reduce(233), // ^, reduce: MlrvalOrFunction
			reduce(233), // &, reduce: MlrvalOrFunction
			reduce(233), // <<, reduce: MlrvalOrFunction
			reduce(233), // >>>, reduce: MlrvalOrFunction
			reduce(233), // +, reduce: MlrvalOrFunction
			reduce(233), // -, reduce: MlrvalOrFunction
			reduce(233), // .+, reduce: MlrvalOrFunction
			reduce(233), // .-, reduce: MlrvalOrFunction
			reduce(233), // *, reduce: MlrvalOrFunction
			reduce(233), // /, reduce: MlrvalOrFunction
			reduce(233), // //, reduce: MlrvalOrFunction
			reduce(233), // %, reduce: MlrvalOrFunction
			reduce(233), // .*, reduce: MlrvalOrFunction
			reduce(233), // ./, reduce: MlrvalOrFunction
			reduce(233), // .//, reduce: MlrvalOrFunction
			reduce(233), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(233), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(233), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(233), // [[, reduce: MlrvalOrFunction
			reduce(233), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1678
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(210), // >, reduce: MlrvalOrFunction
			reduce(210), // >>, reduce: MlrvalOrFunction
			reduce(210), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(210), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(210), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(210), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(210), // ||, reduce: MlrvalOrFunction
			reduce(210), // ^^, reduce: MlrvalOrFunction
			reduce(210), // &&, reduce: MlrvalOrFunction
			reduce(210), // ??, reduce: MlrvalOrFunction
			reduce(210), // ???, reduce: MlrvalOrFunction
			reduce(210), // =~, reduce: MlrvalOrFunction
			reduce(210), // !=~, reduce: MlrvalOrFunction
			reduce(210), // ==, reduce: MlrvalOrFunction
			reduce(210), // !=, reduce: MlrvalOrFunction
			reduce(210), // <=>, reduce: MlrvalOrFunction
			reduce(210), // >=, reduce: MlrvalOrFunction
			reduce(210), // <, reduce: MlrvalOrFunction
			reduce(210), // <=, reduce: MlrvalOrFunction
			reduce(210), // ^, reduce: MlrvalOrFunction
			reduce(210), // &, reduce: MlrvalOrFunction
			reduce(210), // <<, reduce: MlrvalOrFunction
			reduce(210), // >>>, reduce: MlrvalOrFunction
			reduce(210), // +, reduce: MlrvalOrFunction
			reduce(210), // -, reduce: MlrvalOrFunction
			reduce(210), // .+, reduce: MlrvalOrFunction
			reduce(210), // .-, reduce: MlrvalOrFunction
			reduce(210), // *, reduce: MlrvalOrFunction
			reduce(210), // /, reduce: MlrvalOrFunction
			reduce(210), // //, reduce: MlrvalOrFunction
			reduce(210), // %, reduce: MlrvalOrFunction
			reduce(210), // .*, reduce: MlrvalOrFunction
			reduce(210), // ./, reduce: MlrvalOrFunction
			reduce(210), // .//, reduce: MlrvalOrFunction
			reduce(210), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(210), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(210), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(210), // [[, reduce: MlrvalOrFunction
			reduce(210), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1679
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(102), // >, reduce: FieldValue
			reduce(102), // >>, reduce: FieldValue
			reduce(102), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(102), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(102), // ), reduce: FieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(102), // ?, reduce: FieldValue
			nil,         // :
			reduce(102), // ||, reduce: FieldValue
			reduce(102), // ^^, reduce: FieldValue
			reduce(102), // &&, reduce: FieldValue
			reduce(102), // ??, reduce: FieldValue
			reduce(102), // ???, reduce: FieldValue
			reduce(102), // =~, reduce: FieldValue
			reduce(102), // !=~, reduce: FieldValue
			reduce(102), // ==, reduce: FieldValue
			reduce(102), // !=, reduce: FieldValue
			reduce(102), // <=>, reduce: FieldValue
			reduce(102), // >=, reduce: FieldValue
			reduce(102), // <, reduce: FieldValue
			reduce(102), // <=, reduce: FieldValue
			reduce(102), // ^, reduce: FieldValue
			reduce(102), // &, reduce: FieldValue
			reduce(102), // <<, reduce: FieldValue
			reduce(102), // >>>, reduce: FieldValue
			reduce(102), // +, reduce: FieldValue
			reduce(102), // -, reduce: FieldValue
			reduce(102), // .+, reduce: FieldValue
			reduce(102), // .-, reduce: FieldValue
			reduce(102), // *, reduce: FieldValue
			reduce(102), // /, reduce: FieldValue
			reduce(102), // //, reduce: FieldValue
			reduce(102), // %, reduce: FieldValue
			reduce(102), // .*, reduce: FieldValue
			reduce(102), // ./, reduce: FieldValue
			reduce(102), // .//, reduce: FieldValue
			reduce(102), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(102), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(102), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(102), // [[, reduce: FieldValue
			reduce(102), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1680
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(103), // >, reduce: FieldValue
			reduce(103), // >>, reduce: FieldValue
			reduce(103), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(103), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(103), // ), reduce: FieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(103), // ?, reduce: FieldValue
			nil,         // :
			reduce(103), // ||, reduce: FieldValue
			reduce(103), // ^^, reduce: FieldValue
			reduce(103), // &&, reduce: FieldValue
			reduce(103), // ??, reduce: FieldValue
			reduce(103), // ???, reduce: FieldValue
			reduce(103), // =~, reduce: FieldValue
			reduce(103), // !=~, reduce: FieldValue
			reduce(103), // ==, reduce: FieldValue
			reduce(103), // !=, reduce: FieldValue
			reduce(103), // <=>, reduce: FieldValue
			reduce(103), // >=, reduce: FieldValue
			reduce(103), // <, reduce: FieldValue
			reduce(103), // <=, reduce: FieldValue
			reduce(103), // ^, reduce: FieldValue
			reduce(103), // &, reduce: FieldValue
			reduce(103), // <<, reduce: FieldValue
			reduce(103), // >>>, reduce: FieldValue
			reduce(103), // +, reduce: FieldValue
			reduce(103), // -, reduce: FieldValue
			reduce(103), // .+, reduce: FieldValue
			reduce(103), // .-, reduce: FieldValue
			reduce(103), // *, reduce: FieldValue
			reduce(103), // /, reduce: FieldValue
			reduce(103), // //, reduce: FieldValue
			reduce(103), // %, reduce: FieldValue
			reduce(103), // .*, reduce: FieldValue
			reduce(103), // ./, reduce: FieldValue
			reduce(103), // .//, reduce: FieldValue
			reduce(103), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(103), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(103), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(103), // [[, reduce: FieldValue
			reduce(103), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1681
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(104), // >, reduce: DirectFieldValue
			reduce(104), // >>, reduce: DirectFieldValue
			reduce(104), // |, reduce: DirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(104), // ,, reduce: DirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(104), // ), reduce: DirectFieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(104), // ?, reduce: DirectFieldValue
			nil,         // :
			reduce(104), // ||, reduce: DirectFieldValue
			reduce(104), // ^^, reduce: DirectFieldValue
			reduce(104), // &&, reduce: DirectFieldValue
			reduce(104), // ??, reduce: DirectFieldValue
			reduce(104), // ???, reduce: DirectFieldValue
			reduce(104), // =~, reduce: DirectFieldValue
			reduce(104), // !=~, reduce: DirectFieldValue
			reduce(104), // ==, reduce: DirectFieldValue
			reduce(104), // !=, reduce: DirectFieldValue
			reduce(104), // <=>, reduce: DirectFieldValue
			reduce(104), // >=, reduce: DirectFieldValue
			reduce(104), // <, reduce: DirectFieldValue
			reduce(104), // <=, reduce: DirectFieldValue
			reduce(104), // ^, reduce: DirectFieldValue
			reduce(104), // &, reduce: DirectFieldValue
			reduce(104), // <<, reduce: DirectFieldValue
			reduce(104), // >>>, reduce: DirectFieldValue
			reduce(104), // +, reduce: DirectFieldValue
			reduce(104), // -, reduce: DirectFieldValue
			reduce(104), // .+, reduce: DirectFieldValue
			reduce(104), // .-, reduce: DirectFieldValue
			reduce(104), // *, reduce: DirectFieldValue
			reduce(104), // /, reduce: DirectFieldValue
			reduce(104), // //, reduce: DirectFieldValue
			reduce(104), // %, reduce: DirectFieldValue
			reduce(104), // .*, reduce: DirectFieldValue
			reduce(104), // ./, reduce: DirectFieldValue
			reduce(104), // .//, reduce: DirectFieldValue
			reduce(104), // ., reduce: DirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(104), // **, reduce: DirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(104), // [, reduce: DirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(104), // [[, reduce: DirectFieldValue
			reduce(104), // [[[, reduce: DirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1682
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1683
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(106), // >, reduce: BracedFieldValue
			reduce(106), // >>, reduce: BracedFieldValue
			reduce(106), // |, reduce: BracedFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(106), // ,, reduce: BracedFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(106), // ), reduce: BracedFieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(106), // ?, reduce: BracedFieldValue
			nil,         // :
			reduce(106), // ||, reduce: BracedFieldValue
			reduce(106), // ^^, reduce: BracedFieldValue
			reduce(106), // &&, reduce: BracedFieldValue
			reduce(106), // ??, reduce: BracedFieldValue
			reduce(106), // ???, reduce: BracedFieldValue
			reduce(106), // =~, reduce: BracedFieldValue
			reduce(106), // !=~, reduce: BracedFieldValue
			reduce(106), // ==, reduce: BracedFieldValue
			reduce(106), // !=, reduce: BracedFieldValue
			reduce(106), // <=>, reduce: BracedFieldValue
			reduce(106), // >=, reduce: BracedFieldValue
			reduce(106), // <, reduce: BracedFieldValue
			reduce(106), // <=, reduce: BracedFieldValue
			reduce(106), // ^, reduce: BracedFieldValue
			reduce(106), // &, reduce: BracedFieldValue
			reduce(106), // <<, reduce: BracedFieldValue
			reduce(106), // >>>, reduce: BracedFieldValue
			reduce(106), // +, reduce: BracedFieldValue
			reduce(106), // -, reduce: BracedFieldValue
			reduce(106), // .+, reduce: BracedFieldValue
			reduce(106), // .-, reduce: BracedFieldValue
			reduce(106), // *, reduce: BracedFieldValue
			reduce(106), // /, reduce: BracedFieldValue
			reduce(106), // //, reduce: BracedFieldValue
			reduce(106), // %, reduce: BracedFieldValue
			reduce(106), // .*, reduce: BracedFieldValue
			reduce(106), // ./, reduce: BracedFieldValue
			reduce(106), // .//, reduce: BracedFieldValue
			reduce(106), // ., reduce: BracedFieldValue
			nil,         // !
			nil,         // ~
			reduce(106), // **, reduce: BracedFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(106), // [, reduce: BracedFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(106), // [[, reduce: BracedFieldValue
			reduce(106), // [[[, reduce: BracedFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1684
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1685
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1686
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(109), // >, reduce: FullSrec
			reduce(109), // >>, reduce: FullSrec
			reduce(109), // |, reduce: FullSrec
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(109), // ,, reduce: FullSrec
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(109), // ), reduce: FullSrec
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(109), // ?, reduce: FullSrec
			nil,         // :
			reduce(109), // ||, reduce: FullSrec
			reduce(109), // ^^, reduce: FullSrec
			reduce(109), // &&, reduce: FullSrec
			reduce(109), // ??, reduce: FullSrec
			reduce(109), // ???, reduce: FullSrec
			reduce(109), // =~, reduce: FullSrec
			reduce(109), // !=~, reduce: FullSrec
			reduce(109), // ==, reduce: FullSrec
			reduce(109), // !=, reduce: FullSrec
			reduce(109), // <=>, reduce: FullSrec
			reduce(109), // >=, reduce: FullSrec
			reduce(109), // <, reduce: FullSrec
			reduce(109), // <=, reduce: FullSrec
			reduce(109), // ^, reduce: FullSrec
			reduce(109), // &, reduce: FullSrec
			reduce(109), // <<, reduce: FullSrec
			reduce(109), // >>>, reduce: FullSrec
			reduce(109), // +, reduce: FullSrec
			reduce(109), // -, reduce: FullSrec
			reduce(109), // .+, reduce: FullSrec
			reduce(109), // .-, reduce: FullSrec
			reduce(109), // *, reduce: FullSrec
			reduce(109), // /, reduce: FullSrec
			reduce(109), // //, reduce: FullSrec
			reduce(109), // %, reduce: FullSrec
			reduce(109), // .*, reduce: FullSrec
			reduce(109), // ./, reduce: FullSrec
			reduce(109), // .//, reduce: FullSrec
			reduce(109), // ., reduce: FullSrec
			nil,         // !
			nil,         // ~
			reduce(109), // **, reduce: FullSrec
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(109), // [, reduce: FullSrec
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(109), // [[, reduce: FullSrec
			reduce(109), // [[[, reduce: FullSrec
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1687
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(212), // >, reduce: MlrvalOrFunction
			reduce(212), // >>, reduce: MlrvalOrFunction
			reduce(212), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(212), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(212), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(212), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(212), // ||, reduce: MlrvalOrFunction
			reduce(212), // ^^, reduce: MlrvalOrFunction
			reduce(212), // &&, reduce: MlrvalOrFunction
			reduce(212), // ??, reduce: MlrvalOrFunction
			reduce(212), // ???, reduce: MlrvalOrFunction
			reduce(212), // =~, reduce: MlrvalOrFunction
			reduce(212), // !=~, reduce: MlrvalOrFunction
			reduce(212), // ==, reduce: MlrvalOrFunction
			reduce(212), // !=, reduce: MlrvalOrFunction
			reduce(212), // <=>, reduce: MlrvalOrFunction
			reduce(212), // >=, reduce: MlrvalOrFunction
			reduce(212), // <, reduce: MlrvalOrFunction
			reduce(212), // <=, reduce: MlrvalOrFunction
			reduce(212), // ^, reduce: MlrvalOrFunction
			reduce(212), // &, reduce: MlrvalOrFunction
			reduce(212), // <<, reduce: MlrvalOrFunction
			reduce(212), // >>>, reduce: MlrvalOrFunction
			reduce(212), // +, reduce: MlrvalOrFunction
			reduce(212), // -, reduce: MlrvalOrFunction
			reduce(212), // .+, reduce: MlrvalOrFunction
			reduce(212), // .-, reduce: MlrvalOrFunction
			reduce(212), // *, reduce: MlrvalOrFunction
			reduce(212), // /, reduce: MlrvalOrFunction
			reduce(212), // //, reduce: MlrvalOrFunction
			reduce(212), // %, reduce: MlrvalOrFunction
			reduce(212), // .*, reduce: MlrvalOrFunction
			reduce(212), // ./, reduce: MlrvalOrFunction
			reduce(212), // .//, reduce: MlrvalOrFunction
			reduce(212), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(212), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(212), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(212), // [[, reduce: MlrvalOrFunction
			reduce(212), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1688
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(113), // >, reduce: DirectOosvarValue
			reduce(113), // >>, reduce: DirectOosvarValue
			reduce(113), // |, reduce: DirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(113), // ,, reduce: DirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(113), // ), reduce: DirectOosvarValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(113), // ?, reduce: DirectOosvarValue
			nil,         // :
			reduce(113), // ||, reduce: DirectOosvarValue
			reduce(113), // ^^, reduce: DirectOosvarValue
			reduce(113), // &&, reduce: DirectOosvarValue
			reduce(113), // ??, reduce: DirectOosvarValue
			reduce(113), // ???, reduce: DirectOosvarValue
			reduce(113), // =~, reduce: DirectOosvarValue
			reduce(113), // !=~, reduce: DirectOosvarValue
			reduce(113), // ==, reduce: DirectOosvarValue
			reduce(113), // !=, reduce: DirectOosvarValue
			reduce(113), // <=>, reduce: DirectOosvarValue
			reduce(113), // >=, reduce: DirectOosvarValue
			reduce(113), // <, reduce: DirectOosvarValue
			reduce(113), // <=, reduce: DirectOosvarValue
			reduce(113), // ^, reduce: DirectOosvarValue
			reduce(113), // &, reduce: DirectOosvarValue
			reduce(113), // <<, reduce: DirectOosvarValue
			reduce(113), // >>>, reduce: DirectOosvarValue
			reduce(113), // +, reduce: DirectOosvarValue
			reduce(113), // -, reduce: DirectOosvarValue
			reduce(113), // .+, reduce: DirectOosvarValue
			reduce(113), // .-, reduce: DirectOosvarValue
			reduce(113), // *, reduce: DirectOosvarValue
			reduce(113), // /, reduce: DirectOosvarValue
			reduce(113), // //, reduce: DirectOosvarValue
			reduce(113), // %, reduce: DirectOosvarValue
			reduce(113), // .*, reduce: DirectOosvarValue
			reduce(113), // ./, reduce: DirectOosvarValue
			reduce(113), // .//, reduce: DirectOosvarValue
			reduce(113), // ., reduce: DirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(113), // **, reduce: DirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(113), // [, reduce: DirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(113), // [[, reduce: DirectOosvarValue
			reduce(113), // [[[, reduce: DirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1689
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1690
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(115), // >, reduce: BracedOosvarValue
			reduce(115), // >>, reduce: BracedOosvarValue
			reduce(115), // |, reduce: BracedOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(115), // ,, reduce: BracedOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(115), // ), reduce: BracedOosvarValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(115), // ?, reduce: BracedOosvarValue
			nil,         // :
			reduce(115), // ||, reduce: BracedOosvarValue
			reduce(115), // ^^, reduce: BracedOosvarValue
			reduce(115), // &&, reduce: BracedOosvarValue
			reduce(115), // ??, reduce: BracedOosvarValue
			reduce(115), // ???, reduce: BracedOosvarValue
			reduce(115), // =~, reduce: BracedOosvarValue
			reduce(115), // !=~, reduce: BracedOosvarValue
			reduce(115), // ==, reduce: BracedOosvarValue
			reduce(115), // !=, reduce: BracedOosvarValue
			reduce(115), // <=>, reduce: BracedOosvarValue
			reduce(115), // >=, reduce: BracedOosvarValue
			reduce(115), // <, reduce: BracedOosvarValue
			reduce(115), // <=, reduce: BracedOosvarValue
			reduce(115), // ^, reduce: BracedOosvarValue
			reduce(115), // &, reduce: BracedOosvarValue
			reduce(115), // <<, reduce: BracedOosvarValue
			reduce(115), // >>>, reduce: BracedOosvarValue
			reduce(115), // +, reduce: BracedOosvarValue
			reduce(115), // -, reduce: BracedOosvarValue
			reduce(115), // .+, reduce: BracedOosvarValue
			reduce(115), // .-, reduce: BracedOosvarValue
			reduce(115), // *, reduce: BracedOosvarValue
			reduce(115), // /, reduce: BracedOosvarValue
			reduce(115), // //, reduce: BracedOosvarValue
			reduce(115), // %, reduce: BracedOosvarValue
			reduce(115), // .*, reduce: BracedOosvarValue
			reduce(115), // ./, reduce: BracedOosvarValue
			reduce(115), // .//, reduce: BracedOosvarValue
			reduce(115), // ., reduce: BracedOosvarValue
			nil,         // !
			nil,         // ~
			reduce(115), // **, reduce: BracedOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(115), // [, reduce: BracedOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(115), // [[, reduce: BracedOosvarValue
			reduce(115), // [[[, reduce: BracedOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1691
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(116), // >, reduce: FullOosvar
			reduce(116), // >>, reduce: FullOosvar
			reduce(116), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(116), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(116), // ), reduce: FullOosvar
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(116), // ?, reduce: FullOosvar
			nil,         // :
			reduce(116), // ||, reduce: FullOosvar
			reduce(116), // ^^, reduce: FullOosvar
			reduce(116), // &&, reduce: FullOosvar
			reduce(116), // ??, reduce: FullOosvar
			reduce(116), // ???, reduce: FullOosvar
			reduce(116), // =~, reduce: FullOosvar
			reduce(116), // !=~, reduce: FullOosvar
			reduce(116), // ==, reduce: FullOosvar
			reduce(116), // !=, reduce: FullOosvar
			reduce(116), // <=>, reduce: FullOosvar
			reduce(116), // >=, reduce: FullOosvar
			reduce(116), // <, reduce: FullOosvar
			reduce(116), // <=, reduce: FullOosvar
			reduce(116), // ^, reduce: FullOosvar
			reduce(116), // &, reduce: FullOosvar
			reduce(116), // <<, reduce: FullOosvar
			reduce(116), // >>>, reduce: FullOosvar
			reduce(116), // +, reduce: FullOosvar
			reduce(116), // -, reduce: FullOosvar
			reduce(116), // .+, reduce: FullOosvar
			reduce(116), // .-, reduce: FullOosvar
			reduce(116), // *, reduce: FullOosvar
			reduce(116), // /, reduce: FullOosvar
			reduce(116), // //, reduce: FullOosvar
			reduce(116), // %, reduce: FullOosvar
			reduce(116), // .*, reduce: FullOosvar
			reduce(116), // ./, reduce: FullOosvar
			reduce(116), // .//, reduce: FullOosvar
			reduce(116), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(116), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(116), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(116), // [[, reduce: FullOosvar
			reduce(116), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1692
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(117), // >, reduce: FullOosvar
			reduce(117), // >>, reduce: FullOosvar
			reduce(117), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(117), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(117), // ), reduce: FullOosvar
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(117), // ?, reduce: FullOosvar
			nil,         // :
			reduce(117), // ||, reduce: FullOosvar
			reduce(117), // ^^, reduce: FullOosvar
			reduce(117), // &&, reduce: FullOosvar
			reduce(117), // ??, reduce: FullOosvar
			reduce(117), // ???, reduce: FullOosvar
			reduce(117), // =~, reduce: FullOosvar
			reduce(117), // !=~, reduce: FullOosvar
			reduce(117), // ==, reduce: FullOosvar
			reduce(117), // !=, reduce: FullOosvar
			reduce(117), // <=>, reduce: FullOosvar
			reduce(117), // >=, reduce: FullOosvar
			reduce(117), // <, reduce: FullOosvar
			reduce(117), // <=, reduce: FullOosvar
			reduce(117), // ^, reduce: FullOosvar
			reduce(117), // &, reduce: FullOosvar
			reduce(117), // <<, reduce: FullOosvar
			reduce(117), // >>>, reduce: FullOosvar
			reduce(117), // +, reduce: FullOosvar
			reduce(117), // -, reduce: FullOosvar
			reduce(117), // .+, reduce: FullOosvar
			reduce(117), // .-, reduce: FullOosvar
			reduce(117), // *, reduce: FullOosvar
			reduce(117), // /, reduce: FullOosvar
			reduce(117), // //, reduce: FullOosvar
			reduce(117), // %, reduce: FullOosvar
			reduce(117), // .*, reduce: FullOosvar
			reduce(117), // ./, reduce: FullOosvar
			reduce(117), // .//, reduce: FullOosvar
			reduce(117), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(117), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(117), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(117), // [[, reduce: FullOosvar
			reduce(117), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1693
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(118), // >, reduce: LocalVariable
			reduce(118), // >>, reduce: LocalVariable
			reduce(118), // |, reduce: LocalVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(118), // ,, reduce: LocalVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(270), // (, reduce: FunctionName
			reduce(118), // ), reduce: LocalVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(118), // ?, reduce: LocalVariable
			nil,         // :
			reduce(118), // ||, reduce: LocalVariable
			reduce(118), // ^^, reduce: LocalVariable
			reduce(118), // &&, reduce: LocalVariable
			reduce(118), // ??, reduce: LocalVariable
			reduce(118), // ???, reduce: LocalVariable
			reduce(118), // =~, reduce: LocalVariable
			reduce(118), // !=~, reduce: LocalVariable
			reduce(118), // ==, reduce: LocalVariable
			reduce(118), // !=, reduce: LocalVariable
			reduce(118), // <=>, reduce: LocalVariable
			reduce(118), // >=, reduce: LocalVariable
			reduce(118), // <, reduce: LocalVariable
			reduce(118), // <=, reduce: LocalVariable
			reduce(118), // ^, reduce: LocalVariable
			reduce(118), // &, reduce: LocalVariable
			reduce(118), // <<, reduce: LocalVariable
			reduce(118), // >>>, reduce: LocalVariable
			reduce(118), // +, reduce: LocalVariable
			reduce(118), // -, reduce: LocalVariable
			reduce(118), // .+, reduce: LocalVariable
			reduce(118), // .-, reduce: LocalVariable
			reduce(118), // *, reduce: LocalVariable
			reduce(118), // /, reduce: LocalVariable
			reduce(118), // //, reduce: LocalVariable
			reduce(118), // %, reduce: LocalVariable
			reduce(118), // .*, reduce: LocalVariable
			reduce(118), // ./, reduce: LocalVariable
			reduce(118), // .//, reduce: LocalVariable
			reduce(118), // ., reduce: LocalVariable
			nil,         // !
			nil,         // ~
			reduce(118), // **, reduce: LocalVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(118), // [, reduce: LocalVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(118), // [[, reduce: LocalVariable
			reduce(118), // [[[, reduce: LocalVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1694
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(147), // ,, reduce: Rvalue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(147), // ), reduce: Rvalue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1695
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(148), // ,, reduce: PrecedenceChainStart
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(148), // ), reduce: PrecedenceChainStart
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1696
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(150), // ,, reduce: TernaryTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(150), // ), reduce: TernaryTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			shift(2759), // ?
			nil,         // :
			shift(2760), // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1697
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(152), // ,, reduce: LogicalOrTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(152), // ), reduce: LogicalOrTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(152), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(152), // ||, reduce: LogicalOrTerm
			shift(2761), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1698
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(154), // ,, reduce: LogicalXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(154), // ), reduce: LogicalXORTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(154), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(154), // ||, reduce: LogicalXORTerm
			reduce(154), // ^^, reduce: LogicalXORTerm
			shift(2762), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1699
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(156), // ,, reduce: LogicalAndTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(156), // ), reduce: LogicalAndTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(156), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(156), // ||, reduce: LogicalAndTerm
			reduce(156), // ^^, reduce: LogicalAndTerm
			reduce(156), // &&, reduce: LogicalAndTerm
			shift(2763), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1700
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(158), // ,, reduce: AbsentCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(158), // ), reduce: AbsentCoalesceTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(158), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(158), // ||, reduce: AbsentCoalesceTerm
			reduce(158), // ^^, reduce: AbsentCoalesceTerm
			reduce(158), // &&, reduce: AbsentCoalesceTerm
			reduce(158), // ??, reduce: AbsentCoalesceTerm
			shift(2764), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1701
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(160), // ,, reduce: EmptyCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(160), // ), reduce: EmptyCoalesceTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(160), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(160), // ||, reduce: EmptyCoalesceTerm
			reduce(160), // ^^, reduce: EmptyCoalesceTerm
			reduce(160), // &&, reduce: EmptyCoalesceTerm
			reduce(160), // ??, reduce: EmptyCoalesceTerm
			reduce(160), // ???, reduce: EmptyCoalesceTerm
			shift(2765), // =~
			shift(2766), // !=~
			shift(2767), // ==
			shift(2768), // !=
			shift(2769), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1702
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2770), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(166), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(166), // ), reduce: EqneTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(166), // ?, reduce: EqneTerm
			nil,         // :
			reduce(166), // ||, reduce: EqneTerm
			reduce(166), // ^^, reduce: EqneTerm
			reduce(166), // &&, reduce: EqneTerm
			reduce(166), // ??, reduce: EqneTerm
			reduce(166), // ???, reduce: EqneTerm
			reduce(166), // =~, reduce: EqneTerm
			reduce(166), // !=~, reduce: EqneTerm
			reduce(166), // ==, reduce: EqneTerm
			reduce(166), // !=, reduce: EqneTerm
			reduce(166), // <=>, reduce: EqneTerm
			shift(2771), // >=
			shift(2772), // <
			shift(2773), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1703
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(171), // >, reduce: CmpTerm
			nil,         // >>
			shift(2774), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(171), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(171), // ), reduce: CmpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(171), // ?, reduce: CmpTerm
			nil,         // :
			reduce(171), // ||, reduce: CmpTerm
			reduce(171), // ^^, reduce: CmpTerm
			reduce(171), // &&, reduce: CmpTerm
			reduce(171), // ??, reduce: CmpTerm
			reduce(171), // ???, reduce: CmpTerm
			reduce(171), // =~, reduce: CmpTerm
			reduce(171), // !=~, reduce: CmpTerm
			reduce(171), // ==, reduce: CmpTerm
			reduce(171), // !=, reduce: CmpTerm
			reduce(171), // <=>, reduce: CmpTerm
			reduce(171), // >=, reduce: CmpTerm
			reduce(171), // <, reduce: CmpTerm
			reduce(171), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1704
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(173), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(173), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(173), // ,, reduce: BitwiseORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(173), // ), reduce: BitwiseORTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(173), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(173), // ||, reduce: BitwiseORTerm
			reduce(173), // ^^, reduce: BitwiseORTerm
			reduce(173), // &&, reduce: BitwiseORTerm
			reduce(173), // ??, reduce: BitwiseORTerm
			reduce(173), // ???, reduce: BitwiseORTerm
			reduce(173), // =~, reduce: BitwiseORTerm
			reduce(173), // !=~, reduce: BitwiseORTerm
			reduce(173), // ==, reduce: BitwiseORTerm
			reduce(173), // !=, reduce: BitwiseORTerm
			reduce(173), // <=>, reduce: BitwiseORTerm
			reduce(173), // >=, reduce: BitwiseORTerm
			reduce(173), // <, reduce: BitwiseORTerm
			reduce(173), // <=, reduce: BitwiseORTerm
			shift(2775), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1705
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(175), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(175), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(175), // ,, reduce: BitwiseXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(175), // ), reduce: BitwiseXORTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(175), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(175), // ||, reduce: BitwiseXORTerm
			reduce(175), // ^^, reduce: BitwiseXORTerm
			reduce(175), // &&, reduce: BitwiseXORTerm
			reduce(175), // ??, reduce: BitwiseXORTerm
			reduce(175), // ???, reduce: BitwiseXORTerm
			reduce(175), // =~, reduce: BitwiseXORTerm
			reduce(175), // !=~, reduce: BitwiseXORTerm
			reduce(175), // ==, reduce: BitwiseXORTerm
			reduce(175), // !=, reduce: BitwiseXORTerm
			reduce(175), // <=>, reduce: BitwiseXORTerm
			reduce(175), // >=, reduce: BitwiseXORTerm
			reduce(175), // <, reduce: BitwiseXORTerm
			reduce(175), // <=, reduce: BitwiseXORTerm
			reduce(175), // ^, reduce: BitwiseXORTerm
			shift(2776), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1706
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(177), // >, reduce: BitwiseANDTerm
			shift(2777), // >>
			reduce(177), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(177), // ,, reduce: BitwiseANDTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(177), // ), reduce: BitwiseANDTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(177), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(177), // ||, reduce: BitwiseANDTerm
			reduce(177), // ^^, reduce: BitwiseANDTerm
			reduce(177), // &&, reduce: BitwiseANDTerm
			reduce(177), // ??, reduce: BitwiseANDTerm
			reduce(177), // ???, reduce: BitwiseANDTerm
			reduce(177), // =~, reduce: BitwiseANDTerm
			reduce(177), // !=~, reduce: BitwiseANDTerm
			reduce(177), // ==, reduce: BitwiseANDTerm
			reduce(177), // !=, reduce: BitwiseANDTerm
			reduce(177), // <=>, reduce: BitwiseANDTerm
			reduce(177), // >=, reduce: BitwiseANDTerm
			reduce(177), // <, reduce: BitwiseANDTerm
			reduce(177), // <=, reduce: BitwiseANDTerm
			reduce(177), // ^, reduce: BitwiseANDTerm
			reduce(177), // &, reduce: BitwiseANDTerm
			shift(2778), // <<
			shift(2779), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1707
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(181), // >, reduce: BitwiseShiftTerm
			reduce(181), // >>, reduce: BitwiseShiftTerm
			reduce(181), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(181), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(181), // ), reduce: BitwiseShiftTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(181), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(181), // ||, reduce: BitwiseShiftTerm
			reduce(181), // ^^, reduce: BitwiseShiftTerm
			reduce(181), // &&, reduce: BitwiseShiftTerm
			reduce(181), // ??, reduce: BitwiseShiftTerm
			reduce(181), // ???, reduce: BitwiseShiftTerm
			reduce(181), // =~, reduce: BitwiseShiftTerm
			reduce(181), // !=~, reduce: BitwiseShiftTerm
			reduce(181), // ==, reduce: BitwiseShiftTerm
			reduce(181), // !=, reduce: BitwiseShiftTerm
			reduce(181), // <=>, reduce: BitwiseShiftTerm
			reduce(181), // >=, reduce: BitwiseShiftTerm
			reduce(181), // <, reduce: BitwiseShiftTerm
			reduce(181), // <=, reduce: BitwiseShiftTerm
			reduce(181), // ^, reduce: BitwiseShiftTerm
			reduce(181), // &, reduce: BitwiseShiftTerm
			reduce(181), // <<, reduce: BitwiseShiftTerm
			reduce(181), // >>>, reduce: BitwiseShiftTerm
			shift(2780), // +
			shift(2781), // -
			shift(2782), // .+
			shift(2783), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1708
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1709
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(186), // >, reduce: AddsubdotTerm
			reduce(186), // >>, reduce: AddsubdotTerm
			reduce(186), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(186), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(186), // ), reduce: AddsubdotTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(186), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(186), // ||, reduce: AddsubdotTerm
			reduce(186), // ^^, reduce: AddsubdotTerm
			reduce(186), // &&, reduce: AddsubdotTerm
			reduce(186), // ??, reduce: AddsubdotTerm
			reduce(186), // ???, reduce: AddsubdotTerm
			reduce(186), // =~, reduce: AddsubdotTerm
			reduce(186), // !=~, reduce: AddsubdotTerm
			reduce(186), // ==, reduce: AddsubdotTerm
			reduce(186), // !=, reduce: AddsubdotTerm
			reduce(186), // <=>, reduce: AddsubdotTerm
			reduce(186), // >=, reduce: AddsubdotTerm
			reduce(186), // <, reduce: AddsubdotTerm
			reduce(186), // <=, reduce: AddsubdotTerm
			reduce(186), // ^, reduce: AddsubdotTerm
			reduce(186), // &, reduce: AddsubdotTerm
			reduce(186), // <<, reduce: AddsubdotTerm
			reduce(186), // >>>, reduce: AddsubdotTerm
			reduce(186), // +, reduce: AddsubdotTerm
			reduce(186), // -, reduce: AddsubdotTerm
			reduce(186), // .+, reduce: AddsubdotTerm
			reduce(186), // .-, reduce: AddsubdotTerm
			shift(2785), // *
			shift(2786), // /
			shift(2787), // //
			shift(2788), // %
			shift(2789), // .*
			shift(2790), // ./
			shift(2791), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1710
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1711
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1712
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1713
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(194), // >, reduce: MuldivTerm
			reduce(194), // >>, reduce: MuldivTerm
			reduce(194), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(194), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(194), // ), reduce: MuldivTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(194), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(194), // ||, reduce: MuldivTerm
			reduce(194), // ^^, reduce: MuldivTerm
			reduce(194), // &&, reduce: MuldivTerm
			reduce(194), // ??, reduce: MuldivTerm
			reduce(194), // ???, reduce: MuldivTerm
			reduce(194), // =~, reduce: MuldivTerm
			reduce(194), // !=~, reduce: MuldivTerm
			reduce(194), // ==, reduce: MuldivTerm
			reduce(194), // !=, reduce: MuldivTerm
			reduce(194), // <=>, reduce: MuldivTerm
			reduce(194), // >=, reduce: MuldivTerm
			reduce(194), // <, reduce: MuldivTerm
			reduce(194), // <=, reduce: MuldivTerm
			reduce(194), // ^, reduce: MuldivTerm
			reduce(194), // &, reduce: MuldivTerm
			reduce(194), // <<, reduce: MuldivTerm
			reduce(194), // >>>, reduce: MuldivTerm
			reduce(194), // +, reduce: MuldivTerm
			reduce(194), // -, reduce: MuldivTerm
			reduce(194), // .+, reduce: MuldivTerm
			reduce(194), // .-, reduce: MuldivTerm
			reduce(194), // *, reduce: MuldivTerm
			reduce(194), // /, reduce: MuldivTerm
			reduce(194), // //, reduce: MuldivTerm
			reduce(194), // %, reduce: MuldivTerm
			reduce(194), // .*, reduce: MuldivTerm
			reduce(194), // ./, reduce: MuldivTerm
			reduce(194), // .//, reduce: MuldivTerm
			shift(2795), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1714
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(196), // >, reduce: DotTerm
			reduce(196), // >>, reduce: DotTerm
			reduce(196), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(196), // ,, reduce: DotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(196), // ), reduce: DotTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(196), // ?, reduce: DotTerm
			nil,         // :
			reduce(196), // ||, reduce: DotTerm
			reduce(196), // ^^, reduce: DotTerm
			reduce(196), // &&, reduce: DotTerm
			reduce(196), // ??, reduce: DotTerm
			reduce(196), // ???, reduce: DotTerm
			reduce(196), // =~, reduce: DotTerm
			reduce(196), // !=~, reduce: DotTerm
			reduce(196), // ==, reduce: DotTerm
			reduce(196), // !=, reduce: DotTerm
			reduce(196), // <=>, reduce: DotTerm
			reduce(196), // >=, reduce: DotTerm
			reduce(196), // <, reduce: DotTerm
			reduce(196), // <=, reduce: DotTerm
			reduce(196), // ^, reduce: DotTerm
			reduce(196), // &, reduce: DotTerm
			reduce(196), // <<, reduce: DotTerm
			reduce(196), // >>>, reduce: DotTerm
			reduce(196), // +, reduce: DotTerm
			reduce(196), // -, reduce: DotTerm
			reduce(196), // .+, reduce: DotTerm
			reduce(196), // .-, reduce: DotTerm
			reduce(196), // *, reduce: DotTerm
			reduce(196), // /, reduce: DotTerm
			reduce(196), // //, reduce: DotTerm
			reduce(196), // %, reduce: DotTerm
			reduce(196), // .*, reduce: DotTerm
			reduce(196), // ./, reduce: DotTerm
			reduce(196), // .//, reduce: DotTerm
			reduce(196), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1715
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1716
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1717
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(203), // >, reduce: UnaryOpTerm
			reduce(203), // >>, reduce: UnaryOpTerm
			reduce(203), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(203), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(203), // ), reduce: UnaryOpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(203), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(203), // ||, reduce: UnaryOpTerm
			reduce(203), // ^^, reduce: UnaryOpTerm
			reduce(203), // &&, reduce: UnaryOpTerm
			reduce(203), // ??, reduce: UnaryOpTerm
			reduce(203), // ???, reduce: UnaryOpTerm
			reduce(203), // =~, reduce: UnaryOpTerm
			reduce(203), // !=~, reduce: UnaryOpTerm
			reduce(203), // ==, reduce: UnaryOpTerm
			reduce(203), // !=, reduce: UnaryOpTerm
			reduce(203), // <=>, reduce: UnaryOpTerm
			reduce(203), // >=, reduce: UnaryOpTerm
			reduce(203), // <, reduce: UnaryOpTerm
			reduce(203), // <=, reduce: UnaryOpTerm
			reduce(203), // ^, reduce: UnaryOpTerm
			reduce(203), // &, reduce: UnaryOpTerm
			reduce(203), // <<, reduce: UnaryOpTerm
			reduce(203), // >>>, reduce: UnaryOpTerm
			reduce(203), // +, reduce: UnaryOpTerm
			reduce(203), // -, reduce: UnaryOpTerm
			reduce(203), // .+, reduce: UnaryOpTerm
			reduce(203), // .-, reduce: UnaryOpTerm
			reduce(203), // *, reduce: UnaryOpTerm
			reduce(203), // /, reduce: UnaryOpTerm
			reduce(203), // //, reduce: UnaryOpTerm
			reduce(203), // %, reduce: UnaryOpTerm
			reduce(203), // .*, reduce: UnaryOpTerm
			reduce(203), // ./, reduce: UnaryOpTerm
			reduce(203), // .//, reduce: UnaryOpTerm
			reduce(203), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1718
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(207), // >, reduce: PowTerm
			reduce(207), // >>, reduce: PowTerm
			reduce(207), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(207), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(207), // ), reduce: PowTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(207), // ?, reduce: PowTerm
			nil,         // :
			reduce(207), // ||, reduce: PowTerm
			reduce(207), // ^^, reduce: PowTerm
			reduce(207), // &&, reduce: PowTerm
			reduce(207), // ??, reduce: PowTerm
			reduce(207), // ???, reduce: PowTerm
			reduce(207), // =~, reduce: PowTerm
			reduce(207), // !=~, reduce: PowTerm
			reduce(207), // ==, reduce: PowTerm
			reduce(207), // !=, reduce: PowTerm
			reduce(207), // <=>, reduce: PowTerm
			reduce(207), // >=, reduce: PowTerm
			reduce(207), // <, reduce: PowTerm
			reduce(207), // <=, reduce: PowTerm
			reduce(207), // ^, reduce: PowTerm
			reduce(207), // &, reduce: PowTerm
			reduce(207), // <<, reduce: PowTerm
			reduce(207), // >>>, reduce: PowTerm
			reduce(207), // +, reduce: PowTerm
			reduce(207), // -, reduce: PowTerm
			reduce(207), // .+, reduce: PowTerm
			reduce(207), // .-, reduce: PowTerm
			reduce(207), // *, reduce: PowTerm
			reduce(207), // /, reduce: PowTerm
			reduce(207), // //, reduce: PowTerm
			reduce(207), // %, reduce: PowTerm
			reduce(207), // .*, reduce: PowTerm
			reduce(207), // ./, reduce: PowTerm
			reduce(207), // .//, reduce: PowTerm
			reduce(207), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			shift(2798), // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1719
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(209), // >, reduce: PrecedenceChainEnd
			reduce(209), // >>, reduce: PrecedenceChainEnd
			reduce(209), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(209), // ,, reduce: PrecedenceChainEnd
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(209), // ), reduce: PrecedenceChainEnd
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(209), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(209), // ||, reduce: PrecedenceChainEnd
			reduce(209), // ^^, reduce: PrecedenceChainEnd
			reduce(209), // &&, reduce: PrecedenceChainEnd
			reduce(209), // ??, reduce: PrecedenceChainEnd
			reduce(209), // ???, reduce: PrecedenceChainEnd
			reduce(209), // =~, reduce: PrecedenceChainEnd
			reduce(209), // !=~, reduce: PrecedenceChainEnd
			reduce(209), // ==, reduce: PrecedenceChainEnd
			reduce(209), // !=, reduce: PrecedenceChainEnd
			reduce(209), // <=>, reduce: PrecedenceChainEnd
			reduce(209), // >=, reduce: PrecedenceChainEnd
			reduce(209), // <, reduce: PrecedenceChainEnd
			reduce(209), // <=, reduce: PrecedenceChainEnd
			reduce(209), // ^, reduce: PrecedenceChainEnd
			reduce(209), // &, reduce: PrecedenceChainEnd
			reduce(209), // <<, reduce: PrecedenceChainEnd
			reduce(209), // >>>, reduce: PrecedenceChainEnd
			reduce(209), // +, reduce: PrecedenceChainEnd
			reduce(209), // -, reduce: PrecedenceChainEnd
			reduce(209), // .+, reduce: PrecedenceChainEnd
			reduce(209), // .-, reduce: PrecedenceChainEnd
			reduce(209), // *, reduce: PrecedenceChainEnd
			reduce(209), // /, reduce: PrecedenceChainEnd
			reduce(209), // //, reduce: PrecedenceChainEnd
			reduce(209), // %, reduce: PrecedenceChainEnd
			reduce(209), // .*, reduce: PrecedenceChainEnd
			reduce(209), // ./, reduce: PrecedenceChainEnd
			reduce(209), // .//, reduce: PrecedenceChainEnd
			reduce(209), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(209), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(2799), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			shift(2800), // [[
			shift(2801), // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1720
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(215), // >, reduce: MlrvalOrFunction
			reduce(215), // >>, reduce: MlrvalOrFunction
			reduce(215), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(215), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(215), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(215), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(215), // ||, reduce: MlrvalOrFunction
			reduce(215), // ^^, reduce: MlrvalOrFunction
			reduce(215), // &&, reduce: MlrvalOrFunction
			reduce(215), // ??, reduce: MlrvalOrFunction
			reduce(215), // ???, reduce: MlrvalOrFunction
			reduce(215), // =~, reduce: MlrvalOrFunction
			reduce(215), // !=~, reduce: MlrvalOrFunction
			reduce(215), // ==, reduce: MlrvalOrFunction
			reduce(215), // !=, reduce: MlrvalOrFunction
			reduce(215), // <=>, reduce: MlrvalOrFunction
			reduce(215), // >=, reduce: MlrvalOrFunction
			reduce(215), // <, reduce: MlrvalOrFunction
			reduce(215), // <=, reduce: MlrvalOrFunction
			reduce(215), // ^, reduce: MlrvalOrFunction
			reduce(215), // &, reduce: MlrvalOrFunction
			reduce(215), // <<, reduce: MlrvalOrFunction
			reduce(215), // >>>, reduce: MlrvalOrFunction
			reduce(215), // +, reduce: MlrvalOrFunction
			reduce(215), // -, reduce: MlrvalOrFunction
			reduce(215), // .+, reduce: MlrvalOrFunction
			reduce(215), // .-, reduce: MlrvalOrFunction
			reduce(215), // *, reduce: MlrvalOrFunction
			reduce(215), // /, reduce: MlrvalOrFunction
			reduce(215), // //, reduce: MlrvalOrFunction
			reduce(215), // %, reduce: MlrvalOrFunction
			reduce(215), // .*, reduce: MlrvalOrFunction
			reduce(215), // ./, reduce: MlrvalOrFunction
			reduce(215), // .//, reduce: MlrvalOrFunction
			reduce(215), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(215), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(215), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(215), // [[, reduce: MlrvalOrFunction
			reduce(215), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1721
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(216), // >, reduce: MlrvalOrFunction
			reduce(216), // >>, reduce: MlrvalOrFunction
			reduce(216), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(216), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(216), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(216), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(216), // ||, reduce: MlrvalOrFunction
			reduce(216), // ^^, reduce: MlrvalOrFunction
			reduce(216), // &&, reduce: MlrvalOrFunction
			reduce(216), // ??, reduce: MlrvalOrFunction
			reduce(216), // ???, reduce: MlrvalOrFunction
			reduce(216), // =~, reduce: MlrvalOrFunction
			reduce(216), // !=~, reduce: MlrvalOrFunction
			reduce(216), // ==, reduce: MlrvalOrFunction
			reduce(216), // !=, reduce: MlrvalOrFunction
			reduce(216), // <=>, reduce: MlrvalOrFunction
			reduce(216), // >=, reduce: MlrvalOrFunction
			reduce(216), // <, reduce: MlrvalOrFunction
			reduce(216), // <=, reduce: MlrvalOrFunction
			reduce(216), // ^, reduce: MlrvalOrFunction
			reduce(216), // &, reduce: MlrvalOrFunction
			reduce(216), // <<, reduce: MlrvalOrFunction
			reduce(216), // >>>, reduce: MlrvalOrFunction
			reduce(216), // +, reduce: MlrvalOrFunction
			reduce(216), // -, reduce: MlrvalOrFunction
			reduce(216), // .+, reduce: MlrvalOrFunction
			reduce(216), // .-, reduce: MlrvalOrFunction
			reduce(216), // *, reduce: MlrvalOrFunction
			reduce(216), // /, reduce: MlrvalOrFunction
			reduce(216), // //, reduce: MlrvalOrFunction
			reduce(216), // %, reduce: MlrvalOrFunction
			reduce(216), // .*, reduce: MlrvalOrFunction
			reduce(216), // ./, reduce: MlrvalOrFunction
			reduce(216), // .//, reduce: MlrvalOrFunction
			reduce(216), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(216), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(216), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(216), // [[, reduce: MlrvalOrFunction
			reduce(216), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1722
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(217), // >, reduce: MlrvalOrFunction
			reduce(217), // >>, reduce: MlrvalOrFunction
			reduce(217), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(217), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(217), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(217), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(217), // ||, reduce: MlrvalOrFunction
			reduce(217), // ^^, reduce: MlrvalOrFunction
			reduce(217), // &&, reduce: MlrvalOrFunction
			reduce(217), // ??, reduce: MlrvalOrFunction
			reduce(217), // ???, reduce: MlrvalOrFunction
			reduce(217), // =~, reduce: MlrvalOrFunction
			reduce(217), // !=~, reduce: MlrvalOrFunction
			reduce(217), // ==, reduce: MlrvalOrFunction
			reduce(217), // !=, reduce: MlrvalOrFunction
			reduce(217), // <=>, reduce: MlrvalOrFunction
			reduce(217), // >=, reduce: MlrvalOrFunction
			reduce(217), // <, reduce: MlrvalOrFunction
			reduce(217), // <=, reduce: MlrvalOrFunction
			reduce(217), // ^, reduce: MlrvalOrFunction
			reduce(217), // &, reduce: MlrvalOrFunction
			reduce(217), // <<, reduce: MlrvalOrFunction
			reduce(217), // >>>, reduce: MlrvalOrFunction
			reduce(217), // +, reduce: MlrvalOrFunction
			reduce(217), // -, reduce: MlrvalOrFunction
			reduce(217), // .+, reduce: MlrvalOrFunction
			reduce(217), // .-, reduce: MlrvalOrFunction
			reduce(217), // *, reduce: MlrvalOrFunction
			reduce(217), // /, reduce: MlrvalOrFunction
			reduce(217), // //, reduce: MlrvalOrFunction
			reduce(217), // %, reduce: MlrvalOrFunction
			reduce(217), // .*, reduce: MlrvalOrFunction
			reduce(217), // ./, reduce: MlrvalOrFunction
			reduce(217), // .//, reduce: MlrvalOrFunction
			reduce(217), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(217), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(217), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(217), // [[, reduce: MlrvalOrFunction
			reduce(217), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1723
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(218), // >, reduce: MlrvalOrFunction
			reduce(218), // >>, reduce: MlrvalOrFunction
			reduce(218), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(218), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(218), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(218), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(218), // ||, reduce: MlrvalOrFunction
			reduce(218), // ^^, reduce: MlrvalOrFunction
			reduce(218), // &&, reduce: MlrvalOrFunction
			reduce(218), // ??, reduce: MlrvalOrFunction
			reduce(218), // ???, reduce: MlrvalOrFunction
			reduce(218), // =~, reduce: MlrvalOrFunction
			reduce(218), // !=~, reduce: MlrvalOrFunction
			reduce(218), // ==, reduce: MlrvalOrFunction
			reduce(218), // !=, reduce: MlrvalOrFunction
			reduce(218), // <=>, reduce: MlrvalOrFunction
			reduce(218), // >=, reduce: MlrvalOrFunction
			reduce(218), // <, reduce: MlrvalOrFunction
			reduce(218), // <=, reduce: MlrvalOrFunction
			reduce(218), // ^, reduce: MlrvalOrFunction
			reduce(218), // &, reduce: MlrvalOrFunction
			reduce(218), // <<, reduce: MlrvalOrFunction
			reduce(218), // >>>, reduce: MlrvalOrFunction
			reduce(218), // +, reduce: MlrvalOrFunction
			reduce(218), // -, reduce: MlrvalOrFunction
			reduce(218), // .+, reduce: MlrvalOrFunction
			reduce(218), // .-, reduce: MlrvalOrFunction
			reduce(218), // *, reduce: MlrvalOrFunction
			reduce(218), // /, reduce: MlrvalOrFunction
			reduce(218), // //, reduce: MlrvalOrFunction
			reduce(218), // %, reduce: MlrvalOrFunction
			reduce(218), // .*, reduce: MlrvalOrFunction
			reduce(218), // ./, reduce: MlrvalOrFunction
			reduce(218), // .//, reduce: MlrvalOrFunction
			reduce(218), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(218), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(218), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(218), // [[, reduce: MlrvalOrFunction
			reduce(218), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1724
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(219), // >, reduce: MlrvalOrFunction
			reduce(219), // >>, reduce: MlrvalOrFunction
			reduce(219), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(219), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(219), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(219), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(219), // ||, reduce: MlrvalOrFunction
			reduce(219), // ^^, reduce: MlrvalOrFunction
			reduce(219), // &&, reduce: MlrvalOrFunction
			reduce(219), // ??, reduce: MlrvalOrFunction
			reduce(219), // ???, reduce: MlrvalOrFunction
			reduce(219), // =~, reduce: MlrvalOrFunction
			reduce(219), // !=~, reduce: MlrvalOrFunction
			reduce(219), // ==, reduce: MlrvalOrFunction
			reduce(219), // !=, reduce: MlrvalOrFunction
			reduce(219), // <=>, reduce: MlrvalOrFunction
			reduce(219), // >=, reduce: MlrvalOrFunction
			reduce(219), // <, reduce: MlrvalOrFunction
			reduce(219), // <=, reduce: MlrvalOrFunction
			reduce(219), // ^, reduce: MlrvalOrFunction
			reduce(219), // &, reduce: MlrvalOrFunction
			reduce(219), // <<, reduce: MlrvalOrFunction
			reduce(219), // >>>, reduce: MlrvalOrFunction
			reduce(219), // +, reduce: MlrvalOrFunction
			reduce(219), // -, reduce: MlrvalOrFunction
			reduce(219), // .+, reduce: MlrvalOrFunction
			reduce(219), // .-, reduce: MlrvalOrFunction
			reduce(219), // *, reduce: MlrvalOrFunction
			reduce(219), // /, reduce: MlrvalOrFunction
			reduce(219), // //, reduce: MlrvalOrFunction
			reduce(219), // %, reduce: MlrvalOrFunction
			reduce(219), // .*, reduce: MlrvalOrFunction
			reduce(219), // ./, reduce: MlrvalOrFunction
			reduce(219), // .//, reduce: MlrvalOrFunction
			reduce(219), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(219), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(219), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(219), // [[, reduce: MlrvalOrFunction
			reduce(219), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1725
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(220), // >, reduce: MlrvalOrFunction
			reduce(220), // >>, reduce: MlrvalOrFunction
			reduce(220), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(220), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(220), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(220), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(220), // ||, reduce: MlrvalOrFunction
			reduce(220), // ^^, reduce: MlrvalOrFunction
			reduce(220), // &&, reduce: MlrvalOrFunction
			reduce(220), // ??, reduce: MlrvalOrFunction
			reduce(220), // ???, reduce: MlrvalOrFunction
			reduce(220), // =~, reduce: MlrvalOrFunction
			reduce(220), // !=~, reduce: MlrvalOrFunction
			reduce(220), // ==, reduce: MlrvalOrFunction
			reduce(220), // !=, reduce: MlrvalOrFunction
			reduce(220), // <=>, reduce: MlrvalOrFunction
			reduce(220), // >=, reduce: MlrvalOrFunction
			reduce(220), // <, reduce: MlrvalOrFunction
			reduce(220), // <=, reduce: MlrvalOrFunction
			reduce(220), // ^, reduce: MlrvalOrFunction
			reduce(220), // &, reduce: MlrvalOrFunction
			reduce(220), // <<, reduce: MlrvalOrFunction
			reduce(220), // >>>, reduce: MlrvalOrFunction
			reduce(220), // +, reduce: MlrvalOrFunction
			reduce(220), // -, reduce: MlrvalOrFunction
			reduce(220), // .+, reduce: MlrvalOrFunction
			reduce(220), // .-, reduce: MlrvalOrFunction
			reduce(220), // *, reduce: MlrvalOrFunction
			reduce(220), // /, reduce: MlrvalOrFunction
			reduce(220), // //, reduce: MlrvalOrFunction
			reduce(220), // %, reduce: MlrvalOrFunction
			reduce(220), // .*, reduce: MlrvalOrFunction
			reduce(220), // ./, reduce: MlrvalOrFunction
			reduce(220), // .//, reduce: MlrvalOrFunction
			reduce(220), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(220), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(220), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(220), // [[, reduce: MlrvalOrFunction
			reduce(220), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1726
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(221), // >, reduce: MlrvalOrFunction
			reduce(221), // >>, reduce: MlrvalOrFunction
			reduce(221), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(221), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(221), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(221), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(221), // ||, reduce: MlrvalOrFunction
			reduce(221), // ^^, reduce: MlrvalOrFunction
			reduce(221), // &&, reduce: MlrvalOrFunction
			reduce(221), // ??, reduce: MlrvalOrFunction
			reduce(221), // ???, reduce: MlrvalOrFunction
			reduce(221), // =~, reduce: MlrvalOrFunction
			reduce(221), // !=~, reduce: MlrvalOrFunction
			reduce(221), // ==, reduce: MlrvalOrFunction
			reduce(221), // !=, reduce: MlrvalOrFunction
			reduce(221), // <=>, reduce: MlrvalOrFunction
			reduce(221), // >=, reduce: MlrvalOrFunction
			reduce(221), // <, reduce: MlrvalOrFunction
			reduce(221), // <=, reduce: MlrvalOrFunction
			reduce(221), // ^, reduce: MlrvalOrFunction
			reduce(221), // &, reduce: MlrvalOrFunction
			reduce(221), // <<, reduce: MlrvalOrFunction
			reduce(221), // >>>, reduce: MlrvalOrFunction
			reduce(221), // +, reduce: MlrvalOrFunction
			reduce(221), // -, reduce: MlrvalOrFunction
			reduce(221), // .+, reduce: MlrvalOrFunction
			reduce(221), // .-, reduce: MlrvalOrFunction
			reduce(221), // *, reduce: MlrvalOrFunction
			reduce(221), // /, reduce: MlrvalOrFunction
			reduce(221), // //, reduce: MlrvalOrFunction
			reduce(221), // %, reduce: MlrvalOrFunction
			reduce(221), // .*, reduce: MlrvalOrFunction
			reduce(221), // ./, reduce: MlrvalOrFunction
			reduce(221), // .//, reduce: MlrvalOrFunction
			reduce(221), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(221), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(221), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(221), // [[, reduce: MlrvalOrFunction
			reduce(221), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1727
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(222), // >, reduce: MlrvalOrFunction
			reduce(222), // >>, reduce: MlrvalOrFunction
			reduce(222), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(222), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(222), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(222), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(222), // ||, reduce: MlrvalOrFunction
			reduce(222), // ^^, reduce: MlrvalOrFunction
			reduce(222), // &&, reduce: MlrvalOrFunction
			reduce(222), // ??, reduce: MlrvalOrFunction
			reduce(222), // ???, reduce: MlrvalOrFunction
			reduce(222), // =~, reduce: MlrvalOrFunction
			reduce(222), // !=~, reduce: MlrvalOrFunction
			reduce(222), // ==, reduce: MlrvalOrFunction
			reduce(222), // !=, reduce: MlrvalOrFunction
			reduce(222), // <=>, reduce: MlrvalOrFunction
			reduce(222), // >=, reduce: MlrvalOrFunction
			reduce(222), // <, reduce: MlrvalOrFunction
			reduce(222), // <=, reduce: MlrvalOrFunction
			reduce(222), // ^, reduce: MlrvalOrFunction
			reduce(222), // &, reduce: MlrvalOrFunction
			reduce(222), // <<, reduce: MlrvalOrFunction
			reduce(222), // >>>, reduce: MlrvalOrFunction
			reduce(222), // +, reduce: MlrvalOrFunction
			reduce(222), // -, reduce: MlrvalOrFunction
			reduce(222), // .+, reduce: MlrvalOrFunction
			reduce(222), // .-, reduce: MlrvalOrFunction
			reduce(222), // *, reduce: MlrvalOrFunction
			reduce(222), // /, reduce: MlrvalOrFunction
			reduce(222), // //, reduce: MlrvalOrFunction
			reduce(222), // %, reduce: MlrvalOrFunction
			reduce(222), // .*, reduce: MlrvalOrFunction
			reduce(222), // ./, reduce: MlrvalOrFunction
			reduce(222), // .//, reduce: MlrvalOrFunction
			reduce(222), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(222), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(222), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(222), // [[, reduce: MlrvalOrFunction
			reduce(222), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1728
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(223), // >, reduce: MlrvalOrFunction
			reduce(223), // >>, reduce: MlrvalOrFunction
			reduce(223), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(223), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(223), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(223), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(223), // ||, reduce: MlrvalOrFunction
			reduce(223), // ^^, reduce: MlrvalOrFunction
			reduce(223), // &&, reduce: MlrvalOrFunction
			reduce(223), // ??, reduce: MlrvalOrFunction
			reduce(223), // ???, reduce: MlrvalOrFunction
			reduce(223), // =~, reduce: MlrvalOrFunction
			reduce(223), // !=~, reduce: MlrvalOrFunction
			reduce(223), // ==, reduce: MlrvalOrFunction
			reduce(223), // !=, reduce: MlrvalOrFunction
			reduce(223), // <=>, reduce: MlrvalOrFunction
			reduce(223), // >=, reduce: MlrvalOrFunction
			reduce(223), // <, reduce: MlrvalOrFunction
			reduce(223), // <=, reduce: MlrvalOrFunction
			reduce(223), // ^, reduce: MlrvalOrFunction
			reduce(223), // &, reduce: MlrvalOrFunction
			reduce(223), // <<, reduce: MlrvalOrFunction
			reduce(223), // >>>, reduce: MlrvalOrFunction
			reduce(223), // +, reduce: MlrvalOrFunction
			reduce(223), // -, reduce: MlrvalOrFunction
			reduce(223), // .+, reduce: MlrvalOrFunction
			reduce(223), // .-, reduce: MlrvalOrFunction
			reduce(223), // *, reduce: MlrvalOrFunction
			reduce(223), // /, reduce: MlrvalOrFunction
			reduce(223), // //, reduce: MlrvalOrFunction
			reduce(223), // %, reduce: MlrvalOrFunction
			reduce(223), // .*, reduce: MlrvalOrFunction
			reduce(223), // ./, reduce: MlrvalOrFunction
			reduce(223), // .//, reduce: MlrvalOrFunction
			reduce(223), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(223), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(223), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(223), // [[, reduce: MlrvalOrFunction
			reduce(223), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1729
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(224), // >, reduce: MlrvalOrFunction
			reduce(224), // >>, reduce: MlrvalOrFunction
			reduce(224), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(224), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(224), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(224), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(224), // ||, reduce: MlrvalOrFunction
			reduce(224), // ^^, reduce: MlrvalOrFunction
			reduce(224), // &&, reduce: MlrvalOrFunction
			reduce(224), // ??, reduce: MlrvalOrFunction
			reduce(224), // ???, reduce: MlrvalOrFunction
			reduce(224), // =~, reduce: MlrvalOrFunction
			reduce(224), // !=~, reduce: MlrvalOrFunction
			reduce(224), // ==, reduce: MlrvalOrFunction
			reduce(224), // !=, reduce: MlrvalOrFunction
			reduce(224), // <=>, reduce: MlrvalOrFunction
			reduce(224), // >=, reduce: MlrvalOrFunction
			reduce(224), // <, reduce: MlrvalOrFunction
			reduce(224), // <=, reduce: MlrvalOrFunction
			reduce(224), // ^, reduce: MlrvalOrFunction
			reduce(224), // &, reduce: MlrvalOrFunction
			reduce(224), // <<, reduce: MlrvalOrFunction
			reduce(224), // >>>, reduce: MlrvalOrFunction
			reduce(224), // +, reduce: MlrvalOrFunction
			reduce(224), // -, reduce: MlrvalOrFunction
			reduce(224), // .+, reduce: MlrvalOrFunction
			reduce(224), // .-, reduce: MlrvalOrFunction
			reduce(224), // *, reduce: MlrvalOrFunction
			reduce(224), // /, reduce: MlrvalOrFunction
			reduce(224), // //, reduce: MlrvalOrFunction
			reduce(224), // %, reduce: MlrvalOrFunction
			reduce(224), // .*, reduce: MlrvalOrFunction
			reduce(224), // ./, reduce: MlrvalOrFunction
			reduce(224), // .//, reduce: MlrvalOrFunction
			reduce(224), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(224), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(224), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(224), // [[, reduce: MlrvalOrFunction
			reduce(224), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1730
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(225), // >, reduce: MlrvalOrFunction
			reduce(225), // >>, reduce: MlrvalOrFunction
			reduce(225), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(225), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(225), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(225), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(225), // ||, reduce: MlrvalOrFunction
			reduce(225), // ^^, reduce: MlrvalOrFunction
			reduce(225), // &&, reduce: MlrvalOrFunction
			reduce(225), // ??, reduce: MlrvalOrFunction
			reduce(225), // ???, reduce: MlrvalOrFunction
			reduce(225), // =~, reduce: MlrvalOrFunction
			reduce(225), // !=~, reduce: MlrvalOrFunction
			reduce(225), // ==, reduce: MlrvalOrFunction
			reduce(225), // !=, reduce: MlrvalOrFunction
			reduce(225), // <=>, reduce: MlrvalOrFunction
			reduce(225), // >=, reduce: MlrvalOrFunction
			reduce(225), // <, reduce: MlrvalOrFunction
			reduce(225), // <=, reduce: MlrvalOrFunction
			reduce(225), // ^, reduce: MlrvalOrFunction
			reduce(225), // &, reduce: MlrvalOrFunction
			reduce(225), // <<, reduce: MlrvalOrFunction
			reduce(225), // >>>, reduce: MlrvalOrFunction
			reduce(225), // +, reduce: MlrvalOrFunction
			reduce(225), // -, reduce: MlrvalOrFunction
			reduce(225), // .+, reduce: MlrvalOrFunction
			reduce(225), // .-, reduce: MlrvalOrFunction
			reduce(225), // *, reduce: MlrvalOrFunction
			reduce(225), // /, reduce: MlrvalOrFunction
			reduce(225), // //, reduce: MlrvalOrFunction
			reduce(225), // %, reduce: MlrvalOrFunction
			reduce(225), // .*, reduce: MlrvalOrFunction
			reduce(225), // ./, reduce: MlrvalOrFunction
			reduce(225), // .//, reduce: MlrvalOrFunction
			reduce(225), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(225), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(225), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(225), // [[, reduce: MlrvalOrFunction
			reduce(225), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1731
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(226), // >, reduce: MlrvalOrFunction
			reduce(226), // >>, reduce: MlrvalOrFunction
			reduce(226), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(226), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(226), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(226), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(226), // ||, reduce: MlrvalOrFunction
			reduce(226), // ^^, reduce: MlrvalOrFunction
			reduce(226), // &&, reduce: MlrvalOrFunction
			reduce(226), // ??, reduce: MlrvalOrFunction
			reduce(226), // ???, reduce: MlrvalOrFunction
			reduce(226), // =~, reduce: MlrvalOrFunction
			reduce(226), // !=~, reduce: MlrvalOrFunction
			reduce(226), // ==, reduce: MlrvalOrFunction
			reduce(226), // !=, reduce: MlrvalOrFunction
			reduce(226), // <=>, reduce: MlrvalOrFunction
			reduce(226), // >=, reduce: MlrvalOrFunction
			reduce(226), // <, reduce: MlrvalOrFunction
			reduce(226), // <=, reduce: MlrvalOrFunction
			reduce(226), // ^, reduce: MlrvalOrFunction
			reduce(226), // &, reduce: MlrvalOrFunction
			reduce(226), // <<, reduce: MlrvalOrFunction
			reduce(226), // >>>, reduce: MlrvalOrFunction
			reduce(226), // +, reduce: MlrvalOrFunction
			reduce(226), // -, reduce: MlrvalOrFunction
			reduce(226), // .+, reduce: MlrvalOrFunction
			reduce(226), // .-, reduce: MlrvalOrFunction
			reduce(226), // *, reduce: MlrvalOrFunction
			reduce(226), // /, reduce: MlrvalOrFunction
			reduce(226), // //, reduce: MlrvalOrFunction
			reduce(226), // %, reduce: MlrvalOrFunction
			reduce(226), // .*, reduce: MlrvalOrFunction
			reduce(226), // ./, reduce: MlrvalOrFunction
			reduce(226), // .//, reduce: MlrvalOrFunction
			reduce(226), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(226), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(226), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(226), // [[, reduce: MlrvalOrFunction
			reduce(226), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1732
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(227), // >, reduce: MlrvalOrFunction
			reduce(227), // >>, reduce: MlrvalOrFunction
			reduce(227), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(227), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(227), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(227), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(227), // ||, reduce: MlrvalOrFunction
			reduce(227), // ^^, reduce: MlrvalOrFunction
			reduce(227), // &&, reduce: MlrvalOrFunction
			reduce(227), // ??, reduce: MlrvalOrFunction
			reduce(227), // ???, reduce: MlrvalOrFunction
			reduce(227), // =~, reduce: MlrvalOrFunction
			reduce(227), // !=~, reduce: MlrvalOrFunction
			reduce(227), // ==, reduce: MlrvalOrFunction
			reduce(227), // !=, reduce: MlrvalOrFunction
			reduce(227), // <=>, reduce: MlrvalOrFunction
			reduce(227), // >=, reduce: MlrvalOrFunction
			reduce(227), // <, reduce: MlrvalOrFunction
			reduce(227), // <=, reduce: MlrvalOrFunction
			reduce(227), // ^, reduce: MlrvalOrFunction
			reduce(227), // &, reduce: MlrvalOrFunction
			reduce(227), // <<, reduce: MlrvalOrFunction
			reduce(227), // >>>, reduce: MlrvalOrFunction
			reduce(227), // +, reduce: MlrvalOrFunction
			reduce(227), // -, reduce: MlrvalOrFunction
			reduce(227), // .+, reduce: MlrvalOrFunction
			reduce(227), // .-, reduce: MlrvalOrFunction
			reduce(227), // *, reduce: MlrvalOrFunction
			reduce(227), // /, reduce: MlrvalOrFunction
			reduce(227), // //, reduce: MlrvalOrFunction
			reduce(227), // %, reduce: MlrvalOrFunction
			reduce(227), // .*, reduce: MlrvalOrFunction
			reduce(227), // ./, reduce: MlrvalOrFunction
			reduce(227), // .//, reduce: MlrvalOrFunction
			reduce(227), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(227), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(227), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(227), // [[, reduce: MlrvalOrFunction
			reduce(227), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1733
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(752),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(756),  // (
			nil,         // )
			nil,         // emitp
			shift(768),  // field_name
			shift(769),  // $[
			shift(2802), // ]
			shift(771),  // braced_field_name
			shift(772),  // $[[
			shift(773),  // $[[[
			shift(774),  // full_srec
			shift(776),  // oosvar_name
			shift(777),  // @[
			shift(778),  // braced_oosvar_name
			shift(779),  // full_oosvar
			shift(780),  // all
			shift(781),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(796),  // +
			shift(798),  // -
			shift(799),  // .+
			shift(800),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(803),  // !
			shift(804),  // ~
			nil,         // **
			shift(809),  // string_literal
			shift(810),  // regex_case_insensitive
			shift(811),  // int_literal
			shift(812),  // float_literal
			shift(813),  // boolean_literal
			shift(814),  // null_literal
			shift(815),  // inf_literal
			shift(816),  // nan_literal
			shift(817),  // const_M_PI
			shift(818),  // const_M_E
			shift(819),  // panic
			shift(821),  // [
			shift(824),  // ctx_IPS
			shift(825),  // ctx_IFS
			shift(826),  // ctx_IRS
			shift(827),  // ctx_OPS
			shift(828),  // ctx_OFS
			shift(829),  // ctx_ORS
			shift(830),  // ctx_FLATSEP
			shift(831),  // ctx_NF
			shift(832),  // ctx_NR
			shift(833),  // ctx_FNR
			shift(834),  // ctx_FILENAME
			shift(835),  // ctx_FILENUM
			shift(837),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(844),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1734
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(240), // >, reduce: MlrvalOrFunction
			reduce(240), // >>, reduce: MlrvalOrFunction
			reduce(240), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(240), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(240), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(240), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(240), // ||, reduce: MlrvalOrFunction
			reduce(240), // ^^, reduce: MlrvalOrFunction
			reduce(240), // &&, reduce: MlrvalOrFunction
			reduce(240), // ??, reduce: MlrvalOrFunction
			reduce(240), // ???, reduce: MlrvalOrFunction
			reduce(240), // =~, reduce: MlrvalOrFunction
			reduce(240), // !=~, reduce: MlrvalOrFunction
			reduce(240), // ==, reduce: MlrvalOrFunction
			reduce(240), // !=, reduce: MlrvalOrFunction
			reduce(240), // <=>, reduce: MlrvalOrFunction
			reduce(240), // >=, reduce: MlrvalOrFunction
			reduce(240), // <, reduce: MlrvalOrFunction
			reduce(240), // <=, reduce: MlrvalOrFunction
			reduce(240), // ^, reduce: MlrvalOrFunction
			reduce(240), // &, reduce: MlrvalOrFunction
			reduce(240), // <<, reduce: MlrvalOrFunction
			reduce(240), // >>>, reduce: MlrvalOrFunction
			reduce(240), // +, reduce: MlrvalOrFunction
			reduce(240), // -, reduce: MlrvalOrFunction
			reduce(240), // .+, reduce: MlrvalOrFunction
			reduce(240), // .-, reduce: MlrvalOrFunction
			reduce(240), // *, reduce: MlrvalOrFunction
			reduce(240), // /, reduce: MlrvalOrFunction
			reduce(240), // //, reduce: MlrvalOrFunction
			reduce(240), // %, reduce: MlrvalOrFunction
			reduce(240), // .*, reduce: MlrvalOrFunction
			reduce(240), // ./, reduce: MlrvalOrFunction
			reduce(240), // .//, reduce: MlrvalOrFunction
			reduce(240), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(240), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(240), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(240), // [[, reduce: MlrvalOrFunction
			reduce(240), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1735
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(241), // >, reduce: ContextVariable
			reduce(241), // >>, reduce: ContextVariable
			reduce(241), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(241), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(241), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(241), // ?, reduce: ContextVariable
			nil,         // :
			reduce(241), // ||, reduce: ContextVariable
			reduce(241), // ^^, reduce: ContextVariable
			reduce(241), // &&, reduce: ContextVariable
			reduce(241), // ??, reduce: ContextVariable
			reduce(241), // ???, reduce: ContextVariable
			reduce(241), // =~, reduce: ContextVariable
			reduce(241), // !=~, reduce: ContextVariable
			reduce(241), // ==, reduce: ContextVariable
			reduce(241), // !=, reduce: ContextVariable
			reduce(241), // <=>, reduce: ContextVariable
			reduce(241), // >=, reduce: ContextVariable
			reduce(241), // <, reduce: ContextVariable
			reduce(241), // <=, reduce: ContextVariable
			reduce(241), // ^, reduce: ContextVariable
			reduce(241), // &, reduce: ContextVariable
			reduce(241), // <<, reduce: ContextVariable
			reduce(241), // >>>, reduce: ContextVariable
			reduce(241), // +, reduce: ContextVariable
			reduce(241), // -, reduce: ContextVariable
			reduce(241), // .+, reduce: ContextVariable
			reduce(241), // .-, reduce: ContextVariable
			reduce(241), // *, reduce: ContextVariable
			reduce(241), // /, reduce: ContextVariable
			reduce(241), // //, reduce: ContextVariable
			reduce(241), // %, reduce: ContextVariable
			reduce(241), // .*, reduce: ContextVariable
			reduce(241), // ./, reduce: ContextVariable
			reduce(241), // .//, reduce: ContextVariable
			reduce(241), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(241), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(241), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(241), // [[, reduce: ContextVariable
			reduce(241), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1736
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(242), // >, reduce: ContextVariable
			reduce(242), // >>, reduce: ContextVariable
			reduce(242), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(242), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(242), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(242), // ?, reduce: ContextVariable
			nil,         // :
			reduce(242), // ||, reduce: ContextVariable
			reduce(242), // ^^, reduce: ContextVariable
			reduce(242), // &&, reduce: ContextVariable
			reduce(242), // ??, reduce: ContextVariable
			reduce(242), // ???, reduce: ContextVariable
			reduce(242), // =~, reduce: ContextVariable
			reduce(242), // !=~, reduce: ContextVariable
			reduce(242), // ==, reduce: ContextVariable
			reduce(242), // !=, reduce: ContextVariable
			reduce(242), // <=>, reduce: ContextVariable
			reduce(242), // >=, reduce: ContextVariable
			reduce(242), // <, reduce: ContextVariable
			reduce(242), // <=, reduce: ContextVariable
			reduce(242), // ^, reduce: ContextVariable
			reduce(242), // &, reduce: ContextVariable
			reduce(242), // <<, reduce: ContextVariable
			reduce(242), // >>>, reduce: ContextVariable
			reduce(242), // +, reduce: ContextVariable
			reduce(242), // -, reduce: ContextVariable
			reduce(242), // .+, reduce: ContextVariable
			reduce(242), // .-, reduce: ContextVariable
			reduce(242), // *, reduce: ContextVariable
			reduce(242), // /, reduce: ContextVariable
			reduce(242), // //, reduce: ContextVariable
			reduce(242), // %, reduce: ContextVariable
			reduce(242), // .*, reduce: ContextVariable
			reduce(242), // ./, reduce: ContextVariable
			reduce(242), // .//, reduce: ContextVariable
			reduce(242), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(242), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(242), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(242), // [[, reduce: ContextVariable
			reduce(242), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1737
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(243), // >, reduce: ContextVariable
			reduce(243), // >>, reduce: ContextVariable
			reduce(243), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(243), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(243), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(243), // ?, reduce: ContextVariable
			nil,         // :
			reduce(243), // ||, reduce: ContextVariable
			reduce(243), // ^^, reduce: ContextVariable
			reduce(243), // &&, reduce: ContextVariable
			reduce(243), // ??, reduce: ContextVariable
			reduce(243), // ???, reduce: ContextVariable
			reduce(243), // =~, reduce: ContextVariable
			reduce(243), // !=~, reduce: ContextVariable
			reduce(243), // ==, reduce: ContextVariable
			reduce(243), // !=, reduce: ContextVariable
			reduce(243), // <=>, reduce: ContextVariable
			reduce(243), // >=, reduce: ContextVariable
			reduce(243), // <, reduce: ContextVariable
			reduce(243), // <=, reduce: ContextVariable
			reduce(243), // ^, reduce: ContextVariable
			reduce(243), // &, reduce: ContextVariable
			reduce(243), // <<, reduce: ContextVariable
			reduce(243), // >>>, reduce: ContextVariable
			reduce(243), // +, reduce: ContextVariable
			reduce(243), // -, reduce: ContextVariable
			reduce(243), // .+, reduce: ContextVariable
			reduce(243), // .-, reduce: ContextVariable
			reduce(243), // *, reduce: ContextVariable
			reduce(243), // /, reduce: ContextVariable
			reduce(243), // //, reduce: ContextVariable
			reduce(243), // %, reduce: ContextVariable
			reduce(243), // .*, reduce: ContextVariable
			reduce(243), // ./, reduce: ContextVariable
			reduce(243), // .//, reduce: ContextVariable
			reduce(243), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(243), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(243), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(243), // [[, reduce: ContextVariable
			reduce(243), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1738
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(244), // >, reduce: ContextVariable
			reduce(244), // >>, reduce: ContextVariable
			reduce(244), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(244), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(244), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(244), // ?, reduce: ContextVariable
			nil,         // :
			reduce(244), // ||, reduce: ContextVariable
			reduce(244), // ^^, reduce: ContextVariable
			reduce(244), // &&, reduce: ContextVariable
			reduce(244), // ??, reduce: ContextVariable
			reduce(244), // ???, reduce: ContextVariable
			reduce(244), // =~, reduce: ContextVariable
			reduce(244), // !=~, reduce: ContextVariable
			reduce(244), // ==, reduce: ContextVariable
			reduce(244), // !=, reduce: ContextVariable
			reduce(244), // <=>, reduce: ContextVariable
			reduce(244), // >=, reduce: ContextVariable
			reduce(244), // <, reduce: ContextVariable
			reduce(244), // <=, reduce: ContextVariable
			reduce(244), // ^, reduce: ContextVariable
			reduce(244), // &, reduce: ContextVariable
			reduce(244), // <<, reduce: ContextVariable
			reduce(244), // >>>, reduce: ContextVariable
			reduce(244), // +, reduce: ContextVariable
			reduce(244), // -, reduce: ContextVariable
			reduce(244), // .+, reduce: ContextVariable
			reduce(244), // .-, reduce: ContextVariable
			reduce(244), // *, reduce: ContextVariable
			reduce(244), // /, reduce: ContextVariable
			reduce(244), // //, reduce: ContextVariable
			reduce(244), // %, reduce: ContextVariable
			reduce(244), // .*, reduce: ContextVariable
			reduce(244), // ./, reduce: ContextVariable
			reduce(244), // .//, reduce: ContextVariable
			reduce(244), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(244), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(244), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(244), // [[, reduce: ContextVariable
			reduce(244), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1739
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(245), // >, reduce: ContextVariable
			reduce(245), // >>, reduce: ContextVariable
			reduce(245), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(245), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(245), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(245), // ?, reduce: ContextVariable
			nil,         // :
			reduce(245), // ||, reduce: ContextVariable
			reduce(245), // ^^, reduce: ContextVariable
			reduce(245), // &&, reduce: ContextVariable
			reduce(245), // ??, reduce: ContextVariable
			reduce(245), // ???, reduce: ContextVariable
			reduce(245), // =~, reduce: ContextVariable
			reduce(245), // !=~, reduce: ContextVariable
			reduce(245), // ==, reduce: ContextVariable
			reduce(245), // !=, reduce: ContextVariable
			reduce(245), // <=>, reduce: ContextVariable
			reduce(245), // >=, reduce: ContextVariable
			reduce(245), // <, reduce: ContextVariable
			reduce(245), // <=, reduce: ContextVariable
			reduce(245), // ^, reduce: ContextVariable
			reduce(245), // &, reduce: ContextVariable
			reduce(245), // <<, reduce: ContextVariable
			reduce(245), // >>>, reduce: ContextVariable
			reduce(245), // +, reduce: ContextVariable
			reduce(245), // -, reduce: ContextVariable
			reduce(245), // .+, reduce: ContextVariable
			reduce(245), // .-, reduce: ContextVariable
			reduce(245), // *, reduce: ContextVariable
			reduce(245), // /, reduce: ContextVariable
			reduce(245), // //, reduce: ContextVariable
			reduce(245), // %, reduce: ContextVariable
			reduce(245), // .*, reduce: ContextVariable
			reduce(245), // ./, reduce: ContextVariable
			reduce(245), // .//, reduce: ContextVariable
			reduce(245), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(245), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(245), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(245), // [[, reduce: ContextVariable
			reduce(245), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1740
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(246), // >, reduce: ContextVariable
			reduce(246), // >>, reduce: ContextVariable
			reduce(246), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(246), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(246), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(246), // ?, reduce: ContextVariable
			nil,         // :
			reduce(246), // ||, reduce: ContextVariable
			reduce(246), // ^^, reduce: ContextVariable
			reduce(246), // &&, reduce: ContextVariable
			reduce(246), // ??, reduce: ContextVariable
			reduce(246), // ???, reduce: ContextVariable
			reduce(246), // =~, reduce: ContextVariable
			reduce(246), // !=~, reduce: ContextVariable
			reduce(246), // ==, reduce: ContextVariable
			reduce(246), // !=, reduce: ContextVariable
			reduce(246), // <=>, reduce: ContextVariable
			reduce(246), // >=, reduce: ContextVariable
			reduce(246), // <, reduce: ContextVariable
			reduce(246), // <=, reduce: ContextVariable
			reduce(246), // ^, reduce: ContextVariable
			reduce(246), // &, reduce: ContextVariable
			reduce(246), // <<, reduce: ContextVariable
			reduce(246), // >>>, reduce: ContextVariable
			reduce(246), // +, reduce: ContextVariable
			reduce(246), // -, reduce: ContextVariable
			reduce(246), // .+, reduce: ContextVariable
			reduce(246), // .-, reduce: ContextVariable
			reduce(246), // *, reduce: ContextVariable
			reduce(246), // /, reduce: ContextVariable
			reduce(246), // //, reduce: ContextVariable
			reduce(246), // %, reduce: ContextVariable
			reduce(246), // .*, reduce: ContextVariable
			reduce(246), // ./, reduce: ContextVariable
			reduce(246), // .//, reduce: ContextVariable
			reduce(246), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(246), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(246), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(246), // [[, reduce: ContextVariable
			reduce(246), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1741
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(247), // >, reduce: ContextVariable
			reduce(247), // >>, reduce: ContextVariable
			reduce(247), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(247), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(247), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(247), // ?, reduce: ContextVariable
			nil,         // :
			reduce(247), // ||, reduce: ContextVariable
			reduce(247), // ^^, reduce: ContextVariable
			reduce(247), // &&, reduce: ContextVariable
			reduce(247), // ??, reduce: ContextVariable
			reduce(247), // ???, reduce: ContextVariable
			reduce(247), // =~, reduce: ContextVariable
			reduce(247), // !=~, reduce: ContextVariable
			reduce(247), // ==, reduce: ContextVariable
			reduce(247), // !=, reduce: ContextVariable
			reduce(247), // <=>, reduce: ContextVariable
			reduce(247), // >=, reduce: ContextVariable
			reduce(247), // <, reduce: ContextVariable
			reduce(247), // <=, reduce: ContextVariable
			reduce(247), // ^, reduce: ContextVariable
			reduce(247), // &, reduce: ContextVariable
			reduce(247), // <<, reduce: ContextVariable
			reduce(247), // >>>, reduce: ContextVariable
			reduce(247), // +, reduce: ContextVariable
			reduce(247), // -, reduce: ContextVariable
			reduce(247), // .+, reduce: ContextVariable
			reduce(247), // .-, reduce: ContextVariable
			reduce(247), // *, reduce: ContextVariable
			reduce(247), // /, reduce: ContextVariable
			reduce(247), // //, reduce: ContextVariable
			reduce(247), // %, reduce: ContextVariable
			reduce(247), // .*, reduce: ContextVariable
			reduce(247), // ./, reduce: ContextVariable
			reduce(247), // .//, reduce: ContextVariable
			reduce(247), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(247), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(247), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(247), // [[, reduce: ContextVariable
			reduce(247), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1742
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(248), // >, reduce: ContextVariable
			reduce(248), // >>, reduce: ContextVariable
			reduce(248), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(248), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(248), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(248), // ?, reduce: ContextVariable
			nil,         // :
			reduce(248), // ||, reduce: ContextVariable
			reduce(248), // ^^, reduce: ContextVariable
			reduce(248), // &&, reduce: ContextVariable
			reduce(248), // ??, reduce: ContextVariable
			reduce(248), // ???, reduce: ContextVariable
			reduce(248), // =~, reduce: ContextVariable
			reduce(248), // !=~, reduce: ContextVariable
			reduce(248), // ==, reduce: ContextVariable
			reduce(248), // !=, reduce: ContextVariable
			reduce(248), // <=>, reduce: ContextVariable
			reduce(248), // >=, reduce: ContextVariable
			reduce(248), // <, reduce: ContextVariable
			reduce(248), // <=, reduce: ContextVariable
			reduce(248), // ^, reduce: ContextVariable
			reduce(248), // &, reduce: ContextVariable
			reduce(248), // <<, reduce: ContextVariable
			reduce(248), // >>>, reduce: ContextVariable
			reduce(248), // +, reduce: ContextVariable
			reduce(248), // -, reduce: ContextVariable
			reduce(248), // .+, reduce: ContextVariable
			reduce(248), // .-, reduce: ContextVariable
			reduce(248), // *, reduce: ContextVariable
			reduce(248), // /, reduce: ContextVariable
			reduce(248), // //, reduce: ContextVariable
			reduce(248), // %, reduce: ContextVariable
			reduce(248), // .*, reduce: ContextVariable
			reduce(248), // ./, reduce: ContextVariable
			reduce(248), // .//, reduce: ContextVariable
			reduce(248), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(248), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(248), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(248), // [[, reduce: ContextVariable
			reduce(248), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1743
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(249), // >, reduce: ContextVariable
			reduce(249), // >>, reduce: ContextVariable
			reduce(249), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(249), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(249), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(249), // ?, reduce: ContextVariable
			nil,         // :
			reduce(249), // ||, reduce: ContextVariable
			reduce(249), // ^^, reduce: ContextVariable
			reduce(249), // &&, reduce: ContextVariable
			reduce(249), // ??, reduce: ContextVariable
			reduce(249), // ???, reduce: ContextVariable
			reduce(249), // =~, reduce: ContextVariable
			reduce(249), // !=~, reduce: ContextVariable
			reduce(249), // ==, reduce: ContextVariable
			reduce(249), // !=, reduce: ContextVariable
			reduce(249), // <=>, reduce: ContextVariable
			reduce(249), // >=, reduce: ContextVariable
			reduce(249), // <, reduce: ContextVariable
			reduce(249), // <=, reduce: ContextVariable
			reduce(249), // ^, reduce: ContextVariable
			reduce(249), // &, reduce: ContextVariable
			reduce(249), // <<, reduce: ContextVariable
			reduce(249), // >>>, reduce: ContextVariable
			reduce(249), // +, reduce: ContextVariable
			reduce(249), // -, reduce: ContextVariable
			reduce(249), // .+, reduce: ContextVariable
			reduce(249), // .-, reduce: ContextVariable
			reduce(249), // *, reduce: ContextVariable
			reduce(249), // /, reduce: ContextVariable
			reduce(249), // //, reduce: ContextVariable
			reduce(249), // %, reduce: ContextVariable
			reduce(249), // .*, reduce: ContextVariable
			reduce(249), // ./, reduce: ContextVariable
			reduce(249), // .//, reduce: ContextVariable
			reduce(249), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(249), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(249), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(249), // [[, reduce: ContextVariable
			reduce(249), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1744
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(250), // >, reduce: ContextVariable
			reduce(250), // >>, reduce: ContextVariable
			reduce(250), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(250), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(250), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(250), // ?, reduce: ContextVariable
			nil,         // :
			reduce(250), // ||, reduce: ContextVariable
			reduce(250), // ^^, reduce: ContextVariable
			reduce(250), // &&, reduce: ContextVariable
			reduce(250), // ??, reduce: ContextVariable
			reduce(250), // ???, reduce: ContextVariable
			reduce(250), // =~, reduce: ContextVariable
			reduce(250), // !=~, reduce: ContextVariable
			reduce(250), // ==, reduce: ContextVariable
			reduce(250), // !=, reduce: ContextVariable
			reduce(250), // <=>, reduce: ContextVariable
			reduce(250), // >=, reduce: ContextVariable
			reduce(250), // <, reduce: ContextVariable
			reduce(250), // <=, reduce: ContextVariable
			reduce(250), // ^, reduce: ContextVariable
			reduce(250), // &, reduce: ContextVariable
			reduce(250), // <<, reduce: ContextVariable
			reduce(250), // >>>, reduce: ContextVariable
			reduce(250), // +, reduce: ContextVariable
			reduce(250), // -, reduce: ContextVariable
			reduce(250), // .+, reduce: ContextVariable
			reduce(250), // .-, reduce: ContextVariable
			reduce(250), // *, reduce: ContextVariable
			reduce(250), // /, reduce: ContextVariable
			reduce(250), // //, reduce: ContextVariable
			reduce(250), // %, reduce: ContextVariable
			reduce(250), // .*, reduce: ContextVariable
			reduce(250), // ./, reduce: ContextVariable
			reduce(250), // .//, reduce: ContextVariable
			reduce(250), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(250), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(250), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(250), // [[, reduce: ContextVariable
			reduce(250), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1745
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(251), // >, reduce: ContextVariable
			reduce(251), // >>, reduce: ContextVariable
			reduce(251), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(251), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(251), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(251), // ?, reduce: ContextVariable
			nil,         // :
			reduce(251), // ||, reduce: ContextVariable
			reduce(251), // ^^, reduce: ContextVariable
			reduce(251), // &&, reduce: ContextVariable
			reduce(251), // ??, reduce: ContextVariable
			reduce(251), // ???, reduce: ContextVariable
			reduce(251), // =~, reduce: ContextVariable
			reduce(251), // !=~, reduce: ContextVariable
			reduce(251), // ==, reduce: ContextVariable
			reduce(251), // !=, reduce: ContextVariable
			reduce(251), // <=>, reduce: ContextVariable
			reduce(251), // >=, reduce: ContextVariable
			reduce(251), // <, reduce: ContextVariable
			reduce(251), // <=, reduce: ContextVariable
			reduce(251), // ^, reduce: ContextVariable
			reduce(251), // &, reduce: ContextVariable
			reduce(251), // <<, reduce: ContextVariable
			reduce(251), // >>>, reduce: ContextVariable
			reduce(251), // +, reduce: ContextVariable
			reduce(251), // -, reduce: ContextVariable
			reduce(251), // .+, reduce: ContextVariable
			reduce(251), // .-, reduce: ContextVariable
			reduce(251), // *, reduce: ContextVariable
			reduce(251), // /, reduce: ContextVariable
			reduce(251), // //, reduce: ContextVariable
			reduce(251), // %, reduce: ContextVariable
			reduce(251), // .*, reduce: ContextVariable
			reduce(251), // ./, reduce: ContextVariable
			reduce(251), // .//, reduce: ContextVariable
			reduce(251), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(251), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(251), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(251), // [[, reduce: ContextVariable
			reduce(251), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1746
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(252), // >, reduce: ContextVariable
			reduce(252), // >>, reduce: ContextVariable
			reduce(252), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(252), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(252), // ), reduce: ContextVariable
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(252), // ?, reduce: ContextVariable
			nil,         // :
			reduce(252), // ||, reduce: ContextVariable
			reduce(252), // ^^, reduce: ContextVariable
			reduce(252), // &&, reduce: ContextVariable
			reduce(252), // ??, reduce: ContextVariable
			reduce(252), // ???, reduce: ContextVariable
			reduce(252), // =~, reduce: ContextVariable
			reduce(252), // !=~, reduce: ContextVariable
			reduce(252), // ==, reduce: ContextVariable
			reduce(252), // !=, reduce: ContextVariable
			reduce(252), // <=>, reduce: ContextVariable
			reduce(252), // >=, reduce: ContextVariable
			reduce(252), // <, reduce: ContextVariable
			reduce(252), // <=, reduce: ContextVariable
			reduce(252), // ^, reduce: ContextVariable
			reduce(252), // &, reduce: ContextVariable
			reduce(252), // <<, reduce: ContextVariable
			reduce(252), // >>>, reduce: ContextVariable
			reduce(252), // +, reduce: ContextVariable
			reduce(252), // -, reduce: ContextVariable
			reduce(252), // .+, reduce: ContextVariable
			reduce(252), // .-, reduce: ContextVariable
			reduce(252), // *, reduce: ContextVariable
			reduce(252), // /, reduce: ContextVariable
			reduce(252), // //, reduce: ContextVariable
			reduce(252), // %, reduce: ContextVariable
			reduce(252), // .*, reduce: ContextVariable
			reduce(252), // ./, reduce: ContextVariable
			reduce(252), // .//, reduce: ContextVariable
			reduce(252), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(252), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(252), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(252), // [[, reduce: ContextVariable
			reduce(252), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1747
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(253), // >, reduce: MlrvalOrFunction
			reduce(253), // >>, reduce: MlrvalOrFunction
			reduce(253), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(253), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(253), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(253), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(253), // ||, reduce: MlrvalOrFunction
			reduce(253), // ^^, reduce: MlrvalOrFunction
			reduce(253), // &&, reduce: MlrvalOrFunction
			reduce(253), // ??, reduce: MlrvalOrFunction
			reduce(253), // ???, reduce: MlrvalOrFunction
			reduce(253), // =~, reduce: MlrvalOrFunction
			reduce(253), // !=~, reduce: MlrvalOrFunction
			reduce(253), // ==, reduce: MlrvalOrFunction
			reduce(253), // !=, reduce: MlrvalOrFunction
			reduce(253), // <=>, reduce: MlrvalOrFunction
			reduce(253), // >=, reduce: MlrvalOrFunction
			reduce(253), // <, reduce: MlrvalOrFunction
			reduce(253), // <=, reduce: MlrvalOrFunction
			reduce(253), // ^, reduce: MlrvalOrFunction
			reduce(253), // &, reduce: MlrvalOrFunction
			reduce(253), // <<, reduce: MlrvalOrFunction
			reduce(253), // >>>, reduce: MlrvalOrFunction
			reduce(253), // +, reduce: MlrvalOrFunction
			reduce(253), // -, reduce: MlrvalOrFunction
			reduce(253), // .+, reduce: MlrvalOrFunction
			reduce(253), // .-, reduce: MlrvalOrFunction
			reduce(253), // *, reduce: MlrvalOrFunction
			reduce(253), // /, reduce: MlrvalOrFunction
			reduce(253), // //, reduce: MlrvalOrFunction
			reduce(253), // %, reduce: MlrvalOrFunction
			reduce(253), // .*, reduce: MlrvalOrFunction
			reduce(253), // ./, reduce: MlrvalOrFunction
			reduce(253), // .//, reduce: MlrvalOrFunction
			reduce(253), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(253), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(253), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(253), // [[, reduce: MlrvalOrFunction
			reduce(253), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1748
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			shift(2804), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(2805), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1749
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(256), // >, reduce: MlrvalOrFunction
			reduce(256), // >>, reduce: MlrvalOrFunction
			reduce(256), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(256), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(256), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(256), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(256), // ||, reduce: MlrvalOrFunction
			reduce(256), // ^^, reduce: MlrvalOrFunction
			reduce(256), // &&, reduce: MlrvalOrFunction
			reduce(256), // ??, reduce: MlrvalOrFunction
			reduce(256), // ???, reduce: MlrvalOrFunction
			reduce(256), // =~, reduce: MlrvalOrFunction
			reduce(256), // !=~, reduce: MlrvalOrFunction
			reduce(256), // ==, reduce: MlrvalOrFunction
			reduce(256), // !=, reduce: MlrvalOrFunction
			reduce(256), // <=>, reduce: MlrvalOrFunction
			reduce(256), // >=, reduce: MlrvalOrFunction
			reduce(256), // <, reduce: MlrvalOrFunction
			reduce(256), // <=, reduce: MlrvalOrFunction
			reduce(256), // ^, reduce: MlrvalOrFunction
			reduce(256), // &, reduce: MlrvalOrFunction
			reduce(256), // <<, reduce: MlrvalOrFunction
			reduce(256), // >>>, reduce: MlrvalOrFunction
			reduce(256), // +, reduce: MlrvalOrFunction
			reduce(256), // -, reduce: MlrvalOrFunction
			reduce(256), // .+, reduce: MlrvalOrFunction
			reduce(256), // .-, reduce: MlrvalOrFunction
			reduce(256), // *, reduce: MlrvalOrFunction
			reduce(256), // /, reduce: MlrvalOrFunction
			reduce(256), // //, reduce: MlrvalOrFunction
			reduce(256), // %, reduce: MlrvalOrFunction
			reduce(256), // .*, reduce: MlrvalOrFunction
			reduce(256), // ./, reduce: MlrvalOrFunction
			reduce(256), // .//, reduce: MlrvalOrFunction
			reduce(256), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(256), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(256), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(256), // [[, reduce: MlrvalOrFunction
			reduce(256), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1750
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(257), // >, reduce: MlrvalOrFunction
			reduce(257), // >>, reduce: MlrvalOrFunction
			reduce(257), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(257), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(257), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(257), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(257), // ||, reduce: MlrvalOrFunction
			reduce(257), // ^^, reduce: MlrvalOrFunction
			reduce(257), // &&, reduce: MlrvalOrFunction
			reduce(257), // ??, reduce: MlrvalOrFunction
			reduce(257), // ???, reduce: MlrvalOrFunction
			reduce(257), // =~, reduce: MlrvalOrFunction
			reduce(257), // !=~, reduce: MlrvalOrFunction
			reduce(257), // ==, reduce: MlrvalOrFunction
			reduce(257), // !=, reduce: MlrvalOrFunction
			reduce(257), // <=>, reduce: MlrvalOrFunction
			reduce(257), // >=, reduce: MlrvalOrFunction
			reduce(257), // <, reduce: MlrvalOrFunction
			reduce(257), // <=, reduce: MlrvalOrFunction
			reduce(257), // ^, reduce: MlrvalOrFunction
			reduce(257), // &, reduce: MlrvalOrFunction
			reduce(257), // <<, reduce: MlrvalOrFunction
			reduce(257), // >>>, reduce: MlrvalOrFunction
			reduce(257), // +, reduce: MlrvalOrFunction
			reduce(257), // -, reduce: MlrvalOrFunction
			reduce(257), // .+, reduce: MlrvalOrFunction
			reduce(257), // .-, reduce: MlrvalOrFunction
			reduce(257), // *, reduce: MlrvalOrFunction
			reduce(257), // /, reduce: MlrvalOrFunction
			reduce(257), // //, reduce: MlrvalOrFunction
			reduce(257), // %, reduce: MlrvalOrFunction
			reduce(257), // .*, reduce: MlrvalOrFunction
			reduce(257), // ./, reduce: MlrvalOrFunction
			reduce(257), // .//, reduce: MlrvalOrFunction
			reduce(257), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(257), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(257), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(257), // [[, reduce: MlrvalOrFunction
			reduce(257), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1751
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(258), // >, reduce: MlrvalOrFunction
			reduce(258), // >>, reduce: MlrvalOrFunction
			reduce(258), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(258), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(258), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(258), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(258), // ||, reduce: MlrvalOrFunction
			reduce(258), // ^^, reduce: MlrvalOrFunction
			reduce(258), // &&, reduce: MlrvalOrFunction
			reduce(258), // ??, reduce: MlrvalOrFunction
			reduce(258), // ???, reduce: MlrvalOrFunction
			reduce(258), // =~, reduce: MlrvalOrFunction
			reduce(258), // !=~, reduce: MlrvalOrFunction
			reduce(258), // ==, reduce: MlrvalOrFunction
			reduce(258), // !=, reduce: MlrvalOrFunction
			reduce(258), // <=>, reduce: MlrvalOrFunction
			reduce(258), // >=, reduce: MlrvalOrFunction
			reduce(258), // <, reduce: MlrvalOrFunction
			reduce(258), // <=, reduce: MlrvalOrFunction
			reduce(258), // ^, reduce: MlrvalOrFunction
			reduce(258), // &, reduce: MlrvalOrFunction
			reduce(258), // <<, reduce: MlrvalOrFunction
			reduce(258), // >>>, reduce: MlrvalOrFunction
			reduce(258), // +, reduce: MlrvalOrFunction
			reduce(258), // -, reduce: MlrvalOrFunction
			reduce(258), // .+, reduce: MlrvalOrFunction
			reduce(258), // .-, reduce: MlrvalOrFunction
			reduce(258), // *, reduce: MlrvalOrFunction
			reduce(258), // /, reduce: MlrvalOrFunction
			reduce(258), // //, reduce: MlrvalOrFunction
			reduce(258), // %, reduce: MlrvalOrFunction
			reduce(258), // .*, reduce: MlrvalOrFunction
			reduce(258), // ./, reduce: MlrvalOrFunction
			reduce(258), // .//, reduce: MlrvalOrFunction
			reduce(258), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(258), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(258), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(258), // [[, reduce: MlrvalOrFunction
			reduce(258), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1752
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(259), // >, reduce: MlrvalOrFunction
			reduce(259), // >>, reduce: MlrvalOrFunction
			reduce(259), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(259), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(259), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(259), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(259), // ||, reduce: MlrvalOrFunction
			reduce(259), // ^^, reduce: MlrvalOrFunction
			reduce(259), // &&, reduce: MlrvalOrFunction
			reduce(259), // ??, reduce: MlrvalOrFunction
			reduce(259), // ???, reduce: MlrvalOrFunction
			reduce(259), // =~, reduce: MlrvalOrFunction
			reduce(259), // !=~, reduce: MlrvalOrFunction
			reduce(259), // ==, reduce: MlrvalOrFunction
			reduce(259), // !=, reduce: MlrvalOrFunction
			reduce(259), // <=>, reduce: MlrvalOrFunction
			reduce(259), // >=, reduce: MlrvalOrFunction
			reduce(259), // <, reduce: MlrvalOrFunction
			reduce(259), // <=, reduce: MlrvalOrFunction
			reduce(259), // ^, reduce: MlrvalOrFunction
			reduce(259), // &, reduce: MlrvalOrFunction
			reduce(259), // <<, reduce: MlrvalOrFunction
			reduce(259), // >>>, reduce: MlrvalOrFunction
			reduce(259), // +, reduce: MlrvalOrFunction
			reduce(259), // -, reduce: MlrvalOrFunction
			reduce(259), // .+, reduce: MlrvalOrFunction
			reduce(259), // .-, reduce: MlrvalOrFunction
			reduce(259), // *, reduce: MlrvalOrFunction
			reduce(259), // /, reduce: MlrvalOrFunction
			reduce(259), // //, reduce: MlrvalOrFunction
			reduce(259), // %, reduce: MlrvalOrFunction
			reduce(259), // .*, reduce: MlrvalOrFunction
			reduce(259), // ./, reduce: MlrvalOrFunction
			reduce(259), // .//, reduce: MlrvalOrFunction
			reduce(259), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(259), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(259), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(259), // [[, reduce: MlrvalOrFunction
			reduce(259), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1753
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(267), // >, reduce: MlrvalOrFunction
			reduce(267), // >>, reduce: MlrvalOrFunction
			reduce(267), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(267), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(267), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(267), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(267), // ||, reduce: MlrvalOrFunction
			reduce(267), // ^^, reduce: MlrvalOrFunction
			reduce(267), // &&, reduce: MlrvalOrFunction
			reduce(267), // ??, reduce: MlrvalOrFunction
			reduce(267), // ???, reduce: MlrvalOrFunction
			reduce(267), // =~, reduce: MlrvalOrFunction
			reduce(267), // !=~, reduce: MlrvalOrFunction
			reduce(267), // ==, reduce: MlrvalOrFunction
			reduce(267), // !=, reduce: MlrvalOrFunction
			reduce(267), // <=>, reduce: MlrvalOrFunction
			reduce(267), // >=, reduce: MlrvalOrFunction
			reduce(267), // <, reduce: MlrvalOrFunction
			reduce(267), // <=, reduce: MlrvalOrFunction
			reduce(267), // ^, reduce: MlrvalOrFunction
			reduce(267), // &, reduce: MlrvalOrFunction
			reduce(267), // <<, reduce: MlrvalOrFunction
			reduce(267), // >>>, reduce: MlrvalOrFunction
			reduce(267), // +, reduce: MlrvalOrFunction
			reduce(267), // -, reduce: MlrvalOrFunction
			reduce(267), // .+, reduce: MlrvalOrFunction
			reduce(267), // .-, reduce: MlrvalOrFunction
			reduce(267), // *, reduce: MlrvalOrFunction
			reduce(267), // /, reduce: MlrvalOrFunction
			reduce(267), // //, reduce: MlrvalOrFunction
			reduce(267), // %, reduce: MlrvalOrFunction
			reduce(267), // .*, reduce: MlrvalOrFunction
			reduce(267), // ./, reduce: MlrvalOrFunction
			reduce(267), // .//, reduce: MlrvalOrFunction
			reduce(267), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(267), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(267), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(267), // [[, reduce: MlrvalOrFunction
			reduce(267), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1754
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2806), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1755
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2807), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1756
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(2809), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1757
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1758
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(296), // $, reduce: ElseBlock
			nil,         // empty
			reduce(296), // ;, reduce: ElseBlock
			reduce(296), // {, reduce: ElseBlock
			nil,         // }
			nil,         // =
			reduce(296), // unset, reduce: ElseBlock
			reduce(296), // filter, reduce: ElseBlock
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(296), // print, reduce: ElseBlock
			nil,         // ,
			reduce(296), // printn, reduce: ElseBlock
			reduce(296), // eprint, reduce: ElseBlock
			reduce(296), // eprintn, reduce: ElseBlock
			reduce(296), // dump, reduce: ElseBlock
			reduce(296), // edump, reduce: ElseBlock
			reduce(296), // tee, reduce: ElseBlock
			reduce(296), // emitf, reduce: ElseBlock
			reduce(296), // emit1, reduce: ElseBlock
			reduce(296), // emit, reduce: ElseBlock
			reduce(296), // (, reduce: ElseBlock
			nil,         // )
			reduce(296), // emitp, reduce: ElseBlock
			reduce(296), // field_name, reduce: ElseBlock
			reduce(296), // $[, reduce: ElseBlock
			nil,         // ]
			reduce(296), // braced_field_name, reduce: ElseBlock
			reduce(296), // $[[, reduce: ElseBlock
			reduce(296), // $[[[, reduce: ElseBlock
			reduce(296), // full_srec, reduce: ElseBlock
			reduce(296), // oosvar_name, reduce: ElseBlock
			reduce(296), // @[, reduce: ElseBlock
			reduce(296), // braced_oosvar_name, reduce: ElseBlock
			reduce(296), // full_oosvar, reduce: ElseBlock
			reduce(296), // all, reduce: ElseBlock
			reduce(296), // non_sigil_name, reduce: ElseBlock
			reduce(296), // arr, reduce: ElseBlock
			reduce(296), // bool, reduce: ElseBlock
			reduce(296), // float, reduce: ElseBlock
			reduce(296), // int, reduce: ElseBlock
			reduce(296), // map, reduce: ElseBlock
			reduce(296), // num, reduce: ElseBlock
			reduce(296), // str, reduce: ElseBlock
			reduce(296), // var, reduce: ElseBlock
			reduce(296), // funct, reduce: ElseBlock
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(296), // +, reduce: ElseBlock
			reduce(296), // -, reduce: ElseBlock
			reduce(296), // .+, reduce: ElseBlock
			reduce(296), // .-, reduce: ElseBlock
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(296), // !, reduce: ElseBlock
			reduce(296), // ~, reduce: ElseBlock
			nil,         // **
			reduce(296), // string_literal, reduce: ElseBlock
			reduce(296), // regex_case_insensitive, reduce: ElseBlock
			reduce(296), // int_literal, reduce: ElseBlock
			reduce(296), // float_literal, reduce: ElseBlock
			reduce(296), // boolean_literal, reduce: ElseBlock
			reduce(296), // null_literal, reduce: ElseBlock
			reduce(296), // inf_literal, reduce: ElseBlock
			reduce(296), // nan_literal, reduce: ElseBlock
			reduce(296), // const_M_PI, reduce: ElseBlock
			reduce(296), // const_M_E, reduce: ElseBlock
			reduce(296), // panic, reduce: ElseBlock
			reduce(296), // [, reduce: ElseBlock
			reduce(296), // ctx_IPS, reduce: ElseBlock
			reduce(296), // ctx_IFS, reduce: ElseBlock
			reduce(296), // ctx_IRS, reduce: ElseBlock
			reduce(296), // ctx_OPS, reduce: ElseBlock
			reduce(296), // ctx_OFS, reduce: ElseBlock
			reduce(296), // ctx_ORS, reduce: ElseBlock
			reduce(296), // ctx_FLATSEP, reduce: ElseBlock
			reduce(296), // ctx_NF, reduce: ElseBlock
			reduce(296), // ctx_NR, reduce: ElseBlock
			reduce(296), // ctx_FNR, reduce: ElseBlock
			reduce(296), // ctx_FILENAME, reduce: ElseBlock
			reduce(296), // ctx_FILENUM, reduce: ElseBlock
			reduce(296), // env, reduce: ElseBlock
			nil,         // [[
			nil,         // [[[
			reduce(296), // call, reduce: ElseBlock
			reduce(296), // begin, reduce: ElseBlock
			reduce(296), // end, reduce: ElseBlock
			reduce(296), // if, reduce: ElseBlock
			nil,         // elif
			nil,         // else
			reduce(296), // while, reduce: ElseBlock
			reduce(296), // do, reduce: ElseBlock
			reduce(296), // for, reduce: ElseBlock
			nil,         // in
			reduce(296), // break, reduce: ElseBlock
			reduce(296), // continue, reduce: ElseBlock
			reduce(296), // func, reduce: ElseBlock
			reduce(296), // subr, reduce: ElseBlock
			reduce(296), // return, reduce: ElseBlock
		},
	},
	actionRow{ // S1759
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(2811), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1760
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(2812), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1761
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2813), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1762
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(2814), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1763
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			shift(2815), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1764
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(310), // ;, reduce: TripleForStart
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(310), // ,, reduce: TripleForStart
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1765
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			shift(2817), // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			shift(2818), // ||=
			shift(2819), // ^^=
			shift(2820), // &&=
			shift(2821), // ??=
			shift(2822), // ???=
			shift(2823), // |=
			shift(2824), // &=
			shift(2825), // ^=
			shift(2826), // <<=
			shift(2827), // >>=
			shift(2828), // >>>=
			shift(2829), // +=
			shift(2830), // .=
			shift(2831), // -=
			shift(2832), // *=
			shift(2833), // /=
			shift(2834), // //=
			shift(2835), // %=
			shift(2836), // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1766
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			reduce(36), // =, reduce: Lvalue
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			reduce(36), // ||=, reduce: Lvalue
			reduce(36), // ^^=, reduce: Lvalue
			reduce(36), // &&=, reduce: Lvalue
			reduce(36), // ??=, reduce: Lvalue
			reduce(36), // ???=, reduce: Lvalue
			reduce(36), // |=, reduce: Lvalue
			reduce(36), // &=, reduce: Lvalue
			reduce(36), // ^=, reduce: Lvalue
			reduce(36), // <<=, reduce: Lvalue
			reduce(36), // >>=, reduce: Lvalue
			reduce(36), // >>>=, reduce: Lvalue
			reduce(36), // +=, reduce: Lvalue
			reduce(36), // .=, reduce: Lvalue
			reduce(36), // -=, reduce: Lvalue
			reduce(36), // *=, reduce: Lvalue
			reduce(36), // /=, reduce: Lvalue
			reduce(36), // //=, reduce: Lvalue
			reduce(36), // %=, reduce: Lvalue
			reduce(36), // **=, reduce: Lvalue
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1767
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(214), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(2837), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(214), // ||=, reduce: MlrvalOrFunction
			reduce(214), // ^^=, reduce: MlrvalOrFunction
			reduce(214), // &&=, reduce: MlrvalOrFunction
			reduce(214), // ??=, reduce: MlrvalOrFunction
			reduce(214), // ???=, reduce: MlrvalOrFunction
			reduce(214), // |=, reduce: MlrvalOrFunction
			reduce(214), // &=, reduce: MlrvalOrFunction
			reduce(214), // ^=, reduce: MlrvalOrFunction
			reduce(214), // <<=, reduce: MlrvalOrFunction
			reduce(214), // >>=, reduce: MlrvalOrFunction
			reduce(214), // >>>=, reduce: MlrvalOrFunction
			reduce(214), // +=, reduce: MlrvalOrFunction
			reduce(214), // .=, reduce: MlrvalOrFunction
			reduce(214), // -=, reduce: MlrvalOrFunction
			reduce(214), // *=, reduce: MlrvalOrFunction
			reduce(214), // /=, reduce: MlrvalOrFunction
			reduce(214), // //=, reduce: MlrvalOrFunction
			reduce(214), // %=, reduce: MlrvalOrFunction
			reduce(214), // **=, reduce: MlrvalOrFunction
			reduce(214), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			shift(2838), // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1768
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(211), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(211), // >, reduce: MlrvalOrFunction
			reduce(211), // >>, reduce: MlrvalOrFunction
			reduce(211), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(211), // ||=, reduce: MlrvalOrFunction
			reduce(211), // ^^=, reduce: MlrvalOrFunction
			reduce(211), // &&=, reduce: MlrvalOrFunction
			reduce(211), // ??=, reduce: MlrvalOrFunction
			reduce(211), // ???=, reduce: MlrvalOrFunction
			reduce(211), // |=, reduce: MlrvalOrFunction
			reduce(211), // &=, reduce: MlrvalOrFunction
			reduce(211), // ^=, reduce: MlrvalOrFunction
			reduce(211), // <<=, reduce: MlrvalOrFunction
			reduce(211), // >>=, reduce: MlrvalOrFunction
			reduce(211), // >>>=, reduce: MlrvalOrFunction
			reduce(211), // +=, reduce: MlrvalOrFunction
			reduce(211), // .=, reduce: MlrvalOrFunction
			reduce(211), // -=, reduce: MlrvalOrFunction
			reduce(211), // *=, reduce: MlrvalOrFunction
			reduce(211), // /=, reduce: MlrvalOrFunction
			reduce(211), // //=, reduce: MlrvalOrFunction
			reduce(211), // %=, reduce: MlrvalOrFunction
			reduce(211), // **=, reduce: MlrvalOrFunction
			reduce(211), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(211), // ||, reduce: MlrvalOrFunction
			reduce(211), // ^^, reduce: MlrvalOrFunction
			reduce(211), // &&, reduce: MlrvalOrFunction
			reduce(211), // ??, reduce: MlrvalOrFunction
			reduce(211), // ???, reduce: MlrvalOrFunction
			reduce(211), // =~, reduce: MlrvalOrFunction
			reduce(211), // !=~, reduce: MlrvalOrFunction
			reduce(211), // ==, reduce: MlrvalOrFunction
			reduce(211), // !=, reduce: MlrvalOrFunction
			reduce(211), // <=>, reduce: MlrvalOrFunction
			reduce(211), // >=, reduce: MlrvalOrFunction
			reduce(211), // <, reduce: MlrvalOrFunction
			reduce(211), // <=, reduce: MlrvalOrFunction
			reduce(211), // ^, reduce: MlrvalOrFunction
			reduce(211), // &, reduce: MlrvalOrFunction
			reduce(211), // <<, reduce: MlrvalOrFunction
			reduce(211), // >>>, reduce: MlrvalOrFunction
			reduce(211), // +, reduce: MlrvalOrFunction
			reduce(211), // -, reduce: MlrvalOrFunction
			reduce(211), // .+, reduce: MlrvalOrFunction
			reduce(211), // .-, reduce: MlrvalOrFunction
			reduce(211), // *, reduce: MlrvalOrFunction
			reduce(211), // /, reduce: MlrvalOrFunction
			reduce(211), // //, reduce: MlrvalOrFunction
			reduce(211), // %, reduce: MlrvalOrFunction
			reduce(211), // .*, reduce: MlrvalOrFunction
			reduce(211), // ./, reduce: MlrvalOrFunction
			reduce(211), // .//, reduce: MlrvalOrFunction
			reduce(211), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(211), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(211), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(211), // [[, reduce: MlrvalOrFunction
			reduce(211), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1769
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(520),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(524),  // (
			nil,         // )
			nil,         // emitp
			shift(536),  // field_name
			shift(537),  // $[
			nil,         // ]
			shift(538),  // braced_field_name
			shift(539),  // $[[
			shift(540),  // $[[[
			shift(541),  // full_srec
			shift(543),  // oosvar_name
			shift(544),  // @[
			shift(545),  // braced_oosvar_name
			shift(546),  // full_oosvar
			shift(547),  // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(563),  // +
			shift(565),  // -
			shift(566),  // .+
			shift(567),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(570),  // !
			shift(571),  // ~
			nil,         // **
			shift(576),  // string_literal
			shift(577),  // regex_case_insensitive
			shift(578),  // int_literal
			shift(579),  // float_literal
			shift(580),  // boolean_literal
			shift(581),  // null_literal
			shift(582),  // inf_literal
			shift(583),  // nan_literal
			shift(584),  // const_M_PI
			shift(585),  // const_M_E
			shift(586),  // panic
			shift(588),  // [
			shift(590),  // ctx_IPS
			shift(591),  // ctx_IFS
			shift(592),  // ctx_IRS
			shift(593),  // ctx_OPS
			shift(594),  // ctx_OFS
			shift(595),  // ctx_ORS
			shift(596),  // ctx_FLATSEP
			shift(597),  // ctx_NF
			shift(598),  // ctx_NR
			shift(599),  // ctx_FNR
			shift(600),  // ctx_FILENAME
			shift(601),  // ctx_FILENUM
			shift(603),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(610),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1770
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(110), // =, reduce: OosvarValue
			nil,         // unset
			nil,         // filter
			reduce(110), // >, reduce: OosvarValue
			reduce(110), // >>, reduce: OosvarValue
			reduce(110), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(110), // ||=, reduce: OosvarValue
			reduce(110), // ^^=, reduce: OosvarValue
			reduce(110), // &&=, reduce: OosvarValue
			reduce(110), // ??=, reduce: OosvarValue
			reduce(110), // ???=, reduce: OosvarValue
			reduce(110), // |=, reduce: OosvarValue
			reduce(110), // &=, reduce: OosvarValue
			reduce(110), // ^=, reduce: OosvarValue
			reduce(110), // <<=, reduce: OosvarValue
			reduce(110), // >>=, reduce: OosvarValue
			reduce(110), // >>>=, reduce: OosvarValue
			reduce(110), // +=, reduce: OosvarValue
			reduce(110), // .=, reduce: OosvarValue
			reduce(110), // -=, reduce: OosvarValue
			reduce(110), // *=, reduce: OosvarValue
			reduce(110), // /=, reduce: OosvarValue
			reduce(110), // //=, reduce: OosvarValue
			reduce(110), // %=, reduce: OosvarValue
			reduce(110), // **=, reduce: OosvarValue
			reduce(110), // ?, reduce: OosvarValue
			nil,         // :
			reduce(110), // ||, reduce: OosvarValue
			reduce(110), // ^^, reduce: OosvarValue
			reduce(110), // &&, reduce: OosvarValue
			reduce(110), // ??, reduce: OosvarValue
			reduce(110), // ???, reduce: OosvarValue
			reduce(110), // =~, reduce: OosvarValue
			reduce(110), // !=~, reduce: OosvarValue
			reduce(110), // ==, reduce: OosvarValue
			reduce(110), // !=, reduce: OosvarValue
			reduce(110), // <=>, reduce: OosvarValue
			reduce(110), // >=, reduce: OosvarValue
			reduce(110), // <, reduce: OosvarValue
			reduce(110), // <=, reduce: OosvarValue
			reduce(110), // ^, reduce: OosvarValue
			reduce(110), // &, reduce: OosvarValue
			reduce(110), // <<, reduce: OosvarValue
			reduce(110), // >>>, reduce: OosvarValue
			reduce(110), // +, reduce: OosvarValue
			reduce(110), // -, reduce: OosvarValue
			reduce(110), // .+, reduce: OosvarValue
			reduce(110), // .-, reduce: OosvarValue
			reduce(110), // *, reduce: OosvarValue
			reduce(110), // /, reduce: OosvarValue
			reduce(110), // //, reduce: OosvarValue
			reduce(110), // %, reduce: OosvarValue
			reduce(110), // .*, reduce: OosvarValue
			reduce(110), // ./, reduce: OosvarValue
			reduce(110), // .//, reduce: OosvarValue
			reduce(110), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(110), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(110), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(110), // [[, reduce: OosvarValue
			reduce(110), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1771
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(112), // =, reduce: OosvarValue
			nil,         // unset
			nil,         // filter
			reduce(112), // >, reduce: OosvarValue
			reduce(112), // >>, reduce: OosvarValue
			reduce(112), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(112), // ||=, reduce: OosvarValue
			reduce(112), // ^^=, reduce: OosvarValue
			reduce(112), // &&=, reduce: OosvarValue
			reduce(112), // ??=, reduce: OosvarValue
			reduce(112), // ???=, reduce: OosvarValue
			reduce(112), // |=, reduce: OosvarValue
			reduce(112), // &=, reduce: OosvarValue
			reduce(112), // ^=, reduce: OosvarValue
			reduce(112), // <<=, reduce: OosvarValue
			reduce(112), // >>=, reduce: OosvarValue
			reduce(112), // >>>=, reduce: OosvarValue
			reduce(112), // +=, reduce: OosvarValue
			reduce(112), // .=, reduce: OosvarValue
			reduce(112), // -=, reduce: OosvarValue
			reduce(112), // *=, reduce: OosvarValue
			reduce(112), // /=, reduce: OosvarValue
			reduce(112), // //=, reduce: OosvarValue
			reduce(112), // %=, reduce: OosvarValue
			reduce(112), // **=, reduce: OosvarValue
			reduce(112), // ?, reduce: OosvarValue
			nil,         // :
			reduce(112), // ||, reduce: OosvarValue
			reduce(112), // ^^, reduce: OosvarValue
			reduce(112), // &&, reduce: OosvarValue
			reduce(112), // ??, reduce: OosvarValue
			reduce(112), // ???, reduce: OosvarValue
			reduce(112), // =~, reduce: OosvarValue
			reduce(112), // !=~, reduce: OosvarValue
			reduce(112), // ==, reduce: OosvarValue
			reduce(112), // !=, reduce: OosvarValue
			reduce(112), // <=>, reduce: OosvarValue
			reduce(112), // >=, reduce: OosvarValue
			reduce(112), // <, reduce: OosvarValue
			reduce(112), // <=, reduce: OosvarValue
			reduce(112), // ^, reduce: OosvarValue
			reduce(112), // &, reduce: OosvarValue
			reduce(112), // <<, reduce: OosvarValue
			reduce(112), // >>>, reduce: OosvarValue
			reduce(112), // +, reduce: OosvarValue
			reduce(112), // -, reduce: OosvarValue
			reduce(112), // .+, reduce: OosvarValue
			reduce(112), // .-, reduce: OosvarValue
			reduce(112), // *, reduce: OosvarValue
			reduce(112), // /, reduce: OosvarValue
			reduce(112), // //, reduce: OosvarValue
			reduce(112), // %, reduce: OosvarValue
			reduce(112), // .*, reduce: OosvarValue
			reduce(112), // ./, reduce: OosvarValue
			reduce(112), // .//, reduce: OosvarValue
			reduce(112), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(112), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(112), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(112), // [[, reduce: OosvarValue
			reduce(112), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1772
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(111), // =, reduce: OosvarValue
			nil,         // unset
			nil,         // filter
			reduce(111), // >, reduce: OosvarValue
			reduce(111), // >>, reduce: OosvarValue
			reduce(111), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(111), // ||=, reduce: OosvarValue
			reduce(111), // ^^=, reduce: OosvarValue
			reduce(111), // &&=, reduce: OosvarValue
			reduce(111), // ??=, reduce: OosvarValue
			reduce(111), // ???=, reduce: OosvarValue
			reduce(111), // |=, reduce: OosvarValue
			reduce(111), // &=, reduce: OosvarValue
			reduce(111), // ^=, reduce: OosvarValue
			reduce(111), // <<=, reduce: OosvarValue
			reduce(111), // >>=, reduce: OosvarValue
			reduce(111), // >>>=, reduce: OosvarValue
			reduce(111), // +=, reduce: OosvarValue
			reduce(111), // .=, reduce: OosvarValue
			reduce(111), // -=, reduce: OosvarValue
			reduce(111), // *=, reduce: OosvarValue
			reduce(111), // /=, reduce: OosvarValue
			reduce(111), // //=, reduce: OosvarValue
			reduce(111), // %=, reduce: OosvarValue
			reduce(111), // **=, reduce: OosvarValue
			reduce(111), // ?, reduce: OosvarValue
			nil,         // :
			reduce(111), // ||, reduce: OosvarValue
			reduce(111), // ^^, reduce: OosvarValue
			reduce(111), // &&, reduce: OosvarValue
			reduce(111), // ??, reduce: OosvarValue
			reduce(111), // ???, reduce: OosvarValue
			reduce(111), // =~, reduce: OosvarValue
			reduce(111), // !=~, reduce: OosvarValue
			reduce(111), // ==, reduce: OosvarValue
			reduce(111), // !=, reduce: OosvarValue
			reduce(111), // <=>, reduce: OosvarValue
			reduce(111), // >=, reduce: OosvarValue
			reduce(111), // <, reduce: OosvarValue
			reduce(111), // <=, reduce: OosvarValue
			reduce(111), // ^, reduce: OosvarValue
			reduce(111), // &, reduce: OosvarValue
			reduce(111), // <<, reduce: OosvarValue
			reduce(111), // >>>, reduce: OosvarValue
			reduce(111), // +, reduce: OosvarValue
			reduce(111), // -, reduce: OosvarValue
			reduce(111), // .+, reduce: OosvarValue
			reduce(111), // .-, reduce: OosvarValue
			reduce(111), // *, reduce: OosvarValue
			reduce(111), // /, reduce: OosvarValue
			reduce(111), // //, reduce: OosvarValue
			reduce(111), // %, reduce: OosvarValue
			reduce(111), // .*, reduce: OosvarValue
			reduce(111), // ./, reduce: OosvarValue
			reduce(111), // .//, reduce: OosvarValue
			reduce(111), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(111), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(111), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(111), // [[, reduce: OosvarValue
			reduce(111), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1773
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			reduce(99), // =, reduce: FieldValue
			nil,        // unset
			nil,        // filter
			reduce(99), // >, reduce: FieldValue
			reduce(99), // >>, reduce: FieldValue
			reduce(99), // |, reduce: FieldValue
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			reduce(99), // ||=, reduce: FieldValue
			reduce(99), // ^^=, reduce: FieldValue
			reduce(99), // &&=, reduce: FieldValue
			reduce(99), // ??=, reduce: FieldValue
			reduce(99), // ???=, reduce: FieldValue
			reduce(99), // |=, reduce: FieldValue
			reduce(99), // &=, reduce: FieldValue
			reduce(99), // ^=, reduce: FieldValue
			reduce(99), // <<=, reduce: FieldValue
			reduce(99), // >>=, reduce: FieldValue
			reduce(99), // >>>=, reduce: FieldValue
			reduce(99), // +=, reduce: FieldValue
			reduce(99), // .=, reduce: FieldValue
			reduce(99), // -=, reduce: FieldValue
			reduce(99), // *=, reduce: FieldValue
			reduce(99), // /=, reduce: FieldValue
			reduce(99), // //=, reduce: FieldValue
			reduce(99), // %=, reduce: FieldValue
			reduce(99), // **=, reduce: FieldValue
			reduce(99), // ?, reduce: FieldValue
			nil,        // :
			reduce(99), // ||, reduce: FieldValue
			reduce(99), // ^^, reduce: FieldValue
			reduce(99), // &&, reduce: FieldValue
			reduce(99), // ??, reduce: FieldValue
			reduce(99), // ???, reduce: FieldValue
			reduce(99), // =~, reduce: FieldValue
			reduce(99), // !=~, reduce: FieldValue
			reduce(99), // ==, reduce: FieldValue
			reduce(99), // !=, reduce: FieldValue
			reduce(99), // <=>, reduce: FieldValue
			reduce(99), // >=, reduce: FieldValue
			reduce(99), // <, reduce: FieldValue
			reduce(99), // <=, reduce: FieldValue
			reduce(99), // ^, reduce: FieldValue
			reduce(99), // &, reduce: FieldValue
			reduce(99), // <<, reduce: FieldValue
			reduce(99), // >>>, reduce: FieldValue
			reduce(99), // +, reduce: FieldValue
			reduce(99), // -, reduce: FieldValue
			reduce(99), // .+, reduce: FieldValue
			reduce(99), // .-, reduce: FieldValue
			reduce(99), // *, reduce: FieldValue
			reduce(99), // /, reduce: FieldValue
			reduce(99), // //, reduce: FieldValue
			reduce(99), // %, reduce: FieldValue
			reduce(99), // .*, reduce: FieldValue
			reduce(99), // ./, reduce: FieldValue
			reduce(99), // .//, reduce: FieldValue
			reduce(99), // ., reduce: FieldValue
			nil,        // !
			nil,        // ~
			reduce(99), // **, reduce: FieldValue
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(99), // [, reduce: FieldValue
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(99), // [[, reduce: FieldValue
			reduce(99), // [[[, reduce: FieldValue
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1774
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(101), // =, reduce: FieldValue
			nil,         // unset
			nil,         // filter
			reduce(101), // >, reduce: FieldValue
			reduce(101), // >>, reduce: FieldValue
			reduce(101), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(101), // ||=, reduce: FieldValue
			reduce(101), // ^^=, reduce: FieldValue
			reduce(101), // &&=, reduce: FieldValue
			reduce(101), // ??=, reduce: FieldValue
			reduce(101), // ???=, reduce: FieldValue
			reduce(101), // |=, reduce: FieldValue
			reduce(101), // &=, reduce: FieldValue
			reduce(101), // ^=, reduce: FieldValue
			reduce(101), // <<=, reduce: FieldValue
			reduce(101), // >>=, reduce: FieldValue
			reduce(101), // >>>=, reduce: FieldValue
			reduce(101), // +=, reduce: FieldValue
			reduce(101), // .=, reduce: FieldValue
			reduce(101), // -=, reduce: FieldValue
			reduce(101), // *=, reduce: FieldValue
			reduce(101), // /=, reduce: FieldValue
			reduce(101), // //=, reduce: FieldValue
			reduce(101), // %=, reduce: FieldValue
			reduce(101), // **=, reduce: FieldValue
			reduce(101), // ?, reduce: FieldValue
			nil,         // :
			reduce(101), // ||, reduce: FieldValue
			reduce(101), // ^^, reduce: FieldValue
			reduce(101), // &&, reduce: FieldValue
			reduce(101), // ??, reduce: FieldValue
			reduce(101), // ???, reduce: FieldValue
			reduce(101), // =~, reduce: FieldValue
			reduce(101), // !=~, reduce: FieldValue
			reduce(101), // ==, reduce: FieldValue
			reduce(101), // !=, reduce: FieldValue
			reduce(101), // <=>, reduce: FieldValue
			reduce(101), // >=, reduce: FieldValue
			reduce(101), // <, reduce: FieldValue
			reduce(101), // <=, reduce: FieldValue
			reduce(101), // ^, reduce: FieldValue
			reduce(101), // &, reduce: FieldValue
			reduce(101), // <<, reduce: FieldValue
			reduce(101), // >>>, reduce: FieldValue
			reduce(101), // +, reduce: FieldValue
			reduce(101), // -, reduce: FieldValue
			reduce(101), // .+, reduce: FieldValue
			reduce(101), // .-, reduce: FieldValue
			reduce(101), // *, reduce: FieldValue
			reduce(101), // /, reduce: FieldValue
			reduce(101), // //, reduce: FieldValue
			reduce(101), // %, reduce: FieldValue
			reduce(101), // .*, reduce: FieldValue
			reduce(101), // ./, reduce: FieldValue
			reduce(101), // .//, reduce: FieldValue
			reduce(101), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(101), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(101), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(101), // [[, reduce: FieldValue
			reduce(101), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1775
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(100), // =, reduce: FieldValue
			nil,         // unset
			nil,         // filter
			reduce(100), // >, reduce: FieldValue
			reduce(100), // >>, reduce: FieldValue
			reduce(100), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(100), // ||=, reduce: FieldValue
			reduce(100), // ^^=, reduce: FieldValue
			reduce(100), // &&=, reduce: FieldValue
			reduce(100), // ??=, reduce: FieldValue
			reduce(100), // ???=, reduce: FieldValue
			reduce(100), // |=, reduce: FieldValue
			reduce(100), // &=, reduce: FieldValue
			reduce(100), // ^=, reduce: FieldValue
			reduce(100), // <<=, reduce: FieldValue
			reduce(100), // >>=, reduce: FieldValue
			reduce(100), // >>>=, reduce: FieldValue
			reduce(100), // +=, reduce: FieldValue
			reduce(100), // .=, reduce: FieldValue
			reduce(100), // -=, reduce: FieldValue
			reduce(100), // *=, reduce: FieldValue
			reduce(100), // /=, reduce: FieldValue
			reduce(100), // //=, reduce: FieldValue
			reduce(100), // %=, reduce: FieldValue
			reduce(100), // **=, reduce: FieldValue
			reduce(100), // ?, reduce: FieldValue
			nil,         // :
			reduce(100), // ||, reduce: FieldValue
			reduce(100), // ^^, reduce: FieldValue
			reduce(100), // &&, reduce: FieldValue
			reduce(100), // ??, reduce: FieldValue
			reduce(100), // ???, reduce: FieldValue
			reduce(100), // =~, reduce: FieldValue
			reduce(100), // !=~, reduce: FieldValue
			reduce(100), // ==, reduce: FieldValue
			reduce(100), // !=, reduce: FieldValue
			reduce(100), // <=>, reduce: FieldValue
			reduce(100), // >=, reduce: FieldValue
			reduce(100), // <, reduce: FieldValue
			reduce(100), // <=, reduce: FieldValue
			reduce(100), // ^, reduce: FieldValue
			reduce(100), // &, reduce: FieldValue
			reduce(100), // <<, reduce: FieldValue
			reduce(100), // >>>, reduce: FieldValue
			reduce(100), // +, reduce: FieldValue
			reduce(100), // -, reduce: FieldValue
			reduce(100), // .+, reduce: FieldValue
			reduce(100), // .-, reduce: FieldValue
			reduce(100), // *, reduce: FieldValue
			reduce(100), // /, reduce: FieldValue
			reduce(100), // //, reduce: FieldValue
			reduce(100), // %, reduce: FieldValue
			reduce(100), // .*, reduce: FieldValue
			reduce(100), // ./, reduce: FieldValue
			reduce(100), // .//, reduce: FieldValue
			reduce(100), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(100), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(100), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(100), // [[, reduce: FieldValue
			reduce(100), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1776
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(213), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(213), // >, reduce: MlrvalOrFunction
			reduce(213), // >>, reduce: MlrvalOrFunction
			reduce(213), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(213), // ||=, reduce: MlrvalOrFunction
			reduce(213), // ^^=, reduce: MlrvalOrFunction
			reduce(213), // &&=, reduce: MlrvalOrFunction
			reduce(213), // ??=, reduce: MlrvalOrFunction
			reduce(213), // ???=, reduce: MlrvalOrFunction
			reduce(213), // |=, reduce: MlrvalOrFunction
			reduce(213), // &=, reduce: MlrvalOrFunction
			reduce(213), // ^=, reduce: MlrvalOrFunction
			reduce(213), // <<=, reduce: MlrvalOrFunction
			reduce(213), // >>=, reduce: MlrvalOrFunction
			reduce(213), // >>>=, reduce: MlrvalOrFunction
			reduce(213), // +=, reduce: MlrvalOrFunction
			reduce(213), // .=, reduce: MlrvalOrFunction
			reduce(213), // -=, reduce: MlrvalOrFunction
			reduce(213), // *=, reduce: MlrvalOrFunction
			reduce(213), // /=, reduce: MlrvalOrFunction
			reduce(213), // //=, reduce: MlrvalOrFunction
			reduce(213), // %=, reduce: MlrvalOrFunction
			reduce(213), // **=, reduce: MlrvalOrFunction
			reduce(213), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(213), // ||, reduce: MlrvalOrFunction
			reduce(213), // ^^, reduce: MlrvalOrFunction
			reduce(213), // &&, reduce: MlrvalOrFunction
			reduce(213), // ??, reduce: MlrvalOrFunction
			reduce(213), // ???, reduce: MlrvalOrFunction
			reduce(213), // =~, reduce: MlrvalOrFunction
			reduce(213), // !=~, reduce: MlrvalOrFunction
			reduce(213), // ==, reduce: MlrvalOrFunction
			reduce(213), // !=, reduce: MlrvalOrFunction
			reduce(213), // <=>, reduce: MlrvalOrFunction
			reduce(213), // >=, reduce: MlrvalOrFunction
			reduce(213), // <, reduce: MlrvalOrFunction
			reduce(213), // <=, reduce: MlrvalOrFunction
			reduce(213), // ^, reduce: MlrvalOrFunction
			reduce(213), // &, reduce: MlrvalOrFunction
			reduce(213), // <<, reduce: MlrvalOrFunction
			reduce(213), // >>>, reduce: MlrvalOrFunction
			reduce(213), // +, reduce: MlrvalOrFunction
			reduce(213), // -, reduce: MlrvalOrFunction
			reduce(213), // .+, reduce: MlrvalOrFunction
			reduce(213), // .-, reduce: MlrvalOrFunction
			reduce(213), // *, reduce: MlrvalOrFunction
			reduce(213), // /, reduce: MlrvalOrFunction
			reduce(213), // //, reduce: MlrvalOrFunction
			reduce(213), // %, reduce: MlrvalOrFunction
			reduce(213), // .*, reduce: MlrvalOrFunction
			reduce(213), // ./, reduce: MlrvalOrFunction
			reduce(213), // .//, reduce: MlrvalOrFunction
			reduce(213), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(213), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(213), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(213), // [[, reduce: MlrvalOrFunction
			reduce(213), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1777
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(233), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(233), // >, reduce: MlrvalOrFunction
			reduce(233), // >>, reduce: MlrvalOrFunction
			reduce(233), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(233), // ||=, reduce: MlrvalOrFunction
			reduce(233), // ^^=, reduce: MlrvalOrFunction
			reduce(233), // &&=, reduce: MlrvalOrFunction
			reduce(233), // ??=, reduce: MlrvalOrFunction
			reduce(233), // ???=, reduce: MlrvalOrFunction
			reduce(233), // |=, reduce: MlrvalOrFunction
			reduce(233), // &=, reduce: MlrvalOrFunction
			reduce(233), // ^=, reduce: MlrvalOrFunction
			reduce(233), // <<=, reduce: MlrvalOrFunction
			reduce(233), // >>=, reduce: MlrvalOrFunction
			reduce(233), // >>>=, reduce: MlrvalOrFunction
			reduce(233), // +=, reduce: MlrvalOrFunction
			reduce(233), // .=, reduce: MlrvalOrFunction
			reduce(233), // -=, reduce: MlrvalOrFunction
			reduce(233), // *=, reduce: MlrvalOrFunction
			reduce(233), // /=, reduce: MlrvalOrFunction
			reduce(233), // //=, reduce: MlrvalOrFunction
			reduce(233), // %=, reduce: MlrvalOrFunction
			reduce(233), // **=, reduce: MlrvalOrFunction
			reduce(233), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(233), // ||, reduce: MlrvalOrFunction
			reduce(233), // ^^, reduce: MlrvalOrFunction
			reduce(233), // &&, reduce: MlrvalOrFunction
			reduce(233), // ??, reduce: MlrvalOrFunction
			reduce(233), // ???, reduce: MlrvalOrFunction
			reduce(233), // =~, reduce: MlrvalOrFunction
			reduce(233), // !=~, reduce: MlrvalOrFunction
			reduce(233), // ==, reduce: MlrvalOrFunction
			reduce(233), // !=, reduce: MlrvalOrFunction
			reduce(233), // <=>, reduce: MlrvalOrFunction
			reduce(233), // >=, reduce: MlrvalOrFunction
			reduce(233), // <, reduce: MlrvalOrFunction
			reduce(233), // <=, reduce: MlrvalOrFunction
			reduce(233), // ^, reduce: MlrvalOrFunction
			reduce(233), // &, reduce: MlrvalOrFunction
			reduce(233), // <<, reduce: MlrvalOrFunction
			reduce(233), // >>>, reduce: MlrvalOrFunction
			reduce(233), // +, reduce: MlrvalOrFunction
			reduce(233), // -, reduce: MlrvalOrFunction
			reduce(233), // .+, reduce: MlrvalOrFunction
			reduce(233), // .-, reduce: MlrvalOrFunction
			reduce(233), // *, reduce: MlrvalOrFunction
			reduce(233), // /, reduce: MlrvalOrFunction
			reduce(233), // //, reduce: MlrvalOrFunction
			reduce(233), // %, reduce: MlrvalOrFunction
			reduce(233), // .*, reduce: MlrvalOrFunction
			reduce(233), // ./, reduce: MlrvalOrFunction
			reduce(233), // .//, reduce: MlrvalOrFunction
			reduce(233), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(233), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(233), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(233), // [[, reduce: MlrvalOrFunction
			reduce(233), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1778
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(210), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(210), // >, reduce: MlrvalOrFunction
			reduce(210), // >>, reduce: MlrvalOrFunction
			reduce(210), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(210), // ||=, reduce: MlrvalOrFunction
			reduce(210), // ^^=, reduce: MlrvalOrFunction
			reduce(210), // &&=, reduce: MlrvalOrFunction
			reduce(210), // ??=, reduce: MlrvalOrFunction
			reduce(210), // ???=, reduce: MlrvalOrFunction
			reduce(210), // |=, reduce: MlrvalOrFunction
			reduce(210), // &=, reduce: MlrvalOrFunction
			reduce(210), // ^=, reduce: MlrvalOrFunction
			reduce(210), // <<=, reduce: MlrvalOrFunction
			reduce(210), // >>=, reduce: MlrvalOrFunction
			reduce(210), // >>>=, reduce: MlrvalOrFunction
			reduce(210), // +=, reduce: MlrvalOrFunction
			reduce(210), // .=, reduce: MlrvalOrFunction
			reduce(210), // -=, reduce: MlrvalOrFunction
			reduce(210), // *=, reduce: MlrvalOrFunction
			reduce(210), // /=, reduce: MlrvalOrFunction
			reduce(210), // //=, reduce: MlrvalOrFunction
			reduce(210), // %=, reduce: MlrvalOrFunction
			reduce(210), // **=, reduce: MlrvalOrFunction
			reduce(210), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(210), // ||, reduce: MlrvalOrFunction
			reduce(210), // ^^, reduce: MlrvalOrFunction
			reduce(210), // &&, reduce: MlrvalOrFunction
			reduce(210), // ??, reduce: MlrvalOrFunction
			reduce(210), // ???, reduce: MlrvalOrFunction
			reduce(210), // =~, reduce: MlrvalOrFunction
			reduce(210), // !=~, reduce: MlrvalOrFunction
			reduce(210), // ==, reduce: MlrvalOrFunction
			reduce(210), // !=, reduce: MlrvalOrFunction
			reduce(210), // <=>, reduce: MlrvalOrFunction
			reduce(210), // >=, reduce: MlrvalOrFunction
			reduce(210), // <, reduce: MlrvalOrFunction
			reduce(210), // <=, reduce: MlrvalOrFunction
			reduce(210), // ^, reduce: MlrvalOrFunction
			reduce(210), // &, reduce: MlrvalOrFunction
			reduce(210), // <<, reduce: MlrvalOrFunction
			reduce(210), // >>>, reduce: MlrvalOrFunction
			reduce(210), // +, reduce: MlrvalOrFunction
			reduce(210), // -, reduce: MlrvalOrFunction
			reduce(210), // .+, reduce: MlrvalOrFunction
			reduce(210), // .-, reduce: MlrvalOrFunction
			reduce(210), // *, reduce: MlrvalOrFunction
			reduce(210), // /, reduce: MlrvalOrFunction
			reduce(210), // //, reduce: MlrvalOrFunction
			reduce(210), // %, reduce: MlrvalOrFunction
			reduce(210), // .*, reduce: MlrvalOrFunction
			reduce(210), // ./, reduce: MlrvalOrFunction
			reduce(210), // .//, reduce: MlrvalOrFunction
			reduce(210), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(210), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(210), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(210), // [[, reduce: MlrvalOrFunction
			reduce(210), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1779
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(102), // =, reduce: FieldValue
			nil,         // unset
			nil,         // filter
			reduce(102), // >, reduce: FieldValue
			reduce(102), // >>, reduce: FieldValue
			reduce(102), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(102), // ||=, reduce: FieldValue
			reduce(102), // ^^=, reduce: FieldValue
			reduce(102), // &&=, reduce: FieldValue
			reduce(102), // ??=, reduce: FieldValue
			reduce(102), // ???=, reduce: FieldValue
			reduce(102), // |=, reduce: FieldValue
			reduce(102), // &=, reduce: FieldValue
			reduce(102), // ^=, reduce: FieldValue
			reduce(102), // <<=, reduce: FieldValue
			reduce(102), // >>=, reduce: FieldValue
			reduce(102), // >>>=, reduce: FieldValue
			reduce(102), // +=, reduce: FieldValue
			reduce(102), // .=, reduce: FieldValue
			reduce(102), // -=, reduce: FieldValue
			reduce(102), // *=, reduce: FieldValue
			reduce(102), // /=, reduce: FieldValue
			reduce(102), // //=, reduce: FieldValue
			reduce(102), // %=, reduce: FieldValue
			reduce(102), // **=, reduce: FieldValue
			reduce(102), // ?, reduce: FieldValue
			nil,         // :
			reduce(102), // ||, reduce: FieldValue
			reduce(102), // ^^, reduce: FieldValue
			reduce(102), // &&, reduce: FieldValue
			reduce(102), // ??, reduce: FieldValue
			reduce(102), // ???, reduce: FieldValue
			reduce(102), // =~, reduce: FieldValue
			reduce(102), // !=~, reduce: FieldValue
			reduce(102), // ==, reduce: FieldValue
			reduce(102), // !=, reduce: FieldValue
			reduce(102), // <=>, reduce: FieldValue
			reduce(102), // >=, reduce: FieldValue
			reduce(102), // <, reduce: FieldValue
			reduce(102), // <=, reduce: FieldValue
			reduce(102), // ^, reduce: FieldValue
			reduce(102), // &, reduce: FieldValue
			reduce(102), // <<, reduce: FieldValue
			reduce(102), // >>>, reduce: FieldValue
			reduce(102), // +, reduce: FieldValue
			reduce(102), // -, reduce: FieldValue
			reduce(102), // .+, reduce: FieldValue
			reduce(102), // .-, reduce: FieldValue
			reduce(102), // *, reduce: FieldValue
			reduce(102), // /, reduce: FieldValue
			reduce(102), // //, reduce: FieldValue
			reduce(102), // %, reduce: FieldValue
			reduce(102), // .*, reduce: FieldValue
			reduce(102), // ./, reduce: FieldValue
			reduce(102), // .//, reduce: FieldValue
			reduce(102), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(102), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(102), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(102), // [[, reduce: FieldValue
			reduce(102), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1780
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(103), // =, reduce: FieldValue
			nil,         // unset
			nil,         // filter
			reduce(103), // >, reduce: FieldValue
			reduce(103), // >>, reduce: FieldValue
			reduce(103), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(103), // ||=, reduce: FieldValue
			reduce(103), // ^^=, reduce: FieldValue
			reduce(103), // &&=, reduce: FieldValue
			reduce(103), // ??=, reduce: FieldValue
			reduce(103), // ???=, reduce: FieldValue
			reduce(103), // |=, reduce: FieldValue
			reduce(103), // &=, reduce: FieldValue
			reduce(103), // ^=, reduce: FieldValue
			reduce(103), // <<=, reduce: FieldValue
			reduce(103), // >>=, reduce: FieldValue
			reduce(103), // >>>=, reduce: FieldValue
			reduce(103), // +=, reduce: FieldValue
			reduce(103), // .=, reduce: FieldValue
			reduce(103), // -=, reduce: FieldValue
			reduce(103), // *=, reduce: FieldValue
			reduce(103), // /=, reduce: FieldValue
			reduce(103), // //=, reduce: FieldValue
			reduce(103), // %=, reduce: FieldValue
			reduce(103), // **=, reduce: FieldValue
			reduce(103), // ?, reduce: FieldValue
			nil,         // :
			reduce(103), // ||, reduce: FieldValue
			reduce(103), // ^^, reduce: FieldValue
			reduce(103), // &&, reduce: FieldValue
			reduce(103), // ??, reduce: FieldValue
			reduce(103), // ???, reduce: FieldValue
			reduce(103), // =~, reduce: FieldValue
			reduce(103), // !=~, reduce: FieldValue
			reduce(103), // ==, reduce: FieldValue
			reduce(103), // !=, reduce: FieldValue
			reduce(103), // <=>, reduce: FieldValue
			reduce(103), // >=, reduce: FieldValue
			reduce(103), // <, reduce: FieldValue
			reduce(103), // <=, reduce: FieldValue
			reduce(103), // ^, reduce: FieldValue
			reduce(103), // &, reduce: FieldValue
			reduce(103), // <<, reduce: FieldValue
			reduce(103), // >>>, reduce: FieldValue
			reduce(103), // +, reduce: FieldValue
			reduce(103), // -, reduce: FieldValue
			reduce(103), // .+, reduce: FieldValue
			reduce(103), // .-, reduce: FieldValue
			reduce(103), // *, reduce: FieldValue
			reduce(103), // /, reduce: FieldValue
			reduce(103), // //, reduce: FieldValue
			reduce(103), // %, reduce: FieldValue
			reduce(103), // .*, reduce: FieldValue
			reduce(103), // ./, reduce: FieldValue
			reduce(103), // .//, reduce: FieldValue
			reduce(103), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(103), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(103), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(103), // [[, reduce: FieldValue
			reduce(103), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1781
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(104), // =, reduce: DirectFieldValue
			nil,         // unset
			nil,         // filter
			reduce(104), // >, reduce: DirectFieldValue
			reduce(104), // >>, reduce: DirectFieldValue
			reduce(104), // |, reduce: DirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(104), // ||=, reduce: DirectFieldValue
			reduce(104), // ^^=, reduce: DirectFieldValue
			reduce(104), // &&=, reduce: DirectFieldValue
			reduce(104), // ??=, reduce: DirectFieldValue
			reduce(104), // ???=, reduce: DirectFieldValue
			reduce(104), // |=, reduce: DirectFieldValue
			reduce(104), // &=, reduce: DirectFieldValue
			reduce(104), // ^=, reduce: DirectFieldValue
			reduce(104), // <<=, reduce: DirectFieldValue
			reduce(104), // >>=, reduce: DirectFieldValue
			reduce(104), // >>>=, reduce: DirectFieldValue
			reduce(104), // +=, reduce: DirectFieldValue
			reduce(104), // .=, reduce: DirectFieldValue
			reduce(104), // -=, reduce: DirectFieldValue
			reduce(104), // *=, reduce: DirectFieldValue
			reduce(104), // /=, reduce: DirectFieldValue
			reduce(104), // //=, reduce: DirectFieldValue
			reduce(104), // %=, reduce: DirectFieldValue
			reduce(104), // **=, reduce: DirectFieldValue
			reduce(104), // ?, reduce: DirectFieldValue
			nil,         // :
			reduce(104), // ||, reduce: DirectFieldValue
			reduce(104), // ^^, reduce: DirectFieldValue
			reduce(104), // &&, reduce: DirectFieldValue
			reduce(104), // ??, reduce: DirectFieldValue
			reduce(104), // ???, reduce: DirectFieldValue
			reduce(104), // =~, reduce: DirectFieldValue
			reduce(104), // !=~, reduce: DirectFieldValue
			reduce(104), // ==, reduce: DirectFieldValue
			reduce(104), // !=, reduce: DirectFieldValue
			reduce(104), // <=>, reduce: DirectFieldValue
			reduce(104), // >=, reduce: DirectFieldValue
			reduce(104), // <, reduce: DirectFieldValue
			reduce(104), // <=, reduce: DirectFieldValue
			reduce(104), // ^, reduce: DirectFieldValue
			reduce(104), // &, reduce: DirectFieldValue
			reduce(104), // <<, reduce: DirectFieldValue
			reduce(104), // >>>, reduce: DirectFieldValue
			reduce(104), // +, reduce: DirectFieldValue
			reduce(104), // -, reduce: DirectFieldValue
			reduce(104), // .+, reduce: DirectFieldValue
			reduce(104), // .-, reduce: DirectFieldValue
			reduce(104), // *, reduce: DirectFieldValue
			reduce(104), // /, reduce: DirectFieldValue
			reduce(104), // //, reduce: DirectFieldValue
			reduce(104), // %, reduce: DirectFieldValue
			reduce(104), // .*, reduce: DirectFieldValue
			reduce(104), // ./, reduce: DirectFieldValue
			reduce(104), // .//, reduce: DirectFieldValue
			reduce(104), // ., reduce: DirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(104), // **, reduce: DirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(104), // [, reduce: DirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(104), // [[, reduce: DirectFieldValue
			reduce(104), // [[[, reduce: DirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1782
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1783
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(106), // =, reduce: BracedFieldValue
			nil,         // unset
			nil,         // filter
			reduce(106), // >, reduce: BracedFieldValue
			reduce(106), // >>, reduce: BracedFieldValue
			reduce(106), // |, reduce: BracedFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(106), // ||=, reduce: BracedFieldValue
			reduce(106), // ^^=, reduce: BracedFieldValue
			reduce(106), // &&=, reduce: BracedFieldValue
			reduce(106), // ??=, reduce: BracedFieldValue
			reduce(106), // ???=, reduce: BracedFieldValue
			reduce(106), // |=, reduce: BracedFieldValue
			reduce(106), // &=, reduce: BracedFieldValue
			reduce(106), // ^=, reduce: BracedFieldValue
			reduce(106), // <<=, reduce: BracedFieldValue
			reduce(106), // >>=, reduce: BracedFieldValue
			reduce(106), // >>>=, reduce: BracedFieldValue
			reduce(106), // +=, reduce: BracedFieldValue
			reduce(106), // .=, reduce: BracedFieldValue
			reduce(106), // -=, reduce: BracedFieldValue
			reduce(106), // *=, reduce: BracedFieldValue
			reduce(106), // /=, reduce: BracedFieldValue
			reduce(106), // //=, reduce: BracedFieldValue
			reduce(106), // %=, reduce: BracedFieldValue
			reduce(106), // **=, reduce: BracedFieldValue
			reduce(106), // ?, reduce: BracedFieldValue
			nil,         // :
			reduce(106), // ||, reduce: BracedFieldValue
			reduce(106), // ^^, reduce: BracedFieldValue
			reduce(106), // &&, reduce: BracedFieldValue
			reduce(106), // ??, reduce: BracedFieldValue
			reduce(106), // ???, reduce: BracedFieldValue
			reduce(106), // =~, reduce: BracedFieldValue
			reduce(106), // !=~, reduce: BracedFieldValue
			reduce(106), // ==, reduce: BracedFieldValue
			reduce(106), // !=, reduce: BracedFieldValue
			reduce(106), // <=>, reduce: BracedFieldValue
			reduce(106), // >=, reduce: BracedFieldValue
			reduce(106), // <, reduce: BracedFieldValue
			reduce(106), // <=, reduce: BracedFieldValue
			reduce(106), // ^, reduce: BracedFieldValue
			reduce(106), // &, reduce: BracedFieldValue
			reduce(106), // <<, reduce: BracedFieldValue
			reduce(106), // >>>, reduce: BracedFieldValue
			reduce(106), // +, reduce: BracedFieldValue
			reduce(106), // -, reduce: BracedFieldValue
			reduce(106), // .+, reduce: BracedFieldValue
			reduce(106), // .-, reduce: BracedFieldValue
			reduce(106), // *, reduce: BracedFieldValue
			reduce(106), // /, reduce: BracedFieldValue
			reduce(106), // //, reduce: BracedFieldValue
			reduce(106), // %, reduce: BracedFieldValue
			reduce(106), // .*, reduce: BracedFieldValue
			reduce(106), // ./, reduce: BracedFieldValue
			reduce(106), // .//, reduce: BracedFieldValue
			reduce(106), // ., reduce: BracedFieldValue
			nil,         // !
			nil,         // ~
			reduce(106), // **, reduce: BracedFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(106), // [, reduce: BracedFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(106), // [[, reduce: BracedFieldValue
			reduce(106), // [[[, reduce: BracedFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1784
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1785
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1786
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(109), // =, reduce: FullSrec
			nil,         // unset
			nil,         // filter
			reduce(109), // >, reduce: FullSrec
			reduce(109), // >>, reduce: FullSrec
			reduce(109), // |, reduce: FullSrec
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(109), // ||=, reduce: FullSrec
			reduce(109), // ^^=, reduce: FullSrec
			reduce(109), // &&=, reduce: FullSrec
			reduce(109), // ??=, reduce: FullSrec
			reduce(109), // ???=, reduce: FullSrec
			reduce(109), // |=, reduce: FullSrec
			reduce(109), // &=, reduce: FullSrec
			reduce(109), // ^=, reduce: FullSrec
			reduce(109), // <<=, reduce: FullSrec
			reduce(109), // >>=, reduce: FullSrec
			reduce(109), // >>>=, reduce: FullSrec
			reduce(109), // +=, reduce: FullSrec
			reduce(109), // .=, reduce: FullSrec
			reduce(109), // -=, reduce: FullSrec
			reduce(109), // *=, reduce: FullSrec
			reduce(109), // /=, reduce: FullSrec
			reduce(109), // //=, reduce: FullSrec
			reduce(109), // %=, reduce: FullSrec
			reduce(109), // **=, reduce: FullSrec
			reduce(109), // ?, reduce: FullSrec
			nil,         // :
			reduce(109), // ||, reduce: FullSrec
			reduce(109), // ^^, reduce: FullSrec
			reduce(109), // &&, reduce: FullSrec
			reduce(109), // ??, reduce: FullSrec
			reduce(109), // ???, reduce: FullSrec
			reduce(109), // =~, reduce: FullSrec
			reduce(109), // !=~, reduce: FullSrec
			reduce(109), // ==, reduce: FullSrec
			reduce(109), // !=, reduce: FullSrec
			reduce(109), // <=>, reduce: FullSrec
			reduce(109), // >=, reduce: FullSrec
			reduce(109), // <, reduce: FullSrec
			reduce(109), // <=, reduce: FullSrec
			reduce(109), // ^, reduce: FullSrec
			reduce(109), // &, reduce: FullSrec
			reduce(109), // <<, reduce: FullSrec
			reduce(109), // >>>, reduce: FullSrec
			reduce(109), // +, reduce: FullSrec
			reduce(109), // -, reduce: FullSrec
			reduce(109), // .+, reduce: FullSrec
			reduce(109), // .-, reduce: FullSrec
			reduce(109), // *, reduce: FullSrec
			reduce(109), // /, reduce: FullSrec
			reduce(109), // //, reduce: FullSrec
			reduce(109), // %, reduce: FullSrec
			reduce(109), // .*, reduce: FullSrec
			reduce(109), // ./, reduce: FullSrec
			reduce(109), // .//, reduce: FullSrec
			reduce(109), // ., reduce: FullSrec
			nil,         // !
			nil,         // ~
			reduce(109), // **, reduce: FullSrec
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(109), // [, reduce: FullSrec
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(109), // [[, reduce: FullSrec
			reduce(109), // [[[, reduce: FullSrec
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1787
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(212), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(212), // >, reduce: MlrvalOrFunction
			reduce(212), // >>, reduce: MlrvalOrFunction
			reduce(212), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(212), // ||=, reduce: MlrvalOrFunction
			reduce(212), // ^^=, reduce: MlrvalOrFunction
			reduce(212), // &&=, reduce: MlrvalOrFunction
			reduce(212), // ??=, reduce: MlrvalOrFunction
			reduce(212), // ???=, reduce: MlrvalOrFunction
			reduce(212), // |=, reduce: MlrvalOrFunction
			reduce(212), // &=, reduce: MlrvalOrFunction
			reduce(212), // ^=, reduce: MlrvalOrFunction
			reduce(212), // <<=, reduce: MlrvalOrFunction
			reduce(212), // >>=, reduce: MlrvalOrFunction
			reduce(212), // >>>=, reduce: MlrvalOrFunction
			reduce(212), // +=, reduce: MlrvalOrFunction
			reduce(212), // .=, reduce: MlrvalOrFunction
			reduce(212), // -=, reduce: MlrvalOrFunction
			reduce(212), // *=, reduce: MlrvalOrFunction
			reduce(212), // /=, reduce: MlrvalOrFunction
			reduce(212), // //=, reduce: MlrvalOrFunction
			reduce(212), // %=, reduce: MlrvalOrFunction
			reduce(212), // **=, reduce: MlrvalOrFunction
			reduce(212), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(212), // ||, reduce: MlrvalOrFunction
			reduce(212), // ^^, reduce: MlrvalOrFunction
			reduce(212), // &&, reduce: MlrvalOrFunction
			reduce(212), // ??, reduce: MlrvalOrFunction
			reduce(212), // ???, reduce: MlrvalOrFunction
			reduce(212), // =~, reduce: MlrvalOrFunction
			reduce(212), // !=~, reduce: MlrvalOrFunction
			reduce(212), // ==, reduce: MlrvalOrFunction
			reduce(212), // !=, reduce: MlrvalOrFunction
			reduce(212), // <=>, reduce: MlrvalOrFunction
			reduce(212), // >=, reduce: MlrvalOrFunction
			reduce(212), // <, reduce: MlrvalOrFunction
			reduce(212), // <=, reduce: MlrvalOrFunction
			reduce(212), // ^, reduce: MlrvalOrFunction
			reduce(212), // &, reduce: MlrvalOrFunction
			reduce(212), // <<, reduce: MlrvalOrFunction
			reduce(212), // >>>, reduce: MlrvalOrFunction
			reduce(212), // +, reduce: MlrvalOrFunction
			reduce(212), // -, reduce: MlrvalOrFunction
			reduce(212), // .+, reduce: MlrvalOrFunction
			reduce(212), // .-, reduce: MlrvalOrFunction
			reduce(212), // *, reduce: MlrvalOrFunction
			reduce(212), // /, reduce: MlrvalOrFunction
			reduce(212), // //, reduce: MlrvalOrFunction
			reduce(212), // %, reduce: MlrvalOrFunction
			reduce(212), // .*, reduce: MlrvalOrFunction
			reduce(212), // ./, reduce: MlrvalOrFunction
			reduce(212), // .//, reduce: MlrvalOrFunction
			reduce(212), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(212), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(212), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(212), // [[, reduce: MlrvalOrFunction
			reduce(212), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1788
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(113), // =, reduce: DirectOosvarValue
			nil,         // unset
			nil,         // filter
			reduce(113), // >, reduce: DirectOosvarValue
			reduce(113), // >>, reduce: DirectOosvarValue
			reduce(113), // |, reduce: DirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(113), // ||=, reduce: DirectOosvarValue
			reduce(113), // ^^=, reduce: DirectOosvarValue
			reduce(113), // &&=, reduce: DirectOosvarValue
			reduce(113), // ??=, reduce: DirectOosvarValue
			reduce(113), // ???=, reduce: DirectOosvarValue
			reduce(113), // |=, reduce: DirectOosvarValue
			reduce(113), // &=, reduce: DirectOosvarValue
			reduce(113), // ^=, reduce: DirectOosvarValue
			reduce(113), // <<=, reduce: DirectOosvarValue
			reduce(113), // >>=, reduce: DirectOosvarValue
			reduce(113), // >>>=, reduce: DirectOosvarValue
			reduce(113), // +=, reduce: DirectOosvarValue
			reduce(113), // .=, reduce: DirectOosvarValue
			reduce(113), // -=, reduce: DirectOosvarValue
			reduce(113), // *=, reduce: DirectOosvarValue
			reduce(113), // /=, reduce: DirectOosvarValue
			reduce(113), // //=, reduce: DirectOosvarValue
			reduce(113), // %=, reduce: DirectOosvarValue
			reduce(113), // **=, reduce: DirectOosvarValue
			reduce(113), // ?, reduce: DirectOosvarValue
			nil,         // :
			reduce(113), // ||, reduce: DirectOosvarValue
			reduce(113), // ^^, reduce: DirectOosvarValue
			reduce(113), // &&, reduce: DirectOosvarValue
			reduce(113), // ??, reduce: DirectOosvarValue
			reduce(113), // ???, reduce: DirectOosvarValue
			reduce(113), // =~, reduce: DirectOosvarValue
			reduce(113), // !=~, reduce: DirectOosvarValue
			reduce(113), // ==, reduce: DirectOosvarValue
			reduce(113), // !=, reduce: DirectOosvarValue
			reduce(113), // <=>, reduce: DirectOosvarValue
			reduce(113), // >=, reduce: DirectOosvarValue
			reduce(113), // <, reduce: DirectOosvarValue
			reduce(113), // <=, reduce: DirectOosvarValue
			reduce(113), // ^, reduce: DirectOosvarValue
			reduce(113), // &, reduce: DirectOosvarValue
			reduce(113), // <<, reduce: DirectOosvarValue
			reduce(113), // >>>, reduce: DirectOosvarValue
			reduce(113), // +, reduce: DirectOosvarValue
			reduce(113), // -, reduce: DirectOosvarValue
			reduce(113), // .+, reduce: DirectOosvarValue
			reduce(113), // .-, reduce: DirectOosvarValue
			reduce(113), // *, reduce: DirectOosvarValue
			reduce(113), // /, reduce: DirectOosvarValue
			reduce(113), // //, reduce: DirectOosvarValue
			reduce(113), // %, reduce: DirectOosvarValue
			reduce(113), // .*, reduce: DirectOosvarValue
			reduce(113), // ./, reduce: DirectOosvarValue
			reduce(113), // .//, reduce: DirectOosvarValue
			reduce(113), // ., reduce: DirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(113), // **, reduce: DirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(113), // [, reduce: DirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(113), // [[, reduce: DirectOosvarValue
			reduce(113), // [[[, reduce: DirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1789
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1790
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(115), // =, reduce: BracedOosvarValue
			nil,         // unset
			nil,         // filter
			reduce(115), // >, reduce: BracedOosvarValue
			reduce(115), // >>, reduce: BracedOosvarValue
			reduce(115), // |, reduce: BracedOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(115), // ||=, reduce: BracedOosvarValue
			reduce(115), // ^^=, reduce: BracedOosvarValue
			reduce(115), // &&=, reduce: BracedOosvarValue
			reduce(115), // ??=, reduce: BracedOosvarValue
			reduce(115), // ???=, reduce: BracedOosvarValue
			reduce(115), // |=, reduce: BracedOosvarValue
			reduce(115), // &=, reduce: BracedOosvarValue
			reduce(115), // ^=, reduce: BracedOosvarValue
			reduce(115), // <<=, reduce: BracedOosvarValue
			reduce(115), // >>=, reduce: BracedOosvarValue
			reduce(115), // >>>=, reduce: BracedOosvarValue
			reduce(115), // +=, reduce: BracedOosvarValue
			reduce(115), // .=, reduce: BracedOosvarValue
			reduce(115), // -=, reduce: BracedOosvarValue
			reduce(115), // *=, reduce: BracedOosvarValue
			reduce(115), // /=, reduce: BracedOosvarValue
			reduce(115), // //=, reduce: BracedOosvarValue
			reduce(115), // %=, reduce: BracedOosvarValue
			reduce(115), // **=, reduce: BracedOosvarValue
			reduce(115), // ?, reduce: BracedOosvarValue
			nil,         // :
			reduce(115), // ||, reduce: BracedOosvarValue
			reduce(115), // ^^, reduce: BracedOosvarValue
			reduce(115), // &&, reduce: BracedOosvarValue
			reduce(115), // ??, reduce: BracedOosvarValue
			reduce(115), // ???, reduce: BracedOosvarValue
			reduce(115), // =~, reduce: BracedOosvarValue
			reduce(115), // !=~, reduce: BracedOosvarValue
			reduce(115), // ==, reduce: BracedOosvarValue
			reduce(115), // !=, reduce: BracedOosvarValue
			reduce(115), // <=>, reduce: BracedOosvarValue
			reduce(115), // >=, reduce: BracedOosvarValue
			reduce(115), // <, reduce: BracedOosvarValue
			reduce(115), // <=, reduce: BracedOosvarValue
			reduce(115), // ^, reduce: BracedOosvarValue
			reduce(115), // &, reduce: BracedOosvarValue
			reduce(115), // <<, reduce: BracedOosvarValue
			reduce(115), // >>>, reduce: BracedOosvarValue
			reduce(115), // +, reduce: BracedOosvarValue
			reduce(115), // -, reduce: BracedOosvarValue
			reduce(115), // .+, reduce: BracedOosvarValue
			reduce(115), // .-, reduce: BracedOosvarValue
			reduce(115), // *, reduce: BracedOosvarValue
			reduce(115), // /, reduce: BracedOosvarValue
			reduce(115), // //, reduce: BracedOosvarValue
			reduce(115), // %, reduce: BracedOosvarValue
			reduce(115), // .*, reduce: BracedOosvarValue
			reduce(115), // ./, reduce: BracedOosvarValue
			reduce(115), // .//, reduce: BracedOosvarValue
			reduce(115), // ., reduce: BracedOosvarValue
			nil,         // !
			nil,         // ~
			reduce(115), // **, reduce: BracedOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(115), // [, reduce: BracedOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(115), // [[, reduce: BracedOosvarValue
			reduce(115), // [[[, reduce: BracedOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1791
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(116), // =, reduce: FullOosvar
			nil,         // unset
			nil,         // filter
			reduce(116), // >, reduce: FullOosvar
			reduce(116), // >>, reduce: FullOosvar
			reduce(116), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(116), // ||=, reduce: FullOosvar
			reduce(116), // ^^=, reduce: FullOosvar
			reduce(116), // &&=, reduce: FullOosvar
			reduce(116), // ??=, reduce: FullOosvar
			reduce(116), // ???=, reduce: FullOosvar
			reduce(116), // |=, reduce: FullOosvar
			reduce(116), // &=, reduce: FullOosvar
			reduce(116), // ^=, reduce: FullOosvar
			reduce(116), // <<=, reduce: FullOosvar
			reduce(116), // >>=, reduce: FullOosvar
			reduce(116), // >>>=, reduce: FullOosvar
			reduce(116), // +=, reduce: FullOosvar
			reduce(116), // .=, reduce: FullOosvar
			reduce(116), // -=, reduce: FullOosvar
			reduce(116), // *=, reduce: FullOosvar
			reduce(116), // /=, reduce: FullOosvar
			reduce(116), // //=, reduce: FullOosvar
			reduce(116), // %=, reduce: FullOosvar
			reduce(116), // **=, reduce: FullOosvar
			reduce(116), // ?, reduce: FullOosvar
			nil,         // :
			reduce(116), // ||, reduce: FullOosvar
			reduce(116), // ^^, reduce: FullOosvar
			reduce(116), // &&, reduce: FullOosvar
			reduce(116), // ??, reduce: FullOosvar
			reduce(116), // ???, reduce: FullOosvar
			reduce(116), // =~, reduce: FullOosvar
			reduce(116), // !=~, reduce: FullOosvar
			reduce(116), // ==, reduce: FullOosvar
			reduce(116), // !=, reduce: FullOosvar
			reduce(116), // <=>, reduce: FullOosvar
			reduce(116), // >=, reduce: FullOosvar
			reduce(116), // <, reduce: FullOosvar
			reduce(116), // <=, reduce: FullOosvar
			reduce(116), // ^, reduce: FullOosvar
			reduce(116), // &, reduce: FullOosvar
			reduce(116), // <<, reduce: FullOosvar
			reduce(116), // >>>, reduce: FullOosvar
			reduce(116), // +, reduce: FullOosvar
			reduce(116), // -, reduce: FullOosvar
			reduce(116), // .+, reduce: FullOosvar
			reduce(116), // .-, reduce: FullOosvar
			reduce(116), // *, reduce: FullOosvar
			reduce(116), // /, reduce: FullOosvar
			reduce(116), // //, reduce: FullOosvar
			reduce(116), // %, reduce: FullOosvar
			reduce(116), // .*, reduce: FullOosvar
			reduce(116), // ./, reduce: FullOosvar
			reduce(116), // .//, reduce: FullOosvar
			reduce(116), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(116), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(116), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(116), // [[, reduce: FullOosvar
			reduce(116), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1792
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(117), // =, reduce: FullOosvar
			nil,         // unset
			nil,         // filter
			reduce(117), // >, reduce: FullOosvar
			reduce(117), // >>, reduce: FullOosvar
			reduce(117), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(117), // ||=, reduce: FullOosvar
			reduce(117), // ^^=, reduce: FullOosvar
			reduce(117), // &&=, reduce: FullOosvar
			reduce(117), // ??=, reduce: FullOosvar
			reduce(117), // ???=, reduce: FullOosvar
			reduce(117), // |=, reduce: FullOosvar
			reduce(117), // &=, reduce: FullOosvar
			reduce(117), // ^=, reduce: FullOosvar
			reduce(117), // <<=, reduce: FullOosvar
			reduce(117), // >>=, reduce: FullOosvar
			reduce(117), // >>>=, reduce: FullOosvar
			reduce(117), // +=, reduce: FullOosvar
			reduce(117), // .=, reduce: FullOosvar
			reduce(117), // -=, reduce: FullOosvar
			reduce(117), // *=, reduce: FullOosvar
			reduce(117), // /=, reduce: FullOosvar
			reduce(117), // //=, reduce: FullOosvar
			reduce(117), // %=, reduce: FullOosvar
			reduce(117), // **=, reduce: FullOosvar
			reduce(117), // ?, reduce: FullOosvar
			nil,         // :
			reduce(117), // ||, reduce: FullOosvar
			reduce(117), // ^^, reduce: FullOosvar
			reduce(117), // &&, reduce: FullOosvar
			reduce(117), // ??, reduce: FullOosvar
			reduce(117), // ???, reduce: FullOosvar
			reduce(117), // =~, reduce: FullOosvar
			reduce(117), // !=~, reduce: FullOosvar
			reduce(117), // ==, reduce: FullOosvar
			reduce(117), // !=, reduce: FullOosvar
			reduce(117), // <=>, reduce: FullOosvar
			reduce(117), // >=, reduce: FullOosvar
			reduce(117), // <, reduce: FullOosvar
			reduce(117), // <=, reduce: FullOosvar
			reduce(117), // ^, reduce: FullOosvar
			reduce(117), // &, reduce: FullOosvar
			reduce(117), // <<, reduce: FullOosvar
			reduce(117), // >>>, reduce: FullOosvar
			reduce(117), // +, reduce: FullOosvar
			reduce(117), // -, reduce: FullOosvar
			reduce(117), // .+, reduce: FullOosvar
			reduce(117), // .-, reduce: FullOosvar
			reduce(117), // *, reduce: FullOosvar
			reduce(117), // /, reduce: FullOosvar
			reduce(117), // //, reduce: FullOosvar
			reduce(117), // %, reduce: FullOosvar
			reduce(117), // .*, reduce: FullOosvar
			reduce(117), // ./, reduce: FullOosvar
			reduce(117), // .//, reduce: FullOosvar
			reduce(117), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(117), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(117), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(117), // [[, reduce: FullOosvar
			reduce(117), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1793
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(118), // =, reduce: LocalVariable
			nil,         // unset
			nil,         // filter
			reduce(118), // >, reduce: LocalVariable
			reduce(118), // >>, reduce: LocalVariable
			reduce(118), // |, reduce: LocalVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(118), // ,, reduce: LocalVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(270), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(118), // ||=, reduce: LocalVariable
			reduce(118), // ^^=, reduce: LocalVariable
			reduce(118), // &&=, reduce: LocalVariable
			reduce(118), // ??=, reduce: LocalVariable
			reduce(118), // ???=, reduce: LocalVariable
			reduce(118), // |=, reduce: LocalVariable
			reduce(118), // &=, reduce: LocalVariable
			reduce(118), // ^=, reduce: LocalVariable
			reduce(118), // <<=, reduce: LocalVariable
			reduce(118), // >>=, reduce: LocalVariable
			reduce(118), // >>>=, reduce: LocalVariable
			reduce(118), // +=, reduce: LocalVariable
			reduce(118), // .=, reduce: LocalVariable
			reduce(118), // -=, reduce: LocalVariable
			reduce(118), // *=, reduce: LocalVariable
			reduce(118), // /=, reduce: LocalVariable
			reduce(118), // //=, reduce: LocalVariable
			reduce(118), // %=, reduce: LocalVariable
			reduce(118), // **=, reduce: LocalVariable
			reduce(118), // ?, reduce: LocalVariable
			nil,         // :
			reduce(118), // ||, reduce: LocalVariable
			reduce(118), // ^^, reduce: LocalVariable
			reduce(118), // &&, reduce: LocalVariable
			reduce(118), // ??, reduce: LocalVariable
			reduce(118), // ???, reduce: LocalVariable
			reduce(118), // =~, reduce: LocalVariable
			reduce(118), // !=~, reduce: LocalVariable
			reduce(118), // ==, reduce: LocalVariable
			reduce(118), // !=, reduce: LocalVariable
			reduce(118), // <=>, reduce: LocalVariable
			reduce(118), // >=, reduce: LocalVariable
			reduce(118), // <, reduce: LocalVariable
			reduce(118), // <=, reduce: LocalVariable
			reduce(118), // ^, reduce: LocalVariable
			reduce(118), // &, reduce: LocalVariable
			reduce(118), // <<, reduce: LocalVariable
			reduce(118), // >>>, reduce: LocalVariable
			reduce(118), // +, reduce: LocalVariable
			reduce(118), // -, reduce: LocalVariable
			reduce(118), // .+, reduce: LocalVariable
			reduce(118), // .-, reduce: LocalVariable
			reduce(118), // *, reduce: LocalVariable
			reduce(118), // /, reduce: LocalVariable
			reduce(118), // //, reduce: LocalVariable
			reduce(118), // %, reduce: LocalVariable
			reduce(118), // .*, reduce: LocalVariable
			reduce(118), // ./, reduce: LocalVariable
			reduce(118), // .//, reduce: LocalVariable
			reduce(118), // ., reduce: LocalVariable
			nil,         // !
			nil,         // ~
			reduce(118), // **, reduce: LocalVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(118), // [, reduce: LocalVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(118), // [[, reduce: LocalVariable
			reduce(118), // [[[, reduce: LocalVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			reduce(118), // in, reduce: LocalVariable
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1794
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(147), // =, reduce: Rvalue
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(147), // ||=, reduce: Rvalue
			reduce(147), // ^^=, reduce: Rvalue
			reduce(147), // &&=, reduce: Rvalue
			reduce(147), // ??=, reduce: Rvalue
			reduce(147), // ???=, reduce: Rvalue
			reduce(147), // |=, reduce: Rvalue
			reduce(147), // &=, reduce: Rvalue
			reduce(147), // ^=, reduce: Rvalue
			reduce(147), // <<=, reduce: Rvalue
			reduce(147), // >>=, reduce: Rvalue
			reduce(147), // >>>=, reduce: Rvalue
			reduce(147), // +=, reduce: Rvalue
			reduce(147), // .=, reduce: Rvalue
			reduce(147), // -=, reduce: Rvalue
			reduce(147), // *=, reduce: Rvalue
			reduce(147), // /=, reduce: Rvalue
			reduce(147), // //=, reduce: Rvalue
			reduce(147), // %=, reduce: Rvalue
			reduce(147), // **=, reduce: Rvalue
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1795
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(148), // =, reduce: PrecedenceChainStart
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(148), // ||=, reduce: PrecedenceChainStart
			reduce(148), // ^^=, reduce: PrecedenceChainStart
			reduce(148), // &&=, reduce: PrecedenceChainStart
			reduce(148), // ??=, reduce: PrecedenceChainStart
			reduce(148), // ???=, reduce: PrecedenceChainStart
			reduce(148), // |=, reduce: PrecedenceChainStart
			reduce(148), // &=, reduce: PrecedenceChainStart
			reduce(148), // ^=, reduce: PrecedenceChainStart
			reduce(148), // <<=, reduce: PrecedenceChainStart
			reduce(148), // >>=, reduce: PrecedenceChainStart
			reduce(148), // >>>=, reduce: PrecedenceChainStart
			reduce(148), // +=, reduce: PrecedenceChainStart
			reduce(148), // .=, reduce: PrecedenceChainStart
			reduce(148), // -=, reduce: PrecedenceChainStart
			reduce(148), // *=, reduce: PrecedenceChainStart
			reduce(148), // /=, reduce: PrecedenceChainStart
			reduce(148), // //=, reduce: PrecedenceChainStart
			reduce(148), // %=, reduce: PrecedenceChainStart
			reduce(148), // **=, reduce: PrecedenceChainStart
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1796
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(150), // =, reduce: TernaryTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(150), // ||=, reduce: TernaryTerm
			reduce(150), // ^^=, reduce: TernaryTerm
			reduce(150), // &&=, reduce: TernaryTerm
			reduce(150), // ??=, reduce: TernaryTerm
			reduce(150), // ???=, reduce: TernaryTerm
			reduce(150), // |=, reduce: TernaryTerm
			reduce(150), // &=, reduce: TernaryTerm
			reduce(150), // ^=, reduce: TernaryTerm
			reduce(150), // <<=, reduce: TernaryTerm
			reduce(150), // >>=, reduce: TernaryTerm
			reduce(150), // >>>=, reduce: TernaryTerm
			reduce(150), // +=, reduce: TernaryTerm
			reduce(150), // .=, reduce: TernaryTerm
			reduce(150), // -=, reduce: TernaryTerm
			reduce(150), // *=, reduce: TernaryTerm
			reduce(150), // /=, reduce: TernaryTerm
			reduce(150), // //=, reduce: TernaryTerm
			reduce(150), // %=, reduce: TernaryTerm
			reduce(150), // **=, reduce: TernaryTerm
			shift(2846), // ?
			nil,         // :
			shift(2847), // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1797
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(152), // =, reduce: LogicalOrTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(152), // ||=, reduce: LogicalOrTerm
			reduce(152), // ^^=, reduce: LogicalOrTerm
			reduce(152), // &&=, reduce: LogicalOrTerm
			reduce(152), // ??=, reduce: LogicalOrTerm
			reduce(152), // ???=, reduce: LogicalOrTerm
			reduce(152), // |=, reduce: LogicalOrTerm
			reduce(152), // &=, reduce: LogicalOrTerm
			reduce(152), // ^=, reduce: LogicalOrTerm
			reduce(152), // <<=, reduce: LogicalOrTerm
			reduce(152), // >>=, reduce: LogicalOrTerm
			reduce(152), // >>>=, reduce: LogicalOrTerm
			reduce(152), // +=, reduce: LogicalOrTerm
			reduce(152), // .=, reduce: LogicalOrTerm
			reduce(152), // -=, reduce: LogicalOrTerm
			reduce(152), // *=, reduce: LogicalOrTerm
			reduce(152), // /=, reduce: LogicalOrTerm
			reduce(152), // //=, reduce: LogicalOrTerm
			reduce(152), // %=, reduce: LogicalOrTerm
			reduce(152), // **=, reduce: LogicalOrTerm
			reduce(152), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(152), // ||, reduce: LogicalOrTerm
			shift(2848), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1798
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(154), // =, reduce: LogicalXORTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(154), // ||=, reduce: LogicalXORTerm
			reduce(154), // ^^=, reduce: LogicalXORTerm
			reduce(154), // &&=, reduce: LogicalXORTerm
			reduce(154), // ??=, reduce: LogicalXORTerm
			reduce(154), // ???=, reduce: LogicalXORTerm
			reduce(154), // |=, reduce: LogicalXORTerm
			reduce(154), // &=, reduce: LogicalXORTerm
			reduce(154), // ^=, reduce: LogicalXORTerm
			reduce(154), // <<=, reduce: LogicalXORTerm
			reduce(154), // >>=, reduce: LogicalXORTerm
			reduce(154), // >>>=, reduce: LogicalXORTerm
			reduce(154), // +=, reduce: LogicalXORTerm
			reduce(154), // .=, reduce: LogicalXORTerm
			reduce(154), // -=, reduce: LogicalXORTerm
			reduce(154), // *=, reduce: LogicalXORTerm
			reduce(154), // /=, reduce: LogicalXORTerm
			reduce(154), // //=, reduce: LogicalXORTerm
			reduce(154), // %=, reduce: LogicalXORTerm
			reduce(154), // **=, reduce: LogicalXORTerm
			reduce(154), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(154), // ||, reduce: LogicalXORTerm
			reduce(154), // ^^, reduce: LogicalXORTerm
			shift(2849), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1799
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(156), // =, reduce: LogicalAndTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(156), // ||=, reduce: LogicalAndTerm
			reduce(156), // ^^=, reduce: LogicalAndTerm
			reduce(156), // &&=, reduce: LogicalAndTerm
			reduce(156), // ??=, reduce: LogicalAndTerm
			reduce(156), // ???=, reduce: LogicalAndTerm
			reduce(156), // |=, reduce: LogicalAndTerm
			reduce(156), // &=, reduce: LogicalAndTerm
			reduce(156), // ^=, reduce: LogicalAndTerm
			reduce(156), // <<=, reduce: LogicalAndTerm
			reduce(156), // >>=, reduce: LogicalAndTerm
			reduce(156), // >>>=, reduce: LogicalAndTerm
			reduce(156), // +=, reduce: LogicalAndTerm
			reduce(156), // .=, reduce: LogicalAndTerm
			reduce(156), // -=, reduce: LogicalAndTerm
			reduce(156), // *=, reduce: LogicalAndTerm
			reduce(156), // /=, reduce: LogicalAndTerm
			reduce(156), // //=, reduce: LogicalAndTerm
			reduce(156), // %=, reduce: LogicalAndTerm
			reduce(156), // **=, reduce: LogicalAndTerm
			reduce(156), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(156), // ||, reduce: LogicalAndTerm
			reduce(156), // ^^, reduce: LogicalAndTerm
			reduce(156), // &&, reduce: LogicalAndTerm
			shift(2850), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1800
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(158), // =, reduce: AbsentCoalesceTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(158), // ||=, reduce: AbsentCoalesceTerm
			reduce(158), // ^^=, reduce: AbsentCoalesceTerm
			reduce(158), // &&=, reduce: AbsentCoalesceTerm
			reduce(158), // ??=, reduce: AbsentCoalesceTerm
			reduce(158), // ???=, reduce: AbsentCoalesceTerm
			reduce(158), // |=, reduce: AbsentCoalesceTerm
			reduce(158), // &=, reduce: AbsentCoalesceTerm
			reduce(158), // ^=, reduce: AbsentCoalesceTerm
			reduce(158), // <<=, reduce: AbsentCoalesceTerm
			reduce(158), // >>=, reduce: AbsentCoalesceTerm
			reduce(158), // >>>=, reduce: AbsentCoalesceTerm
			reduce(158), // +=, reduce: AbsentCoalesceTerm
			reduce(158), // .=, reduce: AbsentCoalesceTerm
			reduce(158), // -=, reduce: AbsentCoalesceTerm
			reduce(158), // *=, reduce: AbsentCoalesceTerm
			reduce(158), // /=, reduce: AbsentCoalesceTerm
			reduce(158), // //=, reduce: AbsentCoalesceTerm
			reduce(158), // %=, reduce: AbsentCoalesceTerm
			reduce(158), // **=, reduce: AbsentCoalesceTerm
			reduce(158), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(158), // ||, reduce: AbsentCoalesceTerm
			reduce(158), // ^^, reduce: AbsentCoalesceTerm
			reduce(158), // &&, reduce: AbsentCoalesceTerm
			reduce(158), // ??, reduce: AbsentCoalesceTerm
			shift(2851), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1801
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(160), // =, reduce: EmptyCoalesceTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(160), // ||=, reduce: EmptyCoalesceTerm
			reduce(160), // ^^=, reduce: EmptyCoalesceTerm
			reduce(160), // &&=, reduce: EmptyCoalesceTerm
			reduce(160), // ??=, reduce: EmptyCoalesceTerm
			reduce(160), // ???=, reduce: EmptyCoalesceTerm
			reduce(160), // |=, reduce: EmptyCoalesceTerm
			reduce(160), // &=, reduce: EmptyCoalesceTerm
			reduce(160), // ^=, reduce: EmptyCoalesceTerm
			reduce(160), // <<=, reduce: EmptyCoalesceTerm
			reduce(160), // >>=, reduce: EmptyCoalesceTerm
			reduce(160), // >>>=, reduce: EmptyCoalesceTerm
			reduce(160), // +=, reduce: EmptyCoalesceTerm
			reduce(160), // .=, reduce: EmptyCoalesceTerm
			reduce(160), // -=, reduce: EmptyCoalesceTerm
			reduce(160), // *=, reduce: EmptyCoalesceTerm
			reduce(160), // /=, reduce: EmptyCoalesceTerm
			reduce(160), // //=, reduce: EmptyCoalesceTerm
			reduce(160), // %=, reduce: EmptyCoalesceTerm
			reduce(160), // **=, reduce: EmptyCoalesceTerm
			reduce(160), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(160), // ||, reduce: EmptyCoalesceTerm
			reduce(160), // ^^, reduce: EmptyCoalesceTerm
			reduce(160), // &&, reduce: EmptyCoalesceTerm
			reduce(160), // ??, reduce: EmptyCoalesceTerm
			reduce(160), // ???, reduce: EmptyCoalesceTerm
			shift(2852), // =~
			shift(2853), // !=~
			shift(2854), // ==
			shift(2855), // !=
			shift(2856), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1802
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(166), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(2857), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(166), // ||=, reduce: EqneTerm
			reduce(166), // ^^=, reduce: EqneTerm
			reduce(166), // &&=, reduce: EqneTerm
			reduce(166), // ??=, reduce: EqneTerm
			reduce(166), // ???=, reduce: EqneTerm
			reduce(166), // |=, reduce: EqneTerm
			reduce(166), // &=, reduce: EqneTerm
			reduce(166), // ^=, reduce: EqneTerm
			reduce(166), // <<=, reduce: EqneTerm
			reduce(166), // >>=, reduce: EqneTerm
			reduce(166), // >>>=, reduce: EqneTerm
			reduce(166), // +=, reduce: EqneTerm
			reduce(166), // .=, reduce: EqneTerm
			reduce(166), // -=, reduce: EqneTerm
			reduce(166), // *=, reduce: EqneTerm
			reduce(166), // /=, reduce: EqneTerm
			reduce(166), // //=, reduce: EqneTerm
			reduce(166), // %=, reduce: EqneTerm
			reduce(166), // **=, reduce: EqneTerm
			reduce(166), // ?, reduce: EqneTerm
			nil,         // :
			reduce(166), // ||, reduce: EqneTerm
			reduce(166), // ^^, reduce: EqneTerm
			reduce(166), // &&, reduce: EqneTerm
			reduce(166), // ??, reduce: EqneTerm
			reduce(166), // ???, reduce: EqneTerm
			reduce(166), // =~, reduce: EqneTerm
			reduce(166), // !=~, reduce: EqneTerm
			reduce(166), // ==, reduce: EqneTerm
			reduce(166), // !=, reduce: EqneTerm
			reduce(166), // <=>, reduce: EqneTerm
			shift(2858), // >=
			shift(2859), // <
			shift(2860), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1803
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(171), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(171), // >, reduce: CmpTerm
			nil,         // >>
			shift(2861), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(171), // ||=, reduce: CmpTerm
			reduce(171), // ^^=, reduce: CmpTerm
			reduce(171), // &&=, reduce: CmpTerm
			reduce(171), // ??=, reduce: CmpTerm
			reduce(171), // ???=, reduce: CmpTerm
			reduce(171), // |=, reduce: CmpTerm
			reduce(171), // &=, reduce: CmpTerm
			reduce(171), // ^=, reduce: CmpTerm
			reduce(171), // <<=, reduce: CmpTerm
			reduce(171), // >>=, reduce: CmpTerm
			reduce(171), // >>>=, reduce: CmpTerm
			reduce(171), // +=, reduce: CmpTerm
			reduce(171), // .=, reduce: CmpTerm
			reduce(171), // -=, reduce: CmpTerm
			reduce(171), // *=, reduce: CmpTerm
			reduce(171), // /=, reduce: CmpTerm
			reduce(171), // //=, reduce: CmpTerm
			reduce(171), // %=, reduce: CmpTerm
			reduce(171), // **=, reduce: CmpTerm
			reduce(171), // ?, reduce: CmpTerm
			nil,         // :
			reduce(171), // ||, reduce: CmpTerm
			reduce(171), // ^^, reduce: CmpTerm
			reduce(171), // &&, reduce: CmpTerm
			reduce(171), // ??, reduce: CmpTerm
			reduce(171), // ???, reduce: CmpTerm
			reduce(171), // =~, reduce: CmpTerm
			reduce(171), // !=~, reduce: CmpTerm
			reduce(171), // ==, reduce: CmpTerm
			reduce(171), // !=, reduce: CmpTerm
			reduce(171), // <=>, reduce: CmpTerm
			reduce(171), // >=, reduce: CmpTerm
			reduce(171), // <, reduce: CmpTerm
			reduce(171), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1804
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(173), // =, reduce: BitwiseORTerm
			nil,         // unset
			nil,         // filter
			reduce(173), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(173), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(173), // ||=, reduce: BitwiseORTerm
			reduce(173), // ^^=, reduce: BitwiseORTerm
			reduce(173), // &&=, reduce: BitwiseORTerm
			reduce(173), // ??=, reduce: BitwiseORTerm
			reduce(173), // ???=, reduce: BitwiseORTerm
			reduce(173), // |=, reduce: BitwiseORTerm
			reduce(173), // &=, reduce: BitwiseORTerm
			reduce(173), // ^=, reduce: BitwiseORTerm
			reduce(173), // <<=, reduce: BitwiseORTerm
			reduce(173), // >>=, reduce: BitwiseORTerm
			reduce(173), // >>>=, reduce: BitwiseORTerm
			reduce(173), // +=, reduce: BitwiseORTerm
			reduce(173), // .=, reduce: BitwiseORTerm
			reduce(173), // -=, reduce: BitwiseORTerm
			reduce(173), // *=, reduce: BitwiseORTerm
			reduce(173), // /=, reduce: BitwiseORTerm
			reduce(173), // //=, reduce: BitwiseORTerm
			reduce(173), // %=, reduce: BitwiseORTerm
			reduce(173), // **=, reduce: BitwiseORTerm
			reduce(173), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(173), // ||, reduce: BitwiseORTerm
			reduce(173), // ^^, reduce: BitwiseORTerm
			reduce(173), // &&, reduce: BitwiseORTerm
			reduce(173), // ??, reduce: BitwiseORTerm
			reduce(173), // ???, reduce: BitwiseORTerm
			reduce(173), // =~, reduce: BitwiseORTerm
			reduce(173), // !=~, reduce: BitwiseORTerm
			reduce(173), // ==, reduce: BitwiseORTerm
			reduce(173), // !=, reduce: BitwiseORTerm
			reduce(173), // <=>, reduce: BitwiseORTerm
			reduce(173), // >=, reduce: BitwiseORTerm
			reduce(173), // <, reduce: BitwiseORTerm
			reduce(173), // <=, reduce: BitwiseORTerm
			shift(2862), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1805
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(175), // =, reduce: BitwiseXORTerm
			nil,         // unset
			nil,         // filter
			reduce(175), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(175), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(175), // ||=, reduce: BitwiseXORTerm
			reduce(175), // ^^=, reduce: BitwiseXORTerm
			reduce(175), // &&=, reduce: BitwiseXORTerm
			reduce(175), // ??=, reduce: BitwiseXORTerm
			reduce(175), // ???=, reduce: BitwiseXORTerm
			reduce(175), // |=, reduce: BitwiseXORTerm
			reduce(175), // &=, reduce: BitwiseXORTerm
			reduce(175), // ^=, reduce: BitwiseXORTerm
			reduce(175), // <<=, reduce: BitwiseXORTerm
			reduce(175), // >>=, reduce: BitwiseXORTerm
			reduce(175), // >>>=, reduce: BitwiseXORTerm
			reduce(175), // +=, reduce: BitwiseXORTerm
			reduce(175), // .=, reduce: BitwiseXORTerm
			reduce(175), // -=, reduce: BitwiseXORTerm
			reduce(175), // *=, reduce: BitwiseXORTerm
			reduce(175), // /=, reduce: BitwiseXORTerm
			reduce(175), // //=, reduce: BitwiseXORTerm
			reduce(175), // %=, reduce: BitwiseXORTerm
			reduce(175), // **=, reduce: BitwiseXORTerm
			reduce(175), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(175), // ||, reduce: BitwiseXORTerm
			reduce(175), // ^^, reduce: BitwiseXORTerm
			reduce(175), // &&, reduce: BitwiseXORTerm
			reduce(175), // ??, reduce: BitwiseXORTerm
			reduce(175), // ???, reduce: BitwiseXORTerm
			reduce(175), // =~, reduce: BitwiseXORTerm
			reduce(175), // !=~, reduce: BitwiseXORTerm
			reduce(175), // ==, reduce: BitwiseXORTerm
			reduce(175), // !=, reduce: BitwiseXORTerm
			reduce(175), // <=>, reduce: BitwiseXORTerm
			reduce(175), // >=, reduce: BitwiseXORTerm
			reduce(175), // <, reduce: BitwiseXORTerm
			reduce(175), // <=, reduce: BitwiseXORTerm
			reduce(175), // ^, reduce: BitwiseXORTerm
			shift(2863), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1806
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(177), // =, reduce: BitwiseANDTerm
			nil,         // unset
			nil,         // filter
			reduce(177), // >, reduce: BitwiseANDTerm
			shift(2864), // >>
			reduce(177), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(177), // ||=, reduce: BitwiseANDTerm
			reduce(177), // ^^=, reduce: BitwiseANDTerm
			reduce(177), // &&=, reduce: BitwiseANDTerm
			reduce(177), // ??=, reduce: BitwiseANDTerm
			reduce(177), // ???=, reduce: BitwiseANDTerm
			reduce(177), // |=, reduce: BitwiseANDTerm
			reduce(177), // &=, reduce: BitwiseANDTerm
			reduce(177), // ^=, reduce: BitwiseANDTerm
			reduce(177), // <<=, reduce: BitwiseANDTerm
			reduce(177), // >>=, reduce: BitwiseANDTerm
			reduce(177), // >>>=, reduce: BitwiseANDTerm
			reduce(177), // +=, reduce: BitwiseANDTerm
			reduce(177), // .=, reduce: BitwiseANDTerm
			reduce(177), // -=, reduce: BitwiseANDTerm
			reduce(177), // *=, reduce: BitwiseANDTerm
			reduce(177), // /=, reduce: BitwiseANDTerm
			reduce(177), // //=, reduce: BitwiseANDTerm
			reduce(177), // %=, reduce: BitwiseANDTerm
			reduce(177), // **=, reduce: BitwiseANDTerm
			reduce(177), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(177), // ||, reduce: BitwiseANDTerm
			reduce(177), // ^^, reduce: BitwiseANDTerm
			reduce(177), // &&, reduce: BitwiseANDTerm
			reduce(177), // ??, reduce: BitwiseANDTerm
			reduce(177), // ???, reduce: BitwiseANDTerm
			reduce(177), // =~, reduce: BitwiseANDTerm
			reduce(177), // !=~, reduce: BitwiseANDTerm
			reduce(177), // ==, reduce: BitwiseANDTerm
			reduce(177), // !=, reduce: BitwiseANDTerm
			reduce(177), // <=>, reduce: BitwiseANDTerm
			reduce(177), // >=, reduce: BitwiseANDTerm
			reduce(177), // <, reduce: BitwiseANDTerm
			reduce(177), // <=, reduce: BitwiseANDTerm
			reduce(177), // ^, reduce: BitwiseANDTerm
			reduce(177), // &, reduce: BitwiseANDTerm
			shift(2865), // <<
			shift(2866), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1807
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(181), // =, reduce: BitwiseShiftTerm
			nil,         // unset
			nil,         // filter
			reduce(181), // >, reduce: BitwiseShiftTerm
			reduce(181), // >>, reduce: BitwiseShiftTerm
			reduce(181), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(181), // ||=, reduce: BitwiseShiftTerm
			reduce(181), // ^^=, reduce: BitwiseShiftTerm
			reduce(181), // &&=, reduce: BitwiseShiftTerm
			reduce(181), // ??=, reduce: BitwiseShiftTerm
			reduce(181), // ???=, reduce: BitwiseShiftTerm
			reduce(181), // |=, reduce: BitwiseShiftTerm
			reduce(181), // &=, reduce: BitwiseShiftTerm
			reduce(181), // ^=, reduce: BitwiseShiftTerm
			reduce(181), // <<=, reduce: BitwiseShiftTerm
			reduce(181), // >>=, reduce: BitwiseShiftTerm
			reduce(181), // >>>=, reduce: BitwiseShiftTerm
			reduce(181), // +=, reduce: BitwiseShiftTerm
			reduce(181), // .=, reduce: BitwiseShiftTerm
			reduce(181), // -=, reduce: BitwiseShiftTerm
			reduce(181), // *=, reduce: BitwiseShiftTerm
			reduce(181), // /=, reduce: BitwiseShiftTerm
			reduce(181), // //=, reduce: BitwiseShiftTerm
			reduce(181), // %=, reduce: BitwiseShiftTerm
			reduce(181), // **=, reduce: BitwiseShiftTerm
			reduce(181), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(181), // ||, reduce: BitwiseShiftTerm
			reduce(181), // ^^, reduce: BitwiseShiftTerm
			reduce(181), // &&, reduce: BitwiseShiftTerm
			reduce(181), // ??, reduce: BitwiseShiftTerm
			reduce(181), // ???, reduce: BitwiseShiftTerm
			reduce(181), // =~, reduce: BitwiseShiftTerm
			reduce(181), // !=~, reduce: BitwiseShiftTerm
			reduce(181), // ==, reduce: BitwiseShiftTerm
			reduce(181), // !=, reduce: BitwiseShiftTerm
			reduce(181), // <=>, reduce: BitwiseShiftTerm
			reduce(181), // >=, reduce: BitwiseShiftTerm
			reduce(181), // <, reduce: BitwiseShiftTerm
			reduce(181), // <=, reduce: BitwiseShiftTerm
			reduce(181), // ^, reduce: BitwiseShiftTerm
			reduce(181), // &, reduce: BitwiseShiftTerm
			reduce(181), // <<, reduce: BitwiseShiftTerm
			reduce(181), // >>>, reduce: BitwiseShiftTerm
			shift(2867), // +
			shift(2868), // -
			shift(2869), // .+
			shift(2870), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1808
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1809
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(186), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(186), // >, reduce: AddsubdotTerm
			reduce(186), // >>, reduce: AddsubdotTerm
			reduce(186), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(186), // ||=, reduce: AddsubdotTerm
			reduce(186), // ^^=, reduce: AddsubdotTerm
			reduce(186), // &&=, reduce: AddsubdotTerm
			reduce(186), // ??=, reduce: AddsubdotTerm
			reduce(186), // ???=, reduce: AddsubdotTerm
			reduce(186), // |=, reduce: AddsubdotTerm
			reduce(186), // &=, reduce: AddsubdotTerm
			reduce(186), // ^=, reduce: AddsubdotTerm
			reduce(186), // <<=, reduce: AddsubdotTerm
			reduce(186), // >>=, reduce: AddsubdotTerm
			reduce(186), // >>>=, reduce: AddsubdotTerm
			reduce(186), // +=, reduce: AddsubdotTerm
			reduce(186), // .=, reduce: AddsubdotTerm
			reduce(186), // -=, reduce: AddsubdotTerm
			reduce(186), // *=, reduce: AddsubdotTerm
			reduce(186), // /=, reduce: AddsubdotTerm
			reduce(186), // //=, reduce: AddsubdotTerm
			reduce(186), // %=, reduce: AddsubdotTerm
			reduce(186), // **=, reduce: AddsubdotTerm
			reduce(186), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(186), // ||, reduce: AddsubdotTerm
			reduce(186), // ^^, reduce: AddsubdotTerm
			reduce(186), // &&, reduce: AddsubdotTerm
			reduce(186), // ??, reduce: AddsubdotTerm
			reduce(186), // ???, reduce: AddsubdotTerm
			reduce(186), // =~, reduce: AddsubdotTerm
			reduce(186), // !=~, reduce: AddsubdotTerm
			reduce(186), // ==, reduce: AddsubdotTerm
			reduce(186), // !=, reduce: AddsubdotTerm
			reduce(186), // <=>, reduce: AddsubdotTerm
			reduce(186), // >=, reduce: AddsubdotTerm
			reduce(186), // <, reduce: AddsubdotTerm
			reduce(186), // <=, reduce: AddsubdotTerm
			reduce(186), // ^, reduce: AddsubdotTerm
			reduce(186), // &, reduce: AddsubdotTerm
			reduce(186), // <<, reduce: AddsubdotTerm
			reduce(186), // >>>, reduce: AddsubdotTerm
			reduce(186), // +, reduce: AddsubdotTerm
			reduce(186), // -, reduce: AddsubdotTerm
			reduce(186), // .+, reduce: AddsubdotTerm
			reduce(186), // .-, reduce: AddsubdotTerm
			shift(2875), // *
			shift(2876), // /
			shift(2877), // //
			shift(2878), // %
			shift(2879), // .*
			shift(2880), // ./
			shift(2881), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1810
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1811
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1812
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1813
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(194), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(194), // >, reduce: MuldivTerm
			reduce(194), // >>, reduce: MuldivTerm
			reduce(194), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(194), // ||=, reduce: MuldivTerm
			reduce(194), // ^^=, reduce: MuldivTerm
			reduce(194), // &&=, reduce: MuldivTerm
			reduce(194), // ??=, reduce: MuldivTerm
			reduce(194), // ???=, reduce: MuldivTerm
			reduce(194), // |=, reduce: MuldivTerm
			reduce(194), // &=, reduce: MuldivTerm
			reduce(194), // ^=, reduce: MuldivTerm
			reduce(194), // <<=, reduce: MuldivTerm
			reduce(194), // >>=, reduce: MuldivTerm
			reduce(194), // >>>=, reduce: MuldivTerm
			reduce(194), // +=, reduce: MuldivTerm
			reduce(194), // .=, reduce: MuldivTerm
			reduce(194), // -=, reduce: MuldivTerm
			reduce(194), // *=, reduce: MuldivTerm
			reduce(194), // /=, reduce: MuldivTerm
			reduce(194), // //=, reduce: MuldivTerm
			reduce(194), // %=, reduce: MuldivTerm
			reduce(194), // **=, reduce: MuldivTerm
			reduce(194), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(194), // ||, reduce: MuldivTerm
			reduce(194), // ^^, reduce: MuldivTerm
			reduce(194), // &&, reduce: MuldivTerm
			reduce(194), // ??, reduce: MuldivTerm
			reduce(194), // ???, reduce: MuldivTerm
			reduce(194), // =~, reduce: MuldivTerm
			reduce(194), // !=~, reduce: MuldivTerm
			reduce(194), // ==, reduce: MuldivTerm
			reduce(194), // !=, reduce: MuldivTerm
			reduce(194), // <=>, reduce: MuldivTerm
			reduce(194), // >=, reduce: MuldivTerm
			reduce(194), // <, reduce: MuldivTerm
			reduce(194), // <=, reduce: MuldivTerm
			reduce(194), // ^, reduce: MuldivTerm
			reduce(194), // &, reduce: MuldivTerm
			reduce(194), // <<, reduce: MuldivTerm
			reduce(194), // >>>, reduce: MuldivTerm
			reduce(194), // +, reduce: MuldivTerm
			reduce(194), // -, reduce: MuldivTerm
			reduce(194), // .+, reduce: MuldivTerm
			reduce(194), // .-, reduce: MuldivTerm
			reduce(194), // *, reduce: MuldivTerm
			reduce(194), // /, reduce: MuldivTerm
			reduce(194), // //, reduce: MuldivTerm
			reduce(194), // %, reduce: MuldivTerm
			reduce(194), // .*, reduce: MuldivTerm
			reduce(194), // ./, reduce: MuldivTerm
			reduce(194), // .//, reduce: MuldivTerm
			shift(2885), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1814
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(196), // =, reduce: DotTerm
			nil,         // unset
			nil,         // filter
			reduce(196), // >, reduce: DotTerm
			reduce(196), // >>, reduce: DotTerm
			reduce(196), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(196), // ||=, reduce: DotTerm
			reduce(196), // ^^=, reduce: DotTerm
			reduce(196), // &&=, reduce: DotTerm
			reduce(196), // ??=, reduce: DotTerm
			reduce(196), // ???=, reduce: DotTerm
			reduce(196), // |=, reduce: DotTerm
			reduce(196), // &=, reduce: DotTerm
			reduce(196), // ^=, reduce: DotTerm
			reduce(196), // <<=, reduce: DotTerm
			reduce(196), // >>=, reduce: DotTerm
			reduce(196), // >>>=, reduce: DotTerm
			reduce(196), // +=, reduce: DotTerm
			reduce(196), // .=, reduce: DotTerm
			reduce(196), // -=, reduce: DotTerm
			reduce(196), // *=, reduce: DotTerm
			reduce(196), // /=, reduce: DotTerm
			reduce(196), // //=, reduce: DotTerm
			reduce(196), // %=, reduce: DotTerm
			reduce(196), // **=, reduce: DotTerm
			reduce(196), // ?, reduce: DotTerm
			nil,         // :
			reduce(196), // ||, reduce: DotTerm
			reduce(196), // ^^, reduce: DotTerm
			reduce(196), // &&, reduce: DotTerm
			reduce(196), // ??, reduce: DotTerm
			reduce(196), // ???, reduce: DotTerm
			reduce(196), // =~, reduce: DotTerm
			reduce(196), // !=~, reduce: DotTerm
			reduce(196), // ==, reduce: DotTerm
			reduce(196), // !=, reduce: DotTerm
			reduce(196), // <=>, reduce: DotTerm
			reduce(196), // >=, reduce: DotTerm
			reduce(196), // <, reduce: DotTerm
			reduce(196), // <=, reduce: DotTerm
			reduce(196), // ^, reduce: DotTerm
			reduce(196), // &, reduce: DotTerm
			reduce(196), // <<, reduce: DotTerm
			reduce(196), // >>>, reduce: DotTerm
			reduce(196), // +, reduce: DotTerm
			reduce(196), // -, reduce: DotTerm
			reduce(196), // .+, reduce: DotTerm
			reduce(196), // .-, reduce: DotTerm
			reduce(196), // *, reduce: DotTerm
			reduce(196), // /, reduce: DotTerm
			reduce(196), // //, reduce: DotTerm
			reduce(196), // %, reduce: DotTerm
			reduce(196), // .*, reduce: DotTerm
			reduce(196), // ./, reduce: DotTerm
			reduce(196), // .//, reduce: DotTerm
			reduce(196), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1815
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1816
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1817
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(203), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(203), // >, reduce: UnaryOpTerm
			reduce(203), // >>, reduce: UnaryOpTerm
			reduce(203), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(203), // ||=, reduce: UnaryOpTerm
			reduce(203), // ^^=, reduce: UnaryOpTerm
			reduce(203), // &&=, reduce: UnaryOpTerm
			reduce(203), // ??=, reduce: UnaryOpTerm
			reduce(203), // ???=, reduce: UnaryOpTerm
			reduce(203), // |=, reduce: UnaryOpTerm
			reduce(203), // &=, reduce: UnaryOpTerm
			reduce(203), // ^=, reduce: UnaryOpTerm
			reduce(203), // <<=, reduce: UnaryOpTerm
			reduce(203), // >>=, reduce: UnaryOpTerm
			reduce(203), // >>>=, reduce: UnaryOpTerm
			reduce(203), // +=, reduce: UnaryOpTerm
			reduce(203), // .=, reduce: UnaryOpTerm
			reduce(203), // -=, reduce: UnaryOpTerm
			reduce(203), // *=, reduce: UnaryOpTerm
			reduce(203), // /=, reduce: UnaryOpTerm
			reduce(203), // //=, reduce: UnaryOpTerm
			reduce(203), // %=, reduce: UnaryOpTerm
			reduce(203), // **=, reduce: UnaryOpTerm
			reduce(203), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(203), // ||, reduce: UnaryOpTerm
			reduce(203), // ^^, reduce: UnaryOpTerm
			reduce(203), // &&, reduce: UnaryOpTerm
			reduce(203), // ??, reduce: UnaryOpTerm
			reduce(203), // ???, reduce: UnaryOpTerm
			reduce(203), // =~, reduce: UnaryOpTerm
			reduce(203), // !=~, reduce: UnaryOpTerm
			reduce(203), // ==, reduce: UnaryOpTerm
			reduce(203), // !=, reduce: UnaryOpTerm
			reduce(203), // <=>, reduce: UnaryOpTerm
			reduce(203), // >=, reduce: UnaryOpTerm
			reduce(203), // <, reduce: UnaryOpTerm
			reduce(203), // <=, reduce: UnaryOpTerm
			reduce(203), // ^, reduce: UnaryOpTerm
			reduce(203), // &, reduce: UnaryOpTerm
			reduce(203), // <<, reduce: UnaryOpTerm
			reduce(203), // >>>, reduce: UnaryOpTerm
			reduce(203), // +, reduce: UnaryOpTerm
			reduce(203), // -, reduce: UnaryOpTerm
			reduce(203), // .+, reduce: UnaryOpTerm
			reduce(203), // .-, reduce: UnaryOpTerm
			reduce(203), // *, reduce: UnaryOpTerm
			reduce(203), // /, reduce: UnaryOpTerm
			reduce(203), // //, reduce: UnaryOpTerm
			reduce(203), // %, reduce: UnaryOpTerm
			reduce(203), // .*, reduce: UnaryOpTerm
			reduce(203), // ./, reduce: UnaryOpTerm
			reduce(203), // .//, reduce: UnaryOpTerm
			reduce(203), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1818
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(207), // =, reduce: PowTerm
			nil,         // unset
			nil,         // filter
			reduce(207), // >, reduce: PowTerm
			reduce(207), // >>, reduce: PowTerm
			reduce(207), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(207), // ||=, reduce: PowTerm
			reduce(207), // ^^=, reduce: PowTerm
			reduce(207), // &&=, reduce: PowTerm
			reduce(207), // ??=, reduce: PowTerm
			reduce(207), // ???=, reduce: PowTerm
			reduce(207), // |=, reduce: PowTerm
			reduce(207), // &=, reduce: PowTerm
			reduce(207), // ^=, reduce: PowTerm
			reduce(207), // <<=, reduce: PowTerm
			reduce(207), // >>=, reduce: PowTerm
			reduce(207), // >>>=, reduce: PowTerm
			reduce(207), // +=, reduce: PowTerm
			reduce(207), // .=, reduce: PowTerm
			reduce(207), // -=, reduce: PowTerm
			reduce(207), // *=, reduce: PowTerm
			reduce(207), // /=, reduce: PowTerm
			reduce(207), // //=, reduce: PowTerm
			reduce(207), // %=, reduce: PowTerm
			reduce(207), // **=, reduce: PowTerm
			reduce(207), // ?, reduce: PowTerm
			nil,         // :
			reduce(207), // ||, reduce: PowTerm
			reduce(207), // ^^, reduce: PowTerm
			reduce(207), // &&, reduce: PowTerm
			reduce(207), // ??, reduce: PowTerm
			reduce(207), // ???, reduce: PowTerm
			reduce(207), // =~, reduce: PowTerm
			reduce(207), // !=~, reduce: PowTerm
			reduce(207), // ==, reduce: PowTerm
			reduce(207), // !=, reduce: PowTerm
			reduce(207), // <=>, reduce: PowTerm
			reduce(207), // >=, reduce: PowTerm
			reduce(207), // <, reduce: PowTerm
			reduce(207), // <=, reduce: PowTerm
			reduce(207), // ^, reduce: PowTerm
			reduce(207), // &, reduce: PowTerm
			reduce(207), // <<, reduce: PowTerm
			reduce(207), // >>>, reduce: PowTerm
			reduce(207), // +, reduce: PowTerm
			reduce(207), // -, reduce: PowTerm
			reduce(207), // .+, reduce: PowTerm
			reduce(207), // .-, reduce: PowTerm
			reduce(207), // *, reduce: PowTerm
			reduce(207), // /, reduce: PowTerm
			reduce(207), // //, reduce: PowTerm
			reduce(207), // %, reduce: PowTerm
			reduce(207), // .*, reduce: PowTerm
			reduce(207), // ./, reduce: PowTerm
			reduce(207), // .//, reduce: PowTerm
			reduce(207), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			shift(2888), // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1819
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(209), // =, reduce: PrecedenceChainEnd
			nil,         // unset
			nil,         // filter
			reduce(209), // >, reduce: PrecedenceChainEnd
			reduce(209), // >>, reduce: PrecedenceChainEnd
			reduce(209), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(209), // ||=, reduce: PrecedenceChainEnd
			reduce(209), // ^^=, reduce: PrecedenceChainEnd
			reduce(209), // &&=, reduce: PrecedenceChainEnd
			reduce(209), // ??=, reduce: PrecedenceChainEnd
			reduce(209), // ???=, reduce: PrecedenceChainEnd
			reduce(209), // |=, reduce: PrecedenceChainEnd
			reduce(209), // &=, reduce: PrecedenceChainEnd
			reduce(209), // ^=, reduce: PrecedenceChainEnd
			reduce(209), // <<=, reduce: PrecedenceChainEnd
			reduce(209), // >>=, reduce: PrecedenceChainEnd
			reduce(209), // >>>=, reduce: PrecedenceChainEnd
			reduce(209), // +=, reduce: PrecedenceChainEnd
			reduce(209), // .=, reduce: PrecedenceChainEnd
			reduce(209), // -=, reduce: PrecedenceChainEnd
			reduce(209), // *=, reduce: PrecedenceChainEnd
			reduce(209), // /=, reduce: PrecedenceChainEnd
			reduce(209), // //=, reduce: PrecedenceChainEnd
			reduce(209), // %=, reduce: PrecedenceChainEnd
			reduce(209), // **=, reduce: PrecedenceChainEnd
			reduce(209), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(209), // ||, reduce: PrecedenceChainEnd
			reduce(209), // ^^, reduce: PrecedenceChainEnd
			reduce(209), // &&, reduce: PrecedenceChainEnd
			reduce(209), // ??, reduce: PrecedenceChainEnd
			reduce(209), // ???, reduce: PrecedenceChainEnd
			reduce(209), // =~, reduce: PrecedenceChainEnd
			reduce(209), // !=~, reduce: PrecedenceChainEnd
			reduce(209), // ==, reduce: PrecedenceChainEnd
			reduce(209), // !=, reduce: PrecedenceChainEnd
			reduce(209), // <=>, reduce: PrecedenceChainEnd
			reduce(209), // >=, reduce: PrecedenceChainEnd
			reduce(209), // <, reduce: PrecedenceChainEnd
			reduce(209), // <=, reduce: PrecedenceChainEnd
			reduce(209), // ^, reduce: PrecedenceChainEnd
			reduce(209), // &, reduce: PrecedenceChainEnd
			reduce(209), // <<, reduce: PrecedenceChainEnd
			reduce(209), // >>>, reduce: PrecedenceChainEnd
			reduce(209), // +, reduce: PrecedenceChainEnd
			reduce(209), // -, reduce: PrecedenceChainEnd
			reduce(209), // .+, reduce: PrecedenceChainEnd
			reduce(209), // .-, reduce: PrecedenceChainEnd
			reduce(209), // *, reduce: PrecedenceChainEnd
			reduce(209), // /, reduce: PrecedenceChainEnd
			reduce(209), // //, reduce: PrecedenceChainEnd
			reduce(209), // %, reduce: PrecedenceChainEnd
			reduce(209), // .*, reduce: PrecedenceChainEnd
			reduce(209), // ./, reduce: PrecedenceChainEnd
			reduce(209), // .//, reduce: PrecedenceChainEnd
			reduce(209), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(209), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(2889), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			shift(2890), // [[
			shift(2891), // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1820
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(215), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(215), // >, reduce: MlrvalOrFunction
			reduce(215), // >>, reduce: MlrvalOrFunction
			reduce(215), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(215), // ||=, reduce: MlrvalOrFunction
			reduce(215), // ^^=, reduce: MlrvalOrFunction
			reduce(215), // &&=, reduce: MlrvalOrFunction
			reduce(215), // ??=, reduce: MlrvalOrFunction
			reduce(215), // ???=, reduce: MlrvalOrFunction
			reduce(215), // |=, reduce: MlrvalOrFunction
			reduce(215), // &=, reduce: MlrvalOrFunction
			reduce(215), // ^=, reduce: MlrvalOrFunction
			reduce(215), // <<=, reduce: MlrvalOrFunction
			reduce(215), // >>=, reduce: MlrvalOrFunction
			reduce(215), // >>>=, reduce: MlrvalOrFunction
			reduce(215), // +=, reduce: MlrvalOrFunction
			reduce(215), // .=, reduce: MlrvalOrFunction
			reduce(215), // -=, reduce: MlrvalOrFunction
			reduce(215), // *=, reduce: MlrvalOrFunction
			reduce(215), // /=, reduce: MlrvalOrFunction
			reduce(215), // //=, reduce: MlrvalOrFunction
			reduce(215), // %=, reduce: MlrvalOrFunction
			reduce(215), // **=, reduce: MlrvalOrFunction
			reduce(215), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(215), // ||, reduce: MlrvalOrFunction
			reduce(215), // ^^, reduce: MlrvalOrFunction
			reduce(215), // &&, reduce: MlrvalOrFunction
			reduce(215), // ??, reduce: MlrvalOrFunction
			reduce(215), // ???, reduce: MlrvalOrFunction
			reduce(215), // =~, reduce: MlrvalOrFunction
			reduce(215), // !=~, reduce: MlrvalOrFunction
			reduce(215), // ==, reduce: MlrvalOrFunction
			reduce(215), // !=, reduce: MlrvalOrFunction
			reduce(215), // <=>, reduce: MlrvalOrFunction
			reduce(215), // >=, reduce: MlrvalOrFunction
			reduce(215), // <, reduce: MlrvalOrFunction
			reduce(215), // <=, reduce: MlrvalOrFunction
			reduce(215), // ^, reduce: MlrvalOrFunction
			reduce(215), // &, reduce: MlrvalOrFunction
			reduce(215), // <<, reduce: MlrvalOrFunction
			reduce(215), // >>>, reduce: MlrvalOrFunction
			reduce(215), // +, reduce: MlrvalOrFunction
			reduce(215), // -, reduce: MlrvalOrFunction
			reduce(215), // .+, reduce: MlrvalOrFunction
			reduce(215), // .-, reduce: MlrvalOrFunction
			reduce(215), // *, reduce: MlrvalOrFunction
			reduce(215), // /, reduce: MlrvalOrFunction
			reduce(215), // //, reduce: MlrvalOrFunction
			reduce(215), // %, reduce: MlrvalOrFunction
			reduce(215), // .*, reduce: MlrvalOrFunction
			reduce(215), // ./, reduce: MlrvalOrFunction
			reduce(215), // .//, reduce: MlrvalOrFunction
			reduce(215), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(215), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(215), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(215), // [[, reduce: MlrvalOrFunction
			reduce(215), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1821
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(216), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(216), // >, reduce: MlrvalOrFunction
			reduce(216), // >>, reduce: MlrvalOrFunction
			reduce(216), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(216), // ||=, reduce: MlrvalOrFunction
			reduce(216), // ^^=, reduce: MlrvalOrFunction
			reduce(216), // &&=, reduce: MlrvalOrFunction
			reduce(216), // ??=, reduce: MlrvalOrFunction
			reduce(216), // ???=, reduce: MlrvalOrFunction
			reduce(216), // |=, reduce: MlrvalOrFunction
			reduce(216), // &=, reduce: MlrvalOrFunction
			reduce(216), // ^=, reduce: MlrvalOrFunction
			reduce(216), // <<=, reduce: MlrvalOrFunction
			reduce(216), // >>=, reduce: MlrvalOrFunction
			reduce(216), // >>>=, reduce: MlrvalOrFunction
			reduce(216), // +=, reduce: MlrvalOrFunction
			reduce(216), // .=, reduce: MlrvalOrFunction
			reduce(216), // -=, reduce: MlrvalOrFunction
			reduce(216), // *=, reduce: MlrvalOrFunction
			reduce(216), // /=, reduce: MlrvalOrFunction
			reduce(216), // //=, reduce: MlrvalOrFunction
			reduce(216), // %=, reduce: MlrvalOrFunction
			reduce(216), // **=, reduce: MlrvalOrFunction
			reduce(216), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(216), // ||, reduce: MlrvalOrFunction
			reduce(216), // ^^, reduce: MlrvalOrFunction
			reduce(216), // &&, reduce: MlrvalOrFunction
			reduce(216), // ??, reduce: MlrvalOrFunction
			reduce(216), // ???, reduce: MlrvalOrFunction
			reduce(216), // =~, reduce: MlrvalOrFunction
			reduce(216), // !=~, reduce: MlrvalOrFunction
			reduce(216), // ==, reduce: MlrvalOrFunction
			reduce(216), // !=, reduce: MlrvalOrFunction
			reduce(216), // <=>, reduce: MlrvalOrFunction
			reduce(216), // >=, reduce: MlrvalOrFunction
			reduce(216), // <, reduce: MlrvalOrFunction
			reduce(216), // <=, reduce: MlrvalOrFunction
			reduce(216), // ^, reduce: MlrvalOrFunction
			reduce(216), // &, reduce: MlrvalOrFunction
			reduce(216), // <<, reduce: MlrvalOrFunction
			reduce(216), // >>>, reduce: MlrvalOrFunction
			reduce(216), // +, reduce: MlrvalOrFunction
			reduce(216), // -, reduce: MlrvalOrFunction
			reduce(216), // .+, reduce: MlrvalOrFunction
			reduce(216), // .-, reduce: MlrvalOrFunction
			reduce(216), // *, reduce: MlrvalOrFunction
			reduce(216), // /, reduce: MlrvalOrFunction
			reduce(216), // //, reduce: MlrvalOrFunction
			reduce(216), // %, reduce: MlrvalOrFunction
			reduce(216), // .*, reduce: MlrvalOrFunction
			reduce(216), // ./, reduce: MlrvalOrFunction
			reduce(216), // .//, reduce: MlrvalOrFunction
			reduce(216), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(216), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(216), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(216), // [[, reduce: MlrvalOrFunction
			reduce(216), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1822
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(217), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(217), // >, reduce: MlrvalOrFunction
			reduce(217), // >>, reduce: MlrvalOrFunction
			reduce(217), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(217), // ||=, reduce: MlrvalOrFunction
			reduce(217), // ^^=, reduce: MlrvalOrFunction
			reduce(217), // &&=, reduce: MlrvalOrFunction
			reduce(217), // ??=, reduce: MlrvalOrFunction
			reduce(217), // ???=, reduce: MlrvalOrFunction
			reduce(217), // |=, reduce: MlrvalOrFunction
			reduce(217), // &=, reduce: MlrvalOrFunction
			reduce(217), // ^=, reduce: MlrvalOrFunction
			reduce(217), // <<=, reduce: MlrvalOrFunction
			reduce(217), // >>=, reduce: MlrvalOrFunction
			reduce(217), // >>>=, reduce: MlrvalOrFunction
			reduce(217), // +=, reduce: MlrvalOrFunction
			reduce(217), // .=, reduce: MlrvalOrFunction
			reduce(217), // -=, reduce: MlrvalOrFunction
			reduce(217), // *=, reduce: MlrvalOrFunction
			reduce(217), // /=, reduce: MlrvalOrFunction
			reduce(217), // //=, reduce: MlrvalOrFunction
			reduce(217), // %=, reduce: MlrvalOrFunction
			reduce(217), // **=, reduce: MlrvalOrFunction
			reduce(217), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(217), // ||, reduce: MlrvalOrFunction
			reduce(217), // ^^, reduce: MlrvalOrFunction
			reduce(217), // &&, reduce: MlrvalOrFunction
			reduce(217), // ??, reduce: MlrvalOrFunction
			reduce(217), // ???, reduce: MlrvalOrFunction
			reduce(217), // =~, reduce: MlrvalOrFunction
			reduce(217), // !=~, reduce: MlrvalOrFunction
			reduce(217), // ==, reduce: MlrvalOrFunction
			reduce(217), // !=, reduce: MlrvalOrFunction
			reduce(217), // <=>, reduce: MlrvalOrFunction
			reduce(217), // >=, reduce: MlrvalOrFunction
			reduce(217), // <, reduce: MlrvalOrFunction
			reduce(217), // <=, reduce: MlrvalOrFunction
			reduce(217), // ^, reduce: MlrvalOrFunction
			reduce(217), // &, reduce: MlrvalOrFunction
			reduce(217), // <<, reduce: MlrvalOrFunction
			reduce(217), // >>>, reduce: MlrvalOrFunction
			reduce(217), // +, reduce: MlrvalOrFunction
			reduce(217), // -, reduce: MlrvalOrFunction
			reduce(217), // .+, reduce: MlrvalOrFunction
			reduce(217), // .-, reduce: MlrvalOrFunction
			reduce(217), // *, reduce: MlrvalOrFunction
			reduce(217), // /, reduce: MlrvalOrFunction
			reduce(217), // //, reduce: MlrvalOrFunction
			reduce(217), // %, reduce: MlrvalOrFunction
			reduce(217), // .*, reduce: MlrvalOrFunction
			reduce(217), // ./, reduce: MlrvalOrFunction
			reduce(217), // .//, reduce: MlrvalOrFunction
			reduce(217), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(217), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(217), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(217), // [[, reduce: MlrvalOrFunction
			reduce(217), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1823
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(218), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(218), // >, reduce: MlrvalOrFunction
			reduce(218), // >>, reduce: MlrvalOrFunction
			reduce(218), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(218), // ||=, reduce: MlrvalOrFunction
			reduce(218), // ^^=, reduce: MlrvalOrFunction
			reduce(218), // &&=, reduce: MlrvalOrFunction
			reduce(218), // ??=, reduce: MlrvalOrFunction
			reduce(218), // ???=, reduce: MlrvalOrFunction
			reduce(218), // |=, reduce: MlrvalOrFunction
			reduce(218), // &=, reduce: MlrvalOrFunction
			reduce(218), // ^=, reduce: MlrvalOrFunction
			reduce(218), // <<=, reduce: MlrvalOrFunction
			reduce(218), // >>=, reduce: MlrvalOrFunction
			reduce(218), // >>>=, reduce: MlrvalOrFunction
			reduce(218), // +=, reduce: MlrvalOrFunction
			reduce(218), // .=, reduce: MlrvalOrFunction
			reduce(218), // -=, reduce: MlrvalOrFunction
			reduce(218), // *=, reduce: MlrvalOrFunction
			reduce(218), // /=, reduce: MlrvalOrFunction
			reduce(218), // //=, reduce: MlrvalOrFunction
			reduce(218), // %=, reduce: MlrvalOrFunction
			reduce(218), // **=, reduce: MlrvalOrFunction
			reduce(218), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(218), // ||, reduce: MlrvalOrFunction
			reduce(218), // ^^, reduce: MlrvalOrFunction
			reduce(218), // &&, reduce: MlrvalOrFunction
			reduce(218), // ??, reduce: MlrvalOrFunction
			reduce(218), // ???, reduce: MlrvalOrFunction
			reduce(218), // =~, reduce: MlrvalOrFunction
			reduce(218), // !=~, reduce: MlrvalOrFunction
			reduce(218), // ==, reduce: MlrvalOrFunction
			reduce(218), // !=, reduce: MlrvalOrFunction
			reduce(218), // <=>, reduce: MlrvalOrFunction
			reduce(218), // >=, reduce: MlrvalOrFunction
			reduce(218), // <, reduce: MlrvalOrFunction
			reduce(218), // <=, reduce: MlrvalOrFunction
			reduce(218), // ^, reduce: MlrvalOrFunction
			reduce(218), // &, reduce: MlrvalOrFunction
			reduce(218), // <<, reduce: MlrvalOrFunction
			reduce(218), // >>>, reduce: MlrvalOrFunction
			reduce(218), // +, reduce: MlrvalOrFunction
			reduce(218), // -, reduce: MlrvalOrFunction
			reduce(218), // .+, reduce: MlrvalOrFunction
			reduce(218), // .-, reduce: MlrvalOrFunction
			reduce(218), // *, reduce: MlrvalOrFunction
			reduce(218), // /, reduce: MlrvalOrFunction
			reduce(218), // //, reduce: MlrvalOrFunction
			reduce(218), // %, reduce: MlrvalOrFunction
			reduce(218), // .*, reduce: MlrvalOrFunction
			reduce(218), // ./, reduce: MlrvalOrFunction
			reduce(218), // .//, reduce: MlrvalOrFunction
			reduce(218), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(218), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(218), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(218), // [[, reduce: MlrvalOrFunction
			reduce(218), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1824
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(219), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(219), // >, reduce: MlrvalOrFunction
			reduce(219), // >>, reduce: MlrvalOrFunction
			reduce(219), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(219), // ||=, reduce: MlrvalOrFunction
			reduce(219), // ^^=, reduce: MlrvalOrFunction
			reduce(219), // &&=, reduce: MlrvalOrFunction
			reduce(219), // ??=, reduce: MlrvalOrFunction
			reduce(219), // ???=, reduce: MlrvalOrFunction
			reduce(219), // |=, reduce: MlrvalOrFunction
			reduce(219), // &=, reduce: MlrvalOrFunction
			reduce(219), // ^=, reduce: MlrvalOrFunction
			reduce(219), // <<=, reduce: MlrvalOrFunction
			reduce(219), // >>=, reduce: MlrvalOrFunction
			reduce(219), // >>>=, reduce: MlrvalOrFunction
			reduce(219), // +=, reduce: MlrvalOrFunction
			reduce(219), // .=, reduce: MlrvalOrFunction
			reduce(219), // -=, reduce: MlrvalOrFunction
			reduce(219), // *=, reduce: MlrvalOrFunction
			reduce(219), // /=, reduce: MlrvalOrFunction
			reduce(219), // //=, reduce: MlrvalOrFunction
			reduce(219), // %=, reduce: MlrvalOrFunction
			reduce(219), // **=, reduce: MlrvalOrFunction
			reduce(219), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(219), // ||, reduce: MlrvalOrFunction
			reduce(219), // ^^, reduce: MlrvalOrFunction
			reduce(219), // &&, reduce: MlrvalOrFunction
			reduce(219), // ??, reduce: MlrvalOrFunction
			reduce(219), // ???, reduce: MlrvalOrFunction
			reduce(219), // =~, reduce: MlrvalOrFunction
			reduce(219), // !=~, reduce: MlrvalOrFunction
			reduce(219), // ==, reduce: MlrvalOrFunction
			reduce(219), // !=, reduce: MlrvalOrFunction
			reduce(219), // <=>, reduce: MlrvalOrFunction
			reduce(219), // >=, reduce: MlrvalOrFunction
			reduce(219), // <, reduce: MlrvalOrFunction
			reduce(219), // <=, reduce: MlrvalOrFunction
			reduce(219), // ^, reduce: MlrvalOrFunction
			reduce(219), // &, reduce: MlrvalOrFunction
			reduce(219), // <<, reduce: MlrvalOrFunction
			reduce(219), // >>>, reduce: MlrvalOrFunction
			reduce(219), // +, reduce: MlrvalOrFunction
			reduce(219), // -, reduce: MlrvalOrFunction
			reduce(219), // .+, reduce: MlrvalOrFunction
			reduce(219), // .-, reduce: MlrvalOrFunction
			reduce(219), // *, reduce: MlrvalOrFunction
			reduce(219), // /, reduce: MlrvalOrFunction
			reduce(219), // //, reduce: MlrvalOrFunction
			reduce(219), // %, reduce: MlrvalOrFunction
			reduce(219), // .*, reduce: MlrvalOrFunction
			reduce(219), // ./, reduce: MlrvalOrFunction
			reduce(219), // .//, reduce: MlrvalOrFunction
			reduce(219), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(219), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(219), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(219), // [[, reduce: MlrvalOrFunction
			reduce(219), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1825
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(220), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(220), // >, reduce: MlrvalOrFunction
			reduce(220), // >>, reduce: MlrvalOrFunction
			reduce(220), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(220), // ||=, reduce: MlrvalOrFunction
			reduce(220), // ^^=, reduce: MlrvalOrFunction
			reduce(220), // &&=, reduce: MlrvalOrFunction
			reduce(220), // ??=, reduce: MlrvalOrFunction
			reduce(220), // ???=, reduce: MlrvalOrFunction
			reduce(220), // |=, reduce: MlrvalOrFunction
			reduce(220), // &=, reduce: MlrvalOrFunction
			reduce(220), // ^=, reduce: MlrvalOrFunction
			reduce(220), // <<=, reduce: MlrvalOrFunction
			reduce(220), // >>=, reduce: MlrvalOrFunction
			reduce(220), // >>>=, reduce: MlrvalOrFunction
			reduce(220), // +=, reduce: MlrvalOrFunction
			reduce(220), // .=, reduce: MlrvalOrFunction
			reduce(220), // -=, reduce: MlrvalOrFunction
			reduce(220), // *=, reduce: MlrvalOrFunction
			reduce(220), // /=, reduce: MlrvalOrFunction
			reduce(220), // //=, reduce: MlrvalOrFunction
			reduce(220), // %=, reduce: MlrvalOrFunction
			reduce(220), // **=, reduce: MlrvalOrFunction
			reduce(220), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(220), // ||, reduce: MlrvalOrFunction
			reduce(220), // ^^, reduce: MlrvalOrFunction
			reduce(220), // &&, reduce: MlrvalOrFunction
			reduce(220), // ??, reduce: MlrvalOrFunction
			reduce(220), // ???, reduce: MlrvalOrFunction
			reduce(220), // =~, reduce: MlrvalOrFunction
			reduce(220), // !=~, reduce: MlrvalOrFunction
			reduce(220), // ==, reduce: MlrvalOrFunction
			reduce(220), // !=, reduce: MlrvalOrFunction
			reduce(220), // <=>, reduce: MlrvalOrFunction
			reduce(220), // >=, reduce: MlrvalOrFunction
			reduce(220), // <, reduce: MlrvalOrFunction
			reduce(220), // <=, reduce: MlrvalOrFunction
			reduce(220), // ^, reduce: MlrvalOrFunction
			reduce(220), // &, reduce: MlrvalOrFunction
			reduce(220), // <<, reduce: MlrvalOrFunction
			reduce(220), // >>>, reduce: MlrvalOrFunction
			reduce(220), // +, reduce: MlrvalOrFunction
			reduce(220), // -, reduce: MlrvalOrFunction
			reduce(220), // .+, reduce: MlrvalOrFunction
			reduce(220), // .-, reduce: MlrvalOrFunction
			reduce(220), // *, reduce: MlrvalOrFunction
			reduce(220), // /, reduce: MlrvalOrFunction
			reduce(220), // //, reduce: MlrvalOrFunction
			reduce(220), // %, reduce: MlrvalOrFunction
			reduce(220), // .*, reduce: MlrvalOrFunction
			reduce(220), // ./, reduce: MlrvalOrFunction
			reduce(220), // .//, reduce: MlrvalOrFunction
			reduce(220), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(220), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(220), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(220), // [[, reduce: MlrvalOrFunction
			reduce(220), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1826
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(221), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(221), // >, reduce: MlrvalOrFunction
			reduce(221), // >>, reduce: MlrvalOrFunction
			reduce(221), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(221), // ||=, reduce: MlrvalOrFunction
			reduce(221), // ^^=, reduce: MlrvalOrFunction
			reduce(221), // &&=, reduce: MlrvalOrFunction
			reduce(221), // ??=, reduce: MlrvalOrFunction
			reduce(221), // ???=, reduce: MlrvalOrFunction
			reduce(221), // |=, reduce: MlrvalOrFunction
			reduce(221), // &=, reduce: MlrvalOrFunction
			reduce(221), // ^=, reduce: MlrvalOrFunction
			reduce(221), // <<=, reduce: MlrvalOrFunction
			reduce(221), // >>=, reduce: MlrvalOrFunction
			reduce(221), // >>>=, reduce: MlrvalOrFunction
			reduce(221), // +=, reduce: MlrvalOrFunction
			reduce(221), // .=, reduce: MlrvalOrFunction
			reduce(221), // -=, reduce: MlrvalOrFunction
			reduce(221), // *=, reduce: MlrvalOrFunction
			reduce(221), // /=, reduce: MlrvalOrFunction
			reduce(221), // //=, reduce: MlrvalOrFunction
			reduce(221), // %=, reduce: MlrvalOrFunction
			reduce(221), // **=, reduce: MlrvalOrFunction
			reduce(221), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(221), // ||, reduce: MlrvalOrFunction
			reduce(221), // ^^, reduce: MlrvalOrFunction
			reduce(221), // &&, reduce: MlrvalOrFunction
			reduce(221), // ??, reduce: MlrvalOrFunction
			reduce(221), // ???, reduce: MlrvalOrFunction
			reduce(221), // =~, reduce: MlrvalOrFunction
			reduce(221), // !=~, reduce: MlrvalOrFunction
			reduce(221), // ==, reduce: MlrvalOrFunction
			reduce(221), // !=, reduce: MlrvalOrFunction
			reduce(221), // <=>, reduce: MlrvalOrFunction
			reduce(221), // >=, reduce: MlrvalOrFunction
			reduce(221), // <, reduce: MlrvalOrFunction
			reduce(221), // <=, reduce: MlrvalOrFunction
			reduce(221), // ^, reduce: MlrvalOrFunction
			reduce(221), // &, reduce: MlrvalOrFunction
			reduce(221), // <<, reduce: MlrvalOrFunction
			reduce(221), // >>>, reduce: MlrvalOrFunction
			reduce(221), // +, reduce: MlrvalOrFunction
			reduce(221), // -, reduce: MlrvalOrFunction
			reduce(221), // .+, reduce: MlrvalOrFunction
			reduce(221), // .-, reduce: MlrvalOrFunction
			reduce(221), // *, reduce: MlrvalOrFunction
			reduce(221), // /, reduce: MlrvalOrFunction
			reduce(221), // //, reduce: MlrvalOrFunction
			reduce(221), // %, reduce: MlrvalOrFunction
			reduce(221), // .*, reduce: MlrvalOrFunction
			reduce(221), // ./, reduce: MlrvalOrFunction
			reduce(221), // .//, reduce: MlrvalOrFunction
			reduce(221), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(221), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(221), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(221), // [[, reduce: MlrvalOrFunction
			reduce(221), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1827
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(222), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(222), // >, reduce: MlrvalOrFunction
			reduce(222), // >>, reduce: MlrvalOrFunction
			reduce(222), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(222), // ||=, reduce: MlrvalOrFunction
			reduce(222), // ^^=, reduce: MlrvalOrFunction
			reduce(222), // &&=, reduce: MlrvalOrFunction
			reduce(222), // ??=, reduce: MlrvalOrFunction
			reduce(222), // ???=, reduce: MlrvalOrFunction
			reduce(222), // |=, reduce: MlrvalOrFunction
			reduce(222), // &=, reduce: MlrvalOrFunction
			reduce(222), // ^=, reduce: MlrvalOrFunction
			reduce(222), // <<=, reduce: MlrvalOrFunction
			reduce(222), // >>=, reduce: MlrvalOrFunction
			reduce(222), // >>>=, reduce: MlrvalOrFunction
			reduce(222), // +=, reduce: MlrvalOrFunction
			reduce(222), // .=, reduce: MlrvalOrFunction
			reduce(222), // -=, reduce: MlrvalOrFunction
			reduce(222), // *=, reduce: MlrvalOrFunction
			reduce(222), // /=, reduce: MlrvalOrFunction
			reduce(222), // //=, reduce: MlrvalOrFunction
			reduce(222), // %=, reduce: MlrvalOrFunction
			reduce(222), // **=, reduce: MlrvalOrFunction
			reduce(222), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(222), // ||, reduce: MlrvalOrFunction
			reduce(222), // ^^, reduce: MlrvalOrFunction
			reduce(222), // &&, reduce: MlrvalOrFunction
			reduce(222), // ??, reduce: MlrvalOrFunction
			reduce(222), // ???, reduce: MlrvalOrFunction
			reduce(222), // =~, reduce: MlrvalOrFunction
			reduce(222), // !=~, reduce: MlrvalOrFunction
			reduce(222), // ==, reduce: MlrvalOrFunction
			reduce(222), // !=, reduce: MlrvalOrFunction
			reduce(222), // <=>, reduce: MlrvalOrFunction
			reduce(222), // >=, reduce: MlrvalOrFunction
			reduce(222), // <, reduce: MlrvalOrFunction
			reduce(222), // <=, reduce: MlrvalOrFunction
			reduce(222), // ^, reduce: MlrvalOrFunction
			reduce(222), // &, reduce: MlrvalOrFunction
			reduce(222), // <<, reduce: MlrvalOrFunction
			reduce(222), // >>>, reduce: MlrvalOrFunction
			reduce(222), // +, reduce: MlrvalOrFunction
			reduce(222), // -, reduce: MlrvalOrFunction
			reduce(222), // .+, reduce: MlrvalOrFunction
			reduce(222), // .-, reduce: MlrvalOrFunction
			reduce(222), // *, reduce: MlrvalOrFunction
			reduce(222), // /, reduce: MlrvalOrFunction
			reduce(222), // //, reduce: MlrvalOrFunction
			reduce(222), // %, reduce: MlrvalOrFunction
			reduce(222), // .*, reduce: MlrvalOrFunction
			reduce(222), // ./, reduce: MlrvalOrFunction
			reduce(222), // .//, reduce: MlrvalOrFunction
			reduce(222), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(222), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(222), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(222), // [[, reduce: MlrvalOrFunction
			reduce(222), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1828
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(223), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(223), // >, reduce: MlrvalOrFunction
			reduce(223), // >>, reduce: MlrvalOrFunction
			reduce(223), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(223), // ||=, reduce: MlrvalOrFunction
			reduce(223), // ^^=, reduce: MlrvalOrFunction
			reduce(223), // &&=, reduce: MlrvalOrFunction
			reduce(223), // ??=, reduce: MlrvalOrFunction
			reduce(223), // ???=, reduce: MlrvalOrFunction
			reduce(223), // |=, reduce: MlrvalOrFunction
			reduce(223), // &=, reduce: MlrvalOrFunction
			reduce(223), // ^=, reduce: MlrvalOrFunction
			reduce(223), // <<=, reduce: MlrvalOrFunction
			reduce(223), // >>=, reduce: MlrvalOrFunction
			reduce(223), // >>>=, reduce: MlrvalOrFunction
			reduce(223), // +=, reduce: MlrvalOrFunction
			reduce(223), // .=, reduce: MlrvalOrFunction
			reduce(223), // -=, reduce: MlrvalOrFunction
			reduce(223), // *=, reduce: MlrvalOrFunction
			reduce(223), // /=, reduce: MlrvalOrFunction
			reduce(223), // //=, reduce: MlrvalOrFunction
			reduce(223), // %=, reduce: MlrvalOrFunction
			reduce(223), // **=, reduce: MlrvalOrFunction
			reduce(223), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(223), // ||, reduce: MlrvalOrFunction
			reduce(223), // ^^, reduce: MlrvalOrFunction
			reduce(223), // &&, reduce: MlrvalOrFunction
			reduce(223), // ??, reduce: MlrvalOrFunction
			reduce(223), // ???, reduce: MlrvalOrFunction
			reduce(223), // =~, reduce: MlrvalOrFunction
			reduce(223), // !=~, reduce: MlrvalOrFunction
			reduce(223), // ==, reduce: MlrvalOrFunction
			reduce(223), // !=, reduce: MlrvalOrFunction
			reduce(223), // <=>, reduce: MlrvalOrFunction
			reduce(223), // >=, reduce: MlrvalOrFunction
			reduce(223), // <, reduce: MlrvalOrFunction
			reduce(223), // <=, reduce: MlrvalOrFunction
			reduce(223), // ^, reduce: MlrvalOrFunction
			reduce(223), // &, reduce: MlrvalOrFunction
			reduce(223), // <<, reduce: MlrvalOrFunction
			reduce(223), // >>>, reduce: MlrvalOrFunction
			reduce(223), // +, reduce: MlrvalOrFunction
			reduce(223), // -, reduce: MlrvalOrFunction
			reduce(223), // .+, reduce: MlrvalOrFunction
			reduce(223), // .-, reduce: MlrvalOrFunction
			reduce(223), // *, reduce: MlrvalOrFunction
			reduce(223), // /, reduce: MlrvalOrFunction
			reduce(223), // //, reduce: MlrvalOrFunction
			reduce(223), // %, reduce: MlrvalOrFunction
			reduce(223), // .*, reduce: MlrvalOrFunction
			reduce(223), // ./, reduce: MlrvalOrFunction
			reduce(223), // .//, reduce: MlrvalOrFunction
			reduce(223), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(223), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(223), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(223), // [[, reduce: MlrvalOrFunction
			reduce(223), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1829
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(224), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(224), // >, reduce: MlrvalOrFunction
			reduce(224), // >>, reduce: MlrvalOrFunction
			reduce(224), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(224), // ||=, reduce: MlrvalOrFunction
			reduce(224), // ^^=, reduce: MlrvalOrFunction
			reduce(224), // &&=, reduce: MlrvalOrFunction
			reduce(224), // ??=, reduce: MlrvalOrFunction
			reduce(224), // ???=, reduce: MlrvalOrFunction
			reduce(224), // |=, reduce: MlrvalOrFunction
			reduce(224), // &=, reduce: MlrvalOrFunction
			reduce(224), // ^=, reduce: MlrvalOrFunction
			reduce(224), // <<=, reduce: MlrvalOrFunction
			reduce(224), // >>=, reduce: MlrvalOrFunction
			reduce(224), // >>>=, reduce: MlrvalOrFunction
			reduce(224), // +=, reduce: MlrvalOrFunction
			reduce(224), // .=, reduce: MlrvalOrFunction
			reduce(224), // -=, reduce: MlrvalOrFunction
			reduce(224), // *=, reduce: MlrvalOrFunction
			reduce(224), // /=, reduce: MlrvalOrFunction
			reduce(224), // //=, reduce: MlrvalOrFunction
			reduce(224), // %=, reduce: MlrvalOrFunction
			reduce(224), // **=, reduce: MlrvalOrFunction
			reduce(224), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(224), // ||, reduce: MlrvalOrFunction
			reduce(224), // ^^, reduce: MlrvalOrFunction
			reduce(224), // &&, reduce: MlrvalOrFunction
			reduce(224), // ??, reduce: MlrvalOrFunction
			reduce(224), // ???, reduce: MlrvalOrFunction
			reduce(224), // =~, reduce: MlrvalOrFunction
			reduce(224), // !=~, reduce: MlrvalOrFunction
			reduce(224), // ==, reduce: MlrvalOrFunction
			reduce(224), // !=, reduce: MlrvalOrFunction
			reduce(224), // <=>, reduce: MlrvalOrFunction
			reduce(224), // >=, reduce: MlrvalOrFunction
			reduce(224), // <, reduce: MlrvalOrFunction
			reduce(224), // <=, reduce: MlrvalOrFunction
			reduce(224), // ^, reduce: MlrvalOrFunction
			reduce(224), // &, reduce: MlrvalOrFunction
			reduce(224), // <<, reduce: MlrvalOrFunction
			reduce(224), // >>>, reduce: MlrvalOrFunction
			reduce(224), // +, reduce: MlrvalOrFunction
			reduce(224), // -, reduce: MlrvalOrFunction
			reduce(224), // .+, reduce: MlrvalOrFunction
			reduce(224), // .-, reduce: MlrvalOrFunction
			reduce(224), // *, reduce: MlrvalOrFunction
			reduce(224), // /, reduce: MlrvalOrFunction
			reduce(224), // //, reduce: MlrvalOrFunction
			reduce(224), // %, reduce: MlrvalOrFunction
			reduce(224), // .*, reduce: MlrvalOrFunction
			reduce(224), // ./, reduce: MlrvalOrFunction
			reduce(224), // .//, reduce: MlrvalOrFunction
			reduce(224), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(224), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(224), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(224), // [[, reduce: MlrvalOrFunction
			reduce(224), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1830
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(225), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(225), // >, reduce: MlrvalOrFunction
			reduce(225), // >>, reduce: MlrvalOrFunction
			reduce(225), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(225), // ||=, reduce: MlrvalOrFunction
			reduce(225), // ^^=, reduce: MlrvalOrFunction
			reduce(225), // &&=, reduce: MlrvalOrFunction
			reduce(225), // ??=, reduce: MlrvalOrFunction
			reduce(225), // ???=, reduce: MlrvalOrFunction
			reduce(225), // |=, reduce: MlrvalOrFunction
			reduce(225), // &=, reduce: MlrvalOrFunction
			reduce(225), // ^=, reduce: MlrvalOrFunction
			reduce(225), // <<=, reduce: MlrvalOrFunction
			reduce(225), // >>=, reduce: MlrvalOrFunction
			reduce(225), // >>>=, reduce: MlrvalOrFunction
			reduce(225), // +=, reduce: MlrvalOrFunction
			reduce(225), // .=, reduce: MlrvalOrFunction
			reduce(225), // -=, reduce: MlrvalOrFunction
			reduce(225), // *=, reduce: MlrvalOrFunction
			reduce(225), // /=, reduce: MlrvalOrFunction
			reduce(225), // //=, reduce: MlrvalOrFunction
			reduce(225), // %=, reduce: MlrvalOrFunction
			reduce(225), // **=, reduce: MlrvalOrFunction
			reduce(225), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(225), // ||, reduce: MlrvalOrFunction
			reduce(225), // ^^, reduce: MlrvalOrFunction
			reduce(225), // &&, reduce: MlrvalOrFunction
			reduce(225), // ??, reduce: MlrvalOrFunction
			reduce(225), // ???, reduce: MlrvalOrFunction
			reduce(225), // =~, reduce: MlrvalOrFunction
			reduce(225), // !=~, reduce: MlrvalOrFunction
			reduce(225), // ==, reduce: MlrvalOrFunction
			reduce(225), // !=, reduce: MlrvalOrFunction
			reduce(225), // <=>, reduce: MlrvalOrFunction
			reduce(225), // >=, reduce: MlrvalOrFunction
			reduce(225), // <, reduce: MlrvalOrFunction
			reduce(225), // <=, reduce: MlrvalOrFunction
			reduce(225), // ^, reduce: MlrvalOrFunction
			reduce(225), // &, reduce: MlrvalOrFunction
			reduce(225), // <<, reduce: MlrvalOrFunction
			reduce(225), // >>>, reduce: MlrvalOrFunction
			reduce(225), // +, reduce: MlrvalOrFunction
			reduce(225), // -, reduce: MlrvalOrFunction
			reduce(225), // .+, reduce: MlrvalOrFunction
			reduce(225), // .-, reduce: MlrvalOrFunction
			reduce(225), // *, reduce: MlrvalOrFunction
			reduce(225), // /, reduce: MlrvalOrFunction
			reduce(225), // //, reduce: MlrvalOrFunction
			reduce(225), // %, reduce: MlrvalOrFunction
			reduce(225), // .*, reduce: MlrvalOrFunction
			reduce(225), // ./, reduce: MlrvalOrFunction
			reduce(225), // .//, reduce: MlrvalOrFunction
			reduce(225), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(225), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(225), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(225), // [[, reduce: MlrvalOrFunction
			reduce(225), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1831
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(226), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(226), // >, reduce: MlrvalOrFunction
			reduce(226), // >>, reduce: MlrvalOrFunction
			reduce(226), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(226), // ||=, reduce: MlrvalOrFunction
			reduce(226), // ^^=, reduce: MlrvalOrFunction
			reduce(226), // &&=, reduce: MlrvalOrFunction
			reduce(226), // ??=, reduce: MlrvalOrFunction
			reduce(226), // ???=, reduce: MlrvalOrFunction
			reduce(226), // |=, reduce: MlrvalOrFunction
			reduce(226), // &=, reduce: MlrvalOrFunction
			reduce(226), // ^=, reduce: MlrvalOrFunction
			reduce(226), // <<=, reduce: MlrvalOrFunction
			reduce(226), // >>=, reduce: MlrvalOrFunction
			reduce(226), // >>>=, reduce: MlrvalOrFunction
			reduce(226), // +=, reduce: MlrvalOrFunction
			reduce(226), // .=, reduce: MlrvalOrFunction
			reduce(226), // -=, reduce: MlrvalOrFunction
			reduce(226), // *=, reduce: MlrvalOrFunction
			reduce(226), // /=, reduce: MlrvalOrFunction
			reduce(226), // //=, reduce: MlrvalOrFunction
			reduce(226), // %=, reduce: MlrvalOrFunction
			reduce(226), // **=, reduce: MlrvalOrFunction
			reduce(226), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(226), // ||, reduce: MlrvalOrFunction
			reduce(226), // ^^, reduce: MlrvalOrFunction
			reduce(226), // &&, reduce: MlrvalOrFunction
			reduce(226), // ??, reduce: MlrvalOrFunction
			reduce(226), // ???, reduce: MlrvalOrFunction
			reduce(226), // =~, reduce: MlrvalOrFunction
			reduce(226), // !=~, reduce: MlrvalOrFunction
			reduce(226), // ==, reduce: MlrvalOrFunction
			reduce(226), // !=, reduce: MlrvalOrFunction
			reduce(226), // <=>, reduce: MlrvalOrFunction
			reduce(226), // >=, reduce: MlrvalOrFunction
			reduce(226), // <, reduce: MlrvalOrFunction
			reduce(226), // <=, reduce: MlrvalOrFunction
			reduce(226), // ^, reduce: MlrvalOrFunction
			reduce(226), // &, reduce: MlrvalOrFunction
			reduce(226), // <<, reduce: MlrvalOrFunction
			reduce(226), // >>>, reduce: MlrvalOrFunction
			reduce(226), // +, reduce: MlrvalOrFunction
			reduce(226), // -, reduce: MlrvalOrFunction
			reduce(226), // .+, reduce: MlrvalOrFunction
			reduce(226), // .-, reduce: MlrvalOrFunction
			reduce(226), // *, reduce: MlrvalOrFunction
			reduce(226), // /, reduce: MlrvalOrFunction
			reduce(226), // //, reduce: MlrvalOrFunction
			reduce(226), // %, reduce: MlrvalOrFunction
			reduce(226), // .*, reduce: MlrvalOrFunction
			reduce(226), // ./, reduce: MlrvalOrFunction
			reduce(226), // .//, reduce: MlrvalOrFunction
			reduce(226), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(226), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(226), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(226), // [[, reduce: MlrvalOrFunction
			reduce(226), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1832
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(227), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(227), // >, reduce: MlrvalOrFunction
			reduce(227), // >>, reduce: MlrvalOrFunction
			reduce(227), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(227), // ||=, reduce: MlrvalOrFunction
			reduce(227), // ^^=, reduce: MlrvalOrFunction
			reduce(227), // &&=, reduce: MlrvalOrFunction
			reduce(227), // ??=, reduce: MlrvalOrFunction
			reduce(227), // ???=, reduce: MlrvalOrFunction
			reduce(227), // |=, reduce: MlrvalOrFunction
			reduce(227), // &=, reduce: MlrvalOrFunction
			reduce(227), // ^=, reduce: MlrvalOrFunction
			reduce(227), // <<=, reduce: MlrvalOrFunction
			reduce(227), // >>=, reduce: MlrvalOrFunction
			reduce(227), // >>>=, reduce: MlrvalOrFunction
			reduce(227), // +=, reduce: MlrvalOrFunction
			reduce(227), // .=, reduce: MlrvalOrFunction
			reduce(227), // -=, reduce: MlrvalOrFunction
			reduce(227), // *=, reduce: MlrvalOrFunction
			reduce(227), // /=, reduce: MlrvalOrFunction
			reduce(227), // //=, reduce: MlrvalOrFunction
			reduce(227), // %=, reduce: MlrvalOrFunction
			reduce(227), // **=, reduce: MlrvalOrFunction
			reduce(227), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(227), // ||, reduce: MlrvalOrFunction
			reduce(227), // ^^, reduce: MlrvalOrFunction
			reduce(227), // &&, reduce: MlrvalOrFunction
			reduce(227), // ??, reduce: MlrvalOrFunction
			reduce(227), // ???, reduce: MlrvalOrFunction
			reduce(227), // =~, reduce: MlrvalOrFunction
			reduce(227), // !=~, reduce: MlrvalOrFunction
			reduce(227), // ==, reduce: MlrvalOrFunction
			reduce(227), // !=, reduce: MlrvalOrFunction
			reduce(227), // <=>, reduce: MlrvalOrFunction
			reduce(227), // >=, reduce: MlrvalOrFunction
			reduce(227), // <, reduce: MlrvalOrFunction
			reduce(227), // <=, reduce: MlrvalOrFunction
			reduce(227), // ^, reduce: MlrvalOrFunction
			reduce(227), // &, reduce: MlrvalOrFunction
			reduce(227), // <<, reduce: MlrvalOrFunction
			reduce(227), // >>>, reduce: MlrvalOrFunction
			reduce(227), // +, reduce: MlrvalOrFunction
			reduce(227), // -, reduce: MlrvalOrFunction
			reduce(227), // .+, reduce: MlrvalOrFunction
			reduce(227), // .-, reduce: MlrvalOrFunction
			reduce(227), // *, reduce: MlrvalOrFunction
			reduce(227), // /, reduce: MlrvalOrFunction
			reduce(227), // //, reduce: MlrvalOrFunction
			reduce(227), // %, reduce: MlrvalOrFunction
			reduce(227), // .*, reduce: MlrvalOrFunction
			reduce(227), // ./, reduce: MlrvalOrFunction
			reduce(227), // .//, reduce: MlrvalOrFunction
			reduce(227), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(227), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(227), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(227), // [[, reduce: MlrvalOrFunction
			reduce(227), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1833
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(752),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(756),  // (
			nil,         // )
			nil,         // emitp
			shift(768),  // field_name
			shift(769),  // $[
			shift(2892), // ]
			shift(771),  // braced_field_name
			shift(772),  // $[[
			shift(773),  // $[[[
			shift(774),  // full_srec
			shift(776),  // oosvar_name
			shift(777),  // @[
			shift(778),  // braced_oosvar_name
			shift(779),  // full_oosvar
			shift(780),  // all
			shift(781),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(796),  // +
			shift(798),  // -
			shift(799),  // .+
			shift(800),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(803),  // !
			shift(804),  // ~
			nil,         // **
			shift(809),  // string_literal
			shift(810),  // regex_case_insensitive
			shift(811),  // int_literal
			shift(812),  // float_literal
			shift(813),  // boolean_literal
			shift(814),  // null_literal
			shift(815),  // inf_literal
			shift(816),  // nan_literal
			shift(817),  // const_M_PI
			shift(818),  // const_M_E
			shift(819),  // panic
			shift(821),  // [
			shift(824),  // ctx_IPS
			shift(825),  // ctx_IFS
			shift(826),  // ctx_IRS
			shift(827),  // ctx_OPS
			shift(828),  // ctx_OFS
			shift(829),  // ctx_ORS
			shift(830),  // ctx_FLATSEP
			shift(831),  // ctx_NF
			shift(832),  // ctx_NR
			shift(833),  // ctx_FNR
			shift(834),  // ctx_FILENAME
			shift(835),  // ctx_FILENUM
			shift(837),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(844),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1834
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(240), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(240), // >, reduce: MlrvalOrFunction
			reduce(240), // >>, reduce: MlrvalOrFunction
			reduce(240), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(240), // ||=, reduce: MlrvalOrFunction
			reduce(240), // ^^=, reduce: MlrvalOrFunction
			reduce(240), // &&=, reduce: MlrvalOrFunction
			reduce(240), // ??=, reduce: MlrvalOrFunction
			reduce(240), // ???=, reduce: MlrvalOrFunction
			reduce(240), // |=, reduce: MlrvalOrFunction
			reduce(240), // &=, reduce: MlrvalOrFunction
			reduce(240), // ^=, reduce: MlrvalOrFunction
			reduce(240), // <<=, reduce: MlrvalOrFunction
			reduce(240), // >>=, reduce: MlrvalOrFunction
			reduce(240), // >>>=, reduce: MlrvalOrFunction
			reduce(240), // +=, reduce: MlrvalOrFunction
			reduce(240), // .=, reduce: MlrvalOrFunction
			reduce(240), // -=, reduce: MlrvalOrFunction
			reduce(240), // *=, reduce: MlrvalOrFunction
			reduce(240), // /=, reduce: MlrvalOrFunction
			reduce(240), // //=, reduce: MlrvalOrFunction
			reduce(240), // %=, reduce: MlrvalOrFunction
			reduce(240), // **=, reduce: MlrvalOrFunction
			reduce(240), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(240), // ||, reduce: MlrvalOrFunction
			reduce(240), // ^^, reduce: MlrvalOrFunction
			reduce(240), // &&, reduce: MlrvalOrFunction
			reduce(240), // ??, reduce: MlrvalOrFunction
			reduce(240), // ???, reduce: MlrvalOrFunction
			reduce(240), // =~, reduce: MlrvalOrFunction
			reduce(240), // !=~, reduce: MlrvalOrFunction
			reduce(240), // ==, reduce: MlrvalOrFunction
			reduce(240), // !=, reduce: MlrvalOrFunction
			reduce(240), // <=>, reduce: MlrvalOrFunction
			reduce(240), // >=, reduce: MlrvalOrFunction
			reduce(240), // <, reduce: MlrvalOrFunction
			reduce(240), // <=, reduce: MlrvalOrFunction
			reduce(240), // ^, reduce: MlrvalOrFunction
			reduce(240), // &, reduce: MlrvalOrFunction
			reduce(240), // <<, reduce: MlrvalOrFunction
			reduce(240), // >>>, reduce: MlrvalOrFunction
			reduce(240), // +, reduce: MlrvalOrFunction
			reduce(240), // -, reduce: MlrvalOrFunction
			reduce(240), // .+, reduce: MlrvalOrFunction
			reduce(240), // .-, reduce: MlrvalOrFunction
			reduce(240), // *, reduce: MlrvalOrFunction
			reduce(240), // /, reduce: MlrvalOrFunction
			reduce(240), // //, reduce: MlrvalOrFunction
			reduce(240), // %, reduce: MlrvalOrFunction
			reduce(240), // .*, reduce: MlrvalOrFunction
			reduce(240), // ./, reduce: MlrvalOrFunction
			reduce(240), // .//, reduce: MlrvalOrFunction
			reduce(240), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(240), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(240), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(240), // [[, reduce: MlrvalOrFunction
			reduce(240), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1835
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(241), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(241), // >, reduce: ContextVariable
			reduce(241), // >>, reduce: ContextVariable
			reduce(241), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(241), // ||=, reduce: ContextVariable
			reduce(241), // ^^=, reduce: ContextVariable
			reduce(241), // &&=, reduce: ContextVariable
			reduce(241), // ??=, reduce: ContextVariable
			reduce(241), // ???=, reduce: ContextVariable
			reduce(241), // |=, reduce: ContextVariable
			reduce(241), // &=, reduce: ContextVariable
			reduce(241), // ^=, reduce: ContextVariable
			reduce(241), // <<=, reduce: ContextVariable
			reduce(241), // >>=, reduce: ContextVariable
			reduce(241), // >>>=, reduce: ContextVariable
			reduce(241), // +=, reduce: ContextVariable
			reduce(241), // .=, reduce: ContextVariable
			reduce(241), // -=, reduce: ContextVariable
			reduce(241), // *=, reduce: ContextVariable
			reduce(241), // /=, reduce: ContextVariable
			reduce(241), // //=, reduce: ContextVariable
			reduce(241), // %=, reduce: ContextVariable
			reduce(241), // **=, reduce: ContextVariable
			reduce(241), // ?, reduce: ContextVariable
			nil,         // :
			reduce(241), // ||, reduce: ContextVariable
			reduce(241), // ^^, reduce: ContextVariable
			reduce(241), // &&, reduce: ContextVariable
			reduce(241), // ??, reduce: ContextVariable
			reduce(241), // ???, reduce: ContextVariable
			reduce(241), // =~, reduce: ContextVariable
			reduce(241), // !=~, reduce: ContextVariable
			reduce(241), // ==, reduce: ContextVariable
			reduce(241), // !=, reduce: ContextVariable
			reduce(241), // <=>, reduce: ContextVariable
			reduce(241), // >=, reduce: ContextVariable
			reduce(241), // <, reduce: ContextVariable
			reduce(241), // <=, reduce: ContextVariable
			reduce(241), // ^, reduce: ContextVariable
			reduce(241), // &, reduce: ContextVariable
			reduce(241), // <<, reduce: ContextVariable
			reduce(241), // >>>, reduce: ContextVariable
			reduce(241), // +, reduce: ContextVariable
			reduce(241), // -, reduce: ContextVariable
			reduce(241), // .+, reduce: ContextVariable
			reduce(241), // .-, reduce: ContextVariable
			reduce(241), // *, reduce: ContextVariable
			reduce(241), // /, reduce: ContextVariable
			reduce(241), // //, reduce: ContextVariable
			reduce(241), // %, reduce: ContextVariable
			reduce(241), // .*, reduce: ContextVariable
			reduce(241), // ./, reduce: ContextVariable
			reduce(241), // .//, reduce: ContextVariable
			reduce(241), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(241), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(241), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(241), // [[, reduce: ContextVariable
			reduce(241), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1836
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(242), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(242), // >, reduce: ContextVariable
			reduce(242), // >>, reduce: ContextVariable
			reduce(242), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(242), // ||=, reduce: ContextVariable
			reduce(242), // ^^=, reduce: ContextVariable
			reduce(242), // &&=, reduce: ContextVariable
			reduce(242), // ??=, reduce: ContextVariable
			reduce(242), // ???=, reduce: ContextVariable
			reduce(242), // |=, reduce: ContextVariable
			reduce(242), // &=, reduce: ContextVariable
			reduce(242), // ^=, reduce: ContextVariable
			reduce(242), // <<=, reduce: ContextVariable
			reduce(242), // >>=, reduce: ContextVariable
			reduce(242), // >>>=, reduce: ContextVariable
			reduce(242), // +=, reduce: ContextVariable
			reduce(242), // .=, reduce: ContextVariable
			reduce(242), // -=, reduce: ContextVariable
			reduce(242), // *=, reduce: ContextVariable
			reduce(242), // /=, reduce: ContextVariable
			reduce(242), // //=, reduce: ContextVariable
			reduce(242), // %=, reduce: ContextVariable
			reduce(242), // **=, reduce: ContextVariable
			reduce(242), // ?, reduce: ContextVariable
			nil,         // :
			reduce(242), // ||, reduce: ContextVariable
			reduce(242), // ^^, reduce: ContextVariable
			reduce(242), // &&, reduce: ContextVariable
			reduce(242), // ??, reduce: ContextVariable
			reduce(242), // ???, reduce: ContextVariable
			reduce(242), // =~, reduce: ContextVariable
			reduce(242), // !=~, reduce: ContextVariable
			reduce(242), // ==, reduce: ContextVariable
			reduce(242), // !=, reduce: ContextVariable
			reduce(242), // <=>, reduce: ContextVariable
			reduce(242), // >=, reduce: ContextVariable
			reduce(242), // <, reduce: ContextVariable
			reduce(242), // <=, reduce: ContextVariable
			reduce(242), // ^, reduce: ContextVariable
			reduce(242), // &, reduce: ContextVariable
			reduce(242), // <<, reduce: ContextVariable
			reduce(242), // >>>, reduce: ContextVariable
			reduce(242), // +, reduce: ContextVariable
			reduce(242), // -, reduce: ContextVariable
			reduce(242), // .+, reduce: ContextVariable
			reduce(242), // .-, reduce: ContextVariable
			reduce(242), // *, reduce: ContextVariable
			reduce(242), // /, reduce: ContextVariable
			reduce(242), // //, reduce: ContextVariable
			reduce(242), // %, reduce: ContextVariable
			reduce(242), // .*, reduce: ContextVariable
			reduce(242), // ./, reduce: ContextVariable
			reduce(242), // .//, reduce: ContextVariable
			reduce(242), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(242), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(242), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(242), // [[, reduce: ContextVariable
			reduce(242), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1837
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(243), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(243), // >, reduce: ContextVariable
			reduce(243), // >>, reduce: ContextVariable
			reduce(243), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(243), // ||=, reduce: ContextVariable
			reduce(243), // ^^=, reduce: ContextVariable
			reduce(243), // &&=, reduce: ContextVariable
			reduce(243), // ??=, reduce: ContextVariable
			reduce(243), // ???=, reduce: ContextVariable
			reduce(243), // |=, reduce: ContextVariable
			reduce(243), // &=, reduce: ContextVariable
			reduce(243), // ^=, reduce: ContextVariable
			reduce(243), // <<=, reduce: ContextVariable
			reduce(243), // >>=, reduce: ContextVariable
			reduce(243), // >>>=, reduce: ContextVariable
			reduce(243), // +=, reduce: ContextVariable
			reduce(243), // .=, reduce: ContextVariable
			reduce(243), // -=, reduce: ContextVariable
			reduce(243), // *=, reduce: ContextVariable
			reduce(243), // /=, reduce: ContextVariable
			reduce(243), // //=, reduce: ContextVariable
			reduce(243), // %=, reduce: ContextVariable
			reduce(243), // **=, reduce: ContextVariable
			reduce(243), // ?, reduce: ContextVariable
			nil,         // :
			reduce(243), // ||, reduce: ContextVariable
			reduce(243), // ^^, reduce: ContextVariable
			reduce(243), // &&, reduce: ContextVariable
			reduce(243), // ??, reduce: ContextVariable
			reduce(243), // ???, reduce: ContextVariable
			reduce(243), // =~, reduce: ContextVariable
			reduce(243), // !=~, reduce: ContextVariable
			reduce(243), // ==, reduce: ContextVariable
			reduce(243), // !=, reduce: ContextVariable
			reduce(243), // <=>, reduce: ContextVariable
			reduce(243), // >=, reduce: ContextVariable
			reduce(243), // <, reduce: ContextVariable
			reduce(243), // <=, reduce: ContextVariable
			reduce(243), // ^, reduce: ContextVariable
			reduce(243), // &, reduce: ContextVariable
			reduce(243), // <<, reduce: ContextVariable
			reduce(243), // >>>, reduce: ContextVariable
			reduce(243), // +, reduce: ContextVariable
			reduce(243), // -, reduce: ContextVariable
			reduce(243), // .+, reduce: ContextVariable
			reduce(243), // .-, reduce: ContextVariable
			reduce(243), // *, reduce: ContextVariable
			reduce(243), // /, reduce: ContextVariable
			reduce(243), // //, reduce: ContextVariable
			reduce(243), // %, reduce: ContextVariable
			reduce(243), // .*, reduce: ContextVariable
			reduce(243), // ./, reduce: ContextVariable
			reduce(243), // .//, reduce: ContextVariable
			reduce(243), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(243), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(243), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(243), // [[, reduce: ContextVariable
			reduce(243), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1838
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(244), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(244), // >, reduce: ContextVariable
			reduce(244), // >>, reduce: ContextVariable
			reduce(244), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(244), // ||=, reduce: ContextVariable
			reduce(244), // ^^=, reduce: ContextVariable
			reduce(244), // &&=, reduce: ContextVariable
			reduce(244), // ??=, reduce: ContextVariable
			reduce(244), // ???=, reduce: ContextVariable
			reduce(244), // |=, reduce: ContextVariable
			reduce(244), // &=, reduce: ContextVariable
			reduce(244), // ^=, reduce: ContextVariable
			reduce(244), // <<=, reduce: ContextVariable
			reduce(244), // >>=, reduce: ContextVariable
			reduce(244), // >>>=, reduce: ContextVariable
			reduce(244), // +=, reduce: ContextVariable
			reduce(244), // .=, reduce: ContextVariable
			reduce(244), // -=, reduce: ContextVariable
			reduce(244), // *=, reduce: ContextVariable
			reduce(244), // /=, reduce: ContextVariable
			reduce(244), // //=, reduce: ContextVariable
			reduce(244), // %=, reduce: ContextVariable
			reduce(244), // **=, reduce: ContextVariable
			reduce(244), // ?, reduce: ContextVariable
			nil,         // :
			reduce(244), // ||, reduce: ContextVariable
			reduce(244), // ^^, reduce: ContextVariable
			reduce(244), // &&, reduce: ContextVariable
			reduce(244), // ??, reduce: ContextVariable
			reduce(244), // ???, reduce: ContextVariable
			reduce(244), // =~, reduce: ContextVariable
			reduce(244), // !=~, reduce: ContextVariable
			reduce(244), // ==, reduce: ContextVariable
			reduce(244), // !=, reduce: ContextVariable
			reduce(244), // <=>, reduce: ContextVariable
			reduce(244), // >=, reduce: ContextVariable
			reduce(244), // <, reduce: ContextVariable
			reduce(244), // <=, reduce: ContextVariable
			reduce(244), // ^, reduce: ContextVariable
			reduce(244), // &, reduce: ContextVariable
			reduce(244), // <<, reduce: ContextVariable
			reduce(244), // >>>, reduce: ContextVariable
			reduce(244), // +, reduce: ContextVariable
			reduce(244), // -, reduce: ContextVariable
			reduce(244), // .+, reduce: ContextVariable
			reduce(244), // .-, reduce: ContextVariable
			reduce(244), // *, reduce: ContextVariable
			reduce(244), // /, reduce: ContextVariable
			reduce(244), // //, reduce: ContextVariable
			reduce(244), // %, reduce: ContextVariable
			reduce(244), // .*, reduce: ContextVariable
			reduce(244), // ./, reduce: ContextVariable
			reduce(244), // .//, reduce: ContextVariable
			reduce(244), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(244), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(244), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(244), // [[, reduce: ContextVariable
			reduce(244), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1839
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(245), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(245), // >, reduce: ContextVariable
			reduce(245), // >>, reduce: ContextVariable
			reduce(245), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(245), // ||=, reduce: ContextVariable
			reduce(245), // ^^=, reduce: ContextVariable
			reduce(245), // &&=, reduce: ContextVariable
			reduce(245), // ??=, reduce: ContextVariable
			reduce(245), // ???=, reduce: ContextVariable
			reduce(245), // |=, reduce: ContextVariable
			reduce(245), // &=, reduce: ContextVariable
			reduce(245), // ^=, reduce: ContextVariable
			reduce(245), // <<=, reduce: ContextVariable
			reduce(245), // >>=, reduce: ContextVariable
			reduce(245), // >>>=, reduce: ContextVariable
			reduce(245), // +=, reduce: ContextVariable
			reduce(245), // .=, reduce: ContextVariable
			reduce(245), // -=, reduce: ContextVariable
			reduce(245), // *=, reduce: ContextVariable
			reduce(245), // /=, reduce: ContextVariable
			reduce(245), // //=, reduce: ContextVariable
			reduce(245), // %=, reduce: ContextVariable
			reduce(245), // **=, reduce: ContextVariable
			reduce(245), // ?, reduce: ContextVariable
			nil,         // :
			reduce(245), // ||, reduce: ContextVariable
			reduce(245), // ^^, reduce: ContextVariable
			reduce(245), // &&, reduce: ContextVariable
			reduce(245), // ??, reduce: ContextVariable
			reduce(245), // ???, reduce: ContextVariable
			reduce(245), // =~, reduce: ContextVariable
			reduce(245), // !=~, reduce: ContextVariable
			reduce(245), // ==, reduce: ContextVariable
			reduce(245), // !=, reduce: ContextVariable
			reduce(245), // <=>, reduce: ContextVariable
			reduce(245), // >=, reduce: ContextVariable
			reduce(245), // <, reduce: ContextVariable
			reduce(245), // <=, reduce: ContextVariable
			reduce(245), // ^, reduce: ContextVariable
			reduce(245), // &, reduce: ContextVariable
			reduce(245), // <<, reduce: ContextVariable
			reduce(245), // >>>, reduce: ContextVariable
			reduce(245), // +, reduce: ContextVariable
			reduce(245), // -, reduce: ContextVariable
			reduce(245), // .+, reduce: ContextVariable
			reduce(245), // .-, reduce: ContextVariable
			reduce(245), // *, reduce: ContextVariable
			reduce(245), // /, reduce: ContextVariable
			reduce(245), // //, reduce: ContextVariable
			reduce(245), // %, reduce: ContextVariable
			reduce(245), // .*, reduce: ContextVariable
			reduce(245), // ./, reduce: ContextVariable
			reduce(245), // .//, reduce: ContextVariable
			reduce(245), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(245), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(245), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(245), // [[, reduce: ContextVariable
			reduce(245), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1840
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(246), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(246), // >, reduce: ContextVariable
			reduce(246), // >>, reduce: ContextVariable
			reduce(246), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(246), // ||=, reduce: ContextVariable
			reduce(246), // ^^=, reduce: ContextVariable
			reduce(246), // &&=, reduce: ContextVariable
			reduce(246), // ??=, reduce: ContextVariable
			reduce(246), // ???=, reduce: ContextVariable
			reduce(246), // |=, reduce: ContextVariable
			reduce(246), // &=, reduce: ContextVariable
			reduce(246), // ^=, reduce: ContextVariable
			reduce(246), // <<=, reduce: ContextVariable
			reduce(246), // >>=, reduce: ContextVariable
			reduce(246), // >>>=, reduce: ContextVariable
			reduce(246), // +=, reduce: ContextVariable
			reduce(246), // .=, reduce: ContextVariable
			reduce(246), // -=, reduce: ContextVariable
			reduce(246), // *=, reduce: ContextVariable
			reduce(246), // /=, reduce: ContextVariable
			reduce(246), // //=, reduce: ContextVariable
			reduce(246), // %=, reduce: ContextVariable
			reduce(246), // **=, reduce: ContextVariable
			reduce(246), // ?, reduce: ContextVariable
			nil,         // :
			reduce(246), // ||, reduce: ContextVariable
			reduce(246), // ^^, reduce: ContextVariable
			reduce(246), // &&, reduce: ContextVariable
			reduce(246), // ??, reduce: ContextVariable
			reduce(246), // ???, reduce: ContextVariable
			reduce(246), // =~, reduce: ContextVariable
			reduce(246), // !=~, reduce: ContextVariable
			reduce(246), // ==, reduce: ContextVariable
			reduce(246), // !=, reduce: ContextVariable
			reduce(246), // <=>, reduce: ContextVariable
			reduce(246), // >=, reduce: ContextVariable
			reduce(246), // <, reduce: ContextVariable
			reduce(246), // <=, reduce: ContextVariable
			reduce(246), // ^, reduce: ContextVariable
			reduce(246), // &, reduce: ContextVariable
			reduce(246), // <<, reduce: ContextVariable
			reduce(246), // >>>, reduce: ContextVariable
			reduce(246), // +, reduce: ContextVariable
			reduce(246), // -, reduce: ContextVariable
			reduce(246), // .+, reduce: ContextVariable
			reduce(246), // .-, reduce: ContextVariable
			reduce(246), // *, reduce: ContextVariable
			reduce(246), // /, reduce: ContextVariable
			reduce(246), // //, reduce: ContextVariable
			reduce(246), // %, reduce: ContextVariable
			reduce(246), // .*, reduce: ContextVariable
			reduce(246), // ./, reduce: ContextVariable
			reduce(246), // .//, reduce: ContextVariable
			reduce(246), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(246), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(246), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(246), // [[, reduce: ContextVariable
			reduce(246), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1841
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(247), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(247), // >, reduce: ContextVariable
			reduce(247), // >>, reduce: ContextVariable
			reduce(247), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(247), // ||=, reduce: ContextVariable
			reduce(247), // ^^=, reduce: ContextVariable
			reduce(247), // &&=, reduce: ContextVariable
			reduce(247), // ??=, reduce: ContextVariable
			reduce(247), // ???=, reduce: ContextVariable
			reduce(247), // |=, reduce: ContextVariable
			reduce(247), // &=, reduce: ContextVariable
			reduce(247), // ^=, reduce: ContextVariable
			reduce(247), // <<=, reduce: ContextVariable
			reduce(247), // >>=, reduce: ContextVariable
			reduce(247), // >>>=, reduce: ContextVariable
			reduce(247), // +=, reduce: ContextVariable
			reduce(247), // .=, reduce: ContextVariable
			reduce(247), // -=, reduce: ContextVariable
			reduce(247), // *=, reduce: ContextVariable
			reduce(247), // /=, reduce: ContextVariable
			reduce(247), // //=, reduce: ContextVariable
			reduce(247), // %=, reduce: ContextVariable
			reduce(247), // **=, reduce: ContextVariable
			reduce(247), // ?, reduce: ContextVariable
			nil,         // :
			reduce(247), // ||, reduce: ContextVariable
			reduce(247), // ^^, reduce: ContextVariable
			reduce(247), // &&, reduce: ContextVariable
			reduce(247), // ??, reduce: ContextVariable
			reduce(247), // ???, reduce: ContextVariable
			reduce(247), // =~, reduce: ContextVariable
			reduce(247), // !=~, reduce: ContextVariable
			reduce(247), // ==, reduce: ContextVariable
			reduce(247), // !=, reduce: ContextVariable
			reduce(247), // <=>, reduce: ContextVariable
			reduce(247), // >=, reduce: ContextVariable
			reduce(247), // <, reduce: ContextVariable
			reduce(247), // <=, reduce: ContextVariable
			reduce(247), // ^, reduce: ContextVariable
			reduce(247), // &, reduce: ContextVariable
			reduce(247), // <<, reduce: ContextVariable
			reduce(247), // >>>, reduce: ContextVariable
			reduce(247), // +, reduce: ContextVariable
			reduce(247), // -, reduce: ContextVariable
			reduce(247), // .+, reduce: ContextVariable
			reduce(247), // .-, reduce: ContextVariable
			reduce(247), // *, reduce: ContextVariable
			reduce(247), // /, reduce: ContextVariable
			reduce(247), // //, reduce: ContextVariable
			reduce(247), // %, reduce: ContextVariable
			reduce(247), // .*, reduce: ContextVariable
			reduce(247), // ./, reduce: ContextVariable
			reduce(247), // .//, reduce: ContextVariable
			reduce(247), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(247), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(247), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(247), // [[, reduce: ContextVariable
			reduce(247), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1842
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(248), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(248), // >, reduce: ContextVariable
			reduce(248), // >>, reduce: ContextVariable
			reduce(248), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(248), // ||=, reduce: ContextVariable
			reduce(248), // ^^=, reduce: ContextVariable
			reduce(248), // &&=, reduce: ContextVariable
			reduce(248), // ??=, reduce: ContextVariable
			reduce(248), // ???=, reduce: ContextVariable
			reduce(248), // |=, reduce: ContextVariable
			reduce(248), // &=, reduce: ContextVariable
			reduce(248), // ^=, reduce: ContextVariable
			reduce(248), // <<=, reduce: ContextVariable
			reduce(248), // >>=, reduce: ContextVariable
			reduce(248), // >>>=, reduce: ContextVariable
			reduce(248), // +=, reduce: ContextVariable
			reduce(248), // .=, reduce: ContextVariable
			reduce(248), // -=, reduce: ContextVariable
			reduce(248), // *=, reduce: ContextVariable
			reduce(248), // /=, reduce: ContextVariable
			reduce(248), // //=, reduce: ContextVariable
			reduce(248), // %=, reduce: ContextVariable
			reduce(248), // **=, reduce: ContextVariable
			reduce(248), // ?, reduce: ContextVariable
			nil,         // :
			reduce(248), // ||, reduce: ContextVariable
			reduce(248), // ^^, reduce: ContextVariable
			reduce(248), // &&, reduce: ContextVariable
			reduce(248), // ??, reduce: ContextVariable
			reduce(248), // ???, reduce: ContextVariable
			reduce(248), // =~, reduce: ContextVariable
			reduce(248), // !=~, reduce: ContextVariable
			reduce(248), // ==, reduce: ContextVariable
			reduce(248), // !=, reduce: ContextVariable
			reduce(248), // <=>, reduce: ContextVariable
			reduce(248), // >=, reduce: ContextVariable
			reduce(248), // <, reduce: ContextVariable
			reduce(248), // <=, reduce: ContextVariable
			reduce(248), // ^, reduce: ContextVariable
			reduce(248), // &, reduce: ContextVariable
			reduce(248), // <<, reduce: ContextVariable
			reduce(248), // >>>, reduce: ContextVariable
			reduce(248), // +, reduce: ContextVariable
			reduce(248), // -, reduce: ContextVariable
			reduce(248), // .+, reduce: ContextVariable
			reduce(248), // .-, reduce: ContextVariable
			reduce(248), // *, reduce: ContextVariable
			reduce(248), // /, reduce: ContextVariable
			reduce(248), // //, reduce: ContextVariable
			reduce(248), // %, reduce: ContextVariable
			reduce(248), // .*, reduce: ContextVariable
			reduce(248), // ./, reduce: ContextVariable
			reduce(248), // .//, reduce: ContextVariable
			reduce(248), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(248), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(248), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(248), // [[, reduce: ContextVariable
			reduce(248), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1843
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(249), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(249), // >, reduce: ContextVariable
			reduce(249), // >>, reduce: ContextVariable
			reduce(249), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(249), // ||=, reduce: ContextVariable
			reduce(249), // ^^=, reduce: ContextVariable
			reduce(249), // &&=, reduce: ContextVariable
			reduce(249), // ??=, reduce: ContextVariable
			reduce(249), // ???=, reduce: ContextVariable
			reduce(249), // |=, reduce: ContextVariable
			reduce(249), // &=, reduce: ContextVariable
			reduce(249), // ^=, reduce: ContextVariable
			reduce(249), // <<=, reduce: ContextVariable
			reduce(249), // >>=, reduce: ContextVariable
			reduce(249), // >>>=, reduce: ContextVariable
			reduce(249), // +=, reduce: ContextVariable
			reduce(249), // .=, reduce: ContextVariable
			reduce(249), // -=, reduce: ContextVariable
			reduce(249), // *=, reduce: ContextVariable
			reduce(249), // /=, reduce: ContextVariable
			reduce(249), // //=, reduce: ContextVariable
			reduce(249), // %=, reduce: ContextVariable
			reduce(249), // **=, reduce: ContextVariable
			reduce(249), // ?, reduce: ContextVariable
			nil,         // :
			reduce(249), // ||, reduce: ContextVariable
			reduce(249), // ^^, reduce: ContextVariable
			reduce(249), // &&, reduce: ContextVariable
			reduce(249), // ??, reduce: ContextVariable
			reduce(249), // ???, reduce: ContextVariable
			reduce(249), // =~, reduce: ContextVariable
			reduce(249), // !=~, reduce: ContextVariable
			reduce(249), // ==, reduce: ContextVariable
			reduce(249), // !=, reduce: ContextVariable
			reduce(249), // <=>, reduce: ContextVariable
			reduce(249), // >=, reduce: ContextVariable
			reduce(249), // <, reduce: ContextVariable
			reduce(249), // <=, reduce: ContextVariable
			reduce(249), // ^, reduce: ContextVariable
			reduce(249), // &, reduce: ContextVariable
			reduce(249), // <<, reduce: ContextVariable
			reduce(249), // >>>, reduce: ContextVariable
			reduce(249), // +, reduce: ContextVariable
			reduce(249), // -, reduce: ContextVariable
			reduce(249), // .+, reduce: ContextVariable
			reduce(249), // .-, reduce: ContextVariable
			reduce(249), // *, reduce: ContextVariable
			reduce(249), // /, reduce: ContextVariable
			reduce(249), // //, reduce: ContextVariable
			reduce(249), // %, reduce: ContextVariable
			reduce(249), // .*, reduce: ContextVariable
			reduce(249), // ./, reduce: ContextVariable
			reduce(249), // .//, reduce: ContextVariable
			reduce(249), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(249), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(249), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(249), // [[, reduce: ContextVariable
			reduce(249), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1844
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(250), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(250), // >, reduce: ContextVariable
			reduce(250), // >>, reduce: ContextVariable
			reduce(250), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(250), // ||=, reduce: ContextVariable
			reduce(250), // ^^=, reduce: ContextVariable
			reduce(250), // &&=, reduce: ContextVariable
			reduce(250), // ??=, reduce: ContextVariable
			reduce(250), // ???=, reduce: ContextVariable
			reduce(250), // |=, reduce: ContextVariable
			reduce(250), // &=, reduce: ContextVariable
			reduce(250), // ^=, reduce: ContextVariable
			reduce(250), // <<=, reduce: ContextVariable
			reduce(250), // >>=, reduce: ContextVariable
			reduce(250), // >>>=, reduce: ContextVariable
			reduce(250), // +=, reduce: ContextVariable
			reduce(250), // .=, reduce: ContextVariable
			reduce(250), // -=, reduce: ContextVariable
			reduce(250), // *=, reduce: ContextVariable
			reduce(250), // /=, reduce: ContextVariable
			reduce(250), // //=, reduce: ContextVariable
			reduce(250), // %=, reduce: ContextVariable
			reduce(250), // **=, reduce: ContextVariable
			reduce(250), // ?, reduce: ContextVariable
			nil,         // :
			reduce(250), // ||, reduce: ContextVariable
			reduce(250), // ^^, reduce: ContextVariable
			reduce(250), // &&, reduce: ContextVariable
			reduce(250), // ??, reduce: ContextVariable
			reduce(250), // ???, reduce: ContextVariable
			reduce(250), // =~, reduce: ContextVariable
			reduce(250), // !=~, reduce: ContextVariable
			reduce(250), // ==, reduce: ContextVariable
			reduce(250), // !=, reduce: ContextVariable
			reduce(250), // <=>, reduce: ContextVariable
			reduce(250), // >=, reduce: ContextVariable
			reduce(250), // <, reduce: ContextVariable
			reduce(250), // <=, reduce: ContextVariable
			reduce(250), // ^, reduce: ContextVariable
			reduce(250), // &, reduce: ContextVariable
			reduce(250), // <<, reduce: ContextVariable
			reduce(250), // >>>, reduce: ContextVariable
			reduce(250), // +, reduce: ContextVariable
			reduce(250), // -, reduce: ContextVariable
			reduce(250), // .+, reduce: ContextVariable
			reduce(250), // .-, reduce: ContextVariable
			reduce(250), // *, reduce: ContextVariable
			reduce(250), // /, reduce: ContextVariable
			reduce(250), // //, reduce: ContextVariable
			reduce(250), // %, reduce: ContextVariable
			reduce(250), // .*, reduce: ContextVariable
			reduce(250), // ./, reduce: ContextVariable
			reduce(250), // .//, reduce: ContextVariable
			reduce(250), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(250), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(250), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(250), // [[, reduce: ContextVariable
			reduce(250), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1845
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(251), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(251), // >, reduce: ContextVariable
			reduce(251), // >>, reduce: ContextVariable
			reduce(251), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(251), // ||=, reduce: ContextVariable
			reduce(251), // ^^=, reduce: ContextVariable
			reduce(251), // &&=, reduce: ContextVariable
			reduce(251), // ??=, reduce: ContextVariable
			reduce(251), // ???=, reduce: ContextVariable
			reduce(251), // |=, reduce: ContextVariable
			reduce(251), // &=, reduce: ContextVariable
			reduce(251), // ^=, reduce: ContextVariable
			reduce(251), // <<=, reduce: ContextVariable
			reduce(251), // >>=, reduce: ContextVariable
			reduce(251), // >>>=, reduce: ContextVariable
			reduce(251), // +=, reduce: ContextVariable
			reduce(251), // .=, reduce: ContextVariable
			reduce(251), // -=, reduce: ContextVariable
			reduce(251), // *=, reduce: ContextVariable
			reduce(251), // /=, reduce: ContextVariable
			reduce(251), // //=, reduce: ContextVariable
			reduce(251), // %=, reduce: ContextVariable
			reduce(251), // **=, reduce: ContextVariable
			reduce(251), // ?, reduce: ContextVariable
			nil,         // :
			reduce(251), // ||, reduce: ContextVariable
			reduce(251), // ^^, reduce: ContextVariable
			reduce(251), // &&, reduce: ContextVariable
			reduce(251), // ??, reduce: ContextVariable
			reduce(251), // ???, reduce: ContextVariable
			reduce(251), // =~, reduce: ContextVariable
			reduce(251), // !=~, reduce: ContextVariable
			reduce(251), // ==, reduce: ContextVariable
			reduce(251), // !=, reduce: ContextVariable
			reduce(251), // <=>, reduce: ContextVariable
			reduce(251), // >=, reduce: ContextVariable
			reduce(251), // <, reduce: ContextVariable
			reduce(251), // <=, reduce: ContextVariable
			reduce(251), // ^, reduce: ContextVariable
			reduce(251), // &, reduce: ContextVariable
			reduce(251), // <<, reduce: ContextVariable
			reduce(251), // >>>, reduce: ContextVariable
			reduce(251), // +, reduce: ContextVariable
			reduce(251), // -, reduce: ContextVariable
			reduce(251), // .+, reduce: ContextVariable
			reduce(251), // .-, reduce: ContextVariable
			reduce(251), // *, reduce: ContextVariable
			reduce(251), // /, reduce: ContextVariable
			reduce(251), // //, reduce: ContextVariable
			reduce(251), // %, reduce: ContextVariable
			reduce(251), // .*, reduce: ContextVariable
			reduce(251), // ./, reduce: ContextVariable
			reduce(251), // .//, reduce: ContextVariable
			reduce(251), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(251), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(251), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(251), // [[, reduce: ContextVariable
			reduce(251), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1846
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(252), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(252), // >, reduce: ContextVariable
			reduce(252), // >>, reduce: ContextVariable
			reduce(252), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(252), // ||=, reduce: ContextVariable
			reduce(252), // ^^=, reduce: ContextVariable
			reduce(252), // &&=, reduce: ContextVariable
			reduce(252), // ??=, reduce: ContextVariable
			reduce(252), // ???=, reduce: ContextVariable
			reduce(252), // |=, reduce: ContextVariable
			reduce(252), // &=, reduce: ContextVariable
			reduce(252), // ^=, reduce: ContextVariable
			reduce(252), // <<=, reduce: ContextVariable
			reduce(252), // >>=, reduce: ContextVariable
			reduce(252), // >>>=, reduce: ContextVariable
			reduce(252), // +=, reduce: ContextVariable
			reduce(252), // .=, reduce: ContextVariable
			reduce(252), // -=, reduce: ContextVariable
			reduce(252), // *=, reduce: ContextVariable
			reduce(252), // /=, reduce: ContextVariable
			reduce(252), // //=, reduce: ContextVariable
			reduce(252), // %=, reduce: ContextVariable
			reduce(252), // **=, reduce: ContextVariable
			reduce(252), // ?, reduce: ContextVariable
			nil,         // :
			reduce(252), // ||, reduce: ContextVariable
			reduce(252), // ^^, reduce: ContextVariable
			reduce(252), // &&, reduce: ContextVariable
			reduce(252), // ??, reduce: ContextVariable
			reduce(252), // ???, reduce: ContextVariable
			reduce(252), // =~, reduce: ContextVariable
			reduce(252), // !=~, reduce: ContextVariable
			reduce(252), // ==, reduce: ContextVariable
			reduce(252), // !=, reduce: ContextVariable
			reduce(252), // <=>, reduce: ContextVariable
			reduce(252), // >=, reduce: ContextVariable
			reduce(252), // <, reduce: ContextVariable
			reduce(252), // <=, reduce: ContextVariable
			reduce(252), // ^, reduce: ContextVariable
			reduce(252), // &, reduce: ContextVariable
			reduce(252), // <<, reduce: ContextVariable
			reduce(252), // >>>, reduce: ContextVariable
			reduce(252), // +, reduce: ContextVariable
			reduce(252), // -, reduce: ContextVariable
			reduce(252), // .+, reduce: ContextVariable
			reduce(252), // .-, reduce: ContextVariable
			reduce(252), // *, reduce: ContextVariable
			reduce(252), // /, reduce: ContextVariable
			reduce(252), // //, reduce: ContextVariable
			reduce(252), // %, reduce: ContextVariable
			reduce(252), // .*, reduce: ContextVariable
			reduce(252), // ./, reduce: ContextVariable
			reduce(252), // .//, reduce: ContextVariable
			reduce(252), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(252), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(252), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(252), // [[, reduce: ContextVariable
			reduce(252), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1847
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(253), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(253), // >, reduce: MlrvalOrFunction
			reduce(253), // >>, reduce: MlrvalOrFunction
			reduce(253), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(253), // ||=, reduce: MlrvalOrFunction
			reduce(253), // ^^=, reduce: MlrvalOrFunction
			reduce(253), // &&=, reduce: MlrvalOrFunction
			reduce(253), // ??=, reduce: MlrvalOrFunction
			reduce(253), // ???=, reduce: MlrvalOrFunction
			reduce(253), // |=, reduce: MlrvalOrFunction
			reduce(253), // &=, reduce: MlrvalOrFunction
			reduce(253), // ^=, reduce: MlrvalOrFunction
			reduce(253), // <<=, reduce: MlrvalOrFunction
			reduce(253), // >>=, reduce: MlrvalOrFunction
			reduce(253), // >>>=, reduce: MlrvalOrFunction
			reduce(253), // +=, reduce: MlrvalOrFunction
			reduce(253), // .=, reduce: MlrvalOrFunction
			reduce(253), // -=, reduce: MlrvalOrFunction
			reduce(253), // *=, reduce: MlrvalOrFunction
			reduce(253), // /=, reduce: MlrvalOrFunction
			reduce(253), // //=, reduce: MlrvalOrFunction
			reduce(253), // %=, reduce: MlrvalOrFunction
			reduce(253), // **=, reduce: MlrvalOrFunction
			reduce(253), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(253), // ||, reduce: MlrvalOrFunction
			reduce(253), // ^^, reduce: MlrvalOrFunction
			reduce(253), // &&, reduce: MlrvalOrFunction
			reduce(253), // ??, reduce: MlrvalOrFunction
			reduce(253), // ???, reduce: MlrvalOrFunction
			reduce(253), // =~, reduce: MlrvalOrFunction
			reduce(253), // !=~, reduce: MlrvalOrFunction
			reduce(253), // ==, reduce: MlrvalOrFunction
			reduce(253), // !=, reduce: MlrvalOrFunction
			reduce(253), // <=>, reduce: MlrvalOrFunction
			reduce(253), // >=, reduce: MlrvalOrFunction
			reduce(253), // <, reduce: MlrvalOrFunction
			reduce(253), // <=, reduce: MlrvalOrFunction
			reduce(253), // ^, reduce: MlrvalOrFunction
			reduce(253), // &, reduce: MlrvalOrFunction
			reduce(253), // <<, reduce: MlrvalOrFunction
			reduce(253), // >>>, reduce: MlrvalOrFunction
			reduce(253), // +, reduce: MlrvalOrFunction
			reduce(253), // -, reduce: MlrvalOrFunction
			reduce(253), // .+, reduce: MlrvalOrFunction
			reduce(253), // .-, reduce: MlrvalOrFunction
			reduce(253), // *, reduce: MlrvalOrFunction
			reduce(253), // /, reduce: MlrvalOrFunction
			reduce(253), // //, reduce: MlrvalOrFunction
			reduce(253), // %, reduce: MlrvalOrFunction
			reduce(253), // .*, reduce: MlrvalOrFunction
			reduce(253), // ./, reduce: MlrvalOrFunction
			reduce(253), // .//, reduce: MlrvalOrFunction
			reduce(253), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(253), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(253), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(253), // [[, reduce: MlrvalOrFunction
			reduce(253), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1848
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			shift(2894), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(2895), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1849
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(256), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(256), // >, reduce: MlrvalOrFunction
			reduce(256), // >>, reduce: MlrvalOrFunction
			reduce(256), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(256), // ||=, reduce: MlrvalOrFunction
			reduce(256), // ^^=, reduce: MlrvalOrFunction
			reduce(256), // &&=, reduce: MlrvalOrFunction
			reduce(256), // ??=, reduce: MlrvalOrFunction
			reduce(256), // ???=, reduce: MlrvalOrFunction
			reduce(256), // |=, reduce: MlrvalOrFunction
			reduce(256), // &=, reduce: MlrvalOrFunction
			reduce(256), // ^=, reduce: MlrvalOrFunction
			reduce(256), // <<=, reduce: MlrvalOrFunction
			reduce(256), // >>=, reduce: MlrvalOrFunction
			reduce(256), // >>>=, reduce: MlrvalOrFunction
			reduce(256), // +=, reduce: MlrvalOrFunction
			reduce(256), // .=, reduce: MlrvalOrFunction
			reduce(256), // -=, reduce: MlrvalOrFunction
			reduce(256), // *=, reduce: MlrvalOrFunction
			reduce(256), // /=, reduce: MlrvalOrFunction
			reduce(256), // //=, reduce: MlrvalOrFunction
			reduce(256), // %=, reduce: MlrvalOrFunction
			reduce(256), // **=, reduce: MlrvalOrFunction
			reduce(256), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(256), // ||, reduce: MlrvalOrFunction
			reduce(256), // ^^, reduce: MlrvalOrFunction
			reduce(256), // &&, reduce: MlrvalOrFunction
			reduce(256), // ??, reduce: MlrvalOrFunction
			reduce(256), // ???, reduce: MlrvalOrFunction
			reduce(256), // =~, reduce: MlrvalOrFunction
			reduce(256), // !=~, reduce: MlrvalOrFunction
			reduce(256), // ==, reduce: MlrvalOrFunction
			reduce(256), // !=, reduce: MlrvalOrFunction
			reduce(256), // <=>, reduce: MlrvalOrFunction
			reduce(256), // >=, reduce: MlrvalOrFunction
			reduce(256), // <, reduce: MlrvalOrFunction
			reduce(256), // <=, reduce: MlrvalOrFunction
			reduce(256), // ^, reduce: MlrvalOrFunction
			reduce(256), // &, reduce: MlrvalOrFunction
			reduce(256), // <<, reduce: MlrvalOrFunction
			reduce(256), // >>>, reduce: MlrvalOrFunction
			reduce(256), // +, reduce: MlrvalOrFunction
			reduce(256), // -, reduce: MlrvalOrFunction
			reduce(256), // .+, reduce: MlrvalOrFunction
			reduce(256), // .-, reduce: MlrvalOrFunction
			reduce(256), // *, reduce: MlrvalOrFunction
			reduce(256), // /, reduce: MlrvalOrFunction
			reduce(256), // //, reduce: MlrvalOrFunction
			reduce(256), // %, reduce: MlrvalOrFunction
			reduce(256), // .*, reduce: MlrvalOrFunction
			reduce(256), // ./, reduce: MlrvalOrFunction
			reduce(256), // .//, reduce: MlrvalOrFunction
			reduce(256), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(256), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(256), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(256), // [[, reduce: MlrvalOrFunction
			reduce(256), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1850
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(257), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(257), // >, reduce: MlrvalOrFunction
			reduce(257), // >>, reduce: MlrvalOrFunction
			reduce(257), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(257), // ||=, reduce: MlrvalOrFunction
			reduce(257), // ^^=, reduce: MlrvalOrFunction
			reduce(257), // &&=, reduce: MlrvalOrFunction
			reduce(257), // ??=, reduce: MlrvalOrFunction
			reduce(257), // ???=, reduce: MlrvalOrFunction
			reduce(257), // |=, reduce: MlrvalOrFunction
			reduce(257), // &=, reduce: MlrvalOrFunction
			reduce(257), // ^=, reduce: MlrvalOrFunction
			reduce(257), // <<=, reduce: MlrvalOrFunction
			reduce(257), // >>=, reduce: MlrvalOrFunction
			reduce(257), // >>>=, reduce: MlrvalOrFunction
			reduce(257), // +=, reduce: MlrvalOrFunction
			reduce(257), // .=, reduce: MlrvalOrFunction
			reduce(257), // -=, reduce: MlrvalOrFunction
			reduce(257), // *=, reduce: MlrvalOrFunction
			reduce(257), // /=, reduce: MlrvalOrFunction
			reduce(257), // //=, reduce: MlrvalOrFunction
			reduce(257), // %=, reduce: MlrvalOrFunction
			reduce(257), // **=, reduce: MlrvalOrFunction
			reduce(257), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(257), // ||, reduce: MlrvalOrFunction
			reduce(257), // ^^, reduce: MlrvalOrFunction
			reduce(257), // &&, reduce: MlrvalOrFunction
			reduce(257), // ??, reduce: MlrvalOrFunction
			reduce(257), // ???, reduce: MlrvalOrFunction
			reduce(257), // =~, reduce: MlrvalOrFunction
			reduce(257), // !=~, reduce: MlrvalOrFunction
			reduce(257), // ==, reduce: MlrvalOrFunction
			reduce(257), // !=, reduce: MlrvalOrFunction
			reduce(257), // <=>, reduce: MlrvalOrFunction
			reduce(257), // >=, reduce: MlrvalOrFunction
			reduce(257), // <, reduce: MlrvalOrFunction
			reduce(257), // <=, reduce: MlrvalOrFunction
			reduce(257), // ^, reduce: MlrvalOrFunction
			reduce(257), // &, reduce: MlrvalOrFunction
			reduce(257), // <<, reduce: MlrvalOrFunction
			reduce(257), // >>>, reduce: MlrvalOrFunction
			reduce(257), // +, reduce: MlrvalOrFunction
			reduce(257), // -, reduce: MlrvalOrFunction
			reduce(257), // .+, reduce: MlrvalOrFunction
			reduce(257), // .-, reduce: MlrvalOrFunction
			reduce(257), // *, reduce: MlrvalOrFunction
			reduce(257), // /, reduce: MlrvalOrFunction
			reduce(257), // //, reduce: MlrvalOrFunction
			reduce(257), // %, reduce: MlrvalOrFunction
			reduce(257), // .*, reduce: MlrvalOrFunction
			reduce(257), // ./, reduce: MlrvalOrFunction
			reduce(257), // .//, reduce: MlrvalOrFunction
			reduce(257), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(257), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(257), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(257), // [[, reduce: MlrvalOrFunction
			reduce(257), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1851
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(258), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(258), // >, reduce: MlrvalOrFunction
			reduce(258), // >>, reduce: MlrvalOrFunction
			reduce(258), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(258), // ||=, reduce: MlrvalOrFunction
			reduce(258), // ^^=, reduce: MlrvalOrFunction
			reduce(258), // &&=, reduce: MlrvalOrFunction
			reduce(258), // ??=, reduce: MlrvalOrFunction
			reduce(258), // ???=, reduce: MlrvalOrFunction
			reduce(258), // |=, reduce: MlrvalOrFunction
			reduce(258), // &=, reduce: MlrvalOrFunction
			reduce(258), // ^=, reduce: MlrvalOrFunction
			reduce(258), // <<=, reduce: MlrvalOrFunction
			reduce(258), // >>=, reduce: MlrvalOrFunction
			reduce(258), // >>>=, reduce: MlrvalOrFunction
			reduce(258), // +=, reduce: MlrvalOrFunction
			reduce(258), // .=, reduce: MlrvalOrFunction
			reduce(258), // -=, reduce: MlrvalOrFunction
			reduce(258), // *=, reduce: MlrvalOrFunction
			reduce(258), // /=, reduce: MlrvalOrFunction
			reduce(258), // //=, reduce: MlrvalOrFunction
			reduce(258), // %=, reduce: MlrvalOrFunction
			reduce(258), // **=, reduce: MlrvalOrFunction
			reduce(258), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(258), // ||, reduce: MlrvalOrFunction
			reduce(258), // ^^, reduce: MlrvalOrFunction
			reduce(258), // &&, reduce: MlrvalOrFunction
			reduce(258), // ??, reduce: MlrvalOrFunction
			reduce(258), // ???, reduce: MlrvalOrFunction
			reduce(258), // =~, reduce: MlrvalOrFunction
			reduce(258), // !=~, reduce: MlrvalOrFunction
			reduce(258), // ==, reduce: MlrvalOrFunction
			reduce(258), // !=, reduce: MlrvalOrFunction
			reduce(258), // <=>, reduce: MlrvalOrFunction
			reduce(258), // >=, reduce: MlrvalOrFunction
			reduce(258), // <, reduce: MlrvalOrFunction
			reduce(258), // <=, reduce: MlrvalOrFunction
			reduce(258), // ^, reduce: MlrvalOrFunction
			reduce(258), // &, reduce: MlrvalOrFunction
			reduce(258), // <<, reduce: MlrvalOrFunction
			reduce(258), // >>>, reduce: MlrvalOrFunction
			reduce(258), // +, reduce: MlrvalOrFunction
			reduce(258), // -, reduce: MlrvalOrFunction
			reduce(258), // .+, reduce: MlrvalOrFunction
			reduce(258), // .-, reduce: MlrvalOrFunction
			reduce(258), // *, reduce: MlrvalOrFunction
			reduce(258), // /, reduce: MlrvalOrFunction
			reduce(258), // //, reduce: MlrvalOrFunction
			reduce(258), // %, reduce: MlrvalOrFunction
			reduce(258), // .*, reduce: MlrvalOrFunction
			reduce(258), // ./, reduce: MlrvalOrFunction
			reduce(258), // .//, reduce: MlrvalOrFunction
			reduce(258), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(258), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(258), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(258), // [[, reduce: MlrvalOrFunction
			reduce(258), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1852
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(259), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(259), // >, reduce: MlrvalOrFunction
			reduce(259), // >>, reduce: MlrvalOrFunction
			reduce(259), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(259), // ||=, reduce: MlrvalOrFunction
			reduce(259), // ^^=, reduce: MlrvalOrFunction
			reduce(259), // &&=, reduce: MlrvalOrFunction
			reduce(259), // ??=, reduce: MlrvalOrFunction
			reduce(259), // ???=, reduce: MlrvalOrFunction
			reduce(259), // |=, reduce: MlrvalOrFunction
			reduce(259), // &=, reduce: MlrvalOrFunction
			reduce(259), // ^=, reduce: MlrvalOrFunction
			reduce(259), // <<=, reduce: MlrvalOrFunction
			reduce(259), // >>=, reduce: MlrvalOrFunction
			reduce(259), // >>>=, reduce: MlrvalOrFunction
			reduce(259), // +=, reduce: MlrvalOrFunction
			reduce(259), // .=, reduce: MlrvalOrFunction
			reduce(259), // -=, reduce: MlrvalOrFunction
			reduce(259), // *=, reduce: MlrvalOrFunction
			reduce(259), // /=, reduce: MlrvalOrFunction
			reduce(259), // //=, reduce: MlrvalOrFunction
			reduce(259), // %=, reduce: MlrvalOrFunction
			reduce(259), // **=, reduce: MlrvalOrFunction
			reduce(259), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(259), // ||, reduce: MlrvalOrFunction
			reduce(259), // ^^, reduce: MlrvalOrFunction
			reduce(259), // &&, reduce: MlrvalOrFunction
			reduce(259), // ??, reduce: MlrvalOrFunction
			reduce(259), // ???, reduce: MlrvalOrFunction
			reduce(259), // =~, reduce: MlrvalOrFunction
			reduce(259), // !=~, reduce: MlrvalOrFunction
			reduce(259), // ==, reduce: MlrvalOrFunction
			reduce(259), // !=, reduce: MlrvalOrFunction
			reduce(259), // <=>, reduce: MlrvalOrFunction
			reduce(259), // >=, reduce: MlrvalOrFunction
			reduce(259), // <, reduce: MlrvalOrFunction
			reduce(259), // <=, reduce: MlrvalOrFunction
			reduce(259), // ^, reduce: MlrvalOrFunction
			reduce(259), // &, reduce: MlrvalOrFunction
			reduce(259), // <<, reduce: MlrvalOrFunction
			reduce(259), // >>>, reduce: MlrvalOrFunction
			reduce(259), // +, reduce: MlrvalOrFunction
			reduce(259), // -, reduce: MlrvalOrFunction
			reduce(259), // .+, reduce: MlrvalOrFunction
			reduce(259), // .-, reduce: MlrvalOrFunction
			reduce(259), // *, reduce: MlrvalOrFunction
			reduce(259), // /, reduce: MlrvalOrFunction
			reduce(259), // //, reduce: MlrvalOrFunction
			reduce(259), // %, reduce: MlrvalOrFunction
			reduce(259), // .*, reduce: MlrvalOrFunction
			reduce(259), // ./, reduce: MlrvalOrFunction
			reduce(259), // .//, reduce: MlrvalOrFunction
			reduce(259), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(259), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(259), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(259), // [[, reduce: MlrvalOrFunction
			reduce(259), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1853
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(267), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(267), // >, reduce: MlrvalOrFunction
			reduce(267), // >>, reduce: MlrvalOrFunction
			reduce(267), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(267), // ||=, reduce: MlrvalOrFunction
			reduce(267), // ^^=, reduce: MlrvalOrFunction
			reduce(267), // &&=, reduce: MlrvalOrFunction
			reduce(267), // ??=, reduce: MlrvalOrFunction
			reduce(267), // ???=, reduce: MlrvalOrFunction
			reduce(267), // |=, reduce: MlrvalOrFunction
			reduce(267), // &=, reduce: MlrvalOrFunction
			reduce(267), // ^=, reduce: MlrvalOrFunction
			reduce(267), // <<=, reduce: MlrvalOrFunction
			reduce(267), // >>=, reduce: MlrvalOrFunction
			reduce(267), // >>>=, reduce: MlrvalOrFunction
			reduce(267), // +=, reduce: MlrvalOrFunction
			reduce(267), // .=, reduce: MlrvalOrFunction
			reduce(267), // -=, reduce: MlrvalOrFunction
			reduce(267), // *=, reduce: MlrvalOrFunction
			reduce(267), // /=, reduce: MlrvalOrFunction
			reduce(267), // //=, reduce: MlrvalOrFunction
			reduce(267), // %=, reduce: MlrvalOrFunction
			reduce(267), // **=, reduce: MlrvalOrFunction
			reduce(267), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(267), // ||, reduce: MlrvalOrFunction
			reduce(267), // ^^, reduce: MlrvalOrFunction
			reduce(267), // &&, reduce: MlrvalOrFunction
			reduce(267), // ??, reduce: MlrvalOrFunction
			reduce(267), // ???, reduce: MlrvalOrFunction
			reduce(267), // =~, reduce: MlrvalOrFunction
			reduce(267), // !=~, reduce: MlrvalOrFunction
			reduce(267), // ==, reduce: MlrvalOrFunction
			reduce(267), // !=, reduce: MlrvalOrFunction
			reduce(267), // <=>, reduce: MlrvalOrFunction
			reduce(267), // >=, reduce: MlrvalOrFunction
			reduce(267), // <, reduce: MlrvalOrFunction
			reduce(267), // <=, reduce: MlrvalOrFunction
			reduce(267), // ^, reduce: MlrvalOrFunction
			reduce(267), // &, reduce: MlrvalOrFunction
			reduce(267), // <<, reduce: MlrvalOrFunction
			reduce(267), // >>>, reduce: MlrvalOrFunction
			reduce(267), // +, reduce: MlrvalOrFunction
			reduce(267), // -, reduce: MlrvalOrFunction
			reduce(267), // .+, reduce: MlrvalOrFunction
			reduce(267), // .-, reduce: MlrvalOrFunction
			reduce(267), // *, reduce: MlrvalOrFunction
			reduce(267), // /, reduce: MlrvalOrFunction
			reduce(267), // //, reduce: MlrvalOrFunction
			reduce(267), // %, reduce: MlrvalOrFunction
			reduce(267), // .*, reduce: MlrvalOrFunction
			reduce(267), // ./, reduce: MlrvalOrFunction
			reduce(267), // .//, reduce: MlrvalOrFunction
			reduce(267), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(267), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(267), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(267), // [[, reduce: MlrvalOrFunction
			reduce(267), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1854
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2896), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1855
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(2897), // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(2898), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1856
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2899), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1857
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1858
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(334), // ,, reduce: UntypedFuncOrSubrParameterName
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(334), // ), reduce: UntypedFuncOrSubrParameterName
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1859
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			reduce(121), // non_sigil_name, reduce: Typedecl
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1860
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			reduce(122), // non_sigil_name, reduce: Typedecl
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1861
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(2901), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1862
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(328), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1863
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(2902), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(329), // ), reduce: FuncOrSubrNonEmptyParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1864
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(332), // ,, reduce: FuncOrSubrParameter
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(332), // ), reduce: FuncOrSubrParameter
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1865
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(333), // ,, reduce: FuncOrSubrParameter
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(333), // ), reduce: FuncOrSubrParameter
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1866
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1867
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1868
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(12), // $, reduce: NonEmptyStatementBlock
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1869
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(235), // >, reduce: MapLiteral
			reduce(235), // >>, reduce: MapLiteral
			reduce(235), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(235), // ?, reduce: MapLiteral
			reduce(235), // :, reduce: MapLiteral
			reduce(235), // ||, reduce: MapLiteral
			reduce(235), // ^^, reduce: MapLiteral
			reduce(235), // &&, reduce: MapLiteral
			reduce(235), // ??, reduce: MapLiteral
			reduce(235), // ???, reduce: MapLiteral
			reduce(235), // =~, reduce: MapLiteral
			reduce(235), // !=~, reduce: MapLiteral
			reduce(235), // ==, reduce: MapLiteral
			reduce(235), // !=, reduce: MapLiteral
			reduce(235), // <=>, reduce: MapLiteral
			reduce(235), // >=, reduce: MapLiteral
			reduce(235), // <, reduce: MapLiteral
			reduce(235), // <=, reduce: MapLiteral
			reduce(235), // ^, reduce: MapLiteral
			reduce(235), // &, reduce: MapLiteral
			reduce(235), // <<, reduce: MapLiteral
			reduce(235), // >>>, reduce: MapLiteral
			reduce(235), // +, reduce: MapLiteral
			reduce(235), // -, reduce: MapLiteral
			reduce(235), // .+, reduce: MapLiteral
			reduce(235), // .-, reduce: MapLiteral
			reduce(235), // *, reduce: MapLiteral
			reduce(235), // /, reduce: MapLiteral
			reduce(235), // //, reduce: MapLiteral
			reduce(235), // %, reduce: MapLiteral
			reduce(235), // .*, reduce: MapLiteral
			reduce(235), // ./, reduce: MapLiteral
			reduce(235), // .//, reduce: MapLiteral
			reduce(235), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(235), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(235), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(235), // [[, reduce: MapLiteral
			reduce(235), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1870
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			shift(2905), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1871
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(239), // }, reduce: MapLiteralKeyValuePair
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(239), // ,, reduce: MapLiteralKeyValuePair
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1872
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(214), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(214), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(214), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1873
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(211), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(211), // >, reduce: MlrvalOrFunction
			reduce(211), // >>, reduce: MlrvalOrFunction
			reduce(211), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(211), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(211), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(211), // ||, reduce: MlrvalOrFunction
			reduce(211), // ^^, reduce: MlrvalOrFunction
			reduce(211), // &&, reduce: MlrvalOrFunction
			reduce(211), // ??, reduce: MlrvalOrFunction
			reduce(211), // ???, reduce: MlrvalOrFunction
			reduce(211), // =~, reduce: MlrvalOrFunction
			reduce(211), // !=~, reduce: MlrvalOrFunction
			reduce(211), // ==, reduce: MlrvalOrFunction
			reduce(211), // !=, reduce: MlrvalOrFunction
			reduce(211), // <=>, reduce: MlrvalOrFunction
			reduce(211), // >=, reduce: MlrvalOrFunction
			reduce(211), // <, reduce: MlrvalOrFunction
			reduce(211), // <=, reduce: MlrvalOrFunction
			reduce(211), // ^, reduce: MlrvalOrFunction
			reduce(211), // &, reduce: MlrvalOrFunction
			reduce(211), // <<, reduce: MlrvalOrFunction
			reduce(211), // >>>, reduce: MlrvalOrFunction
			reduce(211), // +, reduce: MlrvalOrFunction
			reduce(211), // -, reduce: MlrvalOrFunction
			reduce(211), // .+, reduce: MlrvalOrFunction
			reduce(211), // .-, reduce: MlrvalOrFunction
			reduce(211), // *, reduce: MlrvalOrFunction
			reduce(211), // /, reduce: MlrvalOrFunction
			reduce(211), // //, reduce: MlrvalOrFunction
			reduce(211), // %, reduce: MlrvalOrFunction
			reduce(211), // .*, reduce: MlrvalOrFunction
			reduce(211), // ./, reduce: MlrvalOrFunction
			reduce(211), // .//, reduce: MlrvalOrFunction
			reduce(211), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(211), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(211), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(211), // [[, reduce: MlrvalOrFunction
			reduce(211), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1874
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1875
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(110), // }, reduce: OosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(110), // >, reduce: OosvarValue
			reduce(110), // >>, reduce: OosvarValue
			reduce(110), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(110), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(110), // ?, reduce: OosvarValue
			nil,         // :
			reduce(110), // ||, reduce: OosvarValue
			reduce(110), // ^^, reduce: OosvarValue
			reduce(110), // &&, reduce: OosvarValue
			reduce(110), // ??, reduce: OosvarValue
			reduce(110), // ???, reduce: OosvarValue
			reduce(110), // =~, reduce: OosvarValue
			reduce(110), // !=~, reduce: OosvarValue
			reduce(110), // ==, reduce: OosvarValue
			reduce(110), // !=, reduce: OosvarValue
			reduce(110), // <=>, reduce: OosvarValue
			reduce(110), // >=, reduce: OosvarValue
			reduce(110), // <, reduce: OosvarValue
			reduce(110), // <=, reduce: OosvarValue
			reduce(110), // ^, reduce: OosvarValue
			reduce(110), // &, reduce: OosvarValue
			reduce(110), // <<, reduce: OosvarValue
			reduce(110), // >>>, reduce: OosvarValue
			reduce(110), // +, reduce: OosvarValue
			reduce(110), // -, reduce: OosvarValue
			reduce(110), // .+, reduce: OosvarValue
			reduce(110), // .-, reduce: OosvarValue
			reduce(110), // *, reduce: OosvarValue
			reduce(110), // /, reduce: OosvarValue
			reduce(110), // //, reduce: OosvarValue
			reduce(110), // %, reduce: OosvarValue
			reduce(110), // .*, reduce: OosvarValue
			reduce(110), // ./, reduce: OosvarValue
			reduce(110), // .//, reduce: OosvarValue
			reduce(110), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(110), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(110), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(110), // [[, reduce: OosvarValue
			reduce(110), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1876
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(112), // }, reduce: OosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(112), // >, reduce: OosvarValue
			reduce(112), // >>, reduce: OosvarValue
			reduce(112), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(112), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(112), // ?, reduce: OosvarValue
			nil,         // :
			reduce(112), // ||, reduce: OosvarValue
			reduce(112), // ^^, reduce: OosvarValue
			reduce(112), // &&, reduce: OosvarValue
			reduce(112), // ??, reduce: OosvarValue
			reduce(112), // ???, reduce: OosvarValue
			reduce(112), // =~, reduce: OosvarValue
			reduce(112), // !=~, reduce: OosvarValue
			reduce(112), // ==, reduce: OosvarValue
			reduce(112), // !=, reduce: OosvarValue
			reduce(112), // <=>, reduce: OosvarValue
			reduce(112), // >=, reduce: OosvarValue
			reduce(112), // <, reduce: OosvarValue
			reduce(112), // <=, reduce: OosvarValue
			reduce(112), // ^, reduce: OosvarValue
			reduce(112), // &, reduce: OosvarValue
			reduce(112), // <<, reduce: OosvarValue
			reduce(112), // >>>, reduce: OosvarValue
			reduce(112), // +, reduce: OosvarValue
			reduce(112), // -, reduce: OosvarValue
			reduce(112), // .+, reduce: OosvarValue
			reduce(112), // .-, reduce: OosvarValue
			reduce(112), // *, reduce: OosvarValue
			reduce(112), // /, reduce: OosvarValue
			reduce(112), // //, reduce: OosvarValue
			reduce(112), // %, reduce: OosvarValue
			reduce(112), // .*, reduce: OosvarValue
			reduce(112), // ./, reduce: OosvarValue
			reduce(112), // .//, reduce: OosvarValue
			reduce(112), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(112), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(112), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(112), // [[, reduce: OosvarValue
			reduce(112), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1877
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(111), // }, reduce: OosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(111), // >, reduce: OosvarValue
			reduce(111), // >>, reduce: OosvarValue
			reduce(111), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(111), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(111), // ?, reduce: OosvarValue
			nil,         // :
			reduce(111), // ||, reduce: OosvarValue
			reduce(111), // ^^, reduce: OosvarValue
			reduce(111), // &&, reduce: OosvarValue
			reduce(111), // ??, reduce: OosvarValue
			reduce(111), // ???, reduce: OosvarValue
			reduce(111), // =~, reduce: OosvarValue
			reduce(111), // !=~, reduce: OosvarValue
			reduce(111), // ==, reduce: OosvarValue
			reduce(111), // !=, reduce: OosvarValue
			reduce(111), // <=>, reduce: OosvarValue
			reduce(111), // >=, reduce: OosvarValue
			reduce(111), // <, reduce: OosvarValue
			reduce(111), // <=, reduce: OosvarValue
			reduce(111), // ^, reduce: OosvarValue
			reduce(111), // &, reduce: OosvarValue
			reduce(111), // <<, reduce: OosvarValue
			reduce(111), // >>>, reduce: OosvarValue
			reduce(111), // +, reduce: OosvarValue
			reduce(111), // -, reduce: OosvarValue
			reduce(111), // .+, reduce: OosvarValue
			reduce(111), // .-, reduce: OosvarValue
			reduce(111), // *, reduce: OosvarValue
			reduce(111), // /, reduce: OosvarValue
			reduce(111), // //, reduce: OosvarValue
			reduce(111), // %, reduce: OosvarValue
			reduce(111), // .*, reduce: OosvarValue
			reduce(111), // ./, reduce: OosvarValue
			reduce(111), // .//, reduce: OosvarValue
			reduce(111), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(111), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(111), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(111), // [[, reduce: OosvarValue
			reduce(111), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1878
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			reduce(99), // }, reduce: FieldValue
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(99), // >, reduce: FieldValue
			reduce(99), // >>, reduce: FieldValue
			reduce(99), // |, reduce: FieldValue
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(99), // ,, reduce: FieldValue
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(99), // ?, reduce: FieldValue
			nil,        // :
			reduce(99), // ||, reduce: FieldValue
			reduce(99), // ^^, reduce: FieldValue
			reduce(99), // &&, reduce: FieldValue
			reduce(99), // ??, reduce: FieldValue
			reduce(99), // ???, reduce: FieldValue
			reduce(99), // =~, reduce: FieldValue
			reduce(99), // !=~, reduce: FieldValue
			reduce(99), // ==, reduce: FieldValue
			reduce(99), // !=, reduce: FieldValue
			reduce(99), // <=>, reduce: FieldValue
			reduce(99), // >=, reduce: FieldValue
			reduce(99), // <, reduce: FieldValue
			reduce(99), // <=, reduce: FieldValue
			reduce(99), // ^, reduce: FieldValue
			reduce(99), // &, reduce: FieldValue
			reduce(99), // <<, reduce: FieldValue
			reduce(99), // >>>, reduce: FieldValue
			reduce(99), // +, reduce: FieldValue
			reduce(99), // -, reduce: FieldValue
			reduce(99), // .+, reduce: FieldValue
			reduce(99), // .-, reduce: FieldValue
			reduce(99), // *, reduce: FieldValue
			reduce(99), // /, reduce: FieldValue
			reduce(99), // //, reduce: FieldValue
			reduce(99), // %, reduce: FieldValue
			reduce(99), // .*, reduce: FieldValue
			reduce(99), // ./, reduce: FieldValue
			reduce(99), // .//, reduce: FieldValue
			reduce(99), // ., reduce: FieldValue
			nil,        // !
			nil,        // ~
			reduce(99), // **, reduce: FieldValue
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(99), // [, reduce: FieldValue
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(99), // [[, reduce: FieldValue
			reduce(99), // [[[, reduce: FieldValue
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1879
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(101), // }, reduce: FieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(101), // >, reduce: FieldValue
			reduce(101), // >>, reduce: FieldValue
			reduce(101), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(101), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(101), // ?, reduce: FieldValue
			nil,         // :
			reduce(101), // ||, reduce: FieldValue
			reduce(101), // ^^, reduce: FieldValue
			reduce(101), // &&, reduce: FieldValue
			reduce(101), // ??, reduce: FieldValue
			reduce(101), // ???, reduce: FieldValue
			reduce(101), // =~, reduce: FieldValue
			reduce(101), // !=~, reduce: FieldValue
			reduce(101), // ==, reduce: FieldValue
			reduce(101), // !=, reduce: FieldValue
			reduce(101), // <=>, reduce: FieldValue
			reduce(101), // >=, reduce: FieldValue
			reduce(101), // <, reduce: FieldValue
			reduce(101), // <=, reduce: FieldValue
			reduce(101), // ^, reduce: FieldValue
			reduce(101), // &, reduce: FieldValue
			reduce(101), // <<, reduce: FieldValue
			reduce(101), // >>>, reduce: FieldValue
			reduce(101), // +, reduce: FieldValue
			reduce(101), // -, reduce: FieldValue
			reduce(101), // .+, reduce: FieldValue
			reduce(101), // .-, reduce: FieldValue
			reduce(101), // *, reduce: FieldValue
			reduce(101), // /, reduce: FieldValue
			reduce(101), // //, reduce: FieldValue
			reduce(101), // %, reduce: FieldValue
			reduce(101), // .*, reduce: FieldValue
			reduce(101), // ./, reduce: FieldValue
			reduce(101), // .//, reduce: FieldValue
			reduce(101), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(101), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(101), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(101), // [[, reduce: FieldValue
			reduce(101), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1880
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(100), // }, reduce: FieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(100), // >, reduce: FieldValue
			reduce(100), // >>, reduce: FieldValue
			reduce(100), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(100), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(100), // ?, reduce: FieldValue
			nil,         // :
			reduce(100), // ||, reduce: FieldValue
			reduce(100), // ^^, reduce: FieldValue
			reduce(100), // &&, reduce: FieldValue
			reduce(100), // ??, reduce: FieldValue
			reduce(100), // ???, reduce: FieldValue
			reduce(100), // =~, reduce: FieldValue
			reduce(100), // !=~, reduce: FieldValue
			reduce(100), // ==, reduce: FieldValue
			reduce(100), // !=, reduce: FieldValue
			reduce(100), // <=>, reduce: FieldValue
			reduce(100), // >=, reduce: FieldValue
			reduce(100), // <, reduce: FieldValue
			reduce(100), // <=, reduce: FieldValue
			reduce(100), // ^, reduce: FieldValue
			reduce(100), // &, reduce: FieldValue
			reduce(100), // <<, reduce: FieldValue
			reduce(100), // >>>, reduce: FieldValue
			reduce(100), // +, reduce: FieldValue
			reduce(100), // -, reduce: FieldValue
			reduce(100), // .+, reduce: FieldValue
			reduce(100), // .-, reduce: FieldValue
			reduce(100), // *, reduce: FieldValue
			reduce(100), // /, reduce: FieldValue
			reduce(100), // //, reduce: FieldValue
			reduce(100), // %, reduce: FieldValue
			reduce(100), // .*, reduce: FieldValue
			reduce(100), // ./, reduce: FieldValue
			reduce(100), // .//, reduce: FieldValue
			reduce(100), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(100), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(100), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(100), // [[, reduce: FieldValue
			reduce(100), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1881
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(213), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(213), // >, reduce: MlrvalOrFunction
			reduce(213), // >>, reduce: MlrvalOrFunction
			reduce(213), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(213), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(213), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(213), // ||, reduce: MlrvalOrFunction
			reduce(213), // ^^, reduce: MlrvalOrFunction
			reduce(213), // &&, reduce: MlrvalOrFunction
			reduce(213), // ??, reduce: MlrvalOrFunction
			reduce(213), // ???, reduce: MlrvalOrFunction
			reduce(213), // =~, reduce: MlrvalOrFunction
			reduce(213), // !=~, reduce: MlrvalOrFunction
			reduce(213), // ==, reduce: MlrvalOrFunction
			reduce(213), // !=, reduce: MlrvalOrFunction
			reduce(213), // <=>, reduce: MlrvalOrFunction
			reduce(213), // >=, reduce: MlrvalOrFunction
			reduce(213), // <, reduce: MlrvalOrFunction
			reduce(213), // <=, reduce: MlrvalOrFunction
			reduce(213), // ^, reduce: MlrvalOrFunction
			reduce(213), // &, reduce: MlrvalOrFunction
			reduce(213), // <<, reduce: MlrvalOrFunction
			reduce(213), // >>>, reduce: MlrvalOrFunction
			reduce(213), // +, reduce: MlrvalOrFunction
			reduce(213), // -, reduce: MlrvalOrFunction
			reduce(213), // .+, reduce: MlrvalOrFunction
			reduce(213), // .-, reduce: MlrvalOrFunction
			reduce(213), // *, reduce: MlrvalOrFunction
			reduce(213), // /, reduce: MlrvalOrFunction
			reduce(213), // //, reduce: MlrvalOrFunction
			reduce(213), // %, reduce: MlrvalOrFunction
			reduce(213), // .*, reduce: MlrvalOrFunction
			reduce(213), // ./, reduce: MlrvalOrFunction
			reduce(213), // .//, reduce: MlrvalOrFunction
			reduce(213), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(213), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(213), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(213), // [[, reduce: MlrvalOrFunction
			reduce(213), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1882
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(233), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(233), // >, reduce: MlrvalOrFunction
			reduce(233), // >>, reduce: MlrvalOrFunction
			reduce(233), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(233), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(233), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(233), // ||, reduce: MlrvalOrFunction
			reduce(233), // ^^, reduce: MlrvalOrFunction
			reduce(233), // &&, reduce: MlrvalOrFunction
			reduce(233), // ??, reduce: MlrvalOrFunction
			reduce(233), // ???, reduce: MlrvalOrFunction
			reduce(233), // =~, reduce: MlrvalOrFunction
			reduce(233), // !=~, reduce: MlrvalOrFunction
			reduce(233), // ==, reduce: MlrvalOrFunction
			reduce(233), // !=, reduce: MlrvalOrFunction
			reduce(233), // <=>, reduce: MlrvalOrFunction
			reduce(233), // >=, reduce: MlrvalOrFunction
			reduce(233), // <, reduce: MlrvalOrFunction
			reduce(233), // <=, reduce: MlrvalOrFunction
			reduce(233), // ^, reduce: MlrvalOrFunction
			reduce(233), // &, reduce: MlrvalOrFunction
			reduce(233), // <<, reduce: MlrvalOrFunction
			reduce(233), // >>>, reduce: MlrvalOrFunction
			reduce(233), // +, reduce: MlrvalOrFunction
			reduce(233), // -, reduce: MlrvalOrFunction
			reduce(233), // .+, reduce: MlrvalOrFunction
			reduce(233), // .-, reduce: MlrvalOrFunction
			reduce(233), // *, reduce: MlrvalOrFunction
			reduce(233), // /, reduce: MlrvalOrFunction
			reduce(233), // //, reduce: MlrvalOrFunction
			reduce(233), // %, reduce: MlrvalOrFunction
			reduce(233), // .*, reduce: MlrvalOrFunction
			reduce(233), // ./, reduce: MlrvalOrFunction
			reduce(233), // .//, reduce: MlrvalOrFunction
			reduce(233), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(233), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(233), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(233), // [[, reduce: MlrvalOrFunction
			reduce(233), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1883
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(210), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(210), // >, reduce: MlrvalOrFunction
			reduce(210), // >>, reduce: MlrvalOrFunction
			reduce(210), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(210), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(210), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(210), // ||, reduce: MlrvalOrFunction
			reduce(210), // ^^, reduce: MlrvalOrFunction
			reduce(210), // &&, reduce: MlrvalOrFunction
			reduce(210), // ??, reduce: MlrvalOrFunction
			reduce(210), // ???, reduce: MlrvalOrFunction
			reduce(210), // =~, reduce: MlrvalOrFunction
			reduce(210), // !=~, reduce: MlrvalOrFunction
			reduce(210), // ==, reduce: MlrvalOrFunction
			reduce(210), // !=, reduce: MlrvalOrFunction
			reduce(210), // <=>, reduce: MlrvalOrFunction
			reduce(210), // >=, reduce: MlrvalOrFunction
			reduce(210), // <, reduce: MlrvalOrFunction
			reduce(210), // <=, reduce: MlrvalOrFunction
			reduce(210), // ^, reduce: MlrvalOrFunction
			reduce(210), // &, reduce: MlrvalOrFunction
			reduce(210), // <<, reduce: MlrvalOrFunction
			reduce(210), // >>>, reduce: MlrvalOrFunction
			reduce(210), // +, reduce: MlrvalOrFunction
			reduce(210), // -, reduce: MlrvalOrFunction
			reduce(210), // .+, reduce: MlrvalOrFunction
			reduce(210), // .-, reduce: MlrvalOrFunction
			reduce(210), // *, reduce: MlrvalOrFunction
			reduce(210), // /, reduce: MlrvalOrFunction
			reduce(210), // //, reduce: MlrvalOrFunction
			reduce(210), // %, reduce: MlrvalOrFunction
			reduce(210), // .*, reduce: MlrvalOrFunction
			reduce(210), // ./, reduce: MlrvalOrFunction
			reduce(210), // .//, reduce: MlrvalOrFunction
			reduce(210), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(210), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(210), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(210), // [[, reduce: MlrvalOrFunction
			reduce(210), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1884
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(102), // }, reduce: FieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(102), // >, reduce: FieldValue
			reduce(102), // >>, reduce: FieldValue
			reduce(102), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(102), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(102), // ?, reduce: FieldValue
			nil,         // :
			reduce(102), // ||, reduce: FieldValue
			reduce(102), // ^^, reduce: FieldValue
			reduce(102), // &&, reduce: FieldValue
			reduce(102), // ??, reduce: FieldValue
			reduce(102), // ???, reduce: FieldValue
			reduce(102), // =~, reduce: FieldValue
			reduce(102), // !=~, reduce: FieldValue
			reduce(102), // ==, reduce: FieldValue
			reduce(102), // !=, reduce: FieldValue
			reduce(102), // <=>, reduce: FieldValue
			reduce(102), // >=, reduce: FieldValue
			reduce(102), // <, reduce: FieldValue
			reduce(102), // <=, reduce: FieldValue
			reduce(102), // ^, reduce: FieldValue
			reduce(102), // &, reduce: FieldValue
			reduce(102), // <<, reduce: FieldValue
			reduce(102), // >>>, reduce: FieldValue
			reduce(102), // +, reduce: FieldValue
			reduce(102), // -, reduce: FieldValue
			reduce(102), // .+, reduce: FieldValue
			reduce(102), // .-, reduce: FieldValue
			reduce(102), // *, reduce: FieldValue
			reduce(102), // /, reduce: FieldValue
			reduce(102), // //, reduce: FieldValue
			reduce(102), // %, reduce: FieldValue
			reduce(102), // .*, reduce: FieldValue
			reduce(102), // ./, reduce: FieldValue
			reduce(102), // .//, reduce: FieldValue
			reduce(102), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(102), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(102), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(102), // [[, reduce: FieldValue
			reduce(102), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1885
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(103), // }, reduce: FieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(103), // >, reduce: FieldValue
			reduce(103), // >>, reduce: FieldValue
			reduce(103), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(103), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(103), // ?, reduce: FieldValue
			nil,         // :
			reduce(103), // ||, reduce: FieldValue
			reduce(103), // ^^, reduce: FieldValue
			reduce(103), // &&, reduce: FieldValue
			reduce(103), // ??, reduce: FieldValue
			reduce(103), // ???, reduce: FieldValue
			reduce(103), // =~, reduce: FieldValue
			reduce(103), // !=~, reduce: FieldValue
			reduce(103), // ==, reduce: FieldValue
			reduce(103), // !=, reduce: FieldValue
			reduce(103), // <=>, reduce: FieldValue
			reduce(103), // >=, reduce: FieldValue
			reduce(103), // <, reduce: FieldValue
			reduce(103), // <=, reduce: FieldValue
			reduce(103), // ^, reduce: FieldValue
			reduce(103), // &, reduce: FieldValue
			reduce(103), // <<, reduce: FieldValue
			reduce(103), // >>>, reduce: FieldValue
			reduce(103), // +, reduce: FieldValue
			reduce(103), // -, reduce: FieldValue
			reduce(103), // .+, reduce: FieldValue
			reduce(103), // .-, reduce: FieldValue
			reduce(103), // *, reduce: FieldValue
			reduce(103), // /, reduce: FieldValue
			reduce(103), // //, reduce: FieldValue
			reduce(103), // %, reduce: FieldValue
			reduce(103), // .*, reduce: FieldValue
			reduce(103), // ./, reduce: FieldValue
			reduce(103), // .//, reduce: FieldValue
			reduce(103), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(103), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(103), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(103), // [[, reduce: FieldValue
			reduce(103), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1886
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(104), // }, reduce: DirectFieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(104), // >, reduce: DirectFieldValue
			reduce(104), // >>, reduce: DirectFieldValue
			reduce(104), // |, reduce: DirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(104), // ,, reduce: DirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(104), // ?, reduce: DirectFieldValue
			nil,         // :
			reduce(104), // ||, reduce: DirectFieldValue
			reduce(104), // ^^, reduce: DirectFieldValue
			reduce(104), // &&, reduce: DirectFieldValue
			reduce(104), // ??, reduce: DirectFieldValue
			reduce(104), // ???, reduce: DirectFieldValue
			reduce(104), // =~, reduce: DirectFieldValue
			reduce(104), // !=~, reduce: DirectFieldValue
			reduce(104), // ==, reduce: DirectFieldValue
			reduce(104), // !=, reduce: DirectFieldValue
			reduce(104), // <=>, reduce: DirectFieldValue
			reduce(104), // >=, reduce: DirectFieldValue
			reduce(104), // <, reduce: DirectFieldValue
			reduce(104), // <=, reduce: DirectFieldValue
			reduce(104), // ^, reduce: DirectFieldValue
			reduce(104), // &, reduce: DirectFieldValue
			reduce(104), // <<, reduce: DirectFieldValue
			reduce(104), // >>>, reduce: DirectFieldValue
			reduce(104), // +, reduce: DirectFieldValue
			reduce(104), // -, reduce: DirectFieldValue
			reduce(104), // .+, reduce: DirectFieldValue
			reduce(104), // .-, reduce: DirectFieldValue
			reduce(104), // *, reduce: DirectFieldValue
			reduce(104), // /, reduce: DirectFieldValue
			reduce(104), // //, reduce: DirectFieldValue
			reduce(104), // %, reduce: DirectFieldValue
			reduce(104), // .*, reduce: DirectFieldValue
			reduce(104), // ./, reduce: DirectFieldValue
			reduce(104), // .//, reduce: DirectFieldValue
			reduce(104), // ., reduce: DirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(104), // **, reduce: DirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(104), // [, reduce: DirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(104), // [[, reduce: DirectFieldValue
			reduce(104), // [[[, reduce: DirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1887
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1888
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(106), // }, reduce: BracedFieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(106), // >, reduce: BracedFieldValue
			reduce(106), // >>, reduce: BracedFieldValue
			reduce(106), // |, reduce: BracedFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(106), // ,, reduce: BracedFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(106), // ?, reduce: BracedFieldValue
			nil,         // :
			reduce(106), // ||, reduce: BracedFieldValue
			reduce(106), // ^^, reduce: BracedFieldValue
			reduce(106), // &&, reduce: BracedFieldValue
			reduce(106), // ??, reduce: BracedFieldValue
			reduce(106), // ???, reduce: BracedFieldValue
			reduce(106), // =~, reduce: BracedFieldValue
			reduce(106), // !=~, reduce: BracedFieldValue
			reduce(106), // ==, reduce: BracedFieldValue
			reduce(106), // !=, reduce: BracedFieldValue
			reduce(106), // <=>, reduce: BracedFieldValue
			reduce(106), // >=, reduce: BracedFieldValue
			reduce(106), // <, reduce: BracedFieldValue
			reduce(106), // <=, reduce: BracedFieldValue
			reduce(106), // ^, reduce: BracedFieldValue
			reduce(106), // &, reduce: BracedFieldValue
			reduce(106), // <<, reduce: BracedFieldValue
			reduce(106), // >>>, reduce: BracedFieldValue
			reduce(106), // +, reduce: BracedFieldValue
			reduce(106), // -, reduce: BracedFieldValue
			reduce(106), // .+, reduce: BracedFieldValue
			reduce(106), // .-, reduce: BracedFieldValue
			reduce(106), // *, reduce: BracedFieldValue
			reduce(106), // /, reduce: BracedFieldValue
			reduce(106), // //, reduce: BracedFieldValue
			reduce(106), // %, reduce: BracedFieldValue
			reduce(106), // .*, reduce: BracedFieldValue
			reduce(106), // ./, reduce: BracedFieldValue
			reduce(106), // .//, reduce: BracedFieldValue
			reduce(106), // ., reduce: BracedFieldValue
			nil,         // !
			nil,         // ~
			reduce(106), // **, reduce: BracedFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(106), // [, reduce: BracedFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(106), // [[, reduce: BracedFieldValue
			reduce(106), // [[[, reduce: BracedFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1889
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1890
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1891
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(109), // }, reduce: FullSrec
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(109), // >, reduce: FullSrec
			reduce(109), // >>, reduce: FullSrec
			reduce(109), // |, reduce: FullSrec
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(109), // ,, reduce: FullSrec
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(109), // ?, reduce: FullSrec
			nil,         // :
			reduce(109), // ||, reduce: FullSrec
			reduce(109), // ^^, reduce: FullSrec
			reduce(109), // &&, reduce: FullSrec
			reduce(109), // ??, reduce: FullSrec
			reduce(109), // ???, reduce: FullSrec
			reduce(109), // =~, reduce: FullSrec
			reduce(109), // !=~, reduce: FullSrec
			reduce(109), // ==, reduce: FullSrec
			reduce(109), // !=, reduce: FullSrec
			reduce(109), // <=>, reduce: FullSrec
			reduce(109), // >=, reduce: FullSrec
			reduce(109), // <, reduce: FullSrec
			reduce(109), // <=, reduce: FullSrec
			reduce(109), // ^, reduce: FullSrec
			reduce(109), // &, reduce: FullSrec
			reduce(109), // <<, reduce: FullSrec
			reduce(109), // >>>, reduce: FullSrec
			reduce(109), // +, reduce: FullSrec
			reduce(109), // -, reduce: FullSrec
			reduce(109), // .+, reduce: FullSrec
			reduce(109), // .-, reduce: FullSrec
			reduce(109), // *, reduce: FullSrec
			reduce(109), // /, reduce: FullSrec
			reduce(109), // //, reduce: FullSrec
			reduce(109), // %, reduce: FullSrec
			reduce(109), // .*, reduce: FullSrec
			reduce(109), // ./, reduce: FullSrec
			reduce(109), // .//, reduce: FullSrec
			reduce(109), // ., reduce: FullSrec
			nil,         // !
			nil,         // ~
			reduce(109), // **, reduce: FullSrec
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(109), // [, reduce: FullSrec
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(109), // [[, reduce: FullSrec
			reduce(109), // [[[, reduce: FullSrec
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1892
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(212), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(212), // >, reduce: MlrvalOrFunction
			reduce(212), // >>, reduce: MlrvalOrFunction
			reduce(212), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(212), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(212), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(212), // ||, reduce: MlrvalOrFunction
			reduce(212), // ^^, reduce: MlrvalOrFunction
			reduce(212), // &&, reduce: MlrvalOrFunction
			reduce(212), // ??, reduce: MlrvalOrFunction
			reduce(212), // ???, reduce: MlrvalOrFunction
			reduce(212), // =~, reduce: MlrvalOrFunction
			reduce(212), // !=~, reduce: MlrvalOrFunction
			reduce(212), // ==, reduce: MlrvalOrFunction
			reduce(212), // !=, reduce: MlrvalOrFunction
			reduce(212), // <=>, reduce: MlrvalOrFunction
			reduce(212), // >=, reduce: MlrvalOrFunction
			reduce(212), // <, reduce: MlrvalOrFunction
			reduce(212), // <=, reduce: MlrvalOrFunction
			reduce(212), // ^, reduce: MlrvalOrFunction
			reduce(212), // &, reduce: MlrvalOrFunction
			reduce(212), // <<, reduce: MlrvalOrFunction
			reduce(212), // >>>, reduce: MlrvalOrFunction
			reduce(212), // +, reduce: MlrvalOrFunction
			reduce(212), // -, reduce: MlrvalOrFunction
			reduce(212), // .+, reduce: MlrvalOrFunction
			reduce(212), // .-, reduce: MlrvalOrFunction
			reduce(212), // *, reduce: MlrvalOrFunction
			reduce(212), // /, reduce: MlrvalOrFunction
			reduce(212), // //, reduce: MlrvalOrFunction
			reduce(212), // %, reduce: MlrvalOrFunction
			reduce(212), // .*, reduce: MlrvalOrFunction
			reduce(212), // ./, reduce: MlrvalOrFunction
			reduce(212), // .//, reduce: MlrvalOrFunction
			reduce(212), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(212), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(212), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(212), // [[, reduce: MlrvalOrFunction
			reduce(212), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1893
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(113), // }, reduce: DirectOosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(113), // >, reduce: DirectOosvarValue
			reduce(113), // >>, reduce: DirectOosvarValue
			reduce(113), // |, reduce: DirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(113), // ,, reduce: DirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(113), // ?, reduce: DirectOosvarValue
			nil,         // :
			reduce(113), // ||, reduce: DirectOosvarValue
			reduce(113), // ^^, reduce: DirectOosvarValue
			reduce(113), // &&, reduce: DirectOosvarValue
			reduce(113), // ??, reduce: DirectOosvarValue
			reduce(113), // ???, reduce: DirectOosvarValue
			reduce(113), // =~, reduce: DirectOosvarValue
			reduce(113), // !=~, reduce: DirectOosvarValue
			reduce(113), // ==, reduce: DirectOosvarValue
			reduce(113), // !=, reduce: DirectOosvarValue
			reduce(113), // <=>, reduce: DirectOosvarValue
			reduce(113), // >=, reduce: DirectOosvarValue
			reduce(113), // <, reduce: DirectOosvarValue
			reduce(113), // <=, reduce: DirectOosvarValue
			reduce(113), // ^, reduce: DirectOosvarValue
			reduce(113), // &, reduce: DirectOosvarValue
			reduce(113), // <<, reduce: DirectOosvarValue
			reduce(113), // >>>, reduce: DirectOosvarValue
			reduce(113), // +, reduce: DirectOosvarValue
			reduce(113), // -, reduce: DirectOosvarValue
			reduce(113), // .+, reduce: DirectOosvarValue
			reduce(113), // .-, reduce: DirectOosvarValue
			reduce(113), // *, reduce: DirectOosvarValue
			reduce(113), // /, reduce: DirectOosvarValue
			reduce(113), // //, reduce: DirectOosvarValue
			reduce(113), // %, reduce: DirectOosvarValue
			reduce(113), // .*, reduce: DirectOosvarValue
			reduce(113), // ./, reduce: DirectOosvarValue
			reduce(113), // .//, reduce: DirectOosvarValue
			reduce(113), // ., reduce: DirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(113), // **, reduce: DirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(113), // [, reduce: DirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(113), // [[, reduce: DirectOosvarValue
			reduce(113), // [[[, reduce: DirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1894
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S1895
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(115), // }, reduce: BracedOosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(115), // >, reduce: BracedOosvarValue
			reduce(115), // >>, reduce: BracedOosvarValue
			reduce(115), // |, reduce: BracedOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(115), // ,, reduce: BracedOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(115), // ?, reduce: BracedOosvarValue
			nil,         // :
			reduce(115), // ||, reduce: BracedOosvarValue
			reduce(115), // ^^, reduce: BracedOosvarValue
			reduce(115), // &&, reduce: BracedOosvarValue
			reduce(115), // ??, reduce: BracedOosvarValue
			reduce(115), // ???, reduce: BracedOosvarValue
			reduce(115), // =~, reduce: BracedOosvarValue
			reduce(115), // !=~, reduce: BracedOosvarValue
			reduce(115), // ==, reduce: BracedOosvarValue
			reduce(115), // !=, reduce: BracedOosvarValue
			reduce(115), // <=>, reduce: BracedOosvarValue
			reduce(115), // >=, reduce: BracedOosvarValue
			reduce(115), // <, reduce: BracedOosvarValue
			reduce(115), // <=, reduce: BracedOosvarValue
			reduce(115), // ^, reduce: BracedOosvarValue
			reduce(115), // &, reduce: BracedOosvarValue
			reduce(115), // <<, reduce: BracedOosvarValue
			reduce(115), // >>>, reduce: BracedOosvarValue
			reduce(115), // +, reduce: BracedOosvarValue
			reduce(115), // -, reduce: BracedOosvarValue
			reduce(115), // .+, reduce: BracedOosvarValue
			reduce(115), // .-, reduce: BracedOosvarValue
			reduce(115), // *, reduce: BracedOosvarValue
			reduce(115), // /, reduce: BracedOosvarValue
			reduce(115), // //, reduce: BracedOosvarValue
			reduce(115), // %, reduce: BracedOosvarValue
			reduce(115), // .*, reduce: BracedOosvarValue
			reduce(115), // ./, reduce: BracedOosvarValue
			reduce(115), // .//, reduce: BracedOosvarValue
			reduce(115), // ., reduce: BracedOosvarValue
			nil,         // !
			nil,         // ~
			reduce(115), // **, reduce: BracedOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(115), // [, reduce: BracedOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(115), // [[, reduce: BracedOosvarValue
			reduce(115), // [[[, reduce: BracedOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1896
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(116), // }, reduce: FullOosvar
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(116), // >, reduce: FullOosvar
			reduce(116), // >>, reduce: FullOosvar
			reduce(116), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(116), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(116), // ?, reduce: FullOosvar
			nil,         // :
			reduce(116), // ||, reduce: FullOosvar
			reduce(116), // ^^, reduce: FullOosvar
			reduce(116), // &&, reduce: FullOosvar
			reduce(116), // ??, reduce: FullOosvar
			reduce(116), // ???, reduce: FullOosvar
			reduce(116), // =~, reduce: FullOosvar
			reduce(116), // !=~, reduce: FullOosvar
			reduce(116), // ==, reduce: FullOosvar
			reduce(116), // !=, reduce: FullOosvar
			reduce(116), // <=>, reduce: FullOosvar
			reduce(116), // >=, reduce: FullOosvar
			reduce(116), // <, reduce: FullOosvar
			reduce(116), // <=, reduce: FullOosvar
			reduce(116), // ^, reduce: FullOosvar
			reduce(116), // &, reduce: FullOosvar
			reduce(116), // <<, reduce: FullOosvar
			reduce(116), // >>>, reduce: FullOosvar
			reduce(116), // +, reduce: FullOosvar
			reduce(116), // -, reduce: FullOosvar
			reduce(116), // .+, reduce: FullOosvar
			reduce(116), // .-, reduce: FullOosvar
			reduce(116), // *, reduce: FullOosvar
			reduce(116), // /, reduce: FullOosvar
			reduce(116), // //, reduce: FullOosvar
			reduce(116), // %, reduce: FullOosvar
			reduce(116), // .*, reduce: FullOosvar
			reduce(116), // ./, reduce: FullOosvar
			reduce(116), // .//, reduce: FullOosvar
			reduce(116), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(116), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(116), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(116), // [[, reduce: FullOosvar
			reduce(116), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1897
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(117), // }, reduce: FullOosvar
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(117), // >, reduce: FullOosvar
			reduce(117), // >>, reduce: FullOosvar
			reduce(117), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(117), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(117), // ?, reduce: FullOosvar
			nil,         // :
			reduce(117), // ||, reduce: FullOosvar
			reduce(117), // ^^, reduce: FullOosvar
			reduce(117), // &&, reduce: FullOosvar
			reduce(117), // ??, reduce: FullOosvar
			reduce(117), // ???, reduce: FullOosvar
			reduce(117), // =~, reduce: FullOosvar
			reduce(117), // !=~, reduce: FullOosvar
			reduce(117), // ==, reduce: FullOosvar
			reduce(117), // !=, reduce: FullOosvar
			reduce(117), // <=>, reduce: FullOosvar
			reduce(117), // >=, reduce: FullOosvar
			reduce(117), // <, reduce: FullOosvar
			reduce(117), // <=, reduce: FullOosvar
			reduce(117), // ^, reduce: FullOosvar
			reduce(117), // &, reduce: FullOosvar
			reduce(117), // <<, reduce: FullOosvar
			reduce(117), // >>>, reduce: FullOosvar
			reduce(117), // +, reduce: FullOosvar
			reduce(117), // -, reduce: FullOosvar
			reduce(117), // .+, reduce: FullOosvar
			reduce(117), // .-, reduce: FullOosvar
			reduce(117), // *, reduce: FullOosvar
			reduce(117), // /, reduce: FullOosvar
			reduce(117), // //, reduce: FullOosvar
			reduce(117), // %, reduce: FullOosvar
			reduce(117), // .*, reduce: FullOosvar
			reduce(117), // ./, reduce: FullOosvar
			reduce(117), // .//, reduce: FullOosvar
			reduce(117), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(117), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(117), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(117), // [[, reduce: FullOosvar
			reduce(117), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1898
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(118), // }, reduce: LocalVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(118), // >, reduce: LocalVariable
			reduce(118), // >>, reduce: LocalVariable
			reduce(118), // |, reduce: LocalVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(118), // ,, reduce: LocalVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(270), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(118), // ?, reduce: LocalVariable
			nil,         // :
			reduce(118), // ||, reduce: LocalVariable
			reduce(118), // ^^, reduce: LocalVariable
			reduce(118), // &&, reduce: LocalVariable
			reduce(118), // ??, reduce: LocalVariable
			reduce(118), // ???, reduce: LocalVariable
			reduce(118), // =~, reduce: LocalVariable
			reduce(118), // !=~, reduce: LocalVariable
			reduce(118), // ==, reduce: LocalVariable
			reduce(118), // !=, reduce: LocalVariable
			reduce(118), // <=>, reduce: LocalVariable
			reduce(118), // >=, reduce: LocalVariable
			reduce(118), // <, reduce: LocalVariable
			reduce(118), // <=, reduce: LocalVariable
			reduce(118), // ^, reduce: LocalVariable
			reduce(118), // &, reduce: LocalVariable
			reduce(118), // <<, reduce: LocalVariable
			reduce(118), // >>>, reduce: LocalVariable
			reduce(118), // +, reduce: LocalVariable
			reduce(118), // -, reduce: LocalVariable
			reduce(118), // .+, reduce: LocalVariable
			reduce(118), // .-, reduce: LocalVariable
			reduce(118), // *, reduce: LocalVariable
			reduce(118), // /, reduce: LocalVariable
			reduce(118), // //, reduce: LocalVariable
			reduce(118), // %, reduce: LocalVariable
			reduce(118), // .*, reduce: LocalVariable
			reduce(118), // ./, reduce: LocalVariable
			reduce(118), // .//, reduce: LocalVariable
			reduce(118), // ., reduce: LocalVariable
			nil,         // !
			nil,         // ~
			reduce(118), // **, reduce: LocalVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(118), // [, reduce: LocalVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(118), // [[, reduce: LocalVariable
			reduce(118), // [[[, reduce: LocalVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1899
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(147), // }, reduce: Rvalue
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(147), // ,, reduce: Rvalue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1900
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(148), // }, reduce: PrecedenceChainStart
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(148), // ,, reduce: PrecedenceChainStart
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1901
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(150), // }, reduce: TernaryTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(150), // ,, reduce: TernaryTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			shift(2912), // ?
			nil,         // :
			shift(2913), // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1902
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(152), // }, reduce: LogicalOrTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(152), // ,, reduce: LogicalOrTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(152), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(152), // ||, reduce: LogicalOrTerm
			shift(2914), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1903
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(154), // }, reduce: LogicalXORTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(154), // ,, reduce: LogicalXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(154), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(154), // ||, reduce: LogicalXORTerm
			reduce(154), // ^^, reduce: LogicalXORTerm
			shift(2915), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1904
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(156), // }, reduce: LogicalAndTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(156), // ,, reduce: LogicalAndTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(156), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(156), // ||, reduce: LogicalAndTerm
			reduce(156), // ^^, reduce: LogicalAndTerm
			reduce(156), // &&, reduce: LogicalAndTerm
			shift(2916), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1905
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(158), // }, reduce: AbsentCoalesceTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(158), // ,, reduce: AbsentCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(158), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(158), // ||, reduce: AbsentCoalesceTerm
			reduce(158), // ^^, reduce: AbsentCoalesceTerm
			reduce(158), // &&, reduce: AbsentCoalesceTerm
			reduce(158), // ??, reduce: AbsentCoalesceTerm
			shift(2917), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1906
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(160), // }, reduce: EmptyCoalesceTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(160), // ,, reduce: EmptyCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(160), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(160), // ||, reduce: EmptyCoalesceTerm
			reduce(160), // ^^, reduce: EmptyCoalesceTerm
			reduce(160), // &&, reduce: EmptyCoalesceTerm
			reduce(160), // ??, reduce: EmptyCoalesceTerm
			reduce(160), // ???, reduce: EmptyCoalesceTerm
			shift(2918), // =~
			shift(2919), // !=~
			shift(2920), // ==
			shift(2921), // !=
			shift(2922), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1907
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(166), // }, reduce: EqneTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2923), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(166), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(166), // ?, reduce: EqneTerm
			nil,         // :
			reduce(166), // ||, reduce: EqneTerm
			reduce(166), // ^^, reduce: EqneTerm
			reduce(166), // &&, reduce: EqneTerm
			reduce(166), // ??, reduce: EqneTerm
			reduce(166), // ???, reduce: EqneTerm
			reduce(166), // =~, reduce: EqneTerm
			reduce(166), // !=~, reduce: EqneTerm
			reduce(166), // ==, reduce: EqneTerm
			reduce(166), // !=, reduce: EqneTerm
			reduce(166), // <=>, reduce: EqneTerm
			shift(2924), // >=
			shift(2925), // <
			shift(2926), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1908
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(171), // }, reduce: CmpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(171), // >, reduce: CmpTerm
			nil,         // >>
			shift(2927), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(171), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(171), // ?, reduce: CmpTerm
			nil,         // :
			reduce(171), // ||, reduce: CmpTerm
			reduce(171), // ^^, reduce: CmpTerm
			reduce(171), // &&, reduce: CmpTerm
			reduce(171), // ??, reduce: CmpTerm
			reduce(171), // ???, reduce: CmpTerm
			reduce(171), // =~, reduce: CmpTerm
			reduce(171), // !=~, reduce: CmpTerm
			reduce(171), // ==, reduce: CmpTerm
			reduce(171), // !=, reduce: CmpTerm
			reduce(171), // <=>, reduce: CmpTerm
			reduce(171), // >=, reduce: CmpTerm
			reduce(171), // <, reduce: CmpTerm
			reduce(171), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1909
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(173), // }, reduce: BitwiseORTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(173), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(173), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(173), // ,, reduce: BitwiseORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(173), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(173), // ||, reduce: BitwiseORTerm
			reduce(173), // ^^, reduce: BitwiseORTerm
			reduce(173), // &&, reduce: BitwiseORTerm
			reduce(173), // ??, reduce: BitwiseORTerm
			reduce(173), // ???, reduce: BitwiseORTerm
			reduce(173), // =~, reduce: BitwiseORTerm
			reduce(173), // !=~, reduce: BitwiseORTerm
			reduce(173), // ==, reduce: BitwiseORTerm
			reduce(173), // !=, reduce: BitwiseORTerm
			reduce(173), // <=>, reduce: BitwiseORTerm
			reduce(173), // >=, reduce: BitwiseORTerm
			reduce(173), // <, reduce: BitwiseORTerm
			reduce(173), // <=, reduce: BitwiseORTerm
			shift(2928), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1910
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(175), // }, reduce: BitwiseXORTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(175), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(175), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(175), // ,, reduce: BitwiseXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(175), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(175), // ||, reduce: BitwiseXORTerm
			reduce(175), // ^^, reduce: BitwiseXORTerm
			reduce(175), // &&, reduce: BitwiseXORTerm
			reduce(175), // ??, reduce: BitwiseXORTerm
			reduce(175), // ???, reduce: BitwiseXORTerm
			reduce(175), // =~, reduce: BitwiseXORTerm
			reduce(175), // !=~, reduce: BitwiseXORTerm
			reduce(175), // ==, reduce: BitwiseXORTerm
			reduce(175), // !=, reduce: BitwiseXORTerm
			reduce(175), // <=>, reduce: BitwiseXORTerm
			reduce(175), // >=, reduce: BitwiseXORTerm
			reduce(175), // <, reduce: BitwiseXORTerm
			reduce(175), // <=, reduce: BitwiseXORTerm
			reduce(175), // ^, reduce: BitwiseXORTerm
			shift(2929), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1911
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(177), // }, reduce: BitwiseANDTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(177), // >, reduce: BitwiseANDTerm
			shift(2930), // >>
			reduce(177), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(177), // ,, reduce: BitwiseANDTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(177), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(177), // ||, reduce: BitwiseANDTerm
			reduce(177), // ^^, reduce: BitwiseANDTerm
			reduce(177), // &&, reduce: BitwiseANDTerm
			reduce(177), // ??, reduce: BitwiseANDTerm
			reduce(177), // ???, reduce: BitwiseANDTerm
			reduce(177), // =~, reduce: BitwiseANDTerm
			reduce(177), // !=~, reduce: BitwiseANDTerm
			reduce(177), // ==, reduce: BitwiseANDTerm
			reduce(177), // !=, reduce: BitwiseANDTerm
			reduce(177), // <=>, reduce: BitwiseANDTerm
			reduce(177), // >=, reduce: BitwiseANDTerm
			reduce(177), // <, reduce: BitwiseANDTerm
			reduce(177), // <=, reduce: BitwiseANDTerm
			reduce(177), // ^, reduce: BitwiseANDTerm
			reduce(177), // &, reduce: BitwiseANDTerm
			shift(2931), // <<
			shift(2932), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1912
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(181), // }, reduce: BitwiseShiftTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(181), // >, reduce: BitwiseShiftTerm
			reduce(181), // >>, reduce: BitwiseShiftTerm
			reduce(181), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(181), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(181), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(181), // ||, reduce: BitwiseShiftTerm
			reduce(181), // ^^, reduce: BitwiseShiftTerm
			reduce(181), // &&, reduce: BitwiseShiftTerm
			reduce(181), // ??, reduce: BitwiseShiftTerm
			reduce(181), // ???, reduce: BitwiseShiftTerm
			reduce(181), // =~, reduce: BitwiseShiftTerm
			reduce(181), // !=~, reduce: BitwiseShiftTerm
			reduce(181), // ==, reduce: BitwiseShiftTerm
			reduce(181), // !=, reduce: BitwiseShiftTerm
			reduce(181), // <=>, reduce: BitwiseShiftTerm
			reduce(181), // >=, reduce: BitwiseShiftTerm
			reduce(181), // <, reduce: BitwiseShiftTerm
			reduce(181), // <=, reduce: BitwiseShiftTerm
			reduce(181), // ^, reduce: BitwiseShiftTerm
			reduce(181), // &, reduce: BitwiseShiftTerm
			reduce(181), // <<, reduce: BitwiseShiftTerm
			reduce(181), // >>>, reduce: BitwiseShiftTerm
			shift(2933), // +
			shift(2934), // -
			shift(2935), // .+
			shift(2936), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1913
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1914
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(186), // }, reduce: AddsubdotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(186), // >, reduce: AddsubdotTerm
			reduce(186), // >>, reduce: AddsubdotTerm
			reduce(186), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(186), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(186), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(186), // ||, reduce: AddsubdotTerm
			reduce(186), // ^^, reduce: AddsubdotTerm
			reduce(186), // &&, reduce: AddsubdotTerm
			reduce(186), // ??, reduce: AddsubdotTerm
			reduce(186), // ???, reduce: AddsubdotTerm
			reduce(186), // =~, reduce: AddsubdotTerm
			reduce(186), // !=~, reduce: AddsubdotTerm
			reduce(186), // ==, reduce: AddsubdotTerm
			reduce(186), // !=, reduce: AddsubdotTerm
			reduce(186), // <=>, reduce: AddsubdotTerm
			reduce(186), // >=, reduce: AddsubdotTerm
			reduce(186), // <, reduce: AddsubdotTerm
			reduce(186), // <=, reduce: AddsubdotTerm
			reduce(186), // ^, reduce: AddsubdotTerm
			reduce(186), // &, reduce: AddsubdotTerm
			reduce(186), // <<, reduce: AddsubdotTerm
			reduce(186), // >>>, reduce: AddsubdotTerm
			reduce(186), // +, reduce: AddsubdotTerm
			reduce(186), // -, reduce: AddsubdotTerm
			reduce(186), // .+, reduce: AddsubdotTerm
			reduce(186), // .-, reduce: AddsubdotTerm
			shift(2938), // *
			shift(2939), // /
			shift(2940), // //
			shift(2941), // %
			shift(2942), // .*
			shift(2943), // ./
			shift(2944), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1915
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1916
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1917
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1918
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(194), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(194), // >, reduce: MuldivTerm
			reduce(194), // >>, reduce: MuldivTerm
			reduce(194), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(194), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(194), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(194), // ||, reduce: MuldivTerm
			reduce(194), // ^^, reduce: MuldivTerm
			reduce(194), // &&, reduce: MuldivTerm
			reduce(194), // ??, reduce: MuldivTerm
			reduce(194), // ???, reduce: MuldivTerm
			reduce(194), // =~, reduce: MuldivTerm
			reduce(194), // !=~, reduce: MuldivTerm
			reduce(194), // ==, reduce: MuldivTerm
			reduce(194), // !=, reduce: MuldivTerm
			reduce(194), // <=>, reduce: MuldivTerm
			reduce(194), // >=, reduce: MuldivTerm
			reduce(194), // <, reduce: MuldivTerm
			reduce(194), // <=, reduce: MuldivTerm
			reduce(194), // ^, reduce: MuldivTerm
			reduce(194), // &, reduce: MuldivTerm
			reduce(194), // <<, reduce: MuldivTerm
			reduce(194), // >>>, reduce: MuldivTerm
			reduce(194), // +, reduce: MuldivTerm
			reduce(194), // -, reduce: MuldivTerm
			reduce(194), // .+, reduce: MuldivTerm
			reduce(194), // .-, reduce: MuldivTerm
			reduce(194), // *, reduce: MuldivTerm
			reduce(194), // /, reduce: MuldivTerm
			reduce(194), // //, reduce: MuldivTerm
			reduce(194), // %, reduce: MuldivTerm
			reduce(194), // .*, reduce: MuldivTerm
			reduce(194), // ./, reduce: MuldivTerm
			reduce(194), // .//, reduce: MuldivTerm
			shift(2948), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1919
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(196), // }, reduce: DotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(196), // >, reduce: DotTerm
			reduce(196), // >>, reduce: DotTerm
			reduce(196), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(196), // ,, reduce: DotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(196), // ?, reduce: DotTerm
			nil,         // :
			reduce(196), // ||, reduce: DotTerm
			reduce(196), // ^^, reduce: DotTerm
			reduce(196), // &&, reduce: DotTerm
			reduce(196), // ??, reduce: DotTerm
			reduce(196), // ???, reduce: DotTerm
			reduce(196), // =~, reduce: DotTerm
			reduce(196), // !=~, reduce: DotTerm
			reduce(196), // ==, reduce: DotTerm
			reduce(196), // !=, reduce: DotTerm
			reduce(196), // <=>, reduce: DotTerm
			reduce(196), // >=, reduce: DotTerm
			reduce(196), // <, reduce: DotTerm
			reduce(196), // <=, reduce: DotTerm
			reduce(196), // ^, reduce: DotTerm
			reduce(196), // &, reduce: DotTerm
			reduce(196), // <<, reduce: DotTerm
			reduce(196), // >>>, reduce: DotTerm
			reduce(196), // +, reduce: DotTerm
			reduce(196), // -, reduce: DotTerm
			reduce(196), // .+, reduce: DotTerm
			reduce(196), // .-, reduce: DotTerm
			reduce(196), // *, reduce: DotTerm
			reduce(196), // /, reduce: DotTerm
			reduce(196), // //, reduce: DotTerm
			reduce(196), // %, reduce: DotTerm
			reduce(196), // .*, reduce: DotTerm
			reduce(196), // ./, reduce: DotTerm
			reduce(196), // .//, reduce: DotTerm
			reduce(196), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1920
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1921
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1922
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(203), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(203), // >, reduce: UnaryOpTerm
			reduce(203), // >>, reduce: UnaryOpTerm
			reduce(203), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(203), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(203), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(203), // ||, reduce: UnaryOpTerm
			reduce(203), // ^^, reduce: UnaryOpTerm
			reduce(203), // &&, reduce: UnaryOpTerm
			reduce(203), // ??, reduce: UnaryOpTerm
			reduce(203), // ???, reduce: UnaryOpTerm
			reduce(203), // =~, reduce: UnaryOpTerm
			reduce(203), // !=~, reduce: UnaryOpTerm
			reduce(203), // ==, reduce: UnaryOpTerm
			reduce(203), // !=, reduce: UnaryOpTerm
			reduce(203), // <=>, reduce: UnaryOpTerm
			reduce(203), // >=, reduce: UnaryOpTerm
			reduce(203), // <, reduce: UnaryOpTerm
			reduce(203), // <=, reduce: UnaryOpTerm
			reduce(203), // ^, reduce: UnaryOpTerm
			reduce(203), // &, reduce: UnaryOpTerm
			reduce(203), // <<, reduce: UnaryOpTerm
			reduce(203), // >>>, reduce: UnaryOpTerm
			reduce(203), // +, reduce: UnaryOpTerm
			reduce(203), // -, reduce: UnaryOpTerm
			reduce(203), // .+, reduce: UnaryOpTerm
			reduce(203), // .-, reduce: UnaryOpTerm
			reduce(203), // *, reduce: UnaryOpTerm
			reduce(203), // /, reduce: UnaryOpTerm
			reduce(203), // //, reduce: UnaryOpTerm
			reduce(203), // %, reduce: UnaryOpTerm
			reduce(203), // .*, reduce: UnaryOpTerm
			reduce(203), // ./, reduce: UnaryOpTerm
			reduce(203), // .//, reduce: UnaryOpTerm
			reduce(203), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1923
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(207), // }, reduce: PowTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(207), // >, reduce: PowTerm
			reduce(207), // >>, reduce: PowTerm
			reduce(207), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(207), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(207), // ?, reduce: PowTerm
			nil,         // :
			reduce(207), // ||, reduce: PowTerm
			reduce(207), // ^^, reduce: PowTerm
			reduce(207), // &&, reduce: PowTerm
			reduce(207), // ??, reduce: PowTerm
			reduce(207), // ???, reduce: PowTerm
			reduce(207), // =~, reduce: PowTerm
			reduce(207), // !=~, reduce: PowTerm
			reduce(207), // ==, reduce: PowTerm
			reduce(207), // !=, reduce: PowTerm
			reduce(207), // <=>, reduce: PowTerm
			reduce(207), // >=, reduce: PowTerm
			reduce(207), // <, reduce: PowTerm
			reduce(207), // <=, reduce: PowTerm
			reduce(207), // ^, reduce: PowTerm
			reduce(207), // &, reduce: PowTerm
			reduce(207), // <<, reduce: PowTerm
			reduce(207), // >>>, reduce: PowTerm
			reduce(207), // +, reduce: PowTerm
			reduce(207), // -, reduce: PowTerm
			reduce(207), // .+, reduce: PowTerm
			reduce(207), // .-, reduce: PowTerm
			reduce(207), // *, reduce: PowTerm
			reduce(207), // /, reduce: PowTerm
			reduce(207), // //, reduce: PowTerm
			reduce(207), // %, reduce: PowTerm
			reduce(207), // .*, reduce: PowTerm
			reduce(207), // ./, reduce: PowTerm
			reduce(207), // .//, reduce: PowTerm
			reduce(207), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			shift(2951), // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1924
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(209), // }, reduce: PrecedenceChainEnd
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(209), // >, reduce: PrecedenceChainEnd
			reduce(209), // >>, reduce: PrecedenceChainEnd
			reduce(209), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(209), // ,, reduce: PrecedenceChainEnd
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(209), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(209), // ||, reduce: PrecedenceChainEnd
			reduce(209), // ^^, reduce: PrecedenceChainEnd
			reduce(209), // &&, reduce: PrecedenceChainEnd
			reduce(209), // ??, reduce: PrecedenceChainEnd
			reduce(209), // ???, reduce: PrecedenceChainEnd
			reduce(209), // =~, reduce: PrecedenceChainEnd
			reduce(209), // !=~, reduce: PrecedenceChainEnd
			reduce(209), // ==, reduce: PrecedenceChainEnd
			reduce(209), // !=, reduce: PrecedenceChainEnd
			reduce(209), // <=>, reduce: PrecedenceChainEnd
			reduce(209), // >=, reduce: PrecedenceChainEnd
			reduce(209), // <, reduce: PrecedenceChainEnd
			reduce(209), // <=, reduce: PrecedenceChainEnd
			reduce(209), // ^, reduce: PrecedenceChainEnd
			reduce(209), // &, reduce: PrecedenceChainEnd
			reduce(209), // <<, reduce: PrecedenceChainEnd
			reduce(209), // >>>, reduce: PrecedenceChainEnd
			reduce(209), // +, reduce: PrecedenceChainEnd
			reduce(209), // -, reduce: PrecedenceChainEnd
			reduce(209), // .+, reduce: PrecedenceChainEnd
			reduce(209), // .-, reduce: PrecedenceChainEnd
			reduce(209), // *, reduce: PrecedenceChainEnd
			reduce(209), // /, reduce: PrecedenceChainEnd
			reduce(209), // //, reduce: PrecedenceChainEnd
			reduce(209), // %, reduce: PrecedenceChainEnd
			reduce(209), // .*, reduce: PrecedenceChainEnd
			reduce(209), // ./, reduce: PrecedenceChainEnd
			reduce(209), // .//, reduce: PrecedenceChainEnd
			reduce(209), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(209), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(2952), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			shift(2953), // [[
			shift(2954), // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1925
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(215), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(215), // >, reduce: MlrvalOrFunction
			reduce(215), // >>, reduce: MlrvalOrFunction
			reduce(215), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(215), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(215), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(215), // ||, reduce: MlrvalOrFunction
			reduce(215), // ^^, reduce: MlrvalOrFunction
			reduce(215), // &&, reduce: MlrvalOrFunction
			reduce(215), // ??, reduce: MlrvalOrFunction
			reduce(215), // ???, reduce: MlrvalOrFunction
			reduce(215), // =~, reduce: MlrvalOrFunction
			reduce(215), // !=~, reduce: MlrvalOrFunction
			reduce(215), // ==, reduce: MlrvalOrFunction
			reduce(215), // !=, reduce: MlrvalOrFunction
			reduce(215), // <=>, reduce: MlrvalOrFunction
			reduce(215), // >=, reduce: MlrvalOrFunction
			reduce(215), // <, reduce: MlrvalOrFunction
			reduce(215), // <=, reduce: MlrvalOrFunction
			reduce(215), // ^, reduce: MlrvalOrFunction
			reduce(215), // &, reduce: MlrvalOrFunction
			reduce(215), // <<, reduce: MlrvalOrFunction
			reduce(215), // >>>, reduce: MlrvalOrFunction
			reduce(215), // +, reduce: MlrvalOrFunction
			reduce(215), // -, reduce: MlrvalOrFunction
			reduce(215), // .+, reduce: MlrvalOrFunction
			reduce(215), // .-, reduce: MlrvalOrFunction
			reduce(215), // *, reduce: MlrvalOrFunction
			reduce(215), // /, reduce: MlrvalOrFunction
			reduce(215), // //, reduce: MlrvalOrFunction
			reduce(215), // %, reduce: MlrvalOrFunction
			reduce(215), // .*, reduce: MlrvalOrFunction
			reduce(215), // ./, reduce: MlrvalOrFunction
			reduce(215), // .//, reduce: MlrvalOrFunction
			reduce(215), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(215), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(215), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(215), // [[, reduce: MlrvalOrFunction
			reduce(215), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1926
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(216), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(216), // >, reduce: MlrvalOrFunction
			reduce(216), // >>, reduce: MlrvalOrFunction
			reduce(216), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(216), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(216), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(216), // ||, reduce: MlrvalOrFunction
			reduce(216), // ^^, reduce: MlrvalOrFunction
			reduce(216), // &&, reduce: MlrvalOrFunction
			reduce(216), // ??, reduce: MlrvalOrFunction
			reduce(216), // ???, reduce: MlrvalOrFunction
			reduce(216), // =~, reduce: MlrvalOrFunction
			reduce(216), // !=~, reduce: MlrvalOrFunction
			reduce(216), // ==, reduce: MlrvalOrFunction
			reduce(216), // !=, reduce: MlrvalOrFunction
			reduce(216), // <=>, reduce: MlrvalOrFunction
			reduce(216), // >=, reduce: MlrvalOrFunction
			reduce(216), // <, reduce: MlrvalOrFunction
			reduce(216), // <=, reduce: MlrvalOrFunction
			reduce(216), // ^, reduce: MlrvalOrFunction
			reduce(216), // &, reduce: MlrvalOrFunction
			reduce(216), // <<, reduce: MlrvalOrFunction
			reduce(216), // >>>, reduce: MlrvalOrFunction
			reduce(216), // +, reduce: MlrvalOrFunction
			reduce(216), // -, reduce: MlrvalOrFunction
			reduce(216), // .+, reduce: MlrvalOrFunction
			reduce(216), // .-, reduce: MlrvalOrFunction
			reduce(216), // *, reduce: MlrvalOrFunction
			reduce(216), // /, reduce: MlrvalOrFunction
			reduce(216), // //, reduce: MlrvalOrFunction
			reduce(216), // %, reduce: MlrvalOrFunction
			reduce(216), // .*, reduce: MlrvalOrFunction
			reduce(216), // ./, reduce: MlrvalOrFunction
			reduce(216), // .//, reduce: MlrvalOrFunction
			reduce(216), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(216), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(216), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(216), // [[, reduce: MlrvalOrFunction
			reduce(216), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1927
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(217), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(217), // >, reduce: MlrvalOrFunction
			reduce(217), // >>, reduce: MlrvalOrFunction
			reduce(217), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(217), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(217), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(217), // ||, reduce: MlrvalOrFunction
			reduce(217), // ^^, reduce: MlrvalOrFunction
			reduce(217), // &&, reduce: MlrvalOrFunction
			reduce(217), // ??, reduce: MlrvalOrFunction
			reduce(217), // ???, reduce: MlrvalOrFunction
			reduce(217), // =~, reduce: MlrvalOrFunction
			reduce(217), // !=~, reduce: MlrvalOrFunction
			reduce(217), // ==, reduce: MlrvalOrFunction
			reduce(217), // !=, reduce: MlrvalOrFunction
			reduce(217), // <=>, reduce: MlrvalOrFunction
			reduce(217), // >=, reduce: MlrvalOrFunction
			reduce(217), // <, reduce: MlrvalOrFunction
			reduce(217), // <=, reduce: MlrvalOrFunction
			reduce(217), // ^, reduce: MlrvalOrFunction
			reduce(217), // &, reduce: MlrvalOrFunction
			reduce(217), // <<, reduce: MlrvalOrFunction
			reduce(217), // >>>, reduce: MlrvalOrFunction
			reduce(217), // +, reduce: MlrvalOrFunction
			reduce(217), // -, reduce: MlrvalOrFunction
			reduce(217), // .+, reduce: MlrvalOrFunction
			reduce(217), // .-, reduce: MlrvalOrFunction
			reduce(217), // *, reduce: MlrvalOrFunction
			reduce(217), // /, reduce: MlrvalOrFunction
			reduce(217), // //, reduce: MlrvalOrFunction
			reduce(217), // %, reduce: MlrvalOrFunction
			reduce(217), // .*, reduce: MlrvalOrFunction
			reduce(217), // ./, reduce: MlrvalOrFunction
			reduce(217), // .//, reduce: MlrvalOrFunction
			reduce(217), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(217), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(217), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(217), // [[, reduce: MlrvalOrFunction
			reduce(217), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1928
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(218), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(218), // >, reduce: MlrvalOrFunction
			reduce(218), // >>, reduce: MlrvalOrFunction
			reduce(218), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(218), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(218), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(218), // ||, reduce: MlrvalOrFunction
			reduce(218), // ^^, reduce: MlrvalOrFunction
			reduce(218), // &&, reduce: MlrvalOrFunction
			reduce(218), // ??, reduce: MlrvalOrFunction
			reduce(218), // ???, reduce: MlrvalOrFunction
			reduce(218), // =~, reduce: MlrvalOrFunction
			reduce(218), // !=~, reduce: MlrvalOrFunction
			reduce(218), // ==, reduce: MlrvalOrFunction
			reduce(218), // !=, reduce: MlrvalOrFunction
			reduce(218), // <=>, reduce: MlrvalOrFunction
			reduce(218), // >=, reduce: MlrvalOrFunction
			reduce(218), // <, reduce: MlrvalOrFunction
			reduce(218), // <=, reduce: MlrvalOrFunction
			reduce(218), // ^, reduce: MlrvalOrFunction
			reduce(218), // &, reduce: MlrvalOrFunction
			reduce(218), // <<, reduce: MlrvalOrFunction
			reduce(218), // >>>, reduce: MlrvalOrFunction
			reduce(218), // +, reduce: MlrvalOrFunction
			reduce(218), // -, reduce: MlrvalOrFunction
			reduce(218), // .+, reduce: MlrvalOrFunction
			reduce(218), // .-, reduce: MlrvalOrFunction
			reduce(218), // *, reduce: MlrvalOrFunction
			reduce(218), // /, reduce: MlrvalOrFunction
			reduce(218), // //, reduce: MlrvalOrFunction
			reduce(218), // %, reduce: MlrvalOrFunction
			reduce(218), // .*, reduce: MlrvalOrFunction
			reduce(218), // ./, reduce: MlrvalOrFunction
			reduce(218), // .//, reduce: MlrvalOrFunction
			reduce(218), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(218), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(218), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(218), // [[, reduce: MlrvalOrFunction
			reduce(218), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1929
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(219), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(219), // >, reduce: MlrvalOrFunction
			reduce(219), // >>, reduce: MlrvalOrFunction
			reduce(219), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(219), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(219), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(219), // ||, reduce: MlrvalOrFunction
			reduce(219), // ^^, reduce: MlrvalOrFunction
			reduce(219), // &&, reduce: MlrvalOrFunction
			reduce(219), // ??, reduce: MlrvalOrFunction
			reduce(219), // ???, reduce: MlrvalOrFunction
			reduce(219), // =~, reduce: MlrvalOrFunction
			reduce(219), // !=~, reduce: MlrvalOrFunction
			reduce(219), // ==, reduce: MlrvalOrFunction
			reduce(219), // !=, reduce: MlrvalOrFunction
			reduce(219), // <=>, reduce: MlrvalOrFunction
			reduce(219), // >=, reduce: MlrvalOrFunction
			reduce(219), // <, reduce: MlrvalOrFunction
			reduce(219), // <=, reduce: MlrvalOrFunction
			reduce(219), // ^, reduce: MlrvalOrFunction
			reduce(219), // &, reduce: MlrvalOrFunction
			reduce(219), // <<, reduce: MlrvalOrFunction
			reduce(219), // >>>, reduce: MlrvalOrFunction
			reduce(219), // +, reduce: MlrvalOrFunction
			reduce(219), // -, reduce: MlrvalOrFunction
			reduce(219), // .+, reduce: MlrvalOrFunction
			reduce(219), // .-, reduce: MlrvalOrFunction
			reduce(219), // *, reduce: MlrvalOrFunction
			reduce(219), // /, reduce: MlrvalOrFunction
			reduce(219), // //, reduce: MlrvalOrFunction
			reduce(219), // %, reduce: MlrvalOrFunction
			reduce(219), // .*, reduce: MlrvalOrFunction
			reduce(219), // ./, reduce: MlrvalOrFunction
			reduce(219), // .//, reduce: MlrvalOrFunction
			reduce(219), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(219), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(219), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(219), // [[, reduce: MlrvalOrFunction
			reduce(219), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1930
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(220), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(220), // >, reduce: MlrvalOrFunction
			reduce(220), // >>, reduce: MlrvalOrFunction
			reduce(220), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(220), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(220), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(220), // ||, reduce: MlrvalOrFunction
			reduce(220), // ^^, reduce: MlrvalOrFunction
			reduce(220), // &&, reduce: MlrvalOrFunction
			reduce(220), // ??, reduce: MlrvalOrFunction
			reduce(220), // ???, reduce: MlrvalOrFunction
			reduce(220), // =~, reduce: MlrvalOrFunction
			reduce(220), // !=~, reduce: MlrvalOrFunction
			reduce(220), // ==, reduce: MlrvalOrFunction
			reduce(220), // !=, reduce: MlrvalOrFunction
			reduce(220), // <=>, reduce: MlrvalOrFunction
			reduce(220), // >=, reduce: MlrvalOrFunction
			reduce(220), // <, reduce: MlrvalOrFunction
			reduce(220), // <=, reduce: MlrvalOrFunction
			reduce(220), // ^, reduce: MlrvalOrFunction
			reduce(220), // &, reduce: MlrvalOrFunction
			reduce(220), // <<, reduce: MlrvalOrFunction
			reduce(220), // >>>, reduce: MlrvalOrFunction
			reduce(220), // +, reduce: MlrvalOrFunction
			reduce(220), // -, reduce: MlrvalOrFunction
			reduce(220), // .+, reduce: MlrvalOrFunction
			reduce(220), // .-, reduce: MlrvalOrFunction
			reduce(220), // *, reduce: MlrvalOrFunction
			reduce(220), // /, reduce: MlrvalOrFunction
			reduce(220), // //, reduce: MlrvalOrFunction
			reduce(220), // %, reduce: MlrvalOrFunction
			reduce(220), // .*, reduce: MlrvalOrFunction
			reduce(220), // ./, reduce: MlrvalOrFunction
			reduce(220), // .//, reduce: MlrvalOrFunction
			reduce(220), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(220), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(220), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(220), // [[, reduce: MlrvalOrFunction
			reduce(220), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1931
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(221), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(221), // >, reduce: MlrvalOrFunction
			reduce(221), // >>, reduce: MlrvalOrFunction
			reduce(221), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(221), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(221), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(221), // ||, reduce: MlrvalOrFunction
			reduce(221), // ^^, reduce: MlrvalOrFunction
			reduce(221), // &&, reduce: MlrvalOrFunction
			reduce(221), // ??, reduce: MlrvalOrFunction
			reduce(221), // ???, reduce: MlrvalOrFunction
			reduce(221), // =~, reduce: MlrvalOrFunction
			reduce(221), // !=~, reduce: MlrvalOrFunction
			reduce(221), // ==, reduce: MlrvalOrFunction
			reduce(221), // !=, reduce: MlrvalOrFunction
			reduce(221), // <=>, reduce: MlrvalOrFunction
			reduce(221), // >=, reduce: MlrvalOrFunction
			reduce(221), // <, reduce: MlrvalOrFunction
			reduce(221), // <=, reduce: MlrvalOrFunction
			reduce(221), // ^, reduce: MlrvalOrFunction
			reduce(221), // &, reduce: MlrvalOrFunction
			reduce(221), // <<, reduce: MlrvalOrFunction
			reduce(221), // >>>, reduce: MlrvalOrFunction
			reduce(221), // +, reduce: MlrvalOrFunction
			reduce(221), // -, reduce: MlrvalOrFunction
			reduce(221), // .+, reduce: MlrvalOrFunction
			reduce(221), // .-, reduce: MlrvalOrFunction
			reduce(221), // *, reduce: MlrvalOrFunction
			reduce(221), // /, reduce: MlrvalOrFunction
			reduce(221), // //, reduce: MlrvalOrFunction
			reduce(221), // %, reduce: MlrvalOrFunction
			reduce(221), // .*, reduce: MlrvalOrFunction
			reduce(221), // ./, reduce: MlrvalOrFunction
			reduce(221), // .//, reduce: MlrvalOrFunction
			reduce(221), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(221), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(221), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(221), // [[, reduce: MlrvalOrFunction
			reduce(221), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1932
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(222), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(222), // >, reduce: MlrvalOrFunction
			reduce(222), // >>, reduce: MlrvalOrFunction
			reduce(222), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(222), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(222), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(222), // ||, reduce: MlrvalOrFunction
			reduce(222), // ^^, reduce: MlrvalOrFunction
			reduce(222), // &&, reduce: MlrvalOrFunction
			reduce(222), // ??, reduce: MlrvalOrFunction
			reduce(222), // ???, reduce: MlrvalOrFunction
			reduce(222), // =~, reduce: MlrvalOrFunction
			reduce(222), // !=~, reduce: MlrvalOrFunction
			reduce(222), // ==, reduce: MlrvalOrFunction
			reduce(222), // !=, reduce: MlrvalOrFunction
			reduce(222), // <=>, reduce: MlrvalOrFunction
			reduce(222), // >=, reduce: MlrvalOrFunction
			reduce(222), // <, reduce: MlrvalOrFunction
			reduce(222), // <=, reduce: MlrvalOrFunction
			reduce(222), // ^, reduce: MlrvalOrFunction
			reduce(222), // &, reduce: MlrvalOrFunction
			reduce(222), // <<, reduce: MlrvalOrFunction
			reduce(222), // >>>, reduce: MlrvalOrFunction
			reduce(222), // +, reduce: MlrvalOrFunction
			reduce(222), // -, reduce: MlrvalOrFunction
			reduce(222), // .+, reduce: MlrvalOrFunction
			reduce(222), // .-, reduce: MlrvalOrFunction
			reduce(222), // *, reduce: MlrvalOrFunction
			reduce(222), // /, reduce: MlrvalOrFunction
			reduce(222), // //, reduce: MlrvalOrFunction
			reduce(222), // %, reduce: MlrvalOrFunction
			reduce(222), // .*, reduce: MlrvalOrFunction
			reduce(222), // ./, reduce: MlrvalOrFunction
			reduce(222), // .//, reduce: MlrvalOrFunction
			reduce(222), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(222), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(222), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(222), // [[, reduce: MlrvalOrFunction
			reduce(222), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1933
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(223), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(223), // >, reduce: MlrvalOrFunction
			reduce(223), // >>, reduce: MlrvalOrFunction
			reduce(223), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(223), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(223), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(223), // ||, reduce: MlrvalOrFunction
			reduce(223), // ^^, reduce: MlrvalOrFunction
			reduce(223), // &&, reduce: MlrvalOrFunction
			reduce(223), // ??, reduce: MlrvalOrFunction
			reduce(223), // ???, reduce: MlrvalOrFunction
			reduce(223), // =~, reduce: MlrvalOrFunction
			reduce(223), // !=~, reduce: MlrvalOrFunction
			reduce(223), // ==, reduce: MlrvalOrFunction
			reduce(223), // !=, reduce: MlrvalOrFunction
			reduce(223), // <=>, reduce: MlrvalOrFunction
			reduce(223), // >=, reduce: MlrvalOrFunction
			reduce(223), // <, reduce: MlrvalOrFunction
			reduce(223), // <=, reduce: MlrvalOrFunction
			reduce(223), // ^, reduce: MlrvalOrFunction
			reduce(223), // &, reduce: MlrvalOrFunction
			reduce(223), // <<, reduce: MlrvalOrFunction
			reduce(223), // >>>, reduce: MlrvalOrFunction
			reduce(223), // +, reduce: MlrvalOrFunction
			reduce(223), // -, reduce: MlrvalOrFunction
			reduce(223), // .+, reduce: MlrvalOrFunction
			reduce(223), // .-, reduce: MlrvalOrFunction
			reduce(223), // *, reduce: MlrvalOrFunction
			reduce(223), // /, reduce: MlrvalOrFunction
			reduce(223), // //, reduce: MlrvalOrFunction
			reduce(223), // %, reduce: MlrvalOrFunction
			reduce(223), // .*, reduce: MlrvalOrFunction
			reduce(223), // ./, reduce: MlrvalOrFunction
			reduce(223), // .//, reduce: MlrvalOrFunction
			reduce(223), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(223), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(223), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(223), // [[, reduce: MlrvalOrFunction
			reduce(223), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1934
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(224), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(224), // >, reduce: MlrvalOrFunction
			reduce(224), // >>, reduce: MlrvalOrFunction
			reduce(224), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(224), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(224), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(224), // ||, reduce: MlrvalOrFunction
			reduce(224), // ^^, reduce: MlrvalOrFunction
			reduce(224), // &&, reduce: MlrvalOrFunction
			reduce(224), // ??, reduce: MlrvalOrFunction
			reduce(224), // ???, reduce: MlrvalOrFunction
			reduce(224), // =~, reduce: MlrvalOrFunction
			reduce(224), // !=~, reduce: MlrvalOrFunction
			reduce(224), // ==, reduce: MlrvalOrFunction
			reduce(224), // !=, reduce: MlrvalOrFunction
			reduce(224), // <=>, reduce: MlrvalOrFunction
			reduce(224), // >=, reduce: MlrvalOrFunction
			reduce(224), // <, reduce: MlrvalOrFunction
			reduce(224), // <=, reduce: MlrvalOrFunction
			reduce(224), // ^, reduce: MlrvalOrFunction
			reduce(224), // &, reduce: MlrvalOrFunction
			reduce(224), // <<, reduce: MlrvalOrFunction
			reduce(224), // >>>, reduce: MlrvalOrFunction
			reduce(224), // +, reduce: MlrvalOrFunction
			reduce(224), // -, reduce: MlrvalOrFunction
			reduce(224), // .+, reduce: MlrvalOrFunction
			reduce(224), // .-, reduce: MlrvalOrFunction
			reduce(224), // *, reduce: MlrvalOrFunction
			reduce(224), // /, reduce: MlrvalOrFunction
			reduce(224), // //, reduce: MlrvalOrFunction
			reduce(224), // %, reduce: MlrvalOrFunction
			reduce(224), // .*, reduce: MlrvalOrFunction
			reduce(224), // ./, reduce: MlrvalOrFunction
			reduce(224), // .//, reduce: MlrvalOrFunction
			reduce(224), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(224), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(224), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(224), // [[, reduce: MlrvalOrFunction
			reduce(224), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1935
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(225), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(225), // >, reduce: MlrvalOrFunction
			reduce(225), // >>, reduce: MlrvalOrFunction
			reduce(225), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(225), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(225), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(225), // ||, reduce: MlrvalOrFunction
			reduce(225), // ^^, reduce: MlrvalOrFunction
			reduce(225), // &&, reduce: MlrvalOrFunction
			reduce(225), // ??, reduce: MlrvalOrFunction
			reduce(225), // ???, reduce: MlrvalOrFunction
			reduce(225), // =~, reduce: MlrvalOrFunction
			reduce(225), // !=~, reduce: MlrvalOrFunction
			reduce(225), // ==, reduce: MlrvalOrFunction
			reduce(225), // !=, reduce: MlrvalOrFunction
			reduce(225), // <=>, reduce: MlrvalOrFunction
			reduce(225), // >=, reduce: MlrvalOrFunction
			reduce(225), // <, reduce: MlrvalOrFunction
			reduce(225), // <=, reduce: MlrvalOrFunction
			reduce(225), // ^, reduce: MlrvalOrFunction
			reduce(225), // &, reduce: MlrvalOrFunction
			reduce(225), // <<, reduce: MlrvalOrFunction
			reduce(225), // >>>, reduce: MlrvalOrFunction
			reduce(225), // +, reduce: MlrvalOrFunction
			reduce(225), // -, reduce: MlrvalOrFunction
			reduce(225), // .+, reduce: MlrvalOrFunction
			reduce(225), // .-, reduce: MlrvalOrFunction
			reduce(225), // *, reduce: MlrvalOrFunction
			reduce(225), // /, reduce: MlrvalOrFunction
			reduce(225), // //, reduce: MlrvalOrFunction
			reduce(225), // %, reduce: MlrvalOrFunction
			reduce(225), // .*, reduce: MlrvalOrFunction
			reduce(225), // ./, reduce: MlrvalOrFunction
			reduce(225), // .//, reduce: MlrvalOrFunction
			reduce(225), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(225), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(225), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(225), // [[, reduce: MlrvalOrFunction
			reduce(225), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1936
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(226), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(226), // >, reduce: MlrvalOrFunction
			reduce(226), // >>, reduce: MlrvalOrFunction
			reduce(226), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(226), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(226), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(226), // ||, reduce: MlrvalOrFunction
			reduce(226), // ^^, reduce: MlrvalOrFunction
			reduce(226), // &&, reduce: MlrvalOrFunction
			reduce(226), // ??, reduce: MlrvalOrFunction
			reduce(226), // ???, reduce: MlrvalOrFunction
			reduce(226), // =~, reduce: MlrvalOrFunction
			reduce(226), // !=~, reduce: MlrvalOrFunction
			reduce(226), // ==, reduce: MlrvalOrFunction
			reduce(226), // !=, reduce: MlrvalOrFunction
			reduce(226), // <=>, reduce: MlrvalOrFunction
			reduce(226), // >=, reduce: MlrvalOrFunction
			reduce(226), // <, reduce: MlrvalOrFunction
			reduce(226), // <=, reduce: MlrvalOrFunction
			reduce(226), // ^, reduce: MlrvalOrFunction
			reduce(226), // &, reduce: MlrvalOrFunction
			reduce(226), // <<, reduce: MlrvalOrFunction
			reduce(226), // >>>, reduce: MlrvalOrFunction
			reduce(226), // +, reduce: MlrvalOrFunction
			reduce(226), // -, reduce: MlrvalOrFunction
			reduce(226), // .+, reduce: MlrvalOrFunction
			reduce(226), // .-, reduce: MlrvalOrFunction
			reduce(226), // *, reduce: MlrvalOrFunction
			reduce(226), // /, reduce: MlrvalOrFunction
			reduce(226), // //, reduce: MlrvalOrFunction
			reduce(226), // %, reduce: MlrvalOrFunction
			reduce(226), // .*, reduce: MlrvalOrFunction
			reduce(226), // ./, reduce: MlrvalOrFunction
			reduce(226), // .//, reduce: MlrvalOrFunction
			reduce(226), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(226), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(226), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(226), // [[, reduce: MlrvalOrFunction
			reduce(226), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1937
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(227), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(227), // >, reduce: MlrvalOrFunction
			reduce(227), // >>, reduce: MlrvalOrFunction
			reduce(227), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(227), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(227), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(227), // ||, reduce: MlrvalOrFunction
			reduce(227), // ^^, reduce: MlrvalOrFunction
			reduce(227), // &&, reduce: MlrvalOrFunction
			reduce(227), // ??, reduce: MlrvalOrFunction
			reduce(227), // ???, reduce: MlrvalOrFunction
			reduce(227), // =~, reduce: MlrvalOrFunction
			reduce(227), // !=~, reduce: MlrvalOrFunction
			reduce(227), // ==, reduce: MlrvalOrFunction
			reduce(227), // !=, reduce: MlrvalOrFunction
			reduce(227), // <=>, reduce: MlrvalOrFunction
			reduce(227), // >=, reduce: MlrvalOrFunction
			reduce(227), // <, reduce: MlrvalOrFunction
			reduce(227), // <=, reduce: MlrvalOrFunction
			reduce(227), // ^, reduce: MlrvalOrFunction
			reduce(227), // &, reduce: MlrvalOrFunction
			reduce(227), // <<, reduce: MlrvalOrFunction
			reduce(227), // >>>, reduce: MlrvalOrFunction
			reduce(227), // +, reduce: MlrvalOrFunction
			reduce(227), // -, reduce: MlrvalOrFunction
			reduce(227), // .+, reduce: MlrvalOrFunction
			reduce(227), // .-, reduce: MlrvalOrFunction
			reduce(227), // *, reduce: MlrvalOrFunction
			reduce(227), // /, reduce: MlrvalOrFunction
			reduce(227), // //, reduce: MlrvalOrFunction
			reduce(227), // %, reduce: MlrvalOrFunction
			reduce(227), // .*, reduce: MlrvalOrFunction
			reduce(227), // ./, reduce: MlrvalOrFunction
			reduce(227), // .//, reduce: MlrvalOrFunction
			reduce(227), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(227), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(227), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(227), // [[, reduce: MlrvalOrFunction
			reduce(227), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1938
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(752),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(756),  // (
			nil,         // )
			nil,         // emitp
			shift(768),  // field_name
			shift(769),  // $[
			shift(2955), // ]
			shift(771),  // braced_field_name
			shift(772),  // $[[
			shift(773),  // $[[[
			shift(774),  // full_srec
			shift(776),  // oosvar_name
			shift(777),  // @[
			shift(778),  // braced_oosvar_name
			shift(779),  // full_oosvar
			shift(780),  // all
			shift(781),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(796),  // +
			shift(798),  // -
			shift(799),  // .+
			shift(800),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(803),  // !
			shift(804),  // ~
			nil,         // **
			shift(809),  // string_literal
			shift(810),  // regex_case_insensitive
			shift(811),  // int_literal
			shift(812),  // float_literal
			shift(813),  // boolean_literal
			shift(814),  // null_literal
			shift(815),  // inf_literal
			shift(816),  // nan_literal
			shift(817),  // const_M_PI
			shift(818),  // const_M_E
			shift(819),  // panic
			shift(821),  // [
			shift(824),  // ctx_IPS
			shift(825),  // ctx_IFS
			shift(826),  // ctx_IRS
			shift(827),  // ctx_OPS
			shift(828),  // ctx_OFS
			shift(829),  // ctx_ORS
			shift(830),  // ctx_FLATSEP
			shift(831),  // ctx_NF
			shift(832),  // ctx_NR
			shift(833),  // ctx_FNR
			shift(834),  // ctx_FILENAME
			shift(835),  // ctx_FILENUM
			shift(837),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(844),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1939
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(240), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(240), // >, reduce: MlrvalOrFunction
			reduce(240), // >>, reduce: MlrvalOrFunction
			reduce(240), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(240), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(240), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(240), // ||, reduce: MlrvalOrFunction
			reduce(240), // ^^, reduce: MlrvalOrFunction
			reduce(240), // &&, reduce: MlrvalOrFunction
			reduce(240), // ??, reduce: MlrvalOrFunction
			reduce(240), // ???, reduce: MlrvalOrFunction
			reduce(240), // =~, reduce: MlrvalOrFunction
			reduce(240), // !=~, reduce: MlrvalOrFunction
			reduce(240), // ==, reduce: MlrvalOrFunction
			reduce(240), // !=, reduce: MlrvalOrFunction
			reduce(240), // <=>, reduce: MlrvalOrFunction
			reduce(240), // >=, reduce: MlrvalOrFunction
			reduce(240), // <, reduce: MlrvalOrFunction
			reduce(240), // <=, reduce: MlrvalOrFunction
			reduce(240), // ^, reduce: MlrvalOrFunction
			reduce(240), // &, reduce: MlrvalOrFunction
			reduce(240), // <<, reduce: MlrvalOrFunction
			reduce(240), // >>>, reduce: MlrvalOrFunction
			reduce(240), // +, reduce: MlrvalOrFunction
			reduce(240), // -, reduce: MlrvalOrFunction
			reduce(240), // .+, reduce: MlrvalOrFunction
			reduce(240), // .-, reduce: MlrvalOrFunction
			reduce(240), // *, reduce: MlrvalOrFunction
			reduce(240), // /, reduce: MlrvalOrFunction
			reduce(240), // //, reduce: MlrvalOrFunction
			reduce(240), // %, reduce: MlrvalOrFunction
			reduce(240), // .*, reduce: MlrvalOrFunction
			reduce(240), // ./, reduce: MlrvalOrFunction
			reduce(240), // .//, reduce: MlrvalOrFunction
			reduce(240), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(240), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(240), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(240), // [[, reduce: MlrvalOrFunction
			reduce(240), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1940
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(241), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(241), // >, reduce: ContextVariable
			reduce(241), // >>, reduce: ContextVariable
			reduce(241), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(241), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(241), // ?, reduce: ContextVariable
			nil,         // :
			reduce(241), // ||, reduce: ContextVariable
			reduce(241), // ^^, reduce: ContextVariable
			reduce(241), // &&, reduce: ContextVariable
			reduce(241), // ??, reduce: ContextVariable
			reduce(241), // ???, reduce: ContextVariable
			reduce(241), // =~, reduce: ContextVariable
			reduce(241), // !=~, reduce: ContextVariable
			reduce(241), // ==, reduce: ContextVariable
			reduce(241), // !=, reduce: ContextVariable
			reduce(241), // <=>, reduce: ContextVariable
			reduce(241), // >=, reduce: ContextVariable
			reduce(241), // <, reduce: ContextVariable
			reduce(241), // <=, reduce: ContextVariable
			reduce(241), // ^, reduce: ContextVariable
			reduce(241), // &, reduce: ContextVariable
			reduce(241), // <<, reduce: ContextVariable
			reduce(241), // >>>, reduce: ContextVariable
			reduce(241), // +, reduce: ContextVariable
			reduce(241), // -, reduce: ContextVariable
			reduce(241), // .+, reduce: ContextVariable
			reduce(241), // .-, reduce: ContextVariable
			reduce(241), // *, reduce: ContextVariable
			reduce(241), // /, reduce: ContextVariable
			reduce(241), // //, reduce: ContextVariable
			reduce(241), // %, reduce: ContextVariable
			reduce(241), // .*, reduce: ContextVariable
			reduce(241), // ./, reduce: ContextVariable
			reduce(241), // .//, reduce: ContextVariable
			reduce(241), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(241), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(241), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(241), // [[, reduce: ContextVariable
			reduce(241), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1941
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(242), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(242), // >, reduce: ContextVariable
			reduce(242), // >>, reduce: ContextVariable
			reduce(242), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(242), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(242), // ?, reduce: ContextVariable
			nil,         // :
			reduce(242), // ||, reduce: ContextVariable
			reduce(242), // ^^, reduce: ContextVariable
			reduce(242), // &&, reduce: ContextVariable
			reduce(242), // ??, reduce: ContextVariable
			reduce(242), // ???, reduce: ContextVariable
			reduce(242), // =~, reduce: ContextVariable
			reduce(242), // !=~, reduce: ContextVariable
			reduce(242), // ==, reduce: ContextVariable
			reduce(242), // !=, reduce: ContextVariable
			reduce(242), // <=>, reduce: ContextVariable
			reduce(242), // >=, reduce: ContextVariable
			reduce(242), // <, reduce: ContextVariable
			reduce(242), // <=, reduce: ContextVariable
			reduce(242), // ^, reduce: ContextVariable
			reduce(242), // &, reduce: ContextVariable
			reduce(242), // <<, reduce: ContextVariable
			reduce(242), // >>>, reduce: ContextVariable
			reduce(242), // +, reduce: ContextVariable
			reduce(242), // -, reduce: ContextVariable
			reduce(242), // .+, reduce: ContextVariable
			reduce(242), // .-, reduce: ContextVariable
			reduce(242), // *, reduce: ContextVariable
			reduce(242), // /, reduce: ContextVariable
			reduce(242), // //, reduce: ContextVariable
			reduce(242), // %, reduce: ContextVariable
			reduce(242), // .*, reduce: ContextVariable
			reduce(242), // ./, reduce: ContextVariable
			reduce(242), // .//, reduce: ContextVariable
			reduce(242), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(242), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(242), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(242), // [[, reduce: ContextVariable
			reduce(242), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1942
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(243), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(243), // >, reduce: ContextVariable
			reduce(243), // >>, reduce: ContextVariable
			reduce(243), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(243), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(243), // ?, reduce: ContextVariable
			nil,         // :
			reduce(243), // ||, reduce: ContextVariable
			reduce(243), // ^^, reduce: ContextVariable
			reduce(243), // &&, reduce: ContextVariable
			reduce(243), // ??, reduce: ContextVariable
			reduce(243), // ???, reduce: ContextVariable
			reduce(243), // =~, reduce: ContextVariable
			reduce(243), // !=~, reduce: ContextVariable
			reduce(243), // ==, reduce: ContextVariable
			reduce(243), // !=, reduce: ContextVariable
			reduce(243), // <=>, reduce: ContextVariable
			reduce(243), // >=, reduce: ContextVariable
			reduce(243), // <, reduce: ContextVariable
			reduce(243), // <=, reduce: ContextVariable
			reduce(243), // ^, reduce: ContextVariable
			reduce(243), // &, reduce: ContextVariable
			reduce(243), // <<, reduce: ContextVariable
			reduce(243), // >>>, reduce: ContextVariable
			reduce(243), // +, reduce: ContextVariable
			reduce(243), // -, reduce: ContextVariable
			reduce(243), // .+, reduce: ContextVariable
			reduce(243), // .-, reduce: ContextVariable
			reduce(243), // *, reduce: ContextVariable
			reduce(243), // /, reduce: ContextVariable
			reduce(243), // //, reduce: ContextVariable
			reduce(243), // %, reduce: ContextVariable
			reduce(243), // .*, reduce: ContextVariable
			reduce(243), // ./, reduce: ContextVariable
			reduce(243), // .//, reduce: ContextVariable
			reduce(243), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(243), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(243), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(243), // [[, reduce: ContextVariable
			reduce(243), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1943
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(244), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(244), // >, reduce: ContextVariable
			reduce(244), // >>, reduce: ContextVariable
			reduce(244), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(244), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(244), // ?, reduce: ContextVariable
			nil,         // :
			reduce(244), // ||, reduce: ContextVariable
			reduce(244), // ^^, reduce: ContextVariable
			reduce(244), // &&, reduce: ContextVariable
			reduce(244), // ??, reduce: ContextVariable
			reduce(244), // ???, reduce: ContextVariable
			reduce(244), // =~, reduce: ContextVariable
			reduce(244), // !=~, reduce: ContextVariable
			reduce(244), // ==, reduce: ContextVariable
			reduce(244), // !=, reduce: ContextVariable
			reduce(244), // <=>, reduce: ContextVariable
			reduce(244), // >=, reduce: ContextVariable
			reduce(244), // <, reduce: ContextVariable
			reduce(244), // <=, reduce: ContextVariable
			reduce(244), // ^, reduce: ContextVariable
			reduce(244), // &, reduce: ContextVariable
			reduce(244), // <<, reduce: ContextVariable
			reduce(244), // >>>, reduce: ContextVariable
			reduce(244), // +, reduce: ContextVariable
			reduce(244), // -, reduce: ContextVariable
			reduce(244), // .+, reduce: ContextVariable
			reduce(244), // .-, reduce: ContextVariable
			reduce(244), // *, reduce: ContextVariable
			reduce(244), // /, reduce: ContextVariable
			reduce(244), // //, reduce: ContextVariable
			reduce(244), // %, reduce: ContextVariable
			reduce(244), // .*, reduce: ContextVariable
			reduce(244), // ./, reduce: ContextVariable
			reduce(244), // .//, reduce: ContextVariable
			reduce(244), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(244), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(244), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(244), // [[, reduce: ContextVariable
			reduce(244), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1944
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(245), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(245), // >, reduce: ContextVariable
			reduce(245), // >>, reduce: ContextVariable
			reduce(245), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(245), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(245), // ?, reduce: ContextVariable
			nil,         // :
			reduce(245), // ||, reduce: ContextVariable
			reduce(245), // ^^, reduce: ContextVariable
			reduce(245), // &&, reduce: ContextVariable
			reduce(245), // ??, reduce: ContextVariable
			reduce(245), // ???, reduce: ContextVariable
			reduce(245), // =~, reduce: ContextVariable
			reduce(245), // !=~, reduce: ContextVariable
			reduce(245), // ==, reduce: ContextVariable
			reduce(245), // !=, reduce: ContextVariable
			reduce(245), // <=>, reduce: ContextVariable
			reduce(245), // >=, reduce: ContextVariable
			reduce(245), // <, reduce: ContextVariable
			reduce(245), // <=, reduce: ContextVariable
			reduce(245), // ^, reduce: ContextVariable
			reduce(245), // &, reduce: ContextVariable
			reduce(245), // <<, reduce: ContextVariable
			reduce(245), // >>>, reduce: ContextVariable
			reduce(245), // +, reduce: ContextVariable
			reduce(245), // -, reduce: ContextVariable
			reduce(245), // .+, reduce: ContextVariable
			reduce(245), // .-, reduce: ContextVariable
			reduce(245), // *, reduce: ContextVariable
			reduce(245), // /, reduce: ContextVariable
			reduce(245), // //, reduce: ContextVariable
			reduce(245), // %, reduce: ContextVariable
			reduce(245), // .*, reduce: ContextVariable
			reduce(245), // ./, reduce: ContextVariable
			reduce(245), // .//, reduce: ContextVariable
			reduce(245), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(245), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(245), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(245), // [[, reduce: ContextVariable
			reduce(245), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1945
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(246), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(246), // >, reduce: ContextVariable
			reduce(246), // >>, reduce: ContextVariable
			reduce(246), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(246), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(246), // ?, reduce: ContextVariable
			nil,         // :
			reduce(246), // ||, reduce: ContextVariable
			reduce(246), // ^^, reduce: ContextVariable
			reduce(246), // &&, reduce: ContextVariable
			reduce(246), // ??, reduce: ContextVariable
			reduce(246), // ???, reduce: ContextVariable
			reduce(246), // =~, reduce: ContextVariable
			reduce(246), // !=~, reduce: ContextVariable
			reduce(246), // ==, reduce: ContextVariable
			reduce(246), // !=, reduce: ContextVariable
			reduce(246), // <=>, reduce: ContextVariable
			reduce(246), // >=, reduce: ContextVariable
			reduce(246), // <, reduce: ContextVariable
			reduce(246), // <=, reduce: ContextVariable
			reduce(246), // ^, reduce: ContextVariable
			reduce(246), // &, reduce: ContextVariable
			reduce(246), // <<, reduce: ContextVariable
			reduce(246), // >>>, reduce: ContextVariable
			reduce(246), // +, reduce: ContextVariable
			reduce(246), // -, reduce: ContextVariable
			reduce(246), // .+, reduce: ContextVariable
			reduce(246), // .-, reduce: ContextVariable
			reduce(246), // *, reduce: ContextVariable
			reduce(246), // /, reduce: ContextVariable
			reduce(246), // //, reduce: ContextVariable
			reduce(246), // %, reduce: ContextVariable
			reduce(246), // .*, reduce: ContextVariable
			reduce(246), // ./, reduce: ContextVariable
			reduce(246), // .//, reduce: ContextVariable
			reduce(246), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(246), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(246), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(246), // [[, reduce: ContextVariable
			reduce(246), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1946
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(247), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(247), // >, reduce: ContextVariable
			reduce(247), // >>, reduce: ContextVariable
			reduce(247), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(247), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(247), // ?, reduce: ContextVariable
			nil,         // :
			reduce(247), // ||, reduce: ContextVariable
			reduce(247), // ^^, reduce: ContextVariable
			reduce(247), // &&, reduce: ContextVariable
			reduce(247), // ??, reduce: ContextVariable
			reduce(247), // ???, reduce: ContextVariable
			reduce(247), // =~, reduce: ContextVariable
			reduce(247), // !=~, reduce: ContextVariable
			reduce(247), // ==, reduce: ContextVariable
			reduce(247), // !=, reduce: ContextVariable
			reduce(247), // <=>, reduce: ContextVariable
			reduce(247), // >=, reduce: ContextVariable
			reduce(247), // <, reduce: ContextVariable
			reduce(247), // <=, reduce: ContextVariable
			reduce(247), // ^, reduce: ContextVariable
			reduce(247), // &, reduce: ContextVariable
			reduce(247), // <<, reduce: ContextVariable
			reduce(247), // >>>, reduce: ContextVariable
			reduce(247), // +, reduce: ContextVariable
			reduce(247), // -, reduce: ContextVariable
			reduce(247), // .+, reduce: ContextVariable
			reduce(247), // .-, reduce: ContextVariable
			reduce(247), // *, reduce: ContextVariable
			reduce(247), // /, reduce: ContextVariable
			reduce(247), // //, reduce: ContextVariable
			reduce(247), // %, reduce: ContextVariable
			reduce(247), // .*, reduce: ContextVariable
			reduce(247), // ./, reduce: ContextVariable
			reduce(247), // .//, reduce: ContextVariable
			reduce(247), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(247), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(247), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(247), // [[, reduce: ContextVariable
			reduce(247), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1947
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(248), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(248), // >, reduce: ContextVariable
			reduce(248), // >>, reduce: ContextVariable
			reduce(248), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(248), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(248), // ?, reduce: ContextVariable
			nil,         // :
			reduce(248), // ||, reduce: ContextVariable
			reduce(248), // ^^, reduce: ContextVariable
			reduce(248), // &&, reduce: ContextVariable
			reduce(248), // ??, reduce: ContextVariable
			reduce(248), // ???, reduce: ContextVariable
			reduce(248), // =~, reduce: ContextVariable
			reduce(248), // !=~, reduce: ContextVariable
			reduce(248), // ==, reduce: ContextVariable
			reduce(248), // !=, reduce: ContextVariable
			reduce(248), // <=>, reduce: ContextVariable
			reduce(248), // >=, reduce: ContextVariable
			reduce(248), // <, reduce: ContextVariable
			reduce(248), // <=, reduce: ContextVariable
			reduce(248), // ^, reduce: ContextVariable
			reduce(248), // &, reduce: ContextVariable
			reduce(248), // <<, reduce: ContextVariable
			reduce(248), // >>>, reduce: ContextVariable
			reduce(248), // +, reduce: ContextVariable
			reduce(248), // -, reduce: ContextVariable
			reduce(248), // .+, reduce: ContextVariable
			reduce(248), // .-, reduce: ContextVariable
			reduce(248), // *, reduce: ContextVariable
			reduce(248), // /, reduce: ContextVariable
			reduce(248), // //, reduce: ContextVariable
			reduce(248), // %, reduce: ContextVariable
			reduce(248), // .*, reduce: ContextVariable
			reduce(248), // ./, reduce: ContextVariable
			reduce(248), // .//, reduce: ContextVariable
			reduce(248), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(248), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(248), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(248), // [[, reduce: ContextVariable
			reduce(248), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1948
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(249), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(249), // >, reduce: ContextVariable
			reduce(249), // >>, reduce: ContextVariable
			reduce(249), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(249), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(249), // ?, reduce: ContextVariable
			nil,         // :
			reduce(249), // ||, reduce: ContextVariable
			reduce(249), // ^^, reduce: ContextVariable
			reduce(249), // &&, reduce: ContextVariable
			reduce(249), // ??, reduce: ContextVariable
			reduce(249), // ???, reduce: ContextVariable
			reduce(249), // =~, reduce: ContextVariable
			reduce(249), // !=~, reduce: ContextVariable
			reduce(249), // ==, reduce: ContextVariable
			reduce(249), // !=, reduce: ContextVariable
			reduce(249), // <=>, reduce: ContextVariable
			reduce(249), // >=, reduce: ContextVariable
			reduce(249), // <, reduce: ContextVariable
			reduce(249), // <=, reduce: ContextVariable
			reduce(249), // ^, reduce: ContextVariable
			reduce(249), // &, reduce: ContextVariable
			reduce(249), // <<, reduce: ContextVariable
			reduce(249), // >>>, reduce: ContextVariable
			reduce(249), // +, reduce: ContextVariable
			reduce(249), // -, reduce: ContextVariable
			reduce(249), // .+, reduce: ContextVariable
			reduce(249), // .-, reduce: ContextVariable
			reduce(249), // *, reduce: ContextVariable
			reduce(249), // /, reduce: ContextVariable
			reduce(249), // //, reduce: ContextVariable
			reduce(249), // %, reduce: ContextVariable
			reduce(249), // .*, reduce: ContextVariable
			reduce(249), // ./, reduce: ContextVariable
			reduce(249), // .//, reduce: ContextVariable
			reduce(249), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(249), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(249), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(249), // [[, reduce: ContextVariable
			reduce(249), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1949
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(250), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(250), // >, reduce: ContextVariable
			reduce(250), // >>, reduce: ContextVariable
			reduce(250), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(250), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(250), // ?, reduce: ContextVariable
			nil,         // :
			reduce(250), // ||, reduce: ContextVariable
			reduce(250), // ^^, reduce: ContextVariable
			reduce(250), // &&, reduce: ContextVariable
			reduce(250), // ??, reduce: ContextVariable
			reduce(250), // ???, reduce: ContextVariable
			reduce(250), // =~, reduce: ContextVariable
			reduce(250), // !=~, reduce: ContextVariable
			reduce(250), // ==, reduce: ContextVariable
			reduce(250), // !=, reduce: ContextVariable
			reduce(250), // <=>, reduce: ContextVariable
			reduce(250), // >=, reduce: ContextVariable
			reduce(250), // <, reduce: ContextVariable
			reduce(250), // <=, reduce: ContextVariable
			reduce(250), // ^, reduce: ContextVariable
			reduce(250), // &, reduce: ContextVariable
			reduce(250), // <<, reduce: ContextVariable
			reduce(250), // >>>, reduce: ContextVariable
			reduce(250), // +, reduce: ContextVariable
			reduce(250), // -, reduce: ContextVariable
			reduce(250), // .+, reduce: ContextVariable
			reduce(250), // .-, reduce: ContextVariable
			reduce(250), // *, reduce: ContextVariable
			reduce(250), // /, reduce: ContextVariable
			reduce(250), // //, reduce: ContextVariable
			reduce(250), // %, reduce: ContextVariable
			reduce(250), // .*, reduce: ContextVariable
			reduce(250), // ./, reduce: ContextVariable
			reduce(250), // .//, reduce: ContextVariable
			reduce(250), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(250), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(250), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(250), // [[, reduce: ContextVariable
			reduce(250), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1950
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(251), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(251), // >, reduce: ContextVariable
			reduce(251), // >>, reduce: ContextVariable
			reduce(251), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(251), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(251), // ?, reduce: ContextVariable
			nil,         // :
			reduce(251), // ||, reduce: ContextVariable
			reduce(251), // ^^, reduce: ContextVariable
			reduce(251), // &&, reduce: ContextVariable
			reduce(251), // ??, reduce: ContextVariable
			reduce(251), // ???, reduce: ContextVariable
			reduce(251), // =~, reduce: ContextVariable
			reduce(251), // !=~, reduce: ContextVariable
			reduce(251), // ==, reduce: ContextVariable
			reduce(251), // !=, reduce: ContextVariable
			reduce(251), // <=>, reduce: ContextVariable
			reduce(251), // >=, reduce: ContextVariable
			reduce(251), // <, reduce: ContextVariable
			reduce(251), // <=, reduce: ContextVariable
			reduce(251), // ^, reduce: ContextVariable
			reduce(251), // &, reduce: ContextVariable
			reduce(251), // <<, reduce: ContextVariable
			reduce(251), // >>>, reduce: ContextVariable
			reduce(251), // +, reduce: ContextVariable
			reduce(251), // -, reduce: ContextVariable
			reduce(251), // .+, reduce: ContextVariable
			reduce(251), // .-, reduce: ContextVariable
			reduce(251), // *, reduce: ContextVariable
			reduce(251), // /, reduce: ContextVariable
			reduce(251), // //, reduce: ContextVariable
			reduce(251), // %, reduce: ContextVariable
			reduce(251), // .*, reduce: ContextVariable
			reduce(251), // ./, reduce: ContextVariable
			reduce(251), // .//, reduce: ContextVariable
			reduce(251), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(251), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(251), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(251), // [[, reduce: ContextVariable
			reduce(251), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1951
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(252), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(252), // >, reduce: ContextVariable
			reduce(252), // >>, reduce: ContextVariable
			reduce(252), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(252), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(252), // ?, reduce: ContextVariable
			nil,         // :
			reduce(252), // ||, reduce: ContextVariable
			reduce(252), // ^^, reduce: ContextVariable
			reduce(252), // &&, reduce: ContextVariable
			reduce(252), // ??, reduce: ContextVariable
			reduce(252), // ???, reduce: ContextVariable
			reduce(252), // =~, reduce: ContextVariable
			reduce(252), // !=~, reduce: ContextVariable
			reduce(252), // ==, reduce: ContextVariable
			reduce(252), // !=, reduce: ContextVariable
			reduce(252), // <=>, reduce: ContextVariable
			reduce(252), // >=, reduce: ContextVariable
			reduce(252), // <, reduce: ContextVariable
			reduce(252), // <=, reduce: ContextVariable
			reduce(252), // ^, reduce: ContextVariable
			reduce(252), // &, reduce: ContextVariable
			reduce(252), // <<, reduce: ContextVariable
			reduce(252), // >>>, reduce: ContextVariable
			reduce(252), // +, reduce: ContextVariable
			reduce(252), // -, reduce: ContextVariable
			reduce(252), // .+, reduce: ContextVariable
			reduce(252), // .-, reduce: ContextVariable
			reduce(252), // *, reduce: ContextVariable
			reduce(252), // /, reduce: ContextVariable
			reduce(252), // //, reduce: ContextVariable
			reduce(252), // %, reduce: ContextVariable
			reduce(252), // .*, reduce: ContextVariable
			reduce(252), // ./, reduce: ContextVariable
			reduce(252), // .//, reduce: ContextVariable
			reduce(252), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(252), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(252), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(252), // [[, reduce: ContextVariable
			reduce(252), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1952
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(253), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(253), // >, reduce: MlrvalOrFunction
			reduce(253), // >>, reduce: MlrvalOrFunction
			reduce(253), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(253), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(253), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(253), // ||, reduce: MlrvalOrFunction
			reduce(253), // ^^, reduce: MlrvalOrFunction
			reduce(253), // &&, reduce: MlrvalOrFunction
			reduce(253), // ??, reduce: MlrvalOrFunction
			reduce(253), // ???, reduce: MlrvalOrFunction
			reduce(253), // =~, reduce: MlrvalOrFunction
			reduce(253), // !=~, reduce: MlrvalOrFunction
			reduce(253), // ==, reduce: MlrvalOrFunction
			reduce(253), // !=, reduce: MlrvalOrFunction
			reduce(253), // <=>, reduce: MlrvalOrFunction
			reduce(253), // >=, reduce: MlrvalOrFunction
			reduce(253), // <, reduce: MlrvalOrFunction
			reduce(253), // <=, reduce: MlrvalOrFunction
			reduce(253), // ^, reduce: MlrvalOrFunction
			reduce(253), // &, reduce: MlrvalOrFunction
			reduce(253), // <<, reduce: MlrvalOrFunction
			reduce(253), // >>>, reduce: MlrvalOrFunction
			reduce(253), // +, reduce: MlrvalOrFunction
			reduce(253), // -, reduce: MlrvalOrFunction
			reduce(253), // .+, reduce: MlrvalOrFunction
			reduce(253), // .-, reduce: MlrvalOrFunction
			reduce(253), // *, reduce: MlrvalOrFunction
			reduce(253), // /, reduce: MlrvalOrFunction
			reduce(253), // //, reduce: MlrvalOrFunction
			reduce(253), // %, reduce: MlrvalOrFunction
			reduce(253), // .*, reduce: MlrvalOrFunction
			reduce(253), // ./, reduce: MlrvalOrFunction
			reduce(253), // .//, reduce: MlrvalOrFunction
			reduce(253), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(253), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(253), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(253), // [[, reduce: MlrvalOrFunction
			reduce(253), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1953
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			shift(2957), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(2958), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1954
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(256), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(256), // >, reduce: MlrvalOrFunction
			reduce(256), // >>, reduce: MlrvalOrFunction
			reduce(256), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(256), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(256), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(256), // ||, reduce: MlrvalOrFunction
			reduce(256), // ^^, reduce: MlrvalOrFunction
			reduce(256), // &&, reduce: MlrvalOrFunction
			reduce(256), // ??, reduce: MlrvalOrFunction
			reduce(256), // ???, reduce: MlrvalOrFunction
			reduce(256), // =~, reduce: MlrvalOrFunction
			reduce(256), // !=~, reduce: MlrvalOrFunction
			reduce(256), // ==, reduce: MlrvalOrFunction
			reduce(256), // !=, reduce: MlrvalOrFunction
			reduce(256), // <=>, reduce: MlrvalOrFunction
			reduce(256), // >=, reduce: MlrvalOrFunction
			reduce(256), // <, reduce: MlrvalOrFunction
			reduce(256), // <=, reduce: MlrvalOrFunction
			reduce(256), // ^, reduce: MlrvalOrFunction
			reduce(256), // &, reduce: MlrvalOrFunction
			reduce(256), // <<, reduce: MlrvalOrFunction
			reduce(256), // >>>, reduce: MlrvalOrFunction
			reduce(256), // +, reduce: MlrvalOrFunction
			reduce(256), // -, reduce: MlrvalOrFunction
			reduce(256), // .+, reduce: MlrvalOrFunction
			reduce(256), // .-, reduce: MlrvalOrFunction
			reduce(256), // *, reduce: MlrvalOrFunction
			reduce(256), // /, reduce: MlrvalOrFunction
			reduce(256), // //, reduce: MlrvalOrFunction
			reduce(256), // %, reduce: MlrvalOrFunction
			reduce(256), // .*, reduce: MlrvalOrFunction
			reduce(256), // ./, reduce: MlrvalOrFunction
			reduce(256), // .//, reduce: MlrvalOrFunction
			reduce(256), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(256), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(256), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(256), // [[, reduce: MlrvalOrFunction
			reduce(256), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1955
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(257), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(257), // >, reduce: MlrvalOrFunction
			reduce(257), // >>, reduce: MlrvalOrFunction
			reduce(257), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(257), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(257), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(257), // ||, reduce: MlrvalOrFunction
			reduce(257), // ^^, reduce: MlrvalOrFunction
			reduce(257), // &&, reduce: MlrvalOrFunction
			reduce(257), // ??, reduce: MlrvalOrFunction
			reduce(257), // ???, reduce: MlrvalOrFunction
			reduce(257), // =~, reduce: MlrvalOrFunction
			reduce(257), // !=~, reduce: MlrvalOrFunction
			reduce(257), // ==, reduce: MlrvalOrFunction
			reduce(257), // !=, reduce: MlrvalOrFunction
			reduce(257), // <=>, reduce: MlrvalOrFunction
			reduce(257), // >=, reduce: MlrvalOrFunction
			reduce(257), // <, reduce: MlrvalOrFunction
			reduce(257), // <=, reduce: MlrvalOrFunction
			reduce(257), // ^, reduce: MlrvalOrFunction
			reduce(257), // &, reduce: MlrvalOrFunction
			reduce(257), // <<, reduce: MlrvalOrFunction
			reduce(257), // >>>, reduce: MlrvalOrFunction
			reduce(257), // +, reduce: MlrvalOrFunction
			reduce(257), // -, reduce: MlrvalOrFunction
			reduce(257), // .+, reduce: MlrvalOrFunction
			reduce(257), // .-, reduce: MlrvalOrFunction
			reduce(257), // *, reduce: MlrvalOrFunction
			reduce(257), // /, reduce: MlrvalOrFunction
			reduce(257), // //, reduce: MlrvalOrFunction
			reduce(257), // %, reduce: MlrvalOrFunction
			reduce(257), // .*, reduce: MlrvalOrFunction
			reduce(257), // ./, reduce: MlrvalOrFunction
			reduce(257), // .//, reduce: MlrvalOrFunction
			reduce(257), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(257), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(257), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(257), // [[, reduce: MlrvalOrFunction
			reduce(257), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1956
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(258), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(258), // >, reduce: MlrvalOrFunction
			reduce(258), // >>, reduce: MlrvalOrFunction
			reduce(258), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(258), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(258), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(258), // ||, reduce: MlrvalOrFunction
			reduce(258), // ^^, reduce: MlrvalOrFunction
			reduce(258), // &&, reduce: MlrvalOrFunction
			reduce(258), // ??, reduce: MlrvalOrFunction
			reduce(258), // ???, reduce: MlrvalOrFunction
			reduce(258), // =~, reduce: MlrvalOrFunction
			reduce(258), // !=~, reduce: MlrvalOrFunction
			reduce(258), // ==, reduce: MlrvalOrFunction
			reduce(258), // !=, reduce: MlrvalOrFunction
			reduce(258), // <=>, reduce: MlrvalOrFunction
			reduce(258), // >=, reduce: MlrvalOrFunction
			reduce(258), // <, reduce: MlrvalOrFunction
			reduce(258), // <=, reduce: MlrvalOrFunction
			reduce(258), // ^, reduce: MlrvalOrFunction
			reduce(258), // &, reduce: MlrvalOrFunction
			reduce(258), // <<, reduce: MlrvalOrFunction
			reduce(258), // >>>, reduce: MlrvalOrFunction
			reduce(258), // +, reduce: MlrvalOrFunction
			reduce(258), // -, reduce: MlrvalOrFunction
			reduce(258), // .+, reduce: MlrvalOrFunction
			reduce(258), // .-, reduce: MlrvalOrFunction
			reduce(258), // *, reduce: MlrvalOrFunction
			reduce(258), // /, reduce: MlrvalOrFunction
			reduce(258), // //, reduce: MlrvalOrFunction
			reduce(258), // %, reduce: MlrvalOrFunction
			reduce(258), // .*, reduce: MlrvalOrFunction
			reduce(258), // ./, reduce: MlrvalOrFunction
			reduce(258), // .//, reduce: MlrvalOrFunction
			reduce(258), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(258), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(258), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(258), // [[, reduce: MlrvalOrFunction
			reduce(258), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1957
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(259), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(259), // >, reduce: MlrvalOrFunction
			reduce(259), // >>, reduce: MlrvalOrFunction
			reduce(259), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(259), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(259), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(259), // ||, reduce: MlrvalOrFunction
			reduce(259), // ^^, reduce: MlrvalOrFunction
			reduce(259), // &&, reduce: MlrvalOrFunction
			reduce(259), // ??, reduce: MlrvalOrFunction
			reduce(259), // ???, reduce: MlrvalOrFunction
			reduce(259), // =~, reduce: MlrvalOrFunction
			reduce(259), // !=~, reduce: MlrvalOrFunction
			reduce(259), // ==, reduce: MlrvalOrFunction
			reduce(259), // !=, reduce: MlrvalOrFunction
			reduce(259), // <=>, reduce: MlrvalOrFunction
			reduce(259), // >=, reduce: MlrvalOrFunction
			reduce(259), // <, reduce: MlrvalOrFunction
			reduce(259), // <=, reduce: MlrvalOrFunction
			reduce(259), // ^, reduce: MlrvalOrFunction
			reduce(259), // &, reduce: MlrvalOrFunction
			reduce(259), // <<, reduce: MlrvalOrFunction
			reduce(259), // >>>, reduce: MlrvalOrFunction
			reduce(259), // +, reduce: MlrvalOrFunction
			reduce(259), // -, reduce: MlrvalOrFunction
			reduce(259), // .+, reduce: MlrvalOrFunction
			reduce(259), // .-, reduce: MlrvalOrFunction
			reduce(259), // *, reduce: MlrvalOrFunction
			reduce(259), // /, reduce: MlrvalOrFunction
			reduce(259), // //, reduce: MlrvalOrFunction
			reduce(259), // %, reduce: MlrvalOrFunction
			reduce(259), // .*, reduce: MlrvalOrFunction
			reduce(259), // ./, reduce: MlrvalOrFunction
			reduce(259), // .//, reduce: MlrvalOrFunction
			reduce(259), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(259), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(259), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(259), // [[, reduce: MlrvalOrFunction
			reduce(259), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1958
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(267), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(267), // >, reduce: MlrvalOrFunction
			reduce(267), // >>, reduce: MlrvalOrFunction
			reduce(267), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(267), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(267), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(267), // ||, reduce: MlrvalOrFunction
			reduce(267), // ^^, reduce: MlrvalOrFunction
			reduce(267), // &&, reduce: MlrvalOrFunction
			reduce(267), // ??, reduce: MlrvalOrFunction
			reduce(267), // ???, reduce: MlrvalOrFunction
			reduce(267), // =~, reduce: MlrvalOrFunction
			reduce(267), // !=~, reduce: MlrvalOrFunction
			reduce(267), // ==, reduce: MlrvalOrFunction
			reduce(267), // !=, reduce: MlrvalOrFunction
			reduce(267), // <=>, reduce: MlrvalOrFunction
			reduce(267), // >=, reduce: MlrvalOrFunction
			reduce(267), // <, reduce: MlrvalOrFunction
			reduce(267), // <=, reduce: MlrvalOrFunction
			reduce(267), // ^, reduce: MlrvalOrFunction
			reduce(267), // &, reduce: MlrvalOrFunction
			reduce(267), // <<, reduce: MlrvalOrFunction
			reduce(267), // >>>, reduce: MlrvalOrFunction
			reduce(267), // +, reduce: MlrvalOrFunction
			reduce(267), // -, reduce: MlrvalOrFunction
			reduce(267), // .+, reduce: MlrvalOrFunction
			reduce(267), // .-, reduce: MlrvalOrFunction
			reduce(267), // *, reduce: MlrvalOrFunction
			reduce(267), // /, reduce: MlrvalOrFunction
			reduce(267), // //, reduce: MlrvalOrFunction
			reduce(267), // %, reduce: MlrvalOrFunction
			reduce(267), // .*, reduce: MlrvalOrFunction
			reduce(267), // ./, reduce: MlrvalOrFunction
			reduce(267), // .//, reduce: MlrvalOrFunction
			reduce(267), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(267), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(267), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(267), // [[, reduce: MlrvalOrFunction
			reduce(267), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1959
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2959), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1960
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2960), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1961
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(208), // >, reduce: PrecedenceChainEnd
			reduce(208), // >>, reduce: PrecedenceChainEnd
			reduce(208), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(208), // ?, reduce: PrecedenceChainEnd
			reduce(208), // :, reduce: PrecedenceChainEnd
			reduce(208), // ||, reduce: PrecedenceChainEnd
			reduce(208), // ^^, reduce: PrecedenceChainEnd
			reduce(208), // &&, reduce: PrecedenceChainEnd
			reduce(208), // ??, reduce: PrecedenceChainEnd
			reduce(208), // ???, reduce: PrecedenceChainEnd
			reduce(208), // =~, reduce: PrecedenceChainEnd
			reduce(208), // !=~, reduce: PrecedenceChainEnd
			reduce(208), // ==, reduce: PrecedenceChainEnd
			reduce(208), // !=, reduce: PrecedenceChainEnd
			reduce(208), // <=>, reduce: PrecedenceChainEnd
			reduce(208), // >=, reduce: PrecedenceChainEnd
			reduce(208), // <, reduce: PrecedenceChainEnd
			reduce(208), // <=, reduce: PrecedenceChainEnd
			reduce(208), // ^, reduce: PrecedenceChainEnd
			reduce(208), // &, reduce: PrecedenceChainEnd
			reduce(208), // <<, reduce: PrecedenceChainEnd
			reduce(208), // >>>, reduce: PrecedenceChainEnd
			reduce(208), // +, reduce: PrecedenceChainEnd
			reduce(208), // -, reduce: PrecedenceChainEnd
			reduce(208), // .+, reduce: PrecedenceChainEnd
			reduce(208), // .-, reduce: PrecedenceChainEnd
			reduce(208), // *, reduce: PrecedenceChainEnd
			reduce(208), // /, reduce: PrecedenceChainEnd
			reduce(208), // //, reduce: PrecedenceChainEnd
			reduce(208), // %, reduce: PrecedenceChainEnd
			reduce(208), // .*, reduce: PrecedenceChainEnd
			reduce(208), // ./, reduce: PrecedenceChainEnd
			reduce(208), // .//, reduce: PrecedenceChainEnd
			reduce(208), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(208), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1962
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(105), // >, reduce: IndirectFieldValue
			reduce(105), // >>, reduce: IndirectFieldValue
			reduce(105), // |, reduce: IndirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(105), // ?, reduce: IndirectFieldValue
			reduce(105), // :, reduce: IndirectFieldValue
			reduce(105), // ||, reduce: IndirectFieldValue
			reduce(105), // ^^, reduce: IndirectFieldValue
			reduce(105), // &&, reduce: IndirectFieldValue
			reduce(105), // ??, reduce: IndirectFieldValue
			reduce(105), // ???, reduce: IndirectFieldValue
			reduce(105), // =~, reduce: IndirectFieldValue
			reduce(105), // !=~, reduce: IndirectFieldValue
			reduce(105), // ==, reduce: IndirectFieldValue
			reduce(105), // !=, reduce: IndirectFieldValue
			reduce(105), // <=>, reduce: IndirectFieldValue
			reduce(105), // >=, reduce: IndirectFieldValue
			reduce(105), // <, reduce: IndirectFieldValue
			reduce(105), // <=, reduce: IndirectFieldValue
			reduce(105), // ^, reduce: IndirectFieldValue
			reduce(105), // &, reduce: IndirectFieldValue
			reduce(105), // <<, reduce: IndirectFieldValue
			reduce(105), // >>>, reduce: IndirectFieldValue
			reduce(105), // +, reduce: IndirectFieldValue
			reduce(105), // -, reduce: IndirectFieldValue
			reduce(105), // .+, reduce: IndirectFieldValue
			reduce(105), // .-, reduce: IndirectFieldValue
			reduce(105), // *, reduce: IndirectFieldValue
			reduce(105), // /, reduce: IndirectFieldValue
			reduce(105), // //, reduce: IndirectFieldValue
			reduce(105), // %, reduce: IndirectFieldValue
			reduce(105), // .*, reduce: IndirectFieldValue
			reduce(105), // ./, reduce: IndirectFieldValue
			reduce(105), // .//, reduce: IndirectFieldValue
			reduce(105), // ., reduce: IndirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(105), // **, reduce: IndirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(105), // [, reduce: IndirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(105), // [[, reduce: IndirectFieldValue
			reduce(105), // [[[, reduce: IndirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1963
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2961), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1964
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2962), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1965
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(114), // >, reduce: IndirectOosvarValue
			reduce(114), // >>, reduce: IndirectOosvarValue
			reduce(114), // |, reduce: IndirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(114), // ?, reduce: IndirectOosvarValue
			reduce(114), // :, reduce: IndirectOosvarValue
			reduce(114), // ||, reduce: IndirectOosvarValue
			reduce(114), // ^^, reduce: IndirectOosvarValue
			reduce(114), // &&, reduce: IndirectOosvarValue
			reduce(114), // ??, reduce: IndirectOosvarValue
			reduce(114), // ???, reduce: IndirectOosvarValue
			reduce(114), // =~, reduce: IndirectOosvarValue
			reduce(114), // !=~, reduce: IndirectOosvarValue
			reduce(114), // ==, reduce: IndirectOosvarValue
			reduce(114), // !=, reduce: IndirectOosvarValue
			reduce(114), // <=>, reduce: IndirectOosvarValue
			reduce(114), // >=, reduce: IndirectOosvarValue
			reduce(114), // <, reduce: IndirectOosvarValue
			reduce(114), // <=, reduce: IndirectOosvarValue
			reduce(114), // ^, reduce: IndirectOosvarValue
			reduce(114), // &, reduce: IndirectOosvarValue
			reduce(114), // <<, reduce: IndirectOosvarValue
			reduce(114), // >>>, reduce: IndirectOosvarValue
			reduce(114), // +, reduce: IndirectOosvarValue
			reduce(114), // -, reduce: IndirectOosvarValue
			reduce(114), // .+, reduce: IndirectOosvarValue
			reduce(114), // .-, reduce: IndirectOosvarValue
			reduce(114), // *, reduce: IndirectOosvarValue
			reduce(114), // /, reduce: IndirectOosvarValue
			reduce(114), // //, reduce: IndirectOosvarValue
			reduce(114), // %, reduce: IndirectOosvarValue
			reduce(114), // .*, reduce: IndirectOosvarValue
			reduce(114), // ./, reduce: IndirectOosvarValue
			reduce(114), // .//, reduce: IndirectOosvarValue
			reduce(114), // ., reduce: IndirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(114), // **, reduce: IndirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(114), // [, reduce: IndirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(114), // [[, reduce: IndirectOosvarValue
			reduce(114), // [[[, reduce: IndirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1966
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(2963), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1967
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(151), // ?, reduce: LogicalOrTerm
			reduce(151), // :, reduce: LogicalOrTerm
			reduce(151), // ||, reduce: LogicalOrTerm
			shift(879),  // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1968
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(153), // ?, reduce: LogicalXORTerm
			reduce(153), // :, reduce: LogicalXORTerm
			reduce(153), // ||, reduce: LogicalXORTerm
			reduce(153), // ^^, reduce: LogicalXORTerm
			shift(880),  // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1969
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(155), // ?, reduce: LogicalAndTerm
			reduce(155), // :, reduce: LogicalAndTerm
			reduce(155), // ||, reduce: LogicalAndTerm
			reduce(155), // ^^, reduce: LogicalAndTerm
			reduce(155), // &&, reduce: LogicalAndTerm
			shift(881),  // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1970
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(157), // ?, reduce: AbsentCoalesceTerm
			reduce(157), // :, reduce: AbsentCoalesceTerm
			reduce(157), // ||, reduce: AbsentCoalesceTerm
			reduce(157), // ^^, reduce: AbsentCoalesceTerm
			reduce(157), // &&, reduce: AbsentCoalesceTerm
			reduce(157), // ??, reduce: AbsentCoalesceTerm
			shift(882),  // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1971
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(159), // ?, reduce: EmptyCoalesceTerm
			reduce(159), // :, reduce: EmptyCoalesceTerm
			reduce(159), // ||, reduce: EmptyCoalesceTerm
			reduce(159), // ^^, reduce: EmptyCoalesceTerm
			reduce(159), // &&, reduce: EmptyCoalesceTerm
			reduce(159), // ??, reduce: EmptyCoalesceTerm
			reduce(159), // ???, reduce: EmptyCoalesceTerm
			shift(883),  // =~
			shift(884),  // !=~
			shift(885),  // ==
			shift(886),  // !=
			shift(887),  // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1972
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(888),  // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(161), // ?, reduce: EqneTerm
			reduce(161), // :, reduce: EqneTerm
			reduce(161), // ||, reduce: EqneTerm
			reduce(161), // ^^, reduce: EqneTerm
			reduce(161), // &&, reduce: EqneTerm
			reduce(161), // ??, reduce: EqneTerm
			reduce(161), // ???, reduce: EqneTerm
			reduce(161), // =~, reduce: EqneTerm
			reduce(161), // !=~, reduce: EqneTerm
			reduce(161), // ==, reduce: EqneTerm
			reduce(161), // !=, reduce: EqneTerm
			reduce(161), // <=>, reduce: EqneTerm
			shift(889),  // >=
			shift(890),  // <
			shift(891),  // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1973
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(888),  // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(162), // ?, reduce: EqneTerm
			reduce(162), // :, reduce: EqneTerm
			reduce(162), // ||, reduce: EqneTerm
			reduce(162), // ^^, reduce: EqneTerm
			reduce(162), // &&, reduce: EqneTerm
			reduce(162), // ??, reduce: EqneTerm
			reduce(162), // ???, reduce: EqneTerm
			reduce(162), // =~, reduce: EqneTerm
			reduce(162), // !=~, reduce: EqneTerm
			reduce(162), // ==, reduce: EqneTerm
			reduce(162), // !=, reduce: EqneTerm
			reduce(162), // <=>, reduce: EqneTerm
			shift(889),  // >=
			shift(890),  // <
			shift(891),  // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1974
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(888),  // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(163), // ?, reduce: EqneTerm
			reduce(163), // :, reduce: EqneTerm
			reduce(163), // ||, reduce: EqneTerm
			reduce(163), // ^^, reduce: EqneTerm
			reduce(163), // &&, reduce: EqneTerm
			reduce(163), // ??, reduce: EqneTerm
			reduce(163), // ???, reduce: EqneTerm
			reduce(163), // =~, reduce: EqneTerm
			reduce(163), // !=~, reduce: EqneTerm
			reduce(163), // ==, reduce: EqneTerm
			reduce(163), // !=, reduce: EqneTerm
			reduce(163), // <=>, reduce: EqneTerm
			shift(889),  // >=
			shift(890),  // <
			shift(891),  // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1975
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(888),  // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(164), // ?, reduce: EqneTerm
			reduce(164), // :, reduce: EqneTerm
			reduce(164), // ||, reduce: EqneTerm
			reduce(164), // ^^, reduce: EqneTerm
			reduce(164), // &&, reduce: EqneTerm
			reduce(164), // ??, reduce: EqneTerm
			reduce(164), // ???, reduce: EqneTerm
			reduce(164), // =~, reduce: EqneTerm
			reduce(164), // !=~, reduce: EqneTerm
			reduce(164), // ==, reduce: EqneTerm
			reduce(164), // !=, reduce: EqneTerm
			reduce(164), // <=>, reduce: EqneTerm
			shift(889),  // >=
			shift(890),  // <
			shift(891),  // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1976
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(888),  // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(165), // ?, reduce: EqneTerm
			reduce(165), // :, reduce: EqneTerm
			reduce(165), // ||, reduce: EqneTerm
			reduce(165), // ^^, reduce: EqneTerm
			reduce(165), // &&, reduce: EqneTerm
			reduce(165), // ??, reduce: EqneTerm
			reduce(165), // ???, reduce: EqneTerm
			reduce(165), // =~, reduce: EqneTerm
			reduce(165), // !=~, reduce: EqneTerm
			reduce(165), // ==, reduce: EqneTerm
			reduce(165), // !=, reduce: EqneTerm
			reduce(165), // <=>, reduce: EqneTerm
			shift(889),  // >=
			shift(890),  // <
			shift(891),  // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1977
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(167), // >, reduce: CmpTerm
			nil,         // >>
			shift(892),  // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(167), // ?, reduce: CmpTerm
			reduce(167), // :, reduce: CmpTerm
			reduce(167), // ||, reduce: CmpTerm
			reduce(167), // ^^, reduce: CmpTerm
			reduce(167), // &&, reduce: CmpTerm
			reduce(167), // ??, reduce: CmpTerm
			reduce(167), // ???, reduce: CmpTerm
			reduce(167), // =~, reduce: CmpTerm
			reduce(167), // !=~, reduce: CmpTerm
			reduce(167), // ==, reduce: CmpTerm
			reduce(167), // !=, reduce: CmpTerm
			reduce(167), // <=>, reduce: CmpTerm
			reduce(167), // >=, reduce: CmpTerm
			reduce(167), // <, reduce: CmpTerm
			reduce(167), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1978
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(168), // >, reduce: CmpTerm
			nil,         // >>
			shift(892),  // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(168), // ?, reduce: CmpTerm
			reduce(168), // :, reduce: CmpTerm
			reduce(168), // ||, reduce: CmpTerm
			reduce(168), // ^^, reduce: CmpTerm
			reduce(168), // &&, reduce: CmpTerm
			reduce(168), // ??, reduce: CmpTerm
			reduce(168), // ???, reduce: CmpTerm
			reduce(168), // =~, reduce: CmpTerm
			reduce(168), // !=~, reduce: CmpTerm
			reduce(168), // ==, reduce: CmpTerm
			reduce(168), // !=, reduce: CmpTerm
			reduce(168), // <=>, reduce: CmpTerm
			reduce(168), // >=, reduce: CmpTerm
			reduce(168), // <, reduce: CmpTerm
			reduce(168), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1979
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(169), // >, reduce: CmpTerm
			nil,         // >>
			shift(892),  // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(169), // ?, reduce: CmpTerm
			reduce(169), // :, reduce: CmpTerm
			reduce(169), // ||, reduce: CmpTerm
			reduce(169), // ^^, reduce: CmpTerm
			reduce(169), // &&, reduce: CmpTerm
			reduce(169), // ??, reduce: CmpTerm
			reduce(169), // ???, reduce: CmpTerm
			reduce(169), // =~, reduce: CmpTerm
			reduce(169), // !=~, reduce: CmpTerm
			reduce(169), // ==, reduce: CmpTerm
			reduce(169), // !=, reduce: CmpTerm
			reduce(169), // <=>, reduce: CmpTerm
			reduce(169), // >=, reduce: CmpTerm
			reduce(169), // <, reduce: CmpTerm
			reduce(169), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1980
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(170), // >, reduce: CmpTerm
			nil,         // >>
			shift(892),  // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(170), // ?, reduce: CmpTerm
			reduce(170), // :, reduce: CmpTerm
			reduce(170), // ||, reduce: CmpTerm
			reduce(170), // ^^, reduce: CmpTerm
			reduce(170), // &&, reduce: CmpTerm
			reduce(170), // ??, reduce: CmpTerm
			reduce(170), // ???, reduce: CmpTerm
			reduce(170), // =~, reduce: CmpTerm
			reduce(170), // !=~, reduce: CmpTerm
			reduce(170), // ==, reduce: CmpTerm
			reduce(170), // !=, reduce: CmpTerm
			reduce(170), // <=>, reduce: CmpTerm
			reduce(170), // >=, reduce: CmpTerm
			reduce(170), // <, reduce: CmpTerm
			reduce(170), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1981
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(172), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(172), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(172), // ?, reduce: BitwiseORTerm
			reduce(172), // :, reduce: BitwiseORTerm
			reduce(172), // ||, reduce: BitwiseORTerm
			reduce(172), // ^^, reduce: BitwiseORTerm
			reduce(172), // &&, reduce: BitwiseORTerm
			reduce(172), // ??, reduce: BitwiseORTerm
			reduce(172), // ???, reduce: BitwiseORTerm
			reduce(172), // =~, reduce: BitwiseORTerm
			reduce(172), // !=~, reduce: BitwiseORTerm
			reduce(172), // ==, reduce: BitwiseORTerm
			reduce(172), // !=, reduce: BitwiseORTerm
			reduce(172), // <=>, reduce: BitwiseORTerm
			reduce(172), // >=, reduce: BitwiseORTerm
			reduce(172), // <, reduce: BitwiseORTerm
			reduce(172), // <=, reduce: BitwiseORTerm
			shift(893),  // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1982
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(174), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(174), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(174), // ?, reduce: BitwiseXORTerm
			reduce(174), // :, reduce: BitwiseXORTerm
			reduce(174), // ||, reduce: BitwiseXORTerm
			reduce(174), // ^^, reduce: BitwiseXORTerm
			reduce(174), // &&, reduce: BitwiseXORTerm
			reduce(174), // ??, reduce: BitwiseXORTerm
			reduce(174), // ???, reduce: BitwiseXORTerm
			reduce(174), // =~, reduce: BitwiseXORTerm
			reduce(174), // !=~, reduce: BitwiseXORTerm
			reduce(174), // ==, reduce: BitwiseXORTerm
			reduce(174), // !=, reduce: BitwiseXORTerm
			reduce(174), // <=>, reduce: BitwiseXORTerm
			reduce(174), // >=, reduce: BitwiseXORTerm
			reduce(174), // <, reduce: BitwiseXORTerm
			reduce(174), // <=, reduce: BitwiseXORTerm
			reduce(174), // ^, reduce: BitwiseXORTerm
			shift(894),  // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1983
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(176), // >, reduce: BitwiseANDTerm
			shift(895),  // >>
			reduce(176), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(176), // ?, reduce: BitwiseANDTerm
			reduce(176), // :, reduce: BitwiseANDTerm
			reduce(176), // ||, reduce: BitwiseANDTerm
			reduce(176), // ^^, reduce: BitwiseANDTerm
			reduce(176), // &&, reduce: BitwiseANDTerm
			reduce(176), // ??, reduce: BitwiseANDTerm
			reduce(176), // ???, reduce: BitwiseANDTerm
			reduce(176), // =~, reduce: BitwiseANDTerm
			reduce(176), // !=~, reduce: BitwiseANDTerm
			reduce(176), // ==, reduce: BitwiseANDTerm
			reduce(176), // !=, reduce: BitwiseANDTerm
			reduce(176), // <=>, reduce: BitwiseANDTerm
			reduce(176), // >=, reduce: BitwiseANDTerm
			reduce(176), // <, reduce: BitwiseANDTerm
			reduce(176), // <=, reduce: BitwiseANDTerm
			reduce(176), // ^, reduce: BitwiseANDTerm
			reduce(176), // &, reduce: BitwiseANDTerm
			shift(896),  // <<
			shift(897),  // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1984
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(179), // >, reduce: BitwiseShiftTerm
			reduce(179), // >>, reduce: BitwiseShiftTerm
			reduce(179), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(179), // ?, reduce: BitwiseShiftTerm
			reduce(179), // :, reduce: BitwiseShiftTerm
			reduce(179), // ||, reduce: BitwiseShiftTerm
			reduce(179), // ^^, reduce: BitwiseShiftTerm
			reduce(179), // &&, reduce: BitwiseShiftTerm
			reduce(179), // ??, reduce: BitwiseShiftTerm
			reduce(179), // ???, reduce: BitwiseShiftTerm
			reduce(179), // =~, reduce: BitwiseShiftTerm
			reduce(179), // !=~, reduce: BitwiseShiftTerm
			reduce(179), // ==, reduce: BitwiseShiftTerm
			reduce(179), // !=, reduce: BitwiseShiftTerm
			reduce(179), // <=>, reduce: BitwiseShiftTerm
			reduce(179), // >=, reduce: BitwiseShiftTerm
			reduce(179), // <, reduce: BitwiseShiftTerm
			reduce(179), // <=, reduce: BitwiseShiftTerm
			reduce(179), // ^, reduce: BitwiseShiftTerm
			reduce(179), // &, reduce: BitwiseShiftTerm
			reduce(179), // <<, reduce: BitwiseShiftTerm
			reduce(179), // >>>, reduce: BitwiseShiftTerm
			shift(898),  // +
			shift(899),  // -
			shift(900),  // .+
			shift(901),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1985
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(178), // >, reduce: BitwiseShiftTerm
			reduce(178), // >>, reduce: BitwiseShiftTerm
			reduce(178), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(178), // ?, reduce: BitwiseShiftTerm
			reduce(178), // :, reduce: BitwiseShiftTerm
			reduce(178), // ||, reduce: BitwiseShiftTerm
			reduce(178), // ^^, reduce: BitwiseShiftTerm
			reduce(178), // &&, reduce: BitwiseShiftTerm
			reduce(178), // ??, reduce: BitwiseShiftTerm
			reduce(178), // ???, reduce: BitwiseShiftTerm
			reduce(178), // =~, reduce: BitwiseShiftTerm
			reduce(178), // !=~, reduce: BitwiseShiftTerm
			reduce(178), // ==, reduce: BitwiseShiftTerm
			reduce(178), // !=, reduce: BitwiseShiftTerm
			reduce(178), // <=>, reduce: BitwiseShiftTerm
			reduce(178), // >=, reduce: BitwiseShiftTerm
			reduce(178), // <, reduce: BitwiseShiftTerm
			reduce(178), // <=, reduce: BitwiseShiftTerm
			reduce(178), // ^, reduce: BitwiseShiftTerm
			reduce(178), // &, reduce: BitwiseShiftTerm
			reduce(178), // <<, reduce: BitwiseShiftTerm
			reduce(178), // >>>, reduce: BitwiseShiftTerm
			shift(898),  // +
			shift(899),  // -
			shift(900),  // .+
			shift(901),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1986
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(180), // >, reduce: BitwiseShiftTerm
			reduce(180), // >>, reduce: BitwiseShiftTerm
			reduce(180), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(180), // ?, reduce: BitwiseShiftTerm
			reduce(180), // :, reduce: BitwiseShiftTerm
			reduce(180), // ||, reduce: BitwiseShiftTerm
			reduce(180), // ^^, reduce: BitwiseShiftTerm
			reduce(180), // &&, reduce: BitwiseShiftTerm
			reduce(180), // ??, reduce: BitwiseShiftTerm
			reduce(180), // ???, reduce: BitwiseShiftTerm
			reduce(180), // =~, reduce: BitwiseShiftTerm
			reduce(180), // !=~, reduce: BitwiseShiftTerm
			reduce(180), // ==, reduce: BitwiseShiftTerm
			reduce(180), // !=, reduce: BitwiseShiftTerm
			reduce(180), // <=>, reduce: BitwiseShiftTerm
			reduce(180), // >=, reduce: BitwiseShiftTerm
			reduce(180), // <, reduce: BitwiseShiftTerm
			reduce(180), // <=, reduce: BitwiseShiftTerm
			reduce(180), // ^, reduce: BitwiseShiftTerm
			reduce(180), // &, reduce: BitwiseShiftTerm
			reduce(180), // <<, reduce: BitwiseShiftTerm
			reduce(180), // >>>, reduce: BitwiseShiftTerm
			shift(898),  // +
			shift(899),  // -
			shift(900),  // .+
			shift(901),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1987
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(182), // >, reduce: AddsubdotTerm
			reduce(182), // >>, reduce: AddsubdotTerm
			reduce(182), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(182), // ?, reduce: AddsubdotTerm
			reduce(182), // :, reduce: AddsubdotTerm
			reduce(182), // ||, reduce: AddsubdotTerm
			reduce(182), // ^^, reduce: AddsubdotTerm
			reduce(182), // &&, reduce: AddsubdotTerm
			reduce(182), // ??, reduce: AddsubdotTerm
			reduce(182), // ???, reduce: AddsubdotTerm
			reduce(182), // =~, reduce: AddsubdotTerm
			reduce(182), // !=~, reduce: AddsubdotTerm
			reduce(182), // ==, reduce: AddsubdotTerm
			reduce(182), // !=, reduce: AddsubdotTerm
			reduce(182), // <=>, reduce: AddsubdotTerm
			reduce(182), // >=, reduce: AddsubdotTerm
			reduce(182), // <, reduce: AddsubdotTerm
			reduce(182), // <=, reduce: AddsubdotTerm
			reduce(182), // ^, reduce: AddsubdotTerm
			reduce(182), // &, reduce: AddsubdotTerm
			reduce(182), // <<, reduce: AddsubdotTerm
			reduce(182), // >>>, reduce: AddsubdotTerm
			reduce(182), // +, reduce: AddsubdotTerm
			reduce(182), // -, reduce: AddsubdotTerm
			reduce(182), // .+, reduce: AddsubdotTerm
			reduce(182), // .-, reduce: AddsubdotTerm
			shift(903),  // *
			shift(904),  // /
			shift(905),  // //
			shift(906),  // %
			shift(907),  // .*
			shift(908),  // ./
			shift(909),  // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1988
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(183), // >, reduce: AddsubdotTerm
			reduce(183), // >>, reduce: AddsubdotTerm
			reduce(183), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(183), // ?, reduce: AddsubdotTerm
			reduce(183), // :, reduce: AddsubdotTerm
			reduce(183), // ||, reduce: AddsubdotTerm
			reduce(183), // ^^, reduce: AddsubdotTerm
			reduce(183), // &&, reduce: AddsubdotTerm
			reduce(183), // ??, reduce: AddsubdotTerm
			reduce(183), // ???, reduce: AddsubdotTerm
			reduce(183), // =~, reduce: AddsubdotTerm
			reduce(183), // !=~, reduce: AddsubdotTerm
			reduce(183), // ==, reduce: AddsubdotTerm
			reduce(183), // !=, reduce: AddsubdotTerm
			reduce(183), // <=>, reduce: AddsubdotTerm
			reduce(183), // >=, reduce: AddsubdotTerm
			reduce(183), // <, reduce: AddsubdotTerm
			reduce(183), // <=, reduce: AddsubdotTerm
			reduce(183), // ^, reduce: AddsubdotTerm
			reduce(183), // &, reduce: AddsubdotTerm
			reduce(183), // <<, reduce: AddsubdotTerm
			reduce(183), // >>>, reduce: AddsubdotTerm
			reduce(183), // +, reduce: AddsubdotTerm
			reduce(183), // -, reduce: AddsubdotTerm
			reduce(183), // .+, reduce: AddsubdotTerm
			reduce(183), // .-, reduce: AddsubdotTerm
			shift(903),  // *
			shift(904),  // /
			shift(905),  // //
			shift(906),  // %
			shift(907),  // .*
			shift(908),  // ./
			shift(909),  // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1989
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(184), // >, reduce: AddsubdotTerm
			reduce(184), // >>, reduce: AddsubdotTerm
			reduce(184), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(184), // ?, reduce: AddsubdotTerm
			reduce(184), // :, reduce: AddsubdotTerm
			reduce(184), // ||, reduce: AddsubdotTerm
			reduce(184), // ^^, reduce: AddsubdotTerm
			reduce(184), // &&, reduce: AddsubdotTerm
			reduce(184), // ??, reduce: AddsubdotTerm
			reduce(184), // ???, reduce: AddsubdotTerm
			reduce(184), // =~, reduce: AddsubdotTerm
			reduce(184), // !=~, reduce: AddsubdotTerm
			reduce(184), // ==, reduce: AddsubdotTerm
			reduce(184), // !=, reduce: AddsubdotTerm
			reduce(184), // <=>, reduce: AddsubdotTerm
			reduce(184), // >=, reduce: AddsubdotTerm
			reduce(184), // <, reduce: AddsubdotTerm
			reduce(184), // <=, reduce: AddsubdotTerm
			reduce(184), // ^, reduce: AddsubdotTerm
			reduce(184), // &, reduce: AddsubdotTerm
			reduce(184), // <<, reduce: AddsubdotTerm
			reduce(184), // >>>, reduce: AddsubdotTerm
			reduce(184), // +, reduce: AddsubdotTerm
			reduce(184), // -, reduce: AddsubdotTerm
			reduce(184), // .+, reduce: AddsubdotTerm
			reduce(184), // .-, reduce: AddsubdotTerm
			shift(903),  // *
			shift(904),  // /
			shift(905),  // //
			shift(906),  // %
			shift(907),  // .*
			shift(908),  // ./
			shift(909),  // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1990
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(185), // >, reduce: AddsubdotTerm
			reduce(185), // >>, reduce: AddsubdotTerm
			reduce(185), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(185), // ?, reduce: AddsubdotTerm
			reduce(185), // :, reduce: AddsubdotTerm
			reduce(185), // ||, reduce: AddsubdotTerm
			reduce(185), // ^^, reduce: AddsubdotTerm
			reduce(185), // &&, reduce: AddsubdotTerm
			reduce(185), // ??, reduce: AddsubdotTerm
			reduce(185), // ???, reduce: AddsubdotTerm
			reduce(185), // =~, reduce: AddsubdotTerm
			reduce(185), // !=~, reduce: AddsubdotTerm
			reduce(185), // ==, reduce: AddsubdotTerm
			reduce(185), // !=, reduce: AddsubdotTerm
			reduce(185), // <=>, reduce: AddsubdotTerm
			reduce(185), // >=, reduce: AddsubdotTerm
			reduce(185), // <, reduce: AddsubdotTerm
			reduce(185), // <=, reduce: AddsubdotTerm
			reduce(185), // ^, reduce: AddsubdotTerm
			reduce(185), // &, reduce: AddsubdotTerm
			reduce(185), // <<, reduce: AddsubdotTerm
			reduce(185), // >>>, reduce: AddsubdotTerm
			reduce(185), // +, reduce: AddsubdotTerm
			reduce(185), // -, reduce: AddsubdotTerm
			reduce(185), // .+, reduce: AddsubdotTerm
			reduce(185), // .-, reduce: AddsubdotTerm
			shift(903),  // *
			shift(904),  // /
			shift(905),  // //
			shift(906),  // %
			shift(907),  // .*
			shift(908),  // ./
			shift(909),  // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1991
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(187), // >, reduce: MuldivTerm
			reduce(187), // >>, reduce: MuldivTerm
			reduce(187), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(187), // ?, reduce: MuldivTerm
			reduce(187), // :, reduce: MuldivTerm
			reduce(187), // ||, reduce: MuldivTerm
			reduce(187), // ^^, reduce: MuldivTerm
			reduce(187), // &&, reduce: MuldivTerm
			reduce(187), // ??, reduce: MuldivTerm
			reduce(187), // ???, reduce: MuldivTerm
			reduce(187), // =~, reduce: MuldivTerm
			reduce(187), // !=~, reduce: MuldivTerm
			reduce(187), // ==, reduce: MuldivTerm
			reduce(187), // !=, reduce: MuldivTerm
			reduce(187), // <=>, reduce: MuldivTerm
			reduce(187), // >=, reduce: MuldivTerm
			reduce(187), // <, reduce: MuldivTerm
			reduce(187), // <=, reduce: MuldivTerm
			reduce(187), // ^, reduce: MuldivTerm
			reduce(187), // &, reduce: MuldivTerm
			reduce(187), // <<, reduce: MuldivTerm
			reduce(187), // >>>, reduce: MuldivTerm
			reduce(187), // +, reduce: MuldivTerm
			reduce(187), // -, reduce: MuldivTerm
			reduce(187), // .+, reduce: MuldivTerm
			reduce(187), // .-, reduce: MuldivTerm
			reduce(187), // *, reduce: MuldivTerm
			reduce(187), // /, reduce: MuldivTerm
			reduce(187), // //, reduce: MuldivTerm
			reduce(187), // %, reduce: MuldivTerm
			reduce(187), // .*, reduce: MuldivTerm
			reduce(187), // ./, reduce: MuldivTerm
			reduce(187), // .//, reduce: MuldivTerm
			shift(913),  // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1992
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(188), // >, reduce: MuldivTerm
			reduce(188), // >>, reduce: MuldivTerm
			reduce(188), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(188), // ?, reduce: MuldivTerm
			reduce(188), // :, reduce: MuldivTerm
			reduce(188), // ||, reduce: MuldivTerm
			reduce(188), // ^^, reduce: MuldivTerm
			reduce(188), // &&, reduce: MuldivTerm
			reduce(188), // ??, reduce: MuldivTerm
			reduce(188), // ???, reduce: MuldivTerm
			reduce(188), // =~, reduce: MuldivTerm
			reduce(188), // !=~, reduce: MuldivTerm
			reduce(188), // ==, reduce: MuldivTerm
			reduce(188), // !=, reduce: MuldivTerm
			reduce(188), // <=>, reduce: MuldivTerm
			reduce(188), // >=, reduce: MuldivTerm
			reduce(188), // <, reduce: MuldivTerm
			reduce(188), // <=, reduce: MuldivTerm
			reduce(188), // ^, reduce: MuldivTerm
			reduce(188), // &, reduce: MuldivTerm
			reduce(188), // <<, reduce: MuldivTerm
			reduce(188), // >>>, reduce: MuldivTerm
			reduce(188), // +, reduce: MuldivTerm
			reduce(188), // -, reduce: MuldivTerm
			reduce(188), // .+, reduce: MuldivTerm
			reduce(188), // .-, reduce: MuldivTerm
			reduce(188), // *, reduce: MuldivTerm
			reduce(188), // /, reduce: MuldivTerm
			reduce(188), // //, reduce: MuldivTerm
			reduce(188), // %, reduce: MuldivTerm
			reduce(188), // .*, reduce: MuldivTerm
			reduce(188), // ./, reduce: MuldivTerm
			reduce(188), // .//, reduce: MuldivTerm
			shift(913),  // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1993
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(189), // >, reduce: MuldivTerm
			reduce(189), // >>, reduce: MuldivTerm
			reduce(189), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(189), // ?, reduce: MuldivTerm
			reduce(189), // :, reduce: MuldivTerm
			reduce(189), // ||, reduce: MuldivTerm
			reduce(189), // ^^, reduce: MuldivTerm
			reduce(189), // &&, reduce: MuldivTerm
			reduce(189), // ??, reduce: MuldivTerm
			reduce(189), // ???, reduce: MuldivTerm
			reduce(189), // =~, reduce: MuldivTerm
			reduce(189), // !=~, reduce: MuldivTerm
			reduce(189), // ==, reduce: MuldivTerm
			reduce(189), // !=, reduce: MuldivTerm
			reduce(189), // <=>, reduce: MuldivTerm
			reduce(189), // >=, reduce: MuldivTerm
			reduce(189), // <, reduce: MuldivTerm
			reduce(189), // <=, reduce: MuldivTerm
			reduce(189), // ^, reduce: MuldivTerm
			reduce(189), // &, reduce: MuldivTerm
			reduce(189), // <<, reduce: MuldivTerm
			reduce(189), // >>>, reduce: MuldivTerm
			reduce(189), // +, reduce: MuldivTerm
			reduce(189), // -, reduce: MuldivTerm
			reduce(189), // .+, reduce: MuldivTerm
			reduce(189), // .-, reduce: MuldivTerm
			reduce(189), // *, reduce: MuldivTerm
			reduce(189), // /, reduce: MuldivTerm
			reduce(189), // //, reduce: MuldivTerm
			reduce(189), // %, reduce: MuldivTerm
			reduce(189), // .*, reduce: MuldivTerm
			reduce(189), // ./, reduce: MuldivTerm
			reduce(189), // .//, reduce: MuldivTerm
			shift(913),  // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1994
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(190), // >, reduce: MuldivTerm
			reduce(190), // >>, reduce: MuldivTerm
			reduce(190), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(190), // ?, reduce: MuldivTerm
			reduce(190), // :, reduce: MuldivTerm
			reduce(190), // ||, reduce: MuldivTerm
			reduce(190), // ^^, reduce: MuldivTerm
			reduce(190), // &&, reduce: MuldivTerm
			reduce(190), // ??, reduce: MuldivTerm
			reduce(190), // ???, reduce: MuldivTerm
			reduce(190), // =~, reduce: MuldivTerm
			reduce(190), // !=~, reduce: MuldivTerm
			reduce(190), // ==, reduce: MuldivTerm
			reduce(190), // !=, reduce: MuldivTerm
			reduce(190), // <=>, reduce: MuldivTerm
			reduce(190), // >=, reduce: MuldivTerm
			reduce(190), // <, reduce: MuldivTerm
			reduce(190), // <=, reduce: MuldivTerm
			reduce(190), // ^, reduce: MuldivTerm
			reduce(190), // &, reduce: MuldivTerm
			reduce(190), // <<, reduce: MuldivTerm
			reduce(190), // >>>, reduce: MuldivTerm
			reduce(190), // +, reduce: MuldivTerm
			reduce(190), // -, reduce: MuldivTerm
			reduce(190), // .+, reduce: MuldivTerm
			reduce(190), // .-, reduce: MuldivTerm
			reduce(190), // *, reduce: MuldivTerm
			reduce(190), // /, reduce: MuldivTerm
			reduce(190), // //, reduce: MuldivTerm
			reduce(190), // %, reduce: MuldivTerm
			reduce(190), // .*, reduce: MuldivTerm
			reduce(190), // ./, reduce: MuldivTerm
			reduce(190), // .//, reduce: MuldivTerm
			shift(913),  // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1995
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(191), // >, reduce: MuldivTerm
			reduce(191), // >>, reduce: MuldivTerm
			reduce(191), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(191), // ?, reduce: MuldivTerm
			reduce(191), // :, reduce: MuldivTerm
			reduce(191), // ||, reduce: MuldivTerm
			reduce(191), // ^^, reduce: MuldivTerm
			reduce(191), // &&, reduce: MuldivTerm
			reduce(191), // ??, reduce: MuldivTerm
			reduce(191), // ???, reduce: MuldivTerm
			reduce(191), // =~, reduce: MuldivTerm
			reduce(191), // !=~, reduce: MuldivTerm
			reduce(191), // ==, reduce: MuldivTerm
			reduce(191), // !=, reduce: MuldivTerm
			reduce(191), // <=>, reduce: MuldivTerm
			reduce(191), // >=, reduce: MuldivTerm
			reduce(191), // <, reduce: MuldivTerm
			reduce(191), // <=, reduce: MuldivTerm
			reduce(191), // ^, reduce: MuldivTerm
			reduce(191), // &, reduce: MuldivTerm
			reduce(191), // <<, reduce: MuldivTerm
			reduce(191), // >>>, reduce: MuldivTerm
			reduce(191), // +, reduce: MuldivTerm
			reduce(191), // -, reduce: MuldivTerm
			reduce(191), // .+, reduce: MuldivTerm
			reduce(191), // .-, reduce: MuldivTerm
			reduce(191), // *, reduce: MuldivTerm
			reduce(191), // /, reduce: MuldivTerm
			reduce(191), // //, reduce: MuldivTerm
			reduce(191), // %, reduce: MuldivTerm
			reduce(191), // .*, reduce: MuldivTerm
			reduce(191), // ./, reduce: MuldivTerm
			reduce(191), // .//, reduce: MuldivTerm
			shift(913),  // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1996
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(192), // >, reduce: MuldivTerm
			reduce(192), // >>, reduce: MuldivTerm
			reduce(192), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(192), // ?, reduce: MuldivTerm
			reduce(192), // :, reduce: MuldivTerm
			reduce(192), // ||, reduce: MuldivTerm
			reduce(192), // ^^, reduce: MuldivTerm
			reduce(192), // &&, reduce: MuldivTerm
			reduce(192), // ??, reduce: MuldivTerm
			reduce(192), // ???, reduce: MuldivTerm
			reduce(192), // =~, reduce: MuldivTerm
			reduce(192), // !=~, reduce: MuldivTerm
			reduce(192), // ==, reduce: MuldivTerm
			reduce(192), // !=, reduce: MuldivTerm
			reduce(192), // <=>, reduce: MuldivTerm
			reduce(192), // >=, reduce: MuldivTerm
			reduce(192), // <, reduce: MuldivTerm
			reduce(192), // <=, reduce: MuldivTerm
			reduce(192), // ^, reduce: MuldivTerm
			reduce(192), // &, reduce: MuldivTerm
			reduce(192), // <<, reduce: MuldivTerm
			reduce(192), // >>>, reduce: MuldivTerm
			reduce(192), // +, reduce: MuldivTerm
			reduce(192), // -, reduce: MuldivTerm
			reduce(192), // .+, reduce: MuldivTerm
			reduce(192), // .-, reduce: MuldivTerm
			reduce(192), // *, reduce: MuldivTerm
			reduce(192), // /, reduce: MuldivTerm
			reduce(192), // //, reduce: MuldivTerm
			reduce(192), // %, reduce: MuldivTerm
			reduce(192), // .*, reduce: MuldivTerm
			reduce(192), // ./, reduce: MuldivTerm
			reduce(192), // .//, reduce: MuldivTerm
			shift(913),  // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1997
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(193), // >, reduce: MuldivTerm
			reduce(193), // >>, reduce: MuldivTerm
			reduce(193), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(193), // ?, reduce: MuldivTerm
			reduce(193), // :, reduce: MuldivTerm
			reduce(193), // ||, reduce: MuldivTerm
			reduce(193), // ^^, reduce: MuldivTerm
			reduce(193), // &&, reduce: MuldivTerm
			reduce(193), // ??, reduce: MuldivTerm
			reduce(193), // ???, reduce: MuldivTerm
			reduce(193), // =~, reduce: MuldivTerm
			reduce(193), // !=~, reduce: MuldivTerm
			reduce(193), // ==, reduce: MuldivTerm
			reduce(193), // !=, reduce: MuldivTerm
			reduce(193), // <=>, reduce: MuldivTerm
			reduce(193), // >=, reduce: MuldivTerm
			reduce(193), // <, reduce: MuldivTerm
			reduce(193), // <=, reduce: MuldivTerm
			reduce(193), // ^, reduce: MuldivTerm
			reduce(193), // &, reduce: MuldivTerm
			reduce(193), // <<, reduce: MuldivTerm
			reduce(193), // >>>, reduce: MuldivTerm
			reduce(193), // +, reduce: MuldivTerm
			reduce(193), // -, reduce: MuldivTerm
			reduce(193), // .+, reduce: MuldivTerm
			reduce(193), // .-, reduce: MuldivTerm
			reduce(193), // *, reduce: MuldivTerm
			reduce(193), // /, reduce: MuldivTerm
			reduce(193), // //, reduce: MuldivTerm
			reduce(193), // %, reduce: MuldivTerm
			reduce(193), // .*, reduce: MuldivTerm
			reduce(193), // ./, reduce: MuldivTerm
			reduce(193), // .//, reduce: MuldivTerm
			shift(913),  // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1998
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(195), // >, reduce: DotTerm
			reduce(195), // >>, reduce: DotTerm
			reduce(195), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(195), // ?, reduce: DotTerm
			reduce(195), // :, reduce: DotTerm
			reduce(195), // ||, reduce: DotTerm
			reduce(195), // ^^, reduce: DotTerm
			reduce(195), // &&, reduce: DotTerm
			reduce(195), // ??, reduce: DotTerm
			reduce(195), // ???, reduce: DotTerm
			reduce(195), // =~, reduce: DotTerm
			reduce(195), // !=~, reduce: DotTerm
			reduce(195), // ==, reduce: DotTerm
			reduce(195), // !=, reduce: DotTerm
			reduce(195), // <=>, reduce: DotTerm
			reduce(195), // >=, reduce: DotTerm
			reduce(195), // <, reduce: DotTerm
			reduce(195), // <=, reduce: DotTerm
			reduce(195), // ^, reduce: DotTerm
			reduce(195), // &, reduce: DotTerm
			reduce(195), // <<, reduce: DotTerm
			reduce(195), // >>>, reduce: DotTerm
			reduce(195), // +, reduce: DotTerm
			reduce(195), // -, reduce: DotTerm
			reduce(195), // .+, reduce: DotTerm
			reduce(195), // .-, reduce: DotTerm
			reduce(195), // *, reduce: DotTerm
			reduce(195), // /, reduce: DotTerm
			reduce(195), // //, reduce: DotTerm
			reduce(195), // %, reduce: DotTerm
			reduce(195), // .*, reduce: DotTerm
			reduce(195), // ./, reduce: DotTerm
			reduce(195), // .//, reduce: DotTerm
			reduce(195), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S1999
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2000
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2001
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(204), // >, reduce: PowTerm
			reduce(204), // >>, reduce: PowTerm
			reduce(204), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(204), // ?, reduce: PowTerm
			reduce(204), // :, reduce: PowTerm
			reduce(204), // ||, reduce: PowTerm
			reduce(204), // ^^, reduce: PowTerm
			reduce(204), // &&, reduce: PowTerm
			reduce(204), // ??, reduce: PowTerm
			reduce(204), // ???, reduce: PowTerm
			reduce(204), // =~, reduce: PowTerm
			reduce(204), // !=~, reduce: PowTerm
			reduce(204), // ==, reduce: PowTerm
			reduce(204), // !=, reduce: PowTerm
			reduce(204), // <=>, reduce: PowTerm
			reduce(204), // >=, reduce: PowTerm
			reduce(204), // <, reduce: PowTerm
			reduce(204), // <=, reduce: PowTerm
			reduce(204), // ^, reduce: PowTerm
			reduce(204), // &, reduce: PowTerm
			reduce(204), // <<, reduce: PowTerm
			reduce(204), // >>>, reduce: PowTerm
			reduce(204), // +, reduce: PowTerm
			reduce(204), // -, reduce: PowTerm
			reduce(204), // .+, reduce: PowTerm
			reduce(204), // .-, reduce: PowTerm
			reduce(204), // *, reduce: PowTerm
			reduce(204), // /, reduce: PowTerm
			reduce(204), // //, reduce: PowTerm
			reduce(204), // %, reduce: PowTerm
			reduce(204), // .*, reduce: PowTerm
			reduce(204), // ./, reduce: PowTerm
			reduce(204), // .//, reduce: PowTerm
			reduce(204), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2002
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2966), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(2967), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2003
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(2969), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2004
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2970), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2005
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2971), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2006
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(229), // >, reduce: ArrayLiteral
			reduce(229), // >>, reduce: ArrayLiteral
			reduce(229), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(229), // ?, reduce: ArrayLiteral
			reduce(229), // :, reduce: ArrayLiteral
			reduce(229), // ||, reduce: ArrayLiteral
			reduce(229), // ^^, reduce: ArrayLiteral
			reduce(229), // &&, reduce: ArrayLiteral
			reduce(229), // ??, reduce: ArrayLiteral
			reduce(229), // ???, reduce: ArrayLiteral
			reduce(229), // =~, reduce: ArrayLiteral
			reduce(229), // !=~, reduce: ArrayLiteral
			reduce(229), // ==, reduce: ArrayLiteral
			reduce(229), // !=, reduce: ArrayLiteral
			reduce(229), // <=>, reduce: ArrayLiteral
			reduce(229), // >=, reduce: ArrayLiteral
			reduce(229), // <, reduce: ArrayLiteral
			reduce(229), // <=, reduce: ArrayLiteral
			reduce(229), // ^, reduce: ArrayLiteral
			reduce(229), // &, reduce: ArrayLiteral
			reduce(229), // <<, reduce: ArrayLiteral
			reduce(229), // >>>, reduce: ArrayLiteral
			reduce(229), // +, reduce: ArrayLiteral
			reduce(229), // -, reduce: ArrayLiteral
			reduce(229), // .+, reduce: ArrayLiteral
			reduce(229), // .-, reduce: ArrayLiteral
			reduce(229), // *, reduce: ArrayLiteral
			reduce(229), // /, reduce: ArrayLiteral
			reduce(229), // //, reduce: ArrayLiteral
			reduce(229), // %, reduce: ArrayLiteral
			reduce(229), // .*, reduce: ArrayLiteral
			reduce(229), // ./, reduce: ArrayLiteral
			reduce(229), // .//, reduce: ArrayLiteral
			reduce(229), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(229), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(229), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(229), // [[, reduce: ArrayLiteral
			reduce(229), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2007
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(238), // }, reduce: MapLiteralKeyValuePairs
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2008
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(255), // >, reduce: ENV
			reduce(255), // >>, reduce: ENV
			reduce(255), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(255), // ?, reduce: ENV
			reduce(255), // :, reduce: ENV
			reduce(255), // ||, reduce: ENV
			reduce(255), // ^^, reduce: ENV
			reduce(255), // &&, reduce: ENV
			reduce(255), // ??, reduce: ENV
			reduce(255), // ???, reduce: ENV
			reduce(255), // =~, reduce: ENV
			reduce(255), // !=~, reduce: ENV
			reduce(255), // ==, reduce: ENV
			reduce(255), // !=, reduce: ENV
			reduce(255), // <=>, reduce: ENV
			reduce(255), // >=, reduce: ENV
			reduce(255), // <, reduce: ENV
			reduce(255), // <=, reduce: ENV
			reduce(255), // ^, reduce: ENV
			reduce(255), // &, reduce: ENV
			reduce(255), // <<, reduce: ENV
			reduce(255), // >>>, reduce: ENV
			reduce(255), // +, reduce: ENV
			reduce(255), // -, reduce: ENV
			reduce(255), // .+, reduce: ENV
			reduce(255), // .-, reduce: ENV
			reduce(255), // *, reduce: ENV
			reduce(255), // /, reduce: ENV
			reduce(255), // //, reduce: ENV
			reduce(255), // %, reduce: ENV
			reduce(255), // .*, reduce: ENV
			reduce(255), // ./, reduce: ENV
			reduce(255), // .//, reduce: ENV
			reduce(255), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(255), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(255), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(255), // [[, reduce: ENV
			reduce(255), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2009
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(2972), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2010
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(2973), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2011
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(268), // >, reduce: FunctionCallsite
			reduce(268), // >>, reduce: FunctionCallsite
			reduce(268), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(268), // ?, reduce: FunctionCallsite
			reduce(268), // :, reduce: FunctionCallsite
			reduce(268), // ||, reduce: FunctionCallsite
			reduce(268), // ^^, reduce: FunctionCallsite
			reduce(268), // &&, reduce: FunctionCallsite
			reduce(268), // ??, reduce: FunctionCallsite
			reduce(268), // ???, reduce: FunctionCallsite
			reduce(268), // =~, reduce: FunctionCallsite
			reduce(268), // !=~, reduce: FunctionCallsite
			reduce(268), // ==, reduce: FunctionCallsite
			reduce(268), // !=, reduce: FunctionCallsite
			reduce(268), // <=>, reduce: FunctionCallsite
			reduce(268), // >=, reduce: FunctionCallsite
			reduce(268), // <, reduce: FunctionCallsite
			reduce(268), // <=, reduce: FunctionCallsite
			reduce(268), // ^, reduce: FunctionCallsite
			reduce(268), // &, reduce: FunctionCallsite
			reduce(268), // <<, reduce: FunctionCallsite
			reduce(268), // >>>, reduce: FunctionCallsite
			reduce(268), // +, reduce: FunctionCallsite
			reduce(268), // -, reduce: FunctionCallsite
			reduce(268), // .+, reduce: FunctionCallsite
			reduce(268), // .-, reduce: FunctionCallsite
			reduce(268), // *, reduce: FunctionCallsite
			reduce(268), // /, reduce: FunctionCallsite
			reduce(268), // //, reduce: FunctionCallsite
			reduce(268), // %, reduce: FunctionCallsite
			reduce(268), // .*, reduce: FunctionCallsite
			reduce(268), // ./, reduce: FunctionCallsite
			reduce(268), // .//, reduce: FunctionCallsite
			reduce(268), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(268), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(268), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(268), // [[, reduce: FunctionCallsite
			reduce(268), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2012
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(2974), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2013
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(13), // $, reduce: StatementBlockInBraces
			nil,        // empty
			reduce(13), // ;, reduce: StatementBlockInBraces
			reduce(13), // {, reduce: StatementBlockInBraces
			nil,        // }
			nil,        // =
			reduce(13), // unset, reduce: StatementBlockInBraces
			reduce(13), // filter, reduce: StatementBlockInBraces
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			reduce(13), // print, reduce: StatementBlockInBraces
			nil,        // ,
			reduce(13), // printn, reduce: StatementBlockInBraces
			reduce(13), // eprint, reduce: StatementBlockInBraces
			reduce(13), // eprintn, reduce: StatementBlockInBraces
			reduce(13), // dump, reduce: StatementBlockInBraces
			reduce(13), // edump, reduce: StatementBlockInBraces
			reduce(13), // tee, reduce: StatementBlockInBraces
			reduce(13), // emitf, reduce: StatementBlockInBraces
			reduce(13), // emit1, reduce: StatementBlockInBraces
			reduce(13), // emit, reduce: StatementBlockInBraces
			reduce(13), // (, reduce: StatementBlockInBraces
			nil,        // )
			reduce(13), // emitp, reduce: StatementBlockInBraces
			reduce(13), // field_name, reduce: StatementBlockInBraces
			reduce(13), // $[, reduce: StatementBlockInBraces
			nil,        // ]
			reduce(13), // braced_field_name, reduce: StatementBlockInBraces
			reduce(13), // $[[, reduce: StatementBlockInBraces
			reduce(13), // $[[[, reduce: StatementBlockInBraces
			reduce(13), // full_srec, reduce: StatementBlockInBraces
			reduce(13), // oosvar_name, reduce: StatementBlockInBraces
			reduce(13), // @[, reduce: StatementBlockInBraces
			reduce(13), // braced_oosvar_name, reduce: StatementBlockInBraces
			reduce(13), // full_oosvar, reduce: StatementBlockInBraces
			reduce(13), // all, reduce: StatementBlockInBraces
			reduce(13), // non_sigil_name, reduce: StatementBlockInBraces
			reduce(13), // arr, reduce: StatementBlockInBraces
			reduce(13), // bool, reduce: StatementBlockInBraces
			reduce(13), // float, reduce: StatementBlockInBraces
			reduce(13), // int, reduce: StatementBlockInBraces
			reduce(13), // map, reduce: StatementBlockInBraces
			reduce(13), // num, reduce: StatementBlockInBraces
			reduce(13), // str, reduce: StatementBlockInBraces
			reduce(13), // var, reduce: StatementBlockInBraces
			reduce(13), // funct, reduce: StatementBlockInBraces
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			reduce(13), // !, reduce: StatementBlockInBraces
			reduce(13), // ~, reduce: StatementBlockInBraces
			nil,        // **
			reduce(13), // string_literal, reduce: StatementBlockInBraces
			reduce(13), // regex_case_insensitive, reduce: StatementBlockInBraces
			reduce(13), // int_literal, reduce: StatementBlockInBraces
			reduce(13), // float_literal, reduce: StatementBlockInBraces
			reduce(13), // boolean_literal, reduce: StatementBlockInBraces
			reduce(13), // null_literal, reduce: StatementBlockInBraces
			reduce(13), // inf_literal, reduce: StatementBlockInBraces
			reduce(13), // nan_literal, reduce: StatementBlockInBraces
			reduce(13), // const_M_PI, reduce: StatementBlockInBraces
			reduce(13), // const_M_E, reduce: StatementBlockInBraces
			reduce(13), // panic, reduce: StatementBlockInBraces
			reduce(13), // [, reduce: StatementBlockInBraces
			reduce(13), // ctx_IPS, reduce: StatementBlockInBraces
			reduce(13), // ctx_IFS, reduce: StatementBlockInBraces
			reduce(13), // ctx_IRS, reduce: StatementBlockInBraces
			reduce(13), // ctx_OPS, reduce: StatementBlockInBraces
			reduce(13), // ctx_OFS, reduce: StatementBlockInBraces
			reduce(13), // ctx_ORS, reduce: StatementBlockInBraces
			reduce(13), // ctx_FLATSEP, reduce: StatementBlockInBraces
			reduce(13), // ctx_NF, reduce: StatementBlockInBraces
			reduce(13), // ctx_NR, reduce: StatementBlockInBraces
			reduce(13), // ctx_FNR, reduce: StatementBlockInBraces
			reduce(13), // ctx_FILENAME, reduce: StatementBlockInBraces
			reduce(13), // ctx_FILENUM, reduce: StatementBlockInBraces
			reduce(13), // env, reduce: StatementBlockInBraces
			nil,        // [[
			nil,        // [[[
			reduce(13), // call, reduce: StatementBlockInBraces
			reduce(13), // begin, reduce: StatementBlockInBraces
			reduce(13), // end, reduce: StatementBlockInBraces
			reduce(13), // if, reduce: StatementBlockInBraces
			nil,        // elif
			nil,        // else
			reduce(13), // while, reduce: StatementBlockInBraces
			reduce(13), // do, reduce: StatementBlockInBraces
			reduce(13), // for, reduce: StatementBlockInBraces
			nil,        // in
			reduce(13), // break, reduce: StatementBlockInBraces
			reduce(13), // continue, reduce: StatementBlockInBraces
			reduce(13), // func, reduce: StatementBlockInBraces
			reduce(13), // subr, reduce: StatementBlockInBraces
			reduce(13), // return, reduce: StatementBlockInBraces
		},
	},
	actionRow{ // S2014
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S2015
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(2976), // ;
			nil,         // {
			reduce(10),  // }, reduce: NonEmptyStatementBlock
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2016
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S2017
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S2018
		canRecover: false,
		actions: [numSymbols]action{
			nil,       // INVALID
			nil,       // $
			nil,       // empty
			nil,       // ;
			nil,       // {
			reduce(6), // }, reduce: NonEmptyStatementBlock
			nil,       // =
			nil,       // unset
			nil,       // filter
			nil,       // >
			nil,       // >>
			nil,       // |
			nil,       // stdout
			nil,       // stderr
			nil,       // print
			nil,       // ,
			nil,       // printn
			nil,       // eprint
			nil,       // eprintn
			nil,       // dump
			nil,       // edump
			nil,       // tee
			nil,       // emitf
			nil,       // emit1
			nil,       // emit
			nil,       // (
			nil,       // )
			nil,       // emitp
			nil,       // field_name
			nil,       // $[
			nil,       // ]
			nil,       // braced_field_name
			nil,       // $[[
			nil,       // $[[[
			nil,       // full_srec
			nil,       // oosvar_name
			nil,       // @[
			nil,       // braced_oosvar_name
			nil,       // full_oosvar
			nil,       // all
			nil,       // non_sigil_name
			nil,       // arr
			nil,       // bool
			nil,       // float
			nil,       // int
			nil,       // map
			nil,       // num
			nil,       // str
			nil,       // var
			nil,       // funct
			nil,       // ||=
			nil,       // ^^=
			nil,       // &&=
			nil,       // ??=
			nil,       // ???=
			nil,       // |=
			nil,       // &=
			nil,       // ^=
			nil,       // <<=
			nil,       // >>=
			nil,       // >>>=
			nil,       // +=
			nil,       // .=
			nil,       // -=
			nil,       // *=
			nil,       // /=
			nil,       // //=
			nil,       // %=
			nil,       // **=
			nil,       // ?
			nil,       // :
			nil,       // ||
			nil,       // ^^
			nil,       // &&
			nil,       // ??
			nil,       // ???
			nil,       // =~
			nil,       // !=~
			nil,       // ==
			nil,       // !=
			nil,       // <=>
			nil,       // >=
			nil,       // <
			nil,       // <=
			nil,       // ^
			nil,       // &
			nil,       // <<
			nil,       // >>>
			nil,       // +
			nil,       // -
			nil,       // .+
			nil,       // .-
			nil,       // *
			nil,       // /
			nil,       // //
			nil,       // %
			nil,       // .*
			nil,       // ./
			nil,       // .//
			nil,       // .
			nil,       // !
			nil,       // ~
			nil,       // **
			nil,       // string_literal
			nil,       // regex_case_insensitive
			nil,       // int_literal
			nil,       // float_literal
			nil,       // boolean_literal
			nil,       // null_literal
			nil,       // inf_literal
			nil,       // nan_literal
			nil,       // const_M_PI
			nil,       // const_M_E
			nil,       // panic
			nil,       // [
			nil,       // ctx_IPS
			nil,       // ctx_IFS
			nil,       // ctx_IRS
			nil,       // ctx_OPS
			nil,       // ctx_OFS
			nil,       // ctx_ORS
			nil,       // ctx_FLATSEP
			nil,       // ctx_NF
			nil,       // ctx_NR
			nil,       // ctx_FNR
			nil,       // ctx_FILENAME
			nil,       // ctx_FILENUM
			nil,       // env
			nil,       // [[
			nil,       // [[[
			nil,       // call
			nil,       // begin
			nil,       // end
			nil,       // if
			nil,       // elif
			nil,       // else
			nil,       // while
			nil,       // do
			nil,       // for
			nil,       // in
			nil,       // break
			nil,       // continue
			nil,       // func
			nil,       // subr
			nil,       // return
		},
	},
	actionRow{ // S2019
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(234), // ;, reduce: MapLiteral
			reduce(234), // {, reduce: MapLiteral
			reduce(234), // }, reduce: MapLiteral
			reduce(234), // =, reduce: MapLiteral
			nil,         // unset
			nil,         // filter
			reduce(234), // >, reduce: MapLiteral
			reduce(234), // >>, reduce: MapLiteral
			reduce(234), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(234), // ||=, reduce: MapLiteral
			reduce(234), // ^^=, reduce: MapLiteral
			reduce(234), // &&=, reduce: MapLiteral
			reduce(234), // ??=, reduce: MapLiteral
			reduce(234), // ???=, reduce: MapLiteral
			reduce(234), // |=, reduce: MapLiteral
			reduce(234), // &=, reduce: MapLiteral
			reduce(234), // ^=, reduce: MapLiteral
			reduce(234), // <<=, reduce: MapLiteral
			reduce(234), // >>=, reduce: MapLiteral
			reduce(234), // >>>=, reduce: MapLiteral
			reduce(234), // +=, reduce: MapLiteral
			reduce(234), // .=, reduce: MapLiteral
			reduce(234), // -=, reduce: MapLiteral
			reduce(234), // *=, reduce: MapLiteral
			reduce(234), // /=, reduce: MapLiteral
			reduce(234), // //=, reduce: MapLiteral
			reduce(234), // %=, reduce: MapLiteral
			reduce(234), // **=, reduce: MapLiteral
			reduce(234), // ?, reduce: MapLiteral
			nil,         // :
			reduce(234), // ||, reduce: MapLiteral
			reduce(234), // ^^, reduce: MapLiteral
			reduce(234), // &&, reduce: MapLiteral
			reduce(234), // ??, reduce: MapLiteral
			reduce(234), // ???, reduce: MapLiteral
			reduce(234), // =~, reduce: MapLiteral
			reduce(234), // !=~, reduce: MapLiteral
			reduce(234), // ==, reduce: MapLiteral
			reduce(234), // !=, reduce: MapLiteral
			reduce(234), // <=>, reduce: MapLiteral
			reduce(234), // >=, reduce: MapLiteral
			reduce(234), // <, reduce: MapLiteral
			reduce(234), // <=, reduce: MapLiteral
			reduce(234), // ^, reduce: MapLiteral
			reduce(234), // &, reduce: MapLiteral
			reduce(234), // <<, reduce: MapLiteral
			reduce(234), // >>>, reduce: MapLiteral
			reduce(234), // +, reduce: MapLiteral
			reduce(234), // -, reduce: MapLiteral
			reduce(234), // .+, reduce: MapLiteral
			reduce(234), // .-, reduce: MapLiteral
			reduce(234), // *, reduce: MapLiteral
			reduce(234), // /, reduce: MapLiteral
			reduce(234), // //, reduce: MapLiteral
			reduce(234), // %, reduce: MapLiteral
			reduce(234), // .*, reduce: MapLiteral
			reduce(234), // ./, reduce: MapLiteral
			reduce(234), // .//, reduce: MapLiteral
			reduce(234), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(234), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(234), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(234), // [[, reduce: MapLiteral
			reduce(234), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2020
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(2979), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2021
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2022
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2023
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2024
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2025
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2026
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2027
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2028
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2029
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2030
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2031
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2032
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2033
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2034
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2035
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2036
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2037
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2038
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2039
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2040
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2041
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(289), // ;, reduce: CondBlock
			reduce(289), // {, reduce: CondBlock
			reduce(289), // }, reduce: CondBlock
			nil,         // =
			reduce(289), // unset, reduce: CondBlock
			reduce(289), // filter, reduce: CondBlock
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(289), // print, reduce: CondBlock
			nil,         // ,
			reduce(289), // printn, reduce: CondBlock
			reduce(289), // eprint, reduce: CondBlock
			reduce(289), // eprintn, reduce: CondBlock
			reduce(289), // dump, reduce: CondBlock
			reduce(289), // edump, reduce: CondBlock
			reduce(289), // tee, reduce: CondBlock
			reduce(289), // emitf, reduce: CondBlock
			reduce(289), // emit1, reduce: CondBlock
			reduce(289), // emit, reduce: CondBlock
			reduce(289), // (, reduce: CondBlock
			nil,         // )
			reduce(289), // emitp, reduce: CondBlock
			reduce(289), // field_name, reduce: CondBlock
			reduce(289), // $[, reduce: CondBlock
			nil,         // ]
			reduce(289), // braced_field_name, reduce: CondBlock
			reduce(289), // $[[, reduce: CondBlock
			reduce(289), // $[[[, reduce: CondBlock
			reduce(289), // full_srec, reduce: CondBlock
			reduce(289), // oosvar_name, reduce: CondBlock
			reduce(289), // @[, reduce: CondBlock
			reduce(289), // braced_oosvar_name, reduce: CondBlock
			reduce(289), // full_oosvar, reduce: CondBlock
			reduce(289), // all, reduce: CondBlock
			reduce(289), // non_sigil_name, reduce: CondBlock
			reduce(289), // arr, reduce: CondBlock
			reduce(289), // bool, reduce: CondBlock
			reduce(289), // float, reduce: CondBlock
			reduce(289), // int, reduce: CondBlock
			reduce(289), // map, reduce: CondBlock
			reduce(289), // num, reduce: CondBlock
			reduce(289), // str, reduce: CondBlock
			reduce(289), // var, reduce: CondBlock
			reduce(289), // funct, reduce: CondBlock
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(289), // +, reduce: CondBlock
			reduce(289), // -, reduce: CondBlock
			reduce(289), // .+, reduce: CondBlock
			reduce(289), // .-, reduce: CondBlock
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(289), // !, reduce: CondBlock
			reduce(289), // ~, reduce: CondBlock
			nil,         // **
			reduce(289), // string_literal, reduce: CondBlock
			reduce(289), // regex_case_insensitive, reduce: CondBlock
			reduce(289), // int_literal, reduce: CondBlock
			reduce(289), // float_literal, reduce: CondBlock
			reduce(289), // boolean_literal, reduce: CondBlock
			reduce(289), // null_literal, reduce: CondBlock
			reduce(289), // inf_literal, reduce: CondBlock
			reduce(289), // nan_literal, reduce: CondBlock
			reduce(289), // const_M_PI, reduce: CondBlock
			reduce(289), // const_M_E, reduce: CondBlock
			reduce(289), // panic, reduce: CondBlock
			reduce(289), // [, reduce: CondBlock
			reduce(289), // ctx_IPS, reduce: CondBlock
			reduce(289), // ctx_IFS, reduce: CondBlock
			reduce(289), // ctx_IRS, reduce: CondBlock
			reduce(289), // ctx_OPS, reduce: CondBlock
			reduce(289), // ctx_OFS, reduce: CondBlock
			reduce(289), // ctx_ORS, reduce: CondBlock
			reduce(289), // ctx_FLATSEP, reduce: CondBlock
			reduce(289), // ctx_NF, reduce: CondBlock
			reduce(289), // ctx_NR, reduce: CondBlock
			reduce(289), // ctx_FNR, reduce: CondBlock
			reduce(289), // ctx_FILENAME, reduce: CondBlock
			reduce(289), // ctx_FILENUM, reduce: CondBlock
			reduce(289), // env, reduce: CondBlock
			nil,         // [[
			nil,         // [[[
			reduce(289), // call, reduce: CondBlock
			reduce(289), // begin, reduce: CondBlock
			reduce(289), // end, reduce: CondBlock
			reduce(289), // if, reduce: CondBlock
			nil,         // elif
			nil,         // else
			reduce(289), // while, reduce: CondBlock
			reduce(289), // do, reduce: CondBlock
			reduce(289), // for, reduce: CondBlock
			nil,         // in
			reduce(289), // break, reduce: CondBlock
			reduce(289), // continue, reduce: CondBlock
			reduce(289), // func, reduce: CondBlock
			reduce(289), // subr, reduce: CondBlock
			reduce(289), // return, reduce: CondBlock
		},
	},
	actionRow{ // S2042
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S2043
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			shift(3001), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2044
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(273), // ;, reduce: FcnArgs
			nil,         // {
			reduce(273), // }, reduce: FcnArgs
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3003), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2045
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(35), // ;, reduce: Unset
			nil,        // {
			reduce(35), // }, reduce: Unset
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2046
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(214), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(214), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(214), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(214), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2047
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(211), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(211), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(211), // >, reduce: MlrvalOrFunction
			reduce(211), // >>, reduce: MlrvalOrFunction
			reduce(211), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(211), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(211), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(211), // ||, reduce: MlrvalOrFunction
			reduce(211), // ^^, reduce: MlrvalOrFunction
			reduce(211), // &&, reduce: MlrvalOrFunction
			reduce(211), // ??, reduce: MlrvalOrFunction
			reduce(211), // ???, reduce: MlrvalOrFunction
			reduce(211), // =~, reduce: MlrvalOrFunction
			reduce(211), // !=~, reduce: MlrvalOrFunction
			reduce(211), // ==, reduce: MlrvalOrFunction
			reduce(211), // !=, reduce: MlrvalOrFunction
			reduce(211), // <=>, reduce: MlrvalOrFunction
			reduce(211), // >=, reduce: MlrvalOrFunction
			reduce(211), // <, reduce: MlrvalOrFunction
			reduce(211), // <=, reduce: MlrvalOrFunction
			reduce(211), // ^, reduce: MlrvalOrFunction
			reduce(211), // &, reduce: MlrvalOrFunction
			reduce(211), // <<, reduce: MlrvalOrFunction
			reduce(211), // >>>, reduce: MlrvalOrFunction
			reduce(211), // +, reduce: MlrvalOrFunction
			reduce(211), // -, reduce: MlrvalOrFunction
			reduce(211), // .+, reduce: MlrvalOrFunction
			reduce(211), // .-, reduce: MlrvalOrFunction
			reduce(211), // *, reduce: MlrvalOrFunction
			reduce(211), // /, reduce: MlrvalOrFunction
			reduce(211), // //, reduce: MlrvalOrFunction
			reduce(211), // %, reduce: MlrvalOrFunction
			reduce(211), // .*, reduce: MlrvalOrFunction
			reduce(211), // ./, reduce: MlrvalOrFunction
			reduce(211), // .//, reduce: MlrvalOrFunction
			reduce(211), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(211), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(211), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(211), // [[, reduce: MlrvalOrFunction
			reduce(211), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2048
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2049
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(110), // ;, reduce: OosvarValue
			nil,         // {
			reduce(110), // }, reduce: OosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(110), // >, reduce: OosvarValue
			reduce(110), // >>, reduce: OosvarValue
			reduce(110), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(110), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(110), // ?, reduce: OosvarValue
			nil,         // :
			reduce(110), // ||, reduce: OosvarValue
			reduce(110), // ^^, reduce: OosvarValue
			reduce(110), // &&, reduce: OosvarValue
			reduce(110), // ??, reduce: OosvarValue
			reduce(110), // ???, reduce: OosvarValue
			reduce(110), // =~, reduce: OosvarValue
			reduce(110), // !=~, reduce: OosvarValue
			reduce(110), // ==, reduce: OosvarValue
			reduce(110), // !=, reduce: OosvarValue
			reduce(110), // <=>, reduce: OosvarValue
			reduce(110), // >=, reduce: OosvarValue
			reduce(110), // <, reduce: OosvarValue
			reduce(110), // <=, reduce: OosvarValue
			reduce(110), // ^, reduce: OosvarValue
			reduce(110), // &, reduce: OosvarValue
			reduce(110), // <<, reduce: OosvarValue
			reduce(110), // >>>, reduce: OosvarValue
			reduce(110), // +, reduce: OosvarValue
			reduce(110), // -, reduce: OosvarValue
			reduce(110), // .+, reduce: OosvarValue
			reduce(110), // .-, reduce: OosvarValue
			reduce(110), // *, reduce: OosvarValue
			reduce(110), // /, reduce: OosvarValue
			reduce(110), // //, reduce: OosvarValue
			reduce(110), // %, reduce: OosvarValue
			reduce(110), // .*, reduce: OosvarValue
			reduce(110), // ./, reduce: OosvarValue
			reduce(110), // .//, reduce: OosvarValue
			reduce(110), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(110), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(110), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(110), // [[, reduce: OosvarValue
			reduce(110), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2050
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(112), // ;, reduce: OosvarValue
			nil,         // {
			reduce(112), // }, reduce: OosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(112), // >, reduce: OosvarValue
			reduce(112), // >>, reduce: OosvarValue
			reduce(112), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(112), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(112), // ?, reduce: OosvarValue
			nil,         // :
			reduce(112), // ||, reduce: OosvarValue
			reduce(112), // ^^, reduce: OosvarValue
			reduce(112), // &&, reduce: OosvarValue
			reduce(112), // ??, reduce: OosvarValue
			reduce(112), // ???, reduce: OosvarValue
			reduce(112), // =~, reduce: OosvarValue
			reduce(112), // !=~, reduce: OosvarValue
			reduce(112), // ==, reduce: OosvarValue
			reduce(112), // !=, reduce: OosvarValue
			reduce(112), // <=>, reduce: OosvarValue
			reduce(112), // >=, reduce: OosvarValue
			reduce(112), // <, reduce: OosvarValue
			reduce(112), // <=, reduce: OosvarValue
			reduce(112), // ^, reduce: OosvarValue
			reduce(112), // &, reduce: OosvarValue
			reduce(112), // <<, reduce: OosvarValue
			reduce(112), // >>>, reduce: OosvarValue
			reduce(112), // +, reduce: OosvarValue
			reduce(112), // -, reduce: OosvarValue
			reduce(112), // .+, reduce: OosvarValue
			reduce(112), // .-, reduce: OosvarValue
			reduce(112), // *, reduce: OosvarValue
			reduce(112), // /, reduce: OosvarValue
			reduce(112), // //, reduce: OosvarValue
			reduce(112), // %, reduce: OosvarValue
			reduce(112), // .*, reduce: OosvarValue
			reduce(112), // ./, reduce: OosvarValue
			reduce(112), // .//, reduce: OosvarValue
			reduce(112), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(112), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(112), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(112), // [[, reduce: OosvarValue
			reduce(112), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2051
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(111), // ;, reduce: OosvarValue
			nil,         // {
			reduce(111), // }, reduce: OosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(111), // >, reduce: OosvarValue
			reduce(111), // >>, reduce: OosvarValue
			reduce(111), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(111), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(111), // ?, reduce: OosvarValue
			nil,         // :
			reduce(111), // ||, reduce: OosvarValue
			reduce(111), // ^^, reduce: OosvarValue
			reduce(111), // &&, reduce: OosvarValue
			reduce(111), // ??, reduce: OosvarValue
			reduce(111), // ???, reduce: OosvarValue
			reduce(111), // =~, reduce: OosvarValue
			reduce(111), // !=~, reduce: OosvarValue
			reduce(111), // ==, reduce: OosvarValue
			reduce(111), // !=, reduce: OosvarValue
			reduce(111), // <=>, reduce: OosvarValue
			reduce(111), // >=, reduce: OosvarValue
			reduce(111), // <, reduce: OosvarValue
			reduce(111), // <=, reduce: OosvarValue
			reduce(111), // ^, reduce: OosvarValue
			reduce(111), // &, reduce: OosvarValue
			reduce(111), // <<, reduce: OosvarValue
			reduce(111), // >>>, reduce: OosvarValue
			reduce(111), // +, reduce: OosvarValue
			reduce(111), // -, reduce: OosvarValue
			reduce(111), // .+, reduce: OosvarValue
			reduce(111), // .-, reduce: OosvarValue
			reduce(111), // *, reduce: OosvarValue
			reduce(111), // /, reduce: OosvarValue
			reduce(111), // //, reduce: OosvarValue
			reduce(111), // %, reduce: OosvarValue
			reduce(111), // .*, reduce: OosvarValue
			reduce(111), // ./, reduce: OosvarValue
			reduce(111), // .//, reduce: OosvarValue
			reduce(111), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(111), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(111), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(111), // [[, reduce: OosvarValue
			reduce(111), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2052
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(99), // ;, reduce: FieldValue
			nil,        // {
			reduce(99), // }, reduce: FieldValue
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(99), // >, reduce: FieldValue
			reduce(99), // >>, reduce: FieldValue
			reduce(99), // |, reduce: FieldValue
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(99), // ,, reduce: FieldValue
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(99), // ?, reduce: FieldValue
			nil,        // :
			reduce(99), // ||, reduce: FieldValue
			reduce(99), // ^^, reduce: FieldValue
			reduce(99), // &&, reduce: FieldValue
			reduce(99), // ??, reduce: FieldValue
			reduce(99), // ???, reduce: FieldValue
			reduce(99), // =~, reduce: FieldValue
			reduce(99), // !=~, reduce: FieldValue
			reduce(99), // ==, reduce: FieldValue
			reduce(99), // !=, reduce: FieldValue
			reduce(99), // <=>, reduce: FieldValue
			reduce(99), // >=, reduce: FieldValue
			reduce(99), // <, reduce: FieldValue
			reduce(99), // <=, reduce: FieldValue
			reduce(99), // ^, reduce: FieldValue
			reduce(99), // &, reduce: FieldValue
			reduce(99), // <<, reduce: FieldValue
			reduce(99), // >>>, reduce: FieldValue
			reduce(99), // +, reduce: FieldValue
			reduce(99), // -, reduce: FieldValue
			reduce(99), // .+, reduce: FieldValue
			reduce(99), // .-, reduce: FieldValue
			reduce(99), // *, reduce: FieldValue
			reduce(99), // /, reduce: FieldValue
			reduce(99), // //, reduce: FieldValue
			reduce(99), // %, reduce: FieldValue
			reduce(99), // .*, reduce: FieldValue
			reduce(99), // ./, reduce: FieldValue
			reduce(99), // .//, reduce: FieldValue
			reduce(99), // ., reduce: FieldValue
			nil,        // !
			nil,        // ~
			reduce(99), // **, reduce: FieldValue
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(99), // [, reduce: FieldValue
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(99), // [[, reduce: FieldValue
			reduce(99), // [[[, reduce: FieldValue
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2053
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(101), // ;, reduce: FieldValue
			nil,         // {
			reduce(101), // }, reduce: FieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(101), // >, reduce: FieldValue
			reduce(101), // >>, reduce: FieldValue
			reduce(101), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(101), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(101), // ?, reduce: FieldValue
			nil,         // :
			reduce(101), // ||, reduce: FieldValue
			reduce(101), // ^^, reduce: FieldValue
			reduce(101), // &&, reduce: FieldValue
			reduce(101), // ??, reduce: FieldValue
			reduce(101), // ???, reduce: FieldValue
			reduce(101), // =~, reduce: FieldValue
			reduce(101), // !=~, reduce: FieldValue
			reduce(101), // ==, reduce: FieldValue
			reduce(101), // !=, reduce: FieldValue
			reduce(101), // <=>, reduce: FieldValue
			reduce(101), // >=, reduce: FieldValue
			reduce(101), // <, reduce: FieldValue
			reduce(101), // <=, reduce: FieldValue
			reduce(101), // ^, reduce: FieldValue
			reduce(101), // &, reduce: FieldValue
			reduce(101), // <<, reduce: FieldValue
			reduce(101), // >>>, reduce: FieldValue
			reduce(101), // +, reduce: FieldValue
			reduce(101), // -, reduce: FieldValue
			reduce(101), // .+, reduce: FieldValue
			reduce(101), // .-, reduce: FieldValue
			reduce(101), // *, reduce: FieldValue
			reduce(101), // /, reduce: FieldValue
			reduce(101), // //, reduce: FieldValue
			reduce(101), // %, reduce: FieldValue
			reduce(101), // .*, reduce: FieldValue
			reduce(101), // ./, reduce: FieldValue
			reduce(101), // .//, reduce: FieldValue
			reduce(101), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(101), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(101), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(101), // [[, reduce: FieldValue
			reduce(101), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2054
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(100), // ;, reduce: FieldValue
			nil,         // {
			reduce(100), // }, reduce: FieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(100), // >, reduce: FieldValue
			reduce(100), // >>, reduce: FieldValue
			reduce(100), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(100), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(100), // ?, reduce: FieldValue
			nil,         // :
			reduce(100), // ||, reduce: FieldValue
			reduce(100), // ^^, reduce: FieldValue
			reduce(100), // &&, reduce: FieldValue
			reduce(100), // ??, reduce: FieldValue
			reduce(100), // ???, reduce: FieldValue
			reduce(100), // =~, reduce: FieldValue
			reduce(100), // !=~, reduce: FieldValue
			reduce(100), // ==, reduce: FieldValue
			reduce(100), // !=, reduce: FieldValue
			reduce(100), // <=>, reduce: FieldValue
			reduce(100), // >=, reduce: FieldValue
			reduce(100), // <, reduce: FieldValue
			reduce(100), // <=, reduce: FieldValue
			reduce(100), // ^, reduce: FieldValue
			reduce(100), // &, reduce: FieldValue
			reduce(100), // <<, reduce: FieldValue
			reduce(100), // >>>, reduce: FieldValue
			reduce(100), // +, reduce: FieldValue
			reduce(100), // -, reduce: FieldValue
			reduce(100), // .+, reduce: FieldValue
			reduce(100), // .-, reduce: FieldValue
			reduce(100), // *, reduce: FieldValue
			reduce(100), // /, reduce: FieldValue
			reduce(100), // //, reduce: FieldValue
			reduce(100), // %, reduce: FieldValue
			reduce(100), // .*, reduce: FieldValue
			reduce(100), // ./, reduce: FieldValue
			reduce(100), // .//, reduce: FieldValue
			reduce(100), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(100), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(100), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(100), // [[, reduce: FieldValue
			reduce(100), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2055
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(213), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(213), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(213), // >, reduce: MlrvalOrFunction
			reduce(213), // >>, reduce: MlrvalOrFunction
			reduce(213), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(213), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(213), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(213), // ||, reduce: MlrvalOrFunction
			reduce(213), // ^^, reduce: MlrvalOrFunction
			reduce(213), // &&, reduce: MlrvalOrFunction
			reduce(213), // ??, reduce: MlrvalOrFunction
			reduce(213), // ???, reduce: MlrvalOrFunction
			reduce(213), // =~, reduce: MlrvalOrFunction
			reduce(213), // !=~, reduce: MlrvalOrFunction
			reduce(213), // ==, reduce: MlrvalOrFunction
			reduce(213), // !=, reduce: MlrvalOrFunction
			reduce(213), // <=>, reduce: MlrvalOrFunction
			reduce(213), // >=, reduce: MlrvalOrFunction
			reduce(213), // <, reduce: MlrvalOrFunction
			reduce(213), // <=, reduce: MlrvalOrFunction
			reduce(213), // ^, reduce: MlrvalOrFunction
			reduce(213), // &, reduce: MlrvalOrFunction
			reduce(213), // <<, reduce: MlrvalOrFunction
			reduce(213), // >>>, reduce: MlrvalOrFunction
			reduce(213), // +, reduce: MlrvalOrFunction
			reduce(213), // -, reduce: MlrvalOrFunction
			reduce(213), // .+, reduce: MlrvalOrFunction
			reduce(213), // .-, reduce: MlrvalOrFunction
			reduce(213), // *, reduce: MlrvalOrFunction
			reduce(213), // /, reduce: MlrvalOrFunction
			reduce(213), // //, reduce: MlrvalOrFunction
			reduce(213), // %, reduce: MlrvalOrFunction
			reduce(213), // .*, reduce: MlrvalOrFunction
			reduce(213), // ./, reduce: MlrvalOrFunction
			reduce(213), // .//, reduce: MlrvalOrFunction
			reduce(213), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(213), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(213), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(213), // [[, reduce: MlrvalOrFunction
			reduce(213), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2056
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(233), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(233), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(233), // >, reduce: MlrvalOrFunction
			reduce(233), // >>, reduce: MlrvalOrFunction
			reduce(233), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(233), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(233), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(233), // ||, reduce: MlrvalOrFunction
			reduce(233), // ^^, reduce: MlrvalOrFunction
			reduce(233), // &&, reduce: MlrvalOrFunction
			reduce(233), // ??, reduce: MlrvalOrFunction
			reduce(233), // ???, reduce: MlrvalOrFunction
			reduce(233), // =~, reduce: MlrvalOrFunction
			reduce(233), // !=~, reduce: MlrvalOrFunction
			reduce(233), // ==, reduce: MlrvalOrFunction
			reduce(233), // !=, reduce: MlrvalOrFunction
			reduce(233), // <=>, reduce: MlrvalOrFunction
			reduce(233), // >=, reduce: MlrvalOrFunction
			reduce(233), // <, reduce: MlrvalOrFunction
			reduce(233), // <=, reduce: MlrvalOrFunction
			reduce(233), // ^, reduce: MlrvalOrFunction
			reduce(233), // &, reduce: MlrvalOrFunction
			reduce(233), // <<, reduce: MlrvalOrFunction
			reduce(233), // >>>, reduce: MlrvalOrFunction
			reduce(233), // +, reduce: MlrvalOrFunction
			reduce(233), // -, reduce: MlrvalOrFunction
			reduce(233), // .+, reduce: MlrvalOrFunction
			reduce(233), // .-, reduce: MlrvalOrFunction
			reduce(233), // *, reduce: MlrvalOrFunction
			reduce(233), // /, reduce: MlrvalOrFunction
			reduce(233), // //, reduce: MlrvalOrFunction
			reduce(233), // %, reduce: MlrvalOrFunction
			reduce(233), // .*, reduce: MlrvalOrFunction
			reduce(233), // ./, reduce: MlrvalOrFunction
			reduce(233), // .//, reduce: MlrvalOrFunction
			reduce(233), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(233), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(233), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(233), // [[, reduce: MlrvalOrFunction
			reduce(233), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2057
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(210), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(210), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(210), // >, reduce: MlrvalOrFunction
			reduce(210), // >>, reduce: MlrvalOrFunction
			reduce(210), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(210), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(210), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(210), // ||, reduce: MlrvalOrFunction
			reduce(210), // ^^, reduce: MlrvalOrFunction
			reduce(210), // &&, reduce: MlrvalOrFunction
			reduce(210), // ??, reduce: MlrvalOrFunction
			reduce(210), // ???, reduce: MlrvalOrFunction
			reduce(210), // =~, reduce: MlrvalOrFunction
			reduce(210), // !=~, reduce: MlrvalOrFunction
			reduce(210), // ==, reduce: MlrvalOrFunction
			reduce(210), // !=, reduce: MlrvalOrFunction
			reduce(210), // <=>, reduce: MlrvalOrFunction
			reduce(210), // >=, reduce: MlrvalOrFunction
			reduce(210), // <, reduce: MlrvalOrFunction
			reduce(210), // <=, reduce: MlrvalOrFunction
			reduce(210), // ^, reduce: MlrvalOrFunction
			reduce(210), // &, reduce: MlrvalOrFunction
			reduce(210), // <<, reduce: MlrvalOrFunction
			reduce(210), // >>>, reduce: MlrvalOrFunction
			reduce(210), // +, reduce: MlrvalOrFunction
			reduce(210), // -, reduce: MlrvalOrFunction
			reduce(210), // .+, reduce: MlrvalOrFunction
			reduce(210), // .-, reduce: MlrvalOrFunction
			reduce(210), // *, reduce: MlrvalOrFunction
			reduce(210), // /, reduce: MlrvalOrFunction
			reduce(210), // //, reduce: MlrvalOrFunction
			reduce(210), // %, reduce: MlrvalOrFunction
			reduce(210), // .*, reduce: MlrvalOrFunction
			reduce(210), // ./, reduce: MlrvalOrFunction
			reduce(210), // .//, reduce: MlrvalOrFunction
			reduce(210), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(210), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(210), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(210), // [[, reduce: MlrvalOrFunction
			reduce(210), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2058
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(102), // ;, reduce: FieldValue
			nil,         // {
			reduce(102), // }, reduce: FieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(102), // >, reduce: FieldValue
			reduce(102), // >>, reduce: FieldValue
			reduce(102), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(102), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(102), // ?, reduce: FieldValue
			nil,         // :
			reduce(102), // ||, reduce: FieldValue
			reduce(102), // ^^, reduce: FieldValue
			reduce(102), // &&, reduce: FieldValue
			reduce(102), // ??, reduce: FieldValue
			reduce(102), // ???, reduce: FieldValue
			reduce(102), // =~, reduce: FieldValue
			reduce(102), // !=~, reduce: FieldValue
			reduce(102), // ==, reduce: FieldValue
			reduce(102), // !=, reduce: FieldValue
			reduce(102), // <=>, reduce: FieldValue
			reduce(102), // >=, reduce: FieldValue
			reduce(102), // <, reduce: FieldValue
			reduce(102), // <=, reduce: FieldValue
			reduce(102), // ^, reduce: FieldValue
			reduce(102), // &, reduce: FieldValue
			reduce(102), // <<, reduce: FieldValue
			reduce(102), // >>>, reduce: FieldValue
			reduce(102), // +, reduce: FieldValue
			reduce(102), // -, reduce: FieldValue
			reduce(102), // .+, reduce: FieldValue
			reduce(102), // .-, reduce: FieldValue
			reduce(102), // *, reduce: FieldValue
			reduce(102), // /, reduce: FieldValue
			reduce(102), // //, reduce: FieldValue
			reduce(102), // %, reduce: FieldValue
			reduce(102), // .*, reduce: FieldValue
			reduce(102), // ./, reduce: FieldValue
			reduce(102), // .//, reduce: FieldValue
			reduce(102), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(102), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(102), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(102), // [[, reduce: FieldValue
			reduce(102), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2059
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(103), // ;, reduce: FieldValue
			nil,         // {
			reduce(103), // }, reduce: FieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(103), // >, reduce: FieldValue
			reduce(103), // >>, reduce: FieldValue
			reduce(103), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(103), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(103), // ?, reduce: FieldValue
			nil,         // :
			reduce(103), // ||, reduce: FieldValue
			reduce(103), // ^^, reduce: FieldValue
			reduce(103), // &&, reduce: FieldValue
			reduce(103), // ??, reduce: FieldValue
			reduce(103), // ???, reduce: FieldValue
			reduce(103), // =~, reduce: FieldValue
			reduce(103), // !=~, reduce: FieldValue
			reduce(103), // ==, reduce: FieldValue
			reduce(103), // !=, reduce: FieldValue
			reduce(103), // <=>, reduce: FieldValue
			reduce(103), // >=, reduce: FieldValue
			reduce(103), // <, reduce: FieldValue
			reduce(103), // <=, reduce: FieldValue
			reduce(103), // ^, reduce: FieldValue
			reduce(103), // &, reduce: FieldValue
			reduce(103), // <<, reduce: FieldValue
			reduce(103), // >>>, reduce: FieldValue
			reduce(103), // +, reduce: FieldValue
			reduce(103), // -, reduce: FieldValue
			reduce(103), // .+, reduce: FieldValue
			reduce(103), // .-, reduce: FieldValue
			reduce(103), // *, reduce: FieldValue
			reduce(103), // /, reduce: FieldValue
			reduce(103), // //, reduce: FieldValue
			reduce(103), // %, reduce: FieldValue
			reduce(103), // .*, reduce: FieldValue
			reduce(103), // ./, reduce: FieldValue
			reduce(103), // .//, reduce: FieldValue
			reduce(103), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(103), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(103), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(103), // [[, reduce: FieldValue
			reduce(103), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2060
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(104), // ;, reduce: DirectFieldValue
			nil,         // {
			reduce(104), // }, reduce: DirectFieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(104), // >, reduce: DirectFieldValue
			reduce(104), // >>, reduce: DirectFieldValue
			reduce(104), // |, reduce: DirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(104), // ,, reduce: DirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(104), // ?, reduce: DirectFieldValue
			nil,         // :
			reduce(104), // ||, reduce: DirectFieldValue
			reduce(104), // ^^, reduce: DirectFieldValue
			reduce(104), // &&, reduce: DirectFieldValue
			reduce(104), // ??, reduce: DirectFieldValue
			reduce(104), // ???, reduce: DirectFieldValue
			reduce(104), // =~, reduce: DirectFieldValue
			reduce(104), // !=~, reduce: DirectFieldValue
			reduce(104), // ==, reduce: DirectFieldValue
			reduce(104), // !=, reduce: DirectFieldValue
			reduce(104), // <=>, reduce: DirectFieldValue
			reduce(104), // >=, reduce: DirectFieldValue
			reduce(104), // <, reduce: DirectFieldValue
			reduce(104), // <=, reduce: DirectFieldValue
			reduce(104), // ^, reduce: DirectFieldValue
			reduce(104), // &, reduce: DirectFieldValue
			reduce(104), // <<, reduce: DirectFieldValue
			reduce(104), // >>>, reduce: DirectFieldValue
			reduce(104), // +, reduce: DirectFieldValue
			reduce(104), // -, reduce: DirectFieldValue
			reduce(104), // .+, reduce: DirectFieldValue
			reduce(104), // .-, reduce: DirectFieldValue
			reduce(104), // *, reduce: DirectFieldValue
			reduce(104), // /, reduce: DirectFieldValue
			reduce(104), // //, reduce: DirectFieldValue
			reduce(104), // %, reduce: DirectFieldValue
			reduce(104), // .*, reduce: DirectFieldValue
			reduce(104), // ./, reduce: DirectFieldValue
			reduce(104), // .//, reduce: DirectFieldValue
			reduce(104), // ., reduce: DirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(104), // **, reduce: DirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(104), // [, reduce: DirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(104), // [[, reduce: DirectFieldValue
			reduce(104), // [[[, reduce: DirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2061
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2062
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(106), // ;, reduce: BracedFieldValue
			nil,         // {
			reduce(106), // }, reduce: BracedFieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(106), // >, reduce: BracedFieldValue
			reduce(106), // >>, reduce: BracedFieldValue
			reduce(106), // |, reduce: BracedFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(106), // ,, reduce: BracedFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(106), // ?, reduce: BracedFieldValue
			nil,         // :
			reduce(106), // ||, reduce: BracedFieldValue
			reduce(106), // ^^, reduce: BracedFieldValue
			reduce(106), // &&, reduce: BracedFieldValue
			reduce(106), // ??, reduce: BracedFieldValue
			reduce(106), // ???, reduce: BracedFieldValue
			reduce(106), // =~, reduce: BracedFieldValue
			reduce(106), // !=~, reduce: BracedFieldValue
			reduce(106), // ==, reduce: BracedFieldValue
			reduce(106), // !=, reduce: BracedFieldValue
			reduce(106), // <=>, reduce: BracedFieldValue
			reduce(106), // >=, reduce: BracedFieldValue
			reduce(106), // <, reduce: BracedFieldValue
			reduce(106), // <=, reduce: BracedFieldValue
			reduce(106), // ^, reduce: BracedFieldValue
			reduce(106), // &, reduce: BracedFieldValue
			reduce(106), // <<, reduce: BracedFieldValue
			reduce(106), // >>>, reduce: BracedFieldValue
			reduce(106), // +, reduce: BracedFieldValue
			reduce(106), // -, reduce: BracedFieldValue
			reduce(106), // .+, reduce: BracedFieldValue
			reduce(106), // .-, reduce: BracedFieldValue
			reduce(106), // *, reduce: BracedFieldValue
			reduce(106), // /, reduce: BracedFieldValue
			reduce(106), // //, reduce: BracedFieldValue
			reduce(106), // %, reduce: BracedFieldValue
			reduce(106), // .*, reduce: BracedFieldValue
			reduce(106), // ./, reduce: BracedFieldValue
			reduce(106), // .//, reduce: BracedFieldValue
			reduce(106), // ., reduce: BracedFieldValue
			nil,         // !
			nil,         // ~
			reduce(106), // **, reduce: BracedFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(106), // [, reduce: BracedFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(106), // [[, reduce: BracedFieldValue
			reduce(106), // [[[, reduce: BracedFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2063
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2064
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2065
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(109), // ;, reduce: FullSrec
			nil,         // {
			reduce(109), // }, reduce: FullSrec
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(109), // >, reduce: FullSrec
			reduce(109), // >>, reduce: FullSrec
			reduce(109), // |, reduce: FullSrec
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(109), // ,, reduce: FullSrec
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(109), // ?, reduce: FullSrec
			nil,         // :
			reduce(109), // ||, reduce: FullSrec
			reduce(109), // ^^, reduce: FullSrec
			reduce(109), // &&, reduce: FullSrec
			reduce(109), // ??, reduce: FullSrec
			reduce(109), // ???, reduce: FullSrec
			reduce(109), // =~, reduce: FullSrec
			reduce(109), // !=~, reduce: FullSrec
			reduce(109), // ==, reduce: FullSrec
			reduce(109), // !=, reduce: FullSrec
			reduce(109), // <=>, reduce: FullSrec
			reduce(109), // >=, reduce: FullSrec
			reduce(109), // <, reduce: FullSrec
			reduce(109), // <=, reduce: FullSrec
			reduce(109), // ^, reduce: FullSrec
			reduce(109), // &, reduce: FullSrec
			reduce(109), // <<, reduce: FullSrec
			reduce(109), // >>>, reduce: FullSrec
			reduce(109), // +, reduce: FullSrec
			reduce(109), // -, reduce: FullSrec
			reduce(109), // .+, reduce: FullSrec
			reduce(109), // .-, reduce: FullSrec
			reduce(109), // *, reduce: FullSrec
			reduce(109), // /, reduce: FullSrec
			reduce(109), // //, reduce: FullSrec
			reduce(109), // %, reduce: FullSrec
			reduce(109), // .*, reduce: FullSrec
			reduce(109), // ./, reduce: FullSrec
			reduce(109), // .//, reduce: FullSrec
			reduce(109), // ., reduce: FullSrec
			nil,         // !
			nil,         // ~
			reduce(109), // **, reduce: FullSrec
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(109), // [, reduce: FullSrec
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(109), // [[, reduce: FullSrec
			reduce(109), // [[[, reduce: FullSrec
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2066
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(212), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(212), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(212), // >, reduce: MlrvalOrFunction
			reduce(212), // >>, reduce: MlrvalOrFunction
			reduce(212), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(212), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(212), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(212), // ||, reduce: MlrvalOrFunction
			reduce(212), // ^^, reduce: MlrvalOrFunction
			reduce(212), // &&, reduce: MlrvalOrFunction
			reduce(212), // ??, reduce: MlrvalOrFunction
			reduce(212), // ???, reduce: MlrvalOrFunction
			reduce(212), // =~, reduce: MlrvalOrFunction
			reduce(212), // !=~, reduce: MlrvalOrFunction
			reduce(212), // ==, reduce: MlrvalOrFunction
			reduce(212), // !=, reduce: MlrvalOrFunction
			reduce(212), // <=>, reduce: MlrvalOrFunction
			reduce(212), // >=, reduce: MlrvalOrFunction
			reduce(212), // <, reduce: MlrvalOrFunction
			reduce(212), // <=, reduce: MlrvalOrFunction
			reduce(212), // ^, reduce: MlrvalOrFunction
			reduce(212), // &, reduce: MlrvalOrFunction
			reduce(212), // <<, reduce: MlrvalOrFunction
			reduce(212), // >>>, reduce: MlrvalOrFunction
			reduce(212), // +, reduce: MlrvalOrFunction
			reduce(212), // -, reduce: MlrvalOrFunction
			reduce(212), // .+, reduce: MlrvalOrFunction
			reduce(212), // .-, reduce: MlrvalOrFunction
			reduce(212), // *, reduce: MlrvalOrFunction
			reduce(212), // /, reduce: MlrvalOrFunction
			reduce(212), // //, reduce: MlrvalOrFunction
			reduce(212), // %, reduce: MlrvalOrFunction
			reduce(212), // .*, reduce: MlrvalOrFunction
			reduce(212), // ./, reduce: MlrvalOrFunction
			reduce(212), // .//, reduce: MlrvalOrFunction
			reduce(212), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(212), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(212), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(212), // [[, reduce: MlrvalOrFunction
			reduce(212), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2067
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(113), // ;, reduce: DirectOosvarValue
			nil,         // {
			reduce(113), // }, reduce: DirectOosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(113), // >, reduce: DirectOosvarValue
			reduce(113), // >>, reduce: DirectOosvarValue
			reduce(113), // |, reduce: DirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(113), // ,, reduce: DirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(113), // ?, reduce: DirectOosvarValue
			nil,         // :
			reduce(113), // ||, reduce: DirectOosvarValue
			reduce(113), // ^^, reduce: DirectOosvarValue
			reduce(113), // &&, reduce: DirectOosvarValue
			reduce(113), // ??, reduce: DirectOosvarValue
			reduce(113), // ???, reduce: DirectOosvarValue
			reduce(113), // =~, reduce: DirectOosvarValue
			reduce(113), // !=~, reduce: DirectOosvarValue
			reduce(113), // ==, reduce: DirectOosvarValue
			reduce(113), // !=, reduce: DirectOosvarValue
			reduce(113), // <=>, reduce: DirectOosvarValue
			reduce(113), // >=, reduce: DirectOosvarValue
			reduce(113), // <, reduce: DirectOosvarValue
			reduce(113), // <=, reduce: DirectOosvarValue
			reduce(113), // ^, reduce: DirectOosvarValue
			reduce(113), // &, reduce: DirectOosvarValue
			reduce(113), // <<, reduce: DirectOosvarValue
			reduce(113), // >>>, reduce: DirectOosvarValue
			reduce(113), // +, reduce: DirectOosvarValue
			reduce(113), // -, reduce: DirectOosvarValue
			reduce(113), // .+, reduce: DirectOosvarValue
			reduce(113), // .-, reduce: DirectOosvarValue
			reduce(113), // *, reduce: DirectOosvarValue
			reduce(113), // /, reduce: DirectOosvarValue
			reduce(113), // //, reduce: DirectOosvarValue
			reduce(113), // %, reduce: DirectOosvarValue
			reduce(113), // .*, reduce: DirectOosvarValue
			reduce(113), // ./, reduce: DirectOosvarValue
			reduce(113), // .//, reduce: DirectOosvarValue
			reduce(113), // ., reduce: DirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(113), // **, reduce: DirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(113), // [, reduce: DirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(113), // [[, reduce: DirectOosvarValue
			reduce(113), // [[[, reduce: DirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2068
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2069
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(115), // ;, reduce: BracedOosvarValue
			nil,         // {
			reduce(115), // }, reduce: BracedOosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(115), // >, reduce: BracedOosvarValue
			reduce(115), // >>, reduce: BracedOosvarValue
			reduce(115), // |, reduce: BracedOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(115), // ,, reduce: BracedOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(115), // ?, reduce: BracedOosvarValue
			nil,         // :
			reduce(115), // ||, reduce: BracedOosvarValue
			reduce(115), // ^^, reduce: BracedOosvarValue
			reduce(115), // &&, reduce: BracedOosvarValue
			reduce(115), // ??, reduce: BracedOosvarValue
			reduce(115), // ???, reduce: BracedOosvarValue
			reduce(115), // =~, reduce: BracedOosvarValue
			reduce(115), // !=~, reduce: BracedOosvarValue
			reduce(115), // ==, reduce: BracedOosvarValue
			reduce(115), // !=, reduce: BracedOosvarValue
			reduce(115), // <=>, reduce: BracedOosvarValue
			reduce(115), // >=, reduce: BracedOosvarValue
			reduce(115), // <, reduce: BracedOosvarValue
			reduce(115), // <=, reduce: BracedOosvarValue
			reduce(115), // ^, reduce: BracedOosvarValue
			reduce(115), // &, reduce: BracedOosvarValue
			reduce(115), // <<, reduce: BracedOosvarValue
			reduce(115), // >>>, reduce: BracedOosvarValue
			reduce(115), // +, reduce: BracedOosvarValue
			reduce(115), // -, reduce: BracedOosvarValue
			reduce(115), // .+, reduce: BracedOosvarValue
			reduce(115), // .-, reduce: BracedOosvarValue
			reduce(115), // *, reduce: BracedOosvarValue
			reduce(115), // /, reduce: BracedOosvarValue
			reduce(115), // //, reduce: BracedOosvarValue
			reduce(115), // %, reduce: BracedOosvarValue
			reduce(115), // .*, reduce: BracedOosvarValue
			reduce(115), // ./, reduce: BracedOosvarValue
			reduce(115), // .//, reduce: BracedOosvarValue
			reduce(115), // ., reduce: BracedOosvarValue
			nil,         // !
			nil,         // ~
			reduce(115), // **, reduce: BracedOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(115), // [, reduce: BracedOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(115), // [[, reduce: BracedOosvarValue
			reduce(115), // [[[, reduce: BracedOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2070
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(116), // ;, reduce: FullOosvar
			nil,         // {
			reduce(116), // }, reduce: FullOosvar
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(116), // >, reduce: FullOosvar
			reduce(116), // >>, reduce: FullOosvar
			reduce(116), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(116), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(116), // ?, reduce: FullOosvar
			nil,         // :
			reduce(116), // ||, reduce: FullOosvar
			reduce(116), // ^^, reduce: FullOosvar
			reduce(116), // &&, reduce: FullOosvar
			reduce(116), // ??, reduce: FullOosvar
			reduce(116), // ???, reduce: FullOosvar
			reduce(116), // =~, reduce: FullOosvar
			reduce(116), // !=~, reduce: FullOosvar
			reduce(116), // ==, reduce: FullOosvar
			reduce(116), // !=, reduce: FullOosvar
			reduce(116), // <=>, reduce: FullOosvar
			reduce(116), // >=, reduce: FullOosvar
			reduce(116), // <, reduce: FullOosvar
			reduce(116), // <=, reduce: FullOosvar
			reduce(116), // ^, reduce: FullOosvar
			reduce(116), // &, reduce: FullOosvar
			reduce(116), // <<, reduce: FullOosvar
			reduce(116), // >>>, reduce: FullOosvar
			reduce(116), // +, reduce: FullOosvar
			reduce(116), // -, reduce: FullOosvar
			reduce(116), // .+, reduce: FullOosvar
			reduce(116), // .-, reduce: FullOosvar
			reduce(116), // *, reduce: FullOosvar
			reduce(116), // /, reduce: FullOosvar
			reduce(116), // //, reduce: FullOosvar
			reduce(116), // %, reduce: FullOosvar
			reduce(116), // .*, reduce: FullOosvar
			reduce(116), // ./, reduce: FullOosvar
			reduce(116), // .//, reduce: FullOosvar
			reduce(116), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(116), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(116), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(116), // [[, reduce: FullOosvar
			reduce(116), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2071
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(117), // ;, reduce: FullOosvar
			nil,         // {
			reduce(117), // }, reduce: FullOosvar
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(117), // >, reduce: FullOosvar
			reduce(117), // >>, reduce: FullOosvar
			reduce(117), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(117), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(117), // ?, reduce: FullOosvar
			nil,         // :
			reduce(117), // ||, reduce: FullOosvar
			reduce(117), // ^^, reduce: FullOosvar
			reduce(117), // &&, reduce: FullOosvar
			reduce(117), // ??, reduce: FullOosvar
			reduce(117), // ???, reduce: FullOosvar
			reduce(117), // =~, reduce: FullOosvar
			reduce(117), // !=~, reduce: FullOosvar
			reduce(117), // ==, reduce: FullOosvar
			reduce(117), // !=, reduce: FullOosvar
			reduce(117), // <=>, reduce: FullOosvar
			reduce(117), // >=, reduce: FullOosvar
			reduce(117), // <, reduce: FullOosvar
			reduce(117), // <=, reduce: FullOosvar
			reduce(117), // ^, reduce: FullOosvar
			reduce(117), // &, reduce: FullOosvar
			reduce(117), // <<, reduce: FullOosvar
			reduce(117), // >>>, reduce: FullOosvar
			reduce(117), // +, reduce: FullOosvar
			reduce(117), // -, reduce: FullOosvar
			reduce(117), // .+, reduce: FullOosvar
			reduce(117), // .-, reduce: FullOosvar
			reduce(117), // *, reduce: FullOosvar
			reduce(117), // /, reduce: FullOosvar
			reduce(117), // //, reduce: FullOosvar
			reduce(117), // %, reduce: FullOosvar
			reduce(117), // .*, reduce: FullOosvar
			reduce(117), // ./, reduce: FullOosvar
			reduce(117), // .//, reduce: FullOosvar
			reduce(117), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(117), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(117), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(117), // [[, reduce: FullOosvar
			reduce(117), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2072
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(118), // ;, reduce: LocalVariable
			nil,         // {
			reduce(118), // }, reduce: LocalVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(118), // >, reduce: LocalVariable
			reduce(118), // >>, reduce: LocalVariable
			reduce(118), // |, reduce: LocalVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(118), // ,, reduce: LocalVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(270), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(118), // ?, reduce: LocalVariable
			nil,         // :
			reduce(118), // ||, reduce: LocalVariable
			reduce(118), // ^^, reduce: LocalVariable
			reduce(118), // &&, reduce: LocalVariable
			reduce(118), // ??, reduce: LocalVariable
			reduce(118), // ???, reduce: LocalVariable
			reduce(118), // =~, reduce: LocalVariable
			reduce(118), // !=~, reduce: LocalVariable
			reduce(118), // ==, reduce: LocalVariable
			reduce(118), // !=, reduce: LocalVariable
			reduce(118), // <=>, reduce: LocalVariable
			reduce(118), // >=, reduce: LocalVariable
			reduce(118), // <, reduce: LocalVariable
			reduce(118), // <=, reduce: LocalVariable
			reduce(118), // ^, reduce: LocalVariable
			reduce(118), // &, reduce: LocalVariable
			reduce(118), // <<, reduce: LocalVariable
			reduce(118), // >>>, reduce: LocalVariable
			reduce(118), // +, reduce: LocalVariable
			reduce(118), // -, reduce: LocalVariable
			reduce(118), // .+, reduce: LocalVariable
			reduce(118), // .-, reduce: LocalVariable
			reduce(118), // *, reduce: LocalVariable
			reduce(118), // /, reduce: LocalVariable
			reduce(118), // //, reduce: LocalVariable
			reduce(118), // %, reduce: LocalVariable
			reduce(118), // .*, reduce: LocalVariable
			reduce(118), // ./, reduce: LocalVariable
			reduce(118), // .//, reduce: LocalVariable
			reduce(118), // ., reduce: LocalVariable
			nil,         // !
			nil,         // ~
			reduce(118), // **, reduce: LocalVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(118), // [, reduce: LocalVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(118), // [[, reduce: LocalVariable
			reduce(118), // [[[, reduce: LocalVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2073
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(147), // ;, reduce: Rvalue
			nil,         // {
			reduce(147), // }, reduce: Rvalue
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(147), // ,, reduce: Rvalue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2074
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(148), // ;, reduce: PrecedenceChainStart
			nil,         // {
			reduce(148), // }, reduce: PrecedenceChainStart
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(148), // ,, reduce: PrecedenceChainStart
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2075
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(150), // ;, reduce: TernaryTerm
			nil,         // {
			reduce(150), // }, reduce: TernaryTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(150), // ,, reduce: TernaryTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			shift(3009), // ?
			nil,         // :
			shift(3010), // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2076
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(152), // ;, reduce: LogicalOrTerm
			nil,         // {
			reduce(152), // }, reduce: LogicalOrTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(152), // ,, reduce: LogicalOrTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(152), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(152), // ||, reduce: LogicalOrTerm
			shift(3011), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2077
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(154), // ;, reduce: LogicalXORTerm
			nil,         // {
			reduce(154), // }, reduce: LogicalXORTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(154), // ,, reduce: LogicalXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(154), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(154), // ||, reduce: LogicalXORTerm
			reduce(154), // ^^, reduce: LogicalXORTerm
			shift(3012), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2078
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(156), // ;, reduce: LogicalAndTerm
			nil,         // {
			reduce(156), // }, reduce: LogicalAndTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(156), // ,, reduce: LogicalAndTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(156), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(156), // ||, reduce: LogicalAndTerm
			reduce(156), // ^^, reduce: LogicalAndTerm
			reduce(156), // &&, reduce: LogicalAndTerm
			shift(3013), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2079
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(158), // ;, reduce: AbsentCoalesceTerm
			nil,         // {
			reduce(158), // }, reduce: AbsentCoalesceTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(158), // ,, reduce: AbsentCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(158), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(158), // ||, reduce: AbsentCoalesceTerm
			reduce(158), // ^^, reduce: AbsentCoalesceTerm
			reduce(158), // &&, reduce: AbsentCoalesceTerm
			reduce(158), // ??, reduce: AbsentCoalesceTerm
			shift(3014), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2080
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(160), // ;, reduce: EmptyCoalesceTerm
			nil,         // {
			reduce(160), // }, reduce: EmptyCoalesceTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(160), // ,, reduce: EmptyCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(160), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(160), // ||, reduce: EmptyCoalesceTerm
			reduce(160), // ^^, reduce: EmptyCoalesceTerm
			reduce(160), // &&, reduce: EmptyCoalesceTerm
			reduce(160), // ??, reduce: EmptyCoalesceTerm
			reduce(160), // ???, reduce: EmptyCoalesceTerm
			shift(3015), // =~
			shift(3016), // !=~
			shift(3017), // ==
			shift(3018), // !=
			shift(3019), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2081
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(166), // ;, reduce: EqneTerm
			nil,         // {
			reduce(166), // }, reduce: EqneTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(3020), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(166), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(166), // ?, reduce: EqneTerm
			nil,         // :
			reduce(166), // ||, reduce: EqneTerm
			reduce(166), // ^^, reduce: EqneTerm
			reduce(166), // &&, reduce: EqneTerm
			reduce(166), // ??, reduce: EqneTerm
			reduce(166), // ???, reduce: EqneTerm
			reduce(166), // =~, reduce: EqneTerm
			reduce(166), // !=~, reduce: EqneTerm
			reduce(166), // ==, reduce: EqneTerm
			reduce(166), // !=, reduce: EqneTerm
			reduce(166), // <=>, reduce: EqneTerm
			shift(3021), // >=
			shift(3022), // <
			shift(3023), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2082
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(171), // ;, reduce: CmpTerm
			nil,         // {
			reduce(171), // }, reduce: CmpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(171), // >, reduce: CmpTerm
			nil,         // >>
			shift(3024), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(171), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(171), // ?, reduce: CmpTerm
			nil,         // :
			reduce(171), // ||, reduce: CmpTerm
			reduce(171), // ^^, reduce: CmpTerm
			reduce(171), // &&, reduce: CmpTerm
			reduce(171), // ??, reduce: CmpTerm
			reduce(171), // ???, reduce: CmpTerm
			reduce(171), // =~, reduce: CmpTerm
			reduce(171), // !=~, reduce: CmpTerm
			reduce(171), // ==, reduce: CmpTerm
			reduce(171), // !=, reduce: CmpTerm
			reduce(171), // <=>, reduce: CmpTerm
			reduce(171), // >=, reduce: CmpTerm
			reduce(171), // <, reduce: CmpTerm
			reduce(171), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2083
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(173), // ;, reduce: BitwiseORTerm
			nil,         // {
			reduce(173), // }, reduce: BitwiseORTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(173), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(173), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(173), // ,, reduce: BitwiseORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(173), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(173), // ||, reduce: BitwiseORTerm
			reduce(173), // ^^, reduce: BitwiseORTerm
			reduce(173), // &&, reduce: BitwiseORTerm
			reduce(173), // ??, reduce: BitwiseORTerm
			reduce(173), // ???, reduce: BitwiseORTerm
			reduce(173), // =~, reduce: BitwiseORTerm
			reduce(173), // !=~, reduce: BitwiseORTerm
			reduce(173), // ==, reduce: BitwiseORTerm
			reduce(173), // !=, reduce: BitwiseORTerm
			reduce(173), // <=>, reduce: BitwiseORTerm
			reduce(173), // >=, reduce: BitwiseORTerm
			reduce(173), // <, reduce: BitwiseORTerm
			reduce(173), // <=, reduce: BitwiseORTerm
			shift(3025), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2084
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(175), // ;, reduce: BitwiseXORTerm
			nil,         // {
			reduce(175), // }, reduce: BitwiseXORTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(175), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(175), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(175), // ,, reduce: BitwiseXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(175), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(175), // ||, reduce: BitwiseXORTerm
			reduce(175), // ^^, reduce: BitwiseXORTerm
			reduce(175), // &&, reduce: BitwiseXORTerm
			reduce(175), // ??, reduce: BitwiseXORTerm
			reduce(175), // ???, reduce: BitwiseXORTerm
			reduce(175), // =~, reduce: BitwiseXORTerm
			reduce(175), // !=~, reduce: BitwiseXORTerm
			reduce(175), // ==, reduce: BitwiseXORTerm
			reduce(175), // !=, reduce: BitwiseXORTerm
			reduce(175), // <=>, reduce: BitwiseXORTerm
			reduce(175), // >=, reduce: BitwiseXORTerm
			reduce(175), // <, reduce: BitwiseXORTerm
			reduce(175), // <=, reduce: BitwiseXORTerm
			reduce(175), // ^, reduce: BitwiseXORTerm
			shift(3026), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2085
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(177), // ;, reduce: BitwiseANDTerm
			nil,         // {
			reduce(177), // }, reduce: BitwiseANDTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(177), // >, reduce: BitwiseANDTerm
			shift(3027), // >>
			reduce(177), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(177), // ,, reduce: BitwiseANDTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(177), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(177), // ||, reduce: BitwiseANDTerm
			reduce(177), // ^^, reduce: BitwiseANDTerm
			reduce(177), // &&, reduce: BitwiseANDTerm
			reduce(177), // ??, reduce: BitwiseANDTerm
			reduce(177), // ???, reduce: BitwiseANDTerm
			reduce(177), // =~, reduce: BitwiseANDTerm
			reduce(177), // !=~, reduce: BitwiseANDTerm
			reduce(177), // ==, reduce: BitwiseANDTerm
			reduce(177), // !=, reduce: BitwiseANDTerm
			reduce(177), // <=>, reduce: BitwiseANDTerm
			reduce(177), // >=, reduce: BitwiseANDTerm
			reduce(177), // <, reduce: BitwiseANDTerm
			reduce(177), // <=, reduce: BitwiseANDTerm
			reduce(177), // ^, reduce: BitwiseANDTerm
			reduce(177), // &, reduce: BitwiseANDTerm
			shift(3028), // <<
			shift(3029), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2086
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(181), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			reduce(181), // }, reduce: BitwiseShiftTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(181), // >, reduce: BitwiseShiftTerm
			reduce(181), // >>, reduce: BitwiseShiftTerm
			reduce(181), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(181), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(181), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(181), // ||, reduce: BitwiseShiftTerm
			reduce(181), // ^^, reduce: BitwiseShiftTerm
			reduce(181), // &&, reduce: BitwiseShiftTerm
			reduce(181), // ??, reduce: BitwiseShiftTerm
			reduce(181), // ???, reduce: BitwiseShiftTerm
			reduce(181), // =~, reduce: BitwiseShiftTerm
			reduce(181), // !=~, reduce: BitwiseShiftTerm
			reduce(181), // ==, reduce: BitwiseShiftTerm
			reduce(181), // !=, reduce: BitwiseShiftTerm
			reduce(181), // <=>, reduce: BitwiseShiftTerm
			reduce(181), // >=, reduce: BitwiseShiftTerm
			reduce(181), // <, reduce: BitwiseShiftTerm
			reduce(181), // <=, reduce: BitwiseShiftTerm
			reduce(181), // ^, reduce: BitwiseShiftTerm
			reduce(181), // &, reduce: BitwiseShiftTerm
			reduce(181), // <<, reduce: BitwiseShiftTerm
			reduce(181), // >>>, reduce: BitwiseShiftTerm
			shift(3030), // +
			shift(3031), // -
			shift(3032), // .+
			shift(3033), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2087
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2088
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(186), // ;, reduce: AddsubdotTerm
			nil,         // {
			reduce(186), // }, reduce: AddsubdotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(186), // >, reduce: AddsubdotTerm
			reduce(186), // >>, reduce: AddsubdotTerm
			reduce(186), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(186), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(186), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(186), // ||, reduce: AddsubdotTerm
			reduce(186), // ^^, reduce: AddsubdotTerm
			reduce(186), // &&, reduce: AddsubdotTerm
			reduce(186), // ??, reduce: AddsubdotTerm
			reduce(186), // ???, reduce: AddsubdotTerm
			reduce(186), // =~, reduce: AddsubdotTerm
			reduce(186), // !=~, reduce: AddsubdotTerm
			reduce(186), // ==, reduce: AddsubdotTerm
			reduce(186), // !=, reduce: AddsubdotTerm
			reduce(186), // <=>, reduce: AddsubdotTerm
			reduce(186), // >=, reduce: AddsubdotTerm
			reduce(186), // <, reduce: AddsubdotTerm
			reduce(186), // <=, reduce: AddsubdotTerm
			reduce(186), // ^, reduce: AddsubdotTerm
			reduce(186), // &, reduce: AddsubdotTerm
			reduce(186), // <<, reduce: AddsubdotTerm
			reduce(186), // >>>, reduce: AddsubdotTerm
			reduce(186), // +, reduce: AddsubdotTerm
			reduce(186), // -, reduce: AddsubdotTerm
			reduce(186), // .+, reduce: AddsubdotTerm
			reduce(186), // .-, reduce: AddsubdotTerm
			shift(3035), // *
			shift(3036), // /
			shift(3037), // //
			shift(3038), // %
			shift(3039), // .*
			shift(3040), // ./
			shift(3041), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2089
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2090
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2091
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2092
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(194), // ;, reduce: MuldivTerm
			nil,         // {
			reduce(194), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(194), // >, reduce: MuldivTerm
			reduce(194), // >>, reduce: MuldivTerm
			reduce(194), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(194), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(194), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(194), // ||, reduce: MuldivTerm
			reduce(194), // ^^, reduce: MuldivTerm
			reduce(194), // &&, reduce: MuldivTerm
			reduce(194), // ??, reduce: MuldivTerm
			reduce(194), // ???, reduce: MuldivTerm
			reduce(194), // =~, reduce: MuldivTerm
			reduce(194), // !=~, reduce: MuldivTerm
			reduce(194), // ==, reduce: MuldivTerm
			reduce(194), // !=, reduce: MuldivTerm
			reduce(194), // <=>, reduce: MuldivTerm
			reduce(194), // >=, reduce: MuldivTerm
			reduce(194), // <, reduce: MuldivTerm
			reduce(194), // <=, reduce: MuldivTerm
			reduce(194), // ^, reduce: MuldivTerm
			reduce(194), // &, reduce: MuldivTerm
			reduce(194), // <<, reduce: MuldivTerm
			reduce(194), // >>>, reduce: MuldivTerm
			reduce(194), // +, reduce: MuldivTerm
			reduce(194), // -, reduce: MuldivTerm
			reduce(194), // .+, reduce: MuldivTerm
			reduce(194), // .-, reduce: MuldivTerm
			reduce(194), // *, reduce: MuldivTerm
			reduce(194), // /, reduce: MuldivTerm
			reduce(194), // //, reduce: MuldivTerm
			reduce(194), // %, reduce: MuldivTerm
			reduce(194), // .*, reduce: MuldivTerm
			reduce(194), // ./, reduce: MuldivTerm
			reduce(194), // .//, reduce: MuldivTerm
			shift(3045), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2093
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(196), // ;, reduce: DotTerm
			nil,         // {
			reduce(196), // }, reduce: DotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(196), // >, reduce: DotTerm
			reduce(196), // >>, reduce: DotTerm
			reduce(196), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(196), // ,, reduce: DotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(196), // ?, reduce: DotTerm
			nil,         // :
			reduce(196), // ||, reduce: DotTerm
			reduce(196), // ^^, reduce: DotTerm
			reduce(196), // &&, reduce: DotTerm
			reduce(196), // ??, reduce: DotTerm
			reduce(196), // ???, reduce: DotTerm
			reduce(196), // =~, reduce: DotTerm
			reduce(196), // !=~, reduce: DotTerm
			reduce(196), // ==, reduce: DotTerm
			reduce(196), // !=, reduce: DotTerm
			reduce(196), // <=>, reduce: DotTerm
			reduce(196), // >=, reduce: DotTerm
			reduce(196), // <, reduce: DotTerm
			reduce(196), // <=, reduce: DotTerm
			reduce(196), // ^, reduce: DotTerm
			reduce(196), // &, reduce: DotTerm
			reduce(196), // <<, reduce: DotTerm
			reduce(196), // >>>, reduce: DotTerm
			reduce(196), // +, reduce: DotTerm
			reduce(196), // -, reduce: DotTerm
			reduce(196), // .+, reduce: DotTerm
			reduce(196), // .-, reduce: DotTerm
			reduce(196), // *, reduce: DotTerm
			reduce(196), // /, reduce: DotTerm
			reduce(196), // //, reduce: DotTerm
			reduce(196), // %, reduce: DotTerm
			reduce(196), // .*, reduce: DotTerm
			reduce(196), // ./, reduce: DotTerm
			reduce(196), // .//, reduce: DotTerm
			reduce(196), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2094
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2095
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2096
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(203), // ;, reduce: UnaryOpTerm
			nil,         // {
			reduce(203), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(203), // >, reduce: UnaryOpTerm
			reduce(203), // >>, reduce: UnaryOpTerm
			reduce(203), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(203), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(203), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(203), // ||, reduce: UnaryOpTerm
			reduce(203), // ^^, reduce: UnaryOpTerm
			reduce(203), // &&, reduce: UnaryOpTerm
			reduce(203), // ??, reduce: UnaryOpTerm
			reduce(203), // ???, reduce: UnaryOpTerm
			reduce(203), // =~, reduce: UnaryOpTerm
			reduce(203), // !=~, reduce: UnaryOpTerm
			reduce(203), // ==, reduce: UnaryOpTerm
			reduce(203), // !=, reduce: UnaryOpTerm
			reduce(203), // <=>, reduce: UnaryOpTerm
			reduce(203), // >=, reduce: UnaryOpTerm
			reduce(203), // <, reduce: UnaryOpTerm
			reduce(203), // <=, reduce: UnaryOpTerm
			reduce(203), // ^, reduce: UnaryOpTerm
			reduce(203), // &, reduce: UnaryOpTerm
			reduce(203), // <<, reduce: UnaryOpTerm
			reduce(203), // >>>, reduce: UnaryOpTerm
			reduce(203), // +, reduce: UnaryOpTerm
			reduce(203), // -, reduce: UnaryOpTerm
			reduce(203), // .+, reduce: UnaryOpTerm
			reduce(203), // .-, reduce: UnaryOpTerm
			reduce(203), // *, reduce: UnaryOpTerm
			reduce(203), // /, reduce: UnaryOpTerm
			reduce(203), // //, reduce: UnaryOpTerm
			reduce(203), // %, reduce: UnaryOpTerm
			reduce(203), // .*, reduce: UnaryOpTerm
			reduce(203), // ./, reduce: UnaryOpTerm
			reduce(203), // .//, reduce: UnaryOpTerm
			reduce(203), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2097
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(207), // ;, reduce: PowTerm
			nil,         // {
			reduce(207), // }, reduce: PowTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(207), // >, reduce: PowTerm
			reduce(207), // >>, reduce: PowTerm
			reduce(207), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(207), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(207), // ?, reduce: PowTerm
			nil,         // :
			reduce(207), // ||, reduce: PowTerm
			reduce(207), // ^^, reduce: PowTerm
			reduce(207), // &&, reduce: PowTerm
			reduce(207), // ??, reduce: PowTerm
			reduce(207), // ???, reduce: PowTerm
			reduce(207), // =~, reduce: PowTerm
			reduce(207), // !=~, reduce: PowTerm
			reduce(207), // ==, reduce: PowTerm
			reduce(207), // !=, reduce: PowTerm
			reduce(207), // <=>, reduce: PowTerm
			reduce(207), // >=, reduce: PowTerm
			reduce(207), // <, reduce: PowTerm
			reduce(207), // <=, reduce: PowTerm
			reduce(207), // ^, reduce: PowTerm
			reduce(207), // &, reduce: PowTerm
			reduce(207), // <<, reduce: PowTerm
			reduce(207), // >>>, reduce: PowTerm
			reduce(207), // +, reduce: PowTerm
			reduce(207), // -, reduce: PowTerm
			reduce(207), // .+, reduce: PowTerm
			reduce(207), // .-, reduce: PowTerm
			reduce(207), // *, reduce: PowTerm
			reduce(207), // /, reduce: PowTerm
			reduce(207), // //, reduce: PowTerm
			reduce(207), // %, reduce: PowTerm
			reduce(207), // .*, reduce: PowTerm
			reduce(207), // ./, reduce: PowTerm
			reduce(207), // .//, reduce: PowTerm
			reduce(207), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			shift(3048), // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2098
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(209), // ;, reduce: PrecedenceChainEnd
			nil,         // {
			reduce(209), // }, reduce: PrecedenceChainEnd
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(209), // >, reduce: PrecedenceChainEnd
			reduce(209), // >>, reduce: PrecedenceChainEnd
			reduce(209), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(209), // ,, reduce: PrecedenceChainEnd
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(209), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(209), // ||, reduce: PrecedenceChainEnd
			reduce(209), // ^^, reduce: PrecedenceChainEnd
			reduce(209), // &&, reduce: PrecedenceChainEnd
			reduce(209), // ??, reduce: PrecedenceChainEnd
			reduce(209), // ???, reduce: PrecedenceChainEnd
			reduce(209), // =~, reduce: PrecedenceChainEnd
			reduce(209), // !=~, reduce: PrecedenceChainEnd
			reduce(209), // ==, reduce: PrecedenceChainEnd
			reduce(209), // !=, reduce: PrecedenceChainEnd
			reduce(209), // <=>, reduce: PrecedenceChainEnd
			reduce(209), // >=, reduce: PrecedenceChainEnd
			reduce(209), // <, reduce: PrecedenceChainEnd
			reduce(209), // <=, reduce: PrecedenceChainEnd
			reduce(209), // ^, reduce: PrecedenceChainEnd
			reduce(209), // &, reduce: PrecedenceChainEnd
			reduce(209), // <<, reduce: PrecedenceChainEnd
			reduce(209), // >>>, reduce: PrecedenceChainEnd
			reduce(209), // +, reduce: PrecedenceChainEnd
			reduce(209), // -, reduce: PrecedenceChainEnd
			reduce(209), // .+, reduce: PrecedenceChainEnd
			reduce(209), // .-, reduce: PrecedenceChainEnd
			reduce(209), // *, reduce: PrecedenceChainEnd
			reduce(209), // /, reduce: PrecedenceChainEnd
			reduce(209), // //, reduce: PrecedenceChainEnd
			reduce(209), // %, reduce: PrecedenceChainEnd
			reduce(209), // .*, reduce: PrecedenceChainEnd
			reduce(209), // ./, reduce: PrecedenceChainEnd
			reduce(209), // .//, reduce: PrecedenceChainEnd
			reduce(209), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(209), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(3049), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			shift(3050), // [[
			shift(3051), // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2099
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(215), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(215), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(215), // >, reduce: MlrvalOrFunction
			reduce(215), // >>, reduce: MlrvalOrFunction
			reduce(215), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(215), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(215), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(215), // ||, reduce: MlrvalOrFunction
			reduce(215), // ^^, reduce: MlrvalOrFunction
			reduce(215), // &&, reduce: MlrvalOrFunction
			reduce(215), // ??, reduce: MlrvalOrFunction
			reduce(215), // ???, reduce: MlrvalOrFunction
			reduce(215), // =~, reduce: MlrvalOrFunction
			reduce(215), // !=~, reduce: MlrvalOrFunction
			reduce(215), // ==, reduce: MlrvalOrFunction
			reduce(215), // !=, reduce: MlrvalOrFunction
			reduce(215), // <=>, reduce: MlrvalOrFunction
			reduce(215), // >=, reduce: MlrvalOrFunction
			reduce(215), // <, reduce: MlrvalOrFunction
			reduce(215), // <=, reduce: MlrvalOrFunction
			reduce(215), // ^, reduce: MlrvalOrFunction
			reduce(215), // &, reduce: MlrvalOrFunction
			reduce(215), // <<, reduce: MlrvalOrFunction
			reduce(215), // >>>, reduce: MlrvalOrFunction
			reduce(215), // +, reduce: MlrvalOrFunction
			reduce(215), // -, reduce: MlrvalOrFunction
			reduce(215), // .+, reduce: MlrvalOrFunction
			reduce(215), // .-, reduce: MlrvalOrFunction
			reduce(215), // *, reduce: MlrvalOrFunction
			reduce(215), // /, reduce: MlrvalOrFunction
			reduce(215), // //, reduce: MlrvalOrFunction
			reduce(215), // %, reduce: MlrvalOrFunction
			reduce(215), // .*, reduce: MlrvalOrFunction
			reduce(215), // ./, reduce: MlrvalOrFunction
			reduce(215), // .//, reduce: MlrvalOrFunction
			reduce(215), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(215), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(215), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(215), // [[, reduce: MlrvalOrFunction
			reduce(215), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2100
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(216), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(216), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(216), // >, reduce: MlrvalOrFunction
			reduce(216), // >>, reduce: MlrvalOrFunction
			reduce(216), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(216), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(216), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(216), // ||, reduce: MlrvalOrFunction
			reduce(216), // ^^, reduce: MlrvalOrFunction
			reduce(216), // &&, reduce: MlrvalOrFunction
			reduce(216), // ??, reduce: MlrvalOrFunction
			reduce(216), // ???, reduce: MlrvalOrFunction
			reduce(216), // =~, reduce: MlrvalOrFunction
			reduce(216), // !=~, reduce: MlrvalOrFunction
			reduce(216), // ==, reduce: MlrvalOrFunction
			reduce(216), // !=, reduce: MlrvalOrFunction
			reduce(216), // <=>, reduce: MlrvalOrFunction
			reduce(216), // >=, reduce: MlrvalOrFunction
			reduce(216), // <, reduce: MlrvalOrFunction
			reduce(216), // <=, reduce: MlrvalOrFunction
			reduce(216), // ^, reduce: MlrvalOrFunction
			reduce(216), // &, reduce: MlrvalOrFunction
			reduce(216), // <<, reduce: MlrvalOrFunction
			reduce(216), // >>>, reduce: MlrvalOrFunction
			reduce(216), // +, reduce: MlrvalOrFunction
			reduce(216), // -, reduce: MlrvalOrFunction
			reduce(216), // .+, reduce: MlrvalOrFunction
			reduce(216), // .-, reduce: MlrvalOrFunction
			reduce(216), // *, reduce: MlrvalOrFunction
			reduce(216), // /, reduce: MlrvalOrFunction
			reduce(216), // //, reduce: MlrvalOrFunction
			reduce(216), // %, reduce: MlrvalOrFunction
			reduce(216), // .*, reduce: MlrvalOrFunction
			reduce(216), // ./, reduce: MlrvalOrFunction
			reduce(216), // .//, reduce: MlrvalOrFunction
			reduce(216), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(216), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(216), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(216), // [[, reduce: MlrvalOrFunction
			reduce(216), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2101
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(217), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(217), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(217), // >, reduce: MlrvalOrFunction
			reduce(217), // >>, reduce: MlrvalOrFunction
			reduce(217), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(217), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(217), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(217), // ||, reduce: MlrvalOrFunction
			reduce(217), // ^^, reduce: MlrvalOrFunction
			reduce(217), // &&, reduce: MlrvalOrFunction
			reduce(217), // ??, reduce: MlrvalOrFunction
			reduce(217), // ???, reduce: MlrvalOrFunction
			reduce(217), // =~, reduce: MlrvalOrFunction
			reduce(217), // !=~, reduce: MlrvalOrFunction
			reduce(217), // ==, reduce: MlrvalOrFunction
			reduce(217), // !=, reduce: MlrvalOrFunction
			reduce(217), // <=>, reduce: MlrvalOrFunction
			reduce(217), // >=, reduce: MlrvalOrFunction
			reduce(217), // <, reduce: MlrvalOrFunction
			reduce(217), // <=, reduce: MlrvalOrFunction
			reduce(217), // ^, reduce: MlrvalOrFunction
			reduce(217), // &, reduce: MlrvalOrFunction
			reduce(217), // <<, reduce: MlrvalOrFunction
			reduce(217), // >>>, reduce: MlrvalOrFunction
			reduce(217), // +, reduce: MlrvalOrFunction
			reduce(217), // -, reduce: MlrvalOrFunction
			reduce(217), // .+, reduce: MlrvalOrFunction
			reduce(217), // .-, reduce: MlrvalOrFunction
			reduce(217), // *, reduce: MlrvalOrFunction
			reduce(217), // /, reduce: MlrvalOrFunction
			reduce(217), // //, reduce: MlrvalOrFunction
			reduce(217), // %, reduce: MlrvalOrFunction
			reduce(217), // .*, reduce: MlrvalOrFunction
			reduce(217), // ./, reduce: MlrvalOrFunction
			reduce(217), // .//, reduce: MlrvalOrFunction
			reduce(217), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(217), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(217), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(217), // [[, reduce: MlrvalOrFunction
			reduce(217), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2102
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(218), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(218), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(218), // >, reduce: MlrvalOrFunction
			reduce(218), // >>, reduce: MlrvalOrFunction
			reduce(218), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(218), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(218), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(218), // ||, reduce: MlrvalOrFunction
			reduce(218), // ^^, reduce: MlrvalOrFunction
			reduce(218), // &&, reduce: MlrvalOrFunction
			reduce(218), // ??, reduce: MlrvalOrFunction
			reduce(218), // ???, reduce: MlrvalOrFunction
			reduce(218), // =~, reduce: MlrvalOrFunction
			reduce(218), // !=~, reduce: MlrvalOrFunction
			reduce(218), // ==, reduce: MlrvalOrFunction
			reduce(218), // !=, reduce: MlrvalOrFunction
			reduce(218), // <=>, reduce: MlrvalOrFunction
			reduce(218), // >=, reduce: MlrvalOrFunction
			reduce(218), // <, reduce: MlrvalOrFunction
			reduce(218), // <=, reduce: MlrvalOrFunction
			reduce(218), // ^, reduce: MlrvalOrFunction
			reduce(218), // &, reduce: MlrvalOrFunction
			reduce(218), // <<, reduce: MlrvalOrFunction
			reduce(218), // >>>, reduce: MlrvalOrFunction
			reduce(218), // +, reduce: MlrvalOrFunction
			reduce(218), // -, reduce: MlrvalOrFunction
			reduce(218), // .+, reduce: MlrvalOrFunction
			reduce(218), // .-, reduce: MlrvalOrFunction
			reduce(218), // *, reduce: MlrvalOrFunction
			reduce(218), // /, reduce: MlrvalOrFunction
			reduce(218), // //, reduce: MlrvalOrFunction
			reduce(218), // %, reduce: MlrvalOrFunction
			reduce(218), // .*, reduce: MlrvalOrFunction
			reduce(218), // ./, reduce: MlrvalOrFunction
			reduce(218), // .//, reduce: MlrvalOrFunction
			reduce(218), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(218), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(218), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(218), // [[, reduce: MlrvalOrFunction
			reduce(218), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2103
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(219), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(219), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(219), // >, reduce: MlrvalOrFunction
			reduce(219), // >>, reduce: MlrvalOrFunction
			reduce(219), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(219), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(219), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(219), // ||, reduce: MlrvalOrFunction
			reduce(219), // ^^, reduce: MlrvalOrFunction
			reduce(219), // &&, reduce: MlrvalOrFunction
			reduce(219), // ??, reduce: MlrvalOrFunction
			reduce(219), // ???, reduce: MlrvalOrFunction
			reduce(219), // =~, reduce: MlrvalOrFunction
			reduce(219), // !=~, reduce: MlrvalOrFunction
			reduce(219), // ==, reduce: MlrvalOrFunction
			reduce(219), // !=, reduce: MlrvalOrFunction
			reduce(219), // <=>, reduce: MlrvalOrFunction
			reduce(219), // >=, reduce: MlrvalOrFunction
			reduce(219), // <, reduce: MlrvalOrFunction
			reduce(219), // <=, reduce: MlrvalOrFunction
			reduce(219), // ^, reduce: MlrvalOrFunction
			reduce(219), // &, reduce: MlrvalOrFunction
			reduce(219), // <<, reduce: MlrvalOrFunction
			reduce(219), // >>>, reduce: MlrvalOrFunction
			reduce(219), // +, reduce: MlrvalOrFunction
			reduce(219), // -, reduce: MlrvalOrFunction
			reduce(219), // .+, reduce: MlrvalOrFunction
			reduce(219), // .-, reduce: MlrvalOrFunction
			reduce(219), // *, reduce: MlrvalOrFunction
			reduce(219), // /, reduce: MlrvalOrFunction
			reduce(219), // //, reduce: MlrvalOrFunction
			reduce(219), // %, reduce: MlrvalOrFunction
			reduce(219), // .*, reduce: MlrvalOrFunction
			reduce(219), // ./, reduce: MlrvalOrFunction
			reduce(219), // .//, reduce: MlrvalOrFunction
			reduce(219), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(219), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(219), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(219), // [[, reduce: MlrvalOrFunction
			reduce(219), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2104
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(220), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(220), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(220), // >, reduce: MlrvalOrFunction
			reduce(220), // >>, reduce: MlrvalOrFunction
			reduce(220), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(220), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(220), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(220), // ||, reduce: MlrvalOrFunction
			reduce(220), // ^^, reduce: MlrvalOrFunction
			reduce(220), // &&, reduce: MlrvalOrFunction
			reduce(220), // ??, reduce: MlrvalOrFunction
			reduce(220), // ???, reduce: MlrvalOrFunction
			reduce(220), // =~, reduce: MlrvalOrFunction
			reduce(220), // !=~, reduce: MlrvalOrFunction
			reduce(220), // ==, reduce: MlrvalOrFunction
			reduce(220), // !=, reduce: MlrvalOrFunction
			reduce(220), // <=>, reduce: MlrvalOrFunction
			reduce(220), // >=, reduce: MlrvalOrFunction
			reduce(220), // <, reduce: MlrvalOrFunction
			reduce(220), // <=, reduce: MlrvalOrFunction
			reduce(220), // ^, reduce: MlrvalOrFunction
			reduce(220), // &, reduce: MlrvalOrFunction
			reduce(220), // <<, reduce: MlrvalOrFunction
			reduce(220), // >>>, reduce: MlrvalOrFunction
			reduce(220), // +, reduce: MlrvalOrFunction
			reduce(220), // -, reduce: MlrvalOrFunction
			reduce(220), // .+, reduce: MlrvalOrFunction
			reduce(220), // .-, reduce: MlrvalOrFunction
			reduce(220), // *, reduce: MlrvalOrFunction
			reduce(220), // /, reduce: MlrvalOrFunction
			reduce(220), // //, reduce: MlrvalOrFunction
			reduce(220), // %, reduce: MlrvalOrFunction
			reduce(220), // .*, reduce: MlrvalOrFunction
			reduce(220), // ./, reduce: MlrvalOrFunction
			reduce(220), // .//, reduce: MlrvalOrFunction
			reduce(220), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(220), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(220), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(220), // [[, reduce: MlrvalOrFunction
			reduce(220), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2105
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(221), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(221), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(221), // >, reduce: MlrvalOrFunction
			reduce(221), // >>, reduce: MlrvalOrFunction
			reduce(221), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(221), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(221), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(221), // ||, reduce: MlrvalOrFunction
			reduce(221), // ^^, reduce: MlrvalOrFunction
			reduce(221), // &&, reduce: MlrvalOrFunction
			reduce(221), // ??, reduce: MlrvalOrFunction
			reduce(221), // ???, reduce: MlrvalOrFunction
			reduce(221), // =~, reduce: MlrvalOrFunction
			reduce(221), // !=~, reduce: MlrvalOrFunction
			reduce(221), // ==, reduce: MlrvalOrFunction
			reduce(221), // !=, reduce: MlrvalOrFunction
			reduce(221), // <=>, reduce: MlrvalOrFunction
			reduce(221), // >=, reduce: MlrvalOrFunction
			reduce(221), // <, reduce: MlrvalOrFunction
			reduce(221), // <=, reduce: MlrvalOrFunction
			reduce(221), // ^, reduce: MlrvalOrFunction
			reduce(221), // &, reduce: MlrvalOrFunction
			reduce(221), // <<, reduce: MlrvalOrFunction
			reduce(221), // >>>, reduce: MlrvalOrFunction
			reduce(221), // +, reduce: MlrvalOrFunction
			reduce(221), // -, reduce: MlrvalOrFunction
			reduce(221), // .+, reduce: MlrvalOrFunction
			reduce(221), // .-, reduce: MlrvalOrFunction
			reduce(221), // *, reduce: MlrvalOrFunction
			reduce(221), // /, reduce: MlrvalOrFunction
			reduce(221), // //, reduce: MlrvalOrFunction
			reduce(221), // %, reduce: MlrvalOrFunction
			reduce(221), // .*, reduce: MlrvalOrFunction
			reduce(221), // ./, reduce: MlrvalOrFunction
			reduce(221), // .//, reduce: MlrvalOrFunction
			reduce(221), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(221), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(221), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(221), // [[, reduce: MlrvalOrFunction
			reduce(221), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2106
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(222), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(222), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(222), // >, reduce: MlrvalOrFunction
			reduce(222), // >>, reduce: MlrvalOrFunction
			reduce(222), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(222), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(222), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(222), // ||, reduce: MlrvalOrFunction
			reduce(222), // ^^, reduce: MlrvalOrFunction
			reduce(222), // &&, reduce: MlrvalOrFunction
			reduce(222), // ??, reduce: MlrvalOrFunction
			reduce(222), // ???, reduce: MlrvalOrFunction
			reduce(222), // =~, reduce: MlrvalOrFunction
			reduce(222), // !=~, reduce: MlrvalOrFunction
			reduce(222), // ==, reduce: MlrvalOrFunction
			reduce(222), // !=, reduce: MlrvalOrFunction
			reduce(222), // <=>, reduce: MlrvalOrFunction
			reduce(222), // >=, reduce: MlrvalOrFunction
			reduce(222), // <, reduce: MlrvalOrFunction
			reduce(222), // <=, reduce: MlrvalOrFunction
			reduce(222), // ^, reduce: MlrvalOrFunction
			reduce(222), // &, reduce: MlrvalOrFunction
			reduce(222), // <<, reduce: MlrvalOrFunction
			reduce(222), // >>>, reduce: MlrvalOrFunction
			reduce(222), // +, reduce: MlrvalOrFunction
			reduce(222), // -, reduce: MlrvalOrFunction
			reduce(222), // .+, reduce: MlrvalOrFunction
			reduce(222), // .-, reduce: MlrvalOrFunction
			reduce(222), // *, reduce: MlrvalOrFunction
			reduce(222), // /, reduce: MlrvalOrFunction
			reduce(222), // //, reduce: MlrvalOrFunction
			reduce(222), // %, reduce: MlrvalOrFunction
			reduce(222), // .*, reduce: MlrvalOrFunction
			reduce(222), // ./, reduce: MlrvalOrFunction
			reduce(222), // .//, reduce: MlrvalOrFunction
			reduce(222), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(222), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(222), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(222), // [[, reduce: MlrvalOrFunction
			reduce(222), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2107
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(223), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(223), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(223), // >, reduce: MlrvalOrFunction
			reduce(223), // >>, reduce: MlrvalOrFunction
			reduce(223), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(223), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(223), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(223), // ||, reduce: MlrvalOrFunction
			reduce(223), // ^^, reduce: MlrvalOrFunction
			reduce(223), // &&, reduce: MlrvalOrFunction
			reduce(223), // ??, reduce: MlrvalOrFunction
			reduce(223), // ???, reduce: MlrvalOrFunction
			reduce(223), // =~, reduce: MlrvalOrFunction
			reduce(223), // !=~, reduce: MlrvalOrFunction
			reduce(223), // ==, reduce: MlrvalOrFunction
			reduce(223), // !=, reduce: MlrvalOrFunction
			reduce(223), // <=>, reduce: MlrvalOrFunction
			reduce(223), // >=, reduce: MlrvalOrFunction
			reduce(223), // <, reduce: MlrvalOrFunction
			reduce(223), // <=, reduce: MlrvalOrFunction
			reduce(223), // ^, reduce: MlrvalOrFunction
			reduce(223), // &, reduce: MlrvalOrFunction
			reduce(223), // <<, reduce: MlrvalOrFunction
			reduce(223), // >>>, reduce: MlrvalOrFunction
			reduce(223), // +, reduce: MlrvalOrFunction
			reduce(223), // -, reduce: MlrvalOrFunction
			reduce(223), // .+, reduce: MlrvalOrFunction
			reduce(223), // .-, reduce: MlrvalOrFunction
			reduce(223), // *, reduce: MlrvalOrFunction
			reduce(223), // /, reduce: MlrvalOrFunction
			reduce(223), // //, reduce: MlrvalOrFunction
			reduce(223), // %, reduce: MlrvalOrFunction
			reduce(223), // .*, reduce: MlrvalOrFunction
			reduce(223), // ./, reduce: MlrvalOrFunction
			reduce(223), // .//, reduce: MlrvalOrFunction
			reduce(223), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(223), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(223), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(223), // [[, reduce: MlrvalOrFunction
			reduce(223), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2108
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(224), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(224), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(224), // >, reduce: MlrvalOrFunction
			reduce(224), // >>, reduce: MlrvalOrFunction
			reduce(224), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(224), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(224), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(224), // ||, reduce: MlrvalOrFunction
			reduce(224), // ^^, reduce: MlrvalOrFunction
			reduce(224), // &&, reduce: MlrvalOrFunction
			reduce(224), // ??, reduce: MlrvalOrFunction
			reduce(224), // ???, reduce: MlrvalOrFunction
			reduce(224), // =~, reduce: MlrvalOrFunction
			reduce(224), // !=~, reduce: MlrvalOrFunction
			reduce(224), // ==, reduce: MlrvalOrFunction
			reduce(224), // !=, reduce: MlrvalOrFunction
			reduce(224), // <=>, reduce: MlrvalOrFunction
			reduce(224), // >=, reduce: MlrvalOrFunction
			reduce(224), // <, reduce: MlrvalOrFunction
			reduce(224), // <=, reduce: MlrvalOrFunction
			reduce(224), // ^, reduce: MlrvalOrFunction
			reduce(224), // &, reduce: MlrvalOrFunction
			reduce(224), // <<, reduce: MlrvalOrFunction
			reduce(224), // >>>, reduce: MlrvalOrFunction
			reduce(224), // +, reduce: MlrvalOrFunction
			reduce(224), // -, reduce: MlrvalOrFunction
			reduce(224), // .+, reduce: MlrvalOrFunction
			reduce(224), // .-, reduce: MlrvalOrFunction
			reduce(224), // *, reduce: MlrvalOrFunction
			reduce(224), // /, reduce: MlrvalOrFunction
			reduce(224), // //, reduce: MlrvalOrFunction
			reduce(224), // %, reduce: MlrvalOrFunction
			reduce(224), // .*, reduce: MlrvalOrFunction
			reduce(224), // ./, reduce: MlrvalOrFunction
			reduce(224), // .//, reduce: MlrvalOrFunction
			reduce(224), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(224), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(224), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(224), // [[, reduce: MlrvalOrFunction
			reduce(224), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2109
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(225), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(225), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(225), // >, reduce: MlrvalOrFunction
			reduce(225), // >>, reduce: MlrvalOrFunction
			reduce(225), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(225), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(225), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(225), // ||, reduce: MlrvalOrFunction
			reduce(225), // ^^, reduce: MlrvalOrFunction
			reduce(225), // &&, reduce: MlrvalOrFunction
			reduce(225), // ??, reduce: MlrvalOrFunction
			reduce(225), // ???, reduce: MlrvalOrFunction
			reduce(225), // =~, reduce: MlrvalOrFunction
			reduce(225), // !=~, reduce: MlrvalOrFunction
			reduce(225), // ==, reduce: MlrvalOrFunction
			reduce(225), // !=, reduce: MlrvalOrFunction
			reduce(225), // <=>, reduce: MlrvalOrFunction
			reduce(225), // >=, reduce: MlrvalOrFunction
			reduce(225), // <, reduce: MlrvalOrFunction
			reduce(225), // <=, reduce: MlrvalOrFunction
			reduce(225), // ^, reduce: MlrvalOrFunction
			reduce(225), // &, reduce: MlrvalOrFunction
			reduce(225), // <<, reduce: MlrvalOrFunction
			reduce(225), // >>>, reduce: MlrvalOrFunction
			reduce(225), // +, reduce: MlrvalOrFunction
			reduce(225), // -, reduce: MlrvalOrFunction
			reduce(225), // .+, reduce: MlrvalOrFunction
			reduce(225), // .-, reduce: MlrvalOrFunction
			reduce(225), // *, reduce: MlrvalOrFunction
			reduce(225), // /, reduce: MlrvalOrFunction
			reduce(225), // //, reduce: MlrvalOrFunction
			reduce(225), // %, reduce: MlrvalOrFunction
			reduce(225), // .*, reduce: MlrvalOrFunction
			reduce(225), // ./, reduce: MlrvalOrFunction
			reduce(225), // .//, reduce: MlrvalOrFunction
			reduce(225), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(225), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(225), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(225), // [[, reduce: MlrvalOrFunction
			reduce(225), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2110
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(226), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(226), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(226), // >, reduce: MlrvalOrFunction
			reduce(226), // >>, reduce: MlrvalOrFunction
			reduce(226), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(226), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(226), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(226), // ||, reduce: MlrvalOrFunction
			reduce(226), // ^^, reduce: MlrvalOrFunction
			reduce(226), // &&, reduce: MlrvalOrFunction
			reduce(226), // ??, reduce: MlrvalOrFunction
			reduce(226), // ???, reduce: MlrvalOrFunction
			reduce(226), // =~, reduce: MlrvalOrFunction
			reduce(226), // !=~, reduce: MlrvalOrFunction
			reduce(226), // ==, reduce: MlrvalOrFunction
			reduce(226), // !=, reduce: MlrvalOrFunction
			reduce(226), // <=>, reduce: MlrvalOrFunction
			reduce(226), // >=, reduce: MlrvalOrFunction
			reduce(226), // <, reduce: MlrvalOrFunction
			reduce(226), // <=, reduce: MlrvalOrFunction
			reduce(226), // ^, reduce: MlrvalOrFunction
			reduce(226), // &, reduce: MlrvalOrFunction
			reduce(226), // <<, reduce: MlrvalOrFunction
			reduce(226), // >>>, reduce: MlrvalOrFunction
			reduce(226), // +, reduce: MlrvalOrFunction
			reduce(226), // -, reduce: MlrvalOrFunction
			reduce(226), // .+, reduce: MlrvalOrFunction
			reduce(226), // .-, reduce: MlrvalOrFunction
			reduce(226), // *, reduce: MlrvalOrFunction
			reduce(226), // /, reduce: MlrvalOrFunction
			reduce(226), // //, reduce: MlrvalOrFunction
			reduce(226), // %, reduce: MlrvalOrFunction
			reduce(226), // .*, reduce: MlrvalOrFunction
			reduce(226), // ./, reduce: MlrvalOrFunction
			reduce(226), // .//, reduce: MlrvalOrFunction
			reduce(226), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(226), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(226), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(226), // [[, reduce: MlrvalOrFunction
			reduce(226), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2111
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(227), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(227), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(227), // >, reduce: MlrvalOrFunction
			reduce(227), // >>, reduce: MlrvalOrFunction
			reduce(227), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(227), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(227), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(227), // ||, reduce: MlrvalOrFunction
			reduce(227), // ^^, reduce: MlrvalOrFunction
			reduce(227), // &&, reduce: MlrvalOrFunction
			reduce(227), // ??, reduce: MlrvalOrFunction
			reduce(227), // ???, reduce: MlrvalOrFunction
			reduce(227), // =~, reduce: MlrvalOrFunction
			reduce(227), // !=~, reduce: MlrvalOrFunction
			reduce(227), // ==, reduce: MlrvalOrFunction
			reduce(227), // !=, reduce: MlrvalOrFunction
			reduce(227), // <=>, reduce: MlrvalOrFunction
			reduce(227), // >=, reduce: MlrvalOrFunction
			reduce(227), // <, reduce: MlrvalOrFunction
			reduce(227), // <=, reduce: MlrvalOrFunction
			reduce(227), // ^, reduce: MlrvalOrFunction
			reduce(227), // &, reduce: MlrvalOrFunction
			reduce(227), // <<, reduce: MlrvalOrFunction
			reduce(227), // >>>, reduce: MlrvalOrFunction
			reduce(227), // +, reduce: MlrvalOrFunction
			reduce(227), // -, reduce: MlrvalOrFunction
			reduce(227), // .+, reduce: MlrvalOrFunction
			reduce(227), // .-, reduce: MlrvalOrFunction
			reduce(227), // *, reduce: MlrvalOrFunction
			reduce(227), // /, reduce: MlrvalOrFunction
			reduce(227), // //, reduce: MlrvalOrFunction
			reduce(227), // %, reduce: MlrvalOrFunction
			reduce(227), // .*, reduce: MlrvalOrFunction
			reduce(227), // ./, reduce: MlrvalOrFunction
			reduce(227), // .//, reduce: MlrvalOrFunction
			reduce(227), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(227), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(227), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(227), // [[, reduce: MlrvalOrFunction
			reduce(227), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2112
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(752),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(756),  // (
			nil,         // )
			nil,         // emitp
			shift(768),  // field_name
			shift(769),  // $[
			shift(3052), // ]
			shift(771),  // braced_field_name
			shift(772),  // $[[
			shift(773),  // $[[[
			shift(774),  // full_srec
			shift(776),  // oosvar_name
			shift(777),  // @[
			shift(778),  // braced_oosvar_name
			shift(779),  // full_oosvar
			shift(780),  // all
			shift(781),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(796),  // +
			shift(798),  // -
			shift(799),  // .+
			shift(800),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(803),  // !
			shift(804),  // ~
			nil,         // **
			shift(809),  // string_literal
			shift(810),  // regex_case_insensitive
			shift(811),  // int_literal
			shift(812),  // float_literal
			shift(813),  // boolean_literal
			shift(814),  // null_literal
			shift(815),  // inf_literal
			shift(816),  // nan_literal
			shift(817),  // const_M_PI
			shift(818),  // const_M_E
			shift(819),  // panic
			shift(821),  // [
			shift(824),  // ctx_IPS
			shift(825),  // ctx_IFS
			shift(826),  // ctx_IRS
			shift(827),  // ctx_OPS
			shift(828),  // ctx_OFS
			shift(829),  // ctx_ORS
			shift(830),  // ctx_FLATSEP
			shift(831),  // ctx_NF
			shift(832),  // ctx_NR
			shift(833),  // ctx_FNR
			shift(834),  // ctx_FILENAME
			shift(835),  // ctx_FILENUM
			shift(837),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(844),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2113
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(240), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(240), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(240), // >, reduce: MlrvalOrFunction
			reduce(240), // >>, reduce: MlrvalOrFunction
			reduce(240), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(240), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(240), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(240), // ||, reduce: MlrvalOrFunction
			reduce(240), // ^^, reduce: MlrvalOrFunction
			reduce(240), // &&, reduce: MlrvalOrFunction
			reduce(240), // ??, reduce: MlrvalOrFunction
			reduce(240), // ???, reduce: MlrvalOrFunction
			reduce(240), // =~, reduce: MlrvalOrFunction
			reduce(240), // !=~, reduce: MlrvalOrFunction
			reduce(240), // ==, reduce: MlrvalOrFunction
			reduce(240), // !=, reduce: MlrvalOrFunction
			reduce(240), // <=>, reduce: MlrvalOrFunction
			reduce(240), // >=, reduce: MlrvalOrFunction
			reduce(240), // <, reduce: MlrvalOrFunction
			reduce(240), // <=, reduce: MlrvalOrFunction
			reduce(240), // ^, reduce: MlrvalOrFunction
			reduce(240), // &, reduce: MlrvalOrFunction
			reduce(240), // <<, reduce: MlrvalOrFunction
			reduce(240), // >>>, reduce: MlrvalOrFunction
			reduce(240), // +, reduce: MlrvalOrFunction
			reduce(240), // -, reduce: MlrvalOrFunction
			reduce(240), // .+, reduce: MlrvalOrFunction
			reduce(240), // .-, reduce: MlrvalOrFunction
			reduce(240), // *, reduce: MlrvalOrFunction
			reduce(240), // /, reduce: MlrvalOrFunction
			reduce(240), // //, reduce: MlrvalOrFunction
			reduce(240), // %, reduce: MlrvalOrFunction
			reduce(240), // .*, reduce: MlrvalOrFunction
			reduce(240), // ./, reduce: MlrvalOrFunction
			reduce(240), // .//, reduce: MlrvalOrFunction
			reduce(240), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(240), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(240), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(240), // [[, reduce: MlrvalOrFunction
			reduce(240), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2114
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(241), // ;, reduce: ContextVariable
			nil,         // {
			reduce(241), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(241), // >, reduce: ContextVariable
			reduce(241), // >>, reduce: ContextVariable
			reduce(241), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(241), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(241), // ?, reduce: ContextVariable
			nil,         // :
			reduce(241), // ||, reduce: ContextVariable
			reduce(241), // ^^, reduce: ContextVariable
			reduce(241), // &&, reduce: ContextVariable
			reduce(241), // ??, reduce: ContextVariable
			reduce(241), // ???, reduce: ContextVariable
			reduce(241), // =~, reduce: ContextVariable
			reduce(241), // !=~, reduce: ContextVariable
			reduce(241), // ==, reduce: ContextVariable
			reduce(241), // !=, reduce: ContextVariable
			reduce(241), // <=>, reduce: ContextVariable
			reduce(241), // >=, reduce: ContextVariable
			reduce(241), // <, reduce: ContextVariable
			reduce(241), // <=, reduce: ContextVariable
			reduce(241), // ^, reduce: ContextVariable
			reduce(241), // &, reduce: ContextVariable
			reduce(241), // <<, reduce: ContextVariable
			reduce(241), // >>>, reduce: ContextVariable
			reduce(241), // +, reduce: ContextVariable
			reduce(241), // -, reduce: ContextVariable
			reduce(241), // .+, reduce: ContextVariable
			reduce(241), // .-, reduce: ContextVariable
			reduce(241), // *, reduce: ContextVariable
			reduce(241), // /, reduce: ContextVariable
			reduce(241), // //, reduce: ContextVariable
			reduce(241), // %, reduce: ContextVariable
			reduce(241), // .*, reduce: ContextVariable
			reduce(241), // ./, reduce: ContextVariable
			reduce(241), // .//, reduce: ContextVariable
			reduce(241), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(241), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(241), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(241), // [[, reduce: ContextVariable
			reduce(241), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2115
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(242), // ;, reduce: ContextVariable
			nil,         // {
			reduce(242), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(242), // >, reduce: ContextVariable
			reduce(242), // >>, reduce: ContextVariable
			reduce(242), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(242), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(242), // ?, reduce: ContextVariable
			nil,         // :
			reduce(242), // ||, reduce: ContextVariable
			reduce(242), // ^^, reduce: ContextVariable
			reduce(242), // &&, reduce: ContextVariable
			reduce(242), // ??, reduce: ContextVariable
			reduce(242), // ???, reduce: ContextVariable
			reduce(242), // =~, reduce: ContextVariable
			reduce(242), // !=~, reduce: ContextVariable
			reduce(242), // ==, reduce: ContextVariable
			reduce(242), // !=, reduce: ContextVariable
			reduce(242), // <=>, reduce: ContextVariable
			reduce(242), // >=, reduce: ContextVariable
			reduce(242), // <, reduce: ContextVariable
			reduce(242), // <=, reduce: ContextVariable
			reduce(242), // ^, reduce: ContextVariable
			reduce(242), // &, reduce: ContextVariable
			reduce(242), // <<, reduce: ContextVariable
			reduce(242), // >>>, reduce: ContextVariable
			reduce(242), // +, reduce: ContextVariable
			reduce(242), // -, reduce: ContextVariable
			reduce(242), // .+, reduce: ContextVariable
			reduce(242), // .-, reduce: ContextVariable
			reduce(242), // *, reduce: ContextVariable
			reduce(242), // /, reduce: ContextVariable
			reduce(242), // //, reduce: ContextVariable
			reduce(242), // %, reduce: ContextVariable
			reduce(242), // .*, reduce: ContextVariable
			reduce(242), // ./, reduce: ContextVariable
			reduce(242), // .//, reduce: ContextVariable
			reduce(242), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(242), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(242), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(242), // [[, reduce: ContextVariable
			reduce(242), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2116
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(243), // ;, reduce: ContextVariable
			nil,         // {
			reduce(243), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(243), // >, reduce: ContextVariable
			reduce(243), // >>, reduce: ContextVariable
			reduce(243), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(243), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(243), // ?, reduce: ContextVariable
			nil,         // :
			reduce(243), // ||, reduce: ContextVariable
			reduce(243), // ^^, reduce: ContextVariable
			reduce(243), // &&, reduce: ContextVariable
			reduce(243), // ??, reduce: ContextVariable
			reduce(243), // ???, reduce: ContextVariable
			reduce(243), // =~, reduce: ContextVariable
			reduce(243), // !=~, reduce: ContextVariable
			reduce(243), // ==, reduce: ContextVariable
			reduce(243), // !=, reduce: ContextVariable
			reduce(243), // <=>, reduce: ContextVariable
			reduce(243), // >=, reduce: ContextVariable
			reduce(243), // <, reduce: ContextVariable
			reduce(243), // <=, reduce: ContextVariable
			reduce(243), // ^, reduce: ContextVariable
			reduce(243), // &, reduce: ContextVariable
			reduce(243), // <<, reduce: ContextVariable
			reduce(243), // >>>, reduce: ContextVariable
			reduce(243), // +, reduce: ContextVariable
			reduce(243), // -, reduce: ContextVariable
			reduce(243), // .+, reduce: ContextVariable
			reduce(243), // .-, reduce: ContextVariable
			reduce(243), // *, reduce: ContextVariable
			reduce(243), // /, reduce: ContextVariable
			reduce(243), // //, reduce: ContextVariable
			reduce(243), // %, reduce: ContextVariable
			reduce(243), // .*, reduce: ContextVariable
			reduce(243), // ./, reduce: ContextVariable
			reduce(243), // .//, reduce: ContextVariable
			reduce(243), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(243), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(243), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(243), // [[, reduce: ContextVariable
			reduce(243), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2117
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(244), // ;, reduce: ContextVariable
			nil,         // {
			reduce(244), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(244), // >, reduce: ContextVariable
			reduce(244), // >>, reduce: ContextVariable
			reduce(244), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(244), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(244), // ?, reduce: ContextVariable
			nil,         // :
			reduce(244), // ||, reduce: ContextVariable
			reduce(244), // ^^, reduce: ContextVariable
			reduce(244), // &&, reduce: ContextVariable
			reduce(244), // ??, reduce: ContextVariable
			reduce(244), // ???, reduce: ContextVariable
			reduce(244), // =~, reduce: ContextVariable
			reduce(244), // !=~, reduce: ContextVariable
			reduce(244), // ==, reduce: ContextVariable
			reduce(244), // !=, reduce: ContextVariable
			reduce(244), // <=>, reduce: ContextVariable
			reduce(244), // >=, reduce: ContextVariable
			reduce(244), // <, reduce: ContextVariable
			reduce(244), // <=, reduce: ContextVariable
			reduce(244), // ^, reduce: ContextVariable
			reduce(244), // &, reduce: ContextVariable
			reduce(244), // <<, reduce: ContextVariable
			reduce(244), // >>>, reduce: ContextVariable
			reduce(244), // +, reduce: ContextVariable
			reduce(244), // -, reduce: ContextVariable
			reduce(244), // .+, reduce: ContextVariable
			reduce(244), // .-, reduce: ContextVariable
			reduce(244), // *, reduce: ContextVariable
			reduce(244), // /, reduce: ContextVariable
			reduce(244), // //, reduce: ContextVariable
			reduce(244), // %, reduce: ContextVariable
			reduce(244), // .*, reduce: ContextVariable
			reduce(244), // ./, reduce: ContextVariable
			reduce(244), // .//, reduce: ContextVariable
			reduce(244), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(244), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(244), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(244), // [[, reduce: ContextVariable
			reduce(244), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2118
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(245), // ;, reduce: ContextVariable
			nil,         // {
			reduce(245), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(245), // >, reduce: ContextVariable
			reduce(245), // >>, reduce: ContextVariable
			reduce(245), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(245), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(245), // ?, reduce: ContextVariable
			nil,         // :
			reduce(245), // ||, reduce: ContextVariable
			reduce(245), // ^^, reduce: ContextVariable
			reduce(245), // &&, reduce: ContextVariable
			reduce(245), // ??, reduce: ContextVariable
			reduce(245), // ???, reduce: ContextVariable
			reduce(245), // =~, reduce: ContextVariable
			reduce(245), // !=~, reduce: ContextVariable
			reduce(245), // ==, reduce: ContextVariable
			reduce(245), // !=, reduce: ContextVariable
			reduce(245), // <=>, reduce: ContextVariable
			reduce(245), // >=, reduce: ContextVariable
			reduce(245), // <, reduce: ContextVariable
			reduce(245), // <=, reduce: ContextVariable
			reduce(245), // ^, reduce: ContextVariable
			reduce(245), // &, reduce: ContextVariable
			reduce(245), // <<, reduce: ContextVariable
			reduce(245), // >>>, reduce: ContextVariable
			reduce(245), // +, reduce: ContextVariable
			reduce(245), // -, reduce: ContextVariable
			reduce(245), // .+, reduce: ContextVariable
			reduce(245), // .-, reduce: ContextVariable
			reduce(245), // *, reduce: ContextVariable
			reduce(245), // /, reduce: ContextVariable
			reduce(245), // //, reduce: ContextVariable
			reduce(245), // %, reduce: ContextVariable
			reduce(245), // .*, reduce: ContextVariable
			reduce(245), // ./, reduce: ContextVariable
			reduce(245), // .//, reduce: ContextVariable
			reduce(245), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(245), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(245), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(245), // [[, reduce: ContextVariable
			reduce(245), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2119
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(246), // ;, reduce: ContextVariable
			nil,         // {
			reduce(246), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(246), // >, reduce: ContextVariable
			reduce(246), // >>, reduce: ContextVariable
			reduce(246), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(246), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(246), // ?, reduce: ContextVariable
			nil,         // :
			reduce(246), // ||, reduce: ContextVariable
			reduce(246), // ^^, reduce: ContextVariable
			reduce(246), // &&, reduce: ContextVariable
			reduce(246), // ??, reduce: ContextVariable
			reduce(246), // ???, reduce: ContextVariable
			reduce(246), // =~, reduce: ContextVariable
			reduce(246), // !=~, reduce: ContextVariable
			reduce(246), // ==, reduce: ContextVariable
			reduce(246), // !=, reduce: ContextVariable
			reduce(246), // <=>, reduce: ContextVariable
			reduce(246), // >=, reduce: ContextVariable
			reduce(246), // <, reduce: ContextVariable
			reduce(246), // <=, reduce: ContextVariable
			reduce(246), // ^, reduce: ContextVariable
			reduce(246), // &, reduce: ContextVariable
			reduce(246), // <<, reduce: ContextVariable
			reduce(246), // >>>, reduce: ContextVariable
			reduce(246), // +, reduce: ContextVariable
			reduce(246), // -, reduce: ContextVariable
			reduce(246), // .+, reduce: ContextVariable
			reduce(246), // .-, reduce: ContextVariable
			reduce(246), // *, reduce: ContextVariable
			reduce(246), // /, reduce: ContextVariable
			reduce(246), // //, reduce: ContextVariable
			reduce(246), // %, reduce: ContextVariable
			reduce(246), // .*, reduce: ContextVariable
			reduce(246), // ./, reduce: ContextVariable
			reduce(246), // .//, reduce: ContextVariable
			reduce(246), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(246), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(246), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(246), // [[, reduce: ContextVariable
			reduce(246), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2120
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(247), // ;, reduce: ContextVariable
			nil,         // {
			reduce(247), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(247), // >, reduce: ContextVariable
			reduce(247), // >>, reduce: ContextVariable
			reduce(247), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(247), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(247), // ?, reduce: ContextVariable
			nil,         // :
			reduce(247), // ||, reduce: ContextVariable
			reduce(247), // ^^, reduce: ContextVariable
			reduce(247), // &&, reduce: ContextVariable
			reduce(247), // ??, reduce: ContextVariable
			reduce(247), // ???, reduce: ContextVariable
			reduce(247), // =~, reduce: ContextVariable
			reduce(247), // !=~, reduce: ContextVariable
			reduce(247), // ==, reduce: ContextVariable
			reduce(247), // !=, reduce: ContextVariable
			reduce(247), // <=>, reduce: ContextVariable
			reduce(247), // >=, reduce: ContextVariable
			reduce(247), // <, reduce: ContextVariable
			reduce(247), // <=, reduce: ContextVariable
			reduce(247), // ^, reduce: ContextVariable
			reduce(247), // &, reduce: ContextVariable
			reduce(247), // <<, reduce: ContextVariable
			reduce(247), // >>>, reduce: ContextVariable
			reduce(247), // +, reduce: ContextVariable
			reduce(247), // -, reduce: ContextVariable
			reduce(247), // .+, reduce: ContextVariable
			reduce(247), // .-, reduce: ContextVariable
			reduce(247), // *, reduce: ContextVariable
			reduce(247), // /, reduce: ContextVariable
			reduce(247), // //, reduce: ContextVariable
			reduce(247), // %, reduce: ContextVariable
			reduce(247), // .*, reduce: ContextVariable
			reduce(247), // ./, reduce: ContextVariable
			reduce(247), // .//, reduce: ContextVariable
			reduce(247), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(247), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(247), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(247), // [[, reduce: ContextVariable
			reduce(247), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2121
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(248), // ;, reduce: ContextVariable
			nil,         // {
			reduce(248), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(248), // >, reduce: ContextVariable
			reduce(248), // >>, reduce: ContextVariable
			reduce(248), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(248), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(248), // ?, reduce: ContextVariable
			nil,         // :
			reduce(248), // ||, reduce: ContextVariable
			reduce(248), // ^^, reduce: ContextVariable
			reduce(248), // &&, reduce: ContextVariable
			reduce(248), // ??, reduce: ContextVariable
			reduce(248), // ???, reduce: ContextVariable
			reduce(248), // =~, reduce: ContextVariable
			reduce(248), // !=~, reduce: ContextVariable
			reduce(248), // ==, reduce: ContextVariable
			reduce(248), // !=, reduce: ContextVariable
			reduce(248), // <=>, reduce: ContextVariable
			reduce(248), // >=, reduce: ContextVariable
			reduce(248), // <, reduce: ContextVariable
			reduce(248), // <=, reduce: ContextVariable
			reduce(248), // ^, reduce: ContextVariable
			reduce(248), // &, reduce: ContextVariable
			reduce(248), // <<, reduce: ContextVariable
			reduce(248), // >>>, reduce: ContextVariable
			reduce(248), // +, reduce: ContextVariable
			reduce(248), // -, reduce: ContextVariable
			reduce(248), // .+, reduce: ContextVariable
			reduce(248), // .-, reduce: ContextVariable
			reduce(248), // *, reduce: ContextVariable
			reduce(248), // /, reduce: ContextVariable
			reduce(248), // //, reduce: ContextVariable
			reduce(248), // %, reduce: ContextVariable
			reduce(248), // .*, reduce: ContextVariable
			reduce(248), // ./, reduce: ContextVariable
			reduce(248), // .//, reduce: ContextVariable
			reduce(248), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(248), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(248), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(248), // [[, reduce: ContextVariable
			reduce(248), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2122
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(249), // ;, reduce: ContextVariable
			nil,         // {
			reduce(249), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(249), // >, reduce: ContextVariable
			reduce(249), // >>, reduce: ContextVariable
			reduce(249), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(249), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(249), // ?, reduce: ContextVariable
			nil,         // :
			reduce(249), // ||, reduce: ContextVariable
			reduce(249), // ^^, reduce: ContextVariable
			reduce(249), // &&, reduce: ContextVariable
			reduce(249), // ??, reduce: ContextVariable
			reduce(249), // ???, reduce: ContextVariable
			reduce(249), // =~, reduce: ContextVariable
			reduce(249), // !=~, reduce: ContextVariable
			reduce(249), // ==, reduce: ContextVariable
			reduce(249), // !=, reduce: ContextVariable
			reduce(249), // <=>, reduce: ContextVariable
			reduce(249), // >=, reduce: ContextVariable
			reduce(249), // <, reduce: ContextVariable
			reduce(249), // <=, reduce: ContextVariable
			reduce(249), // ^, reduce: ContextVariable
			reduce(249), // &, reduce: ContextVariable
			reduce(249), // <<, reduce: ContextVariable
			reduce(249), // >>>, reduce: ContextVariable
			reduce(249), // +, reduce: ContextVariable
			reduce(249), // -, reduce: ContextVariable
			reduce(249), // .+, reduce: ContextVariable
			reduce(249), // .-, reduce: ContextVariable
			reduce(249), // *, reduce: ContextVariable
			reduce(249), // /, reduce: ContextVariable
			reduce(249), // //, reduce: ContextVariable
			reduce(249), // %, reduce: ContextVariable
			reduce(249), // .*, reduce: ContextVariable
			reduce(249), // ./, reduce: ContextVariable
			reduce(249), // .//, reduce: ContextVariable
			reduce(249), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(249), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(249), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(249), // [[, reduce: ContextVariable
			reduce(249), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2123
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(250), // ;, reduce: ContextVariable
			nil,         // {
			reduce(250), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(250), // >, reduce: ContextVariable
			reduce(250), // >>, reduce: ContextVariable
			reduce(250), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(250), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(250), // ?, reduce: ContextVariable
			nil,         // :
			reduce(250), // ||, reduce: ContextVariable
			reduce(250), // ^^, reduce: ContextVariable
			reduce(250), // &&, reduce: ContextVariable
			reduce(250), // ??, reduce: ContextVariable
			reduce(250), // ???, reduce: ContextVariable
			reduce(250), // =~, reduce: ContextVariable
			reduce(250), // !=~, reduce: ContextVariable
			reduce(250), // ==, reduce: ContextVariable
			reduce(250), // !=, reduce: ContextVariable
			reduce(250), // <=>, reduce: ContextVariable
			reduce(250), // >=, reduce: ContextVariable
			reduce(250), // <, reduce: ContextVariable
			reduce(250), // <=, reduce: ContextVariable
			reduce(250), // ^, reduce: ContextVariable
			reduce(250), // &, reduce: ContextVariable
			reduce(250), // <<, reduce: ContextVariable
			reduce(250), // >>>, reduce: ContextVariable
			reduce(250), // +, reduce: ContextVariable
			reduce(250), // -, reduce: ContextVariable
			reduce(250), // .+, reduce: ContextVariable
			reduce(250), // .-, reduce: ContextVariable
			reduce(250), // *, reduce: ContextVariable
			reduce(250), // /, reduce: ContextVariable
			reduce(250), // //, reduce: ContextVariable
			reduce(250), // %, reduce: ContextVariable
			reduce(250), // .*, reduce: ContextVariable
			reduce(250), // ./, reduce: ContextVariable
			reduce(250), // .//, reduce: ContextVariable
			reduce(250), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(250), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(250), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(250), // [[, reduce: ContextVariable
			reduce(250), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2124
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(251), // ;, reduce: ContextVariable
			nil,         // {
			reduce(251), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(251), // >, reduce: ContextVariable
			reduce(251), // >>, reduce: ContextVariable
			reduce(251), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(251), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(251), // ?, reduce: ContextVariable
			nil,         // :
			reduce(251), // ||, reduce: ContextVariable
			reduce(251), // ^^, reduce: ContextVariable
			reduce(251), // &&, reduce: ContextVariable
			reduce(251), // ??, reduce: ContextVariable
			reduce(251), // ???, reduce: ContextVariable
			reduce(251), // =~, reduce: ContextVariable
			reduce(251), // !=~, reduce: ContextVariable
			reduce(251), // ==, reduce: ContextVariable
			reduce(251), // !=, reduce: ContextVariable
			reduce(251), // <=>, reduce: ContextVariable
			reduce(251), // >=, reduce: ContextVariable
			reduce(251), // <, reduce: ContextVariable
			reduce(251), // <=, reduce: ContextVariable
			reduce(251), // ^, reduce: ContextVariable
			reduce(251), // &, reduce: ContextVariable
			reduce(251), // <<, reduce: ContextVariable
			reduce(251), // >>>, reduce: ContextVariable
			reduce(251), // +, reduce: ContextVariable
			reduce(251), // -, reduce: ContextVariable
			reduce(251), // .+, reduce: ContextVariable
			reduce(251), // .-, reduce: ContextVariable
			reduce(251), // *, reduce: ContextVariable
			reduce(251), // /, reduce: ContextVariable
			reduce(251), // //, reduce: ContextVariable
			reduce(251), // %, reduce: ContextVariable
			reduce(251), // .*, reduce: ContextVariable
			reduce(251), // ./, reduce: ContextVariable
			reduce(251), // .//, reduce: ContextVariable
			reduce(251), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(251), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(251), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(251), // [[, reduce: ContextVariable
			reduce(251), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2125
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(252), // ;, reduce: ContextVariable
			nil,         // {
			reduce(252), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(252), // >, reduce: ContextVariable
			reduce(252), // >>, reduce: ContextVariable
			reduce(252), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(252), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(252), // ?, reduce: ContextVariable
			nil,         // :
			reduce(252), // ||, reduce: ContextVariable
			reduce(252), // ^^, reduce: ContextVariable
			reduce(252), // &&, reduce: ContextVariable
			reduce(252), // ??, reduce: ContextVariable
			reduce(252), // ???, reduce: ContextVariable
			reduce(252), // =~, reduce: ContextVariable
			reduce(252), // !=~, reduce: ContextVariable
			reduce(252), // ==, reduce: ContextVariable
			reduce(252), // !=, reduce: ContextVariable
			reduce(252), // <=>, reduce: ContextVariable
			reduce(252), // >=, reduce: ContextVariable
			reduce(252), // <, reduce: ContextVariable
			reduce(252), // <=, reduce: ContextVariable
			reduce(252), // ^, reduce: ContextVariable
			reduce(252), // &, reduce: ContextVariable
			reduce(252), // <<, reduce: ContextVariable
			reduce(252), // >>>, reduce: ContextVariable
			reduce(252), // +, reduce: ContextVariable
			reduce(252), // -, reduce: ContextVariable
			reduce(252), // .+, reduce: ContextVariable
			reduce(252), // .-, reduce: ContextVariable
			reduce(252), // *, reduce: ContextVariable
			reduce(252), // /, reduce: ContextVariable
			reduce(252), // //, reduce: ContextVariable
			reduce(252), // %, reduce: ContextVariable
			reduce(252), // .*, reduce: ContextVariable
			reduce(252), // ./, reduce: ContextVariable
			reduce(252), // .//, reduce: ContextVariable
			reduce(252), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(252), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(252), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(252), // [[, reduce: ContextVariable
			reduce(252), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2126
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(253), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(253), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(253), // >, reduce: MlrvalOrFunction
			reduce(253), // >>, reduce: MlrvalOrFunction
			reduce(253), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(253), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(253), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(253), // ||, reduce: MlrvalOrFunction
			reduce(253), // ^^, reduce: MlrvalOrFunction
			reduce(253), // &&, reduce: MlrvalOrFunction
			reduce(253), // ??, reduce: MlrvalOrFunction
			reduce(253), // ???, reduce: MlrvalOrFunction
			reduce(253), // =~, reduce: MlrvalOrFunction
			reduce(253), // !=~, reduce: MlrvalOrFunction
			reduce(253), // ==, reduce: MlrvalOrFunction
			reduce(253), // !=, reduce: MlrvalOrFunction
			reduce(253), // <=>, reduce: MlrvalOrFunction
			reduce(253), // >=, reduce: MlrvalOrFunction
			reduce(253), // <, reduce: MlrvalOrFunction
			reduce(253), // <=, reduce: MlrvalOrFunction
			reduce(253), // ^, reduce: MlrvalOrFunction
			reduce(253), // &, reduce: MlrvalOrFunction
			reduce(253), // <<, reduce: MlrvalOrFunction
			reduce(253), // >>>, reduce: MlrvalOrFunction
			reduce(253), // +, reduce: MlrvalOrFunction
			reduce(253), // -, reduce: MlrvalOrFunction
			reduce(253), // .+, reduce: MlrvalOrFunction
			reduce(253), // .-, reduce: MlrvalOrFunction
			reduce(253), // *, reduce: MlrvalOrFunction
			reduce(253), // /, reduce: MlrvalOrFunction
			reduce(253), // //, reduce: MlrvalOrFunction
			reduce(253), // %, reduce: MlrvalOrFunction
			reduce(253), // .*, reduce: MlrvalOrFunction
			reduce(253), // ./, reduce: MlrvalOrFunction
			reduce(253), // .//, reduce: MlrvalOrFunction
			reduce(253), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(253), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(253), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(253), // [[, reduce: MlrvalOrFunction
			reduce(253), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2127
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			shift(3054), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(3055), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2128
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(256), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(256), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(256), // >, reduce: MlrvalOrFunction
			reduce(256), // >>, reduce: MlrvalOrFunction
			reduce(256), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(256), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(256), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(256), // ||, reduce: MlrvalOrFunction
			reduce(256), // ^^, reduce: MlrvalOrFunction
			reduce(256), // &&, reduce: MlrvalOrFunction
			reduce(256), // ??, reduce: MlrvalOrFunction
			reduce(256), // ???, reduce: MlrvalOrFunction
			reduce(256), // =~, reduce: MlrvalOrFunction
			reduce(256), // !=~, reduce: MlrvalOrFunction
			reduce(256), // ==, reduce: MlrvalOrFunction
			reduce(256), // !=, reduce: MlrvalOrFunction
			reduce(256), // <=>, reduce: MlrvalOrFunction
			reduce(256), // >=, reduce: MlrvalOrFunction
			reduce(256), // <, reduce: MlrvalOrFunction
			reduce(256), // <=, reduce: MlrvalOrFunction
			reduce(256), // ^, reduce: MlrvalOrFunction
			reduce(256), // &, reduce: MlrvalOrFunction
			reduce(256), // <<, reduce: MlrvalOrFunction
			reduce(256), // >>>, reduce: MlrvalOrFunction
			reduce(256), // +, reduce: MlrvalOrFunction
			reduce(256), // -, reduce: MlrvalOrFunction
			reduce(256), // .+, reduce: MlrvalOrFunction
			reduce(256), // .-, reduce: MlrvalOrFunction
			reduce(256), // *, reduce: MlrvalOrFunction
			reduce(256), // /, reduce: MlrvalOrFunction
			reduce(256), // //, reduce: MlrvalOrFunction
			reduce(256), // %, reduce: MlrvalOrFunction
			reduce(256), // .*, reduce: MlrvalOrFunction
			reduce(256), // ./, reduce: MlrvalOrFunction
			reduce(256), // .//, reduce: MlrvalOrFunction
			reduce(256), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(256), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(256), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(256), // [[, reduce: MlrvalOrFunction
			reduce(256), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2129
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(257), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(257), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(257), // >, reduce: MlrvalOrFunction
			reduce(257), // >>, reduce: MlrvalOrFunction
			reduce(257), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(257), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(257), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(257), // ||, reduce: MlrvalOrFunction
			reduce(257), // ^^, reduce: MlrvalOrFunction
			reduce(257), // &&, reduce: MlrvalOrFunction
			reduce(257), // ??, reduce: MlrvalOrFunction
			reduce(257), // ???, reduce: MlrvalOrFunction
			reduce(257), // =~, reduce: MlrvalOrFunction
			reduce(257), // !=~, reduce: MlrvalOrFunction
			reduce(257), // ==, reduce: MlrvalOrFunction
			reduce(257), // !=, reduce: MlrvalOrFunction
			reduce(257), // <=>, reduce: MlrvalOrFunction
			reduce(257), // >=, reduce: MlrvalOrFunction
			reduce(257), // <, reduce: MlrvalOrFunction
			reduce(257), // <=, reduce: MlrvalOrFunction
			reduce(257), // ^, reduce: MlrvalOrFunction
			reduce(257), // &, reduce: MlrvalOrFunction
			reduce(257), // <<, reduce: MlrvalOrFunction
			reduce(257), // >>>, reduce: MlrvalOrFunction
			reduce(257), // +, reduce: MlrvalOrFunction
			reduce(257), // -, reduce: MlrvalOrFunction
			reduce(257), // .+, reduce: MlrvalOrFunction
			reduce(257), // .-, reduce: MlrvalOrFunction
			reduce(257), // *, reduce: MlrvalOrFunction
			reduce(257), // /, reduce: MlrvalOrFunction
			reduce(257), // //, reduce: MlrvalOrFunction
			reduce(257), // %, reduce: MlrvalOrFunction
			reduce(257), // .*, reduce: MlrvalOrFunction
			reduce(257), // ./, reduce: MlrvalOrFunction
			reduce(257), // .//, reduce: MlrvalOrFunction
			reduce(257), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(257), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(257), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(257), // [[, reduce: MlrvalOrFunction
			reduce(257), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2130
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(258), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(258), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(258), // >, reduce: MlrvalOrFunction
			reduce(258), // >>, reduce: MlrvalOrFunction
			reduce(258), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(258), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(258), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(258), // ||, reduce: MlrvalOrFunction
			reduce(258), // ^^, reduce: MlrvalOrFunction
			reduce(258), // &&, reduce: MlrvalOrFunction
			reduce(258), // ??, reduce: MlrvalOrFunction
			reduce(258), // ???, reduce: MlrvalOrFunction
			reduce(258), // =~, reduce: MlrvalOrFunction
			reduce(258), // !=~, reduce: MlrvalOrFunction
			reduce(258), // ==, reduce: MlrvalOrFunction
			reduce(258), // !=, reduce: MlrvalOrFunction
			reduce(258), // <=>, reduce: MlrvalOrFunction
			reduce(258), // >=, reduce: MlrvalOrFunction
			reduce(258), // <, reduce: MlrvalOrFunction
			reduce(258), // <=, reduce: MlrvalOrFunction
			reduce(258), // ^, reduce: MlrvalOrFunction
			reduce(258), // &, reduce: MlrvalOrFunction
			reduce(258), // <<, reduce: MlrvalOrFunction
			reduce(258), // >>>, reduce: MlrvalOrFunction
			reduce(258), // +, reduce: MlrvalOrFunction
			reduce(258), // -, reduce: MlrvalOrFunction
			reduce(258), // .+, reduce: MlrvalOrFunction
			reduce(258), // .-, reduce: MlrvalOrFunction
			reduce(258), // *, reduce: MlrvalOrFunction
			reduce(258), // /, reduce: MlrvalOrFunction
			reduce(258), // //, reduce: MlrvalOrFunction
			reduce(258), // %, reduce: MlrvalOrFunction
			reduce(258), // .*, reduce: MlrvalOrFunction
			reduce(258), // ./, reduce: MlrvalOrFunction
			reduce(258), // .//, reduce: MlrvalOrFunction
			reduce(258), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(258), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(258), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(258), // [[, reduce: MlrvalOrFunction
			reduce(258), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2131
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(259), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(259), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(259), // >, reduce: MlrvalOrFunction
			reduce(259), // >>, reduce: MlrvalOrFunction
			reduce(259), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(259), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(259), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(259), // ||, reduce: MlrvalOrFunction
			reduce(259), // ^^, reduce: MlrvalOrFunction
			reduce(259), // &&, reduce: MlrvalOrFunction
			reduce(259), // ??, reduce: MlrvalOrFunction
			reduce(259), // ???, reduce: MlrvalOrFunction
			reduce(259), // =~, reduce: MlrvalOrFunction
			reduce(259), // !=~, reduce: MlrvalOrFunction
			reduce(259), // ==, reduce: MlrvalOrFunction
			reduce(259), // !=, reduce: MlrvalOrFunction
			reduce(259), // <=>, reduce: MlrvalOrFunction
			reduce(259), // >=, reduce: MlrvalOrFunction
			reduce(259), // <, reduce: MlrvalOrFunction
			reduce(259), // <=, reduce: MlrvalOrFunction
			reduce(259), // ^, reduce: MlrvalOrFunction
			reduce(259), // &, reduce: MlrvalOrFunction
			reduce(259), // <<, reduce: MlrvalOrFunction
			reduce(259), // >>>, reduce: MlrvalOrFunction
			reduce(259), // +, reduce: MlrvalOrFunction
			reduce(259), // -, reduce: MlrvalOrFunction
			reduce(259), // .+, reduce: MlrvalOrFunction
			reduce(259), // .-, reduce: MlrvalOrFunction
			reduce(259), // *, reduce: MlrvalOrFunction
			reduce(259), // /, reduce: MlrvalOrFunction
			reduce(259), // //, reduce: MlrvalOrFunction
			reduce(259), // %, reduce: MlrvalOrFunction
			reduce(259), // .*, reduce: MlrvalOrFunction
			reduce(259), // ./, reduce: MlrvalOrFunction
			reduce(259), // .//, reduce: MlrvalOrFunction
			reduce(259), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(259), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(259), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(259), // [[, reduce: MlrvalOrFunction
			reduce(259), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2132
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(267), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(267), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(267), // >, reduce: MlrvalOrFunction
			reduce(267), // >>, reduce: MlrvalOrFunction
			reduce(267), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(267), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(267), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(267), // ||, reduce: MlrvalOrFunction
			reduce(267), // ^^, reduce: MlrvalOrFunction
			reduce(267), // &&, reduce: MlrvalOrFunction
			reduce(267), // ??, reduce: MlrvalOrFunction
			reduce(267), // ???, reduce: MlrvalOrFunction
			reduce(267), // =~, reduce: MlrvalOrFunction
			reduce(267), // !=~, reduce: MlrvalOrFunction
			reduce(267), // ==, reduce: MlrvalOrFunction
			reduce(267), // !=, reduce: MlrvalOrFunction
			reduce(267), // <=>, reduce: MlrvalOrFunction
			reduce(267), // >=, reduce: MlrvalOrFunction
			reduce(267), // <, reduce: MlrvalOrFunction
			reduce(267), // <=, reduce: MlrvalOrFunction
			reduce(267), // ^, reduce: MlrvalOrFunction
			reduce(267), // &, reduce: MlrvalOrFunction
			reduce(267), // <<, reduce: MlrvalOrFunction
			reduce(267), // >>>, reduce: MlrvalOrFunction
			reduce(267), // +, reduce: MlrvalOrFunction
			reduce(267), // -, reduce: MlrvalOrFunction
			reduce(267), // .+, reduce: MlrvalOrFunction
			reduce(267), // .-, reduce: MlrvalOrFunction
			reduce(267), // *, reduce: MlrvalOrFunction
			reduce(267), // /, reduce: MlrvalOrFunction
			reduce(267), // //, reduce: MlrvalOrFunction
			reduce(267), // %, reduce: MlrvalOrFunction
			reduce(267), // .*, reduce: MlrvalOrFunction
			reduce(267), // ./, reduce: MlrvalOrFunction
			reduce(267), // .//, reduce: MlrvalOrFunction
			reduce(267), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(267), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(267), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(267), // [[, reduce: MlrvalOrFunction
			reduce(267), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2133
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3056), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2134
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3057), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2135
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			shift(3058), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2136
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(39), // ;, reduce: FilterStatement
			nil,        // {
			reduce(39), // }, reduce: FilterStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2137
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(214), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(214), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(214), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2138
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(211), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(211), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(211), // >, reduce: MlrvalOrFunction
			reduce(211), // >>, reduce: MlrvalOrFunction
			reduce(211), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(211), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(211), // ||, reduce: MlrvalOrFunction
			reduce(211), // ^^, reduce: MlrvalOrFunction
			reduce(211), // &&, reduce: MlrvalOrFunction
			reduce(211), // ??, reduce: MlrvalOrFunction
			reduce(211), // ???, reduce: MlrvalOrFunction
			reduce(211), // =~, reduce: MlrvalOrFunction
			reduce(211), // !=~, reduce: MlrvalOrFunction
			reduce(211), // ==, reduce: MlrvalOrFunction
			reduce(211), // !=, reduce: MlrvalOrFunction
			reduce(211), // <=>, reduce: MlrvalOrFunction
			reduce(211), // >=, reduce: MlrvalOrFunction
			reduce(211), // <, reduce: MlrvalOrFunction
			reduce(211), // <=, reduce: MlrvalOrFunction
			reduce(211), // ^, reduce: MlrvalOrFunction
			reduce(211), // &, reduce: MlrvalOrFunction
			reduce(211), // <<, reduce: MlrvalOrFunction
			reduce(211), // >>>, reduce: MlrvalOrFunction
			reduce(211), // +, reduce: MlrvalOrFunction
			reduce(211), // -, reduce: MlrvalOrFunction
			reduce(211), // .+, reduce: MlrvalOrFunction
			reduce(211), // .-, reduce: MlrvalOrFunction
			reduce(211), // *, reduce: MlrvalOrFunction
			reduce(211), // /, reduce: MlrvalOrFunction
			reduce(211), // //, reduce: MlrvalOrFunction
			reduce(211), // %, reduce: MlrvalOrFunction
			reduce(211), // .*, reduce: MlrvalOrFunction
			reduce(211), // ./, reduce: MlrvalOrFunction
			reduce(211), // .//, reduce: MlrvalOrFunction
			reduce(211), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(211), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(211), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(211), // [[, reduce: MlrvalOrFunction
			reduce(211), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2139
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2140
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(110), // ;, reduce: OosvarValue
			nil,         // {
			reduce(110), // }, reduce: OosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(110), // >, reduce: OosvarValue
			reduce(110), // >>, reduce: OosvarValue
			reduce(110), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(110), // ?, reduce: OosvarValue
			nil,         // :
			reduce(110), // ||, reduce: OosvarValue
			reduce(110), // ^^, reduce: OosvarValue
			reduce(110), // &&, reduce: OosvarValue
			reduce(110), // ??, reduce: OosvarValue
			reduce(110), // ???, reduce: OosvarValue
			reduce(110), // =~, reduce: OosvarValue
			reduce(110), // !=~, reduce: OosvarValue
			reduce(110), // ==, reduce: OosvarValue
			reduce(110), // !=, reduce: OosvarValue
			reduce(110), // <=>, reduce: OosvarValue
			reduce(110), // >=, reduce: OosvarValue
			reduce(110), // <, reduce: OosvarValue
			reduce(110), // <=, reduce: OosvarValue
			reduce(110), // ^, reduce: OosvarValue
			reduce(110), // &, reduce: OosvarValue
			reduce(110), // <<, reduce: OosvarValue
			reduce(110), // >>>, reduce: OosvarValue
			reduce(110), // +, reduce: OosvarValue
			reduce(110), // -, reduce: OosvarValue
			reduce(110), // .+, reduce: OosvarValue
			reduce(110), // .-, reduce: OosvarValue
			reduce(110), // *, reduce: OosvarValue
			reduce(110), // /, reduce: OosvarValue
			reduce(110), // //, reduce: OosvarValue
			reduce(110), // %, reduce: OosvarValue
			reduce(110), // .*, reduce: OosvarValue
			reduce(110), // ./, reduce: OosvarValue
			reduce(110), // .//, reduce: OosvarValue
			reduce(110), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(110), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(110), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(110), // [[, reduce: OosvarValue
			reduce(110), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2141
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(112), // ;, reduce: OosvarValue
			nil,         // {
			reduce(112), // }, reduce: OosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(112), // >, reduce: OosvarValue
			reduce(112), // >>, reduce: OosvarValue
			reduce(112), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(112), // ?, reduce: OosvarValue
			nil,         // :
			reduce(112), // ||, reduce: OosvarValue
			reduce(112), // ^^, reduce: OosvarValue
			reduce(112), // &&, reduce: OosvarValue
			reduce(112), // ??, reduce: OosvarValue
			reduce(112), // ???, reduce: OosvarValue
			reduce(112), // =~, reduce: OosvarValue
			reduce(112), // !=~, reduce: OosvarValue
			reduce(112), // ==, reduce: OosvarValue
			reduce(112), // !=, reduce: OosvarValue
			reduce(112), // <=>, reduce: OosvarValue
			reduce(112), // >=, reduce: OosvarValue
			reduce(112), // <, reduce: OosvarValue
			reduce(112), // <=, reduce: OosvarValue
			reduce(112), // ^, reduce: OosvarValue
			reduce(112), // &, reduce: OosvarValue
			reduce(112), // <<, reduce: OosvarValue
			reduce(112), // >>>, reduce: OosvarValue
			reduce(112), // +, reduce: OosvarValue
			reduce(112), // -, reduce: OosvarValue
			reduce(112), // .+, reduce: OosvarValue
			reduce(112), // .-, reduce: OosvarValue
			reduce(112), // *, reduce: OosvarValue
			reduce(112), // /, reduce: OosvarValue
			reduce(112), // //, reduce: OosvarValue
			reduce(112), // %, reduce: OosvarValue
			reduce(112), // .*, reduce: OosvarValue
			reduce(112), // ./, reduce: OosvarValue
			reduce(112), // .//, reduce: OosvarValue
			reduce(112), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(112), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(112), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(112), // [[, reduce: OosvarValue
			reduce(112), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2142
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(111), // ;, reduce: OosvarValue
			nil,         // {
			reduce(111), // }, reduce: OosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(111), // >, reduce: OosvarValue
			reduce(111), // >>, reduce: OosvarValue
			reduce(111), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(111), // ?, reduce: OosvarValue
			nil,         // :
			reduce(111), // ||, reduce: OosvarValue
			reduce(111), // ^^, reduce: OosvarValue
			reduce(111), // &&, reduce: OosvarValue
			reduce(111), // ??, reduce: OosvarValue
			reduce(111), // ???, reduce: OosvarValue
			reduce(111), // =~, reduce: OosvarValue
			reduce(111), // !=~, reduce: OosvarValue
			reduce(111), // ==, reduce: OosvarValue
			reduce(111), // !=, reduce: OosvarValue
			reduce(111), // <=>, reduce: OosvarValue
			reduce(111), // >=, reduce: OosvarValue
			reduce(111), // <, reduce: OosvarValue
			reduce(111), // <=, reduce: OosvarValue
			reduce(111), // ^, reduce: OosvarValue
			reduce(111), // &, reduce: OosvarValue
			reduce(111), // <<, reduce: OosvarValue
			reduce(111), // >>>, reduce: OosvarValue
			reduce(111), // +, reduce: OosvarValue
			reduce(111), // -, reduce: OosvarValue
			reduce(111), // .+, reduce: OosvarValue
			reduce(111), // .-, reduce: OosvarValue
			reduce(111), // *, reduce: OosvarValue
			reduce(111), // /, reduce: OosvarValue
			reduce(111), // //, reduce: OosvarValue
			reduce(111), // %, reduce: OosvarValue
			reduce(111), // .*, reduce: OosvarValue
			reduce(111), // ./, reduce: OosvarValue
			reduce(111), // .//, reduce: OosvarValue
			reduce(111), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(111), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(111), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(111), // [[, reduce: OosvarValue
			reduce(111), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2143
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(99), // ;, reduce: FieldValue
			nil,        // {
			reduce(99), // }, reduce: FieldValue
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(99), // >, reduce: FieldValue
			reduce(99), // >>, reduce: FieldValue
			reduce(99), // |, reduce: FieldValue
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(99), // ?, reduce: FieldValue
			nil,        // :
			reduce(99), // ||, reduce: FieldValue
			reduce(99), // ^^, reduce: FieldValue
			reduce(99), // &&, reduce: FieldValue
			reduce(99), // ??, reduce: FieldValue
			reduce(99), // ???, reduce: FieldValue
			reduce(99), // =~, reduce: FieldValue
			reduce(99), // !=~, reduce: FieldValue
			reduce(99), // ==, reduce: FieldValue
			reduce(99), // !=, reduce: FieldValue
			reduce(99), // <=>, reduce: FieldValue
			reduce(99), // >=, reduce: FieldValue
			reduce(99), // <, reduce: FieldValue
			reduce(99), // <=, reduce: FieldValue
			reduce(99), // ^, reduce: FieldValue
			reduce(99), // &, reduce: FieldValue
			reduce(99), // <<, reduce: FieldValue
			reduce(99), // >>>, reduce: FieldValue
			reduce(99), // +, reduce: FieldValue
			reduce(99), // -, reduce: FieldValue
			reduce(99), // .+, reduce: FieldValue
			reduce(99), // .-, reduce: FieldValue
			reduce(99), // *, reduce: FieldValue
			reduce(99), // /, reduce: FieldValue
			reduce(99), // //, reduce: FieldValue
			reduce(99), // %, reduce: FieldValue
			reduce(99), // .*, reduce: FieldValue
			reduce(99), // ./, reduce: FieldValue
			reduce(99), // .//, reduce: FieldValue
			reduce(99), // ., reduce: FieldValue
			nil,        // !
			nil,        // ~
			reduce(99), // **, reduce: FieldValue
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(99), // [, reduce: FieldValue
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(99), // [[, reduce: FieldValue
			reduce(99), // [[[, reduce: FieldValue
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2144
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(101), // ;, reduce: FieldValue
			nil,         // {
			reduce(101), // }, reduce: FieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(101), // >, reduce: FieldValue
			reduce(101), // >>, reduce: FieldValue
			reduce(101), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(101), // ?, reduce: FieldValue
			nil,         // :
			reduce(101), // ||, reduce: FieldValue
			reduce(101), // ^^, reduce: FieldValue
			reduce(101), // &&, reduce: FieldValue
			reduce(101), // ??, reduce: FieldValue
			reduce(101), // ???, reduce: FieldValue
			reduce(101), // =~, reduce: FieldValue
			reduce(101), // !=~, reduce: FieldValue
			reduce(101), // ==, reduce: FieldValue
			reduce(101), // !=, reduce: FieldValue
			reduce(101), // <=>, reduce: FieldValue
			reduce(101), // >=, reduce: FieldValue
			reduce(101), // <, reduce: FieldValue
			reduce(101), // <=, reduce: FieldValue
			reduce(101), // ^, reduce: FieldValue
			reduce(101), // &, reduce: FieldValue
			reduce(101), // <<, reduce: FieldValue
			reduce(101), // >>>, reduce: FieldValue
			reduce(101), // +, reduce: FieldValue
			reduce(101), // -, reduce: FieldValue
			reduce(101), // .+, reduce: FieldValue
			reduce(101), // .-, reduce: FieldValue
			reduce(101), // *, reduce: FieldValue
			reduce(101), // /, reduce: FieldValue
			reduce(101), // //, reduce: FieldValue
			reduce(101), // %, reduce: FieldValue
			reduce(101), // .*, reduce: FieldValue
			reduce(101), // ./, reduce: FieldValue
			reduce(101), // .//, reduce: FieldValue
			reduce(101), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(101), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(101), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(101), // [[, reduce: FieldValue
			reduce(101), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2145
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(100), // ;, reduce: FieldValue
			nil,         // {
			reduce(100), // }, reduce: FieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(100), // >, reduce: FieldValue
			reduce(100), // >>, reduce: FieldValue
			reduce(100), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(100), // ?, reduce: FieldValue
			nil,         // :
			reduce(100), // ||, reduce: FieldValue
			reduce(100), // ^^, reduce: FieldValue
			reduce(100), // &&, reduce: FieldValue
			reduce(100), // ??, reduce: FieldValue
			reduce(100), // ???, reduce: FieldValue
			reduce(100), // =~, reduce: FieldValue
			reduce(100), // !=~, reduce: FieldValue
			reduce(100), // ==, reduce: FieldValue
			reduce(100), // !=, reduce: FieldValue
			reduce(100), // <=>, reduce: FieldValue
			reduce(100), // >=, reduce: FieldValue
			reduce(100), // <, reduce: FieldValue
			reduce(100), // <=, reduce: FieldValue
			reduce(100), // ^, reduce: FieldValue
			reduce(100), // &, reduce: FieldValue
			reduce(100), // <<, reduce: FieldValue
			reduce(100), // >>>, reduce: FieldValue
			reduce(100), // +, reduce: FieldValue
			reduce(100), // -, reduce: FieldValue
			reduce(100), // .+, reduce: FieldValue
			reduce(100), // .-, reduce: FieldValue
			reduce(100), // *, reduce: FieldValue
			reduce(100), // /, reduce: FieldValue
			reduce(100), // //, reduce: FieldValue
			reduce(100), // %, reduce: FieldValue
			reduce(100), // .*, reduce: FieldValue
			reduce(100), // ./, reduce: FieldValue
			reduce(100), // .//, reduce: FieldValue
			reduce(100), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(100), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(100), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(100), // [[, reduce: FieldValue
			reduce(100), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2146
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(213), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(213), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(213), // >, reduce: MlrvalOrFunction
			reduce(213), // >>, reduce: MlrvalOrFunction
			reduce(213), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(213), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(213), // ||, reduce: MlrvalOrFunction
			reduce(213), // ^^, reduce: MlrvalOrFunction
			reduce(213), // &&, reduce: MlrvalOrFunction
			reduce(213), // ??, reduce: MlrvalOrFunction
			reduce(213), // ???, reduce: MlrvalOrFunction
			reduce(213), // =~, reduce: MlrvalOrFunction
			reduce(213), // !=~, reduce: MlrvalOrFunction
			reduce(213), // ==, reduce: MlrvalOrFunction
			reduce(213), // !=, reduce: MlrvalOrFunction
			reduce(213), // <=>, reduce: MlrvalOrFunction
			reduce(213), // >=, reduce: MlrvalOrFunction
			reduce(213), // <, reduce: MlrvalOrFunction
			reduce(213), // <=, reduce: MlrvalOrFunction
			reduce(213), // ^, reduce: MlrvalOrFunction
			reduce(213), // &, reduce: MlrvalOrFunction
			reduce(213), // <<, reduce: MlrvalOrFunction
			reduce(213), // >>>, reduce: MlrvalOrFunction
			reduce(213), // +, reduce: MlrvalOrFunction
			reduce(213), // -, reduce: MlrvalOrFunction
			reduce(213), // .+, reduce: MlrvalOrFunction
			reduce(213), // .-, reduce: MlrvalOrFunction
			reduce(213), // *, reduce: MlrvalOrFunction
			reduce(213), // /, reduce: MlrvalOrFunction
			reduce(213), // //, reduce: MlrvalOrFunction
			reduce(213), // %, reduce: MlrvalOrFunction
			reduce(213), // .*, reduce: MlrvalOrFunction
			reduce(213), // ./, reduce: MlrvalOrFunction
			reduce(213), // .//, reduce: MlrvalOrFunction
			reduce(213), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(213), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(213), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(213), // [[, reduce: MlrvalOrFunction
			reduce(213), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2147
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(233), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(233), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(233), // >, reduce: MlrvalOrFunction
			reduce(233), // >>, reduce: MlrvalOrFunction
			reduce(233), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(233), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(233), // ||, reduce: MlrvalOrFunction
			reduce(233), // ^^, reduce: MlrvalOrFunction
			reduce(233), // &&, reduce: MlrvalOrFunction
			reduce(233), // ??, reduce: MlrvalOrFunction
			reduce(233), // ???, reduce: MlrvalOrFunction
			reduce(233), // =~, reduce: MlrvalOrFunction
			reduce(233), // !=~, reduce: MlrvalOrFunction
			reduce(233), // ==, reduce: MlrvalOrFunction
			reduce(233), // !=, reduce: MlrvalOrFunction
			reduce(233), // <=>, reduce: MlrvalOrFunction
			reduce(233), // >=, reduce: MlrvalOrFunction
			reduce(233), // <, reduce: MlrvalOrFunction
			reduce(233), // <=, reduce: MlrvalOrFunction
			reduce(233), // ^, reduce: MlrvalOrFunction
			reduce(233), // &, reduce: MlrvalOrFunction
			reduce(233), // <<, reduce: MlrvalOrFunction
			reduce(233), // >>>, reduce: MlrvalOrFunction
			reduce(233), // +, reduce: MlrvalOrFunction
			reduce(233), // -, reduce: MlrvalOrFunction
			reduce(233), // .+, reduce: MlrvalOrFunction
			reduce(233), // .-, reduce: MlrvalOrFunction
			reduce(233), // *, reduce: MlrvalOrFunction
			reduce(233), // /, reduce: MlrvalOrFunction
			reduce(233), // //, reduce: MlrvalOrFunction
			reduce(233), // %, reduce: MlrvalOrFunction
			reduce(233), // .*, reduce: MlrvalOrFunction
			reduce(233), // ./, reduce: MlrvalOrFunction
			reduce(233), // .//, reduce: MlrvalOrFunction
			reduce(233), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(233), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(233), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(233), // [[, reduce: MlrvalOrFunction
			reduce(233), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2148
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(210), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(210), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(210), // >, reduce: MlrvalOrFunction
			reduce(210), // >>, reduce: MlrvalOrFunction
			reduce(210), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(210), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(210), // ||, reduce: MlrvalOrFunction
			reduce(210), // ^^, reduce: MlrvalOrFunction
			reduce(210), // &&, reduce: MlrvalOrFunction
			reduce(210), // ??, reduce: MlrvalOrFunction
			reduce(210), // ???, reduce: MlrvalOrFunction
			reduce(210), // =~, reduce: MlrvalOrFunction
			reduce(210), // !=~, reduce: MlrvalOrFunction
			reduce(210), // ==, reduce: MlrvalOrFunction
			reduce(210), // !=, reduce: MlrvalOrFunction
			reduce(210), // <=>, reduce: MlrvalOrFunction
			reduce(210), // >=, reduce: MlrvalOrFunction
			reduce(210), // <, reduce: MlrvalOrFunction
			reduce(210), // <=, reduce: MlrvalOrFunction
			reduce(210), // ^, reduce: MlrvalOrFunction
			reduce(210), // &, reduce: MlrvalOrFunction
			reduce(210), // <<, reduce: MlrvalOrFunction
			reduce(210), // >>>, reduce: MlrvalOrFunction
			reduce(210), // +, reduce: MlrvalOrFunction
			reduce(210), // -, reduce: MlrvalOrFunction
			reduce(210), // .+, reduce: MlrvalOrFunction
			reduce(210), // .-, reduce: MlrvalOrFunction
			reduce(210), // *, reduce: MlrvalOrFunction
			reduce(210), // /, reduce: MlrvalOrFunction
			reduce(210), // //, reduce: MlrvalOrFunction
			reduce(210), // %, reduce: MlrvalOrFunction
			reduce(210), // .*, reduce: MlrvalOrFunction
			reduce(210), // ./, reduce: MlrvalOrFunction
			reduce(210), // .//, reduce: MlrvalOrFunction
			reduce(210), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(210), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(210), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(210), // [[, reduce: MlrvalOrFunction
			reduce(210), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2149
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(102), // ;, reduce: FieldValue
			nil,         // {
			reduce(102), // }, reduce: FieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(102), // >, reduce: FieldValue
			reduce(102), // >>, reduce: FieldValue
			reduce(102), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(102), // ?, reduce: FieldValue
			nil,         // :
			reduce(102), // ||, reduce: FieldValue
			reduce(102), // ^^, reduce: FieldValue
			reduce(102), // &&, reduce: FieldValue
			reduce(102), // ??, reduce: FieldValue
			reduce(102), // ???, reduce: FieldValue
			reduce(102), // =~, reduce: FieldValue
			reduce(102), // !=~, reduce: FieldValue
			reduce(102), // ==, reduce: FieldValue
			reduce(102), // !=, reduce: FieldValue
			reduce(102), // <=>, reduce: FieldValue
			reduce(102), // >=, reduce: FieldValue
			reduce(102), // <, reduce: FieldValue
			reduce(102), // <=, reduce: FieldValue
			reduce(102), // ^, reduce: FieldValue
			reduce(102), // &, reduce: FieldValue
			reduce(102), // <<, reduce: FieldValue
			reduce(102), // >>>, reduce: FieldValue
			reduce(102), // +, reduce: FieldValue
			reduce(102), // -, reduce: FieldValue
			reduce(102), // .+, reduce: FieldValue
			reduce(102), // .-, reduce: FieldValue
			reduce(102), // *, reduce: FieldValue
			reduce(102), // /, reduce: FieldValue
			reduce(102), // //, reduce: FieldValue
			reduce(102), // %, reduce: FieldValue
			reduce(102), // .*, reduce: FieldValue
			reduce(102), // ./, reduce: FieldValue
			reduce(102), // .//, reduce: FieldValue
			reduce(102), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(102), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(102), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(102), // [[, reduce: FieldValue
			reduce(102), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2150
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(103), // ;, reduce: FieldValue
			nil,         // {
			reduce(103), // }, reduce: FieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(103), // >, reduce: FieldValue
			reduce(103), // >>, reduce: FieldValue
			reduce(103), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(103), // ?, reduce: FieldValue
			nil,         // :
			reduce(103), // ||, reduce: FieldValue
			reduce(103), // ^^, reduce: FieldValue
			reduce(103), // &&, reduce: FieldValue
			reduce(103), // ??, reduce: FieldValue
			reduce(103), // ???, reduce: FieldValue
			reduce(103), // =~, reduce: FieldValue
			reduce(103), // !=~, reduce: FieldValue
			reduce(103), // ==, reduce: FieldValue
			reduce(103), // !=, reduce: FieldValue
			reduce(103), // <=>, reduce: FieldValue
			reduce(103), // >=, reduce: FieldValue
			reduce(103), // <, reduce: FieldValue
			reduce(103), // <=, reduce: FieldValue
			reduce(103), // ^, reduce: FieldValue
			reduce(103), // &, reduce: FieldValue
			reduce(103), // <<, reduce: FieldValue
			reduce(103), // >>>, reduce: FieldValue
			reduce(103), // +, reduce: FieldValue
			reduce(103), // -, reduce: FieldValue
			reduce(103), // .+, reduce: FieldValue
			reduce(103), // .-, reduce: FieldValue
			reduce(103), // *, reduce: FieldValue
			reduce(103), // /, reduce: FieldValue
			reduce(103), // //, reduce: FieldValue
			reduce(103), // %, reduce: FieldValue
			reduce(103), // .*, reduce: FieldValue
			reduce(103), // ./, reduce: FieldValue
			reduce(103), // .//, reduce: FieldValue
			reduce(103), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(103), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(103), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(103), // [[, reduce: FieldValue
			reduce(103), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2151
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(104), // ;, reduce: DirectFieldValue
			nil,         // {
			reduce(104), // }, reduce: DirectFieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(104), // >, reduce: DirectFieldValue
			reduce(104), // >>, reduce: DirectFieldValue
			reduce(104), // |, reduce: DirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(104), // ?, reduce: DirectFieldValue
			nil,         // :
			reduce(104), // ||, reduce: DirectFieldValue
			reduce(104), // ^^, reduce: DirectFieldValue
			reduce(104), // &&, reduce: DirectFieldValue
			reduce(104), // ??, reduce: DirectFieldValue
			reduce(104), // ???, reduce: DirectFieldValue
			reduce(104), // =~, reduce: DirectFieldValue
			reduce(104), // !=~, reduce: DirectFieldValue
			reduce(104), // ==, reduce: DirectFieldValue
			reduce(104), // !=, reduce: DirectFieldValue
			reduce(104), // <=>, reduce: DirectFieldValue
			reduce(104), // >=, reduce: DirectFieldValue
			reduce(104), // <, reduce: DirectFieldValue
			reduce(104), // <=, reduce: DirectFieldValue
			reduce(104), // ^, reduce: DirectFieldValue
			reduce(104), // &, reduce: DirectFieldValue
			reduce(104), // <<, reduce: DirectFieldValue
			reduce(104), // >>>, reduce: DirectFieldValue
			reduce(104), // +, reduce: DirectFieldValue
			reduce(104), // -, reduce: DirectFieldValue
			reduce(104), // .+, reduce: DirectFieldValue
			reduce(104), // .-, reduce: DirectFieldValue
			reduce(104), // *, reduce: DirectFieldValue
			reduce(104), // /, reduce: DirectFieldValue
			reduce(104), // //, reduce: DirectFieldValue
			reduce(104), // %, reduce: DirectFieldValue
			reduce(104), // .*, reduce: DirectFieldValue
			reduce(104), // ./, reduce: DirectFieldValue
			reduce(104), // .//, reduce: DirectFieldValue
			reduce(104), // ., reduce: DirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(104), // **, reduce: DirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(104), // [, reduce: DirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(104), // [[, reduce: DirectFieldValue
			reduce(104), // [[[, reduce: DirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2152
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2153
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(106), // ;, reduce: BracedFieldValue
			nil,         // {
			reduce(106), // }, reduce: BracedFieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(106), // >, reduce: BracedFieldValue
			reduce(106), // >>, reduce: BracedFieldValue
			reduce(106), // |, reduce: BracedFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(106), // ?, reduce: BracedFieldValue
			nil,         // :
			reduce(106), // ||, reduce: BracedFieldValue
			reduce(106), // ^^, reduce: BracedFieldValue
			reduce(106), // &&, reduce: BracedFieldValue
			reduce(106), // ??, reduce: BracedFieldValue
			reduce(106), // ???, reduce: BracedFieldValue
			reduce(106), // =~, reduce: BracedFieldValue
			reduce(106), // !=~, reduce: BracedFieldValue
			reduce(106), // ==, reduce: BracedFieldValue
			reduce(106), // !=, reduce: BracedFieldValue
			reduce(106), // <=>, reduce: BracedFieldValue
			reduce(106), // >=, reduce: BracedFieldValue
			reduce(106), // <, reduce: BracedFieldValue
			reduce(106), // <=, reduce: BracedFieldValue
			reduce(106), // ^, reduce: BracedFieldValue
			reduce(106), // &, reduce: BracedFieldValue
			reduce(106), // <<, reduce: BracedFieldValue
			reduce(106), // >>>, reduce: BracedFieldValue
			reduce(106), // +, reduce: BracedFieldValue
			reduce(106), // -, reduce: BracedFieldValue
			reduce(106), // .+, reduce: BracedFieldValue
			reduce(106), // .-, reduce: BracedFieldValue
			reduce(106), // *, reduce: BracedFieldValue
			reduce(106), // /, reduce: BracedFieldValue
			reduce(106), // //, reduce: BracedFieldValue
			reduce(106), // %, reduce: BracedFieldValue
			reduce(106), // .*, reduce: BracedFieldValue
			reduce(106), // ./, reduce: BracedFieldValue
			reduce(106), // .//, reduce: BracedFieldValue
			reduce(106), // ., reduce: BracedFieldValue
			nil,         // !
			nil,         // ~
			reduce(106), // **, reduce: BracedFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(106), // [, reduce: BracedFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(106), // [[, reduce: BracedFieldValue
			reduce(106), // [[[, reduce: BracedFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2154
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2155
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2156
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(109), // ;, reduce: FullSrec
			nil,         // {
			reduce(109), // }, reduce: FullSrec
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(109), // >, reduce: FullSrec
			reduce(109), // >>, reduce: FullSrec
			reduce(109), // |, reduce: FullSrec
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(109), // ?, reduce: FullSrec
			nil,         // :
			reduce(109), // ||, reduce: FullSrec
			reduce(109), // ^^, reduce: FullSrec
			reduce(109), // &&, reduce: FullSrec
			reduce(109), // ??, reduce: FullSrec
			reduce(109), // ???, reduce: FullSrec
			reduce(109), // =~, reduce: FullSrec
			reduce(109), // !=~, reduce: FullSrec
			reduce(109), // ==, reduce: FullSrec
			reduce(109), // !=, reduce: FullSrec
			reduce(109), // <=>, reduce: FullSrec
			reduce(109), // >=, reduce: FullSrec
			reduce(109), // <, reduce: FullSrec
			reduce(109), // <=, reduce: FullSrec
			reduce(109), // ^, reduce: FullSrec
			reduce(109), // &, reduce: FullSrec
			reduce(109), // <<, reduce: FullSrec
			reduce(109), // >>>, reduce: FullSrec
			reduce(109), // +, reduce: FullSrec
			reduce(109), // -, reduce: FullSrec
			reduce(109), // .+, reduce: FullSrec
			reduce(109), // .-, reduce: FullSrec
			reduce(109), // *, reduce: FullSrec
			reduce(109), // /, reduce: FullSrec
			reduce(109), // //, reduce: FullSrec
			reduce(109), // %, reduce: FullSrec
			reduce(109), // .*, reduce: FullSrec
			reduce(109), // ./, reduce: FullSrec
			reduce(109), // .//, reduce: FullSrec
			reduce(109), // ., reduce: FullSrec
			nil,         // !
			nil,         // ~
			reduce(109), // **, reduce: FullSrec
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(109), // [, reduce: FullSrec
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(109), // [[, reduce: FullSrec
			reduce(109), // [[[, reduce: FullSrec
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2157
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(212), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(212), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(212), // >, reduce: MlrvalOrFunction
			reduce(212), // >>, reduce: MlrvalOrFunction
			reduce(212), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(212), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(212), // ||, reduce: MlrvalOrFunction
			reduce(212), // ^^, reduce: MlrvalOrFunction
			reduce(212), // &&, reduce: MlrvalOrFunction
			reduce(212), // ??, reduce: MlrvalOrFunction
			reduce(212), // ???, reduce: MlrvalOrFunction
			reduce(212), // =~, reduce: MlrvalOrFunction
			reduce(212), // !=~, reduce: MlrvalOrFunction
			reduce(212), // ==, reduce: MlrvalOrFunction
			reduce(212), // !=, reduce: MlrvalOrFunction
			reduce(212), // <=>, reduce: MlrvalOrFunction
			reduce(212), // >=, reduce: MlrvalOrFunction
			reduce(212), // <, reduce: MlrvalOrFunction
			reduce(212), // <=, reduce: MlrvalOrFunction
			reduce(212), // ^, reduce: MlrvalOrFunction
			reduce(212), // &, reduce: MlrvalOrFunction
			reduce(212), // <<, reduce: MlrvalOrFunction
			reduce(212), // >>>, reduce: MlrvalOrFunction
			reduce(212), // +, reduce: MlrvalOrFunction
			reduce(212), // -, reduce: MlrvalOrFunction
			reduce(212), // .+, reduce: MlrvalOrFunction
			reduce(212), // .-, reduce: MlrvalOrFunction
			reduce(212), // *, reduce: MlrvalOrFunction
			reduce(212), // /, reduce: MlrvalOrFunction
			reduce(212), // //, reduce: MlrvalOrFunction
			reduce(212), // %, reduce: MlrvalOrFunction
			reduce(212), // .*, reduce: MlrvalOrFunction
			reduce(212), // ./, reduce: MlrvalOrFunction
			reduce(212), // .//, reduce: MlrvalOrFunction
			reduce(212), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(212), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(212), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(212), // [[, reduce: MlrvalOrFunction
			reduce(212), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2158
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(113), // ;, reduce: DirectOosvarValue
			nil,         // {
			reduce(113), // }, reduce: DirectOosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(113), // >, reduce: DirectOosvarValue
			reduce(113), // >>, reduce: DirectOosvarValue
			reduce(113), // |, reduce: DirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(113), // ?, reduce: DirectOosvarValue
			nil,         // :
			reduce(113), // ||, reduce: DirectOosvarValue
			reduce(113), // ^^, reduce: DirectOosvarValue
			reduce(113), // &&, reduce: DirectOosvarValue
			reduce(113), // ??, reduce: DirectOosvarValue
			reduce(113), // ???, reduce: DirectOosvarValue
			reduce(113), // =~, reduce: DirectOosvarValue
			reduce(113), // !=~, reduce: DirectOosvarValue
			reduce(113), // ==, reduce: DirectOosvarValue
			reduce(113), // !=, reduce: DirectOosvarValue
			reduce(113), // <=>, reduce: DirectOosvarValue
			reduce(113), // >=, reduce: DirectOosvarValue
			reduce(113), // <, reduce: DirectOosvarValue
			reduce(113), // <=, reduce: DirectOosvarValue
			reduce(113), // ^, reduce: DirectOosvarValue
			reduce(113), // &, reduce: DirectOosvarValue
			reduce(113), // <<, reduce: DirectOosvarValue
			reduce(113), // >>>, reduce: DirectOosvarValue
			reduce(113), // +, reduce: DirectOosvarValue
			reduce(113), // -, reduce: DirectOosvarValue
			reduce(113), // .+, reduce: DirectOosvarValue
			reduce(113), // .-, reduce: DirectOosvarValue
			reduce(113), // *, reduce: DirectOosvarValue
			reduce(113), // /, reduce: DirectOosvarValue
			reduce(113), // //, reduce: DirectOosvarValue
			reduce(113), // %, reduce: DirectOosvarValue
			reduce(113), // .*, reduce: DirectOosvarValue
			reduce(113), // ./, reduce: DirectOosvarValue
			reduce(113), // .//, reduce: DirectOosvarValue
			reduce(113), // ., reduce: DirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(113), // **, reduce: DirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(113), // [, reduce: DirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(113), // [[, reduce: DirectOosvarValue
			reduce(113), // [[[, reduce: DirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2159
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2160
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(115), // ;, reduce: BracedOosvarValue
			nil,         // {
			reduce(115), // }, reduce: BracedOosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(115), // >, reduce: BracedOosvarValue
			reduce(115), // >>, reduce: BracedOosvarValue
			reduce(115), // |, reduce: BracedOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(115), // ?, reduce: BracedOosvarValue
			nil,         // :
			reduce(115), // ||, reduce: BracedOosvarValue
			reduce(115), // ^^, reduce: BracedOosvarValue
			reduce(115), // &&, reduce: BracedOosvarValue
			reduce(115), // ??, reduce: BracedOosvarValue
			reduce(115), // ???, reduce: BracedOosvarValue
			reduce(115), // =~, reduce: BracedOosvarValue
			reduce(115), // !=~, reduce: BracedOosvarValue
			reduce(115), // ==, reduce: BracedOosvarValue
			reduce(115), // !=, reduce: BracedOosvarValue
			reduce(115), // <=>, reduce: BracedOosvarValue
			reduce(115), // >=, reduce: BracedOosvarValue
			reduce(115), // <, reduce: BracedOosvarValue
			reduce(115), // <=, reduce: BracedOosvarValue
			reduce(115), // ^, reduce: BracedOosvarValue
			reduce(115), // &, reduce: BracedOosvarValue
			reduce(115), // <<, reduce: BracedOosvarValue
			reduce(115), // >>>, reduce: BracedOosvarValue
			reduce(115), // +, reduce: BracedOosvarValue
			reduce(115), // -, reduce: BracedOosvarValue
			reduce(115), // .+, reduce: BracedOosvarValue
			reduce(115), // .-, reduce: BracedOosvarValue
			reduce(115), // *, reduce: BracedOosvarValue
			reduce(115), // /, reduce: BracedOosvarValue
			reduce(115), // //, reduce: BracedOosvarValue
			reduce(115), // %, reduce: BracedOosvarValue
			reduce(115), // .*, reduce: BracedOosvarValue
			reduce(115), // ./, reduce: BracedOosvarValue
			reduce(115), // .//, reduce: BracedOosvarValue
			reduce(115), // ., reduce: BracedOosvarValue
			nil,         // !
			nil,         // ~
			reduce(115), // **, reduce: BracedOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(115), // [, reduce: BracedOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(115), // [[, reduce: BracedOosvarValue
			reduce(115), // [[[, reduce: BracedOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2161
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(116), // ;, reduce: FullOosvar
			nil,         // {
			reduce(116), // }, reduce: FullOosvar
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(116), // >, reduce: FullOosvar
			reduce(116), // >>, reduce: FullOosvar
			reduce(116), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(116), // ?, reduce: FullOosvar
			nil,         // :
			reduce(116), // ||, reduce: FullOosvar
			reduce(116), // ^^, reduce: FullOosvar
			reduce(116), // &&, reduce: FullOosvar
			reduce(116), // ??, reduce: FullOosvar
			reduce(116), // ???, reduce: FullOosvar
			reduce(116), // =~, reduce: FullOosvar
			reduce(116), // !=~, reduce: FullOosvar
			reduce(116), // ==, reduce: FullOosvar
			reduce(116), // !=, reduce: FullOosvar
			reduce(116), // <=>, reduce: FullOosvar
			reduce(116), // >=, reduce: FullOosvar
			reduce(116), // <, reduce: FullOosvar
			reduce(116), // <=, reduce: FullOosvar
			reduce(116), // ^, reduce: FullOosvar
			reduce(116), // &, reduce: FullOosvar
			reduce(116), // <<, reduce: FullOosvar
			reduce(116), // >>>, reduce: FullOosvar
			reduce(116), // +, reduce: FullOosvar
			reduce(116), // -, reduce: FullOosvar
			reduce(116), // .+, reduce: FullOosvar
			reduce(116), // .-, reduce: FullOosvar
			reduce(116), // *, reduce: FullOosvar
			reduce(116), // /, reduce: FullOosvar
			reduce(116), // //, reduce: FullOosvar
			reduce(116), // %, reduce: FullOosvar
			reduce(116), // .*, reduce: FullOosvar
			reduce(116), // ./, reduce: FullOosvar
			reduce(116), // .//, reduce: FullOosvar
			reduce(116), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(116), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(116), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(116), // [[, reduce: FullOosvar
			reduce(116), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2162
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(117), // ;, reduce: FullOosvar
			nil,         // {
			reduce(117), // }, reduce: FullOosvar
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(117), // >, reduce: FullOosvar
			reduce(117), // >>, reduce: FullOosvar
			reduce(117), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(117), // ?, reduce: FullOosvar
			nil,         // :
			reduce(117), // ||, reduce: FullOosvar
			reduce(117), // ^^, reduce: FullOosvar
			reduce(117), // &&, reduce: FullOosvar
			reduce(117), // ??, reduce: FullOosvar
			reduce(117), // ???, reduce: FullOosvar
			reduce(117), // =~, reduce: FullOosvar
			reduce(117), // !=~, reduce: FullOosvar
			reduce(117), // ==, reduce: FullOosvar
			reduce(117), // !=, reduce: FullOosvar
			reduce(117), // <=>, reduce: FullOosvar
			reduce(117), // >=, reduce: FullOosvar
			reduce(117), // <, reduce: FullOosvar
			reduce(117), // <=, reduce: FullOosvar
			reduce(117), // ^, reduce: FullOosvar
			reduce(117), // &, reduce: FullOosvar
			reduce(117), // <<, reduce: FullOosvar
			reduce(117), // >>>, reduce: FullOosvar
			reduce(117), // +, reduce: FullOosvar
			reduce(117), // -, reduce: FullOosvar
			reduce(117), // .+, reduce: FullOosvar
			reduce(117), // .-, reduce: FullOosvar
			reduce(117), // *, reduce: FullOosvar
			reduce(117), // /, reduce: FullOosvar
			reduce(117), // //, reduce: FullOosvar
			reduce(117), // %, reduce: FullOosvar
			reduce(117), // .*, reduce: FullOosvar
			reduce(117), // ./, reduce: FullOosvar
			reduce(117), // .//, reduce: FullOosvar
			reduce(117), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(117), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(117), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(117), // [[, reduce: FullOosvar
			reduce(117), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2163
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(118), // ;, reduce: LocalVariable
			nil,         // {
			reduce(118), // }, reduce: LocalVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(118), // >, reduce: LocalVariable
			reduce(118), // >>, reduce: LocalVariable
			reduce(118), // |, reduce: LocalVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(270), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(118), // ?, reduce: LocalVariable
			nil,         // :
			reduce(118), // ||, reduce: LocalVariable
			reduce(118), // ^^, reduce: LocalVariable
			reduce(118), // &&, reduce: LocalVariable
			reduce(118), // ??, reduce: LocalVariable
			reduce(118), // ???, reduce: LocalVariable
			reduce(118), // =~, reduce: LocalVariable
			reduce(118), // !=~, reduce: LocalVariable
			reduce(118), // ==, reduce: LocalVariable
			reduce(118), // !=, reduce: LocalVariable
			reduce(118), // <=>, reduce: LocalVariable
			reduce(118), // >=, reduce: LocalVariable
			reduce(118), // <, reduce: LocalVariable
			reduce(118), // <=, reduce: LocalVariable
			reduce(118), // ^, reduce: LocalVariable
			reduce(118), // &, reduce: LocalVariable
			reduce(118), // <<, reduce: LocalVariable
			reduce(118), // >>>, reduce: LocalVariable
			reduce(118), // +, reduce: LocalVariable
			reduce(118), // -, reduce: LocalVariable
			reduce(118), // .+, reduce: LocalVariable
			reduce(118), // .-, reduce: LocalVariable
			reduce(118), // *, reduce: LocalVariable
			reduce(118), // /, reduce: LocalVariable
			reduce(118), // //, reduce: LocalVariable
			reduce(118), // %, reduce: LocalVariable
			reduce(118), // .*, reduce: LocalVariable
			reduce(118), // ./, reduce: LocalVariable
			reduce(118), // .//, reduce: LocalVariable
			reduce(118), // ., reduce: LocalVariable
			nil,         // !
			nil,         // ~
			reduce(118), // **, reduce: LocalVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(118), // [, reduce: LocalVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(118), // [[, reduce: LocalVariable
			reduce(118), // [[[, reduce: LocalVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2164
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(147), // ;, reduce: Rvalue
			nil,         // {
			reduce(147), // }, reduce: Rvalue
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2165
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(148), // ;, reduce: PrecedenceChainStart
			nil,         // {
			reduce(148), // }, reduce: PrecedenceChainStart
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2166
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(150), // ;, reduce: TernaryTerm
			nil,         // {
			reduce(150), // }, reduce: TernaryTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			shift(3065), // ?
			nil,         // :
			shift(3066), // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2167
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(152), // ;, reduce: LogicalOrTerm
			nil,         // {
			reduce(152), // }, reduce: LogicalOrTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(152), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(152), // ||, reduce: LogicalOrTerm
			shift(3067), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2168
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(154), // ;, reduce: LogicalXORTerm
			nil,         // {
			reduce(154), // }, reduce: LogicalXORTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(154), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(154), // ||, reduce: LogicalXORTerm
			reduce(154), // ^^, reduce: LogicalXORTerm
			shift(3068), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2169
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(156), // ;, reduce: LogicalAndTerm
			nil,         // {
			reduce(156), // }, reduce: LogicalAndTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(156), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(156), // ||, reduce: LogicalAndTerm
			reduce(156), // ^^, reduce: LogicalAndTerm
			reduce(156), // &&, reduce: LogicalAndTerm
			shift(3069), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2170
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(158), // ;, reduce: AbsentCoalesceTerm
			nil,         // {
			reduce(158), // }, reduce: AbsentCoalesceTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(158), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(158), // ||, reduce: AbsentCoalesceTerm
			reduce(158), // ^^, reduce: AbsentCoalesceTerm
			reduce(158), // &&, reduce: AbsentCoalesceTerm
			reduce(158), // ??, reduce: AbsentCoalesceTerm
			shift(3070), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2171
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(160), // ;, reduce: EmptyCoalesceTerm
			nil,         // {
			reduce(160), // }, reduce: EmptyCoalesceTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(160), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(160), // ||, reduce: EmptyCoalesceTerm
			reduce(160), // ^^, reduce: EmptyCoalesceTerm
			reduce(160), // &&, reduce: EmptyCoalesceTerm
			reduce(160), // ??, reduce: EmptyCoalesceTerm
			reduce(160), // ???, reduce: EmptyCoalesceTerm
			shift(3071), // =~
			shift(3072), // !=~
			shift(3073), // ==
			shift(3074), // !=
			shift(3075), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2172
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(166), // ;, reduce: EqneTerm
			nil,         // {
			reduce(166), // }, reduce: EqneTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(3076), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(166), // ?, reduce: EqneTerm
			nil,         // :
			reduce(166), // ||, reduce: EqneTerm
			reduce(166), // ^^, reduce: EqneTerm
			reduce(166), // &&, reduce: EqneTerm
			reduce(166), // ??, reduce: EqneTerm
			reduce(166), // ???, reduce: EqneTerm
			reduce(166), // =~, reduce: EqneTerm
			reduce(166), // !=~, reduce: EqneTerm
			reduce(166), // ==, reduce: EqneTerm
			reduce(166), // !=, reduce: EqneTerm
			reduce(166), // <=>, reduce: EqneTerm
			shift(3077), // >=
			shift(3078), // <
			shift(3079), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2173
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(171), // ;, reduce: CmpTerm
			nil,         // {
			reduce(171), // }, reduce: CmpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(171), // >, reduce: CmpTerm
			nil,         // >>
			shift(3080), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(171), // ?, reduce: CmpTerm
			nil,         // :
			reduce(171), // ||, reduce: CmpTerm
			reduce(171), // ^^, reduce: CmpTerm
			reduce(171), // &&, reduce: CmpTerm
			reduce(171), // ??, reduce: CmpTerm
			reduce(171), // ???, reduce: CmpTerm
			reduce(171), // =~, reduce: CmpTerm
			reduce(171), // !=~, reduce: CmpTerm
			reduce(171), // ==, reduce: CmpTerm
			reduce(171), // !=, reduce: CmpTerm
			reduce(171), // <=>, reduce: CmpTerm
			reduce(171), // >=, reduce: CmpTerm
			reduce(171), // <, reduce: CmpTerm
			reduce(171), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2174
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(173), // ;, reduce: BitwiseORTerm
			nil,         // {
			reduce(173), // }, reduce: BitwiseORTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(173), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(173), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(173), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(173), // ||, reduce: BitwiseORTerm
			reduce(173), // ^^, reduce: BitwiseORTerm
			reduce(173), // &&, reduce: BitwiseORTerm
			reduce(173), // ??, reduce: BitwiseORTerm
			reduce(173), // ???, reduce: BitwiseORTerm
			reduce(173), // =~, reduce: BitwiseORTerm
			reduce(173), // !=~, reduce: BitwiseORTerm
			reduce(173), // ==, reduce: BitwiseORTerm
			reduce(173), // !=, reduce: BitwiseORTerm
			reduce(173), // <=>, reduce: BitwiseORTerm
			reduce(173), // >=, reduce: BitwiseORTerm
			reduce(173), // <, reduce: BitwiseORTerm
			reduce(173), // <=, reduce: BitwiseORTerm
			shift(3081), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2175
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(175), // ;, reduce: BitwiseXORTerm
			nil,         // {
			reduce(175), // }, reduce: BitwiseXORTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(175), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(175), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(175), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(175), // ||, reduce: BitwiseXORTerm
			reduce(175), // ^^, reduce: BitwiseXORTerm
			reduce(175), // &&, reduce: BitwiseXORTerm
			reduce(175), // ??, reduce: BitwiseXORTerm
			reduce(175), // ???, reduce: BitwiseXORTerm
			reduce(175), // =~, reduce: BitwiseXORTerm
			reduce(175), // !=~, reduce: BitwiseXORTerm
			reduce(175), // ==, reduce: BitwiseXORTerm
			reduce(175), // !=, reduce: BitwiseXORTerm
			reduce(175), // <=>, reduce: BitwiseXORTerm
			reduce(175), // >=, reduce: BitwiseXORTerm
			reduce(175), // <, reduce: BitwiseXORTerm
			reduce(175), // <=, reduce: BitwiseXORTerm
			reduce(175), // ^, reduce: BitwiseXORTerm
			shift(3082), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2176
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(177), // ;, reduce: BitwiseANDTerm
			nil,         // {
			reduce(177), // }, reduce: BitwiseANDTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(177), // >, reduce: BitwiseANDTerm
			shift(3083), // >>
			reduce(177), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(177), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(177), // ||, reduce: BitwiseANDTerm
			reduce(177), // ^^, reduce: BitwiseANDTerm
			reduce(177), // &&, reduce: BitwiseANDTerm
			reduce(177), // ??, reduce: BitwiseANDTerm
			reduce(177), // ???, reduce: BitwiseANDTerm
			reduce(177), // =~, reduce: BitwiseANDTerm
			reduce(177), // !=~, reduce: BitwiseANDTerm
			reduce(177), // ==, reduce: BitwiseANDTerm
			reduce(177), // !=, reduce: BitwiseANDTerm
			reduce(177), // <=>, reduce: BitwiseANDTerm
			reduce(177), // >=, reduce: BitwiseANDTerm
			reduce(177), // <, reduce: BitwiseANDTerm
			reduce(177), // <=, reduce: BitwiseANDTerm
			reduce(177), // ^, reduce: BitwiseANDTerm
			reduce(177), // &, reduce: BitwiseANDTerm
			shift(3084), // <<
			shift(3085), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2177
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(181), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			reduce(181), // }, reduce: BitwiseShiftTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(181), // >, reduce: BitwiseShiftTerm
			reduce(181), // >>, reduce: BitwiseShiftTerm
			reduce(181), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(181), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(181), // ||, reduce: BitwiseShiftTerm
			reduce(181), // ^^, reduce: BitwiseShiftTerm
			reduce(181), // &&, reduce: BitwiseShiftTerm
			reduce(181), // ??, reduce: BitwiseShiftTerm
			reduce(181), // ???, reduce: BitwiseShiftTerm
			reduce(181), // =~, reduce: BitwiseShiftTerm
			reduce(181), // !=~, reduce: BitwiseShiftTerm
			reduce(181), // ==, reduce: BitwiseShiftTerm
			reduce(181), // !=, reduce: BitwiseShiftTerm
			reduce(181), // <=>, reduce: BitwiseShiftTerm
			reduce(181), // >=, reduce: BitwiseShiftTerm
			reduce(181), // <, reduce: BitwiseShiftTerm
			reduce(181), // <=, reduce: BitwiseShiftTerm
			reduce(181), // ^, reduce: BitwiseShiftTerm
			reduce(181), // &, reduce: BitwiseShiftTerm
			reduce(181), // <<, reduce: BitwiseShiftTerm
			reduce(181), // >>>, reduce: BitwiseShiftTerm
			shift(3086), // +
			shift(3087), // -
			shift(3088), // .+
			shift(3089), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2178
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2179
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(186), // ;, reduce: AddsubdotTerm
			nil,         // {
			reduce(186), // }, reduce: AddsubdotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(186), // >, reduce: AddsubdotTerm
			reduce(186), // >>, reduce: AddsubdotTerm
			reduce(186), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(186), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(186), // ||, reduce: AddsubdotTerm
			reduce(186), // ^^, reduce: AddsubdotTerm
			reduce(186), // &&, reduce: AddsubdotTerm
			reduce(186), // ??, reduce: AddsubdotTerm
			reduce(186), // ???, reduce: AddsubdotTerm
			reduce(186), // =~, reduce: AddsubdotTerm
			reduce(186), // !=~, reduce: AddsubdotTerm
			reduce(186), // ==, reduce: AddsubdotTerm
			reduce(186), // !=, reduce: AddsubdotTerm
			reduce(186), // <=>, reduce: AddsubdotTerm
			reduce(186), // >=, reduce: AddsubdotTerm
			reduce(186), // <, reduce: AddsubdotTerm
			reduce(186), // <=, reduce: AddsubdotTerm
			reduce(186), // ^, reduce: AddsubdotTerm
			reduce(186), // &, reduce: AddsubdotTerm
			reduce(186), // <<, reduce: AddsubdotTerm
			reduce(186), // >>>, reduce: AddsubdotTerm
			reduce(186), // +, reduce: AddsubdotTerm
			reduce(186), // -, reduce: AddsubdotTerm
			reduce(186), // .+, reduce: AddsubdotTerm
			reduce(186), // .-, reduce: AddsubdotTerm
			shift(3091), // *
			shift(3092), // /
			shift(3093), // //
			shift(3094), // %
			shift(3095), // .*
			shift(3096), // ./
			shift(3097), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2180
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2181
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2182
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2183
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(194), // ;, reduce: MuldivTerm
			nil,         // {
			reduce(194), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(194), // >, reduce: MuldivTerm
			reduce(194), // >>, reduce: MuldivTerm
			reduce(194), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(194), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(194), // ||, reduce: MuldivTerm
			reduce(194), // ^^, reduce: MuldivTerm
			reduce(194), // &&, reduce: MuldivTerm
			reduce(194), // ??, reduce: MuldivTerm
			reduce(194), // ???, reduce: MuldivTerm
			reduce(194), // =~, reduce: MuldivTerm
			reduce(194), // !=~, reduce: MuldivTerm
			reduce(194), // ==, reduce: MuldivTerm
			reduce(194), // !=, reduce: MuldivTerm
			reduce(194), // <=>, reduce: MuldivTerm
			reduce(194), // >=, reduce: MuldivTerm
			reduce(194), // <, reduce: MuldivTerm
			reduce(194), // <=, reduce: MuldivTerm
			reduce(194), // ^, reduce: MuldivTerm
			reduce(194), // &, reduce: MuldivTerm
			reduce(194), // <<, reduce: MuldivTerm
			reduce(194), // >>>, reduce: MuldivTerm
			reduce(194), // +, reduce: MuldivTerm
			reduce(194), // -, reduce: MuldivTerm
			reduce(194), // .+, reduce: MuldivTerm
			reduce(194), // .-, reduce: MuldivTerm
			reduce(194), // *, reduce: MuldivTerm
			reduce(194), // /, reduce: MuldivTerm
			reduce(194), // //, reduce: MuldivTerm
			reduce(194), // %, reduce: MuldivTerm
			reduce(194), // .*, reduce: MuldivTerm
			reduce(194), // ./, reduce: MuldivTerm
			reduce(194), // .//, reduce: MuldivTerm
			shift(3101), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2184
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(196), // ;, reduce: DotTerm
			nil,         // {
			reduce(196), // }, reduce: DotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(196), // >, reduce: DotTerm
			reduce(196), // >>, reduce: DotTerm
			reduce(196), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(196), // ?, reduce: DotTerm
			nil,         // :
			reduce(196), // ||, reduce: DotTerm
			reduce(196), // ^^, reduce: DotTerm
			reduce(196), // &&, reduce: DotTerm
			reduce(196), // ??, reduce: DotTerm
			reduce(196), // ???, reduce: DotTerm
			reduce(196), // =~, reduce: DotTerm
			reduce(196), // !=~, reduce: DotTerm
			reduce(196), // ==, reduce: DotTerm
			reduce(196), // !=, reduce: DotTerm
			reduce(196), // <=>, reduce: DotTerm
			reduce(196), // >=, reduce: DotTerm
			reduce(196), // <, reduce: DotTerm
			reduce(196), // <=, reduce: DotTerm
			reduce(196), // ^, reduce: DotTerm
			reduce(196), // &, reduce: DotTerm
			reduce(196), // <<, reduce: DotTerm
			reduce(196), // >>>, reduce: DotTerm
			reduce(196), // +, reduce: DotTerm
			reduce(196), // -, reduce: DotTerm
			reduce(196), // .+, reduce: DotTerm
			reduce(196), // .-, reduce: DotTerm
			reduce(196), // *, reduce: DotTerm
			reduce(196), // /, reduce: DotTerm
			reduce(196), // //, reduce: DotTerm
			reduce(196), // %, reduce: DotTerm
			reduce(196), // .*, reduce: DotTerm
			reduce(196), // ./, reduce: DotTerm
			reduce(196), // .//, reduce: DotTerm
			reduce(196), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2185
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2186
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2187
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(203), // ;, reduce: UnaryOpTerm
			nil,         // {
			reduce(203), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(203), // >, reduce: UnaryOpTerm
			reduce(203), // >>, reduce: UnaryOpTerm
			reduce(203), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(203), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(203), // ||, reduce: UnaryOpTerm
			reduce(203), // ^^, reduce: UnaryOpTerm
			reduce(203), // &&, reduce: UnaryOpTerm
			reduce(203), // ??, reduce: UnaryOpTerm
			reduce(203), // ???, reduce: UnaryOpTerm
			reduce(203), // =~, reduce: UnaryOpTerm
			reduce(203), // !=~, reduce: UnaryOpTerm
			reduce(203), // ==, reduce: UnaryOpTerm
			reduce(203), // !=, reduce: UnaryOpTerm
			reduce(203), // <=>, reduce: UnaryOpTerm
			reduce(203), // >=, reduce: UnaryOpTerm
			reduce(203), // <, reduce: UnaryOpTerm
			reduce(203), // <=, reduce: UnaryOpTerm
			reduce(203), // ^, reduce: UnaryOpTerm
			reduce(203), // &, reduce: UnaryOpTerm
			reduce(203), // <<, reduce: UnaryOpTerm
			reduce(203), // >>>, reduce: UnaryOpTerm
			reduce(203), // +, reduce: UnaryOpTerm
			reduce(203), // -, reduce: UnaryOpTerm
			reduce(203), // .+, reduce: UnaryOpTerm
			reduce(203), // .-, reduce: UnaryOpTerm
			reduce(203), // *, reduce: UnaryOpTerm
			reduce(203), // /, reduce: UnaryOpTerm
			reduce(203), // //, reduce: UnaryOpTerm
			reduce(203), // %, reduce: UnaryOpTerm
			reduce(203), // .*, reduce: UnaryOpTerm
			reduce(203), // ./, reduce: UnaryOpTerm
			reduce(203), // .//, reduce: UnaryOpTerm
			reduce(203), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2188
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(207), // ;, reduce: PowTerm
			nil,         // {
			reduce(207), // }, reduce: PowTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(207), // >, reduce: PowTerm
			reduce(207), // >>, reduce: PowTerm
			reduce(207), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(207), // ?, reduce: PowTerm
			nil,         // :
			reduce(207), // ||, reduce: PowTerm
			reduce(207), // ^^, reduce: PowTerm
			reduce(207), // &&, reduce: PowTerm
			reduce(207), // ??, reduce: PowTerm
			reduce(207), // ???, reduce: PowTerm
			reduce(207), // =~, reduce: PowTerm
			reduce(207), // !=~, reduce: PowTerm
			reduce(207), // ==, reduce: PowTerm
			reduce(207), // !=, reduce: PowTerm
			reduce(207), // <=>, reduce: PowTerm
			reduce(207), // >=, reduce: PowTerm
			reduce(207), // <, reduce: PowTerm
			reduce(207), // <=, reduce: PowTerm
			reduce(207), // ^, reduce: PowTerm
			reduce(207), // &, reduce: PowTerm
			reduce(207), // <<, reduce: PowTerm
			reduce(207), // >>>, reduce: PowTerm
			reduce(207), // +, reduce: PowTerm
			reduce(207), // -, reduce: PowTerm
			reduce(207), // .+, reduce: PowTerm
			reduce(207), // .-, reduce: PowTerm
			reduce(207), // *, reduce: PowTerm
			reduce(207), // /, reduce: PowTerm
			reduce(207), // //, reduce: PowTerm
			reduce(207), // %, reduce: PowTerm
			reduce(207), // .*, reduce: PowTerm
			reduce(207), // ./, reduce: PowTerm
			reduce(207), // .//, reduce: PowTerm
			reduce(207), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			shift(3104), // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2189
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(209), // ;, reduce: PrecedenceChainEnd
			nil,         // {
			reduce(209), // }, reduce: PrecedenceChainEnd
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(209), // >, reduce: PrecedenceChainEnd
			reduce(209), // >>, reduce: PrecedenceChainEnd
			reduce(209), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(209), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(209), // ||, reduce: PrecedenceChainEnd
			reduce(209), // ^^, reduce: PrecedenceChainEnd
			reduce(209), // &&, reduce: PrecedenceChainEnd
			reduce(209), // ??, reduce: PrecedenceChainEnd
			reduce(209), // ???, reduce: PrecedenceChainEnd
			reduce(209), // =~, reduce: PrecedenceChainEnd
			reduce(209), // !=~, reduce: PrecedenceChainEnd
			reduce(209), // ==, reduce: PrecedenceChainEnd
			reduce(209), // !=, reduce: PrecedenceChainEnd
			reduce(209), // <=>, reduce: PrecedenceChainEnd
			reduce(209), // >=, reduce: PrecedenceChainEnd
			reduce(209), // <, reduce: PrecedenceChainEnd
			reduce(209), // <=, reduce: PrecedenceChainEnd
			reduce(209), // ^, reduce: PrecedenceChainEnd
			reduce(209), // &, reduce: PrecedenceChainEnd
			reduce(209), // <<, reduce: PrecedenceChainEnd
			reduce(209), // >>>, reduce: PrecedenceChainEnd
			reduce(209), // +, reduce: PrecedenceChainEnd
			reduce(209), // -, reduce: PrecedenceChainEnd
			reduce(209), // .+, reduce: PrecedenceChainEnd
			reduce(209), // .-, reduce: PrecedenceChainEnd
			reduce(209), // *, reduce: PrecedenceChainEnd
			reduce(209), // /, reduce: PrecedenceChainEnd
			reduce(209), // //, reduce: PrecedenceChainEnd
			reduce(209), // %, reduce: PrecedenceChainEnd
			reduce(209), // .*, reduce: PrecedenceChainEnd
			reduce(209), // ./, reduce: PrecedenceChainEnd
			reduce(209), // .//, reduce: PrecedenceChainEnd
			reduce(209), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(209), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(3105), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			shift(3106), // [[
			shift(3107), // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2190
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(215), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(215), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(215), // >, reduce: MlrvalOrFunction
			reduce(215), // >>, reduce: MlrvalOrFunction
			reduce(215), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(215), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(215), // ||, reduce: MlrvalOrFunction
			reduce(215), // ^^, reduce: MlrvalOrFunction
			reduce(215), // &&, reduce: MlrvalOrFunction
			reduce(215), // ??, reduce: MlrvalOrFunction
			reduce(215), // ???, reduce: MlrvalOrFunction
			reduce(215), // =~, reduce: MlrvalOrFunction
			reduce(215), // !=~, reduce: MlrvalOrFunction
			reduce(215), // ==, reduce: MlrvalOrFunction
			reduce(215), // !=, reduce: MlrvalOrFunction
			reduce(215), // <=>, reduce: MlrvalOrFunction
			reduce(215), // >=, reduce: MlrvalOrFunction
			reduce(215), // <, reduce: MlrvalOrFunction
			reduce(215), // <=, reduce: MlrvalOrFunction
			reduce(215), // ^, reduce: MlrvalOrFunction
			reduce(215), // &, reduce: MlrvalOrFunction
			reduce(215), // <<, reduce: MlrvalOrFunction
			reduce(215), // >>>, reduce: MlrvalOrFunction
			reduce(215), // +, reduce: MlrvalOrFunction
			reduce(215), // -, reduce: MlrvalOrFunction
			reduce(215), // .+, reduce: MlrvalOrFunction
			reduce(215), // .-, reduce: MlrvalOrFunction
			reduce(215), // *, reduce: MlrvalOrFunction
			reduce(215), // /, reduce: MlrvalOrFunction
			reduce(215), // //, reduce: MlrvalOrFunction
			reduce(215), // %, reduce: MlrvalOrFunction
			reduce(215), // .*, reduce: MlrvalOrFunction
			reduce(215), // ./, reduce: MlrvalOrFunction
			reduce(215), // .//, reduce: MlrvalOrFunction
			reduce(215), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(215), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(215), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(215), // [[, reduce: MlrvalOrFunction
			reduce(215), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2191
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(216), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(216), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(216), // >, reduce: MlrvalOrFunction
			reduce(216), // >>, reduce: MlrvalOrFunction
			reduce(216), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(216), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(216), // ||, reduce: MlrvalOrFunction
			reduce(216), // ^^, reduce: MlrvalOrFunction
			reduce(216), // &&, reduce: MlrvalOrFunction
			reduce(216), // ??, reduce: MlrvalOrFunction
			reduce(216), // ???, reduce: MlrvalOrFunction
			reduce(216), // =~, reduce: MlrvalOrFunction
			reduce(216), // !=~, reduce: MlrvalOrFunction
			reduce(216), // ==, reduce: MlrvalOrFunction
			reduce(216), // !=, reduce: MlrvalOrFunction
			reduce(216), // <=>, reduce: MlrvalOrFunction
			reduce(216), // >=, reduce: MlrvalOrFunction
			reduce(216), // <, reduce: MlrvalOrFunction
			reduce(216), // <=, reduce: MlrvalOrFunction
			reduce(216), // ^, reduce: MlrvalOrFunction
			reduce(216), // &, reduce: MlrvalOrFunction
			reduce(216), // <<, reduce: MlrvalOrFunction
			reduce(216), // >>>, reduce: MlrvalOrFunction
			reduce(216), // +, reduce: MlrvalOrFunction
			reduce(216), // -, reduce: MlrvalOrFunction
			reduce(216), // .+, reduce: MlrvalOrFunction
			reduce(216), // .-, reduce: MlrvalOrFunction
			reduce(216), // *, reduce: MlrvalOrFunction
			reduce(216), // /, reduce: MlrvalOrFunction
			reduce(216), // //, reduce: MlrvalOrFunction
			reduce(216), // %, reduce: MlrvalOrFunction
			reduce(216), // .*, reduce: MlrvalOrFunction
			reduce(216), // ./, reduce: MlrvalOrFunction
			reduce(216), // .//, reduce: MlrvalOrFunction
			reduce(216), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(216), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(216), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(216), // [[, reduce: MlrvalOrFunction
			reduce(216), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2192
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(217), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(217), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(217), // >, reduce: MlrvalOrFunction
			reduce(217), // >>, reduce: MlrvalOrFunction
			reduce(217), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(217), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(217), // ||, reduce: MlrvalOrFunction
			reduce(217), // ^^, reduce: MlrvalOrFunction
			reduce(217), // &&, reduce: MlrvalOrFunction
			reduce(217), // ??, reduce: MlrvalOrFunction
			reduce(217), // ???, reduce: MlrvalOrFunction
			reduce(217), // =~, reduce: MlrvalOrFunction
			reduce(217), // !=~, reduce: MlrvalOrFunction
			reduce(217), // ==, reduce: MlrvalOrFunction
			reduce(217), // !=, reduce: MlrvalOrFunction
			reduce(217), // <=>, reduce: MlrvalOrFunction
			reduce(217), // >=, reduce: MlrvalOrFunction
			reduce(217), // <, reduce: MlrvalOrFunction
			reduce(217), // <=, reduce: MlrvalOrFunction
			reduce(217), // ^, reduce: MlrvalOrFunction
			reduce(217), // &, reduce: MlrvalOrFunction
			reduce(217), // <<, reduce: MlrvalOrFunction
			reduce(217), // >>>, reduce: MlrvalOrFunction
			reduce(217), // +, reduce: MlrvalOrFunction
			reduce(217), // -, reduce: MlrvalOrFunction
			reduce(217), // .+, reduce: MlrvalOrFunction
			reduce(217), // .-, reduce: MlrvalOrFunction
			reduce(217), // *, reduce: MlrvalOrFunction
			reduce(217), // /, reduce: MlrvalOrFunction
			reduce(217), // //, reduce: MlrvalOrFunction
			reduce(217), // %, reduce: MlrvalOrFunction
			reduce(217), // .*, reduce: MlrvalOrFunction
			reduce(217), // ./, reduce: MlrvalOrFunction
			reduce(217), // .//, reduce: MlrvalOrFunction
			reduce(217), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(217), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(217), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(217), // [[, reduce: MlrvalOrFunction
			reduce(217), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2193
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(218), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(218), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(218), // >, reduce: MlrvalOrFunction
			reduce(218), // >>, reduce: MlrvalOrFunction
			reduce(218), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(218), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(218), // ||, reduce: MlrvalOrFunction
			reduce(218), // ^^, reduce: MlrvalOrFunction
			reduce(218), // &&, reduce: MlrvalOrFunction
			reduce(218), // ??, reduce: MlrvalOrFunction
			reduce(218), // ???, reduce: MlrvalOrFunction
			reduce(218), // =~, reduce: MlrvalOrFunction
			reduce(218), // !=~, reduce: MlrvalOrFunction
			reduce(218), // ==, reduce: MlrvalOrFunction
			reduce(218), // !=, reduce: MlrvalOrFunction
			reduce(218), // <=>, reduce: MlrvalOrFunction
			reduce(218), // >=, reduce: MlrvalOrFunction
			reduce(218), // <, reduce: MlrvalOrFunction
			reduce(218), // <=, reduce: MlrvalOrFunction
			reduce(218), // ^, reduce: MlrvalOrFunction
			reduce(218), // &, reduce: MlrvalOrFunction
			reduce(218), // <<, reduce: MlrvalOrFunction
			reduce(218), // >>>, reduce: MlrvalOrFunction
			reduce(218), // +, reduce: MlrvalOrFunction
			reduce(218), // -, reduce: MlrvalOrFunction
			reduce(218), // .+, reduce: MlrvalOrFunction
			reduce(218), // .-, reduce: MlrvalOrFunction
			reduce(218), // *, reduce: MlrvalOrFunction
			reduce(218), // /, reduce: MlrvalOrFunction
			reduce(218), // //, reduce: MlrvalOrFunction
			reduce(218), // %, reduce: MlrvalOrFunction
			reduce(218), // .*, reduce: MlrvalOrFunction
			reduce(218), // ./, reduce: MlrvalOrFunction
			reduce(218), // .//, reduce: MlrvalOrFunction
			reduce(218), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(218), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(218), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(218), // [[, reduce: MlrvalOrFunction
			reduce(218), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2194
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(219), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(219), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(219), // >, reduce: MlrvalOrFunction
			reduce(219), // >>, reduce: MlrvalOrFunction
			reduce(219), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(219), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(219), // ||, reduce: MlrvalOrFunction
			reduce(219), // ^^, reduce: MlrvalOrFunction
			reduce(219), // &&, reduce: MlrvalOrFunction
			reduce(219), // ??, reduce: MlrvalOrFunction
			reduce(219), // ???, reduce: MlrvalOrFunction
			reduce(219), // =~, reduce: MlrvalOrFunction
			reduce(219), // !=~, reduce: MlrvalOrFunction
			reduce(219), // ==, reduce: MlrvalOrFunction
			reduce(219), // !=, reduce: MlrvalOrFunction
			reduce(219), // <=>, reduce: MlrvalOrFunction
			reduce(219), // >=, reduce: MlrvalOrFunction
			reduce(219), // <, reduce: MlrvalOrFunction
			reduce(219), // <=, reduce: MlrvalOrFunction
			reduce(219), // ^, reduce: MlrvalOrFunction
			reduce(219), // &, reduce: MlrvalOrFunction
			reduce(219), // <<, reduce: MlrvalOrFunction
			reduce(219), // >>>, reduce: MlrvalOrFunction
			reduce(219), // +, reduce: MlrvalOrFunction
			reduce(219), // -, reduce: MlrvalOrFunction
			reduce(219), // .+, reduce: MlrvalOrFunction
			reduce(219), // .-, reduce: MlrvalOrFunction
			reduce(219), // *, reduce: MlrvalOrFunction
			reduce(219), // /, reduce: MlrvalOrFunction
			reduce(219), // //, reduce: MlrvalOrFunction
			reduce(219), // %, reduce: MlrvalOrFunction
			reduce(219), // .*, reduce: MlrvalOrFunction
			reduce(219), // ./, reduce: MlrvalOrFunction
			reduce(219), // .//, reduce: MlrvalOrFunction
			reduce(219), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(219), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(219), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(219), // [[, reduce: MlrvalOrFunction
			reduce(219), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2195
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(220), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(220), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(220), // >, reduce: MlrvalOrFunction
			reduce(220), // >>, reduce: MlrvalOrFunction
			reduce(220), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(220), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(220), // ||, reduce: MlrvalOrFunction
			reduce(220), // ^^, reduce: MlrvalOrFunction
			reduce(220), // &&, reduce: MlrvalOrFunction
			reduce(220), // ??, reduce: MlrvalOrFunction
			reduce(220), // ???, reduce: MlrvalOrFunction
			reduce(220), // =~, reduce: MlrvalOrFunction
			reduce(220), // !=~, reduce: MlrvalOrFunction
			reduce(220), // ==, reduce: MlrvalOrFunction
			reduce(220), // !=, reduce: MlrvalOrFunction
			reduce(220), // <=>, reduce: MlrvalOrFunction
			reduce(220), // >=, reduce: MlrvalOrFunction
			reduce(220), // <, reduce: MlrvalOrFunction
			reduce(220), // <=, reduce: MlrvalOrFunction
			reduce(220), // ^, reduce: MlrvalOrFunction
			reduce(220), // &, reduce: MlrvalOrFunction
			reduce(220), // <<, reduce: MlrvalOrFunction
			reduce(220), // >>>, reduce: MlrvalOrFunction
			reduce(220), // +, reduce: MlrvalOrFunction
			reduce(220), // -, reduce: MlrvalOrFunction
			reduce(220), // .+, reduce: MlrvalOrFunction
			reduce(220), // .-, reduce: MlrvalOrFunction
			reduce(220), // *, reduce: MlrvalOrFunction
			reduce(220), // /, reduce: MlrvalOrFunction
			reduce(220), // //, reduce: MlrvalOrFunction
			reduce(220), // %, reduce: MlrvalOrFunction
			reduce(220), // .*, reduce: MlrvalOrFunction
			reduce(220), // ./, reduce: MlrvalOrFunction
			reduce(220), // .//, reduce: MlrvalOrFunction
			reduce(220), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(220), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(220), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(220), // [[, reduce: MlrvalOrFunction
			reduce(220), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2196
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(221), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(221), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(221), // >, reduce: MlrvalOrFunction
			reduce(221), // >>, reduce: MlrvalOrFunction
			reduce(221), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(221), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(221), // ||, reduce: MlrvalOrFunction
			reduce(221), // ^^, reduce: MlrvalOrFunction
			reduce(221), // &&, reduce: MlrvalOrFunction
			reduce(221), // ??, reduce: MlrvalOrFunction
			reduce(221), // ???, reduce: MlrvalOrFunction
			reduce(221), // =~, reduce: MlrvalOrFunction
			reduce(221), // !=~, reduce: MlrvalOrFunction
			reduce(221), // ==, reduce: MlrvalOrFunction
			reduce(221), // !=, reduce: MlrvalOrFunction
			reduce(221), // <=>, reduce: MlrvalOrFunction
			reduce(221), // >=, reduce: MlrvalOrFunction
			reduce(221), // <, reduce: MlrvalOrFunction
			reduce(221), // <=, reduce: MlrvalOrFunction
			reduce(221), // ^, reduce: MlrvalOrFunction
			reduce(221), // &, reduce: MlrvalOrFunction
			reduce(221), // <<, reduce: MlrvalOrFunction
			reduce(221), // >>>, reduce: MlrvalOrFunction
			reduce(221), // +, reduce: MlrvalOrFunction
			reduce(221), // -, reduce: MlrvalOrFunction
			reduce(221), // .+, reduce: MlrvalOrFunction
			reduce(221), // .-, reduce: MlrvalOrFunction
			reduce(221), // *, reduce: MlrvalOrFunction
			reduce(221), // /, reduce: MlrvalOrFunction
			reduce(221), // //, reduce: MlrvalOrFunction
			reduce(221), // %, reduce: MlrvalOrFunction
			reduce(221), // .*, reduce: MlrvalOrFunction
			reduce(221), // ./, reduce: MlrvalOrFunction
			reduce(221), // .//, reduce: MlrvalOrFunction
			reduce(221), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(221), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(221), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(221), // [[, reduce: MlrvalOrFunction
			reduce(221), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2197
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(222), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(222), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(222), // >, reduce: MlrvalOrFunction
			reduce(222), // >>, reduce: MlrvalOrFunction
			reduce(222), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(222), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(222), // ||, reduce: MlrvalOrFunction
			reduce(222), // ^^, reduce: MlrvalOrFunction
			reduce(222), // &&, reduce: MlrvalOrFunction
			reduce(222), // ??, reduce: MlrvalOrFunction
			reduce(222), // ???, reduce: MlrvalOrFunction
			reduce(222), // =~, reduce: MlrvalOrFunction
			reduce(222), // !=~, reduce: MlrvalOrFunction
			reduce(222), // ==, reduce: MlrvalOrFunction
			reduce(222), // !=, reduce: MlrvalOrFunction
			reduce(222), // <=>, reduce: MlrvalOrFunction
			reduce(222), // >=, reduce: MlrvalOrFunction
			reduce(222), // <, reduce: MlrvalOrFunction
			reduce(222), // <=, reduce: MlrvalOrFunction
			reduce(222), // ^, reduce: MlrvalOrFunction
			reduce(222), // &, reduce: MlrvalOrFunction
			reduce(222), // <<, reduce: MlrvalOrFunction
			reduce(222), // >>>, reduce: MlrvalOrFunction
			reduce(222), // +, reduce: MlrvalOrFunction
			reduce(222), // -, reduce: MlrvalOrFunction
			reduce(222), // .+, reduce: MlrvalOrFunction
			reduce(222), // .-, reduce: MlrvalOrFunction
			reduce(222), // *, reduce: MlrvalOrFunction
			reduce(222), // /, reduce: MlrvalOrFunction
			reduce(222), // //, reduce: MlrvalOrFunction
			reduce(222), // %, reduce: MlrvalOrFunction
			reduce(222), // .*, reduce: MlrvalOrFunction
			reduce(222), // ./, reduce: MlrvalOrFunction
			reduce(222), // .//, reduce: MlrvalOrFunction
			reduce(222), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(222), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(222), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(222), // [[, reduce: MlrvalOrFunction
			reduce(222), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2198
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(223), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(223), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(223), // >, reduce: MlrvalOrFunction
			reduce(223), // >>, reduce: MlrvalOrFunction
			reduce(223), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(223), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(223), // ||, reduce: MlrvalOrFunction
			reduce(223), // ^^, reduce: MlrvalOrFunction
			reduce(223), // &&, reduce: MlrvalOrFunction
			reduce(223), // ??, reduce: MlrvalOrFunction
			reduce(223), // ???, reduce: MlrvalOrFunction
			reduce(223), // =~, reduce: MlrvalOrFunction
			reduce(223), // !=~, reduce: MlrvalOrFunction
			reduce(223), // ==, reduce: MlrvalOrFunction
			reduce(223), // !=, reduce: MlrvalOrFunction
			reduce(223), // <=>, reduce: MlrvalOrFunction
			reduce(223), // >=, reduce: MlrvalOrFunction
			reduce(223), // <, reduce: MlrvalOrFunction
			reduce(223), // <=, reduce: MlrvalOrFunction
			reduce(223), // ^, reduce: MlrvalOrFunction
			reduce(223), // &, reduce: MlrvalOrFunction
			reduce(223), // <<, reduce: MlrvalOrFunction
			reduce(223), // >>>, reduce: MlrvalOrFunction
			reduce(223), // +, reduce: MlrvalOrFunction
			reduce(223), // -, reduce: MlrvalOrFunction
			reduce(223), // .+, reduce: MlrvalOrFunction
			reduce(223), // .-, reduce: MlrvalOrFunction
			reduce(223), // *, reduce: MlrvalOrFunction
			reduce(223), // /, reduce: MlrvalOrFunction
			reduce(223), // //, reduce: MlrvalOrFunction
			reduce(223), // %, reduce: MlrvalOrFunction
			reduce(223), // .*, reduce: MlrvalOrFunction
			reduce(223), // ./, reduce: MlrvalOrFunction
			reduce(223), // .//, reduce: MlrvalOrFunction
			reduce(223), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(223), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(223), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(223), // [[, reduce: MlrvalOrFunction
			reduce(223), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2199
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(224), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(224), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(224), // >, reduce: MlrvalOrFunction
			reduce(224), // >>, reduce: MlrvalOrFunction
			reduce(224), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(224), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(224), // ||, reduce: MlrvalOrFunction
			reduce(224), // ^^, reduce: MlrvalOrFunction
			reduce(224), // &&, reduce: MlrvalOrFunction
			reduce(224), // ??, reduce: MlrvalOrFunction
			reduce(224), // ???, reduce: MlrvalOrFunction
			reduce(224), // =~, reduce: MlrvalOrFunction
			reduce(224), // !=~, reduce: MlrvalOrFunction
			reduce(224), // ==, reduce: MlrvalOrFunction
			reduce(224), // !=, reduce: MlrvalOrFunction
			reduce(224), // <=>, reduce: MlrvalOrFunction
			reduce(224), // >=, reduce: MlrvalOrFunction
			reduce(224), // <, reduce: MlrvalOrFunction
			reduce(224), // <=, reduce: MlrvalOrFunction
			reduce(224), // ^, reduce: MlrvalOrFunction
			reduce(224), // &, reduce: MlrvalOrFunction
			reduce(224), // <<, reduce: MlrvalOrFunction
			reduce(224), // >>>, reduce: MlrvalOrFunction
			reduce(224), // +, reduce: MlrvalOrFunction
			reduce(224), // -, reduce: MlrvalOrFunction
			reduce(224), // .+, reduce: MlrvalOrFunction
			reduce(224), // .-, reduce: MlrvalOrFunction
			reduce(224), // *, reduce: MlrvalOrFunction
			reduce(224), // /, reduce: MlrvalOrFunction
			reduce(224), // //, reduce: MlrvalOrFunction
			reduce(224), // %, reduce: MlrvalOrFunction
			reduce(224), // .*, reduce: MlrvalOrFunction
			reduce(224), // ./, reduce: MlrvalOrFunction
			reduce(224), // .//, reduce: MlrvalOrFunction
			reduce(224), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(224), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(224), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(224), // [[, reduce: MlrvalOrFunction
			reduce(224), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2200
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(225), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(225), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(225), // >, reduce: MlrvalOrFunction
			reduce(225), // >>, reduce: MlrvalOrFunction
			reduce(225), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(225), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(225), // ||, reduce: MlrvalOrFunction
			reduce(225), // ^^, reduce: MlrvalOrFunction
			reduce(225), // &&, reduce: MlrvalOrFunction
			reduce(225), // ??, reduce: MlrvalOrFunction
			reduce(225), // ???, reduce: MlrvalOrFunction
			reduce(225), // =~, reduce: MlrvalOrFunction
			reduce(225), // !=~, reduce: MlrvalOrFunction
			reduce(225), // ==, reduce: MlrvalOrFunction
			reduce(225), // !=, reduce: MlrvalOrFunction
			reduce(225), // <=>, reduce: MlrvalOrFunction
			reduce(225), // >=, reduce: MlrvalOrFunction
			reduce(225), // <, reduce: MlrvalOrFunction
			reduce(225), // <=, reduce: MlrvalOrFunction
			reduce(225), // ^, reduce: MlrvalOrFunction
			reduce(225), // &, reduce: MlrvalOrFunction
			reduce(225), // <<, reduce: MlrvalOrFunction
			reduce(225), // >>>, reduce: MlrvalOrFunction
			reduce(225), // +, reduce: MlrvalOrFunction
			reduce(225), // -, reduce: MlrvalOrFunction
			reduce(225), // .+, reduce: MlrvalOrFunction
			reduce(225), // .-, reduce: MlrvalOrFunction
			reduce(225), // *, reduce: MlrvalOrFunction
			reduce(225), // /, reduce: MlrvalOrFunction
			reduce(225), // //, reduce: MlrvalOrFunction
			reduce(225), // %, reduce: MlrvalOrFunction
			reduce(225), // .*, reduce: MlrvalOrFunction
			reduce(225), // ./, reduce: MlrvalOrFunction
			reduce(225), // .//, reduce: MlrvalOrFunction
			reduce(225), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(225), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(225), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(225), // [[, reduce: MlrvalOrFunction
			reduce(225), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2201
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(226), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(226), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(226), // >, reduce: MlrvalOrFunction
			reduce(226), // >>, reduce: MlrvalOrFunction
			reduce(226), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(226), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(226), // ||, reduce: MlrvalOrFunction
			reduce(226), // ^^, reduce: MlrvalOrFunction
			reduce(226), // &&, reduce: MlrvalOrFunction
			reduce(226), // ??, reduce: MlrvalOrFunction
			reduce(226), // ???, reduce: MlrvalOrFunction
			reduce(226), // =~, reduce: MlrvalOrFunction
			reduce(226), // !=~, reduce: MlrvalOrFunction
			reduce(226), // ==, reduce: MlrvalOrFunction
			reduce(226), // !=, reduce: MlrvalOrFunction
			reduce(226), // <=>, reduce: MlrvalOrFunction
			reduce(226), // >=, reduce: MlrvalOrFunction
			reduce(226), // <, reduce: MlrvalOrFunction
			reduce(226), // <=, reduce: MlrvalOrFunction
			reduce(226), // ^, reduce: MlrvalOrFunction
			reduce(226), // &, reduce: MlrvalOrFunction
			reduce(226), // <<, reduce: MlrvalOrFunction
			reduce(226), // >>>, reduce: MlrvalOrFunction
			reduce(226), // +, reduce: MlrvalOrFunction
			reduce(226), // -, reduce: MlrvalOrFunction
			reduce(226), // .+, reduce: MlrvalOrFunction
			reduce(226), // .-, reduce: MlrvalOrFunction
			reduce(226), // *, reduce: MlrvalOrFunction
			reduce(226), // /, reduce: MlrvalOrFunction
			reduce(226), // //, reduce: MlrvalOrFunction
			reduce(226), // %, reduce: MlrvalOrFunction
			reduce(226), // .*, reduce: MlrvalOrFunction
			reduce(226), // ./, reduce: MlrvalOrFunction
			reduce(226), // .//, reduce: MlrvalOrFunction
			reduce(226), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(226), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(226), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(226), // [[, reduce: MlrvalOrFunction
			reduce(226), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2202
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(227), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(227), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(227), // >, reduce: MlrvalOrFunction
			reduce(227), // >>, reduce: MlrvalOrFunction
			reduce(227), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(227), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(227), // ||, reduce: MlrvalOrFunction
			reduce(227), // ^^, reduce: MlrvalOrFunction
			reduce(227), // &&, reduce: MlrvalOrFunction
			reduce(227), // ??, reduce: MlrvalOrFunction
			reduce(227), // ???, reduce: MlrvalOrFunction
			reduce(227), // =~, reduce: MlrvalOrFunction
			reduce(227), // !=~, reduce: MlrvalOrFunction
			reduce(227), // ==, reduce: MlrvalOrFunction
			reduce(227), // !=, reduce: MlrvalOrFunction
			reduce(227), // <=>, reduce: MlrvalOrFunction
			reduce(227), // >=, reduce: MlrvalOrFunction
			reduce(227), // <, reduce: MlrvalOrFunction
			reduce(227), // <=, reduce: MlrvalOrFunction
			reduce(227), // ^, reduce: MlrvalOrFunction
			reduce(227), // &, reduce: MlrvalOrFunction
			reduce(227), // <<, reduce: MlrvalOrFunction
			reduce(227), // >>>, reduce: MlrvalOrFunction
			reduce(227), // +, reduce: MlrvalOrFunction
			reduce(227), // -, reduce: MlrvalOrFunction
			reduce(227), // .+, reduce: MlrvalOrFunction
			reduce(227), // .-, reduce: MlrvalOrFunction
			reduce(227), // *, reduce: MlrvalOrFunction
			reduce(227), // /, reduce: MlrvalOrFunction
			reduce(227), // //, reduce: MlrvalOrFunction
			reduce(227), // %, reduce: MlrvalOrFunction
			reduce(227), // .*, reduce: MlrvalOrFunction
			reduce(227), // ./, reduce: MlrvalOrFunction
			reduce(227), // .//, reduce: MlrvalOrFunction
			reduce(227), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(227), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(227), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(227), // [[, reduce: MlrvalOrFunction
			reduce(227), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2203
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(752),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(756),  // (
			nil,         // )
			nil,         // emitp
			shift(768),  // field_name
			shift(769),  // $[
			shift(3108), // ]
			shift(771),  // braced_field_name
			shift(772),  // $[[
			shift(773),  // $[[[
			shift(774),  // full_srec
			shift(776),  // oosvar_name
			shift(777),  // @[
			shift(778),  // braced_oosvar_name
			shift(779),  // full_oosvar
			shift(780),  // all
			shift(781),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(796),  // +
			shift(798),  // -
			shift(799),  // .+
			shift(800),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(803),  // !
			shift(804),  // ~
			nil,         // **
			shift(809),  // string_literal
			shift(810),  // regex_case_insensitive
			shift(811),  // int_literal
			shift(812),  // float_literal
			shift(813),  // boolean_literal
			shift(814),  // null_literal
			shift(815),  // inf_literal
			shift(816),  // nan_literal
			shift(817),  // const_M_PI
			shift(818),  // const_M_E
			shift(819),  // panic
			shift(821),  // [
			shift(824),  // ctx_IPS
			shift(825),  // ctx_IFS
			shift(826),  // ctx_IRS
			shift(827),  // ctx_OPS
			shift(828),  // ctx_OFS
			shift(829),  // ctx_ORS
			shift(830),  // ctx_FLATSEP
			shift(831),  // ctx_NF
			shift(832),  // ctx_NR
			shift(833),  // ctx_FNR
			shift(834),  // ctx_FILENAME
			shift(835),  // ctx_FILENUM
			shift(837),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(844),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2204
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(240), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(240), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(240), // >, reduce: MlrvalOrFunction
			reduce(240), // >>, reduce: MlrvalOrFunction
			reduce(240), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(240), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(240), // ||, reduce: MlrvalOrFunction
			reduce(240), // ^^, reduce: MlrvalOrFunction
			reduce(240), // &&, reduce: MlrvalOrFunction
			reduce(240), // ??, reduce: MlrvalOrFunction
			reduce(240), // ???, reduce: MlrvalOrFunction
			reduce(240), // =~, reduce: MlrvalOrFunction
			reduce(240), // !=~, reduce: MlrvalOrFunction
			reduce(240), // ==, reduce: MlrvalOrFunction
			reduce(240), // !=, reduce: MlrvalOrFunction
			reduce(240), // <=>, reduce: MlrvalOrFunction
			reduce(240), // >=, reduce: MlrvalOrFunction
			reduce(240), // <, reduce: MlrvalOrFunction
			reduce(240), // <=, reduce: MlrvalOrFunction
			reduce(240), // ^, reduce: MlrvalOrFunction
			reduce(240), // &, reduce: MlrvalOrFunction
			reduce(240), // <<, reduce: MlrvalOrFunction
			reduce(240), // >>>, reduce: MlrvalOrFunction
			reduce(240), // +, reduce: MlrvalOrFunction
			reduce(240), // -, reduce: MlrvalOrFunction
			reduce(240), // .+, reduce: MlrvalOrFunction
			reduce(240), // .-, reduce: MlrvalOrFunction
			reduce(240), // *, reduce: MlrvalOrFunction
			reduce(240), // /, reduce: MlrvalOrFunction
			reduce(240), // //, reduce: MlrvalOrFunction
			reduce(240), // %, reduce: MlrvalOrFunction
			reduce(240), // .*, reduce: MlrvalOrFunction
			reduce(240), // ./, reduce: MlrvalOrFunction
			reduce(240), // .//, reduce: MlrvalOrFunction
			reduce(240), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(240), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(240), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(240), // [[, reduce: MlrvalOrFunction
			reduce(240), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2205
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(241), // ;, reduce: ContextVariable
			nil,         // {
			reduce(241), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(241), // >, reduce: ContextVariable
			reduce(241), // >>, reduce: ContextVariable
			reduce(241), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(241), // ?, reduce: ContextVariable
			nil,         // :
			reduce(241), // ||, reduce: ContextVariable
			reduce(241), // ^^, reduce: ContextVariable
			reduce(241), // &&, reduce: ContextVariable
			reduce(241), // ??, reduce: ContextVariable
			reduce(241), // ???, reduce: ContextVariable
			reduce(241), // =~, reduce: ContextVariable
			reduce(241), // !=~, reduce: ContextVariable
			reduce(241), // ==, reduce: ContextVariable
			reduce(241), // !=, reduce: ContextVariable
			reduce(241), // <=>, reduce: ContextVariable
			reduce(241), // >=, reduce: ContextVariable
			reduce(241), // <, reduce: ContextVariable
			reduce(241), // <=, reduce: ContextVariable
			reduce(241), // ^, reduce: ContextVariable
			reduce(241), // &, reduce: ContextVariable
			reduce(241), // <<, reduce: ContextVariable
			reduce(241), // >>>, reduce: ContextVariable
			reduce(241), // +, reduce: ContextVariable
			reduce(241), // -, reduce: ContextVariable
			reduce(241), // .+, reduce: ContextVariable
			reduce(241), // .-, reduce: ContextVariable
			reduce(241), // *, reduce: ContextVariable
			reduce(241), // /, reduce: ContextVariable
			reduce(241), // //, reduce: ContextVariable
			reduce(241), // %, reduce: ContextVariable
			reduce(241), // .*, reduce: ContextVariable
			reduce(241), // ./, reduce: ContextVariable
			reduce(241), // .//, reduce: ContextVariable
			reduce(241), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(241), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(241), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(241), // [[, reduce: ContextVariable
			reduce(241), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2206
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(242), // ;, reduce: ContextVariable
			nil,         // {
			reduce(242), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(242), // >, reduce: ContextVariable
			reduce(242), // >>, reduce: ContextVariable
			reduce(242), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(242), // ?, reduce: ContextVariable
			nil,         // :
			reduce(242), // ||, reduce: ContextVariable
			reduce(242), // ^^, reduce: ContextVariable
			reduce(242), // &&, reduce: ContextVariable
			reduce(242), // ??, reduce: ContextVariable
			reduce(242), // ???, reduce: ContextVariable
			reduce(242), // =~, reduce: ContextVariable
			reduce(242), // !=~, reduce: ContextVariable
			reduce(242), // ==, reduce: ContextVariable
			reduce(242), // !=, reduce: ContextVariable
			reduce(242), // <=>, reduce: ContextVariable
			reduce(242), // >=, reduce: ContextVariable
			reduce(242), // <, reduce: ContextVariable
			reduce(242), // <=, reduce: ContextVariable
			reduce(242), // ^, reduce: ContextVariable
			reduce(242), // &, reduce: ContextVariable
			reduce(242), // <<, reduce: ContextVariable
			reduce(242), // >>>, reduce: ContextVariable
			reduce(242), // +, reduce: ContextVariable
			reduce(242), // -, reduce: ContextVariable
			reduce(242), // .+, reduce: ContextVariable
			reduce(242), // .-, reduce: ContextVariable
			reduce(242), // *, reduce: ContextVariable
			reduce(242), // /, reduce: ContextVariable
			reduce(242), // //, reduce: ContextVariable
			reduce(242), // %, reduce: ContextVariable
			reduce(242), // .*, reduce: ContextVariable
			reduce(242), // ./, reduce: ContextVariable
			reduce(242), // .//, reduce: ContextVariable
			reduce(242), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(242), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(242), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(242), // [[, reduce: ContextVariable
			reduce(242), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2207
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(243), // ;, reduce: ContextVariable
			nil,         // {
			reduce(243), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(243), // >, reduce: ContextVariable
			reduce(243), // >>, reduce: ContextVariable
			reduce(243), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(243), // ?, reduce: ContextVariable
			nil,         // :
			reduce(243), // ||, reduce: ContextVariable
			reduce(243), // ^^, reduce: ContextVariable
			reduce(243), // &&, reduce: ContextVariable
			reduce(243), // ??, reduce: ContextVariable
			reduce(243), // ???, reduce: ContextVariable
			reduce(243), // =~, reduce: ContextVariable
			reduce(243), // !=~, reduce: ContextVariable
			reduce(243), // ==, reduce: ContextVariable
			reduce(243), // !=, reduce: ContextVariable
			reduce(243), // <=>, reduce: ContextVariable
			reduce(243), // >=, reduce: ContextVariable
			reduce(243), // <, reduce: ContextVariable
			reduce(243), // <=, reduce: ContextVariable
			reduce(243), // ^, reduce: ContextVariable
			reduce(243), // &, reduce: ContextVariable
			reduce(243), // <<, reduce: ContextVariable
			reduce(243), // >>>, reduce: ContextVariable
			reduce(243), // +, reduce: ContextVariable
			reduce(243), // -, reduce: ContextVariable
			reduce(243), // .+, reduce: ContextVariable
			reduce(243), // .-, reduce: ContextVariable
			reduce(243), // *, reduce: ContextVariable
			reduce(243), // /, reduce: ContextVariable
			reduce(243), // //, reduce: ContextVariable
			reduce(243), // %, reduce: ContextVariable
			reduce(243), // .*, reduce: ContextVariable
			reduce(243), // ./, reduce: ContextVariable
			reduce(243), // .//, reduce: ContextVariable
			reduce(243), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(243), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(243), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(243), // [[, reduce: ContextVariable
			reduce(243), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2208
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(244), // ;, reduce: ContextVariable
			nil,         // {
			reduce(244), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(244), // >, reduce: ContextVariable
			reduce(244), // >>, reduce: ContextVariable
			reduce(244), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(244), // ?, reduce: ContextVariable
			nil,         // :
			reduce(244), // ||, reduce: ContextVariable
			reduce(244), // ^^, reduce: ContextVariable
			reduce(244), // &&, reduce: ContextVariable
			reduce(244), // ??, reduce: ContextVariable
			reduce(244), // ???, reduce: ContextVariable
			reduce(244), // =~, reduce: ContextVariable
			reduce(244), // !=~, reduce: ContextVariable
			reduce(244), // ==, reduce: ContextVariable
			reduce(244), // !=, reduce: ContextVariable
			reduce(244), // <=>, reduce: ContextVariable
			reduce(244), // >=, reduce: ContextVariable
			reduce(244), // <, reduce: ContextVariable
			reduce(244), // <=, reduce: ContextVariable
			reduce(244), // ^, reduce: ContextVariable
			reduce(244), // &, reduce: ContextVariable
			reduce(244), // <<, reduce: ContextVariable
			reduce(244), // >>>, reduce: ContextVariable
			reduce(244), // +, reduce: ContextVariable
			reduce(244), // -, reduce: ContextVariable
			reduce(244), // .+, reduce: ContextVariable
			reduce(244), // .-, reduce: ContextVariable
			reduce(244), // *, reduce: ContextVariable
			reduce(244), // /, reduce: ContextVariable
			reduce(244), // //, reduce: ContextVariable
			reduce(244), // %, reduce: ContextVariable
			reduce(244), // .*, reduce: ContextVariable
			reduce(244), // ./, reduce: ContextVariable
			reduce(244), // .//, reduce: ContextVariable
			reduce(244), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(244), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(244), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(244), // [[, reduce: ContextVariable
			reduce(244), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2209
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(245), // ;, reduce: ContextVariable
			nil,         // {
			reduce(245), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(245), // >, reduce: ContextVariable
			reduce(245), // >>, reduce: ContextVariable
			reduce(245), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(245), // ?, reduce: ContextVariable
			nil,         // :
			reduce(245), // ||, reduce: ContextVariable
			reduce(245), // ^^, reduce: ContextVariable
			reduce(245), // &&, reduce: ContextVariable
			reduce(245), // ??, reduce: ContextVariable
			reduce(245), // ???, reduce: ContextVariable
			reduce(245), // =~, reduce: ContextVariable
			reduce(245), // !=~, reduce: ContextVariable
			reduce(245), // ==, reduce: ContextVariable
			reduce(245), // !=, reduce: ContextVariable
			reduce(245), // <=>, reduce: ContextVariable
			reduce(245), // >=, reduce: ContextVariable
			reduce(245), // <, reduce: ContextVariable
			reduce(245), // <=, reduce: ContextVariable
			reduce(245), // ^, reduce: ContextVariable
			reduce(245), // &, reduce: ContextVariable
			reduce(245), // <<, reduce: ContextVariable
			reduce(245), // >>>, reduce: ContextVariable
			reduce(245), // +, reduce: ContextVariable
			reduce(245), // -, reduce: ContextVariable
			reduce(245), // .+, reduce: ContextVariable
			reduce(245), // .-, reduce: ContextVariable
			reduce(245), // *, reduce: ContextVariable
			reduce(245), // /, reduce: ContextVariable
			reduce(245), // //, reduce: ContextVariable
			reduce(245), // %, reduce: ContextVariable
			reduce(245), // .*, reduce: ContextVariable
			reduce(245), // ./, reduce: ContextVariable
			reduce(245), // .//, reduce: ContextVariable
			reduce(245), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(245), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(245), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(245), // [[, reduce: ContextVariable
			reduce(245), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2210
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(246), // ;, reduce: ContextVariable
			nil,         // {
			reduce(246), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(246), // >, reduce: ContextVariable
			reduce(246), // >>, reduce: ContextVariable
			reduce(246), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(246), // ?, reduce: ContextVariable
			nil,         // :
			reduce(246), // ||, reduce: ContextVariable
			reduce(246), // ^^, reduce: ContextVariable
			reduce(246), // &&, reduce: ContextVariable
			reduce(246), // ??, reduce: ContextVariable
			reduce(246), // ???, reduce: ContextVariable
			reduce(246), // =~, reduce: ContextVariable
			reduce(246), // !=~, reduce: ContextVariable
			reduce(246), // ==, reduce: ContextVariable
			reduce(246), // !=, reduce: ContextVariable
			reduce(246), // <=>, reduce: ContextVariable
			reduce(246), // >=, reduce: ContextVariable
			reduce(246), // <, reduce: ContextVariable
			reduce(246), // <=, reduce: ContextVariable
			reduce(246), // ^, reduce: ContextVariable
			reduce(246), // &, reduce: ContextVariable
			reduce(246), // <<, reduce: ContextVariable
			reduce(246), // >>>, reduce: ContextVariable
			reduce(246), // +, reduce: ContextVariable
			reduce(246), // -, reduce: ContextVariable
			reduce(246), // .+, reduce: ContextVariable
			reduce(246), // .-, reduce: ContextVariable
			reduce(246), // *, reduce: ContextVariable
			reduce(246), // /, reduce: ContextVariable
			reduce(246), // //, reduce: ContextVariable
			reduce(246), // %, reduce: ContextVariable
			reduce(246), // .*, reduce: ContextVariable
			reduce(246), // ./, reduce: ContextVariable
			reduce(246), // .//, reduce: ContextVariable
			reduce(246), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(246), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(246), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(246), // [[, reduce: ContextVariable
			reduce(246), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2211
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(247), // ;, reduce: ContextVariable
			nil,         // {
			reduce(247), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(247), // >, reduce: ContextVariable
			reduce(247), // >>, reduce: ContextVariable
			reduce(247), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(247), // ?, reduce: ContextVariable
			nil,         // :
			reduce(247), // ||, reduce: ContextVariable
			reduce(247), // ^^, reduce: ContextVariable
			reduce(247), // &&, reduce: ContextVariable
			reduce(247), // ??, reduce: ContextVariable
			reduce(247), // ???, reduce: ContextVariable
			reduce(247), // =~, reduce: ContextVariable
			reduce(247), // !=~, reduce: ContextVariable
			reduce(247), // ==, reduce: ContextVariable
			reduce(247), // !=, reduce: ContextVariable
			reduce(247), // <=>, reduce: ContextVariable
			reduce(247), // >=, reduce: ContextVariable
			reduce(247), // <, reduce: ContextVariable
			reduce(247), // <=, reduce: ContextVariable
			reduce(247), // ^, reduce: ContextVariable
			reduce(247), // &, reduce: ContextVariable
			reduce(247), // <<, reduce: ContextVariable
			reduce(247), // >>>, reduce: ContextVariable
			reduce(247), // +, reduce: ContextVariable
			reduce(247), // -, reduce: ContextVariable
			reduce(247), // .+, reduce: ContextVariable
			reduce(247), // .-, reduce: ContextVariable
			reduce(247), // *, reduce: ContextVariable
			reduce(247), // /, reduce: ContextVariable
			reduce(247), // //, reduce: ContextVariable
			reduce(247), // %, reduce: ContextVariable
			reduce(247), // .*, reduce: ContextVariable
			reduce(247), // ./, reduce: ContextVariable
			reduce(247), // .//, reduce: ContextVariable
			reduce(247), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(247), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(247), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(247), // [[, reduce: ContextVariable
			reduce(247), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2212
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(248), // ;, reduce: ContextVariable
			nil,         // {
			reduce(248), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(248), // >, reduce: ContextVariable
			reduce(248), // >>, reduce: ContextVariable
			reduce(248), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(248), // ?, reduce: ContextVariable
			nil,         // :
			reduce(248), // ||, reduce: ContextVariable
			reduce(248), // ^^, reduce: ContextVariable
			reduce(248), // &&, reduce: ContextVariable
			reduce(248), // ??, reduce: ContextVariable
			reduce(248), // ???, reduce: ContextVariable
			reduce(248), // =~, reduce: ContextVariable
			reduce(248), // !=~, reduce: ContextVariable
			reduce(248), // ==, reduce: ContextVariable
			reduce(248), // !=, reduce: ContextVariable
			reduce(248), // <=>, reduce: ContextVariable
			reduce(248), // >=, reduce: ContextVariable
			reduce(248), // <, reduce: ContextVariable
			reduce(248), // <=, reduce: ContextVariable
			reduce(248), // ^, reduce: ContextVariable
			reduce(248), // &, reduce: ContextVariable
			reduce(248), // <<, reduce: ContextVariable
			reduce(248), // >>>, reduce: ContextVariable
			reduce(248), // +, reduce: ContextVariable
			reduce(248), // -, reduce: ContextVariable
			reduce(248), // .+, reduce: ContextVariable
			reduce(248), // .-, reduce: ContextVariable
			reduce(248), // *, reduce: ContextVariable
			reduce(248), // /, reduce: ContextVariable
			reduce(248), // //, reduce: ContextVariable
			reduce(248), // %, reduce: ContextVariable
			reduce(248), // .*, reduce: ContextVariable
			reduce(248), // ./, reduce: ContextVariable
			reduce(248), // .//, reduce: ContextVariable
			reduce(248), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(248), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(248), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(248), // [[, reduce: ContextVariable
			reduce(248), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2213
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(249), // ;, reduce: ContextVariable
			nil,         // {
			reduce(249), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(249), // >, reduce: ContextVariable
			reduce(249), // >>, reduce: ContextVariable
			reduce(249), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(249), // ?, reduce: ContextVariable
			nil,         // :
			reduce(249), // ||, reduce: ContextVariable
			reduce(249), // ^^, reduce: ContextVariable
			reduce(249), // &&, reduce: ContextVariable
			reduce(249), // ??, reduce: ContextVariable
			reduce(249), // ???, reduce: ContextVariable
			reduce(249), // =~, reduce: ContextVariable
			reduce(249), // !=~, reduce: ContextVariable
			reduce(249), // ==, reduce: ContextVariable
			reduce(249), // !=, reduce: ContextVariable
			reduce(249), // <=>, reduce: ContextVariable
			reduce(249), // >=, reduce: ContextVariable
			reduce(249), // <, reduce: ContextVariable
			reduce(249), // <=, reduce: ContextVariable
			reduce(249), // ^, reduce: ContextVariable
			reduce(249), // &, reduce: ContextVariable
			reduce(249), // <<, reduce: ContextVariable
			reduce(249), // >>>, reduce: ContextVariable
			reduce(249), // +, reduce: ContextVariable
			reduce(249), // -, reduce: ContextVariable
			reduce(249), // .+, reduce: ContextVariable
			reduce(249), // .-, reduce: ContextVariable
			reduce(249), // *, reduce: ContextVariable
			reduce(249), // /, reduce: ContextVariable
			reduce(249), // //, reduce: ContextVariable
			reduce(249), // %, reduce: ContextVariable
			reduce(249), // .*, reduce: ContextVariable
			reduce(249), // ./, reduce: ContextVariable
			reduce(249), // .//, reduce: ContextVariable
			reduce(249), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(249), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(249), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(249), // [[, reduce: ContextVariable
			reduce(249), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2214
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(250), // ;, reduce: ContextVariable
			nil,         // {
			reduce(250), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(250), // >, reduce: ContextVariable
			reduce(250), // >>, reduce: ContextVariable
			reduce(250), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(250), // ?, reduce: ContextVariable
			nil,         // :
			reduce(250), // ||, reduce: ContextVariable
			reduce(250), // ^^, reduce: ContextVariable
			reduce(250), // &&, reduce: ContextVariable
			reduce(250), // ??, reduce: ContextVariable
			reduce(250), // ???, reduce: ContextVariable
			reduce(250), // =~, reduce: ContextVariable
			reduce(250), // !=~, reduce: ContextVariable
			reduce(250), // ==, reduce: ContextVariable
			reduce(250), // !=, reduce: ContextVariable
			reduce(250), // <=>, reduce: ContextVariable
			reduce(250), // >=, reduce: ContextVariable
			reduce(250), // <, reduce: ContextVariable
			reduce(250), // <=, reduce: ContextVariable
			reduce(250), // ^, reduce: ContextVariable
			reduce(250), // &, reduce: ContextVariable
			reduce(250), // <<, reduce: ContextVariable
			reduce(250), // >>>, reduce: ContextVariable
			reduce(250), // +, reduce: ContextVariable
			reduce(250), // -, reduce: ContextVariable
			reduce(250), // .+, reduce: ContextVariable
			reduce(250), // .-, reduce: ContextVariable
			reduce(250), // *, reduce: ContextVariable
			reduce(250), // /, reduce: ContextVariable
			reduce(250), // //, reduce: ContextVariable
			reduce(250), // %, reduce: ContextVariable
			reduce(250), // .*, reduce: ContextVariable
			reduce(250), // ./, reduce: ContextVariable
			reduce(250), // .//, reduce: ContextVariable
			reduce(250), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(250), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(250), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(250), // [[, reduce: ContextVariable
			reduce(250), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2215
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(251), // ;, reduce: ContextVariable
			nil,         // {
			reduce(251), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(251), // >, reduce: ContextVariable
			reduce(251), // >>, reduce: ContextVariable
			reduce(251), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(251), // ?, reduce: ContextVariable
			nil,         // :
			reduce(251), // ||, reduce: ContextVariable
			reduce(251), // ^^, reduce: ContextVariable
			reduce(251), // &&, reduce: ContextVariable
			reduce(251), // ??, reduce: ContextVariable
			reduce(251), // ???, reduce: ContextVariable
			reduce(251), // =~, reduce: ContextVariable
			reduce(251), // !=~, reduce: ContextVariable
			reduce(251), // ==, reduce: ContextVariable
			reduce(251), // !=, reduce: ContextVariable
			reduce(251), // <=>, reduce: ContextVariable
			reduce(251), // >=, reduce: ContextVariable
			reduce(251), // <, reduce: ContextVariable
			reduce(251), // <=, reduce: ContextVariable
			reduce(251), // ^, reduce: ContextVariable
			reduce(251), // &, reduce: ContextVariable
			reduce(251), // <<, reduce: ContextVariable
			reduce(251), // >>>, reduce: ContextVariable
			reduce(251), // +, reduce: ContextVariable
			reduce(251), // -, reduce: ContextVariable
			reduce(251), // .+, reduce: ContextVariable
			reduce(251), // .-, reduce: ContextVariable
			reduce(251), // *, reduce: ContextVariable
			reduce(251), // /, reduce: ContextVariable
			reduce(251), // //, reduce: ContextVariable
			reduce(251), // %, reduce: ContextVariable
			reduce(251), // .*, reduce: ContextVariable
			reduce(251), // ./, reduce: ContextVariable
			reduce(251), // .//, reduce: ContextVariable
			reduce(251), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(251), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(251), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(251), // [[, reduce: ContextVariable
			reduce(251), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2216
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(252), // ;, reduce: ContextVariable
			nil,         // {
			reduce(252), // }, reduce: ContextVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(252), // >, reduce: ContextVariable
			reduce(252), // >>, reduce: ContextVariable
			reduce(252), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(252), // ?, reduce: ContextVariable
			nil,         // :
			reduce(252), // ||, reduce: ContextVariable
			reduce(252), // ^^, reduce: ContextVariable
			reduce(252), // &&, reduce: ContextVariable
			reduce(252), // ??, reduce: ContextVariable
			reduce(252), // ???, reduce: ContextVariable
			reduce(252), // =~, reduce: ContextVariable
			reduce(252), // !=~, reduce: ContextVariable
			reduce(252), // ==, reduce: ContextVariable
			reduce(252), // !=, reduce: ContextVariable
			reduce(252), // <=>, reduce: ContextVariable
			reduce(252), // >=, reduce: ContextVariable
			reduce(252), // <, reduce: ContextVariable
			reduce(252), // <=, reduce: ContextVariable
			reduce(252), // ^, reduce: ContextVariable
			reduce(252), // &, reduce: ContextVariable
			reduce(252), // <<, reduce: ContextVariable
			reduce(252), // >>>, reduce: ContextVariable
			reduce(252), // +, reduce: ContextVariable
			reduce(252), // -, reduce: ContextVariable
			reduce(252), // .+, reduce: ContextVariable
			reduce(252), // .-, reduce: ContextVariable
			reduce(252), // *, reduce: ContextVariable
			reduce(252), // /, reduce: ContextVariable
			reduce(252), // //, reduce: ContextVariable
			reduce(252), // %, reduce: ContextVariable
			reduce(252), // .*, reduce: ContextVariable
			reduce(252), // ./, reduce: ContextVariable
			reduce(252), // .//, reduce: ContextVariable
			reduce(252), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(252), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(252), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(252), // [[, reduce: ContextVariable
			reduce(252), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2217
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(253), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(253), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(253), // >, reduce: MlrvalOrFunction
			reduce(253), // >>, reduce: MlrvalOrFunction
			reduce(253), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(253), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(253), // ||, reduce: MlrvalOrFunction
			reduce(253), // ^^, reduce: MlrvalOrFunction
			reduce(253), // &&, reduce: MlrvalOrFunction
			reduce(253), // ??, reduce: MlrvalOrFunction
			reduce(253), // ???, reduce: MlrvalOrFunction
			reduce(253), // =~, reduce: MlrvalOrFunction
			reduce(253), // !=~, reduce: MlrvalOrFunction
			reduce(253), // ==, reduce: MlrvalOrFunction
			reduce(253), // !=, reduce: MlrvalOrFunction
			reduce(253), // <=>, reduce: MlrvalOrFunction
			reduce(253), // >=, reduce: MlrvalOrFunction
			reduce(253), // <, reduce: MlrvalOrFunction
			reduce(253), // <=, reduce: MlrvalOrFunction
			reduce(253), // ^, reduce: MlrvalOrFunction
			reduce(253), // &, reduce: MlrvalOrFunction
			reduce(253), // <<, reduce: MlrvalOrFunction
			reduce(253), // >>>, reduce: MlrvalOrFunction
			reduce(253), // +, reduce: MlrvalOrFunction
			reduce(253), // -, reduce: MlrvalOrFunction
			reduce(253), // .+, reduce: MlrvalOrFunction
			reduce(253), // .-, reduce: MlrvalOrFunction
			reduce(253), // *, reduce: MlrvalOrFunction
			reduce(253), // /, reduce: MlrvalOrFunction
			reduce(253), // //, reduce: MlrvalOrFunction
			reduce(253), // %, reduce: MlrvalOrFunction
			reduce(253), // .*, reduce: MlrvalOrFunction
			reduce(253), // ./, reduce: MlrvalOrFunction
			reduce(253), // .//, reduce: MlrvalOrFunction
			reduce(253), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(253), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(253), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(253), // [[, reduce: MlrvalOrFunction
			reduce(253), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2218
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			shift(3110), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(3111), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2219
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(256), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(256), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(256), // >, reduce: MlrvalOrFunction
			reduce(256), // >>, reduce: MlrvalOrFunction
			reduce(256), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(256), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(256), // ||, reduce: MlrvalOrFunction
			reduce(256), // ^^, reduce: MlrvalOrFunction
			reduce(256), // &&, reduce: MlrvalOrFunction
			reduce(256), // ??, reduce: MlrvalOrFunction
			reduce(256), // ???, reduce: MlrvalOrFunction
			reduce(256), // =~, reduce: MlrvalOrFunction
			reduce(256), // !=~, reduce: MlrvalOrFunction
			reduce(256), // ==, reduce: MlrvalOrFunction
			reduce(256), // !=, reduce: MlrvalOrFunction
			reduce(256), // <=>, reduce: MlrvalOrFunction
			reduce(256), // >=, reduce: MlrvalOrFunction
			reduce(256), // <, reduce: MlrvalOrFunction
			reduce(256), // <=, reduce: MlrvalOrFunction
			reduce(256), // ^, reduce: MlrvalOrFunction
			reduce(256), // &, reduce: MlrvalOrFunction
			reduce(256), // <<, reduce: MlrvalOrFunction
			reduce(256), // >>>, reduce: MlrvalOrFunction
			reduce(256), // +, reduce: MlrvalOrFunction
			reduce(256), // -, reduce: MlrvalOrFunction
			reduce(256), // .+, reduce: MlrvalOrFunction
			reduce(256), // .-, reduce: MlrvalOrFunction
			reduce(256), // *, reduce: MlrvalOrFunction
			reduce(256), // /, reduce: MlrvalOrFunction
			reduce(256), // //, reduce: MlrvalOrFunction
			reduce(256), // %, reduce: MlrvalOrFunction
			reduce(256), // .*, reduce: MlrvalOrFunction
			reduce(256), // ./, reduce: MlrvalOrFunction
			reduce(256), // .//, reduce: MlrvalOrFunction
			reduce(256), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(256), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(256), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(256), // [[, reduce: MlrvalOrFunction
			reduce(256), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2220
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(257), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(257), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(257), // >, reduce: MlrvalOrFunction
			reduce(257), // >>, reduce: MlrvalOrFunction
			reduce(257), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(257), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(257), // ||, reduce: MlrvalOrFunction
			reduce(257), // ^^, reduce: MlrvalOrFunction
			reduce(257), // &&, reduce: MlrvalOrFunction
			reduce(257), // ??, reduce: MlrvalOrFunction
			reduce(257), // ???, reduce: MlrvalOrFunction
			reduce(257), // =~, reduce: MlrvalOrFunction
			reduce(257), // !=~, reduce: MlrvalOrFunction
			reduce(257), // ==, reduce: MlrvalOrFunction
			reduce(257), // !=, reduce: MlrvalOrFunction
			reduce(257), // <=>, reduce: MlrvalOrFunction
			reduce(257), // >=, reduce: MlrvalOrFunction
			reduce(257), // <, reduce: MlrvalOrFunction
			reduce(257), // <=, reduce: MlrvalOrFunction
			reduce(257), // ^, reduce: MlrvalOrFunction
			reduce(257), // &, reduce: MlrvalOrFunction
			reduce(257), // <<, reduce: MlrvalOrFunction
			reduce(257), // >>>, reduce: MlrvalOrFunction
			reduce(257), // +, reduce: MlrvalOrFunction
			reduce(257), // -, reduce: MlrvalOrFunction
			reduce(257), // .+, reduce: MlrvalOrFunction
			reduce(257), // .-, reduce: MlrvalOrFunction
			reduce(257), // *, reduce: MlrvalOrFunction
			reduce(257), // /, reduce: MlrvalOrFunction
			reduce(257), // //, reduce: MlrvalOrFunction
			reduce(257), // %, reduce: MlrvalOrFunction
			reduce(257), // .*, reduce: MlrvalOrFunction
			reduce(257), // ./, reduce: MlrvalOrFunction
			reduce(257), // .//, reduce: MlrvalOrFunction
			reduce(257), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(257), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(257), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(257), // [[, reduce: MlrvalOrFunction
			reduce(257), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2221
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(258), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(258), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(258), // >, reduce: MlrvalOrFunction
			reduce(258), // >>, reduce: MlrvalOrFunction
			reduce(258), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(258), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(258), // ||, reduce: MlrvalOrFunction
			reduce(258), // ^^, reduce: MlrvalOrFunction
			reduce(258), // &&, reduce: MlrvalOrFunction
			reduce(258), // ??, reduce: MlrvalOrFunction
			reduce(258), // ???, reduce: MlrvalOrFunction
			reduce(258), // =~, reduce: MlrvalOrFunction
			reduce(258), // !=~, reduce: MlrvalOrFunction
			reduce(258), // ==, reduce: MlrvalOrFunction
			reduce(258), // !=, reduce: MlrvalOrFunction
			reduce(258), // <=>, reduce: MlrvalOrFunction
			reduce(258), // >=, reduce: MlrvalOrFunction
			reduce(258), // <, reduce: MlrvalOrFunction
			reduce(258), // <=, reduce: MlrvalOrFunction
			reduce(258), // ^, reduce: MlrvalOrFunction
			reduce(258), // &, reduce: MlrvalOrFunction
			reduce(258), // <<, reduce: MlrvalOrFunction
			reduce(258), // >>>, reduce: MlrvalOrFunction
			reduce(258), // +, reduce: MlrvalOrFunction
			reduce(258), // -, reduce: MlrvalOrFunction
			reduce(258), // .+, reduce: MlrvalOrFunction
			reduce(258), // .-, reduce: MlrvalOrFunction
			reduce(258), // *, reduce: MlrvalOrFunction
			reduce(258), // /, reduce: MlrvalOrFunction
			reduce(258), // //, reduce: MlrvalOrFunction
			reduce(258), // %, reduce: MlrvalOrFunction
			reduce(258), // .*, reduce: MlrvalOrFunction
			reduce(258), // ./, reduce: MlrvalOrFunction
			reduce(258), // .//, reduce: MlrvalOrFunction
			reduce(258), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(258), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(258), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(258), // [[, reduce: MlrvalOrFunction
			reduce(258), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2222
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(259), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(259), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(259), // >, reduce: MlrvalOrFunction
			reduce(259), // >>, reduce: MlrvalOrFunction
			reduce(259), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(259), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(259), // ||, reduce: MlrvalOrFunction
			reduce(259), // ^^, reduce: MlrvalOrFunction
			reduce(259), // &&, reduce: MlrvalOrFunction
			reduce(259), // ??, reduce: MlrvalOrFunction
			reduce(259), // ???, reduce: MlrvalOrFunction
			reduce(259), // =~, reduce: MlrvalOrFunction
			reduce(259), // !=~, reduce: MlrvalOrFunction
			reduce(259), // ==, reduce: MlrvalOrFunction
			reduce(259), // !=, reduce: MlrvalOrFunction
			reduce(259), // <=>, reduce: MlrvalOrFunction
			reduce(259), // >=, reduce: MlrvalOrFunction
			reduce(259), // <, reduce: MlrvalOrFunction
			reduce(259), // <=, reduce: MlrvalOrFunction
			reduce(259), // ^, reduce: MlrvalOrFunction
			reduce(259), // &, reduce: MlrvalOrFunction
			reduce(259), // <<, reduce: MlrvalOrFunction
			reduce(259), // >>>, reduce: MlrvalOrFunction
			reduce(259), // +, reduce: MlrvalOrFunction
			reduce(259), // -, reduce: MlrvalOrFunction
			reduce(259), // .+, reduce: MlrvalOrFunction
			reduce(259), // .-, reduce: MlrvalOrFunction
			reduce(259), // *, reduce: MlrvalOrFunction
			reduce(259), // /, reduce: MlrvalOrFunction
			reduce(259), // //, reduce: MlrvalOrFunction
			reduce(259), // %, reduce: MlrvalOrFunction
			reduce(259), // .*, reduce: MlrvalOrFunction
			reduce(259), // ./, reduce: MlrvalOrFunction
			reduce(259), // .//, reduce: MlrvalOrFunction
			reduce(259), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(259), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(259), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(259), // [[, reduce: MlrvalOrFunction
			reduce(259), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2223
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(267), // ;, reduce: MlrvalOrFunction
			nil,         // {
			reduce(267), // }, reduce: MlrvalOrFunction
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(267), // >, reduce: MlrvalOrFunction
			reduce(267), // >>, reduce: MlrvalOrFunction
			reduce(267), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(267), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(267), // ||, reduce: MlrvalOrFunction
			reduce(267), // ^^, reduce: MlrvalOrFunction
			reduce(267), // &&, reduce: MlrvalOrFunction
			reduce(267), // ??, reduce: MlrvalOrFunction
			reduce(267), // ???, reduce: MlrvalOrFunction
			reduce(267), // =~, reduce: MlrvalOrFunction
			reduce(267), // !=~, reduce: MlrvalOrFunction
			reduce(267), // ==, reduce: MlrvalOrFunction
			reduce(267), // !=, reduce: MlrvalOrFunction
			reduce(267), // <=>, reduce: MlrvalOrFunction
			reduce(267), // >=, reduce: MlrvalOrFunction
			reduce(267), // <, reduce: MlrvalOrFunction
			reduce(267), // <=, reduce: MlrvalOrFunction
			reduce(267), // ^, reduce: MlrvalOrFunction
			reduce(267), // &, reduce: MlrvalOrFunction
			reduce(267), // <<, reduce: MlrvalOrFunction
			reduce(267), // >>>, reduce: MlrvalOrFunction
			reduce(267), // +, reduce: MlrvalOrFunction
			reduce(267), // -, reduce: MlrvalOrFunction
			reduce(267), // .+, reduce: MlrvalOrFunction
			reduce(267), // .-, reduce: MlrvalOrFunction
			reduce(267), // *, reduce: MlrvalOrFunction
			reduce(267), // /, reduce: MlrvalOrFunction
			reduce(267), // //, reduce: MlrvalOrFunction
			reduce(267), // %, reduce: MlrvalOrFunction
			reduce(267), // .*, reduce: MlrvalOrFunction
			reduce(267), // ./, reduce: MlrvalOrFunction
			reduce(267), // .//, reduce: MlrvalOrFunction
			reduce(267), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(267), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(267), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(267), // [[, reduce: MlrvalOrFunction
			reduce(267), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2224
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3112), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2225
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3113), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2226
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(48), // ;, reduce: PrintStatement
			nil,        // {
			reduce(48), // }, reduce: PrintStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2227
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(47),  // ;, reduce: PrintStatement
			nil,         // {
			reduce(47),  // }, reduce: PrintStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3114), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2228
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			shift(3117), // stdout
			shift(3118), // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2229
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			shift(3117), // stdout
			shift(3118), // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2230
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			shift(3117), // stdout
			shift(3118), // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2231
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(52), // ;, reduce: PrintnStatement
			nil,        // {
			reduce(52), // }, reduce: PrintnStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2232
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(51),  // ;, reduce: PrintnStatement
			nil,         // {
			reduce(51),  // }, reduce: PrintnStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3121), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2233
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(55), // ;, reduce: EprintStatement
			nil,        // {
			reduce(55), // }, reduce: EprintStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2234
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(57), // ;, reduce: EprintnStatement
			nil,        // {
			reduce(57), // }, reduce: EprintnStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2235
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(60), // ;, reduce: DumpStatement
			nil,        // {
			reduce(60), // }, reduce: DumpStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2236
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(59),  // ;, reduce: DumpStatement
			nil,         // {
			reduce(59),  // }, reduce: DumpStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3122), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2237
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(63), // ;, reduce: EdumpStatement
			nil,        // {
			reduce(63), // }, reduce: EdumpStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2238
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3123), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2239
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			shift(3124), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2240
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(87), // ;, reduce: Emittable
			nil,        // {
			reduce(87), // }, reduce: Emittable
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(87), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2241
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3126), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2242
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(94), // ;, reduce: Emittable
			nil,        // {
			reduce(94), // }, reduce: Emittable
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(94), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2243
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(65), // ;, reduce: EmitFStatement
			nil,        // {
			reduce(65), // }, reduce: EmitFStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2244
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(84),  // ;, reduce: EmittableList
			nil,         // {
			reduce(84),  // }, reduce: EmittableList
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3127), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2245
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(88), // ;, reduce: Emittable
			nil,        // {
			reduce(88), // }, reduce: Emittable
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(88), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2246
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(89), // ;, reduce: Emittable
			nil,        // {
			reduce(89), // }, reduce: Emittable
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(89), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2247
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(90), // ;, reduce: Emittable
			nil,        // {
			reduce(90), // }, reduce: Emittable
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(90), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2248
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(91), // ;, reduce: Emittable
			nil,        // {
			reduce(91), // }, reduce: Emittable
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(91), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2249
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(92), // ;, reduce: Emittable
			nil,        // {
			reduce(92), // }, reduce: Emittable
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(92), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2250
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(93), // ;, reduce: Emittable
			nil,        // {
			reduce(93), // }, reduce: Emittable
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(93), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2251
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(95), // ;, reduce: Emittable
			nil,        // {
			reduce(95), // }, reduce: Emittable
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(95), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2252
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(96), // ;, reduce: Emittable
			nil,        // {
			reduce(96), // }, reduce: Emittable
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(96), // ,, reduce: Emittable
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2253
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(104), // ;, reduce: DirectFieldValue
			nil,         // {
			reduce(104), // }, reduce: DirectFieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(104), // ,, reduce: DirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2254
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2255
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(106), // ;, reduce: BracedFieldValue
			nil,         // {
			reduce(106), // }, reduce: BracedFieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(106), // ,, reduce: BracedFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2256
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(109), // ;, reduce: FullSrec
			nil,         // {
			reduce(109), // }, reduce: FullSrec
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(109), // ,, reduce: FullSrec
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2257
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(113), // ;, reduce: DirectOosvarValue
			nil,         // {
			reduce(113), // }, reduce: DirectOosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(113), // ,, reduce: DirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2258
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2259
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(115), // ;, reduce: BracedOosvarValue
			nil,         // {
			reduce(115), // }, reduce: BracedOosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(115), // ,, reduce: BracedOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2260
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(116), // ;, reduce: FullOosvar
			nil,         // {
			reduce(116), // }, reduce: FullOosvar
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(116), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2261
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(117), // ;, reduce: FullOosvar
			nil,         // {
			reduce(117), // }, reduce: FullOosvar
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(117), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2262
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(118), // ;, reduce: LocalVariable
			nil,         // {
			reduce(118), // }, reduce: LocalVariable
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(118), // ,, reduce: LocalVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2263
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(67), // ;, reduce: Emit1Statement
			nil,        // {
			reduce(67), // }, reduce: Emit1Statement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2264
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3130), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2265
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(68),  // ;, reduce: EmitStatement
			nil,         // {
			reduce(68),  // }, reduce: EmitStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3131), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2266
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1330), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			shift(1343), // field_name
			shift(1344), // $[
			nil,         // ]
			shift(1345), // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(1346), // full_srec
			shift(1347), // oosvar_name
			shift(1348), // @[
			shift(1349), // braced_oosvar_name
			shift(1350), // full_oosvar
			shift(1351), // all
			shift(1352), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2267
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(86), // ;, reduce: EmittableAsList
			nil,        // {
			reduce(86), // }, reduce: EmittableAsList
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(86), // ,, reduce: EmittableAsList
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2268
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3133), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2269
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3134), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2270
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(76),  // ;, reduce: EmitPStatement
			nil,         // {
			reduce(76),  // }, reduce: EmitPStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3135), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2271
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1330), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			shift(1343), // field_name
			shift(1344), // $[
			nil,         // ]
			shift(1345), // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(1346), // full_srec
			shift(1347), // oosvar_name
			shift(1348), // @[
			shift(1349), // braced_oosvar_name
			shift(1350), // full_oosvar
			shift(1351), // all
			shift(1352), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2272
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3137), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2273
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3138), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2274
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3139), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2275
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3140), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2276
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2277
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2278
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2279
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2280
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2281
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2282
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2283
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2284
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2285
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2286
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2287
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2288
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2289
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2290
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2291
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2292
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2293
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2294
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2295
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2296
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2297
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2298
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2299
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2300
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2301
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(197), // ;, reduce: UnaryOpTerm
			reduce(197), // {, reduce: UnaryOpTerm
			reduce(197), // }, reduce: UnaryOpTerm
			reduce(197), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(197), // >, reduce: UnaryOpTerm
			reduce(197), // >>, reduce: UnaryOpTerm
			reduce(197), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(197), // ||=, reduce: UnaryOpTerm
			reduce(197), // ^^=, reduce: UnaryOpTerm
			reduce(197), // &&=, reduce: UnaryOpTerm
			reduce(197), // ??=, reduce: UnaryOpTerm
			reduce(197), // ???=, reduce: UnaryOpTerm
			reduce(197), // |=, reduce: UnaryOpTerm
			reduce(197), // &=, reduce: UnaryOpTerm
			reduce(197), // ^=, reduce: UnaryOpTerm
			reduce(197), // <<=, reduce: UnaryOpTerm
			reduce(197), // >>=, reduce: UnaryOpTerm
			reduce(197), // >>>=, reduce: UnaryOpTerm
			reduce(197), // +=, reduce: UnaryOpTerm
			reduce(197), // .=, reduce: UnaryOpTerm
			reduce(197), // -=, reduce: UnaryOpTerm
			reduce(197), // *=, reduce: UnaryOpTerm
			reduce(197), // /=, reduce: UnaryOpTerm
			reduce(197), // //=, reduce: UnaryOpTerm
			reduce(197), // %=, reduce: UnaryOpTerm
			reduce(197), // **=, reduce: UnaryOpTerm
			reduce(197), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(197), // ||, reduce: UnaryOpTerm
			reduce(197), // ^^, reduce: UnaryOpTerm
			reduce(197), // &&, reduce: UnaryOpTerm
			reduce(197), // ??, reduce: UnaryOpTerm
			reduce(197), // ???, reduce: UnaryOpTerm
			reduce(197), // =~, reduce: UnaryOpTerm
			reduce(197), // !=~, reduce: UnaryOpTerm
			reduce(197), // ==, reduce: UnaryOpTerm
			reduce(197), // !=, reduce: UnaryOpTerm
			reduce(197), // <=>, reduce: UnaryOpTerm
			reduce(197), // >=, reduce: UnaryOpTerm
			reduce(197), // <, reduce: UnaryOpTerm
			reduce(197), // <=, reduce: UnaryOpTerm
			reduce(197), // ^, reduce: UnaryOpTerm
			reduce(197), // &, reduce: UnaryOpTerm
			reduce(197), // <<, reduce: UnaryOpTerm
			reduce(197), // >>>, reduce: UnaryOpTerm
			reduce(197), // +, reduce: UnaryOpTerm
			reduce(197), // -, reduce: UnaryOpTerm
			reduce(197), // .+, reduce: UnaryOpTerm
			reduce(197), // .-, reduce: UnaryOpTerm
			reduce(197), // *, reduce: UnaryOpTerm
			reduce(197), // /, reduce: UnaryOpTerm
			reduce(197), // //, reduce: UnaryOpTerm
			reduce(197), // %, reduce: UnaryOpTerm
			reduce(197), // .*, reduce: UnaryOpTerm
			reduce(197), // ./, reduce: UnaryOpTerm
			reduce(197), // .//, reduce: UnaryOpTerm
			reduce(197), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2302
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2336), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2303
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2304
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2305
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2306
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2307
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2308
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2309
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2310
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(198), // ;, reduce: UnaryOpTerm
			reduce(198), // {, reduce: UnaryOpTerm
			reduce(198), // }, reduce: UnaryOpTerm
			reduce(198), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(198), // >, reduce: UnaryOpTerm
			reduce(198), // >>, reduce: UnaryOpTerm
			reduce(198), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(198), // ||=, reduce: UnaryOpTerm
			reduce(198), // ^^=, reduce: UnaryOpTerm
			reduce(198), // &&=, reduce: UnaryOpTerm
			reduce(198), // ??=, reduce: UnaryOpTerm
			reduce(198), // ???=, reduce: UnaryOpTerm
			reduce(198), // |=, reduce: UnaryOpTerm
			reduce(198), // &=, reduce: UnaryOpTerm
			reduce(198), // ^=, reduce: UnaryOpTerm
			reduce(198), // <<=, reduce: UnaryOpTerm
			reduce(198), // >>=, reduce: UnaryOpTerm
			reduce(198), // >>>=, reduce: UnaryOpTerm
			reduce(198), // +=, reduce: UnaryOpTerm
			reduce(198), // .=, reduce: UnaryOpTerm
			reduce(198), // -=, reduce: UnaryOpTerm
			reduce(198), // *=, reduce: UnaryOpTerm
			reduce(198), // /=, reduce: UnaryOpTerm
			reduce(198), // //=, reduce: UnaryOpTerm
			reduce(198), // %=, reduce: UnaryOpTerm
			reduce(198), // **=, reduce: UnaryOpTerm
			reduce(198), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(198), // ||, reduce: UnaryOpTerm
			reduce(198), // ^^, reduce: UnaryOpTerm
			reduce(198), // &&, reduce: UnaryOpTerm
			reduce(198), // ??, reduce: UnaryOpTerm
			reduce(198), // ???, reduce: UnaryOpTerm
			reduce(198), // =~, reduce: UnaryOpTerm
			reduce(198), // !=~, reduce: UnaryOpTerm
			reduce(198), // ==, reduce: UnaryOpTerm
			reduce(198), // !=, reduce: UnaryOpTerm
			reduce(198), // <=>, reduce: UnaryOpTerm
			reduce(198), // >=, reduce: UnaryOpTerm
			reduce(198), // <, reduce: UnaryOpTerm
			reduce(198), // <=, reduce: UnaryOpTerm
			reduce(198), // ^, reduce: UnaryOpTerm
			reduce(198), // &, reduce: UnaryOpTerm
			reduce(198), // <<, reduce: UnaryOpTerm
			reduce(198), // >>>, reduce: UnaryOpTerm
			reduce(198), // +, reduce: UnaryOpTerm
			reduce(198), // -, reduce: UnaryOpTerm
			reduce(198), // .+, reduce: UnaryOpTerm
			reduce(198), // .-, reduce: UnaryOpTerm
			reduce(198), // *, reduce: UnaryOpTerm
			reduce(198), // /, reduce: UnaryOpTerm
			reduce(198), // //, reduce: UnaryOpTerm
			reduce(198), // %, reduce: UnaryOpTerm
			reduce(198), // .*, reduce: UnaryOpTerm
			reduce(198), // ./, reduce: UnaryOpTerm
			reduce(198), // .//, reduce: UnaryOpTerm
			reduce(198), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2311
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(199), // ;, reduce: UnaryOpTerm
			reduce(199), // {, reduce: UnaryOpTerm
			reduce(199), // }, reduce: UnaryOpTerm
			reduce(199), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(199), // >, reduce: UnaryOpTerm
			reduce(199), // >>, reduce: UnaryOpTerm
			reduce(199), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(199), // ||=, reduce: UnaryOpTerm
			reduce(199), // ^^=, reduce: UnaryOpTerm
			reduce(199), // &&=, reduce: UnaryOpTerm
			reduce(199), // ??=, reduce: UnaryOpTerm
			reduce(199), // ???=, reduce: UnaryOpTerm
			reduce(199), // |=, reduce: UnaryOpTerm
			reduce(199), // &=, reduce: UnaryOpTerm
			reduce(199), // ^=, reduce: UnaryOpTerm
			reduce(199), // <<=, reduce: UnaryOpTerm
			reduce(199), // >>=, reduce: UnaryOpTerm
			reduce(199), // >>>=, reduce: UnaryOpTerm
			reduce(199), // +=, reduce: UnaryOpTerm
			reduce(199), // .=, reduce: UnaryOpTerm
			reduce(199), // -=, reduce: UnaryOpTerm
			reduce(199), // *=, reduce: UnaryOpTerm
			reduce(199), // /=, reduce: UnaryOpTerm
			reduce(199), // //=, reduce: UnaryOpTerm
			reduce(199), // %=, reduce: UnaryOpTerm
			reduce(199), // **=, reduce: UnaryOpTerm
			reduce(199), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(199), // ||, reduce: UnaryOpTerm
			reduce(199), // ^^, reduce: UnaryOpTerm
			reduce(199), // &&, reduce: UnaryOpTerm
			reduce(199), // ??, reduce: UnaryOpTerm
			reduce(199), // ???, reduce: UnaryOpTerm
			reduce(199), // =~, reduce: UnaryOpTerm
			reduce(199), // !=~, reduce: UnaryOpTerm
			reduce(199), // ==, reduce: UnaryOpTerm
			reduce(199), // !=, reduce: UnaryOpTerm
			reduce(199), // <=>, reduce: UnaryOpTerm
			reduce(199), // >=, reduce: UnaryOpTerm
			reduce(199), // <, reduce: UnaryOpTerm
			reduce(199), // <=, reduce: UnaryOpTerm
			reduce(199), // ^, reduce: UnaryOpTerm
			reduce(199), // &, reduce: UnaryOpTerm
			reduce(199), // <<, reduce: UnaryOpTerm
			reduce(199), // >>>, reduce: UnaryOpTerm
			reduce(199), // +, reduce: UnaryOpTerm
			reduce(199), // -, reduce: UnaryOpTerm
			reduce(199), // .+, reduce: UnaryOpTerm
			reduce(199), // .-, reduce: UnaryOpTerm
			reduce(199), // *, reduce: UnaryOpTerm
			reduce(199), // /, reduce: UnaryOpTerm
			reduce(199), // //, reduce: UnaryOpTerm
			reduce(199), // %, reduce: UnaryOpTerm
			reduce(199), // .*, reduce: UnaryOpTerm
			reduce(199), // ./, reduce: UnaryOpTerm
			reduce(199), // .//, reduce: UnaryOpTerm
			reduce(199), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2312
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(200), // ;, reduce: UnaryOpTerm
			reduce(200), // {, reduce: UnaryOpTerm
			reduce(200), // }, reduce: UnaryOpTerm
			reduce(200), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(200), // >, reduce: UnaryOpTerm
			reduce(200), // >>, reduce: UnaryOpTerm
			reduce(200), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(200), // ||=, reduce: UnaryOpTerm
			reduce(200), // ^^=, reduce: UnaryOpTerm
			reduce(200), // &&=, reduce: UnaryOpTerm
			reduce(200), // ??=, reduce: UnaryOpTerm
			reduce(200), // ???=, reduce: UnaryOpTerm
			reduce(200), // |=, reduce: UnaryOpTerm
			reduce(200), // &=, reduce: UnaryOpTerm
			reduce(200), // ^=, reduce: UnaryOpTerm
			reduce(200), // <<=, reduce: UnaryOpTerm
			reduce(200), // >>=, reduce: UnaryOpTerm
			reduce(200), // >>>=, reduce: UnaryOpTerm
			reduce(200), // +=, reduce: UnaryOpTerm
			reduce(200), // .=, reduce: UnaryOpTerm
			reduce(200), // -=, reduce: UnaryOpTerm
			reduce(200), // *=, reduce: UnaryOpTerm
			reduce(200), // /=, reduce: UnaryOpTerm
			reduce(200), // //=, reduce: UnaryOpTerm
			reduce(200), // %=, reduce: UnaryOpTerm
			reduce(200), // **=, reduce: UnaryOpTerm
			reduce(200), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(200), // ||, reduce: UnaryOpTerm
			reduce(200), // ^^, reduce: UnaryOpTerm
			reduce(200), // &&, reduce: UnaryOpTerm
			reduce(200), // ??, reduce: UnaryOpTerm
			reduce(200), // ???, reduce: UnaryOpTerm
			reduce(200), // =~, reduce: UnaryOpTerm
			reduce(200), // !=~, reduce: UnaryOpTerm
			reduce(200), // ==, reduce: UnaryOpTerm
			reduce(200), // !=, reduce: UnaryOpTerm
			reduce(200), // <=>, reduce: UnaryOpTerm
			reduce(200), // >=, reduce: UnaryOpTerm
			reduce(200), // <, reduce: UnaryOpTerm
			reduce(200), // <=, reduce: UnaryOpTerm
			reduce(200), // ^, reduce: UnaryOpTerm
			reduce(200), // &, reduce: UnaryOpTerm
			reduce(200), // <<, reduce: UnaryOpTerm
			reduce(200), // >>>, reduce: UnaryOpTerm
			reduce(200), // +, reduce: UnaryOpTerm
			reduce(200), // -, reduce: UnaryOpTerm
			reduce(200), // .+, reduce: UnaryOpTerm
			reduce(200), // .-, reduce: UnaryOpTerm
			reduce(200), // *, reduce: UnaryOpTerm
			reduce(200), // /, reduce: UnaryOpTerm
			reduce(200), // //, reduce: UnaryOpTerm
			reduce(200), // %, reduce: UnaryOpTerm
			reduce(200), // .*, reduce: UnaryOpTerm
			reduce(200), // ./, reduce: UnaryOpTerm
			reduce(200), // .//, reduce: UnaryOpTerm
			reduce(200), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2313
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2314
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(201), // ;, reduce: UnaryOpTerm
			reduce(201), // {, reduce: UnaryOpTerm
			reduce(201), // }, reduce: UnaryOpTerm
			reduce(201), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(201), // >, reduce: UnaryOpTerm
			reduce(201), // >>, reduce: UnaryOpTerm
			reduce(201), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(201), // ||=, reduce: UnaryOpTerm
			reduce(201), // ^^=, reduce: UnaryOpTerm
			reduce(201), // &&=, reduce: UnaryOpTerm
			reduce(201), // ??=, reduce: UnaryOpTerm
			reduce(201), // ???=, reduce: UnaryOpTerm
			reduce(201), // |=, reduce: UnaryOpTerm
			reduce(201), // &=, reduce: UnaryOpTerm
			reduce(201), // ^=, reduce: UnaryOpTerm
			reduce(201), // <<=, reduce: UnaryOpTerm
			reduce(201), // >>=, reduce: UnaryOpTerm
			reduce(201), // >>>=, reduce: UnaryOpTerm
			reduce(201), // +=, reduce: UnaryOpTerm
			reduce(201), // .=, reduce: UnaryOpTerm
			reduce(201), // -=, reduce: UnaryOpTerm
			reduce(201), // *=, reduce: UnaryOpTerm
			reduce(201), // /=, reduce: UnaryOpTerm
			reduce(201), // //=, reduce: UnaryOpTerm
			reduce(201), // %=, reduce: UnaryOpTerm
			reduce(201), // **=, reduce: UnaryOpTerm
			reduce(201), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(201), // ||, reduce: UnaryOpTerm
			reduce(201), // ^^, reduce: UnaryOpTerm
			reduce(201), // &&, reduce: UnaryOpTerm
			reduce(201), // ??, reduce: UnaryOpTerm
			reduce(201), // ???, reduce: UnaryOpTerm
			reduce(201), // =~, reduce: UnaryOpTerm
			reduce(201), // !=~, reduce: UnaryOpTerm
			reduce(201), // ==, reduce: UnaryOpTerm
			reduce(201), // !=, reduce: UnaryOpTerm
			reduce(201), // <=>, reduce: UnaryOpTerm
			reduce(201), // >=, reduce: UnaryOpTerm
			reduce(201), // <, reduce: UnaryOpTerm
			reduce(201), // <=, reduce: UnaryOpTerm
			reduce(201), // ^, reduce: UnaryOpTerm
			reduce(201), // &, reduce: UnaryOpTerm
			reduce(201), // <<, reduce: UnaryOpTerm
			reduce(201), // >>>, reduce: UnaryOpTerm
			reduce(201), // +, reduce: UnaryOpTerm
			reduce(201), // -, reduce: UnaryOpTerm
			reduce(201), // .+, reduce: UnaryOpTerm
			reduce(201), // .-, reduce: UnaryOpTerm
			reduce(201), // *, reduce: UnaryOpTerm
			reduce(201), // /, reduce: UnaryOpTerm
			reduce(201), // //, reduce: UnaryOpTerm
			reduce(201), // %, reduce: UnaryOpTerm
			reduce(201), // .*, reduce: UnaryOpTerm
			reduce(201), // ./, reduce: UnaryOpTerm
			reduce(201), // .//, reduce: UnaryOpTerm
			reduce(201), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2315
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(202), // ;, reduce: UnaryOpTerm
			reduce(202), // {, reduce: UnaryOpTerm
			reduce(202), // }, reduce: UnaryOpTerm
			reduce(202), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(202), // >, reduce: UnaryOpTerm
			reduce(202), // >>, reduce: UnaryOpTerm
			reduce(202), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(202), // ||=, reduce: UnaryOpTerm
			reduce(202), // ^^=, reduce: UnaryOpTerm
			reduce(202), // &&=, reduce: UnaryOpTerm
			reduce(202), // ??=, reduce: UnaryOpTerm
			reduce(202), // ???=, reduce: UnaryOpTerm
			reduce(202), // |=, reduce: UnaryOpTerm
			reduce(202), // &=, reduce: UnaryOpTerm
			reduce(202), // ^=, reduce: UnaryOpTerm
			reduce(202), // <<=, reduce: UnaryOpTerm
			reduce(202), // >>=, reduce: UnaryOpTerm
			reduce(202), // >>>=, reduce: UnaryOpTerm
			reduce(202), // +=, reduce: UnaryOpTerm
			reduce(202), // .=, reduce: UnaryOpTerm
			reduce(202), // -=, reduce: UnaryOpTerm
			reduce(202), // *=, reduce: UnaryOpTerm
			reduce(202), // /=, reduce: UnaryOpTerm
			reduce(202), // //=, reduce: UnaryOpTerm
			reduce(202), // %=, reduce: UnaryOpTerm
			reduce(202), // **=, reduce: UnaryOpTerm
			reduce(202), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(202), // ||, reduce: UnaryOpTerm
			reduce(202), // ^^, reduce: UnaryOpTerm
			reduce(202), // &&, reduce: UnaryOpTerm
			reduce(202), // ??, reduce: UnaryOpTerm
			reduce(202), // ???, reduce: UnaryOpTerm
			reduce(202), // =~, reduce: UnaryOpTerm
			reduce(202), // !=~, reduce: UnaryOpTerm
			reduce(202), // ==, reduce: UnaryOpTerm
			reduce(202), // !=, reduce: UnaryOpTerm
			reduce(202), // <=>, reduce: UnaryOpTerm
			reduce(202), // >=, reduce: UnaryOpTerm
			reduce(202), // <, reduce: UnaryOpTerm
			reduce(202), // <=, reduce: UnaryOpTerm
			reduce(202), // ^, reduce: UnaryOpTerm
			reduce(202), // &, reduce: UnaryOpTerm
			reduce(202), // <<, reduce: UnaryOpTerm
			reduce(202), // >>>, reduce: UnaryOpTerm
			reduce(202), // +, reduce: UnaryOpTerm
			reduce(202), // -, reduce: UnaryOpTerm
			reduce(202), // .+, reduce: UnaryOpTerm
			reduce(202), // .-, reduce: UnaryOpTerm
			reduce(202), // *, reduce: UnaryOpTerm
			reduce(202), // /, reduce: UnaryOpTerm
			reduce(202), // //, reduce: UnaryOpTerm
			reduce(202), // %, reduce: UnaryOpTerm
			reduce(202), // .*, reduce: UnaryOpTerm
			reduce(202), // ./, reduce: UnaryOpTerm
			reduce(202), // .//, reduce: UnaryOpTerm
			reduce(202), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2316
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3174), // +
			shift(3175), // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2317
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3178), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2318
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2319
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2320
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(228), // ;, reduce: ArrayLiteral
			reduce(228), // {, reduce: ArrayLiteral
			reduce(228), // }, reduce: ArrayLiteral
			reduce(228), // =, reduce: ArrayLiteral
			nil,         // unset
			nil,         // filter
			reduce(228), // >, reduce: ArrayLiteral
			reduce(228), // >>, reduce: ArrayLiteral
			reduce(228), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(228), // ||=, reduce: ArrayLiteral
			reduce(228), // ^^=, reduce: ArrayLiteral
			reduce(228), // &&=, reduce: ArrayLiteral
			reduce(228), // ??=, reduce: ArrayLiteral
			reduce(228), // ???=, reduce: ArrayLiteral
			reduce(228), // |=, reduce: ArrayLiteral
			reduce(228), // &=, reduce: ArrayLiteral
			reduce(228), // ^=, reduce: ArrayLiteral
			reduce(228), // <<=, reduce: ArrayLiteral
			reduce(228), // >>=, reduce: ArrayLiteral
			reduce(228), // >>>=, reduce: ArrayLiteral
			reduce(228), // +=, reduce: ArrayLiteral
			reduce(228), // .=, reduce: ArrayLiteral
			reduce(228), // -=, reduce: ArrayLiteral
			reduce(228), // *=, reduce: ArrayLiteral
			reduce(228), // /=, reduce: ArrayLiteral
			reduce(228), // //=, reduce: ArrayLiteral
			reduce(228), // %=, reduce: ArrayLiteral
			reduce(228), // **=, reduce: ArrayLiteral
			reduce(228), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(228), // ||, reduce: ArrayLiteral
			reduce(228), // ^^, reduce: ArrayLiteral
			reduce(228), // &&, reduce: ArrayLiteral
			reduce(228), // ??, reduce: ArrayLiteral
			reduce(228), // ???, reduce: ArrayLiteral
			reduce(228), // =~, reduce: ArrayLiteral
			reduce(228), // !=~, reduce: ArrayLiteral
			reduce(228), // ==, reduce: ArrayLiteral
			reduce(228), // !=, reduce: ArrayLiteral
			reduce(228), // <=>, reduce: ArrayLiteral
			reduce(228), // >=, reduce: ArrayLiteral
			reduce(228), // <, reduce: ArrayLiteral
			reduce(228), // <=, reduce: ArrayLiteral
			reduce(228), // ^, reduce: ArrayLiteral
			reduce(228), // &, reduce: ArrayLiteral
			reduce(228), // <<, reduce: ArrayLiteral
			reduce(228), // >>>, reduce: ArrayLiteral
			reduce(228), // +, reduce: ArrayLiteral
			reduce(228), // -, reduce: ArrayLiteral
			reduce(228), // .+, reduce: ArrayLiteral
			reduce(228), // .-, reduce: ArrayLiteral
			reduce(228), // *, reduce: ArrayLiteral
			reduce(228), // /, reduce: ArrayLiteral
			reduce(228), // //, reduce: ArrayLiteral
			reduce(228), // %, reduce: ArrayLiteral
			reduce(228), // .*, reduce: ArrayLiteral
			reduce(228), // ./, reduce: ArrayLiteral
			reduce(228), // .//, reduce: ArrayLiteral
			reduce(228), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(228), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(228), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(228), // [[, reduce: ArrayLiteral
			reduce(228), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2321
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3181), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2322
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(3182), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2323
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2324
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(3185), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2325
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3186), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2326
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(287), // ;, reduce: BeginBlock
			reduce(287), // {, reduce: BeginBlock
			reduce(287), // }, reduce: BeginBlock
			nil,         // =
			reduce(287), // unset, reduce: BeginBlock
			reduce(287), // filter, reduce: BeginBlock
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(287), // print, reduce: BeginBlock
			nil,         // ,
			reduce(287), // printn, reduce: BeginBlock
			reduce(287), // eprint, reduce: BeginBlock
			reduce(287), // eprintn, reduce: BeginBlock
			reduce(287), // dump, reduce: BeginBlock
			reduce(287), // edump, reduce: BeginBlock
			reduce(287), // tee, reduce: BeginBlock
			reduce(287), // emitf, reduce: BeginBlock
			reduce(287), // emit1, reduce: BeginBlock
			reduce(287), // emit, reduce: BeginBlock
			reduce(287), // (, reduce: BeginBlock
			nil,         // )
			reduce(287), // emitp, reduce: BeginBlock
			reduce(287), // field_name, reduce: BeginBlock
			reduce(287), // $[, reduce: BeginBlock
			nil,         // ]
			reduce(287), // braced_field_name, reduce: BeginBlock
			reduce(287), // $[[, reduce: BeginBlock
			reduce(287), // $[[[, reduce: BeginBlock
			reduce(287), // full_srec, reduce: BeginBlock
			reduce(287), // oosvar_name, reduce: BeginBlock
			reduce(287), // @[, reduce: BeginBlock
			reduce(287), // braced_oosvar_name, reduce: BeginBlock
			reduce(287), // full_oosvar, reduce: BeginBlock
			reduce(287), // all, reduce: BeginBlock
			reduce(287), // non_sigil_name, reduce: BeginBlock
			reduce(287), // arr, reduce: BeginBlock
			reduce(287), // bool, reduce: BeginBlock
			reduce(287), // float, reduce: BeginBlock
			reduce(287), // int, reduce: BeginBlock
			reduce(287), // map, reduce: BeginBlock
			reduce(287), // num, reduce: BeginBlock
			reduce(287), // str, reduce: BeginBlock
			reduce(287), // var, reduce: BeginBlock
			reduce(287), // funct, reduce: BeginBlock
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(287), // +, reduce: BeginBlock
			reduce(287), // -, reduce: BeginBlock
			reduce(287), // .+, reduce: BeginBlock
			reduce(287), // .-, reduce: BeginBlock
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(287), // !, reduce: BeginBlock
			reduce(287), // ~, reduce: BeginBlock
			nil,         // **
			reduce(287), // string_literal, reduce: BeginBlock
			reduce(287), // regex_case_insensitive, reduce: BeginBlock
			reduce(287), // int_literal, reduce: BeginBlock
			reduce(287), // float_literal, reduce: BeginBlock
			reduce(287), // boolean_literal, reduce: BeginBlock
			reduce(287), // null_literal, reduce: BeginBlock
			reduce(287), // inf_literal, reduce: BeginBlock
			reduce(287), // nan_literal, reduce: BeginBlock
			reduce(287), // const_M_PI, reduce: BeginBlock
			reduce(287), // const_M_E, reduce: BeginBlock
			reduce(287), // panic, reduce: BeginBlock
			reduce(287), // [, reduce: BeginBlock
			reduce(287), // ctx_IPS, reduce: BeginBlock
			reduce(287), // ctx_IFS, reduce: BeginBlock
			reduce(287), // ctx_IRS, reduce: BeginBlock
			reduce(287), // ctx_OPS, reduce: BeginBlock
			reduce(287), // ctx_OFS, reduce: BeginBlock
			reduce(287), // ctx_ORS, reduce: BeginBlock
			reduce(287), // ctx_FLATSEP, reduce: BeginBlock
			reduce(287), // ctx_NF, reduce: BeginBlock
			reduce(287), // ctx_NR, reduce: BeginBlock
			reduce(287), // ctx_FNR, reduce: BeginBlock
			reduce(287), // ctx_FILENAME, reduce: BeginBlock
			reduce(287), // ctx_FILENUM, reduce: BeginBlock
			reduce(287), // env, reduce: BeginBlock
			nil,         // [[
			nil,         // [[[
			reduce(287), // call, reduce: BeginBlock
			reduce(287), // begin, reduce: BeginBlock
			reduce(287), // end, reduce: BeginBlock
			reduce(287), // if, reduce: BeginBlock
			nil,         // elif
			nil,         // else
			reduce(287), // while, reduce: BeginBlock
			reduce(287), // do, reduce: BeginBlock
			reduce(287), // for, reduce: BeginBlock
			nil,         // in
			reduce(287), // break, reduce: BeginBlock
			reduce(287), // continue, reduce: BeginBlock
			reduce(287), // func, reduce: BeginBlock
			reduce(287), // subr, reduce: BeginBlock
			reduce(287), // return, reduce: BeginBlock
		},
	},
	actionRow{ // S2327
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(288), // ;, reduce: EndBlock
			reduce(288), // {, reduce: EndBlock
			reduce(288), // }, reduce: EndBlock
			nil,         // =
			reduce(288), // unset, reduce: EndBlock
			reduce(288), // filter, reduce: EndBlock
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(288), // print, reduce: EndBlock
			nil,         // ,
			reduce(288), // printn, reduce: EndBlock
			reduce(288), // eprint, reduce: EndBlock
			reduce(288), // eprintn, reduce: EndBlock
			reduce(288), // dump, reduce: EndBlock
			reduce(288), // edump, reduce: EndBlock
			reduce(288), // tee, reduce: EndBlock
			reduce(288), // emitf, reduce: EndBlock
			reduce(288), // emit1, reduce: EndBlock
			reduce(288), // emit, reduce: EndBlock
			reduce(288), // (, reduce: EndBlock
			nil,         // )
			reduce(288), // emitp, reduce: EndBlock
			reduce(288), // field_name, reduce: EndBlock
			reduce(288), // $[, reduce: EndBlock
			nil,         // ]
			reduce(288), // braced_field_name, reduce: EndBlock
			reduce(288), // $[[, reduce: EndBlock
			reduce(288), // $[[[, reduce: EndBlock
			reduce(288), // full_srec, reduce: EndBlock
			reduce(288), // oosvar_name, reduce: EndBlock
			reduce(288), // @[, reduce: EndBlock
			reduce(288), // braced_oosvar_name, reduce: EndBlock
			reduce(288), // full_oosvar, reduce: EndBlock
			reduce(288), // all, reduce: EndBlock
			reduce(288), // non_sigil_name, reduce: EndBlock
			reduce(288), // arr, reduce: EndBlock
			reduce(288), // bool, reduce: EndBlock
			reduce(288), // float, reduce: EndBlock
			reduce(288), // int, reduce: EndBlock
			reduce(288), // map, reduce: EndBlock
			reduce(288), // num, reduce: EndBlock
			reduce(288), // str, reduce: EndBlock
			reduce(288), // var, reduce: EndBlock
			reduce(288), // funct, reduce: EndBlock
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(288), // +, reduce: EndBlock
			reduce(288), // -, reduce: EndBlock
			reduce(288), // .+, reduce: EndBlock
			reduce(288), // .-, reduce: EndBlock
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(288), // !, reduce: EndBlock
			reduce(288), // ~, reduce: EndBlock
			nil,         // **
			reduce(288), // string_literal, reduce: EndBlock
			reduce(288), // regex_case_insensitive, reduce: EndBlock
			reduce(288), // int_literal, reduce: EndBlock
			reduce(288), // float_literal, reduce: EndBlock
			reduce(288), // boolean_literal, reduce: EndBlock
			reduce(288), // null_literal, reduce: EndBlock
			reduce(288), // inf_literal, reduce: EndBlock
			reduce(288), // nan_literal, reduce: EndBlock
			reduce(288), // const_M_PI, reduce: EndBlock
			reduce(288), // const_M_E, reduce: EndBlock
			reduce(288), // panic, reduce: EndBlock
			reduce(288), // [, reduce: EndBlock
			reduce(288), // ctx_IPS, reduce: EndBlock
			reduce(288), // ctx_IFS, reduce: EndBlock
			reduce(288), // ctx_IRS, reduce: EndBlock
			reduce(288), // ctx_OPS, reduce: EndBlock
			reduce(288), // ctx_OFS, reduce: EndBlock
			reduce(288), // ctx_ORS, reduce: EndBlock
			reduce(288), // ctx_FLATSEP, reduce: EndBlock
			reduce(288), // ctx_NF, reduce: EndBlock
			reduce(288), // ctx_NR, reduce: EndBlock
			reduce(288), // ctx_FNR, reduce: EndBlock
			reduce(288), // ctx_FILENAME, reduce: EndBlock
			reduce(288), // ctx_FILENUM, reduce: EndBlock
			reduce(288), // env, reduce: EndBlock
			nil,         // [[
			nil,         // [[[
			reduce(288), // call, reduce: EndBlock
			reduce(288), // begin, reduce: EndBlock
			reduce(288), // end, reduce: EndBlock
			reduce(288), // if, reduce: EndBlock
			nil,         // elif
			nil,         // else
			reduce(288), // while, reduce: EndBlock
			reduce(288), // do, reduce: EndBlock
			reduce(288), // for, reduce: EndBlock
			nil,         // in
			reduce(288), // break, reduce: EndBlock
			reduce(288), // continue, reduce: EndBlock
			reduce(288), // func, reduce: EndBlock
			reduce(288), // subr, reduce: EndBlock
			reduce(288), // return, reduce: EndBlock
		},
	},
	actionRow{ // S2328
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(291), // ;, reduce: IfChain
			reduce(291), // {, reduce: IfChain
			reduce(291), // }, reduce: IfChain
			nil,         // =
			reduce(291), // unset, reduce: IfChain
			reduce(291), // filter, reduce: IfChain
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(291), // print, reduce: IfChain
			nil,         // ,
			reduce(291), // printn, reduce: IfChain
			reduce(291), // eprint, reduce: IfChain
			reduce(291), // eprintn, reduce: IfChain
			reduce(291), // dump, reduce: IfChain
			reduce(291), // edump, reduce: IfChain
			reduce(291), // tee, reduce: IfChain
			reduce(291), // emitf, reduce: IfChain
			reduce(291), // emit1, reduce: IfChain
			reduce(291), // emit, reduce: IfChain
			reduce(291), // (, reduce: IfChain
			nil,         // )
			reduce(291), // emitp, reduce: IfChain
			reduce(291), // field_name, reduce: IfChain
			reduce(291), // $[, reduce: IfChain
			nil,         // ]
			reduce(291), // braced_field_name, reduce: IfChain
			reduce(291), // $[[, reduce: IfChain
			reduce(291), // $[[[, reduce: IfChain
			reduce(291), // full_srec, reduce: IfChain
			reduce(291), // oosvar_name, reduce: IfChain
			reduce(291), // @[, reduce: IfChain
			reduce(291), // braced_oosvar_name, reduce: IfChain
			reduce(291), // full_oosvar, reduce: IfChain
			reduce(291), // all, reduce: IfChain
			reduce(291), // non_sigil_name, reduce: IfChain
			reduce(291), // arr, reduce: IfChain
			reduce(291), // bool, reduce: IfChain
			reduce(291), // float, reduce: IfChain
			reduce(291), // int, reduce: IfChain
			reduce(291), // map, reduce: IfChain
			reduce(291), // num, reduce: IfChain
			reduce(291), // str, reduce: IfChain
			reduce(291), // var, reduce: IfChain
			reduce(291), // funct, reduce: IfChain
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(291), // +, reduce: IfChain
			reduce(291), // -, reduce: IfChain
			reduce(291), // .+, reduce: IfChain
			reduce(291), // .-, reduce: IfChain
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(291), // !, reduce: IfChain
			reduce(291), // ~, reduce: IfChain
			nil,         // **
			reduce(291), // string_literal, reduce: IfChain
			reduce(291), // regex_case_insensitive, reduce: IfChain
			reduce(291), // int_literal, reduce: IfChain
			reduce(291), // float_literal, reduce: IfChain
			reduce(291), // boolean_literal, reduce: IfChain
			reduce(291), // null_literal, reduce: IfChain
			reduce(291), // inf_literal, reduce: IfChain
			reduce(291), // nan_literal, reduce: IfChain
			reduce(291), // const_M_PI, reduce: IfChain
			reduce(291), // const_M_E, reduce: IfChain
			reduce(291), // panic, reduce: IfChain
			reduce(291), // [, reduce: IfChain
			reduce(291), // ctx_IPS, reduce: IfChain
			reduce(291), // ctx_IFS, reduce: IfChain
			reduce(291), // ctx_IRS, reduce: IfChain
			reduce(291), // ctx_OPS, reduce: IfChain
			reduce(291), // ctx_OFS, reduce: IfChain
			reduce(291), // ctx_ORS, reduce: IfChain
			reduce(291), // ctx_FLATSEP, reduce: IfChain
			reduce(291), // ctx_NF, reduce: IfChain
			reduce(291), // ctx_NR, reduce: IfChain
			reduce(291), // ctx_FNR, reduce: IfChain
			reduce(291), // ctx_FILENAME, reduce: IfChain
			reduce(291), // ctx_FILENUM, reduce: IfChain
			reduce(291), // env, reduce: IfChain
			nil,         // [[
			nil,         // [[[
			reduce(291), // call, reduce: IfChain
			reduce(291), // begin, reduce: IfChain
			reduce(291), // end, reduce: IfChain
			reduce(291), // if, reduce: IfChain
			nil,         // elif
			nil,         // else
			reduce(291), // while, reduce: IfChain
			reduce(291), // do, reduce: IfChain
			reduce(291), // for, reduce: IfChain
			nil,         // in
			reduce(291), // break, reduce: IfChain
			reduce(291), // continue, reduce: IfChain
			reduce(291), // func, reduce: IfChain
			reduce(291), // subr, reduce: IfChain
			reduce(291), // return, reduce: IfChain
		},
	},
	actionRow{ // S2329
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(293), // ;, reduce: IfElifStar
			reduce(293), // {, reduce: IfElifStar
			reduce(293), // }, reduce: IfElifStar
			nil,         // =
			reduce(293), // unset, reduce: IfElifStar
			reduce(293), // filter, reduce: IfElifStar
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(293), // print, reduce: IfElifStar
			nil,         // ,
			reduce(293), // printn, reduce: IfElifStar
			reduce(293), // eprint, reduce: IfElifStar
			reduce(293), // eprintn, reduce: IfElifStar
			reduce(293), // dump, reduce: IfElifStar
			reduce(293), // edump, reduce: IfElifStar
			reduce(293), // tee, reduce: IfElifStar
			reduce(293), // emitf, reduce: IfElifStar
			reduce(293), // emit1, reduce: IfElifStar
			reduce(293), // emit, reduce: IfElifStar
			reduce(293), // (, reduce: IfElifStar
			nil,         // )
			reduce(293), // emitp, reduce: IfElifStar
			reduce(293), // field_name, reduce: IfElifStar
			reduce(293), // $[, reduce: IfElifStar
			nil,         // ]
			reduce(293), // braced_field_name, reduce: IfElifStar
			reduce(293), // $[[, reduce: IfElifStar
			reduce(293), // $[[[, reduce: IfElifStar
			reduce(293), // full_srec, reduce: IfElifStar
			reduce(293), // oosvar_name, reduce: IfElifStar
			reduce(293), // @[, reduce: IfElifStar
			reduce(293), // braced_oosvar_name, reduce: IfElifStar
			reduce(293), // full_oosvar, reduce: IfElifStar
			reduce(293), // all, reduce: IfElifStar
			reduce(293), // non_sigil_name, reduce: IfElifStar
			reduce(293), // arr, reduce: IfElifStar
			reduce(293), // bool, reduce: IfElifStar
			reduce(293), // float, reduce: IfElifStar
			reduce(293), // int, reduce: IfElifStar
			reduce(293), // map, reduce: IfElifStar
			reduce(293), // num, reduce: IfElifStar
			reduce(293), // str, reduce: IfElifStar
			reduce(293), // var, reduce: IfElifStar
			reduce(293), // funct, reduce: IfElifStar
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(293), // +, reduce: IfElifStar
			reduce(293), // -, reduce: IfElifStar
			reduce(293), // .+, reduce: IfElifStar
			reduce(293), // .-, reduce: IfElifStar
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(293), // !, reduce: IfElifStar
			reduce(293), // ~, reduce: IfElifStar
			nil,         // **
			reduce(293), // string_literal, reduce: IfElifStar
			reduce(293), // regex_case_insensitive, reduce: IfElifStar
			reduce(293), // int_literal, reduce: IfElifStar
			reduce(293), // float_literal, reduce: IfElifStar
			reduce(293), // boolean_literal, reduce: IfElifStar
			reduce(293), // null_literal, reduce: IfElifStar
			reduce(293), // inf_literal, reduce: IfElifStar
			reduce(293), // nan_literal, reduce: IfElifStar
			reduce(293), // const_M_PI, reduce: IfElifStar
			reduce(293), // const_M_E, reduce: IfElifStar
			reduce(293), // panic, reduce: IfElifStar
			reduce(293), // [, reduce: IfElifStar
			reduce(293), // ctx_IPS, reduce: IfElifStar
			reduce(293), // ctx_IFS, reduce: IfElifStar
			reduce(293), // ctx_IRS, reduce: IfElifStar
			reduce(293), // ctx_OPS, reduce: IfElifStar
			reduce(293), // ctx_OFS, reduce: IfElifStar
			reduce(293), // ctx_ORS, reduce: IfElifStar
			reduce(293), // ctx_FLATSEP, reduce: IfElifStar
			reduce(293), // ctx_NF, reduce: IfElifStar
			reduce(293), // ctx_NR, reduce: IfElifStar
			reduce(293), // ctx_FNR, reduce: IfElifStar
			reduce(293), // ctx_FILENAME, reduce: IfElifStar
			reduce(293), // ctx_FILENUM, reduce: IfElifStar
			reduce(293), // env, reduce: IfElifStar
			nil,         // [[
			nil,         // [[[
			reduce(293), // call, reduce: IfElifStar
			reduce(293), // begin, reduce: IfElifStar
			reduce(293), // end, reduce: IfElifStar
			reduce(293), // if, reduce: IfElifStar
			reduce(293), // elif, reduce: IfElifStar
			reduce(293), // else, reduce: IfElifStar
			reduce(293), // while, reduce: IfElifStar
			reduce(293), // do, reduce: IfElifStar
			reduce(293), // for, reduce: IfElifStar
			nil,         // in
			reduce(293), // break, reduce: IfElifStar
			reduce(293), // continue, reduce: IfElifStar
			reduce(293), // func, reduce: IfElifStar
			reduce(293), // subr, reduce: IfElifStar
			reduce(293), // return, reduce: IfElifStar
		},
	},
	actionRow{ // S2330
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3187), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2331
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2042), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2332
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2333
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2334
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			shift(3191), // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2335
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(309), // ;, reduce: TripleForStart
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(309), // ,, reduce: TripleForStart
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3193), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(1793), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2336
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2337
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3196), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2338
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3197), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2339
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(336), // ;, reduce: ReturnStatement
			nil,         // {
			reduce(336), // }, reduce: ReturnStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2340
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(235), // $, reduce: MapLiteral
			nil,         // empty
			reduce(235), // ;, reduce: MapLiteral
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(235), // >, reduce: MapLiteral
			reduce(235), // >>, reduce: MapLiteral
			reduce(235), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(235), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(235), // ?, reduce: MapLiteral
			nil,         // :
			reduce(235), // ||, reduce: MapLiteral
			reduce(235), // ^^, reduce: MapLiteral
			reduce(235), // &&, reduce: MapLiteral
			reduce(235), // ??, reduce: MapLiteral
			reduce(235), // ???, reduce: MapLiteral
			reduce(235), // =~, reduce: MapLiteral
			reduce(235), // !=~, reduce: MapLiteral
			reduce(235), // ==, reduce: MapLiteral
			reduce(235), // !=, reduce: MapLiteral
			reduce(235), // <=>, reduce: MapLiteral
			reduce(235), // >=, reduce: MapLiteral
			reduce(235), // <, reduce: MapLiteral
			reduce(235), // <=, reduce: MapLiteral
			reduce(235), // ^, reduce: MapLiteral
			reduce(235), // &, reduce: MapLiteral
			reduce(235), // <<, reduce: MapLiteral
			reduce(235), // >>>, reduce: MapLiteral
			reduce(235), // +, reduce: MapLiteral
			reduce(235), // -, reduce: MapLiteral
			reduce(235), // .+, reduce: MapLiteral
			reduce(235), // .-, reduce: MapLiteral
			reduce(235), // *, reduce: MapLiteral
			reduce(235), // /, reduce: MapLiteral
			reduce(235), // //, reduce: MapLiteral
			reduce(235), // %, reduce: MapLiteral
			reduce(235), // .*, reduce: MapLiteral
			reduce(235), // ./, reduce: MapLiteral
			reduce(235), // .//, reduce: MapLiteral
			reduce(235), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(235), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(235), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(235), // [[, reduce: MapLiteral
			reduce(235), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2341
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(275), // $, reduce: FcnArgs
			nil,         // empty
			reduce(275), // ;, reduce: FcnArgs
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2342
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(208), // $, reduce: PrecedenceChainEnd
			nil,         // empty
			reduce(208), // ;, reduce: PrecedenceChainEnd
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(208), // >, reduce: PrecedenceChainEnd
			reduce(208), // >>, reduce: PrecedenceChainEnd
			reduce(208), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(208), // ,, reduce: PrecedenceChainEnd
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(208), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(208), // ||, reduce: PrecedenceChainEnd
			reduce(208), // ^^, reduce: PrecedenceChainEnd
			reduce(208), // &&, reduce: PrecedenceChainEnd
			reduce(208), // ??, reduce: PrecedenceChainEnd
			reduce(208), // ???, reduce: PrecedenceChainEnd
			reduce(208), // =~, reduce: PrecedenceChainEnd
			reduce(208), // !=~, reduce: PrecedenceChainEnd
			reduce(208), // ==, reduce: PrecedenceChainEnd
			reduce(208), // !=, reduce: PrecedenceChainEnd
			reduce(208), // <=>, reduce: PrecedenceChainEnd
			reduce(208), // >=, reduce: PrecedenceChainEnd
			reduce(208), // <, reduce: PrecedenceChainEnd
			reduce(208), // <=, reduce: PrecedenceChainEnd
			reduce(208), // ^, reduce: PrecedenceChainEnd
			reduce(208), // &, reduce: PrecedenceChainEnd
			reduce(208), // <<, reduce: PrecedenceChainEnd
			reduce(208), // >>>, reduce: PrecedenceChainEnd
			reduce(208), // +, reduce: PrecedenceChainEnd
			reduce(208), // -, reduce: PrecedenceChainEnd
			reduce(208), // .+, reduce: PrecedenceChainEnd
			reduce(208), // .-, reduce: PrecedenceChainEnd
			reduce(208), // *, reduce: PrecedenceChainEnd
			reduce(208), // /, reduce: PrecedenceChainEnd
			reduce(208), // //, reduce: PrecedenceChainEnd
			reduce(208), // %, reduce: PrecedenceChainEnd
			reduce(208), // .*, reduce: PrecedenceChainEnd
			reduce(208), // ./, reduce: PrecedenceChainEnd
			reduce(208), // .//, reduce: PrecedenceChainEnd
			reduce(208), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(208), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2343
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(105), // $, reduce: IndirectFieldValue
			nil,         // empty
			reduce(105), // ;, reduce: IndirectFieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(105), // >, reduce: IndirectFieldValue
			reduce(105), // >>, reduce: IndirectFieldValue
			reduce(105), // |, reduce: IndirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(105), // ,, reduce: IndirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(105), // ?, reduce: IndirectFieldValue
			nil,         // :
			reduce(105), // ||, reduce: IndirectFieldValue
			reduce(105), // ^^, reduce: IndirectFieldValue
			reduce(105), // &&, reduce: IndirectFieldValue
			reduce(105), // ??, reduce: IndirectFieldValue
			reduce(105), // ???, reduce: IndirectFieldValue
			reduce(105), // =~, reduce: IndirectFieldValue
			reduce(105), // !=~, reduce: IndirectFieldValue
			reduce(105), // ==, reduce: IndirectFieldValue
			reduce(105), // !=, reduce: IndirectFieldValue
			reduce(105), // <=>, reduce: IndirectFieldValue
			reduce(105), // >=, reduce: IndirectFieldValue
			reduce(105), // <, reduce: IndirectFieldValue
			reduce(105), // <=, reduce: IndirectFieldValue
			reduce(105), // ^, reduce: IndirectFieldValue
			reduce(105), // &, reduce: IndirectFieldValue
			reduce(105), // <<, reduce: IndirectFieldValue
			reduce(105), // >>>, reduce: IndirectFieldValue
			reduce(105), // +, reduce: IndirectFieldValue
			reduce(105), // -, reduce: IndirectFieldValue
			reduce(105), // .+, reduce: IndirectFieldValue
			reduce(105), // .-, reduce: IndirectFieldValue
			reduce(105), // *, reduce: IndirectFieldValue
			reduce(105), // /, reduce: IndirectFieldValue
			reduce(105), // //, reduce: IndirectFieldValue
			reduce(105), // %, reduce: IndirectFieldValue
			reduce(105), // .*, reduce: IndirectFieldValue
			reduce(105), // ./, reduce: IndirectFieldValue
			reduce(105), // .//, reduce: IndirectFieldValue
			reduce(105), // ., reduce: IndirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(105), // **, reduce: IndirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(105), // [, reduce: IndirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(105), // [[, reduce: IndirectFieldValue
			reduce(105), // [[[, reduce: IndirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2344
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3198), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2345
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3199), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2346
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(114), // $, reduce: IndirectOosvarValue
			nil,         // empty
			reduce(114), // ;, reduce: IndirectOosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(114), // >, reduce: IndirectOosvarValue
			reduce(114), // >>, reduce: IndirectOosvarValue
			reduce(114), // |, reduce: IndirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(114), // ,, reduce: IndirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(114), // ?, reduce: IndirectOosvarValue
			nil,         // :
			reduce(114), // ||, reduce: IndirectOosvarValue
			reduce(114), // ^^, reduce: IndirectOosvarValue
			reduce(114), // &&, reduce: IndirectOosvarValue
			reduce(114), // ??, reduce: IndirectOosvarValue
			reduce(114), // ???, reduce: IndirectOosvarValue
			reduce(114), // =~, reduce: IndirectOosvarValue
			reduce(114), // !=~, reduce: IndirectOosvarValue
			reduce(114), // ==, reduce: IndirectOosvarValue
			reduce(114), // !=, reduce: IndirectOosvarValue
			reduce(114), // <=>, reduce: IndirectOosvarValue
			reduce(114), // >=, reduce: IndirectOosvarValue
			reduce(114), // <, reduce: IndirectOosvarValue
			reduce(114), // <=, reduce: IndirectOosvarValue
			reduce(114), // ^, reduce: IndirectOosvarValue
			reduce(114), // &, reduce: IndirectOosvarValue
			reduce(114), // <<, reduce: IndirectOosvarValue
			reduce(114), // >>>, reduce: IndirectOosvarValue
			reduce(114), // +, reduce: IndirectOosvarValue
			reduce(114), // -, reduce: IndirectOosvarValue
			reduce(114), // .+, reduce: IndirectOosvarValue
			reduce(114), // .-, reduce: IndirectOosvarValue
			reduce(114), // *, reduce: IndirectOosvarValue
			reduce(114), // /, reduce: IndirectOosvarValue
			reduce(114), // //, reduce: IndirectOosvarValue
			reduce(114), // %, reduce: IndirectOosvarValue
			reduce(114), // .*, reduce: IndirectOosvarValue
			reduce(114), // ./, reduce: IndirectOosvarValue
			reduce(114), // .//, reduce: IndirectOosvarValue
			reduce(114), // ., reduce: IndirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(114), // **, reduce: IndirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(114), // [, reduce: IndirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(114), // [[, reduce: IndirectOosvarValue
			reduce(114), // [[[, reduce: IndirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2347
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3200), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2348
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(151), // $, reduce: LogicalOrTerm
			nil,         // empty
			reduce(151), // ;, reduce: LogicalOrTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(151), // ,, reduce: LogicalOrTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(151), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(151), // ||, reduce: LogicalOrTerm
			shift(1113), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2349
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(153), // $, reduce: LogicalXORTerm
			nil,         // empty
			reduce(153), // ;, reduce: LogicalXORTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(153), // ,, reduce: LogicalXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(153), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(153), // ||, reduce: LogicalXORTerm
			reduce(153), // ^^, reduce: LogicalXORTerm
			shift(1114), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2350
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(155), // $, reduce: LogicalAndTerm
			nil,         // empty
			reduce(155), // ;, reduce: LogicalAndTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(155), // ,, reduce: LogicalAndTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(155), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(155), // ||, reduce: LogicalAndTerm
			reduce(155), // ^^, reduce: LogicalAndTerm
			reduce(155), // &&, reduce: LogicalAndTerm
			shift(1115), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2351
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(157), // $, reduce: AbsentCoalesceTerm
			nil,         // empty
			reduce(157), // ;, reduce: AbsentCoalesceTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(157), // ,, reduce: AbsentCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(157), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(157), // ||, reduce: AbsentCoalesceTerm
			reduce(157), // ^^, reduce: AbsentCoalesceTerm
			reduce(157), // &&, reduce: AbsentCoalesceTerm
			reduce(157), // ??, reduce: AbsentCoalesceTerm
			shift(1116), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2352
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(159), // $, reduce: EmptyCoalesceTerm
			nil,         // empty
			reduce(159), // ;, reduce: EmptyCoalesceTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(159), // ,, reduce: EmptyCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(159), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(159), // ||, reduce: EmptyCoalesceTerm
			reduce(159), // ^^, reduce: EmptyCoalesceTerm
			reduce(159), // &&, reduce: EmptyCoalesceTerm
			reduce(159), // ??, reduce: EmptyCoalesceTerm
			reduce(159), // ???, reduce: EmptyCoalesceTerm
			shift(1117), // =~
			shift(1118), // !=~
			shift(1119), // ==
			shift(1120), // !=
			shift(1121), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2353
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(161), // $, reduce: EqneTerm
			nil,         // empty
			reduce(161), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1122), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(161), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(161), // ?, reduce: EqneTerm
			nil,         // :
			reduce(161), // ||, reduce: EqneTerm
			reduce(161), // ^^, reduce: EqneTerm
			reduce(161), // &&, reduce: EqneTerm
			reduce(161), // ??, reduce: EqneTerm
			reduce(161), // ???, reduce: EqneTerm
			reduce(161), // =~, reduce: EqneTerm
			reduce(161), // !=~, reduce: EqneTerm
			reduce(161), // ==, reduce: EqneTerm
			reduce(161), // !=, reduce: EqneTerm
			reduce(161), // <=>, reduce: EqneTerm
			shift(1123), // >=
			shift(1124), // <
			shift(1125), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2354
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(162), // $, reduce: EqneTerm
			nil,         // empty
			reduce(162), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1122), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(162), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(162), // ?, reduce: EqneTerm
			nil,         // :
			reduce(162), // ||, reduce: EqneTerm
			reduce(162), // ^^, reduce: EqneTerm
			reduce(162), // &&, reduce: EqneTerm
			reduce(162), // ??, reduce: EqneTerm
			reduce(162), // ???, reduce: EqneTerm
			reduce(162), // =~, reduce: EqneTerm
			reduce(162), // !=~, reduce: EqneTerm
			reduce(162), // ==, reduce: EqneTerm
			reduce(162), // !=, reduce: EqneTerm
			reduce(162), // <=>, reduce: EqneTerm
			shift(1123), // >=
			shift(1124), // <
			shift(1125), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2355
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(163), // $, reduce: EqneTerm
			nil,         // empty
			reduce(163), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1122), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(163), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(163), // ?, reduce: EqneTerm
			nil,         // :
			reduce(163), // ||, reduce: EqneTerm
			reduce(163), // ^^, reduce: EqneTerm
			reduce(163), // &&, reduce: EqneTerm
			reduce(163), // ??, reduce: EqneTerm
			reduce(163), // ???, reduce: EqneTerm
			reduce(163), // =~, reduce: EqneTerm
			reduce(163), // !=~, reduce: EqneTerm
			reduce(163), // ==, reduce: EqneTerm
			reduce(163), // !=, reduce: EqneTerm
			reduce(163), // <=>, reduce: EqneTerm
			shift(1123), // >=
			shift(1124), // <
			shift(1125), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2356
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(164), // $, reduce: EqneTerm
			nil,         // empty
			reduce(164), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1122), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(164), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(164), // ?, reduce: EqneTerm
			nil,         // :
			reduce(164), // ||, reduce: EqneTerm
			reduce(164), // ^^, reduce: EqneTerm
			reduce(164), // &&, reduce: EqneTerm
			reduce(164), // ??, reduce: EqneTerm
			reduce(164), // ???, reduce: EqneTerm
			reduce(164), // =~, reduce: EqneTerm
			reduce(164), // !=~, reduce: EqneTerm
			reduce(164), // ==, reduce: EqneTerm
			reduce(164), // !=, reduce: EqneTerm
			reduce(164), // <=>, reduce: EqneTerm
			shift(1123), // >=
			shift(1124), // <
			shift(1125), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2357
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(165), // $, reduce: EqneTerm
			nil,         // empty
			reduce(165), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1122), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(165), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(165), // ?, reduce: EqneTerm
			nil,         // :
			reduce(165), // ||, reduce: EqneTerm
			reduce(165), // ^^, reduce: EqneTerm
			reduce(165), // &&, reduce: EqneTerm
			reduce(165), // ??, reduce: EqneTerm
			reduce(165), // ???, reduce: EqneTerm
			reduce(165), // =~, reduce: EqneTerm
			reduce(165), // !=~, reduce: EqneTerm
			reduce(165), // ==, reduce: EqneTerm
			reduce(165), // !=, reduce: EqneTerm
			reduce(165), // <=>, reduce: EqneTerm
			shift(1123), // >=
			shift(1124), // <
			shift(1125), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2358
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(167), // $, reduce: CmpTerm
			nil,         // empty
			reduce(167), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(167), // >, reduce: CmpTerm
			nil,         // >>
			shift(1126), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(167), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(167), // ?, reduce: CmpTerm
			nil,         // :
			reduce(167), // ||, reduce: CmpTerm
			reduce(167), // ^^, reduce: CmpTerm
			reduce(167), // &&, reduce: CmpTerm
			reduce(167), // ??, reduce: CmpTerm
			reduce(167), // ???, reduce: CmpTerm
			reduce(167), // =~, reduce: CmpTerm
			reduce(167), // !=~, reduce: CmpTerm
			reduce(167), // ==, reduce: CmpTerm
			reduce(167), // !=, reduce: CmpTerm
			reduce(167), // <=>, reduce: CmpTerm
			reduce(167), // >=, reduce: CmpTerm
			reduce(167), // <, reduce: CmpTerm
			reduce(167), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2359
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(168), // $, reduce: CmpTerm
			nil,         // empty
			reduce(168), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(168), // >, reduce: CmpTerm
			nil,         // >>
			shift(1126), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(168), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(168), // ?, reduce: CmpTerm
			nil,         // :
			reduce(168), // ||, reduce: CmpTerm
			reduce(168), // ^^, reduce: CmpTerm
			reduce(168), // &&, reduce: CmpTerm
			reduce(168), // ??, reduce: CmpTerm
			reduce(168), // ???, reduce: CmpTerm
			reduce(168), // =~, reduce: CmpTerm
			reduce(168), // !=~, reduce: CmpTerm
			reduce(168), // ==, reduce: CmpTerm
			reduce(168), // !=, reduce: CmpTerm
			reduce(168), // <=>, reduce: CmpTerm
			reduce(168), // >=, reduce: CmpTerm
			reduce(168), // <, reduce: CmpTerm
			reduce(168), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2360
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(169), // $, reduce: CmpTerm
			nil,         // empty
			reduce(169), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(169), // >, reduce: CmpTerm
			nil,         // >>
			shift(1126), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(169), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(169), // ?, reduce: CmpTerm
			nil,         // :
			reduce(169), // ||, reduce: CmpTerm
			reduce(169), // ^^, reduce: CmpTerm
			reduce(169), // &&, reduce: CmpTerm
			reduce(169), // ??, reduce: CmpTerm
			reduce(169), // ???, reduce: CmpTerm
			reduce(169), // =~, reduce: CmpTerm
			reduce(169), // !=~, reduce: CmpTerm
			reduce(169), // ==, reduce: CmpTerm
			reduce(169), // !=, reduce: CmpTerm
			reduce(169), // <=>, reduce: CmpTerm
			reduce(169), // >=, reduce: CmpTerm
			reduce(169), // <, reduce: CmpTerm
			reduce(169), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2361
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(170), // $, reduce: CmpTerm
			nil,         // empty
			reduce(170), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(170), // >, reduce: CmpTerm
			nil,         // >>
			shift(1126), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(170), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(170), // ?, reduce: CmpTerm
			nil,         // :
			reduce(170), // ||, reduce: CmpTerm
			reduce(170), // ^^, reduce: CmpTerm
			reduce(170), // &&, reduce: CmpTerm
			reduce(170), // ??, reduce: CmpTerm
			reduce(170), // ???, reduce: CmpTerm
			reduce(170), // =~, reduce: CmpTerm
			reduce(170), // !=~, reduce: CmpTerm
			reduce(170), // ==, reduce: CmpTerm
			reduce(170), // !=, reduce: CmpTerm
			reduce(170), // <=>, reduce: CmpTerm
			reduce(170), // >=, reduce: CmpTerm
			reduce(170), // <, reduce: CmpTerm
			reduce(170), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2362
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(172), // $, reduce: BitwiseORTerm
			nil,         // empty
			reduce(172), // ;, reduce: BitwiseORTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(172), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(172), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(172), // ,, reduce: BitwiseORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(172), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(172), // ||, reduce: BitwiseORTerm
			reduce(172), // ^^, reduce: BitwiseORTerm
			reduce(172), // &&, reduce: BitwiseORTerm
			reduce(172), // ??, reduce: BitwiseORTerm
			reduce(172), // ???, reduce: BitwiseORTerm
			reduce(172), // =~, reduce: BitwiseORTerm
			reduce(172), // !=~, reduce: BitwiseORTerm
			reduce(172), // ==, reduce: BitwiseORTerm
			reduce(172), // !=, reduce: BitwiseORTerm
			reduce(172), // <=>, reduce: BitwiseORTerm
			reduce(172), // >=, reduce: BitwiseORTerm
			reduce(172), // <, reduce: BitwiseORTerm
			reduce(172), // <=, reduce: BitwiseORTerm
			shift(1127), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2363
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(174), // $, reduce: BitwiseXORTerm
			nil,         // empty
			reduce(174), // ;, reduce: BitwiseXORTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(174), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(174), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(174), // ,, reduce: BitwiseXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(174), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(174), // ||, reduce: BitwiseXORTerm
			reduce(174), // ^^, reduce: BitwiseXORTerm
			reduce(174), // &&, reduce: BitwiseXORTerm
			reduce(174), // ??, reduce: BitwiseXORTerm
			reduce(174), // ???, reduce: BitwiseXORTerm
			reduce(174), // =~, reduce: BitwiseXORTerm
			reduce(174), // !=~, reduce: BitwiseXORTerm
			reduce(174), // ==, reduce: BitwiseXORTerm
			reduce(174), // !=, reduce: BitwiseXORTerm
			reduce(174), // <=>, reduce: BitwiseXORTerm
			reduce(174), // >=, reduce: BitwiseXORTerm
			reduce(174), // <, reduce: BitwiseXORTerm
			reduce(174), // <=, reduce: BitwiseXORTerm
			reduce(174), // ^, reduce: BitwiseXORTerm
			shift(1128), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2364
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(176), // $, reduce: BitwiseANDTerm
			nil,         // empty
			reduce(176), // ;, reduce: BitwiseANDTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(176), // >, reduce: BitwiseANDTerm
			shift(1129), // >>
			reduce(176), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(176), // ,, reduce: BitwiseANDTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(176), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(176), // ||, reduce: BitwiseANDTerm
			reduce(176), // ^^, reduce: BitwiseANDTerm
			reduce(176), // &&, reduce: BitwiseANDTerm
			reduce(176), // ??, reduce: BitwiseANDTerm
			reduce(176), // ???, reduce: BitwiseANDTerm
			reduce(176), // =~, reduce: BitwiseANDTerm
			reduce(176), // !=~, reduce: BitwiseANDTerm
			reduce(176), // ==, reduce: BitwiseANDTerm
			reduce(176), // !=, reduce: BitwiseANDTerm
			reduce(176), // <=>, reduce: BitwiseANDTerm
			reduce(176), // >=, reduce: BitwiseANDTerm
			reduce(176), // <, reduce: BitwiseANDTerm
			reduce(176), // <=, reduce: BitwiseANDTerm
			reduce(176), // ^, reduce: BitwiseANDTerm
			reduce(176), // &, reduce: BitwiseANDTerm
			shift(1130), // <<
			shift(1131), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2365
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(179), // $, reduce: BitwiseShiftTerm
			nil,         // empty
			reduce(179), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(179), // >, reduce: BitwiseShiftTerm
			reduce(179), // >>, reduce: BitwiseShiftTerm
			reduce(179), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(179), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(179), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(179), // ||, reduce: BitwiseShiftTerm
			reduce(179), // ^^, reduce: BitwiseShiftTerm
			reduce(179), // &&, reduce: BitwiseShiftTerm
			reduce(179), // ??, reduce: BitwiseShiftTerm
			reduce(179), // ???, reduce: BitwiseShiftTerm
			reduce(179), // =~, reduce: BitwiseShiftTerm
			reduce(179), // !=~, reduce: BitwiseShiftTerm
			reduce(179), // ==, reduce: BitwiseShiftTerm
			reduce(179), // !=, reduce: BitwiseShiftTerm
			reduce(179), // <=>, reduce: BitwiseShiftTerm
			reduce(179), // >=, reduce: BitwiseShiftTerm
			reduce(179), // <, reduce: BitwiseShiftTerm
			reduce(179), // <=, reduce: BitwiseShiftTerm
			reduce(179), // ^, reduce: BitwiseShiftTerm
			reduce(179), // &, reduce: BitwiseShiftTerm
			reduce(179), // <<, reduce: BitwiseShiftTerm
			reduce(179), // >>>, reduce: BitwiseShiftTerm
			shift(1132), // +
			shift(1133), // -
			shift(1134), // .+
			shift(1135), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2366
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(178), // $, reduce: BitwiseShiftTerm
			nil,         // empty
			reduce(178), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(178), // >, reduce: BitwiseShiftTerm
			reduce(178), // >>, reduce: BitwiseShiftTerm
			reduce(178), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(178), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(178), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(178), // ||, reduce: BitwiseShiftTerm
			reduce(178), // ^^, reduce: BitwiseShiftTerm
			reduce(178), // &&, reduce: BitwiseShiftTerm
			reduce(178), // ??, reduce: BitwiseShiftTerm
			reduce(178), // ???, reduce: BitwiseShiftTerm
			reduce(178), // =~, reduce: BitwiseShiftTerm
			reduce(178), // !=~, reduce: BitwiseShiftTerm
			reduce(178), // ==, reduce: BitwiseShiftTerm
			reduce(178), // !=, reduce: BitwiseShiftTerm
			reduce(178), // <=>, reduce: BitwiseShiftTerm
			reduce(178), // >=, reduce: BitwiseShiftTerm
			reduce(178), // <, reduce: BitwiseShiftTerm
			reduce(178), // <=, reduce: BitwiseShiftTerm
			reduce(178), // ^, reduce: BitwiseShiftTerm
			reduce(178), // &, reduce: BitwiseShiftTerm
			reduce(178), // <<, reduce: BitwiseShiftTerm
			reduce(178), // >>>, reduce: BitwiseShiftTerm
			shift(1132), // +
			shift(1133), // -
			shift(1134), // .+
			shift(1135), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2367
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(180), // $, reduce: BitwiseShiftTerm
			nil,         // empty
			reduce(180), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(180), // >, reduce: BitwiseShiftTerm
			reduce(180), // >>, reduce: BitwiseShiftTerm
			reduce(180), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(180), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(180), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(180), // ||, reduce: BitwiseShiftTerm
			reduce(180), // ^^, reduce: BitwiseShiftTerm
			reduce(180), // &&, reduce: BitwiseShiftTerm
			reduce(180), // ??, reduce: BitwiseShiftTerm
			reduce(180), // ???, reduce: BitwiseShiftTerm
			reduce(180), // =~, reduce: BitwiseShiftTerm
			reduce(180), // !=~, reduce: BitwiseShiftTerm
			reduce(180), // ==, reduce: BitwiseShiftTerm
			reduce(180), // !=, reduce: BitwiseShiftTerm
			reduce(180), // <=>, reduce: BitwiseShiftTerm
			reduce(180), // >=, reduce: BitwiseShiftTerm
			reduce(180), // <, reduce: BitwiseShiftTerm
			reduce(180), // <=, reduce: BitwiseShiftTerm
			reduce(180), // ^, reduce: BitwiseShiftTerm
			reduce(180), // &, reduce: BitwiseShiftTerm
			reduce(180), // <<, reduce: BitwiseShiftTerm
			reduce(180), // >>>, reduce: BitwiseShiftTerm
			shift(1132), // +
			shift(1133), // -
			shift(1134), // .+
			shift(1135), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2368
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(182), // $, reduce: AddsubdotTerm
			nil,         // empty
			reduce(182), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(182), // >, reduce: AddsubdotTerm
			reduce(182), // >>, reduce: AddsubdotTerm
			reduce(182), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(182), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(182), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(182), // ||, reduce: AddsubdotTerm
			reduce(182), // ^^, reduce: AddsubdotTerm
			reduce(182), // &&, reduce: AddsubdotTerm
			reduce(182), // ??, reduce: AddsubdotTerm
			reduce(182), // ???, reduce: AddsubdotTerm
			reduce(182), // =~, reduce: AddsubdotTerm
			reduce(182), // !=~, reduce: AddsubdotTerm
			reduce(182), // ==, reduce: AddsubdotTerm
			reduce(182), // !=, reduce: AddsubdotTerm
			reduce(182), // <=>, reduce: AddsubdotTerm
			reduce(182), // >=, reduce: AddsubdotTerm
			reduce(182), // <, reduce: AddsubdotTerm
			reduce(182), // <=, reduce: AddsubdotTerm
			reduce(182), // ^, reduce: AddsubdotTerm
			reduce(182), // &, reduce: AddsubdotTerm
			reduce(182), // <<, reduce: AddsubdotTerm
			reduce(182), // >>>, reduce: AddsubdotTerm
			reduce(182), // +, reduce: AddsubdotTerm
			reduce(182), // -, reduce: AddsubdotTerm
			reduce(182), // .+, reduce: AddsubdotTerm
			reduce(182), // .-, reduce: AddsubdotTerm
			shift(1137), // *
			shift(1138), // /
			shift(1139), // //
			shift(1140), // %
			shift(1141), // .*
			shift(1142), // ./
			shift(1143), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2369
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(183), // $, reduce: AddsubdotTerm
			nil,         // empty
			reduce(183), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(183), // >, reduce: AddsubdotTerm
			reduce(183), // >>, reduce: AddsubdotTerm
			reduce(183), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(183), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(183), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(183), // ||, reduce: AddsubdotTerm
			reduce(183), // ^^, reduce: AddsubdotTerm
			reduce(183), // &&, reduce: AddsubdotTerm
			reduce(183), // ??, reduce: AddsubdotTerm
			reduce(183), // ???, reduce: AddsubdotTerm
			reduce(183), // =~, reduce: AddsubdotTerm
			reduce(183), // !=~, reduce: AddsubdotTerm
			reduce(183), // ==, reduce: AddsubdotTerm
			reduce(183), // !=, reduce: AddsubdotTerm
			reduce(183), // <=>, reduce: AddsubdotTerm
			reduce(183), // >=, reduce: AddsubdotTerm
			reduce(183), // <, reduce: AddsubdotTerm
			reduce(183), // <=, reduce: AddsubdotTerm
			reduce(183), // ^, reduce: AddsubdotTerm
			reduce(183), // &, reduce: AddsubdotTerm
			reduce(183), // <<, reduce: AddsubdotTerm
			reduce(183), // >>>, reduce: AddsubdotTerm
			reduce(183), // +, reduce: AddsubdotTerm
			reduce(183), // -, reduce: AddsubdotTerm
			reduce(183), // .+, reduce: AddsubdotTerm
			reduce(183), // .-, reduce: AddsubdotTerm
			shift(1137), // *
			shift(1138), // /
			shift(1139), // //
			shift(1140), // %
			shift(1141), // .*
			shift(1142), // ./
			shift(1143), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2370
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(184), // $, reduce: AddsubdotTerm
			nil,         // empty
			reduce(184), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(184), // >, reduce: AddsubdotTerm
			reduce(184), // >>, reduce: AddsubdotTerm
			reduce(184), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(184), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(184), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(184), // ||, reduce: AddsubdotTerm
			reduce(184), // ^^, reduce: AddsubdotTerm
			reduce(184), // &&, reduce: AddsubdotTerm
			reduce(184), // ??, reduce: AddsubdotTerm
			reduce(184), // ???, reduce: AddsubdotTerm
			reduce(184), // =~, reduce: AddsubdotTerm
			reduce(184), // !=~, reduce: AddsubdotTerm
			reduce(184), // ==, reduce: AddsubdotTerm
			reduce(184), // !=, reduce: AddsubdotTerm
			reduce(184), // <=>, reduce: AddsubdotTerm
			reduce(184), // >=, reduce: AddsubdotTerm
			reduce(184), // <, reduce: AddsubdotTerm
			reduce(184), // <=, reduce: AddsubdotTerm
			reduce(184), // ^, reduce: AddsubdotTerm
			reduce(184), // &, reduce: AddsubdotTerm
			reduce(184), // <<, reduce: AddsubdotTerm
			reduce(184), // >>>, reduce: AddsubdotTerm
			reduce(184), // +, reduce: AddsubdotTerm
			reduce(184), // -, reduce: AddsubdotTerm
			reduce(184), // .+, reduce: AddsubdotTerm
			reduce(184), // .-, reduce: AddsubdotTerm
			shift(1137), // *
			shift(1138), // /
			shift(1139), // //
			shift(1140), // %
			shift(1141), // .*
			shift(1142), // ./
			shift(1143), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2371
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(185), // $, reduce: AddsubdotTerm
			nil,         // empty
			reduce(185), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(185), // >, reduce: AddsubdotTerm
			reduce(185), // >>, reduce: AddsubdotTerm
			reduce(185), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(185), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(185), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(185), // ||, reduce: AddsubdotTerm
			reduce(185), // ^^, reduce: AddsubdotTerm
			reduce(185), // &&, reduce: AddsubdotTerm
			reduce(185), // ??, reduce: AddsubdotTerm
			reduce(185), // ???, reduce: AddsubdotTerm
			reduce(185), // =~, reduce: AddsubdotTerm
			reduce(185), // !=~, reduce: AddsubdotTerm
			reduce(185), // ==, reduce: AddsubdotTerm
			reduce(185), // !=, reduce: AddsubdotTerm
			reduce(185), // <=>, reduce: AddsubdotTerm
			reduce(185), // >=, reduce: AddsubdotTerm
			reduce(185), // <, reduce: AddsubdotTerm
			reduce(185), // <=, reduce: AddsubdotTerm
			reduce(185), // ^, reduce: AddsubdotTerm
			reduce(185), // &, reduce: AddsubdotTerm
			reduce(185), // <<, reduce: AddsubdotTerm
			reduce(185), // >>>, reduce: AddsubdotTerm
			reduce(185), // +, reduce: AddsubdotTerm
			reduce(185), // -, reduce: AddsubdotTerm
			reduce(185), // .+, reduce: AddsubdotTerm
			reduce(185), // .-, reduce: AddsubdotTerm
			shift(1137), // *
			shift(1138), // /
			shift(1139), // //
			shift(1140), // %
			shift(1141), // .*
			shift(1142), // ./
			shift(1143), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2372
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(187), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(187), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(187), // >, reduce: MuldivTerm
			reduce(187), // >>, reduce: MuldivTerm
			reduce(187), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(187), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(187), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(187), // ||, reduce: MuldivTerm
			reduce(187), // ^^, reduce: MuldivTerm
			reduce(187), // &&, reduce: MuldivTerm
			reduce(187), // ??, reduce: MuldivTerm
			reduce(187), // ???, reduce: MuldivTerm
			reduce(187), // =~, reduce: MuldivTerm
			reduce(187), // !=~, reduce: MuldivTerm
			reduce(187), // ==, reduce: MuldivTerm
			reduce(187), // !=, reduce: MuldivTerm
			reduce(187), // <=>, reduce: MuldivTerm
			reduce(187), // >=, reduce: MuldivTerm
			reduce(187), // <, reduce: MuldivTerm
			reduce(187), // <=, reduce: MuldivTerm
			reduce(187), // ^, reduce: MuldivTerm
			reduce(187), // &, reduce: MuldivTerm
			reduce(187), // <<, reduce: MuldivTerm
			reduce(187), // >>>, reduce: MuldivTerm
			reduce(187), // +, reduce: MuldivTerm
			reduce(187), // -, reduce: MuldivTerm
			reduce(187), // .+, reduce: MuldivTerm
			reduce(187), // .-, reduce: MuldivTerm
			reduce(187), // *, reduce: MuldivTerm
			reduce(187), // /, reduce: MuldivTerm
			reduce(187), // //, reduce: MuldivTerm
			reduce(187), // %, reduce: MuldivTerm
			reduce(187), // .*, reduce: MuldivTerm
			reduce(187), // ./, reduce: MuldivTerm
			reduce(187), // .//, reduce: MuldivTerm
			shift(1147), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2373
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(188), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(188), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(188), // >, reduce: MuldivTerm
			reduce(188), // >>, reduce: MuldivTerm
			reduce(188), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(188), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(188), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(188), // ||, reduce: MuldivTerm
			reduce(188), // ^^, reduce: MuldivTerm
			reduce(188), // &&, reduce: MuldivTerm
			reduce(188), // ??, reduce: MuldivTerm
			reduce(188), // ???, reduce: MuldivTerm
			reduce(188), // =~, reduce: MuldivTerm
			reduce(188), // !=~, reduce: MuldivTerm
			reduce(188), // ==, reduce: MuldivTerm
			reduce(188), // !=, reduce: MuldivTerm
			reduce(188), // <=>, reduce: MuldivTerm
			reduce(188), // >=, reduce: MuldivTerm
			reduce(188), // <, reduce: MuldivTerm
			reduce(188), // <=, reduce: MuldivTerm
			reduce(188), // ^, reduce: MuldivTerm
			reduce(188), // &, reduce: MuldivTerm
			reduce(188), // <<, reduce: MuldivTerm
			reduce(188), // >>>, reduce: MuldivTerm
			reduce(188), // +, reduce: MuldivTerm
			reduce(188), // -, reduce: MuldivTerm
			reduce(188), // .+, reduce: MuldivTerm
			reduce(188), // .-, reduce: MuldivTerm
			reduce(188), // *, reduce: MuldivTerm
			reduce(188), // /, reduce: MuldivTerm
			reduce(188), // //, reduce: MuldivTerm
			reduce(188), // %, reduce: MuldivTerm
			reduce(188), // .*, reduce: MuldivTerm
			reduce(188), // ./, reduce: MuldivTerm
			reduce(188), // .//, reduce: MuldivTerm
			shift(1147), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2374
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(189), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(189), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(189), // >, reduce: MuldivTerm
			reduce(189), // >>, reduce: MuldivTerm
			reduce(189), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(189), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(189), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(189), // ||, reduce: MuldivTerm
			reduce(189), // ^^, reduce: MuldivTerm
			reduce(189), // &&, reduce: MuldivTerm
			reduce(189), // ??, reduce: MuldivTerm
			reduce(189), // ???, reduce: MuldivTerm
			reduce(189), // =~, reduce: MuldivTerm
			reduce(189), // !=~, reduce: MuldivTerm
			reduce(189), // ==, reduce: MuldivTerm
			reduce(189), // !=, reduce: MuldivTerm
			reduce(189), // <=>, reduce: MuldivTerm
			reduce(189), // >=, reduce: MuldivTerm
			reduce(189), // <, reduce: MuldivTerm
			reduce(189), // <=, reduce: MuldivTerm
			reduce(189), // ^, reduce: MuldivTerm
			reduce(189), // &, reduce: MuldivTerm
			reduce(189), // <<, reduce: MuldivTerm
			reduce(189), // >>>, reduce: MuldivTerm
			reduce(189), // +, reduce: MuldivTerm
			reduce(189), // -, reduce: MuldivTerm
			reduce(189), // .+, reduce: MuldivTerm
			reduce(189), // .-, reduce: MuldivTerm
			reduce(189), // *, reduce: MuldivTerm
			reduce(189), // /, reduce: MuldivTerm
			reduce(189), // //, reduce: MuldivTerm
			reduce(189), // %, reduce: MuldivTerm
			reduce(189), // .*, reduce: MuldivTerm
			reduce(189), // ./, reduce: MuldivTerm
			reduce(189), // .//, reduce: MuldivTerm
			shift(1147), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2375
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(190), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(190), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(190), // >, reduce: MuldivTerm
			reduce(190), // >>, reduce: MuldivTerm
			reduce(190), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(190), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(190), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(190), // ||, reduce: MuldivTerm
			reduce(190), // ^^, reduce: MuldivTerm
			reduce(190), // &&, reduce: MuldivTerm
			reduce(190), // ??, reduce: MuldivTerm
			reduce(190), // ???, reduce: MuldivTerm
			reduce(190), // =~, reduce: MuldivTerm
			reduce(190), // !=~, reduce: MuldivTerm
			reduce(190), // ==, reduce: MuldivTerm
			reduce(190), // !=, reduce: MuldivTerm
			reduce(190), // <=>, reduce: MuldivTerm
			reduce(190), // >=, reduce: MuldivTerm
			reduce(190), // <, reduce: MuldivTerm
			reduce(190), // <=, reduce: MuldivTerm
			reduce(190), // ^, reduce: MuldivTerm
			reduce(190), // &, reduce: MuldivTerm
			reduce(190), // <<, reduce: MuldivTerm
			reduce(190), // >>>, reduce: MuldivTerm
			reduce(190), // +, reduce: MuldivTerm
			reduce(190), // -, reduce: MuldivTerm
			reduce(190), // .+, reduce: MuldivTerm
			reduce(190), // .-, reduce: MuldivTerm
			reduce(190), // *, reduce: MuldivTerm
			reduce(190), // /, reduce: MuldivTerm
			reduce(190), // //, reduce: MuldivTerm
			reduce(190), // %, reduce: MuldivTerm
			reduce(190), // .*, reduce: MuldivTerm
			reduce(190), // ./, reduce: MuldivTerm
			reduce(190), // .//, reduce: MuldivTerm
			shift(1147), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2376
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(191), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(191), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(191), // >, reduce: MuldivTerm
			reduce(191), // >>, reduce: MuldivTerm
			reduce(191), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(191), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(191), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(191), // ||, reduce: MuldivTerm
			reduce(191), // ^^, reduce: MuldivTerm
			reduce(191), // &&, reduce: MuldivTerm
			reduce(191), // ??, reduce: MuldivTerm
			reduce(191), // ???, reduce: MuldivTerm
			reduce(191), // =~, reduce: MuldivTerm
			reduce(191), // !=~, reduce: MuldivTerm
			reduce(191), // ==, reduce: MuldivTerm
			reduce(191), // !=, reduce: MuldivTerm
			reduce(191), // <=>, reduce: MuldivTerm
			reduce(191), // >=, reduce: MuldivTerm
			reduce(191), // <, reduce: MuldivTerm
			reduce(191), // <=, reduce: MuldivTerm
			reduce(191), // ^, reduce: MuldivTerm
			reduce(191), // &, reduce: MuldivTerm
			reduce(191), // <<, reduce: MuldivTerm
			reduce(191), // >>>, reduce: MuldivTerm
			reduce(191), // +, reduce: MuldivTerm
			reduce(191), // -, reduce: MuldivTerm
			reduce(191), // .+, reduce: MuldivTerm
			reduce(191), // .-, reduce: MuldivTerm
			reduce(191), // *, reduce: MuldivTerm
			reduce(191), // /, reduce: MuldivTerm
			reduce(191), // //, reduce: MuldivTerm
			reduce(191), // %, reduce: MuldivTerm
			reduce(191), // .*, reduce: MuldivTerm
			reduce(191), // ./, reduce: MuldivTerm
			reduce(191), // .//, reduce: MuldivTerm
			shift(1147), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2377
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(192), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(192), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(192), // >, reduce: MuldivTerm
			reduce(192), // >>, reduce: MuldivTerm
			reduce(192), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(192), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(192), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(192), // ||, reduce: MuldivTerm
			reduce(192), // ^^, reduce: MuldivTerm
			reduce(192), // &&, reduce: MuldivTerm
			reduce(192), // ??, reduce: MuldivTerm
			reduce(192), // ???, reduce: MuldivTerm
			reduce(192), // =~, reduce: MuldivTerm
			reduce(192), // !=~, reduce: MuldivTerm
			reduce(192), // ==, reduce: MuldivTerm
			reduce(192), // !=, reduce: MuldivTerm
			reduce(192), // <=>, reduce: MuldivTerm
			reduce(192), // >=, reduce: MuldivTerm
			reduce(192), // <, reduce: MuldivTerm
			reduce(192), // <=, reduce: MuldivTerm
			reduce(192), // ^, reduce: MuldivTerm
			reduce(192), // &, reduce: MuldivTerm
			reduce(192), // <<, reduce: MuldivTerm
			reduce(192), // >>>, reduce: MuldivTerm
			reduce(192), // +, reduce: MuldivTerm
			reduce(192), // -, reduce: MuldivTerm
			reduce(192), // .+, reduce: MuldivTerm
			reduce(192), // .-, reduce: MuldivTerm
			reduce(192), // *, reduce: MuldivTerm
			reduce(192), // /, reduce: MuldivTerm
			reduce(192), // //, reduce: MuldivTerm
			reduce(192), // %, reduce: MuldivTerm
			reduce(192), // .*, reduce: MuldivTerm
			reduce(192), // ./, reduce: MuldivTerm
			reduce(192), // .//, reduce: MuldivTerm
			shift(1147), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2378
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(193), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(193), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(193), // >, reduce: MuldivTerm
			reduce(193), // >>, reduce: MuldivTerm
			reduce(193), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(193), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(193), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(193), // ||, reduce: MuldivTerm
			reduce(193), // ^^, reduce: MuldivTerm
			reduce(193), // &&, reduce: MuldivTerm
			reduce(193), // ??, reduce: MuldivTerm
			reduce(193), // ???, reduce: MuldivTerm
			reduce(193), // =~, reduce: MuldivTerm
			reduce(193), // !=~, reduce: MuldivTerm
			reduce(193), // ==, reduce: MuldivTerm
			reduce(193), // !=, reduce: MuldivTerm
			reduce(193), // <=>, reduce: MuldivTerm
			reduce(193), // >=, reduce: MuldivTerm
			reduce(193), // <, reduce: MuldivTerm
			reduce(193), // <=, reduce: MuldivTerm
			reduce(193), // ^, reduce: MuldivTerm
			reduce(193), // &, reduce: MuldivTerm
			reduce(193), // <<, reduce: MuldivTerm
			reduce(193), // >>>, reduce: MuldivTerm
			reduce(193), // +, reduce: MuldivTerm
			reduce(193), // -, reduce: MuldivTerm
			reduce(193), // .+, reduce: MuldivTerm
			reduce(193), // .-, reduce: MuldivTerm
			reduce(193), // *, reduce: MuldivTerm
			reduce(193), // /, reduce: MuldivTerm
			reduce(193), // //, reduce: MuldivTerm
			reduce(193), // %, reduce: MuldivTerm
			reduce(193), // .*, reduce: MuldivTerm
			reduce(193), // ./, reduce: MuldivTerm
			reduce(193), // .//, reduce: MuldivTerm
			shift(1147), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2379
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(195), // $, reduce: DotTerm
			nil,         // empty
			reduce(195), // ;, reduce: DotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(195), // >, reduce: DotTerm
			reduce(195), // >>, reduce: DotTerm
			reduce(195), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(195), // ,, reduce: DotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(195), // ?, reduce: DotTerm
			nil,         // :
			reduce(195), // ||, reduce: DotTerm
			reduce(195), // ^^, reduce: DotTerm
			reduce(195), // &&, reduce: DotTerm
			reduce(195), // ??, reduce: DotTerm
			reduce(195), // ???, reduce: DotTerm
			reduce(195), // =~, reduce: DotTerm
			reduce(195), // !=~, reduce: DotTerm
			reduce(195), // ==, reduce: DotTerm
			reduce(195), // !=, reduce: DotTerm
			reduce(195), // <=>, reduce: DotTerm
			reduce(195), // >=, reduce: DotTerm
			reduce(195), // <, reduce: DotTerm
			reduce(195), // <=, reduce: DotTerm
			reduce(195), // ^, reduce: DotTerm
			reduce(195), // &, reduce: DotTerm
			reduce(195), // <<, reduce: DotTerm
			reduce(195), // >>>, reduce: DotTerm
			reduce(195), // +, reduce: DotTerm
			reduce(195), // -, reduce: DotTerm
			reduce(195), // .+, reduce: DotTerm
			reduce(195), // .-, reduce: DotTerm
			reduce(195), // *, reduce: DotTerm
			reduce(195), // /, reduce: DotTerm
			reduce(195), // //, reduce: DotTerm
			reduce(195), // %, reduce: DotTerm
			reduce(195), // .*, reduce: DotTerm
			reduce(195), // ./, reduce: DotTerm
			reduce(195), // .//, reduce: DotTerm
			reduce(195), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2380
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2381
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2382
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(204), // $, reduce: PowTerm
			nil,         // empty
			reduce(204), // ;, reduce: PowTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(204), // >, reduce: PowTerm
			reduce(204), // >>, reduce: PowTerm
			reduce(204), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(204), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(204), // ?, reduce: PowTerm
			nil,         // :
			reduce(204), // ||, reduce: PowTerm
			reduce(204), // ^^, reduce: PowTerm
			reduce(204), // &&, reduce: PowTerm
			reduce(204), // ??, reduce: PowTerm
			reduce(204), // ???, reduce: PowTerm
			reduce(204), // =~, reduce: PowTerm
			reduce(204), // !=~, reduce: PowTerm
			reduce(204), // ==, reduce: PowTerm
			reduce(204), // !=, reduce: PowTerm
			reduce(204), // <=>, reduce: PowTerm
			reduce(204), // >=, reduce: PowTerm
			reduce(204), // <, reduce: PowTerm
			reduce(204), // <=, reduce: PowTerm
			reduce(204), // ^, reduce: PowTerm
			reduce(204), // &, reduce: PowTerm
			reduce(204), // <<, reduce: PowTerm
			reduce(204), // >>>, reduce: PowTerm
			reduce(204), // +, reduce: PowTerm
			reduce(204), // -, reduce: PowTerm
			reduce(204), // .+, reduce: PowTerm
			reduce(204), // .-, reduce: PowTerm
			reduce(204), // *, reduce: PowTerm
			reduce(204), // /, reduce: PowTerm
			reduce(204), // //, reduce: PowTerm
			reduce(204), // %, reduce: PowTerm
			reduce(204), // .*, reduce: PowTerm
			reduce(204), // ./, reduce: PowTerm
			reduce(204), // .//, reduce: PowTerm
			reduce(204), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2383
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3203), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3204), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2384
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(3206), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2385
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3207), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2386
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3208), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2387
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(229), // $, reduce: ArrayLiteral
			nil,         // empty
			reduce(229), // ;, reduce: ArrayLiteral
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(229), // >, reduce: ArrayLiteral
			reduce(229), // >>, reduce: ArrayLiteral
			reduce(229), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(229), // ,, reduce: ArrayLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(229), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(229), // ||, reduce: ArrayLiteral
			reduce(229), // ^^, reduce: ArrayLiteral
			reduce(229), // &&, reduce: ArrayLiteral
			reduce(229), // ??, reduce: ArrayLiteral
			reduce(229), // ???, reduce: ArrayLiteral
			reduce(229), // =~, reduce: ArrayLiteral
			reduce(229), // !=~, reduce: ArrayLiteral
			reduce(229), // ==, reduce: ArrayLiteral
			reduce(229), // !=, reduce: ArrayLiteral
			reduce(229), // <=>, reduce: ArrayLiteral
			reduce(229), // >=, reduce: ArrayLiteral
			reduce(229), // <, reduce: ArrayLiteral
			reduce(229), // <=, reduce: ArrayLiteral
			reduce(229), // ^, reduce: ArrayLiteral
			reduce(229), // &, reduce: ArrayLiteral
			reduce(229), // <<, reduce: ArrayLiteral
			reduce(229), // >>>, reduce: ArrayLiteral
			reduce(229), // +, reduce: ArrayLiteral
			reduce(229), // -, reduce: ArrayLiteral
			reduce(229), // .+, reduce: ArrayLiteral
			reduce(229), // .-, reduce: ArrayLiteral
			reduce(229), // *, reduce: ArrayLiteral
			reduce(229), // /, reduce: ArrayLiteral
			reduce(229), // //, reduce: ArrayLiteral
			reduce(229), // %, reduce: ArrayLiteral
			reduce(229), // .*, reduce: ArrayLiteral
			reduce(229), // ./, reduce: ArrayLiteral
			reduce(229), // .//, reduce: ArrayLiteral
			reduce(229), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(229), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(229), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(229), // [[, reduce: ArrayLiteral
			reduce(229), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2388
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(255), // $, reduce: ENV
			nil,         // empty
			reduce(255), // ;, reduce: ENV
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(255), // >, reduce: ENV
			reduce(255), // >>, reduce: ENV
			reduce(255), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(255), // ,, reduce: ENV
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(255), // ?, reduce: ENV
			nil,         // :
			reduce(255), // ||, reduce: ENV
			reduce(255), // ^^, reduce: ENV
			reduce(255), // &&, reduce: ENV
			reduce(255), // ??, reduce: ENV
			reduce(255), // ???, reduce: ENV
			reduce(255), // =~, reduce: ENV
			reduce(255), // !=~, reduce: ENV
			reduce(255), // ==, reduce: ENV
			reduce(255), // !=, reduce: ENV
			reduce(255), // <=>, reduce: ENV
			reduce(255), // >=, reduce: ENV
			reduce(255), // <, reduce: ENV
			reduce(255), // <=, reduce: ENV
			reduce(255), // ^, reduce: ENV
			reduce(255), // &, reduce: ENV
			reduce(255), // <<, reduce: ENV
			reduce(255), // >>>, reduce: ENV
			reduce(255), // +, reduce: ENV
			reduce(255), // -, reduce: ENV
			reduce(255), // .+, reduce: ENV
			reduce(255), // .-, reduce: ENV
			reduce(255), // *, reduce: ENV
			reduce(255), // /, reduce: ENV
			reduce(255), // //, reduce: ENV
			reduce(255), // %, reduce: ENV
			reduce(255), // .*, reduce: ENV
			reduce(255), // ./, reduce: ENV
			reduce(255), // .//, reduce: ENV
			reduce(255), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(255), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(255), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(255), // [[, reduce: ENV
			reduce(255), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2389
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3209), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2390
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3210), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2391
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(268), // $, reduce: FunctionCallsite
			nil,         // empty
			reduce(268), // ;, reduce: FunctionCallsite
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(268), // >, reduce: FunctionCallsite
			reduce(268), // >>, reduce: FunctionCallsite
			reduce(268), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(268), // ,, reduce: FunctionCallsite
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(268), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(268), // ||, reduce: FunctionCallsite
			reduce(268), // ^^, reduce: FunctionCallsite
			reduce(268), // &&, reduce: FunctionCallsite
			reduce(268), // ??, reduce: FunctionCallsite
			reduce(268), // ???, reduce: FunctionCallsite
			reduce(268), // =~, reduce: FunctionCallsite
			reduce(268), // !=~, reduce: FunctionCallsite
			reduce(268), // ==, reduce: FunctionCallsite
			reduce(268), // !=, reduce: FunctionCallsite
			reduce(268), // <=>, reduce: FunctionCallsite
			reduce(268), // >=, reduce: FunctionCallsite
			reduce(268), // <, reduce: FunctionCallsite
			reduce(268), // <=, reduce: FunctionCallsite
			reduce(268), // ^, reduce: FunctionCallsite
			reduce(268), // &, reduce: FunctionCallsite
			reduce(268), // <<, reduce: FunctionCallsite
			reduce(268), // >>>, reduce: FunctionCallsite
			reduce(268), // +, reduce: FunctionCallsite
			reduce(268), // -, reduce: FunctionCallsite
			reduce(268), // .+, reduce: FunctionCallsite
			reduce(268), // .-, reduce: FunctionCallsite
			reduce(268), // *, reduce: FunctionCallsite
			reduce(268), // /, reduce: FunctionCallsite
			reduce(268), // //, reduce: FunctionCallsite
			reduce(268), // %, reduce: FunctionCallsite
			reduce(268), // .*, reduce: FunctionCallsite
			reduce(268), // ./, reduce: FunctionCallsite
			reduce(268), // .//, reduce: FunctionCallsite
			reduce(268), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(268), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(268), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(268), // [[, reduce: FunctionCallsite
			reduce(268), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2392
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3211), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2393
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(235), // $, reduce: MapLiteral
			nil,         // empty
			reduce(235), // ;, reduce: MapLiteral
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(235), // >, reduce: MapLiteral
			reduce(235), // >>, reduce: MapLiteral
			reduce(235), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(235), // ?, reduce: MapLiteral
			nil,         // :
			reduce(235), // ||, reduce: MapLiteral
			reduce(235), // ^^, reduce: MapLiteral
			reduce(235), // &&, reduce: MapLiteral
			reduce(235), // ??, reduce: MapLiteral
			reduce(235), // ???, reduce: MapLiteral
			reduce(235), // =~, reduce: MapLiteral
			reduce(235), // !=~, reduce: MapLiteral
			reduce(235), // ==, reduce: MapLiteral
			reduce(235), // !=, reduce: MapLiteral
			reduce(235), // <=>, reduce: MapLiteral
			reduce(235), // >=, reduce: MapLiteral
			reduce(235), // <, reduce: MapLiteral
			reduce(235), // <=, reduce: MapLiteral
			reduce(235), // ^, reduce: MapLiteral
			reduce(235), // &, reduce: MapLiteral
			reduce(235), // <<, reduce: MapLiteral
			reduce(235), // >>>, reduce: MapLiteral
			reduce(235), // +, reduce: MapLiteral
			reduce(235), // -, reduce: MapLiteral
			reduce(235), // .+, reduce: MapLiteral
			reduce(235), // .-, reduce: MapLiteral
			reduce(235), // *, reduce: MapLiteral
			reduce(235), // /, reduce: MapLiteral
			reduce(235), // //, reduce: MapLiteral
			reduce(235), // %, reduce: MapLiteral
			reduce(235), // .*, reduce: MapLiteral
			reduce(235), // ./, reduce: MapLiteral
			reduce(235), // .//, reduce: MapLiteral
			reduce(235), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(235), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(235), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(235), // [[, reduce: MapLiteral
			reduce(235), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2394
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(208), // $, reduce: PrecedenceChainEnd
			nil,         // empty
			reduce(208), // ;, reduce: PrecedenceChainEnd
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(208), // >, reduce: PrecedenceChainEnd
			reduce(208), // >>, reduce: PrecedenceChainEnd
			reduce(208), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(208), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(208), // ||, reduce: PrecedenceChainEnd
			reduce(208), // ^^, reduce: PrecedenceChainEnd
			reduce(208), // &&, reduce: PrecedenceChainEnd
			reduce(208), // ??, reduce: PrecedenceChainEnd
			reduce(208), // ???, reduce: PrecedenceChainEnd
			reduce(208), // =~, reduce: PrecedenceChainEnd
			reduce(208), // !=~, reduce: PrecedenceChainEnd
			reduce(208), // ==, reduce: PrecedenceChainEnd
			reduce(208), // !=, reduce: PrecedenceChainEnd
			reduce(208), // <=>, reduce: PrecedenceChainEnd
			reduce(208), // >=, reduce: PrecedenceChainEnd
			reduce(208), // <, reduce: PrecedenceChainEnd
			reduce(208), // <=, reduce: PrecedenceChainEnd
			reduce(208), // ^, reduce: PrecedenceChainEnd
			reduce(208), // &, reduce: PrecedenceChainEnd
			reduce(208), // <<, reduce: PrecedenceChainEnd
			reduce(208), // >>>, reduce: PrecedenceChainEnd
			reduce(208), // +, reduce: PrecedenceChainEnd
			reduce(208), // -, reduce: PrecedenceChainEnd
			reduce(208), // .+, reduce: PrecedenceChainEnd
			reduce(208), // .-, reduce: PrecedenceChainEnd
			reduce(208), // *, reduce: PrecedenceChainEnd
			reduce(208), // /, reduce: PrecedenceChainEnd
			reduce(208), // //, reduce: PrecedenceChainEnd
			reduce(208), // %, reduce: PrecedenceChainEnd
			reduce(208), // .*, reduce: PrecedenceChainEnd
			reduce(208), // ./, reduce: PrecedenceChainEnd
			reduce(208), // .//, reduce: PrecedenceChainEnd
			reduce(208), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(208), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2395
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(105), // $, reduce: IndirectFieldValue
			nil,         // empty
			reduce(105), // ;, reduce: IndirectFieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(105), // >, reduce: IndirectFieldValue
			reduce(105), // >>, reduce: IndirectFieldValue
			reduce(105), // |, reduce: IndirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(105), // ?, reduce: IndirectFieldValue
			nil,         // :
			reduce(105), // ||, reduce: IndirectFieldValue
			reduce(105), // ^^, reduce: IndirectFieldValue
			reduce(105), // &&, reduce: IndirectFieldValue
			reduce(105), // ??, reduce: IndirectFieldValue
			reduce(105), // ???, reduce: IndirectFieldValue
			reduce(105), // =~, reduce: IndirectFieldValue
			reduce(105), // !=~, reduce: IndirectFieldValue
			reduce(105), // ==, reduce: IndirectFieldValue
			reduce(105), // !=, reduce: IndirectFieldValue
			reduce(105), // <=>, reduce: IndirectFieldValue
			reduce(105), // >=, reduce: IndirectFieldValue
			reduce(105), // <, reduce: IndirectFieldValue
			reduce(105), // <=, reduce: IndirectFieldValue
			reduce(105), // ^, reduce: IndirectFieldValue
			reduce(105), // &, reduce: IndirectFieldValue
			reduce(105), // <<, reduce: IndirectFieldValue
			reduce(105), // >>>, reduce: IndirectFieldValue
			reduce(105), // +, reduce: IndirectFieldValue
			reduce(105), // -, reduce: IndirectFieldValue
			reduce(105), // .+, reduce: IndirectFieldValue
			reduce(105), // .-, reduce: IndirectFieldValue
			reduce(105), // *, reduce: IndirectFieldValue
			reduce(105), // /, reduce: IndirectFieldValue
			reduce(105), // //, reduce: IndirectFieldValue
			reduce(105), // %, reduce: IndirectFieldValue
			reduce(105), // .*, reduce: IndirectFieldValue
			reduce(105), // ./, reduce: IndirectFieldValue
			reduce(105), // .//, reduce: IndirectFieldValue
			reduce(105), // ., reduce: IndirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(105), // **, reduce: IndirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(105), // [, reduce: IndirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(105), // [[, reduce: IndirectFieldValue
			reduce(105), // [[[, reduce: IndirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2396
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3212), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2397
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3213), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2398
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(114), // $, reduce: IndirectOosvarValue
			nil,         // empty
			reduce(114), // ;, reduce: IndirectOosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(114), // >, reduce: IndirectOosvarValue
			reduce(114), // >>, reduce: IndirectOosvarValue
			reduce(114), // |, reduce: IndirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(114), // ?, reduce: IndirectOosvarValue
			nil,         // :
			reduce(114), // ||, reduce: IndirectOosvarValue
			reduce(114), // ^^, reduce: IndirectOosvarValue
			reduce(114), // &&, reduce: IndirectOosvarValue
			reduce(114), // ??, reduce: IndirectOosvarValue
			reduce(114), // ???, reduce: IndirectOosvarValue
			reduce(114), // =~, reduce: IndirectOosvarValue
			reduce(114), // !=~, reduce: IndirectOosvarValue
			reduce(114), // ==, reduce: IndirectOosvarValue
			reduce(114), // !=, reduce: IndirectOosvarValue
			reduce(114), // <=>, reduce: IndirectOosvarValue
			reduce(114), // >=, reduce: IndirectOosvarValue
			reduce(114), // <, reduce: IndirectOosvarValue
			reduce(114), // <=, reduce: IndirectOosvarValue
			reduce(114), // ^, reduce: IndirectOosvarValue
			reduce(114), // &, reduce: IndirectOosvarValue
			reduce(114), // <<, reduce: IndirectOosvarValue
			reduce(114), // >>>, reduce: IndirectOosvarValue
			reduce(114), // +, reduce: IndirectOosvarValue
			reduce(114), // -, reduce: IndirectOosvarValue
			reduce(114), // .+, reduce: IndirectOosvarValue
			reduce(114), // .-, reduce: IndirectOosvarValue
			reduce(114), // *, reduce: IndirectOosvarValue
			reduce(114), // /, reduce: IndirectOosvarValue
			reduce(114), // //, reduce: IndirectOosvarValue
			reduce(114), // %, reduce: IndirectOosvarValue
			reduce(114), // .*, reduce: IndirectOosvarValue
			reduce(114), // ./, reduce: IndirectOosvarValue
			reduce(114), // .//, reduce: IndirectOosvarValue
			reduce(114), // ., reduce: IndirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(114), // **, reduce: IndirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(114), // [, reduce: IndirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(114), // [[, reduce: IndirectOosvarValue
			reduce(114), // [[[, reduce: IndirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2399
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3214), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2400
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(151), // $, reduce: LogicalOrTerm
			nil,         // empty
			reduce(151), // ;, reduce: LogicalOrTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(151), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(151), // ||, reduce: LogicalOrTerm
			shift(1169), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2401
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(153), // $, reduce: LogicalXORTerm
			nil,         // empty
			reduce(153), // ;, reduce: LogicalXORTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(153), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(153), // ||, reduce: LogicalXORTerm
			reduce(153), // ^^, reduce: LogicalXORTerm
			shift(1170), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2402
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(155), // $, reduce: LogicalAndTerm
			nil,         // empty
			reduce(155), // ;, reduce: LogicalAndTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(155), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(155), // ||, reduce: LogicalAndTerm
			reduce(155), // ^^, reduce: LogicalAndTerm
			reduce(155), // &&, reduce: LogicalAndTerm
			shift(1171), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2403
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(157), // $, reduce: AbsentCoalesceTerm
			nil,         // empty
			reduce(157), // ;, reduce: AbsentCoalesceTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(157), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(157), // ||, reduce: AbsentCoalesceTerm
			reduce(157), // ^^, reduce: AbsentCoalesceTerm
			reduce(157), // &&, reduce: AbsentCoalesceTerm
			reduce(157), // ??, reduce: AbsentCoalesceTerm
			shift(1172), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2404
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(159), // $, reduce: EmptyCoalesceTerm
			nil,         // empty
			reduce(159), // ;, reduce: EmptyCoalesceTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(159), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(159), // ||, reduce: EmptyCoalesceTerm
			reduce(159), // ^^, reduce: EmptyCoalesceTerm
			reduce(159), // &&, reduce: EmptyCoalesceTerm
			reduce(159), // ??, reduce: EmptyCoalesceTerm
			reduce(159), // ???, reduce: EmptyCoalesceTerm
			shift(1173), // =~
			shift(1174), // !=~
			shift(1175), // ==
			shift(1176), // !=
			shift(1177), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2405
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(161), // $, reduce: EqneTerm
			nil,         // empty
			reduce(161), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1178), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(161), // ?, reduce: EqneTerm
			nil,         // :
			reduce(161), // ||, reduce: EqneTerm
			reduce(161), // ^^, reduce: EqneTerm
			reduce(161), // &&, reduce: EqneTerm
			reduce(161), // ??, reduce: EqneTerm
			reduce(161), // ???, reduce: EqneTerm
			reduce(161), // =~, reduce: EqneTerm
			reduce(161), // !=~, reduce: EqneTerm
			reduce(161), // ==, reduce: EqneTerm
			reduce(161), // !=, reduce: EqneTerm
			reduce(161), // <=>, reduce: EqneTerm
			shift(1179), // >=
			shift(1180), // <
			shift(1181), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2406
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(162), // $, reduce: EqneTerm
			nil,         // empty
			reduce(162), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1178), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(162), // ?, reduce: EqneTerm
			nil,         // :
			reduce(162), // ||, reduce: EqneTerm
			reduce(162), // ^^, reduce: EqneTerm
			reduce(162), // &&, reduce: EqneTerm
			reduce(162), // ??, reduce: EqneTerm
			reduce(162), // ???, reduce: EqneTerm
			reduce(162), // =~, reduce: EqneTerm
			reduce(162), // !=~, reduce: EqneTerm
			reduce(162), // ==, reduce: EqneTerm
			reduce(162), // !=, reduce: EqneTerm
			reduce(162), // <=>, reduce: EqneTerm
			shift(1179), // >=
			shift(1180), // <
			shift(1181), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2407
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(163), // $, reduce: EqneTerm
			nil,         // empty
			reduce(163), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1178), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(163), // ?, reduce: EqneTerm
			nil,         // :
			reduce(163), // ||, reduce: EqneTerm
			reduce(163), // ^^, reduce: EqneTerm
			reduce(163), // &&, reduce: EqneTerm
			reduce(163), // ??, reduce: EqneTerm
			reduce(163), // ???, reduce: EqneTerm
			reduce(163), // =~, reduce: EqneTerm
			reduce(163), // !=~, reduce: EqneTerm
			reduce(163), // ==, reduce: EqneTerm
			reduce(163), // !=, reduce: EqneTerm
			reduce(163), // <=>, reduce: EqneTerm
			shift(1179), // >=
			shift(1180), // <
			shift(1181), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2408
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(164), // $, reduce: EqneTerm
			nil,         // empty
			reduce(164), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1178), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(164), // ?, reduce: EqneTerm
			nil,         // :
			reduce(164), // ||, reduce: EqneTerm
			reduce(164), // ^^, reduce: EqneTerm
			reduce(164), // &&, reduce: EqneTerm
			reduce(164), // ??, reduce: EqneTerm
			reduce(164), // ???, reduce: EqneTerm
			reduce(164), // =~, reduce: EqneTerm
			reduce(164), // !=~, reduce: EqneTerm
			reduce(164), // ==, reduce: EqneTerm
			reduce(164), // !=, reduce: EqneTerm
			reduce(164), // <=>, reduce: EqneTerm
			shift(1179), // >=
			shift(1180), // <
			shift(1181), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2409
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(165), // $, reduce: EqneTerm
			nil,         // empty
			reduce(165), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1178), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(165), // ?, reduce: EqneTerm
			nil,         // :
			reduce(165), // ||, reduce: EqneTerm
			reduce(165), // ^^, reduce: EqneTerm
			reduce(165), // &&, reduce: EqneTerm
			reduce(165), // ??, reduce: EqneTerm
			reduce(165), // ???, reduce: EqneTerm
			reduce(165), // =~, reduce: EqneTerm
			reduce(165), // !=~, reduce: EqneTerm
			reduce(165), // ==, reduce: EqneTerm
			reduce(165), // !=, reduce: EqneTerm
			reduce(165), // <=>, reduce: EqneTerm
			shift(1179), // >=
			shift(1180), // <
			shift(1181), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2410
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(167), // $, reduce: CmpTerm
			nil,         // empty
			reduce(167), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(167), // >, reduce: CmpTerm
			nil,         // >>
			shift(1182), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(167), // ?, reduce: CmpTerm
			nil,         // :
			reduce(167), // ||, reduce: CmpTerm
			reduce(167), // ^^, reduce: CmpTerm
			reduce(167), // &&, reduce: CmpTerm
			reduce(167), // ??, reduce: CmpTerm
			reduce(167), // ???, reduce: CmpTerm
			reduce(167), // =~, reduce: CmpTerm
			reduce(167), // !=~, reduce: CmpTerm
			reduce(167), // ==, reduce: CmpTerm
			reduce(167), // !=, reduce: CmpTerm
			reduce(167), // <=>, reduce: CmpTerm
			reduce(167), // >=, reduce: CmpTerm
			reduce(167), // <, reduce: CmpTerm
			reduce(167), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2411
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(168), // $, reduce: CmpTerm
			nil,         // empty
			reduce(168), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(168), // >, reduce: CmpTerm
			nil,         // >>
			shift(1182), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(168), // ?, reduce: CmpTerm
			nil,         // :
			reduce(168), // ||, reduce: CmpTerm
			reduce(168), // ^^, reduce: CmpTerm
			reduce(168), // &&, reduce: CmpTerm
			reduce(168), // ??, reduce: CmpTerm
			reduce(168), // ???, reduce: CmpTerm
			reduce(168), // =~, reduce: CmpTerm
			reduce(168), // !=~, reduce: CmpTerm
			reduce(168), // ==, reduce: CmpTerm
			reduce(168), // !=, reduce: CmpTerm
			reduce(168), // <=>, reduce: CmpTerm
			reduce(168), // >=, reduce: CmpTerm
			reduce(168), // <, reduce: CmpTerm
			reduce(168), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2412
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(169), // $, reduce: CmpTerm
			nil,         // empty
			reduce(169), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(169), // >, reduce: CmpTerm
			nil,         // >>
			shift(1182), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(169), // ?, reduce: CmpTerm
			nil,         // :
			reduce(169), // ||, reduce: CmpTerm
			reduce(169), // ^^, reduce: CmpTerm
			reduce(169), // &&, reduce: CmpTerm
			reduce(169), // ??, reduce: CmpTerm
			reduce(169), // ???, reduce: CmpTerm
			reduce(169), // =~, reduce: CmpTerm
			reduce(169), // !=~, reduce: CmpTerm
			reduce(169), // ==, reduce: CmpTerm
			reduce(169), // !=, reduce: CmpTerm
			reduce(169), // <=>, reduce: CmpTerm
			reduce(169), // >=, reduce: CmpTerm
			reduce(169), // <, reduce: CmpTerm
			reduce(169), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2413
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(170), // $, reduce: CmpTerm
			nil,         // empty
			reduce(170), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(170), // >, reduce: CmpTerm
			nil,         // >>
			shift(1182), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(170), // ?, reduce: CmpTerm
			nil,         // :
			reduce(170), // ||, reduce: CmpTerm
			reduce(170), // ^^, reduce: CmpTerm
			reduce(170), // &&, reduce: CmpTerm
			reduce(170), // ??, reduce: CmpTerm
			reduce(170), // ???, reduce: CmpTerm
			reduce(170), // =~, reduce: CmpTerm
			reduce(170), // !=~, reduce: CmpTerm
			reduce(170), // ==, reduce: CmpTerm
			reduce(170), // !=, reduce: CmpTerm
			reduce(170), // <=>, reduce: CmpTerm
			reduce(170), // >=, reduce: CmpTerm
			reduce(170), // <, reduce: CmpTerm
			reduce(170), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2414
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(172), // $, reduce: BitwiseORTerm
			nil,         // empty
			reduce(172), // ;, reduce: BitwiseORTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(172), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(172), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(172), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(172), // ||, reduce: BitwiseORTerm
			reduce(172), // ^^, reduce: BitwiseORTerm
			reduce(172), // &&, reduce: BitwiseORTerm
			reduce(172), // ??, reduce: BitwiseORTerm
			reduce(172), // ???, reduce: BitwiseORTerm
			reduce(172), // =~, reduce: BitwiseORTerm
			reduce(172), // !=~, reduce: BitwiseORTerm
			reduce(172), // ==, reduce: BitwiseORTerm
			reduce(172), // !=, reduce: BitwiseORTerm
			reduce(172), // <=>, reduce: BitwiseORTerm
			reduce(172), // >=, reduce: BitwiseORTerm
			reduce(172), // <, reduce: BitwiseORTerm
			reduce(172), // <=, reduce: BitwiseORTerm
			shift(1183), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2415
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(174), // $, reduce: BitwiseXORTerm
			nil,         // empty
			reduce(174), // ;, reduce: BitwiseXORTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(174), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(174), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(174), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(174), // ||, reduce: BitwiseXORTerm
			reduce(174), // ^^, reduce: BitwiseXORTerm
			reduce(174), // &&, reduce: BitwiseXORTerm
			reduce(174), // ??, reduce: BitwiseXORTerm
			reduce(174), // ???, reduce: BitwiseXORTerm
			reduce(174), // =~, reduce: BitwiseXORTerm
			reduce(174), // !=~, reduce: BitwiseXORTerm
			reduce(174), // ==, reduce: BitwiseXORTerm
			reduce(174), // !=, reduce: BitwiseXORTerm
			reduce(174), // <=>, reduce: BitwiseXORTerm
			reduce(174), // >=, reduce: BitwiseXORTerm
			reduce(174), // <, reduce: BitwiseXORTerm
			reduce(174), // <=, reduce: BitwiseXORTerm
			reduce(174), // ^, reduce: BitwiseXORTerm
			shift(1184), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2416
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(176), // $, reduce: BitwiseANDTerm
			nil,         // empty
			reduce(176), // ;, reduce: BitwiseANDTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(176), // >, reduce: BitwiseANDTerm
			shift(1185), // >>
			reduce(176), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(176), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(176), // ||, reduce: BitwiseANDTerm
			reduce(176), // ^^, reduce: BitwiseANDTerm
			reduce(176), // &&, reduce: BitwiseANDTerm
			reduce(176), // ??, reduce: BitwiseANDTerm
			reduce(176), // ???, reduce: BitwiseANDTerm
			reduce(176), // =~, reduce: BitwiseANDTerm
			reduce(176), // !=~, reduce: BitwiseANDTerm
			reduce(176), // ==, reduce: BitwiseANDTerm
			reduce(176), // !=, reduce: BitwiseANDTerm
			reduce(176), // <=>, reduce: BitwiseANDTerm
			reduce(176), // >=, reduce: BitwiseANDTerm
			reduce(176), // <, reduce: BitwiseANDTerm
			reduce(176), // <=, reduce: BitwiseANDTerm
			reduce(176), // ^, reduce: BitwiseANDTerm
			reduce(176), // &, reduce: BitwiseANDTerm
			shift(1186), // <<
			shift(1187), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2417
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(179), // $, reduce: BitwiseShiftTerm
			nil,         // empty
			reduce(179), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(179), // >, reduce: BitwiseShiftTerm
			reduce(179), // >>, reduce: BitwiseShiftTerm
			reduce(179), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(179), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(179), // ||, reduce: BitwiseShiftTerm
			reduce(179), // ^^, reduce: BitwiseShiftTerm
			reduce(179), // &&, reduce: BitwiseShiftTerm
			reduce(179), // ??, reduce: BitwiseShiftTerm
			reduce(179), // ???, reduce: BitwiseShiftTerm
			reduce(179), // =~, reduce: BitwiseShiftTerm
			reduce(179), // !=~, reduce: BitwiseShiftTerm
			reduce(179), // ==, reduce: BitwiseShiftTerm
			reduce(179), // !=, reduce: BitwiseShiftTerm
			reduce(179), // <=>, reduce: BitwiseShiftTerm
			reduce(179), // >=, reduce: BitwiseShiftTerm
			reduce(179), // <, reduce: BitwiseShiftTerm
			reduce(179), // <=, reduce: BitwiseShiftTerm
			reduce(179), // ^, reduce: BitwiseShiftTerm
			reduce(179), // &, reduce: BitwiseShiftTerm
			reduce(179), // <<, reduce: BitwiseShiftTerm
			reduce(179), // >>>, reduce: BitwiseShiftTerm
			shift(1188), // +
			shift(1189), // -
			shift(1190), // .+
			shift(1191), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2418
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(178), // $, reduce: BitwiseShiftTerm
			nil,         // empty
			reduce(178), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(178), // >, reduce: BitwiseShiftTerm
			reduce(178), // >>, reduce: BitwiseShiftTerm
			reduce(178), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(178), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(178), // ||, reduce: BitwiseShiftTerm
			reduce(178), // ^^, reduce: BitwiseShiftTerm
			reduce(178), // &&, reduce: BitwiseShiftTerm
			reduce(178), // ??, reduce: BitwiseShiftTerm
			reduce(178), // ???, reduce: BitwiseShiftTerm
			reduce(178), // =~, reduce: BitwiseShiftTerm
			reduce(178), // !=~, reduce: BitwiseShiftTerm
			reduce(178), // ==, reduce: BitwiseShiftTerm
			reduce(178), // !=, reduce: BitwiseShiftTerm
			reduce(178), // <=>, reduce: BitwiseShiftTerm
			reduce(178), // >=, reduce: BitwiseShiftTerm
			reduce(178), // <, reduce: BitwiseShiftTerm
			reduce(178), // <=, reduce: BitwiseShiftTerm
			reduce(178), // ^, reduce: BitwiseShiftTerm
			reduce(178), // &, reduce: BitwiseShiftTerm
			reduce(178), // <<, reduce: BitwiseShiftTerm
			reduce(178), // >>>, reduce: BitwiseShiftTerm
			shift(1188), // +
			shift(1189), // -
			shift(1190), // .+
			shift(1191), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2419
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(180), // $, reduce: BitwiseShiftTerm
			nil,         // empty
			reduce(180), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(180), // >, reduce: BitwiseShiftTerm
			reduce(180), // >>, reduce: BitwiseShiftTerm
			reduce(180), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(180), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(180), // ||, reduce: BitwiseShiftTerm
			reduce(180), // ^^, reduce: BitwiseShiftTerm
			reduce(180), // &&, reduce: BitwiseShiftTerm
			reduce(180), // ??, reduce: BitwiseShiftTerm
			reduce(180), // ???, reduce: BitwiseShiftTerm
			reduce(180), // =~, reduce: BitwiseShiftTerm
			reduce(180), // !=~, reduce: BitwiseShiftTerm
			reduce(180), // ==, reduce: BitwiseShiftTerm
			reduce(180), // !=, reduce: BitwiseShiftTerm
			reduce(180), // <=>, reduce: BitwiseShiftTerm
			reduce(180), // >=, reduce: BitwiseShiftTerm
			reduce(180), // <, reduce: BitwiseShiftTerm
			reduce(180), // <=, reduce: BitwiseShiftTerm
			reduce(180), // ^, reduce: BitwiseShiftTerm
			reduce(180), // &, reduce: BitwiseShiftTerm
			reduce(180), // <<, reduce: BitwiseShiftTerm
			reduce(180), // >>>, reduce: BitwiseShiftTerm
			shift(1188), // +
			shift(1189), // -
			shift(1190), // .+
			shift(1191), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2420
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(182), // $, reduce: AddsubdotTerm
			nil,         // empty
			reduce(182), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(182), // >, reduce: AddsubdotTerm
			reduce(182), // >>, reduce: AddsubdotTerm
			reduce(182), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(182), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(182), // ||, reduce: AddsubdotTerm
			reduce(182), // ^^, reduce: AddsubdotTerm
			reduce(182), // &&, reduce: AddsubdotTerm
			reduce(182), // ??, reduce: AddsubdotTerm
			reduce(182), // ???, reduce: AddsubdotTerm
			reduce(182), // =~, reduce: AddsubdotTerm
			reduce(182), // !=~, reduce: AddsubdotTerm
			reduce(182), // ==, reduce: AddsubdotTerm
			reduce(182), // !=, reduce: AddsubdotTerm
			reduce(182), // <=>, reduce: AddsubdotTerm
			reduce(182), // >=, reduce: AddsubdotTerm
			reduce(182), // <, reduce: AddsubdotTerm
			reduce(182), // <=, reduce: AddsubdotTerm
			reduce(182), // ^, reduce: AddsubdotTerm
			reduce(182), // &, reduce: AddsubdotTerm
			reduce(182), // <<, reduce: AddsubdotTerm
			reduce(182), // >>>, reduce: AddsubdotTerm
			reduce(182), // +, reduce: AddsubdotTerm
			reduce(182), // -, reduce: AddsubdotTerm
			reduce(182), // .+, reduce: AddsubdotTerm
			reduce(182), // .-, reduce: AddsubdotTerm
			shift(1193), // *
			shift(1194), // /
			shift(1195), // //
			shift(1196), // %
			shift(1197), // .*
			shift(1198), // ./
			shift(1199), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2421
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(183), // $, reduce: AddsubdotTerm
			nil,         // empty
			reduce(183), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(183), // >, reduce: AddsubdotTerm
			reduce(183), // >>, reduce: AddsubdotTerm
			reduce(183), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(183), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(183), // ||, reduce: AddsubdotTerm
			reduce(183), // ^^, reduce: AddsubdotTerm
			reduce(183), // &&, reduce: AddsubdotTerm
			reduce(183), // ??, reduce: AddsubdotTerm
			reduce(183), // ???, reduce: AddsubdotTerm
			reduce(183), // =~, reduce: AddsubdotTerm
			reduce(183), // !=~, reduce: AddsubdotTerm
			reduce(183), // ==, reduce: AddsubdotTerm
			reduce(183), // !=, reduce: AddsubdotTerm
			reduce(183), // <=>, reduce: AddsubdotTerm
			reduce(183), // >=, reduce: AddsubdotTerm
			reduce(183), // <, reduce: AddsubdotTerm
			reduce(183), // <=, reduce: AddsubdotTerm
			reduce(183), // ^, reduce: AddsubdotTerm
			reduce(183), // &, reduce: AddsubdotTerm
			reduce(183), // <<, reduce: AddsubdotTerm
			reduce(183), // >>>, reduce: AddsubdotTerm
			reduce(183), // +, reduce: AddsubdotTerm
			reduce(183), // -, reduce: AddsubdotTerm
			reduce(183), // .+, reduce: AddsubdotTerm
			reduce(183), // .-, reduce: AddsubdotTerm
			shift(1193), // *
			shift(1194), // /
			shift(1195), // //
			shift(1196), // %
			shift(1197), // .*
			shift(1198), // ./
			shift(1199), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2422
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(184), // $, reduce: AddsubdotTerm
			nil,         // empty
			reduce(184), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(184), // >, reduce: AddsubdotTerm
			reduce(184), // >>, reduce: AddsubdotTerm
			reduce(184), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(184), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(184), // ||, reduce: AddsubdotTerm
			reduce(184), // ^^, reduce: AddsubdotTerm
			reduce(184), // &&, reduce: AddsubdotTerm
			reduce(184), // ??, reduce: AddsubdotTerm
			reduce(184), // ???, reduce: AddsubdotTerm
			reduce(184), // =~, reduce: AddsubdotTerm
			reduce(184), // !=~, reduce: AddsubdotTerm
			reduce(184), // ==, reduce: AddsubdotTerm
			reduce(184), // !=, reduce: AddsubdotTerm
			reduce(184), // <=>, reduce: AddsubdotTerm
			reduce(184), // >=, reduce: AddsubdotTerm
			reduce(184), // <, reduce: AddsubdotTerm
			reduce(184), // <=, reduce: AddsubdotTerm
			reduce(184), // ^, reduce: AddsubdotTerm
			reduce(184), // &, reduce: AddsubdotTerm
			reduce(184), // <<, reduce: AddsubdotTerm
			reduce(184), // >>>, reduce: AddsubdotTerm
			reduce(184), // +, reduce: AddsubdotTerm
			reduce(184), // -, reduce: AddsubdotTerm
			reduce(184), // .+, reduce: AddsubdotTerm
			reduce(184), // .-, reduce: AddsubdotTerm
			shift(1193), // *
			shift(1194), // /
			shift(1195), // //
			shift(1196), // %
			shift(1197), // .*
			shift(1198), // ./
			shift(1199), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2423
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(185), // $, reduce: AddsubdotTerm
			nil,         // empty
			reduce(185), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(185), // >, reduce: AddsubdotTerm
			reduce(185), // >>, reduce: AddsubdotTerm
			reduce(185), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(185), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(185), // ||, reduce: AddsubdotTerm
			reduce(185), // ^^, reduce: AddsubdotTerm
			reduce(185), // &&, reduce: AddsubdotTerm
			reduce(185), // ??, reduce: AddsubdotTerm
			reduce(185), // ???, reduce: AddsubdotTerm
			reduce(185), // =~, reduce: AddsubdotTerm
			reduce(185), // !=~, reduce: AddsubdotTerm
			reduce(185), // ==, reduce: AddsubdotTerm
			reduce(185), // !=, reduce: AddsubdotTerm
			reduce(185), // <=>, reduce: AddsubdotTerm
			reduce(185), // >=, reduce: AddsubdotTerm
			reduce(185), // <, reduce: AddsubdotTerm
			reduce(185), // <=, reduce: AddsubdotTerm
			reduce(185), // ^, reduce: AddsubdotTerm
			reduce(185), // &, reduce: AddsubdotTerm
			reduce(185), // <<, reduce: AddsubdotTerm
			reduce(185), // >>>, reduce: AddsubdotTerm
			reduce(185), // +, reduce: AddsubdotTerm
			reduce(185), // -, reduce: AddsubdotTerm
			reduce(185), // .+, reduce: AddsubdotTerm
			reduce(185), // .-, reduce: AddsubdotTerm
			shift(1193), // *
			shift(1194), // /
			shift(1195), // //
			shift(1196), // %
			shift(1197), // .*
			shift(1198), // ./
			shift(1199), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2424
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(187), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(187), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(187), // >, reduce: MuldivTerm
			reduce(187), // >>, reduce: MuldivTerm
			reduce(187), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(187), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(187), // ||, reduce: MuldivTerm
			reduce(187), // ^^, reduce: MuldivTerm
			reduce(187), // &&, reduce: MuldivTerm
			reduce(187), // ??, reduce: MuldivTerm
			reduce(187), // ???, reduce: MuldivTerm
			reduce(187), // =~, reduce: MuldivTerm
			reduce(187), // !=~, reduce: MuldivTerm
			reduce(187), // ==, reduce: MuldivTerm
			reduce(187), // !=, reduce: MuldivTerm
			reduce(187), // <=>, reduce: MuldivTerm
			reduce(187), // >=, reduce: MuldivTerm
			reduce(187), // <, reduce: MuldivTerm
			reduce(187), // <=, reduce: MuldivTerm
			reduce(187), // ^, reduce: MuldivTerm
			reduce(187), // &, reduce: MuldivTerm
			reduce(187), // <<, reduce: MuldivTerm
			reduce(187), // >>>, reduce: MuldivTerm
			reduce(187), // +, reduce: MuldivTerm
			reduce(187), // -, reduce: MuldivTerm
			reduce(187), // .+, reduce: MuldivTerm
			reduce(187), // .-, reduce: MuldivTerm
			reduce(187), // *, reduce: MuldivTerm
			reduce(187), // /, reduce: MuldivTerm
			reduce(187), // //, reduce: MuldivTerm
			reduce(187), // %, reduce: MuldivTerm
			reduce(187), // .*, reduce: MuldivTerm
			reduce(187), // ./, reduce: MuldivTerm
			reduce(187), // .//, reduce: MuldivTerm
			shift(1203), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2425
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(188), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(188), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(188), // >, reduce: MuldivTerm
			reduce(188), // >>, reduce: MuldivTerm
			reduce(188), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(188), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(188), // ||, reduce: MuldivTerm
			reduce(188), // ^^, reduce: MuldivTerm
			reduce(188), // &&, reduce: MuldivTerm
			reduce(188), // ??, reduce: MuldivTerm
			reduce(188), // ???, reduce: MuldivTerm
			reduce(188), // =~, reduce: MuldivTerm
			reduce(188), // !=~, reduce: MuldivTerm
			reduce(188), // ==, reduce: MuldivTerm
			reduce(188), // !=, reduce: MuldivTerm
			reduce(188), // <=>, reduce: MuldivTerm
			reduce(188), // >=, reduce: MuldivTerm
			reduce(188), // <, reduce: MuldivTerm
			reduce(188), // <=, reduce: MuldivTerm
			reduce(188), // ^, reduce: MuldivTerm
			reduce(188), // &, reduce: MuldivTerm
			reduce(188), // <<, reduce: MuldivTerm
			reduce(188), // >>>, reduce: MuldivTerm
			reduce(188), // +, reduce: MuldivTerm
			reduce(188), // -, reduce: MuldivTerm
			reduce(188), // .+, reduce: MuldivTerm
			reduce(188), // .-, reduce: MuldivTerm
			reduce(188), // *, reduce: MuldivTerm
			reduce(188), // /, reduce: MuldivTerm
			reduce(188), // //, reduce: MuldivTerm
			reduce(188), // %, reduce: MuldivTerm
			reduce(188), // .*, reduce: MuldivTerm
			reduce(188), // ./, reduce: MuldivTerm
			reduce(188), // .//, reduce: MuldivTerm
			shift(1203), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2426
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(189), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(189), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(189), // >, reduce: MuldivTerm
			reduce(189), // >>, reduce: MuldivTerm
			reduce(189), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(189), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(189), // ||, reduce: MuldivTerm
			reduce(189), // ^^, reduce: MuldivTerm
			reduce(189), // &&, reduce: MuldivTerm
			reduce(189), // ??, reduce: MuldivTerm
			reduce(189), // ???, reduce: MuldivTerm
			reduce(189), // =~, reduce: MuldivTerm
			reduce(189), // !=~, reduce: MuldivTerm
			reduce(189), // ==, reduce: MuldivTerm
			reduce(189), // !=, reduce: MuldivTerm
			reduce(189), // <=>, reduce: MuldivTerm
			reduce(189), // >=, reduce: MuldivTerm
			reduce(189), // <, reduce: MuldivTerm
			reduce(189), // <=, reduce: MuldivTerm
			reduce(189), // ^, reduce: MuldivTerm
			reduce(189), // &, reduce: MuldivTerm
			reduce(189), // <<, reduce: MuldivTerm
			reduce(189), // >>>, reduce: MuldivTerm
			reduce(189), // +, reduce: MuldivTerm
			reduce(189), // -, reduce: MuldivTerm
			reduce(189), // .+, reduce: MuldivTerm
			reduce(189), // .-, reduce: MuldivTerm
			reduce(189), // *, reduce: MuldivTerm
			reduce(189), // /, reduce: MuldivTerm
			reduce(189), // //, reduce: MuldivTerm
			reduce(189), // %, reduce: MuldivTerm
			reduce(189), // .*, reduce: MuldivTerm
			reduce(189), // ./, reduce: MuldivTerm
			reduce(189), // .//, reduce: MuldivTerm
			shift(1203), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2427
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(190), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(190), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(190), // >, reduce: MuldivTerm
			reduce(190), // >>, reduce: MuldivTerm
			reduce(190), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(190), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(190), // ||, reduce: MuldivTerm
			reduce(190), // ^^, reduce: MuldivTerm
			reduce(190), // &&, reduce: MuldivTerm
			reduce(190), // ??, reduce: MuldivTerm
			reduce(190), // ???, reduce: MuldivTerm
			reduce(190), // =~, reduce: MuldivTerm
			reduce(190), // !=~, reduce: MuldivTerm
			reduce(190), // ==, reduce: MuldivTerm
			reduce(190), // !=, reduce: MuldivTerm
			reduce(190), // <=>, reduce: MuldivTerm
			reduce(190), // >=, reduce: MuldivTerm
			reduce(190), // <, reduce: MuldivTerm
			reduce(190), // <=, reduce: MuldivTerm
			reduce(190), // ^, reduce: MuldivTerm
			reduce(190), // &, reduce: MuldivTerm
			reduce(190), // <<, reduce: MuldivTerm
			reduce(190), // >>>, reduce: MuldivTerm
			reduce(190), // +, reduce: MuldivTerm
			reduce(190), // -, reduce: MuldivTerm
			reduce(190), // .+, reduce: MuldivTerm
			reduce(190), // .-, reduce: MuldivTerm
			reduce(190), // *, reduce: MuldivTerm
			reduce(190), // /, reduce: MuldivTerm
			reduce(190), // //, reduce: MuldivTerm
			reduce(190), // %, reduce: MuldivTerm
			reduce(190), // .*, reduce: MuldivTerm
			reduce(190), // ./, reduce: MuldivTerm
			reduce(190), // .//, reduce: MuldivTerm
			shift(1203), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2428
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(191), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(191), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(191), // >, reduce: MuldivTerm
			reduce(191), // >>, reduce: MuldivTerm
			reduce(191), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(191), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(191), // ||, reduce: MuldivTerm
			reduce(191), // ^^, reduce: MuldivTerm
			reduce(191), // &&, reduce: MuldivTerm
			reduce(191), // ??, reduce: MuldivTerm
			reduce(191), // ???, reduce: MuldivTerm
			reduce(191), // =~, reduce: MuldivTerm
			reduce(191), // !=~, reduce: MuldivTerm
			reduce(191), // ==, reduce: MuldivTerm
			reduce(191), // !=, reduce: MuldivTerm
			reduce(191), // <=>, reduce: MuldivTerm
			reduce(191), // >=, reduce: MuldivTerm
			reduce(191), // <, reduce: MuldivTerm
			reduce(191), // <=, reduce: MuldivTerm
			reduce(191), // ^, reduce: MuldivTerm
			reduce(191), // &, reduce: MuldivTerm
			reduce(191), // <<, reduce: MuldivTerm
			reduce(191), // >>>, reduce: MuldivTerm
			reduce(191), // +, reduce: MuldivTerm
			reduce(191), // -, reduce: MuldivTerm
			reduce(191), // .+, reduce: MuldivTerm
			reduce(191), // .-, reduce: MuldivTerm
			reduce(191), // *, reduce: MuldivTerm
			reduce(191), // /, reduce: MuldivTerm
			reduce(191), // //, reduce: MuldivTerm
			reduce(191), // %, reduce: MuldivTerm
			reduce(191), // .*, reduce: MuldivTerm
			reduce(191), // ./, reduce: MuldivTerm
			reduce(191), // .//, reduce: MuldivTerm
			shift(1203), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2429
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(192), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(192), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(192), // >, reduce: MuldivTerm
			reduce(192), // >>, reduce: MuldivTerm
			reduce(192), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(192), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(192), // ||, reduce: MuldivTerm
			reduce(192), // ^^, reduce: MuldivTerm
			reduce(192), // &&, reduce: MuldivTerm
			reduce(192), // ??, reduce: MuldivTerm
			reduce(192), // ???, reduce: MuldivTerm
			reduce(192), // =~, reduce: MuldivTerm
			reduce(192), // !=~, reduce: MuldivTerm
			reduce(192), // ==, reduce: MuldivTerm
			reduce(192), // !=, reduce: MuldivTerm
			reduce(192), // <=>, reduce: MuldivTerm
			reduce(192), // >=, reduce: MuldivTerm
			reduce(192), // <, reduce: MuldivTerm
			reduce(192), // <=, reduce: MuldivTerm
			reduce(192), // ^, reduce: MuldivTerm
			reduce(192), // &, reduce: MuldivTerm
			reduce(192), // <<, reduce: MuldivTerm
			reduce(192), // >>>, reduce: MuldivTerm
			reduce(192), // +, reduce: MuldivTerm
			reduce(192), // -, reduce: MuldivTerm
			reduce(192), // .+, reduce: MuldivTerm
			reduce(192), // .-, reduce: MuldivTerm
			reduce(192), // *, reduce: MuldivTerm
			reduce(192), // /, reduce: MuldivTerm
			reduce(192), // //, reduce: MuldivTerm
			reduce(192), // %, reduce: MuldivTerm
			reduce(192), // .*, reduce: MuldivTerm
			reduce(192), // ./, reduce: MuldivTerm
			reduce(192), // .//, reduce: MuldivTerm
			shift(1203), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2430
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(193), // $, reduce: MuldivTerm
			nil,         // empty
			reduce(193), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(193), // >, reduce: MuldivTerm
			reduce(193), // >>, reduce: MuldivTerm
			reduce(193), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(193), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(193), // ||, reduce: MuldivTerm
			reduce(193), // ^^, reduce: MuldivTerm
			reduce(193), // &&, reduce: MuldivTerm
			reduce(193), // ??, reduce: MuldivTerm
			reduce(193), // ???, reduce: MuldivTerm
			reduce(193), // =~, reduce: MuldivTerm
			reduce(193), // !=~, reduce: MuldivTerm
			reduce(193), // ==, reduce: MuldivTerm
			reduce(193), // !=, reduce: MuldivTerm
			reduce(193), // <=>, reduce: MuldivTerm
			reduce(193), // >=, reduce: MuldivTerm
			reduce(193), // <, reduce: MuldivTerm
			reduce(193), // <=, reduce: MuldivTerm
			reduce(193), // ^, reduce: MuldivTerm
			reduce(193), // &, reduce: MuldivTerm
			reduce(193), // <<, reduce: MuldivTerm
			reduce(193), // >>>, reduce: MuldivTerm
			reduce(193), // +, reduce: MuldivTerm
			reduce(193), // -, reduce: MuldivTerm
			reduce(193), // .+, reduce: MuldivTerm
			reduce(193), // .-, reduce: MuldivTerm
			reduce(193), // *, reduce: MuldivTerm
			reduce(193), // /, reduce: MuldivTerm
			reduce(193), // //, reduce: MuldivTerm
			reduce(193), // %, reduce: MuldivTerm
			reduce(193), // .*, reduce: MuldivTerm
			reduce(193), // ./, reduce: MuldivTerm
			reduce(193), // .//, reduce: MuldivTerm
			shift(1203), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2431
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(195), // $, reduce: DotTerm
			nil,         // empty
			reduce(195), // ;, reduce: DotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(195), // >, reduce: DotTerm
			reduce(195), // >>, reduce: DotTerm
			reduce(195), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(195), // ?, reduce: DotTerm
			nil,         // :
			reduce(195), // ||, reduce: DotTerm
			reduce(195), // ^^, reduce: DotTerm
			reduce(195), // &&, reduce: DotTerm
			reduce(195), // ??, reduce: DotTerm
			reduce(195), // ???, reduce: DotTerm
			reduce(195), // =~, reduce: DotTerm
			reduce(195), // !=~, reduce: DotTerm
			reduce(195), // ==, reduce: DotTerm
			reduce(195), // !=, reduce: DotTerm
			reduce(195), // <=>, reduce: DotTerm
			reduce(195), // >=, reduce: DotTerm
			reduce(195), // <, reduce: DotTerm
			reduce(195), // <=, reduce: DotTerm
			reduce(195), // ^, reduce: DotTerm
			reduce(195), // &, reduce: DotTerm
			reduce(195), // <<, reduce: DotTerm
			reduce(195), // >>>, reduce: DotTerm
			reduce(195), // +, reduce: DotTerm
			reduce(195), // -, reduce: DotTerm
			reduce(195), // .+, reduce: DotTerm
			reduce(195), // .-, reduce: DotTerm
			reduce(195), // *, reduce: DotTerm
			reduce(195), // /, reduce: DotTerm
			reduce(195), // //, reduce: DotTerm
			reduce(195), // %, reduce: DotTerm
			reduce(195), // .*, reduce: DotTerm
			reduce(195), // ./, reduce: DotTerm
			reduce(195), // .//, reduce: DotTerm
			reduce(195), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2432
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2433
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2434
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(204), // $, reduce: PowTerm
			nil,         // empty
			reduce(204), // ;, reduce: PowTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(204), // >, reduce: PowTerm
			reduce(204), // >>, reduce: PowTerm
			reduce(204), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(204), // ?, reduce: PowTerm
			nil,         // :
			reduce(204), // ||, reduce: PowTerm
			reduce(204), // ^^, reduce: PowTerm
			reduce(204), // &&, reduce: PowTerm
			reduce(204), // ??, reduce: PowTerm
			reduce(204), // ???, reduce: PowTerm
			reduce(204), // =~, reduce: PowTerm
			reduce(204), // !=~, reduce: PowTerm
			reduce(204), // ==, reduce: PowTerm
			reduce(204), // !=, reduce: PowTerm
			reduce(204), // <=>, reduce: PowTerm
			reduce(204), // >=, reduce: PowTerm
			reduce(204), // <, reduce: PowTerm
			reduce(204), // <=, reduce: PowTerm
			reduce(204), // ^, reduce: PowTerm
			reduce(204), // &, reduce: PowTerm
			reduce(204), // <<, reduce: PowTerm
			reduce(204), // >>>, reduce: PowTerm
			reduce(204), // +, reduce: PowTerm
			reduce(204), // -, reduce: PowTerm
			reduce(204), // .+, reduce: PowTerm
			reduce(204), // .-, reduce: PowTerm
			reduce(204), // *, reduce: PowTerm
			reduce(204), // /, reduce: PowTerm
			reduce(204), // //, reduce: PowTerm
			reduce(204), // %, reduce: PowTerm
			reduce(204), // .*, reduce: PowTerm
			reduce(204), // ./, reduce: PowTerm
			reduce(204), // .//, reduce: PowTerm
			reduce(204), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2435
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3217), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3218), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2436
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(3220), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2437
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3221), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2438
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3222), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2439
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(229), // $, reduce: ArrayLiteral
			nil,         // empty
			reduce(229), // ;, reduce: ArrayLiteral
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(229), // >, reduce: ArrayLiteral
			reduce(229), // >>, reduce: ArrayLiteral
			reduce(229), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(229), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(229), // ||, reduce: ArrayLiteral
			reduce(229), // ^^, reduce: ArrayLiteral
			reduce(229), // &&, reduce: ArrayLiteral
			reduce(229), // ??, reduce: ArrayLiteral
			reduce(229), // ???, reduce: ArrayLiteral
			reduce(229), // =~, reduce: ArrayLiteral
			reduce(229), // !=~, reduce: ArrayLiteral
			reduce(229), // ==, reduce: ArrayLiteral
			reduce(229), // !=, reduce: ArrayLiteral
			reduce(229), // <=>, reduce: ArrayLiteral
			reduce(229), // >=, reduce: ArrayLiteral
			reduce(229), // <, reduce: ArrayLiteral
			reduce(229), // <=, reduce: ArrayLiteral
			reduce(229), // ^, reduce: ArrayLiteral
			reduce(229), // &, reduce: ArrayLiteral
			reduce(229), // <<, reduce: ArrayLiteral
			reduce(229), // >>>, reduce: ArrayLiteral
			reduce(229), // +, reduce: ArrayLiteral
			reduce(229), // -, reduce: ArrayLiteral
			reduce(229), // .+, reduce: ArrayLiteral
			reduce(229), // .-, reduce: ArrayLiteral
			reduce(229), // *, reduce: ArrayLiteral
			reduce(229), // /, reduce: ArrayLiteral
			reduce(229), // //, reduce: ArrayLiteral
			reduce(229), // %, reduce: ArrayLiteral
			reduce(229), // .*, reduce: ArrayLiteral
			reduce(229), // ./, reduce: ArrayLiteral
			reduce(229), // .//, reduce: ArrayLiteral
			reduce(229), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(229), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(229), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(229), // [[, reduce: ArrayLiteral
			reduce(229), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2440
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(255), // $, reduce: ENV
			nil,         // empty
			reduce(255), // ;, reduce: ENV
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(255), // >, reduce: ENV
			reduce(255), // >>, reduce: ENV
			reduce(255), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(255), // ?, reduce: ENV
			nil,         // :
			reduce(255), // ||, reduce: ENV
			reduce(255), // ^^, reduce: ENV
			reduce(255), // &&, reduce: ENV
			reduce(255), // ??, reduce: ENV
			reduce(255), // ???, reduce: ENV
			reduce(255), // =~, reduce: ENV
			reduce(255), // !=~, reduce: ENV
			reduce(255), // ==, reduce: ENV
			reduce(255), // !=, reduce: ENV
			reduce(255), // <=>, reduce: ENV
			reduce(255), // >=, reduce: ENV
			reduce(255), // <, reduce: ENV
			reduce(255), // <=, reduce: ENV
			reduce(255), // ^, reduce: ENV
			reduce(255), // &, reduce: ENV
			reduce(255), // <<, reduce: ENV
			reduce(255), // >>>, reduce: ENV
			reduce(255), // +, reduce: ENV
			reduce(255), // -, reduce: ENV
			reduce(255), // .+, reduce: ENV
			reduce(255), // .-, reduce: ENV
			reduce(255), // *, reduce: ENV
			reduce(255), // /, reduce: ENV
			reduce(255), // //, reduce: ENV
			reduce(255), // %, reduce: ENV
			reduce(255), // .*, reduce: ENV
			reduce(255), // ./, reduce: ENV
			reduce(255), // .//, reduce: ENV
			reduce(255), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(255), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(255), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(255), // [[, reduce: ENV
			reduce(255), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2441
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3223), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2442
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3224), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2443
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(268), // $, reduce: FunctionCallsite
			nil,         // empty
			reduce(268), // ;, reduce: FunctionCallsite
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(268), // >, reduce: FunctionCallsite
			reduce(268), // >>, reduce: FunctionCallsite
			reduce(268), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(268), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(268), // ||, reduce: FunctionCallsite
			reduce(268), // ^^, reduce: FunctionCallsite
			reduce(268), // &&, reduce: FunctionCallsite
			reduce(268), // ??, reduce: FunctionCallsite
			reduce(268), // ???, reduce: FunctionCallsite
			reduce(268), // =~, reduce: FunctionCallsite
			reduce(268), // !=~, reduce: FunctionCallsite
			reduce(268), // ==, reduce: FunctionCallsite
			reduce(268), // !=, reduce: FunctionCallsite
			reduce(268), // <=>, reduce: FunctionCallsite
			reduce(268), // >=, reduce: FunctionCallsite
			reduce(268), // <, reduce: FunctionCallsite
			reduce(268), // <=, reduce: FunctionCallsite
			reduce(268), // ^, reduce: FunctionCallsite
			reduce(268), // &, reduce: FunctionCallsite
			reduce(268), // <<, reduce: FunctionCallsite
			reduce(268), // >>>, reduce: FunctionCallsite
			reduce(268), // +, reduce: FunctionCallsite
			reduce(268), // -, reduce: FunctionCallsite
			reduce(268), // .+, reduce: FunctionCallsite
			reduce(268), // .-, reduce: FunctionCallsite
			reduce(268), // *, reduce: FunctionCallsite
			reduce(268), // /, reduce: FunctionCallsite
			reduce(268), // //, reduce: FunctionCallsite
			reduce(268), // %, reduce: FunctionCallsite
			reduce(268), // .*, reduce: FunctionCallsite
			reduce(268), // ./, reduce: FunctionCallsite
			reduce(268), // .//, reduce: FunctionCallsite
			reduce(268), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(268), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(268), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(268), // [[, reduce: FunctionCallsite
			reduce(268), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2444
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3225), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2445
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(49), // $, reduce: PrintStatement
			nil,        // empty
			reduce(49), // ;, reduce: PrintStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2446
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(53), // $, reduce: PrintnStatement
			nil,        // empty
			reduce(53), // ;, reduce: PrintnStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2447
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(61), // $, reduce: DumpStatement
			nil,        // empty
			reduce(61), // ;, reduce: DumpStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2448
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(64), // $, reduce: TeeStatement
			nil,        // empty
			reduce(64), // ;, reduce: TeeStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2449
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(109), // $, reduce: FullSrec
			nil,         // empty
			reduce(109), // ;, reduce: FullSrec
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2450
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(234), // >, reduce: MapLiteral
			reduce(234), // >>, reduce: MapLiteral
			reduce(234), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(234), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(234), // ?, reduce: MapLiteral
			nil,         // :
			reduce(234), // ||, reduce: MapLiteral
			reduce(234), // ^^, reduce: MapLiteral
			reduce(234), // &&, reduce: MapLiteral
			reduce(234), // ??, reduce: MapLiteral
			reduce(234), // ???, reduce: MapLiteral
			reduce(234), // =~, reduce: MapLiteral
			reduce(234), // !=~, reduce: MapLiteral
			reduce(234), // ==, reduce: MapLiteral
			reduce(234), // !=, reduce: MapLiteral
			reduce(234), // <=>, reduce: MapLiteral
			reduce(234), // >=, reduce: MapLiteral
			reduce(234), // <, reduce: MapLiteral
			reduce(234), // <=, reduce: MapLiteral
			reduce(234), // ^, reduce: MapLiteral
			reduce(234), // &, reduce: MapLiteral
			reduce(234), // <<, reduce: MapLiteral
			reduce(234), // >>>, reduce: MapLiteral
			reduce(234), // +, reduce: MapLiteral
			reduce(234), // -, reduce: MapLiteral
			reduce(234), // .+, reduce: MapLiteral
			reduce(234), // .-, reduce: MapLiteral
			reduce(234), // *, reduce: MapLiteral
			reduce(234), // /, reduce: MapLiteral
			reduce(234), // //, reduce: MapLiteral
			reduce(234), // %, reduce: MapLiteral
			reduce(234), // .*, reduce: MapLiteral
			reduce(234), // ./, reduce: MapLiteral
			reduce(234), // .//, reduce: MapLiteral
			reduce(234), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(234), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(234), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(234), // [[, reduce: MapLiteral
			reduce(234), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2451
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(3226), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2452
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3227), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2453
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3228), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2454
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3229), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2455
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3230), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2456
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3231), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2457
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2458
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2459
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2460
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2461
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2462
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2463
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2464
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2465
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2466
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2467
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2468
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2469
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2470
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2471
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2472
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2473
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2474
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2475
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2476
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2477
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2478
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2479
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2480
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2481
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2482
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(197), // >, reduce: UnaryOpTerm
			reduce(197), // >>, reduce: UnaryOpTerm
			reduce(197), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(197), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(197), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(197), // ||, reduce: UnaryOpTerm
			reduce(197), // ^^, reduce: UnaryOpTerm
			reduce(197), // &&, reduce: UnaryOpTerm
			reduce(197), // ??, reduce: UnaryOpTerm
			reduce(197), // ???, reduce: UnaryOpTerm
			reduce(197), // =~, reduce: UnaryOpTerm
			reduce(197), // !=~, reduce: UnaryOpTerm
			reduce(197), // ==, reduce: UnaryOpTerm
			reduce(197), // !=, reduce: UnaryOpTerm
			reduce(197), // <=>, reduce: UnaryOpTerm
			reduce(197), // >=, reduce: UnaryOpTerm
			reduce(197), // <, reduce: UnaryOpTerm
			reduce(197), // <=, reduce: UnaryOpTerm
			reduce(197), // ^, reduce: UnaryOpTerm
			reduce(197), // &, reduce: UnaryOpTerm
			reduce(197), // <<, reduce: UnaryOpTerm
			reduce(197), // >>>, reduce: UnaryOpTerm
			reduce(197), // +, reduce: UnaryOpTerm
			reduce(197), // -, reduce: UnaryOpTerm
			reduce(197), // .+, reduce: UnaryOpTerm
			reduce(197), // .-, reduce: UnaryOpTerm
			reduce(197), // *, reduce: UnaryOpTerm
			reduce(197), // /, reduce: UnaryOpTerm
			reduce(197), // //, reduce: UnaryOpTerm
			reduce(197), // %, reduce: UnaryOpTerm
			reduce(197), // .*, reduce: UnaryOpTerm
			reduce(197), // ./, reduce: UnaryOpTerm
			reduce(197), // .//, reduce: UnaryOpTerm
			reduce(197), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2483
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2484
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2485
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2486
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2487
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2488
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2489
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2490
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(198), // >, reduce: UnaryOpTerm
			reduce(198), // >>, reduce: UnaryOpTerm
			reduce(198), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(198), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(198), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(198), // ||, reduce: UnaryOpTerm
			reduce(198), // ^^, reduce: UnaryOpTerm
			reduce(198), // &&, reduce: UnaryOpTerm
			reduce(198), // ??, reduce: UnaryOpTerm
			reduce(198), // ???, reduce: UnaryOpTerm
			reduce(198), // =~, reduce: UnaryOpTerm
			reduce(198), // !=~, reduce: UnaryOpTerm
			reduce(198), // ==, reduce: UnaryOpTerm
			reduce(198), // !=, reduce: UnaryOpTerm
			reduce(198), // <=>, reduce: UnaryOpTerm
			reduce(198), // >=, reduce: UnaryOpTerm
			reduce(198), // <, reduce: UnaryOpTerm
			reduce(198), // <=, reduce: UnaryOpTerm
			reduce(198), // ^, reduce: UnaryOpTerm
			reduce(198), // &, reduce: UnaryOpTerm
			reduce(198), // <<, reduce: UnaryOpTerm
			reduce(198), // >>>, reduce: UnaryOpTerm
			reduce(198), // +, reduce: UnaryOpTerm
			reduce(198), // -, reduce: UnaryOpTerm
			reduce(198), // .+, reduce: UnaryOpTerm
			reduce(198), // .-, reduce: UnaryOpTerm
			reduce(198), // *, reduce: UnaryOpTerm
			reduce(198), // /, reduce: UnaryOpTerm
			reduce(198), // //, reduce: UnaryOpTerm
			reduce(198), // %, reduce: UnaryOpTerm
			reduce(198), // .*, reduce: UnaryOpTerm
			reduce(198), // ./, reduce: UnaryOpTerm
			reduce(198), // .//, reduce: UnaryOpTerm
			reduce(198), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2491
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(199), // >, reduce: UnaryOpTerm
			reduce(199), // >>, reduce: UnaryOpTerm
			reduce(199), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(199), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(199), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(199), // ||, reduce: UnaryOpTerm
			reduce(199), // ^^, reduce: UnaryOpTerm
			reduce(199), // &&, reduce: UnaryOpTerm
			reduce(199), // ??, reduce: UnaryOpTerm
			reduce(199), // ???, reduce: UnaryOpTerm
			reduce(199), // =~, reduce: UnaryOpTerm
			reduce(199), // !=~, reduce: UnaryOpTerm
			reduce(199), // ==, reduce: UnaryOpTerm
			reduce(199), // !=, reduce: UnaryOpTerm
			reduce(199), // <=>, reduce: UnaryOpTerm
			reduce(199), // >=, reduce: UnaryOpTerm
			reduce(199), // <, reduce: UnaryOpTerm
			reduce(199), // <=, reduce: UnaryOpTerm
			reduce(199), // ^, reduce: UnaryOpTerm
			reduce(199), // &, reduce: UnaryOpTerm
			reduce(199), // <<, reduce: UnaryOpTerm
			reduce(199), // >>>, reduce: UnaryOpTerm
			reduce(199), // +, reduce: UnaryOpTerm
			reduce(199), // -, reduce: UnaryOpTerm
			reduce(199), // .+, reduce: UnaryOpTerm
			reduce(199), // .-, reduce: UnaryOpTerm
			reduce(199), // *, reduce: UnaryOpTerm
			reduce(199), // /, reduce: UnaryOpTerm
			reduce(199), // //, reduce: UnaryOpTerm
			reduce(199), // %, reduce: UnaryOpTerm
			reduce(199), // .*, reduce: UnaryOpTerm
			reduce(199), // ./, reduce: UnaryOpTerm
			reduce(199), // .//, reduce: UnaryOpTerm
			reduce(199), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2492
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(200), // >, reduce: UnaryOpTerm
			reduce(200), // >>, reduce: UnaryOpTerm
			reduce(200), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(200), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(200), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(200), // ||, reduce: UnaryOpTerm
			reduce(200), // ^^, reduce: UnaryOpTerm
			reduce(200), // &&, reduce: UnaryOpTerm
			reduce(200), // ??, reduce: UnaryOpTerm
			reduce(200), // ???, reduce: UnaryOpTerm
			reduce(200), // =~, reduce: UnaryOpTerm
			reduce(200), // !=~, reduce: UnaryOpTerm
			reduce(200), // ==, reduce: UnaryOpTerm
			reduce(200), // !=, reduce: UnaryOpTerm
			reduce(200), // <=>, reduce: UnaryOpTerm
			reduce(200), // >=, reduce: UnaryOpTerm
			reduce(200), // <, reduce: UnaryOpTerm
			reduce(200), // <=, reduce: UnaryOpTerm
			reduce(200), // ^, reduce: UnaryOpTerm
			reduce(200), // &, reduce: UnaryOpTerm
			reduce(200), // <<, reduce: UnaryOpTerm
			reduce(200), // >>>, reduce: UnaryOpTerm
			reduce(200), // +, reduce: UnaryOpTerm
			reduce(200), // -, reduce: UnaryOpTerm
			reduce(200), // .+, reduce: UnaryOpTerm
			reduce(200), // .-, reduce: UnaryOpTerm
			reduce(200), // *, reduce: UnaryOpTerm
			reduce(200), // /, reduce: UnaryOpTerm
			reduce(200), // //, reduce: UnaryOpTerm
			reduce(200), // %, reduce: UnaryOpTerm
			reduce(200), // .*, reduce: UnaryOpTerm
			reduce(200), // ./, reduce: UnaryOpTerm
			reduce(200), // .//, reduce: UnaryOpTerm
			reduce(200), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2493
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2494
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(201), // >, reduce: UnaryOpTerm
			reduce(201), // >>, reduce: UnaryOpTerm
			reduce(201), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(201), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(201), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(201), // ||, reduce: UnaryOpTerm
			reduce(201), // ^^, reduce: UnaryOpTerm
			reduce(201), // &&, reduce: UnaryOpTerm
			reduce(201), // ??, reduce: UnaryOpTerm
			reduce(201), // ???, reduce: UnaryOpTerm
			reduce(201), // =~, reduce: UnaryOpTerm
			reduce(201), // !=~, reduce: UnaryOpTerm
			reduce(201), // ==, reduce: UnaryOpTerm
			reduce(201), // !=, reduce: UnaryOpTerm
			reduce(201), // <=>, reduce: UnaryOpTerm
			reduce(201), // >=, reduce: UnaryOpTerm
			reduce(201), // <, reduce: UnaryOpTerm
			reduce(201), // <=, reduce: UnaryOpTerm
			reduce(201), // ^, reduce: UnaryOpTerm
			reduce(201), // &, reduce: UnaryOpTerm
			reduce(201), // <<, reduce: UnaryOpTerm
			reduce(201), // >>>, reduce: UnaryOpTerm
			reduce(201), // +, reduce: UnaryOpTerm
			reduce(201), // -, reduce: UnaryOpTerm
			reduce(201), // .+, reduce: UnaryOpTerm
			reduce(201), // .-, reduce: UnaryOpTerm
			reduce(201), // *, reduce: UnaryOpTerm
			reduce(201), // /, reduce: UnaryOpTerm
			reduce(201), // //, reduce: UnaryOpTerm
			reduce(201), // %, reduce: UnaryOpTerm
			reduce(201), // .*, reduce: UnaryOpTerm
			reduce(201), // ./, reduce: UnaryOpTerm
			reduce(201), // .//, reduce: UnaryOpTerm
			reduce(201), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2495
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(202), // >, reduce: UnaryOpTerm
			reduce(202), // >>, reduce: UnaryOpTerm
			reduce(202), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(202), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(202), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(202), // ||, reduce: UnaryOpTerm
			reduce(202), // ^^, reduce: UnaryOpTerm
			reduce(202), // &&, reduce: UnaryOpTerm
			reduce(202), // ??, reduce: UnaryOpTerm
			reduce(202), // ???, reduce: UnaryOpTerm
			reduce(202), // =~, reduce: UnaryOpTerm
			reduce(202), // !=~, reduce: UnaryOpTerm
			reduce(202), // ==, reduce: UnaryOpTerm
			reduce(202), // !=, reduce: UnaryOpTerm
			reduce(202), // <=>, reduce: UnaryOpTerm
			reduce(202), // >=, reduce: UnaryOpTerm
			reduce(202), // <, reduce: UnaryOpTerm
			reduce(202), // <=, reduce: UnaryOpTerm
			reduce(202), // ^, reduce: UnaryOpTerm
			reduce(202), // &, reduce: UnaryOpTerm
			reduce(202), // <<, reduce: UnaryOpTerm
			reduce(202), // >>>, reduce: UnaryOpTerm
			reduce(202), // +, reduce: UnaryOpTerm
			reduce(202), // -, reduce: UnaryOpTerm
			reduce(202), // .+, reduce: UnaryOpTerm
			reduce(202), // .-, reduce: UnaryOpTerm
			reduce(202), // *, reduce: UnaryOpTerm
			reduce(202), // /, reduce: UnaryOpTerm
			reduce(202), // //, reduce: UnaryOpTerm
			reduce(202), // %, reduce: UnaryOpTerm
			reduce(202), // .*, reduce: UnaryOpTerm
			reduce(202), // ./, reduce: UnaryOpTerm
			reduce(202), // .//, reduce: UnaryOpTerm
			reduce(202), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2496
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3265), // +
			shift(3266), // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2497
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3269), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2498
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2499
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2500
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(228), // >, reduce: ArrayLiteral
			reduce(228), // >>, reduce: ArrayLiteral
			reduce(228), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(228), // ,, reduce: ArrayLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(228), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(228), // ||, reduce: ArrayLiteral
			reduce(228), // ^^, reduce: ArrayLiteral
			reduce(228), // &&, reduce: ArrayLiteral
			reduce(228), // ??, reduce: ArrayLiteral
			reduce(228), // ???, reduce: ArrayLiteral
			reduce(228), // =~, reduce: ArrayLiteral
			reduce(228), // !=~, reduce: ArrayLiteral
			reduce(228), // ==, reduce: ArrayLiteral
			reduce(228), // !=, reduce: ArrayLiteral
			reduce(228), // <=>, reduce: ArrayLiteral
			reduce(228), // >=, reduce: ArrayLiteral
			reduce(228), // <, reduce: ArrayLiteral
			reduce(228), // <=, reduce: ArrayLiteral
			reduce(228), // ^, reduce: ArrayLiteral
			reduce(228), // &, reduce: ArrayLiteral
			reduce(228), // <<, reduce: ArrayLiteral
			reduce(228), // >>>, reduce: ArrayLiteral
			reduce(228), // +, reduce: ArrayLiteral
			reduce(228), // -, reduce: ArrayLiteral
			reduce(228), // .+, reduce: ArrayLiteral
			reduce(228), // .-, reduce: ArrayLiteral
			reduce(228), // *, reduce: ArrayLiteral
			reduce(228), // /, reduce: ArrayLiteral
			reduce(228), // //, reduce: ArrayLiteral
			reduce(228), // %, reduce: ArrayLiteral
			reduce(228), // .*, reduce: ArrayLiteral
			reduce(228), // ./, reduce: ArrayLiteral
			reduce(228), // .//, reduce: ArrayLiteral
			reduce(228), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(228), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(228), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(228), // [[, reduce: ArrayLiteral
			reduce(228), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2501
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3272), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2502
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(3273), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2503
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2504
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(3276), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2505
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2506
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(235), // $, reduce: MapLiteral
			nil,         // empty
			reduce(235), // ;, reduce: MapLiteral
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(235), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2507
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(66), // $, reduce: EmitFStatement
			nil,        // empty
			reduce(66), // ;, reduce: EmitFStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2508
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(85), // $, reduce: EmittableList
			nil,        // empty
			reduce(85), // ;, reduce: EmittableList
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2509
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(105), // $, reduce: IndirectFieldValue
			nil,         // empty
			reduce(105), // ;, reduce: IndirectFieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(105), // ,, reduce: IndirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2510
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(114), // $, reduce: IndirectOosvarValue
			nil,         // empty
			reduce(114), // ;, reduce: IndirectOosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(114), // ,, reduce: IndirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2511
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(69),  // $, reduce: EmitStatement
			nil,         // empty
			reduce(69),  // ;, reduce: EmitStatement
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3278), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2512
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1330), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			shift(1343), // field_name
			shift(1344), // $[
			nil,         // ]
			shift(1345), // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(1346), // full_srec
			shift(1347), // oosvar_name
			shift(1348), // @[
			shift(1349), // braced_oosvar_name
			shift(1350), // full_oosvar
			shift(1351), // all
			shift(1352), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2513
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(97),  // $, reduce: EmitKeys
			nil,         // empty
			reduce(97),  // ;, reduce: EmitKeys
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3280), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2514
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(72), // $, reduce: EmitStatement
			nil,        // empty
			reduce(72), // ;, reduce: EmitStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2515
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(234), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(234), // ), reduce: MapLiteral
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2516
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(3281), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2517
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(70),  // $, reduce: EmitStatement
			nil,         // empty
			reduce(70),  // ;, reduce: EmitStatement
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3282), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2518
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1330), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			shift(1343), // field_name
			shift(1344), // $[
			nil,         // ]
			shift(1345), // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(1346), // full_srec
			shift(1347), // oosvar_name
			shift(1348), // @[
			shift(1349), // braced_oosvar_name
			shift(1350), // full_oosvar
			shift(1351), // all
			shift(1352), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2519
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3284), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2520
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3285), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2521
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(235), // >, reduce: MapLiteral
			reduce(235), // >>, reduce: MapLiteral
			reduce(235), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(235), // ), reduce: MapLiteral
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(235), // ?, reduce: MapLiteral
			nil,         // :
			reduce(235), // ||, reduce: MapLiteral
			reduce(235), // ^^, reduce: MapLiteral
			reduce(235), // &&, reduce: MapLiteral
			reduce(235), // ??, reduce: MapLiteral
			reduce(235), // ???, reduce: MapLiteral
			reduce(235), // =~, reduce: MapLiteral
			reduce(235), // !=~, reduce: MapLiteral
			reduce(235), // ==, reduce: MapLiteral
			reduce(235), // !=, reduce: MapLiteral
			reduce(235), // <=>, reduce: MapLiteral
			reduce(235), // >=, reduce: MapLiteral
			reduce(235), // <, reduce: MapLiteral
			reduce(235), // <=, reduce: MapLiteral
			reduce(235), // ^, reduce: MapLiteral
			reduce(235), // &, reduce: MapLiteral
			reduce(235), // <<, reduce: MapLiteral
			reduce(235), // >>>, reduce: MapLiteral
			reduce(235), // +, reduce: MapLiteral
			reduce(235), // -, reduce: MapLiteral
			reduce(235), // .+, reduce: MapLiteral
			reduce(235), // .-, reduce: MapLiteral
			reduce(235), // *, reduce: MapLiteral
			reduce(235), // /, reduce: MapLiteral
			reduce(235), // //, reduce: MapLiteral
			reduce(235), // %, reduce: MapLiteral
			reduce(235), // .*, reduce: MapLiteral
			reduce(235), // ./, reduce: MapLiteral
			reduce(235), // .//, reduce: MapLiteral
			reduce(235), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(235), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(235), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(235), // [[, reduce: MapLiteral
			reduce(235), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2522
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(208), // >, reduce: PrecedenceChainEnd
			reduce(208), // >>, reduce: PrecedenceChainEnd
			reduce(208), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(208), // ), reduce: PrecedenceChainEnd
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(208), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(208), // ||, reduce: PrecedenceChainEnd
			reduce(208), // ^^, reduce: PrecedenceChainEnd
			reduce(208), // &&, reduce: PrecedenceChainEnd
			reduce(208), // ??, reduce: PrecedenceChainEnd
			reduce(208), // ???, reduce: PrecedenceChainEnd
			reduce(208), // =~, reduce: PrecedenceChainEnd
			reduce(208), // !=~, reduce: PrecedenceChainEnd
			reduce(208), // ==, reduce: PrecedenceChainEnd
			reduce(208), // !=, reduce: PrecedenceChainEnd
			reduce(208), // <=>, reduce: PrecedenceChainEnd
			reduce(208), // >=, reduce: PrecedenceChainEnd
			reduce(208), // <, reduce: PrecedenceChainEnd
			reduce(208), // <=, reduce: PrecedenceChainEnd
			reduce(208), // ^, reduce: PrecedenceChainEnd
			reduce(208), // &, reduce: PrecedenceChainEnd
			reduce(208), // <<, reduce: PrecedenceChainEnd
			reduce(208), // >>>, reduce: PrecedenceChainEnd
			reduce(208), // +, reduce: PrecedenceChainEnd
			reduce(208), // -, reduce: PrecedenceChainEnd
			reduce(208), // .+, reduce: PrecedenceChainEnd
			reduce(208), // .-, reduce: PrecedenceChainEnd
			reduce(208), // *, reduce: PrecedenceChainEnd
			reduce(208), // /, reduce: PrecedenceChainEnd
			reduce(208), // //, reduce: PrecedenceChainEnd
			reduce(208), // %, reduce: PrecedenceChainEnd
			reduce(208), // .*, reduce: PrecedenceChainEnd
			reduce(208), // ./, reduce: PrecedenceChainEnd
			reduce(208), // .//, reduce: PrecedenceChainEnd
			reduce(208), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(208), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2523
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(105), // >, reduce: IndirectFieldValue
			reduce(105), // >>, reduce: IndirectFieldValue
			reduce(105), // |, reduce: IndirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(105), // ), reduce: IndirectFieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(105), // ?, reduce: IndirectFieldValue
			nil,         // :
			reduce(105), // ||, reduce: IndirectFieldValue
			reduce(105), // ^^, reduce: IndirectFieldValue
			reduce(105), // &&, reduce: IndirectFieldValue
			reduce(105), // ??, reduce: IndirectFieldValue
			reduce(105), // ???, reduce: IndirectFieldValue
			reduce(105), // =~, reduce: IndirectFieldValue
			reduce(105), // !=~, reduce: IndirectFieldValue
			reduce(105), // ==, reduce: IndirectFieldValue
			reduce(105), // !=, reduce: IndirectFieldValue
			reduce(105), // <=>, reduce: IndirectFieldValue
			reduce(105), // >=, reduce: IndirectFieldValue
			reduce(105), // <, reduce: IndirectFieldValue
			reduce(105), // <=, reduce: IndirectFieldValue
			reduce(105), // ^, reduce: IndirectFieldValue
			reduce(105), // &, reduce: IndirectFieldValue
			reduce(105), // <<, reduce: IndirectFieldValue
			reduce(105), // >>>, reduce: IndirectFieldValue
			reduce(105), // +, reduce: IndirectFieldValue
			reduce(105), // -, reduce: IndirectFieldValue
			reduce(105), // .+, reduce: IndirectFieldValue
			reduce(105), // .-, reduce: IndirectFieldValue
			reduce(105), // *, reduce: IndirectFieldValue
			reduce(105), // /, reduce: IndirectFieldValue
			reduce(105), // //, reduce: IndirectFieldValue
			reduce(105), // %, reduce: IndirectFieldValue
			reduce(105), // .*, reduce: IndirectFieldValue
			reduce(105), // ./, reduce: IndirectFieldValue
			reduce(105), // .//, reduce: IndirectFieldValue
			reduce(105), // ., reduce: IndirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(105), // **, reduce: IndirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(105), // [, reduce: IndirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(105), // [[, reduce: IndirectFieldValue
			reduce(105), // [[[, reduce: IndirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2524
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3286), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2525
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3287), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2526
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(114), // >, reduce: IndirectOosvarValue
			reduce(114), // >>, reduce: IndirectOosvarValue
			reduce(114), // |, reduce: IndirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(114), // ), reduce: IndirectOosvarValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(114), // ?, reduce: IndirectOosvarValue
			nil,         // :
			reduce(114), // ||, reduce: IndirectOosvarValue
			reduce(114), // ^^, reduce: IndirectOosvarValue
			reduce(114), // &&, reduce: IndirectOosvarValue
			reduce(114), // ??, reduce: IndirectOosvarValue
			reduce(114), // ???, reduce: IndirectOosvarValue
			reduce(114), // =~, reduce: IndirectOosvarValue
			reduce(114), // !=~, reduce: IndirectOosvarValue
			reduce(114), // ==, reduce: IndirectOosvarValue
			reduce(114), // !=, reduce: IndirectOosvarValue
			reduce(114), // <=>, reduce: IndirectOosvarValue
			reduce(114), // >=, reduce: IndirectOosvarValue
			reduce(114), // <, reduce: IndirectOosvarValue
			reduce(114), // <=, reduce: IndirectOosvarValue
			reduce(114), // ^, reduce: IndirectOosvarValue
			reduce(114), // &, reduce: IndirectOosvarValue
			reduce(114), // <<, reduce: IndirectOosvarValue
			reduce(114), // >>>, reduce: IndirectOosvarValue
			reduce(114), // +, reduce: IndirectOosvarValue
			reduce(114), // -, reduce: IndirectOosvarValue
			reduce(114), // .+, reduce: IndirectOosvarValue
			reduce(114), // .-, reduce: IndirectOosvarValue
			reduce(114), // *, reduce: IndirectOosvarValue
			reduce(114), // /, reduce: IndirectOosvarValue
			reduce(114), // //, reduce: IndirectOosvarValue
			reduce(114), // %, reduce: IndirectOosvarValue
			reduce(114), // .*, reduce: IndirectOosvarValue
			reduce(114), // ./, reduce: IndirectOosvarValue
			reduce(114), // .//, reduce: IndirectOosvarValue
			reduce(114), // ., reduce: IndirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(114), // **, reduce: IndirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(114), // [, reduce: IndirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(114), // [[, reduce: IndirectOosvarValue
			reduce(114), // [[[, reduce: IndirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2527
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3288), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2528
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(151), // ), reduce: LogicalOrTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(151), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(151), // ||, reduce: LogicalOrTerm
			shift(1363), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2529
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(153), // ), reduce: LogicalXORTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(153), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(153), // ||, reduce: LogicalXORTerm
			reduce(153), // ^^, reduce: LogicalXORTerm
			shift(1364), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2530
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(155), // ), reduce: LogicalAndTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(155), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(155), // ||, reduce: LogicalAndTerm
			reduce(155), // ^^, reduce: LogicalAndTerm
			reduce(155), // &&, reduce: LogicalAndTerm
			shift(1365), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2531
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(157), // ), reduce: AbsentCoalesceTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(157), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(157), // ||, reduce: AbsentCoalesceTerm
			reduce(157), // ^^, reduce: AbsentCoalesceTerm
			reduce(157), // &&, reduce: AbsentCoalesceTerm
			reduce(157), // ??, reduce: AbsentCoalesceTerm
			shift(1366), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2532
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(159), // ), reduce: EmptyCoalesceTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(159), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(159), // ||, reduce: EmptyCoalesceTerm
			reduce(159), // ^^, reduce: EmptyCoalesceTerm
			reduce(159), // &&, reduce: EmptyCoalesceTerm
			reduce(159), // ??, reduce: EmptyCoalesceTerm
			reduce(159), // ???, reduce: EmptyCoalesceTerm
			shift(1367), // =~
			shift(1368), // !=~
			shift(1369), // ==
			shift(1370), // !=
			shift(1371), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2533
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1372), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(161), // ), reduce: EqneTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(161), // ?, reduce: EqneTerm
			nil,         // :
			reduce(161), // ||, reduce: EqneTerm
			reduce(161), // ^^, reduce: EqneTerm
			reduce(161), // &&, reduce: EqneTerm
			reduce(161), // ??, reduce: EqneTerm
			reduce(161), // ???, reduce: EqneTerm
			reduce(161), // =~, reduce: EqneTerm
			reduce(161), // !=~, reduce: EqneTerm
			reduce(161), // ==, reduce: EqneTerm
			reduce(161), // !=, reduce: EqneTerm
			reduce(161), // <=>, reduce: EqneTerm
			shift(1373), // >=
			shift(1374), // <
			shift(1375), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2534
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1372), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(162), // ), reduce: EqneTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(162), // ?, reduce: EqneTerm
			nil,         // :
			reduce(162), // ||, reduce: EqneTerm
			reduce(162), // ^^, reduce: EqneTerm
			reduce(162), // &&, reduce: EqneTerm
			reduce(162), // ??, reduce: EqneTerm
			reduce(162), // ???, reduce: EqneTerm
			reduce(162), // =~, reduce: EqneTerm
			reduce(162), // !=~, reduce: EqneTerm
			reduce(162), // ==, reduce: EqneTerm
			reduce(162), // !=, reduce: EqneTerm
			reduce(162), // <=>, reduce: EqneTerm
			shift(1373), // >=
			shift(1374), // <
			shift(1375), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2535
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1372), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(163), // ), reduce: EqneTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(163), // ?, reduce: EqneTerm
			nil,         // :
			reduce(163), // ||, reduce: EqneTerm
			reduce(163), // ^^, reduce: EqneTerm
			reduce(163), // &&, reduce: EqneTerm
			reduce(163), // ??, reduce: EqneTerm
			reduce(163), // ???, reduce: EqneTerm
			reduce(163), // =~, reduce: EqneTerm
			reduce(163), // !=~, reduce: EqneTerm
			reduce(163), // ==, reduce: EqneTerm
			reduce(163), // !=, reduce: EqneTerm
			reduce(163), // <=>, reduce: EqneTerm
			shift(1373), // >=
			shift(1374), // <
			shift(1375), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2536
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1372), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(164), // ), reduce: EqneTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(164), // ?, reduce: EqneTerm
			nil,         // :
			reduce(164), // ||, reduce: EqneTerm
			reduce(164), // ^^, reduce: EqneTerm
			reduce(164), // &&, reduce: EqneTerm
			reduce(164), // ??, reduce: EqneTerm
			reduce(164), // ???, reduce: EqneTerm
			reduce(164), // =~, reduce: EqneTerm
			reduce(164), // !=~, reduce: EqneTerm
			reduce(164), // ==, reduce: EqneTerm
			reduce(164), // !=, reduce: EqneTerm
			reduce(164), // <=>, reduce: EqneTerm
			shift(1373), // >=
			shift(1374), // <
			shift(1375), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2537
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1372), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(165), // ), reduce: EqneTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(165), // ?, reduce: EqneTerm
			nil,         // :
			reduce(165), // ||, reduce: EqneTerm
			reduce(165), // ^^, reduce: EqneTerm
			reduce(165), // &&, reduce: EqneTerm
			reduce(165), // ??, reduce: EqneTerm
			reduce(165), // ???, reduce: EqneTerm
			reduce(165), // =~, reduce: EqneTerm
			reduce(165), // !=~, reduce: EqneTerm
			reduce(165), // ==, reduce: EqneTerm
			reduce(165), // !=, reduce: EqneTerm
			reduce(165), // <=>, reduce: EqneTerm
			shift(1373), // >=
			shift(1374), // <
			shift(1375), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2538
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(167), // >, reduce: CmpTerm
			nil,         // >>
			shift(1376), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(167), // ), reduce: CmpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(167), // ?, reduce: CmpTerm
			nil,         // :
			reduce(167), // ||, reduce: CmpTerm
			reduce(167), // ^^, reduce: CmpTerm
			reduce(167), // &&, reduce: CmpTerm
			reduce(167), // ??, reduce: CmpTerm
			reduce(167), // ???, reduce: CmpTerm
			reduce(167), // =~, reduce: CmpTerm
			reduce(167), // !=~, reduce: CmpTerm
			reduce(167), // ==, reduce: CmpTerm
			reduce(167), // !=, reduce: CmpTerm
			reduce(167), // <=>, reduce: CmpTerm
			reduce(167), // >=, reduce: CmpTerm
			reduce(167), // <, reduce: CmpTerm
			reduce(167), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2539
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(168), // >, reduce: CmpTerm
			nil,         // >>
			shift(1376), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(168), // ), reduce: CmpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(168), // ?, reduce: CmpTerm
			nil,         // :
			reduce(168), // ||, reduce: CmpTerm
			reduce(168), // ^^, reduce: CmpTerm
			reduce(168), // &&, reduce: CmpTerm
			reduce(168), // ??, reduce: CmpTerm
			reduce(168), // ???, reduce: CmpTerm
			reduce(168), // =~, reduce: CmpTerm
			reduce(168), // !=~, reduce: CmpTerm
			reduce(168), // ==, reduce: CmpTerm
			reduce(168), // !=, reduce: CmpTerm
			reduce(168), // <=>, reduce: CmpTerm
			reduce(168), // >=, reduce: CmpTerm
			reduce(168), // <, reduce: CmpTerm
			reduce(168), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2540
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(169), // >, reduce: CmpTerm
			nil,         // >>
			shift(1376), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(169), // ), reduce: CmpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(169), // ?, reduce: CmpTerm
			nil,         // :
			reduce(169), // ||, reduce: CmpTerm
			reduce(169), // ^^, reduce: CmpTerm
			reduce(169), // &&, reduce: CmpTerm
			reduce(169), // ??, reduce: CmpTerm
			reduce(169), // ???, reduce: CmpTerm
			reduce(169), // =~, reduce: CmpTerm
			reduce(169), // !=~, reduce: CmpTerm
			reduce(169), // ==, reduce: CmpTerm
			reduce(169), // !=, reduce: CmpTerm
			reduce(169), // <=>, reduce: CmpTerm
			reduce(169), // >=, reduce: CmpTerm
			reduce(169), // <, reduce: CmpTerm
			reduce(169), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2541
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(170), // >, reduce: CmpTerm
			nil,         // >>
			shift(1376), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(170), // ), reduce: CmpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(170), // ?, reduce: CmpTerm
			nil,         // :
			reduce(170), // ||, reduce: CmpTerm
			reduce(170), // ^^, reduce: CmpTerm
			reduce(170), // &&, reduce: CmpTerm
			reduce(170), // ??, reduce: CmpTerm
			reduce(170), // ???, reduce: CmpTerm
			reduce(170), // =~, reduce: CmpTerm
			reduce(170), // !=~, reduce: CmpTerm
			reduce(170), // ==, reduce: CmpTerm
			reduce(170), // !=, reduce: CmpTerm
			reduce(170), // <=>, reduce: CmpTerm
			reduce(170), // >=, reduce: CmpTerm
			reduce(170), // <, reduce: CmpTerm
			reduce(170), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2542
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(172), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(172), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(172), // ), reduce: BitwiseORTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(172), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(172), // ||, reduce: BitwiseORTerm
			reduce(172), // ^^, reduce: BitwiseORTerm
			reduce(172), // &&, reduce: BitwiseORTerm
			reduce(172), // ??, reduce: BitwiseORTerm
			reduce(172), // ???, reduce: BitwiseORTerm
			reduce(172), // =~, reduce: BitwiseORTerm
			reduce(172), // !=~, reduce: BitwiseORTerm
			reduce(172), // ==, reduce: BitwiseORTerm
			reduce(172), // !=, reduce: BitwiseORTerm
			reduce(172), // <=>, reduce: BitwiseORTerm
			reduce(172), // >=, reduce: BitwiseORTerm
			reduce(172), // <, reduce: BitwiseORTerm
			reduce(172), // <=, reduce: BitwiseORTerm
			shift(1377), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2543
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(174), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(174), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(174), // ), reduce: BitwiseXORTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(174), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(174), // ||, reduce: BitwiseXORTerm
			reduce(174), // ^^, reduce: BitwiseXORTerm
			reduce(174), // &&, reduce: BitwiseXORTerm
			reduce(174), // ??, reduce: BitwiseXORTerm
			reduce(174), // ???, reduce: BitwiseXORTerm
			reduce(174), // =~, reduce: BitwiseXORTerm
			reduce(174), // !=~, reduce: BitwiseXORTerm
			reduce(174), // ==, reduce: BitwiseXORTerm
			reduce(174), // !=, reduce: BitwiseXORTerm
			reduce(174), // <=>, reduce: BitwiseXORTerm
			reduce(174), // >=, reduce: BitwiseXORTerm
			reduce(174), // <, reduce: BitwiseXORTerm
			reduce(174), // <=, reduce: BitwiseXORTerm
			reduce(174), // ^, reduce: BitwiseXORTerm
			shift(1378), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2544
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(176), // >, reduce: BitwiseANDTerm
			shift(1379), // >>
			reduce(176), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(176), // ), reduce: BitwiseANDTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(176), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(176), // ||, reduce: BitwiseANDTerm
			reduce(176), // ^^, reduce: BitwiseANDTerm
			reduce(176), // &&, reduce: BitwiseANDTerm
			reduce(176), // ??, reduce: BitwiseANDTerm
			reduce(176), // ???, reduce: BitwiseANDTerm
			reduce(176), // =~, reduce: BitwiseANDTerm
			reduce(176), // !=~, reduce: BitwiseANDTerm
			reduce(176), // ==, reduce: BitwiseANDTerm
			reduce(176), // !=, reduce: BitwiseANDTerm
			reduce(176), // <=>, reduce: BitwiseANDTerm
			reduce(176), // >=, reduce: BitwiseANDTerm
			reduce(176), // <, reduce: BitwiseANDTerm
			reduce(176), // <=, reduce: BitwiseANDTerm
			reduce(176), // ^, reduce: BitwiseANDTerm
			reduce(176), // &, reduce: BitwiseANDTerm
			shift(1380), // <<
			shift(1381), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2545
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(179), // >, reduce: BitwiseShiftTerm
			reduce(179), // >>, reduce: BitwiseShiftTerm
			reduce(179), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(179), // ), reduce: BitwiseShiftTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(179), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(179), // ||, reduce: BitwiseShiftTerm
			reduce(179), // ^^, reduce: BitwiseShiftTerm
			reduce(179), // &&, reduce: BitwiseShiftTerm
			reduce(179), // ??, reduce: BitwiseShiftTerm
			reduce(179), // ???, reduce: BitwiseShiftTerm
			reduce(179), // =~, reduce: BitwiseShiftTerm
			reduce(179), // !=~, reduce: BitwiseShiftTerm
			reduce(179), // ==, reduce: BitwiseShiftTerm
			reduce(179), // !=, reduce: BitwiseShiftTerm
			reduce(179), // <=>, reduce: BitwiseShiftTerm
			reduce(179), // >=, reduce: BitwiseShiftTerm
			reduce(179), // <, reduce: BitwiseShiftTerm
			reduce(179), // <=, reduce: BitwiseShiftTerm
			reduce(179), // ^, reduce: BitwiseShiftTerm
			reduce(179), // &, reduce: BitwiseShiftTerm
			reduce(179), // <<, reduce: BitwiseShiftTerm
			reduce(179), // >>>, reduce: BitwiseShiftTerm
			shift(1382), // +
			shift(1383), // -
			shift(1384), // .+
			shift(1385), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2546
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(178), // >, reduce: BitwiseShiftTerm
			reduce(178), // >>, reduce: BitwiseShiftTerm
			reduce(178), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(178), // ), reduce: BitwiseShiftTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(178), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(178), // ||, reduce: BitwiseShiftTerm
			reduce(178), // ^^, reduce: BitwiseShiftTerm
			reduce(178), // &&, reduce: BitwiseShiftTerm
			reduce(178), // ??, reduce: BitwiseShiftTerm
			reduce(178), // ???, reduce: BitwiseShiftTerm
			reduce(178), // =~, reduce: BitwiseShiftTerm
			reduce(178), // !=~, reduce: BitwiseShiftTerm
			reduce(178), // ==, reduce: BitwiseShiftTerm
			reduce(178), // !=, reduce: BitwiseShiftTerm
			reduce(178), // <=>, reduce: BitwiseShiftTerm
			reduce(178), // >=, reduce: BitwiseShiftTerm
			reduce(178), // <, reduce: BitwiseShiftTerm
			reduce(178), // <=, reduce: BitwiseShiftTerm
			reduce(178), // ^, reduce: BitwiseShiftTerm
			reduce(178), // &, reduce: BitwiseShiftTerm
			reduce(178), // <<, reduce: BitwiseShiftTerm
			reduce(178), // >>>, reduce: BitwiseShiftTerm
			shift(1382), // +
			shift(1383), // -
			shift(1384), // .+
			shift(1385), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2547
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(180), // >, reduce: BitwiseShiftTerm
			reduce(180), // >>, reduce: BitwiseShiftTerm
			reduce(180), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(180), // ), reduce: BitwiseShiftTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(180), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(180), // ||, reduce: BitwiseShiftTerm
			reduce(180), // ^^, reduce: BitwiseShiftTerm
			reduce(180), // &&, reduce: BitwiseShiftTerm
			reduce(180), // ??, reduce: BitwiseShiftTerm
			reduce(180), // ???, reduce: BitwiseShiftTerm
			reduce(180), // =~, reduce: BitwiseShiftTerm
			reduce(180), // !=~, reduce: BitwiseShiftTerm
			reduce(180), // ==, reduce: BitwiseShiftTerm
			reduce(180), // !=, reduce: BitwiseShiftTerm
			reduce(180), // <=>, reduce: BitwiseShiftTerm
			reduce(180), // >=, reduce: BitwiseShiftTerm
			reduce(180), // <, reduce: BitwiseShiftTerm
			reduce(180), // <=, reduce: BitwiseShiftTerm
			reduce(180), // ^, reduce: BitwiseShiftTerm
			reduce(180), // &, reduce: BitwiseShiftTerm
			reduce(180), // <<, reduce: BitwiseShiftTerm
			reduce(180), // >>>, reduce: BitwiseShiftTerm
			shift(1382), // +
			shift(1383), // -
			shift(1384), // .+
			shift(1385), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2548
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(182), // >, reduce: AddsubdotTerm
			reduce(182), // >>, reduce: AddsubdotTerm
			reduce(182), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(182), // ), reduce: AddsubdotTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(182), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(182), // ||, reduce: AddsubdotTerm
			reduce(182), // ^^, reduce: AddsubdotTerm
			reduce(182), // &&, reduce: AddsubdotTerm
			reduce(182), // ??, reduce: AddsubdotTerm
			reduce(182), // ???, reduce: AddsubdotTerm
			reduce(182), // =~, reduce: AddsubdotTerm
			reduce(182), // !=~, reduce: AddsubdotTerm
			reduce(182), // ==, reduce: AddsubdotTerm
			reduce(182), // !=, reduce: AddsubdotTerm
			reduce(182), // <=>, reduce: AddsubdotTerm
			reduce(182), // >=, reduce: AddsubdotTerm
			reduce(182), // <, reduce: AddsubdotTerm
			reduce(182), // <=, reduce: AddsubdotTerm
			reduce(182), // ^, reduce: AddsubdotTerm
			reduce(182), // &, reduce: AddsubdotTerm
			reduce(182), // <<, reduce: AddsubdotTerm
			reduce(182), // >>>, reduce: AddsubdotTerm
			reduce(182), // +, reduce: AddsubdotTerm
			reduce(182), // -, reduce: AddsubdotTerm
			reduce(182), // .+, reduce: AddsubdotTerm
			reduce(182), // .-, reduce: AddsubdotTerm
			shift(1387), // *
			shift(1388), // /
			shift(1389), // //
			shift(1390), // %
			shift(1391), // .*
			shift(1392), // ./
			shift(1393), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2549
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(183), // >, reduce: AddsubdotTerm
			reduce(183), // >>, reduce: AddsubdotTerm
			reduce(183), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(183), // ), reduce: AddsubdotTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(183), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(183), // ||, reduce: AddsubdotTerm
			reduce(183), // ^^, reduce: AddsubdotTerm
			reduce(183), // &&, reduce: AddsubdotTerm
			reduce(183), // ??, reduce: AddsubdotTerm
			reduce(183), // ???, reduce: AddsubdotTerm
			reduce(183), // =~, reduce: AddsubdotTerm
			reduce(183), // !=~, reduce: AddsubdotTerm
			reduce(183), // ==, reduce: AddsubdotTerm
			reduce(183), // !=, reduce: AddsubdotTerm
			reduce(183), // <=>, reduce: AddsubdotTerm
			reduce(183), // >=, reduce: AddsubdotTerm
			reduce(183), // <, reduce: AddsubdotTerm
			reduce(183), // <=, reduce: AddsubdotTerm
			reduce(183), // ^, reduce: AddsubdotTerm
			reduce(183), // &, reduce: AddsubdotTerm
			reduce(183), // <<, reduce: AddsubdotTerm
			reduce(183), // >>>, reduce: AddsubdotTerm
			reduce(183), // +, reduce: AddsubdotTerm
			reduce(183), // -, reduce: AddsubdotTerm
			reduce(183), // .+, reduce: AddsubdotTerm
			reduce(183), // .-, reduce: AddsubdotTerm
			shift(1387), // *
			shift(1388), // /
			shift(1389), // //
			shift(1390), // %
			shift(1391), // .*
			shift(1392), // ./
			shift(1393), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2550
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(184), // >, reduce: AddsubdotTerm
			reduce(184), // >>, reduce: AddsubdotTerm
			reduce(184), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(184), // ), reduce: AddsubdotTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(184), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(184), // ||, reduce: AddsubdotTerm
			reduce(184), // ^^, reduce: AddsubdotTerm
			reduce(184), // &&, reduce: AddsubdotTerm
			reduce(184), // ??, reduce: AddsubdotTerm
			reduce(184), // ???, reduce: AddsubdotTerm
			reduce(184), // =~, reduce: AddsubdotTerm
			reduce(184), // !=~, reduce: AddsubdotTerm
			reduce(184), // ==, reduce: AddsubdotTerm
			reduce(184), // !=, reduce: AddsubdotTerm
			reduce(184), // <=>, reduce: AddsubdotTerm
			reduce(184), // >=, reduce: AddsubdotTerm
			reduce(184), // <, reduce: AddsubdotTerm
			reduce(184), // <=, reduce: AddsubdotTerm
			reduce(184), // ^, reduce: AddsubdotTerm
			reduce(184), // &, reduce: AddsubdotTerm
			reduce(184), // <<, reduce: AddsubdotTerm
			reduce(184), // >>>, reduce: AddsubdotTerm
			reduce(184), // +, reduce: AddsubdotTerm
			reduce(184), // -, reduce: AddsubdotTerm
			reduce(184), // .+, reduce: AddsubdotTerm
			reduce(184), // .-, reduce: AddsubdotTerm
			shift(1387), // *
			shift(1388), // /
			shift(1389), // //
			shift(1390), // %
			shift(1391), // .*
			shift(1392), // ./
			shift(1393), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2551
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(185), // >, reduce: AddsubdotTerm
			reduce(185), // >>, reduce: AddsubdotTerm
			reduce(185), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(185), // ), reduce: AddsubdotTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(185), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(185), // ||, reduce: AddsubdotTerm
			reduce(185), // ^^, reduce: AddsubdotTerm
			reduce(185), // &&, reduce: AddsubdotTerm
			reduce(185), // ??, reduce: AddsubdotTerm
			reduce(185), // ???, reduce: AddsubdotTerm
			reduce(185), // =~, reduce: AddsubdotTerm
			reduce(185), // !=~, reduce: AddsubdotTerm
			reduce(185), // ==, reduce: AddsubdotTerm
			reduce(185), // !=, reduce: AddsubdotTerm
			reduce(185), // <=>, reduce: AddsubdotTerm
			reduce(185), // >=, reduce: AddsubdotTerm
			reduce(185), // <, reduce: AddsubdotTerm
			reduce(185), // <=, reduce: AddsubdotTerm
			reduce(185), // ^, reduce: AddsubdotTerm
			reduce(185), // &, reduce: AddsubdotTerm
			reduce(185), // <<, reduce: AddsubdotTerm
			reduce(185), // >>>, reduce: AddsubdotTerm
			reduce(185), // +, reduce: AddsubdotTerm
			reduce(185), // -, reduce: AddsubdotTerm
			reduce(185), // .+, reduce: AddsubdotTerm
			reduce(185), // .-, reduce: AddsubdotTerm
			shift(1387), // *
			shift(1388), // /
			shift(1389), // //
			shift(1390), // %
			shift(1391), // .*
			shift(1392), // ./
			shift(1393), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2552
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(187), // >, reduce: MuldivTerm
			reduce(187), // >>, reduce: MuldivTerm
			reduce(187), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(187), // ), reduce: MuldivTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(187), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(187), // ||, reduce: MuldivTerm
			reduce(187), // ^^, reduce: MuldivTerm
			reduce(187), // &&, reduce: MuldivTerm
			reduce(187), // ??, reduce: MuldivTerm
			reduce(187), // ???, reduce: MuldivTerm
			reduce(187), // =~, reduce: MuldivTerm
			reduce(187), // !=~, reduce: MuldivTerm
			reduce(187), // ==, reduce: MuldivTerm
			reduce(187), // !=, reduce: MuldivTerm
			reduce(187), // <=>, reduce: MuldivTerm
			reduce(187), // >=, reduce: MuldivTerm
			reduce(187), // <, reduce: MuldivTerm
			reduce(187), // <=, reduce: MuldivTerm
			reduce(187), // ^, reduce: MuldivTerm
			reduce(187), // &, reduce: MuldivTerm
			reduce(187), // <<, reduce: MuldivTerm
			reduce(187), // >>>, reduce: MuldivTerm
			reduce(187), // +, reduce: MuldivTerm
			reduce(187), // -, reduce: MuldivTerm
			reduce(187), // .+, reduce: MuldivTerm
			reduce(187), // .-, reduce: MuldivTerm
			reduce(187), // *, reduce: MuldivTerm
			reduce(187), // /, reduce: MuldivTerm
			reduce(187), // //, reduce: MuldivTerm
			reduce(187), // %, reduce: MuldivTerm
			reduce(187), // .*, reduce: MuldivTerm
			reduce(187), // ./, reduce: MuldivTerm
			reduce(187), // .//, reduce: MuldivTerm
			shift(1397), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2553
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(188), // >, reduce: MuldivTerm
			reduce(188), // >>, reduce: MuldivTerm
			reduce(188), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(188), // ), reduce: MuldivTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(188), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(188), // ||, reduce: MuldivTerm
			reduce(188), // ^^, reduce: MuldivTerm
			reduce(188), // &&, reduce: MuldivTerm
			reduce(188), // ??, reduce: MuldivTerm
			reduce(188), // ???, reduce: MuldivTerm
			reduce(188), // =~, reduce: MuldivTerm
			reduce(188), // !=~, reduce: MuldivTerm
			reduce(188), // ==, reduce: MuldivTerm
			reduce(188), // !=, reduce: MuldivTerm
			reduce(188), // <=>, reduce: MuldivTerm
			reduce(188), // >=, reduce: MuldivTerm
			reduce(188), // <, reduce: MuldivTerm
			reduce(188), // <=, reduce: MuldivTerm
			reduce(188), // ^, reduce: MuldivTerm
			reduce(188), // &, reduce: MuldivTerm
			reduce(188), // <<, reduce: MuldivTerm
			reduce(188), // >>>, reduce: MuldivTerm
			reduce(188), // +, reduce: MuldivTerm
			reduce(188), // -, reduce: MuldivTerm
			reduce(188), // .+, reduce: MuldivTerm
			reduce(188), // .-, reduce: MuldivTerm
			reduce(188), // *, reduce: MuldivTerm
			reduce(188), // /, reduce: MuldivTerm
			reduce(188), // //, reduce: MuldivTerm
			reduce(188), // %, reduce: MuldivTerm
			reduce(188), // .*, reduce: MuldivTerm
			reduce(188), // ./, reduce: MuldivTerm
			reduce(188), // .//, reduce: MuldivTerm
			shift(1397), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2554
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(189), // >, reduce: MuldivTerm
			reduce(189), // >>, reduce: MuldivTerm
			reduce(189), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(189), // ), reduce: MuldivTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(189), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(189), // ||, reduce: MuldivTerm
			reduce(189), // ^^, reduce: MuldivTerm
			reduce(189), // &&, reduce: MuldivTerm
			reduce(189), // ??, reduce: MuldivTerm
			reduce(189), // ???, reduce: MuldivTerm
			reduce(189), // =~, reduce: MuldivTerm
			reduce(189), // !=~, reduce: MuldivTerm
			reduce(189), // ==, reduce: MuldivTerm
			reduce(189), // !=, reduce: MuldivTerm
			reduce(189), // <=>, reduce: MuldivTerm
			reduce(189), // >=, reduce: MuldivTerm
			reduce(189), // <, reduce: MuldivTerm
			reduce(189), // <=, reduce: MuldivTerm
			reduce(189), // ^, reduce: MuldivTerm
			reduce(189), // &, reduce: MuldivTerm
			reduce(189), // <<, reduce: MuldivTerm
			reduce(189), // >>>, reduce: MuldivTerm
			reduce(189), // +, reduce: MuldivTerm
			reduce(189), // -, reduce: MuldivTerm
			reduce(189), // .+, reduce: MuldivTerm
			reduce(189), // .-, reduce: MuldivTerm
			reduce(189), // *, reduce: MuldivTerm
			reduce(189), // /, reduce: MuldivTerm
			reduce(189), // //, reduce: MuldivTerm
			reduce(189), // %, reduce: MuldivTerm
			reduce(189), // .*, reduce: MuldivTerm
			reduce(189), // ./, reduce: MuldivTerm
			reduce(189), // .//, reduce: MuldivTerm
			shift(1397), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2555
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(190), // >, reduce: MuldivTerm
			reduce(190), // >>, reduce: MuldivTerm
			reduce(190), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(190), // ), reduce: MuldivTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(190), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(190), // ||, reduce: MuldivTerm
			reduce(190), // ^^, reduce: MuldivTerm
			reduce(190), // &&, reduce: MuldivTerm
			reduce(190), // ??, reduce: MuldivTerm
			reduce(190), // ???, reduce: MuldivTerm
			reduce(190), // =~, reduce: MuldivTerm
			reduce(190), // !=~, reduce: MuldivTerm
			reduce(190), // ==, reduce: MuldivTerm
			reduce(190), // !=, reduce: MuldivTerm
			reduce(190), // <=>, reduce: MuldivTerm
			reduce(190), // >=, reduce: MuldivTerm
			reduce(190), // <, reduce: MuldivTerm
			reduce(190), // <=, reduce: MuldivTerm
			reduce(190), // ^, reduce: MuldivTerm
			reduce(190), // &, reduce: MuldivTerm
			reduce(190), // <<, reduce: MuldivTerm
			reduce(190), // >>>, reduce: MuldivTerm
			reduce(190), // +, reduce: MuldivTerm
			reduce(190), // -, reduce: MuldivTerm
			reduce(190), // .+, reduce: MuldivTerm
			reduce(190), // .-, reduce: MuldivTerm
			reduce(190), // *, reduce: MuldivTerm
			reduce(190), // /, reduce: MuldivTerm
			reduce(190), // //, reduce: MuldivTerm
			reduce(190), // %, reduce: MuldivTerm
			reduce(190), // .*, reduce: MuldivTerm
			reduce(190), // ./, reduce: MuldivTerm
			reduce(190), // .//, reduce: MuldivTerm
			shift(1397), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2556
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(191), // >, reduce: MuldivTerm
			reduce(191), // >>, reduce: MuldivTerm
			reduce(191), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(191), // ), reduce: MuldivTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(191), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(191), // ||, reduce: MuldivTerm
			reduce(191), // ^^, reduce: MuldivTerm
			reduce(191), // &&, reduce: MuldivTerm
			reduce(191), // ??, reduce: MuldivTerm
			reduce(191), // ???, reduce: MuldivTerm
			reduce(191), // =~, reduce: MuldivTerm
			reduce(191), // !=~, reduce: MuldivTerm
			reduce(191), // ==, reduce: MuldivTerm
			reduce(191), // !=, reduce: MuldivTerm
			reduce(191), // <=>, reduce: MuldivTerm
			reduce(191), // >=, reduce: MuldivTerm
			reduce(191), // <, reduce: MuldivTerm
			reduce(191), // <=, reduce: MuldivTerm
			reduce(191), // ^, reduce: MuldivTerm
			reduce(191), // &, reduce: MuldivTerm
			reduce(191), // <<, reduce: MuldivTerm
			reduce(191), // >>>, reduce: MuldivTerm
			reduce(191), // +, reduce: MuldivTerm
			reduce(191), // -, reduce: MuldivTerm
			reduce(191), // .+, reduce: MuldivTerm
			reduce(191), // .-, reduce: MuldivTerm
			reduce(191), // *, reduce: MuldivTerm
			reduce(191), // /, reduce: MuldivTerm
			reduce(191), // //, reduce: MuldivTerm
			reduce(191), // %, reduce: MuldivTerm
			reduce(191), // .*, reduce: MuldivTerm
			reduce(191), // ./, reduce: MuldivTerm
			reduce(191), // .//, reduce: MuldivTerm
			shift(1397), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2557
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(192), // >, reduce: MuldivTerm
			reduce(192), // >>, reduce: MuldivTerm
			reduce(192), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(192), // ), reduce: MuldivTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(192), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(192), // ||, reduce: MuldivTerm
			reduce(192), // ^^, reduce: MuldivTerm
			reduce(192), // &&, reduce: MuldivTerm
			reduce(192), // ??, reduce: MuldivTerm
			reduce(192), // ???, reduce: MuldivTerm
			reduce(192), // =~, reduce: MuldivTerm
			reduce(192), // !=~, reduce: MuldivTerm
			reduce(192), // ==, reduce: MuldivTerm
			reduce(192), // !=, reduce: MuldivTerm
			reduce(192), // <=>, reduce: MuldivTerm
			reduce(192), // >=, reduce: MuldivTerm
			reduce(192), // <, reduce: MuldivTerm
			reduce(192), // <=, reduce: MuldivTerm
			reduce(192), // ^, reduce: MuldivTerm
			reduce(192), // &, reduce: MuldivTerm
			reduce(192), // <<, reduce: MuldivTerm
			reduce(192), // >>>, reduce: MuldivTerm
			reduce(192), // +, reduce: MuldivTerm
			reduce(192), // -, reduce: MuldivTerm
			reduce(192), // .+, reduce: MuldivTerm
			reduce(192), // .-, reduce: MuldivTerm
			reduce(192), // *, reduce: MuldivTerm
			reduce(192), // /, reduce: MuldivTerm
			reduce(192), // //, reduce: MuldivTerm
			reduce(192), // %, reduce: MuldivTerm
			reduce(192), // .*, reduce: MuldivTerm
			reduce(192), // ./, reduce: MuldivTerm
			reduce(192), // .//, reduce: MuldivTerm
			shift(1397), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2558
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(193), // >, reduce: MuldivTerm
			reduce(193), // >>, reduce: MuldivTerm
			reduce(193), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(193), // ), reduce: MuldivTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(193), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(193), // ||, reduce: MuldivTerm
			reduce(193), // ^^, reduce: MuldivTerm
			reduce(193), // &&, reduce: MuldivTerm
			reduce(193), // ??, reduce: MuldivTerm
			reduce(193), // ???, reduce: MuldivTerm
			reduce(193), // =~, reduce: MuldivTerm
			reduce(193), // !=~, reduce: MuldivTerm
			reduce(193), // ==, reduce: MuldivTerm
			reduce(193), // !=, reduce: MuldivTerm
			reduce(193), // <=>, reduce: MuldivTerm
			reduce(193), // >=, reduce: MuldivTerm
			reduce(193), // <, reduce: MuldivTerm
			reduce(193), // <=, reduce: MuldivTerm
			reduce(193), // ^, reduce: MuldivTerm
			reduce(193), // &, reduce: MuldivTerm
			reduce(193), // <<, reduce: MuldivTerm
			reduce(193), // >>>, reduce: MuldivTerm
			reduce(193), // +, reduce: MuldivTerm
			reduce(193), // -, reduce: MuldivTerm
			reduce(193), // .+, reduce: MuldivTerm
			reduce(193), // .-, reduce: MuldivTerm
			reduce(193), // *, reduce: MuldivTerm
			reduce(193), // /, reduce: MuldivTerm
			reduce(193), // //, reduce: MuldivTerm
			reduce(193), // %, reduce: MuldivTerm
			reduce(193), // .*, reduce: MuldivTerm
			reduce(193), // ./, reduce: MuldivTerm
			reduce(193), // .//, reduce: MuldivTerm
			shift(1397), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2559
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(195), // >, reduce: DotTerm
			reduce(195), // >>, reduce: DotTerm
			reduce(195), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(195), // ), reduce: DotTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(195), // ?, reduce: DotTerm
			nil,         // :
			reduce(195), // ||, reduce: DotTerm
			reduce(195), // ^^, reduce: DotTerm
			reduce(195), // &&, reduce: DotTerm
			reduce(195), // ??, reduce: DotTerm
			reduce(195), // ???, reduce: DotTerm
			reduce(195), // =~, reduce: DotTerm
			reduce(195), // !=~, reduce: DotTerm
			reduce(195), // ==, reduce: DotTerm
			reduce(195), // !=, reduce: DotTerm
			reduce(195), // <=>, reduce: DotTerm
			reduce(195), // >=, reduce: DotTerm
			reduce(195), // <, reduce: DotTerm
			reduce(195), // <=, reduce: DotTerm
			reduce(195), // ^, reduce: DotTerm
			reduce(195), // &, reduce: DotTerm
			reduce(195), // <<, reduce: DotTerm
			reduce(195), // >>>, reduce: DotTerm
			reduce(195), // +, reduce: DotTerm
			reduce(195), // -, reduce: DotTerm
			reduce(195), // .+, reduce: DotTerm
			reduce(195), // .-, reduce: DotTerm
			reduce(195), // *, reduce: DotTerm
			reduce(195), // /, reduce: DotTerm
			reduce(195), // //, reduce: DotTerm
			reduce(195), // %, reduce: DotTerm
			reduce(195), // .*, reduce: DotTerm
			reduce(195), // ./, reduce: DotTerm
			reduce(195), // .//, reduce: DotTerm
			reduce(195), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2560
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2561
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2562
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(204), // >, reduce: PowTerm
			reduce(204), // >>, reduce: PowTerm
			reduce(204), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(204), // ), reduce: PowTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(204), // ?, reduce: PowTerm
			nil,         // :
			reduce(204), // ||, reduce: PowTerm
			reduce(204), // ^^, reduce: PowTerm
			reduce(204), // &&, reduce: PowTerm
			reduce(204), // ??, reduce: PowTerm
			reduce(204), // ???, reduce: PowTerm
			reduce(204), // =~, reduce: PowTerm
			reduce(204), // !=~, reduce: PowTerm
			reduce(204), // ==, reduce: PowTerm
			reduce(204), // !=, reduce: PowTerm
			reduce(204), // <=>, reduce: PowTerm
			reduce(204), // >=, reduce: PowTerm
			reduce(204), // <, reduce: PowTerm
			reduce(204), // <=, reduce: PowTerm
			reduce(204), // ^, reduce: PowTerm
			reduce(204), // &, reduce: PowTerm
			reduce(204), // <<, reduce: PowTerm
			reduce(204), // >>>, reduce: PowTerm
			reduce(204), // +, reduce: PowTerm
			reduce(204), // -, reduce: PowTerm
			reduce(204), // .+, reduce: PowTerm
			reduce(204), // .-, reduce: PowTerm
			reduce(204), // *, reduce: PowTerm
			reduce(204), // /, reduce: PowTerm
			reduce(204), // //, reduce: PowTerm
			reduce(204), // %, reduce: PowTerm
			reduce(204), // .*, reduce: PowTerm
			reduce(204), // ./, reduce: PowTerm
			reduce(204), // .//, reduce: PowTerm
			reduce(204), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2563
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3291), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3292), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2564
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(3294), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2565
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3295), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2566
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3296), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2567
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(229), // >, reduce: ArrayLiteral
			reduce(229), // >>, reduce: ArrayLiteral
			reduce(229), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(229), // ), reduce: ArrayLiteral
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(229), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(229), // ||, reduce: ArrayLiteral
			reduce(229), // ^^, reduce: ArrayLiteral
			reduce(229), // &&, reduce: ArrayLiteral
			reduce(229), // ??, reduce: ArrayLiteral
			reduce(229), // ???, reduce: ArrayLiteral
			reduce(229), // =~, reduce: ArrayLiteral
			reduce(229), // !=~, reduce: ArrayLiteral
			reduce(229), // ==, reduce: ArrayLiteral
			reduce(229), // !=, reduce: ArrayLiteral
			reduce(229), // <=>, reduce: ArrayLiteral
			reduce(229), // >=, reduce: ArrayLiteral
			reduce(229), // <, reduce: ArrayLiteral
			reduce(229), // <=, reduce: ArrayLiteral
			reduce(229), // ^, reduce: ArrayLiteral
			reduce(229), // &, reduce: ArrayLiteral
			reduce(229), // <<, reduce: ArrayLiteral
			reduce(229), // >>>, reduce: ArrayLiteral
			reduce(229), // +, reduce: ArrayLiteral
			reduce(229), // -, reduce: ArrayLiteral
			reduce(229), // .+, reduce: ArrayLiteral
			reduce(229), // .-, reduce: ArrayLiteral
			reduce(229), // *, reduce: ArrayLiteral
			reduce(229), // /, reduce: ArrayLiteral
			reduce(229), // //, reduce: ArrayLiteral
			reduce(229), // %, reduce: ArrayLiteral
			reduce(229), // .*, reduce: ArrayLiteral
			reduce(229), // ./, reduce: ArrayLiteral
			reduce(229), // .//, reduce: ArrayLiteral
			reduce(229), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(229), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(229), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(229), // [[, reduce: ArrayLiteral
			reduce(229), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2568
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(255), // >, reduce: ENV
			reduce(255), // >>, reduce: ENV
			reduce(255), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(255), // ), reduce: ENV
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(255), // ?, reduce: ENV
			nil,         // :
			reduce(255), // ||, reduce: ENV
			reduce(255), // ^^, reduce: ENV
			reduce(255), // &&, reduce: ENV
			reduce(255), // ??, reduce: ENV
			reduce(255), // ???, reduce: ENV
			reduce(255), // =~, reduce: ENV
			reduce(255), // !=~, reduce: ENV
			reduce(255), // ==, reduce: ENV
			reduce(255), // !=, reduce: ENV
			reduce(255), // <=>, reduce: ENV
			reduce(255), // >=, reduce: ENV
			reduce(255), // <, reduce: ENV
			reduce(255), // <=, reduce: ENV
			reduce(255), // ^, reduce: ENV
			reduce(255), // &, reduce: ENV
			reduce(255), // <<, reduce: ENV
			reduce(255), // >>>, reduce: ENV
			reduce(255), // +, reduce: ENV
			reduce(255), // -, reduce: ENV
			reduce(255), // .+, reduce: ENV
			reduce(255), // .-, reduce: ENV
			reduce(255), // *, reduce: ENV
			reduce(255), // /, reduce: ENV
			reduce(255), // //, reduce: ENV
			reduce(255), // %, reduce: ENV
			reduce(255), // .*, reduce: ENV
			reduce(255), // ./, reduce: ENV
			reduce(255), // .//, reduce: ENV
			reduce(255), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(255), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(255), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(255), // [[, reduce: ENV
			reduce(255), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2569
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3297), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2570
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3298), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2571
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(268), // >, reduce: FunctionCallsite
			reduce(268), // >>, reduce: FunctionCallsite
			reduce(268), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(268), // ), reduce: FunctionCallsite
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(268), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(268), // ||, reduce: FunctionCallsite
			reduce(268), // ^^, reduce: FunctionCallsite
			reduce(268), // &&, reduce: FunctionCallsite
			reduce(268), // ??, reduce: FunctionCallsite
			reduce(268), // ???, reduce: FunctionCallsite
			reduce(268), // =~, reduce: FunctionCallsite
			reduce(268), // !=~, reduce: FunctionCallsite
			reduce(268), // ==, reduce: FunctionCallsite
			reduce(268), // !=, reduce: FunctionCallsite
			reduce(268), // <=>, reduce: FunctionCallsite
			reduce(268), // >=, reduce: FunctionCallsite
			reduce(268), // <, reduce: FunctionCallsite
			reduce(268), // <=, reduce: FunctionCallsite
			reduce(268), // ^, reduce: FunctionCallsite
			reduce(268), // &, reduce: FunctionCallsite
			reduce(268), // <<, reduce: FunctionCallsite
			reduce(268), // >>>, reduce: FunctionCallsite
			reduce(268), // +, reduce: FunctionCallsite
			reduce(268), // -, reduce: FunctionCallsite
			reduce(268), // .+, reduce: FunctionCallsite
			reduce(268), // .-, reduce: FunctionCallsite
			reduce(268), // *, reduce: FunctionCallsite
			reduce(268), // /, reduce: FunctionCallsite
			reduce(268), // //, reduce: FunctionCallsite
			reduce(268), // %, reduce: FunctionCallsite
			reduce(268), // .*, reduce: FunctionCallsite
			reduce(268), // ./, reduce: FunctionCallsite
			reduce(268), // .//, reduce: FunctionCallsite
			reduce(268), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(268), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(268), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(268), // [[, reduce: FunctionCallsite
			reduce(268), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2572
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3299), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2573
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(77),  // $, reduce: EmitPStatement
			nil,         // empty
			reduce(77),  // ;, reduce: EmitPStatement
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3300), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2574
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1330), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			shift(1343), // field_name
			shift(1344), // $[
			nil,         // ]
			shift(1345), // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(1346), // full_srec
			shift(1347), // oosvar_name
			shift(1348), // @[
			shift(1349), // braced_oosvar_name
			shift(1350), // full_oosvar
			shift(1351), // all
			shift(1352), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2575
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(80), // $, reduce: EmitPStatement
			nil,        // empty
			reduce(80), // ;, reduce: EmitPStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2576
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(78),  // $, reduce: EmitPStatement
			nil,         // empty
			reduce(78),  // ;, reduce: EmitPStatement
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3302), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2577
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(235), // >, reduce: MapLiteral
			reduce(235), // >>, reduce: MapLiteral
			reduce(235), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(235), // ], reduce: MapLiteral
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(235), // ?, reduce: MapLiteral
			nil,         // :
			reduce(235), // ||, reduce: MapLiteral
			reduce(235), // ^^, reduce: MapLiteral
			reduce(235), // &&, reduce: MapLiteral
			reduce(235), // ??, reduce: MapLiteral
			reduce(235), // ???, reduce: MapLiteral
			reduce(235), // =~, reduce: MapLiteral
			reduce(235), // !=~, reduce: MapLiteral
			reduce(235), // ==, reduce: MapLiteral
			reduce(235), // !=, reduce: MapLiteral
			reduce(235), // <=>, reduce: MapLiteral
			reduce(235), // >=, reduce: MapLiteral
			reduce(235), // <, reduce: MapLiteral
			reduce(235), // <=, reduce: MapLiteral
			reduce(235), // ^, reduce: MapLiteral
			reduce(235), // &, reduce: MapLiteral
			reduce(235), // <<, reduce: MapLiteral
			reduce(235), // >>>, reduce: MapLiteral
			reduce(235), // +, reduce: MapLiteral
			reduce(235), // -, reduce: MapLiteral
			reduce(235), // .+, reduce: MapLiteral
			reduce(235), // .-, reduce: MapLiteral
			reduce(235), // *, reduce: MapLiteral
			reduce(235), // /, reduce: MapLiteral
			reduce(235), // //, reduce: MapLiteral
			reduce(235), // %, reduce: MapLiteral
			reduce(235), // .*, reduce: MapLiteral
			reduce(235), // ./, reduce: MapLiteral
			reduce(235), // .//, reduce: MapLiteral
			reduce(235), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(235), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(235), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(235), // [[, reduce: MapLiteral
			reduce(235), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2578
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(208), // >, reduce: PrecedenceChainEnd
			reduce(208), // >>, reduce: PrecedenceChainEnd
			reduce(208), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(208), // ], reduce: PrecedenceChainEnd
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(208), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(208), // ||, reduce: PrecedenceChainEnd
			reduce(208), // ^^, reduce: PrecedenceChainEnd
			reduce(208), // &&, reduce: PrecedenceChainEnd
			reduce(208), // ??, reduce: PrecedenceChainEnd
			reduce(208), // ???, reduce: PrecedenceChainEnd
			reduce(208), // =~, reduce: PrecedenceChainEnd
			reduce(208), // !=~, reduce: PrecedenceChainEnd
			reduce(208), // ==, reduce: PrecedenceChainEnd
			reduce(208), // !=, reduce: PrecedenceChainEnd
			reduce(208), // <=>, reduce: PrecedenceChainEnd
			reduce(208), // >=, reduce: PrecedenceChainEnd
			reduce(208), // <, reduce: PrecedenceChainEnd
			reduce(208), // <=, reduce: PrecedenceChainEnd
			reduce(208), // ^, reduce: PrecedenceChainEnd
			reduce(208), // &, reduce: PrecedenceChainEnd
			reduce(208), // <<, reduce: PrecedenceChainEnd
			reduce(208), // >>>, reduce: PrecedenceChainEnd
			reduce(208), // +, reduce: PrecedenceChainEnd
			reduce(208), // -, reduce: PrecedenceChainEnd
			reduce(208), // .+, reduce: PrecedenceChainEnd
			reduce(208), // .-, reduce: PrecedenceChainEnd
			reduce(208), // *, reduce: PrecedenceChainEnd
			reduce(208), // /, reduce: PrecedenceChainEnd
			reduce(208), // //, reduce: PrecedenceChainEnd
			reduce(208), // %, reduce: PrecedenceChainEnd
			reduce(208), // .*, reduce: PrecedenceChainEnd
			reduce(208), // ./, reduce: PrecedenceChainEnd
			reduce(208), // .//, reduce: PrecedenceChainEnd
			reduce(208), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(208), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2579
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(105), // >, reduce: IndirectFieldValue
			reduce(105), // >>, reduce: IndirectFieldValue
			reduce(105), // |, reduce: IndirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(105), // ], reduce: IndirectFieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(105), // ?, reduce: IndirectFieldValue
			nil,         // :
			reduce(105), // ||, reduce: IndirectFieldValue
			reduce(105), // ^^, reduce: IndirectFieldValue
			reduce(105), // &&, reduce: IndirectFieldValue
			reduce(105), // ??, reduce: IndirectFieldValue
			reduce(105), // ???, reduce: IndirectFieldValue
			reduce(105), // =~, reduce: IndirectFieldValue
			reduce(105), // !=~, reduce: IndirectFieldValue
			reduce(105), // ==, reduce: IndirectFieldValue
			reduce(105), // !=, reduce: IndirectFieldValue
			reduce(105), // <=>, reduce: IndirectFieldValue
			reduce(105), // >=, reduce: IndirectFieldValue
			reduce(105), // <, reduce: IndirectFieldValue
			reduce(105), // <=, reduce: IndirectFieldValue
			reduce(105), // ^, reduce: IndirectFieldValue
			reduce(105), // &, reduce: IndirectFieldValue
			reduce(105), // <<, reduce: IndirectFieldValue
			reduce(105), // >>>, reduce: IndirectFieldValue
			reduce(105), // +, reduce: IndirectFieldValue
			reduce(105), // -, reduce: IndirectFieldValue
			reduce(105), // .+, reduce: IndirectFieldValue
			reduce(105), // .-, reduce: IndirectFieldValue
			reduce(105), // *, reduce: IndirectFieldValue
			reduce(105), // /, reduce: IndirectFieldValue
			reduce(105), // //, reduce: IndirectFieldValue
			reduce(105), // %, reduce: IndirectFieldValue
			reduce(105), // .*, reduce: IndirectFieldValue
			reduce(105), // ./, reduce: IndirectFieldValue
			reduce(105), // .//, reduce: IndirectFieldValue
			reduce(105), // ., reduce: IndirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(105), // **, reduce: IndirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(105), // [, reduce: IndirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(105), // [[, reduce: IndirectFieldValue
			reduce(105), // [[[, reduce: IndirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2580
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3303), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2581
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3304), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2582
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(114), // >, reduce: IndirectOosvarValue
			reduce(114), // >>, reduce: IndirectOosvarValue
			reduce(114), // |, reduce: IndirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(114), // ], reduce: IndirectOosvarValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(114), // ?, reduce: IndirectOosvarValue
			nil,         // :
			reduce(114), // ||, reduce: IndirectOosvarValue
			reduce(114), // ^^, reduce: IndirectOosvarValue
			reduce(114), // &&, reduce: IndirectOosvarValue
			reduce(114), // ??, reduce: IndirectOosvarValue
			reduce(114), // ???, reduce: IndirectOosvarValue
			reduce(114), // =~, reduce: IndirectOosvarValue
			reduce(114), // !=~, reduce: IndirectOosvarValue
			reduce(114), // ==, reduce: IndirectOosvarValue
			reduce(114), // !=, reduce: IndirectOosvarValue
			reduce(114), // <=>, reduce: IndirectOosvarValue
			reduce(114), // >=, reduce: IndirectOosvarValue
			reduce(114), // <, reduce: IndirectOosvarValue
			reduce(114), // <=, reduce: IndirectOosvarValue
			reduce(114), // ^, reduce: IndirectOosvarValue
			reduce(114), // &, reduce: IndirectOosvarValue
			reduce(114), // <<, reduce: IndirectOosvarValue
			reduce(114), // >>>, reduce: IndirectOosvarValue
			reduce(114), // +, reduce: IndirectOosvarValue
			reduce(114), // -, reduce: IndirectOosvarValue
			reduce(114), // .+, reduce: IndirectOosvarValue
			reduce(114), // .-, reduce: IndirectOosvarValue
			reduce(114), // *, reduce: IndirectOosvarValue
			reduce(114), // /, reduce: IndirectOosvarValue
			reduce(114), // //, reduce: IndirectOosvarValue
			reduce(114), // %, reduce: IndirectOosvarValue
			reduce(114), // .*, reduce: IndirectOosvarValue
			reduce(114), // ./, reduce: IndirectOosvarValue
			reduce(114), // .//, reduce: IndirectOosvarValue
			reduce(114), // ., reduce: IndirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(114), // **, reduce: IndirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(114), // [, reduce: IndirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(114), // [[, reduce: IndirectOosvarValue
			reduce(114), // [[[, reduce: IndirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2583
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3305), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2584
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(151), // ], reduce: LogicalOrTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(151), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(151), // ||, reduce: LogicalOrTerm
			shift(1423), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2585
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(153), // ], reduce: LogicalXORTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(153), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(153), // ||, reduce: LogicalXORTerm
			reduce(153), // ^^, reduce: LogicalXORTerm
			shift(1424), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2586
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(155), // ], reduce: LogicalAndTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(155), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(155), // ||, reduce: LogicalAndTerm
			reduce(155), // ^^, reduce: LogicalAndTerm
			reduce(155), // &&, reduce: LogicalAndTerm
			shift(1425), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2587
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(157), // ], reduce: AbsentCoalesceTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(157), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(157), // ||, reduce: AbsentCoalesceTerm
			reduce(157), // ^^, reduce: AbsentCoalesceTerm
			reduce(157), // &&, reduce: AbsentCoalesceTerm
			reduce(157), // ??, reduce: AbsentCoalesceTerm
			shift(1426), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2588
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(159), // ], reduce: EmptyCoalesceTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(159), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(159), // ||, reduce: EmptyCoalesceTerm
			reduce(159), // ^^, reduce: EmptyCoalesceTerm
			reduce(159), // &&, reduce: EmptyCoalesceTerm
			reduce(159), // ??, reduce: EmptyCoalesceTerm
			reduce(159), // ???, reduce: EmptyCoalesceTerm
			shift(1427), // =~
			shift(1428), // !=~
			shift(1429), // ==
			shift(1430), // !=
			shift(1431), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2589
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1432), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(161), // ], reduce: EqneTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(161), // ?, reduce: EqneTerm
			nil,         // :
			reduce(161), // ||, reduce: EqneTerm
			reduce(161), // ^^, reduce: EqneTerm
			reduce(161), // &&, reduce: EqneTerm
			reduce(161), // ??, reduce: EqneTerm
			reduce(161), // ???, reduce: EqneTerm
			reduce(161), // =~, reduce: EqneTerm
			reduce(161), // !=~, reduce: EqneTerm
			reduce(161), // ==, reduce: EqneTerm
			reduce(161), // !=, reduce: EqneTerm
			reduce(161), // <=>, reduce: EqneTerm
			shift(1433), // >=
			shift(1434), // <
			shift(1435), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2590
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1432), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(162), // ], reduce: EqneTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(162), // ?, reduce: EqneTerm
			nil,         // :
			reduce(162), // ||, reduce: EqneTerm
			reduce(162), // ^^, reduce: EqneTerm
			reduce(162), // &&, reduce: EqneTerm
			reduce(162), // ??, reduce: EqneTerm
			reduce(162), // ???, reduce: EqneTerm
			reduce(162), // =~, reduce: EqneTerm
			reduce(162), // !=~, reduce: EqneTerm
			reduce(162), // ==, reduce: EqneTerm
			reduce(162), // !=, reduce: EqneTerm
			reduce(162), // <=>, reduce: EqneTerm
			shift(1433), // >=
			shift(1434), // <
			shift(1435), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2591
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1432), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(163), // ], reduce: EqneTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(163), // ?, reduce: EqneTerm
			nil,         // :
			reduce(163), // ||, reduce: EqneTerm
			reduce(163), // ^^, reduce: EqneTerm
			reduce(163), // &&, reduce: EqneTerm
			reduce(163), // ??, reduce: EqneTerm
			reduce(163), // ???, reduce: EqneTerm
			reduce(163), // =~, reduce: EqneTerm
			reduce(163), // !=~, reduce: EqneTerm
			reduce(163), // ==, reduce: EqneTerm
			reduce(163), // !=, reduce: EqneTerm
			reduce(163), // <=>, reduce: EqneTerm
			shift(1433), // >=
			shift(1434), // <
			shift(1435), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2592
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1432), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(164), // ], reduce: EqneTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(164), // ?, reduce: EqneTerm
			nil,         // :
			reduce(164), // ||, reduce: EqneTerm
			reduce(164), // ^^, reduce: EqneTerm
			reduce(164), // &&, reduce: EqneTerm
			reduce(164), // ??, reduce: EqneTerm
			reduce(164), // ???, reduce: EqneTerm
			reduce(164), // =~, reduce: EqneTerm
			reduce(164), // !=~, reduce: EqneTerm
			reduce(164), // ==, reduce: EqneTerm
			reduce(164), // !=, reduce: EqneTerm
			reduce(164), // <=>, reduce: EqneTerm
			shift(1433), // >=
			shift(1434), // <
			shift(1435), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2593
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1432), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(165), // ], reduce: EqneTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(165), // ?, reduce: EqneTerm
			nil,         // :
			reduce(165), // ||, reduce: EqneTerm
			reduce(165), // ^^, reduce: EqneTerm
			reduce(165), // &&, reduce: EqneTerm
			reduce(165), // ??, reduce: EqneTerm
			reduce(165), // ???, reduce: EqneTerm
			reduce(165), // =~, reduce: EqneTerm
			reduce(165), // !=~, reduce: EqneTerm
			reduce(165), // ==, reduce: EqneTerm
			reduce(165), // !=, reduce: EqneTerm
			reduce(165), // <=>, reduce: EqneTerm
			shift(1433), // >=
			shift(1434), // <
			shift(1435), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2594
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(167), // >, reduce: CmpTerm
			nil,         // >>
			shift(1436), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(167), // ], reduce: CmpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(167), // ?, reduce: CmpTerm
			nil,         // :
			reduce(167), // ||, reduce: CmpTerm
			reduce(167), // ^^, reduce: CmpTerm
			reduce(167), // &&, reduce: CmpTerm
			reduce(167), // ??, reduce: CmpTerm
			reduce(167), // ???, reduce: CmpTerm
			reduce(167), // =~, reduce: CmpTerm
			reduce(167), // !=~, reduce: CmpTerm
			reduce(167), // ==, reduce: CmpTerm
			reduce(167), // !=, reduce: CmpTerm
			reduce(167), // <=>, reduce: CmpTerm
			reduce(167), // >=, reduce: CmpTerm
			reduce(167), // <, reduce: CmpTerm
			reduce(167), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2595
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(168), // >, reduce: CmpTerm
			nil,         // >>
			shift(1436), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(168), // ], reduce: CmpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(168), // ?, reduce: CmpTerm
			nil,         // :
			reduce(168), // ||, reduce: CmpTerm
			reduce(168), // ^^, reduce: CmpTerm
			reduce(168), // &&, reduce: CmpTerm
			reduce(168), // ??, reduce: CmpTerm
			reduce(168), // ???, reduce: CmpTerm
			reduce(168), // =~, reduce: CmpTerm
			reduce(168), // !=~, reduce: CmpTerm
			reduce(168), // ==, reduce: CmpTerm
			reduce(168), // !=, reduce: CmpTerm
			reduce(168), // <=>, reduce: CmpTerm
			reduce(168), // >=, reduce: CmpTerm
			reduce(168), // <, reduce: CmpTerm
			reduce(168), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2596
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(169), // >, reduce: CmpTerm
			nil,         // >>
			shift(1436), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(169), // ], reduce: CmpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(169), // ?, reduce: CmpTerm
			nil,         // :
			reduce(169), // ||, reduce: CmpTerm
			reduce(169), // ^^, reduce: CmpTerm
			reduce(169), // &&, reduce: CmpTerm
			reduce(169), // ??, reduce: CmpTerm
			reduce(169), // ???, reduce: CmpTerm
			reduce(169), // =~, reduce: CmpTerm
			reduce(169), // !=~, reduce: CmpTerm
			reduce(169), // ==, reduce: CmpTerm
			reduce(169), // !=, reduce: CmpTerm
			reduce(169), // <=>, reduce: CmpTerm
			reduce(169), // >=, reduce: CmpTerm
			reduce(169), // <, reduce: CmpTerm
			reduce(169), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2597
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(170), // >, reduce: CmpTerm
			nil,         // >>
			shift(1436), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(170), // ], reduce: CmpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(170), // ?, reduce: CmpTerm
			nil,         // :
			reduce(170), // ||, reduce: CmpTerm
			reduce(170), // ^^, reduce: CmpTerm
			reduce(170), // &&, reduce: CmpTerm
			reduce(170), // ??, reduce: CmpTerm
			reduce(170), // ???, reduce: CmpTerm
			reduce(170), // =~, reduce: CmpTerm
			reduce(170), // !=~, reduce: CmpTerm
			reduce(170), // ==, reduce: CmpTerm
			reduce(170), // !=, reduce: CmpTerm
			reduce(170), // <=>, reduce: CmpTerm
			reduce(170), // >=, reduce: CmpTerm
			reduce(170), // <, reduce: CmpTerm
			reduce(170), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2598
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(172), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(172), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(172), // ], reduce: BitwiseORTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(172), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(172), // ||, reduce: BitwiseORTerm
			reduce(172), // ^^, reduce: BitwiseORTerm
			reduce(172), // &&, reduce: BitwiseORTerm
			reduce(172), // ??, reduce: BitwiseORTerm
			reduce(172), // ???, reduce: BitwiseORTerm
			reduce(172), // =~, reduce: BitwiseORTerm
			reduce(172), // !=~, reduce: BitwiseORTerm
			reduce(172), // ==, reduce: BitwiseORTerm
			reduce(172), // !=, reduce: BitwiseORTerm
			reduce(172), // <=>, reduce: BitwiseORTerm
			reduce(172), // >=, reduce: BitwiseORTerm
			reduce(172), // <, reduce: BitwiseORTerm
			reduce(172), // <=, reduce: BitwiseORTerm
			shift(1437), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2599
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(174), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(174), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(174), // ], reduce: BitwiseXORTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(174), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(174), // ||, reduce: BitwiseXORTerm
			reduce(174), // ^^, reduce: BitwiseXORTerm
			reduce(174), // &&, reduce: BitwiseXORTerm
			reduce(174), // ??, reduce: BitwiseXORTerm
			reduce(174), // ???, reduce: BitwiseXORTerm
			reduce(174), // =~, reduce: BitwiseXORTerm
			reduce(174), // !=~, reduce: BitwiseXORTerm
			reduce(174), // ==, reduce: BitwiseXORTerm
			reduce(174), // !=, reduce: BitwiseXORTerm
			reduce(174), // <=>, reduce: BitwiseXORTerm
			reduce(174), // >=, reduce: BitwiseXORTerm
			reduce(174), // <, reduce: BitwiseXORTerm
			reduce(174), // <=, reduce: BitwiseXORTerm
			reduce(174), // ^, reduce: BitwiseXORTerm
			shift(1438), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2600
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(176), // >, reduce: BitwiseANDTerm
			shift(1439), // >>
			reduce(176), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(176), // ], reduce: BitwiseANDTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(176), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(176), // ||, reduce: BitwiseANDTerm
			reduce(176), // ^^, reduce: BitwiseANDTerm
			reduce(176), // &&, reduce: BitwiseANDTerm
			reduce(176), // ??, reduce: BitwiseANDTerm
			reduce(176), // ???, reduce: BitwiseANDTerm
			reduce(176), // =~, reduce: BitwiseANDTerm
			reduce(176), // !=~, reduce: BitwiseANDTerm
			reduce(176), // ==, reduce: BitwiseANDTerm
			reduce(176), // !=, reduce: BitwiseANDTerm
			reduce(176), // <=>, reduce: BitwiseANDTerm
			reduce(176), // >=, reduce: BitwiseANDTerm
			reduce(176), // <, reduce: BitwiseANDTerm
			reduce(176), // <=, reduce: BitwiseANDTerm
			reduce(176), // ^, reduce: BitwiseANDTerm
			reduce(176), // &, reduce: BitwiseANDTerm
			shift(1440), // <<
			shift(1441), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2601
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(179), // >, reduce: BitwiseShiftTerm
			reduce(179), // >>, reduce: BitwiseShiftTerm
			reduce(179), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(179), // ], reduce: BitwiseShiftTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(179), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(179), // ||, reduce: BitwiseShiftTerm
			reduce(179), // ^^, reduce: BitwiseShiftTerm
			reduce(179), // &&, reduce: BitwiseShiftTerm
			reduce(179), // ??, reduce: BitwiseShiftTerm
			reduce(179), // ???, reduce: BitwiseShiftTerm
			reduce(179), // =~, reduce: BitwiseShiftTerm
			reduce(179), // !=~, reduce: BitwiseShiftTerm
			reduce(179), // ==, reduce: BitwiseShiftTerm
			reduce(179), // !=, reduce: BitwiseShiftTerm
			reduce(179), // <=>, reduce: BitwiseShiftTerm
			reduce(179), // >=, reduce: BitwiseShiftTerm
			reduce(179), // <, reduce: BitwiseShiftTerm
			reduce(179), // <=, reduce: BitwiseShiftTerm
			reduce(179), // ^, reduce: BitwiseShiftTerm
			reduce(179), // &, reduce: BitwiseShiftTerm
			reduce(179), // <<, reduce: BitwiseShiftTerm
			reduce(179), // >>>, reduce: BitwiseShiftTerm
			shift(1442), // +
			shift(1443), // -
			shift(1444), // .+
			shift(1445), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2602
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(178), // >, reduce: BitwiseShiftTerm
			reduce(178), // >>, reduce: BitwiseShiftTerm
			reduce(178), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(178), // ], reduce: BitwiseShiftTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(178), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(178), // ||, reduce: BitwiseShiftTerm
			reduce(178), // ^^, reduce: BitwiseShiftTerm
			reduce(178), // &&, reduce: BitwiseShiftTerm
			reduce(178), // ??, reduce: BitwiseShiftTerm
			reduce(178), // ???, reduce: BitwiseShiftTerm
			reduce(178), // =~, reduce: BitwiseShiftTerm
			reduce(178), // !=~, reduce: BitwiseShiftTerm
			reduce(178), // ==, reduce: BitwiseShiftTerm
			reduce(178), // !=, reduce: BitwiseShiftTerm
			reduce(178), // <=>, reduce: BitwiseShiftTerm
			reduce(178), // >=, reduce: BitwiseShiftTerm
			reduce(178), // <, reduce: BitwiseShiftTerm
			reduce(178), // <=, reduce: BitwiseShiftTerm
			reduce(178), // ^, reduce: BitwiseShiftTerm
			reduce(178), // &, reduce: BitwiseShiftTerm
			reduce(178), // <<, reduce: BitwiseShiftTerm
			reduce(178), // >>>, reduce: BitwiseShiftTerm
			shift(1442), // +
			shift(1443), // -
			shift(1444), // .+
			shift(1445), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2603
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(180), // >, reduce: BitwiseShiftTerm
			reduce(180), // >>, reduce: BitwiseShiftTerm
			reduce(180), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(180), // ], reduce: BitwiseShiftTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(180), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(180), // ||, reduce: BitwiseShiftTerm
			reduce(180), // ^^, reduce: BitwiseShiftTerm
			reduce(180), // &&, reduce: BitwiseShiftTerm
			reduce(180), // ??, reduce: BitwiseShiftTerm
			reduce(180), // ???, reduce: BitwiseShiftTerm
			reduce(180), // =~, reduce: BitwiseShiftTerm
			reduce(180), // !=~, reduce: BitwiseShiftTerm
			reduce(180), // ==, reduce: BitwiseShiftTerm
			reduce(180), // !=, reduce: BitwiseShiftTerm
			reduce(180), // <=>, reduce: BitwiseShiftTerm
			reduce(180), // >=, reduce: BitwiseShiftTerm
			reduce(180), // <, reduce: BitwiseShiftTerm
			reduce(180), // <=, reduce: BitwiseShiftTerm
			reduce(180), // ^, reduce: BitwiseShiftTerm
			reduce(180), // &, reduce: BitwiseShiftTerm
			reduce(180), // <<, reduce: BitwiseShiftTerm
			reduce(180), // >>>, reduce: BitwiseShiftTerm
			shift(1442), // +
			shift(1443), // -
			shift(1444), // .+
			shift(1445), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2604
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(182), // >, reduce: AddsubdotTerm
			reduce(182), // >>, reduce: AddsubdotTerm
			reduce(182), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(182), // ], reduce: AddsubdotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(182), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(182), // ||, reduce: AddsubdotTerm
			reduce(182), // ^^, reduce: AddsubdotTerm
			reduce(182), // &&, reduce: AddsubdotTerm
			reduce(182), // ??, reduce: AddsubdotTerm
			reduce(182), // ???, reduce: AddsubdotTerm
			reduce(182), // =~, reduce: AddsubdotTerm
			reduce(182), // !=~, reduce: AddsubdotTerm
			reduce(182), // ==, reduce: AddsubdotTerm
			reduce(182), // !=, reduce: AddsubdotTerm
			reduce(182), // <=>, reduce: AddsubdotTerm
			reduce(182), // >=, reduce: AddsubdotTerm
			reduce(182), // <, reduce: AddsubdotTerm
			reduce(182), // <=, reduce: AddsubdotTerm
			reduce(182), // ^, reduce: AddsubdotTerm
			reduce(182), // &, reduce: AddsubdotTerm
			reduce(182), // <<, reduce: AddsubdotTerm
			reduce(182), // >>>, reduce: AddsubdotTerm
			reduce(182), // +, reduce: AddsubdotTerm
			reduce(182), // -, reduce: AddsubdotTerm
			reduce(182), // .+, reduce: AddsubdotTerm
			reduce(182), // .-, reduce: AddsubdotTerm
			shift(1447), // *
			shift(1448), // /
			shift(1449), // //
			shift(1450), // %
			shift(1451), // .*
			shift(1452), // ./
			shift(1453), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2605
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(183), // >, reduce: AddsubdotTerm
			reduce(183), // >>, reduce: AddsubdotTerm
			reduce(183), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(183), // ], reduce: AddsubdotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(183), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(183), // ||, reduce: AddsubdotTerm
			reduce(183), // ^^, reduce: AddsubdotTerm
			reduce(183), // &&, reduce: AddsubdotTerm
			reduce(183), // ??, reduce: AddsubdotTerm
			reduce(183), // ???, reduce: AddsubdotTerm
			reduce(183), // =~, reduce: AddsubdotTerm
			reduce(183), // !=~, reduce: AddsubdotTerm
			reduce(183), // ==, reduce: AddsubdotTerm
			reduce(183), // !=, reduce: AddsubdotTerm
			reduce(183), // <=>, reduce: AddsubdotTerm
			reduce(183), // >=, reduce: AddsubdotTerm
			reduce(183), // <, reduce: AddsubdotTerm
			reduce(183), // <=, reduce: AddsubdotTerm
			reduce(183), // ^, reduce: AddsubdotTerm
			reduce(183), // &, reduce: AddsubdotTerm
			reduce(183), // <<, reduce: AddsubdotTerm
			reduce(183), // >>>, reduce: AddsubdotTerm
			reduce(183), // +, reduce: AddsubdotTerm
			reduce(183), // -, reduce: AddsubdotTerm
			reduce(183), // .+, reduce: AddsubdotTerm
			reduce(183), // .-, reduce: AddsubdotTerm
			shift(1447), // *
			shift(1448), // /
			shift(1449), // //
			shift(1450), // %
			shift(1451), // .*
			shift(1452), // ./
			shift(1453), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2606
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(184), // >, reduce: AddsubdotTerm
			reduce(184), // >>, reduce: AddsubdotTerm
			reduce(184), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(184), // ], reduce: AddsubdotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(184), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(184), // ||, reduce: AddsubdotTerm
			reduce(184), // ^^, reduce: AddsubdotTerm
			reduce(184), // &&, reduce: AddsubdotTerm
			reduce(184), // ??, reduce: AddsubdotTerm
			reduce(184), // ???, reduce: AddsubdotTerm
			reduce(184), // =~, reduce: AddsubdotTerm
			reduce(184), // !=~, reduce: AddsubdotTerm
			reduce(184), // ==, reduce: AddsubdotTerm
			reduce(184), // !=, reduce: AddsubdotTerm
			reduce(184), // <=>, reduce: AddsubdotTerm
			reduce(184), // >=, reduce: AddsubdotTerm
			reduce(184), // <, reduce: AddsubdotTerm
			reduce(184), // <=, reduce: AddsubdotTerm
			reduce(184), // ^, reduce: AddsubdotTerm
			reduce(184), // &, reduce: AddsubdotTerm
			reduce(184), // <<, reduce: AddsubdotTerm
			reduce(184), // >>>, reduce: AddsubdotTerm
			reduce(184), // +, reduce: AddsubdotTerm
			reduce(184), // -, reduce: AddsubdotTerm
			reduce(184), // .+, reduce: AddsubdotTerm
			reduce(184), // .-, reduce: AddsubdotTerm
			shift(1447), // *
			shift(1448), // /
			shift(1449), // //
			shift(1450), // %
			shift(1451), // .*
			shift(1452), // ./
			shift(1453), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2607
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(185), // >, reduce: AddsubdotTerm
			reduce(185), // >>, reduce: AddsubdotTerm
			reduce(185), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(185), // ], reduce: AddsubdotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(185), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(185), // ||, reduce: AddsubdotTerm
			reduce(185), // ^^, reduce: AddsubdotTerm
			reduce(185), // &&, reduce: AddsubdotTerm
			reduce(185), // ??, reduce: AddsubdotTerm
			reduce(185), // ???, reduce: AddsubdotTerm
			reduce(185), // =~, reduce: AddsubdotTerm
			reduce(185), // !=~, reduce: AddsubdotTerm
			reduce(185), // ==, reduce: AddsubdotTerm
			reduce(185), // !=, reduce: AddsubdotTerm
			reduce(185), // <=>, reduce: AddsubdotTerm
			reduce(185), // >=, reduce: AddsubdotTerm
			reduce(185), // <, reduce: AddsubdotTerm
			reduce(185), // <=, reduce: AddsubdotTerm
			reduce(185), // ^, reduce: AddsubdotTerm
			reduce(185), // &, reduce: AddsubdotTerm
			reduce(185), // <<, reduce: AddsubdotTerm
			reduce(185), // >>>, reduce: AddsubdotTerm
			reduce(185), // +, reduce: AddsubdotTerm
			reduce(185), // -, reduce: AddsubdotTerm
			reduce(185), // .+, reduce: AddsubdotTerm
			reduce(185), // .-, reduce: AddsubdotTerm
			shift(1447), // *
			shift(1448), // /
			shift(1449), // //
			shift(1450), // %
			shift(1451), // .*
			shift(1452), // ./
			shift(1453), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2608
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(187), // >, reduce: MuldivTerm
			reduce(187), // >>, reduce: MuldivTerm
			reduce(187), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(187), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(187), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(187), // ||, reduce: MuldivTerm
			reduce(187), // ^^, reduce: MuldivTerm
			reduce(187), // &&, reduce: MuldivTerm
			reduce(187), // ??, reduce: MuldivTerm
			reduce(187), // ???, reduce: MuldivTerm
			reduce(187), // =~, reduce: MuldivTerm
			reduce(187), // !=~, reduce: MuldivTerm
			reduce(187), // ==, reduce: MuldivTerm
			reduce(187), // !=, reduce: MuldivTerm
			reduce(187), // <=>, reduce: MuldivTerm
			reduce(187), // >=, reduce: MuldivTerm
			reduce(187), // <, reduce: MuldivTerm
			reduce(187), // <=, reduce: MuldivTerm
			reduce(187), // ^, reduce: MuldivTerm
			reduce(187), // &, reduce: MuldivTerm
			reduce(187), // <<, reduce: MuldivTerm
			reduce(187), // >>>, reduce: MuldivTerm
			reduce(187), // +, reduce: MuldivTerm
			reduce(187), // -, reduce: MuldivTerm
			reduce(187), // .+, reduce: MuldivTerm
			reduce(187), // .-, reduce: MuldivTerm
			reduce(187), // *, reduce: MuldivTerm
			reduce(187), // /, reduce: MuldivTerm
			reduce(187), // //, reduce: MuldivTerm
			reduce(187), // %, reduce: MuldivTerm
			reduce(187), // .*, reduce: MuldivTerm
			reduce(187), // ./, reduce: MuldivTerm
			reduce(187), // .//, reduce: MuldivTerm
			shift(1457), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2609
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(188), // >, reduce: MuldivTerm
			reduce(188), // >>, reduce: MuldivTerm
			reduce(188), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(188), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(188), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(188), // ||, reduce: MuldivTerm
			reduce(188), // ^^, reduce: MuldivTerm
			reduce(188), // &&, reduce: MuldivTerm
			reduce(188), // ??, reduce: MuldivTerm
			reduce(188), // ???, reduce: MuldivTerm
			reduce(188), // =~, reduce: MuldivTerm
			reduce(188), // !=~, reduce: MuldivTerm
			reduce(188), // ==, reduce: MuldivTerm
			reduce(188), // !=, reduce: MuldivTerm
			reduce(188), // <=>, reduce: MuldivTerm
			reduce(188), // >=, reduce: MuldivTerm
			reduce(188), // <, reduce: MuldivTerm
			reduce(188), // <=, reduce: MuldivTerm
			reduce(188), // ^, reduce: MuldivTerm
			reduce(188), // &, reduce: MuldivTerm
			reduce(188), // <<, reduce: MuldivTerm
			reduce(188), // >>>, reduce: MuldivTerm
			reduce(188), // +, reduce: MuldivTerm
			reduce(188), // -, reduce: MuldivTerm
			reduce(188), // .+, reduce: MuldivTerm
			reduce(188), // .-, reduce: MuldivTerm
			reduce(188), // *, reduce: MuldivTerm
			reduce(188), // /, reduce: MuldivTerm
			reduce(188), // //, reduce: MuldivTerm
			reduce(188), // %, reduce: MuldivTerm
			reduce(188), // .*, reduce: MuldivTerm
			reduce(188), // ./, reduce: MuldivTerm
			reduce(188), // .//, reduce: MuldivTerm
			shift(1457), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2610
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(189), // >, reduce: MuldivTerm
			reduce(189), // >>, reduce: MuldivTerm
			reduce(189), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(189), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(189), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(189), // ||, reduce: MuldivTerm
			reduce(189), // ^^, reduce: MuldivTerm
			reduce(189), // &&, reduce: MuldivTerm
			reduce(189), // ??, reduce: MuldivTerm
			reduce(189), // ???, reduce: MuldivTerm
			reduce(189), // =~, reduce: MuldivTerm
			reduce(189), // !=~, reduce: MuldivTerm
			reduce(189), // ==, reduce: MuldivTerm
			reduce(189), // !=, reduce: MuldivTerm
			reduce(189), // <=>, reduce: MuldivTerm
			reduce(189), // >=, reduce: MuldivTerm
			reduce(189), // <, reduce: MuldivTerm
			reduce(189), // <=, reduce: MuldivTerm
			reduce(189), // ^, reduce: MuldivTerm
			reduce(189), // &, reduce: MuldivTerm
			reduce(189), // <<, reduce: MuldivTerm
			reduce(189), // >>>, reduce: MuldivTerm
			reduce(189), // +, reduce: MuldivTerm
			reduce(189), // -, reduce: MuldivTerm
			reduce(189), // .+, reduce: MuldivTerm
			reduce(189), // .-, reduce: MuldivTerm
			reduce(189), // *, reduce: MuldivTerm
			reduce(189), // /, reduce: MuldivTerm
			reduce(189), // //, reduce: MuldivTerm
			reduce(189), // %, reduce: MuldivTerm
			reduce(189), // .*, reduce: MuldivTerm
			reduce(189), // ./, reduce: MuldivTerm
			reduce(189), // .//, reduce: MuldivTerm
			shift(1457), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2611
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(190), // >, reduce: MuldivTerm
			reduce(190), // >>, reduce: MuldivTerm
			reduce(190), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(190), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(190), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(190), // ||, reduce: MuldivTerm
			reduce(190), // ^^, reduce: MuldivTerm
			reduce(190), // &&, reduce: MuldivTerm
			reduce(190), // ??, reduce: MuldivTerm
			reduce(190), // ???, reduce: MuldivTerm
			reduce(190), // =~, reduce: MuldivTerm
			reduce(190), // !=~, reduce: MuldivTerm
			reduce(190), // ==, reduce: MuldivTerm
			reduce(190), // !=, reduce: MuldivTerm
			reduce(190), // <=>, reduce: MuldivTerm
			reduce(190), // >=, reduce: MuldivTerm
			reduce(190), // <, reduce: MuldivTerm
			reduce(190), // <=, reduce: MuldivTerm
			reduce(190), // ^, reduce: MuldivTerm
			reduce(190), // &, reduce: MuldivTerm
			reduce(190), // <<, reduce: MuldivTerm
			reduce(190), // >>>, reduce: MuldivTerm
			reduce(190), // +, reduce: MuldivTerm
			reduce(190), // -, reduce: MuldivTerm
			reduce(190), // .+, reduce: MuldivTerm
			reduce(190), // .-, reduce: MuldivTerm
			reduce(190), // *, reduce: MuldivTerm
			reduce(190), // /, reduce: MuldivTerm
			reduce(190), // //, reduce: MuldivTerm
			reduce(190), // %, reduce: MuldivTerm
			reduce(190), // .*, reduce: MuldivTerm
			reduce(190), // ./, reduce: MuldivTerm
			reduce(190), // .//, reduce: MuldivTerm
			shift(1457), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2612
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(191), // >, reduce: MuldivTerm
			reduce(191), // >>, reduce: MuldivTerm
			reduce(191), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(191), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(191), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(191), // ||, reduce: MuldivTerm
			reduce(191), // ^^, reduce: MuldivTerm
			reduce(191), // &&, reduce: MuldivTerm
			reduce(191), // ??, reduce: MuldivTerm
			reduce(191), // ???, reduce: MuldivTerm
			reduce(191), // =~, reduce: MuldivTerm
			reduce(191), // !=~, reduce: MuldivTerm
			reduce(191), // ==, reduce: MuldivTerm
			reduce(191), // !=, reduce: MuldivTerm
			reduce(191), // <=>, reduce: MuldivTerm
			reduce(191), // >=, reduce: MuldivTerm
			reduce(191), // <, reduce: MuldivTerm
			reduce(191), // <=, reduce: MuldivTerm
			reduce(191), // ^, reduce: MuldivTerm
			reduce(191), // &, reduce: MuldivTerm
			reduce(191), // <<, reduce: MuldivTerm
			reduce(191), // >>>, reduce: MuldivTerm
			reduce(191), // +, reduce: MuldivTerm
			reduce(191), // -, reduce: MuldivTerm
			reduce(191), // .+, reduce: MuldivTerm
			reduce(191), // .-, reduce: MuldivTerm
			reduce(191), // *, reduce: MuldivTerm
			reduce(191), // /, reduce: MuldivTerm
			reduce(191), // //, reduce: MuldivTerm
			reduce(191), // %, reduce: MuldivTerm
			reduce(191), // .*, reduce: MuldivTerm
			reduce(191), // ./, reduce: MuldivTerm
			reduce(191), // .//, reduce: MuldivTerm
			shift(1457), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2613
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(192), // >, reduce: MuldivTerm
			reduce(192), // >>, reduce: MuldivTerm
			reduce(192), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(192), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(192), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(192), // ||, reduce: MuldivTerm
			reduce(192), // ^^, reduce: MuldivTerm
			reduce(192), // &&, reduce: MuldivTerm
			reduce(192), // ??, reduce: MuldivTerm
			reduce(192), // ???, reduce: MuldivTerm
			reduce(192), // =~, reduce: MuldivTerm
			reduce(192), // !=~, reduce: MuldivTerm
			reduce(192), // ==, reduce: MuldivTerm
			reduce(192), // !=, reduce: MuldivTerm
			reduce(192), // <=>, reduce: MuldivTerm
			reduce(192), // >=, reduce: MuldivTerm
			reduce(192), // <, reduce: MuldivTerm
			reduce(192), // <=, reduce: MuldivTerm
			reduce(192), // ^, reduce: MuldivTerm
			reduce(192), // &, reduce: MuldivTerm
			reduce(192), // <<, reduce: MuldivTerm
			reduce(192), // >>>, reduce: MuldivTerm
			reduce(192), // +, reduce: MuldivTerm
			reduce(192), // -, reduce: MuldivTerm
			reduce(192), // .+, reduce: MuldivTerm
			reduce(192), // .-, reduce: MuldivTerm
			reduce(192), // *, reduce: MuldivTerm
			reduce(192), // /, reduce: MuldivTerm
			reduce(192), // //, reduce: MuldivTerm
			reduce(192), // %, reduce: MuldivTerm
			reduce(192), // .*, reduce: MuldivTerm
			reduce(192), // ./, reduce: MuldivTerm
			reduce(192), // .//, reduce: MuldivTerm
			shift(1457), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2614
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(193), // >, reduce: MuldivTerm
			reduce(193), // >>, reduce: MuldivTerm
			reduce(193), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(193), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(193), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(193), // ||, reduce: MuldivTerm
			reduce(193), // ^^, reduce: MuldivTerm
			reduce(193), // &&, reduce: MuldivTerm
			reduce(193), // ??, reduce: MuldivTerm
			reduce(193), // ???, reduce: MuldivTerm
			reduce(193), // =~, reduce: MuldivTerm
			reduce(193), // !=~, reduce: MuldivTerm
			reduce(193), // ==, reduce: MuldivTerm
			reduce(193), // !=, reduce: MuldivTerm
			reduce(193), // <=>, reduce: MuldivTerm
			reduce(193), // >=, reduce: MuldivTerm
			reduce(193), // <, reduce: MuldivTerm
			reduce(193), // <=, reduce: MuldivTerm
			reduce(193), // ^, reduce: MuldivTerm
			reduce(193), // &, reduce: MuldivTerm
			reduce(193), // <<, reduce: MuldivTerm
			reduce(193), // >>>, reduce: MuldivTerm
			reduce(193), // +, reduce: MuldivTerm
			reduce(193), // -, reduce: MuldivTerm
			reduce(193), // .+, reduce: MuldivTerm
			reduce(193), // .-, reduce: MuldivTerm
			reduce(193), // *, reduce: MuldivTerm
			reduce(193), // /, reduce: MuldivTerm
			reduce(193), // //, reduce: MuldivTerm
			reduce(193), // %, reduce: MuldivTerm
			reduce(193), // .*, reduce: MuldivTerm
			reduce(193), // ./, reduce: MuldivTerm
			reduce(193), // .//, reduce: MuldivTerm
			shift(1457), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2615
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(195), // >, reduce: DotTerm
			reduce(195), // >>, reduce: DotTerm
			reduce(195), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(195), // ], reduce: DotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(195), // ?, reduce: DotTerm
			nil,         // :
			reduce(195), // ||, reduce: DotTerm
			reduce(195), // ^^, reduce: DotTerm
			reduce(195), // &&, reduce: DotTerm
			reduce(195), // ??, reduce: DotTerm
			reduce(195), // ???, reduce: DotTerm
			reduce(195), // =~, reduce: DotTerm
			reduce(195), // !=~, reduce: DotTerm
			reduce(195), // ==, reduce: DotTerm
			reduce(195), // !=, reduce: DotTerm
			reduce(195), // <=>, reduce: DotTerm
			reduce(195), // >=, reduce: DotTerm
			reduce(195), // <, reduce: DotTerm
			reduce(195), // <=, reduce: DotTerm
			reduce(195), // ^, reduce: DotTerm
			reduce(195), // &, reduce: DotTerm
			reduce(195), // <<, reduce: DotTerm
			reduce(195), // >>>, reduce: DotTerm
			reduce(195), // +, reduce: DotTerm
			reduce(195), // -, reduce: DotTerm
			reduce(195), // .+, reduce: DotTerm
			reduce(195), // .-, reduce: DotTerm
			reduce(195), // *, reduce: DotTerm
			reduce(195), // /, reduce: DotTerm
			reduce(195), // //, reduce: DotTerm
			reduce(195), // %, reduce: DotTerm
			reduce(195), // .*, reduce: DotTerm
			reduce(195), // ./, reduce: DotTerm
			reduce(195), // .//, reduce: DotTerm
			reduce(195), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2616
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2617
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2618
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(204), // >, reduce: PowTerm
			reduce(204), // >>, reduce: PowTerm
			reduce(204), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(204), // ], reduce: PowTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(204), // ?, reduce: PowTerm
			nil,         // :
			reduce(204), // ||, reduce: PowTerm
			reduce(204), // ^^, reduce: PowTerm
			reduce(204), // &&, reduce: PowTerm
			reduce(204), // ??, reduce: PowTerm
			reduce(204), // ???, reduce: PowTerm
			reduce(204), // =~, reduce: PowTerm
			reduce(204), // !=~, reduce: PowTerm
			reduce(204), // ==, reduce: PowTerm
			reduce(204), // !=, reduce: PowTerm
			reduce(204), // <=>, reduce: PowTerm
			reduce(204), // >=, reduce: PowTerm
			reduce(204), // <, reduce: PowTerm
			reduce(204), // <=, reduce: PowTerm
			reduce(204), // ^, reduce: PowTerm
			reduce(204), // &, reduce: PowTerm
			reduce(204), // <<, reduce: PowTerm
			reduce(204), // >>>, reduce: PowTerm
			reduce(204), // +, reduce: PowTerm
			reduce(204), // -, reduce: PowTerm
			reduce(204), // .+, reduce: PowTerm
			reduce(204), // .-, reduce: PowTerm
			reduce(204), // *, reduce: PowTerm
			reduce(204), // /, reduce: PowTerm
			reduce(204), // //, reduce: PowTerm
			reduce(204), // %, reduce: PowTerm
			reduce(204), // .*, reduce: PowTerm
			reduce(204), // ./, reduce: PowTerm
			reduce(204), // .//, reduce: PowTerm
			reduce(204), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2619
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3308), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3309), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2620
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(3311), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2621
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3312), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2622
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3313), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2623
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(229), // >, reduce: ArrayLiteral
			reduce(229), // >>, reduce: ArrayLiteral
			reduce(229), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(229), // ], reduce: ArrayLiteral
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(229), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(229), // ||, reduce: ArrayLiteral
			reduce(229), // ^^, reduce: ArrayLiteral
			reduce(229), // &&, reduce: ArrayLiteral
			reduce(229), // ??, reduce: ArrayLiteral
			reduce(229), // ???, reduce: ArrayLiteral
			reduce(229), // =~, reduce: ArrayLiteral
			reduce(229), // !=~, reduce: ArrayLiteral
			reduce(229), // ==, reduce: ArrayLiteral
			reduce(229), // !=, reduce: ArrayLiteral
			reduce(229), // <=>, reduce: ArrayLiteral
			reduce(229), // >=, reduce: ArrayLiteral
			reduce(229), // <, reduce: ArrayLiteral
			reduce(229), // <=, reduce: ArrayLiteral
			reduce(229), // ^, reduce: ArrayLiteral
			reduce(229), // &, reduce: ArrayLiteral
			reduce(229), // <<, reduce: ArrayLiteral
			reduce(229), // >>>, reduce: ArrayLiteral
			reduce(229), // +, reduce: ArrayLiteral
			reduce(229), // -, reduce: ArrayLiteral
			reduce(229), // .+, reduce: ArrayLiteral
			reduce(229), // .-, reduce: ArrayLiteral
			reduce(229), // *, reduce: ArrayLiteral
			reduce(229), // /, reduce: ArrayLiteral
			reduce(229), // //, reduce: ArrayLiteral
			reduce(229), // %, reduce: ArrayLiteral
			reduce(229), // .*, reduce: ArrayLiteral
			reduce(229), // ./, reduce: ArrayLiteral
			reduce(229), // .//, reduce: ArrayLiteral
			reduce(229), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(229), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(229), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(229), // [[, reduce: ArrayLiteral
			reduce(229), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2624
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(255), // >, reduce: ENV
			reduce(255), // >>, reduce: ENV
			reduce(255), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(255), // ], reduce: ENV
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(255), // ?, reduce: ENV
			nil,         // :
			reduce(255), // ||, reduce: ENV
			reduce(255), // ^^, reduce: ENV
			reduce(255), // &&, reduce: ENV
			reduce(255), // ??, reduce: ENV
			reduce(255), // ???, reduce: ENV
			reduce(255), // =~, reduce: ENV
			reduce(255), // !=~, reduce: ENV
			reduce(255), // ==, reduce: ENV
			reduce(255), // !=, reduce: ENV
			reduce(255), // <=>, reduce: ENV
			reduce(255), // >=, reduce: ENV
			reduce(255), // <, reduce: ENV
			reduce(255), // <=, reduce: ENV
			reduce(255), // ^, reduce: ENV
			reduce(255), // &, reduce: ENV
			reduce(255), // <<, reduce: ENV
			reduce(255), // >>>, reduce: ENV
			reduce(255), // +, reduce: ENV
			reduce(255), // -, reduce: ENV
			reduce(255), // .+, reduce: ENV
			reduce(255), // .-, reduce: ENV
			reduce(255), // *, reduce: ENV
			reduce(255), // /, reduce: ENV
			reduce(255), // //, reduce: ENV
			reduce(255), // %, reduce: ENV
			reduce(255), // .*, reduce: ENV
			reduce(255), // ./, reduce: ENV
			reduce(255), // .//, reduce: ENV
			reduce(255), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(255), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(255), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(255), // [[, reduce: ENV
			reduce(255), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2625
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3314), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2626
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3315), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2627
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(268), // >, reduce: FunctionCallsite
			reduce(268), // >>, reduce: FunctionCallsite
			reduce(268), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(268), // ], reduce: FunctionCallsite
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(268), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(268), // ||, reduce: FunctionCallsite
			reduce(268), // ^^, reduce: FunctionCallsite
			reduce(268), // &&, reduce: FunctionCallsite
			reduce(268), // ??, reduce: FunctionCallsite
			reduce(268), // ???, reduce: FunctionCallsite
			reduce(268), // =~, reduce: FunctionCallsite
			reduce(268), // !=~, reduce: FunctionCallsite
			reduce(268), // ==, reduce: FunctionCallsite
			reduce(268), // !=, reduce: FunctionCallsite
			reduce(268), // <=>, reduce: FunctionCallsite
			reduce(268), // >=, reduce: FunctionCallsite
			reduce(268), // <, reduce: FunctionCallsite
			reduce(268), // <=, reduce: FunctionCallsite
			reduce(268), // ^, reduce: FunctionCallsite
			reduce(268), // &, reduce: FunctionCallsite
			reduce(268), // <<, reduce: FunctionCallsite
			reduce(268), // >>>, reduce: FunctionCallsite
			reduce(268), // +, reduce: FunctionCallsite
			reduce(268), // -, reduce: FunctionCallsite
			reduce(268), // .+, reduce: FunctionCallsite
			reduce(268), // .-, reduce: FunctionCallsite
			reduce(268), // *, reduce: FunctionCallsite
			reduce(268), // /, reduce: FunctionCallsite
			reduce(268), // //, reduce: FunctionCallsite
			reduce(268), // %, reduce: FunctionCallsite
			reduce(268), // .*, reduce: FunctionCallsite
			reduce(268), // ./, reduce: FunctionCallsite
			reduce(268), // .//, reduce: FunctionCallsite
			reduce(268), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(268), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(268), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(268), // [[, reduce: FunctionCallsite
			reduce(268), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2628
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3316), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2629
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(107), // $, reduce: PositionalFieldName
			nil,         // empty
			reduce(107), // ;, reduce: PositionalFieldName
			reduce(107), // {, reduce: PositionalFieldName
			nil,         // }
			reduce(107), // =, reduce: PositionalFieldName
			nil,         // unset
			nil,         // filter
			reduce(107), // >, reduce: PositionalFieldName
			reduce(107), // >>, reduce: PositionalFieldName
			reduce(107), // |, reduce: PositionalFieldName
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(107), // ||=, reduce: PositionalFieldName
			reduce(107), // ^^=, reduce: PositionalFieldName
			reduce(107), // &&=, reduce: PositionalFieldName
			reduce(107), // ??=, reduce: PositionalFieldName
			reduce(107), // ???=, reduce: PositionalFieldName
			reduce(107), // |=, reduce: PositionalFieldName
			reduce(107), // &=, reduce: PositionalFieldName
			reduce(107), // ^=, reduce: PositionalFieldName
			reduce(107), // <<=, reduce: PositionalFieldName
			reduce(107), // >>=, reduce: PositionalFieldName
			reduce(107), // >>>=, reduce: PositionalFieldName
			reduce(107), // +=, reduce: PositionalFieldName
			reduce(107), // .=, reduce: PositionalFieldName
			reduce(107), // -=, reduce: PositionalFieldName
			reduce(107), // *=, reduce: PositionalFieldName
			reduce(107), // /=, reduce: PositionalFieldName
			reduce(107), // //=, reduce: PositionalFieldName
			reduce(107), // %=, reduce: PositionalFieldName
			reduce(107), // **=, reduce: PositionalFieldName
			reduce(107), // ?, reduce: PositionalFieldName
			nil,         // :
			reduce(107), // ||, reduce: PositionalFieldName
			reduce(107), // ^^, reduce: PositionalFieldName
			reduce(107), // &&, reduce: PositionalFieldName
			reduce(107), // ??, reduce: PositionalFieldName
			reduce(107), // ???, reduce: PositionalFieldName
			reduce(107), // =~, reduce: PositionalFieldName
			reduce(107), // !=~, reduce: PositionalFieldName
			reduce(107), // ==, reduce: PositionalFieldName
			reduce(107), // !=, reduce: PositionalFieldName
			reduce(107), // <=>, reduce: PositionalFieldName
			reduce(107), // >=, reduce: PositionalFieldName
			reduce(107), // <, reduce: PositionalFieldName
			reduce(107), // <=, reduce: PositionalFieldName
			reduce(107), // ^, reduce: PositionalFieldName
			reduce(107), // &, reduce: PositionalFieldName
			reduce(107), // <<, reduce: PositionalFieldName
			reduce(107), // >>>, reduce: PositionalFieldName
			reduce(107), // +, reduce: PositionalFieldName
			reduce(107), // -, reduce: PositionalFieldName
			reduce(107), // .+, reduce: PositionalFieldName
			reduce(107), // .-, reduce: PositionalFieldName
			reduce(107), // *, reduce: PositionalFieldName
			reduce(107), // /, reduce: PositionalFieldName
			reduce(107), // //, reduce: PositionalFieldName
			reduce(107), // %, reduce: PositionalFieldName
			reduce(107), // .*, reduce: PositionalFieldName
			reduce(107), // ./, reduce: PositionalFieldName
			reduce(107), // .//, reduce: PositionalFieldName
			reduce(107), // ., reduce: PositionalFieldName
			nil,         // !
			nil,         // ~
			reduce(107), // **, reduce: PositionalFieldName
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(107), // [, reduce: PositionalFieldName
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(107), // [[, reduce: PositionalFieldName
			reduce(107), // [[[, reduce: PositionalFieldName
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2630
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3317), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2631
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(7),   // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(45),  // (
			nil,        // )
			nil,        // emitp
			shift(58),  // field_name
			shift(59),  // $[
			nil,        // ]
			shift(60),  // braced_field_name
			shift(61),  // $[[
			shift(62),  // $[[[
			shift(63),  // full_srec
			shift(65),  // oosvar_name
			shift(66),  // @[
			shift(67),  // braced_oosvar_name
			shift(68),  // full_oosvar
			shift(69),  // all
			shift(70),  // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(94),  // +
			shift(96),  // -
			shift(97),  // .+
			shift(98),  // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(101), // !
			shift(102), // ~
			nil,        // **
			shift(107), // string_literal
			shift(108), // regex_case_insensitive
			shift(109), // int_literal
			shift(110), // float_literal
			shift(111), // boolean_literal
			shift(112), // null_literal
			shift(113), // inf_literal
			shift(114), // nan_literal
			shift(115), // const_M_PI
			shift(116), // const_M_E
			shift(117), // panic
			shift(119), // [
			shift(121), // ctx_IPS
			shift(122), // ctx_IFS
			shift(123), // ctx_IRS
			shift(124), // ctx_OPS
			shift(125), // ctx_OFS
			shift(126), // ctx_ORS
			shift(127), // ctx_FLATSEP
			shift(128), // ctx_NF
			shift(129), // ctx_NR
			shift(130), // ctx_FNR
			shift(131), // ctx_FILENAME
			shift(132), // ctx_FILENUM
			shift(134), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(734), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2632
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(206), // $, reduce: PowTerm
			nil,         // empty
			reduce(206), // ;, reduce: PowTerm
			reduce(206), // {, reduce: PowTerm
			nil,         // }
			reduce(206), // =, reduce: PowTerm
			nil,         // unset
			nil,         // filter
			reduce(206), // >, reduce: PowTerm
			reduce(206), // >>, reduce: PowTerm
			reduce(206), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(206), // ||=, reduce: PowTerm
			reduce(206), // ^^=, reduce: PowTerm
			reduce(206), // &&=, reduce: PowTerm
			reduce(206), // ??=, reduce: PowTerm
			reduce(206), // ???=, reduce: PowTerm
			reduce(206), // |=, reduce: PowTerm
			reduce(206), // &=, reduce: PowTerm
			reduce(206), // ^=, reduce: PowTerm
			reduce(206), // <<=, reduce: PowTerm
			reduce(206), // >>=, reduce: PowTerm
			reduce(206), // >>>=, reduce: PowTerm
			reduce(206), // +=, reduce: PowTerm
			reduce(206), // .=, reduce: PowTerm
			reduce(206), // -=, reduce: PowTerm
			reduce(206), // *=, reduce: PowTerm
			reduce(206), // /=, reduce: PowTerm
			reduce(206), // //=, reduce: PowTerm
			reduce(206), // %=, reduce: PowTerm
			reduce(206), // **=, reduce: PowTerm
			reduce(206), // ?, reduce: PowTerm
			nil,         // :
			reduce(206), // ||, reduce: PowTerm
			reduce(206), // ^^, reduce: PowTerm
			reduce(206), // &&, reduce: PowTerm
			reduce(206), // ??, reduce: PowTerm
			reduce(206), // ???, reduce: PowTerm
			reduce(206), // =~, reduce: PowTerm
			reduce(206), // !=~, reduce: PowTerm
			reduce(206), // ==, reduce: PowTerm
			reduce(206), // !=, reduce: PowTerm
			reduce(206), // <=>, reduce: PowTerm
			reduce(206), // >=, reduce: PowTerm
			reduce(206), // <, reduce: PowTerm
			reduce(206), // <=, reduce: PowTerm
			reduce(206), // ^, reduce: PowTerm
			reduce(206), // &, reduce: PowTerm
			reduce(206), // <<, reduce: PowTerm
			reduce(206), // >>>, reduce: PowTerm
			reduce(206), // +, reduce: PowTerm
			reduce(206), // -, reduce: PowTerm
			reduce(206), // .+, reduce: PowTerm
			reduce(206), // .-, reduce: PowTerm
			reduce(206), // *, reduce: PowTerm
			reduce(206), // /, reduce: PowTerm
			reduce(206), // //, reduce: PowTerm
			reduce(206), // %, reduce: PowTerm
			reduce(206), // .*, reduce: PowTerm
			reduce(206), // ./, reduce: PowTerm
			reduce(206), // .//, reduce: PowTerm
			reduce(206), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2633
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(205), // $, reduce: PowTerm
			nil,         // empty
			reduce(205), // ;, reduce: PowTerm
			reduce(205), // {, reduce: PowTerm
			nil,         // }
			reduce(205), // =, reduce: PowTerm
			nil,         // unset
			nil,         // filter
			reduce(205), // >, reduce: PowTerm
			reduce(205), // >>, reduce: PowTerm
			reduce(205), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(205), // ||=, reduce: PowTerm
			reduce(205), // ^^=, reduce: PowTerm
			reduce(205), // &&=, reduce: PowTerm
			reduce(205), // ??=, reduce: PowTerm
			reduce(205), // ???=, reduce: PowTerm
			reduce(205), // |=, reduce: PowTerm
			reduce(205), // &=, reduce: PowTerm
			reduce(205), // ^=, reduce: PowTerm
			reduce(205), // <<=, reduce: PowTerm
			reduce(205), // >>=, reduce: PowTerm
			reduce(205), // >>>=, reduce: PowTerm
			reduce(205), // +=, reduce: PowTerm
			reduce(205), // .=, reduce: PowTerm
			reduce(205), // -=, reduce: PowTerm
			reduce(205), // *=, reduce: PowTerm
			reduce(205), // /=, reduce: PowTerm
			reduce(205), // //=, reduce: PowTerm
			reduce(205), // %=, reduce: PowTerm
			reduce(205), // **=, reduce: PowTerm
			reduce(205), // ?, reduce: PowTerm
			nil,         // :
			reduce(205), // ||, reduce: PowTerm
			reduce(205), // ^^, reduce: PowTerm
			reduce(205), // &&, reduce: PowTerm
			reduce(205), // ??, reduce: PowTerm
			reduce(205), // ???, reduce: PowTerm
			reduce(205), // =~, reduce: PowTerm
			reduce(205), // !=~, reduce: PowTerm
			reduce(205), // ==, reduce: PowTerm
			reduce(205), // !=, reduce: PowTerm
			reduce(205), // <=>, reduce: PowTerm
			reduce(205), // >=, reduce: PowTerm
			reduce(205), // <, reduce: PowTerm
			reduce(205), // <=, reduce: PowTerm
			reduce(205), // ^, reduce: PowTerm
			reduce(205), // &, reduce: PowTerm
			reduce(205), // <<, reduce: PowTerm
			reduce(205), // >>>, reduce: PowTerm
			reduce(205), // +, reduce: PowTerm
			reduce(205), // -, reduce: PowTerm
			reduce(205), // .+, reduce: PowTerm
			reduce(205), // .-, reduce: PowTerm
			reduce(205), // *, reduce: PowTerm
			reduce(205), // /, reduce: PowTerm
			reduce(205), // //, reduce: PowTerm
			reduce(205), // %, reduce: PowTerm
			reduce(205), // .*, reduce: PowTerm
			reduce(205), // ./, reduce: PowTerm
			reduce(205), // .//, reduce: PowTerm
			reduce(205), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2634
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(234), // >, reduce: MapLiteral
			reduce(234), // >>, reduce: MapLiteral
			reduce(234), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(234), // ], reduce: MapLiteral
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(234), // ?, reduce: MapLiteral
			reduce(234), // :, reduce: MapLiteral
			reduce(234), // ||, reduce: MapLiteral
			reduce(234), // ^^, reduce: MapLiteral
			reduce(234), // &&, reduce: MapLiteral
			reduce(234), // ??, reduce: MapLiteral
			reduce(234), // ???, reduce: MapLiteral
			reduce(234), // =~, reduce: MapLiteral
			reduce(234), // !=~, reduce: MapLiteral
			reduce(234), // ==, reduce: MapLiteral
			reduce(234), // !=, reduce: MapLiteral
			reduce(234), // <=>, reduce: MapLiteral
			reduce(234), // >=, reduce: MapLiteral
			reduce(234), // <, reduce: MapLiteral
			reduce(234), // <=, reduce: MapLiteral
			reduce(234), // ^, reduce: MapLiteral
			reduce(234), // &, reduce: MapLiteral
			reduce(234), // <<, reduce: MapLiteral
			reduce(234), // >>>, reduce: MapLiteral
			reduce(234), // +, reduce: MapLiteral
			reduce(234), // -, reduce: MapLiteral
			reduce(234), // .+, reduce: MapLiteral
			reduce(234), // .-, reduce: MapLiteral
			reduce(234), // *, reduce: MapLiteral
			reduce(234), // /, reduce: MapLiteral
			reduce(234), // //, reduce: MapLiteral
			reduce(234), // %, reduce: MapLiteral
			reduce(234), // .*, reduce: MapLiteral
			reduce(234), // ./, reduce: MapLiteral
			reduce(234), // .//, reduce: MapLiteral
			reduce(234), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(234), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(234), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(234), // [[, reduce: MapLiteral
			reduce(234), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2635
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(3319), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2636
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(260), // $, reduce: ArrayOrMapIndexAccess
			nil,         // empty
			reduce(260), // ;, reduce: ArrayOrMapIndexAccess
			reduce(260), // {, reduce: ArrayOrMapIndexAccess
			nil,         // }
			reduce(260), // =, reduce: ArrayOrMapIndexAccess
			nil,         // unset
			nil,         // filter
			reduce(260), // >, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>, reduce: ArrayOrMapIndexAccess
			reduce(260), // |, reduce: ArrayOrMapIndexAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(260), // ||=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^=, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???=, reduce: ArrayOrMapIndexAccess
			reduce(260), // |=, reduce: ArrayOrMapIndexAccess
			reduce(260), // &=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<=, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>=, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>=, reduce: ArrayOrMapIndexAccess
			reduce(260), // +=, reduce: ArrayOrMapIndexAccess
			reduce(260), // .=, reduce: ArrayOrMapIndexAccess
			reduce(260), // -=, reduce: ArrayOrMapIndexAccess
			reduce(260), // *=, reduce: ArrayOrMapIndexAccess
			reduce(260), // /=, reduce: ArrayOrMapIndexAccess
			reduce(260), // //=, reduce: ArrayOrMapIndexAccess
			reduce(260), // %=, reduce: ArrayOrMapIndexAccess
			reduce(260), // **=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ?, reduce: ArrayOrMapIndexAccess
			nil,         // :
			reduce(260), // ||, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???, reduce: ArrayOrMapIndexAccess
			reduce(260), // =~, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=~, reduce: ArrayOrMapIndexAccess
			reduce(260), // ==, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=>, reduce: ArrayOrMapIndexAccess
			reduce(260), // >=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>, reduce: ArrayOrMapIndexAccess
			reduce(260), // +, reduce: ArrayOrMapIndexAccess
			reduce(260), // -, reduce: ArrayOrMapIndexAccess
			reduce(260), // .+, reduce: ArrayOrMapIndexAccess
			reduce(260), // .-, reduce: ArrayOrMapIndexAccess
			reduce(260), // *, reduce: ArrayOrMapIndexAccess
			reduce(260), // /, reduce: ArrayOrMapIndexAccess
			reduce(260), // //, reduce: ArrayOrMapIndexAccess
			reduce(260), // %, reduce: ArrayOrMapIndexAccess
			reduce(260), // .*, reduce: ArrayOrMapIndexAccess
			reduce(260), // ./, reduce: ArrayOrMapIndexAccess
			reduce(260), // .//, reduce: ArrayOrMapIndexAccess
			reduce(260), // ., reduce: ArrayOrMapIndexAccess
			nil,         // !
			nil,         // ~
			reduce(260), // **, reduce: ArrayOrMapIndexAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(260), // [, reduce: ArrayOrMapIndexAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(260), // [[, reduce: ArrayOrMapIndexAccess
			reduce(260), // [[[, reduce: ArrayOrMapIndexAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2637
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(3321), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2638
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3322), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2639
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3323), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2640
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3324), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2641
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3325), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2642
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3326), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2643
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2644
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2645
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3329), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2646
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(266), // $, reduce: ArraySliceAccess
			nil,         // empty
			reduce(266), // ;, reduce: ArraySliceAccess
			reduce(266), // {, reduce: ArraySliceAccess
			nil,         // }
			reduce(266), // =, reduce: ArraySliceAccess
			nil,         // unset
			nil,         // filter
			reduce(266), // >, reduce: ArraySliceAccess
			reduce(266), // >>, reduce: ArraySliceAccess
			reduce(266), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(266), // ||=, reduce: ArraySliceAccess
			reduce(266), // ^^=, reduce: ArraySliceAccess
			reduce(266), // &&=, reduce: ArraySliceAccess
			reduce(266), // ??=, reduce: ArraySliceAccess
			reduce(266), // ???=, reduce: ArraySliceAccess
			reduce(266), // |=, reduce: ArraySliceAccess
			reduce(266), // &=, reduce: ArraySliceAccess
			reduce(266), // ^=, reduce: ArraySliceAccess
			reduce(266), // <<=, reduce: ArraySliceAccess
			reduce(266), // >>=, reduce: ArraySliceAccess
			reduce(266), // >>>=, reduce: ArraySliceAccess
			reduce(266), // +=, reduce: ArraySliceAccess
			reduce(266), // .=, reduce: ArraySliceAccess
			reduce(266), // -=, reduce: ArraySliceAccess
			reduce(266), // *=, reduce: ArraySliceAccess
			reduce(266), // /=, reduce: ArraySliceAccess
			reduce(266), // //=, reduce: ArraySliceAccess
			reduce(266), // %=, reduce: ArraySliceAccess
			reduce(266), // **=, reduce: ArraySliceAccess
			reduce(266), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(266), // ||, reduce: ArraySliceAccess
			reduce(266), // ^^, reduce: ArraySliceAccess
			reduce(266), // &&, reduce: ArraySliceAccess
			reduce(266), // ??, reduce: ArraySliceAccess
			reduce(266), // ???, reduce: ArraySliceAccess
			reduce(266), // =~, reduce: ArraySliceAccess
			reduce(266), // !=~, reduce: ArraySliceAccess
			reduce(266), // ==, reduce: ArraySliceAccess
			reduce(266), // !=, reduce: ArraySliceAccess
			reduce(266), // <=>, reduce: ArraySliceAccess
			reduce(266), // >=, reduce: ArraySliceAccess
			reduce(266), // <, reduce: ArraySliceAccess
			reduce(266), // <=, reduce: ArraySliceAccess
			reduce(266), // ^, reduce: ArraySliceAccess
			reduce(266), // &, reduce: ArraySliceAccess
			reduce(266), // <<, reduce: ArraySliceAccess
			reduce(266), // >>>, reduce: ArraySliceAccess
			reduce(266), // +, reduce: ArraySliceAccess
			reduce(266), // -, reduce: ArraySliceAccess
			reduce(266), // .+, reduce: ArraySliceAccess
			reduce(266), // .-, reduce: ArraySliceAccess
			reduce(266), // *, reduce: ArraySliceAccess
			reduce(266), // /, reduce: ArraySliceAccess
			reduce(266), // //, reduce: ArraySliceAccess
			reduce(266), // %, reduce: ArraySliceAccess
			reduce(266), // .*, reduce: ArraySliceAccess
			reduce(266), // ./, reduce: ArraySliceAccess
			reduce(266), // .//, reduce: ArraySliceAccess
			reduce(266), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(266), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(266), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(266), // [[, reduce: ArraySliceAccess
			reduce(266), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2647
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2648
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2649
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2650
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2651
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2652
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2653
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2654
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2655
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2656
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2657
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2658
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2659
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2660
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2661
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2662
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2663
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2664
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2665
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2666
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2667
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2668
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2669
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2670
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(197), // >, reduce: UnaryOpTerm
			reduce(197), // >>, reduce: UnaryOpTerm
			reduce(197), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(197), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(197), // ?, reduce: UnaryOpTerm
			reduce(197), // :, reduce: UnaryOpTerm
			reduce(197), // ||, reduce: UnaryOpTerm
			reduce(197), // ^^, reduce: UnaryOpTerm
			reduce(197), // &&, reduce: UnaryOpTerm
			reduce(197), // ??, reduce: UnaryOpTerm
			reduce(197), // ???, reduce: UnaryOpTerm
			reduce(197), // =~, reduce: UnaryOpTerm
			reduce(197), // !=~, reduce: UnaryOpTerm
			reduce(197), // ==, reduce: UnaryOpTerm
			reduce(197), // !=, reduce: UnaryOpTerm
			reduce(197), // <=>, reduce: UnaryOpTerm
			reduce(197), // >=, reduce: UnaryOpTerm
			reduce(197), // <, reduce: UnaryOpTerm
			reduce(197), // <=, reduce: UnaryOpTerm
			reduce(197), // ^, reduce: UnaryOpTerm
			reduce(197), // &, reduce: UnaryOpTerm
			reduce(197), // <<, reduce: UnaryOpTerm
			reduce(197), // >>>, reduce: UnaryOpTerm
			reduce(197), // +, reduce: UnaryOpTerm
			reduce(197), // -, reduce: UnaryOpTerm
			reduce(197), // .+, reduce: UnaryOpTerm
			reduce(197), // .-, reduce: UnaryOpTerm
			reduce(197), // *, reduce: UnaryOpTerm
			reduce(197), // /, reduce: UnaryOpTerm
			reduce(197), // //, reduce: UnaryOpTerm
			reduce(197), // %, reduce: UnaryOpTerm
			reduce(197), // .*, reduce: UnaryOpTerm
			reduce(197), // ./, reduce: UnaryOpTerm
			reduce(197), // .//, reduce: UnaryOpTerm
			reduce(197), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2671
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2672
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2673
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2674
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2675
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2676
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2677
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2678
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(198), // >, reduce: UnaryOpTerm
			reduce(198), // >>, reduce: UnaryOpTerm
			reduce(198), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(198), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(198), // ?, reduce: UnaryOpTerm
			reduce(198), // :, reduce: UnaryOpTerm
			reduce(198), // ||, reduce: UnaryOpTerm
			reduce(198), // ^^, reduce: UnaryOpTerm
			reduce(198), // &&, reduce: UnaryOpTerm
			reduce(198), // ??, reduce: UnaryOpTerm
			reduce(198), // ???, reduce: UnaryOpTerm
			reduce(198), // =~, reduce: UnaryOpTerm
			reduce(198), // !=~, reduce: UnaryOpTerm
			reduce(198), // ==, reduce: UnaryOpTerm
			reduce(198), // !=, reduce: UnaryOpTerm
			reduce(198), // <=>, reduce: UnaryOpTerm
			reduce(198), // >=, reduce: UnaryOpTerm
			reduce(198), // <, reduce: UnaryOpTerm
			reduce(198), // <=, reduce: UnaryOpTerm
			reduce(198), // ^, reduce: UnaryOpTerm
			reduce(198), // &, reduce: UnaryOpTerm
			reduce(198), // <<, reduce: UnaryOpTerm
			reduce(198), // >>>, reduce: UnaryOpTerm
			reduce(198), // +, reduce: UnaryOpTerm
			reduce(198), // -, reduce: UnaryOpTerm
			reduce(198), // .+, reduce: UnaryOpTerm
			reduce(198), // .-, reduce: UnaryOpTerm
			reduce(198), // *, reduce: UnaryOpTerm
			reduce(198), // /, reduce: UnaryOpTerm
			reduce(198), // //, reduce: UnaryOpTerm
			reduce(198), // %, reduce: UnaryOpTerm
			reduce(198), // .*, reduce: UnaryOpTerm
			reduce(198), // ./, reduce: UnaryOpTerm
			reduce(198), // .//, reduce: UnaryOpTerm
			reduce(198), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2679
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(199), // >, reduce: UnaryOpTerm
			reduce(199), // >>, reduce: UnaryOpTerm
			reduce(199), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(199), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(199), // ?, reduce: UnaryOpTerm
			reduce(199), // :, reduce: UnaryOpTerm
			reduce(199), // ||, reduce: UnaryOpTerm
			reduce(199), // ^^, reduce: UnaryOpTerm
			reduce(199), // &&, reduce: UnaryOpTerm
			reduce(199), // ??, reduce: UnaryOpTerm
			reduce(199), // ???, reduce: UnaryOpTerm
			reduce(199), // =~, reduce: UnaryOpTerm
			reduce(199), // !=~, reduce: UnaryOpTerm
			reduce(199), // ==, reduce: UnaryOpTerm
			reduce(199), // !=, reduce: UnaryOpTerm
			reduce(199), // <=>, reduce: UnaryOpTerm
			reduce(199), // >=, reduce: UnaryOpTerm
			reduce(199), // <, reduce: UnaryOpTerm
			reduce(199), // <=, reduce: UnaryOpTerm
			reduce(199), // ^, reduce: UnaryOpTerm
			reduce(199), // &, reduce: UnaryOpTerm
			reduce(199), // <<, reduce: UnaryOpTerm
			reduce(199), // >>>, reduce: UnaryOpTerm
			reduce(199), // +, reduce: UnaryOpTerm
			reduce(199), // -, reduce: UnaryOpTerm
			reduce(199), // .+, reduce: UnaryOpTerm
			reduce(199), // .-, reduce: UnaryOpTerm
			reduce(199), // *, reduce: UnaryOpTerm
			reduce(199), // /, reduce: UnaryOpTerm
			reduce(199), // //, reduce: UnaryOpTerm
			reduce(199), // %, reduce: UnaryOpTerm
			reduce(199), // .*, reduce: UnaryOpTerm
			reduce(199), // ./, reduce: UnaryOpTerm
			reduce(199), // .//, reduce: UnaryOpTerm
			reduce(199), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2680
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(200), // >, reduce: UnaryOpTerm
			reduce(200), // >>, reduce: UnaryOpTerm
			reduce(200), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(200), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(200), // ?, reduce: UnaryOpTerm
			reduce(200), // :, reduce: UnaryOpTerm
			reduce(200), // ||, reduce: UnaryOpTerm
			reduce(200), // ^^, reduce: UnaryOpTerm
			reduce(200), // &&, reduce: UnaryOpTerm
			reduce(200), // ??, reduce: UnaryOpTerm
			reduce(200), // ???, reduce: UnaryOpTerm
			reduce(200), // =~, reduce: UnaryOpTerm
			reduce(200), // !=~, reduce: UnaryOpTerm
			reduce(200), // ==, reduce: UnaryOpTerm
			reduce(200), // !=, reduce: UnaryOpTerm
			reduce(200), // <=>, reduce: UnaryOpTerm
			reduce(200), // >=, reduce: UnaryOpTerm
			reduce(200), // <, reduce: UnaryOpTerm
			reduce(200), // <=, reduce: UnaryOpTerm
			reduce(200), // ^, reduce: UnaryOpTerm
			reduce(200), // &, reduce: UnaryOpTerm
			reduce(200), // <<, reduce: UnaryOpTerm
			reduce(200), // >>>, reduce: UnaryOpTerm
			reduce(200), // +, reduce: UnaryOpTerm
			reduce(200), // -, reduce: UnaryOpTerm
			reduce(200), // .+, reduce: UnaryOpTerm
			reduce(200), // .-, reduce: UnaryOpTerm
			reduce(200), // *, reduce: UnaryOpTerm
			reduce(200), // /, reduce: UnaryOpTerm
			reduce(200), // //, reduce: UnaryOpTerm
			reduce(200), // %, reduce: UnaryOpTerm
			reduce(200), // .*, reduce: UnaryOpTerm
			reduce(200), // ./, reduce: UnaryOpTerm
			reduce(200), // .//, reduce: UnaryOpTerm
			reduce(200), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2681
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2682
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(201), // >, reduce: UnaryOpTerm
			reduce(201), // >>, reduce: UnaryOpTerm
			reduce(201), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(201), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(201), // ?, reduce: UnaryOpTerm
			reduce(201), // :, reduce: UnaryOpTerm
			reduce(201), // ||, reduce: UnaryOpTerm
			reduce(201), // ^^, reduce: UnaryOpTerm
			reduce(201), // &&, reduce: UnaryOpTerm
			reduce(201), // ??, reduce: UnaryOpTerm
			reduce(201), // ???, reduce: UnaryOpTerm
			reduce(201), // =~, reduce: UnaryOpTerm
			reduce(201), // !=~, reduce: UnaryOpTerm
			reduce(201), // ==, reduce: UnaryOpTerm
			reduce(201), // !=, reduce: UnaryOpTerm
			reduce(201), // <=>, reduce: UnaryOpTerm
			reduce(201), // >=, reduce: UnaryOpTerm
			reduce(201), // <, reduce: UnaryOpTerm
			reduce(201), // <=, reduce: UnaryOpTerm
			reduce(201), // ^, reduce: UnaryOpTerm
			reduce(201), // &, reduce: UnaryOpTerm
			reduce(201), // <<, reduce: UnaryOpTerm
			reduce(201), // >>>, reduce: UnaryOpTerm
			reduce(201), // +, reduce: UnaryOpTerm
			reduce(201), // -, reduce: UnaryOpTerm
			reduce(201), // .+, reduce: UnaryOpTerm
			reduce(201), // .-, reduce: UnaryOpTerm
			reduce(201), // *, reduce: UnaryOpTerm
			reduce(201), // /, reduce: UnaryOpTerm
			reduce(201), // //, reduce: UnaryOpTerm
			reduce(201), // %, reduce: UnaryOpTerm
			reduce(201), // .*, reduce: UnaryOpTerm
			reduce(201), // ./, reduce: UnaryOpTerm
			reduce(201), // .//, reduce: UnaryOpTerm
			reduce(201), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2683
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(202), // >, reduce: UnaryOpTerm
			reduce(202), // >>, reduce: UnaryOpTerm
			reduce(202), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(202), // ], reduce: UnaryOpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(202), // ?, reduce: UnaryOpTerm
			reduce(202), // :, reduce: UnaryOpTerm
			reduce(202), // ||, reduce: UnaryOpTerm
			reduce(202), // ^^, reduce: UnaryOpTerm
			reduce(202), // &&, reduce: UnaryOpTerm
			reduce(202), // ??, reduce: UnaryOpTerm
			reduce(202), // ???, reduce: UnaryOpTerm
			reduce(202), // =~, reduce: UnaryOpTerm
			reduce(202), // !=~, reduce: UnaryOpTerm
			reduce(202), // ==, reduce: UnaryOpTerm
			reduce(202), // !=, reduce: UnaryOpTerm
			reduce(202), // <=>, reduce: UnaryOpTerm
			reduce(202), // >=, reduce: UnaryOpTerm
			reduce(202), // <, reduce: UnaryOpTerm
			reduce(202), // <=, reduce: UnaryOpTerm
			reduce(202), // ^, reduce: UnaryOpTerm
			reduce(202), // &, reduce: UnaryOpTerm
			reduce(202), // <<, reduce: UnaryOpTerm
			reduce(202), // >>>, reduce: UnaryOpTerm
			reduce(202), // +, reduce: UnaryOpTerm
			reduce(202), // -, reduce: UnaryOpTerm
			reduce(202), // .+, reduce: UnaryOpTerm
			reduce(202), // .-, reduce: UnaryOpTerm
			reduce(202), // *, reduce: UnaryOpTerm
			reduce(202), // /, reduce: UnaryOpTerm
			reduce(202), // //, reduce: UnaryOpTerm
			reduce(202), // %, reduce: UnaryOpTerm
			reduce(202), // .*, reduce: UnaryOpTerm
			reduce(202), // ./, reduce: UnaryOpTerm
			reduce(202), // .//, reduce: UnaryOpTerm
			reduce(202), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2684
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3361), // +
			shift(3362), // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2685
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3365), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2686
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2687
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2688
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(228), // >, reduce: ArrayLiteral
			reduce(228), // >>, reduce: ArrayLiteral
			reduce(228), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(228), // ], reduce: ArrayLiteral
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(228), // ?, reduce: ArrayLiteral
			reduce(228), // :, reduce: ArrayLiteral
			reduce(228), // ||, reduce: ArrayLiteral
			reduce(228), // ^^, reduce: ArrayLiteral
			reduce(228), // &&, reduce: ArrayLiteral
			reduce(228), // ??, reduce: ArrayLiteral
			reduce(228), // ???, reduce: ArrayLiteral
			reduce(228), // =~, reduce: ArrayLiteral
			reduce(228), // !=~, reduce: ArrayLiteral
			reduce(228), // ==, reduce: ArrayLiteral
			reduce(228), // !=, reduce: ArrayLiteral
			reduce(228), // <=>, reduce: ArrayLiteral
			reduce(228), // >=, reduce: ArrayLiteral
			reduce(228), // <, reduce: ArrayLiteral
			reduce(228), // <=, reduce: ArrayLiteral
			reduce(228), // ^, reduce: ArrayLiteral
			reduce(228), // &, reduce: ArrayLiteral
			reduce(228), // <<, reduce: ArrayLiteral
			reduce(228), // >>>, reduce: ArrayLiteral
			reduce(228), // +, reduce: ArrayLiteral
			reduce(228), // -, reduce: ArrayLiteral
			reduce(228), // .+, reduce: ArrayLiteral
			reduce(228), // .-, reduce: ArrayLiteral
			reduce(228), // *, reduce: ArrayLiteral
			reduce(228), // /, reduce: ArrayLiteral
			reduce(228), // //, reduce: ArrayLiteral
			reduce(228), // %, reduce: ArrayLiteral
			reduce(228), // .*, reduce: ArrayLiteral
			reduce(228), // ./, reduce: ArrayLiteral
			reduce(228), // .//, reduce: ArrayLiteral
			reduce(228), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(228), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(228), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(228), // [[, reduce: ArrayLiteral
			reduce(228), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2689
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3368), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2690
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(3369), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2691
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2692
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(3372), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2693
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2694
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3374), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2695
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3375), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2696
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(235), // >, reduce: MapLiteral
			reduce(235), // >>, reduce: MapLiteral
			reduce(235), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(235), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(235), // ], reduce: MapLiteral
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(235), // ?, reduce: MapLiteral
			nil,         // :
			reduce(235), // ||, reduce: MapLiteral
			reduce(235), // ^^, reduce: MapLiteral
			reduce(235), // &&, reduce: MapLiteral
			reduce(235), // ??, reduce: MapLiteral
			reduce(235), // ???, reduce: MapLiteral
			reduce(235), // =~, reduce: MapLiteral
			reduce(235), // !=~, reduce: MapLiteral
			reduce(235), // ==, reduce: MapLiteral
			reduce(235), // !=, reduce: MapLiteral
			reduce(235), // <=>, reduce: MapLiteral
			reduce(235), // >=, reduce: MapLiteral
			reduce(235), // <, reduce: MapLiteral
			reduce(235), // <=, reduce: MapLiteral
			reduce(235), // ^, reduce: MapLiteral
			reduce(235), // &, reduce: MapLiteral
			reduce(235), // <<, reduce: MapLiteral
			reduce(235), // >>>, reduce: MapLiteral
			reduce(235), // +, reduce: MapLiteral
			reduce(235), // -, reduce: MapLiteral
			reduce(235), // .+, reduce: MapLiteral
			reduce(235), // .-, reduce: MapLiteral
			reduce(235), // *, reduce: MapLiteral
			reduce(235), // /, reduce: MapLiteral
			reduce(235), // //, reduce: MapLiteral
			reduce(235), // %, reduce: MapLiteral
			reduce(235), // .*, reduce: MapLiteral
			reduce(235), // ./, reduce: MapLiteral
			reduce(235), // .//, reduce: MapLiteral
			reduce(235), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(235), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(235), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(235), // [[, reduce: MapLiteral
			reduce(235), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2697
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(232), // ], reduce: ArrayLiteralElements
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2698
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(208), // >, reduce: PrecedenceChainEnd
			reduce(208), // >>, reduce: PrecedenceChainEnd
			reduce(208), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(208), // ,, reduce: PrecedenceChainEnd
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(208), // ], reduce: PrecedenceChainEnd
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(208), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(208), // ||, reduce: PrecedenceChainEnd
			reduce(208), // ^^, reduce: PrecedenceChainEnd
			reduce(208), // &&, reduce: PrecedenceChainEnd
			reduce(208), // ??, reduce: PrecedenceChainEnd
			reduce(208), // ???, reduce: PrecedenceChainEnd
			reduce(208), // =~, reduce: PrecedenceChainEnd
			reduce(208), // !=~, reduce: PrecedenceChainEnd
			reduce(208), // ==, reduce: PrecedenceChainEnd
			reduce(208), // !=, reduce: PrecedenceChainEnd
			reduce(208), // <=>, reduce: PrecedenceChainEnd
			reduce(208), // >=, reduce: PrecedenceChainEnd
			reduce(208), // <, reduce: PrecedenceChainEnd
			reduce(208), // <=, reduce: PrecedenceChainEnd
			reduce(208), // ^, reduce: PrecedenceChainEnd
			reduce(208), // &, reduce: PrecedenceChainEnd
			reduce(208), // <<, reduce: PrecedenceChainEnd
			reduce(208), // >>>, reduce: PrecedenceChainEnd
			reduce(208), // +, reduce: PrecedenceChainEnd
			reduce(208), // -, reduce: PrecedenceChainEnd
			reduce(208), // .+, reduce: PrecedenceChainEnd
			reduce(208), // .-, reduce: PrecedenceChainEnd
			reduce(208), // *, reduce: PrecedenceChainEnd
			reduce(208), // /, reduce: PrecedenceChainEnd
			reduce(208), // //, reduce: PrecedenceChainEnd
			reduce(208), // %, reduce: PrecedenceChainEnd
			reduce(208), // .*, reduce: PrecedenceChainEnd
			reduce(208), // ./, reduce: PrecedenceChainEnd
			reduce(208), // .//, reduce: PrecedenceChainEnd
			reduce(208), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(208), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2699
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(105), // >, reduce: IndirectFieldValue
			reduce(105), // >>, reduce: IndirectFieldValue
			reduce(105), // |, reduce: IndirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(105), // ,, reduce: IndirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(105), // ], reduce: IndirectFieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(105), // ?, reduce: IndirectFieldValue
			nil,         // :
			reduce(105), // ||, reduce: IndirectFieldValue
			reduce(105), // ^^, reduce: IndirectFieldValue
			reduce(105), // &&, reduce: IndirectFieldValue
			reduce(105), // ??, reduce: IndirectFieldValue
			reduce(105), // ???, reduce: IndirectFieldValue
			reduce(105), // =~, reduce: IndirectFieldValue
			reduce(105), // !=~, reduce: IndirectFieldValue
			reduce(105), // ==, reduce: IndirectFieldValue
			reduce(105), // !=, reduce: IndirectFieldValue
			reduce(105), // <=>, reduce: IndirectFieldValue
			reduce(105), // >=, reduce: IndirectFieldValue
			reduce(105), // <, reduce: IndirectFieldValue
			reduce(105), // <=, reduce: IndirectFieldValue
			reduce(105), // ^, reduce: IndirectFieldValue
			reduce(105), // &, reduce: IndirectFieldValue
			reduce(105), // <<, reduce: IndirectFieldValue
			reduce(105), // >>>, reduce: IndirectFieldValue
			reduce(105), // +, reduce: IndirectFieldValue
			reduce(105), // -, reduce: IndirectFieldValue
			reduce(105), // .+, reduce: IndirectFieldValue
			reduce(105), // .-, reduce: IndirectFieldValue
			reduce(105), // *, reduce: IndirectFieldValue
			reduce(105), // /, reduce: IndirectFieldValue
			reduce(105), // //, reduce: IndirectFieldValue
			reduce(105), // %, reduce: IndirectFieldValue
			reduce(105), // .*, reduce: IndirectFieldValue
			reduce(105), // ./, reduce: IndirectFieldValue
			reduce(105), // .//, reduce: IndirectFieldValue
			reduce(105), // ., reduce: IndirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(105), // **, reduce: IndirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(105), // [, reduce: IndirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(105), // [[, reduce: IndirectFieldValue
			reduce(105), // [[[, reduce: IndirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2700
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3376), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2701
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3377), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2702
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(114), // >, reduce: IndirectOosvarValue
			reduce(114), // >>, reduce: IndirectOosvarValue
			reduce(114), // |, reduce: IndirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(114), // ,, reduce: IndirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(114), // ], reduce: IndirectOosvarValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(114), // ?, reduce: IndirectOosvarValue
			nil,         // :
			reduce(114), // ||, reduce: IndirectOosvarValue
			reduce(114), // ^^, reduce: IndirectOosvarValue
			reduce(114), // &&, reduce: IndirectOosvarValue
			reduce(114), // ??, reduce: IndirectOosvarValue
			reduce(114), // ???, reduce: IndirectOosvarValue
			reduce(114), // =~, reduce: IndirectOosvarValue
			reduce(114), // !=~, reduce: IndirectOosvarValue
			reduce(114), // ==, reduce: IndirectOosvarValue
			reduce(114), // !=, reduce: IndirectOosvarValue
			reduce(114), // <=>, reduce: IndirectOosvarValue
			reduce(114), // >=, reduce: IndirectOosvarValue
			reduce(114), // <, reduce: IndirectOosvarValue
			reduce(114), // <=, reduce: IndirectOosvarValue
			reduce(114), // ^, reduce: IndirectOosvarValue
			reduce(114), // &, reduce: IndirectOosvarValue
			reduce(114), // <<, reduce: IndirectOosvarValue
			reduce(114), // >>>, reduce: IndirectOosvarValue
			reduce(114), // +, reduce: IndirectOosvarValue
			reduce(114), // -, reduce: IndirectOosvarValue
			reduce(114), // .+, reduce: IndirectOosvarValue
			reduce(114), // .-, reduce: IndirectOosvarValue
			reduce(114), // *, reduce: IndirectOosvarValue
			reduce(114), // /, reduce: IndirectOosvarValue
			reduce(114), // //, reduce: IndirectOosvarValue
			reduce(114), // %, reduce: IndirectOosvarValue
			reduce(114), // .*, reduce: IndirectOosvarValue
			reduce(114), // ./, reduce: IndirectOosvarValue
			reduce(114), // .//, reduce: IndirectOosvarValue
			reduce(114), // ., reduce: IndirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(114), // **, reduce: IndirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(114), // [, reduce: IndirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(114), // [[, reduce: IndirectOosvarValue
			reduce(114), // [[[, reduce: IndirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2703
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3378), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2704
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(151), // ,, reduce: LogicalOrTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(151), // ], reduce: LogicalOrTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(151), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(151), // ||, reduce: LogicalOrTerm
			shift(1613), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2705
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(153), // ,, reduce: LogicalXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(153), // ], reduce: LogicalXORTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(153), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(153), // ||, reduce: LogicalXORTerm
			reduce(153), // ^^, reduce: LogicalXORTerm
			shift(1614), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2706
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(155), // ,, reduce: LogicalAndTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(155), // ], reduce: LogicalAndTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(155), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(155), // ||, reduce: LogicalAndTerm
			reduce(155), // ^^, reduce: LogicalAndTerm
			reduce(155), // &&, reduce: LogicalAndTerm
			shift(1615), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2707
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(157), // ,, reduce: AbsentCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(157), // ], reduce: AbsentCoalesceTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(157), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(157), // ||, reduce: AbsentCoalesceTerm
			reduce(157), // ^^, reduce: AbsentCoalesceTerm
			reduce(157), // &&, reduce: AbsentCoalesceTerm
			reduce(157), // ??, reduce: AbsentCoalesceTerm
			shift(1616), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2708
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(159), // ,, reduce: EmptyCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(159), // ], reduce: EmptyCoalesceTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(159), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(159), // ||, reduce: EmptyCoalesceTerm
			reduce(159), // ^^, reduce: EmptyCoalesceTerm
			reduce(159), // &&, reduce: EmptyCoalesceTerm
			reduce(159), // ??, reduce: EmptyCoalesceTerm
			reduce(159), // ???, reduce: EmptyCoalesceTerm
			shift(1617), // =~
			shift(1618), // !=~
			shift(1619), // ==
			shift(1620), // !=
			shift(1621), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2709
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1622), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(161), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(161), // ], reduce: EqneTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(161), // ?, reduce: EqneTerm
			nil,         // :
			reduce(161), // ||, reduce: EqneTerm
			reduce(161), // ^^, reduce: EqneTerm
			reduce(161), // &&, reduce: EqneTerm
			reduce(161), // ??, reduce: EqneTerm
			reduce(161), // ???, reduce: EqneTerm
			reduce(161), // =~, reduce: EqneTerm
			reduce(161), // !=~, reduce: EqneTerm
			reduce(161), // ==, reduce: EqneTerm
			reduce(161), // !=, reduce: EqneTerm
			reduce(161), // <=>, reduce: EqneTerm
			shift(1623), // >=
			shift(1624), // <
			shift(1625), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2710
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1622), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(162), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(162), // ], reduce: EqneTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(162), // ?, reduce: EqneTerm
			nil,         // :
			reduce(162), // ||, reduce: EqneTerm
			reduce(162), // ^^, reduce: EqneTerm
			reduce(162), // &&, reduce: EqneTerm
			reduce(162), // ??, reduce: EqneTerm
			reduce(162), // ???, reduce: EqneTerm
			reduce(162), // =~, reduce: EqneTerm
			reduce(162), // !=~, reduce: EqneTerm
			reduce(162), // ==, reduce: EqneTerm
			reduce(162), // !=, reduce: EqneTerm
			reduce(162), // <=>, reduce: EqneTerm
			shift(1623), // >=
			shift(1624), // <
			shift(1625), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2711
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1622), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(163), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(163), // ], reduce: EqneTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(163), // ?, reduce: EqneTerm
			nil,         // :
			reduce(163), // ||, reduce: EqneTerm
			reduce(163), // ^^, reduce: EqneTerm
			reduce(163), // &&, reduce: EqneTerm
			reduce(163), // ??, reduce: EqneTerm
			reduce(163), // ???, reduce: EqneTerm
			reduce(163), // =~, reduce: EqneTerm
			reduce(163), // !=~, reduce: EqneTerm
			reduce(163), // ==, reduce: EqneTerm
			reduce(163), // !=, reduce: EqneTerm
			reduce(163), // <=>, reduce: EqneTerm
			shift(1623), // >=
			shift(1624), // <
			shift(1625), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2712
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1622), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(164), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(164), // ], reduce: EqneTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(164), // ?, reduce: EqneTerm
			nil,         // :
			reduce(164), // ||, reduce: EqneTerm
			reduce(164), // ^^, reduce: EqneTerm
			reduce(164), // &&, reduce: EqneTerm
			reduce(164), // ??, reduce: EqneTerm
			reduce(164), // ???, reduce: EqneTerm
			reduce(164), // =~, reduce: EqneTerm
			reduce(164), // !=~, reduce: EqneTerm
			reduce(164), // ==, reduce: EqneTerm
			reduce(164), // !=, reduce: EqneTerm
			reduce(164), // <=>, reduce: EqneTerm
			shift(1623), // >=
			shift(1624), // <
			shift(1625), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2713
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(1622), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(165), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(165), // ], reduce: EqneTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(165), // ?, reduce: EqneTerm
			nil,         // :
			reduce(165), // ||, reduce: EqneTerm
			reduce(165), // ^^, reduce: EqneTerm
			reduce(165), // &&, reduce: EqneTerm
			reduce(165), // ??, reduce: EqneTerm
			reduce(165), // ???, reduce: EqneTerm
			reduce(165), // =~, reduce: EqneTerm
			reduce(165), // !=~, reduce: EqneTerm
			reduce(165), // ==, reduce: EqneTerm
			reduce(165), // !=, reduce: EqneTerm
			reduce(165), // <=>, reduce: EqneTerm
			shift(1623), // >=
			shift(1624), // <
			shift(1625), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2714
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(167), // >, reduce: CmpTerm
			nil,         // >>
			shift(1626), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(167), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(167), // ], reduce: CmpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(167), // ?, reduce: CmpTerm
			nil,         // :
			reduce(167), // ||, reduce: CmpTerm
			reduce(167), // ^^, reduce: CmpTerm
			reduce(167), // &&, reduce: CmpTerm
			reduce(167), // ??, reduce: CmpTerm
			reduce(167), // ???, reduce: CmpTerm
			reduce(167), // =~, reduce: CmpTerm
			reduce(167), // !=~, reduce: CmpTerm
			reduce(167), // ==, reduce: CmpTerm
			reduce(167), // !=, reduce: CmpTerm
			reduce(167), // <=>, reduce: CmpTerm
			reduce(167), // >=, reduce: CmpTerm
			reduce(167), // <, reduce: CmpTerm
			reduce(167), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2715
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(168), // >, reduce: CmpTerm
			nil,         // >>
			shift(1626), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(168), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(168), // ], reduce: CmpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(168), // ?, reduce: CmpTerm
			nil,         // :
			reduce(168), // ||, reduce: CmpTerm
			reduce(168), // ^^, reduce: CmpTerm
			reduce(168), // &&, reduce: CmpTerm
			reduce(168), // ??, reduce: CmpTerm
			reduce(168), // ???, reduce: CmpTerm
			reduce(168), // =~, reduce: CmpTerm
			reduce(168), // !=~, reduce: CmpTerm
			reduce(168), // ==, reduce: CmpTerm
			reduce(168), // !=, reduce: CmpTerm
			reduce(168), // <=>, reduce: CmpTerm
			reduce(168), // >=, reduce: CmpTerm
			reduce(168), // <, reduce: CmpTerm
			reduce(168), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2716
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(169), // >, reduce: CmpTerm
			nil,         // >>
			shift(1626), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(169), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(169), // ], reduce: CmpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(169), // ?, reduce: CmpTerm
			nil,         // :
			reduce(169), // ||, reduce: CmpTerm
			reduce(169), // ^^, reduce: CmpTerm
			reduce(169), // &&, reduce: CmpTerm
			reduce(169), // ??, reduce: CmpTerm
			reduce(169), // ???, reduce: CmpTerm
			reduce(169), // =~, reduce: CmpTerm
			reduce(169), // !=~, reduce: CmpTerm
			reduce(169), // ==, reduce: CmpTerm
			reduce(169), // !=, reduce: CmpTerm
			reduce(169), // <=>, reduce: CmpTerm
			reduce(169), // >=, reduce: CmpTerm
			reduce(169), // <, reduce: CmpTerm
			reduce(169), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2717
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(170), // >, reduce: CmpTerm
			nil,         // >>
			shift(1626), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(170), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(170), // ], reduce: CmpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(170), // ?, reduce: CmpTerm
			nil,         // :
			reduce(170), // ||, reduce: CmpTerm
			reduce(170), // ^^, reduce: CmpTerm
			reduce(170), // &&, reduce: CmpTerm
			reduce(170), // ??, reduce: CmpTerm
			reduce(170), // ???, reduce: CmpTerm
			reduce(170), // =~, reduce: CmpTerm
			reduce(170), // !=~, reduce: CmpTerm
			reduce(170), // ==, reduce: CmpTerm
			reduce(170), // !=, reduce: CmpTerm
			reduce(170), // <=>, reduce: CmpTerm
			reduce(170), // >=, reduce: CmpTerm
			reduce(170), // <, reduce: CmpTerm
			reduce(170), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2718
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(172), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(172), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(172), // ,, reduce: BitwiseORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(172), // ], reduce: BitwiseORTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(172), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(172), // ||, reduce: BitwiseORTerm
			reduce(172), // ^^, reduce: BitwiseORTerm
			reduce(172), // &&, reduce: BitwiseORTerm
			reduce(172), // ??, reduce: BitwiseORTerm
			reduce(172), // ???, reduce: BitwiseORTerm
			reduce(172), // =~, reduce: BitwiseORTerm
			reduce(172), // !=~, reduce: BitwiseORTerm
			reduce(172), // ==, reduce: BitwiseORTerm
			reduce(172), // !=, reduce: BitwiseORTerm
			reduce(172), // <=>, reduce: BitwiseORTerm
			reduce(172), // >=, reduce: BitwiseORTerm
			reduce(172), // <, reduce: BitwiseORTerm
			reduce(172), // <=, reduce: BitwiseORTerm
			shift(1627), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2719
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(174), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(174), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(174), // ,, reduce: BitwiseXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(174), // ], reduce: BitwiseXORTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(174), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(174), // ||, reduce: BitwiseXORTerm
			reduce(174), // ^^, reduce: BitwiseXORTerm
			reduce(174), // &&, reduce: BitwiseXORTerm
			reduce(174), // ??, reduce: BitwiseXORTerm
			reduce(174), // ???, reduce: BitwiseXORTerm
			reduce(174), // =~, reduce: BitwiseXORTerm
			reduce(174), // !=~, reduce: BitwiseXORTerm
			reduce(174), // ==, reduce: BitwiseXORTerm
			reduce(174), // !=, reduce: BitwiseXORTerm
			reduce(174), // <=>, reduce: BitwiseXORTerm
			reduce(174), // >=, reduce: BitwiseXORTerm
			reduce(174), // <, reduce: BitwiseXORTerm
			reduce(174), // <=, reduce: BitwiseXORTerm
			reduce(174), // ^, reduce: BitwiseXORTerm
			shift(1628), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2720
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(176), // >, reduce: BitwiseANDTerm
			shift(1629), // >>
			reduce(176), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(176), // ,, reduce: BitwiseANDTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(176), // ], reduce: BitwiseANDTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(176), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(176), // ||, reduce: BitwiseANDTerm
			reduce(176), // ^^, reduce: BitwiseANDTerm
			reduce(176), // &&, reduce: BitwiseANDTerm
			reduce(176), // ??, reduce: BitwiseANDTerm
			reduce(176), // ???, reduce: BitwiseANDTerm
			reduce(176), // =~, reduce: BitwiseANDTerm
			reduce(176), // !=~, reduce: BitwiseANDTerm
			reduce(176), // ==, reduce: BitwiseANDTerm
			reduce(176), // !=, reduce: BitwiseANDTerm
			reduce(176), // <=>, reduce: BitwiseANDTerm
			reduce(176), // >=, reduce: BitwiseANDTerm
			reduce(176), // <, reduce: BitwiseANDTerm
			reduce(176), // <=, reduce: BitwiseANDTerm
			reduce(176), // ^, reduce: BitwiseANDTerm
			reduce(176), // &, reduce: BitwiseANDTerm
			shift(1630), // <<
			shift(1631), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2721
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(179), // >, reduce: BitwiseShiftTerm
			reduce(179), // >>, reduce: BitwiseShiftTerm
			reduce(179), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(179), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(179), // ], reduce: BitwiseShiftTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(179), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(179), // ||, reduce: BitwiseShiftTerm
			reduce(179), // ^^, reduce: BitwiseShiftTerm
			reduce(179), // &&, reduce: BitwiseShiftTerm
			reduce(179), // ??, reduce: BitwiseShiftTerm
			reduce(179), // ???, reduce: BitwiseShiftTerm
			reduce(179), // =~, reduce: BitwiseShiftTerm
			reduce(179), // !=~, reduce: BitwiseShiftTerm
			reduce(179), // ==, reduce: BitwiseShiftTerm
			reduce(179), // !=, reduce: BitwiseShiftTerm
			reduce(179), // <=>, reduce: BitwiseShiftTerm
			reduce(179), // >=, reduce: BitwiseShiftTerm
			reduce(179), // <, reduce: BitwiseShiftTerm
			reduce(179), // <=, reduce: BitwiseShiftTerm
			reduce(179), // ^, reduce: BitwiseShiftTerm
			reduce(179), // &, reduce: BitwiseShiftTerm
			reduce(179), // <<, reduce: BitwiseShiftTerm
			reduce(179), // >>>, reduce: BitwiseShiftTerm
			shift(1632), // +
			shift(1633), // -
			shift(1634), // .+
			shift(1635), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2722
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(178), // >, reduce: BitwiseShiftTerm
			reduce(178), // >>, reduce: BitwiseShiftTerm
			reduce(178), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(178), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(178), // ], reduce: BitwiseShiftTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(178), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(178), // ||, reduce: BitwiseShiftTerm
			reduce(178), // ^^, reduce: BitwiseShiftTerm
			reduce(178), // &&, reduce: BitwiseShiftTerm
			reduce(178), // ??, reduce: BitwiseShiftTerm
			reduce(178), // ???, reduce: BitwiseShiftTerm
			reduce(178), // =~, reduce: BitwiseShiftTerm
			reduce(178), // !=~, reduce: BitwiseShiftTerm
			reduce(178), // ==, reduce: BitwiseShiftTerm
			reduce(178), // !=, reduce: BitwiseShiftTerm
			reduce(178), // <=>, reduce: BitwiseShiftTerm
			reduce(178), // >=, reduce: BitwiseShiftTerm
			reduce(178), // <, reduce: BitwiseShiftTerm
			reduce(178), // <=, reduce: BitwiseShiftTerm
			reduce(178), // ^, reduce: BitwiseShiftTerm
			reduce(178), // &, reduce: BitwiseShiftTerm
			reduce(178), // <<, reduce: BitwiseShiftTerm
			reduce(178), // >>>, reduce: BitwiseShiftTerm
			shift(1632), // +
			shift(1633), // -
			shift(1634), // .+
			shift(1635), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2723
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(180), // >, reduce: BitwiseShiftTerm
			reduce(180), // >>, reduce: BitwiseShiftTerm
			reduce(180), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(180), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(180), // ], reduce: BitwiseShiftTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(180), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(180), // ||, reduce: BitwiseShiftTerm
			reduce(180), // ^^, reduce: BitwiseShiftTerm
			reduce(180), // &&, reduce: BitwiseShiftTerm
			reduce(180), // ??, reduce: BitwiseShiftTerm
			reduce(180), // ???, reduce: BitwiseShiftTerm
			reduce(180), // =~, reduce: BitwiseShiftTerm
			reduce(180), // !=~, reduce: BitwiseShiftTerm
			reduce(180), // ==, reduce: BitwiseShiftTerm
			reduce(180), // !=, reduce: BitwiseShiftTerm
			reduce(180), // <=>, reduce: BitwiseShiftTerm
			reduce(180), // >=, reduce: BitwiseShiftTerm
			reduce(180), // <, reduce: BitwiseShiftTerm
			reduce(180), // <=, reduce: BitwiseShiftTerm
			reduce(180), // ^, reduce: BitwiseShiftTerm
			reduce(180), // &, reduce: BitwiseShiftTerm
			reduce(180), // <<, reduce: BitwiseShiftTerm
			reduce(180), // >>>, reduce: BitwiseShiftTerm
			shift(1632), // +
			shift(1633), // -
			shift(1634), // .+
			shift(1635), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2724
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(182), // >, reduce: AddsubdotTerm
			reduce(182), // >>, reduce: AddsubdotTerm
			reduce(182), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(182), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(182), // ], reduce: AddsubdotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(182), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(182), // ||, reduce: AddsubdotTerm
			reduce(182), // ^^, reduce: AddsubdotTerm
			reduce(182), // &&, reduce: AddsubdotTerm
			reduce(182), // ??, reduce: AddsubdotTerm
			reduce(182), // ???, reduce: AddsubdotTerm
			reduce(182), // =~, reduce: AddsubdotTerm
			reduce(182), // !=~, reduce: AddsubdotTerm
			reduce(182), // ==, reduce: AddsubdotTerm
			reduce(182), // !=, reduce: AddsubdotTerm
			reduce(182), // <=>, reduce: AddsubdotTerm
			reduce(182), // >=, reduce: AddsubdotTerm
			reduce(182), // <, reduce: AddsubdotTerm
			reduce(182), // <=, reduce: AddsubdotTerm
			reduce(182), // ^, reduce: AddsubdotTerm
			reduce(182), // &, reduce: AddsubdotTerm
			reduce(182), // <<, reduce: AddsubdotTerm
			reduce(182), // >>>, reduce: AddsubdotTerm
			reduce(182), // +, reduce: AddsubdotTerm
			reduce(182), // -, reduce: AddsubdotTerm
			reduce(182), // .+, reduce: AddsubdotTerm
			reduce(182), // .-, reduce: AddsubdotTerm
			shift(1637), // *
			shift(1638), // /
			shift(1639), // //
			shift(1640), // %
			shift(1641), // .*
			shift(1642), // ./
			shift(1643), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2725
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(183), // >, reduce: AddsubdotTerm
			reduce(183), // >>, reduce: AddsubdotTerm
			reduce(183), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(183), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(183), // ], reduce: AddsubdotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(183), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(183), // ||, reduce: AddsubdotTerm
			reduce(183), // ^^, reduce: AddsubdotTerm
			reduce(183), // &&, reduce: AddsubdotTerm
			reduce(183), // ??, reduce: AddsubdotTerm
			reduce(183), // ???, reduce: AddsubdotTerm
			reduce(183), // =~, reduce: AddsubdotTerm
			reduce(183), // !=~, reduce: AddsubdotTerm
			reduce(183), // ==, reduce: AddsubdotTerm
			reduce(183), // !=, reduce: AddsubdotTerm
			reduce(183), // <=>, reduce: AddsubdotTerm
			reduce(183), // >=, reduce: AddsubdotTerm
			reduce(183), // <, reduce: AddsubdotTerm
			reduce(183), // <=, reduce: AddsubdotTerm
			reduce(183), // ^, reduce: AddsubdotTerm
			reduce(183), // &, reduce: AddsubdotTerm
			reduce(183), // <<, reduce: AddsubdotTerm
			reduce(183), // >>>, reduce: AddsubdotTerm
			reduce(183), // +, reduce: AddsubdotTerm
			reduce(183), // -, reduce: AddsubdotTerm
			reduce(183), // .+, reduce: AddsubdotTerm
			reduce(183), // .-, reduce: AddsubdotTerm
			shift(1637), // *
			shift(1638), // /
			shift(1639), // //
			shift(1640), // %
			shift(1641), // .*
			shift(1642), // ./
			shift(1643), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2726
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(184), // >, reduce: AddsubdotTerm
			reduce(184), // >>, reduce: AddsubdotTerm
			reduce(184), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(184), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(184), // ], reduce: AddsubdotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(184), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(184), // ||, reduce: AddsubdotTerm
			reduce(184), // ^^, reduce: AddsubdotTerm
			reduce(184), // &&, reduce: AddsubdotTerm
			reduce(184), // ??, reduce: AddsubdotTerm
			reduce(184), // ???, reduce: AddsubdotTerm
			reduce(184), // =~, reduce: AddsubdotTerm
			reduce(184), // !=~, reduce: AddsubdotTerm
			reduce(184), // ==, reduce: AddsubdotTerm
			reduce(184), // !=, reduce: AddsubdotTerm
			reduce(184), // <=>, reduce: AddsubdotTerm
			reduce(184), // >=, reduce: AddsubdotTerm
			reduce(184), // <, reduce: AddsubdotTerm
			reduce(184), // <=, reduce: AddsubdotTerm
			reduce(184), // ^, reduce: AddsubdotTerm
			reduce(184), // &, reduce: AddsubdotTerm
			reduce(184), // <<, reduce: AddsubdotTerm
			reduce(184), // >>>, reduce: AddsubdotTerm
			reduce(184), // +, reduce: AddsubdotTerm
			reduce(184), // -, reduce: AddsubdotTerm
			reduce(184), // .+, reduce: AddsubdotTerm
			reduce(184), // .-, reduce: AddsubdotTerm
			shift(1637), // *
			shift(1638), // /
			shift(1639), // //
			shift(1640), // %
			shift(1641), // .*
			shift(1642), // ./
			shift(1643), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2727
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(185), // >, reduce: AddsubdotTerm
			reduce(185), // >>, reduce: AddsubdotTerm
			reduce(185), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(185), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(185), // ], reduce: AddsubdotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(185), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(185), // ||, reduce: AddsubdotTerm
			reduce(185), // ^^, reduce: AddsubdotTerm
			reduce(185), // &&, reduce: AddsubdotTerm
			reduce(185), // ??, reduce: AddsubdotTerm
			reduce(185), // ???, reduce: AddsubdotTerm
			reduce(185), // =~, reduce: AddsubdotTerm
			reduce(185), // !=~, reduce: AddsubdotTerm
			reduce(185), // ==, reduce: AddsubdotTerm
			reduce(185), // !=, reduce: AddsubdotTerm
			reduce(185), // <=>, reduce: AddsubdotTerm
			reduce(185), // >=, reduce: AddsubdotTerm
			reduce(185), // <, reduce: AddsubdotTerm
			reduce(185), // <=, reduce: AddsubdotTerm
			reduce(185), // ^, reduce: AddsubdotTerm
			reduce(185), // &, reduce: AddsubdotTerm
			reduce(185), // <<, reduce: AddsubdotTerm
			reduce(185), // >>>, reduce: AddsubdotTerm
			reduce(185), // +, reduce: AddsubdotTerm
			reduce(185), // -, reduce: AddsubdotTerm
			reduce(185), // .+, reduce: AddsubdotTerm
			reduce(185), // .-, reduce: AddsubdotTerm
			shift(1637), // *
			shift(1638), // /
			shift(1639), // //
			shift(1640), // %
			shift(1641), // .*
			shift(1642), // ./
			shift(1643), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2728
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(187), // >, reduce: MuldivTerm
			reduce(187), // >>, reduce: MuldivTerm
			reduce(187), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(187), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(187), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(187), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(187), // ||, reduce: MuldivTerm
			reduce(187), // ^^, reduce: MuldivTerm
			reduce(187), // &&, reduce: MuldivTerm
			reduce(187), // ??, reduce: MuldivTerm
			reduce(187), // ???, reduce: MuldivTerm
			reduce(187), // =~, reduce: MuldivTerm
			reduce(187), // !=~, reduce: MuldivTerm
			reduce(187), // ==, reduce: MuldivTerm
			reduce(187), // !=, reduce: MuldivTerm
			reduce(187), // <=>, reduce: MuldivTerm
			reduce(187), // >=, reduce: MuldivTerm
			reduce(187), // <, reduce: MuldivTerm
			reduce(187), // <=, reduce: MuldivTerm
			reduce(187), // ^, reduce: MuldivTerm
			reduce(187), // &, reduce: MuldivTerm
			reduce(187), // <<, reduce: MuldivTerm
			reduce(187), // >>>, reduce: MuldivTerm
			reduce(187), // +, reduce: MuldivTerm
			reduce(187), // -, reduce: MuldivTerm
			reduce(187), // .+, reduce: MuldivTerm
			reduce(187), // .-, reduce: MuldivTerm
			reduce(187), // *, reduce: MuldivTerm
			reduce(187), // /, reduce: MuldivTerm
			reduce(187), // //, reduce: MuldivTerm
			reduce(187), // %, reduce: MuldivTerm
			reduce(187), // .*, reduce: MuldivTerm
			reduce(187), // ./, reduce: MuldivTerm
			reduce(187), // .//, reduce: MuldivTerm
			shift(1647), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2729
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(188), // >, reduce: MuldivTerm
			reduce(188), // >>, reduce: MuldivTerm
			reduce(188), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(188), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(188), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(188), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(188), // ||, reduce: MuldivTerm
			reduce(188), // ^^, reduce: MuldivTerm
			reduce(188), // &&, reduce: MuldivTerm
			reduce(188), // ??, reduce: MuldivTerm
			reduce(188), // ???, reduce: MuldivTerm
			reduce(188), // =~, reduce: MuldivTerm
			reduce(188), // !=~, reduce: MuldivTerm
			reduce(188), // ==, reduce: MuldivTerm
			reduce(188), // !=, reduce: MuldivTerm
			reduce(188), // <=>, reduce: MuldivTerm
			reduce(188), // >=, reduce: MuldivTerm
			reduce(188), // <, reduce: MuldivTerm
			reduce(188), // <=, reduce: MuldivTerm
			reduce(188), // ^, reduce: MuldivTerm
			reduce(188), // &, reduce: MuldivTerm
			reduce(188), // <<, reduce: MuldivTerm
			reduce(188), // >>>, reduce: MuldivTerm
			reduce(188), // +, reduce: MuldivTerm
			reduce(188), // -, reduce: MuldivTerm
			reduce(188), // .+, reduce: MuldivTerm
			reduce(188), // .-, reduce: MuldivTerm
			reduce(188), // *, reduce: MuldivTerm
			reduce(188), // /, reduce: MuldivTerm
			reduce(188), // //, reduce: MuldivTerm
			reduce(188), // %, reduce: MuldivTerm
			reduce(188), // .*, reduce: MuldivTerm
			reduce(188), // ./, reduce: MuldivTerm
			reduce(188), // .//, reduce: MuldivTerm
			shift(1647), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2730
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(189), // >, reduce: MuldivTerm
			reduce(189), // >>, reduce: MuldivTerm
			reduce(189), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(189), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(189), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(189), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(189), // ||, reduce: MuldivTerm
			reduce(189), // ^^, reduce: MuldivTerm
			reduce(189), // &&, reduce: MuldivTerm
			reduce(189), // ??, reduce: MuldivTerm
			reduce(189), // ???, reduce: MuldivTerm
			reduce(189), // =~, reduce: MuldivTerm
			reduce(189), // !=~, reduce: MuldivTerm
			reduce(189), // ==, reduce: MuldivTerm
			reduce(189), // !=, reduce: MuldivTerm
			reduce(189), // <=>, reduce: MuldivTerm
			reduce(189), // >=, reduce: MuldivTerm
			reduce(189), // <, reduce: MuldivTerm
			reduce(189), // <=, reduce: MuldivTerm
			reduce(189), // ^, reduce: MuldivTerm
			reduce(189), // &, reduce: MuldivTerm
			reduce(189), // <<, reduce: MuldivTerm
			reduce(189), // >>>, reduce: MuldivTerm
			reduce(189), // +, reduce: MuldivTerm
			reduce(189), // -, reduce: MuldivTerm
			reduce(189), // .+, reduce: MuldivTerm
			reduce(189), // .-, reduce: MuldivTerm
			reduce(189), // *, reduce: MuldivTerm
			reduce(189), // /, reduce: MuldivTerm
			reduce(189), // //, reduce: MuldivTerm
			reduce(189), // %, reduce: MuldivTerm
			reduce(189), // .*, reduce: MuldivTerm
			reduce(189), // ./, reduce: MuldivTerm
			reduce(189), // .//, reduce: MuldivTerm
			shift(1647), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2731
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(190), // >, reduce: MuldivTerm
			reduce(190), // >>, reduce: MuldivTerm
			reduce(190), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(190), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(190), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(190), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(190), // ||, reduce: MuldivTerm
			reduce(190), // ^^, reduce: MuldivTerm
			reduce(190), // &&, reduce: MuldivTerm
			reduce(190), // ??, reduce: MuldivTerm
			reduce(190), // ???, reduce: MuldivTerm
			reduce(190), // =~, reduce: MuldivTerm
			reduce(190), // !=~, reduce: MuldivTerm
			reduce(190), // ==, reduce: MuldivTerm
			reduce(190), // !=, reduce: MuldivTerm
			reduce(190), // <=>, reduce: MuldivTerm
			reduce(190), // >=, reduce: MuldivTerm
			reduce(190), // <, reduce: MuldivTerm
			reduce(190), // <=, reduce: MuldivTerm
			reduce(190), // ^, reduce: MuldivTerm
			reduce(190), // &, reduce: MuldivTerm
			reduce(190), // <<, reduce: MuldivTerm
			reduce(190), // >>>, reduce: MuldivTerm
			reduce(190), // +, reduce: MuldivTerm
			reduce(190), // -, reduce: MuldivTerm
			reduce(190), // .+, reduce: MuldivTerm
			reduce(190), // .-, reduce: MuldivTerm
			reduce(190), // *, reduce: MuldivTerm
			reduce(190), // /, reduce: MuldivTerm
			reduce(190), // //, reduce: MuldivTerm
			reduce(190), // %, reduce: MuldivTerm
			reduce(190), // .*, reduce: MuldivTerm
			reduce(190), // ./, reduce: MuldivTerm
			reduce(190), // .//, reduce: MuldivTerm
			shift(1647), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2732
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(191), // >, reduce: MuldivTerm
			reduce(191), // >>, reduce: MuldivTerm
			reduce(191), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(191), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(191), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(191), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(191), // ||, reduce: MuldivTerm
			reduce(191), // ^^, reduce: MuldivTerm
			reduce(191), // &&, reduce: MuldivTerm
			reduce(191), // ??, reduce: MuldivTerm
			reduce(191), // ???, reduce: MuldivTerm
			reduce(191), // =~, reduce: MuldivTerm
			reduce(191), // !=~, reduce: MuldivTerm
			reduce(191), // ==, reduce: MuldivTerm
			reduce(191), // !=, reduce: MuldivTerm
			reduce(191), // <=>, reduce: MuldivTerm
			reduce(191), // >=, reduce: MuldivTerm
			reduce(191), // <, reduce: MuldivTerm
			reduce(191), // <=, reduce: MuldivTerm
			reduce(191), // ^, reduce: MuldivTerm
			reduce(191), // &, reduce: MuldivTerm
			reduce(191), // <<, reduce: MuldivTerm
			reduce(191), // >>>, reduce: MuldivTerm
			reduce(191), // +, reduce: MuldivTerm
			reduce(191), // -, reduce: MuldivTerm
			reduce(191), // .+, reduce: MuldivTerm
			reduce(191), // .-, reduce: MuldivTerm
			reduce(191), // *, reduce: MuldivTerm
			reduce(191), // /, reduce: MuldivTerm
			reduce(191), // //, reduce: MuldivTerm
			reduce(191), // %, reduce: MuldivTerm
			reduce(191), // .*, reduce: MuldivTerm
			reduce(191), // ./, reduce: MuldivTerm
			reduce(191), // .//, reduce: MuldivTerm
			shift(1647), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2733
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(192), // >, reduce: MuldivTerm
			reduce(192), // >>, reduce: MuldivTerm
			reduce(192), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(192), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(192), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(192), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(192), // ||, reduce: MuldivTerm
			reduce(192), // ^^, reduce: MuldivTerm
			reduce(192), // &&, reduce: MuldivTerm
			reduce(192), // ??, reduce: MuldivTerm
			reduce(192), // ???, reduce: MuldivTerm
			reduce(192), // =~, reduce: MuldivTerm
			reduce(192), // !=~, reduce: MuldivTerm
			reduce(192), // ==, reduce: MuldivTerm
			reduce(192), // !=, reduce: MuldivTerm
			reduce(192), // <=>, reduce: MuldivTerm
			reduce(192), // >=, reduce: MuldivTerm
			reduce(192), // <, reduce: MuldivTerm
			reduce(192), // <=, reduce: MuldivTerm
			reduce(192), // ^, reduce: MuldivTerm
			reduce(192), // &, reduce: MuldivTerm
			reduce(192), // <<, reduce: MuldivTerm
			reduce(192), // >>>, reduce: MuldivTerm
			reduce(192), // +, reduce: MuldivTerm
			reduce(192), // -, reduce: MuldivTerm
			reduce(192), // .+, reduce: MuldivTerm
			reduce(192), // .-, reduce: MuldivTerm
			reduce(192), // *, reduce: MuldivTerm
			reduce(192), // /, reduce: MuldivTerm
			reduce(192), // //, reduce: MuldivTerm
			reduce(192), // %, reduce: MuldivTerm
			reduce(192), // .*, reduce: MuldivTerm
			reduce(192), // ./, reduce: MuldivTerm
			reduce(192), // .//, reduce: MuldivTerm
			shift(1647), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2734
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(193), // >, reduce: MuldivTerm
			reduce(193), // >>, reduce: MuldivTerm
			reduce(193), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(193), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(193), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(193), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(193), // ||, reduce: MuldivTerm
			reduce(193), // ^^, reduce: MuldivTerm
			reduce(193), // &&, reduce: MuldivTerm
			reduce(193), // ??, reduce: MuldivTerm
			reduce(193), // ???, reduce: MuldivTerm
			reduce(193), // =~, reduce: MuldivTerm
			reduce(193), // !=~, reduce: MuldivTerm
			reduce(193), // ==, reduce: MuldivTerm
			reduce(193), // !=, reduce: MuldivTerm
			reduce(193), // <=>, reduce: MuldivTerm
			reduce(193), // >=, reduce: MuldivTerm
			reduce(193), // <, reduce: MuldivTerm
			reduce(193), // <=, reduce: MuldivTerm
			reduce(193), // ^, reduce: MuldivTerm
			reduce(193), // &, reduce: MuldivTerm
			reduce(193), // <<, reduce: MuldivTerm
			reduce(193), // >>>, reduce: MuldivTerm
			reduce(193), // +, reduce: MuldivTerm
			reduce(193), // -, reduce: MuldivTerm
			reduce(193), // .+, reduce: MuldivTerm
			reduce(193), // .-, reduce: MuldivTerm
			reduce(193), // *, reduce: MuldivTerm
			reduce(193), // /, reduce: MuldivTerm
			reduce(193), // //, reduce: MuldivTerm
			reduce(193), // %, reduce: MuldivTerm
			reduce(193), // .*, reduce: MuldivTerm
			reduce(193), // ./, reduce: MuldivTerm
			reduce(193), // .//, reduce: MuldivTerm
			shift(1647), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2735
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(195), // >, reduce: DotTerm
			reduce(195), // >>, reduce: DotTerm
			reduce(195), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(195), // ,, reduce: DotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(195), // ], reduce: DotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(195), // ?, reduce: DotTerm
			nil,         // :
			reduce(195), // ||, reduce: DotTerm
			reduce(195), // ^^, reduce: DotTerm
			reduce(195), // &&, reduce: DotTerm
			reduce(195), // ??, reduce: DotTerm
			reduce(195), // ???, reduce: DotTerm
			reduce(195), // =~, reduce: DotTerm
			reduce(195), // !=~, reduce: DotTerm
			reduce(195), // ==, reduce: DotTerm
			reduce(195), // !=, reduce: DotTerm
			reduce(195), // <=>, reduce: DotTerm
			reduce(195), // >=, reduce: DotTerm
			reduce(195), // <, reduce: DotTerm
			reduce(195), // <=, reduce: DotTerm
			reduce(195), // ^, reduce: DotTerm
			reduce(195), // &, reduce: DotTerm
			reduce(195), // <<, reduce: DotTerm
			reduce(195), // >>>, reduce: DotTerm
			reduce(195), // +, reduce: DotTerm
			reduce(195), // -, reduce: DotTerm
			reduce(195), // .+, reduce: DotTerm
			reduce(195), // .-, reduce: DotTerm
			reduce(195), // *, reduce: DotTerm
			reduce(195), // /, reduce: DotTerm
			reduce(195), // //, reduce: DotTerm
			reduce(195), // %, reduce: DotTerm
			reduce(195), // .*, reduce: DotTerm
			reduce(195), // ./, reduce: DotTerm
			reduce(195), // .//, reduce: DotTerm
			reduce(195), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2736
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2737
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2738
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(204), // >, reduce: PowTerm
			reduce(204), // >>, reduce: PowTerm
			reduce(204), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(204), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(204), // ], reduce: PowTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(204), // ?, reduce: PowTerm
			nil,         // :
			reduce(204), // ||, reduce: PowTerm
			reduce(204), // ^^, reduce: PowTerm
			reduce(204), // &&, reduce: PowTerm
			reduce(204), // ??, reduce: PowTerm
			reduce(204), // ???, reduce: PowTerm
			reduce(204), // =~, reduce: PowTerm
			reduce(204), // !=~, reduce: PowTerm
			reduce(204), // ==, reduce: PowTerm
			reduce(204), // !=, reduce: PowTerm
			reduce(204), // <=>, reduce: PowTerm
			reduce(204), // >=, reduce: PowTerm
			reduce(204), // <, reduce: PowTerm
			reduce(204), // <=, reduce: PowTerm
			reduce(204), // ^, reduce: PowTerm
			reduce(204), // &, reduce: PowTerm
			reduce(204), // <<, reduce: PowTerm
			reduce(204), // >>>, reduce: PowTerm
			reduce(204), // +, reduce: PowTerm
			reduce(204), // -, reduce: PowTerm
			reduce(204), // .+, reduce: PowTerm
			reduce(204), // .-, reduce: PowTerm
			reduce(204), // *, reduce: PowTerm
			reduce(204), // /, reduce: PowTerm
			reduce(204), // //, reduce: PowTerm
			reduce(204), // %, reduce: PowTerm
			reduce(204), // .*, reduce: PowTerm
			reduce(204), // ./, reduce: PowTerm
			reduce(204), // .//, reduce: PowTerm
			reduce(204), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2739
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3381), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3382), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2740
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(3384), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2741
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3385), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2742
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3386), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2743
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(229), // >, reduce: ArrayLiteral
			reduce(229), // >>, reduce: ArrayLiteral
			reduce(229), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(229), // ,, reduce: ArrayLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(229), // ], reduce: ArrayLiteral
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(229), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(229), // ||, reduce: ArrayLiteral
			reduce(229), // ^^, reduce: ArrayLiteral
			reduce(229), // &&, reduce: ArrayLiteral
			reduce(229), // ??, reduce: ArrayLiteral
			reduce(229), // ???, reduce: ArrayLiteral
			reduce(229), // =~, reduce: ArrayLiteral
			reduce(229), // !=~, reduce: ArrayLiteral
			reduce(229), // ==, reduce: ArrayLiteral
			reduce(229), // !=, reduce: ArrayLiteral
			reduce(229), // <=>, reduce: ArrayLiteral
			reduce(229), // >=, reduce: ArrayLiteral
			reduce(229), // <, reduce: ArrayLiteral
			reduce(229), // <=, reduce: ArrayLiteral
			reduce(229), // ^, reduce: ArrayLiteral
			reduce(229), // &, reduce: ArrayLiteral
			reduce(229), // <<, reduce: ArrayLiteral
			reduce(229), // >>>, reduce: ArrayLiteral
			reduce(229), // +, reduce: ArrayLiteral
			reduce(229), // -, reduce: ArrayLiteral
			reduce(229), // .+, reduce: ArrayLiteral
			reduce(229), // .-, reduce: ArrayLiteral
			reduce(229), // *, reduce: ArrayLiteral
			reduce(229), // /, reduce: ArrayLiteral
			reduce(229), // //, reduce: ArrayLiteral
			reduce(229), // %, reduce: ArrayLiteral
			reduce(229), // .*, reduce: ArrayLiteral
			reduce(229), // ./, reduce: ArrayLiteral
			reduce(229), // .//, reduce: ArrayLiteral
			reduce(229), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(229), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(229), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(229), // [[, reduce: ArrayLiteral
			reduce(229), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2744
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(255), // >, reduce: ENV
			reduce(255), // >>, reduce: ENV
			reduce(255), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(255), // ,, reduce: ENV
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(255), // ], reduce: ENV
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(255), // ?, reduce: ENV
			nil,         // :
			reduce(255), // ||, reduce: ENV
			reduce(255), // ^^, reduce: ENV
			reduce(255), // &&, reduce: ENV
			reduce(255), // ??, reduce: ENV
			reduce(255), // ???, reduce: ENV
			reduce(255), // =~, reduce: ENV
			reduce(255), // !=~, reduce: ENV
			reduce(255), // ==, reduce: ENV
			reduce(255), // !=, reduce: ENV
			reduce(255), // <=>, reduce: ENV
			reduce(255), // >=, reduce: ENV
			reduce(255), // <, reduce: ENV
			reduce(255), // <=, reduce: ENV
			reduce(255), // ^, reduce: ENV
			reduce(255), // &, reduce: ENV
			reduce(255), // <<, reduce: ENV
			reduce(255), // >>>, reduce: ENV
			reduce(255), // +, reduce: ENV
			reduce(255), // -, reduce: ENV
			reduce(255), // .+, reduce: ENV
			reduce(255), // .-, reduce: ENV
			reduce(255), // *, reduce: ENV
			reduce(255), // /, reduce: ENV
			reduce(255), // //, reduce: ENV
			reduce(255), // %, reduce: ENV
			reduce(255), // .*, reduce: ENV
			reduce(255), // ./, reduce: ENV
			reduce(255), // .//, reduce: ENV
			reduce(255), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(255), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(255), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(255), // [[, reduce: ENV
			reduce(255), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2745
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3387), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2746
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3388), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2747
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(268), // >, reduce: FunctionCallsite
			reduce(268), // >>, reduce: FunctionCallsite
			reduce(268), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(268), // ,, reduce: FunctionCallsite
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(268), // ], reduce: FunctionCallsite
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(268), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(268), // ||, reduce: FunctionCallsite
			reduce(268), // ^^, reduce: FunctionCallsite
			reduce(268), // &&, reduce: FunctionCallsite
			reduce(268), // ??, reduce: FunctionCallsite
			reduce(268), // ???, reduce: FunctionCallsite
			reduce(268), // =~, reduce: FunctionCallsite
			reduce(268), // !=~, reduce: FunctionCallsite
			reduce(268), // ==, reduce: FunctionCallsite
			reduce(268), // !=, reduce: FunctionCallsite
			reduce(268), // <=>, reduce: FunctionCallsite
			reduce(268), // >=, reduce: FunctionCallsite
			reduce(268), // <, reduce: FunctionCallsite
			reduce(268), // <=, reduce: FunctionCallsite
			reduce(268), // ^, reduce: FunctionCallsite
			reduce(268), // &, reduce: FunctionCallsite
			reduce(268), // <<, reduce: FunctionCallsite
			reduce(268), // >>>, reduce: FunctionCallsite
			reduce(268), // +, reduce: FunctionCallsite
			reduce(268), // -, reduce: FunctionCallsite
			reduce(268), // .+, reduce: FunctionCallsite
			reduce(268), // .-, reduce: FunctionCallsite
			reduce(268), // *, reduce: FunctionCallsite
			reduce(268), // /, reduce: FunctionCallsite
			reduce(268), // //, reduce: FunctionCallsite
			reduce(268), // %, reduce: FunctionCallsite
			reduce(268), // .*, reduce: FunctionCallsite
			reduce(268), // ./, reduce: FunctionCallsite
			reduce(268), // .//, reduce: FunctionCallsite
			reduce(268), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(268), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(268), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(268), // [[, reduce: FunctionCallsite
			reduce(268), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2748
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3389), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2749
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(254), // $, reduce: ENV
			nil,         // empty
			reduce(254), // ;, reduce: ENV
			reduce(254), // {, reduce: ENV
			nil,         // }
			reduce(254), // =, reduce: ENV
			nil,         // unset
			nil,         // filter
			reduce(254), // >, reduce: ENV
			reduce(254), // >>, reduce: ENV
			reduce(254), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(254), // ||=, reduce: ENV
			reduce(254), // ^^=, reduce: ENV
			reduce(254), // &&=, reduce: ENV
			reduce(254), // ??=, reduce: ENV
			reduce(254), // ???=, reduce: ENV
			reduce(254), // |=, reduce: ENV
			reduce(254), // &=, reduce: ENV
			reduce(254), // ^=, reduce: ENV
			reduce(254), // <<=, reduce: ENV
			reduce(254), // >>=, reduce: ENV
			reduce(254), // >>>=, reduce: ENV
			reduce(254), // +=, reduce: ENV
			reduce(254), // .=, reduce: ENV
			reduce(254), // -=, reduce: ENV
			reduce(254), // *=, reduce: ENV
			reduce(254), // /=, reduce: ENV
			reduce(254), // //=, reduce: ENV
			reduce(254), // %=, reduce: ENV
			reduce(254), // **=, reduce: ENV
			reduce(254), // ?, reduce: ENV
			nil,         // :
			reduce(254), // ||, reduce: ENV
			reduce(254), // ^^, reduce: ENV
			reduce(254), // &&, reduce: ENV
			reduce(254), // ??, reduce: ENV
			reduce(254), // ???, reduce: ENV
			reduce(254), // =~, reduce: ENV
			reduce(254), // !=~, reduce: ENV
			reduce(254), // ==, reduce: ENV
			reduce(254), // !=, reduce: ENV
			reduce(254), // <=>, reduce: ENV
			reduce(254), // >=, reduce: ENV
			reduce(254), // <, reduce: ENV
			reduce(254), // <=, reduce: ENV
			reduce(254), // ^, reduce: ENV
			reduce(254), // &, reduce: ENV
			reduce(254), // <<, reduce: ENV
			reduce(254), // >>>, reduce: ENV
			reduce(254), // +, reduce: ENV
			reduce(254), // -, reduce: ENV
			reduce(254), // .+, reduce: ENV
			reduce(254), // .-, reduce: ENV
			reduce(254), // *, reduce: ENV
			reduce(254), // /, reduce: ENV
			reduce(254), // //, reduce: ENV
			reduce(254), // %, reduce: ENV
			reduce(254), // .*, reduce: ENV
			reduce(254), // ./, reduce: ENV
			reduce(254), // .//, reduce: ENV
			reduce(254), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(254), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(254), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(254), // [[, reduce: ENV
			reduce(254), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2750
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(234), // >, reduce: MapLiteral
			reduce(234), // >>, reduce: MapLiteral
			reduce(234), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(234), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(234), // ), reduce: MapLiteral
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(234), // ?, reduce: MapLiteral
			nil,         // :
			reduce(234), // ||, reduce: MapLiteral
			reduce(234), // ^^, reduce: MapLiteral
			reduce(234), // &&, reduce: MapLiteral
			reduce(234), // ??, reduce: MapLiteral
			reduce(234), // ???, reduce: MapLiteral
			reduce(234), // =~, reduce: MapLiteral
			reduce(234), // !=~, reduce: MapLiteral
			reduce(234), // ==, reduce: MapLiteral
			reduce(234), // !=, reduce: MapLiteral
			reduce(234), // <=>, reduce: MapLiteral
			reduce(234), // >=, reduce: MapLiteral
			reduce(234), // <, reduce: MapLiteral
			reduce(234), // <=, reduce: MapLiteral
			reduce(234), // ^, reduce: MapLiteral
			reduce(234), // &, reduce: MapLiteral
			reduce(234), // <<, reduce: MapLiteral
			reduce(234), // >>>, reduce: MapLiteral
			reduce(234), // +, reduce: MapLiteral
			reduce(234), // -, reduce: MapLiteral
			reduce(234), // .+, reduce: MapLiteral
			reduce(234), // .-, reduce: MapLiteral
			reduce(234), // *, reduce: MapLiteral
			reduce(234), // /, reduce: MapLiteral
			reduce(234), // //, reduce: MapLiteral
			reduce(234), // %, reduce: MapLiteral
			reduce(234), // .*, reduce: MapLiteral
			reduce(234), // ./, reduce: MapLiteral
			reduce(234), // .//, reduce: MapLiteral
			reduce(234), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(234), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(234), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(234), // [[, reduce: MapLiteral
			reduce(234), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2751
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(3390), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2752
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			reduce(274), // ), reduce: FcnArgs
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2753
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(269), // $, reduce: FunctionCallsite
			nil,         // empty
			reduce(269), // ;, reduce: FunctionCallsite
			reduce(269), // {, reduce: FunctionCallsite
			nil,         // }
			reduce(269), // =, reduce: FunctionCallsite
			nil,         // unset
			nil,         // filter
			reduce(269), // >, reduce: FunctionCallsite
			reduce(269), // >>, reduce: FunctionCallsite
			reduce(269), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(269), // ||=, reduce: FunctionCallsite
			reduce(269), // ^^=, reduce: FunctionCallsite
			reduce(269), // &&=, reduce: FunctionCallsite
			reduce(269), // ??=, reduce: FunctionCallsite
			reduce(269), // ???=, reduce: FunctionCallsite
			reduce(269), // |=, reduce: FunctionCallsite
			reduce(269), // &=, reduce: FunctionCallsite
			reduce(269), // ^=, reduce: FunctionCallsite
			reduce(269), // <<=, reduce: FunctionCallsite
			reduce(269), // >>=, reduce: FunctionCallsite
			reduce(269), // >>>=, reduce: FunctionCallsite
			reduce(269), // +=, reduce: FunctionCallsite
			reduce(269), // .=, reduce: FunctionCallsite
			reduce(269), // -=, reduce: FunctionCallsite
			reduce(269), // *=, reduce: FunctionCallsite
			reduce(269), // /=, reduce: FunctionCallsite
			reduce(269), // //=, reduce: FunctionCallsite
			reduce(269), // %=, reduce: FunctionCallsite
			reduce(269), // **=, reduce: FunctionCallsite
			reduce(269), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(269), // ||, reduce: FunctionCallsite
			reduce(269), // ^^, reduce: FunctionCallsite
			reduce(269), // &&, reduce: FunctionCallsite
			reduce(269), // ??, reduce: FunctionCallsite
			reduce(269), // ???, reduce: FunctionCallsite
			reduce(269), // =~, reduce: FunctionCallsite
			reduce(269), // !=~, reduce: FunctionCallsite
			reduce(269), // ==, reduce: FunctionCallsite
			reduce(269), // !=, reduce: FunctionCallsite
			reduce(269), // <=>, reduce: FunctionCallsite
			reduce(269), // >=, reduce: FunctionCallsite
			reduce(269), // <, reduce: FunctionCallsite
			reduce(269), // <=, reduce: FunctionCallsite
			reduce(269), // ^, reduce: FunctionCallsite
			reduce(269), // &, reduce: FunctionCallsite
			reduce(269), // <<, reduce: FunctionCallsite
			reduce(269), // >>>, reduce: FunctionCallsite
			reduce(269), // +, reduce: FunctionCallsite
			reduce(269), // -, reduce: FunctionCallsite
			reduce(269), // .+, reduce: FunctionCallsite
			reduce(269), // .-, reduce: FunctionCallsite
			reduce(269), // *, reduce: FunctionCallsite
			reduce(269), // /, reduce: FunctionCallsite
			reduce(269), // //, reduce: FunctionCallsite
			reduce(269), // %, reduce: FunctionCallsite
			reduce(269), // .*, reduce: FunctionCallsite
			reduce(269), // ./, reduce: FunctionCallsite
			reduce(269), // .//, reduce: FunctionCallsite
			reduce(269), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(269), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(269), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(269), // [[, reduce: FunctionCallsite
			reduce(269), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2754
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3392), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2755
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3393), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2756
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3394), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2757
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3395), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2758
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3396), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2759
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2760
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2761
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2762
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2763
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2764
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2765
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2766
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2767
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2768
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2769
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2770
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2771
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2772
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2773
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2774
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2775
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2776
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2777
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2778
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2779
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2780
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2781
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2782
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2783
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2784
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(197), // >, reduce: UnaryOpTerm
			reduce(197), // >>, reduce: UnaryOpTerm
			reduce(197), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(197), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(197), // ), reduce: UnaryOpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(197), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(197), // ||, reduce: UnaryOpTerm
			reduce(197), // ^^, reduce: UnaryOpTerm
			reduce(197), // &&, reduce: UnaryOpTerm
			reduce(197), // ??, reduce: UnaryOpTerm
			reduce(197), // ???, reduce: UnaryOpTerm
			reduce(197), // =~, reduce: UnaryOpTerm
			reduce(197), // !=~, reduce: UnaryOpTerm
			reduce(197), // ==, reduce: UnaryOpTerm
			reduce(197), // !=, reduce: UnaryOpTerm
			reduce(197), // <=>, reduce: UnaryOpTerm
			reduce(197), // >=, reduce: UnaryOpTerm
			reduce(197), // <, reduce: UnaryOpTerm
			reduce(197), // <=, reduce: UnaryOpTerm
			reduce(197), // ^, reduce: UnaryOpTerm
			reduce(197), // &, reduce: UnaryOpTerm
			reduce(197), // <<, reduce: UnaryOpTerm
			reduce(197), // >>>, reduce: UnaryOpTerm
			reduce(197), // +, reduce: UnaryOpTerm
			reduce(197), // -, reduce: UnaryOpTerm
			reduce(197), // .+, reduce: UnaryOpTerm
			reduce(197), // .-, reduce: UnaryOpTerm
			reduce(197), // *, reduce: UnaryOpTerm
			reduce(197), // /, reduce: UnaryOpTerm
			reduce(197), // //, reduce: UnaryOpTerm
			reduce(197), // %, reduce: UnaryOpTerm
			reduce(197), // .*, reduce: UnaryOpTerm
			reduce(197), // ./, reduce: UnaryOpTerm
			reduce(197), // .//, reduce: UnaryOpTerm
			reduce(197), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2785
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2786
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2787
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2788
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2789
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2790
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2791
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2792
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(198), // >, reduce: UnaryOpTerm
			reduce(198), // >>, reduce: UnaryOpTerm
			reduce(198), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(198), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(198), // ), reduce: UnaryOpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(198), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(198), // ||, reduce: UnaryOpTerm
			reduce(198), // ^^, reduce: UnaryOpTerm
			reduce(198), // &&, reduce: UnaryOpTerm
			reduce(198), // ??, reduce: UnaryOpTerm
			reduce(198), // ???, reduce: UnaryOpTerm
			reduce(198), // =~, reduce: UnaryOpTerm
			reduce(198), // !=~, reduce: UnaryOpTerm
			reduce(198), // ==, reduce: UnaryOpTerm
			reduce(198), // !=, reduce: UnaryOpTerm
			reduce(198), // <=>, reduce: UnaryOpTerm
			reduce(198), // >=, reduce: UnaryOpTerm
			reduce(198), // <, reduce: UnaryOpTerm
			reduce(198), // <=, reduce: UnaryOpTerm
			reduce(198), // ^, reduce: UnaryOpTerm
			reduce(198), // &, reduce: UnaryOpTerm
			reduce(198), // <<, reduce: UnaryOpTerm
			reduce(198), // >>>, reduce: UnaryOpTerm
			reduce(198), // +, reduce: UnaryOpTerm
			reduce(198), // -, reduce: UnaryOpTerm
			reduce(198), // .+, reduce: UnaryOpTerm
			reduce(198), // .-, reduce: UnaryOpTerm
			reduce(198), // *, reduce: UnaryOpTerm
			reduce(198), // /, reduce: UnaryOpTerm
			reduce(198), // //, reduce: UnaryOpTerm
			reduce(198), // %, reduce: UnaryOpTerm
			reduce(198), // .*, reduce: UnaryOpTerm
			reduce(198), // ./, reduce: UnaryOpTerm
			reduce(198), // .//, reduce: UnaryOpTerm
			reduce(198), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2793
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(199), // >, reduce: UnaryOpTerm
			reduce(199), // >>, reduce: UnaryOpTerm
			reduce(199), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(199), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(199), // ), reduce: UnaryOpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(199), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(199), // ||, reduce: UnaryOpTerm
			reduce(199), // ^^, reduce: UnaryOpTerm
			reduce(199), // &&, reduce: UnaryOpTerm
			reduce(199), // ??, reduce: UnaryOpTerm
			reduce(199), // ???, reduce: UnaryOpTerm
			reduce(199), // =~, reduce: UnaryOpTerm
			reduce(199), // !=~, reduce: UnaryOpTerm
			reduce(199), // ==, reduce: UnaryOpTerm
			reduce(199), // !=, reduce: UnaryOpTerm
			reduce(199), // <=>, reduce: UnaryOpTerm
			reduce(199), // >=, reduce: UnaryOpTerm
			reduce(199), // <, reduce: UnaryOpTerm
			reduce(199), // <=, reduce: UnaryOpTerm
			reduce(199), // ^, reduce: UnaryOpTerm
			reduce(199), // &, reduce: UnaryOpTerm
			reduce(199), // <<, reduce: UnaryOpTerm
			reduce(199), // >>>, reduce: UnaryOpTerm
			reduce(199), // +, reduce: UnaryOpTerm
			reduce(199), // -, reduce: UnaryOpTerm
			reduce(199), // .+, reduce: UnaryOpTerm
			reduce(199), // .-, reduce: UnaryOpTerm
			reduce(199), // *, reduce: UnaryOpTerm
			reduce(199), // /, reduce: UnaryOpTerm
			reduce(199), // //, reduce: UnaryOpTerm
			reduce(199), // %, reduce: UnaryOpTerm
			reduce(199), // .*, reduce: UnaryOpTerm
			reduce(199), // ./, reduce: UnaryOpTerm
			reduce(199), // .//, reduce: UnaryOpTerm
			reduce(199), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2794
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(200), // >, reduce: UnaryOpTerm
			reduce(200), // >>, reduce: UnaryOpTerm
			reduce(200), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(200), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(200), // ), reduce: UnaryOpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(200), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(200), // ||, reduce: UnaryOpTerm
			reduce(200), // ^^, reduce: UnaryOpTerm
			reduce(200), // &&, reduce: UnaryOpTerm
			reduce(200), // ??, reduce: UnaryOpTerm
			reduce(200), // ???, reduce: UnaryOpTerm
			reduce(200), // =~, reduce: UnaryOpTerm
			reduce(200), // !=~, reduce: UnaryOpTerm
			reduce(200), // ==, reduce: UnaryOpTerm
			reduce(200), // !=, reduce: UnaryOpTerm
			reduce(200), // <=>, reduce: UnaryOpTerm
			reduce(200), // >=, reduce: UnaryOpTerm
			reduce(200), // <, reduce: UnaryOpTerm
			reduce(200), // <=, reduce: UnaryOpTerm
			reduce(200), // ^, reduce: UnaryOpTerm
			reduce(200), // &, reduce: UnaryOpTerm
			reduce(200), // <<, reduce: UnaryOpTerm
			reduce(200), // >>>, reduce: UnaryOpTerm
			reduce(200), // +, reduce: UnaryOpTerm
			reduce(200), // -, reduce: UnaryOpTerm
			reduce(200), // .+, reduce: UnaryOpTerm
			reduce(200), // .-, reduce: UnaryOpTerm
			reduce(200), // *, reduce: UnaryOpTerm
			reduce(200), // /, reduce: UnaryOpTerm
			reduce(200), // //, reduce: UnaryOpTerm
			reduce(200), // %, reduce: UnaryOpTerm
			reduce(200), // .*, reduce: UnaryOpTerm
			reduce(200), // ./, reduce: UnaryOpTerm
			reduce(200), // .//, reduce: UnaryOpTerm
			reduce(200), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2795
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2796
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(201), // >, reduce: UnaryOpTerm
			reduce(201), // >>, reduce: UnaryOpTerm
			reduce(201), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(201), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(201), // ), reduce: UnaryOpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(201), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(201), // ||, reduce: UnaryOpTerm
			reduce(201), // ^^, reduce: UnaryOpTerm
			reduce(201), // &&, reduce: UnaryOpTerm
			reduce(201), // ??, reduce: UnaryOpTerm
			reduce(201), // ???, reduce: UnaryOpTerm
			reduce(201), // =~, reduce: UnaryOpTerm
			reduce(201), // !=~, reduce: UnaryOpTerm
			reduce(201), // ==, reduce: UnaryOpTerm
			reduce(201), // !=, reduce: UnaryOpTerm
			reduce(201), // <=>, reduce: UnaryOpTerm
			reduce(201), // >=, reduce: UnaryOpTerm
			reduce(201), // <, reduce: UnaryOpTerm
			reduce(201), // <=, reduce: UnaryOpTerm
			reduce(201), // ^, reduce: UnaryOpTerm
			reduce(201), // &, reduce: UnaryOpTerm
			reduce(201), // <<, reduce: UnaryOpTerm
			reduce(201), // >>>, reduce: UnaryOpTerm
			reduce(201), // +, reduce: UnaryOpTerm
			reduce(201), // -, reduce: UnaryOpTerm
			reduce(201), // .+, reduce: UnaryOpTerm
			reduce(201), // .-, reduce: UnaryOpTerm
			reduce(201), // *, reduce: UnaryOpTerm
			reduce(201), // /, reduce: UnaryOpTerm
			reduce(201), // //, reduce: UnaryOpTerm
			reduce(201), // %, reduce: UnaryOpTerm
			reduce(201), // .*, reduce: UnaryOpTerm
			reduce(201), // ./, reduce: UnaryOpTerm
			reduce(201), // .//, reduce: UnaryOpTerm
			reduce(201), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2797
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(202), // >, reduce: UnaryOpTerm
			reduce(202), // >>, reduce: UnaryOpTerm
			reduce(202), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(202), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(202), // ), reduce: UnaryOpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(202), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(202), // ||, reduce: UnaryOpTerm
			reduce(202), // ^^, reduce: UnaryOpTerm
			reduce(202), // &&, reduce: UnaryOpTerm
			reduce(202), // ??, reduce: UnaryOpTerm
			reduce(202), // ???, reduce: UnaryOpTerm
			reduce(202), // =~, reduce: UnaryOpTerm
			reduce(202), // !=~, reduce: UnaryOpTerm
			reduce(202), // ==, reduce: UnaryOpTerm
			reduce(202), // !=, reduce: UnaryOpTerm
			reduce(202), // <=>, reduce: UnaryOpTerm
			reduce(202), // >=, reduce: UnaryOpTerm
			reduce(202), // <, reduce: UnaryOpTerm
			reduce(202), // <=, reduce: UnaryOpTerm
			reduce(202), // ^, reduce: UnaryOpTerm
			reduce(202), // &, reduce: UnaryOpTerm
			reduce(202), // <<, reduce: UnaryOpTerm
			reduce(202), // >>>, reduce: UnaryOpTerm
			reduce(202), // +, reduce: UnaryOpTerm
			reduce(202), // -, reduce: UnaryOpTerm
			reduce(202), // .+, reduce: UnaryOpTerm
			reduce(202), // .-, reduce: UnaryOpTerm
			reduce(202), // *, reduce: UnaryOpTerm
			reduce(202), // /, reduce: UnaryOpTerm
			reduce(202), // //, reduce: UnaryOpTerm
			reduce(202), // %, reduce: UnaryOpTerm
			reduce(202), // .*, reduce: UnaryOpTerm
			reduce(202), // ./, reduce: UnaryOpTerm
			reduce(202), // .//, reduce: UnaryOpTerm
			reduce(202), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2798
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3430), // +
			shift(3431), // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2799
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3434), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2800
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2801
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2802
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(228), // >, reduce: ArrayLiteral
			reduce(228), // >>, reduce: ArrayLiteral
			reduce(228), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(228), // ,, reduce: ArrayLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(228), // ), reduce: ArrayLiteral
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(228), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(228), // ||, reduce: ArrayLiteral
			reduce(228), // ^^, reduce: ArrayLiteral
			reduce(228), // &&, reduce: ArrayLiteral
			reduce(228), // ??, reduce: ArrayLiteral
			reduce(228), // ???, reduce: ArrayLiteral
			reduce(228), // =~, reduce: ArrayLiteral
			reduce(228), // !=~, reduce: ArrayLiteral
			reduce(228), // ==, reduce: ArrayLiteral
			reduce(228), // !=, reduce: ArrayLiteral
			reduce(228), // <=>, reduce: ArrayLiteral
			reduce(228), // >=, reduce: ArrayLiteral
			reduce(228), // <, reduce: ArrayLiteral
			reduce(228), // <=, reduce: ArrayLiteral
			reduce(228), // ^, reduce: ArrayLiteral
			reduce(228), // &, reduce: ArrayLiteral
			reduce(228), // <<, reduce: ArrayLiteral
			reduce(228), // >>>, reduce: ArrayLiteral
			reduce(228), // +, reduce: ArrayLiteral
			reduce(228), // -, reduce: ArrayLiteral
			reduce(228), // .+, reduce: ArrayLiteral
			reduce(228), // .-, reduce: ArrayLiteral
			reduce(228), // *, reduce: ArrayLiteral
			reduce(228), // /, reduce: ArrayLiteral
			reduce(228), // //, reduce: ArrayLiteral
			reduce(228), // %, reduce: ArrayLiteral
			reduce(228), // .*, reduce: ArrayLiteral
			reduce(228), // ./, reduce: ArrayLiteral
			reduce(228), // .//, reduce: ArrayLiteral
			reduce(228), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(228), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(228), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(228), // [[, reduce: ArrayLiteral
			reduce(228), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2803
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3437), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2804
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(3438), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2805
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2806
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(3441), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2807
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2808
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3443), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2809
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(276), // $, reduce: SubroutineCallsite
			nil,         // empty
			reduce(276), // ;, reduce: SubroutineCallsite
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2810
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3444), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2811
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3446), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2812
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(289), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2813
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2814
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			reduce(13), // while, reduce: StatementBlockInBraces
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2815
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(234), // =, reduce: MapLiteral
			nil,         // unset
			nil,         // filter
			reduce(234), // >, reduce: MapLiteral
			reduce(234), // >>, reduce: MapLiteral
			reduce(234), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(234), // ||=, reduce: MapLiteral
			reduce(234), // ^^=, reduce: MapLiteral
			reduce(234), // &&=, reduce: MapLiteral
			reduce(234), // ??=, reduce: MapLiteral
			reduce(234), // ???=, reduce: MapLiteral
			reduce(234), // |=, reduce: MapLiteral
			reduce(234), // &=, reduce: MapLiteral
			reduce(234), // ^=, reduce: MapLiteral
			reduce(234), // <<=, reduce: MapLiteral
			reduce(234), // >>=, reduce: MapLiteral
			reduce(234), // >>>=, reduce: MapLiteral
			reduce(234), // +=, reduce: MapLiteral
			reduce(234), // .=, reduce: MapLiteral
			reduce(234), // -=, reduce: MapLiteral
			reduce(234), // *=, reduce: MapLiteral
			reduce(234), // /=, reduce: MapLiteral
			reduce(234), // //=, reduce: MapLiteral
			reduce(234), // %=, reduce: MapLiteral
			reduce(234), // **=, reduce: MapLiteral
			reduce(234), // ?, reduce: MapLiteral
			nil,         // :
			reduce(234), // ||, reduce: MapLiteral
			reduce(234), // ^^, reduce: MapLiteral
			reduce(234), // &&, reduce: MapLiteral
			reduce(234), // ??, reduce: MapLiteral
			reduce(234), // ???, reduce: MapLiteral
			reduce(234), // =~, reduce: MapLiteral
			reduce(234), // !=~, reduce: MapLiteral
			reduce(234), // ==, reduce: MapLiteral
			reduce(234), // !=, reduce: MapLiteral
			reduce(234), // <=>, reduce: MapLiteral
			reduce(234), // >=, reduce: MapLiteral
			reduce(234), // <, reduce: MapLiteral
			reduce(234), // <=, reduce: MapLiteral
			reduce(234), // ^, reduce: MapLiteral
			reduce(234), // &, reduce: MapLiteral
			reduce(234), // <<, reduce: MapLiteral
			reduce(234), // >>>, reduce: MapLiteral
			reduce(234), // +, reduce: MapLiteral
			reduce(234), // -, reduce: MapLiteral
			reduce(234), // .+, reduce: MapLiteral
			reduce(234), // .-, reduce: MapLiteral
			reduce(234), // *, reduce: MapLiteral
			reduce(234), // /, reduce: MapLiteral
			reduce(234), // //, reduce: MapLiteral
			reduce(234), // %, reduce: MapLiteral
			reduce(234), // .*, reduce: MapLiteral
			reduce(234), // ./, reduce: MapLiteral
			reduce(234), // .//, reduce: MapLiteral
			reduce(234), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(234), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(234), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(234), // [[, reduce: MapLiteral
			reduce(234), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2816
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(3449), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2817
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2818
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2819
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2820
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2821
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2822
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2823
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2824
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2825
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2826
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2827
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2828
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2829
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2830
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2831
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2832
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2833
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2834
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2835
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2836
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2837
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(3561), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2838
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2839
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3563), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2840
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3564), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(214), // ), reduce: MlrvalOrFunction
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(214), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2841
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3565), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3566), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2842
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3567), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2843
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3568), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2844
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3569), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2845
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3570), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2846
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2847
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2848
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2849
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2850
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2851
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2852
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2853
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2854
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2855
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2856
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2857
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2858
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2859
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2860
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2861
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2862
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2863
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2864
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2865
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2866
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2867
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2868
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2869
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2870
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2871
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(214), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(214), // ||=, reduce: MlrvalOrFunction
			reduce(214), // ^^=, reduce: MlrvalOrFunction
			reduce(214), // &&=, reduce: MlrvalOrFunction
			reduce(214), // ??=, reduce: MlrvalOrFunction
			reduce(214), // ???=, reduce: MlrvalOrFunction
			reduce(214), // |=, reduce: MlrvalOrFunction
			reduce(214), // &=, reduce: MlrvalOrFunction
			reduce(214), // ^=, reduce: MlrvalOrFunction
			reduce(214), // <<=, reduce: MlrvalOrFunction
			reduce(214), // >>=, reduce: MlrvalOrFunction
			reduce(214), // >>>=, reduce: MlrvalOrFunction
			reduce(214), // +=, reduce: MlrvalOrFunction
			reduce(214), // .=, reduce: MlrvalOrFunction
			reduce(214), // -=, reduce: MlrvalOrFunction
			reduce(214), // *=, reduce: MlrvalOrFunction
			reduce(214), // /=, reduce: MlrvalOrFunction
			reduce(214), // //=, reduce: MlrvalOrFunction
			reduce(214), // %=, reduce: MlrvalOrFunction
			reduce(214), // **=, reduce: MlrvalOrFunction
			reduce(214), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2872
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2873
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(118), // =, reduce: LocalVariable
			nil,         // unset
			nil,         // filter
			reduce(118), // >, reduce: LocalVariable
			reduce(118), // >>, reduce: LocalVariable
			reduce(118), // |, reduce: LocalVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(270), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(118), // ||=, reduce: LocalVariable
			reduce(118), // ^^=, reduce: LocalVariable
			reduce(118), // &&=, reduce: LocalVariable
			reduce(118), // ??=, reduce: LocalVariable
			reduce(118), // ???=, reduce: LocalVariable
			reduce(118), // |=, reduce: LocalVariable
			reduce(118), // &=, reduce: LocalVariable
			reduce(118), // ^=, reduce: LocalVariable
			reduce(118), // <<=, reduce: LocalVariable
			reduce(118), // >>=, reduce: LocalVariable
			reduce(118), // >>>=, reduce: LocalVariable
			reduce(118), // +=, reduce: LocalVariable
			reduce(118), // .=, reduce: LocalVariable
			reduce(118), // -=, reduce: LocalVariable
			reduce(118), // *=, reduce: LocalVariable
			reduce(118), // /=, reduce: LocalVariable
			reduce(118), // //=, reduce: LocalVariable
			reduce(118), // %=, reduce: LocalVariable
			reduce(118), // **=, reduce: LocalVariable
			reduce(118), // ?, reduce: LocalVariable
			nil,         // :
			reduce(118), // ||, reduce: LocalVariable
			reduce(118), // ^^, reduce: LocalVariable
			reduce(118), // &&, reduce: LocalVariable
			reduce(118), // ??, reduce: LocalVariable
			reduce(118), // ???, reduce: LocalVariable
			reduce(118), // =~, reduce: LocalVariable
			reduce(118), // !=~, reduce: LocalVariable
			reduce(118), // ==, reduce: LocalVariable
			reduce(118), // !=, reduce: LocalVariable
			reduce(118), // <=>, reduce: LocalVariable
			reduce(118), // >=, reduce: LocalVariable
			reduce(118), // <, reduce: LocalVariable
			reduce(118), // <=, reduce: LocalVariable
			reduce(118), // ^, reduce: LocalVariable
			reduce(118), // &, reduce: LocalVariable
			reduce(118), // <<, reduce: LocalVariable
			reduce(118), // >>>, reduce: LocalVariable
			reduce(118), // +, reduce: LocalVariable
			reduce(118), // -, reduce: LocalVariable
			reduce(118), // .+, reduce: LocalVariable
			reduce(118), // .-, reduce: LocalVariable
			reduce(118), // *, reduce: LocalVariable
			reduce(118), // /, reduce: LocalVariable
			reduce(118), // //, reduce: LocalVariable
			reduce(118), // %, reduce: LocalVariable
			reduce(118), // .*, reduce: LocalVariable
			reduce(118), // ./, reduce: LocalVariable
			reduce(118), // .//, reduce: LocalVariable
			reduce(118), // ., reduce: LocalVariable
			nil,         // !
			nil,         // ~
			reduce(118), // **, reduce: LocalVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(118), // [, reduce: LocalVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(118), // [[, reduce: LocalVariable
			reduce(118), // [[[, reduce: LocalVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2874
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(197), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(197), // >, reduce: UnaryOpTerm
			reduce(197), // >>, reduce: UnaryOpTerm
			reduce(197), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(197), // ||=, reduce: UnaryOpTerm
			reduce(197), // ^^=, reduce: UnaryOpTerm
			reduce(197), // &&=, reduce: UnaryOpTerm
			reduce(197), // ??=, reduce: UnaryOpTerm
			reduce(197), // ???=, reduce: UnaryOpTerm
			reduce(197), // |=, reduce: UnaryOpTerm
			reduce(197), // &=, reduce: UnaryOpTerm
			reduce(197), // ^=, reduce: UnaryOpTerm
			reduce(197), // <<=, reduce: UnaryOpTerm
			reduce(197), // >>=, reduce: UnaryOpTerm
			reduce(197), // >>>=, reduce: UnaryOpTerm
			reduce(197), // +=, reduce: UnaryOpTerm
			reduce(197), // .=, reduce: UnaryOpTerm
			reduce(197), // -=, reduce: UnaryOpTerm
			reduce(197), // *=, reduce: UnaryOpTerm
			reduce(197), // /=, reduce: UnaryOpTerm
			reduce(197), // //=, reduce: UnaryOpTerm
			reduce(197), // %=, reduce: UnaryOpTerm
			reduce(197), // **=, reduce: UnaryOpTerm
			reduce(197), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(197), // ||, reduce: UnaryOpTerm
			reduce(197), // ^^, reduce: UnaryOpTerm
			reduce(197), // &&, reduce: UnaryOpTerm
			reduce(197), // ??, reduce: UnaryOpTerm
			reduce(197), // ???, reduce: UnaryOpTerm
			reduce(197), // =~, reduce: UnaryOpTerm
			reduce(197), // !=~, reduce: UnaryOpTerm
			reduce(197), // ==, reduce: UnaryOpTerm
			reduce(197), // !=, reduce: UnaryOpTerm
			reduce(197), // <=>, reduce: UnaryOpTerm
			reduce(197), // >=, reduce: UnaryOpTerm
			reduce(197), // <, reduce: UnaryOpTerm
			reduce(197), // <=, reduce: UnaryOpTerm
			reduce(197), // ^, reduce: UnaryOpTerm
			reduce(197), // &, reduce: UnaryOpTerm
			reduce(197), // <<, reduce: UnaryOpTerm
			reduce(197), // >>>, reduce: UnaryOpTerm
			reduce(197), // +, reduce: UnaryOpTerm
			reduce(197), // -, reduce: UnaryOpTerm
			reduce(197), // .+, reduce: UnaryOpTerm
			reduce(197), // .-, reduce: UnaryOpTerm
			reduce(197), // *, reduce: UnaryOpTerm
			reduce(197), // /, reduce: UnaryOpTerm
			reduce(197), // //, reduce: UnaryOpTerm
			reduce(197), // %, reduce: UnaryOpTerm
			reduce(197), // .*, reduce: UnaryOpTerm
			reduce(197), // ./, reduce: UnaryOpTerm
			reduce(197), // .//, reduce: UnaryOpTerm
			reduce(197), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2875
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2876
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2877
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2878
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2879
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2880
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2881
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2882
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(198), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(198), // >, reduce: UnaryOpTerm
			reduce(198), // >>, reduce: UnaryOpTerm
			reduce(198), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(198), // ||=, reduce: UnaryOpTerm
			reduce(198), // ^^=, reduce: UnaryOpTerm
			reduce(198), // &&=, reduce: UnaryOpTerm
			reduce(198), // ??=, reduce: UnaryOpTerm
			reduce(198), // ???=, reduce: UnaryOpTerm
			reduce(198), // |=, reduce: UnaryOpTerm
			reduce(198), // &=, reduce: UnaryOpTerm
			reduce(198), // ^=, reduce: UnaryOpTerm
			reduce(198), // <<=, reduce: UnaryOpTerm
			reduce(198), // >>=, reduce: UnaryOpTerm
			reduce(198), // >>>=, reduce: UnaryOpTerm
			reduce(198), // +=, reduce: UnaryOpTerm
			reduce(198), // .=, reduce: UnaryOpTerm
			reduce(198), // -=, reduce: UnaryOpTerm
			reduce(198), // *=, reduce: UnaryOpTerm
			reduce(198), // /=, reduce: UnaryOpTerm
			reduce(198), // //=, reduce: UnaryOpTerm
			reduce(198), // %=, reduce: UnaryOpTerm
			reduce(198), // **=, reduce: UnaryOpTerm
			reduce(198), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(198), // ||, reduce: UnaryOpTerm
			reduce(198), // ^^, reduce: UnaryOpTerm
			reduce(198), // &&, reduce: UnaryOpTerm
			reduce(198), // ??, reduce: UnaryOpTerm
			reduce(198), // ???, reduce: UnaryOpTerm
			reduce(198), // =~, reduce: UnaryOpTerm
			reduce(198), // !=~, reduce: UnaryOpTerm
			reduce(198), // ==, reduce: UnaryOpTerm
			reduce(198), // !=, reduce: UnaryOpTerm
			reduce(198), // <=>, reduce: UnaryOpTerm
			reduce(198), // >=, reduce: UnaryOpTerm
			reduce(198), // <, reduce: UnaryOpTerm
			reduce(198), // <=, reduce: UnaryOpTerm
			reduce(198), // ^, reduce: UnaryOpTerm
			reduce(198), // &, reduce: UnaryOpTerm
			reduce(198), // <<, reduce: UnaryOpTerm
			reduce(198), // >>>, reduce: UnaryOpTerm
			reduce(198), // +, reduce: UnaryOpTerm
			reduce(198), // -, reduce: UnaryOpTerm
			reduce(198), // .+, reduce: UnaryOpTerm
			reduce(198), // .-, reduce: UnaryOpTerm
			reduce(198), // *, reduce: UnaryOpTerm
			reduce(198), // /, reduce: UnaryOpTerm
			reduce(198), // //, reduce: UnaryOpTerm
			reduce(198), // %, reduce: UnaryOpTerm
			reduce(198), // .*, reduce: UnaryOpTerm
			reduce(198), // ./, reduce: UnaryOpTerm
			reduce(198), // .//, reduce: UnaryOpTerm
			reduce(198), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2883
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(199), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(199), // >, reduce: UnaryOpTerm
			reduce(199), // >>, reduce: UnaryOpTerm
			reduce(199), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(199), // ||=, reduce: UnaryOpTerm
			reduce(199), // ^^=, reduce: UnaryOpTerm
			reduce(199), // &&=, reduce: UnaryOpTerm
			reduce(199), // ??=, reduce: UnaryOpTerm
			reduce(199), // ???=, reduce: UnaryOpTerm
			reduce(199), // |=, reduce: UnaryOpTerm
			reduce(199), // &=, reduce: UnaryOpTerm
			reduce(199), // ^=, reduce: UnaryOpTerm
			reduce(199), // <<=, reduce: UnaryOpTerm
			reduce(199), // >>=, reduce: UnaryOpTerm
			reduce(199), // >>>=, reduce: UnaryOpTerm
			reduce(199), // +=, reduce: UnaryOpTerm
			reduce(199), // .=, reduce: UnaryOpTerm
			reduce(199), // -=, reduce: UnaryOpTerm
			reduce(199), // *=, reduce: UnaryOpTerm
			reduce(199), // /=, reduce: UnaryOpTerm
			reduce(199), // //=, reduce: UnaryOpTerm
			reduce(199), // %=, reduce: UnaryOpTerm
			reduce(199), // **=, reduce: UnaryOpTerm
			reduce(199), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(199), // ||, reduce: UnaryOpTerm
			reduce(199), // ^^, reduce: UnaryOpTerm
			reduce(199), // &&, reduce: UnaryOpTerm
			reduce(199), // ??, reduce: UnaryOpTerm
			reduce(199), // ???, reduce: UnaryOpTerm
			reduce(199), // =~, reduce: UnaryOpTerm
			reduce(199), // !=~, reduce: UnaryOpTerm
			reduce(199), // ==, reduce: UnaryOpTerm
			reduce(199), // !=, reduce: UnaryOpTerm
			reduce(199), // <=>, reduce: UnaryOpTerm
			reduce(199), // >=, reduce: UnaryOpTerm
			reduce(199), // <, reduce: UnaryOpTerm
			reduce(199), // <=, reduce: UnaryOpTerm
			reduce(199), // ^, reduce: UnaryOpTerm
			reduce(199), // &, reduce: UnaryOpTerm
			reduce(199), // <<, reduce: UnaryOpTerm
			reduce(199), // >>>, reduce: UnaryOpTerm
			reduce(199), // +, reduce: UnaryOpTerm
			reduce(199), // -, reduce: UnaryOpTerm
			reduce(199), // .+, reduce: UnaryOpTerm
			reduce(199), // .-, reduce: UnaryOpTerm
			reduce(199), // *, reduce: UnaryOpTerm
			reduce(199), // /, reduce: UnaryOpTerm
			reduce(199), // //, reduce: UnaryOpTerm
			reduce(199), // %, reduce: UnaryOpTerm
			reduce(199), // .*, reduce: UnaryOpTerm
			reduce(199), // ./, reduce: UnaryOpTerm
			reduce(199), // .//, reduce: UnaryOpTerm
			reduce(199), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2884
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(200), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(200), // >, reduce: UnaryOpTerm
			reduce(200), // >>, reduce: UnaryOpTerm
			reduce(200), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(200), // ||=, reduce: UnaryOpTerm
			reduce(200), // ^^=, reduce: UnaryOpTerm
			reduce(200), // &&=, reduce: UnaryOpTerm
			reduce(200), // ??=, reduce: UnaryOpTerm
			reduce(200), // ???=, reduce: UnaryOpTerm
			reduce(200), // |=, reduce: UnaryOpTerm
			reduce(200), // &=, reduce: UnaryOpTerm
			reduce(200), // ^=, reduce: UnaryOpTerm
			reduce(200), // <<=, reduce: UnaryOpTerm
			reduce(200), // >>=, reduce: UnaryOpTerm
			reduce(200), // >>>=, reduce: UnaryOpTerm
			reduce(200), // +=, reduce: UnaryOpTerm
			reduce(200), // .=, reduce: UnaryOpTerm
			reduce(200), // -=, reduce: UnaryOpTerm
			reduce(200), // *=, reduce: UnaryOpTerm
			reduce(200), // /=, reduce: UnaryOpTerm
			reduce(200), // //=, reduce: UnaryOpTerm
			reduce(200), // %=, reduce: UnaryOpTerm
			reduce(200), // **=, reduce: UnaryOpTerm
			reduce(200), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(200), // ||, reduce: UnaryOpTerm
			reduce(200), // ^^, reduce: UnaryOpTerm
			reduce(200), // &&, reduce: UnaryOpTerm
			reduce(200), // ??, reduce: UnaryOpTerm
			reduce(200), // ???, reduce: UnaryOpTerm
			reduce(200), // =~, reduce: UnaryOpTerm
			reduce(200), // !=~, reduce: UnaryOpTerm
			reduce(200), // ==, reduce: UnaryOpTerm
			reduce(200), // !=, reduce: UnaryOpTerm
			reduce(200), // <=>, reduce: UnaryOpTerm
			reduce(200), // >=, reduce: UnaryOpTerm
			reduce(200), // <, reduce: UnaryOpTerm
			reduce(200), // <=, reduce: UnaryOpTerm
			reduce(200), // ^, reduce: UnaryOpTerm
			reduce(200), // &, reduce: UnaryOpTerm
			reduce(200), // <<, reduce: UnaryOpTerm
			reduce(200), // >>>, reduce: UnaryOpTerm
			reduce(200), // +, reduce: UnaryOpTerm
			reduce(200), // -, reduce: UnaryOpTerm
			reduce(200), // .+, reduce: UnaryOpTerm
			reduce(200), // .-, reduce: UnaryOpTerm
			reduce(200), // *, reduce: UnaryOpTerm
			reduce(200), // /, reduce: UnaryOpTerm
			reduce(200), // //, reduce: UnaryOpTerm
			reduce(200), // %, reduce: UnaryOpTerm
			reduce(200), // .*, reduce: UnaryOpTerm
			reduce(200), // ./, reduce: UnaryOpTerm
			reduce(200), // .//, reduce: UnaryOpTerm
			reduce(200), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2885
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2886
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(201), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(201), // >, reduce: UnaryOpTerm
			reduce(201), // >>, reduce: UnaryOpTerm
			reduce(201), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(201), // ||=, reduce: UnaryOpTerm
			reduce(201), // ^^=, reduce: UnaryOpTerm
			reduce(201), // &&=, reduce: UnaryOpTerm
			reduce(201), // ??=, reduce: UnaryOpTerm
			reduce(201), // ???=, reduce: UnaryOpTerm
			reduce(201), // |=, reduce: UnaryOpTerm
			reduce(201), // &=, reduce: UnaryOpTerm
			reduce(201), // ^=, reduce: UnaryOpTerm
			reduce(201), // <<=, reduce: UnaryOpTerm
			reduce(201), // >>=, reduce: UnaryOpTerm
			reduce(201), // >>>=, reduce: UnaryOpTerm
			reduce(201), // +=, reduce: UnaryOpTerm
			reduce(201), // .=, reduce: UnaryOpTerm
			reduce(201), // -=, reduce: UnaryOpTerm
			reduce(201), // *=, reduce: UnaryOpTerm
			reduce(201), // /=, reduce: UnaryOpTerm
			reduce(201), // //=, reduce: UnaryOpTerm
			reduce(201), // %=, reduce: UnaryOpTerm
			reduce(201), // **=, reduce: UnaryOpTerm
			reduce(201), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(201), // ||, reduce: UnaryOpTerm
			reduce(201), // ^^, reduce: UnaryOpTerm
			reduce(201), // &&, reduce: UnaryOpTerm
			reduce(201), // ??, reduce: UnaryOpTerm
			reduce(201), // ???, reduce: UnaryOpTerm
			reduce(201), // =~, reduce: UnaryOpTerm
			reduce(201), // !=~, reduce: UnaryOpTerm
			reduce(201), // ==, reduce: UnaryOpTerm
			reduce(201), // !=, reduce: UnaryOpTerm
			reduce(201), // <=>, reduce: UnaryOpTerm
			reduce(201), // >=, reduce: UnaryOpTerm
			reduce(201), // <, reduce: UnaryOpTerm
			reduce(201), // <=, reduce: UnaryOpTerm
			reduce(201), // ^, reduce: UnaryOpTerm
			reduce(201), // &, reduce: UnaryOpTerm
			reduce(201), // <<, reduce: UnaryOpTerm
			reduce(201), // >>>, reduce: UnaryOpTerm
			reduce(201), // +, reduce: UnaryOpTerm
			reduce(201), // -, reduce: UnaryOpTerm
			reduce(201), // .+, reduce: UnaryOpTerm
			reduce(201), // .-, reduce: UnaryOpTerm
			reduce(201), // *, reduce: UnaryOpTerm
			reduce(201), // /, reduce: UnaryOpTerm
			reduce(201), // //, reduce: UnaryOpTerm
			reduce(201), // %, reduce: UnaryOpTerm
			reduce(201), // .*, reduce: UnaryOpTerm
			reduce(201), // ./, reduce: UnaryOpTerm
			reduce(201), // .//, reduce: UnaryOpTerm
			reduce(201), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2887
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(202), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(202), // >, reduce: UnaryOpTerm
			reduce(202), // >>, reduce: UnaryOpTerm
			reduce(202), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(202), // ||=, reduce: UnaryOpTerm
			reduce(202), // ^^=, reduce: UnaryOpTerm
			reduce(202), // &&=, reduce: UnaryOpTerm
			reduce(202), // ??=, reduce: UnaryOpTerm
			reduce(202), // ???=, reduce: UnaryOpTerm
			reduce(202), // |=, reduce: UnaryOpTerm
			reduce(202), // &=, reduce: UnaryOpTerm
			reduce(202), // ^=, reduce: UnaryOpTerm
			reduce(202), // <<=, reduce: UnaryOpTerm
			reduce(202), // >>=, reduce: UnaryOpTerm
			reduce(202), // >>>=, reduce: UnaryOpTerm
			reduce(202), // +=, reduce: UnaryOpTerm
			reduce(202), // .=, reduce: UnaryOpTerm
			reduce(202), // -=, reduce: UnaryOpTerm
			reduce(202), // *=, reduce: UnaryOpTerm
			reduce(202), // /=, reduce: UnaryOpTerm
			reduce(202), // //=, reduce: UnaryOpTerm
			reduce(202), // %=, reduce: UnaryOpTerm
			reduce(202), // **=, reduce: UnaryOpTerm
			reduce(202), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(202), // ||, reduce: UnaryOpTerm
			reduce(202), // ^^, reduce: UnaryOpTerm
			reduce(202), // &&, reduce: UnaryOpTerm
			reduce(202), // ??, reduce: UnaryOpTerm
			reduce(202), // ???, reduce: UnaryOpTerm
			reduce(202), // =~, reduce: UnaryOpTerm
			reduce(202), // !=~, reduce: UnaryOpTerm
			reduce(202), // ==, reduce: UnaryOpTerm
			reduce(202), // !=, reduce: UnaryOpTerm
			reduce(202), // <=>, reduce: UnaryOpTerm
			reduce(202), // >=, reduce: UnaryOpTerm
			reduce(202), // <, reduce: UnaryOpTerm
			reduce(202), // <=, reduce: UnaryOpTerm
			reduce(202), // ^, reduce: UnaryOpTerm
			reduce(202), // &, reduce: UnaryOpTerm
			reduce(202), // <<, reduce: UnaryOpTerm
			reduce(202), // >>>, reduce: UnaryOpTerm
			reduce(202), // +, reduce: UnaryOpTerm
			reduce(202), // -, reduce: UnaryOpTerm
			reduce(202), // .+, reduce: UnaryOpTerm
			reduce(202), // .-, reduce: UnaryOpTerm
			reduce(202), // *, reduce: UnaryOpTerm
			reduce(202), // /, reduce: UnaryOpTerm
			reduce(202), // //, reduce: UnaryOpTerm
			reduce(202), // %, reduce: UnaryOpTerm
			reduce(202), // .*, reduce: UnaryOpTerm
			reduce(202), // ./, reduce: UnaryOpTerm
			reduce(202), // .//, reduce: UnaryOpTerm
			reduce(202), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2888
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3604), // +
			shift(3605), // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2889
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3608), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2890
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2891
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2892
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(228), // =, reduce: ArrayLiteral
			nil,         // unset
			nil,         // filter
			reduce(228), // >, reduce: ArrayLiteral
			reduce(228), // >>, reduce: ArrayLiteral
			reduce(228), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(228), // ||=, reduce: ArrayLiteral
			reduce(228), // ^^=, reduce: ArrayLiteral
			reduce(228), // &&=, reduce: ArrayLiteral
			reduce(228), // ??=, reduce: ArrayLiteral
			reduce(228), // ???=, reduce: ArrayLiteral
			reduce(228), // |=, reduce: ArrayLiteral
			reduce(228), // &=, reduce: ArrayLiteral
			reduce(228), // ^=, reduce: ArrayLiteral
			reduce(228), // <<=, reduce: ArrayLiteral
			reduce(228), // >>=, reduce: ArrayLiteral
			reduce(228), // >>>=, reduce: ArrayLiteral
			reduce(228), // +=, reduce: ArrayLiteral
			reduce(228), // .=, reduce: ArrayLiteral
			reduce(228), // -=, reduce: ArrayLiteral
			reduce(228), // *=, reduce: ArrayLiteral
			reduce(228), // /=, reduce: ArrayLiteral
			reduce(228), // //=, reduce: ArrayLiteral
			reduce(228), // %=, reduce: ArrayLiteral
			reduce(228), // **=, reduce: ArrayLiteral
			reduce(228), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(228), // ||, reduce: ArrayLiteral
			reduce(228), // ^^, reduce: ArrayLiteral
			reduce(228), // &&, reduce: ArrayLiteral
			reduce(228), // ??, reduce: ArrayLiteral
			reduce(228), // ???, reduce: ArrayLiteral
			reduce(228), // =~, reduce: ArrayLiteral
			reduce(228), // !=~, reduce: ArrayLiteral
			reduce(228), // ==, reduce: ArrayLiteral
			reduce(228), // !=, reduce: ArrayLiteral
			reduce(228), // <=>, reduce: ArrayLiteral
			reduce(228), // >=, reduce: ArrayLiteral
			reduce(228), // <, reduce: ArrayLiteral
			reduce(228), // <=, reduce: ArrayLiteral
			reduce(228), // ^, reduce: ArrayLiteral
			reduce(228), // &, reduce: ArrayLiteral
			reduce(228), // <<, reduce: ArrayLiteral
			reduce(228), // >>>, reduce: ArrayLiteral
			reduce(228), // +, reduce: ArrayLiteral
			reduce(228), // -, reduce: ArrayLiteral
			reduce(228), // .+, reduce: ArrayLiteral
			reduce(228), // .-, reduce: ArrayLiteral
			reduce(228), // *, reduce: ArrayLiteral
			reduce(228), // /, reduce: ArrayLiteral
			reduce(228), // //, reduce: ArrayLiteral
			reduce(228), // %, reduce: ArrayLiteral
			reduce(228), // .*, reduce: ArrayLiteral
			reduce(228), // ./, reduce: ArrayLiteral
			reduce(228), // .//, reduce: ArrayLiteral
			reduce(228), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(228), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(228), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(228), // [[, reduce: ArrayLiteral
			reduce(228), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2893
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3611), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2894
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(3612), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2895
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2896
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(3615), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2897
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(312), // ;, reduce: TripleForContinuation
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(312), // ,, reduce: TripleForContinuation
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2898
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2899
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2900
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(335), // ,, reduce: TypedFuncOrSubrParameterName
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(335), // ), reduce: TypedFuncOrSubrParameterName
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2901
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3714), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3715), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2902
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(330), // ), reduce: FuncOrSubrNonEmptyParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2903
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3717), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2904
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3718), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2905
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(234), // }, reduce: MapLiteral
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(234), // >, reduce: MapLiteral
			reduce(234), // >>, reduce: MapLiteral
			reduce(234), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(234), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(234), // ?, reduce: MapLiteral
			nil,         // :
			reduce(234), // ||, reduce: MapLiteral
			reduce(234), // ^^, reduce: MapLiteral
			reduce(234), // &&, reduce: MapLiteral
			reduce(234), // ??, reduce: MapLiteral
			reduce(234), // ???, reduce: MapLiteral
			reduce(234), // =~, reduce: MapLiteral
			reduce(234), // !=~, reduce: MapLiteral
			reduce(234), // ==, reduce: MapLiteral
			reduce(234), // !=, reduce: MapLiteral
			reduce(234), // <=>, reduce: MapLiteral
			reduce(234), // >=, reduce: MapLiteral
			reduce(234), // <, reduce: MapLiteral
			reduce(234), // <=, reduce: MapLiteral
			reduce(234), // ^, reduce: MapLiteral
			reduce(234), // &, reduce: MapLiteral
			reduce(234), // <<, reduce: MapLiteral
			reduce(234), // >>>, reduce: MapLiteral
			reduce(234), // +, reduce: MapLiteral
			reduce(234), // -, reduce: MapLiteral
			reduce(234), // .+, reduce: MapLiteral
			reduce(234), // .-, reduce: MapLiteral
			reduce(234), // *, reduce: MapLiteral
			reduce(234), // /, reduce: MapLiteral
			reduce(234), // //, reduce: MapLiteral
			reduce(234), // %, reduce: MapLiteral
			reduce(234), // .*, reduce: MapLiteral
			reduce(234), // ./, reduce: MapLiteral
			reduce(234), // .//, reduce: MapLiteral
			reduce(234), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(234), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(234), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(234), // [[, reduce: MapLiteral
			reduce(234), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2906
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(3719), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2907
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3720), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2908
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3721), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2909
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3722), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2910
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3723), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2911
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3724), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2912
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2913
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2914
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2915
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2916
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2917
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2918
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2919
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2920
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2921
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2922
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2923
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2924
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2925
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2926
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2927
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2928
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2929
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2930
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2931
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2932
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2933
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2934
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2935
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2936
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2937
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(197), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(197), // >, reduce: UnaryOpTerm
			reduce(197), // >>, reduce: UnaryOpTerm
			reduce(197), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(197), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(197), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(197), // ||, reduce: UnaryOpTerm
			reduce(197), // ^^, reduce: UnaryOpTerm
			reduce(197), // &&, reduce: UnaryOpTerm
			reduce(197), // ??, reduce: UnaryOpTerm
			reduce(197), // ???, reduce: UnaryOpTerm
			reduce(197), // =~, reduce: UnaryOpTerm
			reduce(197), // !=~, reduce: UnaryOpTerm
			reduce(197), // ==, reduce: UnaryOpTerm
			reduce(197), // !=, reduce: UnaryOpTerm
			reduce(197), // <=>, reduce: UnaryOpTerm
			reduce(197), // >=, reduce: UnaryOpTerm
			reduce(197), // <, reduce: UnaryOpTerm
			reduce(197), // <=, reduce: UnaryOpTerm
			reduce(197), // ^, reduce: UnaryOpTerm
			reduce(197), // &, reduce: UnaryOpTerm
			reduce(197), // <<, reduce: UnaryOpTerm
			reduce(197), // >>>, reduce: UnaryOpTerm
			reduce(197), // +, reduce: UnaryOpTerm
			reduce(197), // -, reduce: UnaryOpTerm
			reduce(197), // .+, reduce: UnaryOpTerm
			reduce(197), // .-, reduce: UnaryOpTerm
			reduce(197), // *, reduce: UnaryOpTerm
			reduce(197), // /, reduce: UnaryOpTerm
			reduce(197), // //, reduce: UnaryOpTerm
			reduce(197), // %, reduce: UnaryOpTerm
			reduce(197), // .*, reduce: UnaryOpTerm
			reduce(197), // ./, reduce: UnaryOpTerm
			reduce(197), // .//, reduce: UnaryOpTerm
			reduce(197), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2938
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2939
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2940
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2941
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2942
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2943
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2944
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2945
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(198), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(198), // >, reduce: UnaryOpTerm
			reduce(198), // >>, reduce: UnaryOpTerm
			reduce(198), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(198), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(198), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(198), // ||, reduce: UnaryOpTerm
			reduce(198), // ^^, reduce: UnaryOpTerm
			reduce(198), // &&, reduce: UnaryOpTerm
			reduce(198), // ??, reduce: UnaryOpTerm
			reduce(198), // ???, reduce: UnaryOpTerm
			reduce(198), // =~, reduce: UnaryOpTerm
			reduce(198), // !=~, reduce: UnaryOpTerm
			reduce(198), // ==, reduce: UnaryOpTerm
			reduce(198), // !=, reduce: UnaryOpTerm
			reduce(198), // <=>, reduce: UnaryOpTerm
			reduce(198), // >=, reduce: UnaryOpTerm
			reduce(198), // <, reduce: UnaryOpTerm
			reduce(198), // <=, reduce: UnaryOpTerm
			reduce(198), // ^, reduce: UnaryOpTerm
			reduce(198), // &, reduce: UnaryOpTerm
			reduce(198), // <<, reduce: UnaryOpTerm
			reduce(198), // >>>, reduce: UnaryOpTerm
			reduce(198), // +, reduce: UnaryOpTerm
			reduce(198), // -, reduce: UnaryOpTerm
			reduce(198), // .+, reduce: UnaryOpTerm
			reduce(198), // .-, reduce: UnaryOpTerm
			reduce(198), // *, reduce: UnaryOpTerm
			reduce(198), // /, reduce: UnaryOpTerm
			reduce(198), // //, reduce: UnaryOpTerm
			reduce(198), // %, reduce: UnaryOpTerm
			reduce(198), // .*, reduce: UnaryOpTerm
			reduce(198), // ./, reduce: UnaryOpTerm
			reduce(198), // .//, reduce: UnaryOpTerm
			reduce(198), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2946
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(199), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(199), // >, reduce: UnaryOpTerm
			reduce(199), // >>, reduce: UnaryOpTerm
			reduce(199), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(199), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(199), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(199), // ||, reduce: UnaryOpTerm
			reduce(199), // ^^, reduce: UnaryOpTerm
			reduce(199), // &&, reduce: UnaryOpTerm
			reduce(199), // ??, reduce: UnaryOpTerm
			reduce(199), // ???, reduce: UnaryOpTerm
			reduce(199), // =~, reduce: UnaryOpTerm
			reduce(199), // !=~, reduce: UnaryOpTerm
			reduce(199), // ==, reduce: UnaryOpTerm
			reduce(199), // !=, reduce: UnaryOpTerm
			reduce(199), // <=>, reduce: UnaryOpTerm
			reduce(199), // >=, reduce: UnaryOpTerm
			reduce(199), // <, reduce: UnaryOpTerm
			reduce(199), // <=, reduce: UnaryOpTerm
			reduce(199), // ^, reduce: UnaryOpTerm
			reduce(199), // &, reduce: UnaryOpTerm
			reduce(199), // <<, reduce: UnaryOpTerm
			reduce(199), // >>>, reduce: UnaryOpTerm
			reduce(199), // +, reduce: UnaryOpTerm
			reduce(199), // -, reduce: UnaryOpTerm
			reduce(199), // .+, reduce: UnaryOpTerm
			reduce(199), // .-, reduce: UnaryOpTerm
			reduce(199), // *, reduce: UnaryOpTerm
			reduce(199), // /, reduce: UnaryOpTerm
			reduce(199), // //, reduce: UnaryOpTerm
			reduce(199), // %, reduce: UnaryOpTerm
			reduce(199), // .*, reduce: UnaryOpTerm
			reduce(199), // ./, reduce: UnaryOpTerm
			reduce(199), // .//, reduce: UnaryOpTerm
			reduce(199), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2947
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(200), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(200), // >, reduce: UnaryOpTerm
			reduce(200), // >>, reduce: UnaryOpTerm
			reduce(200), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(200), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(200), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(200), // ||, reduce: UnaryOpTerm
			reduce(200), // ^^, reduce: UnaryOpTerm
			reduce(200), // &&, reduce: UnaryOpTerm
			reduce(200), // ??, reduce: UnaryOpTerm
			reduce(200), // ???, reduce: UnaryOpTerm
			reduce(200), // =~, reduce: UnaryOpTerm
			reduce(200), // !=~, reduce: UnaryOpTerm
			reduce(200), // ==, reduce: UnaryOpTerm
			reduce(200), // !=, reduce: UnaryOpTerm
			reduce(200), // <=>, reduce: UnaryOpTerm
			reduce(200), // >=, reduce: UnaryOpTerm
			reduce(200), // <, reduce: UnaryOpTerm
			reduce(200), // <=, reduce: UnaryOpTerm
			reduce(200), // ^, reduce: UnaryOpTerm
			reduce(200), // &, reduce: UnaryOpTerm
			reduce(200), // <<, reduce: UnaryOpTerm
			reduce(200), // >>>, reduce: UnaryOpTerm
			reduce(200), // +, reduce: UnaryOpTerm
			reduce(200), // -, reduce: UnaryOpTerm
			reduce(200), // .+, reduce: UnaryOpTerm
			reduce(200), // .-, reduce: UnaryOpTerm
			reduce(200), // *, reduce: UnaryOpTerm
			reduce(200), // /, reduce: UnaryOpTerm
			reduce(200), // //, reduce: UnaryOpTerm
			reduce(200), // %, reduce: UnaryOpTerm
			reduce(200), // .*, reduce: UnaryOpTerm
			reduce(200), // ./, reduce: UnaryOpTerm
			reduce(200), // .//, reduce: UnaryOpTerm
			reduce(200), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2948
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2949
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(201), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(201), // >, reduce: UnaryOpTerm
			reduce(201), // >>, reduce: UnaryOpTerm
			reduce(201), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(201), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(201), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(201), // ||, reduce: UnaryOpTerm
			reduce(201), // ^^, reduce: UnaryOpTerm
			reduce(201), // &&, reduce: UnaryOpTerm
			reduce(201), // ??, reduce: UnaryOpTerm
			reduce(201), // ???, reduce: UnaryOpTerm
			reduce(201), // =~, reduce: UnaryOpTerm
			reduce(201), // !=~, reduce: UnaryOpTerm
			reduce(201), // ==, reduce: UnaryOpTerm
			reduce(201), // !=, reduce: UnaryOpTerm
			reduce(201), // <=>, reduce: UnaryOpTerm
			reduce(201), // >=, reduce: UnaryOpTerm
			reduce(201), // <, reduce: UnaryOpTerm
			reduce(201), // <=, reduce: UnaryOpTerm
			reduce(201), // ^, reduce: UnaryOpTerm
			reduce(201), // &, reduce: UnaryOpTerm
			reduce(201), // <<, reduce: UnaryOpTerm
			reduce(201), // >>>, reduce: UnaryOpTerm
			reduce(201), // +, reduce: UnaryOpTerm
			reduce(201), // -, reduce: UnaryOpTerm
			reduce(201), // .+, reduce: UnaryOpTerm
			reduce(201), // .-, reduce: UnaryOpTerm
			reduce(201), // *, reduce: UnaryOpTerm
			reduce(201), // /, reduce: UnaryOpTerm
			reduce(201), // //, reduce: UnaryOpTerm
			reduce(201), // %, reduce: UnaryOpTerm
			reduce(201), // .*, reduce: UnaryOpTerm
			reduce(201), // ./, reduce: UnaryOpTerm
			reduce(201), // .//, reduce: UnaryOpTerm
			reduce(201), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2950
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(202), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(202), // >, reduce: UnaryOpTerm
			reduce(202), // >>, reduce: UnaryOpTerm
			reduce(202), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(202), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(202), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(202), // ||, reduce: UnaryOpTerm
			reduce(202), // ^^, reduce: UnaryOpTerm
			reduce(202), // &&, reduce: UnaryOpTerm
			reduce(202), // ??, reduce: UnaryOpTerm
			reduce(202), // ???, reduce: UnaryOpTerm
			reduce(202), // =~, reduce: UnaryOpTerm
			reduce(202), // !=~, reduce: UnaryOpTerm
			reduce(202), // ==, reduce: UnaryOpTerm
			reduce(202), // !=, reduce: UnaryOpTerm
			reduce(202), // <=>, reduce: UnaryOpTerm
			reduce(202), // >=, reduce: UnaryOpTerm
			reduce(202), // <, reduce: UnaryOpTerm
			reduce(202), // <=, reduce: UnaryOpTerm
			reduce(202), // ^, reduce: UnaryOpTerm
			reduce(202), // &, reduce: UnaryOpTerm
			reduce(202), // <<, reduce: UnaryOpTerm
			reduce(202), // >>>, reduce: UnaryOpTerm
			reduce(202), // +, reduce: UnaryOpTerm
			reduce(202), // -, reduce: UnaryOpTerm
			reduce(202), // .+, reduce: UnaryOpTerm
			reduce(202), // .-, reduce: UnaryOpTerm
			reduce(202), // *, reduce: UnaryOpTerm
			reduce(202), // /, reduce: UnaryOpTerm
			reduce(202), // //, reduce: UnaryOpTerm
			reduce(202), // %, reduce: UnaryOpTerm
			reduce(202), // .*, reduce: UnaryOpTerm
			reduce(202), // ./, reduce: UnaryOpTerm
			reduce(202), // .//, reduce: UnaryOpTerm
			reduce(202), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2951
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3758), // +
			shift(3759), // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2952
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3762), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2953
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2954
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2955
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(228), // }, reduce: ArrayLiteral
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(228), // >, reduce: ArrayLiteral
			reduce(228), // >>, reduce: ArrayLiteral
			reduce(228), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(228), // ,, reduce: ArrayLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(228), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(228), // ||, reduce: ArrayLiteral
			reduce(228), // ^^, reduce: ArrayLiteral
			reduce(228), // &&, reduce: ArrayLiteral
			reduce(228), // ??, reduce: ArrayLiteral
			reduce(228), // ???, reduce: ArrayLiteral
			reduce(228), // =~, reduce: ArrayLiteral
			reduce(228), // !=~, reduce: ArrayLiteral
			reduce(228), // ==, reduce: ArrayLiteral
			reduce(228), // !=, reduce: ArrayLiteral
			reduce(228), // <=>, reduce: ArrayLiteral
			reduce(228), // >=, reduce: ArrayLiteral
			reduce(228), // <, reduce: ArrayLiteral
			reduce(228), // <=, reduce: ArrayLiteral
			reduce(228), // ^, reduce: ArrayLiteral
			reduce(228), // &, reduce: ArrayLiteral
			reduce(228), // <<, reduce: ArrayLiteral
			reduce(228), // >>>, reduce: ArrayLiteral
			reduce(228), // +, reduce: ArrayLiteral
			reduce(228), // -, reduce: ArrayLiteral
			reduce(228), // .+, reduce: ArrayLiteral
			reduce(228), // .-, reduce: ArrayLiteral
			reduce(228), // *, reduce: ArrayLiteral
			reduce(228), // /, reduce: ArrayLiteral
			reduce(228), // //, reduce: ArrayLiteral
			reduce(228), // %, reduce: ArrayLiteral
			reduce(228), // .*, reduce: ArrayLiteral
			reduce(228), // ./, reduce: ArrayLiteral
			reduce(228), // .//, reduce: ArrayLiteral
			reduce(228), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(228), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(228), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(228), // [[, reduce: ArrayLiteral
			reduce(228), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2956
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3765), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2957
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(3766), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2958
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2959
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(3769), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2960
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2961
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(107), // >, reduce: PositionalFieldName
			reduce(107), // >>, reduce: PositionalFieldName
			reduce(107), // |, reduce: PositionalFieldName
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(107), // ?, reduce: PositionalFieldName
			reduce(107), // :, reduce: PositionalFieldName
			reduce(107), // ||, reduce: PositionalFieldName
			reduce(107), // ^^, reduce: PositionalFieldName
			reduce(107), // &&, reduce: PositionalFieldName
			reduce(107), // ??, reduce: PositionalFieldName
			reduce(107), // ???, reduce: PositionalFieldName
			reduce(107), // =~, reduce: PositionalFieldName
			reduce(107), // !=~, reduce: PositionalFieldName
			reduce(107), // ==, reduce: PositionalFieldName
			reduce(107), // !=, reduce: PositionalFieldName
			reduce(107), // <=>, reduce: PositionalFieldName
			reduce(107), // >=, reduce: PositionalFieldName
			reduce(107), // <, reduce: PositionalFieldName
			reduce(107), // <=, reduce: PositionalFieldName
			reduce(107), // ^, reduce: PositionalFieldName
			reduce(107), // &, reduce: PositionalFieldName
			reduce(107), // <<, reduce: PositionalFieldName
			reduce(107), // >>>, reduce: PositionalFieldName
			reduce(107), // +, reduce: PositionalFieldName
			reduce(107), // -, reduce: PositionalFieldName
			reduce(107), // .+, reduce: PositionalFieldName
			reduce(107), // .-, reduce: PositionalFieldName
			reduce(107), // *, reduce: PositionalFieldName
			reduce(107), // /, reduce: PositionalFieldName
			reduce(107), // //, reduce: PositionalFieldName
			reduce(107), // %, reduce: PositionalFieldName
			reduce(107), // .*, reduce: PositionalFieldName
			reduce(107), // ./, reduce: PositionalFieldName
			reduce(107), // .//, reduce: PositionalFieldName
			reduce(107), // ., reduce: PositionalFieldName
			nil,         // !
			nil,         // ~
			reduce(107), // **, reduce: PositionalFieldName
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(107), // [, reduce: PositionalFieldName
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(107), // [[, reduce: PositionalFieldName
			reduce(107), // [[[, reduce: PositionalFieldName
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2962
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3771), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2963
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2964
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(206), // >, reduce: PowTerm
			reduce(206), // >>, reduce: PowTerm
			reduce(206), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(206), // ?, reduce: PowTerm
			reduce(206), // :, reduce: PowTerm
			reduce(206), // ||, reduce: PowTerm
			reduce(206), // ^^, reduce: PowTerm
			reduce(206), // &&, reduce: PowTerm
			reduce(206), // ??, reduce: PowTerm
			reduce(206), // ???, reduce: PowTerm
			reduce(206), // =~, reduce: PowTerm
			reduce(206), // !=~, reduce: PowTerm
			reduce(206), // ==, reduce: PowTerm
			reduce(206), // !=, reduce: PowTerm
			reduce(206), // <=>, reduce: PowTerm
			reduce(206), // >=, reduce: PowTerm
			reduce(206), // <, reduce: PowTerm
			reduce(206), // <=, reduce: PowTerm
			reduce(206), // ^, reduce: PowTerm
			reduce(206), // &, reduce: PowTerm
			reduce(206), // <<, reduce: PowTerm
			reduce(206), // >>>, reduce: PowTerm
			reduce(206), // +, reduce: PowTerm
			reduce(206), // -, reduce: PowTerm
			reduce(206), // .+, reduce: PowTerm
			reduce(206), // .-, reduce: PowTerm
			reduce(206), // *, reduce: PowTerm
			reduce(206), // /, reduce: PowTerm
			reduce(206), // //, reduce: PowTerm
			reduce(206), // %, reduce: PowTerm
			reduce(206), // .*, reduce: PowTerm
			reduce(206), // ./, reduce: PowTerm
			reduce(206), // .//, reduce: PowTerm
			reduce(206), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2965
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(205), // >, reduce: PowTerm
			reduce(205), // >>, reduce: PowTerm
			reduce(205), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(205), // ?, reduce: PowTerm
			reduce(205), // :, reduce: PowTerm
			reduce(205), // ||, reduce: PowTerm
			reduce(205), // ^^, reduce: PowTerm
			reduce(205), // &&, reduce: PowTerm
			reduce(205), // ??, reduce: PowTerm
			reduce(205), // ???, reduce: PowTerm
			reduce(205), // =~, reduce: PowTerm
			reduce(205), // !=~, reduce: PowTerm
			reduce(205), // ==, reduce: PowTerm
			reduce(205), // !=, reduce: PowTerm
			reduce(205), // <=>, reduce: PowTerm
			reduce(205), // >=, reduce: PowTerm
			reduce(205), // <, reduce: PowTerm
			reduce(205), // <=, reduce: PowTerm
			reduce(205), // ^, reduce: PowTerm
			reduce(205), // &, reduce: PowTerm
			reduce(205), // <<, reduce: PowTerm
			reduce(205), // >>>, reduce: PowTerm
			reduce(205), // +, reduce: PowTerm
			reduce(205), // -, reduce: PowTerm
			reduce(205), // .+, reduce: PowTerm
			reduce(205), // .-, reduce: PowTerm
			reduce(205), // *, reduce: PowTerm
			reduce(205), // /, reduce: PowTerm
			reduce(205), // //, reduce: PowTerm
			reduce(205), // %, reduce: PowTerm
			reduce(205), // .*, reduce: PowTerm
			reduce(205), // ./, reduce: PowTerm
			reduce(205), // .//, reduce: PowTerm
			reduce(205), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2966
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(260), // >, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>, reduce: ArrayOrMapIndexAccess
			reduce(260), // |, reduce: ArrayOrMapIndexAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(260), // ?, reduce: ArrayOrMapIndexAccess
			reduce(260), // :, reduce: ArrayOrMapIndexAccess
			reduce(260), // ||, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???, reduce: ArrayOrMapIndexAccess
			reduce(260), // =~, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=~, reduce: ArrayOrMapIndexAccess
			reduce(260), // ==, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=>, reduce: ArrayOrMapIndexAccess
			reduce(260), // >=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>, reduce: ArrayOrMapIndexAccess
			reduce(260), // +, reduce: ArrayOrMapIndexAccess
			reduce(260), // -, reduce: ArrayOrMapIndexAccess
			reduce(260), // .+, reduce: ArrayOrMapIndexAccess
			reduce(260), // .-, reduce: ArrayOrMapIndexAccess
			reduce(260), // *, reduce: ArrayOrMapIndexAccess
			reduce(260), // /, reduce: ArrayOrMapIndexAccess
			reduce(260), // //, reduce: ArrayOrMapIndexAccess
			reduce(260), // %, reduce: ArrayOrMapIndexAccess
			reduce(260), // .*, reduce: ArrayOrMapIndexAccess
			reduce(260), // ./, reduce: ArrayOrMapIndexAccess
			reduce(260), // .//, reduce: ArrayOrMapIndexAccess
			reduce(260), // ., reduce: ArrayOrMapIndexAccess
			nil,         // !
			nil,         // ~
			reduce(260), // **, reduce: ArrayOrMapIndexAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(260), // [, reduce: ArrayOrMapIndexAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(260), // [[, reduce: ArrayOrMapIndexAccess
			reduce(260), // [[[, reduce: ArrayOrMapIndexAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2967
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(3774), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2968
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3775), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2969
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(266), // >, reduce: ArraySliceAccess
			reduce(266), // >>, reduce: ArraySliceAccess
			reduce(266), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(266), // ?, reduce: ArraySliceAccess
			reduce(266), // :, reduce: ArraySliceAccess
			reduce(266), // ||, reduce: ArraySliceAccess
			reduce(266), // ^^, reduce: ArraySliceAccess
			reduce(266), // &&, reduce: ArraySliceAccess
			reduce(266), // ??, reduce: ArraySliceAccess
			reduce(266), // ???, reduce: ArraySliceAccess
			reduce(266), // =~, reduce: ArraySliceAccess
			reduce(266), // !=~, reduce: ArraySliceAccess
			reduce(266), // ==, reduce: ArraySliceAccess
			reduce(266), // !=, reduce: ArraySliceAccess
			reduce(266), // <=>, reduce: ArraySliceAccess
			reduce(266), // >=, reduce: ArraySliceAccess
			reduce(266), // <, reduce: ArraySliceAccess
			reduce(266), // <=, reduce: ArraySliceAccess
			reduce(266), // ^, reduce: ArraySliceAccess
			reduce(266), // &, reduce: ArraySliceAccess
			reduce(266), // <<, reduce: ArraySliceAccess
			reduce(266), // >>>, reduce: ArraySliceAccess
			reduce(266), // +, reduce: ArraySliceAccess
			reduce(266), // -, reduce: ArraySliceAccess
			reduce(266), // .+, reduce: ArraySliceAccess
			reduce(266), // .-, reduce: ArraySliceAccess
			reduce(266), // *, reduce: ArraySliceAccess
			reduce(266), // /, reduce: ArraySliceAccess
			reduce(266), // //, reduce: ArraySliceAccess
			reduce(266), // %, reduce: ArraySliceAccess
			reduce(266), // .*, reduce: ArraySliceAccess
			reduce(266), // ./, reduce: ArraySliceAccess
			reduce(266), // .//, reduce: ArraySliceAccess
			reduce(266), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(266), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(266), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(266), // [[, reduce: ArraySliceAccess
			reduce(266), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2970
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3776), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2971
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3777), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2972
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(254), // >, reduce: ENV
			reduce(254), // >>, reduce: ENV
			reduce(254), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(254), // ?, reduce: ENV
			reduce(254), // :, reduce: ENV
			reduce(254), // ||, reduce: ENV
			reduce(254), // ^^, reduce: ENV
			reduce(254), // &&, reduce: ENV
			reduce(254), // ??, reduce: ENV
			reduce(254), // ???, reduce: ENV
			reduce(254), // =~, reduce: ENV
			reduce(254), // !=~, reduce: ENV
			reduce(254), // ==, reduce: ENV
			reduce(254), // !=, reduce: ENV
			reduce(254), // <=>, reduce: ENV
			reduce(254), // >=, reduce: ENV
			reduce(254), // <, reduce: ENV
			reduce(254), // <=, reduce: ENV
			reduce(254), // ^, reduce: ENV
			reduce(254), // &, reduce: ENV
			reduce(254), // <<, reduce: ENV
			reduce(254), // >>>, reduce: ENV
			reduce(254), // +, reduce: ENV
			reduce(254), // -, reduce: ENV
			reduce(254), // .+, reduce: ENV
			reduce(254), // .-, reduce: ENV
			reduce(254), // *, reduce: ENV
			reduce(254), // /, reduce: ENV
			reduce(254), // //, reduce: ENV
			reduce(254), // %, reduce: ENV
			reduce(254), // .*, reduce: ENV
			reduce(254), // ./, reduce: ENV
			reduce(254), // .//, reduce: ENV
			reduce(254), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(254), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(254), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(254), // [[, reduce: ENV
			reduce(254), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2973
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(269), // >, reduce: FunctionCallsite
			reduce(269), // >>, reduce: FunctionCallsite
			reduce(269), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(269), // ?, reduce: FunctionCallsite
			reduce(269), // :, reduce: FunctionCallsite
			reduce(269), // ||, reduce: FunctionCallsite
			reduce(269), // ^^, reduce: FunctionCallsite
			reduce(269), // &&, reduce: FunctionCallsite
			reduce(269), // ??, reduce: FunctionCallsite
			reduce(269), // ???, reduce: FunctionCallsite
			reduce(269), // =~, reduce: FunctionCallsite
			reduce(269), // !=~, reduce: FunctionCallsite
			reduce(269), // ==, reduce: FunctionCallsite
			reduce(269), // !=, reduce: FunctionCallsite
			reduce(269), // <=>, reduce: FunctionCallsite
			reduce(269), // >=, reduce: FunctionCallsite
			reduce(269), // <, reduce: FunctionCallsite
			reduce(269), // <=, reduce: FunctionCallsite
			reduce(269), // ^, reduce: FunctionCallsite
			reduce(269), // &, reduce: FunctionCallsite
			reduce(269), // <<, reduce: FunctionCallsite
			reduce(269), // >>>, reduce: FunctionCallsite
			reduce(269), // +, reduce: FunctionCallsite
			reduce(269), // -, reduce: FunctionCallsite
			reduce(269), // .+, reduce: FunctionCallsite
			reduce(269), // .-, reduce: FunctionCallsite
			reduce(269), // *, reduce: FunctionCallsite
			reduce(269), // /, reduce: FunctionCallsite
			reduce(269), // //, reduce: FunctionCallsite
			reduce(269), // %, reduce: FunctionCallsite
			reduce(269), // .*, reduce: FunctionCallsite
			reduce(269), // ./, reduce: FunctionCallsite
			reduce(269), // .//, reduce: FunctionCallsite
			reduce(269), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(269), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(269), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(269), // [[, reduce: FunctionCallsite
			reduce(269), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2974
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3779), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3780), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2975
		canRecover: false,
		actions: [numSymbols]action{
			nil,       // INVALID
			nil,       // $
			nil,       // empty
			nil,       // ;
			nil,       // {
			reduce(7), // }, reduce: NonEmptyStatementBlock
			nil,       // =
			nil,       // unset
			nil,       // filter
			nil,       // >
			nil,       // >>
			nil,       // |
			nil,       // stdout
			nil,       // stderr
			nil,       // print
			nil,       // ,
			nil,       // printn
			nil,       // eprint
			nil,       // eprintn
			nil,       // dump
			nil,       // edump
			nil,       // tee
			nil,       // emitf
			nil,       // emit1
			nil,       // emit
			nil,       // (
			nil,       // )
			nil,       // emitp
			nil,       // field_name
			nil,       // $[
			nil,       // ]
			nil,       // braced_field_name
			nil,       // $[[
			nil,       // $[[[
			nil,       // full_srec
			nil,       // oosvar_name
			nil,       // @[
			nil,       // braced_oosvar_name
			nil,       // full_oosvar
			nil,       // all
			nil,       // non_sigil_name
			nil,       // arr
			nil,       // bool
			nil,       // float
			nil,       // int
			nil,       // map
			nil,       // num
			nil,       // str
			nil,       // var
			nil,       // funct
			nil,       // ||=
			nil,       // ^^=
			nil,       // &&=
			nil,       // ??=
			nil,       // ???=
			nil,       // |=
			nil,       // &=
			nil,       // ^=
			nil,       // <<=
			nil,       // >>=
			nil,       // >>>=
			nil,       // +=
			nil,       // .=
			nil,       // -=
			nil,       // *=
			nil,       // /=
			nil,       // //=
			nil,       // %=
			nil,       // **=
			nil,       // ?
			nil,       // :
			nil,       // ||
			nil,       // ^^
			nil,       // &&
			nil,       // ??
			nil,       // ???
			nil,       // =~
			nil,       // !=~
			nil,       // ==
			nil,       // !=
			nil,       // <=>
			nil,       // >=
			nil,       // <
			nil,       // <=
			nil,       // ^
			nil,       // &
			nil,       // <<
			nil,       // >>>
			nil,       // +
			nil,       // -
			nil,       // .+
			nil,       // .-
			nil,       // *
			nil,       // /
			nil,       // //
			nil,       // %
			nil,       // .*
			nil,       // ./
			nil,       // .//
			nil,       // .
			nil,       // !
			nil,       // ~
			nil,       // **
			nil,       // string_literal
			nil,       // regex_case_insensitive
			nil,       // int_literal
			nil,       // float_literal
			nil,       // boolean_literal
			nil,       // null_literal
			nil,       // inf_literal
			nil,       // nan_literal
			nil,       // const_M_PI
			nil,       // const_M_E
			nil,       // panic
			nil,       // [
			nil,       // ctx_IPS
			nil,       // ctx_IFS
			nil,       // ctx_IRS
			nil,       // ctx_OPS
			nil,       // ctx_OFS
			nil,       // ctx_ORS
			nil,       // ctx_FLATSEP
			nil,       // ctx_NF
			nil,       // ctx_NR
			nil,       // ctx_FNR
			nil,       // ctx_FILENAME
			nil,       // ctx_FILENUM
			nil,       // env
			nil,       // [[
			nil,       // [[[
			nil,       // call
			nil,       // begin
			nil,       // end
			nil,       // if
			nil,       // elif
			nil,       // else
			nil,       // while
			nil,       // do
			nil,       // for
			nil,       // in
			nil,       // break
			nil,       // continue
			nil,       // func
			nil,       // subr
			nil,       // return
		},
	},
	actionRow{ // S2976
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(11),  // }, reduce: NonEmptyStatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S2977
		canRecover: false,
		actions: [numSymbols]action{
			nil,       // INVALID
			nil,       // $
			nil,       // empty
			nil,       // ;
			nil,       // {
			reduce(9), // }, reduce: NonEmptyStatementBlock
			nil,       // =
			nil,       // unset
			nil,       // filter
			nil,       // >
			nil,       // >>
			nil,       // |
			nil,       // stdout
			nil,       // stderr
			nil,       // print
			nil,       // ,
			nil,       // printn
			nil,       // eprint
			nil,       // eprintn
			nil,       // dump
			nil,       // edump
			nil,       // tee
			nil,       // emitf
			nil,       // emit1
			nil,       // emit
			nil,       // (
			nil,       // )
			nil,       // emitp
			nil,       // field_name
			nil,       // $[
			nil,       // ]
			nil,       // braced_field_name
			nil,       // $[[
			nil,       // $[[[
			nil,       // full_srec
			nil,       // oosvar_name
			nil,       // @[
			nil,       // braced_oosvar_name
			nil,       // full_oosvar
			nil,       // all
			nil,       // non_sigil_name
			nil,       // arr
			nil,       // bool
			nil,       // float
			nil,       // int
			nil,       // map
			nil,       // num
			nil,       // str
			nil,       // var
			nil,       // funct
			nil,       // ||=
			nil,       // ^^=
			nil,       // &&=
			nil,       // ??=
			nil,       // ???=
			nil,       // |=
			nil,       // &=
			nil,       // ^=
			nil,       // <<=
			nil,       // >>=
			nil,       // >>>=
			nil,       // +=
			nil,       // .=
			nil,       // -=
			nil,       // *=
			nil,       // /=
			nil,       // //=
			nil,       // %=
			nil,       // **=
			nil,       // ?
			nil,       // :
			nil,       // ||
			nil,       // ^^
			nil,       // &&
			nil,       // ??
			nil,       // ???
			nil,       // =~
			nil,       // !=~
			nil,       // ==
			nil,       // !=
			nil,       // <=>
			nil,       // >=
			nil,       // <
			nil,       // <=
			nil,       // ^
			nil,       // &
			nil,       // <<
			nil,       // >>>
			nil,       // +
			nil,       // -
			nil,       // .+
			nil,       // .-
			nil,       // *
			nil,       // /
			nil,       // //
			nil,       // %
			nil,       // .*
			nil,       // ./
			nil,       // .//
			nil,       // .
			nil,       // !
			nil,       // ~
			nil,       // **
			nil,       // string_literal
			nil,       // regex_case_insensitive
			nil,       // int_literal
			nil,       // float_literal
			nil,       // boolean_literal
			nil,       // null_literal
			nil,       // inf_literal
			nil,       // nan_literal
			nil,       // const_M_PI
			nil,       // const_M_E
			nil,       // panic
			nil,       // [
			nil,       // ctx_IPS
			nil,       // ctx_IFS
			nil,       // ctx_IRS
			nil,       // ctx_OPS
			nil,       // ctx_OFS
			nil,       // ctx_ORS
			nil,       // ctx_FLATSEP
			nil,       // ctx_NF
			nil,       // ctx_NR
			nil,       // ctx_FNR
			nil,       // ctx_FILENAME
			nil,       // ctx_FILENUM
			nil,       // env
			nil,       // [[
			nil,       // [[[
			nil,       // call
			nil,       // begin
			nil,       // end
			nil,       // if
			nil,       // elif
			nil,       // else
			nil,       // while
			nil,       // do
			nil,       // for
			nil,       // in
			nil,       // break
			nil,       // continue
			nil,       // func
			nil,       // subr
			nil,       // return
		},
	},
	actionRow{ // S2978
		canRecover: false,
		actions: [numSymbols]action{
			nil,       // INVALID
			nil,       // $
			nil,       // empty
			nil,       // ;
			nil,       // {
			reduce(8), // }, reduce: NonEmptyStatementBlock
			nil,       // =
			nil,       // unset
			nil,       // filter
			nil,       // >
			nil,       // >>
			nil,       // |
			nil,       // stdout
			nil,       // stderr
			nil,       // print
			nil,       // ,
			nil,       // printn
			nil,       // eprint
			nil,       // eprintn
			nil,       // dump
			nil,       // edump
			nil,       // tee
			nil,       // emitf
			nil,       // emit1
			nil,       // emit
			nil,       // (
			nil,       // )
			nil,       // emitp
			nil,       // field_name
			nil,       // $[
			nil,       // ]
			nil,       // braced_field_name
			nil,       // $[[
			nil,       // $[[[
			nil,       // full_srec
			nil,       // oosvar_name
			nil,       // @[
			nil,       // braced_oosvar_name
			nil,       // full_oosvar
			nil,       // all
			nil,       // non_sigil_name
			nil,       // arr
			nil,       // bool
			nil,       // float
			nil,       // int
			nil,       // map
			nil,       // num
			nil,       // str
			nil,       // var
			nil,       // funct
			nil,       // ||=
			nil,       // ^^=
			nil,       // &&=
			nil,       // ??=
			nil,       // ???=
			nil,       // |=
			nil,       // &=
			nil,       // ^=
			nil,       // <<=
			nil,       // >>=
			nil,       // >>>=
			nil,       // +=
			nil,       // .=
			nil,       // -=
			nil,       // *=
			nil,       // /=
			nil,       // //=
			nil,       // %=
			nil,       // **=
			nil,       // ?
			nil,       // :
			nil,       // ||
			nil,       // ^^
			nil,       // &&
			nil,       // ??
			nil,       // ???
			nil,       // =~
			nil,       // !=~
			nil,       // ==
			nil,       // !=
			nil,       // <=>
			nil,       // >=
			nil,       // <
			nil,       // <=
			nil,       // ^
			nil,       // &
			nil,       // <<
			nil,       // >>>
			nil,       // +
			nil,       // -
			nil,       // .+
			nil,       // .-
			nil,       // *
			nil,       // /
			nil,       // //
			nil,       // %
			nil,       // .*
			nil,       // ./
			nil,       // .//
			nil,       // .
			nil,       // !
			nil,       // ~
			nil,       // **
			nil,       // string_literal
			nil,       // regex_case_insensitive
			nil,       // int_literal
			nil,       // float_literal
			nil,       // boolean_literal
			nil,       // null_literal
			nil,       // inf_literal
			nil,       // nan_literal
			nil,       // const_M_PI
			nil,       // const_M_E
			nil,       // panic
			nil,       // [
			nil,       // ctx_IPS
			nil,       // ctx_IFS
			nil,       // ctx_IRS
			nil,       // ctx_OPS
			nil,       // ctx_OFS
			nil,       // ctx_ORS
			nil,       // ctx_FLATSEP
			nil,       // ctx_NF
			nil,       // ctx_NR
			nil,       // ctx_FNR
			nil,       // ctx_FILENAME
			nil,       // ctx_FILENUM
			nil,       // env
			nil,       // [[
			nil,       // [[[
			nil,       // call
			nil,       // begin
			nil,       // end
			nil,       // if
			nil,       // elif
			nil,       // else
			nil,       // while
			nil,       // do
			nil,       // for
			nil,       // in
			nil,       // break
			nil,       // continue
			nil,       // func
			nil,       // subr
			nil,       // return
		},
	},
	actionRow{ // S2979
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(235), // ;, reduce: MapLiteral
			reduce(235), // {, reduce: MapLiteral
			reduce(235), // }, reduce: MapLiteral
			reduce(235), // =, reduce: MapLiteral
			nil,         // unset
			nil,         // filter
			reduce(235), // >, reduce: MapLiteral
			reduce(235), // >>, reduce: MapLiteral
			reduce(235), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(235), // ||=, reduce: MapLiteral
			reduce(235), // ^^=, reduce: MapLiteral
			reduce(235), // &&=, reduce: MapLiteral
			reduce(235), // ??=, reduce: MapLiteral
			reduce(235), // ???=, reduce: MapLiteral
			reduce(235), // |=, reduce: MapLiteral
			reduce(235), // &=, reduce: MapLiteral
			reduce(235), // ^=, reduce: MapLiteral
			reduce(235), // <<=, reduce: MapLiteral
			reduce(235), // >>=, reduce: MapLiteral
			reduce(235), // >>>=, reduce: MapLiteral
			reduce(235), // +=, reduce: MapLiteral
			reduce(235), // .=, reduce: MapLiteral
			reduce(235), // -=, reduce: MapLiteral
			reduce(235), // *=, reduce: MapLiteral
			reduce(235), // /=, reduce: MapLiteral
			reduce(235), // //=, reduce: MapLiteral
			reduce(235), // %=, reduce: MapLiteral
			reduce(235), // **=, reduce: MapLiteral
			reduce(235), // ?, reduce: MapLiteral
			nil,         // :
			reduce(235), // ||, reduce: MapLiteral
			reduce(235), // ^^, reduce: MapLiteral
			reduce(235), // &&, reduce: MapLiteral
			reduce(235), // ??, reduce: MapLiteral
			reduce(235), // ???, reduce: MapLiteral
			reduce(235), // =~, reduce: MapLiteral
			reduce(235), // !=~, reduce: MapLiteral
			reduce(235), // ==, reduce: MapLiteral
			reduce(235), // !=, reduce: MapLiteral
			reduce(235), // <=>, reduce: MapLiteral
			reduce(235), // >=, reduce: MapLiteral
			reduce(235), // <, reduce: MapLiteral
			reduce(235), // <=, reduce: MapLiteral
			reduce(235), // ^, reduce: MapLiteral
			reduce(235), // &, reduce: MapLiteral
			reduce(235), // <<, reduce: MapLiteral
			reduce(235), // >>>, reduce: MapLiteral
			reduce(235), // +, reduce: MapLiteral
			reduce(235), // -, reduce: MapLiteral
			reduce(235), // .+, reduce: MapLiteral
			reduce(235), // .-, reduce: MapLiteral
			reduce(235), // *, reduce: MapLiteral
			reduce(235), // /, reduce: MapLiteral
			reduce(235), // //, reduce: MapLiteral
			reduce(235), // %, reduce: MapLiteral
			reduce(235), // .*, reduce: MapLiteral
			reduce(235), // ./, reduce: MapLiteral
			reduce(235), // .//, reduce: MapLiteral
			reduce(235), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(235), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(235), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(235), // [[, reduce: MapLiteral
			reduce(235), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2980
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(34), // ;, reduce: Assignment
			nil,        // {
			reduce(34), // }, reduce: Assignment
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S2981
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(128), // ;, reduce: Assignment
			nil,         // {
			reduce(128), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2982
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(129), // ;, reduce: Assignment
			nil,         // {
			reduce(129), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2983
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(130), // ;, reduce: Assignment
			nil,         // {
			reduce(130), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2984
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(131), // ;, reduce: Assignment
			nil,         // {
			reduce(131), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2985
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(132), // ;, reduce: Assignment
			nil,         // {
			reduce(132), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2986
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(133), // ;, reduce: Assignment
			nil,         // {
			reduce(133), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2987
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(134), // ;, reduce: Assignment
			nil,         // {
			reduce(134), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2988
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(135), // ;, reduce: Assignment
			nil,         // {
			reduce(135), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2989
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(136), // ;, reduce: Assignment
			nil,         // {
			reduce(136), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2990
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(137), // ;, reduce: Assignment
			nil,         // {
			reduce(137), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2991
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(138), // ;, reduce: Assignment
			nil,         // {
			reduce(138), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2992
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(139), // ;, reduce: Assignment
			nil,         // {
			reduce(139), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2993
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(140), // ;, reduce: Assignment
			nil,         // {
			reduce(140), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2994
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(141), // ;, reduce: Assignment
			nil,         // {
			reduce(141), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2995
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(142), // ;, reduce: Assignment
			nil,         // {
			reduce(142), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2996
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(143), // ;, reduce: Assignment
			nil,         // {
			reduce(143), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2997
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(144), // ;, reduce: Assignment
			nil,         // {
			reduce(144), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2998
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(145), // ;, reduce: Assignment
			nil,         // {
			reduce(145), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S2999
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(146), // ;, reduce: Assignment
			nil,         // {
			reduce(146), // }, reduce: Assignment
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3000
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(3782), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3001
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(234), // ;, reduce: MapLiteral
			nil,         // {
			reduce(234), // }, reduce: MapLiteral
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(234), // >, reduce: MapLiteral
			reduce(234), // >>, reduce: MapLiteral
			reduce(234), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(234), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(234), // ?, reduce: MapLiteral
			nil,         // :
			reduce(234), // ||, reduce: MapLiteral
			reduce(234), // ^^, reduce: MapLiteral
			reduce(234), // &&, reduce: MapLiteral
			reduce(234), // ??, reduce: MapLiteral
			reduce(234), // ???, reduce: MapLiteral
			reduce(234), // =~, reduce: MapLiteral
			reduce(234), // !=~, reduce: MapLiteral
			reduce(234), // ==, reduce: MapLiteral
			reduce(234), // !=, reduce: MapLiteral
			reduce(234), // <=>, reduce: MapLiteral
			reduce(234), // >=, reduce: MapLiteral
			reduce(234), // <, reduce: MapLiteral
			reduce(234), // <=, reduce: MapLiteral
			reduce(234), // ^, reduce: MapLiteral
			reduce(234), // &, reduce: MapLiteral
			reduce(234), // <<, reduce: MapLiteral
			reduce(234), // >>>, reduce: MapLiteral
			reduce(234), // +, reduce: MapLiteral
			reduce(234), // -, reduce: MapLiteral
			reduce(234), // .+, reduce: MapLiteral
			reduce(234), // .-, reduce: MapLiteral
			reduce(234), // *, reduce: MapLiteral
			reduce(234), // /, reduce: MapLiteral
			reduce(234), // //, reduce: MapLiteral
			reduce(234), // %, reduce: MapLiteral
			reduce(234), // .*, reduce: MapLiteral
			reduce(234), // ./, reduce: MapLiteral
			reduce(234), // .//, reduce: MapLiteral
			reduce(234), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(234), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(234), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(234), // [[, reduce: MapLiteral
			reduce(234), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3002
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(3783), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3003
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(274), // ;, reduce: FcnArgs
			shift(2043), // {
			reduce(274), // }, reduce: FcnArgs
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3004
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3785), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3005
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3786), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3006
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3787), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3007
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3788), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3008
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3789), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3009
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3010
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3011
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3012
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3013
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3014
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3015
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3016
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3017
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3018
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3019
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3020
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3021
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3022
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3023
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3024
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3025
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3026
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3027
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3028
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3029
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3030
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3031
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3032
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3033
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3034
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(197), // ;, reduce: UnaryOpTerm
			nil,         // {
			reduce(197), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(197), // >, reduce: UnaryOpTerm
			reduce(197), // >>, reduce: UnaryOpTerm
			reduce(197), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(197), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(197), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(197), // ||, reduce: UnaryOpTerm
			reduce(197), // ^^, reduce: UnaryOpTerm
			reduce(197), // &&, reduce: UnaryOpTerm
			reduce(197), // ??, reduce: UnaryOpTerm
			reduce(197), // ???, reduce: UnaryOpTerm
			reduce(197), // =~, reduce: UnaryOpTerm
			reduce(197), // !=~, reduce: UnaryOpTerm
			reduce(197), // ==, reduce: UnaryOpTerm
			reduce(197), // !=, reduce: UnaryOpTerm
			reduce(197), // <=>, reduce: UnaryOpTerm
			reduce(197), // >=, reduce: UnaryOpTerm
			reduce(197), // <, reduce: UnaryOpTerm
			reduce(197), // <=, reduce: UnaryOpTerm
			reduce(197), // ^, reduce: UnaryOpTerm
			reduce(197), // &, reduce: UnaryOpTerm
			reduce(197), // <<, reduce: UnaryOpTerm
			reduce(197), // >>>, reduce: UnaryOpTerm
			reduce(197), // +, reduce: UnaryOpTerm
			reduce(197), // -, reduce: UnaryOpTerm
			reduce(197), // .+, reduce: UnaryOpTerm
			reduce(197), // .-, reduce: UnaryOpTerm
			reduce(197), // *, reduce: UnaryOpTerm
			reduce(197), // /, reduce: UnaryOpTerm
			reduce(197), // //, reduce: UnaryOpTerm
			reduce(197), // %, reduce: UnaryOpTerm
			reduce(197), // .*, reduce: UnaryOpTerm
			reduce(197), // ./, reduce: UnaryOpTerm
			reduce(197), // .//, reduce: UnaryOpTerm
			reduce(197), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3035
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3036
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3037
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3038
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3039
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3040
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3041
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3042
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(198), // ;, reduce: UnaryOpTerm
			nil,         // {
			reduce(198), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(198), // >, reduce: UnaryOpTerm
			reduce(198), // >>, reduce: UnaryOpTerm
			reduce(198), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(198), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(198), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(198), // ||, reduce: UnaryOpTerm
			reduce(198), // ^^, reduce: UnaryOpTerm
			reduce(198), // &&, reduce: UnaryOpTerm
			reduce(198), // ??, reduce: UnaryOpTerm
			reduce(198), // ???, reduce: UnaryOpTerm
			reduce(198), // =~, reduce: UnaryOpTerm
			reduce(198), // !=~, reduce: UnaryOpTerm
			reduce(198), // ==, reduce: UnaryOpTerm
			reduce(198), // !=, reduce: UnaryOpTerm
			reduce(198), // <=>, reduce: UnaryOpTerm
			reduce(198), // >=, reduce: UnaryOpTerm
			reduce(198), // <, reduce: UnaryOpTerm
			reduce(198), // <=, reduce: UnaryOpTerm
			reduce(198), // ^, reduce: UnaryOpTerm
			reduce(198), // &, reduce: UnaryOpTerm
			reduce(198), // <<, reduce: UnaryOpTerm
			reduce(198), // >>>, reduce: UnaryOpTerm
			reduce(198), // +, reduce: UnaryOpTerm
			reduce(198), // -, reduce: UnaryOpTerm
			reduce(198), // .+, reduce: UnaryOpTerm
			reduce(198), // .-, reduce: UnaryOpTerm
			reduce(198), // *, reduce: UnaryOpTerm
			reduce(198), // /, reduce: UnaryOpTerm
			reduce(198), // //, reduce: UnaryOpTerm
			reduce(198), // %, reduce: UnaryOpTerm
			reduce(198), // .*, reduce: UnaryOpTerm
			reduce(198), // ./, reduce: UnaryOpTerm
			reduce(198), // .//, reduce: UnaryOpTerm
			reduce(198), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3043
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(199), // ;, reduce: UnaryOpTerm
			nil,         // {
			reduce(199), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(199), // >, reduce: UnaryOpTerm
			reduce(199), // >>, reduce: UnaryOpTerm
			reduce(199), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(199), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(199), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(199), // ||, reduce: UnaryOpTerm
			reduce(199), // ^^, reduce: UnaryOpTerm
			reduce(199), // &&, reduce: UnaryOpTerm
			reduce(199), // ??, reduce: UnaryOpTerm
			reduce(199), // ???, reduce: UnaryOpTerm
			reduce(199), // =~, reduce: UnaryOpTerm
			reduce(199), // !=~, reduce: UnaryOpTerm
			reduce(199), // ==, reduce: UnaryOpTerm
			reduce(199), // !=, reduce: UnaryOpTerm
			reduce(199), // <=>, reduce: UnaryOpTerm
			reduce(199), // >=, reduce: UnaryOpTerm
			reduce(199), // <, reduce: UnaryOpTerm
			reduce(199), // <=, reduce: UnaryOpTerm
			reduce(199), // ^, reduce: UnaryOpTerm
			reduce(199), // &, reduce: UnaryOpTerm
			reduce(199), // <<, reduce: UnaryOpTerm
			reduce(199), // >>>, reduce: UnaryOpTerm
			reduce(199), // +, reduce: UnaryOpTerm
			reduce(199), // -, reduce: UnaryOpTerm
			reduce(199), // .+, reduce: UnaryOpTerm
			reduce(199), // .-, reduce: UnaryOpTerm
			reduce(199), // *, reduce: UnaryOpTerm
			reduce(199), // /, reduce: UnaryOpTerm
			reduce(199), // //, reduce: UnaryOpTerm
			reduce(199), // %, reduce: UnaryOpTerm
			reduce(199), // .*, reduce: UnaryOpTerm
			reduce(199), // ./, reduce: UnaryOpTerm
			reduce(199), // .//, reduce: UnaryOpTerm
			reduce(199), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3044
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(200), // ;, reduce: UnaryOpTerm
			nil,         // {
			reduce(200), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(200), // >, reduce: UnaryOpTerm
			reduce(200), // >>, reduce: UnaryOpTerm
			reduce(200), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(200), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(200), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(200), // ||, reduce: UnaryOpTerm
			reduce(200), // ^^, reduce: UnaryOpTerm
			reduce(200), // &&, reduce: UnaryOpTerm
			reduce(200), // ??, reduce: UnaryOpTerm
			reduce(200), // ???, reduce: UnaryOpTerm
			reduce(200), // =~, reduce: UnaryOpTerm
			reduce(200), // !=~, reduce: UnaryOpTerm
			reduce(200), // ==, reduce: UnaryOpTerm
			reduce(200), // !=, reduce: UnaryOpTerm
			reduce(200), // <=>, reduce: UnaryOpTerm
			reduce(200), // >=, reduce: UnaryOpTerm
			reduce(200), // <, reduce: UnaryOpTerm
			reduce(200), // <=, reduce: UnaryOpTerm
			reduce(200), // ^, reduce: UnaryOpTerm
			reduce(200), // &, reduce: UnaryOpTerm
			reduce(200), // <<, reduce: UnaryOpTerm
			reduce(200), // >>>, reduce: UnaryOpTerm
			reduce(200), // +, reduce: UnaryOpTerm
			reduce(200), // -, reduce: UnaryOpTerm
			reduce(200), // .+, reduce: UnaryOpTerm
			reduce(200), // .-, reduce: UnaryOpTerm
			reduce(200), // *, reduce: UnaryOpTerm
			reduce(200), // /, reduce: UnaryOpTerm
			reduce(200), // //, reduce: UnaryOpTerm
			reduce(200), // %, reduce: UnaryOpTerm
			reduce(200), // .*, reduce: UnaryOpTerm
			reduce(200), // ./, reduce: UnaryOpTerm
			reduce(200), // .//, reduce: UnaryOpTerm
			reduce(200), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3045
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3046
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(201), // ;, reduce: UnaryOpTerm
			nil,         // {
			reduce(201), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(201), // >, reduce: UnaryOpTerm
			reduce(201), // >>, reduce: UnaryOpTerm
			reduce(201), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(201), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(201), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(201), // ||, reduce: UnaryOpTerm
			reduce(201), // ^^, reduce: UnaryOpTerm
			reduce(201), // &&, reduce: UnaryOpTerm
			reduce(201), // ??, reduce: UnaryOpTerm
			reduce(201), // ???, reduce: UnaryOpTerm
			reduce(201), // =~, reduce: UnaryOpTerm
			reduce(201), // !=~, reduce: UnaryOpTerm
			reduce(201), // ==, reduce: UnaryOpTerm
			reduce(201), // !=, reduce: UnaryOpTerm
			reduce(201), // <=>, reduce: UnaryOpTerm
			reduce(201), // >=, reduce: UnaryOpTerm
			reduce(201), // <, reduce: UnaryOpTerm
			reduce(201), // <=, reduce: UnaryOpTerm
			reduce(201), // ^, reduce: UnaryOpTerm
			reduce(201), // &, reduce: UnaryOpTerm
			reduce(201), // <<, reduce: UnaryOpTerm
			reduce(201), // >>>, reduce: UnaryOpTerm
			reduce(201), // +, reduce: UnaryOpTerm
			reduce(201), // -, reduce: UnaryOpTerm
			reduce(201), // .+, reduce: UnaryOpTerm
			reduce(201), // .-, reduce: UnaryOpTerm
			reduce(201), // *, reduce: UnaryOpTerm
			reduce(201), // /, reduce: UnaryOpTerm
			reduce(201), // //, reduce: UnaryOpTerm
			reduce(201), // %, reduce: UnaryOpTerm
			reduce(201), // .*, reduce: UnaryOpTerm
			reduce(201), // ./, reduce: UnaryOpTerm
			reduce(201), // .//, reduce: UnaryOpTerm
			reduce(201), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3047
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(202), // ;, reduce: UnaryOpTerm
			nil,         // {
			reduce(202), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(202), // >, reduce: UnaryOpTerm
			reduce(202), // >>, reduce: UnaryOpTerm
			reduce(202), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(202), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(202), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(202), // ||, reduce: UnaryOpTerm
			reduce(202), // ^^, reduce: UnaryOpTerm
			reduce(202), // &&, reduce: UnaryOpTerm
			reduce(202), // ??, reduce: UnaryOpTerm
			reduce(202), // ???, reduce: UnaryOpTerm
			reduce(202), // =~, reduce: UnaryOpTerm
			reduce(202), // !=~, reduce: UnaryOpTerm
			reduce(202), // ==, reduce: UnaryOpTerm
			reduce(202), // !=, reduce: UnaryOpTerm
			reduce(202), // <=>, reduce: UnaryOpTerm
			reduce(202), // >=, reduce: UnaryOpTerm
			reduce(202), // <, reduce: UnaryOpTerm
			reduce(202), // <=, reduce: UnaryOpTerm
			reduce(202), // ^, reduce: UnaryOpTerm
			reduce(202), // &, reduce: UnaryOpTerm
			reduce(202), // <<, reduce: UnaryOpTerm
			reduce(202), // >>>, reduce: UnaryOpTerm
			reduce(202), // +, reduce: UnaryOpTerm
			reduce(202), // -, reduce: UnaryOpTerm
			reduce(202), // .+, reduce: UnaryOpTerm
			reduce(202), // .-, reduce: UnaryOpTerm
			reduce(202), // *, reduce: UnaryOpTerm
			reduce(202), // /, reduce: UnaryOpTerm
			reduce(202), // //, reduce: UnaryOpTerm
			reduce(202), // %, reduce: UnaryOpTerm
			reduce(202), // .*, reduce: UnaryOpTerm
			reduce(202), // ./, reduce: UnaryOpTerm
			reduce(202), // .//, reduce: UnaryOpTerm
			reduce(202), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3048
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3823), // +
			shift(3824), // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3049
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3827), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3050
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3051
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3052
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(228), // ;, reduce: ArrayLiteral
			nil,         // {
			reduce(228), // }, reduce: ArrayLiteral
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(228), // >, reduce: ArrayLiteral
			reduce(228), // >>, reduce: ArrayLiteral
			reduce(228), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(228), // ,, reduce: ArrayLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(228), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(228), // ||, reduce: ArrayLiteral
			reduce(228), // ^^, reduce: ArrayLiteral
			reduce(228), // &&, reduce: ArrayLiteral
			reduce(228), // ??, reduce: ArrayLiteral
			reduce(228), // ???, reduce: ArrayLiteral
			reduce(228), // =~, reduce: ArrayLiteral
			reduce(228), // !=~, reduce: ArrayLiteral
			reduce(228), // ==, reduce: ArrayLiteral
			reduce(228), // !=, reduce: ArrayLiteral
			reduce(228), // <=>, reduce: ArrayLiteral
			reduce(228), // >=, reduce: ArrayLiteral
			reduce(228), // <, reduce: ArrayLiteral
			reduce(228), // <=, reduce: ArrayLiteral
			reduce(228), // ^, reduce: ArrayLiteral
			reduce(228), // &, reduce: ArrayLiteral
			reduce(228), // <<, reduce: ArrayLiteral
			reduce(228), // >>>, reduce: ArrayLiteral
			reduce(228), // +, reduce: ArrayLiteral
			reduce(228), // -, reduce: ArrayLiteral
			reduce(228), // .+, reduce: ArrayLiteral
			reduce(228), // .-, reduce: ArrayLiteral
			reduce(228), // *, reduce: ArrayLiteral
			reduce(228), // /, reduce: ArrayLiteral
			reduce(228), // //, reduce: ArrayLiteral
			reduce(228), // %, reduce: ArrayLiteral
			reduce(228), // .*, reduce: ArrayLiteral
			reduce(228), // ./, reduce: ArrayLiteral
			reduce(228), // .//, reduce: ArrayLiteral
			reduce(228), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(228), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(228), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(228), // [[, reduce: ArrayLiteral
			reduce(228), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3053
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3830), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3054
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(3831), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3055
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3056
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(3834), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3057
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3058
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(234), // ;, reduce: MapLiteral
			nil,         // {
			reduce(234), // }, reduce: MapLiteral
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(234), // >, reduce: MapLiteral
			reduce(234), // >>, reduce: MapLiteral
			reduce(234), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(234), // ?, reduce: MapLiteral
			nil,         // :
			reduce(234), // ||, reduce: MapLiteral
			reduce(234), // ^^, reduce: MapLiteral
			reduce(234), // &&, reduce: MapLiteral
			reduce(234), // ??, reduce: MapLiteral
			reduce(234), // ???, reduce: MapLiteral
			reduce(234), // =~, reduce: MapLiteral
			reduce(234), // !=~, reduce: MapLiteral
			reduce(234), // ==, reduce: MapLiteral
			reduce(234), // !=, reduce: MapLiteral
			reduce(234), // <=>, reduce: MapLiteral
			reduce(234), // >=, reduce: MapLiteral
			reduce(234), // <, reduce: MapLiteral
			reduce(234), // <=, reduce: MapLiteral
			reduce(234), // ^, reduce: MapLiteral
			reduce(234), // &, reduce: MapLiteral
			reduce(234), // <<, reduce: MapLiteral
			reduce(234), // >>>, reduce: MapLiteral
			reduce(234), // +, reduce: MapLiteral
			reduce(234), // -, reduce: MapLiteral
			reduce(234), // .+, reduce: MapLiteral
			reduce(234), // .-, reduce: MapLiteral
			reduce(234), // *, reduce: MapLiteral
			reduce(234), // /, reduce: MapLiteral
			reduce(234), // //, reduce: MapLiteral
			reduce(234), // %, reduce: MapLiteral
			reduce(234), // .*, reduce: MapLiteral
			reduce(234), // ./, reduce: MapLiteral
			reduce(234), // .//, reduce: MapLiteral
			reduce(234), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(234), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(234), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(234), // [[, reduce: MapLiteral
			reduce(234), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3059
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(3836), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3060
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3837), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3061
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3838), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3062
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3839), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3063
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3840), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3064
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3841), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3065
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3066
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3067
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3068
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3069
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3070
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3071
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3072
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3073
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3074
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3075
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3076
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3077
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3078
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3079
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3080
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3081
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3082
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3083
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3084
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3085
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3086
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3087
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3088
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3089
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3090
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(197), // ;, reduce: UnaryOpTerm
			nil,         // {
			reduce(197), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(197), // >, reduce: UnaryOpTerm
			reduce(197), // >>, reduce: UnaryOpTerm
			reduce(197), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(197), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(197), // ||, reduce: UnaryOpTerm
			reduce(197), // ^^, reduce: UnaryOpTerm
			reduce(197), // &&, reduce: UnaryOpTerm
			reduce(197), // ??, reduce: UnaryOpTerm
			reduce(197), // ???, reduce: UnaryOpTerm
			reduce(197), // =~, reduce: UnaryOpTerm
			reduce(197), // !=~, reduce: UnaryOpTerm
			reduce(197), // ==, reduce: UnaryOpTerm
			reduce(197), // !=, reduce: UnaryOpTerm
			reduce(197), // <=>, reduce: UnaryOpTerm
			reduce(197), // >=, reduce: UnaryOpTerm
			reduce(197), // <, reduce: UnaryOpTerm
			reduce(197), // <=, reduce: UnaryOpTerm
			reduce(197), // ^, reduce: UnaryOpTerm
			reduce(197), // &, reduce: UnaryOpTerm
			reduce(197), // <<, reduce: UnaryOpTerm
			reduce(197), // >>>, reduce: UnaryOpTerm
			reduce(197), // +, reduce: UnaryOpTerm
			reduce(197), // -, reduce: UnaryOpTerm
			reduce(197), // .+, reduce: UnaryOpTerm
			reduce(197), // .-, reduce: UnaryOpTerm
			reduce(197), // *, reduce: UnaryOpTerm
			reduce(197), // /, reduce: UnaryOpTerm
			reduce(197), // //, reduce: UnaryOpTerm
			reduce(197), // %, reduce: UnaryOpTerm
			reduce(197), // .*, reduce: UnaryOpTerm
			reduce(197), // ./, reduce: UnaryOpTerm
			reduce(197), // .//, reduce: UnaryOpTerm
			reduce(197), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3091
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3092
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3093
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3094
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3095
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3096
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3097
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3098
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(198), // ;, reduce: UnaryOpTerm
			nil,         // {
			reduce(198), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(198), // >, reduce: UnaryOpTerm
			reduce(198), // >>, reduce: UnaryOpTerm
			reduce(198), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(198), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(198), // ||, reduce: UnaryOpTerm
			reduce(198), // ^^, reduce: UnaryOpTerm
			reduce(198), // &&, reduce: UnaryOpTerm
			reduce(198), // ??, reduce: UnaryOpTerm
			reduce(198), // ???, reduce: UnaryOpTerm
			reduce(198), // =~, reduce: UnaryOpTerm
			reduce(198), // !=~, reduce: UnaryOpTerm
			reduce(198), // ==, reduce: UnaryOpTerm
			reduce(198), // !=, reduce: UnaryOpTerm
			reduce(198), // <=>, reduce: UnaryOpTerm
			reduce(198), // >=, reduce: UnaryOpTerm
			reduce(198), // <, reduce: UnaryOpTerm
			reduce(198), // <=, reduce: UnaryOpTerm
			reduce(198), // ^, reduce: UnaryOpTerm
			reduce(198), // &, reduce: UnaryOpTerm
			reduce(198), // <<, reduce: UnaryOpTerm
			reduce(198), // >>>, reduce: UnaryOpTerm
			reduce(198), // +, reduce: UnaryOpTerm
			reduce(198), // -, reduce: UnaryOpTerm
			reduce(198), // .+, reduce: UnaryOpTerm
			reduce(198), // .-, reduce: UnaryOpTerm
			reduce(198), // *, reduce: UnaryOpTerm
			reduce(198), // /, reduce: UnaryOpTerm
			reduce(198), // //, reduce: UnaryOpTerm
			reduce(198), // %, reduce: UnaryOpTerm
			reduce(198), // .*, reduce: UnaryOpTerm
			reduce(198), // ./, reduce: UnaryOpTerm
			reduce(198), // .//, reduce: UnaryOpTerm
			reduce(198), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3099
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(199), // ;, reduce: UnaryOpTerm
			nil,         // {
			reduce(199), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(199), // >, reduce: UnaryOpTerm
			reduce(199), // >>, reduce: UnaryOpTerm
			reduce(199), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(199), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(199), // ||, reduce: UnaryOpTerm
			reduce(199), // ^^, reduce: UnaryOpTerm
			reduce(199), // &&, reduce: UnaryOpTerm
			reduce(199), // ??, reduce: UnaryOpTerm
			reduce(199), // ???, reduce: UnaryOpTerm
			reduce(199), // =~, reduce: UnaryOpTerm
			reduce(199), // !=~, reduce: UnaryOpTerm
			reduce(199), // ==, reduce: UnaryOpTerm
			reduce(199), // !=, reduce: UnaryOpTerm
			reduce(199), // <=>, reduce: UnaryOpTerm
			reduce(199), // >=, reduce: UnaryOpTerm
			reduce(199), // <, reduce: UnaryOpTerm
			reduce(199), // <=, reduce: UnaryOpTerm
			reduce(199), // ^, reduce: UnaryOpTerm
			reduce(199), // &, reduce: UnaryOpTerm
			reduce(199), // <<, reduce: UnaryOpTerm
			reduce(199), // >>>, reduce: UnaryOpTerm
			reduce(199), // +, reduce: UnaryOpTerm
			reduce(199), // -, reduce: UnaryOpTerm
			reduce(199), // .+, reduce: UnaryOpTerm
			reduce(199), // .-, reduce: UnaryOpTerm
			reduce(199), // *, reduce: UnaryOpTerm
			reduce(199), // /, reduce: UnaryOpTerm
			reduce(199), // //, reduce: UnaryOpTerm
			reduce(199), // %, reduce: UnaryOpTerm
			reduce(199), // .*, reduce: UnaryOpTerm
			reduce(199), // ./, reduce: UnaryOpTerm
			reduce(199), // .//, reduce: UnaryOpTerm
			reduce(199), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3100
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(200), // ;, reduce: UnaryOpTerm
			nil,         // {
			reduce(200), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(200), // >, reduce: UnaryOpTerm
			reduce(200), // >>, reduce: UnaryOpTerm
			reduce(200), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(200), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(200), // ||, reduce: UnaryOpTerm
			reduce(200), // ^^, reduce: UnaryOpTerm
			reduce(200), // &&, reduce: UnaryOpTerm
			reduce(200), // ??, reduce: UnaryOpTerm
			reduce(200), // ???, reduce: UnaryOpTerm
			reduce(200), // =~, reduce: UnaryOpTerm
			reduce(200), // !=~, reduce: UnaryOpTerm
			reduce(200), // ==, reduce: UnaryOpTerm
			reduce(200), // !=, reduce: UnaryOpTerm
			reduce(200), // <=>, reduce: UnaryOpTerm
			reduce(200), // >=, reduce: UnaryOpTerm
			reduce(200), // <, reduce: UnaryOpTerm
			reduce(200), // <=, reduce: UnaryOpTerm
			reduce(200), // ^, reduce: UnaryOpTerm
			reduce(200), // &, reduce: UnaryOpTerm
			reduce(200), // <<, reduce: UnaryOpTerm
			reduce(200), // >>>, reduce: UnaryOpTerm
			reduce(200), // +, reduce: UnaryOpTerm
			reduce(200), // -, reduce: UnaryOpTerm
			reduce(200), // .+, reduce: UnaryOpTerm
			reduce(200), // .-, reduce: UnaryOpTerm
			reduce(200), // *, reduce: UnaryOpTerm
			reduce(200), // /, reduce: UnaryOpTerm
			reduce(200), // //, reduce: UnaryOpTerm
			reduce(200), // %, reduce: UnaryOpTerm
			reduce(200), // .*, reduce: UnaryOpTerm
			reduce(200), // ./, reduce: UnaryOpTerm
			reduce(200), // .//, reduce: UnaryOpTerm
			reduce(200), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3101
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3102
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(201), // ;, reduce: UnaryOpTerm
			nil,         // {
			reduce(201), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(201), // >, reduce: UnaryOpTerm
			reduce(201), // >>, reduce: UnaryOpTerm
			reduce(201), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(201), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(201), // ||, reduce: UnaryOpTerm
			reduce(201), // ^^, reduce: UnaryOpTerm
			reduce(201), // &&, reduce: UnaryOpTerm
			reduce(201), // ??, reduce: UnaryOpTerm
			reduce(201), // ???, reduce: UnaryOpTerm
			reduce(201), // =~, reduce: UnaryOpTerm
			reduce(201), // !=~, reduce: UnaryOpTerm
			reduce(201), // ==, reduce: UnaryOpTerm
			reduce(201), // !=, reduce: UnaryOpTerm
			reduce(201), // <=>, reduce: UnaryOpTerm
			reduce(201), // >=, reduce: UnaryOpTerm
			reduce(201), // <, reduce: UnaryOpTerm
			reduce(201), // <=, reduce: UnaryOpTerm
			reduce(201), // ^, reduce: UnaryOpTerm
			reduce(201), // &, reduce: UnaryOpTerm
			reduce(201), // <<, reduce: UnaryOpTerm
			reduce(201), // >>>, reduce: UnaryOpTerm
			reduce(201), // +, reduce: UnaryOpTerm
			reduce(201), // -, reduce: UnaryOpTerm
			reduce(201), // .+, reduce: UnaryOpTerm
			reduce(201), // .-, reduce: UnaryOpTerm
			reduce(201), // *, reduce: UnaryOpTerm
			reduce(201), // /, reduce: UnaryOpTerm
			reduce(201), // //, reduce: UnaryOpTerm
			reduce(201), // %, reduce: UnaryOpTerm
			reduce(201), // .*, reduce: UnaryOpTerm
			reduce(201), // ./, reduce: UnaryOpTerm
			reduce(201), // .//, reduce: UnaryOpTerm
			reduce(201), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3103
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(202), // ;, reduce: UnaryOpTerm
			nil,         // {
			reduce(202), // }, reduce: UnaryOpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(202), // >, reduce: UnaryOpTerm
			reduce(202), // >>, reduce: UnaryOpTerm
			reduce(202), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(202), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(202), // ||, reduce: UnaryOpTerm
			reduce(202), // ^^, reduce: UnaryOpTerm
			reduce(202), // &&, reduce: UnaryOpTerm
			reduce(202), // ??, reduce: UnaryOpTerm
			reduce(202), // ???, reduce: UnaryOpTerm
			reduce(202), // =~, reduce: UnaryOpTerm
			reduce(202), // !=~, reduce: UnaryOpTerm
			reduce(202), // ==, reduce: UnaryOpTerm
			reduce(202), // !=, reduce: UnaryOpTerm
			reduce(202), // <=>, reduce: UnaryOpTerm
			reduce(202), // >=, reduce: UnaryOpTerm
			reduce(202), // <, reduce: UnaryOpTerm
			reduce(202), // <=, reduce: UnaryOpTerm
			reduce(202), // ^, reduce: UnaryOpTerm
			reduce(202), // &, reduce: UnaryOpTerm
			reduce(202), // <<, reduce: UnaryOpTerm
			reduce(202), // >>>, reduce: UnaryOpTerm
			reduce(202), // +, reduce: UnaryOpTerm
			reduce(202), // -, reduce: UnaryOpTerm
			reduce(202), // .+, reduce: UnaryOpTerm
			reduce(202), // .-, reduce: UnaryOpTerm
			reduce(202), // *, reduce: UnaryOpTerm
			reduce(202), // /, reduce: UnaryOpTerm
			reduce(202), // //, reduce: UnaryOpTerm
			reduce(202), // %, reduce: UnaryOpTerm
			reduce(202), // .*, reduce: UnaryOpTerm
			reduce(202), // ./, reduce: UnaryOpTerm
			reduce(202), // .//, reduce: UnaryOpTerm
			reduce(202), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3104
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3875), // +
			shift(3876), // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3105
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3879), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3106
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3107
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3108
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(228), // ;, reduce: ArrayLiteral
			nil,         // {
			reduce(228), // }, reduce: ArrayLiteral
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(228), // >, reduce: ArrayLiteral
			reduce(228), // >>, reduce: ArrayLiteral
			reduce(228), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(228), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(228), // ||, reduce: ArrayLiteral
			reduce(228), // ^^, reduce: ArrayLiteral
			reduce(228), // &&, reduce: ArrayLiteral
			reduce(228), // ??, reduce: ArrayLiteral
			reduce(228), // ???, reduce: ArrayLiteral
			reduce(228), // =~, reduce: ArrayLiteral
			reduce(228), // !=~, reduce: ArrayLiteral
			reduce(228), // ==, reduce: ArrayLiteral
			reduce(228), // !=, reduce: ArrayLiteral
			reduce(228), // <=>, reduce: ArrayLiteral
			reduce(228), // >=, reduce: ArrayLiteral
			reduce(228), // <, reduce: ArrayLiteral
			reduce(228), // <=, reduce: ArrayLiteral
			reduce(228), // ^, reduce: ArrayLiteral
			reduce(228), // &, reduce: ArrayLiteral
			reduce(228), // <<, reduce: ArrayLiteral
			reduce(228), // >>>, reduce: ArrayLiteral
			reduce(228), // +, reduce: ArrayLiteral
			reduce(228), // -, reduce: ArrayLiteral
			reduce(228), // .+, reduce: ArrayLiteral
			reduce(228), // .-, reduce: ArrayLiteral
			reduce(228), // *, reduce: ArrayLiteral
			reduce(228), // /, reduce: ArrayLiteral
			reduce(228), // //, reduce: ArrayLiteral
			reduce(228), // %, reduce: ArrayLiteral
			reduce(228), // .*, reduce: ArrayLiteral
			reduce(228), // ./, reduce: ArrayLiteral
			reduce(228), // .//, reduce: ArrayLiteral
			reduce(228), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(228), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(228), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(228), // [[, reduce: ArrayLiteral
			reduce(228), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3109
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3882), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3110
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(3883), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3111
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3112
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(3886), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3113
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3114
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3115
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(45), // ;, reduce: RedirectTarget
			nil,        // {
			reduce(45), // }, reduce: RedirectTarget
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(45), // ,, reduce: RedirectTarget
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3116
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(40), // ;, reduce: Redirector
			nil,        // {
			reduce(40), // }, reduce: Redirector
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(40), // ,, reduce: Redirector
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3117
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(43), // ;, reduce: RedirectTarget
			nil,        // {
			reduce(43), // }, reduce: RedirectTarget
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(43), // ,, reduce: RedirectTarget
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3118
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(44), // ;, reduce: RedirectTarget
			nil,        // {
			reduce(44), // }, reduce: RedirectTarget
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(44), // ,, reduce: RedirectTarget
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3119
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(41), // ;, reduce: Redirector
			nil,        // {
			reduce(41), // }, reduce: Redirector
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(41), // ,, reduce: Redirector
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3120
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(42), // ;, reduce: Redirector
			nil,        // {
			reduce(42), // }, reduce: Redirector
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(42), // ,, reduce: Redirector
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3121
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3122
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3123
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(3892), // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3124
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(234), // ;, reduce: MapLiteral
			nil,         // {
			reduce(234), // }, reduce: MapLiteral
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(234), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3125
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(3893), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3126
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2239), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			shift(2253), // field_name
			shift(2254), // $[
			nil,         // ]
			shift(2255), // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(2256), // full_srec
			shift(2257), // oosvar_name
			shift(2258), // @[
			shift(2259), // braced_oosvar_name
			shift(2260), // full_oosvar
			shift(2261), // all
			shift(2262), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3127
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2239), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			shift(2253), // field_name
			shift(2254), // $[
			nil,         // ]
			shift(2255), // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(2256), // full_srec
			shift(2257), // oosvar_name
			shift(2258), // @[
			shift(2259), // braced_oosvar_name
			shift(2260), // full_oosvar
			shift(2261), // all
			shift(2262), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3128
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3896), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3129
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3897), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3130
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2239), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3899), // (
			nil,         // )
			nil,         // emitp
			shift(2253), // field_name
			shift(2254), // $[
			nil,         // ]
			shift(2255), // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(2256), // full_srec
			shift(2257), // oosvar_name
			shift(2258), // @[
			shift(2259), // braced_oosvar_name
			shift(2260), // full_oosvar
			shift(2261), // all
			shift(2262), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3131
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3132
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3902), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3133
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(208), // ;, reduce: PrecedenceChainEnd
			reduce(208), // {, reduce: PrecedenceChainEnd
			reduce(208), // }, reduce: PrecedenceChainEnd
			reduce(208), // =, reduce: PrecedenceChainEnd
			nil,         // unset
			nil,         // filter
			reduce(208), // >, reduce: PrecedenceChainEnd
			reduce(208), // >>, reduce: PrecedenceChainEnd
			reduce(208), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(208), // ||=, reduce: PrecedenceChainEnd
			reduce(208), // ^^=, reduce: PrecedenceChainEnd
			reduce(208), // &&=, reduce: PrecedenceChainEnd
			reduce(208), // ??=, reduce: PrecedenceChainEnd
			reduce(208), // ???=, reduce: PrecedenceChainEnd
			reduce(208), // |=, reduce: PrecedenceChainEnd
			reduce(208), // &=, reduce: PrecedenceChainEnd
			reduce(208), // ^=, reduce: PrecedenceChainEnd
			reduce(208), // <<=, reduce: PrecedenceChainEnd
			reduce(208), // >>=, reduce: PrecedenceChainEnd
			reduce(208), // >>>=, reduce: PrecedenceChainEnd
			reduce(208), // +=, reduce: PrecedenceChainEnd
			reduce(208), // .=, reduce: PrecedenceChainEnd
			reduce(208), // -=, reduce: PrecedenceChainEnd
			reduce(208), // *=, reduce: PrecedenceChainEnd
			reduce(208), // /=, reduce: PrecedenceChainEnd
			reduce(208), // //=, reduce: PrecedenceChainEnd
			reduce(208), // %=, reduce: PrecedenceChainEnd
			reduce(208), // **=, reduce: PrecedenceChainEnd
			reduce(208), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(208), // ||, reduce: PrecedenceChainEnd
			reduce(208), // ^^, reduce: PrecedenceChainEnd
			reduce(208), // &&, reduce: PrecedenceChainEnd
			reduce(208), // ??, reduce: PrecedenceChainEnd
			reduce(208), // ???, reduce: PrecedenceChainEnd
			reduce(208), // =~, reduce: PrecedenceChainEnd
			reduce(208), // !=~, reduce: PrecedenceChainEnd
			reduce(208), // ==, reduce: PrecedenceChainEnd
			reduce(208), // !=, reduce: PrecedenceChainEnd
			reduce(208), // <=>, reduce: PrecedenceChainEnd
			reduce(208), // >=, reduce: PrecedenceChainEnd
			reduce(208), // <, reduce: PrecedenceChainEnd
			reduce(208), // <=, reduce: PrecedenceChainEnd
			reduce(208), // ^, reduce: PrecedenceChainEnd
			reduce(208), // &, reduce: PrecedenceChainEnd
			reduce(208), // <<, reduce: PrecedenceChainEnd
			reduce(208), // >>>, reduce: PrecedenceChainEnd
			reduce(208), // +, reduce: PrecedenceChainEnd
			reduce(208), // -, reduce: PrecedenceChainEnd
			reduce(208), // .+, reduce: PrecedenceChainEnd
			reduce(208), // .-, reduce: PrecedenceChainEnd
			reduce(208), // *, reduce: PrecedenceChainEnd
			reduce(208), // /, reduce: PrecedenceChainEnd
			reduce(208), // //, reduce: PrecedenceChainEnd
			reduce(208), // %, reduce: PrecedenceChainEnd
			reduce(208), // .*, reduce: PrecedenceChainEnd
			reduce(208), // ./, reduce: PrecedenceChainEnd
			reduce(208), // .//, reduce: PrecedenceChainEnd
			reduce(208), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(208), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3134
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2239), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3904), // (
			nil,         // )
			nil,         // emitp
			shift(2253), // field_name
			shift(2254), // $[
			nil,         // ]
			shift(2255), // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(2256), // full_srec
			shift(2257), // oosvar_name
			shift(2258), // @[
			shift(2259), // braced_oosvar_name
			shift(2260), // full_oosvar
			shift(2261), // all
			shift(2262), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3135
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3136
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3906), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3137
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(105), // ;, reduce: IndirectFieldValue
			reduce(105), // {, reduce: IndirectFieldValue
			reduce(105), // }, reduce: IndirectFieldValue
			reduce(105), // =, reduce: IndirectFieldValue
			nil,         // unset
			nil,         // filter
			reduce(105), // >, reduce: IndirectFieldValue
			reduce(105), // >>, reduce: IndirectFieldValue
			reduce(105), // |, reduce: IndirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(105), // ||=, reduce: IndirectFieldValue
			reduce(105), // ^^=, reduce: IndirectFieldValue
			reduce(105), // &&=, reduce: IndirectFieldValue
			reduce(105), // ??=, reduce: IndirectFieldValue
			reduce(105), // ???=, reduce: IndirectFieldValue
			reduce(105), // |=, reduce: IndirectFieldValue
			reduce(105), // &=, reduce: IndirectFieldValue
			reduce(105), // ^=, reduce: IndirectFieldValue
			reduce(105), // <<=, reduce: IndirectFieldValue
			reduce(105), // >>=, reduce: IndirectFieldValue
			reduce(105), // >>>=, reduce: IndirectFieldValue
			reduce(105), // +=, reduce: IndirectFieldValue
			reduce(105), // .=, reduce: IndirectFieldValue
			reduce(105), // -=, reduce: IndirectFieldValue
			reduce(105), // *=, reduce: IndirectFieldValue
			reduce(105), // /=, reduce: IndirectFieldValue
			reduce(105), // //=, reduce: IndirectFieldValue
			reduce(105), // %=, reduce: IndirectFieldValue
			reduce(105), // **=, reduce: IndirectFieldValue
			reduce(105), // ?, reduce: IndirectFieldValue
			nil,         // :
			reduce(105), // ||, reduce: IndirectFieldValue
			reduce(105), // ^^, reduce: IndirectFieldValue
			reduce(105), // &&, reduce: IndirectFieldValue
			reduce(105), // ??, reduce: IndirectFieldValue
			reduce(105), // ???, reduce: IndirectFieldValue
			reduce(105), // =~, reduce: IndirectFieldValue
			reduce(105), // !=~, reduce: IndirectFieldValue
			reduce(105), // ==, reduce: IndirectFieldValue
			reduce(105), // !=, reduce: IndirectFieldValue
			reduce(105), // <=>, reduce: IndirectFieldValue
			reduce(105), // >=, reduce: IndirectFieldValue
			reduce(105), // <, reduce: IndirectFieldValue
			reduce(105), // <=, reduce: IndirectFieldValue
			reduce(105), // ^, reduce: IndirectFieldValue
			reduce(105), // &, reduce: IndirectFieldValue
			reduce(105), // <<, reduce: IndirectFieldValue
			reduce(105), // >>>, reduce: IndirectFieldValue
			reduce(105), // +, reduce: IndirectFieldValue
			reduce(105), // -, reduce: IndirectFieldValue
			reduce(105), // .+, reduce: IndirectFieldValue
			reduce(105), // .-, reduce: IndirectFieldValue
			reduce(105), // *, reduce: IndirectFieldValue
			reduce(105), // /, reduce: IndirectFieldValue
			reduce(105), // //, reduce: IndirectFieldValue
			reduce(105), // %, reduce: IndirectFieldValue
			reduce(105), // .*, reduce: IndirectFieldValue
			reduce(105), // ./, reduce: IndirectFieldValue
			reduce(105), // .//, reduce: IndirectFieldValue
			reduce(105), // ., reduce: IndirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(105), // **, reduce: IndirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(105), // [, reduce: IndirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(105), // [[, reduce: IndirectFieldValue
			reduce(105), // [[[, reduce: IndirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3138
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3907), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3139
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3908), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3140
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(114), // ;, reduce: IndirectOosvarValue
			reduce(114), // {, reduce: IndirectOosvarValue
			reduce(114), // }, reduce: IndirectOosvarValue
			reduce(114), // =, reduce: IndirectOosvarValue
			nil,         // unset
			nil,         // filter
			reduce(114), // >, reduce: IndirectOosvarValue
			reduce(114), // >>, reduce: IndirectOosvarValue
			reduce(114), // |, reduce: IndirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(114), // ||=, reduce: IndirectOosvarValue
			reduce(114), // ^^=, reduce: IndirectOosvarValue
			reduce(114), // &&=, reduce: IndirectOosvarValue
			reduce(114), // ??=, reduce: IndirectOosvarValue
			reduce(114), // ???=, reduce: IndirectOosvarValue
			reduce(114), // |=, reduce: IndirectOosvarValue
			reduce(114), // &=, reduce: IndirectOosvarValue
			reduce(114), // ^=, reduce: IndirectOosvarValue
			reduce(114), // <<=, reduce: IndirectOosvarValue
			reduce(114), // >>=, reduce: IndirectOosvarValue
			reduce(114), // >>>=, reduce: IndirectOosvarValue
			reduce(114), // +=, reduce: IndirectOosvarValue
			reduce(114), // .=, reduce: IndirectOosvarValue
			reduce(114), // -=, reduce: IndirectOosvarValue
			reduce(114), // *=, reduce: IndirectOosvarValue
			reduce(114), // /=, reduce: IndirectOosvarValue
			reduce(114), // //=, reduce: IndirectOosvarValue
			reduce(114), // %=, reduce: IndirectOosvarValue
			reduce(114), // **=, reduce: IndirectOosvarValue
			reduce(114), // ?, reduce: IndirectOosvarValue
			nil,         // :
			reduce(114), // ||, reduce: IndirectOosvarValue
			reduce(114), // ^^, reduce: IndirectOosvarValue
			reduce(114), // &&, reduce: IndirectOosvarValue
			reduce(114), // ??, reduce: IndirectOosvarValue
			reduce(114), // ???, reduce: IndirectOosvarValue
			reduce(114), // =~, reduce: IndirectOosvarValue
			reduce(114), // !=~, reduce: IndirectOosvarValue
			reduce(114), // ==, reduce: IndirectOosvarValue
			reduce(114), // !=, reduce: IndirectOosvarValue
			reduce(114), // <=>, reduce: IndirectOosvarValue
			reduce(114), // >=, reduce: IndirectOosvarValue
			reduce(114), // <, reduce: IndirectOosvarValue
			reduce(114), // <=, reduce: IndirectOosvarValue
			reduce(114), // ^, reduce: IndirectOosvarValue
			reduce(114), // &, reduce: IndirectOosvarValue
			reduce(114), // <<, reduce: IndirectOosvarValue
			reduce(114), // >>>, reduce: IndirectOosvarValue
			reduce(114), // +, reduce: IndirectOosvarValue
			reduce(114), // -, reduce: IndirectOosvarValue
			reduce(114), // .+, reduce: IndirectOosvarValue
			reduce(114), // .-, reduce: IndirectOosvarValue
			reduce(114), // *, reduce: IndirectOosvarValue
			reduce(114), // /, reduce: IndirectOosvarValue
			reduce(114), // //, reduce: IndirectOosvarValue
			reduce(114), // %, reduce: IndirectOosvarValue
			reduce(114), // .*, reduce: IndirectOosvarValue
			reduce(114), // ./, reduce: IndirectOosvarValue
			reduce(114), // .//, reduce: IndirectOosvarValue
			reduce(114), // ., reduce: IndirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(114), // **, reduce: IndirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(114), // [, reduce: IndirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(114), // [[, reduce: IndirectOosvarValue
			reduce(114), // [[[, reduce: IndirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3141
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3909), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3142
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(151), // ;, reduce: LogicalOrTerm
			reduce(151), // {, reduce: LogicalOrTerm
			reduce(151), // }, reduce: LogicalOrTerm
			reduce(151), // =, reduce: LogicalOrTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(151), // ||=, reduce: LogicalOrTerm
			reduce(151), // ^^=, reduce: LogicalOrTerm
			reduce(151), // &&=, reduce: LogicalOrTerm
			reduce(151), // ??=, reduce: LogicalOrTerm
			reduce(151), // ???=, reduce: LogicalOrTerm
			reduce(151), // |=, reduce: LogicalOrTerm
			reduce(151), // &=, reduce: LogicalOrTerm
			reduce(151), // ^=, reduce: LogicalOrTerm
			reduce(151), // <<=, reduce: LogicalOrTerm
			reduce(151), // >>=, reduce: LogicalOrTerm
			reduce(151), // >>>=, reduce: LogicalOrTerm
			reduce(151), // +=, reduce: LogicalOrTerm
			reduce(151), // .=, reduce: LogicalOrTerm
			reduce(151), // -=, reduce: LogicalOrTerm
			reduce(151), // *=, reduce: LogicalOrTerm
			reduce(151), // /=, reduce: LogicalOrTerm
			reduce(151), // //=, reduce: LogicalOrTerm
			reduce(151), // %=, reduce: LogicalOrTerm
			reduce(151), // **=, reduce: LogicalOrTerm
			reduce(151), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(151), // ||, reduce: LogicalOrTerm
			shift(2278), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3143
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(153), // ;, reduce: LogicalXORTerm
			reduce(153), // {, reduce: LogicalXORTerm
			reduce(153), // }, reduce: LogicalXORTerm
			reduce(153), // =, reduce: LogicalXORTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(153), // ||=, reduce: LogicalXORTerm
			reduce(153), // ^^=, reduce: LogicalXORTerm
			reduce(153), // &&=, reduce: LogicalXORTerm
			reduce(153), // ??=, reduce: LogicalXORTerm
			reduce(153), // ???=, reduce: LogicalXORTerm
			reduce(153), // |=, reduce: LogicalXORTerm
			reduce(153), // &=, reduce: LogicalXORTerm
			reduce(153), // ^=, reduce: LogicalXORTerm
			reduce(153), // <<=, reduce: LogicalXORTerm
			reduce(153), // >>=, reduce: LogicalXORTerm
			reduce(153), // >>>=, reduce: LogicalXORTerm
			reduce(153), // +=, reduce: LogicalXORTerm
			reduce(153), // .=, reduce: LogicalXORTerm
			reduce(153), // -=, reduce: LogicalXORTerm
			reduce(153), // *=, reduce: LogicalXORTerm
			reduce(153), // /=, reduce: LogicalXORTerm
			reduce(153), // //=, reduce: LogicalXORTerm
			reduce(153), // %=, reduce: LogicalXORTerm
			reduce(153), // **=, reduce: LogicalXORTerm
			reduce(153), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(153), // ||, reduce: LogicalXORTerm
			reduce(153), // ^^, reduce: LogicalXORTerm
			shift(2279), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3144
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(155), // ;, reduce: LogicalAndTerm
			reduce(155), // {, reduce: LogicalAndTerm
			reduce(155), // }, reduce: LogicalAndTerm
			reduce(155), // =, reduce: LogicalAndTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(155), // ||=, reduce: LogicalAndTerm
			reduce(155), // ^^=, reduce: LogicalAndTerm
			reduce(155), // &&=, reduce: LogicalAndTerm
			reduce(155), // ??=, reduce: LogicalAndTerm
			reduce(155), // ???=, reduce: LogicalAndTerm
			reduce(155), // |=, reduce: LogicalAndTerm
			reduce(155), // &=, reduce: LogicalAndTerm
			reduce(155), // ^=, reduce: LogicalAndTerm
			reduce(155), // <<=, reduce: LogicalAndTerm
			reduce(155), // >>=, reduce: LogicalAndTerm
			reduce(155), // >>>=, reduce: LogicalAndTerm
			reduce(155), // +=, reduce: LogicalAndTerm
			reduce(155), // .=, reduce: LogicalAndTerm
			reduce(155), // -=, reduce: LogicalAndTerm
			reduce(155), // *=, reduce: LogicalAndTerm
			reduce(155), // /=, reduce: LogicalAndTerm
			reduce(155), // //=, reduce: LogicalAndTerm
			reduce(155), // %=, reduce: LogicalAndTerm
			reduce(155), // **=, reduce: LogicalAndTerm
			reduce(155), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(155), // ||, reduce: LogicalAndTerm
			reduce(155), // ^^, reduce: LogicalAndTerm
			reduce(155), // &&, reduce: LogicalAndTerm
			shift(2280), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3145
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(157), // ;, reduce: AbsentCoalesceTerm
			reduce(157), // {, reduce: AbsentCoalesceTerm
			reduce(157), // }, reduce: AbsentCoalesceTerm
			reduce(157), // =, reduce: AbsentCoalesceTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(157), // ||=, reduce: AbsentCoalesceTerm
			reduce(157), // ^^=, reduce: AbsentCoalesceTerm
			reduce(157), // &&=, reduce: AbsentCoalesceTerm
			reduce(157), // ??=, reduce: AbsentCoalesceTerm
			reduce(157), // ???=, reduce: AbsentCoalesceTerm
			reduce(157), // |=, reduce: AbsentCoalesceTerm
			reduce(157), // &=, reduce: AbsentCoalesceTerm
			reduce(157), // ^=, reduce: AbsentCoalesceTerm
			reduce(157), // <<=, reduce: AbsentCoalesceTerm
			reduce(157), // >>=, reduce: AbsentCoalesceTerm
			reduce(157), // >>>=, reduce: AbsentCoalesceTerm
			reduce(157), // +=, reduce: AbsentCoalesceTerm
			reduce(157), // .=, reduce: AbsentCoalesceTerm
			reduce(157), // -=, reduce: AbsentCoalesceTerm
			reduce(157), // *=, reduce: AbsentCoalesceTerm
			reduce(157), // /=, reduce: AbsentCoalesceTerm
			reduce(157), // //=, reduce: AbsentCoalesceTerm
			reduce(157), // %=, reduce: AbsentCoalesceTerm
			reduce(157), // **=, reduce: AbsentCoalesceTerm
			reduce(157), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(157), // ||, reduce: AbsentCoalesceTerm
			reduce(157), // ^^, reduce: AbsentCoalesceTerm
			reduce(157), // &&, reduce: AbsentCoalesceTerm
			reduce(157), // ??, reduce: AbsentCoalesceTerm
			shift(2281), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3146
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(159), // ;, reduce: EmptyCoalesceTerm
			reduce(159), // {, reduce: EmptyCoalesceTerm
			reduce(159), // }, reduce: EmptyCoalesceTerm
			reduce(159), // =, reduce: EmptyCoalesceTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(159), // ||=, reduce: EmptyCoalesceTerm
			reduce(159), // ^^=, reduce: EmptyCoalesceTerm
			reduce(159), // &&=, reduce: EmptyCoalesceTerm
			reduce(159), // ??=, reduce: EmptyCoalesceTerm
			reduce(159), // ???=, reduce: EmptyCoalesceTerm
			reduce(159), // |=, reduce: EmptyCoalesceTerm
			reduce(159), // &=, reduce: EmptyCoalesceTerm
			reduce(159), // ^=, reduce: EmptyCoalesceTerm
			reduce(159), // <<=, reduce: EmptyCoalesceTerm
			reduce(159), // >>=, reduce: EmptyCoalesceTerm
			reduce(159), // >>>=, reduce: EmptyCoalesceTerm
			reduce(159), // +=, reduce: EmptyCoalesceTerm
			reduce(159), // .=, reduce: EmptyCoalesceTerm
			reduce(159), // -=, reduce: EmptyCoalesceTerm
			reduce(159), // *=, reduce: EmptyCoalesceTerm
			reduce(159), // /=, reduce: EmptyCoalesceTerm
			reduce(159), // //=, reduce: EmptyCoalesceTerm
			reduce(159), // %=, reduce: EmptyCoalesceTerm
			reduce(159), // **=, reduce: EmptyCoalesceTerm
			reduce(159), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(159), // ||, reduce: EmptyCoalesceTerm
			reduce(159), // ^^, reduce: EmptyCoalesceTerm
			reduce(159), // &&, reduce: EmptyCoalesceTerm
			reduce(159), // ??, reduce: EmptyCoalesceTerm
			reduce(159), // ???, reduce: EmptyCoalesceTerm
			shift(2282), // =~
			shift(2283), // !=~
			shift(2284), // ==
			shift(2285), // !=
			shift(2286), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3147
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(161), // ;, reduce: EqneTerm
			reduce(161), // {, reduce: EqneTerm
			reduce(161), // }, reduce: EqneTerm
			reduce(161), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(2287), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(161), // ||=, reduce: EqneTerm
			reduce(161), // ^^=, reduce: EqneTerm
			reduce(161), // &&=, reduce: EqneTerm
			reduce(161), // ??=, reduce: EqneTerm
			reduce(161), // ???=, reduce: EqneTerm
			reduce(161), // |=, reduce: EqneTerm
			reduce(161), // &=, reduce: EqneTerm
			reduce(161), // ^=, reduce: EqneTerm
			reduce(161), // <<=, reduce: EqneTerm
			reduce(161), // >>=, reduce: EqneTerm
			reduce(161), // >>>=, reduce: EqneTerm
			reduce(161), // +=, reduce: EqneTerm
			reduce(161), // .=, reduce: EqneTerm
			reduce(161), // -=, reduce: EqneTerm
			reduce(161), // *=, reduce: EqneTerm
			reduce(161), // /=, reduce: EqneTerm
			reduce(161), // //=, reduce: EqneTerm
			reduce(161), // %=, reduce: EqneTerm
			reduce(161), // **=, reduce: EqneTerm
			reduce(161), // ?, reduce: EqneTerm
			nil,         // :
			reduce(161), // ||, reduce: EqneTerm
			reduce(161), // ^^, reduce: EqneTerm
			reduce(161), // &&, reduce: EqneTerm
			reduce(161), // ??, reduce: EqneTerm
			reduce(161), // ???, reduce: EqneTerm
			reduce(161), // =~, reduce: EqneTerm
			reduce(161), // !=~, reduce: EqneTerm
			reduce(161), // ==, reduce: EqneTerm
			reduce(161), // !=, reduce: EqneTerm
			reduce(161), // <=>, reduce: EqneTerm
			shift(2288), // >=
			shift(2289), // <
			shift(2290), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3148
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(162), // ;, reduce: EqneTerm
			reduce(162), // {, reduce: EqneTerm
			reduce(162), // }, reduce: EqneTerm
			reduce(162), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(2287), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(162), // ||=, reduce: EqneTerm
			reduce(162), // ^^=, reduce: EqneTerm
			reduce(162), // &&=, reduce: EqneTerm
			reduce(162), // ??=, reduce: EqneTerm
			reduce(162), // ???=, reduce: EqneTerm
			reduce(162), // |=, reduce: EqneTerm
			reduce(162), // &=, reduce: EqneTerm
			reduce(162), // ^=, reduce: EqneTerm
			reduce(162), // <<=, reduce: EqneTerm
			reduce(162), // >>=, reduce: EqneTerm
			reduce(162), // >>>=, reduce: EqneTerm
			reduce(162), // +=, reduce: EqneTerm
			reduce(162), // .=, reduce: EqneTerm
			reduce(162), // -=, reduce: EqneTerm
			reduce(162), // *=, reduce: EqneTerm
			reduce(162), // /=, reduce: EqneTerm
			reduce(162), // //=, reduce: EqneTerm
			reduce(162), // %=, reduce: EqneTerm
			reduce(162), // **=, reduce: EqneTerm
			reduce(162), // ?, reduce: EqneTerm
			nil,         // :
			reduce(162), // ||, reduce: EqneTerm
			reduce(162), // ^^, reduce: EqneTerm
			reduce(162), // &&, reduce: EqneTerm
			reduce(162), // ??, reduce: EqneTerm
			reduce(162), // ???, reduce: EqneTerm
			reduce(162), // =~, reduce: EqneTerm
			reduce(162), // !=~, reduce: EqneTerm
			reduce(162), // ==, reduce: EqneTerm
			reduce(162), // !=, reduce: EqneTerm
			reduce(162), // <=>, reduce: EqneTerm
			shift(2288), // >=
			shift(2289), // <
			shift(2290), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3149
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(163), // ;, reduce: EqneTerm
			reduce(163), // {, reduce: EqneTerm
			reduce(163), // }, reduce: EqneTerm
			reduce(163), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(2287), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(163), // ||=, reduce: EqneTerm
			reduce(163), // ^^=, reduce: EqneTerm
			reduce(163), // &&=, reduce: EqneTerm
			reduce(163), // ??=, reduce: EqneTerm
			reduce(163), // ???=, reduce: EqneTerm
			reduce(163), // |=, reduce: EqneTerm
			reduce(163), // &=, reduce: EqneTerm
			reduce(163), // ^=, reduce: EqneTerm
			reduce(163), // <<=, reduce: EqneTerm
			reduce(163), // >>=, reduce: EqneTerm
			reduce(163), // >>>=, reduce: EqneTerm
			reduce(163), // +=, reduce: EqneTerm
			reduce(163), // .=, reduce: EqneTerm
			reduce(163), // -=, reduce: EqneTerm
			reduce(163), // *=, reduce: EqneTerm
			reduce(163), // /=, reduce: EqneTerm
			reduce(163), // //=, reduce: EqneTerm
			reduce(163), // %=, reduce: EqneTerm
			reduce(163), // **=, reduce: EqneTerm
			reduce(163), // ?, reduce: EqneTerm
			nil,         // :
			reduce(163), // ||, reduce: EqneTerm
			reduce(163), // ^^, reduce: EqneTerm
			reduce(163), // &&, reduce: EqneTerm
			reduce(163), // ??, reduce: EqneTerm
			reduce(163), // ???, reduce: EqneTerm
			reduce(163), // =~, reduce: EqneTerm
			reduce(163), // !=~, reduce: EqneTerm
			reduce(163), // ==, reduce: EqneTerm
			reduce(163), // !=, reduce: EqneTerm
			reduce(163), // <=>, reduce: EqneTerm
			shift(2288), // >=
			shift(2289), // <
			shift(2290), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3150
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(164), // ;, reduce: EqneTerm
			reduce(164), // {, reduce: EqneTerm
			reduce(164), // }, reduce: EqneTerm
			reduce(164), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(2287), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(164), // ||=, reduce: EqneTerm
			reduce(164), // ^^=, reduce: EqneTerm
			reduce(164), // &&=, reduce: EqneTerm
			reduce(164), // ??=, reduce: EqneTerm
			reduce(164), // ???=, reduce: EqneTerm
			reduce(164), // |=, reduce: EqneTerm
			reduce(164), // &=, reduce: EqneTerm
			reduce(164), // ^=, reduce: EqneTerm
			reduce(164), // <<=, reduce: EqneTerm
			reduce(164), // >>=, reduce: EqneTerm
			reduce(164), // >>>=, reduce: EqneTerm
			reduce(164), // +=, reduce: EqneTerm
			reduce(164), // .=, reduce: EqneTerm
			reduce(164), // -=, reduce: EqneTerm
			reduce(164), // *=, reduce: EqneTerm
			reduce(164), // /=, reduce: EqneTerm
			reduce(164), // //=, reduce: EqneTerm
			reduce(164), // %=, reduce: EqneTerm
			reduce(164), // **=, reduce: EqneTerm
			reduce(164), // ?, reduce: EqneTerm
			nil,         // :
			reduce(164), // ||, reduce: EqneTerm
			reduce(164), // ^^, reduce: EqneTerm
			reduce(164), // &&, reduce: EqneTerm
			reduce(164), // ??, reduce: EqneTerm
			reduce(164), // ???, reduce: EqneTerm
			reduce(164), // =~, reduce: EqneTerm
			reduce(164), // !=~, reduce: EqneTerm
			reduce(164), // ==, reduce: EqneTerm
			reduce(164), // !=, reduce: EqneTerm
			reduce(164), // <=>, reduce: EqneTerm
			shift(2288), // >=
			shift(2289), // <
			shift(2290), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3151
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(165), // ;, reduce: EqneTerm
			reduce(165), // {, reduce: EqneTerm
			reduce(165), // }, reduce: EqneTerm
			reduce(165), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(2287), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(165), // ||=, reduce: EqneTerm
			reduce(165), // ^^=, reduce: EqneTerm
			reduce(165), // &&=, reduce: EqneTerm
			reduce(165), // ??=, reduce: EqneTerm
			reduce(165), // ???=, reduce: EqneTerm
			reduce(165), // |=, reduce: EqneTerm
			reduce(165), // &=, reduce: EqneTerm
			reduce(165), // ^=, reduce: EqneTerm
			reduce(165), // <<=, reduce: EqneTerm
			reduce(165), // >>=, reduce: EqneTerm
			reduce(165), // >>>=, reduce: EqneTerm
			reduce(165), // +=, reduce: EqneTerm
			reduce(165), // .=, reduce: EqneTerm
			reduce(165), // -=, reduce: EqneTerm
			reduce(165), // *=, reduce: EqneTerm
			reduce(165), // /=, reduce: EqneTerm
			reduce(165), // //=, reduce: EqneTerm
			reduce(165), // %=, reduce: EqneTerm
			reduce(165), // **=, reduce: EqneTerm
			reduce(165), // ?, reduce: EqneTerm
			nil,         // :
			reduce(165), // ||, reduce: EqneTerm
			reduce(165), // ^^, reduce: EqneTerm
			reduce(165), // &&, reduce: EqneTerm
			reduce(165), // ??, reduce: EqneTerm
			reduce(165), // ???, reduce: EqneTerm
			reduce(165), // =~, reduce: EqneTerm
			reduce(165), // !=~, reduce: EqneTerm
			reduce(165), // ==, reduce: EqneTerm
			reduce(165), // !=, reduce: EqneTerm
			reduce(165), // <=>, reduce: EqneTerm
			shift(2288), // >=
			shift(2289), // <
			shift(2290), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3152
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(167), // ;, reduce: CmpTerm
			reduce(167), // {, reduce: CmpTerm
			reduce(167), // }, reduce: CmpTerm
			reduce(167), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(167), // >, reduce: CmpTerm
			nil,         // >>
			shift(2291), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(167), // ||=, reduce: CmpTerm
			reduce(167), // ^^=, reduce: CmpTerm
			reduce(167), // &&=, reduce: CmpTerm
			reduce(167), // ??=, reduce: CmpTerm
			reduce(167), // ???=, reduce: CmpTerm
			reduce(167), // |=, reduce: CmpTerm
			reduce(167), // &=, reduce: CmpTerm
			reduce(167), // ^=, reduce: CmpTerm
			reduce(167), // <<=, reduce: CmpTerm
			reduce(167), // >>=, reduce: CmpTerm
			reduce(167), // >>>=, reduce: CmpTerm
			reduce(167), // +=, reduce: CmpTerm
			reduce(167), // .=, reduce: CmpTerm
			reduce(167), // -=, reduce: CmpTerm
			reduce(167), // *=, reduce: CmpTerm
			reduce(167), // /=, reduce: CmpTerm
			reduce(167), // //=, reduce: CmpTerm
			reduce(167), // %=, reduce: CmpTerm
			reduce(167), // **=, reduce: CmpTerm
			reduce(167), // ?, reduce: CmpTerm
			nil,         // :
			reduce(167), // ||, reduce: CmpTerm
			reduce(167), // ^^, reduce: CmpTerm
			reduce(167), // &&, reduce: CmpTerm
			reduce(167), // ??, reduce: CmpTerm
			reduce(167), // ???, reduce: CmpTerm
			reduce(167), // =~, reduce: CmpTerm
			reduce(167), // !=~, reduce: CmpTerm
			reduce(167), // ==, reduce: CmpTerm
			reduce(167), // !=, reduce: CmpTerm
			reduce(167), // <=>, reduce: CmpTerm
			reduce(167), // >=, reduce: CmpTerm
			reduce(167), // <, reduce: CmpTerm
			reduce(167), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3153
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(168), // ;, reduce: CmpTerm
			reduce(168), // {, reduce: CmpTerm
			reduce(168), // }, reduce: CmpTerm
			reduce(168), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(168), // >, reduce: CmpTerm
			nil,         // >>
			shift(2291), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(168), // ||=, reduce: CmpTerm
			reduce(168), // ^^=, reduce: CmpTerm
			reduce(168), // &&=, reduce: CmpTerm
			reduce(168), // ??=, reduce: CmpTerm
			reduce(168), // ???=, reduce: CmpTerm
			reduce(168), // |=, reduce: CmpTerm
			reduce(168), // &=, reduce: CmpTerm
			reduce(168), // ^=, reduce: CmpTerm
			reduce(168), // <<=, reduce: CmpTerm
			reduce(168), // >>=, reduce: CmpTerm
			reduce(168), // >>>=, reduce: CmpTerm
			reduce(168), // +=, reduce: CmpTerm
			reduce(168), // .=, reduce: CmpTerm
			reduce(168), // -=, reduce: CmpTerm
			reduce(168), // *=, reduce: CmpTerm
			reduce(168), // /=, reduce: CmpTerm
			reduce(168), // //=, reduce: CmpTerm
			reduce(168), // %=, reduce: CmpTerm
			reduce(168), // **=, reduce: CmpTerm
			reduce(168), // ?, reduce: CmpTerm
			nil,         // :
			reduce(168), // ||, reduce: CmpTerm
			reduce(168), // ^^, reduce: CmpTerm
			reduce(168), // &&, reduce: CmpTerm
			reduce(168), // ??, reduce: CmpTerm
			reduce(168), // ???, reduce: CmpTerm
			reduce(168), // =~, reduce: CmpTerm
			reduce(168), // !=~, reduce: CmpTerm
			reduce(168), // ==, reduce: CmpTerm
			reduce(168), // !=, reduce: CmpTerm
			reduce(168), // <=>, reduce: CmpTerm
			reduce(168), // >=, reduce: CmpTerm
			reduce(168), // <, reduce: CmpTerm
			reduce(168), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3154
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(169), // ;, reduce: CmpTerm
			reduce(169), // {, reduce: CmpTerm
			reduce(169), // }, reduce: CmpTerm
			reduce(169), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(169), // >, reduce: CmpTerm
			nil,         // >>
			shift(2291), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(169), // ||=, reduce: CmpTerm
			reduce(169), // ^^=, reduce: CmpTerm
			reduce(169), // &&=, reduce: CmpTerm
			reduce(169), // ??=, reduce: CmpTerm
			reduce(169), // ???=, reduce: CmpTerm
			reduce(169), // |=, reduce: CmpTerm
			reduce(169), // &=, reduce: CmpTerm
			reduce(169), // ^=, reduce: CmpTerm
			reduce(169), // <<=, reduce: CmpTerm
			reduce(169), // >>=, reduce: CmpTerm
			reduce(169), // >>>=, reduce: CmpTerm
			reduce(169), // +=, reduce: CmpTerm
			reduce(169), // .=, reduce: CmpTerm
			reduce(169), // -=, reduce: CmpTerm
			reduce(169), // *=, reduce: CmpTerm
			reduce(169), // /=, reduce: CmpTerm
			reduce(169), // //=, reduce: CmpTerm
			reduce(169), // %=, reduce: CmpTerm
			reduce(169), // **=, reduce: CmpTerm
			reduce(169), // ?, reduce: CmpTerm
			nil,         // :
			reduce(169), // ||, reduce: CmpTerm
			reduce(169), // ^^, reduce: CmpTerm
			reduce(169), // &&, reduce: CmpTerm
			reduce(169), // ??, reduce: CmpTerm
			reduce(169), // ???, reduce: CmpTerm
			reduce(169), // =~, reduce: CmpTerm
			reduce(169), // !=~, reduce: CmpTerm
			reduce(169), // ==, reduce: CmpTerm
			reduce(169), // !=, reduce: CmpTerm
			reduce(169), // <=>, reduce: CmpTerm
			reduce(169), // >=, reduce: CmpTerm
			reduce(169), // <, reduce: CmpTerm
			reduce(169), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3155
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(170), // ;, reduce: CmpTerm
			reduce(170), // {, reduce: CmpTerm
			reduce(170), // }, reduce: CmpTerm
			reduce(170), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(170), // >, reduce: CmpTerm
			nil,         // >>
			shift(2291), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(170), // ||=, reduce: CmpTerm
			reduce(170), // ^^=, reduce: CmpTerm
			reduce(170), // &&=, reduce: CmpTerm
			reduce(170), // ??=, reduce: CmpTerm
			reduce(170), // ???=, reduce: CmpTerm
			reduce(170), // |=, reduce: CmpTerm
			reduce(170), // &=, reduce: CmpTerm
			reduce(170), // ^=, reduce: CmpTerm
			reduce(170), // <<=, reduce: CmpTerm
			reduce(170), // >>=, reduce: CmpTerm
			reduce(170), // >>>=, reduce: CmpTerm
			reduce(170), // +=, reduce: CmpTerm
			reduce(170), // .=, reduce: CmpTerm
			reduce(170), // -=, reduce: CmpTerm
			reduce(170), // *=, reduce: CmpTerm
			reduce(170), // /=, reduce: CmpTerm
			reduce(170), // //=, reduce: CmpTerm
			reduce(170), // %=, reduce: CmpTerm
			reduce(170), // **=, reduce: CmpTerm
			reduce(170), // ?, reduce: CmpTerm
			nil,         // :
			reduce(170), // ||, reduce: CmpTerm
			reduce(170), // ^^, reduce: CmpTerm
			reduce(170), // &&, reduce: CmpTerm
			reduce(170), // ??, reduce: CmpTerm
			reduce(170), // ???, reduce: CmpTerm
			reduce(170), // =~, reduce: CmpTerm
			reduce(170), // !=~, reduce: CmpTerm
			reduce(170), // ==, reduce: CmpTerm
			reduce(170), // !=, reduce: CmpTerm
			reduce(170), // <=>, reduce: CmpTerm
			reduce(170), // >=, reduce: CmpTerm
			reduce(170), // <, reduce: CmpTerm
			reduce(170), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3156
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(172), // ;, reduce: BitwiseORTerm
			reduce(172), // {, reduce: BitwiseORTerm
			reduce(172), // }, reduce: BitwiseORTerm
			reduce(172), // =, reduce: BitwiseORTerm
			nil,         // unset
			nil,         // filter
			reduce(172), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(172), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(172), // ||=, reduce: BitwiseORTerm
			reduce(172), // ^^=, reduce: BitwiseORTerm
			reduce(172), // &&=, reduce: BitwiseORTerm
			reduce(172), // ??=, reduce: BitwiseORTerm
			reduce(172), // ???=, reduce: BitwiseORTerm
			reduce(172), // |=, reduce: BitwiseORTerm
			reduce(172), // &=, reduce: BitwiseORTerm
			reduce(172), // ^=, reduce: BitwiseORTerm
			reduce(172), // <<=, reduce: BitwiseORTerm
			reduce(172), // >>=, reduce: BitwiseORTerm
			reduce(172), // >>>=, reduce: BitwiseORTerm
			reduce(172), // +=, reduce: BitwiseORTerm
			reduce(172), // .=, reduce: BitwiseORTerm
			reduce(172), // -=, reduce: BitwiseORTerm
			reduce(172), // *=, reduce: BitwiseORTerm
			reduce(172), // /=, reduce: BitwiseORTerm
			reduce(172), // //=, reduce: BitwiseORTerm
			reduce(172), // %=, reduce: BitwiseORTerm
			reduce(172), // **=, reduce: BitwiseORTerm
			reduce(172), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(172), // ||, reduce: BitwiseORTerm
			reduce(172), // ^^, reduce: BitwiseORTerm
			reduce(172), // &&, reduce: BitwiseORTerm
			reduce(172), // ??, reduce: BitwiseORTerm
			reduce(172), // ???, reduce: BitwiseORTerm
			reduce(172), // =~, reduce: BitwiseORTerm
			reduce(172), // !=~, reduce: BitwiseORTerm
			reduce(172), // ==, reduce: BitwiseORTerm
			reduce(172), // !=, reduce: BitwiseORTerm
			reduce(172), // <=>, reduce: BitwiseORTerm
			reduce(172), // >=, reduce: BitwiseORTerm
			reduce(172), // <, reduce: BitwiseORTerm
			reduce(172), // <=, reduce: BitwiseORTerm
			shift(2292), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3157
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(174), // ;, reduce: BitwiseXORTerm
			reduce(174), // {, reduce: BitwiseXORTerm
			reduce(174), // }, reduce: BitwiseXORTerm
			reduce(174), // =, reduce: BitwiseXORTerm
			nil,         // unset
			nil,         // filter
			reduce(174), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(174), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(174), // ||=, reduce: BitwiseXORTerm
			reduce(174), // ^^=, reduce: BitwiseXORTerm
			reduce(174), // &&=, reduce: BitwiseXORTerm
			reduce(174), // ??=, reduce: BitwiseXORTerm
			reduce(174), // ???=, reduce: BitwiseXORTerm
			reduce(174), // |=, reduce: BitwiseXORTerm
			reduce(174), // &=, reduce: BitwiseXORTerm
			reduce(174), // ^=, reduce: BitwiseXORTerm
			reduce(174), // <<=, reduce: BitwiseXORTerm
			reduce(174), // >>=, reduce: BitwiseXORTerm
			reduce(174), // >>>=, reduce: BitwiseXORTerm
			reduce(174), // +=, reduce: BitwiseXORTerm
			reduce(174), // .=, reduce: BitwiseXORTerm
			reduce(174), // -=, reduce: BitwiseXORTerm
			reduce(174), // *=, reduce: BitwiseXORTerm
			reduce(174), // /=, reduce: BitwiseXORTerm
			reduce(174), // //=, reduce: BitwiseXORTerm
			reduce(174), // %=, reduce: BitwiseXORTerm
			reduce(174), // **=, reduce: BitwiseXORTerm
			reduce(174), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(174), // ||, reduce: BitwiseXORTerm
			reduce(174), // ^^, reduce: BitwiseXORTerm
			reduce(174), // &&, reduce: BitwiseXORTerm
			reduce(174), // ??, reduce: BitwiseXORTerm
			reduce(174), // ???, reduce: BitwiseXORTerm
			reduce(174), // =~, reduce: BitwiseXORTerm
			reduce(174), // !=~, reduce: BitwiseXORTerm
			reduce(174), // ==, reduce: BitwiseXORTerm
			reduce(174), // !=, reduce: BitwiseXORTerm
			reduce(174), // <=>, reduce: BitwiseXORTerm
			reduce(174), // >=, reduce: BitwiseXORTerm
			reduce(174), // <, reduce: BitwiseXORTerm
			reduce(174), // <=, reduce: BitwiseXORTerm
			reduce(174), // ^, reduce: BitwiseXORTerm
			shift(2293), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3158
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(176), // ;, reduce: BitwiseANDTerm
			reduce(176), // {, reduce: BitwiseANDTerm
			reduce(176), // }, reduce: BitwiseANDTerm
			reduce(176), // =, reduce: BitwiseANDTerm
			nil,         // unset
			nil,         // filter
			reduce(176), // >, reduce: BitwiseANDTerm
			shift(2294), // >>
			reduce(176), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(176), // ||=, reduce: BitwiseANDTerm
			reduce(176), // ^^=, reduce: BitwiseANDTerm
			reduce(176), // &&=, reduce: BitwiseANDTerm
			reduce(176), // ??=, reduce: BitwiseANDTerm
			reduce(176), // ???=, reduce: BitwiseANDTerm
			reduce(176), // |=, reduce: BitwiseANDTerm
			reduce(176), // &=, reduce: BitwiseANDTerm
			reduce(176), // ^=, reduce: BitwiseANDTerm
			reduce(176), // <<=, reduce: BitwiseANDTerm
			reduce(176), // >>=, reduce: BitwiseANDTerm
			reduce(176), // >>>=, reduce: BitwiseANDTerm
			reduce(176), // +=, reduce: BitwiseANDTerm
			reduce(176), // .=, reduce: BitwiseANDTerm
			reduce(176), // -=, reduce: BitwiseANDTerm
			reduce(176), // *=, reduce: BitwiseANDTerm
			reduce(176), // /=, reduce: BitwiseANDTerm
			reduce(176), // //=, reduce: BitwiseANDTerm
			reduce(176), // %=, reduce: BitwiseANDTerm
			reduce(176), // **=, reduce: BitwiseANDTerm
			reduce(176), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(176), // ||, reduce: BitwiseANDTerm
			reduce(176), // ^^, reduce: BitwiseANDTerm
			reduce(176), // &&, reduce: BitwiseANDTerm
			reduce(176), // ??, reduce: BitwiseANDTerm
			reduce(176), // ???, reduce: BitwiseANDTerm
			reduce(176), // =~, reduce: BitwiseANDTerm
			reduce(176), // !=~, reduce: BitwiseANDTerm
			reduce(176), // ==, reduce: BitwiseANDTerm
			reduce(176), // !=, reduce: BitwiseANDTerm
			reduce(176), // <=>, reduce: BitwiseANDTerm
			reduce(176), // >=, reduce: BitwiseANDTerm
			reduce(176), // <, reduce: BitwiseANDTerm
			reduce(176), // <=, reduce: BitwiseANDTerm
			reduce(176), // ^, reduce: BitwiseANDTerm
			reduce(176), // &, reduce: BitwiseANDTerm
			shift(2295), // <<
			shift(2296), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3159
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(179), // ;, reduce: BitwiseShiftTerm
			reduce(179), // {, reduce: BitwiseShiftTerm
			reduce(179), // }, reduce: BitwiseShiftTerm
			reduce(179), // =, reduce: BitwiseShiftTerm
			nil,         // unset
			nil,         // filter
			reduce(179), // >, reduce: BitwiseShiftTerm
			reduce(179), // >>, reduce: BitwiseShiftTerm
			reduce(179), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(179), // ||=, reduce: BitwiseShiftTerm
			reduce(179), // ^^=, reduce: BitwiseShiftTerm
			reduce(179), // &&=, reduce: BitwiseShiftTerm
			reduce(179), // ??=, reduce: BitwiseShiftTerm
			reduce(179), // ???=, reduce: BitwiseShiftTerm
			reduce(179), // |=, reduce: BitwiseShiftTerm
			reduce(179), // &=, reduce: BitwiseShiftTerm
			reduce(179), // ^=, reduce: BitwiseShiftTerm
			reduce(179), // <<=, reduce: BitwiseShiftTerm
			reduce(179), // >>=, reduce: BitwiseShiftTerm
			reduce(179), // >>>=, reduce: BitwiseShiftTerm
			reduce(179), // +=, reduce: BitwiseShiftTerm
			reduce(179), // .=, reduce: BitwiseShiftTerm
			reduce(179), // -=, reduce: BitwiseShiftTerm
			reduce(179), // *=, reduce: BitwiseShiftTerm
			reduce(179), // /=, reduce: BitwiseShiftTerm
			reduce(179), // //=, reduce: BitwiseShiftTerm
			reduce(179), // %=, reduce: BitwiseShiftTerm
			reduce(179), // **=, reduce: BitwiseShiftTerm
			reduce(179), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(179), // ||, reduce: BitwiseShiftTerm
			reduce(179), // ^^, reduce: BitwiseShiftTerm
			reduce(179), // &&, reduce: BitwiseShiftTerm
			reduce(179), // ??, reduce: BitwiseShiftTerm
			reduce(179), // ???, reduce: BitwiseShiftTerm
			reduce(179), // =~, reduce: BitwiseShiftTerm
			reduce(179), // !=~, reduce: BitwiseShiftTerm
			reduce(179), // ==, reduce: BitwiseShiftTerm
			reduce(179), // !=, reduce: BitwiseShiftTerm
			reduce(179), // <=>, reduce: BitwiseShiftTerm
			reduce(179), // >=, reduce: BitwiseShiftTerm
			reduce(179), // <, reduce: BitwiseShiftTerm
			reduce(179), // <=, reduce: BitwiseShiftTerm
			reduce(179), // ^, reduce: BitwiseShiftTerm
			reduce(179), // &, reduce: BitwiseShiftTerm
			reduce(179), // <<, reduce: BitwiseShiftTerm
			reduce(179), // >>>, reduce: BitwiseShiftTerm
			shift(2297), // +
			shift(2298), // -
			shift(2299), // .+
			shift(2300), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3160
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(178), // ;, reduce: BitwiseShiftTerm
			reduce(178), // {, reduce: BitwiseShiftTerm
			reduce(178), // }, reduce: BitwiseShiftTerm
			reduce(178), // =, reduce: BitwiseShiftTerm
			nil,         // unset
			nil,         // filter
			reduce(178), // >, reduce: BitwiseShiftTerm
			reduce(178), // >>, reduce: BitwiseShiftTerm
			reduce(178), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(178), // ||=, reduce: BitwiseShiftTerm
			reduce(178), // ^^=, reduce: BitwiseShiftTerm
			reduce(178), // &&=, reduce: BitwiseShiftTerm
			reduce(178), // ??=, reduce: BitwiseShiftTerm
			reduce(178), // ???=, reduce: BitwiseShiftTerm
			reduce(178), // |=, reduce: BitwiseShiftTerm
			reduce(178), // &=, reduce: BitwiseShiftTerm
			reduce(178), // ^=, reduce: BitwiseShiftTerm
			reduce(178), // <<=, reduce: BitwiseShiftTerm
			reduce(178), // >>=, reduce: BitwiseShiftTerm
			reduce(178), // >>>=, reduce: BitwiseShiftTerm
			reduce(178), // +=, reduce: BitwiseShiftTerm
			reduce(178), // .=, reduce: BitwiseShiftTerm
			reduce(178), // -=, reduce: BitwiseShiftTerm
			reduce(178), // *=, reduce: BitwiseShiftTerm
			reduce(178), // /=, reduce: BitwiseShiftTerm
			reduce(178), // //=, reduce: BitwiseShiftTerm
			reduce(178), // %=, reduce: BitwiseShiftTerm
			reduce(178), // **=, reduce: BitwiseShiftTerm
			reduce(178), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(178), // ||, reduce: BitwiseShiftTerm
			reduce(178), // ^^, reduce: BitwiseShiftTerm
			reduce(178), // &&, reduce: BitwiseShiftTerm
			reduce(178), // ??, reduce: BitwiseShiftTerm
			reduce(178), // ???, reduce: BitwiseShiftTerm
			reduce(178), // =~, reduce: BitwiseShiftTerm
			reduce(178), // !=~, reduce: BitwiseShiftTerm
			reduce(178), // ==, reduce: BitwiseShiftTerm
			reduce(178), // !=, reduce: BitwiseShiftTerm
			reduce(178), // <=>, reduce: BitwiseShiftTerm
			reduce(178), // >=, reduce: BitwiseShiftTerm
			reduce(178), // <, reduce: BitwiseShiftTerm
			reduce(178), // <=, reduce: BitwiseShiftTerm
			reduce(178), // ^, reduce: BitwiseShiftTerm
			reduce(178), // &, reduce: BitwiseShiftTerm
			reduce(178), // <<, reduce: BitwiseShiftTerm
			reduce(178), // >>>, reduce: BitwiseShiftTerm
			shift(2297), // +
			shift(2298), // -
			shift(2299), // .+
			shift(2300), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3161
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(180), // ;, reduce: BitwiseShiftTerm
			reduce(180), // {, reduce: BitwiseShiftTerm
			reduce(180), // }, reduce: BitwiseShiftTerm
			reduce(180), // =, reduce: BitwiseShiftTerm
			nil,         // unset
			nil,         // filter
			reduce(180), // >, reduce: BitwiseShiftTerm
			reduce(180), // >>, reduce: BitwiseShiftTerm
			reduce(180), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(180), // ||=, reduce: BitwiseShiftTerm
			reduce(180), // ^^=, reduce: BitwiseShiftTerm
			reduce(180), // &&=, reduce: BitwiseShiftTerm
			reduce(180), // ??=, reduce: BitwiseShiftTerm
			reduce(180), // ???=, reduce: BitwiseShiftTerm
			reduce(180), // |=, reduce: BitwiseShiftTerm
			reduce(180), // &=, reduce: BitwiseShiftTerm
			reduce(180), // ^=, reduce: BitwiseShiftTerm
			reduce(180), // <<=, reduce: BitwiseShiftTerm
			reduce(180), // >>=, reduce: BitwiseShiftTerm
			reduce(180), // >>>=, reduce: BitwiseShiftTerm
			reduce(180), // +=, reduce: BitwiseShiftTerm
			reduce(180), // .=, reduce: BitwiseShiftTerm
			reduce(180), // -=, reduce: BitwiseShiftTerm
			reduce(180), // *=, reduce: BitwiseShiftTerm
			reduce(180), // /=, reduce: BitwiseShiftTerm
			reduce(180), // //=, reduce: BitwiseShiftTerm
			reduce(180), // %=, reduce: BitwiseShiftTerm
			reduce(180), // **=, reduce: BitwiseShiftTerm
			reduce(180), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(180), // ||, reduce: BitwiseShiftTerm
			reduce(180), // ^^, reduce: BitwiseShiftTerm
			reduce(180), // &&, reduce: BitwiseShiftTerm
			reduce(180), // ??, reduce: BitwiseShiftTerm
			reduce(180), // ???, reduce: BitwiseShiftTerm
			reduce(180), // =~, reduce: BitwiseShiftTerm
			reduce(180), // !=~, reduce: BitwiseShiftTerm
			reduce(180), // ==, reduce: BitwiseShiftTerm
			reduce(180), // !=, reduce: BitwiseShiftTerm
			reduce(180), // <=>, reduce: BitwiseShiftTerm
			reduce(180), // >=, reduce: BitwiseShiftTerm
			reduce(180), // <, reduce: BitwiseShiftTerm
			reduce(180), // <=, reduce: BitwiseShiftTerm
			reduce(180), // ^, reduce: BitwiseShiftTerm
			reduce(180), // &, reduce: BitwiseShiftTerm
			reduce(180), // <<, reduce: BitwiseShiftTerm
			reduce(180), // >>>, reduce: BitwiseShiftTerm
			shift(2297), // +
			shift(2298), // -
			shift(2299), // .+
			shift(2300), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3162
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(182), // ;, reduce: AddsubdotTerm
			reduce(182), // {, reduce: AddsubdotTerm
			reduce(182), // }, reduce: AddsubdotTerm
			reduce(182), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(182), // >, reduce: AddsubdotTerm
			reduce(182), // >>, reduce: AddsubdotTerm
			reduce(182), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(182), // ||=, reduce: AddsubdotTerm
			reduce(182), // ^^=, reduce: AddsubdotTerm
			reduce(182), // &&=, reduce: AddsubdotTerm
			reduce(182), // ??=, reduce: AddsubdotTerm
			reduce(182), // ???=, reduce: AddsubdotTerm
			reduce(182), // |=, reduce: AddsubdotTerm
			reduce(182), // &=, reduce: AddsubdotTerm
			reduce(182), // ^=, reduce: AddsubdotTerm
			reduce(182), // <<=, reduce: AddsubdotTerm
			reduce(182), // >>=, reduce: AddsubdotTerm
			reduce(182), // >>>=, reduce: AddsubdotTerm
			reduce(182), // +=, reduce: AddsubdotTerm
			reduce(182), // .=, reduce: AddsubdotTerm
			reduce(182), // -=, reduce: AddsubdotTerm
			reduce(182), // *=, reduce: AddsubdotTerm
			reduce(182), // /=, reduce: AddsubdotTerm
			reduce(182), // //=, reduce: AddsubdotTerm
			reduce(182), // %=, reduce: AddsubdotTerm
			reduce(182), // **=, reduce: AddsubdotTerm
			reduce(182), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(182), // ||, reduce: AddsubdotTerm
			reduce(182), // ^^, reduce: AddsubdotTerm
			reduce(182), // &&, reduce: AddsubdotTerm
			reduce(182), // ??, reduce: AddsubdotTerm
			reduce(182), // ???, reduce: AddsubdotTerm
			reduce(182), // =~, reduce: AddsubdotTerm
			reduce(182), // !=~, reduce: AddsubdotTerm
			reduce(182), // ==, reduce: AddsubdotTerm
			reduce(182), // !=, reduce: AddsubdotTerm
			reduce(182), // <=>, reduce: AddsubdotTerm
			reduce(182), // >=, reduce: AddsubdotTerm
			reduce(182), // <, reduce: AddsubdotTerm
			reduce(182), // <=, reduce: AddsubdotTerm
			reduce(182), // ^, reduce: AddsubdotTerm
			reduce(182), // &, reduce: AddsubdotTerm
			reduce(182), // <<, reduce: AddsubdotTerm
			reduce(182), // >>>, reduce: AddsubdotTerm
			reduce(182), // +, reduce: AddsubdotTerm
			reduce(182), // -, reduce: AddsubdotTerm
			reduce(182), // .+, reduce: AddsubdotTerm
			reduce(182), // .-, reduce: AddsubdotTerm
			shift(2303), // *
			shift(2304), // /
			shift(2305), // //
			shift(2306), // %
			shift(2307), // .*
			shift(2308), // ./
			shift(2309), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3163
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(183), // ;, reduce: AddsubdotTerm
			reduce(183), // {, reduce: AddsubdotTerm
			reduce(183), // }, reduce: AddsubdotTerm
			reduce(183), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(183), // >, reduce: AddsubdotTerm
			reduce(183), // >>, reduce: AddsubdotTerm
			reduce(183), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(183), // ||=, reduce: AddsubdotTerm
			reduce(183), // ^^=, reduce: AddsubdotTerm
			reduce(183), // &&=, reduce: AddsubdotTerm
			reduce(183), // ??=, reduce: AddsubdotTerm
			reduce(183), // ???=, reduce: AddsubdotTerm
			reduce(183), // |=, reduce: AddsubdotTerm
			reduce(183), // &=, reduce: AddsubdotTerm
			reduce(183), // ^=, reduce: AddsubdotTerm
			reduce(183), // <<=, reduce: AddsubdotTerm
			reduce(183), // >>=, reduce: AddsubdotTerm
			reduce(183), // >>>=, reduce: AddsubdotTerm
			reduce(183), // +=, reduce: AddsubdotTerm
			reduce(183), // .=, reduce: AddsubdotTerm
			reduce(183), // -=, reduce: AddsubdotTerm
			reduce(183), // *=, reduce: AddsubdotTerm
			reduce(183), // /=, reduce: AddsubdotTerm
			reduce(183), // //=, reduce: AddsubdotTerm
			reduce(183), // %=, reduce: AddsubdotTerm
			reduce(183), // **=, reduce: AddsubdotTerm
			reduce(183), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(183), // ||, reduce: AddsubdotTerm
			reduce(183), // ^^, reduce: AddsubdotTerm
			reduce(183), // &&, reduce: AddsubdotTerm
			reduce(183), // ??, reduce: AddsubdotTerm
			reduce(183), // ???, reduce: AddsubdotTerm
			reduce(183), // =~, reduce: AddsubdotTerm
			reduce(183), // !=~, reduce: AddsubdotTerm
			reduce(183), // ==, reduce: AddsubdotTerm
			reduce(183), // !=, reduce: AddsubdotTerm
			reduce(183), // <=>, reduce: AddsubdotTerm
			reduce(183), // >=, reduce: AddsubdotTerm
			reduce(183), // <, reduce: AddsubdotTerm
			reduce(183), // <=, reduce: AddsubdotTerm
			reduce(183), // ^, reduce: AddsubdotTerm
			reduce(183), // &, reduce: AddsubdotTerm
			reduce(183), // <<, reduce: AddsubdotTerm
			reduce(183), // >>>, reduce: AddsubdotTerm
			reduce(183), // +, reduce: AddsubdotTerm
			reduce(183), // -, reduce: AddsubdotTerm
			reduce(183), // .+, reduce: AddsubdotTerm
			reduce(183), // .-, reduce: AddsubdotTerm
			shift(2303), // *
			shift(2304), // /
			shift(2305), // //
			shift(2306), // %
			shift(2307), // .*
			shift(2308), // ./
			shift(2309), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3164
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(184), // ;, reduce: AddsubdotTerm
			reduce(184), // {, reduce: AddsubdotTerm
			reduce(184), // }, reduce: AddsubdotTerm
			reduce(184), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(184), // >, reduce: AddsubdotTerm
			reduce(184), // >>, reduce: AddsubdotTerm
			reduce(184), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(184), // ||=, reduce: AddsubdotTerm
			reduce(184), // ^^=, reduce: AddsubdotTerm
			reduce(184), // &&=, reduce: AddsubdotTerm
			reduce(184), // ??=, reduce: AddsubdotTerm
			reduce(184), // ???=, reduce: AddsubdotTerm
			reduce(184), // |=, reduce: AddsubdotTerm
			reduce(184), // &=, reduce: AddsubdotTerm
			reduce(184), // ^=, reduce: AddsubdotTerm
			reduce(184), // <<=, reduce: AddsubdotTerm
			reduce(184), // >>=, reduce: AddsubdotTerm
			reduce(184), // >>>=, reduce: AddsubdotTerm
			reduce(184), // +=, reduce: AddsubdotTerm
			reduce(184), // .=, reduce: AddsubdotTerm
			reduce(184), // -=, reduce: AddsubdotTerm
			reduce(184), // *=, reduce: AddsubdotTerm
			reduce(184), // /=, reduce: AddsubdotTerm
			reduce(184), // //=, reduce: AddsubdotTerm
			reduce(184), // %=, reduce: AddsubdotTerm
			reduce(184), // **=, reduce: AddsubdotTerm
			reduce(184), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(184), // ||, reduce: AddsubdotTerm
			reduce(184), // ^^, reduce: AddsubdotTerm
			reduce(184), // &&, reduce: AddsubdotTerm
			reduce(184), // ??, reduce: AddsubdotTerm
			reduce(184), // ???, reduce: AddsubdotTerm
			reduce(184), // =~, reduce: AddsubdotTerm
			reduce(184), // !=~, reduce: AddsubdotTerm
			reduce(184), // ==, reduce: AddsubdotTerm
			reduce(184), // !=, reduce: AddsubdotTerm
			reduce(184), // <=>, reduce: AddsubdotTerm
			reduce(184), // >=, reduce: AddsubdotTerm
			reduce(184), // <, reduce: AddsubdotTerm
			reduce(184), // <=, reduce: AddsubdotTerm
			reduce(184), // ^, reduce: AddsubdotTerm
			reduce(184), // &, reduce: AddsubdotTerm
			reduce(184), // <<, reduce: AddsubdotTerm
			reduce(184), // >>>, reduce: AddsubdotTerm
			reduce(184), // +, reduce: AddsubdotTerm
			reduce(184), // -, reduce: AddsubdotTerm
			reduce(184), // .+, reduce: AddsubdotTerm
			reduce(184), // .-, reduce: AddsubdotTerm
			shift(2303), // *
			shift(2304), // /
			shift(2305), // //
			shift(2306), // %
			shift(2307), // .*
			shift(2308), // ./
			shift(2309), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3165
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(185), // ;, reduce: AddsubdotTerm
			reduce(185), // {, reduce: AddsubdotTerm
			reduce(185), // }, reduce: AddsubdotTerm
			reduce(185), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(185), // >, reduce: AddsubdotTerm
			reduce(185), // >>, reduce: AddsubdotTerm
			reduce(185), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(185), // ||=, reduce: AddsubdotTerm
			reduce(185), // ^^=, reduce: AddsubdotTerm
			reduce(185), // &&=, reduce: AddsubdotTerm
			reduce(185), // ??=, reduce: AddsubdotTerm
			reduce(185), // ???=, reduce: AddsubdotTerm
			reduce(185), // |=, reduce: AddsubdotTerm
			reduce(185), // &=, reduce: AddsubdotTerm
			reduce(185), // ^=, reduce: AddsubdotTerm
			reduce(185), // <<=, reduce: AddsubdotTerm
			reduce(185), // >>=, reduce: AddsubdotTerm
			reduce(185), // >>>=, reduce: AddsubdotTerm
			reduce(185), // +=, reduce: AddsubdotTerm
			reduce(185), // .=, reduce: AddsubdotTerm
			reduce(185), // -=, reduce: AddsubdotTerm
			reduce(185), // *=, reduce: AddsubdotTerm
			reduce(185), // /=, reduce: AddsubdotTerm
			reduce(185), // //=, reduce: AddsubdotTerm
			reduce(185), // %=, reduce: AddsubdotTerm
			reduce(185), // **=, reduce: AddsubdotTerm
			reduce(185), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(185), // ||, reduce: AddsubdotTerm
			reduce(185), // ^^, reduce: AddsubdotTerm
			reduce(185), // &&, reduce: AddsubdotTerm
			reduce(185), // ??, reduce: AddsubdotTerm
			reduce(185), // ???, reduce: AddsubdotTerm
			reduce(185), // =~, reduce: AddsubdotTerm
			reduce(185), // !=~, reduce: AddsubdotTerm
			reduce(185), // ==, reduce: AddsubdotTerm
			reduce(185), // !=, reduce: AddsubdotTerm
			reduce(185), // <=>, reduce: AddsubdotTerm
			reduce(185), // >=, reduce: AddsubdotTerm
			reduce(185), // <, reduce: AddsubdotTerm
			reduce(185), // <=, reduce: AddsubdotTerm
			reduce(185), // ^, reduce: AddsubdotTerm
			reduce(185), // &, reduce: AddsubdotTerm
			reduce(185), // <<, reduce: AddsubdotTerm
			reduce(185), // >>>, reduce: AddsubdotTerm
			reduce(185), // +, reduce: AddsubdotTerm
			reduce(185), // -, reduce: AddsubdotTerm
			reduce(185), // .+, reduce: AddsubdotTerm
			reduce(185), // .-, reduce: AddsubdotTerm
			shift(2303), // *
			shift(2304), // /
			shift(2305), // //
			shift(2306), // %
			shift(2307), // .*
			shift(2308), // ./
			shift(2309), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3166
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(187), // ;, reduce: MuldivTerm
			reduce(187), // {, reduce: MuldivTerm
			reduce(187), // }, reduce: MuldivTerm
			reduce(187), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(187), // >, reduce: MuldivTerm
			reduce(187), // >>, reduce: MuldivTerm
			reduce(187), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(187), // ||=, reduce: MuldivTerm
			reduce(187), // ^^=, reduce: MuldivTerm
			reduce(187), // &&=, reduce: MuldivTerm
			reduce(187), // ??=, reduce: MuldivTerm
			reduce(187), // ???=, reduce: MuldivTerm
			reduce(187), // |=, reduce: MuldivTerm
			reduce(187), // &=, reduce: MuldivTerm
			reduce(187), // ^=, reduce: MuldivTerm
			reduce(187), // <<=, reduce: MuldivTerm
			reduce(187), // >>=, reduce: MuldivTerm
			reduce(187), // >>>=, reduce: MuldivTerm
			reduce(187), // +=, reduce: MuldivTerm
			reduce(187), // .=, reduce: MuldivTerm
			reduce(187), // -=, reduce: MuldivTerm
			reduce(187), // *=, reduce: MuldivTerm
			reduce(187), // /=, reduce: MuldivTerm
			reduce(187), // //=, reduce: MuldivTerm
			reduce(187), // %=, reduce: MuldivTerm
			reduce(187), // **=, reduce: MuldivTerm
			reduce(187), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(187), // ||, reduce: MuldivTerm
			reduce(187), // ^^, reduce: MuldivTerm
			reduce(187), // &&, reduce: MuldivTerm
			reduce(187), // ??, reduce: MuldivTerm
			reduce(187), // ???, reduce: MuldivTerm
			reduce(187), // =~, reduce: MuldivTerm
			reduce(187), // !=~, reduce: MuldivTerm
			reduce(187), // ==, reduce: MuldivTerm
			reduce(187), // !=, reduce: MuldivTerm
			reduce(187), // <=>, reduce: MuldivTerm
			reduce(187), // >=, reduce: MuldivTerm
			reduce(187), // <, reduce: MuldivTerm
			reduce(187), // <=, reduce: MuldivTerm
			reduce(187), // ^, reduce: MuldivTerm
			reduce(187), // &, reduce: MuldivTerm
			reduce(187), // <<, reduce: MuldivTerm
			reduce(187), // >>>, reduce: MuldivTerm
			reduce(187), // +, reduce: MuldivTerm
			reduce(187), // -, reduce: MuldivTerm
			reduce(187), // .+, reduce: MuldivTerm
			reduce(187), // .-, reduce: MuldivTerm
			reduce(187), // *, reduce: MuldivTerm
			reduce(187), // /, reduce: MuldivTerm
			reduce(187), // //, reduce: MuldivTerm
			reduce(187), // %, reduce: MuldivTerm
			reduce(187), // .*, reduce: MuldivTerm
			reduce(187), // ./, reduce: MuldivTerm
			reduce(187), // .//, reduce: MuldivTerm
			shift(2313), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3167
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(188), // ;, reduce: MuldivTerm
			reduce(188), // {, reduce: MuldivTerm
			reduce(188), // }, reduce: MuldivTerm
			reduce(188), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(188), // >, reduce: MuldivTerm
			reduce(188), // >>, reduce: MuldivTerm
			reduce(188), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(188), // ||=, reduce: MuldivTerm
			reduce(188), // ^^=, reduce: MuldivTerm
			reduce(188), // &&=, reduce: MuldivTerm
			reduce(188), // ??=, reduce: MuldivTerm
			reduce(188), // ???=, reduce: MuldivTerm
			reduce(188), // |=, reduce: MuldivTerm
			reduce(188), // &=, reduce: MuldivTerm
			reduce(188), // ^=, reduce: MuldivTerm
			reduce(188), // <<=, reduce: MuldivTerm
			reduce(188), // >>=, reduce: MuldivTerm
			reduce(188), // >>>=, reduce: MuldivTerm
			reduce(188), // +=, reduce: MuldivTerm
			reduce(188), // .=, reduce: MuldivTerm
			reduce(188), // -=, reduce: MuldivTerm
			reduce(188), // *=, reduce: MuldivTerm
			reduce(188), // /=, reduce: MuldivTerm
			reduce(188), // //=, reduce: MuldivTerm
			reduce(188), // %=, reduce: MuldivTerm
			reduce(188), // **=, reduce: MuldivTerm
			reduce(188), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(188), // ||, reduce: MuldivTerm
			reduce(188), // ^^, reduce: MuldivTerm
			reduce(188), // &&, reduce: MuldivTerm
			reduce(188), // ??, reduce: MuldivTerm
			reduce(188), // ???, reduce: MuldivTerm
			reduce(188), // =~, reduce: MuldivTerm
			reduce(188), // !=~, reduce: MuldivTerm
			reduce(188), // ==, reduce: MuldivTerm
			reduce(188), // !=, reduce: MuldivTerm
			reduce(188), // <=>, reduce: MuldivTerm
			reduce(188), // >=, reduce: MuldivTerm
			reduce(188), // <, reduce: MuldivTerm
			reduce(188), // <=, reduce: MuldivTerm
			reduce(188), // ^, reduce: MuldivTerm
			reduce(188), // &, reduce: MuldivTerm
			reduce(188), // <<, reduce: MuldivTerm
			reduce(188), // >>>, reduce: MuldivTerm
			reduce(188), // +, reduce: MuldivTerm
			reduce(188), // -, reduce: MuldivTerm
			reduce(188), // .+, reduce: MuldivTerm
			reduce(188), // .-, reduce: MuldivTerm
			reduce(188), // *, reduce: MuldivTerm
			reduce(188), // /, reduce: MuldivTerm
			reduce(188), // //, reduce: MuldivTerm
			reduce(188), // %, reduce: MuldivTerm
			reduce(188), // .*, reduce: MuldivTerm
			reduce(188), // ./, reduce: MuldivTerm
			reduce(188), // .//, reduce: MuldivTerm
			shift(2313), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3168
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(189), // ;, reduce: MuldivTerm
			reduce(189), // {, reduce: MuldivTerm
			reduce(189), // }, reduce: MuldivTerm
			reduce(189), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(189), // >, reduce: MuldivTerm
			reduce(189), // >>, reduce: MuldivTerm
			reduce(189), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(189), // ||=, reduce: MuldivTerm
			reduce(189), // ^^=, reduce: MuldivTerm
			reduce(189), // &&=, reduce: MuldivTerm
			reduce(189), // ??=, reduce: MuldivTerm
			reduce(189), // ???=, reduce: MuldivTerm
			reduce(189), // |=, reduce: MuldivTerm
			reduce(189), // &=, reduce: MuldivTerm
			reduce(189), // ^=, reduce: MuldivTerm
			reduce(189), // <<=, reduce: MuldivTerm
			reduce(189), // >>=, reduce: MuldivTerm
			reduce(189), // >>>=, reduce: MuldivTerm
			reduce(189), // +=, reduce: MuldivTerm
			reduce(189), // .=, reduce: MuldivTerm
			reduce(189), // -=, reduce: MuldivTerm
			reduce(189), // *=, reduce: MuldivTerm
			reduce(189), // /=, reduce: MuldivTerm
			reduce(189), // //=, reduce: MuldivTerm
			reduce(189), // %=, reduce: MuldivTerm
			reduce(189), // **=, reduce: MuldivTerm
			reduce(189), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(189), // ||, reduce: MuldivTerm
			reduce(189), // ^^, reduce: MuldivTerm
			reduce(189), // &&, reduce: MuldivTerm
			reduce(189), // ??, reduce: MuldivTerm
			reduce(189), // ???, reduce: MuldivTerm
			reduce(189), // =~, reduce: MuldivTerm
			reduce(189), // !=~, reduce: MuldivTerm
			reduce(189), // ==, reduce: MuldivTerm
			reduce(189), // !=, reduce: MuldivTerm
			reduce(189), // <=>, reduce: MuldivTerm
			reduce(189), // >=, reduce: MuldivTerm
			reduce(189), // <, reduce: MuldivTerm
			reduce(189), // <=, reduce: MuldivTerm
			reduce(189), // ^, reduce: MuldivTerm
			reduce(189), // &, reduce: MuldivTerm
			reduce(189), // <<, reduce: MuldivTerm
			reduce(189), // >>>, reduce: MuldivTerm
			reduce(189), // +, reduce: MuldivTerm
			reduce(189), // -, reduce: MuldivTerm
			reduce(189), // .+, reduce: MuldivTerm
			reduce(189), // .-, reduce: MuldivTerm
			reduce(189), // *, reduce: MuldivTerm
			reduce(189), // /, reduce: MuldivTerm
			reduce(189), // //, reduce: MuldivTerm
			reduce(189), // %, reduce: MuldivTerm
			reduce(189), // .*, reduce: MuldivTerm
			reduce(189), // ./, reduce: MuldivTerm
			reduce(189), // .//, reduce: MuldivTerm
			shift(2313), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3169
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(190), // ;, reduce: MuldivTerm
			reduce(190), // {, reduce: MuldivTerm
			reduce(190), // }, reduce: MuldivTerm
			reduce(190), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(190), // >, reduce: MuldivTerm
			reduce(190), // >>, reduce: MuldivTerm
			reduce(190), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(190), // ||=, reduce: MuldivTerm
			reduce(190), // ^^=, reduce: MuldivTerm
			reduce(190), // &&=, reduce: MuldivTerm
			reduce(190), // ??=, reduce: MuldivTerm
			reduce(190), // ???=, reduce: MuldivTerm
			reduce(190), // |=, reduce: MuldivTerm
			reduce(190), // &=, reduce: MuldivTerm
			reduce(190), // ^=, reduce: MuldivTerm
			reduce(190), // <<=, reduce: MuldivTerm
			reduce(190), // >>=, reduce: MuldivTerm
			reduce(190), // >>>=, reduce: MuldivTerm
			reduce(190), // +=, reduce: MuldivTerm
			reduce(190), // .=, reduce: MuldivTerm
			reduce(190), // -=, reduce: MuldivTerm
			reduce(190), // *=, reduce: MuldivTerm
			reduce(190), // /=, reduce: MuldivTerm
			reduce(190), // //=, reduce: MuldivTerm
			reduce(190), // %=, reduce: MuldivTerm
			reduce(190), // **=, reduce: MuldivTerm
			reduce(190), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(190), // ||, reduce: MuldivTerm
			reduce(190), // ^^, reduce: MuldivTerm
			reduce(190), // &&, reduce: MuldivTerm
			reduce(190), // ??, reduce: MuldivTerm
			reduce(190), // ???, reduce: MuldivTerm
			reduce(190), // =~, reduce: MuldivTerm
			reduce(190), // !=~, reduce: MuldivTerm
			reduce(190), // ==, reduce: MuldivTerm
			reduce(190), // !=, reduce: MuldivTerm
			reduce(190), // <=>, reduce: MuldivTerm
			reduce(190), // >=, reduce: MuldivTerm
			reduce(190), // <, reduce: MuldivTerm
			reduce(190), // <=, reduce: MuldivTerm
			reduce(190), // ^, reduce: MuldivTerm
			reduce(190), // &, reduce: MuldivTerm
			reduce(190), // <<, reduce: MuldivTerm
			reduce(190), // >>>, reduce: MuldivTerm
			reduce(190), // +, reduce: MuldivTerm
			reduce(190), // -, reduce: MuldivTerm
			reduce(190), // .+, reduce: MuldivTerm
			reduce(190), // .-, reduce: MuldivTerm
			reduce(190), // *, reduce: MuldivTerm
			reduce(190), // /, reduce: MuldivTerm
			reduce(190), // //, reduce: MuldivTerm
			reduce(190), // %, reduce: MuldivTerm
			reduce(190), // .*, reduce: MuldivTerm
			reduce(190), // ./, reduce: MuldivTerm
			reduce(190), // .//, reduce: MuldivTerm
			shift(2313), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3170
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(191), // ;, reduce: MuldivTerm
			reduce(191), // {, reduce: MuldivTerm
			reduce(191), // }, reduce: MuldivTerm
			reduce(191), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(191), // >, reduce: MuldivTerm
			reduce(191), // >>, reduce: MuldivTerm
			reduce(191), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(191), // ||=, reduce: MuldivTerm
			reduce(191), // ^^=, reduce: MuldivTerm
			reduce(191), // &&=, reduce: MuldivTerm
			reduce(191), // ??=, reduce: MuldivTerm
			reduce(191), // ???=, reduce: MuldivTerm
			reduce(191), // |=, reduce: MuldivTerm
			reduce(191), // &=, reduce: MuldivTerm
			reduce(191), // ^=, reduce: MuldivTerm
			reduce(191), // <<=, reduce: MuldivTerm
			reduce(191), // >>=, reduce: MuldivTerm
			reduce(191), // >>>=, reduce: MuldivTerm
			reduce(191), // +=, reduce: MuldivTerm
			reduce(191), // .=, reduce: MuldivTerm
			reduce(191), // -=, reduce: MuldivTerm
			reduce(191), // *=, reduce: MuldivTerm
			reduce(191), // /=, reduce: MuldivTerm
			reduce(191), // //=, reduce: MuldivTerm
			reduce(191), // %=, reduce: MuldivTerm
			reduce(191), // **=, reduce: MuldivTerm
			reduce(191), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(191), // ||, reduce: MuldivTerm
			reduce(191), // ^^, reduce: MuldivTerm
			reduce(191), // &&, reduce: MuldivTerm
			reduce(191), // ??, reduce: MuldivTerm
			reduce(191), // ???, reduce: MuldivTerm
			reduce(191), // =~, reduce: MuldivTerm
			reduce(191), // !=~, reduce: MuldivTerm
			reduce(191), // ==, reduce: MuldivTerm
			reduce(191), // !=, reduce: MuldivTerm
			reduce(191), // <=>, reduce: MuldivTerm
			reduce(191), // >=, reduce: MuldivTerm
			reduce(191), // <, reduce: MuldivTerm
			reduce(191), // <=, reduce: MuldivTerm
			reduce(191), // ^, reduce: MuldivTerm
			reduce(191), // &, reduce: MuldivTerm
			reduce(191), // <<, reduce: MuldivTerm
			reduce(191), // >>>, reduce: MuldivTerm
			reduce(191), // +, reduce: MuldivTerm
			reduce(191), // -, reduce: MuldivTerm
			reduce(191), // .+, reduce: MuldivTerm
			reduce(191), // .-, reduce: MuldivTerm
			reduce(191), // *, reduce: MuldivTerm
			reduce(191), // /, reduce: MuldivTerm
			reduce(191), // //, reduce: MuldivTerm
			reduce(191), // %, reduce: MuldivTerm
			reduce(191), // .*, reduce: MuldivTerm
			reduce(191), // ./, reduce: MuldivTerm
			reduce(191), // .//, reduce: MuldivTerm
			shift(2313), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3171
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(192), // ;, reduce: MuldivTerm
			reduce(192), // {, reduce: MuldivTerm
			reduce(192), // }, reduce: MuldivTerm
			reduce(192), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(192), // >, reduce: MuldivTerm
			reduce(192), // >>, reduce: MuldivTerm
			reduce(192), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(192), // ||=, reduce: MuldivTerm
			reduce(192), // ^^=, reduce: MuldivTerm
			reduce(192), // &&=, reduce: MuldivTerm
			reduce(192), // ??=, reduce: MuldivTerm
			reduce(192), // ???=, reduce: MuldivTerm
			reduce(192), // |=, reduce: MuldivTerm
			reduce(192), // &=, reduce: MuldivTerm
			reduce(192), // ^=, reduce: MuldivTerm
			reduce(192), // <<=, reduce: MuldivTerm
			reduce(192), // >>=, reduce: MuldivTerm
			reduce(192), // >>>=, reduce: MuldivTerm
			reduce(192), // +=, reduce: MuldivTerm
			reduce(192), // .=, reduce: MuldivTerm
			reduce(192), // -=, reduce: MuldivTerm
			reduce(192), // *=, reduce: MuldivTerm
			reduce(192), // /=, reduce: MuldivTerm
			reduce(192), // //=, reduce: MuldivTerm
			reduce(192), // %=, reduce: MuldivTerm
			reduce(192), // **=, reduce: MuldivTerm
			reduce(192), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(192), // ||, reduce: MuldivTerm
			reduce(192), // ^^, reduce: MuldivTerm
			reduce(192), // &&, reduce: MuldivTerm
			reduce(192), // ??, reduce: MuldivTerm
			reduce(192), // ???, reduce: MuldivTerm
			reduce(192), // =~, reduce: MuldivTerm
			reduce(192), // !=~, reduce: MuldivTerm
			reduce(192), // ==, reduce: MuldivTerm
			reduce(192), // !=, reduce: MuldivTerm
			reduce(192), // <=>, reduce: MuldivTerm
			reduce(192), // >=, reduce: MuldivTerm
			reduce(192), // <, reduce: MuldivTerm
			reduce(192), // <=, reduce: MuldivTerm
			reduce(192), // ^, reduce: MuldivTerm
			reduce(192), // &, reduce: MuldivTerm
			reduce(192), // <<, reduce: MuldivTerm
			reduce(192), // >>>, reduce: MuldivTerm
			reduce(192), // +, reduce: MuldivTerm
			reduce(192), // -, reduce: MuldivTerm
			reduce(192), // .+, reduce: MuldivTerm
			reduce(192), // .-, reduce: MuldivTerm
			reduce(192), // *, reduce: MuldivTerm
			reduce(192), // /, reduce: MuldivTerm
			reduce(192), // //, reduce: MuldivTerm
			reduce(192), // %, reduce: MuldivTerm
			reduce(192), // .*, reduce: MuldivTerm
			reduce(192), // ./, reduce: MuldivTerm
			reduce(192), // .//, reduce: MuldivTerm
			shift(2313), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3172
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(193), // ;, reduce: MuldivTerm
			reduce(193), // {, reduce: MuldivTerm
			reduce(193), // }, reduce: MuldivTerm
			reduce(193), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(193), // >, reduce: MuldivTerm
			reduce(193), // >>, reduce: MuldivTerm
			reduce(193), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(193), // ||=, reduce: MuldivTerm
			reduce(193), // ^^=, reduce: MuldivTerm
			reduce(193), // &&=, reduce: MuldivTerm
			reduce(193), // ??=, reduce: MuldivTerm
			reduce(193), // ???=, reduce: MuldivTerm
			reduce(193), // |=, reduce: MuldivTerm
			reduce(193), // &=, reduce: MuldivTerm
			reduce(193), // ^=, reduce: MuldivTerm
			reduce(193), // <<=, reduce: MuldivTerm
			reduce(193), // >>=, reduce: MuldivTerm
			reduce(193), // >>>=, reduce: MuldivTerm
			reduce(193), // +=, reduce: MuldivTerm
			reduce(193), // .=, reduce: MuldivTerm
			reduce(193), // -=, reduce: MuldivTerm
			reduce(193), // *=, reduce: MuldivTerm
			reduce(193), // /=, reduce: MuldivTerm
			reduce(193), // //=, reduce: MuldivTerm
			reduce(193), // %=, reduce: MuldivTerm
			reduce(193), // **=, reduce: MuldivTerm
			reduce(193), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(193), // ||, reduce: MuldivTerm
			reduce(193), // ^^, reduce: MuldivTerm
			reduce(193), // &&, reduce: MuldivTerm
			reduce(193), // ??, reduce: MuldivTerm
			reduce(193), // ???, reduce: MuldivTerm
			reduce(193), // =~, reduce: MuldivTerm
			reduce(193), // !=~, reduce: MuldivTerm
			reduce(193), // ==, reduce: MuldivTerm
			reduce(193), // !=, reduce: MuldivTerm
			reduce(193), // <=>, reduce: MuldivTerm
			reduce(193), // >=, reduce: MuldivTerm
			reduce(193), // <, reduce: MuldivTerm
			reduce(193), // <=, reduce: MuldivTerm
			reduce(193), // ^, reduce: MuldivTerm
			reduce(193), // &, reduce: MuldivTerm
			reduce(193), // <<, reduce: MuldivTerm
			reduce(193), // >>>, reduce: MuldivTerm
			reduce(193), // +, reduce: MuldivTerm
			reduce(193), // -, reduce: MuldivTerm
			reduce(193), // .+, reduce: MuldivTerm
			reduce(193), // .-, reduce: MuldivTerm
			reduce(193), // *, reduce: MuldivTerm
			reduce(193), // /, reduce: MuldivTerm
			reduce(193), // //, reduce: MuldivTerm
			reduce(193), // %, reduce: MuldivTerm
			reduce(193), // .*, reduce: MuldivTerm
			reduce(193), // ./, reduce: MuldivTerm
			reduce(193), // .//, reduce: MuldivTerm
			shift(2313), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3173
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(195), // ;, reduce: DotTerm
			reduce(195), // {, reduce: DotTerm
			reduce(195), // }, reduce: DotTerm
			reduce(195), // =, reduce: DotTerm
			nil,         // unset
			nil,         // filter
			reduce(195), // >, reduce: DotTerm
			reduce(195), // >>, reduce: DotTerm
			reduce(195), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(195), // ||=, reduce: DotTerm
			reduce(195), // ^^=, reduce: DotTerm
			reduce(195), // &&=, reduce: DotTerm
			reduce(195), // ??=, reduce: DotTerm
			reduce(195), // ???=, reduce: DotTerm
			reduce(195), // |=, reduce: DotTerm
			reduce(195), // &=, reduce: DotTerm
			reduce(195), // ^=, reduce: DotTerm
			reduce(195), // <<=, reduce: DotTerm
			reduce(195), // >>=, reduce: DotTerm
			reduce(195), // >>>=, reduce: DotTerm
			reduce(195), // +=, reduce: DotTerm
			reduce(195), // .=, reduce: DotTerm
			reduce(195), // -=, reduce: DotTerm
			reduce(195), // *=, reduce: DotTerm
			reduce(195), // /=, reduce: DotTerm
			reduce(195), // //=, reduce: DotTerm
			reduce(195), // %=, reduce: DotTerm
			reduce(195), // **=, reduce: DotTerm
			reduce(195), // ?, reduce: DotTerm
			nil,         // :
			reduce(195), // ||, reduce: DotTerm
			reduce(195), // ^^, reduce: DotTerm
			reduce(195), // &&, reduce: DotTerm
			reduce(195), // ??, reduce: DotTerm
			reduce(195), // ???, reduce: DotTerm
			reduce(195), // =~, reduce: DotTerm
			reduce(195), // !=~, reduce: DotTerm
			reduce(195), // ==, reduce: DotTerm
			reduce(195), // !=, reduce: DotTerm
			reduce(195), // <=>, reduce: DotTerm
			reduce(195), // >=, reduce: DotTerm
			reduce(195), // <, reduce: DotTerm
			reduce(195), // <=, reduce: DotTerm
			reduce(195), // ^, reduce: DotTerm
			reduce(195), // &, reduce: DotTerm
			reduce(195), // <<, reduce: DotTerm
			reduce(195), // >>>, reduce: DotTerm
			reduce(195), // +, reduce: DotTerm
			reduce(195), // -, reduce: DotTerm
			reduce(195), // .+, reduce: DotTerm
			reduce(195), // .-, reduce: DotTerm
			reduce(195), // *, reduce: DotTerm
			reduce(195), // /, reduce: DotTerm
			reduce(195), // //, reduce: DotTerm
			reduce(195), // %, reduce: DotTerm
			reduce(195), // .*, reduce: DotTerm
			reduce(195), // ./, reduce: DotTerm
			reduce(195), // .//, reduce: DotTerm
			reduce(195), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3174
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3175
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3176
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(204), // ;, reduce: PowTerm
			reduce(204), // {, reduce: PowTerm
			reduce(204), // }, reduce: PowTerm
			reduce(204), // =, reduce: PowTerm
			nil,         // unset
			nil,         // filter
			reduce(204), // >, reduce: PowTerm
			reduce(204), // >>, reduce: PowTerm
			reduce(204), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(204), // ||=, reduce: PowTerm
			reduce(204), // ^^=, reduce: PowTerm
			reduce(204), // &&=, reduce: PowTerm
			reduce(204), // ??=, reduce: PowTerm
			reduce(204), // ???=, reduce: PowTerm
			reduce(204), // |=, reduce: PowTerm
			reduce(204), // &=, reduce: PowTerm
			reduce(204), // ^=, reduce: PowTerm
			reduce(204), // <<=, reduce: PowTerm
			reduce(204), // >>=, reduce: PowTerm
			reduce(204), // >>>=, reduce: PowTerm
			reduce(204), // +=, reduce: PowTerm
			reduce(204), // .=, reduce: PowTerm
			reduce(204), // -=, reduce: PowTerm
			reduce(204), // *=, reduce: PowTerm
			reduce(204), // /=, reduce: PowTerm
			reduce(204), // //=, reduce: PowTerm
			reduce(204), // %=, reduce: PowTerm
			reduce(204), // **=, reduce: PowTerm
			reduce(204), // ?, reduce: PowTerm
			nil,         // :
			reduce(204), // ||, reduce: PowTerm
			reduce(204), // ^^, reduce: PowTerm
			reduce(204), // &&, reduce: PowTerm
			reduce(204), // ??, reduce: PowTerm
			reduce(204), // ???, reduce: PowTerm
			reduce(204), // =~, reduce: PowTerm
			reduce(204), // !=~, reduce: PowTerm
			reduce(204), // ==, reduce: PowTerm
			reduce(204), // !=, reduce: PowTerm
			reduce(204), // <=>, reduce: PowTerm
			reduce(204), // >=, reduce: PowTerm
			reduce(204), // <, reduce: PowTerm
			reduce(204), // <=, reduce: PowTerm
			reduce(204), // ^, reduce: PowTerm
			reduce(204), // &, reduce: PowTerm
			reduce(204), // <<, reduce: PowTerm
			reduce(204), // >>>, reduce: PowTerm
			reduce(204), // +, reduce: PowTerm
			reduce(204), // -, reduce: PowTerm
			reduce(204), // .+, reduce: PowTerm
			reduce(204), // .-, reduce: PowTerm
			reduce(204), // *, reduce: PowTerm
			reduce(204), // /, reduce: PowTerm
			reduce(204), // //, reduce: PowTerm
			reduce(204), // %, reduce: PowTerm
			reduce(204), // .*, reduce: PowTerm
			reduce(204), // ./, reduce: PowTerm
			reduce(204), // .//, reduce: PowTerm
			reduce(204), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3177
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3912), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3913), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3178
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(3915), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3179
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3916), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3180
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3917), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3181
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(229), // ;, reduce: ArrayLiteral
			reduce(229), // {, reduce: ArrayLiteral
			reduce(229), // }, reduce: ArrayLiteral
			reduce(229), // =, reduce: ArrayLiteral
			nil,         // unset
			nil,         // filter
			reduce(229), // >, reduce: ArrayLiteral
			reduce(229), // >>, reduce: ArrayLiteral
			reduce(229), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(229), // ||=, reduce: ArrayLiteral
			reduce(229), // ^^=, reduce: ArrayLiteral
			reduce(229), // &&=, reduce: ArrayLiteral
			reduce(229), // ??=, reduce: ArrayLiteral
			reduce(229), // ???=, reduce: ArrayLiteral
			reduce(229), // |=, reduce: ArrayLiteral
			reduce(229), // &=, reduce: ArrayLiteral
			reduce(229), // ^=, reduce: ArrayLiteral
			reduce(229), // <<=, reduce: ArrayLiteral
			reduce(229), // >>=, reduce: ArrayLiteral
			reduce(229), // >>>=, reduce: ArrayLiteral
			reduce(229), // +=, reduce: ArrayLiteral
			reduce(229), // .=, reduce: ArrayLiteral
			reduce(229), // -=, reduce: ArrayLiteral
			reduce(229), // *=, reduce: ArrayLiteral
			reduce(229), // /=, reduce: ArrayLiteral
			reduce(229), // //=, reduce: ArrayLiteral
			reduce(229), // %=, reduce: ArrayLiteral
			reduce(229), // **=, reduce: ArrayLiteral
			reduce(229), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(229), // ||, reduce: ArrayLiteral
			reduce(229), // ^^, reduce: ArrayLiteral
			reduce(229), // &&, reduce: ArrayLiteral
			reduce(229), // ??, reduce: ArrayLiteral
			reduce(229), // ???, reduce: ArrayLiteral
			reduce(229), // =~, reduce: ArrayLiteral
			reduce(229), // !=~, reduce: ArrayLiteral
			reduce(229), // ==, reduce: ArrayLiteral
			reduce(229), // !=, reduce: ArrayLiteral
			reduce(229), // <=>, reduce: ArrayLiteral
			reduce(229), // >=, reduce: ArrayLiteral
			reduce(229), // <, reduce: ArrayLiteral
			reduce(229), // <=, reduce: ArrayLiteral
			reduce(229), // ^, reduce: ArrayLiteral
			reduce(229), // &, reduce: ArrayLiteral
			reduce(229), // <<, reduce: ArrayLiteral
			reduce(229), // >>>, reduce: ArrayLiteral
			reduce(229), // +, reduce: ArrayLiteral
			reduce(229), // -, reduce: ArrayLiteral
			reduce(229), // .+, reduce: ArrayLiteral
			reduce(229), // .-, reduce: ArrayLiteral
			reduce(229), // *, reduce: ArrayLiteral
			reduce(229), // /, reduce: ArrayLiteral
			reduce(229), // //, reduce: ArrayLiteral
			reduce(229), // %, reduce: ArrayLiteral
			reduce(229), // .*, reduce: ArrayLiteral
			reduce(229), // ./, reduce: ArrayLiteral
			reduce(229), // .//, reduce: ArrayLiteral
			reduce(229), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(229), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(229), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(229), // [[, reduce: ArrayLiteral
			reduce(229), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3182
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(255), // ;, reduce: ENV
			reduce(255), // {, reduce: ENV
			reduce(255), // }, reduce: ENV
			reduce(255), // =, reduce: ENV
			nil,         // unset
			nil,         // filter
			reduce(255), // >, reduce: ENV
			reduce(255), // >>, reduce: ENV
			reduce(255), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(255), // ||=, reduce: ENV
			reduce(255), // ^^=, reduce: ENV
			reduce(255), // &&=, reduce: ENV
			reduce(255), // ??=, reduce: ENV
			reduce(255), // ???=, reduce: ENV
			reduce(255), // |=, reduce: ENV
			reduce(255), // &=, reduce: ENV
			reduce(255), // ^=, reduce: ENV
			reduce(255), // <<=, reduce: ENV
			reduce(255), // >>=, reduce: ENV
			reduce(255), // >>>=, reduce: ENV
			reduce(255), // +=, reduce: ENV
			reduce(255), // .=, reduce: ENV
			reduce(255), // -=, reduce: ENV
			reduce(255), // *=, reduce: ENV
			reduce(255), // /=, reduce: ENV
			reduce(255), // //=, reduce: ENV
			reduce(255), // %=, reduce: ENV
			reduce(255), // **=, reduce: ENV
			reduce(255), // ?, reduce: ENV
			nil,         // :
			reduce(255), // ||, reduce: ENV
			reduce(255), // ^^, reduce: ENV
			reduce(255), // &&, reduce: ENV
			reduce(255), // ??, reduce: ENV
			reduce(255), // ???, reduce: ENV
			reduce(255), // =~, reduce: ENV
			reduce(255), // !=~, reduce: ENV
			reduce(255), // ==, reduce: ENV
			reduce(255), // !=, reduce: ENV
			reduce(255), // <=>, reduce: ENV
			reduce(255), // >=, reduce: ENV
			reduce(255), // <, reduce: ENV
			reduce(255), // <=, reduce: ENV
			reduce(255), // ^, reduce: ENV
			reduce(255), // &, reduce: ENV
			reduce(255), // <<, reduce: ENV
			reduce(255), // >>>, reduce: ENV
			reduce(255), // +, reduce: ENV
			reduce(255), // -, reduce: ENV
			reduce(255), // .+, reduce: ENV
			reduce(255), // .-, reduce: ENV
			reduce(255), // *, reduce: ENV
			reduce(255), // /, reduce: ENV
			reduce(255), // //, reduce: ENV
			reduce(255), // %, reduce: ENV
			reduce(255), // .*, reduce: ENV
			reduce(255), // ./, reduce: ENV
			reduce(255), // .//, reduce: ENV
			reduce(255), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(255), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(255), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(255), // [[, reduce: ENV
			reduce(255), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3183
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3918), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3184
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3919), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3185
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(268), // ;, reduce: FunctionCallsite
			reduce(268), // {, reduce: FunctionCallsite
			reduce(268), // }, reduce: FunctionCallsite
			reduce(268), // =, reduce: FunctionCallsite
			nil,         // unset
			nil,         // filter
			reduce(268), // >, reduce: FunctionCallsite
			reduce(268), // >>, reduce: FunctionCallsite
			reduce(268), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(268), // ||=, reduce: FunctionCallsite
			reduce(268), // ^^=, reduce: FunctionCallsite
			reduce(268), // &&=, reduce: FunctionCallsite
			reduce(268), // ??=, reduce: FunctionCallsite
			reduce(268), // ???=, reduce: FunctionCallsite
			reduce(268), // |=, reduce: FunctionCallsite
			reduce(268), // &=, reduce: FunctionCallsite
			reduce(268), // ^=, reduce: FunctionCallsite
			reduce(268), // <<=, reduce: FunctionCallsite
			reduce(268), // >>=, reduce: FunctionCallsite
			reduce(268), // >>>=, reduce: FunctionCallsite
			reduce(268), // +=, reduce: FunctionCallsite
			reduce(268), // .=, reduce: FunctionCallsite
			reduce(268), // -=, reduce: FunctionCallsite
			reduce(268), // *=, reduce: FunctionCallsite
			reduce(268), // /=, reduce: FunctionCallsite
			reduce(268), // //=, reduce: FunctionCallsite
			reduce(268), // %=, reduce: FunctionCallsite
			reduce(268), // **=, reduce: FunctionCallsite
			reduce(268), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(268), // ||, reduce: FunctionCallsite
			reduce(268), // ^^, reduce: FunctionCallsite
			reduce(268), // &&, reduce: FunctionCallsite
			reduce(268), // ??, reduce: FunctionCallsite
			reduce(268), // ???, reduce: FunctionCallsite
			reduce(268), // =~, reduce: FunctionCallsite
			reduce(268), // !=~, reduce: FunctionCallsite
			reduce(268), // ==, reduce: FunctionCallsite
			reduce(268), // !=, reduce: FunctionCallsite
			reduce(268), // <=>, reduce: FunctionCallsite
			reduce(268), // >=, reduce: FunctionCallsite
			reduce(268), // <, reduce: FunctionCallsite
			reduce(268), // <=, reduce: FunctionCallsite
			reduce(268), // ^, reduce: FunctionCallsite
			reduce(268), // &, reduce: FunctionCallsite
			reduce(268), // <<, reduce: FunctionCallsite
			reduce(268), // >>>, reduce: FunctionCallsite
			reduce(268), // +, reduce: FunctionCallsite
			reduce(268), // -, reduce: FunctionCallsite
			reduce(268), // .+, reduce: FunctionCallsite
			reduce(268), // .-, reduce: FunctionCallsite
			reduce(268), // *, reduce: FunctionCallsite
			reduce(268), // /, reduce: FunctionCallsite
			reduce(268), // //, reduce: FunctionCallsite
			reduce(268), // %, reduce: FunctionCallsite
			reduce(268), // .*, reduce: FunctionCallsite
			reduce(268), // ./, reduce: FunctionCallsite
			reduce(268), // .//, reduce: FunctionCallsite
			reduce(268), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(268), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(268), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(268), // [[, reduce: FunctionCallsite
			reduce(268), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3186
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(3921), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3187
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3188
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(296), // ;, reduce: ElseBlock
			reduce(296), // {, reduce: ElseBlock
			reduce(296), // }, reduce: ElseBlock
			nil,         // =
			reduce(296), // unset, reduce: ElseBlock
			reduce(296), // filter, reduce: ElseBlock
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(296), // print, reduce: ElseBlock
			nil,         // ,
			reduce(296), // printn, reduce: ElseBlock
			reduce(296), // eprint, reduce: ElseBlock
			reduce(296), // eprintn, reduce: ElseBlock
			reduce(296), // dump, reduce: ElseBlock
			reduce(296), // edump, reduce: ElseBlock
			reduce(296), // tee, reduce: ElseBlock
			reduce(296), // emitf, reduce: ElseBlock
			reduce(296), // emit1, reduce: ElseBlock
			reduce(296), // emit, reduce: ElseBlock
			reduce(296), // (, reduce: ElseBlock
			nil,         // )
			reduce(296), // emitp, reduce: ElseBlock
			reduce(296), // field_name, reduce: ElseBlock
			reduce(296), // $[, reduce: ElseBlock
			nil,         // ]
			reduce(296), // braced_field_name, reduce: ElseBlock
			reduce(296), // $[[, reduce: ElseBlock
			reduce(296), // $[[[, reduce: ElseBlock
			reduce(296), // full_srec, reduce: ElseBlock
			reduce(296), // oosvar_name, reduce: ElseBlock
			reduce(296), // @[, reduce: ElseBlock
			reduce(296), // braced_oosvar_name, reduce: ElseBlock
			reduce(296), // full_oosvar, reduce: ElseBlock
			reduce(296), // all, reduce: ElseBlock
			reduce(296), // non_sigil_name, reduce: ElseBlock
			reduce(296), // arr, reduce: ElseBlock
			reduce(296), // bool, reduce: ElseBlock
			reduce(296), // float, reduce: ElseBlock
			reduce(296), // int, reduce: ElseBlock
			reduce(296), // map, reduce: ElseBlock
			reduce(296), // num, reduce: ElseBlock
			reduce(296), // str, reduce: ElseBlock
			reduce(296), // var, reduce: ElseBlock
			reduce(296), // funct, reduce: ElseBlock
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(296), // +, reduce: ElseBlock
			reduce(296), // -, reduce: ElseBlock
			reduce(296), // .+, reduce: ElseBlock
			reduce(296), // .-, reduce: ElseBlock
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(296), // !, reduce: ElseBlock
			reduce(296), // ~, reduce: ElseBlock
			nil,         // **
			reduce(296), // string_literal, reduce: ElseBlock
			reduce(296), // regex_case_insensitive, reduce: ElseBlock
			reduce(296), // int_literal, reduce: ElseBlock
			reduce(296), // float_literal, reduce: ElseBlock
			reduce(296), // boolean_literal, reduce: ElseBlock
			reduce(296), // null_literal, reduce: ElseBlock
			reduce(296), // inf_literal, reduce: ElseBlock
			reduce(296), // nan_literal, reduce: ElseBlock
			reduce(296), // const_M_PI, reduce: ElseBlock
			reduce(296), // const_M_E, reduce: ElseBlock
			reduce(296), // panic, reduce: ElseBlock
			reduce(296), // [, reduce: ElseBlock
			reduce(296), // ctx_IPS, reduce: ElseBlock
			reduce(296), // ctx_IFS, reduce: ElseBlock
			reduce(296), // ctx_IRS, reduce: ElseBlock
			reduce(296), // ctx_OPS, reduce: ElseBlock
			reduce(296), // ctx_OFS, reduce: ElseBlock
			reduce(296), // ctx_ORS, reduce: ElseBlock
			reduce(296), // ctx_FLATSEP, reduce: ElseBlock
			reduce(296), // ctx_NF, reduce: ElseBlock
			reduce(296), // ctx_NR, reduce: ElseBlock
			reduce(296), // ctx_FNR, reduce: ElseBlock
			reduce(296), // ctx_FILENAME, reduce: ElseBlock
			reduce(296), // ctx_FILENUM, reduce: ElseBlock
			reduce(296), // env, reduce: ElseBlock
			nil,         // [[
			nil,         // [[[
			reduce(296), // call, reduce: ElseBlock
			reduce(296), // begin, reduce: ElseBlock
			reduce(296), // end, reduce: ElseBlock
			reduce(296), // if, reduce: ElseBlock
			nil,         // elif
			nil,         // else
			reduce(296), // while, reduce: ElseBlock
			reduce(296), // do, reduce: ElseBlock
			reduce(296), // for, reduce: ElseBlock
			nil,         // in
			reduce(296), // break, reduce: ElseBlock
			reduce(296), // continue, reduce: ElseBlock
			reduce(296), // func, reduce: ElseBlock
			reduce(296), // subr, reduce: ElseBlock
			reduce(296), // return, reduce: ElseBlock
		},
	},
	actionRow{ // S3189
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3923), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3190
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3924), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3191
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3925), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3192
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(214), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3926), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(214), // ||=, reduce: MlrvalOrFunction
			reduce(214), // ^^=, reduce: MlrvalOrFunction
			reduce(214), // &&=, reduce: MlrvalOrFunction
			reduce(214), // ??=, reduce: MlrvalOrFunction
			reduce(214), // ???=, reduce: MlrvalOrFunction
			reduce(214), // |=, reduce: MlrvalOrFunction
			reduce(214), // &=, reduce: MlrvalOrFunction
			reduce(214), // ^=, reduce: MlrvalOrFunction
			reduce(214), // <<=, reduce: MlrvalOrFunction
			reduce(214), // >>=, reduce: MlrvalOrFunction
			reduce(214), // >>>=, reduce: MlrvalOrFunction
			reduce(214), // +=, reduce: MlrvalOrFunction
			reduce(214), // .=, reduce: MlrvalOrFunction
			reduce(214), // -=, reduce: MlrvalOrFunction
			reduce(214), // *=, reduce: MlrvalOrFunction
			reduce(214), // /=, reduce: MlrvalOrFunction
			reduce(214), // //=, reduce: MlrvalOrFunction
			reduce(214), // %=, reduce: MlrvalOrFunction
			reduce(214), // **=, reduce: MlrvalOrFunction
			reduce(214), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			shift(3927), // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3193
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(520),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(524),  // (
			nil,         // )
			nil,         // emitp
			shift(536),  // field_name
			shift(537),  // $[
			nil,         // ]
			shift(538),  // braced_field_name
			shift(539),  // $[[
			shift(540),  // $[[[
			shift(541),  // full_srec
			shift(543),  // oosvar_name
			shift(544),  // @[
			shift(545),  // braced_oosvar_name
			shift(546),  // full_oosvar
			shift(547),  // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(563),  // +
			shift(565),  // -
			shift(566),  // .+
			shift(567),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(570),  // !
			shift(571),  // ~
			nil,         // **
			shift(576),  // string_literal
			shift(577),  // regex_case_insensitive
			shift(578),  // int_literal
			shift(579),  // float_literal
			shift(580),  // boolean_literal
			shift(581),  // null_literal
			shift(582),  // inf_literal
			shift(583),  // nan_literal
			shift(584),  // const_M_PI
			shift(585),  // const_M_E
			shift(586),  // panic
			shift(588),  // [
			shift(590),  // ctx_IPS
			shift(591),  // ctx_IFS
			shift(592),  // ctx_IRS
			shift(593),  // ctx_OPS
			shift(594),  // ctx_OFS
			shift(595),  // ctx_ORS
			shift(596),  // ctx_FLATSEP
			shift(597),  // ctx_NF
			shift(598),  // ctx_NR
			shift(599),  // ctx_FNR
			shift(600),  // ctx_FILENAME
			shift(601),  // ctx_FILENUM
			shift(603),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(610),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3194
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(3929), // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(2898), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3195
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3930), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3196
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3197
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3198
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(107), // $, reduce: PositionalFieldName
			nil,         // empty
			reduce(107), // ;, reduce: PositionalFieldName
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(107), // >, reduce: PositionalFieldName
			reduce(107), // >>, reduce: PositionalFieldName
			reduce(107), // |, reduce: PositionalFieldName
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(107), // ,, reduce: PositionalFieldName
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(107), // ?, reduce: PositionalFieldName
			nil,         // :
			reduce(107), // ||, reduce: PositionalFieldName
			reduce(107), // ^^, reduce: PositionalFieldName
			reduce(107), // &&, reduce: PositionalFieldName
			reduce(107), // ??, reduce: PositionalFieldName
			reduce(107), // ???, reduce: PositionalFieldName
			reduce(107), // =~, reduce: PositionalFieldName
			reduce(107), // !=~, reduce: PositionalFieldName
			reduce(107), // ==, reduce: PositionalFieldName
			reduce(107), // !=, reduce: PositionalFieldName
			reduce(107), // <=>, reduce: PositionalFieldName
			reduce(107), // >=, reduce: PositionalFieldName
			reduce(107), // <, reduce: PositionalFieldName
			reduce(107), // <=, reduce: PositionalFieldName
			reduce(107), // ^, reduce: PositionalFieldName
			reduce(107), // &, reduce: PositionalFieldName
			reduce(107), // <<, reduce: PositionalFieldName
			reduce(107), // >>>, reduce: PositionalFieldName
			reduce(107), // +, reduce: PositionalFieldName
			reduce(107), // -, reduce: PositionalFieldName
			reduce(107), // .+, reduce: PositionalFieldName
			reduce(107), // .-, reduce: PositionalFieldName
			reduce(107), // *, reduce: PositionalFieldName
			reduce(107), // /, reduce: PositionalFieldName
			reduce(107), // //, reduce: PositionalFieldName
			reduce(107), // %, reduce: PositionalFieldName
			reduce(107), // .*, reduce: PositionalFieldName
			reduce(107), // ./, reduce: PositionalFieldName
			reduce(107), // .//, reduce: PositionalFieldName
			reduce(107), // ., reduce: PositionalFieldName
			nil,         // !
			nil,         // ~
			reduce(107), // **, reduce: PositionalFieldName
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(107), // [, reduce: PositionalFieldName
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(107), // [[, reduce: PositionalFieldName
			reduce(107), // [[[, reduce: PositionalFieldName
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3199
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3933), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3200
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3201
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(206), // $, reduce: PowTerm
			nil,         // empty
			reduce(206), // ;, reduce: PowTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(206), // >, reduce: PowTerm
			reduce(206), // >>, reduce: PowTerm
			reduce(206), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(206), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(206), // ?, reduce: PowTerm
			nil,         // :
			reduce(206), // ||, reduce: PowTerm
			reduce(206), // ^^, reduce: PowTerm
			reduce(206), // &&, reduce: PowTerm
			reduce(206), // ??, reduce: PowTerm
			reduce(206), // ???, reduce: PowTerm
			reduce(206), // =~, reduce: PowTerm
			reduce(206), // !=~, reduce: PowTerm
			reduce(206), // ==, reduce: PowTerm
			reduce(206), // !=, reduce: PowTerm
			reduce(206), // <=>, reduce: PowTerm
			reduce(206), // >=, reduce: PowTerm
			reduce(206), // <, reduce: PowTerm
			reduce(206), // <=, reduce: PowTerm
			reduce(206), // ^, reduce: PowTerm
			reduce(206), // &, reduce: PowTerm
			reduce(206), // <<, reduce: PowTerm
			reduce(206), // >>>, reduce: PowTerm
			reduce(206), // +, reduce: PowTerm
			reduce(206), // -, reduce: PowTerm
			reduce(206), // .+, reduce: PowTerm
			reduce(206), // .-, reduce: PowTerm
			reduce(206), // *, reduce: PowTerm
			reduce(206), // /, reduce: PowTerm
			reduce(206), // //, reduce: PowTerm
			reduce(206), // %, reduce: PowTerm
			reduce(206), // .*, reduce: PowTerm
			reduce(206), // ./, reduce: PowTerm
			reduce(206), // .//, reduce: PowTerm
			reduce(206), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3202
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(205), // $, reduce: PowTerm
			nil,         // empty
			reduce(205), // ;, reduce: PowTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(205), // >, reduce: PowTerm
			reduce(205), // >>, reduce: PowTerm
			reduce(205), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(205), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(205), // ?, reduce: PowTerm
			nil,         // :
			reduce(205), // ||, reduce: PowTerm
			reduce(205), // ^^, reduce: PowTerm
			reduce(205), // &&, reduce: PowTerm
			reduce(205), // ??, reduce: PowTerm
			reduce(205), // ???, reduce: PowTerm
			reduce(205), // =~, reduce: PowTerm
			reduce(205), // !=~, reduce: PowTerm
			reduce(205), // ==, reduce: PowTerm
			reduce(205), // !=, reduce: PowTerm
			reduce(205), // <=>, reduce: PowTerm
			reduce(205), // >=, reduce: PowTerm
			reduce(205), // <, reduce: PowTerm
			reduce(205), // <=, reduce: PowTerm
			reduce(205), // ^, reduce: PowTerm
			reduce(205), // &, reduce: PowTerm
			reduce(205), // <<, reduce: PowTerm
			reduce(205), // >>>, reduce: PowTerm
			reduce(205), // +, reduce: PowTerm
			reduce(205), // -, reduce: PowTerm
			reduce(205), // .+, reduce: PowTerm
			reduce(205), // .-, reduce: PowTerm
			reduce(205), // *, reduce: PowTerm
			reduce(205), // /, reduce: PowTerm
			reduce(205), // //, reduce: PowTerm
			reduce(205), // %, reduce: PowTerm
			reduce(205), // .*, reduce: PowTerm
			reduce(205), // ./, reduce: PowTerm
			reduce(205), // .//, reduce: PowTerm
			reduce(205), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3203
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(260), // $, reduce: ArrayOrMapIndexAccess
			nil,         // empty
			reduce(260), // ;, reduce: ArrayOrMapIndexAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(260), // >, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>, reduce: ArrayOrMapIndexAccess
			reduce(260), // |, reduce: ArrayOrMapIndexAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(260), // ,, reduce: ArrayOrMapIndexAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(260), // ?, reduce: ArrayOrMapIndexAccess
			nil,         // :
			reduce(260), // ||, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???, reduce: ArrayOrMapIndexAccess
			reduce(260), // =~, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=~, reduce: ArrayOrMapIndexAccess
			reduce(260), // ==, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=>, reduce: ArrayOrMapIndexAccess
			reduce(260), // >=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>, reduce: ArrayOrMapIndexAccess
			reduce(260), // +, reduce: ArrayOrMapIndexAccess
			reduce(260), // -, reduce: ArrayOrMapIndexAccess
			reduce(260), // .+, reduce: ArrayOrMapIndexAccess
			reduce(260), // .-, reduce: ArrayOrMapIndexAccess
			reduce(260), // *, reduce: ArrayOrMapIndexAccess
			reduce(260), // /, reduce: ArrayOrMapIndexAccess
			reduce(260), // //, reduce: ArrayOrMapIndexAccess
			reduce(260), // %, reduce: ArrayOrMapIndexAccess
			reduce(260), // .*, reduce: ArrayOrMapIndexAccess
			reduce(260), // ./, reduce: ArrayOrMapIndexAccess
			reduce(260), // .//, reduce: ArrayOrMapIndexAccess
			reduce(260), // ., reduce: ArrayOrMapIndexAccess
			nil,         // !
			nil,         // ~
			reduce(260), // **, reduce: ArrayOrMapIndexAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(260), // [, reduce: ArrayOrMapIndexAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(260), // [[, reduce: ArrayOrMapIndexAccess
			reduce(260), // [[[, reduce: ArrayOrMapIndexAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3204
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(3936), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3205
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3937), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3206
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(266), // $, reduce: ArraySliceAccess
			nil,         // empty
			reduce(266), // ;, reduce: ArraySliceAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(266), // >, reduce: ArraySliceAccess
			reduce(266), // >>, reduce: ArraySliceAccess
			reduce(266), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(266), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(266), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(266), // ||, reduce: ArraySliceAccess
			reduce(266), // ^^, reduce: ArraySliceAccess
			reduce(266), // &&, reduce: ArraySliceAccess
			reduce(266), // ??, reduce: ArraySliceAccess
			reduce(266), // ???, reduce: ArraySliceAccess
			reduce(266), // =~, reduce: ArraySliceAccess
			reduce(266), // !=~, reduce: ArraySliceAccess
			reduce(266), // ==, reduce: ArraySliceAccess
			reduce(266), // !=, reduce: ArraySliceAccess
			reduce(266), // <=>, reduce: ArraySliceAccess
			reduce(266), // >=, reduce: ArraySliceAccess
			reduce(266), // <, reduce: ArraySliceAccess
			reduce(266), // <=, reduce: ArraySliceAccess
			reduce(266), // ^, reduce: ArraySliceAccess
			reduce(266), // &, reduce: ArraySliceAccess
			reduce(266), // <<, reduce: ArraySliceAccess
			reduce(266), // >>>, reduce: ArraySliceAccess
			reduce(266), // +, reduce: ArraySliceAccess
			reduce(266), // -, reduce: ArraySliceAccess
			reduce(266), // .+, reduce: ArraySliceAccess
			reduce(266), // .-, reduce: ArraySliceAccess
			reduce(266), // *, reduce: ArraySliceAccess
			reduce(266), // /, reduce: ArraySliceAccess
			reduce(266), // //, reduce: ArraySliceAccess
			reduce(266), // %, reduce: ArraySliceAccess
			reduce(266), // .*, reduce: ArraySliceAccess
			reduce(266), // ./, reduce: ArraySliceAccess
			reduce(266), // .//, reduce: ArraySliceAccess
			reduce(266), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(266), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(266), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(266), // [[, reduce: ArraySliceAccess
			reduce(266), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3207
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3938), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3208
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3939), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3209
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(254), // $, reduce: ENV
			nil,         // empty
			reduce(254), // ;, reduce: ENV
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(254), // >, reduce: ENV
			reduce(254), // >>, reduce: ENV
			reduce(254), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(254), // ,, reduce: ENV
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(254), // ?, reduce: ENV
			nil,         // :
			reduce(254), // ||, reduce: ENV
			reduce(254), // ^^, reduce: ENV
			reduce(254), // &&, reduce: ENV
			reduce(254), // ??, reduce: ENV
			reduce(254), // ???, reduce: ENV
			reduce(254), // =~, reduce: ENV
			reduce(254), // !=~, reduce: ENV
			reduce(254), // ==, reduce: ENV
			reduce(254), // !=, reduce: ENV
			reduce(254), // <=>, reduce: ENV
			reduce(254), // >=, reduce: ENV
			reduce(254), // <, reduce: ENV
			reduce(254), // <=, reduce: ENV
			reduce(254), // ^, reduce: ENV
			reduce(254), // &, reduce: ENV
			reduce(254), // <<, reduce: ENV
			reduce(254), // >>>, reduce: ENV
			reduce(254), // +, reduce: ENV
			reduce(254), // -, reduce: ENV
			reduce(254), // .+, reduce: ENV
			reduce(254), // .-, reduce: ENV
			reduce(254), // *, reduce: ENV
			reduce(254), // /, reduce: ENV
			reduce(254), // //, reduce: ENV
			reduce(254), // %, reduce: ENV
			reduce(254), // .*, reduce: ENV
			reduce(254), // ./, reduce: ENV
			reduce(254), // .//, reduce: ENV
			reduce(254), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(254), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(254), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(254), // [[, reduce: ENV
			reduce(254), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3210
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(269), // $, reduce: FunctionCallsite
			nil,         // empty
			reduce(269), // ;, reduce: FunctionCallsite
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(269), // >, reduce: FunctionCallsite
			reduce(269), // >>, reduce: FunctionCallsite
			reduce(269), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(269), // ,, reduce: FunctionCallsite
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(269), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(269), // ||, reduce: FunctionCallsite
			reduce(269), // ^^, reduce: FunctionCallsite
			reduce(269), // &&, reduce: FunctionCallsite
			reduce(269), // ??, reduce: FunctionCallsite
			reduce(269), // ???, reduce: FunctionCallsite
			reduce(269), // =~, reduce: FunctionCallsite
			reduce(269), // !=~, reduce: FunctionCallsite
			reduce(269), // ==, reduce: FunctionCallsite
			reduce(269), // !=, reduce: FunctionCallsite
			reduce(269), // <=>, reduce: FunctionCallsite
			reduce(269), // >=, reduce: FunctionCallsite
			reduce(269), // <, reduce: FunctionCallsite
			reduce(269), // <=, reduce: FunctionCallsite
			reduce(269), // ^, reduce: FunctionCallsite
			reduce(269), // &, reduce: FunctionCallsite
			reduce(269), // <<, reduce: FunctionCallsite
			reduce(269), // >>>, reduce: FunctionCallsite
			reduce(269), // +, reduce: FunctionCallsite
			reduce(269), // -, reduce: FunctionCallsite
			reduce(269), // .+, reduce: FunctionCallsite
			reduce(269), // .-, reduce: FunctionCallsite
			reduce(269), // *, reduce: FunctionCallsite
			reduce(269), // /, reduce: FunctionCallsite
			reduce(269), // //, reduce: FunctionCallsite
			reduce(269), // %, reduce: FunctionCallsite
			reduce(269), // .*, reduce: FunctionCallsite
			reduce(269), // ./, reduce: FunctionCallsite
			reduce(269), // .//, reduce: FunctionCallsite
			reduce(269), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(269), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(269), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(269), // [[, reduce: FunctionCallsite
			reduce(269), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3211
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3941), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3942), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3212
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(107), // $, reduce: PositionalFieldName
			nil,         // empty
			reduce(107), // ;, reduce: PositionalFieldName
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(107), // >, reduce: PositionalFieldName
			reduce(107), // >>, reduce: PositionalFieldName
			reduce(107), // |, reduce: PositionalFieldName
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(107), // ?, reduce: PositionalFieldName
			nil,         // :
			reduce(107), // ||, reduce: PositionalFieldName
			reduce(107), // ^^, reduce: PositionalFieldName
			reduce(107), // &&, reduce: PositionalFieldName
			reduce(107), // ??, reduce: PositionalFieldName
			reduce(107), // ???, reduce: PositionalFieldName
			reduce(107), // =~, reduce: PositionalFieldName
			reduce(107), // !=~, reduce: PositionalFieldName
			reduce(107), // ==, reduce: PositionalFieldName
			reduce(107), // !=, reduce: PositionalFieldName
			reduce(107), // <=>, reduce: PositionalFieldName
			reduce(107), // >=, reduce: PositionalFieldName
			reduce(107), // <, reduce: PositionalFieldName
			reduce(107), // <=, reduce: PositionalFieldName
			reduce(107), // ^, reduce: PositionalFieldName
			reduce(107), // &, reduce: PositionalFieldName
			reduce(107), // <<, reduce: PositionalFieldName
			reduce(107), // >>>, reduce: PositionalFieldName
			reduce(107), // +, reduce: PositionalFieldName
			reduce(107), // -, reduce: PositionalFieldName
			reduce(107), // .+, reduce: PositionalFieldName
			reduce(107), // .-, reduce: PositionalFieldName
			reduce(107), // *, reduce: PositionalFieldName
			reduce(107), // /, reduce: PositionalFieldName
			reduce(107), // //, reduce: PositionalFieldName
			reduce(107), // %, reduce: PositionalFieldName
			reduce(107), // .*, reduce: PositionalFieldName
			reduce(107), // ./, reduce: PositionalFieldName
			reduce(107), // .//, reduce: PositionalFieldName
			reduce(107), // ., reduce: PositionalFieldName
			nil,         // !
			nil,         // ~
			reduce(107), // **, reduce: PositionalFieldName
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(107), // [, reduce: PositionalFieldName
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(107), // [[, reduce: PositionalFieldName
			reduce(107), // [[[, reduce: PositionalFieldName
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3213
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3943), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3214
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(384), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(388), // (
			nil,        // )
			nil,        // emitp
			shift(400), // field_name
			shift(401), // $[
			nil,        // ]
			shift(402), // braced_field_name
			shift(403), // $[[
			shift(404), // $[[[
			shift(405), // full_srec
			shift(407), // oosvar_name
			shift(408), // @[
			shift(409), // braced_oosvar_name
			shift(410), // full_oosvar
			shift(411), // all
			shift(412), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(427), // +
			shift(429), // -
			shift(430), // .+
			shift(431), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(434), // !
			shift(435), // ~
			nil,        // **
			shift(440), // string_literal
			shift(441), // regex_case_insensitive
			shift(442), // int_literal
			shift(443), // float_literal
			shift(444), // boolean_literal
			shift(445), // null_literal
			shift(446), // inf_literal
			shift(447), // nan_literal
			shift(448), // const_M_PI
			shift(449), // const_M_E
			shift(450), // panic
			shift(452), // [
			shift(454), // ctx_IPS
			shift(455), // ctx_IFS
			shift(456), // ctx_IRS
			shift(457), // ctx_OPS
			shift(458), // ctx_OFS
			shift(459), // ctx_ORS
			shift(460), // ctx_FLATSEP
			shift(461), // ctx_NF
			shift(462), // ctx_NR
			shift(463), // ctx_FNR
			shift(464), // ctx_FILENAME
			shift(465), // ctx_FILENUM
			shift(467), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(474), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3215
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(206), // $, reduce: PowTerm
			nil,         // empty
			reduce(206), // ;, reduce: PowTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(206), // >, reduce: PowTerm
			reduce(206), // >>, reduce: PowTerm
			reduce(206), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(206), // ?, reduce: PowTerm
			nil,         // :
			reduce(206), // ||, reduce: PowTerm
			reduce(206), // ^^, reduce: PowTerm
			reduce(206), // &&, reduce: PowTerm
			reduce(206), // ??, reduce: PowTerm
			reduce(206), // ???, reduce: PowTerm
			reduce(206), // =~, reduce: PowTerm
			reduce(206), // !=~, reduce: PowTerm
			reduce(206), // ==, reduce: PowTerm
			reduce(206), // !=, reduce: PowTerm
			reduce(206), // <=>, reduce: PowTerm
			reduce(206), // >=, reduce: PowTerm
			reduce(206), // <, reduce: PowTerm
			reduce(206), // <=, reduce: PowTerm
			reduce(206), // ^, reduce: PowTerm
			reduce(206), // &, reduce: PowTerm
			reduce(206), // <<, reduce: PowTerm
			reduce(206), // >>>, reduce: PowTerm
			reduce(206), // +, reduce: PowTerm
			reduce(206), // -, reduce: PowTerm
			reduce(206), // .+, reduce: PowTerm
			reduce(206), // .-, reduce: PowTerm
			reduce(206), // *, reduce: PowTerm
			reduce(206), // /, reduce: PowTerm
			reduce(206), // //, reduce: PowTerm
			reduce(206), // %, reduce: PowTerm
			reduce(206), // .*, reduce: PowTerm
			reduce(206), // ./, reduce: PowTerm
			reduce(206), // .//, reduce: PowTerm
			reduce(206), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3216
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(205), // $, reduce: PowTerm
			nil,         // empty
			reduce(205), // ;, reduce: PowTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(205), // >, reduce: PowTerm
			reduce(205), // >>, reduce: PowTerm
			reduce(205), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(205), // ?, reduce: PowTerm
			nil,         // :
			reduce(205), // ||, reduce: PowTerm
			reduce(205), // ^^, reduce: PowTerm
			reduce(205), // &&, reduce: PowTerm
			reduce(205), // ??, reduce: PowTerm
			reduce(205), // ???, reduce: PowTerm
			reduce(205), // =~, reduce: PowTerm
			reduce(205), // !=~, reduce: PowTerm
			reduce(205), // ==, reduce: PowTerm
			reduce(205), // !=, reduce: PowTerm
			reduce(205), // <=>, reduce: PowTerm
			reduce(205), // >=, reduce: PowTerm
			reduce(205), // <, reduce: PowTerm
			reduce(205), // <=, reduce: PowTerm
			reduce(205), // ^, reduce: PowTerm
			reduce(205), // &, reduce: PowTerm
			reduce(205), // <<, reduce: PowTerm
			reduce(205), // >>>, reduce: PowTerm
			reduce(205), // +, reduce: PowTerm
			reduce(205), // -, reduce: PowTerm
			reduce(205), // .+, reduce: PowTerm
			reduce(205), // .-, reduce: PowTerm
			reduce(205), // *, reduce: PowTerm
			reduce(205), // /, reduce: PowTerm
			reduce(205), // //, reduce: PowTerm
			reduce(205), // %, reduce: PowTerm
			reduce(205), // .*, reduce: PowTerm
			reduce(205), // ./, reduce: PowTerm
			reduce(205), // .//, reduce: PowTerm
			reduce(205), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3217
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(260), // $, reduce: ArrayOrMapIndexAccess
			nil,         // empty
			reduce(260), // ;, reduce: ArrayOrMapIndexAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(260), // >, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>, reduce: ArrayOrMapIndexAccess
			reduce(260), // |, reduce: ArrayOrMapIndexAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(260), // ?, reduce: ArrayOrMapIndexAccess
			nil,         // :
			reduce(260), // ||, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???, reduce: ArrayOrMapIndexAccess
			reduce(260), // =~, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=~, reduce: ArrayOrMapIndexAccess
			reduce(260), // ==, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=>, reduce: ArrayOrMapIndexAccess
			reduce(260), // >=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>, reduce: ArrayOrMapIndexAccess
			reduce(260), // +, reduce: ArrayOrMapIndexAccess
			reduce(260), // -, reduce: ArrayOrMapIndexAccess
			reduce(260), // .+, reduce: ArrayOrMapIndexAccess
			reduce(260), // .-, reduce: ArrayOrMapIndexAccess
			reduce(260), // *, reduce: ArrayOrMapIndexAccess
			reduce(260), // /, reduce: ArrayOrMapIndexAccess
			reduce(260), // //, reduce: ArrayOrMapIndexAccess
			reduce(260), // %, reduce: ArrayOrMapIndexAccess
			reduce(260), // .*, reduce: ArrayOrMapIndexAccess
			reduce(260), // ./, reduce: ArrayOrMapIndexAccess
			reduce(260), // .//, reduce: ArrayOrMapIndexAccess
			reduce(260), // ., reduce: ArrayOrMapIndexAccess
			nil,         // !
			nil,         // ~
			reduce(260), // **, reduce: ArrayOrMapIndexAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(260), // [, reduce: ArrayOrMapIndexAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(260), // [[, reduce: ArrayOrMapIndexAccess
			reduce(260), // [[[, reduce: ArrayOrMapIndexAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3218
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(3946), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3219
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3947), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3220
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(266), // $, reduce: ArraySliceAccess
			nil,         // empty
			reduce(266), // ;, reduce: ArraySliceAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(266), // >, reduce: ArraySliceAccess
			reduce(266), // >>, reduce: ArraySliceAccess
			reduce(266), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(266), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(266), // ||, reduce: ArraySliceAccess
			reduce(266), // ^^, reduce: ArraySliceAccess
			reduce(266), // &&, reduce: ArraySliceAccess
			reduce(266), // ??, reduce: ArraySliceAccess
			reduce(266), // ???, reduce: ArraySliceAccess
			reduce(266), // =~, reduce: ArraySliceAccess
			reduce(266), // !=~, reduce: ArraySliceAccess
			reduce(266), // ==, reduce: ArraySliceAccess
			reduce(266), // !=, reduce: ArraySliceAccess
			reduce(266), // <=>, reduce: ArraySliceAccess
			reduce(266), // >=, reduce: ArraySliceAccess
			reduce(266), // <, reduce: ArraySliceAccess
			reduce(266), // <=, reduce: ArraySliceAccess
			reduce(266), // ^, reduce: ArraySliceAccess
			reduce(266), // &, reduce: ArraySliceAccess
			reduce(266), // <<, reduce: ArraySliceAccess
			reduce(266), // >>>, reduce: ArraySliceAccess
			reduce(266), // +, reduce: ArraySliceAccess
			reduce(266), // -, reduce: ArraySliceAccess
			reduce(266), // .+, reduce: ArraySliceAccess
			reduce(266), // .-, reduce: ArraySliceAccess
			reduce(266), // *, reduce: ArraySliceAccess
			reduce(266), // /, reduce: ArraySliceAccess
			reduce(266), // //, reduce: ArraySliceAccess
			reduce(266), // %, reduce: ArraySliceAccess
			reduce(266), // .*, reduce: ArraySliceAccess
			reduce(266), // ./, reduce: ArraySliceAccess
			reduce(266), // .//, reduce: ArraySliceAccess
			reduce(266), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(266), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(266), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(266), // [[, reduce: ArraySliceAccess
			reduce(266), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3221
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3948), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3222
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3949), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3223
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(254), // $, reduce: ENV
			nil,         // empty
			reduce(254), // ;, reduce: ENV
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(254), // >, reduce: ENV
			reduce(254), // >>, reduce: ENV
			reduce(254), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(254), // ?, reduce: ENV
			nil,         // :
			reduce(254), // ||, reduce: ENV
			reduce(254), // ^^, reduce: ENV
			reduce(254), // &&, reduce: ENV
			reduce(254), // ??, reduce: ENV
			reduce(254), // ???, reduce: ENV
			reduce(254), // =~, reduce: ENV
			reduce(254), // !=~, reduce: ENV
			reduce(254), // ==, reduce: ENV
			reduce(254), // !=, reduce: ENV
			reduce(254), // <=>, reduce: ENV
			reduce(254), // >=, reduce: ENV
			reduce(254), // <, reduce: ENV
			reduce(254), // <=, reduce: ENV
			reduce(254), // ^, reduce: ENV
			reduce(254), // &, reduce: ENV
			reduce(254), // <<, reduce: ENV
			reduce(254), // >>>, reduce: ENV
			reduce(254), // +, reduce: ENV
			reduce(254), // -, reduce: ENV
			reduce(254), // .+, reduce: ENV
			reduce(254), // .-, reduce: ENV
			reduce(254), // *, reduce: ENV
			reduce(254), // /, reduce: ENV
			reduce(254), // //, reduce: ENV
			reduce(254), // %, reduce: ENV
			reduce(254), // .*, reduce: ENV
			reduce(254), // ./, reduce: ENV
			reduce(254), // .//, reduce: ENV
			reduce(254), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(254), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(254), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(254), // [[, reduce: ENV
			reduce(254), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3224
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(269), // $, reduce: FunctionCallsite
			nil,         // empty
			reduce(269), // ;, reduce: FunctionCallsite
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(269), // >, reduce: FunctionCallsite
			reduce(269), // >>, reduce: FunctionCallsite
			reduce(269), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(269), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(269), // ||, reduce: FunctionCallsite
			reduce(269), // ^^, reduce: FunctionCallsite
			reduce(269), // &&, reduce: FunctionCallsite
			reduce(269), // ??, reduce: FunctionCallsite
			reduce(269), // ???, reduce: FunctionCallsite
			reduce(269), // =~, reduce: FunctionCallsite
			reduce(269), // !=~, reduce: FunctionCallsite
			reduce(269), // ==, reduce: FunctionCallsite
			reduce(269), // !=, reduce: FunctionCallsite
			reduce(269), // <=>, reduce: FunctionCallsite
			reduce(269), // >=, reduce: FunctionCallsite
			reduce(269), // <, reduce: FunctionCallsite
			reduce(269), // <=, reduce: FunctionCallsite
			reduce(269), // ^, reduce: FunctionCallsite
			reduce(269), // &, reduce: FunctionCallsite
			reduce(269), // <<, reduce: FunctionCallsite
			reduce(269), // >>>, reduce: FunctionCallsite
			reduce(269), // +, reduce: FunctionCallsite
			reduce(269), // -, reduce: FunctionCallsite
			reduce(269), // .+, reduce: FunctionCallsite
			reduce(269), // .-, reduce: FunctionCallsite
			reduce(269), // *, reduce: FunctionCallsite
			reduce(269), // /, reduce: FunctionCallsite
			reduce(269), // //, reduce: FunctionCallsite
			reduce(269), // %, reduce: FunctionCallsite
			reduce(269), // .*, reduce: FunctionCallsite
			reduce(269), // ./, reduce: FunctionCallsite
			reduce(269), // .//, reduce: FunctionCallsite
			reduce(269), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(269), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(269), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(269), // [[, reduce: FunctionCallsite
			reduce(269), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3225
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3951), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3952), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3226
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(235), // >, reduce: MapLiteral
			reduce(235), // >>, reduce: MapLiteral
			reduce(235), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(235), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(235), // ?, reduce: MapLiteral
			nil,         // :
			reduce(235), // ||, reduce: MapLiteral
			reduce(235), // ^^, reduce: MapLiteral
			reduce(235), // &&, reduce: MapLiteral
			reduce(235), // ??, reduce: MapLiteral
			reduce(235), // ???, reduce: MapLiteral
			reduce(235), // =~, reduce: MapLiteral
			reduce(235), // !=~, reduce: MapLiteral
			reduce(235), // ==, reduce: MapLiteral
			reduce(235), // !=, reduce: MapLiteral
			reduce(235), // <=>, reduce: MapLiteral
			reduce(235), // >=, reduce: MapLiteral
			reduce(235), // <, reduce: MapLiteral
			reduce(235), // <=, reduce: MapLiteral
			reduce(235), // ^, reduce: MapLiteral
			reduce(235), // &, reduce: MapLiteral
			reduce(235), // <<, reduce: MapLiteral
			reduce(235), // >>>, reduce: MapLiteral
			reduce(235), // +, reduce: MapLiteral
			reduce(235), // -, reduce: MapLiteral
			reduce(235), // .+, reduce: MapLiteral
			reduce(235), // .-, reduce: MapLiteral
			reduce(235), // *, reduce: MapLiteral
			reduce(235), // /, reduce: MapLiteral
			reduce(235), // //, reduce: MapLiteral
			reduce(235), // %, reduce: MapLiteral
			reduce(235), // .*, reduce: MapLiteral
			reduce(235), // ./, reduce: MapLiteral
			reduce(235), // .//, reduce: MapLiteral
			reduce(235), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(235), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(235), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(235), // [[, reduce: MapLiteral
			reduce(235), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3227
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(208), // >, reduce: PrecedenceChainEnd
			reduce(208), // >>, reduce: PrecedenceChainEnd
			reduce(208), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(208), // ,, reduce: PrecedenceChainEnd
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(208), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(208), // ||, reduce: PrecedenceChainEnd
			reduce(208), // ^^, reduce: PrecedenceChainEnd
			reduce(208), // &&, reduce: PrecedenceChainEnd
			reduce(208), // ??, reduce: PrecedenceChainEnd
			reduce(208), // ???, reduce: PrecedenceChainEnd
			reduce(208), // =~, reduce: PrecedenceChainEnd
			reduce(208), // !=~, reduce: PrecedenceChainEnd
			reduce(208), // ==, reduce: PrecedenceChainEnd
			reduce(208), // !=, reduce: PrecedenceChainEnd
			reduce(208), // <=>, reduce: PrecedenceChainEnd
			reduce(208), // >=, reduce: PrecedenceChainEnd
			reduce(208), // <, reduce: PrecedenceChainEnd
			reduce(208), // <=, reduce: PrecedenceChainEnd
			reduce(208), // ^, reduce: PrecedenceChainEnd
			reduce(208), // &, reduce: PrecedenceChainEnd
			reduce(208), // <<, reduce: PrecedenceChainEnd
			reduce(208), // >>>, reduce: PrecedenceChainEnd
			reduce(208), // +, reduce: PrecedenceChainEnd
			reduce(208), // -, reduce: PrecedenceChainEnd
			reduce(208), // .+, reduce: PrecedenceChainEnd
			reduce(208), // .-, reduce: PrecedenceChainEnd
			reduce(208), // *, reduce: PrecedenceChainEnd
			reduce(208), // /, reduce: PrecedenceChainEnd
			reduce(208), // //, reduce: PrecedenceChainEnd
			reduce(208), // %, reduce: PrecedenceChainEnd
			reduce(208), // .*, reduce: PrecedenceChainEnd
			reduce(208), // ./, reduce: PrecedenceChainEnd
			reduce(208), // .//, reduce: PrecedenceChainEnd
			reduce(208), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(208), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3228
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(105), // >, reduce: IndirectFieldValue
			reduce(105), // >>, reduce: IndirectFieldValue
			reduce(105), // |, reduce: IndirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(105), // ,, reduce: IndirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(105), // ?, reduce: IndirectFieldValue
			nil,         // :
			reduce(105), // ||, reduce: IndirectFieldValue
			reduce(105), // ^^, reduce: IndirectFieldValue
			reduce(105), // &&, reduce: IndirectFieldValue
			reduce(105), // ??, reduce: IndirectFieldValue
			reduce(105), // ???, reduce: IndirectFieldValue
			reduce(105), // =~, reduce: IndirectFieldValue
			reduce(105), // !=~, reduce: IndirectFieldValue
			reduce(105), // ==, reduce: IndirectFieldValue
			reduce(105), // !=, reduce: IndirectFieldValue
			reduce(105), // <=>, reduce: IndirectFieldValue
			reduce(105), // >=, reduce: IndirectFieldValue
			reduce(105), // <, reduce: IndirectFieldValue
			reduce(105), // <=, reduce: IndirectFieldValue
			reduce(105), // ^, reduce: IndirectFieldValue
			reduce(105), // &, reduce: IndirectFieldValue
			reduce(105), // <<, reduce: IndirectFieldValue
			reduce(105), // >>>, reduce: IndirectFieldValue
			reduce(105), // +, reduce: IndirectFieldValue
			reduce(105), // -, reduce: IndirectFieldValue
			reduce(105), // .+, reduce: IndirectFieldValue
			reduce(105), // .-, reduce: IndirectFieldValue
			reduce(105), // *, reduce: IndirectFieldValue
			reduce(105), // /, reduce: IndirectFieldValue
			reduce(105), // //, reduce: IndirectFieldValue
			reduce(105), // %, reduce: IndirectFieldValue
			reduce(105), // .*, reduce: IndirectFieldValue
			reduce(105), // ./, reduce: IndirectFieldValue
			reduce(105), // .//, reduce: IndirectFieldValue
			reduce(105), // ., reduce: IndirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(105), // **, reduce: IndirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(105), // [, reduce: IndirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(105), // [[, reduce: IndirectFieldValue
			reduce(105), // [[[, reduce: IndirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3229
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3953), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3230
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3954), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3231
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(114), // >, reduce: IndirectOosvarValue
			reduce(114), // >>, reduce: IndirectOosvarValue
			reduce(114), // |, reduce: IndirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(114), // ,, reduce: IndirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(114), // ?, reduce: IndirectOosvarValue
			nil,         // :
			reduce(114), // ||, reduce: IndirectOosvarValue
			reduce(114), // ^^, reduce: IndirectOosvarValue
			reduce(114), // &&, reduce: IndirectOosvarValue
			reduce(114), // ??, reduce: IndirectOosvarValue
			reduce(114), // ???, reduce: IndirectOosvarValue
			reduce(114), // =~, reduce: IndirectOosvarValue
			reduce(114), // !=~, reduce: IndirectOosvarValue
			reduce(114), // ==, reduce: IndirectOosvarValue
			reduce(114), // !=, reduce: IndirectOosvarValue
			reduce(114), // <=>, reduce: IndirectOosvarValue
			reduce(114), // >=, reduce: IndirectOosvarValue
			reduce(114), // <, reduce: IndirectOosvarValue
			reduce(114), // <=, reduce: IndirectOosvarValue
			reduce(114), // ^, reduce: IndirectOosvarValue
			reduce(114), // &, reduce: IndirectOosvarValue
			reduce(114), // <<, reduce: IndirectOosvarValue
			reduce(114), // >>>, reduce: IndirectOosvarValue
			reduce(114), // +, reduce: IndirectOosvarValue
			reduce(114), // -, reduce: IndirectOosvarValue
			reduce(114), // .+, reduce: IndirectOosvarValue
			reduce(114), // .-, reduce: IndirectOosvarValue
			reduce(114), // *, reduce: IndirectOosvarValue
			reduce(114), // /, reduce: IndirectOosvarValue
			reduce(114), // //, reduce: IndirectOosvarValue
			reduce(114), // %, reduce: IndirectOosvarValue
			reduce(114), // .*, reduce: IndirectOosvarValue
			reduce(114), // ./, reduce: IndirectOosvarValue
			reduce(114), // .//, reduce: IndirectOosvarValue
			reduce(114), // ., reduce: IndirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(114), // **, reduce: IndirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(114), // [, reduce: IndirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(114), // [[, reduce: IndirectOosvarValue
			reduce(114), // [[[, reduce: IndirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3232
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3955), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3233
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(151), // ,, reduce: LogicalOrTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(151), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(151), // ||, reduce: LogicalOrTerm
			shift(2459), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3234
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(153), // ,, reduce: LogicalXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(153), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(153), // ||, reduce: LogicalXORTerm
			reduce(153), // ^^, reduce: LogicalXORTerm
			shift(2460), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3235
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(155), // ,, reduce: LogicalAndTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(155), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(155), // ||, reduce: LogicalAndTerm
			reduce(155), // ^^, reduce: LogicalAndTerm
			reduce(155), // &&, reduce: LogicalAndTerm
			shift(2461), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3236
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(157), // ,, reduce: AbsentCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(157), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(157), // ||, reduce: AbsentCoalesceTerm
			reduce(157), // ^^, reduce: AbsentCoalesceTerm
			reduce(157), // &&, reduce: AbsentCoalesceTerm
			reduce(157), // ??, reduce: AbsentCoalesceTerm
			shift(2462), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3237
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(159), // ,, reduce: EmptyCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(159), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(159), // ||, reduce: EmptyCoalesceTerm
			reduce(159), // ^^, reduce: EmptyCoalesceTerm
			reduce(159), // &&, reduce: EmptyCoalesceTerm
			reduce(159), // ??, reduce: EmptyCoalesceTerm
			reduce(159), // ???, reduce: EmptyCoalesceTerm
			shift(2463), // =~
			shift(2464), // !=~
			shift(2465), // ==
			shift(2466), // !=
			shift(2467), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3238
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2468), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(161), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(161), // ?, reduce: EqneTerm
			nil,         // :
			reduce(161), // ||, reduce: EqneTerm
			reduce(161), // ^^, reduce: EqneTerm
			reduce(161), // &&, reduce: EqneTerm
			reduce(161), // ??, reduce: EqneTerm
			reduce(161), // ???, reduce: EqneTerm
			reduce(161), // =~, reduce: EqneTerm
			reduce(161), // !=~, reduce: EqneTerm
			reduce(161), // ==, reduce: EqneTerm
			reduce(161), // !=, reduce: EqneTerm
			reduce(161), // <=>, reduce: EqneTerm
			shift(2469), // >=
			shift(2470), // <
			shift(2471), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3239
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2468), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(162), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(162), // ?, reduce: EqneTerm
			nil,         // :
			reduce(162), // ||, reduce: EqneTerm
			reduce(162), // ^^, reduce: EqneTerm
			reduce(162), // &&, reduce: EqneTerm
			reduce(162), // ??, reduce: EqneTerm
			reduce(162), // ???, reduce: EqneTerm
			reduce(162), // =~, reduce: EqneTerm
			reduce(162), // !=~, reduce: EqneTerm
			reduce(162), // ==, reduce: EqneTerm
			reduce(162), // !=, reduce: EqneTerm
			reduce(162), // <=>, reduce: EqneTerm
			shift(2469), // >=
			shift(2470), // <
			shift(2471), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3240
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2468), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(163), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(163), // ?, reduce: EqneTerm
			nil,         // :
			reduce(163), // ||, reduce: EqneTerm
			reduce(163), // ^^, reduce: EqneTerm
			reduce(163), // &&, reduce: EqneTerm
			reduce(163), // ??, reduce: EqneTerm
			reduce(163), // ???, reduce: EqneTerm
			reduce(163), // =~, reduce: EqneTerm
			reduce(163), // !=~, reduce: EqneTerm
			reduce(163), // ==, reduce: EqneTerm
			reduce(163), // !=, reduce: EqneTerm
			reduce(163), // <=>, reduce: EqneTerm
			shift(2469), // >=
			shift(2470), // <
			shift(2471), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3241
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2468), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(164), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(164), // ?, reduce: EqneTerm
			nil,         // :
			reduce(164), // ||, reduce: EqneTerm
			reduce(164), // ^^, reduce: EqneTerm
			reduce(164), // &&, reduce: EqneTerm
			reduce(164), // ??, reduce: EqneTerm
			reduce(164), // ???, reduce: EqneTerm
			reduce(164), // =~, reduce: EqneTerm
			reduce(164), // !=~, reduce: EqneTerm
			reduce(164), // ==, reduce: EqneTerm
			reduce(164), // !=, reduce: EqneTerm
			reduce(164), // <=>, reduce: EqneTerm
			shift(2469), // >=
			shift(2470), // <
			shift(2471), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3242
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2468), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(165), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(165), // ?, reduce: EqneTerm
			nil,         // :
			reduce(165), // ||, reduce: EqneTerm
			reduce(165), // ^^, reduce: EqneTerm
			reduce(165), // &&, reduce: EqneTerm
			reduce(165), // ??, reduce: EqneTerm
			reduce(165), // ???, reduce: EqneTerm
			reduce(165), // =~, reduce: EqneTerm
			reduce(165), // !=~, reduce: EqneTerm
			reduce(165), // ==, reduce: EqneTerm
			reduce(165), // !=, reduce: EqneTerm
			reduce(165), // <=>, reduce: EqneTerm
			shift(2469), // >=
			shift(2470), // <
			shift(2471), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3243
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(167), // >, reduce: CmpTerm
			nil,         // >>
			shift(2472), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(167), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(167), // ?, reduce: CmpTerm
			nil,         // :
			reduce(167), // ||, reduce: CmpTerm
			reduce(167), // ^^, reduce: CmpTerm
			reduce(167), // &&, reduce: CmpTerm
			reduce(167), // ??, reduce: CmpTerm
			reduce(167), // ???, reduce: CmpTerm
			reduce(167), // =~, reduce: CmpTerm
			reduce(167), // !=~, reduce: CmpTerm
			reduce(167), // ==, reduce: CmpTerm
			reduce(167), // !=, reduce: CmpTerm
			reduce(167), // <=>, reduce: CmpTerm
			reduce(167), // >=, reduce: CmpTerm
			reduce(167), // <, reduce: CmpTerm
			reduce(167), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3244
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(168), // >, reduce: CmpTerm
			nil,         // >>
			shift(2472), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(168), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(168), // ?, reduce: CmpTerm
			nil,         // :
			reduce(168), // ||, reduce: CmpTerm
			reduce(168), // ^^, reduce: CmpTerm
			reduce(168), // &&, reduce: CmpTerm
			reduce(168), // ??, reduce: CmpTerm
			reduce(168), // ???, reduce: CmpTerm
			reduce(168), // =~, reduce: CmpTerm
			reduce(168), // !=~, reduce: CmpTerm
			reduce(168), // ==, reduce: CmpTerm
			reduce(168), // !=, reduce: CmpTerm
			reduce(168), // <=>, reduce: CmpTerm
			reduce(168), // >=, reduce: CmpTerm
			reduce(168), // <, reduce: CmpTerm
			reduce(168), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3245
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(169), // >, reduce: CmpTerm
			nil,         // >>
			shift(2472), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(169), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(169), // ?, reduce: CmpTerm
			nil,         // :
			reduce(169), // ||, reduce: CmpTerm
			reduce(169), // ^^, reduce: CmpTerm
			reduce(169), // &&, reduce: CmpTerm
			reduce(169), // ??, reduce: CmpTerm
			reduce(169), // ???, reduce: CmpTerm
			reduce(169), // =~, reduce: CmpTerm
			reduce(169), // !=~, reduce: CmpTerm
			reduce(169), // ==, reduce: CmpTerm
			reduce(169), // !=, reduce: CmpTerm
			reduce(169), // <=>, reduce: CmpTerm
			reduce(169), // >=, reduce: CmpTerm
			reduce(169), // <, reduce: CmpTerm
			reduce(169), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3246
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(170), // >, reduce: CmpTerm
			nil,         // >>
			shift(2472), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(170), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(170), // ?, reduce: CmpTerm
			nil,         // :
			reduce(170), // ||, reduce: CmpTerm
			reduce(170), // ^^, reduce: CmpTerm
			reduce(170), // &&, reduce: CmpTerm
			reduce(170), // ??, reduce: CmpTerm
			reduce(170), // ???, reduce: CmpTerm
			reduce(170), // =~, reduce: CmpTerm
			reduce(170), // !=~, reduce: CmpTerm
			reduce(170), // ==, reduce: CmpTerm
			reduce(170), // !=, reduce: CmpTerm
			reduce(170), // <=>, reduce: CmpTerm
			reduce(170), // >=, reduce: CmpTerm
			reduce(170), // <, reduce: CmpTerm
			reduce(170), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3247
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(172), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(172), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(172), // ,, reduce: BitwiseORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(172), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(172), // ||, reduce: BitwiseORTerm
			reduce(172), // ^^, reduce: BitwiseORTerm
			reduce(172), // &&, reduce: BitwiseORTerm
			reduce(172), // ??, reduce: BitwiseORTerm
			reduce(172), // ???, reduce: BitwiseORTerm
			reduce(172), // =~, reduce: BitwiseORTerm
			reduce(172), // !=~, reduce: BitwiseORTerm
			reduce(172), // ==, reduce: BitwiseORTerm
			reduce(172), // !=, reduce: BitwiseORTerm
			reduce(172), // <=>, reduce: BitwiseORTerm
			reduce(172), // >=, reduce: BitwiseORTerm
			reduce(172), // <, reduce: BitwiseORTerm
			reduce(172), // <=, reduce: BitwiseORTerm
			shift(2473), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3248
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(174), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(174), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(174), // ,, reduce: BitwiseXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(174), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(174), // ||, reduce: BitwiseXORTerm
			reduce(174), // ^^, reduce: BitwiseXORTerm
			reduce(174), // &&, reduce: BitwiseXORTerm
			reduce(174), // ??, reduce: BitwiseXORTerm
			reduce(174), // ???, reduce: BitwiseXORTerm
			reduce(174), // =~, reduce: BitwiseXORTerm
			reduce(174), // !=~, reduce: BitwiseXORTerm
			reduce(174), // ==, reduce: BitwiseXORTerm
			reduce(174), // !=, reduce: BitwiseXORTerm
			reduce(174), // <=>, reduce: BitwiseXORTerm
			reduce(174), // >=, reduce: BitwiseXORTerm
			reduce(174), // <, reduce: BitwiseXORTerm
			reduce(174), // <=, reduce: BitwiseXORTerm
			reduce(174), // ^, reduce: BitwiseXORTerm
			shift(2474), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3249
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(176), // >, reduce: BitwiseANDTerm
			shift(2475), // >>
			reduce(176), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(176), // ,, reduce: BitwiseANDTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(176), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(176), // ||, reduce: BitwiseANDTerm
			reduce(176), // ^^, reduce: BitwiseANDTerm
			reduce(176), // &&, reduce: BitwiseANDTerm
			reduce(176), // ??, reduce: BitwiseANDTerm
			reduce(176), // ???, reduce: BitwiseANDTerm
			reduce(176), // =~, reduce: BitwiseANDTerm
			reduce(176), // !=~, reduce: BitwiseANDTerm
			reduce(176), // ==, reduce: BitwiseANDTerm
			reduce(176), // !=, reduce: BitwiseANDTerm
			reduce(176), // <=>, reduce: BitwiseANDTerm
			reduce(176), // >=, reduce: BitwiseANDTerm
			reduce(176), // <, reduce: BitwiseANDTerm
			reduce(176), // <=, reduce: BitwiseANDTerm
			reduce(176), // ^, reduce: BitwiseANDTerm
			reduce(176), // &, reduce: BitwiseANDTerm
			shift(2476), // <<
			shift(2477), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3250
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(179), // >, reduce: BitwiseShiftTerm
			reduce(179), // >>, reduce: BitwiseShiftTerm
			reduce(179), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(179), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(179), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(179), // ||, reduce: BitwiseShiftTerm
			reduce(179), // ^^, reduce: BitwiseShiftTerm
			reduce(179), // &&, reduce: BitwiseShiftTerm
			reduce(179), // ??, reduce: BitwiseShiftTerm
			reduce(179), // ???, reduce: BitwiseShiftTerm
			reduce(179), // =~, reduce: BitwiseShiftTerm
			reduce(179), // !=~, reduce: BitwiseShiftTerm
			reduce(179), // ==, reduce: BitwiseShiftTerm
			reduce(179), // !=, reduce: BitwiseShiftTerm
			reduce(179), // <=>, reduce: BitwiseShiftTerm
			reduce(179), // >=, reduce: BitwiseShiftTerm
			reduce(179), // <, reduce: BitwiseShiftTerm
			reduce(179), // <=, reduce: BitwiseShiftTerm
			reduce(179), // ^, reduce: BitwiseShiftTerm
			reduce(179), // &, reduce: BitwiseShiftTerm
			reduce(179), // <<, reduce: BitwiseShiftTerm
			reduce(179), // >>>, reduce: BitwiseShiftTerm
			shift(2478), // +
			shift(2479), // -
			shift(2480), // .+
			shift(2481), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3251
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(178), // >, reduce: BitwiseShiftTerm
			reduce(178), // >>, reduce: BitwiseShiftTerm
			reduce(178), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(178), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(178), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(178), // ||, reduce: BitwiseShiftTerm
			reduce(178), // ^^, reduce: BitwiseShiftTerm
			reduce(178), // &&, reduce: BitwiseShiftTerm
			reduce(178), // ??, reduce: BitwiseShiftTerm
			reduce(178), // ???, reduce: BitwiseShiftTerm
			reduce(178), // =~, reduce: BitwiseShiftTerm
			reduce(178), // !=~, reduce: BitwiseShiftTerm
			reduce(178), // ==, reduce: BitwiseShiftTerm
			reduce(178), // !=, reduce: BitwiseShiftTerm
			reduce(178), // <=>, reduce: BitwiseShiftTerm
			reduce(178), // >=, reduce: BitwiseShiftTerm
			reduce(178), // <, reduce: BitwiseShiftTerm
			reduce(178), // <=, reduce: BitwiseShiftTerm
			reduce(178), // ^, reduce: BitwiseShiftTerm
			reduce(178), // &, reduce: BitwiseShiftTerm
			reduce(178), // <<, reduce: BitwiseShiftTerm
			reduce(178), // >>>, reduce: BitwiseShiftTerm
			shift(2478), // +
			shift(2479), // -
			shift(2480), // .+
			shift(2481), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3252
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(180), // >, reduce: BitwiseShiftTerm
			reduce(180), // >>, reduce: BitwiseShiftTerm
			reduce(180), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(180), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(180), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(180), // ||, reduce: BitwiseShiftTerm
			reduce(180), // ^^, reduce: BitwiseShiftTerm
			reduce(180), // &&, reduce: BitwiseShiftTerm
			reduce(180), // ??, reduce: BitwiseShiftTerm
			reduce(180), // ???, reduce: BitwiseShiftTerm
			reduce(180), // =~, reduce: BitwiseShiftTerm
			reduce(180), // !=~, reduce: BitwiseShiftTerm
			reduce(180), // ==, reduce: BitwiseShiftTerm
			reduce(180), // !=, reduce: BitwiseShiftTerm
			reduce(180), // <=>, reduce: BitwiseShiftTerm
			reduce(180), // >=, reduce: BitwiseShiftTerm
			reduce(180), // <, reduce: BitwiseShiftTerm
			reduce(180), // <=, reduce: BitwiseShiftTerm
			reduce(180), // ^, reduce: BitwiseShiftTerm
			reduce(180), // &, reduce: BitwiseShiftTerm
			reduce(180), // <<, reduce: BitwiseShiftTerm
			reduce(180), // >>>, reduce: BitwiseShiftTerm
			shift(2478), // +
			shift(2479), // -
			shift(2480), // .+
			shift(2481), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3253
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(182), // >, reduce: AddsubdotTerm
			reduce(182), // >>, reduce: AddsubdotTerm
			reduce(182), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(182), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(182), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(182), // ||, reduce: AddsubdotTerm
			reduce(182), // ^^, reduce: AddsubdotTerm
			reduce(182), // &&, reduce: AddsubdotTerm
			reduce(182), // ??, reduce: AddsubdotTerm
			reduce(182), // ???, reduce: AddsubdotTerm
			reduce(182), // =~, reduce: AddsubdotTerm
			reduce(182), // !=~, reduce: AddsubdotTerm
			reduce(182), // ==, reduce: AddsubdotTerm
			reduce(182), // !=, reduce: AddsubdotTerm
			reduce(182), // <=>, reduce: AddsubdotTerm
			reduce(182), // >=, reduce: AddsubdotTerm
			reduce(182), // <, reduce: AddsubdotTerm
			reduce(182), // <=, reduce: AddsubdotTerm
			reduce(182), // ^, reduce: AddsubdotTerm
			reduce(182), // &, reduce: AddsubdotTerm
			reduce(182), // <<, reduce: AddsubdotTerm
			reduce(182), // >>>, reduce: AddsubdotTerm
			reduce(182), // +, reduce: AddsubdotTerm
			reduce(182), // -, reduce: AddsubdotTerm
			reduce(182), // .+, reduce: AddsubdotTerm
			reduce(182), // .-, reduce: AddsubdotTerm
			shift(2483), // *
			shift(2484), // /
			shift(2485), // //
			shift(2486), // %
			shift(2487), // .*
			shift(2488), // ./
			shift(2489), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3254
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(183), // >, reduce: AddsubdotTerm
			reduce(183), // >>, reduce: AddsubdotTerm
			reduce(183), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(183), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(183), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(183), // ||, reduce: AddsubdotTerm
			reduce(183), // ^^, reduce: AddsubdotTerm
			reduce(183), // &&, reduce: AddsubdotTerm
			reduce(183), // ??, reduce: AddsubdotTerm
			reduce(183), // ???, reduce: AddsubdotTerm
			reduce(183), // =~, reduce: AddsubdotTerm
			reduce(183), // !=~, reduce: AddsubdotTerm
			reduce(183), // ==, reduce: AddsubdotTerm
			reduce(183), // !=, reduce: AddsubdotTerm
			reduce(183), // <=>, reduce: AddsubdotTerm
			reduce(183), // >=, reduce: AddsubdotTerm
			reduce(183), // <, reduce: AddsubdotTerm
			reduce(183), // <=, reduce: AddsubdotTerm
			reduce(183), // ^, reduce: AddsubdotTerm
			reduce(183), // &, reduce: AddsubdotTerm
			reduce(183), // <<, reduce: AddsubdotTerm
			reduce(183), // >>>, reduce: AddsubdotTerm
			reduce(183), // +, reduce: AddsubdotTerm
			reduce(183), // -, reduce: AddsubdotTerm
			reduce(183), // .+, reduce: AddsubdotTerm
			reduce(183), // .-, reduce: AddsubdotTerm
			shift(2483), // *
			shift(2484), // /
			shift(2485), // //
			shift(2486), // %
			shift(2487), // .*
			shift(2488), // ./
			shift(2489), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3255
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(184), // >, reduce: AddsubdotTerm
			reduce(184), // >>, reduce: AddsubdotTerm
			reduce(184), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(184), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(184), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(184), // ||, reduce: AddsubdotTerm
			reduce(184), // ^^, reduce: AddsubdotTerm
			reduce(184), // &&, reduce: AddsubdotTerm
			reduce(184), // ??, reduce: AddsubdotTerm
			reduce(184), // ???, reduce: AddsubdotTerm
			reduce(184), // =~, reduce: AddsubdotTerm
			reduce(184), // !=~, reduce: AddsubdotTerm
			reduce(184), // ==, reduce: AddsubdotTerm
			reduce(184), // !=, reduce: AddsubdotTerm
			reduce(184), // <=>, reduce: AddsubdotTerm
			reduce(184), // >=, reduce: AddsubdotTerm
			reduce(184), // <, reduce: AddsubdotTerm
			reduce(184), // <=, reduce: AddsubdotTerm
			reduce(184), // ^, reduce: AddsubdotTerm
			reduce(184), // &, reduce: AddsubdotTerm
			reduce(184), // <<, reduce: AddsubdotTerm
			reduce(184), // >>>, reduce: AddsubdotTerm
			reduce(184), // +, reduce: AddsubdotTerm
			reduce(184), // -, reduce: AddsubdotTerm
			reduce(184), // .+, reduce: AddsubdotTerm
			reduce(184), // .-, reduce: AddsubdotTerm
			shift(2483), // *
			shift(2484), // /
			shift(2485), // //
			shift(2486), // %
			shift(2487), // .*
			shift(2488), // ./
			shift(2489), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3256
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(185), // >, reduce: AddsubdotTerm
			reduce(185), // >>, reduce: AddsubdotTerm
			reduce(185), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(185), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(185), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(185), // ||, reduce: AddsubdotTerm
			reduce(185), // ^^, reduce: AddsubdotTerm
			reduce(185), // &&, reduce: AddsubdotTerm
			reduce(185), // ??, reduce: AddsubdotTerm
			reduce(185), // ???, reduce: AddsubdotTerm
			reduce(185), // =~, reduce: AddsubdotTerm
			reduce(185), // !=~, reduce: AddsubdotTerm
			reduce(185), // ==, reduce: AddsubdotTerm
			reduce(185), // !=, reduce: AddsubdotTerm
			reduce(185), // <=>, reduce: AddsubdotTerm
			reduce(185), // >=, reduce: AddsubdotTerm
			reduce(185), // <, reduce: AddsubdotTerm
			reduce(185), // <=, reduce: AddsubdotTerm
			reduce(185), // ^, reduce: AddsubdotTerm
			reduce(185), // &, reduce: AddsubdotTerm
			reduce(185), // <<, reduce: AddsubdotTerm
			reduce(185), // >>>, reduce: AddsubdotTerm
			reduce(185), // +, reduce: AddsubdotTerm
			reduce(185), // -, reduce: AddsubdotTerm
			reduce(185), // .+, reduce: AddsubdotTerm
			reduce(185), // .-, reduce: AddsubdotTerm
			shift(2483), // *
			shift(2484), // /
			shift(2485), // //
			shift(2486), // %
			shift(2487), // .*
			shift(2488), // ./
			shift(2489), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3257
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(187), // >, reduce: MuldivTerm
			reduce(187), // >>, reduce: MuldivTerm
			reduce(187), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(187), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(187), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(187), // ||, reduce: MuldivTerm
			reduce(187), // ^^, reduce: MuldivTerm
			reduce(187), // &&, reduce: MuldivTerm
			reduce(187), // ??, reduce: MuldivTerm
			reduce(187), // ???, reduce: MuldivTerm
			reduce(187), // =~, reduce: MuldivTerm
			reduce(187), // !=~, reduce: MuldivTerm
			reduce(187), // ==, reduce: MuldivTerm
			reduce(187), // !=, reduce: MuldivTerm
			reduce(187), // <=>, reduce: MuldivTerm
			reduce(187), // >=, reduce: MuldivTerm
			reduce(187), // <, reduce: MuldivTerm
			reduce(187), // <=, reduce: MuldivTerm
			reduce(187), // ^, reduce: MuldivTerm
			reduce(187), // &, reduce: MuldivTerm
			reduce(187), // <<, reduce: MuldivTerm
			reduce(187), // >>>, reduce: MuldivTerm
			reduce(187), // +, reduce: MuldivTerm
			reduce(187), // -, reduce: MuldivTerm
			reduce(187), // .+, reduce: MuldivTerm
			reduce(187), // .-, reduce: MuldivTerm
			reduce(187), // *, reduce: MuldivTerm
			reduce(187), // /, reduce: MuldivTerm
			reduce(187), // //, reduce: MuldivTerm
			reduce(187), // %, reduce: MuldivTerm
			reduce(187), // .*, reduce: MuldivTerm
			reduce(187), // ./, reduce: MuldivTerm
			reduce(187), // .//, reduce: MuldivTerm
			shift(2493), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3258
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(188), // >, reduce: MuldivTerm
			reduce(188), // >>, reduce: MuldivTerm
			reduce(188), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(188), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(188), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(188), // ||, reduce: MuldivTerm
			reduce(188), // ^^, reduce: MuldivTerm
			reduce(188), // &&, reduce: MuldivTerm
			reduce(188), // ??, reduce: MuldivTerm
			reduce(188), // ???, reduce: MuldivTerm
			reduce(188), // =~, reduce: MuldivTerm
			reduce(188), // !=~, reduce: MuldivTerm
			reduce(188), // ==, reduce: MuldivTerm
			reduce(188), // !=, reduce: MuldivTerm
			reduce(188), // <=>, reduce: MuldivTerm
			reduce(188), // >=, reduce: MuldivTerm
			reduce(188), // <, reduce: MuldivTerm
			reduce(188), // <=, reduce: MuldivTerm
			reduce(188), // ^, reduce: MuldivTerm
			reduce(188), // &, reduce: MuldivTerm
			reduce(188), // <<, reduce: MuldivTerm
			reduce(188), // >>>, reduce: MuldivTerm
			reduce(188), // +, reduce: MuldivTerm
			reduce(188), // -, reduce: MuldivTerm
			reduce(188), // .+, reduce: MuldivTerm
			reduce(188), // .-, reduce: MuldivTerm
			reduce(188), // *, reduce: MuldivTerm
			reduce(188), // /, reduce: MuldivTerm
			reduce(188), // //, reduce: MuldivTerm
			reduce(188), // %, reduce: MuldivTerm
			reduce(188), // .*, reduce: MuldivTerm
			reduce(188), // ./, reduce: MuldivTerm
			reduce(188), // .//, reduce: MuldivTerm
			shift(2493), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3259
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(189), // >, reduce: MuldivTerm
			reduce(189), // >>, reduce: MuldivTerm
			reduce(189), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(189), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(189), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(189), // ||, reduce: MuldivTerm
			reduce(189), // ^^, reduce: MuldivTerm
			reduce(189), // &&, reduce: MuldivTerm
			reduce(189), // ??, reduce: MuldivTerm
			reduce(189), // ???, reduce: MuldivTerm
			reduce(189), // =~, reduce: MuldivTerm
			reduce(189), // !=~, reduce: MuldivTerm
			reduce(189), // ==, reduce: MuldivTerm
			reduce(189), // !=, reduce: MuldivTerm
			reduce(189), // <=>, reduce: MuldivTerm
			reduce(189), // >=, reduce: MuldivTerm
			reduce(189), // <, reduce: MuldivTerm
			reduce(189), // <=, reduce: MuldivTerm
			reduce(189), // ^, reduce: MuldivTerm
			reduce(189), // &, reduce: MuldivTerm
			reduce(189), // <<, reduce: MuldivTerm
			reduce(189), // >>>, reduce: MuldivTerm
			reduce(189), // +, reduce: MuldivTerm
			reduce(189), // -, reduce: MuldivTerm
			reduce(189), // .+, reduce: MuldivTerm
			reduce(189), // .-, reduce: MuldivTerm
			reduce(189), // *, reduce: MuldivTerm
			reduce(189), // /, reduce: MuldivTerm
			reduce(189), // //, reduce: MuldivTerm
			reduce(189), // %, reduce: MuldivTerm
			reduce(189), // .*, reduce: MuldivTerm
			reduce(189), // ./, reduce: MuldivTerm
			reduce(189), // .//, reduce: MuldivTerm
			shift(2493), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3260
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(190), // >, reduce: MuldivTerm
			reduce(190), // >>, reduce: MuldivTerm
			reduce(190), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(190), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(190), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(190), // ||, reduce: MuldivTerm
			reduce(190), // ^^, reduce: MuldivTerm
			reduce(190), // &&, reduce: MuldivTerm
			reduce(190), // ??, reduce: MuldivTerm
			reduce(190), // ???, reduce: MuldivTerm
			reduce(190), // =~, reduce: MuldivTerm
			reduce(190), // !=~, reduce: MuldivTerm
			reduce(190), // ==, reduce: MuldivTerm
			reduce(190), // !=, reduce: MuldivTerm
			reduce(190), // <=>, reduce: MuldivTerm
			reduce(190), // >=, reduce: MuldivTerm
			reduce(190), // <, reduce: MuldivTerm
			reduce(190), // <=, reduce: MuldivTerm
			reduce(190), // ^, reduce: MuldivTerm
			reduce(190), // &, reduce: MuldivTerm
			reduce(190), // <<, reduce: MuldivTerm
			reduce(190), // >>>, reduce: MuldivTerm
			reduce(190), // +, reduce: MuldivTerm
			reduce(190), // -, reduce: MuldivTerm
			reduce(190), // .+, reduce: MuldivTerm
			reduce(190), // .-, reduce: MuldivTerm
			reduce(190), // *, reduce: MuldivTerm
			reduce(190), // /, reduce: MuldivTerm
			reduce(190), // //, reduce: MuldivTerm
			reduce(190), // %, reduce: MuldivTerm
			reduce(190), // .*, reduce: MuldivTerm
			reduce(190), // ./, reduce: MuldivTerm
			reduce(190), // .//, reduce: MuldivTerm
			shift(2493), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3261
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(191), // >, reduce: MuldivTerm
			reduce(191), // >>, reduce: MuldivTerm
			reduce(191), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(191), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(191), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(191), // ||, reduce: MuldivTerm
			reduce(191), // ^^, reduce: MuldivTerm
			reduce(191), // &&, reduce: MuldivTerm
			reduce(191), // ??, reduce: MuldivTerm
			reduce(191), // ???, reduce: MuldivTerm
			reduce(191), // =~, reduce: MuldivTerm
			reduce(191), // !=~, reduce: MuldivTerm
			reduce(191), // ==, reduce: MuldivTerm
			reduce(191), // !=, reduce: MuldivTerm
			reduce(191), // <=>, reduce: MuldivTerm
			reduce(191), // >=, reduce: MuldivTerm
			reduce(191), // <, reduce: MuldivTerm
			reduce(191), // <=, reduce: MuldivTerm
			reduce(191), // ^, reduce: MuldivTerm
			reduce(191), // &, reduce: MuldivTerm
			reduce(191), // <<, reduce: MuldivTerm
			reduce(191), // >>>, reduce: MuldivTerm
			reduce(191), // +, reduce: MuldivTerm
			reduce(191), // -, reduce: MuldivTerm
			reduce(191), // .+, reduce: MuldivTerm
			reduce(191), // .-, reduce: MuldivTerm
			reduce(191), // *, reduce: MuldivTerm
			reduce(191), // /, reduce: MuldivTerm
			reduce(191), // //, reduce: MuldivTerm
			reduce(191), // %, reduce: MuldivTerm
			reduce(191), // .*, reduce: MuldivTerm
			reduce(191), // ./, reduce: MuldivTerm
			reduce(191), // .//, reduce: MuldivTerm
			shift(2493), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3262
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(192), // >, reduce: MuldivTerm
			reduce(192), // >>, reduce: MuldivTerm
			reduce(192), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(192), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(192), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(192), // ||, reduce: MuldivTerm
			reduce(192), // ^^, reduce: MuldivTerm
			reduce(192), // &&, reduce: MuldivTerm
			reduce(192), // ??, reduce: MuldivTerm
			reduce(192), // ???, reduce: MuldivTerm
			reduce(192), // =~, reduce: MuldivTerm
			reduce(192), // !=~, reduce: MuldivTerm
			reduce(192), // ==, reduce: MuldivTerm
			reduce(192), // !=, reduce: MuldivTerm
			reduce(192), // <=>, reduce: MuldivTerm
			reduce(192), // >=, reduce: MuldivTerm
			reduce(192), // <, reduce: MuldivTerm
			reduce(192), // <=, reduce: MuldivTerm
			reduce(192), // ^, reduce: MuldivTerm
			reduce(192), // &, reduce: MuldivTerm
			reduce(192), // <<, reduce: MuldivTerm
			reduce(192), // >>>, reduce: MuldivTerm
			reduce(192), // +, reduce: MuldivTerm
			reduce(192), // -, reduce: MuldivTerm
			reduce(192), // .+, reduce: MuldivTerm
			reduce(192), // .-, reduce: MuldivTerm
			reduce(192), // *, reduce: MuldivTerm
			reduce(192), // /, reduce: MuldivTerm
			reduce(192), // //, reduce: MuldivTerm
			reduce(192), // %, reduce: MuldivTerm
			reduce(192), // .*, reduce: MuldivTerm
			reduce(192), // ./, reduce: MuldivTerm
			reduce(192), // .//, reduce: MuldivTerm
			shift(2493), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3263
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(193), // >, reduce: MuldivTerm
			reduce(193), // >>, reduce: MuldivTerm
			reduce(193), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(193), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(193), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(193), // ||, reduce: MuldivTerm
			reduce(193), // ^^, reduce: MuldivTerm
			reduce(193), // &&, reduce: MuldivTerm
			reduce(193), // ??, reduce: MuldivTerm
			reduce(193), // ???, reduce: MuldivTerm
			reduce(193), // =~, reduce: MuldivTerm
			reduce(193), // !=~, reduce: MuldivTerm
			reduce(193), // ==, reduce: MuldivTerm
			reduce(193), // !=, reduce: MuldivTerm
			reduce(193), // <=>, reduce: MuldivTerm
			reduce(193), // >=, reduce: MuldivTerm
			reduce(193), // <, reduce: MuldivTerm
			reduce(193), // <=, reduce: MuldivTerm
			reduce(193), // ^, reduce: MuldivTerm
			reduce(193), // &, reduce: MuldivTerm
			reduce(193), // <<, reduce: MuldivTerm
			reduce(193), // >>>, reduce: MuldivTerm
			reduce(193), // +, reduce: MuldivTerm
			reduce(193), // -, reduce: MuldivTerm
			reduce(193), // .+, reduce: MuldivTerm
			reduce(193), // .-, reduce: MuldivTerm
			reduce(193), // *, reduce: MuldivTerm
			reduce(193), // /, reduce: MuldivTerm
			reduce(193), // //, reduce: MuldivTerm
			reduce(193), // %, reduce: MuldivTerm
			reduce(193), // .*, reduce: MuldivTerm
			reduce(193), // ./, reduce: MuldivTerm
			reduce(193), // .//, reduce: MuldivTerm
			shift(2493), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3264
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(195), // >, reduce: DotTerm
			reduce(195), // >>, reduce: DotTerm
			reduce(195), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(195), // ,, reduce: DotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(195), // ?, reduce: DotTerm
			nil,         // :
			reduce(195), // ||, reduce: DotTerm
			reduce(195), // ^^, reduce: DotTerm
			reduce(195), // &&, reduce: DotTerm
			reduce(195), // ??, reduce: DotTerm
			reduce(195), // ???, reduce: DotTerm
			reduce(195), // =~, reduce: DotTerm
			reduce(195), // !=~, reduce: DotTerm
			reduce(195), // ==, reduce: DotTerm
			reduce(195), // !=, reduce: DotTerm
			reduce(195), // <=>, reduce: DotTerm
			reduce(195), // >=, reduce: DotTerm
			reduce(195), // <, reduce: DotTerm
			reduce(195), // <=, reduce: DotTerm
			reduce(195), // ^, reduce: DotTerm
			reduce(195), // &, reduce: DotTerm
			reduce(195), // <<, reduce: DotTerm
			reduce(195), // >>>, reduce: DotTerm
			reduce(195), // +, reduce: DotTerm
			reduce(195), // -, reduce: DotTerm
			reduce(195), // .+, reduce: DotTerm
			reduce(195), // .-, reduce: DotTerm
			reduce(195), // *, reduce: DotTerm
			reduce(195), // /, reduce: DotTerm
			reduce(195), // //, reduce: DotTerm
			reduce(195), // %, reduce: DotTerm
			reduce(195), // .*, reduce: DotTerm
			reduce(195), // ./, reduce: DotTerm
			reduce(195), // .//, reduce: DotTerm
			reduce(195), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3265
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3266
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3267
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(204), // >, reduce: PowTerm
			reduce(204), // >>, reduce: PowTerm
			reduce(204), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(204), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(204), // ?, reduce: PowTerm
			nil,         // :
			reduce(204), // ||, reduce: PowTerm
			reduce(204), // ^^, reduce: PowTerm
			reduce(204), // &&, reduce: PowTerm
			reduce(204), // ??, reduce: PowTerm
			reduce(204), // ???, reduce: PowTerm
			reduce(204), // =~, reduce: PowTerm
			reduce(204), // !=~, reduce: PowTerm
			reduce(204), // ==, reduce: PowTerm
			reduce(204), // !=, reduce: PowTerm
			reduce(204), // <=>, reduce: PowTerm
			reduce(204), // >=, reduce: PowTerm
			reduce(204), // <, reduce: PowTerm
			reduce(204), // <=, reduce: PowTerm
			reduce(204), // ^, reduce: PowTerm
			reduce(204), // &, reduce: PowTerm
			reduce(204), // <<, reduce: PowTerm
			reduce(204), // >>>, reduce: PowTerm
			reduce(204), // +, reduce: PowTerm
			reduce(204), // -, reduce: PowTerm
			reduce(204), // .+, reduce: PowTerm
			reduce(204), // .-, reduce: PowTerm
			reduce(204), // *, reduce: PowTerm
			reduce(204), // /, reduce: PowTerm
			reduce(204), // //, reduce: PowTerm
			reduce(204), // %, reduce: PowTerm
			reduce(204), // .*, reduce: PowTerm
			reduce(204), // ./, reduce: PowTerm
			reduce(204), // .//, reduce: PowTerm
			reduce(204), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3268
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3958), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3959), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3269
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(3961), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3270
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3962), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3271
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3963), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3272
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(229), // >, reduce: ArrayLiteral
			reduce(229), // >>, reduce: ArrayLiteral
			reduce(229), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(229), // ,, reduce: ArrayLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(229), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(229), // ||, reduce: ArrayLiteral
			reduce(229), // ^^, reduce: ArrayLiteral
			reduce(229), // &&, reduce: ArrayLiteral
			reduce(229), // ??, reduce: ArrayLiteral
			reduce(229), // ???, reduce: ArrayLiteral
			reduce(229), // =~, reduce: ArrayLiteral
			reduce(229), // !=~, reduce: ArrayLiteral
			reduce(229), // ==, reduce: ArrayLiteral
			reduce(229), // !=, reduce: ArrayLiteral
			reduce(229), // <=>, reduce: ArrayLiteral
			reduce(229), // >=, reduce: ArrayLiteral
			reduce(229), // <, reduce: ArrayLiteral
			reduce(229), // <=, reduce: ArrayLiteral
			reduce(229), // ^, reduce: ArrayLiteral
			reduce(229), // &, reduce: ArrayLiteral
			reduce(229), // <<, reduce: ArrayLiteral
			reduce(229), // >>>, reduce: ArrayLiteral
			reduce(229), // +, reduce: ArrayLiteral
			reduce(229), // -, reduce: ArrayLiteral
			reduce(229), // .+, reduce: ArrayLiteral
			reduce(229), // .-, reduce: ArrayLiteral
			reduce(229), // *, reduce: ArrayLiteral
			reduce(229), // /, reduce: ArrayLiteral
			reduce(229), // //, reduce: ArrayLiteral
			reduce(229), // %, reduce: ArrayLiteral
			reduce(229), // .*, reduce: ArrayLiteral
			reduce(229), // ./, reduce: ArrayLiteral
			reduce(229), // .//, reduce: ArrayLiteral
			reduce(229), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(229), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(229), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(229), // [[, reduce: ArrayLiteral
			reduce(229), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3273
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(255), // >, reduce: ENV
			reduce(255), // >>, reduce: ENV
			reduce(255), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(255), // ,, reduce: ENV
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(255), // ?, reduce: ENV
			nil,         // :
			reduce(255), // ||, reduce: ENV
			reduce(255), // ^^, reduce: ENV
			reduce(255), // &&, reduce: ENV
			reduce(255), // ??, reduce: ENV
			reduce(255), // ???, reduce: ENV
			reduce(255), // =~, reduce: ENV
			reduce(255), // !=~, reduce: ENV
			reduce(255), // ==, reduce: ENV
			reduce(255), // !=, reduce: ENV
			reduce(255), // <=>, reduce: ENV
			reduce(255), // >=, reduce: ENV
			reduce(255), // <, reduce: ENV
			reduce(255), // <=, reduce: ENV
			reduce(255), // ^, reduce: ENV
			reduce(255), // &, reduce: ENV
			reduce(255), // <<, reduce: ENV
			reduce(255), // >>>, reduce: ENV
			reduce(255), // +, reduce: ENV
			reduce(255), // -, reduce: ENV
			reduce(255), // .+, reduce: ENV
			reduce(255), // .-, reduce: ENV
			reduce(255), // *, reduce: ENV
			reduce(255), // /, reduce: ENV
			reduce(255), // //, reduce: ENV
			reduce(255), // %, reduce: ENV
			reduce(255), // .*, reduce: ENV
			reduce(255), // ./, reduce: ENV
			reduce(255), // .//, reduce: ENV
			reduce(255), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(255), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(255), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(255), // [[, reduce: ENV
			reduce(255), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3274
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3964), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3275
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3965), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3276
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(268), // >, reduce: FunctionCallsite
			reduce(268), // >>, reduce: FunctionCallsite
			reduce(268), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(268), // ,, reduce: FunctionCallsite
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(268), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(268), // ||, reduce: FunctionCallsite
			reduce(268), // ^^, reduce: FunctionCallsite
			reduce(268), // &&, reduce: FunctionCallsite
			reduce(268), // ??, reduce: FunctionCallsite
			reduce(268), // ???, reduce: FunctionCallsite
			reduce(268), // =~, reduce: FunctionCallsite
			reduce(268), // !=~, reduce: FunctionCallsite
			reduce(268), // ==, reduce: FunctionCallsite
			reduce(268), // !=, reduce: FunctionCallsite
			reduce(268), // <=>, reduce: FunctionCallsite
			reduce(268), // >=, reduce: FunctionCallsite
			reduce(268), // <, reduce: FunctionCallsite
			reduce(268), // <=, reduce: FunctionCallsite
			reduce(268), // ^, reduce: FunctionCallsite
			reduce(268), // &, reduce: FunctionCallsite
			reduce(268), // <<, reduce: FunctionCallsite
			reduce(268), // >>>, reduce: FunctionCallsite
			reduce(268), // +, reduce: FunctionCallsite
			reduce(268), // -, reduce: FunctionCallsite
			reduce(268), // .+, reduce: FunctionCallsite
			reduce(268), // .-, reduce: FunctionCallsite
			reduce(268), // *, reduce: FunctionCallsite
			reduce(268), // /, reduce: FunctionCallsite
			reduce(268), // //, reduce: FunctionCallsite
			reduce(268), // %, reduce: FunctionCallsite
			reduce(268), // .*, reduce: FunctionCallsite
			reduce(268), // ./, reduce: FunctionCallsite
			reduce(268), // .//, reduce: FunctionCallsite
			reduce(268), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(268), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(268), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(268), // [[, reduce: FunctionCallsite
			reduce(268), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3277
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3966), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3278
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3279
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3968), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3280
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3281
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(235), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(235), // ), reduce: MapLiteral
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3282
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3283
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			reduce(85), // ), reduce: EmittableList
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3284
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(105), // ,, reduce: IndirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(105), // ), reduce: IndirectFieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3285
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(114), // ,, reduce: IndirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(114), // ), reduce: IndirectOosvarValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3286
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(107), // >, reduce: PositionalFieldName
			reduce(107), // >>, reduce: PositionalFieldName
			reduce(107), // |, reduce: PositionalFieldName
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(107), // ), reduce: PositionalFieldName
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(107), // ?, reduce: PositionalFieldName
			nil,         // :
			reduce(107), // ||, reduce: PositionalFieldName
			reduce(107), // ^^, reduce: PositionalFieldName
			reduce(107), // &&, reduce: PositionalFieldName
			reduce(107), // ??, reduce: PositionalFieldName
			reduce(107), // ???, reduce: PositionalFieldName
			reduce(107), // =~, reduce: PositionalFieldName
			reduce(107), // !=~, reduce: PositionalFieldName
			reduce(107), // ==, reduce: PositionalFieldName
			reduce(107), // !=, reduce: PositionalFieldName
			reduce(107), // <=>, reduce: PositionalFieldName
			reduce(107), // >=, reduce: PositionalFieldName
			reduce(107), // <, reduce: PositionalFieldName
			reduce(107), // <=, reduce: PositionalFieldName
			reduce(107), // ^, reduce: PositionalFieldName
			reduce(107), // &, reduce: PositionalFieldName
			reduce(107), // <<, reduce: PositionalFieldName
			reduce(107), // >>>, reduce: PositionalFieldName
			reduce(107), // +, reduce: PositionalFieldName
			reduce(107), // -, reduce: PositionalFieldName
			reduce(107), // .+, reduce: PositionalFieldName
			reduce(107), // .-, reduce: PositionalFieldName
			reduce(107), // *, reduce: PositionalFieldName
			reduce(107), // /, reduce: PositionalFieldName
			reduce(107), // //, reduce: PositionalFieldName
			reduce(107), // %, reduce: PositionalFieldName
			reduce(107), // .*, reduce: PositionalFieldName
			reduce(107), // ./, reduce: PositionalFieldName
			reduce(107), // .//, reduce: PositionalFieldName
			reduce(107), // ., reduce: PositionalFieldName
			nil,         // !
			nil,         // ~
			reduce(107), // **, reduce: PositionalFieldName
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(107), // [, reduce: PositionalFieldName
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(107), // [[, reduce: PositionalFieldName
			reduce(107), // [[[, reduce: PositionalFieldName
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3287
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3971), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3288
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3289
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(206), // >, reduce: PowTerm
			reduce(206), // >>, reduce: PowTerm
			reduce(206), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(206), // ), reduce: PowTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(206), // ?, reduce: PowTerm
			nil,         // :
			reduce(206), // ||, reduce: PowTerm
			reduce(206), // ^^, reduce: PowTerm
			reduce(206), // &&, reduce: PowTerm
			reduce(206), // ??, reduce: PowTerm
			reduce(206), // ???, reduce: PowTerm
			reduce(206), // =~, reduce: PowTerm
			reduce(206), // !=~, reduce: PowTerm
			reduce(206), // ==, reduce: PowTerm
			reduce(206), // !=, reduce: PowTerm
			reduce(206), // <=>, reduce: PowTerm
			reduce(206), // >=, reduce: PowTerm
			reduce(206), // <, reduce: PowTerm
			reduce(206), // <=, reduce: PowTerm
			reduce(206), // ^, reduce: PowTerm
			reduce(206), // &, reduce: PowTerm
			reduce(206), // <<, reduce: PowTerm
			reduce(206), // >>>, reduce: PowTerm
			reduce(206), // +, reduce: PowTerm
			reduce(206), // -, reduce: PowTerm
			reduce(206), // .+, reduce: PowTerm
			reduce(206), // .-, reduce: PowTerm
			reduce(206), // *, reduce: PowTerm
			reduce(206), // /, reduce: PowTerm
			reduce(206), // //, reduce: PowTerm
			reduce(206), // %, reduce: PowTerm
			reduce(206), // .*, reduce: PowTerm
			reduce(206), // ./, reduce: PowTerm
			reduce(206), // .//, reduce: PowTerm
			reduce(206), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3290
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(205), // >, reduce: PowTerm
			reduce(205), // >>, reduce: PowTerm
			reduce(205), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(205), // ), reduce: PowTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(205), // ?, reduce: PowTerm
			nil,         // :
			reduce(205), // ||, reduce: PowTerm
			reduce(205), // ^^, reduce: PowTerm
			reduce(205), // &&, reduce: PowTerm
			reduce(205), // ??, reduce: PowTerm
			reduce(205), // ???, reduce: PowTerm
			reduce(205), // =~, reduce: PowTerm
			reduce(205), // !=~, reduce: PowTerm
			reduce(205), // ==, reduce: PowTerm
			reduce(205), // !=, reduce: PowTerm
			reduce(205), // <=>, reduce: PowTerm
			reduce(205), // >=, reduce: PowTerm
			reduce(205), // <, reduce: PowTerm
			reduce(205), // <=, reduce: PowTerm
			reduce(205), // ^, reduce: PowTerm
			reduce(205), // &, reduce: PowTerm
			reduce(205), // <<, reduce: PowTerm
			reduce(205), // >>>, reduce: PowTerm
			reduce(205), // +, reduce: PowTerm
			reduce(205), // -, reduce: PowTerm
			reduce(205), // .+, reduce: PowTerm
			reduce(205), // .-, reduce: PowTerm
			reduce(205), // *, reduce: PowTerm
			reduce(205), // /, reduce: PowTerm
			reduce(205), // //, reduce: PowTerm
			reduce(205), // %, reduce: PowTerm
			reduce(205), // .*, reduce: PowTerm
			reduce(205), // ./, reduce: PowTerm
			reduce(205), // .//, reduce: PowTerm
			reduce(205), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3291
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(260), // >, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>, reduce: ArrayOrMapIndexAccess
			reduce(260), // |, reduce: ArrayOrMapIndexAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(260), // ), reduce: ArrayOrMapIndexAccess
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(260), // ?, reduce: ArrayOrMapIndexAccess
			nil,         // :
			reduce(260), // ||, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???, reduce: ArrayOrMapIndexAccess
			reduce(260), // =~, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=~, reduce: ArrayOrMapIndexAccess
			reduce(260), // ==, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=>, reduce: ArrayOrMapIndexAccess
			reduce(260), // >=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>, reduce: ArrayOrMapIndexAccess
			reduce(260), // +, reduce: ArrayOrMapIndexAccess
			reduce(260), // -, reduce: ArrayOrMapIndexAccess
			reduce(260), // .+, reduce: ArrayOrMapIndexAccess
			reduce(260), // .-, reduce: ArrayOrMapIndexAccess
			reduce(260), // *, reduce: ArrayOrMapIndexAccess
			reduce(260), // /, reduce: ArrayOrMapIndexAccess
			reduce(260), // //, reduce: ArrayOrMapIndexAccess
			reduce(260), // %, reduce: ArrayOrMapIndexAccess
			reduce(260), // .*, reduce: ArrayOrMapIndexAccess
			reduce(260), // ./, reduce: ArrayOrMapIndexAccess
			reduce(260), // .//, reduce: ArrayOrMapIndexAccess
			reduce(260), // ., reduce: ArrayOrMapIndexAccess
			nil,         // !
			nil,         // ~
			reduce(260), // **, reduce: ArrayOrMapIndexAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(260), // [, reduce: ArrayOrMapIndexAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(260), // [[, reduce: ArrayOrMapIndexAccess
			reduce(260), // [[[, reduce: ArrayOrMapIndexAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3292
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(3974), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3293
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3975), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3294
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(266), // >, reduce: ArraySliceAccess
			reduce(266), // >>, reduce: ArraySliceAccess
			reduce(266), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(266), // ), reduce: ArraySliceAccess
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(266), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(266), // ||, reduce: ArraySliceAccess
			reduce(266), // ^^, reduce: ArraySliceAccess
			reduce(266), // &&, reduce: ArraySliceAccess
			reduce(266), // ??, reduce: ArraySliceAccess
			reduce(266), // ???, reduce: ArraySliceAccess
			reduce(266), // =~, reduce: ArraySliceAccess
			reduce(266), // !=~, reduce: ArraySliceAccess
			reduce(266), // ==, reduce: ArraySliceAccess
			reduce(266), // !=, reduce: ArraySliceAccess
			reduce(266), // <=>, reduce: ArraySliceAccess
			reduce(266), // >=, reduce: ArraySliceAccess
			reduce(266), // <, reduce: ArraySliceAccess
			reduce(266), // <=, reduce: ArraySliceAccess
			reduce(266), // ^, reduce: ArraySliceAccess
			reduce(266), // &, reduce: ArraySliceAccess
			reduce(266), // <<, reduce: ArraySliceAccess
			reduce(266), // >>>, reduce: ArraySliceAccess
			reduce(266), // +, reduce: ArraySliceAccess
			reduce(266), // -, reduce: ArraySliceAccess
			reduce(266), // .+, reduce: ArraySliceAccess
			reduce(266), // .-, reduce: ArraySliceAccess
			reduce(266), // *, reduce: ArraySliceAccess
			reduce(266), // /, reduce: ArraySliceAccess
			reduce(266), // //, reduce: ArraySliceAccess
			reduce(266), // %, reduce: ArraySliceAccess
			reduce(266), // .*, reduce: ArraySliceAccess
			reduce(266), // ./, reduce: ArraySliceAccess
			reduce(266), // .//, reduce: ArraySliceAccess
			reduce(266), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(266), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(266), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(266), // [[, reduce: ArraySliceAccess
			reduce(266), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3295
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3976), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3296
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3977), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3297
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(254), // >, reduce: ENV
			reduce(254), // >>, reduce: ENV
			reduce(254), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(254), // ), reduce: ENV
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(254), // ?, reduce: ENV
			nil,         // :
			reduce(254), // ||, reduce: ENV
			reduce(254), // ^^, reduce: ENV
			reduce(254), // &&, reduce: ENV
			reduce(254), // ??, reduce: ENV
			reduce(254), // ???, reduce: ENV
			reduce(254), // =~, reduce: ENV
			reduce(254), // !=~, reduce: ENV
			reduce(254), // ==, reduce: ENV
			reduce(254), // !=, reduce: ENV
			reduce(254), // <=>, reduce: ENV
			reduce(254), // >=, reduce: ENV
			reduce(254), // <, reduce: ENV
			reduce(254), // <=, reduce: ENV
			reduce(254), // ^, reduce: ENV
			reduce(254), // &, reduce: ENV
			reduce(254), // <<, reduce: ENV
			reduce(254), // >>>, reduce: ENV
			reduce(254), // +, reduce: ENV
			reduce(254), // -, reduce: ENV
			reduce(254), // .+, reduce: ENV
			reduce(254), // .-, reduce: ENV
			reduce(254), // *, reduce: ENV
			reduce(254), // /, reduce: ENV
			reduce(254), // //, reduce: ENV
			reduce(254), // %, reduce: ENV
			reduce(254), // .*, reduce: ENV
			reduce(254), // ./, reduce: ENV
			reduce(254), // .//, reduce: ENV
			reduce(254), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(254), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(254), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(254), // [[, reduce: ENV
			reduce(254), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3298
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(269), // >, reduce: FunctionCallsite
			reduce(269), // >>, reduce: FunctionCallsite
			reduce(269), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(269), // ), reduce: FunctionCallsite
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(269), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(269), // ||, reduce: FunctionCallsite
			reduce(269), // ^^, reduce: FunctionCallsite
			reduce(269), // &&, reduce: FunctionCallsite
			reduce(269), // ??, reduce: FunctionCallsite
			reduce(269), // ???, reduce: FunctionCallsite
			reduce(269), // =~, reduce: FunctionCallsite
			reduce(269), // !=~, reduce: FunctionCallsite
			reduce(269), // ==, reduce: FunctionCallsite
			reduce(269), // !=, reduce: FunctionCallsite
			reduce(269), // <=>, reduce: FunctionCallsite
			reduce(269), // >=, reduce: FunctionCallsite
			reduce(269), // <, reduce: FunctionCallsite
			reduce(269), // <=, reduce: FunctionCallsite
			reduce(269), // ^, reduce: FunctionCallsite
			reduce(269), // &, reduce: FunctionCallsite
			reduce(269), // <<, reduce: FunctionCallsite
			reduce(269), // >>>, reduce: FunctionCallsite
			reduce(269), // +, reduce: FunctionCallsite
			reduce(269), // -, reduce: FunctionCallsite
			reduce(269), // .+, reduce: FunctionCallsite
			reduce(269), // .-, reduce: FunctionCallsite
			reduce(269), // *, reduce: FunctionCallsite
			reduce(269), // /, reduce: FunctionCallsite
			reduce(269), // //, reduce: FunctionCallsite
			reduce(269), // %, reduce: FunctionCallsite
			reduce(269), // .*, reduce: FunctionCallsite
			reduce(269), // ./, reduce: FunctionCallsite
			reduce(269), // .//, reduce: FunctionCallsite
			reduce(269), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(269), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(269), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(269), // [[, reduce: FunctionCallsite
			reduce(269), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3299
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3979), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3980), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3300
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3301
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(3982), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3302
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3303
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(107), // >, reduce: PositionalFieldName
			reduce(107), // >>, reduce: PositionalFieldName
			reduce(107), // |, reduce: PositionalFieldName
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(107), // ], reduce: PositionalFieldName
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(107), // ?, reduce: PositionalFieldName
			nil,         // :
			reduce(107), // ||, reduce: PositionalFieldName
			reduce(107), // ^^, reduce: PositionalFieldName
			reduce(107), // &&, reduce: PositionalFieldName
			reduce(107), // ??, reduce: PositionalFieldName
			reduce(107), // ???, reduce: PositionalFieldName
			reduce(107), // =~, reduce: PositionalFieldName
			reduce(107), // !=~, reduce: PositionalFieldName
			reduce(107), // ==, reduce: PositionalFieldName
			reduce(107), // !=, reduce: PositionalFieldName
			reduce(107), // <=>, reduce: PositionalFieldName
			reduce(107), // >=, reduce: PositionalFieldName
			reduce(107), // <, reduce: PositionalFieldName
			reduce(107), // <=, reduce: PositionalFieldName
			reduce(107), // ^, reduce: PositionalFieldName
			reduce(107), // &, reduce: PositionalFieldName
			reduce(107), // <<, reduce: PositionalFieldName
			reduce(107), // >>>, reduce: PositionalFieldName
			reduce(107), // +, reduce: PositionalFieldName
			reduce(107), // -, reduce: PositionalFieldName
			reduce(107), // .+, reduce: PositionalFieldName
			reduce(107), // .-, reduce: PositionalFieldName
			reduce(107), // *, reduce: PositionalFieldName
			reduce(107), // /, reduce: PositionalFieldName
			reduce(107), // //, reduce: PositionalFieldName
			reduce(107), // %, reduce: PositionalFieldName
			reduce(107), // .*, reduce: PositionalFieldName
			reduce(107), // ./, reduce: PositionalFieldName
			reduce(107), // .//, reduce: PositionalFieldName
			reduce(107), // ., reduce: PositionalFieldName
			nil,         // !
			nil,         // ~
			reduce(107), // **, reduce: PositionalFieldName
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(107), // [, reduce: PositionalFieldName
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(107), // [[, reduce: PositionalFieldName
			reduce(107), // [[[, reduce: PositionalFieldName
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3304
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3984), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3305
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3306
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(206), // >, reduce: PowTerm
			reduce(206), // >>, reduce: PowTerm
			reduce(206), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(206), // ], reduce: PowTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(206), // ?, reduce: PowTerm
			nil,         // :
			reduce(206), // ||, reduce: PowTerm
			reduce(206), // ^^, reduce: PowTerm
			reduce(206), // &&, reduce: PowTerm
			reduce(206), // ??, reduce: PowTerm
			reduce(206), // ???, reduce: PowTerm
			reduce(206), // =~, reduce: PowTerm
			reduce(206), // !=~, reduce: PowTerm
			reduce(206), // ==, reduce: PowTerm
			reduce(206), // !=, reduce: PowTerm
			reduce(206), // <=>, reduce: PowTerm
			reduce(206), // >=, reduce: PowTerm
			reduce(206), // <, reduce: PowTerm
			reduce(206), // <=, reduce: PowTerm
			reduce(206), // ^, reduce: PowTerm
			reduce(206), // &, reduce: PowTerm
			reduce(206), // <<, reduce: PowTerm
			reduce(206), // >>>, reduce: PowTerm
			reduce(206), // +, reduce: PowTerm
			reduce(206), // -, reduce: PowTerm
			reduce(206), // .+, reduce: PowTerm
			reduce(206), // .-, reduce: PowTerm
			reduce(206), // *, reduce: PowTerm
			reduce(206), // /, reduce: PowTerm
			reduce(206), // //, reduce: PowTerm
			reduce(206), // %, reduce: PowTerm
			reduce(206), // .*, reduce: PowTerm
			reduce(206), // ./, reduce: PowTerm
			reduce(206), // .//, reduce: PowTerm
			reduce(206), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3307
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(205), // >, reduce: PowTerm
			reduce(205), // >>, reduce: PowTerm
			reduce(205), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(205), // ], reduce: PowTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(205), // ?, reduce: PowTerm
			nil,         // :
			reduce(205), // ||, reduce: PowTerm
			reduce(205), // ^^, reduce: PowTerm
			reduce(205), // &&, reduce: PowTerm
			reduce(205), // ??, reduce: PowTerm
			reduce(205), // ???, reduce: PowTerm
			reduce(205), // =~, reduce: PowTerm
			reduce(205), // !=~, reduce: PowTerm
			reduce(205), // ==, reduce: PowTerm
			reduce(205), // !=, reduce: PowTerm
			reduce(205), // <=>, reduce: PowTerm
			reduce(205), // >=, reduce: PowTerm
			reduce(205), // <, reduce: PowTerm
			reduce(205), // <=, reduce: PowTerm
			reduce(205), // ^, reduce: PowTerm
			reduce(205), // &, reduce: PowTerm
			reduce(205), // <<, reduce: PowTerm
			reduce(205), // >>>, reduce: PowTerm
			reduce(205), // +, reduce: PowTerm
			reduce(205), // -, reduce: PowTerm
			reduce(205), // .+, reduce: PowTerm
			reduce(205), // .-, reduce: PowTerm
			reduce(205), // *, reduce: PowTerm
			reduce(205), // /, reduce: PowTerm
			reduce(205), // //, reduce: PowTerm
			reduce(205), // %, reduce: PowTerm
			reduce(205), // .*, reduce: PowTerm
			reduce(205), // ./, reduce: PowTerm
			reduce(205), // .//, reduce: PowTerm
			reduce(205), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3308
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(260), // >, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>, reduce: ArrayOrMapIndexAccess
			reduce(260), // |, reduce: ArrayOrMapIndexAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(260), // ], reduce: ArrayOrMapIndexAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(260), // ?, reduce: ArrayOrMapIndexAccess
			nil,         // :
			reduce(260), // ||, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???, reduce: ArrayOrMapIndexAccess
			reduce(260), // =~, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=~, reduce: ArrayOrMapIndexAccess
			reduce(260), // ==, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=>, reduce: ArrayOrMapIndexAccess
			reduce(260), // >=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>, reduce: ArrayOrMapIndexAccess
			reduce(260), // +, reduce: ArrayOrMapIndexAccess
			reduce(260), // -, reduce: ArrayOrMapIndexAccess
			reduce(260), // .+, reduce: ArrayOrMapIndexAccess
			reduce(260), // .-, reduce: ArrayOrMapIndexAccess
			reduce(260), // *, reduce: ArrayOrMapIndexAccess
			reduce(260), // /, reduce: ArrayOrMapIndexAccess
			reduce(260), // //, reduce: ArrayOrMapIndexAccess
			reduce(260), // %, reduce: ArrayOrMapIndexAccess
			reduce(260), // .*, reduce: ArrayOrMapIndexAccess
			reduce(260), // ./, reduce: ArrayOrMapIndexAccess
			reduce(260), // .//, reduce: ArrayOrMapIndexAccess
			reduce(260), // ., reduce: ArrayOrMapIndexAccess
			nil,         // !
			nil,         // ~
			reduce(260), // **, reduce: ArrayOrMapIndexAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(260), // [, reduce: ArrayOrMapIndexAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(260), // [[, reduce: ArrayOrMapIndexAccess
			reduce(260), // [[[, reduce: ArrayOrMapIndexAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3309
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(3987), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3310
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3988), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3311
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(266), // >, reduce: ArraySliceAccess
			reduce(266), // >>, reduce: ArraySliceAccess
			reduce(266), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(266), // ], reduce: ArraySliceAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(266), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(266), // ||, reduce: ArraySliceAccess
			reduce(266), // ^^, reduce: ArraySliceAccess
			reduce(266), // &&, reduce: ArraySliceAccess
			reduce(266), // ??, reduce: ArraySliceAccess
			reduce(266), // ???, reduce: ArraySliceAccess
			reduce(266), // =~, reduce: ArraySliceAccess
			reduce(266), // !=~, reduce: ArraySliceAccess
			reduce(266), // ==, reduce: ArraySliceAccess
			reduce(266), // !=, reduce: ArraySliceAccess
			reduce(266), // <=>, reduce: ArraySliceAccess
			reduce(266), // >=, reduce: ArraySliceAccess
			reduce(266), // <, reduce: ArraySliceAccess
			reduce(266), // <=, reduce: ArraySliceAccess
			reduce(266), // ^, reduce: ArraySliceAccess
			reduce(266), // &, reduce: ArraySliceAccess
			reduce(266), // <<, reduce: ArraySliceAccess
			reduce(266), // >>>, reduce: ArraySliceAccess
			reduce(266), // +, reduce: ArraySliceAccess
			reduce(266), // -, reduce: ArraySliceAccess
			reduce(266), // .+, reduce: ArraySliceAccess
			reduce(266), // .-, reduce: ArraySliceAccess
			reduce(266), // *, reduce: ArraySliceAccess
			reduce(266), // /, reduce: ArraySliceAccess
			reduce(266), // //, reduce: ArraySliceAccess
			reduce(266), // %, reduce: ArraySliceAccess
			reduce(266), // .*, reduce: ArraySliceAccess
			reduce(266), // ./, reduce: ArraySliceAccess
			reduce(266), // .//, reduce: ArraySliceAccess
			reduce(266), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(266), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(266), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(266), // [[, reduce: ArraySliceAccess
			reduce(266), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3312
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3989), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3313
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3990), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3314
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(254), // >, reduce: ENV
			reduce(254), // >>, reduce: ENV
			reduce(254), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(254), // ], reduce: ENV
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(254), // ?, reduce: ENV
			nil,         // :
			reduce(254), // ||, reduce: ENV
			reduce(254), // ^^, reduce: ENV
			reduce(254), // &&, reduce: ENV
			reduce(254), // ??, reduce: ENV
			reduce(254), // ???, reduce: ENV
			reduce(254), // =~, reduce: ENV
			reduce(254), // !=~, reduce: ENV
			reduce(254), // ==, reduce: ENV
			reduce(254), // !=, reduce: ENV
			reduce(254), // <=>, reduce: ENV
			reduce(254), // >=, reduce: ENV
			reduce(254), // <, reduce: ENV
			reduce(254), // <=, reduce: ENV
			reduce(254), // ^, reduce: ENV
			reduce(254), // &, reduce: ENV
			reduce(254), // <<, reduce: ENV
			reduce(254), // >>>, reduce: ENV
			reduce(254), // +, reduce: ENV
			reduce(254), // -, reduce: ENV
			reduce(254), // .+, reduce: ENV
			reduce(254), // .-, reduce: ENV
			reduce(254), // *, reduce: ENV
			reduce(254), // /, reduce: ENV
			reduce(254), // //, reduce: ENV
			reduce(254), // %, reduce: ENV
			reduce(254), // .*, reduce: ENV
			reduce(254), // ./, reduce: ENV
			reduce(254), // .//, reduce: ENV
			reduce(254), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(254), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(254), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(254), // [[, reduce: ENV
			reduce(254), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3315
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(269), // >, reduce: FunctionCallsite
			reduce(269), // >>, reduce: FunctionCallsite
			reduce(269), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(269), // ], reduce: FunctionCallsite
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(269), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(269), // ||, reduce: FunctionCallsite
			reduce(269), // ^^, reduce: FunctionCallsite
			reduce(269), // &&, reduce: FunctionCallsite
			reduce(269), // ??, reduce: FunctionCallsite
			reduce(269), // ???, reduce: FunctionCallsite
			reduce(269), // =~, reduce: FunctionCallsite
			reduce(269), // !=~, reduce: FunctionCallsite
			reduce(269), // ==, reduce: FunctionCallsite
			reduce(269), // !=, reduce: FunctionCallsite
			reduce(269), // <=>, reduce: FunctionCallsite
			reduce(269), // >=, reduce: FunctionCallsite
			reduce(269), // <, reduce: FunctionCallsite
			reduce(269), // <=, reduce: FunctionCallsite
			reduce(269), // ^, reduce: FunctionCallsite
			reduce(269), // &, reduce: FunctionCallsite
			reduce(269), // <<, reduce: FunctionCallsite
			reduce(269), // >>>, reduce: FunctionCallsite
			reduce(269), // +, reduce: FunctionCallsite
			reduce(269), // -, reduce: FunctionCallsite
			reduce(269), // .+, reduce: FunctionCallsite
			reduce(269), // .-, reduce: FunctionCallsite
			reduce(269), // *, reduce: FunctionCallsite
			reduce(269), // /, reduce: FunctionCallsite
			reduce(269), // //, reduce: FunctionCallsite
			reduce(269), // %, reduce: FunctionCallsite
			reduce(269), // .*, reduce: FunctionCallsite
			reduce(269), // ./, reduce: FunctionCallsite
			reduce(269), // .//, reduce: FunctionCallsite
			reduce(269), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(269), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(269), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(269), // [[, reduce: FunctionCallsite
			reduce(269), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3316
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3992), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3993), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3317
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(108), // $, reduce: PositionalFieldValue
			nil,         // empty
			reduce(108), // ;, reduce: PositionalFieldValue
			reduce(108), // {, reduce: PositionalFieldValue
			nil,         // }
			reduce(108), // =, reduce: PositionalFieldValue
			nil,         // unset
			nil,         // filter
			reduce(108), // >, reduce: PositionalFieldValue
			reduce(108), // >>, reduce: PositionalFieldValue
			reduce(108), // |, reduce: PositionalFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(108), // ||=, reduce: PositionalFieldValue
			reduce(108), // ^^=, reduce: PositionalFieldValue
			reduce(108), // &&=, reduce: PositionalFieldValue
			reduce(108), // ??=, reduce: PositionalFieldValue
			reduce(108), // ???=, reduce: PositionalFieldValue
			reduce(108), // |=, reduce: PositionalFieldValue
			reduce(108), // &=, reduce: PositionalFieldValue
			reduce(108), // ^=, reduce: PositionalFieldValue
			reduce(108), // <<=, reduce: PositionalFieldValue
			reduce(108), // >>=, reduce: PositionalFieldValue
			reduce(108), // >>>=, reduce: PositionalFieldValue
			reduce(108), // +=, reduce: PositionalFieldValue
			reduce(108), // .=, reduce: PositionalFieldValue
			reduce(108), // -=, reduce: PositionalFieldValue
			reduce(108), // *=, reduce: PositionalFieldValue
			reduce(108), // /=, reduce: PositionalFieldValue
			reduce(108), // //=, reduce: PositionalFieldValue
			reduce(108), // %=, reduce: PositionalFieldValue
			reduce(108), // **=, reduce: PositionalFieldValue
			reduce(108), // ?, reduce: PositionalFieldValue
			nil,         // :
			reduce(108), // ||, reduce: PositionalFieldValue
			reduce(108), // ^^, reduce: PositionalFieldValue
			reduce(108), // &&, reduce: PositionalFieldValue
			reduce(108), // ??, reduce: PositionalFieldValue
			reduce(108), // ???, reduce: PositionalFieldValue
			reduce(108), // =~, reduce: PositionalFieldValue
			reduce(108), // !=~, reduce: PositionalFieldValue
			reduce(108), // ==, reduce: PositionalFieldValue
			reduce(108), // !=, reduce: PositionalFieldValue
			reduce(108), // <=>, reduce: PositionalFieldValue
			reduce(108), // >=, reduce: PositionalFieldValue
			reduce(108), // <, reduce: PositionalFieldValue
			reduce(108), // <=, reduce: PositionalFieldValue
			reduce(108), // ^, reduce: PositionalFieldValue
			reduce(108), // &, reduce: PositionalFieldValue
			reduce(108), // <<, reduce: PositionalFieldValue
			reduce(108), // >>>, reduce: PositionalFieldValue
			reduce(108), // +, reduce: PositionalFieldValue
			reduce(108), // -, reduce: PositionalFieldValue
			reduce(108), // .+, reduce: PositionalFieldValue
			reduce(108), // .-, reduce: PositionalFieldValue
			reduce(108), // *, reduce: PositionalFieldValue
			reduce(108), // /, reduce: PositionalFieldValue
			reduce(108), // //, reduce: PositionalFieldValue
			reduce(108), // %, reduce: PositionalFieldValue
			reduce(108), // .*, reduce: PositionalFieldValue
			reduce(108), // ./, reduce: PositionalFieldValue
			reduce(108), // .//, reduce: PositionalFieldValue
			reduce(108), // ., reduce: PositionalFieldValue
			nil,         // !
			nil,         // ~
			reduce(108), // **, reduce: PositionalFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(108), // [, reduce: PositionalFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(108), // [[, reduce: PositionalFieldValue
			reduce(108), // [[[, reduce: PositionalFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3318
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(149), // $, reduce: TernaryTerm
			nil,         // empty
			reduce(149), // ;, reduce: TernaryTerm
			reduce(149), // {, reduce: TernaryTerm
			nil,         // }
			reduce(149), // =, reduce: TernaryTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(149), // ||=, reduce: TernaryTerm
			reduce(149), // ^^=, reduce: TernaryTerm
			reduce(149), // &&=, reduce: TernaryTerm
			reduce(149), // ??=, reduce: TernaryTerm
			reduce(149), // ???=, reduce: TernaryTerm
			reduce(149), // |=, reduce: TernaryTerm
			reduce(149), // &=, reduce: TernaryTerm
			reduce(149), // ^=, reduce: TernaryTerm
			reduce(149), // <<=, reduce: TernaryTerm
			reduce(149), // >>=, reduce: TernaryTerm
			reduce(149), // >>>=, reduce: TernaryTerm
			reduce(149), // +=, reduce: TernaryTerm
			reduce(149), // .=, reduce: TernaryTerm
			reduce(149), // -=, reduce: TernaryTerm
			reduce(149), // *=, reduce: TernaryTerm
			reduce(149), // /=, reduce: TernaryTerm
			reduce(149), // //=, reduce: TernaryTerm
			reduce(149), // %=, reduce: TernaryTerm
			reduce(149), // **=, reduce: TernaryTerm
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3319
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(235), // >, reduce: MapLiteral
			reduce(235), // >>, reduce: MapLiteral
			reduce(235), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(235), // ], reduce: MapLiteral
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(235), // ?, reduce: MapLiteral
			reduce(235), // :, reduce: MapLiteral
			reduce(235), // ||, reduce: MapLiteral
			reduce(235), // ^^, reduce: MapLiteral
			reduce(235), // &&, reduce: MapLiteral
			reduce(235), // ??, reduce: MapLiteral
			reduce(235), // ???, reduce: MapLiteral
			reduce(235), // =~, reduce: MapLiteral
			reduce(235), // !=~, reduce: MapLiteral
			reduce(235), // ==, reduce: MapLiteral
			reduce(235), // !=, reduce: MapLiteral
			reduce(235), // <=>, reduce: MapLiteral
			reduce(235), // >=, reduce: MapLiteral
			reduce(235), // <, reduce: MapLiteral
			reduce(235), // <=, reduce: MapLiteral
			reduce(235), // ^, reduce: MapLiteral
			reduce(235), // &, reduce: MapLiteral
			reduce(235), // <<, reduce: MapLiteral
			reduce(235), // >>>, reduce: MapLiteral
			reduce(235), // +, reduce: MapLiteral
			reduce(235), // -, reduce: MapLiteral
			reduce(235), // .+, reduce: MapLiteral
			reduce(235), // .-, reduce: MapLiteral
			reduce(235), // *, reduce: MapLiteral
			reduce(235), // /, reduce: MapLiteral
			reduce(235), // //, reduce: MapLiteral
			reduce(235), // %, reduce: MapLiteral
			reduce(235), // .*, reduce: MapLiteral
			reduce(235), // ./, reduce: MapLiteral
			reduce(235), // .//, reduce: MapLiteral
			reduce(235), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(235), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(235), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(235), // [[, reduce: MapLiteral
			reduce(235), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3320
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3994), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3321
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(265), // $, reduce: ArraySliceAccess
			nil,         // empty
			reduce(265), // ;, reduce: ArraySliceAccess
			reduce(265), // {, reduce: ArraySliceAccess
			nil,         // }
			reduce(265), // =, reduce: ArraySliceAccess
			nil,         // unset
			nil,         // filter
			reduce(265), // >, reduce: ArraySliceAccess
			reduce(265), // >>, reduce: ArraySliceAccess
			reduce(265), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(265), // ||=, reduce: ArraySliceAccess
			reduce(265), // ^^=, reduce: ArraySliceAccess
			reduce(265), // &&=, reduce: ArraySliceAccess
			reduce(265), // ??=, reduce: ArraySliceAccess
			reduce(265), // ???=, reduce: ArraySliceAccess
			reduce(265), // |=, reduce: ArraySliceAccess
			reduce(265), // &=, reduce: ArraySliceAccess
			reduce(265), // ^=, reduce: ArraySliceAccess
			reduce(265), // <<=, reduce: ArraySliceAccess
			reduce(265), // >>=, reduce: ArraySliceAccess
			reduce(265), // >>>=, reduce: ArraySliceAccess
			reduce(265), // +=, reduce: ArraySliceAccess
			reduce(265), // .=, reduce: ArraySliceAccess
			reduce(265), // -=, reduce: ArraySliceAccess
			reduce(265), // *=, reduce: ArraySliceAccess
			reduce(265), // /=, reduce: ArraySliceAccess
			reduce(265), // //=, reduce: ArraySliceAccess
			reduce(265), // %=, reduce: ArraySliceAccess
			reduce(265), // **=, reduce: ArraySliceAccess
			reduce(265), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(265), // ||, reduce: ArraySliceAccess
			reduce(265), // ^^, reduce: ArraySliceAccess
			reduce(265), // &&, reduce: ArraySliceAccess
			reduce(265), // ??, reduce: ArraySliceAccess
			reduce(265), // ???, reduce: ArraySliceAccess
			reduce(265), // =~, reduce: ArraySliceAccess
			reduce(265), // !=~, reduce: ArraySliceAccess
			reduce(265), // ==, reduce: ArraySliceAccess
			reduce(265), // !=, reduce: ArraySliceAccess
			reduce(265), // <=>, reduce: ArraySliceAccess
			reduce(265), // >=, reduce: ArraySliceAccess
			reduce(265), // <, reduce: ArraySliceAccess
			reduce(265), // <=, reduce: ArraySliceAccess
			reduce(265), // ^, reduce: ArraySliceAccess
			reduce(265), // &, reduce: ArraySliceAccess
			reduce(265), // <<, reduce: ArraySliceAccess
			reduce(265), // >>>, reduce: ArraySliceAccess
			reduce(265), // +, reduce: ArraySliceAccess
			reduce(265), // -, reduce: ArraySliceAccess
			reduce(265), // .+, reduce: ArraySliceAccess
			reduce(265), // .-, reduce: ArraySliceAccess
			reduce(265), // *, reduce: ArraySliceAccess
			reduce(265), // /, reduce: ArraySliceAccess
			reduce(265), // //, reduce: ArraySliceAccess
			reduce(265), // %, reduce: ArraySliceAccess
			reduce(265), // .*, reduce: ArraySliceAccess
			reduce(265), // ./, reduce: ArraySliceAccess
			reduce(265), // .//, reduce: ArraySliceAccess
			reduce(265), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(265), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(265), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(265), // [[, reduce: ArraySliceAccess
			reduce(265), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3322
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(208), // >, reduce: PrecedenceChainEnd
			reduce(208), // >>, reduce: PrecedenceChainEnd
			reduce(208), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(208), // ], reduce: PrecedenceChainEnd
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(208), // ?, reduce: PrecedenceChainEnd
			reduce(208), // :, reduce: PrecedenceChainEnd
			reduce(208), // ||, reduce: PrecedenceChainEnd
			reduce(208), // ^^, reduce: PrecedenceChainEnd
			reduce(208), // &&, reduce: PrecedenceChainEnd
			reduce(208), // ??, reduce: PrecedenceChainEnd
			reduce(208), // ???, reduce: PrecedenceChainEnd
			reduce(208), // =~, reduce: PrecedenceChainEnd
			reduce(208), // !=~, reduce: PrecedenceChainEnd
			reduce(208), // ==, reduce: PrecedenceChainEnd
			reduce(208), // !=, reduce: PrecedenceChainEnd
			reduce(208), // <=>, reduce: PrecedenceChainEnd
			reduce(208), // >=, reduce: PrecedenceChainEnd
			reduce(208), // <, reduce: PrecedenceChainEnd
			reduce(208), // <=, reduce: PrecedenceChainEnd
			reduce(208), // ^, reduce: PrecedenceChainEnd
			reduce(208), // &, reduce: PrecedenceChainEnd
			reduce(208), // <<, reduce: PrecedenceChainEnd
			reduce(208), // >>>, reduce: PrecedenceChainEnd
			reduce(208), // +, reduce: PrecedenceChainEnd
			reduce(208), // -, reduce: PrecedenceChainEnd
			reduce(208), // .+, reduce: PrecedenceChainEnd
			reduce(208), // .-, reduce: PrecedenceChainEnd
			reduce(208), // *, reduce: PrecedenceChainEnd
			reduce(208), // /, reduce: PrecedenceChainEnd
			reduce(208), // //, reduce: PrecedenceChainEnd
			reduce(208), // %, reduce: PrecedenceChainEnd
			reduce(208), // .*, reduce: PrecedenceChainEnd
			reduce(208), // ./, reduce: PrecedenceChainEnd
			reduce(208), // .//, reduce: PrecedenceChainEnd
			reduce(208), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(208), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3323
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(105), // >, reduce: IndirectFieldValue
			reduce(105), // >>, reduce: IndirectFieldValue
			reduce(105), // |, reduce: IndirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(105), // ], reduce: IndirectFieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(105), // ?, reduce: IndirectFieldValue
			reduce(105), // :, reduce: IndirectFieldValue
			reduce(105), // ||, reduce: IndirectFieldValue
			reduce(105), // ^^, reduce: IndirectFieldValue
			reduce(105), // &&, reduce: IndirectFieldValue
			reduce(105), // ??, reduce: IndirectFieldValue
			reduce(105), // ???, reduce: IndirectFieldValue
			reduce(105), // =~, reduce: IndirectFieldValue
			reduce(105), // !=~, reduce: IndirectFieldValue
			reduce(105), // ==, reduce: IndirectFieldValue
			reduce(105), // !=, reduce: IndirectFieldValue
			reduce(105), // <=>, reduce: IndirectFieldValue
			reduce(105), // >=, reduce: IndirectFieldValue
			reduce(105), // <, reduce: IndirectFieldValue
			reduce(105), // <=, reduce: IndirectFieldValue
			reduce(105), // ^, reduce: IndirectFieldValue
			reduce(105), // &, reduce: IndirectFieldValue
			reduce(105), // <<, reduce: IndirectFieldValue
			reduce(105), // >>>, reduce: IndirectFieldValue
			reduce(105), // +, reduce: IndirectFieldValue
			reduce(105), // -, reduce: IndirectFieldValue
			reduce(105), // .+, reduce: IndirectFieldValue
			reduce(105), // .-, reduce: IndirectFieldValue
			reduce(105), // *, reduce: IndirectFieldValue
			reduce(105), // /, reduce: IndirectFieldValue
			reduce(105), // //, reduce: IndirectFieldValue
			reduce(105), // %, reduce: IndirectFieldValue
			reduce(105), // .*, reduce: IndirectFieldValue
			reduce(105), // ./, reduce: IndirectFieldValue
			reduce(105), // .//, reduce: IndirectFieldValue
			reduce(105), // ., reduce: IndirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(105), // **, reduce: IndirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(105), // [, reduce: IndirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(105), // [[, reduce: IndirectFieldValue
			reduce(105), // [[[, reduce: IndirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3324
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3995), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3325
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(3996), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3326
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(114), // >, reduce: IndirectOosvarValue
			reduce(114), // >>, reduce: IndirectOosvarValue
			reduce(114), // |, reduce: IndirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(114), // ], reduce: IndirectOosvarValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(114), // ?, reduce: IndirectOosvarValue
			reduce(114), // :, reduce: IndirectOosvarValue
			reduce(114), // ||, reduce: IndirectOosvarValue
			reduce(114), // ^^, reduce: IndirectOosvarValue
			reduce(114), // &&, reduce: IndirectOosvarValue
			reduce(114), // ??, reduce: IndirectOosvarValue
			reduce(114), // ???, reduce: IndirectOosvarValue
			reduce(114), // =~, reduce: IndirectOosvarValue
			reduce(114), // !=~, reduce: IndirectOosvarValue
			reduce(114), // ==, reduce: IndirectOosvarValue
			reduce(114), // !=, reduce: IndirectOosvarValue
			reduce(114), // <=>, reduce: IndirectOosvarValue
			reduce(114), // >=, reduce: IndirectOosvarValue
			reduce(114), // <, reduce: IndirectOosvarValue
			reduce(114), // <=, reduce: IndirectOosvarValue
			reduce(114), // ^, reduce: IndirectOosvarValue
			reduce(114), // &, reduce: IndirectOosvarValue
			reduce(114), // <<, reduce: IndirectOosvarValue
			reduce(114), // >>>, reduce: IndirectOosvarValue
			reduce(114), // +, reduce: IndirectOosvarValue
			reduce(114), // -, reduce: IndirectOosvarValue
			reduce(114), // .+, reduce: IndirectOosvarValue
			reduce(114), // .-, reduce: IndirectOosvarValue
			reduce(114), // *, reduce: IndirectOosvarValue
			reduce(114), // /, reduce: IndirectOosvarValue
			reduce(114), // //, reduce: IndirectOosvarValue
			reduce(114), // %, reduce: IndirectOosvarValue
			reduce(114), // .*, reduce: IndirectOosvarValue
			reduce(114), // ./, reduce: IndirectOosvarValue
			reduce(114), // .//, reduce: IndirectOosvarValue
			reduce(114), // ., reduce: IndirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(114), // **, reduce: IndirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(114), // [, reduce: IndirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(114), // [[, reduce: IndirectOosvarValue
			reduce(114), // [[[, reduce: IndirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3327
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(3997), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3328
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(151), // ], reduce: LogicalOrTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(151), // ?, reduce: LogicalOrTerm
			reduce(151), // :, reduce: LogicalOrTerm
			reduce(151), // ||, reduce: LogicalOrTerm
			shift(2647), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3329
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(264), // $, reduce: ArraySliceAccess
			nil,         // empty
			reduce(264), // ;, reduce: ArraySliceAccess
			reduce(264), // {, reduce: ArraySliceAccess
			nil,         // }
			reduce(264), // =, reduce: ArraySliceAccess
			nil,         // unset
			nil,         // filter
			reduce(264), // >, reduce: ArraySliceAccess
			reduce(264), // >>, reduce: ArraySliceAccess
			reduce(264), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(264), // ||=, reduce: ArraySliceAccess
			reduce(264), // ^^=, reduce: ArraySliceAccess
			reduce(264), // &&=, reduce: ArraySliceAccess
			reduce(264), // ??=, reduce: ArraySliceAccess
			reduce(264), // ???=, reduce: ArraySliceAccess
			reduce(264), // |=, reduce: ArraySliceAccess
			reduce(264), // &=, reduce: ArraySliceAccess
			reduce(264), // ^=, reduce: ArraySliceAccess
			reduce(264), // <<=, reduce: ArraySliceAccess
			reduce(264), // >>=, reduce: ArraySliceAccess
			reduce(264), // >>>=, reduce: ArraySliceAccess
			reduce(264), // +=, reduce: ArraySliceAccess
			reduce(264), // .=, reduce: ArraySliceAccess
			reduce(264), // -=, reduce: ArraySliceAccess
			reduce(264), // *=, reduce: ArraySliceAccess
			reduce(264), // /=, reduce: ArraySliceAccess
			reduce(264), // //=, reduce: ArraySliceAccess
			reduce(264), // %=, reduce: ArraySliceAccess
			reduce(264), // **=, reduce: ArraySliceAccess
			reduce(264), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(264), // ||, reduce: ArraySliceAccess
			reduce(264), // ^^, reduce: ArraySliceAccess
			reduce(264), // &&, reduce: ArraySliceAccess
			reduce(264), // ??, reduce: ArraySliceAccess
			reduce(264), // ???, reduce: ArraySliceAccess
			reduce(264), // =~, reduce: ArraySliceAccess
			reduce(264), // !=~, reduce: ArraySliceAccess
			reduce(264), // ==, reduce: ArraySliceAccess
			reduce(264), // !=, reduce: ArraySliceAccess
			reduce(264), // <=>, reduce: ArraySliceAccess
			reduce(264), // >=, reduce: ArraySliceAccess
			reduce(264), // <, reduce: ArraySliceAccess
			reduce(264), // <=, reduce: ArraySliceAccess
			reduce(264), // ^, reduce: ArraySliceAccess
			reduce(264), // &, reduce: ArraySliceAccess
			reduce(264), // <<, reduce: ArraySliceAccess
			reduce(264), // >>>, reduce: ArraySliceAccess
			reduce(264), // +, reduce: ArraySliceAccess
			reduce(264), // -, reduce: ArraySliceAccess
			reduce(264), // .+, reduce: ArraySliceAccess
			reduce(264), // .-, reduce: ArraySliceAccess
			reduce(264), // *, reduce: ArraySliceAccess
			reduce(264), // /, reduce: ArraySliceAccess
			reduce(264), // //, reduce: ArraySliceAccess
			reduce(264), // %, reduce: ArraySliceAccess
			reduce(264), // .*, reduce: ArraySliceAccess
			reduce(264), // ./, reduce: ArraySliceAccess
			reduce(264), // .//, reduce: ArraySliceAccess
			reduce(264), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(264), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(264), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(264), // [[, reduce: ArraySliceAccess
			reduce(264), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3330
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(153), // ], reduce: LogicalXORTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(153), // ?, reduce: LogicalXORTerm
			reduce(153), // :, reduce: LogicalXORTerm
			reduce(153), // ||, reduce: LogicalXORTerm
			reduce(153), // ^^, reduce: LogicalXORTerm
			shift(2648), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3331
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(155), // ], reduce: LogicalAndTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(155), // ?, reduce: LogicalAndTerm
			reduce(155), // :, reduce: LogicalAndTerm
			reduce(155), // ||, reduce: LogicalAndTerm
			reduce(155), // ^^, reduce: LogicalAndTerm
			reduce(155), // &&, reduce: LogicalAndTerm
			shift(2649), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3332
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(157), // ], reduce: AbsentCoalesceTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(157), // ?, reduce: AbsentCoalesceTerm
			reduce(157), // :, reduce: AbsentCoalesceTerm
			reduce(157), // ||, reduce: AbsentCoalesceTerm
			reduce(157), // ^^, reduce: AbsentCoalesceTerm
			reduce(157), // &&, reduce: AbsentCoalesceTerm
			reduce(157), // ??, reduce: AbsentCoalesceTerm
			shift(2650), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3333
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(159), // ], reduce: EmptyCoalesceTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(159), // ?, reduce: EmptyCoalesceTerm
			reduce(159), // :, reduce: EmptyCoalesceTerm
			reduce(159), // ||, reduce: EmptyCoalesceTerm
			reduce(159), // ^^, reduce: EmptyCoalesceTerm
			reduce(159), // &&, reduce: EmptyCoalesceTerm
			reduce(159), // ??, reduce: EmptyCoalesceTerm
			reduce(159), // ???, reduce: EmptyCoalesceTerm
			shift(2651), // =~
			shift(2652), // !=~
			shift(2653), // ==
			shift(2654), // !=
			shift(2655), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3334
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2656), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(161), // ], reduce: EqneTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(161), // ?, reduce: EqneTerm
			reduce(161), // :, reduce: EqneTerm
			reduce(161), // ||, reduce: EqneTerm
			reduce(161), // ^^, reduce: EqneTerm
			reduce(161), // &&, reduce: EqneTerm
			reduce(161), // ??, reduce: EqneTerm
			reduce(161), // ???, reduce: EqneTerm
			reduce(161), // =~, reduce: EqneTerm
			reduce(161), // !=~, reduce: EqneTerm
			reduce(161), // ==, reduce: EqneTerm
			reduce(161), // !=, reduce: EqneTerm
			reduce(161), // <=>, reduce: EqneTerm
			shift(2657), // >=
			shift(2658), // <
			shift(2659), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3335
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2656), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(162), // ], reduce: EqneTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(162), // ?, reduce: EqneTerm
			reduce(162), // :, reduce: EqneTerm
			reduce(162), // ||, reduce: EqneTerm
			reduce(162), // ^^, reduce: EqneTerm
			reduce(162), // &&, reduce: EqneTerm
			reduce(162), // ??, reduce: EqneTerm
			reduce(162), // ???, reduce: EqneTerm
			reduce(162), // =~, reduce: EqneTerm
			reduce(162), // !=~, reduce: EqneTerm
			reduce(162), // ==, reduce: EqneTerm
			reduce(162), // !=, reduce: EqneTerm
			reduce(162), // <=>, reduce: EqneTerm
			shift(2657), // >=
			shift(2658), // <
			shift(2659), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3336
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2656), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(163), // ], reduce: EqneTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(163), // ?, reduce: EqneTerm
			reduce(163), // :, reduce: EqneTerm
			reduce(163), // ||, reduce: EqneTerm
			reduce(163), // ^^, reduce: EqneTerm
			reduce(163), // &&, reduce: EqneTerm
			reduce(163), // ??, reduce: EqneTerm
			reduce(163), // ???, reduce: EqneTerm
			reduce(163), // =~, reduce: EqneTerm
			reduce(163), // !=~, reduce: EqneTerm
			reduce(163), // ==, reduce: EqneTerm
			reduce(163), // !=, reduce: EqneTerm
			reduce(163), // <=>, reduce: EqneTerm
			shift(2657), // >=
			shift(2658), // <
			shift(2659), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3337
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2656), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(164), // ], reduce: EqneTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(164), // ?, reduce: EqneTerm
			reduce(164), // :, reduce: EqneTerm
			reduce(164), // ||, reduce: EqneTerm
			reduce(164), // ^^, reduce: EqneTerm
			reduce(164), // &&, reduce: EqneTerm
			reduce(164), // ??, reduce: EqneTerm
			reduce(164), // ???, reduce: EqneTerm
			reduce(164), // =~, reduce: EqneTerm
			reduce(164), // !=~, reduce: EqneTerm
			reduce(164), // ==, reduce: EqneTerm
			reduce(164), // !=, reduce: EqneTerm
			reduce(164), // <=>, reduce: EqneTerm
			shift(2657), // >=
			shift(2658), // <
			shift(2659), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3338
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2656), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(165), // ], reduce: EqneTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(165), // ?, reduce: EqneTerm
			reduce(165), // :, reduce: EqneTerm
			reduce(165), // ||, reduce: EqneTerm
			reduce(165), // ^^, reduce: EqneTerm
			reduce(165), // &&, reduce: EqneTerm
			reduce(165), // ??, reduce: EqneTerm
			reduce(165), // ???, reduce: EqneTerm
			reduce(165), // =~, reduce: EqneTerm
			reduce(165), // !=~, reduce: EqneTerm
			reduce(165), // ==, reduce: EqneTerm
			reduce(165), // !=, reduce: EqneTerm
			reduce(165), // <=>, reduce: EqneTerm
			shift(2657), // >=
			shift(2658), // <
			shift(2659), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3339
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(167), // >, reduce: CmpTerm
			nil,         // >>
			shift(2660), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(167), // ], reduce: CmpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(167), // ?, reduce: CmpTerm
			reduce(167), // :, reduce: CmpTerm
			reduce(167), // ||, reduce: CmpTerm
			reduce(167), // ^^, reduce: CmpTerm
			reduce(167), // &&, reduce: CmpTerm
			reduce(167), // ??, reduce: CmpTerm
			reduce(167), // ???, reduce: CmpTerm
			reduce(167), // =~, reduce: CmpTerm
			reduce(167), // !=~, reduce: CmpTerm
			reduce(167), // ==, reduce: CmpTerm
			reduce(167), // !=, reduce: CmpTerm
			reduce(167), // <=>, reduce: CmpTerm
			reduce(167), // >=, reduce: CmpTerm
			reduce(167), // <, reduce: CmpTerm
			reduce(167), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3340
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(168), // >, reduce: CmpTerm
			nil,         // >>
			shift(2660), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(168), // ], reduce: CmpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(168), // ?, reduce: CmpTerm
			reduce(168), // :, reduce: CmpTerm
			reduce(168), // ||, reduce: CmpTerm
			reduce(168), // ^^, reduce: CmpTerm
			reduce(168), // &&, reduce: CmpTerm
			reduce(168), // ??, reduce: CmpTerm
			reduce(168), // ???, reduce: CmpTerm
			reduce(168), // =~, reduce: CmpTerm
			reduce(168), // !=~, reduce: CmpTerm
			reduce(168), // ==, reduce: CmpTerm
			reduce(168), // !=, reduce: CmpTerm
			reduce(168), // <=>, reduce: CmpTerm
			reduce(168), // >=, reduce: CmpTerm
			reduce(168), // <, reduce: CmpTerm
			reduce(168), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3341
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(169), // >, reduce: CmpTerm
			nil,         // >>
			shift(2660), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(169), // ], reduce: CmpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(169), // ?, reduce: CmpTerm
			reduce(169), // :, reduce: CmpTerm
			reduce(169), // ||, reduce: CmpTerm
			reduce(169), // ^^, reduce: CmpTerm
			reduce(169), // &&, reduce: CmpTerm
			reduce(169), // ??, reduce: CmpTerm
			reduce(169), // ???, reduce: CmpTerm
			reduce(169), // =~, reduce: CmpTerm
			reduce(169), // !=~, reduce: CmpTerm
			reduce(169), // ==, reduce: CmpTerm
			reduce(169), // !=, reduce: CmpTerm
			reduce(169), // <=>, reduce: CmpTerm
			reduce(169), // >=, reduce: CmpTerm
			reduce(169), // <, reduce: CmpTerm
			reduce(169), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3342
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(170), // >, reduce: CmpTerm
			nil,         // >>
			shift(2660), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(170), // ], reduce: CmpTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(170), // ?, reduce: CmpTerm
			reduce(170), // :, reduce: CmpTerm
			reduce(170), // ||, reduce: CmpTerm
			reduce(170), // ^^, reduce: CmpTerm
			reduce(170), // &&, reduce: CmpTerm
			reduce(170), // ??, reduce: CmpTerm
			reduce(170), // ???, reduce: CmpTerm
			reduce(170), // =~, reduce: CmpTerm
			reduce(170), // !=~, reduce: CmpTerm
			reduce(170), // ==, reduce: CmpTerm
			reduce(170), // !=, reduce: CmpTerm
			reduce(170), // <=>, reduce: CmpTerm
			reduce(170), // >=, reduce: CmpTerm
			reduce(170), // <, reduce: CmpTerm
			reduce(170), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3343
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(172), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(172), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(172), // ], reduce: BitwiseORTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(172), // ?, reduce: BitwiseORTerm
			reduce(172), // :, reduce: BitwiseORTerm
			reduce(172), // ||, reduce: BitwiseORTerm
			reduce(172), // ^^, reduce: BitwiseORTerm
			reduce(172), // &&, reduce: BitwiseORTerm
			reduce(172), // ??, reduce: BitwiseORTerm
			reduce(172), // ???, reduce: BitwiseORTerm
			reduce(172), // =~, reduce: BitwiseORTerm
			reduce(172), // !=~, reduce: BitwiseORTerm
			reduce(172), // ==, reduce: BitwiseORTerm
			reduce(172), // !=, reduce: BitwiseORTerm
			reduce(172), // <=>, reduce: BitwiseORTerm
			reduce(172), // >=, reduce: BitwiseORTerm
			reduce(172), // <, reduce: BitwiseORTerm
			reduce(172), // <=, reduce: BitwiseORTerm
			shift(2661), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3344
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(174), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(174), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(174), // ], reduce: BitwiseXORTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(174), // ?, reduce: BitwiseXORTerm
			reduce(174), // :, reduce: BitwiseXORTerm
			reduce(174), // ||, reduce: BitwiseXORTerm
			reduce(174), // ^^, reduce: BitwiseXORTerm
			reduce(174), // &&, reduce: BitwiseXORTerm
			reduce(174), // ??, reduce: BitwiseXORTerm
			reduce(174), // ???, reduce: BitwiseXORTerm
			reduce(174), // =~, reduce: BitwiseXORTerm
			reduce(174), // !=~, reduce: BitwiseXORTerm
			reduce(174), // ==, reduce: BitwiseXORTerm
			reduce(174), // !=, reduce: BitwiseXORTerm
			reduce(174), // <=>, reduce: BitwiseXORTerm
			reduce(174), // >=, reduce: BitwiseXORTerm
			reduce(174), // <, reduce: BitwiseXORTerm
			reduce(174), // <=, reduce: BitwiseXORTerm
			reduce(174), // ^, reduce: BitwiseXORTerm
			shift(2662), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3345
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(176), // >, reduce: BitwiseANDTerm
			shift(2663), // >>
			reduce(176), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(176), // ], reduce: BitwiseANDTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(176), // ?, reduce: BitwiseANDTerm
			reduce(176), // :, reduce: BitwiseANDTerm
			reduce(176), // ||, reduce: BitwiseANDTerm
			reduce(176), // ^^, reduce: BitwiseANDTerm
			reduce(176), // &&, reduce: BitwiseANDTerm
			reduce(176), // ??, reduce: BitwiseANDTerm
			reduce(176), // ???, reduce: BitwiseANDTerm
			reduce(176), // =~, reduce: BitwiseANDTerm
			reduce(176), // !=~, reduce: BitwiseANDTerm
			reduce(176), // ==, reduce: BitwiseANDTerm
			reduce(176), // !=, reduce: BitwiseANDTerm
			reduce(176), // <=>, reduce: BitwiseANDTerm
			reduce(176), // >=, reduce: BitwiseANDTerm
			reduce(176), // <, reduce: BitwiseANDTerm
			reduce(176), // <=, reduce: BitwiseANDTerm
			reduce(176), // ^, reduce: BitwiseANDTerm
			reduce(176), // &, reduce: BitwiseANDTerm
			shift(2664), // <<
			shift(2665), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3346
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(179), // >, reduce: BitwiseShiftTerm
			reduce(179), // >>, reduce: BitwiseShiftTerm
			reduce(179), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(179), // ], reduce: BitwiseShiftTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(179), // ?, reduce: BitwiseShiftTerm
			reduce(179), // :, reduce: BitwiseShiftTerm
			reduce(179), // ||, reduce: BitwiseShiftTerm
			reduce(179), // ^^, reduce: BitwiseShiftTerm
			reduce(179), // &&, reduce: BitwiseShiftTerm
			reduce(179), // ??, reduce: BitwiseShiftTerm
			reduce(179), // ???, reduce: BitwiseShiftTerm
			reduce(179), // =~, reduce: BitwiseShiftTerm
			reduce(179), // !=~, reduce: BitwiseShiftTerm
			reduce(179), // ==, reduce: BitwiseShiftTerm
			reduce(179), // !=, reduce: BitwiseShiftTerm
			reduce(179), // <=>, reduce: BitwiseShiftTerm
			reduce(179), // >=, reduce: BitwiseShiftTerm
			reduce(179), // <, reduce: BitwiseShiftTerm
			reduce(179), // <=, reduce: BitwiseShiftTerm
			reduce(179), // ^, reduce: BitwiseShiftTerm
			reduce(179), // &, reduce: BitwiseShiftTerm
			reduce(179), // <<, reduce: BitwiseShiftTerm
			reduce(179), // >>>, reduce: BitwiseShiftTerm
			shift(2666), // +
			shift(2667), // -
			shift(2668), // .+
			shift(2669), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3347
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(178), // >, reduce: BitwiseShiftTerm
			reduce(178), // >>, reduce: BitwiseShiftTerm
			reduce(178), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(178), // ], reduce: BitwiseShiftTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(178), // ?, reduce: BitwiseShiftTerm
			reduce(178), // :, reduce: BitwiseShiftTerm
			reduce(178), // ||, reduce: BitwiseShiftTerm
			reduce(178), // ^^, reduce: BitwiseShiftTerm
			reduce(178), // &&, reduce: BitwiseShiftTerm
			reduce(178), // ??, reduce: BitwiseShiftTerm
			reduce(178), // ???, reduce: BitwiseShiftTerm
			reduce(178), // =~, reduce: BitwiseShiftTerm
			reduce(178), // !=~, reduce: BitwiseShiftTerm
			reduce(178), // ==, reduce: BitwiseShiftTerm
			reduce(178), // !=, reduce: BitwiseShiftTerm
			reduce(178), // <=>, reduce: BitwiseShiftTerm
			reduce(178), // >=, reduce: BitwiseShiftTerm
			reduce(178), // <, reduce: BitwiseShiftTerm
			reduce(178), // <=, reduce: BitwiseShiftTerm
			reduce(178), // ^, reduce: BitwiseShiftTerm
			reduce(178), // &, reduce: BitwiseShiftTerm
			reduce(178), // <<, reduce: BitwiseShiftTerm
			reduce(178), // >>>, reduce: BitwiseShiftTerm
			shift(2666), // +
			shift(2667), // -
			shift(2668), // .+
			shift(2669), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3348
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(180), // >, reduce: BitwiseShiftTerm
			reduce(180), // >>, reduce: BitwiseShiftTerm
			reduce(180), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(180), // ], reduce: BitwiseShiftTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(180), // ?, reduce: BitwiseShiftTerm
			reduce(180), // :, reduce: BitwiseShiftTerm
			reduce(180), // ||, reduce: BitwiseShiftTerm
			reduce(180), // ^^, reduce: BitwiseShiftTerm
			reduce(180), // &&, reduce: BitwiseShiftTerm
			reduce(180), // ??, reduce: BitwiseShiftTerm
			reduce(180), // ???, reduce: BitwiseShiftTerm
			reduce(180), // =~, reduce: BitwiseShiftTerm
			reduce(180), // !=~, reduce: BitwiseShiftTerm
			reduce(180), // ==, reduce: BitwiseShiftTerm
			reduce(180), // !=, reduce: BitwiseShiftTerm
			reduce(180), // <=>, reduce: BitwiseShiftTerm
			reduce(180), // >=, reduce: BitwiseShiftTerm
			reduce(180), // <, reduce: BitwiseShiftTerm
			reduce(180), // <=, reduce: BitwiseShiftTerm
			reduce(180), // ^, reduce: BitwiseShiftTerm
			reduce(180), // &, reduce: BitwiseShiftTerm
			reduce(180), // <<, reduce: BitwiseShiftTerm
			reduce(180), // >>>, reduce: BitwiseShiftTerm
			shift(2666), // +
			shift(2667), // -
			shift(2668), // .+
			shift(2669), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3349
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(182), // >, reduce: AddsubdotTerm
			reduce(182), // >>, reduce: AddsubdotTerm
			reduce(182), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(182), // ], reduce: AddsubdotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(182), // ?, reduce: AddsubdotTerm
			reduce(182), // :, reduce: AddsubdotTerm
			reduce(182), // ||, reduce: AddsubdotTerm
			reduce(182), // ^^, reduce: AddsubdotTerm
			reduce(182), // &&, reduce: AddsubdotTerm
			reduce(182), // ??, reduce: AddsubdotTerm
			reduce(182), // ???, reduce: AddsubdotTerm
			reduce(182), // =~, reduce: AddsubdotTerm
			reduce(182), // !=~, reduce: AddsubdotTerm
			reduce(182), // ==, reduce: AddsubdotTerm
			reduce(182), // !=, reduce: AddsubdotTerm
			reduce(182), // <=>, reduce: AddsubdotTerm
			reduce(182), // >=, reduce: AddsubdotTerm
			reduce(182), // <, reduce: AddsubdotTerm
			reduce(182), // <=, reduce: AddsubdotTerm
			reduce(182), // ^, reduce: AddsubdotTerm
			reduce(182), // &, reduce: AddsubdotTerm
			reduce(182), // <<, reduce: AddsubdotTerm
			reduce(182), // >>>, reduce: AddsubdotTerm
			reduce(182), // +, reduce: AddsubdotTerm
			reduce(182), // -, reduce: AddsubdotTerm
			reduce(182), // .+, reduce: AddsubdotTerm
			reduce(182), // .-, reduce: AddsubdotTerm
			shift(2671), // *
			shift(2672), // /
			shift(2673), // //
			shift(2674), // %
			shift(2675), // .*
			shift(2676), // ./
			shift(2677), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3350
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(183), // >, reduce: AddsubdotTerm
			reduce(183), // >>, reduce: AddsubdotTerm
			reduce(183), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(183), // ], reduce: AddsubdotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(183), // ?, reduce: AddsubdotTerm
			reduce(183), // :, reduce: AddsubdotTerm
			reduce(183), // ||, reduce: AddsubdotTerm
			reduce(183), // ^^, reduce: AddsubdotTerm
			reduce(183), // &&, reduce: AddsubdotTerm
			reduce(183), // ??, reduce: AddsubdotTerm
			reduce(183), // ???, reduce: AddsubdotTerm
			reduce(183), // =~, reduce: AddsubdotTerm
			reduce(183), // !=~, reduce: AddsubdotTerm
			reduce(183), // ==, reduce: AddsubdotTerm
			reduce(183), // !=, reduce: AddsubdotTerm
			reduce(183), // <=>, reduce: AddsubdotTerm
			reduce(183), // >=, reduce: AddsubdotTerm
			reduce(183), // <, reduce: AddsubdotTerm
			reduce(183), // <=, reduce: AddsubdotTerm
			reduce(183), // ^, reduce: AddsubdotTerm
			reduce(183), // &, reduce: AddsubdotTerm
			reduce(183), // <<, reduce: AddsubdotTerm
			reduce(183), // >>>, reduce: AddsubdotTerm
			reduce(183), // +, reduce: AddsubdotTerm
			reduce(183), // -, reduce: AddsubdotTerm
			reduce(183), // .+, reduce: AddsubdotTerm
			reduce(183), // .-, reduce: AddsubdotTerm
			shift(2671), // *
			shift(2672), // /
			shift(2673), // //
			shift(2674), // %
			shift(2675), // .*
			shift(2676), // ./
			shift(2677), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3351
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(184), // >, reduce: AddsubdotTerm
			reduce(184), // >>, reduce: AddsubdotTerm
			reduce(184), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(184), // ], reduce: AddsubdotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(184), // ?, reduce: AddsubdotTerm
			reduce(184), // :, reduce: AddsubdotTerm
			reduce(184), // ||, reduce: AddsubdotTerm
			reduce(184), // ^^, reduce: AddsubdotTerm
			reduce(184), // &&, reduce: AddsubdotTerm
			reduce(184), // ??, reduce: AddsubdotTerm
			reduce(184), // ???, reduce: AddsubdotTerm
			reduce(184), // =~, reduce: AddsubdotTerm
			reduce(184), // !=~, reduce: AddsubdotTerm
			reduce(184), // ==, reduce: AddsubdotTerm
			reduce(184), // !=, reduce: AddsubdotTerm
			reduce(184), // <=>, reduce: AddsubdotTerm
			reduce(184), // >=, reduce: AddsubdotTerm
			reduce(184), // <, reduce: AddsubdotTerm
			reduce(184), // <=, reduce: AddsubdotTerm
			reduce(184), // ^, reduce: AddsubdotTerm
			reduce(184), // &, reduce: AddsubdotTerm
			reduce(184), // <<, reduce: AddsubdotTerm
			reduce(184), // >>>, reduce: AddsubdotTerm
			reduce(184), // +, reduce: AddsubdotTerm
			reduce(184), // -, reduce: AddsubdotTerm
			reduce(184), // .+, reduce: AddsubdotTerm
			reduce(184), // .-, reduce: AddsubdotTerm
			shift(2671), // *
			shift(2672), // /
			shift(2673), // //
			shift(2674), // %
			shift(2675), // .*
			shift(2676), // ./
			shift(2677), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3352
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(185), // >, reduce: AddsubdotTerm
			reduce(185), // >>, reduce: AddsubdotTerm
			reduce(185), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(185), // ], reduce: AddsubdotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(185), // ?, reduce: AddsubdotTerm
			reduce(185), // :, reduce: AddsubdotTerm
			reduce(185), // ||, reduce: AddsubdotTerm
			reduce(185), // ^^, reduce: AddsubdotTerm
			reduce(185), // &&, reduce: AddsubdotTerm
			reduce(185), // ??, reduce: AddsubdotTerm
			reduce(185), // ???, reduce: AddsubdotTerm
			reduce(185), // =~, reduce: AddsubdotTerm
			reduce(185), // !=~, reduce: AddsubdotTerm
			reduce(185), // ==, reduce: AddsubdotTerm
			reduce(185), // !=, reduce: AddsubdotTerm
			reduce(185), // <=>, reduce: AddsubdotTerm
			reduce(185), // >=, reduce: AddsubdotTerm
			reduce(185), // <, reduce: AddsubdotTerm
			reduce(185), // <=, reduce: AddsubdotTerm
			reduce(185), // ^, reduce: AddsubdotTerm
			reduce(185), // &, reduce: AddsubdotTerm
			reduce(185), // <<, reduce: AddsubdotTerm
			reduce(185), // >>>, reduce: AddsubdotTerm
			reduce(185), // +, reduce: AddsubdotTerm
			reduce(185), // -, reduce: AddsubdotTerm
			reduce(185), // .+, reduce: AddsubdotTerm
			reduce(185), // .-, reduce: AddsubdotTerm
			shift(2671), // *
			shift(2672), // /
			shift(2673), // //
			shift(2674), // %
			shift(2675), // .*
			shift(2676), // ./
			shift(2677), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3353
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(187), // >, reduce: MuldivTerm
			reduce(187), // >>, reduce: MuldivTerm
			reduce(187), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(187), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(187), // ?, reduce: MuldivTerm
			reduce(187), // :, reduce: MuldivTerm
			reduce(187), // ||, reduce: MuldivTerm
			reduce(187), // ^^, reduce: MuldivTerm
			reduce(187), // &&, reduce: MuldivTerm
			reduce(187), // ??, reduce: MuldivTerm
			reduce(187), // ???, reduce: MuldivTerm
			reduce(187), // =~, reduce: MuldivTerm
			reduce(187), // !=~, reduce: MuldivTerm
			reduce(187), // ==, reduce: MuldivTerm
			reduce(187), // !=, reduce: MuldivTerm
			reduce(187), // <=>, reduce: MuldivTerm
			reduce(187), // >=, reduce: MuldivTerm
			reduce(187), // <, reduce: MuldivTerm
			reduce(187), // <=, reduce: MuldivTerm
			reduce(187), // ^, reduce: MuldivTerm
			reduce(187), // &, reduce: MuldivTerm
			reduce(187), // <<, reduce: MuldivTerm
			reduce(187), // >>>, reduce: MuldivTerm
			reduce(187), // +, reduce: MuldivTerm
			reduce(187), // -, reduce: MuldivTerm
			reduce(187), // .+, reduce: MuldivTerm
			reduce(187), // .-, reduce: MuldivTerm
			reduce(187), // *, reduce: MuldivTerm
			reduce(187), // /, reduce: MuldivTerm
			reduce(187), // //, reduce: MuldivTerm
			reduce(187), // %, reduce: MuldivTerm
			reduce(187), // .*, reduce: MuldivTerm
			reduce(187), // ./, reduce: MuldivTerm
			reduce(187), // .//, reduce: MuldivTerm
			shift(2681), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3354
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(188), // >, reduce: MuldivTerm
			reduce(188), // >>, reduce: MuldivTerm
			reduce(188), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(188), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(188), // ?, reduce: MuldivTerm
			reduce(188), // :, reduce: MuldivTerm
			reduce(188), // ||, reduce: MuldivTerm
			reduce(188), // ^^, reduce: MuldivTerm
			reduce(188), // &&, reduce: MuldivTerm
			reduce(188), // ??, reduce: MuldivTerm
			reduce(188), // ???, reduce: MuldivTerm
			reduce(188), // =~, reduce: MuldivTerm
			reduce(188), // !=~, reduce: MuldivTerm
			reduce(188), // ==, reduce: MuldivTerm
			reduce(188), // !=, reduce: MuldivTerm
			reduce(188), // <=>, reduce: MuldivTerm
			reduce(188), // >=, reduce: MuldivTerm
			reduce(188), // <, reduce: MuldivTerm
			reduce(188), // <=, reduce: MuldivTerm
			reduce(188), // ^, reduce: MuldivTerm
			reduce(188), // &, reduce: MuldivTerm
			reduce(188), // <<, reduce: MuldivTerm
			reduce(188), // >>>, reduce: MuldivTerm
			reduce(188), // +, reduce: MuldivTerm
			reduce(188), // -, reduce: MuldivTerm
			reduce(188), // .+, reduce: MuldivTerm
			reduce(188), // .-, reduce: MuldivTerm
			reduce(188), // *, reduce: MuldivTerm
			reduce(188), // /, reduce: MuldivTerm
			reduce(188), // //, reduce: MuldivTerm
			reduce(188), // %, reduce: MuldivTerm
			reduce(188), // .*, reduce: MuldivTerm
			reduce(188), // ./, reduce: MuldivTerm
			reduce(188), // .//, reduce: MuldivTerm
			shift(2681), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3355
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(189), // >, reduce: MuldivTerm
			reduce(189), // >>, reduce: MuldivTerm
			reduce(189), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(189), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(189), // ?, reduce: MuldivTerm
			reduce(189), // :, reduce: MuldivTerm
			reduce(189), // ||, reduce: MuldivTerm
			reduce(189), // ^^, reduce: MuldivTerm
			reduce(189), // &&, reduce: MuldivTerm
			reduce(189), // ??, reduce: MuldivTerm
			reduce(189), // ???, reduce: MuldivTerm
			reduce(189), // =~, reduce: MuldivTerm
			reduce(189), // !=~, reduce: MuldivTerm
			reduce(189), // ==, reduce: MuldivTerm
			reduce(189), // !=, reduce: MuldivTerm
			reduce(189), // <=>, reduce: MuldivTerm
			reduce(189), // >=, reduce: MuldivTerm
			reduce(189), // <, reduce: MuldivTerm
			reduce(189), // <=, reduce: MuldivTerm
			reduce(189), // ^, reduce: MuldivTerm
			reduce(189), // &, reduce: MuldivTerm
			reduce(189), // <<, reduce: MuldivTerm
			reduce(189), // >>>, reduce: MuldivTerm
			reduce(189), // +, reduce: MuldivTerm
			reduce(189), // -, reduce: MuldivTerm
			reduce(189), // .+, reduce: MuldivTerm
			reduce(189), // .-, reduce: MuldivTerm
			reduce(189), // *, reduce: MuldivTerm
			reduce(189), // /, reduce: MuldivTerm
			reduce(189), // //, reduce: MuldivTerm
			reduce(189), // %, reduce: MuldivTerm
			reduce(189), // .*, reduce: MuldivTerm
			reduce(189), // ./, reduce: MuldivTerm
			reduce(189), // .//, reduce: MuldivTerm
			shift(2681), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3356
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(190), // >, reduce: MuldivTerm
			reduce(190), // >>, reduce: MuldivTerm
			reduce(190), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(190), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(190), // ?, reduce: MuldivTerm
			reduce(190), // :, reduce: MuldivTerm
			reduce(190), // ||, reduce: MuldivTerm
			reduce(190), // ^^, reduce: MuldivTerm
			reduce(190), // &&, reduce: MuldivTerm
			reduce(190), // ??, reduce: MuldivTerm
			reduce(190), // ???, reduce: MuldivTerm
			reduce(190), // =~, reduce: MuldivTerm
			reduce(190), // !=~, reduce: MuldivTerm
			reduce(190), // ==, reduce: MuldivTerm
			reduce(190), // !=, reduce: MuldivTerm
			reduce(190), // <=>, reduce: MuldivTerm
			reduce(190), // >=, reduce: MuldivTerm
			reduce(190), // <, reduce: MuldivTerm
			reduce(190), // <=, reduce: MuldivTerm
			reduce(190), // ^, reduce: MuldivTerm
			reduce(190), // &, reduce: MuldivTerm
			reduce(190), // <<, reduce: MuldivTerm
			reduce(190), // >>>, reduce: MuldivTerm
			reduce(190), // +, reduce: MuldivTerm
			reduce(190), // -, reduce: MuldivTerm
			reduce(190), // .+, reduce: MuldivTerm
			reduce(190), // .-, reduce: MuldivTerm
			reduce(190), // *, reduce: MuldivTerm
			reduce(190), // /, reduce: MuldivTerm
			reduce(190), // //, reduce: MuldivTerm
			reduce(190), // %, reduce: MuldivTerm
			reduce(190), // .*, reduce: MuldivTerm
			reduce(190), // ./, reduce: MuldivTerm
			reduce(190), // .//, reduce: MuldivTerm
			shift(2681), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3357
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(191), // >, reduce: MuldivTerm
			reduce(191), // >>, reduce: MuldivTerm
			reduce(191), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(191), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(191), // ?, reduce: MuldivTerm
			reduce(191), // :, reduce: MuldivTerm
			reduce(191), // ||, reduce: MuldivTerm
			reduce(191), // ^^, reduce: MuldivTerm
			reduce(191), // &&, reduce: MuldivTerm
			reduce(191), // ??, reduce: MuldivTerm
			reduce(191), // ???, reduce: MuldivTerm
			reduce(191), // =~, reduce: MuldivTerm
			reduce(191), // !=~, reduce: MuldivTerm
			reduce(191), // ==, reduce: MuldivTerm
			reduce(191), // !=, reduce: MuldivTerm
			reduce(191), // <=>, reduce: MuldivTerm
			reduce(191), // >=, reduce: MuldivTerm
			reduce(191), // <, reduce: MuldivTerm
			reduce(191), // <=, reduce: MuldivTerm
			reduce(191), // ^, reduce: MuldivTerm
			reduce(191), // &, reduce: MuldivTerm
			reduce(191), // <<, reduce: MuldivTerm
			reduce(191), // >>>, reduce: MuldivTerm
			reduce(191), // +, reduce: MuldivTerm
			reduce(191), // -, reduce: MuldivTerm
			reduce(191), // .+, reduce: MuldivTerm
			reduce(191), // .-, reduce: MuldivTerm
			reduce(191), // *, reduce: MuldivTerm
			reduce(191), // /, reduce: MuldivTerm
			reduce(191), // //, reduce: MuldivTerm
			reduce(191), // %, reduce: MuldivTerm
			reduce(191), // .*, reduce: MuldivTerm
			reduce(191), // ./, reduce: MuldivTerm
			reduce(191), // .//, reduce: MuldivTerm
			shift(2681), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3358
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(192), // >, reduce: MuldivTerm
			reduce(192), // >>, reduce: MuldivTerm
			reduce(192), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(192), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(192), // ?, reduce: MuldivTerm
			reduce(192), // :, reduce: MuldivTerm
			reduce(192), // ||, reduce: MuldivTerm
			reduce(192), // ^^, reduce: MuldivTerm
			reduce(192), // &&, reduce: MuldivTerm
			reduce(192), // ??, reduce: MuldivTerm
			reduce(192), // ???, reduce: MuldivTerm
			reduce(192), // =~, reduce: MuldivTerm
			reduce(192), // !=~, reduce: MuldivTerm
			reduce(192), // ==, reduce: MuldivTerm
			reduce(192), // !=, reduce: MuldivTerm
			reduce(192), // <=>, reduce: MuldivTerm
			reduce(192), // >=, reduce: MuldivTerm
			reduce(192), // <, reduce: MuldivTerm
			reduce(192), // <=, reduce: MuldivTerm
			reduce(192), // ^, reduce: MuldivTerm
			reduce(192), // &, reduce: MuldivTerm
			reduce(192), // <<, reduce: MuldivTerm
			reduce(192), // >>>, reduce: MuldivTerm
			reduce(192), // +, reduce: MuldivTerm
			reduce(192), // -, reduce: MuldivTerm
			reduce(192), // .+, reduce: MuldivTerm
			reduce(192), // .-, reduce: MuldivTerm
			reduce(192), // *, reduce: MuldivTerm
			reduce(192), // /, reduce: MuldivTerm
			reduce(192), // //, reduce: MuldivTerm
			reduce(192), // %, reduce: MuldivTerm
			reduce(192), // .*, reduce: MuldivTerm
			reduce(192), // ./, reduce: MuldivTerm
			reduce(192), // .//, reduce: MuldivTerm
			shift(2681), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3359
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(193), // >, reduce: MuldivTerm
			reduce(193), // >>, reduce: MuldivTerm
			reduce(193), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(193), // ], reduce: MuldivTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(193), // ?, reduce: MuldivTerm
			reduce(193), // :, reduce: MuldivTerm
			reduce(193), // ||, reduce: MuldivTerm
			reduce(193), // ^^, reduce: MuldivTerm
			reduce(193), // &&, reduce: MuldivTerm
			reduce(193), // ??, reduce: MuldivTerm
			reduce(193), // ???, reduce: MuldivTerm
			reduce(193), // =~, reduce: MuldivTerm
			reduce(193), // !=~, reduce: MuldivTerm
			reduce(193), // ==, reduce: MuldivTerm
			reduce(193), // !=, reduce: MuldivTerm
			reduce(193), // <=>, reduce: MuldivTerm
			reduce(193), // >=, reduce: MuldivTerm
			reduce(193), // <, reduce: MuldivTerm
			reduce(193), // <=, reduce: MuldivTerm
			reduce(193), // ^, reduce: MuldivTerm
			reduce(193), // &, reduce: MuldivTerm
			reduce(193), // <<, reduce: MuldivTerm
			reduce(193), // >>>, reduce: MuldivTerm
			reduce(193), // +, reduce: MuldivTerm
			reduce(193), // -, reduce: MuldivTerm
			reduce(193), // .+, reduce: MuldivTerm
			reduce(193), // .-, reduce: MuldivTerm
			reduce(193), // *, reduce: MuldivTerm
			reduce(193), // /, reduce: MuldivTerm
			reduce(193), // //, reduce: MuldivTerm
			reduce(193), // %, reduce: MuldivTerm
			reduce(193), // .*, reduce: MuldivTerm
			reduce(193), // ./, reduce: MuldivTerm
			reduce(193), // .//, reduce: MuldivTerm
			shift(2681), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3360
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(195), // >, reduce: DotTerm
			reduce(195), // >>, reduce: DotTerm
			reduce(195), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(195), // ], reduce: DotTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(195), // ?, reduce: DotTerm
			reduce(195), // :, reduce: DotTerm
			reduce(195), // ||, reduce: DotTerm
			reduce(195), // ^^, reduce: DotTerm
			reduce(195), // &&, reduce: DotTerm
			reduce(195), // ??, reduce: DotTerm
			reduce(195), // ???, reduce: DotTerm
			reduce(195), // =~, reduce: DotTerm
			reduce(195), // !=~, reduce: DotTerm
			reduce(195), // ==, reduce: DotTerm
			reduce(195), // !=, reduce: DotTerm
			reduce(195), // <=>, reduce: DotTerm
			reduce(195), // >=, reduce: DotTerm
			reduce(195), // <, reduce: DotTerm
			reduce(195), // <=, reduce: DotTerm
			reduce(195), // ^, reduce: DotTerm
			reduce(195), // &, reduce: DotTerm
			reduce(195), // <<, reduce: DotTerm
			reduce(195), // >>>, reduce: DotTerm
			reduce(195), // +, reduce: DotTerm
			reduce(195), // -, reduce: DotTerm
			reduce(195), // .+, reduce: DotTerm
			reduce(195), // .-, reduce: DotTerm
			reduce(195), // *, reduce: DotTerm
			reduce(195), // /, reduce: DotTerm
			reduce(195), // //, reduce: DotTerm
			reduce(195), // %, reduce: DotTerm
			reduce(195), // .*, reduce: DotTerm
			reduce(195), // ./, reduce: DotTerm
			reduce(195), // .//, reduce: DotTerm
			reduce(195), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3361
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3362
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3363
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(204), // >, reduce: PowTerm
			reduce(204), // >>, reduce: PowTerm
			reduce(204), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(204), // ], reduce: PowTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(204), // ?, reduce: PowTerm
			reduce(204), // :, reduce: PowTerm
			reduce(204), // ||, reduce: PowTerm
			reduce(204), // ^^, reduce: PowTerm
			reduce(204), // &&, reduce: PowTerm
			reduce(204), // ??, reduce: PowTerm
			reduce(204), // ???, reduce: PowTerm
			reduce(204), // =~, reduce: PowTerm
			reduce(204), // !=~, reduce: PowTerm
			reduce(204), // ==, reduce: PowTerm
			reduce(204), // !=, reduce: PowTerm
			reduce(204), // <=>, reduce: PowTerm
			reduce(204), // >=, reduce: PowTerm
			reduce(204), // <, reduce: PowTerm
			reduce(204), // <=, reduce: PowTerm
			reduce(204), // ^, reduce: PowTerm
			reduce(204), // &, reduce: PowTerm
			reduce(204), // <<, reduce: PowTerm
			reduce(204), // >>>, reduce: PowTerm
			reduce(204), // +, reduce: PowTerm
			reduce(204), // -, reduce: PowTerm
			reduce(204), // .+, reduce: PowTerm
			reduce(204), // .-, reduce: PowTerm
			reduce(204), // *, reduce: PowTerm
			reduce(204), // /, reduce: PowTerm
			reduce(204), // //, reduce: PowTerm
			reduce(204), // %, reduce: PowTerm
			reduce(204), // .*, reduce: PowTerm
			reduce(204), // ./, reduce: PowTerm
			reduce(204), // .//, reduce: PowTerm
			reduce(204), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3364
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4000), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4001), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3365
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4003), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3366
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4004), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3367
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4005), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3368
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(229), // >, reduce: ArrayLiteral
			reduce(229), // >>, reduce: ArrayLiteral
			reduce(229), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(229), // ], reduce: ArrayLiteral
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(229), // ?, reduce: ArrayLiteral
			reduce(229), // :, reduce: ArrayLiteral
			reduce(229), // ||, reduce: ArrayLiteral
			reduce(229), // ^^, reduce: ArrayLiteral
			reduce(229), // &&, reduce: ArrayLiteral
			reduce(229), // ??, reduce: ArrayLiteral
			reduce(229), // ???, reduce: ArrayLiteral
			reduce(229), // =~, reduce: ArrayLiteral
			reduce(229), // !=~, reduce: ArrayLiteral
			reduce(229), // ==, reduce: ArrayLiteral
			reduce(229), // !=, reduce: ArrayLiteral
			reduce(229), // <=>, reduce: ArrayLiteral
			reduce(229), // >=, reduce: ArrayLiteral
			reduce(229), // <, reduce: ArrayLiteral
			reduce(229), // <=, reduce: ArrayLiteral
			reduce(229), // ^, reduce: ArrayLiteral
			reduce(229), // &, reduce: ArrayLiteral
			reduce(229), // <<, reduce: ArrayLiteral
			reduce(229), // >>>, reduce: ArrayLiteral
			reduce(229), // +, reduce: ArrayLiteral
			reduce(229), // -, reduce: ArrayLiteral
			reduce(229), // .+, reduce: ArrayLiteral
			reduce(229), // .-, reduce: ArrayLiteral
			reduce(229), // *, reduce: ArrayLiteral
			reduce(229), // /, reduce: ArrayLiteral
			reduce(229), // //, reduce: ArrayLiteral
			reduce(229), // %, reduce: ArrayLiteral
			reduce(229), // .*, reduce: ArrayLiteral
			reduce(229), // ./, reduce: ArrayLiteral
			reduce(229), // .//, reduce: ArrayLiteral
			reduce(229), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(229), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(229), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(229), // [[, reduce: ArrayLiteral
			reduce(229), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3369
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(255), // >, reduce: ENV
			reduce(255), // >>, reduce: ENV
			reduce(255), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(255), // ], reduce: ENV
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(255), // ?, reduce: ENV
			reduce(255), // :, reduce: ENV
			reduce(255), // ||, reduce: ENV
			reduce(255), // ^^, reduce: ENV
			reduce(255), // &&, reduce: ENV
			reduce(255), // ??, reduce: ENV
			reduce(255), // ???, reduce: ENV
			reduce(255), // =~, reduce: ENV
			reduce(255), // !=~, reduce: ENV
			reduce(255), // ==, reduce: ENV
			reduce(255), // !=, reduce: ENV
			reduce(255), // <=>, reduce: ENV
			reduce(255), // >=, reduce: ENV
			reduce(255), // <, reduce: ENV
			reduce(255), // <=, reduce: ENV
			reduce(255), // ^, reduce: ENV
			reduce(255), // &, reduce: ENV
			reduce(255), // <<, reduce: ENV
			reduce(255), // >>>, reduce: ENV
			reduce(255), // +, reduce: ENV
			reduce(255), // -, reduce: ENV
			reduce(255), // .+, reduce: ENV
			reduce(255), // .-, reduce: ENV
			reduce(255), // *, reduce: ENV
			reduce(255), // /, reduce: ENV
			reduce(255), // //, reduce: ENV
			reduce(255), // %, reduce: ENV
			reduce(255), // .*, reduce: ENV
			reduce(255), // ./, reduce: ENV
			reduce(255), // .//, reduce: ENV
			reduce(255), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(255), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(255), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(255), // [[, reduce: ENV
			reduce(255), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3370
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4006), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3371
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4007), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3372
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(268), // >, reduce: FunctionCallsite
			reduce(268), // >>, reduce: FunctionCallsite
			reduce(268), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(268), // ], reduce: FunctionCallsite
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(268), // ?, reduce: FunctionCallsite
			reduce(268), // :, reduce: FunctionCallsite
			reduce(268), // ||, reduce: FunctionCallsite
			reduce(268), // ^^, reduce: FunctionCallsite
			reduce(268), // &&, reduce: FunctionCallsite
			reduce(268), // ??, reduce: FunctionCallsite
			reduce(268), // ???, reduce: FunctionCallsite
			reduce(268), // =~, reduce: FunctionCallsite
			reduce(268), // !=~, reduce: FunctionCallsite
			reduce(268), // ==, reduce: FunctionCallsite
			reduce(268), // !=, reduce: FunctionCallsite
			reduce(268), // <=>, reduce: FunctionCallsite
			reduce(268), // >=, reduce: FunctionCallsite
			reduce(268), // <, reduce: FunctionCallsite
			reduce(268), // <=, reduce: FunctionCallsite
			reduce(268), // ^, reduce: FunctionCallsite
			reduce(268), // &, reduce: FunctionCallsite
			reduce(268), // <<, reduce: FunctionCallsite
			reduce(268), // >>>, reduce: FunctionCallsite
			reduce(268), // +, reduce: FunctionCallsite
			reduce(268), // -, reduce: FunctionCallsite
			reduce(268), // .+, reduce: FunctionCallsite
			reduce(268), // .-, reduce: FunctionCallsite
			reduce(268), // *, reduce: FunctionCallsite
			reduce(268), // /, reduce: FunctionCallsite
			reduce(268), // //, reduce: FunctionCallsite
			reduce(268), // %, reduce: FunctionCallsite
			reduce(268), // .*, reduce: FunctionCallsite
			reduce(268), // ./, reduce: FunctionCallsite
			reduce(268), // .//, reduce: FunctionCallsite
			reduce(268), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(268), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(268), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(268), // [[, reduce: FunctionCallsite
			reduce(268), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3373
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4008), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3374
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(261), // $, reduce: ArrayOrMapPositionalNameAccess
			nil,         // empty
			reduce(261), // ;, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // {, reduce: ArrayOrMapPositionalNameAccess
			nil,         // }
			reduce(261), // =, reduce: ArrayOrMapPositionalNameAccess
			nil,         // unset
			nil,         // filter
			reduce(261), // >, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |, reduce: ArrayOrMapPositionalNameAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(261), // ||=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // **=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ?, reduce: ArrayOrMapPositionalNameAccess
			nil,         // :
			reduce(261), // ||, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // =~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ==, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .+, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .-, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .*, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ./, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .//, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ., reduce: ArrayOrMapPositionalNameAccess
			nil,         // !
			nil,         // ~
			reduce(261), // **, reduce: ArrayOrMapPositionalNameAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(261), // [, reduce: ArrayOrMapPositionalNameAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(261), // [[, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // [[[, reduce: ArrayOrMapPositionalNameAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3375
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4009), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3376
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(107), // >, reduce: PositionalFieldName
			reduce(107), // >>, reduce: PositionalFieldName
			reduce(107), // |, reduce: PositionalFieldName
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(107), // ,, reduce: PositionalFieldName
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(107), // ], reduce: PositionalFieldName
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(107), // ?, reduce: PositionalFieldName
			nil,         // :
			reduce(107), // ||, reduce: PositionalFieldName
			reduce(107), // ^^, reduce: PositionalFieldName
			reduce(107), // &&, reduce: PositionalFieldName
			reduce(107), // ??, reduce: PositionalFieldName
			reduce(107), // ???, reduce: PositionalFieldName
			reduce(107), // =~, reduce: PositionalFieldName
			reduce(107), // !=~, reduce: PositionalFieldName
			reduce(107), // ==, reduce: PositionalFieldName
			reduce(107), // !=, reduce: PositionalFieldName
			reduce(107), // <=>, reduce: PositionalFieldName
			reduce(107), // >=, reduce: PositionalFieldName
			reduce(107), // <, reduce: PositionalFieldName
			reduce(107), // <=, reduce: PositionalFieldName
			reduce(107), // ^, reduce: PositionalFieldName
			reduce(107), // &, reduce: PositionalFieldName
			reduce(107), // <<, reduce: PositionalFieldName
			reduce(107), // >>>, reduce: PositionalFieldName
			reduce(107), // +, reduce: PositionalFieldName
			reduce(107), // -, reduce: PositionalFieldName
			reduce(107), // .+, reduce: PositionalFieldName
			reduce(107), // .-, reduce: PositionalFieldName
			reduce(107), // *, reduce: PositionalFieldName
			reduce(107), // /, reduce: PositionalFieldName
			reduce(107), // //, reduce: PositionalFieldName
			reduce(107), // %, reduce: PositionalFieldName
			reduce(107), // .*, reduce: PositionalFieldName
			reduce(107), // ./, reduce: PositionalFieldName
			reduce(107), // .//, reduce: PositionalFieldName
			reduce(107), // ., reduce: PositionalFieldName
			nil,         // !
			nil,         // ~
			reduce(107), // **, reduce: PositionalFieldName
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(107), // [, reduce: PositionalFieldName
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(107), // [[, reduce: PositionalFieldName
			reduce(107), // [[[, reduce: PositionalFieldName
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3377
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4010), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3378
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(752), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(756), // (
			nil,        // )
			nil,        // emitp
			shift(768), // field_name
			shift(769), // $[
			nil,        // ]
			shift(771), // braced_field_name
			shift(772), // $[[
			shift(773), // $[[[
			shift(774), // full_srec
			shift(776), // oosvar_name
			shift(777), // @[
			shift(778), // braced_oosvar_name
			shift(779), // full_oosvar
			shift(780), // all
			shift(781), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(796), // +
			shift(798), // -
			shift(799), // .+
			shift(800), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(803), // !
			shift(804), // ~
			nil,        // **
			shift(809), // string_literal
			shift(810), // regex_case_insensitive
			shift(811), // int_literal
			shift(812), // float_literal
			shift(813), // boolean_literal
			shift(814), // null_literal
			shift(815), // inf_literal
			shift(816), // nan_literal
			shift(817), // const_M_PI
			shift(818), // const_M_E
			shift(819), // panic
			shift(821), // [
			shift(824), // ctx_IPS
			shift(825), // ctx_IFS
			shift(826), // ctx_IRS
			shift(827), // ctx_OPS
			shift(828), // ctx_OFS
			shift(829), // ctx_ORS
			shift(830), // ctx_FLATSEP
			shift(831), // ctx_NF
			shift(832), // ctx_NR
			shift(833), // ctx_FNR
			shift(834), // ctx_FILENAME
			shift(835), // ctx_FILENUM
			shift(837), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(844), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3379
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(206), // >, reduce: PowTerm
			reduce(206), // >>, reduce: PowTerm
			reduce(206), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(206), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(206), // ], reduce: PowTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(206), // ?, reduce: PowTerm
			nil,         // :
			reduce(206), // ||, reduce: PowTerm
			reduce(206), // ^^, reduce: PowTerm
			reduce(206), // &&, reduce: PowTerm
			reduce(206), // ??, reduce: PowTerm
			reduce(206), // ???, reduce: PowTerm
			reduce(206), // =~, reduce: PowTerm
			reduce(206), // !=~, reduce: PowTerm
			reduce(206), // ==, reduce: PowTerm
			reduce(206), // !=, reduce: PowTerm
			reduce(206), // <=>, reduce: PowTerm
			reduce(206), // >=, reduce: PowTerm
			reduce(206), // <, reduce: PowTerm
			reduce(206), // <=, reduce: PowTerm
			reduce(206), // ^, reduce: PowTerm
			reduce(206), // &, reduce: PowTerm
			reduce(206), // <<, reduce: PowTerm
			reduce(206), // >>>, reduce: PowTerm
			reduce(206), // +, reduce: PowTerm
			reduce(206), // -, reduce: PowTerm
			reduce(206), // .+, reduce: PowTerm
			reduce(206), // .-, reduce: PowTerm
			reduce(206), // *, reduce: PowTerm
			reduce(206), // /, reduce: PowTerm
			reduce(206), // //, reduce: PowTerm
			reduce(206), // %, reduce: PowTerm
			reduce(206), // .*, reduce: PowTerm
			reduce(206), // ./, reduce: PowTerm
			reduce(206), // .//, reduce: PowTerm
			reduce(206), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3380
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(205), // >, reduce: PowTerm
			reduce(205), // >>, reduce: PowTerm
			reduce(205), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(205), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(205), // ], reduce: PowTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(205), // ?, reduce: PowTerm
			nil,         // :
			reduce(205), // ||, reduce: PowTerm
			reduce(205), // ^^, reduce: PowTerm
			reduce(205), // &&, reduce: PowTerm
			reduce(205), // ??, reduce: PowTerm
			reduce(205), // ???, reduce: PowTerm
			reduce(205), // =~, reduce: PowTerm
			reduce(205), // !=~, reduce: PowTerm
			reduce(205), // ==, reduce: PowTerm
			reduce(205), // !=, reduce: PowTerm
			reduce(205), // <=>, reduce: PowTerm
			reduce(205), // >=, reduce: PowTerm
			reduce(205), // <, reduce: PowTerm
			reduce(205), // <=, reduce: PowTerm
			reduce(205), // ^, reduce: PowTerm
			reduce(205), // &, reduce: PowTerm
			reduce(205), // <<, reduce: PowTerm
			reduce(205), // >>>, reduce: PowTerm
			reduce(205), // +, reduce: PowTerm
			reduce(205), // -, reduce: PowTerm
			reduce(205), // .+, reduce: PowTerm
			reduce(205), // .-, reduce: PowTerm
			reduce(205), // *, reduce: PowTerm
			reduce(205), // /, reduce: PowTerm
			reduce(205), // //, reduce: PowTerm
			reduce(205), // %, reduce: PowTerm
			reduce(205), // .*, reduce: PowTerm
			reduce(205), // ./, reduce: PowTerm
			reduce(205), // .//, reduce: PowTerm
			reduce(205), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3381
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(260), // >, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>, reduce: ArrayOrMapIndexAccess
			reduce(260), // |, reduce: ArrayOrMapIndexAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(260), // ,, reduce: ArrayOrMapIndexAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(260), // ], reduce: ArrayOrMapIndexAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(260), // ?, reduce: ArrayOrMapIndexAccess
			nil,         // :
			reduce(260), // ||, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???, reduce: ArrayOrMapIndexAccess
			reduce(260), // =~, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=~, reduce: ArrayOrMapIndexAccess
			reduce(260), // ==, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=>, reduce: ArrayOrMapIndexAccess
			reduce(260), // >=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>, reduce: ArrayOrMapIndexAccess
			reduce(260), // +, reduce: ArrayOrMapIndexAccess
			reduce(260), // -, reduce: ArrayOrMapIndexAccess
			reduce(260), // .+, reduce: ArrayOrMapIndexAccess
			reduce(260), // .-, reduce: ArrayOrMapIndexAccess
			reduce(260), // *, reduce: ArrayOrMapIndexAccess
			reduce(260), // /, reduce: ArrayOrMapIndexAccess
			reduce(260), // //, reduce: ArrayOrMapIndexAccess
			reduce(260), // %, reduce: ArrayOrMapIndexAccess
			reduce(260), // .*, reduce: ArrayOrMapIndexAccess
			reduce(260), // ./, reduce: ArrayOrMapIndexAccess
			reduce(260), // .//, reduce: ArrayOrMapIndexAccess
			reduce(260), // ., reduce: ArrayOrMapIndexAccess
			nil,         // !
			nil,         // ~
			reduce(260), // **, reduce: ArrayOrMapIndexAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(260), // [, reduce: ArrayOrMapIndexAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(260), // [[, reduce: ArrayOrMapIndexAccess
			reduce(260), // [[[, reduce: ArrayOrMapIndexAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3382
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4013), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3383
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4014), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3384
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(266), // >, reduce: ArraySliceAccess
			reduce(266), // >>, reduce: ArraySliceAccess
			reduce(266), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(266), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(266), // ], reduce: ArraySliceAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(266), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(266), // ||, reduce: ArraySliceAccess
			reduce(266), // ^^, reduce: ArraySliceAccess
			reduce(266), // &&, reduce: ArraySliceAccess
			reduce(266), // ??, reduce: ArraySliceAccess
			reduce(266), // ???, reduce: ArraySliceAccess
			reduce(266), // =~, reduce: ArraySliceAccess
			reduce(266), // !=~, reduce: ArraySliceAccess
			reduce(266), // ==, reduce: ArraySliceAccess
			reduce(266), // !=, reduce: ArraySliceAccess
			reduce(266), // <=>, reduce: ArraySliceAccess
			reduce(266), // >=, reduce: ArraySliceAccess
			reduce(266), // <, reduce: ArraySliceAccess
			reduce(266), // <=, reduce: ArraySliceAccess
			reduce(266), // ^, reduce: ArraySliceAccess
			reduce(266), // &, reduce: ArraySliceAccess
			reduce(266), // <<, reduce: ArraySliceAccess
			reduce(266), // >>>, reduce: ArraySliceAccess
			reduce(266), // +, reduce: ArraySliceAccess
			reduce(266), // -, reduce: ArraySliceAccess
			reduce(266), // .+, reduce: ArraySliceAccess
			reduce(266), // .-, reduce: ArraySliceAccess
			reduce(266), // *, reduce: ArraySliceAccess
			reduce(266), // /, reduce: ArraySliceAccess
			reduce(266), // //, reduce: ArraySliceAccess
			reduce(266), // %, reduce: ArraySliceAccess
			reduce(266), // .*, reduce: ArraySliceAccess
			reduce(266), // ./, reduce: ArraySliceAccess
			reduce(266), // .//, reduce: ArraySliceAccess
			reduce(266), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(266), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(266), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(266), // [[, reduce: ArraySliceAccess
			reduce(266), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3385
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4015), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3386
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4016), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3387
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(254), // >, reduce: ENV
			reduce(254), // >>, reduce: ENV
			reduce(254), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(254), // ,, reduce: ENV
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(254), // ], reduce: ENV
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(254), // ?, reduce: ENV
			nil,         // :
			reduce(254), // ||, reduce: ENV
			reduce(254), // ^^, reduce: ENV
			reduce(254), // &&, reduce: ENV
			reduce(254), // ??, reduce: ENV
			reduce(254), // ???, reduce: ENV
			reduce(254), // =~, reduce: ENV
			reduce(254), // !=~, reduce: ENV
			reduce(254), // ==, reduce: ENV
			reduce(254), // !=, reduce: ENV
			reduce(254), // <=>, reduce: ENV
			reduce(254), // >=, reduce: ENV
			reduce(254), // <, reduce: ENV
			reduce(254), // <=, reduce: ENV
			reduce(254), // ^, reduce: ENV
			reduce(254), // &, reduce: ENV
			reduce(254), // <<, reduce: ENV
			reduce(254), // >>>, reduce: ENV
			reduce(254), // +, reduce: ENV
			reduce(254), // -, reduce: ENV
			reduce(254), // .+, reduce: ENV
			reduce(254), // .-, reduce: ENV
			reduce(254), // *, reduce: ENV
			reduce(254), // /, reduce: ENV
			reduce(254), // //, reduce: ENV
			reduce(254), // %, reduce: ENV
			reduce(254), // .*, reduce: ENV
			reduce(254), // ./, reduce: ENV
			reduce(254), // .//, reduce: ENV
			reduce(254), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(254), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(254), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(254), // [[, reduce: ENV
			reduce(254), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3388
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(269), // >, reduce: FunctionCallsite
			reduce(269), // >>, reduce: FunctionCallsite
			reduce(269), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(269), // ,, reduce: FunctionCallsite
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(269), // ], reduce: FunctionCallsite
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(269), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(269), // ||, reduce: FunctionCallsite
			reduce(269), // ^^, reduce: FunctionCallsite
			reduce(269), // &&, reduce: FunctionCallsite
			reduce(269), // ??, reduce: FunctionCallsite
			reduce(269), // ???, reduce: FunctionCallsite
			reduce(269), // =~, reduce: FunctionCallsite
			reduce(269), // !=~, reduce: FunctionCallsite
			reduce(269), // ==, reduce: FunctionCallsite
			reduce(269), // !=, reduce: FunctionCallsite
			reduce(269), // <=>, reduce: FunctionCallsite
			reduce(269), // >=, reduce: FunctionCallsite
			reduce(269), // <, reduce: FunctionCallsite
			reduce(269), // <=, reduce: FunctionCallsite
			reduce(269), // ^, reduce: FunctionCallsite
			reduce(269), // &, reduce: FunctionCallsite
			reduce(269), // <<, reduce: FunctionCallsite
			reduce(269), // >>>, reduce: FunctionCallsite
			reduce(269), // +, reduce: FunctionCallsite
			reduce(269), // -, reduce: FunctionCallsite
			reduce(269), // .+, reduce: FunctionCallsite
			reduce(269), // .-, reduce: FunctionCallsite
			reduce(269), // *, reduce: FunctionCallsite
			reduce(269), // /, reduce: FunctionCallsite
			reduce(269), // //, reduce: FunctionCallsite
			reduce(269), // %, reduce: FunctionCallsite
			reduce(269), // .*, reduce: FunctionCallsite
			reduce(269), // ./, reduce: FunctionCallsite
			reduce(269), // .//, reduce: FunctionCallsite
			reduce(269), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(269), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(269), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(269), // [[, reduce: FunctionCallsite
			reduce(269), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3389
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4018), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4019), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3390
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(235), // >, reduce: MapLiteral
			reduce(235), // >>, reduce: MapLiteral
			reduce(235), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(235), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(235), // ), reduce: MapLiteral
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(235), // ?, reduce: MapLiteral
			nil,         // :
			reduce(235), // ||, reduce: MapLiteral
			reduce(235), // ^^, reduce: MapLiteral
			reduce(235), // &&, reduce: MapLiteral
			reduce(235), // ??, reduce: MapLiteral
			reduce(235), // ???, reduce: MapLiteral
			reduce(235), // =~, reduce: MapLiteral
			reduce(235), // !=~, reduce: MapLiteral
			reduce(235), // ==, reduce: MapLiteral
			reduce(235), // !=, reduce: MapLiteral
			reduce(235), // <=>, reduce: MapLiteral
			reduce(235), // >=, reduce: MapLiteral
			reduce(235), // <, reduce: MapLiteral
			reduce(235), // <=, reduce: MapLiteral
			reduce(235), // ^, reduce: MapLiteral
			reduce(235), // &, reduce: MapLiteral
			reduce(235), // <<, reduce: MapLiteral
			reduce(235), // >>>, reduce: MapLiteral
			reduce(235), // +, reduce: MapLiteral
			reduce(235), // -, reduce: MapLiteral
			reduce(235), // .+, reduce: MapLiteral
			reduce(235), // .-, reduce: MapLiteral
			reduce(235), // *, reduce: MapLiteral
			reduce(235), // /, reduce: MapLiteral
			reduce(235), // //, reduce: MapLiteral
			reduce(235), // %, reduce: MapLiteral
			reduce(235), // .*, reduce: MapLiteral
			reduce(235), // ./, reduce: MapLiteral
			reduce(235), // .//, reduce: MapLiteral
			reduce(235), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(235), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(235), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(235), // [[, reduce: MapLiteral
			reduce(235), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3391
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(275), // ), reduce: FcnArgs
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3392
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(208), // >, reduce: PrecedenceChainEnd
			reduce(208), // >>, reduce: PrecedenceChainEnd
			reduce(208), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(208), // ,, reduce: PrecedenceChainEnd
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(208), // ), reduce: PrecedenceChainEnd
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(208), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(208), // ||, reduce: PrecedenceChainEnd
			reduce(208), // ^^, reduce: PrecedenceChainEnd
			reduce(208), // &&, reduce: PrecedenceChainEnd
			reduce(208), // ??, reduce: PrecedenceChainEnd
			reduce(208), // ???, reduce: PrecedenceChainEnd
			reduce(208), // =~, reduce: PrecedenceChainEnd
			reduce(208), // !=~, reduce: PrecedenceChainEnd
			reduce(208), // ==, reduce: PrecedenceChainEnd
			reduce(208), // !=, reduce: PrecedenceChainEnd
			reduce(208), // <=>, reduce: PrecedenceChainEnd
			reduce(208), // >=, reduce: PrecedenceChainEnd
			reduce(208), // <, reduce: PrecedenceChainEnd
			reduce(208), // <=, reduce: PrecedenceChainEnd
			reduce(208), // ^, reduce: PrecedenceChainEnd
			reduce(208), // &, reduce: PrecedenceChainEnd
			reduce(208), // <<, reduce: PrecedenceChainEnd
			reduce(208), // >>>, reduce: PrecedenceChainEnd
			reduce(208), // +, reduce: PrecedenceChainEnd
			reduce(208), // -, reduce: PrecedenceChainEnd
			reduce(208), // .+, reduce: PrecedenceChainEnd
			reduce(208), // .-, reduce: PrecedenceChainEnd
			reduce(208), // *, reduce: PrecedenceChainEnd
			reduce(208), // /, reduce: PrecedenceChainEnd
			reduce(208), // //, reduce: PrecedenceChainEnd
			reduce(208), // %, reduce: PrecedenceChainEnd
			reduce(208), // .*, reduce: PrecedenceChainEnd
			reduce(208), // ./, reduce: PrecedenceChainEnd
			reduce(208), // .//, reduce: PrecedenceChainEnd
			reduce(208), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(208), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3393
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(105), // >, reduce: IndirectFieldValue
			reduce(105), // >>, reduce: IndirectFieldValue
			reduce(105), // |, reduce: IndirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(105), // ,, reduce: IndirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(105), // ), reduce: IndirectFieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(105), // ?, reduce: IndirectFieldValue
			nil,         // :
			reduce(105), // ||, reduce: IndirectFieldValue
			reduce(105), // ^^, reduce: IndirectFieldValue
			reduce(105), // &&, reduce: IndirectFieldValue
			reduce(105), // ??, reduce: IndirectFieldValue
			reduce(105), // ???, reduce: IndirectFieldValue
			reduce(105), // =~, reduce: IndirectFieldValue
			reduce(105), // !=~, reduce: IndirectFieldValue
			reduce(105), // ==, reduce: IndirectFieldValue
			reduce(105), // !=, reduce: IndirectFieldValue
			reduce(105), // <=>, reduce: IndirectFieldValue
			reduce(105), // >=, reduce: IndirectFieldValue
			reduce(105), // <, reduce: IndirectFieldValue
			reduce(105), // <=, reduce: IndirectFieldValue
			reduce(105), // ^, reduce: IndirectFieldValue
			reduce(105), // &, reduce: IndirectFieldValue
			reduce(105), // <<, reduce: IndirectFieldValue
			reduce(105), // >>>, reduce: IndirectFieldValue
			reduce(105), // +, reduce: IndirectFieldValue
			reduce(105), // -, reduce: IndirectFieldValue
			reduce(105), // .+, reduce: IndirectFieldValue
			reduce(105), // .-, reduce: IndirectFieldValue
			reduce(105), // *, reduce: IndirectFieldValue
			reduce(105), // /, reduce: IndirectFieldValue
			reduce(105), // //, reduce: IndirectFieldValue
			reduce(105), // %, reduce: IndirectFieldValue
			reduce(105), // .*, reduce: IndirectFieldValue
			reduce(105), // ./, reduce: IndirectFieldValue
			reduce(105), // .//, reduce: IndirectFieldValue
			reduce(105), // ., reduce: IndirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(105), // **, reduce: IndirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(105), // [, reduce: IndirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(105), // [[, reduce: IndirectFieldValue
			reduce(105), // [[[, reduce: IndirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3394
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4020), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3395
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4021), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3396
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(114), // >, reduce: IndirectOosvarValue
			reduce(114), // >>, reduce: IndirectOosvarValue
			reduce(114), // |, reduce: IndirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(114), // ,, reduce: IndirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(114), // ), reduce: IndirectOosvarValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(114), // ?, reduce: IndirectOosvarValue
			nil,         // :
			reduce(114), // ||, reduce: IndirectOosvarValue
			reduce(114), // ^^, reduce: IndirectOosvarValue
			reduce(114), // &&, reduce: IndirectOosvarValue
			reduce(114), // ??, reduce: IndirectOosvarValue
			reduce(114), // ???, reduce: IndirectOosvarValue
			reduce(114), // =~, reduce: IndirectOosvarValue
			reduce(114), // !=~, reduce: IndirectOosvarValue
			reduce(114), // ==, reduce: IndirectOosvarValue
			reduce(114), // !=, reduce: IndirectOosvarValue
			reduce(114), // <=>, reduce: IndirectOosvarValue
			reduce(114), // >=, reduce: IndirectOosvarValue
			reduce(114), // <, reduce: IndirectOosvarValue
			reduce(114), // <=, reduce: IndirectOosvarValue
			reduce(114), // ^, reduce: IndirectOosvarValue
			reduce(114), // &, reduce: IndirectOosvarValue
			reduce(114), // <<, reduce: IndirectOosvarValue
			reduce(114), // >>>, reduce: IndirectOosvarValue
			reduce(114), // +, reduce: IndirectOosvarValue
			reduce(114), // -, reduce: IndirectOosvarValue
			reduce(114), // .+, reduce: IndirectOosvarValue
			reduce(114), // .-, reduce: IndirectOosvarValue
			reduce(114), // *, reduce: IndirectOosvarValue
			reduce(114), // /, reduce: IndirectOosvarValue
			reduce(114), // //, reduce: IndirectOosvarValue
			reduce(114), // %, reduce: IndirectOosvarValue
			reduce(114), // .*, reduce: IndirectOosvarValue
			reduce(114), // ./, reduce: IndirectOosvarValue
			reduce(114), // .//, reduce: IndirectOosvarValue
			reduce(114), // ., reduce: IndirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(114), // **, reduce: IndirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(114), // [, reduce: IndirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(114), // [[, reduce: IndirectOosvarValue
			reduce(114), // [[[, reduce: IndirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3397
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4022), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3398
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(151), // ,, reduce: LogicalOrTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(151), // ), reduce: LogicalOrTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(151), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(151), // ||, reduce: LogicalOrTerm
			shift(2761), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3399
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(153), // ,, reduce: LogicalXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(153), // ), reduce: LogicalXORTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(153), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(153), // ||, reduce: LogicalXORTerm
			reduce(153), // ^^, reduce: LogicalXORTerm
			shift(2762), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3400
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(155), // ,, reduce: LogicalAndTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(155), // ), reduce: LogicalAndTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(155), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(155), // ||, reduce: LogicalAndTerm
			reduce(155), // ^^, reduce: LogicalAndTerm
			reduce(155), // &&, reduce: LogicalAndTerm
			shift(2763), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3401
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(157), // ,, reduce: AbsentCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(157), // ), reduce: AbsentCoalesceTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(157), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(157), // ||, reduce: AbsentCoalesceTerm
			reduce(157), // ^^, reduce: AbsentCoalesceTerm
			reduce(157), // &&, reduce: AbsentCoalesceTerm
			reduce(157), // ??, reduce: AbsentCoalesceTerm
			shift(2764), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3402
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(159), // ,, reduce: EmptyCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(159), // ), reduce: EmptyCoalesceTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(159), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(159), // ||, reduce: EmptyCoalesceTerm
			reduce(159), // ^^, reduce: EmptyCoalesceTerm
			reduce(159), // &&, reduce: EmptyCoalesceTerm
			reduce(159), // ??, reduce: EmptyCoalesceTerm
			reduce(159), // ???, reduce: EmptyCoalesceTerm
			shift(2765), // =~
			shift(2766), // !=~
			shift(2767), // ==
			shift(2768), // !=
			shift(2769), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3403
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2770), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(161), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(161), // ), reduce: EqneTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(161), // ?, reduce: EqneTerm
			nil,         // :
			reduce(161), // ||, reduce: EqneTerm
			reduce(161), // ^^, reduce: EqneTerm
			reduce(161), // &&, reduce: EqneTerm
			reduce(161), // ??, reduce: EqneTerm
			reduce(161), // ???, reduce: EqneTerm
			reduce(161), // =~, reduce: EqneTerm
			reduce(161), // !=~, reduce: EqneTerm
			reduce(161), // ==, reduce: EqneTerm
			reduce(161), // !=, reduce: EqneTerm
			reduce(161), // <=>, reduce: EqneTerm
			shift(2771), // >=
			shift(2772), // <
			shift(2773), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3404
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2770), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(162), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(162), // ), reduce: EqneTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(162), // ?, reduce: EqneTerm
			nil,         // :
			reduce(162), // ||, reduce: EqneTerm
			reduce(162), // ^^, reduce: EqneTerm
			reduce(162), // &&, reduce: EqneTerm
			reduce(162), // ??, reduce: EqneTerm
			reduce(162), // ???, reduce: EqneTerm
			reduce(162), // =~, reduce: EqneTerm
			reduce(162), // !=~, reduce: EqneTerm
			reduce(162), // ==, reduce: EqneTerm
			reduce(162), // !=, reduce: EqneTerm
			reduce(162), // <=>, reduce: EqneTerm
			shift(2771), // >=
			shift(2772), // <
			shift(2773), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3405
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2770), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(163), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(163), // ), reduce: EqneTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(163), // ?, reduce: EqneTerm
			nil,         // :
			reduce(163), // ||, reduce: EqneTerm
			reduce(163), // ^^, reduce: EqneTerm
			reduce(163), // &&, reduce: EqneTerm
			reduce(163), // ??, reduce: EqneTerm
			reduce(163), // ???, reduce: EqneTerm
			reduce(163), // =~, reduce: EqneTerm
			reduce(163), // !=~, reduce: EqneTerm
			reduce(163), // ==, reduce: EqneTerm
			reduce(163), // !=, reduce: EqneTerm
			reduce(163), // <=>, reduce: EqneTerm
			shift(2771), // >=
			shift(2772), // <
			shift(2773), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3406
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2770), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(164), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(164), // ), reduce: EqneTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(164), // ?, reduce: EqneTerm
			nil,         // :
			reduce(164), // ||, reduce: EqneTerm
			reduce(164), // ^^, reduce: EqneTerm
			reduce(164), // &&, reduce: EqneTerm
			reduce(164), // ??, reduce: EqneTerm
			reduce(164), // ???, reduce: EqneTerm
			reduce(164), // =~, reduce: EqneTerm
			reduce(164), // !=~, reduce: EqneTerm
			reduce(164), // ==, reduce: EqneTerm
			reduce(164), // !=, reduce: EqneTerm
			reduce(164), // <=>, reduce: EqneTerm
			shift(2771), // >=
			shift(2772), // <
			shift(2773), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3407
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2770), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(165), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(165), // ), reduce: EqneTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(165), // ?, reduce: EqneTerm
			nil,         // :
			reduce(165), // ||, reduce: EqneTerm
			reduce(165), // ^^, reduce: EqneTerm
			reduce(165), // &&, reduce: EqneTerm
			reduce(165), // ??, reduce: EqneTerm
			reduce(165), // ???, reduce: EqneTerm
			reduce(165), // =~, reduce: EqneTerm
			reduce(165), // !=~, reduce: EqneTerm
			reduce(165), // ==, reduce: EqneTerm
			reduce(165), // !=, reduce: EqneTerm
			reduce(165), // <=>, reduce: EqneTerm
			shift(2771), // >=
			shift(2772), // <
			shift(2773), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3408
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(167), // >, reduce: CmpTerm
			nil,         // >>
			shift(2774), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(167), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(167), // ), reduce: CmpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(167), // ?, reduce: CmpTerm
			nil,         // :
			reduce(167), // ||, reduce: CmpTerm
			reduce(167), // ^^, reduce: CmpTerm
			reduce(167), // &&, reduce: CmpTerm
			reduce(167), // ??, reduce: CmpTerm
			reduce(167), // ???, reduce: CmpTerm
			reduce(167), // =~, reduce: CmpTerm
			reduce(167), // !=~, reduce: CmpTerm
			reduce(167), // ==, reduce: CmpTerm
			reduce(167), // !=, reduce: CmpTerm
			reduce(167), // <=>, reduce: CmpTerm
			reduce(167), // >=, reduce: CmpTerm
			reduce(167), // <, reduce: CmpTerm
			reduce(167), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3409
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(168), // >, reduce: CmpTerm
			nil,         // >>
			shift(2774), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(168), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(168), // ), reduce: CmpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(168), // ?, reduce: CmpTerm
			nil,         // :
			reduce(168), // ||, reduce: CmpTerm
			reduce(168), // ^^, reduce: CmpTerm
			reduce(168), // &&, reduce: CmpTerm
			reduce(168), // ??, reduce: CmpTerm
			reduce(168), // ???, reduce: CmpTerm
			reduce(168), // =~, reduce: CmpTerm
			reduce(168), // !=~, reduce: CmpTerm
			reduce(168), // ==, reduce: CmpTerm
			reduce(168), // !=, reduce: CmpTerm
			reduce(168), // <=>, reduce: CmpTerm
			reduce(168), // >=, reduce: CmpTerm
			reduce(168), // <, reduce: CmpTerm
			reduce(168), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3410
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(169), // >, reduce: CmpTerm
			nil,         // >>
			shift(2774), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(169), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(169), // ), reduce: CmpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(169), // ?, reduce: CmpTerm
			nil,         // :
			reduce(169), // ||, reduce: CmpTerm
			reduce(169), // ^^, reduce: CmpTerm
			reduce(169), // &&, reduce: CmpTerm
			reduce(169), // ??, reduce: CmpTerm
			reduce(169), // ???, reduce: CmpTerm
			reduce(169), // =~, reduce: CmpTerm
			reduce(169), // !=~, reduce: CmpTerm
			reduce(169), // ==, reduce: CmpTerm
			reduce(169), // !=, reduce: CmpTerm
			reduce(169), // <=>, reduce: CmpTerm
			reduce(169), // >=, reduce: CmpTerm
			reduce(169), // <, reduce: CmpTerm
			reduce(169), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3411
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(170), // >, reduce: CmpTerm
			nil,         // >>
			shift(2774), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(170), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(170), // ), reduce: CmpTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(170), // ?, reduce: CmpTerm
			nil,         // :
			reduce(170), // ||, reduce: CmpTerm
			reduce(170), // ^^, reduce: CmpTerm
			reduce(170), // &&, reduce: CmpTerm
			reduce(170), // ??, reduce: CmpTerm
			reduce(170), // ???, reduce: CmpTerm
			reduce(170), // =~, reduce: CmpTerm
			reduce(170), // !=~, reduce: CmpTerm
			reduce(170), // ==, reduce: CmpTerm
			reduce(170), // !=, reduce: CmpTerm
			reduce(170), // <=>, reduce: CmpTerm
			reduce(170), // >=, reduce: CmpTerm
			reduce(170), // <, reduce: CmpTerm
			reduce(170), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3412
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(172), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(172), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(172), // ,, reduce: BitwiseORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(172), // ), reduce: BitwiseORTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(172), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(172), // ||, reduce: BitwiseORTerm
			reduce(172), // ^^, reduce: BitwiseORTerm
			reduce(172), // &&, reduce: BitwiseORTerm
			reduce(172), // ??, reduce: BitwiseORTerm
			reduce(172), // ???, reduce: BitwiseORTerm
			reduce(172), // =~, reduce: BitwiseORTerm
			reduce(172), // !=~, reduce: BitwiseORTerm
			reduce(172), // ==, reduce: BitwiseORTerm
			reduce(172), // !=, reduce: BitwiseORTerm
			reduce(172), // <=>, reduce: BitwiseORTerm
			reduce(172), // >=, reduce: BitwiseORTerm
			reduce(172), // <, reduce: BitwiseORTerm
			reduce(172), // <=, reduce: BitwiseORTerm
			shift(2775), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3413
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(174), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(174), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(174), // ,, reduce: BitwiseXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(174), // ), reduce: BitwiseXORTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(174), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(174), // ||, reduce: BitwiseXORTerm
			reduce(174), // ^^, reduce: BitwiseXORTerm
			reduce(174), // &&, reduce: BitwiseXORTerm
			reduce(174), // ??, reduce: BitwiseXORTerm
			reduce(174), // ???, reduce: BitwiseXORTerm
			reduce(174), // =~, reduce: BitwiseXORTerm
			reduce(174), // !=~, reduce: BitwiseXORTerm
			reduce(174), // ==, reduce: BitwiseXORTerm
			reduce(174), // !=, reduce: BitwiseXORTerm
			reduce(174), // <=>, reduce: BitwiseXORTerm
			reduce(174), // >=, reduce: BitwiseXORTerm
			reduce(174), // <, reduce: BitwiseXORTerm
			reduce(174), // <=, reduce: BitwiseXORTerm
			reduce(174), // ^, reduce: BitwiseXORTerm
			shift(2776), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3414
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(176), // >, reduce: BitwiseANDTerm
			shift(2777), // >>
			reduce(176), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(176), // ,, reduce: BitwiseANDTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(176), // ), reduce: BitwiseANDTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(176), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(176), // ||, reduce: BitwiseANDTerm
			reduce(176), // ^^, reduce: BitwiseANDTerm
			reduce(176), // &&, reduce: BitwiseANDTerm
			reduce(176), // ??, reduce: BitwiseANDTerm
			reduce(176), // ???, reduce: BitwiseANDTerm
			reduce(176), // =~, reduce: BitwiseANDTerm
			reduce(176), // !=~, reduce: BitwiseANDTerm
			reduce(176), // ==, reduce: BitwiseANDTerm
			reduce(176), // !=, reduce: BitwiseANDTerm
			reduce(176), // <=>, reduce: BitwiseANDTerm
			reduce(176), // >=, reduce: BitwiseANDTerm
			reduce(176), // <, reduce: BitwiseANDTerm
			reduce(176), // <=, reduce: BitwiseANDTerm
			reduce(176), // ^, reduce: BitwiseANDTerm
			reduce(176), // &, reduce: BitwiseANDTerm
			shift(2778), // <<
			shift(2779), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3415
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(179), // >, reduce: BitwiseShiftTerm
			reduce(179), // >>, reduce: BitwiseShiftTerm
			reduce(179), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(179), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(179), // ), reduce: BitwiseShiftTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(179), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(179), // ||, reduce: BitwiseShiftTerm
			reduce(179), // ^^, reduce: BitwiseShiftTerm
			reduce(179), // &&, reduce: BitwiseShiftTerm
			reduce(179), // ??, reduce: BitwiseShiftTerm
			reduce(179), // ???, reduce: BitwiseShiftTerm
			reduce(179), // =~, reduce: BitwiseShiftTerm
			reduce(179), // !=~, reduce: BitwiseShiftTerm
			reduce(179), // ==, reduce: BitwiseShiftTerm
			reduce(179), // !=, reduce: BitwiseShiftTerm
			reduce(179), // <=>, reduce: BitwiseShiftTerm
			reduce(179), // >=, reduce: BitwiseShiftTerm
			reduce(179), // <, reduce: BitwiseShiftTerm
			reduce(179), // <=, reduce: BitwiseShiftTerm
			reduce(179), // ^, reduce: BitwiseShiftTerm
			reduce(179), // &, reduce: BitwiseShiftTerm
			reduce(179), // <<, reduce: BitwiseShiftTerm
			reduce(179), // >>>, reduce: BitwiseShiftTerm
			shift(2780), // +
			shift(2781), // -
			shift(2782), // .+
			shift(2783), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3416
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(178), // >, reduce: BitwiseShiftTerm
			reduce(178), // >>, reduce: BitwiseShiftTerm
			reduce(178), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(178), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(178), // ), reduce: BitwiseShiftTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(178), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(178), // ||, reduce: BitwiseShiftTerm
			reduce(178), // ^^, reduce: BitwiseShiftTerm
			reduce(178), // &&, reduce: BitwiseShiftTerm
			reduce(178), // ??, reduce: BitwiseShiftTerm
			reduce(178), // ???, reduce: BitwiseShiftTerm
			reduce(178), // =~, reduce: BitwiseShiftTerm
			reduce(178), // !=~, reduce: BitwiseShiftTerm
			reduce(178), // ==, reduce: BitwiseShiftTerm
			reduce(178), // !=, reduce: BitwiseShiftTerm
			reduce(178), // <=>, reduce: BitwiseShiftTerm
			reduce(178), // >=, reduce: BitwiseShiftTerm
			reduce(178), // <, reduce: BitwiseShiftTerm
			reduce(178), // <=, reduce: BitwiseShiftTerm
			reduce(178), // ^, reduce: BitwiseShiftTerm
			reduce(178), // &, reduce: BitwiseShiftTerm
			reduce(178), // <<, reduce: BitwiseShiftTerm
			reduce(178), // >>>, reduce: BitwiseShiftTerm
			shift(2780), // +
			shift(2781), // -
			shift(2782), // .+
			shift(2783), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3417
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(180), // >, reduce: BitwiseShiftTerm
			reduce(180), // >>, reduce: BitwiseShiftTerm
			reduce(180), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(180), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(180), // ), reduce: BitwiseShiftTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(180), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(180), // ||, reduce: BitwiseShiftTerm
			reduce(180), // ^^, reduce: BitwiseShiftTerm
			reduce(180), // &&, reduce: BitwiseShiftTerm
			reduce(180), // ??, reduce: BitwiseShiftTerm
			reduce(180), // ???, reduce: BitwiseShiftTerm
			reduce(180), // =~, reduce: BitwiseShiftTerm
			reduce(180), // !=~, reduce: BitwiseShiftTerm
			reduce(180), // ==, reduce: BitwiseShiftTerm
			reduce(180), // !=, reduce: BitwiseShiftTerm
			reduce(180), // <=>, reduce: BitwiseShiftTerm
			reduce(180), // >=, reduce: BitwiseShiftTerm
			reduce(180), // <, reduce: BitwiseShiftTerm
			reduce(180), // <=, reduce: BitwiseShiftTerm
			reduce(180), // ^, reduce: BitwiseShiftTerm
			reduce(180), // &, reduce: BitwiseShiftTerm
			reduce(180), // <<, reduce: BitwiseShiftTerm
			reduce(180), // >>>, reduce: BitwiseShiftTerm
			shift(2780), // +
			shift(2781), // -
			shift(2782), // .+
			shift(2783), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3418
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(182), // >, reduce: AddsubdotTerm
			reduce(182), // >>, reduce: AddsubdotTerm
			reduce(182), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(182), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(182), // ), reduce: AddsubdotTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(182), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(182), // ||, reduce: AddsubdotTerm
			reduce(182), // ^^, reduce: AddsubdotTerm
			reduce(182), // &&, reduce: AddsubdotTerm
			reduce(182), // ??, reduce: AddsubdotTerm
			reduce(182), // ???, reduce: AddsubdotTerm
			reduce(182), // =~, reduce: AddsubdotTerm
			reduce(182), // !=~, reduce: AddsubdotTerm
			reduce(182), // ==, reduce: AddsubdotTerm
			reduce(182), // !=, reduce: AddsubdotTerm
			reduce(182), // <=>, reduce: AddsubdotTerm
			reduce(182), // >=, reduce: AddsubdotTerm
			reduce(182), // <, reduce: AddsubdotTerm
			reduce(182), // <=, reduce: AddsubdotTerm
			reduce(182), // ^, reduce: AddsubdotTerm
			reduce(182), // &, reduce: AddsubdotTerm
			reduce(182), // <<, reduce: AddsubdotTerm
			reduce(182), // >>>, reduce: AddsubdotTerm
			reduce(182), // +, reduce: AddsubdotTerm
			reduce(182), // -, reduce: AddsubdotTerm
			reduce(182), // .+, reduce: AddsubdotTerm
			reduce(182), // .-, reduce: AddsubdotTerm
			shift(2785), // *
			shift(2786), // /
			shift(2787), // //
			shift(2788), // %
			shift(2789), // .*
			shift(2790), // ./
			shift(2791), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3419
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(183), // >, reduce: AddsubdotTerm
			reduce(183), // >>, reduce: AddsubdotTerm
			reduce(183), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(183), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(183), // ), reduce: AddsubdotTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(183), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(183), // ||, reduce: AddsubdotTerm
			reduce(183), // ^^, reduce: AddsubdotTerm
			reduce(183), // &&, reduce: AddsubdotTerm
			reduce(183), // ??, reduce: AddsubdotTerm
			reduce(183), // ???, reduce: AddsubdotTerm
			reduce(183), // =~, reduce: AddsubdotTerm
			reduce(183), // !=~, reduce: AddsubdotTerm
			reduce(183), // ==, reduce: AddsubdotTerm
			reduce(183), // !=, reduce: AddsubdotTerm
			reduce(183), // <=>, reduce: AddsubdotTerm
			reduce(183), // >=, reduce: AddsubdotTerm
			reduce(183), // <, reduce: AddsubdotTerm
			reduce(183), // <=, reduce: AddsubdotTerm
			reduce(183), // ^, reduce: AddsubdotTerm
			reduce(183), // &, reduce: AddsubdotTerm
			reduce(183), // <<, reduce: AddsubdotTerm
			reduce(183), // >>>, reduce: AddsubdotTerm
			reduce(183), // +, reduce: AddsubdotTerm
			reduce(183), // -, reduce: AddsubdotTerm
			reduce(183), // .+, reduce: AddsubdotTerm
			reduce(183), // .-, reduce: AddsubdotTerm
			shift(2785), // *
			shift(2786), // /
			shift(2787), // //
			shift(2788), // %
			shift(2789), // .*
			shift(2790), // ./
			shift(2791), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3420
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(184), // >, reduce: AddsubdotTerm
			reduce(184), // >>, reduce: AddsubdotTerm
			reduce(184), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(184), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(184), // ), reduce: AddsubdotTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(184), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(184), // ||, reduce: AddsubdotTerm
			reduce(184), // ^^, reduce: AddsubdotTerm
			reduce(184), // &&, reduce: AddsubdotTerm
			reduce(184), // ??, reduce: AddsubdotTerm
			reduce(184), // ???, reduce: AddsubdotTerm
			reduce(184), // =~, reduce: AddsubdotTerm
			reduce(184), // !=~, reduce: AddsubdotTerm
			reduce(184), // ==, reduce: AddsubdotTerm
			reduce(184), // !=, reduce: AddsubdotTerm
			reduce(184), // <=>, reduce: AddsubdotTerm
			reduce(184), // >=, reduce: AddsubdotTerm
			reduce(184), // <, reduce: AddsubdotTerm
			reduce(184), // <=, reduce: AddsubdotTerm
			reduce(184), // ^, reduce: AddsubdotTerm
			reduce(184), // &, reduce: AddsubdotTerm
			reduce(184), // <<, reduce: AddsubdotTerm
			reduce(184), // >>>, reduce: AddsubdotTerm
			reduce(184), // +, reduce: AddsubdotTerm
			reduce(184), // -, reduce: AddsubdotTerm
			reduce(184), // .+, reduce: AddsubdotTerm
			reduce(184), // .-, reduce: AddsubdotTerm
			shift(2785), // *
			shift(2786), // /
			shift(2787), // //
			shift(2788), // %
			shift(2789), // .*
			shift(2790), // ./
			shift(2791), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3421
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(185), // >, reduce: AddsubdotTerm
			reduce(185), // >>, reduce: AddsubdotTerm
			reduce(185), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(185), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(185), // ), reduce: AddsubdotTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(185), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(185), // ||, reduce: AddsubdotTerm
			reduce(185), // ^^, reduce: AddsubdotTerm
			reduce(185), // &&, reduce: AddsubdotTerm
			reduce(185), // ??, reduce: AddsubdotTerm
			reduce(185), // ???, reduce: AddsubdotTerm
			reduce(185), // =~, reduce: AddsubdotTerm
			reduce(185), // !=~, reduce: AddsubdotTerm
			reduce(185), // ==, reduce: AddsubdotTerm
			reduce(185), // !=, reduce: AddsubdotTerm
			reduce(185), // <=>, reduce: AddsubdotTerm
			reduce(185), // >=, reduce: AddsubdotTerm
			reduce(185), // <, reduce: AddsubdotTerm
			reduce(185), // <=, reduce: AddsubdotTerm
			reduce(185), // ^, reduce: AddsubdotTerm
			reduce(185), // &, reduce: AddsubdotTerm
			reduce(185), // <<, reduce: AddsubdotTerm
			reduce(185), // >>>, reduce: AddsubdotTerm
			reduce(185), // +, reduce: AddsubdotTerm
			reduce(185), // -, reduce: AddsubdotTerm
			reduce(185), // .+, reduce: AddsubdotTerm
			reduce(185), // .-, reduce: AddsubdotTerm
			shift(2785), // *
			shift(2786), // /
			shift(2787), // //
			shift(2788), // %
			shift(2789), // .*
			shift(2790), // ./
			shift(2791), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3422
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(187), // >, reduce: MuldivTerm
			reduce(187), // >>, reduce: MuldivTerm
			reduce(187), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(187), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(187), // ), reduce: MuldivTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(187), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(187), // ||, reduce: MuldivTerm
			reduce(187), // ^^, reduce: MuldivTerm
			reduce(187), // &&, reduce: MuldivTerm
			reduce(187), // ??, reduce: MuldivTerm
			reduce(187), // ???, reduce: MuldivTerm
			reduce(187), // =~, reduce: MuldivTerm
			reduce(187), // !=~, reduce: MuldivTerm
			reduce(187), // ==, reduce: MuldivTerm
			reduce(187), // !=, reduce: MuldivTerm
			reduce(187), // <=>, reduce: MuldivTerm
			reduce(187), // >=, reduce: MuldivTerm
			reduce(187), // <, reduce: MuldivTerm
			reduce(187), // <=, reduce: MuldivTerm
			reduce(187), // ^, reduce: MuldivTerm
			reduce(187), // &, reduce: MuldivTerm
			reduce(187), // <<, reduce: MuldivTerm
			reduce(187), // >>>, reduce: MuldivTerm
			reduce(187), // +, reduce: MuldivTerm
			reduce(187), // -, reduce: MuldivTerm
			reduce(187), // .+, reduce: MuldivTerm
			reduce(187), // .-, reduce: MuldivTerm
			reduce(187), // *, reduce: MuldivTerm
			reduce(187), // /, reduce: MuldivTerm
			reduce(187), // //, reduce: MuldivTerm
			reduce(187), // %, reduce: MuldivTerm
			reduce(187), // .*, reduce: MuldivTerm
			reduce(187), // ./, reduce: MuldivTerm
			reduce(187), // .//, reduce: MuldivTerm
			shift(2795), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3423
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(188), // >, reduce: MuldivTerm
			reduce(188), // >>, reduce: MuldivTerm
			reduce(188), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(188), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(188), // ), reduce: MuldivTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(188), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(188), // ||, reduce: MuldivTerm
			reduce(188), // ^^, reduce: MuldivTerm
			reduce(188), // &&, reduce: MuldivTerm
			reduce(188), // ??, reduce: MuldivTerm
			reduce(188), // ???, reduce: MuldivTerm
			reduce(188), // =~, reduce: MuldivTerm
			reduce(188), // !=~, reduce: MuldivTerm
			reduce(188), // ==, reduce: MuldivTerm
			reduce(188), // !=, reduce: MuldivTerm
			reduce(188), // <=>, reduce: MuldivTerm
			reduce(188), // >=, reduce: MuldivTerm
			reduce(188), // <, reduce: MuldivTerm
			reduce(188), // <=, reduce: MuldivTerm
			reduce(188), // ^, reduce: MuldivTerm
			reduce(188), // &, reduce: MuldivTerm
			reduce(188), // <<, reduce: MuldivTerm
			reduce(188), // >>>, reduce: MuldivTerm
			reduce(188), // +, reduce: MuldivTerm
			reduce(188), // -, reduce: MuldivTerm
			reduce(188), // .+, reduce: MuldivTerm
			reduce(188), // .-, reduce: MuldivTerm
			reduce(188), // *, reduce: MuldivTerm
			reduce(188), // /, reduce: MuldivTerm
			reduce(188), // //, reduce: MuldivTerm
			reduce(188), // %, reduce: MuldivTerm
			reduce(188), // .*, reduce: MuldivTerm
			reduce(188), // ./, reduce: MuldivTerm
			reduce(188), // .//, reduce: MuldivTerm
			shift(2795), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3424
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(189), // >, reduce: MuldivTerm
			reduce(189), // >>, reduce: MuldivTerm
			reduce(189), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(189), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(189), // ), reduce: MuldivTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(189), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(189), // ||, reduce: MuldivTerm
			reduce(189), // ^^, reduce: MuldivTerm
			reduce(189), // &&, reduce: MuldivTerm
			reduce(189), // ??, reduce: MuldivTerm
			reduce(189), // ???, reduce: MuldivTerm
			reduce(189), // =~, reduce: MuldivTerm
			reduce(189), // !=~, reduce: MuldivTerm
			reduce(189), // ==, reduce: MuldivTerm
			reduce(189), // !=, reduce: MuldivTerm
			reduce(189), // <=>, reduce: MuldivTerm
			reduce(189), // >=, reduce: MuldivTerm
			reduce(189), // <, reduce: MuldivTerm
			reduce(189), // <=, reduce: MuldivTerm
			reduce(189), // ^, reduce: MuldivTerm
			reduce(189), // &, reduce: MuldivTerm
			reduce(189), // <<, reduce: MuldivTerm
			reduce(189), // >>>, reduce: MuldivTerm
			reduce(189), // +, reduce: MuldivTerm
			reduce(189), // -, reduce: MuldivTerm
			reduce(189), // .+, reduce: MuldivTerm
			reduce(189), // .-, reduce: MuldivTerm
			reduce(189), // *, reduce: MuldivTerm
			reduce(189), // /, reduce: MuldivTerm
			reduce(189), // //, reduce: MuldivTerm
			reduce(189), // %, reduce: MuldivTerm
			reduce(189), // .*, reduce: MuldivTerm
			reduce(189), // ./, reduce: MuldivTerm
			reduce(189), // .//, reduce: MuldivTerm
			shift(2795), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3425
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(190), // >, reduce: MuldivTerm
			reduce(190), // >>, reduce: MuldivTerm
			reduce(190), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(190), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(190), // ), reduce: MuldivTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(190), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(190), // ||, reduce: MuldivTerm
			reduce(190), // ^^, reduce: MuldivTerm
			reduce(190), // &&, reduce: MuldivTerm
			reduce(190), // ??, reduce: MuldivTerm
			reduce(190), // ???, reduce: MuldivTerm
			reduce(190), // =~, reduce: MuldivTerm
			reduce(190), // !=~, reduce: MuldivTerm
			reduce(190), // ==, reduce: MuldivTerm
			reduce(190), // !=, reduce: MuldivTerm
			reduce(190), // <=>, reduce: MuldivTerm
			reduce(190), // >=, reduce: MuldivTerm
			reduce(190), // <, reduce: MuldivTerm
			reduce(190), // <=, reduce: MuldivTerm
			reduce(190), // ^, reduce: MuldivTerm
			reduce(190), // &, reduce: MuldivTerm
			reduce(190), // <<, reduce: MuldivTerm
			reduce(190), // >>>, reduce: MuldivTerm
			reduce(190), // +, reduce: MuldivTerm
			reduce(190), // -, reduce: MuldivTerm
			reduce(190), // .+, reduce: MuldivTerm
			reduce(190), // .-, reduce: MuldivTerm
			reduce(190), // *, reduce: MuldivTerm
			reduce(190), // /, reduce: MuldivTerm
			reduce(190), // //, reduce: MuldivTerm
			reduce(190), // %, reduce: MuldivTerm
			reduce(190), // .*, reduce: MuldivTerm
			reduce(190), // ./, reduce: MuldivTerm
			reduce(190), // .//, reduce: MuldivTerm
			shift(2795), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3426
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(191), // >, reduce: MuldivTerm
			reduce(191), // >>, reduce: MuldivTerm
			reduce(191), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(191), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(191), // ), reduce: MuldivTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(191), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(191), // ||, reduce: MuldivTerm
			reduce(191), // ^^, reduce: MuldivTerm
			reduce(191), // &&, reduce: MuldivTerm
			reduce(191), // ??, reduce: MuldivTerm
			reduce(191), // ???, reduce: MuldivTerm
			reduce(191), // =~, reduce: MuldivTerm
			reduce(191), // !=~, reduce: MuldivTerm
			reduce(191), // ==, reduce: MuldivTerm
			reduce(191), // !=, reduce: MuldivTerm
			reduce(191), // <=>, reduce: MuldivTerm
			reduce(191), // >=, reduce: MuldivTerm
			reduce(191), // <, reduce: MuldivTerm
			reduce(191), // <=, reduce: MuldivTerm
			reduce(191), // ^, reduce: MuldivTerm
			reduce(191), // &, reduce: MuldivTerm
			reduce(191), // <<, reduce: MuldivTerm
			reduce(191), // >>>, reduce: MuldivTerm
			reduce(191), // +, reduce: MuldivTerm
			reduce(191), // -, reduce: MuldivTerm
			reduce(191), // .+, reduce: MuldivTerm
			reduce(191), // .-, reduce: MuldivTerm
			reduce(191), // *, reduce: MuldivTerm
			reduce(191), // /, reduce: MuldivTerm
			reduce(191), // //, reduce: MuldivTerm
			reduce(191), // %, reduce: MuldivTerm
			reduce(191), // .*, reduce: MuldivTerm
			reduce(191), // ./, reduce: MuldivTerm
			reduce(191), // .//, reduce: MuldivTerm
			shift(2795), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3427
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(192), // >, reduce: MuldivTerm
			reduce(192), // >>, reduce: MuldivTerm
			reduce(192), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(192), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(192), // ), reduce: MuldivTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(192), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(192), // ||, reduce: MuldivTerm
			reduce(192), // ^^, reduce: MuldivTerm
			reduce(192), // &&, reduce: MuldivTerm
			reduce(192), // ??, reduce: MuldivTerm
			reduce(192), // ???, reduce: MuldivTerm
			reduce(192), // =~, reduce: MuldivTerm
			reduce(192), // !=~, reduce: MuldivTerm
			reduce(192), // ==, reduce: MuldivTerm
			reduce(192), // !=, reduce: MuldivTerm
			reduce(192), // <=>, reduce: MuldivTerm
			reduce(192), // >=, reduce: MuldivTerm
			reduce(192), // <, reduce: MuldivTerm
			reduce(192), // <=, reduce: MuldivTerm
			reduce(192), // ^, reduce: MuldivTerm
			reduce(192), // &, reduce: MuldivTerm
			reduce(192), // <<, reduce: MuldivTerm
			reduce(192), // >>>, reduce: MuldivTerm
			reduce(192), // +, reduce: MuldivTerm
			reduce(192), // -, reduce: MuldivTerm
			reduce(192), // .+, reduce: MuldivTerm
			reduce(192), // .-, reduce: MuldivTerm
			reduce(192), // *, reduce: MuldivTerm
			reduce(192), // /, reduce: MuldivTerm
			reduce(192), // //, reduce: MuldivTerm
			reduce(192), // %, reduce: MuldivTerm
			reduce(192), // .*, reduce: MuldivTerm
			reduce(192), // ./, reduce: MuldivTerm
			reduce(192), // .//, reduce: MuldivTerm
			shift(2795), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3428
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(193), // >, reduce: MuldivTerm
			reduce(193), // >>, reduce: MuldivTerm
			reduce(193), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(193), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(193), // ), reduce: MuldivTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(193), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(193), // ||, reduce: MuldivTerm
			reduce(193), // ^^, reduce: MuldivTerm
			reduce(193), // &&, reduce: MuldivTerm
			reduce(193), // ??, reduce: MuldivTerm
			reduce(193), // ???, reduce: MuldivTerm
			reduce(193), // =~, reduce: MuldivTerm
			reduce(193), // !=~, reduce: MuldivTerm
			reduce(193), // ==, reduce: MuldivTerm
			reduce(193), // !=, reduce: MuldivTerm
			reduce(193), // <=>, reduce: MuldivTerm
			reduce(193), // >=, reduce: MuldivTerm
			reduce(193), // <, reduce: MuldivTerm
			reduce(193), // <=, reduce: MuldivTerm
			reduce(193), // ^, reduce: MuldivTerm
			reduce(193), // &, reduce: MuldivTerm
			reduce(193), // <<, reduce: MuldivTerm
			reduce(193), // >>>, reduce: MuldivTerm
			reduce(193), // +, reduce: MuldivTerm
			reduce(193), // -, reduce: MuldivTerm
			reduce(193), // .+, reduce: MuldivTerm
			reduce(193), // .-, reduce: MuldivTerm
			reduce(193), // *, reduce: MuldivTerm
			reduce(193), // /, reduce: MuldivTerm
			reduce(193), // //, reduce: MuldivTerm
			reduce(193), // %, reduce: MuldivTerm
			reduce(193), // .*, reduce: MuldivTerm
			reduce(193), // ./, reduce: MuldivTerm
			reduce(193), // .//, reduce: MuldivTerm
			shift(2795), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3429
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(195), // >, reduce: DotTerm
			reduce(195), // >>, reduce: DotTerm
			reduce(195), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(195), // ,, reduce: DotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(195), // ), reduce: DotTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(195), // ?, reduce: DotTerm
			nil,         // :
			reduce(195), // ||, reduce: DotTerm
			reduce(195), // ^^, reduce: DotTerm
			reduce(195), // &&, reduce: DotTerm
			reduce(195), // ??, reduce: DotTerm
			reduce(195), // ???, reduce: DotTerm
			reduce(195), // =~, reduce: DotTerm
			reduce(195), // !=~, reduce: DotTerm
			reduce(195), // ==, reduce: DotTerm
			reduce(195), // !=, reduce: DotTerm
			reduce(195), // <=>, reduce: DotTerm
			reduce(195), // >=, reduce: DotTerm
			reduce(195), // <, reduce: DotTerm
			reduce(195), // <=, reduce: DotTerm
			reduce(195), // ^, reduce: DotTerm
			reduce(195), // &, reduce: DotTerm
			reduce(195), // <<, reduce: DotTerm
			reduce(195), // >>>, reduce: DotTerm
			reduce(195), // +, reduce: DotTerm
			reduce(195), // -, reduce: DotTerm
			reduce(195), // .+, reduce: DotTerm
			reduce(195), // .-, reduce: DotTerm
			reduce(195), // *, reduce: DotTerm
			reduce(195), // /, reduce: DotTerm
			reduce(195), // //, reduce: DotTerm
			reduce(195), // %, reduce: DotTerm
			reduce(195), // .*, reduce: DotTerm
			reduce(195), // ./, reduce: DotTerm
			reduce(195), // .//, reduce: DotTerm
			reduce(195), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3430
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3431
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3432
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(204), // >, reduce: PowTerm
			reduce(204), // >>, reduce: PowTerm
			reduce(204), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(204), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(204), // ), reduce: PowTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(204), // ?, reduce: PowTerm
			nil,         // :
			reduce(204), // ||, reduce: PowTerm
			reduce(204), // ^^, reduce: PowTerm
			reduce(204), // &&, reduce: PowTerm
			reduce(204), // ??, reduce: PowTerm
			reduce(204), // ???, reduce: PowTerm
			reduce(204), // =~, reduce: PowTerm
			reduce(204), // !=~, reduce: PowTerm
			reduce(204), // ==, reduce: PowTerm
			reduce(204), // !=, reduce: PowTerm
			reduce(204), // <=>, reduce: PowTerm
			reduce(204), // >=, reduce: PowTerm
			reduce(204), // <, reduce: PowTerm
			reduce(204), // <=, reduce: PowTerm
			reduce(204), // ^, reduce: PowTerm
			reduce(204), // &, reduce: PowTerm
			reduce(204), // <<, reduce: PowTerm
			reduce(204), // >>>, reduce: PowTerm
			reduce(204), // +, reduce: PowTerm
			reduce(204), // -, reduce: PowTerm
			reduce(204), // .+, reduce: PowTerm
			reduce(204), // .-, reduce: PowTerm
			reduce(204), // *, reduce: PowTerm
			reduce(204), // /, reduce: PowTerm
			reduce(204), // //, reduce: PowTerm
			reduce(204), // %, reduce: PowTerm
			reduce(204), // .*, reduce: PowTerm
			reduce(204), // ./, reduce: PowTerm
			reduce(204), // .//, reduce: PowTerm
			reduce(204), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3433
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4025), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4026), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3434
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4028), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3435
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4029), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3436
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4030), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3437
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(229), // >, reduce: ArrayLiteral
			reduce(229), // >>, reduce: ArrayLiteral
			reduce(229), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(229), // ,, reduce: ArrayLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(229), // ), reduce: ArrayLiteral
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(229), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(229), // ||, reduce: ArrayLiteral
			reduce(229), // ^^, reduce: ArrayLiteral
			reduce(229), // &&, reduce: ArrayLiteral
			reduce(229), // ??, reduce: ArrayLiteral
			reduce(229), // ???, reduce: ArrayLiteral
			reduce(229), // =~, reduce: ArrayLiteral
			reduce(229), // !=~, reduce: ArrayLiteral
			reduce(229), // ==, reduce: ArrayLiteral
			reduce(229), // !=, reduce: ArrayLiteral
			reduce(229), // <=>, reduce: ArrayLiteral
			reduce(229), // >=, reduce: ArrayLiteral
			reduce(229), // <, reduce: ArrayLiteral
			reduce(229), // <=, reduce: ArrayLiteral
			reduce(229), // ^, reduce: ArrayLiteral
			reduce(229), // &, reduce: ArrayLiteral
			reduce(229), // <<, reduce: ArrayLiteral
			reduce(229), // >>>, reduce: ArrayLiteral
			reduce(229), // +, reduce: ArrayLiteral
			reduce(229), // -, reduce: ArrayLiteral
			reduce(229), // .+, reduce: ArrayLiteral
			reduce(229), // .-, reduce: ArrayLiteral
			reduce(229), // *, reduce: ArrayLiteral
			reduce(229), // /, reduce: ArrayLiteral
			reduce(229), // //, reduce: ArrayLiteral
			reduce(229), // %, reduce: ArrayLiteral
			reduce(229), // .*, reduce: ArrayLiteral
			reduce(229), // ./, reduce: ArrayLiteral
			reduce(229), // .//, reduce: ArrayLiteral
			reduce(229), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(229), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(229), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(229), // [[, reduce: ArrayLiteral
			reduce(229), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3438
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(255), // >, reduce: ENV
			reduce(255), // >>, reduce: ENV
			reduce(255), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(255), // ,, reduce: ENV
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(255), // ), reduce: ENV
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(255), // ?, reduce: ENV
			nil,         // :
			reduce(255), // ||, reduce: ENV
			reduce(255), // ^^, reduce: ENV
			reduce(255), // &&, reduce: ENV
			reduce(255), // ??, reduce: ENV
			reduce(255), // ???, reduce: ENV
			reduce(255), // =~, reduce: ENV
			reduce(255), // !=~, reduce: ENV
			reduce(255), // ==, reduce: ENV
			reduce(255), // !=, reduce: ENV
			reduce(255), // <=>, reduce: ENV
			reduce(255), // >=, reduce: ENV
			reduce(255), // <, reduce: ENV
			reduce(255), // <=, reduce: ENV
			reduce(255), // ^, reduce: ENV
			reduce(255), // &, reduce: ENV
			reduce(255), // <<, reduce: ENV
			reduce(255), // >>>, reduce: ENV
			reduce(255), // +, reduce: ENV
			reduce(255), // -, reduce: ENV
			reduce(255), // .+, reduce: ENV
			reduce(255), // .-, reduce: ENV
			reduce(255), // *, reduce: ENV
			reduce(255), // /, reduce: ENV
			reduce(255), // //, reduce: ENV
			reduce(255), // %, reduce: ENV
			reduce(255), // .*, reduce: ENV
			reduce(255), // ./, reduce: ENV
			reduce(255), // .//, reduce: ENV
			reduce(255), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(255), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(255), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(255), // [[, reduce: ENV
			reduce(255), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3439
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4031), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3440
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4032), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3441
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(268), // >, reduce: FunctionCallsite
			reduce(268), // >>, reduce: FunctionCallsite
			reduce(268), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(268), // ,, reduce: FunctionCallsite
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(268), // ), reduce: FunctionCallsite
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(268), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(268), // ||, reduce: FunctionCallsite
			reduce(268), // ^^, reduce: FunctionCallsite
			reduce(268), // &&, reduce: FunctionCallsite
			reduce(268), // ??, reduce: FunctionCallsite
			reduce(268), // ???, reduce: FunctionCallsite
			reduce(268), // =~, reduce: FunctionCallsite
			reduce(268), // !=~, reduce: FunctionCallsite
			reduce(268), // ==, reduce: FunctionCallsite
			reduce(268), // !=, reduce: FunctionCallsite
			reduce(268), // <=>, reduce: FunctionCallsite
			reduce(268), // >=, reduce: FunctionCallsite
			reduce(268), // <, reduce: FunctionCallsite
			reduce(268), // <=, reduce: FunctionCallsite
			reduce(268), // ^, reduce: FunctionCallsite
			reduce(268), // &, reduce: FunctionCallsite
			reduce(268), // <<, reduce: FunctionCallsite
			reduce(268), // >>>, reduce: FunctionCallsite
			reduce(268), // +, reduce: FunctionCallsite
			reduce(268), // -, reduce: FunctionCallsite
			reduce(268), // .+, reduce: FunctionCallsite
			reduce(268), // .-, reduce: FunctionCallsite
			reduce(268), // *, reduce: FunctionCallsite
			reduce(268), // /, reduce: FunctionCallsite
			reduce(268), // //, reduce: FunctionCallsite
			reduce(268), // %, reduce: FunctionCallsite
			reduce(268), // .*, reduce: FunctionCallsite
			reduce(268), // ./, reduce: FunctionCallsite
			reduce(268), // .//, reduce: FunctionCallsite
			reduce(268), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(268), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(268), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(268), // [[, reduce: FunctionCallsite
			reduce(268), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3442
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4033), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3443
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(277), // $, reduce: SubroutineCallsite
			nil,         // empty
			reduce(277), // ;, reduce: SubroutineCallsite
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3444
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3446), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3445
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(294), // $, reduce: IfBlock
			nil,         // empty
			reduce(294), // ;, reduce: IfBlock
			reduce(294), // {, reduce: IfBlock
			nil,         // }
			nil,         // =
			reduce(294), // unset, reduce: IfBlock
			reduce(294), // filter, reduce: IfBlock
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(294), // print, reduce: IfBlock
			nil,         // ,
			reduce(294), // printn, reduce: IfBlock
			reduce(294), // eprint, reduce: IfBlock
			reduce(294), // eprintn, reduce: IfBlock
			reduce(294), // dump, reduce: IfBlock
			reduce(294), // edump, reduce: IfBlock
			reduce(294), // tee, reduce: IfBlock
			reduce(294), // emitf, reduce: IfBlock
			reduce(294), // emit1, reduce: IfBlock
			reduce(294), // emit, reduce: IfBlock
			reduce(294), // (, reduce: IfBlock
			nil,         // )
			reduce(294), // emitp, reduce: IfBlock
			reduce(294), // field_name, reduce: IfBlock
			reduce(294), // $[, reduce: IfBlock
			nil,         // ]
			reduce(294), // braced_field_name, reduce: IfBlock
			reduce(294), // $[[, reduce: IfBlock
			reduce(294), // $[[[, reduce: IfBlock
			reduce(294), // full_srec, reduce: IfBlock
			reduce(294), // oosvar_name, reduce: IfBlock
			reduce(294), // @[, reduce: IfBlock
			reduce(294), // braced_oosvar_name, reduce: IfBlock
			reduce(294), // full_oosvar, reduce: IfBlock
			reduce(294), // all, reduce: IfBlock
			reduce(294), // non_sigil_name, reduce: IfBlock
			reduce(294), // arr, reduce: IfBlock
			reduce(294), // bool, reduce: IfBlock
			reduce(294), // float, reduce: IfBlock
			reduce(294), // int, reduce: IfBlock
			reduce(294), // map, reduce: IfBlock
			reduce(294), // num, reduce: IfBlock
			reduce(294), // str, reduce: IfBlock
			reduce(294), // var, reduce: IfBlock
			reduce(294), // funct, reduce: IfBlock
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(294), // +, reduce: IfBlock
			reduce(294), // -, reduce: IfBlock
			reduce(294), // .+, reduce: IfBlock
			reduce(294), // .-, reduce: IfBlock
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(294), // !, reduce: IfBlock
			reduce(294), // ~, reduce: IfBlock
			nil,         // **
			reduce(294), // string_literal, reduce: IfBlock
			reduce(294), // regex_case_insensitive, reduce: IfBlock
			reduce(294), // int_literal, reduce: IfBlock
			reduce(294), // float_literal, reduce: IfBlock
			reduce(294), // boolean_literal, reduce: IfBlock
			reduce(294), // null_literal, reduce: IfBlock
			reduce(294), // inf_literal, reduce: IfBlock
			reduce(294), // nan_literal, reduce: IfBlock
			reduce(294), // const_M_PI, reduce: IfBlock
			reduce(294), // const_M_E, reduce: IfBlock
			reduce(294), // panic, reduce: IfBlock
			reduce(294), // [, reduce: IfBlock
			reduce(294), // ctx_IPS, reduce: IfBlock
			reduce(294), // ctx_IFS, reduce: IfBlock
			reduce(294), // ctx_IRS, reduce: IfBlock
			reduce(294), // ctx_OPS, reduce: IfBlock
			reduce(294), // ctx_OFS, reduce: IfBlock
			reduce(294), // ctx_ORS, reduce: IfBlock
			reduce(294), // ctx_FLATSEP, reduce: IfBlock
			reduce(294), // ctx_NF, reduce: IfBlock
			reduce(294), // ctx_NR, reduce: IfBlock
			reduce(294), // ctx_FNR, reduce: IfBlock
			reduce(294), // ctx_FILENAME, reduce: IfBlock
			reduce(294), // ctx_FILENUM, reduce: IfBlock
			reduce(294), // env, reduce: IfBlock
			nil,         // [[
			nil,         // [[[
			reduce(294), // call, reduce: IfBlock
			reduce(294), // begin, reduce: IfBlock
			reduce(294), // end, reduce: IfBlock
			reduce(294), // if, reduce: IfBlock
			reduce(294), // elif, reduce: IfBlock
			reduce(294), // else, reduce: IfBlock
			reduce(294), // while, reduce: IfBlock
			reduce(294), // do, reduce: IfBlock
			reduce(294), // for, reduce: IfBlock
			nil,         // in
			reduce(294), // break, reduce: IfBlock
			reduce(294), // continue, reduce: IfBlock
			reduce(294), // func, reduce: IfBlock
			reduce(294), // subr, reduce: IfBlock
			reduce(294), // return, reduce: IfBlock
		},
	},
	actionRow{ // S3446
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S3447
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(297), // $, reduce: WhileLoop
			nil,         // empty
			reduce(297), // ;, reduce: WhileLoop
			reduce(297), // {, reduce: WhileLoop
			nil,         // }
			nil,         // =
			reduce(297), // unset, reduce: WhileLoop
			reduce(297), // filter, reduce: WhileLoop
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(297), // print, reduce: WhileLoop
			nil,         // ,
			reduce(297), // printn, reduce: WhileLoop
			reduce(297), // eprint, reduce: WhileLoop
			reduce(297), // eprintn, reduce: WhileLoop
			reduce(297), // dump, reduce: WhileLoop
			reduce(297), // edump, reduce: WhileLoop
			reduce(297), // tee, reduce: WhileLoop
			reduce(297), // emitf, reduce: WhileLoop
			reduce(297), // emit1, reduce: WhileLoop
			reduce(297), // emit, reduce: WhileLoop
			reduce(297), // (, reduce: WhileLoop
			nil,         // )
			reduce(297), // emitp, reduce: WhileLoop
			reduce(297), // field_name, reduce: WhileLoop
			reduce(297), // $[, reduce: WhileLoop
			nil,         // ]
			reduce(297), // braced_field_name, reduce: WhileLoop
			reduce(297), // $[[, reduce: WhileLoop
			reduce(297), // $[[[, reduce: WhileLoop
			reduce(297), // full_srec, reduce: WhileLoop
			reduce(297), // oosvar_name, reduce: WhileLoop
			reduce(297), // @[, reduce: WhileLoop
			reduce(297), // braced_oosvar_name, reduce: WhileLoop
			reduce(297), // full_oosvar, reduce: WhileLoop
			reduce(297), // all, reduce: WhileLoop
			reduce(297), // non_sigil_name, reduce: WhileLoop
			reduce(297), // arr, reduce: WhileLoop
			reduce(297), // bool, reduce: WhileLoop
			reduce(297), // float, reduce: WhileLoop
			reduce(297), // int, reduce: WhileLoop
			reduce(297), // map, reduce: WhileLoop
			reduce(297), // num, reduce: WhileLoop
			reduce(297), // str, reduce: WhileLoop
			reduce(297), // var, reduce: WhileLoop
			reduce(297), // funct, reduce: WhileLoop
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(297), // +, reduce: WhileLoop
			reduce(297), // -, reduce: WhileLoop
			reduce(297), // .+, reduce: WhileLoop
			reduce(297), // .-, reduce: WhileLoop
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(297), // !, reduce: WhileLoop
			reduce(297), // ~, reduce: WhileLoop
			nil,         // **
			reduce(297), // string_literal, reduce: WhileLoop
			reduce(297), // regex_case_insensitive, reduce: WhileLoop
			reduce(297), // int_literal, reduce: WhileLoop
			reduce(297), // float_literal, reduce: WhileLoop
			reduce(297), // boolean_literal, reduce: WhileLoop
			reduce(297), // null_literal, reduce: WhileLoop
			reduce(297), // inf_literal, reduce: WhileLoop
			reduce(297), // nan_literal, reduce: WhileLoop
			reduce(297), // const_M_PI, reduce: WhileLoop
			reduce(297), // const_M_E, reduce: WhileLoop
			reduce(297), // panic, reduce: WhileLoop
			reduce(297), // [, reduce: WhileLoop
			reduce(297), // ctx_IPS, reduce: WhileLoop
			reduce(297), // ctx_IFS, reduce: WhileLoop
			reduce(297), // ctx_IRS, reduce: WhileLoop
			reduce(297), // ctx_OPS, reduce: WhileLoop
			reduce(297), // ctx_OFS, reduce: WhileLoop
			reduce(297), // ctx_ORS, reduce: WhileLoop
			reduce(297), // ctx_FLATSEP, reduce: WhileLoop
			reduce(297), // ctx_NF, reduce: WhileLoop
			reduce(297), // ctx_NR, reduce: WhileLoop
			reduce(297), // ctx_FNR, reduce: WhileLoop
			reduce(297), // ctx_FILENAME, reduce: WhileLoop
			reduce(297), // ctx_FILENUM, reduce: WhileLoop
			reduce(297), // env, reduce: WhileLoop
			nil,         // [[
			nil,         // [[[
			reduce(297), // call, reduce: WhileLoop
			reduce(297), // begin, reduce: WhileLoop
			reduce(297), // end, reduce: WhileLoop
			reduce(297), // if, reduce: WhileLoop
			nil,         // elif
			nil,         // else
			reduce(297), // while, reduce: WhileLoop
			reduce(297), // do, reduce: WhileLoop
			reduce(297), // for, reduce: WhileLoop
			nil,         // in
			reduce(297), // break, reduce: WhileLoop
			reduce(297), // continue, reduce: WhileLoop
			reduce(297), // func, reduce: WhileLoop
			reduce(297), // subr, reduce: WhileLoop
			reduce(297), // return, reduce: WhileLoop
		},
	},
	actionRow{ // S3448
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4036), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3449
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(235), // =, reduce: MapLiteral
			nil,         // unset
			nil,         // filter
			reduce(235), // >, reduce: MapLiteral
			reduce(235), // >>, reduce: MapLiteral
			reduce(235), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(235), // ||=, reduce: MapLiteral
			reduce(235), // ^^=, reduce: MapLiteral
			reduce(235), // &&=, reduce: MapLiteral
			reduce(235), // ??=, reduce: MapLiteral
			reduce(235), // ???=, reduce: MapLiteral
			reduce(235), // |=, reduce: MapLiteral
			reduce(235), // &=, reduce: MapLiteral
			reduce(235), // ^=, reduce: MapLiteral
			reduce(235), // <<=, reduce: MapLiteral
			reduce(235), // >>=, reduce: MapLiteral
			reduce(235), // >>>=, reduce: MapLiteral
			reduce(235), // +=, reduce: MapLiteral
			reduce(235), // .=, reduce: MapLiteral
			reduce(235), // -=, reduce: MapLiteral
			reduce(235), // *=, reduce: MapLiteral
			reduce(235), // /=, reduce: MapLiteral
			reduce(235), // //=, reduce: MapLiteral
			reduce(235), // %=, reduce: MapLiteral
			reduce(235), // **=, reduce: MapLiteral
			reduce(235), // ?, reduce: MapLiteral
			nil,         // :
			reduce(235), // ||, reduce: MapLiteral
			reduce(235), // ^^, reduce: MapLiteral
			reduce(235), // &&, reduce: MapLiteral
			reduce(235), // ??, reduce: MapLiteral
			reduce(235), // ???, reduce: MapLiteral
			reduce(235), // =~, reduce: MapLiteral
			reduce(235), // !=~, reduce: MapLiteral
			reduce(235), // ==, reduce: MapLiteral
			reduce(235), // !=, reduce: MapLiteral
			reduce(235), // <=>, reduce: MapLiteral
			reduce(235), // >=, reduce: MapLiteral
			reduce(235), // <, reduce: MapLiteral
			reduce(235), // <=, reduce: MapLiteral
			reduce(235), // ^, reduce: MapLiteral
			reduce(235), // &, reduce: MapLiteral
			reduce(235), // <<, reduce: MapLiteral
			reduce(235), // >>>, reduce: MapLiteral
			reduce(235), // +, reduce: MapLiteral
			reduce(235), // -, reduce: MapLiteral
			reduce(235), // .+, reduce: MapLiteral
			reduce(235), // .-, reduce: MapLiteral
			reduce(235), // *, reduce: MapLiteral
			reduce(235), // /, reduce: MapLiteral
			reduce(235), // //, reduce: MapLiteral
			reduce(235), // %, reduce: MapLiteral
			reduce(235), // .*, reduce: MapLiteral
			reduce(235), // ./, reduce: MapLiteral
			reduce(235), // .//, reduce: MapLiteral
			reduce(235), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(235), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(235), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(235), // [[, reduce: MapLiteral
			reduce(235), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3450
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			shift(4037), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3451
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(34), // ;, reduce: Assignment
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(34), // ,, reduce: Assignment
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3452
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(214), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(214), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(214), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3453
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(211), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(211), // >, reduce: MlrvalOrFunction
			reduce(211), // >>, reduce: MlrvalOrFunction
			reduce(211), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(211), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(211), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(211), // ||, reduce: MlrvalOrFunction
			reduce(211), // ^^, reduce: MlrvalOrFunction
			reduce(211), // &&, reduce: MlrvalOrFunction
			reduce(211), // ??, reduce: MlrvalOrFunction
			reduce(211), // ???, reduce: MlrvalOrFunction
			reduce(211), // =~, reduce: MlrvalOrFunction
			reduce(211), // !=~, reduce: MlrvalOrFunction
			reduce(211), // ==, reduce: MlrvalOrFunction
			reduce(211), // !=, reduce: MlrvalOrFunction
			reduce(211), // <=>, reduce: MlrvalOrFunction
			reduce(211), // >=, reduce: MlrvalOrFunction
			reduce(211), // <, reduce: MlrvalOrFunction
			reduce(211), // <=, reduce: MlrvalOrFunction
			reduce(211), // ^, reduce: MlrvalOrFunction
			reduce(211), // &, reduce: MlrvalOrFunction
			reduce(211), // <<, reduce: MlrvalOrFunction
			reduce(211), // >>>, reduce: MlrvalOrFunction
			reduce(211), // +, reduce: MlrvalOrFunction
			reduce(211), // -, reduce: MlrvalOrFunction
			reduce(211), // .+, reduce: MlrvalOrFunction
			reduce(211), // .-, reduce: MlrvalOrFunction
			reduce(211), // *, reduce: MlrvalOrFunction
			reduce(211), // /, reduce: MlrvalOrFunction
			reduce(211), // //, reduce: MlrvalOrFunction
			reduce(211), // %, reduce: MlrvalOrFunction
			reduce(211), // .*, reduce: MlrvalOrFunction
			reduce(211), // ./, reduce: MlrvalOrFunction
			reduce(211), // .//, reduce: MlrvalOrFunction
			reduce(211), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(211), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(211), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(211), // [[, reduce: MlrvalOrFunction
			reduce(211), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3454
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3455
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(110), // ;, reduce: OosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(110), // >, reduce: OosvarValue
			reduce(110), // >>, reduce: OosvarValue
			reduce(110), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(110), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(110), // ?, reduce: OosvarValue
			nil,         // :
			reduce(110), // ||, reduce: OosvarValue
			reduce(110), // ^^, reduce: OosvarValue
			reduce(110), // &&, reduce: OosvarValue
			reduce(110), // ??, reduce: OosvarValue
			reduce(110), // ???, reduce: OosvarValue
			reduce(110), // =~, reduce: OosvarValue
			reduce(110), // !=~, reduce: OosvarValue
			reduce(110), // ==, reduce: OosvarValue
			reduce(110), // !=, reduce: OosvarValue
			reduce(110), // <=>, reduce: OosvarValue
			reduce(110), // >=, reduce: OosvarValue
			reduce(110), // <, reduce: OosvarValue
			reduce(110), // <=, reduce: OosvarValue
			reduce(110), // ^, reduce: OosvarValue
			reduce(110), // &, reduce: OosvarValue
			reduce(110), // <<, reduce: OosvarValue
			reduce(110), // >>>, reduce: OosvarValue
			reduce(110), // +, reduce: OosvarValue
			reduce(110), // -, reduce: OosvarValue
			reduce(110), // .+, reduce: OosvarValue
			reduce(110), // .-, reduce: OosvarValue
			reduce(110), // *, reduce: OosvarValue
			reduce(110), // /, reduce: OosvarValue
			reduce(110), // //, reduce: OosvarValue
			reduce(110), // %, reduce: OosvarValue
			reduce(110), // .*, reduce: OosvarValue
			reduce(110), // ./, reduce: OosvarValue
			reduce(110), // .//, reduce: OosvarValue
			reduce(110), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(110), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(110), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(110), // [[, reduce: OosvarValue
			reduce(110), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3456
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(112), // ;, reduce: OosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(112), // >, reduce: OosvarValue
			reduce(112), // >>, reduce: OosvarValue
			reduce(112), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(112), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(112), // ?, reduce: OosvarValue
			nil,         // :
			reduce(112), // ||, reduce: OosvarValue
			reduce(112), // ^^, reduce: OosvarValue
			reduce(112), // &&, reduce: OosvarValue
			reduce(112), // ??, reduce: OosvarValue
			reduce(112), // ???, reduce: OosvarValue
			reduce(112), // =~, reduce: OosvarValue
			reduce(112), // !=~, reduce: OosvarValue
			reduce(112), // ==, reduce: OosvarValue
			reduce(112), // !=, reduce: OosvarValue
			reduce(112), // <=>, reduce: OosvarValue
			reduce(112), // >=, reduce: OosvarValue
			reduce(112), // <, reduce: OosvarValue
			reduce(112), // <=, reduce: OosvarValue
			reduce(112), // ^, reduce: OosvarValue
			reduce(112), // &, reduce: OosvarValue
			reduce(112), // <<, reduce: OosvarValue
			reduce(112), // >>>, reduce: OosvarValue
			reduce(112), // +, reduce: OosvarValue
			reduce(112), // -, reduce: OosvarValue
			reduce(112), // .+, reduce: OosvarValue
			reduce(112), // .-, reduce: OosvarValue
			reduce(112), // *, reduce: OosvarValue
			reduce(112), // /, reduce: OosvarValue
			reduce(112), // //, reduce: OosvarValue
			reduce(112), // %, reduce: OosvarValue
			reduce(112), // .*, reduce: OosvarValue
			reduce(112), // ./, reduce: OosvarValue
			reduce(112), // .//, reduce: OosvarValue
			reduce(112), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(112), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(112), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(112), // [[, reduce: OosvarValue
			reduce(112), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3457
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(111), // ;, reduce: OosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(111), // >, reduce: OosvarValue
			reduce(111), // >>, reduce: OosvarValue
			reduce(111), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(111), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(111), // ?, reduce: OosvarValue
			nil,         // :
			reduce(111), // ||, reduce: OosvarValue
			reduce(111), // ^^, reduce: OosvarValue
			reduce(111), // &&, reduce: OosvarValue
			reduce(111), // ??, reduce: OosvarValue
			reduce(111), // ???, reduce: OosvarValue
			reduce(111), // =~, reduce: OosvarValue
			reduce(111), // !=~, reduce: OosvarValue
			reduce(111), // ==, reduce: OosvarValue
			reduce(111), // !=, reduce: OosvarValue
			reduce(111), // <=>, reduce: OosvarValue
			reduce(111), // >=, reduce: OosvarValue
			reduce(111), // <, reduce: OosvarValue
			reduce(111), // <=, reduce: OosvarValue
			reduce(111), // ^, reduce: OosvarValue
			reduce(111), // &, reduce: OosvarValue
			reduce(111), // <<, reduce: OosvarValue
			reduce(111), // >>>, reduce: OosvarValue
			reduce(111), // +, reduce: OosvarValue
			reduce(111), // -, reduce: OosvarValue
			reduce(111), // .+, reduce: OosvarValue
			reduce(111), // .-, reduce: OosvarValue
			reduce(111), // *, reduce: OosvarValue
			reduce(111), // /, reduce: OosvarValue
			reduce(111), // //, reduce: OosvarValue
			reduce(111), // %, reduce: OosvarValue
			reduce(111), // .*, reduce: OosvarValue
			reduce(111), // ./, reduce: OosvarValue
			reduce(111), // .//, reduce: OosvarValue
			reduce(111), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(111), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(111), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(111), // [[, reduce: OosvarValue
			reduce(111), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3458
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(99), // ;, reduce: FieldValue
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(99), // >, reduce: FieldValue
			reduce(99), // >>, reduce: FieldValue
			reduce(99), // |, reduce: FieldValue
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(99), // ,, reduce: FieldValue
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(99), // ?, reduce: FieldValue
			nil,        // :
			reduce(99), // ||, reduce: FieldValue
			reduce(99), // ^^, reduce: FieldValue
			reduce(99), // &&, reduce: FieldValue
			reduce(99), // ??, reduce: FieldValue
			reduce(99), // ???, reduce: FieldValue
			reduce(99), // =~, reduce: FieldValue
			reduce(99), // !=~, reduce: FieldValue
			reduce(99), // ==, reduce: FieldValue
			reduce(99), // !=, reduce: FieldValue
			reduce(99), // <=>, reduce: FieldValue
			reduce(99), // >=, reduce: FieldValue
			reduce(99), // <, reduce: FieldValue
			reduce(99), // <=, reduce: FieldValue
			reduce(99), // ^, reduce: FieldValue
			reduce(99), // &, reduce: FieldValue
			reduce(99), // <<, reduce: FieldValue
			reduce(99), // >>>, reduce: FieldValue
			reduce(99), // +, reduce: FieldValue
			reduce(99), // -, reduce: FieldValue
			reduce(99), // .+, reduce: FieldValue
			reduce(99), // .-, reduce: FieldValue
			reduce(99), // *, reduce: FieldValue
			reduce(99), // /, reduce: FieldValue
			reduce(99), // //, reduce: FieldValue
			reduce(99), // %, reduce: FieldValue
			reduce(99), // .*, reduce: FieldValue
			reduce(99), // ./, reduce: FieldValue
			reduce(99), // .//, reduce: FieldValue
			reduce(99), // ., reduce: FieldValue
			nil,        // !
			nil,        // ~
			reduce(99), // **, reduce: FieldValue
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(99), // [, reduce: FieldValue
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(99), // [[, reduce: FieldValue
			reduce(99), // [[[, reduce: FieldValue
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3459
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(101), // ;, reduce: FieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(101), // >, reduce: FieldValue
			reduce(101), // >>, reduce: FieldValue
			reduce(101), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(101), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(101), // ?, reduce: FieldValue
			nil,         // :
			reduce(101), // ||, reduce: FieldValue
			reduce(101), // ^^, reduce: FieldValue
			reduce(101), // &&, reduce: FieldValue
			reduce(101), // ??, reduce: FieldValue
			reduce(101), // ???, reduce: FieldValue
			reduce(101), // =~, reduce: FieldValue
			reduce(101), // !=~, reduce: FieldValue
			reduce(101), // ==, reduce: FieldValue
			reduce(101), // !=, reduce: FieldValue
			reduce(101), // <=>, reduce: FieldValue
			reduce(101), // >=, reduce: FieldValue
			reduce(101), // <, reduce: FieldValue
			reduce(101), // <=, reduce: FieldValue
			reduce(101), // ^, reduce: FieldValue
			reduce(101), // &, reduce: FieldValue
			reduce(101), // <<, reduce: FieldValue
			reduce(101), // >>>, reduce: FieldValue
			reduce(101), // +, reduce: FieldValue
			reduce(101), // -, reduce: FieldValue
			reduce(101), // .+, reduce: FieldValue
			reduce(101), // .-, reduce: FieldValue
			reduce(101), // *, reduce: FieldValue
			reduce(101), // /, reduce: FieldValue
			reduce(101), // //, reduce: FieldValue
			reduce(101), // %, reduce: FieldValue
			reduce(101), // .*, reduce: FieldValue
			reduce(101), // ./, reduce: FieldValue
			reduce(101), // .//, reduce: FieldValue
			reduce(101), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(101), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(101), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(101), // [[, reduce: FieldValue
			reduce(101), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3460
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(100), // ;, reduce: FieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(100), // >, reduce: FieldValue
			reduce(100), // >>, reduce: FieldValue
			reduce(100), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(100), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(100), // ?, reduce: FieldValue
			nil,         // :
			reduce(100), // ||, reduce: FieldValue
			reduce(100), // ^^, reduce: FieldValue
			reduce(100), // &&, reduce: FieldValue
			reduce(100), // ??, reduce: FieldValue
			reduce(100), // ???, reduce: FieldValue
			reduce(100), // =~, reduce: FieldValue
			reduce(100), // !=~, reduce: FieldValue
			reduce(100), // ==, reduce: FieldValue
			reduce(100), // !=, reduce: FieldValue
			reduce(100), // <=>, reduce: FieldValue
			reduce(100), // >=, reduce: FieldValue
			reduce(100), // <, reduce: FieldValue
			reduce(100), // <=, reduce: FieldValue
			reduce(100), // ^, reduce: FieldValue
			reduce(100), // &, reduce: FieldValue
			reduce(100), // <<, reduce: FieldValue
			reduce(100), // >>>, reduce: FieldValue
			reduce(100), // +, reduce: FieldValue
			reduce(100), // -, reduce: FieldValue
			reduce(100), // .+, reduce: FieldValue
			reduce(100), // .-, reduce: FieldValue
			reduce(100), // *, reduce: FieldValue
			reduce(100), // /, reduce: FieldValue
			reduce(100), // //, reduce: FieldValue
			reduce(100), // %, reduce: FieldValue
			reduce(100), // .*, reduce: FieldValue
			reduce(100), // ./, reduce: FieldValue
			reduce(100), // .//, reduce: FieldValue
			reduce(100), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(100), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(100), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(100), // [[, reduce: FieldValue
			reduce(100), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3461
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(213), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(213), // >, reduce: MlrvalOrFunction
			reduce(213), // >>, reduce: MlrvalOrFunction
			reduce(213), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(213), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(213), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(213), // ||, reduce: MlrvalOrFunction
			reduce(213), // ^^, reduce: MlrvalOrFunction
			reduce(213), // &&, reduce: MlrvalOrFunction
			reduce(213), // ??, reduce: MlrvalOrFunction
			reduce(213), // ???, reduce: MlrvalOrFunction
			reduce(213), // =~, reduce: MlrvalOrFunction
			reduce(213), // !=~, reduce: MlrvalOrFunction
			reduce(213), // ==, reduce: MlrvalOrFunction
			reduce(213), // !=, reduce: MlrvalOrFunction
			reduce(213), // <=>, reduce: MlrvalOrFunction
			reduce(213), // >=, reduce: MlrvalOrFunction
			reduce(213), // <, reduce: MlrvalOrFunction
			reduce(213), // <=, reduce: MlrvalOrFunction
			reduce(213), // ^, reduce: MlrvalOrFunction
			reduce(213), // &, reduce: MlrvalOrFunction
			reduce(213), // <<, reduce: MlrvalOrFunction
			reduce(213), // >>>, reduce: MlrvalOrFunction
			reduce(213), // +, reduce: MlrvalOrFunction
			reduce(213), // -, reduce: MlrvalOrFunction
			reduce(213), // .+, reduce: MlrvalOrFunction
			reduce(213), // .-, reduce: MlrvalOrFunction
			reduce(213), // *, reduce: MlrvalOrFunction
			reduce(213), // /, reduce: MlrvalOrFunction
			reduce(213), // //, reduce: MlrvalOrFunction
			reduce(213), // %, reduce: MlrvalOrFunction
			reduce(213), // .*, reduce: MlrvalOrFunction
			reduce(213), // ./, reduce: MlrvalOrFunction
			reduce(213), // .//, reduce: MlrvalOrFunction
			reduce(213), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(213), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(213), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(213), // [[, reduce: MlrvalOrFunction
			reduce(213), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3462
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(233), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(233), // >, reduce: MlrvalOrFunction
			reduce(233), // >>, reduce: MlrvalOrFunction
			reduce(233), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(233), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(233), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(233), // ||, reduce: MlrvalOrFunction
			reduce(233), // ^^, reduce: MlrvalOrFunction
			reduce(233), // &&, reduce: MlrvalOrFunction
			reduce(233), // ??, reduce: MlrvalOrFunction
			reduce(233), // ???, reduce: MlrvalOrFunction
			reduce(233), // =~, reduce: MlrvalOrFunction
			reduce(233), // !=~, reduce: MlrvalOrFunction
			reduce(233), // ==, reduce: MlrvalOrFunction
			reduce(233), // !=, reduce: MlrvalOrFunction
			reduce(233), // <=>, reduce: MlrvalOrFunction
			reduce(233), // >=, reduce: MlrvalOrFunction
			reduce(233), // <, reduce: MlrvalOrFunction
			reduce(233), // <=, reduce: MlrvalOrFunction
			reduce(233), // ^, reduce: MlrvalOrFunction
			reduce(233), // &, reduce: MlrvalOrFunction
			reduce(233), // <<, reduce: MlrvalOrFunction
			reduce(233), // >>>, reduce: MlrvalOrFunction
			reduce(233), // +, reduce: MlrvalOrFunction
			reduce(233), // -, reduce: MlrvalOrFunction
			reduce(233), // .+, reduce: MlrvalOrFunction
			reduce(233), // .-, reduce: MlrvalOrFunction
			reduce(233), // *, reduce: MlrvalOrFunction
			reduce(233), // /, reduce: MlrvalOrFunction
			reduce(233), // //, reduce: MlrvalOrFunction
			reduce(233), // %, reduce: MlrvalOrFunction
			reduce(233), // .*, reduce: MlrvalOrFunction
			reduce(233), // ./, reduce: MlrvalOrFunction
			reduce(233), // .//, reduce: MlrvalOrFunction
			reduce(233), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(233), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(233), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(233), // [[, reduce: MlrvalOrFunction
			reduce(233), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3463
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(210), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(210), // >, reduce: MlrvalOrFunction
			reduce(210), // >>, reduce: MlrvalOrFunction
			reduce(210), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(210), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(210), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(210), // ||, reduce: MlrvalOrFunction
			reduce(210), // ^^, reduce: MlrvalOrFunction
			reduce(210), // &&, reduce: MlrvalOrFunction
			reduce(210), // ??, reduce: MlrvalOrFunction
			reduce(210), // ???, reduce: MlrvalOrFunction
			reduce(210), // =~, reduce: MlrvalOrFunction
			reduce(210), // !=~, reduce: MlrvalOrFunction
			reduce(210), // ==, reduce: MlrvalOrFunction
			reduce(210), // !=, reduce: MlrvalOrFunction
			reduce(210), // <=>, reduce: MlrvalOrFunction
			reduce(210), // >=, reduce: MlrvalOrFunction
			reduce(210), // <, reduce: MlrvalOrFunction
			reduce(210), // <=, reduce: MlrvalOrFunction
			reduce(210), // ^, reduce: MlrvalOrFunction
			reduce(210), // &, reduce: MlrvalOrFunction
			reduce(210), // <<, reduce: MlrvalOrFunction
			reduce(210), // >>>, reduce: MlrvalOrFunction
			reduce(210), // +, reduce: MlrvalOrFunction
			reduce(210), // -, reduce: MlrvalOrFunction
			reduce(210), // .+, reduce: MlrvalOrFunction
			reduce(210), // .-, reduce: MlrvalOrFunction
			reduce(210), // *, reduce: MlrvalOrFunction
			reduce(210), // /, reduce: MlrvalOrFunction
			reduce(210), // //, reduce: MlrvalOrFunction
			reduce(210), // %, reduce: MlrvalOrFunction
			reduce(210), // .*, reduce: MlrvalOrFunction
			reduce(210), // ./, reduce: MlrvalOrFunction
			reduce(210), // .//, reduce: MlrvalOrFunction
			reduce(210), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(210), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(210), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(210), // [[, reduce: MlrvalOrFunction
			reduce(210), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3464
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(102), // ;, reduce: FieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(102), // >, reduce: FieldValue
			reduce(102), // >>, reduce: FieldValue
			reduce(102), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(102), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(102), // ?, reduce: FieldValue
			nil,         // :
			reduce(102), // ||, reduce: FieldValue
			reduce(102), // ^^, reduce: FieldValue
			reduce(102), // &&, reduce: FieldValue
			reduce(102), // ??, reduce: FieldValue
			reduce(102), // ???, reduce: FieldValue
			reduce(102), // =~, reduce: FieldValue
			reduce(102), // !=~, reduce: FieldValue
			reduce(102), // ==, reduce: FieldValue
			reduce(102), // !=, reduce: FieldValue
			reduce(102), // <=>, reduce: FieldValue
			reduce(102), // >=, reduce: FieldValue
			reduce(102), // <, reduce: FieldValue
			reduce(102), // <=, reduce: FieldValue
			reduce(102), // ^, reduce: FieldValue
			reduce(102), // &, reduce: FieldValue
			reduce(102), // <<, reduce: FieldValue
			reduce(102), // >>>, reduce: FieldValue
			reduce(102), // +, reduce: FieldValue
			reduce(102), // -, reduce: FieldValue
			reduce(102), // .+, reduce: FieldValue
			reduce(102), // .-, reduce: FieldValue
			reduce(102), // *, reduce: FieldValue
			reduce(102), // /, reduce: FieldValue
			reduce(102), // //, reduce: FieldValue
			reduce(102), // %, reduce: FieldValue
			reduce(102), // .*, reduce: FieldValue
			reduce(102), // ./, reduce: FieldValue
			reduce(102), // .//, reduce: FieldValue
			reduce(102), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(102), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(102), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(102), // [[, reduce: FieldValue
			reduce(102), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3465
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(103), // ;, reduce: FieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(103), // >, reduce: FieldValue
			reduce(103), // >>, reduce: FieldValue
			reduce(103), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(103), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(103), // ?, reduce: FieldValue
			nil,         // :
			reduce(103), // ||, reduce: FieldValue
			reduce(103), // ^^, reduce: FieldValue
			reduce(103), // &&, reduce: FieldValue
			reduce(103), // ??, reduce: FieldValue
			reduce(103), // ???, reduce: FieldValue
			reduce(103), // =~, reduce: FieldValue
			reduce(103), // !=~, reduce: FieldValue
			reduce(103), // ==, reduce: FieldValue
			reduce(103), // !=, reduce: FieldValue
			reduce(103), // <=>, reduce: FieldValue
			reduce(103), // >=, reduce: FieldValue
			reduce(103), // <, reduce: FieldValue
			reduce(103), // <=, reduce: FieldValue
			reduce(103), // ^, reduce: FieldValue
			reduce(103), // &, reduce: FieldValue
			reduce(103), // <<, reduce: FieldValue
			reduce(103), // >>>, reduce: FieldValue
			reduce(103), // +, reduce: FieldValue
			reduce(103), // -, reduce: FieldValue
			reduce(103), // .+, reduce: FieldValue
			reduce(103), // .-, reduce: FieldValue
			reduce(103), // *, reduce: FieldValue
			reduce(103), // /, reduce: FieldValue
			reduce(103), // //, reduce: FieldValue
			reduce(103), // %, reduce: FieldValue
			reduce(103), // .*, reduce: FieldValue
			reduce(103), // ./, reduce: FieldValue
			reduce(103), // .//, reduce: FieldValue
			reduce(103), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(103), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(103), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(103), // [[, reduce: FieldValue
			reduce(103), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3466
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(104), // ;, reduce: DirectFieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(104), // >, reduce: DirectFieldValue
			reduce(104), // >>, reduce: DirectFieldValue
			reduce(104), // |, reduce: DirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(104), // ,, reduce: DirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(104), // ?, reduce: DirectFieldValue
			nil,         // :
			reduce(104), // ||, reduce: DirectFieldValue
			reduce(104), // ^^, reduce: DirectFieldValue
			reduce(104), // &&, reduce: DirectFieldValue
			reduce(104), // ??, reduce: DirectFieldValue
			reduce(104), // ???, reduce: DirectFieldValue
			reduce(104), // =~, reduce: DirectFieldValue
			reduce(104), // !=~, reduce: DirectFieldValue
			reduce(104), // ==, reduce: DirectFieldValue
			reduce(104), // !=, reduce: DirectFieldValue
			reduce(104), // <=>, reduce: DirectFieldValue
			reduce(104), // >=, reduce: DirectFieldValue
			reduce(104), // <, reduce: DirectFieldValue
			reduce(104), // <=, reduce: DirectFieldValue
			reduce(104), // ^, reduce: DirectFieldValue
			reduce(104), // &, reduce: DirectFieldValue
			reduce(104), // <<, reduce: DirectFieldValue
			reduce(104), // >>>, reduce: DirectFieldValue
			reduce(104), // +, reduce: DirectFieldValue
			reduce(104), // -, reduce: DirectFieldValue
			reduce(104), // .+, reduce: DirectFieldValue
			reduce(104), // .-, reduce: DirectFieldValue
			reduce(104), // *, reduce: DirectFieldValue
			reduce(104), // /, reduce: DirectFieldValue
			reduce(104), // //, reduce: DirectFieldValue
			reduce(104), // %, reduce: DirectFieldValue
			reduce(104), // .*, reduce: DirectFieldValue
			reduce(104), // ./, reduce: DirectFieldValue
			reduce(104), // .//, reduce: DirectFieldValue
			reduce(104), // ., reduce: DirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(104), // **, reduce: DirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(104), // [, reduce: DirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(104), // [[, reduce: DirectFieldValue
			reduce(104), // [[[, reduce: DirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3467
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3468
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(106), // ;, reduce: BracedFieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(106), // >, reduce: BracedFieldValue
			reduce(106), // >>, reduce: BracedFieldValue
			reduce(106), // |, reduce: BracedFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(106), // ,, reduce: BracedFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(106), // ?, reduce: BracedFieldValue
			nil,         // :
			reduce(106), // ||, reduce: BracedFieldValue
			reduce(106), // ^^, reduce: BracedFieldValue
			reduce(106), // &&, reduce: BracedFieldValue
			reduce(106), // ??, reduce: BracedFieldValue
			reduce(106), // ???, reduce: BracedFieldValue
			reduce(106), // =~, reduce: BracedFieldValue
			reduce(106), // !=~, reduce: BracedFieldValue
			reduce(106), // ==, reduce: BracedFieldValue
			reduce(106), // !=, reduce: BracedFieldValue
			reduce(106), // <=>, reduce: BracedFieldValue
			reduce(106), // >=, reduce: BracedFieldValue
			reduce(106), // <, reduce: BracedFieldValue
			reduce(106), // <=, reduce: BracedFieldValue
			reduce(106), // ^, reduce: BracedFieldValue
			reduce(106), // &, reduce: BracedFieldValue
			reduce(106), // <<, reduce: BracedFieldValue
			reduce(106), // >>>, reduce: BracedFieldValue
			reduce(106), // +, reduce: BracedFieldValue
			reduce(106), // -, reduce: BracedFieldValue
			reduce(106), // .+, reduce: BracedFieldValue
			reduce(106), // .-, reduce: BracedFieldValue
			reduce(106), // *, reduce: BracedFieldValue
			reduce(106), // /, reduce: BracedFieldValue
			reduce(106), // //, reduce: BracedFieldValue
			reduce(106), // %, reduce: BracedFieldValue
			reduce(106), // .*, reduce: BracedFieldValue
			reduce(106), // ./, reduce: BracedFieldValue
			reduce(106), // .//, reduce: BracedFieldValue
			reduce(106), // ., reduce: BracedFieldValue
			nil,         // !
			nil,         // ~
			reduce(106), // **, reduce: BracedFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(106), // [, reduce: BracedFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(106), // [[, reduce: BracedFieldValue
			reduce(106), // [[[, reduce: BracedFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3469
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3470
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3471
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(109), // ;, reduce: FullSrec
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(109), // >, reduce: FullSrec
			reduce(109), // >>, reduce: FullSrec
			reduce(109), // |, reduce: FullSrec
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(109), // ,, reduce: FullSrec
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(109), // ?, reduce: FullSrec
			nil,         // :
			reduce(109), // ||, reduce: FullSrec
			reduce(109), // ^^, reduce: FullSrec
			reduce(109), // &&, reduce: FullSrec
			reduce(109), // ??, reduce: FullSrec
			reduce(109), // ???, reduce: FullSrec
			reduce(109), // =~, reduce: FullSrec
			reduce(109), // !=~, reduce: FullSrec
			reduce(109), // ==, reduce: FullSrec
			reduce(109), // !=, reduce: FullSrec
			reduce(109), // <=>, reduce: FullSrec
			reduce(109), // >=, reduce: FullSrec
			reduce(109), // <, reduce: FullSrec
			reduce(109), // <=, reduce: FullSrec
			reduce(109), // ^, reduce: FullSrec
			reduce(109), // &, reduce: FullSrec
			reduce(109), // <<, reduce: FullSrec
			reduce(109), // >>>, reduce: FullSrec
			reduce(109), // +, reduce: FullSrec
			reduce(109), // -, reduce: FullSrec
			reduce(109), // .+, reduce: FullSrec
			reduce(109), // .-, reduce: FullSrec
			reduce(109), // *, reduce: FullSrec
			reduce(109), // /, reduce: FullSrec
			reduce(109), // //, reduce: FullSrec
			reduce(109), // %, reduce: FullSrec
			reduce(109), // .*, reduce: FullSrec
			reduce(109), // ./, reduce: FullSrec
			reduce(109), // .//, reduce: FullSrec
			reduce(109), // ., reduce: FullSrec
			nil,         // !
			nil,         // ~
			reduce(109), // **, reduce: FullSrec
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(109), // [, reduce: FullSrec
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(109), // [[, reduce: FullSrec
			reduce(109), // [[[, reduce: FullSrec
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3472
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(212), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(212), // >, reduce: MlrvalOrFunction
			reduce(212), // >>, reduce: MlrvalOrFunction
			reduce(212), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(212), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(212), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(212), // ||, reduce: MlrvalOrFunction
			reduce(212), // ^^, reduce: MlrvalOrFunction
			reduce(212), // &&, reduce: MlrvalOrFunction
			reduce(212), // ??, reduce: MlrvalOrFunction
			reduce(212), // ???, reduce: MlrvalOrFunction
			reduce(212), // =~, reduce: MlrvalOrFunction
			reduce(212), // !=~, reduce: MlrvalOrFunction
			reduce(212), // ==, reduce: MlrvalOrFunction
			reduce(212), // !=, reduce: MlrvalOrFunction
			reduce(212), // <=>, reduce: MlrvalOrFunction
			reduce(212), // >=, reduce: MlrvalOrFunction
			reduce(212), // <, reduce: MlrvalOrFunction
			reduce(212), // <=, reduce: MlrvalOrFunction
			reduce(212), // ^, reduce: MlrvalOrFunction
			reduce(212), // &, reduce: MlrvalOrFunction
			reduce(212), // <<, reduce: MlrvalOrFunction
			reduce(212), // >>>, reduce: MlrvalOrFunction
			reduce(212), // +, reduce: MlrvalOrFunction
			reduce(212), // -, reduce: MlrvalOrFunction
			reduce(212), // .+, reduce: MlrvalOrFunction
			reduce(212), // .-, reduce: MlrvalOrFunction
			reduce(212), // *, reduce: MlrvalOrFunction
			reduce(212), // /, reduce: MlrvalOrFunction
			reduce(212), // //, reduce: MlrvalOrFunction
			reduce(212), // %, reduce: MlrvalOrFunction
			reduce(212), // .*, reduce: MlrvalOrFunction
			reduce(212), // ./, reduce: MlrvalOrFunction
			reduce(212), // .//, reduce: MlrvalOrFunction
			reduce(212), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(212), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(212), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(212), // [[, reduce: MlrvalOrFunction
			reduce(212), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3473
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(113), // ;, reduce: DirectOosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(113), // >, reduce: DirectOosvarValue
			reduce(113), // >>, reduce: DirectOosvarValue
			reduce(113), // |, reduce: DirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(113), // ,, reduce: DirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(113), // ?, reduce: DirectOosvarValue
			nil,         // :
			reduce(113), // ||, reduce: DirectOosvarValue
			reduce(113), // ^^, reduce: DirectOosvarValue
			reduce(113), // &&, reduce: DirectOosvarValue
			reduce(113), // ??, reduce: DirectOosvarValue
			reduce(113), // ???, reduce: DirectOosvarValue
			reduce(113), // =~, reduce: DirectOosvarValue
			reduce(113), // !=~, reduce: DirectOosvarValue
			reduce(113), // ==, reduce: DirectOosvarValue
			reduce(113), // !=, reduce: DirectOosvarValue
			reduce(113), // <=>, reduce: DirectOosvarValue
			reduce(113), // >=, reduce: DirectOosvarValue
			reduce(113), // <, reduce: DirectOosvarValue
			reduce(113), // <=, reduce: DirectOosvarValue
			reduce(113), // ^, reduce: DirectOosvarValue
			reduce(113), // &, reduce: DirectOosvarValue
			reduce(113), // <<, reduce: DirectOosvarValue
			reduce(113), // >>>, reduce: DirectOosvarValue
			reduce(113), // +, reduce: DirectOosvarValue
			reduce(113), // -, reduce: DirectOosvarValue
			reduce(113), // .+, reduce: DirectOosvarValue
			reduce(113), // .-, reduce: DirectOosvarValue
			reduce(113), // *, reduce: DirectOosvarValue
			reduce(113), // /, reduce: DirectOosvarValue
			reduce(113), // //, reduce: DirectOosvarValue
			reduce(113), // %, reduce: DirectOosvarValue
			reduce(113), // .*, reduce: DirectOosvarValue
			reduce(113), // ./, reduce: DirectOosvarValue
			reduce(113), // .//, reduce: DirectOosvarValue
			reduce(113), // ., reduce: DirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(113), // **, reduce: DirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(113), // [, reduce: DirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(113), // [[, reduce: DirectOosvarValue
			reduce(113), // [[[, reduce: DirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3474
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3475
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(115), // ;, reduce: BracedOosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(115), // >, reduce: BracedOosvarValue
			reduce(115), // >>, reduce: BracedOosvarValue
			reduce(115), // |, reduce: BracedOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(115), // ,, reduce: BracedOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(115), // ?, reduce: BracedOosvarValue
			nil,         // :
			reduce(115), // ||, reduce: BracedOosvarValue
			reduce(115), // ^^, reduce: BracedOosvarValue
			reduce(115), // &&, reduce: BracedOosvarValue
			reduce(115), // ??, reduce: BracedOosvarValue
			reduce(115), // ???, reduce: BracedOosvarValue
			reduce(115), // =~, reduce: BracedOosvarValue
			reduce(115), // !=~, reduce: BracedOosvarValue
			reduce(115), // ==, reduce: BracedOosvarValue
			reduce(115), // !=, reduce: BracedOosvarValue
			reduce(115), // <=>, reduce: BracedOosvarValue
			reduce(115), // >=, reduce: BracedOosvarValue
			reduce(115), // <, reduce: BracedOosvarValue
			reduce(115), // <=, reduce: BracedOosvarValue
			reduce(115), // ^, reduce: BracedOosvarValue
			reduce(115), // &, reduce: BracedOosvarValue
			reduce(115), // <<, reduce: BracedOosvarValue
			reduce(115), // >>>, reduce: BracedOosvarValue
			reduce(115), // +, reduce: BracedOosvarValue
			reduce(115), // -, reduce: BracedOosvarValue
			reduce(115), // .+, reduce: BracedOosvarValue
			reduce(115), // .-, reduce: BracedOosvarValue
			reduce(115), // *, reduce: BracedOosvarValue
			reduce(115), // /, reduce: BracedOosvarValue
			reduce(115), // //, reduce: BracedOosvarValue
			reduce(115), // %, reduce: BracedOosvarValue
			reduce(115), // .*, reduce: BracedOosvarValue
			reduce(115), // ./, reduce: BracedOosvarValue
			reduce(115), // .//, reduce: BracedOosvarValue
			reduce(115), // ., reduce: BracedOosvarValue
			nil,         // !
			nil,         // ~
			reduce(115), // **, reduce: BracedOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(115), // [, reduce: BracedOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(115), // [[, reduce: BracedOosvarValue
			reduce(115), // [[[, reduce: BracedOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3476
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(116), // ;, reduce: FullOosvar
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(116), // >, reduce: FullOosvar
			reduce(116), // >>, reduce: FullOosvar
			reduce(116), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(116), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(116), // ?, reduce: FullOosvar
			nil,         // :
			reduce(116), // ||, reduce: FullOosvar
			reduce(116), // ^^, reduce: FullOosvar
			reduce(116), // &&, reduce: FullOosvar
			reduce(116), // ??, reduce: FullOosvar
			reduce(116), // ???, reduce: FullOosvar
			reduce(116), // =~, reduce: FullOosvar
			reduce(116), // !=~, reduce: FullOosvar
			reduce(116), // ==, reduce: FullOosvar
			reduce(116), // !=, reduce: FullOosvar
			reduce(116), // <=>, reduce: FullOosvar
			reduce(116), // >=, reduce: FullOosvar
			reduce(116), // <, reduce: FullOosvar
			reduce(116), // <=, reduce: FullOosvar
			reduce(116), // ^, reduce: FullOosvar
			reduce(116), // &, reduce: FullOosvar
			reduce(116), // <<, reduce: FullOosvar
			reduce(116), // >>>, reduce: FullOosvar
			reduce(116), // +, reduce: FullOosvar
			reduce(116), // -, reduce: FullOosvar
			reduce(116), // .+, reduce: FullOosvar
			reduce(116), // .-, reduce: FullOosvar
			reduce(116), // *, reduce: FullOosvar
			reduce(116), // /, reduce: FullOosvar
			reduce(116), // //, reduce: FullOosvar
			reduce(116), // %, reduce: FullOosvar
			reduce(116), // .*, reduce: FullOosvar
			reduce(116), // ./, reduce: FullOosvar
			reduce(116), // .//, reduce: FullOosvar
			reduce(116), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(116), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(116), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(116), // [[, reduce: FullOosvar
			reduce(116), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3477
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(117), // ;, reduce: FullOosvar
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(117), // >, reduce: FullOosvar
			reduce(117), // >>, reduce: FullOosvar
			reduce(117), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(117), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(117), // ?, reduce: FullOosvar
			nil,         // :
			reduce(117), // ||, reduce: FullOosvar
			reduce(117), // ^^, reduce: FullOosvar
			reduce(117), // &&, reduce: FullOosvar
			reduce(117), // ??, reduce: FullOosvar
			reduce(117), // ???, reduce: FullOosvar
			reduce(117), // =~, reduce: FullOosvar
			reduce(117), // !=~, reduce: FullOosvar
			reduce(117), // ==, reduce: FullOosvar
			reduce(117), // !=, reduce: FullOosvar
			reduce(117), // <=>, reduce: FullOosvar
			reduce(117), // >=, reduce: FullOosvar
			reduce(117), // <, reduce: FullOosvar
			reduce(117), // <=, reduce: FullOosvar
			reduce(117), // ^, reduce: FullOosvar
			reduce(117), // &, reduce: FullOosvar
			reduce(117), // <<, reduce: FullOosvar
			reduce(117), // >>>, reduce: FullOosvar
			reduce(117), // +, reduce: FullOosvar
			reduce(117), // -, reduce: FullOosvar
			reduce(117), // .+, reduce: FullOosvar
			reduce(117), // .-, reduce: FullOosvar
			reduce(117), // *, reduce: FullOosvar
			reduce(117), // /, reduce: FullOosvar
			reduce(117), // //, reduce: FullOosvar
			reduce(117), // %, reduce: FullOosvar
			reduce(117), // .*, reduce: FullOosvar
			reduce(117), // ./, reduce: FullOosvar
			reduce(117), // .//, reduce: FullOosvar
			reduce(117), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(117), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(117), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(117), // [[, reduce: FullOosvar
			reduce(117), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3478
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(118), // ;, reduce: LocalVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(118), // >, reduce: LocalVariable
			reduce(118), // >>, reduce: LocalVariable
			reduce(118), // |, reduce: LocalVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(118), // ,, reduce: LocalVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(270), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(118), // ?, reduce: LocalVariable
			nil,         // :
			reduce(118), // ||, reduce: LocalVariable
			reduce(118), // ^^, reduce: LocalVariable
			reduce(118), // &&, reduce: LocalVariable
			reduce(118), // ??, reduce: LocalVariable
			reduce(118), // ???, reduce: LocalVariable
			reduce(118), // =~, reduce: LocalVariable
			reduce(118), // !=~, reduce: LocalVariable
			reduce(118), // ==, reduce: LocalVariable
			reduce(118), // !=, reduce: LocalVariable
			reduce(118), // <=>, reduce: LocalVariable
			reduce(118), // >=, reduce: LocalVariable
			reduce(118), // <, reduce: LocalVariable
			reduce(118), // <=, reduce: LocalVariable
			reduce(118), // ^, reduce: LocalVariable
			reduce(118), // &, reduce: LocalVariable
			reduce(118), // <<, reduce: LocalVariable
			reduce(118), // >>>, reduce: LocalVariable
			reduce(118), // +, reduce: LocalVariable
			reduce(118), // -, reduce: LocalVariable
			reduce(118), // .+, reduce: LocalVariable
			reduce(118), // .-, reduce: LocalVariable
			reduce(118), // *, reduce: LocalVariable
			reduce(118), // /, reduce: LocalVariable
			reduce(118), // //, reduce: LocalVariable
			reduce(118), // %, reduce: LocalVariable
			reduce(118), // .*, reduce: LocalVariable
			reduce(118), // ./, reduce: LocalVariable
			reduce(118), // .//, reduce: LocalVariable
			reduce(118), // ., reduce: LocalVariable
			nil,         // !
			nil,         // ~
			reduce(118), // **, reduce: LocalVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(118), // [, reduce: LocalVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(118), // [[, reduce: LocalVariable
			reduce(118), // [[[, reduce: LocalVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3479
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(147), // ;, reduce: Rvalue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(147), // ,, reduce: Rvalue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3480
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(148), // ;, reduce: PrecedenceChainStart
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(148), // ,, reduce: PrecedenceChainStart
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3481
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(150), // ;, reduce: TernaryTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(150), // ,, reduce: TernaryTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			shift(4044), // ?
			nil,         // :
			shift(4045), // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3482
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(152), // ;, reduce: LogicalOrTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(152), // ,, reduce: LogicalOrTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(152), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(152), // ||, reduce: LogicalOrTerm
			shift(4046), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3483
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(154), // ;, reduce: LogicalXORTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(154), // ,, reduce: LogicalXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(154), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(154), // ||, reduce: LogicalXORTerm
			reduce(154), // ^^, reduce: LogicalXORTerm
			shift(4047), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3484
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(156), // ;, reduce: LogicalAndTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(156), // ,, reduce: LogicalAndTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(156), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(156), // ||, reduce: LogicalAndTerm
			reduce(156), // ^^, reduce: LogicalAndTerm
			reduce(156), // &&, reduce: LogicalAndTerm
			shift(4048), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3485
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(158), // ;, reduce: AbsentCoalesceTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(158), // ,, reduce: AbsentCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(158), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(158), // ||, reduce: AbsentCoalesceTerm
			reduce(158), // ^^, reduce: AbsentCoalesceTerm
			reduce(158), // &&, reduce: AbsentCoalesceTerm
			reduce(158), // ??, reduce: AbsentCoalesceTerm
			shift(4049), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3486
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(160), // ;, reduce: EmptyCoalesceTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(160), // ,, reduce: EmptyCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(160), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(160), // ||, reduce: EmptyCoalesceTerm
			reduce(160), // ^^, reduce: EmptyCoalesceTerm
			reduce(160), // &&, reduce: EmptyCoalesceTerm
			reduce(160), // ??, reduce: EmptyCoalesceTerm
			reduce(160), // ???, reduce: EmptyCoalesceTerm
			shift(4050), // =~
			shift(4051), // !=~
			shift(4052), // ==
			shift(4053), // !=
			shift(4054), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3487
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(166), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(4055), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(166), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(166), // ?, reduce: EqneTerm
			nil,         // :
			reduce(166), // ||, reduce: EqneTerm
			reduce(166), // ^^, reduce: EqneTerm
			reduce(166), // &&, reduce: EqneTerm
			reduce(166), // ??, reduce: EqneTerm
			reduce(166), // ???, reduce: EqneTerm
			reduce(166), // =~, reduce: EqneTerm
			reduce(166), // !=~, reduce: EqneTerm
			reduce(166), // ==, reduce: EqneTerm
			reduce(166), // !=, reduce: EqneTerm
			reduce(166), // <=>, reduce: EqneTerm
			shift(4056), // >=
			shift(4057), // <
			shift(4058), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3488
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(171), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(171), // >, reduce: CmpTerm
			nil,         // >>
			shift(4059), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(171), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(171), // ?, reduce: CmpTerm
			nil,         // :
			reduce(171), // ||, reduce: CmpTerm
			reduce(171), // ^^, reduce: CmpTerm
			reduce(171), // &&, reduce: CmpTerm
			reduce(171), // ??, reduce: CmpTerm
			reduce(171), // ???, reduce: CmpTerm
			reduce(171), // =~, reduce: CmpTerm
			reduce(171), // !=~, reduce: CmpTerm
			reduce(171), // ==, reduce: CmpTerm
			reduce(171), // !=, reduce: CmpTerm
			reduce(171), // <=>, reduce: CmpTerm
			reduce(171), // >=, reduce: CmpTerm
			reduce(171), // <, reduce: CmpTerm
			reduce(171), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3489
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(173), // ;, reduce: BitwiseORTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(173), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(173), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(173), // ,, reduce: BitwiseORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(173), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(173), // ||, reduce: BitwiseORTerm
			reduce(173), // ^^, reduce: BitwiseORTerm
			reduce(173), // &&, reduce: BitwiseORTerm
			reduce(173), // ??, reduce: BitwiseORTerm
			reduce(173), // ???, reduce: BitwiseORTerm
			reduce(173), // =~, reduce: BitwiseORTerm
			reduce(173), // !=~, reduce: BitwiseORTerm
			reduce(173), // ==, reduce: BitwiseORTerm
			reduce(173), // !=, reduce: BitwiseORTerm
			reduce(173), // <=>, reduce: BitwiseORTerm
			reduce(173), // >=, reduce: BitwiseORTerm
			reduce(173), // <, reduce: BitwiseORTerm
			reduce(173), // <=, reduce: BitwiseORTerm
			shift(4060), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3490
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(175), // ;, reduce: BitwiseXORTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(175), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(175), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(175), // ,, reduce: BitwiseXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(175), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(175), // ||, reduce: BitwiseXORTerm
			reduce(175), // ^^, reduce: BitwiseXORTerm
			reduce(175), // &&, reduce: BitwiseXORTerm
			reduce(175), // ??, reduce: BitwiseXORTerm
			reduce(175), // ???, reduce: BitwiseXORTerm
			reduce(175), // =~, reduce: BitwiseXORTerm
			reduce(175), // !=~, reduce: BitwiseXORTerm
			reduce(175), // ==, reduce: BitwiseXORTerm
			reduce(175), // !=, reduce: BitwiseXORTerm
			reduce(175), // <=>, reduce: BitwiseXORTerm
			reduce(175), // >=, reduce: BitwiseXORTerm
			reduce(175), // <, reduce: BitwiseXORTerm
			reduce(175), // <=, reduce: BitwiseXORTerm
			reduce(175), // ^, reduce: BitwiseXORTerm
			shift(4061), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3491
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(177), // ;, reduce: BitwiseANDTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(177), // >, reduce: BitwiseANDTerm
			shift(4062), // >>
			reduce(177), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(177), // ,, reduce: BitwiseANDTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(177), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(177), // ||, reduce: BitwiseANDTerm
			reduce(177), // ^^, reduce: BitwiseANDTerm
			reduce(177), // &&, reduce: BitwiseANDTerm
			reduce(177), // ??, reduce: BitwiseANDTerm
			reduce(177), // ???, reduce: BitwiseANDTerm
			reduce(177), // =~, reduce: BitwiseANDTerm
			reduce(177), // !=~, reduce: BitwiseANDTerm
			reduce(177), // ==, reduce: BitwiseANDTerm
			reduce(177), // !=, reduce: BitwiseANDTerm
			reduce(177), // <=>, reduce: BitwiseANDTerm
			reduce(177), // >=, reduce: BitwiseANDTerm
			reduce(177), // <, reduce: BitwiseANDTerm
			reduce(177), // <=, reduce: BitwiseANDTerm
			reduce(177), // ^, reduce: BitwiseANDTerm
			reduce(177), // &, reduce: BitwiseANDTerm
			shift(4063), // <<
			shift(4064), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3492
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(181), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(181), // >, reduce: BitwiseShiftTerm
			reduce(181), // >>, reduce: BitwiseShiftTerm
			reduce(181), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(181), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(181), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(181), // ||, reduce: BitwiseShiftTerm
			reduce(181), // ^^, reduce: BitwiseShiftTerm
			reduce(181), // &&, reduce: BitwiseShiftTerm
			reduce(181), // ??, reduce: BitwiseShiftTerm
			reduce(181), // ???, reduce: BitwiseShiftTerm
			reduce(181), // =~, reduce: BitwiseShiftTerm
			reduce(181), // !=~, reduce: BitwiseShiftTerm
			reduce(181), // ==, reduce: BitwiseShiftTerm
			reduce(181), // !=, reduce: BitwiseShiftTerm
			reduce(181), // <=>, reduce: BitwiseShiftTerm
			reduce(181), // >=, reduce: BitwiseShiftTerm
			reduce(181), // <, reduce: BitwiseShiftTerm
			reduce(181), // <=, reduce: BitwiseShiftTerm
			reduce(181), // ^, reduce: BitwiseShiftTerm
			reduce(181), // &, reduce: BitwiseShiftTerm
			reduce(181), // <<, reduce: BitwiseShiftTerm
			reduce(181), // >>>, reduce: BitwiseShiftTerm
			shift(4065), // +
			shift(4066), // -
			shift(4067), // .+
			shift(4068), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3493
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3494
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(186), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(186), // >, reduce: AddsubdotTerm
			reduce(186), // >>, reduce: AddsubdotTerm
			reduce(186), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(186), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(186), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(186), // ||, reduce: AddsubdotTerm
			reduce(186), // ^^, reduce: AddsubdotTerm
			reduce(186), // &&, reduce: AddsubdotTerm
			reduce(186), // ??, reduce: AddsubdotTerm
			reduce(186), // ???, reduce: AddsubdotTerm
			reduce(186), // =~, reduce: AddsubdotTerm
			reduce(186), // !=~, reduce: AddsubdotTerm
			reduce(186), // ==, reduce: AddsubdotTerm
			reduce(186), // !=, reduce: AddsubdotTerm
			reduce(186), // <=>, reduce: AddsubdotTerm
			reduce(186), // >=, reduce: AddsubdotTerm
			reduce(186), // <, reduce: AddsubdotTerm
			reduce(186), // <=, reduce: AddsubdotTerm
			reduce(186), // ^, reduce: AddsubdotTerm
			reduce(186), // &, reduce: AddsubdotTerm
			reduce(186), // <<, reduce: AddsubdotTerm
			reduce(186), // >>>, reduce: AddsubdotTerm
			reduce(186), // +, reduce: AddsubdotTerm
			reduce(186), // -, reduce: AddsubdotTerm
			reduce(186), // .+, reduce: AddsubdotTerm
			reduce(186), // .-, reduce: AddsubdotTerm
			shift(4070), // *
			shift(4071), // /
			shift(4072), // //
			shift(4073), // %
			shift(4074), // .*
			shift(4075), // ./
			shift(4076), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3495
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3496
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3497
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3498
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(194), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(194), // >, reduce: MuldivTerm
			reduce(194), // >>, reduce: MuldivTerm
			reduce(194), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(194), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(194), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(194), // ||, reduce: MuldivTerm
			reduce(194), // ^^, reduce: MuldivTerm
			reduce(194), // &&, reduce: MuldivTerm
			reduce(194), // ??, reduce: MuldivTerm
			reduce(194), // ???, reduce: MuldivTerm
			reduce(194), // =~, reduce: MuldivTerm
			reduce(194), // !=~, reduce: MuldivTerm
			reduce(194), // ==, reduce: MuldivTerm
			reduce(194), // !=, reduce: MuldivTerm
			reduce(194), // <=>, reduce: MuldivTerm
			reduce(194), // >=, reduce: MuldivTerm
			reduce(194), // <, reduce: MuldivTerm
			reduce(194), // <=, reduce: MuldivTerm
			reduce(194), // ^, reduce: MuldivTerm
			reduce(194), // &, reduce: MuldivTerm
			reduce(194), // <<, reduce: MuldivTerm
			reduce(194), // >>>, reduce: MuldivTerm
			reduce(194), // +, reduce: MuldivTerm
			reduce(194), // -, reduce: MuldivTerm
			reduce(194), // .+, reduce: MuldivTerm
			reduce(194), // .-, reduce: MuldivTerm
			reduce(194), // *, reduce: MuldivTerm
			reduce(194), // /, reduce: MuldivTerm
			reduce(194), // //, reduce: MuldivTerm
			reduce(194), // %, reduce: MuldivTerm
			reduce(194), // .*, reduce: MuldivTerm
			reduce(194), // ./, reduce: MuldivTerm
			reduce(194), // .//, reduce: MuldivTerm
			shift(4080), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3499
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(196), // ;, reduce: DotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(196), // >, reduce: DotTerm
			reduce(196), // >>, reduce: DotTerm
			reduce(196), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(196), // ,, reduce: DotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(196), // ?, reduce: DotTerm
			nil,         // :
			reduce(196), // ||, reduce: DotTerm
			reduce(196), // ^^, reduce: DotTerm
			reduce(196), // &&, reduce: DotTerm
			reduce(196), // ??, reduce: DotTerm
			reduce(196), // ???, reduce: DotTerm
			reduce(196), // =~, reduce: DotTerm
			reduce(196), // !=~, reduce: DotTerm
			reduce(196), // ==, reduce: DotTerm
			reduce(196), // !=, reduce: DotTerm
			reduce(196), // <=>, reduce: DotTerm
			reduce(196), // >=, reduce: DotTerm
			reduce(196), // <, reduce: DotTerm
			reduce(196), // <=, reduce: DotTerm
			reduce(196), // ^, reduce: DotTerm
			reduce(196), // &, reduce: DotTerm
			reduce(196), // <<, reduce: DotTerm
			reduce(196), // >>>, reduce: DotTerm
			reduce(196), // +, reduce: DotTerm
			reduce(196), // -, reduce: DotTerm
			reduce(196), // .+, reduce: DotTerm
			reduce(196), // .-, reduce: DotTerm
			reduce(196), // *, reduce: DotTerm
			reduce(196), // /, reduce: DotTerm
			reduce(196), // //, reduce: DotTerm
			reduce(196), // %, reduce: DotTerm
			reduce(196), // .*, reduce: DotTerm
			reduce(196), // ./, reduce: DotTerm
			reduce(196), // .//, reduce: DotTerm
			reduce(196), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3500
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3501
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3502
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(203), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(203), // >, reduce: UnaryOpTerm
			reduce(203), // >>, reduce: UnaryOpTerm
			reduce(203), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(203), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(203), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(203), // ||, reduce: UnaryOpTerm
			reduce(203), // ^^, reduce: UnaryOpTerm
			reduce(203), // &&, reduce: UnaryOpTerm
			reduce(203), // ??, reduce: UnaryOpTerm
			reduce(203), // ???, reduce: UnaryOpTerm
			reduce(203), // =~, reduce: UnaryOpTerm
			reduce(203), // !=~, reduce: UnaryOpTerm
			reduce(203), // ==, reduce: UnaryOpTerm
			reduce(203), // !=, reduce: UnaryOpTerm
			reduce(203), // <=>, reduce: UnaryOpTerm
			reduce(203), // >=, reduce: UnaryOpTerm
			reduce(203), // <, reduce: UnaryOpTerm
			reduce(203), // <=, reduce: UnaryOpTerm
			reduce(203), // ^, reduce: UnaryOpTerm
			reduce(203), // &, reduce: UnaryOpTerm
			reduce(203), // <<, reduce: UnaryOpTerm
			reduce(203), // >>>, reduce: UnaryOpTerm
			reduce(203), // +, reduce: UnaryOpTerm
			reduce(203), // -, reduce: UnaryOpTerm
			reduce(203), // .+, reduce: UnaryOpTerm
			reduce(203), // .-, reduce: UnaryOpTerm
			reduce(203), // *, reduce: UnaryOpTerm
			reduce(203), // /, reduce: UnaryOpTerm
			reduce(203), // //, reduce: UnaryOpTerm
			reduce(203), // %, reduce: UnaryOpTerm
			reduce(203), // .*, reduce: UnaryOpTerm
			reduce(203), // ./, reduce: UnaryOpTerm
			reduce(203), // .//, reduce: UnaryOpTerm
			reduce(203), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3503
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(207), // ;, reduce: PowTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(207), // >, reduce: PowTerm
			reduce(207), // >>, reduce: PowTerm
			reduce(207), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(207), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(207), // ?, reduce: PowTerm
			nil,         // :
			reduce(207), // ||, reduce: PowTerm
			reduce(207), // ^^, reduce: PowTerm
			reduce(207), // &&, reduce: PowTerm
			reduce(207), // ??, reduce: PowTerm
			reduce(207), // ???, reduce: PowTerm
			reduce(207), // =~, reduce: PowTerm
			reduce(207), // !=~, reduce: PowTerm
			reduce(207), // ==, reduce: PowTerm
			reduce(207), // !=, reduce: PowTerm
			reduce(207), // <=>, reduce: PowTerm
			reduce(207), // >=, reduce: PowTerm
			reduce(207), // <, reduce: PowTerm
			reduce(207), // <=, reduce: PowTerm
			reduce(207), // ^, reduce: PowTerm
			reduce(207), // &, reduce: PowTerm
			reduce(207), // <<, reduce: PowTerm
			reduce(207), // >>>, reduce: PowTerm
			reduce(207), // +, reduce: PowTerm
			reduce(207), // -, reduce: PowTerm
			reduce(207), // .+, reduce: PowTerm
			reduce(207), // .-, reduce: PowTerm
			reduce(207), // *, reduce: PowTerm
			reduce(207), // /, reduce: PowTerm
			reduce(207), // //, reduce: PowTerm
			reduce(207), // %, reduce: PowTerm
			reduce(207), // .*, reduce: PowTerm
			reduce(207), // ./, reduce: PowTerm
			reduce(207), // .//, reduce: PowTerm
			reduce(207), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			shift(4083), // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3504
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(209), // ;, reduce: PrecedenceChainEnd
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(209), // >, reduce: PrecedenceChainEnd
			reduce(209), // >>, reduce: PrecedenceChainEnd
			reduce(209), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(209), // ,, reduce: PrecedenceChainEnd
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(209), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(209), // ||, reduce: PrecedenceChainEnd
			reduce(209), // ^^, reduce: PrecedenceChainEnd
			reduce(209), // &&, reduce: PrecedenceChainEnd
			reduce(209), // ??, reduce: PrecedenceChainEnd
			reduce(209), // ???, reduce: PrecedenceChainEnd
			reduce(209), // =~, reduce: PrecedenceChainEnd
			reduce(209), // !=~, reduce: PrecedenceChainEnd
			reduce(209), // ==, reduce: PrecedenceChainEnd
			reduce(209), // !=, reduce: PrecedenceChainEnd
			reduce(209), // <=>, reduce: PrecedenceChainEnd
			reduce(209), // >=, reduce: PrecedenceChainEnd
			reduce(209), // <, reduce: PrecedenceChainEnd
			reduce(209), // <=, reduce: PrecedenceChainEnd
			reduce(209), // ^, reduce: PrecedenceChainEnd
			reduce(209), // &, reduce: PrecedenceChainEnd
			reduce(209), // <<, reduce: PrecedenceChainEnd
			reduce(209), // >>>, reduce: PrecedenceChainEnd
			reduce(209), // +, reduce: PrecedenceChainEnd
			reduce(209), // -, reduce: PrecedenceChainEnd
			reduce(209), // .+, reduce: PrecedenceChainEnd
			reduce(209), // .-, reduce: PrecedenceChainEnd
			reduce(209), // *, reduce: PrecedenceChainEnd
			reduce(209), // /, reduce: PrecedenceChainEnd
			reduce(209), // //, reduce: PrecedenceChainEnd
			reduce(209), // %, reduce: PrecedenceChainEnd
			reduce(209), // .*, reduce: PrecedenceChainEnd
			reduce(209), // ./, reduce: PrecedenceChainEnd
			reduce(209), // .//, reduce: PrecedenceChainEnd
			reduce(209), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(209), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(4084), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			shift(4085), // [[
			shift(4086), // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3505
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(215), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(215), // >, reduce: MlrvalOrFunction
			reduce(215), // >>, reduce: MlrvalOrFunction
			reduce(215), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(215), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(215), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(215), // ||, reduce: MlrvalOrFunction
			reduce(215), // ^^, reduce: MlrvalOrFunction
			reduce(215), // &&, reduce: MlrvalOrFunction
			reduce(215), // ??, reduce: MlrvalOrFunction
			reduce(215), // ???, reduce: MlrvalOrFunction
			reduce(215), // =~, reduce: MlrvalOrFunction
			reduce(215), // !=~, reduce: MlrvalOrFunction
			reduce(215), // ==, reduce: MlrvalOrFunction
			reduce(215), // !=, reduce: MlrvalOrFunction
			reduce(215), // <=>, reduce: MlrvalOrFunction
			reduce(215), // >=, reduce: MlrvalOrFunction
			reduce(215), // <, reduce: MlrvalOrFunction
			reduce(215), // <=, reduce: MlrvalOrFunction
			reduce(215), // ^, reduce: MlrvalOrFunction
			reduce(215), // &, reduce: MlrvalOrFunction
			reduce(215), // <<, reduce: MlrvalOrFunction
			reduce(215), // >>>, reduce: MlrvalOrFunction
			reduce(215), // +, reduce: MlrvalOrFunction
			reduce(215), // -, reduce: MlrvalOrFunction
			reduce(215), // .+, reduce: MlrvalOrFunction
			reduce(215), // .-, reduce: MlrvalOrFunction
			reduce(215), // *, reduce: MlrvalOrFunction
			reduce(215), // /, reduce: MlrvalOrFunction
			reduce(215), // //, reduce: MlrvalOrFunction
			reduce(215), // %, reduce: MlrvalOrFunction
			reduce(215), // .*, reduce: MlrvalOrFunction
			reduce(215), // ./, reduce: MlrvalOrFunction
			reduce(215), // .//, reduce: MlrvalOrFunction
			reduce(215), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(215), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(215), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(215), // [[, reduce: MlrvalOrFunction
			reduce(215), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3506
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(216), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(216), // >, reduce: MlrvalOrFunction
			reduce(216), // >>, reduce: MlrvalOrFunction
			reduce(216), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(216), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(216), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(216), // ||, reduce: MlrvalOrFunction
			reduce(216), // ^^, reduce: MlrvalOrFunction
			reduce(216), // &&, reduce: MlrvalOrFunction
			reduce(216), // ??, reduce: MlrvalOrFunction
			reduce(216), // ???, reduce: MlrvalOrFunction
			reduce(216), // =~, reduce: MlrvalOrFunction
			reduce(216), // !=~, reduce: MlrvalOrFunction
			reduce(216), // ==, reduce: MlrvalOrFunction
			reduce(216), // !=, reduce: MlrvalOrFunction
			reduce(216), // <=>, reduce: MlrvalOrFunction
			reduce(216), // >=, reduce: MlrvalOrFunction
			reduce(216), // <, reduce: MlrvalOrFunction
			reduce(216), // <=, reduce: MlrvalOrFunction
			reduce(216), // ^, reduce: MlrvalOrFunction
			reduce(216), // &, reduce: MlrvalOrFunction
			reduce(216), // <<, reduce: MlrvalOrFunction
			reduce(216), // >>>, reduce: MlrvalOrFunction
			reduce(216), // +, reduce: MlrvalOrFunction
			reduce(216), // -, reduce: MlrvalOrFunction
			reduce(216), // .+, reduce: MlrvalOrFunction
			reduce(216), // .-, reduce: MlrvalOrFunction
			reduce(216), // *, reduce: MlrvalOrFunction
			reduce(216), // /, reduce: MlrvalOrFunction
			reduce(216), // //, reduce: MlrvalOrFunction
			reduce(216), // %, reduce: MlrvalOrFunction
			reduce(216), // .*, reduce: MlrvalOrFunction
			reduce(216), // ./, reduce: MlrvalOrFunction
			reduce(216), // .//, reduce: MlrvalOrFunction
			reduce(216), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(216), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(216), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(216), // [[, reduce: MlrvalOrFunction
			reduce(216), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3507
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(217), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(217), // >, reduce: MlrvalOrFunction
			reduce(217), // >>, reduce: MlrvalOrFunction
			reduce(217), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(217), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(217), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(217), // ||, reduce: MlrvalOrFunction
			reduce(217), // ^^, reduce: MlrvalOrFunction
			reduce(217), // &&, reduce: MlrvalOrFunction
			reduce(217), // ??, reduce: MlrvalOrFunction
			reduce(217), // ???, reduce: MlrvalOrFunction
			reduce(217), // =~, reduce: MlrvalOrFunction
			reduce(217), // !=~, reduce: MlrvalOrFunction
			reduce(217), // ==, reduce: MlrvalOrFunction
			reduce(217), // !=, reduce: MlrvalOrFunction
			reduce(217), // <=>, reduce: MlrvalOrFunction
			reduce(217), // >=, reduce: MlrvalOrFunction
			reduce(217), // <, reduce: MlrvalOrFunction
			reduce(217), // <=, reduce: MlrvalOrFunction
			reduce(217), // ^, reduce: MlrvalOrFunction
			reduce(217), // &, reduce: MlrvalOrFunction
			reduce(217), // <<, reduce: MlrvalOrFunction
			reduce(217), // >>>, reduce: MlrvalOrFunction
			reduce(217), // +, reduce: MlrvalOrFunction
			reduce(217), // -, reduce: MlrvalOrFunction
			reduce(217), // .+, reduce: MlrvalOrFunction
			reduce(217), // .-, reduce: MlrvalOrFunction
			reduce(217), // *, reduce: MlrvalOrFunction
			reduce(217), // /, reduce: MlrvalOrFunction
			reduce(217), // //, reduce: MlrvalOrFunction
			reduce(217), // %, reduce: MlrvalOrFunction
			reduce(217), // .*, reduce: MlrvalOrFunction
			reduce(217), // ./, reduce: MlrvalOrFunction
			reduce(217), // .//, reduce: MlrvalOrFunction
			reduce(217), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(217), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(217), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(217), // [[, reduce: MlrvalOrFunction
			reduce(217), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3508
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(218), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(218), // >, reduce: MlrvalOrFunction
			reduce(218), // >>, reduce: MlrvalOrFunction
			reduce(218), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(218), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(218), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(218), // ||, reduce: MlrvalOrFunction
			reduce(218), // ^^, reduce: MlrvalOrFunction
			reduce(218), // &&, reduce: MlrvalOrFunction
			reduce(218), // ??, reduce: MlrvalOrFunction
			reduce(218), // ???, reduce: MlrvalOrFunction
			reduce(218), // =~, reduce: MlrvalOrFunction
			reduce(218), // !=~, reduce: MlrvalOrFunction
			reduce(218), // ==, reduce: MlrvalOrFunction
			reduce(218), // !=, reduce: MlrvalOrFunction
			reduce(218), // <=>, reduce: MlrvalOrFunction
			reduce(218), // >=, reduce: MlrvalOrFunction
			reduce(218), // <, reduce: MlrvalOrFunction
			reduce(218), // <=, reduce: MlrvalOrFunction
			reduce(218), // ^, reduce: MlrvalOrFunction
			reduce(218), // &, reduce: MlrvalOrFunction
			reduce(218), // <<, reduce: MlrvalOrFunction
			reduce(218), // >>>, reduce: MlrvalOrFunction
			reduce(218), // +, reduce: MlrvalOrFunction
			reduce(218), // -, reduce: MlrvalOrFunction
			reduce(218), // .+, reduce: MlrvalOrFunction
			reduce(218), // .-, reduce: MlrvalOrFunction
			reduce(218), // *, reduce: MlrvalOrFunction
			reduce(218), // /, reduce: MlrvalOrFunction
			reduce(218), // //, reduce: MlrvalOrFunction
			reduce(218), // %, reduce: MlrvalOrFunction
			reduce(218), // .*, reduce: MlrvalOrFunction
			reduce(218), // ./, reduce: MlrvalOrFunction
			reduce(218), // .//, reduce: MlrvalOrFunction
			reduce(218), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(218), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(218), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(218), // [[, reduce: MlrvalOrFunction
			reduce(218), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3509
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(219), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(219), // >, reduce: MlrvalOrFunction
			reduce(219), // >>, reduce: MlrvalOrFunction
			reduce(219), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(219), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(219), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(219), // ||, reduce: MlrvalOrFunction
			reduce(219), // ^^, reduce: MlrvalOrFunction
			reduce(219), // &&, reduce: MlrvalOrFunction
			reduce(219), // ??, reduce: MlrvalOrFunction
			reduce(219), // ???, reduce: MlrvalOrFunction
			reduce(219), // =~, reduce: MlrvalOrFunction
			reduce(219), // !=~, reduce: MlrvalOrFunction
			reduce(219), // ==, reduce: MlrvalOrFunction
			reduce(219), // !=, reduce: MlrvalOrFunction
			reduce(219), // <=>, reduce: MlrvalOrFunction
			reduce(219), // >=, reduce: MlrvalOrFunction
			reduce(219), // <, reduce: MlrvalOrFunction
			reduce(219), // <=, reduce: MlrvalOrFunction
			reduce(219), // ^, reduce: MlrvalOrFunction
			reduce(219), // &, reduce: MlrvalOrFunction
			reduce(219), // <<, reduce: MlrvalOrFunction
			reduce(219), // >>>, reduce: MlrvalOrFunction
			reduce(219), // +, reduce: MlrvalOrFunction
			reduce(219), // -, reduce: MlrvalOrFunction
			reduce(219), // .+, reduce: MlrvalOrFunction
			reduce(219), // .-, reduce: MlrvalOrFunction
			reduce(219), // *, reduce: MlrvalOrFunction
			reduce(219), // /, reduce: MlrvalOrFunction
			reduce(219), // //, reduce: MlrvalOrFunction
			reduce(219), // %, reduce: MlrvalOrFunction
			reduce(219), // .*, reduce: MlrvalOrFunction
			reduce(219), // ./, reduce: MlrvalOrFunction
			reduce(219), // .//, reduce: MlrvalOrFunction
			reduce(219), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(219), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(219), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(219), // [[, reduce: MlrvalOrFunction
			reduce(219), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3510
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(220), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(220), // >, reduce: MlrvalOrFunction
			reduce(220), // >>, reduce: MlrvalOrFunction
			reduce(220), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(220), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(220), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(220), // ||, reduce: MlrvalOrFunction
			reduce(220), // ^^, reduce: MlrvalOrFunction
			reduce(220), // &&, reduce: MlrvalOrFunction
			reduce(220), // ??, reduce: MlrvalOrFunction
			reduce(220), // ???, reduce: MlrvalOrFunction
			reduce(220), // =~, reduce: MlrvalOrFunction
			reduce(220), // !=~, reduce: MlrvalOrFunction
			reduce(220), // ==, reduce: MlrvalOrFunction
			reduce(220), // !=, reduce: MlrvalOrFunction
			reduce(220), // <=>, reduce: MlrvalOrFunction
			reduce(220), // >=, reduce: MlrvalOrFunction
			reduce(220), // <, reduce: MlrvalOrFunction
			reduce(220), // <=, reduce: MlrvalOrFunction
			reduce(220), // ^, reduce: MlrvalOrFunction
			reduce(220), // &, reduce: MlrvalOrFunction
			reduce(220), // <<, reduce: MlrvalOrFunction
			reduce(220), // >>>, reduce: MlrvalOrFunction
			reduce(220), // +, reduce: MlrvalOrFunction
			reduce(220), // -, reduce: MlrvalOrFunction
			reduce(220), // .+, reduce: MlrvalOrFunction
			reduce(220), // .-, reduce: MlrvalOrFunction
			reduce(220), // *, reduce: MlrvalOrFunction
			reduce(220), // /, reduce: MlrvalOrFunction
			reduce(220), // //, reduce: MlrvalOrFunction
			reduce(220), // %, reduce: MlrvalOrFunction
			reduce(220), // .*, reduce: MlrvalOrFunction
			reduce(220), // ./, reduce: MlrvalOrFunction
			reduce(220), // .//, reduce: MlrvalOrFunction
			reduce(220), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(220), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(220), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(220), // [[, reduce: MlrvalOrFunction
			reduce(220), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3511
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(221), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(221), // >, reduce: MlrvalOrFunction
			reduce(221), // >>, reduce: MlrvalOrFunction
			reduce(221), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(221), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(221), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(221), // ||, reduce: MlrvalOrFunction
			reduce(221), // ^^, reduce: MlrvalOrFunction
			reduce(221), // &&, reduce: MlrvalOrFunction
			reduce(221), // ??, reduce: MlrvalOrFunction
			reduce(221), // ???, reduce: MlrvalOrFunction
			reduce(221), // =~, reduce: MlrvalOrFunction
			reduce(221), // !=~, reduce: MlrvalOrFunction
			reduce(221), // ==, reduce: MlrvalOrFunction
			reduce(221), // !=, reduce: MlrvalOrFunction
			reduce(221), // <=>, reduce: MlrvalOrFunction
			reduce(221), // >=, reduce: MlrvalOrFunction
			reduce(221), // <, reduce: MlrvalOrFunction
			reduce(221), // <=, reduce: MlrvalOrFunction
			reduce(221), // ^, reduce: MlrvalOrFunction
			reduce(221), // &, reduce: MlrvalOrFunction
			reduce(221), // <<, reduce: MlrvalOrFunction
			reduce(221), // >>>, reduce: MlrvalOrFunction
			reduce(221), // +, reduce: MlrvalOrFunction
			reduce(221), // -, reduce: MlrvalOrFunction
			reduce(221), // .+, reduce: MlrvalOrFunction
			reduce(221), // .-, reduce: MlrvalOrFunction
			reduce(221), // *, reduce: MlrvalOrFunction
			reduce(221), // /, reduce: MlrvalOrFunction
			reduce(221), // //, reduce: MlrvalOrFunction
			reduce(221), // %, reduce: MlrvalOrFunction
			reduce(221), // .*, reduce: MlrvalOrFunction
			reduce(221), // ./, reduce: MlrvalOrFunction
			reduce(221), // .//, reduce: MlrvalOrFunction
			reduce(221), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(221), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(221), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(221), // [[, reduce: MlrvalOrFunction
			reduce(221), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3512
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(222), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(222), // >, reduce: MlrvalOrFunction
			reduce(222), // >>, reduce: MlrvalOrFunction
			reduce(222), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(222), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(222), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(222), // ||, reduce: MlrvalOrFunction
			reduce(222), // ^^, reduce: MlrvalOrFunction
			reduce(222), // &&, reduce: MlrvalOrFunction
			reduce(222), // ??, reduce: MlrvalOrFunction
			reduce(222), // ???, reduce: MlrvalOrFunction
			reduce(222), // =~, reduce: MlrvalOrFunction
			reduce(222), // !=~, reduce: MlrvalOrFunction
			reduce(222), // ==, reduce: MlrvalOrFunction
			reduce(222), // !=, reduce: MlrvalOrFunction
			reduce(222), // <=>, reduce: MlrvalOrFunction
			reduce(222), // >=, reduce: MlrvalOrFunction
			reduce(222), // <, reduce: MlrvalOrFunction
			reduce(222), // <=, reduce: MlrvalOrFunction
			reduce(222), // ^, reduce: MlrvalOrFunction
			reduce(222), // &, reduce: MlrvalOrFunction
			reduce(222), // <<, reduce: MlrvalOrFunction
			reduce(222), // >>>, reduce: MlrvalOrFunction
			reduce(222), // +, reduce: MlrvalOrFunction
			reduce(222), // -, reduce: MlrvalOrFunction
			reduce(222), // .+, reduce: MlrvalOrFunction
			reduce(222), // .-, reduce: MlrvalOrFunction
			reduce(222), // *, reduce: MlrvalOrFunction
			reduce(222), // /, reduce: MlrvalOrFunction
			reduce(222), // //, reduce: MlrvalOrFunction
			reduce(222), // %, reduce: MlrvalOrFunction
			reduce(222), // .*, reduce: MlrvalOrFunction
			reduce(222), // ./, reduce: MlrvalOrFunction
			reduce(222), // .//, reduce: MlrvalOrFunction
			reduce(222), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(222), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(222), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(222), // [[, reduce: MlrvalOrFunction
			reduce(222), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3513
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(223), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(223), // >, reduce: MlrvalOrFunction
			reduce(223), // >>, reduce: MlrvalOrFunction
			reduce(223), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(223), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(223), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(223), // ||, reduce: MlrvalOrFunction
			reduce(223), // ^^, reduce: MlrvalOrFunction
			reduce(223), // &&, reduce: MlrvalOrFunction
			reduce(223), // ??, reduce: MlrvalOrFunction
			reduce(223), // ???, reduce: MlrvalOrFunction
			reduce(223), // =~, reduce: MlrvalOrFunction
			reduce(223), // !=~, reduce: MlrvalOrFunction
			reduce(223), // ==, reduce: MlrvalOrFunction
			reduce(223), // !=, reduce: MlrvalOrFunction
			reduce(223), // <=>, reduce: MlrvalOrFunction
			reduce(223), // >=, reduce: MlrvalOrFunction
			reduce(223), // <, reduce: MlrvalOrFunction
			reduce(223), // <=, reduce: MlrvalOrFunction
			reduce(223), // ^, reduce: MlrvalOrFunction
			reduce(223), // &, reduce: MlrvalOrFunction
			reduce(223), // <<, reduce: MlrvalOrFunction
			reduce(223), // >>>, reduce: MlrvalOrFunction
			reduce(223), // +, reduce: MlrvalOrFunction
			reduce(223), // -, reduce: MlrvalOrFunction
			reduce(223), // .+, reduce: MlrvalOrFunction
			reduce(223), // .-, reduce: MlrvalOrFunction
			reduce(223), // *, reduce: MlrvalOrFunction
			reduce(223), // /, reduce: MlrvalOrFunction
			reduce(223), // //, reduce: MlrvalOrFunction
			reduce(223), // %, reduce: MlrvalOrFunction
			reduce(223), // .*, reduce: MlrvalOrFunction
			reduce(223), // ./, reduce: MlrvalOrFunction
			reduce(223), // .//, reduce: MlrvalOrFunction
			reduce(223), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(223), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(223), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(223), // [[, reduce: MlrvalOrFunction
			reduce(223), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3514
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(224), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(224), // >, reduce: MlrvalOrFunction
			reduce(224), // >>, reduce: MlrvalOrFunction
			reduce(224), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(224), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(224), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(224), // ||, reduce: MlrvalOrFunction
			reduce(224), // ^^, reduce: MlrvalOrFunction
			reduce(224), // &&, reduce: MlrvalOrFunction
			reduce(224), // ??, reduce: MlrvalOrFunction
			reduce(224), // ???, reduce: MlrvalOrFunction
			reduce(224), // =~, reduce: MlrvalOrFunction
			reduce(224), // !=~, reduce: MlrvalOrFunction
			reduce(224), // ==, reduce: MlrvalOrFunction
			reduce(224), // !=, reduce: MlrvalOrFunction
			reduce(224), // <=>, reduce: MlrvalOrFunction
			reduce(224), // >=, reduce: MlrvalOrFunction
			reduce(224), // <, reduce: MlrvalOrFunction
			reduce(224), // <=, reduce: MlrvalOrFunction
			reduce(224), // ^, reduce: MlrvalOrFunction
			reduce(224), // &, reduce: MlrvalOrFunction
			reduce(224), // <<, reduce: MlrvalOrFunction
			reduce(224), // >>>, reduce: MlrvalOrFunction
			reduce(224), // +, reduce: MlrvalOrFunction
			reduce(224), // -, reduce: MlrvalOrFunction
			reduce(224), // .+, reduce: MlrvalOrFunction
			reduce(224), // .-, reduce: MlrvalOrFunction
			reduce(224), // *, reduce: MlrvalOrFunction
			reduce(224), // /, reduce: MlrvalOrFunction
			reduce(224), // //, reduce: MlrvalOrFunction
			reduce(224), // %, reduce: MlrvalOrFunction
			reduce(224), // .*, reduce: MlrvalOrFunction
			reduce(224), // ./, reduce: MlrvalOrFunction
			reduce(224), // .//, reduce: MlrvalOrFunction
			reduce(224), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(224), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(224), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(224), // [[, reduce: MlrvalOrFunction
			reduce(224), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3515
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(225), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(225), // >, reduce: MlrvalOrFunction
			reduce(225), // >>, reduce: MlrvalOrFunction
			reduce(225), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(225), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(225), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(225), // ||, reduce: MlrvalOrFunction
			reduce(225), // ^^, reduce: MlrvalOrFunction
			reduce(225), // &&, reduce: MlrvalOrFunction
			reduce(225), // ??, reduce: MlrvalOrFunction
			reduce(225), // ???, reduce: MlrvalOrFunction
			reduce(225), // =~, reduce: MlrvalOrFunction
			reduce(225), // !=~, reduce: MlrvalOrFunction
			reduce(225), // ==, reduce: MlrvalOrFunction
			reduce(225), // !=, reduce: MlrvalOrFunction
			reduce(225), // <=>, reduce: MlrvalOrFunction
			reduce(225), // >=, reduce: MlrvalOrFunction
			reduce(225), // <, reduce: MlrvalOrFunction
			reduce(225), // <=, reduce: MlrvalOrFunction
			reduce(225), // ^, reduce: MlrvalOrFunction
			reduce(225), // &, reduce: MlrvalOrFunction
			reduce(225), // <<, reduce: MlrvalOrFunction
			reduce(225), // >>>, reduce: MlrvalOrFunction
			reduce(225), // +, reduce: MlrvalOrFunction
			reduce(225), // -, reduce: MlrvalOrFunction
			reduce(225), // .+, reduce: MlrvalOrFunction
			reduce(225), // .-, reduce: MlrvalOrFunction
			reduce(225), // *, reduce: MlrvalOrFunction
			reduce(225), // /, reduce: MlrvalOrFunction
			reduce(225), // //, reduce: MlrvalOrFunction
			reduce(225), // %, reduce: MlrvalOrFunction
			reduce(225), // .*, reduce: MlrvalOrFunction
			reduce(225), // ./, reduce: MlrvalOrFunction
			reduce(225), // .//, reduce: MlrvalOrFunction
			reduce(225), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(225), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(225), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(225), // [[, reduce: MlrvalOrFunction
			reduce(225), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3516
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(226), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(226), // >, reduce: MlrvalOrFunction
			reduce(226), // >>, reduce: MlrvalOrFunction
			reduce(226), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(226), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(226), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(226), // ||, reduce: MlrvalOrFunction
			reduce(226), // ^^, reduce: MlrvalOrFunction
			reduce(226), // &&, reduce: MlrvalOrFunction
			reduce(226), // ??, reduce: MlrvalOrFunction
			reduce(226), // ???, reduce: MlrvalOrFunction
			reduce(226), // =~, reduce: MlrvalOrFunction
			reduce(226), // !=~, reduce: MlrvalOrFunction
			reduce(226), // ==, reduce: MlrvalOrFunction
			reduce(226), // !=, reduce: MlrvalOrFunction
			reduce(226), // <=>, reduce: MlrvalOrFunction
			reduce(226), // >=, reduce: MlrvalOrFunction
			reduce(226), // <, reduce: MlrvalOrFunction
			reduce(226), // <=, reduce: MlrvalOrFunction
			reduce(226), // ^, reduce: MlrvalOrFunction
			reduce(226), // &, reduce: MlrvalOrFunction
			reduce(226), // <<, reduce: MlrvalOrFunction
			reduce(226), // >>>, reduce: MlrvalOrFunction
			reduce(226), // +, reduce: MlrvalOrFunction
			reduce(226), // -, reduce: MlrvalOrFunction
			reduce(226), // .+, reduce: MlrvalOrFunction
			reduce(226), // .-, reduce: MlrvalOrFunction
			reduce(226), // *, reduce: MlrvalOrFunction
			reduce(226), // /, reduce: MlrvalOrFunction
			reduce(226), // //, reduce: MlrvalOrFunction
			reduce(226), // %, reduce: MlrvalOrFunction
			reduce(226), // .*, reduce: MlrvalOrFunction
			reduce(226), // ./, reduce: MlrvalOrFunction
			reduce(226), // .//, reduce: MlrvalOrFunction
			reduce(226), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(226), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(226), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(226), // [[, reduce: MlrvalOrFunction
			reduce(226), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3517
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(227), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(227), // >, reduce: MlrvalOrFunction
			reduce(227), // >>, reduce: MlrvalOrFunction
			reduce(227), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(227), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(227), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(227), // ||, reduce: MlrvalOrFunction
			reduce(227), // ^^, reduce: MlrvalOrFunction
			reduce(227), // &&, reduce: MlrvalOrFunction
			reduce(227), // ??, reduce: MlrvalOrFunction
			reduce(227), // ???, reduce: MlrvalOrFunction
			reduce(227), // =~, reduce: MlrvalOrFunction
			reduce(227), // !=~, reduce: MlrvalOrFunction
			reduce(227), // ==, reduce: MlrvalOrFunction
			reduce(227), // !=, reduce: MlrvalOrFunction
			reduce(227), // <=>, reduce: MlrvalOrFunction
			reduce(227), // >=, reduce: MlrvalOrFunction
			reduce(227), // <, reduce: MlrvalOrFunction
			reduce(227), // <=, reduce: MlrvalOrFunction
			reduce(227), // ^, reduce: MlrvalOrFunction
			reduce(227), // &, reduce: MlrvalOrFunction
			reduce(227), // <<, reduce: MlrvalOrFunction
			reduce(227), // >>>, reduce: MlrvalOrFunction
			reduce(227), // +, reduce: MlrvalOrFunction
			reduce(227), // -, reduce: MlrvalOrFunction
			reduce(227), // .+, reduce: MlrvalOrFunction
			reduce(227), // .-, reduce: MlrvalOrFunction
			reduce(227), // *, reduce: MlrvalOrFunction
			reduce(227), // /, reduce: MlrvalOrFunction
			reduce(227), // //, reduce: MlrvalOrFunction
			reduce(227), // %, reduce: MlrvalOrFunction
			reduce(227), // .*, reduce: MlrvalOrFunction
			reduce(227), // ./, reduce: MlrvalOrFunction
			reduce(227), // .//, reduce: MlrvalOrFunction
			reduce(227), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(227), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(227), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(227), // [[, reduce: MlrvalOrFunction
			reduce(227), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3518
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(752),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(756),  // (
			nil,         // )
			nil,         // emitp
			shift(768),  // field_name
			shift(769),  // $[
			shift(4087), // ]
			shift(771),  // braced_field_name
			shift(772),  // $[[
			shift(773),  // $[[[
			shift(774),  // full_srec
			shift(776),  // oosvar_name
			shift(777),  // @[
			shift(778),  // braced_oosvar_name
			shift(779),  // full_oosvar
			shift(780),  // all
			shift(781),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(796),  // +
			shift(798),  // -
			shift(799),  // .+
			shift(800),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(803),  // !
			shift(804),  // ~
			nil,         // **
			shift(809),  // string_literal
			shift(810),  // regex_case_insensitive
			shift(811),  // int_literal
			shift(812),  // float_literal
			shift(813),  // boolean_literal
			shift(814),  // null_literal
			shift(815),  // inf_literal
			shift(816),  // nan_literal
			shift(817),  // const_M_PI
			shift(818),  // const_M_E
			shift(819),  // panic
			shift(821),  // [
			shift(824),  // ctx_IPS
			shift(825),  // ctx_IFS
			shift(826),  // ctx_IRS
			shift(827),  // ctx_OPS
			shift(828),  // ctx_OFS
			shift(829),  // ctx_ORS
			shift(830),  // ctx_FLATSEP
			shift(831),  // ctx_NF
			shift(832),  // ctx_NR
			shift(833),  // ctx_FNR
			shift(834),  // ctx_FILENAME
			shift(835),  // ctx_FILENUM
			shift(837),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(844),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3519
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(240), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(240), // >, reduce: MlrvalOrFunction
			reduce(240), // >>, reduce: MlrvalOrFunction
			reduce(240), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(240), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(240), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(240), // ||, reduce: MlrvalOrFunction
			reduce(240), // ^^, reduce: MlrvalOrFunction
			reduce(240), // &&, reduce: MlrvalOrFunction
			reduce(240), // ??, reduce: MlrvalOrFunction
			reduce(240), // ???, reduce: MlrvalOrFunction
			reduce(240), // =~, reduce: MlrvalOrFunction
			reduce(240), // !=~, reduce: MlrvalOrFunction
			reduce(240), // ==, reduce: MlrvalOrFunction
			reduce(240), // !=, reduce: MlrvalOrFunction
			reduce(240), // <=>, reduce: MlrvalOrFunction
			reduce(240), // >=, reduce: MlrvalOrFunction
			reduce(240), // <, reduce: MlrvalOrFunction
			reduce(240), // <=, reduce: MlrvalOrFunction
			reduce(240), // ^, reduce: MlrvalOrFunction
			reduce(240), // &, reduce: MlrvalOrFunction
			reduce(240), // <<, reduce: MlrvalOrFunction
			reduce(240), // >>>, reduce: MlrvalOrFunction
			reduce(240), // +, reduce: MlrvalOrFunction
			reduce(240), // -, reduce: MlrvalOrFunction
			reduce(240), // .+, reduce: MlrvalOrFunction
			reduce(240), // .-, reduce: MlrvalOrFunction
			reduce(240), // *, reduce: MlrvalOrFunction
			reduce(240), // /, reduce: MlrvalOrFunction
			reduce(240), // //, reduce: MlrvalOrFunction
			reduce(240), // %, reduce: MlrvalOrFunction
			reduce(240), // .*, reduce: MlrvalOrFunction
			reduce(240), // ./, reduce: MlrvalOrFunction
			reduce(240), // .//, reduce: MlrvalOrFunction
			reduce(240), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(240), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(240), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(240), // [[, reduce: MlrvalOrFunction
			reduce(240), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3520
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(241), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(241), // >, reduce: ContextVariable
			reduce(241), // >>, reduce: ContextVariable
			reduce(241), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(241), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(241), // ?, reduce: ContextVariable
			nil,         // :
			reduce(241), // ||, reduce: ContextVariable
			reduce(241), // ^^, reduce: ContextVariable
			reduce(241), // &&, reduce: ContextVariable
			reduce(241), // ??, reduce: ContextVariable
			reduce(241), // ???, reduce: ContextVariable
			reduce(241), // =~, reduce: ContextVariable
			reduce(241), // !=~, reduce: ContextVariable
			reduce(241), // ==, reduce: ContextVariable
			reduce(241), // !=, reduce: ContextVariable
			reduce(241), // <=>, reduce: ContextVariable
			reduce(241), // >=, reduce: ContextVariable
			reduce(241), // <, reduce: ContextVariable
			reduce(241), // <=, reduce: ContextVariable
			reduce(241), // ^, reduce: ContextVariable
			reduce(241), // &, reduce: ContextVariable
			reduce(241), // <<, reduce: ContextVariable
			reduce(241), // >>>, reduce: ContextVariable
			reduce(241), // +, reduce: ContextVariable
			reduce(241), // -, reduce: ContextVariable
			reduce(241), // .+, reduce: ContextVariable
			reduce(241), // .-, reduce: ContextVariable
			reduce(241), // *, reduce: ContextVariable
			reduce(241), // /, reduce: ContextVariable
			reduce(241), // //, reduce: ContextVariable
			reduce(241), // %, reduce: ContextVariable
			reduce(241), // .*, reduce: ContextVariable
			reduce(241), // ./, reduce: ContextVariable
			reduce(241), // .//, reduce: ContextVariable
			reduce(241), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(241), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(241), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(241), // [[, reduce: ContextVariable
			reduce(241), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3521
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(242), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(242), // >, reduce: ContextVariable
			reduce(242), // >>, reduce: ContextVariable
			reduce(242), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(242), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(242), // ?, reduce: ContextVariable
			nil,         // :
			reduce(242), // ||, reduce: ContextVariable
			reduce(242), // ^^, reduce: ContextVariable
			reduce(242), // &&, reduce: ContextVariable
			reduce(242), // ??, reduce: ContextVariable
			reduce(242), // ???, reduce: ContextVariable
			reduce(242), // =~, reduce: ContextVariable
			reduce(242), // !=~, reduce: ContextVariable
			reduce(242), // ==, reduce: ContextVariable
			reduce(242), // !=, reduce: ContextVariable
			reduce(242), // <=>, reduce: ContextVariable
			reduce(242), // >=, reduce: ContextVariable
			reduce(242), // <, reduce: ContextVariable
			reduce(242), // <=, reduce: ContextVariable
			reduce(242), // ^, reduce: ContextVariable
			reduce(242), // &, reduce: ContextVariable
			reduce(242), // <<, reduce: ContextVariable
			reduce(242), // >>>, reduce: ContextVariable
			reduce(242), // +, reduce: ContextVariable
			reduce(242), // -, reduce: ContextVariable
			reduce(242), // .+, reduce: ContextVariable
			reduce(242), // .-, reduce: ContextVariable
			reduce(242), // *, reduce: ContextVariable
			reduce(242), // /, reduce: ContextVariable
			reduce(242), // //, reduce: ContextVariable
			reduce(242), // %, reduce: ContextVariable
			reduce(242), // .*, reduce: ContextVariable
			reduce(242), // ./, reduce: ContextVariable
			reduce(242), // .//, reduce: ContextVariable
			reduce(242), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(242), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(242), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(242), // [[, reduce: ContextVariable
			reduce(242), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3522
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(243), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(243), // >, reduce: ContextVariable
			reduce(243), // >>, reduce: ContextVariable
			reduce(243), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(243), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(243), // ?, reduce: ContextVariable
			nil,         // :
			reduce(243), // ||, reduce: ContextVariable
			reduce(243), // ^^, reduce: ContextVariable
			reduce(243), // &&, reduce: ContextVariable
			reduce(243), // ??, reduce: ContextVariable
			reduce(243), // ???, reduce: ContextVariable
			reduce(243), // =~, reduce: ContextVariable
			reduce(243), // !=~, reduce: ContextVariable
			reduce(243), // ==, reduce: ContextVariable
			reduce(243), // !=, reduce: ContextVariable
			reduce(243), // <=>, reduce: ContextVariable
			reduce(243), // >=, reduce: ContextVariable
			reduce(243), // <, reduce: ContextVariable
			reduce(243), // <=, reduce: ContextVariable
			reduce(243), // ^, reduce: ContextVariable
			reduce(243), // &, reduce: ContextVariable
			reduce(243), // <<, reduce: ContextVariable
			reduce(243), // >>>, reduce: ContextVariable
			reduce(243), // +, reduce: ContextVariable
			reduce(243), // -, reduce: ContextVariable
			reduce(243), // .+, reduce: ContextVariable
			reduce(243), // .-, reduce: ContextVariable
			reduce(243), // *, reduce: ContextVariable
			reduce(243), // /, reduce: ContextVariable
			reduce(243), // //, reduce: ContextVariable
			reduce(243), // %, reduce: ContextVariable
			reduce(243), // .*, reduce: ContextVariable
			reduce(243), // ./, reduce: ContextVariable
			reduce(243), // .//, reduce: ContextVariable
			reduce(243), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(243), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(243), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(243), // [[, reduce: ContextVariable
			reduce(243), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3523
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(244), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(244), // >, reduce: ContextVariable
			reduce(244), // >>, reduce: ContextVariable
			reduce(244), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(244), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(244), // ?, reduce: ContextVariable
			nil,         // :
			reduce(244), // ||, reduce: ContextVariable
			reduce(244), // ^^, reduce: ContextVariable
			reduce(244), // &&, reduce: ContextVariable
			reduce(244), // ??, reduce: ContextVariable
			reduce(244), // ???, reduce: ContextVariable
			reduce(244), // =~, reduce: ContextVariable
			reduce(244), // !=~, reduce: ContextVariable
			reduce(244), // ==, reduce: ContextVariable
			reduce(244), // !=, reduce: ContextVariable
			reduce(244), // <=>, reduce: ContextVariable
			reduce(244), // >=, reduce: ContextVariable
			reduce(244), // <, reduce: ContextVariable
			reduce(244), // <=, reduce: ContextVariable
			reduce(244), // ^, reduce: ContextVariable
			reduce(244), // &, reduce: ContextVariable
			reduce(244), // <<, reduce: ContextVariable
			reduce(244), // >>>, reduce: ContextVariable
			reduce(244), // +, reduce: ContextVariable
			reduce(244), // -, reduce: ContextVariable
			reduce(244), // .+, reduce: ContextVariable
			reduce(244), // .-, reduce: ContextVariable
			reduce(244), // *, reduce: ContextVariable
			reduce(244), // /, reduce: ContextVariable
			reduce(244), // //, reduce: ContextVariable
			reduce(244), // %, reduce: ContextVariable
			reduce(244), // .*, reduce: ContextVariable
			reduce(244), // ./, reduce: ContextVariable
			reduce(244), // .//, reduce: ContextVariable
			reduce(244), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(244), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(244), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(244), // [[, reduce: ContextVariable
			reduce(244), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3524
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(245), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(245), // >, reduce: ContextVariable
			reduce(245), // >>, reduce: ContextVariable
			reduce(245), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(245), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(245), // ?, reduce: ContextVariable
			nil,         // :
			reduce(245), // ||, reduce: ContextVariable
			reduce(245), // ^^, reduce: ContextVariable
			reduce(245), // &&, reduce: ContextVariable
			reduce(245), // ??, reduce: ContextVariable
			reduce(245), // ???, reduce: ContextVariable
			reduce(245), // =~, reduce: ContextVariable
			reduce(245), // !=~, reduce: ContextVariable
			reduce(245), // ==, reduce: ContextVariable
			reduce(245), // !=, reduce: ContextVariable
			reduce(245), // <=>, reduce: ContextVariable
			reduce(245), // >=, reduce: ContextVariable
			reduce(245), // <, reduce: ContextVariable
			reduce(245), // <=, reduce: ContextVariable
			reduce(245), // ^, reduce: ContextVariable
			reduce(245), // &, reduce: ContextVariable
			reduce(245), // <<, reduce: ContextVariable
			reduce(245), // >>>, reduce: ContextVariable
			reduce(245), // +, reduce: ContextVariable
			reduce(245), // -, reduce: ContextVariable
			reduce(245), // .+, reduce: ContextVariable
			reduce(245), // .-, reduce: ContextVariable
			reduce(245), // *, reduce: ContextVariable
			reduce(245), // /, reduce: ContextVariable
			reduce(245), // //, reduce: ContextVariable
			reduce(245), // %, reduce: ContextVariable
			reduce(245), // .*, reduce: ContextVariable
			reduce(245), // ./, reduce: ContextVariable
			reduce(245), // .//, reduce: ContextVariable
			reduce(245), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(245), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(245), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(245), // [[, reduce: ContextVariable
			reduce(245), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3525
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(246), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(246), // >, reduce: ContextVariable
			reduce(246), // >>, reduce: ContextVariable
			reduce(246), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(246), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(246), // ?, reduce: ContextVariable
			nil,         // :
			reduce(246), // ||, reduce: ContextVariable
			reduce(246), // ^^, reduce: ContextVariable
			reduce(246), // &&, reduce: ContextVariable
			reduce(246), // ??, reduce: ContextVariable
			reduce(246), // ???, reduce: ContextVariable
			reduce(246), // =~, reduce: ContextVariable
			reduce(246), // !=~, reduce: ContextVariable
			reduce(246), // ==, reduce: ContextVariable
			reduce(246), // !=, reduce: ContextVariable
			reduce(246), // <=>, reduce: ContextVariable
			reduce(246), // >=, reduce: ContextVariable
			reduce(246), // <, reduce: ContextVariable
			reduce(246), // <=, reduce: ContextVariable
			reduce(246), // ^, reduce: ContextVariable
			reduce(246), // &, reduce: ContextVariable
			reduce(246), // <<, reduce: ContextVariable
			reduce(246), // >>>, reduce: ContextVariable
			reduce(246), // +, reduce: ContextVariable
			reduce(246), // -, reduce: ContextVariable
			reduce(246), // .+, reduce: ContextVariable
			reduce(246), // .-, reduce: ContextVariable
			reduce(246), // *, reduce: ContextVariable
			reduce(246), // /, reduce: ContextVariable
			reduce(246), // //, reduce: ContextVariable
			reduce(246), // %, reduce: ContextVariable
			reduce(246), // .*, reduce: ContextVariable
			reduce(246), // ./, reduce: ContextVariable
			reduce(246), // .//, reduce: ContextVariable
			reduce(246), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(246), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(246), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(246), // [[, reduce: ContextVariable
			reduce(246), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3526
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(247), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(247), // >, reduce: ContextVariable
			reduce(247), // >>, reduce: ContextVariable
			reduce(247), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(247), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(247), // ?, reduce: ContextVariable
			nil,         // :
			reduce(247), // ||, reduce: ContextVariable
			reduce(247), // ^^, reduce: ContextVariable
			reduce(247), // &&, reduce: ContextVariable
			reduce(247), // ??, reduce: ContextVariable
			reduce(247), // ???, reduce: ContextVariable
			reduce(247), // =~, reduce: ContextVariable
			reduce(247), // !=~, reduce: ContextVariable
			reduce(247), // ==, reduce: ContextVariable
			reduce(247), // !=, reduce: ContextVariable
			reduce(247), // <=>, reduce: ContextVariable
			reduce(247), // >=, reduce: ContextVariable
			reduce(247), // <, reduce: ContextVariable
			reduce(247), // <=, reduce: ContextVariable
			reduce(247), // ^, reduce: ContextVariable
			reduce(247), // &, reduce: ContextVariable
			reduce(247), // <<, reduce: ContextVariable
			reduce(247), // >>>, reduce: ContextVariable
			reduce(247), // +, reduce: ContextVariable
			reduce(247), // -, reduce: ContextVariable
			reduce(247), // .+, reduce: ContextVariable
			reduce(247), // .-, reduce: ContextVariable
			reduce(247), // *, reduce: ContextVariable
			reduce(247), // /, reduce: ContextVariable
			reduce(247), // //, reduce: ContextVariable
			reduce(247), // %, reduce: ContextVariable
			reduce(247), // .*, reduce: ContextVariable
			reduce(247), // ./, reduce: ContextVariable
			reduce(247), // .//, reduce: ContextVariable
			reduce(247), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(247), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(247), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(247), // [[, reduce: ContextVariable
			reduce(247), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3527
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(248), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(248), // >, reduce: ContextVariable
			reduce(248), // >>, reduce: ContextVariable
			reduce(248), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(248), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(248), // ?, reduce: ContextVariable
			nil,         // :
			reduce(248), // ||, reduce: ContextVariable
			reduce(248), // ^^, reduce: ContextVariable
			reduce(248), // &&, reduce: ContextVariable
			reduce(248), // ??, reduce: ContextVariable
			reduce(248), // ???, reduce: ContextVariable
			reduce(248), // =~, reduce: ContextVariable
			reduce(248), // !=~, reduce: ContextVariable
			reduce(248), // ==, reduce: ContextVariable
			reduce(248), // !=, reduce: ContextVariable
			reduce(248), // <=>, reduce: ContextVariable
			reduce(248), // >=, reduce: ContextVariable
			reduce(248), // <, reduce: ContextVariable
			reduce(248), // <=, reduce: ContextVariable
			reduce(248), // ^, reduce: ContextVariable
			reduce(248), // &, reduce: ContextVariable
			reduce(248), // <<, reduce: ContextVariable
			reduce(248), // >>>, reduce: ContextVariable
			reduce(248), // +, reduce: ContextVariable
			reduce(248), // -, reduce: ContextVariable
			reduce(248), // .+, reduce: ContextVariable
			reduce(248), // .-, reduce: ContextVariable
			reduce(248), // *, reduce: ContextVariable
			reduce(248), // /, reduce: ContextVariable
			reduce(248), // //, reduce: ContextVariable
			reduce(248), // %, reduce: ContextVariable
			reduce(248), // .*, reduce: ContextVariable
			reduce(248), // ./, reduce: ContextVariable
			reduce(248), // .//, reduce: ContextVariable
			reduce(248), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(248), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(248), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(248), // [[, reduce: ContextVariable
			reduce(248), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3528
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(249), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(249), // >, reduce: ContextVariable
			reduce(249), // >>, reduce: ContextVariable
			reduce(249), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(249), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(249), // ?, reduce: ContextVariable
			nil,         // :
			reduce(249), // ||, reduce: ContextVariable
			reduce(249), // ^^, reduce: ContextVariable
			reduce(249), // &&, reduce: ContextVariable
			reduce(249), // ??, reduce: ContextVariable
			reduce(249), // ???, reduce: ContextVariable
			reduce(249), // =~, reduce: ContextVariable
			reduce(249), // !=~, reduce: ContextVariable
			reduce(249), // ==, reduce: ContextVariable
			reduce(249), // !=, reduce: ContextVariable
			reduce(249), // <=>, reduce: ContextVariable
			reduce(249), // >=, reduce: ContextVariable
			reduce(249), // <, reduce: ContextVariable
			reduce(249), // <=, reduce: ContextVariable
			reduce(249), // ^, reduce: ContextVariable
			reduce(249), // &, reduce: ContextVariable
			reduce(249), // <<, reduce: ContextVariable
			reduce(249), // >>>, reduce: ContextVariable
			reduce(249), // +, reduce: ContextVariable
			reduce(249), // -, reduce: ContextVariable
			reduce(249), // .+, reduce: ContextVariable
			reduce(249), // .-, reduce: ContextVariable
			reduce(249), // *, reduce: ContextVariable
			reduce(249), // /, reduce: ContextVariable
			reduce(249), // //, reduce: ContextVariable
			reduce(249), // %, reduce: ContextVariable
			reduce(249), // .*, reduce: ContextVariable
			reduce(249), // ./, reduce: ContextVariable
			reduce(249), // .//, reduce: ContextVariable
			reduce(249), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(249), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(249), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(249), // [[, reduce: ContextVariable
			reduce(249), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3529
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(250), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(250), // >, reduce: ContextVariable
			reduce(250), // >>, reduce: ContextVariable
			reduce(250), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(250), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(250), // ?, reduce: ContextVariable
			nil,         // :
			reduce(250), // ||, reduce: ContextVariable
			reduce(250), // ^^, reduce: ContextVariable
			reduce(250), // &&, reduce: ContextVariable
			reduce(250), // ??, reduce: ContextVariable
			reduce(250), // ???, reduce: ContextVariable
			reduce(250), // =~, reduce: ContextVariable
			reduce(250), // !=~, reduce: ContextVariable
			reduce(250), // ==, reduce: ContextVariable
			reduce(250), // !=, reduce: ContextVariable
			reduce(250), // <=>, reduce: ContextVariable
			reduce(250), // >=, reduce: ContextVariable
			reduce(250), // <, reduce: ContextVariable
			reduce(250), // <=, reduce: ContextVariable
			reduce(250), // ^, reduce: ContextVariable
			reduce(250), // &, reduce: ContextVariable
			reduce(250), // <<, reduce: ContextVariable
			reduce(250), // >>>, reduce: ContextVariable
			reduce(250), // +, reduce: ContextVariable
			reduce(250), // -, reduce: ContextVariable
			reduce(250), // .+, reduce: ContextVariable
			reduce(250), // .-, reduce: ContextVariable
			reduce(250), // *, reduce: ContextVariable
			reduce(250), // /, reduce: ContextVariable
			reduce(250), // //, reduce: ContextVariable
			reduce(250), // %, reduce: ContextVariable
			reduce(250), // .*, reduce: ContextVariable
			reduce(250), // ./, reduce: ContextVariable
			reduce(250), // .//, reduce: ContextVariable
			reduce(250), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(250), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(250), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(250), // [[, reduce: ContextVariable
			reduce(250), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3530
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(251), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(251), // >, reduce: ContextVariable
			reduce(251), // >>, reduce: ContextVariable
			reduce(251), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(251), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(251), // ?, reduce: ContextVariable
			nil,         // :
			reduce(251), // ||, reduce: ContextVariable
			reduce(251), // ^^, reduce: ContextVariable
			reduce(251), // &&, reduce: ContextVariable
			reduce(251), // ??, reduce: ContextVariable
			reduce(251), // ???, reduce: ContextVariable
			reduce(251), // =~, reduce: ContextVariable
			reduce(251), // !=~, reduce: ContextVariable
			reduce(251), // ==, reduce: ContextVariable
			reduce(251), // !=, reduce: ContextVariable
			reduce(251), // <=>, reduce: ContextVariable
			reduce(251), // >=, reduce: ContextVariable
			reduce(251), // <, reduce: ContextVariable
			reduce(251), // <=, reduce: ContextVariable
			reduce(251), // ^, reduce: ContextVariable
			reduce(251), // &, reduce: ContextVariable
			reduce(251), // <<, reduce: ContextVariable
			reduce(251), // >>>, reduce: ContextVariable
			reduce(251), // +, reduce: ContextVariable
			reduce(251), // -, reduce: ContextVariable
			reduce(251), // .+, reduce: ContextVariable
			reduce(251), // .-, reduce: ContextVariable
			reduce(251), // *, reduce: ContextVariable
			reduce(251), // /, reduce: ContextVariable
			reduce(251), // //, reduce: ContextVariable
			reduce(251), // %, reduce: ContextVariable
			reduce(251), // .*, reduce: ContextVariable
			reduce(251), // ./, reduce: ContextVariable
			reduce(251), // .//, reduce: ContextVariable
			reduce(251), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(251), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(251), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(251), // [[, reduce: ContextVariable
			reduce(251), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3531
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(252), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(252), // >, reduce: ContextVariable
			reduce(252), // >>, reduce: ContextVariable
			reduce(252), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(252), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(252), // ?, reduce: ContextVariable
			nil,         // :
			reduce(252), // ||, reduce: ContextVariable
			reduce(252), // ^^, reduce: ContextVariable
			reduce(252), // &&, reduce: ContextVariable
			reduce(252), // ??, reduce: ContextVariable
			reduce(252), // ???, reduce: ContextVariable
			reduce(252), // =~, reduce: ContextVariable
			reduce(252), // !=~, reduce: ContextVariable
			reduce(252), // ==, reduce: ContextVariable
			reduce(252), // !=, reduce: ContextVariable
			reduce(252), // <=>, reduce: ContextVariable
			reduce(252), // >=, reduce: ContextVariable
			reduce(252), // <, reduce: ContextVariable
			reduce(252), // <=, reduce: ContextVariable
			reduce(252), // ^, reduce: ContextVariable
			reduce(252), // &, reduce: ContextVariable
			reduce(252), // <<, reduce: ContextVariable
			reduce(252), // >>>, reduce: ContextVariable
			reduce(252), // +, reduce: ContextVariable
			reduce(252), // -, reduce: ContextVariable
			reduce(252), // .+, reduce: ContextVariable
			reduce(252), // .-, reduce: ContextVariable
			reduce(252), // *, reduce: ContextVariable
			reduce(252), // /, reduce: ContextVariable
			reduce(252), // //, reduce: ContextVariable
			reduce(252), // %, reduce: ContextVariable
			reduce(252), // .*, reduce: ContextVariable
			reduce(252), // ./, reduce: ContextVariable
			reduce(252), // .//, reduce: ContextVariable
			reduce(252), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(252), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(252), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(252), // [[, reduce: ContextVariable
			reduce(252), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3532
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(253), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(253), // >, reduce: MlrvalOrFunction
			reduce(253), // >>, reduce: MlrvalOrFunction
			reduce(253), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(253), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(253), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(253), // ||, reduce: MlrvalOrFunction
			reduce(253), // ^^, reduce: MlrvalOrFunction
			reduce(253), // &&, reduce: MlrvalOrFunction
			reduce(253), // ??, reduce: MlrvalOrFunction
			reduce(253), // ???, reduce: MlrvalOrFunction
			reduce(253), // =~, reduce: MlrvalOrFunction
			reduce(253), // !=~, reduce: MlrvalOrFunction
			reduce(253), // ==, reduce: MlrvalOrFunction
			reduce(253), // !=, reduce: MlrvalOrFunction
			reduce(253), // <=>, reduce: MlrvalOrFunction
			reduce(253), // >=, reduce: MlrvalOrFunction
			reduce(253), // <, reduce: MlrvalOrFunction
			reduce(253), // <=, reduce: MlrvalOrFunction
			reduce(253), // ^, reduce: MlrvalOrFunction
			reduce(253), // &, reduce: MlrvalOrFunction
			reduce(253), // <<, reduce: MlrvalOrFunction
			reduce(253), // >>>, reduce: MlrvalOrFunction
			reduce(253), // +, reduce: MlrvalOrFunction
			reduce(253), // -, reduce: MlrvalOrFunction
			reduce(253), // .+, reduce: MlrvalOrFunction
			reduce(253), // .-, reduce: MlrvalOrFunction
			reduce(253), // *, reduce: MlrvalOrFunction
			reduce(253), // /, reduce: MlrvalOrFunction
			reduce(253), // //, reduce: MlrvalOrFunction
			reduce(253), // %, reduce: MlrvalOrFunction
			reduce(253), // .*, reduce: MlrvalOrFunction
			reduce(253), // ./, reduce: MlrvalOrFunction
			reduce(253), // .//, reduce: MlrvalOrFunction
			reduce(253), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(253), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(253), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(253), // [[, reduce: MlrvalOrFunction
			reduce(253), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3533
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			shift(4089), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(4090), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3534
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(256), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(256), // >, reduce: MlrvalOrFunction
			reduce(256), // >>, reduce: MlrvalOrFunction
			reduce(256), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(256), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(256), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(256), // ||, reduce: MlrvalOrFunction
			reduce(256), // ^^, reduce: MlrvalOrFunction
			reduce(256), // &&, reduce: MlrvalOrFunction
			reduce(256), // ??, reduce: MlrvalOrFunction
			reduce(256), // ???, reduce: MlrvalOrFunction
			reduce(256), // =~, reduce: MlrvalOrFunction
			reduce(256), // !=~, reduce: MlrvalOrFunction
			reduce(256), // ==, reduce: MlrvalOrFunction
			reduce(256), // !=, reduce: MlrvalOrFunction
			reduce(256), // <=>, reduce: MlrvalOrFunction
			reduce(256), // >=, reduce: MlrvalOrFunction
			reduce(256), // <, reduce: MlrvalOrFunction
			reduce(256), // <=, reduce: MlrvalOrFunction
			reduce(256), // ^, reduce: MlrvalOrFunction
			reduce(256), // &, reduce: MlrvalOrFunction
			reduce(256), // <<, reduce: MlrvalOrFunction
			reduce(256), // >>>, reduce: MlrvalOrFunction
			reduce(256), // +, reduce: MlrvalOrFunction
			reduce(256), // -, reduce: MlrvalOrFunction
			reduce(256), // .+, reduce: MlrvalOrFunction
			reduce(256), // .-, reduce: MlrvalOrFunction
			reduce(256), // *, reduce: MlrvalOrFunction
			reduce(256), // /, reduce: MlrvalOrFunction
			reduce(256), // //, reduce: MlrvalOrFunction
			reduce(256), // %, reduce: MlrvalOrFunction
			reduce(256), // .*, reduce: MlrvalOrFunction
			reduce(256), // ./, reduce: MlrvalOrFunction
			reduce(256), // .//, reduce: MlrvalOrFunction
			reduce(256), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(256), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(256), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(256), // [[, reduce: MlrvalOrFunction
			reduce(256), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3535
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(257), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(257), // >, reduce: MlrvalOrFunction
			reduce(257), // >>, reduce: MlrvalOrFunction
			reduce(257), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(257), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(257), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(257), // ||, reduce: MlrvalOrFunction
			reduce(257), // ^^, reduce: MlrvalOrFunction
			reduce(257), // &&, reduce: MlrvalOrFunction
			reduce(257), // ??, reduce: MlrvalOrFunction
			reduce(257), // ???, reduce: MlrvalOrFunction
			reduce(257), // =~, reduce: MlrvalOrFunction
			reduce(257), // !=~, reduce: MlrvalOrFunction
			reduce(257), // ==, reduce: MlrvalOrFunction
			reduce(257), // !=, reduce: MlrvalOrFunction
			reduce(257), // <=>, reduce: MlrvalOrFunction
			reduce(257), // >=, reduce: MlrvalOrFunction
			reduce(257), // <, reduce: MlrvalOrFunction
			reduce(257), // <=, reduce: MlrvalOrFunction
			reduce(257), // ^, reduce: MlrvalOrFunction
			reduce(257), // &, reduce: MlrvalOrFunction
			reduce(257), // <<, reduce: MlrvalOrFunction
			reduce(257), // >>>, reduce: MlrvalOrFunction
			reduce(257), // +, reduce: MlrvalOrFunction
			reduce(257), // -, reduce: MlrvalOrFunction
			reduce(257), // .+, reduce: MlrvalOrFunction
			reduce(257), // .-, reduce: MlrvalOrFunction
			reduce(257), // *, reduce: MlrvalOrFunction
			reduce(257), // /, reduce: MlrvalOrFunction
			reduce(257), // //, reduce: MlrvalOrFunction
			reduce(257), // %, reduce: MlrvalOrFunction
			reduce(257), // .*, reduce: MlrvalOrFunction
			reduce(257), // ./, reduce: MlrvalOrFunction
			reduce(257), // .//, reduce: MlrvalOrFunction
			reduce(257), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(257), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(257), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(257), // [[, reduce: MlrvalOrFunction
			reduce(257), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3536
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(258), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(258), // >, reduce: MlrvalOrFunction
			reduce(258), // >>, reduce: MlrvalOrFunction
			reduce(258), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(258), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(258), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(258), // ||, reduce: MlrvalOrFunction
			reduce(258), // ^^, reduce: MlrvalOrFunction
			reduce(258), // &&, reduce: MlrvalOrFunction
			reduce(258), // ??, reduce: MlrvalOrFunction
			reduce(258), // ???, reduce: MlrvalOrFunction
			reduce(258), // =~, reduce: MlrvalOrFunction
			reduce(258), // !=~, reduce: MlrvalOrFunction
			reduce(258), // ==, reduce: MlrvalOrFunction
			reduce(258), // !=, reduce: MlrvalOrFunction
			reduce(258), // <=>, reduce: MlrvalOrFunction
			reduce(258), // >=, reduce: MlrvalOrFunction
			reduce(258), // <, reduce: MlrvalOrFunction
			reduce(258), // <=, reduce: MlrvalOrFunction
			reduce(258), // ^, reduce: MlrvalOrFunction
			reduce(258), // &, reduce: MlrvalOrFunction
			reduce(258), // <<, reduce: MlrvalOrFunction
			reduce(258), // >>>, reduce: MlrvalOrFunction
			reduce(258), // +, reduce: MlrvalOrFunction
			reduce(258), // -, reduce: MlrvalOrFunction
			reduce(258), // .+, reduce: MlrvalOrFunction
			reduce(258), // .-, reduce: MlrvalOrFunction
			reduce(258), // *, reduce: MlrvalOrFunction
			reduce(258), // /, reduce: MlrvalOrFunction
			reduce(258), // //, reduce: MlrvalOrFunction
			reduce(258), // %, reduce: MlrvalOrFunction
			reduce(258), // .*, reduce: MlrvalOrFunction
			reduce(258), // ./, reduce: MlrvalOrFunction
			reduce(258), // .//, reduce: MlrvalOrFunction
			reduce(258), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(258), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(258), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(258), // [[, reduce: MlrvalOrFunction
			reduce(258), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3537
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(259), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(259), // >, reduce: MlrvalOrFunction
			reduce(259), // >>, reduce: MlrvalOrFunction
			reduce(259), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(259), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(259), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(259), // ||, reduce: MlrvalOrFunction
			reduce(259), // ^^, reduce: MlrvalOrFunction
			reduce(259), // &&, reduce: MlrvalOrFunction
			reduce(259), // ??, reduce: MlrvalOrFunction
			reduce(259), // ???, reduce: MlrvalOrFunction
			reduce(259), // =~, reduce: MlrvalOrFunction
			reduce(259), // !=~, reduce: MlrvalOrFunction
			reduce(259), // ==, reduce: MlrvalOrFunction
			reduce(259), // !=, reduce: MlrvalOrFunction
			reduce(259), // <=>, reduce: MlrvalOrFunction
			reduce(259), // >=, reduce: MlrvalOrFunction
			reduce(259), // <, reduce: MlrvalOrFunction
			reduce(259), // <=, reduce: MlrvalOrFunction
			reduce(259), // ^, reduce: MlrvalOrFunction
			reduce(259), // &, reduce: MlrvalOrFunction
			reduce(259), // <<, reduce: MlrvalOrFunction
			reduce(259), // >>>, reduce: MlrvalOrFunction
			reduce(259), // +, reduce: MlrvalOrFunction
			reduce(259), // -, reduce: MlrvalOrFunction
			reduce(259), // .+, reduce: MlrvalOrFunction
			reduce(259), // .-, reduce: MlrvalOrFunction
			reduce(259), // *, reduce: MlrvalOrFunction
			reduce(259), // /, reduce: MlrvalOrFunction
			reduce(259), // //, reduce: MlrvalOrFunction
			reduce(259), // %, reduce: MlrvalOrFunction
			reduce(259), // .*, reduce: MlrvalOrFunction
			reduce(259), // ./, reduce: MlrvalOrFunction
			reduce(259), // .//, reduce: MlrvalOrFunction
			reduce(259), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(259), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(259), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(259), // [[, reduce: MlrvalOrFunction
			reduce(259), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3538
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(267), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(267), // >, reduce: MlrvalOrFunction
			reduce(267), // >>, reduce: MlrvalOrFunction
			reduce(267), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(267), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(267), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(267), // ||, reduce: MlrvalOrFunction
			reduce(267), // ^^, reduce: MlrvalOrFunction
			reduce(267), // &&, reduce: MlrvalOrFunction
			reduce(267), // ??, reduce: MlrvalOrFunction
			reduce(267), // ???, reduce: MlrvalOrFunction
			reduce(267), // =~, reduce: MlrvalOrFunction
			reduce(267), // !=~, reduce: MlrvalOrFunction
			reduce(267), // ==, reduce: MlrvalOrFunction
			reduce(267), // !=, reduce: MlrvalOrFunction
			reduce(267), // <=>, reduce: MlrvalOrFunction
			reduce(267), // >=, reduce: MlrvalOrFunction
			reduce(267), // <, reduce: MlrvalOrFunction
			reduce(267), // <=, reduce: MlrvalOrFunction
			reduce(267), // ^, reduce: MlrvalOrFunction
			reduce(267), // &, reduce: MlrvalOrFunction
			reduce(267), // <<, reduce: MlrvalOrFunction
			reduce(267), // >>>, reduce: MlrvalOrFunction
			reduce(267), // +, reduce: MlrvalOrFunction
			reduce(267), // -, reduce: MlrvalOrFunction
			reduce(267), // .+, reduce: MlrvalOrFunction
			reduce(267), // .-, reduce: MlrvalOrFunction
			reduce(267), // *, reduce: MlrvalOrFunction
			reduce(267), // /, reduce: MlrvalOrFunction
			reduce(267), // //, reduce: MlrvalOrFunction
			reduce(267), // %, reduce: MlrvalOrFunction
			reduce(267), // .*, reduce: MlrvalOrFunction
			reduce(267), // ./, reduce: MlrvalOrFunction
			reduce(267), // .//, reduce: MlrvalOrFunction
			reduce(267), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(267), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(267), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(267), // [[, reduce: MlrvalOrFunction
			reduce(267), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3539
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(4091), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3540
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(4092), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3541
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(128), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(128), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3542
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(129), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(129), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3543
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(130), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(130), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3544
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(131), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(131), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3545
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(132), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(132), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3546
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(133), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(133), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3547
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(134), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(134), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3548
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(135), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(135), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3549
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(136), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(136), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3550
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(137), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(137), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3551
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(138), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(138), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3552
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(139), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(139), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3553
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(140), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(140), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3554
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(141), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(141), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3555
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(142), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(142), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3556
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(143), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(143), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3557
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(144), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(144), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3558
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(145), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(145), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3559
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(146), // ;, reduce: Assignment
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(146), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3560
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			shift(4093), // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3561
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			reduce(118), // in, reduce: LocalVariable
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3562
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4094), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3563
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(208), // =, reduce: PrecedenceChainEnd
			nil,         // unset
			nil,         // filter
			reduce(208), // >, reduce: PrecedenceChainEnd
			reduce(208), // >>, reduce: PrecedenceChainEnd
			reduce(208), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(208), // ||=, reduce: PrecedenceChainEnd
			reduce(208), // ^^=, reduce: PrecedenceChainEnd
			reduce(208), // &&=, reduce: PrecedenceChainEnd
			reduce(208), // ??=, reduce: PrecedenceChainEnd
			reduce(208), // ???=, reduce: PrecedenceChainEnd
			reduce(208), // |=, reduce: PrecedenceChainEnd
			reduce(208), // &=, reduce: PrecedenceChainEnd
			reduce(208), // ^=, reduce: PrecedenceChainEnd
			reduce(208), // <<=, reduce: PrecedenceChainEnd
			reduce(208), // >>=, reduce: PrecedenceChainEnd
			reduce(208), // >>>=, reduce: PrecedenceChainEnd
			reduce(208), // +=, reduce: PrecedenceChainEnd
			reduce(208), // .=, reduce: PrecedenceChainEnd
			reduce(208), // -=, reduce: PrecedenceChainEnd
			reduce(208), // *=, reduce: PrecedenceChainEnd
			reduce(208), // /=, reduce: PrecedenceChainEnd
			reduce(208), // //=, reduce: PrecedenceChainEnd
			reduce(208), // %=, reduce: PrecedenceChainEnd
			reduce(208), // **=, reduce: PrecedenceChainEnd
			reduce(208), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(208), // ||, reduce: PrecedenceChainEnd
			reduce(208), // ^^, reduce: PrecedenceChainEnd
			reduce(208), // &&, reduce: PrecedenceChainEnd
			reduce(208), // ??, reduce: PrecedenceChainEnd
			reduce(208), // ???, reduce: PrecedenceChainEnd
			reduce(208), // =~, reduce: PrecedenceChainEnd
			reduce(208), // !=~, reduce: PrecedenceChainEnd
			reduce(208), // ==, reduce: PrecedenceChainEnd
			reduce(208), // !=, reduce: PrecedenceChainEnd
			reduce(208), // <=>, reduce: PrecedenceChainEnd
			reduce(208), // >=, reduce: PrecedenceChainEnd
			reduce(208), // <, reduce: PrecedenceChainEnd
			reduce(208), // <=, reduce: PrecedenceChainEnd
			reduce(208), // ^, reduce: PrecedenceChainEnd
			reduce(208), // &, reduce: PrecedenceChainEnd
			reduce(208), // <<, reduce: PrecedenceChainEnd
			reduce(208), // >>>, reduce: PrecedenceChainEnd
			reduce(208), // +, reduce: PrecedenceChainEnd
			reduce(208), // -, reduce: PrecedenceChainEnd
			reduce(208), // .+, reduce: PrecedenceChainEnd
			reduce(208), // .-, reduce: PrecedenceChainEnd
			reduce(208), // *, reduce: PrecedenceChainEnd
			reduce(208), // /, reduce: PrecedenceChainEnd
			reduce(208), // //, reduce: PrecedenceChainEnd
			reduce(208), // %, reduce: PrecedenceChainEnd
			reduce(208), // .*, reduce: PrecedenceChainEnd
			reduce(208), // ./, reduce: PrecedenceChainEnd
			reduce(208), // .//, reduce: PrecedenceChainEnd
			reduce(208), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(208), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3564
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1352), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3565
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1352), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3566
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(4097), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3567
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(105), // =, reduce: IndirectFieldValue
			nil,         // unset
			nil,         // filter
			reduce(105), // >, reduce: IndirectFieldValue
			reduce(105), // >>, reduce: IndirectFieldValue
			reduce(105), // |, reduce: IndirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(105), // ||=, reduce: IndirectFieldValue
			reduce(105), // ^^=, reduce: IndirectFieldValue
			reduce(105), // &&=, reduce: IndirectFieldValue
			reduce(105), // ??=, reduce: IndirectFieldValue
			reduce(105), // ???=, reduce: IndirectFieldValue
			reduce(105), // |=, reduce: IndirectFieldValue
			reduce(105), // &=, reduce: IndirectFieldValue
			reduce(105), // ^=, reduce: IndirectFieldValue
			reduce(105), // <<=, reduce: IndirectFieldValue
			reduce(105), // >>=, reduce: IndirectFieldValue
			reduce(105), // >>>=, reduce: IndirectFieldValue
			reduce(105), // +=, reduce: IndirectFieldValue
			reduce(105), // .=, reduce: IndirectFieldValue
			reduce(105), // -=, reduce: IndirectFieldValue
			reduce(105), // *=, reduce: IndirectFieldValue
			reduce(105), // /=, reduce: IndirectFieldValue
			reduce(105), // //=, reduce: IndirectFieldValue
			reduce(105), // %=, reduce: IndirectFieldValue
			reduce(105), // **=, reduce: IndirectFieldValue
			reduce(105), // ?, reduce: IndirectFieldValue
			nil,         // :
			reduce(105), // ||, reduce: IndirectFieldValue
			reduce(105), // ^^, reduce: IndirectFieldValue
			reduce(105), // &&, reduce: IndirectFieldValue
			reduce(105), // ??, reduce: IndirectFieldValue
			reduce(105), // ???, reduce: IndirectFieldValue
			reduce(105), // =~, reduce: IndirectFieldValue
			reduce(105), // !=~, reduce: IndirectFieldValue
			reduce(105), // ==, reduce: IndirectFieldValue
			reduce(105), // !=, reduce: IndirectFieldValue
			reduce(105), // <=>, reduce: IndirectFieldValue
			reduce(105), // >=, reduce: IndirectFieldValue
			reduce(105), // <, reduce: IndirectFieldValue
			reduce(105), // <=, reduce: IndirectFieldValue
			reduce(105), // ^, reduce: IndirectFieldValue
			reduce(105), // &, reduce: IndirectFieldValue
			reduce(105), // <<, reduce: IndirectFieldValue
			reduce(105), // >>>, reduce: IndirectFieldValue
			reduce(105), // +, reduce: IndirectFieldValue
			reduce(105), // -, reduce: IndirectFieldValue
			reduce(105), // .+, reduce: IndirectFieldValue
			reduce(105), // .-, reduce: IndirectFieldValue
			reduce(105), // *, reduce: IndirectFieldValue
			reduce(105), // /, reduce: IndirectFieldValue
			reduce(105), // //, reduce: IndirectFieldValue
			reduce(105), // %, reduce: IndirectFieldValue
			reduce(105), // .*, reduce: IndirectFieldValue
			reduce(105), // ./, reduce: IndirectFieldValue
			reduce(105), // .//, reduce: IndirectFieldValue
			reduce(105), // ., reduce: IndirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(105), // **, reduce: IndirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(105), // [, reduce: IndirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(105), // [[, reduce: IndirectFieldValue
			reduce(105), // [[[, reduce: IndirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3568
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4098), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3569
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4099), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3570
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(114), // =, reduce: IndirectOosvarValue
			nil,         // unset
			nil,         // filter
			reduce(114), // >, reduce: IndirectOosvarValue
			reduce(114), // >>, reduce: IndirectOosvarValue
			reduce(114), // |, reduce: IndirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(114), // ||=, reduce: IndirectOosvarValue
			reduce(114), // ^^=, reduce: IndirectOosvarValue
			reduce(114), // &&=, reduce: IndirectOosvarValue
			reduce(114), // ??=, reduce: IndirectOosvarValue
			reduce(114), // ???=, reduce: IndirectOosvarValue
			reduce(114), // |=, reduce: IndirectOosvarValue
			reduce(114), // &=, reduce: IndirectOosvarValue
			reduce(114), // ^=, reduce: IndirectOosvarValue
			reduce(114), // <<=, reduce: IndirectOosvarValue
			reduce(114), // >>=, reduce: IndirectOosvarValue
			reduce(114), // >>>=, reduce: IndirectOosvarValue
			reduce(114), // +=, reduce: IndirectOosvarValue
			reduce(114), // .=, reduce: IndirectOosvarValue
			reduce(114), // -=, reduce: IndirectOosvarValue
			reduce(114), // *=, reduce: IndirectOosvarValue
			reduce(114), // /=, reduce: IndirectOosvarValue
			reduce(114), // //=, reduce: IndirectOosvarValue
			reduce(114), // %=, reduce: IndirectOosvarValue
			reduce(114), // **=, reduce: IndirectOosvarValue
			reduce(114), // ?, reduce: IndirectOosvarValue
			nil,         // :
			reduce(114), // ||, reduce: IndirectOosvarValue
			reduce(114), // ^^, reduce: IndirectOosvarValue
			reduce(114), // &&, reduce: IndirectOosvarValue
			reduce(114), // ??, reduce: IndirectOosvarValue
			reduce(114), // ???, reduce: IndirectOosvarValue
			reduce(114), // =~, reduce: IndirectOosvarValue
			reduce(114), // !=~, reduce: IndirectOosvarValue
			reduce(114), // ==, reduce: IndirectOosvarValue
			reduce(114), // !=, reduce: IndirectOosvarValue
			reduce(114), // <=>, reduce: IndirectOosvarValue
			reduce(114), // >=, reduce: IndirectOosvarValue
			reduce(114), // <, reduce: IndirectOosvarValue
			reduce(114), // <=, reduce: IndirectOosvarValue
			reduce(114), // ^, reduce: IndirectOosvarValue
			reduce(114), // &, reduce: IndirectOosvarValue
			reduce(114), // <<, reduce: IndirectOosvarValue
			reduce(114), // >>>, reduce: IndirectOosvarValue
			reduce(114), // +, reduce: IndirectOosvarValue
			reduce(114), // -, reduce: IndirectOosvarValue
			reduce(114), // .+, reduce: IndirectOosvarValue
			reduce(114), // .-, reduce: IndirectOosvarValue
			reduce(114), // *, reduce: IndirectOosvarValue
			reduce(114), // /, reduce: IndirectOosvarValue
			reduce(114), // //, reduce: IndirectOosvarValue
			reduce(114), // %, reduce: IndirectOosvarValue
			reduce(114), // .*, reduce: IndirectOosvarValue
			reduce(114), // ./, reduce: IndirectOosvarValue
			reduce(114), // .//, reduce: IndirectOosvarValue
			reduce(114), // ., reduce: IndirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(114), // **, reduce: IndirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(114), // [, reduce: IndirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(114), // [[, reduce: IndirectOosvarValue
			reduce(114), // [[[, reduce: IndirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3571
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4100), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3572
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(151), // =, reduce: LogicalOrTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(151), // ||=, reduce: LogicalOrTerm
			reduce(151), // ^^=, reduce: LogicalOrTerm
			reduce(151), // &&=, reduce: LogicalOrTerm
			reduce(151), // ??=, reduce: LogicalOrTerm
			reduce(151), // ???=, reduce: LogicalOrTerm
			reduce(151), // |=, reduce: LogicalOrTerm
			reduce(151), // &=, reduce: LogicalOrTerm
			reduce(151), // ^=, reduce: LogicalOrTerm
			reduce(151), // <<=, reduce: LogicalOrTerm
			reduce(151), // >>=, reduce: LogicalOrTerm
			reduce(151), // >>>=, reduce: LogicalOrTerm
			reduce(151), // +=, reduce: LogicalOrTerm
			reduce(151), // .=, reduce: LogicalOrTerm
			reduce(151), // -=, reduce: LogicalOrTerm
			reduce(151), // *=, reduce: LogicalOrTerm
			reduce(151), // /=, reduce: LogicalOrTerm
			reduce(151), // //=, reduce: LogicalOrTerm
			reduce(151), // %=, reduce: LogicalOrTerm
			reduce(151), // **=, reduce: LogicalOrTerm
			reduce(151), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(151), // ||, reduce: LogicalOrTerm
			shift(2848), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3573
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(153), // =, reduce: LogicalXORTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(153), // ||=, reduce: LogicalXORTerm
			reduce(153), // ^^=, reduce: LogicalXORTerm
			reduce(153), // &&=, reduce: LogicalXORTerm
			reduce(153), // ??=, reduce: LogicalXORTerm
			reduce(153), // ???=, reduce: LogicalXORTerm
			reduce(153), // |=, reduce: LogicalXORTerm
			reduce(153), // &=, reduce: LogicalXORTerm
			reduce(153), // ^=, reduce: LogicalXORTerm
			reduce(153), // <<=, reduce: LogicalXORTerm
			reduce(153), // >>=, reduce: LogicalXORTerm
			reduce(153), // >>>=, reduce: LogicalXORTerm
			reduce(153), // +=, reduce: LogicalXORTerm
			reduce(153), // .=, reduce: LogicalXORTerm
			reduce(153), // -=, reduce: LogicalXORTerm
			reduce(153), // *=, reduce: LogicalXORTerm
			reduce(153), // /=, reduce: LogicalXORTerm
			reduce(153), // //=, reduce: LogicalXORTerm
			reduce(153), // %=, reduce: LogicalXORTerm
			reduce(153), // **=, reduce: LogicalXORTerm
			reduce(153), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(153), // ||, reduce: LogicalXORTerm
			reduce(153), // ^^, reduce: LogicalXORTerm
			shift(2849), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3574
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(155), // =, reduce: LogicalAndTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(155), // ||=, reduce: LogicalAndTerm
			reduce(155), // ^^=, reduce: LogicalAndTerm
			reduce(155), // &&=, reduce: LogicalAndTerm
			reduce(155), // ??=, reduce: LogicalAndTerm
			reduce(155), // ???=, reduce: LogicalAndTerm
			reduce(155), // |=, reduce: LogicalAndTerm
			reduce(155), // &=, reduce: LogicalAndTerm
			reduce(155), // ^=, reduce: LogicalAndTerm
			reduce(155), // <<=, reduce: LogicalAndTerm
			reduce(155), // >>=, reduce: LogicalAndTerm
			reduce(155), // >>>=, reduce: LogicalAndTerm
			reduce(155), // +=, reduce: LogicalAndTerm
			reduce(155), // .=, reduce: LogicalAndTerm
			reduce(155), // -=, reduce: LogicalAndTerm
			reduce(155), // *=, reduce: LogicalAndTerm
			reduce(155), // /=, reduce: LogicalAndTerm
			reduce(155), // //=, reduce: LogicalAndTerm
			reduce(155), // %=, reduce: LogicalAndTerm
			reduce(155), // **=, reduce: LogicalAndTerm
			reduce(155), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(155), // ||, reduce: LogicalAndTerm
			reduce(155), // ^^, reduce: LogicalAndTerm
			reduce(155), // &&, reduce: LogicalAndTerm
			shift(2850), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3575
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(157), // =, reduce: AbsentCoalesceTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(157), // ||=, reduce: AbsentCoalesceTerm
			reduce(157), // ^^=, reduce: AbsentCoalesceTerm
			reduce(157), // &&=, reduce: AbsentCoalesceTerm
			reduce(157), // ??=, reduce: AbsentCoalesceTerm
			reduce(157), // ???=, reduce: AbsentCoalesceTerm
			reduce(157), // |=, reduce: AbsentCoalesceTerm
			reduce(157), // &=, reduce: AbsentCoalesceTerm
			reduce(157), // ^=, reduce: AbsentCoalesceTerm
			reduce(157), // <<=, reduce: AbsentCoalesceTerm
			reduce(157), // >>=, reduce: AbsentCoalesceTerm
			reduce(157), // >>>=, reduce: AbsentCoalesceTerm
			reduce(157), // +=, reduce: AbsentCoalesceTerm
			reduce(157), // .=, reduce: AbsentCoalesceTerm
			reduce(157), // -=, reduce: AbsentCoalesceTerm
			reduce(157), // *=, reduce: AbsentCoalesceTerm
			reduce(157), // /=, reduce: AbsentCoalesceTerm
			reduce(157), // //=, reduce: AbsentCoalesceTerm
			reduce(157), // %=, reduce: AbsentCoalesceTerm
			reduce(157), // **=, reduce: AbsentCoalesceTerm
			reduce(157), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(157), // ||, reduce: AbsentCoalesceTerm
			reduce(157), // ^^, reduce: AbsentCoalesceTerm
			reduce(157), // &&, reduce: AbsentCoalesceTerm
			reduce(157), // ??, reduce: AbsentCoalesceTerm
			shift(2851), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3576
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(159), // =, reduce: EmptyCoalesceTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(159), // ||=, reduce: EmptyCoalesceTerm
			reduce(159), // ^^=, reduce: EmptyCoalesceTerm
			reduce(159), // &&=, reduce: EmptyCoalesceTerm
			reduce(159), // ??=, reduce: EmptyCoalesceTerm
			reduce(159), // ???=, reduce: EmptyCoalesceTerm
			reduce(159), // |=, reduce: EmptyCoalesceTerm
			reduce(159), // &=, reduce: EmptyCoalesceTerm
			reduce(159), // ^=, reduce: EmptyCoalesceTerm
			reduce(159), // <<=, reduce: EmptyCoalesceTerm
			reduce(159), // >>=, reduce: EmptyCoalesceTerm
			reduce(159), // >>>=, reduce: EmptyCoalesceTerm
			reduce(159), // +=, reduce: EmptyCoalesceTerm
			reduce(159), // .=, reduce: EmptyCoalesceTerm
			reduce(159), // -=, reduce: EmptyCoalesceTerm
			reduce(159), // *=, reduce: EmptyCoalesceTerm
			reduce(159), // /=, reduce: EmptyCoalesceTerm
			reduce(159), // //=, reduce: EmptyCoalesceTerm
			reduce(159), // %=, reduce: EmptyCoalesceTerm
			reduce(159), // **=, reduce: EmptyCoalesceTerm
			reduce(159), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(159), // ||, reduce: EmptyCoalesceTerm
			reduce(159), // ^^, reduce: EmptyCoalesceTerm
			reduce(159), // &&, reduce: EmptyCoalesceTerm
			reduce(159), // ??, reduce: EmptyCoalesceTerm
			reduce(159), // ???, reduce: EmptyCoalesceTerm
			shift(2852), // =~
			shift(2853), // !=~
			shift(2854), // ==
			shift(2855), // !=
			shift(2856), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3577
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(161), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(2857), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(161), // ||=, reduce: EqneTerm
			reduce(161), // ^^=, reduce: EqneTerm
			reduce(161), // &&=, reduce: EqneTerm
			reduce(161), // ??=, reduce: EqneTerm
			reduce(161), // ???=, reduce: EqneTerm
			reduce(161), // |=, reduce: EqneTerm
			reduce(161), // &=, reduce: EqneTerm
			reduce(161), // ^=, reduce: EqneTerm
			reduce(161), // <<=, reduce: EqneTerm
			reduce(161), // >>=, reduce: EqneTerm
			reduce(161), // >>>=, reduce: EqneTerm
			reduce(161), // +=, reduce: EqneTerm
			reduce(161), // .=, reduce: EqneTerm
			reduce(161), // -=, reduce: EqneTerm
			reduce(161), // *=, reduce: EqneTerm
			reduce(161), // /=, reduce: EqneTerm
			reduce(161), // //=, reduce: EqneTerm
			reduce(161), // %=, reduce: EqneTerm
			reduce(161), // **=, reduce: EqneTerm
			reduce(161), // ?, reduce: EqneTerm
			nil,         // :
			reduce(161), // ||, reduce: EqneTerm
			reduce(161), // ^^, reduce: EqneTerm
			reduce(161), // &&, reduce: EqneTerm
			reduce(161), // ??, reduce: EqneTerm
			reduce(161), // ???, reduce: EqneTerm
			reduce(161), // =~, reduce: EqneTerm
			reduce(161), // !=~, reduce: EqneTerm
			reduce(161), // ==, reduce: EqneTerm
			reduce(161), // !=, reduce: EqneTerm
			reduce(161), // <=>, reduce: EqneTerm
			shift(2858), // >=
			shift(2859), // <
			shift(2860), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3578
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(162), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(2857), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(162), // ||=, reduce: EqneTerm
			reduce(162), // ^^=, reduce: EqneTerm
			reduce(162), // &&=, reduce: EqneTerm
			reduce(162), // ??=, reduce: EqneTerm
			reduce(162), // ???=, reduce: EqneTerm
			reduce(162), // |=, reduce: EqneTerm
			reduce(162), // &=, reduce: EqneTerm
			reduce(162), // ^=, reduce: EqneTerm
			reduce(162), // <<=, reduce: EqneTerm
			reduce(162), // >>=, reduce: EqneTerm
			reduce(162), // >>>=, reduce: EqneTerm
			reduce(162), // +=, reduce: EqneTerm
			reduce(162), // .=, reduce: EqneTerm
			reduce(162), // -=, reduce: EqneTerm
			reduce(162), // *=, reduce: EqneTerm
			reduce(162), // /=, reduce: EqneTerm
			reduce(162), // //=, reduce: EqneTerm
			reduce(162), // %=, reduce: EqneTerm
			reduce(162), // **=, reduce: EqneTerm
			reduce(162), // ?, reduce: EqneTerm
			nil,         // :
			reduce(162), // ||, reduce: EqneTerm
			reduce(162), // ^^, reduce: EqneTerm
			reduce(162), // &&, reduce: EqneTerm
			reduce(162), // ??, reduce: EqneTerm
			reduce(162), // ???, reduce: EqneTerm
			reduce(162), // =~, reduce: EqneTerm
			reduce(162), // !=~, reduce: EqneTerm
			reduce(162), // ==, reduce: EqneTerm
			reduce(162), // !=, reduce: EqneTerm
			reduce(162), // <=>, reduce: EqneTerm
			shift(2858), // >=
			shift(2859), // <
			shift(2860), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3579
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(163), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(2857), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(163), // ||=, reduce: EqneTerm
			reduce(163), // ^^=, reduce: EqneTerm
			reduce(163), // &&=, reduce: EqneTerm
			reduce(163), // ??=, reduce: EqneTerm
			reduce(163), // ???=, reduce: EqneTerm
			reduce(163), // |=, reduce: EqneTerm
			reduce(163), // &=, reduce: EqneTerm
			reduce(163), // ^=, reduce: EqneTerm
			reduce(163), // <<=, reduce: EqneTerm
			reduce(163), // >>=, reduce: EqneTerm
			reduce(163), // >>>=, reduce: EqneTerm
			reduce(163), // +=, reduce: EqneTerm
			reduce(163), // .=, reduce: EqneTerm
			reduce(163), // -=, reduce: EqneTerm
			reduce(163), // *=, reduce: EqneTerm
			reduce(163), // /=, reduce: EqneTerm
			reduce(163), // //=, reduce: EqneTerm
			reduce(163), // %=, reduce: EqneTerm
			reduce(163), // **=, reduce: EqneTerm
			reduce(163), // ?, reduce: EqneTerm
			nil,         // :
			reduce(163), // ||, reduce: EqneTerm
			reduce(163), // ^^, reduce: EqneTerm
			reduce(163), // &&, reduce: EqneTerm
			reduce(163), // ??, reduce: EqneTerm
			reduce(163), // ???, reduce: EqneTerm
			reduce(163), // =~, reduce: EqneTerm
			reduce(163), // !=~, reduce: EqneTerm
			reduce(163), // ==, reduce: EqneTerm
			reduce(163), // !=, reduce: EqneTerm
			reduce(163), // <=>, reduce: EqneTerm
			shift(2858), // >=
			shift(2859), // <
			shift(2860), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3580
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(164), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(2857), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(164), // ||=, reduce: EqneTerm
			reduce(164), // ^^=, reduce: EqneTerm
			reduce(164), // &&=, reduce: EqneTerm
			reduce(164), // ??=, reduce: EqneTerm
			reduce(164), // ???=, reduce: EqneTerm
			reduce(164), // |=, reduce: EqneTerm
			reduce(164), // &=, reduce: EqneTerm
			reduce(164), // ^=, reduce: EqneTerm
			reduce(164), // <<=, reduce: EqneTerm
			reduce(164), // >>=, reduce: EqneTerm
			reduce(164), // >>>=, reduce: EqneTerm
			reduce(164), // +=, reduce: EqneTerm
			reduce(164), // .=, reduce: EqneTerm
			reduce(164), // -=, reduce: EqneTerm
			reduce(164), // *=, reduce: EqneTerm
			reduce(164), // /=, reduce: EqneTerm
			reduce(164), // //=, reduce: EqneTerm
			reduce(164), // %=, reduce: EqneTerm
			reduce(164), // **=, reduce: EqneTerm
			reduce(164), // ?, reduce: EqneTerm
			nil,         // :
			reduce(164), // ||, reduce: EqneTerm
			reduce(164), // ^^, reduce: EqneTerm
			reduce(164), // &&, reduce: EqneTerm
			reduce(164), // ??, reduce: EqneTerm
			reduce(164), // ???, reduce: EqneTerm
			reduce(164), // =~, reduce: EqneTerm
			reduce(164), // !=~, reduce: EqneTerm
			reduce(164), // ==, reduce: EqneTerm
			reduce(164), // !=, reduce: EqneTerm
			reduce(164), // <=>, reduce: EqneTerm
			shift(2858), // >=
			shift(2859), // <
			shift(2860), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3581
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(165), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(2857), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(165), // ||=, reduce: EqneTerm
			reduce(165), // ^^=, reduce: EqneTerm
			reduce(165), // &&=, reduce: EqneTerm
			reduce(165), // ??=, reduce: EqneTerm
			reduce(165), // ???=, reduce: EqneTerm
			reduce(165), // |=, reduce: EqneTerm
			reduce(165), // &=, reduce: EqneTerm
			reduce(165), // ^=, reduce: EqneTerm
			reduce(165), // <<=, reduce: EqneTerm
			reduce(165), // >>=, reduce: EqneTerm
			reduce(165), // >>>=, reduce: EqneTerm
			reduce(165), // +=, reduce: EqneTerm
			reduce(165), // .=, reduce: EqneTerm
			reduce(165), // -=, reduce: EqneTerm
			reduce(165), // *=, reduce: EqneTerm
			reduce(165), // /=, reduce: EqneTerm
			reduce(165), // //=, reduce: EqneTerm
			reduce(165), // %=, reduce: EqneTerm
			reduce(165), // **=, reduce: EqneTerm
			reduce(165), // ?, reduce: EqneTerm
			nil,         // :
			reduce(165), // ||, reduce: EqneTerm
			reduce(165), // ^^, reduce: EqneTerm
			reduce(165), // &&, reduce: EqneTerm
			reduce(165), // ??, reduce: EqneTerm
			reduce(165), // ???, reduce: EqneTerm
			reduce(165), // =~, reduce: EqneTerm
			reduce(165), // !=~, reduce: EqneTerm
			reduce(165), // ==, reduce: EqneTerm
			reduce(165), // !=, reduce: EqneTerm
			reduce(165), // <=>, reduce: EqneTerm
			shift(2858), // >=
			shift(2859), // <
			shift(2860), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3582
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(167), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(167), // >, reduce: CmpTerm
			nil,         // >>
			shift(2861), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(167), // ||=, reduce: CmpTerm
			reduce(167), // ^^=, reduce: CmpTerm
			reduce(167), // &&=, reduce: CmpTerm
			reduce(167), // ??=, reduce: CmpTerm
			reduce(167), // ???=, reduce: CmpTerm
			reduce(167), // |=, reduce: CmpTerm
			reduce(167), // &=, reduce: CmpTerm
			reduce(167), // ^=, reduce: CmpTerm
			reduce(167), // <<=, reduce: CmpTerm
			reduce(167), // >>=, reduce: CmpTerm
			reduce(167), // >>>=, reduce: CmpTerm
			reduce(167), // +=, reduce: CmpTerm
			reduce(167), // .=, reduce: CmpTerm
			reduce(167), // -=, reduce: CmpTerm
			reduce(167), // *=, reduce: CmpTerm
			reduce(167), // /=, reduce: CmpTerm
			reduce(167), // //=, reduce: CmpTerm
			reduce(167), // %=, reduce: CmpTerm
			reduce(167), // **=, reduce: CmpTerm
			reduce(167), // ?, reduce: CmpTerm
			nil,         // :
			reduce(167), // ||, reduce: CmpTerm
			reduce(167), // ^^, reduce: CmpTerm
			reduce(167), // &&, reduce: CmpTerm
			reduce(167), // ??, reduce: CmpTerm
			reduce(167), // ???, reduce: CmpTerm
			reduce(167), // =~, reduce: CmpTerm
			reduce(167), // !=~, reduce: CmpTerm
			reduce(167), // ==, reduce: CmpTerm
			reduce(167), // !=, reduce: CmpTerm
			reduce(167), // <=>, reduce: CmpTerm
			reduce(167), // >=, reduce: CmpTerm
			reduce(167), // <, reduce: CmpTerm
			reduce(167), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3583
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(168), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(168), // >, reduce: CmpTerm
			nil,         // >>
			shift(2861), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(168), // ||=, reduce: CmpTerm
			reduce(168), // ^^=, reduce: CmpTerm
			reduce(168), // &&=, reduce: CmpTerm
			reduce(168), // ??=, reduce: CmpTerm
			reduce(168), // ???=, reduce: CmpTerm
			reduce(168), // |=, reduce: CmpTerm
			reduce(168), // &=, reduce: CmpTerm
			reduce(168), // ^=, reduce: CmpTerm
			reduce(168), // <<=, reduce: CmpTerm
			reduce(168), // >>=, reduce: CmpTerm
			reduce(168), // >>>=, reduce: CmpTerm
			reduce(168), // +=, reduce: CmpTerm
			reduce(168), // .=, reduce: CmpTerm
			reduce(168), // -=, reduce: CmpTerm
			reduce(168), // *=, reduce: CmpTerm
			reduce(168), // /=, reduce: CmpTerm
			reduce(168), // //=, reduce: CmpTerm
			reduce(168), // %=, reduce: CmpTerm
			reduce(168), // **=, reduce: CmpTerm
			reduce(168), // ?, reduce: CmpTerm
			nil,         // :
			reduce(168), // ||, reduce: CmpTerm
			reduce(168), // ^^, reduce: CmpTerm
			reduce(168), // &&, reduce: CmpTerm
			reduce(168), // ??, reduce: CmpTerm
			reduce(168), // ???, reduce: CmpTerm
			reduce(168), // =~, reduce: CmpTerm
			reduce(168), // !=~, reduce: CmpTerm
			reduce(168), // ==, reduce: CmpTerm
			reduce(168), // !=, reduce: CmpTerm
			reduce(168), // <=>, reduce: CmpTerm
			reduce(168), // >=, reduce: CmpTerm
			reduce(168), // <, reduce: CmpTerm
			reduce(168), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3584
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(169), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(169), // >, reduce: CmpTerm
			nil,         // >>
			shift(2861), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(169), // ||=, reduce: CmpTerm
			reduce(169), // ^^=, reduce: CmpTerm
			reduce(169), // &&=, reduce: CmpTerm
			reduce(169), // ??=, reduce: CmpTerm
			reduce(169), // ???=, reduce: CmpTerm
			reduce(169), // |=, reduce: CmpTerm
			reduce(169), // &=, reduce: CmpTerm
			reduce(169), // ^=, reduce: CmpTerm
			reduce(169), // <<=, reduce: CmpTerm
			reduce(169), // >>=, reduce: CmpTerm
			reduce(169), // >>>=, reduce: CmpTerm
			reduce(169), // +=, reduce: CmpTerm
			reduce(169), // .=, reduce: CmpTerm
			reduce(169), // -=, reduce: CmpTerm
			reduce(169), // *=, reduce: CmpTerm
			reduce(169), // /=, reduce: CmpTerm
			reduce(169), // //=, reduce: CmpTerm
			reduce(169), // %=, reduce: CmpTerm
			reduce(169), // **=, reduce: CmpTerm
			reduce(169), // ?, reduce: CmpTerm
			nil,         // :
			reduce(169), // ||, reduce: CmpTerm
			reduce(169), // ^^, reduce: CmpTerm
			reduce(169), // &&, reduce: CmpTerm
			reduce(169), // ??, reduce: CmpTerm
			reduce(169), // ???, reduce: CmpTerm
			reduce(169), // =~, reduce: CmpTerm
			reduce(169), // !=~, reduce: CmpTerm
			reduce(169), // ==, reduce: CmpTerm
			reduce(169), // !=, reduce: CmpTerm
			reduce(169), // <=>, reduce: CmpTerm
			reduce(169), // >=, reduce: CmpTerm
			reduce(169), // <, reduce: CmpTerm
			reduce(169), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3585
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(170), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(170), // >, reduce: CmpTerm
			nil,         // >>
			shift(2861), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(170), // ||=, reduce: CmpTerm
			reduce(170), // ^^=, reduce: CmpTerm
			reduce(170), // &&=, reduce: CmpTerm
			reduce(170), // ??=, reduce: CmpTerm
			reduce(170), // ???=, reduce: CmpTerm
			reduce(170), // |=, reduce: CmpTerm
			reduce(170), // &=, reduce: CmpTerm
			reduce(170), // ^=, reduce: CmpTerm
			reduce(170), // <<=, reduce: CmpTerm
			reduce(170), // >>=, reduce: CmpTerm
			reduce(170), // >>>=, reduce: CmpTerm
			reduce(170), // +=, reduce: CmpTerm
			reduce(170), // .=, reduce: CmpTerm
			reduce(170), // -=, reduce: CmpTerm
			reduce(170), // *=, reduce: CmpTerm
			reduce(170), // /=, reduce: CmpTerm
			reduce(170), // //=, reduce: CmpTerm
			reduce(170), // %=, reduce: CmpTerm
			reduce(170), // **=, reduce: CmpTerm
			reduce(170), // ?, reduce: CmpTerm
			nil,         // :
			reduce(170), // ||, reduce: CmpTerm
			reduce(170), // ^^, reduce: CmpTerm
			reduce(170), // &&, reduce: CmpTerm
			reduce(170), // ??, reduce: CmpTerm
			reduce(170), // ???, reduce: CmpTerm
			reduce(170), // =~, reduce: CmpTerm
			reduce(170), // !=~, reduce: CmpTerm
			reduce(170), // ==, reduce: CmpTerm
			reduce(170), // !=, reduce: CmpTerm
			reduce(170), // <=>, reduce: CmpTerm
			reduce(170), // >=, reduce: CmpTerm
			reduce(170), // <, reduce: CmpTerm
			reduce(170), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3586
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(172), // =, reduce: BitwiseORTerm
			nil,         // unset
			nil,         // filter
			reduce(172), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(172), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(172), // ||=, reduce: BitwiseORTerm
			reduce(172), // ^^=, reduce: BitwiseORTerm
			reduce(172), // &&=, reduce: BitwiseORTerm
			reduce(172), // ??=, reduce: BitwiseORTerm
			reduce(172), // ???=, reduce: BitwiseORTerm
			reduce(172), // |=, reduce: BitwiseORTerm
			reduce(172), // &=, reduce: BitwiseORTerm
			reduce(172), // ^=, reduce: BitwiseORTerm
			reduce(172), // <<=, reduce: BitwiseORTerm
			reduce(172), // >>=, reduce: BitwiseORTerm
			reduce(172), // >>>=, reduce: BitwiseORTerm
			reduce(172), // +=, reduce: BitwiseORTerm
			reduce(172), // .=, reduce: BitwiseORTerm
			reduce(172), // -=, reduce: BitwiseORTerm
			reduce(172), // *=, reduce: BitwiseORTerm
			reduce(172), // /=, reduce: BitwiseORTerm
			reduce(172), // //=, reduce: BitwiseORTerm
			reduce(172), // %=, reduce: BitwiseORTerm
			reduce(172), // **=, reduce: BitwiseORTerm
			reduce(172), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(172), // ||, reduce: BitwiseORTerm
			reduce(172), // ^^, reduce: BitwiseORTerm
			reduce(172), // &&, reduce: BitwiseORTerm
			reduce(172), // ??, reduce: BitwiseORTerm
			reduce(172), // ???, reduce: BitwiseORTerm
			reduce(172), // =~, reduce: BitwiseORTerm
			reduce(172), // !=~, reduce: BitwiseORTerm
			reduce(172), // ==, reduce: BitwiseORTerm
			reduce(172), // !=, reduce: BitwiseORTerm
			reduce(172), // <=>, reduce: BitwiseORTerm
			reduce(172), // >=, reduce: BitwiseORTerm
			reduce(172), // <, reduce: BitwiseORTerm
			reduce(172), // <=, reduce: BitwiseORTerm
			shift(2862), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3587
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(174), // =, reduce: BitwiseXORTerm
			nil,         // unset
			nil,         // filter
			reduce(174), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(174), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(174), // ||=, reduce: BitwiseXORTerm
			reduce(174), // ^^=, reduce: BitwiseXORTerm
			reduce(174), // &&=, reduce: BitwiseXORTerm
			reduce(174), // ??=, reduce: BitwiseXORTerm
			reduce(174), // ???=, reduce: BitwiseXORTerm
			reduce(174), // |=, reduce: BitwiseXORTerm
			reduce(174), // &=, reduce: BitwiseXORTerm
			reduce(174), // ^=, reduce: BitwiseXORTerm
			reduce(174), // <<=, reduce: BitwiseXORTerm
			reduce(174), // >>=, reduce: BitwiseXORTerm
			reduce(174), // >>>=, reduce: BitwiseXORTerm
			reduce(174), // +=, reduce: BitwiseXORTerm
			reduce(174), // .=, reduce: BitwiseXORTerm
			reduce(174), // -=, reduce: BitwiseXORTerm
			reduce(174), // *=, reduce: BitwiseXORTerm
			reduce(174), // /=, reduce: BitwiseXORTerm
			reduce(174), // //=, reduce: BitwiseXORTerm
			reduce(174), // %=, reduce: BitwiseXORTerm
			reduce(174), // **=, reduce: BitwiseXORTerm
			reduce(174), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(174), // ||, reduce: BitwiseXORTerm
			reduce(174), // ^^, reduce: BitwiseXORTerm
			reduce(174), // &&, reduce: BitwiseXORTerm
			reduce(174), // ??, reduce: BitwiseXORTerm
			reduce(174), // ???, reduce: BitwiseXORTerm
			reduce(174), // =~, reduce: BitwiseXORTerm
			reduce(174), // !=~, reduce: BitwiseXORTerm
			reduce(174), // ==, reduce: BitwiseXORTerm
			reduce(174), // !=, reduce: BitwiseXORTerm
			reduce(174), // <=>, reduce: BitwiseXORTerm
			reduce(174), // >=, reduce: BitwiseXORTerm
			reduce(174), // <, reduce: BitwiseXORTerm
			reduce(174), // <=, reduce: BitwiseXORTerm
			reduce(174), // ^, reduce: BitwiseXORTerm
			shift(2863), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3588
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(176), // =, reduce: BitwiseANDTerm
			nil,         // unset
			nil,         // filter
			reduce(176), // >, reduce: BitwiseANDTerm
			shift(2864), // >>
			reduce(176), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(176), // ||=, reduce: BitwiseANDTerm
			reduce(176), // ^^=, reduce: BitwiseANDTerm
			reduce(176), // &&=, reduce: BitwiseANDTerm
			reduce(176), // ??=, reduce: BitwiseANDTerm
			reduce(176), // ???=, reduce: BitwiseANDTerm
			reduce(176), // |=, reduce: BitwiseANDTerm
			reduce(176), // &=, reduce: BitwiseANDTerm
			reduce(176), // ^=, reduce: BitwiseANDTerm
			reduce(176), // <<=, reduce: BitwiseANDTerm
			reduce(176), // >>=, reduce: BitwiseANDTerm
			reduce(176), // >>>=, reduce: BitwiseANDTerm
			reduce(176), // +=, reduce: BitwiseANDTerm
			reduce(176), // .=, reduce: BitwiseANDTerm
			reduce(176), // -=, reduce: BitwiseANDTerm
			reduce(176), // *=, reduce: BitwiseANDTerm
			reduce(176), // /=, reduce: BitwiseANDTerm
			reduce(176), // //=, reduce: BitwiseANDTerm
			reduce(176), // %=, reduce: BitwiseANDTerm
			reduce(176), // **=, reduce: BitwiseANDTerm
			reduce(176), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(176), // ||, reduce: BitwiseANDTerm
			reduce(176), // ^^, reduce: BitwiseANDTerm
			reduce(176), // &&, reduce: BitwiseANDTerm
			reduce(176), // ??, reduce: BitwiseANDTerm
			reduce(176), // ???, reduce: BitwiseANDTerm
			reduce(176), // =~, reduce: BitwiseANDTerm
			reduce(176), // !=~, reduce: BitwiseANDTerm
			reduce(176), // ==, reduce: BitwiseANDTerm
			reduce(176), // !=, reduce: BitwiseANDTerm
			reduce(176), // <=>, reduce: BitwiseANDTerm
			reduce(176), // >=, reduce: BitwiseANDTerm
			reduce(176), // <, reduce: BitwiseANDTerm
			reduce(176), // <=, reduce: BitwiseANDTerm
			reduce(176), // ^, reduce: BitwiseANDTerm
			reduce(176), // &, reduce: BitwiseANDTerm
			shift(2865), // <<
			shift(2866), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3589
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(179), // =, reduce: BitwiseShiftTerm
			nil,         // unset
			nil,         // filter
			reduce(179), // >, reduce: BitwiseShiftTerm
			reduce(179), // >>, reduce: BitwiseShiftTerm
			reduce(179), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(179), // ||=, reduce: BitwiseShiftTerm
			reduce(179), // ^^=, reduce: BitwiseShiftTerm
			reduce(179), // &&=, reduce: BitwiseShiftTerm
			reduce(179), // ??=, reduce: BitwiseShiftTerm
			reduce(179), // ???=, reduce: BitwiseShiftTerm
			reduce(179), // |=, reduce: BitwiseShiftTerm
			reduce(179), // &=, reduce: BitwiseShiftTerm
			reduce(179), // ^=, reduce: BitwiseShiftTerm
			reduce(179), // <<=, reduce: BitwiseShiftTerm
			reduce(179), // >>=, reduce: BitwiseShiftTerm
			reduce(179), // >>>=, reduce: BitwiseShiftTerm
			reduce(179), // +=, reduce: BitwiseShiftTerm
			reduce(179), // .=, reduce: BitwiseShiftTerm
			reduce(179), // -=, reduce: BitwiseShiftTerm
			reduce(179), // *=, reduce: BitwiseShiftTerm
			reduce(179), // /=, reduce: BitwiseShiftTerm
			reduce(179), // //=, reduce: BitwiseShiftTerm
			reduce(179), // %=, reduce: BitwiseShiftTerm
			reduce(179), // **=, reduce: BitwiseShiftTerm
			reduce(179), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(179), // ||, reduce: BitwiseShiftTerm
			reduce(179), // ^^, reduce: BitwiseShiftTerm
			reduce(179), // &&, reduce: BitwiseShiftTerm
			reduce(179), // ??, reduce: BitwiseShiftTerm
			reduce(179), // ???, reduce: BitwiseShiftTerm
			reduce(179), // =~, reduce: BitwiseShiftTerm
			reduce(179), // !=~, reduce: BitwiseShiftTerm
			reduce(179), // ==, reduce: BitwiseShiftTerm
			reduce(179), // !=, reduce: BitwiseShiftTerm
			reduce(179), // <=>, reduce: BitwiseShiftTerm
			reduce(179), // >=, reduce: BitwiseShiftTerm
			reduce(179), // <, reduce: BitwiseShiftTerm
			reduce(179), // <=, reduce: BitwiseShiftTerm
			reduce(179), // ^, reduce: BitwiseShiftTerm
			reduce(179), // &, reduce: BitwiseShiftTerm
			reduce(179), // <<, reduce: BitwiseShiftTerm
			reduce(179), // >>>, reduce: BitwiseShiftTerm
			shift(2867), // +
			shift(2868), // -
			shift(2869), // .+
			shift(2870), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3590
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(178), // =, reduce: BitwiseShiftTerm
			nil,         // unset
			nil,         // filter
			reduce(178), // >, reduce: BitwiseShiftTerm
			reduce(178), // >>, reduce: BitwiseShiftTerm
			reduce(178), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(178), // ||=, reduce: BitwiseShiftTerm
			reduce(178), // ^^=, reduce: BitwiseShiftTerm
			reduce(178), // &&=, reduce: BitwiseShiftTerm
			reduce(178), // ??=, reduce: BitwiseShiftTerm
			reduce(178), // ???=, reduce: BitwiseShiftTerm
			reduce(178), // |=, reduce: BitwiseShiftTerm
			reduce(178), // &=, reduce: BitwiseShiftTerm
			reduce(178), // ^=, reduce: BitwiseShiftTerm
			reduce(178), // <<=, reduce: BitwiseShiftTerm
			reduce(178), // >>=, reduce: BitwiseShiftTerm
			reduce(178), // >>>=, reduce: BitwiseShiftTerm
			reduce(178), // +=, reduce: BitwiseShiftTerm
			reduce(178), // .=, reduce: BitwiseShiftTerm
			reduce(178), // -=, reduce: BitwiseShiftTerm
			reduce(178), // *=, reduce: BitwiseShiftTerm
			reduce(178), // /=, reduce: BitwiseShiftTerm
			reduce(178), // //=, reduce: BitwiseShiftTerm
			reduce(178), // %=, reduce: BitwiseShiftTerm
			reduce(178), // **=, reduce: BitwiseShiftTerm
			reduce(178), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(178), // ||, reduce: BitwiseShiftTerm
			reduce(178), // ^^, reduce: BitwiseShiftTerm
			reduce(178), // &&, reduce: BitwiseShiftTerm
			reduce(178), // ??, reduce: BitwiseShiftTerm
			reduce(178), // ???, reduce: BitwiseShiftTerm
			reduce(178), // =~, reduce: BitwiseShiftTerm
			reduce(178), // !=~, reduce: BitwiseShiftTerm
			reduce(178), // ==, reduce: BitwiseShiftTerm
			reduce(178), // !=, reduce: BitwiseShiftTerm
			reduce(178), // <=>, reduce: BitwiseShiftTerm
			reduce(178), // >=, reduce: BitwiseShiftTerm
			reduce(178), // <, reduce: BitwiseShiftTerm
			reduce(178), // <=, reduce: BitwiseShiftTerm
			reduce(178), // ^, reduce: BitwiseShiftTerm
			reduce(178), // &, reduce: BitwiseShiftTerm
			reduce(178), // <<, reduce: BitwiseShiftTerm
			reduce(178), // >>>, reduce: BitwiseShiftTerm
			shift(2867), // +
			shift(2868), // -
			shift(2869), // .+
			shift(2870), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3591
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(180), // =, reduce: BitwiseShiftTerm
			nil,         // unset
			nil,         // filter
			reduce(180), // >, reduce: BitwiseShiftTerm
			reduce(180), // >>, reduce: BitwiseShiftTerm
			reduce(180), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(180), // ||=, reduce: BitwiseShiftTerm
			reduce(180), // ^^=, reduce: BitwiseShiftTerm
			reduce(180), // &&=, reduce: BitwiseShiftTerm
			reduce(180), // ??=, reduce: BitwiseShiftTerm
			reduce(180), // ???=, reduce: BitwiseShiftTerm
			reduce(180), // |=, reduce: BitwiseShiftTerm
			reduce(180), // &=, reduce: BitwiseShiftTerm
			reduce(180), // ^=, reduce: BitwiseShiftTerm
			reduce(180), // <<=, reduce: BitwiseShiftTerm
			reduce(180), // >>=, reduce: BitwiseShiftTerm
			reduce(180), // >>>=, reduce: BitwiseShiftTerm
			reduce(180), // +=, reduce: BitwiseShiftTerm
			reduce(180), // .=, reduce: BitwiseShiftTerm
			reduce(180), // -=, reduce: BitwiseShiftTerm
			reduce(180), // *=, reduce: BitwiseShiftTerm
			reduce(180), // /=, reduce: BitwiseShiftTerm
			reduce(180), // //=, reduce: BitwiseShiftTerm
			reduce(180), // %=, reduce: BitwiseShiftTerm
			reduce(180), // **=, reduce: BitwiseShiftTerm
			reduce(180), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(180), // ||, reduce: BitwiseShiftTerm
			reduce(180), // ^^, reduce: BitwiseShiftTerm
			reduce(180), // &&, reduce: BitwiseShiftTerm
			reduce(180), // ??, reduce: BitwiseShiftTerm
			reduce(180), // ???, reduce: BitwiseShiftTerm
			reduce(180), // =~, reduce: BitwiseShiftTerm
			reduce(180), // !=~, reduce: BitwiseShiftTerm
			reduce(180), // ==, reduce: BitwiseShiftTerm
			reduce(180), // !=, reduce: BitwiseShiftTerm
			reduce(180), // <=>, reduce: BitwiseShiftTerm
			reduce(180), // >=, reduce: BitwiseShiftTerm
			reduce(180), // <, reduce: BitwiseShiftTerm
			reduce(180), // <=, reduce: BitwiseShiftTerm
			reduce(180), // ^, reduce: BitwiseShiftTerm
			reduce(180), // &, reduce: BitwiseShiftTerm
			reduce(180), // <<, reduce: BitwiseShiftTerm
			reduce(180), // >>>, reduce: BitwiseShiftTerm
			shift(2867), // +
			shift(2868), // -
			shift(2869), // .+
			shift(2870), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3592
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(182), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(182), // >, reduce: AddsubdotTerm
			reduce(182), // >>, reduce: AddsubdotTerm
			reduce(182), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(182), // ||=, reduce: AddsubdotTerm
			reduce(182), // ^^=, reduce: AddsubdotTerm
			reduce(182), // &&=, reduce: AddsubdotTerm
			reduce(182), // ??=, reduce: AddsubdotTerm
			reduce(182), // ???=, reduce: AddsubdotTerm
			reduce(182), // |=, reduce: AddsubdotTerm
			reduce(182), // &=, reduce: AddsubdotTerm
			reduce(182), // ^=, reduce: AddsubdotTerm
			reduce(182), // <<=, reduce: AddsubdotTerm
			reduce(182), // >>=, reduce: AddsubdotTerm
			reduce(182), // >>>=, reduce: AddsubdotTerm
			reduce(182), // +=, reduce: AddsubdotTerm
			reduce(182), // .=, reduce: AddsubdotTerm
			reduce(182), // -=, reduce: AddsubdotTerm
			reduce(182), // *=, reduce: AddsubdotTerm
			reduce(182), // /=, reduce: AddsubdotTerm
			reduce(182), // //=, reduce: AddsubdotTerm
			reduce(182), // %=, reduce: AddsubdotTerm
			reduce(182), // **=, reduce: AddsubdotTerm
			reduce(182), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(182), // ||, reduce: AddsubdotTerm
			reduce(182), // ^^, reduce: AddsubdotTerm
			reduce(182), // &&, reduce: AddsubdotTerm
			reduce(182), // ??, reduce: AddsubdotTerm
			reduce(182), // ???, reduce: AddsubdotTerm
			reduce(182), // =~, reduce: AddsubdotTerm
			reduce(182), // !=~, reduce: AddsubdotTerm
			reduce(182), // ==, reduce: AddsubdotTerm
			reduce(182), // !=, reduce: AddsubdotTerm
			reduce(182), // <=>, reduce: AddsubdotTerm
			reduce(182), // >=, reduce: AddsubdotTerm
			reduce(182), // <, reduce: AddsubdotTerm
			reduce(182), // <=, reduce: AddsubdotTerm
			reduce(182), // ^, reduce: AddsubdotTerm
			reduce(182), // &, reduce: AddsubdotTerm
			reduce(182), // <<, reduce: AddsubdotTerm
			reduce(182), // >>>, reduce: AddsubdotTerm
			reduce(182), // +, reduce: AddsubdotTerm
			reduce(182), // -, reduce: AddsubdotTerm
			reduce(182), // .+, reduce: AddsubdotTerm
			reduce(182), // .-, reduce: AddsubdotTerm
			shift(2875), // *
			shift(2876), // /
			shift(2877), // //
			shift(2878), // %
			shift(2879), // .*
			shift(2880), // ./
			shift(2881), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3593
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(183), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(183), // >, reduce: AddsubdotTerm
			reduce(183), // >>, reduce: AddsubdotTerm
			reduce(183), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(183), // ||=, reduce: AddsubdotTerm
			reduce(183), // ^^=, reduce: AddsubdotTerm
			reduce(183), // &&=, reduce: AddsubdotTerm
			reduce(183), // ??=, reduce: AddsubdotTerm
			reduce(183), // ???=, reduce: AddsubdotTerm
			reduce(183), // |=, reduce: AddsubdotTerm
			reduce(183), // &=, reduce: AddsubdotTerm
			reduce(183), // ^=, reduce: AddsubdotTerm
			reduce(183), // <<=, reduce: AddsubdotTerm
			reduce(183), // >>=, reduce: AddsubdotTerm
			reduce(183), // >>>=, reduce: AddsubdotTerm
			reduce(183), // +=, reduce: AddsubdotTerm
			reduce(183), // .=, reduce: AddsubdotTerm
			reduce(183), // -=, reduce: AddsubdotTerm
			reduce(183), // *=, reduce: AddsubdotTerm
			reduce(183), // /=, reduce: AddsubdotTerm
			reduce(183), // //=, reduce: AddsubdotTerm
			reduce(183), // %=, reduce: AddsubdotTerm
			reduce(183), // **=, reduce: AddsubdotTerm
			reduce(183), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(183), // ||, reduce: AddsubdotTerm
			reduce(183), // ^^, reduce: AddsubdotTerm
			reduce(183), // &&, reduce: AddsubdotTerm
			reduce(183), // ??, reduce: AddsubdotTerm
			reduce(183), // ???, reduce: AddsubdotTerm
			reduce(183), // =~, reduce: AddsubdotTerm
			reduce(183), // !=~, reduce: AddsubdotTerm
			reduce(183), // ==, reduce: AddsubdotTerm
			reduce(183), // !=, reduce: AddsubdotTerm
			reduce(183), // <=>, reduce: AddsubdotTerm
			reduce(183), // >=, reduce: AddsubdotTerm
			reduce(183), // <, reduce: AddsubdotTerm
			reduce(183), // <=, reduce: AddsubdotTerm
			reduce(183), // ^, reduce: AddsubdotTerm
			reduce(183), // &, reduce: AddsubdotTerm
			reduce(183), // <<, reduce: AddsubdotTerm
			reduce(183), // >>>, reduce: AddsubdotTerm
			reduce(183), // +, reduce: AddsubdotTerm
			reduce(183), // -, reduce: AddsubdotTerm
			reduce(183), // .+, reduce: AddsubdotTerm
			reduce(183), // .-, reduce: AddsubdotTerm
			shift(2875), // *
			shift(2876), // /
			shift(2877), // //
			shift(2878), // %
			shift(2879), // .*
			shift(2880), // ./
			shift(2881), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3594
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(184), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(184), // >, reduce: AddsubdotTerm
			reduce(184), // >>, reduce: AddsubdotTerm
			reduce(184), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(184), // ||=, reduce: AddsubdotTerm
			reduce(184), // ^^=, reduce: AddsubdotTerm
			reduce(184), // &&=, reduce: AddsubdotTerm
			reduce(184), // ??=, reduce: AddsubdotTerm
			reduce(184), // ???=, reduce: AddsubdotTerm
			reduce(184), // |=, reduce: AddsubdotTerm
			reduce(184), // &=, reduce: AddsubdotTerm
			reduce(184), // ^=, reduce: AddsubdotTerm
			reduce(184), // <<=, reduce: AddsubdotTerm
			reduce(184), // >>=, reduce: AddsubdotTerm
			reduce(184), // >>>=, reduce: AddsubdotTerm
			reduce(184), // +=, reduce: AddsubdotTerm
			reduce(184), // .=, reduce: AddsubdotTerm
			reduce(184), // -=, reduce: AddsubdotTerm
			reduce(184), // *=, reduce: AddsubdotTerm
			reduce(184), // /=, reduce: AddsubdotTerm
			reduce(184), // //=, reduce: AddsubdotTerm
			reduce(184), // %=, reduce: AddsubdotTerm
			reduce(184), // **=, reduce: AddsubdotTerm
			reduce(184), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(184), // ||, reduce: AddsubdotTerm
			reduce(184), // ^^, reduce: AddsubdotTerm
			reduce(184), // &&, reduce: AddsubdotTerm
			reduce(184), // ??, reduce: AddsubdotTerm
			reduce(184), // ???, reduce: AddsubdotTerm
			reduce(184), // =~, reduce: AddsubdotTerm
			reduce(184), // !=~, reduce: AddsubdotTerm
			reduce(184), // ==, reduce: AddsubdotTerm
			reduce(184), // !=, reduce: AddsubdotTerm
			reduce(184), // <=>, reduce: AddsubdotTerm
			reduce(184), // >=, reduce: AddsubdotTerm
			reduce(184), // <, reduce: AddsubdotTerm
			reduce(184), // <=, reduce: AddsubdotTerm
			reduce(184), // ^, reduce: AddsubdotTerm
			reduce(184), // &, reduce: AddsubdotTerm
			reduce(184), // <<, reduce: AddsubdotTerm
			reduce(184), // >>>, reduce: AddsubdotTerm
			reduce(184), // +, reduce: AddsubdotTerm
			reduce(184), // -, reduce: AddsubdotTerm
			reduce(184), // .+, reduce: AddsubdotTerm
			reduce(184), // .-, reduce: AddsubdotTerm
			shift(2875), // *
			shift(2876), // /
			shift(2877), // //
			shift(2878), // %
			shift(2879), // .*
			shift(2880), // ./
			shift(2881), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3595
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(185), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(185), // >, reduce: AddsubdotTerm
			reduce(185), // >>, reduce: AddsubdotTerm
			reduce(185), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(185), // ||=, reduce: AddsubdotTerm
			reduce(185), // ^^=, reduce: AddsubdotTerm
			reduce(185), // &&=, reduce: AddsubdotTerm
			reduce(185), // ??=, reduce: AddsubdotTerm
			reduce(185), // ???=, reduce: AddsubdotTerm
			reduce(185), // |=, reduce: AddsubdotTerm
			reduce(185), // &=, reduce: AddsubdotTerm
			reduce(185), // ^=, reduce: AddsubdotTerm
			reduce(185), // <<=, reduce: AddsubdotTerm
			reduce(185), // >>=, reduce: AddsubdotTerm
			reduce(185), // >>>=, reduce: AddsubdotTerm
			reduce(185), // +=, reduce: AddsubdotTerm
			reduce(185), // .=, reduce: AddsubdotTerm
			reduce(185), // -=, reduce: AddsubdotTerm
			reduce(185), // *=, reduce: AddsubdotTerm
			reduce(185), // /=, reduce: AddsubdotTerm
			reduce(185), // //=, reduce: AddsubdotTerm
			reduce(185), // %=, reduce: AddsubdotTerm
			reduce(185), // **=, reduce: AddsubdotTerm
			reduce(185), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(185), // ||, reduce: AddsubdotTerm
			reduce(185), // ^^, reduce: AddsubdotTerm
			reduce(185), // &&, reduce: AddsubdotTerm
			reduce(185), // ??, reduce: AddsubdotTerm
			reduce(185), // ???, reduce: AddsubdotTerm
			reduce(185), // =~, reduce: AddsubdotTerm
			reduce(185), // !=~, reduce: AddsubdotTerm
			reduce(185), // ==, reduce: AddsubdotTerm
			reduce(185), // !=, reduce: AddsubdotTerm
			reduce(185), // <=>, reduce: AddsubdotTerm
			reduce(185), // >=, reduce: AddsubdotTerm
			reduce(185), // <, reduce: AddsubdotTerm
			reduce(185), // <=, reduce: AddsubdotTerm
			reduce(185), // ^, reduce: AddsubdotTerm
			reduce(185), // &, reduce: AddsubdotTerm
			reduce(185), // <<, reduce: AddsubdotTerm
			reduce(185), // >>>, reduce: AddsubdotTerm
			reduce(185), // +, reduce: AddsubdotTerm
			reduce(185), // -, reduce: AddsubdotTerm
			reduce(185), // .+, reduce: AddsubdotTerm
			reduce(185), // .-, reduce: AddsubdotTerm
			shift(2875), // *
			shift(2876), // /
			shift(2877), // //
			shift(2878), // %
			shift(2879), // .*
			shift(2880), // ./
			shift(2881), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3596
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(187), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(187), // >, reduce: MuldivTerm
			reduce(187), // >>, reduce: MuldivTerm
			reduce(187), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(187), // ||=, reduce: MuldivTerm
			reduce(187), // ^^=, reduce: MuldivTerm
			reduce(187), // &&=, reduce: MuldivTerm
			reduce(187), // ??=, reduce: MuldivTerm
			reduce(187), // ???=, reduce: MuldivTerm
			reduce(187), // |=, reduce: MuldivTerm
			reduce(187), // &=, reduce: MuldivTerm
			reduce(187), // ^=, reduce: MuldivTerm
			reduce(187), // <<=, reduce: MuldivTerm
			reduce(187), // >>=, reduce: MuldivTerm
			reduce(187), // >>>=, reduce: MuldivTerm
			reduce(187), // +=, reduce: MuldivTerm
			reduce(187), // .=, reduce: MuldivTerm
			reduce(187), // -=, reduce: MuldivTerm
			reduce(187), // *=, reduce: MuldivTerm
			reduce(187), // /=, reduce: MuldivTerm
			reduce(187), // //=, reduce: MuldivTerm
			reduce(187), // %=, reduce: MuldivTerm
			reduce(187), // **=, reduce: MuldivTerm
			reduce(187), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(187), // ||, reduce: MuldivTerm
			reduce(187), // ^^, reduce: MuldivTerm
			reduce(187), // &&, reduce: MuldivTerm
			reduce(187), // ??, reduce: MuldivTerm
			reduce(187), // ???, reduce: MuldivTerm
			reduce(187), // =~, reduce: MuldivTerm
			reduce(187), // !=~, reduce: MuldivTerm
			reduce(187), // ==, reduce: MuldivTerm
			reduce(187), // !=, reduce: MuldivTerm
			reduce(187), // <=>, reduce: MuldivTerm
			reduce(187), // >=, reduce: MuldivTerm
			reduce(187), // <, reduce: MuldivTerm
			reduce(187), // <=, reduce: MuldivTerm
			reduce(187), // ^, reduce: MuldivTerm
			reduce(187), // &, reduce: MuldivTerm
			reduce(187), // <<, reduce: MuldivTerm
			reduce(187), // >>>, reduce: MuldivTerm
			reduce(187), // +, reduce: MuldivTerm
			reduce(187), // -, reduce: MuldivTerm
			reduce(187), // .+, reduce: MuldivTerm
			reduce(187), // .-, reduce: MuldivTerm
			reduce(187), // *, reduce: MuldivTerm
			reduce(187), // /, reduce: MuldivTerm
			reduce(187), // //, reduce: MuldivTerm
			reduce(187), // %, reduce: MuldivTerm
			reduce(187), // .*, reduce: MuldivTerm
			reduce(187), // ./, reduce: MuldivTerm
			reduce(187), // .//, reduce: MuldivTerm
			shift(2885), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3597
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(188), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(188), // >, reduce: MuldivTerm
			reduce(188), // >>, reduce: MuldivTerm
			reduce(188), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(188), // ||=, reduce: MuldivTerm
			reduce(188), // ^^=, reduce: MuldivTerm
			reduce(188), // &&=, reduce: MuldivTerm
			reduce(188), // ??=, reduce: MuldivTerm
			reduce(188), // ???=, reduce: MuldivTerm
			reduce(188), // |=, reduce: MuldivTerm
			reduce(188), // &=, reduce: MuldivTerm
			reduce(188), // ^=, reduce: MuldivTerm
			reduce(188), // <<=, reduce: MuldivTerm
			reduce(188), // >>=, reduce: MuldivTerm
			reduce(188), // >>>=, reduce: MuldivTerm
			reduce(188), // +=, reduce: MuldivTerm
			reduce(188), // .=, reduce: MuldivTerm
			reduce(188), // -=, reduce: MuldivTerm
			reduce(188), // *=, reduce: MuldivTerm
			reduce(188), // /=, reduce: MuldivTerm
			reduce(188), // //=, reduce: MuldivTerm
			reduce(188), // %=, reduce: MuldivTerm
			reduce(188), // **=, reduce: MuldivTerm
			reduce(188), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(188), // ||, reduce: MuldivTerm
			reduce(188), // ^^, reduce: MuldivTerm
			reduce(188), // &&, reduce: MuldivTerm
			reduce(188), // ??, reduce: MuldivTerm
			reduce(188), // ???, reduce: MuldivTerm
			reduce(188), // =~, reduce: MuldivTerm
			reduce(188), // !=~, reduce: MuldivTerm
			reduce(188), // ==, reduce: MuldivTerm
			reduce(188), // !=, reduce: MuldivTerm
			reduce(188), // <=>, reduce: MuldivTerm
			reduce(188), // >=, reduce: MuldivTerm
			reduce(188), // <, reduce: MuldivTerm
			reduce(188), // <=, reduce: MuldivTerm
			reduce(188), // ^, reduce: MuldivTerm
			reduce(188), // &, reduce: MuldivTerm
			reduce(188), // <<, reduce: MuldivTerm
			reduce(188), // >>>, reduce: MuldivTerm
			reduce(188), // +, reduce: MuldivTerm
			reduce(188), // -, reduce: MuldivTerm
			reduce(188), // .+, reduce: MuldivTerm
			reduce(188), // .-, reduce: MuldivTerm
			reduce(188), // *, reduce: MuldivTerm
			reduce(188), // /, reduce: MuldivTerm
			reduce(188), // //, reduce: MuldivTerm
			reduce(188), // %, reduce: MuldivTerm
			reduce(188), // .*, reduce: MuldivTerm
			reduce(188), // ./, reduce: MuldivTerm
			reduce(188), // .//, reduce: MuldivTerm
			shift(2885), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3598
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(189), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(189), // >, reduce: MuldivTerm
			reduce(189), // >>, reduce: MuldivTerm
			reduce(189), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(189), // ||=, reduce: MuldivTerm
			reduce(189), // ^^=, reduce: MuldivTerm
			reduce(189), // &&=, reduce: MuldivTerm
			reduce(189), // ??=, reduce: MuldivTerm
			reduce(189), // ???=, reduce: MuldivTerm
			reduce(189), // |=, reduce: MuldivTerm
			reduce(189), // &=, reduce: MuldivTerm
			reduce(189), // ^=, reduce: MuldivTerm
			reduce(189), // <<=, reduce: MuldivTerm
			reduce(189), // >>=, reduce: MuldivTerm
			reduce(189), // >>>=, reduce: MuldivTerm
			reduce(189), // +=, reduce: MuldivTerm
			reduce(189), // .=, reduce: MuldivTerm
			reduce(189), // -=, reduce: MuldivTerm
			reduce(189), // *=, reduce: MuldivTerm
			reduce(189), // /=, reduce: MuldivTerm
			reduce(189), // //=, reduce: MuldivTerm
			reduce(189), // %=, reduce: MuldivTerm
			reduce(189), // **=, reduce: MuldivTerm
			reduce(189), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(189), // ||, reduce: MuldivTerm
			reduce(189), // ^^, reduce: MuldivTerm
			reduce(189), // &&, reduce: MuldivTerm
			reduce(189), // ??, reduce: MuldivTerm
			reduce(189), // ???, reduce: MuldivTerm
			reduce(189), // =~, reduce: MuldivTerm
			reduce(189), // !=~, reduce: MuldivTerm
			reduce(189), // ==, reduce: MuldivTerm
			reduce(189), // !=, reduce: MuldivTerm
			reduce(189), // <=>, reduce: MuldivTerm
			reduce(189), // >=, reduce: MuldivTerm
			reduce(189), // <, reduce: MuldivTerm
			reduce(189), // <=, reduce: MuldivTerm
			reduce(189), // ^, reduce: MuldivTerm
			reduce(189), // &, reduce: MuldivTerm
			reduce(189), // <<, reduce: MuldivTerm
			reduce(189), // >>>, reduce: MuldivTerm
			reduce(189), // +, reduce: MuldivTerm
			reduce(189), // -, reduce: MuldivTerm
			reduce(189), // .+, reduce: MuldivTerm
			reduce(189), // .-, reduce: MuldivTerm
			reduce(189), // *, reduce: MuldivTerm
			reduce(189), // /, reduce: MuldivTerm
			reduce(189), // //, reduce: MuldivTerm
			reduce(189), // %, reduce: MuldivTerm
			reduce(189), // .*, reduce: MuldivTerm
			reduce(189), // ./, reduce: MuldivTerm
			reduce(189), // .//, reduce: MuldivTerm
			shift(2885), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3599
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(190), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(190), // >, reduce: MuldivTerm
			reduce(190), // >>, reduce: MuldivTerm
			reduce(190), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(190), // ||=, reduce: MuldivTerm
			reduce(190), // ^^=, reduce: MuldivTerm
			reduce(190), // &&=, reduce: MuldivTerm
			reduce(190), // ??=, reduce: MuldivTerm
			reduce(190), // ???=, reduce: MuldivTerm
			reduce(190), // |=, reduce: MuldivTerm
			reduce(190), // &=, reduce: MuldivTerm
			reduce(190), // ^=, reduce: MuldivTerm
			reduce(190), // <<=, reduce: MuldivTerm
			reduce(190), // >>=, reduce: MuldivTerm
			reduce(190), // >>>=, reduce: MuldivTerm
			reduce(190), // +=, reduce: MuldivTerm
			reduce(190), // .=, reduce: MuldivTerm
			reduce(190), // -=, reduce: MuldivTerm
			reduce(190), // *=, reduce: MuldivTerm
			reduce(190), // /=, reduce: MuldivTerm
			reduce(190), // //=, reduce: MuldivTerm
			reduce(190), // %=, reduce: MuldivTerm
			reduce(190), // **=, reduce: MuldivTerm
			reduce(190), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(190), // ||, reduce: MuldivTerm
			reduce(190), // ^^, reduce: MuldivTerm
			reduce(190), // &&, reduce: MuldivTerm
			reduce(190), // ??, reduce: MuldivTerm
			reduce(190), // ???, reduce: MuldivTerm
			reduce(190), // =~, reduce: MuldivTerm
			reduce(190), // !=~, reduce: MuldivTerm
			reduce(190), // ==, reduce: MuldivTerm
			reduce(190), // !=, reduce: MuldivTerm
			reduce(190), // <=>, reduce: MuldivTerm
			reduce(190), // >=, reduce: MuldivTerm
			reduce(190), // <, reduce: MuldivTerm
			reduce(190), // <=, reduce: MuldivTerm
			reduce(190), // ^, reduce: MuldivTerm
			reduce(190), // &, reduce: MuldivTerm
			reduce(190), // <<, reduce: MuldivTerm
			reduce(190), // >>>, reduce: MuldivTerm
			reduce(190), // +, reduce: MuldivTerm
			reduce(190), // -, reduce: MuldivTerm
			reduce(190), // .+, reduce: MuldivTerm
			reduce(190), // .-, reduce: MuldivTerm
			reduce(190), // *, reduce: MuldivTerm
			reduce(190), // /, reduce: MuldivTerm
			reduce(190), // //, reduce: MuldivTerm
			reduce(190), // %, reduce: MuldivTerm
			reduce(190), // .*, reduce: MuldivTerm
			reduce(190), // ./, reduce: MuldivTerm
			reduce(190), // .//, reduce: MuldivTerm
			shift(2885), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3600
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(191), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(191), // >, reduce: MuldivTerm
			reduce(191), // >>, reduce: MuldivTerm
			reduce(191), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(191), // ||=, reduce: MuldivTerm
			reduce(191), // ^^=, reduce: MuldivTerm
			reduce(191), // &&=, reduce: MuldivTerm
			reduce(191), // ??=, reduce: MuldivTerm
			reduce(191), // ???=, reduce: MuldivTerm
			reduce(191), // |=, reduce: MuldivTerm
			reduce(191), // &=, reduce: MuldivTerm
			reduce(191), // ^=, reduce: MuldivTerm
			reduce(191), // <<=, reduce: MuldivTerm
			reduce(191), // >>=, reduce: MuldivTerm
			reduce(191), // >>>=, reduce: MuldivTerm
			reduce(191), // +=, reduce: MuldivTerm
			reduce(191), // .=, reduce: MuldivTerm
			reduce(191), // -=, reduce: MuldivTerm
			reduce(191), // *=, reduce: MuldivTerm
			reduce(191), // /=, reduce: MuldivTerm
			reduce(191), // //=, reduce: MuldivTerm
			reduce(191), // %=, reduce: MuldivTerm
			reduce(191), // **=, reduce: MuldivTerm
			reduce(191), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(191), // ||, reduce: MuldivTerm
			reduce(191), // ^^, reduce: MuldivTerm
			reduce(191), // &&, reduce: MuldivTerm
			reduce(191), // ??, reduce: MuldivTerm
			reduce(191), // ???, reduce: MuldivTerm
			reduce(191), // =~, reduce: MuldivTerm
			reduce(191), // !=~, reduce: MuldivTerm
			reduce(191), // ==, reduce: MuldivTerm
			reduce(191), // !=, reduce: MuldivTerm
			reduce(191), // <=>, reduce: MuldivTerm
			reduce(191), // >=, reduce: MuldivTerm
			reduce(191), // <, reduce: MuldivTerm
			reduce(191), // <=, reduce: MuldivTerm
			reduce(191), // ^, reduce: MuldivTerm
			reduce(191), // &, reduce: MuldivTerm
			reduce(191), // <<, reduce: MuldivTerm
			reduce(191), // >>>, reduce: MuldivTerm
			reduce(191), // +, reduce: MuldivTerm
			reduce(191), // -, reduce: MuldivTerm
			reduce(191), // .+, reduce: MuldivTerm
			reduce(191), // .-, reduce: MuldivTerm
			reduce(191), // *, reduce: MuldivTerm
			reduce(191), // /, reduce: MuldivTerm
			reduce(191), // //, reduce: MuldivTerm
			reduce(191), // %, reduce: MuldivTerm
			reduce(191), // .*, reduce: MuldivTerm
			reduce(191), // ./, reduce: MuldivTerm
			reduce(191), // .//, reduce: MuldivTerm
			shift(2885), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3601
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(192), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(192), // >, reduce: MuldivTerm
			reduce(192), // >>, reduce: MuldivTerm
			reduce(192), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(192), // ||=, reduce: MuldivTerm
			reduce(192), // ^^=, reduce: MuldivTerm
			reduce(192), // &&=, reduce: MuldivTerm
			reduce(192), // ??=, reduce: MuldivTerm
			reduce(192), // ???=, reduce: MuldivTerm
			reduce(192), // |=, reduce: MuldivTerm
			reduce(192), // &=, reduce: MuldivTerm
			reduce(192), // ^=, reduce: MuldivTerm
			reduce(192), // <<=, reduce: MuldivTerm
			reduce(192), // >>=, reduce: MuldivTerm
			reduce(192), // >>>=, reduce: MuldivTerm
			reduce(192), // +=, reduce: MuldivTerm
			reduce(192), // .=, reduce: MuldivTerm
			reduce(192), // -=, reduce: MuldivTerm
			reduce(192), // *=, reduce: MuldivTerm
			reduce(192), // /=, reduce: MuldivTerm
			reduce(192), // //=, reduce: MuldivTerm
			reduce(192), // %=, reduce: MuldivTerm
			reduce(192), // **=, reduce: MuldivTerm
			reduce(192), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(192), // ||, reduce: MuldivTerm
			reduce(192), // ^^, reduce: MuldivTerm
			reduce(192), // &&, reduce: MuldivTerm
			reduce(192), // ??, reduce: MuldivTerm
			reduce(192), // ???, reduce: MuldivTerm
			reduce(192), // =~, reduce: MuldivTerm
			reduce(192), // !=~, reduce: MuldivTerm
			reduce(192), // ==, reduce: MuldivTerm
			reduce(192), // !=, reduce: MuldivTerm
			reduce(192), // <=>, reduce: MuldivTerm
			reduce(192), // >=, reduce: MuldivTerm
			reduce(192), // <, reduce: MuldivTerm
			reduce(192), // <=, reduce: MuldivTerm
			reduce(192), // ^, reduce: MuldivTerm
			reduce(192), // &, reduce: MuldivTerm
			reduce(192), // <<, reduce: MuldivTerm
			reduce(192), // >>>, reduce: MuldivTerm
			reduce(192), // +, reduce: MuldivTerm
			reduce(192), // -, reduce: MuldivTerm
			reduce(192), // .+, reduce: MuldivTerm
			reduce(192), // .-, reduce: MuldivTerm
			reduce(192), // *, reduce: MuldivTerm
			reduce(192), // /, reduce: MuldivTerm
			reduce(192), // //, reduce: MuldivTerm
			reduce(192), // %, reduce: MuldivTerm
			reduce(192), // .*, reduce: MuldivTerm
			reduce(192), // ./, reduce: MuldivTerm
			reduce(192), // .//, reduce: MuldivTerm
			shift(2885), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3602
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(193), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(193), // >, reduce: MuldivTerm
			reduce(193), // >>, reduce: MuldivTerm
			reduce(193), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(193), // ||=, reduce: MuldivTerm
			reduce(193), // ^^=, reduce: MuldivTerm
			reduce(193), // &&=, reduce: MuldivTerm
			reduce(193), // ??=, reduce: MuldivTerm
			reduce(193), // ???=, reduce: MuldivTerm
			reduce(193), // |=, reduce: MuldivTerm
			reduce(193), // &=, reduce: MuldivTerm
			reduce(193), // ^=, reduce: MuldivTerm
			reduce(193), // <<=, reduce: MuldivTerm
			reduce(193), // >>=, reduce: MuldivTerm
			reduce(193), // >>>=, reduce: MuldivTerm
			reduce(193), // +=, reduce: MuldivTerm
			reduce(193), // .=, reduce: MuldivTerm
			reduce(193), // -=, reduce: MuldivTerm
			reduce(193), // *=, reduce: MuldivTerm
			reduce(193), // /=, reduce: MuldivTerm
			reduce(193), // //=, reduce: MuldivTerm
			reduce(193), // %=, reduce: MuldivTerm
			reduce(193), // **=, reduce: MuldivTerm
			reduce(193), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(193), // ||, reduce: MuldivTerm
			reduce(193), // ^^, reduce: MuldivTerm
			reduce(193), // &&, reduce: MuldivTerm
			reduce(193), // ??, reduce: MuldivTerm
			reduce(193), // ???, reduce: MuldivTerm
			reduce(193), // =~, reduce: MuldivTerm
			reduce(193), // !=~, reduce: MuldivTerm
			reduce(193), // ==, reduce: MuldivTerm
			reduce(193), // !=, reduce: MuldivTerm
			reduce(193), // <=>, reduce: MuldivTerm
			reduce(193), // >=, reduce: MuldivTerm
			reduce(193), // <, reduce: MuldivTerm
			reduce(193), // <=, reduce: MuldivTerm
			reduce(193), // ^, reduce: MuldivTerm
			reduce(193), // &, reduce: MuldivTerm
			reduce(193), // <<, reduce: MuldivTerm
			reduce(193), // >>>, reduce: MuldivTerm
			reduce(193), // +, reduce: MuldivTerm
			reduce(193), // -, reduce: MuldivTerm
			reduce(193), // .+, reduce: MuldivTerm
			reduce(193), // .-, reduce: MuldivTerm
			reduce(193), // *, reduce: MuldivTerm
			reduce(193), // /, reduce: MuldivTerm
			reduce(193), // //, reduce: MuldivTerm
			reduce(193), // %, reduce: MuldivTerm
			reduce(193), // .*, reduce: MuldivTerm
			reduce(193), // ./, reduce: MuldivTerm
			reduce(193), // .//, reduce: MuldivTerm
			shift(2885), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3603
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(195), // =, reduce: DotTerm
			nil,         // unset
			nil,         // filter
			reduce(195), // >, reduce: DotTerm
			reduce(195), // >>, reduce: DotTerm
			reduce(195), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(195), // ||=, reduce: DotTerm
			reduce(195), // ^^=, reduce: DotTerm
			reduce(195), // &&=, reduce: DotTerm
			reduce(195), // ??=, reduce: DotTerm
			reduce(195), // ???=, reduce: DotTerm
			reduce(195), // |=, reduce: DotTerm
			reduce(195), // &=, reduce: DotTerm
			reduce(195), // ^=, reduce: DotTerm
			reduce(195), // <<=, reduce: DotTerm
			reduce(195), // >>=, reduce: DotTerm
			reduce(195), // >>>=, reduce: DotTerm
			reduce(195), // +=, reduce: DotTerm
			reduce(195), // .=, reduce: DotTerm
			reduce(195), // -=, reduce: DotTerm
			reduce(195), // *=, reduce: DotTerm
			reduce(195), // /=, reduce: DotTerm
			reduce(195), // //=, reduce: DotTerm
			reduce(195), // %=, reduce: DotTerm
			reduce(195), // **=, reduce: DotTerm
			reduce(195), // ?, reduce: DotTerm
			nil,         // :
			reduce(195), // ||, reduce: DotTerm
			reduce(195), // ^^, reduce: DotTerm
			reduce(195), // &&, reduce: DotTerm
			reduce(195), // ??, reduce: DotTerm
			reduce(195), // ???, reduce: DotTerm
			reduce(195), // =~, reduce: DotTerm
			reduce(195), // !=~, reduce: DotTerm
			reduce(195), // ==, reduce: DotTerm
			reduce(195), // !=, reduce: DotTerm
			reduce(195), // <=>, reduce: DotTerm
			reduce(195), // >=, reduce: DotTerm
			reduce(195), // <, reduce: DotTerm
			reduce(195), // <=, reduce: DotTerm
			reduce(195), // ^, reduce: DotTerm
			reduce(195), // &, reduce: DotTerm
			reduce(195), // <<, reduce: DotTerm
			reduce(195), // >>>, reduce: DotTerm
			reduce(195), // +, reduce: DotTerm
			reduce(195), // -, reduce: DotTerm
			reduce(195), // .+, reduce: DotTerm
			reduce(195), // .-, reduce: DotTerm
			reduce(195), // *, reduce: DotTerm
			reduce(195), // /, reduce: DotTerm
			reduce(195), // //, reduce: DotTerm
			reduce(195), // %, reduce: DotTerm
			reduce(195), // .*, reduce: DotTerm
			reduce(195), // ./, reduce: DotTerm
			reduce(195), // .//, reduce: DotTerm
			reduce(195), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3604
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3605
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3606
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(204), // =, reduce: PowTerm
			nil,         // unset
			nil,         // filter
			reduce(204), // >, reduce: PowTerm
			reduce(204), // >>, reduce: PowTerm
			reduce(204), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(204), // ||=, reduce: PowTerm
			reduce(204), // ^^=, reduce: PowTerm
			reduce(204), // &&=, reduce: PowTerm
			reduce(204), // ??=, reduce: PowTerm
			reduce(204), // ???=, reduce: PowTerm
			reduce(204), // |=, reduce: PowTerm
			reduce(204), // &=, reduce: PowTerm
			reduce(204), // ^=, reduce: PowTerm
			reduce(204), // <<=, reduce: PowTerm
			reduce(204), // >>=, reduce: PowTerm
			reduce(204), // >>>=, reduce: PowTerm
			reduce(204), // +=, reduce: PowTerm
			reduce(204), // .=, reduce: PowTerm
			reduce(204), // -=, reduce: PowTerm
			reduce(204), // *=, reduce: PowTerm
			reduce(204), // /=, reduce: PowTerm
			reduce(204), // //=, reduce: PowTerm
			reduce(204), // %=, reduce: PowTerm
			reduce(204), // **=, reduce: PowTerm
			reduce(204), // ?, reduce: PowTerm
			nil,         // :
			reduce(204), // ||, reduce: PowTerm
			reduce(204), // ^^, reduce: PowTerm
			reduce(204), // &&, reduce: PowTerm
			reduce(204), // ??, reduce: PowTerm
			reduce(204), // ???, reduce: PowTerm
			reduce(204), // =~, reduce: PowTerm
			reduce(204), // !=~, reduce: PowTerm
			reduce(204), // ==, reduce: PowTerm
			reduce(204), // !=, reduce: PowTerm
			reduce(204), // <=>, reduce: PowTerm
			reduce(204), // >=, reduce: PowTerm
			reduce(204), // <, reduce: PowTerm
			reduce(204), // <=, reduce: PowTerm
			reduce(204), // ^, reduce: PowTerm
			reduce(204), // &, reduce: PowTerm
			reduce(204), // <<, reduce: PowTerm
			reduce(204), // >>>, reduce: PowTerm
			reduce(204), // +, reduce: PowTerm
			reduce(204), // -, reduce: PowTerm
			reduce(204), // .+, reduce: PowTerm
			reduce(204), // .-, reduce: PowTerm
			reduce(204), // *, reduce: PowTerm
			reduce(204), // /, reduce: PowTerm
			reduce(204), // //, reduce: PowTerm
			reduce(204), // %, reduce: PowTerm
			reduce(204), // .*, reduce: PowTerm
			reduce(204), // ./, reduce: PowTerm
			reduce(204), // .//, reduce: PowTerm
			reduce(204), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3607
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4103), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4104), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3608
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4106), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3609
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4107), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3610
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4108), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3611
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(229), // =, reduce: ArrayLiteral
			nil,         // unset
			nil,         // filter
			reduce(229), // >, reduce: ArrayLiteral
			reduce(229), // >>, reduce: ArrayLiteral
			reduce(229), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(229), // ||=, reduce: ArrayLiteral
			reduce(229), // ^^=, reduce: ArrayLiteral
			reduce(229), // &&=, reduce: ArrayLiteral
			reduce(229), // ??=, reduce: ArrayLiteral
			reduce(229), // ???=, reduce: ArrayLiteral
			reduce(229), // |=, reduce: ArrayLiteral
			reduce(229), // &=, reduce: ArrayLiteral
			reduce(229), // ^=, reduce: ArrayLiteral
			reduce(229), // <<=, reduce: ArrayLiteral
			reduce(229), // >>=, reduce: ArrayLiteral
			reduce(229), // >>>=, reduce: ArrayLiteral
			reduce(229), // +=, reduce: ArrayLiteral
			reduce(229), // .=, reduce: ArrayLiteral
			reduce(229), // -=, reduce: ArrayLiteral
			reduce(229), // *=, reduce: ArrayLiteral
			reduce(229), // /=, reduce: ArrayLiteral
			reduce(229), // //=, reduce: ArrayLiteral
			reduce(229), // %=, reduce: ArrayLiteral
			reduce(229), // **=, reduce: ArrayLiteral
			reduce(229), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(229), // ||, reduce: ArrayLiteral
			reduce(229), // ^^, reduce: ArrayLiteral
			reduce(229), // &&, reduce: ArrayLiteral
			reduce(229), // ??, reduce: ArrayLiteral
			reduce(229), // ???, reduce: ArrayLiteral
			reduce(229), // =~, reduce: ArrayLiteral
			reduce(229), // !=~, reduce: ArrayLiteral
			reduce(229), // ==, reduce: ArrayLiteral
			reduce(229), // !=, reduce: ArrayLiteral
			reduce(229), // <=>, reduce: ArrayLiteral
			reduce(229), // >=, reduce: ArrayLiteral
			reduce(229), // <, reduce: ArrayLiteral
			reduce(229), // <=, reduce: ArrayLiteral
			reduce(229), // ^, reduce: ArrayLiteral
			reduce(229), // &, reduce: ArrayLiteral
			reduce(229), // <<, reduce: ArrayLiteral
			reduce(229), // >>>, reduce: ArrayLiteral
			reduce(229), // +, reduce: ArrayLiteral
			reduce(229), // -, reduce: ArrayLiteral
			reduce(229), // .+, reduce: ArrayLiteral
			reduce(229), // .-, reduce: ArrayLiteral
			reduce(229), // *, reduce: ArrayLiteral
			reduce(229), // /, reduce: ArrayLiteral
			reduce(229), // //, reduce: ArrayLiteral
			reduce(229), // %, reduce: ArrayLiteral
			reduce(229), // .*, reduce: ArrayLiteral
			reduce(229), // ./, reduce: ArrayLiteral
			reduce(229), // .//, reduce: ArrayLiteral
			reduce(229), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(229), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(229), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(229), // [[, reduce: ArrayLiteral
			reduce(229), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3612
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(255), // =, reduce: ENV
			nil,         // unset
			nil,         // filter
			reduce(255), // >, reduce: ENV
			reduce(255), // >>, reduce: ENV
			reduce(255), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(255), // ||=, reduce: ENV
			reduce(255), // ^^=, reduce: ENV
			reduce(255), // &&=, reduce: ENV
			reduce(255), // ??=, reduce: ENV
			reduce(255), // ???=, reduce: ENV
			reduce(255), // |=, reduce: ENV
			reduce(255), // &=, reduce: ENV
			reduce(255), // ^=, reduce: ENV
			reduce(255), // <<=, reduce: ENV
			reduce(255), // >>=, reduce: ENV
			reduce(255), // >>>=, reduce: ENV
			reduce(255), // +=, reduce: ENV
			reduce(255), // .=, reduce: ENV
			reduce(255), // -=, reduce: ENV
			reduce(255), // *=, reduce: ENV
			reduce(255), // /=, reduce: ENV
			reduce(255), // //=, reduce: ENV
			reduce(255), // %=, reduce: ENV
			reduce(255), // **=, reduce: ENV
			reduce(255), // ?, reduce: ENV
			nil,         // :
			reduce(255), // ||, reduce: ENV
			reduce(255), // ^^, reduce: ENV
			reduce(255), // &&, reduce: ENV
			reduce(255), // ??, reduce: ENV
			reduce(255), // ???, reduce: ENV
			reduce(255), // =~, reduce: ENV
			reduce(255), // !=~, reduce: ENV
			reduce(255), // ==, reduce: ENV
			reduce(255), // !=, reduce: ENV
			reduce(255), // <=>, reduce: ENV
			reduce(255), // >=, reduce: ENV
			reduce(255), // <, reduce: ENV
			reduce(255), // <=, reduce: ENV
			reduce(255), // ^, reduce: ENV
			reduce(255), // &, reduce: ENV
			reduce(255), // <<, reduce: ENV
			reduce(255), // >>>, reduce: ENV
			reduce(255), // +, reduce: ENV
			reduce(255), // -, reduce: ENV
			reduce(255), // .+, reduce: ENV
			reduce(255), // .-, reduce: ENV
			reduce(255), // *, reduce: ENV
			reduce(255), // /, reduce: ENV
			reduce(255), // //, reduce: ENV
			reduce(255), // %, reduce: ENV
			reduce(255), // .*, reduce: ENV
			reduce(255), // ./, reduce: ENV
			reduce(255), // .//, reduce: ENV
			reduce(255), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(255), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(255), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(255), // [[, reduce: ENV
			reduce(255), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3613
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4109), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3614
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4110), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3615
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(268), // =, reduce: FunctionCallsite
			nil,         // unset
			nil,         // filter
			reduce(268), // >, reduce: FunctionCallsite
			reduce(268), // >>, reduce: FunctionCallsite
			reduce(268), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(268), // ||=, reduce: FunctionCallsite
			reduce(268), // ^^=, reduce: FunctionCallsite
			reduce(268), // &&=, reduce: FunctionCallsite
			reduce(268), // ??=, reduce: FunctionCallsite
			reduce(268), // ???=, reduce: FunctionCallsite
			reduce(268), // |=, reduce: FunctionCallsite
			reduce(268), // &=, reduce: FunctionCallsite
			reduce(268), // ^=, reduce: FunctionCallsite
			reduce(268), // <<=, reduce: FunctionCallsite
			reduce(268), // >>=, reduce: FunctionCallsite
			reduce(268), // >>>=, reduce: FunctionCallsite
			reduce(268), // +=, reduce: FunctionCallsite
			reduce(268), // .=, reduce: FunctionCallsite
			reduce(268), // -=, reduce: FunctionCallsite
			reduce(268), // *=, reduce: FunctionCallsite
			reduce(268), // /=, reduce: FunctionCallsite
			reduce(268), // //=, reduce: FunctionCallsite
			reduce(268), // %=, reduce: FunctionCallsite
			reduce(268), // **=, reduce: FunctionCallsite
			reduce(268), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(268), // ||, reduce: FunctionCallsite
			reduce(268), // ^^, reduce: FunctionCallsite
			reduce(268), // &&, reduce: FunctionCallsite
			reduce(268), // ??, reduce: FunctionCallsite
			reduce(268), // ???, reduce: FunctionCallsite
			reduce(268), // =~, reduce: FunctionCallsite
			reduce(268), // !=~, reduce: FunctionCallsite
			reduce(268), // ==, reduce: FunctionCallsite
			reduce(268), // !=, reduce: FunctionCallsite
			reduce(268), // <=>, reduce: FunctionCallsite
			reduce(268), // >=, reduce: FunctionCallsite
			reduce(268), // <, reduce: FunctionCallsite
			reduce(268), // <=, reduce: FunctionCallsite
			reduce(268), // ^, reduce: FunctionCallsite
			reduce(268), // &, reduce: FunctionCallsite
			reduce(268), // <<, reduce: FunctionCallsite
			reduce(268), // >>>, reduce: FunctionCallsite
			reduce(268), // +, reduce: FunctionCallsite
			reduce(268), // -, reduce: FunctionCallsite
			reduce(268), // .+, reduce: FunctionCallsite
			reduce(268), // .-, reduce: FunctionCallsite
			reduce(268), // *, reduce: FunctionCallsite
			reduce(268), // /, reduce: FunctionCallsite
			reduce(268), // //, reduce: FunctionCallsite
			reduce(268), // %, reduce: FunctionCallsite
			reduce(268), // .*, reduce: FunctionCallsite
			reduce(268), // ./, reduce: FunctionCallsite
			reduce(268), // .//, reduce: FunctionCallsite
			reduce(268), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(268), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(268), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(268), // [[, reduce: FunctionCallsite
			reduce(268), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3616
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(172),  // {
			shift(4111), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(177),  // (
			nil,         // )
			nil,         // emitp
			shift(189),  // field_name
			shift(190),  // $[
			nil,         // ]
			shift(191),  // braced_field_name
			shift(192),  // $[[
			shift(193),  // $[[[
			shift(194),  // full_srec
			shift(196),  // oosvar_name
			shift(197),  // @[
			shift(198),  // braced_oosvar_name
			shift(199),  // full_oosvar
			shift(200),  // all
			shift(201),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(218),  // +
			shift(220),  // -
			shift(221),  // .+
			shift(222),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(225),  // !
			shift(226),  // ~
			nil,         // **
			shift(231),  // string_literal
			shift(232),  // regex_case_insensitive
			shift(233),  // int_literal
			shift(234),  // float_literal
			shift(235),  // boolean_literal
			shift(236),  // null_literal
			shift(237),  // inf_literal
			shift(238),  // nan_literal
			shift(239),  // const_M_PI
			shift(240),  // const_M_E
			shift(241),  // panic
			shift(243),  // [
			shift(247),  // ctx_IPS
			shift(248),  // ctx_IFS
			shift(249),  // ctx_IRS
			shift(250),  // ctx_OPS
			shift(251),  // ctx_OFS
			shift(252),  // ctx_ORS
			shift(253),  // ctx_FLATSEP
			shift(254),  // ctx_NF
			shift(255),  // ctx_NR
			shift(256),  // ctx_FNR
			shift(257),  // ctx_FILENAME
			shift(258),  // ctx_FILENUM
			shift(260),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(267),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3617
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(315), // ;, reduce: TripleForContinuationItem
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(315), // ,, reduce: TripleForContinuationItem
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3618
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(316), // ;, reduce: TripleForContinuationItem
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(316), // ,, reduce: TripleForContinuationItem
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3619
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(38), // ;, reduce: BareBoolean
			nil,        // {
			nil,        // }
			reduce(36), // =, reduce: Lvalue
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(38), // ,, reduce: BareBoolean
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			reduce(36), // ||=, reduce: Lvalue
			reduce(36), // ^^=, reduce: Lvalue
			reduce(36), // &&=, reduce: Lvalue
			reduce(36), // ??=, reduce: Lvalue
			reduce(36), // ???=, reduce: Lvalue
			reduce(36), // |=, reduce: Lvalue
			reduce(36), // &=, reduce: Lvalue
			reduce(36), // ^=, reduce: Lvalue
			reduce(36), // <<=, reduce: Lvalue
			reduce(36), // >>=, reduce: Lvalue
			reduce(36), // >>>=, reduce: Lvalue
			reduce(36), // +=, reduce: Lvalue
			reduce(36), // .=, reduce: Lvalue
			reduce(36), // -=, reduce: Lvalue
			reduce(36), // *=, reduce: Lvalue
			reduce(36), // /=, reduce: Lvalue
			reduce(36), // //=, reduce: Lvalue
			reduce(36), // %=, reduce: Lvalue
			reduce(36), // **=, reduce: Lvalue
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3620
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(214), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(214), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(214), // >, reduce: MlrvalOrFunction
			reduce(214), // >>, reduce: MlrvalOrFunction
			reduce(214), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(214), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(214), // ||=, reduce: MlrvalOrFunction
			reduce(214), // ^^=, reduce: MlrvalOrFunction
			reduce(214), // &&=, reduce: MlrvalOrFunction
			reduce(214), // ??=, reduce: MlrvalOrFunction
			reduce(214), // ???=, reduce: MlrvalOrFunction
			reduce(214), // |=, reduce: MlrvalOrFunction
			reduce(214), // &=, reduce: MlrvalOrFunction
			reduce(214), // ^=, reduce: MlrvalOrFunction
			reduce(214), // <<=, reduce: MlrvalOrFunction
			reduce(214), // >>=, reduce: MlrvalOrFunction
			reduce(214), // >>>=, reduce: MlrvalOrFunction
			reduce(214), // +=, reduce: MlrvalOrFunction
			reduce(214), // .=, reduce: MlrvalOrFunction
			reduce(214), // -=, reduce: MlrvalOrFunction
			reduce(214), // *=, reduce: MlrvalOrFunction
			reduce(214), // /=, reduce: MlrvalOrFunction
			reduce(214), // //=, reduce: MlrvalOrFunction
			reduce(214), // %=, reduce: MlrvalOrFunction
			reduce(214), // **=, reduce: MlrvalOrFunction
			reduce(214), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(214), // ||, reduce: MlrvalOrFunction
			reduce(214), // ^^, reduce: MlrvalOrFunction
			reduce(214), // &&, reduce: MlrvalOrFunction
			reduce(214), // ??, reduce: MlrvalOrFunction
			reduce(214), // ???, reduce: MlrvalOrFunction
			reduce(214), // =~, reduce: MlrvalOrFunction
			reduce(214), // !=~, reduce: MlrvalOrFunction
			reduce(214), // ==, reduce: MlrvalOrFunction
			reduce(214), // !=, reduce: MlrvalOrFunction
			reduce(214), // <=>, reduce: MlrvalOrFunction
			reduce(214), // >=, reduce: MlrvalOrFunction
			reduce(214), // <, reduce: MlrvalOrFunction
			reduce(214), // <=, reduce: MlrvalOrFunction
			reduce(214), // ^, reduce: MlrvalOrFunction
			reduce(214), // &, reduce: MlrvalOrFunction
			reduce(214), // <<, reduce: MlrvalOrFunction
			reduce(214), // >>>, reduce: MlrvalOrFunction
			reduce(214), // +, reduce: MlrvalOrFunction
			reduce(214), // -, reduce: MlrvalOrFunction
			reduce(214), // .+, reduce: MlrvalOrFunction
			reduce(214), // .-, reduce: MlrvalOrFunction
			reduce(214), // *, reduce: MlrvalOrFunction
			reduce(214), // /, reduce: MlrvalOrFunction
			reduce(214), // //, reduce: MlrvalOrFunction
			reduce(214), // %, reduce: MlrvalOrFunction
			reduce(214), // .*, reduce: MlrvalOrFunction
			reduce(214), // ./, reduce: MlrvalOrFunction
			reduce(214), // .//, reduce: MlrvalOrFunction
			reduce(214), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(214), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(214), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(214), // [[, reduce: MlrvalOrFunction
			reduce(214), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3621
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(211), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(211), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(211), // >, reduce: MlrvalOrFunction
			reduce(211), // >>, reduce: MlrvalOrFunction
			reduce(211), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(211), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(211), // ||=, reduce: MlrvalOrFunction
			reduce(211), // ^^=, reduce: MlrvalOrFunction
			reduce(211), // &&=, reduce: MlrvalOrFunction
			reduce(211), // ??=, reduce: MlrvalOrFunction
			reduce(211), // ???=, reduce: MlrvalOrFunction
			reduce(211), // |=, reduce: MlrvalOrFunction
			reduce(211), // &=, reduce: MlrvalOrFunction
			reduce(211), // ^=, reduce: MlrvalOrFunction
			reduce(211), // <<=, reduce: MlrvalOrFunction
			reduce(211), // >>=, reduce: MlrvalOrFunction
			reduce(211), // >>>=, reduce: MlrvalOrFunction
			reduce(211), // +=, reduce: MlrvalOrFunction
			reduce(211), // .=, reduce: MlrvalOrFunction
			reduce(211), // -=, reduce: MlrvalOrFunction
			reduce(211), // *=, reduce: MlrvalOrFunction
			reduce(211), // /=, reduce: MlrvalOrFunction
			reduce(211), // //=, reduce: MlrvalOrFunction
			reduce(211), // %=, reduce: MlrvalOrFunction
			reduce(211), // **=, reduce: MlrvalOrFunction
			reduce(211), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(211), // ||, reduce: MlrvalOrFunction
			reduce(211), // ^^, reduce: MlrvalOrFunction
			reduce(211), // &&, reduce: MlrvalOrFunction
			reduce(211), // ??, reduce: MlrvalOrFunction
			reduce(211), // ???, reduce: MlrvalOrFunction
			reduce(211), // =~, reduce: MlrvalOrFunction
			reduce(211), // !=~, reduce: MlrvalOrFunction
			reduce(211), // ==, reduce: MlrvalOrFunction
			reduce(211), // !=, reduce: MlrvalOrFunction
			reduce(211), // <=>, reduce: MlrvalOrFunction
			reduce(211), // >=, reduce: MlrvalOrFunction
			reduce(211), // <, reduce: MlrvalOrFunction
			reduce(211), // <=, reduce: MlrvalOrFunction
			reduce(211), // ^, reduce: MlrvalOrFunction
			reduce(211), // &, reduce: MlrvalOrFunction
			reduce(211), // <<, reduce: MlrvalOrFunction
			reduce(211), // >>>, reduce: MlrvalOrFunction
			reduce(211), // +, reduce: MlrvalOrFunction
			reduce(211), // -, reduce: MlrvalOrFunction
			reduce(211), // .+, reduce: MlrvalOrFunction
			reduce(211), // .-, reduce: MlrvalOrFunction
			reduce(211), // *, reduce: MlrvalOrFunction
			reduce(211), // /, reduce: MlrvalOrFunction
			reduce(211), // //, reduce: MlrvalOrFunction
			reduce(211), // %, reduce: MlrvalOrFunction
			reduce(211), // .*, reduce: MlrvalOrFunction
			reduce(211), // ./, reduce: MlrvalOrFunction
			reduce(211), // .//, reduce: MlrvalOrFunction
			reduce(211), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(211), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(211), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(211), // [[, reduce: MlrvalOrFunction
			reduce(211), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3622
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3623
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(110), // ;, reduce: OosvarValue
			nil,         // {
			nil,         // }
			reduce(110), // =, reduce: OosvarValue
			nil,         // unset
			nil,         // filter
			reduce(110), // >, reduce: OosvarValue
			reduce(110), // >>, reduce: OosvarValue
			reduce(110), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(110), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(110), // ||=, reduce: OosvarValue
			reduce(110), // ^^=, reduce: OosvarValue
			reduce(110), // &&=, reduce: OosvarValue
			reduce(110), // ??=, reduce: OosvarValue
			reduce(110), // ???=, reduce: OosvarValue
			reduce(110), // |=, reduce: OosvarValue
			reduce(110), // &=, reduce: OosvarValue
			reduce(110), // ^=, reduce: OosvarValue
			reduce(110), // <<=, reduce: OosvarValue
			reduce(110), // >>=, reduce: OosvarValue
			reduce(110), // >>>=, reduce: OosvarValue
			reduce(110), // +=, reduce: OosvarValue
			reduce(110), // .=, reduce: OosvarValue
			reduce(110), // -=, reduce: OosvarValue
			reduce(110), // *=, reduce: OosvarValue
			reduce(110), // /=, reduce: OosvarValue
			reduce(110), // //=, reduce: OosvarValue
			reduce(110), // %=, reduce: OosvarValue
			reduce(110), // **=, reduce: OosvarValue
			reduce(110), // ?, reduce: OosvarValue
			nil,         // :
			reduce(110), // ||, reduce: OosvarValue
			reduce(110), // ^^, reduce: OosvarValue
			reduce(110), // &&, reduce: OosvarValue
			reduce(110), // ??, reduce: OosvarValue
			reduce(110), // ???, reduce: OosvarValue
			reduce(110), // =~, reduce: OosvarValue
			reduce(110), // !=~, reduce: OosvarValue
			reduce(110), // ==, reduce: OosvarValue
			reduce(110), // !=, reduce: OosvarValue
			reduce(110), // <=>, reduce: OosvarValue
			reduce(110), // >=, reduce: OosvarValue
			reduce(110), // <, reduce: OosvarValue
			reduce(110), // <=, reduce: OosvarValue
			reduce(110), // ^, reduce: OosvarValue
			reduce(110), // &, reduce: OosvarValue
			reduce(110), // <<, reduce: OosvarValue
			reduce(110), // >>>, reduce: OosvarValue
			reduce(110), // +, reduce: OosvarValue
			reduce(110), // -, reduce: OosvarValue
			reduce(110), // .+, reduce: OosvarValue
			reduce(110), // .-, reduce: OosvarValue
			reduce(110), // *, reduce: OosvarValue
			reduce(110), // /, reduce: OosvarValue
			reduce(110), // //, reduce: OosvarValue
			reduce(110), // %, reduce: OosvarValue
			reduce(110), // .*, reduce: OosvarValue
			reduce(110), // ./, reduce: OosvarValue
			reduce(110), // .//, reduce: OosvarValue
			reduce(110), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(110), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(110), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(110), // [[, reduce: OosvarValue
			reduce(110), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3624
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(112), // ;, reduce: OosvarValue
			nil,         // {
			nil,         // }
			reduce(112), // =, reduce: OosvarValue
			nil,         // unset
			nil,         // filter
			reduce(112), // >, reduce: OosvarValue
			reduce(112), // >>, reduce: OosvarValue
			reduce(112), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(112), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(112), // ||=, reduce: OosvarValue
			reduce(112), // ^^=, reduce: OosvarValue
			reduce(112), // &&=, reduce: OosvarValue
			reduce(112), // ??=, reduce: OosvarValue
			reduce(112), // ???=, reduce: OosvarValue
			reduce(112), // |=, reduce: OosvarValue
			reduce(112), // &=, reduce: OosvarValue
			reduce(112), // ^=, reduce: OosvarValue
			reduce(112), // <<=, reduce: OosvarValue
			reduce(112), // >>=, reduce: OosvarValue
			reduce(112), // >>>=, reduce: OosvarValue
			reduce(112), // +=, reduce: OosvarValue
			reduce(112), // .=, reduce: OosvarValue
			reduce(112), // -=, reduce: OosvarValue
			reduce(112), // *=, reduce: OosvarValue
			reduce(112), // /=, reduce: OosvarValue
			reduce(112), // //=, reduce: OosvarValue
			reduce(112), // %=, reduce: OosvarValue
			reduce(112), // **=, reduce: OosvarValue
			reduce(112), // ?, reduce: OosvarValue
			nil,         // :
			reduce(112), // ||, reduce: OosvarValue
			reduce(112), // ^^, reduce: OosvarValue
			reduce(112), // &&, reduce: OosvarValue
			reduce(112), // ??, reduce: OosvarValue
			reduce(112), // ???, reduce: OosvarValue
			reduce(112), // =~, reduce: OosvarValue
			reduce(112), // !=~, reduce: OosvarValue
			reduce(112), // ==, reduce: OosvarValue
			reduce(112), // !=, reduce: OosvarValue
			reduce(112), // <=>, reduce: OosvarValue
			reduce(112), // >=, reduce: OosvarValue
			reduce(112), // <, reduce: OosvarValue
			reduce(112), // <=, reduce: OosvarValue
			reduce(112), // ^, reduce: OosvarValue
			reduce(112), // &, reduce: OosvarValue
			reduce(112), // <<, reduce: OosvarValue
			reduce(112), // >>>, reduce: OosvarValue
			reduce(112), // +, reduce: OosvarValue
			reduce(112), // -, reduce: OosvarValue
			reduce(112), // .+, reduce: OosvarValue
			reduce(112), // .-, reduce: OosvarValue
			reduce(112), // *, reduce: OosvarValue
			reduce(112), // /, reduce: OosvarValue
			reduce(112), // //, reduce: OosvarValue
			reduce(112), // %, reduce: OosvarValue
			reduce(112), // .*, reduce: OosvarValue
			reduce(112), // ./, reduce: OosvarValue
			reduce(112), // .//, reduce: OosvarValue
			reduce(112), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(112), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(112), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(112), // [[, reduce: OosvarValue
			reduce(112), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3625
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(111), // ;, reduce: OosvarValue
			nil,         // {
			nil,         // }
			reduce(111), // =, reduce: OosvarValue
			nil,         // unset
			nil,         // filter
			reduce(111), // >, reduce: OosvarValue
			reduce(111), // >>, reduce: OosvarValue
			reduce(111), // |, reduce: OosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(111), // ,, reduce: OosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(111), // ||=, reduce: OosvarValue
			reduce(111), // ^^=, reduce: OosvarValue
			reduce(111), // &&=, reduce: OosvarValue
			reduce(111), // ??=, reduce: OosvarValue
			reduce(111), // ???=, reduce: OosvarValue
			reduce(111), // |=, reduce: OosvarValue
			reduce(111), // &=, reduce: OosvarValue
			reduce(111), // ^=, reduce: OosvarValue
			reduce(111), // <<=, reduce: OosvarValue
			reduce(111), // >>=, reduce: OosvarValue
			reduce(111), // >>>=, reduce: OosvarValue
			reduce(111), // +=, reduce: OosvarValue
			reduce(111), // .=, reduce: OosvarValue
			reduce(111), // -=, reduce: OosvarValue
			reduce(111), // *=, reduce: OosvarValue
			reduce(111), // /=, reduce: OosvarValue
			reduce(111), // //=, reduce: OosvarValue
			reduce(111), // %=, reduce: OosvarValue
			reduce(111), // **=, reduce: OosvarValue
			reduce(111), // ?, reduce: OosvarValue
			nil,         // :
			reduce(111), // ||, reduce: OosvarValue
			reduce(111), // ^^, reduce: OosvarValue
			reduce(111), // &&, reduce: OosvarValue
			reduce(111), // ??, reduce: OosvarValue
			reduce(111), // ???, reduce: OosvarValue
			reduce(111), // =~, reduce: OosvarValue
			reduce(111), // !=~, reduce: OosvarValue
			reduce(111), // ==, reduce: OosvarValue
			reduce(111), // !=, reduce: OosvarValue
			reduce(111), // <=>, reduce: OosvarValue
			reduce(111), // >=, reduce: OosvarValue
			reduce(111), // <, reduce: OosvarValue
			reduce(111), // <=, reduce: OosvarValue
			reduce(111), // ^, reduce: OosvarValue
			reduce(111), // &, reduce: OosvarValue
			reduce(111), // <<, reduce: OosvarValue
			reduce(111), // >>>, reduce: OosvarValue
			reduce(111), // +, reduce: OosvarValue
			reduce(111), // -, reduce: OosvarValue
			reduce(111), // .+, reduce: OosvarValue
			reduce(111), // .-, reduce: OosvarValue
			reduce(111), // *, reduce: OosvarValue
			reduce(111), // /, reduce: OosvarValue
			reduce(111), // //, reduce: OosvarValue
			reduce(111), // %, reduce: OosvarValue
			reduce(111), // .*, reduce: OosvarValue
			reduce(111), // ./, reduce: OosvarValue
			reduce(111), // .//, reduce: OosvarValue
			reduce(111), // ., reduce: OosvarValue
			nil,         // !
			nil,         // ~
			reduce(111), // **, reduce: OosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(111), // [, reduce: OosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(111), // [[, reduce: OosvarValue
			reduce(111), // [[[, reduce: OosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3626
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(99), // ;, reduce: FieldValue
			nil,        // {
			nil,        // }
			reduce(99), // =, reduce: FieldValue
			nil,        // unset
			nil,        // filter
			reduce(99), // >, reduce: FieldValue
			reduce(99), // >>, reduce: FieldValue
			reduce(99), // |, reduce: FieldValue
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(99), // ,, reduce: FieldValue
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			reduce(99), // ||=, reduce: FieldValue
			reduce(99), // ^^=, reduce: FieldValue
			reduce(99), // &&=, reduce: FieldValue
			reduce(99), // ??=, reduce: FieldValue
			reduce(99), // ???=, reduce: FieldValue
			reduce(99), // |=, reduce: FieldValue
			reduce(99), // &=, reduce: FieldValue
			reduce(99), // ^=, reduce: FieldValue
			reduce(99), // <<=, reduce: FieldValue
			reduce(99), // >>=, reduce: FieldValue
			reduce(99), // >>>=, reduce: FieldValue
			reduce(99), // +=, reduce: FieldValue
			reduce(99), // .=, reduce: FieldValue
			reduce(99), // -=, reduce: FieldValue
			reduce(99), // *=, reduce: FieldValue
			reduce(99), // /=, reduce: FieldValue
			reduce(99), // //=, reduce: FieldValue
			reduce(99), // %=, reduce: FieldValue
			reduce(99), // **=, reduce: FieldValue
			reduce(99), // ?, reduce: FieldValue
			nil,        // :
			reduce(99), // ||, reduce: FieldValue
			reduce(99), // ^^, reduce: FieldValue
			reduce(99), // &&, reduce: FieldValue
			reduce(99), // ??, reduce: FieldValue
			reduce(99), // ???, reduce: FieldValue
			reduce(99), // =~, reduce: FieldValue
			reduce(99), // !=~, reduce: FieldValue
			reduce(99), // ==, reduce: FieldValue
			reduce(99), // !=, reduce: FieldValue
			reduce(99), // <=>, reduce: FieldValue
			reduce(99), // >=, reduce: FieldValue
			reduce(99), // <, reduce: FieldValue
			reduce(99), // <=, reduce: FieldValue
			reduce(99), // ^, reduce: FieldValue
			reduce(99), // &, reduce: FieldValue
			reduce(99), // <<, reduce: FieldValue
			reduce(99), // >>>, reduce: FieldValue
			reduce(99), // +, reduce: FieldValue
			reduce(99), // -, reduce: FieldValue
			reduce(99), // .+, reduce: FieldValue
			reduce(99), // .-, reduce: FieldValue
			reduce(99), // *, reduce: FieldValue
			reduce(99), // /, reduce: FieldValue
			reduce(99), // //, reduce: FieldValue
			reduce(99), // %, reduce: FieldValue
			reduce(99), // .*, reduce: FieldValue
			reduce(99), // ./, reduce: FieldValue
			reduce(99), // .//, reduce: FieldValue
			reduce(99), // ., reduce: FieldValue
			nil,        // !
			nil,        // ~
			reduce(99), // **, reduce: FieldValue
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(99), // [, reduce: FieldValue
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(99), // [[, reduce: FieldValue
			reduce(99), // [[[, reduce: FieldValue
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3627
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(101), // ;, reduce: FieldValue
			nil,         // {
			nil,         // }
			reduce(101), // =, reduce: FieldValue
			nil,         // unset
			nil,         // filter
			reduce(101), // >, reduce: FieldValue
			reduce(101), // >>, reduce: FieldValue
			reduce(101), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(101), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(101), // ||=, reduce: FieldValue
			reduce(101), // ^^=, reduce: FieldValue
			reduce(101), // &&=, reduce: FieldValue
			reduce(101), // ??=, reduce: FieldValue
			reduce(101), // ???=, reduce: FieldValue
			reduce(101), // |=, reduce: FieldValue
			reduce(101), // &=, reduce: FieldValue
			reduce(101), // ^=, reduce: FieldValue
			reduce(101), // <<=, reduce: FieldValue
			reduce(101), // >>=, reduce: FieldValue
			reduce(101), // >>>=, reduce: FieldValue
			reduce(101), // +=, reduce: FieldValue
			reduce(101), // .=, reduce: FieldValue
			reduce(101), // -=, reduce: FieldValue
			reduce(101), // *=, reduce: FieldValue
			reduce(101), // /=, reduce: FieldValue
			reduce(101), // //=, reduce: FieldValue
			reduce(101), // %=, reduce: FieldValue
			reduce(101), // **=, reduce: FieldValue
			reduce(101), // ?, reduce: FieldValue
			nil,         // :
			reduce(101), // ||, reduce: FieldValue
			reduce(101), // ^^, reduce: FieldValue
			reduce(101), // &&, reduce: FieldValue
			reduce(101), // ??, reduce: FieldValue
			reduce(101), // ???, reduce: FieldValue
			reduce(101), // =~, reduce: FieldValue
			reduce(101), // !=~, reduce: FieldValue
			reduce(101), // ==, reduce: FieldValue
			reduce(101), // !=, reduce: FieldValue
			reduce(101), // <=>, reduce: FieldValue
			reduce(101), // >=, reduce: FieldValue
			reduce(101), // <, reduce: FieldValue
			reduce(101), // <=, reduce: FieldValue
			reduce(101), // ^, reduce: FieldValue
			reduce(101), // &, reduce: FieldValue
			reduce(101), // <<, reduce: FieldValue
			reduce(101), // >>>, reduce: FieldValue
			reduce(101), // +, reduce: FieldValue
			reduce(101), // -, reduce: FieldValue
			reduce(101), // .+, reduce: FieldValue
			reduce(101), // .-, reduce: FieldValue
			reduce(101), // *, reduce: FieldValue
			reduce(101), // /, reduce: FieldValue
			reduce(101), // //, reduce: FieldValue
			reduce(101), // %, reduce: FieldValue
			reduce(101), // .*, reduce: FieldValue
			reduce(101), // ./, reduce: FieldValue
			reduce(101), // .//, reduce: FieldValue
			reduce(101), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(101), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(101), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(101), // [[, reduce: FieldValue
			reduce(101), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3628
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(100), // ;, reduce: FieldValue
			nil,         // {
			nil,         // }
			reduce(100), // =, reduce: FieldValue
			nil,         // unset
			nil,         // filter
			reduce(100), // >, reduce: FieldValue
			reduce(100), // >>, reduce: FieldValue
			reduce(100), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(100), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(100), // ||=, reduce: FieldValue
			reduce(100), // ^^=, reduce: FieldValue
			reduce(100), // &&=, reduce: FieldValue
			reduce(100), // ??=, reduce: FieldValue
			reduce(100), // ???=, reduce: FieldValue
			reduce(100), // |=, reduce: FieldValue
			reduce(100), // &=, reduce: FieldValue
			reduce(100), // ^=, reduce: FieldValue
			reduce(100), // <<=, reduce: FieldValue
			reduce(100), // >>=, reduce: FieldValue
			reduce(100), // >>>=, reduce: FieldValue
			reduce(100), // +=, reduce: FieldValue
			reduce(100), // .=, reduce: FieldValue
			reduce(100), // -=, reduce: FieldValue
			reduce(100), // *=, reduce: FieldValue
			reduce(100), // /=, reduce: FieldValue
			reduce(100), // //=, reduce: FieldValue
			reduce(100), // %=, reduce: FieldValue
			reduce(100), // **=, reduce: FieldValue
			reduce(100), // ?, reduce: FieldValue
			nil,         // :
			reduce(100), // ||, reduce: FieldValue
			reduce(100), // ^^, reduce: FieldValue
			reduce(100), // &&, reduce: FieldValue
			reduce(100), // ??, reduce: FieldValue
			reduce(100), // ???, reduce: FieldValue
			reduce(100), // =~, reduce: FieldValue
			reduce(100), // !=~, reduce: FieldValue
			reduce(100), // ==, reduce: FieldValue
			reduce(100), // !=, reduce: FieldValue
			reduce(100), // <=>, reduce: FieldValue
			reduce(100), // >=, reduce: FieldValue
			reduce(100), // <, reduce: FieldValue
			reduce(100), // <=, reduce: FieldValue
			reduce(100), // ^, reduce: FieldValue
			reduce(100), // &, reduce: FieldValue
			reduce(100), // <<, reduce: FieldValue
			reduce(100), // >>>, reduce: FieldValue
			reduce(100), // +, reduce: FieldValue
			reduce(100), // -, reduce: FieldValue
			reduce(100), // .+, reduce: FieldValue
			reduce(100), // .-, reduce: FieldValue
			reduce(100), // *, reduce: FieldValue
			reduce(100), // /, reduce: FieldValue
			reduce(100), // //, reduce: FieldValue
			reduce(100), // %, reduce: FieldValue
			reduce(100), // .*, reduce: FieldValue
			reduce(100), // ./, reduce: FieldValue
			reduce(100), // .//, reduce: FieldValue
			reduce(100), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(100), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(100), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(100), // [[, reduce: FieldValue
			reduce(100), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3629
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(213), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(213), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(213), // >, reduce: MlrvalOrFunction
			reduce(213), // >>, reduce: MlrvalOrFunction
			reduce(213), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(213), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(213), // ||=, reduce: MlrvalOrFunction
			reduce(213), // ^^=, reduce: MlrvalOrFunction
			reduce(213), // &&=, reduce: MlrvalOrFunction
			reduce(213), // ??=, reduce: MlrvalOrFunction
			reduce(213), // ???=, reduce: MlrvalOrFunction
			reduce(213), // |=, reduce: MlrvalOrFunction
			reduce(213), // &=, reduce: MlrvalOrFunction
			reduce(213), // ^=, reduce: MlrvalOrFunction
			reduce(213), // <<=, reduce: MlrvalOrFunction
			reduce(213), // >>=, reduce: MlrvalOrFunction
			reduce(213), // >>>=, reduce: MlrvalOrFunction
			reduce(213), // +=, reduce: MlrvalOrFunction
			reduce(213), // .=, reduce: MlrvalOrFunction
			reduce(213), // -=, reduce: MlrvalOrFunction
			reduce(213), // *=, reduce: MlrvalOrFunction
			reduce(213), // /=, reduce: MlrvalOrFunction
			reduce(213), // //=, reduce: MlrvalOrFunction
			reduce(213), // %=, reduce: MlrvalOrFunction
			reduce(213), // **=, reduce: MlrvalOrFunction
			reduce(213), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(213), // ||, reduce: MlrvalOrFunction
			reduce(213), // ^^, reduce: MlrvalOrFunction
			reduce(213), // &&, reduce: MlrvalOrFunction
			reduce(213), // ??, reduce: MlrvalOrFunction
			reduce(213), // ???, reduce: MlrvalOrFunction
			reduce(213), // =~, reduce: MlrvalOrFunction
			reduce(213), // !=~, reduce: MlrvalOrFunction
			reduce(213), // ==, reduce: MlrvalOrFunction
			reduce(213), // !=, reduce: MlrvalOrFunction
			reduce(213), // <=>, reduce: MlrvalOrFunction
			reduce(213), // >=, reduce: MlrvalOrFunction
			reduce(213), // <, reduce: MlrvalOrFunction
			reduce(213), // <=, reduce: MlrvalOrFunction
			reduce(213), // ^, reduce: MlrvalOrFunction
			reduce(213), // &, reduce: MlrvalOrFunction
			reduce(213), // <<, reduce: MlrvalOrFunction
			reduce(213), // >>>, reduce: MlrvalOrFunction
			reduce(213), // +, reduce: MlrvalOrFunction
			reduce(213), // -, reduce: MlrvalOrFunction
			reduce(213), // .+, reduce: MlrvalOrFunction
			reduce(213), // .-, reduce: MlrvalOrFunction
			reduce(213), // *, reduce: MlrvalOrFunction
			reduce(213), // /, reduce: MlrvalOrFunction
			reduce(213), // //, reduce: MlrvalOrFunction
			reduce(213), // %, reduce: MlrvalOrFunction
			reduce(213), // .*, reduce: MlrvalOrFunction
			reduce(213), // ./, reduce: MlrvalOrFunction
			reduce(213), // .//, reduce: MlrvalOrFunction
			reduce(213), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(213), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(213), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(213), // [[, reduce: MlrvalOrFunction
			reduce(213), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3630
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(233), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(233), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(233), // >, reduce: MlrvalOrFunction
			reduce(233), // >>, reduce: MlrvalOrFunction
			reduce(233), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(233), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(233), // ||=, reduce: MlrvalOrFunction
			reduce(233), // ^^=, reduce: MlrvalOrFunction
			reduce(233), // &&=, reduce: MlrvalOrFunction
			reduce(233), // ??=, reduce: MlrvalOrFunction
			reduce(233), // ???=, reduce: MlrvalOrFunction
			reduce(233), // |=, reduce: MlrvalOrFunction
			reduce(233), // &=, reduce: MlrvalOrFunction
			reduce(233), // ^=, reduce: MlrvalOrFunction
			reduce(233), // <<=, reduce: MlrvalOrFunction
			reduce(233), // >>=, reduce: MlrvalOrFunction
			reduce(233), // >>>=, reduce: MlrvalOrFunction
			reduce(233), // +=, reduce: MlrvalOrFunction
			reduce(233), // .=, reduce: MlrvalOrFunction
			reduce(233), // -=, reduce: MlrvalOrFunction
			reduce(233), // *=, reduce: MlrvalOrFunction
			reduce(233), // /=, reduce: MlrvalOrFunction
			reduce(233), // //=, reduce: MlrvalOrFunction
			reduce(233), // %=, reduce: MlrvalOrFunction
			reduce(233), // **=, reduce: MlrvalOrFunction
			reduce(233), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(233), // ||, reduce: MlrvalOrFunction
			reduce(233), // ^^, reduce: MlrvalOrFunction
			reduce(233), // &&, reduce: MlrvalOrFunction
			reduce(233), // ??, reduce: MlrvalOrFunction
			reduce(233), // ???, reduce: MlrvalOrFunction
			reduce(233), // =~, reduce: MlrvalOrFunction
			reduce(233), // !=~, reduce: MlrvalOrFunction
			reduce(233), // ==, reduce: MlrvalOrFunction
			reduce(233), // !=, reduce: MlrvalOrFunction
			reduce(233), // <=>, reduce: MlrvalOrFunction
			reduce(233), // >=, reduce: MlrvalOrFunction
			reduce(233), // <, reduce: MlrvalOrFunction
			reduce(233), // <=, reduce: MlrvalOrFunction
			reduce(233), // ^, reduce: MlrvalOrFunction
			reduce(233), // &, reduce: MlrvalOrFunction
			reduce(233), // <<, reduce: MlrvalOrFunction
			reduce(233), // >>>, reduce: MlrvalOrFunction
			reduce(233), // +, reduce: MlrvalOrFunction
			reduce(233), // -, reduce: MlrvalOrFunction
			reduce(233), // .+, reduce: MlrvalOrFunction
			reduce(233), // .-, reduce: MlrvalOrFunction
			reduce(233), // *, reduce: MlrvalOrFunction
			reduce(233), // /, reduce: MlrvalOrFunction
			reduce(233), // //, reduce: MlrvalOrFunction
			reduce(233), // %, reduce: MlrvalOrFunction
			reduce(233), // .*, reduce: MlrvalOrFunction
			reduce(233), // ./, reduce: MlrvalOrFunction
			reduce(233), // .//, reduce: MlrvalOrFunction
			reduce(233), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(233), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(233), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(233), // [[, reduce: MlrvalOrFunction
			reduce(233), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3631
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(210), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(210), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(210), // >, reduce: MlrvalOrFunction
			reduce(210), // >>, reduce: MlrvalOrFunction
			reduce(210), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(210), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(210), // ||=, reduce: MlrvalOrFunction
			reduce(210), // ^^=, reduce: MlrvalOrFunction
			reduce(210), // &&=, reduce: MlrvalOrFunction
			reduce(210), // ??=, reduce: MlrvalOrFunction
			reduce(210), // ???=, reduce: MlrvalOrFunction
			reduce(210), // |=, reduce: MlrvalOrFunction
			reduce(210), // &=, reduce: MlrvalOrFunction
			reduce(210), // ^=, reduce: MlrvalOrFunction
			reduce(210), // <<=, reduce: MlrvalOrFunction
			reduce(210), // >>=, reduce: MlrvalOrFunction
			reduce(210), // >>>=, reduce: MlrvalOrFunction
			reduce(210), // +=, reduce: MlrvalOrFunction
			reduce(210), // .=, reduce: MlrvalOrFunction
			reduce(210), // -=, reduce: MlrvalOrFunction
			reduce(210), // *=, reduce: MlrvalOrFunction
			reduce(210), // /=, reduce: MlrvalOrFunction
			reduce(210), // //=, reduce: MlrvalOrFunction
			reduce(210), // %=, reduce: MlrvalOrFunction
			reduce(210), // **=, reduce: MlrvalOrFunction
			reduce(210), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(210), // ||, reduce: MlrvalOrFunction
			reduce(210), // ^^, reduce: MlrvalOrFunction
			reduce(210), // &&, reduce: MlrvalOrFunction
			reduce(210), // ??, reduce: MlrvalOrFunction
			reduce(210), // ???, reduce: MlrvalOrFunction
			reduce(210), // =~, reduce: MlrvalOrFunction
			reduce(210), // !=~, reduce: MlrvalOrFunction
			reduce(210), // ==, reduce: MlrvalOrFunction
			reduce(210), // !=, reduce: MlrvalOrFunction
			reduce(210), // <=>, reduce: MlrvalOrFunction
			reduce(210), // >=, reduce: MlrvalOrFunction
			reduce(210), // <, reduce: MlrvalOrFunction
			reduce(210), // <=, reduce: MlrvalOrFunction
			reduce(210), // ^, reduce: MlrvalOrFunction
			reduce(210), // &, reduce: MlrvalOrFunction
			reduce(210), // <<, reduce: MlrvalOrFunction
			reduce(210), // >>>, reduce: MlrvalOrFunction
			reduce(210), // +, reduce: MlrvalOrFunction
			reduce(210), // -, reduce: MlrvalOrFunction
			reduce(210), // .+, reduce: MlrvalOrFunction
			reduce(210), // .-, reduce: MlrvalOrFunction
			reduce(210), // *, reduce: MlrvalOrFunction
			reduce(210), // /, reduce: MlrvalOrFunction
			reduce(210), // //, reduce: MlrvalOrFunction
			reduce(210), // %, reduce: MlrvalOrFunction
			reduce(210), // .*, reduce: MlrvalOrFunction
			reduce(210), // ./, reduce: MlrvalOrFunction
			reduce(210), // .//, reduce: MlrvalOrFunction
			reduce(210), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(210), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(210), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(210), // [[, reduce: MlrvalOrFunction
			reduce(210), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3632
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(102), // ;, reduce: FieldValue
			nil,         // {
			nil,         // }
			reduce(102), // =, reduce: FieldValue
			nil,         // unset
			nil,         // filter
			reduce(102), // >, reduce: FieldValue
			reduce(102), // >>, reduce: FieldValue
			reduce(102), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(102), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(102), // ||=, reduce: FieldValue
			reduce(102), // ^^=, reduce: FieldValue
			reduce(102), // &&=, reduce: FieldValue
			reduce(102), // ??=, reduce: FieldValue
			reduce(102), // ???=, reduce: FieldValue
			reduce(102), // |=, reduce: FieldValue
			reduce(102), // &=, reduce: FieldValue
			reduce(102), // ^=, reduce: FieldValue
			reduce(102), // <<=, reduce: FieldValue
			reduce(102), // >>=, reduce: FieldValue
			reduce(102), // >>>=, reduce: FieldValue
			reduce(102), // +=, reduce: FieldValue
			reduce(102), // .=, reduce: FieldValue
			reduce(102), // -=, reduce: FieldValue
			reduce(102), // *=, reduce: FieldValue
			reduce(102), // /=, reduce: FieldValue
			reduce(102), // //=, reduce: FieldValue
			reduce(102), // %=, reduce: FieldValue
			reduce(102), // **=, reduce: FieldValue
			reduce(102), // ?, reduce: FieldValue
			nil,         // :
			reduce(102), // ||, reduce: FieldValue
			reduce(102), // ^^, reduce: FieldValue
			reduce(102), // &&, reduce: FieldValue
			reduce(102), // ??, reduce: FieldValue
			reduce(102), // ???, reduce: FieldValue
			reduce(102), // =~, reduce: FieldValue
			reduce(102), // !=~, reduce: FieldValue
			reduce(102), // ==, reduce: FieldValue
			reduce(102), // !=, reduce: FieldValue
			reduce(102), // <=>, reduce: FieldValue
			reduce(102), // >=, reduce: FieldValue
			reduce(102), // <, reduce: FieldValue
			reduce(102), // <=, reduce: FieldValue
			reduce(102), // ^, reduce: FieldValue
			reduce(102), // &, reduce: FieldValue
			reduce(102), // <<, reduce: FieldValue
			reduce(102), // >>>, reduce: FieldValue
			reduce(102), // +, reduce: FieldValue
			reduce(102), // -, reduce: FieldValue
			reduce(102), // .+, reduce: FieldValue
			reduce(102), // .-, reduce: FieldValue
			reduce(102), // *, reduce: FieldValue
			reduce(102), // /, reduce: FieldValue
			reduce(102), // //, reduce: FieldValue
			reduce(102), // %, reduce: FieldValue
			reduce(102), // .*, reduce: FieldValue
			reduce(102), // ./, reduce: FieldValue
			reduce(102), // .//, reduce: FieldValue
			reduce(102), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(102), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(102), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(102), // [[, reduce: FieldValue
			reduce(102), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3633
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(103), // ;, reduce: FieldValue
			nil,         // {
			nil,         // }
			reduce(103), // =, reduce: FieldValue
			nil,         // unset
			nil,         // filter
			reduce(103), // >, reduce: FieldValue
			reduce(103), // >>, reduce: FieldValue
			reduce(103), // |, reduce: FieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(103), // ,, reduce: FieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(103), // ||=, reduce: FieldValue
			reduce(103), // ^^=, reduce: FieldValue
			reduce(103), // &&=, reduce: FieldValue
			reduce(103), // ??=, reduce: FieldValue
			reduce(103), // ???=, reduce: FieldValue
			reduce(103), // |=, reduce: FieldValue
			reduce(103), // &=, reduce: FieldValue
			reduce(103), // ^=, reduce: FieldValue
			reduce(103), // <<=, reduce: FieldValue
			reduce(103), // >>=, reduce: FieldValue
			reduce(103), // >>>=, reduce: FieldValue
			reduce(103), // +=, reduce: FieldValue
			reduce(103), // .=, reduce: FieldValue
			reduce(103), // -=, reduce: FieldValue
			reduce(103), // *=, reduce: FieldValue
			reduce(103), // /=, reduce: FieldValue
			reduce(103), // //=, reduce: FieldValue
			reduce(103), // %=, reduce: FieldValue
			reduce(103), // **=, reduce: FieldValue
			reduce(103), // ?, reduce: FieldValue
			nil,         // :
			reduce(103), // ||, reduce: FieldValue
			reduce(103), // ^^, reduce: FieldValue
			reduce(103), // &&, reduce: FieldValue
			reduce(103), // ??, reduce: FieldValue
			reduce(103), // ???, reduce: FieldValue
			reduce(103), // =~, reduce: FieldValue
			reduce(103), // !=~, reduce: FieldValue
			reduce(103), // ==, reduce: FieldValue
			reduce(103), // !=, reduce: FieldValue
			reduce(103), // <=>, reduce: FieldValue
			reduce(103), // >=, reduce: FieldValue
			reduce(103), // <, reduce: FieldValue
			reduce(103), // <=, reduce: FieldValue
			reduce(103), // ^, reduce: FieldValue
			reduce(103), // &, reduce: FieldValue
			reduce(103), // <<, reduce: FieldValue
			reduce(103), // >>>, reduce: FieldValue
			reduce(103), // +, reduce: FieldValue
			reduce(103), // -, reduce: FieldValue
			reduce(103), // .+, reduce: FieldValue
			reduce(103), // .-, reduce: FieldValue
			reduce(103), // *, reduce: FieldValue
			reduce(103), // /, reduce: FieldValue
			reduce(103), // //, reduce: FieldValue
			reduce(103), // %, reduce: FieldValue
			reduce(103), // .*, reduce: FieldValue
			reduce(103), // ./, reduce: FieldValue
			reduce(103), // .//, reduce: FieldValue
			reduce(103), // ., reduce: FieldValue
			nil,         // !
			nil,         // ~
			reduce(103), // **, reduce: FieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(103), // [, reduce: FieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(103), // [[, reduce: FieldValue
			reduce(103), // [[[, reduce: FieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3634
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(104), // ;, reduce: DirectFieldValue
			nil,         // {
			nil,         // }
			reduce(104), // =, reduce: DirectFieldValue
			nil,         // unset
			nil,         // filter
			reduce(104), // >, reduce: DirectFieldValue
			reduce(104), // >>, reduce: DirectFieldValue
			reduce(104), // |, reduce: DirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(104), // ,, reduce: DirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(104), // ||=, reduce: DirectFieldValue
			reduce(104), // ^^=, reduce: DirectFieldValue
			reduce(104), // &&=, reduce: DirectFieldValue
			reduce(104), // ??=, reduce: DirectFieldValue
			reduce(104), // ???=, reduce: DirectFieldValue
			reduce(104), // |=, reduce: DirectFieldValue
			reduce(104), // &=, reduce: DirectFieldValue
			reduce(104), // ^=, reduce: DirectFieldValue
			reduce(104), // <<=, reduce: DirectFieldValue
			reduce(104), // >>=, reduce: DirectFieldValue
			reduce(104), // >>>=, reduce: DirectFieldValue
			reduce(104), // +=, reduce: DirectFieldValue
			reduce(104), // .=, reduce: DirectFieldValue
			reduce(104), // -=, reduce: DirectFieldValue
			reduce(104), // *=, reduce: DirectFieldValue
			reduce(104), // /=, reduce: DirectFieldValue
			reduce(104), // //=, reduce: DirectFieldValue
			reduce(104), // %=, reduce: DirectFieldValue
			reduce(104), // **=, reduce: DirectFieldValue
			reduce(104), // ?, reduce: DirectFieldValue
			nil,         // :
			reduce(104), // ||, reduce: DirectFieldValue
			reduce(104), // ^^, reduce: DirectFieldValue
			reduce(104), // &&, reduce: DirectFieldValue
			reduce(104), // ??, reduce: DirectFieldValue
			reduce(104), // ???, reduce: DirectFieldValue
			reduce(104), // =~, reduce: DirectFieldValue
			reduce(104), // !=~, reduce: DirectFieldValue
			reduce(104), // ==, reduce: DirectFieldValue
			reduce(104), // !=, reduce: DirectFieldValue
			reduce(104), // <=>, reduce: DirectFieldValue
			reduce(104), // >=, reduce: DirectFieldValue
			reduce(104), // <, reduce: DirectFieldValue
			reduce(104), // <=, reduce: DirectFieldValue
			reduce(104), // ^, reduce: DirectFieldValue
			reduce(104), // &, reduce: DirectFieldValue
			reduce(104), // <<, reduce: DirectFieldValue
			reduce(104), // >>>, reduce: DirectFieldValue
			reduce(104), // +, reduce: DirectFieldValue
			reduce(104), // -, reduce: DirectFieldValue
			reduce(104), // .+, reduce: DirectFieldValue
			reduce(104), // .-, reduce: DirectFieldValue
			reduce(104), // *, reduce: DirectFieldValue
			reduce(104), // /, reduce: DirectFieldValue
			reduce(104), // //, reduce: DirectFieldValue
			reduce(104), // %, reduce: DirectFieldValue
			reduce(104), // .*, reduce: DirectFieldValue
			reduce(104), // ./, reduce: DirectFieldValue
			reduce(104), // .//, reduce: DirectFieldValue
			reduce(104), // ., reduce: DirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(104), // **, reduce: DirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(104), // [, reduce: DirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(104), // [[, reduce: DirectFieldValue
			reduce(104), // [[[, reduce: DirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3635
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3636
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(106), // ;, reduce: BracedFieldValue
			nil,         // {
			nil,         // }
			reduce(106), // =, reduce: BracedFieldValue
			nil,         // unset
			nil,         // filter
			reduce(106), // >, reduce: BracedFieldValue
			reduce(106), // >>, reduce: BracedFieldValue
			reduce(106), // |, reduce: BracedFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(106), // ,, reduce: BracedFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(106), // ||=, reduce: BracedFieldValue
			reduce(106), // ^^=, reduce: BracedFieldValue
			reduce(106), // &&=, reduce: BracedFieldValue
			reduce(106), // ??=, reduce: BracedFieldValue
			reduce(106), // ???=, reduce: BracedFieldValue
			reduce(106), // |=, reduce: BracedFieldValue
			reduce(106), // &=, reduce: BracedFieldValue
			reduce(106), // ^=, reduce: BracedFieldValue
			reduce(106), // <<=, reduce: BracedFieldValue
			reduce(106), // >>=, reduce: BracedFieldValue
			reduce(106), // >>>=, reduce: BracedFieldValue
			reduce(106), // +=, reduce: BracedFieldValue
			reduce(106), // .=, reduce: BracedFieldValue
			reduce(106), // -=, reduce: BracedFieldValue
			reduce(106), // *=, reduce: BracedFieldValue
			reduce(106), // /=, reduce: BracedFieldValue
			reduce(106), // //=, reduce: BracedFieldValue
			reduce(106), // %=, reduce: BracedFieldValue
			reduce(106), // **=, reduce: BracedFieldValue
			reduce(106), // ?, reduce: BracedFieldValue
			nil,         // :
			reduce(106), // ||, reduce: BracedFieldValue
			reduce(106), // ^^, reduce: BracedFieldValue
			reduce(106), // &&, reduce: BracedFieldValue
			reduce(106), // ??, reduce: BracedFieldValue
			reduce(106), // ???, reduce: BracedFieldValue
			reduce(106), // =~, reduce: BracedFieldValue
			reduce(106), // !=~, reduce: BracedFieldValue
			reduce(106), // ==, reduce: BracedFieldValue
			reduce(106), // !=, reduce: BracedFieldValue
			reduce(106), // <=>, reduce: BracedFieldValue
			reduce(106), // >=, reduce: BracedFieldValue
			reduce(106), // <, reduce: BracedFieldValue
			reduce(106), // <=, reduce: BracedFieldValue
			reduce(106), // ^, reduce: BracedFieldValue
			reduce(106), // &, reduce: BracedFieldValue
			reduce(106), // <<, reduce: BracedFieldValue
			reduce(106), // >>>, reduce: BracedFieldValue
			reduce(106), // +, reduce: BracedFieldValue
			reduce(106), // -, reduce: BracedFieldValue
			reduce(106), // .+, reduce: BracedFieldValue
			reduce(106), // .-, reduce: BracedFieldValue
			reduce(106), // *, reduce: BracedFieldValue
			reduce(106), // /, reduce: BracedFieldValue
			reduce(106), // //, reduce: BracedFieldValue
			reduce(106), // %, reduce: BracedFieldValue
			reduce(106), // .*, reduce: BracedFieldValue
			reduce(106), // ./, reduce: BracedFieldValue
			reduce(106), // .//, reduce: BracedFieldValue
			reduce(106), // ., reduce: BracedFieldValue
			nil,         // !
			nil,         // ~
			reduce(106), // **, reduce: BracedFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(106), // [, reduce: BracedFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(106), // [[, reduce: BracedFieldValue
			reduce(106), // [[[, reduce: BracedFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3637
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3638
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3639
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(109), // ;, reduce: FullSrec
			nil,         // {
			nil,         // }
			reduce(109), // =, reduce: FullSrec
			nil,         // unset
			nil,         // filter
			reduce(109), // >, reduce: FullSrec
			reduce(109), // >>, reduce: FullSrec
			reduce(109), // |, reduce: FullSrec
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(109), // ,, reduce: FullSrec
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(109), // ||=, reduce: FullSrec
			reduce(109), // ^^=, reduce: FullSrec
			reduce(109), // &&=, reduce: FullSrec
			reduce(109), // ??=, reduce: FullSrec
			reduce(109), // ???=, reduce: FullSrec
			reduce(109), // |=, reduce: FullSrec
			reduce(109), // &=, reduce: FullSrec
			reduce(109), // ^=, reduce: FullSrec
			reduce(109), // <<=, reduce: FullSrec
			reduce(109), // >>=, reduce: FullSrec
			reduce(109), // >>>=, reduce: FullSrec
			reduce(109), // +=, reduce: FullSrec
			reduce(109), // .=, reduce: FullSrec
			reduce(109), // -=, reduce: FullSrec
			reduce(109), // *=, reduce: FullSrec
			reduce(109), // /=, reduce: FullSrec
			reduce(109), // //=, reduce: FullSrec
			reduce(109), // %=, reduce: FullSrec
			reduce(109), // **=, reduce: FullSrec
			reduce(109), // ?, reduce: FullSrec
			nil,         // :
			reduce(109), // ||, reduce: FullSrec
			reduce(109), // ^^, reduce: FullSrec
			reduce(109), // &&, reduce: FullSrec
			reduce(109), // ??, reduce: FullSrec
			reduce(109), // ???, reduce: FullSrec
			reduce(109), // =~, reduce: FullSrec
			reduce(109), // !=~, reduce: FullSrec
			reduce(109), // ==, reduce: FullSrec
			reduce(109), // !=, reduce: FullSrec
			reduce(109), // <=>, reduce: FullSrec
			reduce(109), // >=, reduce: FullSrec
			reduce(109), // <, reduce: FullSrec
			reduce(109), // <=, reduce: FullSrec
			reduce(109), // ^, reduce: FullSrec
			reduce(109), // &, reduce: FullSrec
			reduce(109), // <<, reduce: FullSrec
			reduce(109), // >>>, reduce: FullSrec
			reduce(109), // +, reduce: FullSrec
			reduce(109), // -, reduce: FullSrec
			reduce(109), // .+, reduce: FullSrec
			reduce(109), // .-, reduce: FullSrec
			reduce(109), // *, reduce: FullSrec
			reduce(109), // /, reduce: FullSrec
			reduce(109), // //, reduce: FullSrec
			reduce(109), // %, reduce: FullSrec
			reduce(109), // .*, reduce: FullSrec
			reduce(109), // ./, reduce: FullSrec
			reduce(109), // .//, reduce: FullSrec
			reduce(109), // ., reduce: FullSrec
			nil,         // !
			nil,         // ~
			reduce(109), // **, reduce: FullSrec
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(109), // [, reduce: FullSrec
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(109), // [[, reduce: FullSrec
			reduce(109), // [[[, reduce: FullSrec
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3640
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(212), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(212), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(212), // >, reduce: MlrvalOrFunction
			reduce(212), // >>, reduce: MlrvalOrFunction
			reduce(212), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(212), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(212), // ||=, reduce: MlrvalOrFunction
			reduce(212), // ^^=, reduce: MlrvalOrFunction
			reduce(212), // &&=, reduce: MlrvalOrFunction
			reduce(212), // ??=, reduce: MlrvalOrFunction
			reduce(212), // ???=, reduce: MlrvalOrFunction
			reduce(212), // |=, reduce: MlrvalOrFunction
			reduce(212), // &=, reduce: MlrvalOrFunction
			reduce(212), // ^=, reduce: MlrvalOrFunction
			reduce(212), // <<=, reduce: MlrvalOrFunction
			reduce(212), // >>=, reduce: MlrvalOrFunction
			reduce(212), // >>>=, reduce: MlrvalOrFunction
			reduce(212), // +=, reduce: MlrvalOrFunction
			reduce(212), // .=, reduce: MlrvalOrFunction
			reduce(212), // -=, reduce: MlrvalOrFunction
			reduce(212), // *=, reduce: MlrvalOrFunction
			reduce(212), // /=, reduce: MlrvalOrFunction
			reduce(212), // //=, reduce: MlrvalOrFunction
			reduce(212), // %=, reduce: MlrvalOrFunction
			reduce(212), // **=, reduce: MlrvalOrFunction
			reduce(212), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(212), // ||, reduce: MlrvalOrFunction
			reduce(212), // ^^, reduce: MlrvalOrFunction
			reduce(212), // &&, reduce: MlrvalOrFunction
			reduce(212), // ??, reduce: MlrvalOrFunction
			reduce(212), // ???, reduce: MlrvalOrFunction
			reduce(212), // =~, reduce: MlrvalOrFunction
			reduce(212), // !=~, reduce: MlrvalOrFunction
			reduce(212), // ==, reduce: MlrvalOrFunction
			reduce(212), // !=, reduce: MlrvalOrFunction
			reduce(212), // <=>, reduce: MlrvalOrFunction
			reduce(212), // >=, reduce: MlrvalOrFunction
			reduce(212), // <, reduce: MlrvalOrFunction
			reduce(212), // <=, reduce: MlrvalOrFunction
			reduce(212), // ^, reduce: MlrvalOrFunction
			reduce(212), // &, reduce: MlrvalOrFunction
			reduce(212), // <<, reduce: MlrvalOrFunction
			reduce(212), // >>>, reduce: MlrvalOrFunction
			reduce(212), // +, reduce: MlrvalOrFunction
			reduce(212), // -, reduce: MlrvalOrFunction
			reduce(212), // .+, reduce: MlrvalOrFunction
			reduce(212), // .-, reduce: MlrvalOrFunction
			reduce(212), // *, reduce: MlrvalOrFunction
			reduce(212), // /, reduce: MlrvalOrFunction
			reduce(212), // //, reduce: MlrvalOrFunction
			reduce(212), // %, reduce: MlrvalOrFunction
			reduce(212), // .*, reduce: MlrvalOrFunction
			reduce(212), // ./, reduce: MlrvalOrFunction
			reduce(212), // .//, reduce: MlrvalOrFunction
			reduce(212), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(212), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(212), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(212), // [[, reduce: MlrvalOrFunction
			reduce(212), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3641
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(113), // ;, reduce: DirectOosvarValue
			nil,         // {
			nil,         // }
			reduce(113), // =, reduce: DirectOosvarValue
			nil,         // unset
			nil,         // filter
			reduce(113), // >, reduce: DirectOosvarValue
			reduce(113), // >>, reduce: DirectOosvarValue
			reduce(113), // |, reduce: DirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(113), // ,, reduce: DirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(113), // ||=, reduce: DirectOosvarValue
			reduce(113), // ^^=, reduce: DirectOosvarValue
			reduce(113), // &&=, reduce: DirectOosvarValue
			reduce(113), // ??=, reduce: DirectOosvarValue
			reduce(113), // ???=, reduce: DirectOosvarValue
			reduce(113), // |=, reduce: DirectOosvarValue
			reduce(113), // &=, reduce: DirectOosvarValue
			reduce(113), // ^=, reduce: DirectOosvarValue
			reduce(113), // <<=, reduce: DirectOosvarValue
			reduce(113), // >>=, reduce: DirectOosvarValue
			reduce(113), // >>>=, reduce: DirectOosvarValue
			reduce(113), // +=, reduce: DirectOosvarValue
			reduce(113), // .=, reduce: DirectOosvarValue
			reduce(113), // -=, reduce: DirectOosvarValue
			reduce(113), // *=, reduce: DirectOosvarValue
			reduce(113), // /=, reduce: DirectOosvarValue
			reduce(113), // //=, reduce: DirectOosvarValue
			reduce(113), // %=, reduce: DirectOosvarValue
			reduce(113), // **=, reduce: DirectOosvarValue
			reduce(113), // ?, reduce: DirectOosvarValue
			nil,         // :
			reduce(113), // ||, reduce: DirectOosvarValue
			reduce(113), // ^^, reduce: DirectOosvarValue
			reduce(113), // &&, reduce: DirectOosvarValue
			reduce(113), // ??, reduce: DirectOosvarValue
			reduce(113), // ???, reduce: DirectOosvarValue
			reduce(113), // =~, reduce: DirectOosvarValue
			reduce(113), // !=~, reduce: DirectOosvarValue
			reduce(113), // ==, reduce: DirectOosvarValue
			reduce(113), // !=, reduce: DirectOosvarValue
			reduce(113), // <=>, reduce: DirectOosvarValue
			reduce(113), // >=, reduce: DirectOosvarValue
			reduce(113), // <, reduce: DirectOosvarValue
			reduce(113), // <=, reduce: DirectOosvarValue
			reduce(113), // ^, reduce: DirectOosvarValue
			reduce(113), // &, reduce: DirectOosvarValue
			reduce(113), // <<, reduce: DirectOosvarValue
			reduce(113), // >>>, reduce: DirectOosvarValue
			reduce(113), // +, reduce: DirectOosvarValue
			reduce(113), // -, reduce: DirectOosvarValue
			reduce(113), // .+, reduce: DirectOosvarValue
			reduce(113), // .-, reduce: DirectOosvarValue
			reduce(113), // *, reduce: DirectOosvarValue
			reduce(113), // /, reduce: DirectOosvarValue
			reduce(113), // //, reduce: DirectOosvarValue
			reduce(113), // %, reduce: DirectOosvarValue
			reduce(113), // .*, reduce: DirectOosvarValue
			reduce(113), // ./, reduce: DirectOosvarValue
			reduce(113), // .//, reduce: DirectOosvarValue
			reduce(113), // ., reduce: DirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(113), // **, reduce: DirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(113), // [, reduce: DirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(113), // [[, reduce: DirectOosvarValue
			reduce(113), // [[[, reduce: DirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3642
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3643
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(115), // ;, reduce: BracedOosvarValue
			nil,         // {
			nil,         // }
			reduce(115), // =, reduce: BracedOosvarValue
			nil,         // unset
			nil,         // filter
			reduce(115), // >, reduce: BracedOosvarValue
			reduce(115), // >>, reduce: BracedOosvarValue
			reduce(115), // |, reduce: BracedOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(115), // ,, reduce: BracedOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(115), // ||=, reduce: BracedOosvarValue
			reduce(115), // ^^=, reduce: BracedOosvarValue
			reduce(115), // &&=, reduce: BracedOosvarValue
			reduce(115), // ??=, reduce: BracedOosvarValue
			reduce(115), // ???=, reduce: BracedOosvarValue
			reduce(115), // |=, reduce: BracedOosvarValue
			reduce(115), // &=, reduce: BracedOosvarValue
			reduce(115), // ^=, reduce: BracedOosvarValue
			reduce(115), // <<=, reduce: BracedOosvarValue
			reduce(115), // >>=, reduce: BracedOosvarValue
			reduce(115), // >>>=, reduce: BracedOosvarValue
			reduce(115), // +=, reduce: BracedOosvarValue
			reduce(115), // .=, reduce: BracedOosvarValue
			reduce(115), // -=, reduce: BracedOosvarValue
			reduce(115), // *=, reduce: BracedOosvarValue
			reduce(115), // /=, reduce: BracedOosvarValue
			reduce(115), // //=, reduce: BracedOosvarValue
			reduce(115), // %=, reduce: BracedOosvarValue
			reduce(115), // **=, reduce: BracedOosvarValue
			reduce(115), // ?, reduce: BracedOosvarValue
			nil,         // :
			reduce(115), // ||, reduce: BracedOosvarValue
			reduce(115), // ^^, reduce: BracedOosvarValue
			reduce(115), // &&, reduce: BracedOosvarValue
			reduce(115), // ??, reduce: BracedOosvarValue
			reduce(115), // ???, reduce: BracedOosvarValue
			reduce(115), // =~, reduce: BracedOosvarValue
			reduce(115), // !=~, reduce: BracedOosvarValue
			reduce(115), // ==, reduce: BracedOosvarValue
			reduce(115), // !=, reduce: BracedOosvarValue
			reduce(115), // <=>, reduce: BracedOosvarValue
			reduce(115), // >=, reduce: BracedOosvarValue
			reduce(115), // <, reduce: BracedOosvarValue
			reduce(115), // <=, reduce: BracedOosvarValue
			reduce(115), // ^, reduce: BracedOosvarValue
			reduce(115), // &, reduce: BracedOosvarValue
			reduce(115), // <<, reduce: BracedOosvarValue
			reduce(115), // >>>, reduce: BracedOosvarValue
			reduce(115), // +, reduce: BracedOosvarValue
			reduce(115), // -, reduce: BracedOosvarValue
			reduce(115), // .+, reduce: BracedOosvarValue
			reduce(115), // .-, reduce: BracedOosvarValue
			reduce(115), // *, reduce: BracedOosvarValue
			reduce(115), // /, reduce: BracedOosvarValue
			reduce(115), // //, reduce: BracedOosvarValue
			reduce(115), // %, reduce: BracedOosvarValue
			reduce(115), // .*, reduce: BracedOosvarValue
			reduce(115), // ./, reduce: BracedOosvarValue
			reduce(115), // .//, reduce: BracedOosvarValue
			reduce(115), // ., reduce: BracedOosvarValue
			nil,         // !
			nil,         // ~
			reduce(115), // **, reduce: BracedOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(115), // [, reduce: BracedOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(115), // [[, reduce: BracedOosvarValue
			reduce(115), // [[[, reduce: BracedOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3644
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(116), // ;, reduce: FullOosvar
			nil,         // {
			nil,         // }
			reduce(116), // =, reduce: FullOosvar
			nil,         // unset
			nil,         // filter
			reduce(116), // >, reduce: FullOosvar
			reduce(116), // >>, reduce: FullOosvar
			reduce(116), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(116), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(116), // ||=, reduce: FullOosvar
			reduce(116), // ^^=, reduce: FullOosvar
			reduce(116), // &&=, reduce: FullOosvar
			reduce(116), // ??=, reduce: FullOosvar
			reduce(116), // ???=, reduce: FullOosvar
			reduce(116), // |=, reduce: FullOosvar
			reduce(116), // &=, reduce: FullOosvar
			reduce(116), // ^=, reduce: FullOosvar
			reduce(116), // <<=, reduce: FullOosvar
			reduce(116), // >>=, reduce: FullOosvar
			reduce(116), // >>>=, reduce: FullOosvar
			reduce(116), // +=, reduce: FullOosvar
			reduce(116), // .=, reduce: FullOosvar
			reduce(116), // -=, reduce: FullOosvar
			reduce(116), // *=, reduce: FullOosvar
			reduce(116), // /=, reduce: FullOosvar
			reduce(116), // //=, reduce: FullOosvar
			reduce(116), // %=, reduce: FullOosvar
			reduce(116), // **=, reduce: FullOosvar
			reduce(116), // ?, reduce: FullOosvar
			nil,         // :
			reduce(116), // ||, reduce: FullOosvar
			reduce(116), // ^^, reduce: FullOosvar
			reduce(116), // &&, reduce: FullOosvar
			reduce(116), // ??, reduce: FullOosvar
			reduce(116), // ???, reduce: FullOosvar
			reduce(116), // =~, reduce: FullOosvar
			reduce(116), // !=~, reduce: FullOosvar
			reduce(116), // ==, reduce: FullOosvar
			reduce(116), // !=, reduce: FullOosvar
			reduce(116), // <=>, reduce: FullOosvar
			reduce(116), // >=, reduce: FullOosvar
			reduce(116), // <, reduce: FullOosvar
			reduce(116), // <=, reduce: FullOosvar
			reduce(116), // ^, reduce: FullOosvar
			reduce(116), // &, reduce: FullOosvar
			reduce(116), // <<, reduce: FullOosvar
			reduce(116), // >>>, reduce: FullOosvar
			reduce(116), // +, reduce: FullOosvar
			reduce(116), // -, reduce: FullOosvar
			reduce(116), // .+, reduce: FullOosvar
			reduce(116), // .-, reduce: FullOosvar
			reduce(116), // *, reduce: FullOosvar
			reduce(116), // /, reduce: FullOosvar
			reduce(116), // //, reduce: FullOosvar
			reduce(116), // %, reduce: FullOosvar
			reduce(116), // .*, reduce: FullOosvar
			reduce(116), // ./, reduce: FullOosvar
			reduce(116), // .//, reduce: FullOosvar
			reduce(116), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(116), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(116), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(116), // [[, reduce: FullOosvar
			reduce(116), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3645
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(117), // ;, reduce: FullOosvar
			nil,         // {
			nil,         // }
			reduce(117), // =, reduce: FullOosvar
			nil,         // unset
			nil,         // filter
			reduce(117), // >, reduce: FullOosvar
			reduce(117), // >>, reduce: FullOosvar
			reduce(117), // |, reduce: FullOosvar
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(117), // ,, reduce: FullOosvar
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(117), // ||=, reduce: FullOosvar
			reduce(117), // ^^=, reduce: FullOosvar
			reduce(117), // &&=, reduce: FullOosvar
			reduce(117), // ??=, reduce: FullOosvar
			reduce(117), // ???=, reduce: FullOosvar
			reduce(117), // |=, reduce: FullOosvar
			reduce(117), // &=, reduce: FullOosvar
			reduce(117), // ^=, reduce: FullOosvar
			reduce(117), // <<=, reduce: FullOosvar
			reduce(117), // >>=, reduce: FullOosvar
			reduce(117), // >>>=, reduce: FullOosvar
			reduce(117), // +=, reduce: FullOosvar
			reduce(117), // .=, reduce: FullOosvar
			reduce(117), // -=, reduce: FullOosvar
			reduce(117), // *=, reduce: FullOosvar
			reduce(117), // /=, reduce: FullOosvar
			reduce(117), // //=, reduce: FullOosvar
			reduce(117), // %=, reduce: FullOosvar
			reduce(117), // **=, reduce: FullOosvar
			reduce(117), // ?, reduce: FullOosvar
			nil,         // :
			reduce(117), // ||, reduce: FullOosvar
			reduce(117), // ^^, reduce: FullOosvar
			reduce(117), // &&, reduce: FullOosvar
			reduce(117), // ??, reduce: FullOosvar
			reduce(117), // ???, reduce: FullOosvar
			reduce(117), // =~, reduce: FullOosvar
			reduce(117), // !=~, reduce: FullOosvar
			reduce(117), // ==, reduce: FullOosvar
			reduce(117), // !=, reduce: FullOosvar
			reduce(117), // <=>, reduce: FullOosvar
			reduce(117), // >=, reduce: FullOosvar
			reduce(117), // <, reduce: FullOosvar
			reduce(117), // <=, reduce: FullOosvar
			reduce(117), // ^, reduce: FullOosvar
			reduce(117), // &, reduce: FullOosvar
			reduce(117), // <<, reduce: FullOosvar
			reduce(117), // >>>, reduce: FullOosvar
			reduce(117), // +, reduce: FullOosvar
			reduce(117), // -, reduce: FullOosvar
			reduce(117), // .+, reduce: FullOosvar
			reduce(117), // .-, reduce: FullOosvar
			reduce(117), // *, reduce: FullOosvar
			reduce(117), // /, reduce: FullOosvar
			reduce(117), // //, reduce: FullOosvar
			reduce(117), // %, reduce: FullOosvar
			reduce(117), // .*, reduce: FullOosvar
			reduce(117), // ./, reduce: FullOosvar
			reduce(117), // .//, reduce: FullOosvar
			reduce(117), // ., reduce: FullOosvar
			nil,         // !
			nil,         // ~
			reduce(117), // **, reduce: FullOosvar
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(117), // [, reduce: FullOosvar
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(117), // [[, reduce: FullOosvar
			reduce(117), // [[[, reduce: FullOosvar
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3646
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(118), // ;, reduce: LocalVariable
			nil,         // {
			nil,         // }
			reduce(118), // =, reduce: LocalVariable
			nil,         // unset
			nil,         // filter
			reduce(118), // >, reduce: LocalVariable
			reduce(118), // >>, reduce: LocalVariable
			reduce(118), // |, reduce: LocalVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(118), // ,, reduce: LocalVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			reduce(270), // (, reduce: FunctionName
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(118), // ||=, reduce: LocalVariable
			reduce(118), // ^^=, reduce: LocalVariable
			reduce(118), // &&=, reduce: LocalVariable
			reduce(118), // ??=, reduce: LocalVariable
			reduce(118), // ???=, reduce: LocalVariable
			reduce(118), // |=, reduce: LocalVariable
			reduce(118), // &=, reduce: LocalVariable
			reduce(118), // ^=, reduce: LocalVariable
			reduce(118), // <<=, reduce: LocalVariable
			reduce(118), // >>=, reduce: LocalVariable
			reduce(118), // >>>=, reduce: LocalVariable
			reduce(118), // +=, reduce: LocalVariable
			reduce(118), // .=, reduce: LocalVariable
			reduce(118), // -=, reduce: LocalVariable
			reduce(118), // *=, reduce: LocalVariable
			reduce(118), // /=, reduce: LocalVariable
			reduce(118), // //=, reduce: LocalVariable
			reduce(118), // %=, reduce: LocalVariable
			reduce(118), // **=, reduce: LocalVariable
			reduce(118), // ?, reduce: LocalVariable
			nil,         // :
			reduce(118), // ||, reduce: LocalVariable
			reduce(118), // ^^, reduce: LocalVariable
			reduce(118), // &&, reduce: LocalVariable
			reduce(118), // ??, reduce: LocalVariable
			reduce(118), // ???, reduce: LocalVariable
			reduce(118), // =~, reduce: LocalVariable
			reduce(118), // !=~, reduce: LocalVariable
			reduce(118), // ==, reduce: LocalVariable
			reduce(118), // !=, reduce: LocalVariable
			reduce(118), // <=>, reduce: LocalVariable
			reduce(118), // >=, reduce: LocalVariable
			reduce(118), // <, reduce: LocalVariable
			reduce(118), // <=, reduce: LocalVariable
			reduce(118), // ^, reduce: LocalVariable
			reduce(118), // &, reduce: LocalVariable
			reduce(118), // <<, reduce: LocalVariable
			reduce(118), // >>>, reduce: LocalVariable
			reduce(118), // +, reduce: LocalVariable
			reduce(118), // -, reduce: LocalVariable
			reduce(118), // .+, reduce: LocalVariable
			reduce(118), // .-, reduce: LocalVariable
			reduce(118), // *, reduce: LocalVariable
			reduce(118), // /, reduce: LocalVariable
			reduce(118), // //, reduce: LocalVariable
			reduce(118), // %, reduce: LocalVariable
			reduce(118), // .*, reduce: LocalVariable
			reduce(118), // ./, reduce: LocalVariable
			reduce(118), // .//, reduce: LocalVariable
			reduce(118), // ., reduce: LocalVariable
			nil,         // !
			nil,         // ~
			reduce(118), // **, reduce: LocalVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(118), // [, reduce: LocalVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(118), // [[, reduce: LocalVariable
			reduce(118), // [[[, reduce: LocalVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3647
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(147), // ;, reduce: Rvalue
			nil,         // {
			nil,         // }
			reduce(147), // =, reduce: Rvalue
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(147), // ,, reduce: Rvalue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(147), // ||=, reduce: Rvalue
			reduce(147), // ^^=, reduce: Rvalue
			reduce(147), // &&=, reduce: Rvalue
			reduce(147), // ??=, reduce: Rvalue
			reduce(147), // ???=, reduce: Rvalue
			reduce(147), // |=, reduce: Rvalue
			reduce(147), // &=, reduce: Rvalue
			reduce(147), // ^=, reduce: Rvalue
			reduce(147), // <<=, reduce: Rvalue
			reduce(147), // >>=, reduce: Rvalue
			reduce(147), // >>>=, reduce: Rvalue
			reduce(147), // +=, reduce: Rvalue
			reduce(147), // .=, reduce: Rvalue
			reduce(147), // -=, reduce: Rvalue
			reduce(147), // *=, reduce: Rvalue
			reduce(147), // /=, reduce: Rvalue
			reduce(147), // //=, reduce: Rvalue
			reduce(147), // %=, reduce: Rvalue
			reduce(147), // **=, reduce: Rvalue
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3648
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(148), // ;, reduce: PrecedenceChainStart
			nil,         // {
			nil,         // }
			reduce(148), // =, reduce: PrecedenceChainStart
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(148), // ,, reduce: PrecedenceChainStart
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(148), // ||=, reduce: PrecedenceChainStart
			reduce(148), // ^^=, reduce: PrecedenceChainStart
			reduce(148), // &&=, reduce: PrecedenceChainStart
			reduce(148), // ??=, reduce: PrecedenceChainStart
			reduce(148), // ???=, reduce: PrecedenceChainStart
			reduce(148), // |=, reduce: PrecedenceChainStart
			reduce(148), // &=, reduce: PrecedenceChainStart
			reduce(148), // ^=, reduce: PrecedenceChainStart
			reduce(148), // <<=, reduce: PrecedenceChainStart
			reduce(148), // >>=, reduce: PrecedenceChainStart
			reduce(148), // >>>=, reduce: PrecedenceChainStart
			reduce(148), // +=, reduce: PrecedenceChainStart
			reduce(148), // .=, reduce: PrecedenceChainStart
			reduce(148), // -=, reduce: PrecedenceChainStart
			reduce(148), // *=, reduce: PrecedenceChainStart
			reduce(148), // /=, reduce: PrecedenceChainStart
			reduce(148), // //=, reduce: PrecedenceChainStart
			reduce(148), // %=, reduce: PrecedenceChainStart
			reduce(148), // **=, reduce: PrecedenceChainStart
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3649
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(150), // ;, reduce: TernaryTerm
			nil,         // {
			nil,         // }
			reduce(150), // =, reduce: TernaryTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(150), // ,, reduce: TernaryTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(150), // ||=, reduce: TernaryTerm
			reduce(150), // ^^=, reduce: TernaryTerm
			reduce(150), // &&=, reduce: TernaryTerm
			reduce(150), // ??=, reduce: TernaryTerm
			reduce(150), // ???=, reduce: TernaryTerm
			reduce(150), // |=, reduce: TernaryTerm
			reduce(150), // &=, reduce: TernaryTerm
			reduce(150), // ^=, reduce: TernaryTerm
			reduce(150), // <<=, reduce: TernaryTerm
			reduce(150), // >>=, reduce: TernaryTerm
			reduce(150), // >>>=, reduce: TernaryTerm
			reduce(150), // +=, reduce: TernaryTerm
			reduce(150), // .=, reduce: TernaryTerm
			reduce(150), // -=, reduce: TernaryTerm
			reduce(150), // *=, reduce: TernaryTerm
			reduce(150), // /=, reduce: TernaryTerm
			reduce(150), // //=, reduce: TernaryTerm
			reduce(150), // %=, reduce: TernaryTerm
			reduce(150), // **=, reduce: TernaryTerm
			shift(4118), // ?
			nil,         // :
			shift(4119), // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3650
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(152), // ;, reduce: LogicalOrTerm
			nil,         // {
			nil,         // }
			reduce(152), // =, reduce: LogicalOrTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(152), // ,, reduce: LogicalOrTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(152), // ||=, reduce: LogicalOrTerm
			reduce(152), // ^^=, reduce: LogicalOrTerm
			reduce(152), // &&=, reduce: LogicalOrTerm
			reduce(152), // ??=, reduce: LogicalOrTerm
			reduce(152), // ???=, reduce: LogicalOrTerm
			reduce(152), // |=, reduce: LogicalOrTerm
			reduce(152), // &=, reduce: LogicalOrTerm
			reduce(152), // ^=, reduce: LogicalOrTerm
			reduce(152), // <<=, reduce: LogicalOrTerm
			reduce(152), // >>=, reduce: LogicalOrTerm
			reduce(152), // >>>=, reduce: LogicalOrTerm
			reduce(152), // +=, reduce: LogicalOrTerm
			reduce(152), // .=, reduce: LogicalOrTerm
			reduce(152), // -=, reduce: LogicalOrTerm
			reduce(152), // *=, reduce: LogicalOrTerm
			reduce(152), // /=, reduce: LogicalOrTerm
			reduce(152), // //=, reduce: LogicalOrTerm
			reduce(152), // %=, reduce: LogicalOrTerm
			reduce(152), // **=, reduce: LogicalOrTerm
			reduce(152), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(152), // ||, reduce: LogicalOrTerm
			shift(4120), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3651
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(154), // ;, reduce: LogicalXORTerm
			nil,         // {
			nil,         // }
			reduce(154), // =, reduce: LogicalXORTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(154), // ,, reduce: LogicalXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(154), // ||=, reduce: LogicalXORTerm
			reduce(154), // ^^=, reduce: LogicalXORTerm
			reduce(154), // &&=, reduce: LogicalXORTerm
			reduce(154), // ??=, reduce: LogicalXORTerm
			reduce(154), // ???=, reduce: LogicalXORTerm
			reduce(154), // |=, reduce: LogicalXORTerm
			reduce(154), // &=, reduce: LogicalXORTerm
			reduce(154), // ^=, reduce: LogicalXORTerm
			reduce(154), // <<=, reduce: LogicalXORTerm
			reduce(154), // >>=, reduce: LogicalXORTerm
			reduce(154), // >>>=, reduce: LogicalXORTerm
			reduce(154), // +=, reduce: LogicalXORTerm
			reduce(154), // .=, reduce: LogicalXORTerm
			reduce(154), // -=, reduce: LogicalXORTerm
			reduce(154), // *=, reduce: LogicalXORTerm
			reduce(154), // /=, reduce: LogicalXORTerm
			reduce(154), // //=, reduce: LogicalXORTerm
			reduce(154), // %=, reduce: LogicalXORTerm
			reduce(154), // **=, reduce: LogicalXORTerm
			reduce(154), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(154), // ||, reduce: LogicalXORTerm
			reduce(154), // ^^, reduce: LogicalXORTerm
			shift(4121), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3652
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(156), // ;, reduce: LogicalAndTerm
			nil,         // {
			nil,         // }
			reduce(156), // =, reduce: LogicalAndTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(156), // ,, reduce: LogicalAndTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(156), // ||=, reduce: LogicalAndTerm
			reduce(156), // ^^=, reduce: LogicalAndTerm
			reduce(156), // &&=, reduce: LogicalAndTerm
			reduce(156), // ??=, reduce: LogicalAndTerm
			reduce(156), // ???=, reduce: LogicalAndTerm
			reduce(156), // |=, reduce: LogicalAndTerm
			reduce(156), // &=, reduce: LogicalAndTerm
			reduce(156), // ^=, reduce: LogicalAndTerm
			reduce(156), // <<=, reduce: LogicalAndTerm
			reduce(156), // >>=, reduce: LogicalAndTerm
			reduce(156), // >>>=, reduce: LogicalAndTerm
			reduce(156), // +=, reduce: LogicalAndTerm
			reduce(156), // .=, reduce: LogicalAndTerm
			reduce(156), // -=, reduce: LogicalAndTerm
			reduce(156), // *=, reduce: LogicalAndTerm
			reduce(156), // /=, reduce: LogicalAndTerm
			reduce(156), // //=, reduce: LogicalAndTerm
			reduce(156), // %=, reduce: LogicalAndTerm
			reduce(156), // **=, reduce: LogicalAndTerm
			reduce(156), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(156), // ||, reduce: LogicalAndTerm
			reduce(156), // ^^, reduce: LogicalAndTerm
			reduce(156), // &&, reduce: LogicalAndTerm
			shift(4122), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3653
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(158), // ;, reduce: AbsentCoalesceTerm
			nil,         // {
			nil,         // }
			reduce(158), // =, reduce: AbsentCoalesceTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(158), // ,, reduce: AbsentCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(158), // ||=, reduce: AbsentCoalesceTerm
			reduce(158), // ^^=, reduce: AbsentCoalesceTerm
			reduce(158), // &&=, reduce: AbsentCoalesceTerm
			reduce(158), // ??=, reduce: AbsentCoalesceTerm
			reduce(158), // ???=, reduce: AbsentCoalesceTerm
			reduce(158), // |=, reduce: AbsentCoalesceTerm
			reduce(158), // &=, reduce: AbsentCoalesceTerm
			reduce(158), // ^=, reduce: AbsentCoalesceTerm
			reduce(158), // <<=, reduce: AbsentCoalesceTerm
			reduce(158), // >>=, reduce: AbsentCoalesceTerm
			reduce(158), // >>>=, reduce: AbsentCoalesceTerm
			reduce(158), // +=, reduce: AbsentCoalesceTerm
			reduce(158), // .=, reduce: AbsentCoalesceTerm
			reduce(158), // -=, reduce: AbsentCoalesceTerm
			reduce(158), // *=, reduce: AbsentCoalesceTerm
			reduce(158), // /=, reduce: AbsentCoalesceTerm
			reduce(158), // //=, reduce: AbsentCoalesceTerm
			reduce(158), // %=, reduce: AbsentCoalesceTerm
			reduce(158), // **=, reduce: AbsentCoalesceTerm
			reduce(158), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(158), // ||, reduce: AbsentCoalesceTerm
			reduce(158), // ^^, reduce: AbsentCoalesceTerm
			reduce(158), // &&, reduce: AbsentCoalesceTerm
			reduce(158), // ??, reduce: AbsentCoalesceTerm
			shift(4123), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3654
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(160), // ;, reduce: EmptyCoalesceTerm
			nil,         // {
			nil,         // }
			reduce(160), // =, reduce: EmptyCoalesceTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(160), // ,, reduce: EmptyCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(160), // ||=, reduce: EmptyCoalesceTerm
			reduce(160), // ^^=, reduce: EmptyCoalesceTerm
			reduce(160), // &&=, reduce: EmptyCoalesceTerm
			reduce(160), // ??=, reduce: EmptyCoalesceTerm
			reduce(160), // ???=, reduce: EmptyCoalesceTerm
			reduce(160), // |=, reduce: EmptyCoalesceTerm
			reduce(160), // &=, reduce: EmptyCoalesceTerm
			reduce(160), // ^=, reduce: EmptyCoalesceTerm
			reduce(160), // <<=, reduce: EmptyCoalesceTerm
			reduce(160), // >>=, reduce: EmptyCoalesceTerm
			reduce(160), // >>>=, reduce: EmptyCoalesceTerm
			reduce(160), // +=, reduce: EmptyCoalesceTerm
			reduce(160), // .=, reduce: EmptyCoalesceTerm
			reduce(160), // -=, reduce: EmptyCoalesceTerm
			reduce(160), // *=, reduce: EmptyCoalesceTerm
			reduce(160), // /=, reduce: EmptyCoalesceTerm
			reduce(160), // //=, reduce: EmptyCoalesceTerm
			reduce(160), // %=, reduce: EmptyCoalesceTerm
			reduce(160), // **=, reduce: EmptyCoalesceTerm
			reduce(160), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(160), // ||, reduce: EmptyCoalesceTerm
			reduce(160), // ^^, reduce: EmptyCoalesceTerm
			reduce(160), // &&, reduce: EmptyCoalesceTerm
			reduce(160), // ??, reduce: EmptyCoalesceTerm
			reduce(160), // ???, reduce: EmptyCoalesceTerm
			shift(4124), // =~
			shift(4125), // !=~
			shift(4126), // ==
			shift(4127), // !=
			shift(4128), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3655
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(166), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			reduce(166), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(4129), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(166), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(166), // ||=, reduce: EqneTerm
			reduce(166), // ^^=, reduce: EqneTerm
			reduce(166), // &&=, reduce: EqneTerm
			reduce(166), // ??=, reduce: EqneTerm
			reduce(166), // ???=, reduce: EqneTerm
			reduce(166), // |=, reduce: EqneTerm
			reduce(166), // &=, reduce: EqneTerm
			reduce(166), // ^=, reduce: EqneTerm
			reduce(166), // <<=, reduce: EqneTerm
			reduce(166), // >>=, reduce: EqneTerm
			reduce(166), // >>>=, reduce: EqneTerm
			reduce(166), // +=, reduce: EqneTerm
			reduce(166), // .=, reduce: EqneTerm
			reduce(166), // -=, reduce: EqneTerm
			reduce(166), // *=, reduce: EqneTerm
			reduce(166), // /=, reduce: EqneTerm
			reduce(166), // //=, reduce: EqneTerm
			reduce(166), // %=, reduce: EqneTerm
			reduce(166), // **=, reduce: EqneTerm
			reduce(166), // ?, reduce: EqneTerm
			nil,         // :
			reduce(166), // ||, reduce: EqneTerm
			reduce(166), // ^^, reduce: EqneTerm
			reduce(166), // &&, reduce: EqneTerm
			reduce(166), // ??, reduce: EqneTerm
			reduce(166), // ???, reduce: EqneTerm
			reduce(166), // =~, reduce: EqneTerm
			reduce(166), // !=~, reduce: EqneTerm
			reduce(166), // ==, reduce: EqneTerm
			reduce(166), // !=, reduce: EqneTerm
			reduce(166), // <=>, reduce: EqneTerm
			shift(4130), // >=
			shift(4131), // <
			shift(4132), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3656
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(171), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			reduce(171), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(171), // >, reduce: CmpTerm
			nil,         // >>
			shift(4133), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(171), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(171), // ||=, reduce: CmpTerm
			reduce(171), // ^^=, reduce: CmpTerm
			reduce(171), // &&=, reduce: CmpTerm
			reduce(171), // ??=, reduce: CmpTerm
			reduce(171), // ???=, reduce: CmpTerm
			reduce(171), // |=, reduce: CmpTerm
			reduce(171), // &=, reduce: CmpTerm
			reduce(171), // ^=, reduce: CmpTerm
			reduce(171), // <<=, reduce: CmpTerm
			reduce(171), // >>=, reduce: CmpTerm
			reduce(171), // >>>=, reduce: CmpTerm
			reduce(171), // +=, reduce: CmpTerm
			reduce(171), // .=, reduce: CmpTerm
			reduce(171), // -=, reduce: CmpTerm
			reduce(171), // *=, reduce: CmpTerm
			reduce(171), // /=, reduce: CmpTerm
			reduce(171), // //=, reduce: CmpTerm
			reduce(171), // %=, reduce: CmpTerm
			reduce(171), // **=, reduce: CmpTerm
			reduce(171), // ?, reduce: CmpTerm
			nil,         // :
			reduce(171), // ||, reduce: CmpTerm
			reduce(171), // ^^, reduce: CmpTerm
			reduce(171), // &&, reduce: CmpTerm
			reduce(171), // ??, reduce: CmpTerm
			reduce(171), // ???, reduce: CmpTerm
			reduce(171), // =~, reduce: CmpTerm
			reduce(171), // !=~, reduce: CmpTerm
			reduce(171), // ==, reduce: CmpTerm
			reduce(171), // !=, reduce: CmpTerm
			reduce(171), // <=>, reduce: CmpTerm
			reduce(171), // >=, reduce: CmpTerm
			reduce(171), // <, reduce: CmpTerm
			reduce(171), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3657
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(173), // ;, reduce: BitwiseORTerm
			nil,         // {
			nil,         // }
			reduce(173), // =, reduce: BitwiseORTerm
			nil,         // unset
			nil,         // filter
			reduce(173), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(173), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(173), // ,, reduce: BitwiseORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(173), // ||=, reduce: BitwiseORTerm
			reduce(173), // ^^=, reduce: BitwiseORTerm
			reduce(173), // &&=, reduce: BitwiseORTerm
			reduce(173), // ??=, reduce: BitwiseORTerm
			reduce(173), // ???=, reduce: BitwiseORTerm
			reduce(173), // |=, reduce: BitwiseORTerm
			reduce(173), // &=, reduce: BitwiseORTerm
			reduce(173), // ^=, reduce: BitwiseORTerm
			reduce(173), // <<=, reduce: BitwiseORTerm
			reduce(173), // >>=, reduce: BitwiseORTerm
			reduce(173), // >>>=, reduce: BitwiseORTerm
			reduce(173), // +=, reduce: BitwiseORTerm
			reduce(173), // .=, reduce: BitwiseORTerm
			reduce(173), // -=, reduce: BitwiseORTerm
			reduce(173), // *=, reduce: BitwiseORTerm
			reduce(173), // /=, reduce: BitwiseORTerm
			reduce(173), // //=, reduce: BitwiseORTerm
			reduce(173), // %=, reduce: BitwiseORTerm
			reduce(173), // **=, reduce: BitwiseORTerm
			reduce(173), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(173), // ||, reduce: BitwiseORTerm
			reduce(173), // ^^, reduce: BitwiseORTerm
			reduce(173), // &&, reduce: BitwiseORTerm
			reduce(173), // ??, reduce: BitwiseORTerm
			reduce(173), // ???, reduce: BitwiseORTerm
			reduce(173), // =~, reduce: BitwiseORTerm
			reduce(173), // !=~, reduce: BitwiseORTerm
			reduce(173), // ==, reduce: BitwiseORTerm
			reduce(173), // !=, reduce: BitwiseORTerm
			reduce(173), // <=>, reduce: BitwiseORTerm
			reduce(173), // >=, reduce: BitwiseORTerm
			reduce(173), // <, reduce: BitwiseORTerm
			reduce(173), // <=, reduce: BitwiseORTerm
			shift(4134), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3658
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(175), // ;, reduce: BitwiseXORTerm
			nil,         // {
			nil,         // }
			reduce(175), // =, reduce: BitwiseXORTerm
			nil,         // unset
			nil,         // filter
			reduce(175), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(175), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(175), // ,, reduce: BitwiseXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(175), // ||=, reduce: BitwiseXORTerm
			reduce(175), // ^^=, reduce: BitwiseXORTerm
			reduce(175), // &&=, reduce: BitwiseXORTerm
			reduce(175), // ??=, reduce: BitwiseXORTerm
			reduce(175), // ???=, reduce: BitwiseXORTerm
			reduce(175), // |=, reduce: BitwiseXORTerm
			reduce(175), // &=, reduce: BitwiseXORTerm
			reduce(175), // ^=, reduce: BitwiseXORTerm
			reduce(175), // <<=, reduce: BitwiseXORTerm
			reduce(175), // >>=, reduce: BitwiseXORTerm
			reduce(175), // >>>=, reduce: BitwiseXORTerm
			reduce(175), // +=, reduce: BitwiseXORTerm
			reduce(175), // .=, reduce: BitwiseXORTerm
			reduce(175), // -=, reduce: BitwiseXORTerm
			reduce(175), // *=, reduce: BitwiseXORTerm
			reduce(175), // /=, reduce: BitwiseXORTerm
			reduce(175), // //=, reduce: BitwiseXORTerm
			reduce(175), // %=, reduce: BitwiseXORTerm
			reduce(175), // **=, reduce: BitwiseXORTerm
			reduce(175), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(175), // ||, reduce: BitwiseXORTerm
			reduce(175), // ^^, reduce: BitwiseXORTerm
			reduce(175), // &&, reduce: BitwiseXORTerm
			reduce(175), // ??, reduce: BitwiseXORTerm
			reduce(175), // ???, reduce: BitwiseXORTerm
			reduce(175), // =~, reduce: BitwiseXORTerm
			reduce(175), // !=~, reduce: BitwiseXORTerm
			reduce(175), // ==, reduce: BitwiseXORTerm
			reduce(175), // !=, reduce: BitwiseXORTerm
			reduce(175), // <=>, reduce: BitwiseXORTerm
			reduce(175), // >=, reduce: BitwiseXORTerm
			reduce(175), // <, reduce: BitwiseXORTerm
			reduce(175), // <=, reduce: BitwiseXORTerm
			reduce(175), // ^, reduce: BitwiseXORTerm
			shift(4135), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3659
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(177), // ;, reduce: BitwiseANDTerm
			nil,         // {
			nil,         // }
			reduce(177), // =, reduce: BitwiseANDTerm
			nil,         // unset
			nil,         // filter
			reduce(177), // >, reduce: BitwiseANDTerm
			shift(4136), // >>
			reduce(177), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(177), // ,, reduce: BitwiseANDTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(177), // ||=, reduce: BitwiseANDTerm
			reduce(177), // ^^=, reduce: BitwiseANDTerm
			reduce(177), // &&=, reduce: BitwiseANDTerm
			reduce(177), // ??=, reduce: BitwiseANDTerm
			reduce(177), // ???=, reduce: BitwiseANDTerm
			reduce(177), // |=, reduce: BitwiseANDTerm
			reduce(177), // &=, reduce: BitwiseANDTerm
			reduce(177), // ^=, reduce: BitwiseANDTerm
			reduce(177), // <<=, reduce: BitwiseANDTerm
			reduce(177), // >>=, reduce: BitwiseANDTerm
			reduce(177), // >>>=, reduce: BitwiseANDTerm
			reduce(177), // +=, reduce: BitwiseANDTerm
			reduce(177), // .=, reduce: BitwiseANDTerm
			reduce(177), // -=, reduce: BitwiseANDTerm
			reduce(177), // *=, reduce: BitwiseANDTerm
			reduce(177), // /=, reduce: BitwiseANDTerm
			reduce(177), // //=, reduce: BitwiseANDTerm
			reduce(177), // %=, reduce: BitwiseANDTerm
			reduce(177), // **=, reduce: BitwiseANDTerm
			reduce(177), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(177), // ||, reduce: BitwiseANDTerm
			reduce(177), // ^^, reduce: BitwiseANDTerm
			reduce(177), // &&, reduce: BitwiseANDTerm
			reduce(177), // ??, reduce: BitwiseANDTerm
			reduce(177), // ???, reduce: BitwiseANDTerm
			reduce(177), // =~, reduce: BitwiseANDTerm
			reduce(177), // !=~, reduce: BitwiseANDTerm
			reduce(177), // ==, reduce: BitwiseANDTerm
			reduce(177), // !=, reduce: BitwiseANDTerm
			reduce(177), // <=>, reduce: BitwiseANDTerm
			reduce(177), // >=, reduce: BitwiseANDTerm
			reduce(177), // <, reduce: BitwiseANDTerm
			reduce(177), // <=, reduce: BitwiseANDTerm
			reduce(177), // ^, reduce: BitwiseANDTerm
			reduce(177), // &, reduce: BitwiseANDTerm
			shift(4137), // <<
			shift(4138), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3660
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(181), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			nil,         // }
			reduce(181), // =, reduce: BitwiseShiftTerm
			nil,         // unset
			nil,         // filter
			reduce(181), // >, reduce: BitwiseShiftTerm
			reduce(181), // >>, reduce: BitwiseShiftTerm
			reduce(181), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(181), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(181), // ||=, reduce: BitwiseShiftTerm
			reduce(181), // ^^=, reduce: BitwiseShiftTerm
			reduce(181), // &&=, reduce: BitwiseShiftTerm
			reduce(181), // ??=, reduce: BitwiseShiftTerm
			reduce(181), // ???=, reduce: BitwiseShiftTerm
			reduce(181), // |=, reduce: BitwiseShiftTerm
			reduce(181), // &=, reduce: BitwiseShiftTerm
			reduce(181), // ^=, reduce: BitwiseShiftTerm
			reduce(181), // <<=, reduce: BitwiseShiftTerm
			reduce(181), // >>=, reduce: BitwiseShiftTerm
			reduce(181), // >>>=, reduce: BitwiseShiftTerm
			reduce(181), // +=, reduce: BitwiseShiftTerm
			reduce(181), // .=, reduce: BitwiseShiftTerm
			reduce(181), // -=, reduce: BitwiseShiftTerm
			reduce(181), // *=, reduce: BitwiseShiftTerm
			reduce(181), // /=, reduce: BitwiseShiftTerm
			reduce(181), // //=, reduce: BitwiseShiftTerm
			reduce(181), // %=, reduce: BitwiseShiftTerm
			reduce(181), // **=, reduce: BitwiseShiftTerm
			reduce(181), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(181), // ||, reduce: BitwiseShiftTerm
			reduce(181), // ^^, reduce: BitwiseShiftTerm
			reduce(181), // &&, reduce: BitwiseShiftTerm
			reduce(181), // ??, reduce: BitwiseShiftTerm
			reduce(181), // ???, reduce: BitwiseShiftTerm
			reduce(181), // =~, reduce: BitwiseShiftTerm
			reduce(181), // !=~, reduce: BitwiseShiftTerm
			reduce(181), // ==, reduce: BitwiseShiftTerm
			reduce(181), // !=, reduce: BitwiseShiftTerm
			reduce(181), // <=>, reduce: BitwiseShiftTerm
			reduce(181), // >=, reduce: BitwiseShiftTerm
			reduce(181), // <, reduce: BitwiseShiftTerm
			reduce(181), // <=, reduce: BitwiseShiftTerm
			reduce(181), // ^, reduce: BitwiseShiftTerm
			reduce(181), // &, reduce: BitwiseShiftTerm
			reduce(181), // <<, reduce: BitwiseShiftTerm
			reduce(181), // >>>, reduce: BitwiseShiftTerm
			shift(4139), // +
			shift(4140), // -
			shift(4141), // .+
			shift(4142), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3661
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3662
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(186), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			reduce(186), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(186), // >, reduce: AddsubdotTerm
			reduce(186), // >>, reduce: AddsubdotTerm
			reduce(186), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(186), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(186), // ||=, reduce: AddsubdotTerm
			reduce(186), // ^^=, reduce: AddsubdotTerm
			reduce(186), // &&=, reduce: AddsubdotTerm
			reduce(186), // ??=, reduce: AddsubdotTerm
			reduce(186), // ???=, reduce: AddsubdotTerm
			reduce(186), // |=, reduce: AddsubdotTerm
			reduce(186), // &=, reduce: AddsubdotTerm
			reduce(186), // ^=, reduce: AddsubdotTerm
			reduce(186), // <<=, reduce: AddsubdotTerm
			reduce(186), // >>=, reduce: AddsubdotTerm
			reduce(186), // >>>=, reduce: AddsubdotTerm
			reduce(186), // +=, reduce: AddsubdotTerm
			reduce(186), // .=, reduce: AddsubdotTerm
			reduce(186), // -=, reduce: AddsubdotTerm
			reduce(186), // *=, reduce: AddsubdotTerm
			reduce(186), // /=, reduce: AddsubdotTerm
			reduce(186), // //=, reduce: AddsubdotTerm
			reduce(186), // %=, reduce: AddsubdotTerm
			reduce(186), // **=, reduce: AddsubdotTerm
			reduce(186), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(186), // ||, reduce: AddsubdotTerm
			reduce(186), // ^^, reduce: AddsubdotTerm
			reduce(186), // &&, reduce: AddsubdotTerm
			reduce(186), // ??, reduce: AddsubdotTerm
			reduce(186), // ???, reduce: AddsubdotTerm
			reduce(186), // =~, reduce: AddsubdotTerm
			reduce(186), // !=~, reduce: AddsubdotTerm
			reduce(186), // ==, reduce: AddsubdotTerm
			reduce(186), // !=, reduce: AddsubdotTerm
			reduce(186), // <=>, reduce: AddsubdotTerm
			reduce(186), // >=, reduce: AddsubdotTerm
			reduce(186), // <, reduce: AddsubdotTerm
			reduce(186), // <=, reduce: AddsubdotTerm
			reduce(186), // ^, reduce: AddsubdotTerm
			reduce(186), // &, reduce: AddsubdotTerm
			reduce(186), // <<, reduce: AddsubdotTerm
			reduce(186), // >>>, reduce: AddsubdotTerm
			reduce(186), // +, reduce: AddsubdotTerm
			reduce(186), // -, reduce: AddsubdotTerm
			reduce(186), // .+, reduce: AddsubdotTerm
			reduce(186), // .-, reduce: AddsubdotTerm
			shift(4144), // *
			shift(4145), // /
			shift(4146), // //
			shift(4147), // %
			shift(4148), // .*
			shift(4149), // ./
			shift(4150), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3663
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3664
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3665
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3666
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(194), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			reduce(194), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(194), // >, reduce: MuldivTerm
			reduce(194), // >>, reduce: MuldivTerm
			reduce(194), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(194), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(194), // ||=, reduce: MuldivTerm
			reduce(194), // ^^=, reduce: MuldivTerm
			reduce(194), // &&=, reduce: MuldivTerm
			reduce(194), // ??=, reduce: MuldivTerm
			reduce(194), // ???=, reduce: MuldivTerm
			reduce(194), // |=, reduce: MuldivTerm
			reduce(194), // &=, reduce: MuldivTerm
			reduce(194), // ^=, reduce: MuldivTerm
			reduce(194), // <<=, reduce: MuldivTerm
			reduce(194), // >>=, reduce: MuldivTerm
			reduce(194), // >>>=, reduce: MuldivTerm
			reduce(194), // +=, reduce: MuldivTerm
			reduce(194), // .=, reduce: MuldivTerm
			reduce(194), // -=, reduce: MuldivTerm
			reduce(194), // *=, reduce: MuldivTerm
			reduce(194), // /=, reduce: MuldivTerm
			reduce(194), // //=, reduce: MuldivTerm
			reduce(194), // %=, reduce: MuldivTerm
			reduce(194), // **=, reduce: MuldivTerm
			reduce(194), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(194), // ||, reduce: MuldivTerm
			reduce(194), // ^^, reduce: MuldivTerm
			reduce(194), // &&, reduce: MuldivTerm
			reduce(194), // ??, reduce: MuldivTerm
			reduce(194), // ???, reduce: MuldivTerm
			reduce(194), // =~, reduce: MuldivTerm
			reduce(194), // !=~, reduce: MuldivTerm
			reduce(194), // ==, reduce: MuldivTerm
			reduce(194), // !=, reduce: MuldivTerm
			reduce(194), // <=>, reduce: MuldivTerm
			reduce(194), // >=, reduce: MuldivTerm
			reduce(194), // <, reduce: MuldivTerm
			reduce(194), // <=, reduce: MuldivTerm
			reduce(194), // ^, reduce: MuldivTerm
			reduce(194), // &, reduce: MuldivTerm
			reduce(194), // <<, reduce: MuldivTerm
			reduce(194), // >>>, reduce: MuldivTerm
			reduce(194), // +, reduce: MuldivTerm
			reduce(194), // -, reduce: MuldivTerm
			reduce(194), // .+, reduce: MuldivTerm
			reduce(194), // .-, reduce: MuldivTerm
			reduce(194), // *, reduce: MuldivTerm
			reduce(194), // /, reduce: MuldivTerm
			reduce(194), // //, reduce: MuldivTerm
			reduce(194), // %, reduce: MuldivTerm
			reduce(194), // .*, reduce: MuldivTerm
			reduce(194), // ./, reduce: MuldivTerm
			reduce(194), // .//, reduce: MuldivTerm
			shift(4154), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3667
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(196), // ;, reduce: DotTerm
			nil,         // {
			nil,         // }
			reduce(196), // =, reduce: DotTerm
			nil,         // unset
			nil,         // filter
			reduce(196), // >, reduce: DotTerm
			reduce(196), // >>, reduce: DotTerm
			reduce(196), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(196), // ,, reduce: DotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(196), // ||=, reduce: DotTerm
			reduce(196), // ^^=, reduce: DotTerm
			reduce(196), // &&=, reduce: DotTerm
			reduce(196), // ??=, reduce: DotTerm
			reduce(196), // ???=, reduce: DotTerm
			reduce(196), // |=, reduce: DotTerm
			reduce(196), // &=, reduce: DotTerm
			reduce(196), // ^=, reduce: DotTerm
			reduce(196), // <<=, reduce: DotTerm
			reduce(196), // >>=, reduce: DotTerm
			reduce(196), // >>>=, reduce: DotTerm
			reduce(196), // +=, reduce: DotTerm
			reduce(196), // .=, reduce: DotTerm
			reduce(196), // -=, reduce: DotTerm
			reduce(196), // *=, reduce: DotTerm
			reduce(196), // /=, reduce: DotTerm
			reduce(196), // //=, reduce: DotTerm
			reduce(196), // %=, reduce: DotTerm
			reduce(196), // **=, reduce: DotTerm
			reduce(196), // ?, reduce: DotTerm
			nil,         // :
			reduce(196), // ||, reduce: DotTerm
			reduce(196), // ^^, reduce: DotTerm
			reduce(196), // &&, reduce: DotTerm
			reduce(196), // ??, reduce: DotTerm
			reduce(196), // ???, reduce: DotTerm
			reduce(196), // =~, reduce: DotTerm
			reduce(196), // !=~, reduce: DotTerm
			reduce(196), // ==, reduce: DotTerm
			reduce(196), // !=, reduce: DotTerm
			reduce(196), // <=>, reduce: DotTerm
			reduce(196), // >=, reduce: DotTerm
			reduce(196), // <, reduce: DotTerm
			reduce(196), // <=, reduce: DotTerm
			reduce(196), // ^, reduce: DotTerm
			reduce(196), // &, reduce: DotTerm
			reduce(196), // <<, reduce: DotTerm
			reduce(196), // >>>, reduce: DotTerm
			reduce(196), // +, reduce: DotTerm
			reduce(196), // -, reduce: DotTerm
			reduce(196), // .+, reduce: DotTerm
			reduce(196), // .-, reduce: DotTerm
			reduce(196), // *, reduce: DotTerm
			reduce(196), // /, reduce: DotTerm
			reduce(196), // //, reduce: DotTerm
			reduce(196), // %, reduce: DotTerm
			reduce(196), // .*, reduce: DotTerm
			reduce(196), // ./, reduce: DotTerm
			reduce(196), // .//, reduce: DotTerm
			reduce(196), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3668
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3669
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3670
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(203), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			reduce(203), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(203), // >, reduce: UnaryOpTerm
			reduce(203), // >>, reduce: UnaryOpTerm
			reduce(203), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(203), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(203), // ||=, reduce: UnaryOpTerm
			reduce(203), // ^^=, reduce: UnaryOpTerm
			reduce(203), // &&=, reduce: UnaryOpTerm
			reduce(203), // ??=, reduce: UnaryOpTerm
			reduce(203), // ???=, reduce: UnaryOpTerm
			reduce(203), // |=, reduce: UnaryOpTerm
			reduce(203), // &=, reduce: UnaryOpTerm
			reduce(203), // ^=, reduce: UnaryOpTerm
			reduce(203), // <<=, reduce: UnaryOpTerm
			reduce(203), // >>=, reduce: UnaryOpTerm
			reduce(203), // >>>=, reduce: UnaryOpTerm
			reduce(203), // +=, reduce: UnaryOpTerm
			reduce(203), // .=, reduce: UnaryOpTerm
			reduce(203), // -=, reduce: UnaryOpTerm
			reduce(203), // *=, reduce: UnaryOpTerm
			reduce(203), // /=, reduce: UnaryOpTerm
			reduce(203), // //=, reduce: UnaryOpTerm
			reduce(203), // %=, reduce: UnaryOpTerm
			reduce(203), // **=, reduce: UnaryOpTerm
			reduce(203), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(203), // ||, reduce: UnaryOpTerm
			reduce(203), // ^^, reduce: UnaryOpTerm
			reduce(203), // &&, reduce: UnaryOpTerm
			reduce(203), // ??, reduce: UnaryOpTerm
			reduce(203), // ???, reduce: UnaryOpTerm
			reduce(203), // =~, reduce: UnaryOpTerm
			reduce(203), // !=~, reduce: UnaryOpTerm
			reduce(203), // ==, reduce: UnaryOpTerm
			reduce(203), // !=, reduce: UnaryOpTerm
			reduce(203), // <=>, reduce: UnaryOpTerm
			reduce(203), // >=, reduce: UnaryOpTerm
			reduce(203), // <, reduce: UnaryOpTerm
			reduce(203), // <=, reduce: UnaryOpTerm
			reduce(203), // ^, reduce: UnaryOpTerm
			reduce(203), // &, reduce: UnaryOpTerm
			reduce(203), // <<, reduce: UnaryOpTerm
			reduce(203), // >>>, reduce: UnaryOpTerm
			reduce(203), // +, reduce: UnaryOpTerm
			reduce(203), // -, reduce: UnaryOpTerm
			reduce(203), // .+, reduce: UnaryOpTerm
			reduce(203), // .-, reduce: UnaryOpTerm
			reduce(203), // *, reduce: UnaryOpTerm
			reduce(203), // /, reduce: UnaryOpTerm
			reduce(203), // //, reduce: UnaryOpTerm
			reduce(203), // %, reduce: UnaryOpTerm
			reduce(203), // .*, reduce: UnaryOpTerm
			reduce(203), // ./, reduce: UnaryOpTerm
			reduce(203), // .//, reduce: UnaryOpTerm
			reduce(203), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3671
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(207), // ;, reduce: PowTerm
			nil,         // {
			nil,         // }
			reduce(207), // =, reduce: PowTerm
			nil,         // unset
			nil,         // filter
			reduce(207), // >, reduce: PowTerm
			reduce(207), // >>, reduce: PowTerm
			reduce(207), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(207), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(207), // ||=, reduce: PowTerm
			reduce(207), // ^^=, reduce: PowTerm
			reduce(207), // &&=, reduce: PowTerm
			reduce(207), // ??=, reduce: PowTerm
			reduce(207), // ???=, reduce: PowTerm
			reduce(207), // |=, reduce: PowTerm
			reduce(207), // &=, reduce: PowTerm
			reduce(207), // ^=, reduce: PowTerm
			reduce(207), // <<=, reduce: PowTerm
			reduce(207), // >>=, reduce: PowTerm
			reduce(207), // >>>=, reduce: PowTerm
			reduce(207), // +=, reduce: PowTerm
			reduce(207), // .=, reduce: PowTerm
			reduce(207), // -=, reduce: PowTerm
			reduce(207), // *=, reduce: PowTerm
			reduce(207), // /=, reduce: PowTerm
			reduce(207), // //=, reduce: PowTerm
			reduce(207), // %=, reduce: PowTerm
			reduce(207), // **=, reduce: PowTerm
			reduce(207), // ?, reduce: PowTerm
			nil,         // :
			reduce(207), // ||, reduce: PowTerm
			reduce(207), // ^^, reduce: PowTerm
			reduce(207), // &&, reduce: PowTerm
			reduce(207), // ??, reduce: PowTerm
			reduce(207), // ???, reduce: PowTerm
			reduce(207), // =~, reduce: PowTerm
			reduce(207), // !=~, reduce: PowTerm
			reduce(207), // ==, reduce: PowTerm
			reduce(207), // !=, reduce: PowTerm
			reduce(207), // <=>, reduce: PowTerm
			reduce(207), // >=, reduce: PowTerm
			reduce(207), // <, reduce: PowTerm
			reduce(207), // <=, reduce: PowTerm
			reduce(207), // ^, reduce: PowTerm
			reduce(207), // &, reduce: PowTerm
			reduce(207), // <<, reduce: PowTerm
			reduce(207), // >>>, reduce: PowTerm
			reduce(207), // +, reduce: PowTerm
			reduce(207), // -, reduce: PowTerm
			reduce(207), // .+, reduce: PowTerm
			reduce(207), // .-, reduce: PowTerm
			reduce(207), // *, reduce: PowTerm
			reduce(207), // /, reduce: PowTerm
			reduce(207), // //, reduce: PowTerm
			reduce(207), // %, reduce: PowTerm
			reduce(207), // .*, reduce: PowTerm
			reduce(207), // ./, reduce: PowTerm
			reduce(207), // .//, reduce: PowTerm
			reduce(207), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			shift(4157), // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3672
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(209), // ;, reduce: PrecedenceChainEnd
			nil,         // {
			nil,         // }
			reduce(209), // =, reduce: PrecedenceChainEnd
			nil,         // unset
			nil,         // filter
			reduce(209), // >, reduce: PrecedenceChainEnd
			reduce(209), // >>, reduce: PrecedenceChainEnd
			reduce(209), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(209), // ,, reduce: PrecedenceChainEnd
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(209), // ||=, reduce: PrecedenceChainEnd
			reduce(209), // ^^=, reduce: PrecedenceChainEnd
			reduce(209), // &&=, reduce: PrecedenceChainEnd
			reduce(209), // ??=, reduce: PrecedenceChainEnd
			reduce(209), // ???=, reduce: PrecedenceChainEnd
			reduce(209), // |=, reduce: PrecedenceChainEnd
			reduce(209), // &=, reduce: PrecedenceChainEnd
			reduce(209), // ^=, reduce: PrecedenceChainEnd
			reduce(209), // <<=, reduce: PrecedenceChainEnd
			reduce(209), // >>=, reduce: PrecedenceChainEnd
			reduce(209), // >>>=, reduce: PrecedenceChainEnd
			reduce(209), // +=, reduce: PrecedenceChainEnd
			reduce(209), // .=, reduce: PrecedenceChainEnd
			reduce(209), // -=, reduce: PrecedenceChainEnd
			reduce(209), // *=, reduce: PrecedenceChainEnd
			reduce(209), // /=, reduce: PrecedenceChainEnd
			reduce(209), // //=, reduce: PrecedenceChainEnd
			reduce(209), // %=, reduce: PrecedenceChainEnd
			reduce(209), // **=, reduce: PrecedenceChainEnd
			reduce(209), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(209), // ||, reduce: PrecedenceChainEnd
			reduce(209), // ^^, reduce: PrecedenceChainEnd
			reduce(209), // &&, reduce: PrecedenceChainEnd
			reduce(209), // ??, reduce: PrecedenceChainEnd
			reduce(209), // ???, reduce: PrecedenceChainEnd
			reduce(209), // =~, reduce: PrecedenceChainEnd
			reduce(209), // !=~, reduce: PrecedenceChainEnd
			reduce(209), // ==, reduce: PrecedenceChainEnd
			reduce(209), // !=, reduce: PrecedenceChainEnd
			reduce(209), // <=>, reduce: PrecedenceChainEnd
			reduce(209), // >=, reduce: PrecedenceChainEnd
			reduce(209), // <, reduce: PrecedenceChainEnd
			reduce(209), // <=, reduce: PrecedenceChainEnd
			reduce(209), // ^, reduce: PrecedenceChainEnd
			reduce(209), // &, reduce: PrecedenceChainEnd
			reduce(209), // <<, reduce: PrecedenceChainEnd
			reduce(209), // >>>, reduce: PrecedenceChainEnd
			reduce(209), // +, reduce: PrecedenceChainEnd
			reduce(209), // -, reduce: PrecedenceChainEnd
			reduce(209), // .+, reduce: PrecedenceChainEnd
			reduce(209), // .-, reduce: PrecedenceChainEnd
			reduce(209), // *, reduce: PrecedenceChainEnd
			reduce(209), // /, reduce: PrecedenceChainEnd
			reduce(209), // //, reduce: PrecedenceChainEnd
			reduce(209), // %, reduce: PrecedenceChainEnd
			reduce(209), // .*, reduce: PrecedenceChainEnd
			reduce(209), // ./, reduce: PrecedenceChainEnd
			reduce(209), // .//, reduce: PrecedenceChainEnd
			reduce(209), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(209), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(4158), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			shift(4159), // [[
			shift(4160), // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3673
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(215), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(215), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(215), // >, reduce: MlrvalOrFunction
			reduce(215), // >>, reduce: MlrvalOrFunction
			reduce(215), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(215), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(215), // ||=, reduce: MlrvalOrFunction
			reduce(215), // ^^=, reduce: MlrvalOrFunction
			reduce(215), // &&=, reduce: MlrvalOrFunction
			reduce(215), // ??=, reduce: MlrvalOrFunction
			reduce(215), // ???=, reduce: MlrvalOrFunction
			reduce(215), // |=, reduce: MlrvalOrFunction
			reduce(215), // &=, reduce: MlrvalOrFunction
			reduce(215), // ^=, reduce: MlrvalOrFunction
			reduce(215), // <<=, reduce: MlrvalOrFunction
			reduce(215), // >>=, reduce: MlrvalOrFunction
			reduce(215), // >>>=, reduce: MlrvalOrFunction
			reduce(215), // +=, reduce: MlrvalOrFunction
			reduce(215), // .=, reduce: MlrvalOrFunction
			reduce(215), // -=, reduce: MlrvalOrFunction
			reduce(215), // *=, reduce: MlrvalOrFunction
			reduce(215), // /=, reduce: MlrvalOrFunction
			reduce(215), // //=, reduce: MlrvalOrFunction
			reduce(215), // %=, reduce: MlrvalOrFunction
			reduce(215), // **=, reduce: MlrvalOrFunction
			reduce(215), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(215), // ||, reduce: MlrvalOrFunction
			reduce(215), // ^^, reduce: MlrvalOrFunction
			reduce(215), // &&, reduce: MlrvalOrFunction
			reduce(215), // ??, reduce: MlrvalOrFunction
			reduce(215), // ???, reduce: MlrvalOrFunction
			reduce(215), // =~, reduce: MlrvalOrFunction
			reduce(215), // !=~, reduce: MlrvalOrFunction
			reduce(215), // ==, reduce: MlrvalOrFunction
			reduce(215), // !=, reduce: MlrvalOrFunction
			reduce(215), // <=>, reduce: MlrvalOrFunction
			reduce(215), // >=, reduce: MlrvalOrFunction
			reduce(215), // <, reduce: MlrvalOrFunction
			reduce(215), // <=, reduce: MlrvalOrFunction
			reduce(215), // ^, reduce: MlrvalOrFunction
			reduce(215), // &, reduce: MlrvalOrFunction
			reduce(215), // <<, reduce: MlrvalOrFunction
			reduce(215), // >>>, reduce: MlrvalOrFunction
			reduce(215), // +, reduce: MlrvalOrFunction
			reduce(215), // -, reduce: MlrvalOrFunction
			reduce(215), // .+, reduce: MlrvalOrFunction
			reduce(215), // .-, reduce: MlrvalOrFunction
			reduce(215), // *, reduce: MlrvalOrFunction
			reduce(215), // /, reduce: MlrvalOrFunction
			reduce(215), // //, reduce: MlrvalOrFunction
			reduce(215), // %, reduce: MlrvalOrFunction
			reduce(215), // .*, reduce: MlrvalOrFunction
			reduce(215), // ./, reduce: MlrvalOrFunction
			reduce(215), // .//, reduce: MlrvalOrFunction
			reduce(215), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(215), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(215), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(215), // [[, reduce: MlrvalOrFunction
			reduce(215), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3674
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(216), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(216), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(216), // >, reduce: MlrvalOrFunction
			reduce(216), // >>, reduce: MlrvalOrFunction
			reduce(216), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(216), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(216), // ||=, reduce: MlrvalOrFunction
			reduce(216), // ^^=, reduce: MlrvalOrFunction
			reduce(216), // &&=, reduce: MlrvalOrFunction
			reduce(216), // ??=, reduce: MlrvalOrFunction
			reduce(216), // ???=, reduce: MlrvalOrFunction
			reduce(216), // |=, reduce: MlrvalOrFunction
			reduce(216), // &=, reduce: MlrvalOrFunction
			reduce(216), // ^=, reduce: MlrvalOrFunction
			reduce(216), // <<=, reduce: MlrvalOrFunction
			reduce(216), // >>=, reduce: MlrvalOrFunction
			reduce(216), // >>>=, reduce: MlrvalOrFunction
			reduce(216), // +=, reduce: MlrvalOrFunction
			reduce(216), // .=, reduce: MlrvalOrFunction
			reduce(216), // -=, reduce: MlrvalOrFunction
			reduce(216), // *=, reduce: MlrvalOrFunction
			reduce(216), // /=, reduce: MlrvalOrFunction
			reduce(216), // //=, reduce: MlrvalOrFunction
			reduce(216), // %=, reduce: MlrvalOrFunction
			reduce(216), // **=, reduce: MlrvalOrFunction
			reduce(216), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(216), // ||, reduce: MlrvalOrFunction
			reduce(216), // ^^, reduce: MlrvalOrFunction
			reduce(216), // &&, reduce: MlrvalOrFunction
			reduce(216), // ??, reduce: MlrvalOrFunction
			reduce(216), // ???, reduce: MlrvalOrFunction
			reduce(216), // =~, reduce: MlrvalOrFunction
			reduce(216), // !=~, reduce: MlrvalOrFunction
			reduce(216), // ==, reduce: MlrvalOrFunction
			reduce(216), // !=, reduce: MlrvalOrFunction
			reduce(216), // <=>, reduce: MlrvalOrFunction
			reduce(216), // >=, reduce: MlrvalOrFunction
			reduce(216), // <, reduce: MlrvalOrFunction
			reduce(216), // <=, reduce: MlrvalOrFunction
			reduce(216), // ^, reduce: MlrvalOrFunction
			reduce(216), // &, reduce: MlrvalOrFunction
			reduce(216), // <<, reduce: MlrvalOrFunction
			reduce(216), // >>>, reduce: MlrvalOrFunction
			reduce(216), // +, reduce: MlrvalOrFunction
			reduce(216), // -, reduce: MlrvalOrFunction
			reduce(216), // .+, reduce: MlrvalOrFunction
			reduce(216), // .-, reduce: MlrvalOrFunction
			reduce(216), // *, reduce: MlrvalOrFunction
			reduce(216), // /, reduce: MlrvalOrFunction
			reduce(216), // //, reduce: MlrvalOrFunction
			reduce(216), // %, reduce: MlrvalOrFunction
			reduce(216), // .*, reduce: MlrvalOrFunction
			reduce(216), // ./, reduce: MlrvalOrFunction
			reduce(216), // .//, reduce: MlrvalOrFunction
			reduce(216), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(216), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(216), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(216), // [[, reduce: MlrvalOrFunction
			reduce(216), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3675
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(217), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(217), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(217), // >, reduce: MlrvalOrFunction
			reduce(217), // >>, reduce: MlrvalOrFunction
			reduce(217), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(217), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(217), // ||=, reduce: MlrvalOrFunction
			reduce(217), // ^^=, reduce: MlrvalOrFunction
			reduce(217), // &&=, reduce: MlrvalOrFunction
			reduce(217), // ??=, reduce: MlrvalOrFunction
			reduce(217), // ???=, reduce: MlrvalOrFunction
			reduce(217), // |=, reduce: MlrvalOrFunction
			reduce(217), // &=, reduce: MlrvalOrFunction
			reduce(217), // ^=, reduce: MlrvalOrFunction
			reduce(217), // <<=, reduce: MlrvalOrFunction
			reduce(217), // >>=, reduce: MlrvalOrFunction
			reduce(217), // >>>=, reduce: MlrvalOrFunction
			reduce(217), // +=, reduce: MlrvalOrFunction
			reduce(217), // .=, reduce: MlrvalOrFunction
			reduce(217), // -=, reduce: MlrvalOrFunction
			reduce(217), // *=, reduce: MlrvalOrFunction
			reduce(217), // /=, reduce: MlrvalOrFunction
			reduce(217), // //=, reduce: MlrvalOrFunction
			reduce(217), // %=, reduce: MlrvalOrFunction
			reduce(217), // **=, reduce: MlrvalOrFunction
			reduce(217), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(217), // ||, reduce: MlrvalOrFunction
			reduce(217), // ^^, reduce: MlrvalOrFunction
			reduce(217), // &&, reduce: MlrvalOrFunction
			reduce(217), // ??, reduce: MlrvalOrFunction
			reduce(217), // ???, reduce: MlrvalOrFunction
			reduce(217), // =~, reduce: MlrvalOrFunction
			reduce(217), // !=~, reduce: MlrvalOrFunction
			reduce(217), // ==, reduce: MlrvalOrFunction
			reduce(217), // !=, reduce: MlrvalOrFunction
			reduce(217), // <=>, reduce: MlrvalOrFunction
			reduce(217), // >=, reduce: MlrvalOrFunction
			reduce(217), // <, reduce: MlrvalOrFunction
			reduce(217), // <=, reduce: MlrvalOrFunction
			reduce(217), // ^, reduce: MlrvalOrFunction
			reduce(217), // &, reduce: MlrvalOrFunction
			reduce(217), // <<, reduce: MlrvalOrFunction
			reduce(217), // >>>, reduce: MlrvalOrFunction
			reduce(217), // +, reduce: MlrvalOrFunction
			reduce(217), // -, reduce: MlrvalOrFunction
			reduce(217), // .+, reduce: MlrvalOrFunction
			reduce(217), // .-, reduce: MlrvalOrFunction
			reduce(217), // *, reduce: MlrvalOrFunction
			reduce(217), // /, reduce: MlrvalOrFunction
			reduce(217), // //, reduce: MlrvalOrFunction
			reduce(217), // %, reduce: MlrvalOrFunction
			reduce(217), // .*, reduce: MlrvalOrFunction
			reduce(217), // ./, reduce: MlrvalOrFunction
			reduce(217), // .//, reduce: MlrvalOrFunction
			reduce(217), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(217), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(217), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(217), // [[, reduce: MlrvalOrFunction
			reduce(217), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3676
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(218), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(218), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(218), // >, reduce: MlrvalOrFunction
			reduce(218), // >>, reduce: MlrvalOrFunction
			reduce(218), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(218), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(218), // ||=, reduce: MlrvalOrFunction
			reduce(218), // ^^=, reduce: MlrvalOrFunction
			reduce(218), // &&=, reduce: MlrvalOrFunction
			reduce(218), // ??=, reduce: MlrvalOrFunction
			reduce(218), // ???=, reduce: MlrvalOrFunction
			reduce(218), // |=, reduce: MlrvalOrFunction
			reduce(218), // &=, reduce: MlrvalOrFunction
			reduce(218), // ^=, reduce: MlrvalOrFunction
			reduce(218), // <<=, reduce: MlrvalOrFunction
			reduce(218), // >>=, reduce: MlrvalOrFunction
			reduce(218), // >>>=, reduce: MlrvalOrFunction
			reduce(218), // +=, reduce: MlrvalOrFunction
			reduce(218), // .=, reduce: MlrvalOrFunction
			reduce(218), // -=, reduce: MlrvalOrFunction
			reduce(218), // *=, reduce: MlrvalOrFunction
			reduce(218), // /=, reduce: MlrvalOrFunction
			reduce(218), // //=, reduce: MlrvalOrFunction
			reduce(218), // %=, reduce: MlrvalOrFunction
			reduce(218), // **=, reduce: MlrvalOrFunction
			reduce(218), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(218), // ||, reduce: MlrvalOrFunction
			reduce(218), // ^^, reduce: MlrvalOrFunction
			reduce(218), // &&, reduce: MlrvalOrFunction
			reduce(218), // ??, reduce: MlrvalOrFunction
			reduce(218), // ???, reduce: MlrvalOrFunction
			reduce(218), // =~, reduce: MlrvalOrFunction
			reduce(218), // !=~, reduce: MlrvalOrFunction
			reduce(218), // ==, reduce: MlrvalOrFunction
			reduce(218), // !=, reduce: MlrvalOrFunction
			reduce(218), // <=>, reduce: MlrvalOrFunction
			reduce(218), // >=, reduce: MlrvalOrFunction
			reduce(218), // <, reduce: MlrvalOrFunction
			reduce(218), // <=, reduce: MlrvalOrFunction
			reduce(218), // ^, reduce: MlrvalOrFunction
			reduce(218), // &, reduce: MlrvalOrFunction
			reduce(218), // <<, reduce: MlrvalOrFunction
			reduce(218), // >>>, reduce: MlrvalOrFunction
			reduce(218), // +, reduce: MlrvalOrFunction
			reduce(218), // -, reduce: MlrvalOrFunction
			reduce(218), // .+, reduce: MlrvalOrFunction
			reduce(218), // .-, reduce: MlrvalOrFunction
			reduce(218), // *, reduce: MlrvalOrFunction
			reduce(218), // /, reduce: MlrvalOrFunction
			reduce(218), // //, reduce: MlrvalOrFunction
			reduce(218), // %, reduce: MlrvalOrFunction
			reduce(218), // .*, reduce: MlrvalOrFunction
			reduce(218), // ./, reduce: MlrvalOrFunction
			reduce(218), // .//, reduce: MlrvalOrFunction
			reduce(218), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(218), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(218), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(218), // [[, reduce: MlrvalOrFunction
			reduce(218), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3677
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(219), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(219), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(219), // >, reduce: MlrvalOrFunction
			reduce(219), // >>, reduce: MlrvalOrFunction
			reduce(219), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(219), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(219), // ||=, reduce: MlrvalOrFunction
			reduce(219), // ^^=, reduce: MlrvalOrFunction
			reduce(219), // &&=, reduce: MlrvalOrFunction
			reduce(219), // ??=, reduce: MlrvalOrFunction
			reduce(219), // ???=, reduce: MlrvalOrFunction
			reduce(219), // |=, reduce: MlrvalOrFunction
			reduce(219), // &=, reduce: MlrvalOrFunction
			reduce(219), // ^=, reduce: MlrvalOrFunction
			reduce(219), // <<=, reduce: MlrvalOrFunction
			reduce(219), // >>=, reduce: MlrvalOrFunction
			reduce(219), // >>>=, reduce: MlrvalOrFunction
			reduce(219), // +=, reduce: MlrvalOrFunction
			reduce(219), // .=, reduce: MlrvalOrFunction
			reduce(219), // -=, reduce: MlrvalOrFunction
			reduce(219), // *=, reduce: MlrvalOrFunction
			reduce(219), // /=, reduce: MlrvalOrFunction
			reduce(219), // //=, reduce: MlrvalOrFunction
			reduce(219), // %=, reduce: MlrvalOrFunction
			reduce(219), // **=, reduce: MlrvalOrFunction
			reduce(219), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(219), // ||, reduce: MlrvalOrFunction
			reduce(219), // ^^, reduce: MlrvalOrFunction
			reduce(219), // &&, reduce: MlrvalOrFunction
			reduce(219), // ??, reduce: MlrvalOrFunction
			reduce(219), // ???, reduce: MlrvalOrFunction
			reduce(219), // =~, reduce: MlrvalOrFunction
			reduce(219), // !=~, reduce: MlrvalOrFunction
			reduce(219), // ==, reduce: MlrvalOrFunction
			reduce(219), // !=, reduce: MlrvalOrFunction
			reduce(219), // <=>, reduce: MlrvalOrFunction
			reduce(219), // >=, reduce: MlrvalOrFunction
			reduce(219), // <, reduce: MlrvalOrFunction
			reduce(219), // <=, reduce: MlrvalOrFunction
			reduce(219), // ^, reduce: MlrvalOrFunction
			reduce(219), // &, reduce: MlrvalOrFunction
			reduce(219), // <<, reduce: MlrvalOrFunction
			reduce(219), // >>>, reduce: MlrvalOrFunction
			reduce(219), // +, reduce: MlrvalOrFunction
			reduce(219), // -, reduce: MlrvalOrFunction
			reduce(219), // .+, reduce: MlrvalOrFunction
			reduce(219), // .-, reduce: MlrvalOrFunction
			reduce(219), // *, reduce: MlrvalOrFunction
			reduce(219), // /, reduce: MlrvalOrFunction
			reduce(219), // //, reduce: MlrvalOrFunction
			reduce(219), // %, reduce: MlrvalOrFunction
			reduce(219), // .*, reduce: MlrvalOrFunction
			reduce(219), // ./, reduce: MlrvalOrFunction
			reduce(219), // .//, reduce: MlrvalOrFunction
			reduce(219), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(219), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(219), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(219), // [[, reduce: MlrvalOrFunction
			reduce(219), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3678
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(220), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(220), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(220), // >, reduce: MlrvalOrFunction
			reduce(220), // >>, reduce: MlrvalOrFunction
			reduce(220), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(220), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(220), // ||=, reduce: MlrvalOrFunction
			reduce(220), // ^^=, reduce: MlrvalOrFunction
			reduce(220), // &&=, reduce: MlrvalOrFunction
			reduce(220), // ??=, reduce: MlrvalOrFunction
			reduce(220), // ???=, reduce: MlrvalOrFunction
			reduce(220), // |=, reduce: MlrvalOrFunction
			reduce(220), // &=, reduce: MlrvalOrFunction
			reduce(220), // ^=, reduce: MlrvalOrFunction
			reduce(220), // <<=, reduce: MlrvalOrFunction
			reduce(220), // >>=, reduce: MlrvalOrFunction
			reduce(220), // >>>=, reduce: MlrvalOrFunction
			reduce(220), // +=, reduce: MlrvalOrFunction
			reduce(220), // .=, reduce: MlrvalOrFunction
			reduce(220), // -=, reduce: MlrvalOrFunction
			reduce(220), // *=, reduce: MlrvalOrFunction
			reduce(220), // /=, reduce: MlrvalOrFunction
			reduce(220), // //=, reduce: MlrvalOrFunction
			reduce(220), // %=, reduce: MlrvalOrFunction
			reduce(220), // **=, reduce: MlrvalOrFunction
			reduce(220), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(220), // ||, reduce: MlrvalOrFunction
			reduce(220), // ^^, reduce: MlrvalOrFunction
			reduce(220), // &&, reduce: MlrvalOrFunction
			reduce(220), // ??, reduce: MlrvalOrFunction
			reduce(220), // ???, reduce: MlrvalOrFunction
			reduce(220), // =~, reduce: MlrvalOrFunction
			reduce(220), // !=~, reduce: MlrvalOrFunction
			reduce(220), // ==, reduce: MlrvalOrFunction
			reduce(220), // !=, reduce: MlrvalOrFunction
			reduce(220), // <=>, reduce: MlrvalOrFunction
			reduce(220), // >=, reduce: MlrvalOrFunction
			reduce(220), // <, reduce: MlrvalOrFunction
			reduce(220), // <=, reduce: MlrvalOrFunction
			reduce(220), // ^, reduce: MlrvalOrFunction
			reduce(220), // &, reduce: MlrvalOrFunction
			reduce(220), // <<, reduce: MlrvalOrFunction
			reduce(220), // >>>, reduce: MlrvalOrFunction
			reduce(220), // +, reduce: MlrvalOrFunction
			reduce(220), // -, reduce: MlrvalOrFunction
			reduce(220), // .+, reduce: MlrvalOrFunction
			reduce(220), // .-, reduce: MlrvalOrFunction
			reduce(220), // *, reduce: MlrvalOrFunction
			reduce(220), // /, reduce: MlrvalOrFunction
			reduce(220), // //, reduce: MlrvalOrFunction
			reduce(220), // %, reduce: MlrvalOrFunction
			reduce(220), // .*, reduce: MlrvalOrFunction
			reduce(220), // ./, reduce: MlrvalOrFunction
			reduce(220), // .//, reduce: MlrvalOrFunction
			reduce(220), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(220), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(220), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(220), // [[, reduce: MlrvalOrFunction
			reduce(220), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3679
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(221), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(221), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(221), // >, reduce: MlrvalOrFunction
			reduce(221), // >>, reduce: MlrvalOrFunction
			reduce(221), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(221), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(221), // ||=, reduce: MlrvalOrFunction
			reduce(221), // ^^=, reduce: MlrvalOrFunction
			reduce(221), // &&=, reduce: MlrvalOrFunction
			reduce(221), // ??=, reduce: MlrvalOrFunction
			reduce(221), // ???=, reduce: MlrvalOrFunction
			reduce(221), // |=, reduce: MlrvalOrFunction
			reduce(221), // &=, reduce: MlrvalOrFunction
			reduce(221), // ^=, reduce: MlrvalOrFunction
			reduce(221), // <<=, reduce: MlrvalOrFunction
			reduce(221), // >>=, reduce: MlrvalOrFunction
			reduce(221), // >>>=, reduce: MlrvalOrFunction
			reduce(221), // +=, reduce: MlrvalOrFunction
			reduce(221), // .=, reduce: MlrvalOrFunction
			reduce(221), // -=, reduce: MlrvalOrFunction
			reduce(221), // *=, reduce: MlrvalOrFunction
			reduce(221), // /=, reduce: MlrvalOrFunction
			reduce(221), // //=, reduce: MlrvalOrFunction
			reduce(221), // %=, reduce: MlrvalOrFunction
			reduce(221), // **=, reduce: MlrvalOrFunction
			reduce(221), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(221), // ||, reduce: MlrvalOrFunction
			reduce(221), // ^^, reduce: MlrvalOrFunction
			reduce(221), // &&, reduce: MlrvalOrFunction
			reduce(221), // ??, reduce: MlrvalOrFunction
			reduce(221), // ???, reduce: MlrvalOrFunction
			reduce(221), // =~, reduce: MlrvalOrFunction
			reduce(221), // !=~, reduce: MlrvalOrFunction
			reduce(221), // ==, reduce: MlrvalOrFunction
			reduce(221), // !=, reduce: MlrvalOrFunction
			reduce(221), // <=>, reduce: MlrvalOrFunction
			reduce(221), // >=, reduce: MlrvalOrFunction
			reduce(221), // <, reduce: MlrvalOrFunction
			reduce(221), // <=, reduce: MlrvalOrFunction
			reduce(221), // ^, reduce: MlrvalOrFunction
			reduce(221), // &, reduce: MlrvalOrFunction
			reduce(221), // <<, reduce: MlrvalOrFunction
			reduce(221), // >>>, reduce: MlrvalOrFunction
			reduce(221), // +, reduce: MlrvalOrFunction
			reduce(221), // -, reduce: MlrvalOrFunction
			reduce(221), // .+, reduce: MlrvalOrFunction
			reduce(221), // .-, reduce: MlrvalOrFunction
			reduce(221), // *, reduce: MlrvalOrFunction
			reduce(221), // /, reduce: MlrvalOrFunction
			reduce(221), // //, reduce: MlrvalOrFunction
			reduce(221), // %, reduce: MlrvalOrFunction
			reduce(221), // .*, reduce: MlrvalOrFunction
			reduce(221), // ./, reduce: MlrvalOrFunction
			reduce(221), // .//, reduce: MlrvalOrFunction
			reduce(221), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(221), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(221), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(221), // [[, reduce: MlrvalOrFunction
			reduce(221), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3680
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(222), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(222), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(222), // >, reduce: MlrvalOrFunction
			reduce(222), // >>, reduce: MlrvalOrFunction
			reduce(222), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(222), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(222), // ||=, reduce: MlrvalOrFunction
			reduce(222), // ^^=, reduce: MlrvalOrFunction
			reduce(222), // &&=, reduce: MlrvalOrFunction
			reduce(222), // ??=, reduce: MlrvalOrFunction
			reduce(222), // ???=, reduce: MlrvalOrFunction
			reduce(222), // |=, reduce: MlrvalOrFunction
			reduce(222), // &=, reduce: MlrvalOrFunction
			reduce(222), // ^=, reduce: MlrvalOrFunction
			reduce(222), // <<=, reduce: MlrvalOrFunction
			reduce(222), // >>=, reduce: MlrvalOrFunction
			reduce(222), // >>>=, reduce: MlrvalOrFunction
			reduce(222), // +=, reduce: MlrvalOrFunction
			reduce(222), // .=, reduce: MlrvalOrFunction
			reduce(222), // -=, reduce: MlrvalOrFunction
			reduce(222), // *=, reduce: MlrvalOrFunction
			reduce(222), // /=, reduce: MlrvalOrFunction
			reduce(222), // //=, reduce: MlrvalOrFunction
			reduce(222), // %=, reduce: MlrvalOrFunction
			reduce(222), // **=, reduce: MlrvalOrFunction
			reduce(222), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(222), // ||, reduce: MlrvalOrFunction
			reduce(222), // ^^, reduce: MlrvalOrFunction
			reduce(222), // &&, reduce: MlrvalOrFunction
			reduce(222), // ??, reduce: MlrvalOrFunction
			reduce(222), // ???, reduce: MlrvalOrFunction
			reduce(222), // =~, reduce: MlrvalOrFunction
			reduce(222), // !=~, reduce: MlrvalOrFunction
			reduce(222), // ==, reduce: MlrvalOrFunction
			reduce(222), // !=, reduce: MlrvalOrFunction
			reduce(222), // <=>, reduce: MlrvalOrFunction
			reduce(222), // >=, reduce: MlrvalOrFunction
			reduce(222), // <, reduce: MlrvalOrFunction
			reduce(222), // <=, reduce: MlrvalOrFunction
			reduce(222), // ^, reduce: MlrvalOrFunction
			reduce(222), // &, reduce: MlrvalOrFunction
			reduce(222), // <<, reduce: MlrvalOrFunction
			reduce(222), // >>>, reduce: MlrvalOrFunction
			reduce(222), // +, reduce: MlrvalOrFunction
			reduce(222), // -, reduce: MlrvalOrFunction
			reduce(222), // .+, reduce: MlrvalOrFunction
			reduce(222), // .-, reduce: MlrvalOrFunction
			reduce(222), // *, reduce: MlrvalOrFunction
			reduce(222), // /, reduce: MlrvalOrFunction
			reduce(222), // //, reduce: MlrvalOrFunction
			reduce(222), // %, reduce: MlrvalOrFunction
			reduce(222), // .*, reduce: MlrvalOrFunction
			reduce(222), // ./, reduce: MlrvalOrFunction
			reduce(222), // .//, reduce: MlrvalOrFunction
			reduce(222), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(222), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(222), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(222), // [[, reduce: MlrvalOrFunction
			reduce(222), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3681
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(223), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(223), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(223), // >, reduce: MlrvalOrFunction
			reduce(223), // >>, reduce: MlrvalOrFunction
			reduce(223), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(223), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(223), // ||=, reduce: MlrvalOrFunction
			reduce(223), // ^^=, reduce: MlrvalOrFunction
			reduce(223), // &&=, reduce: MlrvalOrFunction
			reduce(223), // ??=, reduce: MlrvalOrFunction
			reduce(223), // ???=, reduce: MlrvalOrFunction
			reduce(223), // |=, reduce: MlrvalOrFunction
			reduce(223), // &=, reduce: MlrvalOrFunction
			reduce(223), // ^=, reduce: MlrvalOrFunction
			reduce(223), // <<=, reduce: MlrvalOrFunction
			reduce(223), // >>=, reduce: MlrvalOrFunction
			reduce(223), // >>>=, reduce: MlrvalOrFunction
			reduce(223), // +=, reduce: MlrvalOrFunction
			reduce(223), // .=, reduce: MlrvalOrFunction
			reduce(223), // -=, reduce: MlrvalOrFunction
			reduce(223), // *=, reduce: MlrvalOrFunction
			reduce(223), // /=, reduce: MlrvalOrFunction
			reduce(223), // //=, reduce: MlrvalOrFunction
			reduce(223), // %=, reduce: MlrvalOrFunction
			reduce(223), // **=, reduce: MlrvalOrFunction
			reduce(223), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(223), // ||, reduce: MlrvalOrFunction
			reduce(223), // ^^, reduce: MlrvalOrFunction
			reduce(223), // &&, reduce: MlrvalOrFunction
			reduce(223), // ??, reduce: MlrvalOrFunction
			reduce(223), // ???, reduce: MlrvalOrFunction
			reduce(223), // =~, reduce: MlrvalOrFunction
			reduce(223), // !=~, reduce: MlrvalOrFunction
			reduce(223), // ==, reduce: MlrvalOrFunction
			reduce(223), // !=, reduce: MlrvalOrFunction
			reduce(223), // <=>, reduce: MlrvalOrFunction
			reduce(223), // >=, reduce: MlrvalOrFunction
			reduce(223), // <, reduce: MlrvalOrFunction
			reduce(223), // <=, reduce: MlrvalOrFunction
			reduce(223), // ^, reduce: MlrvalOrFunction
			reduce(223), // &, reduce: MlrvalOrFunction
			reduce(223), // <<, reduce: MlrvalOrFunction
			reduce(223), // >>>, reduce: MlrvalOrFunction
			reduce(223), // +, reduce: MlrvalOrFunction
			reduce(223), // -, reduce: MlrvalOrFunction
			reduce(223), // .+, reduce: MlrvalOrFunction
			reduce(223), // .-, reduce: MlrvalOrFunction
			reduce(223), // *, reduce: MlrvalOrFunction
			reduce(223), // /, reduce: MlrvalOrFunction
			reduce(223), // //, reduce: MlrvalOrFunction
			reduce(223), // %, reduce: MlrvalOrFunction
			reduce(223), // .*, reduce: MlrvalOrFunction
			reduce(223), // ./, reduce: MlrvalOrFunction
			reduce(223), // .//, reduce: MlrvalOrFunction
			reduce(223), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(223), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(223), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(223), // [[, reduce: MlrvalOrFunction
			reduce(223), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3682
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(224), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(224), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(224), // >, reduce: MlrvalOrFunction
			reduce(224), // >>, reduce: MlrvalOrFunction
			reduce(224), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(224), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(224), // ||=, reduce: MlrvalOrFunction
			reduce(224), // ^^=, reduce: MlrvalOrFunction
			reduce(224), // &&=, reduce: MlrvalOrFunction
			reduce(224), // ??=, reduce: MlrvalOrFunction
			reduce(224), // ???=, reduce: MlrvalOrFunction
			reduce(224), // |=, reduce: MlrvalOrFunction
			reduce(224), // &=, reduce: MlrvalOrFunction
			reduce(224), // ^=, reduce: MlrvalOrFunction
			reduce(224), // <<=, reduce: MlrvalOrFunction
			reduce(224), // >>=, reduce: MlrvalOrFunction
			reduce(224), // >>>=, reduce: MlrvalOrFunction
			reduce(224), // +=, reduce: MlrvalOrFunction
			reduce(224), // .=, reduce: MlrvalOrFunction
			reduce(224), // -=, reduce: MlrvalOrFunction
			reduce(224), // *=, reduce: MlrvalOrFunction
			reduce(224), // /=, reduce: MlrvalOrFunction
			reduce(224), // //=, reduce: MlrvalOrFunction
			reduce(224), // %=, reduce: MlrvalOrFunction
			reduce(224), // **=, reduce: MlrvalOrFunction
			reduce(224), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(224), // ||, reduce: MlrvalOrFunction
			reduce(224), // ^^, reduce: MlrvalOrFunction
			reduce(224), // &&, reduce: MlrvalOrFunction
			reduce(224), // ??, reduce: MlrvalOrFunction
			reduce(224), // ???, reduce: MlrvalOrFunction
			reduce(224), // =~, reduce: MlrvalOrFunction
			reduce(224), // !=~, reduce: MlrvalOrFunction
			reduce(224), // ==, reduce: MlrvalOrFunction
			reduce(224), // !=, reduce: MlrvalOrFunction
			reduce(224), // <=>, reduce: MlrvalOrFunction
			reduce(224), // >=, reduce: MlrvalOrFunction
			reduce(224), // <, reduce: MlrvalOrFunction
			reduce(224), // <=, reduce: MlrvalOrFunction
			reduce(224), // ^, reduce: MlrvalOrFunction
			reduce(224), // &, reduce: MlrvalOrFunction
			reduce(224), // <<, reduce: MlrvalOrFunction
			reduce(224), // >>>, reduce: MlrvalOrFunction
			reduce(224), // +, reduce: MlrvalOrFunction
			reduce(224), // -, reduce: MlrvalOrFunction
			reduce(224), // .+, reduce: MlrvalOrFunction
			reduce(224), // .-, reduce: MlrvalOrFunction
			reduce(224), // *, reduce: MlrvalOrFunction
			reduce(224), // /, reduce: MlrvalOrFunction
			reduce(224), // //, reduce: MlrvalOrFunction
			reduce(224), // %, reduce: MlrvalOrFunction
			reduce(224), // .*, reduce: MlrvalOrFunction
			reduce(224), // ./, reduce: MlrvalOrFunction
			reduce(224), // .//, reduce: MlrvalOrFunction
			reduce(224), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(224), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(224), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(224), // [[, reduce: MlrvalOrFunction
			reduce(224), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3683
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(225), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(225), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(225), // >, reduce: MlrvalOrFunction
			reduce(225), // >>, reduce: MlrvalOrFunction
			reduce(225), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(225), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(225), // ||=, reduce: MlrvalOrFunction
			reduce(225), // ^^=, reduce: MlrvalOrFunction
			reduce(225), // &&=, reduce: MlrvalOrFunction
			reduce(225), // ??=, reduce: MlrvalOrFunction
			reduce(225), // ???=, reduce: MlrvalOrFunction
			reduce(225), // |=, reduce: MlrvalOrFunction
			reduce(225), // &=, reduce: MlrvalOrFunction
			reduce(225), // ^=, reduce: MlrvalOrFunction
			reduce(225), // <<=, reduce: MlrvalOrFunction
			reduce(225), // >>=, reduce: MlrvalOrFunction
			reduce(225), // >>>=, reduce: MlrvalOrFunction
			reduce(225), // +=, reduce: MlrvalOrFunction
			reduce(225), // .=, reduce: MlrvalOrFunction
			reduce(225), // -=, reduce: MlrvalOrFunction
			reduce(225), // *=, reduce: MlrvalOrFunction
			reduce(225), // /=, reduce: MlrvalOrFunction
			reduce(225), // //=, reduce: MlrvalOrFunction
			reduce(225), // %=, reduce: MlrvalOrFunction
			reduce(225), // **=, reduce: MlrvalOrFunction
			reduce(225), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(225), // ||, reduce: MlrvalOrFunction
			reduce(225), // ^^, reduce: MlrvalOrFunction
			reduce(225), // &&, reduce: MlrvalOrFunction
			reduce(225), // ??, reduce: MlrvalOrFunction
			reduce(225), // ???, reduce: MlrvalOrFunction
			reduce(225), // =~, reduce: MlrvalOrFunction
			reduce(225), // !=~, reduce: MlrvalOrFunction
			reduce(225), // ==, reduce: MlrvalOrFunction
			reduce(225), // !=, reduce: MlrvalOrFunction
			reduce(225), // <=>, reduce: MlrvalOrFunction
			reduce(225), // >=, reduce: MlrvalOrFunction
			reduce(225), // <, reduce: MlrvalOrFunction
			reduce(225), // <=, reduce: MlrvalOrFunction
			reduce(225), // ^, reduce: MlrvalOrFunction
			reduce(225), // &, reduce: MlrvalOrFunction
			reduce(225), // <<, reduce: MlrvalOrFunction
			reduce(225), // >>>, reduce: MlrvalOrFunction
			reduce(225), // +, reduce: MlrvalOrFunction
			reduce(225), // -, reduce: MlrvalOrFunction
			reduce(225), // .+, reduce: MlrvalOrFunction
			reduce(225), // .-, reduce: MlrvalOrFunction
			reduce(225), // *, reduce: MlrvalOrFunction
			reduce(225), // /, reduce: MlrvalOrFunction
			reduce(225), // //, reduce: MlrvalOrFunction
			reduce(225), // %, reduce: MlrvalOrFunction
			reduce(225), // .*, reduce: MlrvalOrFunction
			reduce(225), // ./, reduce: MlrvalOrFunction
			reduce(225), // .//, reduce: MlrvalOrFunction
			reduce(225), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(225), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(225), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(225), // [[, reduce: MlrvalOrFunction
			reduce(225), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3684
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(226), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(226), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(226), // >, reduce: MlrvalOrFunction
			reduce(226), // >>, reduce: MlrvalOrFunction
			reduce(226), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(226), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(226), // ||=, reduce: MlrvalOrFunction
			reduce(226), // ^^=, reduce: MlrvalOrFunction
			reduce(226), // &&=, reduce: MlrvalOrFunction
			reduce(226), // ??=, reduce: MlrvalOrFunction
			reduce(226), // ???=, reduce: MlrvalOrFunction
			reduce(226), // |=, reduce: MlrvalOrFunction
			reduce(226), // &=, reduce: MlrvalOrFunction
			reduce(226), // ^=, reduce: MlrvalOrFunction
			reduce(226), // <<=, reduce: MlrvalOrFunction
			reduce(226), // >>=, reduce: MlrvalOrFunction
			reduce(226), // >>>=, reduce: MlrvalOrFunction
			reduce(226), // +=, reduce: MlrvalOrFunction
			reduce(226), // .=, reduce: MlrvalOrFunction
			reduce(226), // -=, reduce: MlrvalOrFunction
			reduce(226), // *=, reduce: MlrvalOrFunction
			reduce(226), // /=, reduce: MlrvalOrFunction
			reduce(226), // //=, reduce: MlrvalOrFunction
			reduce(226), // %=, reduce: MlrvalOrFunction
			reduce(226), // **=, reduce: MlrvalOrFunction
			reduce(226), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(226), // ||, reduce: MlrvalOrFunction
			reduce(226), // ^^, reduce: MlrvalOrFunction
			reduce(226), // &&, reduce: MlrvalOrFunction
			reduce(226), // ??, reduce: MlrvalOrFunction
			reduce(226), // ???, reduce: MlrvalOrFunction
			reduce(226), // =~, reduce: MlrvalOrFunction
			reduce(226), // !=~, reduce: MlrvalOrFunction
			reduce(226), // ==, reduce: MlrvalOrFunction
			reduce(226), // !=, reduce: MlrvalOrFunction
			reduce(226), // <=>, reduce: MlrvalOrFunction
			reduce(226), // >=, reduce: MlrvalOrFunction
			reduce(226), // <, reduce: MlrvalOrFunction
			reduce(226), // <=, reduce: MlrvalOrFunction
			reduce(226), // ^, reduce: MlrvalOrFunction
			reduce(226), // &, reduce: MlrvalOrFunction
			reduce(226), // <<, reduce: MlrvalOrFunction
			reduce(226), // >>>, reduce: MlrvalOrFunction
			reduce(226), // +, reduce: MlrvalOrFunction
			reduce(226), // -, reduce: MlrvalOrFunction
			reduce(226), // .+, reduce: MlrvalOrFunction
			reduce(226), // .-, reduce: MlrvalOrFunction
			reduce(226), // *, reduce: MlrvalOrFunction
			reduce(226), // /, reduce: MlrvalOrFunction
			reduce(226), // //, reduce: MlrvalOrFunction
			reduce(226), // %, reduce: MlrvalOrFunction
			reduce(226), // .*, reduce: MlrvalOrFunction
			reduce(226), // ./, reduce: MlrvalOrFunction
			reduce(226), // .//, reduce: MlrvalOrFunction
			reduce(226), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(226), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(226), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(226), // [[, reduce: MlrvalOrFunction
			reduce(226), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3685
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(227), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(227), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(227), // >, reduce: MlrvalOrFunction
			reduce(227), // >>, reduce: MlrvalOrFunction
			reduce(227), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(227), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(227), // ||=, reduce: MlrvalOrFunction
			reduce(227), // ^^=, reduce: MlrvalOrFunction
			reduce(227), // &&=, reduce: MlrvalOrFunction
			reduce(227), // ??=, reduce: MlrvalOrFunction
			reduce(227), // ???=, reduce: MlrvalOrFunction
			reduce(227), // |=, reduce: MlrvalOrFunction
			reduce(227), // &=, reduce: MlrvalOrFunction
			reduce(227), // ^=, reduce: MlrvalOrFunction
			reduce(227), // <<=, reduce: MlrvalOrFunction
			reduce(227), // >>=, reduce: MlrvalOrFunction
			reduce(227), // >>>=, reduce: MlrvalOrFunction
			reduce(227), // +=, reduce: MlrvalOrFunction
			reduce(227), // .=, reduce: MlrvalOrFunction
			reduce(227), // -=, reduce: MlrvalOrFunction
			reduce(227), // *=, reduce: MlrvalOrFunction
			reduce(227), // /=, reduce: MlrvalOrFunction
			reduce(227), // //=, reduce: MlrvalOrFunction
			reduce(227), // %=, reduce: MlrvalOrFunction
			reduce(227), // **=, reduce: MlrvalOrFunction
			reduce(227), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(227), // ||, reduce: MlrvalOrFunction
			reduce(227), // ^^, reduce: MlrvalOrFunction
			reduce(227), // &&, reduce: MlrvalOrFunction
			reduce(227), // ??, reduce: MlrvalOrFunction
			reduce(227), // ???, reduce: MlrvalOrFunction
			reduce(227), // =~, reduce: MlrvalOrFunction
			reduce(227), // !=~, reduce: MlrvalOrFunction
			reduce(227), // ==, reduce: MlrvalOrFunction
			reduce(227), // !=, reduce: MlrvalOrFunction
			reduce(227), // <=>, reduce: MlrvalOrFunction
			reduce(227), // >=, reduce: MlrvalOrFunction
			reduce(227), // <, reduce: MlrvalOrFunction
			reduce(227), // <=, reduce: MlrvalOrFunction
			reduce(227), // ^, reduce: MlrvalOrFunction
			reduce(227), // &, reduce: MlrvalOrFunction
			reduce(227), // <<, reduce: MlrvalOrFunction
			reduce(227), // >>>, reduce: MlrvalOrFunction
			reduce(227), // +, reduce: MlrvalOrFunction
			reduce(227), // -, reduce: MlrvalOrFunction
			reduce(227), // .+, reduce: MlrvalOrFunction
			reduce(227), // .-, reduce: MlrvalOrFunction
			reduce(227), // *, reduce: MlrvalOrFunction
			reduce(227), // /, reduce: MlrvalOrFunction
			reduce(227), // //, reduce: MlrvalOrFunction
			reduce(227), // %, reduce: MlrvalOrFunction
			reduce(227), // .*, reduce: MlrvalOrFunction
			reduce(227), // ./, reduce: MlrvalOrFunction
			reduce(227), // .//, reduce: MlrvalOrFunction
			reduce(227), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(227), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(227), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(227), // [[, reduce: MlrvalOrFunction
			reduce(227), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3686
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(752),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(756),  // (
			nil,         // )
			nil,         // emitp
			shift(768),  // field_name
			shift(769),  // $[
			shift(4161), // ]
			shift(771),  // braced_field_name
			shift(772),  // $[[
			shift(773),  // $[[[
			shift(774),  // full_srec
			shift(776),  // oosvar_name
			shift(777),  // @[
			shift(778),  // braced_oosvar_name
			shift(779),  // full_oosvar
			shift(780),  // all
			shift(781),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(796),  // +
			shift(798),  // -
			shift(799),  // .+
			shift(800),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(803),  // !
			shift(804),  // ~
			nil,         // **
			shift(809),  // string_literal
			shift(810),  // regex_case_insensitive
			shift(811),  // int_literal
			shift(812),  // float_literal
			shift(813),  // boolean_literal
			shift(814),  // null_literal
			shift(815),  // inf_literal
			shift(816),  // nan_literal
			shift(817),  // const_M_PI
			shift(818),  // const_M_E
			shift(819),  // panic
			shift(821),  // [
			shift(824),  // ctx_IPS
			shift(825),  // ctx_IFS
			shift(826),  // ctx_IRS
			shift(827),  // ctx_OPS
			shift(828),  // ctx_OFS
			shift(829),  // ctx_ORS
			shift(830),  // ctx_FLATSEP
			shift(831),  // ctx_NF
			shift(832),  // ctx_NR
			shift(833),  // ctx_FNR
			shift(834),  // ctx_FILENAME
			shift(835),  // ctx_FILENUM
			shift(837),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(844),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3687
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(240), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(240), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(240), // >, reduce: MlrvalOrFunction
			reduce(240), // >>, reduce: MlrvalOrFunction
			reduce(240), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(240), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(240), // ||=, reduce: MlrvalOrFunction
			reduce(240), // ^^=, reduce: MlrvalOrFunction
			reduce(240), // &&=, reduce: MlrvalOrFunction
			reduce(240), // ??=, reduce: MlrvalOrFunction
			reduce(240), // ???=, reduce: MlrvalOrFunction
			reduce(240), // |=, reduce: MlrvalOrFunction
			reduce(240), // &=, reduce: MlrvalOrFunction
			reduce(240), // ^=, reduce: MlrvalOrFunction
			reduce(240), // <<=, reduce: MlrvalOrFunction
			reduce(240), // >>=, reduce: MlrvalOrFunction
			reduce(240), // >>>=, reduce: MlrvalOrFunction
			reduce(240), // +=, reduce: MlrvalOrFunction
			reduce(240), // .=, reduce: MlrvalOrFunction
			reduce(240), // -=, reduce: MlrvalOrFunction
			reduce(240), // *=, reduce: MlrvalOrFunction
			reduce(240), // /=, reduce: MlrvalOrFunction
			reduce(240), // //=, reduce: MlrvalOrFunction
			reduce(240), // %=, reduce: MlrvalOrFunction
			reduce(240), // **=, reduce: MlrvalOrFunction
			reduce(240), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(240), // ||, reduce: MlrvalOrFunction
			reduce(240), // ^^, reduce: MlrvalOrFunction
			reduce(240), // &&, reduce: MlrvalOrFunction
			reduce(240), // ??, reduce: MlrvalOrFunction
			reduce(240), // ???, reduce: MlrvalOrFunction
			reduce(240), // =~, reduce: MlrvalOrFunction
			reduce(240), // !=~, reduce: MlrvalOrFunction
			reduce(240), // ==, reduce: MlrvalOrFunction
			reduce(240), // !=, reduce: MlrvalOrFunction
			reduce(240), // <=>, reduce: MlrvalOrFunction
			reduce(240), // >=, reduce: MlrvalOrFunction
			reduce(240), // <, reduce: MlrvalOrFunction
			reduce(240), // <=, reduce: MlrvalOrFunction
			reduce(240), // ^, reduce: MlrvalOrFunction
			reduce(240), // &, reduce: MlrvalOrFunction
			reduce(240), // <<, reduce: MlrvalOrFunction
			reduce(240), // >>>, reduce: MlrvalOrFunction
			reduce(240), // +, reduce: MlrvalOrFunction
			reduce(240), // -, reduce: MlrvalOrFunction
			reduce(240), // .+, reduce: MlrvalOrFunction
			reduce(240), // .-, reduce: MlrvalOrFunction
			reduce(240), // *, reduce: MlrvalOrFunction
			reduce(240), // /, reduce: MlrvalOrFunction
			reduce(240), // //, reduce: MlrvalOrFunction
			reduce(240), // %, reduce: MlrvalOrFunction
			reduce(240), // .*, reduce: MlrvalOrFunction
			reduce(240), // ./, reduce: MlrvalOrFunction
			reduce(240), // .//, reduce: MlrvalOrFunction
			reduce(240), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(240), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(240), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(240), // [[, reduce: MlrvalOrFunction
			reduce(240), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3688
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(241), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			reduce(241), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(241), // >, reduce: ContextVariable
			reduce(241), // >>, reduce: ContextVariable
			reduce(241), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(241), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(241), // ||=, reduce: ContextVariable
			reduce(241), // ^^=, reduce: ContextVariable
			reduce(241), // &&=, reduce: ContextVariable
			reduce(241), // ??=, reduce: ContextVariable
			reduce(241), // ???=, reduce: ContextVariable
			reduce(241), // |=, reduce: ContextVariable
			reduce(241), // &=, reduce: ContextVariable
			reduce(241), // ^=, reduce: ContextVariable
			reduce(241), // <<=, reduce: ContextVariable
			reduce(241), // >>=, reduce: ContextVariable
			reduce(241), // >>>=, reduce: ContextVariable
			reduce(241), // +=, reduce: ContextVariable
			reduce(241), // .=, reduce: ContextVariable
			reduce(241), // -=, reduce: ContextVariable
			reduce(241), // *=, reduce: ContextVariable
			reduce(241), // /=, reduce: ContextVariable
			reduce(241), // //=, reduce: ContextVariable
			reduce(241), // %=, reduce: ContextVariable
			reduce(241), // **=, reduce: ContextVariable
			reduce(241), // ?, reduce: ContextVariable
			nil,         // :
			reduce(241), // ||, reduce: ContextVariable
			reduce(241), // ^^, reduce: ContextVariable
			reduce(241), // &&, reduce: ContextVariable
			reduce(241), // ??, reduce: ContextVariable
			reduce(241), // ???, reduce: ContextVariable
			reduce(241), // =~, reduce: ContextVariable
			reduce(241), // !=~, reduce: ContextVariable
			reduce(241), // ==, reduce: ContextVariable
			reduce(241), // !=, reduce: ContextVariable
			reduce(241), // <=>, reduce: ContextVariable
			reduce(241), // >=, reduce: ContextVariable
			reduce(241), // <, reduce: ContextVariable
			reduce(241), // <=, reduce: ContextVariable
			reduce(241), // ^, reduce: ContextVariable
			reduce(241), // &, reduce: ContextVariable
			reduce(241), // <<, reduce: ContextVariable
			reduce(241), // >>>, reduce: ContextVariable
			reduce(241), // +, reduce: ContextVariable
			reduce(241), // -, reduce: ContextVariable
			reduce(241), // .+, reduce: ContextVariable
			reduce(241), // .-, reduce: ContextVariable
			reduce(241), // *, reduce: ContextVariable
			reduce(241), // /, reduce: ContextVariable
			reduce(241), // //, reduce: ContextVariable
			reduce(241), // %, reduce: ContextVariable
			reduce(241), // .*, reduce: ContextVariable
			reduce(241), // ./, reduce: ContextVariable
			reduce(241), // .//, reduce: ContextVariable
			reduce(241), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(241), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(241), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(241), // [[, reduce: ContextVariable
			reduce(241), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3689
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(242), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			reduce(242), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(242), // >, reduce: ContextVariable
			reduce(242), // >>, reduce: ContextVariable
			reduce(242), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(242), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(242), // ||=, reduce: ContextVariable
			reduce(242), // ^^=, reduce: ContextVariable
			reduce(242), // &&=, reduce: ContextVariable
			reduce(242), // ??=, reduce: ContextVariable
			reduce(242), // ???=, reduce: ContextVariable
			reduce(242), // |=, reduce: ContextVariable
			reduce(242), // &=, reduce: ContextVariable
			reduce(242), // ^=, reduce: ContextVariable
			reduce(242), // <<=, reduce: ContextVariable
			reduce(242), // >>=, reduce: ContextVariable
			reduce(242), // >>>=, reduce: ContextVariable
			reduce(242), // +=, reduce: ContextVariable
			reduce(242), // .=, reduce: ContextVariable
			reduce(242), // -=, reduce: ContextVariable
			reduce(242), // *=, reduce: ContextVariable
			reduce(242), // /=, reduce: ContextVariable
			reduce(242), // //=, reduce: ContextVariable
			reduce(242), // %=, reduce: ContextVariable
			reduce(242), // **=, reduce: ContextVariable
			reduce(242), // ?, reduce: ContextVariable
			nil,         // :
			reduce(242), // ||, reduce: ContextVariable
			reduce(242), // ^^, reduce: ContextVariable
			reduce(242), // &&, reduce: ContextVariable
			reduce(242), // ??, reduce: ContextVariable
			reduce(242), // ???, reduce: ContextVariable
			reduce(242), // =~, reduce: ContextVariable
			reduce(242), // !=~, reduce: ContextVariable
			reduce(242), // ==, reduce: ContextVariable
			reduce(242), // !=, reduce: ContextVariable
			reduce(242), // <=>, reduce: ContextVariable
			reduce(242), // >=, reduce: ContextVariable
			reduce(242), // <, reduce: ContextVariable
			reduce(242), // <=, reduce: ContextVariable
			reduce(242), // ^, reduce: ContextVariable
			reduce(242), // &, reduce: ContextVariable
			reduce(242), // <<, reduce: ContextVariable
			reduce(242), // >>>, reduce: ContextVariable
			reduce(242), // +, reduce: ContextVariable
			reduce(242), // -, reduce: ContextVariable
			reduce(242), // .+, reduce: ContextVariable
			reduce(242), // .-, reduce: ContextVariable
			reduce(242), // *, reduce: ContextVariable
			reduce(242), // /, reduce: ContextVariable
			reduce(242), // //, reduce: ContextVariable
			reduce(242), // %, reduce: ContextVariable
			reduce(242), // .*, reduce: ContextVariable
			reduce(242), // ./, reduce: ContextVariable
			reduce(242), // .//, reduce: ContextVariable
			reduce(242), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(242), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(242), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(242), // [[, reduce: ContextVariable
			reduce(242), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3690
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(243), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			reduce(243), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(243), // >, reduce: ContextVariable
			reduce(243), // >>, reduce: ContextVariable
			reduce(243), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(243), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(243), // ||=, reduce: ContextVariable
			reduce(243), // ^^=, reduce: ContextVariable
			reduce(243), // &&=, reduce: ContextVariable
			reduce(243), // ??=, reduce: ContextVariable
			reduce(243), // ???=, reduce: ContextVariable
			reduce(243), // |=, reduce: ContextVariable
			reduce(243), // &=, reduce: ContextVariable
			reduce(243), // ^=, reduce: ContextVariable
			reduce(243), // <<=, reduce: ContextVariable
			reduce(243), // >>=, reduce: ContextVariable
			reduce(243), // >>>=, reduce: ContextVariable
			reduce(243), // +=, reduce: ContextVariable
			reduce(243), // .=, reduce: ContextVariable
			reduce(243), // -=, reduce: ContextVariable
			reduce(243), // *=, reduce: ContextVariable
			reduce(243), // /=, reduce: ContextVariable
			reduce(243), // //=, reduce: ContextVariable
			reduce(243), // %=, reduce: ContextVariable
			reduce(243), // **=, reduce: ContextVariable
			reduce(243), // ?, reduce: ContextVariable
			nil,         // :
			reduce(243), // ||, reduce: ContextVariable
			reduce(243), // ^^, reduce: ContextVariable
			reduce(243), // &&, reduce: ContextVariable
			reduce(243), // ??, reduce: ContextVariable
			reduce(243), // ???, reduce: ContextVariable
			reduce(243), // =~, reduce: ContextVariable
			reduce(243), // !=~, reduce: ContextVariable
			reduce(243), // ==, reduce: ContextVariable
			reduce(243), // !=, reduce: ContextVariable
			reduce(243), // <=>, reduce: ContextVariable
			reduce(243), // >=, reduce: ContextVariable
			reduce(243), // <, reduce: ContextVariable
			reduce(243), // <=, reduce: ContextVariable
			reduce(243), // ^, reduce: ContextVariable
			reduce(243), // &, reduce: ContextVariable
			reduce(243), // <<, reduce: ContextVariable
			reduce(243), // >>>, reduce: ContextVariable
			reduce(243), // +, reduce: ContextVariable
			reduce(243), // -, reduce: ContextVariable
			reduce(243), // .+, reduce: ContextVariable
			reduce(243), // .-, reduce: ContextVariable
			reduce(243), // *, reduce: ContextVariable
			reduce(243), // /, reduce: ContextVariable
			reduce(243), // //, reduce: ContextVariable
			reduce(243), // %, reduce: ContextVariable
			reduce(243), // .*, reduce: ContextVariable
			reduce(243), // ./, reduce: ContextVariable
			reduce(243), // .//, reduce: ContextVariable
			reduce(243), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(243), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(243), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(243), // [[, reduce: ContextVariable
			reduce(243), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3691
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(244), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			reduce(244), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(244), // >, reduce: ContextVariable
			reduce(244), // >>, reduce: ContextVariable
			reduce(244), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(244), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(244), // ||=, reduce: ContextVariable
			reduce(244), // ^^=, reduce: ContextVariable
			reduce(244), // &&=, reduce: ContextVariable
			reduce(244), // ??=, reduce: ContextVariable
			reduce(244), // ???=, reduce: ContextVariable
			reduce(244), // |=, reduce: ContextVariable
			reduce(244), // &=, reduce: ContextVariable
			reduce(244), // ^=, reduce: ContextVariable
			reduce(244), // <<=, reduce: ContextVariable
			reduce(244), // >>=, reduce: ContextVariable
			reduce(244), // >>>=, reduce: ContextVariable
			reduce(244), // +=, reduce: ContextVariable
			reduce(244), // .=, reduce: ContextVariable
			reduce(244), // -=, reduce: ContextVariable
			reduce(244), // *=, reduce: ContextVariable
			reduce(244), // /=, reduce: ContextVariable
			reduce(244), // //=, reduce: ContextVariable
			reduce(244), // %=, reduce: ContextVariable
			reduce(244), // **=, reduce: ContextVariable
			reduce(244), // ?, reduce: ContextVariable
			nil,         // :
			reduce(244), // ||, reduce: ContextVariable
			reduce(244), // ^^, reduce: ContextVariable
			reduce(244), // &&, reduce: ContextVariable
			reduce(244), // ??, reduce: ContextVariable
			reduce(244), // ???, reduce: ContextVariable
			reduce(244), // =~, reduce: ContextVariable
			reduce(244), // !=~, reduce: ContextVariable
			reduce(244), // ==, reduce: ContextVariable
			reduce(244), // !=, reduce: ContextVariable
			reduce(244), // <=>, reduce: ContextVariable
			reduce(244), // >=, reduce: ContextVariable
			reduce(244), // <, reduce: ContextVariable
			reduce(244), // <=, reduce: ContextVariable
			reduce(244), // ^, reduce: ContextVariable
			reduce(244), // &, reduce: ContextVariable
			reduce(244), // <<, reduce: ContextVariable
			reduce(244), // >>>, reduce: ContextVariable
			reduce(244), // +, reduce: ContextVariable
			reduce(244), // -, reduce: ContextVariable
			reduce(244), // .+, reduce: ContextVariable
			reduce(244), // .-, reduce: ContextVariable
			reduce(244), // *, reduce: ContextVariable
			reduce(244), // /, reduce: ContextVariable
			reduce(244), // //, reduce: ContextVariable
			reduce(244), // %, reduce: ContextVariable
			reduce(244), // .*, reduce: ContextVariable
			reduce(244), // ./, reduce: ContextVariable
			reduce(244), // .//, reduce: ContextVariable
			reduce(244), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(244), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(244), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(244), // [[, reduce: ContextVariable
			reduce(244), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3692
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(245), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			reduce(245), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(245), // >, reduce: ContextVariable
			reduce(245), // >>, reduce: ContextVariable
			reduce(245), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(245), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(245), // ||=, reduce: ContextVariable
			reduce(245), // ^^=, reduce: ContextVariable
			reduce(245), // &&=, reduce: ContextVariable
			reduce(245), // ??=, reduce: ContextVariable
			reduce(245), // ???=, reduce: ContextVariable
			reduce(245), // |=, reduce: ContextVariable
			reduce(245), // &=, reduce: ContextVariable
			reduce(245), // ^=, reduce: ContextVariable
			reduce(245), // <<=, reduce: ContextVariable
			reduce(245), // >>=, reduce: ContextVariable
			reduce(245), // >>>=, reduce: ContextVariable
			reduce(245), // +=, reduce: ContextVariable
			reduce(245), // .=, reduce: ContextVariable
			reduce(245), // -=, reduce: ContextVariable
			reduce(245), // *=, reduce: ContextVariable
			reduce(245), // /=, reduce: ContextVariable
			reduce(245), // //=, reduce: ContextVariable
			reduce(245), // %=, reduce: ContextVariable
			reduce(245), // **=, reduce: ContextVariable
			reduce(245), // ?, reduce: ContextVariable
			nil,         // :
			reduce(245), // ||, reduce: ContextVariable
			reduce(245), // ^^, reduce: ContextVariable
			reduce(245), // &&, reduce: ContextVariable
			reduce(245), // ??, reduce: ContextVariable
			reduce(245), // ???, reduce: ContextVariable
			reduce(245), // =~, reduce: ContextVariable
			reduce(245), // !=~, reduce: ContextVariable
			reduce(245), // ==, reduce: ContextVariable
			reduce(245), // !=, reduce: ContextVariable
			reduce(245), // <=>, reduce: ContextVariable
			reduce(245), // >=, reduce: ContextVariable
			reduce(245), // <, reduce: ContextVariable
			reduce(245), // <=, reduce: ContextVariable
			reduce(245), // ^, reduce: ContextVariable
			reduce(245), // &, reduce: ContextVariable
			reduce(245), // <<, reduce: ContextVariable
			reduce(245), // >>>, reduce: ContextVariable
			reduce(245), // +, reduce: ContextVariable
			reduce(245), // -, reduce: ContextVariable
			reduce(245), // .+, reduce: ContextVariable
			reduce(245), // .-, reduce: ContextVariable
			reduce(245), // *, reduce: ContextVariable
			reduce(245), // /, reduce: ContextVariable
			reduce(245), // //, reduce: ContextVariable
			reduce(245), // %, reduce: ContextVariable
			reduce(245), // .*, reduce: ContextVariable
			reduce(245), // ./, reduce: ContextVariable
			reduce(245), // .//, reduce: ContextVariable
			reduce(245), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(245), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(245), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(245), // [[, reduce: ContextVariable
			reduce(245), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3693
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(246), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			reduce(246), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(246), // >, reduce: ContextVariable
			reduce(246), // >>, reduce: ContextVariable
			reduce(246), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(246), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(246), // ||=, reduce: ContextVariable
			reduce(246), // ^^=, reduce: ContextVariable
			reduce(246), // &&=, reduce: ContextVariable
			reduce(246), // ??=, reduce: ContextVariable
			reduce(246), // ???=, reduce: ContextVariable
			reduce(246), // |=, reduce: ContextVariable
			reduce(246), // &=, reduce: ContextVariable
			reduce(246), // ^=, reduce: ContextVariable
			reduce(246), // <<=, reduce: ContextVariable
			reduce(246), // >>=, reduce: ContextVariable
			reduce(246), // >>>=, reduce: ContextVariable
			reduce(246), // +=, reduce: ContextVariable
			reduce(246), // .=, reduce: ContextVariable
			reduce(246), // -=, reduce: ContextVariable
			reduce(246), // *=, reduce: ContextVariable
			reduce(246), // /=, reduce: ContextVariable
			reduce(246), // //=, reduce: ContextVariable
			reduce(246), // %=, reduce: ContextVariable
			reduce(246), // **=, reduce: ContextVariable
			reduce(246), // ?, reduce: ContextVariable
			nil,         // :
			reduce(246), // ||, reduce: ContextVariable
			reduce(246), // ^^, reduce: ContextVariable
			reduce(246), // &&, reduce: ContextVariable
			reduce(246), // ??, reduce: ContextVariable
			reduce(246), // ???, reduce: ContextVariable
			reduce(246), // =~, reduce: ContextVariable
			reduce(246), // !=~, reduce: ContextVariable
			reduce(246), // ==, reduce: ContextVariable
			reduce(246), // !=, reduce: ContextVariable
			reduce(246), // <=>, reduce: ContextVariable
			reduce(246), // >=, reduce: ContextVariable
			reduce(246), // <, reduce: ContextVariable
			reduce(246), // <=, reduce: ContextVariable
			reduce(246), // ^, reduce: ContextVariable
			reduce(246), // &, reduce: ContextVariable
			reduce(246), // <<, reduce: ContextVariable
			reduce(246), // >>>, reduce: ContextVariable
			reduce(246), // +, reduce: ContextVariable
			reduce(246), // -, reduce: ContextVariable
			reduce(246), // .+, reduce: ContextVariable
			reduce(246), // .-, reduce: ContextVariable
			reduce(246), // *, reduce: ContextVariable
			reduce(246), // /, reduce: ContextVariable
			reduce(246), // //, reduce: ContextVariable
			reduce(246), // %, reduce: ContextVariable
			reduce(246), // .*, reduce: ContextVariable
			reduce(246), // ./, reduce: ContextVariable
			reduce(246), // .//, reduce: ContextVariable
			reduce(246), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(246), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(246), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(246), // [[, reduce: ContextVariable
			reduce(246), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3694
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(247), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			reduce(247), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(247), // >, reduce: ContextVariable
			reduce(247), // >>, reduce: ContextVariable
			reduce(247), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(247), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(247), // ||=, reduce: ContextVariable
			reduce(247), // ^^=, reduce: ContextVariable
			reduce(247), // &&=, reduce: ContextVariable
			reduce(247), // ??=, reduce: ContextVariable
			reduce(247), // ???=, reduce: ContextVariable
			reduce(247), // |=, reduce: ContextVariable
			reduce(247), // &=, reduce: ContextVariable
			reduce(247), // ^=, reduce: ContextVariable
			reduce(247), // <<=, reduce: ContextVariable
			reduce(247), // >>=, reduce: ContextVariable
			reduce(247), // >>>=, reduce: ContextVariable
			reduce(247), // +=, reduce: ContextVariable
			reduce(247), // .=, reduce: ContextVariable
			reduce(247), // -=, reduce: ContextVariable
			reduce(247), // *=, reduce: ContextVariable
			reduce(247), // /=, reduce: ContextVariable
			reduce(247), // //=, reduce: ContextVariable
			reduce(247), // %=, reduce: ContextVariable
			reduce(247), // **=, reduce: ContextVariable
			reduce(247), // ?, reduce: ContextVariable
			nil,         // :
			reduce(247), // ||, reduce: ContextVariable
			reduce(247), // ^^, reduce: ContextVariable
			reduce(247), // &&, reduce: ContextVariable
			reduce(247), // ??, reduce: ContextVariable
			reduce(247), // ???, reduce: ContextVariable
			reduce(247), // =~, reduce: ContextVariable
			reduce(247), // !=~, reduce: ContextVariable
			reduce(247), // ==, reduce: ContextVariable
			reduce(247), // !=, reduce: ContextVariable
			reduce(247), // <=>, reduce: ContextVariable
			reduce(247), // >=, reduce: ContextVariable
			reduce(247), // <, reduce: ContextVariable
			reduce(247), // <=, reduce: ContextVariable
			reduce(247), // ^, reduce: ContextVariable
			reduce(247), // &, reduce: ContextVariable
			reduce(247), // <<, reduce: ContextVariable
			reduce(247), // >>>, reduce: ContextVariable
			reduce(247), // +, reduce: ContextVariable
			reduce(247), // -, reduce: ContextVariable
			reduce(247), // .+, reduce: ContextVariable
			reduce(247), // .-, reduce: ContextVariable
			reduce(247), // *, reduce: ContextVariable
			reduce(247), // /, reduce: ContextVariable
			reduce(247), // //, reduce: ContextVariable
			reduce(247), // %, reduce: ContextVariable
			reduce(247), // .*, reduce: ContextVariable
			reduce(247), // ./, reduce: ContextVariable
			reduce(247), // .//, reduce: ContextVariable
			reduce(247), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(247), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(247), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(247), // [[, reduce: ContextVariable
			reduce(247), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3695
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(248), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			reduce(248), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(248), // >, reduce: ContextVariable
			reduce(248), // >>, reduce: ContextVariable
			reduce(248), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(248), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(248), // ||=, reduce: ContextVariable
			reduce(248), // ^^=, reduce: ContextVariable
			reduce(248), // &&=, reduce: ContextVariable
			reduce(248), // ??=, reduce: ContextVariable
			reduce(248), // ???=, reduce: ContextVariable
			reduce(248), // |=, reduce: ContextVariable
			reduce(248), // &=, reduce: ContextVariable
			reduce(248), // ^=, reduce: ContextVariable
			reduce(248), // <<=, reduce: ContextVariable
			reduce(248), // >>=, reduce: ContextVariable
			reduce(248), // >>>=, reduce: ContextVariable
			reduce(248), // +=, reduce: ContextVariable
			reduce(248), // .=, reduce: ContextVariable
			reduce(248), // -=, reduce: ContextVariable
			reduce(248), // *=, reduce: ContextVariable
			reduce(248), // /=, reduce: ContextVariable
			reduce(248), // //=, reduce: ContextVariable
			reduce(248), // %=, reduce: ContextVariable
			reduce(248), // **=, reduce: ContextVariable
			reduce(248), // ?, reduce: ContextVariable
			nil,         // :
			reduce(248), // ||, reduce: ContextVariable
			reduce(248), // ^^, reduce: ContextVariable
			reduce(248), // &&, reduce: ContextVariable
			reduce(248), // ??, reduce: ContextVariable
			reduce(248), // ???, reduce: ContextVariable
			reduce(248), // =~, reduce: ContextVariable
			reduce(248), // !=~, reduce: ContextVariable
			reduce(248), // ==, reduce: ContextVariable
			reduce(248), // !=, reduce: ContextVariable
			reduce(248), // <=>, reduce: ContextVariable
			reduce(248), // >=, reduce: ContextVariable
			reduce(248), // <, reduce: ContextVariable
			reduce(248), // <=, reduce: ContextVariable
			reduce(248), // ^, reduce: ContextVariable
			reduce(248), // &, reduce: ContextVariable
			reduce(248), // <<, reduce: ContextVariable
			reduce(248), // >>>, reduce: ContextVariable
			reduce(248), // +, reduce: ContextVariable
			reduce(248), // -, reduce: ContextVariable
			reduce(248), // .+, reduce: ContextVariable
			reduce(248), // .-, reduce: ContextVariable
			reduce(248), // *, reduce: ContextVariable
			reduce(248), // /, reduce: ContextVariable
			reduce(248), // //, reduce: ContextVariable
			reduce(248), // %, reduce: ContextVariable
			reduce(248), // .*, reduce: ContextVariable
			reduce(248), // ./, reduce: ContextVariable
			reduce(248), // .//, reduce: ContextVariable
			reduce(248), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(248), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(248), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(248), // [[, reduce: ContextVariable
			reduce(248), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3696
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(249), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			reduce(249), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(249), // >, reduce: ContextVariable
			reduce(249), // >>, reduce: ContextVariable
			reduce(249), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(249), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(249), // ||=, reduce: ContextVariable
			reduce(249), // ^^=, reduce: ContextVariable
			reduce(249), // &&=, reduce: ContextVariable
			reduce(249), // ??=, reduce: ContextVariable
			reduce(249), // ???=, reduce: ContextVariable
			reduce(249), // |=, reduce: ContextVariable
			reduce(249), // &=, reduce: ContextVariable
			reduce(249), // ^=, reduce: ContextVariable
			reduce(249), // <<=, reduce: ContextVariable
			reduce(249), // >>=, reduce: ContextVariable
			reduce(249), // >>>=, reduce: ContextVariable
			reduce(249), // +=, reduce: ContextVariable
			reduce(249), // .=, reduce: ContextVariable
			reduce(249), // -=, reduce: ContextVariable
			reduce(249), // *=, reduce: ContextVariable
			reduce(249), // /=, reduce: ContextVariable
			reduce(249), // //=, reduce: ContextVariable
			reduce(249), // %=, reduce: ContextVariable
			reduce(249), // **=, reduce: ContextVariable
			reduce(249), // ?, reduce: ContextVariable
			nil,         // :
			reduce(249), // ||, reduce: ContextVariable
			reduce(249), // ^^, reduce: ContextVariable
			reduce(249), // &&, reduce: ContextVariable
			reduce(249), // ??, reduce: ContextVariable
			reduce(249), // ???, reduce: ContextVariable
			reduce(249), // =~, reduce: ContextVariable
			reduce(249), // !=~, reduce: ContextVariable
			reduce(249), // ==, reduce: ContextVariable
			reduce(249), // !=, reduce: ContextVariable
			reduce(249), // <=>, reduce: ContextVariable
			reduce(249), // >=, reduce: ContextVariable
			reduce(249), // <, reduce: ContextVariable
			reduce(249), // <=, reduce: ContextVariable
			reduce(249), // ^, reduce: ContextVariable
			reduce(249), // &, reduce: ContextVariable
			reduce(249), // <<, reduce: ContextVariable
			reduce(249), // >>>, reduce: ContextVariable
			reduce(249), // +, reduce: ContextVariable
			reduce(249), // -, reduce: ContextVariable
			reduce(249), // .+, reduce: ContextVariable
			reduce(249), // .-, reduce: ContextVariable
			reduce(249), // *, reduce: ContextVariable
			reduce(249), // /, reduce: ContextVariable
			reduce(249), // //, reduce: ContextVariable
			reduce(249), // %, reduce: ContextVariable
			reduce(249), // .*, reduce: ContextVariable
			reduce(249), // ./, reduce: ContextVariable
			reduce(249), // .//, reduce: ContextVariable
			reduce(249), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(249), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(249), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(249), // [[, reduce: ContextVariable
			reduce(249), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3697
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(250), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			reduce(250), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(250), // >, reduce: ContextVariable
			reduce(250), // >>, reduce: ContextVariable
			reduce(250), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(250), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(250), // ||=, reduce: ContextVariable
			reduce(250), // ^^=, reduce: ContextVariable
			reduce(250), // &&=, reduce: ContextVariable
			reduce(250), // ??=, reduce: ContextVariable
			reduce(250), // ???=, reduce: ContextVariable
			reduce(250), // |=, reduce: ContextVariable
			reduce(250), // &=, reduce: ContextVariable
			reduce(250), // ^=, reduce: ContextVariable
			reduce(250), // <<=, reduce: ContextVariable
			reduce(250), // >>=, reduce: ContextVariable
			reduce(250), // >>>=, reduce: ContextVariable
			reduce(250), // +=, reduce: ContextVariable
			reduce(250), // .=, reduce: ContextVariable
			reduce(250), // -=, reduce: ContextVariable
			reduce(250), // *=, reduce: ContextVariable
			reduce(250), // /=, reduce: ContextVariable
			reduce(250), // //=, reduce: ContextVariable
			reduce(250), // %=, reduce: ContextVariable
			reduce(250), // **=, reduce: ContextVariable
			reduce(250), // ?, reduce: ContextVariable
			nil,         // :
			reduce(250), // ||, reduce: ContextVariable
			reduce(250), // ^^, reduce: ContextVariable
			reduce(250), // &&, reduce: ContextVariable
			reduce(250), // ??, reduce: ContextVariable
			reduce(250), // ???, reduce: ContextVariable
			reduce(250), // =~, reduce: ContextVariable
			reduce(250), // !=~, reduce: ContextVariable
			reduce(250), // ==, reduce: ContextVariable
			reduce(250), // !=, reduce: ContextVariable
			reduce(250), // <=>, reduce: ContextVariable
			reduce(250), // >=, reduce: ContextVariable
			reduce(250), // <, reduce: ContextVariable
			reduce(250), // <=, reduce: ContextVariable
			reduce(250), // ^, reduce: ContextVariable
			reduce(250), // &, reduce: ContextVariable
			reduce(250), // <<, reduce: ContextVariable
			reduce(250), // >>>, reduce: ContextVariable
			reduce(250), // +, reduce: ContextVariable
			reduce(250), // -, reduce: ContextVariable
			reduce(250), // .+, reduce: ContextVariable
			reduce(250), // .-, reduce: ContextVariable
			reduce(250), // *, reduce: ContextVariable
			reduce(250), // /, reduce: ContextVariable
			reduce(250), // //, reduce: ContextVariable
			reduce(250), // %, reduce: ContextVariable
			reduce(250), // .*, reduce: ContextVariable
			reduce(250), // ./, reduce: ContextVariable
			reduce(250), // .//, reduce: ContextVariable
			reduce(250), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(250), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(250), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(250), // [[, reduce: ContextVariable
			reduce(250), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3698
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(251), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			reduce(251), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(251), // >, reduce: ContextVariable
			reduce(251), // >>, reduce: ContextVariable
			reduce(251), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(251), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(251), // ||=, reduce: ContextVariable
			reduce(251), // ^^=, reduce: ContextVariable
			reduce(251), // &&=, reduce: ContextVariable
			reduce(251), // ??=, reduce: ContextVariable
			reduce(251), // ???=, reduce: ContextVariable
			reduce(251), // |=, reduce: ContextVariable
			reduce(251), // &=, reduce: ContextVariable
			reduce(251), // ^=, reduce: ContextVariable
			reduce(251), // <<=, reduce: ContextVariable
			reduce(251), // >>=, reduce: ContextVariable
			reduce(251), // >>>=, reduce: ContextVariable
			reduce(251), // +=, reduce: ContextVariable
			reduce(251), // .=, reduce: ContextVariable
			reduce(251), // -=, reduce: ContextVariable
			reduce(251), // *=, reduce: ContextVariable
			reduce(251), // /=, reduce: ContextVariable
			reduce(251), // //=, reduce: ContextVariable
			reduce(251), // %=, reduce: ContextVariable
			reduce(251), // **=, reduce: ContextVariable
			reduce(251), // ?, reduce: ContextVariable
			nil,         // :
			reduce(251), // ||, reduce: ContextVariable
			reduce(251), // ^^, reduce: ContextVariable
			reduce(251), // &&, reduce: ContextVariable
			reduce(251), // ??, reduce: ContextVariable
			reduce(251), // ???, reduce: ContextVariable
			reduce(251), // =~, reduce: ContextVariable
			reduce(251), // !=~, reduce: ContextVariable
			reduce(251), // ==, reduce: ContextVariable
			reduce(251), // !=, reduce: ContextVariable
			reduce(251), // <=>, reduce: ContextVariable
			reduce(251), // >=, reduce: ContextVariable
			reduce(251), // <, reduce: ContextVariable
			reduce(251), // <=, reduce: ContextVariable
			reduce(251), // ^, reduce: ContextVariable
			reduce(251), // &, reduce: ContextVariable
			reduce(251), // <<, reduce: ContextVariable
			reduce(251), // >>>, reduce: ContextVariable
			reduce(251), // +, reduce: ContextVariable
			reduce(251), // -, reduce: ContextVariable
			reduce(251), // .+, reduce: ContextVariable
			reduce(251), // .-, reduce: ContextVariable
			reduce(251), // *, reduce: ContextVariable
			reduce(251), // /, reduce: ContextVariable
			reduce(251), // //, reduce: ContextVariable
			reduce(251), // %, reduce: ContextVariable
			reduce(251), // .*, reduce: ContextVariable
			reduce(251), // ./, reduce: ContextVariable
			reduce(251), // .//, reduce: ContextVariable
			reduce(251), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(251), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(251), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(251), // [[, reduce: ContextVariable
			reduce(251), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3699
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(252), // ;, reduce: ContextVariable
			nil,         // {
			nil,         // }
			reduce(252), // =, reduce: ContextVariable
			nil,         // unset
			nil,         // filter
			reduce(252), // >, reduce: ContextVariable
			reduce(252), // >>, reduce: ContextVariable
			reduce(252), // |, reduce: ContextVariable
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(252), // ,, reduce: ContextVariable
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(252), // ||=, reduce: ContextVariable
			reduce(252), // ^^=, reduce: ContextVariable
			reduce(252), // &&=, reduce: ContextVariable
			reduce(252), // ??=, reduce: ContextVariable
			reduce(252), // ???=, reduce: ContextVariable
			reduce(252), // |=, reduce: ContextVariable
			reduce(252), // &=, reduce: ContextVariable
			reduce(252), // ^=, reduce: ContextVariable
			reduce(252), // <<=, reduce: ContextVariable
			reduce(252), // >>=, reduce: ContextVariable
			reduce(252), // >>>=, reduce: ContextVariable
			reduce(252), // +=, reduce: ContextVariable
			reduce(252), // .=, reduce: ContextVariable
			reduce(252), // -=, reduce: ContextVariable
			reduce(252), // *=, reduce: ContextVariable
			reduce(252), // /=, reduce: ContextVariable
			reduce(252), // //=, reduce: ContextVariable
			reduce(252), // %=, reduce: ContextVariable
			reduce(252), // **=, reduce: ContextVariable
			reduce(252), // ?, reduce: ContextVariable
			nil,         // :
			reduce(252), // ||, reduce: ContextVariable
			reduce(252), // ^^, reduce: ContextVariable
			reduce(252), // &&, reduce: ContextVariable
			reduce(252), // ??, reduce: ContextVariable
			reduce(252), // ???, reduce: ContextVariable
			reduce(252), // =~, reduce: ContextVariable
			reduce(252), // !=~, reduce: ContextVariable
			reduce(252), // ==, reduce: ContextVariable
			reduce(252), // !=, reduce: ContextVariable
			reduce(252), // <=>, reduce: ContextVariable
			reduce(252), // >=, reduce: ContextVariable
			reduce(252), // <, reduce: ContextVariable
			reduce(252), // <=, reduce: ContextVariable
			reduce(252), // ^, reduce: ContextVariable
			reduce(252), // &, reduce: ContextVariable
			reduce(252), // <<, reduce: ContextVariable
			reduce(252), // >>>, reduce: ContextVariable
			reduce(252), // +, reduce: ContextVariable
			reduce(252), // -, reduce: ContextVariable
			reduce(252), // .+, reduce: ContextVariable
			reduce(252), // .-, reduce: ContextVariable
			reduce(252), // *, reduce: ContextVariable
			reduce(252), // /, reduce: ContextVariable
			reduce(252), // //, reduce: ContextVariable
			reduce(252), // %, reduce: ContextVariable
			reduce(252), // .*, reduce: ContextVariable
			reduce(252), // ./, reduce: ContextVariable
			reduce(252), // .//, reduce: ContextVariable
			reduce(252), // ., reduce: ContextVariable
			nil,         // !
			nil,         // ~
			reduce(252), // **, reduce: ContextVariable
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(252), // [, reduce: ContextVariable
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(252), // [[, reduce: ContextVariable
			reduce(252), // [[[, reduce: ContextVariable
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3700
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(253), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(253), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(253), // >, reduce: MlrvalOrFunction
			reduce(253), // >>, reduce: MlrvalOrFunction
			reduce(253), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(253), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(253), // ||=, reduce: MlrvalOrFunction
			reduce(253), // ^^=, reduce: MlrvalOrFunction
			reduce(253), // &&=, reduce: MlrvalOrFunction
			reduce(253), // ??=, reduce: MlrvalOrFunction
			reduce(253), // ???=, reduce: MlrvalOrFunction
			reduce(253), // |=, reduce: MlrvalOrFunction
			reduce(253), // &=, reduce: MlrvalOrFunction
			reduce(253), // ^=, reduce: MlrvalOrFunction
			reduce(253), // <<=, reduce: MlrvalOrFunction
			reduce(253), // >>=, reduce: MlrvalOrFunction
			reduce(253), // >>>=, reduce: MlrvalOrFunction
			reduce(253), // +=, reduce: MlrvalOrFunction
			reduce(253), // .=, reduce: MlrvalOrFunction
			reduce(253), // -=, reduce: MlrvalOrFunction
			reduce(253), // *=, reduce: MlrvalOrFunction
			reduce(253), // /=, reduce: MlrvalOrFunction
			reduce(253), // //=, reduce: MlrvalOrFunction
			reduce(253), // %=, reduce: MlrvalOrFunction
			reduce(253), // **=, reduce: MlrvalOrFunction
			reduce(253), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(253), // ||, reduce: MlrvalOrFunction
			reduce(253), // ^^, reduce: MlrvalOrFunction
			reduce(253), // &&, reduce: MlrvalOrFunction
			reduce(253), // ??, reduce: MlrvalOrFunction
			reduce(253), // ???, reduce: MlrvalOrFunction
			reduce(253), // =~, reduce: MlrvalOrFunction
			reduce(253), // !=~, reduce: MlrvalOrFunction
			reduce(253), // ==, reduce: MlrvalOrFunction
			reduce(253), // !=, reduce: MlrvalOrFunction
			reduce(253), // <=>, reduce: MlrvalOrFunction
			reduce(253), // >=, reduce: MlrvalOrFunction
			reduce(253), // <, reduce: MlrvalOrFunction
			reduce(253), // <=, reduce: MlrvalOrFunction
			reduce(253), // ^, reduce: MlrvalOrFunction
			reduce(253), // &, reduce: MlrvalOrFunction
			reduce(253), // <<, reduce: MlrvalOrFunction
			reduce(253), // >>>, reduce: MlrvalOrFunction
			reduce(253), // +, reduce: MlrvalOrFunction
			reduce(253), // -, reduce: MlrvalOrFunction
			reduce(253), // .+, reduce: MlrvalOrFunction
			reduce(253), // .-, reduce: MlrvalOrFunction
			reduce(253), // *, reduce: MlrvalOrFunction
			reduce(253), // /, reduce: MlrvalOrFunction
			reduce(253), // //, reduce: MlrvalOrFunction
			reduce(253), // %, reduce: MlrvalOrFunction
			reduce(253), // .*, reduce: MlrvalOrFunction
			reduce(253), // ./, reduce: MlrvalOrFunction
			reduce(253), // .//, reduce: MlrvalOrFunction
			reduce(253), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(253), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(253), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(253), // [[, reduce: MlrvalOrFunction
			reduce(253), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3701
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			shift(4163), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			shift(4164), // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3702
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(256), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(256), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(256), // >, reduce: MlrvalOrFunction
			reduce(256), // >>, reduce: MlrvalOrFunction
			reduce(256), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(256), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(256), // ||=, reduce: MlrvalOrFunction
			reduce(256), // ^^=, reduce: MlrvalOrFunction
			reduce(256), // &&=, reduce: MlrvalOrFunction
			reduce(256), // ??=, reduce: MlrvalOrFunction
			reduce(256), // ???=, reduce: MlrvalOrFunction
			reduce(256), // |=, reduce: MlrvalOrFunction
			reduce(256), // &=, reduce: MlrvalOrFunction
			reduce(256), // ^=, reduce: MlrvalOrFunction
			reduce(256), // <<=, reduce: MlrvalOrFunction
			reduce(256), // >>=, reduce: MlrvalOrFunction
			reduce(256), // >>>=, reduce: MlrvalOrFunction
			reduce(256), // +=, reduce: MlrvalOrFunction
			reduce(256), // .=, reduce: MlrvalOrFunction
			reduce(256), // -=, reduce: MlrvalOrFunction
			reduce(256), // *=, reduce: MlrvalOrFunction
			reduce(256), // /=, reduce: MlrvalOrFunction
			reduce(256), // //=, reduce: MlrvalOrFunction
			reduce(256), // %=, reduce: MlrvalOrFunction
			reduce(256), // **=, reduce: MlrvalOrFunction
			reduce(256), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(256), // ||, reduce: MlrvalOrFunction
			reduce(256), // ^^, reduce: MlrvalOrFunction
			reduce(256), // &&, reduce: MlrvalOrFunction
			reduce(256), // ??, reduce: MlrvalOrFunction
			reduce(256), // ???, reduce: MlrvalOrFunction
			reduce(256), // =~, reduce: MlrvalOrFunction
			reduce(256), // !=~, reduce: MlrvalOrFunction
			reduce(256), // ==, reduce: MlrvalOrFunction
			reduce(256), // !=, reduce: MlrvalOrFunction
			reduce(256), // <=>, reduce: MlrvalOrFunction
			reduce(256), // >=, reduce: MlrvalOrFunction
			reduce(256), // <, reduce: MlrvalOrFunction
			reduce(256), // <=, reduce: MlrvalOrFunction
			reduce(256), // ^, reduce: MlrvalOrFunction
			reduce(256), // &, reduce: MlrvalOrFunction
			reduce(256), // <<, reduce: MlrvalOrFunction
			reduce(256), // >>>, reduce: MlrvalOrFunction
			reduce(256), // +, reduce: MlrvalOrFunction
			reduce(256), // -, reduce: MlrvalOrFunction
			reduce(256), // .+, reduce: MlrvalOrFunction
			reduce(256), // .-, reduce: MlrvalOrFunction
			reduce(256), // *, reduce: MlrvalOrFunction
			reduce(256), // /, reduce: MlrvalOrFunction
			reduce(256), // //, reduce: MlrvalOrFunction
			reduce(256), // %, reduce: MlrvalOrFunction
			reduce(256), // .*, reduce: MlrvalOrFunction
			reduce(256), // ./, reduce: MlrvalOrFunction
			reduce(256), // .//, reduce: MlrvalOrFunction
			reduce(256), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(256), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(256), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(256), // [[, reduce: MlrvalOrFunction
			reduce(256), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3703
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(257), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(257), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(257), // >, reduce: MlrvalOrFunction
			reduce(257), // >>, reduce: MlrvalOrFunction
			reduce(257), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(257), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(257), // ||=, reduce: MlrvalOrFunction
			reduce(257), // ^^=, reduce: MlrvalOrFunction
			reduce(257), // &&=, reduce: MlrvalOrFunction
			reduce(257), // ??=, reduce: MlrvalOrFunction
			reduce(257), // ???=, reduce: MlrvalOrFunction
			reduce(257), // |=, reduce: MlrvalOrFunction
			reduce(257), // &=, reduce: MlrvalOrFunction
			reduce(257), // ^=, reduce: MlrvalOrFunction
			reduce(257), // <<=, reduce: MlrvalOrFunction
			reduce(257), // >>=, reduce: MlrvalOrFunction
			reduce(257), // >>>=, reduce: MlrvalOrFunction
			reduce(257), // +=, reduce: MlrvalOrFunction
			reduce(257), // .=, reduce: MlrvalOrFunction
			reduce(257), // -=, reduce: MlrvalOrFunction
			reduce(257), // *=, reduce: MlrvalOrFunction
			reduce(257), // /=, reduce: MlrvalOrFunction
			reduce(257), // //=, reduce: MlrvalOrFunction
			reduce(257), // %=, reduce: MlrvalOrFunction
			reduce(257), // **=, reduce: MlrvalOrFunction
			reduce(257), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(257), // ||, reduce: MlrvalOrFunction
			reduce(257), // ^^, reduce: MlrvalOrFunction
			reduce(257), // &&, reduce: MlrvalOrFunction
			reduce(257), // ??, reduce: MlrvalOrFunction
			reduce(257), // ???, reduce: MlrvalOrFunction
			reduce(257), // =~, reduce: MlrvalOrFunction
			reduce(257), // !=~, reduce: MlrvalOrFunction
			reduce(257), // ==, reduce: MlrvalOrFunction
			reduce(257), // !=, reduce: MlrvalOrFunction
			reduce(257), // <=>, reduce: MlrvalOrFunction
			reduce(257), // >=, reduce: MlrvalOrFunction
			reduce(257), // <, reduce: MlrvalOrFunction
			reduce(257), // <=, reduce: MlrvalOrFunction
			reduce(257), // ^, reduce: MlrvalOrFunction
			reduce(257), // &, reduce: MlrvalOrFunction
			reduce(257), // <<, reduce: MlrvalOrFunction
			reduce(257), // >>>, reduce: MlrvalOrFunction
			reduce(257), // +, reduce: MlrvalOrFunction
			reduce(257), // -, reduce: MlrvalOrFunction
			reduce(257), // .+, reduce: MlrvalOrFunction
			reduce(257), // .-, reduce: MlrvalOrFunction
			reduce(257), // *, reduce: MlrvalOrFunction
			reduce(257), // /, reduce: MlrvalOrFunction
			reduce(257), // //, reduce: MlrvalOrFunction
			reduce(257), // %, reduce: MlrvalOrFunction
			reduce(257), // .*, reduce: MlrvalOrFunction
			reduce(257), // ./, reduce: MlrvalOrFunction
			reduce(257), // .//, reduce: MlrvalOrFunction
			reduce(257), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(257), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(257), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(257), // [[, reduce: MlrvalOrFunction
			reduce(257), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3704
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(258), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(258), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(258), // >, reduce: MlrvalOrFunction
			reduce(258), // >>, reduce: MlrvalOrFunction
			reduce(258), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(258), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(258), // ||=, reduce: MlrvalOrFunction
			reduce(258), // ^^=, reduce: MlrvalOrFunction
			reduce(258), // &&=, reduce: MlrvalOrFunction
			reduce(258), // ??=, reduce: MlrvalOrFunction
			reduce(258), // ???=, reduce: MlrvalOrFunction
			reduce(258), // |=, reduce: MlrvalOrFunction
			reduce(258), // &=, reduce: MlrvalOrFunction
			reduce(258), // ^=, reduce: MlrvalOrFunction
			reduce(258), // <<=, reduce: MlrvalOrFunction
			reduce(258), // >>=, reduce: MlrvalOrFunction
			reduce(258), // >>>=, reduce: MlrvalOrFunction
			reduce(258), // +=, reduce: MlrvalOrFunction
			reduce(258), // .=, reduce: MlrvalOrFunction
			reduce(258), // -=, reduce: MlrvalOrFunction
			reduce(258), // *=, reduce: MlrvalOrFunction
			reduce(258), // /=, reduce: MlrvalOrFunction
			reduce(258), // //=, reduce: MlrvalOrFunction
			reduce(258), // %=, reduce: MlrvalOrFunction
			reduce(258), // **=, reduce: MlrvalOrFunction
			reduce(258), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(258), // ||, reduce: MlrvalOrFunction
			reduce(258), // ^^, reduce: MlrvalOrFunction
			reduce(258), // &&, reduce: MlrvalOrFunction
			reduce(258), // ??, reduce: MlrvalOrFunction
			reduce(258), // ???, reduce: MlrvalOrFunction
			reduce(258), // =~, reduce: MlrvalOrFunction
			reduce(258), // !=~, reduce: MlrvalOrFunction
			reduce(258), // ==, reduce: MlrvalOrFunction
			reduce(258), // !=, reduce: MlrvalOrFunction
			reduce(258), // <=>, reduce: MlrvalOrFunction
			reduce(258), // >=, reduce: MlrvalOrFunction
			reduce(258), // <, reduce: MlrvalOrFunction
			reduce(258), // <=, reduce: MlrvalOrFunction
			reduce(258), // ^, reduce: MlrvalOrFunction
			reduce(258), // &, reduce: MlrvalOrFunction
			reduce(258), // <<, reduce: MlrvalOrFunction
			reduce(258), // >>>, reduce: MlrvalOrFunction
			reduce(258), // +, reduce: MlrvalOrFunction
			reduce(258), // -, reduce: MlrvalOrFunction
			reduce(258), // .+, reduce: MlrvalOrFunction
			reduce(258), // .-, reduce: MlrvalOrFunction
			reduce(258), // *, reduce: MlrvalOrFunction
			reduce(258), // /, reduce: MlrvalOrFunction
			reduce(258), // //, reduce: MlrvalOrFunction
			reduce(258), // %, reduce: MlrvalOrFunction
			reduce(258), // .*, reduce: MlrvalOrFunction
			reduce(258), // ./, reduce: MlrvalOrFunction
			reduce(258), // .//, reduce: MlrvalOrFunction
			reduce(258), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(258), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(258), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(258), // [[, reduce: MlrvalOrFunction
			reduce(258), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3705
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(259), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(259), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(259), // >, reduce: MlrvalOrFunction
			reduce(259), // >>, reduce: MlrvalOrFunction
			reduce(259), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(259), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(259), // ||=, reduce: MlrvalOrFunction
			reduce(259), // ^^=, reduce: MlrvalOrFunction
			reduce(259), // &&=, reduce: MlrvalOrFunction
			reduce(259), // ??=, reduce: MlrvalOrFunction
			reduce(259), // ???=, reduce: MlrvalOrFunction
			reduce(259), // |=, reduce: MlrvalOrFunction
			reduce(259), // &=, reduce: MlrvalOrFunction
			reduce(259), // ^=, reduce: MlrvalOrFunction
			reduce(259), // <<=, reduce: MlrvalOrFunction
			reduce(259), // >>=, reduce: MlrvalOrFunction
			reduce(259), // >>>=, reduce: MlrvalOrFunction
			reduce(259), // +=, reduce: MlrvalOrFunction
			reduce(259), // .=, reduce: MlrvalOrFunction
			reduce(259), // -=, reduce: MlrvalOrFunction
			reduce(259), // *=, reduce: MlrvalOrFunction
			reduce(259), // /=, reduce: MlrvalOrFunction
			reduce(259), // //=, reduce: MlrvalOrFunction
			reduce(259), // %=, reduce: MlrvalOrFunction
			reduce(259), // **=, reduce: MlrvalOrFunction
			reduce(259), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(259), // ||, reduce: MlrvalOrFunction
			reduce(259), // ^^, reduce: MlrvalOrFunction
			reduce(259), // &&, reduce: MlrvalOrFunction
			reduce(259), // ??, reduce: MlrvalOrFunction
			reduce(259), // ???, reduce: MlrvalOrFunction
			reduce(259), // =~, reduce: MlrvalOrFunction
			reduce(259), // !=~, reduce: MlrvalOrFunction
			reduce(259), // ==, reduce: MlrvalOrFunction
			reduce(259), // !=, reduce: MlrvalOrFunction
			reduce(259), // <=>, reduce: MlrvalOrFunction
			reduce(259), // >=, reduce: MlrvalOrFunction
			reduce(259), // <, reduce: MlrvalOrFunction
			reduce(259), // <=, reduce: MlrvalOrFunction
			reduce(259), // ^, reduce: MlrvalOrFunction
			reduce(259), // &, reduce: MlrvalOrFunction
			reduce(259), // <<, reduce: MlrvalOrFunction
			reduce(259), // >>>, reduce: MlrvalOrFunction
			reduce(259), // +, reduce: MlrvalOrFunction
			reduce(259), // -, reduce: MlrvalOrFunction
			reduce(259), // .+, reduce: MlrvalOrFunction
			reduce(259), // .-, reduce: MlrvalOrFunction
			reduce(259), // *, reduce: MlrvalOrFunction
			reduce(259), // /, reduce: MlrvalOrFunction
			reduce(259), // //, reduce: MlrvalOrFunction
			reduce(259), // %, reduce: MlrvalOrFunction
			reduce(259), // .*, reduce: MlrvalOrFunction
			reduce(259), // ./, reduce: MlrvalOrFunction
			reduce(259), // .//, reduce: MlrvalOrFunction
			reduce(259), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(259), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(259), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(259), // [[, reduce: MlrvalOrFunction
			reduce(259), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3706
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(267), // ;, reduce: MlrvalOrFunction
			nil,         // {
			nil,         // }
			reduce(267), // =, reduce: MlrvalOrFunction
			nil,         // unset
			nil,         // filter
			reduce(267), // >, reduce: MlrvalOrFunction
			reduce(267), // >>, reduce: MlrvalOrFunction
			reduce(267), // |, reduce: MlrvalOrFunction
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(267), // ,, reduce: MlrvalOrFunction
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(267), // ||=, reduce: MlrvalOrFunction
			reduce(267), // ^^=, reduce: MlrvalOrFunction
			reduce(267), // &&=, reduce: MlrvalOrFunction
			reduce(267), // ??=, reduce: MlrvalOrFunction
			reduce(267), // ???=, reduce: MlrvalOrFunction
			reduce(267), // |=, reduce: MlrvalOrFunction
			reduce(267), // &=, reduce: MlrvalOrFunction
			reduce(267), // ^=, reduce: MlrvalOrFunction
			reduce(267), // <<=, reduce: MlrvalOrFunction
			reduce(267), // >>=, reduce: MlrvalOrFunction
			reduce(267), // >>>=, reduce: MlrvalOrFunction
			reduce(267), // +=, reduce: MlrvalOrFunction
			reduce(267), // .=, reduce: MlrvalOrFunction
			reduce(267), // -=, reduce: MlrvalOrFunction
			reduce(267), // *=, reduce: MlrvalOrFunction
			reduce(267), // /=, reduce: MlrvalOrFunction
			reduce(267), // //=, reduce: MlrvalOrFunction
			reduce(267), // %=, reduce: MlrvalOrFunction
			reduce(267), // **=, reduce: MlrvalOrFunction
			reduce(267), // ?, reduce: MlrvalOrFunction
			nil,         // :
			reduce(267), // ||, reduce: MlrvalOrFunction
			reduce(267), // ^^, reduce: MlrvalOrFunction
			reduce(267), // &&, reduce: MlrvalOrFunction
			reduce(267), // ??, reduce: MlrvalOrFunction
			reduce(267), // ???, reduce: MlrvalOrFunction
			reduce(267), // =~, reduce: MlrvalOrFunction
			reduce(267), // !=~, reduce: MlrvalOrFunction
			reduce(267), // ==, reduce: MlrvalOrFunction
			reduce(267), // !=, reduce: MlrvalOrFunction
			reduce(267), // <=>, reduce: MlrvalOrFunction
			reduce(267), // >=, reduce: MlrvalOrFunction
			reduce(267), // <, reduce: MlrvalOrFunction
			reduce(267), // <=, reduce: MlrvalOrFunction
			reduce(267), // ^, reduce: MlrvalOrFunction
			reduce(267), // &, reduce: MlrvalOrFunction
			reduce(267), // <<, reduce: MlrvalOrFunction
			reduce(267), // >>>, reduce: MlrvalOrFunction
			reduce(267), // +, reduce: MlrvalOrFunction
			reduce(267), // -, reduce: MlrvalOrFunction
			reduce(267), // .+, reduce: MlrvalOrFunction
			reduce(267), // .-, reduce: MlrvalOrFunction
			reduce(267), // *, reduce: MlrvalOrFunction
			reduce(267), // /, reduce: MlrvalOrFunction
			reduce(267), // //, reduce: MlrvalOrFunction
			reduce(267), // %, reduce: MlrvalOrFunction
			reduce(267), // .*, reduce: MlrvalOrFunction
			reduce(267), // ./, reduce: MlrvalOrFunction
			reduce(267), // .//, reduce: MlrvalOrFunction
			reduce(267), // ., reduce: MlrvalOrFunction
			nil,         // !
			nil,         // ~
			reduce(267), // **, reduce: MlrvalOrFunction
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(267), // [, reduce: MlrvalOrFunction
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(267), // [[, reduce: MlrvalOrFunction
			reduce(267), // [[[, reduce: MlrvalOrFunction
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3707
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(4165), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3708
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(4166), // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(4167), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3709
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(313), // ;, reduce: TripleForContinuation
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(313), // ,, reduce: TripleForContinuation
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3710
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(4168), // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3711
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(311), // ;, reduce: TripleForStart
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(311), // ,, reduce: TripleForStart
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3712
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4169), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3713
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(324), // $, reduce: UnnamedFunctionDefinition
			nil,         // empty
			reduce(324), // ;, reduce: UnnamedFunctionDefinition
			reduce(324), // {, reduce: UnnamedFunctionDefinition
			nil,         // }
			reduce(324), // =, reduce: UnnamedFunctionDefinition
			nil,         // unset
			nil,         // filter
			reduce(324), // >, reduce: UnnamedFunctionDefinition
			reduce(324), // >>, reduce: UnnamedFunctionDefinition
			reduce(324), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(324), // ||=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^=, reduce: UnnamedFunctionDefinition
			reduce(324), // &&=, reduce: UnnamedFunctionDefinition
			reduce(324), // ??=, reduce: UnnamedFunctionDefinition
			reduce(324), // ???=, reduce: UnnamedFunctionDefinition
			reduce(324), // |=, reduce: UnnamedFunctionDefinition
			reduce(324), // &=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^=, reduce: UnnamedFunctionDefinition
			reduce(324), // <<=, reduce: UnnamedFunctionDefinition
			reduce(324), // >>=, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>=, reduce: UnnamedFunctionDefinition
			reduce(324), // +=, reduce: UnnamedFunctionDefinition
			reduce(324), // .=, reduce: UnnamedFunctionDefinition
			reduce(324), // -=, reduce: UnnamedFunctionDefinition
			reduce(324), // *=, reduce: UnnamedFunctionDefinition
			reduce(324), // /=, reduce: UnnamedFunctionDefinition
			reduce(324), // //=, reduce: UnnamedFunctionDefinition
			reduce(324), // %=, reduce: UnnamedFunctionDefinition
			reduce(324), // **=, reduce: UnnamedFunctionDefinition
			reduce(324), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(324), // ||, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^, reduce: UnnamedFunctionDefinition
			reduce(324), // &&, reduce: UnnamedFunctionDefinition
			reduce(324), // ??, reduce: UnnamedFunctionDefinition
			reduce(324), // ???, reduce: UnnamedFunctionDefinition
			reduce(324), // =~, reduce: UnnamedFunctionDefinition
			reduce(324), // !=~, reduce: UnnamedFunctionDefinition
			reduce(324), // ==, reduce: UnnamedFunctionDefinition
			reduce(324), // !=, reduce: UnnamedFunctionDefinition
			reduce(324), // <=>, reduce: UnnamedFunctionDefinition
			reduce(324), // >=, reduce: UnnamedFunctionDefinition
			reduce(324), // <, reduce: UnnamedFunctionDefinition
			reduce(324), // <=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^, reduce: UnnamedFunctionDefinition
			reduce(324), // &, reduce: UnnamedFunctionDefinition
			reduce(324), // <<, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>, reduce: UnnamedFunctionDefinition
			reduce(324), // +, reduce: UnnamedFunctionDefinition
			reduce(324), // -, reduce: UnnamedFunctionDefinition
			reduce(324), // .+, reduce: UnnamedFunctionDefinition
			reduce(324), // .-, reduce: UnnamedFunctionDefinition
			reduce(324), // *, reduce: UnnamedFunctionDefinition
			reduce(324), // /, reduce: UnnamedFunctionDefinition
			reduce(324), // //, reduce: UnnamedFunctionDefinition
			reduce(324), // %, reduce: UnnamedFunctionDefinition
			reduce(324), // .*, reduce: UnnamedFunctionDefinition
			reduce(324), // ./, reduce: UnnamedFunctionDefinition
			reduce(324), // .//, reduce: UnnamedFunctionDefinition
			reduce(324), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(324), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(324), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(324), // [[, reduce: UnnamedFunctionDefinition
			reduce(324), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3714
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S3715
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3716
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(331), // ), reduce: FuncOrSubrNonEmptyParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3717
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(289),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4182), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3718
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(289), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3719
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(235), // }, reduce: MapLiteral
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(235), // >, reduce: MapLiteral
			reduce(235), // >>, reduce: MapLiteral
			reduce(235), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(235), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(235), // ?, reduce: MapLiteral
			nil,         // :
			reduce(235), // ||, reduce: MapLiteral
			reduce(235), // ^^, reduce: MapLiteral
			reduce(235), // &&, reduce: MapLiteral
			reduce(235), // ??, reduce: MapLiteral
			reduce(235), // ???, reduce: MapLiteral
			reduce(235), // =~, reduce: MapLiteral
			reduce(235), // !=~, reduce: MapLiteral
			reduce(235), // ==, reduce: MapLiteral
			reduce(235), // !=, reduce: MapLiteral
			reduce(235), // <=>, reduce: MapLiteral
			reduce(235), // >=, reduce: MapLiteral
			reduce(235), // <, reduce: MapLiteral
			reduce(235), // <=, reduce: MapLiteral
			reduce(235), // ^, reduce: MapLiteral
			reduce(235), // &, reduce: MapLiteral
			reduce(235), // <<, reduce: MapLiteral
			reduce(235), // >>>, reduce: MapLiteral
			reduce(235), // +, reduce: MapLiteral
			reduce(235), // -, reduce: MapLiteral
			reduce(235), // .+, reduce: MapLiteral
			reduce(235), // .-, reduce: MapLiteral
			reduce(235), // *, reduce: MapLiteral
			reduce(235), // /, reduce: MapLiteral
			reduce(235), // //, reduce: MapLiteral
			reduce(235), // %, reduce: MapLiteral
			reduce(235), // .*, reduce: MapLiteral
			reduce(235), // ./, reduce: MapLiteral
			reduce(235), // .//, reduce: MapLiteral
			reduce(235), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(235), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(235), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(235), // [[, reduce: MapLiteral
			reduce(235), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3720
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(208), // }, reduce: PrecedenceChainEnd
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(208), // >, reduce: PrecedenceChainEnd
			reduce(208), // >>, reduce: PrecedenceChainEnd
			reduce(208), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(208), // ,, reduce: PrecedenceChainEnd
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(208), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(208), // ||, reduce: PrecedenceChainEnd
			reduce(208), // ^^, reduce: PrecedenceChainEnd
			reduce(208), // &&, reduce: PrecedenceChainEnd
			reduce(208), // ??, reduce: PrecedenceChainEnd
			reduce(208), // ???, reduce: PrecedenceChainEnd
			reduce(208), // =~, reduce: PrecedenceChainEnd
			reduce(208), // !=~, reduce: PrecedenceChainEnd
			reduce(208), // ==, reduce: PrecedenceChainEnd
			reduce(208), // !=, reduce: PrecedenceChainEnd
			reduce(208), // <=>, reduce: PrecedenceChainEnd
			reduce(208), // >=, reduce: PrecedenceChainEnd
			reduce(208), // <, reduce: PrecedenceChainEnd
			reduce(208), // <=, reduce: PrecedenceChainEnd
			reduce(208), // ^, reduce: PrecedenceChainEnd
			reduce(208), // &, reduce: PrecedenceChainEnd
			reduce(208), // <<, reduce: PrecedenceChainEnd
			reduce(208), // >>>, reduce: PrecedenceChainEnd
			reduce(208), // +, reduce: PrecedenceChainEnd
			reduce(208), // -, reduce: PrecedenceChainEnd
			reduce(208), // .+, reduce: PrecedenceChainEnd
			reduce(208), // .-, reduce: PrecedenceChainEnd
			reduce(208), // *, reduce: PrecedenceChainEnd
			reduce(208), // /, reduce: PrecedenceChainEnd
			reduce(208), // //, reduce: PrecedenceChainEnd
			reduce(208), // %, reduce: PrecedenceChainEnd
			reduce(208), // .*, reduce: PrecedenceChainEnd
			reduce(208), // ./, reduce: PrecedenceChainEnd
			reduce(208), // .//, reduce: PrecedenceChainEnd
			reduce(208), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(208), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3721
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(105), // }, reduce: IndirectFieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(105), // >, reduce: IndirectFieldValue
			reduce(105), // >>, reduce: IndirectFieldValue
			reduce(105), // |, reduce: IndirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(105), // ,, reduce: IndirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(105), // ?, reduce: IndirectFieldValue
			nil,         // :
			reduce(105), // ||, reduce: IndirectFieldValue
			reduce(105), // ^^, reduce: IndirectFieldValue
			reduce(105), // &&, reduce: IndirectFieldValue
			reduce(105), // ??, reduce: IndirectFieldValue
			reduce(105), // ???, reduce: IndirectFieldValue
			reduce(105), // =~, reduce: IndirectFieldValue
			reduce(105), // !=~, reduce: IndirectFieldValue
			reduce(105), // ==, reduce: IndirectFieldValue
			reduce(105), // !=, reduce: IndirectFieldValue
			reduce(105), // <=>, reduce: IndirectFieldValue
			reduce(105), // >=, reduce: IndirectFieldValue
			reduce(105), // <, reduce: IndirectFieldValue
			reduce(105), // <=, reduce: IndirectFieldValue
			reduce(105), // ^, reduce: IndirectFieldValue
			reduce(105), // &, reduce: IndirectFieldValue
			reduce(105), // <<, reduce: IndirectFieldValue
			reduce(105), // >>>, reduce: IndirectFieldValue
			reduce(105), // +, reduce: IndirectFieldValue
			reduce(105), // -, reduce: IndirectFieldValue
			reduce(105), // .+, reduce: IndirectFieldValue
			reduce(105), // .-, reduce: IndirectFieldValue
			reduce(105), // *, reduce: IndirectFieldValue
			reduce(105), // /, reduce: IndirectFieldValue
			reduce(105), // //, reduce: IndirectFieldValue
			reduce(105), // %, reduce: IndirectFieldValue
			reduce(105), // .*, reduce: IndirectFieldValue
			reduce(105), // ./, reduce: IndirectFieldValue
			reduce(105), // .//, reduce: IndirectFieldValue
			reduce(105), // ., reduce: IndirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(105), // **, reduce: IndirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(105), // [, reduce: IndirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(105), // [[, reduce: IndirectFieldValue
			reduce(105), // [[[, reduce: IndirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3722
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4184), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3723
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4185), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3724
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(114), // }, reduce: IndirectOosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(114), // >, reduce: IndirectOosvarValue
			reduce(114), // >>, reduce: IndirectOosvarValue
			reduce(114), // |, reduce: IndirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(114), // ,, reduce: IndirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(114), // ?, reduce: IndirectOosvarValue
			nil,         // :
			reduce(114), // ||, reduce: IndirectOosvarValue
			reduce(114), // ^^, reduce: IndirectOosvarValue
			reduce(114), // &&, reduce: IndirectOosvarValue
			reduce(114), // ??, reduce: IndirectOosvarValue
			reduce(114), // ???, reduce: IndirectOosvarValue
			reduce(114), // =~, reduce: IndirectOosvarValue
			reduce(114), // !=~, reduce: IndirectOosvarValue
			reduce(114), // ==, reduce: IndirectOosvarValue
			reduce(114), // !=, reduce: IndirectOosvarValue
			reduce(114), // <=>, reduce: IndirectOosvarValue
			reduce(114), // >=, reduce: IndirectOosvarValue
			reduce(114), // <, reduce: IndirectOosvarValue
			reduce(114), // <=, reduce: IndirectOosvarValue
			reduce(114), // ^, reduce: IndirectOosvarValue
			reduce(114), // &, reduce: IndirectOosvarValue
			reduce(114), // <<, reduce: IndirectOosvarValue
			reduce(114), // >>>, reduce: IndirectOosvarValue
			reduce(114), // +, reduce: IndirectOosvarValue
			reduce(114), // -, reduce: IndirectOosvarValue
			reduce(114), // .+, reduce: IndirectOosvarValue
			reduce(114), // .-, reduce: IndirectOosvarValue
			reduce(114), // *, reduce: IndirectOosvarValue
			reduce(114), // /, reduce: IndirectOosvarValue
			reduce(114), // //, reduce: IndirectOosvarValue
			reduce(114), // %, reduce: IndirectOosvarValue
			reduce(114), // .*, reduce: IndirectOosvarValue
			reduce(114), // ./, reduce: IndirectOosvarValue
			reduce(114), // .//, reduce: IndirectOosvarValue
			reduce(114), // ., reduce: IndirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(114), // **, reduce: IndirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(114), // [, reduce: IndirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(114), // [[, reduce: IndirectOosvarValue
			reduce(114), // [[[, reduce: IndirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3725
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4186), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3726
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(151), // }, reduce: LogicalOrTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(151), // ,, reduce: LogicalOrTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(151), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(151), // ||, reduce: LogicalOrTerm
			shift(2914), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3727
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(153), // }, reduce: LogicalXORTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(153), // ,, reduce: LogicalXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(153), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(153), // ||, reduce: LogicalXORTerm
			reduce(153), // ^^, reduce: LogicalXORTerm
			shift(2915), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3728
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(155), // }, reduce: LogicalAndTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(155), // ,, reduce: LogicalAndTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(155), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(155), // ||, reduce: LogicalAndTerm
			reduce(155), // ^^, reduce: LogicalAndTerm
			reduce(155), // &&, reduce: LogicalAndTerm
			shift(2916), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3729
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(157), // }, reduce: AbsentCoalesceTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(157), // ,, reduce: AbsentCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(157), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(157), // ||, reduce: AbsentCoalesceTerm
			reduce(157), // ^^, reduce: AbsentCoalesceTerm
			reduce(157), // &&, reduce: AbsentCoalesceTerm
			reduce(157), // ??, reduce: AbsentCoalesceTerm
			shift(2917), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3730
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(159), // }, reduce: EmptyCoalesceTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(159), // ,, reduce: EmptyCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(159), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(159), // ||, reduce: EmptyCoalesceTerm
			reduce(159), // ^^, reduce: EmptyCoalesceTerm
			reduce(159), // &&, reduce: EmptyCoalesceTerm
			reduce(159), // ??, reduce: EmptyCoalesceTerm
			reduce(159), // ???, reduce: EmptyCoalesceTerm
			shift(2918), // =~
			shift(2919), // !=~
			shift(2920), // ==
			shift(2921), // !=
			shift(2922), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3731
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(161), // }, reduce: EqneTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2923), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(161), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(161), // ?, reduce: EqneTerm
			nil,         // :
			reduce(161), // ||, reduce: EqneTerm
			reduce(161), // ^^, reduce: EqneTerm
			reduce(161), // &&, reduce: EqneTerm
			reduce(161), // ??, reduce: EqneTerm
			reduce(161), // ???, reduce: EqneTerm
			reduce(161), // =~, reduce: EqneTerm
			reduce(161), // !=~, reduce: EqneTerm
			reduce(161), // ==, reduce: EqneTerm
			reduce(161), // !=, reduce: EqneTerm
			reduce(161), // <=>, reduce: EqneTerm
			shift(2924), // >=
			shift(2925), // <
			shift(2926), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3732
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(162), // }, reduce: EqneTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2923), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(162), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(162), // ?, reduce: EqneTerm
			nil,         // :
			reduce(162), // ||, reduce: EqneTerm
			reduce(162), // ^^, reduce: EqneTerm
			reduce(162), // &&, reduce: EqneTerm
			reduce(162), // ??, reduce: EqneTerm
			reduce(162), // ???, reduce: EqneTerm
			reduce(162), // =~, reduce: EqneTerm
			reduce(162), // !=~, reduce: EqneTerm
			reduce(162), // ==, reduce: EqneTerm
			reduce(162), // !=, reduce: EqneTerm
			reduce(162), // <=>, reduce: EqneTerm
			shift(2924), // >=
			shift(2925), // <
			shift(2926), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3733
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(163), // }, reduce: EqneTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2923), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(163), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(163), // ?, reduce: EqneTerm
			nil,         // :
			reduce(163), // ||, reduce: EqneTerm
			reduce(163), // ^^, reduce: EqneTerm
			reduce(163), // &&, reduce: EqneTerm
			reduce(163), // ??, reduce: EqneTerm
			reduce(163), // ???, reduce: EqneTerm
			reduce(163), // =~, reduce: EqneTerm
			reduce(163), // !=~, reduce: EqneTerm
			reduce(163), // ==, reduce: EqneTerm
			reduce(163), // !=, reduce: EqneTerm
			reduce(163), // <=>, reduce: EqneTerm
			shift(2924), // >=
			shift(2925), // <
			shift(2926), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3734
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(164), // }, reduce: EqneTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2923), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(164), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(164), // ?, reduce: EqneTerm
			nil,         // :
			reduce(164), // ||, reduce: EqneTerm
			reduce(164), // ^^, reduce: EqneTerm
			reduce(164), // &&, reduce: EqneTerm
			reduce(164), // ??, reduce: EqneTerm
			reduce(164), // ???, reduce: EqneTerm
			reduce(164), // =~, reduce: EqneTerm
			reduce(164), // !=~, reduce: EqneTerm
			reduce(164), // ==, reduce: EqneTerm
			reduce(164), // !=, reduce: EqneTerm
			reduce(164), // <=>, reduce: EqneTerm
			shift(2924), // >=
			shift(2925), // <
			shift(2926), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3735
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(165), // }, reduce: EqneTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(2923), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(165), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(165), // ?, reduce: EqneTerm
			nil,         // :
			reduce(165), // ||, reduce: EqneTerm
			reduce(165), // ^^, reduce: EqneTerm
			reduce(165), // &&, reduce: EqneTerm
			reduce(165), // ??, reduce: EqneTerm
			reduce(165), // ???, reduce: EqneTerm
			reduce(165), // =~, reduce: EqneTerm
			reduce(165), // !=~, reduce: EqneTerm
			reduce(165), // ==, reduce: EqneTerm
			reduce(165), // !=, reduce: EqneTerm
			reduce(165), // <=>, reduce: EqneTerm
			shift(2924), // >=
			shift(2925), // <
			shift(2926), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3736
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(167), // }, reduce: CmpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(167), // >, reduce: CmpTerm
			nil,         // >>
			shift(2927), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(167), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(167), // ?, reduce: CmpTerm
			nil,         // :
			reduce(167), // ||, reduce: CmpTerm
			reduce(167), // ^^, reduce: CmpTerm
			reduce(167), // &&, reduce: CmpTerm
			reduce(167), // ??, reduce: CmpTerm
			reduce(167), // ???, reduce: CmpTerm
			reduce(167), // =~, reduce: CmpTerm
			reduce(167), // !=~, reduce: CmpTerm
			reduce(167), // ==, reduce: CmpTerm
			reduce(167), // !=, reduce: CmpTerm
			reduce(167), // <=>, reduce: CmpTerm
			reduce(167), // >=, reduce: CmpTerm
			reduce(167), // <, reduce: CmpTerm
			reduce(167), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3737
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(168), // }, reduce: CmpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(168), // >, reduce: CmpTerm
			nil,         // >>
			shift(2927), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(168), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(168), // ?, reduce: CmpTerm
			nil,         // :
			reduce(168), // ||, reduce: CmpTerm
			reduce(168), // ^^, reduce: CmpTerm
			reduce(168), // &&, reduce: CmpTerm
			reduce(168), // ??, reduce: CmpTerm
			reduce(168), // ???, reduce: CmpTerm
			reduce(168), // =~, reduce: CmpTerm
			reduce(168), // !=~, reduce: CmpTerm
			reduce(168), // ==, reduce: CmpTerm
			reduce(168), // !=, reduce: CmpTerm
			reduce(168), // <=>, reduce: CmpTerm
			reduce(168), // >=, reduce: CmpTerm
			reduce(168), // <, reduce: CmpTerm
			reduce(168), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3738
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(169), // }, reduce: CmpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(169), // >, reduce: CmpTerm
			nil,         // >>
			shift(2927), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(169), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(169), // ?, reduce: CmpTerm
			nil,         // :
			reduce(169), // ||, reduce: CmpTerm
			reduce(169), // ^^, reduce: CmpTerm
			reduce(169), // &&, reduce: CmpTerm
			reduce(169), // ??, reduce: CmpTerm
			reduce(169), // ???, reduce: CmpTerm
			reduce(169), // =~, reduce: CmpTerm
			reduce(169), // !=~, reduce: CmpTerm
			reduce(169), // ==, reduce: CmpTerm
			reduce(169), // !=, reduce: CmpTerm
			reduce(169), // <=>, reduce: CmpTerm
			reduce(169), // >=, reduce: CmpTerm
			reduce(169), // <, reduce: CmpTerm
			reduce(169), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3739
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(170), // }, reduce: CmpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(170), // >, reduce: CmpTerm
			nil,         // >>
			shift(2927), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(170), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(170), // ?, reduce: CmpTerm
			nil,         // :
			reduce(170), // ||, reduce: CmpTerm
			reduce(170), // ^^, reduce: CmpTerm
			reduce(170), // &&, reduce: CmpTerm
			reduce(170), // ??, reduce: CmpTerm
			reduce(170), // ???, reduce: CmpTerm
			reduce(170), // =~, reduce: CmpTerm
			reduce(170), // !=~, reduce: CmpTerm
			reduce(170), // ==, reduce: CmpTerm
			reduce(170), // !=, reduce: CmpTerm
			reduce(170), // <=>, reduce: CmpTerm
			reduce(170), // >=, reduce: CmpTerm
			reduce(170), // <, reduce: CmpTerm
			reduce(170), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3740
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(172), // }, reduce: BitwiseORTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(172), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(172), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(172), // ,, reduce: BitwiseORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(172), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(172), // ||, reduce: BitwiseORTerm
			reduce(172), // ^^, reduce: BitwiseORTerm
			reduce(172), // &&, reduce: BitwiseORTerm
			reduce(172), // ??, reduce: BitwiseORTerm
			reduce(172), // ???, reduce: BitwiseORTerm
			reduce(172), // =~, reduce: BitwiseORTerm
			reduce(172), // !=~, reduce: BitwiseORTerm
			reduce(172), // ==, reduce: BitwiseORTerm
			reduce(172), // !=, reduce: BitwiseORTerm
			reduce(172), // <=>, reduce: BitwiseORTerm
			reduce(172), // >=, reduce: BitwiseORTerm
			reduce(172), // <, reduce: BitwiseORTerm
			reduce(172), // <=, reduce: BitwiseORTerm
			shift(2928), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3741
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(174), // }, reduce: BitwiseXORTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(174), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(174), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(174), // ,, reduce: BitwiseXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(174), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(174), // ||, reduce: BitwiseXORTerm
			reduce(174), // ^^, reduce: BitwiseXORTerm
			reduce(174), // &&, reduce: BitwiseXORTerm
			reduce(174), // ??, reduce: BitwiseXORTerm
			reduce(174), // ???, reduce: BitwiseXORTerm
			reduce(174), // =~, reduce: BitwiseXORTerm
			reduce(174), // !=~, reduce: BitwiseXORTerm
			reduce(174), // ==, reduce: BitwiseXORTerm
			reduce(174), // !=, reduce: BitwiseXORTerm
			reduce(174), // <=>, reduce: BitwiseXORTerm
			reduce(174), // >=, reduce: BitwiseXORTerm
			reduce(174), // <, reduce: BitwiseXORTerm
			reduce(174), // <=, reduce: BitwiseXORTerm
			reduce(174), // ^, reduce: BitwiseXORTerm
			shift(2929), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3742
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(176), // }, reduce: BitwiseANDTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(176), // >, reduce: BitwiseANDTerm
			shift(2930), // >>
			reduce(176), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(176), // ,, reduce: BitwiseANDTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(176), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(176), // ||, reduce: BitwiseANDTerm
			reduce(176), // ^^, reduce: BitwiseANDTerm
			reduce(176), // &&, reduce: BitwiseANDTerm
			reduce(176), // ??, reduce: BitwiseANDTerm
			reduce(176), // ???, reduce: BitwiseANDTerm
			reduce(176), // =~, reduce: BitwiseANDTerm
			reduce(176), // !=~, reduce: BitwiseANDTerm
			reduce(176), // ==, reduce: BitwiseANDTerm
			reduce(176), // !=, reduce: BitwiseANDTerm
			reduce(176), // <=>, reduce: BitwiseANDTerm
			reduce(176), // >=, reduce: BitwiseANDTerm
			reduce(176), // <, reduce: BitwiseANDTerm
			reduce(176), // <=, reduce: BitwiseANDTerm
			reduce(176), // ^, reduce: BitwiseANDTerm
			reduce(176), // &, reduce: BitwiseANDTerm
			shift(2931), // <<
			shift(2932), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3743
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(179), // }, reduce: BitwiseShiftTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(179), // >, reduce: BitwiseShiftTerm
			reduce(179), // >>, reduce: BitwiseShiftTerm
			reduce(179), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(179), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(179), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(179), // ||, reduce: BitwiseShiftTerm
			reduce(179), // ^^, reduce: BitwiseShiftTerm
			reduce(179), // &&, reduce: BitwiseShiftTerm
			reduce(179), // ??, reduce: BitwiseShiftTerm
			reduce(179), // ???, reduce: BitwiseShiftTerm
			reduce(179), // =~, reduce: BitwiseShiftTerm
			reduce(179), // !=~, reduce: BitwiseShiftTerm
			reduce(179), // ==, reduce: BitwiseShiftTerm
			reduce(179), // !=, reduce: BitwiseShiftTerm
			reduce(179), // <=>, reduce: BitwiseShiftTerm
			reduce(179), // >=, reduce: BitwiseShiftTerm
			reduce(179), // <, reduce: BitwiseShiftTerm
			reduce(179), // <=, reduce: BitwiseShiftTerm
			reduce(179), // ^, reduce: BitwiseShiftTerm
			reduce(179), // &, reduce: BitwiseShiftTerm
			reduce(179), // <<, reduce: BitwiseShiftTerm
			reduce(179), // >>>, reduce: BitwiseShiftTerm
			shift(2933), // +
			shift(2934), // -
			shift(2935), // .+
			shift(2936), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3744
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(178), // }, reduce: BitwiseShiftTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(178), // >, reduce: BitwiseShiftTerm
			reduce(178), // >>, reduce: BitwiseShiftTerm
			reduce(178), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(178), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(178), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(178), // ||, reduce: BitwiseShiftTerm
			reduce(178), // ^^, reduce: BitwiseShiftTerm
			reduce(178), // &&, reduce: BitwiseShiftTerm
			reduce(178), // ??, reduce: BitwiseShiftTerm
			reduce(178), // ???, reduce: BitwiseShiftTerm
			reduce(178), // =~, reduce: BitwiseShiftTerm
			reduce(178), // !=~, reduce: BitwiseShiftTerm
			reduce(178), // ==, reduce: BitwiseShiftTerm
			reduce(178), // !=, reduce: BitwiseShiftTerm
			reduce(178), // <=>, reduce: BitwiseShiftTerm
			reduce(178), // >=, reduce: BitwiseShiftTerm
			reduce(178), // <, reduce: BitwiseShiftTerm
			reduce(178), // <=, reduce: BitwiseShiftTerm
			reduce(178), // ^, reduce: BitwiseShiftTerm
			reduce(178), // &, reduce: BitwiseShiftTerm
			reduce(178), // <<, reduce: BitwiseShiftTerm
			reduce(178), // >>>, reduce: BitwiseShiftTerm
			shift(2933), // +
			shift(2934), // -
			shift(2935), // .+
			shift(2936), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3745
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(180), // }, reduce: BitwiseShiftTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(180), // >, reduce: BitwiseShiftTerm
			reduce(180), // >>, reduce: BitwiseShiftTerm
			reduce(180), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(180), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(180), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(180), // ||, reduce: BitwiseShiftTerm
			reduce(180), // ^^, reduce: BitwiseShiftTerm
			reduce(180), // &&, reduce: BitwiseShiftTerm
			reduce(180), // ??, reduce: BitwiseShiftTerm
			reduce(180), // ???, reduce: BitwiseShiftTerm
			reduce(180), // =~, reduce: BitwiseShiftTerm
			reduce(180), // !=~, reduce: BitwiseShiftTerm
			reduce(180), // ==, reduce: BitwiseShiftTerm
			reduce(180), // !=, reduce: BitwiseShiftTerm
			reduce(180), // <=>, reduce: BitwiseShiftTerm
			reduce(180), // >=, reduce: BitwiseShiftTerm
			reduce(180), // <, reduce: BitwiseShiftTerm
			reduce(180), // <=, reduce: BitwiseShiftTerm
			reduce(180), // ^, reduce: BitwiseShiftTerm
			reduce(180), // &, reduce: BitwiseShiftTerm
			reduce(180), // <<, reduce: BitwiseShiftTerm
			reduce(180), // >>>, reduce: BitwiseShiftTerm
			shift(2933), // +
			shift(2934), // -
			shift(2935), // .+
			shift(2936), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3746
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(182), // }, reduce: AddsubdotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(182), // >, reduce: AddsubdotTerm
			reduce(182), // >>, reduce: AddsubdotTerm
			reduce(182), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(182), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(182), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(182), // ||, reduce: AddsubdotTerm
			reduce(182), // ^^, reduce: AddsubdotTerm
			reduce(182), // &&, reduce: AddsubdotTerm
			reduce(182), // ??, reduce: AddsubdotTerm
			reduce(182), // ???, reduce: AddsubdotTerm
			reduce(182), // =~, reduce: AddsubdotTerm
			reduce(182), // !=~, reduce: AddsubdotTerm
			reduce(182), // ==, reduce: AddsubdotTerm
			reduce(182), // !=, reduce: AddsubdotTerm
			reduce(182), // <=>, reduce: AddsubdotTerm
			reduce(182), // >=, reduce: AddsubdotTerm
			reduce(182), // <, reduce: AddsubdotTerm
			reduce(182), // <=, reduce: AddsubdotTerm
			reduce(182), // ^, reduce: AddsubdotTerm
			reduce(182), // &, reduce: AddsubdotTerm
			reduce(182), // <<, reduce: AddsubdotTerm
			reduce(182), // >>>, reduce: AddsubdotTerm
			reduce(182), // +, reduce: AddsubdotTerm
			reduce(182), // -, reduce: AddsubdotTerm
			reduce(182), // .+, reduce: AddsubdotTerm
			reduce(182), // .-, reduce: AddsubdotTerm
			shift(2938), // *
			shift(2939), // /
			shift(2940), // //
			shift(2941), // %
			shift(2942), // .*
			shift(2943), // ./
			shift(2944), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3747
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(183), // }, reduce: AddsubdotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(183), // >, reduce: AddsubdotTerm
			reduce(183), // >>, reduce: AddsubdotTerm
			reduce(183), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(183), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(183), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(183), // ||, reduce: AddsubdotTerm
			reduce(183), // ^^, reduce: AddsubdotTerm
			reduce(183), // &&, reduce: AddsubdotTerm
			reduce(183), // ??, reduce: AddsubdotTerm
			reduce(183), // ???, reduce: AddsubdotTerm
			reduce(183), // =~, reduce: AddsubdotTerm
			reduce(183), // !=~, reduce: AddsubdotTerm
			reduce(183), // ==, reduce: AddsubdotTerm
			reduce(183), // !=, reduce: AddsubdotTerm
			reduce(183), // <=>, reduce: AddsubdotTerm
			reduce(183), // >=, reduce: AddsubdotTerm
			reduce(183), // <, reduce: AddsubdotTerm
			reduce(183), // <=, reduce: AddsubdotTerm
			reduce(183), // ^, reduce: AddsubdotTerm
			reduce(183), // &, reduce: AddsubdotTerm
			reduce(183), // <<, reduce: AddsubdotTerm
			reduce(183), // >>>, reduce: AddsubdotTerm
			reduce(183), // +, reduce: AddsubdotTerm
			reduce(183), // -, reduce: AddsubdotTerm
			reduce(183), // .+, reduce: AddsubdotTerm
			reduce(183), // .-, reduce: AddsubdotTerm
			shift(2938), // *
			shift(2939), // /
			shift(2940), // //
			shift(2941), // %
			shift(2942), // .*
			shift(2943), // ./
			shift(2944), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3748
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(184), // }, reduce: AddsubdotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(184), // >, reduce: AddsubdotTerm
			reduce(184), // >>, reduce: AddsubdotTerm
			reduce(184), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(184), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(184), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(184), // ||, reduce: AddsubdotTerm
			reduce(184), // ^^, reduce: AddsubdotTerm
			reduce(184), // &&, reduce: AddsubdotTerm
			reduce(184), // ??, reduce: AddsubdotTerm
			reduce(184), // ???, reduce: AddsubdotTerm
			reduce(184), // =~, reduce: AddsubdotTerm
			reduce(184), // !=~, reduce: AddsubdotTerm
			reduce(184), // ==, reduce: AddsubdotTerm
			reduce(184), // !=, reduce: AddsubdotTerm
			reduce(184), // <=>, reduce: AddsubdotTerm
			reduce(184), // >=, reduce: AddsubdotTerm
			reduce(184), // <, reduce: AddsubdotTerm
			reduce(184), // <=, reduce: AddsubdotTerm
			reduce(184), // ^, reduce: AddsubdotTerm
			reduce(184), // &, reduce: AddsubdotTerm
			reduce(184), // <<, reduce: AddsubdotTerm
			reduce(184), // >>>, reduce: AddsubdotTerm
			reduce(184), // +, reduce: AddsubdotTerm
			reduce(184), // -, reduce: AddsubdotTerm
			reduce(184), // .+, reduce: AddsubdotTerm
			reduce(184), // .-, reduce: AddsubdotTerm
			shift(2938), // *
			shift(2939), // /
			shift(2940), // //
			shift(2941), // %
			shift(2942), // .*
			shift(2943), // ./
			shift(2944), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3749
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(185), // }, reduce: AddsubdotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(185), // >, reduce: AddsubdotTerm
			reduce(185), // >>, reduce: AddsubdotTerm
			reduce(185), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(185), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(185), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(185), // ||, reduce: AddsubdotTerm
			reduce(185), // ^^, reduce: AddsubdotTerm
			reduce(185), // &&, reduce: AddsubdotTerm
			reduce(185), // ??, reduce: AddsubdotTerm
			reduce(185), // ???, reduce: AddsubdotTerm
			reduce(185), // =~, reduce: AddsubdotTerm
			reduce(185), // !=~, reduce: AddsubdotTerm
			reduce(185), // ==, reduce: AddsubdotTerm
			reduce(185), // !=, reduce: AddsubdotTerm
			reduce(185), // <=>, reduce: AddsubdotTerm
			reduce(185), // >=, reduce: AddsubdotTerm
			reduce(185), // <, reduce: AddsubdotTerm
			reduce(185), // <=, reduce: AddsubdotTerm
			reduce(185), // ^, reduce: AddsubdotTerm
			reduce(185), // &, reduce: AddsubdotTerm
			reduce(185), // <<, reduce: AddsubdotTerm
			reduce(185), // >>>, reduce: AddsubdotTerm
			reduce(185), // +, reduce: AddsubdotTerm
			reduce(185), // -, reduce: AddsubdotTerm
			reduce(185), // .+, reduce: AddsubdotTerm
			reduce(185), // .-, reduce: AddsubdotTerm
			shift(2938), // *
			shift(2939), // /
			shift(2940), // //
			shift(2941), // %
			shift(2942), // .*
			shift(2943), // ./
			shift(2944), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3750
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(187), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(187), // >, reduce: MuldivTerm
			reduce(187), // >>, reduce: MuldivTerm
			reduce(187), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(187), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(187), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(187), // ||, reduce: MuldivTerm
			reduce(187), // ^^, reduce: MuldivTerm
			reduce(187), // &&, reduce: MuldivTerm
			reduce(187), // ??, reduce: MuldivTerm
			reduce(187), // ???, reduce: MuldivTerm
			reduce(187), // =~, reduce: MuldivTerm
			reduce(187), // !=~, reduce: MuldivTerm
			reduce(187), // ==, reduce: MuldivTerm
			reduce(187), // !=, reduce: MuldivTerm
			reduce(187), // <=>, reduce: MuldivTerm
			reduce(187), // >=, reduce: MuldivTerm
			reduce(187), // <, reduce: MuldivTerm
			reduce(187), // <=, reduce: MuldivTerm
			reduce(187), // ^, reduce: MuldivTerm
			reduce(187), // &, reduce: MuldivTerm
			reduce(187), // <<, reduce: MuldivTerm
			reduce(187), // >>>, reduce: MuldivTerm
			reduce(187), // +, reduce: MuldivTerm
			reduce(187), // -, reduce: MuldivTerm
			reduce(187), // .+, reduce: MuldivTerm
			reduce(187), // .-, reduce: MuldivTerm
			reduce(187), // *, reduce: MuldivTerm
			reduce(187), // /, reduce: MuldivTerm
			reduce(187), // //, reduce: MuldivTerm
			reduce(187), // %, reduce: MuldivTerm
			reduce(187), // .*, reduce: MuldivTerm
			reduce(187), // ./, reduce: MuldivTerm
			reduce(187), // .//, reduce: MuldivTerm
			shift(2948), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3751
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(188), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(188), // >, reduce: MuldivTerm
			reduce(188), // >>, reduce: MuldivTerm
			reduce(188), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(188), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(188), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(188), // ||, reduce: MuldivTerm
			reduce(188), // ^^, reduce: MuldivTerm
			reduce(188), // &&, reduce: MuldivTerm
			reduce(188), // ??, reduce: MuldivTerm
			reduce(188), // ???, reduce: MuldivTerm
			reduce(188), // =~, reduce: MuldivTerm
			reduce(188), // !=~, reduce: MuldivTerm
			reduce(188), // ==, reduce: MuldivTerm
			reduce(188), // !=, reduce: MuldivTerm
			reduce(188), // <=>, reduce: MuldivTerm
			reduce(188), // >=, reduce: MuldivTerm
			reduce(188), // <, reduce: MuldivTerm
			reduce(188), // <=, reduce: MuldivTerm
			reduce(188), // ^, reduce: MuldivTerm
			reduce(188), // &, reduce: MuldivTerm
			reduce(188), // <<, reduce: MuldivTerm
			reduce(188), // >>>, reduce: MuldivTerm
			reduce(188), // +, reduce: MuldivTerm
			reduce(188), // -, reduce: MuldivTerm
			reduce(188), // .+, reduce: MuldivTerm
			reduce(188), // .-, reduce: MuldivTerm
			reduce(188), // *, reduce: MuldivTerm
			reduce(188), // /, reduce: MuldivTerm
			reduce(188), // //, reduce: MuldivTerm
			reduce(188), // %, reduce: MuldivTerm
			reduce(188), // .*, reduce: MuldivTerm
			reduce(188), // ./, reduce: MuldivTerm
			reduce(188), // .//, reduce: MuldivTerm
			shift(2948), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3752
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(189), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(189), // >, reduce: MuldivTerm
			reduce(189), // >>, reduce: MuldivTerm
			reduce(189), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(189), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(189), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(189), // ||, reduce: MuldivTerm
			reduce(189), // ^^, reduce: MuldivTerm
			reduce(189), // &&, reduce: MuldivTerm
			reduce(189), // ??, reduce: MuldivTerm
			reduce(189), // ???, reduce: MuldivTerm
			reduce(189), // =~, reduce: MuldivTerm
			reduce(189), // !=~, reduce: MuldivTerm
			reduce(189), // ==, reduce: MuldivTerm
			reduce(189), // !=, reduce: MuldivTerm
			reduce(189), // <=>, reduce: MuldivTerm
			reduce(189), // >=, reduce: MuldivTerm
			reduce(189), // <, reduce: MuldivTerm
			reduce(189), // <=, reduce: MuldivTerm
			reduce(189), // ^, reduce: MuldivTerm
			reduce(189), // &, reduce: MuldivTerm
			reduce(189), // <<, reduce: MuldivTerm
			reduce(189), // >>>, reduce: MuldivTerm
			reduce(189), // +, reduce: MuldivTerm
			reduce(189), // -, reduce: MuldivTerm
			reduce(189), // .+, reduce: MuldivTerm
			reduce(189), // .-, reduce: MuldivTerm
			reduce(189), // *, reduce: MuldivTerm
			reduce(189), // /, reduce: MuldivTerm
			reduce(189), // //, reduce: MuldivTerm
			reduce(189), // %, reduce: MuldivTerm
			reduce(189), // .*, reduce: MuldivTerm
			reduce(189), // ./, reduce: MuldivTerm
			reduce(189), // .//, reduce: MuldivTerm
			shift(2948), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3753
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(190), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(190), // >, reduce: MuldivTerm
			reduce(190), // >>, reduce: MuldivTerm
			reduce(190), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(190), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(190), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(190), // ||, reduce: MuldivTerm
			reduce(190), // ^^, reduce: MuldivTerm
			reduce(190), // &&, reduce: MuldivTerm
			reduce(190), // ??, reduce: MuldivTerm
			reduce(190), // ???, reduce: MuldivTerm
			reduce(190), // =~, reduce: MuldivTerm
			reduce(190), // !=~, reduce: MuldivTerm
			reduce(190), // ==, reduce: MuldivTerm
			reduce(190), // !=, reduce: MuldivTerm
			reduce(190), // <=>, reduce: MuldivTerm
			reduce(190), // >=, reduce: MuldivTerm
			reduce(190), // <, reduce: MuldivTerm
			reduce(190), // <=, reduce: MuldivTerm
			reduce(190), // ^, reduce: MuldivTerm
			reduce(190), // &, reduce: MuldivTerm
			reduce(190), // <<, reduce: MuldivTerm
			reduce(190), // >>>, reduce: MuldivTerm
			reduce(190), // +, reduce: MuldivTerm
			reduce(190), // -, reduce: MuldivTerm
			reduce(190), // .+, reduce: MuldivTerm
			reduce(190), // .-, reduce: MuldivTerm
			reduce(190), // *, reduce: MuldivTerm
			reduce(190), // /, reduce: MuldivTerm
			reduce(190), // //, reduce: MuldivTerm
			reduce(190), // %, reduce: MuldivTerm
			reduce(190), // .*, reduce: MuldivTerm
			reduce(190), // ./, reduce: MuldivTerm
			reduce(190), // .//, reduce: MuldivTerm
			shift(2948), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3754
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(191), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(191), // >, reduce: MuldivTerm
			reduce(191), // >>, reduce: MuldivTerm
			reduce(191), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(191), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(191), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(191), // ||, reduce: MuldivTerm
			reduce(191), // ^^, reduce: MuldivTerm
			reduce(191), // &&, reduce: MuldivTerm
			reduce(191), // ??, reduce: MuldivTerm
			reduce(191), // ???, reduce: MuldivTerm
			reduce(191), // =~, reduce: MuldivTerm
			reduce(191), // !=~, reduce: MuldivTerm
			reduce(191), // ==, reduce: MuldivTerm
			reduce(191), // !=, reduce: MuldivTerm
			reduce(191), // <=>, reduce: MuldivTerm
			reduce(191), // >=, reduce: MuldivTerm
			reduce(191), // <, reduce: MuldivTerm
			reduce(191), // <=, reduce: MuldivTerm
			reduce(191), // ^, reduce: MuldivTerm
			reduce(191), // &, reduce: MuldivTerm
			reduce(191), // <<, reduce: MuldivTerm
			reduce(191), // >>>, reduce: MuldivTerm
			reduce(191), // +, reduce: MuldivTerm
			reduce(191), // -, reduce: MuldivTerm
			reduce(191), // .+, reduce: MuldivTerm
			reduce(191), // .-, reduce: MuldivTerm
			reduce(191), // *, reduce: MuldivTerm
			reduce(191), // /, reduce: MuldivTerm
			reduce(191), // //, reduce: MuldivTerm
			reduce(191), // %, reduce: MuldivTerm
			reduce(191), // .*, reduce: MuldivTerm
			reduce(191), // ./, reduce: MuldivTerm
			reduce(191), // .//, reduce: MuldivTerm
			shift(2948), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3755
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(192), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(192), // >, reduce: MuldivTerm
			reduce(192), // >>, reduce: MuldivTerm
			reduce(192), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(192), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(192), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(192), // ||, reduce: MuldivTerm
			reduce(192), // ^^, reduce: MuldivTerm
			reduce(192), // &&, reduce: MuldivTerm
			reduce(192), // ??, reduce: MuldivTerm
			reduce(192), // ???, reduce: MuldivTerm
			reduce(192), // =~, reduce: MuldivTerm
			reduce(192), // !=~, reduce: MuldivTerm
			reduce(192), // ==, reduce: MuldivTerm
			reduce(192), // !=, reduce: MuldivTerm
			reduce(192), // <=>, reduce: MuldivTerm
			reduce(192), // >=, reduce: MuldivTerm
			reduce(192), // <, reduce: MuldivTerm
			reduce(192), // <=, reduce: MuldivTerm
			reduce(192), // ^, reduce: MuldivTerm
			reduce(192), // &, reduce: MuldivTerm
			reduce(192), // <<, reduce: MuldivTerm
			reduce(192), // >>>, reduce: MuldivTerm
			reduce(192), // +, reduce: MuldivTerm
			reduce(192), // -, reduce: MuldivTerm
			reduce(192), // .+, reduce: MuldivTerm
			reduce(192), // .-, reduce: MuldivTerm
			reduce(192), // *, reduce: MuldivTerm
			reduce(192), // /, reduce: MuldivTerm
			reduce(192), // //, reduce: MuldivTerm
			reduce(192), // %, reduce: MuldivTerm
			reduce(192), // .*, reduce: MuldivTerm
			reduce(192), // ./, reduce: MuldivTerm
			reduce(192), // .//, reduce: MuldivTerm
			shift(2948), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3756
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(193), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(193), // >, reduce: MuldivTerm
			reduce(193), // >>, reduce: MuldivTerm
			reduce(193), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(193), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(193), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(193), // ||, reduce: MuldivTerm
			reduce(193), // ^^, reduce: MuldivTerm
			reduce(193), // &&, reduce: MuldivTerm
			reduce(193), // ??, reduce: MuldivTerm
			reduce(193), // ???, reduce: MuldivTerm
			reduce(193), // =~, reduce: MuldivTerm
			reduce(193), // !=~, reduce: MuldivTerm
			reduce(193), // ==, reduce: MuldivTerm
			reduce(193), // !=, reduce: MuldivTerm
			reduce(193), // <=>, reduce: MuldivTerm
			reduce(193), // >=, reduce: MuldivTerm
			reduce(193), // <, reduce: MuldivTerm
			reduce(193), // <=, reduce: MuldivTerm
			reduce(193), // ^, reduce: MuldivTerm
			reduce(193), // &, reduce: MuldivTerm
			reduce(193), // <<, reduce: MuldivTerm
			reduce(193), // >>>, reduce: MuldivTerm
			reduce(193), // +, reduce: MuldivTerm
			reduce(193), // -, reduce: MuldivTerm
			reduce(193), // .+, reduce: MuldivTerm
			reduce(193), // .-, reduce: MuldivTerm
			reduce(193), // *, reduce: MuldivTerm
			reduce(193), // /, reduce: MuldivTerm
			reduce(193), // //, reduce: MuldivTerm
			reduce(193), // %, reduce: MuldivTerm
			reduce(193), // .*, reduce: MuldivTerm
			reduce(193), // ./, reduce: MuldivTerm
			reduce(193), // .//, reduce: MuldivTerm
			shift(2948), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3757
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(195), // }, reduce: DotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(195), // >, reduce: DotTerm
			reduce(195), // >>, reduce: DotTerm
			reduce(195), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(195), // ,, reduce: DotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(195), // ?, reduce: DotTerm
			nil,         // :
			reduce(195), // ||, reduce: DotTerm
			reduce(195), // ^^, reduce: DotTerm
			reduce(195), // &&, reduce: DotTerm
			reduce(195), // ??, reduce: DotTerm
			reduce(195), // ???, reduce: DotTerm
			reduce(195), // =~, reduce: DotTerm
			reduce(195), // !=~, reduce: DotTerm
			reduce(195), // ==, reduce: DotTerm
			reduce(195), // !=, reduce: DotTerm
			reduce(195), // <=>, reduce: DotTerm
			reduce(195), // >=, reduce: DotTerm
			reduce(195), // <, reduce: DotTerm
			reduce(195), // <=, reduce: DotTerm
			reduce(195), // ^, reduce: DotTerm
			reduce(195), // &, reduce: DotTerm
			reduce(195), // <<, reduce: DotTerm
			reduce(195), // >>>, reduce: DotTerm
			reduce(195), // +, reduce: DotTerm
			reduce(195), // -, reduce: DotTerm
			reduce(195), // .+, reduce: DotTerm
			reduce(195), // .-, reduce: DotTerm
			reduce(195), // *, reduce: DotTerm
			reduce(195), // /, reduce: DotTerm
			reduce(195), // //, reduce: DotTerm
			reduce(195), // %, reduce: DotTerm
			reduce(195), // .*, reduce: DotTerm
			reduce(195), // ./, reduce: DotTerm
			reduce(195), // .//, reduce: DotTerm
			reduce(195), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3758
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3759
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3760
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(204), // }, reduce: PowTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(204), // >, reduce: PowTerm
			reduce(204), // >>, reduce: PowTerm
			reduce(204), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(204), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(204), // ?, reduce: PowTerm
			nil,         // :
			reduce(204), // ||, reduce: PowTerm
			reduce(204), // ^^, reduce: PowTerm
			reduce(204), // &&, reduce: PowTerm
			reduce(204), // ??, reduce: PowTerm
			reduce(204), // ???, reduce: PowTerm
			reduce(204), // =~, reduce: PowTerm
			reduce(204), // !=~, reduce: PowTerm
			reduce(204), // ==, reduce: PowTerm
			reduce(204), // !=, reduce: PowTerm
			reduce(204), // <=>, reduce: PowTerm
			reduce(204), // >=, reduce: PowTerm
			reduce(204), // <, reduce: PowTerm
			reduce(204), // <=, reduce: PowTerm
			reduce(204), // ^, reduce: PowTerm
			reduce(204), // &, reduce: PowTerm
			reduce(204), // <<, reduce: PowTerm
			reduce(204), // >>>, reduce: PowTerm
			reduce(204), // +, reduce: PowTerm
			reduce(204), // -, reduce: PowTerm
			reduce(204), // .+, reduce: PowTerm
			reduce(204), // .-, reduce: PowTerm
			reduce(204), // *, reduce: PowTerm
			reduce(204), // /, reduce: PowTerm
			reduce(204), // //, reduce: PowTerm
			reduce(204), // %, reduce: PowTerm
			reduce(204), // .*, reduce: PowTerm
			reduce(204), // ./, reduce: PowTerm
			reduce(204), // .//, reduce: PowTerm
			reduce(204), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3761
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4189), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4190), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3762
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4192), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3763
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4193), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3764
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4194), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3765
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(229), // }, reduce: ArrayLiteral
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(229), // >, reduce: ArrayLiteral
			reduce(229), // >>, reduce: ArrayLiteral
			reduce(229), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(229), // ,, reduce: ArrayLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(229), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(229), // ||, reduce: ArrayLiteral
			reduce(229), // ^^, reduce: ArrayLiteral
			reduce(229), // &&, reduce: ArrayLiteral
			reduce(229), // ??, reduce: ArrayLiteral
			reduce(229), // ???, reduce: ArrayLiteral
			reduce(229), // =~, reduce: ArrayLiteral
			reduce(229), // !=~, reduce: ArrayLiteral
			reduce(229), // ==, reduce: ArrayLiteral
			reduce(229), // !=, reduce: ArrayLiteral
			reduce(229), // <=>, reduce: ArrayLiteral
			reduce(229), // >=, reduce: ArrayLiteral
			reduce(229), // <, reduce: ArrayLiteral
			reduce(229), // <=, reduce: ArrayLiteral
			reduce(229), // ^, reduce: ArrayLiteral
			reduce(229), // &, reduce: ArrayLiteral
			reduce(229), // <<, reduce: ArrayLiteral
			reduce(229), // >>>, reduce: ArrayLiteral
			reduce(229), // +, reduce: ArrayLiteral
			reduce(229), // -, reduce: ArrayLiteral
			reduce(229), // .+, reduce: ArrayLiteral
			reduce(229), // .-, reduce: ArrayLiteral
			reduce(229), // *, reduce: ArrayLiteral
			reduce(229), // /, reduce: ArrayLiteral
			reduce(229), // //, reduce: ArrayLiteral
			reduce(229), // %, reduce: ArrayLiteral
			reduce(229), // .*, reduce: ArrayLiteral
			reduce(229), // ./, reduce: ArrayLiteral
			reduce(229), // .//, reduce: ArrayLiteral
			reduce(229), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(229), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(229), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(229), // [[, reduce: ArrayLiteral
			reduce(229), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3766
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(255), // }, reduce: ENV
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(255), // >, reduce: ENV
			reduce(255), // >>, reduce: ENV
			reduce(255), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(255), // ,, reduce: ENV
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(255), // ?, reduce: ENV
			nil,         // :
			reduce(255), // ||, reduce: ENV
			reduce(255), // ^^, reduce: ENV
			reduce(255), // &&, reduce: ENV
			reduce(255), // ??, reduce: ENV
			reduce(255), // ???, reduce: ENV
			reduce(255), // =~, reduce: ENV
			reduce(255), // !=~, reduce: ENV
			reduce(255), // ==, reduce: ENV
			reduce(255), // !=, reduce: ENV
			reduce(255), // <=>, reduce: ENV
			reduce(255), // >=, reduce: ENV
			reduce(255), // <, reduce: ENV
			reduce(255), // <=, reduce: ENV
			reduce(255), // ^, reduce: ENV
			reduce(255), // &, reduce: ENV
			reduce(255), // <<, reduce: ENV
			reduce(255), // >>>, reduce: ENV
			reduce(255), // +, reduce: ENV
			reduce(255), // -, reduce: ENV
			reduce(255), // .+, reduce: ENV
			reduce(255), // .-, reduce: ENV
			reduce(255), // *, reduce: ENV
			reduce(255), // /, reduce: ENV
			reduce(255), // //, reduce: ENV
			reduce(255), // %, reduce: ENV
			reduce(255), // .*, reduce: ENV
			reduce(255), // ./, reduce: ENV
			reduce(255), // .//, reduce: ENV
			reduce(255), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(255), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(255), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(255), // [[, reduce: ENV
			reduce(255), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3767
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4195), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3768
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4196), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3769
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(268), // }, reduce: FunctionCallsite
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(268), // >, reduce: FunctionCallsite
			reduce(268), // >>, reduce: FunctionCallsite
			reduce(268), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(268), // ,, reduce: FunctionCallsite
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(268), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(268), // ||, reduce: FunctionCallsite
			reduce(268), // ^^, reduce: FunctionCallsite
			reduce(268), // &&, reduce: FunctionCallsite
			reduce(268), // ??, reduce: FunctionCallsite
			reduce(268), // ???, reduce: FunctionCallsite
			reduce(268), // =~, reduce: FunctionCallsite
			reduce(268), // !=~, reduce: FunctionCallsite
			reduce(268), // ==, reduce: FunctionCallsite
			reduce(268), // !=, reduce: FunctionCallsite
			reduce(268), // <=>, reduce: FunctionCallsite
			reduce(268), // >=, reduce: FunctionCallsite
			reduce(268), // <, reduce: FunctionCallsite
			reduce(268), // <=, reduce: FunctionCallsite
			reduce(268), // ^, reduce: FunctionCallsite
			reduce(268), // &, reduce: FunctionCallsite
			reduce(268), // <<, reduce: FunctionCallsite
			reduce(268), // >>>, reduce: FunctionCallsite
			reduce(268), // +, reduce: FunctionCallsite
			reduce(268), // -, reduce: FunctionCallsite
			reduce(268), // .+, reduce: FunctionCallsite
			reduce(268), // .-, reduce: FunctionCallsite
			reduce(268), // *, reduce: FunctionCallsite
			reduce(268), // /, reduce: FunctionCallsite
			reduce(268), // //, reduce: FunctionCallsite
			reduce(268), // %, reduce: FunctionCallsite
			reduce(268), // .*, reduce: FunctionCallsite
			reduce(268), // ./, reduce: FunctionCallsite
			reduce(268), // .//, reduce: FunctionCallsite
			reduce(268), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(268), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(268), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(268), // [[, reduce: FunctionCallsite
			reduce(268), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3770
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4197), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3771
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(108), // >, reduce: PositionalFieldValue
			reduce(108), // >>, reduce: PositionalFieldValue
			reduce(108), // |, reduce: PositionalFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(108), // ?, reduce: PositionalFieldValue
			reduce(108), // :, reduce: PositionalFieldValue
			reduce(108), // ||, reduce: PositionalFieldValue
			reduce(108), // ^^, reduce: PositionalFieldValue
			reduce(108), // &&, reduce: PositionalFieldValue
			reduce(108), // ??, reduce: PositionalFieldValue
			reduce(108), // ???, reduce: PositionalFieldValue
			reduce(108), // =~, reduce: PositionalFieldValue
			reduce(108), // !=~, reduce: PositionalFieldValue
			reduce(108), // ==, reduce: PositionalFieldValue
			reduce(108), // !=, reduce: PositionalFieldValue
			reduce(108), // <=>, reduce: PositionalFieldValue
			reduce(108), // >=, reduce: PositionalFieldValue
			reduce(108), // <, reduce: PositionalFieldValue
			reduce(108), // <=, reduce: PositionalFieldValue
			reduce(108), // ^, reduce: PositionalFieldValue
			reduce(108), // &, reduce: PositionalFieldValue
			reduce(108), // <<, reduce: PositionalFieldValue
			reduce(108), // >>>, reduce: PositionalFieldValue
			reduce(108), // +, reduce: PositionalFieldValue
			reduce(108), // -, reduce: PositionalFieldValue
			reduce(108), // .+, reduce: PositionalFieldValue
			reduce(108), // .-, reduce: PositionalFieldValue
			reduce(108), // *, reduce: PositionalFieldValue
			reduce(108), // /, reduce: PositionalFieldValue
			reduce(108), // //, reduce: PositionalFieldValue
			reduce(108), // %, reduce: PositionalFieldValue
			reduce(108), // .*, reduce: PositionalFieldValue
			reduce(108), // ./, reduce: PositionalFieldValue
			reduce(108), // .//, reduce: PositionalFieldValue
			reduce(108), // ., reduce: PositionalFieldValue
			nil,         // !
			nil,         // ~
			reduce(108), // **, reduce: PositionalFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(108), // [, reduce: PositionalFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(108), // [[, reduce: PositionalFieldValue
			reduce(108), // [[[, reduce: PositionalFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3772
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			reduce(149), // :, reduce: TernaryTerm
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3773
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4198), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3774
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(265), // >, reduce: ArraySliceAccess
			reduce(265), // >>, reduce: ArraySliceAccess
			reduce(265), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(265), // ?, reduce: ArraySliceAccess
			reduce(265), // :, reduce: ArraySliceAccess
			reduce(265), // ||, reduce: ArraySliceAccess
			reduce(265), // ^^, reduce: ArraySliceAccess
			reduce(265), // &&, reduce: ArraySliceAccess
			reduce(265), // ??, reduce: ArraySliceAccess
			reduce(265), // ???, reduce: ArraySliceAccess
			reduce(265), // =~, reduce: ArraySliceAccess
			reduce(265), // !=~, reduce: ArraySliceAccess
			reduce(265), // ==, reduce: ArraySliceAccess
			reduce(265), // !=, reduce: ArraySliceAccess
			reduce(265), // <=>, reduce: ArraySliceAccess
			reduce(265), // >=, reduce: ArraySliceAccess
			reduce(265), // <, reduce: ArraySliceAccess
			reduce(265), // <=, reduce: ArraySliceAccess
			reduce(265), // ^, reduce: ArraySliceAccess
			reduce(265), // &, reduce: ArraySliceAccess
			reduce(265), // <<, reduce: ArraySliceAccess
			reduce(265), // >>>, reduce: ArraySliceAccess
			reduce(265), // +, reduce: ArraySliceAccess
			reduce(265), // -, reduce: ArraySliceAccess
			reduce(265), // .+, reduce: ArraySliceAccess
			reduce(265), // .-, reduce: ArraySliceAccess
			reduce(265), // *, reduce: ArraySliceAccess
			reduce(265), // /, reduce: ArraySliceAccess
			reduce(265), // //, reduce: ArraySliceAccess
			reduce(265), // %, reduce: ArraySliceAccess
			reduce(265), // .*, reduce: ArraySliceAccess
			reduce(265), // ./, reduce: ArraySliceAccess
			reduce(265), // .//, reduce: ArraySliceAccess
			reduce(265), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(265), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(265), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(265), // [[, reduce: ArraySliceAccess
			reduce(265), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3775
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(264), // >, reduce: ArraySliceAccess
			reduce(264), // >>, reduce: ArraySliceAccess
			reduce(264), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(264), // ?, reduce: ArraySliceAccess
			reduce(264), // :, reduce: ArraySliceAccess
			reduce(264), // ||, reduce: ArraySliceAccess
			reduce(264), // ^^, reduce: ArraySliceAccess
			reduce(264), // &&, reduce: ArraySliceAccess
			reduce(264), // ??, reduce: ArraySliceAccess
			reduce(264), // ???, reduce: ArraySliceAccess
			reduce(264), // =~, reduce: ArraySliceAccess
			reduce(264), // !=~, reduce: ArraySliceAccess
			reduce(264), // ==, reduce: ArraySliceAccess
			reduce(264), // !=, reduce: ArraySliceAccess
			reduce(264), // <=>, reduce: ArraySliceAccess
			reduce(264), // >=, reduce: ArraySliceAccess
			reduce(264), // <, reduce: ArraySliceAccess
			reduce(264), // <=, reduce: ArraySliceAccess
			reduce(264), // ^, reduce: ArraySliceAccess
			reduce(264), // &, reduce: ArraySliceAccess
			reduce(264), // <<, reduce: ArraySliceAccess
			reduce(264), // >>>, reduce: ArraySliceAccess
			reduce(264), // +, reduce: ArraySliceAccess
			reduce(264), // -, reduce: ArraySliceAccess
			reduce(264), // .+, reduce: ArraySliceAccess
			reduce(264), // .-, reduce: ArraySliceAccess
			reduce(264), // *, reduce: ArraySliceAccess
			reduce(264), // /, reduce: ArraySliceAccess
			reduce(264), // //, reduce: ArraySliceAccess
			reduce(264), // %, reduce: ArraySliceAccess
			reduce(264), // .*, reduce: ArraySliceAccess
			reduce(264), // ./, reduce: ArraySliceAccess
			reduce(264), // .//, reduce: ArraySliceAccess
			reduce(264), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(264), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(264), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(264), // [[, reduce: ArraySliceAccess
			reduce(264), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3776
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(261), // >, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |, reduce: ArrayOrMapPositionalNameAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(261), // ?, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // :, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ||, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // =~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ==, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .+, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .-, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .*, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ./, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .//, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ., reduce: ArrayOrMapPositionalNameAccess
			nil,         // !
			nil,         // ~
			reduce(261), // **, reduce: ArrayOrMapPositionalNameAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(261), // [, reduce: ArrayOrMapPositionalNameAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(261), // [[, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // [[[, reduce: ArrayOrMapPositionalNameAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3777
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4199), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3778
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(324), // >, reduce: UnnamedFunctionDefinition
			reduce(324), // >>, reduce: UnnamedFunctionDefinition
			reduce(324), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(324), // ?, reduce: UnnamedFunctionDefinition
			reduce(324), // :, reduce: UnnamedFunctionDefinition
			reduce(324), // ||, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^, reduce: UnnamedFunctionDefinition
			reduce(324), // &&, reduce: UnnamedFunctionDefinition
			reduce(324), // ??, reduce: UnnamedFunctionDefinition
			reduce(324), // ???, reduce: UnnamedFunctionDefinition
			reduce(324), // =~, reduce: UnnamedFunctionDefinition
			reduce(324), // !=~, reduce: UnnamedFunctionDefinition
			reduce(324), // ==, reduce: UnnamedFunctionDefinition
			reduce(324), // !=, reduce: UnnamedFunctionDefinition
			reduce(324), // <=>, reduce: UnnamedFunctionDefinition
			reduce(324), // >=, reduce: UnnamedFunctionDefinition
			reduce(324), // <, reduce: UnnamedFunctionDefinition
			reduce(324), // <=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^, reduce: UnnamedFunctionDefinition
			reduce(324), // &, reduce: UnnamedFunctionDefinition
			reduce(324), // <<, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>, reduce: UnnamedFunctionDefinition
			reduce(324), // +, reduce: UnnamedFunctionDefinition
			reduce(324), // -, reduce: UnnamedFunctionDefinition
			reduce(324), // .+, reduce: UnnamedFunctionDefinition
			reduce(324), // .-, reduce: UnnamedFunctionDefinition
			reduce(324), // *, reduce: UnnamedFunctionDefinition
			reduce(324), // /, reduce: UnnamedFunctionDefinition
			reduce(324), // //, reduce: UnnamedFunctionDefinition
			reduce(324), // %, reduce: UnnamedFunctionDefinition
			reduce(324), // .*, reduce: UnnamedFunctionDefinition
			reduce(324), // ./, reduce: UnnamedFunctionDefinition
			reduce(324), // .//, reduce: UnnamedFunctionDefinition
			reduce(324), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(324), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(324), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(324), // [[, reduce: UnnamedFunctionDefinition
			reduce(324), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3779
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S3780
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3781
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			reduce(12), // }, reduce: NonEmptyStatementBlock
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3782
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(13), // ;, reduce: StatementBlockInBraces
			reduce(13), // {, reduce: StatementBlockInBraces
			reduce(13), // }, reduce: StatementBlockInBraces
			nil,        // =
			reduce(13), // unset, reduce: StatementBlockInBraces
			reduce(13), // filter, reduce: StatementBlockInBraces
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			reduce(13), // print, reduce: StatementBlockInBraces
			nil,        // ,
			reduce(13), // printn, reduce: StatementBlockInBraces
			reduce(13), // eprint, reduce: StatementBlockInBraces
			reduce(13), // eprintn, reduce: StatementBlockInBraces
			reduce(13), // dump, reduce: StatementBlockInBraces
			reduce(13), // edump, reduce: StatementBlockInBraces
			reduce(13), // tee, reduce: StatementBlockInBraces
			reduce(13), // emitf, reduce: StatementBlockInBraces
			reduce(13), // emit1, reduce: StatementBlockInBraces
			reduce(13), // emit, reduce: StatementBlockInBraces
			reduce(13), // (, reduce: StatementBlockInBraces
			nil,        // )
			reduce(13), // emitp, reduce: StatementBlockInBraces
			reduce(13), // field_name, reduce: StatementBlockInBraces
			reduce(13), // $[, reduce: StatementBlockInBraces
			nil,        // ]
			reduce(13), // braced_field_name, reduce: StatementBlockInBraces
			reduce(13), // $[[, reduce: StatementBlockInBraces
			reduce(13), // $[[[, reduce: StatementBlockInBraces
			reduce(13), // full_srec, reduce: StatementBlockInBraces
			reduce(13), // oosvar_name, reduce: StatementBlockInBraces
			reduce(13), // @[, reduce: StatementBlockInBraces
			reduce(13), // braced_oosvar_name, reduce: StatementBlockInBraces
			reduce(13), // full_oosvar, reduce: StatementBlockInBraces
			reduce(13), // all, reduce: StatementBlockInBraces
			reduce(13), // non_sigil_name, reduce: StatementBlockInBraces
			reduce(13), // arr, reduce: StatementBlockInBraces
			reduce(13), // bool, reduce: StatementBlockInBraces
			reduce(13), // float, reduce: StatementBlockInBraces
			reduce(13), // int, reduce: StatementBlockInBraces
			reduce(13), // map, reduce: StatementBlockInBraces
			reduce(13), // num, reduce: StatementBlockInBraces
			reduce(13), // str, reduce: StatementBlockInBraces
			reduce(13), // var, reduce: StatementBlockInBraces
			reduce(13), // funct, reduce: StatementBlockInBraces
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			reduce(13), // !, reduce: StatementBlockInBraces
			reduce(13), // ~, reduce: StatementBlockInBraces
			nil,        // **
			reduce(13), // string_literal, reduce: StatementBlockInBraces
			reduce(13), // regex_case_insensitive, reduce: StatementBlockInBraces
			reduce(13), // int_literal, reduce: StatementBlockInBraces
			reduce(13), // float_literal, reduce: StatementBlockInBraces
			reduce(13), // boolean_literal, reduce: StatementBlockInBraces
			reduce(13), // null_literal, reduce: StatementBlockInBraces
			reduce(13), // inf_literal, reduce: StatementBlockInBraces
			reduce(13), // nan_literal, reduce: StatementBlockInBraces
			reduce(13), // const_M_PI, reduce: StatementBlockInBraces
			reduce(13), // const_M_E, reduce: StatementBlockInBraces
			reduce(13), // panic, reduce: StatementBlockInBraces
			reduce(13), // [, reduce: StatementBlockInBraces
			reduce(13), // ctx_IPS, reduce: StatementBlockInBraces
			reduce(13), // ctx_IFS, reduce: StatementBlockInBraces
			reduce(13), // ctx_IRS, reduce: StatementBlockInBraces
			reduce(13), // ctx_OPS, reduce: StatementBlockInBraces
			reduce(13), // ctx_OFS, reduce: StatementBlockInBraces
			reduce(13), // ctx_ORS, reduce: StatementBlockInBraces
			reduce(13), // ctx_FLATSEP, reduce: StatementBlockInBraces
			reduce(13), // ctx_NF, reduce: StatementBlockInBraces
			reduce(13), // ctx_NR, reduce: StatementBlockInBraces
			reduce(13), // ctx_FNR, reduce: StatementBlockInBraces
			reduce(13), // ctx_FILENAME, reduce: StatementBlockInBraces
			reduce(13), // ctx_FILENUM, reduce: StatementBlockInBraces
			reduce(13), // env, reduce: StatementBlockInBraces
			nil,        // [[
			nil,        // [[[
			reduce(13), // call, reduce: StatementBlockInBraces
			reduce(13), // begin, reduce: StatementBlockInBraces
			reduce(13), // end, reduce: StatementBlockInBraces
			reduce(13), // if, reduce: StatementBlockInBraces
			nil,        // elif
			nil,        // else
			reduce(13), // while, reduce: StatementBlockInBraces
			reduce(13), // do, reduce: StatementBlockInBraces
			reduce(13), // for, reduce: StatementBlockInBraces
			nil,        // in
			reduce(13), // break, reduce: StatementBlockInBraces
			reduce(13), // continue, reduce: StatementBlockInBraces
			reduce(13), // func, reduce: StatementBlockInBraces
			reduce(13), // subr, reduce: StatementBlockInBraces
			reduce(13), // return, reduce: StatementBlockInBraces
		},
	},
	actionRow{ // S3783
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(235), // ;, reduce: MapLiteral
			nil,         // {
			reduce(235), // }, reduce: MapLiteral
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(235), // >, reduce: MapLiteral
			reduce(235), // >>, reduce: MapLiteral
			reduce(235), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(235), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(235), // ?, reduce: MapLiteral
			nil,         // :
			reduce(235), // ||, reduce: MapLiteral
			reduce(235), // ^^, reduce: MapLiteral
			reduce(235), // &&, reduce: MapLiteral
			reduce(235), // ??, reduce: MapLiteral
			reduce(235), // ???, reduce: MapLiteral
			reduce(235), // =~, reduce: MapLiteral
			reduce(235), // !=~, reduce: MapLiteral
			reduce(235), // ==, reduce: MapLiteral
			reduce(235), // !=, reduce: MapLiteral
			reduce(235), // <=>, reduce: MapLiteral
			reduce(235), // >=, reduce: MapLiteral
			reduce(235), // <, reduce: MapLiteral
			reduce(235), // <=, reduce: MapLiteral
			reduce(235), // ^, reduce: MapLiteral
			reduce(235), // &, reduce: MapLiteral
			reduce(235), // <<, reduce: MapLiteral
			reduce(235), // >>>, reduce: MapLiteral
			reduce(235), // +, reduce: MapLiteral
			reduce(235), // -, reduce: MapLiteral
			reduce(235), // .+, reduce: MapLiteral
			reduce(235), // .-, reduce: MapLiteral
			reduce(235), // *, reduce: MapLiteral
			reduce(235), // /, reduce: MapLiteral
			reduce(235), // //, reduce: MapLiteral
			reduce(235), // %, reduce: MapLiteral
			reduce(235), // .*, reduce: MapLiteral
			reduce(235), // ./, reduce: MapLiteral
			reduce(235), // .//, reduce: MapLiteral
			reduce(235), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(235), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(235), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(235), // [[, reduce: MapLiteral
			reduce(235), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3784
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(275), // ;, reduce: FcnArgs
			nil,         // {
			reduce(275), // }, reduce: FcnArgs
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3785
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(208), // ;, reduce: PrecedenceChainEnd
			nil,         // {
			reduce(208), // }, reduce: PrecedenceChainEnd
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(208), // >, reduce: PrecedenceChainEnd
			reduce(208), // >>, reduce: PrecedenceChainEnd
			reduce(208), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(208), // ,, reduce: PrecedenceChainEnd
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(208), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(208), // ||, reduce: PrecedenceChainEnd
			reduce(208), // ^^, reduce: PrecedenceChainEnd
			reduce(208), // &&, reduce: PrecedenceChainEnd
			reduce(208), // ??, reduce: PrecedenceChainEnd
			reduce(208), // ???, reduce: PrecedenceChainEnd
			reduce(208), // =~, reduce: PrecedenceChainEnd
			reduce(208), // !=~, reduce: PrecedenceChainEnd
			reduce(208), // ==, reduce: PrecedenceChainEnd
			reduce(208), // !=, reduce: PrecedenceChainEnd
			reduce(208), // <=>, reduce: PrecedenceChainEnd
			reduce(208), // >=, reduce: PrecedenceChainEnd
			reduce(208), // <, reduce: PrecedenceChainEnd
			reduce(208), // <=, reduce: PrecedenceChainEnd
			reduce(208), // ^, reduce: PrecedenceChainEnd
			reduce(208), // &, reduce: PrecedenceChainEnd
			reduce(208), // <<, reduce: PrecedenceChainEnd
			reduce(208), // >>>, reduce: PrecedenceChainEnd
			reduce(208), // +, reduce: PrecedenceChainEnd
			reduce(208), // -, reduce: PrecedenceChainEnd
			reduce(208), // .+, reduce: PrecedenceChainEnd
			reduce(208), // .-, reduce: PrecedenceChainEnd
			reduce(208), // *, reduce: PrecedenceChainEnd
			reduce(208), // /, reduce: PrecedenceChainEnd
			reduce(208), // //, reduce: PrecedenceChainEnd
			reduce(208), // %, reduce: PrecedenceChainEnd
			reduce(208), // .*, reduce: PrecedenceChainEnd
			reduce(208), // ./, reduce: PrecedenceChainEnd
			reduce(208), // .//, reduce: PrecedenceChainEnd
			reduce(208), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(208), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3786
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(105), // ;, reduce: IndirectFieldValue
			nil,         // {
			reduce(105), // }, reduce: IndirectFieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(105), // >, reduce: IndirectFieldValue
			reduce(105), // >>, reduce: IndirectFieldValue
			reduce(105), // |, reduce: IndirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(105), // ,, reduce: IndirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(105), // ?, reduce: IndirectFieldValue
			nil,         // :
			reduce(105), // ||, reduce: IndirectFieldValue
			reduce(105), // ^^, reduce: IndirectFieldValue
			reduce(105), // &&, reduce: IndirectFieldValue
			reduce(105), // ??, reduce: IndirectFieldValue
			reduce(105), // ???, reduce: IndirectFieldValue
			reduce(105), // =~, reduce: IndirectFieldValue
			reduce(105), // !=~, reduce: IndirectFieldValue
			reduce(105), // ==, reduce: IndirectFieldValue
			reduce(105), // !=, reduce: IndirectFieldValue
			reduce(105), // <=>, reduce: IndirectFieldValue
			reduce(105), // >=, reduce: IndirectFieldValue
			reduce(105), // <, reduce: IndirectFieldValue
			reduce(105), // <=, reduce: IndirectFieldValue
			reduce(105), // ^, reduce: IndirectFieldValue
			reduce(105), // &, reduce: IndirectFieldValue
			reduce(105), // <<, reduce: IndirectFieldValue
			reduce(105), // >>>, reduce: IndirectFieldValue
			reduce(105), // +, reduce: IndirectFieldValue
			reduce(105), // -, reduce: IndirectFieldValue
			reduce(105), // .+, reduce: IndirectFieldValue
			reduce(105), // .-, reduce: IndirectFieldValue
			reduce(105), // *, reduce: IndirectFieldValue
			reduce(105), // /, reduce: IndirectFieldValue
			reduce(105), // //, reduce: IndirectFieldValue
			reduce(105), // %, reduce: IndirectFieldValue
			reduce(105), // .*, reduce: IndirectFieldValue
			reduce(105), // ./, reduce: IndirectFieldValue
			reduce(105), // .//, reduce: IndirectFieldValue
			reduce(105), // ., reduce: IndirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(105), // **, reduce: IndirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(105), // [, reduce: IndirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(105), // [[, reduce: IndirectFieldValue
			reduce(105), // [[[, reduce: IndirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3787
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4202), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3788
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4203), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3789
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(114), // ;, reduce: IndirectOosvarValue
			nil,         // {
			reduce(114), // }, reduce: IndirectOosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(114), // >, reduce: IndirectOosvarValue
			reduce(114), // >>, reduce: IndirectOosvarValue
			reduce(114), // |, reduce: IndirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(114), // ,, reduce: IndirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(114), // ?, reduce: IndirectOosvarValue
			nil,         // :
			reduce(114), // ||, reduce: IndirectOosvarValue
			reduce(114), // ^^, reduce: IndirectOosvarValue
			reduce(114), // &&, reduce: IndirectOosvarValue
			reduce(114), // ??, reduce: IndirectOosvarValue
			reduce(114), // ???, reduce: IndirectOosvarValue
			reduce(114), // =~, reduce: IndirectOosvarValue
			reduce(114), // !=~, reduce: IndirectOosvarValue
			reduce(114), // ==, reduce: IndirectOosvarValue
			reduce(114), // !=, reduce: IndirectOosvarValue
			reduce(114), // <=>, reduce: IndirectOosvarValue
			reduce(114), // >=, reduce: IndirectOosvarValue
			reduce(114), // <, reduce: IndirectOosvarValue
			reduce(114), // <=, reduce: IndirectOosvarValue
			reduce(114), // ^, reduce: IndirectOosvarValue
			reduce(114), // &, reduce: IndirectOosvarValue
			reduce(114), // <<, reduce: IndirectOosvarValue
			reduce(114), // >>>, reduce: IndirectOosvarValue
			reduce(114), // +, reduce: IndirectOosvarValue
			reduce(114), // -, reduce: IndirectOosvarValue
			reduce(114), // .+, reduce: IndirectOosvarValue
			reduce(114), // .-, reduce: IndirectOosvarValue
			reduce(114), // *, reduce: IndirectOosvarValue
			reduce(114), // /, reduce: IndirectOosvarValue
			reduce(114), // //, reduce: IndirectOosvarValue
			reduce(114), // %, reduce: IndirectOosvarValue
			reduce(114), // .*, reduce: IndirectOosvarValue
			reduce(114), // ./, reduce: IndirectOosvarValue
			reduce(114), // .//, reduce: IndirectOosvarValue
			reduce(114), // ., reduce: IndirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(114), // **, reduce: IndirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(114), // [, reduce: IndirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(114), // [[, reduce: IndirectOosvarValue
			reduce(114), // [[[, reduce: IndirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3790
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4204), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3791
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(151), // ;, reduce: LogicalOrTerm
			nil,         // {
			reduce(151), // }, reduce: LogicalOrTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(151), // ,, reduce: LogicalOrTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(151), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(151), // ||, reduce: LogicalOrTerm
			shift(3011), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3792
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(153), // ;, reduce: LogicalXORTerm
			nil,         // {
			reduce(153), // }, reduce: LogicalXORTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(153), // ,, reduce: LogicalXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(153), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(153), // ||, reduce: LogicalXORTerm
			reduce(153), // ^^, reduce: LogicalXORTerm
			shift(3012), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3793
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(155), // ;, reduce: LogicalAndTerm
			nil,         // {
			reduce(155), // }, reduce: LogicalAndTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(155), // ,, reduce: LogicalAndTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(155), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(155), // ||, reduce: LogicalAndTerm
			reduce(155), // ^^, reduce: LogicalAndTerm
			reduce(155), // &&, reduce: LogicalAndTerm
			shift(3013), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3794
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(157), // ;, reduce: AbsentCoalesceTerm
			nil,         // {
			reduce(157), // }, reduce: AbsentCoalesceTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(157), // ,, reduce: AbsentCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(157), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(157), // ||, reduce: AbsentCoalesceTerm
			reduce(157), // ^^, reduce: AbsentCoalesceTerm
			reduce(157), // &&, reduce: AbsentCoalesceTerm
			reduce(157), // ??, reduce: AbsentCoalesceTerm
			shift(3014), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3795
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(159), // ;, reduce: EmptyCoalesceTerm
			nil,         // {
			reduce(159), // }, reduce: EmptyCoalesceTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(159), // ,, reduce: EmptyCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(159), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(159), // ||, reduce: EmptyCoalesceTerm
			reduce(159), // ^^, reduce: EmptyCoalesceTerm
			reduce(159), // &&, reduce: EmptyCoalesceTerm
			reduce(159), // ??, reduce: EmptyCoalesceTerm
			reduce(159), // ???, reduce: EmptyCoalesceTerm
			shift(3015), // =~
			shift(3016), // !=~
			shift(3017), // ==
			shift(3018), // !=
			shift(3019), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3796
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(161), // ;, reduce: EqneTerm
			nil,         // {
			reduce(161), // }, reduce: EqneTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(3020), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(161), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(161), // ?, reduce: EqneTerm
			nil,         // :
			reduce(161), // ||, reduce: EqneTerm
			reduce(161), // ^^, reduce: EqneTerm
			reduce(161), // &&, reduce: EqneTerm
			reduce(161), // ??, reduce: EqneTerm
			reduce(161), // ???, reduce: EqneTerm
			reduce(161), // =~, reduce: EqneTerm
			reduce(161), // !=~, reduce: EqneTerm
			reduce(161), // ==, reduce: EqneTerm
			reduce(161), // !=, reduce: EqneTerm
			reduce(161), // <=>, reduce: EqneTerm
			shift(3021), // >=
			shift(3022), // <
			shift(3023), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3797
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(162), // ;, reduce: EqneTerm
			nil,         // {
			reduce(162), // }, reduce: EqneTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(3020), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(162), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(162), // ?, reduce: EqneTerm
			nil,         // :
			reduce(162), // ||, reduce: EqneTerm
			reduce(162), // ^^, reduce: EqneTerm
			reduce(162), // &&, reduce: EqneTerm
			reduce(162), // ??, reduce: EqneTerm
			reduce(162), // ???, reduce: EqneTerm
			reduce(162), // =~, reduce: EqneTerm
			reduce(162), // !=~, reduce: EqneTerm
			reduce(162), // ==, reduce: EqneTerm
			reduce(162), // !=, reduce: EqneTerm
			reduce(162), // <=>, reduce: EqneTerm
			shift(3021), // >=
			shift(3022), // <
			shift(3023), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3798
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(163), // ;, reduce: EqneTerm
			nil,         // {
			reduce(163), // }, reduce: EqneTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(3020), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(163), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(163), // ?, reduce: EqneTerm
			nil,         // :
			reduce(163), // ||, reduce: EqneTerm
			reduce(163), // ^^, reduce: EqneTerm
			reduce(163), // &&, reduce: EqneTerm
			reduce(163), // ??, reduce: EqneTerm
			reduce(163), // ???, reduce: EqneTerm
			reduce(163), // =~, reduce: EqneTerm
			reduce(163), // !=~, reduce: EqneTerm
			reduce(163), // ==, reduce: EqneTerm
			reduce(163), // !=, reduce: EqneTerm
			reduce(163), // <=>, reduce: EqneTerm
			shift(3021), // >=
			shift(3022), // <
			shift(3023), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3799
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(164), // ;, reduce: EqneTerm
			nil,         // {
			reduce(164), // }, reduce: EqneTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(3020), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(164), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(164), // ?, reduce: EqneTerm
			nil,         // :
			reduce(164), // ||, reduce: EqneTerm
			reduce(164), // ^^, reduce: EqneTerm
			reduce(164), // &&, reduce: EqneTerm
			reduce(164), // ??, reduce: EqneTerm
			reduce(164), // ???, reduce: EqneTerm
			reduce(164), // =~, reduce: EqneTerm
			reduce(164), // !=~, reduce: EqneTerm
			reduce(164), // ==, reduce: EqneTerm
			reduce(164), // !=, reduce: EqneTerm
			reduce(164), // <=>, reduce: EqneTerm
			shift(3021), // >=
			shift(3022), // <
			shift(3023), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3800
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(165), // ;, reduce: EqneTerm
			nil,         // {
			reduce(165), // }, reduce: EqneTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(3020), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(165), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(165), // ?, reduce: EqneTerm
			nil,         // :
			reduce(165), // ||, reduce: EqneTerm
			reduce(165), // ^^, reduce: EqneTerm
			reduce(165), // &&, reduce: EqneTerm
			reduce(165), // ??, reduce: EqneTerm
			reduce(165), // ???, reduce: EqneTerm
			reduce(165), // =~, reduce: EqneTerm
			reduce(165), // !=~, reduce: EqneTerm
			reduce(165), // ==, reduce: EqneTerm
			reduce(165), // !=, reduce: EqneTerm
			reduce(165), // <=>, reduce: EqneTerm
			shift(3021), // >=
			shift(3022), // <
			shift(3023), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3801
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(167), // ;, reduce: CmpTerm
			nil,         // {
			reduce(167), // }, reduce: CmpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(167), // >, reduce: CmpTerm
			nil,         // >>
			shift(3024), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(167), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(167), // ?, reduce: CmpTerm
			nil,         // :
			reduce(167), // ||, reduce: CmpTerm
			reduce(167), // ^^, reduce: CmpTerm
			reduce(167), // &&, reduce: CmpTerm
			reduce(167), // ??, reduce: CmpTerm
			reduce(167), // ???, reduce: CmpTerm
			reduce(167), // =~, reduce: CmpTerm
			reduce(167), // !=~, reduce: CmpTerm
			reduce(167), // ==, reduce: CmpTerm
			reduce(167), // !=, reduce: CmpTerm
			reduce(167), // <=>, reduce: CmpTerm
			reduce(167), // >=, reduce: CmpTerm
			reduce(167), // <, reduce: CmpTerm
			reduce(167), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3802
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(168), // ;, reduce: CmpTerm
			nil,         // {
			reduce(168), // }, reduce: CmpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(168), // >, reduce: CmpTerm
			nil,         // >>
			shift(3024), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(168), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(168), // ?, reduce: CmpTerm
			nil,         // :
			reduce(168), // ||, reduce: CmpTerm
			reduce(168), // ^^, reduce: CmpTerm
			reduce(168), // &&, reduce: CmpTerm
			reduce(168), // ??, reduce: CmpTerm
			reduce(168), // ???, reduce: CmpTerm
			reduce(168), // =~, reduce: CmpTerm
			reduce(168), // !=~, reduce: CmpTerm
			reduce(168), // ==, reduce: CmpTerm
			reduce(168), // !=, reduce: CmpTerm
			reduce(168), // <=>, reduce: CmpTerm
			reduce(168), // >=, reduce: CmpTerm
			reduce(168), // <, reduce: CmpTerm
			reduce(168), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3803
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(169), // ;, reduce: CmpTerm
			nil,         // {
			reduce(169), // }, reduce: CmpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(169), // >, reduce: CmpTerm
			nil,         // >>
			shift(3024), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(169), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(169), // ?, reduce: CmpTerm
			nil,         // :
			reduce(169), // ||, reduce: CmpTerm
			reduce(169), // ^^, reduce: CmpTerm
			reduce(169), // &&, reduce: CmpTerm
			reduce(169), // ??, reduce: CmpTerm
			reduce(169), // ???, reduce: CmpTerm
			reduce(169), // =~, reduce: CmpTerm
			reduce(169), // !=~, reduce: CmpTerm
			reduce(169), // ==, reduce: CmpTerm
			reduce(169), // !=, reduce: CmpTerm
			reduce(169), // <=>, reduce: CmpTerm
			reduce(169), // >=, reduce: CmpTerm
			reduce(169), // <, reduce: CmpTerm
			reduce(169), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3804
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(170), // ;, reduce: CmpTerm
			nil,         // {
			reduce(170), // }, reduce: CmpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(170), // >, reduce: CmpTerm
			nil,         // >>
			shift(3024), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(170), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(170), // ?, reduce: CmpTerm
			nil,         // :
			reduce(170), // ||, reduce: CmpTerm
			reduce(170), // ^^, reduce: CmpTerm
			reduce(170), // &&, reduce: CmpTerm
			reduce(170), // ??, reduce: CmpTerm
			reduce(170), // ???, reduce: CmpTerm
			reduce(170), // =~, reduce: CmpTerm
			reduce(170), // !=~, reduce: CmpTerm
			reduce(170), // ==, reduce: CmpTerm
			reduce(170), // !=, reduce: CmpTerm
			reduce(170), // <=>, reduce: CmpTerm
			reduce(170), // >=, reduce: CmpTerm
			reduce(170), // <, reduce: CmpTerm
			reduce(170), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3805
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(172), // ;, reduce: BitwiseORTerm
			nil,         // {
			reduce(172), // }, reduce: BitwiseORTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(172), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(172), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(172), // ,, reduce: BitwiseORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(172), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(172), // ||, reduce: BitwiseORTerm
			reduce(172), // ^^, reduce: BitwiseORTerm
			reduce(172), // &&, reduce: BitwiseORTerm
			reduce(172), // ??, reduce: BitwiseORTerm
			reduce(172), // ???, reduce: BitwiseORTerm
			reduce(172), // =~, reduce: BitwiseORTerm
			reduce(172), // !=~, reduce: BitwiseORTerm
			reduce(172), // ==, reduce: BitwiseORTerm
			reduce(172), // !=, reduce: BitwiseORTerm
			reduce(172), // <=>, reduce: BitwiseORTerm
			reduce(172), // >=, reduce: BitwiseORTerm
			reduce(172), // <, reduce: BitwiseORTerm
			reduce(172), // <=, reduce: BitwiseORTerm
			shift(3025), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3806
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(174), // ;, reduce: BitwiseXORTerm
			nil,         // {
			reduce(174), // }, reduce: BitwiseXORTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(174), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(174), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(174), // ,, reduce: BitwiseXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(174), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(174), // ||, reduce: BitwiseXORTerm
			reduce(174), // ^^, reduce: BitwiseXORTerm
			reduce(174), // &&, reduce: BitwiseXORTerm
			reduce(174), // ??, reduce: BitwiseXORTerm
			reduce(174), // ???, reduce: BitwiseXORTerm
			reduce(174), // =~, reduce: BitwiseXORTerm
			reduce(174), // !=~, reduce: BitwiseXORTerm
			reduce(174), // ==, reduce: BitwiseXORTerm
			reduce(174), // !=, reduce: BitwiseXORTerm
			reduce(174), // <=>, reduce: BitwiseXORTerm
			reduce(174), // >=, reduce: BitwiseXORTerm
			reduce(174), // <, reduce: BitwiseXORTerm
			reduce(174), // <=, reduce: BitwiseXORTerm
			reduce(174), // ^, reduce: BitwiseXORTerm
			shift(3026), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3807
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(176), // ;, reduce: BitwiseANDTerm
			nil,         // {
			reduce(176), // }, reduce: BitwiseANDTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(176), // >, reduce: BitwiseANDTerm
			shift(3027), // >>
			reduce(176), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(176), // ,, reduce: BitwiseANDTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(176), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(176), // ||, reduce: BitwiseANDTerm
			reduce(176), // ^^, reduce: BitwiseANDTerm
			reduce(176), // &&, reduce: BitwiseANDTerm
			reduce(176), // ??, reduce: BitwiseANDTerm
			reduce(176), // ???, reduce: BitwiseANDTerm
			reduce(176), // =~, reduce: BitwiseANDTerm
			reduce(176), // !=~, reduce: BitwiseANDTerm
			reduce(176), // ==, reduce: BitwiseANDTerm
			reduce(176), // !=, reduce: BitwiseANDTerm
			reduce(176), // <=>, reduce: BitwiseANDTerm
			reduce(176), // >=, reduce: BitwiseANDTerm
			reduce(176), // <, reduce: BitwiseANDTerm
			reduce(176), // <=, reduce: BitwiseANDTerm
			reduce(176), // ^, reduce: BitwiseANDTerm
			reduce(176), // &, reduce: BitwiseANDTerm
			shift(3028), // <<
			shift(3029), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3808
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(179), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			reduce(179), // }, reduce: BitwiseShiftTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(179), // >, reduce: BitwiseShiftTerm
			reduce(179), // >>, reduce: BitwiseShiftTerm
			reduce(179), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(179), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(179), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(179), // ||, reduce: BitwiseShiftTerm
			reduce(179), // ^^, reduce: BitwiseShiftTerm
			reduce(179), // &&, reduce: BitwiseShiftTerm
			reduce(179), // ??, reduce: BitwiseShiftTerm
			reduce(179), // ???, reduce: BitwiseShiftTerm
			reduce(179), // =~, reduce: BitwiseShiftTerm
			reduce(179), // !=~, reduce: BitwiseShiftTerm
			reduce(179), // ==, reduce: BitwiseShiftTerm
			reduce(179), // !=, reduce: BitwiseShiftTerm
			reduce(179), // <=>, reduce: BitwiseShiftTerm
			reduce(179), // >=, reduce: BitwiseShiftTerm
			reduce(179), // <, reduce: BitwiseShiftTerm
			reduce(179), // <=, reduce: BitwiseShiftTerm
			reduce(179), // ^, reduce: BitwiseShiftTerm
			reduce(179), // &, reduce: BitwiseShiftTerm
			reduce(179), // <<, reduce: BitwiseShiftTerm
			reduce(179), // >>>, reduce: BitwiseShiftTerm
			shift(3030), // +
			shift(3031), // -
			shift(3032), // .+
			shift(3033), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3809
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(178), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			reduce(178), // }, reduce: BitwiseShiftTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(178), // >, reduce: BitwiseShiftTerm
			reduce(178), // >>, reduce: BitwiseShiftTerm
			reduce(178), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(178), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(178), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(178), // ||, reduce: BitwiseShiftTerm
			reduce(178), // ^^, reduce: BitwiseShiftTerm
			reduce(178), // &&, reduce: BitwiseShiftTerm
			reduce(178), // ??, reduce: BitwiseShiftTerm
			reduce(178), // ???, reduce: BitwiseShiftTerm
			reduce(178), // =~, reduce: BitwiseShiftTerm
			reduce(178), // !=~, reduce: BitwiseShiftTerm
			reduce(178), // ==, reduce: BitwiseShiftTerm
			reduce(178), // !=, reduce: BitwiseShiftTerm
			reduce(178), // <=>, reduce: BitwiseShiftTerm
			reduce(178), // >=, reduce: BitwiseShiftTerm
			reduce(178), // <, reduce: BitwiseShiftTerm
			reduce(178), // <=, reduce: BitwiseShiftTerm
			reduce(178), // ^, reduce: BitwiseShiftTerm
			reduce(178), // &, reduce: BitwiseShiftTerm
			reduce(178), // <<, reduce: BitwiseShiftTerm
			reduce(178), // >>>, reduce: BitwiseShiftTerm
			shift(3030), // +
			shift(3031), // -
			shift(3032), // .+
			shift(3033), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3810
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(180), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			reduce(180), // }, reduce: BitwiseShiftTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(180), // >, reduce: BitwiseShiftTerm
			reduce(180), // >>, reduce: BitwiseShiftTerm
			reduce(180), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(180), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(180), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(180), // ||, reduce: BitwiseShiftTerm
			reduce(180), // ^^, reduce: BitwiseShiftTerm
			reduce(180), // &&, reduce: BitwiseShiftTerm
			reduce(180), // ??, reduce: BitwiseShiftTerm
			reduce(180), // ???, reduce: BitwiseShiftTerm
			reduce(180), // =~, reduce: BitwiseShiftTerm
			reduce(180), // !=~, reduce: BitwiseShiftTerm
			reduce(180), // ==, reduce: BitwiseShiftTerm
			reduce(180), // !=, reduce: BitwiseShiftTerm
			reduce(180), // <=>, reduce: BitwiseShiftTerm
			reduce(180), // >=, reduce: BitwiseShiftTerm
			reduce(180), // <, reduce: BitwiseShiftTerm
			reduce(180), // <=, reduce: BitwiseShiftTerm
			reduce(180), // ^, reduce: BitwiseShiftTerm
			reduce(180), // &, reduce: BitwiseShiftTerm
			reduce(180), // <<, reduce: BitwiseShiftTerm
			reduce(180), // >>>, reduce: BitwiseShiftTerm
			shift(3030), // +
			shift(3031), // -
			shift(3032), // .+
			shift(3033), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3811
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(182), // ;, reduce: AddsubdotTerm
			nil,         // {
			reduce(182), // }, reduce: AddsubdotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(182), // >, reduce: AddsubdotTerm
			reduce(182), // >>, reduce: AddsubdotTerm
			reduce(182), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(182), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(182), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(182), // ||, reduce: AddsubdotTerm
			reduce(182), // ^^, reduce: AddsubdotTerm
			reduce(182), // &&, reduce: AddsubdotTerm
			reduce(182), // ??, reduce: AddsubdotTerm
			reduce(182), // ???, reduce: AddsubdotTerm
			reduce(182), // =~, reduce: AddsubdotTerm
			reduce(182), // !=~, reduce: AddsubdotTerm
			reduce(182), // ==, reduce: AddsubdotTerm
			reduce(182), // !=, reduce: AddsubdotTerm
			reduce(182), // <=>, reduce: AddsubdotTerm
			reduce(182), // >=, reduce: AddsubdotTerm
			reduce(182), // <, reduce: AddsubdotTerm
			reduce(182), // <=, reduce: AddsubdotTerm
			reduce(182), // ^, reduce: AddsubdotTerm
			reduce(182), // &, reduce: AddsubdotTerm
			reduce(182), // <<, reduce: AddsubdotTerm
			reduce(182), // >>>, reduce: AddsubdotTerm
			reduce(182), // +, reduce: AddsubdotTerm
			reduce(182), // -, reduce: AddsubdotTerm
			reduce(182), // .+, reduce: AddsubdotTerm
			reduce(182), // .-, reduce: AddsubdotTerm
			shift(3035), // *
			shift(3036), // /
			shift(3037), // //
			shift(3038), // %
			shift(3039), // .*
			shift(3040), // ./
			shift(3041), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3812
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(183), // ;, reduce: AddsubdotTerm
			nil,         // {
			reduce(183), // }, reduce: AddsubdotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(183), // >, reduce: AddsubdotTerm
			reduce(183), // >>, reduce: AddsubdotTerm
			reduce(183), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(183), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(183), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(183), // ||, reduce: AddsubdotTerm
			reduce(183), // ^^, reduce: AddsubdotTerm
			reduce(183), // &&, reduce: AddsubdotTerm
			reduce(183), // ??, reduce: AddsubdotTerm
			reduce(183), // ???, reduce: AddsubdotTerm
			reduce(183), // =~, reduce: AddsubdotTerm
			reduce(183), // !=~, reduce: AddsubdotTerm
			reduce(183), // ==, reduce: AddsubdotTerm
			reduce(183), // !=, reduce: AddsubdotTerm
			reduce(183), // <=>, reduce: AddsubdotTerm
			reduce(183), // >=, reduce: AddsubdotTerm
			reduce(183), // <, reduce: AddsubdotTerm
			reduce(183), // <=, reduce: AddsubdotTerm
			reduce(183), // ^, reduce: AddsubdotTerm
			reduce(183), // &, reduce: AddsubdotTerm
			reduce(183), // <<, reduce: AddsubdotTerm
			reduce(183), // >>>, reduce: AddsubdotTerm
			reduce(183), // +, reduce: AddsubdotTerm
			reduce(183), // -, reduce: AddsubdotTerm
			reduce(183), // .+, reduce: AddsubdotTerm
			reduce(183), // .-, reduce: AddsubdotTerm
			shift(3035), // *
			shift(3036), // /
			shift(3037), // //
			shift(3038), // %
			shift(3039), // .*
			shift(3040), // ./
			shift(3041), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3813
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(184), // ;, reduce: AddsubdotTerm
			nil,         // {
			reduce(184), // }, reduce: AddsubdotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(184), // >, reduce: AddsubdotTerm
			reduce(184), // >>, reduce: AddsubdotTerm
			reduce(184), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(184), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(184), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(184), // ||, reduce: AddsubdotTerm
			reduce(184), // ^^, reduce: AddsubdotTerm
			reduce(184), // &&, reduce: AddsubdotTerm
			reduce(184), // ??, reduce: AddsubdotTerm
			reduce(184), // ???, reduce: AddsubdotTerm
			reduce(184), // =~, reduce: AddsubdotTerm
			reduce(184), // !=~, reduce: AddsubdotTerm
			reduce(184), // ==, reduce: AddsubdotTerm
			reduce(184), // !=, reduce: AddsubdotTerm
			reduce(184), // <=>, reduce: AddsubdotTerm
			reduce(184), // >=, reduce: AddsubdotTerm
			reduce(184), // <, reduce: AddsubdotTerm
			reduce(184), // <=, reduce: AddsubdotTerm
			reduce(184), // ^, reduce: AddsubdotTerm
			reduce(184), // &, reduce: AddsubdotTerm
			reduce(184), // <<, reduce: AddsubdotTerm
			reduce(184), // >>>, reduce: AddsubdotTerm
			reduce(184), // +, reduce: AddsubdotTerm
			reduce(184), // -, reduce: AddsubdotTerm
			reduce(184), // .+, reduce: AddsubdotTerm
			reduce(184), // .-, reduce: AddsubdotTerm
			shift(3035), // *
			shift(3036), // /
			shift(3037), // //
			shift(3038), // %
			shift(3039), // .*
			shift(3040), // ./
			shift(3041), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3814
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(185), // ;, reduce: AddsubdotTerm
			nil,         // {
			reduce(185), // }, reduce: AddsubdotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(185), // >, reduce: AddsubdotTerm
			reduce(185), // >>, reduce: AddsubdotTerm
			reduce(185), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(185), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(185), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(185), // ||, reduce: AddsubdotTerm
			reduce(185), // ^^, reduce: AddsubdotTerm
			reduce(185), // &&, reduce: AddsubdotTerm
			reduce(185), // ??, reduce: AddsubdotTerm
			reduce(185), // ???, reduce: AddsubdotTerm
			reduce(185), // =~, reduce: AddsubdotTerm
			reduce(185), // !=~, reduce: AddsubdotTerm
			reduce(185), // ==, reduce: AddsubdotTerm
			reduce(185), // !=, reduce: AddsubdotTerm
			reduce(185), // <=>, reduce: AddsubdotTerm
			reduce(185), // >=, reduce: AddsubdotTerm
			reduce(185), // <, reduce: AddsubdotTerm
			reduce(185), // <=, reduce: AddsubdotTerm
			reduce(185), // ^, reduce: AddsubdotTerm
			reduce(185), // &, reduce: AddsubdotTerm
			reduce(185), // <<, reduce: AddsubdotTerm
			reduce(185), // >>>, reduce: AddsubdotTerm
			reduce(185), // +, reduce: AddsubdotTerm
			reduce(185), // -, reduce: AddsubdotTerm
			reduce(185), // .+, reduce: AddsubdotTerm
			reduce(185), // .-, reduce: AddsubdotTerm
			shift(3035), // *
			shift(3036), // /
			shift(3037), // //
			shift(3038), // %
			shift(3039), // .*
			shift(3040), // ./
			shift(3041), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3815
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(187), // ;, reduce: MuldivTerm
			nil,         // {
			reduce(187), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(187), // >, reduce: MuldivTerm
			reduce(187), // >>, reduce: MuldivTerm
			reduce(187), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(187), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(187), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(187), // ||, reduce: MuldivTerm
			reduce(187), // ^^, reduce: MuldivTerm
			reduce(187), // &&, reduce: MuldivTerm
			reduce(187), // ??, reduce: MuldivTerm
			reduce(187), // ???, reduce: MuldivTerm
			reduce(187), // =~, reduce: MuldivTerm
			reduce(187), // !=~, reduce: MuldivTerm
			reduce(187), // ==, reduce: MuldivTerm
			reduce(187), // !=, reduce: MuldivTerm
			reduce(187), // <=>, reduce: MuldivTerm
			reduce(187), // >=, reduce: MuldivTerm
			reduce(187), // <, reduce: MuldivTerm
			reduce(187), // <=, reduce: MuldivTerm
			reduce(187), // ^, reduce: MuldivTerm
			reduce(187), // &, reduce: MuldivTerm
			reduce(187), // <<, reduce: MuldivTerm
			reduce(187), // >>>, reduce: MuldivTerm
			reduce(187), // +, reduce: MuldivTerm
			reduce(187), // -, reduce: MuldivTerm
			reduce(187), // .+, reduce: MuldivTerm
			reduce(187), // .-, reduce: MuldivTerm
			reduce(187), // *, reduce: MuldivTerm
			reduce(187), // /, reduce: MuldivTerm
			reduce(187), // //, reduce: MuldivTerm
			reduce(187), // %, reduce: MuldivTerm
			reduce(187), // .*, reduce: MuldivTerm
			reduce(187), // ./, reduce: MuldivTerm
			reduce(187), // .//, reduce: MuldivTerm
			shift(3045), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3816
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(188), // ;, reduce: MuldivTerm
			nil,         // {
			reduce(188), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(188), // >, reduce: MuldivTerm
			reduce(188), // >>, reduce: MuldivTerm
			reduce(188), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(188), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(188), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(188), // ||, reduce: MuldivTerm
			reduce(188), // ^^, reduce: MuldivTerm
			reduce(188), // &&, reduce: MuldivTerm
			reduce(188), // ??, reduce: MuldivTerm
			reduce(188), // ???, reduce: MuldivTerm
			reduce(188), // =~, reduce: MuldivTerm
			reduce(188), // !=~, reduce: MuldivTerm
			reduce(188), // ==, reduce: MuldivTerm
			reduce(188), // !=, reduce: MuldivTerm
			reduce(188), // <=>, reduce: MuldivTerm
			reduce(188), // >=, reduce: MuldivTerm
			reduce(188), // <, reduce: MuldivTerm
			reduce(188), // <=, reduce: MuldivTerm
			reduce(188), // ^, reduce: MuldivTerm
			reduce(188), // &, reduce: MuldivTerm
			reduce(188), // <<, reduce: MuldivTerm
			reduce(188), // >>>, reduce: MuldivTerm
			reduce(188), // +, reduce: MuldivTerm
			reduce(188), // -, reduce: MuldivTerm
			reduce(188), // .+, reduce: MuldivTerm
			reduce(188), // .-, reduce: MuldivTerm
			reduce(188), // *, reduce: MuldivTerm
			reduce(188), // /, reduce: MuldivTerm
			reduce(188), // //, reduce: MuldivTerm
			reduce(188), // %, reduce: MuldivTerm
			reduce(188), // .*, reduce: MuldivTerm
			reduce(188), // ./, reduce: MuldivTerm
			reduce(188), // .//, reduce: MuldivTerm
			shift(3045), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3817
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(189), // ;, reduce: MuldivTerm
			nil,         // {
			reduce(189), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(189), // >, reduce: MuldivTerm
			reduce(189), // >>, reduce: MuldivTerm
			reduce(189), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(189), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(189), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(189), // ||, reduce: MuldivTerm
			reduce(189), // ^^, reduce: MuldivTerm
			reduce(189), // &&, reduce: MuldivTerm
			reduce(189), // ??, reduce: MuldivTerm
			reduce(189), // ???, reduce: MuldivTerm
			reduce(189), // =~, reduce: MuldivTerm
			reduce(189), // !=~, reduce: MuldivTerm
			reduce(189), // ==, reduce: MuldivTerm
			reduce(189), // !=, reduce: MuldivTerm
			reduce(189), // <=>, reduce: MuldivTerm
			reduce(189), // >=, reduce: MuldivTerm
			reduce(189), // <, reduce: MuldivTerm
			reduce(189), // <=, reduce: MuldivTerm
			reduce(189), // ^, reduce: MuldivTerm
			reduce(189), // &, reduce: MuldivTerm
			reduce(189), // <<, reduce: MuldivTerm
			reduce(189), // >>>, reduce: MuldivTerm
			reduce(189), // +, reduce: MuldivTerm
			reduce(189), // -, reduce: MuldivTerm
			reduce(189), // .+, reduce: MuldivTerm
			reduce(189), // .-, reduce: MuldivTerm
			reduce(189), // *, reduce: MuldivTerm
			reduce(189), // /, reduce: MuldivTerm
			reduce(189), // //, reduce: MuldivTerm
			reduce(189), // %, reduce: MuldivTerm
			reduce(189), // .*, reduce: MuldivTerm
			reduce(189), // ./, reduce: MuldivTerm
			reduce(189), // .//, reduce: MuldivTerm
			shift(3045), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3818
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(190), // ;, reduce: MuldivTerm
			nil,         // {
			reduce(190), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(190), // >, reduce: MuldivTerm
			reduce(190), // >>, reduce: MuldivTerm
			reduce(190), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(190), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(190), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(190), // ||, reduce: MuldivTerm
			reduce(190), // ^^, reduce: MuldivTerm
			reduce(190), // &&, reduce: MuldivTerm
			reduce(190), // ??, reduce: MuldivTerm
			reduce(190), // ???, reduce: MuldivTerm
			reduce(190), // =~, reduce: MuldivTerm
			reduce(190), // !=~, reduce: MuldivTerm
			reduce(190), // ==, reduce: MuldivTerm
			reduce(190), // !=, reduce: MuldivTerm
			reduce(190), // <=>, reduce: MuldivTerm
			reduce(190), // >=, reduce: MuldivTerm
			reduce(190), // <, reduce: MuldivTerm
			reduce(190), // <=, reduce: MuldivTerm
			reduce(190), // ^, reduce: MuldivTerm
			reduce(190), // &, reduce: MuldivTerm
			reduce(190), // <<, reduce: MuldivTerm
			reduce(190), // >>>, reduce: MuldivTerm
			reduce(190), // +, reduce: MuldivTerm
			reduce(190), // -, reduce: MuldivTerm
			reduce(190), // .+, reduce: MuldivTerm
			reduce(190), // .-, reduce: MuldivTerm
			reduce(190), // *, reduce: MuldivTerm
			reduce(190), // /, reduce: MuldivTerm
			reduce(190), // //, reduce: MuldivTerm
			reduce(190), // %, reduce: MuldivTerm
			reduce(190), // .*, reduce: MuldivTerm
			reduce(190), // ./, reduce: MuldivTerm
			reduce(190), // .//, reduce: MuldivTerm
			shift(3045), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3819
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(191), // ;, reduce: MuldivTerm
			nil,         // {
			reduce(191), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(191), // >, reduce: MuldivTerm
			reduce(191), // >>, reduce: MuldivTerm
			reduce(191), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(191), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(191), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(191), // ||, reduce: MuldivTerm
			reduce(191), // ^^, reduce: MuldivTerm
			reduce(191), // &&, reduce: MuldivTerm
			reduce(191), // ??, reduce: MuldivTerm
			reduce(191), // ???, reduce: MuldivTerm
			reduce(191), // =~, reduce: MuldivTerm
			reduce(191), // !=~, reduce: MuldivTerm
			reduce(191), // ==, reduce: MuldivTerm
			reduce(191), // !=, reduce: MuldivTerm
			reduce(191), // <=>, reduce: MuldivTerm
			reduce(191), // >=, reduce: MuldivTerm
			reduce(191), // <, reduce: MuldivTerm
			reduce(191), // <=, reduce: MuldivTerm
			reduce(191), // ^, reduce: MuldivTerm
			reduce(191), // &, reduce: MuldivTerm
			reduce(191), // <<, reduce: MuldivTerm
			reduce(191), // >>>, reduce: MuldivTerm
			reduce(191), // +, reduce: MuldivTerm
			reduce(191), // -, reduce: MuldivTerm
			reduce(191), // .+, reduce: MuldivTerm
			reduce(191), // .-, reduce: MuldivTerm
			reduce(191), // *, reduce: MuldivTerm
			reduce(191), // /, reduce: MuldivTerm
			reduce(191), // //, reduce: MuldivTerm
			reduce(191), // %, reduce: MuldivTerm
			reduce(191), // .*, reduce: MuldivTerm
			reduce(191), // ./, reduce: MuldivTerm
			reduce(191), // .//, reduce: MuldivTerm
			shift(3045), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3820
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(192), // ;, reduce: MuldivTerm
			nil,         // {
			reduce(192), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(192), // >, reduce: MuldivTerm
			reduce(192), // >>, reduce: MuldivTerm
			reduce(192), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(192), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(192), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(192), // ||, reduce: MuldivTerm
			reduce(192), // ^^, reduce: MuldivTerm
			reduce(192), // &&, reduce: MuldivTerm
			reduce(192), // ??, reduce: MuldivTerm
			reduce(192), // ???, reduce: MuldivTerm
			reduce(192), // =~, reduce: MuldivTerm
			reduce(192), // !=~, reduce: MuldivTerm
			reduce(192), // ==, reduce: MuldivTerm
			reduce(192), // !=, reduce: MuldivTerm
			reduce(192), // <=>, reduce: MuldivTerm
			reduce(192), // >=, reduce: MuldivTerm
			reduce(192), // <, reduce: MuldivTerm
			reduce(192), // <=, reduce: MuldivTerm
			reduce(192), // ^, reduce: MuldivTerm
			reduce(192), // &, reduce: MuldivTerm
			reduce(192), // <<, reduce: MuldivTerm
			reduce(192), // >>>, reduce: MuldivTerm
			reduce(192), // +, reduce: MuldivTerm
			reduce(192), // -, reduce: MuldivTerm
			reduce(192), // .+, reduce: MuldivTerm
			reduce(192), // .-, reduce: MuldivTerm
			reduce(192), // *, reduce: MuldivTerm
			reduce(192), // /, reduce: MuldivTerm
			reduce(192), // //, reduce: MuldivTerm
			reduce(192), // %, reduce: MuldivTerm
			reduce(192), // .*, reduce: MuldivTerm
			reduce(192), // ./, reduce: MuldivTerm
			reduce(192), // .//, reduce: MuldivTerm
			shift(3045), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3821
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(193), // ;, reduce: MuldivTerm
			nil,         // {
			reduce(193), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(193), // >, reduce: MuldivTerm
			reduce(193), // >>, reduce: MuldivTerm
			reduce(193), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(193), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(193), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(193), // ||, reduce: MuldivTerm
			reduce(193), // ^^, reduce: MuldivTerm
			reduce(193), // &&, reduce: MuldivTerm
			reduce(193), // ??, reduce: MuldivTerm
			reduce(193), // ???, reduce: MuldivTerm
			reduce(193), // =~, reduce: MuldivTerm
			reduce(193), // !=~, reduce: MuldivTerm
			reduce(193), // ==, reduce: MuldivTerm
			reduce(193), // !=, reduce: MuldivTerm
			reduce(193), // <=>, reduce: MuldivTerm
			reduce(193), // >=, reduce: MuldivTerm
			reduce(193), // <, reduce: MuldivTerm
			reduce(193), // <=, reduce: MuldivTerm
			reduce(193), // ^, reduce: MuldivTerm
			reduce(193), // &, reduce: MuldivTerm
			reduce(193), // <<, reduce: MuldivTerm
			reduce(193), // >>>, reduce: MuldivTerm
			reduce(193), // +, reduce: MuldivTerm
			reduce(193), // -, reduce: MuldivTerm
			reduce(193), // .+, reduce: MuldivTerm
			reduce(193), // .-, reduce: MuldivTerm
			reduce(193), // *, reduce: MuldivTerm
			reduce(193), // /, reduce: MuldivTerm
			reduce(193), // //, reduce: MuldivTerm
			reduce(193), // %, reduce: MuldivTerm
			reduce(193), // .*, reduce: MuldivTerm
			reduce(193), // ./, reduce: MuldivTerm
			reduce(193), // .//, reduce: MuldivTerm
			shift(3045), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3822
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(195), // ;, reduce: DotTerm
			nil,         // {
			reduce(195), // }, reduce: DotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(195), // >, reduce: DotTerm
			reduce(195), // >>, reduce: DotTerm
			reduce(195), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(195), // ,, reduce: DotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(195), // ?, reduce: DotTerm
			nil,         // :
			reduce(195), // ||, reduce: DotTerm
			reduce(195), // ^^, reduce: DotTerm
			reduce(195), // &&, reduce: DotTerm
			reduce(195), // ??, reduce: DotTerm
			reduce(195), // ???, reduce: DotTerm
			reduce(195), // =~, reduce: DotTerm
			reduce(195), // !=~, reduce: DotTerm
			reduce(195), // ==, reduce: DotTerm
			reduce(195), // !=, reduce: DotTerm
			reduce(195), // <=>, reduce: DotTerm
			reduce(195), // >=, reduce: DotTerm
			reduce(195), // <, reduce: DotTerm
			reduce(195), // <=, reduce: DotTerm
			reduce(195), // ^, reduce: DotTerm
			reduce(195), // &, reduce: DotTerm
			reduce(195), // <<, reduce: DotTerm
			reduce(195), // >>>, reduce: DotTerm
			reduce(195), // +, reduce: DotTerm
			reduce(195), // -, reduce: DotTerm
			reduce(195), // .+, reduce: DotTerm
			reduce(195), // .-, reduce: DotTerm
			reduce(195), // *, reduce: DotTerm
			reduce(195), // /, reduce: DotTerm
			reduce(195), // //, reduce: DotTerm
			reduce(195), // %, reduce: DotTerm
			reduce(195), // .*, reduce: DotTerm
			reduce(195), // ./, reduce: DotTerm
			reduce(195), // .//, reduce: DotTerm
			reduce(195), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3823
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3824
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3825
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(204), // ;, reduce: PowTerm
			nil,         // {
			reduce(204), // }, reduce: PowTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(204), // >, reduce: PowTerm
			reduce(204), // >>, reduce: PowTerm
			reduce(204), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(204), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(204), // ?, reduce: PowTerm
			nil,         // :
			reduce(204), // ||, reduce: PowTerm
			reduce(204), // ^^, reduce: PowTerm
			reduce(204), // &&, reduce: PowTerm
			reduce(204), // ??, reduce: PowTerm
			reduce(204), // ???, reduce: PowTerm
			reduce(204), // =~, reduce: PowTerm
			reduce(204), // !=~, reduce: PowTerm
			reduce(204), // ==, reduce: PowTerm
			reduce(204), // !=, reduce: PowTerm
			reduce(204), // <=>, reduce: PowTerm
			reduce(204), // >=, reduce: PowTerm
			reduce(204), // <, reduce: PowTerm
			reduce(204), // <=, reduce: PowTerm
			reduce(204), // ^, reduce: PowTerm
			reduce(204), // &, reduce: PowTerm
			reduce(204), // <<, reduce: PowTerm
			reduce(204), // >>>, reduce: PowTerm
			reduce(204), // +, reduce: PowTerm
			reduce(204), // -, reduce: PowTerm
			reduce(204), // .+, reduce: PowTerm
			reduce(204), // .-, reduce: PowTerm
			reduce(204), // *, reduce: PowTerm
			reduce(204), // /, reduce: PowTerm
			reduce(204), // //, reduce: PowTerm
			reduce(204), // %, reduce: PowTerm
			reduce(204), // .*, reduce: PowTerm
			reduce(204), // ./, reduce: PowTerm
			reduce(204), // .//, reduce: PowTerm
			reduce(204), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3826
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4207), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4208), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3827
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4210), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3828
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4211), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3829
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4212), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3830
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(229), // ;, reduce: ArrayLiteral
			nil,         // {
			reduce(229), // }, reduce: ArrayLiteral
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(229), // >, reduce: ArrayLiteral
			reduce(229), // >>, reduce: ArrayLiteral
			reduce(229), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(229), // ,, reduce: ArrayLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(229), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(229), // ||, reduce: ArrayLiteral
			reduce(229), // ^^, reduce: ArrayLiteral
			reduce(229), // &&, reduce: ArrayLiteral
			reduce(229), // ??, reduce: ArrayLiteral
			reduce(229), // ???, reduce: ArrayLiteral
			reduce(229), // =~, reduce: ArrayLiteral
			reduce(229), // !=~, reduce: ArrayLiteral
			reduce(229), // ==, reduce: ArrayLiteral
			reduce(229), // !=, reduce: ArrayLiteral
			reduce(229), // <=>, reduce: ArrayLiteral
			reduce(229), // >=, reduce: ArrayLiteral
			reduce(229), // <, reduce: ArrayLiteral
			reduce(229), // <=, reduce: ArrayLiteral
			reduce(229), // ^, reduce: ArrayLiteral
			reduce(229), // &, reduce: ArrayLiteral
			reduce(229), // <<, reduce: ArrayLiteral
			reduce(229), // >>>, reduce: ArrayLiteral
			reduce(229), // +, reduce: ArrayLiteral
			reduce(229), // -, reduce: ArrayLiteral
			reduce(229), // .+, reduce: ArrayLiteral
			reduce(229), // .-, reduce: ArrayLiteral
			reduce(229), // *, reduce: ArrayLiteral
			reduce(229), // /, reduce: ArrayLiteral
			reduce(229), // //, reduce: ArrayLiteral
			reduce(229), // %, reduce: ArrayLiteral
			reduce(229), // .*, reduce: ArrayLiteral
			reduce(229), // ./, reduce: ArrayLiteral
			reduce(229), // .//, reduce: ArrayLiteral
			reduce(229), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(229), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(229), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(229), // [[, reduce: ArrayLiteral
			reduce(229), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3831
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(255), // ;, reduce: ENV
			nil,         // {
			reduce(255), // }, reduce: ENV
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(255), // >, reduce: ENV
			reduce(255), // >>, reduce: ENV
			reduce(255), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(255), // ,, reduce: ENV
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(255), // ?, reduce: ENV
			nil,         // :
			reduce(255), // ||, reduce: ENV
			reduce(255), // ^^, reduce: ENV
			reduce(255), // &&, reduce: ENV
			reduce(255), // ??, reduce: ENV
			reduce(255), // ???, reduce: ENV
			reduce(255), // =~, reduce: ENV
			reduce(255), // !=~, reduce: ENV
			reduce(255), // ==, reduce: ENV
			reduce(255), // !=, reduce: ENV
			reduce(255), // <=>, reduce: ENV
			reduce(255), // >=, reduce: ENV
			reduce(255), // <, reduce: ENV
			reduce(255), // <=, reduce: ENV
			reduce(255), // ^, reduce: ENV
			reduce(255), // &, reduce: ENV
			reduce(255), // <<, reduce: ENV
			reduce(255), // >>>, reduce: ENV
			reduce(255), // +, reduce: ENV
			reduce(255), // -, reduce: ENV
			reduce(255), // .+, reduce: ENV
			reduce(255), // .-, reduce: ENV
			reduce(255), // *, reduce: ENV
			reduce(255), // /, reduce: ENV
			reduce(255), // //, reduce: ENV
			reduce(255), // %, reduce: ENV
			reduce(255), // .*, reduce: ENV
			reduce(255), // ./, reduce: ENV
			reduce(255), // .//, reduce: ENV
			reduce(255), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(255), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(255), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(255), // [[, reduce: ENV
			reduce(255), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3832
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4213), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3833
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4214), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3834
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(268), // ;, reduce: FunctionCallsite
			nil,         // {
			reduce(268), // }, reduce: FunctionCallsite
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(268), // >, reduce: FunctionCallsite
			reduce(268), // >>, reduce: FunctionCallsite
			reduce(268), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(268), // ,, reduce: FunctionCallsite
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(268), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(268), // ||, reduce: FunctionCallsite
			reduce(268), // ^^, reduce: FunctionCallsite
			reduce(268), // &&, reduce: FunctionCallsite
			reduce(268), // ??, reduce: FunctionCallsite
			reduce(268), // ???, reduce: FunctionCallsite
			reduce(268), // =~, reduce: FunctionCallsite
			reduce(268), // !=~, reduce: FunctionCallsite
			reduce(268), // ==, reduce: FunctionCallsite
			reduce(268), // !=, reduce: FunctionCallsite
			reduce(268), // <=>, reduce: FunctionCallsite
			reduce(268), // >=, reduce: FunctionCallsite
			reduce(268), // <, reduce: FunctionCallsite
			reduce(268), // <=, reduce: FunctionCallsite
			reduce(268), // ^, reduce: FunctionCallsite
			reduce(268), // &, reduce: FunctionCallsite
			reduce(268), // <<, reduce: FunctionCallsite
			reduce(268), // >>>, reduce: FunctionCallsite
			reduce(268), // +, reduce: FunctionCallsite
			reduce(268), // -, reduce: FunctionCallsite
			reduce(268), // .+, reduce: FunctionCallsite
			reduce(268), // .-, reduce: FunctionCallsite
			reduce(268), // *, reduce: FunctionCallsite
			reduce(268), // /, reduce: FunctionCallsite
			reduce(268), // //, reduce: FunctionCallsite
			reduce(268), // %, reduce: FunctionCallsite
			reduce(268), // .*, reduce: FunctionCallsite
			reduce(268), // ./, reduce: FunctionCallsite
			reduce(268), // .//, reduce: FunctionCallsite
			reduce(268), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(268), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(268), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(268), // [[, reduce: FunctionCallsite
			reduce(268), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3835
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4215), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3836
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(235), // ;, reduce: MapLiteral
			nil,         // {
			reduce(235), // }, reduce: MapLiteral
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(235), // >, reduce: MapLiteral
			reduce(235), // >>, reduce: MapLiteral
			reduce(235), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(235), // ?, reduce: MapLiteral
			nil,         // :
			reduce(235), // ||, reduce: MapLiteral
			reduce(235), // ^^, reduce: MapLiteral
			reduce(235), // &&, reduce: MapLiteral
			reduce(235), // ??, reduce: MapLiteral
			reduce(235), // ???, reduce: MapLiteral
			reduce(235), // =~, reduce: MapLiteral
			reduce(235), // !=~, reduce: MapLiteral
			reduce(235), // ==, reduce: MapLiteral
			reduce(235), // !=, reduce: MapLiteral
			reduce(235), // <=>, reduce: MapLiteral
			reduce(235), // >=, reduce: MapLiteral
			reduce(235), // <, reduce: MapLiteral
			reduce(235), // <=, reduce: MapLiteral
			reduce(235), // ^, reduce: MapLiteral
			reduce(235), // &, reduce: MapLiteral
			reduce(235), // <<, reduce: MapLiteral
			reduce(235), // >>>, reduce: MapLiteral
			reduce(235), // +, reduce: MapLiteral
			reduce(235), // -, reduce: MapLiteral
			reduce(235), // .+, reduce: MapLiteral
			reduce(235), // .-, reduce: MapLiteral
			reduce(235), // *, reduce: MapLiteral
			reduce(235), // /, reduce: MapLiteral
			reduce(235), // //, reduce: MapLiteral
			reduce(235), // %, reduce: MapLiteral
			reduce(235), // .*, reduce: MapLiteral
			reduce(235), // ./, reduce: MapLiteral
			reduce(235), // .//, reduce: MapLiteral
			reduce(235), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(235), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(235), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(235), // [[, reduce: MapLiteral
			reduce(235), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3837
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(208), // ;, reduce: PrecedenceChainEnd
			nil,         // {
			reduce(208), // }, reduce: PrecedenceChainEnd
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(208), // >, reduce: PrecedenceChainEnd
			reduce(208), // >>, reduce: PrecedenceChainEnd
			reduce(208), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(208), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(208), // ||, reduce: PrecedenceChainEnd
			reduce(208), // ^^, reduce: PrecedenceChainEnd
			reduce(208), // &&, reduce: PrecedenceChainEnd
			reduce(208), // ??, reduce: PrecedenceChainEnd
			reduce(208), // ???, reduce: PrecedenceChainEnd
			reduce(208), // =~, reduce: PrecedenceChainEnd
			reduce(208), // !=~, reduce: PrecedenceChainEnd
			reduce(208), // ==, reduce: PrecedenceChainEnd
			reduce(208), // !=, reduce: PrecedenceChainEnd
			reduce(208), // <=>, reduce: PrecedenceChainEnd
			reduce(208), // >=, reduce: PrecedenceChainEnd
			reduce(208), // <, reduce: PrecedenceChainEnd
			reduce(208), // <=, reduce: PrecedenceChainEnd
			reduce(208), // ^, reduce: PrecedenceChainEnd
			reduce(208), // &, reduce: PrecedenceChainEnd
			reduce(208), // <<, reduce: PrecedenceChainEnd
			reduce(208), // >>>, reduce: PrecedenceChainEnd
			reduce(208), // +, reduce: PrecedenceChainEnd
			reduce(208), // -, reduce: PrecedenceChainEnd
			reduce(208), // .+, reduce: PrecedenceChainEnd
			reduce(208), // .-, reduce: PrecedenceChainEnd
			reduce(208), // *, reduce: PrecedenceChainEnd
			reduce(208), // /, reduce: PrecedenceChainEnd
			reduce(208), // //, reduce: PrecedenceChainEnd
			reduce(208), // %, reduce: PrecedenceChainEnd
			reduce(208), // .*, reduce: PrecedenceChainEnd
			reduce(208), // ./, reduce: PrecedenceChainEnd
			reduce(208), // .//, reduce: PrecedenceChainEnd
			reduce(208), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(208), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3838
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(105), // ;, reduce: IndirectFieldValue
			nil,         // {
			reduce(105), // }, reduce: IndirectFieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(105), // >, reduce: IndirectFieldValue
			reduce(105), // >>, reduce: IndirectFieldValue
			reduce(105), // |, reduce: IndirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(105), // ?, reduce: IndirectFieldValue
			nil,         // :
			reduce(105), // ||, reduce: IndirectFieldValue
			reduce(105), // ^^, reduce: IndirectFieldValue
			reduce(105), // &&, reduce: IndirectFieldValue
			reduce(105), // ??, reduce: IndirectFieldValue
			reduce(105), // ???, reduce: IndirectFieldValue
			reduce(105), // =~, reduce: IndirectFieldValue
			reduce(105), // !=~, reduce: IndirectFieldValue
			reduce(105), // ==, reduce: IndirectFieldValue
			reduce(105), // !=, reduce: IndirectFieldValue
			reduce(105), // <=>, reduce: IndirectFieldValue
			reduce(105), // >=, reduce: IndirectFieldValue
			reduce(105), // <, reduce: IndirectFieldValue
			reduce(105), // <=, reduce: IndirectFieldValue
			reduce(105), // ^, reduce: IndirectFieldValue
			reduce(105), // &, reduce: IndirectFieldValue
			reduce(105), // <<, reduce: IndirectFieldValue
			reduce(105), // >>>, reduce: IndirectFieldValue
			reduce(105), // +, reduce: IndirectFieldValue
			reduce(105), // -, reduce: IndirectFieldValue
			reduce(105), // .+, reduce: IndirectFieldValue
			reduce(105), // .-, reduce: IndirectFieldValue
			reduce(105), // *, reduce: IndirectFieldValue
			reduce(105), // /, reduce: IndirectFieldValue
			reduce(105), // //, reduce: IndirectFieldValue
			reduce(105), // %, reduce: IndirectFieldValue
			reduce(105), // .*, reduce: IndirectFieldValue
			reduce(105), // ./, reduce: IndirectFieldValue
			reduce(105), // .//, reduce: IndirectFieldValue
			reduce(105), // ., reduce: IndirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(105), // **, reduce: IndirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(105), // [, reduce: IndirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(105), // [[, reduce: IndirectFieldValue
			reduce(105), // [[[, reduce: IndirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3839
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4216), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3840
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4217), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3841
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(114), // ;, reduce: IndirectOosvarValue
			nil,         // {
			reduce(114), // }, reduce: IndirectOosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(114), // >, reduce: IndirectOosvarValue
			reduce(114), // >>, reduce: IndirectOosvarValue
			reduce(114), // |, reduce: IndirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(114), // ?, reduce: IndirectOosvarValue
			nil,         // :
			reduce(114), // ||, reduce: IndirectOosvarValue
			reduce(114), // ^^, reduce: IndirectOosvarValue
			reduce(114), // &&, reduce: IndirectOosvarValue
			reduce(114), // ??, reduce: IndirectOosvarValue
			reduce(114), // ???, reduce: IndirectOosvarValue
			reduce(114), // =~, reduce: IndirectOosvarValue
			reduce(114), // !=~, reduce: IndirectOosvarValue
			reduce(114), // ==, reduce: IndirectOosvarValue
			reduce(114), // !=, reduce: IndirectOosvarValue
			reduce(114), // <=>, reduce: IndirectOosvarValue
			reduce(114), // >=, reduce: IndirectOosvarValue
			reduce(114), // <, reduce: IndirectOosvarValue
			reduce(114), // <=, reduce: IndirectOosvarValue
			reduce(114), // ^, reduce: IndirectOosvarValue
			reduce(114), // &, reduce: IndirectOosvarValue
			reduce(114), // <<, reduce: IndirectOosvarValue
			reduce(114), // >>>, reduce: IndirectOosvarValue
			reduce(114), // +, reduce: IndirectOosvarValue
			reduce(114), // -, reduce: IndirectOosvarValue
			reduce(114), // .+, reduce: IndirectOosvarValue
			reduce(114), // .-, reduce: IndirectOosvarValue
			reduce(114), // *, reduce: IndirectOosvarValue
			reduce(114), // /, reduce: IndirectOosvarValue
			reduce(114), // //, reduce: IndirectOosvarValue
			reduce(114), // %, reduce: IndirectOosvarValue
			reduce(114), // .*, reduce: IndirectOosvarValue
			reduce(114), // ./, reduce: IndirectOosvarValue
			reduce(114), // .//, reduce: IndirectOosvarValue
			reduce(114), // ., reduce: IndirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(114), // **, reduce: IndirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(114), // [, reduce: IndirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(114), // [[, reduce: IndirectOosvarValue
			reduce(114), // [[[, reduce: IndirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3842
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4218), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3843
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(151), // ;, reduce: LogicalOrTerm
			nil,         // {
			reduce(151), // }, reduce: LogicalOrTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(151), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(151), // ||, reduce: LogicalOrTerm
			shift(3067), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3844
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(153), // ;, reduce: LogicalXORTerm
			nil,         // {
			reduce(153), // }, reduce: LogicalXORTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(153), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(153), // ||, reduce: LogicalXORTerm
			reduce(153), // ^^, reduce: LogicalXORTerm
			shift(3068), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3845
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(155), // ;, reduce: LogicalAndTerm
			nil,         // {
			reduce(155), // }, reduce: LogicalAndTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(155), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(155), // ||, reduce: LogicalAndTerm
			reduce(155), // ^^, reduce: LogicalAndTerm
			reduce(155), // &&, reduce: LogicalAndTerm
			shift(3069), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3846
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(157), // ;, reduce: AbsentCoalesceTerm
			nil,         // {
			reduce(157), // }, reduce: AbsentCoalesceTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(157), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(157), // ||, reduce: AbsentCoalesceTerm
			reduce(157), // ^^, reduce: AbsentCoalesceTerm
			reduce(157), // &&, reduce: AbsentCoalesceTerm
			reduce(157), // ??, reduce: AbsentCoalesceTerm
			shift(3070), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3847
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(159), // ;, reduce: EmptyCoalesceTerm
			nil,         // {
			reduce(159), // }, reduce: EmptyCoalesceTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(159), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(159), // ||, reduce: EmptyCoalesceTerm
			reduce(159), // ^^, reduce: EmptyCoalesceTerm
			reduce(159), // &&, reduce: EmptyCoalesceTerm
			reduce(159), // ??, reduce: EmptyCoalesceTerm
			reduce(159), // ???, reduce: EmptyCoalesceTerm
			shift(3071), // =~
			shift(3072), // !=~
			shift(3073), // ==
			shift(3074), // !=
			shift(3075), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3848
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(161), // ;, reduce: EqneTerm
			nil,         // {
			reduce(161), // }, reduce: EqneTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(3076), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(161), // ?, reduce: EqneTerm
			nil,         // :
			reduce(161), // ||, reduce: EqneTerm
			reduce(161), // ^^, reduce: EqneTerm
			reduce(161), // &&, reduce: EqneTerm
			reduce(161), // ??, reduce: EqneTerm
			reduce(161), // ???, reduce: EqneTerm
			reduce(161), // =~, reduce: EqneTerm
			reduce(161), // !=~, reduce: EqneTerm
			reduce(161), // ==, reduce: EqneTerm
			reduce(161), // !=, reduce: EqneTerm
			reduce(161), // <=>, reduce: EqneTerm
			shift(3077), // >=
			shift(3078), // <
			shift(3079), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3849
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(162), // ;, reduce: EqneTerm
			nil,         // {
			reduce(162), // }, reduce: EqneTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(3076), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(162), // ?, reduce: EqneTerm
			nil,         // :
			reduce(162), // ||, reduce: EqneTerm
			reduce(162), // ^^, reduce: EqneTerm
			reduce(162), // &&, reduce: EqneTerm
			reduce(162), // ??, reduce: EqneTerm
			reduce(162), // ???, reduce: EqneTerm
			reduce(162), // =~, reduce: EqneTerm
			reduce(162), // !=~, reduce: EqneTerm
			reduce(162), // ==, reduce: EqneTerm
			reduce(162), // !=, reduce: EqneTerm
			reduce(162), // <=>, reduce: EqneTerm
			shift(3077), // >=
			shift(3078), // <
			shift(3079), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3850
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(163), // ;, reduce: EqneTerm
			nil,         // {
			reduce(163), // }, reduce: EqneTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(3076), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(163), // ?, reduce: EqneTerm
			nil,         // :
			reduce(163), // ||, reduce: EqneTerm
			reduce(163), // ^^, reduce: EqneTerm
			reduce(163), // &&, reduce: EqneTerm
			reduce(163), // ??, reduce: EqneTerm
			reduce(163), // ???, reduce: EqneTerm
			reduce(163), // =~, reduce: EqneTerm
			reduce(163), // !=~, reduce: EqneTerm
			reduce(163), // ==, reduce: EqneTerm
			reduce(163), // !=, reduce: EqneTerm
			reduce(163), // <=>, reduce: EqneTerm
			shift(3077), // >=
			shift(3078), // <
			shift(3079), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3851
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(164), // ;, reduce: EqneTerm
			nil,         // {
			reduce(164), // }, reduce: EqneTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(3076), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(164), // ?, reduce: EqneTerm
			nil,         // :
			reduce(164), // ||, reduce: EqneTerm
			reduce(164), // ^^, reduce: EqneTerm
			reduce(164), // &&, reduce: EqneTerm
			reduce(164), // ??, reduce: EqneTerm
			reduce(164), // ???, reduce: EqneTerm
			reduce(164), // =~, reduce: EqneTerm
			reduce(164), // !=~, reduce: EqneTerm
			reduce(164), // ==, reduce: EqneTerm
			reduce(164), // !=, reduce: EqneTerm
			reduce(164), // <=>, reduce: EqneTerm
			shift(3077), // >=
			shift(3078), // <
			shift(3079), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3852
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(165), // ;, reduce: EqneTerm
			nil,         // {
			reduce(165), // }, reduce: EqneTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(3076), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(165), // ?, reduce: EqneTerm
			nil,         // :
			reduce(165), // ||, reduce: EqneTerm
			reduce(165), // ^^, reduce: EqneTerm
			reduce(165), // &&, reduce: EqneTerm
			reduce(165), // ??, reduce: EqneTerm
			reduce(165), // ???, reduce: EqneTerm
			reduce(165), // =~, reduce: EqneTerm
			reduce(165), // !=~, reduce: EqneTerm
			reduce(165), // ==, reduce: EqneTerm
			reduce(165), // !=, reduce: EqneTerm
			reduce(165), // <=>, reduce: EqneTerm
			shift(3077), // >=
			shift(3078), // <
			shift(3079), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3853
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(167), // ;, reduce: CmpTerm
			nil,         // {
			reduce(167), // }, reduce: CmpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(167), // >, reduce: CmpTerm
			nil,         // >>
			shift(3080), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(167), // ?, reduce: CmpTerm
			nil,         // :
			reduce(167), // ||, reduce: CmpTerm
			reduce(167), // ^^, reduce: CmpTerm
			reduce(167), // &&, reduce: CmpTerm
			reduce(167), // ??, reduce: CmpTerm
			reduce(167), // ???, reduce: CmpTerm
			reduce(167), // =~, reduce: CmpTerm
			reduce(167), // !=~, reduce: CmpTerm
			reduce(167), // ==, reduce: CmpTerm
			reduce(167), // !=, reduce: CmpTerm
			reduce(167), // <=>, reduce: CmpTerm
			reduce(167), // >=, reduce: CmpTerm
			reduce(167), // <, reduce: CmpTerm
			reduce(167), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3854
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(168), // ;, reduce: CmpTerm
			nil,         // {
			reduce(168), // }, reduce: CmpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(168), // >, reduce: CmpTerm
			nil,         // >>
			shift(3080), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(168), // ?, reduce: CmpTerm
			nil,         // :
			reduce(168), // ||, reduce: CmpTerm
			reduce(168), // ^^, reduce: CmpTerm
			reduce(168), // &&, reduce: CmpTerm
			reduce(168), // ??, reduce: CmpTerm
			reduce(168), // ???, reduce: CmpTerm
			reduce(168), // =~, reduce: CmpTerm
			reduce(168), // !=~, reduce: CmpTerm
			reduce(168), // ==, reduce: CmpTerm
			reduce(168), // !=, reduce: CmpTerm
			reduce(168), // <=>, reduce: CmpTerm
			reduce(168), // >=, reduce: CmpTerm
			reduce(168), // <, reduce: CmpTerm
			reduce(168), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3855
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(169), // ;, reduce: CmpTerm
			nil,         // {
			reduce(169), // }, reduce: CmpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(169), // >, reduce: CmpTerm
			nil,         // >>
			shift(3080), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(169), // ?, reduce: CmpTerm
			nil,         // :
			reduce(169), // ||, reduce: CmpTerm
			reduce(169), // ^^, reduce: CmpTerm
			reduce(169), // &&, reduce: CmpTerm
			reduce(169), // ??, reduce: CmpTerm
			reduce(169), // ???, reduce: CmpTerm
			reduce(169), // =~, reduce: CmpTerm
			reduce(169), // !=~, reduce: CmpTerm
			reduce(169), // ==, reduce: CmpTerm
			reduce(169), // !=, reduce: CmpTerm
			reduce(169), // <=>, reduce: CmpTerm
			reduce(169), // >=, reduce: CmpTerm
			reduce(169), // <, reduce: CmpTerm
			reduce(169), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3856
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(170), // ;, reduce: CmpTerm
			nil,         // {
			reduce(170), // }, reduce: CmpTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(170), // >, reduce: CmpTerm
			nil,         // >>
			shift(3080), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(170), // ?, reduce: CmpTerm
			nil,         // :
			reduce(170), // ||, reduce: CmpTerm
			reduce(170), // ^^, reduce: CmpTerm
			reduce(170), // &&, reduce: CmpTerm
			reduce(170), // ??, reduce: CmpTerm
			reduce(170), // ???, reduce: CmpTerm
			reduce(170), // =~, reduce: CmpTerm
			reduce(170), // !=~, reduce: CmpTerm
			reduce(170), // ==, reduce: CmpTerm
			reduce(170), // !=, reduce: CmpTerm
			reduce(170), // <=>, reduce: CmpTerm
			reduce(170), // >=, reduce: CmpTerm
			reduce(170), // <, reduce: CmpTerm
			reduce(170), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3857
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(172), // ;, reduce: BitwiseORTerm
			nil,         // {
			reduce(172), // }, reduce: BitwiseORTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(172), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(172), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(172), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(172), // ||, reduce: BitwiseORTerm
			reduce(172), // ^^, reduce: BitwiseORTerm
			reduce(172), // &&, reduce: BitwiseORTerm
			reduce(172), // ??, reduce: BitwiseORTerm
			reduce(172), // ???, reduce: BitwiseORTerm
			reduce(172), // =~, reduce: BitwiseORTerm
			reduce(172), // !=~, reduce: BitwiseORTerm
			reduce(172), // ==, reduce: BitwiseORTerm
			reduce(172), // !=, reduce: BitwiseORTerm
			reduce(172), // <=>, reduce: BitwiseORTerm
			reduce(172), // >=, reduce: BitwiseORTerm
			reduce(172), // <, reduce: BitwiseORTerm
			reduce(172), // <=, reduce: BitwiseORTerm
			shift(3081), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3858
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(174), // ;, reduce: BitwiseXORTerm
			nil,         // {
			reduce(174), // }, reduce: BitwiseXORTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(174), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(174), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(174), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(174), // ||, reduce: BitwiseXORTerm
			reduce(174), // ^^, reduce: BitwiseXORTerm
			reduce(174), // &&, reduce: BitwiseXORTerm
			reduce(174), // ??, reduce: BitwiseXORTerm
			reduce(174), // ???, reduce: BitwiseXORTerm
			reduce(174), // =~, reduce: BitwiseXORTerm
			reduce(174), // !=~, reduce: BitwiseXORTerm
			reduce(174), // ==, reduce: BitwiseXORTerm
			reduce(174), // !=, reduce: BitwiseXORTerm
			reduce(174), // <=>, reduce: BitwiseXORTerm
			reduce(174), // >=, reduce: BitwiseXORTerm
			reduce(174), // <, reduce: BitwiseXORTerm
			reduce(174), // <=, reduce: BitwiseXORTerm
			reduce(174), // ^, reduce: BitwiseXORTerm
			shift(3082), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3859
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(176), // ;, reduce: BitwiseANDTerm
			nil,         // {
			reduce(176), // }, reduce: BitwiseANDTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(176), // >, reduce: BitwiseANDTerm
			shift(3083), // >>
			reduce(176), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(176), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(176), // ||, reduce: BitwiseANDTerm
			reduce(176), // ^^, reduce: BitwiseANDTerm
			reduce(176), // &&, reduce: BitwiseANDTerm
			reduce(176), // ??, reduce: BitwiseANDTerm
			reduce(176), // ???, reduce: BitwiseANDTerm
			reduce(176), // =~, reduce: BitwiseANDTerm
			reduce(176), // !=~, reduce: BitwiseANDTerm
			reduce(176), // ==, reduce: BitwiseANDTerm
			reduce(176), // !=, reduce: BitwiseANDTerm
			reduce(176), // <=>, reduce: BitwiseANDTerm
			reduce(176), // >=, reduce: BitwiseANDTerm
			reduce(176), // <, reduce: BitwiseANDTerm
			reduce(176), // <=, reduce: BitwiseANDTerm
			reduce(176), // ^, reduce: BitwiseANDTerm
			reduce(176), // &, reduce: BitwiseANDTerm
			shift(3084), // <<
			shift(3085), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3860
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(179), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			reduce(179), // }, reduce: BitwiseShiftTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(179), // >, reduce: BitwiseShiftTerm
			reduce(179), // >>, reduce: BitwiseShiftTerm
			reduce(179), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(179), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(179), // ||, reduce: BitwiseShiftTerm
			reduce(179), // ^^, reduce: BitwiseShiftTerm
			reduce(179), // &&, reduce: BitwiseShiftTerm
			reduce(179), // ??, reduce: BitwiseShiftTerm
			reduce(179), // ???, reduce: BitwiseShiftTerm
			reduce(179), // =~, reduce: BitwiseShiftTerm
			reduce(179), // !=~, reduce: BitwiseShiftTerm
			reduce(179), // ==, reduce: BitwiseShiftTerm
			reduce(179), // !=, reduce: BitwiseShiftTerm
			reduce(179), // <=>, reduce: BitwiseShiftTerm
			reduce(179), // >=, reduce: BitwiseShiftTerm
			reduce(179), // <, reduce: BitwiseShiftTerm
			reduce(179), // <=, reduce: BitwiseShiftTerm
			reduce(179), // ^, reduce: BitwiseShiftTerm
			reduce(179), // &, reduce: BitwiseShiftTerm
			reduce(179), // <<, reduce: BitwiseShiftTerm
			reduce(179), // >>>, reduce: BitwiseShiftTerm
			shift(3086), // +
			shift(3087), // -
			shift(3088), // .+
			shift(3089), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3861
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(178), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			reduce(178), // }, reduce: BitwiseShiftTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(178), // >, reduce: BitwiseShiftTerm
			reduce(178), // >>, reduce: BitwiseShiftTerm
			reduce(178), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(178), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(178), // ||, reduce: BitwiseShiftTerm
			reduce(178), // ^^, reduce: BitwiseShiftTerm
			reduce(178), // &&, reduce: BitwiseShiftTerm
			reduce(178), // ??, reduce: BitwiseShiftTerm
			reduce(178), // ???, reduce: BitwiseShiftTerm
			reduce(178), // =~, reduce: BitwiseShiftTerm
			reduce(178), // !=~, reduce: BitwiseShiftTerm
			reduce(178), // ==, reduce: BitwiseShiftTerm
			reduce(178), // !=, reduce: BitwiseShiftTerm
			reduce(178), // <=>, reduce: BitwiseShiftTerm
			reduce(178), // >=, reduce: BitwiseShiftTerm
			reduce(178), // <, reduce: BitwiseShiftTerm
			reduce(178), // <=, reduce: BitwiseShiftTerm
			reduce(178), // ^, reduce: BitwiseShiftTerm
			reduce(178), // &, reduce: BitwiseShiftTerm
			reduce(178), // <<, reduce: BitwiseShiftTerm
			reduce(178), // >>>, reduce: BitwiseShiftTerm
			shift(3086), // +
			shift(3087), // -
			shift(3088), // .+
			shift(3089), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3862
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(180), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			reduce(180), // }, reduce: BitwiseShiftTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(180), // >, reduce: BitwiseShiftTerm
			reduce(180), // >>, reduce: BitwiseShiftTerm
			reduce(180), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(180), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(180), // ||, reduce: BitwiseShiftTerm
			reduce(180), // ^^, reduce: BitwiseShiftTerm
			reduce(180), // &&, reduce: BitwiseShiftTerm
			reduce(180), // ??, reduce: BitwiseShiftTerm
			reduce(180), // ???, reduce: BitwiseShiftTerm
			reduce(180), // =~, reduce: BitwiseShiftTerm
			reduce(180), // !=~, reduce: BitwiseShiftTerm
			reduce(180), // ==, reduce: BitwiseShiftTerm
			reduce(180), // !=, reduce: BitwiseShiftTerm
			reduce(180), // <=>, reduce: BitwiseShiftTerm
			reduce(180), // >=, reduce: BitwiseShiftTerm
			reduce(180), // <, reduce: BitwiseShiftTerm
			reduce(180), // <=, reduce: BitwiseShiftTerm
			reduce(180), // ^, reduce: BitwiseShiftTerm
			reduce(180), // &, reduce: BitwiseShiftTerm
			reduce(180), // <<, reduce: BitwiseShiftTerm
			reduce(180), // >>>, reduce: BitwiseShiftTerm
			shift(3086), // +
			shift(3087), // -
			shift(3088), // .+
			shift(3089), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3863
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(182), // ;, reduce: AddsubdotTerm
			nil,         // {
			reduce(182), // }, reduce: AddsubdotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(182), // >, reduce: AddsubdotTerm
			reduce(182), // >>, reduce: AddsubdotTerm
			reduce(182), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(182), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(182), // ||, reduce: AddsubdotTerm
			reduce(182), // ^^, reduce: AddsubdotTerm
			reduce(182), // &&, reduce: AddsubdotTerm
			reduce(182), // ??, reduce: AddsubdotTerm
			reduce(182), // ???, reduce: AddsubdotTerm
			reduce(182), // =~, reduce: AddsubdotTerm
			reduce(182), // !=~, reduce: AddsubdotTerm
			reduce(182), // ==, reduce: AddsubdotTerm
			reduce(182), // !=, reduce: AddsubdotTerm
			reduce(182), // <=>, reduce: AddsubdotTerm
			reduce(182), // >=, reduce: AddsubdotTerm
			reduce(182), // <, reduce: AddsubdotTerm
			reduce(182), // <=, reduce: AddsubdotTerm
			reduce(182), // ^, reduce: AddsubdotTerm
			reduce(182), // &, reduce: AddsubdotTerm
			reduce(182), // <<, reduce: AddsubdotTerm
			reduce(182), // >>>, reduce: AddsubdotTerm
			reduce(182), // +, reduce: AddsubdotTerm
			reduce(182), // -, reduce: AddsubdotTerm
			reduce(182), // .+, reduce: AddsubdotTerm
			reduce(182), // .-, reduce: AddsubdotTerm
			shift(3091), // *
			shift(3092), // /
			shift(3093), // //
			shift(3094), // %
			shift(3095), // .*
			shift(3096), // ./
			shift(3097), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3864
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(183), // ;, reduce: AddsubdotTerm
			nil,         // {
			reduce(183), // }, reduce: AddsubdotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(183), // >, reduce: AddsubdotTerm
			reduce(183), // >>, reduce: AddsubdotTerm
			reduce(183), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(183), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(183), // ||, reduce: AddsubdotTerm
			reduce(183), // ^^, reduce: AddsubdotTerm
			reduce(183), // &&, reduce: AddsubdotTerm
			reduce(183), // ??, reduce: AddsubdotTerm
			reduce(183), // ???, reduce: AddsubdotTerm
			reduce(183), // =~, reduce: AddsubdotTerm
			reduce(183), // !=~, reduce: AddsubdotTerm
			reduce(183), // ==, reduce: AddsubdotTerm
			reduce(183), // !=, reduce: AddsubdotTerm
			reduce(183), // <=>, reduce: AddsubdotTerm
			reduce(183), // >=, reduce: AddsubdotTerm
			reduce(183), // <, reduce: AddsubdotTerm
			reduce(183), // <=, reduce: AddsubdotTerm
			reduce(183), // ^, reduce: AddsubdotTerm
			reduce(183), // &, reduce: AddsubdotTerm
			reduce(183), // <<, reduce: AddsubdotTerm
			reduce(183), // >>>, reduce: AddsubdotTerm
			reduce(183), // +, reduce: AddsubdotTerm
			reduce(183), // -, reduce: AddsubdotTerm
			reduce(183), // .+, reduce: AddsubdotTerm
			reduce(183), // .-, reduce: AddsubdotTerm
			shift(3091), // *
			shift(3092), // /
			shift(3093), // //
			shift(3094), // %
			shift(3095), // .*
			shift(3096), // ./
			shift(3097), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3865
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(184), // ;, reduce: AddsubdotTerm
			nil,         // {
			reduce(184), // }, reduce: AddsubdotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(184), // >, reduce: AddsubdotTerm
			reduce(184), // >>, reduce: AddsubdotTerm
			reduce(184), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(184), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(184), // ||, reduce: AddsubdotTerm
			reduce(184), // ^^, reduce: AddsubdotTerm
			reduce(184), // &&, reduce: AddsubdotTerm
			reduce(184), // ??, reduce: AddsubdotTerm
			reduce(184), // ???, reduce: AddsubdotTerm
			reduce(184), // =~, reduce: AddsubdotTerm
			reduce(184), // !=~, reduce: AddsubdotTerm
			reduce(184), // ==, reduce: AddsubdotTerm
			reduce(184), // !=, reduce: AddsubdotTerm
			reduce(184), // <=>, reduce: AddsubdotTerm
			reduce(184), // >=, reduce: AddsubdotTerm
			reduce(184), // <, reduce: AddsubdotTerm
			reduce(184), // <=, reduce: AddsubdotTerm
			reduce(184), // ^, reduce: AddsubdotTerm
			reduce(184), // &, reduce: AddsubdotTerm
			reduce(184), // <<, reduce: AddsubdotTerm
			reduce(184), // >>>, reduce: AddsubdotTerm
			reduce(184), // +, reduce: AddsubdotTerm
			reduce(184), // -, reduce: AddsubdotTerm
			reduce(184), // .+, reduce: AddsubdotTerm
			reduce(184), // .-, reduce: AddsubdotTerm
			shift(3091), // *
			shift(3092), // /
			shift(3093), // //
			shift(3094), // %
			shift(3095), // .*
			shift(3096), // ./
			shift(3097), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3866
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(185), // ;, reduce: AddsubdotTerm
			nil,         // {
			reduce(185), // }, reduce: AddsubdotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(185), // >, reduce: AddsubdotTerm
			reduce(185), // >>, reduce: AddsubdotTerm
			reduce(185), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(185), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(185), // ||, reduce: AddsubdotTerm
			reduce(185), // ^^, reduce: AddsubdotTerm
			reduce(185), // &&, reduce: AddsubdotTerm
			reduce(185), // ??, reduce: AddsubdotTerm
			reduce(185), // ???, reduce: AddsubdotTerm
			reduce(185), // =~, reduce: AddsubdotTerm
			reduce(185), // !=~, reduce: AddsubdotTerm
			reduce(185), // ==, reduce: AddsubdotTerm
			reduce(185), // !=, reduce: AddsubdotTerm
			reduce(185), // <=>, reduce: AddsubdotTerm
			reduce(185), // >=, reduce: AddsubdotTerm
			reduce(185), // <, reduce: AddsubdotTerm
			reduce(185), // <=, reduce: AddsubdotTerm
			reduce(185), // ^, reduce: AddsubdotTerm
			reduce(185), // &, reduce: AddsubdotTerm
			reduce(185), // <<, reduce: AddsubdotTerm
			reduce(185), // >>>, reduce: AddsubdotTerm
			reduce(185), // +, reduce: AddsubdotTerm
			reduce(185), // -, reduce: AddsubdotTerm
			reduce(185), // .+, reduce: AddsubdotTerm
			reduce(185), // .-, reduce: AddsubdotTerm
			shift(3091), // *
			shift(3092), // /
			shift(3093), // //
			shift(3094), // %
			shift(3095), // .*
			shift(3096), // ./
			shift(3097), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3867
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(187), // ;, reduce: MuldivTerm
			nil,         // {
			reduce(187), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(187), // >, reduce: MuldivTerm
			reduce(187), // >>, reduce: MuldivTerm
			reduce(187), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(187), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(187), // ||, reduce: MuldivTerm
			reduce(187), // ^^, reduce: MuldivTerm
			reduce(187), // &&, reduce: MuldivTerm
			reduce(187), // ??, reduce: MuldivTerm
			reduce(187), // ???, reduce: MuldivTerm
			reduce(187), // =~, reduce: MuldivTerm
			reduce(187), // !=~, reduce: MuldivTerm
			reduce(187), // ==, reduce: MuldivTerm
			reduce(187), // !=, reduce: MuldivTerm
			reduce(187), // <=>, reduce: MuldivTerm
			reduce(187), // >=, reduce: MuldivTerm
			reduce(187), // <, reduce: MuldivTerm
			reduce(187), // <=, reduce: MuldivTerm
			reduce(187), // ^, reduce: MuldivTerm
			reduce(187), // &, reduce: MuldivTerm
			reduce(187), // <<, reduce: MuldivTerm
			reduce(187), // >>>, reduce: MuldivTerm
			reduce(187), // +, reduce: MuldivTerm
			reduce(187), // -, reduce: MuldivTerm
			reduce(187), // .+, reduce: MuldivTerm
			reduce(187), // .-, reduce: MuldivTerm
			reduce(187), // *, reduce: MuldivTerm
			reduce(187), // /, reduce: MuldivTerm
			reduce(187), // //, reduce: MuldivTerm
			reduce(187), // %, reduce: MuldivTerm
			reduce(187), // .*, reduce: MuldivTerm
			reduce(187), // ./, reduce: MuldivTerm
			reduce(187), // .//, reduce: MuldivTerm
			shift(3101), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3868
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(188), // ;, reduce: MuldivTerm
			nil,         // {
			reduce(188), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(188), // >, reduce: MuldivTerm
			reduce(188), // >>, reduce: MuldivTerm
			reduce(188), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(188), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(188), // ||, reduce: MuldivTerm
			reduce(188), // ^^, reduce: MuldivTerm
			reduce(188), // &&, reduce: MuldivTerm
			reduce(188), // ??, reduce: MuldivTerm
			reduce(188), // ???, reduce: MuldivTerm
			reduce(188), // =~, reduce: MuldivTerm
			reduce(188), // !=~, reduce: MuldivTerm
			reduce(188), // ==, reduce: MuldivTerm
			reduce(188), // !=, reduce: MuldivTerm
			reduce(188), // <=>, reduce: MuldivTerm
			reduce(188), // >=, reduce: MuldivTerm
			reduce(188), // <, reduce: MuldivTerm
			reduce(188), // <=, reduce: MuldivTerm
			reduce(188), // ^, reduce: MuldivTerm
			reduce(188), // &, reduce: MuldivTerm
			reduce(188), // <<, reduce: MuldivTerm
			reduce(188), // >>>, reduce: MuldivTerm
			reduce(188), // +, reduce: MuldivTerm
			reduce(188), // -, reduce: MuldivTerm
			reduce(188), // .+, reduce: MuldivTerm
			reduce(188), // .-, reduce: MuldivTerm
			reduce(188), // *, reduce: MuldivTerm
			reduce(188), // /, reduce: MuldivTerm
			reduce(188), // //, reduce: MuldivTerm
			reduce(188), // %, reduce: MuldivTerm
			reduce(188), // .*, reduce: MuldivTerm
			reduce(188), // ./, reduce: MuldivTerm
			reduce(188), // .//, reduce: MuldivTerm
			shift(3101), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3869
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(189), // ;, reduce: MuldivTerm
			nil,         // {
			reduce(189), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(189), // >, reduce: MuldivTerm
			reduce(189), // >>, reduce: MuldivTerm
			reduce(189), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(189), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(189), // ||, reduce: MuldivTerm
			reduce(189), // ^^, reduce: MuldivTerm
			reduce(189), // &&, reduce: MuldivTerm
			reduce(189), // ??, reduce: MuldivTerm
			reduce(189), // ???, reduce: MuldivTerm
			reduce(189), // =~, reduce: MuldivTerm
			reduce(189), // !=~, reduce: MuldivTerm
			reduce(189), // ==, reduce: MuldivTerm
			reduce(189), // !=, reduce: MuldivTerm
			reduce(189), // <=>, reduce: MuldivTerm
			reduce(189), // >=, reduce: MuldivTerm
			reduce(189), // <, reduce: MuldivTerm
			reduce(189), // <=, reduce: MuldivTerm
			reduce(189), // ^, reduce: MuldivTerm
			reduce(189), // &, reduce: MuldivTerm
			reduce(189), // <<, reduce: MuldivTerm
			reduce(189), // >>>, reduce: MuldivTerm
			reduce(189), // +, reduce: MuldivTerm
			reduce(189), // -, reduce: MuldivTerm
			reduce(189), // .+, reduce: MuldivTerm
			reduce(189), // .-, reduce: MuldivTerm
			reduce(189), // *, reduce: MuldivTerm
			reduce(189), // /, reduce: MuldivTerm
			reduce(189), // //, reduce: MuldivTerm
			reduce(189), // %, reduce: MuldivTerm
			reduce(189), // .*, reduce: MuldivTerm
			reduce(189), // ./, reduce: MuldivTerm
			reduce(189), // .//, reduce: MuldivTerm
			shift(3101), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3870
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(190), // ;, reduce: MuldivTerm
			nil,         // {
			reduce(190), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(190), // >, reduce: MuldivTerm
			reduce(190), // >>, reduce: MuldivTerm
			reduce(190), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(190), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(190), // ||, reduce: MuldivTerm
			reduce(190), // ^^, reduce: MuldivTerm
			reduce(190), // &&, reduce: MuldivTerm
			reduce(190), // ??, reduce: MuldivTerm
			reduce(190), // ???, reduce: MuldivTerm
			reduce(190), // =~, reduce: MuldivTerm
			reduce(190), // !=~, reduce: MuldivTerm
			reduce(190), // ==, reduce: MuldivTerm
			reduce(190), // !=, reduce: MuldivTerm
			reduce(190), // <=>, reduce: MuldivTerm
			reduce(190), // >=, reduce: MuldivTerm
			reduce(190), // <, reduce: MuldivTerm
			reduce(190), // <=, reduce: MuldivTerm
			reduce(190), // ^, reduce: MuldivTerm
			reduce(190), // &, reduce: MuldivTerm
			reduce(190), // <<, reduce: MuldivTerm
			reduce(190), // >>>, reduce: MuldivTerm
			reduce(190), // +, reduce: MuldivTerm
			reduce(190), // -, reduce: MuldivTerm
			reduce(190), // .+, reduce: MuldivTerm
			reduce(190), // .-, reduce: MuldivTerm
			reduce(190), // *, reduce: MuldivTerm
			reduce(190), // /, reduce: MuldivTerm
			reduce(190), // //, reduce: MuldivTerm
			reduce(190), // %, reduce: MuldivTerm
			reduce(190), // .*, reduce: MuldivTerm
			reduce(190), // ./, reduce: MuldivTerm
			reduce(190), // .//, reduce: MuldivTerm
			shift(3101), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3871
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(191), // ;, reduce: MuldivTerm
			nil,         // {
			reduce(191), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(191), // >, reduce: MuldivTerm
			reduce(191), // >>, reduce: MuldivTerm
			reduce(191), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(191), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(191), // ||, reduce: MuldivTerm
			reduce(191), // ^^, reduce: MuldivTerm
			reduce(191), // &&, reduce: MuldivTerm
			reduce(191), // ??, reduce: MuldivTerm
			reduce(191), // ???, reduce: MuldivTerm
			reduce(191), // =~, reduce: MuldivTerm
			reduce(191), // !=~, reduce: MuldivTerm
			reduce(191), // ==, reduce: MuldivTerm
			reduce(191), // !=, reduce: MuldivTerm
			reduce(191), // <=>, reduce: MuldivTerm
			reduce(191), // >=, reduce: MuldivTerm
			reduce(191), // <, reduce: MuldivTerm
			reduce(191), // <=, reduce: MuldivTerm
			reduce(191), // ^, reduce: MuldivTerm
			reduce(191), // &, reduce: MuldivTerm
			reduce(191), // <<, reduce: MuldivTerm
			reduce(191), // >>>, reduce: MuldivTerm
			reduce(191), // +, reduce: MuldivTerm
			reduce(191), // -, reduce: MuldivTerm
			reduce(191), // .+, reduce: MuldivTerm
			reduce(191), // .-, reduce: MuldivTerm
			reduce(191), // *, reduce: MuldivTerm
			reduce(191), // /, reduce: MuldivTerm
			reduce(191), // //, reduce: MuldivTerm
			reduce(191), // %, reduce: MuldivTerm
			reduce(191), // .*, reduce: MuldivTerm
			reduce(191), // ./, reduce: MuldivTerm
			reduce(191), // .//, reduce: MuldivTerm
			shift(3101), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3872
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(192), // ;, reduce: MuldivTerm
			nil,         // {
			reduce(192), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(192), // >, reduce: MuldivTerm
			reduce(192), // >>, reduce: MuldivTerm
			reduce(192), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(192), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(192), // ||, reduce: MuldivTerm
			reduce(192), // ^^, reduce: MuldivTerm
			reduce(192), // &&, reduce: MuldivTerm
			reduce(192), // ??, reduce: MuldivTerm
			reduce(192), // ???, reduce: MuldivTerm
			reduce(192), // =~, reduce: MuldivTerm
			reduce(192), // !=~, reduce: MuldivTerm
			reduce(192), // ==, reduce: MuldivTerm
			reduce(192), // !=, reduce: MuldivTerm
			reduce(192), // <=>, reduce: MuldivTerm
			reduce(192), // >=, reduce: MuldivTerm
			reduce(192), // <, reduce: MuldivTerm
			reduce(192), // <=, reduce: MuldivTerm
			reduce(192), // ^, reduce: MuldivTerm
			reduce(192), // &, reduce: MuldivTerm
			reduce(192), // <<, reduce: MuldivTerm
			reduce(192), // >>>, reduce: MuldivTerm
			reduce(192), // +, reduce: MuldivTerm
			reduce(192), // -, reduce: MuldivTerm
			reduce(192), // .+, reduce: MuldivTerm
			reduce(192), // .-, reduce: MuldivTerm
			reduce(192), // *, reduce: MuldivTerm
			reduce(192), // /, reduce: MuldivTerm
			reduce(192), // //, reduce: MuldivTerm
			reduce(192), // %, reduce: MuldivTerm
			reduce(192), // .*, reduce: MuldivTerm
			reduce(192), // ./, reduce: MuldivTerm
			reduce(192), // .//, reduce: MuldivTerm
			shift(3101), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3873
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(193), // ;, reduce: MuldivTerm
			nil,         // {
			reduce(193), // }, reduce: MuldivTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(193), // >, reduce: MuldivTerm
			reduce(193), // >>, reduce: MuldivTerm
			reduce(193), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(193), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(193), // ||, reduce: MuldivTerm
			reduce(193), // ^^, reduce: MuldivTerm
			reduce(193), // &&, reduce: MuldivTerm
			reduce(193), // ??, reduce: MuldivTerm
			reduce(193), // ???, reduce: MuldivTerm
			reduce(193), // =~, reduce: MuldivTerm
			reduce(193), // !=~, reduce: MuldivTerm
			reduce(193), // ==, reduce: MuldivTerm
			reduce(193), // !=, reduce: MuldivTerm
			reduce(193), // <=>, reduce: MuldivTerm
			reduce(193), // >=, reduce: MuldivTerm
			reduce(193), // <, reduce: MuldivTerm
			reduce(193), // <=, reduce: MuldivTerm
			reduce(193), // ^, reduce: MuldivTerm
			reduce(193), // &, reduce: MuldivTerm
			reduce(193), // <<, reduce: MuldivTerm
			reduce(193), // >>>, reduce: MuldivTerm
			reduce(193), // +, reduce: MuldivTerm
			reduce(193), // -, reduce: MuldivTerm
			reduce(193), // .+, reduce: MuldivTerm
			reduce(193), // .-, reduce: MuldivTerm
			reduce(193), // *, reduce: MuldivTerm
			reduce(193), // /, reduce: MuldivTerm
			reduce(193), // //, reduce: MuldivTerm
			reduce(193), // %, reduce: MuldivTerm
			reduce(193), // .*, reduce: MuldivTerm
			reduce(193), // ./, reduce: MuldivTerm
			reduce(193), // .//, reduce: MuldivTerm
			shift(3101), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3874
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(195), // ;, reduce: DotTerm
			nil,         // {
			reduce(195), // }, reduce: DotTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(195), // >, reduce: DotTerm
			reduce(195), // >>, reduce: DotTerm
			reduce(195), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(195), // ?, reduce: DotTerm
			nil,         // :
			reduce(195), // ||, reduce: DotTerm
			reduce(195), // ^^, reduce: DotTerm
			reduce(195), // &&, reduce: DotTerm
			reduce(195), // ??, reduce: DotTerm
			reduce(195), // ???, reduce: DotTerm
			reduce(195), // =~, reduce: DotTerm
			reduce(195), // !=~, reduce: DotTerm
			reduce(195), // ==, reduce: DotTerm
			reduce(195), // !=, reduce: DotTerm
			reduce(195), // <=>, reduce: DotTerm
			reduce(195), // >=, reduce: DotTerm
			reduce(195), // <, reduce: DotTerm
			reduce(195), // <=, reduce: DotTerm
			reduce(195), // ^, reduce: DotTerm
			reduce(195), // &, reduce: DotTerm
			reduce(195), // <<, reduce: DotTerm
			reduce(195), // >>>, reduce: DotTerm
			reduce(195), // +, reduce: DotTerm
			reduce(195), // -, reduce: DotTerm
			reduce(195), // .+, reduce: DotTerm
			reduce(195), // .-, reduce: DotTerm
			reduce(195), // *, reduce: DotTerm
			reduce(195), // /, reduce: DotTerm
			reduce(195), // //, reduce: DotTerm
			reduce(195), // %, reduce: DotTerm
			reduce(195), // .*, reduce: DotTerm
			reduce(195), // ./, reduce: DotTerm
			reduce(195), // .//, reduce: DotTerm
			reduce(195), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3875
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3876
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3877
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(204), // ;, reduce: PowTerm
			nil,         // {
			reduce(204), // }, reduce: PowTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(204), // >, reduce: PowTerm
			reduce(204), // >>, reduce: PowTerm
			reduce(204), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(204), // ?, reduce: PowTerm
			nil,         // :
			reduce(204), // ||, reduce: PowTerm
			reduce(204), // ^^, reduce: PowTerm
			reduce(204), // &&, reduce: PowTerm
			reduce(204), // ??, reduce: PowTerm
			reduce(204), // ???, reduce: PowTerm
			reduce(204), // =~, reduce: PowTerm
			reduce(204), // !=~, reduce: PowTerm
			reduce(204), // ==, reduce: PowTerm
			reduce(204), // !=, reduce: PowTerm
			reduce(204), // <=>, reduce: PowTerm
			reduce(204), // >=, reduce: PowTerm
			reduce(204), // <, reduce: PowTerm
			reduce(204), // <=, reduce: PowTerm
			reduce(204), // ^, reduce: PowTerm
			reduce(204), // &, reduce: PowTerm
			reduce(204), // <<, reduce: PowTerm
			reduce(204), // >>>, reduce: PowTerm
			reduce(204), // +, reduce: PowTerm
			reduce(204), // -, reduce: PowTerm
			reduce(204), // .+, reduce: PowTerm
			reduce(204), // .-, reduce: PowTerm
			reduce(204), // *, reduce: PowTerm
			reduce(204), // /, reduce: PowTerm
			reduce(204), // //, reduce: PowTerm
			reduce(204), // %, reduce: PowTerm
			reduce(204), // .*, reduce: PowTerm
			reduce(204), // ./, reduce: PowTerm
			reduce(204), // .//, reduce: PowTerm
			reduce(204), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3878
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4221), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4222), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3879
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4224), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3880
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4225), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3881
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4226), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3882
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(229), // ;, reduce: ArrayLiteral
			nil,         // {
			reduce(229), // }, reduce: ArrayLiteral
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(229), // >, reduce: ArrayLiteral
			reduce(229), // >>, reduce: ArrayLiteral
			reduce(229), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(229), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(229), // ||, reduce: ArrayLiteral
			reduce(229), // ^^, reduce: ArrayLiteral
			reduce(229), // &&, reduce: ArrayLiteral
			reduce(229), // ??, reduce: ArrayLiteral
			reduce(229), // ???, reduce: ArrayLiteral
			reduce(229), // =~, reduce: ArrayLiteral
			reduce(229), // !=~, reduce: ArrayLiteral
			reduce(229), // ==, reduce: ArrayLiteral
			reduce(229), // !=, reduce: ArrayLiteral
			reduce(229), // <=>, reduce: ArrayLiteral
			reduce(229), // >=, reduce: ArrayLiteral
			reduce(229), // <, reduce: ArrayLiteral
			reduce(229), // <=, reduce: ArrayLiteral
			reduce(229), // ^, reduce: ArrayLiteral
			reduce(229), // &, reduce: ArrayLiteral
			reduce(229), // <<, reduce: ArrayLiteral
			reduce(229), // >>>, reduce: ArrayLiteral
			reduce(229), // +, reduce: ArrayLiteral
			reduce(229), // -, reduce: ArrayLiteral
			reduce(229), // .+, reduce: ArrayLiteral
			reduce(229), // .-, reduce: ArrayLiteral
			reduce(229), // *, reduce: ArrayLiteral
			reduce(229), // /, reduce: ArrayLiteral
			reduce(229), // //, reduce: ArrayLiteral
			reduce(229), // %, reduce: ArrayLiteral
			reduce(229), // .*, reduce: ArrayLiteral
			reduce(229), // ./, reduce: ArrayLiteral
			reduce(229), // .//, reduce: ArrayLiteral
			reduce(229), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(229), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(229), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(229), // [[, reduce: ArrayLiteral
			reduce(229), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3883
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(255), // ;, reduce: ENV
			nil,         // {
			reduce(255), // }, reduce: ENV
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(255), // >, reduce: ENV
			reduce(255), // >>, reduce: ENV
			reduce(255), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(255), // ?, reduce: ENV
			nil,         // :
			reduce(255), // ||, reduce: ENV
			reduce(255), // ^^, reduce: ENV
			reduce(255), // &&, reduce: ENV
			reduce(255), // ??, reduce: ENV
			reduce(255), // ???, reduce: ENV
			reduce(255), // =~, reduce: ENV
			reduce(255), // !=~, reduce: ENV
			reduce(255), // ==, reduce: ENV
			reduce(255), // !=, reduce: ENV
			reduce(255), // <=>, reduce: ENV
			reduce(255), // >=, reduce: ENV
			reduce(255), // <, reduce: ENV
			reduce(255), // <=, reduce: ENV
			reduce(255), // ^, reduce: ENV
			reduce(255), // &, reduce: ENV
			reduce(255), // <<, reduce: ENV
			reduce(255), // >>>, reduce: ENV
			reduce(255), // +, reduce: ENV
			reduce(255), // -, reduce: ENV
			reduce(255), // .+, reduce: ENV
			reduce(255), // .-, reduce: ENV
			reduce(255), // *, reduce: ENV
			reduce(255), // /, reduce: ENV
			reduce(255), // //, reduce: ENV
			reduce(255), // %, reduce: ENV
			reduce(255), // .*, reduce: ENV
			reduce(255), // ./, reduce: ENV
			reduce(255), // .//, reduce: ENV
			reduce(255), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(255), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(255), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(255), // [[, reduce: ENV
			reduce(255), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3884
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4227), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3885
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4228), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3886
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(268), // ;, reduce: FunctionCallsite
			nil,         // {
			reduce(268), // }, reduce: FunctionCallsite
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(268), // >, reduce: FunctionCallsite
			reduce(268), // >>, reduce: FunctionCallsite
			reduce(268), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(268), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(268), // ||, reduce: FunctionCallsite
			reduce(268), // ^^, reduce: FunctionCallsite
			reduce(268), // &&, reduce: FunctionCallsite
			reduce(268), // ??, reduce: FunctionCallsite
			reduce(268), // ???, reduce: FunctionCallsite
			reduce(268), // =~, reduce: FunctionCallsite
			reduce(268), // !=~, reduce: FunctionCallsite
			reduce(268), // ==, reduce: FunctionCallsite
			reduce(268), // !=, reduce: FunctionCallsite
			reduce(268), // <=>, reduce: FunctionCallsite
			reduce(268), // >=, reduce: FunctionCallsite
			reduce(268), // <, reduce: FunctionCallsite
			reduce(268), // <=, reduce: FunctionCallsite
			reduce(268), // ^, reduce: FunctionCallsite
			reduce(268), // &, reduce: FunctionCallsite
			reduce(268), // <<, reduce: FunctionCallsite
			reduce(268), // >>>, reduce: FunctionCallsite
			reduce(268), // +, reduce: FunctionCallsite
			reduce(268), // -, reduce: FunctionCallsite
			reduce(268), // .+, reduce: FunctionCallsite
			reduce(268), // .-, reduce: FunctionCallsite
			reduce(268), // *, reduce: FunctionCallsite
			reduce(268), // /, reduce: FunctionCallsite
			reduce(268), // //, reduce: FunctionCallsite
			reduce(268), // %, reduce: FunctionCallsite
			reduce(268), // .*, reduce: FunctionCallsite
			reduce(268), // ./, reduce: FunctionCallsite
			reduce(268), // .//, reduce: FunctionCallsite
			reduce(268), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(268), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(268), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(268), // [[, reduce: FunctionCallsite
			reduce(268), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3887
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4229), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3888
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(49), // ;, reduce: PrintStatement
			nil,        // {
			reduce(49), // }, reduce: PrintStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3889
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(53), // ;, reduce: PrintnStatement
			nil,        // {
			reduce(53), // }, reduce: PrintnStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3890
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(61), // ;, reduce: DumpStatement
			nil,        // {
			reduce(61), // }, reduce: DumpStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3891
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(64), // ;, reduce: TeeStatement
			nil,        // {
			reduce(64), // }, reduce: TeeStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3892
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(109), // ;, reduce: FullSrec
			nil,         // {
			reduce(109), // }, reduce: FullSrec
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3893
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(235), // ;, reduce: MapLiteral
			nil,         // {
			reduce(235), // }, reduce: MapLiteral
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(235), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3894
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(66), // ;, reduce: EmitFStatement
			nil,        // {
			reduce(66), // }, reduce: EmitFStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3895
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(85), // ;, reduce: EmittableList
			nil,        // {
			reduce(85), // }, reduce: EmittableList
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3896
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(105), // ;, reduce: IndirectFieldValue
			nil,         // {
			reduce(105), // }, reduce: IndirectFieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(105), // ,, reduce: IndirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3897
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(114), // ;, reduce: IndirectOosvarValue
			nil,         // {
			reduce(114), // }, reduce: IndirectOosvarValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(114), // ,, reduce: IndirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3898
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(69),  // ;, reduce: EmitStatement
			nil,         // {
			reduce(69),  // }, reduce: EmitStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(4230), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3899
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1330), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			shift(1343), // field_name
			shift(1344), // $[
			nil,         // ]
			shift(1345), // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(1346), // full_srec
			shift(1347), // oosvar_name
			shift(1348), // @[
			shift(1349), // braced_oosvar_name
			shift(1350), // full_oosvar
			shift(1351), // all
			shift(1352), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3900
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(97),  // ;, reduce: EmitKeys
			nil,         // {
			reduce(97),  // }, reduce: EmitKeys
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(4232), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3901
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(72), // ;, reduce: EmitStatement
			nil,        // {
			reduce(72), // }, reduce: EmitStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3902
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(70),  // ;, reduce: EmitStatement
			nil,         // {
			reduce(70),  // }, reduce: EmitStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(4233), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3903
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(77),  // ;, reduce: EmitPStatement
			nil,         // {
			reduce(77),  // }, reduce: EmitPStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(4234), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3904
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1330), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			shift(1343), // field_name
			shift(1344), // $[
			nil,         // ]
			shift(1345), // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			shift(1346), // full_srec
			shift(1347), // oosvar_name
			shift(1348), // @[
			shift(1349), // braced_oosvar_name
			shift(1350), // full_oosvar
			shift(1351), // all
			shift(1352), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3905
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(80), // ;, reduce: EmitPStatement
			nil,        // {
			reduce(80), // }, reduce: EmitPStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3906
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(78),  // ;, reduce: EmitPStatement
			nil,         // {
			reduce(78),  // }, reduce: EmitPStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(4236), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3907
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(107), // ;, reduce: PositionalFieldName
			reduce(107), // {, reduce: PositionalFieldName
			reduce(107), // }, reduce: PositionalFieldName
			reduce(107), // =, reduce: PositionalFieldName
			nil,         // unset
			nil,         // filter
			reduce(107), // >, reduce: PositionalFieldName
			reduce(107), // >>, reduce: PositionalFieldName
			reduce(107), // |, reduce: PositionalFieldName
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(107), // ||=, reduce: PositionalFieldName
			reduce(107), // ^^=, reduce: PositionalFieldName
			reduce(107), // &&=, reduce: PositionalFieldName
			reduce(107), // ??=, reduce: PositionalFieldName
			reduce(107), // ???=, reduce: PositionalFieldName
			reduce(107), // |=, reduce: PositionalFieldName
			reduce(107), // &=, reduce: PositionalFieldName
			reduce(107), // ^=, reduce: PositionalFieldName
			reduce(107), // <<=, reduce: PositionalFieldName
			reduce(107), // >>=, reduce: PositionalFieldName
			reduce(107), // >>>=, reduce: PositionalFieldName
			reduce(107), // +=, reduce: PositionalFieldName
			reduce(107), // .=, reduce: PositionalFieldName
			reduce(107), // -=, reduce: PositionalFieldName
			reduce(107), // *=, reduce: PositionalFieldName
			reduce(107), // /=, reduce: PositionalFieldName
			reduce(107), // //=, reduce: PositionalFieldName
			reduce(107), // %=, reduce: PositionalFieldName
			reduce(107), // **=, reduce: PositionalFieldName
			reduce(107), // ?, reduce: PositionalFieldName
			nil,         // :
			reduce(107), // ||, reduce: PositionalFieldName
			reduce(107), // ^^, reduce: PositionalFieldName
			reduce(107), // &&, reduce: PositionalFieldName
			reduce(107), // ??, reduce: PositionalFieldName
			reduce(107), // ???, reduce: PositionalFieldName
			reduce(107), // =~, reduce: PositionalFieldName
			reduce(107), // !=~, reduce: PositionalFieldName
			reduce(107), // ==, reduce: PositionalFieldName
			reduce(107), // !=, reduce: PositionalFieldName
			reduce(107), // <=>, reduce: PositionalFieldName
			reduce(107), // >=, reduce: PositionalFieldName
			reduce(107), // <, reduce: PositionalFieldName
			reduce(107), // <=, reduce: PositionalFieldName
			reduce(107), // ^, reduce: PositionalFieldName
			reduce(107), // &, reduce: PositionalFieldName
			reduce(107), // <<, reduce: PositionalFieldName
			reduce(107), // >>>, reduce: PositionalFieldName
			reduce(107), // +, reduce: PositionalFieldName
			reduce(107), // -, reduce: PositionalFieldName
			reduce(107), // .+, reduce: PositionalFieldName
			reduce(107), // .-, reduce: PositionalFieldName
			reduce(107), // *, reduce: PositionalFieldName
			reduce(107), // /, reduce: PositionalFieldName
			reduce(107), // //, reduce: PositionalFieldName
			reduce(107), // %, reduce: PositionalFieldName
			reduce(107), // .*, reduce: PositionalFieldName
			reduce(107), // ./, reduce: PositionalFieldName
			reduce(107), // .//, reduce: PositionalFieldName
			reduce(107), // ., reduce: PositionalFieldName
			nil,         // !
			nil,         // ~
			reduce(107), // **, reduce: PositionalFieldName
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(107), // [, reduce: PositionalFieldName
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(107), // [[, reduce: PositionalFieldName
			reduce(107), // [[[, reduce: PositionalFieldName
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3908
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4237), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3909
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(953),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(990),  // (
			nil,         // )
			nil,         // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2302), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3910
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(206), // ;, reduce: PowTerm
			reduce(206), // {, reduce: PowTerm
			reduce(206), // }, reduce: PowTerm
			reduce(206), // =, reduce: PowTerm
			nil,         // unset
			nil,         // filter
			reduce(206), // >, reduce: PowTerm
			reduce(206), // >>, reduce: PowTerm
			reduce(206), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(206), // ||=, reduce: PowTerm
			reduce(206), // ^^=, reduce: PowTerm
			reduce(206), // &&=, reduce: PowTerm
			reduce(206), // ??=, reduce: PowTerm
			reduce(206), // ???=, reduce: PowTerm
			reduce(206), // |=, reduce: PowTerm
			reduce(206), // &=, reduce: PowTerm
			reduce(206), // ^=, reduce: PowTerm
			reduce(206), // <<=, reduce: PowTerm
			reduce(206), // >>=, reduce: PowTerm
			reduce(206), // >>>=, reduce: PowTerm
			reduce(206), // +=, reduce: PowTerm
			reduce(206), // .=, reduce: PowTerm
			reduce(206), // -=, reduce: PowTerm
			reduce(206), // *=, reduce: PowTerm
			reduce(206), // /=, reduce: PowTerm
			reduce(206), // //=, reduce: PowTerm
			reduce(206), // %=, reduce: PowTerm
			reduce(206), // **=, reduce: PowTerm
			reduce(206), // ?, reduce: PowTerm
			nil,         // :
			reduce(206), // ||, reduce: PowTerm
			reduce(206), // ^^, reduce: PowTerm
			reduce(206), // &&, reduce: PowTerm
			reduce(206), // ??, reduce: PowTerm
			reduce(206), // ???, reduce: PowTerm
			reduce(206), // =~, reduce: PowTerm
			reduce(206), // !=~, reduce: PowTerm
			reduce(206), // ==, reduce: PowTerm
			reduce(206), // !=, reduce: PowTerm
			reduce(206), // <=>, reduce: PowTerm
			reduce(206), // >=, reduce: PowTerm
			reduce(206), // <, reduce: PowTerm
			reduce(206), // <=, reduce: PowTerm
			reduce(206), // ^, reduce: PowTerm
			reduce(206), // &, reduce: PowTerm
			reduce(206), // <<, reduce: PowTerm
			reduce(206), // >>>, reduce: PowTerm
			reduce(206), // +, reduce: PowTerm
			reduce(206), // -, reduce: PowTerm
			reduce(206), // .+, reduce: PowTerm
			reduce(206), // .-, reduce: PowTerm
			reduce(206), // *, reduce: PowTerm
			reduce(206), // /, reduce: PowTerm
			reduce(206), // //, reduce: PowTerm
			reduce(206), // %, reduce: PowTerm
			reduce(206), // .*, reduce: PowTerm
			reduce(206), // ./, reduce: PowTerm
			reduce(206), // .//, reduce: PowTerm
			reduce(206), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3911
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(205), // ;, reduce: PowTerm
			reduce(205), // {, reduce: PowTerm
			reduce(205), // }, reduce: PowTerm
			reduce(205), // =, reduce: PowTerm
			nil,         // unset
			nil,         // filter
			reduce(205), // >, reduce: PowTerm
			reduce(205), // >>, reduce: PowTerm
			reduce(205), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(205), // ||=, reduce: PowTerm
			reduce(205), // ^^=, reduce: PowTerm
			reduce(205), // &&=, reduce: PowTerm
			reduce(205), // ??=, reduce: PowTerm
			reduce(205), // ???=, reduce: PowTerm
			reduce(205), // |=, reduce: PowTerm
			reduce(205), // &=, reduce: PowTerm
			reduce(205), // ^=, reduce: PowTerm
			reduce(205), // <<=, reduce: PowTerm
			reduce(205), // >>=, reduce: PowTerm
			reduce(205), // >>>=, reduce: PowTerm
			reduce(205), // +=, reduce: PowTerm
			reduce(205), // .=, reduce: PowTerm
			reduce(205), // -=, reduce: PowTerm
			reduce(205), // *=, reduce: PowTerm
			reduce(205), // /=, reduce: PowTerm
			reduce(205), // //=, reduce: PowTerm
			reduce(205), // %=, reduce: PowTerm
			reduce(205), // **=, reduce: PowTerm
			reduce(205), // ?, reduce: PowTerm
			nil,         // :
			reduce(205), // ||, reduce: PowTerm
			reduce(205), // ^^, reduce: PowTerm
			reduce(205), // &&, reduce: PowTerm
			reduce(205), // ??, reduce: PowTerm
			reduce(205), // ???, reduce: PowTerm
			reduce(205), // =~, reduce: PowTerm
			reduce(205), // !=~, reduce: PowTerm
			reduce(205), // ==, reduce: PowTerm
			reduce(205), // !=, reduce: PowTerm
			reduce(205), // <=>, reduce: PowTerm
			reduce(205), // >=, reduce: PowTerm
			reduce(205), // <, reduce: PowTerm
			reduce(205), // <=, reduce: PowTerm
			reduce(205), // ^, reduce: PowTerm
			reduce(205), // &, reduce: PowTerm
			reduce(205), // <<, reduce: PowTerm
			reduce(205), // >>>, reduce: PowTerm
			reduce(205), // +, reduce: PowTerm
			reduce(205), // -, reduce: PowTerm
			reduce(205), // .+, reduce: PowTerm
			reduce(205), // .-, reduce: PowTerm
			reduce(205), // *, reduce: PowTerm
			reduce(205), // /, reduce: PowTerm
			reduce(205), // //, reduce: PowTerm
			reduce(205), // %, reduce: PowTerm
			reduce(205), // .*, reduce: PowTerm
			reduce(205), // ./, reduce: PowTerm
			reduce(205), // .//, reduce: PowTerm
			reduce(205), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3912
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(260), // ;, reduce: ArrayOrMapIndexAccess
			reduce(260), // {, reduce: ArrayOrMapIndexAccess
			reduce(260), // }, reduce: ArrayOrMapIndexAccess
			reduce(260), // =, reduce: ArrayOrMapIndexAccess
			nil,         // unset
			nil,         // filter
			reduce(260), // >, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>, reduce: ArrayOrMapIndexAccess
			reduce(260), // |, reduce: ArrayOrMapIndexAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(260), // ||=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^=, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???=, reduce: ArrayOrMapIndexAccess
			reduce(260), // |=, reduce: ArrayOrMapIndexAccess
			reduce(260), // &=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<=, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>=, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>=, reduce: ArrayOrMapIndexAccess
			reduce(260), // +=, reduce: ArrayOrMapIndexAccess
			reduce(260), // .=, reduce: ArrayOrMapIndexAccess
			reduce(260), // -=, reduce: ArrayOrMapIndexAccess
			reduce(260), // *=, reduce: ArrayOrMapIndexAccess
			reduce(260), // /=, reduce: ArrayOrMapIndexAccess
			reduce(260), // //=, reduce: ArrayOrMapIndexAccess
			reduce(260), // %=, reduce: ArrayOrMapIndexAccess
			reduce(260), // **=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ?, reduce: ArrayOrMapIndexAccess
			nil,         // :
			reduce(260), // ||, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???, reduce: ArrayOrMapIndexAccess
			reduce(260), // =~, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=~, reduce: ArrayOrMapIndexAccess
			reduce(260), // ==, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=>, reduce: ArrayOrMapIndexAccess
			reduce(260), // >=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>, reduce: ArrayOrMapIndexAccess
			reduce(260), // +, reduce: ArrayOrMapIndexAccess
			reduce(260), // -, reduce: ArrayOrMapIndexAccess
			reduce(260), // .+, reduce: ArrayOrMapIndexAccess
			reduce(260), // .-, reduce: ArrayOrMapIndexAccess
			reduce(260), // *, reduce: ArrayOrMapIndexAccess
			reduce(260), // /, reduce: ArrayOrMapIndexAccess
			reduce(260), // //, reduce: ArrayOrMapIndexAccess
			reduce(260), // %, reduce: ArrayOrMapIndexAccess
			reduce(260), // .*, reduce: ArrayOrMapIndexAccess
			reduce(260), // ./, reduce: ArrayOrMapIndexAccess
			reduce(260), // .//, reduce: ArrayOrMapIndexAccess
			reduce(260), // ., reduce: ArrayOrMapIndexAccess
			nil,         // !
			nil,         // ~
			reduce(260), // **, reduce: ArrayOrMapIndexAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(260), // [, reduce: ArrayOrMapIndexAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(260), // [[, reduce: ArrayOrMapIndexAccess
			reduce(260), // [[[, reduce: ArrayOrMapIndexAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3913
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4240), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3914
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4241), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3915
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(266), // ;, reduce: ArraySliceAccess
			reduce(266), // {, reduce: ArraySliceAccess
			reduce(266), // }, reduce: ArraySliceAccess
			reduce(266), // =, reduce: ArraySliceAccess
			nil,         // unset
			nil,         // filter
			reduce(266), // >, reduce: ArraySliceAccess
			reduce(266), // >>, reduce: ArraySliceAccess
			reduce(266), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(266), // ||=, reduce: ArraySliceAccess
			reduce(266), // ^^=, reduce: ArraySliceAccess
			reduce(266), // &&=, reduce: ArraySliceAccess
			reduce(266), // ??=, reduce: ArraySliceAccess
			reduce(266), // ???=, reduce: ArraySliceAccess
			reduce(266), // |=, reduce: ArraySliceAccess
			reduce(266), // &=, reduce: ArraySliceAccess
			reduce(266), // ^=, reduce: ArraySliceAccess
			reduce(266), // <<=, reduce: ArraySliceAccess
			reduce(266), // >>=, reduce: ArraySliceAccess
			reduce(266), // >>>=, reduce: ArraySliceAccess
			reduce(266), // +=, reduce: ArraySliceAccess
			reduce(266), // .=, reduce: ArraySliceAccess
			reduce(266), // -=, reduce: ArraySliceAccess
			reduce(266), // *=, reduce: ArraySliceAccess
			reduce(266), // /=, reduce: ArraySliceAccess
			reduce(266), // //=, reduce: ArraySliceAccess
			reduce(266), // %=, reduce: ArraySliceAccess
			reduce(266), // **=, reduce: ArraySliceAccess
			reduce(266), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(266), // ||, reduce: ArraySliceAccess
			reduce(266), // ^^, reduce: ArraySliceAccess
			reduce(266), // &&, reduce: ArraySliceAccess
			reduce(266), // ??, reduce: ArraySliceAccess
			reduce(266), // ???, reduce: ArraySliceAccess
			reduce(266), // =~, reduce: ArraySliceAccess
			reduce(266), // !=~, reduce: ArraySliceAccess
			reduce(266), // ==, reduce: ArraySliceAccess
			reduce(266), // !=, reduce: ArraySliceAccess
			reduce(266), // <=>, reduce: ArraySliceAccess
			reduce(266), // >=, reduce: ArraySliceAccess
			reduce(266), // <, reduce: ArraySliceAccess
			reduce(266), // <=, reduce: ArraySliceAccess
			reduce(266), // ^, reduce: ArraySliceAccess
			reduce(266), // &, reduce: ArraySliceAccess
			reduce(266), // <<, reduce: ArraySliceAccess
			reduce(266), // >>>, reduce: ArraySliceAccess
			reduce(266), // +, reduce: ArraySliceAccess
			reduce(266), // -, reduce: ArraySliceAccess
			reduce(266), // .+, reduce: ArraySliceAccess
			reduce(266), // .-, reduce: ArraySliceAccess
			reduce(266), // *, reduce: ArraySliceAccess
			reduce(266), // /, reduce: ArraySliceAccess
			reduce(266), // //, reduce: ArraySliceAccess
			reduce(266), // %, reduce: ArraySliceAccess
			reduce(266), // .*, reduce: ArraySliceAccess
			reduce(266), // ./, reduce: ArraySliceAccess
			reduce(266), // .//, reduce: ArraySliceAccess
			reduce(266), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(266), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(266), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(266), // [[, reduce: ArraySliceAccess
			reduce(266), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3916
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4242), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3917
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4243), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3918
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(254), // ;, reduce: ENV
			reduce(254), // {, reduce: ENV
			reduce(254), // }, reduce: ENV
			reduce(254), // =, reduce: ENV
			nil,         // unset
			nil,         // filter
			reduce(254), // >, reduce: ENV
			reduce(254), // >>, reduce: ENV
			reduce(254), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(254), // ||=, reduce: ENV
			reduce(254), // ^^=, reduce: ENV
			reduce(254), // &&=, reduce: ENV
			reduce(254), // ??=, reduce: ENV
			reduce(254), // ???=, reduce: ENV
			reduce(254), // |=, reduce: ENV
			reduce(254), // &=, reduce: ENV
			reduce(254), // ^=, reduce: ENV
			reduce(254), // <<=, reduce: ENV
			reduce(254), // >>=, reduce: ENV
			reduce(254), // >>>=, reduce: ENV
			reduce(254), // +=, reduce: ENV
			reduce(254), // .=, reduce: ENV
			reduce(254), // -=, reduce: ENV
			reduce(254), // *=, reduce: ENV
			reduce(254), // /=, reduce: ENV
			reduce(254), // //=, reduce: ENV
			reduce(254), // %=, reduce: ENV
			reduce(254), // **=, reduce: ENV
			reduce(254), // ?, reduce: ENV
			nil,         // :
			reduce(254), // ||, reduce: ENV
			reduce(254), // ^^, reduce: ENV
			reduce(254), // &&, reduce: ENV
			reduce(254), // ??, reduce: ENV
			reduce(254), // ???, reduce: ENV
			reduce(254), // =~, reduce: ENV
			reduce(254), // !=~, reduce: ENV
			reduce(254), // ==, reduce: ENV
			reduce(254), // !=, reduce: ENV
			reduce(254), // <=>, reduce: ENV
			reduce(254), // >=, reduce: ENV
			reduce(254), // <, reduce: ENV
			reduce(254), // <=, reduce: ENV
			reduce(254), // ^, reduce: ENV
			reduce(254), // &, reduce: ENV
			reduce(254), // <<, reduce: ENV
			reduce(254), // >>>, reduce: ENV
			reduce(254), // +, reduce: ENV
			reduce(254), // -, reduce: ENV
			reduce(254), // .+, reduce: ENV
			reduce(254), // .-, reduce: ENV
			reduce(254), // *, reduce: ENV
			reduce(254), // /, reduce: ENV
			reduce(254), // //, reduce: ENV
			reduce(254), // %, reduce: ENV
			reduce(254), // .*, reduce: ENV
			reduce(254), // ./, reduce: ENV
			reduce(254), // .//, reduce: ENV
			reduce(254), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(254), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(254), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(254), // [[, reduce: ENV
			reduce(254), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3919
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(269), // ;, reduce: FunctionCallsite
			reduce(269), // {, reduce: FunctionCallsite
			reduce(269), // }, reduce: FunctionCallsite
			reduce(269), // =, reduce: FunctionCallsite
			nil,         // unset
			nil,         // filter
			reduce(269), // >, reduce: FunctionCallsite
			reduce(269), // >>, reduce: FunctionCallsite
			reduce(269), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(269), // ||=, reduce: FunctionCallsite
			reduce(269), // ^^=, reduce: FunctionCallsite
			reduce(269), // &&=, reduce: FunctionCallsite
			reduce(269), // ??=, reduce: FunctionCallsite
			reduce(269), // ???=, reduce: FunctionCallsite
			reduce(269), // |=, reduce: FunctionCallsite
			reduce(269), // &=, reduce: FunctionCallsite
			reduce(269), // ^=, reduce: FunctionCallsite
			reduce(269), // <<=, reduce: FunctionCallsite
			reduce(269), // >>=, reduce: FunctionCallsite
			reduce(269), // >>>=, reduce: FunctionCallsite
			reduce(269), // +=, reduce: FunctionCallsite
			reduce(269), // .=, reduce: FunctionCallsite
			reduce(269), // -=, reduce: FunctionCallsite
			reduce(269), // *=, reduce: FunctionCallsite
			reduce(269), // /=, reduce: FunctionCallsite
			reduce(269), // //=, reduce: FunctionCallsite
			reduce(269), // %=, reduce: FunctionCallsite
			reduce(269), // **=, reduce: FunctionCallsite
			reduce(269), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(269), // ||, reduce: FunctionCallsite
			reduce(269), // ^^, reduce: FunctionCallsite
			reduce(269), // &&, reduce: FunctionCallsite
			reduce(269), // ??, reduce: FunctionCallsite
			reduce(269), // ???, reduce: FunctionCallsite
			reduce(269), // =~, reduce: FunctionCallsite
			reduce(269), // !=~, reduce: FunctionCallsite
			reduce(269), // ==, reduce: FunctionCallsite
			reduce(269), // !=, reduce: FunctionCallsite
			reduce(269), // <=>, reduce: FunctionCallsite
			reduce(269), // >=, reduce: FunctionCallsite
			reduce(269), // <, reduce: FunctionCallsite
			reduce(269), // <=, reduce: FunctionCallsite
			reduce(269), // ^, reduce: FunctionCallsite
			reduce(269), // &, reduce: FunctionCallsite
			reduce(269), // <<, reduce: FunctionCallsite
			reduce(269), // >>>, reduce: FunctionCallsite
			reduce(269), // +, reduce: FunctionCallsite
			reduce(269), // -, reduce: FunctionCallsite
			reduce(269), // .+, reduce: FunctionCallsite
			reduce(269), // .-, reduce: FunctionCallsite
			reduce(269), // *, reduce: FunctionCallsite
			reduce(269), // /, reduce: FunctionCallsite
			reduce(269), // //, reduce: FunctionCallsite
			reduce(269), // %, reduce: FunctionCallsite
			reduce(269), // .*, reduce: FunctionCallsite
			reduce(269), // ./, reduce: FunctionCallsite
			reduce(269), // .//, reduce: FunctionCallsite
			reduce(269), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(269), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(269), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(269), // [[, reduce: FunctionCallsite
			reduce(269), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3920
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4244), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3921
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(276), // ;, reduce: SubroutineCallsite
			nil,         // {
			reduce(276), // }, reduce: SubroutineCallsite
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3922
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4245), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3923
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4247), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3924
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2042), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3925
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3926
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(3561), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3927
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3928
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(3565), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4252), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3929
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(312), // ;, reduce: TripleForContinuation
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(312), // ,, reduce: TripleForContinuation
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3930
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4255), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4256), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3931
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4257), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3932
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4258), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3933
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(108), // $, reduce: PositionalFieldValue
			nil,         // empty
			reduce(108), // ;, reduce: PositionalFieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(108), // >, reduce: PositionalFieldValue
			reduce(108), // >>, reduce: PositionalFieldValue
			reduce(108), // |, reduce: PositionalFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(108), // ,, reduce: PositionalFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(108), // ?, reduce: PositionalFieldValue
			nil,         // :
			reduce(108), // ||, reduce: PositionalFieldValue
			reduce(108), // ^^, reduce: PositionalFieldValue
			reduce(108), // &&, reduce: PositionalFieldValue
			reduce(108), // ??, reduce: PositionalFieldValue
			reduce(108), // ???, reduce: PositionalFieldValue
			reduce(108), // =~, reduce: PositionalFieldValue
			reduce(108), // !=~, reduce: PositionalFieldValue
			reduce(108), // ==, reduce: PositionalFieldValue
			reduce(108), // !=, reduce: PositionalFieldValue
			reduce(108), // <=>, reduce: PositionalFieldValue
			reduce(108), // >=, reduce: PositionalFieldValue
			reduce(108), // <, reduce: PositionalFieldValue
			reduce(108), // <=, reduce: PositionalFieldValue
			reduce(108), // ^, reduce: PositionalFieldValue
			reduce(108), // &, reduce: PositionalFieldValue
			reduce(108), // <<, reduce: PositionalFieldValue
			reduce(108), // >>>, reduce: PositionalFieldValue
			reduce(108), // +, reduce: PositionalFieldValue
			reduce(108), // -, reduce: PositionalFieldValue
			reduce(108), // .+, reduce: PositionalFieldValue
			reduce(108), // .-, reduce: PositionalFieldValue
			reduce(108), // *, reduce: PositionalFieldValue
			reduce(108), // /, reduce: PositionalFieldValue
			reduce(108), // //, reduce: PositionalFieldValue
			reduce(108), // %, reduce: PositionalFieldValue
			reduce(108), // .*, reduce: PositionalFieldValue
			reduce(108), // ./, reduce: PositionalFieldValue
			reduce(108), // .//, reduce: PositionalFieldValue
			reduce(108), // ., reduce: PositionalFieldValue
			nil,         // !
			nil,         // ~
			reduce(108), // **, reduce: PositionalFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(108), // [, reduce: PositionalFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(108), // [[, reduce: PositionalFieldValue
			reduce(108), // [[[, reduce: PositionalFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3934
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(149), // $, reduce: TernaryTerm
			nil,         // empty
			reduce(149), // ;, reduce: TernaryTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(149), // ,, reduce: TernaryTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3935
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4259), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3936
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(265), // $, reduce: ArraySliceAccess
			nil,         // empty
			reduce(265), // ;, reduce: ArraySliceAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(265), // >, reduce: ArraySliceAccess
			reduce(265), // >>, reduce: ArraySliceAccess
			reduce(265), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(265), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(265), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(265), // ||, reduce: ArraySliceAccess
			reduce(265), // ^^, reduce: ArraySliceAccess
			reduce(265), // &&, reduce: ArraySliceAccess
			reduce(265), // ??, reduce: ArraySliceAccess
			reduce(265), // ???, reduce: ArraySliceAccess
			reduce(265), // =~, reduce: ArraySliceAccess
			reduce(265), // !=~, reduce: ArraySliceAccess
			reduce(265), // ==, reduce: ArraySliceAccess
			reduce(265), // !=, reduce: ArraySliceAccess
			reduce(265), // <=>, reduce: ArraySliceAccess
			reduce(265), // >=, reduce: ArraySliceAccess
			reduce(265), // <, reduce: ArraySliceAccess
			reduce(265), // <=, reduce: ArraySliceAccess
			reduce(265), // ^, reduce: ArraySliceAccess
			reduce(265), // &, reduce: ArraySliceAccess
			reduce(265), // <<, reduce: ArraySliceAccess
			reduce(265), // >>>, reduce: ArraySliceAccess
			reduce(265), // +, reduce: ArraySliceAccess
			reduce(265), // -, reduce: ArraySliceAccess
			reduce(265), // .+, reduce: ArraySliceAccess
			reduce(265), // .-, reduce: ArraySliceAccess
			reduce(265), // *, reduce: ArraySliceAccess
			reduce(265), // /, reduce: ArraySliceAccess
			reduce(265), // //, reduce: ArraySliceAccess
			reduce(265), // %, reduce: ArraySliceAccess
			reduce(265), // .*, reduce: ArraySliceAccess
			reduce(265), // ./, reduce: ArraySliceAccess
			reduce(265), // .//, reduce: ArraySliceAccess
			reduce(265), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(265), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(265), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(265), // [[, reduce: ArraySliceAccess
			reduce(265), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3937
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(264), // $, reduce: ArraySliceAccess
			nil,         // empty
			reduce(264), // ;, reduce: ArraySliceAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(264), // >, reduce: ArraySliceAccess
			reduce(264), // >>, reduce: ArraySliceAccess
			reduce(264), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(264), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(264), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(264), // ||, reduce: ArraySliceAccess
			reduce(264), // ^^, reduce: ArraySliceAccess
			reduce(264), // &&, reduce: ArraySliceAccess
			reduce(264), // ??, reduce: ArraySliceAccess
			reduce(264), // ???, reduce: ArraySliceAccess
			reduce(264), // =~, reduce: ArraySliceAccess
			reduce(264), // !=~, reduce: ArraySliceAccess
			reduce(264), // ==, reduce: ArraySliceAccess
			reduce(264), // !=, reduce: ArraySliceAccess
			reduce(264), // <=>, reduce: ArraySliceAccess
			reduce(264), // >=, reduce: ArraySliceAccess
			reduce(264), // <, reduce: ArraySliceAccess
			reduce(264), // <=, reduce: ArraySliceAccess
			reduce(264), // ^, reduce: ArraySliceAccess
			reduce(264), // &, reduce: ArraySliceAccess
			reduce(264), // <<, reduce: ArraySliceAccess
			reduce(264), // >>>, reduce: ArraySliceAccess
			reduce(264), // +, reduce: ArraySliceAccess
			reduce(264), // -, reduce: ArraySliceAccess
			reduce(264), // .+, reduce: ArraySliceAccess
			reduce(264), // .-, reduce: ArraySliceAccess
			reduce(264), // *, reduce: ArraySliceAccess
			reduce(264), // /, reduce: ArraySliceAccess
			reduce(264), // //, reduce: ArraySliceAccess
			reduce(264), // %, reduce: ArraySliceAccess
			reduce(264), // .*, reduce: ArraySliceAccess
			reduce(264), // ./, reduce: ArraySliceAccess
			reduce(264), // .//, reduce: ArraySliceAccess
			reduce(264), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(264), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(264), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(264), // [[, reduce: ArraySliceAccess
			reduce(264), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3938
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(261), // $, reduce: ArrayOrMapPositionalNameAccess
			nil,         // empty
			reduce(261), // ;, reduce: ArrayOrMapPositionalNameAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(261), // >, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |, reduce: ArrayOrMapPositionalNameAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(261), // ,, reduce: ArrayOrMapPositionalNameAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(261), // ?, reduce: ArrayOrMapPositionalNameAccess
			nil,         // :
			reduce(261), // ||, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // =~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ==, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .+, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .-, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .*, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ./, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .//, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ., reduce: ArrayOrMapPositionalNameAccess
			nil,         // !
			nil,         // ~
			reduce(261), // **, reduce: ArrayOrMapPositionalNameAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(261), // [, reduce: ArrayOrMapPositionalNameAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(261), // [[, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // [[[, reduce: ArrayOrMapPositionalNameAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3939
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4260), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3940
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(324), // $, reduce: UnnamedFunctionDefinition
			nil,         // empty
			reduce(324), // ;, reduce: UnnamedFunctionDefinition
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(324), // >, reduce: UnnamedFunctionDefinition
			reduce(324), // >>, reduce: UnnamedFunctionDefinition
			reduce(324), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(324), // ,, reduce: UnnamedFunctionDefinition
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(324), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(324), // ||, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^, reduce: UnnamedFunctionDefinition
			reduce(324), // &&, reduce: UnnamedFunctionDefinition
			reduce(324), // ??, reduce: UnnamedFunctionDefinition
			reduce(324), // ???, reduce: UnnamedFunctionDefinition
			reduce(324), // =~, reduce: UnnamedFunctionDefinition
			reduce(324), // !=~, reduce: UnnamedFunctionDefinition
			reduce(324), // ==, reduce: UnnamedFunctionDefinition
			reduce(324), // !=, reduce: UnnamedFunctionDefinition
			reduce(324), // <=>, reduce: UnnamedFunctionDefinition
			reduce(324), // >=, reduce: UnnamedFunctionDefinition
			reduce(324), // <, reduce: UnnamedFunctionDefinition
			reduce(324), // <=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^, reduce: UnnamedFunctionDefinition
			reduce(324), // &, reduce: UnnamedFunctionDefinition
			reduce(324), // <<, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>, reduce: UnnamedFunctionDefinition
			reduce(324), // +, reduce: UnnamedFunctionDefinition
			reduce(324), // -, reduce: UnnamedFunctionDefinition
			reduce(324), // .+, reduce: UnnamedFunctionDefinition
			reduce(324), // .-, reduce: UnnamedFunctionDefinition
			reduce(324), // *, reduce: UnnamedFunctionDefinition
			reduce(324), // /, reduce: UnnamedFunctionDefinition
			reduce(324), // //, reduce: UnnamedFunctionDefinition
			reduce(324), // %, reduce: UnnamedFunctionDefinition
			reduce(324), // .*, reduce: UnnamedFunctionDefinition
			reduce(324), // ./, reduce: UnnamedFunctionDefinition
			reduce(324), // .//, reduce: UnnamedFunctionDefinition
			reduce(324), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(324), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(324), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(324), // [[, reduce: UnnamedFunctionDefinition
			reduce(324), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3941
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S3942
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3943
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(108), // $, reduce: PositionalFieldValue
			nil,         // empty
			reduce(108), // ;, reduce: PositionalFieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(108), // >, reduce: PositionalFieldValue
			reduce(108), // >>, reduce: PositionalFieldValue
			reduce(108), // |, reduce: PositionalFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(108), // ?, reduce: PositionalFieldValue
			nil,         // :
			reduce(108), // ||, reduce: PositionalFieldValue
			reduce(108), // ^^, reduce: PositionalFieldValue
			reduce(108), // &&, reduce: PositionalFieldValue
			reduce(108), // ??, reduce: PositionalFieldValue
			reduce(108), // ???, reduce: PositionalFieldValue
			reduce(108), // =~, reduce: PositionalFieldValue
			reduce(108), // !=~, reduce: PositionalFieldValue
			reduce(108), // ==, reduce: PositionalFieldValue
			reduce(108), // !=, reduce: PositionalFieldValue
			reduce(108), // <=>, reduce: PositionalFieldValue
			reduce(108), // >=, reduce: PositionalFieldValue
			reduce(108), // <, reduce: PositionalFieldValue
			reduce(108), // <=, reduce: PositionalFieldValue
			reduce(108), // ^, reduce: PositionalFieldValue
			reduce(108), // &, reduce: PositionalFieldValue
			reduce(108), // <<, reduce: PositionalFieldValue
			reduce(108), // >>>, reduce: PositionalFieldValue
			reduce(108), // +, reduce: PositionalFieldValue
			reduce(108), // -, reduce: PositionalFieldValue
			reduce(108), // .+, reduce: PositionalFieldValue
			reduce(108), // .-, reduce: PositionalFieldValue
			reduce(108), // *, reduce: PositionalFieldValue
			reduce(108), // /, reduce: PositionalFieldValue
			reduce(108), // //, reduce: PositionalFieldValue
			reduce(108), // %, reduce: PositionalFieldValue
			reduce(108), // .*, reduce: PositionalFieldValue
			reduce(108), // ./, reduce: PositionalFieldValue
			reduce(108), // .//, reduce: PositionalFieldValue
			reduce(108), // ., reduce: PositionalFieldValue
			nil,         // !
			nil,         // ~
			reduce(108), // **, reduce: PositionalFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(108), // [, reduce: PositionalFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(108), // [[, reduce: PositionalFieldValue
			reduce(108), // [[[, reduce: PositionalFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3944
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(149), // $, reduce: TernaryTerm
			nil,         // empty
			reduce(149), // ;, reduce: TernaryTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3945
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4263), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3946
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(265), // $, reduce: ArraySliceAccess
			nil,         // empty
			reduce(265), // ;, reduce: ArraySliceAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(265), // >, reduce: ArraySliceAccess
			reduce(265), // >>, reduce: ArraySliceAccess
			reduce(265), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(265), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(265), // ||, reduce: ArraySliceAccess
			reduce(265), // ^^, reduce: ArraySliceAccess
			reduce(265), // &&, reduce: ArraySliceAccess
			reduce(265), // ??, reduce: ArraySliceAccess
			reduce(265), // ???, reduce: ArraySliceAccess
			reduce(265), // =~, reduce: ArraySliceAccess
			reduce(265), // !=~, reduce: ArraySliceAccess
			reduce(265), // ==, reduce: ArraySliceAccess
			reduce(265), // !=, reduce: ArraySliceAccess
			reduce(265), // <=>, reduce: ArraySliceAccess
			reduce(265), // >=, reduce: ArraySliceAccess
			reduce(265), // <, reduce: ArraySliceAccess
			reduce(265), // <=, reduce: ArraySliceAccess
			reduce(265), // ^, reduce: ArraySliceAccess
			reduce(265), // &, reduce: ArraySliceAccess
			reduce(265), // <<, reduce: ArraySliceAccess
			reduce(265), // >>>, reduce: ArraySliceAccess
			reduce(265), // +, reduce: ArraySliceAccess
			reduce(265), // -, reduce: ArraySliceAccess
			reduce(265), // .+, reduce: ArraySliceAccess
			reduce(265), // .-, reduce: ArraySliceAccess
			reduce(265), // *, reduce: ArraySliceAccess
			reduce(265), // /, reduce: ArraySliceAccess
			reduce(265), // //, reduce: ArraySliceAccess
			reduce(265), // %, reduce: ArraySliceAccess
			reduce(265), // .*, reduce: ArraySliceAccess
			reduce(265), // ./, reduce: ArraySliceAccess
			reduce(265), // .//, reduce: ArraySliceAccess
			reduce(265), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(265), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(265), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(265), // [[, reduce: ArraySliceAccess
			reduce(265), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3947
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(264), // $, reduce: ArraySliceAccess
			nil,         // empty
			reduce(264), // ;, reduce: ArraySliceAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(264), // >, reduce: ArraySliceAccess
			reduce(264), // >>, reduce: ArraySliceAccess
			reduce(264), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(264), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(264), // ||, reduce: ArraySliceAccess
			reduce(264), // ^^, reduce: ArraySliceAccess
			reduce(264), // &&, reduce: ArraySliceAccess
			reduce(264), // ??, reduce: ArraySliceAccess
			reduce(264), // ???, reduce: ArraySliceAccess
			reduce(264), // =~, reduce: ArraySliceAccess
			reduce(264), // !=~, reduce: ArraySliceAccess
			reduce(264), // ==, reduce: ArraySliceAccess
			reduce(264), // !=, reduce: ArraySliceAccess
			reduce(264), // <=>, reduce: ArraySliceAccess
			reduce(264), // >=, reduce: ArraySliceAccess
			reduce(264), // <, reduce: ArraySliceAccess
			reduce(264), // <=, reduce: ArraySliceAccess
			reduce(264), // ^, reduce: ArraySliceAccess
			reduce(264), // &, reduce: ArraySliceAccess
			reduce(264), // <<, reduce: ArraySliceAccess
			reduce(264), // >>>, reduce: ArraySliceAccess
			reduce(264), // +, reduce: ArraySliceAccess
			reduce(264), // -, reduce: ArraySliceAccess
			reduce(264), // .+, reduce: ArraySliceAccess
			reduce(264), // .-, reduce: ArraySliceAccess
			reduce(264), // *, reduce: ArraySliceAccess
			reduce(264), // /, reduce: ArraySliceAccess
			reduce(264), // //, reduce: ArraySliceAccess
			reduce(264), // %, reduce: ArraySliceAccess
			reduce(264), // .*, reduce: ArraySliceAccess
			reduce(264), // ./, reduce: ArraySliceAccess
			reduce(264), // .//, reduce: ArraySliceAccess
			reduce(264), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(264), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(264), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(264), // [[, reduce: ArraySliceAccess
			reduce(264), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3948
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(261), // $, reduce: ArrayOrMapPositionalNameAccess
			nil,         // empty
			reduce(261), // ;, reduce: ArrayOrMapPositionalNameAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(261), // >, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |, reduce: ArrayOrMapPositionalNameAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(261), // ?, reduce: ArrayOrMapPositionalNameAccess
			nil,         // :
			reduce(261), // ||, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // =~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ==, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .+, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .-, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .*, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ./, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .//, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ., reduce: ArrayOrMapPositionalNameAccess
			nil,         // !
			nil,         // ~
			reduce(261), // **, reduce: ArrayOrMapPositionalNameAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(261), // [, reduce: ArrayOrMapPositionalNameAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(261), // [[, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // [[[, reduce: ArrayOrMapPositionalNameAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3949
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4264), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3950
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(324), // $, reduce: UnnamedFunctionDefinition
			nil,         // empty
			reduce(324), // ;, reduce: UnnamedFunctionDefinition
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(324), // >, reduce: UnnamedFunctionDefinition
			reduce(324), // >>, reduce: UnnamedFunctionDefinition
			reduce(324), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(324), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(324), // ||, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^, reduce: UnnamedFunctionDefinition
			reduce(324), // &&, reduce: UnnamedFunctionDefinition
			reduce(324), // ??, reduce: UnnamedFunctionDefinition
			reduce(324), // ???, reduce: UnnamedFunctionDefinition
			reduce(324), // =~, reduce: UnnamedFunctionDefinition
			reduce(324), // !=~, reduce: UnnamedFunctionDefinition
			reduce(324), // ==, reduce: UnnamedFunctionDefinition
			reduce(324), // !=, reduce: UnnamedFunctionDefinition
			reduce(324), // <=>, reduce: UnnamedFunctionDefinition
			reduce(324), // >=, reduce: UnnamedFunctionDefinition
			reduce(324), // <, reduce: UnnamedFunctionDefinition
			reduce(324), // <=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^, reduce: UnnamedFunctionDefinition
			reduce(324), // &, reduce: UnnamedFunctionDefinition
			reduce(324), // <<, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>, reduce: UnnamedFunctionDefinition
			reduce(324), // +, reduce: UnnamedFunctionDefinition
			reduce(324), // -, reduce: UnnamedFunctionDefinition
			reduce(324), // .+, reduce: UnnamedFunctionDefinition
			reduce(324), // .-, reduce: UnnamedFunctionDefinition
			reduce(324), // *, reduce: UnnamedFunctionDefinition
			reduce(324), // /, reduce: UnnamedFunctionDefinition
			reduce(324), // //, reduce: UnnamedFunctionDefinition
			reduce(324), // %, reduce: UnnamedFunctionDefinition
			reduce(324), // .*, reduce: UnnamedFunctionDefinition
			reduce(324), // ./, reduce: UnnamedFunctionDefinition
			reduce(324), // .//, reduce: UnnamedFunctionDefinition
			reduce(324), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(324), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(324), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(324), // [[, reduce: UnnamedFunctionDefinition
			reduce(324), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3951
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S3952
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3953
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(107), // >, reduce: PositionalFieldName
			reduce(107), // >>, reduce: PositionalFieldName
			reduce(107), // |, reduce: PositionalFieldName
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(107), // ,, reduce: PositionalFieldName
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(107), // ?, reduce: PositionalFieldName
			nil,         // :
			reduce(107), // ||, reduce: PositionalFieldName
			reduce(107), // ^^, reduce: PositionalFieldName
			reduce(107), // &&, reduce: PositionalFieldName
			reduce(107), // ??, reduce: PositionalFieldName
			reduce(107), // ???, reduce: PositionalFieldName
			reduce(107), // =~, reduce: PositionalFieldName
			reduce(107), // !=~, reduce: PositionalFieldName
			reduce(107), // ==, reduce: PositionalFieldName
			reduce(107), // !=, reduce: PositionalFieldName
			reduce(107), // <=>, reduce: PositionalFieldName
			reduce(107), // >=, reduce: PositionalFieldName
			reduce(107), // <, reduce: PositionalFieldName
			reduce(107), // <=, reduce: PositionalFieldName
			reduce(107), // ^, reduce: PositionalFieldName
			reduce(107), // &, reduce: PositionalFieldName
			reduce(107), // <<, reduce: PositionalFieldName
			reduce(107), // >>>, reduce: PositionalFieldName
			reduce(107), // +, reduce: PositionalFieldName
			reduce(107), // -, reduce: PositionalFieldName
			reduce(107), // .+, reduce: PositionalFieldName
			reduce(107), // .-, reduce: PositionalFieldName
			reduce(107), // *, reduce: PositionalFieldName
			reduce(107), // /, reduce: PositionalFieldName
			reduce(107), // //, reduce: PositionalFieldName
			reduce(107), // %, reduce: PositionalFieldName
			reduce(107), // .*, reduce: PositionalFieldName
			reduce(107), // ./, reduce: PositionalFieldName
			reduce(107), // .//, reduce: PositionalFieldName
			reduce(107), // ., reduce: PositionalFieldName
			nil,         // !
			nil,         // ~
			reduce(107), // **, reduce: PositionalFieldName
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(107), // [, reduce: PositionalFieldName
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(107), // [[, reduce: PositionalFieldName
			reduce(107), // [[[, reduce: PositionalFieldName
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3954
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4267), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3955
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1226), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1233), // (
			nil,         // )
			nil,         // emitp
			shift(1245), // field_name
			shift(1246), // $[
			nil,         // ]
			shift(1247), // braced_field_name
			shift(1248), // $[[
			shift(1249), // $[[[
			shift(1250), // full_srec
			shift(1252), // oosvar_name
			shift(1253), // @[
			shift(1254), // braced_oosvar_name
			shift(1255), // full_oosvar
			shift(1256), // all
			shift(1257), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1272), // +
			shift(1274), // -
			shift(1275), // .+
			shift(1276), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1279), // !
			shift(1280), // ~
			nil,         // **
			shift(1285), // string_literal
			shift(1286), // regex_case_insensitive
			shift(1287), // int_literal
			shift(1288), // float_literal
			shift(1289), // boolean_literal
			shift(1290), // null_literal
			shift(1291), // inf_literal
			shift(1292), // nan_literal
			shift(1293), // const_M_PI
			shift(1294), // const_M_E
			shift(1295), // panic
			shift(1297), // [
			shift(1299), // ctx_IPS
			shift(1300), // ctx_IFS
			shift(1301), // ctx_IRS
			shift(1302), // ctx_OPS
			shift(1303), // ctx_OFS
			shift(1304), // ctx_ORS
			shift(1305), // ctx_FLATSEP
			shift(1306), // ctx_NF
			shift(1307), // ctx_NR
			shift(1308), // ctx_FNR
			shift(1309), // ctx_FILENAME
			shift(1310), // ctx_FILENUM
			shift(1312), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1319), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3956
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(206), // >, reduce: PowTerm
			reduce(206), // >>, reduce: PowTerm
			reduce(206), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(206), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(206), // ?, reduce: PowTerm
			nil,         // :
			reduce(206), // ||, reduce: PowTerm
			reduce(206), // ^^, reduce: PowTerm
			reduce(206), // &&, reduce: PowTerm
			reduce(206), // ??, reduce: PowTerm
			reduce(206), // ???, reduce: PowTerm
			reduce(206), // =~, reduce: PowTerm
			reduce(206), // !=~, reduce: PowTerm
			reduce(206), // ==, reduce: PowTerm
			reduce(206), // !=, reduce: PowTerm
			reduce(206), // <=>, reduce: PowTerm
			reduce(206), // >=, reduce: PowTerm
			reduce(206), // <, reduce: PowTerm
			reduce(206), // <=, reduce: PowTerm
			reduce(206), // ^, reduce: PowTerm
			reduce(206), // &, reduce: PowTerm
			reduce(206), // <<, reduce: PowTerm
			reduce(206), // >>>, reduce: PowTerm
			reduce(206), // +, reduce: PowTerm
			reduce(206), // -, reduce: PowTerm
			reduce(206), // .+, reduce: PowTerm
			reduce(206), // .-, reduce: PowTerm
			reduce(206), // *, reduce: PowTerm
			reduce(206), // /, reduce: PowTerm
			reduce(206), // //, reduce: PowTerm
			reduce(206), // %, reduce: PowTerm
			reduce(206), // .*, reduce: PowTerm
			reduce(206), // ./, reduce: PowTerm
			reduce(206), // .//, reduce: PowTerm
			reduce(206), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3957
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(205), // >, reduce: PowTerm
			reduce(205), // >>, reduce: PowTerm
			reduce(205), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(205), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(205), // ?, reduce: PowTerm
			nil,         // :
			reduce(205), // ||, reduce: PowTerm
			reduce(205), // ^^, reduce: PowTerm
			reduce(205), // &&, reduce: PowTerm
			reduce(205), // ??, reduce: PowTerm
			reduce(205), // ???, reduce: PowTerm
			reduce(205), // =~, reduce: PowTerm
			reduce(205), // !=~, reduce: PowTerm
			reduce(205), // ==, reduce: PowTerm
			reduce(205), // !=, reduce: PowTerm
			reduce(205), // <=>, reduce: PowTerm
			reduce(205), // >=, reduce: PowTerm
			reduce(205), // <, reduce: PowTerm
			reduce(205), // <=, reduce: PowTerm
			reduce(205), // ^, reduce: PowTerm
			reduce(205), // &, reduce: PowTerm
			reduce(205), // <<, reduce: PowTerm
			reduce(205), // >>>, reduce: PowTerm
			reduce(205), // +, reduce: PowTerm
			reduce(205), // -, reduce: PowTerm
			reduce(205), // .+, reduce: PowTerm
			reduce(205), // .-, reduce: PowTerm
			reduce(205), // *, reduce: PowTerm
			reduce(205), // /, reduce: PowTerm
			reduce(205), // //, reduce: PowTerm
			reduce(205), // %, reduce: PowTerm
			reduce(205), // .*, reduce: PowTerm
			reduce(205), // ./, reduce: PowTerm
			reduce(205), // .//, reduce: PowTerm
			reduce(205), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3958
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(260), // >, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>, reduce: ArrayOrMapIndexAccess
			reduce(260), // |, reduce: ArrayOrMapIndexAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(260), // ,, reduce: ArrayOrMapIndexAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(260), // ?, reduce: ArrayOrMapIndexAccess
			nil,         // :
			reduce(260), // ||, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???, reduce: ArrayOrMapIndexAccess
			reduce(260), // =~, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=~, reduce: ArrayOrMapIndexAccess
			reduce(260), // ==, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=>, reduce: ArrayOrMapIndexAccess
			reduce(260), // >=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>, reduce: ArrayOrMapIndexAccess
			reduce(260), // +, reduce: ArrayOrMapIndexAccess
			reduce(260), // -, reduce: ArrayOrMapIndexAccess
			reduce(260), // .+, reduce: ArrayOrMapIndexAccess
			reduce(260), // .-, reduce: ArrayOrMapIndexAccess
			reduce(260), // *, reduce: ArrayOrMapIndexAccess
			reduce(260), // /, reduce: ArrayOrMapIndexAccess
			reduce(260), // //, reduce: ArrayOrMapIndexAccess
			reduce(260), // %, reduce: ArrayOrMapIndexAccess
			reduce(260), // .*, reduce: ArrayOrMapIndexAccess
			reduce(260), // ./, reduce: ArrayOrMapIndexAccess
			reduce(260), // .//, reduce: ArrayOrMapIndexAccess
			reduce(260), // ., reduce: ArrayOrMapIndexAccess
			nil,         // !
			nil,         // ~
			reduce(260), // **, reduce: ArrayOrMapIndexAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(260), // [, reduce: ArrayOrMapIndexAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(260), // [[, reduce: ArrayOrMapIndexAccess
			reduce(260), // [[[, reduce: ArrayOrMapIndexAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3959
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4270), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3960
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4271), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3961
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(266), // >, reduce: ArraySliceAccess
			reduce(266), // >>, reduce: ArraySliceAccess
			reduce(266), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(266), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(266), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(266), // ||, reduce: ArraySliceAccess
			reduce(266), // ^^, reduce: ArraySliceAccess
			reduce(266), // &&, reduce: ArraySliceAccess
			reduce(266), // ??, reduce: ArraySliceAccess
			reduce(266), // ???, reduce: ArraySliceAccess
			reduce(266), // =~, reduce: ArraySliceAccess
			reduce(266), // !=~, reduce: ArraySliceAccess
			reduce(266), // ==, reduce: ArraySliceAccess
			reduce(266), // !=, reduce: ArraySliceAccess
			reduce(266), // <=>, reduce: ArraySliceAccess
			reduce(266), // >=, reduce: ArraySliceAccess
			reduce(266), // <, reduce: ArraySliceAccess
			reduce(266), // <=, reduce: ArraySliceAccess
			reduce(266), // ^, reduce: ArraySliceAccess
			reduce(266), // &, reduce: ArraySliceAccess
			reduce(266), // <<, reduce: ArraySliceAccess
			reduce(266), // >>>, reduce: ArraySliceAccess
			reduce(266), // +, reduce: ArraySliceAccess
			reduce(266), // -, reduce: ArraySliceAccess
			reduce(266), // .+, reduce: ArraySliceAccess
			reduce(266), // .-, reduce: ArraySliceAccess
			reduce(266), // *, reduce: ArraySliceAccess
			reduce(266), // /, reduce: ArraySliceAccess
			reduce(266), // //, reduce: ArraySliceAccess
			reduce(266), // %, reduce: ArraySliceAccess
			reduce(266), // .*, reduce: ArraySliceAccess
			reduce(266), // ./, reduce: ArraySliceAccess
			reduce(266), // .//, reduce: ArraySliceAccess
			reduce(266), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(266), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(266), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(266), // [[, reduce: ArraySliceAccess
			reduce(266), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3962
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4272), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3963
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4273), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3964
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(254), // >, reduce: ENV
			reduce(254), // >>, reduce: ENV
			reduce(254), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(254), // ,, reduce: ENV
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(254), // ?, reduce: ENV
			nil,         // :
			reduce(254), // ||, reduce: ENV
			reduce(254), // ^^, reduce: ENV
			reduce(254), // &&, reduce: ENV
			reduce(254), // ??, reduce: ENV
			reduce(254), // ???, reduce: ENV
			reduce(254), // =~, reduce: ENV
			reduce(254), // !=~, reduce: ENV
			reduce(254), // ==, reduce: ENV
			reduce(254), // !=, reduce: ENV
			reduce(254), // <=>, reduce: ENV
			reduce(254), // >=, reduce: ENV
			reduce(254), // <, reduce: ENV
			reduce(254), // <=, reduce: ENV
			reduce(254), // ^, reduce: ENV
			reduce(254), // &, reduce: ENV
			reduce(254), // <<, reduce: ENV
			reduce(254), // >>>, reduce: ENV
			reduce(254), // +, reduce: ENV
			reduce(254), // -, reduce: ENV
			reduce(254), // .+, reduce: ENV
			reduce(254), // .-, reduce: ENV
			reduce(254), // *, reduce: ENV
			reduce(254), // /, reduce: ENV
			reduce(254), // //, reduce: ENV
			reduce(254), // %, reduce: ENV
			reduce(254), // .*, reduce: ENV
			reduce(254), // ./, reduce: ENV
			reduce(254), // .//, reduce: ENV
			reduce(254), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(254), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(254), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(254), // [[, reduce: ENV
			reduce(254), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3965
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(269), // >, reduce: FunctionCallsite
			reduce(269), // >>, reduce: FunctionCallsite
			reduce(269), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(269), // ,, reduce: FunctionCallsite
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(269), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(269), // ||, reduce: FunctionCallsite
			reduce(269), // ^^, reduce: FunctionCallsite
			reduce(269), // &&, reduce: FunctionCallsite
			reduce(269), // ??, reduce: FunctionCallsite
			reduce(269), // ???, reduce: FunctionCallsite
			reduce(269), // =~, reduce: FunctionCallsite
			reduce(269), // !=~, reduce: FunctionCallsite
			reduce(269), // ==, reduce: FunctionCallsite
			reduce(269), // !=, reduce: FunctionCallsite
			reduce(269), // <=>, reduce: FunctionCallsite
			reduce(269), // >=, reduce: FunctionCallsite
			reduce(269), // <, reduce: FunctionCallsite
			reduce(269), // <=, reduce: FunctionCallsite
			reduce(269), // ^, reduce: FunctionCallsite
			reduce(269), // &, reduce: FunctionCallsite
			reduce(269), // <<, reduce: FunctionCallsite
			reduce(269), // >>>, reduce: FunctionCallsite
			reduce(269), // +, reduce: FunctionCallsite
			reduce(269), // -, reduce: FunctionCallsite
			reduce(269), // .+, reduce: FunctionCallsite
			reduce(269), // .-, reduce: FunctionCallsite
			reduce(269), // *, reduce: FunctionCallsite
			reduce(269), // /, reduce: FunctionCallsite
			reduce(269), // //, reduce: FunctionCallsite
			reduce(269), // %, reduce: FunctionCallsite
			reduce(269), // .*, reduce: FunctionCallsite
			reduce(269), // ./, reduce: FunctionCallsite
			reduce(269), // .//, reduce: FunctionCallsite
			reduce(269), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(269), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(269), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(269), // [[, reduce: FunctionCallsite
			reduce(269), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3966
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4275), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4276), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3967
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(73), // $, reduce: EmitStatement
			nil,        // empty
			reduce(73), // ;, reduce: EmitStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3968
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(71),  // $, reduce: EmitStatement
			nil,         // empty
			reduce(71),  // ;, reduce: EmitStatement
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(4277), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3969
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(98), // $, reduce: EmitKeys
			nil,        // empty
			reduce(98), // ;, reduce: EmitKeys
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3970
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(74), // $, reduce: EmitStatement
			nil,        // empty
			reduce(74), // ;, reduce: EmitStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3971
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(108), // >, reduce: PositionalFieldValue
			reduce(108), // >>, reduce: PositionalFieldValue
			reduce(108), // |, reduce: PositionalFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(108), // ), reduce: PositionalFieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(108), // ?, reduce: PositionalFieldValue
			nil,         // :
			reduce(108), // ||, reduce: PositionalFieldValue
			reduce(108), // ^^, reduce: PositionalFieldValue
			reduce(108), // &&, reduce: PositionalFieldValue
			reduce(108), // ??, reduce: PositionalFieldValue
			reduce(108), // ???, reduce: PositionalFieldValue
			reduce(108), // =~, reduce: PositionalFieldValue
			reduce(108), // !=~, reduce: PositionalFieldValue
			reduce(108), // ==, reduce: PositionalFieldValue
			reduce(108), // !=, reduce: PositionalFieldValue
			reduce(108), // <=>, reduce: PositionalFieldValue
			reduce(108), // >=, reduce: PositionalFieldValue
			reduce(108), // <, reduce: PositionalFieldValue
			reduce(108), // <=, reduce: PositionalFieldValue
			reduce(108), // ^, reduce: PositionalFieldValue
			reduce(108), // &, reduce: PositionalFieldValue
			reduce(108), // <<, reduce: PositionalFieldValue
			reduce(108), // >>>, reduce: PositionalFieldValue
			reduce(108), // +, reduce: PositionalFieldValue
			reduce(108), // -, reduce: PositionalFieldValue
			reduce(108), // .+, reduce: PositionalFieldValue
			reduce(108), // .-, reduce: PositionalFieldValue
			reduce(108), // *, reduce: PositionalFieldValue
			reduce(108), // /, reduce: PositionalFieldValue
			reduce(108), // //, reduce: PositionalFieldValue
			reduce(108), // %, reduce: PositionalFieldValue
			reduce(108), // .*, reduce: PositionalFieldValue
			reduce(108), // ./, reduce: PositionalFieldValue
			reduce(108), // .//, reduce: PositionalFieldValue
			reduce(108), // ., reduce: PositionalFieldValue
			nil,         // !
			nil,         // ~
			reduce(108), // **, reduce: PositionalFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(108), // [, reduce: PositionalFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(108), // [[, reduce: PositionalFieldValue
			reduce(108), // [[[, reduce: PositionalFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3972
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(149), // ), reduce: TernaryTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3973
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4278), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3974
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(265), // >, reduce: ArraySliceAccess
			reduce(265), // >>, reduce: ArraySliceAccess
			reduce(265), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(265), // ), reduce: ArraySliceAccess
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(265), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(265), // ||, reduce: ArraySliceAccess
			reduce(265), // ^^, reduce: ArraySliceAccess
			reduce(265), // &&, reduce: ArraySliceAccess
			reduce(265), // ??, reduce: ArraySliceAccess
			reduce(265), // ???, reduce: ArraySliceAccess
			reduce(265), // =~, reduce: ArraySliceAccess
			reduce(265), // !=~, reduce: ArraySliceAccess
			reduce(265), // ==, reduce: ArraySliceAccess
			reduce(265), // !=, reduce: ArraySliceAccess
			reduce(265), // <=>, reduce: ArraySliceAccess
			reduce(265), // >=, reduce: ArraySliceAccess
			reduce(265), // <, reduce: ArraySliceAccess
			reduce(265), // <=, reduce: ArraySliceAccess
			reduce(265), // ^, reduce: ArraySliceAccess
			reduce(265), // &, reduce: ArraySliceAccess
			reduce(265), // <<, reduce: ArraySliceAccess
			reduce(265), // >>>, reduce: ArraySliceAccess
			reduce(265), // +, reduce: ArraySliceAccess
			reduce(265), // -, reduce: ArraySliceAccess
			reduce(265), // .+, reduce: ArraySliceAccess
			reduce(265), // .-, reduce: ArraySliceAccess
			reduce(265), // *, reduce: ArraySliceAccess
			reduce(265), // /, reduce: ArraySliceAccess
			reduce(265), // //, reduce: ArraySliceAccess
			reduce(265), // %, reduce: ArraySliceAccess
			reduce(265), // .*, reduce: ArraySliceAccess
			reduce(265), // ./, reduce: ArraySliceAccess
			reduce(265), // .//, reduce: ArraySliceAccess
			reduce(265), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(265), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(265), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(265), // [[, reduce: ArraySliceAccess
			reduce(265), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3975
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(264), // >, reduce: ArraySliceAccess
			reduce(264), // >>, reduce: ArraySliceAccess
			reduce(264), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(264), // ), reduce: ArraySliceAccess
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(264), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(264), // ||, reduce: ArraySliceAccess
			reduce(264), // ^^, reduce: ArraySliceAccess
			reduce(264), // &&, reduce: ArraySliceAccess
			reduce(264), // ??, reduce: ArraySliceAccess
			reduce(264), // ???, reduce: ArraySliceAccess
			reduce(264), // =~, reduce: ArraySliceAccess
			reduce(264), // !=~, reduce: ArraySliceAccess
			reduce(264), // ==, reduce: ArraySliceAccess
			reduce(264), // !=, reduce: ArraySliceAccess
			reduce(264), // <=>, reduce: ArraySliceAccess
			reduce(264), // >=, reduce: ArraySliceAccess
			reduce(264), // <, reduce: ArraySliceAccess
			reduce(264), // <=, reduce: ArraySliceAccess
			reduce(264), // ^, reduce: ArraySliceAccess
			reduce(264), // &, reduce: ArraySliceAccess
			reduce(264), // <<, reduce: ArraySliceAccess
			reduce(264), // >>>, reduce: ArraySliceAccess
			reduce(264), // +, reduce: ArraySliceAccess
			reduce(264), // -, reduce: ArraySliceAccess
			reduce(264), // .+, reduce: ArraySliceAccess
			reduce(264), // .-, reduce: ArraySliceAccess
			reduce(264), // *, reduce: ArraySliceAccess
			reduce(264), // /, reduce: ArraySliceAccess
			reduce(264), // //, reduce: ArraySliceAccess
			reduce(264), // %, reduce: ArraySliceAccess
			reduce(264), // .*, reduce: ArraySliceAccess
			reduce(264), // ./, reduce: ArraySliceAccess
			reduce(264), // .//, reduce: ArraySliceAccess
			reduce(264), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(264), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(264), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(264), // [[, reduce: ArraySliceAccess
			reduce(264), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3976
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(261), // >, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |, reduce: ArrayOrMapPositionalNameAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(261), // ), reduce: ArrayOrMapPositionalNameAccess
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(261), // ?, reduce: ArrayOrMapPositionalNameAccess
			nil,         // :
			reduce(261), // ||, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // =~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ==, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .+, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .-, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .*, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ./, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .//, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ., reduce: ArrayOrMapPositionalNameAccess
			nil,         // !
			nil,         // ~
			reduce(261), // **, reduce: ArrayOrMapPositionalNameAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(261), // [, reduce: ArrayOrMapPositionalNameAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(261), // [[, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // [[[, reduce: ArrayOrMapPositionalNameAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3977
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4279), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3978
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(324), // >, reduce: UnnamedFunctionDefinition
			reduce(324), // >>, reduce: UnnamedFunctionDefinition
			reduce(324), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(324), // ), reduce: UnnamedFunctionDefinition
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(324), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(324), // ||, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^, reduce: UnnamedFunctionDefinition
			reduce(324), // &&, reduce: UnnamedFunctionDefinition
			reduce(324), // ??, reduce: UnnamedFunctionDefinition
			reduce(324), // ???, reduce: UnnamedFunctionDefinition
			reduce(324), // =~, reduce: UnnamedFunctionDefinition
			reduce(324), // !=~, reduce: UnnamedFunctionDefinition
			reduce(324), // ==, reduce: UnnamedFunctionDefinition
			reduce(324), // !=, reduce: UnnamedFunctionDefinition
			reduce(324), // <=>, reduce: UnnamedFunctionDefinition
			reduce(324), // >=, reduce: UnnamedFunctionDefinition
			reduce(324), // <, reduce: UnnamedFunctionDefinition
			reduce(324), // <=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^, reduce: UnnamedFunctionDefinition
			reduce(324), // &, reduce: UnnamedFunctionDefinition
			reduce(324), // <<, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>, reduce: UnnamedFunctionDefinition
			reduce(324), // +, reduce: UnnamedFunctionDefinition
			reduce(324), // -, reduce: UnnamedFunctionDefinition
			reduce(324), // .+, reduce: UnnamedFunctionDefinition
			reduce(324), // .-, reduce: UnnamedFunctionDefinition
			reduce(324), // *, reduce: UnnamedFunctionDefinition
			reduce(324), // /, reduce: UnnamedFunctionDefinition
			reduce(324), // //, reduce: UnnamedFunctionDefinition
			reduce(324), // %, reduce: UnnamedFunctionDefinition
			reduce(324), // .*, reduce: UnnamedFunctionDefinition
			reduce(324), // ./, reduce: UnnamedFunctionDefinition
			reduce(324), // .//, reduce: UnnamedFunctionDefinition
			reduce(324), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(324), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(324), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(324), // [[, reduce: UnnamedFunctionDefinition
			reduce(324), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3979
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S3980
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3981
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(81), // $, reduce: EmitPStatement
			nil,        // empty
			reduce(81), // ;, reduce: EmitPStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3982
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(79),  // $, reduce: EmitPStatement
			nil,         // empty
			reduce(79),  // ;, reduce: EmitPStatement
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(4282), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3983
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(82), // $, reduce: EmitPStatement
			nil,        // empty
			reduce(82), // ;, reduce: EmitPStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S3984
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(108), // >, reduce: PositionalFieldValue
			reduce(108), // >>, reduce: PositionalFieldValue
			reduce(108), // |, reduce: PositionalFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(108), // ], reduce: PositionalFieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(108), // ?, reduce: PositionalFieldValue
			nil,         // :
			reduce(108), // ||, reduce: PositionalFieldValue
			reduce(108), // ^^, reduce: PositionalFieldValue
			reduce(108), // &&, reduce: PositionalFieldValue
			reduce(108), // ??, reduce: PositionalFieldValue
			reduce(108), // ???, reduce: PositionalFieldValue
			reduce(108), // =~, reduce: PositionalFieldValue
			reduce(108), // !=~, reduce: PositionalFieldValue
			reduce(108), // ==, reduce: PositionalFieldValue
			reduce(108), // !=, reduce: PositionalFieldValue
			reduce(108), // <=>, reduce: PositionalFieldValue
			reduce(108), // >=, reduce: PositionalFieldValue
			reduce(108), // <, reduce: PositionalFieldValue
			reduce(108), // <=, reduce: PositionalFieldValue
			reduce(108), // ^, reduce: PositionalFieldValue
			reduce(108), // &, reduce: PositionalFieldValue
			reduce(108), // <<, reduce: PositionalFieldValue
			reduce(108), // >>>, reduce: PositionalFieldValue
			reduce(108), // +, reduce: PositionalFieldValue
			reduce(108), // -, reduce: PositionalFieldValue
			reduce(108), // .+, reduce: PositionalFieldValue
			reduce(108), // .-, reduce: PositionalFieldValue
			reduce(108), // *, reduce: PositionalFieldValue
			reduce(108), // /, reduce: PositionalFieldValue
			reduce(108), // //, reduce: PositionalFieldValue
			reduce(108), // %, reduce: PositionalFieldValue
			reduce(108), // .*, reduce: PositionalFieldValue
			reduce(108), // ./, reduce: PositionalFieldValue
			reduce(108), // .//, reduce: PositionalFieldValue
			reduce(108), // ., reduce: PositionalFieldValue
			nil,         // !
			nil,         // ~
			reduce(108), // **, reduce: PositionalFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(108), // [, reduce: PositionalFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(108), // [[, reduce: PositionalFieldValue
			reduce(108), // [[[, reduce: PositionalFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3985
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(149), // ], reduce: TernaryTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3986
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4283), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3987
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(265), // >, reduce: ArraySliceAccess
			reduce(265), // >>, reduce: ArraySliceAccess
			reduce(265), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(265), // ], reduce: ArraySliceAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(265), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(265), // ||, reduce: ArraySliceAccess
			reduce(265), // ^^, reduce: ArraySliceAccess
			reduce(265), // &&, reduce: ArraySliceAccess
			reduce(265), // ??, reduce: ArraySliceAccess
			reduce(265), // ???, reduce: ArraySliceAccess
			reduce(265), // =~, reduce: ArraySliceAccess
			reduce(265), // !=~, reduce: ArraySliceAccess
			reduce(265), // ==, reduce: ArraySliceAccess
			reduce(265), // !=, reduce: ArraySliceAccess
			reduce(265), // <=>, reduce: ArraySliceAccess
			reduce(265), // >=, reduce: ArraySliceAccess
			reduce(265), // <, reduce: ArraySliceAccess
			reduce(265), // <=, reduce: ArraySliceAccess
			reduce(265), // ^, reduce: ArraySliceAccess
			reduce(265), // &, reduce: ArraySliceAccess
			reduce(265), // <<, reduce: ArraySliceAccess
			reduce(265), // >>>, reduce: ArraySliceAccess
			reduce(265), // +, reduce: ArraySliceAccess
			reduce(265), // -, reduce: ArraySliceAccess
			reduce(265), // .+, reduce: ArraySliceAccess
			reduce(265), // .-, reduce: ArraySliceAccess
			reduce(265), // *, reduce: ArraySliceAccess
			reduce(265), // /, reduce: ArraySliceAccess
			reduce(265), // //, reduce: ArraySliceAccess
			reduce(265), // %, reduce: ArraySliceAccess
			reduce(265), // .*, reduce: ArraySliceAccess
			reduce(265), // ./, reduce: ArraySliceAccess
			reduce(265), // .//, reduce: ArraySliceAccess
			reduce(265), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(265), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(265), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(265), // [[, reduce: ArraySliceAccess
			reduce(265), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3988
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(264), // >, reduce: ArraySliceAccess
			reduce(264), // >>, reduce: ArraySliceAccess
			reduce(264), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(264), // ], reduce: ArraySliceAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(264), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(264), // ||, reduce: ArraySliceAccess
			reduce(264), // ^^, reduce: ArraySliceAccess
			reduce(264), // &&, reduce: ArraySliceAccess
			reduce(264), // ??, reduce: ArraySliceAccess
			reduce(264), // ???, reduce: ArraySliceAccess
			reduce(264), // =~, reduce: ArraySliceAccess
			reduce(264), // !=~, reduce: ArraySliceAccess
			reduce(264), // ==, reduce: ArraySliceAccess
			reduce(264), // !=, reduce: ArraySliceAccess
			reduce(264), // <=>, reduce: ArraySliceAccess
			reduce(264), // >=, reduce: ArraySliceAccess
			reduce(264), // <, reduce: ArraySliceAccess
			reduce(264), // <=, reduce: ArraySliceAccess
			reduce(264), // ^, reduce: ArraySliceAccess
			reduce(264), // &, reduce: ArraySliceAccess
			reduce(264), // <<, reduce: ArraySliceAccess
			reduce(264), // >>>, reduce: ArraySliceAccess
			reduce(264), // +, reduce: ArraySliceAccess
			reduce(264), // -, reduce: ArraySliceAccess
			reduce(264), // .+, reduce: ArraySliceAccess
			reduce(264), // .-, reduce: ArraySliceAccess
			reduce(264), // *, reduce: ArraySliceAccess
			reduce(264), // /, reduce: ArraySliceAccess
			reduce(264), // //, reduce: ArraySliceAccess
			reduce(264), // %, reduce: ArraySliceAccess
			reduce(264), // .*, reduce: ArraySliceAccess
			reduce(264), // ./, reduce: ArraySliceAccess
			reduce(264), // .//, reduce: ArraySliceAccess
			reduce(264), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(264), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(264), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(264), // [[, reduce: ArraySliceAccess
			reduce(264), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3989
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(261), // >, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |, reduce: ArrayOrMapPositionalNameAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(261), // ], reduce: ArrayOrMapPositionalNameAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(261), // ?, reduce: ArrayOrMapPositionalNameAccess
			nil,         // :
			reduce(261), // ||, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // =~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ==, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .+, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .-, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .*, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ./, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .//, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ., reduce: ArrayOrMapPositionalNameAccess
			nil,         // !
			nil,         // ~
			reduce(261), // **, reduce: ArrayOrMapPositionalNameAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(261), // [, reduce: ArrayOrMapPositionalNameAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(261), // [[, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // [[[, reduce: ArrayOrMapPositionalNameAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3990
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4284), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3991
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(324), // >, reduce: UnnamedFunctionDefinition
			reduce(324), // >>, reduce: UnnamedFunctionDefinition
			reduce(324), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(324), // ], reduce: UnnamedFunctionDefinition
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(324), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(324), // ||, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^, reduce: UnnamedFunctionDefinition
			reduce(324), // &&, reduce: UnnamedFunctionDefinition
			reduce(324), // ??, reduce: UnnamedFunctionDefinition
			reduce(324), // ???, reduce: UnnamedFunctionDefinition
			reduce(324), // =~, reduce: UnnamedFunctionDefinition
			reduce(324), // !=~, reduce: UnnamedFunctionDefinition
			reduce(324), // ==, reduce: UnnamedFunctionDefinition
			reduce(324), // !=, reduce: UnnamedFunctionDefinition
			reduce(324), // <=>, reduce: UnnamedFunctionDefinition
			reduce(324), // >=, reduce: UnnamedFunctionDefinition
			reduce(324), // <, reduce: UnnamedFunctionDefinition
			reduce(324), // <=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^, reduce: UnnamedFunctionDefinition
			reduce(324), // &, reduce: UnnamedFunctionDefinition
			reduce(324), // <<, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>, reduce: UnnamedFunctionDefinition
			reduce(324), // +, reduce: UnnamedFunctionDefinition
			reduce(324), // -, reduce: UnnamedFunctionDefinition
			reduce(324), // .+, reduce: UnnamedFunctionDefinition
			reduce(324), // .-, reduce: UnnamedFunctionDefinition
			reduce(324), // *, reduce: UnnamedFunctionDefinition
			reduce(324), // /, reduce: UnnamedFunctionDefinition
			reduce(324), // //, reduce: UnnamedFunctionDefinition
			reduce(324), // %, reduce: UnnamedFunctionDefinition
			reduce(324), // .*, reduce: UnnamedFunctionDefinition
			reduce(324), // ./, reduce: UnnamedFunctionDefinition
			reduce(324), // .//, reduce: UnnamedFunctionDefinition
			reduce(324), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(324), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(324), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(324), // [[, reduce: UnnamedFunctionDefinition
			reduce(324), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3992
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S3993
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3994
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(263), // $, reduce: ArraySliceAccess
			nil,         // empty
			reduce(263), // ;, reduce: ArraySliceAccess
			reduce(263), // {, reduce: ArraySliceAccess
			nil,         // }
			reduce(263), // =, reduce: ArraySliceAccess
			nil,         // unset
			nil,         // filter
			reduce(263), // >, reduce: ArraySliceAccess
			reduce(263), // >>, reduce: ArraySliceAccess
			reduce(263), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(263), // ||=, reduce: ArraySliceAccess
			reduce(263), // ^^=, reduce: ArraySliceAccess
			reduce(263), // &&=, reduce: ArraySliceAccess
			reduce(263), // ??=, reduce: ArraySliceAccess
			reduce(263), // ???=, reduce: ArraySliceAccess
			reduce(263), // |=, reduce: ArraySliceAccess
			reduce(263), // &=, reduce: ArraySliceAccess
			reduce(263), // ^=, reduce: ArraySliceAccess
			reduce(263), // <<=, reduce: ArraySliceAccess
			reduce(263), // >>=, reduce: ArraySliceAccess
			reduce(263), // >>>=, reduce: ArraySliceAccess
			reduce(263), // +=, reduce: ArraySliceAccess
			reduce(263), // .=, reduce: ArraySliceAccess
			reduce(263), // -=, reduce: ArraySliceAccess
			reduce(263), // *=, reduce: ArraySliceAccess
			reduce(263), // /=, reduce: ArraySliceAccess
			reduce(263), // //=, reduce: ArraySliceAccess
			reduce(263), // %=, reduce: ArraySliceAccess
			reduce(263), // **=, reduce: ArraySliceAccess
			reduce(263), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(263), // ||, reduce: ArraySliceAccess
			reduce(263), // ^^, reduce: ArraySliceAccess
			reduce(263), // &&, reduce: ArraySliceAccess
			reduce(263), // ??, reduce: ArraySliceAccess
			reduce(263), // ???, reduce: ArraySliceAccess
			reduce(263), // =~, reduce: ArraySliceAccess
			reduce(263), // !=~, reduce: ArraySliceAccess
			reduce(263), // ==, reduce: ArraySliceAccess
			reduce(263), // !=, reduce: ArraySliceAccess
			reduce(263), // <=>, reduce: ArraySliceAccess
			reduce(263), // >=, reduce: ArraySliceAccess
			reduce(263), // <, reduce: ArraySliceAccess
			reduce(263), // <=, reduce: ArraySliceAccess
			reduce(263), // ^, reduce: ArraySliceAccess
			reduce(263), // &, reduce: ArraySliceAccess
			reduce(263), // <<, reduce: ArraySliceAccess
			reduce(263), // >>>, reduce: ArraySliceAccess
			reduce(263), // +, reduce: ArraySliceAccess
			reduce(263), // -, reduce: ArraySliceAccess
			reduce(263), // .+, reduce: ArraySliceAccess
			reduce(263), // .-, reduce: ArraySliceAccess
			reduce(263), // *, reduce: ArraySliceAccess
			reduce(263), // /, reduce: ArraySliceAccess
			reduce(263), // //, reduce: ArraySliceAccess
			reduce(263), // %, reduce: ArraySliceAccess
			reduce(263), // .*, reduce: ArraySliceAccess
			reduce(263), // ./, reduce: ArraySliceAccess
			reduce(263), // .//, reduce: ArraySliceAccess
			reduce(263), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(263), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(263), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(263), // [[, reduce: ArraySliceAccess
			reduce(263), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3995
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(107), // >, reduce: PositionalFieldName
			reduce(107), // >>, reduce: PositionalFieldName
			reduce(107), // |, reduce: PositionalFieldName
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(107), // ], reduce: PositionalFieldName
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(107), // ?, reduce: PositionalFieldName
			reduce(107), // :, reduce: PositionalFieldName
			reduce(107), // ||, reduce: PositionalFieldName
			reduce(107), // ^^, reduce: PositionalFieldName
			reduce(107), // &&, reduce: PositionalFieldName
			reduce(107), // ??, reduce: PositionalFieldName
			reduce(107), // ???, reduce: PositionalFieldName
			reduce(107), // =~, reduce: PositionalFieldName
			reduce(107), // !=~, reduce: PositionalFieldName
			reduce(107), // ==, reduce: PositionalFieldName
			reduce(107), // !=, reduce: PositionalFieldName
			reduce(107), // <=>, reduce: PositionalFieldName
			reduce(107), // >=, reduce: PositionalFieldName
			reduce(107), // <, reduce: PositionalFieldName
			reduce(107), // <=, reduce: PositionalFieldName
			reduce(107), // ^, reduce: PositionalFieldName
			reduce(107), // &, reduce: PositionalFieldName
			reduce(107), // <<, reduce: PositionalFieldName
			reduce(107), // >>>, reduce: PositionalFieldName
			reduce(107), // +, reduce: PositionalFieldName
			reduce(107), // -, reduce: PositionalFieldName
			reduce(107), // .+, reduce: PositionalFieldName
			reduce(107), // .-, reduce: PositionalFieldName
			reduce(107), // *, reduce: PositionalFieldName
			reduce(107), // /, reduce: PositionalFieldName
			reduce(107), // //, reduce: PositionalFieldName
			reduce(107), // %, reduce: PositionalFieldName
			reduce(107), // .*, reduce: PositionalFieldName
			reduce(107), // ./, reduce: PositionalFieldName
			reduce(107), // .//, reduce: PositionalFieldName
			reduce(107), // ., reduce: PositionalFieldName
			nil,         // !
			nil,         // ~
			reduce(107), // **, reduce: PositionalFieldName
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(107), // [, reduce: PositionalFieldName
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(107), // [[, reduce: PositionalFieldName
			reduce(107), // [[[, reduce: PositionalFieldName
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3996
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4287), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3997
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3998
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(206), // >, reduce: PowTerm
			reduce(206), // >>, reduce: PowTerm
			reduce(206), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(206), // ], reduce: PowTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(206), // ?, reduce: PowTerm
			reduce(206), // :, reduce: PowTerm
			reduce(206), // ||, reduce: PowTerm
			reduce(206), // ^^, reduce: PowTerm
			reduce(206), // &&, reduce: PowTerm
			reduce(206), // ??, reduce: PowTerm
			reduce(206), // ???, reduce: PowTerm
			reduce(206), // =~, reduce: PowTerm
			reduce(206), // !=~, reduce: PowTerm
			reduce(206), // ==, reduce: PowTerm
			reduce(206), // !=, reduce: PowTerm
			reduce(206), // <=>, reduce: PowTerm
			reduce(206), // >=, reduce: PowTerm
			reduce(206), // <, reduce: PowTerm
			reduce(206), // <=, reduce: PowTerm
			reduce(206), // ^, reduce: PowTerm
			reduce(206), // &, reduce: PowTerm
			reduce(206), // <<, reduce: PowTerm
			reduce(206), // >>>, reduce: PowTerm
			reduce(206), // +, reduce: PowTerm
			reduce(206), // -, reduce: PowTerm
			reduce(206), // .+, reduce: PowTerm
			reduce(206), // .-, reduce: PowTerm
			reduce(206), // *, reduce: PowTerm
			reduce(206), // /, reduce: PowTerm
			reduce(206), // //, reduce: PowTerm
			reduce(206), // %, reduce: PowTerm
			reduce(206), // .*, reduce: PowTerm
			reduce(206), // ./, reduce: PowTerm
			reduce(206), // .//, reduce: PowTerm
			reduce(206), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S3999
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(205), // >, reduce: PowTerm
			reduce(205), // >>, reduce: PowTerm
			reduce(205), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(205), // ], reduce: PowTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(205), // ?, reduce: PowTerm
			reduce(205), // :, reduce: PowTerm
			reduce(205), // ||, reduce: PowTerm
			reduce(205), // ^^, reduce: PowTerm
			reduce(205), // &&, reduce: PowTerm
			reduce(205), // ??, reduce: PowTerm
			reduce(205), // ???, reduce: PowTerm
			reduce(205), // =~, reduce: PowTerm
			reduce(205), // !=~, reduce: PowTerm
			reduce(205), // ==, reduce: PowTerm
			reduce(205), // !=, reduce: PowTerm
			reduce(205), // <=>, reduce: PowTerm
			reduce(205), // >=, reduce: PowTerm
			reduce(205), // <, reduce: PowTerm
			reduce(205), // <=, reduce: PowTerm
			reduce(205), // ^, reduce: PowTerm
			reduce(205), // &, reduce: PowTerm
			reduce(205), // <<, reduce: PowTerm
			reduce(205), // >>>, reduce: PowTerm
			reduce(205), // +, reduce: PowTerm
			reduce(205), // -, reduce: PowTerm
			reduce(205), // .+, reduce: PowTerm
			reduce(205), // .-, reduce: PowTerm
			reduce(205), // *, reduce: PowTerm
			reduce(205), // /, reduce: PowTerm
			reduce(205), // //, reduce: PowTerm
			reduce(205), // %, reduce: PowTerm
			reduce(205), // .*, reduce: PowTerm
			reduce(205), // ./, reduce: PowTerm
			reduce(205), // .//, reduce: PowTerm
			reduce(205), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4000
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(260), // >, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>, reduce: ArrayOrMapIndexAccess
			reduce(260), // |, reduce: ArrayOrMapIndexAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(260), // ], reduce: ArrayOrMapIndexAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(260), // ?, reduce: ArrayOrMapIndexAccess
			reduce(260), // :, reduce: ArrayOrMapIndexAccess
			reduce(260), // ||, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???, reduce: ArrayOrMapIndexAccess
			reduce(260), // =~, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=~, reduce: ArrayOrMapIndexAccess
			reduce(260), // ==, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=>, reduce: ArrayOrMapIndexAccess
			reduce(260), // >=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>, reduce: ArrayOrMapIndexAccess
			reduce(260), // +, reduce: ArrayOrMapIndexAccess
			reduce(260), // -, reduce: ArrayOrMapIndexAccess
			reduce(260), // .+, reduce: ArrayOrMapIndexAccess
			reduce(260), // .-, reduce: ArrayOrMapIndexAccess
			reduce(260), // *, reduce: ArrayOrMapIndexAccess
			reduce(260), // /, reduce: ArrayOrMapIndexAccess
			reduce(260), // //, reduce: ArrayOrMapIndexAccess
			reduce(260), // %, reduce: ArrayOrMapIndexAccess
			reduce(260), // .*, reduce: ArrayOrMapIndexAccess
			reduce(260), // ./, reduce: ArrayOrMapIndexAccess
			reduce(260), // .//, reduce: ArrayOrMapIndexAccess
			reduce(260), // ., reduce: ArrayOrMapIndexAccess
			nil,         // !
			nil,         // ~
			reduce(260), // **, reduce: ArrayOrMapIndexAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(260), // [, reduce: ArrayOrMapIndexAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(260), // [[, reduce: ArrayOrMapIndexAccess
			reduce(260), // [[[, reduce: ArrayOrMapIndexAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4001
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4290), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4002
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4291), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4003
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(266), // >, reduce: ArraySliceAccess
			reduce(266), // >>, reduce: ArraySliceAccess
			reduce(266), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(266), // ], reduce: ArraySliceAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(266), // ?, reduce: ArraySliceAccess
			reduce(266), // :, reduce: ArraySliceAccess
			reduce(266), // ||, reduce: ArraySliceAccess
			reduce(266), // ^^, reduce: ArraySliceAccess
			reduce(266), // &&, reduce: ArraySliceAccess
			reduce(266), // ??, reduce: ArraySliceAccess
			reduce(266), // ???, reduce: ArraySliceAccess
			reduce(266), // =~, reduce: ArraySliceAccess
			reduce(266), // !=~, reduce: ArraySliceAccess
			reduce(266), // ==, reduce: ArraySliceAccess
			reduce(266), // !=, reduce: ArraySliceAccess
			reduce(266), // <=>, reduce: ArraySliceAccess
			reduce(266), // >=, reduce: ArraySliceAccess
			reduce(266), // <, reduce: ArraySliceAccess
			reduce(266), // <=, reduce: ArraySliceAccess
			reduce(266), // ^, reduce: ArraySliceAccess
			reduce(266), // &, reduce: ArraySliceAccess
			reduce(266), // <<, reduce: ArraySliceAccess
			reduce(266), // >>>, reduce: ArraySliceAccess
			reduce(266), // +, reduce: ArraySliceAccess
			reduce(266), // -, reduce: ArraySliceAccess
			reduce(266), // .+, reduce: ArraySliceAccess
			reduce(266), // .-, reduce: ArraySliceAccess
			reduce(266), // *, reduce: ArraySliceAccess
			reduce(266), // /, reduce: ArraySliceAccess
			reduce(266), // //, reduce: ArraySliceAccess
			reduce(266), // %, reduce: ArraySliceAccess
			reduce(266), // .*, reduce: ArraySliceAccess
			reduce(266), // ./, reduce: ArraySliceAccess
			reduce(266), // .//, reduce: ArraySliceAccess
			reduce(266), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(266), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(266), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(266), // [[, reduce: ArraySliceAccess
			reduce(266), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4004
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4292), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4005
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4293), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4006
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(254), // >, reduce: ENV
			reduce(254), // >>, reduce: ENV
			reduce(254), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(254), // ], reduce: ENV
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(254), // ?, reduce: ENV
			reduce(254), // :, reduce: ENV
			reduce(254), // ||, reduce: ENV
			reduce(254), // ^^, reduce: ENV
			reduce(254), // &&, reduce: ENV
			reduce(254), // ??, reduce: ENV
			reduce(254), // ???, reduce: ENV
			reduce(254), // =~, reduce: ENV
			reduce(254), // !=~, reduce: ENV
			reduce(254), // ==, reduce: ENV
			reduce(254), // !=, reduce: ENV
			reduce(254), // <=>, reduce: ENV
			reduce(254), // >=, reduce: ENV
			reduce(254), // <, reduce: ENV
			reduce(254), // <=, reduce: ENV
			reduce(254), // ^, reduce: ENV
			reduce(254), // &, reduce: ENV
			reduce(254), // <<, reduce: ENV
			reduce(254), // >>>, reduce: ENV
			reduce(254), // +, reduce: ENV
			reduce(254), // -, reduce: ENV
			reduce(254), // .+, reduce: ENV
			reduce(254), // .-, reduce: ENV
			reduce(254), // *, reduce: ENV
			reduce(254), // /, reduce: ENV
			reduce(254), // //, reduce: ENV
			reduce(254), // %, reduce: ENV
			reduce(254), // .*, reduce: ENV
			reduce(254), // ./, reduce: ENV
			reduce(254), // .//, reduce: ENV
			reduce(254), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(254), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(254), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(254), // [[, reduce: ENV
			reduce(254), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4007
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(269), // >, reduce: FunctionCallsite
			reduce(269), // >>, reduce: FunctionCallsite
			reduce(269), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(269), // ], reduce: FunctionCallsite
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(269), // ?, reduce: FunctionCallsite
			reduce(269), // :, reduce: FunctionCallsite
			reduce(269), // ||, reduce: FunctionCallsite
			reduce(269), // ^^, reduce: FunctionCallsite
			reduce(269), // &&, reduce: FunctionCallsite
			reduce(269), // ??, reduce: FunctionCallsite
			reduce(269), // ???, reduce: FunctionCallsite
			reduce(269), // =~, reduce: FunctionCallsite
			reduce(269), // !=~, reduce: FunctionCallsite
			reduce(269), // ==, reduce: FunctionCallsite
			reduce(269), // !=, reduce: FunctionCallsite
			reduce(269), // <=>, reduce: FunctionCallsite
			reduce(269), // >=, reduce: FunctionCallsite
			reduce(269), // <, reduce: FunctionCallsite
			reduce(269), // <=, reduce: FunctionCallsite
			reduce(269), // ^, reduce: FunctionCallsite
			reduce(269), // &, reduce: FunctionCallsite
			reduce(269), // <<, reduce: FunctionCallsite
			reduce(269), // >>>, reduce: FunctionCallsite
			reduce(269), // +, reduce: FunctionCallsite
			reduce(269), // -, reduce: FunctionCallsite
			reduce(269), // .+, reduce: FunctionCallsite
			reduce(269), // .-, reduce: FunctionCallsite
			reduce(269), // *, reduce: FunctionCallsite
			reduce(269), // /, reduce: FunctionCallsite
			reduce(269), // //, reduce: FunctionCallsite
			reduce(269), // %, reduce: FunctionCallsite
			reduce(269), // .*, reduce: FunctionCallsite
			reduce(269), // ./, reduce: FunctionCallsite
			reduce(269), // .//, reduce: FunctionCallsite
			reduce(269), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(269), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(269), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(269), // [[, reduce: FunctionCallsite
			reduce(269), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4008
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4295), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4296), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4009
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(262), // $, reduce: ArrayOrMapPositionalValueAccess
			nil,         // empty
			reduce(262), // ;, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // {, reduce: ArrayOrMapPositionalValueAccess
			nil,         // }
			reduce(262), // =, reduce: ArrayOrMapPositionalValueAccess
			nil,         // unset
			nil,         // filter
			reduce(262), // >, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |, reduce: ArrayOrMapPositionalValueAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(262), // ||=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // **=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ?, reduce: ArrayOrMapPositionalValueAccess
			nil,         // :
			reduce(262), // ||, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // =~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ==, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .+, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .-, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .*, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ./, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .//, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ., reduce: ArrayOrMapPositionalValueAccess
			nil,         // !
			nil,         // ~
			reduce(262), // **, reduce: ArrayOrMapPositionalValueAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(262), // [, reduce: ArrayOrMapPositionalValueAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(262), // [[, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // [[[, reduce: ArrayOrMapPositionalValueAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4010
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(108), // >, reduce: PositionalFieldValue
			reduce(108), // >>, reduce: PositionalFieldValue
			reduce(108), // |, reduce: PositionalFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(108), // ,, reduce: PositionalFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(108), // ], reduce: PositionalFieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(108), // ?, reduce: PositionalFieldValue
			nil,         // :
			reduce(108), // ||, reduce: PositionalFieldValue
			reduce(108), // ^^, reduce: PositionalFieldValue
			reduce(108), // &&, reduce: PositionalFieldValue
			reduce(108), // ??, reduce: PositionalFieldValue
			reduce(108), // ???, reduce: PositionalFieldValue
			reduce(108), // =~, reduce: PositionalFieldValue
			reduce(108), // !=~, reduce: PositionalFieldValue
			reduce(108), // ==, reduce: PositionalFieldValue
			reduce(108), // !=, reduce: PositionalFieldValue
			reduce(108), // <=>, reduce: PositionalFieldValue
			reduce(108), // >=, reduce: PositionalFieldValue
			reduce(108), // <, reduce: PositionalFieldValue
			reduce(108), // <=, reduce: PositionalFieldValue
			reduce(108), // ^, reduce: PositionalFieldValue
			reduce(108), // &, reduce: PositionalFieldValue
			reduce(108), // <<, reduce: PositionalFieldValue
			reduce(108), // >>>, reduce: PositionalFieldValue
			reduce(108), // +, reduce: PositionalFieldValue
			reduce(108), // -, reduce: PositionalFieldValue
			reduce(108), // .+, reduce: PositionalFieldValue
			reduce(108), // .-, reduce: PositionalFieldValue
			reduce(108), // *, reduce: PositionalFieldValue
			reduce(108), // /, reduce: PositionalFieldValue
			reduce(108), // //, reduce: PositionalFieldValue
			reduce(108), // %, reduce: PositionalFieldValue
			reduce(108), // .*, reduce: PositionalFieldValue
			reduce(108), // ./, reduce: PositionalFieldValue
			reduce(108), // .//, reduce: PositionalFieldValue
			reduce(108), // ., reduce: PositionalFieldValue
			nil,         // !
			nil,         // ~
			reduce(108), // **, reduce: PositionalFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(108), // [, reduce: PositionalFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(108), // [[, reduce: PositionalFieldValue
			reduce(108), // [[[, reduce: PositionalFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4011
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(149), // ,, reduce: TernaryTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(149), // ], reduce: TernaryTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4012
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4297), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4013
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(265), // >, reduce: ArraySliceAccess
			reduce(265), // >>, reduce: ArraySliceAccess
			reduce(265), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(265), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(265), // ], reduce: ArraySliceAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(265), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(265), // ||, reduce: ArraySliceAccess
			reduce(265), // ^^, reduce: ArraySliceAccess
			reduce(265), // &&, reduce: ArraySliceAccess
			reduce(265), // ??, reduce: ArraySliceAccess
			reduce(265), // ???, reduce: ArraySliceAccess
			reduce(265), // =~, reduce: ArraySliceAccess
			reduce(265), // !=~, reduce: ArraySliceAccess
			reduce(265), // ==, reduce: ArraySliceAccess
			reduce(265), // !=, reduce: ArraySliceAccess
			reduce(265), // <=>, reduce: ArraySliceAccess
			reduce(265), // >=, reduce: ArraySliceAccess
			reduce(265), // <, reduce: ArraySliceAccess
			reduce(265), // <=, reduce: ArraySliceAccess
			reduce(265), // ^, reduce: ArraySliceAccess
			reduce(265), // &, reduce: ArraySliceAccess
			reduce(265), // <<, reduce: ArraySliceAccess
			reduce(265), // >>>, reduce: ArraySliceAccess
			reduce(265), // +, reduce: ArraySliceAccess
			reduce(265), // -, reduce: ArraySliceAccess
			reduce(265), // .+, reduce: ArraySliceAccess
			reduce(265), // .-, reduce: ArraySliceAccess
			reduce(265), // *, reduce: ArraySliceAccess
			reduce(265), // /, reduce: ArraySliceAccess
			reduce(265), // //, reduce: ArraySliceAccess
			reduce(265), // %, reduce: ArraySliceAccess
			reduce(265), // .*, reduce: ArraySliceAccess
			reduce(265), // ./, reduce: ArraySliceAccess
			reduce(265), // .//, reduce: ArraySliceAccess
			reduce(265), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(265), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(265), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(265), // [[, reduce: ArraySliceAccess
			reduce(265), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4014
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(264), // >, reduce: ArraySliceAccess
			reduce(264), // >>, reduce: ArraySliceAccess
			reduce(264), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(264), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(264), // ], reduce: ArraySliceAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(264), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(264), // ||, reduce: ArraySliceAccess
			reduce(264), // ^^, reduce: ArraySliceAccess
			reduce(264), // &&, reduce: ArraySliceAccess
			reduce(264), // ??, reduce: ArraySliceAccess
			reduce(264), // ???, reduce: ArraySliceAccess
			reduce(264), // =~, reduce: ArraySliceAccess
			reduce(264), // !=~, reduce: ArraySliceAccess
			reduce(264), // ==, reduce: ArraySliceAccess
			reduce(264), // !=, reduce: ArraySliceAccess
			reduce(264), // <=>, reduce: ArraySliceAccess
			reduce(264), // >=, reduce: ArraySliceAccess
			reduce(264), // <, reduce: ArraySliceAccess
			reduce(264), // <=, reduce: ArraySliceAccess
			reduce(264), // ^, reduce: ArraySliceAccess
			reduce(264), // &, reduce: ArraySliceAccess
			reduce(264), // <<, reduce: ArraySliceAccess
			reduce(264), // >>>, reduce: ArraySliceAccess
			reduce(264), // +, reduce: ArraySliceAccess
			reduce(264), // -, reduce: ArraySliceAccess
			reduce(264), // .+, reduce: ArraySliceAccess
			reduce(264), // .-, reduce: ArraySliceAccess
			reduce(264), // *, reduce: ArraySliceAccess
			reduce(264), // /, reduce: ArraySliceAccess
			reduce(264), // //, reduce: ArraySliceAccess
			reduce(264), // %, reduce: ArraySliceAccess
			reduce(264), // .*, reduce: ArraySliceAccess
			reduce(264), // ./, reduce: ArraySliceAccess
			reduce(264), // .//, reduce: ArraySliceAccess
			reduce(264), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(264), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(264), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(264), // [[, reduce: ArraySliceAccess
			reduce(264), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4015
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(261), // >, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |, reduce: ArrayOrMapPositionalNameAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(261), // ,, reduce: ArrayOrMapPositionalNameAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(261), // ], reduce: ArrayOrMapPositionalNameAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(261), // ?, reduce: ArrayOrMapPositionalNameAccess
			nil,         // :
			reduce(261), // ||, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // =~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ==, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .+, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .-, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .*, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ./, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .//, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ., reduce: ArrayOrMapPositionalNameAccess
			nil,         // !
			nil,         // ~
			reduce(261), // **, reduce: ArrayOrMapPositionalNameAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(261), // [, reduce: ArrayOrMapPositionalNameAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(261), // [[, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // [[[, reduce: ArrayOrMapPositionalNameAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4016
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4298), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4017
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(324), // >, reduce: UnnamedFunctionDefinition
			reduce(324), // >>, reduce: UnnamedFunctionDefinition
			reduce(324), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(324), // ,, reduce: UnnamedFunctionDefinition
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(324), // ], reduce: UnnamedFunctionDefinition
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(324), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(324), // ||, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^, reduce: UnnamedFunctionDefinition
			reduce(324), // &&, reduce: UnnamedFunctionDefinition
			reduce(324), // ??, reduce: UnnamedFunctionDefinition
			reduce(324), // ???, reduce: UnnamedFunctionDefinition
			reduce(324), // =~, reduce: UnnamedFunctionDefinition
			reduce(324), // !=~, reduce: UnnamedFunctionDefinition
			reduce(324), // ==, reduce: UnnamedFunctionDefinition
			reduce(324), // !=, reduce: UnnamedFunctionDefinition
			reduce(324), // <=>, reduce: UnnamedFunctionDefinition
			reduce(324), // >=, reduce: UnnamedFunctionDefinition
			reduce(324), // <, reduce: UnnamedFunctionDefinition
			reduce(324), // <=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^, reduce: UnnamedFunctionDefinition
			reduce(324), // &, reduce: UnnamedFunctionDefinition
			reduce(324), // <<, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>, reduce: UnnamedFunctionDefinition
			reduce(324), // +, reduce: UnnamedFunctionDefinition
			reduce(324), // -, reduce: UnnamedFunctionDefinition
			reduce(324), // .+, reduce: UnnamedFunctionDefinition
			reduce(324), // .-, reduce: UnnamedFunctionDefinition
			reduce(324), // *, reduce: UnnamedFunctionDefinition
			reduce(324), // /, reduce: UnnamedFunctionDefinition
			reduce(324), // //, reduce: UnnamedFunctionDefinition
			reduce(324), // %, reduce: UnnamedFunctionDefinition
			reduce(324), // .*, reduce: UnnamedFunctionDefinition
			reduce(324), // ./, reduce: UnnamedFunctionDefinition
			reduce(324), // .//, reduce: UnnamedFunctionDefinition
			reduce(324), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(324), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(324), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(324), // [[, reduce: UnnamedFunctionDefinition
			reduce(324), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4018
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S4019
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4020
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(107), // >, reduce: PositionalFieldName
			reduce(107), // >>, reduce: PositionalFieldName
			reduce(107), // |, reduce: PositionalFieldName
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(107), // ,, reduce: PositionalFieldName
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(107), // ), reduce: PositionalFieldName
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(107), // ?, reduce: PositionalFieldName
			nil,         // :
			reduce(107), // ||, reduce: PositionalFieldName
			reduce(107), // ^^, reduce: PositionalFieldName
			reduce(107), // &&, reduce: PositionalFieldName
			reduce(107), // ??, reduce: PositionalFieldName
			reduce(107), // ???, reduce: PositionalFieldName
			reduce(107), // =~, reduce: PositionalFieldName
			reduce(107), // !=~, reduce: PositionalFieldName
			reduce(107), // ==, reduce: PositionalFieldName
			reduce(107), // !=, reduce: PositionalFieldName
			reduce(107), // <=>, reduce: PositionalFieldName
			reduce(107), // >=, reduce: PositionalFieldName
			reduce(107), // <, reduce: PositionalFieldName
			reduce(107), // <=, reduce: PositionalFieldName
			reduce(107), // ^, reduce: PositionalFieldName
			reduce(107), // &, reduce: PositionalFieldName
			reduce(107), // <<, reduce: PositionalFieldName
			reduce(107), // >>>, reduce: PositionalFieldName
			reduce(107), // +, reduce: PositionalFieldName
			reduce(107), // -, reduce: PositionalFieldName
			reduce(107), // .+, reduce: PositionalFieldName
			reduce(107), // .-, reduce: PositionalFieldName
			reduce(107), // *, reduce: PositionalFieldName
			reduce(107), // /, reduce: PositionalFieldName
			reduce(107), // //, reduce: PositionalFieldName
			reduce(107), // %, reduce: PositionalFieldName
			reduce(107), // .*, reduce: PositionalFieldName
			reduce(107), // ./, reduce: PositionalFieldName
			reduce(107), // .//, reduce: PositionalFieldName
			reduce(107), // ., reduce: PositionalFieldName
			nil,         // !
			nil,         // ~
			reduce(107), // **, reduce: PositionalFieldName
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(107), // [, reduce: PositionalFieldName
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(107), // [[, reduce: PositionalFieldName
			reduce(107), // [[[, reduce: PositionalFieldName
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4021
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4301), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4022
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4023
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(206), // >, reduce: PowTerm
			reduce(206), // >>, reduce: PowTerm
			reduce(206), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(206), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(206), // ), reduce: PowTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(206), // ?, reduce: PowTerm
			nil,         // :
			reduce(206), // ||, reduce: PowTerm
			reduce(206), // ^^, reduce: PowTerm
			reduce(206), // &&, reduce: PowTerm
			reduce(206), // ??, reduce: PowTerm
			reduce(206), // ???, reduce: PowTerm
			reduce(206), // =~, reduce: PowTerm
			reduce(206), // !=~, reduce: PowTerm
			reduce(206), // ==, reduce: PowTerm
			reduce(206), // !=, reduce: PowTerm
			reduce(206), // <=>, reduce: PowTerm
			reduce(206), // >=, reduce: PowTerm
			reduce(206), // <, reduce: PowTerm
			reduce(206), // <=, reduce: PowTerm
			reduce(206), // ^, reduce: PowTerm
			reduce(206), // &, reduce: PowTerm
			reduce(206), // <<, reduce: PowTerm
			reduce(206), // >>>, reduce: PowTerm
			reduce(206), // +, reduce: PowTerm
			reduce(206), // -, reduce: PowTerm
			reduce(206), // .+, reduce: PowTerm
			reduce(206), // .-, reduce: PowTerm
			reduce(206), // *, reduce: PowTerm
			reduce(206), // /, reduce: PowTerm
			reduce(206), // //, reduce: PowTerm
			reduce(206), // %, reduce: PowTerm
			reduce(206), // .*, reduce: PowTerm
			reduce(206), // ./, reduce: PowTerm
			reduce(206), // .//, reduce: PowTerm
			reduce(206), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4024
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(205), // >, reduce: PowTerm
			reduce(205), // >>, reduce: PowTerm
			reduce(205), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(205), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(205), // ), reduce: PowTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(205), // ?, reduce: PowTerm
			nil,         // :
			reduce(205), // ||, reduce: PowTerm
			reduce(205), // ^^, reduce: PowTerm
			reduce(205), // &&, reduce: PowTerm
			reduce(205), // ??, reduce: PowTerm
			reduce(205), // ???, reduce: PowTerm
			reduce(205), // =~, reduce: PowTerm
			reduce(205), // !=~, reduce: PowTerm
			reduce(205), // ==, reduce: PowTerm
			reduce(205), // !=, reduce: PowTerm
			reduce(205), // <=>, reduce: PowTerm
			reduce(205), // >=, reduce: PowTerm
			reduce(205), // <, reduce: PowTerm
			reduce(205), // <=, reduce: PowTerm
			reduce(205), // ^, reduce: PowTerm
			reduce(205), // &, reduce: PowTerm
			reduce(205), // <<, reduce: PowTerm
			reduce(205), // >>>, reduce: PowTerm
			reduce(205), // +, reduce: PowTerm
			reduce(205), // -, reduce: PowTerm
			reduce(205), // .+, reduce: PowTerm
			reduce(205), // .-, reduce: PowTerm
			reduce(205), // *, reduce: PowTerm
			reduce(205), // /, reduce: PowTerm
			reduce(205), // //, reduce: PowTerm
			reduce(205), // %, reduce: PowTerm
			reduce(205), // .*, reduce: PowTerm
			reduce(205), // ./, reduce: PowTerm
			reduce(205), // .//, reduce: PowTerm
			reduce(205), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4025
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(260), // >, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>, reduce: ArrayOrMapIndexAccess
			reduce(260), // |, reduce: ArrayOrMapIndexAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(260), // ,, reduce: ArrayOrMapIndexAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(260), // ), reduce: ArrayOrMapIndexAccess
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(260), // ?, reduce: ArrayOrMapIndexAccess
			nil,         // :
			reduce(260), // ||, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???, reduce: ArrayOrMapIndexAccess
			reduce(260), // =~, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=~, reduce: ArrayOrMapIndexAccess
			reduce(260), // ==, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=>, reduce: ArrayOrMapIndexAccess
			reduce(260), // >=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>, reduce: ArrayOrMapIndexAccess
			reduce(260), // +, reduce: ArrayOrMapIndexAccess
			reduce(260), // -, reduce: ArrayOrMapIndexAccess
			reduce(260), // .+, reduce: ArrayOrMapIndexAccess
			reduce(260), // .-, reduce: ArrayOrMapIndexAccess
			reduce(260), // *, reduce: ArrayOrMapIndexAccess
			reduce(260), // /, reduce: ArrayOrMapIndexAccess
			reduce(260), // //, reduce: ArrayOrMapIndexAccess
			reduce(260), // %, reduce: ArrayOrMapIndexAccess
			reduce(260), // .*, reduce: ArrayOrMapIndexAccess
			reduce(260), // ./, reduce: ArrayOrMapIndexAccess
			reduce(260), // .//, reduce: ArrayOrMapIndexAccess
			reduce(260), // ., reduce: ArrayOrMapIndexAccess
			nil,         // !
			nil,         // ~
			reduce(260), // **, reduce: ArrayOrMapIndexAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(260), // [, reduce: ArrayOrMapIndexAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(260), // [[, reduce: ArrayOrMapIndexAccess
			reduce(260), // [[[, reduce: ArrayOrMapIndexAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4026
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4304), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4027
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4305), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4028
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(266), // >, reduce: ArraySliceAccess
			reduce(266), // >>, reduce: ArraySliceAccess
			reduce(266), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(266), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(266), // ), reduce: ArraySliceAccess
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(266), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(266), // ||, reduce: ArraySliceAccess
			reduce(266), // ^^, reduce: ArraySliceAccess
			reduce(266), // &&, reduce: ArraySliceAccess
			reduce(266), // ??, reduce: ArraySliceAccess
			reduce(266), // ???, reduce: ArraySliceAccess
			reduce(266), // =~, reduce: ArraySliceAccess
			reduce(266), // !=~, reduce: ArraySliceAccess
			reduce(266), // ==, reduce: ArraySliceAccess
			reduce(266), // !=, reduce: ArraySliceAccess
			reduce(266), // <=>, reduce: ArraySliceAccess
			reduce(266), // >=, reduce: ArraySliceAccess
			reduce(266), // <, reduce: ArraySliceAccess
			reduce(266), // <=, reduce: ArraySliceAccess
			reduce(266), // ^, reduce: ArraySliceAccess
			reduce(266), // &, reduce: ArraySliceAccess
			reduce(266), // <<, reduce: ArraySliceAccess
			reduce(266), // >>>, reduce: ArraySliceAccess
			reduce(266), // +, reduce: ArraySliceAccess
			reduce(266), // -, reduce: ArraySliceAccess
			reduce(266), // .+, reduce: ArraySliceAccess
			reduce(266), // .-, reduce: ArraySliceAccess
			reduce(266), // *, reduce: ArraySliceAccess
			reduce(266), // /, reduce: ArraySliceAccess
			reduce(266), // //, reduce: ArraySliceAccess
			reduce(266), // %, reduce: ArraySliceAccess
			reduce(266), // .*, reduce: ArraySliceAccess
			reduce(266), // ./, reduce: ArraySliceAccess
			reduce(266), // .//, reduce: ArraySliceAccess
			reduce(266), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(266), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(266), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(266), // [[, reduce: ArraySliceAccess
			reduce(266), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4029
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4306), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4030
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4307), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4031
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(254), // >, reduce: ENV
			reduce(254), // >>, reduce: ENV
			reduce(254), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(254), // ,, reduce: ENV
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(254), // ), reduce: ENV
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(254), // ?, reduce: ENV
			nil,         // :
			reduce(254), // ||, reduce: ENV
			reduce(254), // ^^, reduce: ENV
			reduce(254), // &&, reduce: ENV
			reduce(254), // ??, reduce: ENV
			reduce(254), // ???, reduce: ENV
			reduce(254), // =~, reduce: ENV
			reduce(254), // !=~, reduce: ENV
			reduce(254), // ==, reduce: ENV
			reduce(254), // !=, reduce: ENV
			reduce(254), // <=>, reduce: ENV
			reduce(254), // >=, reduce: ENV
			reduce(254), // <, reduce: ENV
			reduce(254), // <=, reduce: ENV
			reduce(254), // ^, reduce: ENV
			reduce(254), // &, reduce: ENV
			reduce(254), // <<, reduce: ENV
			reduce(254), // >>>, reduce: ENV
			reduce(254), // +, reduce: ENV
			reduce(254), // -, reduce: ENV
			reduce(254), // .+, reduce: ENV
			reduce(254), // .-, reduce: ENV
			reduce(254), // *, reduce: ENV
			reduce(254), // /, reduce: ENV
			reduce(254), // //, reduce: ENV
			reduce(254), // %, reduce: ENV
			reduce(254), // .*, reduce: ENV
			reduce(254), // ./, reduce: ENV
			reduce(254), // .//, reduce: ENV
			reduce(254), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(254), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(254), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(254), // [[, reduce: ENV
			reduce(254), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4032
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(269), // >, reduce: FunctionCallsite
			reduce(269), // >>, reduce: FunctionCallsite
			reduce(269), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(269), // ,, reduce: FunctionCallsite
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(269), // ), reduce: FunctionCallsite
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(269), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(269), // ||, reduce: FunctionCallsite
			reduce(269), // ^^, reduce: FunctionCallsite
			reduce(269), // &&, reduce: FunctionCallsite
			reduce(269), // ??, reduce: FunctionCallsite
			reduce(269), // ???, reduce: FunctionCallsite
			reduce(269), // =~, reduce: FunctionCallsite
			reduce(269), // !=~, reduce: FunctionCallsite
			reduce(269), // ==, reduce: FunctionCallsite
			reduce(269), // !=, reduce: FunctionCallsite
			reduce(269), // <=>, reduce: FunctionCallsite
			reduce(269), // >=, reduce: FunctionCallsite
			reduce(269), // <, reduce: FunctionCallsite
			reduce(269), // <=, reduce: FunctionCallsite
			reduce(269), // ^, reduce: FunctionCallsite
			reduce(269), // &, reduce: FunctionCallsite
			reduce(269), // <<, reduce: FunctionCallsite
			reduce(269), // >>>, reduce: FunctionCallsite
			reduce(269), // +, reduce: FunctionCallsite
			reduce(269), // -, reduce: FunctionCallsite
			reduce(269), // .+, reduce: FunctionCallsite
			reduce(269), // .-, reduce: FunctionCallsite
			reduce(269), // *, reduce: FunctionCallsite
			reduce(269), // /, reduce: FunctionCallsite
			reduce(269), // //, reduce: FunctionCallsite
			reduce(269), // %, reduce: FunctionCallsite
			reduce(269), // .*, reduce: FunctionCallsite
			reduce(269), // ./, reduce: FunctionCallsite
			reduce(269), // .//, reduce: FunctionCallsite
			reduce(269), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(269), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(269), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(269), // [[, reduce: FunctionCallsite
			reduce(269), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4033
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4309), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4310), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4034
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(295), // $, reduce: ElifBlock
			nil,         // empty
			reduce(295), // ;, reduce: ElifBlock
			reduce(295), // {, reduce: ElifBlock
			nil,         // }
			nil,         // =
			reduce(295), // unset, reduce: ElifBlock
			reduce(295), // filter, reduce: ElifBlock
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(295), // print, reduce: ElifBlock
			nil,         // ,
			reduce(295), // printn, reduce: ElifBlock
			reduce(295), // eprint, reduce: ElifBlock
			reduce(295), // eprintn, reduce: ElifBlock
			reduce(295), // dump, reduce: ElifBlock
			reduce(295), // edump, reduce: ElifBlock
			reduce(295), // tee, reduce: ElifBlock
			reduce(295), // emitf, reduce: ElifBlock
			reduce(295), // emit1, reduce: ElifBlock
			reduce(295), // emit, reduce: ElifBlock
			reduce(295), // (, reduce: ElifBlock
			nil,         // )
			reduce(295), // emitp, reduce: ElifBlock
			reduce(295), // field_name, reduce: ElifBlock
			reduce(295), // $[, reduce: ElifBlock
			nil,         // ]
			reduce(295), // braced_field_name, reduce: ElifBlock
			reduce(295), // $[[, reduce: ElifBlock
			reduce(295), // $[[[, reduce: ElifBlock
			reduce(295), // full_srec, reduce: ElifBlock
			reduce(295), // oosvar_name, reduce: ElifBlock
			reduce(295), // @[, reduce: ElifBlock
			reduce(295), // braced_oosvar_name, reduce: ElifBlock
			reduce(295), // full_oosvar, reduce: ElifBlock
			reduce(295), // all, reduce: ElifBlock
			reduce(295), // non_sigil_name, reduce: ElifBlock
			reduce(295), // arr, reduce: ElifBlock
			reduce(295), // bool, reduce: ElifBlock
			reduce(295), // float, reduce: ElifBlock
			reduce(295), // int, reduce: ElifBlock
			reduce(295), // map, reduce: ElifBlock
			reduce(295), // num, reduce: ElifBlock
			reduce(295), // str, reduce: ElifBlock
			reduce(295), // var, reduce: ElifBlock
			reduce(295), // funct, reduce: ElifBlock
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(295), // +, reduce: ElifBlock
			reduce(295), // -, reduce: ElifBlock
			reduce(295), // .+, reduce: ElifBlock
			reduce(295), // .-, reduce: ElifBlock
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(295), // !, reduce: ElifBlock
			reduce(295), // ~, reduce: ElifBlock
			nil,         // **
			reduce(295), // string_literal, reduce: ElifBlock
			reduce(295), // regex_case_insensitive, reduce: ElifBlock
			reduce(295), // int_literal, reduce: ElifBlock
			reduce(295), // float_literal, reduce: ElifBlock
			reduce(295), // boolean_literal, reduce: ElifBlock
			reduce(295), // null_literal, reduce: ElifBlock
			reduce(295), // inf_literal, reduce: ElifBlock
			reduce(295), // nan_literal, reduce: ElifBlock
			reduce(295), // const_M_PI, reduce: ElifBlock
			reduce(295), // const_M_E, reduce: ElifBlock
			reduce(295), // panic, reduce: ElifBlock
			reduce(295), // [, reduce: ElifBlock
			reduce(295), // ctx_IPS, reduce: ElifBlock
			reduce(295), // ctx_IFS, reduce: ElifBlock
			reduce(295), // ctx_IRS, reduce: ElifBlock
			reduce(295), // ctx_OPS, reduce: ElifBlock
			reduce(295), // ctx_OFS, reduce: ElifBlock
			reduce(295), // ctx_ORS, reduce: ElifBlock
			reduce(295), // ctx_FLATSEP, reduce: ElifBlock
			reduce(295), // ctx_NF, reduce: ElifBlock
			reduce(295), // ctx_NR, reduce: ElifBlock
			reduce(295), // ctx_FNR, reduce: ElifBlock
			reduce(295), // ctx_FILENAME, reduce: ElifBlock
			reduce(295), // ctx_FILENUM, reduce: ElifBlock
			reduce(295), // env, reduce: ElifBlock
			nil,         // [[
			nil,         // [[[
			reduce(295), // call, reduce: ElifBlock
			reduce(295), // begin, reduce: ElifBlock
			reduce(295), // end, reduce: ElifBlock
			reduce(295), // if, reduce: ElifBlock
			reduce(295), // elif, reduce: ElifBlock
			reduce(295), // else, reduce: ElifBlock
			reduce(295), // while, reduce: ElifBlock
			reduce(295), // do, reduce: ElifBlock
			reduce(295), // for, reduce: ElifBlock
			nil,         // in
			reduce(295), // break, reduce: ElifBlock
			reduce(295), // continue, reduce: ElifBlock
			reduce(295), // func, reduce: ElifBlock
			reduce(295), // subr, reduce: ElifBlock
			reduce(295), // return, reduce: ElifBlock
		},
	},
	actionRow{ // S4035
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4311), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4036
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(298), // $, reduce: DoWhileLoop
			nil,         // empty
			reduce(298), // ;, reduce: DoWhileLoop
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4037
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(234), // ;, reduce: MapLiteral
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(234), // >, reduce: MapLiteral
			reduce(234), // >>, reduce: MapLiteral
			reduce(234), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(234), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(234), // ?, reduce: MapLiteral
			nil,         // :
			reduce(234), // ||, reduce: MapLiteral
			reduce(234), // ^^, reduce: MapLiteral
			reduce(234), // &&, reduce: MapLiteral
			reduce(234), // ??, reduce: MapLiteral
			reduce(234), // ???, reduce: MapLiteral
			reduce(234), // =~, reduce: MapLiteral
			reduce(234), // !=~, reduce: MapLiteral
			reduce(234), // ==, reduce: MapLiteral
			reduce(234), // !=, reduce: MapLiteral
			reduce(234), // <=>, reduce: MapLiteral
			reduce(234), // >=, reduce: MapLiteral
			reduce(234), // <, reduce: MapLiteral
			reduce(234), // <=, reduce: MapLiteral
			reduce(234), // ^, reduce: MapLiteral
			reduce(234), // &, reduce: MapLiteral
			reduce(234), // <<, reduce: MapLiteral
			reduce(234), // >>>, reduce: MapLiteral
			reduce(234), // +, reduce: MapLiteral
			reduce(234), // -, reduce: MapLiteral
			reduce(234), // .+, reduce: MapLiteral
			reduce(234), // .-, reduce: MapLiteral
			reduce(234), // *, reduce: MapLiteral
			reduce(234), // /, reduce: MapLiteral
			reduce(234), // //, reduce: MapLiteral
			reduce(234), // %, reduce: MapLiteral
			reduce(234), // .*, reduce: MapLiteral
			reduce(234), // ./, reduce: MapLiteral
			reduce(234), // .//, reduce: MapLiteral
			reduce(234), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(234), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(234), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(234), // [[, reduce: MapLiteral
			reduce(234), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4038
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4312), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4039
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4313), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4040
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4314), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4041
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4315), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4042
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4316), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4043
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4317), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4044
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4045
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4046
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4047
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4048
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4049
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4050
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4051
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4052
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4053
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4054
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4055
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4056
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4057
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4058
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4059
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4060
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4061
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4062
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4063
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4064
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4065
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4066
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4067
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4068
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4069
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(197), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(197), // >, reduce: UnaryOpTerm
			reduce(197), // >>, reduce: UnaryOpTerm
			reduce(197), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(197), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(197), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(197), // ||, reduce: UnaryOpTerm
			reduce(197), // ^^, reduce: UnaryOpTerm
			reduce(197), // &&, reduce: UnaryOpTerm
			reduce(197), // ??, reduce: UnaryOpTerm
			reduce(197), // ???, reduce: UnaryOpTerm
			reduce(197), // =~, reduce: UnaryOpTerm
			reduce(197), // !=~, reduce: UnaryOpTerm
			reduce(197), // ==, reduce: UnaryOpTerm
			reduce(197), // !=, reduce: UnaryOpTerm
			reduce(197), // <=>, reduce: UnaryOpTerm
			reduce(197), // >=, reduce: UnaryOpTerm
			reduce(197), // <, reduce: UnaryOpTerm
			reduce(197), // <=, reduce: UnaryOpTerm
			reduce(197), // ^, reduce: UnaryOpTerm
			reduce(197), // &, reduce: UnaryOpTerm
			reduce(197), // <<, reduce: UnaryOpTerm
			reduce(197), // >>>, reduce: UnaryOpTerm
			reduce(197), // +, reduce: UnaryOpTerm
			reduce(197), // -, reduce: UnaryOpTerm
			reduce(197), // .+, reduce: UnaryOpTerm
			reduce(197), // .-, reduce: UnaryOpTerm
			reduce(197), // *, reduce: UnaryOpTerm
			reduce(197), // /, reduce: UnaryOpTerm
			reduce(197), // //, reduce: UnaryOpTerm
			reduce(197), // %, reduce: UnaryOpTerm
			reduce(197), // .*, reduce: UnaryOpTerm
			reduce(197), // ./, reduce: UnaryOpTerm
			reduce(197), // .//, reduce: UnaryOpTerm
			reduce(197), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4070
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4071
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4072
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4073
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4074
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4075
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4076
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4077
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(198), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(198), // >, reduce: UnaryOpTerm
			reduce(198), // >>, reduce: UnaryOpTerm
			reduce(198), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(198), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(198), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(198), // ||, reduce: UnaryOpTerm
			reduce(198), // ^^, reduce: UnaryOpTerm
			reduce(198), // &&, reduce: UnaryOpTerm
			reduce(198), // ??, reduce: UnaryOpTerm
			reduce(198), // ???, reduce: UnaryOpTerm
			reduce(198), // =~, reduce: UnaryOpTerm
			reduce(198), // !=~, reduce: UnaryOpTerm
			reduce(198), // ==, reduce: UnaryOpTerm
			reduce(198), // !=, reduce: UnaryOpTerm
			reduce(198), // <=>, reduce: UnaryOpTerm
			reduce(198), // >=, reduce: UnaryOpTerm
			reduce(198), // <, reduce: UnaryOpTerm
			reduce(198), // <=, reduce: UnaryOpTerm
			reduce(198), // ^, reduce: UnaryOpTerm
			reduce(198), // &, reduce: UnaryOpTerm
			reduce(198), // <<, reduce: UnaryOpTerm
			reduce(198), // >>>, reduce: UnaryOpTerm
			reduce(198), // +, reduce: UnaryOpTerm
			reduce(198), // -, reduce: UnaryOpTerm
			reduce(198), // .+, reduce: UnaryOpTerm
			reduce(198), // .-, reduce: UnaryOpTerm
			reduce(198), // *, reduce: UnaryOpTerm
			reduce(198), // /, reduce: UnaryOpTerm
			reduce(198), // //, reduce: UnaryOpTerm
			reduce(198), // %, reduce: UnaryOpTerm
			reduce(198), // .*, reduce: UnaryOpTerm
			reduce(198), // ./, reduce: UnaryOpTerm
			reduce(198), // .//, reduce: UnaryOpTerm
			reduce(198), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4078
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(199), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(199), // >, reduce: UnaryOpTerm
			reduce(199), // >>, reduce: UnaryOpTerm
			reduce(199), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(199), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(199), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(199), // ||, reduce: UnaryOpTerm
			reduce(199), // ^^, reduce: UnaryOpTerm
			reduce(199), // &&, reduce: UnaryOpTerm
			reduce(199), // ??, reduce: UnaryOpTerm
			reduce(199), // ???, reduce: UnaryOpTerm
			reduce(199), // =~, reduce: UnaryOpTerm
			reduce(199), // !=~, reduce: UnaryOpTerm
			reduce(199), // ==, reduce: UnaryOpTerm
			reduce(199), // !=, reduce: UnaryOpTerm
			reduce(199), // <=>, reduce: UnaryOpTerm
			reduce(199), // >=, reduce: UnaryOpTerm
			reduce(199), // <, reduce: UnaryOpTerm
			reduce(199), // <=, reduce: UnaryOpTerm
			reduce(199), // ^, reduce: UnaryOpTerm
			reduce(199), // &, reduce: UnaryOpTerm
			reduce(199), // <<, reduce: UnaryOpTerm
			reduce(199), // >>>, reduce: UnaryOpTerm
			reduce(199), // +, reduce: UnaryOpTerm
			reduce(199), // -, reduce: UnaryOpTerm
			reduce(199), // .+, reduce: UnaryOpTerm
			reduce(199), // .-, reduce: UnaryOpTerm
			reduce(199), // *, reduce: UnaryOpTerm
			reduce(199), // /, reduce: UnaryOpTerm
			reduce(199), // //, reduce: UnaryOpTerm
			reduce(199), // %, reduce: UnaryOpTerm
			reduce(199), // .*, reduce: UnaryOpTerm
			reduce(199), // ./, reduce: UnaryOpTerm
			reduce(199), // .//, reduce: UnaryOpTerm
			reduce(199), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4079
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(200), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(200), // >, reduce: UnaryOpTerm
			reduce(200), // >>, reduce: UnaryOpTerm
			reduce(200), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(200), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(200), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(200), // ||, reduce: UnaryOpTerm
			reduce(200), // ^^, reduce: UnaryOpTerm
			reduce(200), // &&, reduce: UnaryOpTerm
			reduce(200), // ??, reduce: UnaryOpTerm
			reduce(200), // ???, reduce: UnaryOpTerm
			reduce(200), // =~, reduce: UnaryOpTerm
			reduce(200), // !=~, reduce: UnaryOpTerm
			reduce(200), // ==, reduce: UnaryOpTerm
			reduce(200), // !=, reduce: UnaryOpTerm
			reduce(200), // <=>, reduce: UnaryOpTerm
			reduce(200), // >=, reduce: UnaryOpTerm
			reduce(200), // <, reduce: UnaryOpTerm
			reduce(200), // <=, reduce: UnaryOpTerm
			reduce(200), // ^, reduce: UnaryOpTerm
			reduce(200), // &, reduce: UnaryOpTerm
			reduce(200), // <<, reduce: UnaryOpTerm
			reduce(200), // >>>, reduce: UnaryOpTerm
			reduce(200), // +, reduce: UnaryOpTerm
			reduce(200), // -, reduce: UnaryOpTerm
			reduce(200), // .+, reduce: UnaryOpTerm
			reduce(200), // .-, reduce: UnaryOpTerm
			reduce(200), // *, reduce: UnaryOpTerm
			reduce(200), // /, reduce: UnaryOpTerm
			reduce(200), // //, reduce: UnaryOpTerm
			reduce(200), // %, reduce: UnaryOpTerm
			reduce(200), // .*, reduce: UnaryOpTerm
			reduce(200), // ./, reduce: UnaryOpTerm
			reduce(200), // .//, reduce: UnaryOpTerm
			reduce(200), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4080
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4081
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(201), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(201), // >, reduce: UnaryOpTerm
			reduce(201), // >>, reduce: UnaryOpTerm
			reduce(201), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(201), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(201), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(201), // ||, reduce: UnaryOpTerm
			reduce(201), // ^^, reduce: UnaryOpTerm
			reduce(201), // &&, reduce: UnaryOpTerm
			reduce(201), // ??, reduce: UnaryOpTerm
			reduce(201), // ???, reduce: UnaryOpTerm
			reduce(201), // =~, reduce: UnaryOpTerm
			reduce(201), // !=~, reduce: UnaryOpTerm
			reduce(201), // ==, reduce: UnaryOpTerm
			reduce(201), // !=, reduce: UnaryOpTerm
			reduce(201), // <=>, reduce: UnaryOpTerm
			reduce(201), // >=, reduce: UnaryOpTerm
			reduce(201), // <, reduce: UnaryOpTerm
			reduce(201), // <=, reduce: UnaryOpTerm
			reduce(201), // ^, reduce: UnaryOpTerm
			reduce(201), // &, reduce: UnaryOpTerm
			reduce(201), // <<, reduce: UnaryOpTerm
			reduce(201), // >>>, reduce: UnaryOpTerm
			reduce(201), // +, reduce: UnaryOpTerm
			reduce(201), // -, reduce: UnaryOpTerm
			reduce(201), // .+, reduce: UnaryOpTerm
			reduce(201), // .-, reduce: UnaryOpTerm
			reduce(201), // *, reduce: UnaryOpTerm
			reduce(201), // /, reduce: UnaryOpTerm
			reduce(201), // //, reduce: UnaryOpTerm
			reduce(201), // %, reduce: UnaryOpTerm
			reduce(201), // .*, reduce: UnaryOpTerm
			reduce(201), // ./, reduce: UnaryOpTerm
			reduce(201), // .//, reduce: UnaryOpTerm
			reduce(201), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4082
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(202), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(202), // >, reduce: UnaryOpTerm
			reduce(202), // >>, reduce: UnaryOpTerm
			reduce(202), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(202), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(202), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(202), // ||, reduce: UnaryOpTerm
			reduce(202), // ^^, reduce: UnaryOpTerm
			reduce(202), // &&, reduce: UnaryOpTerm
			reduce(202), // ??, reduce: UnaryOpTerm
			reduce(202), // ???, reduce: UnaryOpTerm
			reduce(202), // =~, reduce: UnaryOpTerm
			reduce(202), // !=~, reduce: UnaryOpTerm
			reduce(202), // ==, reduce: UnaryOpTerm
			reduce(202), // !=, reduce: UnaryOpTerm
			reduce(202), // <=>, reduce: UnaryOpTerm
			reduce(202), // >=, reduce: UnaryOpTerm
			reduce(202), // <, reduce: UnaryOpTerm
			reduce(202), // <=, reduce: UnaryOpTerm
			reduce(202), // ^, reduce: UnaryOpTerm
			reduce(202), // &, reduce: UnaryOpTerm
			reduce(202), // <<, reduce: UnaryOpTerm
			reduce(202), // >>>, reduce: UnaryOpTerm
			reduce(202), // +, reduce: UnaryOpTerm
			reduce(202), // -, reduce: UnaryOpTerm
			reduce(202), // .+, reduce: UnaryOpTerm
			reduce(202), // .-, reduce: UnaryOpTerm
			reduce(202), // *, reduce: UnaryOpTerm
			reduce(202), // /, reduce: UnaryOpTerm
			reduce(202), // //, reduce: UnaryOpTerm
			reduce(202), // %, reduce: UnaryOpTerm
			reduce(202), // .*, reduce: UnaryOpTerm
			reduce(202), // ./, reduce: UnaryOpTerm
			reduce(202), // .//, reduce: UnaryOpTerm
			reduce(202), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4083
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(4351), // +
			shift(4352), // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4084
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4355), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4085
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4086
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4087
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(228), // ;, reduce: ArrayLiteral
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(228), // >, reduce: ArrayLiteral
			reduce(228), // >>, reduce: ArrayLiteral
			reduce(228), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(228), // ,, reduce: ArrayLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(228), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(228), // ||, reduce: ArrayLiteral
			reduce(228), // ^^, reduce: ArrayLiteral
			reduce(228), // &&, reduce: ArrayLiteral
			reduce(228), // ??, reduce: ArrayLiteral
			reduce(228), // ???, reduce: ArrayLiteral
			reduce(228), // =~, reduce: ArrayLiteral
			reduce(228), // !=~, reduce: ArrayLiteral
			reduce(228), // ==, reduce: ArrayLiteral
			reduce(228), // !=, reduce: ArrayLiteral
			reduce(228), // <=>, reduce: ArrayLiteral
			reduce(228), // >=, reduce: ArrayLiteral
			reduce(228), // <, reduce: ArrayLiteral
			reduce(228), // <=, reduce: ArrayLiteral
			reduce(228), // ^, reduce: ArrayLiteral
			reduce(228), // &, reduce: ArrayLiteral
			reduce(228), // <<, reduce: ArrayLiteral
			reduce(228), // >>>, reduce: ArrayLiteral
			reduce(228), // +, reduce: ArrayLiteral
			reduce(228), // -, reduce: ArrayLiteral
			reduce(228), // .+, reduce: ArrayLiteral
			reduce(228), // .-, reduce: ArrayLiteral
			reduce(228), // *, reduce: ArrayLiteral
			reduce(228), // /, reduce: ArrayLiteral
			reduce(228), // //, reduce: ArrayLiteral
			reduce(228), // %, reduce: ArrayLiteral
			reduce(228), // .*, reduce: ArrayLiteral
			reduce(228), // ./, reduce: ArrayLiteral
			reduce(228), // .//, reduce: ArrayLiteral
			reduce(228), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(228), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(228), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(228), // [[, reduce: ArrayLiteral
			reduce(228), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4088
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4358), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4089
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(4359), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4090
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4091
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(4362), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4092
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4093
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4094
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(289), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4095
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(306), // ,, reduce: MultiIndex
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(306), // ), reduce: MultiIndex
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4096
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(307), // ,, reduce: MultiIndex
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(307), // ), reduce: MultiIndex
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4097
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(3561), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4098
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(107), // =, reduce: PositionalFieldName
			nil,         // unset
			nil,         // filter
			reduce(107), // >, reduce: PositionalFieldName
			reduce(107), // >>, reduce: PositionalFieldName
			reduce(107), // |, reduce: PositionalFieldName
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(107), // ||=, reduce: PositionalFieldName
			reduce(107), // ^^=, reduce: PositionalFieldName
			reduce(107), // &&=, reduce: PositionalFieldName
			reduce(107), // ??=, reduce: PositionalFieldName
			reduce(107), // ???=, reduce: PositionalFieldName
			reduce(107), // |=, reduce: PositionalFieldName
			reduce(107), // &=, reduce: PositionalFieldName
			reduce(107), // ^=, reduce: PositionalFieldName
			reduce(107), // <<=, reduce: PositionalFieldName
			reduce(107), // >>=, reduce: PositionalFieldName
			reduce(107), // >>>=, reduce: PositionalFieldName
			reduce(107), // +=, reduce: PositionalFieldName
			reduce(107), // .=, reduce: PositionalFieldName
			reduce(107), // -=, reduce: PositionalFieldName
			reduce(107), // *=, reduce: PositionalFieldName
			reduce(107), // /=, reduce: PositionalFieldName
			reduce(107), // //=, reduce: PositionalFieldName
			reduce(107), // %=, reduce: PositionalFieldName
			reduce(107), // **=, reduce: PositionalFieldName
			reduce(107), // ?, reduce: PositionalFieldName
			nil,         // :
			reduce(107), // ||, reduce: PositionalFieldName
			reduce(107), // ^^, reduce: PositionalFieldName
			reduce(107), // &&, reduce: PositionalFieldName
			reduce(107), // ??, reduce: PositionalFieldName
			reduce(107), // ???, reduce: PositionalFieldName
			reduce(107), // =~, reduce: PositionalFieldName
			reduce(107), // !=~, reduce: PositionalFieldName
			reduce(107), // ==, reduce: PositionalFieldName
			reduce(107), // !=, reduce: PositionalFieldName
			reduce(107), // <=>, reduce: PositionalFieldName
			reduce(107), // >=, reduce: PositionalFieldName
			reduce(107), // <, reduce: PositionalFieldName
			reduce(107), // <=, reduce: PositionalFieldName
			reduce(107), // ^, reduce: PositionalFieldName
			reduce(107), // &, reduce: PositionalFieldName
			reduce(107), // <<, reduce: PositionalFieldName
			reduce(107), // >>>, reduce: PositionalFieldName
			reduce(107), // +, reduce: PositionalFieldName
			reduce(107), // -, reduce: PositionalFieldName
			reduce(107), // .+, reduce: PositionalFieldName
			reduce(107), // .-, reduce: PositionalFieldName
			reduce(107), // *, reduce: PositionalFieldName
			reduce(107), // /, reduce: PositionalFieldName
			reduce(107), // //, reduce: PositionalFieldName
			reduce(107), // %, reduce: PositionalFieldName
			reduce(107), // .*, reduce: PositionalFieldName
			reduce(107), // ./, reduce: PositionalFieldName
			reduce(107), // .//, reduce: PositionalFieldName
			reduce(107), // ., reduce: PositionalFieldName
			nil,         // !
			nil,         // ~
			reduce(107), // **, reduce: PositionalFieldName
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(107), // [, reduce: PositionalFieldName
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(107), // [[, reduce: PositionalFieldName
			reduce(107), // [[[, reduce: PositionalFieldName
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4099
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4367), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4100
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4101
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(206), // =, reduce: PowTerm
			nil,         // unset
			nil,         // filter
			reduce(206), // >, reduce: PowTerm
			reduce(206), // >>, reduce: PowTerm
			reduce(206), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(206), // ||=, reduce: PowTerm
			reduce(206), // ^^=, reduce: PowTerm
			reduce(206), // &&=, reduce: PowTerm
			reduce(206), // ??=, reduce: PowTerm
			reduce(206), // ???=, reduce: PowTerm
			reduce(206), // |=, reduce: PowTerm
			reduce(206), // &=, reduce: PowTerm
			reduce(206), // ^=, reduce: PowTerm
			reduce(206), // <<=, reduce: PowTerm
			reduce(206), // >>=, reduce: PowTerm
			reduce(206), // >>>=, reduce: PowTerm
			reduce(206), // +=, reduce: PowTerm
			reduce(206), // .=, reduce: PowTerm
			reduce(206), // -=, reduce: PowTerm
			reduce(206), // *=, reduce: PowTerm
			reduce(206), // /=, reduce: PowTerm
			reduce(206), // //=, reduce: PowTerm
			reduce(206), // %=, reduce: PowTerm
			reduce(206), // **=, reduce: PowTerm
			reduce(206), // ?, reduce: PowTerm
			nil,         // :
			reduce(206), // ||, reduce: PowTerm
			reduce(206), // ^^, reduce: PowTerm
			reduce(206), // &&, reduce: PowTerm
			reduce(206), // ??, reduce: PowTerm
			reduce(206), // ???, reduce: PowTerm
			reduce(206), // =~, reduce: PowTerm
			reduce(206), // !=~, reduce: PowTerm
			reduce(206), // ==, reduce: PowTerm
			reduce(206), // !=, reduce: PowTerm
			reduce(206), // <=>, reduce: PowTerm
			reduce(206), // >=, reduce: PowTerm
			reduce(206), // <, reduce: PowTerm
			reduce(206), // <=, reduce: PowTerm
			reduce(206), // ^, reduce: PowTerm
			reduce(206), // &, reduce: PowTerm
			reduce(206), // <<, reduce: PowTerm
			reduce(206), // >>>, reduce: PowTerm
			reduce(206), // +, reduce: PowTerm
			reduce(206), // -, reduce: PowTerm
			reduce(206), // .+, reduce: PowTerm
			reduce(206), // .-, reduce: PowTerm
			reduce(206), // *, reduce: PowTerm
			reduce(206), // /, reduce: PowTerm
			reduce(206), // //, reduce: PowTerm
			reduce(206), // %, reduce: PowTerm
			reduce(206), // .*, reduce: PowTerm
			reduce(206), // ./, reduce: PowTerm
			reduce(206), // .//, reduce: PowTerm
			reduce(206), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4102
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(205), // =, reduce: PowTerm
			nil,         // unset
			nil,         // filter
			reduce(205), // >, reduce: PowTerm
			reduce(205), // >>, reduce: PowTerm
			reduce(205), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(205), // ||=, reduce: PowTerm
			reduce(205), // ^^=, reduce: PowTerm
			reduce(205), // &&=, reduce: PowTerm
			reduce(205), // ??=, reduce: PowTerm
			reduce(205), // ???=, reduce: PowTerm
			reduce(205), // |=, reduce: PowTerm
			reduce(205), // &=, reduce: PowTerm
			reduce(205), // ^=, reduce: PowTerm
			reduce(205), // <<=, reduce: PowTerm
			reduce(205), // >>=, reduce: PowTerm
			reduce(205), // >>>=, reduce: PowTerm
			reduce(205), // +=, reduce: PowTerm
			reduce(205), // .=, reduce: PowTerm
			reduce(205), // -=, reduce: PowTerm
			reduce(205), // *=, reduce: PowTerm
			reduce(205), // /=, reduce: PowTerm
			reduce(205), // //=, reduce: PowTerm
			reduce(205), // %=, reduce: PowTerm
			reduce(205), // **=, reduce: PowTerm
			reduce(205), // ?, reduce: PowTerm
			nil,         // :
			reduce(205), // ||, reduce: PowTerm
			reduce(205), // ^^, reduce: PowTerm
			reduce(205), // &&, reduce: PowTerm
			reduce(205), // ??, reduce: PowTerm
			reduce(205), // ???, reduce: PowTerm
			reduce(205), // =~, reduce: PowTerm
			reduce(205), // !=~, reduce: PowTerm
			reduce(205), // ==, reduce: PowTerm
			reduce(205), // !=, reduce: PowTerm
			reduce(205), // <=>, reduce: PowTerm
			reduce(205), // >=, reduce: PowTerm
			reduce(205), // <, reduce: PowTerm
			reduce(205), // <=, reduce: PowTerm
			reduce(205), // ^, reduce: PowTerm
			reduce(205), // &, reduce: PowTerm
			reduce(205), // <<, reduce: PowTerm
			reduce(205), // >>>, reduce: PowTerm
			reduce(205), // +, reduce: PowTerm
			reduce(205), // -, reduce: PowTerm
			reduce(205), // .+, reduce: PowTerm
			reduce(205), // .-, reduce: PowTerm
			reduce(205), // *, reduce: PowTerm
			reduce(205), // /, reduce: PowTerm
			reduce(205), // //, reduce: PowTerm
			reduce(205), // %, reduce: PowTerm
			reduce(205), // .*, reduce: PowTerm
			reduce(205), // ./, reduce: PowTerm
			reduce(205), // .//, reduce: PowTerm
			reduce(205), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4103
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(260), // =, reduce: ArrayOrMapIndexAccess
			nil,         // unset
			nil,         // filter
			reduce(260), // >, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>, reduce: ArrayOrMapIndexAccess
			reduce(260), // |, reduce: ArrayOrMapIndexAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(260), // ||=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^=, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???=, reduce: ArrayOrMapIndexAccess
			reduce(260), // |=, reduce: ArrayOrMapIndexAccess
			reduce(260), // &=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<=, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>=, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>=, reduce: ArrayOrMapIndexAccess
			reduce(260), // +=, reduce: ArrayOrMapIndexAccess
			reduce(260), // .=, reduce: ArrayOrMapIndexAccess
			reduce(260), // -=, reduce: ArrayOrMapIndexAccess
			reduce(260), // *=, reduce: ArrayOrMapIndexAccess
			reduce(260), // /=, reduce: ArrayOrMapIndexAccess
			reduce(260), // //=, reduce: ArrayOrMapIndexAccess
			reduce(260), // %=, reduce: ArrayOrMapIndexAccess
			reduce(260), // **=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ?, reduce: ArrayOrMapIndexAccess
			nil,         // :
			reduce(260), // ||, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???, reduce: ArrayOrMapIndexAccess
			reduce(260), // =~, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=~, reduce: ArrayOrMapIndexAccess
			reduce(260), // ==, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=>, reduce: ArrayOrMapIndexAccess
			reduce(260), // >=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>, reduce: ArrayOrMapIndexAccess
			reduce(260), // +, reduce: ArrayOrMapIndexAccess
			reduce(260), // -, reduce: ArrayOrMapIndexAccess
			reduce(260), // .+, reduce: ArrayOrMapIndexAccess
			reduce(260), // .-, reduce: ArrayOrMapIndexAccess
			reduce(260), // *, reduce: ArrayOrMapIndexAccess
			reduce(260), // /, reduce: ArrayOrMapIndexAccess
			reduce(260), // //, reduce: ArrayOrMapIndexAccess
			reduce(260), // %, reduce: ArrayOrMapIndexAccess
			reduce(260), // .*, reduce: ArrayOrMapIndexAccess
			reduce(260), // ./, reduce: ArrayOrMapIndexAccess
			reduce(260), // .//, reduce: ArrayOrMapIndexAccess
			reduce(260), // ., reduce: ArrayOrMapIndexAccess
			nil,         // !
			nil,         // ~
			reduce(260), // **, reduce: ArrayOrMapIndexAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(260), // [, reduce: ArrayOrMapIndexAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(260), // [[, reduce: ArrayOrMapIndexAccess
			reduce(260), // [[[, reduce: ArrayOrMapIndexAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4104
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4370), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4105
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4371), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4106
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(266), // =, reduce: ArraySliceAccess
			nil,         // unset
			nil,         // filter
			reduce(266), // >, reduce: ArraySliceAccess
			reduce(266), // >>, reduce: ArraySliceAccess
			reduce(266), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(266), // ||=, reduce: ArraySliceAccess
			reduce(266), // ^^=, reduce: ArraySliceAccess
			reduce(266), // &&=, reduce: ArraySliceAccess
			reduce(266), // ??=, reduce: ArraySliceAccess
			reduce(266), // ???=, reduce: ArraySliceAccess
			reduce(266), // |=, reduce: ArraySliceAccess
			reduce(266), // &=, reduce: ArraySliceAccess
			reduce(266), // ^=, reduce: ArraySliceAccess
			reduce(266), // <<=, reduce: ArraySliceAccess
			reduce(266), // >>=, reduce: ArraySliceAccess
			reduce(266), // >>>=, reduce: ArraySliceAccess
			reduce(266), // +=, reduce: ArraySliceAccess
			reduce(266), // .=, reduce: ArraySliceAccess
			reduce(266), // -=, reduce: ArraySliceAccess
			reduce(266), // *=, reduce: ArraySliceAccess
			reduce(266), // /=, reduce: ArraySliceAccess
			reduce(266), // //=, reduce: ArraySliceAccess
			reduce(266), // %=, reduce: ArraySliceAccess
			reduce(266), // **=, reduce: ArraySliceAccess
			reduce(266), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(266), // ||, reduce: ArraySliceAccess
			reduce(266), // ^^, reduce: ArraySliceAccess
			reduce(266), // &&, reduce: ArraySliceAccess
			reduce(266), // ??, reduce: ArraySliceAccess
			reduce(266), // ???, reduce: ArraySliceAccess
			reduce(266), // =~, reduce: ArraySliceAccess
			reduce(266), // !=~, reduce: ArraySliceAccess
			reduce(266), // ==, reduce: ArraySliceAccess
			reduce(266), // !=, reduce: ArraySliceAccess
			reduce(266), // <=>, reduce: ArraySliceAccess
			reduce(266), // >=, reduce: ArraySliceAccess
			reduce(266), // <, reduce: ArraySliceAccess
			reduce(266), // <=, reduce: ArraySliceAccess
			reduce(266), // ^, reduce: ArraySliceAccess
			reduce(266), // &, reduce: ArraySliceAccess
			reduce(266), // <<, reduce: ArraySliceAccess
			reduce(266), // >>>, reduce: ArraySliceAccess
			reduce(266), // +, reduce: ArraySliceAccess
			reduce(266), // -, reduce: ArraySliceAccess
			reduce(266), // .+, reduce: ArraySliceAccess
			reduce(266), // .-, reduce: ArraySliceAccess
			reduce(266), // *, reduce: ArraySliceAccess
			reduce(266), // /, reduce: ArraySliceAccess
			reduce(266), // //, reduce: ArraySliceAccess
			reduce(266), // %, reduce: ArraySliceAccess
			reduce(266), // .*, reduce: ArraySliceAccess
			reduce(266), // ./, reduce: ArraySliceAccess
			reduce(266), // .//, reduce: ArraySliceAccess
			reduce(266), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(266), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(266), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(266), // [[, reduce: ArraySliceAccess
			reduce(266), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4107
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4372), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4108
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4373), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4109
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(254), // =, reduce: ENV
			nil,         // unset
			nil,         // filter
			reduce(254), // >, reduce: ENV
			reduce(254), // >>, reduce: ENV
			reduce(254), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(254), // ||=, reduce: ENV
			reduce(254), // ^^=, reduce: ENV
			reduce(254), // &&=, reduce: ENV
			reduce(254), // ??=, reduce: ENV
			reduce(254), // ???=, reduce: ENV
			reduce(254), // |=, reduce: ENV
			reduce(254), // &=, reduce: ENV
			reduce(254), // ^=, reduce: ENV
			reduce(254), // <<=, reduce: ENV
			reduce(254), // >>=, reduce: ENV
			reduce(254), // >>>=, reduce: ENV
			reduce(254), // +=, reduce: ENV
			reduce(254), // .=, reduce: ENV
			reduce(254), // -=, reduce: ENV
			reduce(254), // *=, reduce: ENV
			reduce(254), // /=, reduce: ENV
			reduce(254), // //=, reduce: ENV
			reduce(254), // %=, reduce: ENV
			reduce(254), // **=, reduce: ENV
			reduce(254), // ?, reduce: ENV
			nil,         // :
			reduce(254), // ||, reduce: ENV
			reduce(254), // ^^, reduce: ENV
			reduce(254), // &&, reduce: ENV
			reduce(254), // ??, reduce: ENV
			reduce(254), // ???, reduce: ENV
			reduce(254), // =~, reduce: ENV
			reduce(254), // !=~, reduce: ENV
			reduce(254), // ==, reduce: ENV
			reduce(254), // !=, reduce: ENV
			reduce(254), // <=>, reduce: ENV
			reduce(254), // >=, reduce: ENV
			reduce(254), // <, reduce: ENV
			reduce(254), // <=, reduce: ENV
			reduce(254), // ^, reduce: ENV
			reduce(254), // &, reduce: ENV
			reduce(254), // <<, reduce: ENV
			reduce(254), // >>>, reduce: ENV
			reduce(254), // +, reduce: ENV
			reduce(254), // -, reduce: ENV
			reduce(254), // .+, reduce: ENV
			reduce(254), // .-, reduce: ENV
			reduce(254), // *, reduce: ENV
			reduce(254), // /, reduce: ENV
			reduce(254), // //, reduce: ENV
			reduce(254), // %, reduce: ENV
			reduce(254), // .*, reduce: ENV
			reduce(254), // ./, reduce: ENV
			reduce(254), // .//, reduce: ENV
			reduce(254), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(254), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(254), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(254), // [[, reduce: ENV
			reduce(254), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4110
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(269), // =, reduce: FunctionCallsite
			nil,         // unset
			nil,         // filter
			reduce(269), // >, reduce: FunctionCallsite
			reduce(269), // >>, reduce: FunctionCallsite
			reduce(269), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(269), // ||=, reduce: FunctionCallsite
			reduce(269), // ^^=, reduce: FunctionCallsite
			reduce(269), // &&=, reduce: FunctionCallsite
			reduce(269), // ??=, reduce: FunctionCallsite
			reduce(269), // ???=, reduce: FunctionCallsite
			reduce(269), // |=, reduce: FunctionCallsite
			reduce(269), // &=, reduce: FunctionCallsite
			reduce(269), // ^=, reduce: FunctionCallsite
			reduce(269), // <<=, reduce: FunctionCallsite
			reduce(269), // >>=, reduce: FunctionCallsite
			reduce(269), // >>>=, reduce: FunctionCallsite
			reduce(269), // +=, reduce: FunctionCallsite
			reduce(269), // .=, reduce: FunctionCallsite
			reduce(269), // -=, reduce: FunctionCallsite
			reduce(269), // *=, reduce: FunctionCallsite
			reduce(269), // /=, reduce: FunctionCallsite
			reduce(269), // //=, reduce: FunctionCallsite
			reduce(269), // %=, reduce: FunctionCallsite
			reduce(269), // **=, reduce: FunctionCallsite
			reduce(269), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(269), // ||, reduce: FunctionCallsite
			reduce(269), // ^^, reduce: FunctionCallsite
			reduce(269), // &&, reduce: FunctionCallsite
			reduce(269), // ??, reduce: FunctionCallsite
			reduce(269), // ???, reduce: FunctionCallsite
			reduce(269), // =~, reduce: FunctionCallsite
			reduce(269), // !=~, reduce: FunctionCallsite
			reduce(269), // ==, reduce: FunctionCallsite
			reduce(269), // !=, reduce: FunctionCallsite
			reduce(269), // <=>, reduce: FunctionCallsite
			reduce(269), // >=, reduce: FunctionCallsite
			reduce(269), // <, reduce: FunctionCallsite
			reduce(269), // <=, reduce: FunctionCallsite
			reduce(269), // ^, reduce: FunctionCallsite
			reduce(269), // &, reduce: FunctionCallsite
			reduce(269), // <<, reduce: FunctionCallsite
			reduce(269), // >>>, reduce: FunctionCallsite
			reduce(269), // +, reduce: FunctionCallsite
			reduce(269), // -, reduce: FunctionCallsite
			reduce(269), // .+, reduce: FunctionCallsite
			reduce(269), // .-, reduce: FunctionCallsite
			reduce(269), // *, reduce: FunctionCallsite
			reduce(269), // /, reduce: FunctionCallsite
			reduce(269), // //, reduce: FunctionCallsite
			reduce(269), // %, reduce: FunctionCallsite
			reduce(269), // .*, reduce: FunctionCallsite
			reduce(269), // ./, reduce: FunctionCallsite
			reduce(269), // .//, reduce: FunctionCallsite
			reduce(269), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(269), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(269), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(269), // [[, reduce: FunctionCallsite
			reduce(269), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4111
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(234), // ;, reduce: MapLiteral
			nil,         // {
			nil,         // }
			reduce(234), // =, reduce: MapLiteral
			nil,         // unset
			nil,         // filter
			reduce(234), // >, reduce: MapLiteral
			reduce(234), // >>, reduce: MapLiteral
			reduce(234), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(234), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(234), // ||=, reduce: MapLiteral
			reduce(234), // ^^=, reduce: MapLiteral
			reduce(234), // &&=, reduce: MapLiteral
			reduce(234), // ??=, reduce: MapLiteral
			reduce(234), // ???=, reduce: MapLiteral
			reduce(234), // |=, reduce: MapLiteral
			reduce(234), // &=, reduce: MapLiteral
			reduce(234), // ^=, reduce: MapLiteral
			reduce(234), // <<=, reduce: MapLiteral
			reduce(234), // >>=, reduce: MapLiteral
			reduce(234), // >>>=, reduce: MapLiteral
			reduce(234), // +=, reduce: MapLiteral
			reduce(234), // .=, reduce: MapLiteral
			reduce(234), // -=, reduce: MapLiteral
			reduce(234), // *=, reduce: MapLiteral
			reduce(234), // /=, reduce: MapLiteral
			reduce(234), // //=, reduce: MapLiteral
			reduce(234), // %=, reduce: MapLiteral
			reduce(234), // **=, reduce: MapLiteral
			reduce(234), // ?, reduce: MapLiteral
			nil,         // :
			reduce(234), // ||, reduce: MapLiteral
			reduce(234), // ^^, reduce: MapLiteral
			reduce(234), // &&, reduce: MapLiteral
			reduce(234), // ??, reduce: MapLiteral
			reduce(234), // ???, reduce: MapLiteral
			reduce(234), // =~, reduce: MapLiteral
			reduce(234), // !=~, reduce: MapLiteral
			reduce(234), // ==, reduce: MapLiteral
			reduce(234), // !=, reduce: MapLiteral
			reduce(234), // <=>, reduce: MapLiteral
			reduce(234), // >=, reduce: MapLiteral
			reduce(234), // <, reduce: MapLiteral
			reduce(234), // <=, reduce: MapLiteral
			reduce(234), // ^, reduce: MapLiteral
			reduce(234), // &, reduce: MapLiteral
			reduce(234), // <<, reduce: MapLiteral
			reduce(234), // >>>, reduce: MapLiteral
			reduce(234), // +, reduce: MapLiteral
			reduce(234), // -, reduce: MapLiteral
			reduce(234), // .+, reduce: MapLiteral
			reduce(234), // .-, reduce: MapLiteral
			reduce(234), // *, reduce: MapLiteral
			reduce(234), // /, reduce: MapLiteral
			reduce(234), // //, reduce: MapLiteral
			reduce(234), // %, reduce: MapLiteral
			reduce(234), // .*, reduce: MapLiteral
			reduce(234), // ./, reduce: MapLiteral
			reduce(234), // .//, reduce: MapLiteral
			reduce(234), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(234), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(234), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(234), // [[, reduce: MapLiteral
			reduce(234), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4112
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4374), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4113
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4375), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4114
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4376), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4115
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4377), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4116
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4378), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4117
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4379), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4118
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(172), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(177), // (
			nil,        // )
			nil,        // emitp
			shift(189), // field_name
			shift(190), // $[
			nil,        // ]
			shift(191), // braced_field_name
			shift(192), // $[[
			shift(193), // $[[[
			shift(194), // full_srec
			shift(196), // oosvar_name
			shift(197), // @[
			shift(198), // braced_oosvar_name
			shift(199), // full_oosvar
			shift(200), // all
			shift(201), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(218), // +
			shift(220), // -
			shift(221), // .+
			shift(222), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(225), // !
			shift(226), // ~
			nil,        // **
			shift(231), // string_literal
			shift(232), // regex_case_insensitive
			shift(233), // int_literal
			shift(234), // float_literal
			shift(235), // boolean_literal
			shift(236), // null_literal
			shift(237), // inf_literal
			shift(238), // nan_literal
			shift(239), // const_M_PI
			shift(240), // const_M_E
			shift(241), // panic
			shift(243), // [
			shift(247), // ctx_IPS
			shift(248), // ctx_IFS
			shift(249), // ctx_IRS
			shift(250), // ctx_OPS
			shift(251), // ctx_OFS
			shift(252), // ctx_ORS
			shift(253), // ctx_FLATSEP
			shift(254), // ctx_NF
			shift(255), // ctx_NR
			shift(256), // ctx_FNR
			shift(257), // ctx_FILENAME
			shift(258), // ctx_FILENUM
			shift(260), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(267), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4119
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4120
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4121
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4122
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4123
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4124
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4125
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4126
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4127
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4128
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4129
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4130
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4131
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4132
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4133
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4134
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4135
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4136
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4137
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4138
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4139
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4140
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4141
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4142
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4143
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(197), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			reduce(197), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(197), // >, reduce: UnaryOpTerm
			reduce(197), // >>, reduce: UnaryOpTerm
			reduce(197), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(197), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(197), // ||=, reduce: UnaryOpTerm
			reduce(197), // ^^=, reduce: UnaryOpTerm
			reduce(197), // &&=, reduce: UnaryOpTerm
			reduce(197), // ??=, reduce: UnaryOpTerm
			reduce(197), // ???=, reduce: UnaryOpTerm
			reduce(197), // |=, reduce: UnaryOpTerm
			reduce(197), // &=, reduce: UnaryOpTerm
			reduce(197), // ^=, reduce: UnaryOpTerm
			reduce(197), // <<=, reduce: UnaryOpTerm
			reduce(197), // >>=, reduce: UnaryOpTerm
			reduce(197), // >>>=, reduce: UnaryOpTerm
			reduce(197), // +=, reduce: UnaryOpTerm
			reduce(197), // .=, reduce: UnaryOpTerm
			reduce(197), // -=, reduce: UnaryOpTerm
			reduce(197), // *=, reduce: UnaryOpTerm
			reduce(197), // /=, reduce: UnaryOpTerm
			reduce(197), // //=, reduce: UnaryOpTerm
			reduce(197), // %=, reduce: UnaryOpTerm
			reduce(197), // **=, reduce: UnaryOpTerm
			reduce(197), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(197), // ||, reduce: UnaryOpTerm
			reduce(197), // ^^, reduce: UnaryOpTerm
			reduce(197), // &&, reduce: UnaryOpTerm
			reduce(197), // ??, reduce: UnaryOpTerm
			reduce(197), // ???, reduce: UnaryOpTerm
			reduce(197), // =~, reduce: UnaryOpTerm
			reduce(197), // !=~, reduce: UnaryOpTerm
			reduce(197), // ==, reduce: UnaryOpTerm
			reduce(197), // !=, reduce: UnaryOpTerm
			reduce(197), // <=>, reduce: UnaryOpTerm
			reduce(197), // >=, reduce: UnaryOpTerm
			reduce(197), // <, reduce: UnaryOpTerm
			reduce(197), // <=, reduce: UnaryOpTerm
			reduce(197), // ^, reduce: UnaryOpTerm
			reduce(197), // &, reduce: UnaryOpTerm
			reduce(197), // <<, reduce: UnaryOpTerm
			reduce(197), // >>>, reduce: UnaryOpTerm
			reduce(197), // +, reduce: UnaryOpTerm
			reduce(197), // -, reduce: UnaryOpTerm
			reduce(197), // .+, reduce: UnaryOpTerm
			reduce(197), // .-, reduce: UnaryOpTerm
			reduce(197), // *, reduce: UnaryOpTerm
			reduce(197), // /, reduce: UnaryOpTerm
			reduce(197), // //, reduce: UnaryOpTerm
			reduce(197), // %, reduce: UnaryOpTerm
			reduce(197), // .*, reduce: UnaryOpTerm
			reduce(197), // ./, reduce: UnaryOpTerm
			reduce(197), // .//, reduce: UnaryOpTerm
			reduce(197), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4144
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4145
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4146
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4147
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4148
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4149
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4150
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4151
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(198), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			reduce(198), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(198), // >, reduce: UnaryOpTerm
			reduce(198), // >>, reduce: UnaryOpTerm
			reduce(198), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(198), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(198), // ||=, reduce: UnaryOpTerm
			reduce(198), // ^^=, reduce: UnaryOpTerm
			reduce(198), // &&=, reduce: UnaryOpTerm
			reduce(198), // ??=, reduce: UnaryOpTerm
			reduce(198), // ???=, reduce: UnaryOpTerm
			reduce(198), // |=, reduce: UnaryOpTerm
			reduce(198), // &=, reduce: UnaryOpTerm
			reduce(198), // ^=, reduce: UnaryOpTerm
			reduce(198), // <<=, reduce: UnaryOpTerm
			reduce(198), // >>=, reduce: UnaryOpTerm
			reduce(198), // >>>=, reduce: UnaryOpTerm
			reduce(198), // +=, reduce: UnaryOpTerm
			reduce(198), // .=, reduce: UnaryOpTerm
			reduce(198), // -=, reduce: UnaryOpTerm
			reduce(198), // *=, reduce: UnaryOpTerm
			reduce(198), // /=, reduce: UnaryOpTerm
			reduce(198), // //=, reduce: UnaryOpTerm
			reduce(198), // %=, reduce: UnaryOpTerm
			reduce(198), // **=, reduce: UnaryOpTerm
			reduce(198), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(198), // ||, reduce: UnaryOpTerm
			reduce(198), // ^^, reduce: UnaryOpTerm
			reduce(198), // &&, reduce: UnaryOpTerm
			reduce(198), // ??, reduce: UnaryOpTerm
			reduce(198), // ???, reduce: UnaryOpTerm
			reduce(198), // =~, reduce: UnaryOpTerm
			reduce(198), // !=~, reduce: UnaryOpTerm
			reduce(198), // ==, reduce: UnaryOpTerm
			reduce(198), // !=, reduce: UnaryOpTerm
			reduce(198), // <=>, reduce: UnaryOpTerm
			reduce(198), // >=, reduce: UnaryOpTerm
			reduce(198), // <, reduce: UnaryOpTerm
			reduce(198), // <=, reduce: UnaryOpTerm
			reduce(198), // ^, reduce: UnaryOpTerm
			reduce(198), // &, reduce: UnaryOpTerm
			reduce(198), // <<, reduce: UnaryOpTerm
			reduce(198), // >>>, reduce: UnaryOpTerm
			reduce(198), // +, reduce: UnaryOpTerm
			reduce(198), // -, reduce: UnaryOpTerm
			reduce(198), // .+, reduce: UnaryOpTerm
			reduce(198), // .-, reduce: UnaryOpTerm
			reduce(198), // *, reduce: UnaryOpTerm
			reduce(198), // /, reduce: UnaryOpTerm
			reduce(198), // //, reduce: UnaryOpTerm
			reduce(198), // %, reduce: UnaryOpTerm
			reduce(198), // .*, reduce: UnaryOpTerm
			reduce(198), // ./, reduce: UnaryOpTerm
			reduce(198), // .//, reduce: UnaryOpTerm
			reduce(198), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4152
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(199), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			reduce(199), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(199), // >, reduce: UnaryOpTerm
			reduce(199), // >>, reduce: UnaryOpTerm
			reduce(199), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(199), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(199), // ||=, reduce: UnaryOpTerm
			reduce(199), // ^^=, reduce: UnaryOpTerm
			reduce(199), // &&=, reduce: UnaryOpTerm
			reduce(199), // ??=, reduce: UnaryOpTerm
			reduce(199), // ???=, reduce: UnaryOpTerm
			reduce(199), // |=, reduce: UnaryOpTerm
			reduce(199), // &=, reduce: UnaryOpTerm
			reduce(199), // ^=, reduce: UnaryOpTerm
			reduce(199), // <<=, reduce: UnaryOpTerm
			reduce(199), // >>=, reduce: UnaryOpTerm
			reduce(199), // >>>=, reduce: UnaryOpTerm
			reduce(199), // +=, reduce: UnaryOpTerm
			reduce(199), // .=, reduce: UnaryOpTerm
			reduce(199), // -=, reduce: UnaryOpTerm
			reduce(199), // *=, reduce: UnaryOpTerm
			reduce(199), // /=, reduce: UnaryOpTerm
			reduce(199), // //=, reduce: UnaryOpTerm
			reduce(199), // %=, reduce: UnaryOpTerm
			reduce(199), // **=, reduce: UnaryOpTerm
			reduce(199), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(199), // ||, reduce: UnaryOpTerm
			reduce(199), // ^^, reduce: UnaryOpTerm
			reduce(199), // &&, reduce: UnaryOpTerm
			reduce(199), // ??, reduce: UnaryOpTerm
			reduce(199), // ???, reduce: UnaryOpTerm
			reduce(199), // =~, reduce: UnaryOpTerm
			reduce(199), // !=~, reduce: UnaryOpTerm
			reduce(199), // ==, reduce: UnaryOpTerm
			reduce(199), // !=, reduce: UnaryOpTerm
			reduce(199), // <=>, reduce: UnaryOpTerm
			reduce(199), // >=, reduce: UnaryOpTerm
			reduce(199), // <, reduce: UnaryOpTerm
			reduce(199), // <=, reduce: UnaryOpTerm
			reduce(199), // ^, reduce: UnaryOpTerm
			reduce(199), // &, reduce: UnaryOpTerm
			reduce(199), // <<, reduce: UnaryOpTerm
			reduce(199), // >>>, reduce: UnaryOpTerm
			reduce(199), // +, reduce: UnaryOpTerm
			reduce(199), // -, reduce: UnaryOpTerm
			reduce(199), // .+, reduce: UnaryOpTerm
			reduce(199), // .-, reduce: UnaryOpTerm
			reduce(199), // *, reduce: UnaryOpTerm
			reduce(199), // /, reduce: UnaryOpTerm
			reduce(199), // //, reduce: UnaryOpTerm
			reduce(199), // %, reduce: UnaryOpTerm
			reduce(199), // .*, reduce: UnaryOpTerm
			reduce(199), // ./, reduce: UnaryOpTerm
			reduce(199), // .//, reduce: UnaryOpTerm
			reduce(199), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4153
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(200), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			reduce(200), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(200), // >, reduce: UnaryOpTerm
			reduce(200), // >>, reduce: UnaryOpTerm
			reduce(200), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(200), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(200), // ||=, reduce: UnaryOpTerm
			reduce(200), // ^^=, reduce: UnaryOpTerm
			reduce(200), // &&=, reduce: UnaryOpTerm
			reduce(200), // ??=, reduce: UnaryOpTerm
			reduce(200), // ???=, reduce: UnaryOpTerm
			reduce(200), // |=, reduce: UnaryOpTerm
			reduce(200), // &=, reduce: UnaryOpTerm
			reduce(200), // ^=, reduce: UnaryOpTerm
			reduce(200), // <<=, reduce: UnaryOpTerm
			reduce(200), // >>=, reduce: UnaryOpTerm
			reduce(200), // >>>=, reduce: UnaryOpTerm
			reduce(200), // +=, reduce: UnaryOpTerm
			reduce(200), // .=, reduce: UnaryOpTerm
			reduce(200), // -=, reduce: UnaryOpTerm
			reduce(200), // *=, reduce: UnaryOpTerm
			reduce(200), // /=, reduce: UnaryOpTerm
			reduce(200), // //=, reduce: UnaryOpTerm
			reduce(200), // %=, reduce: UnaryOpTerm
			reduce(200), // **=, reduce: UnaryOpTerm
			reduce(200), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(200), // ||, reduce: UnaryOpTerm
			reduce(200), // ^^, reduce: UnaryOpTerm
			reduce(200), // &&, reduce: UnaryOpTerm
			reduce(200), // ??, reduce: UnaryOpTerm
			reduce(200), // ???, reduce: UnaryOpTerm
			reduce(200), // =~, reduce: UnaryOpTerm
			reduce(200), // !=~, reduce: UnaryOpTerm
			reduce(200), // ==, reduce: UnaryOpTerm
			reduce(200), // !=, reduce: UnaryOpTerm
			reduce(200), // <=>, reduce: UnaryOpTerm
			reduce(200), // >=, reduce: UnaryOpTerm
			reduce(200), // <, reduce: UnaryOpTerm
			reduce(200), // <=, reduce: UnaryOpTerm
			reduce(200), // ^, reduce: UnaryOpTerm
			reduce(200), // &, reduce: UnaryOpTerm
			reduce(200), // <<, reduce: UnaryOpTerm
			reduce(200), // >>>, reduce: UnaryOpTerm
			reduce(200), // +, reduce: UnaryOpTerm
			reduce(200), // -, reduce: UnaryOpTerm
			reduce(200), // .+, reduce: UnaryOpTerm
			reduce(200), // .-, reduce: UnaryOpTerm
			reduce(200), // *, reduce: UnaryOpTerm
			reduce(200), // /, reduce: UnaryOpTerm
			reduce(200), // //, reduce: UnaryOpTerm
			reduce(200), // %, reduce: UnaryOpTerm
			reduce(200), // .*, reduce: UnaryOpTerm
			reduce(200), // ./, reduce: UnaryOpTerm
			reduce(200), // .//, reduce: UnaryOpTerm
			reduce(200), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4154
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4155
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(201), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			reduce(201), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(201), // >, reduce: UnaryOpTerm
			reduce(201), // >>, reduce: UnaryOpTerm
			reduce(201), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(201), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(201), // ||=, reduce: UnaryOpTerm
			reduce(201), // ^^=, reduce: UnaryOpTerm
			reduce(201), // &&=, reduce: UnaryOpTerm
			reduce(201), // ??=, reduce: UnaryOpTerm
			reduce(201), // ???=, reduce: UnaryOpTerm
			reduce(201), // |=, reduce: UnaryOpTerm
			reduce(201), // &=, reduce: UnaryOpTerm
			reduce(201), // ^=, reduce: UnaryOpTerm
			reduce(201), // <<=, reduce: UnaryOpTerm
			reduce(201), // >>=, reduce: UnaryOpTerm
			reduce(201), // >>>=, reduce: UnaryOpTerm
			reduce(201), // +=, reduce: UnaryOpTerm
			reduce(201), // .=, reduce: UnaryOpTerm
			reduce(201), // -=, reduce: UnaryOpTerm
			reduce(201), // *=, reduce: UnaryOpTerm
			reduce(201), // /=, reduce: UnaryOpTerm
			reduce(201), // //=, reduce: UnaryOpTerm
			reduce(201), // %=, reduce: UnaryOpTerm
			reduce(201), // **=, reduce: UnaryOpTerm
			reduce(201), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(201), // ||, reduce: UnaryOpTerm
			reduce(201), // ^^, reduce: UnaryOpTerm
			reduce(201), // &&, reduce: UnaryOpTerm
			reduce(201), // ??, reduce: UnaryOpTerm
			reduce(201), // ???, reduce: UnaryOpTerm
			reduce(201), // =~, reduce: UnaryOpTerm
			reduce(201), // !=~, reduce: UnaryOpTerm
			reduce(201), // ==, reduce: UnaryOpTerm
			reduce(201), // !=, reduce: UnaryOpTerm
			reduce(201), // <=>, reduce: UnaryOpTerm
			reduce(201), // >=, reduce: UnaryOpTerm
			reduce(201), // <, reduce: UnaryOpTerm
			reduce(201), // <=, reduce: UnaryOpTerm
			reduce(201), // ^, reduce: UnaryOpTerm
			reduce(201), // &, reduce: UnaryOpTerm
			reduce(201), // <<, reduce: UnaryOpTerm
			reduce(201), // >>>, reduce: UnaryOpTerm
			reduce(201), // +, reduce: UnaryOpTerm
			reduce(201), // -, reduce: UnaryOpTerm
			reduce(201), // .+, reduce: UnaryOpTerm
			reduce(201), // .-, reduce: UnaryOpTerm
			reduce(201), // *, reduce: UnaryOpTerm
			reduce(201), // /, reduce: UnaryOpTerm
			reduce(201), // //, reduce: UnaryOpTerm
			reduce(201), // %, reduce: UnaryOpTerm
			reduce(201), // .*, reduce: UnaryOpTerm
			reduce(201), // ./, reduce: UnaryOpTerm
			reduce(201), // .//, reduce: UnaryOpTerm
			reduce(201), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4156
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(202), // ;, reduce: UnaryOpTerm
			nil,         // {
			nil,         // }
			reduce(202), // =, reduce: UnaryOpTerm
			nil,         // unset
			nil,         // filter
			reduce(202), // >, reduce: UnaryOpTerm
			reduce(202), // >>, reduce: UnaryOpTerm
			reduce(202), // |, reduce: UnaryOpTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(202), // ,, reduce: UnaryOpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(202), // ||=, reduce: UnaryOpTerm
			reduce(202), // ^^=, reduce: UnaryOpTerm
			reduce(202), // &&=, reduce: UnaryOpTerm
			reduce(202), // ??=, reduce: UnaryOpTerm
			reduce(202), // ???=, reduce: UnaryOpTerm
			reduce(202), // |=, reduce: UnaryOpTerm
			reduce(202), // &=, reduce: UnaryOpTerm
			reduce(202), // ^=, reduce: UnaryOpTerm
			reduce(202), // <<=, reduce: UnaryOpTerm
			reduce(202), // >>=, reduce: UnaryOpTerm
			reduce(202), // >>>=, reduce: UnaryOpTerm
			reduce(202), // +=, reduce: UnaryOpTerm
			reduce(202), // .=, reduce: UnaryOpTerm
			reduce(202), // -=, reduce: UnaryOpTerm
			reduce(202), // *=, reduce: UnaryOpTerm
			reduce(202), // /=, reduce: UnaryOpTerm
			reduce(202), // //=, reduce: UnaryOpTerm
			reduce(202), // %=, reduce: UnaryOpTerm
			reduce(202), // **=, reduce: UnaryOpTerm
			reduce(202), // ?, reduce: UnaryOpTerm
			nil,         // :
			reduce(202), // ||, reduce: UnaryOpTerm
			reduce(202), // ^^, reduce: UnaryOpTerm
			reduce(202), // &&, reduce: UnaryOpTerm
			reduce(202), // ??, reduce: UnaryOpTerm
			reduce(202), // ???, reduce: UnaryOpTerm
			reduce(202), // =~, reduce: UnaryOpTerm
			reduce(202), // !=~, reduce: UnaryOpTerm
			reduce(202), // ==, reduce: UnaryOpTerm
			reduce(202), // !=, reduce: UnaryOpTerm
			reduce(202), // <=>, reduce: UnaryOpTerm
			reduce(202), // >=, reduce: UnaryOpTerm
			reduce(202), // <, reduce: UnaryOpTerm
			reduce(202), // <=, reduce: UnaryOpTerm
			reduce(202), // ^, reduce: UnaryOpTerm
			reduce(202), // &, reduce: UnaryOpTerm
			reduce(202), // <<, reduce: UnaryOpTerm
			reduce(202), // >>>, reduce: UnaryOpTerm
			reduce(202), // +, reduce: UnaryOpTerm
			reduce(202), // -, reduce: UnaryOpTerm
			reduce(202), // .+, reduce: UnaryOpTerm
			reduce(202), // .-, reduce: UnaryOpTerm
			reduce(202), // *, reduce: UnaryOpTerm
			reduce(202), // /, reduce: UnaryOpTerm
			reduce(202), // //, reduce: UnaryOpTerm
			reduce(202), // %, reduce: UnaryOpTerm
			reduce(202), // .*, reduce: UnaryOpTerm
			reduce(202), // ./, reduce: UnaryOpTerm
			reduce(202), // .//, reduce: UnaryOpTerm
			reduce(202), // ., reduce: UnaryOpTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4157
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(4413), // +
			shift(4414), // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4158
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1509), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1513), // (
			nil,         // )
			nil,         // emitp
			shift(1525), // field_name
			shift(1526), // $[
			nil,         // ]
			shift(1527), // braced_field_name
			shift(1528), // $[[
			shift(1529), // $[[[
			shift(1530), // full_srec
			shift(1532), // oosvar_name
			shift(1533), // @[
			shift(1534), // braced_oosvar_name
			shift(1535), // full_oosvar
			shift(1536), // all
			shift(1537), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4417), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1553), // +
			shift(1555), // -
			shift(1556), // .+
			shift(1557), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1560), // !
			shift(1561), // ~
			nil,         // **
			shift(1566), // string_literal
			shift(1567), // regex_case_insensitive
			shift(1568), // int_literal
			shift(1569), // float_literal
			shift(1570), // boolean_literal
			shift(1571), // null_literal
			shift(1572), // inf_literal
			shift(1573), // nan_literal
			shift(1574), // const_M_PI
			shift(1575), // const_M_E
			shift(1576), // panic
			shift(1578), // [
			shift(1580), // ctx_IPS
			shift(1581), // ctx_IFS
			shift(1582), // ctx_IRS
			shift(1583), // ctx_OPS
			shift(1584), // ctx_OFS
			shift(1585), // ctx_ORS
			shift(1586), // ctx_FLATSEP
			shift(1587), // ctx_NF
			shift(1588), // ctx_NR
			shift(1589), // ctx_FNR
			shift(1590), // ctx_FILENAME
			shift(1591), // ctx_FILENUM
			shift(1593), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1600), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4159
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4160
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4161
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(228), // ;, reduce: ArrayLiteral
			nil,         // {
			nil,         // }
			reduce(228), // =, reduce: ArrayLiteral
			nil,         // unset
			nil,         // filter
			reduce(228), // >, reduce: ArrayLiteral
			reduce(228), // >>, reduce: ArrayLiteral
			reduce(228), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(228), // ,, reduce: ArrayLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(228), // ||=, reduce: ArrayLiteral
			reduce(228), // ^^=, reduce: ArrayLiteral
			reduce(228), // &&=, reduce: ArrayLiteral
			reduce(228), // ??=, reduce: ArrayLiteral
			reduce(228), // ???=, reduce: ArrayLiteral
			reduce(228), // |=, reduce: ArrayLiteral
			reduce(228), // &=, reduce: ArrayLiteral
			reduce(228), // ^=, reduce: ArrayLiteral
			reduce(228), // <<=, reduce: ArrayLiteral
			reduce(228), // >>=, reduce: ArrayLiteral
			reduce(228), // >>>=, reduce: ArrayLiteral
			reduce(228), // +=, reduce: ArrayLiteral
			reduce(228), // .=, reduce: ArrayLiteral
			reduce(228), // -=, reduce: ArrayLiteral
			reduce(228), // *=, reduce: ArrayLiteral
			reduce(228), // /=, reduce: ArrayLiteral
			reduce(228), // //=, reduce: ArrayLiteral
			reduce(228), // %=, reduce: ArrayLiteral
			reduce(228), // **=, reduce: ArrayLiteral
			reduce(228), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(228), // ||, reduce: ArrayLiteral
			reduce(228), // ^^, reduce: ArrayLiteral
			reduce(228), // &&, reduce: ArrayLiteral
			reduce(228), // ??, reduce: ArrayLiteral
			reduce(228), // ???, reduce: ArrayLiteral
			reduce(228), // =~, reduce: ArrayLiteral
			reduce(228), // !=~, reduce: ArrayLiteral
			reduce(228), // ==, reduce: ArrayLiteral
			reduce(228), // !=, reduce: ArrayLiteral
			reduce(228), // <=>, reduce: ArrayLiteral
			reduce(228), // >=, reduce: ArrayLiteral
			reduce(228), // <, reduce: ArrayLiteral
			reduce(228), // <=, reduce: ArrayLiteral
			reduce(228), // ^, reduce: ArrayLiteral
			reduce(228), // &, reduce: ArrayLiteral
			reduce(228), // <<, reduce: ArrayLiteral
			reduce(228), // >>>, reduce: ArrayLiteral
			reduce(228), // +, reduce: ArrayLiteral
			reduce(228), // -, reduce: ArrayLiteral
			reduce(228), // .+, reduce: ArrayLiteral
			reduce(228), // .-, reduce: ArrayLiteral
			reduce(228), // *, reduce: ArrayLiteral
			reduce(228), // /, reduce: ArrayLiteral
			reduce(228), // //, reduce: ArrayLiteral
			reduce(228), // %, reduce: ArrayLiteral
			reduce(228), // .*, reduce: ArrayLiteral
			reduce(228), // ./, reduce: ArrayLiteral
			reduce(228), // .//, reduce: ArrayLiteral
			reduce(228), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(228), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(228), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(228), // [[, reduce: ArrayLiteral
			reduce(228), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4162
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4420), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4163
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(4421), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4164
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(614), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(618), // (
			nil,        // )
			nil,        // emitp
			shift(630), // field_name
			shift(631), // $[
			nil,        // ]
			shift(632), // braced_field_name
			shift(633), // $[[
			shift(634), // $[[[
			shift(635), // full_srec
			shift(637), // oosvar_name
			shift(638), // @[
			shift(639), // braced_oosvar_name
			shift(640), // full_oosvar
			shift(641), // all
			shift(642), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(657), // +
			shift(659), // -
			shift(660), // .+
			shift(661), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(664), // !
			shift(665), // ~
			nil,        // **
			shift(670), // string_literal
			shift(671), // regex_case_insensitive
			shift(672), // int_literal
			shift(673), // float_literal
			shift(674), // boolean_literal
			shift(675), // null_literal
			shift(676), // inf_literal
			shift(677), // nan_literal
			shift(678), // const_M_PI
			shift(679), // const_M_E
			shift(680), // panic
			shift(682), // [
			shift(684), // ctx_IPS
			shift(685), // ctx_IFS
			shift(686), // ctx_IRS
			shift(687), // ctx_OPS
			shift(688), // ctx_OFS
			shift(689), // ctx_ORS
			shift(690), // ctx_FLATSEP
			shift(691), // ctx_NF
			shift(692), // ctx_NR
			shift(693), // ctx_FNR
			shift(694), // ctx_FILENAME
			shift(695), // ctx_FILENUM
			shift(697), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(704), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4165
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			shift(4424), // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4166
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(317), // ,, reduce: TripleForUpdate
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			reduce(317), // ), reduce: TripleForUpdate
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4167
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4168
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(327), // ), reduce: FuncOrSubrParameterList
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(1858), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(1859), // float
			shift(1860), // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4169
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4431), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4432), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4170
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4433), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4171
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3714), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4172
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			reduce(119), // {, reduce: Typedecl
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4173
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			reduce(120), // {, reduce: Typedecl
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4174
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			reduce(121), // {, reduce: Typedecl
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4175
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			reduce(122), // {, reduce: Typedecl
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4176
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			reduce(123), // {, reduce: Typedecl
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4177
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			reduce(124), // {, reduce: Typedecl
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4178
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			reduce(125), // {, reduce: Typedecl
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4179
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			reduce(126), // {, reduce: Typedecl
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4180
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			reduce(127), // {, reduce: Typedecl
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4181
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(322), // $, reduce: NamedFunctionDefinition
			nil,         // empty
			reduce(322), // ;, reduce: NamedFunctionDefinition
			reduce(322), // {, reduce: NamedFunctionDefinition
			nil,         // }
			nil,         // =
			reduce(322), // unset, reduce: NamedFunctionDefinition
			reduce(322), // filter, reduce: NamedFunctionDefinition
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(322), // print, reduce: NamedFunctionDefinition
			nil,         // ,
			reduce(322), // printn, reduce: NamedFunctionDefinition
			reduce(322), // eprint, reduce: NamedFunctionDefinition
			reduce(322), // eprintn, reduce: NamedFunctionDefinition
			reduce(322), // dump, reduce: NamedFunctionDefinition
			reduce(322), // edump, reduce: NamedFunctionDefinition
			reduce(322), // tee, reduce: NamedFunctionDefinition
			reduce(322), // emitf, reduce: NamedFunctionDefinition
			reduce(322), // emit1, reduce: NamedFunctionDefinition
			reduce(322), // emit, reduce: NamedFunctionDefinition
			reduce(322), // (, reduce: NamedFunctionDefinition
			nil,         // )
			reduce(322), // emitp, reduce: NamedFunctionDefinition
			reduce(322), // field_name, reduce: NamedFunctionDefinition
			reduce(322), // $[, reduce: NamedFunctionDefinition
			nil,         // ]
			reduce(322), // braced_field_name, reduce: NamedFunctionDefinition
			reduce(322), // $[[, reduce: NamedFunctionDefinition
			reduce(322), // $[[[, reduce: NamedFunctionDefinition
			reduce(322), // full_srec, reduce: NamedFunctionDefinition
			reduce(322), // oosvar_name, reduce: NamedFunctionDefinition
			reduce(322), // @[, reduce: NamedFunctionDefinition
			reduce(322), // braced_oosvar_name, reduce: NamedFunctionDefinition
			reduce(322), // full_oosvar, reduce: NamedFunctionDefinition
			reduce(322), // all, reduce: NamedFunctionDefinition
			reduce(322), // non_sigil_name, reduce: NamedFunctionDefinition
			reduce(322), // arr, reduce: NamedFunctionDefinition
			reduce(322), // bool, reduce: NamedFunctionDefinition
			reduce(322), // float, reduce: NamedFunctionDefinition
			reduce(322), // int, reduce: NamedFunctionDefinition
			reduce(322), // map, reduce: NamedFunctionDefinition
			reduce(322), // num, reduce: NamedFunctionDefinition
			reduce(322), // str, reduce: NamedFunctionDefinition
			reduce(322), // var, reduce: NamedFunctionDefinition
			reduce(322), // funct, reduce: NamedFunctionDefinition
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(322), // +, reduce: NamedFunctionDefinition
			reduce(322), // -, reduce: NamedFunctionDefinition
			reduce(322), // .+, reduce: NamedFunctionDefinition
			reduce(322), // .-, reduce: NamedFunctionDefinition
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(322), // !, reduce: NamedFunctionDefinition
			reduce(322), // ~, reduce: NamedFunctionDefinition
			nil,         // **
			reduce(322), // string_literal, reduce: NamedFunctionDefinition
			reduce(322), // regex_case_insensitive, reduce: NamedFunctionDefinition
			reduce(322), // int_literal, reduce: NamedFunctionDefinition
			reduce(322), // float_literal, reduce: NamedFunctionDefinition
			reduce(322), // boolean_literal, reduce: NamedFunctionDefinition
			reduce(322), // null_literal, reduce: NamedFunctionDefinition
			reduce(322), // inf_literal, reduce: NamedFunctionDefinition
			reduce(322), // nan_literal, reduce: NamedFunctionDefinition
			reduce(322), // const_M_PI, reduce: NamedFunctionDefinition
			reduce(322), // const_M_E, reduce: NamedFunctionDefinition
			reduce(322), // panic, reduce: NamedFunctionDefinition
			reduce(322), // [, reduce: NamedFunctionDefinition
			reduce(322), // ctx_IPS, reduce: NamedFunctionDefinition
			reduce(322), // ctx_IFS, reduce: NamedFunctionDefinition
			reduce(322), // ctx_IRS, reduce: NamedFunctionDefinition
			reduce(322), // ctx_OPS, reduce: NamedFunctionDefinition
			reduce(322), // ctx_OFS, reduce: NamedFunctionDefinition
			reduce(322), // ctx_ORS, reduce: NamedFunctionDefinition
			reduce(322), // ctx_FLATSEP, reduce: NamedFunctionDefinition
			reduce(322), // ctx_NF, reduce: NamedFunctionDefinition
			reduce(322), // ctx_NR, reduce: NamedFunctionDefinition
			reduce(322), // ctx_FNR, reduce: NamedFunctionDefinition
			reduce(322), // ctx_FILENAME, reduce: NamedFunctionDefinition
			reduce(322), // ctx_FILENUM, reduce: NamedFunctionDefinition
			reduce(322), // env, reduce: NamedFunctionDefinition
			nil,         // [[
			nil,         // [[[
			reduce(322), // call, reduce: NamedFunctionDefinition
			reduce(322), // begin, reduce: NamedFunctionDefinition
			reduce(322), // end, reduce: NamedFunctionDefinition
			reduce(322), // if, reduce: NamedFunctionDefinition
			nil,         // elif
			nil,         // else
			reduce(322), // while, reduce: NamedFunctionDefinition
			reduce(322), // do, reduce: NamedFunctionDefinition
			reduce(322), // for, reduce: NamedFunctionDefinition
			nil,         // in
			reduce(322), // break, reduce: NamedFunctionDefinition
			reduce(322), // continue, reduce: NamedFunctionDefinition
			reduce(322), // func, reduce: NamedFunctionDefinition
			reduce(322), // subr, reduce: NamedFunctionDefinition
			reduce(322), // return, reduce: NamedFunctionDefinition
		},
	},
	actionRow{ // S4182
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4183
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(326), // $, reduce: SubroutineDefinition
			nil,         // empty
			reduce(326), // ;, reduce: SubroutineDefinition
			reduce(326), // {, reduce: SubroutineDefinition
			nil,         // }
			nil,         // =
			reduce(326), // unset, reduce: SubroutineDefinition
			reduce(326), // filter, reduce: SubroutineDefinition
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(326), // print, reduce: SubroutineDefinition
			nil,         // ,
			reduce(326), // printn, reduce: SubroutineDefinition
			reduce(326), // eprint, reduce: SubroutineDefinition
			reduce(326), // eprintn, reduce: SubroutineDefinition
			reduce(326), // dump, reduce: SubroutineDefinition
			reduce(326), // edump, reduce: SubroutineDefinition
			reduce(326), // tee, reduce: SubroutineDefinition
			reduce(326), // emitf, reduce: SubroutineDefinition
			reduce(326), // emit1, reduce: SubroutineDefinition
			reduce(326), // emit, reduce: SubroutineDefinition
			reduce(326), // (, reduce: SubroutineDefinition
			nil,         // )
			reduce(326), // emitp, reduce: SubroutineDefinition
			reduce(326), // field_name, reduce: SubroutineDefinition
			reduce(326), // $[, reduce: SubroutineDefinition
			nil,         // ]
			reduce(326), // braced_field_name, reduce: SubroutineDefinition
			reduce(326), // $[[, reduce: SubroutineDefinition
			reduce(326), // $[[[, reduce: SubroutineDefinition
			reduce(326), // full_srec, reduce: SubroutineDefinition
			reduce(326), // oosvar_name, reduce: SubroutineDefinition
			reduce(326), // @[, reduce: SubroutineDefinition
			reduce(326), // braced_oosvar_name, reduce: SubroutineDefinition
			reduce(326), // full_oosvar, reduce: SubroutineDefinition
			reduce(326), // all, reduce: SubroutineDefinition
			reduce(326), // non_sigil_name, reduce: SubroutineDefinition
			reduce(326), // arr, reduce: SubroutineDefinition
			reduce(326), // bool, reduce: SubroutineDefinition
			reduce(326), // float, reduce: SubroutineDefinition
			reduce(326), // int, reduce: SubroutineDefinition
			reduce(326), // map, reduce: SubroutineDefinition
			reduce(326), // num, reduce: SubroutineDefinition
			reduce(326), // str, reduce: SubroutineDefinition
			reduce(326), // var, reduce: SubroutineDefinition
			reduce(326), // funct, reduce: SubroutineDefinition
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(326), // +, reduce: SubroutineDefinition
			reduce(326), // -, reduce: SubroutineDefinition
			reduce(326), // .+, reduce: SubroutineDefinition
			reduce(326), // .-, reduce: SubroutineDefinition
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(326), // !, reduce: SubroutineDefinition
			reduce(326), // ~, reduce: SubroutineDefinition
			nil,         // **
			reduce(326), // string_literal, reduce: SubroutineDefinition
			reduce(326), // regex_case_insensitive, reduce: SubroutineDefinition
			reduce(326), // int_literal, reduce: SubroutineDefinition
			reduce(326), // float_literal, reduce: SubroutineDefinition
			reduce(326), // boolean_literal, reduce: SubroutineDefinition
			reduce(326), // null_literal, reduce: SubroutineDefinition
			reduce(326), // inf_literal, reduce: SubroutineDefinition
			reduce(326), // nan_literal, reduce: SubroutineDefinition
			reduce(326), // const_M_PI, reduce: SubroutineDefinition
			reduce(326), // const_M_E, reduce: SubroutineDefinition
			reduce(326), // panic, reduce: SubroutineDefinition
			reduce(326), // [, reduce: SubroutineDefinition
			reduce(326), // ctx_IPS, reduce: SubroutineDefinition
			reduce(326), // ctx_IFS, reduce: SubroutineDefinition
			reduce(326), // ctx_IRS, reduce: SubroutineDefinition
			reduce(326), // ctx_OPS, reduce: SubroutineDefinition
			reduce(326), // ctx_OFS, reduce: SubroutineDefinition
			reduce(326), // ctx_ORS, reduce: SubroutineDefinition
			reduce(326), // ctx_FLATSEP, reduce: SubroutineDefinition
			reduce(326), // ctx_NF, reduce: SubroutineDefinition
			reduce(326), // ctx_NR, reduce: SubroutineDefinition
			reduce(326), // ctx_FNR, reduce: SubroutineDefinition
			reduce(326), // ctx_FILENAME, reduce: SubroutineDefinition
			reduce(326), // ctx_FILENUM, reduce: SubroutineDefinition
			reduce(326), // env, reduce: SubroutineDefinition
			nil,         // [[
			nil,         // [[[
			reduce(326), // call, reduce: SubroutineDefinition
			reduce(326), // begin, reduce: SubroutineDefinition
			reduce(326), // end, reduce: SubroutineDefinition
			reduce(326), // if, reduce: SubroutineDefinition
			nil,         // elif
			nil,         // else
			reduce(326), // while, reduce: SubroutineDefinition
			reduce(326), // do, reduce: SubroutineDefinition
			reduce(326), // for, reduce: SubroutineDefinition
			nil,         // in
			reduce(326), // break, reduce: SubroutineDefinition
			reduce(326), // continue, reduce: SubroutineDefinition
			reduce(326), // func, reduce: SubroutineDefinition
			reduce(326), // subr, reduce: SubroutineDefinition
			reduce(326), // return, reduce: SubroutineDefinition
		},
	},
	actionRow{ // S4184
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(107), // }, reduce: PositionalFieldName
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(107), // >, reduce: PositionalFieldName
			reduce(107), // >>, reduce: PositionalFieldName
			reduce(107), // |, reduce: PositionalFieldName
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(107), // ,, reduce: PositionalFieldName
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(107), // ?, reduce: PositionalFieldName
			nil,         // :
			reduce(107), // ||, reduce: PositionalFieldName
			reduce(107), // ^^, reduce: PositionalFieldName
			reduce(107), // &&, reduce: PositionalFieldName
			reduce(107), // ??, reduce: PositionalFieldName
			reduce(107), // ???, reduce: PositionalFieldName
			reduce(107), // =~, reduce: PositionalFieldName
			reduce(107), // !=~, reduce: PositionalFieldName
			reduce(107), // ==, reduce: PositionalFieldName
			reduce(107), // !=, reduce: PositionalFieldName
			reduce(107), // <=>, reduce: PositionalFieldName
			reduce(107), // >=, reduce: PositionalFieldName
			reduce(107), // <, reduce: PositionalFieldName
			reduce(107), // <=, reduce: PositionalFieldName
			reduce(107), // ^, reduce: PositionalFieldName
			reduce(107), // &, reduce: PositionalFieldName
			reduce(107), // <<, reduce: PositionalFieldName
			reduce(107), // >>>, reduce: PositionalFieldName
			reduce(107), // +, reduce: PositionalFieldName
			reduce(107), // -, reduce: PositionalFieldName
			reduce(107), // .+, reduce: PositionalFieldName
			reduce(107), // .-, reduce: PositionalFieldName
			reduce(107), // *, reduce: PositionalFieldName
			reduce(107), // /, reduce: PositionalFieldName
			reduce(107), // //, reduce: PositionalFieldName
			reduce(107), // %, reduce: PositionalFieldName
			reduce(107), // .*, reduce: PositionalFieldName
			reduce(107), // ./, reduce: PositionalFieldName
			reduce(107), // .//, reduce: PositionalFieldName
			reduce(107), // ., reduce: PositionalFieldName
			nil,         // !
			nil,         // ~
			reduce(107), // **, reduce: PositionalFieldName
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(107), // [, reduce: PositionalFieldName
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(107), // [[, reduce: PositionalFieldName
			reduce(107), // [[[, reduce: PositionalFieldName
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4185
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4436), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4186
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1870), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1874), // (
			nil,         // )
			nil,         // emitp
			shift(1886), // field_name
			shift(1887), // $[
			nil,         // ]
			shift(1888), // braced_field_name
			shift(1889), // $[[
			shift(1890), // $[[[
			shift(1891), // full_srec
			shift(1893), // oosvar_name
			shift(1894), // @[
			shift(1895), // braced_oosvar_name
			shift(1896), // full_oosvar
			shift(1897), // all
			shift(1898), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1913), // +
			shift(1915), // -
			shift(1916), // .+
			shift(1917), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1920), // !
			shift(1921), // ~
			nil,         // **
			shift(1926), // string_literal
			shift(1927), // regex_case_insensitive
			shift(1928), // int_literal
			shift(1929), // float_literal
			shift(1930), // boolean_literal
			shift(1931), // null_literal
			shift(1932), // inf_literal
			shift(1933), // nan_literal
			shift(1934), // const_M_PI
			shift(1935), // const_M_E
			shift(1936), // panic
			shift(1938), // [
			shift(1940), // ctx_IPS
			shift(1941), // ctx_IFS
			shift(1942), // ctx_IRS
			shift(1943), // ctx_OPS
			shift(1944), // ctx_OFS
			shift(1945), // ctx_ORS
			shift(1946), // ctx_FLATSEP
			shift(1947), // ctx_NF
			shift(1948), // ctx_NR
			shift(1949), // ctx_FNR
			shift(1950), // ctx_FILENAME
			shift(1951), // ctx_FILENUM
			shift(1953), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1960), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4187
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(206), // }, reduce: PowTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(206), // >, reduce: PowTerm
			reduce(206), // >>, reduce: PowTerm
			reduce(206), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(206), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(206), // ?, reduce: PowTerm
			nil,         // :
			reduce(206), // ||, reduce: PowTerm
			reduce(206), // ^^, reduce: PowTerm
			reduce(206), // &&, reduce: PowTerm
			reduce(206), // ??, reduce: PowTerm
			reduce(206), // ???, reduce: PowTerm
			reduce(206), // =~, reduce: PowTerm
			reduce(206), // !=~, reduce: PowTerm
			reduce(206), // ==, reduce: PowTerm
			reduce(206), // !=, reduce: PowTerm
			reduce(206), // <=>, reduce: PowTerm
			reduce(206), // >=, reduce: PowTerm
			reduce(206), // <, reduce: PowTerm
			reduce(206), // <=, reduce: PowTerm
			reduce(206), // ^, reduce: PowTerm
			reduce(206), // &, reduce: PowTerm
			reduce(206), // <<, reduce: PowTerm
			reduce(206), // >>>, reduce: PowTerm
			reduce(206), // +, reduce: PowTerm
			reduce(206), // -, reduce: PowTerm
			reduce(206), // .+, reduce: PowTerm
			reduce(206), // .-, reduce: PowTerm
			reduce(206), // *, reduce: PowTerm
			reduce(206), // /, reduce: PowTerm
			reduce(206), // //, reduce: PowTerm
			reduce(206), // %, reduce: PowTerm
			reduce(206), // .*, reduce: PowTerm
			reduce(206), // ./, reduce: PowTerm
			reduce(206), // .//, reduce: PowTerm
			reduce(206), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4188
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(205), // }, reduce: PowTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(205), // >, reduce: PowTerm
			reduce(205), // >>, reduce: PowTerm
			reduce(205), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(205), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(205), // ?, reduce: PowTerm
			nil,         // :
			reduce(205), // ||, reduce: PowTerm
			reduce(205), // ^^, reduce: PowTerm
			reduce(205), // &&, reduce: PowTerm
			reduce(205), // ??, reduce: PowTerm
			reduce(205), // ???, reduce: PowTerm
			reduce(205), // =~, reduce: PowTerm
			reduce(205), // !=~, reduce: PowTerm
			reduce(205), // ==, reduce: PowTerm
			reduce(205), // !=, reduce: PowTerm
			reduce(205), // <=>, reduce: PowTerm
			reduce(205), // >=, reduce: PowTerm
			reduce(205), // <, reduce: PowTerm
			reduce(205), // <=, reduce: PowTerm
			reduce(205), // ^, reduce: PowTerm
			reduce(205), // &, reduce: PowTerm
			reduce(205), // <<, reduce: PowTerm
			reduce(205), // >>>, reduce: PowTerm
			reduce(205), // +, reduce: PowTerm
			reduce(205), // -, reduce: PowTerm
			reduce(205), // .+, reduce: PowTerm
			reduce(205), // .-, reduce: PowTerm
			reduce(205), // *, reduce: PowTerm
			reduce(205), // /, reduce: PowTerm
			reduce(205), // //, reduce: PowTerm
			reduce(205), // %, reduce: PowTerm
			reduce(205), // .*, reduce: PowTerm
			reduce(205), // ./, reduce: PowTerm
			reduce(205), // .//, reduce: PowTerm
			reduce(205), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4189
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(260), // }, reduce: ArrayOrMapIndexAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(260), // >, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>, reduce: ArrayOrMapIndexAccess
			reduce(260), // |, reduce: ArrayOrMapIndexAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(260), // ,, reduce: ArrayOrMapIndexAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(260), // ?, reduce: ArrayOrMapIndexAccess
			nil,         // :
			reduce(260), // ||, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???, reduce: ArrayOrMapIndexAccess
			reduce(260), // =~, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=~, reduce: ArrayOrMapIndexAccess
			reduce(260), // ==, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=>, reduce: ArrayOrMapIndexAccess
			reduce(260), // >=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>, reduce: ArrayOrMapIndexAccess
			reduce(260), // +, reduce: ArrayOrMapIndexAccess
			reduce(260), // -, reduce: ArrayOrMapIndexAccess
			reduce(260), // .+, reduce: ArrayOrMapIndexAccess
			reduce(260), // .-, reduce: ArrayOrMapIndexAccess
			reduce(260), // *, reduce: ArrayOrMapIndexAccess
			reduce(260), // /, reduce: ArrayOrMapIndexAccess
			reduce(260), // //, reduce: ArrayOrMapIndexAccess
			reduce(260), // %, reduce: ArrayOrMapIndexAccess
			reduce(260), // .*, reduce: ArrayOrMapIndexAccess
			reduce(260), // ./, reduce: ArrayOrMapIndexAccess
			reduce(260), // .//, reduce: ArrayOrMapIndexAccess
			reduce(260), // ., reduce: ArrayOrMapIndexAccess
			nil,         // !
			nil,         // ~
			reduce(260), // **, reduce: ArrayOrMapIndexAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(260), // [, reduce: ArrayOrMapIndexAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(260), // [[, reduce: ArrayOrMapIndexAccess
			reduce(260), // [[[, reduce: ArrayOrMapIndexAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4190
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4439), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4191
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4440), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4192
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(266), // }, reduce: ArraySliceAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(266), // >, reduce: ArraySliceAccess
			reduce(266), // >>, reduce: ArraySliceAccess
			reduce(266), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(266), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(266), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(266), // ||, reduce: ArraySliceAccess
			reduce(266), // ^^, reduce: ArraySliceAccess
			reduce(266), // &&, reduce: ArraySliceAccess
			reduce(266), // ??, reduce: ArraySliceAccess
			reduce(266), // ???, reduce: ArraySliceAccess
			reduce(266), // =~, reduce: ArraySliceAccess
			reduce(266), // !=~, reduce: ArraySliceAccess
			reduce(266), // ==, reduce: ArraySliceAccess
			reduce(266), // !=, reduce: ArraySliceAccess
			reduce(266), // <=>, reduce: ArraySliceAccess
			reduce(266), // >=, reduce: ArraySliceAccess
			reduce(266), // <, reduce: ArraySliceAccess
			reduce(266), // <=, reduce: ArraySliceAccess
			reduce(266), // ^, reduce: ArraySliceAccess
			reduce(266), // &, reduce: ArraySliceAccess
			reduce(266), // <<, reduce: ArraySliceAccess
			reduce(266), // >>>, reduce: ArraySliceAccess
			reduce(266), // +, reduce: ArraySliceAccess
			reduce(266), // -, reduce: ArraySliceAccess
			reduce(266), // .+, reduce: ArraySliceAccess
			reduce(266), // .-, reduce: ArraySliceAccess
			reduce(266), // *, reduce: ArraySliceAccess
			reduce(266), // /, reduce: ArraySliceAccess
			reduce(266), // //, reduce: ArraySliceAccess
			reduce(266), // %, reduce: ArraySliceAccess
			reduce(266), // .*, reduce: ArraySliceAccess
			reduce(266), // ./, reduce: ArraySliceAccess
			reduce(266), // .//, reduce: ArraySliceAccess
			reduce(266), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(266), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(266), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(266), // [[, reduce: ArraySliceAccess
			reduce(266), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4193
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4441), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4194
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4442), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4195
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(254), // }, reduce: ENV
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(254), // >, reduce: ENV
			reduce(254), // >>, reduce: ENV
			reduce(254), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(254), // ,, reduce: ENV
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(254), // ?, reduce: ENV
			nil,         // :
			reduce(254), // ||, reduce: ENV
			reduce(254), // ^^, reduce: ENV
			reduce(254), // &&, reduce: ENV
			reduce(254), // ??, reduce: ENV
			reduce(254), // ???, reduce: ENV
			reduce(254), // =~, reduce: ENV
			reduce(254), // !=~, reduce: ENV
			reduce(254), // ==, reduce: ENV
			reduce(254), // !=, reduce: ENV
			reduce(254), // <=>, reduce: ENV
			reduce(254), // >=, reduce: ENV
			reduce(254), // <, reduce: ENV
			reduce(254), // <=, reduce: ENV
			reduce(254), // ^, reduce: ENV
			reduce(254), // &, reduce: ENV
			reduce(254), // <<, reduce: ENV
			reduce(254), // >>>, reduce: ENV
			reduce(254), // +, reduce: ENV
			reduce(254), // -, reduce: ENV
			reduce(254), // .+, reduce: ENV
			reduce(254), // .-, reduce: ENV
			reduce(254), // *, reduce: ENV
			reduce(254), // /, reduce: ENV
			reduce(254), // //, reduce: ENV
			reduce(254), // %, reduce: ENV
			reduce(254), // .*, reduce: ENV
			reduce(254), // ./, reduce: ENV
			reduce(254), // .//, reduce: ENV
			reduce(254), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(254), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(254), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(254), // [[, reduce: ENV
			reduce(254), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4196
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(269), // }, reduce: FunctionCallsite
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(269), // >, reduce: FunctionCallsite
			reduce(269), // >>, reduce: FunctionCallsite
			reduce(269), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(269), // ,, reduce: FunctionCallsite
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(269), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(269), // ||, reduce: FunctionCallsite
			reduce(269), // ^^, reduce: FunctionCallsite
			reduce(269), // &&, reduce: FunctionCallsite
			reduce(269), // ??, reduce: FunctionCallsite
			reduce(269), // ???, reduce: FunctionCallsite
			reduce(269), // =~, reduce: FunctionCallsite
			reduce(269), // !=~, reduce: FunctionCallsite
			reduce(269), // ==, reduce: FunctionCallsite
			reduce(269), // !=, reduce: FunctionCallsite
			reduce(269), // <=>, reduce: FunctionCallsite
			reduce(269), // >=, reduce: FunctionCallsite
			reduce(269), // <, reduce: FunctionCallsite
			reduce(269), // <=, reduce: FunctionCallsite
			reduce(269), // ^, reduce: FunctionCallsite
			reduce(269), // &, reduce: FunctionCallsite
			reduce(269), // <<, reduce: FunctionCallsite
			reduce(269), // >>>, reduce: FunctionCallsite
			reduce(269), // +, reduce: FunctionCallsite
			reduce(269), // -, reduce: FunctionCallsite
			reduce(269), // .+, reduce: FunctionCallsite
			reduce(269), // .-, reduce: FunctionCallsite
			reduce(269), // *, reduce: FunctionCallsite
			reduce(269), // /, reduce: FunctionCallsite
			reduce(269), // //, reduce: FunctionCallsite
			reduce(269), // %, reduce: FunctionCallsite
			reduce(269), // .*, reduce: FunctionCallsite
			reduce(269), // ./, reduce: FunctionCallsite
			reduce(269), // .//, reduce: FunctionCallsite
			reduce(269), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(269), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(269), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(269), // [[, reduce: FunctionCallsite
			reduce(269), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4197
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4444), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4445), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4198
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(263), // >, reduce: ArraySliceAccess
			reduce(263), // >>, reduce: ArraySliceAccess
			reduce(263), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(263), // ?, reduce: ArraySliceAccess
			reduce(263), // :, reduce: ArraySliceAccess
			reduce(263), // ||, reduce: ArraySliceAccess
			reduce(263), // ^^, reduce: ArraySliceAccess
			reduce(263), // &&, reduce: ArraySliceAccess
			reduce(263), // ??, reduce: ArraySliceAccess
			reduce(263), // ???, reduce: ArraySliceAccess
			reduce(263), // =~, reduce: ArraySliceAccess
			reduce(263), // !=~, reduce: ArraySliceAccess
			reduce(263), // ==, reduce: ArraySliceAccess
			reduce(263), // !=, reduce: ArraySliceAccess
			reduce(263), // <=>, reduce: ArraySliceAccess
			reduce(263), // >=, reduce: ArraySliceAccess
			reduce(263), // <, reduce: ArraySliceAccess
			reduce(263), // <=, reduce: ArraySliceAccess
			reduce(263), // ^, reduce: ArraySliceAccess
			reduce(263), // &, reduce: ArraySliceAccess
			reduce(263), // <<, reduce: ArraySliceAccess
			reduce(263), // >>>, reduce: ArraySliceAccess
			reduce(263), // +, reduce: ArraySliceAccess
			reduce(263), // -, reduce: ArraySliceAccess
			reduce(263), // .+, reduce: ArraySliceAccess
			reduce(263), // .-, reduce: ArraySliceAccess
			reduce(263), // *, reduce: ArraySliceAccess
			reduce(263), // /, reduce: ArraySliceAccess
			reduce(263), // //, reduce: ArraySliceAccess
			reduce(263), // %, reduce: ArraySliceAccess
			reduce(263), // .*, reduce: ArraySliceAccess
			reduce(263), // ./, reduce: ArraySliceAccess
			reduce(263), // .//, reduce: ArraySliceAccess
			reduce(263), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(263), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(263), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(263), // [[, reduce: ArraySliceAccess
			reduce(263), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4199
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(262), // >, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |, reduce: ArrayOrMapPositionalValueAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(262), // ?, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // :, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ||, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // =~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ==, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .+, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .-, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .*, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ./, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .//, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ., reduce: ArrayOrMapPositionalValueAccess
			nil,         // !
			nil,         // ~
			reduce(262), // **, reduce: ArrayOrMapPositionalValueAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(262), // [, reduce: ArrayOrMapPositionalValueAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(262), // [[, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // [[[, reduce: ArrayOrMapPositionalValueAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4200
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4446), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4201
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3779), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4202
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(107), // ;, reduce: PositionalFieldName
			nil,         // {
			reduce(107), // }, reduce: PositionalFieldName
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(107), // >, reduce: PositionalFieldName
			reduce(107), // >>, reduce: PositionalFieldName
			reduce(107), // |, reduce: PositionalFieldName
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(107), // ,, reduce: PositionalFieldName
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(107), // ?, reduce: PositionalFieldName
			nil,         // :
			reduce(107), // ||, reduce: PositionalFieldName
			reduce(107), // ^^, reduce: PositionalFieldName
			reduce(107), // &&, reduce: PositionalFieldName
			reduce(107), // ??, reduce: PositionalFieldName
			reduce(107), // ???, reduce: PositionalFieldName
			reduce(107), // =~, reduce: PositionalFieldName
			reduce(107), // !=~, reduce: PositionalFieldName
			reduce(107), // ==, reduce: PositionalFieldName
			reduce(107), // !=, reduce: PositionalFieldName
			reduce(107), // <=>, reduce: PositionalFieldName
			reduce(107), // >=, reduce: PositionalFieldName
			reduce(107), // <, reduce: PositionalFieldName
			reduce(107), // <=, reduce: PositionalFieldName
			reduce(107), // ^, reduce: PositionalFieldName
			reduce(107), // &, reduce: PositionalFieldName
			reduce(107), // <<, reduce: PositionalFieldName
			reduce(107), // >>>, reduce: PositionalFieldName
			reduce(107), // +, reduce: PositionalFieldName
			reduce(107), // -, reduce: PositionalFieldName
			reduce(107), // .+, reduce: PositionalFieldName
			reduce(107), // .-, reduce: PositionalFieldName
			reduce(107), // *, reduce: PositionalFieldName
			reduce(107), // /, reduce: PositionalFieldName
			reduce(107), // //, reduce: PositionalFieldName
			reduce(107), // %, reduce: PositionalFieldName
			reduce(107), // .*, reduce: PositionalFieldName
			reduce(107), // ./, reduce: PositionalFieldName
			reduce(107), // .//, reduce: PositionalFieldName
			reduce(107), // ., reduce: PositionalFieldName
			nil,         // !
			nil,         // ~
			reduce(107), // **, reduce: PositionalFieldName
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(107), // [, reduce: PositionalFieldName
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(107), // [[, reduce: PositionalFieldName
			reduce(107), // [[[, reduce: PositionalFieldName
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4203
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4448), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4204
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4205
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(206), // ;, reduce: PowTerm
			nil,         // {
			reduce(206), // }, reduce: PowTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(206), // >, reduce: PowTerm
			reduce(206), // >>, reduce: PowTerm
			reduce(206), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(206), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(206), // ?, reduce: PowTerm
			nil,         // :
			reduce(206), // ||, reduce: PowTerm
			reduce(206), // ^^, reduce: PowTerm
			reduce(206), // &&, reduce: PowTerm
			reduce(206), // ??, reduce: PowTerm
			reduce(206), // ???, reduce: PowTerm
			reduce(206), // =~, reduce: PowTerm
			reduce(206), // !=~, reduce: PowTerm
			reduce(206), // ==, reduce: PowTerm
			reduce(206), // !=, reduce: PowTerm
			reduce(206), // <=>, reduce: PowTerm
			reduce(206), // >=, reduce: PowTerm
			reduce(206), // <, reduce: PowTerm
			reduce(206), // <=, reduce: PowTerm
			reduce(206), // ^, reduce: PowTerm
			reduce(206), // &, reduce: PowTerm
			reduce(206), // <<, reduce: PowTerm
			reduce(206), // >>>, reduce: PowTerm
			reduce(206), // +, reduce: PowTerm
			reduce(206), // -, reduce: PowTerm
			reduce(206), // .+, reduce: PowTerm
			reduce(206), // .-, reduce: PowTerm
			reduce(206), // *, reduce: PowTerm
			reduce(206), // /, reduce: PowTerm
			reduce(206), // //, reduce: PowTerm
			reduce(206), // %, reduce: PowTerm
			reduce(206), // .*, reduce: PowTerm
			reduce(206), // ./, reduce: PowTerm
			reduce(206), // .//, reduce: PowTerm
			reduce(206), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4206
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(205), // ;, reduce: PowTerm
			nil,         // {
			reduce(205), // }, reduce: PowTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(205), // >, reduce: PowTerm
			reduce(205), // >>, reduce: PowTerm
			reduce(205), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(205), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(205), // ?, reduce: PowTerm
			nil,         // :
			reduce(205), // ||, reduce: PowTerm
			reduce(205), // ^^, reduce: PowTerm
			reduce(205), // &&, reduce: PowTerm
			reduce(205), // ??, reduce: PowTerm
			reduce(205), // ???, reduce: PowTerm
			reduce(205), // =~, reduce: PowTerm
			reduce(205), // !=~, reduce: PowTerm
			reduce(205), // ==, reduce: PowTerm
			reduce(205), // !=, reduce: PowTerm
			reduce(205), // <=>, reduce: PowTerm
			reduce(205), // >=, reduce: PowTerm
			reduce(205), // <, reduce: PowTerm
			reduce(205), // <=, reduce: PowTerm
			reduce(205), // ^, reduce: PowTerm
			reduce(205), // &, reduce: PowTerm
			reduce(205), // <<, reduce: PowTerm
			reduce(205), // >>>, reduce: PowTerm
			reduce(205), // +, reduce: PowTerm
			reduce(205), // -, reduce: PowTerm
			reduce(205), // .+, reduce: PowTerm
			reduce(205), // .-, reduce: PowTerm
			reduce(205), // *, reduce: PowTerm
			reduce(205), // /, reduce: PowTerm
			reduce(205), // //, reduce: PowTerm
			reduce(205), // %, reduce: PowTerm
			reduce(205), // .*, reduce: PowTerm
			reduce(205), // ./, reduce: PowTerm
			reduce(205), // .//, reduce: PowTerm
			reduce(205), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4207
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(260), // ;, reduce: ArrayOrMapIndexAccess
			nil,         // {
			reduce(260), // }, reduce: ArrayOrMapIndexAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(260), // >, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>, reduce: ArrayOrMapIndexAccess
			reduce(260), // |, reduce: ArrayOrMapIndexAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(260), // ,, reduce: ArrayOrMapIndexAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(260), // ?, reduce: ArrayOrMapIndexAccess
			nil,         // :
			reduce(260), // ||, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???, reduce: ArrayOrMapIndexAccess
			reduce(260), // =~, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=~, reduce: ArrayOrMapIndexAccess
			reduce(260), // ==, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=>, reduce: ArrayOrMapIndexAccess
			reduce(260), // >=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>, reduce: ArrayOrMapIndexAccess
			reduce(260), // +, reduce: ArrayOrMapIndexAccess
			reduce(260), // -, reduce: ArrayOrMapIndexAccess
			reduce(260), // .+, reduce: ArrayOrMapIndexAccess
			reduce(260), // .-, reduce: ArrayOrMapIndexAccess
			reduce(260), // *, reduce: ArrayOrMapIndexAccess
			reduce(260), // /, reduce: ArrayOrMapIndexAccess
			reduce(260), // //, reduce: ArrayOrMapIndexAccess
			reduce(260), // %, reduce: ArrayOrMapIndexAccess
			reduce(260), // .*, reduce: ArrayOrMapIndexAccess
			reduce(260), // ./, reduce: ArrayOrMapIndexAccess
			reduce(260), // .//, reduce: ArrayOrMapIndexAccess
			reduce(260), // ., reduce: ArrayOrMapIndexAccess
			nil,         // !
			nil,         // ~
			reduce(260), // **, reduce: ArrayOrMapIndexAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(260), // [, reduce: ArrayOrMapIndexAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(260), // [[, reduce: ArrayOrMapIndexAccess
			reduce(260), // [[[, reduce: ArrayOrMapIndexAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4208
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4451), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4209
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4452), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4210
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(266), // ;, reduce: ArraySliceAccess
			nil,         // {
			reduce(266), // }, reduce: ArraySliceAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(266), // >, reduce: ArraySliceAccess
			reduce(266), // >>, reduce: ArraySliceAccess
			reduce(266), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(266), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(266), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(266), // ||, reduce: ArraySliceAccess
			reduce(266), // ^^, reduce: ArraySliceAccess
			reduce(266), // &&, reduce: ArraySliceAccess
			reduce(266), // ??, reduce: ArraySliceAccess
			reduce(266), // ???, reduce: ArraySliceAccess
			reduce(266), // =~, reduce: ArraySliceAccess
			reduce(266), // !=~, reduce: ArraySliceAccess
			reduce(266), // ==, reduce: ArraySliceAccess
			reduce(266), // !=, reduce: ArraySliceAccess
			reduce(266), // <=>, reduce: ArraySliceAccess
			reduce(266), // >=, reduce: ArraySliceAccess
			reduce(266), // <, reduce: ArraySliceAccess
			reduce(266), // <=, reduce: ArraySliceAccess
			reduce(266), // ^, reduce: ArraySliceAccess
			reduce(266), // &, reduce: ArraySliceAccess
			reduce(266), // <<, reduce: ArraySliceAccess
			reduce(266), // >>>, reduce: ArraySliceAccess
			reduce(266), // +, reduce: ArraySliceAccess
			reduce(266), // -, reduce: ArraySliceAccess
			reduce(266), // .+, reduce: ArraySliceAccess
			reduce(266), // .-, reduce: ArraySliceAccess
			reduce(266), // *, reduce: ArraySliceAccess
			reduce(266), // /, reduce: ArraySliceAccess
			reduce(266), // //, reduce: ArraySliceAccess
			reduce(266), // %, reduce: ArraySliceAccess
			reduce(266), // .*, reduce: ArraySliceAccess
			reduce(266), // ./, reduce: ArraySliceAccess
			reduce(266), // .//, reduce: ArraySliceAccess
			reduce(266), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(266), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(266), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(266), // [[, reduce: ArraySliceAccess
			reduce(266), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4211
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4453), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4212
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4454), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4213
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(254), // ;, reduce: ENV
			nil,         // {
			reduce(254), // }, reduce: ENV
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(254), // >, reduce: ENV
			reduce(254), // >>, reduce: ENV
			reduce(254), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(254), // ,, reduce: ENV
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(254), // ?, reduce: ENV
			nil,         // :
			reduce(254), // ||, reduce: ENV
			reduce(254), // ^^, reduce: ENV
			reduce(254), // &&, reduce: ENV
			reduce(254), // ??, reduce: ENV
			reduce(254), // ???, reduce: ENV
			reduce(254), // =~, reduce: ENV
			reduce(254), // !=~, reduce: ENV
			reduce(254), // ==, reduce: ENV
			reduce(254), // !=, reduce: ENV
			reduce(254), // <=>, reduce: ENV
			reduce(254), // >=, reduce: ENV
			reduce(254), // <, reduce: ENV
			reduce(254), // <=, reduce: ENV
			reduce(254), // ^, reduce: ENV
			reduce(254), // &, reduce: ENV
			reduce(254), // <<, reduce: ENV
			reduce(254), // >>>, reduce: ENV
			reduce(254), // +, reduce: ENV
			reduce(254), // -, reduce: ENV
			reduce(254), // .+, reduce: ENV
			reduce(254), // .-, reduce: ENV
			reduce(254), // *, reduce: ENV
			reduce(254), // /, reduce: ENV
			reduce(254), // //, reduce: ENV
			reduce(254), // %, reduce: ENV
			reduce(254), // .*, reduce: ENV
			reduce(254), // ./, reduce: ENV
			reduce(254), // .//, reduce: ENV
			reduce(254), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(254), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(254), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(254), // [[, reduce: ENV
			reduce(254), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4214
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(269), // ;, reduce: FunctionCallsite
			nil,         // {
			reduce(269), // }, reduce: FunctionCallsite
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(269), // >, reduce: FunctionCallsite
			reduce(269), // >>, reduce: FunctionCallsite
			reduce(269), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(269), // ,, reduce: FunctionCallsite
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(269), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(269), // ||, reduce: FunctionCallsite
			reduce(269), // ^^, reduce: FunctionCallsite
			reduce(269), // &&, reduce: FunctionCallsite
			reduce(269), // ??, reduce: FunctionCallsite
			reduce(269), // ???, reduce: FunctionCallsite
			reduce(269), // =~, reduce: FunctionCallsite
			reduce(269), // !=~, reduce: FunctionCallsite
			reduce(269), // ==, reduce: FunctionCallsite
			reduce(269), // !=, reduce: FunctionCallsite
			reduce(269), // <=>, reduce: FunctionCallsite
			reduce(269), // >=, reduce: FunctionCallsite
			reduce(269), // <, reduce: FunctionCallsite
			reduce(269), // <=, reduce: FunctionCallsite
			reduce(269), // ^, reduce: FunctionCallsite
			reduce(269), // &, reduce: FunctionCallsite
			reduce(269), // <<, reduce: FunctionCallsite
			reduce(269), // >>>, reduce: FunctionCallsite
			reduce(269), // +, reduce: FunctionCallsite
			reduce(269), // -, reduce: FunctionCallsite
			reduce(269), // .+, reduce: FunctionCallsite
			reduce(269), // .-, reduce: FunctionCallsite
			reduce(269), // *, reduce: FunctionCallsite
			reduce(269), // /, reduce: FunctionCallsite
			reduce(269), // //, reduce: FunctionCallsite
			reduce(269), // %, reduce: FunctionCallsite
			reduce(269), // .*, reduce: FunctionCallsite
			reduce(269), // ./, reduce: FunctionCallsite
			reduce(269), // .//, reduce: FunctionCallsite
			reduce(269), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(269), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(269), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(269), // [[, reduce: FunctionCallsite
			reduce(269), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4215
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4456), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4457), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4216
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(107), // ;, reduce: PositionalFieldName
			nil,         // {
			reduce(107), // }, reduce: PositionalFieldName
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(107), // >, reduce: PositionalFieldName
			reduce(107), // >>, reduce: PositionalFieldName
			reduce(107), // |, reduce: PositionalFieldName
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(107), // ?, reduce: PositionalFieldName
			nil,         // :
			reduce(107), // ||, reduce: PositionalFieldName
			reduce(107), // ^^, reduce: PositionalFieldName
			reduce(107), // &&, reduce: PositionalFieldName
			reduce(107), // ??, reduce: PositionalFieldName
			reduce(107), // ???, reduce: PositionalFieldName
			reduce(107), // =~, reduce: PositionalFieldName
			reduce(107), // !=~, reduce: PositionalFieldName
			reduce(107), // ==, reduce: PositionalFieldName
			reduce(107), // !=, reduce: PositionalFieldName
			reduce(107), // <=>, reduce: PositionalFieldName
			reduce(107), // >=, reduce: PositionalFieldName
			reduce(107), // <, reduce: PositionalFieldName
			reduce(107), // <=, reduce: PositionalFieldName
			reduce(107), // ^, reduce: PositionalFieldName
			reduce(107), // &, reduce: PositionalFieldName
			reduce(107), // <<, reduce: PositionalFieldName
			reduce(107), // >>>, reduce: PositionalFieldName
			reduce(107), // +, reduce: PositionalFieldName
			reduce(107), // -, reduce: PositionalFieldName
			reduce(107), // .+, reduce: PositionalFieldName
			reduce(107), // .-, reduce: PositionalFieldName
			reduce(107), // *, reduce: PositionalFieldName
			reduce(107), // /, reduce: PositionalFieldName
			reduce(107), // //, reduce: PositionalFieldName
			reduce(107), // %, reduce: PositionalFieldName
			reduce(107), // .*, reduce: PositionalFieldName
			reduce(107), // ./, reduce: PositionalFieldName
			reduce(107), // .//, reduce: PositionalFieldName
			reduce(107), // ., reduce: PositionalFieldName
			nil,         // !
			nil,         // ~
			reduce(107), // **, reduce: PositionalFieldName
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(107), // [, reduce: PositionalFieldName
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(107), // [[, reduce: PositionalFieldName
			reduce(107), // [[[, reduce: PositionalFieldName
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4217
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4458), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4218
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2135), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2139), // (
			nil,         // )
			nil,         // emitp
			shift(2151), // field_name
			shift(2152), // $[
			nil,         // ]
			shift(2153), // braced_field_name
			shift(2154), // $[[
			shift(2155), // $[[[
			shift(2156), // full_srec
			shift(2158), // oosvar_name
			shift(2159), // @[
			shift(2160), // braced_oosvar_name
			shift(2161), // full_oosvar
			shift(2162), // all
			shift(2163), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2178), // +
			shift(2180), // -
			shift(2181), // .+
			shift(2182), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2185), // !
			shift(2186), // ~
			nil,         // **
			shift(2191), // string_literal
			shift(2192), // regex_case_insensitive
			shift(2193), // int_literal
			shift(2194), // float_literal
			shift(2195), // boolean_literal
			shift(2196), // null_literal
			shift(2197), // inf_literal
			shift(2198), // nan_literal
			shift(2199), // const_M_PI
			shift(2200), // const_M_E
			shift(2201), // panic
			shift(2203), // [
			shift(2205), // ctx_IPS
			shift(2206), // ctx_IFS
			shift(2207), // ctx_IRS
			shift(2208), // ctx_OPS
			shift(2209), // ctx_OFS
			shift(2210), // ctx_ORS
			shift(2211), // ctx_FLATSEP
			shift(2212), // ctx_NF
			shift(2213), // ctx_NR
			shift(2214), // ctx_FNR
			shift(2215), // ctx_FILENAME
			shift(2216), // ctx_FILENUM
			shift(2218), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2225), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4219
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(206), // ;, reduce: PowTerm
			nil,         // {
			reduce(206), // }, reduce: PowTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(206), // >, reduce: PowTerm
			reduce(206), // >>, reduce: PowTerm
			reduce(206), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(206), // ?, reduce: PowTerm
			nil,         // :
			reduce(206), // ||, reduce: PowTerm
			reduce(206), // ^^, reduce: PowTerm
			reduce(206), // &&, reduce: PowTerm
			reduce(206), // ??, reduce: PowTerm
			reduce(206), // ???, reduce: PowTerm
			reduce(206), // =~, reduce: PowTerm
			reduce(206), // !=~, reduce: PowTerm
			reduce(206), // ==, reduce: PowTerm
			reduce(206), // !=, reduce: PowTerm
			reduce(206), // <=>, reduce: PowTerm
			reduce(206), // >=, reduce: PowTerm
			reduce(206), // <, reduce: PowTerm
			reduce(206), // <=, reduce: PowTerm
			reduce(206), // ^, reduce: PowTerm
			reduce(206), // &, reduce: PowTerm
			reduce(206), // <<, reduce: PowTerm
			reduce(206), // >>>, reduce: PowTerm
			reduce(206), // +, reduce: PowTerm
			reduce(206), // -, reduce: PowTerm
			reduce(206), // .+, reduce: PowTerm
			reduce(206), // .-, reduce: PowTerm
			reduce(206), // *, reduce: PowTerm
			reduce(206), // /, reduce: PowTerm
			reduce(206), // //, reduce: PowTerm
			reduce(206), // %, reduce: PowTerm
			reduce(206), // .*, reduce: PowTerm
			reduce(206), // ./, reduce: PowTerm
			reduce(206), // .//, reduce: PowTerm
			reduce(206), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4220
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(205), // ;, reduce: PowTerm
			nil,         // {
			reduce(205), // }, reduce: PowTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(205), // >, reduce: PowTerm
			reduce(205), // >>, reduce: PowTerm
			reduce(205), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(205), // ?, reduce: PowTerm
			nil,         // :
			reduce(205), // ||, reduce: PowTerm
			reduce(205), // ^^, reduce: PowTerm
			reduce(205), // &&, reduce: PowTerm
			reduce(205), // ??, reduce: PowTerm
			reduce(205), // ???, reduce: PowTerm
			reduce(205), // =~, reduce: PowTerm
			reduce(205), // !=~, reduce: PowTerm
			reduce(205), // ==, reduce: PowTerm
			reduce(205), // !=, reduce: PowTerm
			reduce(205), // <=>, reduce: PowTerm
			reduce(205), // >=, reduce: PowTerm
			reduce(205), // <, reduce: PowTerm
			reduce(205), // <=, reduce: PowTerm
			reduce(205), // ^, reduce: PowTerm
			reduce(205), // &, reduce: PowTerm
			reduce(205), // <<, reduce: PowTerm
			reduce(205), // >>>, reduce: PowTerm
			reduce(205), // +, reduce: PowTerm
			reduce(205), // -, reduce: PowTerm
			reduce(205), // .+, reduce: PowTerm
			reduce(205), // .-, reduce: PowTerm
			reduce(205), // *, reduce: PowTerm
			reduce(205), // /, reduce: PowTerm
			reduce(205), // //, reduce: PowTerm
			reduce(205), // %, reduce: PowTerm
			reduce(205), // .*, reduce: PowTerm
			reduce(205), // ./, reduce: PowTerm
			reduce(205), // .//, reduce: PowTerm
			reduce(205), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4221
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(260), // ;, reduce: ArrayOrMapIndexAccess
			nil,         // {
			reduce(260), // }, reduce: ArrayOrMapIndexAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(260), // >, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>, reduce: ArrayOrMapIndexAccess
			reduce(260), // |, reduce: ArrayOrMapIndexAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(260), // ?, reduce: ArrayOrMapIndexAccess
			nil,         // :
			reduce(260), // ||, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???, reduce: ArrayOrMapIndexAccess
			reduce(260), // =~, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=~, reduce: ArrayOrMapIndexAccess
			reduce(260), // ==, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=>, reduce: ArrayOrMapIndexAccess
			reduce(260), // >=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>, reduce: ArrayOrMapIndexAccess
			reduce(260), // +, reduce: ArrayOrMapIndexAccess
			reduce(260), // -, reduce: ArrayOrMapIndexAccess
			reduce(260), // .+, reduce: ArrayOrMapIndexAccess
			reduce(260), // .-, reduce: ArrayOrMapIndexAccess
			reduce(260), // *, reduce: ArrayOrMapIndexAccess
			reduce(260), // /, reduce: ArrayOrMapIndexAccess
			reduce(260), // //, reduce: ArrayOrMapIndexAccess
			reduce(260), // %, reduce: ArrayOrMapIndexAccess
			reduce(260), // .*, reduce: ArrayOrMapIndexAccess
			reduce(260), // ./, reduce: ArrayOrMapIndexAccess
			reduce(260), // .//, reduce: ArrayOrMapIndexAccess
			reduce(260), // ., reduce: ArrayOrMapIndexAccess
			nil,         // !
			nil,         // ~
			reduce(260), // **, reduce: ArrayOrMapIndexAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(260), // [, reduce: ArrayOrMapIndexAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(260), // [[, reduce: ArrayOrMapIndexAccess
			reduce(260), // [[[, reduce: ArrayOrMapIndexAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4222
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4461), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4223
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4462), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4224
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(266), // ;, reduce: ArraySliceAccess
			nil,         // {
			reduce(266), // }, reduce: ArraySliceAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(266), // >, reduce: ArraySliceAccess
			reduce(266), // >>, reduce: ArraySliceAccess
			reduce(266), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(266), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(266), // ||, reduce: ArraySliceAccess
			reduce(266), // ^^, reduce: ArraySliceAccess
			reduce(266), // &&, reduce: ArraySliceAccess
			reduce(266), // ??, reduce: ArraySliceAccess
			reduce(266), // ???, reduce: ArraySliceAccess
			reduce(266), // =~, reduce: ArraySliceAccess
			reduce(266), // !=~, reduce: ArraySliceAccess
			reduce(266), // ==, reduce: ArraySliceAccess
			reduce(266), // !=, reduce: ArraySliceAccess
			reduce(266), // <=>, reduce: ArraySliceAccess
			reduce(266), // >=, reduce: ArraySliceAccess
			reduce(266), // <, reduce: ArraySliceAccess
			reduce(266), // <=, reduce: ArraySliceAccess
			reduce(266), // ^, reduce: ArraySliceAccess
			reduce(266), // &, reduce: ArraySliceAccess
			reduce(266), // <<, reduce: ArraySliceAccess
			reduce(266), // >>>, reduce: ArraySliceAccess
			reduce(266), // +, reduce: ArraySliceAccess
			reduce(266), // -, reduce: ArraySliceAccess
			reduce(266), // .+, reduce: ArraySliceAccess
			reduce(266), // .-, reduce: ArraySliceAccess
			reduce(266), // *, reduce: ArraySliceAccess
			reduce(266), // /, reduce: ArraySliceAccess
			reduce(266), // //, reduce: ArraySliceAccess
			reduce(266), // %, reduce: ArraySliceAccess
			reduce(266), // .*, reduce: ArraySliceAccess
			reduce(266), // ./, reduce: ArraySliceAccess
			reduce(266), // .//, reduce: ArraySliceAccess
			reduce(266), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(266), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(266), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(266), // [[, reduce: ArraySliceAccess
			reduce(266), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4225
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4463), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4226
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4464), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4227
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(254), // ;, reduce: ENV
			nil,         // {
			reduce(254), // }, reduce: ENV
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(254), // >, reduce: ENV
			reduce(254), // >>, reduce: ENV
			reduce(254), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(254), // ?, reduce: ENV
			nil,         // :
			reduce(254), // ||, reduce: ENV
			reduce(254), // ^^, reduce: ENV
			reduce(254), // &&, reduce: ENV
			reduce(254), // ??, reduce: ENV
			reduce(254), // ???, reduce: ENV
			reduce(254), // =~, reduce: ENV
			reduce(254), // !=~, reduce: ENV
			reduce(254), // ==, reduce: ENV
			reduce(254), // !=, reduce: ENV
			reduce(254), // <=>, reduce: ENV
			reduce(254), // >=, reduce: ENV
			reduce(254), // <, reduce: ENV
			reduce(254), // <=, reduce: ENV
			reduce(254), // ^, reduce: ENV
			reduce(254), // &, reduce: ENV
			reduce(254), // <<, reduce: ENV
			reduce(254), // >>>, reduce: ENV
			reduce(254), // +, reduce: ENV
			reduce(254), // -, reduce: ENV
			reduce(254), // .+, reduce: ENV
			reduce(254), // .-, reduce: ENV
			reduce(254), // *, reduce: ENV
			reduce(254), // /, reduce: ENV
			reduce(254), // //, reduce: ENV
			reduce(254), // %, reduce: ENV
			reduce(254), // .*, reduce: ENV
			reduce(254), // ./, reduce: ENV
			reduce(254), // .//, reduce: ENV
			reduce(254), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(254), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(254), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(254), // [[, reduce: ENV
			reduce(254), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4228
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(269), // ;, reduce: FunctionCallsite
			nil,         // {
			reduce(269), // }, reduce: FunctionCallsite
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(269), // >, reduce: FunctionCallsite
			reduce(269), // >>, reduce: FunctionCallsite
			reduce(269), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(269), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(269), // ||, reduce: FunctionCallsite
			reduce(269), // ^^, reduce: FunctionCallsite
			reduce(269), // &&, reduce: FunctionCallsite
			reduce(269), // ??, reduce: FunctionCallsite
			reduce(269), // ???, reduce: FunctionCallsite
			reduce(269), // =~, reduce: FunctionCallsite
			reduce(269), // !=~, reduce: FunctionCallsite
			reduce(269), // ==, reduce: FunctionCallsite
			reduce(269), // !=, reduce: FunctionCallsite
			reduce(269), // <=>, reduce: FunctionCallsite
			reduce(269), // >=, reduce: FunctionCallsite
			reduce(269), // <, reduce: FunctionCallsite
			reduce(269), // <=, reduce: FunctionCallsite
			reduce(269), // ^, reduce: FunctionCallsite
			reduce(269), // &, reduce: FunctionCallsite
			reduce(269), // <<, reduce: FunctionCallsite
			reduce(269), // >>>, reduce: FunctionCallsite
			reduce(269), // +, reduce: FunctionCallsite
			reduce(269), // -, reduce: FunctionCallsite
			reduce(269), // .+, reduce: FunctionCallsite
			reduce(269), // .-, reduce: FunctionCallsite
			reduce(269), // *, reduce: FunctionCallsite
			reduce(269), // /, reduce: FunctionCallsite
			reduce(269), // //, reduce: FunctionCallsite
			reduce(269), // %, reduce: FunctionCallsite
			reduce(269), // .*, reduce: FunctionCallsite
			reduce(269), // ./, reduce: FunctionCallsite
			reduce(269), // .//, reduce: FunctionCallsite
			reduce(269), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(269), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(269), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(269), // [[, reduce: FunctionCallsite
			reduce(269), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4229
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4466), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4467), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4230
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4231
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4469), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4232
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4233
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4234
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4235
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4473), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4236
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4237
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(108), // ;, reduce: PositionalFieldValue
			reduce(108), // {, reduce: PositionalFieldValue
			reduce(108), // }, reduce: PositionalFieldValue
			reduce(108), // =, reduce: PositionalFieldValue
			nil,         // unset
			nil,         // filter
			reduce(108), // >, reduce: PositionalFieldValue
			reduce(108), // >>, reduce: PositionalFieldValue
			reduce(108), // |, reduce: PositionalFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(108), // ||=, reduce: PositionalFieldValue
			reduce(108), // ^^=, reduce: PositionalFieldValue
			reduce(108), // &&=, reduce: PositionalFieldValue
			reduce(108), // ??=, reduce: PositionalFieldValue
			reduce(108), // ???=, reduce: PositionalFieldValue
			reduce(108), // |=, reduce: PositionalFieldValue
			reduce(108), // &=, reduce: PositionalFieldValue
			reduce(108), // ^=, reduce: PositionalFieldValue
			reduce(108), // <<=, reduce: PositionalFieldValue
			reduce(108), // >>=, reduce: PositionalFieldValue
			reduce(108), // >>>=, reduce: PositionalFieldValue
			reduce(108), // +=, reduce: PositionalFieldValue
			reduce(108), // .=, reduce: PositionalFieldValue
			reduce(108), // -=, reduce: PositionalFieldValue
			reduce(108), // *=, reduce: PositionalFieldValue
			reduce(108), // /=, reduce: PositionalFieldValue
			reduce(108), // //=, reduce: PositionalFieldValue
			reduce(108), // %=, reduce: PositionalFieldValue
			reduce(108), // **=, reduce: PositionalFieldValue
			reduce(108), // ?, reduce: PositionalFieldValue
			nil,         // :
			reduce(108), // ||, reduce: PositionalFieldValue
			reduce(108), // ^^, reduce: PositionalFieldValue
			reduce(108), // &&, reduce: PositionalFieldValue
			reduce(108), // ??, reduce: PositionalFieldValue
			reduce(108), // ???, reduce: PositionalFieldValue
			reduce(108), // =~, reduce: PositionalFieldValue
			reduce(108), // !=~, reduce: PositionalFieldValue
			reduce(108), // ==, reduce: PositionalFieldValue
			reduce(108), // !=, reduce: PositionalFieldValue
			reduce(108), // <=>, reduce: PositionalFieldValue
			reduce(108), // >=, reduce: PositionalFieldValue
			reduce(108), // <, reduce: PositionalFieldValue
			reduce(108), // <=, reduce: PositionalFieldValue
			reduce(108), // ^, reduce: PositionalFieldValue
			reduce(108), // &, reduce: PositionalFieldValue
			reduce(108), // <<, reduce: PositionalFieldValue
			reduce(108), // >>>, reduce: PositionalFieldValue
			reduce(108), // +, reduce: PositionalFieldValue
			reduce(108), // -, reduce: PositionalFieldValue
			reduce(108), // .+, reduce: PositionalFieldValue
			reduce(108), // .-, reduce: PositionalFieldValue
			reduce(108), // *, reduce: PositionalFieldValue
			reduce(108), // /, reduce: PositionalFieldValue
			reduce(108), // //, reduce: PositionalFieldValue
			reduce(108), // %, reduce: PositionalFieldValue
			reduce(108), // .*, reduce: PositionalFieldValue
			reduce(108), // ./, reduce: PositionalFieldValue
			reduce(108), // .//, reduce: PositionalFieldValue
			reduce(108), // ., reduce: PositionalFieldValue
			nil,         // !
			nil,         // ~
			reduce(108), // **, reduce: PositionalFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(108), // [, reduce: PositionalFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(108), // [[, reduce: PositionalFieldValue
			reduce(108), // [[[, reduce: PositionalFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4238
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(149), // ;, reduce: TernaryTerm
			reduce(149), // {, reduce: TernaryTerm
			reduce(149), // }, reduce: TernaryTerm
			reduce(149), // =, reduce: TernaryTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(149), // ||=, reduce: TernaryTerm
			reduce(149), // ^^=, reduce: TernaryTerm
			reduce(149), // &&=, reduce: TernaryTerm
			reduce(149), // ??=, reduce: TernaryTerm
			reduce(149), // ???=, reduce: TernaryTerm
			reduce(149), // |=, reduce: TernaryTerm
			reduce(149), // &=, reduce: TernaryTerm
			reduce(149), // ^=, reduce: TernaryTerm
			reduce(149), // <<=, reduce: TernaryTerm
			reduce(149), // >>=, reduce: TernaryTerm
			reduce(149), // >>>=, reduce: TernaryTerm
			reduce(149), // +=, reduce: TernaryTerm
			reduce(149), // .=, reduce: TernaryTerm
			reduce(149), // -=, reduce: TernaryTerm
			reduce(149), // *=, reduce: TernaryTerm
			reduce(149), // /=, reduce: TernaryTerm
			reduce(149), // //=, reduce: TernaryTerm
			reduce(149), // %=, reduce: TernaryTerm
			reduce(149), // **=, reduce: TernaryTerm
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4239
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4475), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4240
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(265), // ;, reduce: ArraySliceAccess
			reduce(265), // {, reduce: ArraySliceAccess
			reduce(265), // }, reduce: ArraySliceAccess
			reduce(265), // =, reduce: ArraySliceAccess
			nil,         // unset
			nil,         // filter
			reduce(265), // >, reduce: ArraySliceAccess
			reduce(265), // >>, reduce: ArraySliceAccess
			reduce(265), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(265), // ||=, reduce: ArraySliceAccess
			reduce(265), // ^^=, reduce: ArraySliceAccess
			reduce(265), // &&=, reduce: ArraySliceAccess
			reduce(265), // ??=, reduce: ArraySliceAccess
			reduce(265), // ???=, reduce: ArraySliceAccess
			reduce(265), // |=, reduce: ArraySliceAccess
			reduce(265), // &=, reduce: ArraySliceAccess
			reduce(265), // ^=, reduce: ArraySliceAccess
			reduce(265), // <<=, reduce: ArraySliceAccess
			reduce(265), // >>=, reduce: ArraySliceAccess
			reduce(265), // >>>=, reduce: ArraySliceAccess
			reduce(265), // +=, reduce: ArraySliceAccess
			reduce(265), // .=, reduce: ArraySliceAccess
			reduce(265), // -=, reduce: ArraySliceAccess
			reduce(265), // *=, reduce: ArraySliceAccess
			reduce(265), // /=, reduce: ArraySliceAccess
			reduce(265), // //=, reduce: ArraySliceAccess
			reduce(265), // %=, reduce: ArraySliceAccess
			reduce(265), // **=, reduce: ArraySliceAccess
			reduce(265), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(265), // ||, reduce: ArraySliceAccess
			reduce(265), // ^^, reduce: ArraySliceAccess
			reduce(265), // &&, reduce: ArraySliceAccess
			reduce(265), // ??, reduce: ArraySliceAccess
			reduce(265), // ???, reduce: ArraySliceAccess
			reduce(265), // =~, reduce: ArraySliceAccess
			reduce(265), // !=~, reduce: ArraySliceAccess
			reduce(265), // ==, reduce: ArraySliceAccess
			reduce(265), // !=, reduce: ArraySliceAccess
			reduce(265), // <=>, reduce: ArraySliceAccess
			reduce(265), // >=, reduce: ArraySliceAccess
			reduce(265), // <, reduce: ArraySliceAccess
			reduce(265), // <=, reduce: ArraySliceAccess
			reduce(265), // ^, reduce: ArraySliceAccess
			reduce(265), // &, reduce: ArraySliceAccess
			reduce(265), // <<, reduce: ArraySliceAccess
			reduce(265), // >>>, reduce: ArraySliceAccess
			reduce(265), // +, reduce: ArraySliceAccess
			reduce(265), // -, reduce: ArraySliceAccess
			reduce(265), // .+, reduce: ArraySliceAccess
			reduce(265), // .-, reduce: ArraySliceAccess
			reduce(265), // *, reduce: ArraySliceAccess
			reduce(265), // /, reduce: ArraySliceAccess
			reduce(265), // //, reduce: ArraySliceAccess
			reduce(265), // %, reduce: ArraySliceAccess
			reduce(265), // .*, reduce: ArraySliceAccess
			reduce(265), // ./, reduce: ArraySliceAccess
			reduce(265), // .//, reduce: ArraySliceAccess
			reduce(265), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(265), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(265), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(265), // [[, reduce: ArraySliceAccess
			reduce(265), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4241
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(264), // ;, reduce: ArraySliceAccess
			reduce(264), // {, reduce: ArraySliceAccess
			reduce(264), // }, reduce: ArraySliceAccess
			reduce(264), // =, reduce: ArraySliceAccess
			nil,         // unset
			nil,         // filter
			reduce(264), // >, reduce: ArraySliceAccess
			reduce(264), // >>, reduce: ArraySliceAccess
			reduce(264), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(264), // ||=, reduce: ArraySliceAccess
			reduce(264), // ^^=, reduce: ArraySliceAccess
			reduce(264), // &&=, reduce: ArraySliceAccess
			reduce(264), // ??=, reduce: ArraySliceAccess
			reduce(264), // ???=, reduce: ArraySliceAccess
			reduce(264), // |=, reduce: ArraySliceAccess
			reduce(264), // &=, reduce: ArraySliceAccess
			reduce(264), // ^=, reduce: ArraySliceAccess
			reduce(264), // <<=, reduce: ArraySliceAccess
			reduce(264), // >>=, reduce: ArraySliceAccess
			reduce(264), // >>>=, reduce: ArraySliceAccess
			reduce(264), // +=, reduce: ArraySliceAccess
			reduce(264), // .=, reduce: ArraySliceAccess
			reduce(264), // -=, reduce: ArraySliceAccess
			reduce(264), // *=, reduce: ArraySliceAccess
			reduce(264), // /=, reduce: ArraySliceAccess
			reduce(264), // //=, reduce: ArraySliceAccess
			reduce(264), // %=, reduce: ArraySliceAccess
			reduce(264), // **=, reduce: ArraySliceAccess
			reduce(264), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(264), // ||, reduce: ArraySliceAccess
			reduce(264), // ^^, reduce: ArraySliceAccess
			reduce(264), // &&, reduce: ArraySliceAccess
			reduce(264), // ??, reduce: ArraySliceAccess
			reduce(264), // ???, reduce: ArraySliceAccess
			reduce(264), // =~, reduce: ArraySliceAccess
			reduce(264), // !=~, reduce: ArraySliceAccess
			reduce(264), // ==, reduce: ArraySliceAccess
			reduce(264), // !=, reduce: ArraySliceAccess
			reduce(264), // <=>, reduce: ArraySliceAccess
			reduce(264), // >=, reduce: ArraySliceAccess
			reduce(264), // <, reduce: ArraySliceAccess
			reduce(264), // <=, reduce: ArraySliceAccess
			reduce(264), // ^, reduce: ArraySliceAccess
			reduce(264), // &, reduce: ArraySliceAccess
			reduce(264), // <<, reduce: ArraySliceAccess
			reduce(264), // >>>, reduce: ArraySliceAccess
			reduce(264), // +, reduce: ArraySliceAccess
			reduce(264), // -, reduce: ArraySliceAccess
			reduce(264), // .+, reduce: ArraySliceAccess
			reduce(264), // .-, reduce: ArraySliceAccess
			reduce(264), // *, reduce: ArraySliceAccess
			reduce(264), // /, reduce: ArraySliceAccess
			reduce(264), // //, reduce: ArraySliceAccess
			reduce(264), // %, reduce: ArraySliceAccess
			reduce(264), // .*, reduce: ArraySliceAccess
			reduce(264), // ./, reduce: ArraySliceAccess
			reduce(264), // .//, reduce: ArraySliceAccess
			reduce(264), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(264), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(264), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(264), // [[, reduce: ArraySliceAccess
			reduce(264), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4242
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(261), // ;, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // {, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // }, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // =, reduce: ArrayOrMapPositionalNameAccess
			nil,         // unset
			nil,         // filter
			reduce(261), // >, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |, reduce: ArrayOrMapPositionalNameAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(261), // ||=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // **=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ?, reduce: ArrayOrMapPositionalNameAccess
			nil,         // :
			reduce(261), // ||, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // =~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ==, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .+, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .-, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .*, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ./, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .//, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ., reduce: ArrayOrMapPositionalNameAccess
			nil,         // !
			nil,         // ~
			reduce(261), // **, reduce: ArrayOrMapPositionalNameAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(261), // [, reduce: ArrayOrMapPositionalNameAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(261), // [[, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // [[[, reduce: ArrayOrMapPositionalNameAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4243
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4476), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4244
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(277), // ;, reduce: SubroutineCallsite
			nil,         // {
			reduce(277), // }, reduce: SubroutineCallsite
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4245
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4247), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4246
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(294), // ;, reduce: IfBlock
			reduce(294), // {, reduce: IfBlock
			reduce(294), // }, reduce: IfBlock
			nil,         // =
			reduce(294), // unset, reduce: IfBlock
			reduce(294), // filter, reduce: IfBlock
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(294), // print, reduce: IfBlock
			nil,         // ,
			reduce(294), // printn, reduce: IfBlock
			reduce(294), // eprint, reduce: IfBlock
			reduce(294), // eprintn, reduce: IfBlock
			reduce(294), // dump, reduce: IfBlock
			reduce(294), // edump, reduce: IfBlock
			reduce(294), // tee, reduce: IfBlock
			reduce(294), // emitf, reduce: IfBlock
			reduce(294), // emit1, reduce: IfBlock
			reduce(294), // emit, reduce: IfBlock
			reduce(294), // (, reduce: IfBlock
			nil,         // )
			reduce(294), // emitp, reduce: IfBlock
			reduce(294), // field_name, reduce: IfBlock
			reduce(294), // $[, reduce: IfBlock
			nil,         // ]
			reduce(294), // braced_field_name, reduce: IfBlock
			reduce(294), // $[[, reduce: IfBlock
			reduce(294), // $[[[, reduce: IfBlock
			reduce(294), // full_srec, reduce: IfBlock
			reduce(294), // oosvar_name, reduce: IfBlock
			reduce(294), // @[, reduce: IfBlock
			reduce(294), // braced_oosvar_name, reduce: IfBlock
			reduce(294), // full_oosvar, reduce: IfBlock
			reduce(294), // all, reduce: IfBlock
			reduce(294), // non_sigil_name, reduce: IfBlock
			reduce(294), // arr, reduce: IfBlock
			reduce(294), // bool, reduce: IfBlock
			reduce(294), // float, reduce: IfBlock
			reduce(294), // int, reduce: IfBlock
			reduce(294), // map, reduce: IfBlock
			reduce(294), // num, reduce: IfBlock
			reduce(294), // str, reduce: IfBlock
			reduce(294), // var, reduce: IfBlock
			reduce(294), // funct, reduce: IfBlock
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(294), // +, reduce: IfBlock
			reduce(294), // -, reduce: IfBlock
			reduce(294), // .+, reduce: IfBlock
			reduce(294), // .-, reduce: IfBlock
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(294), // !, reduce: IfBlock
			reduce(294), // ~, reduce: IfBlock
			nil,         // **
			reduce(294), // string_literal, reduce: IfBlock
			reduce(294), // regex_case_insensitive, reduce: IfBlock
			reduce(294), // int_literal, reduce: IfBlock
			reduce(294), // float_literal, reduce: IfBlock
			reduce(294), // boolean_literal, reduce: IfBlock
			reduce(294), // null_literal, reduce: IfBlock
			reduce(294), // inf_literal, reduce: IfBlock
			reduce(294), // nan_literal, reduce: IfBlock
			reduce(294), // const_M_PI, reduce: IfBlock
			reduce(294), // const_M_E, reduce: IfBlock
			reduce(294), // panic, reduce: IfBlock
			reduce(294), // [, reduce: IfBlock
			reduce(294), // ctx_IPS, reduce: IfBlock
			reduce(294), // ctx_IFS, reduce: IfBlock
			reduce(294), // ctx_IRS, reduce: IfBlock
			reduce(294), // ctx_OPS, reduce: IfBlock
			reduce(294), // ctx_OFS, reduce: IfBlock
			reduce(294), // ctx_ORS, reduce: IfBlock
			reduce(294), // ctx_FLATSEP, reduce: IfBlock
			reduce(294), // ctx_NF, reduce: IfBlock
			reduce(294), // ctx_NR, reduce: IfBlock
			reduce(294), // ctx_FNR, reduce: IfBlock
			reduce(294), // ctx_FILENAME, reduce: IfBlock
			reduce(294), // ctx_FILENUM, reduce: IfBlock
			reduce(294), // env, reduce: IfBlock
			nil,         // [[
			nil,         // [[[
			reduce(294), // call, reduce: IfBlock
			reduce(294), // begin, reduce: IfBlock
			reduce(294), // end, reduce: IfBlock
			reduce(294), // if, reduce: IfBlock
			reduce(294), // elif, reduce: IfBlock
			reduce(294), // else, reduce: IfBlock
			reduce(294), // while, reduce: IfBlock
			reduce(294), // do, reduce: IfBlock
			reduce(294), // for, reduce: IfBlock
			nil,         // in
			reduce(294), // break, reduce: IfBlock
			reduce(294), // continue, reduce: IfBlock
			reduce(294), // func, reduce: IfBlock
			reduce(294), // subr, reduce: IfBlock
			reduce(294), // return, reduce: IfBlock
		},
	},
	actionRow{ // S4247
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S4248
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(297), // ;, reduce: WhileLoop
			reduce(297), // {, reduce: WhileLoop
			reduce(297), // }, reduce: WhileLoop
			nil,         // =
			reduce(297), // unset, reduce: WhileLoop
			reduce(297), // filter, reduce: WhileLoop
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(297), // print, reduce: WhileLoop
			nil,         // ,
			reduce(297), // printn, reduce: WhileLoop
			reduce(297), // eprint, reduce: WhileLoop
			reduce(297), // eprintn, reduce: WhileLoop
			reduce(297), // dump, reduce: WhileLoop
			reduce(297), // edump, reduce: WhileLoop
			reduce(297), // tee, reduce: WhileLoop
			reduce(297), // emitf, reduce: WhileLoop
			reduce(297), // emit1, reduce: WhileLoop
			reduce(297), // emit, reduce: WhileLoop
			reduce(297), // (, reduce: WhileLoop
			nil,         // )
			reduce(297), // emitp, reduce: WhileLoop
			reduce(297), // field_name, reduce: WhileLoop
			reduce(297), // $[, reduce: WhileLoop
			nil,         // ]
			reduce(297), // braced_field_name, reduce: WhileLoop
			reduce(297), // $[[, reduce: WhileLoop
			reduce(297), // $[[[, reduce: WhileLoop
			reduce(297), // full_srec, reduce: WhileLoop
			reduce(297), // oosvar_name, reduce: WhileLoop
			reduce(297), // @[, reduce: WhileLoop
			reduce(297), // braced_oosvar_name, reduce: WhileLoop
			reduce(297), // full_oosvar, reduce: WhileLoop
			reduce(297), // all, reduce: WhileLoop
			reduce(297), // non_sigil_name, reduce: WhileLoop
			reduce(297), // arr, reduce: WhileLoop
			reduce(297), // bool, reduce: WhileLoop
			reduce(297), // float, reduce: WhileLoop
			reduce(297), // int, reduce: WhileLoop
			reduce(297), // map, reduce: WhileLoop
			reduce(297), // num, reduce: WhileLoop
			reduce(297), // str, reduce: WhileLoop
			reduce(297), // var, reduce: WhileLoop
			reduce(297), // funct, reduce: WhileLoop
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(297), // +, reduce: WhileLoop
			reduce(297), // -, reduce: WhileLoop
			reduce(297), // .+, reduce: WhileLoop
			reduce(297), // .-, reduce: WhileLoop
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(297), // !, reduce: WhileLoop
			reduce(297), // ~, reduce: WhileLoop
			nil,         // **
			reduce(297), // string_literal, reduce: WhileLoop
			reduce(297), // regex_case_insensitive, reduce: WhileLoop
			reduce(297), // int_literal, reduce: WhileLoop
			reduce(297), // float_literal, reduce: WhileLoop
			reduce(297), // boolean_literal, reduce: WhileLoop
			reduce(297), // null_literal, reduce: WhileLoop
			reduce(297), // inf_literal, reduce: WhileLoop
			reduce(297), // nan_literal, reduce: WhileLoop
			reduce(297), // const_M_PI, reduce: WhileLoop
			reduce(297), // const_M_E, reduce: WhileLoop
			reduce(297), // panic, reduce: WhileLoop
			reduce(297), // [, reduce: WhileLoop
			reduce(297), // ctx_IPS, reduce: WhileLoop
			reduce(297), // ctx_IFS, reduce: WhileLoop
			reduce(297), // ctx_IRS, reduce: WhileLoop
			reduce(297), // ctx_OPS, reduce: WhileLoop
			reduce(297), // ctx_OFS, reduce: WhileLoop
			reduce(297), // ctx_ORS, reduce: WhileLoop
			reduce(297), // ctx_FLATSEP, reduce: WhileLoop
			reduce(297), // ctx_NF, reduce: WhileLoop
			reduce(297), // ctx_NR, reduce: WhileLoop
			reduce(297), // ctx_FNR, reduce: WhileLoop
			reduce(297), // ctx_FILENAME, reduce: WhileLoop
			reduce(297), // ctx_FILENUM, reduce: WhileLoop
			reduce(297), // env, reduce: WhileLoop
			nil,         // [[
			nil,         // [[[
			reduce(297), // call, reduce: WhileLoop
			reduce(297), // begin, reduce: WhileLoop
			reduce(297), // end, reduce: WhileLoop
			reduce(297), // if, reduce: WhileLoop
			nil,         // elif
			nil,         // else
			reduce(297), // while, reduce: WhileLoop
			reduce(297), // do, reduce: WhileLoop
			reduce(297), // for, reduce: WhileLoop
			nil,         // in
			reduce(297), // break, reduce: WhileLoop
			reduce(297), // continue, reduce: WhileLoop
			reduce(297), // func, reduce: WhileLoop
			reduce(297), // subr, reduce: WhileLoop
			reduce(297), // return, reduce: WhileLoop
		},
	},
	actionRow{ // S4249
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4479), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4250
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			shift(4480), // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4251
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4481), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4252
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(4482), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4253
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(4483), // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(4167), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4254
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(324), // ;, reduce: UnnamedFunctionDefinition
			reduce(324), // {, reduce: UnnamedFunctionDefinition
			reduce(324), // }, reduce: UnnamedFunctionDefinition
			reduce(324), // =, reduce: UnnamedFunctionDefinition
			nil,         // unset
			nil,         // filter
			reduce(324), // >, reduce: UnnamedFunctionDefinition
			reduce(324), // >>, reduce: UnnamedFunctionDefinition
			reduce(324), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(324), // ||=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^=, reduce: UnnamedFunctionDefinition
			reduce(324), // &&=, reduce: UnnamedFunctionDefinition
			reduce(324), // ??=, reduce: UnnamedFunctionDefinition
			reduce(324), // ???=, reduce: UnnamedFunctionDefinition
			reduce(324), // |=, reduce: UnnamedFunctionDefinition
			reduce(324), // &=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^=, reduce: UnnamedFunctionDefinition
			reduce(324), // <<=, reduce: UnnamedFunctionDefinition
			reduce(324), // >>=, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>=, reduce: UnnamedFunctionDefinition
			reduce(324), // +=, reduce: UnnamedFunctionDefinition
			reduce(324), // .=, reduce: UnnamedFunctionDefinition
			reduce(324), // -=, reduce: UnnamedFunctionDefinition
			reduce(324), // *=, reduce: UnnamedFunctionDefinition
			reduce(324), // /=, reduce: UnnamedFunctionDefinition
			reduce(324), // //=, reduce: UnnamedFunctionDefinition
			reduce(324), // %=, reduce: UnnamedFunctionDefinition
			reduce(324), // **=, reduce: UnnamedFunctionDefinition
			reduce(324), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(324), // ||, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^, reduce: UnnamedFunctionDefinition
			reduce(324), // &&, reduce: UnnamedFunctionDefinition
			reduce(324), // ??, reduce: UnnamedFunctionDefinition
			reduce(324), // ???, reduce: UnnamedFunctionDefinition
			reduce(324), // =~, reduce: UnnamedFunctionDefinition
			reduce(324), // !=~, reduce: UnnamedFunctionDefinition
			reduce(324), // ==, reduce: UnnamedFunctionDefinition
			reduce(324), // !=, reduce: UnnamedFunctionDefinition
			reduce(324), // <=>, reduce: UnnamedFunctionDefinition
			reduce(324), // >=, reduce: UnnamedFunctionDefinition
			reduce(324), // <, reduce: UnnamedFunctionDefinition
			reduce(324), // <=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^, reduce: UnnamedFunctionDefinition
			reduce(324), // &, reduce: UnnamedFunctionDefinition
			reduce(324), // <<, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>, reduce: UnnamedFunctionDefinition
			reduce(324), // +, reduce: UnnamedFunctionDefinition
			reduce(324), // -, reduce: UnnamedFunctionDefinition
			reduce(324), // .+, reduce: UnnamedFunctionDefinition
			reduce(324), // .-, reduce: UnnamedFunctionDefinition
			reduce(324), // *, reduce: UnnamedFunctionDefinition
			reduce(324), // /, reduce: UnnamedFunctionDefinition
			reduce(324), // //, reduce: UnnamedFunctionDefinition
			reduce(324), // %, reduce: UnnamedFunctionDefinition
			reduce(324), // .*, reduce: UnnamedFunctionDefinition
			reduce(324), // ./, reduce: UnnamedFunctionDefinition
			reduce(324), // .//, reduce: UnnamedFunctionDefinition
			reduce(324), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(324), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(324), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(324), // [[, reduce: UnnamedFunctionDefinition
			reduce(324), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4255
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S4256
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4257
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2042), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4487), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4258
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2042), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4259
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(263), // $, reduce: ArraySliceAccess
			nil,         // empty
			reduce(263), // ;, reduce: ArraySliceAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(263), // >, reduce: ArraySliceAccess
			reduce(263), // >>, reduce: ArraySliceAccess
			reduce(263), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(263), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(263), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(263), // ||, reduce: ArraySliceAccess
			reduce(263), // ^^, reduce: ArraySliceAccess
			reduce(263), // &&, reduce: ArraySliceAccess
			reduce(263), // ??, reduce: ArraySliceAccess
			reduce(263), // ???, reduce: ArraySliceAccess
			reduce(263), // =~, reduce: ArraySliceAccess
			reduce(263), // !=~, reduce: ArraySliceAccess
			reduce(263), // ==, reduce: ArraySliceAccess
			reduce(263), // !=, reduce: ArraySliceAccess
			reduce(263), // <=>, reduce: ArraySliceAccess
			reduce(263), // >=, reduce: ArraySliceAccess
			reduce(263), // <, reduce: ArraySliceAccess
			reduce(263), // <=, reduce: ArraySliceAccess
			reduce(263), // ^, reduce: ArraySliceAccess
			reduce(263), // &, reduce: ArraySliceAccess
			reduce(263), // <<, reduce: ArraySliceAccess
			reduce(263), // >>>, reduce: ArraySliceAccess
			reduce(263), // +, reduce: ArraySliceAccess
			reduce(263), // -, reduce: ArraySliceAccess
			reduce(263), // .+, reduce: ArraySliceAccess
			reduce(263), // .-, reduce: ArraySliceAccess
			reduce(263), // *, reduce: ArraySliceAccess
			reduce(263), // /, reduce: ArraySliceAccess
			reduce(263), // //, reduce: ArraySliceAccess
			reduce(263), // %, reduce: ArraySliceAccess
			reduce(263), // .*, reduce: ArraySliceAccess
			reduce(263), // ./, reduce: ArraySliceAccess
			reduce(263), // .//, reduce: ArraySliceAccess
			reduce(263), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(263), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(263), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(263), // [[, reduce: ArraySliceAccess
			reduce(263), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4260
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(262), // $, reduce: ArrayOrMapPositionalValueAccess
			nil,         // empty
			reduce(262), // ;, reduce: ArrayOrMapPositionalValueAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(262), // >, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |, reduce: ArrayOrMapPositionalValueAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(262), // ,, reduce: ArrayOrMapPositionalValueAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(262), // ?, reduce: ArrayOrMapPositionalValueAccess
			nil,         // :
			reduce(262), // ||, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // =~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ==, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .+, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .-, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .*, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ./, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .//, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ., reduce: ArrayOrMapPositionalValueAccess
			nil,         // !
			nil,         // ~
			reduce(262), // **, reduce: ArrayOrMapPositionalValueAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(262), // [, reduce: ArrayOrMapPositionalValueAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(262), // [[, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // [[[, reduce: ArrayOrMapPositionalValueAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4261
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4489), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4262
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3941), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4263
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(263), // $, reduce: ArraySliceAccess
			nil,         // empty
			reduce(263), // ;, reduce: ArraySliceAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(263), // >, reduce: ArraySliceAccess
			reduce(263), // >>, reduce: ArraySliceAccess
			reduce(263), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(263), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(263), // ||, reduce: ArraySliceAccess
			reduce(263), // ^^, reduce: ArraySliceAccess
			reduce(263), // &&, reduce: ArraySliceAccess
			reduce(263), // ??, reduce: ArraySliceAccess
			reduce(263), // ???, reduce: ArraySliceAccess
			reduce(263), // =~, reduce: ArraySliceAccess
			reduce(263), // !=~, reduce: ArraySliceAccess
			reduce(263), // ==, reduce: ArraySliceAccess
			reduce(263), // !=, reduce: ArraySliceAccess
			reduce(263), // <=>, reduce: ArraySliceAccess
			reduce(263), // >=, reduce: ArraySliceAccess
			reduce(263), // <, reduce: ArraySliceAccess
			reduce(263), // <=, reduce: ArraySliceAccess
			reduce(263), // ^, reduce: ArraySliceAccess
			reduce(263), // &, reduce: ArraySliceAccess
			reduce(263), // <<, reduce: ArraySliceAccess
			reduce(263), // >>>, reduce: ArraySliceAccess
			reduce(263), // +, reduce: ArraySliceAccess
			reduce(263), // -, reduce: ArraySliceAccess
			reduce(263), // .+, reduce: ArraySliceAccess
			reduce(263), // .-, reduce: ArraySliceAccess
			reduce(263), // *, reduce: ArraySliceAccess
			reduce(263), // /, reduce: ArraySliceAccess
			reduce(263), // //, reduce: ArraySliceAccess
			reduce(263), // %, reduce: ArraySliceAccess
			reduce(263), // .*, reduce: ArraySliceAccess
			reduce(263), // ./, reduce: ArraySliceAccess
			reduce(263), // .//, reduce: ArraySliceAccess
			reduce(263), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(263), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(263), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(263), // [[, reduce: ArraySliceAccess
			reduce(263), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4264
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(262), // $, reduce: ArrayOrMapPositionalValueAccess
			nil,         // empty
			reduce(262), // ;, reduce: ArrayOrMapPositionalValueAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(262), // >, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |, reduce: ArrayOrMapPositionalValueAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(262), // ?, reduce: ArrayOrMapPositionalValueAccess
			nil,         // :
			reduce(262), // ||, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // =~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ==, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .+, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .-, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .*, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ./, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .//, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ., reduce: ArrayOrMapPositionalValueAccess
			nil,         // !
			nil,         // ~
			reduce(262), // **, reduce: ArrayOrMapPositionalValueAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(262), // [, reduce: ArrayOrMapPositionalValueAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(262), // [[, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // [[[, reduce: ArrayOrMapPositionalValueAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4265
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4491), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4266
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3951), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4267
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(108), // >, reduce: PositionalFieldValue
			reduce(108), // >>, reduce: PositionalFieldValue
			reduce(108), // |, reduce: PositionalFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(108), // ,, reduce: PositionalFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(108), // ?, reduce: PositionalFieldValue
			nil,         // :
			reduce(108), // ||, reduce: PositionalFieldValue
			reduce(108), // ^^, reduce: PositionalFieldValue
			reduce(108), // &&, reduce: PositionalFieldValue
			reduce(108), // ??, reduce: PositionalFieldValue
			reduce(108), // ???, reduce: PositionalFieldValue
			reduce(108), // =~, reduce: PositionalFieldValue
			reduce(108), // !=~, reduce: PositionalFieldValue
			reduce(108), // ==, reduce: PositionalFieldValue
			reduce(108), // !=, reduce: PositionalFieldValue
			reduce(108), // <=>, reduce: PositionalFieldValue
			reduce(108), // >=, reduce: PositionalFieldValue
			reduce(108), // <, reduce: PositionalFieldValue
			reduce(108), // <=, reduce: PositionalFieldValue
			reduce(108), // ^, reduce: PositionalFieldValue
			reduce(108), // &, reduce: PositionalFieldValue
			reduce(108), // <<, reduce: PositionalFieldValue
			reduce(108), // >>>, reduce: PositionalFieldValue
			reduce(108), // +, reduce: PositionalFieldValue
			reduce(108), // -, reduce: PositionalFieldValue
			reduce(108), // .+, reduce: PositionalFieldValue
			reduce(108), // .-, reduce: PositionalFieldValue
			reduce(108), // *, reduce: PositionalFieldValue
			reduce(108), // /, reduce: PositionalFieldValue
			reduce(108), // //, reduce: PositionalFieldValue
			reduce(108), // %, reduce: PositionalFieldValue
			reduce(108), // .*, reduce: PositionalFieldValue
			reduce(108), // ./, reduce: PositionalFieldValue
			reduce(108), // .//, reduce: PositionalFieldValue
			reduce(108), // ., reduce: PositionalFieldValue
			nil,         // !
			nil,         // ~
			reduce(108), // **, reduce: PositionalFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(108), // [, reduce: PositionalFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(108), // [[, reduce: PositionalFieldValue
			reduce(108), // [[[, reduce: PositionalFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4268
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(149), // ,, reduce: TernaryTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4269
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4493), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4270
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(265), // >, reduce: ArraySliceAccess
			reduce(265), // >>, reduce: ArraySliceAccess
			reduce(265), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(265), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(265), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(265), // ||, reduce: ArraySliceAccess
			reduce(265), // ^^, reduce: ArraySliceAccess
			reduce(265), // &&, reduce: ArraySliceAccess
			reduce(265), // ??, reduce: ArraySliceAccess
			reduce(265), // ???, reduce: ArraySliceAccess
			reduce(265), // =~, reduce: ArraySliceAccess
			reduce(265), // !=~, reduce: ArraySliceAccess
			reduce(265), // ==, reduce: ArraySliceAccess
			reduce(265), // !=, reduce: ArraySliceAccess
			reduce(265), // <=>, reduce: ArraySliceAccess
			reduce(265), // >=, reduce: ArraySliceAccess
			reduce(265), // <, reduce: ArraySliceAccess
			reduce(265), // <=, reduce: ArraySliceAccess
			reduce(265), // ^, reduce: ArraySliceAccess
			reduce(265), // &, reduce: ArraySliceAccess
			reduce(265), // <<, reduce: ArraySliceAccess
			reduce(265), // >>>, reduce: ArraySliceAccess
			reduce(265), // +, reduce: ArraySliceAccess
			reduce(265), // -, reduce: ArraySliceAccess
			reduce(265), // .+, reduce: ArraySliceAccess
			reduce(265), // .-, reduce: ArraySliceAccess
			reduce(265), // *, reduce: ArraySliceAccess
			reduce(265), // /, reduce: ArraySliceAccess
			reduce(265), // //, reduce: ArraySliceAccess
			reduce(265), // %, reduce: ArraySliceAccess
			reduce(265), // .*, reduce: ArraySliceAccess
			reduce(265), // ./, reduce: ArraySliceAccess
			reduce(265), // .//, reduce: ArraySliceAccess
			reduce(265), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(265), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(265), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(265), // [[, reduce: ArraySliceAccess
			reduce(265), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4271
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(264), // >, reduce: ArraySliceAccess
			reduce(264), // >>, reduce: ArraySliceAccess
			reduce(264), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(264), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(264), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(264), // ||, reduce: ArraySliceAccess
			reduce(264), // ^^, reduce: ArraySliceAccess
			reduce(264), // &&, reduce: ArraySliceAccess
			reduce(264), // ??, reduce: ArraySliceAccess
			reduce(264), // ???, reduce: ArraySliceAccess
			reduce(264), // =~, reduce: ArraySliceAccess
			reduce(264), // !=~, reduce: ArraySliceAccess
			reduce(264), // ==, reduce: ArraySliceAccess
			reduce(264), // !=, reduce: ArraySliceAccess
			reduce(264), // <=>, reduce: ArraySliceAccess
			reduce(264), // >=, reduce: ArraySliceAccess
			reduce(264), // <, reduce: ArraySliceAccess
			reduce(264), // <=, reduce: ArraySliceAccess
			reduce(264), // ^, reduce: ArraySliceAccess
			reduce(264), // &, reduce: ArraySliceAccess
			reduce(264), // <<, reduce: ArraySliceAccess
			reduce(264), // >>>, reduce: ArraySliceAccess
			reduce(264), // +, reduce: ArraySliceAccess
			reduce(264), // -, reduce: ArraySliceAccess
			reduce(264), // .+, reduce: ArraySliceAccess
			reduce(264), // .-, reduce: ArraySliceAccess
			reduce(264), // *, reduce: ArraySliceAccess
			reduce(264), // /, reduce: ArraySliceAccess
			reduce(264), // //, reduce: ArraySliceAccess
			reduce(264), // %, reduce: ArraySliceAccess
			reduce(264), // .*, reduce: ArraySliceAccess
			reduce(264), // ./, reduce: ArraySliceAccess
			reduce(264), // .//, reduce: ArraySliceAccess
			reduce(264), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(264), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(264), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(264), // [[, reduce: ArraySliceAccess
			reduce(264), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4272
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(261), // >, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |, reduce: ArrayOrMapPositionalNameAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(261), // ,, reduce: ArrayOrMapPositionalNameAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(261), // ?, reduce: ArrayOrMapPositionalNameAccess
			nil,         // :
			reduce(261), // ||, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // =~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ==, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .+, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .-, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .*, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ./, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .//, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ., reduce: ArrayOrMapPositionalNameAccess
			nil,         // !
			nil,         // ~
			reduce(261), // **, reduce: ArrayOrMapPositionalNameAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(261), // [, reduce: ArrayOrMapPositionalNameAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(261), // [[, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // [[[, reduce: ArrayOrMapPositionalNameAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4273
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4494), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4274
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(324), // >, reduce: UnnamedFunctionDefinition
			reduce(324), // >>, reduce: UnnamedFunctionDefinition
			reduce(324), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(324), // ,, reduce: UnnamedFunctionDefinition
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(324), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(324), // ||, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^, reduce: UnnamedFunctionDefinition
			reduce(324), // &&, reduce: UnnamedFunctionDefinition
			reduce(324), // ??, reduce: UnnamedFunctionDefinition
			reduce(324), // ???, reduce: UnnamedFunctionDefinition
			reduce(324), // =~, reduce: UnnamedFunctionDefinition
			reduce(324), // !=~, reduce: UnnamedFunctionDefinition
			reduce(324), // ==, reduce: UnnamedFunctionDefinition
			reduce(324), // !=, reduce: UnnamedFunctionDefinition
			reduce(324), // <=>, reduce: UnnamedFunctionDefinition
			reduce(324), // >=, reduce: UnnamedFunctionDefinition
			reduce(324), // <, reduce: UnnamedFunctionDefinition
			reduce(324), // <=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^, reduce: UnnamedFunctionDefinition
			reduce(324), // &, reduce: UnnamedFunctionDefinition
			reduce(324), // <<, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>, reduce: UnnamedFunctionDefinition
			reduce(324), // +, reduce: UnnamedFunctionDefinition
			reduce(324), // -, reduce: UnnamedFunctionDefinition
			reduce(324), // .+, reduce: UnnamedFunctionDefinition
			reduce(324), // .-, reduce: UnnamedFunctionDefinition
			reduce(324), // *, reduce: UnnamedFunctionDefinition
			reduce(324), // /, reduce: UnnamedFunctionDefinition
			reduce(324), // //, reduce: UnnamedFunctionDefinition
			reduce(324), // %, reduce: UnnamedFunctionDefinition
			reduce(324), // .*, reduce: UnnamedFunctionDefinition
			reduce(324), // ./, reduce: UnnamedFunctionDefinition
			reduce(324), // .//, reduce: UnnamedFunctionDefinition
			reduce(324), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(324), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(324), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(324), // [[, reduce: UnnamedFunctionDefinition
			reduce(324), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4275
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S4276
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4277
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4278
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(263), // >, reduce: ArraySliceAccess
			reduce(263), // >>, reduce: ArraySliceAccess
			reduce(263), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(263), // ), reduce: ArraySliceAccess
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(263), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(263), // ||, reduce: ArraySliceAccess
			reduce(263), // ^^, reduce: ArraySliceAccess
			reduce(263), // &&, reduce: ArraySliceAccess
			reduce(263), // ??, reduce: ArraySliceAccess
			reduce(263), // ???, reduce: ArraySliceAccess
			reduce(263), // =~, reduce: ArraySliceAccess
			reduce(263), // !=~, reduce: ArraySliceAccess
			reduce(263), // ==, reduce: ArraySliceAccess
			reduce(263), // !=, reduce: ArraySliceAccess
			reduce(263), // <=>, reduce: ArraySliceAccess
			reduce(263), // >=, reduce: ArraySliceAccess
			reduce(263), // <, reduce: ArraySliceAccess
			reduce(263), // <=, reduce: ArraySliceAccess
			reduce(263), // ^, reduce: ArraySliceAccess
			reduce(263), // &, reduce: ArraySliceAccess
			reduce(263), // <<, reduce: ArraySliceAccess
			reduce(263), // >>>, reduce: ArraySliceAccess
			reduce(263), // +, reduce: ArraySliceAccess
			reduce(263), // -, reduce: ArraySliceAccess
			reduce(263), // .+, reduce: ArraySliceAccess
			reduce(263), // .-, reduce: ArraySliceAccess
			reduce(263), // *, reduce: ArraySliceAccess
			reduce(263), // /, reduce: ArraySliceAccess
			reduce(263), // //, reduce: ArraySliceAccess
			reduce(263), // %, reduce: ArraySliceAccess
			reduce(263), // .*, reduce: ArraySliceAccess
			reduce(263), // ./, reduce: ArraySliceAccess
			reduce(263), // .//, reduce: ArraySliceAccess
			reduce(263), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(263), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(263), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(263), // [[, reduce: ArraySliceAccess
			reduce(263), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4279
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(262), // >, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |, reduce: ArrayOrMapPositionalValueAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(262), // ), reduce: ArrayOrMapPositionalValueAccess
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(262), // ?, reduce: ArrayOrMapPositionalValueAccess
			nil,         // :
			reduce(262), // ||, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // =~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ==, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .+, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .-, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .*, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ./, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .//, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ., reduce: ArrayOrMapPositionalValueAccess
			nil,         // !
			nil,         // ~
			reduce(262), // **, reduce: ArrayOrMapPositionalValueAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(262), // [, reduce: ArrayOrMapPositionalValueAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(262), // [[, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // [[[, reduce: ArrayOrMapPositionalValueAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4280
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4498), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4281
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3979), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4282
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(290), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(295), // (
			nil,        // )
			nil,        // emitp
			shift(307), // field_name
			shift(308), // $[
			nil,        // ]
			shift(309), // braced_field_name
			shift(310), // $[[
			shift(311), // $[[[
			shift(312), // full_srec
			shift(314), // oosvar_name
			shift(315), // @[
			shift(316), // braced_oosvar_name
			shift(317), // full_oosvar
			shift(318), // all
			shift(319), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(334), // +
			shift(336), // -
			shift(337), // .+
			shift(338), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(341), // !
			shift(342), // ~
			nil,        // **
			shift(347), // string_literal
			shift(348), // regex_case_insensitive
			shift(349), // int_literal
			shift(350), // float_literal
			shift(351), // boolean_literal
			shift(352), // null_literal
			shift(353), // inf_literal
			shift(354), // nan_literal
			shift(355), // const_M_PI
			shift(356), // const_M_E
			shift(357), // panic
			shift(359), // [
			shift(361), // ctx_IPS
			shift(362), // ctx_IFS
			shift(363), // ctx_IRS
			shift(364), // ctx_OPS
			shift(365), // ctx_OFS
			shift(366), // ctx_ORS
			shift(367), // ctx_FLATSEP
			shift(368), // ctx_NF
			shift(369), // ctx_NR
			shift(370), // ctx_FNR
			shift(371), // ctx_FILENAME
			shift(372), // ctx_FILENUM
			shift(374), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(381), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4283
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(263), // >, reduce: ArraySliceAccess
			reduce(263), // >>, reduce: ArraySliceAccess
			reduce(263), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(263), // ], reduce: ArraySliceAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(263), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(263), // ||, reduce: ArraySliceAccess
			reduce(263), // ^^, reduce: ArraySliceAccess
			reduce(263), // &&, reduce: ArraySliceAccess
			reduce(263), // ??, reduce: ArraySliceAccess
			reduce(263), // ???, reduce: ArraySliceAccess
			reduce(263), // =~, reduce: ArraySliceAccess
			reduce(263), // !=~, reduce: ArraySliceAccess
			reduce(263), // ==, reduce: ArraySliceAccess
			reduce(263), // !=, reduce: ArraySliceAccess
			reduce(263), // <=>, reduce: ArraySliceAccess
			reduce(263), // >=, reduce: ArraySliceAccess
			reduce(263), // <, reduce: ArraySliceAccess
			reduce(263), // <=, reduce: ArraySliceAccess
			reduce(263), // ^, reduce: ArraySliceAccess
			reduce(263), // &, reduce: ArraySliceAccess
			reduce(263), // <<, reduce: ArraySliceAccess
			reduce(263), // >>>, reduce: ArraySliceAccess
			reduce(263), // +, reduce: ArraySliceAccess
			reduce(263), // -, reduce: ArraySliceAccess
			reduce(263), // .+, reduce: ArraySliceAccess
			reduce(263), // .-, reduce: ArraySliceAccess
			reduce(263), // *, reduce: ArraySliceAccess
			reduce(263), // /, reduce: ArraySliceAccess
			reduce(263), // //, reduce: ArraySliceAccess
			reduce(263), // %, reduce: ArraySliceAccess
			reduce(263), // .*, reduce: ArraySliceAccess
			reduce(263), // ./, reduce: ArraySliceAccess
			reduce(263), // .//, reduce: ArraySliceAccess
			reduce(263), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(263), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(263), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(263), // [[, reduce: ArraySliceAccess
			reduce(263), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4284
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(262), // >, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |, reduce: ArrayOrMapPositionalValueAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(262), // ], reduce: ArrayOrMapPositionalValueAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(262), // ?, reduce: ArrayOrMapPositionalValueAccess
			nil,         // :
			reduce(262), // ||, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // =~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ==, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .+, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .-, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .*, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ./, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .//, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ., reduce: ArrayOrMapPositionalValueAccess
			nil,         // !
			nil,         // ~
			reduce(262), // **, reduce: ArrayOrMapPositionalValueAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(262), // [, reduce: ArrayOrMapPositionalValueAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(262), // [[, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // [[[, reduce: ArrayOrMapPositionalValueAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4285
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4501), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4286
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3992), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4287
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(108), // >, reduce: PositionalFieldValue
			reduce(108), // >>, reduce: PositionalFieldValue
			reduce(108), // |, reduce: PositionalFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(108), // ], reduce: PositionalFieldValue
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(108), // ?, reduce: PositionalFieldValue
			reduce(108), // :, reduce: PositionalFieldValue
			reduce(108), // ||, reduce: PositionalFieldValue
			reduce(108), // ^^, reduce: PositionalFieldValue
			reduce(108), // &&, reduce: PositionalFieldValue
			reduce(108), // ??, reduce: PositionalFieldValue
			reduce(108), // ???, reduce: PositionalFieldValue
			reduce(108), // =~, reduce: PositionalFieldValue
			reduce(108), // !=~, reduce: PositionalFieldValue
			reduce(108), // ==, reduce: PositionalFieldValue
			reduce(108), // !=, reduce: PositionalFieldValue
			reduce(108), // <=>, reduce: PositionalFieldValue
			reduce(108), // >=, reduce: PositionalFieldValue
			reduce(108), // <, reduce: PositionalFieldValue
			reduce(108), // <=, reduce: PositionalFieldValue
			reduce(108), // ^, reduce: PositionalFieldValue
			reduce(108), // &, reduce: PositionalFieldValue
			reduce(108), // <<, reduce: PositionalFieldValue
			reduce(108), // >>>, reduce: PositionalFieldValue
			reduce(108), // +, reduce: PositionalFieldValue
			reduce(108), // -, reduce: PositionalFieldValue
			reduce(108), // .+, reduce: PositionalFieldValue
			reduce(108), // .-, reduce: PositionalFieldValue
			reduce(108), // *, reduce: PositionalFieldValue
			reduce(108), // /, reduce: PositionalFieldValue
			reduce(108), // //, reduce: PositionalFieldValue
			reduce(108), // %, reduce: PositionalFieldValue
			reduce(108), // .*, reduce: PositionalFieldValue
			reduce(108), // ./, reduce: PositionalFieldValue
			reduce(108), // .//, reduce: PositionalFieldValue
			reduce(108), // ., reduce: PositionalFieldValue
			nil,         // !
			nil,         // ~
			reduce(108), // **, reduce: PositionalFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(108), // [, reduce: PositionalFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(108), // [[, reduce: PositionalFieldValue
			reduce(108), // [[[, reduce: PositionalFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4288
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(149), // ], reduce: TernaryTerm
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			reduce(149), // :, reduce: TernaryTerm
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4289
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4503), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4290
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(265), // >, reduce: ArraySliceAccess
			reduce(265), // >>, reduce: ArraySliceAccess
			reduce(265), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(265), // ], reduce: ArraySliceAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(265), // ?, reduce: ArraySliceAccess
			reduce(265), // :, reduce: ArraySliceAccess
			reduce(265), // ||, reduce: ArraySliceAccess
			reduce(265), // ^^, reduce: ArraySliceAccess
			reduce(265), // &&, reduce: ArraySliceAccess
			reduce(265), // ??, reduce: ArraySliceAccess
			reduce(265), // ???, reduce: ArraySliceAccess
			reduce(265), // =~, reduce: ArraySliceAccess
			reduce(265), // !=~, reduce: ArraySliceAccess
			reduce(265), // ==, reduce: ArraySliceAccess
			reduce(265), // !=, reduce: ArraySliceAccess
			reduce(265), // <=>, reduce: ArraySliceAccess
			reduce(265), // >=, reduce: ArraySliceAccess
			reduce(265), // <, reduce: ArraySliceAccess
			reduce(265), // <=, reduce: ArraySliceAccess
			reduce(265), // ^, reduce: ArraySliceAccess
			reduce(265), // &, reduce: ArraySliceAccess
			reduce(265), // <<, reduce: ArraySliceAccess
			reduce(265), // >>>, reduce: ArraySliceAccess
			reduce(265), // +, reduce: ArraySliceAccess
			reduce(265), // -, reduce: ArraySliceAccess
			reduce(265), // .+, reduce: ArraySliceAccess
			reduce(265), // .-, reduce: ArraySliceAccess
			reduce(265), // *, reduce: ArraySliceAccess
			reduce(265), // /, reduce: ArraySliceAccess
			reduce(265), // //, reduce: ArraySliceAccess
			reduce(265), // %, reduce: ArraySliceAccess
			reduce(265), // .*, reduce: ArraySliceAccess
			reduce(265), // ./, reduce: ArraySliceAccess
			reduce(265), // .//, reduce: ArraySliceAccess
			reduce(265), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(265), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(265), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(265), // [[, reduce: ArraySliceAccess
			reduce(265), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4291
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(264), // >, reduce: ArraySliceAccess
			reduce(264), // >>, reduce: ArraySliceAccess
			reduce(264), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(264), // ], reduce: ArraySliceAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(264), // ?, reduce: ArraySliceAccess
			reduce(264), // :, reduce: ArraySliceAccess
			reduce(264), // ||, reduce: ArraySliceAccess
			reduce(264), // ^^, reduce: ArraySliceAccess
			reduce(264), // &&, reduce: ArraySliceAccess
			reduce(264), // ??, reduce: ArraySliceAccess
			reduce(264), // ???, reduce: ArraySliceAccess
			reduce(264), // =~, reduce: ArraySliceAccess
			reduce(264), // !=~, reduce: ArraySliceAccess
			reduce(264), // ==, reduce: ArraySliceAccess
			reduce(264), // !=, reduce: ArraySliceAccess
			reduce(264), // <=>, reduce: ArraySliceAccess
			reduce(264), // >=, reduce: ArraySliceAccess
			reduce(264), // <, reduce: ArraySliceAccess
			reduce(264), // <=, reduce: ArraySliceAccess
			reduce(264), // ^, reduce: ArraySliceAccess
			reduce(264), // &, reduce: ArraySliceAccess
			reduce(264), // <<, reduce: ArraySliceAccess
			reduce(264), // >>>, reduce: ArraySliceAccess
			reduce(264), // +, reduce: ArraySliceAccess
			reduce(264), // -, reduce: ArraySliceAccess
			reduce(264), // .+, reduce: ArraySliceAccess
			reduce(264), // .-, reduce: ArraySliceAccess
			reduce(264), // *, reduce: ArraySliceAccess
			reduce(264), // /, reduce: ArraySliceAccess
			reduce(264), // //, reduce: ArraySliceAccess
			reduce(264), // %, reduce: ArraySliceAccess
			reduce(264), // .*, reduce: ArraySliceAccess
			reduce(264), // ./, reduce: ArraySliceAccess
			reduce(264), // .//, reduce: ArraySliceAccess
			reduce(264), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(264), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(264), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(264), // [[, reduce: ArraySliceAccess
			reduce(264), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4292
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(261), // >, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |, reduce: ArrayOrMapPositionalNameAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(261), // ], reduce: ArrayOrMapPositionalNameAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(261), // ?, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // :, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ||, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // =~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ==, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .+, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .-, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .*, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ./, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .//, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ., reduce: ArrayOrMapPositionalNameAccess
			nil,         // !
			nil,         // ~
			reduce(261), // **, reduce: ArrayOrMapPositionalNameAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(261), // [, reduce: ArrayOrMapPositionalNameAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(261), // [[, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // [[[, reduce: ArrayOrMapPositionalNameAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4293
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4504), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4294
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(324), // >, reduce: UnnamedFunctionDefinition
			reduce(324), // >>, reduce: UnnamedFunctionDefinition
			reduce(324), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(324), // ], reduce: UnnamedFunctionDefinition
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(324), // ?, reduce: UnnamedFunctionDefinition
			reduce(324), // :, reduce: UnnamedFunctionDefinition
			reduce(324), // ||, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^, reduce: UnnamedFunctionDefinition
			reduce(324), // &&, reduce: UnnamedFunctionDefinition
			reduce(324), // ??, reduce: UnnamedFunctionDefinition
			reduce(324), // ???, reduce: UnnamedFunctionDefinition
			reduce(324), // =~, reduce: UnnamedFunctionDefinition
			reduce(324), // !=~, reduce: UnnamedFunctionDefinition
			reduce(324), // ==, reduce: UnnamedFunctionDefinition
			reduce(324), // !=, reduce: UnnamedFunctionDefinition
			reduce(324), // <=>, reduce: UnnamedFunctionDefinition
			reduce(324), // >=, reduce: UnnamedFunctionDefinition
			reduce(324), // <, reduce: UnnamedFunctionDefinition
			reduce(324), // <=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^, reduce: UnnamedFunctionDefinition
			reduce(324), // &, reduce: UnnamedFunctionDefinition
			reduce(324), // <<, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>, reduce: UnnamedFunctionDefinition
			reduce(324), // +, reduce: UnnamedFunctionDefinition
			reduce(324), // -, reduce: UnnamedFunctionDefinition
			reduce(324), // .+, reduce: UnnamedFunctionDefinition
			reduce(324), // .-, reduce: UnnamedFunctionDefinition
			reduce(324), // *, reduce: UnnamedFunctionDefinition
			reduce(324), // /, reduce: UnnamedFunctionDefinition
			reduce(324), // //, reduce: UnnamedFunctionDefinition
			reduce(324), // %, reduce: UnnamedFunctionDefinition
			reduce(324), // .*, reduce: UnnamedFunctionDefinition
			reduce(324), // ./, reduce: UnnamedFunctionDefinition
			reduce(324), // .//, reduce: UnnamedFunctionDefinition
			reduce(324), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(324), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(324), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(324), // [[, reduce: UnnamedFunctionDefinition
			reduce(324), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4295
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S4296
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4297
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(263), // >, reduce: ArraySliceAccess
			reduce(263), // >>, reduce: ArraySliceAccess
			reduce(263), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(263), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(263), // ], reduce: ArraySliceAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(263), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(263), // ||, reduce: ArraySliceAccess
			reduce(263), // ^^, reduce: ArraySliceAccess
			reduce(263), // &&, reduce: ArraySliceAccess
			reduce(263), // ??, reduce: ArraySliceAccess
			reduce(263), // ???, reduce: ArraySliceAccess
			reduce(263), // =~, reduce: ArraySliceAccess
			reduce(263), // !=~, reduce: ArraySliceAccess
			reduce(263), // ==, reduce: ArraySliceAccess
			reduce(263), // !=, reduce: ArraySliceAccess
			reduce(263), // <=>, reduce: ArraySliceAccess
			reduce(263), // >=, reduce: ArraySliceAccess
			reduce(263), // <, reduce: ArraySliceAccess
			reduce(263), // <=, reduce: ArraySliceAccess
			reduce(263), // ^, reduce: ArraySliceAccess
			reduce(263), // &, reduce: ArraySliceAccess
			reduce(263), // <<, reduce: ArraySliceAccess
			reduce(263), // >>>, reduce: ArraySliceAccess
			reduce(263), // +, reduce: ArraySliceAccess
			reduce(263), // -, reduce: ArraySliceAccess
			reduce(263), // .+, reduce: ArraySliceAccess
			reduce(263), // .-, reduce: ArraySliceAccess
			reduce(263), // *, reduce: ArraySliceAccess
			reduce(263), // /, reduce: ArraySliceAccess
			reduce(263), // //, reduce: ArraySliceAccess
			reduce(263), // %, reduce: ArraySliceAccess
			reduce(263), // .*, reduce: ArraySliceAccess
			reduce(263), // ./, reduce: ArraySliceAccess
			reduce(263), // .//, reduce: ArraySliceAccess
			reduce(263), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(263), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(263), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(263), // [[, reduce: ArraySliceAccess
			reduce(263), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4298
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(262), // >, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |, reduce: ArrayOrMapPositionalValueAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(262), // ,, reduce: ArrayOrMapPositionalValueAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(262), // ], reduce: ArrayOrMapPositionalValueAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(262), // ?, reduce: ArrayOrMapPositionalValueAccess
			nil,         // :
			reduce(262), // ||, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // =~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ==, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .+, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .-, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .*, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ./, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .//, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ., reduce: ArrayOrMapPositionalValueAccess
			nil,         // !
			nil,         // ~
			reduce(262), // **, reduce: ArrayOrMapPositionalValueAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(262), // [, reduce: ArrayOrMapPositionalValueAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(262), // [[, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // [[[, reduce: ArrayOrMapPositionalValueAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4299
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4507), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4300
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4018), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4301
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(108), // >, reduce: PositionalFieldValue
			reduce(108), // >>, reduce: PositionalFieldValue
			reduce(108), // |, reduce: PositionalFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(108), // ,, reduce: PositionalFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(108), // ), reduce: PositionalFieldValue
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(108), // ?, reduce: PositionalFieldValue
			nil,         // :
			reduce(108), // ||, reduce: PositionalFieldValue
			reduce(108), // ^^, reduce: PositionalFieldValue
			reduce(108), // &&, reduce: PositionalFieldValue
			reduce(108), // ??, reduce: PositionalFieldValue
			reduce(108), // ???, reduce: PositionalFieldValue
			reduce(108), // =~, reduce: PositionalFieldValue
			reduce(108), // !=~, reduce: PositionalFieldValue
			reduce(108), // ==, reduce: PositionalFieldValue
			reduce(108), // !=, reduce: PositionalFieldValue
			reduce(108), // <=>, reduce: PositionalFieldValue
			reduce(108), // >=, reduce: PositionalFieldValue
			reduce(108), // <, reduce: PositionalFieldValue
			reduce(108), // <=, reduce: PositionalFieldValue
			reduce(108), // ^, reduce: PositionalFieldValue
			reduce(108), // &, reduce: PositionalFieldValue
			reduce(108), // <<, reduce: PositionalFieldValue
			reduce(108), // >>>, reduce: PositionalFieldValue
			reduce(108), // +, reduce: PositionalFieldValue
			reduce(108), // -, reduce: PositionalFieldValue
			reduce(108), // .+, reduce: PositionalFieldValue
			reduce(108), // .-, reduce: PositionalFieldValue
			reduce(108), // *, reduce: PositionalFieldValue
			reduce(108), // /, reduce: PositionalFieldValue
			reduce(108), // //, reduce: PositionalFieldValue
			reduce(108), // %, reduce: PositionalFieldValue
			reduce(108), // .*, reduce: PositionalFieldValue
			reduce(108), // ./, reduce: PositionalFieldValue
			reduce(108), // .//, reduce: PositionalFieldValue
			reduce(108), // ., reduce: PositionalFieldValue
			nil,         // !
			nil,         // ~
			reduce(108), // **, reduce: PositionalFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(108), // [, reduce: PositionalFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(108), // [[, reduce: PositionalFieldValue
			reduce(108), // [[[, reduce: PositionalFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4302
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(149), // ,, reduce: TernaryTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(149), // ), reduce: TernaryTerm
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4303
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4509), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4304
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(265), // >, reduce: ArraySliceAccess
			reduce(265), // >>, reduce: ArraySliceAccess
			reduce(265), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(265), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(265), // ), reduce: ArraySliceAccess
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(265), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(265), // ||, reduce: ArraySliceAccess
			reduce(265), // ^^, reduce: ArraySliceAccess
			reduce(265), // &&, reduce: ArraySliceAccess
			reduce(265), // ??, reduce: ArraySliceAccess
			reduce(265), // ???, reduce: ArraySliceAccess
			reduce(265), // =~, reduce: ArraySliceAccess
			reduce(265), // !=~, reduce: ArraySliceAccess
			reduce(265), // ==, reduce: ArraySliceAccess
			reduce(265), // !=, reduce: ArraySliceAccess
			reduce(265), // <=>, reduce: ArraySliceAccess
			reduce(265), // >=, reduce: ArraySliceAccess
			reduce(265), // <, reduce: ArraySliceAccess
			reduce(265), // <=, reduce: ArraySliceAccess
			reduce(265), // ^, reduce: ArraySliceAccess
			reduce(265), // &, reduce: ArraySliceAccess
			reduce(265), // <<, reduce: ArraySliceAccess
			reduce(265), // >>>, reduce: ArraySliceAccess
			reduce(265), // +, reduce: ArraySliceAccess
			reduce(265), // -, reduce: ArraySliceAccess
			reduce(265), // .+, reduce: ArraySliceAccess
			reduce(265), // .-, reduce: ArraySliceAccess
			reduce(265), // *, reduce: ArraySliceAccess
			reduce(265), // /, reduce: ArraySliceAccess
			reduce(265), // //, reduce: ArraySliceAccess
			reduce(265), // %, reduce: ArraySliceAccess
			reduce(265), // .*, reduce: ArraySliceAccess
			reduce(265), // ./, reduce: ArraySliceAccess
			reduce(265), // .//, reduce: ArraySliceAccess
			reduce(265), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(265), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(265), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(265), // [[, reduce: ArraySliceAccess
			reduce(265), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4305
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(264), // >, reduce: ArraySliceAccess
			reduce(264), // >>, reduce: ArraySliceAccess
			reduce(264), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(264), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(264), // ), reduce: ArraySliceAccess
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(264), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(264), // ||, reduce: ArraySliceAccess
			reduce(264), // ^^, reduce: ArraySliceAccess
			reduce(264), // &&, reduce: ArraySliceAccess
			reduce(264), // ??, reduce: ArraySliceAccess
			reduce(264), // ???, reduce: ArraySliceAccess
			reduce(264), // =~, reduce: ArraySliceAccess
			reduce(264), // !=~, reduce: ArraySliceAccess
			reduce(264), // ==, reduce: ArraySliceAccess
			reduce(264), // !=, reduce: ArraySliceAccess
			reduce(264), // <=>, reduce: ArraySliceAccess
			reduce(264), // >=, reduce: ArraySliceAccess
			reduce(264), // <, reduce: ArraySliceAccess
			reduce(264), // <=, reduce: ArraySliceAccess
			reduce(264), // ^, reduce: ArraySliceAccess
			reduce(264), // &, reduce: ArraySliceAccess
			reduce(264), // <<, reduce: ArraySliceAccess
			reduce(264), // >>>, reduce: ArraySliceAccess
			reduce(264), // +, reduce: ArraySliceAccess
			reduce(264), // -, reduce: ArraySliceAccess
			reduce(264), // .+, reduce: ArraySliceAccess
			reduce(264), // .-, reduce: ArraySliceAccess
			reduce(264), // *, reduce: ArraySliceAccess
			reduce(264), // /, reduce: ArraySliceAccess
			reduce(264), // //, reduce: ArraySliceAccess
			reduce(264), // %, reduce: ArraySliceAccess
			reduce(264), // .*, reduce: ArraySliceAccess
			reduce(264), // ./, reduce: ArraySliceAccess
			reduce(264), // .//, reduce: ArraySliceAccess
			reduce(264), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(264), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(264), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(264), // [[, reduce: ArraySliceAccess
			reduce(264), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4306
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(261), // >, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |, reduce: ArrayOrMapPositionalNameAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(261), // ,, reduce: ArrayOrMapPositionalNameAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(261), // ), reduce: ArrayOrMapPositionalNameAccess
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(261), // ?, reduce: ArrayOrMapPositionalNameAccess
			nil,         // :
			reduce(261), // ||, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // =~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ==, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .+, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .-, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .*, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ./, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .//, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ., reduce: ArrayOrMapPositionalNameAccess
			nil,         // !
			nil,         // ~
			reduce(261), // **, reduce: ArrayOrMapPositionalNameAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(261), // [, reduce: ArrayOrMapPositionalNameAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(261), // [[, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // [[[, reduce: ArrayOrMapPositionalNameAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4307
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4510), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4308
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(324), // >, reduce: UnnamedFunctionDefinition
			reduce(324), // >>, reduce: UnnamedFunctionDefinition
			reduce(324), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(324), // ,, reduce: UnnamedFunctionDefinition
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(324), // ), reduce: UnnamedFunctionDefinition
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(324), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(324), // ||, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^, reduce: UnnamedFunctionDefinition
			reduce(324), // &&, reduce: UnnamedFunctionDefinition
			reduce(324), // ??, reduce: UnnamedFunctionDefinition
			reduce(324), // ???, reduce: UnnamedFunctionDefinition
			reduce(324), // =~, reduce: UnnamedFunctionDefinition
			reduce(324), // !=~, reduce: UnnamedFunctionDefinition
			reduce(324), // ==, reduce: UnnamedFunctionDefinition
			reduce(324), // !=, reduce: UnnamedFunctionDefinition
			reduce(324), // <=>, reduce: UnnamedFunctionDefinition
			reduce(324), // >=, reduce: UnnamedFunctionDefinition
			reduce(324), // <, reduce: UnnamedFunctionDefinition
			reduce(324), // <=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^, reduce: UnnamedFunctionDefinition
			reduce(324), // &, reduce: UnnamedFunctionDefinition
			reduce(324), // <<, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>, reduce: UnnamedFunctionDefinition
			reduce(324), // +, reduce: UnnamedFunctionDefinition
			reduce(324), // -, reduce: UnnamedFunctionDefinition
			reduce(324), // .+, reduce: UnnamedFunctionDefinition
			reduce(324), // .-, reduce: UnnamedFunctionDefinition
			reduce(324), // *, reduce: UnnamedFunctionDefinition
			reduce(324), // /, reduce: UnnamedFunctionDefinition
			reduce(324), // //, reduce: UnnamedFunctionDefinition
			reduce(324), // %, reduce: UnnamedFunctionDefinition
			reduce(324), // .*, reduce: UnnamedFunctionDefinition
			reduce(324), // ./, reduce: UnnamedFunctionDefinition
			reduce(324), // .//, reduce: UnnamedFunctionDefinition
			reduce(324), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(324), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(324), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(324), // [[, reduce: UnnamedFunctionDefinition
			reduce(324), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4309
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S4310
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4311
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(13), // $, reduce: StatementBlockInBraces
			nil,        // empty
			reduce(13), // ;, reduce: StatementBlockInBraces
			reduce(13), // {, reduce: StatementBlockInBraces
			nil,        // }
			nil,        // =
			reduce(13), // unset, reduce: StatementBlockInBraces
			reduce(13), // filter, reduce: StatementBlockInBraces
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			reduce(13), // print, reduce: StatementBlockInBraces
			nil,        // ,
			reduce(13), // printn, reduce: StatementBlockInBraces
			reduce(13), // eprint, reduce: StatementBlockInBraces
			reduce(13), // eprintn, reduce: StatementBlockInBraces
			reduce(13), // dump, reduce: StatementBlockInBraces
			reduce(13), // edump, reduce: StatementBlockInBraces
			reduce(13), // tee, reduce: StatementBlockInBraces
			reduce(13), // emitf, reduce: StatementBlockInBraces
			reduce(13), // emit1, reduce: StatementBlockInBraces
			reduce(13), // emit, reduce: StatementBlockInBraces
			reduce(13), // (, reduce: StatementBlockInBraces
			nil,        // )
			reduce(13), // emitp, reduce: StatementBlockInBraces
			reduce(13), // field_name, reduce: StatementBlockInBraces
			reduce(13), // $[, reduce: StatementBlockInBraces
			nil,        // ]
			reduce(13), // braced_field_name, reduce: StatementBlockInBraces
			reduce(13), // $[[, reduce: StatementBlockInBraces
			reduce(13), // $[[[, reduce: StatementBlockInBraces
			reduce(13), // full_srec, reduce: StatementBlockInBraces
			reduce(13), // oosvar_name, reduce: StatementBlockInBraces
			reduce(13), // @[, reduce: StatementBlockInBraces
			reduce(13), // braced_oosvar_name, reduce: StatementBlockInBraces
			reduce(13), // full_oosvar, reduce: StatementBlockInBraces
			reduce(13), // all, reduce: StatementBlockInBraces
			reduce(13), // non_sigil_name, reduce: StatementBlockInBraces
			reduce(13), // arr, reduce: StatementBlockInBraces
			reduce(13), // bool, reduce: StatementBlockInBraces
			reduce(13), // float, reduce: StatementBlockInBraces
			reduce(13), // int, reduce: StatementBlockInBraces
			reduce(13), // map, reduce: StatementBlockInBraces
			reduce(13), // num, reduce: StatementBlockInBraces
			reduce(13), // str, reduce: StatementBlockInBraces
			reduce(13), // var, reduce: StatementBlockInBraces
			reduce(13), // funct, reduce: StatementBlockInBraces
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			reduce(13), // !, reduce: StatementBlockInBraces
			reduce(13), // ~, reduce: StatementBlockInBraces
			nil,        // **
			reduce(13), // string_literal, reduce: StatementBlockInBraces
			reduce(13), // regex_case_insensitive, reduce: StatementBlockInBraces
			reduce(13), // int_literal, reduce: StatementBlockInBraces
			reduce(13), // float_literal, reduce: StatementBlockInBraces
			reduce(13), // boolean_literal, reduce: StatementBlockInBraces
			reduce(13), // null_literal, reduce: StatementBlockInBraces
			reduce(13), // inf_literal, reduce: StatementBlockInBraces
			reduce(13), // nan_literal, reduce: StatementBlockInBraces
			reduce(13), // const_M_PI, reduce: StatementBlockInBraces
			reduce(13), // const_M_E, reduce: StatementBlockInBraces
			reduce(13), // panic, reduce: StatementBlockInBraces
			reduce(13), // [, reduce: StatementBlockInBraces
			reduce(13), // ctx_IPS, reduce: StatementBlockInBraces
			reduce(13), // ctx_IFS, reduce: StatementBlockInBraces
			reduce(13), // ctx_IRS, reduce: StatementBlockInBraces
			reduce(13), // ctx_OPS, reduce: StatementBlockInBraces
			reduce(13), // ctx_OFS, reduce: StatementBlockInBraces
			reduce(13), // ctx_ORS, reduce: StatementBlockInBraces
			reduce(13), // ctx_FLATSEP, reduce: StatementBlockInBraces
			reduce(13), // ctx_NF, reduce: StatementBlockInBraces
			reduce(13), // ctx_NR, reduce: StatementBlockInBraces
			reduce(13), // ctx_FNR, reduce: StatementBlockInBraces
			reduce(13), // ctx_FILENAME, reduce: StatementBlockInBraces
			reduce(13), // ctx_FILENUM, reduce: StatementBlockInBraces
			reduce(13), // env, reduce: StatementBlockInBraces
			nil,        // [[
			nil,        // [[[
			reduce(13), // call, reduce: StatementBlockInBraces
			reduce(13), // begin, reduce: StatementBlockInBraces
			reduce(13), // end, reduce: StatementBlockInBraces
			reduce(13), // if, reduce: StatementBlockInBraces
			reduce(13), // elif, reduce: StatementBlockInBraces
			reduce(13), // else, reduce: StatementBlockInBraces
			reduce(13), // while, reduce: StatementBlockInBraces
			reduce(13), // do, reduce: StatementBlockInBraces
			reduce(13), // for, reduce: StatementBlockInBraces
			nil,        // in
			reduce(13), // break, reduce: StatementBlockInBraces
			reduce(13), // continue, reduce: StatementBlockInBraces
			reduce(13), // func, reduce: StatementBlockInBraces
			reduce(13), // subr, reduce: StatementBlockInBraces
			reduce(13), // return, reduce: StatementBlockInBraces
		},
	},
	actionRow{ // S4312
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(235), // ;, reduce: MapLiteral
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(235), // >, reduce: MapLiteral
			reduce(235), // >>, reduce: MapLiteral
			reduce(235), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(235), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(235), // ?, reduce: MapLiteral
			nil,         // :
			reduce(235), // ||, reduce: MapLiteral
			reduce(235), // ^^, reduce: MapLiteral
			reduce(235), // &&, reduce: MapLiteral
			reduce(235), // ??, reduce: MapLiteral
			reduce(235), // ???, reduce: MapLiteral
			reduce(235), // =~, reduce: MapLiteral
			reduce(235), // !=~, reduce: MapLiteral
			reduce(235), // ==, reduce: MapLiteral
			reduce(235), // !=, reduce: MapLiteral
			reduce(235), // <=>, reduce: MapLiteral
			reduce(235), // >=, reduce: MapLiteral
			reduce(235), // <, reduce: MapLiteral
			reduce(235), // <=, reduce: MapLiteral
			reduce(235), // ^, reduce: MapLiteral
			reduce(235), // &, reduce: MapLiteral
			reduce(235), // <<, reduce: MapLiteral
			reduce(235), // >>>, reduce: MapLiteral
			reduce(235), // +, reduce: MapLiteral
			reduce(235), // -, reduce: MapLiteral
			reduce(235), // .+, reduce: MapLiteral
			reduce(235), // .-, reduce: MapLiteral
			reduce(235), // *, reduce: MapLiteral
			reduce(235), // /, reduce: MapLiteral
			reduce(235), // //, reduce: MapLiteral
			reduce(235), // %, reduce: MapLiteral
			reduce(235), // .*, reduce: MapLiteral
			reduce(235), // ./, reduce: MapLiteral
			reduce(235), // .//, reduce: MapLiteral
			reduce(235), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(235), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(235), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(235), // [[, reduce: MapLiteral
			reduce(235), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4313
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(208), // ;, reduce: PrecedenceChainEnd
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(208), // >, reduce: PrecedenceChainEnd
			reduce(208), // >>, reduce: PrecedenceChainEnd
			reduce(208), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(208), // ,, reduce: PrecedenceChainEnd
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(208), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(208), // ||, reduce: PrecedenceChainEnd
			reduce(208), // ^^, reduce: PrecedenceChainEnd
			reduce(208), // &&, reduce: PrecedenceChainEnd
			reduce(208), // ??, reduce: PrecedenceChainEnd
			reduce(208), // ???, reduce: PrecedenceChainEnd
			reduce(208), // =~, reduce: PrecedenceChainEnd
			reduce(208), // !=~, reduce: PrecedenceChainEnd
			reduce(208), // ==, reduce: PrecedenceChainEnd
			reduce(208), // !=, reduce: PrecedenceChainEnd
			reduce(208), // <=>, reduce: PrecedenceChainEnd
			reduce(208), // >=, reduce: PrecedenceChainEnd
			reduce(208), // <, reduce: PrecedenceChainEnd
			reduce(208), // <=, reduce: PrecedenceChainEnd
			reduce(208), // ^, reduce: PrecedenceChainEnd
			reduce(208), // &, reduce: PrecedenceChainEnd
			reduce(208), // <<, reduce: PrecedenceChainEnd
			reduce(208), // >>>, reduce: PrecedenceChainEnd
			reduce(208), // +, reduce: PrecedenceChainEnd
			reduce(208), // -, reduce: PrecedenceChainEnd
			reduce(208), // .+, reduce: PrecedenceChainEnd
			reduce(208), // .-, reduce: PrecedenceChainEnd
			reduce(208), // *, reduce: PrecedenceChainEnd
			reduce(208), // /, reduce: PrecedenceChainEnd
			reduce(208), // //, reduce: PrecedenceChainEnd
			reduce(208), // %, reduce: PrecedenceChainEnd
			reduce(208), // .*, reduce: PrecedenceChainEnd
			reduce(208), // ./, reduce: PrecedenceChainEnd
			reduce(208), // .//, reduce: PrecedenceChainEnd
			reduce(208), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(208), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4314
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(105), // ;, reduce: IndirectFieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(105), // >, reduce: IndirectFieldValue
			reduce(105), // >>, reduce: IndirectFieldValue
			reduce(105), // |, reduce: IndirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(105), // ,, reduce: IndirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(105), // ?, reduce: IndirectFieldValue
			nil,         // :
			reduce(105), // ||, reduce: IndirectFieldValue
			reduce(105), // ^^, reduce: IndirectFieldValue
			reduce(105), // &&, reduce: IndirectFieldValue
			reduce(105), // ??, reduce: IndirectFieldValue
			reduce(105), // ???, reduce: IndirectFieldValue
			reduce(105), // =~, reduce: IndirectFieldValue
			reduce(105), // !=~, reduce: IndirectFieldValue
			reduce(105), // ==, reduce: IndirectFieldValue
			reduce(105), // !=, reduce: IndirectFieldValue
			reduce(105), // <=>, reduce: IndirectFieldValue
			reduce(105), // >=, reduce: IndirectFieldValue
			reduce(105), // <, reduce: IndirectFieldValue
			reduce(105), // <=, reduce: IndirectFieldValue
			reduce(105), // ^, reduce: IndirectFieldValue
			reduce(105), // &, reduce: IndirectFieldValue
			reduce(105), // <<, reduce: IndirectFieldValue
			reduce(105), // >>>, reduce: IndirectFieldValue
			reduce(105), // +, reduce: IndirectFieldValue
			reduce(105), // -, reduce: IndirectFieldValue
			reduce(105), // .+, reduce: IndirectFieldValue
			reduce(105), // .-, reduce: IndirectFieldValue
			reduce(105), // *, reduce: IndirectFieldValue
			reduce(105), // /, reduce: IndirectFieldValue
			reduce(105), // //, reduce: IndirectFieldValue
			reduce(105), // %, reduce: IndirectFieldValue
			reduce(105), // .*, reduce: IndirectFieldValue
			reduce(105), // ./, reduce: IndirectFieldValue
			reduce(105), // .//, reduce: IndirectFieldValue
			reduce(105), // ., reduce: IndirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(105), // **, reduce: IndirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(105), // [, reduce: IndirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(105), // [[, reduce: IndirectFieldValue
			reduce(105), // [[[, reduce: IndirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4315
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4513), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4316
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4514), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4317
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(114), // ;, reduce: IndirectOosvarValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(114), // >, reduce: IndirectOosvarValue
			reduce(114), // >>, reduce: IndirectOosvarValue
			reduce(114), // |, reduce: IndirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(114), // ,, reduce: IndirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(114), // ?, reduce: IndirectOosvarValue
			nil,         // :
			reduce(114), // ||, reduce: IndirectOosvarValue
			reduce(114), // ^^, reduce: IndirectOosvarValue
			reduce(114), // &&, reduce: IndirectOosvarValue
			reduce(114), // ??, reduce: IndirectOosvarValue
			reduce(114), // ???, reduce: IndirectOosvarValue
			reduce(114), // =~, reduce: IndirectOosvarValue
			reduce(114), // !=~, reduce: IndirectOosvarValue
			reduce(114), // ==, reduce: IndirectOosvarValue
			reduce(114), // !=, reduce: IndirectOosvarValue
			reduce(114), // <=>, reduce: IndirectOosvarValue
			reduce(114), // >=, reduce: IndirectOosvarValue
			reduce(114), // <, reduce: IndirectOosvarValue
			reduce(114), // <=, reduce: IndirectOosvarValue
			reduce(114), // ^, reduce: IndirectOosvarValue
			reduce(114), // &, reduce: IndirectOosvarValue
			reduce(114), // <<, reduce: IndirectOosvarValue
			reduce(114), // >>>, reduce: IndirectOosvarValue
			reduce(114), // +, reduce: IndirectOosvarValue
			reduce(114), // -, reduce: IndirectOosvarValue
			reduce(114), // .+, reduce: IndirectOosvarValue
			reduce(114), // .-, reduce: IndirectOosvarValue
			reduce(114), // *, reduce: IndirectOosvarValue
			reduce(114), // /, reduce: IndirectOosvarValue
			reduce(114), // //, reduce: IndirectOosvarValue
			reduce(114), // %, reduce: IndirectOosvarValue
			reduce(114), // .*, reduce: IndirectOosvarValue
			reduce(114), // ./, reduce: IndirectOosvarValue
			reduce(114), // .//, reduce: IndirectOosvarValue
			reduce(114), // ., reduce: IndirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(114), // **, reduce: IndirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(114), // [, reduce: IndirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(114), // [[, reduce: IndirectOosvarValue
			reduce(114), // [[[, reduce: IndirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4318
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4515), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4319
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(151), // ;, reduce: LogicalOrTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(151), // ,, reduce: LogicalOrTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(151), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(151), // ||, reduce: LogicalOrTerm
			shift(4046), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4320
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(153), // ;, reduce: LogicalXORTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(153), // ,, reduce: LogicalXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(153), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(153), // ||, reduce: LogicalXORTerm
			reduce(153), // ^^, reduce: LogicalXORTerm
			shift(4047), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4321
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(155), // ;, reduce: LogicalAndTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(155), // ,, reduce: LogicalAndTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(155), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(155), // ||, reduce: LogicalAndTerm
			reduce(155), // ^^, reduce: LogicalAndTerm
			reduce(155), // &&, reduce: LogicalAndTerm
			shift(4048), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4322
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(157), // ;, reduce: AbsentCoalesceTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(157), // ,, reduce: AbsentCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(157), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(157), // ||, reduce: AbsentCoalesceTerm
			reduce(157), // ^^, reduce: AbsentCoalesceTerm
			reduce(157), // &&, reduce: AbsentCoalesceTerm
			reduce(157), // ??, reduce: AbsentCoalesceTerm
			shift(4049), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4323
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(159), // ;, reduce: EmptyCoalesceTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(159), // ,, reduce: EmptyCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(159), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(159), // ||, reduce: EmptyCoalesceTerm
			reduce(159), // ^^, reduce: EmptyCoalesceTerm
			reduce(159), // &&, reduce: EmptyCoalesceTerm
			reduce(159), // ??, reduce: EmptyCoalesceTerm
			reduce(159), // ???, reduce: EmptyCoalesceTerm
			shift(4050), // =~
			shift(4051), // !=~
			shift(4052), // ==
			shift(4053), // !=
			shift(4054), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4324
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(161), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(4055), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(161), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(161), // ?, reduce: EqneTerm
			nil,         // :
			reduce(161), // ||, reduce: EqneTerm
			reduce(161), // ^^, reduce: EqneTerm
			reduce(161), // &&, reduce: EqneTerm
			reduce(161), // ??, reduce: EqneTerm
			reduce(161), // ???, reduce: EqneTerm
			reduce(161), // =~, reduce: EqneTerm
			reduce(161), // !=~, reduce: EqneTerm
			reduce(161), // ==, reduce: EqneTerm
			reduce(161), // !=, reduce: EqneTerm
			reduce(161), // <=>, reduce: EqneTerm
			shift(4056), // >=
			shift(4057), // <
			shift(4058), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4325
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(162), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(4055), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(162), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(162), // ?, reduce: EqneTerm
			nil,         // :
			reduce(162), // ||, reduce: EqneTerm
			reduce(162), // ^^, reduce: EqneTerm
			reduce(162), // &&, reduce: EqneTerm
			reduce(162), // ??, reduce: EqneTerm
			reduce(162), // ???, reduce: EqneTerm
			reduce(162), // =~, reduce: EqneTerm
			reduce(162), // !=~, reduce: EqneTerm
			reduce(162), // ==, reduce: EqneTerm
			reduce(162), // !=, reduce: EqneTerm
			reduce(162), // <=>, reduce: EqneTerm
			shift(4056), // >=
			shift(4057), // <
			shift(4058), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4326
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(163), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(4055), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(163), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(163), // ?, reduce: EqneTerm
			nil,         // :
			reduce(163), // ||, reduce: EqneTerm
			reduce(163), // ^^, reduce: EqneTerm
			reduce(163), // &&, reduce: EqneTerm
			reduce(163), // ??, reduce: EqneTerm
			reduce(163), // ???, reduce: EqneTerm
			reduce(163), // =~, reduce: EqneTerm
			reduce(163), // !=~, reduce: EqneTerm
			reduce(163), // ==, reduce: EqneTerm
			reduce(163), // !=, reduce: EqneTerm
			reduce(163), // <=>, reduce: EqneTerm
			shift(4056), // >=
			shift(4057), // <
			shift(4058), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4327
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(164), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(4055), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(164), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(164), // ?, reduce: EqneTerm
			nil,         // :
			reduce(164), // ||, reduce: EqneTerm
			reduce(164), // ^^, reduce: EqneTerm
			reduce(164), // &&, reduce: EqneTerm
			reduce(164), // ??, reduce: EqneTerm
			reduce(164), // ???, reduce: EqneTerm
			reduce(164), // =~, reduce: EqneTerm
			reduce(164), // !=~, reduce: EqneTerm
			reduce(164), // ==, reduce: EqneTerm
			reduce(164), // !=, reduce: EqneTerm
			reduce(164), // <=>, reduce: EqneTerm
			shift(4056), // >=
			shift(4057), // <
			shift(4058), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4328
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(165), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			shift(4055), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(165), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(165), // ?, reduce: EqneTerm
			nil,         // :
			reduce(165), // ||, reduce: EqneTerm
			reduce(165), // ^^, reduce: EqneTerm
			reduce(165), // &&, reduce: EqneTerm
			reduce(165), // ??, reduce: EqneTerm
			reduce(165), // ???, reduce: EqneTerm
			reduce(165), // =~, reduce: EqneTerm
			reduce(165), // !=~, reduce: EqneTerm
			reduce(165), // ==, reduce: EqneTerm
			reduce(165), // !=, reduce: EqneTerm
			reduce(165), // <=>, reduce: EqneTerm
			shift(4056), // >=
			shift(4057), // <
			shift(4058), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4329
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(167), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(167), // >, reduce: CmpTerm
			nil,         // >>
			shift(4059), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(167), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(167), // ?, reduce: CmpTerm
			nil,         // :
			reduce(167), // ||, reduce: CmpTerm
			reduce(167), // ^^, reduce: CmpTerm
			reduce(167), // &&, reduce: CmpTerm
			reduce(167), // ??, reduce: CmpTerm
			reduce(167), // ???, reduce: CmpTerm
			reduce(167), // =~, reduce: CmpTerm
			reduce(167), // !=~, reduce: CmpTerm
			reduce(167), // ==, reduce: CmpTerm
			reduce(167), // !=, reduce: CmpTerm
			reduce(167), // <=>, reduce: CmpTerm
			reduce(167), // >=, reduce: CmpTerm
			reduce(167), // <, reduce: CmpTerm
			reduce(167), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4330
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(168), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(168), // >, reduce: CmpTerm
			nil,         // >>
			shift(4059), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(168), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(168), // ?, reduce: CmpTerm
			nil,         // :
			reduce(168), // ||, reduce: CmpTerm
			reduce(168), // ^^, reduce: CmpTerm
			reduce(168), // &&, reduce: CmpTerm
			reduce(168), // ??, reduce: CmpTerm
			reduce(168), // ???, reduce: CmpTerm
			reduce(168), // =~, reduce: CmpTerm
			reduce(168), // !=~, reduce: CmpTerm
			reduce(168), // ==, reduce: CmpTerm
			reduce(168), // !=, reduce: CmpTerm
			reduce(168), // <=>, reduce: CmpTerm
			reduce(168), // >=, reduce: CmpTerm
			reduce(168), // <, reduce: CmpTerm
			reduce(168), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4331
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(169), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(169), // >, reduce: CmpTerm
			nil,         // >>
			shift(4059), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(169), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(169), // ?, reduce: CmpTerm
			nil,         // :
			reduce(169), // ||, reduce: CmpTerm
			reduce(169), // ^^, reduce: CmpTerm
			reduce(169), // &&, reduce: CmpTerm
			reduce(169), // ??, reduce: CmpTerm
			reduce(169), // ???, reduce: CmpTerm
			reduce(169), // =~, reduce: CmpTerm
			reduce(169), // !=~, reduce: CmpTerm
			reduce(169), // ==, reduce: CmpTerm
			reduce(169), // !=, reduce: CmpTerm
			reduce(169), // <=>, reduce: CmpTerm
			reduce(169), // >=, reduce: CmpTerm
			reduce(169), // <, reduce: CmpTerm
			reduce(169), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4332
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(170), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(170), // >, reduce: CmpTerm
			nil,         // >>
			shift(4059), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(170), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(170), // ?, reduce: CmpTerm
			nil,         // :
			reduce(170), // ||, reduce: CmpTerm
			reduce(170), // ^^, reduce: CmpTerm
			reduce(170), // &&, reduce: CmpTerm
			reduce(170), // ??, reduce: CmpTerm
			reduce(170), // ???, reduce: CmpTerm
			reduce(170), // =~, reduce: CmpTerm
			reduce(170), // !=~, reduce: CmpTerm
			reduce(170), // ==, reduce: CmpTerm
			reduce(170), // !=, reduce: CmpTerm
			reduce(170), // <=>, reduce: CmpTerm
			reduce(170), // >=, reduce: CmpTerm
			reduce(170), // <, reduce: CmpTerm
			reduce(170), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4333
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(172), // ;, reduce: BitwiseORTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(172), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(172), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(172), // ,, reduce: BitwiseORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(172), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(172), // ||, reduce: BitwiseORTerm
			reduce(172), // ^^, reduce: BitwiseORTerm
			reduce(172), // &&, reduce: BitwiseORTerm
			reduce(172), // ??, reduce: BitwiseORTerm
			reduce(172), // ???, reduce: BitwiseORTerm
			reduce(172), // =~, reduce: BitwiseORTerm
			reduce(172), // !=~, reduce: BitwiseORTerm
			reduce(172), // ==, reduce: BitwiseORTerm
			reduce(172), // !=, reduce: BitwiseORTerm
			reduce(172), // <=>, reduce: BitwiseORTerm
			reduce(172), // >=, reduce: BitwiseORTerm
			reduce(172), // <, reduce: BitwiseORTerm
			reduce(172), // <=, reduce: BitwiseORTerm
			shift(4060), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4334
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(174), // ;, reduce: BitwiseXORTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(174), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(174), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(174), // ,, reduce: BitwiseXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(174), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(174), // ||, reduce: BitwiseXORTerm
			reduce(174), // ^^, reduce: BitwiseXORTerm
			reduce(174), // &&, reduce: BitwiseXORTerm
			reduce(174), // ??, reduce: BitwiseXORTerm
			reduce(174), // ???, reduce: BitwiseXORTerm
			reduce(174), // =~, reduce: BitwiseXORTerm
			reduce(174), // !=~, reduce: BitwiseXORTerm
			reduce(174), // ==, reduce: BitwiseXORTerm
			reduce(174), // !=, reduce: BitwiseXORTerm
			reduce(174), // <=>, reduce: BitwiseXORTerm
			reduce(174), // >=, reduce: BitwiseXORTerm
			reduce(174), // <, reduce: BitwiseXORTerm
			reduce(174), // <=, reduce: BitwiseXORTerm
			reduce(174), // ^, reduce: BitwiseXORTerm
			shift(4061), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4335
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(176), // ;, reduce: BitwiseANDTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(176), // >, reduce: BitwiseANDTerm
			shift(4062), // >>
			reduce(176), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(176), // ,, reduce: BitwiseANDTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(176), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(176), // ||, reduce: BitwiseANDTerm
			reduce(176), // ^^, reduce: BitwiseANDTerm
			reduce(176), // &&, reduce: BitwiseANDTerm
			reduce(176), // ??, reduce: BitwiseANDTerm
			reduce(176), // ???, reduce: BitwiseANDTerm
			reduce(176), // =~, reduce: BitwiseANDTerm
			reduce(176), // !=~, reduce: BitwiseANDTerm
			reduce(176), // ==, reduce: BitwiseANDTerm
			reduce(176), // !=, reduce: BitwiseANDTerm
			reduce(176), // <=>, reduce: BitwiseANDTerm
			reduce(176), // >=, reduce: BitwiseANDTerm
			reduce(176), // <, reduce: BitwiseANDTerm
			reduce(176), // <=, reduce: BitwiseANDTerm
			reduce(176), // ^, reduce: BitwiseANDTerm
			reduce(176), // &, reduce: BitwiseANDTerm
			shift(4063), // <<
			shift(4064), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4336
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(179), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(179), // >, reduce: BitwiseShiftTerm
			reduce(179), // >>, reduce: BitwiseShiftTerm
			reduce(179), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(179), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(179), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(179), // ||, reduce: BitwiseShiftTerm
			reduce(179), // ^^, reduce: BitwiseShiftTerm
			reduce(179), // &&, reduce: BitwiseShiftTerm
			reduce(179), // ??, reduce: BitwiseShiftTerm
			reduce(179), // ???, reduce: BitwiseShiftTerm
			reduce(179), // =~, reduce: BitwiseShiftTerm
			reduce(179), // !=~, reduce: BitwiseShiftTerm
			reduce(179), // ==, reduce: BitwiseShiftTerm
			reduce(179), // !=, reduce: BitwiseShiftTerm
			reduce(179), // <=>, reduce: BitwiseShiftTerm
			reduce(179), // >=, reduce: BitwiseShiftTerm
			reduce(179), // <, reduce: BitwiseShiftTerm
			reduce(179), // <=, reduce: BitwiseShiftTerm
			reduce(179), // ^, reduce: BitwiseShiftTerm
			reduce(179), // &, reduce: BitwiseShiftTerm
			reduce(179), // <<, reduce: BitwiseShiftTerm
			reduce(179), // >>>, reduce: BitwiseShiftTerm
			shift(4065), // +
			shift(4066), // -
			shift(4067), // .+
			shift(4068), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4337
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(178), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(178), // >, reduce: BitwiseShiftTerm
			reduce(178), // >>, reduce: BitwiseShiftTerm
			reduce(178), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(178), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(178), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(178), // ||, reduce: BitwiseShiftTerm
			reduce(178), // ^^, reduce: BitwiseShiftTerm
			reduce(178), // &&, reduce: BitwiseShiftTerm
			reduce(178), // ??, reduce: BitwiseShiftTerm
			reduce(178), // ???, reduce: BitwiseShiftTerm
			reduce(178), // =~, reduce: BitwiseShiftTerm
			reduce(178), // !=~, reduce: BitwiseShiftTerm
			reduce(178), // ==, reduce: BitwiseShiftTerm
			reduce(178), // !=, reduce: BitwiseShiftTerm
			reduce(178), // <=>, reduce: BitwiseShiftTerm
			reduce(178), // >=, reduce: BitwiseShiftTerm
			reduce(178), // <, reduce: BitwiseShiftTerm
			reduce(178), // <=, reduce: BitwiseShiftTerm
			reduce(178), // ^, reduce: BitwiseShiftTerm
			reduce(178), // &, reduce: BitwiseShiftTerm
			reduce(178), // <<, reduce: BitwiseShiftTerm
			reduce(178), // >>>, reduce: BitwiseShiftTerm
			shift(4065), // +
			shift(4066), // -
			shift(4067), // .+
			shift(4068), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4338
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(180), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(180), // >, reduce: BitwiseShiftTerm
			reduce(180), // >>, reduce: BitwiseShiftTerm
			reduce(180), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(180), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(180), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(180), // ||, reduce: BitwiseShiftTerm
			reduce(180), // ^^, reduce: BitwiseShiftTerm
			reduce(180), // &&, reduce: BitwiseShiftTerm
			reduce(180), // ??, reduce: BitwiseShiftTerm
			reduce(180), // ???, reduce: BitwiseShiftTerm
			reduce(180), // =~, reduce: BitwiseShiftTerm
			reduce(180), // !=~, reduce: BitwiseShiftTerm
			reduce(180), // ==, reduce: BitwiseShiftTerm
			reduce(180), // !=, reduce: BitwiseShiftTerm
			reduce(180), // <=>, reduce: BitwiseShiftTerm
			reduce(180), // >=, reduce: BitwiseShiftTerm
			reduce(180), // <, reduce: BitwiseShiftTerm
			reduce(180), // <=, reduce: BitwiseShiftTerm
			reduce(180), // ^, reduce: BitwiseShiftTerm
			reduce(180), // &, reduce: BitwiseShiftTerm
			reduce(180), // <<, reduce: BitwiseShiftTerm
			reduce(180), // >>>, reduce: BitwiseShiftTerm
			shift(4065), // +
			shift(4066), // -
			shift(4067), // .+
			shift(4068), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4339
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(182), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(182), // >, reduce: AddsubdotTerm
			reduce(182), // >>, reduce: AddsubdotTerm
			reduce(182), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(182), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(182), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(182), // ||, reduce: AddsubdotTerm
			reduce(182), // ^^, reduce: AddsubdotTerm
			reduce(182), // &&, reduce: AddsubdotTerm
			reduce(182), // ??, reduce: AddsubdotTerm
			reduce(182), // ???, reduce: AddsubdotTerm
			reduce(182), // =~, reduce: AddsubdotTerm
			reduce(182), // !=~, reduce: AddsubdotTerm
			reduce(182), // ==, reduce: AddsubdotTerm
			reduce(182), // !=, reduce: AddsubdotTerm
			reduce(182), // <=>, reduce: AddsubdotTerm
			reduce(182), // >=, reduce: AddsubdotTerm
			reduce(182), // <, reduce: AddsubdotTerm
			reduce(182), // <=, reduce: AddsubdotTerm
			reduce(182), // ^, reduce: AddsubdotTerm
			reduce(182), // &, reduce: AddsubdotTerm
			reduce(182), // <<, reduce: AddsubdotTerm
			reduce(182), // >>>, reduce: AddsubdotTerm
			reduce(182), // +, reduce: AddsubdotTerm
			reduce(182), // -, reduce: AddsubdotTerm
			reduce(182), // .+, reduce: AddsubdotTerm
			reduce(182), // .-, reduce: AddsubdotTerm
			shift(4070), // *
			shift(4071), // /
			shift(4072), // //
			shift(4073), // %
			shift(4074), // .*
			shift(4075), // ./
			shift(4076), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4340
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(183), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(183), // >, reduce: AddsubdotTerm
			reduce(183), // >>, reduce: AddsubdotTerm
			reduce(183), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(183), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(183), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(183), // ||, reduce: AddsubdotTerm
			reduce(183), // ^^, reduce: AddsubdotTerm
			reduce(183), // &&, reduce: AddsubdotTerm
			reduce(183), // ??, reduce: AddsubdotTerm
			reduce(183), // ???, reduce: AddsubdotTerm
			reduce(183), // =~, reduce: AddsubdotTerm
			reduce(183), // !=~, reduce: AddsubdotTerm
			reduce(183), // ==, reduce: AddsubdotTerm
			reduce(183), // !=, reduce: AddsubdotTerm
			reduce(183), // <=>, reduce: AddsubdotTerm
			reduce(183), // >=, reduce: AddsubdotTerm
			reduce(183), // <, reduce: AddsubdotTerm
			reduce(183), // <=, reduce: AddsubdotTerm
			reduce(183), // ^, reduce: AddsubdotTerm
			reduce(183), // &, reduce: AddsubdotTerm
			reduce(183), // <<, reduce: AddsubdotTerm
			reduce(183), // >>>, reduce: AddsubdotTerm
			reduce(183), // +, reduce: AddsubdotTerm
			reduce(183), // -, reduce: AddsubdotTerm
			reduce(183), // .+, reduce: AddsubdotTerm
			reduce(183), // .-, reduce: AddsubdotTerm
			shift(4070), // *
			shift(4071), // /
			shift(4072), // //
			shift(4073), // %
			shift(4074), // .*
			shift(4075), // ./
			shift(4076), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4341
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(184), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(184), // >, reduce: AddsubdotTerm
			reduce(184), // >>, reduce: AddsubdotTerm
			reduce(184), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(184), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(184), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(184), // ||, reduce: AddsubdotTerm
			reduce(184), // ^^, reduce: AddsubdotTerm
			reduce(184), // &&, reduce: AddsubdotTerm
			reduce(184), // ??, reduce: AddsubdotTerm
			reduce(184), // ???, reduce: AddsubdotTerm
			reduce(184), // =~, reduce: AddsubdotTerm
			reduce(184), // !=~, reduce: AddsubdotTerm
			reduce(184), // ==, reduce: AddsubdotTerm
			reduce(184), // !=, reduce: AddsubdotTerm
			reduce(184), // <=>, reduce: AddsubdotTerm
			reduce(184), // >=, reduce: AddsubdotTerm
			reduce(184), // <, reduce: AddsubdotTerm
			reduce(184), // <=, reduce: AddsubdotTerm
			reduce(184), // ^, reduce: AddsubdotTerm
			reduce(184), // &, reduce: AddsubdotTerm
			reduce(184), // <<, reduce: AddsubdotTerm
			reduce(184), // >>>, reduce: AddsubdotTerm
			reduce(184), // +, reduce: AddsubdotTerm
			reduce(184), // -, reduce: AddsubdotTerm
			reduce(184), // .+, reduce: AddsubdotTerm
			reduce(184), // .-, reduce: AddsubdotTerm
			shift(4070), // *
			shift(4071), // /
			shift(4072), // //
			shift(4073), // %
			shift(4074), // .*
			shift(4075), // ./
			shift(4076), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4342
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(185), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(185), // >, reduce: AddsubdotTerm
			reduce(185), // >>, reduce: AddsubdotTerm
			reduce(185), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(185), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(185), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(185), // ||, reduce: AddsubdotTerm
			reduce(185), // ^^, reduce: AddsubdotTerm
			reduce(185), // &&, reduce: AddsubdotTerm
			reduce(185), // ??, reduce: AddsubdotTerm
			reduce(185), // ???, reduce: AddsubdotTerm
			reduce(185), // =~, reduce: AddsubdotTerm
			reduce(185), // !=~, reduce: AddsubdotTerm
			reduce(185), // ==, reduce: AddsubdotTerm
			reduce(185), // !=, reduce: AddsubdotTerm
			reduce(185), // <=>, reduce: AddsubdotTerm
			reduce(185), // >=, reduce: AddsubdotTerm
			reduce(185), // <, reduce: AddsubdotTerm
			reduce(185), // <=, reduce: AddsubdotTerm
			reduce(185), // ^, reduce: AddsubdotTerm
			reduce(185), // &, reduce: AddsubdotTerm
			reduce(185), // <<, reduce: AddsubdotTerm
			reduce(185), // >>>, reduce: AddsubdotTerm
			reduce(185), // +, reduce: AddsubdotTerm
			reduce(185), // -, reduce: AddsubdotTerm
			reduce(185), // .+, reduce: AddsubdotTerm
			reduce(185), // .-, reduce: AddsubdotTerm
			shift(4070), // *
			shift(4071), // /
			shift(4072), // //
			shift(4073), // %
			shift(4074), // .*
			shift(4075), // ./
			shift(4076), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4343
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(187), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(187), // >, reduce: MuldivTerm
			reduce(187), // >>, reduce: MuldivTerm
			reduce(187), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(187), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(187), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(187), // ||, reduce: MuldivTerm
			reduce(187), // ^^, reduce: MuldivTerm
			reduce(187), // &&, reduce: MuldivTerm
			reduce(187), // ??, reduce: MuldivTerm
			reduce(187), // ???, reduce: MuldivTerm
			reduce(187), // =~, reduce: MuldivTerm
			reduce(187), // !=~, reduce: MuldivTerm
			reduce(187), // ==, reduce: MuldivTerm
			reduce(187), // !=, reduce: MuldivTerm
			reduce(187), // <=>, reduce: MuldivTerm
			reduce(187), // >=, reduce: MuldivTerm
			reduce(187), // <, reduce: MuldivTerm
			reduce(187), // <=, reduce: MuldivTerm
			reduce(187), // ^, reduce: MuldivTerm
			reduce(187), // &, reduce: MuldivTerm
			reduce(187), // <<, reduce: MuldivTerm
			reduce(187), // >>>, reduce: MuldivTerm
			reduce(187), // +, reduce: MuldivTerm
			reduce(187), // -, reduce: MuldivTerm
			reduce(187), // .+, reduce: MuldivTerm
			reduce(187), // .-, reduce: MuldivTerm
			reduce(187), // *, reduce: MuldivTerm
			reduce(187), // /, reduce: MuldivTerm
			reduce(187), // //, reduce: MuldivTerm
			reduce(187), // %, reduce: MuldivTerm
			reduce(187), // .*, reduce: MuldivTerm
			reduce(187), // ./, reduce: MuldivTerm
			reduce(187), // .//, reduce: MuldivTerm
			shift(4080), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4344
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(188), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(188), // >, reduce: MuldivTerm
			reduce(188), // >>, reduce: MuldivTerm
			reduce(188), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(188), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(188), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(188), // ||, reduce: MuldivTerm
			reduce(188), // ^^, reduce: MuldivTerm
			reduce(188), // &&, reduce: MuldivTerm
			reduce(188), // ??, reduce: MuldivTerm
			reduce(188), // ???, reduce: MuldivTerm
			reduce(188), // =~, reduce: MuldivTerm
			reduce(188), // !=~, reduce: MuldivTerm
			reduce(188), // ==, reduce: MuldivTerm
			reduce(188), // !=, reduce: MuldivTerm
			reduce(188), // <=>, reduce: MuldivTerm
			reduce(188), // >=, reduce: MuldivTerm
			reduce(188), // <, reduce: MuldivTerm
			reduce(188), // <=, reduce: MuldivTerm
			reduce(188), // ^, reduce: MuldivTerm
			reduce(188), // &, reduce: MuldivTerm
			reduce(188), // <<, reduce: MuldivTerm
			reduce(188), // >>>, reduce: MuldivTerm
			reduce(188), // +, reduce: MuldivTerm
			reduce(188), // -, reduce: MuldivTerm
			reduce(188), // .+, reduce: MuldivTerm
			reduce(188), // .-, reduce: MuldivTerm
			reduce(188), // *, reduce: MuldivTerm
			reduce(188), // /, reduce: MuldivTerm
			reduce(188), // //, reduce: MuldivTerm
			reduce(188), // %, reduce: MuldivTerm
			reduce(188), // .*, reduce: MuldivTerm
			reduce(188), // ./, reduce: MuldivTerm
			reduce(188), // .//, reduce: MuldivTerm
			shift(4080), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4345
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(189), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(189), // >, reduce: MuldivTerm
			reduce(189), // >>, reduce: MuldivTerm
			reduce(189), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(189), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(189), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(189), // ||, reduce: MuldivTerm
			reduce(189), // ^^, reduce: MuldivTerm
			reduce(189), // &&, reduce: MuldivTerm
			reduce(189), // ??, reduce: MuldivTerm
			reduce(189), // ???, reduce: MuldivTerm
			reduce(189), // =~, reduce: MuldivTerm
			reduce(189), // !=~, reduce: MuldivTerm
			reduce(189), // ==, reduce: MuldivTerm
			reduce(189), // !=, reduce: MuldivTerm
			reduce(189), // <=>, reduce: MuldivTerm
			reduce(189), // >=, reduce: MuldivTerm
			reduce(189), // <, reduce: MuldivTerm
			reduce(189), // <=, reduce: MuldivTerm
			reduce(189), // ^, reduce: MuldivTerm
			reduce(189), // &, reduce: MuldivTerm
			reduce(189), // <<, reduce: MuldivTerm
			reduce(189), // >>>, reduce: MuldivTerm
			reduce(189), // +, reduce: MuldivTerm
			reduce(189), // -, reduce: MuldivTerm
			reduce(189), // .+, reduce: MuldivTerm
			reduce(189), // .-, reduce: MuldivTerm
			reduce(189), // *, reduce: MuldivTerm
			reduce(189), // /, reduce: MuldivTerm
			reduce(189), // //, reduce: MuldivTerm
			reduce(189), // %, reduce: MuldivTerm
			reduce(189), // .*, reduce: MuldivTerm
			reduce(189), // ./, reduce: MuldivTerm
			reduce(189), // .//, reduce: MuldivTerm
			shift(4080), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4346
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(190), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(190), // >, reduce: MuldivTerm
			reduce(190), // >>, reduce: MuldivTerm
			reduce(190), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(190), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(190), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(190), // ||, reduce: MuldivTerm
			reduce(190), // ^^, reduce: MuldivTerm
			reduce(190), // &&, reduce: MuldivTerm
			reduce(190), // ??, reduce: MuldivTerm
			reduce(190), // ???, reduce: MuldivTerm
			reduce(190), // =~, reduce: MuldivTerm
			reduce(190), // !=~, reduce: MuldivTerm
			reduce(190), // ==, reduce: MuldivTerm
			reduce(190), // !=, reduce: MuldivTerm
			reduce(190), // <=>, reduce: MuldivTerm
			reduce(190), // >=, reduce: MuldivTerm
			reduce(190), // <, reduce: MuldivTerm
			reduce(190), // <=, reduce: MuldivTerm
			reduce(190), // ^, reduce: MuldivTerm
			reduce(190), // &, reduce: MuldivTerm
			reduce(190), // <<, reduce: MuldivTerm
			reduce(190), // >>>, reduce: MuldivTerm
			reduce(190), // +, reduce: MuldivTerm
			reduce(190), // -, reduce: MuldivTerm
			reduce(190), // .+, reduce: MuldivTerm
			reduce(190), // .-, reduce: MuldivTerm
			reduce(190), // *, reduce: MuldivTerm
			reduce(190), // /, reduce: MuldivTerm
			reduce(190), // //, reduce: MuldivTerm
			reduce(190), // %, reduce: MuldivTerm
			reduce(190), // .*, reduce: MuldivTerm
			reduce(190), // ./, reduce: MuldivTerm
			reduce(190), // .//, reduce: MuldivTerm
			shift(4080), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4347
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(191), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(191), // >, reduce: MuldivTerm
			reduce(191), // >>, reduce: MuldivTerm
			reduce(191), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(191), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(191), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(191), // ||, reduce: MuldivTerm
			reduce(191), // ^^, reduce: MuldivTerm
			reduce(191), // &&, reduce: MuldivTerm
			reduce(191), // ??, reduce: MuldivTerm
			reduce(191), // ???, reduce: MuldivTerm
			reduce(191), // =~, reduce: MuldivTerm
			reduce(191), // !=~, reduce: MuldivTerm
			reduce(191), // ==, reduce: MuldivTerm
			reduce(191), // !=, reduce: MuldivTerm
			reduce(191), // <=>, reduce: MuldivTerm
			reduce(191), // >=, reduce: MuldivTerm
			reduce(191), // <, reduce: MuldivTerm
			reduce(191), // <=, reduce: MuldivTerm
			reduce(191), // ^, reduce: MuldivTerm
			reduce(191), // &, reduce: MuldivTerm
			reduce(191), // <<, reduce: MuldivTerm
			reduce(191), // >>>, reduce: MuldivTerm
			reduce(191), // +, reduce: MuldivTerm
			reduce(191), // -, reduce: MuldivTerm
			reduce(191), // .+, reduce: MuldivTerm
			reduce(191), // .-, reduce: MuldivTerm
			reduce(191), // *, reduce: MuldivTerm
			reduce(191), // /, reduce: MuldivTerm
			reduce(191), // //, reduce: MuldivTerm
			reduce(191), // %, reduce: MuldivTerm
			reduce(191), // .*, reduce: MuldivTerm
			reduce(191), // ./, reduce: MuldivTerm
			reduce(191), // .//, reduce: MuldivTerm
			shift(4080), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4348
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(192), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(192), // >, reduce: MuldivTerm
			reduce(192), // >>, reduce: MuldivTerm
			reduce(192), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(192), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(192), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(192), // ||, reduce: MuldivTerm
			reduce(192), // ^^, reduce: MuldivTerm
			reduce(192), // &&, reduce: MuldivTerm
			reduce(192), // ??, reduce: MuldivTerm
			reduce(192), // ???, reduce: MuldivTerm
			reduce(192), // =~, reduce: MuldivTerm
			reduce(192), // !=~, reduce: MuldivTerm
			reduce(192), // ==, reduce: MuldivTerm
			reduce(192), // !=, reduce: MuldivTerm
			reduce(192), // <=>, reduce: MuldivTerm
			reduce(192), // >=, reduce: MuldivTerm
			reduce(192), // <, reduce: MuldivTerm
			reduce(192), // <=, reduce: MuldivTerm
			reduce(192), // ^, reduce: MuldivTerm
			reduce(192), // &, reduce: MuldivTerm
			reduce(192), // <<, reduce: MuldivTerm
			reduce(192), // >>>, reduce: MuldivTerm
			reduce(192), // +, reduce: MuldivTerm
			reduce(192), // -, reduce: MuldivTerm
			reduce(192), // .+, reduce: MuldivTerm
			reduce(192), // .-, reduce: MuldivTerm
			reduce(192), // *, reduce: MuldivTerm
			reduce(192), // /, reduce: MuldivTerm
			reduce(192), // //, reduce: MuldivTerm
			reduce(192), // %, reduce: MuldivTerm
			reduce(192), // .*, reduce: MuldivTerm
			reduce(192), // ./, reduce: MuldivTerm
			reduce(192), // .//, reduce: MuldivTerm
			shift(4080), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4349
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(193), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(193), // >, reduce: MuldivTerm
			reduce(193), // >>, reduce: MuldivTerm
			reduce(193), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(193), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(193), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(193), // ||, reduce: MuldivTerm
			reduce(193), // ^^, reduce: MuldivTerm
			reduce(193), // &&, reduce: MuldivTerm
			reduce(193), // ??, reduce: MuldivTerm
			reduce(193), // ???, reduce: MuldivTerm
			reduce(193), // =~, reduce: MuldivTerm
			reduce(193), // !=~, reduce: MuldivTerm
			reduce(193), // ==, reduce: MuldivTerm
			reduce(193), // !=, reduce: MuldivTerm
			reduce(193), // <=>, reduce: MuldivTerm
			reduce(193), // >=, reduce: MuldivTerm
			reduce(193), // <, reduce: MuldivTerm
			reduce(193), // <=, reduce: MuldivTerm
			reduce(193), // ^, reduce: MuldivTerm
			reduce(193), // &, reduce: MuldivTerm
			reduce(193), // <<, reduce: MuldivTerm
			reduce(193), // >>>, reduce: MuldivTerm
			reduce(193), // +, reduce: MuldivTerm
			reduce(193), // -, reduce: MuldivTerm
			reduce(193), // .+, reduce: MuldivTerm
			reduce(193), // .-, reduce: MuldivTerm
			reduce(193), // *, reduce: MuldivTerm
			reduce(193), // /, reduce: MuldivTerm
			reduce(193), // //, reduce: MuldivTerm
			reduce(193), // %, reduce: MuldivTerm
			reduce(193), // .*, reduce: MuldivTerm
			reduce(193), // ./, reduce: MuldivTerm
			reduce(193), // .//, reduce: MuldivTerm
			shift(4080), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4350
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(195), // ;, reduce: DotTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(195), // >, reduce: DotTerm
			reduce(195), // >>, reduce: DotTerm
			reduce(195), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(195), // ,, reduce: DotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(195), // ?, reduce: DotTerm
			nil,         // :
			reduce(195), // ||, reduce: DotTerm
			reduce(195), // ^^, reduce: DotTerm
			reduce(195), // &&, reduce: DotTerm
			reduce(195), // ??, reduce: DotTerm
			reduce(195), // ???, reduce: DotTerm
			reduce(195), // =~, reduce: DotTerm
			reduce(195), // !=~, reduce: DotTerm
			reduce(195), // ==, reduce: DotTerm
			reduce(195), // !=, reduce: DotTerm
			reduce(195), // <=>, reduce: DotTerm
			reduce(195), // >=, reduce: DotTerm
			reduce(195), // <, reduce: DotTerm
			reduce(195), // <=, reduce: DotTerm
			reduce(195), // ^, reduce: DotTerm
			reduce(195), // &, reduce: DotTerm
			reduce(195), // <<, reduce: DotTerm
			reduce(195), // >>>, reduce: DotTerm
			reduce(195), // +, reduce: DotTerm
			reduce(195), // -, reduce: DotTerm
			reduce(195), // .+, reduce: DotTerm
			reduce(195), // .-, reduce: DotTerm
			reduce(195), // *, reduce: DotTerm
			reduce(195), // /, reduce: DotTerm
			reduce(195), // //, reduce: DotTerm
			reduce(195), // %, reduce: DotTerm
			reduce(195), // .*, reduce: DotTerm
			reduce(195), // ./, reduce: DotTerm
			reduce(195), // .//, reduce: DotTerm
			reduce(195), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4351
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4352
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4353
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(204), // ;, reduce: PowTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(204), // >, reduce: PowTerm
			reduce(204), // >>, reduce: PowTerm
			reduce(204), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(204), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(204), // ?, reduce: PowTerm
			nil,         // :
			reduce(204), // ||, reduce: PowTerm
			reduce(204), // ^^, reduce: PowTerm
			reduce(204), // &&, reduce: PowTerm
			reduce(204), // ??, reduce: PowTerm
			reduce(204), // ???, reduce: PowTerm
			reduce(204), // =~, reduce: PowTerm
			reduce(204), // !=~, reduce: PowTerm
			reduce(204), // ==, reduce: PowTerm
			reduce(204), // !=, reduce: PowTerm
			reduce(204), // <=>, reduce: PowTerm
			reduce(204), // >=, reduce: PowTerm
			reduce(204), // <, reduce: PowTerm
			reduce(204), // <=, reduce: PowTerm
			reduce(204), // ^, reduce: PowTerm
			reduce(204), // &, reduce: PowTerm
			reduce(204), // <<, reduce: PowTerm
			reduce(204), // >>>, reduce: PowTerm
			reduce(204), // +, reduce: PowTerm
			reduce(204), // -, reduce: PowTerm
			reduce(204), // .+, reduce: PowTerm
			reduce(204), // .-, reduce: PowTerm
			reduce(204), // *, reduce: PowTerm
			reduce(204), // /, reduce: PowTerm
			reduce(204), // //, reduce: PowTerm
			reduce(204), // %, reduce: PowTerm
			reduce(204), // .*, reduce: PowTerm
			reduce(204), // ./, reduce: PowTerm
			reduce(204), // .//, reduce: PowTerm
			reduce(204), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4354
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4518), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4519), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4355
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4521), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4356
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4522), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4357
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4523), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4358
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(229), // ;, reduce: ArrayLiteral
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(229), // >, reduce: ArrayLiteral
			reduce(229), // >>, reduce: ArrayLiteral
			reduce(229), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(229), // ,, reduce: ArrayLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(229), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(229), // ||, reduce: ArrayLiteral
			reduce(229), // ^^, reduce: ArrayLiteral
			reduce(229), // &&, reduce: ArrayLiteral
			reduce(229), // ??, reduce: ArrayLiteral
			reduce(229), // ???, reduce: ArrayLiteral
			reduce(229), // =~, reduce: ArrayLiteral
			reduce(229), // !=~, reduce: ArrayLiteral
			reduce(229), // ==, reduce: ArrayLiteral
			reduce(229), // !=, reduce: ArrayLiteral
			reduce(229), // <=>, reduce: ArrayLiteral
			reduce(229), // >=, reduce: ArrayLiteral
			reduce(229), // <, reduce: ArrayLiteral
			reduce(229), // <=, reduce: ArrayLiteral
			reduce(229), // ^, reduce: ArrayLiteral
			reduce(229), // &, reduce: ArrayLiteral
			reduce(229), // <<, reduce: ArrayLiteral
			reduce(229), // >>>, reduce: ArrayLiteral
			reduce(229), // +, reduce: ArrayLiteral
			reduce(229), // -, reduce: ArrayLiteral
			reduce(229), // .+, reduce: ArrayLiteral
			reduce(229), // .-, reduce: ArrayLiteral
			reduce(229), // *, reduce: ArrayLiteral
			reduce(229), // /, reduce: ArrayLiteral
			reduce(229), // //, reduce: ArrayLiteral
			reduce(229), // %, reduce: ArrayLiteral
			reduce(229), // .*, reduce: ArrayLiteral
			reduce(229), // ./, reduce: ArrayLiteral
			reduce(229), // .//, reduce: ArrayLiteral
			reduce(229), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(229), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(229), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(229), // [[, reduce: ArrayLiteral
			reduce(229), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4359
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(255), // ;, reduce: ENV
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(255), // >, reduce: ENV
			reduce(255), // >>, reduce: ENV
			reduce(255), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(255), // ,, reduce: ENV
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(255), // ?, reduce: ENV
			nil,         // :
			reduce(255), // ||, reduce: ENV
			reduce(255), // ^^, reduce: ENV
			reduce(255), // &&, reduce: ENV
			reduce(255), // ??, reduce: ENV
			reduce(255), // ???, reduce: ENV
			reduce(255), // =~, reduce: ENV
			reduce(255), // !=~, reduce: ENV
			reduce(255), // ==, reduce: ENV
			reduce(255), // !=, reduce: ENV
			reduce(255), // <=>, reduce: ENV
			reduce(255), // >=, reduce: ENV
			reduce(255), // <, reduce: ENV
			reduce(255), // <=, reduce: ENV
			reduce(255), // ^, reduce: ENV
			reduce(255), // &, reduce: ENV
			reduce(255), // <<, reduce: ENV
			reduce(255), // >>>, reduce: ENV
			reduce(255), // +, reduce: ENV
			reduce(255), // -, reduce: ENV
			reduce(255), // .+, reduce: ENV
			reduce(255), // .-, reduce: ENV
			reduce(255), // *, reduce: ENV
			reduce(255), // /, reduce: ENV
			reduce(255), // //, reduce: ENV
			reduce(255), // %, reduce: ENV
			reduce(255), // .*, reduce: ENV
			reduce(255), // ./, reduce: ENV
			reduce(255), // .//, reduce: ENV
			reduce(255), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(255), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(255), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(255), // [[, reduce: ENV
			reduce(255), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4360
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4524), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4361
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4525), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4362
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(268), // ;, reduce: FunctionCallsite
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(268), // >, reduce: FunctionCallsite
			reduce(268), // >>, reduce: FunctionCallsite
			reduce(268), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(268), // ,, reduce: FunctionCallsite
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(268), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(268), // ||, reduce: FunctionCallsite
			reduce(268), // ^^, reduce: FunctionCallsite
			reduce(268), // &&, reduce: FunctionCallsite
			reduce(268), // ??, reduce: FunctionCallsite
			reduce(268), // ???, reduce: FunctionCallsite
			reduce(268), // =~, reduce: FunctionCallsite
			reduce(268), // !=~, reduce: FunctionCallsite
			reduce(268), // ==, reduce: FunctionCallsite
			reduce(268), // !=, reduce: FunctionCallsite
			reduce(268), // <=>, reduce: FunctionCallsite
			reduce(268), // >=, reduce: FunctionCallsite
			reduce(268), // <, reduce: FunctionCallsite
			reduce(268), // <=, reduce: FunctionCallsite
			reduce(268), // ^, reduce: FunctionCallsite
			reduce(268), // &, reduce: FunctionCallsite
			reduce(268), // <<, reduce: FunctionCallsite
			reduce(268), // >>>, reduce: FunctionCallsite
			reduce(268), // +, reduce: FunctionCallsite
			reduce(268), // -, reduce: FunctionCallsite
			reduce(268), // .+, reduce: FunctionCallsite
			reduce(268), // .-, reduce: FunctionCallsite
			reduce(268), // *, reduce: FunctionCallsite
			reduce(268), // /, reduce: FunctionCallsite
			reduce(268), // //, reduce: FunctionCallsite
			reduce(268), // %, reduce: FunctionCallsite
			reduce(268), // .*, reduce: FunctionCallsite
			reduce(268), // ./, reduce: FunctionCallsite
			reduce(268), // .//, reduce: FunctionCallsite
			reduce(268), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(268), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(268), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(268), // [[, reduce: FunctionCallsite
			reduce(268), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4363
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4526), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4364
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4527), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4365
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(303), // $, reduce: ForLoopOneVariable
			nil,         // empty
			reduce(303), // ;, reduce: ForLoopOneVariable
			reduce(303), // {, reduce: ForLoopOneVariable
			nil,         // }
			nil,         // =
			reduce(303), // unset, reduce: ForLoopOneVariable
			reduce(303), // filter, reduce: ForLoopOneVariable
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(303), // print, reduce: ForLoopOneVariable
			nil,         // ,
			reduce(303), // printn, reduce: ForLoopOneVariable
			reduce(303), // eprint, reduce: ForLoopOneVariable
			reduce(303), // eprintn, reduce: ForLoopOneVariable
			reduce(303), // dump, reduce: ForLoopOneVariable
			reduce(303), // edump, reduce: ForLoopOneVariable
			reduce(303), // tee, reduce: ForLoopOneVariable
			reduce(303), // emitf, reduce: ForLoopOneVariable
			reduce(303), // emit1, reduce: ForLoopOneVariable
			reduce(303), // emit, reduce: ForLoopOneVariable
			reduce(303), // (, reduce: ForLoopOneVariable
			nil,         // )
			reduce(303), // emitp, reduce: ForLoopOneVariable
			reduce(303), // field_name, reduce: ForLoopOneVariable
			reduce(303), // $[, reduce: ForLoopOneVariable
			nil,         // ]
			reduce(303), // braced_field_name, reduce: ForLoopOneVariable
			reduce(303), // $[[, reduce: ForLoopOneVariable
			reduce(303), // $[[[, reduce: ForLoopOneVariable
			reduce(303), // full_srec, reduce: ForLoopOneVariable
			reduce(303), // oosvar_name, reduce: ForLoopOneVariable
			reduce(303), // @[, reduce: ForLoopOneVariable
			reduce(303), // braced_oosvar_name, reduce: ForLoopOneVariable
			reduce(303), // full_oosvar, reduce: ForLoopOneVariable
			reduce(303), // all, reduce: ForLoopOneVariable
			reduce(303), // non_sigil_name, reduce: ForLoopOneVariable
			reduce(303), // arr, reduce: ForLoopOneVariable
			reduce(303), // bool, reduce: ForLoopOneVariable
			reduce(303), // float, reduce: ForLoopOneVariable
			reduce(303), // int, reduce: ForLoopOneVariable
			reduce(303), // map, reduce: ForLoopOneVariable
			reduce(303), // num, reduce: ForLoopOneVariable
			reduce(303), // str, reduce: ForLoopOneVariable
			reduce(303), // var, reduce: ForLoopOneVariable
			reduce(303), // funct, reduce: ForLoopOneVariable
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(303), // +, reduce: ForLoopOneVariable
			reduce(303), // -, reduce: ForLoopOneVariable
			reduce(303), // .+, reduce: ForLoopOneVariable
			reduce(303), // .-, reduce: ForLoopOneVariable
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(303), // !, reduce: ForLoopOneVariable
			reduce(303), // ~, reduce: ForLoopOneVariable
			nil,         // **
			reduce(303), // string_literal, reduce: ForLoopOneVariable
			reduce(303), // regex_case_insensitive, reduce: ForLoopOneVariable
			reduce(303), // int_literal, reduce: ForLoopOneVariable
			reduce(303), // float_literal, reduce: ForLoopOneVariable
			reduce(303), // boolean_literal, reduce: ForLoopOneVariable
			reduce(303), // null_literal, reduce: ForLoopOneVariable
			reduce(303), // inf_literal, reduce: ForLoopOneVariable
			reduce(303), // nan_literal, reduce: ForLoopOneVariable
			reduce(303), // const_M_PI, reduce: ForLoopOneVariable
			reduce(303), // const_M_E, reduce: ForLoopOneVariable
			reduce(303), // panic, reduce: ForLoopOneVariable
			reduce(303), // [, reduce: ForLoopOneVariable
			reduce(303), // ctx_IPS, reduce: ForLoopOneVariable
			reduce(303), // ctx_IFS, reduce: ForLoopOneVariable
			reduce(303), // ctx_IRS, reduce: ForLoopOneVariable
			reduce(303), // ctx_OPS, reduce: ForLoopOneVariable
			reduce(303), // ctx_OFS, reduce: ForLoopOneVariable
			reduce(303), // ctx_ORS, reduce: ForLoopOneVariable
			reduce(303), // ctx_FLATSEP, reduce: ForLoopOneVariable
			reduce(303), // ctx_NF, reduce: ForLoopOneVariable
			reduce(303), // ctx_NR, reduce: ForLoopOneVariable
			reduce(303), // ctx_FNR, reduce: ForLoopOneVariable
			reduce(303), // ctx_FILENAME, reduce: ForLoopOneVariable
			reduce(303), // ctx_FILENUM, reduce: ForLoopOneVariable
			reduce(303), // env, reduce: ForLoopOneVariable
			nil,         // [[
			nil,         // [[[
			reduce(303), // call, reduce: ForLoopOneVariable
			reduce(303), // begin, reduce: ForLoopOneVariable
			reduce(303), // end, reduce: ForLoopOneVariable
			reduce(303), // if, reduce: ForLoopOneVariable
			nil,         // elif
			nil,         // else
			reduce(303), // while, reduce: ForLoopOneVariable
			reduce(303), // do, reduce: ForLoopOneVariable
			reduce(303), // for, reduce: ForLoopOneVariable
			nil,         // in
			reduce(303), // break, reduce: ForLoopOneVariable
			reduce(303), // continue, reduce: ForLoopOneVariable
			reduce(303), // func, reduce: ForLoopOneVariable
			reduce(303), // subr, reduce: ForLoopOneVariable
			reduce(303), // return, reduce: ForLoopOneVariable
		},
	},
	actionRow{ // S4366
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			shift(4528), // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4367
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(108), // =, reduce: PositionalFieldValue
			nil,         // unset
			nil,         // filter
			reduce(108), // >, reduce: PositionalFieldValue
			reduce(108), // >>, reduce: PositionalFieldValue
			reduce(108), // |, reduce: PositionalFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(108), // ||=, reduce: PositionalFieldValue
			reduce(108), // ^^=, reduce: PositionalFieldValue
			reduce(108), // &&=, reduce: PositionalFieldValue
			reduce(108), // ??=, reduce: PositionalFieldValue
			reduce(108), // ???=, reduce: PositionalFieldValue
			reduce(108), // |=, reduce: PositionalFieldValue
			reduce(108), // &=, reduce: PositionalFieldValue
			reduce(108), // ^=, reduce: PositionalFieldValue
			reduce(108), // <<=, reduce: PositionalFieldValue
			reduce(108), // >>=, reduce: PositionalFieldValue
			reduce(108), // >>>=, reduce: PositionalFieldValue
			reduce(108), // +=, reduce: PositionalFieldValue
			reduce(108), // .=, reduce: PositionalFieldValue
			reduce(108), // -=, reduce: PositionalFieldValue
			reduce(108), // *=, reduce: PositionalFieldValue
			reduce(108), // /=, reduce: PositionalFieldValue
			reduce(108), // //=, reduce: PositionalFieldValue
			reduce(108), // %=, reduce: PositionalFieldValue
			reduce(108), // **=, reduce: PositionalFieldValue
			reduce(108), // ?, reduce: PositionalFieldValue
			nil,         // :
			reduce(108), // ||, reduce: PositionalFieldValue
			reduce(108), // ^^, reduce: PositionalFieldValue
			reduce(108), // &&, reduce: PositionalFieldValue
			reduce(108), // ??, reduce: PositionalFieldValue
			reduce(108), // ???, reduce: PositionalFieldValue
			reduce(108), // =~, reduce: PositionalFieldValue
			reduce(108), // !=~, reduce: PositionalFieldValue
			reduce(108), // ==, reduce: PositionalFieldValue
			reduce(108), // !=, reduce: PositionalFieldValue
			reduce(108), // <=>, reduce: PositionalFieldValue
			reduce(108), // >=, reduce: PositionalFieldValue
			reduce(108), // <, reduce: PositionalFieldValue
			reduce(108), // <=, reduce: PositionalFieldValue
			reduce(108), // ^, reduce: PositionalFieldValue
			reduce(108), // &, reduce: PositionalFieldValue
			reduce(108), // <<, reduce: PositionalFieldValue
			reduce(108), // >>>, reduce: PositionalFieldValue
			reduce(108), // +, reduce: PositionalFieldValue
			reduce(108), // -, reduce: PositionalFieldValue
			reduce(108), // .+, reduce: PositionalFieldValue
			reduce(108), // .-, reduce: PositionalFieldValue
			reduce(108), // *, reduce: PositionalFieldValue
			reduce(108), // /, reduce: PositionalFieldValue
			reduce(108), // //, reduce: PositionalFieldValue
			reduce(108), // %, reduce: PositionalFieldValue
			reduce(108), // .*, reduce: PositionalFieldValue
			reduce(108), // ./, reduce: PositionalFieldValue
			reduce(108), // .//, reduce: PositionalFieldValue
			reduce(108), // ., reduce: PositionalFieldValue
			nil,         // !
			nil,         // ~
			reduce(108), // **, reduce: PositionalFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(108), // [, reduce: PositionalFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(108), // [[, reduce: PositionalFieldValue
			reduce(108), // [[[, reduce: PositionalFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4368
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(149), // =, reduce: TernaryTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(149), // ||=, reduce: TernaryTerm
			reduce(149), // ^^=, reduce: TernaryTerm
			reduce(149), // &&=, reduce: TernaryTerm
			reduce(149), // ??=, reduce: TernaryTerm
			reduce(149), // ???=, reduce: TernaryTerm
			reduce(149), // |=, reduce: TernaryTerm
			reduce(149), // &=, reduce: TernaryTerm
			reduce(149), // ^=, reduce: TernaryTerm
			reduce(149), // <<=, reduce: TernaryTerm
			reduce(149), // >>=, reduce: TernaryTerm
			reduce(149), // >>>=, reduce: TernaryTerm
			reduce(149), // +=, reduce: TernaryTerm
			reduce(149), // .=, reduce: TernaryTerm
			reduce(149), // -=, reduce: TernaryTerm
			reduce(149), // *=, reduce: TernaryTerm
			reduce(149), // /=, reduce: TernaryTerm
			reduce(149), // //=, reduce: TernaryTerm
			reduce(149), // %=, reduce: TernaryTerm
			reduce(149), // **=, reduce: TernaryTerm
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4369
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4529), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4370
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(265), // =, reduce: ArraySliceAccess
			nil,         // unset
			nil,         // filter
			reduce(265), // >, reduce: ArraySliceAccess
			reduce(265), // >>, reduce: ArraySliceAccess
			reduce(265), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(265), // ||=, reduce: ArraySliceAccess
			reduce(265), // ^^=, reduce: ArraySliceAccess
			reduce(265), // &&=, reduce: ArraySliceAccess
			reduce(265), // ??=, reduce: ArraySliceAccess
			reduce(265), // ???=, reduce: ArraySliceAccess
			reduce(265), // |=, reduce: ArraySliceAccess
			reduce(265), // &=, reduce: ArraySliceAccess
			reduce(265), // ^=, reduce: ArraySliceAccess
			reduce(265), // <<=, reduce: ArraySliceAccess
			reduce(265), // >>=, reduce: ArraySliceAccess
			reduce(265), // >>>=, reduce: ArraySliceAccess
			reduce(265), // +=, reduce: ArraySliceAccess
			reduce(265), // .=, reduce: ArraySliceAccess
			reduce(265), // -=, reduce: ArraySliceAccess
			reduce(265), // *=, reduce: ArraySliceAccess
			reduce(265), // /=, reduce: ArraySliceAccess
			reduce(265), // //=, reduce: ArraySliceAccess
			reduce(265), // %=, reduce: ArraySliceAccess
			reduce(265), // **=, reduce: ArraySliceAccess
			reduce(265), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(265), // ||, reduce: ArraySliceAccess
			reduce(265), // ^^, reduce: ArraySliceAccess
			reduce(265), // &&, reduce: ArraySliceAccess
			reduce(265), // ??, reduce: ArraySliceAccess
			reduce(265), // ???, reduce: ArraySliceAccess
			reduce(265), // =~, reduce: ArraySliceAccess
			reduce(265), // !=~, reduce: ArraySliceAccess
			reduce(265), // ==, reduce: ArraySliceAccess
			reduce(265), // !=, reduce: ArraySliceAccess
			reduce(265), // <=>, reduce: ArraySliceAccess
			reduce(265), // >=, reduce: ArraySliceAccess
			reduce(265), // <, reduce: ArraySliceAccess
			reduce(265), // <=, reduce: ArraySliceAccess
			reduce(265), // ^, reduce: ArraySliceAccess
			reduce(265), // &, reduce: ArraySliceAccess
			reduce(265), // <<, reduce: ArraySliceAccess
			reduce(265), // >>>, reduce: ArraySliceAccess
			reduce(265), // +, reduce: ArraySliceAccess
			reduce(265), // -, reduce: ArraySliceAccess
			reduce(265), // .+, reduce: ArraySliceAccess
			reduce(265), // .-, reduce: ArraySliceAccess
			reduce(265), // *, reduce: ArraySliceAccess
			reduce(265), // /, reduce: ArraySliceAccess
			reduce(265), // //, reduce: ArraySliceAccess
			reduce(265), // %, reduce: ArraySliceAccess
			reduce(265), // .*, reduce: ArraySliceAccess
			reduce(265), // ./, reduce: ArraySliceAccess
			reduce(265), // .//, reduce: ArraySliceAccess
			reduce(265), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(265), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(265), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(265), // [[, reduce: ArraySliceAccess
			reduce(265), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4371
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(264), // =, reduce: ArraySliceAccess
			nil,         // unset
			nil,         // filter
			reduce(264), // >, reduce: ArraySliceAccess
			reduce(264), // >>, reduce: ArraySliceAccess
			reduce(264), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(264), // ||=, reduce: ArraySliceAccess
			reduce(264), // ^^=, reduce: ArraySliceAccess
			reduce(264), // &&=, reduce: ArraySliceAccess
			reduce(264), // ??=, reduce: ArraySliceAccess
			reduce(264), // ???=, reduce: ArraySliceAccess
			reduce(264), // |=, reduce: ArraySliceAccess
			reduce(264), // &=, reduce: ArraySliceAccess
			reduce(264), // ^=, reduce: ArraySliceAccess
			reduce(264), // <<=, reduce: ArraySliceAccess
			reduce(264), // >>=, reduce: ArraySliceAccess
			reduce(264), // >>>=, reduce: ArraySliceAccess
			reduce(264), // +=, reduce: ArraySliceAccess
			reduce(264), // .=, reduce: ArraySliceAccess
			reduce(264), // -=, reduce: ArraySliceAccess
			reduce(264), // *=, reduce: ArraySliceAccess
			reduce(264), // /=, reduce: ArraySliceAccess
			reduce(264), // //=, reduce: ArraySliceAccess
			reduce(264), // %=, reduce: ArraySliceAccess
			reduce(264), // **=, reduce: ArraySliceAccess
			reduce(264), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(264), // ||, reduce: ArraySliceAccess
			reduce(264), // ^^, reduce: ArraySliceAccess
			reduce(264), // &&, reduce: ArraySliceAccess
			reduce(264), // ??, reduce: ArraySliceAccess
			reduce(264), // ???, reduce: ArraySliceAccess
			reduce(264), // =~, reduce: ArraySliceAccess
			reduce(264), // !=~, reduce: ArraySliceAccess
			reduce(264), // ==, reduce: ArraySliceAccess
			reduce(264), // !=, reduce: ArraySliceAccess
			reduce(264), // <=>, reduce: ArraySliceAccess
			reduce(264), // >=, reduce: ArraySliceAccess
			reduce(264), // <, reduce: ArraySliceAccess
			reduce(264), // <=, reduce: ArraySliceAccess
			reduce(264), // ^, reduce: ArraySliceAccess
			reduce(264), // &, reduce: ArraySliceAccess
			reduce(264), // <<, reduce: ArraySliceAccess
			reduce(264), // >>>, reduce: ArraySliceAccess
			reduce(264), // +, reduce: ArraySliceAccess
			reduce(264), // -, reduce: ArraySliceAccess
			reduce(264), // .+, reduce: ArraySliceAccess
			reduce(264), // .-, reduce: ArraySliceAccess
			reduce(264), // *, reduce: ArraySliceAccess
			reduce(264), // /, reduce: ArraySliceAccess
			reduce(264), // //, reduce: ArraySliceAccess
			reduce(264), // %, reduce: ArraySliceAccess
			reduce(264), // .*, reduce: ArraySliceAccess
			reduce(264), // ./, reduce: ArraySliceAccess
			reduce(264), // .//, reduce: ArraySliceAccess
			reduce(264), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(264), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(264), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(264), // [[, reduce: ArraySliceAccess
			reduce(264), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4372
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(261), // =, reduce: ArrayOrMapPositionalNameAccess
			nil,         // unset
			nil,         // filter
			reduce(261), // >, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |, reduce: ArrayOrMapPositionalNameAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(261), // ||=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // **=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ?, reduce: ArrayOrMapPositionalNameAccess
			nil,         // :
			reduce(261), // ||, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // =~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ==, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .+, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .-, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .*, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ./, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .//, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ., reduce: ArrayOrMapPositionalNameAccess
			nil,         // !
			nil,         // ~
			reduce(261), // **, reduce: ArrayOrMapPositionalNameAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(261), // [, reduce: ArrayOrMapPositionalNameAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(261), // [[, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // [[[, reduce: ArrayOrMapPositionalNameAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4373
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4530), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4374
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(235), // ;, reduce: MapLiteral
			nil,         // {
			nil,         // }
			reduce(235), // =, reduce: MapLiteral
			nil,         // unset
			nil,         // filter
			reduce(235), // >, reduce: MapLiteral
			reduce(235), // >>, reduce: MapLiteral
			reduce(235), // |, reduce: MapLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(235), // ,, reduce: MapLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(235), // ||=, reduce: MapLiteral
			reduce(235), // ^^=, reduce: MapLiteral
			reduce(235), // &&=, reduce: MapLiteral
			reduce(235), // ??=, reduce: MapLiteral
			reduce(235), // ???=, reduce: MapLiteral
			reduce(235), // |=, reduce: MapLiteral
			reduce(235), // &=, reduce: MapLiteral
			reduce(235), // ^=, reduce: MapLiteral
			reduce(235), // <<=, reduce: MapLiteral
			reduce(235), // >>=, reduce: MapLiteral
			reduce(235), // >>>=, reduce: MapLiteral
			reduce(235), // +=, reduce: MapLiteral
			reduce(235), // .=, reduce: MapLiteral
			reduce(235), // -=, reduce: MapLiteral
			reduce(235), // *=, reduce: MapLiteral
			reduce(235), // /=, reduce: MapLiteral
			reduce(235), // //=, reduce: MapLiteral
			reduce(235), // %=, reduce: MapLiteral
			reduce(235), // **=, reduce: MapLiteral
			reduce(235), // ?, reduce: MapLiteral
			nil,         // :
			reduce(235), // ||, reduce: MapLiteral
			reduce(235), // ^^, reduce: MapLiteral
			reduce(235), // &&, reduce: MapLiteral
			reduce(235), // ??, reduce: MapLiteral
			reduce(235), // ???, reduce: MapLiteral
			reduce(235), // =~, reduce: MapLiteral
			reduce(235), // !=~, reduce: MapLiteral
			reduce(235), // ==, reduce: MapLiteral
			reduce(235), // !=, reduce: MapLiteral
			reduce(235), // <=>, reduce: MapLiteral
			reduce(235), // >=, reduce: MapLiteral
			reduce(235), // <, reduce: MapLiteral
			reduce(235), // <=, reduce: MapLiteral
			reduce(235), // ^, reduce: MapLiteral
			reduce(235), // &, reduce: MapLiteral
			reduce(235), // <<, reduce: MapLiteral
			reduce(235), // >>>, reduce: MapLiteral
			reduce(235), // +, reduce: MapLiteral
			reduce(235), // -, reduce: MapLiteral
			reduce(235), // .+, reduce: MapLiteral
			reduce(235), // .-, reduce: MapLiteral
			reduce(235), // *, reduce: MapLiteral
			reduce(235), // /, reduce: MapLiteral
			reduce(235), // //, reduce: MapLiteral
			reduce(235), // %, reduce: MapLiteral
			reduce(235), // .*, reduce: MapLiteral
			reduce(235), // ./, reduce: MapLiteral
			reduce(235), // .//, reduce: MapLiteral
			reduce(235), // ., reduce: MapLiteral
			nil,         // !
			nil,         // ~
			reduce(235), // **, reduce: MapLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(235), // [, reduce: MapLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(235), // [[, reduce: MapLiteral
			reduce(235), // [[[, reduce: MapLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4375
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(208), // ;, reduce: PrecedenceChainEnd
			nil,         // {
			nil,         // }
			reduce(208), // =, reduce: PrecedenceChainEnd
			nil,         // unset
			nil,         // filter
			reduce(208), // >, reduce: PrecedenceChainEnd
			reduce(208), // >>, reduce: PrecedenceChainEnd
			reduce(208), // |, reduce: PrecedenceChainEnd
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(208), // ,, reduce: PrecedenceChainEnd
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(208), // ||=, reduce: PrecedenceChainEnd
			reduce(208), // ^^=, reduce: PrecedenceChainEnd
			reduce(208), // &&=, reduce: PrecedenceChainEnd
			reduce(208), // ??=, reduce: PrecedenceChainEnd
			reduce(208), // ???=, reduce: PrecedenceChainEnd
			reduce(208), // |=, reduce: PrecedenceChainEnd
			reduce(208), // &=, reduce: PrecedenceChainEnd
			reduce(208), // ^=, reduce: PrecedenceChainEnd
			reduce(208), // <<=, reduce: PrecedenceChainEnd
			reduce(208), // >>=, reduce: PrecedenceChainEnd
			reduce(208), // >>>=, reduce: PrecedenceChainEnd
			reduce(208), // +=, reduce: PrecedenceChainEnd
			reduce(208), // .=, reduce: PrecedenceChainEnd
			reduce(208), // -=, reduce: PrecedenceChainEnd
			reduce(208), // *=, reduce: PrecedenceChainEnd
			reduce(208), // /=, reduce: PrecedenceChainEnd
			reduce(208), // //=, reduce: PrecedenceChainEnd
			reduce(208), // %=, reduce: PrecedenceChainEnd
			reduce(208), // **=, reduce: PrecedenceChainEnd
			reduce(208), // ?, reduce: PrecedenceChainEnd
			nil,         // :
			reduce(208), // ||, reduce: PrecedenceChainEnd
			reduce(208), // ^^, reduce: PrecedenceChainEnd
			reduce(208), // &&, reduce: PrecedenceChainEnd
			reduce(208), // ??, reduce: PrecedenceChainEnd
			reduce(208), // ???, reduce: PrecedenceChainEnd
			reduce(208), // =~, reduce: PrecedenceChainEnd
			reduce(208), // !=~, reduce: PrecedenceChainEnd
			reduce(208), // ==, reduce: PrecedenceChainEnd
			reduce(208), // !=, reduce: PrecedenceChainEnd
			reduce(208), // <=>, reduce: PrecedenceChainEnd
			reduce(208), // >=, reduce: PrecedenceChainEnd
			reduce(208), // <, reduce: PrecedenceChainEnd
			reduce(208), // <=, reduce: PrecedenceChainEnd
			reduce(208), // ^, reduce: PrecedenceChainEnd
			reduce(208), // &, reduce: PrecedenceChainEnd
			reduce(208), // <<, reduce: PrecedenceChainEnd
			reduce(208), // >>>, reduce: PrecedenceChainEnd
			reduce(208), // +, reduce: PrecedenceChainEnd
			reduce(208), // -, reduce: PrecedenceChainEnd
			reduce(208), // .+, reduce: PrecedenceChainEnd
			reduce(208), // .-, reduce: PrecedenceChainEnd
			reduce(208), // *, reduce: PrecedenceChainEnd
			reduce(208), // /, reduce: PrecedenceChainEnd
			reduce(208), // //, reduce: PrecedenceChainEnd
			reduce(208), // %, reduce: PrecedenceChainEnd
			reduce(208), // .*, reduce: PrecedenceChainEnd
			reduce(208), // ./, reduce: PrecedenceChainEnd
			reduce(208), // .//, reduce: PrecedenceChainEnd
			reduce(208), // ., reduce: PrecedenceChainEnd
			nil,         // !
			nil,         // ~
			reduce(208), // **, reduce: PrecedenceChainEnd
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4376
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(105), // ;, reduce: IndirectFieldValue
			nil,         // {
			nil,         // }
			reduce(105), // =, reduce: IndirectFieldValue
			nil,         // unset
			nil,         // filter
			reduce(105), // >, reduce: IndirectFieldValue
			reduce(105), // >>, reduce: IndirectFieldValue
			reduce(105), // |, reduce: IndirectFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(105), // ,, reduce: IndirectFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(105), // ||=, reduce: IndirectFieldValue
			reduce(105), // ^^=, reduce: IndirectFieldValue
			reduce(105), // &&=, reduce: IndirectFieldValue
			reduce(105), // ??=, reduce: IndirectFieldValue
			reduce(105), // ???=, reduce: IndirectFieldValue
			reduce(105), // |=, reduce: IndirectFieldValue
			reduce(105), // &=, reduce: IndirectFieldValue
			reduce(105), // ^=, reduce: IndirectFieldValue
			reduce(105), // <<=, reduce: IndirectFieldValue
			reduce(105), // >>=, reduce: IndirectFieldValue
			reduce(105), // >>>=, reduce: IndirectFieldValue
			reduce(105), // +=, reduce: IndirectFieldValue
			reduce(105), // .=, reduce: IndirectFieldValue
			reduce(105), // -=, reduce: IndirectFieldValue
			reduce(105), // *=, reduce: IndirectFieldValue
			reduce(105), // /=, reduce: IndirectFieldValue
			reduce(105), // //=, reduce: IndirectFieldValue
			reduce(105), // %=, reduce: IndirectFieldValue
			reduce(105), // **=, reduce: IndirectFieldValue
			reduce(105), // ?, reduce: IndirectFieldValue
			nil,         // :
			reduce(105), // ||, reduce: IndirectFieldValue
			reduce(105), // ^^, reduce: IndirectFieldValue
			reduce(105), // &&, reduce: IndirectFieldValue
			reduce(105), // ??, reduce: IndirectFieldValue
			reduce(105), // ???, reduce: IndirectFieldValue
			reduce(105), // =~, reduce: IndirectFieldValue
			reduce(105), // !=~, reduce: IndirectFieldValue
			reduce(105), // ==, reduce: IndirectFieldValue
			reduce(105), // !=, reduce: IndirectFieldValue
			reduce(105), // <=>, reduce: IndirectFieldValue
			reduce(105), // >=, reduce: IndirectFieldValue
			reduce(105), // <, reduce: IndirectFieldValue
			reduce(105), // <=, reduce: IndirectFieldValue
			reduce(105), // ^, reduce: IndirectFieldValue
			reduce(105), // &, reduce: IndirectFieldValue
			reduce(105), // <<, reduce: IndirectFieldValue
			reduce(105), // >>>, reduce: IndirectFieldValue
			reduce(105), // +, reduce: IndirectFieldValue
			reduce(105), // -, reduce: IndirectFieldValue
			reduce(105), // .+, reduce: IndirectFieldValue
			reduce(105), // .-, reduce: IndirectFieldValue
			reduce(105), // *, reduce: IndirectFieldValue
			reduce(105), // /, reduce: IndirectFieldValue
			reduce(105), // //, reduce: IndirectFieldValue
			reduce(105), // %, reduce: IndirectFieldValue
			reduce(105), // .*, reduce: IndirectFieldValue
			reduce(105), // ./, reduce: IndirectFieldValue
			reduce(105), // .//, reduce: IndirectFieldValue
			reduce(105), // ., reduce: IndirectFieldValue
			nil,         // !
			nil,         // ~
			reduce(105), // **, reduce: IndirectFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(105), // [, reduce: IndirectFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(105), // [[, reduce: IndirectFieldValue
			reduce(105), // [[[, reduce: IndirectFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4377
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4531), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4378
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4532), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4379
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(114), // ;, reduce: IndirectOosvarValue
			nil,         // {
			nil,         // }
			reduce(114), // =, reduce: IndirectOosvarValue
			nil,         // unset
			nil,         // filter
			reduce(114), // >, reduce: IndirectOosvarValue
			reduce(114), // >>, reduce: IndirectOosvarValue
			reduce(114), // |, reduce: IndirectOosvarValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(114), // ,, reduce: IndirectOosvarValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(114), // ||=, reduce: IndirectOosvarValue
			reduce(114), // ^^=, reduce: IndirectOosvarValue
			reduce(114), // &&=, reduce: IndirectOosvarValue
			reduce(114), // ??=, reduce: IndirectOosvarValue
			reduce(114), // ???=, reduce: IndirectOosvarValue
			reduce(114), // |=, reduce: IndirectOosvarValue
			reduce(114), // &=, reduce: IndirectOosvarValue
			reduce(114), // ^=, reduce: IndirectOosvarValue
			reduce(114), // <<=, reduce: IndirectOosvarValue
			reduce(114), // >>=, reduce: IndirectOosvarValue
			reduce(114), // >>>=, reduce: IndirectOosvarValue
			reduce(114), // +=, reduce: IndirectOosvarValue
			reduce(114), // .=, reduce: IndirectOosvarValue
			reduce(114), // -=, reduce: IndirectOosvarValue
			reduce(114), // *=, reduce: IndirectOosvarValue
			reduce(114), // /=, reduce: IndirectOosvarValue
			reduce(114), // //=, reduce: IndirectOosvarValue
			reduce(114), // %=, reduce: IndirectOosvarValue
			reduce(114), // **=, reduce: IndirectOosvarValue
			reduce(114), // ?, reduce: IndirectOosvarValue
			nil,         // :
			reduce(114), // ||, reduce: IndirectOosvarValue
			reduce(114), // ^^, reduce: IndirectOosvarValue
			reduce(114), // &&, reduce: IndirectOosvarValue
			reduce(114), // ??, reduce: IndirectOosvarValue
			reduce(114), // ???, reduce: IndirectOosvarValue
			reduce(114), // =~, reduce: IndirectOosvarValue
			reduce(114), // !=~, reduce: IndirectOosvarValue
			reduce(114), // ==, reduce: IndirectOosvarValue
			reduce(114), // !=, reduce: IndirectOosvarValue
			reduce(114), // <=>, reduce: IndirectOosvarValue
			reduce(114), // >=, reduce: IndirectOosvarValue
			reduce(114), // <, reduce: IndirectOosvarValue
			reduce(114), // <=, reduce: IndirectOosvarValue
			reduce(114), // ^, reduce: IndirectOosvarValue
			reduce(114), // &, reduce: IndirectOosvarValue
			reduce(114), // <<, reduce: IndirectOosvarValue
			reduce(114), // >>>, reduce: IndirectOosvarValue
			reduce(114), // +, reduce: IndirectOosvarValue
			reduce(114), // -, reduce: IndirectOosvarValue
			reduce(114), // .+, reduce: IndirectOosvarValue
			reduce(114), // .-, reduce: IndirectOosvarValue
			reduce(114), // *, reduce: IndirectOosvarValue
			reduce(114), // /, reduce: IndirectOosvarValue
			reduce(114), // //, reduce: IndirectOosvarValue
			reduce(114), // %, reduce: IndirectOosvarValue
			reduce(114), // .*, reduce: IndirectOosvarValue
			reduce(114), // ./, reduce: IndirectOosvarValue
			reduce(114), // .//, reduce: IndirectOosvarValue
			reduce(114), // ., reduce: IndirectOosvarValue
			nil,         // !
			nil,         // ~
			reduce(114), // **, reduce: IndirectOosvarValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(114), // [, reduce: IndirectOosvarValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(114), // [[, reduce: IndirectOosvarValue
			reduce(114), // [[[, reduce: IndirectOosvarValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4380
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4533), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4381
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(151), // ;, reduce: LogicalOrTerm
			nil,         // {
			nil,         // }
			reduce(151), // =, reduce: LogicalOrTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(151), // ,, reduce: LogicalOrTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(151), // ||=, reduce: LogicalOrTerm
			reduce(151), // ^^=, reduce: LogicalOrTerm
			reduce(151), // &&=, reduce: LogicalOrTerm
			reduce(151), // ??=, reduce: LogicalOrTerm
			reduce(151), // ???=, reduce: LogicalOrTerm
			reduce(151), // |=, reduce: LogicalOrTerm
			reduce(151), // &=, reduce: LogicalOrTerm
			reduce(151), // ^=, reduce: LogicalOrTerm
			reduce(151), // <<=, reduce: LogicalOrTerm
			reduce(151), // >>=, reduce: LogicalOrTerm
			reduce(151), // >>>=, reduce: LogicalOrTerm
			reduce(151), // +=, reduce: LogicalOrTerm
			reduce(151), // .=, reduce: LogicalOrTerm
			reduce(151), // -=, reduce: LogicalOrTerm
			reduce(151), // *=, reduce: LogicalOrTerm
			reduce(151), // /=, reduce: LogicalOrTerm
			reduce(151), // //=, reduce: LogicalOrTerm
			reduce(151), // %=, reduce: LogicalOrTerm
			reduce(151), // **=, reduce: LogicalOrTerm
			reduce(151), // ?, reduce: LogicalOrTerm
			nil,         // :
			reduce(151), // ||, reduce: LogicalOrTerm
			shift(4120), // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4382
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(153), // ;, reduce: LogicalXORTerm
			nil,         // {
			nil,         // }
			reduce(153), // =, reduce: LogicalXORTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(153), // ,, reduce: LogicalXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(153), // ||=, reduce: LogicalXORTerm
			reduce(153), // ^^=, reduce: LogicalXORTerm
			reduce(153), // &&=, reduce: LogicalXORTerm
			reduce(153), // ??=, reduce: LogicalXORTerm
			reduce(153), // ???=, reduce: LogicalXORTerm
			reduce(153), // |=, reduce: LogicalXORTerm
			reduce(153), // &=, reduce: LogicalXORTerm
			reduce(153), // ^=, reduce: LogicalXORTerm
			reduce(153), // <<=, reduce: LogicalXORTerm
			reduce(153), // >>=, reduce: LogicalXORTerm
			reduce(153), // >>>=, reduce: LogicalXORTerm
			reduce(153), // +=, reduce: LogicalXORTerm
			reduce(153), // .=, reduce: LogicalXORTerm
			reduce(153), // -=, reduce: LogicalXORTerm
			reduce(153), // *=, reduce: LogicalXORTerm
			reduce(153), // /=, reduce: LogicalXORTerm
			reduce(153), // //=, reduce: LogicalXORTerm
			reduce(153), // %=, reduce: LogicalXORTerm
			reduce(153), // **=, reduce: LogicalXORTerm
			reduce(153), // ?, reduce: LogicalXORTerm
			nil,         // :
			reduce(153), // ||, reduce: LogicalXORTerm
			reduce(153), // ^^, reduce: LogicalXORTerm
			shift(4121), // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4383
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(155), // ;, reduce: LogicalAndTerm
			nil,         // {
			nil,         // }
			reduce(155), // =, reduce: LogicalAndTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(155), // ,, reduce: LogicalAndTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(155), // ||=, reduce: LogicalAndTerm
			reduce(155), // ^^=, reduce: LogicalAndTerm
			reduce(155), // &&=, reduce: LogicalAndTerm
			reduce(155), // ??=, reduce: LogicalAndTerm
			reduce(155), // ???=, reduce: LogicalAndTerm
			reduce(155), // |=, reduce: LogicalAndTerm
			reduce(155), // &=, reduce: LogicalAndTerm
			reduce(155), // ^=, reduce: LogicalAndTerm
			reduce(155), // <<=, reduce: LogicalAndTerm
			reduce(155), // >>=, reduce: LogicalAndTerm
			reduce(155), // >>>=, reduce: LogicalAndTerm
			reduce(155), // +=, reduce: LogicalAndTerm
			reduce(155), // .=, reduce: LogicalAndTerm
			reduce(155), // -=, reduce: LogicalAndTerm
			reduce(155), // *=, reduce: LogicalAndTerm
			reduce(155), // /=, reduce: LogicalAndTerm
			reduce(155), // //=, reduce: LogicalAndTerm
			reduce(155), // %=, reduce: LogicalAndTerm
			reduce(155), // **=, reduce: LogicalAndTerm
			reduce(155), // ?, reduce: LogicalAndTerm
			nil,         // :
			reduce(155), // ||, reduce: LogicalAndTerm
			reduce(155), // ^^, reduce: LogicalAndTerm
			reduce(155), // &&, reduce: LogicalAndTerm
			shift(4122), // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4384
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(157), // ;, reduce: AbsentCoalesceTerm
			nil,         // {
			nil,         // }
			reduce(157), // =, reduce: AbsentCoalesceTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(157), // ,, reduce: AbsentCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(157), // ||=, reduce: AbsentCoalesceTerm
			reduce(157), // ^^=, reduce: AbsentCoalesceTerm
			reduce(157), // &&=, reduce: AbsentCoalesceTerm
			reduce(157), // ??=, reduce: AbsentCoalesceTerm
			reduce(157), // ???=, reduce: AbsentCoalesceTerm
			reduce(157), // |=, reduce: AbsentCoalesceTerm
			reduce(157), // &=, reduce: AbsentCoalesceTerm
			reduce(157), // ^=, reduce: AbsentCoalesceTerm
			reduce(157), // <<=, reduce: AbsentCoalesceTerm
			reduce(157), // >>=, reduce: AbsentCoalesceTerm
			reduce(157), // >>>=, reduce: AbsentCoalesceTerm
			reduce(157), // +=, reduce: AbsentCoalesceTerm
			reduce(157), // .=, reduce: AbsentCoalesceTerm
			reduce(157), // -=, reduce: AbsentCoalesceTerm
			reduce(157), // *=, reduce: AbsentCoalesceTerm
			reduce(157), // /=, reduce: AbsentCoalesceTerm
			reduce(157), // //=, reduce: AbsentCoalesceTerm
			reduce(157), // %=, reduce: AbsentCoalesceTerm
			reduce(157), // **=, reduce: AbsentCoalesceTerm
			reduce(157), // ?, reduce: AbsentCoalesceTerm
			nil,         // :
			reduce(157), // ||, reduce: AbsentCoalesceTerm
			reduce(157), // ^^, reduce: AbsentCoalesceTerm
			reduce(157), // &&, reduce: AbsentCoalesceTerm
			reduce(157), // ??, reduce: AbsentCoalesceTerm
			shift(4123), // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4385
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(159), // ;, reduce: EmptyCoalesceTerm
			nil,         // {
			nil,         // }
			reduce(159), // =, reduce: EmptyCoalesceTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(159), // ,, reduce: EmptyCoalesceTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(159), // ||=, reduce: EmptyCoalesceTerm
			reduce(159), // ^^=, reduce: EmptyCoalesceTerm
			reduce(159), // &&=, reduce: EmptyCoalesceTerm
			reduce(159), // ??=, reduce: EmptyCoalesceTerm
			reduce(159), // ???=, reduce: EmptyCoalesceTerm
			reduce(159), // |=, reduce: EmptyCoalesceTerm
			reduce(159), // &=, reduce: EmptyCoalesceTerm
			reduce(159), // ^=, reduce: EmptyCoalesceTerm
			reduce(159), // <<=, reduce: EmptyCoalesceTerm
			reduce(159), // >>=, reduce: EmptyCoalesceTerm
			reduce(159), // >>>=, reduce: EmptyCoalesceTerm
			reduce(159), // +=, reduce: EmptyCoalesceTerm
			reduce(159), // .=, reduce: EmptyCoalesceTerm
			reduce(159), // -=, reduce: EmptyCoalesceTerm
			reduce(159), // *=, reduce: EmptyCoalesceTerm
			reduce(159), // /=, reduce: EmptyCoalesceTerm
			reduce(159), // //=, reduce: EmptyCoalesceTerm
			reduce(159), // %=, reduce: EmptyCoalesceTerm
			reduce(159), // **=, reduce: EmptyCoalesceTerm
			reduce(159), // ?, reduce: EmptyCoalesceTerm
			nil,         // :
			reduce(159), // ||, reduce: EmptyCoalesceTerm
			reduce(159), // ^^, reduce: EmptyCoalesceTerm
			reduce(159), // &&, reduce: EmptyCoalesceTerm
			reduce(159), // ??, reduce: EmptyCoalesceTerm
			reduce(159), // ???, reduce: EmptyCoalesceTerm
			shift(4124), // =~
			shift(4125), // !=~
			shift(4126), // ==
			shift(4127), // !=
			shift(4128), // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4386
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(161), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			reduce(161), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(4129), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(161), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(161), // ||=, reduce: EqneTerm
			reduce(161), // ^^=, reduce: EqneTerm
			reduce(161), // &&=, reduce: EqneTerm
			reduce(161), // ??=, reduce: EqneTerm
			reduce(161), // ???=, reduce: EqneTerm
			reduce(161), // |=, reduce: EqneTerm
			reduce(161), // &=, reduce: EqneTerm
			reduce(161), // ^=, reduce: EqneTerm
			reduce(161), // <<=, reduce: EqneTerm
			reduce(161), // >>=, reduce: EqneTerm
			reduce(161), // >>>=, reduce: EqneTerm
			reduce(161), // +=, reduce: EqneTerm
			reduce(161), // .=, reduce: EqneTerm
			reduce(161), // -=, reduce: EqneTerm
			reduce(161), // *=, reduce: EqneTerm
			reduce(161), // /=, reduce: EqneTerm
			reduce(161), // //=, reduce: EqneTerm
			reduce(161), // %=, reduce: EqneTerm
			reduce(161), // **=, reduce: EqneTerm
			reduce(161), // ?, reduce: EqneTerm
			nil,         // :
			reduce(161), // ||, reduce: EqneTerm
			reduce(161), // ^^, reduce: EqneTerm
			reduce(161), // &&, reduce: EqneTerm
			reduce(161), // ??, reduce: EqneTerm
			reduce(161), // ???, reduce: EqneTerm
			reduce(161), // =~, reduce: EqneTerm
			reduce(161), // !=~, reduce: EqneTerm
			reduce(161), // ==, reduce: EqneTerm
			reduce(161), // !=, reduce: EqneTerm
			reduce(161), // <=>, reduce: EqneTerm
			shift(4130), // >=
			shift(4131), // <
			shift(4132), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4387
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(162), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			reduce(162), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(4129), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(162), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(162), // ||=, reduce: EqneTerm
			reduce(162), // ^^=, reduce: EqneTerm
			reduce(162), // &&=, reduce: EqneTerm
			reduce(162), // ??=, reduce: EqneTerm
			reduce(162), // ???=, reduce: EqneTerm
			reduce(162), // |=, reduce: EqneTerm
			reduce(162), // &=, reduce: EqneTerm
			reduce(162), // ^=, reduce: EqneTerm
			reduce(162), // <<=, reduce: EqneTerm
			reduce(162), // >>=, reduce: EqneTerm
			reduce(162), // >>>=, reduce: EqneTerm
			reduce(162), // +=, reduce: EqneTerm
			reduce(162), // .=, reduce: EqneTerm
			reduce(162), // -=, reduce: EqneTerm
			reduce(162), // *=, reduce: EqneTerm
			reduce(162), // /=, reduce: EqneTerm
			reduce(162), // //=, reduce: EqneTerm
			reduce(162), // %=, reduce: EqneTerm
			reduce(162), // **=, reduce: EqneTerm
			reduce(162), // ?, reduce: EqneTerm
			nil,         // :
			reduce(162), // ||, reduce: EqneTerm
			reduce(162), // ^^, reduce: EqneTerm
			reduce(162), // &&, reduce: EqneTerm
			reduce(162), // ??, reduce: EqneTerm
			reduce(162), // ???, reduce: EqneTerm
			reduce(162), // =~, reduce: EqneTerm
			reduce(162), // !=~, reduce: EqneTerm
			reduce(162), // ==, reduce: EqneTerm
			reduce(162), // !=, reduce: EqneTerm
			reduce(162), // <=>, reduce: EqneTerm
			shift(4130), // >=
			shift(4131), // <
			shift(4132), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4388
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(163), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			reduce(163), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(4129), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(163), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(163), // ||=, reduce: EqneTerm
			reduce(163), // ^^=, reduce: EqneTerm
			reduce(163), // &&=, reduce: EqneTerm
			reduce(163), // ??=, reduce: EqneTerm
			reduce(163), // ???=, reduce: EqneTerm
			reduce(163), // |=, reduce: EqneTerm
			reduce(163), // &=, reduce: EqneTerm
			reduce(163), // ^=, reduce: EqneTerm
			reduce(163), // <<=, reduce: EqneTerm
			reduce(163), // >>=, reduce: EqneTerm
			reduce(163), // >>>=, reduce: EqneTerm
			reduce(163), // +=, reduce: EqneTerm
			reduce(163), // .=, reduce: EqneTerm
			reduce(163), // -=, reduce: EqneTerm
			reduce(163), // *=, reduce: EqneTerm
			reduce(163), // /=, reduce: EqneTerm
			reduce(163), // //=, reduce: EqneTerm
			reduce(163), // %=, reduce: EqneTerm
			reduce(163), // **=, reduce: EqneTerm
			reduce(163), // ?, reduce: EqneTerm
			nil,         // :
			reduce(163), // ||, reduce: EqneTerm
			reduce(163), // ^^, reduce: EqneTerm
			reduce(163), // &&, reduce: EqneTerm
			reduce(163), // ??, reduce: EqneTerm
			reduce(163), // ???, reduce: EqneTerm
			reduce(163), // =~, reduce: EqneTerm
			reduce(163), // !=~, reduce: EqneTerm
			reduce(163), // ==, reduce: EqneTerm
			reduce(163), // !=, reduce: EqneTerm
			reduce(163), // <=>, reduce: EqneTerm
			shift(4130), // >=
			shift(4131), // <
			shift(4132), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4389
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(164), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			reduce(164), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(4129), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(164), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(164), // ||=, reduce: EqneTerm
			reduce(164), // ^^=, reduce: EqneTerm
			reduce(164), // &&=, reduce: EqneTerm
			reduce(164), // ??=, reduce: EqneTerm
			reduce(164), // ???=, reduce: EqneTerm
			reduce(164), // |=, reduce: EqneTerm
			reduce(164), // &=, reduce: EqneTerm
			reduce(164), // ^=, reduce: EqneTerm
			reduce(164), // <<=, reduce: EqneTerm
			reduce(164), // >>=, reduce: EqneTerm
			reduce(164), // >>>=, reduce: EqneTerm
			reduce(164), // +=, reduce: EqneTerm
			reduce(164), // .=, reduce: EqneTerm
			reduce(164), // -=, reduce: EqneTerm
			reduce(164), // *=, reduce: EqneTerm
			reduce(164), // /=, reduce: EqneTerm
			reduce(164), // //=, reduce: EqneTerm
			reduce(164), // %=, reduce: EqneTerm
			reduce(164), // **=, reduce: EqneTerm
			reduce(164), // ?, reduce: EqneTerm
			nil,         // :
			reduce(164), // ||, reduce: EqneTerm
			reduce(164), // ^^, reduce: EqneTerm
			reduce(164), // &&, reduce: EqneTerm
			reduce(164), // ??, reduce: EqneTerm
			reduce(164), // ???, reduce: EqneTerm
			reduce(164), // =~, reduce: EqneTerm
			reduce(164), // !=~, reduce: EqneTerm
			reduce(164), // ==, reduce: EqneTerm
			reduce(164), // !=, reduce: EqneTerm
			reduce(164), // <=>, reduce: EqneTerm
			shift(4130), // >=
			shift(4131), // <
			shift(4132), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4390
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(165), // ;, reduce: EqneTerm
			nil,         // {
			nil,         // }
			reduce(165), // =, reduce: EqneTerm
			nil,         // unset
			nil,         // filter
			shift(4129), // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(165), // ,, reduce: EqneTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(165), // ||=, reduce: EqneTerm
			reduce(165), // ^^=, reduce: EqneTerm
			reduce(165), // &&=, reduce: EqneTerm
			reduce(165), // ??=, reduce: EqneTerm
			reduce(165), // ???=, reduce: EqneTerm
			reduce(165), // |=, reduce: EqneTerm
			reduce(165), // &=, reduce: EqneTerm
			reduce(165), // ^=, reduce: EqneTerm
			reduce(165), // <<=, reduce: EqneTerm
			reduce(165), // >>=, reduce: EqneTerm
			reduce(165), // >>>=, reduce: EqneTerm
			reduce(165), // +=, reduce: EqneTerm
			reduce(165), // .=, reduce: EqneTerm
			reduce(165), // -=, reduce: EqneTerm
			reduce(165), // *=, reduce: EqneTerm
			reduce(165), // /=, reduce: EqneTerm
			reduce(165), // //=, reduce: EqneTerm
			reduce(165), // %=, reduce: EqneTerm
			reduce(165), // **=, reduce: EqneTerm
			reduce(165), // ?, reduce: EqneTerm
			nil,         // :
			reduce(165), // ||, reduce: EqneTerm
			reduce(165), // ^^, reduce: EqneTerm
			reduce(165), // &&, reduce: EqneTerm
			reduce(165), // ??, reduce: EqneTerm
			reduce(165), // ???, reduce: EqneTerm
			reduce(165), // =~, reduce: EqneTerm
			reduce(165), // !=~, reduce: EqneTerm
			reduce(165), // ==, reduce: EqneTerm
			reduce(165), // !=, reduce: EqneTerm
			reduce(165), // <=>, reduce: EqneTerm
			shift(4130), // >=
			shift(4131), // <
			shift(4132), // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4391
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(167), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			reduce(167), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(167), // >, reduce: CmpTerm
			nil,         // >>
			shift(4133), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(167), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(167), // ||=, reduce: CmpTerm
			reduce(167), // ^^=, reduce: CmpTerm
			reduce(167), // &&=, reduce: CmpTerm
			reduce(167), // ??=, reduce: CmpTerm
			reduce(167), // ???=, reduce: CmpTerm
			reduce(167), // |=, reduce: CmpTerm
			reduce(167), // &=, reduce: CmpTerm
			reduce(167), // ^=, reduce: CmpTerm
			reduce(167), // <<=, reduce: CmpTerm
			reduce(167), // >>=, reduce: CmpTerm
			reduce(167), // >>>=, reduce: CmpTerm
			reduce(167), // +=, reduce: CmpTerm
			reduce(167), // .=, reduce: CmpTerm
			reduce(167), // -=, reduce: CmpTerm
			reduce(167), // *=, reduce: CmpTerm
			reduce(167), // /=, reduce: CmpTerm
			reduce(167), // //=, reduce: CmpTerm
			reduce(167), // %=, reduce: CmpTerm
			reduce(167), // **=, reduce: CmpTerm
			reduce(167), // ?, reduce: CmpTerm
			nil,         // :
			reduce(167), // ||, reduce: CmpTerm
			reduce(167), // ^^, reduce: CmpTerm
			reduce(167), // &&, reduce: CmpTerm
			reduce(167), // ??, reduce: CmpTerm
			reduce(167), // ???, reduce: CmpTerm
			reduce(167), // =~, reduce: CmpTerm
			reduce(167), // !=~, reduce: CmpTerm
			reduce(167), // ==, reduce: CmpTerm
			reduce(167), // !=, reduce: CmpTerm
			reduce(167), // <=>, reduce: CmpTerm
			reduce(167), // >=, reduce: CmpTerm
			reduce(167), // <, reduce: CmpTerm
			reduce(167), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4392
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(168), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			reduce(168), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(168), // >, reduce: CmpTerm
			nil,         // >>
			shift(4133), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(168), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(168), // ||=, reduce: CmpTerm
			reduce(168), // ^^=, reduce: CmpTerm
			reduce(168), // &&=, reduce: CmpTerm
			reduce(168), // ??=, reduce: CmpTerm
			reduce(168), // ???=, reduce: CmpTerm
			reduce(168), // |=, reduce: CmpTerm
			reduce(168), // &=, reduce: CmpTerm
			reduce(168), // ^=, reduce: CmpTerm
			reduce(168), // <<=, reduce: CmpTerm
			reduce(168), // >>=, reduce: CmpTerm
			reduce(168), // >>>=, reduce: CmpTerm
			reduce(168), // +=, reduce: CmpTerm
			reduce(168), // .=, reduce: CmpTerm
			reduce(168), // -=, reduce: CmpTerm
			reduce(168), // *=, reduce: CmpTerm
			reduce(168), // /=, reduce: CmpTerm
			reduce(168), // //=, reduce: CmpTerm
			reduce(168), // %=, reduce: CmpTerm
			reduce(168), // **=, reduce: CmpTerm
			reduce(168), // ?, reduce: CmpTerm
			nil,         // :
			reduce(168), // ||, reduce: CmpTerm
			reduce(168), // ^^, reduce: CmpTerm
			reduce(168), // &&, reduce: CmpTerm
			reduce(168), // ??, reduce: CmpTerm
			reduce(168), // ???, reduce: CmpTerm
			reduce(168), // =~, reduce: CmpTerm
			reduce(168), // !=~, reduce: CmpTerm
			reduce(168), // ==, reduce: CmpTerm
			reduce(168), // !=, reduce: CmpTerm
			reduce(168), // <=>, reduce: CmpTerm
			reduce(168), // >=, reduce: CmpTerm
			reduce(168), // <, reduce: CmpTerm
			reduce(168), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4393
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(169), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			reduce(169), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(169), // >, reduce: CmpTerm
			nil,         // >>
			shift(4133), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(169), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(169), // ||=, reduce: CmpTerm
			reduce(169), // ^^=, reduce: CmpTerm
			reduce(169), // &&=, reduce: CmpTerm
			reduce(169), // ??=, reduce: CmpTerm
			reduce(169), // ???=, reduce: CmpTerm
			reduce(169), // |=, reduce: CmpTerm
			reduce(169), // &=, reduce: CmpTerm
			reduce(169), // ^=, reduce: CmpTerm
			reduce(169), // <<=, reduce: CmpTerm
			reduce(169), // >>=, reduce: CmpTerm
			reduce(169), // >>>=, reduce: CmpTerm
			reduce(169), // +=, reduce: CmpTerm
			reduce(169), // .=, reduce: CmpTerm
			reduce(169), // -=, reduce: CmpTerm
			reduce(169), // *=, reduce: CmpTerm
			reduce(169), // /=, reduce: CmpTerm
			reduce(169), // //=, reduce: CmpTerm
			reduce(169), // %=, reduce: CmpTerm
			reduce(169), // **=, reduce: CmpTerm
			reduce(169), // ?, reduce: CmpTerm
			nil,         // :
			reduce(169), // ||, reduce: CmpTerm
			reduce(169), // ^^, reduce: CmpTerm
			reduce(169), // &&, reduce: CmpTerm
			reduce(169), // ??, reduce: CmpTerm
			reduce(169), // ???, reduce: CmpTerm
			reduce(169), // =~, reduce: CmpTerm
			reduce(169), // !=~, reduce: CmpTerm
			reduce(169), // ==, reduce: CmpTerm
			reduce(169), // !=, reduce: CmpTerm
			reduce(169), // <=>, reduce: CmpTerm
			reduce(169), // >=, reduce: CmpTerm
			reduce(169), // <, reduce: CmpTerm
			reduce(169), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4394
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(170), // ;, reduce: CmpTerm
			nil,         // {
			nil,         // }
			reduce(170), // =, reduce: CmpTerm
			nil,         // unset
			nil,         // filter
			reduce(170), // >, reduce: CmpTerm
			nil,         // >>
			shift(4133), // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(170), // ,, reduce: CmpTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(170), // ||=, reduce: CmpTerm
			reduce(170), // ^^=, reduce: CmpTerm
			reduce(170), // &&=, reduce: CmpTerm
			reduce(170), // ??=, reduce: CmpTerm
			reduce(170), // ???=, reduce: CmpTerm
			reduce(170), // |=, reduce: CmpTerm
			reduce(170), // &=, reduce: CmpTerm
			reduce(170), // ^=, reduce: CmpTerm
			reduce(170), // <<=, reduce: CmpTerm
			reduce(170), // >>=, reduce: CmpTerm
			reduce(170), // >>>=, reduce: CmpTerm
			reduce(170), // +=, reduce: CmpTerm
			reduce(170), // .=, reduce: CmpTerm
			reduce(170), // -=, reduce: CmpTerm
			reduce(170), // *=, reduce: CmpTerm
			reduce(170), // /=, reduce: CmpTerm
			reduce(170), // //=, reduce: CmpTerm
			reduce(170), // %=, reduce: CmpTerm
			reduce(170), // **=, reduce: CmpTerm
			reduce(170), // ?, reduce: CmpTerm
			nil,         // :
			reduce(170), // ||, reduce: CmpTerm
			reduce(170), // ^^, reduce: CmpTerm
			reduce(170), // &&, reduce: CmpTerm
			reduce(170), // ??, reduce: CmpTerm
			reduce(170), // ???, reduce: CmpTerm
			reduce(170), // =~, reduce: CmpTerm
			reduce(170), // !=~, reduce: CmpTerm
			reduce(170), // ==, reduce: CmpTerm
			reduce(170), // !=, reduce: CmpTerm
			reduce(170), // <=>, reduce: CmpTerm
			reduce(170), // >=, reduce: CmpTerm
			reduce(170), // <, reduce: CmpTerm
			reduce(170), // <=, reduce: CmpTerm
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4395
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(172), // ;, reduce: BitwiseORTerm
			nil,         // {
			nil,         // }
			reduce(172), // =, reduce: BitwiseORTerm
			nil,         // unset
			nil,         // filter
			reduce(172), // >, reduce: BitwiseORTerm
			nil,         // >>
			reduce(172), // |, reduce: BitwiseORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(172), // ,, reduce: BitwiseORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(172), // ||=, reduce: BitwiseORTerm
			reduce(172), // ^^=, reduce: BitwiseORTerm
			reduce(172), // &&=, reduce: BitwiseORTerm
			reduce(172), // ??=, reduce: BitwiseORTerm
			reduce(172), // ???=, reduce: BitwiseORTerm
			reduce(172), // |=, reduce: BitwiseORTerm
			reduce(172), // &=, reduce: BitwiseORTerm
			reduce(172), // ^=, reduce: BitwiseORTerm
			reduce(172), // <<=, reduce: BitwiseORTerm
			reduce(172), // >>=, reduce: BitwiseORTerm
			reduce(172), // >>>=, reduce: BitwiseORTerm
			reduce(172), // +=, reduce: BitwiseORTerm
			reduce(172), // .=, reduce: BitwiseORTerm
			reduce(172), // -=, reduce: BitwiseORTerm
			reduce(172), // *=, reduce: BitwiseORTerm
			reduce(172), // /=, reduce: BitwiseORTerm
			reduce(172), // //=, reduce: BitwiseORTerm
			reduce(172), // %=, reduce: BitwiseORTerm
			reduce(172), // **=, reduce: BitwiseORTerm
			reduce(172), // ?, reduce: BitwiseORTerm
			nil,         // :
			reduce(172), // ||, reduce: BitwiseORTerm
			reduce(172), // ^^, reduce: BitwiseORTerm
			reduce(172), // &&, reduce: BitwiseORTerm
			reduce(172), // ??, reduce: BitwiseORTerm
			reduce(172), // ???, reduce: BitwiseORTerm
			reduce(172), // =~, reduce: BitwiseORTerm
			reduce(172), // !=~, reduce: BitwiseORTerm
			reduce(172), // ==, reduce: BitwiseORTerm
			reduce(172), // !=, reduce: BitwiseORTerm
			reduce(172), // <=>, reduce: BitwiseORTerm
			reduce(172), // >=, reduce: BitwiseORTerm
			reduce(172), // <, reduce: BitwiseORTerm
			reduce(172), // <=, reduce: BitwiseORTerm
			shift(4134), // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4396
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(174), // ;, reduce: BitwiseXORTerm
			nil,         // {
			nil,         // }
			reduce(174), // =, reduce: BitwiseXORTerm
			nil,         // unset
			nil,         // filter
			reduce(174), // >, reduce: BitwiseXORTerm
			nil,         // >>
			reduce(174), // |, reduce: BitwiseXORTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(174), // ,, reduce: BitwiseXORTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(174), // ||=, reduce: BitwiseXORTerm
			reduce(174), // ^^=, reduce: BitwiseXORTerm
			reduce(174), // &&=, reduce: BitwiseXORTerm
			reduce(174), // ??=, reduce: BitwiseXORTerm
			reduce(174), // ???=, reduce: BitwiseXORTerm
			reduce(174), // |=, reduce: BitwiseXORTerm
			reduce(174), // &=, reduce: BitwiseXORTerm
			reduce(174), // ^=, reduce: BitwiseXORTerm
			reduce(174), // <<=, reduce: BitwiseXORTerm
			reduce(174), // >>=, reduce: BitwiseXORTerm
			reduce(174), // >>>=, reduce: BitwiseXORTerm
			reduce(174), // +=, reduce: BitwiseXORTerm
			reduce(174), // .=, reduce: BitwiseXORTerm
			reduce(174), // -=, reduce: BitwiseXORTerm
			reduce(174), // *=, reduce: BitwiseXORTerm
			reduce(174), // /=, reduce: BitwiseXORTerm
			reduce(174), // //=, reduce: BitwiseXORTerm
			reduce(174), // %=, reduce: BitwiseXORTerm
			reduce(174), // **=, reduce: BitwiseXORTerm
			reduce(174), // ?, reduce: BitwiseXORTerm
			nil,         // :
			reduce(174), // ||, reduce: BitwiseXORTerm
			reduce(174), // ^^, reduce: BitwiseXORTerm
			reduce(174), // &&, reduce: BitwiseXORTerm
			reduce(174), // ??, reduce: BitwiseXORTerm
			reduce(174), // ???, reduce: BitwiseXORTerm
			reduce(174), // =~, reduce: BitwiseXORTerm
			reduce(174), // !=~, reduce: BitwiseXORTerm
			reduce(174), // ==, reduce: BitwiseXORTerm
			reduce(174), // !=, reduce: BitwiseXORTerm
			reduce(174), // <=>, reduce: BitwiseXORTerm
			reduce(174), // >=, reduce: BitwiseXORTerm
			reduce(174), // <, reduce: BitwiseXORTerm
			reduce(174), // <=, reduce: BitwiseXORTerm
			reduce(174), // ^, reduce: BitwiseXORTerm
			shift(4135), // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4397
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(176), // ;, reduce: BitwiseANDTerm
			nil,         // {
			nil,         // }
			reduce(176), // =, reduce: BitwiseANDTerm
			nil,         // unset
			nil,         // filter
			reduce(176), // >, reduce: BitwiseANDTerm
			shift(4136), // >>
			reduce(176), // |, reduce: BitwiseANDTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(176), // ,, reduce: BitwiseANDTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(176), // ||=, reduce: BitwiseANDTerm
			reduce(176), // ^^=, reduce: BitwiseANDTerm
			reduce(176), // &&=, reduce: BitwiseANDTerm
			reduce(176), // ??=, reduce: BitwiseANDTerm
			reduce(176), // ???=, reduce: BitwiseANDTerm
			reduce(176), // |=, reduce: BitwiseANDTerm
			reduce(176), // &=, reduce: BitwiseANDTerm
			reduce(176), // ^=, reduce: BitwiseANDTerm
			reduce(176), // <<=, reduce: BitwiseANDTerm
			reduce(176), // >>=, reduce: BitwiseANDTerm
			reduce(176), // >>>=, reduce: BitwiseANDTerm
			reduce(176), // +=, reduce: BitwiseANDTerm
			reduce(176), // .=, reduce: BitwiseANDTerm
			reduce(176), // -=, reduce: BitwiseANDTerm
			reduce(176), // *=, reduce: BitwiseANDTerm
			reduce(176), // /=, reduce: BitwiseANDTerm
			reduce(176), // //=, reduce: BitwiseANDTerm
			reduce(176), // %=, reduce: BitwiseANDTerm
			reduce(176), // **=, reduce: BitwiseANDTerm
			reduce(176), // ?, reduce: BitwiseANDTerm
			nil,         // :
			reduce(176), // ||, reduce: BitwiseANDTerm
			reduce(176), // ^^, reduce: BitwiseANDTerm
			reduce(176), // &&, reduce: BitwiseANDTerm
			reduce(176), // ??, reduce: BitwiseANDTerm
			reduce(176), // ???, reduce: BitwiseANDTerm
			reduce(176), // =~, reduce: BitwiseANDTerm
			reduce(176), // !=~, reduce: BitwiseANDTerm
			reduce(176), // ==, reduce: BitwiseANDTerm
			reduce(176), // !=, reduce: BitwiseANDTerm
			reduce(176), // <=>, reduce: BitwiseANDTerm
			reduce(176), // >=, reduce: BitwiseANDTerm
			reduce(176), // <, reduce: BitwiseANDTerm
			reduce(176), // <=, reduce: BitwiseANDTerm
			reduce(176), // ^, reduce: BitwiseANDTerm
			reduce(176), // &, reduce: BitwiseANDTerm
			shift(4137), // <<
			shift(4138), // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4398
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(179), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			nil,         // }
			reduce(179), // =, reduce: BitwiseShiftTerm
			nil,         // unset
			nil,         // filter
			reduce(179), // >, reduce: BitwiseShiftTerm
			reduce(179), // >>, reduce: BitwiseShiftTerm
			reduce(179), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(179), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(179), // ||=, reduce: BitwiseShiftTerm
			reduce(179), // ^^=, reduce: BitwiseShiftTerm
			reduce(179), // &&=, reduce: BitwiseShiftTerm
			reduce(179), // ??=, reduce: BitwiseShiftTerm
			reduce(179), // ???=, reduce: BitwiseShiftTerm
			reduce(179), // |=, reduce: BitwiseShiftTerm
			reduce(179), // &=, reduce: BitwiseShiftTerm
			reduce(179), // ^=, reduce: BitwiseShiftTerm
			reduce(179), // <<=, reduce: BitwiseShiftTerm
			reduce(179), // >>=, reduce: BitwiseShiftTerm
			reduce(179), // >>>=, reduce: BitwiseShiftTerm
			reduce(179), // +=, reduce: BitwiseShiftTerm
			reduce(179), // .=, reduce: BitwiseShiftTerm
			reduce(179), // -=, reduce: BitwiseShiftTerm
			reduce(179), // *=, reduce: BitwiseShiftTerm
			reduce(179), // /=, reduce: BitwiseShiftTerm
			reduce(179), // //=, reduce: BitwiseShiftTerm
			reduce(179), // %=, reduce: BitwiseShiftTerm
			reduce(179), // **=, reduce: BitwiseShiftTerm
			reduce(179), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(179), // ||, reduce: BitwiseShiftTerm
			reduce(179), // ^^, reduce: BitwiseShiftTerm
			reduce(179), // &&, reduce: BitwiseShiftTerm
			reduce(179), // ??, reduce: BitwiseShiftTerm
			reduce(179), // ???, reduce: BitwiseShiftTerm
			reduce(179), // =~, reduce: BitwiseShiftTerm
			reduce(179), // !=~, reduce: BitwiseShiftTerm
			reduce(179), // ==, reduce: BitwiseShiftTerm
			reduce(179), // !=, reduce: BitwiseShiftTerm
			reduce(179), // <=>, reduce: BitwiseShiftTerm
			reduce(179), // >=, reduce: BitwiseShiftTerm
			reduce(179), // <, reduce: BitwiseShiftTerm
			reduce(179), // <=, reduce: BitwiseShiftTerm
			reduce(179), // ^, reduce: BitwiseShiftTerm
			reduce(179), // &, reduce: BitwiseShiftTerm
			reduce(179), // <<, reduce: BitwiseShiftTerm
			reduce(179), // >>>, reduce: BitwiseShiftTerm
			shift(4139), // +
			shift(4140), // -
			shift(4141), // .+
			shift(4142), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4399
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(178), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			nil,         // }
			reduce(178), // =, reduce: BitwiseShiftTerm
			nil,         // unset
			nil,         // filter
			reduce(178), // >, reduce: BitwiseShiftTerm
			reduce(178), // >>, reduce: BitwiseShiftTerm
			reduce(178), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(178), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(178), // ||=, reduce: BitwiseShiftTerm
			reduce(178), // ^^=, reduce: BitwiseShiftTerm
			reduce(178), // &&=, reduce: BitwiseShiftTerm
			reduce(178), // ??=, reduce: BitwiseShiftTerm
			reduce(178), // ???=, reduce: BitwiseShiftTerm
			reduce(178), // |=, reduce: BitwiseShiftTerm
			reduce(178), // &=, reduce: BitwiseShiftTerm
			reduce(178), // ^=, reduce: BitwiseShiftTerm
			reduce(178), // <<=, reduce: BitwiseShiftTerm
			reduce(178), // >>=, reduce: BitwiseShiftTerm
			reduce(178), // >>>=, reduce: BitwiseShiftTerm
			reduce(178), // +=, reduce: BitwiseShiftTerm
			reduce(178), // .=, reduce: BitwiseShiftTerm
			reduce(178), // -=, reduce: BitwiseShiftTerm
			reduce(178), // *=, reduce: BitwiseShiftTerm
			reduce(178), // /=, reduce: BitwiseShiftTerm
			reduce(178), // //=, reduce: BitwiseShiftTerm
			reduce(178), // %=, reduce: BitwiseShiftTerm
			reduce(178), // **=, reduce: BitwiseShiftTerm
			reduce(178), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(178), // ||, reduce: BitwiseShiftTerm
			reduce(178), // ^^, reduce: BitwiseShiftTerm
			reduce(178), // &&, reduce: BitwiseShiftTerm
			reduce(178), // ??, reduce: BitwiseShiftTerm
			reduce(178), // ???, reduce: BitwiseShiftTerm
			reduce(178), // =~, reduce: BitwiseShiftTerm
			reduce(178), // !=~, reduce: BitwiseShiftTerm
			reduce(178), // ==, reduce: BitwiseShiftTerm
			reduce(178), // !=, reduce: BitwiseShiftTerm
			reduce(178), // <=>, reduce: BitwiseShiftTerm
			reduce(178), // >=, reduce: BitwiseShiftTerm
			reduce(178), // <, reduce: BitwiseShiftTerm
			reduce(178), // <=, reduce: BitwiseShiftTerm
			reduce(178), // ^, reduce: BitwiseShiftTerm
			reduce(178), // &, reduce: BitwiseShiftTerm
			reduce(178), // <<, reduce: BitwiseShiftTerm
			reduce(178), // >>>, reduce: BitwiseShiftTerm
			shift(4139), // +
			shift(4140), // -
			shift(4141), // .+
			shift(4142), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4400
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(180), // ;, reduce: BitwiseShiftTerm
			nil,         // {
			nil,         // }
			reduce(180), // =, reduce: BitwiseShiftTerm
			nil,         // unset
			nil,         // filter
			reduce(180), // >, reduce: BitwiseShiftTerm
			reduce(180), // >>, reduce: BitwiseShiftTerm
			reduce(180), // |, reduce: BitwiseShiftTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(180), // ,, reduce: BitwiseShiftTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(180), // ||=, reduce: BitwiseShiftTerm
			reduce(180), // ^^=, reduce: BitwiseShiftTerm
			reduce(180), // &&=, reduce: BitwiseShiftTerm
			reduce(180), // ??=, reduce: BitwiseShiftTerm
			reduce(180), // ???=, reduce: BitwiseShiftTerm
			reduce(180), // |=, reduce: BitwiseShiftTerm
			reduce(180), // &=, reduce: BitwiseShiftTerm
			reduce(180), // ^=, reduce: BitwiseShiftTerm
			reduce(180), // <<=, reduce: BitwiseShiftTerm
			reduce(180), // >>=, reduce: BitwiseShiftTerm
			reduce(180), // >>>=, reduce: BitwiseShiftTerm
			reduce(180), // +=, reduce: BitwiseShiftTerm
			reduce(180), // .=, reduce: BitwiseShiftTerm
			reduce(180), // -=, reduce: BitwiseShiftTerm
			reduce(180), // *=, reduce: BitwiseShiftTerm
			reduce(180), // /=, reduce: BitwiseShiftTerm
			reduce(180), // //=, reduce: BitwiseShiftTerm
			reduce(180), // %=, reduce: BitwiseShiftTerm
			reduce(180), // **=, reduce: BitwiseShiftTerm
			reduce(180), // ?, reduce: BitwiseShiftTerm
			nil,         // :
			reduce(180), // ||, reduce: BitwiseShiftTerm
			reduce(180), // ^^, reduce: BitwiseShiftTerm
			reduce(180), // &&, reduce: BitwiseShiftTerm
			reduce(180), // ??, reduce: BitwiseShiftTerm
			reduce(180), // ???, reduce: BitwiseShiftTerm
			reduce(180), // =~, reduce: BitwiseShiftTerm
			reduce(180), // !=~, reduce: BitwiseShiftTerm
			reduce(180), // ==, reduce: BitwiseShiftTerm
			reduce(180), // !=, reduce: BitwiseShiftTerm
			reduce(180), // <=>, reduce: BitwiseShiftTerm
			reduce(180), // >=, reduce: BitwiseShiftTerm
			reduce(180), // <, reduce: BitwiseShiftTerm
			reduce(180), // <=, reduce: BitwiseShiftTerm
			reduce(180), // ^, reduce: BitwiseShiftTerm
			reduce(180), // &, reduce: BitwiseShiftTerm
			reduce(180), // <<, reduce: BitwiseShiftTerm
			reduce(180), // >>>, reduce: BitwiseShiftTerm
			shift(4139), // +
			shift(4140), // -
			shift(4141), // .+
			shift(4142), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4401
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(182), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			reduce(182), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(182), // >, reduce: AddsubdotTerm
			reduce(182), // >>, reduce: AddsubdotTerm
			reduce(182), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(182), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(182), // ||=, reduce: AddsubdotTerm
			reduce(182), // ^^=, reduce: AddsubdotTerm
			reduce(182), // &&=, reduce: AddsubdotTerm
			reduce(182), // ??=, reduce: AddsubdotTerm
			reduce(182), // ???=, reduce: AddsubdotTerm
			reduce(182), // |=, reduce: AddsubdotTerm
			reduce(182), // &=, reduce: AddsubdotTerm
			reduce(182), // ^=, reduce: AddsubdotTerm
			reduce(182), // <<=, reduce: AddsubdotTerm
			reduce(182), // >>=, reduce: AddsubdotTerm
			reduce(182), // >>>=, reduce: AddsubdotTerm
			reduce(182), // +=, reduce: AddsubdotTerm
			reduce(182), // .=, reduce: AddsubdotTerm
			reduce(182), // -=, reduce: AddsubdotTerm
			reduce(182), // *=, reduce: AddsubdotTerm
			reduce(182), // /=, reduce: AddsubdotTerm
			reduce(182), // //=, reduce: AddsubdotTerm
			reduce(182), // %=, reduce: AddsubdotTerm
			reduce(182), // **=, reduce: AddsubdotTerm
			reduce(182), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(182), // ||, reduce: AddsubdotTerm
			reduce(182), // ^^, reduce: AddsubdotTerm
			reduce(182), // &&, reduce: AddsubdotTerm
			reduce(182), // ??, reduce: AddsubdotTerm
			reduce(182), // ???, reduce: AddsubdotTerm
			reduce(182), // =~, reduce: AddsubdotTerm
			reduce(182), // !=~, reduce: AddsubdotTerm
			reduce(182), // ==, reduce: AddsubdotTerm
			reduce(182), // !=, reduce: AddsubdotTerm
			reduce(182), // <=>, reduce: AddsubdotTerm
			reduce(182), // >=, reduce: AddsubdotTerm
			reduce(182), // <, reduce: AddsubdotTerm
			reduce(182), // <=, reduce: AddsubdotTerm
			reduce(182), // ^, reduce: AddsubdotTerm
			reduce(182), // &, reduce: AddsubdotTerm
			reduce(182), // <<, reduce: AddsubdotTerm
			reduce(182), // >>>, reduce: AddsubdotTerm
			reduce(182), // +, reduce: AddsubdotTerm
			reduce(182), // -, reduce: AddsubdotTerm
			reduce(182), // .+, reduce: AddsubdotTerm
			reduce(182), // .-, reduce: AddsubdotTerm
			shift(4144), // *
			shift(4145), // /
			shift(4146), // //
			shift(4147), // %
			shift(4148), // .*
			shift(4149), // ./
			shift(4150), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4402
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(183), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			reduce(183), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(183), // >, reduce: AddsubdotTerm
			reduce(183), // >>, reduce: AddsubdotTerm
			reduce(183), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(183), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(183), // ||=, reduce: AddsubdotTerm
			reduce(183), // ^^=, reduce: AddsubdotTerm
			reduce(183), // &&=, reduce: AddsubdotTerm
			reduce(183), // ??=, reduce: AddsubdotTerm
			reduce(183), // ???=, reduce: AddsubdotTerm
			reduce(183), // |=, reduce: AddsubdotTerm
			reduce(183), // &=, reduce: AddsubdotTerm
			reduce(183), // ^=, reduce: AddsubdotTerm
			reduce(183), // <<=, reduce: AddsubdotTerm
			reduce(183), // >>=, reduce: AddsubdotTerm
			reduce(183), // >>>=, reduce: AddsubdotTerm
			reduce(183), // +=, reduce: AddsubdotTerm
			reduce(183), // .=, reduce: AddsubdotTerm
			reduce(183), // -=, reduce: AddsubdotTerm
			reduce(183), // *=, reduce: AddsubdotTerm
			reduce(183), // /=, reduce: AddsubdotTerm
			reduce(183), // //=, reduce: AddsubdotTerm
			reduce(183), // %=, reduce: AddsubdotTerm
			reduce(183), // **=, reduce: AddsubdotTerm
			reduce(183), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(183), // ||, reduce: AddsubdotTerm
			reduce(183), // ^^, reduce: AddsubdotTerm
			reduce(183), // &&, reduce: AddsubdotTerm
			reduce(183), // ??, reduce: AddsubdotTerm
			reduce(183), // ???, reduce: AddsubdotTerm
			reduce(183), // =~, reduce: AddsubdotTerm
			reduce(183), // !=~, reduce: AddsubdotTerm
			reduce(183), // ==, reduce: AddsubdotTerm
			reduce(183), // !=, reduce: AddsubdotTerm
			reduce(183), // <=>, reduce: AddsubdotTerm
			reduce(183), // >=, reduce: AddsubdotTerm
			reduce(183), // <, reduce: AddsubdotTerm
			reduce(183), // <=, reduce: AddsubdotTerm
			reduce(183), // ^, reduce: AddsubdotTerm
			reduce(183), // &, reduce: AddsubdotTerm
			reduce(183), // <<, reduce: AddsubdotTerm
			reduce(183), // >>>, reduce: AddsubdotTerm
			reduce(183), // +, reduce: AddsubdotTerm
			reduce(183), // -, reduce: AddsubdotTerm
			reduce(183), // .+, reduce: AddsubdotTerm
			reduce(183), // .-, reduce: AddsubdotTerm
			shift(4144), // *
			shift(4145), // /
			shift(4146), // //
			shift(4147), // %
			shift(4148), // .*
			shift(4149), // ./
			shift(4150), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4403
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(184), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			reduce(184), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(184), // >, reduce: AddsubdotTerm
			reduce(184), // >>, reduce: AddsubdotTerm
			reduce(184), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(184), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(184), // ||=, reduce: AddsubdotTerm
			reduce(184), // ^^=, reduce: AddsubdotTerm
			reduce(184), // &&=, reduce: AddsubdotTerm
			reduce(184), // ??=, reduce: AddsubdotTerm
			reduce(184), // ???=, reduce: AddsubdotTerm
			reduce(184), // |=, reduce: AddsubdotTerm
			reduce(184), // &=, reduce: AddsubdotTerm
			reduce(184), // ^=, reduce: AddsubdotTerm
			reduce(184), // <<=, reduce: AddsubdotTerm
			reduce(184), // >>=, reduce: AddsubdotTerm
			reduce(184), // >>>=, reduce: AddsubdotTerm
			reduce(184), // +=, reduce: AddsubdotTerm
			reduce(184), // .=, reduce: AddsubdotTerm
			reduce(184), // -=, reduce: AddsubdotTerm
			reduce(184), // *=, reduce: AddsubdotTerm
			reduce(184), // /=, reduce: AddsubdotTerm
			reduce(184), // //=, reduce: AddsubdotTerm
			reduce(184), // %=, reduce: AddsubdotTerm
			reduce(184), // **=, reduce: AddsubdotTerm
			reduce(184), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(184), // ||, reduce: AddsubdotTerm
			reduce(184), // ^^, reduce: AddsubdotTerm
			reduce(184), // &&, reduce: AddsubdotTerm
			reduce(184), // ??, reduce: AddsubdotTerm
			reduce(184), // ???, reduce: AddsubdotTerm
			reduce(184), // =~, reduce: AddsubdotTerm
			reduce(184), // !=~, reduce: AddsubdotTerm
			reduce(184), // ==, reduce: AddsubdotTerm
			reduce(184), // !=, reduce: AddsubdotTerm
			reduce(184), // <=>, reduce: AddsubdotTerm
			reduce(184), // >=, reduce: AddsubdotTerm
			reduce(184), // <, reduce: AddsubdotTerm
			reduce(184), // <=, reduce: AddsubdotTerm
			reduce(184), // ^, reduce: AddsubdotTerm
			reduce(184), // &, reduce: AddsubdotTerm
			reduce(184), // <<, reduce: AddsubdotTerm
			reduce(184), // >>>, reduce: AddsubdotTerm
			reduce(184), // +, reduce: AddsubdotTerm
			reduce(184), // -, reduce: AddsubdotTerm
			reduce(184), // .+, reduce: AddsubdotTerm
			reduce(184), // .-, reduce: AddsubdotTerm
			shift(4144), // *
			shift(4145), // /
			shift(4146), // //
			shift(4147), // %
			shift(4148), // .*
			shift(4149), // ./
			shift(4150), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4404
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(185), // ;, reduce: AddsubdotTerm
			nil,         // {
			nil,         // }
			reduce(185), // =, reduce: AddsubdotTerm
			nil,         // unset
			nil,         // filter
			reduce(185), // >, reduce: AddsubdotTerm
			reduce(185), // >>, reduce: AddsubdotTerm
			reduce(185), // |, reduce: AddsubdotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(185), // ,, reduce: AddsubdotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(185), // ||=, reduce: AddsubdotTerm
			reduce(185), // ^^=, reduce: AddsubdotTerm
			reduce(185), // &&=, reduce: AddsubdotTerm
			reduce(185), // ??=, reduce: AddsubdotTerm
			reduce(185), // ???=, reduce: AddsubdotTerm
			reduce(185), // |=, reduce: AddsubdotTerm
			reduce(185), // &=, reduce: AddsubdotTerm
			reduce(185), // ^=, reduce: AddsubdotTerm
			reduce(185), // <<=, reduce: AddsubdotTerm
			reduce(185), // >>=, reduce: AddsubdotTerm
			reduce(185), // >>>=, reduce: AddsubdotTerm
			reduce(185), // +=, reduce: AddsubdotTerm
			reduce(185), // .=, reduce: AddsubdotTerm
			reduce(185), // -=, reduce: AddsubdotTerm
			reduce(185), // *=, reduce: AddsubdotTerm
			reduce(185), // /=, reduce: AddsubdotTerm
			reduce(185), // //=, reduce: AddsubdotTerm
			reduce(185), // %=, reduce: AddsubdotTerm
			reduce(185), // **=, reduce: AddsubdotTerm
			reduce(185), // ?, reduce: AddsubdotTerm
			nil,         // :
			reduce(185), // ||, reduce: AddsubdotTerm
			reduce(185), // ^^, reduce: AddsubdotTerm
			reduce(185), // &&, reduce: AddsubdotTerm
			reduce(185), // ??, reduce: AddsubdotTerm
			reduce(185), // ???, reduce: AddsubdotTerm
			reduce(185), // =~, reduce: AddsubdotTerm
			reduce(185), // !=~, reduce: AddsubdotTerm
			reduce(185), // ==, reduce: AddsubdotTerm
			reduce(185), // !=, reduce: AddsubdotTerm
			reduce(185), // <=>, reduce: AddsubdotTerm
			reduce(185), // >=, reduce: AddsubdotTerm
			reduce(185), // <, reduce: AddsubdotTerm
			reduce(185), // <=, reduce: AddsubdotTerm
			reduce(185), // ^, reduce: AddsubdotTerm
			reduce(185), // &, reduce: AddsubdotTerm
			reduce(185), // <<, reduce: AddsubdotTerm
			reduce(185), // >>>, reduce: AddsubdotTerm
			reduce(185), // +, reduce: AddsubdotTerm
			reduce(185), // -, reduce: AddsubdotTerm
			reduce(185), // .+, reduce: AddsubdotTerm
			reduce(185), // .-, reduce: AddsubdotTerm
			shift(4144), // *
			shift(4145), // /
			shift(4146), // //
			shift(4147), // %
			shift(4148), // .*
			shift(4149), // ./
			shift(4150), // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4405
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(187), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			reduce(187), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(187), // >, reduce: MuldivTerm
			reduce(187), // >>, reduce: MuldivTerm
			reduce(187), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(187), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(187), // ||=, reduce: MuldivTerm
			reduce(187), // ^^=, reduce: MuldivTerm
			reduce(187), // &&=, reduce: MuldivTerm
			reduce(187), // ??=, reduce: MuldivTerm
			reduce(187), // ???=, reduce: MuldivTerm
			reduce(187), // |=, reduce: MuldivTerm
			reduce(187), // &=, reduce: MuldivTerm
			reduce(187), // ^=, reduce: MuldivTerm
			reduce(187), // <<=, reduce: MuldivTerm
			reduce(187), // >>=, reduce: MuldivTerm
			reduce(187), // >>>=, reduce: MuldivTerm
			reduce(187), // +=, reduce: MuldivTerm
			reduce(187), // .=, reduce: MuldivTerm
			reduce(187), // -=, reduce: MuldivTerm
			reduce(187), // *=, reduce: MuldivTerm
			reduce(187), // /=, reduce: MuldivTerm
			reduce(187), // //=, reduce: MuldivTerm
			reduce(187), // %=, reduce: MuldivTerm
			reduce(187), // **=, reduce: MuldivTerm
			reduce(187), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(187), // ||, reduce: MuldivTerm
			reduce(187), // ^^, reduce: MuldivTerm
			reduce(187), // &&, reduce: MuldivTerm
			reduce(187), // ??, reduce: MuldivTerm
			reduce(187), // ???, reduce: MuldivTerm
			reduce(187), // =~, reduce: MuldivTerm
			reduce(187), // !=~, reduce: MuldivTerm
			reduce(187), // ==, reduce: MuldivTerm
			reduce(187), // !=, reduce: MuldivTerm
			reduce(187), // <=>, reduce: MuldivTerm
			reduce(187), // >=, reduce: MuldivTerm
			reduce(187), // <, reduce: MuldivTerm
			reduce(187), // <=, reduce: MuldivTerm
			reduce(187), // ^, reduce: MuldivTerm
			reduce(187), // &, reduce: MuldivTerm
			reduce(187), // <<, reduce: MuldivTerm
			reduce(187), // >>>, reduce: MuldivTerm
			reduce(187), // +, reduce: MuldivTerm
			reduce(187), // -, reduce: MuldivTerm
			reduce(187), // .+, reduce: MuldivTerm
			reduce(187), // .-, reduce: MuldivTerm
			reduce(187), // *, reduce: MuldivTerm
			reduce(187), // /, reduce: MuldivTerm
			reduce(187), // //, reduce: MuldivTerm
			reduce(187), // %, reduce: MuldivTerm
			reduce(187), // .*, reduce: MuldivTerm
			reduce(187), // ./, reduce: MuldivTerm
			reduce(187), // .//, reduce: MuldivTerm
			shift(4154), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4406
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(188), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			reduce(188), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(188), // >, reduce: MuldivTerm
			reduce(188), // >>, reduce: MuldivTerm
			reduce(188), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(188), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(188), // ||=, reduce: MuldivTerm
			reduce(188), // ^^=, reduce: MuldivTerm
			reduce(188), // &&=, reduce: MuldivTerm
			reduce(188), // ??=, reduce: MuldivTerm
			reduce(188), // ???=, reduce: MuldivTerm
			reduce(188), // |=, reduce: MuldivTerm
			reduce(188), // &=, reduce: MuldivTerm
			reduce(188), // ^=, reduce: MuldivTerm
			reduce(188), // <<=, reduce: MuldivTerm
			reduce(188), // >>=, reduce: MuldivTerm
			reduce(188), // >>>=, reduce: MuldivTerm
			reduce(188), // +=, reduce: MuldivTerm
			reduce(188), // .=, reduce: MuldivTerm
			reduce(188), // -=, reduce: MuldivTerm
			reduce(188), // *=, reduce: MuldivTerm
			reduce(188), // /=, reduce: MuldivTerm
			reduce(188), // //=, reduce: MuldivTerm
			reduce(188), // %=, reduce: MuldivTerm
			reduce(188), // **=, reduce: MuldivTerm
			reduce(188), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(188), // ||, reduce: MuldivTerm
			reduce(188), // ^^, reduce: MuldivTerm
			reduce(188), // &&, reduce: MuldivTerm
			reduce(188), // ??, reduce: MuldivTerm
			reduce(188), // ???, reduce: MuldivTerm
			reduce(188), // =~, reduce: MuldivTerm
			reduce(188), // !=~, reduce: MuldivTerm
			reduce(188), // ==, reduce: MuldivTerm
			reduce(188), // !=, reduce: MuldivTerm
			reduce(188), // <=>, reduce: MuldivTerm
			reduce(188), // >=, reduce: MuldivTerm
			reduce(188), // <, reduce: MuldivTerm
			reduce(188), // <=, reduce: MuldivTerm
			reduce(188), // ^, reduce: MuldivTerm
			reduce(188), // &, reduce: MuldivTerm
			reduce(188), // <<, reduce: MuldivTerm
			reduce(188), // >>>, reduce: MuldivTerm
			reduce(188), // +, reduce: MuldivTerm
			reduce(188), // -, reduce: MuldivTerm
			reduce(188), // .+, reduce: MuldivTerm
			reduce(188), // .-, reduce: MuldivTerm
			reduce(188), // *, reduce: MuldivTerm
			reduce(188), // /, reduce: MuldivTerm
			reduce(188), // //, reduce: MuldivTerm
			reduce(188), // %, reduce: MuldivTerm
			reduce(188), // .*, reduce: MuldivTerm
			reduce(188), // ./, reduce: MuldivTerm
			reduce(188), // .//, reduce: MuldivTerm
			shift(4154), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4407
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(189), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			reduce(189), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(189), // >, reduce: MuldivTerm
			reduce(189), // >>, reduce: MuldivTerm
			reduce(189), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(189), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(189), // ||=, reduce: MuldivTerm
			reduce(189), // ^^=, reduce: MuldivTerm
			reduce(189), // &&=, reduce: MuldivTerm
			reduce(189), // ??=, reduce: MuldivTerm
			reduce(189), // ???=, reduce: MuldivTerm
			reduce(189), // |=, reduce: MuldivTerm
			reduce(189), // &=, reduce: MuldivTerm
			reduce(189), // ^=, reduce: MuldivTerm
			reduce(189), // <<=, reduce: MuldivTerm
			reduce(189), // >>=, reduce: MuldivTerm
			reduce(189), // >>>=, reduce: MuldivTerm
			reduce(189), // +=, reduce: MuldivTerm
			reduce(189), // .=, reduce: MuldivTerm
			reduce(189), // -=, reduce: MuldivTerm
			reduce(189), // *=, reduce: MuldivTerm
			reduce(189), // /=, reduce: MuldivTerm
			reduce(189), // //=, reduce: MuldivTerm
			reduce(189), // %=, reduce: MuldivTerm
			reduce(189), // **=, reduce: MuldivTerm
			reduce(189), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(189), // ||, reduce: MuldivTerm
			reduce(189), // ^^, reduce: MuldivTerm
			reduce(189), // &&, reduce: MuldivTerm
			reduce(189), // ??, reduce: MuldivTerm
			reduce(189), // ???, reduce: MuldivTerm
			reduce(189), // =~, reduce: MuldivTerm
			reduce(189), // !=~, reduce: MuldivTerm
			reduce(189), // ==, reduce: MuldivTerm
			reduce(189), // !=, reduce: MuldivTerm
			reduce(189), // <=>, reduce: MuldivTerm
			reduce(189), // >=, reduce: MuldivTerm
			reduce(189), // <, reduce: MuldivTerm
			reduce(189), // <=, reduce: MuldivTerm
			reduce(189), // ^, reduce: MuldivTerm
			reduce(189), // &, reduce: MuldivTerm
			reduce(189), // <<, reduce: MuldivTerm
			reduce(189), // >>>, reduce: MuldivTerm
			reduce(189), // +, reduce: MuldivTerm
			reduce(189), // -, reduce: MuldivTerm
			reduce(189), // .+, reduce: MuldivTerm
			reduce(189), // .-, reduce: MuldivTerm
			reduce(189), // *, reduce: MuldivTerm
			reduce(189), // /, reduce: MuldivTerm
			reduce(189), // //, reduce: MuldivTerm
			reduce(189), // %, reduce: MuldivTerm
			reduce(189), // .*, reduce: MuldivTerm
			reduce(189), // ./, reduce: MuldivTerm
			reduce(189), // .//, reduce: MuldivTerm
			shift(4154), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4408
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(190), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			reduce(190), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(190), // >, reduce: MuldivTerm
			reduce(190), // >>, reduce: MuldivTerm
			reduce(190), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(190), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(190), // ||=, reduce: MuldivTerm
			reduce(190), // ^^=, reduce: MuldivTerm
			reduce(190), // &&=, reduce: MuldivTerm
			reduce(190), // ??=, reduce: MuldivTerm
			reduce(190), // ???=, reduce: MuldivTerm
			reduce(190), // |=, reduce: MuldivTerm
			reduce(190), // &=, reduce: MuldivTerm
			reduce(190), // ^=, reduce: MuldivTerm
			reduce(190), // <<=, reduce: MuldivTerm
			reduce(190), // >>=, reduce: MuldivTerm
			reduce(190), // >>>=, reduce: MuldivTerm
			reduce(190), // +=, reduce: MuldivTerm
			reduce(190), // .=, reduce: MuldivTerm
			reduce(190), // -=, reduce: MuldivTerm
			reduce(190), // *=, reduce: MuldivTerm
			reduce(190), // /=, reduce: MuldivTerm
			reduce(190), // //=, reduce: MuldivTerm
			reduce(190), // %=, reduce: MuldivTerm
			reduce(190), // **=, reduce: MuldivTerm
			reduce(190), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(190), // ||, reduce: MuldivTerm
			reduce(190), // ^^, reduce: MuldivTerm
			reduce(190), // &&, reduce: MuldivTerm
			reduce(190), // ??, reduce: MuldivTerm
			reduce(190), // ???, reduce: MuldivTerm
			reduce(190), // =~, reduce: MuldivTerm
			reduce(190), // !=~, reduce: MuldivTerm
			reduce(190), // ==, reduce: MuldivTerm
			reduce(190), // !=, reduce: MuldivTerm
			reduce(190), // <=>, reduce: MuldivTerm
			reduce(190), // >=, reduce: MuldivTerm
			reduce(190), // <, reduce: MuldivTerm
			reduce(190), // <=, reduce: MuldivTerm
			reduce(190), // ^, reduce: MuldivTerm
			reduce(190), // &, reduce: MuldivTerm
			reduce(190), // <<, reduce: MuldivTerm
			reduce(190), // >>>, reduce: MuldivTerm
			reduce(190), // +, reduce: MuldivTerm
			reduce(190), // -, reduce: MuldivTerm
			reduce(190), // .+, reduce: MuldivTerm
			reduce(190), // .-, reduce: MuldivTerm
			reduce(190), // *, reduce: MuldivTerm
			reduce(190), // /, reduce: MuldivTerm
			reduce(190), // //, reduce: MuldivTerm
			reduce(190), // %, reduce: MuldivTerm
			reduce(190), // .*, reduce: MuldivTerm
			reduce(190), // ./, reduce: MuldivTerm
			reduce(190), // .//, reduce: MuldivTerm
			shift(4154), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4409
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(191), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			reduce(191), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(191), // >, reduce: MuldivTerm
			reduce(191), // >>, reduce: MuldivTerm
			reduce(191), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(191), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(191), // ||=, reduce: MuldivTerm
			reduce(191), // ^^=, reduce: MuldivTerm
			reduce(191), // &&=, reduce: MuldivTerm
			reduce(191), // ??=, reduce: MuldivTerm
			reduce(191), // ???=, reduce: MuldivTerm
			reduce(191), // |=, reduce: MuldivTerm
			reduce(191), // &=, reduce: MuldivTerm
			reduce(191), // ^=, reduce: MuldivTerm
			reduce(191), // <<=, reduce: MuldivTerm
			reduce(191), // >>=, reduce: MuldivTerm
			reduce(191), // >>>=, reduce: MuldivTerm
			reduce(191), // +=, reduce: MuldivTerm
			reduce(191), // .=, reduce: MuldivTerm
			reduce(191), // -=, reduce: MuldivTerm
			reduce(191), // *=, reduce: MuldivTerm
			reduce(191), // /=, reduce: MuldivTerm
			reduce(191), // //=, reduce: MuldivTerm
			reduce(191), // %=, reduce: MuldivTerm
			reduce(191), // **=, reduce: MuldivTerm
			reduce(191), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(191), // ||, reduce: MuldivTerm
			reduce(191), // ^^, reduce: MuldivTerm
			reduce(191), // &&, reduce: MuldivTerm
			reduce(191), // ??, reduce: MuldivTerm
			reduce(191), // ???, reduce: MuldivTerm
			reduce(191), // =~, reduce: MuldivTerm
			reduce(191), // !=~, reduce: MuldivTerm
			reduce(191), // ==, reduce: MuldivTerm
			reduce(191), // !=, reduce: MuldivTerm
			reduce(191), // <=>, reduce: MuldivTerm
			reduce(191), // >=, reduce: MuldivTerm
			reduce(191), // <, reduce: MuldivTerm
			reduce(191), // <=, reduce: MuldivTerm
			reduce(191), // ^, reduce: MuldivTerm
			reduce(191), // &, reduce: MuldivTerm
			reduce(191), // <<, reduce: MuldivTerm
			reduce(191), // >>>, reduce: MuldivTerm
			reduce(191), // +, reduce: MuldivTerm
			reduce(191), // -, reduce: MuldivTerm
			reduce(191), // .+, reduce: MuldivTerm
			reduce(191), // .-, reduce: MuldivTerm
			reduce(191), // *, reduce: MuldivTerm
			reduce(191), // /, reduce: MuldivTerm
			reduce(191), // //, reduce: MuldivTerm
			reduce(191), // %, reduce: MuldivTerm
			reduce(191), // .*, reduce: MuldivTerm
			reduce(191), // ./, reduce: MuldivTerm
			reduce(191), // .//, reduce: MuldivTerm
			shift(4154), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4410
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(192), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			reduce(192), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(192), // >, reduce: MuldivTerm
			reduce(192), // >>, reduce: MuldivTerm
			reduce(192), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(192), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(192), // ||=, reduce: MuldivTerm
			reduce(192), // ^^=, reduce: MuldivTerm
			reduce(192), // &&=, reduce: MuldivTerm
			reduce(192), // ??=, reduce: MuldivTerm
			reduce(192), // ???=, reduce: MuldivTerm
			reduce(192), // |=, reduce: MuldivTerm
			reduce(192), // &=, reduce: MuldivTerm
			reduce(192), // ^=, reduce: MuldivTerm
			reduce(192), // <<=, reduce: MuldivTerm
			reduce(192), // >>=, reduce: MuldivTerm
			reduce(192), // >>>=, reduce: MuldivTerm
			reduce(192), // +=, reduce: MuldivTerm
			reduce(192), // .=, reduce: MuldivTerm
			reduce(192), // -=, reduce: MuldivTerm
			reduce(192), // *=, reduce: MuldivTerm
			reduce(192), // /=, reduce: MuldivTerm
			reduce(192), // //=, reduce: MuldivTerm
			reduce(192), // %=, reduce: MuldivTerm
			reduce(192), // **=, reduce: MuldivTerm
			reduce(192), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(192), // ||, reduce: MuldivTerm
			reduce(192), // ^^, reduce: MuldivTerm
			reduce(192), // &&, reduce: MuldivTerm
			reduce(192), // ??, reduce: MuldivTerm
			reduce(192), // ???, reduce: MuldivTerm
			reduce(192), // =~, reduce: MuldivTerm
			reduce(192), // !=~, reduce: MuldivTerm
			reduce(192), // ==, reduce: MuldivTerm
			reduce(192), // !=, reduce: MuldivTerm
			reduce(192), // <=>, reduce: MuldivTerm
			reduce(192), // >=, reduce: MuldivTerm
			reduce(192), // <, reduce: MuldivTerm
			reduce(192), // <=, reduce: MuldivTerm
			reduce(192), // ^, reduce: MuldivTerm
			reduce(192), // &, reduce: MuldivTerm
			reduce(192), // <<, reduce: MuldivTerm
			reduce(192), // >>>, reduce: MuldivTerm
			reduce(192), // +, reduce: MuldivTerm
			reduce(192), // -, reduce: MuldivTerm
			reduce(192), // .+, reduce: MuldivTerm
			reduce(192), // .-, reduce: MuldivTerm
			reduce(192), // *, reduce: MuldivTerm
			reduce(192), // /, reduce: MuldivTerm
			reduce(192), // //, reduce: MuldivTerm
			reduce(192), // %, reduce: MuldivTerm
			reduce(192), // .*, reduce: MuldivTerm
			reduce(192), // ./, reduce: MuldivTerm
			reduce(192), // .//, reduce: MuldivTerm
			shift(4154), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4411
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(193), // ;, reduce: MuldivTerm
			nil,         // {
			nil,         // }
			reduce(193), // =, reduce: MuldivTerm
			nil,         // unset
			nil,         // filter
			reduce(193), // >, reduce: MuldivTerm
			reduce(193), // >>, reduce: MuldivTerm
			reduce(193), // |, reduce: MuldivTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(193), // ,, reduce: MuldivTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(193), // ||=, reduce: MuldivTerm
			reduce(193), // ^^=, reduce: MuldivTerm
			reduce(193), // &&=, reduce: MuldivTerm
			reduce(193), // ??=, reduce: MuldivTerm
			reduce(193), // ???=, reduce: MuldivTerm
			reduce(193), // |=, reduce: MuldivTerm
			reduce(193), // &=, reduce: MuldivTerm
			reduce(193), // ^=, reduce: MuldivTerm
			reduce(193), // <<=, reduce: MuldivTerm
			reduce(193), // >>=, reduce: MuldivTerm
			reduce(193), // >>>=, reduce: MuldivTerm
			reduce(193), // +=, reduce: MuldivTerm
			reduce(193), // .=, reduce: MuldivTerm
			reduce(193), // -=, reduce: MuldivTerm
			reduce(193), // *=, reduce: MuldivTerm
			reduce(193), // /=, reduce: MuldivTerm
			reduce(193), // //=, reduce: MuldivTerm
			reduce(193), // %=, reduce: MuldivTerm
			reduce(193), // **=, reduce: MuldivTerm
			reduce(193), // ?, reduce: MuldivTerm
			nil,         // :
			reduce(193), // ||, reduce: MuldivTerm
			reduce(193), // ^^, reduce: MuldivTerm
			reduce(193), // &&, reduce: MuldivTerm
			reduce(193), // ??, reduce: MuldivTerm
			reduce(193), // ???, reduce: MuldivTerm
			reduce(193), // =~, reduce: MuldivTerm
			reduce(193), // !=~, reduce: MuldivTerm
			reduce(193), // ==, reduce: MuldivTerm
			reduce(193), // !=, reduce: MuldivTerm
			reduce(193), // <=>, reduce: MuldivTerm
			reduce(193), // >=, reduce: MuldivTerm
			reduce(193), // <, reduce: MuldivTerm
			reduce(193), // <=, reduce: MuldivTerm
			reduce(193), // ^, reduce: MuldivTerm
			reduce(193), // &, reduce: MuldivTerm
			reduce(193), // <<, reduce: MuldivTerm
			reduce(193), // >>>, reduce: MuldivTerm
			reduce(193), // +, reduce: MuldivTerm
			reduce(193), // -, reduce: MuldivTerm
			reduce(193), // .+, reduce: MuldivTerm
			reduce(193), // .-, reduce: MuldivTerm
			reduce(193), // *, reduce: MuldivTerm
			reduce(193), // /, reduce: MuldivTerm
			reduce(193), // //, reduce: MuldivTerm
			reduce(193), // %, reduce: MuldivTerm
			reduce(193), // .*, reduce: MuldivTerm
			reduce(193), // ./, reduce: MuldivTerm
			reduce(193), // .//, reduce: MuldivTerm
			shift(4154), // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4412
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(195), // ;, reduce: DotTerm
			nil,         // {
			nil,         // }
			reduce(195), // =, reduce: DotTerm
			nil,         // unset
			nil,         // filter
			reduce(195), // >, reduce: DotTerm
			reduce(195), // >>, reduce: DotTerm
			reduce(195), // |, reduce: DotTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(195), // ,, reduce: DotTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(195), // ||=, reduce: DotTerm
			reduce(195), // ^^=, reduce: DotTerm
			reduce(195), // &&=, reduce: DotTerm
			reduce(195), // ??=, reduce: DotTerm
			reduce(195), // ???=, reduce: DotTerm
			reduce(195), // |=, reduce: DotTerm
			reduce(195), // &=, reduce: DotTerm
			reduce(195), // ^=, reduce: DotTerm
			reduce(195), // <<=, reduce: DotTerm
			reduce(195), // >>=, reduce: DotTerm
			reduce(195), // >>>=, reduce: DotTerm
			reduce(195), // +=, reduce: DotTerm
			reduce(195), // .=, reduce: DotTerm
			reduce(195), // -=, reduce: DotTerm
			reduce(195), // *=, reduce: DotTerm
			reduce(195), // /=, reduce: DotTerm
			reduce(195), // //=, reduce: DotTerm
			reduce(195), // %=, reduce: DotTerm
			reduce(195), // **=, reduce: DotTerm
			reduce(195), // ?, reduce: DotTerm
			nil,         // :
			reduce(195), // ||, reduce: DotTerm
			reduce(195), // ^^, reduce: DotTerm
			reduce(195), // &&, reduce: DotTerm
			reduce(195), // ??, reduce: DotTerm
			reduce(195), // ???, reduce: DotTerm
			reduce(195), // =~, reduce: DotTerm
			reduce(195), // !=~, reduce: DotTerm
			reduce(195), // ==, reduce: DotTerm
			reduce(195), // !=, reduce: DotTerm
			reduce(195), // <=>, reduce: DotTerm
			reduce(195), // >=, reduce: DotTerm
			reduce(195), // <, reduce: DotTerm
			reduce(195), // <=, reduce: DotTerm
			reduce(195), // ^, reduce: DotTerm
			reduce(195), // &, reduce: DotTerm
			reduce(195), // <<, reduce: DotTerm
			reduce(195), // >>>, reduce: DotTerm
			reduce(195), // +, reduce: DotTerm
			reduce(195), // -, reduce: DotTerm
			reduce(195), // .+, reduce: DotTerm
			reduce(195), // .-, reduce: DotTerm
			reduce(195), // *, reduce: DotTerm
			reduce(195), // /, reduce: DotTerm
			reduce(195), // //, reduce: DotTerm
			reduce(195), // %, reduce: DotTerm
			reduce(195), // .*, reduce: DotTerm
			reduce(195), // ./, reduce: DotTerm
			reduce(195), // .//, reduce: DotTerm
			reduce(195), // ., reduce: DotTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4413
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4414
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4415
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(204), // ;, reduce: PowTerm
			nil,         // {
			nil,         // }
			reduce(204), // =, reduce: PowTerm
			nil,         // unset
			nil,         // filter
			reduce(204), // >, reduce: PowTerm
			reduce(204), // >>, reduce: PowTerm
			reduce(204), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(204), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(204), // ||=, reduce: PowTerm
			reduce(204), // ^^=, reduce: PowTerm
			reduce(204), // &&=, reduce: PowTerm
			reduce(204), // ??=, reduce: PowTerm
			reduce(204), // ???=, reduce: PowTerm
			reduce(204), // |=, reduce: PowTerm
			reduce(204), // &=, reduce: PowTerm
			reduce(204), // ^=, reduce: PowTerm
			reduce(204), // <<=, reduce: PowTerm
			reduce(204), // >>=, reduce: PowTerm
			reduce(204), // >>>=, reduce: PowTerm
			reduce(204), // +=, reduce: PowTerm
			reduce(204), // .=, reduce: PowTerm
			reduce(204), // -=, reduce: PowTerm
			reduce(204), // *=, reduce: PowTerm
			reduce(204), // /=, reduce: PowTerm
			reduce(204), // //=, reduce: PowTerm
			reduce(204), // %=, reduce: PowTerm
			reduce(204), // **=, reduce: PowTerm
			reduce(204), // ?, reduce: PowTerm
			nil,         // :
			reduce(204), // ||, reduce: PowTerm
			reduce(204), // ^^, reduce: PowTerm
			reduce(204), // &&, reduce: PowTerm
			reduce(204), // ??, reduce: PowTerm
			reduce(204), // ???, reduce: PowTerm
			reduce(204), // =~, reduce: PowTerm
			reduce(204), // !=~, reduce: PowTerm
			reduce(204), // ==, reduce: PowTerm
			reduce(204), // !=, reduce: PowTerm
			reduce(204), // <=>, reduce: PowTerm
			reduce(204), // >=, reduce: PowTerm
			reduce(204), // <, reduce: PowTerm
			reduce(204), // <=, reduce: PowTerm
			reduce(204), // ^, reduce: PowTerm
			reduce(204), // &, reduce: PowTerm
			reduce(204), // <<, reduce: PowTerm
			reduce(204), // >>>, reduce: PowTerm
			reduce(204), // +, reduce: PowTerm
			reduce(204), // -, reduce: PowTerm
			reduce(204), // .+, reduce: PowTerm
			reduce(204), // .-, reduce: PowTerm
			reduce(204), // *, reduce: PowTerm
			reduce(204), // /, reduce: PowTerm
			reduce(204), // //, reduce: PowTerm
			reduce(204), // %, reduce: PowTerm
			reduce(204), // .*, reduce: PowTerm
			reduce(204), // ./, reduce: PowTerm
			reduce(204), // .//, reduce: PowTerm
			reduce(204), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4416
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4536), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4537), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4417
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4539), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4418
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4540), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4419
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4541), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4420
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(229), // ;, reduce: ArrayLiteral
			nil,         // {
			nil,         // }
			reduce(229), // =, reduce: ArrayLiteral
			nil,         // unset
			nil,         // filter
			reduce(229), // >, reduce: ArrayLiteral
			reduce(229), // >>, reduce: ArrayLiteral
			reduce(229), // |, reduce: ArrayLiteral
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(229), // ,, reduce: ArrayLiteral
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(229), // ||=, reduce: ArrayLiteral
			reduce(229), // ^^=, reduce: ArrayLiteral
			reduce(229), // &&=, reduce: ArrayLiteral
			reduce(229), // ??=, reduce: ArrayLiteral
			reduce(229), // ???=, reduce: ArrayLiteral
			reduce(229), // |=, reduce: ArrayLiteral
			reduce(229), // &=, reduce: ArrayLiteral
			reduce(229), // ^=, reduce: ArrayLiteral
			reduce(229), // <<=, reduce: ArrayLiteral
			reduce(229), // >>=, reduce: ArrayLiteral
			reduce(229), // >>>=, reduce: ArrayLiteral
			reduce(229), // +=, reduce: ArrayLiteral
			reduce(229), // .=, reduce: ArrayLiteral
			reduce(229), // -=, reduce: ArrayLiteral
			reduce(229), // *=, reduce: ArrayLiteral
			reduce(229), // /=, reduce: ArrayLiteral
			reduce(229), // //=, reduce: ArrayLiteral
			reduce(229), // %=, reduce: ArrayLiteral
			reduce(229), // **=, reduce: ArrayLiteral
			reduce(229), // ?, reduce: ArrayLiteral
			nil,         // :
			reduce(229), // ||, reduce: ArrayLiteral
			reduce(229), // ^^, reduce: ArrayLiteral
			reduce(229), // &&, reduce: ArrayLiteral
			reduce(229), // ??, reduce: ArrayLiteral
			reduce(229), // ???, reduce: ArrayLiteral
			reduce(229), // =~, reduce: ArrayLiteral
			reduce(229), // !=~, reduce: ArrayLiteral
			reduce(229), // ==, reduce: ArrayLiteral
			reduce(229), // !=, reduce: ArrayLiteral
			reduce(229), // <=>, reduce: ArrayLiteral
			reduce(229), // >=, reduce: ArrayLiteral
			reduce(229), // <, reduce: ArrayLiteral
			reduce(229), // <=, reduce: ArrayLiteral
			reduce(229), // ^, reduce: ArrayLiteral
			reduce(229), // &, reduce: ArrayLiteral
			reduce(229), // <<, reduce: ArrayLiteral
			reduce(229), // >>>, reduce: ArrayLiteral
			reduce(229), // +, reduce: ArrayLiteral
			reduce(229), // -, reduce: ArrayLiteral
			reduce(229), // .+, reduce: ArrayLiteral
			reduce(229), // .-, reduce: ArrayLiteral
			reduce(229), // *, reduce: ArrayLiteral
			reduce(229), // /, reduce: ArrayLiteral
			reduce(229), // //, reduce: ArrayLiteral
			reduce(229), // %, reduce: ArrayLiteral
			reduce(229), // .*, reduce: ArrayLiteral
			reduce(229), // ./, reduce: ArrayLiteral
			reduce(229), // .//, reduce: ArrayLiteral
			reduce(229), // ., reduce: ArrayLiteral
			nil,         // !
			nil,         // ~
			reduce(229), // **, reduce: ArrayLiteral
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(229), // [, reduce: ArrayLiteral
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(229), // [[, reduce: ArrayLiteral
			reduce(229), // [[[, reduce: ArrayLiteral
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4421
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(255), // ;, reduce: ENV
			nil,         // {
			nil,         // }
			reduce(255), // =, reduce: ENV
			nil,         // unset
			nil,         // filter
			reduce(255), // >, reduce: ENV
			reduce(255), // >>, reduce: ENV
			reduce(255), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(255), // ,, reduce: ENV
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(255), // ||=, reduce: ENV
			reduce(255), // ^^=, reduce: ENV
			reduce(255), // &&=, reduce: ENV
			reduce(255), // ??=, reduce: ENV
			reduce(255), // ???=, reduce: ENV
			reduce(255), // |=, reduce: ENV
			reduce(255), // &=, reduce: ENV
			reduce(255), // ^=, reduce: ENV
			reduce(255), // <<=, reduce: ENV
			reduce(255), // >>=, reduce: ENV
			reduce(255), // >>>=, reduce: ENV
			reduce(255), // +=, reduce: ENV
			reduce(255), // .=, reduce: ENV
			reduce(255), // -=, reduce: ENV
			reduce(255), // *=, reduce: ENV
			reduce(255), // /=, reduce: ENV
			reduce(255), // //=, reduce: ENV
			reduce(255), // %=, reduce: ENV
			reduce(255), // **=, reduce: ENV
			reduce(255), // ?, reduce: ENV
			nil,         // :
			reduce(255), // ||, reduce: ENV
			reduce(255), // ^^, reduce: ENV
			reduce(255), // &&, reduce: ENV
			reduce(255), // ??, reduce: ENV
			reduce(255), // ???, reduce: ENV
			reduce(255), // =~, reduce: ENV
			reduce(255), // !=~, reduce: ENV
			reduce(255), // ==, reduce: ENV
			reduce(255), // !=, reduce: ENV
			reduce(255), // <=>, reduce: ENV
			reduce(255), // >=, reduce: ENV
			reduce(255), // <, reduce: ENV
			reduce(255), // <=, reduce: ENV
			reduce(255), // ^, reduce: ENV
			reduce(255), // &, reduce: ENV
			reduce(255), // <<, reduce: ENV
			reduce(255), // >>>, reduce: ENV
			reduce(255), // +, reduce: ENV
			reduce(255), // -, reduce: ENV
			reduce(255), // .+, reduce: ENV
			reduce(255), // .-, reduce: ENV
			reduce(255), // *, reduce: ENV
			reduce(255), // /, reduce: ENV
			reduce(255), // //, reduce: ENV
			reduce(255), // %, reduce: ENV
			reduce(255), // .*, reduce: ENV
			reduce(255), // ./, reduce: ENV
			reduce(255), // .//, reduce: ENV
			reduce(255), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(255), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(255), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(255), // [[, reduce: ENV
			reduce(255), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4422
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4542), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4423
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4543), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4424
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(268), // ;, reduce: FunctionCallsite
			nil,         // {
			nil,         // }
			reduce(268), // =, reduce: FunctionCallsite
			nil,         // unset
			nil,         // filter
			reduce(268), // >, reduce: FunctionCallsite
			reduce(268), // >>, reduce: FunctionCallsite
			reduce(268), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(268), // ,, reduce: FunctionCallsite
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(268), // ||=, reduce: FunctionCallsite
			reduce(268), // ^^=, reduce: FunctionCallsite
			reduce(268), // &&=, reduce: FunctionCallsite
			reduce(268), // ??=, reduce: FunctionCallsite
			reduce(268), // ???=, reduce: FunctionCallsite
			reduce(268), // |=, reduce: FunctionCallsite
			reduce(268), // &=, reduce: FunctionCallsite
			reduce(268), // ^=, reduce: FunctionCallsite
			reduce(268), // <<=, reduce: FunctionCallsite
			reduce(268), // >>=, reduce: FunctionCallsite
			reduce(268), // >>>=, reduce: FunctionCallsite
			reduce(268), // +=, reduce: FunctionCallsite
			reduce(268), // .=, reduce: FunctionCallsite
			reduce(268), // -=, reduce: FunctionCallsite
			reduce(268), // *=, reduce: FunctionCallsite
			reduce(268), // /=, reduce: FunctionCallsite
			reduce(268), // //=, reduce: FunctionCallsite
			reduce(268), // %=, reduce: FunctionCallsite
			reduce(268), // **=, reduce: FunctionCallsite
			reduce(268), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(268), // ||, reduce: FunctionCallsite
			reduce(268), // ^^, reduce: FunctionCallsite
			reduce(268), // &&, reduce: FunctionCallsite
			reduce(268), // ??, reduce: FunctionCallsite
			reduce(268), // ???, reduce: FunctionCallsite
			reduce(268), // =~, reduce: FunctionCallsite
			reduce(268), // !=~, reduce: FunctionCallsite
			reduce(268), // ==, reduce: FunctionCallsite
			reduce(268), // !=, reduce: FunctionCallsite
			reduce(268), // <=>, reduce: FunctionCallsite
			reduce(268), // >=, reduce: FunctionCallsite
			reduce(268), // <, reduce: FunctionCallsite
			reduce(268), // <=, reduce: FunctionCallsite
			reduce(268), // ^, reduce: FunctionCallsite
			reduce(268), // &, reduce: FunctionCallsite
			reduce(268), // <<, reduce: FunctionCallsite
			reduce(268), // >>>, reduce: FunctionCallsite
			reduce(268), // +, reduce: FunctionCallsite
			reduce(268), // -, reduce: FunctionCallsite
			reduce(268), // .+, reduce: FunctionCallsite
			reduce(268), // .-, reduce: FunctionCallsite
			reduce(268), // *, reduce: FunctionCallsite
			reduce(268), // /, reduce: FunctionCallsite
			reduce(268), // //, reduce: FunctionCallsite
			reduce(268), // %, reduce: FunctionCallsite
			reduce(268), // .*, reduce: FunctionCallsite
			reduce(268), // ./, reduce: FunctionCallsite
			reduce(268), // .//, reduce: FunctionCallsite
			reduce(268), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(268), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(268), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(268), // [[, reduce: FunctionCallsite
			reduce(268), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4425
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(318), // ,, reduce: TripleForUpdate
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(318), // ), reduce: TripleForUpdate
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4426
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			shift(4544), // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			shift(4545), // ||=
			shift(4546), // ^^=
			shift(4547), // &&=
			shift(4548), // ??=
			shift(4549), // ???=
			shift(4550), // |=
			shift(4551), // &=
			shift(4552), // ^=
			shift(4553), // <<=
			shift(4554), // >>=
			shift(4555), // >>>=
			shift(4556), // +=
			shift(4557), // .=
			shift(4558), // -=
			shift(4559), // *=
			shift(4560), // /=
			shift(4561), // //=
			shift(4562), // %=
			shift(4563), // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4427
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(4564), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4565), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4428
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(314), // ;, reduce: TripleForContinuation
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(314), // ,, reduce: TripleForContinuation
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4429
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4566), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4430
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(324), // =, reduce: UnnamedFunctionDefinition
			nil,         // unset
			nil,         // filter
			reduce(324), // >, reduce: UnnamedFunctionDefinition
			reduce(324), // >>, reduce: UnnamedFunctionDefinition
			reduce(324), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(324), // ||=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^=, reduce: UnnamedFunctionDefinition
			reduce(324), // &&=, reduce: UnnamedFunctionDefinition
			reduce(324), // ??=, reduce: UnnamedFunctionDefinition
			reduce(324), // ???=, reduce: UnnamedFunctionDefinition
			reduce(324), // |=, reduce: UnnamedFunctionDefinition
			reduce(324), // &=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^=, reduce: UnnamedFunctionDefinition
			reduce(324), // <<=, reduce: UnnamedFunctionDefinition
			reduce(324), // >>=, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>=, reduce: UnnamedFunctionDefinition
			reduce(324), // +=, reduce: UnnamedFunctionDefinition
			reduce(324), // .=, reduce: UnnamedFunctionDefinition
			reduce(324), // -=, reduce: UnnamedFunctionDefinition
			reduce(324), // *=, reduce: UnnamedFunctionDefinition
			reduce(324), // /=, reduce: UnnamedFunctionDefinition
			reduce(324), // //=, reduce: UnnamedFunctionDefinition
			reduce(324), // %=, reduce: UnnamedFunctionDefinition
			reduce(324), // **=, reduce: UnnamedFunctionDefinition
			reduce(324), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(324), // ||, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^, reduce: UnnamedFunctionDefinition
			reduce(324), // &&, reduce: UnnamedFunctionDefinition
			reduce(324), // ??, reduce: UnnamedFunctionDefinition
			reduce(324), // ???, reduce: UnnamedFunctionDefinition
			reduce(324), // =~, reduce: UnnamedFunctionDefinition
			reduce(324), // !=~, reduce: UnnamedFunctionDefinition
			reduce(324), // ==, reduce: UnnamedFunctionDefinition
			reduce(324), // !=, reduce: UnnamedFunctionDefinition
			reduce(324), // <=>, reduce: UnnamedFunctionDefinition
			reduce(324), // >=, reduce: UnnamedFunctionDefinition
			reduce(324), // <, reduce: UnnamedFunctionDefinition
			reduce(324), // <=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^, reduce: UnnamedFunctionDefinition
			reduce(324), // &, reduce: UnnamedFunctionDefinition
			reduce(324), // <<, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>, reduce: UnnamedFunctionDefinition
			reduce(324), // +, reduce: UnnamedFunctionDefinition
			reduce(324), // -, reduce: UnnamedFunctionDefinition
			reduce(324), // .+, reduce: UnnamedFunctionDefinition
			reduce(324), // .-, reduce: UnnamedFunctionDefinition
			reduce(324), // *, reduce: UnnamedFunctionDefinition
			reduce(324), // /, reduce: UnnamedFunctionDefinition
			reduce(324), // //, reduce: UnnamedFunctionDefinition
			reduce(324), // %, reduce: UnnamedFunctionDefinition
			reduce(324), // .*, reduce: UnnamedFunctionDefinition
			reduce(324), // ./, reduce: UnnamedFunctionDefinition
			reduce(324), // .//, reduce: UnnamedFunctionDefinition
			reduce(324), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(324), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(324), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(324), // [[, reduce: UnnamedFunctionDefinition
			reduce(324), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4431
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S4432
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4433
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(13), // $, reduce: StatementBlockInBraces
			nil,        // empty
			reduce(13), // ;, reduce: StatementBlockInBraces
			reduce(13), // {, reduce: StatementBlockInBraces
			nil,        // }
			reduce(13), // =, reduce: StatementBlockInBraces
			nil,        // unset
			nil,        // filter
			reduce(13), // >, reduce: StatementBlockInBraces
			reduce(13), // >>, reduce: StatementBlockInBraces
			reduce(13), // |, reduce: StatementBlockInBraces
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			reduce(13), // ||=, reduce: StatementBlockInBraces
			reduce(13), // ^^=, reduce: StatementBlockInBraces
			reduce(13), // &&=, reduce: StatementBlockInBraces
			reduce(13), // ??=, reduce: StatementBlockInBraces
			reduce(13), // ???=, reduce: StatementBlockInBraces
			reduce(13), // |=, reduce: StatementBlockInBraces
			reduce(13), // &=, reduce: StatementBlockInBraces
			reduce(13), // ^=, reduce: StatementBlockInBraces
			reduce(13), // <<=, reduce: StatementBlockInBraces
			reduce(13), // >>=, reduce: StatementBlockInBraces
			reduce(13), // >>>=, reduce: StatementBlockInBraces
			reduce(13), // +=, reduce: StatementBlockInBraces
			reduce(13), // .=, reduce: StatementBlockInBraces
			reduce(13), // -=, reduce: StatementBlockInBraces
			reduce(13), // *=, reduce: StatementBlockInBraces
			reduce(13), // /=, reduce: StatementBlockInBraces
			reduce(13), // //=, reduce: StatementBlockInBraces
			reduce(13), // %=, reduce: StatementBlockInBraces
			reduce(13), // **=, reduce: StatementBlockInBraces
			reduce(13), // ?, reduce: StatementBlockInBraces
			nil,        // :
			reduce(13), // ||, reduce: StatementBlockInBraces
			reduce(13), // ^^, reduce: StatementBlockInBraces
			reduce(13), // &&, reduce: StatementBlockInBraces
			reduce(13), // ??, reduce: StatementBlockInBraces
			reduce(13), // ???, reduce: StatementBlockInBraces
			reduce(13), // =~, reduce: StatementBlockInBraces
			reduce(13), // !=~, reduce: StatementBlockInBraces
			reduce(13), // ==, reduce: StatementBlockInBraces
			reduce(13), // !=, reduce: StatementBlockInBraces
			reduce(13), // <=>, reduce: StatementBlockInBraces
			reduce(13), // >=, reduce: StatementBlockInBraces
			reduce(13), // <, reduce: StatementBlockInBraces
			reduce(13), // <=, reduce: StatementBlockInBraces
			reduce(13), // ^, reduce: StatementBlockInBraces
			reduce(13), // &, reduce: StatementBlockInBraces
			reduce(13), // <<, reduce: StatementBlockInBraces
			reduce(13), // >>>, reduce: StatementBlockInBraces
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			reduce(13), // *, reduce: StatementBlockInBraces
			reduce(13), // /, reduce: StatementBlockInBraces
			reduce(13), // //, reduce: StatementBlockInBraces
			reduce(13), // %, reduce: StatementBlockInBraces
			reduce(13), // .*, reduce: StatementBlockInBraces
			reduce(13), // ./, reduce: StatementBlockInBraces
			reduce(13), // .//, reduce: StatementBlockInBraces
			reduce(13), // ., reduce: StatementBlockInBraces
			nil,        // !
			nil,        // ~
			reduce(13), // **, reduce: StatementBlockInBraces
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(13), // [, reduce: StatementBlockInBraces
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(13), // [[, reduce: StatementBlockInBraces
			reduce(13), // [[[, reduce: StatementBlockInBraces
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4434
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(325), // $, reduce: UnnamedFunctionDefinition
			nil,         // empty
			reduce(325), // ;, reduce: UnnamedFunctionDefinition
			reduce(325), // {, reduce: UnnamedFunctionDefinition
			nil,         // }
			reduce(325), // =, reduce: UnnamedFunctionDefinition
			nil,         // unset
			nil,         // filter
			reduce(325), // >, reduce: UnnamedFunctionDefinition
			reduce(325), // >>, reduce: UnnamedFunctionDefinition
			reduce(325), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(325), // ||=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^=, reduce: UnnamedFunctionDefinition
			reduce(325), // &&=, reduce: UnnamedFunctionDefinition
			reduce(325), // ??=, reduce: UnnamedFunctionDefinition
			reduce(325), // ???=, reduce: UnnamedFunctionDefinition
			reduce(325), // |=, reduce: UnnamedFunctionDefinition
			reduce(325), // &=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^=, reduce: UnnamedFunctionDefinition
			reduce(325), // <<=, reduce: UnnamedFunctionDefinition
			reduce(325), // >>=, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>=, reduce: UnnamedFunctionDefinition
			reduce(325), // +=, reduce: UnnamedFunctionDefinition
			reduce(325), // .=, reduce: UnnamedFunctionDefinition
			reduce(325), // -=, reduce: UnnamedFunctionDefinition
			reduce(325), // *=, reduce: UnnamedFunctionDefinition
			reduce(325), // /=, reduce: UnnamedFunctionDefinition
			reduce(325), // //=, reduce: UnnamedFunctionDefinition
			reduce(325), // %=, reduce: UnnamedFunctionDefinition
			reduce(325), // **=, reduce: UnnamedFunctionDefinition
			reduce(325), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(325), // ||, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^, reduce: UnnamedFunctionDefinition
			reduce(325), // &&, reduce: UnnamedFunctionDefinition
			reduce(325), // ??, reduce: UnnamedFunctionDefinition
			reduce(325), // ???, reduce: UnnamedFunctionDefinition
			reduce(325), // =~, reduce: UnnamedFunctionDefinition
			reduce(325), // !=~, reduce: UnnamedFunctionDefinition
			reduce(325), // ==, reduce: UnnamedFunctionDefinition
			reduce(325), // !=, reduce: UnnamedFunctionDefinition
			reduce(325), // <=>, reduce: UnnamedFunctionDefinition
			reduce(325), // >=, reduce: UnnamedFunctionDefinition
			reduce(325), // <, reduce: UnnamedFunctionDefinition
			reduce(325), // <=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^, reduce: UnnamedFunctionDefinition
			reduce(325), // &, reduce: UnnamedFunctionDefinition
			reduce(325), // <<, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>, reduce: UnnamedFunctionDefinition
			reduce(325), // +, reduce: UnnamedFunctionDefinition
			reduce(325), // -, reduce: UnnamedFunctionDefinition
			reduce(325), // .+, reduce: UnnamedFunctionDefinition
			reduce(325), // .-, reduce: UnnamedFunctionDefinition
			reduce(325), // *, reduce: UnnamedFunctionDefinition
			reduce(325), // /, reduce: UnnamedFunctionDefinition
			reduce(325), // //, reduce: UnnamedFunctionDefinition
			reduce(325), // %, reduce: UnnamedFunctionDefinition
			reduce(325), // .*, reduce: UnnamedFunctionDefinition
			reduce(325), // ./, reduce: UnnamedFunctionDefinition
			reduce(325), // .//, reduce: UnnamedFunctionDefinition
			reduce(325), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(325), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(325), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(325), // [[, reduce: UnnamedFunctionDefinition
			reduce(325), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4435
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(289), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4436
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(108), // }, reduce: PositionalFieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(108), // >, reduce: PositionalFieldValue
			reduce(108), // >>, reduce: PositionalFieldValue
			reduce(108), // |, reduce: PositionalFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(108), // ,, reduce: PositionalFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(108), // ?, reduce: PositionalFieldValue
			nil,         // :
			reduce(108), // ||, reduce: PositionalFieldValue
			reduce(108), // ^^, reduce: PositionalFieldValue
			reduce(108), // &&, reduce: PositionalFieldValue
			reduce(108), // ??, reduce: PositionalFieldValue
			reduce(108), // ???, reduce: PositionalFieldValue
			reduce(108), // =~, reduce: PositionalFieldValue
			reduce(108), // !=~, reduce: PositionalFieldValue
			reduce(108), // ==, reduce: PositionalFieldValue
			reduce(108), // !=, reduce: PositionalFieldValue
			reduce(108), // <=>, reduce: PositionalFieldValue
			reduce(108), // >=, reduce: PositionalFieldValue
			reduce(108), // <, reduce: PositionalFieldValue
			reduce(108), // <=, reduce: PositionalFieldValue
			reduce(108), // ^, reduce: PositionalFieldValue
			reduce(108), // &, reduce: PositionalFieldValue
			reduce(108), // <<, reduce: PositionalFieldValue
			reduce(108), // >>>, reduce: PositionalFieldValue
			reduce(108), // +, reduce: PositionalFieldValue
			reduce(108), // -, reduce: PositionalFieldValue
			reduce(108), // .+, reduce: PositionalFieldValue
			reduce(108), // .-, reduce: PositionalFieldValue
			reduce(108), // *, reduce: PositionalFieldValue
			reduce(108), // /, reduce: PositionalFieldValue
			reduce(108), // //, reduce: PositionalFieldValue
			reduce(108), // %, reduce: PositionalFieldValue
			reduce(108), // .*, reduce: PositionalFieldValue
			reduce(108), // ./, reduce: PositionalFieldValue
			reduce(108), // .//, reduce: PositionalFieldValue
			reduce(108), // ., reduce: PositionalFieldValue
			nil,         // !
			nil,         // ~
			reduce(108), // **, reduce: PositionalFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(108), // [, reduce: PositionalFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(108), // [[, reduce: PositionalFieldValue
			reduce(108), // [[[, reduce: PositionalFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4437
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(149), // }, reduce: TernaryTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(149), // ,, reduce: TernaryTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4438
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4570), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4439
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(265), // }, reduce: ArraySliceAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(265), // >, reduce: ArraySliceAccess
			reduce(265), // >>, reduce: ArraySliceAccess
			reduce(265), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(265), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(265), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(265), // ||, reduce: ArraySliceAccess
			reduce(265), // ^^, reduce: ArraySliceAccess
			reduce(265), // &&, reduce: ArraySliceAccess
			reduce(265), // ??, reduce: ArraySliceAccess
			reduce(265), // ???, reduce: ArraySliceAccess
			reduce(265), // =~, reduce: ArraySliceAccess
			reduce(265), // !=~, reduce: ArraySliceAccess
			reduce(265), // ==, reduce: ArraySliceAccess
			reduce(265), // !=, reduce: ArraySliceAccess
			reduce(265), // <=>, reduce: ArraySliceAccess
			reduce(265), // >=, reduce: ArraySliceAccess
			reduce(265), // <, reduce: ArraySliceAccess
			reduce(265), // <=, reduce: ArraySliceAccess
			reduce(265), // ^, reduce: ArraySliceAccess
			reduce(265), // &, reduce: ArraySliceAccess
			reduce(265), // <<, reduce: ArraySliceAccess
			reduce(265), // >>>, reduce: ArraySliceAccess
			reduce(265), // +, reduce: ArraySliceAccess
			reduce(265), // -, reduce: ArraySliceAccess
			reduce(265), // .+, reduce: ArraySliceAccess
			reduce(265), // .-, reduce: ArraySliceAccess
			reduce(265), // *, reduce: ArraySliceAccess
			reduce(265), // /, reduce: ArraySliceAccess
			reduce(265), // //, reduce: ArraySliceAccess
			reduce(265), // %, reduce: ArraySliceAccess
			reduce(265), // .*, reduce: ArraySliceAccess
			reduce(265), // ./, reduce: ArraySliceAccess
			reduce(265), // .//, reduce: ArraySliceAccess
			reduce(265), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(265), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(265), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(265), // [[, reduce: ArraySliceAccess
			reduce(265), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4440
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(264), // }, reduce: ArraySliceAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(264), // >, reduce: ArraySliceAccess
			reduce(264), // >>, reduce: ArraySliceAccess
			reduce(264), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(264), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(264), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(264), // ||, reduce: ArraySliceAccess
			reduce(264), // ^^, reduce: ArraySliceAccess
			reduce(264), // &&, reduce: ArraySliceAccess
			reduce(264), // ??, reduce: ArraySliceAccess
			reduce(264), // ???, reduce: ArraySliceAccess
			reduce(264), // =~, reduce: ArraySliceAccess
			reduce(264), // !=~, reduce: ArraySliceAccess
			reduce(264), // ==, reduce: ArraySliceAccess
			reduce(264), // !=, reduce: ArraySliceAccess
			reduce(264), // <=>, reduce: ArraySliceAccess
			reduce(264), // >=, reduce: ArraySliceAccess
			reduce(264), // <, reduce: ArraySliceAccess
			reduce(264), // <=, reduce: ArraySliceAccess
			reduce(264), // ^, reduce: ArraySliceAccess
			reduce(264), // &, reduce: ArraySliceAccess
			reduce(264), // <<, reduce: ArraySliceAccess
			reduce(264), // >>>, reduce: ArraySliceAccess
			reduce(264), // +, reduce: ArraySliceAccess
			reduce(264), // -, reduce: ArraySliceAccess
			reduce(264), // .+, reduce: ArraySliceAccess
			reduce(264), // .-, reduce: ArraySliceAccess
			reduce(264), // *, reduce: ArraySliceAccess
			reduce(264), // /, reduce: ArraySliceAccess
			reduce(264), // //, reduce: ArraySliceAccess
			reduce(264), // %, reduce: ArraySliceAccess
			reduce(264), // .*, reduce: ArraySliceAccess
			reduce(264), // ./, reduce: ArraySliceAccess
			reduce(264), // .//, reduce: ArraySliceAccess
			reduce(264), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(264), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(264), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(264), // [[, reduce: ArraySliceAccess
			reduce(264), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4441
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(261), // }, reduce: ArrayOrMapPositionalNameAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(261), // >, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |, reduce: ArrayOrMapPositionalNameAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(261), // ,, reduce: ArrayOrMapPositionalNameAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(261), // ?, reduce: ArrayOrMapPositionalNameAccess
			nil,         // :
			reduce(261), // ||, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // =~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ==, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .+, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .-, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .*, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ./, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .//, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ., reduce: ArrayOrMapPositionalNameAccess
			nil,         // !
			nil,         // ~
			reduce(261), // **, reduce: ArrayOrMapPositionalNameAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(261), // [, reduce: ArrayOrMapPositionalNameAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(261), // [[, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // [[[, reduce: ArrayOrMapPositionalNameAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4442
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4571), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4443
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(324), // }, reduce: UnnamedFunctionDefinition
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(324), // >, reduce: UnnamedFunctionDefinition
			reduce(324), // >>, reduce: UnnamedFunctionDefinition
			reduce(324), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(324), // ,, reduce: UnnamedFunctionDefinition
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(324), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(324), // ||, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^, reduce: UnnamedFunctionDefinition
			reduce(324), // &&, reduce: UnnamedFunctionDefinition
			reduce(324), // ??, reduce: UnnamedFunctionDefinition
			reduce(324), // ???, reduce: UnnamedFunctionDefinition
			reduce(324), // =~, reduce: UnnamedFunctionDefinition
			reduce(324), // !=~, reduce: UnnamedFunctionDefinition
			reduce(324), // ==, reduce: UnnamedFunctionDefinition
			reduce(324), // !=, reduce: UnnamedFunctionDefinition
			reduce(324), // <=>, reduce: UnnamedFunctionDefinition
			reduce(324), // >=, reduce: UnnamedFunctionDefinition
			reduce(324), // <, reduce: UnnamedFunctionDefinition
			reduce(324), // <=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^, reduce: UnnamedFunctionDefinition
			reduce(324), // &, reduce: UnnamedFunctionDefinition
			reduce(324), // <<, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>, reduce: UnnamedFunctionDefinition
			reduce(324), // +, reduce: UnnamedFunctionDefinition
			reduce(324), // -, reduce: UnnamedFunctionDefinition
			reduce(324), // .+, reduce: UnnamedFunctionDefinition
			reduce(324), // .-, reduce: UnnamedFunctionDefinition
			reduce(324), // *, reduce: UnnamedFunctionDefinition
			reduce(324), // /, reduce: UnnamedFunctionDefinition
			reduce(324), // //, reduce: UnnamedFunctionDefinition
			reduce(324), // %, reduce: UnnamedFunctionDefinition
			reduce(324), // .*, reduce: UnnamedFunctionDefinition
			reduce(324), // ./, reduce: UnnamedFunctionDefinition
			reduce(324), // .//, reduce: UnnamedFunctionDefinition
			reduce(324), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(324), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(324), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(324), // [[, reduce: UnnamedFunctionDefinition
			reduce(324), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4444
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S4445
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4446
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(13), // >, reduce: StatementBlockInBraces
			reduce(13), // >>, reduce: StatementBlockInBraces
			reduce(13), // |, reduce: StatementBlockInBraces
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(13), // ?, reduce: StatementBlockInBraces
			reduce(13), // :, reduce: StatementBlockInBraces
			reduce(13), // ||, reduce: StatementBlockInBraces
			reduce(13), // ^^, reduce: StatementBlockInBraces
			reduce(13), // &&, reduce: StatementBlockInBraces
			reduce(13), // ??, reduce: StatementBlockInBraces
			reduce(13), // ???, reduce: StatementBlockInBraces
			reduce(13), // =~, reduce: StatementBlockInBraces
			reduce(13), // !=~, reduce: StatementBlockInBraces
			reduce(13), // ==, reduce: StatementBlockInBraces
			reduce(13), // !=, reduce: StatementBlockInBraces
			reduce(13), // <=>, reduce: StatementBlockInBraces
			reduce(13), // >=, reduce: StatementBlockInBraces
			reduce(13), // <, reduce: StatementBlockInBraces
			reduce(13), // <=, reduce: StatementBlockInBraces
			reduce(13), // ^, reduce: StatementBlockInBraces
			reduce(13), // &, reduce: StatementBlockInBraces
			reduce(13), // <<, reduce: StatementBlockInBraces
			reduce(13), // >>>, reduce: StatementBlockInBraces
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			reduce(13), // *, reduce: StatementBlockInBraces
			reduce(13), // /, reduce: StatementBlockInBraces
			reduce(13), // //, reduce: StatementBlockInBraces
			reduce(13), // %, reduce: StatementBlockInBraces
			reduce(13), // .*, reduce: StatementBlockInBraces
			reduce(13), // ./, reduce: StatementBlockInBraces
			reduce(13), // .//, reduce: StatementBlockInBraces
			reduce(13), // ., reduce: StatementBlockInBraces
			nil,        // !
			nil,        // ~
			reduce(13), // **, reduce: StatementBlockInBraces
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(13), // [, reduce: StatementBlockInBraces
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(13), // [[, reduce: StatementBlockInBraces
			reduce(13), // [[[, reduce: StatementBlockInBraces
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4447
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(325), // >, reduce: UnnamedFunctionDefinition
			reduce(325), // >>, reduce: UnnamedFunctionDefinition
			reduce(325), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(325), // ?, reduce: UnnamedFunctionDefinition
			reduce(325), // :, reduce: UnnamedFunctionDefinition
			reduce(325), // ||, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^, reduce: UnnamedFunctionDefinition
			reduce(325), // &&, reduce: UnnamedFunctionDefinition
			reduce(325), // ??, reduce: UnnamedFunctionDefinition
			reduce(325), // ???, reduce: UnnamedFunctionDefinition
			reduce(325), // =~, reduce: UnnamedFunctionDefinition
			reduce(325), // !=~, reduce: UnnamedFunctionDefinition
			reduce(325), // ==, reduce: UnnamedFunctionDefinition
			reduce(325), // !=, reduce: UnnamedFunctionDefinition
			reduce(325), // <=>, reduce: UnnamedFunctionDefinition
			reduce(325), // >=, reduce: UnnamedFunctionDefinition
			reduce(325), // <, reduce: UnnamedFunctionDefinition
			reduce(325), // <=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^, reduce: UnnamedFunctionDefinition
			reduce(325), // &, reduce: UnnamedFunctionDefinition
			reduce(325), // <<, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>, reduce: UnnamedFunctionDefinition
			reduce(325), // +, reduce: UnnamedFunctionDefinition
			reduce(325), // -, reduce: UnnamedFunctionDefinition
			reduce(325), // .+, reduce: UnnamedFunctionDefinition
			reduce(325), // .-, reduce: UnnamedFunctionDefinition
			reduce(325), // *, reduce: UnnamedFunctionDefinition
			reduce(325), // /, reduce: UnnamedFunctionDefinition
			reduce(325), // //, reduce: UnnamedFunctionDefinition
			reduce(325), // %, reduce: UnnamedFunctionDefinition
			reduce(325), // .*, reduce: UnnamedFunctionDefinition
			reduce(325), // ./, reduce: UnnamedFunctionDefinition
			reduce(325), // .//, reduce: UnnamedFunctionDefinition
			reduce(325), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(325), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(325), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(325), // [[, reduce: UnnamedFunctionDefinition
			reduce(325), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4448
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(108), // ;, reduce: PositionalFieldValue
			nil,         // {
			reduce(108), // }, reduce: PositionalFieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(108), // >, reduce: PositionalFieldValue
			reduce(108), // >>, reduce: PositionalFieldValue
			reduce(108), // |, reduce: PositionalFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(108), // ,, reduce: PositionalFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(108), // ?, reduce: PositionalFieldValue
			nil,         // :
			reduce(108), // ||, reduce: PositionalFieldValue
			reduce(108), // ^^, reduce: PositionalFieldValue
			reduce(108), // &&, reduce: PositionalFieldValue
			reduce(108), // ??, reduce: PositionalFieldValue
			reduce(108), // ???, reduce: PositionalFieldValue
			reduce(108), // =~, reduce: PositionalFieldValue
			reduce(108), // !=~, reduce: PositionalFieldValue
			reduce(108), // ==, reduce: PositionalFieldValue
			reduce(108), // !=, reduce: PositionalFieldValue
			reduce(108), // <=>, reduce: PositionalFieldValue
			reduce(108), // >=, reduce: PositionalFieldValue
			reduce(108), // <, reduce: PositionalFieldValue
			reduce(108), // <=, reduce: PositionalFieldValue
			reduce(108), // ^, reduce: PositionalFieldValue
			reduce(108), // &, reduce: PositionalFieldValue
			reduce(108), // <<, reduce: PositionalFieldValue
			reduce(108), // >>>, reduce: PositionalFieldValue
			reduce(108), // +, reduce: PositionalFieldValue
			reduce(108), // -, reduce: PositionalFieldValue
			reduce(108), // .+, reduce: PositionalFieldValue
			reduce(108), // .-, reduce: PositionalFieldValue
			reduce(108), // *, reduce: PositionalFieldValue
			reduce(108), // /, reduce: PositionalFieldValue
			reduce(108), // //, reduce: PositionalFieldValue
			reduce(108), // %, reduce: PositionalFieldValue
			reduce(108), // .*, reduce: PositionalFieldValue
			reduce(108), // ./, reduce: PositionalFieldValue
			reduce(108), // .//, reduce: PositionalFieldValue
			reduce(108), // ., reduce: PositionalFieldValue
			nil,         // !
			nil,         // ~
			reduce(108), // **, reduce: PositionalFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(108), // [, reduce: PositionalFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(108), // [[, reduce: PositionalFieldValue
			reduce(108), // [[[, reduce: PositionalFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4449
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(149), // ;, reduce: TernaryTerm
			nil,         // {
			reduce(149), // }, reduce: TernaryTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(149), // ,, reduce: TernaryTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4450
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4574), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4451
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(265), // ;, reduce: ArraySliceAccess
			nil,         // {
			reduce(265), // }, reduce: ArraySliceAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(265), // >, reduce: ArraySliceAccess
			reduce(265), // >>, reduce: ArraySliceAccess
			reduce(265), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(265), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(265), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(265), // ||, reduce: ArraySliceAccess
			reduce(265), // ^^, reduce: ArraySliceAccess
			reduce(265), // &&, reduce: ArraySliceAccess
			reduce(265), // ??, reduce: ArraySliceAccess
			reduce(265), // ???, reduce: ArraySliceAccess
			reduce(265), // =~, reduce: ArraySliceAccess
			reduce(265), // !=~, reduce: ArraySliceAccess
			reduce(265), // ==, reduce: ArraySliceAccess
			reduce(265), // !=, reduce: ArraySliceAccess
			reduce(265), // <=>, reduce: ArraySliceAccess
			reduce(265), // >=, reduce: ArraySliceAccess
			reduce(265), // <, reduce: ArraySliceAccess
			reduce(265), // <=, reduce: ArraySliceAccess
			reduce(265), // ^, reduce: ArraySliceAccess
			reduce(265), // &, reduce: ArraySliceAccess
			reduce(265), // <<, reduce: ArraySliceAccess
			reduce(265), // >>>, reduce: ArraySliceAccess
			reduce(265), // +, reduce: ArraySliceAccess
			reduce(265), // -, reduce: ArraySliceAccess
			reduce(265), // .+, reduce: ArraySliceAccess
			reduce(265), // .-, reduce: ArraySliceAccess
			reduce(265), // *, reduce: ArraySliceAccess
			reduce(265), // /, reduce: ArraySliceAccess
			reduce(265), // //, reduce: ArraySliceAccess
			reduce(265), // %, reduce: ArraySliceAccess
			reduce(265), // .*, reduce: ArraySliceAccess
			reduce(265), // ./, reduce: ArraySliceAccess
			reduce(265), // .//, reduce: ArraySliceAccess
			reduce(265), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(265), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(265), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(265), // [[, reduce: ArraySliceAccess
			reduce(265), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4452
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(264), // ;, reduce: ArraySliceAccess
			nil,         // {
			reduce(264), // }, reduce: ArraySliceAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(264), // >, reduce: ArraySliceAccess
			reduce(264), // >>, reduce: ArraySliceAccess
			reduce(264), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(264), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(264), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(264), // ||, reduce: ArraySliceAccess
			reduce(264), // ^^, reduce: ArraySliceAccess
			reduce(264), // &&, reduce: ArraySliceAccess
			reduce(264), // ??, reduce: ArraySliceAccess
			reduce(264), // ???, reduce: ArraySliceAccess
			reduce(264), // =~, reduce: ArraySliceAccess
			reduce(264), // !=~, reduce: ArraySliceAccess
			reduce(264), // ==, reduce: ArraySliceAccess
			reduce(264), // !=, reduce: ArraySliceAccess
			reduce(264), // <=>, reduce: ArraySliceAccess
			reduce(264), // >=, reduce: ArraySliceAccess
			reduce(264), // <, reduce: ArraySliceAccess
			reduce(264), // <=, reduce: ArraySliceAccess
			reduce(264), // ^, reduce: ArraySliceAccess
			reduce(264), // &, reduce: ArraySliceAccess
			reduce(264), // <<, reduce: ArraySliceAccess
			reduce(264), // >>>, reduce: ArraySliceAccess
			reduce(264), // +, reduce: ArraySliceAccess
			reduce(264), // -, reduce: ArraySliceAccess
			reduce(264), // .+, reduce: ArraySliceAccess
			reduce(264), // .-, reduce: ArraySliceAccess
			reduce(264), // *, reduce: ArraySliceAccess
			reduce(264), // /, reduce: ArraySliceAccess
			reduce(264), // //, reduce: ArraySliceAccess
			reduce(264), // %, reduce: ArraySliceAccess
			reduce(264), // .*, reduce: ArraySliceAccess
			reduce(264), // ./, reduce: ArraySliceAccess
			reduce(264), // .//, reduce: ArraySliceAccess
			reduce(264), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(264), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(264), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(264), // [[, reduce: ArraySliceAccess
			reduce(264), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4453
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(261), // ;, reduce: ArrayOrMapPositionalNameAccess
			nil,         // {
			reduce(261), // }, reduce: ArrayOrMapPositionalNameAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(261), // >, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |, reduce: ArrayOrMapPositionalNameAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(261), // ,, reduce: ArrayOrMapPositionalNameAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(261), // ?, reduce: ArrayOrMapPositionalNameAccess
			nil,         // :
			reduce(261), // ||, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // =~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ==, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .+, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .-, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .*, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ./, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .//, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ., reduce: ArrayOrMapPositionalNameAccess
			nil,         // !
			nil,         // ~
			reduce(261), // **, reduce: ArrayOrMapPositionalNameAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(261), // [, reduce: ArrayOrMapPositionalNameAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(261), // [[, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // [[[, reduce: ArrayOrMapPositionalNameAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4454
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4575), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4455
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(324), // ;, reduce: UnnamedFunctionDefinition
			nil,         // {
			reduce(324), // }, reduce: UnnamedFunctionDefinition
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(324), // >, reduce: UnnamedFunctionDefinition
			reduce(324), // >>, reduce: UnnamedFunctionDefinition
			reduce(324), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(324), // ,, reduce: UnnamedFunctionDefinition
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(324), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(324), // ||, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^, reduce: UnnamedFunctionDefinition
			reduce(324), // &&, reduce: UnnamedFunctionDefinition
			reduce(324), // ??, reduce: UnnamedFunctionDefinition
			reduce(324), // ???, reduce: UnnamedFunctionDefinition
			reduce(324), // =~, reduce: UnnamedFunctionDefinition
			reduce(324), // !=~, reduce: UnnamedFunctionDefinition
			reduce(324), // ==, reduce: UnnamedFunctionDefinition
			reduce(324), // !=, reduce: UnnamedFunctionDefinition
			reduce(324), // <=>, reduce: UnnamedFunctionDefinition
			reduce(324), // >=, reduce: UnnamedFunctionDefinition
			reduce(324), // <, reduce: UnnamedFunctionDefinition
			reduce(324), // <=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^, reduce: UnnamedFunctionDefinition
			reduce(324), // &, reduce: UnnamedFunctionDefinition
			reduce(324), // <<, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>, reduce: UnnamedFunctionDefinition
			reduce(324), // +, reduce: UnnamedFunctionDefinition
			reduce(324), // -, reduce: UnnamedFunctionDefinition
			reduce(324), // .+, reduce: UnnamedFunctionDefinition
			reduce(324), // .-, reduce: UnnamedFunctionDefinition
			reduce(324), // *, reduce: UnnamedFunctionDefinition
			reduce(324), // /, reduce: UnnamedFunctionDefinition
			reduce(324), // //, reduce: UnnamedFunctionDefinition
			reduce(324), // %, reduce: UnnamedFunctionDefinition
			reduce(324), // .*, reduce: UnnamedFunctionDefinition
			reduce(324), // ./, reduce: UnnamedFunctionDefinition
			reduce(324), // .//, reduce: UnnamedFunctionDefinition
			reduce(324), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(324), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(324), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(324), // [[, reduce: UnnamedFunctionDefinition
			reduce(324), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4456
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S4457
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4458
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(108), // ;, reduce: PositionalFieldValue
			nil,         // {
			reduce(108), // }, reduce: PositionalFieldValue
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(108), // >, reduce: PositionalFieldValue
			reduce(108), // >>, reduce: PositionalFieldValue
			reduce(108), // |, reduce: PositionalFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(108), // ?, reduce: PositionalFieldValue
			nil,         // :
			reduce(108), // ||, reduce: PositionalFieldValue
			reduce(108), // ^^, reduce: PositionalFieldValue
			reduce(108), // &&, reduce: PositionalFieldValue
			reduce(108), // ??, reduce: PositionalFieldValue
			reduce(108), // ???, reduce: PositionalFieldValue
			reduce(108), // =~, reduce: PositionalFieldValue
			reduce(108), // !=~, reduce: PositionalFieldValue
			reduce(108), // ==, reduce: PositionalFieldValue
			reduce(108), // !=, reduce: PositionalFieldValue
			reduce(108), // <=>, reduce: PositionalFieldValue
			reduce(108), // >=, reduce: PositionalFieldValue
			reduce(108), // <, reduce: PositionalFieldValue
			reduce(108), // <=, reduce: PositionalFieldValue
			reduce(108), // ^, reduce: PositionalFieldValue
			reduce(108), // &, reduce: PositionalFieldValue
			reduce(108), // <<, reduce: PositionalFieldValue
			reduce(108), // >>>, reduce: PositionalFieldValue
			reduce(108), // +, reduce: PositionalFieldValue
			reduce(108), // -, reduce: PositionalFieldValue
			reduce(108), // .+, reduce: PositionalFieldValue
			reduce(108), // .-, reduce: PositionalFieldValue
			reduce(108), // *, reduce: PositionalFieldValue
			reduce(108), // /, reduce: PositionalFieldValue
			reduce(108), // //, reduce: PositionalFieldValue
			reduce(108), // %, reduce: PositionalFieldValue
			reduce(108), // .*, reduce: PositionalFieldValue
			reduce(108), // ./, reduce: PositionalFieldValue
			reduce(108), // .//, reduce: PositionalFieldValue
			reduce(108), // ., reduce: PositionalFieldValue
			nil,         // !
			nil,         // ~
			reduce(108), // **, reduce: PositionalFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(108), // [, reduce: PositionalFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(108), // [[, reduce: PositionalFieldValue
			reduce(108), // [[[, reduce: PositionalFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4459
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(149), // ;, reduce: TernaryTerm
			nil,         // {
			reduce(149), // }, reduce: TernaryTerm
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4460
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4578), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4461
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(265), // ;, reduce: ArraySliceAccess
			nil,         // {
			reduce(265), // }, reduce: ArraySliceAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(265), // >, reduce: ArraySliceAccess
			reduce(265), // >>, reduce: ArraySliceAccess
			reduce(265), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(265), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(265), // ||, reduce: ArraySliceAccess
			reduce(265), // ^^, reduce: ArraySliceAccess
			reduce(265), // &&, reduce: ArraySliceAccess
			reduce(265), // ??, reduce: ArraySliceAccess
			reduce(265), // ???, reduce: ArraySliceAccess
			reduce(265), // =~, reduce: ArraySliceAccess
			reduce(265), // !=~, reduce: ArraySliceAccess
			reduce(265), // ==, reduce: ArraySliceAccess
			reduce(265), // !=, reduce: ArraySliceAccess
			reduce(265), // <=>, reduce: ArraySliceAccess
			reduce(265), // >=, reduce: ArraySliceAccess
			reduce(265), // <, reduce: ArraySliceAccess
			reduce(265), // <=, reduce: ArraySliceAccess
			reduce(265), // ^, reduce: ArraySliceAccess
			reduce(265), // &, reduce: ArraySliceAccess
			reduce(265), // <<, reduce: ArraySliceAccess
			reduce(265), // >>>, reduce: ArraySliceAccess
			reduce(265), // +, reduce: ArraySliceAccess
			reduce(265), // -, reduce: ArraySliceAccess
			reduce(265), // .+, reduce: ArraySliceAccess
			reduce(265), // .-, reduce: ArraySliceAccess
			reduce(265), // *, reduce: ArraySliceAccess
			reduce(265), // /, reduce: ArraySliceAccess
			reduce(265), // //, reduce: ArraySliceAccess
			reduce(265), // %, reduce: ArraySliceAccess
			reduce(265), // .*, reduce: ArraySliceAccess
			reduce(265), // ./, reduce: ArraySliceAccess
			reduce(265), // .//, reduce: ArraySliceAccess
			reduce(265), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(265), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(265), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(265), // [[, reduce: ArraySliceAccess
			reduce(265), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4462
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(264), // ;, reduce: ArraySliceAccess
			nil,         // {
			reduce(264), // }, reduce: ArraySliceAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(264), // >, reduce: ArraySliceAccess
			reduce(264), // >>, reduce: ArraySliceAccess
			reduce(264), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(264), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(264), // ||, reduce: ArraySliceAccess
			reduce(264), // ^^, reduce: ArraySliceAccess
			reduce(264), // &&, reduce: ArraySliceAccess
			reduce(264), // ??, reduce: ArraySliceAccess
			reduce(264), // ???, reduce: ArraySliceAccess
			reduce(264), // =~, reduce: ArraySliceAccess
			reduce(264), // !=~, reduce: ArraySliceAccess
			reduce(264), // ==, reduce: ArraySliceAccess
			reduce(264), // !=, reduce: ArraySliceAccess
			reduce(264), // <=>, reduce: ArraySliceAccess
			reduce(264), // >=, reduce: ArraySliceAccess
			reduce(264), // <, reduce: ArraySliceAccess
			reduce(264), // <=, reduce: ArraySliceAccess
			reduce(264), // ^, reduce: ArraySliceAccess
			reduce(264), // &, reduce: ArraySliceAccess
			reduce(264), // <<, reduce: ArraySliceAccess
			reduce(264), // >>>, reduce: ArraySliceAccess
			reduce(264), // +, reduce: ArraySliceAccess
			reduce(264), // -, reduce: ArraySliceAccess
			reduce(264), // .+, reduce: ArraySliceAccess
			reduce(264), // .-, reduce: ArraySliceAccess
			reduce(264), // *, reduce: ArraySliceAccess
			reduce(264), // /, reduce: ArraySliceAccess
			reduce(264), // //, reduce: ArraySliceAccess
			reduce(264), // %, reduce: ArraySliceAccess
			reduce(264), // .*, reduce: ArraySliceAccess
			reduce(264), // ./, reduce: ArraySliceAccess
			reduce(264), // .//, reduce: ArraySliceAccess
			reduce(264), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(264), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(264), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(264), // [[, reduce: ArraySliceAccess
			reduce(264), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4463
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(261), // ;, reduce: ArrayOrMapPositionalNameAccess
			nil,         // {
			reduce(261), // }, reduce: ArrayOrMapPositionalNameAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(261), // >, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |, reduce: ArrayOrMapPositionalNameAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(261), // ?, reduce: ArrayOrMapPositionalNameAccess
			nil,         // :
			reduce(261), // ||, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // =~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ==, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .+, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .-, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .*, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ./, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .//, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ., reduce: ArrayOrMapPositionalNameAccess
			nil,         // !
			nil,         // ~
			reduce(261), // **, reduce: ArrayOrMapPositionalNameAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(261), // [, reduce: ArrayOrMapPositionalNameAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(261), // [[, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // [[[, reduce: ArrayOrMapPositionalNameAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4464
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4579), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4465
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(324), // ;, reduce: UnnamedFunctionDefinition
			nil,         // {
			reduce(324), // }, reduce: UnnamedFunctionDefinition
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(324), // >, reduce: UnnamedFunctionDefinition
			reduce(324), // >>, reduce: UnnamedFunctionDefinition
			reduce(324), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(324), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(324), // ||, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^, reduce: UnnamedFunctionDefinition
			reduce(324), // &&, reduce: UnnamedFunctionDefinition
			reduce(324), // ??, reduce: UnnamedFunctionDefinition
			reduce(324), // ???, reduce: UnnamedFunctionDefinition
			reduce(324), // =~, reduce: UnnamedFunctionDefinition
			reduce(324), // !=~, reduce: UnnamedFunctionDefinition
			reduce(324), // ==, reduce: UnnamedFunctionDefinition
			reduce(324), // !=, reduce: UnnamedFunctionDefinition
			reduce(324), // <=>, reduce: UnnamedFunctionDefinition
			reduce(324), // >=, reduce: UnnamedFunctionDefinition
			reduce(324), // <, reduce: UnnamedFunctionDefinition
			reduce(324), // <=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^, reduce: UnnamedFunctionDefinition
			reduce(324), // &, reduce: UnnamedFunctionDefinition
			reduce(324), // <<, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>, reduce: UnnamedFunctionDefinition
			reduce(324), // +, reduce: UnnamedFunctionDefinition
			reduce(324), // -, reduce: UnnamedFunctionDefinition
			reduce(324), // .+, reduce: UnnamedFunctionDefinition
			reduce(324), // .-, reduce: UnnamedFunctionDefinition
			reduce(324), // *, reduce: UnnamedFunctionDefinition
			reduce(324), // /, reduce: UnnamedFunctionDefinition
			reduce(324), // //, reduce: UnnamedFunctionDefinition
			reduce(324), // %, reduce: UnnamedFunctionDefinition
			reduce(324), // .*, reduce: UnnamedFunctionDefinition
			reduce(324), // ./, reduce: UnnamedFunctionDefinition
			reduce(324), // .//, reduce: UnnamedFunctionDefinition
			reduce(324), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(324), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(324), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(324), // [[, reduce: UnnamedFunctionDefinition
			reduce(324), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4466
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S4467
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4468
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(73), // ;, reduce: EmitStatement
			nil,        // {
			reduce(73), // }, reduce: EmitStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4469
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(71),  // ;, reduce: EmitStatement
			nil,         // {
			reduce(71),  // }, reduce: EmitStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(4582), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4470
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(98), // ;, reduce: EmitKeys
			nil,        // {
			reduce(98), // }, reduce: EmitKeys
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4471
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(74), // ;, reduce: EmitStatement
			nil,        // {
			reduce(74), // }, reduce: EmitStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4472
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(81), // ;, reduce: EmitPStatement
			nil,        // {
			reduce(81), // }, reduce: EmitPStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4473
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(79),  // ;, reduce: EmitPStatement
			nil,         // {
			reduce(79),  // }, reduce: EmitPStatement
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(4583), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4474
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(82), // ;, reduce: EmitPStatement
			nil,        // {
			reduce(82), // }, reduce: EmitPStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4475
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(263), // ;, reduce: ArraySliceAccess
			reduce(263), // {, reduce: ArraySliceAccess
			reduce(263), // }, reduce: ArraySliceAccess
			reduce(263), // =, reduce: ArraySliceAccess
			nil,         // unset
			nil,         // filter
			reduce(263), // >, reduce: ArraySliceAccess
			reduce(263), // >>, reduce: ArraySliceAccess
			reduce(263), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(263), // ||=, reduce: ArraySliceAccess
			reduce(263), // ^^=, reduce: ArraySliceAccess
			reduce(263), // &&=, reduce: ArraySliceAccess
			reduce(263), // ??=, reduce: ArraySliceAccess
			reduce(263), // ???=, reduce: ArraySliceAccess
			reduce(263), // |=, reduce: ArraySliceAccess
			reduce(263), // &=, reduce: ArraySliceAccess
			reduce(263), // ^=, reduce: ArraySliceAccess
			reduce(263), // <<=, reduce: ArraySliceAccess
			reduce(263), // >>=, reduce: ArraySliceAccess
			reduce(263), // >>>=, reduce: ArraySliceAccess
			reduce(263), // +=, reduce: ArraySliceAccess
			reduce(263), // .=, reduce: ArraySliceAccess
			reduce(263), // -=, reduce: ArraySliceAccess
			reduce(263), // *=, reduce: ArraySliceAccess
			reduce(263), // /=, reduce: ArraySliceAccess
			reduce(263), // //=, reduce: ArraySliceAccess
			reduce(263), // %=, reduce: ArraySliceAccess
			reduce(263), // **=, reduce: ArraySliceAccess
			reduce(263), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(263), // ||, reduce: ArraySliceAccess
			reduce(263), // ^^, reduce: ArraySliceAccess
			reduce(263), // &&, reduce: ArraySliceAccess
			reduce(263), // ??, reduce: ArraySliceAccess
			reduce(263), // ???, reduce: ArraySliceAccess
			reduce(263), // =~, reduce: ArraySliceAccess
			reduce(263), // !=~, reduce: ArraySliceAccess
			reduce(263), // ==, reduce: ArraySliceAccess
			reduce(263), // !=, reduce: ArraySliceAccess
			reduce(263), // <=>, reduce: ArraySliceAccess
			reduce(263), // >=, reduce: ArraySliceAccess
			reduce(263), // <, reduce: ArraySliceAccess
			reduce(263), // <=, reduce: ArraySliceAccess
			reduce(263), // ^, reduce: ArraySliceAccess
			reduce(263), // &, reduce: ArraySliceAccess
			reduce(263), // <<, reduce: ArraySliceAccess
			reduce(263), // >>>, reduce: ArraySliceAccess
			reduce(263), // +, reduce: ArraySliceAccess
			reduce(263), // -, reduce: ArraySliceAccess
			reduce(263), // .+, reduce: ArraySliceAccess
			reduce(263), // .-, reduce: ArraySliceAccess
			reduce(263), // *, reduce: ArraySliceAccess
			reduce(263), // /, reduce: ArraySliceAccess
			reduce(263), // //, reduce: ArraySliceAccess
			reduce(263), // %, reduce: ArraySliceAccess
			reduce(263), // .*, reduce: ArraySliceAccess
			reduce(263), // ./, reduce: ArraySliceAccess
			reduce(263), // .//, reduce: ArraySliceAccess
			reduce(263), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(263), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(263), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(263), // [[, reduce: ArraySliceAccess
			reduce(263), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4476
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(262), // ;, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // {, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // }, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // =, reduce: ArrayOrMapPositionalValueAccess
			nil,         // unset
			nil,         // filter
			reduce(262), // >, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |, reduce: ArrayOrMapPositionalValueAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(262), // ||=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // **=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ?, reduce: ArrayOrMapPositionalValueAccess
			nil,         // :
			reduce(262), // ||, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // =~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ==, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .+, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .-, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .*, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ./, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .//, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ., reduce: ArrayOrMapPositionalValueAccess
			nil,         // !
			nil,         // ~
			reduce(262), // **, reduce: ArrayOrMapPositionalValueAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(262), // [, reduce: ArrayOrMapPositionalValueAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(262), // [[, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // [[[, reduce: ArrayOrMapPositionalValueAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4477
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(295), // ;, reduce: ElifBlock
			reduce(295), // {, reduce: ElifBlock
			reduce(295), // }, reduce: ElifBlock
			nil,         // =
			reduce(295), // unset, reduce: ElifBlock
			reduce(295), // filter, reduce: ElifBlock
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(295), // print, reduce: ElifBlock
			nil,         // ,
			reduce(295), // printn, reduce: ElifBlock
			reduce(295), // eprint, reduce: ElifBlock
			reduce(295), // eprintn, reduce: ElifBlock
			reduce(295), // dump, reduce: ElifBlock
			reduce(295), // edump, reduce: ElifBlock
			reduce(295), // tee, reduce: ElifBlock
			reduce(295), // emitf, reduce: ElifBlock
			reduce(295), // emit1, reduce: ElifBlock
			reduce(295), // emit, reduce: ElifBlock
			reduce(295), // (, reduce: ElifBlock
			nil,         // )
			reduce(295), // emitp, reduce: ElifBlock
			reduce(295), // field_name, reduce: ElifBlock
			reduce(295), // $[, reduce: ElifBlock
			nil,         // ]
			reduce(295), // braced_field_name, reduce: ElifBlock
			reduce(295), // $[[, reduce: ElifBlock
			reduce(295), // $[[[, reduce: ElifBlock
			reduce(295), // full_srec, reduce: ElifBlock
			reduce(295), // oosvar_name, reduce: ElifBlock
			reduce(295), // @[, reduce: ElifBlock
			reduce(295), // braced_oosvar_name, reduce: ElifBlock
			reduce(295), // full_oosvar, reduce: ElifBlock
			reduce(295), // all, reduce: ElifBlock
			reduce(295), // non_sigil_name, reduce: ElifBlock
			reduce(295), // arr, reduce: ElifBlock
			reduce(295), // bool, reduce: ElifBlock
			reduce(295), // float, reduce: ElifBlock
			reduce(295), // int, reduce: ElifBlock
			reduce(295), // map, reduce: ElifBlock
			reduce(295), // num, reduce: ElifBlock
			reduce(295), // str, reduce: ElifBlock
			reduce(295), // var, reduce: ElifBlock
			reduce(295), // funct, reduce: ElifBlock
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(295), // +, reduce: ElifBlock
			reduce(295), // -, reduce: ElifBlock
			reduce(295), // .+, reduce: ElifBlock
			reduce(295), // .-, reduce: ElifBlock
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(295), // !, reduce: ElifBlock
			reduce(295), // ~, reduce: ElifBlock
			nil,         // **
			reduce(295), // string_literal, reduce: ElifBlock
			reduce(295), // regex_case_insensitive, reduce: ElifBlock
			reduce(295), // int_literal, reduce: ElifBlock
			reduce(295), // float_literal, reduce: ElifBlock
			reduce(295), // boolean_literal, reduce: ElifBlock
			reduce(295), // null_literal, reduce: ElifBlock
			reduce(295), // inf_literal, reduce: ElifBlock
			reduce(295), // nan_literal, reduce: ElifBlock
			reduce(295), // const_M_PI, reduce: ElifBlock
			reduce(295), // const_M_E, reduce: ElifBlock
			reduce(295), // panic, reduce: ElifBlock
			reduce(295), // [, reduce: ElifBlock
			reduce(295), // ctx_IPS, reduce: ElifBlock
			reduce(295), // ctx_IFS, reduce: ElifBlock
			reduce(295), // ctx_IRS, reduce: ElifBlock
			reduce(295), // ctx_OPS, reduce: ElifBlock
			reduce(295), // ctx_OFS, reduce: ElifBlock
			reduce(295), // ctx_ORS, reduce: ElifBlock
			reduce(295), // ctx_FLATSEP, reduce: ElifBlock
			reduce(295), // ctx_NF, reduce: ElifBlock
			reduce(295), // ctx_NR, reduce: ElifBlock
			reduce(295), // ctx_FNR, reduce: ElifBlock
			reduce(295), // ctx_FILENAME, reduce: ElifBlock
			reduce(295), // ctx_FILENUM, reduce: ElifBlock
			reduce(295), // env, reduce: ElifBlock
			nil,         // [[
			nil,         // [[[
			reduce(295), // call, reduce: ElifBlock
			reduce(295), // begin, reduce: ElifBlock
			reduce(295), // end, reduce: ElifBlock
			reduce(295), // if, reduce: ElifBlock
			reduce(295), // elif, reduce: ElifBlock
			reduce(295), // else, reduce: ElifBlock
			reduce(295), // while, reduce: ElifBlock
			reduce(295), // do, reduce: ElifBlock
			reduce(295), // for, reduce: ElifBlock
			nil,         // in
			reduce(295), // break, reduce: ElifBlock
			reduce(295), // continue, reduce: ElifBlock
			reduce(295), // func, reduce: ElifBlock
			reduce(295), // subr, reduce: ElifBlock
			reduce(295), // return, reduce: ElifBlock
		},
	},
	actionRow{ // S4478
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4584), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4479
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(298), // ;, reduce: DoWhileLoop
			nil,         // {
			reduce(298), // }, reduce: DoWhileLoop
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4480
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4481
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2042), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4482
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			shift(3561), // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4483
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(317), // ,, reduce: TripleForUpdate
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			reduce(317), // ), reduce: TripleForUpdate
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4484
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4589), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4485
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4255), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4486
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(322), // ;, reduce: NamedFunctionDefinition
			reduce(322), // {, reduce: NamedFunctionDefinition
			reduce(322), // }, reduce: NamedFunctionDefinition
			nil,         // =
			reduce(322), // unset, reduce: NamedFunctionDefinition
			reduce(322), // filter, reduce: NamedFunctionDefinition
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(322), // print, reduce: NamedFunctionDefinition
			nil,         // ,
			reduce(322), // printn, reduce: NamedFunctionDefinition
			reduce(322), // eprint, reduce: NamedFunctionDefinition
			reduce(322), // eprintn, reduce: NamedFunctionDefinition
			reduce(322), // dump, reduce: NamedFunctionDefinition
			reduce(322), // edump, reduce: NamedFunctionDefinition
			reduce(322), // tee, reduce: NamedFunctionDefinition
			reduce(322), // emitf, reduce: NamedFunctionDefinition
			reduce(322), // emit1, reduce: NamedFunctionDefinition
			reduce(322), // emit, reduce: NamedFunctionDefinition
			reduce(322), // (, reduce: NamedFunctionDefinition
			nil,         // )
			reduce(322), // emitp, reduce: NamedFunctionDefinition
			reduce(322), // field_name, reduce: NamedFunctionDefinition
			reduce(322), // $[, reduce: NamedFunctionDefinition
			nil,         // ]
			reduce(322), // braced_field_name, reduce: NamedFunctionDefinition
			reduce(322), // $[[, reduce: NamedFunctionDefinition
			reduce(322), // $[[[, reduce: NamedFunctionDefinition
			reduce(322), // full_srec, reduce: NamedFunctionDefinition
			reduce(322), // oosvar_name, reduce: NamedFunctionDefinition
			reduce(322), // @[, reduce: NamedFunctionDefinition
			reduce(322), // braced_oosvar_name, reduce: NamedFunctionDefinition
			reduce(322), // full_oosvar, reduce: NamedFunctionDefinition
			reduce(322), // all, reduce: NamedFunctionDefinition
			reduce(322), // non_sigil_name, reduce: NamedFunctionDefinition
			reduce(322), // arr, reduce: NamedFunctionDefinition
			reduce(322), // bool, reduce: NamedFunctionDefinition
			reduce(322), // float, reduce: NamedFunctionDefinition
			reduce(322), // int, reduce: NamedFunctionDefinition
			reduce(322), // map, reduce: NamedFunctionDefinition
			reduce(322), // num, reduce: NamedFunctionDefinition
			reduce(322), // str, reduce: NamedFunctionDefinition
			reduce(322), // var, reduce: NamedFunctionDefinition
			reduce(322), // funct, reduce: NamedFunctionDefinition
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(322), // +, reduce: NamedFunctionDefinition
			reduce(322), // -, reduce: NamedFunctionDefinition
			reduce(322), // .+, reduce: NamedFunctionDefinition
			reduce(322), // .-, reduce: NamedFunctionDefinition
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(322), // !, reduce: NamedFunctionDefinition
			reduce(322), // ~, reduce: NamedFunctionDefinition
			nil,         // **
			reduce(322), // string_literal, reduce: NamedFunctionDefinition
			reduce(322), // regex_case_insensitive, reduce: NamedFunctionDefinition
			reduce(322), // int_literal, reduce: NamedFunctionDefinition
			reduce(322), // float_literal, reduce: NamedFunctionDefinition
			reduce(322), // boolean_literal, reduce: NamedFunctionDefinition
			reduce(322), // null_literal, reduce: NamedFunctionDefinition
			reduce(322), // inf_literal, reduce: NamedFunctionDefinition
			reduce(322), // nan_literal, reduce: NamedFunctionDefinition
			reduce(322), // const_M_PI, reduce: NamedFunctionDefinition
			reduce(322), // const_M_E, reduce: NamedFunctionDefinition
			reduce(322), // panic, reduce: NamedFunctionDefinition
			reduce(322), // [, reduce: NamedFunctionDefinition
			reduce(322), // ctx_IPS, reduce: NamedFunctionDefinition
			reduce(322), // ctx_IFS, reduce: NamedFunctionDefinition
			reduce(322), // ctx_IRS, reduce: NamedFunctionDefinition
			reduce(322), // ctx_OPS, reduce: NamedFunctionDefinition
			reduce(322), // ctx_OFS, reduce: NamedFunctionDefinition
			reduce(322), // ctx_ORS, reduce: NamedFunctionDefinition
			reduce(322), // ctx_FLATSEP, reduce: NamedFunctionDefinition
			reduce(322), // ctx_NF, reduce: NamedFunctionDefinition
			reduce(322), // ctx_NR, reduce: NamedFunctionDefinition
			reduce(322), // ctx_FNR, reduce: NamedFunctionDefinition
			reduce(322), // ctx_FILENAME, reduce: NamedFunctionDefinition
			reduce(322), // ctx_FILENUM, reduce: NamedFunctionDefinition
			reduce(322), // env, reduce: NamedFunctionDefinition
			nil,         // [[
			nil,         // [[[
			reduce(322), // call, reduce: NamedFunctionDefinition
			reduce(322), // begin, reduce: NamedFunctionDefinition
			reduce(322), // end, reduce: NamedFunctionDefinition
			reduce(322), // if, reduce: NamedFunctionDefinition
			nil,         // elif
			nil,         // else
			reduce(322), // while, reduce: NamedFunctionDefinition
			reduce(322), // do, reduce: NamedFunctionDefinition
			reduce(322), // for, reduce: NamedFunctionDefinition
			nil,         // in
			reduce(322), // break, reduce: NamedFunctionDefinition
			reduce(322), // continue, reduce: NamedFunctionDefinition
			reduce(322), // func, reduce: NamedFunctionDefinition
			reduce(322), // subr, reduce: NamedFunctionDefinition
			reduce(322), // return, reduce: NamedFunctionDefinition
		},
	},
	actionRow{ // S4487
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4488
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(326), // ;, reduce: SubroutineDefinition
			reduce(326), // {, reduce: SubroutineDefinition
			reduce(326), // }, reduce: SubroutineDefinition
			nil,         // =
			reduce(326), // unset, reduce: SubroutineDefinition
			reduce(326), // filter, reduce: SubroutineDefinition
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(326), // print, reduce: SubroutineDefinition
			nil,         // ,
			reduce(326), // printn, reduce: SubroutineDefinition
			reduce(326), // eprint, reduce: SubroutineDefinition
			reduce(326), // eprintn, reduce: SubroutineDefinition
			reduce(326), // dump, reduce: SubroutineDefinition
			reduce(326), // edump, reduce: SubroutineDefinition
			reduce(326), // tee, reduce: SubroutineDefinition
			reduce(326), // emitf, reduce: SubroutineDefinition
			reduce(326), // emit1, reduce: SubroutineDefinition
			reduce(326), // emit, reduce: SubroutineDefinition
			reduce(326), // (, reduce: SubroutineDefinition
			nil,         // )
			reduce(326), // emitp, reduce: SubroutineDefinition
			reduce(326), // field_name, reduce: SubroutineDefinition
			reduce(326), // $[, reduce: SubroutineDefinition
			nil,         // ]
			reduce(326), // braced_field_name, reduce: SubroutineDefinition
			reduce(326), // $[[, reduce: SubroutineDefinition
			reduce(326), // $[[[, reduce: SubroutineDefinition
			reduce(326), // full_srec, reduce: SubroutineDefinition
			reduce(326), // oosvar_name, reduce: SubroutineDefinition
			reduce(326), // @[, reduce: SubroutineDefinition
			reduce(326), // braced_oosvar_name, reduce: SubroutineDefinition
			reduce(326), // full_oosvar, reduce: SubroutineDefinition
			reduce(326), // all, reduce: SubroutineDefinition
			reduce(326), // non_sigil_name, reduce: SubroutineDefinition
			reduce(326), // arr, reduce: SubroutineDefinition
			reduce(326), // bool, reduce: SubroutineDefinition
			reduce(326), // float, reduce: SubroutineDefinition
			reduce(326), // int, reduce: SubroutineDefinition
			reduce(326), // map, reduce: SubroutineDefinition
			reduce(326), // num, reduce: SubroutineDefinition
			reduce(326), // str, reduce: SubroutineDefinition
			reduce(326), // var, reduce: SubroutineDefinition
			reduce(326), // funct, reduce: SubroutineDefinition
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(326), // +, reduce: SubroutineDefinition
			reduce(326), // -, reduce: SubroutineDefinition
			reduce(326), // .+, reduce: SubroutineDefinition
			reduce(326), // .-, reduce: SubroutineDefinition
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(326), // !, reduce: SubroutineDefinition
			reduce(326), // ~, reduce: SubroutineDefinition
			nil,         // **
			reduce(326), // string_literal, reduce: SubroutineDefinition
			reduce(326), // regex_case_insensitive, reduce: SubroutineDefinition
			reduce(326), // int_literal, reduce: SubroutineDefinition
			reduce(326), // float_literal, reduce: SubroutineDefinition
			reduce(326), // boolean_literal, reduce: SubroutineDefinition
			reduce(326), // null_literal, reduce: SubroutineDefinition
			reduce(326), // inf_literal, reduce: SubroutineDefinition
			reduce(326), // nan_literal, reduce: SubroutineDefinition
			reduce(326), // const_M_PI, reduce: SubroutineDefinition
			reduce(326), // const_M_E, reduce: SubroutineDefinition
			reduce(326), // panic, reduce: SubroutineDefinition
			reduce(326), // [, reduce: SubroutineDefinition
			reduce(326), // ctx_IPS, reduce: SubroutineDefinition
			reduce(326), // ctx_IFS, reduce: SubroutineDefinition
			reduce(326), // ctx_IRS, reduce: SubroutineDefinition
			reduce(326), // ctx_OPS, reduce: SubroutineDefinition
			reduce(326), // ctx_OFS, reduce: SubroutineDefinition
			reduce(326), // ctx_ORS, reduce: SubroutineDefinition
			reduce(326), // ctx_FLATSEP, reduce: SubroutineDefinition
			reduce(326), // ctx_NF, reduce: SubroutineDefinition
			reduce(326), // ctx_NR, reduce: SubroutineDefinition
			reduce(326), // ctx_FNR, reduce: SubroutineDefinition
			reduce(326), // ctx_FILENAME, reduce: SubroutineDefinition
			reduce(326), // ctx_FILENUM, reduce: SubroutineDefinition
			reduce(326), // env, reduce: SubroutineDefinition
			nil,         // [[
			nil,         // [[[
			reduce(326), // call, reduce: SubroutineDefinition
			reduce(326), // begin, reduce: SubroutineDefinition
			reduce(326), // end, reduce: SubroutineDefinition
			reduce(326), // if, reduce: SubroutineDefinition
			nil,         // elif
			nil,         // else
			reduce(326), // while, reduce: SubroutineDefinition
			reduce(326), // do, reduce: SubroutineDefinition
			reduce(326), // for, reduce: SubroutineDefinition
			nil,         // in
			reduce(326), // break, reduce: SubroutineDefinition
			reduce(326), // continue, reduce: SubroutineDefinition
			reduce(326), // func, reduce: SubroutineDefinition
			reduce(326), // subr, reduce: SubroutineDefinition
			reduce(326), // return, reduce: SubroutineDefinition
		},
	},
	actionRow{ // S4489
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(13), // $, reduce: StatementBlockInBraces
			nil,        // empty
			reduce(13), // ;, reduce: StatementBlockInBraces
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(13), // >, reduce: StatementBlockInBraces
			reduce(13), // >>, reduce: StatementBlockInBraces
			reduce(13), // |, reduce: StatementBlockInBraces
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(13), // ,, reduce: StatementBlockInBraces
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(13), // ?, reduce: StatementBlockInBraces
			nil,        // :
			reduce(13), // ||, reduce: StatementBlockInBraces
			reduce(13), // ^^, reduce: StatementBlockInBraces
			reduce(13), // &&, reduce: StatementBlockInBraces
			reduce(13), // ??, reduce: StatementBlockInBraces
			reduce(13), // ???, reduce: StatementBlockInBraces
			reduce(13), // =~, reduce: StatementBlockInBraces
			reduce(13), // !=~, reduce: StatementBlockInBraces
			reduce(13), // ==, reduce: StatementBlockInBraces
			reduce(13), // !=, reduce: StatementBlockInBraces
			reduce(13), // <=>, reduce: StatementBlockInBraces
			reduce(13), // >=, reduce: StatementBlockInBraces
			reduce(13), // <, reduce: StatementBlockInBraces
			reduce(13), // <=, reduce: StatementBlockInBraces
			reduce(13), // ^, reduce: StatementBlockInBraces
			reduce(13), // &, reduce: StatementBlockInBraces
			reduce(13), // <<, reduce: StatementBlockInBraces
			reduce(13), // >>>, reduce: StatementBlockInBraces
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			reduce(13), // *, reduce: StatementBlockInBraces
			reduce(13), // /, reduce: StatementBlockInBraces
			reduce(13), // //, reduce: StatementBlockInBraces
			reduce(13), // %, reduce: StatementBlockInBraces
			reduce(13), // .*, reduce: StatementBlockInBraces
			reduce(13), // ./, reduce: StatementBlockInBraces
			reduce(13), // .//, reduce: StatementBlockInBraces
			reduce(13), // ., reduce: StatementBlockInBraces
			nil,        // !
			nil,        // ~
			reduce(13), // **, reduce: StatementBlockInBraces
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(13), // [, reduce: StatementBlockInBraces
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(13), // [[, reduce: StatementBlockInBraces
			reduce(13), // [[[, reduce: StatementBlockInBraces
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4490
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(325), // $, reduce: UnnamedFunctionDefinition
			nil,         // empty
			reduce(325), // ;, reduce: UnnamedFunctionDefinition
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(325), // >, reduce: UnnamedFunctionDefinition
			reduce(325), // >>, reduce: UnnamedFunctionDefinition
			reduce(325), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(325), // ,, reduce: UnnamedFunctionDefinition
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(325), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(325), // ||, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^, reduce: UnnamedFunctionDefinition
			reduce(325), // &&, reduce: UnnamedFunctionDefinition
			reduce(325), // ??, reduce: UnnamedFunctionDefinition
			reduce(325), // ???, reduce: UnnamedFunctionDefinition
			reduce(325), // =~, reduce: UnnamedFunctionDefinition
			reduce(325), // !=~, reduce: UnnamedFunctionDefinition
			reduce(325), // ==, reduce: UnnamedFunctionDefinition
			reduce(325), // !=, reduce: UnnamedFunctionDefinition
			reduce(325), // <=>, reduce: UnnamedFunctionDefinition
			reduce(325), // >=, reduce: UnnamedFunctionDefinition
			reduce(325), // <, reduce: UnnamedFunctionDefinition
			reduce(325), // <=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^, reduce: UnnamedFunctionDefinition
			reduce(325), // &, reduce: UnnamedFunctionDefinition
			reduce(325), // <<, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>, reduce: UnnamedFunctionDefinition
			reduce(325), // +, reduce: UnnamedFunctionDefinition
			reduce(325), // -, reduce: UnnamedFunctionDefinition
			reduce(325), // .+, reduce: UnnamedFunctionDefinition
			reduce(325), // .-, reduce: UnnamedFunctionDefinition
			reduce(325), // *, reduce: UnnamedFunctionDefinition
			reduce(325), // /, reduce: UnnamedFunctionDefinition
			reduce(325), // //, reduce: UnnamedFunctionDefinition
			reduce(325), // %, reduce: UnnamedFunctionDefinition
			reduce(325), // .*, reduce: UnnamedFunctionDefinition
			reduce(325), // ./, reduce: UnnamedFunctionDefinition
			reduce(325), // .//, reduce: UnnamedFunctionDefinition
			reduce(325), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(325), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(325), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(325), // [[, reduce: UnnamedFunctionDefinition
			reduce(325), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4491
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(13), // $, reduce: StatementBlockInBraces
			nil,        // empty
			reduce(13), // ;, reduce: StatementBlockInBraces
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(13), // >, reduce: StatementBlockInBraces
			reduce(13), // >>, reduce: StatementBlockInBraces
			reduce(13), // |, reduce: StatementBlockInBraces
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(13), // ?, reduce: StatementBlockInBraces
			nil,        // :
			reduce(13), // ||, reduce: StatementBlockInBraces
			reduce(13), // ^^, reduce: StatementBlockInBraces
			reduce(13), // &&, reduce: StatementBlockInBraces
			reduce(13), // ??, reduce: StatementBlockInBraces
			reduce(13), // ???, reduce: StatementBlockInBraces
			reduce(13), // =~, reduce: StatementBlockInBraces
			reduce(13), // !=~, reduce: StatementBlockInBraces
			reduce(13), // ==, reduce: StatementBlockInBraces
			reduce(13), // !=, reduce: StatementBlockInBraces
			reduce(13), // <=>, reduce: StatementBlockInBraces
			reduce(13), // >=, reduce: StatementBlockInBraces
			reduce(13), // <, reduce: StatementBlockInBraces
			reduce(13), // <=, reduce: StatementBlockInBraces
			reduce(13), // ^, reduce: StatementBlockInBraces
			reduce(13), // &, reduce: StatementBlockInBraces
			reduce(13), // <<, reduce: StatementBlockInBraces
			reduce(13), // >>>, reduce: StatementBlockInBraces
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			reduce(13), // *, reduce: StatementBlockInBraces
			reduce(13), // /, reduce: StatementBlockInBraces
			reduce(13), // //, reduce: StatementBlockInBraces
			reduce(13), // %, reduce: StatementBlockInBraces
			reduce(13), // .*, reduce: StatementBlockInBraces
			reduce(13), // ./, reduce: StatementBlockInBraces
			reduce(13), // .//, reduce: StatementBlockInBraces
			reduce(13), // ., reduce: StatementBlockInBraces
			nil,        // !
			nil,        // ~
			reduce(13), // **, reduce: StatementBlockInBraces
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(13), // [, reduce: StatementBlockInBraces
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(13), // [[, reduce: StatementBlockInBraces
			reduce(13), // [[[, reduce: StatementBlockInBraces
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4492
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(325), // $, reduce: UnnamedFunctionDefinition
			nil,         // empty
			reduce(325), // ;, reduce: UnnamedFunctionDefinition
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(325), // >, reduce: UnnamedFunctionDefinition
			reduce(325), // >>, reduce: UnnamedFunctionDefinition
			reduce(325), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(325), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(325), // ||, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^, reduce: UnnamedFunctionDefinition
			reduce(325), // &&, reduce: UnnamedFunctionDefinition
			reduce(325), // ??, reduce: UnnamedFunctionDefinition
			reduce(325), // ???, reduce: UnnamedFunctionDefinition
			reduce(325), // =~, reduce: UnnamedFunctionDefinition
			reduce(325), // !=~, reduce: UnnamedFunctionDefinition
			reduce(325), // ==, reduce: UnnamedFunctionDefinition
			reduce(325), // !=, reduce: UnnamedFunctionDefinition
			reduce(325), // <=>, reduce: UnnamedFunctionDefinition
			reduce(325), // >=, reduce: UnnamedFunctionDefinition
			reduce(325), // <, reduce: UnnamedFunctionDefinition
			reduce(325), // <=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^, reduce: UnnamedFunctionDefinition
			reduce(325), // &, reduce: UnnamedFunctionDefinition
			reduce(325), // <<, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>, reduce: UnnamedFunctionDefinition
			reduce(325), // +, reduce: UnnamedFunctionDefinition
			reduce(325), // -, reduce: UnnamedFunctionDefinition
			reduce(325), // .+, reduce: UnnamedFunctionDefinition
			reduce(325), // .-, reduce: UnnamedFunctionDefinition
			reduce(325), // *, reduce: UnnamedFunctionDefinition
			reduce(325), // /, reduce: UnnamedFunctionDefinition
			reduce(325), // //, reduce: UnnamedFunctionDefinition
			reduce(325), // %, reduce: UnnamedFunctionDefinition
			reduce(325), // .*, reduce: UnnamedFunctionDefinition
			reduce(325), // ./, reduce: UnnamedFunctionDefinition
			reduce(325), // .//, reduce: UnnamedFunctionDefinition
			reduce(325), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(325), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(325), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(325), // [[, reduce: UnnamedFunctionDefinition
			reduce(325), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4493
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(263), // >, reduce: ArraySliceAccess
			reduce(263), // >>, reduce: ArraySliceAccess
			reduce(263), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(263), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(263), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(263), // ||, reduce: ArraySliceAccess
			reduce(263), // ^^, reduce: ArraySliceAccess
			reduce(263), // &&, reduce: ArraySliceAccess
			reduce(263), // ??, reduce: ArraySliceAccess
			reduce(263), // ???, reduce: ArraySliceAccess
			reduce(263), // =~, reduce: ArraySliceAccess
			reduce(263), // !=~, reduce: ArraySliceAccess
			reduce(263), // ==, reduce: ArraySliceAccess
			reduce(263), // !=, reduce: ArraySliceAccess
			reduce(263), // <=>, reduce: ArraySliceAccess
			reduce(263), // >=, reduce: ArraySliceAccess
			reduce(263), // <, reduce: ArraySliceAccess
			reduce(263), // <=, reduce: ArraySliceAccess
			reduce(263), // ^, reduce: ArraySliceAccess
			reduce(263), // &, reduce: ArraySliceAccess
			reduce(263), // <<, reduce: ArraySliceAccess
			reduce(263), // >>>, reduce: ArraySliceAccess
			reduce(263), // +, reduce: ArraySliceAccess
			reduce(263), // -, reduce: ArraySliceAccess
			reduce(263), // .+, reduce: ArraySliceAccess
			reduce(263), // .-, reduce: ArraySliceAccess
			reduce(263), // *, reduce: ArraySliceAccess
			reduce(263), // /, reduce: ArraySliceAccess
			reduce(263), // //, reduce: ArraySliceAccess
			reduce(263), // %, reduce: ArraySliceAccess
			reduce(263), // .*, reduce: ArraySliceAccess
			reduce(263), // ./, reduce: ArraySliceAccess
			reduce(263), // .//, reduce: ArraySliceAccess
			reduce(263), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(263), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(263), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(263), // [[, reduce: ArraySliceAccess
			reduce(263), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4494
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(262), // >, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |, reduce: ArrayOrMapPositionalValueAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(262), // ,, reduce: ArrayOrMapPositionalValueAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(262), // ?, reduce: ArrayOrMapPositionalValueAccess
			nil,         // :
			reduce(262), // ||, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // =~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ==, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .+, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .-, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .*, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ./, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .//, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ., reduce: ArrayOrMapPositionalValueAccess
			nil,         // !
			nil,         // ~
			reduce(262), // **, reduce: ArrayOrMapPositionalValueAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(262), // [, reduce: ArrayOrMapPositionalValueAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(262), // [[, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // [[[, reduce: ArrayOrMapPositionalValueAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4495
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4592), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4496
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4275), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4497
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(75), // $, reduce: EmitStatement
			nil,        // empty
			reduce(75), // ;, reduce: EmitStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4498
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(13), // >, reduce: StatementBlockInBraces
			reduce(13), // >>, reduce: StatementBlockInBraces
			reduce(13), // |, reduce: StatementBlockInBraces
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			reduce(13), // ), reduce: StatementBlockInBraces
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(13), // ?, reduce: StatementBlockInBraces
			nil,        // :
			reduce(13), // ||, reduce: StatementBlockInBraces
			reduce(13), // ^^, reduce: StatementBlockInBraces
			reduce(13), // &&, reduce: StatementBlockInBraces
			reduce(13), // ??, reduce: StatementBlockInBraces
			reduce(13), // ???, reduce: StatementBlockInBraces
			reduce(13), // =~, reduce: StatementBlockInBraces
			reduce(13), // !=~, reduce: StatementBlockInBraces
			reduce(13), // ==, reduce: StatementBlockInBraces
			reduce(13), // !=, reduce: StatementBlockInBraces
			reduce(13), // <=>, reduce: StatementBlockInBraces
			reduce(13), // >=, reduce: StatementBlockInBraces
			reduce(13), // <, reduce: StatementBlockInBraces
			reduce(13), // <=, reduce: StatementBlockInBraces
			reduce(13), // ^, reduce: StatementBlockInBraces
			reduce(13), // &, reduce: StatementBlockInBraces
			reduce(13), // <<, reduce: StatementBlockInBraces
			reduce(13), // >>>, reduce: StatementBlockInBraces
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			reduce(13), // *, reduce: StatementBlockInBraces
			reduce(13), // /, reduce: StatementBlockInBraces
			reduce(13), // //, reduce: StatementBlockInBraces
			reduce(13), // %, reduce: StatementBlockInBraces
			reduce(13), // .*, reduce: StatementBlockInBraces
			reduce(13), // ./, reduce: StatementBlockInBraces
			reduce(13), // .//, reduce: StatementBlockInBraces
			reduce(13), // ., reduce: StatementBlockInBraces
			nil,        // !
			nil,        // ~
			reduce(13), // **, reduce: StatementBlockInBraces
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(13), // [, reduce: StatementBlockInBraces
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(13), // [[, reduce: StatementBlockInBraces
			reduce(13), // [[[, reduce: StatementBlockInBraces
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4499
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(325), // >, reduce: UnnamedFunctionDefinition
			reduce(325), // >>, reduce: UnnamedFunctionDefinition
			reduce(325), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(325), // ), reduce: UnnamedFunctionDefinition
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(325), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(325), // ||, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^, reduce: UnnamedFunctionDefinition
			reduce(325), // &&, reduce: UnnamedFunctionDefinition
			reduce(325), // ??, reduce: UnnamedFunctionDefinition
			reduce(325), // ???, reduce: UnnamedFunctionDefinition
			reduce(325), // =~, reduce: UnnamedFunctionDefinition
			reduce(325), // !=~, reduce: UnnamedFunctionDefinition
			reduce(325), // ==, reduce: UnnamedFunctionDefinition
			reduce(325), // !=, reduce: UnnamedFunctionDefinition
			reduce(325), // <=>, reduce: UnnamedFunctionDefinition
			reduce(325), // >=, reduce: UnnamedFunctionDefinition
			reduce(325), // <, reduce: UnnamedFunctionDefinition
			reduce(325), // <=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^, reduce: UnnamedFunctionDefinition
			reduce(325), // &, reduce: UnnamedFunctionDefinition
			reduce(325), // <<, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>, reduce: UnnamedFunctionDefinition
			reduce(325), // +, reduce: UnnamedFunctionDefinition
			reduce(325), // -, reduce: UnnamedFunctionDefinition
			reduce(325), // .+, reduce: UnnamedFunctionDefinition
			reduce(325), // .-, reduce: UnnamedFunctionDefinition
			reduce(325), // *, reduce: UnnamedFunctionDefinition
			reduce(325), // /, reduce: UnnamedFunctionDefinition
			reduce(325), // //, reduce: UnnamedFunctionDefinition
			reduce(325), // %, reduce: UnnamedFunctionDefinition
			reduce(325), // .*, reduce: UnnamedFunctionDefinition
			reduce(325), // ./, reduce: UnnamedFunctionDefinition
			reduce(325), // .//, reduce: UnnamedFunctionDefinition
			reduce(325), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(325), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(325), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(325), // [[, reduce: UnnamedFunctionDefinition
			reduce(325), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4500
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			reduce(83), // $, reduce: EmitPStatement
			nil,        // empty
			reduce(83), // ;, reduce: EmitPStatement
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4501
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(13), // >, reduce: StatementBlockInBraces
			reduce(13), // >>, reduce: StatementBlockInBraces
			reduce(13), // |, reduce: StatementBlockInBraces
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			reduce(13), // ], reduce: StatementBlockInBraces
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(13), // ?, reduce: StatementBlockInBraces
			nil,        // :
			reduce(13), // ||, reduce: StatementBlockInBraces
			reduce(13), // ^^, reduce: StatementBlockInBraces
			reduce(13), // &&, reduce: StatementBlockInBraces
			reduce(13), // ??, reduce: StatementBlockInBraces
			reduce(13), // ???, reduce: StatementBlockInBraces
			reduce(13), // =~, reduce: StatementBlockInBraces
			reduce(13), // !=~, reduce: StatementBlockInBraces
			reduce(13), // ==, reduce: StatementBlockInBraces
			reduce(13), // !=, reduce: StatementBlockInBraces
			reduce(13), // <=>, reduce: StatementBlockInBraces
			reduce(13), // >=, reduce: StatementBlockInBraces
			reduce(13), // <, reduce: StatementBlockInBraces
			reduce(13), // <=, reduce: StatementBlockInBraces
			reduce(13), // ^, reduce: StatementBlockInBraces
			reduce(13), // &, reduce: StatementBlockInBraces
			reduce(13), // <<, reduce: StatementBlockInBraces
			reduce(13), // >>>, reduce: StatementBlockInBraces
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			reduce(13), // *, reduce: StatementBlockInBraces
			reduce(13), // /, reduce: StatementBlockInBraces
			reduce(13), // //, reduce: StatementBlockInBraces
			reduce(13), // %, reduce: StatementBlockInBraces
			reduce(13), // .*, reduce: StatementBlockInBraces
			reduce(13), // ./, reduce: StatementBlockInBraces
			reduce(13), // .//, reduce: StatementBlockInBraces
			reduce(13), // ., reduce: StatementBlockInBraces
			nil,        // !
			nil,        // ~
			reduce(13), // **, reduce: StatementBlockInBraces
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(13), // [, reduce: StatementBlockInBraces
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(13), // [[, reduce: StatementBlockInBraces
			reduce(13), // [[[, reduce: StatementBlockInBraces
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4502
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(325), // >, reduce: UnnamedFunctionDefinition
			reduce(325), // >>, reduce: UnnamedFunctionDefinition
			reduce(325), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(325), // ], reduce: UnnamedFunctionDefinition
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(325), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(325), // ||, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^, reduce: UnnamedFunctionDefinition
			reduce(325), // &&, reduce: UnnamedFunctionDefinition
			reduce(325), // ??, reduce: UnnamedFunctionDefinition
			reduce(325), // ???, reduce: UnnamedFunctionDefinition
			reduce(325), // =~, reduce: UnnamedFunctionDefinition
			reduce(325), // !=~, reduce: UnnamedFunctionDefinition
			reduce(325), // ==, reduce: UnnamedFunctionDefinition
			reduce(325), // !=, reduce: UnnamedFunctionDefinition
			reduce(325), // <=>, reduce: UnnamedFunctionDefinition
			reduce(325), // >=, reduce: UnnamedFunctionDefinition
			reduce(325), // <, reduce: UnnamedFunctionDefinition
			reduce(325), // <=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^, reduce: UnnamedFunctionDefinition
			reduce(325), // &, reduce: UnnamedFunctionDefinition
			reduce(325), // <<, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>, reduce: UnnamedFunctionDefinition
			reduce(325), // +, reduce: UnnamedFunctionDefinition
			reduce(325), // -, reduce: UnnamedFunctionDefinition
			reduce(325), // .+, reduce: UnnamedFunctionDefinition
			reduce(325), // .-, reduce: UnnamedFunctionDefinition
			reduce(325), // *, reduce: UnnamedFunctionDefinition
			reduce(325), // /, reduce: UnnamedFunctionDefinition
			reduce(325), // //, reduce: UnnamedFunctionDefinition
			reduce(325), // %, reduce: UnnamedFunctionDefinition
			reduce(325), // .*, reduce: UnnamedFunctionDefinition
			reduce(325), // ./, reduce: UnnamedFunctionDefinition
			reduce(325), // .//, reduce: UnnamedFunctionDefinition
			reduce(325), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(325), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(325), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(325), // [[, reduce: UnnamedFunctionDefinition
			reduce(325), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4503
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(263), // >, reduce: ArraySliceAccess
			reduce(263), // >>, reduce: ArraySliceAccess
			reduce(263), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(263), // ], reduce: ArraySliceAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(263), // ?, reduce: ArraySliceAccess
			reduce(263), // :, reduce: ArraySliceAccess
			reduce(263), // ||, reduce: ArraySliceAccess
			reduce(263), // ^^, reduce: ArraySliceAccess
			reduce(263), // &&, reduce: ArraySliceAccess
			reduce(263), // ??, reduce: ArraySliceAccess
			reduce(263), // ???, reduce: ArraySliceAccess
			reduce(263), // =~, reduce: ArraySliceAccess
			reduce(263), // !=~, reduce: ArraySliceAccess
			reduce(263), // ==, reduce: ArraySliceAccess
			reduce(263), // !=, reduce: ArraySliceAccess
			reduce(263), // <=>, reduce: ArraySliceAccess
			reduce(263), // >=, reduce: ArraySliceAccess
			reduce(263), // <, reduce: ArraySliceAccess
			reduce(263), // <=, reduce: ArraySliceAccess
			reduce(263), // ^, reduce: ArraySliceAccess
			reduce(263), // &, reduce: ArraySliceAccess
			reduce(263), // <<, reduce: ArraySliceAccess
			reduce(263), // >>>, reduce: ArraySliceAccess
			reduce(263), // +, reduce: ArraySliceAccess
			reduce(263), // -, reduce: ArraySliceAccess
			reduce(263), // .+, reduce: ArraySliceAccess
			reduce(263), // .-, reduce: ArraySliceAccess
			reduce(263), // *, reduce: ArraySliceAccess
			reduce(263), // /, reduce: ArraySliceAccess
			reduce(263), // //, reduce: ArraySliceAccess
			reduce(263), // %, reduce: ArraySliceAccess
			reduce(263), // .*, reduce: ArraySliceAccess
			reduce(263), // ./, reduce: ArraySliceAccess
			reduce(263), // .//, reduce: ArraySliceAccess
			reduce(263), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(263), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(263), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(263), // [[, reduce: ArraySliceAccess
			reduce(263), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4504
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(262), // >, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |, reduce: ArrayOrMapPositionalValueAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(262), // ], reduce: ArrayOrMapPositionalValueAccess
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(262), // ?, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // :, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ||, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // =~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ==, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .+, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .-, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .*, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ./, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .//, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ., reduce: ArrayOrMapPositionalValueAccess
			nil,         // !
			nil,         // ~
			reduce(262), // **, reduce: ArrayOrMapPositionalValueAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(262), // [, reduce: ArrayOrMapPositionalValueAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(262), // [[, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // [[[, reduce: ArrayOrMapPositionalValueAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4505
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4594), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4506
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4295), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4507
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(13), // >, reduce: StatementBlockInBraces
			reduce(13), // >>, reduce: StatementBlockInBraces
			reduce(13), // |, reduce: StatementBlockInBraces
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(13), // ,, reduce: StatementBlockInBraces
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			reduce(13), // ], reduce: StatementBlockInBraces
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(13), // ?, reduce: StatementBlockInBraces
			nil,        // :
			reduce(13), // ||, reduce: StatementBlockInBraces
			reduce(13), // ^^, reduce: StatementBlockInBraces
			reduce(13), // &&, reduce: StatementBlockInBraces
			reduce(13), // ??, reduce: StatementBlockInBraces
			reduce(13), // ???, reduce: StatementBlockInBraces
			reduce(13), // =~, reduce: StatementBlockInBraces
			reduce(13), // !=~, reduce: StatementBlockInBraces
			reduce(13), // ==, reduce: StatementBlockInBraces
			reduce(13), // !=, reduce: StatementBlockInBraces
			reduce(13), // <=>, reduce: StatementBlockInBraces
			reduce(13), // >=, reduce: StatementBlockInBraces
			reduce(13), // <, reduce: StatementBlockInBraces
			reduce(13), // <=, reduce: StatementBlockInBraces
			reduce(13), // ^, reduce: StatementBlockInBraces
			reduce(13), // &, reduce: StatementBlockInBraces
			reduce(13), // <<, reduce: StatementBlockInBraces
			reduce(13), // >>>, reduce: StatementBlockInBraces
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			reduce(13), // *, reduce: StatementBlockInBraces
			reduce(13), // /, reduce: StatementBlockInBraces
			reduce(13), // //, reduce: StatementBlockInBraces
			reduce(13), // %, reduce: StatementBlockInBraces
			reduce(13), // .*, reduce: StatementBlockInBraces
			reduce(13), // ./, reduce: StatementBlockInBraces
			reduce(13), // .//, reduce: StatementBlockInBraces
			reduce(13), // ., reduce: StatementBlockInBraces
			nil,        // !
			nil,        // ~
			reduce(13), // **, reduce: StatementBlockInBraces
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(13), // [, reduce: StatementBlockInBraces
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(13), // [[, reduce: StatementBlockInBraces
			reduce(13), // [[[, reduce: StatementBlockInBraces
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4508
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(325), // >, reduce: UnnamedFunctionDefinition
			reduce(325), // >>, reduce: UnnamedFunctionDefinition
			reduce(325), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(325), // ,, reduce: UnnamedFunctionDefinition
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(325), // ], reduce: UnnamedFunctionDefinition
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(325), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(325), // ||, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^, reduce: UnnamedFunctionDefinition
			reduce(325), // &&, reduce: UnnamedFunctionDefinition
			reduce(325), // ??, reduce: UnnamedFunctionDefinition
			reduce(325), // ???, reduce: UnnamedFunctionDefinition
			reduce(325), // =~, reduce: UnnamedFunctionDefinition
			reduce(325), // !=~, reduce: UnnamedFunctionDefinition
			reduce(325), // ==, reduce: UnnamedFunctionDefinition
			reduce(325), // !=, reduce: UnnamedFunctionDefinition
			reduce(325), // <=>, reduce: UnnamedFunctionDefinition
			reduce(325), // >=, reduce: UnnamedFunctionDefinition
			reduce(325), // <, reduce: UnnamedFunctionDefinition
			reduce(325), // <=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^, reduce: UnnamedFunctionDefinition
			reduce(325), // &, reduce: UnnamedFunctionDefinition
			reduce(325), // <<, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>, reduce: UnnamedFunctionDefinition
			reduce(325), // +, reduce: UnnamedFunctionDefinition
			reduce(325), // -, reduce: UnnamedFunctionDefinition
			reduce(325), // .+, reduce: UnnamedFunctionDefinition
			reduce(325), // .-, reduce: UnnamedFunctionDefinition
			reduce(325), // *, reduce: UnnamedFunctionDefinition
			reduce(325), // /, reduce: UnnamedFunctionDefinition
			reduce(325), // //, reduce: UnnamedFunctionDefinition
			reduce(325), // %, reduce: UnnamedFunctionDefinition
			reduce(325), // .*, reduce: UnnamedFunctionDefinition
			reduce(325), // ./, reduce: UnnamedFunctionDefinition
			reduce(325), // .//, reduce: UnnamedFunctionDefinition
			reduce(325), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(325), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(325), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(325), // [[, reduce: UnnamedFunctionDefinition
			reduce(325), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4509
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(263), // >, reduce: ArraySliceAccess
			reduce(263), // >>, reduce: ArraySliceAccess
			reduce(263), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(263), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(263), // ), reduce: ArraySliceAccess
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(263), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(263), // ||, reduce: ArraySliceAccess
			reduce(263), // ^^, reduce: ArraySliceAccess
			reduce(263), // &&, reduce: ArraySliceAccess
			reduce(263), // ??, reduce: ArraySliceAccess
			reduce(263), // ???, reduce: ArraySliceAccess
			reduce(263), // =~, reduce: ArraySliceAccess
			reduce(263), // !=~, reduce: ArraySliceAccess
			reduce(263), // ==, reduce: ArraySliceAccess
			reduce(263), // !=, reduce: ArraySliceAccess
			reduce(263), // <=>, reduce: ArraySliceAccess
			reduce(263), // >=, reduce: ArraySliceAccess
			reduce(263), // <, reduce: ArraySliceAccess
			reduce(263), // <=, reduce: ArraySliceAccess
			reduce(263), // ^, reduce: ArraySliceAccess
			reduce(263), // &, reduce: ArraySliceAccess
			reduce(263), // <<, reduce: ArraySliceAccess
			reduce(263), // >>>, reduce: ArraySliceAccess
			reduce(263), // +, reduce: ArraySliceAccess
			reduce(263), // -, reduce: ArraySliceAccess
			reduce(263), // .+, reduce: ArraySliceAccess
			reduce(263), // .-, reduce: ArraySliceAccess
			reduce(263), // *, reduce: ArraySliceAccess
			reduce(263), // /, reduce: ArraySliceAccess
			reduce(263), // //, reduce: ArraySliceAccess
			reduce(263), // %, reduce: ArraySliceAccess
			reduce(263), // .*, reduce: ArraySliceAccess
			reduce(263), // ./, reduce: ArraySliceAccess
			reduce(263), // .//, reduce: ArraySliceAccess
			reduce(263), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(263), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(263), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(263), // [[, reduce: ArraySliceAccess
			reduce(263), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4510
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(262), // >, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |, reduce: ArrayOrMapPositionalValueAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(262), // ,, reduce: ArrayOrMapPositionalValueAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(262), // ), reduce: ArrayOrMapPositionalValueAccess
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(262), // ?, reduce: ArrayOrMapPositionalValueAccess
			nil,         // :
			reduce(262), // ||, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // =~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ==, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .+, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .-, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .*, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ./, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .//, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ., reduce: ArrayOrMapPositionalValueAccess
			nil,         // !
			nil,         // ~
			reduce(262), // **, reduce: ArrayOrMapPositionalValueAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(262), // [, reduce: ArrayOrMapPositionalValueAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(262), // [[, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // [[[, reduce: ArrayOrMapPositionalValueAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4511
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4596), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4512
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4309), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4513
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(107), // ;, reduce: PositionalFieldName
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(107), // >, reduce: PositionalFieldName
			reduce(107), // >>, reduce: PositionalFieldName
			reduce(107), // |, reduce: PositionalFieldName
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(107), // ,, reduce: PositionalFieldName
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(107), // ?, reduce: PositionalFieldName
			nil,         // :
			reduce(107), // ||, reduce: PositionalFieldName
			reduce(107), // ^^, reduce: PositionalFieldName
			reduce(107), // &&, reduce: PositionalFieldName
			reduce(107), // ??, reduce: PositionalFieldName
			reduce(107), // ???, reduce: PositionalFieldName
			reduce(107), // =~, reduce: PositionalFieldName
			reduce(107), // !=~, reduce: PositionalFieldName
			reduce(107), // ==, reduce: PositionalFieldName
			reduce(107), // !=, reduce: PositionalFieldName
			reduce(107), // <=>, reduce: PositionalFieldName
			reduce(107), // >=, reduce: PositionalFieldName
			reduce(107), // <, reduce: PositionalFieldName
			reduce(107), // <=, reduce: PositionalFieldName
			reduce(107), // ^, reduce: PositionalFieldName
			reduce(107), // &, reduce: PositionalFieldName
			reduce(107), // <<, reduce: PositionalFieldName
			reduce(107), // >>>, reduce: PositionalFieldName
			reduce(107), // +, reduce: PositionalFieldName
			reduce(107), // -, reduce: PositionalFieldName
			reduce(107), // .+, reduce: PositionalFieldName
			reduce(107), // .-, reduce: PositionalFieldName
			reduce(107), // *, reduce: PositionalFieldName
			reduce(107), // /, reduce: PositionalFieldName
			reduce(107), // //, reduce: PositionalFieldName
			reduce(107), // %, reduce: PositionalFieldName
			reduce(107), // .*, reduce: PositionalFieldName
			reduce(107), // ./, reduce: PositionalFieldName
			reduce(107), // .//, reduce: PositionalFieldName
			reduce(107), // ., reduce: PositionalFieldName
			nil,         // !
			nil,         // ~
			reduce(107), // **, reduce: PositionalFieldName
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(107), // [, reduce: PositionalFieldName
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(107), // [[, reduce: PositionalFieldName
			reduce(107), // [[[, reduce: PositionalFieldName
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4514
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4598), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4515
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3450), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3454), // (
			nil,         // )
			nil,         // emitp
			shift(3466), // field_name
			shift(3467), // $[
			nil,         // ]
			shift(3468), // braced_field_name
			shift(3469), // $[[
			shift(3470), // $[[[
			shift(3471), // full_srec
			shift(3473), // oosvar_name
			shift(3474), // @[
			shift(3475), // braced_oosvar_name
			shift(3476), // full_oosvar
			shift(3477), // all
			shift(3478), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3493), // +
			shift(3495), // -
			shift(3496), // .+
			shift(3497), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3500), // !
			shift(3501), // ~
			nil,         // **
			shift(3506), // string_literal
			shift(3507), // regex_case_insensitive
			shift(3508), // int_literal
			shift(3509), // float_literal
			shift(3510), // boolean_literal
			shift(3511), // null_literal
			shift(3512), // inf_literal
			shift(3513), // nan_literal
			shift(3514), // const_M_PI
			shift(3515), // const_M_E
			shift(3516), // panic
			shift(3518), // [
			shift(3520), // ctx_IPS
			shift(3521), // ctx_IFS
			shift(3522), // ctx_IRS
			shift(3523), // ctx_OPS
			shift(3524), // ctx_OFS
			shift(3525), // ctx_ORS
			shift(3526), // ctx_FLATSEP
			shift(3527), // ctx_NF
			shift(3528), // ctx_NR
			shift(3529), // ctx_FNR
			shift(3530), // ctx_FILENAME
			shift(3531), // ctx_FILENUM
			shift(3533), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3540), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4516
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(206), // ;, reduce: PowTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(206), // >, reduce: PowTerm
			reduce(206), // >>, reduce: PowTerm
			reduce(206), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(206), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(206), // ?, reduce: PowTerm
			nil,         // :
			reduce(206), // ||, reduce: PowTerm
			reduce(206), // ^^, reduce: PowTerm
			reduce(206), // &&, reduce: PowTerm
			reduce(206), // ??, reduce: PowTerm
			reduce(206), // ???, reduce: PowTerm
			reduce(206), // =~, reduce: PowTerm
			reduce(206), // !=~, reduce: PowTerm
			reduce(206), // ==, reduce: PowTerm
			reduce(206), // !=, reduce: PowTerm
			reduce(206), // <=>, reduce: PowTerm
			reduce(206), // >=, reduce: PowTerm
			reduce(206), // <, reduce: PowTerm
			reduce(206), // <=, reduce: PowTerm
			reduce(206), // ^, reduce: PowTerm
			reduce(206), // &, reduce: PowTerm
			reduce(206), // <<, reduce: PowTerm
			reduce(206), // >>>, reduce: PowTerm
			reduce(206), // +, reduce: PowTerm
			reduce(206), // -, reduce: PowTerm
			reduce(206), // .+, reduce: PowTerm
			reduce(206), // .-, reduce: PowTerm
			reduce(206), // *, reduce: PowTerm
			reduce(206), // /, reduce: PowTerm
			reduce(206), // //, reduce: PowTerm
			reduce(206), // %, reduce: PowTerm
			reduce(206), // .*, reduce: PowTerm
			reduce(206), // ./, reduce: PowTerm
			reduce(206), // .//, reduce: PowTerm
			reduce(206), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4517
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(205), // ;, reduce: PowTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(205), // >, reduce: PowTerm
			reduce(205), // >>, reduce: PowTerm
			reduce(205), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(205), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(205), // ?, reduce: PowTerm
			nil,         // :
			reduce(205), // ||, reduce: PowTerm
			reduce(205), // ^^, reduce: PowTerm
			reduce(205), // &&, reduce: PowTerm
			reduce(205), // ??, reduce: PowTerm
			reduce(205), // ???, reduce: PowTerm
			reduce(205), // =~, reduce: PowTerm
			reduce(205), // !=~, reduce: PowTerm
			reduce(205), // ==, reduce: PowTerm
			reduce(205), // !=, reduce: PowTerm
			reduce(205), // <=>, reduce: PowTerm
			reduce(205), // >=, reduce: PowTerm
			reduce(205), // <, reduce: PowTerm
			reduce(205), // <=, reduce: PowTerm
			reduce(205), // ^, reduce: PowTerm
			reduce(205), // &, reduce: PowTerm
			reduce(205), // <<, reduce: PowTerm
			reduce(205), // >>>, reduce: PowTerm
			reduce(205), // +, reduce: PowTerm
			reduce(205), // -, reduce: PowTerm
			reduce(205), // .+, reduce: PowTerm
			reduce(205), // .-, reduce: PowTerm
			reduce(205), // *, reduce: PowTerm
			reduce(205), // /, reduce: PowTerm
			reduce(205), // //, reduce: PowTerm
			reduce(205), // %, reduce: PowTerm
			reduce(205), // .*, reduce: PowTerm
			reduce(205), // ./, reduce: PowTerm
			reduce(205), // .//, reduce: PowTerm
			reduce(205), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4518
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(260), // ;, reduce: ArrayOrMapIndexAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(260), // >, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>, reduce: ArrayOrMapIndexAccess
			reduce(260), // |, reduce: ArrayOrMapIndexAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(260), // ,, reduce: ArrayOrMapIndexAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(260), // ?, reduce: ArrayOrMapIndexAccess
			nil,         // :
			reduce(260), // ||, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???, reduce: ArrayOrMapIndexAccess
			reduce(260), // =~, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=~, reduce: ArrayOrMapIndexAccess
			reduce(260), // ==, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=>, reduce: ArrayOrMapIndexAccess
			reduce(260), // >=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>, reduce: ArrayOrMapIndexAccess
			reduce(260), // +, reduce: ArrayOrMapIndexAccess
			reduce(260), // -, reduce: ArrayOrMapIndexAccess
			reduce(260), // .+, reduce: ArrayOrMapIndexAccess
			reduce(260), // .-, reduce: ArrayOrMapIndexAccess
			reduce(260), // *, reduce: ArrayOrMapIndexAccess
			reduce(260), // /, reduce: ArrayOrMapIndexAccess
			reduce(260), // //, reduce: ArrayOrMapIndexAccess
			reduce(260), // %, reduce: ArrayOrMapIndexAccess
			reduce(260), // .*, reduce: ArrayOrMapIndexAccess
			reduce(260), // ./, reduce: ArrayOrMapIndexAccess
			reduce(260), // .//, reduce: ArrayOrMapIndexAccess
			reduce(260), // ., reduce: ArrayOrMapIndexAccess
			nil,         // !
			nil,         // ~
			reduce(260), // **, reduce: ArrayOrMapIndexAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(260), // [, reduce: ArrayOrMapIndexAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(260), // [[, reduce: ArrayOrMapIndexAccess
			reduce(260), // [[[, reduce: ArrayOrMapIndexAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4519
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4601), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4520
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4602), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4521
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(266), // ;, reduce: ArraySliceAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(266), // >, reduce: ArraySliceAccess
			reduce(266), // >>, reduce: ArraySliceAccess
			reduce(266), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(266), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(266), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(266), // ||, reduce: ArraySliceAccess
			reduce(266), // ^^, reduce: ArraySliceAccess
			reduce(266), // &&, reduce: ArraySliceAccess
			reduce(266), // ??, reduce: ArraySliceAccess
			reduce(266), // ???, reduce: ArraySliceAccess
			reduce(266), // =~, reduce: ArraySliceAccess
			reduce(266), // !=~, reduce: ArraySliceAccess
			reduce(266), // ==, reduce: ArraySliceAccess
			reduce(266), // !=, reduce: ArraySliceAccess
			reduce(266), // <=>, reduce: ArraySliceAccess
			reduce(266), // >=, reduce: ArraySliceAccess
			reduce(266), // <, reduce: ArraySliceAccess
			reduce(266), // <=, reduce: ArraySliceAccess
			reduce(266), // ^, reduce: ArraySliceAccess
			reduce(266), // &, reduce: ArraySliceAccess
			reduce(266), // <<, reduce: ArraySliceAccess
			reduce(266), // >>>, reduce: ArraySliceAccess
			reduce(266), // +, reduce: ArraySliceAccess
			reduce(266), // -, reduce: ArraySliceAccess
			reduce(266), // .+, reduce: ArraySliceAccess
			reduce(266), // .-, reduce: ArraySliceAccess
			reduce(266), // *, reduce: ArraySliceAccess
			reduce(266), // /, reduce: ArraySliceAccess
			reduce(266), // //, reduce: ArraySliceAccess
			reduce(266), // %, reduce: ArraySliceAccess
			reduce(266), // .*, reduce: ArraySliceAccess
			reduce(266), // ./, reduce: ArraySliceAccess
			reduce(266), // .//, reduce: ArraySliceAccess
			reduce(266), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(266), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(266), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(266), // [[, reduce: ArraySliceAccess
			reduce(266), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4522
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4603), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4523
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4604), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4524
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(254), // ;, reduce: ENV
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(254), // >, reduce: ENV
			reduce(254), // >>, reduce: ENV
			reduce(254), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(254), // ,, reduce: ENV
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(254), // ?, reduce: ENV
			nil,         // :
			reduce(254), // ||, reduce: ENV
			reduce(254), // ^^, reduce: ENV
			reduce(254), // &&, reduce: ENV
			reduce(254), // ??, reduce: ENV
			reduce(254), // ???, reduce: ENV
			reduce(254), // =~, reduce: ENV
			reduce(254), // !=~, reduce: ENV
			reduce(254), // ==, reduce: ENV
			reduce(254), // !=, reduce: ENV
			reduce(254), // <=>, reduce: ENV
			reduce(254), // >=, reduce: ENV
			reduce(254), // <, reduce: ENV
			reduce(254), // <=, reduce: ENV
			reduce(254), // ^, reduce: ENV
			reduce(254), // &, reduce: ENV
			reduce(254), // <<, reduce: ENV
			reduce(254), // >>>, reduce: ENV
			reduce(254), // +, reduce: ENV
			reduce(254), // -, reduce: ENV
			reduce(254), // .+, reduce: ENV
			reduce(254), // .-, reduce: ENV
			reduce(254), // *, reduce: ENV
			reduce(254), // /, reduce: ENV
			reduce(254), // //, reduce: ENV
			reduce(254), // %, reduce: ENV
			reduce(254), // .*, reduce: ENV
			reduce(254), // ./, reduce: ENV
			reduce(254), // .//, reduce: ENV
			reduce(254), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(254), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(254), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(254), // [[, reduce: ENV
			reduce(254), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4525
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(269), // ;, reduce: FunctionCallsite
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(269), // >, reduce: FunctionCallsite
			reduce(269), // >>, reduce: FunctionCallsite
			reduce(269), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(269), // ,, reduce: FunctionCallsite
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(269), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(269), // ||, reduce: FunctionCallsite
			reduce(269), // ^^, reduce: FunctionCallsite
			reduce(269), // &&, reduce: FunctionCallsite
			reduce(269), // ??, reduce: FunctionCallsite
			reduce(269), // ???, reduce: FunctionCallsite
			reduce(269), // =~, reduce: FunctionCallsite
			reduce(269), // !=~, reduce: FunctionCallsite
			reduce(269), // ==, reduce: FunctionCallsite
			reduce(269), // !=, reduce: FunctionCallsite
			reduce(269), // <=>, reduce: FunctionCallsite
			reduce(269), // >=, reduce: FunctionCallsite
			reduce(269), // <, reduce: FunctionCallsite
			reduce(269), // <=, reduce: FunctionCallsite
			reduce(269), // ^, reduce: FunctionCallsite
			reduce(269), // &, reduce: FunctionCallsite
			reduce(269), // <<, reduce: FunctionCallsite
			reduce(269), // >>>, reduce: FunctionCallsite
			reduce(269), // +, reduce: FunctionCallsite
			reduce(269), // -, reduce: FunctionCallsite
			reduce(269), // .+, reduce: FunctionCallsite
			reduce(269), // .-, reduce: FunctionCallsite
			reduce(269), // *, reduce: FunctionCallsite
			reduce(269), // /, reduce: FunctionCallsite
			reduce(269), // //, reduce: FunctionCallsite
			reduce(269), // %, reduce: FunctionCallsite
			reduce(269), // .*, reduce: FunctionCallsite
			reduce(269), // ./, reduce: FunctionCallsite
			reduce(269), // .//, reduce: FunctionCallsite
			reduce(269), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(269), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(269), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(269), // [[, reduce: FunctionCallsite
			reduce(269), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4526
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4606), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4607), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4527
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(289), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4528
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4529
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(263), // =, reduce: ArraySliceAccess
			nil,         // unset
			nil,         // filter
			reduce(263), // >, reduce: ArraySliceAccess
			reduce(263), // >>, reduce: ArraySliceAccess
			reduce(263), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(263), // ||=, reduce: ArraySliceAccess
			reduce(263), // ^^=, reduce: ArraySliceAccess
			reduce(263), // &&=, reduce: ArraySliceAccess
			reduce(263), // ??=, reduce: ArraySliceAccess
			reduce(263), // ???=, reduce: ArraySliceAccess
			reduce(263), // |=, reduce: ArraySliceAccess
			reduce(263), // &=, reduce: ArraySliceAccess
			reduce(263), // ^=, reduce: ArraySliceAccess
			reduce(263), // <<=, reduce: ArraySliceAccess
			reduce(263), // >>=, reduce: ArraySliceAccess
			reduce(263), // >>>=, reduce: ArraySliceAccess
			reduce(263), // +=, reduce: ArraySliceAccess
			reduce(263), // .=, reduce: ArraySliceAccess
			reduce(263), // -=, reduce: ArraySliceAccess
			reduce(263), // *=, reduce: ArraySliceAccess
			reduce(263), // /=, reduce: ArraySliceAccess
			reduce(263), // //=, reduce: ArraySliceAccess
			reduce(263), // %=, reduce: ArraySliceAccess
			reduce(263), // **=, reduce: ArraySliceAccess
			reduce(263), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(263), // ||, reduce: ArraySliceAccess
			reduce(263), // ^^, reduce: ArraySliceAccess
			reduce(263), // &&, reduce: ArraySliceAccess
			reduce(263), // ??, reduce: ArraySliceAccess
			reduce(263), // ???, reduce: ArraySliceAccess
			reduce(263), // =~, reduce: ArraySliceAccess
			reduce(263), // !=~, reduce: ArraySliceAccess
			reduce(263), // ==, reduce: ArraySliceAccess
			reduce(263), // !=, reduce: ArraySliceAccess
			reduce(263), // <=>, reduce: ArraySliceAccess
			reduce(263), // >=, reduce: ArraySliceAccess
			reduce(263), // <, reduce: ArraySliceAccess
			reduce(263), // <=, reduce: ArraySliceAccess
			reduce(263), // ^, reduce: ArraySliceAccess
			reduce(263), // &, reduce: ArraySliceAccess
			reduce(263), // <<, reduce: ArraySliceAccess
			reduce(263), // >>>, reduce: ArraySliceAccess
			reduce(263), // +, reduce: ArraySliceAccess
			reduce(263), // -, reduce: ArraySliceAccess
			reduce(263), // .+, reduce: ArraySliceAccess
			reduce(263), // .-, reduce: ArraySliceAccess
			reduce(263), // *, reduce: ArraySliceAccess
			reduce(263), // /, reduce: ArraySliceAccess
			reduce(263), // //, reduce: ArraySliceAccess
			reduce(263), // %, reduce: ArraySliceAccess
			reduce(263), // .*, reduce: ArraySliceAccess
			reduce(263), // ./, reduce: ArraySliceAccess
			reduce(263), // .//, reduce: ArraySliceAccess
			reduce(263), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(263), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(263), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(263), // [[, reduce: ArraySliceAccess
			reduce(263), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4530
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(262), // =, reduce: ArrayOrMapPositionalValueAccess
			nil,         // unset
			nil,         // filter
			reduce(262), // >, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |, reduce: ArrayOrMapPositionalValueAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(262), // ||=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // **=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ?, reduce: ArrayOrMapPositionalValueAccess
			nil,         // :
			reduce(262), // ||, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // =~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ==, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .+, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .-, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .*, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ./, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .//, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ., reduce: ArrayOrMapPositionalValueAccess
			nil,         // !
			nil,         // ~
			reduce(262), // **, reduce: ArrayOrMapPositionalValueAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(262), // [, reduce: ArrayOrMapPositionalValueAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(262), // [[, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // [[[, reduce: ArrayOrMapPositionalValueAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4531
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(107), // ;, reduce: PositionalFieldName
			nil,         // {
			nil,         // }
			reduce(107), // =, reduce: PositionalFieldName
			nil,         // unset
			nil,         // filter
			reduce(107), // >, reduce: PositionalFieldName
			reduce(107), // >>, reduce: PositionalFieldName
			reduce(107), // |, reduce: PositionalFieldName
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(107), // ,, reduce: PositionalFieldName
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(107), // ||=, reduce: PositionalFieldName
			reduce(107), // ^^=, reduce: PositionalFieldName
			reduce(107), // &&=, reduce: PositionalFieldName
			reduce(107), // ??=, reduce: PositionalFieldName
			reduce(107), // ???=, reduce: PositionalFieldName
			reduce(107), // |=, reduce: PositionalFieldName
			reduce(107), // &=, reduce: PositionalFieldName
			reduce(107), // ^=, reduce: PositionalFieldName
			reduce(107), // <<=, reduce: PositionalFieldName
			reduce(107), // >>=, reduce: PositionalFieldName
			reduce(107), // >>>=, reduce: PositionalFieldName
			reduce(107), // +=, reduce: PositionalFieldName
			reduce(107), // .=, reduce: PositionalFieldName
			reduce(107), // -=, reduce: PositionalFieldName
			reduce(107), // *=, reduce: PositionalFieldName
			reduce(107), // /=, reduce: PositionalFieldName
			reduce(107), // //=, reduce: PositionalFieldName
			reduce(107), // %=, reduce: PositionalFieldName
			reduce(107), // **=, reduce: PositionalFieldName
			reduce(107), // ?, reduce: PositionalFieldName
			nil,         // :
			reduce(107), // ||, reduce: PositionalFieldName
			reduce(107), // ^^, reduce: PositionalFieldName
			reduce(107), // &&, reduce: PositionalFieldName
			reduce(107), // ??, reduce: PositionalFieldName
			reduce(107), // ???, reduce: PositionalFieldName
			reduce(107), // =~, reduce: PositionalFieldName
			reduce(107), // !=~, reduce: PositionalFieldName
			reduce(107), // ==, reduce: PositionalFieldName
			reduce(107), // !=, reduce: PositionalFieldName
			reduce(107), // <=>, reduce: PositionalFieldName
			reduce(107), // >=, reduce: PositionalFieldName
			reduce(107), // <, reduce: PositionalFieldName
			reduce(107), // <=, reduce: PositionalFieldName
			reduce(107), // ^, reduce: PositionalFieldName
			reduce(107), // &, reduce: PositionalFieldName
			reduce(107), // <<, reduce: PositionalFieldName
			reduce(107), // >>>, reduce: PositionalFieldName
			reduce(107), // +, reduce: PositionalFieldName
			reduce(107), // -, reduce: PositionalFieldName
			reduce(107), // .+, reduce: PositionalFieldName
			reduce(107), // .-, reduce: PositionalFieldName
			reduce(107), // *, reduce: PositionalFieldName
			reduce(107), // /, reduce: PositionalFieldName
			reduce(107), // //, reduce: PositionalFieldName
			reduce(107), // %, reduce: PositionalFieldName
			reduce(107), // .*, reduce: PositionalFieldName
			reduce(107), // ./, reduce: PositionalFieldName
			reduce(107), // .//, reduce: PositionalFieldName
			reduce(107), // ., reduce: PositionalFieldName
			nil,         // !
			nil,         // ~
			reduce(107), // **, reduce: PositionalFieldName
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(107), // [, reduce: PositionalFieldName
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(107), // [[, reduce: PositionalFieldName
			reduce(107), // [[[, reduce: PositionalFieldName
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4532
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4610), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4533
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(3616), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(3622), // (
			nil,         // )
			nil,         // emitp
			shift(3634), // field_name
			shift(3635), // $[
			nil,         // ]
			shift(3636), // braced_field_name
			shift(3637), // $[[
			shift(3638), // $[[[
			shift(3639), // full_srec
			shift(3641), // oosvar_name
			shift(3642), // @[
			shift(3643), // braced_oosvar_name
			shift(3644), // full_oosvar
			shift(3645), // all
			shift(3646), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(3661), // +
			shift(3663), // -
			shift(3664), // .+
			shift(3665), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(3668), // !
			shift(3669), // ~
			nil,         // **
			shift(3674), // string_literal
			shift(3675), // regex_case_insensitive
			shift(3676), // int_literal
			shift(3677), // float_literal
			shift(3678), // boolean_literal
			shift(3679), // null_literal
			shift(3680), // inf_literal
			shift(3681), // nan_literal
			shift(3682), // const_M_PI
			shift(3683), // const_M_E
			shift(3684), // panic
			shift(3686), // [
			shift(3688), // ctx_IPS
			shift(3689), // ctx_IFS
			shift(3690), // ctx_IRS
			shift(3691), // ctx_OPS
			shift(3692), // ctx_OFS
			shift(3693), // ctx_ORS
			shift(3694), // ctx_FLATSEP
			shift(3695), // ctx_NF
			shift(3696), // ctx_NR
			shift(3697), // ctx_FNR
			shift(3698), // ctx_FILENAME
			shift(3699), // ctx_FILENUM
			shift(3701), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(3710), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4534
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(206), // ;, reduce: PowTerm
			nil,         // {
			nil,         // }
			reduce(206), // =, reduce: PowTerm
			nil,         // unset
			nil,         // filter
			reduce(206), // >, reduce: PowTerm
			reduce(206), // >>, reduce: PowTerm
			reduce(206), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(206), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(206), // ||=, reduce: PowTerm
			reduce(206), // ^^=, reduce: PowTerm
			reduce(206), // &&=, reduce: PowTerm
			reduce(206), // ??=, reduce: PowTerm
			reduce(206), // ???=, reduce: PowTerm
			reduce(206), // |=, reduce: PowTerm
			reduce(206), // &=, reduce: PowTerm
			reduce(206), // ^=, reduce: PowTerm
			reduce(206), // <<=, reduce: PowTerm
			reduce(206), // >>=, reduce: PowTerm
			reduce(206), // >>>=, reduce: PowTerm
			reduce(206), // +=, reduce: PowTerm
			reduce(206), // .=, reduce: PowTerm
			reduce(206), // -=, reduce: PowTerm
			reduce(206), // *=, reduce: PowTerm
			reduce(206), // /=, reduce: PowTerm
			reduce(206), // //=, reduce: PowTerm
			reduce(206), // %=, reduce: PowTerm
			reduce(206), // **=, reduce: PowTerm
			reduce(206), // ?, reduce: PowTerm
			nil,         // :
			reduce(206), // ||, reduce: PowTerm
			reduce(206), // ^^, reduce: PowTerm
			reduce(206), // &&, reduce: PowTerm
			reduce(206), // ??, reduce: PowTerm
			reduce(206), // ???, reduce: PowTerm
			reduce(206), // =~, reduce: PowTerm
			reduce(206), // !=~, reduce: PowTerm
			reduce(206), // ==, reduce: PowTerm
			reduce(206), // !=, reduce: PowTerm
			reduce(206), // <=>, reduce: PowTerm
			reduce(206), // >=, reduce: PowTerm
			reduce(206), // <, reduce: PowTerm
			reduce(206), // <=, reduce: PowTerm
			reduce(206), // ^, reduce: PowTerm
			reduce(206), // &, reduce: PowTerm
			reduce(206), // <<, reduce: PowTerm
			reduce(206), // >>>, reduce: PowTerm
			reduce(206), // +, reduce: PowTerm
			reduce(206), // -, reduce: PowTerm
			reduce(206), // .+, reduce: PowTerm
			reduce(206), // .-, reduce: PowTerm
			reduce(206), // *, reduce: PowTerm
			reduce(206), // /, reduce: PowTerm
			reduce(206), // //, reduce: PowTerm
			reduce(206), // %, reduce: PowTerm
			reduce(206), // .*, reduce: PowTerm
			reduce(206), // ./, reduce: PowTerm
			reduce(206), // .//, reduce: PowTerm
			reduce(206), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4535
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(205), // ;, reduce: PowTerm
			nil,         // {
			nil,         // }
			reduce(205), // =, reduce: PowTerm
			nil,         // unset
			nil,         // filter
			reduce(205), // >, reduce: PowTerm
			reduce(205), // >>, reduce: PowTerm
			reduce(205), // |, reduce: PowTerm
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(205), // ,, reduce: PowTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(205), // ||=, reduce: PowTerm
			reduce(205), // ^^=, reduce: PowTerm
			reduce(205), // &&=, reduce: PowTerm
			reduce(205), // ??=, reduce: PowTerm
			reduce(205), // ???=, reduce: PowTerm
			reduce(205), // |=, reduce: PowTerm
			reduce(205), // &=, reduce: PowTerm
			reduce(205), // ^=, reduce: PowTerm
			reduce(205), // <<=, reduce: PowTerm
			reduce(205), // >>=, reduce: PowTerm
			reduce(205), // >>>=, reduce: PowTerm
			reduce(205), // +=, reduce: PowTerm
			reduce(205), // .=, reduce: PowTerm
			reduce(205), // -=, reduce: PowTerm
			reduce(205), // *=, reduce: PowTerm
			reduce(205), // /=, reduce: PowTerm
			reduce(205), // //=, reduce: PowTerm
			reduce(205), // %=, reduce: PowTerm
			reduce(205), // **=, reduce: PowTerm
			reduce(205), // ?, reduce: PowTerm
			nil,         // :
			reduce(205), // ||, reduce: PowTerm
			reduce(205), // ^^, reduce: PowTerm
			reduce(205), // &&, reduce: PowTerm
			reduce(205), // ??, reduce: PowTerm
			reduce(205), // ???, reduce: PowTerm
			reduce(205), // =~, reduce: PowTerm
			reduce(205), // !=~, reduce: PowTerm
			reduce(205), // ==, reduce: PowTerm
			reduce(205), // !=, reduce: PowTerm
			reduce(205), // <=>, reduce: PowTerm
			reduce(205), // >=, reduce: PowTerm
			reduce(205), // <, reduce: PowTerm
			reduce(205), // <=, reduce: PowTerm
			reduce(205), // ^, reduce: PowTerm
			reduce(205), // &, reduce: PowTerm
			reduce(205), // <<, reduce: PowTerm
			reduce(205), // >>>, reduce: PowTerm
			reduce(205), // +, reduce: PowTerm
			reduce(205), // -, reduce: PowTerm
			reduce(205), // .+, reduce: PowTerm
			reduce(205), // .-, reduce: PowTerm
			reduce(205), // *, reduce: PowTerm
			reduce(205), // /, reduce: PowTerm
			reduce(205), // //, reduce: PowTerm
			reduce(205), // %, reduce: PowTerm
			reduce(205), // .*, reduce: PowTerm
			reduce(205), // ./, reduce: PowTerm
			reduce(205), // .//, reduce: PowTerm
			reduce(205), // ., reduce: PowTerm
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4536
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(260), // ;, reduce: ArrayOrMapIndexAccess
			nil,         // {
			nil,         // }
			reduce(260), // =, reduce: ArrayOrMapIndexAccess
			nil,         // unset
			nil,         // filter
			reduce(260), // >, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>, reduce: ArrayOrMapIndexAccess
			reduce(260), // |, reduce: ArrayOrMapIndexAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(260), // ,, reduce: ArrayOrMapIndexAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(260), // ||=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^=, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???=, reduce: ArrayOrMapIndexAccess
			reduce(260), // |=, reduce: ArrayOrMapIndexAccess
			reduce(260), // &=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<=, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>=, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>=, reduce: ArrayOrMapIndexAccess
			reduce(260), // +=, reduce: ArrayOrMapIndexAccess
			reduce(260), // .=, reduce: ArrayOrMapIndexAccess
			reduce(260), // -=, reduce: ArrayOrMapIndexAccess
			reduce(260), // *=, reduce: ArrayOrMapIndexAccess
			reduce(260), // /=, reduce: ArrayOrMapIndexAccess
			reduce(260), // //=, reduce: ArrayOrMapIndexAccess
			reduce(260), // %=, reduce: ArrayOrMapIndexAccess
			reduce(260), // **=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ?, reduce: ArrayOrMapIndexAccess
			nil,         // :
			reduce(260), // ||, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &&, reduce: ArrayOrMapIndexAccess
			reduce(260), // ??, reduce: ArrayOrMapIndexAccess
			reduce(260), // ???, reduce: ArrayOrMapIndexAccess
			reduce(260), // =~, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=~, reduce: ArrayOrMapIndexAccess
			reduce(260), // ==, reduce: ArrayOrMapIndexAccess
			reduce(260), // !=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=>, reduce: ArrayOrMapIndexAccess
			reduce(260), // >=, reduce: ArrayOrMapIndexAccess
			reduce(260), // <, reduce: ArrayOrMapIndexAccess
			reduce(260), // <=, reduce: ArrayOrMapIndexAccess
			reduce(260), // ^, reduce: ArrayOrMapIndexAccess
			reduce(260), // &, reduce: ArrayOrMapIndexAccess
			reduce(260), // <<, reduce: ArrayOrMapIndexAccess
			reduce(260), // >>>, reduce: ArrayOrMapIndexAccess
			reduce(260), // +, reduce: ArrayOrMapIndexAccess
			reduce(260), // -, reduce: ArrayOrMapIndexAccess
			reduce(260), // .+, reduce: ArrayOrMapIndexAccess
			reduce(260), // .-, reduce: ArrayOrMapIndexAccess
			reduce(260), // *, reduce: ArrayOrMapIndexAccess
			reduce(260), // /, reduce: ArrayOrMapIndexAccess
			reduce(260), // //, reduce: ArrayOrMapIndexAccess
			reduce(260), // %, reduce: ArrayOrMapIndexAccess
			reduce(260), // .*, reduce: ArrayOrMapIndexAccess
			reduce(260), // ./, reduce: ArrayOrMapIndexAccess
			reduce(260), // .//, reduce: ArrayOrMapIndexAccess
			reduce(260), // ., reduce: ArrayOrMapIndexAccess
			nil,         // !
			nil,         // ~
			reduce(260), // **, reduce: ArrayOrMapIndexAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(260), // [, reduce: ArrayOrMapIndexAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(260), // [[, reduce: ArrayOrMapIndexAccess
			reduce(260), // [[[, reduce: ArrayOrMapIndexAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4537
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(614),  // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(618),  // (
			nil,         // )
			nil,         // emitp
			shift(630),  // field_name
			shift(631),  // $[
			shift(4613), // ]
			shift(632),  // braced_field_name
			shift(633),  // $[[
			shift(634),  // $[[[
			shift(635),  // full_srec
			shift(637),  // oosvar_name
			shift(638),  // @[
			shift(639),  // braced_oosvar_name
			shift(640),  // full_oosvar
			shift(641),  // all
			shift(642),  // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(657),  // +
			shift(659),  // -
			shift(660),  // .+
			shift(661),  // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(664),  // !
			shift(665),  // ~
			nil,         // **
			shift(670),  // string_literal
			shift(671),  // regex_case_insensitive
			shift(672),  // int_literal
			shift(673),  // float_literal
			shift(674),  // boolean_literal
			shift(675),  // null_literal
			shift(676),  // inf_literal
			shift(677),  // nan_literal
			shift(678),  // const_M_PI
			shift(679),  // const_M_E
			shift(680),  // panic
			shift(682),  // [
			shift(684),  // ctx_IPS
			shift(685),  // ctx_IFS
			shift(686),  // ctx_IRS
			shift(687),  // ctx_OPS
			shift(688),  // ctx_OFS
			shift(689),  // ctx_ORS
			shift(690),  // ctx_FLATSEP
			shift(691),  // ctx_NF
			shift(692),  // ctx_NR
			shift(693),  // ctx_FNR
			shift(694),  // ctx_FILENAME
			shift(695),  // ctx_FILENUM
			shift(697),  // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(704),  // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4538
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4614), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4539
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(266), // ;, reduce: ArraySliceAccess
			nil,         // {
			nil,         // }
			reduce(266), // =, reduce: ArraySliceAccess
			nil,         // unset
			nil,         // filter
			reduce(266), // >, reduce: ArraySliceAccess
			reduce(266), // >>, reduce: ArraySliceAccess
			reduce(266), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(266), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(266), // ||=, reduce: ArraySliceAccess
			reduce(266), // ^^=, reduce: ArraySliceAccess
			reduce(266), // &&=, reduce: ArraySliceAccess
			reduce(266), // ??=, reduce: ArraySliceAccess
			reduce(266), // ???=, reduce: ArraySliceAccess
			reduce(266), // |=, reduce: ArraySliceAccess
			reduce(266), // &=, reduce: ArraySliceAccess
			reduce(266), // ^=, reduce: ArraySliceAccess
			reduce(266), // <<=, reduce: ArraySliceAccess
			reduce(266), // >>=, reduce: ArraySliceAccess
			reduce(266), // >>>=, reduce: ArraySliceAccess
			reduce(266), // +=, reduce: ArraySliceAccess
			reduce(266), // .=, reduce: ArraySliceAccess
			reduce(266), // -=, reduce: ArraySliceAccess
			reduce(266), // *=, reduce: ArraySliceAccess
			reduce(266), // /=, reduce: ArraySliceAccess
			reduce(266), // //=, reduce: ArraySliceAccess
			reduce(266), // %=, reduce: ArraySliceAccess
			reduce(266), // **=, reduce: ArraySliceAccess
			reduce(266), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(266), // ||, reduce: ArraySliceAccess
			reduce(266), // ^^, reduce: ArraySliceAccess
			reduce(266), // &&, reduce: ArraySliceAccess
			reduce(266), // ??, reduce: ArraySliceAccess
			reduce(266), // ???, reduce: ArraySliceAccess
			reduce(266), // =~, reduce: ArraySliceAccess
			reduce(266), // !=~, reduce: ArraySliceAccess
			reduce(266), // ==, reduce: ArraySliceAccess
			reduce(266), // !=, reduce: ArraySliceAccess
			reduce(266), // <=>, reduce: ArraySliceAccess
			reduce(266), // >=, reduce: ArraySliceAccess
			reduce(266), // <, reduce: ArraySliceAccess
			reduce(266), // <=, reduce: ArraySliceAccess
			reduce(266), // ^, reduce: ArraySliceAccess
			reduce(266), // &, reduce: ArraySliceAccess
			reduce(266), // <<, reduce: ArraySliceAccess
			reduce(266), // >>>, reduce: ArraySliceAccess
			reduce(266), // +, reduce: ArraySliceAccess
			reduce(266), // -, reduce: ArraySliceAccess
			reduce(266), // .+, reduce: ArraySliceAccess
			reduce(266), // .-, reduce: ArraySliceAccess
			reduce(266), // *, reduce: ArraySliceAccess
			reduce(266), // /, reduce: ArraySliceAccess
			reduce(266), // //, reduce: ArraySliceAccess
			reduce(266), // %, reduce: ArraySliceAccess
			reduce(266), // .*, reduce: ArraySliceAccess
			reduce(266), // ./, reduce: ArraySliceAccess
			reduce(266), // .//, reduce: ArraySliceAccess
			reduce(266), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(266), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(266), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(266), // [[, reduce: ArraySliceAccess
			reduce(266), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4540
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4615), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4541
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4616), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4542
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(254), // ;, reduce: ENV
			nil,         // {
			nil,         // }
			reduce(254), // =, reduce: ENV
			nil,         // unset
			nil,         // filter
			reduce(254), // >, reduce: ENV
			reduce(254), // >>, reduce: ENV
			reduce(254), // |, reduce: ENV
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(254), // ,, reduce: ENV
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(254), // ||=, reduce: ENV
			reduce(254), // ^^=, reduce: ENV
			reduce(254), // &&=, reduce: ENV
			reduce(254), // ??=, reduce: ENV
			reduce(254), // ???=, reduce: ENV
			reduce(254), // |=, reduce: ENV
			reduce(254), // &=, reduce: ENV
			reduce(254), // ^=, reduce: ENV
			reduce(254), // <<=, reduce: ENV
			reduce(254), // >>=, reduce: ENV
			reduce(254), // >>>=, reduce: ENV
			reduce(254), // +=, reduce: ENV
			reduce(254), // .=, reduce: ENV
			reduce(254), // -=, reduce: ENV
			reduce(254), // *=, reduce: ENV
			reduce(254), // /=, reduce: ENV
			reduce(254), // //=, reduce: ENV
			reduce(254), // %=, reduce: ENV
			reduce(254), // **=, reduce: ENV
			reduce(254), // ?, reduce: ENV
			nil,         // :
			reduce(254), // ||, reduce: ENV
			reduce(254), // ^^, reduce: ENV
			reduce(254), // &&, reduce: ENV
			reduce(254), // ??, reduce: ENV
			reduce(254), // ???, reduce: ENV
			reduce(254), // =~, reduce: ENV
			reduce(254), // !=~, reduce: ENV
			reduce(254), // ==, reduce: ENV
			reduce(254), // !=, reduce: ENV
			reduce(254), // <=>, reduce: ENV
			reduce(254), // >=, reduce: ENV
			reduce(254), // <, reduce: ENV
			reduce(254), // <=, reduce: ENV
			reduce(254), // ^, reduce: ENV
			reduce(254), // &, reduce: ENV
			reduce(254), // <<, reduce: ENV
			reduce(254), // >>>, reduce: ENV
			reduce(254), // +, reduce: ENV
			reduce(254), // -, reduce: ENV
			reduce(254), // .+, reduce: ENV
			reduce(254), // .-, reduce: ENV
			reduce(254), // *, reduce: ENV
			reduce(254), // /, reduce: ENV
			reduce(254), // //, reduce: ENV
			reduce(254), // %, reduce: ENV
			reduce(254), // .*, reduce: ENV
			reduce(254), // ./, reduce: ENV
			reduce(254), // .//, reduce: ENV
			reduce(254), // ., reduce: ENV
			nil,         // !
			nil,         // ~
			reduce(254), // **, reduce: ENV
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(254), // [, reduce: ENV
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(254), // [[, reduce: ENV
			reduce(254), // [[[, reduce: ENV
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4543
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(269), // ;, reduce: FunctionCallsite
			nil,         // {
			nil,         // }
			reduce(269), // =, reduce: FunctionCallsite
			nil,         // unset
			nil,         // filter
			reduce(269), // >, reduce: FunctionCallsite
			reduce(269), // >>, reduce: FunctionCallsite
			reduce(269), // |, reduce: FunctionCallsite
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(269), // ,, reduce: FunctionCallsite
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(269), // ||=, reduce: FunctionCallsite
			reduce(269), // ^^=, reduce: FunctionCallsite
			reduce(269), // &&=, reduce: FunctionCallsite
			reduce(269), // ??=, reduce: FunctionCallsite
			reduce(269), // ???=, reduce: FunctionCallsite
			reduce(269), // |=, reduce: FunctionCallsite
			reduce(269), // &=, reduce: FunctionCallsite
			reduce(269), // ^=, reduce: FunctionCallsite
			reduce(269), // <<=, reduce: FunctionCallsite
			reduce(269), // >>=, reduce: FunctionCallsite
			reduce(269), // >>>=, reduce: FunctionCallsite
			reduce(269), // +=, reduce: FunctionCallsite
			reduce(269), // .=, reduce: FunctionCallsite
			reduce(269), // -=, reduce: FunctionCallsite
			reduce(269), // *=, reduce: FunctionCallsite
			reduce(269), // /=, reduce: FunctionCallsite
			reduce(269), // //=, reduce: FunctionCallsite
			reduce(269), // %=, reduce: FunctionCallsite
			reduce(269), // **=, reduce: FunctionCallsite
			reduce(269), // ?, reduce: FunctionCallsite
			nil,         // :
			reduce(269), // ||, reduce: FunctionCallsite
			reduce(269), // ^^, reduce: FunctionCallsite
			reduce(269), // &&, reduce: FunctionCallsite
			reduce(269), // ??, reduce: FunctionCallsite
			reduce(269), // ???, reduce: FunctionCallsite
			reduce(269), // =~, reduce: FunctionCallsite
			reduce(269), // !=~, reduce: FunctionCallsite
			reduce(269), // ==, reduce: FunctionCallsite
			reduce(269), // !=, reduce: FunctionCallsite
			reduce(269), // <=>, reduce: FunctionCallsite
			reduce(269), // >=, reduce: FunctionCallsite
			reduce(269), // <, reduce: FunctionCallsite
			reduce(269), // <=, reduce: FunctionCallsite
			reduce(269), // ^, reduce: FunctionCallsite
			reduce(269), // &, reduce: FunctionCallsite
			reduce(269), // <<, reduce: FunctionCallsite
			reduce(269), // >>>, reduce: FunctionCallsite
			reduce(269), // +, reduce: FunctionCallsite
			reduce(269), // -, reduce: FunctionCallsite
			reduce(269), // .+, reduce: FunctionCallsite
			reduce(269), // .-, reduce: FunctionCallsite
			reduce(269), // *, reduce: FunctionCallsite
			reduce(269), // /, reduce: FunctionCallsite
			reduce(269), // //, reduce: FunctionCallsite
			reduce(269), // %, reduce: FunctionCallsite
			reduce(269), // .*, reduce: FunctionCallsite
			reduce(269), // ./, reduce: FunctionCallsite
			reduce(269), // .//, reduce: FunctionCallsite
			reduce(269), // ., reduce: FunctionCallsite
			nil,         // !
			nil,         // ~
			reduce(269), // **, reduce: FunctionCallsite
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(269), // [, reduce: FunctionCallsite
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(269), // [[, reduce: FunctionCallsite
			reduce(269), // [[[, reduce: FunctionCallsite
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4544
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4545
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4546
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4547
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4548
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4549
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4550
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4551
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4552
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4553
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4554
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4555
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4556
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4557
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4558
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4559
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4560
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4561
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4562
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4563
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1663), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(1668), // (
			nil,         // )
			nil,         // emitp
			shift(1681), // field_name
			shift(1682), // $[
			nil,         // ]
			shift(1683), // braced_field_name
			shift(1684), // $[[
			shift(1685), // $[[[
			shift(1686), // full_srec
			shift(1688), // oosvar_name
			shift(1689), // @[
			shift(1690), // braced_oosvar_name
			shift(1691), // full_oosvar
			shift(1692), // all
			shift(1693), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1708), // +
			shift(1710), // -
			shift(1711), // .+
			shift(1712), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1715), // !
			shift(1716), // ~
			nil,         // **
			shift(1721), // string_literal
			shift(1722), // regex_case_insensitive
			shift(1723), // int_literal
			shift(1724), // float_literal
			shift(1725), // boolean_literal
			shift(1726), // null_literal
			shift(1727), // inf_literal
			shift(1728), // nan_literal
			shift(1729), // const_M_PI
			shift(1730), // const_M_E
			shift(1731), // panic
			shift(1733), // [
			shift(1735), // ctx_IPS
			shift(1736), // ctx_IFS
			shift(1737), // ctx_IRS
			shift(1738), // ctx_OPS
			shift(1739), // ctx_OFS
			shift(1740), // ctx_ORS
			shift(1741), // ctx_FLATSEP
			shift(1742), // ctx_NF
			shift(1743), // ctx_NR
			shift(1744), // ctx_FNR
			shift(1745), // ctx_FILENAME
			shift(1746), // ctx_FILENUM
			shift(1748), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1755), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4564
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(1763), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2872), // (
			nil,         // )
			nil,         // emitp
			shift(1781), // field_name
			shift(1782), // $[
			nil,         // ]
			shift(1783), // braced_field_name
			shift(1784), // $[[
			shift(1785), // $[[[
			shift(1786), // full_srec
			shift(1788), // oosvar_name
			shift(1789), // @[
			shift(1790), // braced_oosvar_name
			shift(1791), // full_oosvar
			shift(1792), // all
			shift(2873), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1808), // +
			shift(1810), // -
			shift(1811), // .+
			shift(1812), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1815), // !
			shift(1816), // ~
			nil,         // **
			shift(1821), // string_literal
			shift(1822), // regex_case_insensitive
			shift(1823), // int_literal
			shift(1824), // float_literal
			shift(1825), // boolean_literal
			shift(1826), // null_literal
			shift(1827), // inf_literal
			shift(1828), // nan_literal
			shift(1829), // const_M_PI
			shift(1830), // const_M_E
			shift(1831), // panic
			shift(1833), // [
			shift(1835), // ctx_IPS
			shift(1836), // ctx_IFS
			shift(1837), // ctx_IRS
			shift(1838), // ctx_OPS
			shift(1839), // ctx_OFS
			shift(1840), // ctx_ORS
			shift(1841), // ctx_FLATSEP
			shift(1842), // ctx_NF
			shift(1843), // ctx_NR
			shift(1844), // ctx_FNR
			shift(1845), // ctx_FILENAME
			shift(1846), // ctx_FILENUM
			shift(1848), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(1856), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4565
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(289), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4566
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4640), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			shift(4641), // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4567
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4642), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4568
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4431), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4569
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(323), // $, reduce: NamedFunctionDefinition
			nil,         // empty
			reduce(323), // ;, reduce: NamedFunctionDefinition
			reduce(323), // {, reduce: NamedFunctionDefinition
			nil,         // }
			nil,         // =
			reduce(323), // unset, reduce: NamedFunctionDefinition
			reduce(323), // filter, reduce: NamedFunctionDefinition
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(323), // print, reduce: NamedFunctionDefinition
			nil,         // ,
			reduce(323), // printn, reduce: NamedFunctionDefinition
			reduce(323), // eprint, reduce: NamedFunctionDefinition
			reduce(323), // eprintn, reduce: NamedFunctionDefinition
			reduce(323), // dump, reduce: NamedFunctionDefinition
			reduce(323), // edump, reduce: NamedFunctionDefinition
			reduce(323), // tee, reduce: NamedFunctionDefinition
			reduce(323), // emitf, reduce: NamedFunctionDefinition
			reduce(323), // emit1, reduce: NamedFunctionDefinition
			reduce(323), // emit, reduce: NamedFunctionDefinition
			reduce(323), // (, reduce: NamedFunctionDefinition
			nil,         // )
			reduce(323), // emitp, reduce: NamedFunctionDefinition
			reduce(323), // field_name, reduce: NamedFunctionDefinition
			reduce(323), // $[, reduce: NamedFunctionDefinition
			nil,         // ]
			reduce(323), // braced_field_name, reduce: NamedFunctionDefinition
			reduce(323), // $[[, reduce: NamedFunctionDefinition
			reduce(323), // $[[[, reduce: NamedFunctionDefinition
			reduce(323), // full_srec, reduce: NamedFunctionDefinition
			reduce(323), // oosvar_name, reduce: NamedFunctionDefinition
			reduce(323), // @[, reduce: NamedFunctionDefinition
			reduce(323), // braced_oosvar_name, reduce: NamedFunctionDefinition
			reduce(323), // full_oosvar, reduce: NamedFunctionDefinition
			reduce(323), // all, reduce: NamedFunctionDefinition
			reduce(323), // non_sigil_name, reduce: NamedFunctionDefinition
			reduce(323), // arr, reduce: NamedFunctionDefinition
			reduce(323), // bool, reduce: NamedFunctionDefinition
			reduce(323), // float, reduce: NamedFunctionDefinition
			reduce(323), // int, reduce: NamedFunctionDefinition
			reduce(323), // map, reduce: NamedFunctionDefinition
			reduce(323), // num, reduce: NamedFunctionDefinition
			reduce(323), // str, reduce: NamedFunctionDefinition
			reduce(323), // var, reduce: NamedFunctionDefinition
			reduce(323), // funct, reduce: NamedFunctionDefinition
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(323), // +, reduce: NamedFunctionDefinition
			reduce(323), // -, reduce: NamedFunctionDefinition
			reduce(323), // .+, reduce: NamedFunctionDefinition
			reduce(323), // .-, reduce: NamedFunctionDefinition
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(323), // !, reduce: NamedFunctionDefinition
			reduce(323), // ~, reduce: NamedFunctionDefinition
			nil,         // **
			reduce(323), // string_literal, reduce: NamedFunctionDefinition
			reduce(323), // regex_case_insensitive, reduce: NamedFunctionDefinition
			reduce(323), // int_literal, reduce: NamedFunctionDefinition
			reduce(323), // float_literal, reduce: NamedFunctionDefinition
			reduce(323), // boolean_literal, reduce: NamedFunctionDefinition
			reduce(323), // null_literal, reduce: NamedFunctionDefinition
			reduce(323), // inf_literal, reduce: NamedFunctionDefinition
			reduce(323), // nan_literal, reduce: NamedFunctionDefinition
			reduce(323), // const_M_PI, reduce: NamedFunctionDefinition
			reduce(323), // const_M_E, reduce: NamedFunctionDefinition
			reduce(323), // panic, reduce: NamedFunctionDefinition
			reduce(323), // [, reduce: NamedFunctionDefinition
			reduce(323), // ctx_IPS, reduce: NamedFunctionDefinition
			reduce(323), // ctx_IFS, reduce: NamedFunctionDefinition
			reduce(323), // ctx_IRS, reduce: NamedFunctionDefinition
			reduce(323), // ctx_OPS, reduce: NamedFunctionDefinition
			reduce(323), // ctx_OFS, reduce: NamedFunctionDefinition
			reduce(323), // ctx_ORS, reduce: NamedFunctionDefinition
			reduce(323), // ctx_FLATSEP, reduce: NamedFunctionDefinition
			reduce(323), // ctx_NF, reduce: NamedFunctionDefinition
			reduce(323), // ctx_NR, reduce: NamedFunctionDefinition
			reduce(323), // ctx_FNR, reduce: NamedFunctionDefinition
			reduce(323), // ctx_FILENAME, reduce: NamedFunctionDefinition
			reduce(323), // ctx_FILENUM, reduce: NamedFunctionDefinition
			reduce(323), // env, reduce: NamedFunctionDefinition
			nil,         // [[
			nil,         // [[[
			reduce(323), // call, reduce: NamedFunctionDefinition
			reduce(323), // begin, reduce: NamedFunctionDefinition
			reduce(323), // end, reduce: NamedFunctionDefinition
			reduce(323), // if, reduce: NamedFunctionDefinition
			nil,         // elif
			nil,         // else
			reduce(323), // while, reduce: NamedFunctionDefinition
			reduce(323), // do, reduce: NamedFunctionDefinition
			reduce(323), // for, reduce: NamedFunctionDefinition
			nil,         // in
			reduce(323), // break, reduce: NamedFunctionDefinition
			reduce(323), // continue, reduce: NamedFunctionDefinition
			reduce(323), // func, reduce: NamedFunctionDefinition
			reduce(323), // subr, reduce: NamedFunctionDefinition
			reduce(323), // return, reduce: NamedFunctionDefinition
		},
	},
	actionRow{ // S4570
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(263), // }, reduce: ArraySliceAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(263), // >, reduce: ArraySliceAccess
			reduce(263), // >>, reduce: ArraySliceAccess
			reduce(263), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(263), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(263), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(263), // ||, reduce: ArraySliceAccess
			reduce(263), // ^^, reduce: ArraySliceAccess
			reduce(263), // &&, reduce: ArraySliceAccess
			reduce(263), // ??, reduce: ArraySliceAccess
			reduce(263), // ???, reduce: ArraySliceAccess
			reduce(263), // =~, reduce: ArraySliceAccess
			reduce(263), // !=~, reduce: ArraySliceAccess
			reduce(263), // ==, reduce: ArraySliceAccess
			reduce(263), // !=, reduce: ArraySliceAccess
			reduce(263), // <=>, reduce: ArraySliceAccess
			reduce(263), // >=, reduce: ArraySliceAccess
			reduce(263), // <, reduce: ArraySliceAccess
			reduce(263), // <=, reduce: ArraySliceAccess
			reduce(263), // ^, reduce: ArraySliceAccess
			reduce(263), // &, reduce: ArraySliceAccess
			reduce(263), // <<, reduce: ArraySliceAccess
			reduce(263), // >>>, reduce: ArraySliceAccess
			reduce(263), // +, reduce: ArraySliceAccess
			reduce(263), // -, reduce: ArraySliceAccess
			reduce(263), // .+, reduce: ArraySliceAccess
			reduce(263), // .-, reduce: ArraySliceAccess
			reduce(263), // *, reduce: ArraySliceAccess
			reduce(263), // /, reduce: ArraySliceAccess
			reduce(263), // //, reduce: ArraySliceAccess
			reduce(263), // %, reduce: ArraySliceAccess
			reduce(263), // .*, reduce: ArraySliceAccess
			reduce(263), // ./, reduce: ArraySliceAccess
			reduce(263), // .//, reduce: ArraySliceAccess
			reduce(263), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(263), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(263), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(263), // [[, reduce: ArraySliceAccess
			reduce(263), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4571
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(262), // }, reduce: ArrayOrMapPositionalValueAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(262), // >, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |, reduce: ArrayOrMapPositionalValueAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(262), // ,, reduce: ArrayOrMapPositionalValueAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(262), // ?, reduce: ArrayOrMapPositionalValueAccess
			nil,         // :
			reduce(262), // ||, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // =~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ==, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .+, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .-, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .*, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ./, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .//, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ., reduce: ArrayOrMapPositionalValueAccess
			nil,         // !
			nil,         // ~
			reduce(262), // **, reduce: ArrayOrMapPositionalValueAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(262), // [, reduce: ArrayOrMapPositionalValueAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(262), // [[, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // [[[, reduce: ArrayOrMapPositionalValueAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4572
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4644), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4573
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4444), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4574
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(263), // ;, reduce: ArraySliceAccess
			nil,         // {
			reduce(263), // }, reduce: ArraySliceAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(263), // >, reduce: ArraySliceAccess
			reduce(263), // >>, reduce: ArraySliceAccess
			reduce(263), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(263), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(263), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(263), // ||, reduce: ArraySliceAccess
			reduce(263), // ^^, reduce: ArraySliceAccess
			reduce(263), // &&, reduce: ArraySliceAccess
			reduce(263), // ??, reduce: ArraySliceAccess
			reduce(263), // ???, reduce: ArraySliceAccess
			reduce(263), // =~, reduce: ArraySliceAccess
			reduce(263), // !=~, reduce: ArraySliceAccess
			reduce(263), // ==, reduce: ArraySliceAccess
			reduce(263), // !=, reduce: ArraySliceAccess
			reduce(263), // <=>, reduce: ArraySliceAccess
			reduce(263), // >=, reduce: ArraySliceAccess
			reduce(263), // <, reduce: ArraySliceAccess
			reduce(263), // <=, reduce: ArraySliceAccess
			reduce(263), // ^, reduce: ArraySliceAccess
			reduce(263), // &, reduce: ArraySliceAccess
			reduce(263), // <<, reduce: ArraySliceAccess
			reduce(263), // >>>, reduce: ArraySliceAccess
			reduce(263), // +, reduce: ArraySliceAccess
			reduce(263), // -, reduce: ArraySliceAccess
			reduce(263), // .+, reduce: ArraySliceAccess
			reduce(263), // .-, reduce: ArraySliceAccess
			reduce(263), // *, reduce: ArraySliceAccess
			reduce(263), // /, reduce: ArraySliceAccess
			reduce(263), // //, reduce: ArraySliceAccess
			reduce(263), // %, reduce: ArraySliceAccess
			reduce(263), // .*, reduce: ArraySliceAccess
			reduce(263), // ./, reduce: ArraySliceAccess
			reduce(263), // .//, reduce: ArraySliceAccess
			reduce(263), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(263), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(263), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(263), // [[, reduce: ArraySliceAccess
			reduce(263), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4575
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(262), // ;, reduce: ArrayOrMapPositionalValueAccess
			nil,         // {
			reduce(262), // }, reduce: ArrayOrMapPositionalValueAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(262), // >, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |, reduce: ArrayOrMapPositionalValueAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(262), // ,, reduce: ArrayOrMapPositionalValueAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(262), // ?, reduce: ArrayOrMapPositionalValueAccess
			nil,         // :
			reduce(262), // ||, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // =~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ==, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .+, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .-, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .*, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ./, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .//, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ., reduce: ArrayOrMapPositionalValueAccess
			nil,         // !
			nil,         // ~
			reduce(262), // **, reduce: ArrayOrMapPositionalValueAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(262), // [, reduce: ArrayOrMapPositionalValueAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(262), // [[, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // [[[, reduce: ArrayOrMapPositionalValueAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4576
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4646), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4577
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4456), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4578
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(263), // ;, reduce: ArraySliceAccess
			nil,         // {
			reduce(263), // }, reduce: ArraySliceAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(263), // >, reduce: ArraySliceAccess
			reduce(263), // >>, reduce: ArraySliceAccess
			reduce(263), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(263), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(263), // ||, reduce: ArraySliceAccess
			reduce(263), // ^^, reduce: ArraySliceAccess
			reduce(263), // &&, reduce: ArraySliceAccess
			reduce(263), // ??, reduce: ArraySliceAccess
			reduce(263), // ???, reduce: ArraySliceAccess
			reduce(263), // =~, reduce: ArraySliceAccess
			reduce(263), // !=~, reduce: ArraySliceAccess
			reduce(263), // ==, reduce: ArraySliceAccess
			reduce(263), // !=, reduce: ArraySliceAccess
			reduce(263), // <=>, reduce: ArraySliceAccess
			reduce(263), // >=, reduce: ArraySliceAccess
			reduce(263), // <, reduce: ArraySliceAccess
			reduce(263), // <=, reduce: ArraySliceAccess
			reduce(263), // ^, reduce: ArraySliceAccess
			reduce(263), // &, reduce: ArraySliceAccess
			reduce(263), // <<, reduce: ArraySliceAccess
			reduce(263), // >>>, reduce: ArraySliceAccess
			reduce(263), // +, reduce: ArraySliceAccess
			reduce(263), // -, reduce: ArraySliceAccess
			reduce(263), // .+, reduce: ArraySliceAccess
			reduce(263), // .-, reduce: ArraySliceAccess
			reduce(263), // *, reduce: ArraySliceAccess
			reduce(263), // /, reduce: ArraySliceAccess
			reduce(263), // //, reduce: ArraySliceAccess
			reduce(263), // %, reduce: ArraySliceAccess
			reduce(263), // .*, reduce: ArraySliceAccess
			reduce(263), // ./, reduce: ArraySliceAccess
			reduce(263), // .//, reduce: ArraySliceAccess
			reduce(263), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(263), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(263), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(263), // [[, reduce: ArraySliceAccess
			reduce(263), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4579
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(262), // ;, reduce: ArrayOrMapPositionalValueAccess
			nil,         // {
			reduce(262), // }, reduce: ArrayOrMapPositionalValueAccess
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(262), // >, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |, reduce: ArrayOrMapPositionalValueAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(262), // ?, reduce: ArrayOrMapPositionalValueAccess
			nil,         // :
			reduce(262), // ||, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // =~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ==, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .+, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .-, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .*, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ./, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .//, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ., reduce: ArrayOrMapPositionalValueAccess
			nil,         // !
			nil,         // ~
			reduce(262), // **, reduce: ArrayOrMapPositionalValueAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(262), // [, reduce: ArrayOrMapPositionalValueAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(262), // [[, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // [[[, reduce: ArrayOrMapPositionalValueAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4580
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4648), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4581
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4466), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4582
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4583
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2043), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			shift(2048), // (
			nil,         // )
			nil,         // emitp
			shift(2060), // field_name
			shift(2061), // $[
			nil,         // ]
			shift(2062), // braced_field_name
			shift(2063), // $[[
			shift(2064), // $[[[
			shift(2065), // full_srec
			shift(2067), // oosvar_name
			shift(2068), // @[
			shift(2069), // braced_oosvar_name
			shift(2070), // full_oosvar
			shift(2071), // all
			shift(2072), // non_sigil_name
			nil,         // arr
			nil,         // bool
			shift(202),  // float
			shift(203),  // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(2087), // +
			shift(2089), // -
			shift(2090), // .+
			shift(2091), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(2094), // !
			shift(2095), // ~
			nil,         // **
			shift(2100), // string_literal
			shift(2101), // regex_case_insensitive
			shift(2102), // int_literal
			shift(2103), // float_literal
			shift(2104), // boolean_literal
			shift(2105), // null_literal
			shift(2106), // inf_literal
			shift(2107), // nan_literal
			shift(2108), // const_M_PI
			shift(2109), // const_M_E
			shift(2110), // panic
			shift(2112), // [
			shift(2114), // ctx_IPS
			shift(2115), // ctx_IFS
			shift(2116), // ctx_IRS
			shift(2117), // ctx_OPS
			shift(2118), // ctx_OFS
			shift(2119), // ctx_ORS
			shift(2120), // ctx_FLATSEP
			shift(2121), // ctx_NF
			shift(2122), // ctx_NR
			shift(2123), // ctx_FNR
			shift(2124), // ctx_FILENAME
			shift(2125), // ctx_FILENUM
			shift(2127), // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			shift(2134), // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4584
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(13), // ;, reduce: StatementBlockInBraces
			reduce(13), // {, reduce: StatementBlockInBraces
			reduce(13), // }, reduce: StatementBlockInBraces
			nil,        // =
			reduce(13), // unset, reduce: StatementBlockInBraces
			reduce(13), // filter, reduce: StatementBlockInBraces
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			reduce(13), // print, reduce: StatementBlockInBraces
			nil,        // ,
			reduce(13), // printn, reduce: StatementBlockInBraces
			reduce(13), // eprint, reduce: StatementBlockInBraces
			reduce(13), // eprintn, reduce: StatementBlockInBraces
			reduce(13), // dump, reduce: StatementBlockInBraces
			reduce(13), // edump, reduce: StatementBlockInBraces
			reduce(13), // tee, reduce: StatementBlockInBraces
			reduce(13), // emitf, reduce: StatementBlockInBraces
			reduce(13), // emit1, reduce: StatementBlockInBraces
			reduce(13), // emit, reduce: StatementBlockInBraces
			reduce(13), // (, reduce: StatementBlockInBraces
			nil,        // )
			reduce(13), // emitp, reduce: StatementBlockInBraces
			reduce(13), // field_name, reduce: StatementBlockInBraces
			reduce(13), // $[, reduce: StatementBlockInBraces
			nil,        // ]
			reduce(13), // braced_field_name, reduce: StatementBlockInBraces
			reduce(13), // $[[, reduce: StatementBlockInBraces
			reduce(13), // $[[[, reduce: StatementBlockInBraces
			reduce(13), // full_srec, reduce: StatementBlockInBraces
			reduce(13), // oosvar_name, reduce: StatementBlockInBraces
			reduce(13), // @[, reduce: StatementBlockInBraces
			reduce(13), // braced_oosvar_name, reduce: StatementBlockInBraces
			reduce(13), // full_oosvar, reduce: StatementBlockInBraces
			reduce(13), // all, reduce: StatementBlockInBraces
			reduce(13), // non_sigil_name, reduce: StatementBlockInBraces
			reduce(13), // arr, reduce: StatementBlockInBraces
			reduce(13), // bool, reduce: StatementBlockInBraces
			reduce(13), // float, reduce: StatementBlockInBraces
			reduce(13), // int, reduce: StatementBlockInBraces
			reduce(13), // map, reduce: StatementBlockInBraces
			reduce(13), // num, reduce: StatementBlockInBraces
			reduce(13), // str, reduce: StatementBlockInBraces
			reduce(13), // var, reduce: StatementBlockInBraces
			reduce(13), // funct, reduce: StatementBlockInBraces
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			reduce(13), // !, reduce: StatementBlockInBraces
			reduce(13), // ~, reduce: StatementBlockInBraces
			nil,        // **
			reduce(13), // string_literal, reduce: StatementBlockInBraces
			reduce(13), // regex_case_insensitive, reduce: StatementBlockInBraces
			reduce(13), // int_literal, reduce: StatementBlockInBraces
			reduce(13), // float_literal, reduce: StatementBlockInBraces
			reduce(13), // boolean_literal, reduce: StatementBlockInBraces
			reduce(13), // null_literal, reduce: StatementBlockInBraces
			reduce(13), // inf_literal, reduce: StatementBlockInBraces
			reduce(13), // nan_literal, reduce: StatementBlockInBraces
			reduce(13), // const_M_PI, reduce: StatementBlockInBraces
			reduce(13), // const_M_E, reduce: StatementBlockInBraces
			reduce(13), // panic, reduce: StatementBlockInBraces
			reduce(13), // [, reduce: StatementBlockInBraces
			reduce(13), // ctx_IPS, reduce: StatementBlockInBraces
			reduce(13), // ctx_IFS, reduce: StatementBlockInBraces
			reduce(13), // ctx_IRS, reduce: StatementBlockInBraces
			reduce(13), // ctx_OPS, reduce: StatementBlockInBraces
			reduce(13), // ctx_OFS, reduce: StatementBlockInBraces
			reduce(13), // ctx_ORS, reduce: StatementBlockInBraces
			reduce(13), // ctx_FLATSEP, reduce: StatementBlockInBraces
			reduce(13), // ctx_NF, reduce: StatementBlockInBraces
			reduce(13), // ctx_NR, reduce: StatementBlockInBraces
			reduce(13), // ctx_FNR, reduce: StatementBlockInBraces
			reduce(13), // ctx_FILENAME, reduce: StatementBlockInBraces
			reduce(13), // ctx_FILENUM, reduce: StatementBlockInBraces
			reduce(13), // env, reduce: StatementBlockInBraces
			nil,        // [[
			nil,        // [[[
			reduce(13), // call, reduce: StatementBlockInBraces
			reduce(13), // begin, reduce: StatementBlockInBraces
			reduce(13), // end, reduce: StatementBlockInBraces
			reduce(13), // if, reduce: StatementBlockInBraces
			reduce(13), // elif, reduce: StatementBlockInBraces
			reduce(13), // else, reduce: StatementBlockInBraces
			reduce(13), // while, reduce: StatementBlockInBraces
			reduce(13), // do, reduce: StatementBlockInBraces
			reduce(13), // for, reduce: StatementBlockInBraces
			nil,        // in
			reduce(13), // break, reduce: StatementBlockInBraces
			reduce(13), // continue, reduce: StatementBlockInBraces
			reduce(13), // func, reduce: StatementBlockInBraces
			reduce(13), // subr, reduce: StatementBlockInBraces
			reduce(13), // return, reduce: StatementBlockInBraces
		},
	},
	actionRow{ // S4585
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4652), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4586
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(303), // ;, reduce: ForLoopOneVariable
			reduce(303), // {, reduce: ForLoopOneVariable
			reduce(303), // }, reduce: ForLoopOneVariable
			nil,         // =
			reduce(303), // unset, reduce: ForLoopOneVariable
			reduce(303), // filter, reduce: ForLoopOneVariable
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(303), // print, reduce: ForLoopOneVariable
			nil,         // ,
			reduce(303), // printn, reduce: ForLoopOneVariable
			reduce(303), // eprint, reduce: ForLoopOneVariable
			reduce(303), // eprintn, reduce: ForLoopOneVariable
			reduce(303), // dump, reduce: ForLoopOneVariable
			reduce(303), // edump, reduce: ForLoopOneVariable
			reduce(303), // tee, reduce: ForLoopOneVariable
			reduce(303), // emitf, reduce: ForLoopOneVariable
			reduce(303), // emit1, reduce: ForLoopOneVariable
			reduce(303), // emit, reduce: ForLoopOneVariable
			reduce(303), // (, reduce: ForLoopOneVariable
			nil,         // )
			reduce(303), // emitp, reduce: ForLoopOneVariable
			reduce(303), // field_name, reduce: ForLoopOneVariable
			reduce(303), // $[, reduce: ForLoopOneVariable
			nil,         // ]
			reduce(303), // braced_field_name, reduce: ForLoopOneVariable
			reduce(303), // $[[, reduce: ForLoopOneVariable
			reduce(303), // $[[[, reduce: ForLoopOneVariable
			reduce(303), // full_srec, reduce: ForLoopOneVariable
			reduce(303), // oosvar_name, reduce: ForLoopOneVariable
			reduce(303), // @[, reduce: ForLoopOneVariable
			reduce(303), // braced_oosvar_name, reduce: ForLoopOneVariable
			reduce(303), // full_oosvar, reduce: ForLoopOneVariable
			reduce(303), // all, reduce: ForLoopOneVariable
			reduce(303), // non_sigil_name, reduce: ForLoopOneVariable
			reduce(303), // arr, reduce: ForLoopOneVariable
			reduce(303), // bool, reduce: ForLoopOneVariable
			reduce(303), // float, reduce: ForLoopOneVariable
			reduce(303), // int, reduce: ForLoopOneVariable
			reduce(303), // map, reduce: ForLoopOneVariable
			reduce(303), // num, reduce: ForLoopOneVariable
			reduce(303), // str, reduce: ForLoopOneVariable
			reduce(303), // var, reduce: ForLoopOneVariable
			reduce(303), // funct, reduce: ForLoopOneVariable
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(303), // +, reduce: ForLoopOneVariable
			reduce(303), // -, reduce: ForLoopOneVariable
			reduce(303), // .+, reduce: ForLoopOneVariable
			reduce(303), // .-, reduce: ForLoopOneVariable
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(303), // !, reduce: ForLoopOneVariable
			reduce(303), // ~, reduce: ForLoopOneVariable
			nil,         // **
			reduce(303), // string_literal, reduce: ForLoopOneVariable
			reduce(303), // regex_case_insensitive, reduce: ForLoopOneVariable
			reduce(303), // int_literal, reduce: ForLoopOneVariable
			reduce(303), // float_literal, reduce: ForLoopOneVariable
			reduce(303), // boolean_literal, reduce: ForLoopOneVariable
			reduce(303), // null_literal, reduce: ForLoopOneVariable
			reduce(303), // inf_literal, reduce: ForLoopOneVariable
			reduce(303), // nan_literal, reduce: ForLoopOneVariable
			reduce(303), // const_M_PI, reduce: ForLoopOneVariable
			reduce(303), // const_M_E, reduce: ForLoopOneVariable
			reduce(303), // panic, reduce: ForLoopOneVariable
			reduce(303), // [, reduce: ForLoopOneVariable
			reduce(303), // ctx_IPS, reduce: ForLoopOneVariable
			reduce(303), // ctx_IFS, reduce: ForLoopOneVariable
			reduce(303), // ctx_IRS, reduce: ForLoopOneVariable
			reduce(303), // ctx_OPS, reduce: ForLoopOneVariable
			reduce(303), // ctx_OFS, reduce: ForLoopOneVariable
			reduce(303), // ctx_ORS, reduce: ForLoopOneVariable
			reduce(303), // ctx_FLATSEP, reduce: ForLoopOneVariable
			reduce(303), // ctx_NF, reduce: ForLoopOneVariable
			reduce(303), // ctx_NR, reduce: ForLoopOneVariable
			reduce(303), // ctx_FNR, reduce: ForLoopOneVariable
			reduce(303), // ctx_FILENAME, reduce: ForLoopOneVariable
			reduce(303), // ctx_FILENUM, reduce: ForLoopOneVariable
			reduce(303), // env, reduce: ForLoopOneVariable
			nil,         // [[
			nil,         // [[[
			reduce(303), // call, reduce: ForLoopOneVariable
			reduce(303), // begin, reduce: ForLoopOneVariable
			reduce(303), // end, reduce: ForLoopOneVariable
			reduce(303), // if, reduce: ForLoopOneVariable
			nil,         // elif
			nil,         // else
			reduce(303), // while, reduce: ForLoopOneVariable
			reduce(303), // do, reduce: ForLoopOneVariable
			reduce(303), // for, reduce: ForLoopOneVariable
			nil,         // in
			reduce(303), // break, reduce: ForLoopOneVariable
			reduce(303), // continue, reduce: ForLoopOneVariable
			reduce(303), // func, reduce: ForLoopOneVariable
			reduce(303), // subr, reduce: ForLoopOneVariable
			reduce(303), // return, reduce: ForLoopOneVariable
		},
	},
	actionRow{ // S4587
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			shift(4653), // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4588
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			shift(4564), // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4654), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4589
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(13), // ;, reduce: StatementBlockInBraces
			reduce(13), // {, reduce: StatementBlockInBraces
			reduce(13), // }, reduce: StatementBlockInBraces
			reduce(13), // =, reduce: StatementBlockInBraces
			nil,        // unset
			nil,        // filter
			reduce(13), // >, reduce: StatementBlockInBraces
			reduce(13), // >>, reduce: StatementBlockInBraces
			reduce(13), // |, reduce: StatementBlockInBraces
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			reduce(13), // ||=, reduce: StatementBlockInBraces
			reduce(13), // ^^=, reduce: StatementBlockInBraces
			reduce(13), // &&=, reduce: StatementBlockInBraces
			reduce(13), // ??=, reduce: StatementBlockInBraces
			reduce(13), // ???=, reduce: StatementBlockInBraces
			reduce(13), // |=, reduce: StatementBlockInBraces
			reduce(13), // &=, reduce: StatementBlockInBraces
			reduce(13), // ^=, reduce: StatementBlockInBraces
			reduce(13), // <<=, reduce: StatementBlockInBraces
			reduce(13), // >>=, reduce: StatementBlockInBraces
			reduce(13), // >>>=, reduce: StatementBlockInBraces
			reduce(13), // +=, reduce: StatementBlockInBraces
			reduce(13), // .=, reduce: StatementBlockInBraces
			reduce(13), // -=, reduce: StatementBlockInBraces
			reduce(13), // *=, reduce: StatementBlockInBraces
			reduce(13), // /=, reduce: StatementBlockInBraces
			reduce(13), // //=, reduce: StatementBlockInBraces
			reduce(13), // %=, reduce: StatementBlockInBraces
			reduce(13), // **=, reduce: StatementBlockInBraces
			reduce(13), // ?, reduce: StatementBlockInBraces
			nil,        // :
			reduce(13), // ||, reduce: StatementBlockInBraces
			reduce(13), // ^^, reduce: StatementBlockInBraces
			reduce(13), // &&, reduce: StatementBlockInBraces
			reduce(13), // ??, reduce: StatementBlockInBraces
			reduce(13), // ???, reduce: StatementBlockInBraces
			reduce(13), // =~, reduce: StatementBlockInBraces
			reduce(13), // !=~, reduce: StatementBlockInBraces
			reduce(13), // ==, reduce: StatementBlockInBraces
			reduce(13), // !=, reduce: StatementBlockInBraces
			reduce(13), // <=>, reduce: StatementBlockInBraces
			reduce(13), // >=, reduce: StatementBlockInBraces
			reduce(13), // <, reduce: StatementBlockInBraces
			reduce(13), // <=, reduce: StatementBlockInBraces
			reduce(13), // ^, reduce: StatementBlockInBraces
			reduce(13), // &, reduce: StatementBlockInBraces
			reduce(13), // <<, reduce: StatementBlockInBraces
			reduce(13), // >>>, reduce: StatementBlockInBraces
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			reduce(13), // *, reduce: StatementBlockInBraces
			reduce(13), // /, reduce: StatementBlockInBraces
			reduce(13), // //, reduce: StatementBlockInBraces
			reduce(13), // %, reduce: StatementBlockInBraces
			reduce(13), // .*, reduce: StatementBlockInBraces
			reduce(13), // ./, reduce: StatementBlockInBraces
			reduce(13), // .//, reduce: StatementBlockInBraces
			reduce(13), // ., reduce: StatementBlockInBraces
			nil,        // !
			nil,        // ~
			reduce(13), // **, reduce: StatementBlockInBraces
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(13), // [, reduce: StatementBlockInBraces
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(13), // [[, reduce: StatementBlockInBraces
			reduce(13), // [[[, reduce: StatementBlockInBraces
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4590
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(325), // ;, reduce: UnnamedFunctionDefinition
			reduce(325), // {, reduce: UnnamedFunctionDefinition
			reduce(325), // }, reduce: UnnamedFunctionDefinition
			reduce(325), // =, reduce: UnnamedFunctionDefinition
			nil,         // unset
			nil,         // filter
			reduce(325), // >, reduce: UnnamedFunctionDefinition
			reduce(325), // >>, reduce: UnnamedFunctionDefinition
			reduce(325), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(325), // ||=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^=, reduce: UnnamedFunctionDefinition
			reduce(325), // &&=, reduce: UnnamedFunctionDefinition
			reduce(325), // ??=, reduce: UnnamedFunctionDefinition
			reduce(325), // ???=, reduce: UnnamedFunctionDefinition
			reduce(325), // |=, reduce: UnnamedFunctionDefinition
			reduce(325), // &=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^=, reduce: UnnamedFunctionDefinition
			reduce(325), // <<=, reduce: UnnamedFunctionDefinition
			reduce(325), // >>=, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>=, reduce: UnnamedFunctionDefinition
			reduce(325), // +=, reduce: UnnamedFunctionDefinition
			reduce(325), // .=, reduce: UnnamedFunctionDefinition
			reduce(325), // -=, reduce: UnnamedFunctionDefinition
			reduce(325), // *=, reduce: UnnamedFunctionDefinition
			reduce(325), // /=, reduce: UnnamedFunctionDefinition
			reduce(325), // //=, reduce: UnnamedFunctionDefinition
			reduce(325), // %=, reduce: UnnamedFunctionDefinition
			reduce(325), // **=, reduce: UnnamedFunctionDefinition
			reduce(325), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(325), // ||, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^, reduce: UnnamedFunctionDefinition
			reduce(325), // &&, reduce: UnnamedFunctionDefinition
			reduce(325), // ??, reduce: UnnamedFunctionDefinition
			reduce(325), // ???, reduce: UnnamedFunctionDefinition
			reduce(325), // =~, reduce: UnnamedFunctionDefinition
			reduce(325), // !=~, reduce: UnnamedFunctionDefinition
			reduce(325), // ==, reduce: UnnamedFunctionDefinition
			reduce(325), // !=, reduce: UnnamedFunctionDefinition
			reduce(325), // <=>, reduce: UnnamedFunctionDefinition
			reduce(325), // >=, reduce: UnnamedFunctionDefinition
			reduce(325), // <, reduce: UnnamedFunctionDefinition
			reduce(325), // <=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^, reduce: UnnamedFunctionDefinition
			reduce(325), // &, reduce: UnnamedFunctionDefinition
			reduce(325), // <<, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>, reduce: UnnamedFunctionDefinition
			reduce(325), // +, reduce: UnnamedFunctionDefinition
			reduce(325), // -, reduce: UnnamedFunctionDefinition
			reduce(325), // .+, reduce: UnnamedFunctionDefinition
			reduce(325), // .-, reduce: UnnamedFunctionDefinition
			reduce(325), // *, reduce: UnnamedFunctionDefinition
			reduce(325), // /, reduce: UnnamedFunctionDefinition
			reduce(325), // //, reduce: UnnamedFunctionDefinition
			reduce(325), // %, reduce: UnnamedFunctionDefinition
			reduce(325), // .*, reduce: UnnamedFunctionDefinition
			reduce(325), // ./, reduce: UnnamedFunctionDefinition
			reduce(325), // .//, reduce: UnnamedFunctionDefinition
			reduce(325), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(325), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(325), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(325), // [[, reduce: UnnamedFunctionDefinition
			reduce(325), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4591
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2042), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4592
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(13), // >, reduce: StatementBlockInBraces
			reduce(13), // >>, reduce: StatementBlockInBraces
			reduce(13), // |, reduce: StatementBlockInBraces
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(13), // ,, reduce: StatementBlockInBraces
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(13), // ?, reduce: StatementBlockInBraces
			nil,        // :
			reduce(13), // ||, reduce: StatementBlockInBraces
			reduce(13), // ^^, reduce: StatementBlockInBraces
			reduce(13), // &&, reduce: StatementBlockInBraces
			reduce(13), // ??, reduce: StatementBlockInBraces
			reduce(13), // ???, reduce: StatementBlockInBraces
			reduce(13), // =~, reduce: StatementBlockInBraces
			reduce(13), // !=~, reduce: StatementBlockInBraces
			reduce(13), // ==, reduce: StatementBlockInBraces
			reduce(13), // !=, reduce: StatementBlockInBraces
			reduce(13), // <=>, reduce: StatementBlockInBraces
			reduce(13), // >=, reduce: StatementBlockInBraces
			reduce(13), // <, reduce: StatementBlockInBraces
			reduce(13), // <=, reduce: StatementBlockInBraces
			reduce(13), // ^, reduce: StatementBlockInBraces
			reduce(13), // &, reduce: StatementBlockInBraces
			reduce(13), // <<, reduce: StatementBlockInBraces
			reduce(13), // >>>, reduce: StatementBlockInBraces
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			reduce(13), // *, reduce: StatementBlockInBraces
			reduce(13), // /, reduce: StatementBlockInBraces
			reduce(13), // //, reduce: StatementBlockInBraces
			reduce(13), // %, reduce: StatementBlockInBraces
			reduce(13), // .*, reduce: StatementBlockInBraces
			reduce(13), // ./, reduce: StatementBlockInBraces
			reduce(13), // .//, reduce: StatementBlockInBraces
			reduce(13), // ., reduce: StatementBlockInBraces
			nil,        // !
			nil,        // ~
			reduce(13), // **, reduce: StatementBlockInBraces
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(13), // [, reduce: StatementBlockInBraces
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(13), // [[, reduce: StatementBlockInBraces
			reduce(13), // [[[, reduce: StatementBlockInBraces
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4593
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(325), // >, reduce: UnnamedFunctionDefinition
			reduce(325), // >>, reduce: UnnamedFunctionDefinition
			reduce(325), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(325), // ,, reduce: UnnamedFunctionDefinition
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(325), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(325), // ||, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^, reduce: UnnamedFunctionDefinition
			reduce(325), // &&, reduce: UnnamedFunctionDefinition
			reduce(325), // ??, reduce: UnnamedFunctionDefinition
			reduce(325), // ???, reduce: UnnamedFunctionDefinition
			reduce(325), // =~, reduce: UnnamedFunctionDefinition
			reduce(325), // !=~, reduce: UnnamedFunctionDefinition
			reduce(325), // ==, reduce: UnnamedFunctionDefinition
			reduce(325), // !=, reduce: UnnamedFunctionDefinition
			reduce(325), // <=>, reduce: UnnamedFunctionDefinition
			reduce(325), // >=, reduce: UnnamedFunctionDefinition
			reduce(325), // <, reduce: UnnamedFunctionDefinition
			reduce(325), // <=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^, reduce: UnnamedFunctionDefinition
			reduce(325), // &, reduce: UnnamedFunctionDefinition
			reduce(325), // <<, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>, reduce: UnnamedFunctionDefinition
			reduce(325), // +, reduce: UnnamedFunctionDefinition
			reduce(325), // -, reduce: UnnamedFunctionDefinition
			reduce(325), // .+, reduce: UnnamedFunctionDefinition
			reduce(325), // .-, reduce: UnnamedFunctionDefinition
			reduce(325), // *, reduce: UnnamedFunctionDefinition
			reduce(325), // /, reduce: UnnamedFunctionDefinition
			reduce(325), // //, reduce: UnnamedFunctionDefinition
			reduce(325), // %, reduce: UnnamedFunctionDefinition
			reduce(325), // .*, reduce: UnnamedFunctionDefinition
			reduce(325), // ./, reduce: UnnamedFunctionDefinition
			reduce(325), // .//, reduce: UnnamedFunctionDefinition
			reduce(325), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(325), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(325), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(325), // [[, reduce: UnnamedFunctionDefinition
			reduce(325), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4594
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(13), // >, reduce: StatementBlockInBraces
			reduce(13), // >>, reduce: StatementBlockInBraces
			reduce(13), // |, reduce: StatementBlockInBraces
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			reduce(13), // ], reduce: StatementBlockInBraces
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(13), // ?, reduce: StatementBlockInBraces
			reduce(13), // :, reduce: StatementBlockInBraces
			reduce(13), // ||, reduce: StatementBlockInBraces
			reduce(13), // ^^, reduce: StatementBlockInBraces
			reduce(13), // &&, reduce: StatementBlockInBraces
			reduce(13), // ??, reduce: StatementBlockInBraces
			reduce(13), // ???, reduce: StatementBlockInBraces
			reduce(13), // =~, reduce: StatementBlockInBraces
			reduce(13), // !=~, reduce: StatementBlockInBraces
			reduce(13), // ==, reduce: StatementBlockInBraces
			reduce(13), // !=, reduce: StatementBlockInBraces
			reduce(13), // <=>, reduce: StatementBlockInBraces
			reduce(13), // >=, reduce: StatementBlockInBraces
			reduce(13), // <, reduce: StatementBlockInBraces
			reduce(13), // <=, reduce: StatementBlockInBraces
			reduce(13), // ^, reduce: StatementBlockInBraces
			reduce(13), // &, reduce: StatementBlockInBraces
			reduce(13), // <<, reduce: StatementBlockInBraces
			reduce(13), // >>>, reduce: StatementBlockInBraces
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			reduce(13), // *, reduce: StatementBlockInBraces
			reduce(13), // /, reduce: StatementBlockInBraces
			reduce(13), // //, reduce: StatementBlockInBraces
			reduce(13), // %, reduce: StatementBlockInBraces
			reduce(13), // .*, reduce: StatementBlockInBraces
			reduce(13), // ./, reduce: StatementBlockInBraces
			reduce(13), // .//, reduce: StatementBlockInBraces
			reduce(13), // ., reduce: StatementBlockInBraces
			nil,        // !
			nil,        // ~
			reduce(13), // **, reduce: StatementBlockInBraces
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(13), // [, reduce: StatementBlockInBraces
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(13), // [[, reduce: StatementBlockInBraces
			reduce(13), // [[[, reduce: StatementBlockInBraces
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4595
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(325), // >, reduce: UnnamedFunctionDefinition
			reduce(325), // >>, reduce: UnnamedFunctionDefinition
			reduce(325), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			reduce(325), // ], reduce: UnnamedFunctionDefinition
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(325), // ?, reduce: UnnamedFunctionDefinition
			reduce(325), // :, reduce: UnnamedFunctionDefinition
			reduce(325), // ||, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^, reduce: UnnamedFunctionDefinition
			reduce(325), // &&, reduce: UnnamedFunctionDefinition
			reduce(325), // ??, reduce: UnnamedFunctionDefinition
			reduce(325), // ???, reduce: UnnamedFunctionDefinition
			reduce(325), // =~, reduce: UnnamedFunctionDefinition
			reduce(325), // !=~, reduce: UnnamedFunctionDefinition
			reduce(325), // ==, reduce: UnnamedFunctionDefinition
			reduce(325), // !=, reduce: UnnamedFunctionDefinition
			reduce(325), // <=>, reduce: UnnamedFunctionDefinition
			reduce(325), // >=, reduce: UnnamedFunctionDefinition
			reduce(325), // <, reduce: UnnamedFunctionDefinition
			reduce(325), // <=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^, reduce: UnnamedFunctionDefinition
			reduce(325), // &, reduce: UnnamedFunctionDefinition
			reduce(325), // <<, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>, reduce: UnnamedFunctionDefinition
			reduce(325), // +, reduce: UnnamedFunctionDefinition
			reduce(325), // -, reduce: UnnamedFunctionDefinition
			reduce(325), // .+, reduce: UnnamedFunctionDefinition
			reduce(325), // .-, reduce: UnnamedFunctionDefinition
			reduce(325), // *, reduce: UnnamedFunctionDefinition
			reduce(325), // /, reduce: UnnamedFunctionDefinition
			reduce(325), // //, reduce: UnnamedFunctionDefinition
			reduce(325), // %, reduce: UnnamedFunctionDefinition
			reduce(325), // .*, reduce: UnnamedFunctionDefinition
			reduce(325), // ./, reduce: UnnamedFunctionDefinition
			reduce(325), // .//, reduce: UnnamedFunctionDefinition
			reduce(325), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(325), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(325), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(325), // [[, reduce: UnnamedFunctionDefinition
			reduce(325), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4596
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(13), // >, reduce: StatementBlockInBraces
			reduce(13), // >>, reduce: StatementBlockInBraces
			reduce(13), // |, reduce: StatementBlockInBraces
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(13), // ,, reduce: StatementBlockInBraces
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			reduce(13), // ), reduce: StatementBlockInBraces
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(13), // ?, reduce: StatementBlockInBraces
			nil,        // :
			reduce(13), // ||, reduce: StatementBlockInBraces
			reduce(13), // ^^, reduce: StatementBlockInBraces
			reduce(13), // &&, reduce: StatementBlockInBraces
			reduce(13), // ??, reduce: StatementBlockInBraces
			reduce(13), // ???, reduce: StatementBlockInBraces
			reduce(13), // =~, reduce: StatementBlockInBraces
			reduce(13), // !=~, reduce: StatementBlockInBraces
			reduce(13), // ==, reduce: StatementBlockInBraces
			reduce(13), // !=, reduce: StatementBlockInBraces
			reduce(13), // <=>, reduce: StatementBlockInBraces
			reduce(13), // >=, reduce: StatementBlockInBraces
			reduce(13), // <, reduce: StatementBlockInBraces
			reduce(13), // <=, reduce: StatementBlockInBraces
			reduce(13), // ^, reduce: StatementBlockInBraces
			reduce(13), // &, reduce: StatementBlockInBraces
			reduce(13), // <<, reduce: StatementBlockInBraces
			reduce(13), // >>>, reduce: StatementBlockInBraces
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			reduce(13), // *, reduce: StatementBlockInBraces
			reduce(13), // /, reduce: StatementBlockInBraces
			reduce(13), // //, reduce: StatementBlockInBraces
			reduce(13), // %, reduce: StatementBlockInBraces
			reduce(13), // .*, reduce: StatementBlockInBraces
			reduce(13), // ./, reduce: StatementBlockInBraces
			reduce(13), // .//, reduce: StatementBlockInBraces
			reduce(13), // ., reduce: StatementBlockInBraces
			nil,        // !
			nil,        // ~
			reduce(13), // **, reduce: StatementBlockInBraces
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(13), // [, reduce: StatementBlockInBraces
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(13), // [[, reduce: StatementBlockInBraces
			reduce(13), // [[[, reduce: StatementBlockInBraces
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4597
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(325), // >, reduce: UnnamedFunctionDefinition
			reduce(325), // >>, reduce: UnnamedFunctionDefinition
			reduce(325), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(325), // ,, reduce: UnnamedFunctionDefinition
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(325), // ), reduce: UnnamedFunctionDefinition
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(325), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(325), // ||, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^, reduce: UnnamedFunctionDefinition
			reduce(325), // &&, reduce: UnnamedFunctionDefinition
			reduce(325), // ??, reduce: UnnamedFunctionDefinition
			reduce(325), // ???, reduce: UnnamedFunctionDefinition
			reduce(325), // =~, reduce: UnnamedFunctionDefinition
			reduce(325), // !=~, reduce: UnnamedFunctionDefinition
			reduce(325), // ==, reduce: UnnamedFunctionDefinition
			reduce(325), // !=, reduce: UnnamedFunctionDefinition
			reduce(325), // <=>, reduce: UnnamedFunctionDefinition
			reduce(325), // >=, reduce: UnnamedFunctionDefinition
			reduce(325), // <, reduce: UnnamedFunctionDefinition
			reduce(325), // <=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^, reduce: UnnamedFunctionDefinition
			reduce(325), // &, reduce: UnnamedFunctionDefinition
			reduce(325), // <<, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>, reduce: UnnamedFunctionDefinition
			reduce(325), // +, reduce: UnnamedFunctionDefinition
			reduce(325), // -, reduce: UnnamedFunctionDefinition
			reduce(325), // .+, reduce: UnnamedFunctionDefinition
			reduce(325), // .-, reduce: UnnamedFunctionDefinition
			reduce(325), // *, reduce: UnnamedFunctionDefinition
			reduce(325), // /, reduce: UnnamedFunctionDefinition
			reduce(325), // //, reduce: UnnamedFunctionDefinition
			reduce(325), // %, reduce: UnnamedFunctionDefinition
			reduce(325), // .*, reduce: UnnamedFunctionDefinition
			reduce(325), // ./, reduce: UnnamedFunctionDefinition
			reduce(325), // .//, reduce: UnnamedFunctionDefinition
			reduce(325), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(325), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(325), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(325), // [[, reduce: UnnamedFunctionDefinition
			reduce(325), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4598
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(108), // ;, reduce: PositionalFieldValue
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(108), // >, reduce: PositionalFieldValue
			reduce(108), // >>, reduce: PositionalFieldValue
			reduce(108), // |, reduce: PositionalFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(108), // ,, reduce: PositionalFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(108), // ?, reduce: PositionalFieldValue
			nil,         // :
			reduce(108), // ||, reduce: PositionalFieldValue
			reduce(108), // ^^, reduce: PositionalFieldValue
			reduce(108), // &&, reduce: PositionalFieldValue
			reduce(108), // ??, reduce: PositionalFieldValue
			reduce(108), // ???, reduce: PositionalFieldValue
			reduce(108), // =~, reduce: PositionalFieldValue
			reduce(108), // !=~, reduce: PositionalFieldValue
			reduce(108), // ==, reduce: PositionalFieldValue
			reduce(108), // !=, reduce: PositionalFieldValue
			reduce(108), // <=>, reduce: PositionalFieldValue
			reduce(108), // >=, reduce: PositionalFieldValue
			reduce(108), // <, reduce: PositionalFieldValue
			reduce(108), // <=, reduce: PositionalFieldValue
			reduce(108), // ^, reduce: PositionalFieldValue
			reduce(108), // &, reduce: PositionalFieldValue
			reduce(108), // <<, reduce: PositionalFieldValue
			reduce(108), // >>>, reduce: PositionalFieldValue
			reduce(108), // +, reduce: PositionalFieldValue
			reduce(108), // -, reduce: PositionalFieldValue
			reduce(108), // .+, reduce: PositionalFieldValue
			reduce(108), // .-, reduce: PositionalFieldValue
			reduce(108), // *, reduce: PositionalFieldValue
			reduce(108), // /, reduce: PositionalFieldValue
			reduce(108), // //, reduce: PositionalFieldValue
			reduce(108), // %, reduce: PositionalFieldValue
			reduce(108), // .*, reduce: PositionalFieldValue
			reduce(108), // ./, reduce: PositionalFieldValue
			reduce(108), // .//, reduce: PositionalFieldValue
			reduce(108), // ., reduce: PositionalFieldValue
			nil,         // !
			nil,         // ~
			reduce(108), // **, reduce: PositionalFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(108), // [, reduce: PositionalFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(108), // [[, reduce: PositionalFieldValue
			reduce(108), // [[[, reduce: PositionalFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4599
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(149), // ;, reduce: TernaryTerm
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(149), // ,, reduce: TernaryTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4600
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4656), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4601
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(265), // ;, reduce: ArraySliceAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(265), // >, reduce: ArraySliceAccess
			reduce(265), // >>, reduce: ArraySliceAccess
			reduce(265), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(265), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(265), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(265), // ||, reduce: ArraySliceAccess
			reduce(265), // ^^, reduce: ArraySliceAccess
			reduce(265), // &&, reduce: ArraySliceAccess
			reduce(265), // ??, reduce: ArraySliceAccess
			reduce(265), // ???, reduce: ArraySliceAccess
			reduce(265), // =~, reduce: ArraySliceAccess
			reduce(265), // !=~, reduce: ArraySliceAccess
			reduce(265), // ==, reduce: ArraySliceAccess
			reduce(265), // !=, reduce: ArraySliceAccess
			reduce(265), // <=>, reduce: ArraySliceAccess
			reduce(265), // >=, reduce: ArraySliceAccess
			reduce(265), // <, reduce: ArraySliceAccess
			reduce(265), // <=, reduce: ArraySliceAccess
			reduce(265), // ^, reduce: ArraySliceAccess
			reduce(265), // &, reduce: ArraySliceAccess
			reduce(265), // <<, reduce: ArraySliceAccess
			reduce(265), // >>>, reduce: ArraySliceAccess
			reduce(265), // +, reduce: ArraySliceAccess
			reduce(265), // -, reduce: ArraySliceAccess
			reduce(265), // .+, reduce: ArraySliceAccess
			reduce(265), // .-, reduce: ArraySliceAccess
			reduce(265), // *, reduce: ArraySliceAccess
			reduce(265), // /, reduce: ArraySliceAccess
			reduce(265), // //, reduce: ArraySliceAccess
			reduce(265), // %, reduce: ArraySliceAccess
			reduce(265), // .*, reduce: ArraySliceAccess
			reduce(265), // ./, reduce: ArraySliceAccess
			reduce(265), // .//, reduce: ArraySliceAccess
			reduce(265), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(265), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(265), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(265), // [[, reduce: ArraySliceAccess
			reduce(265), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4602
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(264), // ;, reduce: ArraySliceAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(264), // >, reduce: ArraySliceAccess
			reduce(264), // >>, reduce: ArraySliceAccess
			reduce(264), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(264), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(264), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(264), // ||, reduce: ArraySliceAccess
			reduce(264), // ^^, reduce: ArraySliceAccess
			reduce(264), // &&, reduce: ArraySliceAccess
			reduce(264), // ??, reduce: ArraySliceAccess
			reduce(264), // ???, reduce: ArraySliceAccess
			reduce(264), // =~, reduce: ArraySliceAccess
			reduce(264), // !=~, reduce: ArraySliceAccess
			reduce(264), // ==, reduce: ArraySliceAccess
			reduce(264), // !=, reduce: ArraySliceAccess
			reduce(264), // <=>, reduce: ArraySliceAccess
			reduce(264), // >=, reduce: ArraySliceAccess
			reduce(264), // <, reduce: ArraySliceAccess
			reduce(264), // <=, reduce: ArraySliceAccess
			reduce(264), // ^, reduce: ArraySliceAccess
			reduce(264), // &, reduce: ArraySliceAccess
			reduce(264), // <<, reduce: ArraySliceAccess
			reduce(264), // >>>, reduce: ArraySliceAccess
			reduce(264), // +, reduce: ArraySliceAccess
			reduce(264), // -, reduce: ArraySliceAccess
			reduce(264), // .+, reduce: ArraySliceAccess
			reduce(264), // .-, reduce: ArraySliceAccess
			reduce(264), // *, reduce: ArraySliceAccess
			reduce(264), // /, reduce: ArraySliceAccess
			reduce(264), // //, reduce: ArraySliceAccess
			reduce(264), // %, reduce: ArraySliceAccess
			reduce(264), // .*, reduce: ArraySliceAccess
			reduce(264), // ./, reduce: ArraySliceAccess
			reduce(264), // .//, reduce: ArraySliceAccess
			reduce(264), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(264), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(264), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(264), // [[, reduce: ArraySliceAccess
			reduce(264), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4603
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(261), // ;, reduce: ArrayOrMapPositionalNameAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(261), // >, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |, reduce: ArrayOrMapPositionalNameAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(261), // ,, reduce: ArrayOrMapPositionalNameAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(261), // ?, reduce: ArrayOrMapPositionalNameAccess
			nil,         // :
			reduce(261), // ||, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // =~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ==, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .+, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .-, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .*, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ./, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .//, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ., reduce: ArrayOrMapPositionalNameAccess
			nil,         // !
			nil,         // ~
			reduce(261), // **, reduce: ArrayOrMapPositionalNameAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(261), // [, reduce: ArrayOrMapPositionalNameAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(261), // [[, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // [[[, reduce: ArrayOrMapPositionalNameAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4604
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4657), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4605
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(324), // ;, reduce: UnnamedFunctionDefinition
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(324), // >, reduce: UnnamedFunctionDefinition
			reduce(324), // >>, reduce: UnnamedFunctionDefinition
			reduce(324), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(324), // ,, reduce: UnnamedFunctionDefinition
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(324), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(324), // ||, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^, reduce: UnnamedFunctionDefinition
			reduce(324), // &&, reduce: UnnamedFunctionDefinition
			reduce(324), // ??, reduce: UnnamedFunctionDefinition
			reduce(324), // ???, reduce: UnnamedFunctionDefinition
			reduce(324), // =~, reduce: UnnamedFunctionDefinition
			reduce(324), // !=~, reduce: UnnamedFunctionDefinition
			reduce(324), // ==, reduce: UnnamedFunctionDefinition
			reduce(324), // !=, reduce: UnnamedFunctionDefinition
			reduce(324), // <=>, reduce: UnnamedFunctionDefinition
			reduce(324), // >=, reduce: UnnamedFunctionDefinition
			reduce(324), // <, reduce: UnnamedFunctionDefinition
			reduce(324), // <=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^, reduce: UnnamedFunctionDefinition
			reduce(324), // &, reduce: UnnamedFunctionDefinition
			reduce(324), // <<, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>, reduce: UnnamedFunctionDefinition
			reduce(324), // +, reduce: UnnamedFunctionDefinition
			reduce(324), // -, reduce: UnnamedFunctionDefinition
			reduce(324), // .+, reduce: UnnamedFunctionDefinition
			reduce(324), // .-, reduce: UnnamedFunctionDefinition
			reduce(324), // *, reduce: UnnamedFunctionDefinition
			reduce(324), // /, reduce: UnnamedFunctionDefinition
			reduce(324), // //, reduce: UnnamedFunctionDefinition
			reduce(324), // %, reduce: UnnamedFunctionDefinition
			reduce(324), // .*, reduce: UnnamedFunctionDefinition
			reduce(324), // ./, reduce: UnnamedFunctionDefinition
			reduce(324), // .//, reduce: UnnamedFunctionDefinition
			reduce(324), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(324), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(324), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(324), // [[, reduce: UnnamedFunctionDefinition
			reduce(324), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4606
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S4607
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4608
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(304), // $, reduce: ForLoopTwoVariable
			nil,         // empty
			reduce(304), // ;, reduce: ForLoopTwoVariable
			reduce(304), // {, reduce: ForLoopTwoVariable
			nil,         // }
			nil,         // =
			reduce(304), // unset, reduce: ForLoopTwoVariable
			reduce(304), // filter, reduce: ForLoopTwoVariable
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(304), // print, reduce: ForLoopTwoVariable
			nil,         // ,
			reduce(304), // printn, reduce: ForLoopTwoVariable
			reduce(304), // eprint, reduce: ForLoopTwoVariable
			reduce(304), // eprintn, reduce: ForLoopTwoVariable
			reduce(304), // dump, reduce: ForLoopTwoVariable
			reduce(304), // edump, reduce: ForLoopTwoVariable
			reduce(304), // tee, reduce: ForLoopTwoVariable
			reduce(304), // emitf, reduce: ForLoopTwoVariable
			reduce(304), // emit1, reduce: ForLoopTwoVariable
			reduce(304), // emit, reduce: ForLoopTwoVariable
			reduce(304), // (, reduce: ForLoopTwoVariable
			nil,         // )
			reduce(304), // emitp, reduce: ForLoopTwoVariable
			reduce(304), // field_name, reduce: ForLoopTwoVariable
			reduce(304), // $[, reduce: ForLoopTwoVariable
			nil,         // ]
			reduce(304), // braced_field_name, reduce: ForLoopTwoVariable
			reduce(304), // $[[, reduce: ForLoopTwoVariable
			reduce(304), // $[[[, reduce: ForLoopTwoVariable
			reduce(304), // full_srec, reduce: ForLoopTwoVariable
			reduce(304), // oosvar_name, reduce: ForLoopTwoVariable
			reduce(304), // @[, reduce: ForLoopTwoVariable
			reduce(304), // braced_oosvar_name, reduce: ForLoopTwoVariable
			reduce(304), // full_oosvar, reduce: ForLoopTwoVariable
			reduce(304), // all, reduce: ForLoopTwoVariable
			reduce(304), // non_sigil_name, reduce: ForLoopTwoVariable
			reduce(304), // arr, reduce: ForLoopTwoVariable
			reduce(304), // bool, reduce: ForLoopTwoVariable
			reduce(304), // float, reduce: ForLoopTwoVariable
			reduce(304), // int, reduce: ForLoopTwoVariable
			reduce(304), // map, reduce: ForLoopTwoVariable
			reduce(304), // num, reduce: ForLoopTwoVariable
			reduce(304), // str, reduce: ForLoopTwoVariable
			reduce(304), // var, reduce: ForLoopTwoVariable
			reduce(304), // funct, reduce: ForLoopTwoVariable
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(304), // +, reduce: ForLoopTwoVariable
			reduce(304), // -, reduce: ForLoopTwoVariable
			reduce(304), // .+, reduce: ForLoopTwoVariable
			reduce(304), // .-, reduce: ForLoopTwoVariable
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(304), // !, reduce: ForLoopTwoVariable
			reduce(304), // ~, reduce: ForLoopTwoVariable
			nil,         // **
			reduce(304), // string_literal, reduce: ForLoopTwoVariable
			reduce(304), // regex_case_insensitive, reduce: ForLoopTwoVariable
			reduce(304), // int_literal, reduce: ForLoopTwoVariable
			reduce(304), // float_literal, reduce: ForLoopTwoVariable
			reduce(304), // boolean_literal, reduce: ForLoopTwoVariable
			reduce(304), // null_literal, reduce: ForLoopTwoVariable
			reduce(304), // inf_literal, reduce: ForLoopTwoVariable
			reduce(304), // nan_literal, reduce: ForLoopTwoVariable
			reduce(304), // const_M_PI, reduce: ForLoopTwoVariable
			reduce(304), // const_M_E, reduce: ForLoopTwoVariable
			reduce(304), // panic, reduce: ForLoopTwoVariable
			reduce(304), // [, reduce: ForLoopTwoVariable
			reduce(304), // ctx_IPS, reduce: ForLoopTwoVariable
			reduce(304), // ctx_IFS, reduce: ForLoopTwoVariable
			reduce(304), // ctx_IRS, reduce: ForLoopTwoVariable
			reduce(304), // ctx_OPS, reduce: ForLoopTwoVariable
			reduce(304), // ctx_OFS, reduce: ForLoopTwoVariable
			reduce(304), // ctx_ORS, reduce: ForLoopTwoVariable
			reduce(304), // ctx_FLATSEP, reduce: ForLoopTwoVariable
			reduce(304), // ctx_NF, reduce: ForLoopTwoVariable
			reduce(304), // ctx_NR, reduce: ForLoopTwoVariable
			reduce(304), // ctx_FNR, reduce: ForLoopTwoVariable
			reduce(304), // ctx_FILENAME, reduce: ForLoopTwoVariable
			reduce(304), // ctx_FILENUM, reduce: ForLoopTwoVariable
			reduce(304), // env, reduce: ForLoopTwoVariable
			nil,         // [[
			nil,         // [[[
			reduce(304), // call, reduce: ForLoopTwoVariable
			reduce(304), // begin, reduce: ForLoopTwoVariable
			reduce(304), // end, reduce: ForLoopTwoVariable
			reduce(304), // if, reduce: ForLoopTwoVariable
			nil,         // elif
			nil,         // else
			reduce(304), // while, reduce: ForLoopTwoVariable
			reduce(304), // do, reduce: ForLoopTwoVariable
			reduce(304), // for, reduce: ForLoopTwoVariable
			nil,         // in
			reduce(304), // break, reduce: ForLoopTwoVariable
			reduce(304), // continue, reduce: ForLoopTwoVariable
			reduce(304), // func, reduce: ForLoopTwoVariable
			reduce(304), // subr, reduce: ForLoopTwoVariable
			reduce(304), // return, reduce: ForLoopTwoVariable
		},
	},
	actionRow{ // S4609
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4660), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4610
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(108), // ;, reduce: PositionalFieldValue
			nil,         // {
			nil,         // }
			reduce(108), // =, reduce: PositionalFieldValue
			nil,         // unset
			nil,         // filter
			reduce(108), // >, reduce: PositionalFieldValue
			reduce(108), // >>, reduce: PositionalFieldValue
			reduce(108), // |, reduce: PositionalFieldValue
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(108), // ,, reduce: PositionalFieldValue
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(108), // ||=, reduce: PositionalFieldValue
			reduce(108), // ^^=, reduce: PositionalFieldValue
			reduce(108), // &&=, reduce: PositionalFieldValue
			reduce(108), // ??=, reduce: PositionalFieldValue
			reduce(108), // ???=, reduce: PositionalFieldValue
			reduce(108), // |=, reduce: PositionalFieldValue
			reduce(108), // &=, reduce: PositionalFieldValue
			reduce(108), // ^=, reduce: PositionalFieldValue
			reduce(108), // <<=, reduce: PositionalFieldValue
			reduce(108), // >>=, reduce: PositionalFieldValue
			reduce(108), // >>>=, reduce: PositionalFieldValue
			reduce(108), // +=, reduce: PositionalFieldValue
			reduce(108), // .=, reduce: PositionalFieldValue
			reduce(108), // -=, reduce: PositionalFieldValue
			reduce(108), // *=, reduce: PositionalFieldValue
			reduce(108), // /=, reduce: PositionalFieldValue
			reduce(108), // //=, reduce: PositionalFieldValue
			reduce(108), // %=, reduce: PositionalFieldValue
			reduce(108), // **=, reduce: PositionalFieldValue
			reduce(108), // ?, reduce: PositionalFieldValue
			nil,         // :
			reduce(108), // ||, reduce: PositionalFieldValue
			reduce(108), // ^^, reduce: PositionalFieldValue
			reduce(108), // &&, reduce: PositionalFieldValue
			reduce(108), // ??, reduce: PositionalFieldValue
			reduce(108), // ???, reduce: PositionalFieldValue
			reduce(108), // =~, reduce: PositionalFieldValue
			reduce(108), // !=~, reduce: PositionalFieldValue
			reduce(108), // ==, reduce: PositionalFieldValue
			reduce(108), // !=, reduce: PositionalFieldValue
			reduce(108), // <=>, reduce: PositionalFieldValue
			reduce(108), // >=, reduce: PositionalFieldValue
			reduce(108), // <, reduce: PositionalFieldValue
			reduce(108), // <=, reduce: PositionalFieldValue
			reduce(108), // ^, reduce: PositionalFieldValue
			reduce(108), // &, reduce: PositionalFieldValue
			reduce(108), // <<, reduce: PositionalFieldValue
			reduce(108), // >>>, reduce: PositionalFieldValue
			reduce(108), // +, reduce: PositionalFieldValue
			reduce(108), // -, reduce: PositionalFieldValue
			reduce(108), // .+, reduce: PositionalFieldValue
			reduce(108), // .-, reduce: PositionalFieldValue
			reduce(108), // *, reduce: PositionalFieldValue
			reduce(108), // /, reduce: PositionalFieldValue
			reduce(108), // //, reduce: PositionalFieldValue
			reduce(108), // %, reduce: PositionalFieldValue
			reduce(108), // .*, reduce: PositionalFieldValue
			reduce(108), // ./, reduce: PositionalFieldValue
			reduce(108), // .//, reduce: PositionalFieldValue
			reduce(108), // ., reduce: PositionalFieldValue
			nil,         // !
			nil,         // ~
			reduce(108), // **, reduce: PositionalFieldValue
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(108), // [, reduce: PositionalFieldValue
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(108), // [[, reduce: PositionalFieldValue
			reduce(108), // [[[, reduce: PositionalFieldValue
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4611
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(149), // ;, reduce: TernaryTerm
			nil,         // {
			nil,         // }
			reduce(149), // =, reduce: TernaryTerm
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(149), // ,, reduce: TernaryTerm
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(149), // ||=, reduce: TernaryTerm
			reduce(149), // ^^=, reduce: TernaryTerm
			reduce(149), // &&=, reduce: TernaryTerm
			reduce(149), // ??=, reduce: TernaryTerm
			reduce(149), // ???=, reduce: TernaryTerm
			reduce(149), // |=, reduce: TernaryTerm
			reduce(149), // &=, reduce: TernaryTerm
			reduce(149), // ^=, reduce: TernaryTerm
			reduce(149), // <<=, reduce: TernaryTerm
			reduce(149), // >>=, reduce: TernaryTerm
			reduce(149), // >>>=, reduce: TernaryTerm
			reduce(149), // +=, reduce: TernaryTerm
			reduce(149), // .=, reduce: TernaryTerm
			reduce(149), // -=, reduce: TernaryTerm
			reduce(149), // *=, reduce: TernaryTerm
			reduce(149), // /=, reduce: TernaryTerm
			reduce(149), // //=, reduce: TernaryTerm
			reduce(149), // %=, reduce: TernaryTerm
			reduce(149), // **=, reduce: TernaryTerm
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4612
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4661), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4613
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(265), // ;, reduce: ArraySliceAccess
			nil,         // {
			nil,         // }
			reduce(265), // =, reduce: ArraySliceAccess
			nil,         // unset
			nil,         // filter
			reduce(265), // >, reduce: ArraySliceAccess
			reduce(265), // >>, reduce: ArraySliceAccess
			reduce(265), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(265), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(265), // ||=, reduce: ArraySliceAccess
			reduce(265), // ^^=, reduce: ArraySliceAccess
			reduce(265), // &&=, reduce: ArraySliceAccess
			reduce(265), // ??=, reduce: ArraySliceAccess
			reduce(265), // ???=, reduce: ArraySliceAccess
			reduce(265), // |=, reduce: ArraySliceAccess
			reduce(265), // &=, reduce: ArraySliceAccess
			reduce(265), // ^=, reduce: ArraySliceAccess
			reduce(265), // <<=, reduce: ArraySliceAccess
			reduce(265), // >>=, reduce: ArraySliceAccess
			reduce(265), // >>>=, reduce: ArraySliceAccess
			reduce(265), // +=, reduce: ArraySliceAccess
			reduce(265), // .=, reduce: ArraySliceAccess
			reduce(265), // -=, reduce: ArraySliceAccess
			reduce(265), // *=, reduce: ArraySliceAccess
			reduce(265), // /=, reduce: ArraySliceAccess
			reduce(265), // //=, reduce: ArraySliceAccess
			reduce(265), // %=, reduce: ArraySliceAccess
			reduce(265), // **=, reduce: ArraySliceAccess
			reduce(265), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(265), // ||, reduce: ArraySliceAccess
			reduce(265), // ^^, reduce: ArraySliceAccess
			reduce(265), // &&, reduce: ArraySliceAccess
			reduce(265), // ??, reduce: ArraySliceAccess
			reduce(265), // ???, reduce: ArraySliceAccess
			reduce(265), // =~, reduce: ArraySliceAccess
			reduce(265), // !=~, reduce: ArraySliceAccess
			reduce(265), // ==, reduce: ArraySliceAccess
			reduce(265), // !=, reduce: ArraySliceAccess
			reduce(265), // <=>, reduce: ArraySliceAccess
			reduce(265), // >=, reduce: ArraySliceAccess
			reduce(265), // <, reduce: ArraySliceAccess
			reduce(265), // <=, reduce: ArraySliceAccess
			reduce(265), // ^, reduce: ArraySliceAccess
			reduce(265), // &, reduce: ArraySliceAccess
			reduce(265), // <<, reduce: ArraySliceAccess
			reduce(265), // >>>, reduce: ArraySliceAccess
			reduce(265), // +, reduce: ArraySliceAccess
			reduce(265), // -, reduce: ArraySliceAccess
			reduce(265), // .+, reduce: ArraySliceAccess
			reduce(265), // .-, reduce: ArraySliceAccess
			reduce(265), // *, reduce: ArraySliceAccess
			reduce(265), // /, reduce: ArraySliceAccess
			reduce(265), // //, reduce: ArraySliceAccess
			reduce(265), // %, reduce: ArraySliceAccess
			reduce(265), // .*, reduce: ArraySliceAccess
			reduce(265), // ./, reduce: ArraySliceAccess
			reduce(265), // .//, reduce: ArraySliceAccess
			reduce(265), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(265), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(265), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(265), // [[, reduce: ArraySliceAccess
			reduce(265), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4614
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(264), // ;, reduce: ArraySliceAccess
			nil,         // {
			nil,         // }
			reduce(264), // =, reduce: ArraySliceAccess
			nil,         // unset
			nil,         // filter
			reduce(264), // >, reduce: ArraySliceAccess
			reduce(264), // >>, reduce: ArraySliceAccess
			reduce(264), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(264), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(264), // ||=, reduce: ArraySliceAccess
			reduce(264), // ^^=, reduce: ArraySliceAccess
			reduce(264), // &&=, reduce: ArraySliceAccess
			reduce(264), // ??=, reduce: ArraySliceAccess
			reduce(264), // ???=, reduce: ArraySliceAccess
			reduce(264), // |=, reduce: ArraySliceAccess
			reduce(264), // &=, reduce: ArraySliceAccess
			reduce(264), // ^=, reduce: ArraySliceAccess
			reduce(264), // <<=, reduce: ArraySliceAccess
			reduce(264), // >>=, reduce: ArraySliceAccess
			reduce(264), // >>>=, reduce: ArraySliceAccess
			reduce(264), // +=, reduce: ArraySliceAccess
			reduce(264), // .=, reduce: ArraySliceAccess
			reduce(264), // -=, reduce: ArraySliceAccess
			reduce(264), // *=, reduce: ArraySliceAccess
			reduce(264), // /=, reduce: ArraySliceAccess
			reduce(264), // //=, reduce: ArraySliceAccess
			reduce(264), // %=, reduce: ArraySliceAccess
			reduce(264), // **=, reduce: ArraySliceAccess
			reduce(264), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(264), // ||, reduce: ArraySliceAccess
			reduce(264), // ^^, reduce: ArraySliceAccess
			reduce(264), // &&, reduce: ArraySliceAccess
			reduce(264), // ??, reduce: ArraySliceAccess
			reduce(264), // ???, reduce: ArraySliceAccess
			reduce(264), // =~, reduce: ArraySliceAccess
			reduce(264), // !=~, reduce: ArraySliceAccess
			reduce(264), // ==, reduce: ArraySliceAccess
			reduce(264), // !=, reduce: ArraySliceAccess
			reduce(264), // <=>, reduce: ArraySliceAccess
			reduce(264), // >=, reduce: ArraySliceAccess
			reduce(264), // <, reduce: ArraySliceAccess
			reduce(264), // <=, reduce: ArraySliceAccess
			reduce(264), // ^, reduce: ArraySliceAccess
			reduce(264), // &, reduce: ArraySliceAccess
			reduce(264), // <<, reduce: ArraySliceAccess
			reduce(264), // >>>, reduce: ArraySliceAccess
			reduce(264), // +, reduce: ArraySliceAccess
			reduce(264), // -, reduce: ArraySliceAccess
			reduce(264), // .+, reduce: ArraySliceAccess
			reduce(264), // .-, reduce: ArraySliceAccess
			reduce(264), // *, reduce: ArraySliceAccess
			reduce(264), // /, reduce: ArraySliceAccess
			reduce(264), // //, reduce: ArraySliceAccess
			reduce(264), // %, reduce: ArraySliceAccess
			reduce(264), // .*, reduce: ArraySliceAccess
			reduce(264), // ./, reduce: ArraySliceAccess
			reduce(264), // .//, reduce: ArraySliceAccess
			reduce(264), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(264), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(264), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(264), // [[, reduce: ArraySliceAccess
			reduce(264), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4615
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(261), // ;, reduce: ArrayOrMapPositionalNameAccess
			nil,         // {
			nil,         // }
			reduce(261), // =, reduce: ArrayOrMapPositionalNameAccess
			nil,         // unset
			nil,         // filter
			reduce(261), // >, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |, reduce: ArrayOrMapPositionalNameAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(261), // ,, reduce: ArrayOrMapPositionalNameAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(261), // ||=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // |=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // **=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ?, reduce: ArrayOrMapPositionalNameAccess
			nil,         // :
			reduce(261), // ||, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &&, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ??, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ???, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // =~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=~, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ==, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // !=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <=, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ^, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // &, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // <<, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // >>>, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // +, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // -, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .+, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .-, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // *, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // /, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // //, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // %, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .*, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ./, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // .//, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // ., reduce: ArrayOrMapPositionalNameAccess
			nil,         // !
			nil,         // ~
			reduce(261), // **, reduce: ArrayOrMapPositionalNameAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(261), // [, reduce: ArrayOrMapPositionalNameAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(261), // [[, reduce: ArrayOrMapPositionalNameAccess
			reduce(261), // [[[, reduce: ArrayOrMapPositionalNameAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4616
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			shift(4662), // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4617
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(34), // ,, reduce: Assignment
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			reduce(34), // ), reduce: Assignment
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4618
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(128), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(128), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4619
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(129), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(129), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4620
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(130), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(130), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4621
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(131), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(131), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4622
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(132), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(132), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4623
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(133), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(133), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4624
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(134), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(134), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4625
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(135), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(135), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4626
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(136), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(136), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4627
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(137), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(137), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4628
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(138), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(138), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4629
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(139), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(139), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4630
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(140), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(140), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4631
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(141), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(141), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4632
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(142), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(142), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4633
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(143), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(143), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4634
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(144), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(144), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4635
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(145), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(145), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4636
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(146), // ,, reduce: Assignment
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(146), // ), reduce: Assignment
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4637
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(319), // ,, reduce: TripleForUpdate
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			reduce(319), // ), reduce: TripleForUpdate
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4638
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(308), // $, reduce: TripleForLoop
			nil,         // empty
			reduce(308), // ;, reduce: TripleForLoop
			reduce(308), // {, reduce: TripleForLoop
			nil,         // }
			nil,         // =
			reduce(308), // unset, reduce: TripleForLoop
			reduce(308), // filter, reduce: TripleForLoop
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(308), // print, reduce: TripleForLoop
			nil,         // ,
			reduce(308), // printn, reduce: TripleForLoop
			reduce(308), // eprint, reduce: TripleForLoop
			reduce(308), // eprintn, reduce: TripleForLoop
			reduce(308), // dump, reduce: TripleForLoop
			reduce(308), // edump, reduce: TripleForLoop
			reduce(308), // tee, reduce: TripleForLoop
			reduce(308), // emitf, reduce: TripleForLoop
			reduce(308), // emit1, reduce: TripleForLoop
			reduce(308), // emit, reduce: TripleForLoop
			reduce(308), // (, reduce: TripleForLoop
			nil,         // )
			reduce(308), // emitp, reduce: TripleForLoop
			reduce(308), // field_name, reduce: TripleForLoop
			reduce(308), // $[, reduce: TripleForLoop
			nil,         // ]
			reduce(308), // braced_field_name, reduce: TripleForLoop
			reduce(308), // $[[, reduce: TripleForLoop
			reduce(308), // $[[[, reduce: TripleForLoop
			reduce(308), // full_srec, reduce: TripleForLoop
			reduce(308), // oosvar_name, reduce: TripleForLoop
			reduce(308), // @[, reduce: TripleForLoop
			reduce(308), // braced_oosvar_name, reduce: TripleForLoop
			reduce(308), // full_oosvar, reduce: TripleForLoop
			reduce(308), // all, reduce: TripleForLoop
			reduce(308), // non_sigil_name, reduce: TripleForLoop
			reduce(308), // arr, reduce: TripleForLoop
			reduce(308), // bool, reduce: TripleForLoop
			reduce(308), // float, reduce: TripleForLoop
			reduce(308), // int, reduce: TripleForLoop
			reduce(308), // map, reduce: TripleForLoop
			reduce(308), // num, reduce: TripleForLoop
			reduce(308), // str, reduce: TripleForLoop
			reduce(308), // var, reduce: TripleForLoop
			reduce(308), // funct, reduce: TripleForLoop
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(308), // +, reduce: TripleForLoop
			reduce(308), // -, reduce: TripleForLoop
			reduce(308), // .+, reduce: TripleForLoop
			reduce(308), // .-, reduce: TripleForLoop
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(308), // !, reduce: TripleForLoop
			reduce(308), // ~, reduce: TripleForLoop
			nil,         // **
			reduce(308), // string_literal, reduce: TripleForLoop
			reduce(308), // regex_case_insensitive, reduce: TripleForLoop
			reduce(308), // int_literal, reduce: TripleForLoop
			reduce(308), // float_literal, reduce: TripleForLoop
			reduce(308), // boolean_literal, reduce: TripleForLoop
			reduce(308), // null_literal, reduce: TripleForLoop
			reduce(308), // inf_literal, reduce: TripleForLoop
			reduce(308), // nan_literal, reduce: TripleForLoop
			reduce(308), // const_M_PI, reduce: TripleForLoop
			reduce(308), // const_M_E, reduce: TripleForLoop
			reduce(308), // panic, reduce: TripleForLoop
			reduce(308), // [, reduce: TripleForLoop
			reduce(308), // ctx_IPS, reduce: TripleForLoop
			reduce(308), // ctx_IFS, reduce: TripleForLoop
			reduce(308), // ctx_IRS, reduce: TripleForLoop
			reduce(308), // ctx_OPS, reduce: TripleForLoop
			reduce(308), // ctx_OFS, reduce: TripleForLoop
			reduce(308), // ctx_ORS, reduce: TripleForLoop
			reduce(308), // ctx_FLATSEP, reduce: TripleForLoop
			reduce(308), // ctx_NF, reduce: TripleForLoop
			reduce(308), // ctx_NR, reduce: TripleForLoop
			reduce(308), // ctx_FNR, reduce: TripleForLoop
			reduce(308), // ctx_FILENAME, reduce: TripleForLoop
			reduce(308), // ctx_FILENUM, reduce: TripleForLoop
			reduce(308), // env, reduce: TripleForLoop
			nil,         // [[
			nil,         // [[[
			reduce(308), // call, reduce: TripleForLoop
			reduce(308), // begin, reduce: TripleForLoop
			reduce(308), // end, reduce: TripleForLoop
			reduce(308), // if, reduce: TripleForLoop
			nil,         // elif
			nil,         // else
			reduce(308), // while, reduce: TripleForLoop
			reduce(308), // do, reduce: TripleForLoop
			reduce(308), // for, reduce: TripleForLoop
			nil,         // in
			reduce(308), // break, reduce: TripleForLoop
			reduce(308), // continue, reduce: TripleForLoop
			reduce(308), // func, reduce: TripleForLoop
			reduce(308), // subr, reduce: TripleForLoop
			reduce(308), // return, reduce: TripleForLoop
		},
	},
	actionRow{ // S4639
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(324), // ;, reduce: UnnamedFunctionDefinition
			nil,         // {
			nil,         // }
			reduce(324), // =, reduce: UnnamedFunctionDefinition
			nil,         // unset
			nil,         // filter
			reduce(324), // >, reduce: UnnamedFunctionDefinition
			reduce(324), // >>, reduce: UnnamedFunctionDefinition
			reduce(324), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(324), // ,, reduce: UnnamedFunctionDefinition
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(324), // ||=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^=, reduce: UnnamedFunctionDefinition
			reduce(324), // &&=, reduce: UnnamedFunctionDefinition
			reduce(324), // ??=, reduce: UnnamedFunctionDefinition
			reduce(324), // ???=, reduce: UnnamedFunctionDefinition
			reduce(324), // |=, reduce: UnnamedFunctionDefinition
			reduce(324), // &=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^=, reduce: UnnamedFunctionDefinition
			reduce(324), // <<=, reduce: UnnamedFunctionDefinition
			reduce(324), // >>=, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>=, reduce: UnnamedFunctionDefinition
			reduce(324), // +=, reduce: UnnamedFunctionDefinition
			reduce(324), // .=, reduce: UnnamedFunctionDefinition
			reduce(324), // -=, reduce: UnnamedFunctionDefinition
			reduce(324), // *=, reduce: UnnamedFunctionDefinition
			reduce(324), // /=, reduce: UnnamedFunctionDefinition
			reduce(324), // //=, reduce: UnnamedFunctionDefinition
			reduce(324), // %=, reduce: UnnamedFunctionDefinition
			reduce(324), // **=, reduce: UnnamedFunctionDefinition
			reduce(324), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(324), // ||, reduce: UnnamedFunctionDefinition
			reduce(324), // ^^, reduce: UnnamedFunctionDefinition
			reduce(324), // &&, reduce: UnnamedFunctionDefinition
			reduce(324), // ??, reduce: UnnamedFunctionDefinition
			reduce(324), // ???, reduce: UnnamedFunctionDefinition
			reduce(324), // =~, reduce: UnnamedFunctionDefinition
			reduce(324), // !=~, reduce: UnnamedFunctionDefinition
			reduce(324), // ==, reduce: UnnamedFunctionDefinition
			reduce(324), // !=, reduce: UnnamedFunctionDefinition
			reduce(324), // <=>, reduce: UnnamedFunctionDefinition
			reduce(324), // >=, reduce: UnnamedFunctionDefinition
			reduce(324), // <, reduce: UnnamedFunctionDefinition
			reduce(324), // <=, reduce: UnnamedFunctionDefinition
			reduce(324), // ^, reduce: UnnamedFunctionDefinition
			reduce(324), // &, reduce: UnnamedFunctionDefinition
			reduce(324), // <<, reduce: UnnamedFunctionDefinition
			reduce(324), // >>>, reduce: UnnamedFunctionDefinition
			reduce(324), // +, reduce: UnnamedFunctionDefinition
			reduce(324), // -, reduce: UnnamedFunctionDefinition
			reduce(324), // .+, reduce: UnnamedFunctionDefinition
			reduce(324), // .-, reduce: UnnamedFunctionDefinition
			reduce(324), // *, reduce: UnnamedFunctionDefinition
			reduce(324), // /, reduce: UnnamedFunctionDefinition
			reduce(324), // //, reduce: UnnamedFunctionDefinition
			reduce(324), // %, reduce: UnnamedFunctionDefinition
			reduce(324), // .*, reduce: UnnamedFunctionDefinition
			reduce(324), // ./, reduce: UnnamedFunctionDefinition
			reduce(324), // .//, reduce: UnnamedFunctionDefinition
			reduce(324), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(324), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(324), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(324), // [[, reduce: UnnamedFunctionDefinition
			reduce(324), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4640
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			shift(952),  // ;
			shift(953),  // {
			reduce(2),   // }, reduce: StatementBlock
			nil,         // =
			shift(976),  // unset
			shift(978),  // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			shift(979),  // print
			nil,         // ,
			shift(980),  // printn
			shift(981),  // eprint
			shift(982),  // eprintn
			shift(983),  // dump
			shift(984),  // edump
			shift(985),  // tee
			shift(987),  // emitf
			shift(988),  // emit1
			shift(989),  // emit
			shift(990),  // (
			nil,         // )
			shift(991),  // emitp
			shift(1003), // field_name
			shift(1004), // $[
			nil,         // ]
			shift(1005), // braced_field_name
			shift(1006), // $[[
			shift(1007), // $[[[
			shift(1008), // full_srec
			shift(1010), // oosvar_name
			shift(1011), // @[
			shift(1012), // braced_oosvar_name
			shift(1013), // full_oosvar
			shift(1014), // all
			shift(1015), // non_sigil_name
			shift(71),   // arr
			shift(72),   // bool
			shift(73),   // float
			shift(74),   // int
			shift(75),   // map
			shift(76),   // num
			shift(77),   // str
			shift(78),   // var
			shift(79),   // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			shift(1030), // +
			shift(1032), // -
			shift(1033), // .+
			shift(1034), // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			shift(1037), // !
			shift(1038), // ~
			nil,         // **
			shift(1043), // string_literal
			shift(1044), // regex_case_insensitive
			shift(1045), // int_literal
			shift(1046), // float_literal
			shift(1047), // boolean_literal
			shift(1048), // null_literal
			shift(1049), // inf_literal
			shift(1050), // nan_literal
			shift(1051), // const_M_PI
			shift(1052), // const_M_E
			shift(1053), // panic
			shift(1055), // [
			shift(1057), // ctx_IPS
			shift(1058), // ctx_IFS
			shift(1059), // ctx_IRS
			shift(1060), // ctx_OPS
			shift(1061), // ctx_OFS
			shift(1062), // ctx_ORS
			shift(1063), // ctx_FLATSEP
			shift(1064), // ctx_NF
			shift(1065), // ctx_NR
			shift(1066), // ctx_FNR
			shift(1067), // ctx_FILENAME
			shift(1068), // ctx_FILENUM
			shift(1070), // env
			nil,         // [[
			nil,         // [[[
			shift(1077), // call
			shift(1086), // begin
			shift(1087), // end
			shift(1090), // if
			nil,         // elif
			nil,         // else
			shift(1091), // while
			shift(1092), // do
			shift(1097), // for
			nil,         // in
			shift(1098), // break
			shift(1099), // continue
			shift(1100), // func
			shift(1101), // subr
			shift(1102), // return
		},
	},
	actionRow{ // S4641
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			shift(4172), // arr
			shift(4173), // bool
			shift(4174), // float
			shift(4175), // int
			shift(4176), // map
			shift(4177), // num
			shift(4178), // str
			shift(4179), // var
			shift(4180), // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4642
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			nil,        // }
			reduce(13), // =, reduce: StatementBlockInBraces
			nil,        // unset
			nil,        // filter
			reduce(13), // >, reduce: StatementBlockInBraces
			reduce(13), // >>, reduce: StatementBlockInBraces
			reduce(13), // |, reduce: StatementBlockInBraces
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			reduce(13), // ||=, reduce: StatementBlockInBraces
			reduce(13), // ^^=, reduce: StatementBlockInBraces
			reduce(13), // &&=, reduce: StatementBlockInBraces
			reduce(13), // ??=, reduce: StatementBlockInBraces
			reduce(13), // ???=, reduce: StatementBlockInBraces
			reduce(13), // |=, reduce: StatementBlockInBraces
			reduce(13), // &=, reduce: StatementBlockInBraces
			reduce(13), // ^=, reduce: StatementBlockInBraces
			reduce(13), // <<=, reduce: StatementBlockInBraces
			reduce(13), // >>=, reduce: StatementBlockInBraces
			reduce(13), // >>>=, reduce: StatementBlockInBraces
			reduce(13), // +=, reduce: StatementBlockInBraces
			reduce(13), // .=, reduce: StatementBlockInBraces
			reduce(13), // -=, reduce: StatementBlockInBraces
			reduce(13), // *=, reduce: StatementBlockInBraces
			reduce(13), // /=, reduce: StatementBlockInBraces
			reduce(13), // //=, reduce: StatementBlockInBraces
			reduce(13), // %=, reduce: StatementBlockInBraces
			reduce(13), // **=, reduce: StatementBlockInBraces
			reduce(13), // ?, reduce: StatementBlockInBraces
			nil,        // :
			reduce(13), // ||, reduce: StatementBlockInBraces
			reduce(13), // ^^, reduce: StatementBlockInBraces
			reduce(13), // &&, reduce: StatementBlockInBraces
			reduce(13), // ??, reduce: StatementBlockInBraces
			reduce(13), // ???, reduce: StatementBlockInBraces
			reduce(13), // =~, reduce: StatementBlockInBraces
			reduce(13), // !=~, reduce: StatementBlockInBraces
			reduce(13), // ==, reduce: StatementBlockInBraces
			reduce(13), // !=, reduce: StatementBlockInBraces
			reduce(13), // <=>, reduce: StatementBlockInBraces
			reduce(13), // >=, reduce: StatementBlockInBraces
			reduce(13), // <, reduce: StatementBlockInBraces
			reduce(13), // <=, reduce: StatementBlockInBraces
			reduce(13), // ^, reduce: StatementBlockInBraces
			reduce(13), // &, reduce: StatementBlockInBraces
			reduce(13), // <<, reduce: StatementBlockInBraces
			reduce(13), // >>>, reduce: StatementBlockInBraces
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			reduce(13), // *, reduce: StatementBlockInBraces
			reduce(13), // /, reduce: StatementBlockInBraces
			reduce(13), // //, reduce: StatementBlockInBraces
			reduce(13), // %, reduce: StatementBlockInBraces
			reduce(13), // .*, reduce: StatementBlockInBraces
			reduce(13), // ./, reduce: StatementBlockInBraces
			reduce(13), // .//, reduce: StatementBlockInBraces
			reduce(13), // ., reduce: StatementBlockInBraces
			nil,        // !
			nil,        // ~
			reduce(13), // **, reduce: StatementBlockInBraces
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(13), // [, reduce: StatementBlockInBraces
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(13), // [[, reduce: StatementBlockInBraces
			reduce(13), // [[[, reduce: StatementBlockInBraces
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4643
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			reduce(325), // =, reduce: UnnamedFunctionDefinition
			nil,         // unset
			nil,         // filter
			reduce(325), // >, reduce: UnnamedFunctionDefinition
			reduce(325), // >>, reduce: UnnamedFunctionDefinition
			reduce(325), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(325), // ||=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^=, reduce: UnnamedFunctionDefinition
			reduce(325), // &&=, reduce: UnnamedFunctionDefinition
			reduce(325), // ??=, reduce: UnnamedFunctionDefinition
			reduce(325), // ???=, reduce: UnnamedFunctionDefinition
			reduce(325), // |=, reduce: UnnamedFunctionDefinition
			reduce(325), // &=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^=, reduce: UnnamedFunctionDefinition
			reduce(325), // <<=, reduce: UnnamedFunctionDefinition
			reduce(325), // >>=, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>=, reduce: UnnamedFunctionDefinition
			reduce(325), // +=, reduce: UnnamedFunctionDefinition
			reduce(325), // .=, reduce: UnnamedFunctionDefinition
			reduce(325), // -=, reduce: UnnamedFunctionDefinition
			reduce(325), // *=, reduce: UnnamedFunctionDefinition
			reduce(325), // /=, reduce: UnnamedFunctionDefinition
			reduce(325), // //=, reduce: UnnamedFunctionDefinition
			reduce(325), // %=, reduce: UnnamedFunctionDefinition
			reduce(325), // **=, reduce: UnnamedFunctionDefinition
			reduce(325), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(325), // ||, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^, reduce: UnnamedFunctionDefinition
			reduce(325), // &&, reduce: UnnamedFunctionDefinition
			reduce(325), // ??, reduce: UnnamedFunctionDefinition
			reduce(325), // ???, reduce: UnnamedFunctionDefinition
			reduce(325), // =~, reduce: UnnamedFunctionDefinition
			reduce(325), // !=~, reduce: UnnamedFunctionDefinition
			reduce(325), // ==, reduce: UnnamedFunctionDefinition
			reduce(325), // !=, reduce: UnnamedFunctionDefinition
			reduce(325), // <=>, reduce: UnnamedFunctionDefinition
			reduce(325), // >=, reduce: UnnamedFunctionDefinition
			reduce(325), // <, reduce: UnnamedFunctionDefinition
			reduce(325), // <=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^, reduce: UnnamedFunctionDefinition
			reduce(325), // &, reduce: UnnamedFunctionDefinition
			reduce(325), // <<, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>, reduce: UnnamedFunctionDefinition
			reduce(325), // +, reduce: UnnamedFunctionDefinition
			reduce(325), // -, reduce: UnnamedFunctionDefinition
			reduce(325), // .+, reduce: UnnamedFunctionDefinition
			reduce(325), // .-, reduce: UnnamedFunctionDefinition
			reduce(325), // *, reduce: UnnamedFunctionDefinition
			reduce(325), // /, reduce: UnnamedFunctionDefinition
			reduce(325), // //, reduce: UnnamedFunctionDefinition
			reduce(325), // %, reduce: UnnamedFunctionDefinition
			reduce(325), // .*, reduce: UnnamedFunctionDefinition
			reduce(325), // ./, reduce: UnnamedFunctionDefinition
			reduce(325), // .//, reduce: UnnamedFunctionDefinition
			reduce(325), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(325), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(325), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(325), // [[, reduce: UnnamedFunctionDefinition
			reduce(325), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4644
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			nil,        // {
			reduce(13), // }, reduce: StatementBlockInBraces
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(13), // >, reduce: StatementBlockInBraces
			reduce(13), // >>, reduce: StatementBlockInBraces
			reduce(13), // |, reduce: StatementBlockInBraces
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(13), // ,, reduce: StatementBlockInBraces
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(13), // ?, reduce: StatementBlockInBraces
			nil,        // :
			reduce(13), // ||, reduce: StatementBlockInBraces
			reduce(13), // ^^, reduce: StatementBlockInBraces
			reduce(13), // &&, reduce: StatementBlockInBraces
			reduce(13), // ??, reduce: StatementBlockInBraces
			reduce(13), // ???, reduce: StatementBlockInBraces
			reduce(13), // =~, reduce: StatementBlockInBraces
			reduce(13), // !=~, reduce: StatementBlockInBraces
			reduce(13), // ==, reduce: StatementBlockInBraces
			reduce(13), // !=, reduce: StatementBlockInBraces
			reduce(13), // <=>, reduce: StatementBlockInBraces
			reduce(13), // >=, reduce: StatementBlockInBraces
			reduce(13), // <, reduce: StatementBlockInBraces
			reduce(13), // <=, reduce: StatementBlockInBraces
			reduce(13), // ^, reduce: StatementBlockInBraces
			reduce(13), // &, reduce: StatementBlockInBraces
			reduce(13), // <<, reduce: StatementBlockInBraces
			reduce(13), // >>>, reduce: StatementBlockInBraces
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			reduce(13), // *, reduce: StatementBlockInBraces
			reduce(13), // /, reduce: StatementBlockInBraces
			reduce(13), // //, reduce: StatementBlockInBraces
			reduce(13), // %, reduce: StatementBlockInBraces
			reduce(13), // .*, reduce: StatementBlockInBraces
			reduce(13), // ./, reduce: StatementBlockInBraces
			reduce(13), // .//, reduce: StatementBlockInBraces
			reduce(13), // ., reduce: StatementBlockInBraces
			nil,        // !
			nil,        // ~
			reduce(13), // **, reduce: StatementBlockInBraces
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(13), // [, reduce: StatementBlockInBraces
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(13), // [[, reduce: StatementBlockInBraces
			reduce(13), // [[[, reduce: StatementBlockInBraces
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4645
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			reduce(325), // }, reduce: UnnamedFunctionDefinition
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(325), // >, reduce: UnnamedFunctionDefinition
			reduce(325), // >>, reduce: UnnamedFunctionDefinition
			reduce(325), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(325), // ,, reduce: UnnamedFunctionDefinition
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(325), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(325), // ||, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^, reduce: UnnamedFunctionDefinition
			reduce(325), // &&, reduce: UnnamedFunctionDefinition
			reduce(325), // ??, reduce: UnnamedFunctionDefinition
			reduce(325), // ???, reduce: UnnamedFunctionDefinition
			reduce(325), // =~, reduce: UnnamedFunctionDefinition
			reduce(325), // !=~, reduce: UnnamedFunctionDefinition
			reduce(325), // ==, reduce: UnnamedFunctionDefinition
			reduce(325), // !=, reduce: UnnamedFunctionDefinition
			reduce(325), // <=>, reduce: UnnamedFunctionDefinition
			reduce(325), // >=, reduce: UnnamedFunctionDefinition
			reduce(325), // <, reduce: UnnamedFunctionDefinition
			reduce(325), // <=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^, reduce: UnnamedFunctionDefinition
			reduce(325), // &, reduce: UnnamedFunctionDefinition
			reduce(325), // <<, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>, reduce: UnnamedFunctionDefinition
			reduce(325), // +, reduce: UnnamedFunctionDefinition
			reduce(325), // -, reduce: UnnamedFunctionDefinition
			reduce(325), // .+, reduce: UnnamedFunctionDefinition
			reduce(325), // .-, reduce: UnnamedFunctionDefinition
			reduce(325), // *, reduce: UnnamedFunctionDefinition
			reduce(325), // /, reduce: UnnamedFunctionDefinition
			reduce(325), // //, reduce: UnnamedFunctionDefinition
			reduce(325), // %, reduce: UnnamedFunctionDefinition
			reduce(325), // .*, reduce: UnnamedFunctionDefinition
			reduce(325), // ./, reduce: UnnamedFunctionDefinition
			reduce(325), // .//, reduce: UnnamedFunctionDefinition
			reduce(325), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(325), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(325), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(325), // [[, reduce: UnnamedFunctionDefinition
			reduce(325), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4646
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(13), // ;, reduce: StatementBlockInBraces
			nil,        // {
			reduce(13), // }, reduce: StatementBlockInBraces
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(13), // >, reduce: StatementBlockInBraces
			reduce(13), // >>, reduce: StatementBlockInBraces
			reduce(13), // |, reduce: StatementBlockInBraces
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(13), // ,, reduce: StatementBlockInBraces
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(13), // ?, reduce: StatementBlockInBraces
			nil,        // :
			reduce(13), // ||, reduce: StatementBlockInBraces
			reduce(13), // ^^, reduce: StatementBlockInBraces
			reduce(13), // &&, reduce: StatementBlockInBraces
			reduce(13), // ??, reduce: StatementBlockInBraces
			reduce(13), // ???, reduce: StatementBlockInBraces
			reduce(13), // =~, reduce: StatementBlockInBraces
			reduce(13), // !=~, reduce: StatementBlockInBraces
			reduce(13), // ==, reduce: StatementBlockInBraces
			reduce(13), // !=, reduce: StatementBlockInBraces
			reduce(13), // <=>, reduce: StatementBlockInBraces
			reduce(13), // >=, reduce: StatementBlockInBraces
			reduce(13), // <, reduce: StatementBlockInBraces
			reduce(13), // <=, reduce: StatementBlockInBraces
			reduce(13), // ^, reduce: StatementBlockInBraces
			reduce(13), // &, reduce: StatementBlockInBraces
			reduce(13), // <<, reduce: StatementBlockInBraces
			reduce(13), // >>>, reduce: StatementBlockInBraces
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			reduce(13), // *, reduce: StatementBlockInBraces
			reduce(13), // /, reduce: StatementBlockInBraces
			reduce(13), // //, reduce: StatementBlockInBraces
			reduce(13), // %, reduce: StatementBlockInBraces
			reduce(13), // .*, reduce: StatementBlockInBraces
			reduce(13), // ./, reduce: StatementBlockInBraces
			reduce(13), // .//, reduce: StatementBlockInBraces
			reduce(13), // ., reduce: StatementBlockInBraces
			nil,        // !
			nil,        // ~
			reduce(13), // **, reduce: StatementBlockInBraces
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(13), // [, reduce: StatementBlockInBraces
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(13), // [[, reduce: StatementBlockInBraces
			reduce(13), // [[[, reduce: StatementBlockInBraces
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4647
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(325), // ;, reduce: UnnamedFunctionDefinition
			nil,         // {
			reduce(325), // }, reduce: UnnamedFunctionDefinition
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(325), // >, reduce: UnnamedFunctionDefinition
			reduce(325), // >>, reduce: UnnamedFunctionDefinition
			reduce(325), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(325), // ,, reduce: UnnamedFunctionDefinition
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(325), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(325), // ||, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^, reduce: UnnamedFunctionDefinition
			reduce(325), // &&, reduce: UnnamedFunctionDefinition
			reduce(325), // ??, reduce: UnnamedFunctionDefinition
			reduce(325), // ???, reduce: UnnamedFunctionDefinition
			reduce(325), // =~, reduce: UnnamedFunctionDefinition
			reduce(325), // !=~, reduce: UnnamedFunctionDefinition
			reduce(325), // ==, reduce: UnnamedFunctionDefinition
			reduce(325), // !=, reduce: UnnamedFunctionDefinition
			reduce(325), // <=>, reduce: UnnamedFunctionDefinition
			reduce(325), // >=, reduce: UnnamedFunctionDefinition
			reduce(325), // <, reduce: UnnamedFunctionDefinition
			reduce(325), // <=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^, reduce: UnnamedFunctionDefinition
			reduce(325), // &, reduce: UnnamedFunctionDefinition
			reduce(325), // <<, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>, reduce: UnnamedFunctionDefinition
			reduce(325), // +, reduce: UnnamedFunctionDefinition
			reduce(325), // -, reduce: UnnamedFunctionDefinition
			reduce(325), // .+, reduce: UnnamedFunctionDefinition
			reduce(325), // .-, reduce: UnnamedFunctionDefinition
			reduce(325), // *, reduce: UnnamedFunctionDefinition
			reduce(325), // /, reduce: UnnamedFunctionDefinition
			reduce(325), // //, reduce: UnnamedFunctionDefinition
			reduce(325), // %, reduce: UnnamedFunctionDefinition
			reduce(325), // .*, reduce: UnnamedFunctionDefinition
			reduce(325), // ./, reduce: UnnamedFunctionDefinition
			reduce(325), // .//, reduce: UnnamedFunctionDefinition
			reduce(325), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(325), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(325), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(325), // [[, reduce: UnnamedFunctionDefinition
			reduce(325), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4648
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(13), // ;, reduce: StatementBlockInBraces
			nil,        // {
			reduce(13), // }, reduce: StatementBlockInBraces
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(13), // >, reduce: StatementBlockInBraces
			reduce(13), // >>, reduce: StatementBlockInBraces
			reduce(13), // |, reduce: StatementBlockInBraces
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(13), // ?, reduce: StatementBlockInBraces
			nil,        // :
			reduce(13), // ||, reduce: StatementBlockInBraces
			reduce(13), // ^^, reduce: StatementBlockInBraces
			reduce(13), // &&, reduce: StatementBlockInBraces
			reduce(13), // ??, reduce: StatementBlockInBraces
			reduce(13), // ???, reduce: StatementBlockInBraces
			reduce(13), // =~, reduce: StatementBlockInBraces
			reduce(13), // !=~, reduce: StatementBlockInBraces
			reduce(13), // ==, reduce: StatementBlockInBraces
			reduce(13), // !=, reduce: StatementBlockInBraces
			reduce(13), // <=>, reduce: StatementBlockInBraces
			reduce(13), // >=, reduce: StatementBlockInBraces
			reduce(13), // <, reduce: StatementBlockInBraces
			reduce(13), // <=, reduce: StatementBlockInBraces
			reduce(13), // ^, reduce: StatementBlockInBraces
			reduce(13), // &, reduce: StatementBlockInBraces
			reduce(13), // <<, reduce: StatementBlockInBraces
			reduce(13), // >>>, reduce: StatementBlockInBraces
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			reduce(13), // *, reduce: StatementBlockInBraces
			reduce(13), // /, reduce: StatementBlockInBraces
			reduce(13), // //, reduce: StatementBlockInBraces
			reduce(13), // %, reduce: StatementBlockInBraces
			reduce(13), // .*, reduce: StatementBlockInBraces
			reduce(13), // ./, reduce: StatementBlockInBraces
			reduce(13), // .//, reduce: StatementBlockInBraces
			reduce(13), // ., reduce: StatementBlockInBraces
			nil,        // !
			nil,        // ~
			reduce(13), // **, reduce: StatementBlockInBraces
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(13), // [, reduce: StatementBlockInBraces
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(13), // [[, reduce: StatementBlockInBraces
			reduce(13), // [[[, reduce: StatementBlockInBraces
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4649
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(325), // ;, reduce: UnnamedFunctionDefinition
			nil,         // {
			reduce(325), // }, reduce: UnnamedFunctionDefinition
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(325), // >, reduce: UnnamedFunctionDefinition
			reduce(325), // >>, reduce: UnnamedFunctionDefinition
			reduce(325), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(325), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(325), // ||, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^, reduce: UnnamedFunctionDefinition
			reduce(325), // &&, reduce: UnnamedFunctionDefinition
			reduce(325), // ??, reduce: UnnamedFunctionDefinition
			reduce(325), // ???, reduce: UnnamedFunctionDefinition
			reduce(325), // =~, reduce: UnnamedFunctionDefinition
			reduce(325), // !=~, reduce: UnnamedFunctionDefinition
			reduce(325), // ==, reduce: UnnamedFunctionDefinition
			reduce(325), // !=, reduce: UnnamedFunctionDefinition
			reduce(325), // <=>, reduce: UnnamedFunctionDefinition
			reduce(325), // >=, reduce: UnnamedFunctionDefinition
			reduce(325), // <, reduce: UnnamedFunctionDefinition
			reduce(325), // <=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^, reduce: UnnamedFunctionDefinition
			reduce(325), // &, reduce: UnnamedFunctionDefinition
			reduce(325), // <<, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>, reduce: UnnamedFunctionDefinition
			reduce(325), // +, reduce: UnnamedFunctionDefinition
			reduce(325), // -, reduce: UnnamedFunctionDefinition
			reduce(325), // .+, reduce: UnnamedFunctionDefinition
			reduce(325), // .-, reduce: UnnamedFunctionDefinition
			reduce(325), // *, reduce: UnnamedFunctionDefinition
			reduce(325), // /, reduce: UnnamedFunctionDefinition
			reduce(325), // //, reduce: UnnamedFunctionDefinition
			reduce(325), // %, reduce: UnnamedFunctionDefinition
			reduce(325), // .*, reduce: UnnamedFunctionDefinition
			reduce(325), // ./, reduce: UnnamedFunctionDefinition
			reduce(325), // .//, reduce: UnnamedFunctionDefinition
			reduce(325), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(325), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(325), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(325), // [[, reduce: UnnamedFunctionDefinition
			reduce(325), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4650
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(75), // ;, reduce: EmitStatement
			nil,        // {
			reduce(75), // }, reduce: EmitStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4651
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(83), // ;, reduce: EmitPStatement
			nil,        // {
			reduce(83), // }, reduce: EmitPStatement
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4652
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2042), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4653
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(520), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			shift(524), // (
			nil,        // )
			nil,        // emitp
			shift(536), // field_name
			shift(537), // $[
			nil,        // ]
			shift(538), // braced_field_name
			shift(539), // $[[
			shift(540), // $[[[
			shift(541), // full_srec
			shift(543), // oosvar_name
			shift(544), // @[
			shift(545), // braced_oosvar_name
			shift(546), // full_oosvar
			shift(547), // all
			shift(548), // non_sigil_name
			nil,        // arr
			nil,        // bool
			shift(202), // float
			shift(203), // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			shift(563), // +
			shift(565), // -
			shift(566), // .+
			shift(567), // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			shift(570), // !
			shift(571), // ~
			nil,        // **
			shift(576), // string_literal
			shift(577), // regex_case_insensitive
			shift(578), // int_literal
			shift(579), // float_literal
			shift(580), // boolean_literal
			shift(581), // null_literal
			shift(582), // inf_literal
			shift(583), // nan_literal
			shift(584), // const_M_PI
			shift(585), // const_M_E
			shift(586), // panic
			shift(588), // [
			shift(590), // ctx_IPS
			shift(591), // ctx_IFS
			shift(592), // ctx_IRS
			shift(593), // ctx_OPS
			shift(594), // ctx_OFS
			shift(595), // ctx_ORS
			shift(596), // ctx_FLATSEP
			shift(597), // ctx_NF
			shift(598), // ctx_NR
			shift(599), // ctx_FNR
			shift(600), // ctx_FILENAME
			shift(601), // ctx_FILENUM
			shift(603), // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			shift(610), // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4654
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2042), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4655
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(323), // ;, reduce: NamedFunctionDefinition
			reduce(323), // {, reduce: NamedFunctionDefinition
			reduce(323), // }, reduce: NamedFunctionDefinition
			nil,         // =
			reduce(323), // unset, reduce: NamedFunctionDefinition
			reduce(323), // filter, reduce: NamedFunctionDefinition
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(323), // print, reduce: NamedFunctionDefinition
			nil,         // ,
			reduce(323), // printn, reduce: NamedFunctionDefinition
			reduce(323), // eprint, reduce: NamedFunctionDefinition
			reduce(323), // eprintn, reduce: NamedFunctionDefinition
			reduce(323), // dump, reduce: NamedFunctionDefinition
			reduce(323), // edump, reduce: NamedFunctionDefinition
			reduce(323), // tee, reduce: NamedFunctionDefinition
			reduce(323), // emitf, reduce: NamedFunctionDefinition
			reduce(323), // emit1, reduce: NamedFunctionDefinition
			reduce(323), // emit, reduce: NamedFunctionDefinition
			reduce(323), // (, reduce: NamedFunctionDefinition
			nil,         // )
			reduce(323), // emitp, reduce: NamedFunctionDefinition
			reduce(323), // field_name, reduce: NamedFunctionDefinition
			reduce(323), // $[, reduce: NamedFunctionDefinition
			nil,         // ]
			reduce(323), // braced_field_name, reduce: NamedFunctionDefinition
			reduce(323), // $[[, reduce: NamedFunctionDefinition
			reduce(323), // $[[[, reduce: NamedFunctionDefinition
			reduce(323), // full_srec, reduce: NamedFunctionDefinition
			reduce(323), // oosvar_name, reduce: NamedFunctionDefinition
			reduce(323), // @[, reduce: NamedFunctionDefinition
			reduce(323), // braced_oosvar_name, reduce: NamedFunctionDefinition
			reduce(323), // full_oosvar, reduce: NamedFunctionDefinition
			reduce(323), // all, reduce: NamedFunctionDefinition
			reduce(323), // non_sigil_name, reduce: NamedFunctionDefinition
			reduce(323), // arr, reduce: NamedFunctionDefinition
			reduce(323), // bool, reduce: NamedFunctionDefinition
			reduce(323), // float, reduce: NamedFunctionDefinition
			reduce(323), // int, reduce: NamedFunctionDefinition
			reduce(323), // map, reduce: NamedFunctionDefinition
			reduce(323), // num, reduce: NamedFunctionDefinition
			reduce(323), // str, reduce: NamedFunctionDefinition
			reduce(323), // var, reduce: NamedFunctionDefinition
			reduce(323), // funct, reduce: NamedFunctionDefinition
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(323), // +, reduce: NamedFunctionDefinition
			reduce(323), // -, reduce: NamedFunctionDefinition
			reduce(323), // .+, reduce: NamedFunctionDefinition
			reduce(323), // .-, reduce: NamedFunctionDefinition
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(323), // !, reduce: NamedFunctionDefinition
			reduce(323), // ~, reduce: NamedFunctionDefinition
			nil,         // **
			reduce(323), // string_literal, reduce: NamedFunctionDefinition
			reduce(323), // regex_case_insensitive, reduce: NamedFunctionDefinition
			reduce(323), // int_literal, reduce: NamedFunctionDefinition
			reduce(323), // float_literal, reduce: NamedFunctionDefinition
			reduce(323), // boolean_literal, reduce: NamedFunctionDefinition
			reduce(323), // null_literal, reduce: NamedFunctionDefinition
			reduce(323), // inf_literal, reduce: NamedFunctionDefinition
			reduce(323), // nan_literal, reduce: NamedFunctionDefinition
			reduce(323), // const_M_PI, reduce: NamedFunctionDefinition
			reduce(323), // const_M_E, reduce: NamedFunctionDefinition
			reduce(323), // panic, reduce: NamedFunctionDefinition
			reduce(323), // [, reduce: NamedFunctionDefinition
			reduce(323), // ctx_IPS, reduce: NamedFunctionDefinition
			reduce(323), // ctx_IFS, reduce: NamedFunctionDefinition
			reduce(323), // ctx_IRS, reduce: NamedFunctionDefinition
			reduce(323), // ctx_OPS, reduce: NamedFunctionDefinition
			reduce(323), // ctx_OFS, reduce: NamedFunctionDefinition
			reduce(323), // ctx_ORS, reduce: NamedFunctionDefinition
			reduce(323), // ctx_FLATSEP, reduce: NamedFunctionDefinition
			reduce(323), // ctx_NF, reduce: NamedFunctionDefinition
			reduce(323), // ctx_NR, reduce: NamedFunctionDefinition
			reduce(323), // ctx_FNR, reduce: NamedFunctionDefinition
			reduce(323), // ctx_FILENAME, reduce: NamedFunctionDefinition
			reduce(323), // ctx_FILENUM, reduce: NamedFunctionDefinition
			reduce(323), // env, reduce: NamedFunctionDefinition
			nil,         // [[
			nil,         // [[[
			reduce(323), // call, reduce: NamedFunctionDefinition
			reduce(323), // begin, reduce: NamedFunctionDefinition
			reduce(323), // end, reduce: NamedFunctionDefinition
			reduce(323), // if, reduce: NamedFunctionDefinition
			nil,         // elif
			nil,         // else
			reduce(323), // while, reduce: NamedFunctionDefinition
			reduce(323), // do, reduce: NamedFunctionDefinition
			reduce(323), // for, reduce: NamedFunctionDefinition
			nil,         // in
			reduce(323), // break, reduce: NamedFunctionDefinition
			reduce(323), // continue, reduce: NamedFunctionDefinition
			reduce(323), // func, reduce: NamedFunctionDefinition
			reduce(323), // subr, reduce: NamedFunctionDefinition
			reduce(323), // return, reduce: NamedFunctionDefinition
		},
	},
	actionRow{ // S4656
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(263), // ;, reduce: ArraySliceAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(263), // >, reduce: ArraySliceAccess
			reduce(263), // >>, reduce: ArraySliceAccess
			reduce(263), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(263), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(263), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(263), // ||, reduce: ArraySliceAccess
			reduce(263), // ^^, reduce: ArraySliceAccess
			reduce(263), // &&, reduce: ArraySliceAccess
			reduce(263), // ??, reduce: ArraySliceAccess
			reduce(263), // ???, reduce: ArraySliceAccess
			reduce(263), // =~, reduce: ArraySliceAccess
			reduce(263), // !=~, reduce: ArraySliceAccess
			reduce(263), // ==, reduce: ArraySliceAccess
			reduce(263), // !=, reduce: ArraySliceAccess
			reduce(263), // <=>, reduce: ArraySliceAccess
			reduce(263), // >=, reduce: ArraySliceAccess
			reduce(263), // <, reduce: ArraySliceAccess
			reduce(263), // <=, reduce: ArraySliceAccess
			reduce(263), // ^, reduce: ArraySliceAccess
			reduce(263), // &, reduce: ArraySliceAccess
			reduce(263), // <<, reduce: ArraySliceAccess
			reduce(263), // >>>, reduce: ArraySliceAccess
			reduce(263), // +, reduce: ArraySliceAccess
			reduce(263), // -, reduce: ArraySliceAccess
			reduce(263), // .+, reduce: ArraySliceAccess
			reduce(263), // .-, reduce: ArraySliceAccess
			reduce(263), // *, reduce: ArraySliceAccess
			reduce(263), // /, reduce: ArraySliceAccess
			reduce(263), // //, reduce: ArraySliceAccess
			reduce(263), // %, reduce: ArraySliceAccess
			reduce(263), // .*, reduce: ArraySliceAccess
			reduce(263), // ./, reduce: ArraySliceAccess
			reduce(263), // .//, reduce: ArraySliceAccess
			reduce(263), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(263), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(263), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(263), // [[, reduce: ArraySliceAccess
			reduce(263), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4657
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(262), // ;, reduce: ArrayOrMapPositionalValueAccess
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(262), // >, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |, reduce: ArrayOrMapPositionalValueAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(262), // ,, reduce: ArrayOrMapPositionalValueAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(262), // ?, reduce: ArrayOrMapPositionalValueAccess
			nil,         // :
			reduce(262), // ||, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // =~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ==, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .+, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .-, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .*, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ./, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .//, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ., reduce: ArrayOrMapPositionalValueAccess
			nil,         // !
			nil,         // ~
			reduce(262), // **, reduce: ArrayOrMapPositionalValueAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(262), // [, reduce: ArrayOrMapPositionalValueAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(262), // [[, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // [[[, reduce: ArrayOrMapPositionalValueAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4658
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4668), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4659
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4606), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4660
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			nil,        // ;
			shift(289), // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			nil,        // >
			nil,        // >>
			nil,        // |
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			nil,        // ,
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			nil,        // ?
			nil,        // :
			nil,        // ||
			nil,        // ^^
			nil,        // &&
			nil,        // ??
			nil,        // ???
			nil,        // =~
			nil,        // !=~
			nil,        // ==
			nil,        // !=
			nil,        // <=>
			nil,        // >=
			nil,        // <
			nil,        // <=
			nil,        // ^
			nil,        // &
			nil,        // <<
			nil,        // >>>
			nil,        // +
			nil,        // -
			nil,        // .+
			nil,        // .-
			nil,        // *
			nil,        // /
			nil,        // //
			nil,        // %
			nil,        // .*
			nil,        // ./
			nil,        // .//
			nil,        // .
			nil,        // !
			nil,        // ~
			nil,        // **
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			nil,        // [
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			nil,        // [[
			nil,        // [[[
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4661
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(263), // ;, reduce: ArraySliceAccess
			nil,         // {
			nil,         // }
			reduce(263), // =, reduce: ArraySliceAccess
			nil,         // unset
			nil,         // filter
			reduce(263), // >, reduce: ArraySliceAccess
			reduce(263), // >>, reduce: ArraySliceAccess
			reduce(263), // |, reduce: ArraySliceAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(263), // ,, reduce: ArraySliceAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(263), // ||=, reduce: ArraySliceAccess
			reduce(263), // ^^=, reduce: ArraySliceAccess
			reduce(263), // &&=, reduce: ArraySliceAccess
			reduce(263), // ??=, reduce: ArraySliceAccess
			reduce(263), // ???=, reduce: ArraySliceAccess
			reduce(263), // |=, reduce: ArraySliceAccess
			reduce(263), // &=, reduce: ArraySliceAccess
			reduce(263), // ^=, reduce: ArraySliceAccess
			reduce(263), // <<=, reduce: ArraySliceAccess
			reduce(263), // >>=, reduce: ArraySliceAccess
			reduce(263), // >>>=, reduce: ArraySliceAccess
			reduce(263), // +=, reduce: ArraySliceAccess
			reduce(263), // .=, reduce: ArraySliceAccess
			reduce(263), // -=, reduce: ArraySliceAccess
			reduce(263), // *=, reduce: ArraySliceAccess
			reduce(263), // /=, reduce: ArraySliceAccess
			reduce(263), // //=, reduce: ArraySliceAccess
			reduce(263), // %=, reduce: ArraySliceAccess
			reduce(263), // **=, reduce: ArraySliceAccess
			reduce(263), // ?, reduce: ArraySliceAccess
			nil,         // :
			reduce(263), // ||, reduce: ArraySliceAccess
			reduce(263), // ^^, reduce: ArraySliceAccess
			reduce(263), // &&, reduce: ArraySliceAccess
			reduce(263), // ??, reduce: ArraySliceAccess
			reduce(263), // ???, reduce: ArraySliceAccess
			reduce(263), // =~, reduce: ArraySliceAccess
			reduce(263), // !=~, reduce: ArraySliceAccess
			reduce(263), // ==, reduce: ArraySliceAccess
			reduce(263), // !=, reduce: ArraySliceAccess
			reduce(263), // <=>, reduce: ArraySliceAccess
			reduce(263), // >=, reduce: ArraySliceAccess
			reduce(263), // <, reduce: ArraySliceAccess
			reduce(263), // <=, reduce: ArraySliceAccess
			reduce(263), // ^, reduce: ArraySliceAccess
			reduce(263), // &, reduce: ArraySliceAccess
			reduce(263), // <<, reduce: ArraySliceAccess
			reduce(263), // >>>, reduce: ArraySliceAccess
			reduce(263), // +, reduce: ArraySliceAccess
			reduce(263), // -, reduce: ArraySliceAccess
			reduce(263), // .+, reduce: ArraySliceAccess
			reduce(263), // .-, reduce: ArraySliceAccess
			reduce(263), // *, reduce: ArraySliceAccess
			reduce(263), // /, reduce: ArraySliceAccess
			reduce(263), // //, reduce: ArraySliceAccess
			reduce(263), // %, reduce: ArraySliceAccess
			reduce(263), // .*, reduce: ArraySliceAccess
			reduce(263), // ./, reduce: ArraySliceAccess
			reduce(263), // .//, reduce: ArraySliceAccess
			reduce(263), // ., reduce: ArraySliceAccess
			nil,         // !
			nil,         // ~
			reduce(263), // **, reduce: ArraySliceAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(263), // [, reduce: ArraySliceAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(263), // [[, reduce: ArraySliceAccess
			reduce(263), // [[[, reduce: ArraySliceAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4662
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(262), // ;, reduce: ArrayOrMapPositionalValueAccess
			nil,         // {
			nil,         // }
			reduce(262), // =, reduce: ArrayOrMapPositionalValueAccess
			nil,         // unset
			nil,         // filter
			reduce(262), // >, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |, reduce: ArrayOrMapPositionalValueAccess
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(262), // ,, reduce: ArrayOrMapPositionalValueAccess
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(262), // ||=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // |=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // **=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ?, reduce: ArrayOrMapPositionalValueAccess
			nil,         // :
			reduce(262), // ||, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &&, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ??, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ???, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // =~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=~, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ==, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // !=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <=, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ^, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // &, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // <<, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // >>>, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // +, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // -, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .+, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .-, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // *, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // /, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // //, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // %, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .*, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ./, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // .//, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // ., reduce: ArrayOrMapPositionalValueAccess
			nil,         // !
			nil,         // ~
			reduce(262), // **, reduce: ArrayOrMapPositionalValueAccess
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(262), // [, reduce: ArrayOrMapPositionalValueAccess
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(262), // [[, reduce: ArrayOrMapPositionalValueAccess
			reduce(262), // [[[, reduce: ArrayOrMapPositionalValueAccess
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4663
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			shift(4671), // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4664
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(4640), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4665
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(304), // ;, reduce: ForLoopTwoVariable
			reduce(304), // {, reduce: ForLoopTwoVariable
			reduce(304), // }, reduce: ForLoopTwoVariable
			nil,         // =
			reduce(304), // unset, reduce: ForLoopTwoVariable
			reduce(304), // filter, reduce: ForLoopTwoVariable
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(304), // print, reduce: ForLoopTwoVariable
			nil,         // ,
			reduce(304), // printn, reduce: ForLoopTwoVariable
			reduce(304), // eprint, reduce: ForLoopTwoVariable
			reduce(304), // eprintn, reduce: ForLoopTwoVariable
			reduce(304), // dump, reduce: ForLoopTwoVariable
			reduce(304), // edump, reduce: ForLoopTwoVariable
			reduce(304), // tee, reduce: ForLoopTwoVariable
			reduce(304), // emitf, reduce: ForLoopTwoVariable
			reduce(304), // emit1, reduce: ForLoopTwoVariable
			reduce(304), // emit, reduce: ForLoopTwoVariable
			reduce(304), // (, reduce: ForLoopTwoVariable
			nil,         // )
			reduce(304), // emitp, reduce: ForLoopTwoVariable
			reduce(304), // field_name, reduce: ForLoopTwoVariable
			reduce(304), // $[, reduce: ForLoopTwoVariable
			nil,         // ]
			reduce(304), // braced_field_name, reduce: ForLoopTwoVariable
			reduce(304), // $[[, reduce: ForLoopTwoVariable
			reduce(304), // $[[[, reduce: ForLoopTwoVariable
			reduce(304), // full_srec, reduce: ForLoopTwoVariable
			reduce(304), // oosvar_name, reduce: ForLoopTwoVariable
			reduce(304), // @[, reduce: ForLoopTwoVariable
			reduce(304), // braced_oosvar_name, reduce: ForLoopTwoVariable
			reduce(304), // full_oosvar, reduce: ForLoopTwoVariable
			reduce(304), // all, reduce: ForLoopTwoVariable
			reduce(304), // non_sigil_name, reduce: ForLoopTwoVariable
			reduce(304), // arr, reduce: ForLoopTwoVariable
			reduce(304), // bool, reduce: ForLoopTwoVariable
			reduce(304), // float, reduce: ForLoopTwoVariable
			reduce(304), // int, reduce: ForLoopTwoVariable
			reduce(304), // map, reduce: ForLoopTwoVariable
			reduce(304), // num, reduce: ForLoopTwoVariable
			reduce(304), // str, reduce: ForLoopTwoVariable
			reduce(304), // var, reduce: ForLoopTwoVariable
			reduce(304), // funct, reduce: ForLoopTwoVariable
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(304), // +, reduce: ForLoopTwoVariable
			reduce(304), // -, reduce: ForLoopTwoVariable
			reduce(304), // .+, reduce: ForLoopTwoVariable
			reduce(304), // .-, reduce: ForLoopTwoVariable
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(304), // !, reduce: ForLoopTwoVariable
			reduce(304), // ~, reduce: ForLoopTwoVariable
			nil,         // **
			reduce(304), // string_literal, reduce: ForLoopTwoVariable
			reduce(304), // regex_case_insensitive, reduce: ForLoopTwoVariable
			reduce(304), // int_literal, reduce: ForLoopTwoVariable
			reduce(304), // float_literal, reduce: ForLoopTwoVariable
			reduce(304), // boolean_literal, reduce: ForLoopTwoVariable
			reduce(304), // null_literal, reduce: ForLoopTwoVariable
			reduce(304), // inf_literal, reduce: ForLoopTwoVariable
			reduce(304), // nan_literal, reduce: ForLoopTwoVariable
			reduce(304), // const_M_PI, reduce: ForLoopTwoVariable
			reduce(304), // const_M_E, reduce: ForLoopTwoVariable
			reduce(304), // panic, reduce: ForLoopTwoVariable
			reduce(304), // [, reduce: ForLoopTwoVariable
			reduce(304), // ctx_IPS, reduce: ForLoopTwoVariable
			reduce(304), // ctx_IFS, reduce: ForLoopTwoVariable
			reduce(304), // ctx_IRS, reduce: ForLoopTwoVariable
			reduce(304), // ctx_OPS, reduce: ForLoopTwoVariable
			reduce(304), // ctx_OFS, reduce: ForLoopTwoVariable
			reduce(304), // ctx_ORS, reduce: ForLoopTwoVariable
			reduce(304), // ctx_FLATSEP, reduce: ForLoopTwoVariable
			reduce(304), // ctx_NF, reduce: ForLoopTwoVariable
			reduce(304), // ctx_NR, reduce: ForLoopTwoVariable
			reduce(304), // ctx_FNR, reduce: ForLoopTwoVariable
			reduce(304), // ctx_FILENAME, reduce: ForLoopTwoVariable
			reduce(304), // ctx_FILENUM, reduce: ForLoopTwoVariable
			reduce(304), // env, reduce: ForLoopTwoVariable
			nil,         // [[
			nil,         // [[[
			reduce(304), // call, reduce: ForLoopTwoVariable
			reduce(304), // begin, reduce: ForLoopTwoVariable
			reduce(304), // end, reduce: ForLoopTwoVariable
			reduce(304), // if, reduce: ForLoopTwoVariable
			nil,         // elif
			nil,         // else
			reduce(304), // while, reduce: ForLoopTwoVariable
			reduce(304), // do, reduce: ForLoopTwoVariable
			reduce(304), // for, reduce: ForLoopTwoVariable
			nil,         // in
			reduce(304), // break, reduce: ForLoopTwoVariable
			reduce(304), // continue, reduce: ForLoopTwoVariable
			reduce(304), // func, reduce: ForLoopTwoVariable
			reduce(304), // subr, reduce: ForLoopTwoVariable
			reduce(304), // return, reduce: ForLoopTwoVariable
		},
	},
	actionRow{ // S4666
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			shift(4673), // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4667
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(308), // ;, reduce: TripleForLoop
			reduce(308), // {, reduce: TripleForLoop
			reduce(308), // }, reduce: TripleForLoop
			nil,         // =
			reduce(308), // unset, reduce: TripleForLoop
			reduce(308), // filter, reduce: TripleForLoop
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(308), // print, reduce: TripleForLoop
			nil,         // ,
			reduce(308), // printn, reduce: TripleForLoop
			reduce(308), // eprint, reduce: TripleForLoop
			reduce(308), // eprintn, reduce: TripleForLoop
			reduce(308), // dump, reduce: TripleForLoop
			reduce(308), // edump, reduce: TripleForLoop
			reduce(308), // tee, reduce: TripleForLoop
			reduce(308), // emitf, reduce: TripleForLoop
			reduce(308), // emit1, reduce: TripleForLoop
			reduce(308), // emit, reduce: TripleForLoop
			reduce(308), // (, reduce: TripleForLoop
			nil,         // )
			reduce(308), // emitp, reduce: TripleForLoop
			reduce(308), // field_name, reduce: TripleForLoop
			reduce(308), // $[, reduce: TripleForLoop
			nil,         // ]
			reduce(308), // braced_field_name, reduce: TripleForLoop
			reduce(308), // $[[, reduce: TripleForLoop
			reduce(308), // $[[[, reduce: TripleForLoop
			reduce(308), // full_srec, reduce: TripleForLoop
			reduce(308), // oosvar_name, reduce: TripleForLoop
			reduce(308), // @[, reduce: TripleForLoop
			reduce(308), // braced_oosvar_name, reduce: TripleForLoop
			reduce(308), // full_oosvar, reduce: TripleForLoop
			reduce(308), // all, reduce: TripleForLoop
			reduce(308), // non_sigil_name, reduce: TripleForLoop
			reduce(308), // arr, reduce: TripleForLoop
			reduce(308), // bool, reduce: TripleForLoop
			reduce(308), // float, reduce: TripleForLoop
			reduce(308), // int, reduce: TripleForLoop
			reduce(308), // map, reduce: TripleForLoop
			reduce(308), // num, reduce: TripleForLoop
			reduce(308), // str, reduce: TripleForLoop
			reduce(308), // var, reduce: TripleForLoop
			reduce(308), // funct, reduce: TripleForLoop
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(308), // +, reduce: TripleForLoop
			reduce(308), // -, reduce: TripleForLoop
			reduce(308), // .+, reduce: TripleForLoop
			reduce(308), // .-, reduce: TripleForLoop
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(308), // !, reduce: TripleForLoop
			reduce(308), // ~, reduce: TripleForLoop
			nil,         // **
			reduce(308), // string_literal, reduce: TripleForLoop
			reduce(308), // regex_case_insensitive, reduce: TripleForLoop
			reduce(308), // int_literal, reduce: TripleForLoop
			reduce(308), // float_literal, reduce: TripleForLoop
			reduce(308), // boolean_literal, reduce: TripleForLoop
			reduce(308), // null_literal, reduce: TripleForLoop
			reduce(308), // inf_literal, reduce: TripleForLoop
			reduce(308), // nan_literal, reduce: TripleForLoop
			reduce(308), // const_M_PI, reduce: TripleForLoop
			reduce(308), // const_M_E, reduce: TripleForLoop
			reduce(308), // panic, reduce: TripleForLoop
			reduce(308), // [, reduce: TripleForLoop
			reduce(308), // ctx_IPS, reduce: TripleForLoop
			reduce(308), // ctx_IFS, reduce: TripleForLoop
			reduce(308), // ctx_IRS, reduce: TripleForLoop
			reduce(308), // ctx_OPS, reduce: TripleForLoop
			reduce(308), // ctx_OFS, reduce: TripleForLoop
			reduce(308), // ctx_ORS, reduce: TripleForLoop
			reduce(308), // ctx_FLATSEP, reduce: TripleForLoop
			reduce(308), // ctx_NF, reduce: TripleForLoop
			reduce(308), // ctx_NR, reduce: TripleForLoop
			reduce(308), // ctx_FNR, reduce: TripleForLoop
			reduce(308), // ctx_FILENAME, reduce: TripleForLoop
			reduce(308), // ctx_FILENUM, reduce: TripleForLoop
			reduce(308), // env, reduce: TripleForLoop
			nil,         // [[
			nil,         // [[[
			reduce(308), // call, reduce: TripleForLoop
			reduce(308), // begin, reduce: TripleForLoop
			reduce(308), // end, reduce: TripleForLoop
			reduce(308), // if, reduce: TripleForLoop
			nil,         // elif
			nil,         // else
			reduce(308), // while, reduce: TripleForLoop
			reduce(308), // do, reduce: TripleForLoop
			reduce(308), // for, reduce: TripleForLoop
			nil,         // in
			reduce(308), // break, reduce: TripleForLoop
			reduce(308), // continue, reduce: TripleForLoop
			reduce(308), // func, reduce: TripleForLoop
			reduce(308), // subr, reduce: TripleForLoop
			reduce(308), // return, reduce: TripleForLoop
		},
	},
	actionRow{ // S4668
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(13), // ;, reduce: StatementBlockInBraces
			nil,        // {
			nil,        // }
			nil,        // =
			nil,        // unset
			nil,        // filter
			reduce(13), // >, reduce: StatementBlockInBraces
			reduce(13), // >>, reduce: StatementBlockInBraces
			reduce(13), // |, reduce: StatementBlockInBraces
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(13), // ,, reduce: StatementBlockInBraces
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			nil,        // ||=
			nil,        // ^^=
			nil,        // &&=
			nil,        // ??=
			nil,        // ???=
			nil,        // |=
			nil,        // &=
			nil,        // ^=
			nil,        // <<=
			nil,        // >>=
			nil,        // >>>=
			nil,        // +=
			nil,        // .=
			nil,        // -=
			nil,        // *=
			nil,        // /=
			nil,        // //=
			nil,        // %=
			nil,        // **=
			reduce(13), // ?, reduce: StatementBlockInBraces
			nil,        // :
			reduce(13), // ||, reduce: StatementBlockInBraces
			reduce(13), // ^^, reduce: StatementBlockInBraces
			reduce(13), // &&, reduce: StatementBlockInBraces
			reduce(13), // ??, reduce: StatementBlockInBraces
			reduce(13), // ???, reduce: StatementBlockInBraces
			reduce(13), // =~, reduce: StatementBlockInBraces
			reduce(13), // !=~, reduce: StatementBlockInBraces
			reduce(13), // ==, reduce: StatementBlockInBraces
			reduce(13), // !=, reduce: StatementBlockInBraces
			reduce(13), // <=>, reduce: StatementBlockInBraces
			reduce(13), // >=, reduce: StatementBlockInBraces
			reduce(13), // <, reduce: StatementBlockInBraces
			reduce(13), // <=, reduce: StatementBlockInBraces
			reduce(13), // ^, reduce: StatementBlockInBraces
			reduce(13), // &, reduce: StatementBlockInBraces
			reduce(13), // <<, reduce: StatementBlockInBraces
			reduce(13), // >>>, reduce: StatementBlockInBraces
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			reduce(13), // *, reduce: StatementBlockInBraces
			reduce(13), // /, reduce: StatementBlockInBraces
			reduce(13), // //, reduce: StatementBlockInBraces
			reduce(13), // %, reduce: StatementBlockInBraces
			reduce(13), // .*, reduce: StatementBlockInBraces
			reduce(13), // ./, reduce: StatementBlockInBraces
			reduce(13), // .//, reduce: StatementBlockInBraces
			reduce(13), // ., reduce: StatementBlockInBraces
			nil,        // !
			nil,        // ~
			reduce(13), // **, reduce: StatementBlockInBraces
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(13), // [, reduce: StatementBlockInBraces
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(13), // [[, reduce: StatementBlockInBraces
			reduce(13), // [[[, reduce: StatementBlockInBraces
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4669
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(325), // ;, reduce: UnnamedFunctionDefinition
			nil,         // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			reduce(325), // >, reduce: UnnamedFunctionDefinition
			reduce(325), // >>, reduce: UnnamedFunctionDefinition
			reduce(325), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(325), // ,, reduce: UnnamedFunctionDefinition
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			reduce(325), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(325), // ||, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^, reduce: UnnamedFunctionDefinition
			reduce(325), // &&, reduce: UnnamedFunctionDefinition
			reduce(325), // ??, reduce: UnnamedFunctionDefinition
			reduce(325), // ???, reduce: UnnamedFunctionDefinition
			reduce(325), // =~, reduce: UnnamedFunctionDefinition
			reduce(325), // !=~, reduce: UnnamedFunctionDefinition
			reduce(325), // ==, reduce: UnnamedFunctionDefinition
			reduce(325), // !=, reduce: UnnamedFunctionDefinition
			reduce(325), // <=>, reduce: UnnamedFunctionDefinition
			reduce(325), // >=, reduce: UnnamedFunctionDefinition
			reduce(325), // <, reduce: UnnamedFunctionDefinition
			reduce(325), // <=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^, reduce: UnnamedFunctionDefinition
			reduce(325), // &, reduce: UnnamedFunctionDefinition
			reduce(325), // <<, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>, reduce: UnnamedFunctionDefinition
			reduce(325), // +, reduce: UnnamedFunctionDefinition
			reduce(325), // -, reduce: UnnamedFunctionDefinition
			reduce(325), // .+, reduce: UnnamedFunctionDefinition
			reduce(325), // .-, reduce: UnnamedFunctionDefinition
			reduce(325), // *, reduce: UnnamedFunctionDefinition
			reduce(325), // /, reduce: UnnamedFunctionDefinition
			reduce(325), // //, reduce: UnnamedFunctionDefinition
			reduce(325), // %, reduce: UnnamedFunctionDefinition
			reduce(325), // .*, reduce: UnnamedFunctionDefinition
			reduce(325), // ./, reduce: UnnamedFunctionDefinition
			reduce(325), // .//, reduce: UnnamedFunctionDefinition
			reduce(325), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(325), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(325), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(325), // [[, reduce: UnnamedFunctionDefinition
			reduce(325), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4670
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			reduce(305), // $, reduce: ForLoopMultivariable
			nil,         // empty
			reduce(305), // ;, reduce: ForLoopMultivariable
			reduce(305), // {, reduce: ForLoopMultivariable
			nil,         // }
			nil,         // =
			reduce(305), // unset, reduce: ForLoopMultivariable
			reduce(305), // filter, reduce: ForLoopMultivariable
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(305), // print, reduce: ForLoopMultivariable
			nil,         // ,
			reduce(305), // printn, reduce: ForLoopMultivariable
			reduce(305), // eprint, reduce: ForLoopMultivariable
			reduce(305), // eprintn, reduce: ForLoopMultivariable
			reduce(305), // dump, reduce: ForLoopMultivariable
			reduce(305), // edump, reduce: ForLoopMultivariable
			reduce(305), // tee, reduce: ForLoopMultivariable
			reduce(305), // emitf, reduce: ForLoopMultivariable
			reduce(305), // emit1, reduce: ForLoopMultivariable
			reduce(305), // emit, reduce: ForLoopMultivariable
			reduce(305), // (, reduce: ForLoopMultivariable
			nil,         // )
			reduce(305), // emitp, reduce: ForLoopMultivariable
			reduce(305), // field_name, reduce: ForLoopMultivariable
			reduce(305), // $[, reduce: ForLoopMultivariable
			nil,         // ]
			reduce(305), // braced_field_name, reduce: ForLoopMultivariable
			reduce(305), // $[[, reduce: ForLoopMultivariable
			reduce(305), // $[[[, reduce: ForLoopMultivariable
			reduce(305), // full_srec, reduce: ForLoopMultivariable
			reduce(305), // oosvar_name, reduce: ForLoopMultivariable
			reduce(305), // @[, reduce: ForLoopMultivariable
			reduce(305), // braced_oosvar_name, reduce: ForLoopMultivariable
			reduce(305), // full_oosvar, reduce: ForLoopMultivariable
			reduce(305), // all, reduce: ForLoopMultivariable
			reduce(305), // non_sigil_name, reduce: ForLoopMultivariable
			reduce(305), // arr, reduce: ForLoopMultivariable
			reduce(305), // bool, reduce: ForLoopMultivariable
			reduce(305), // float, reduce: ForLoopMultivariable
			reduce(305), // int, reduce: ForLoopMultivariable
			reduce(305), // map, reduce: ForLoopMultivariable
			reduce(305), // num, reduce: ForLoopMultivariable
			reduce(305), // str, reduce: ForLoopMultivariable
			reduce(305), // var, reduce: ForLoopMultivariable
			reduce(305), // funct, reduce: ForLoopMultivariable
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(305), // +, reduce: ForLoopMultivariable
			reduce(305), // -, reduce: ForLoopMultivariable
			reduce(305), // .+, reduce: ForLoopMultivariable
			reduce(305), // .-, reduce: ForLoopMultivariable
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(305), // !, reduce: ForLoopMultivariable
			reduce(305), // ~, reduce: ForLoopMultivariable
			nil,         // **
			reduce(305), // string_literal, reduce: ForLoopMultivariable
			reduce(305), // regex_case_insensitive, reduce: ForLoopMultivariable
			reduce(305), // int_literal, reduce: ForLoopMultivariable
			reduce(305), // float_literal, reduce: ForLoopMultivariable
			reduce(305), // boolean_literal, reduce: ForLoopMultivariable
			reduce(305), // null_literal, reduce: ForLoopMultivariable
			reduce(305), // inf_literal, reduce: ForLoopMultivariable
			reduce(305), // nan_literal, reduce: ForLoopMultivariable
			reduce(305), // const_M_PI, reduce: ForLoopMultivariable
			reduce(305), // const_M_E, reduce: ForLoopMultivariable
			reduce(305), // panic, reduce: ForLoopMultivariable
			reduce(305), // [, reduce: ForLoopMultivariable
			reduce(305), // ctx_IPS, reduce: ForLoopMultivariable
			reduce(305), // ctx_IFS, reduce: ForLoopMultivariable
			reduce(305), // ctx_IRS, reduce: ForLoopMultivariable
			reduce(305), // ctx_OPS, reduce: ForLoopMultivariable
			reduce(305), // ctx_OFS, reduce: ForLoopMultivariable
			reduce(305), // ctx_ORS, reduce: ForLoopMultivariable
			reduce(305), // ctx_FLATSEP, reduce: ForLoopMultivariable
			reduce(305), // ctx_NF, reduce: ForLoopMultivariable
			reduce(305), // ctx_NR, reduce: ForLoopMultivariable
			reduce(305), // ctx_FNR, reduce: ForLoopMultivariable
			reduce(305), // ctx_FILENAME, reduce: ForLoopMultivariable
			reduce(305), // ctx_FILENUM, reduce: ForLoopMultivariable
			reduce(305), // env, reduce: ForLoopMultivariable
			nil,         // [[
			nil,         // [[[
			reduce(305), // call, reduce: ForLoopMultivariable
			reduce(305), // begin, reduce: ForLoopMultivariable
			reduce(305), // end, reduce: ForLoopMultivariable
			reduce(305), // if, reduce: ForLoopMultivariable
			nil,         // elif
			nil,         // else
			reduce(305), // while, reduce: ForLoopMultivariable
			reduce(305), // do, reduce: ForLoopMultivariable
			reduce(305), // for, reduce: ForLoopMultivariable
			nil,         // in
			reduce(305), // break, reduce: ForLoopMultivariable
			reduce(305), // continue, reduce: ForLoopMultivariable
			reduce(305), // func, reduce: ForLoopMultivariable
			reduce(305), // subr, reduce: ForLoopMultivariable
			reduce(305), // return, reduce: ForLoopMultivariable
		},
	},
	actionRow{ // S4671
		canRecover: false,
		actions: [numSymbols]action{
			nil,        // INVALID
			nil,        // $
			nil,        // empty
			reduce(13), // ;, reduce: StatementBlockInBraces
			nil,        // {
			nil,        // }
			reduce(13), // =, reduce: StatementBlockInBraces
			nil,        // unset
			nil,        // filter
			reduce(13), // >, reduce: StatementBlockInBraces
			reduce(13), // >>, reduce: StatementBlockInBraces
			reduce(13), // |, reduce: StatementBlockInBraces
			nil,        // stdout
			nil,        // stderr
			nil,        // print
			reduce(13), // ,, reduce: StatementBlockInBraces
			nil,        // printn
			nil,        // eprint
			nil,        // eprintn
			nil,        // dump
			nil,        // edump
			nil,        // tee
			nil,        // emitf
			nil,        // emit1
			nil,        // emit
			nil,        // (
			nil,        // )
			nil,        // emitp
			nil,        // field_name
			nil,        // $[
			nil,        // ]
			nil,        // braced_field_name
			nil,        // $[[
			nil,        // $[[[
			nil,        // full_srec
			nil,        // oosvar_name
			nil,        // @[
			nil,        // braced_oosvar_name
			nil,        // full_oosvar
			nil,        // all
			nil,        // non_sigil_name
			nil,        // arr
			nil,        // bool
			nil,        // float
			nil,        // int
			nil,        // map
			nil,        // num
			nil,        // str
			nil,        // var
			nil,        // funct
			reduce(13), // ||=, reduce: StatementBlockInBraces
			reduce(13), // ^^=, reduce: StatementBlockInBraces
			reduce(13), // &&=, reduce: StatementBlockInBraces
			reduce(13), // ??=, reduce: StatementBlockInBraces
			reduce(13), // ???=, reduce: StatementBlockInBraces
			reduce(13), // |=, reduce: StatementBlockInBraces
			reduce(13), // &=, reduce: StatementBlockInBraces
			reduce(13), // ^=, reduce: StatementBlockInBraces
			reduce(13), // <<=, reduce: StatementBlockInBraces
			reduce(13), // >>=, reduce: StatementBlockInBraces
			reduce(13), // >>>=, reduce: StatementBlockInBraces
			reduce(13), // +=, reduce: StatementBlockInBraces
			reduce(13), // .=, reduce: StatementBlockInBraces
			reduce(13), // -=, reduce: StatementBlockInBraces
			reduce(13), // *=, reduce: StatementBlockInBraces
			reduce(13), // /=, reduce: StatementBlockInBraces
			reduce(13), // //=, reduce: StatementBlockInBraces
			reduce(13), // %=, reduce: StatementBlockInBraces
			reduce(13), // **=, reduce: StatementBlockInBraces
			reduce(13), // ?, reduce: StatementBlockInBraces
			nil,        // :
			reduce(13), // ||, reduce: StatementBlockInBraces
			reduce(13), // ^^, reduce: StatementBlockInBraces
			reduce(13), // &&, reduce: StatementBlockInBraces
			reduce(13), // ??, reduce: StatementBlockInBraces
			reduce(13), // ???, reduce: StatementBlockInBraces
			reduce(13), // =~, reduce: StatementBlockInBraces
			reduce(13), // !=~, reduce: StatementBlockInBraces
			reduce(13), // ==, reduce: StatementBlockInBraces
			reduce(13), // !=, reduce: StatementBlockInBraces
			reduce(13), // <=>, reduce: StatementBlockInBraces
			reduce(13), // >=, reduce: StatementBlockInBraces
			reduce(13), // <, reduce: StatementBlockInBraces
			reduce(13), // <=, reduce: StatementBlockInBraces
			reduce(13), // ^, reduce: StatementBlockInBraces
			reduce(13), // &, reduce: StatementBlockInBraces
			reduce(13), // <<, reduce: StatementBlockInBraces
			reduce(13), // >>>, reduce: StatementBlockInBraces
			reduce(13), // +, reduce: StatementBlockInBraces
			reduce(13), // -, reduce: StatementBlockInBraces
			reduce(13), // .+, reduce: StatementBlockInBraces
			reduce(13), // .-, reduce: StatementBlockInBraces
			reduce(13), // *, reduce: StatementBlockInBraces
			reduce(13), // /, reduce: StatementBlockInBraces
			reduce(13), // //, reduce: StatementBlockInBraces
			reduce(13), // %, reduce: StatementBlockInBraces
			reduce(13), // .*, reduce: StatementBlockInBraces
			reduce(13), // ./, reduce: StatementBlockInBraces
			reduce(13), // .//, reduce: StatementBlockInBraces
			reduce(13), // ., reduce: StatementBlockInBraces
			nil,        // !
			nil,        // ~
			reduce(13), // **, reduce: StatementBlockInBraces
			nil,        // string_literal
			nil,        // regex_case_insensitive
			nil,        // int_literal
			nil,        // float_literal
			nil,        // boolean_literal
			nil,        // null_literal
			nil,        // inf_literal
			nil,        // nan_literal
			nil,        // const_M_PI
			nil,        // const_M_E
			nil,        // panic
			reduce(13), // [, reduce: StatementBlockInBraces
			nil,        // ctx_IPS
			nil,        // ctx_IFS
			nil,        // ctx_IRS
			nil,        // ctx_OPS
			nil,        // ctx_OFS
			nil,        // ctx_ORS
			nil,        // ctx_FLATSEP
			nil,        // ctx_NF
			nil,        // ctx_NR
			nil,        // ctx_FNR
			nil,        // ctx_FILENAME
			nil,        // ctx_FILENUM
			nil,        // env
			reduce(13), // [[, reduce: StatementBlockInBraces
			reduce(13), // [[[, reduce: StatementBlockInBraces
			nil,        // call
			nil,        // begin
			nil,        // end
			nil,        // if
			nil,        // elif
			nil,        // else
			nil,        // while
			nil,        // do
			nil,        // for
			nil,        // in
			nil,        // break
			nil,        // continue
			nil,        // func
			nil,        // subr
			nil,        // return
		},
	},
	actionRow{ // S4672
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(325), // ;, reduce: UnnamedFunctionDefinition
			nil,         // {
			nil,         // }
			reduce(325), // =, reduce: UnnamedFunctionDefinition
			nil,         // unset
			nil,         // filter
			reduce(325), // >, reduce: UnnamedFunctionDefinition
			reduce(325), // >>, reduce: UnnamedFunctionDefinition
			reduce(325), // |, reduce: UnnamedFunctionDefinition
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			reduce(325), // ,, reduce: UnnamedFunctionDefinition
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			reduce(325), // ||=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^=, reduce: UnnamedFunctionDefinition
			reduce(325), // &&=, reduce: UnnamedFunctionDefinition
			reduce(325), // ??=, reduce: UnnamedFunctionDefinition
			reduce(325), // ???=, reduce: UnnamedFunctionDefinition
			reduce(325), // |=, reduce: UnnamedFunctionDefinition
			reduce(325), // &=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^=, reduce: UnnamedFunctionDefinition
			reduce(325), // <<=, reduce: UnnamedFunctionDefinition
			reduce(325), // >>=, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>=, reduce: UnnamedFunctionDefinition
			reduce(325), // +=, reduce: UnnamedFunctionDefinition
			reduce(325), // .=, reduce: UnnamedFunctionDefinition
			reduce(325), // -=, reduce: UnnamedFunctionDefinition
			reduce(325), // *=, reduce: UnnamedFunctionDefinition
			reduce(325), // /=, reduce: UnnamedFunctionDefinition
			reduce(325), // //=, reduce: UnnamedFunctionDefinition
			reduce(325), // %=, reduce: UnnamedFunctionDefinition
			reduce(325), // **=, reduce: UnnamedFunctionDefinition
			reduce(325), // ?, reduce: UnnamedFunctionDefinition
			nil,         // :
			reduce(325), // ||, reduce: UnnamedFunctionDefinition
			reduce(325), // ^^, reduce: UnnamedFunctionDefinition
			reduce(325), // &&, reduce: UnnamedFunctionDefinition
			reduce(325), // ??, reduce: UnnamedFunctionDefinition
			reduce(325), // ???, reduce: UnnamedFunctionDefinition
			reduce(325), // =~, reduce: UnnamedFunctionDefinition
			reduce(325), // !=~, reduce: UnnamedFunctionDefinition
			reduce(325), // ==, reduce: UnnamedFunctionDefinition
			reduce(325), // !=, reduce: UnnamedFunctionDefinition
			reduce(325), // <=>, reduce: UnnamedFunctionDefinition
			reduce(325), // >=, reduce: UnnamedFunctionDefinition
			reduce(325), // <, reduce: UnnamedFunctionDefinition
			reduce(325), // <=, reduce: UnnamedFunctionDefinition
			reduce(325), // ^, reduce: UnnamedFunctionDefinition
			reduce(325), // &, reduce: UnnamedFunctionDefinition
			reduce(325), // <<, reduce: UnnamedFunctionDefinition
			reduce(325), // >>>, reduce: UnnamedFunctionDefinition
			reduce(325), // +, reduce: UnnamedFunctionDefinition
			reduce(325), // -, reduce: UnnamedFunctionDefinition
			reduce(325), // .+, reduce: UnnamedFunctionDefinition
			reduce(325), // .-, reduce: UnnamedFunctionDefinition
			reduce(325), // *, reduce: UnnamedFunctionDefinition
			reduce(325), // /, reduce: UnnamedFunctionDefinition
			reduce(325), // //, reduce: UnnamedFunctionDefinition
			reduce(325), // %, reduce: UnnamedFunctionDefinition
			reduce(325), // .*, reduce: UnnamedFunctionDefinition
			reduce(325), // ./, reduce: UnnamedFunctionDefinition
			reduce(325), // .//, reduce: UnnamedFunctionDefinition
			reduce(325), // ., reduce: UnnamedFunctionDefinition
			nil,         // !
			nil,         // ~
			reduce(325), // **, reduce: UnnamedFunctionDefinition
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			reduce(325), // [, reduce: UnnamedFunctionDefinition
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			reduce(325), // [[, reduce: UnnamedFunctionDefinition
			reduce(325), // [[[, reduce: UnnamedFunctionDefinition
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4673
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			nil,         // ;
			shift(2042), // {
			nil,         // }
			nil,         // =
			nil,         // unset
			nil,         // filter
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			nil,         // print
			nil,         // ,
			nil,         // printn
			nil,         // eprint
			nil,         // eprintn
			nil,         // dump
			nil,         // edump
			nil,         // tee
			nil,         // emitf
			nil,         // emit1
			nil,         // emit
			nil,         // (
			nil,         // )
			nil,         // emitp
			nil,         // field_name
			nil,         // $[
			nil,         // ]
			nil,         // braced_field_name
			nil,         // $[[
			nil,         // $[[[
			nil,         // full_srec
			nil,         // oosvar_name
			nil,         // @[
			nil,         // braced_oosvar_name
			nil,         // full_oosvar
			nil,         // all
			nil,         // non_sigil_name
			nil,         // arr
			nil,         // bool
			nil,         // float
			nil,         // int
			nil,         // map
			nil,         // num
			nil,         // str
			nil,         // var
			nil,         // funct
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			nil,         // +
			nil,         // -
			nil,         // .+
			nil,         // .-
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			nil,         // !
			nil,         // ~
			nil,         // **
			nil,         // string_literal
			nil,         // regex_case_insensitive
			nil,         // int_literal
			nil,         // float_literal
			nil,         // boolean_literal
			nil,         // null_literal
			nil,         // inf_literal
			nil,         // nan_literal
			nil,         // const_M_PI
			nil,         // const_M_E
			nil,         // panic
			nil,         // [
			nil,         // ctx_IPS
			nil,         // ctx_IFS
			nil,         // ctx_IRS
			nil,         // ctx_OPS
			nil,         // ctx_OFS
			nil,         // ctx_ORS
			nil,         // ctx_FLATSEP
			nil,         // ctx_NF
			nil,         // ctx_NR
			nil,         // ctx_FNR
			nil,         // ctx_FILENAME
			nil,         // ctx_FILENUM
			nil,         // env
			nil,         // [[
			nil,         // [[[
			nil,         // call
			nil,         // begin
			nil,         // end
			nil,         // if
			nil,         // elif
			nil,         // else
			nil,         // while
			nil,         // do
			nil,         // for
			nil,         // in
			nil,         // break
			nil,         // continue
			nil,         // func
			nil,         // subr
			nil,         // return
		},
	},
	actionRow{ // S4674
		canRecover: false,
		actions: [numSymbols]action{
			nil,         // INVALID
			nil,         // $
			nil,         // empty
			reduce(305), // ;, reduce: ForLoopMultivariable
			reduce(305), // {, reduce: ForLoopMultivariable
			reduce(305), // }, reduce: ForLoopMultivariable
			nil,         // =
			reduce(305), // unset, reduce: ForLoopMultivariable
			reduce(305), // filter, reduce: ForLoopMultivariable
			nil,         // >
			nil,         // >>
			nil,         // |
			nil,         // stdout
			nil,         // stderr
			reduce(305), // print, reduce: ForLoopMultivariable
			nil,         // ,
			reduce(305), // printn, reduce: ForLoopMultivariable
			reduce(305), // eprint, reduce: ForLoopMultivariable
			reduce(305), // eprintn, reduce: ForLoopMultivariable
			reduce(305), // dump, reduce: ForLoopMultivariable
			reduce(305), // edump, reduce: ForLoopMultivariable
			reduce(305), // tee, reduce: ForLoopMultivariable
			reduce(305), // emitf, reduce: ForLoopMultivariable
			reduce(305), // emit1, reduce: ForLoopMultivariable
			reduce(305), // emit, reduce: ForLoopMultivariable
			reduce(305), // (, reduce: ForLoopMultivariable
			nil,         // )
			reduce(305), // emitp, reduce: ForLoopMultivariable
			reduce(305), // field_name, reduce: ForLoopMultivariable
			reduce(305), // $[, reduce: ForLoopMultivariable
			nil,         // ]
			reduce(305), // braced_field_name, reduce: ForLoopMultivariable
			reduce(305), // $[[, reduce: ForLoopMultivariable
			reduce(305), // $[[[, reduce: ForLoopMultivariable
			reduce(305), // full_srec, reduce: ForLoopMultivariable
			reduce(305), // oosvar_name, reduce: ForLoopMultivariable
			reduce(305), // @[, reduce: ForLoopMultivariable
			reduce(305), // braced_oosvar_name, reduce: ForLoopMultivariable
			reduce(305), // full_oosvar, reduce: ForLoopMultivariable
			reduce(305), // all, reduce: ForLoopMultivariable
			reduce(305), // non_sigil_name, reduce: ForLoopMultivariable
			reduce(305), // arr, reduce: ForLoopMultivariable
			reduce(305), // bool, reduce: ForLoopMultivariable
			reduce(305), // float, reduce: ForLoopMultivariable
			reduce(305), // int, reduce: ForLoopMultivariable
			reduce(305), // map, reduce: ForLoopMultivariable
			reduce(305), // num, reduce: ForLoopMultivariable
			reduce(305), // str, reduce: ForLoopMultivariable
			reduce(305), // var, reduce: ForLoopMultivariable
			reduce(305), // funct, reduce: ForLoopMultivariable
			nil,         // ||=
			nil,         // ^^=
			nil,         // &&=
			nil,         // ??=
			nil,         // ???=
			nil,         // |=
			nil,         // &=
			nil,         // ^=
			nil,         // <<=
			nil,         // >>=
			nil,         // >>>=
			nil,         // +=
			nil,         // .=
			nil,         // -=
			nil,         // *=
			nil,         // /=
			nil,         // //=
			nil,         // %=
			nil,         // **=
			nil,         // ?
			nil,         // :
			nil,         // ||
			nil,         // ^^
			nil,         // &&
			nil,         // ??
			nil,         // ???
			nil,         // =~
			nil,         // !=~
			nil,         // ==
			nil,         // !=
			nil,         // <=>
			nil,         // >=
			nil,         // <
			nil,         // <=
			nil,         // ^
			nil,         // &
			nil,         // <<
			nil,         // >>>
			reduce(305), // +, reduce: ForLoopMultivariable
			reduce(305), // -, reduce: ForLoopMultivariable
			reduce(305), // .+, reduce: ForLoopMultivariable
			reduce(305), // .-, reduce: ForLoopMultivariable
			nil,         // *
			nil,         // /
			nil,         // //
			nil,         // %
			nil,         // .*
			nil,         // ./
			nil,         // .//
			nil,         // .
			reduce(305), // !, reduce: ForLoopMultivariable
			reduce(305), // ~, reduce: ForLoopMultivariable
			nil,         // **
			reduce(305), // string_literal, reduce: ForLoopMultivariable
			reduce(305), // regex_case_insensitive, reduce: ForLoopMultivariable
			reduce(305), // int_literal, reduce: ForLoopMultivariable
			reduce(305), // float_literal, reduce: ForLoopMultivariable
			reduce(305), // boolean_literal, reduce: ForLoopMultivariable
			reduce(305), // null_literal, reduce: ForLoopMultivariable
			reduce(305), // inf_literal, reduce: ForLoopMultivariable
			reduce(305), // nan_literal, reduce: ForLoopMultivariable
			reduce(305), // const_M_PI, reduce: ForLoopMultivariable
			reduce(305), // const_M_E, reduce: ForLoopMultivariable
			reduce(305), // panic, reduce: ForLoopMultivariable
			reduce(305), // [, reduce: ForLoopMultivariable
			reduce(305), // ctx_IPS, reduce: ForLoopMultivariable
			reduce(305), // ctx_IFS, reduce: ForLoopMultivariable
			reduce(305), // ctx_IRS, reduce: ForLoopMultivariable
			reduce(305), // ctx_OPS, reduce: ForLoopMultivariable
			reduce(305), // ctx_OFS, reduce: ForLoopMultivariable
			reduce(305), // ctx_ORS, reduce: ForLoopMultivariable
			reduce(305), // ctx_FLATSEP, reduce: ForLoopMultivariable
			reduce(305), // ctx_NF, reduce: ForLoopMultivariable
			reduce(305), // ctx_NR, reduce: ForLoopMultivariable
			reduce(305), // ctx_FNR, reduce: ForLoopMultivariable
			reduce(305), // ctx_FILENAME, reduce: ForLoopMultivariable
			reduce(305), // ctx_FILENUM, reduce: ForLoopMultivariable
			reduce(305), // env, reduce: ForLoopMultivariable
			nil,         // [[
			nil,         // [[[
			reduce(305), // call, reduce: ForLoopMultivariable
			reduce(305), // begin, reduce: ForLoopMultivariable
			reduce(305), // end, reduce: ForLoopMultivariable
			reduce(305), // if, reduce: ForLoopMultivariable
			nil,         // elif
			nil,         // else
			reduce(305), // while, reduce: ForLoopMultivariable
			reduce(305), // do, reduce: ForLoopMultivariable
			reduce(305), // for, reduce: ForLoopMultivariable
			nil,         // in
			reduce(305), // break, reduce: ForLoopMultivariable
			reduce(305), // continue, reduce: ForLoopMultivariable
			reduce(305), // func, reduce: ForLoopMultivariable
			reduce(305), // subr, reduce: ForLoopMultivariable
			reduce(305), // return, reduce: ForLoopMultivariable
		},
	},
}
